/*
MIT License

Copyright (c) 2018 LiteSpeed Technologies Inc

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

#include <assert.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <sys/queue.h>

#include <lsr/xxhash.h>

#include "lshpack.h"

#define HPACK_STATIC_TABLE_SIZE   61
#define INITIAL_DYNAMIC_TABLE_SIZE  4096

/* RFC 7541, Section 4.1:
 *
 * " The size of the dynamic table is the sum of the size of its entries.
 * "
 * " The size of an entry is the sum of its name's length in octets (as
 * " defined in Section 5.2), its value's length in octets, and 32.
 */
#define DYNAMIC_ENTRY_OVERHEAD 32

#define NAME_VAL(a, b) sizeof(a) - 1, sizeof(b) - 1, (a), (b)

static const struct
{
    unsigned          name_len;
    unsigned          val_len;
    const char       *name;
    const char       *val;
}
static_table[HPACK_STATIC_TABLE_SIZE] =
{
    { NAME_VAL(":authority",                    "") },
    { NAME_VAL(":method",                       "GET") },
    { NAME_VAL(":method",                       "POST") },
    { NAME_VAL(":path",                         "/") },
    { NAME_VAL(":path",                         "/index.html") },
    { NAME_VAL(":scheme",                       "http") },
    { NAME_VAL(":scheme",                       "https") },
    { NAME_VAL(":status",                       "200") },
    { NAME_VAL(":status",                       "204") },
    { NAME_VAL(":status",                       "206") },
    { NAME_VAL(":status",                       "304") },
    { NAME_VAL(":status",                       "400") },
    { NAME_VAL(":status",                       "404") },
    { NAME_VAL(":status",                       "500") },
    { NAME_VAL("accept-charset",                "") },
    { NAME_VAL("accept-encoding",               "gzip, deflate") },
    { NAME_VAL("accept-language",               "") },
    { NAME_VAL("accept-ranges",                 "") },
    { NAME_VAL("accept",                        "") },
    { NAME_VAL("access-control-allow-origin",   "") },
    { NAME_VAL("age",                           "") },
    { NAME_VAL("allow",                         "") },
    { NAME_VAL("authorization",                 "") },
    { NAME_VAL("cache-control",                 "") },
    { NAME_VAL("content-disposition",           "") },
    { NAME_VAL("content-encoding",              "") },
    { NAME_VAL("content-language",              "") },
    { NAME_VAL("content-length",                "") },
    { NAME_VAL("content-location",              "") },
    { NAME_VAL("content-range",                 "") },
    { NAME_VAL("content-type",                  "") },
    { NAME_VAL("cookie",                        "") },
    { NAME_VAL("date",                          "") },
    { NAME_VAL("etag",                          "") },
    { NAME_VAL("expect",                        "") },
    { NAME_VAL("expires",                       "") },
    { NAME_VAL("from",                          "") },
    { NAME_VAL("host",                          "") },
    { NAME_VAL("if-match",                      "") },
    { NAME_VAL("if-modified-since",             "") },
    { NAME_VAL("if-none-match",                 "") },
    { NAME_VAL("if-range",                      "") },
    { NAME_VAL("if-unmodified-since",           "") },
    { NAME_VAL("last-modified",                 "") },
    { NAME_VAL("link",                          "") },
    { NAME_VAL("location",                      "") },
    { NAME_VAL("max-forwards",                  "") },
    { NAME_VAL("proxy-authenticate",            "") },
    { NAME_VAL("proxy-authorization",           "") },
    { NAME_VAL("range",                         "") },
    { NAME_VAL("referer",                       "") },
    { NAME_VAL("refresh",                       "") },
    { NAME_VAL("retry-after",                   "") },
    { NAME_VAL("server",                        "") },
    { NAME_VAL("set-cookie",                    "") },
    { NAME_VAL("strict-transport-security",     "") },
    { NAME_VAL("transfer-encoding",             "") },
    { NAME_VAL("user-agent",                    "") },
    { NAME_VAL("vary",                          "") },
    { NAME_VAL("via",                           "") },
    { NAME_VAL("www-authenticate",              "") }
};


struct encode_el
{
    uint32_t code;
    int      bits;
};

static const struct encode_el encode_table[257] =
{
    {     0x1ff8,    13},    //        (  0)
    {   0x7fffd8,    23},    //        (  1)
    {  0xfffffe2,    28},    //        (  2)
    {  0xfffffe3,    28},    //        (  3)
    {  0xfffffe4,    28},    //        (  4)
    {  0xfffffe5,    28},    //        (  5)
    {  0xfffffe6,    28},    //        (  6)
    {  0xfffffe7,    28},    //        (  7)
    {  0xfffffe8,    28},    //        (  8)
    {   0xffffea,    24},    //        (  9)
    { 0x3ffffffc,    30},    //        ( 10)
    {  0xfffffe9,    28},    //        ( 11)
    {  0xfffffea,    28},    //        ( 12)
    { 0x3ffffffd,    30},    //        ( 13)
    {  0xfffffeb,    28},    //        ( 14)
    {  0xfffffec,    28},    //        ( 15)
    {  0xfffffed,    28},    //        ( 16)
    {  0xfffffee,    28},    //        ( 17)
    {  0xfffffef,    28},    //        ( 18)
    {  0xffffff0,    28},    //        ( 19)
    {  0xffffff1,    28},    //        ( 20)
    {  0xffffff2,    28},    //        ( 21)
    { 0x3ffffffe,    30},    //        ( 22)
    {  0xffffff3,    28},    //        ( 23)
    {  0xffffff4,    28},    //        ( 24)
    {  0xffffff5,    28},    //        ( 25)
    {  0xffffff6,    28},    //        ( 26)
    {  0xffffff7,    28},    //        ( 27)
    {  0xffffff8,    28},    //        ( 28)
    {  0xffffff9,    28},    //        ( 29)
    {  0xffffffa,    28},    //        ( 30)
    {  0xffffffb,    28},    //        ( 31)
    {       0x14,     6},    //    ' ' ( 32)
    {      0x3f8,    10},    //    '!' ( 33)
    {      0x3f9,    10},    //    '"' ( 34)
    {      0xffa,    12},    //    '#' ( 35)
    {     0x1ff9,    13},    //    '$' ( 36)
    {       0x15,     6},    //    '%' ( 37)
    {       0xf8,     8},    //    '&' ( 38)
    {      0x7fa,    11},    //    ''' ( 39)
    {      0x3fa,    10},    //    '(' ( 40)
    {      0x3fb,    10},    //    ')' ( 41)
    {       0xf9,     8},    //    '*' ( 42)
    {      0x7fb,    11},    //    '+' ( 43)
    {       0xfa,     8},    //    ',' ( 44)
    {       0x16,     6},    //    '-' ( 45)
    {       0x17,     6},    //    '.' ( 46)
    {       0x18,     6},    //    '/' ( 47)
    {        0x0,     5},    //    '0' ( 48)
    {        0x1,     5},    //    '1' ( 49)
    {        0x2,     5},    //    '2' ( 50)
    {       0x19,     6},    //    '3' ( 51)
    {       0x1a,     6},    //    '4' ( 52)
    {       0x1b,     6},    //    '5' ( 53)
    {       0x1c,     6},    //    '6' ( 54)
    {       0x1d,     6},    //    '7' ( 55)
    {       0x1e,     6},    //    '8' ( 56)
    {       0x1f,     6},    //    '9' ( 57)
    {       0x5c,     7},    //    ':' ( 58)
    {       0xfb,     8},    //    ';' ( 59)
    {     0x7ffc,    15},    //    '<' ( 60)
    {       0x20,     6},    //    '=' ( 61)
    {      0xffb,    12},    //    '>' ( 62)
    {      0x3fc,    10},    //    '?' ( 63)
    {     0x1ffa,    13},    //    '@' ( 64)
    {       0x21,     6},    //    'A' ( 65)
    {       0x5d,     7},    //    'B' ( 66)
    {       0x5e,     7},    //    'C' ( 67)
    {       0x5f,     7},    //    'D' ( 68)
    {       0x60,     7},    //    'E' ( 69)
    {       0x61,     7},    //    'F' ( 70)
    {       0x62,     7},    //    'G' ( 71)
    {       0x63,     7},    //    'H' ( 72)
    {       0x64,     7},    //    'I' ( 73)
    {       0x65,     7},    //    'J' ( 74)
    {       0x66,     7},    //    'K' ( 75)
    {       0x67,     7},    //    'L' ( 76)
    {       0x68,     7},    //    'M' ( 77)
    {       0x69,     7},    //    'N' ( 78)
    {       0x6a,     7},    //    'O' ( 79)
    {       0x6b,     7},    //    'P' ( 80)
    {       0x6c,     7},    //    'Q' ( 81)
    {       0x6d,     7},    //    'R' ( 82)
    {       0x6e,     7},    //    'S' ( 83)
    {       0x6f,     7},    //    'T' ( 84)
    {       0x70,     7},    //    'U' ( 85)
    {       0x71,     7},    //    'V' ( 86)
    {       0x72,     7},    //    'W' ( 87)
    {       0xfc,     8},    //    'X' ( 88)
    {       0x73,     7},    //    'Y' ( 89)
    {       0xfd,     8},    //    'Z' ( 90)
    {     0x1ffb,    13},    //    '[' ( 91)
    {    0x7fff0,    19},    //    '\' ( 92)
    {     0x1ffc,    13},    //    ']' ( 93)
    {     0x3ffc,    14},    //    '^' ( 94)
    {       0x22,     6},    //    '_' ( 95)
    {     0x7ffd,    15},    //    '`' ( 96)
    {        0x3,     5},    //    'a' ( 97)
    {       0x23,     6},    //    'b' ( 98)
    {        0x4,     5},    //    'c' ( 99)
    {       0x24,     6},    //    'd' (100)
    {        0x5,     5},    //    'e' (101)
    {       0x25,     6},    //    'f' (102)
    {       0x26,     6},    //    'g' (103)
    {       0x27,     6},    //    'h' (104)
    {        0x6,     5},    //    'i' (105)
    {       0x74,     7},    //    'j' (106)
    {       0x75,     7},    //    'k' (107)
    {       0x28,     6},    //    'l' (108)
    {       0x29,     6},    //    'm' (109)
    {       0x2a,     6},    //    'n' (110)
    {        0x7,     5},    //    'o' (111)
    {       0x2b,     6},    //    'p' (112)
    {       0x76,     7},    //    'q' (113)
    {       0x2c,     6},    //    'r' (114)
    {        0x8,     5},    //    's' (115)
    {        0x9,     5},    //    't' (116)
    {       0x2d,     6},    //    'u' (117)
    {       0x77,     7},    //    'v' (118)
    {       0x78,     7},    //    'w' (119)
    {       0x79,     7},    //    'x' (120)
    {       0x7a,     7},    //    'y' (121)
    {       0x7b,     7},    //    'z' (122)
    {     0x7ffe,    15},    //    '{' (123)
    {      0x7fc,    11},    //    '|' (124)
    {     0x3ffd,    14},    //    '}' (125)
    {     0x1ffd,    13},    //    '~' (126)
    {  0xffffffc,    28},    //        (127)
    {    0xfffe6,    20},    //        (128)
    {   0x3fffd2,    22},    //        (129)
    {    0xfffe7,    20},    //        (130)
    {    0xfffe8,    20},    //        (131)
    {   0x3fffd3,    22},    //        (132)
    {   0x3fffd4,    22},    //        (133)
    {   0x3fffd5,    22},    //        (134)
    {   0x7fffd9,    23},    //        (135)
    {   0x3fffd6,    22},    //        (136)
    {   0x7fffda,    23},    //        (137)
    {   0x7fffdb,    23},    //        (138)
    {   0x7fffdc,    23},    //        (139)
    {   0x7fffdd,    23},    //        (140)
    {   0x7fffde,    23},    //        (141)
    {   0xffffeb,    24},    //        (142)
    {   0x7fffdf,    23},    //        (143)
    {   0xffffec,    24},    //        (144)
    {   0xffffed,    24},    //        (145)
    {   0x3fffd7,    22},    //        (146)
    {   0x7fffe0,    23},    //        (147)
    {   0xffffee,    24},    //        (148)
    {   0x7fffe1,    23},    //        (149)
    {   0x7fffe2,    23},    //        (150)
    {   0x7fffe3,    23},    //        (151)
    {   0x7fffe4,    23},    //        (152)
    {   0x1fffdc,    21},    //        (153)
    {   0x3fffd8,    22},    //        (154)
    {   0x7fffe5,    23},    //        (155)
    {   0x3fffd9,    22},    //        (156)
    {   0x7fffe6,    23},    //        (157)
    {   0x7fffe7,    23},    //        (158)
    {   0xffffef,    24},    //        (159)
    {   0x3fffda,    22},    //        (160)
    {   0x1fffdd,    21},    //        (161)
    {    0xfffe9,    20},    //        (162)
    {   0x3fffdb,    22},    //        (163)
    {   0x3fffdc,    22},    //        (164)
    {   0x7fffe8,    23},    //        (165)
    {   0x7fffe9,    23},    //        (166)
    {   0x1fffde,    21},    //        (167)
    {   0x7fffea,    23},    //        (168)
    {   0x3fffdd,    22},    //        (169)
    {   0x3fffde,    22},    //        (170)
    {   0xfffff0,    24},    //        (171)
    {   0x1fffdf,    21},    //        (172)
    {   0x3fffdf,    22},    //        (173)
    {   0x7fffeb,    23},    //        (174)
    {   0x7fffec,    23},    //        (175)
    {   0x1fffe0,    21},    //        (176)
    {   0x1fffe1,    21},    //        (177)
    {   0x3fffe0,    22},    //        (178)
    {   0x1fffe2,    21},    //        (179)
    {   0x7fffed,    23},    //        (180)
    {   0x3fffe1,    22},    //        (181)
    {   0x7fffee,    23},    //        (182)
    {   0x7fffef,    23},    //        (183)
    {    0xfffea,    20},    //        (184)
    {   0x3fffe2,    22},    //        (185)
    {   0x3fffe3,    22},    //        (186)
    {   0x3fffe4,    22},    //        (187)
    {   0x7ffff0,    23},    //        (188)
    {   0x3fffe5,    22},    //        (189)
    {   0x3fffe6,    22},    //        (190)
    {   0x7ffff1,    23},    //        (191)
    {  0x3ffffe0,    26},    //        (192)
    {  0x3ffffe1,    26},    //        (193)
    {    0xfffeb,    20},    //        (194)
    {    0x7fff1,    19},    //        (195)
    {   0x3fffe7,    22},    //        (196)
    {   0x7ffff2,    23},    //        (197)
    {   0x3fffe8,    22},    //        (198)
    {  0x1ffffec,    25},    //        (199)
    {  0x3ffffe2,    26},    //        (200)
    {  0x3ffffe3,    26},    //        (201)
    {  0x3ffffe4,    26},    //        (202)
    {  0x7ffffde,    27},    //        (203)
    {  0x7ffffdf,    27},    //        (204)
    {  0x3ffffe5,    26},    //        (205)
    {   0xfffff1,    24},    //        (206)
    {  0x1ffffed,    25},    //        (207)
    {    0x7fff2,    19},    //        (208)
    {   0x1fffe3,    21},    //        (209)
    {  0x3ffffe6,    26},    //        (210)
    {  0x7ffffe0,    27},    //        (211)
    {  0x7ffffe1,    27},    //        (212)
    {  0x3ffffe7,    26},    //        (213)
    {  0x7ffffe2,    27},    //        (214)
    {   0xfffff2,    24},    //        (215)
    {   0x1fffe4,    21},    //        (216)
    {   0x1fffe5,    21},    //        (217)
    {  0x3ffffe8,    26},    //        (218)
    {  0x3ffffe9,    26},    //        (219)
    {  0xffffffd,    28},    //        (220)
    {  0x7ffffe3,    27},    //        (221)
    {  0x7ffffe4,    27},    //        (222)
    {  0x7ffffe5,    27},    //        (223)
    {    0xfffec,    20},    //        (224)
    {   0xfffff3,    24},    //        (225)
    {    0xfffed,    20},    //        (226)
    {   0x1fffe6,    21},    //        (227)
    {   0x3fffe9,    22},    //        (228)
    {   0x1fffe7,    21},    //        (229)
    {   0x1fffe8,    21},    //        (230)
    {   0x7ffff3,    23},    //        (231)
    {   0x3fffea,    22},    //        (232)
    {   0x3fffeb,    22},    //        (233)
    {  0x1ffffee,    25},    //        (234)
    {  0x1ffffef,    25},    //        (235)
    {   0xfffff4,    24},    //        (236)
    {   0xfffff5,    24},    //        (237)
    {  0x3ffffea,    26},    //        (238)
    {   0x7ffff4,    23},    //        (239)
    {  0x3ffffeb,    26},    //        (240)
    {  0x7ffffe6,    27},    //        (241)
    {  0x3ffffec,    26},    //        (242)
    {  0x3ffffed,    26},    //        (243)
    {  0x7ffffe7,    27},    //        (244)
    {  0x7ffffe8,    27},    //        (245)
    {  0x7ffffe9,    27},    //        (246)
    {  0x7ffffea,    27},    //        (247)
    {  0x7ffffeb,    27},    //        (248)
    {  0xffffffe,    28},    //        (249)
    {  0x7ffffec,    27},    //        (250)
    {  0x7ffffed,    27},    //        (251)
    {  0x7ffffee,    27},    //        (252)
    {  0x7ffffef,    27},    //        (253)
    {  0x7fffff0,    27},    //        (254)
    {  0x3ffffee,    26},    //        (255)
    { 0x3fffffff,    30}    //    EOS (256)
};


struct decode_el
{
    uint8_t state;
    uint8_t flags;
    uint8_t sym;
};

static const struct decode_el decode_tables[256][16] =
{
    /* 0 */
    {
        { 4, 0x00, 0 },
        { 5, 0x00, 0 },
        { 7, 0x00, 0 },
        { 8, 0x00, 0 },
        { 11, 0x00, 0 },
        { 12, 0x00, 0 },
        { 16, 0x00, 0 },
        { 19, 0x00, 0 },
        { 25, 0x00, 0 },
        { 28, 0x00, 0 },
        { 32, 0x00, 0 },
        { 35, 0x00, 0 },
        { 42, 0x00, 0 },
        { 49, 0x00, 0 },
        { 57, 0x00, 0 },
        { 64, 0x01, 0 },
    },
    /* 1 */
    {
        { 0, 0x03, 48 },
        { 0, 0x03, 49 },
        { 0, 0x03, 50 },
        { 0, 0x03, 97 },
        { 0, 0x03, 99 },
        { 0, 0x03, 101 },
        { 0, 0x03, 105 },
        { 0, 0x03, 111 },
        { 0, 0x03, 115 },
        { 0, 0x03, 116 },
        { 13, 0x00, 0 },
        { 14, 0x00, 0 },
        { 17, 0x00, 0 },
        { 18, 0x00, 0 },
        { 20, 0x00, 0 },
        { 21, 0x00, 0 },
    },
    /* 2 */
    {
        { 1, 0x02, 48 },
        { 22, 0x03, 48 },
        { 1, 0x02, 49 },
        { 22, 0x03, 49 },
        { 1, 0x02, 50 },
        { 22, 0x03, 50 },
        { 1, 0x02, 97 },
        { 22, 0x03, 97 },
        { 1, 0x02, 99 },
        { 22, 0x03, 99 },
        { 1, 0x02, 101 },
        { 22, 0x03, 101 },
        { 1, 0x02, 105 },
        { 22, 0x03, 105 },
        { 1, 0x02, 111 },
        { 22, 0x03, 111 },
    },
    /* 3 */
    {
        { 2, 0x02, 48 },
        { 9, 0x02, 48 },
        { 23, 0x02, 48 },
        { 40, 0x03, 48 },
        { 2, 0x02, 49 },
        { 9, 0x02, 49 },
        { 23, 0x02, 49 },
        { 40, 0x03, 49 },
        { 2, 0x02, 50 },
        { 9, 0x02, 50 },
        { 23, 0x02, 50 },
        { 40, 0x03, 50 },
        { 2, 0x02, 97 },
        { 9, 0x02, 97 },
        { 23, 0x02, 97 },
        { 40, 0x03, 97 },
    },
    /* 4 */
    {
        { 3, 0x02, 48 },
        { 6, 0x02, 48 },
        { 10, 0x02, 48 },
        { 15, 0x02, 48 },
        { 24, 0x02, 48 },
        { 31, 0x02, 48 },
        { 41, 0x02, 48 },
        { 56, 0x03, 48 },
        { 3, 0x02, 49 },
        { 6, 0x02, 49 },
        { 10, 0x02, 49 },
        { 15, 0x02, 49 },
        { 24, 0x02, 49 },
        { 31, 0x02, 49 },
        { 41, 0x02, 49 },
        { 56, 0x03, 49 },
    },
    /* 5 */
    {
        { 3, 0x02, 50 },
        { 6, 0x02, 50 },
        { 10, 0x02, 50 },
        { 15, 0x02, 50 },
        { 24, 0x02, 50 },
        { 31, 0x02, 50 },
        { 41, 0x02, 50 },
        { 56, 0x03, 50 },
        { 3, 0x02, 97 },
        { 6, 0x02, 97 },
        { 10, 0x02, 97 },
        { 15, 0x02, 97 },
        { 24, 0x02, 97 },
        { 31, 0x02, 97 },
        { 41, 0x02, 97 },
        { 56, 0x03, 97 },
    },
    /* 6 */
    {
        { 2, 0x02, 99 },
        { 9, 0x02, 99 },
        { 23, 0x02, 99 },
        { 40, 0x03, 99 },
        { 2, 0x02, 101 },
        { 9, 0x02, 101 },
        { 23, 0x02, 101 },
        { 40, 0x03, 101 },
        { 2, 0x02, 105 },
        { 9, 0x02, 105 },
        { 23, 0x02, 105 },
        { 40, 0x03, 105 },
        { 2, 0x02, 111 },
        { 9, 0x02, 111 },
        { 23, 0x02, 111 },
        { 40, 0x03, 111 },
    },
    /* 7 */
    {
        { 3, 0x02, 99 },
        { 6, 0x02, 99 },
        { 10, 0x02, 99 },
        { 15, 0x02, 99 },
        { 24, 0x02, 99 },
        { 31, 0x02, 99 },
        { 41, 0x02, 99 },
        { 56, 0x03, 99 },
        { 3, 0x02, 101 },
        { 6, 0x02, 101 },
        { 10, 0x02, 101 },
        { 15, 0x02, 101 },
        { 24, 0x02, 101 },
        { 31, 0x02, 101 },
        { 41, 0x02, 101 },
        { 56, 0x03, 101 },
    },
    /* 8 */
    {
        { 3, 0x02, 105 },
        { 6, 0x02, 105 },
        { 10, 0x02, 105 },
        { 15, 0x02, 105 },
        { 24, 0x02, 105 },
        { 31, 0x02, 105 },
        { 41, 0x02, 105 },
        { 56, 0x03, 105 },
        { 3, 0x02, 111 },
        { 6, 0x02, 111 },
        { 10, 0x02, 111 },
        { 15, 0x02, 111 },
        { 24, 0x02, 111 },
        { 31, 0x02, 111 },
        { 41, 0x02, 111 },
        { 56, 0x03, 111 },
    },
    /* 9 */
    {
        { 1, 0x02, 115 },
        { 22, 0x03, 115 },
        { 1, 0x02, 116 },
        { 22, 0x03, 116 },
        { 0, 0x03, 32 },
        { 0, 0x03, 37 },
        { 0, 0x03, 45 },
        { 0, 0x03, 46 },
        { 0, 0x03, 47 },
        { 0, 0x03, 51 },
        { 0, 0x03, 52 },
        { 0, 0x03, 53 },
        { 0, 0x03, 54 },
        { 0, 0x03, 55 },
        { 0, 0x03, 56 },
        { 0, 0x03, 57 },
    },
    /* 10 */
    {
        { 2, 0x02, 115 },
        { 9, 0x02, 115 },
        { 23, 0x02, 115 },
        { 40, 0x03, 115 },
        { 2, 0x02, 116 },
        { 9, 0x02, 116 },
        { 23, 0x02, 116 },
        { 40, 0x03, 116 },
        { 1, 0x02, 32 },
        { 22, 0x03, 32 },
        { 1, 0x02, 37 },
        { 22, 0x03, 37 },
        { 1, 0x02, 45 },
        { 22, 0x03, 45 },
        { 1, 0x02, 46 },
        { 22, 0x03, 46 },
    },
    /* 11 */
    {
        { 3, 0x02, 115 },
        { 6, 0x02, 115 },
        { 10, 0x02, 115 },
        { 15, 0x02, 115 },
        { 24, 0x02, 115 },
        { 31, 0x02, 115 },
        { 41, 0x02, 115 },
        { 56, 0x03, 115 },
        { 3, 0x02, 116 },
        { 6, 0x02, 116 },
        { 10, 0x02, 116 },
        { 15, 0x02, 116 },
        { 24, 0x02, 116 },
        { 31, 0x02, 116 },
        { 41, 0x02, 116 },
        { 56, 0x03, 116 },
    },
    /* 12 */
    {
        { 2, 0x02, 32 },
        { 9, 0x02, 32 },
        { 23, 0x02, 32 },
        { 40, 0x03, 32 },
        { 2, 0x02, 37 },
        { 9, 0x02, 37 },
        { 23, 0x02, 37 },
        { 40, 0x03, 37 },
        { 2, 0x02, 45 },
        { 9, 0x02, 45 },
        { 23, 0x02, 45 },
        { 40, 0x03, 45 },
        { 2, 0x02, 46 },
        { 9, 0x02, 46 },
        { 23, 0x02, 46 },
        { 40, 0x03, 46 },
    },
    /* 13 */
    {
        { 3, 0x02, 32 },
        { 6, 0x02, 32 },
        { 10, 0x02, 32 },
        { 15, 0x02, 32 },
        { 24, 0x02, 32 },
        { 31, 0x02, 32 },
        { 41, 0x02, 32 },
        { 56, 0x03, 32 },
        { 3, 0x02, 37 },
        { 6, 0x02, 37 },
        { 10, 0x02, 37 },
        { 15, 0x02, 37 },
        { 24, 0x02, 37 },
        { 31, 0x02, 37 },
        { 41, 0x02, 37 },
        { 56, 0x03, 37 },
    },
    /* 14 */
    {
        { 3, 0x02, 45 },
        { 6, 0x02, 45 },
        { 10, 0x02, 45 },
        { 15, 0x02, 45 },
        { 24, 0x02, 45 },
        { 31, 0x02, 45 },
        { 41, 0x02, 45 },
        { 56, 0x03, 45 },
        { 3, 0x02, 46 },
        { 6, 0x02, 46 },
        { 10, 0x02, 46 },
        { 15, 0x02, 46 },
        { 24, 0x02, 46 },
        { 31, 0x02, 46 },
        { 41, 0x02, 46 },
        { 56, 0x03, 46 },
    },
    /* 15 */
    {
        { 1, 0x02, 47 },
        { 22, 0x03, 47 },
        { 1, 0x02, 51 },
        { 22, 0x03, 51 },
        { 1, 0x02, 52 },
        { 22, 0x03, 52 },
        { 1, 0x02, 53 },
        { 22, 0x03, 53 },
        { 1, 0x02, 54 },
        { 22, 0x03, 54 },
        { 1, 0x02, 55 },
        { 22, 0x03, 55 },
        { 1, 0x02, 56 },
        { 22, 0x03, 56 },
        { 1, 0x02, 57 },
        { 22, 0x03, 57 },
    },
    /* 16 */
    {
        { 2, 0x02, 47 },
        { 9, 0x02, 47 },
        { 23, 0x02, 47 },
        { 40, 0x03, 47 },
        { 2, 0x02, 51 },
        { 9, 0x02, 51 },
        { 23, 0x02, 51 },
        { 40, 0x03, 51 },
        { 2, 0x02, 52 },
        { 9, 0x02, 52 },
        { 23, 0x02, 52 },
        { 40, 0x03, 52 },
        { 2, 0x02, 53 },
        { 9, 0x02, 53 },
        { 23, 0x02, 53 },
        { 40, 0x03, 53 },
    },
    /* 17 */
    {
        { 3, 0x02, 47 },
        { 6, 0x02, 47 },
        { 10, 0x02, 47 },
        { 15, 0x02, 47 },
        { 24, 0x02, 47 },
        { 31, 0x02, 47 },
        { 41, 0x02, 47 },
        { 56, 0x03, 47 },
        { 3, 0x02, 51 },
        { 6, 0x02, 51 },
        { 10, 0x02, 51 },
        { 15, 0x02, 51 },
        { 24, 0x02, 51 },
        { 31, 0x02, 51 },
        { 41, 0x02, 51 },
        { 56, 0x03, 51 },
    },
    /* 18 */
    {
        { 3, 0x02, 52 },
        { 6, 0x02, 52 },
        { 10, 0x02, 52 },
        { 15, 0x02, 52 },
        { 24, 0x02, 52 },
        { 31, 0x02, 52 },
        { 41, 0x02, 52 },
        { 56, 0x03, 52 },
        { 3, 0x02, 53 },
        { 6, 0x02, 53 },
        { 10, 0x02, 53 },
        { 15, 0x02, 53 },
        { 24, 0x02, 53 },
        { 31, 0x02, 53 },
        { 41, 0x02, 53 },
        { 56, 0x03, 53 },
    },
    /* 19 */
    {
        { 2, 0x02, 54 },
        { 9, 0x02, 54 },
        { 23, 0x02, 54 },
        { 40, 0x03, 54 },
        { 2, 0x02, 55 },
        { 9, 0x02, 55 },
        { 23, 0x02, 55 },
        { 40, 0x03, 55 },
        { 2, 0x02, 56 },
        { 9, 0x02, 56 },
        { 23, 0x02, 56 },
        { 40, 0x03, 56 },
        { 2, 0x02, 57 },
        { 9, 0x02, 57 },
        { 23, 0x02, 57 },
        { 40, 0x03, 57 },
    },
    /* 20 */
    {
        { 3, 0x02, 54 },
        { 6, 0x02, 54 },
        { 10, 0x02, 54 },
        { 15, 0x02, 54 },
        { 24, 0x02, 54 },
        { 31, 0x02, 54 },
        { 41, 0x02, 54 },
        { 56, 0x03, 54 },
        { 3, 0x02, 55 },
        { 6, 0x02, 55 },
        { 10, 0x02, 55 },
        { 15, 0x02, 55 },
        { 24, 0x02, 55 },
        { 31, 0x02, 55 },
        { 41, 0x02, 55 },
        { 56, 0x03, 55 },
    },
    /* 21 */
    {
        { 3, 0x02, 56 },
        { 6, 0x02, 56 },
        { 10, 0x02, 56 },
        { 15, 0x02, 56 },
        { 24, 0x02, 56 },
        { 31, 0x02, 56 },
        { 41, 0x02, 56 },
        { 56, 0x03, 56 },
        { 3, 0x02, 57 },
        { 6, 0x02, 57 },
        { 10, 0x02, 57 },
        { 15, 0x02, 57 },
        { 24, 0x02, 57 },
        { 31, 0x02, 57 },
        { 41, 0x02, 57 },
        { 56, 0x03, 57 },
    },
    /* 22 */
    {
        { 26, 0x00, 0 },
        { 27, 0x00, 0 },
        { 29, 0x00, 0 },
        { 30, 0x00, 0 },
        { 33, 0x00, 0 },
        { 34, 0x00, 0 },
        { 36, 0x00, 0 },
        { 37, 0x00, 0 },
        { 43, 0x00, 0 },
        { 46, 0x00, 0 },
        { 50, 0x00, 0 },
        { 53, 0x00, 0 },
        { 58, 0x00, 0 },
        { 61, 0x00, 0 },
        { 65, 0x00, 0 },
        { 68, 0x01, 0 },
    },
    /* 23 */
    {
        { 0, 0x03, 61 },
        { 0, 0x03, 65 },
        { 0, 0x03, 95 },
        { 0, 0x03, 98 },
        { 0, 0x03, 100 },
        { 0, 0x03, 102 },
        { 0, 0x03, 103 },
        { 0, 0x03, 104 },
        { 0, 0x03, 108 },
        { 0, 0x03, 109 },
        { 0, 0x03, 110 },
        { 0, 0x03, 112 },
        { 0, 0x03, 114 },
        { 0, 0x03, 117 },
        { 38, 0x00, 0 },
        { 39, 0x00, 0 },
    },
    /* 24 */
    {
        { 1, 0x02, 61 },
        { 22, 0x03, 61 },
        { 1, 0x02, 65 },
        { 22, 0x03, 65 },
        { 1, 0x02, 95 },
        { 22, 0x03, 95 },
        { 1, 0x02, 98 },
        { 22, 0x03, 98 },
        { 1, 0x02, 100 },
        { 22, 0x03, 100 },
        { 1, 0x02, 102 },
        { 22, 0x03, 102 },
        { 1, 0x02, 103 },
        { 22, 0x03, 103 },
        { 1, 0x02, 104 },
        { 22, 0x03, 104 },
    },
    /* 25 */
    {
        { 2, 0x02, 61 },
        { 9, 0x02, 61 },
        { 23, 0x02, 61 },
        { 40, 0x03, 61 },
        { 2, 0x02, 65 },
        { 9, 0x02, 65 },
        { 23, 0x02, 65 },
        { 40, 0x03, 65 },
        { 2, 0x02, 95 },
        { 9, 0x02, 95 },
        { 23, 0x02, 95 },
        { 40, 0x03, 95 },
        { 2, 0x02, 98 },
        { 9, 0x02, 98 },
        { 23, 0x02, 98 },
        { 40, 0x03, 98 },
    },
    /* 26 */
    {
        { 3, 0x02, 61 },
        { 6, 0x02, 61 },
        { 10, 0x02, 61 },
        { 15, 0x02, 61 },
        { 24, 0x02, 61 },
        { 31, 0x02, 61 },
        { 41, 0x02, 61 },
        { 56, 0x03, 61 },
        { 3, 0x02, 65 },
        { 6, 0x02, 65 },
        { 10, 0x02, 65 },
        { 15, 0x02, 65 },
        { 24, 0x02, 65 },
        { 31, 0x02, 65 },
        { 41, 0x02, 65 },
        { 56, 0x03, 65 },
    },
    /* 27 */
    {
        { 3, 0x02, 95 },
        { 6, 0x02, 95 },
        { 10, 0x02, 95 },
        { 15, 0x02, 95 },
        { 24, 0x02, 95 },
        { 31, 0x02, 95 },
        { 41, 0x02, 95 },
        { 56, 0x03, 95 },
        { 3, 0x02, 98 },
        { 6, 0x02, 98 },
        { 10, 0x02, 98 },
        { 15, 0x02, 98 },
        { 24, 0x02, 98 },
        { 31, 0x02, 98 },
        { 41, 0x02, 98 },
        { 56, 0x03, 98 },
    },
    /* 28 */
    {
        { 2, 0x02, 100 },
        { 9, 0x02, 100 },
        { 23, 0x02, 100 },
        { 40, 0x03, 100 },
        { 2, 0x02, 102 },
        { 9, 0x02, 102 },
        { 23, 0x02, 102 },
        { 40, 0x03, 102 },
        { 2, 0x02, 103 },
        { 9, 0x02, 103 },
        { 23, 0x02, 103 },
        { 40, 0x03, 103 },
        { 2, 0x02, 104 },
        { 9, 0x02, 104 },
        { 23, 0x02, 104 },
        { 40, 0x03, 104 },
    },
    /* 29 */
    {
        { 3, 0x02, 100 },
        { 6, 0x02, 100 },
        { 10, 0x02, 100 },
        { 15, 0x02, 100 },
        { 24, 0x02, 100 },
        { 31, 0x02, 100 },
        { 41, 0x02, 100 },
        { 56, 0x03, 100 },
        { 3, 0x02, 102 },
        { 6, 0x02, 102 },
        { 10, 0x02, 102 },
        { 15, 0x02, 102 },
        { 24, 0x02, 102 },
        { 31, 0x02, 102 },
        { 41, 0x02, 102 },
        { 56, 0x03, 102 },
    },
    /* 30 */
    {
        { 3, 0x02, 103 },
        { 6, 0x02, 103 },
        { 10, 0x02, 103 },
        { 15, 0x02, 103 },
        { 24, 0x02, 103 },
        { 31, 0x02, 103 },
        { 41, 0x02, 103 },
        { 56, 0x03, 103 },
        { 3, 0x02, 104 },
        { 6, 0x02, 104 },
        { 10, 0x02, 104 },
        { 15, 0x02, 104 },
        { 24, 0x02, 104 },
        { 31, 0x02, 104 },
        { 41, 0x02, 104 },
        { 56, 0x03, 104 },
    },
    /* 31 */
    {
        { 1, 0x02, 108 },
        { 22, 0x03, 108 },
        { 1, 0x02, 109 },
        { 22, 0x03, 109 },
        { 1, 0x02, 110 },
        { 22, 0x03, 110 },
        { 1, 0x02, 112 },
        { 22, 0x03, 112 },
        { 1, 0x02, 114 },
        { 22, 0x03, 114 },
        { 1, 0x02, 117 },
        { 22, 0x03, 117 },
        { 0, 0x03, 58 },
        { 0, 0x03, 66 },
        { 0, 0x03, 67 },
        { 0, 0x03, 68 },
    },
    /* 32 */
    {
        { 2, 0x02, 108 },
        { 9, 0x02, 108 },
        { 23, 0x02, 108 },
        { 40, 0x03, 108 },
        { 2, 0x02, 109 },
        { 9, 0x02, 109 },
        { 23, 0x02, 109 },
        { 40, 0x03, 109 },
        { 2, 0x02, 110 },
        { 9, 0x02, 110 },
        { 23, 0x02, 110 },
        { 40, 0x03, 110 },
        { 2, 0x02, 112 },
        { 9, 0x02, 112 },
        { 23, 0x02, 112 },
        { 40, 0x03, 112 },
    },
    /* 33 */
    {
        { 3, 0x02, 108 },
        { 6, 0x02, 108 },
        { 10, 0x02, 108 },
        { 15, 0x02, 108 },
        { 24, 0x02, 108 },
        { 31, 0x02, 108 },
        { 41, 0x02, 108 },
        { 56, 0x03, 108 },
        { 3, 0x02, 109 },
        { 6, 0x02, 109 },
        { 10, 0x02, 109 },
        { 15, 0x02, 109 },
        { 24, 0x02, 109 },
        { 31, 0x02, 109 },
        { 41, 0x02, 109 },
        { 56, 0x03, 109 },
    },
    /* 34 */
    {
        { 3, 0x02, 110 },
        { 6, 0x02, 110 },
        { 10, 0x02, 110 },
        { 15, 0x02, 110 },
        { 24, 0x02, 110 },
        { 31, 0x02, 110 },
        { 41, 0x02, 110 },
        { 56, 0x03, 110 },
        { 3, 0x02, 112 },
        { 6, 0x02, 112 },
        { 10, 0x02, 112 },
        { 15, 0x02, 112 },
        { 24, 0x02, 112 },
        { 31, 0x02, 112 },
        { 41, 0x02, 112 },
        { 56, 0x03, 112 },
    },
    /* 35 */
    {
        { 2, 0x02, 114 },
        { 9, 0x02, 114 },
        { 23, 0x02, 114 },
        { 40, 0x03, 114 },
        { 2, 0x02, 117 },
        { 9, 0x02, 117 },
        { 23, 0x02, 117 },
        { 40, 0x03, 117 },
        { 1, 0x02, 58 },
        { 22, 0x03, 58 },
        { 1, 0x02, 66 },
        { 22, 0x03, 66 },
        { 1, 0x02, 67 },
        { 22, 0x03, 67 },
        { 1, 0x02, 68 },
        { 22, 0x03, 68 },
    },
    /* 36 */
    {
        { 3, 0x02, 114 },
        { 6, 0x02, 114 },
        { 10, 0x02, 114 },
        { 15, 0x02, 114 },
        { 24, 0x02, 114 },
        { 31, 0x02, 114 },
        { 41, 0x02, 114 },
        { 56, 0x03, 114 },
        { 3, 0x02, 117 },
        { 6, 0x02, 117 },
        { 10, 0x02, 117 },
        { 15, 0x02, 117 },
        { 24, 0x02, 117 },
        { 31, 0x02, 117 },
        { 41, 0x02, 117 },
        { 56, 0x03, 117 },
    },
    /* 37 */
    {
        { 2, 0x02, 58 },
        { 9, 0x02, 58 },
        { 23, 0x02, 58 },
        { 40, 0x03, 58 },
        { 2, 0x02, 66 },
        { 9, 0x02, 66 },
        { 23, 0x02, 66 },
        { 40, 0x03, 66 },
        { 2, 0x02, 67 },
        { 9, 0x02, 67 },
        { 23, 0x02, 67 },
        { 40, 0x03, 67 },
        { 2, 0x02, 68 },
        { 9, 0x02, 68 },
        { 23, 0x02, 68 },
        { 40, 0x03, 68 },
    },
    /* 38 */
    {
        { 3, 0x02, 58 },
        { 6, 0x02, 58 },
        { 10, 0x02, 58 },
        { 15, 0x02, 58 },
        { 24, 0x02, 58 },
        { 31, 0x02, 58 },
        { 41, 0x02, 58 },
        { 56, 0x03, 58 },
        { 3, 0x02, 66 },
        { 6, 0x02, 66 },
        { 10, 0x02, 66 },
        { 15, 0x02, 66 },
        { 24, 0x02, 66 },
        { 31, 0x02, 66 },
        { 41, 0x02, 66 },
        { 56, 0x03, 66 },
    },
    /* 39 */
    {
        { 3, 0x02, 67 },
        { 6, 0x02, 67 },
        { 10, 0x02, 67 },
        { 15, 0x02, 67 },
        { 24, 0x02, 67 },
        { 31, 0x02, 67 },
        { 41, 0x02, 67 },
        { 56, 0x03, 67 },
        { 3, 0x02, 68 },
        { 6, 0x02, 68 },
        { 10, 0x02, 68 },
        { 15, 0x02, 68 },
        { 24, 0x02, 68 },
        { 31, 0x02, 68 },
        { 41, 0x02, 68 },
        { 56, 0x03, 68 },
    },
    /* 40 */
    {
        { 44, 0x00, 0 },
        { 45, 0x00, 0 },
        { 47, 0x00, 0 },
        { 48, 0x00, 0 },
        { 51, 0x00, 0 },
        { 52, 0x00, 0 },
        { 54, 0x00, 0 },
        { 55, 0x00, 0 },
        { 59, 0x00, 0 },
        { 60, 0x00, 0 },
        { 62, 0x00, 0 },
        { 63, 0x00, 0 },
        { 66, 0x00, 0 },
        { 67, 0x00, 0 },
        { 69, 0x00, 0 },
        { 72, 0x01, 0 },
    },
    /* 41 */
    {
        { 0, 0x03, 69 },
        { 0, 0x03, 70 },
        { 0, 0x03, 71 },
        { 0, 0x03, 72 },
        { 0, 0x03, 73 },
        { 0, 0x03, 74 },
        { 0, 0x03, 75 },
        { 0, 0x03, 76 },
        { 0, 0x03, 77 },
        { 0, 0x03, 78 },
        { 0, 0x03, 79 },
        { 0, 0x03, 80 },
        { 0, 0x03, 81 },
        { 0, 0x03, 82 },
        { 0, 0x03, 83 },
        { 0, 0x03, 84 },
    },
    /* 42 */
    {
        { 1, 0x02, 69 },
        { 22, 0x03, 69 },
        { 1, 0x02, 70 },
        { 22, 0x03, 70 },
        { 1, 0x02, 71 },
        { 22, 0x03, 71 },
        { 1, 0x02, 72 },
        { 22, 0x03, 72 },
        { 1, 0x02, 73 },
        { 22, 0x03, 73 },
        { 1, 0x02, 74 },
        { 22, 0x03, 74 },
        { 1, 0x02, 75 },
        { 22, 0x03, 75 },
        { 1, 0x02, 76 },
        { 22, 0x03, 76 },
    },
    /* 43 */
    {
        { 2, 0x02, 69 },
        { 9, 0x02, 69 },
        { 23, 0x02, 69 },
        { 40, 0x03, 69 },
        { 2, 0x02, 70 },
        { 9, 0x02, 70 },
        { 23, 0x02, 70 },
        { 40, 0x03, 70 },
        { 2, 0x02, 71 },
        { 9, 0x02, 71 },
        { 23, 0x02, 71 },
        { 40, 0x03, 71 },
        { 2, 0x02, 72 },
        { 9, 0x02, 72 },
        { 23, 0x02, 72 },
        { 40, 0x03, 72 },
    },
    /* 44 */
    {
        { 3, 0x02, 69 },
        { 6, 0x02, 69 },
        { 10, 0x02, 69 },
        { 15, 0x02, 69 },
        { 24, 0x02, 69 },
        { 31, 0x02, 69 },
        { 41, 0x02, 69 },
        { 56, 0x03, 69 },
        { 3, 0x02, 70 },
        { 6, 0x02, 70 },
        { 10, 0x02, 70 },
        { 15, 0x02, 70 },
        { 24, 0x02, 70 },
        { 31, 0x02, 70 },
        { 41, 0x02, 70 },
        { 56, 0x03, 70 },
    },
    /* 45 */
    {
        { 3, 0x02, 71 },
        { 6, 0x02, 71 },
        { 10, 0x02, 71 },
        { 15, 0x02, 71 },
        { 24, 0x02, 71 },
        { 31, 0x02, 71 },
        { 41, 0x02, 71 },
        { 56, 0x03, 71 },
        { 3, 0x02, 72 },
        { 6, 0x02, 72 },
        { 10, 0x02, 72 },
        { 15, 0x02, 72 },
        { 24, 0x02, 72 },
        { 31, 0x02, 72 },
        { 41, 0x02, 72 },
        { 56, 0x03, 72 },
    },
    /* 46 */
    {
        { 2, 0x02, 73 },
        { 9, 0x02, 73 },
        { 23, 0x02, 73 },
        { 40, 0x03, 73 },
        { 2, 0x02, 74 },
        { 9, 0x02, 74 },
        { 23, 0x02, 74 },
        { 40, 0x03, 74 },
        { 2, 0x02, 75 },
        { 9, 0x02, 75 },
        { 23, 0x02, 75 },
        { 40, 0x03, 75 },
        { 2, 0x02, 76 },
        { 9, 0x02, 76 },
        { 23, 0x02, 76 },
        { 40, 0x03, 76 },
    },
    /* 47 */
    {
        { 3, 0x02, 73 },
        { 6, 0x02, 73 },
        { 10, 0x02, 73 },
        { 15, 0x02, 73 },
        { 24, 0x02, 73 },
        { 31, 0x02, 73 },
        { 41, 0x02, 73 },
        { 56, 0x03, 73 },
        { 3, 0x02, 74 },
        { 6, 0x02, 74 },
        { 10, 0x02, 74 },
        { 15, 0x02, 74 },
        { 24, 0x02, 74 },
        { 31, 0x02, 74 },
        { 41, 0x02, 74 },
        { 56, 0x03, 74 },
    },
    /* 48 */
    {
        { 3, 0x02, 75 },
        { 6, 0x02, 75 },
        { 10, 0x02, 75 },
        { 15, 0x02, 75 },
        { 24, 0x02, 75 },
        { 31, 0x02, 75 },
        { 41, 0x02, 75 },
        { 56, 0x03, 75 },
        { 3, 0x02, 76 },
        { 6, 0x02, 76 },
        { 10, 0x02, 76 },
        { 15, 0x02, 76 },
        { 24, 0x02, 76 },
        { 31, 0x02, 76 },
        { 41, 0x02, 76 },
        { 56, 0x03, 76 },
    },
    /* 49 */
    {
        { 1, 0x02, 77 },
        { 22, 0x03, 77 },
        { 1, 0x02, 78 },
        { 22, 0x03, 78 },
        { 1, 0x02, 79 },
        { 22, 0x03, 79 },
        { 1, 0x02, 80 },
        { 22, 0x03, 80 },
        { 1, 0x02, 81 },
        { 22, 0x03, 81 },
        { 1, 0x02, 82 },
        { 22, 0x03, 82 },
        { 1, 0x02, 83 },
        { 22, 0x03, 83 },
        { 1, 0x02, 84 },
        { 22, 0x03, 84 },
    },
    /* 50 */
    {
        { 2, 0x02, 77 },
        { 9, 0x02, 77 },
        { 23, 0x02, 77 },
        { 40, 0x03, 77 },
        { 2, 0x02, 78 },
        { 9, 0x02, 78 },
        { 23, 0x02, 78 },
        { 40, 0x03, 78 },
        { 2, 0x02, 79 },
        { 9, 0x02, 79 },
        { 23, 0x02, 79 },
        { 40, 0x03, 79 },
        { 2, 0x02, 80 },
        { 9, 0x02, 80 },
        { 23, 0x02, 80 },
        { 40, 0x03, 80 },
    },
    /* 51 */
    {
        { 3, 0x02, 77 },
        { 6, 0x02, 77 },
        { 10, 0x02, 77 },
        { 15, 0x02, 77 },
        { 24, 0x02, 77 },
        { 31, 0x02, 77 },
        { 41, 0x02, 77 },
        { 56, 0x03, 77 },
        { 3, 0x02, 78 },
        { 6, 0x02, 78 },
        { 10, 0x02, 78 },
        { 15, 0x02, 78 },
        { 24, 0x02, 78 },
        { 31, 0x02, 78 },
        { 41, 0x02, 78 },
        { 56, 0x03, 78 },
    },
    /* 52 */
    {
        { 3, 0x02, 79 },
        { 6, 0x02, 79 },
        { 10, 0x02, 79 },
        { 15, 0x02, 79 },
        { 24, 0x02, 79 },
        { 31, 0x02, 79 },
        { 41, 0x02, 79 },
        { 56, 0x03, 79 },
        { 3, 0x02, 80 },
        { 6, 0x02, 80 },
        { 10, 0x02, 80 },
        { 15, 0x02, 80 },
        { 24, 0x02, 80 },
        { 31, 0x02, 80 },
        { 41, 0x02, 80 },
        { 56, 0x03, 80 },
    },
    /* 53 */
    {
        { 2, 0x02, 81 },
        { 9, 0x02, 81 },
        { 23, 0x02, 81 },
        { 40, 0x03, 81 },
        { 2, 0x02, 82 },
        { 9, 0x02, 82 },
        { 23, 0x02, 82 },
        { 40, 0x03, 82 },
        { 2, 0x02, 83 },
        { 9, 0x02, 83 },
        { 23, 0x02, 83 },
        { 40, 0x03, 83 },
        { 2, 0x02, 84 },
        { 9, 0x02, 84 },
        { 23, 0x02, 84 },
        { 40, 0x03, 84 },
    },
    /* 54 */
    {
        { 3, 0x02, 81 },
        { 6, 0x02, 81 },
        { 10, 0x02, 81 },
        { 15, 0x02, 81 },
        { 24, 0x02, 81 },
        { 31, 0x02, 81 },
        { 41, 0x02, 81 },
        { 56, 0x03, 81 },
        { 3, 0x02, 82 },
        { 6, 0x02, 82 },
        { 10, 0x02, 82 },
        { 15, 0x02, 82 },
        { 24, 0x02, 82 },
        { 31, 0x02, 82 },
        { 41, 0x02, 82 },
        { 56, 0x03, 82 },
    },
    /* 55 */
    {
        { 3, 0x02, 83 },
        { 6, 0x02, 83 },
        { 10, 0x02, 83 },
        { 15, 0x02, 83 },
        { 24, 0x02, 83 },
        { 31, 0x02, 83 },
        { 41, 0x02, 83 },
        { 56, 0x03, 83 },
        { 3, 0x02, 84 },
        { 6, 0x02, 84 },
        { 10, 0x02, 84 },
        { 15, 0x02, 84 },
        { 24, 0x02, 84 },
        { 31, 0x02, 84 },
        { 41, 0x02, 84 },
        { 56, 0x03, 84 },
    },
    /* 56 */
    {
        { 0, 0x03, 85 },
        { 0, 0x03, 86 },
        { 0, 0x03, 87 },
        { 0, 0x03, 89 },
        { 0, 0x03, 106 },
        { 0, 0x03, 107 },
        { 0, 0x03, 113 },
        { 0, 0x03, 118 },
        { 0, 0x03, 119 },
        { 0, 0x03, 120 },
        { 0, 0x03, 121 },
        { 0, 0x03, 122 },
        { 70, 0x00, 0 },
        { 71, 0x00, 0 },
        { 73, 0x00, 0 },
        { 74, 0x01, 0 },
    },
    /* 57 */
    {
        { 1, 0x02, 85 },
        { 22, 0x03, 85 },
        { 1, 0x02, 86 },
        { 22, 0x03, 86 },
        { 1, 0x02, 87 },
        { 22, 0x03, 87 },
        { 1, 0x02, 89 },
        { 22, 0x03, 89 },
        { 1, 0x02, 106 },
        { 22, 0x03, 106 },
        { 1, 0x02, 107 },
        { 22, 0x03, 107 },
        { 1, 0x02, 113 },
        { 22, 0x03, 113 },
        { 1, 0x02, 118 },
        { 22, 0x03, 118 },
    },
    /* 58 */
    {
        { 2, 0x02, 85 },
        { 9, 0x02, 85 },
        { 23, 0x02, 85 },
        { 40, 0x03, 85 },
        { 2, 0x02, 86 },
        { 9, 0x02, 86 },
        { 23, 0x02, 86 },
        { 40, 0x03, 86 },
        { 2, 0x02, 87 },
        { 9, 0x02, 87 },
        { 23, 0x02, 87 },
        { 40, 0x03, 87 },
        { 2, 0x02, 89 },
        { 9, 0x02, 89 },
        { 23, 0x02, 89 },
        { 40, 0x03, 89 },
    },
    /* 59 */
    {
        { 3, 0x02, 85 },
        { 6, 0x02, 85 },
        { 10, 0x02, 85 },
        { 15, 0x02, 85 },
        { 24, 0x02, 85 },
        { 31, 0x02, 85 },
        { 41, 0x02, 85 },
        { 56, 0x03, 85 },
        { 3, 0x02, 86 },
        { 6, 0x02, 86 },
        { 10, 0x02, 86 },
        { 15, 0x02, 86 },
        { 24, 0x02, 86 },
        { 31, 0x02, 86 },
        { 41, 0x02, 86 },
        { 56, 0x03, 86 },
    },
    /* 60 */
    {
        { 3, 0x02, 87 },
        { 6, 0x02, 87 },
        { 10, 0x02, 87 },
        { 15, 0x02, 87 },
        { 24, 0x02, 87 },
        { 31, 0x02, 87 },
        { 41, 0x02, 87 },
        { 56, 0x03, 87 },
        { 3, 0x02, 89 },
        { 6, 0x02, 89 },
        { 10, 0x02, 89 },
        { 15, 0x02, 89 },
        { 24, 0x02, 89 },
        { 31, 0x02, 89 },
        { 41, 0x02, 89 },
        { 56, 0x03, 89 },
    },
    /* 61 */
    {
        { 2, 0x02, 106 },
        { 9, 0x02, 106 },
        { 23, 0x02, 106 },
        { 40, 0x03, 106 },
        { 2, 0x02, 107 },
        { 9, 0x02, 107 },
        { 23, 0x02, 107 },
        { 40, 0x03, 107 },
        { 2, 0x02, 113 },
        { 9, 0x02, 113 },
        { 23, 0x02, 113 },
        { 40, 0x03, 113 },
        { 2, 0x02, 118 },
        { 9, 0x02, 118 },
        { 23, 0x02, 118 },
        { 40, 0x03, 118 },
    },
    /* 62 */
    {
        { 3, 0x02, 106 },
        { 6, 0x02, 106 },
        { 10, 0x02, 106 },
        { 15, 0x02, 106 },
        { 24, 0x02, 106 },
        { 31, 0x02, 106 },
        { 41, 0x02, 106 },
        { 56, 0x03, 106 },
        { 3, 0x02, 107 },
        { 6, 0x02, 107 },
        { 10, 0x02, 107 },
        { 15, 0x02, 107 },
        { 24, 0x02, 107 },
        { 31, 0x02, 107 },
        { 41, 0x02, 107 },
        { 56, 0x03, 107 },
    },
    /* 63 */
    {
        { 3, 0x02, 113 },
        { 6, 0x02, 113 },
        { 10, 0x02, 113 },
        { 15, 0x02, 113 },
        { 24, 0x02, 113 },
        { 31, 0x02, 113 },
        { 41, 0x02, 113 },
        { 56, 0x03, 113 },
        { 3, 0x02, 118 },
        { 6, 0x02, 118 },
        { 10, 0x02, 118 },
        { 15, 0x02, 118 },
        { 24, 0x02, 118 },
        { 31, 0x02, 118 },
        { 41, 0x02, 118 },
        { 56, 0x03, 118 },
    },
    /* 64 */
    {
        { 1, 0x02, 119 },
        { 22, 0x03, 119 },
        { 1, 0x02, 120 },
        { 22, 0x03, 120 },
        { 1, 0x02, 121 },
        { 22, 0x03, 121 },
        { 1, 0x02, 122 },
        { 22, 0x03, 122 },
        { 0, 0x03, 38 },
        { 0, 0x03, 42 },
        { 0, 0x03, 44 },
        { 0, 0x03, 59 },
        { 0, 0x03, 88 },
        { 0, 0x03, 90 },
        { 75, 0x00, 0 },
        { 78, 0x00, 0 },
    },
    /* 65 */
    {
        { 2, 0x02, 119 },
        { 9, 0x02, 119 },
        { 23, 0x02, 119 },
        { 40, 0x03, 119 },
        { 2, 0x02, 120 },
        { 9, 0x02, 120 },
        { 23, 0x02, 120 },
        { 40, 0x03, 120 },
        { 2, 0x02, 121 },
        { 9, 0x02, 121 },
        { 23, 0x02, 121 },
        { 40, 0x03, 121 },
        { 2, 0x02, 122 },
        { 9, 0x02, 122 },
        { 23, 0x02, 122 },
        { 40, 0x03, 122 },
    },
    /* 66 */
    {
        { 3, 0x02, 119 },
        { 6, 0x02, 119 },
        { 10, 0x02, 119 },
        { 15, 0x02, 119 },
        { 24, 0x02, 119 },
        { 31, 0x02, 119 },
        { 41, 0x02, 119 },
        { 56, 0x03, 119 },
        { 3, 0x02, 120 },
        { 6, 0x02, 120 },
        { 10, 0x02, 120 },
        { 15, 0x02, 120 },
        { 24, 0x02, 120 },
        { 31, 0x02, 120 },
        { 41, 0x02, 120 },
        { 56, 0x03, 120 },
    },
    /* 67 */
    {
        { 3, 0x02, 121 },
        { 6, 0x02, 121 },
        { 10, 0x02, 121 },
        { 15, 0x02, 121 },
        { 24, 0x02, 121 },
        { 31, 0x02, 121 },
        { 41, 0x02, 121 },
        { 56, 0x03, 121 },
        { 3, 0x02, 122 },
        { 6, 0x02, 122 },
        { 10, 0x02, 122 },
        { 15, 0x02, 122 },
        { 24, 0x02, 122 },
        { 31, 0x02, 122 },
        { 41, 0x02, 122 },
        { 56, 0x03, 122 },
    },
    /* 68 */
    {
        { 1, 0x02, 38 },
        { 22, 0x03, 38 },
        { 1, 0x02, 42 },
        { 22, 0x03, 42 },
        { 1, 0x02, 44 },
        { 22, 0x03, 44 },
        { 1, 0x02, 59 },
        { 22, 0x03, 59 },
        { 1, 0x02, 88 },
        { 22, 0x03, 88 },
        { 1, 0x02, 90 },
        { 22, 0x03, 90 },
        { 76, 0x00, 0 },
        { 77, 0x00, 0 },
        { 79, 0x00, 0 },
        { 81, 0x00, 0 },
    },
    /* 69 */
    {
        { 2, 0x02, 38 },
        { 9, 0x02, 38 },
        { 23, 0x02, 38 },
        { 40, 0x03, 38 },
        { 2, 0x02, 42 },
        { 9, 0x02, 42 },
        { 23, 0x02, 42 },
        { 40, 0x03, 42 },
        { 2, 0x02, 44 },
        { 9, 0x02, 44 },
        { 23, 0x02, 44 },
        { 40, 0x03, 44 },
        { 2, 0x02, 59 },
        { 9, 0x02, 59 },
        { 23, 0x02, 59 },
        { 40, 0x03, 59 },
    },
    /* 70 */
    {
        { 3, 0x02, 38 },
        { 6, 0x02, 38 },
        { 10, 0x02, 38 },
        { 15, 0x02, 38 },
        { 24, 0x02, 38 },
        { 31, 0x02, 38 },
        { 41, 0x02, 38 },
        { 56, 0x03, 38 },
        { 3, 0x02, 42 },
        { 6, 0x02, 42 },
        { 10, 0x02, 42 },
        { 15, 0x02, 42 },
        { 24, 0x02, 42 },
        { 31, 0x02, 42 },
        { 41, 0x02, 42 },
        { 56, 0x03, 42 },
    },
    /* 71 */
    {
        { 3, 0x02, 44 },
        { 6, 0x02, 44 },
        { 10, 0x02, 44 },
        { 15, 0x02, 44 },
        { 24, 0x02, 44 },
        { 31, 0x02, 44 },
        { 41, 0x02, 44 },
        { 56, 0x03, 44 },
        { 3, 0x02, 59 },
        { 6, 0x02, 59 },
        { 10, 0x02, 59 },
        { 15, 0x02, 59 },
        { 24, 0x02, 59 },
        { 31, 0x02, 59 },
        { 41, 0x02, 59 },
        { 56, 0x03, 59 },
    },
    /* 72 */
    {
        { 2, 0x02, 88 },
        { 9, 0x02, 88 },
        { 23, 0x02, 88 },
        { 40, 0x03, 88 },
        { 2, 0x02, 90 },
        { 9, 0x02, 90 },
        { 23, 0x02, 90 },
        { 40, 0x03, 90 },
        { 0, 0x03, 33 },
        { 0, 0x03, 34 },
        { 0, 0x03, 40 },
        { 0, 0x03, 41 },
        { 0, 0x03, 63 },
        { 80, 0x00, 0 },
        { 82, 0x00, 0 },
        { 84, 0x00, 0 },
    },
    /* 73 */
    {
        { 3, 0x02, 88 },
        { 6, 0x02, 88 },
        { 10, 0x02, 88 },
        { 15, 0x02, 88 },
        { 24, 0x02, 88 },
        { 31, 0x02, 88 },
        { 41, 0x02, 88 },
        { 56, 0x03, 88 },
        { 3, 0x02, 90 },
        { 6, 0x02, 90 },
        { 10, 0x02, 90 },
        { 15, 0x02, 90 },
        { 24, 0x02, 90 },
        { 31, 0x02, 90 },
        { 41, 0x02, 90 },
        { 56, 0x03, 90 },
    },
    /* 74 */
    {
        { 1, 0x02, 33 },
        { 22, 0x03, 33 },
        { 1, 0x02, 34 },
        { 22, 0x03, 34 },
        { 1, 0x02, 40 },
        { 22, 0x03, 40 },
        { 1, 0x02, 41 },
        { 22, 0x03, 41 },
        { 1, 0x02, 63 },
        { 22, 0x03, 63 },
        { 0, 0x03, 39 },
        { 0, 0x03, 43 },
        { 0, 0x03, 124 },
        { 83, 0x00, 0 },
        { 85, 0x00, 0 },
        { 88, 0x00, 0 },
    },
    /* 75 */
    {
        { 2, 0x02, 33 },
        { 9, 0x02, 33 },
        { 23, 0x02, 33 },
        { 40, 0x03, 33 },
        { 2, 0x02, 34 },
        { 9, 0x02, 34 },
        { 23, 0x02, 34 },
        { 40, 0x03, 34 },
        { 2, 0x02, 40 },
        { 9, 0x02, 40 },
        { 23, 0x02, 40 },
        { 40, 0x03, 40 },
        { 2, 0x02, 41 },
        { 9, 0x02, 41 },
        { 23, 0x02, 41 },
        { 40, 0x03, 41 },
    },
    /* 76 */
    {
        { 3, 0x02, 33 },
        { 6, 0x02, 33 },
        { 10, 0x02, 33 },
        { 15, 0x02, 33 },
        { 24, 0x02, 33 },
        { 31, 0x02, 33 },
        { 41, 0x02, 33 },
        { 56, 0x03, 33 },
        { 3, 0x02, 34 },
        { 6, 0x02, 34 },
        { 10, 0x02, 34 },
        { 15, 0x02, 34 },
        { 24, 0x02, 34 },
        { 31, 0x02, 34 },
        { 41, 0x02, 34 },
        { 56, 0x03, 34 },
    },
    /* 77 */
    {
        { 3, 0x02, 40 },
        { 6, 0x02, 40 },
        { 10, 0x02, 40 },
        { 15, 0x02, 40 },
        { 24, 0x02, 40 },
        { 31, 0x02, 40 },
        { 41, 0x02, 40 },
        { 56, 0x03, 40 },
        { 3, 0x02, 41 },
        { 6, 0x02, 41 },
        { 10, 0x02, 41 },
        { 15, 0x02, 41 },
        { 24, 0x02, 41 },
        { 31, 0x02, 41 },
        { 41, 0x02, 41 },
        { 56, 0x03, 41 },
    },
    /* 78 */
    {
        { 2, 0x02, 63 },
        { 9, 0x02, 63 },
        { 23, 0x02, 63 },
        { 40, 0x03, 63 },
        { 1, 0x02, 39 },
        { 22, 0x03, 39 },
        { 1, 0x02, 43 },
        { 22, 0x03, 43 },
        { 1, 0x02, 124 },
        { 22, 0x03, 124 },
        { 0, 0x03, 35 },
        { 0, 0x03, 62 },
        { 86, 0x00, 0 },
        { 87, 0x00, 0 },
        { 89, 0x00, 0 },
        { 90, 0x00, 0 },
    },
    /* 79 */
    {
        { 3, 0x02, 63 },
        { 6, 0x02, 63 },
        { 10, 0x02, 63 },
        { 15, 0x02, 63 },
        { 24, 0x02, 63 },
        { 31, 0x02, 63 },
        { 41, 0x02, 63 },
        { 56, 0x03, 63 },
        { 2, 0x02, 39 },
        { 9, 0x02, 39 },
        { 23, 0x02, 39 },
        { 40, 0x03, 39 },
        { 2, 0x02, 43 },
        { 9, 0x02, 43 },
        { 23, 0x02, 43 },
        { 40, 0x03, 43 },
    },
    /* 80 */
    {
        { 3, 0x02, 39 },
        { 6, 0x02, 39 },
        { 10, 0x02, 39 },
        { 15, 0x02, 39 },
        { 24, 0x02, 39 },
        { 31, 0x02, 39 },
        { 41, 0x02, 39 },
        { 56, 0x03, 39 },
        { 3, 0x02, 43 },
        { 6, 0x02, 43 },
        { 10, 0x02, 43 },
        { 15, 0x02, 43 },
        { 24, 0x02, 43 },
        { 31, 0x02, 43 },
        { 41, 0x02, 43 },
        { 56, 0x03, 43 },
    },
    /* 81 */
    {
        { 2, 0x02, 124 },
        { 9, 0x02, 124 },
        { 23, 0x02, 124 },
        { 40, 0x03, 124 },
        { 1, 0x02, 35 },
        { 22, 0x03, 35 },
        { 1, 0x02, 62 },
        { 22, 0x03, 62 },
        { 0, 0x03, 0 },
        { 0, 0x03, 36 },
        { 0, 0x03, 64 },
        { 0, 0x03, 91 },
        { 0, 0x03, 93 },
        { 0, 0x03, 126 },
        { 91, 0x00, 0 },
        { 92, 0x00, 0 },
    },
    /* 82 */
    {
        { 3, 0x02, 124 },
        { 6, 0x02, 124 },
        { 10, 0x02, 124 },
        { 15, 0x02, 124 },
        { 24, 0x02, 124 },
        { 31, 0x02, 124 },
        { 41, 0x02, 124 },
        { 56, 0x03, 124 },
        { 2, 0x02, 35 },
        { 9, 0x02, 35 },
        { 23, 0x02, 35 },
        { 40, 0x03, 35 },
        { 2, 0x02, 62 },
        { 9, 0x02, 62 },
        { 23, 0x02, 62 },
        { 40, 0x03, 62 },
    },
    /* 83 */
    {
        { 3, 0x02, 35 },
        { 6, 0x02, 35 },
        { 10, 0x02, 35 },
        { 15, 0x02, 35 },
        { 24, 0x02, 35 },
        { 31, 0x02, 35 },
        { 41, 0x02, 35 },
        { 56, 0x03, 35 },
        { 3, 0x02, 62 },
        { 6, 0x02, 62 },
        { 10, 0x02, 62 },
        { 15, 0x02, 62 },
        { 24, 0x02, 62 },
        { 31, 0x02, 62 },
        { 41, 0x02, 62 },
        { 56, 0x03, 62 },
    },
    /* 84 */
    {
        { 1, 0x02, 0 },
        { 22, 0x03, 0 },
        { 1, 0x02, 36 },
        { 22, 0x03, 36 },
        { 1, 0x02, 64 },
        { 22, 0x03, 64 },
        { 1, 0x02, 91 },
        { 22, 0x03, 91 },
        { 1, 0x02, 93 },
        { 22, 0x03, 93 },
        { 1, 0x02, 126 },
        { 22, 0x03, 126 },
        { 0, 0x03, 94 },
        { 0, 0x03, 125 },
        { 93, 0x00, 0 },
        { 94, 0x00, 0 },
    },
    /* 85 */
    {
        { 2, 0x02, 0 },
        { 9, 0x02, 0 },
        { 23, 0x02, 0 },
        { 40, 0x03, 0 },
        { 2, 0x02, 36 },
        { 9, 0x02, 36 },
        { 23, 0x02, 36 },
        { 40, 0x03, 36 },
        { 2, 0x02, 64 },
        { 9, 0x02, 64 },
        { 23, 0x02, 64 },
        { 40, 0x03, 64 },
        { 2, 0x02, 91 },
        { 9, 0x02, 91 },
        { 23, 0x02, 91 },
        { 40, 0x03, 91 },
    },
    /* 86 */
    {
        { 3, 0x02, 0 },
        { 6, 0x02, 0 },
        { 10, 0x02, 0 },
        { 15, 0x02, 0 },
        { 24, 0x02, 0 },
        { 31, 0x02, 0 },
        { 41, 0x02, 0 },
        { 56, 0x03, 0 },
        { 3, 0x02, 36 },
        { 6, 0x02, 36 },
        { 10, 0x02, 36 },
        { 15, 0x02, 36 },
        { 24, 0x02, 36 },
        { 31, 0x02, 36 },
        { 41, 0x02, 36 },
        { 56, 0x03, 36 },
    },
    /* 87 */
    {
        { 3, 0x02, 64 },
        { 6, 0x02, 64 },
        { 10, 0x02, 64 },
        { 15, 0x02, 64 },
        { 24, 0x02, 64 },
        { 31, 0x02, 64 },
        { 41, 0x02, 64 },
        { 56, 0x03, 64 },
        { 3, 0x02, 91 },
        { 6, 0x02, 91 },
        { 10, 0x02, 91 },
        { 15, 0x02, 91 },
        { 24, 0x02, 91 },
        { 31, 0x02, 91 },
        { 41, 0x02, 91 },
        { 56, 0x03, 91 },
    },
    /* 88 */
    {
        { 2, 0x02, 93 },
        { 9, 0x02, 93 },
        { 23, 0x02, 93 },
        { 40, 0x03, 93 },
        { 2, 0x02, 126 },
        { 9, 0x02, 126 },
        { 23, 0x02, 126 },
        { 40, 0x03, 126 },
        { 1, 0x02, 94 },
        { 22, 0x03, 94 },
        { 1, 0x02, 125 },
        { 22, 0x03, 125 },
        { 0, 0x03, 60 },
        { 0, 0x03, 96 },
        { 0, 0x03, 123 },
        { 95, 0x00, 0 },
    },
    /* 89 */
    {
        { 3, 0x02, 93 },
        { 6, 0x02, 93 },
        { 10, 0x02, 93 },
        { 15, 0x02, 93 },
        { 24, 0x02, 93 },
        { 31, 0x02, 93 },
        { 41, 0x02, 93 },
        { 56, 0x03, 93 },
        { 3, 0x02, 126 },
        { 6, 0x02, 126 },
        { 10, 0x02, 126 },
        { 15, 0x02, 126 },
        { 24, 0x02, 126 },
        { 31, 0x02, 126 },
        { 41, 0x02, 126 },
        { 56, 0x03, 126 },
    },
    /* 90 */
    {
        { 2, 0x02, 94 },
        { 9, 0x02, 94 },
        { 23, 0x02, 94 },
        { 40, 0x03, 94 },
        { 2, 0x02, 125 },
        { 9, 0x02, 125 },
        { 23, 0x02, 125 },
        { 40, 0x03, 125 },
        { 1, 0x02, 60 },
        { 22, 0x03, 60 },
        { 1, 0x02, 96 },
        { 22, 0x03, 96 },
        { 1, 0x02, 123 },
        { 22, 0x03, 123 },
        { 96, 0x00, 0 },
        { 110, 0x00, 0 },
    },
    /* 91 */
    {
        { 3, 0x02, 94 },
        { 6, 0x02, 94 },
        { 10, 0x02, 94 },
        { 15, 0x02, 94 },
        { 24, 0x02, 94 },
        { 31, 0x02, 94 },
        { 41, 0x02, 94 },
        { 56, 0x03, 94 },
        { 3, 0x02, 125 },
        { 6, 0x02, 125 },
        { 10, 0x02, 125 },
        { 15, 0x02, 125 },
        { 24, 0x02, 125 },
        { 31, 0x02, 125 },
        { 41, 0x02, 125 },
        { 56, 0x03, 125 },
    },
    /* 92 */
    {
        { 2, 0x02, 60 },
        { 9, 0x02, 60 },
        { 23, 0x02, 60 },
        { 40, 0x03, 60 },
        { 2, 0x02, 96 },
        { 9, 0x02, 96 },
        { 23, 0x02, 96 },
        { 40, 0x03, 96 },
        { 2, 0x02, 123 },
        { 9, 0x02, 123 },
        { 23, 0x02, 123 },
        { 40, 0x03, 123 },
        { 97, 0x00, 0 },
        { 101, 0x00, 0 },
        { 111, 0x00, 0 },
        { 133, 0x00, 0 },
    },
    /* 93 */
    {
        { 3, 0x02, 60 },
        { 6, 0x02, 60 },
        { 10, 0x02, 60 },
        { 15, 0x02, 60 },
        { 24, 0x02, 60 },
        { 31, 0x02, 60 },
        { 41, 0x02, 60 },
        { 56, 0x03, 60 },
        { 3, 0x02, 96 },
        { 6, 0x02, 96 },
        { 10, 0x02, 96 },
        { 15, 0x02, 96 },
        { 24, 0x02, 96 },
        { 31, 0x02, 96 },
        { 41, 0x02, 96 },
        { 56, 0x03, 96 },
    },
    /* 94 */
    {
        { 3, 0x02, 123 },
        { 6, 0x02, 123 },
        { 10, 0x02, 123 },
        { 15, 0x02, 123 },
        { 24, 0x02, 123 },
        { 31, 0x02, 123 },
        { 41, 0x02, 123 },
        { 56, 0x03, 123 },
        { 98, 0x00, 0 },
        { 99, 0x00, 0 },
        { 102, 0x00, 0 },
        { 105, 0x00, 0 },
        { 112, 0x00, 0 },
        { 119, 0x00, 0 },
        { 134, 0x00, 0 },
        { 153, 0x00, 0 },
    },
    /* 95 */
    {
        { 0, 0x03, 92 },
        { 0, 0x03, 195 },
        { 0, 0x03, 208 },
        { 100, 0x00, 0 },
        { 103, 0x00, 0 },
        { 104, 0x00, 0 },
        { 106, 0x00, 0 },
        { 107, 0x00, 0 },
        { 113, 0x00, 0 },
        { 116, 0x00, 0 },
        { 120, 0x00, 0 },
        { 126, 0x00, 0 },
        { 135, 0x00, 0 },
        { 142, 0x00, 0 },
        { 154, 0x00, 0 },
        { 169, 0x00, 0 },
    },
    /* 96 */
    {
        { 1, 0x02, 92 },
        { 22, 0x03, 92 },
        { 1, 0x02, 195 },
        { 22, 0x03, 195 },
        { 1, 0x02, 208 },
        { 22, 0x03, 208 },
        { 0, 0x03, 128 },
        { 0, 0x03, 130 },
        { 0, 0x03, 131 },
        { 0, 0x03, 162 },
        { 0, 0x03, 184 },
        { 0, 0x03, 194 },
        { 0, 0x03, 224 },
        { 0, 0x03, 226 },
        { 108, 0x00, 0 },
        { 109, 0x00, 0 },
    },
    /* 97 */
    {
        { 2, 0x02, 92 },
        { 9, 0x02, 92 },
        { 23, 0x02, 92 },
        { 40, 0x03, 92 },
        { 2, 0x02, 195 },
        { 9, 0x02, 195 },
        { 23, 0x02, 195 },
        { 40, 0x03, 195 },
        { 2, 0x02, 208 },
        { 9, 0x02, 208 },
        { 23, 0x02, 208 },
        { 40, 0x03, 208 },
        { 1, 0x02, 128 },
        { 22, 0x03, 128 },
        { 1, 0x02, 130 },
        { 22, 0x03, 130 },
    },
    /* 98 */
    {
        { 3, 0x02, 92 },
        { 6, 0x02, 92 },
        { 10, 0x02, 92 },
        { 15, 0x02, 92 },
        { 24, 0x02, 92 },
        { 31, 0x02, 92 },
        { 41, 0x02, 92 },
        { 56, 0x03, 92 },
        { 3, 0x02, 195 },
        { 6, 0x02, 195 },
        { 10, 0x02, 195 },
        { 15, 0x02, 195 },
        { 24, 0x02, 195 },
        { 31, 0x02, 195 },
        { 41, 0x02, 195 },
        { 56, 0x03, 195 },
    },
    /* 99 */
    {
        { 3, 0x02, 208 },
        { 6, 0x02, 208 },
        { 10, 0x02, 208 },
        { 15, 0x02, 208 },
        { 24, 0x02, 208 },
        { 31, 0x02, 208 },
        { 41, 0x02, 208 },
        { 56, 0x03, 208 },
        { 2, 0x02, 128 },
        { 9, 0x02, 128 },
        { 23, 0x02, 128 },
        { 40, 0x03, 128 },
        { 2, 0x02, 130 },
        { 9, 0x02, 130 },
        { 23, 0x02, 130 },
        { 40, 0x03, 130 },
    },
    /* 100 */
    {
        { 3, 0x02, 128 },
        { 6, 0x02, 128 },
        { 10, 0x02, 128 },
        { 15, 0x02, 128 },
        { 24, 0x02, 128 },
        { 31, 0x02, 128 },
        { 41, 0x02, 128 },
        { 56, 0x03, 128 },
        { 3, 0x02, 130 },
        { 6, 0x02, 130 },
        { 10, 0x02, 130 },
        { 15, 0x02, 130 },
        { 24, 0x02, 130 },
        { 31, 0x02, 130 },
        { 41, 0x02, 130 },
        { 56, 0x03, 130 },
    },
    /* 101 */
    {
        { 1, 0x02, 131 },
        { 22, 0x03, 131 },
        { 1, 0x02, 162 },
        { 22, 0x03, 162 },
        { 1, 0x02, 184 },
        { 22, 0x03, 184 },
        { 1, 0x02, 194 },
        { 22, 0x03, 194 },
        { 1, 0x02, 224 },
        { 22, 0x03, 224 },
        { 1, 0x02, 226 },
        { 22, 0x03, 226 },
        { 0, 0x03, 153 },
        { 0, 0x03, 161 },
        { 0, 0x03, 167 },
        { 0, 0x03, 172 },
    },
    /* 102 */
    {
        { 2, 0x02, 131 },
        { 9, 0x02, 131 },
        { 23, 0x02, 131 },
        { 40, 0x03, 131 },
        { 2, 0x02, 162 },
        { 9, 0x02, 162 },
        { 23, 0x02, 162 },
        { 40, 0x03, 162 },
        { 2, 0x02, 184 },
        { 9, 0x02, 184 },
        { 23, 0x02, 184 },
        { 40, 0x03, 184 },
        { 2, 0x02, 194 },
        { 9, 0x02, 194 },
        { 23, 0x02, 194 },
        { 40, 0x03, 194 },
    },
    /* 103 */
    {
        { 3, 0x02, 131 },
        { 6, 0x02, 131 },
        { 10, 0x02, 131 },
        { 15, 0x02, 131 },
        { 24, 0x02, 131 },
        { 31, 0x02, 131 },
        { 41, 0x02, 131 },
        { 56, 0x03, 131 },
        { 3, 0x02, 162 },
        { 6, 0x02, 162 },
        { 10, 0x02, 162 },
        { 15, 0x02, 162 },
        { 24, 0x02, 162 },
        { 31, 0x02, 162 },
        { 41, 0x02, 162 },
        { 56, 0x03, 162 },
    },
    /* 104 */
    {
        { 3, 0x02, 184 },
        { 6, 0x02, 184 },
        { 10, 0x02, 184 },
        { 15, 0x02, 184 },
        { 24, 0x02, 184 },
        { 31, 0x02, 184 },
        { 41, 0x02, 184 },
        { 56, 0x03, 184 },
        { 3, 0x02, 194 },
        { 6, 0x02, 194 },
        { 10, 0x02, 194 },
        { 15, 0x02, 194 },
        { 24, 0x02, 194 },
        { 31, 0x02, 194 },
        { 41, 0x02, 194 },
        { 56, 0x03, 194 },
    },
    /* 105 */
    {
        { 2, 0x02, 224 },
        { 9, 0x02, 224 },
        { 23, 0x02, 224 },
        { 40, 0x03, 224 },
        { 2, 0x02, 226 },
        { 9, 0x02, 226 },
        { 23, 0x02, 226 },
        { 40, 0x03, 226 },
        { 1, 0x02, 153 },
        { 22, 0x03, 153 },
        { 1, 0x02, 161 },
        { 22, 0x03, 161 },
        { 1, 0x02, 167 },
        { 22, 0x03, 167 },
        { 1, 0x02, 172 },
        { 22, 0x03, 172 },
    },
    /* 106 */
    {
        { 3, 0x02, 224 },
        { 6, 0x02, 224 },
        { 10, 0x02, 224 },
        { 15, 0x02, 224 },
        { 24, 0x02, 224 },
        { 31, 0x02, 224 },
        { 41, 0x02, 224 },
        { 56, 0x03, 224 },
        { 3, 0x02, 226 },
        { 6, 0x02, 226 },
        { 10, 0x02, 226 },
        { 15, 0x02, 226 },
        { 24, 0x02, 226 },
        { 31, 0x02, 226 },
        { 41, 0x02, 226 },
        { 56, 0x03, 226 },
    },
    /* 107 */
    {
        { 2, 0x02, 153 },
        { 9, 0x02, 153 },
        { 23, 0x02, 153 },
        { 40, 0x03, 153 },
        { 2, 0x02, 161 },
        { 9, 0x02, 161 },
        { 23, 0x02, 161 },
        { 40, 0x03, 161 },
        { 2, 0x02, 167 },
        { 9, 0x02, 167 },
        { 23, 0x02, 167 },
        { 40, 0x03, 167 },
        { 2, 0x02, 172 },
        { 9, 0x02, 172 },
        { 23, 0x02, 172 },
        { 40, 0x03, 172 },
    },
    /* 108 */
    {
        { 3, 0x02, 153 },
        { 6, 0x02, 153 },
        { 10, 0x02, 153 },
        { 15, 0x02, 153 },
        { 24, 0x02, 153 },
        { 31, 0x02, 153 },
        { 41, 0x02, 153 },
        { 56, 0x03, 153 },
        { 3, 0x02, 161 },
        { 6, 0x02, 161 },
        { 10, 0x02, 161 },
        { 15, 0x02, 161 },
        { 24, 0x02, 161 },
        { 31, 0x02, 161 },
        { 41, 0x02, 161 },
        { 56, 0x03, 161 },
    },
    /* 109 */
    {
        { 3, 0x02, 167 },
        { 6, 0x02, 167 },
        { 10, 0x02, 167 },
        { 15, 0x02, 167 },
        { 24, 0x02, 167 },
        { 31, 0x02, 167 },
        { 41, 0x02, 167 },
        { 56, 0x03, 167 },
        { 3, 0x02, 172 },
        { 6, 0x02, 172 },
        { 10, 0x02, 172 },
        { 15, 0x02, 172 },
        { 24, 0x02, 172 },
        { 31, 0x02, 172 },
        { 41, 0x02, 172 },
        { 56, 0x03, 172 },
    },
    /* 110 */
    {
        { 114, 0x00, 0 },
        { 115, 0x00, 0 },
        { 117, 0x00, 0 },
        { 118, 0x00, 0 },
        { 121, 0x00, 0 },
        { 123, 0x00, 0 },
        { 127, 0x00, 0 },
        { 130, 0x00, 0 },
        { 136, 0x00, 0 },
        { 139, 0x00, 0 },
        { 143, 0x00, 0 },
        { 146, 0x00, 0 },
        { 155, 0x00, 0 },
        { 162, 0x00, 0 },
        { 170, 0x00, 0 },
        { 180, 0x00, 0 },
    },
    /* 111 */
    {
        { 0, 0x03, 176 },
        { 0, 0x03, 177 },
        { 0, 0x03, 179 },
        { 0, 0x03, 209 },
        { 0, 0x03, 216 },
        { 0, 0x03, 217 },
        { 0, 0x03, 227 },
        { 0, 0x03, 229 },
        { 0, 0x03, 230 },
        { 122, 0x00, 0 },
        { 124, 0x00, 0 },
        { 125, 0x00, 0 },
        { 128, 0x00, 0 },
        { 129, 0x00, 0 },
        { 131, 0x00, 0 },
        { 132, 0x00, 0 },
    },
    /* 112 */
    {
        { 1, 0x02, 176 },
        { 22, 0x03, 176 },
        { 1, 0x02, 177 },
        { 22, 0x03, 177 },
        { 1, 0x02, 179 },
        { 22, 0x03, 179 },
        { 1, 0x02, 209 },
        { 22, 0x03, 209 },
        { 1, 0x02, 216 },
        { 22, 0x03, 216 },
        { 1, 0x02, 217 },
        { 22, 0x03, 217 },
        { 1, 0x02, 227 },
        { 22, 0x03, 227 },
        { 1, 0x02, 229 },
        { 22, 0x03, 229 },
    },
    /* 113 */
    {
        { 2, 0x02, 176 },
        { 9, 0x02, 176 },
        { 23, 0x02, 176 },
        { 40, 0x03, 176 },
        { 2, 0x02, 177 },
        { 9, 0x02, 177 },
        { 23, 0x02, 177 },
        { 40, 0x03, 177 },
        { 2, 0x02, 179 },
        { 9, 0x02, 179 },
        { 23, 0x02, 179 },
        { 40, 0x03, 179 },
        { 2, 0x02, 209 },
        { 9, 0x02, 209 },
        { 23, 0x02, 209 },
        { 40, 0x03, 209 },
    },
    /* 114 */
    {
        { 3, 0x02, 176 },
        { 6, 0x02, 176 },
        { 10, 0x02, 176 },
        { 15, 0x02, 176 },
        { 24, 0x02, 176 },
        { 31, 0x02, 176 },
        { 41, 0x02, 176 },
        { 56, 0x03, 176 },
        { 3, 0x02, 177 },
        { 6, 0x02, 177 },
        { 10, 0x02, 177 },
        { 15, 0x02, 177 },
        { 24, 0x02, 177 },
        { 31, 0x02, 177 },
        { 41, 0x02, 177 },
        { 56, 0x03, 177 },
    },
    /* 115 */
    {
        { 3, 0x02, 179 },
        { 6, 0x02, 179 },
        { 10, 0x02, 179 },
        { 15, 0x02, 179 },
        { 24, 0x02, 179 },
        { 31, 0x02, 179 },
        { 41, 0x02, 179 },
        { 56, 0x03, 179 },
        { 3, 0x02, 209 },
        { 6, 0x02, 209 },
        { 10, 0x02, 209 },
        { 15, 0x02, 209 },
        { 24, 0x02, 209 },
        { 31, 0x02, 209 },
        { 41, 0x02, 209 },
        { 56, 0x03, 209 },
    },
    /* 116 */
    {
        { 2, 0x02, 216 },
        { 9, 0x02, 216 },
        { 23, 0x02, 216 },
        { 40, 0x03, 216 },
        { 2, 0x02, 217 },
        { 9, 0x02, 217 },
        { 23, 0x02, 217 },
        { 40, 0x03, 217 },
        { 2, 0x02, 227 },
        { 9, 0x02, 227 },
        { 23, 0x02, 227 },
        { 40, 0x03, 227 },
        { 2, 0x02, 229 },
        { 9, 0x02, 229 },
        { 23, 0x02, 229 },
        { 40, 0x03, 229 },
    },
    /* 117 */
    {
        { 3, 0x02, 216 },
        { 6, 0x02, 216 },
        { 10, 0x02, 216 },
        { 15, 0x02, 216 },
        { 24, 0x02, 216 },
        { 31, 0x02, 216 },
        { 41, 0x02, 216 },
        { 56, 0x03, 216 },
        { 3, 0x02, 217 },
        { 6, 0x02, 217 },
        { 10, 0x02, 217 },
        { 15, 0x02, 217 },
        { 24, 0x02, 217 },
        { 31, 0x02, 217 },
        { 41, 0x02, 217 },
        { 56, 0x03, 217 },
    },
    /* 118 */
    {
        { 3, 0x02, 227 },
        { 6, 0x02, 227 },
        { 10, 0x02, 227 },
        { 15, 0x02, 227 },
        { 24, 0x02, 227 },
        { 31, 0x02, 227 },
        { 41, 0x02, 227 },
        { 56, 0x03, 227 },
        { 3, 0x02, 229 },
        { 6, 0x02, 229 },
        { 10, 0x02, 229 },
        { 15, 0x02, 229 },
        { 24, 0x02, 229 },
        { 31, 0x02, 229 },
        { 41, 0x02, 229 },
        { 56, 0x03, 229 },
    },
    /* 119 */
    {
        { 1, 0x02, 230 },
        { 22, 0x03, 230 },
        { 0, 0x03, 129 },
        { 0, 0x03, 132 },
        { 0, 0x03, 133 },
        { 0, 0x03, 134 },
        { 0, 0x03, 136 },
        { 0, 0x03, 146 },
        { 0, 0x03, 154 },
        { 0, 0x03, 156 },
        { 0, 0x03, 160 },
        { 0, 0x03, 163 },
        { 0, 0x03, 164 },
        { 0, 0x03, 169 },
        { 0, 0x03, 170 },
        { 0, 0x03, 173 },
    },
    /* 120 */
    {
        { 2, 0x02, 230 },
        { 9, 0x02, 230 },
        { 23, 0x02, 230 },
        { 40, 0x03, 230 },
        { 1, 0x02, 129 },
        { 22, 0x03, 129 },
        { 1, 0x02, 132 },
        { 22, 0x03, 132 },
        { 1, 0x02, 133 },
        { 22, 0x03, 133 },
        { 1, 0x02, 134 },
        { 22, 0x03, 134 },
        { 1, 0x02, 136 },
        { 22, 0x03, 136 },
        { 1, 0x02, 146 },
        { 22, 0x03, 146 },
    },
    /* 121 */
    {
        { 3, 0x02, 230 },
        { 6, 0x02, 230 },
        { 10, 0x02, 230 },
        { 15, 0x02, 230 },
        { 24, 0x02, 230 },
        { 31, 0x02, 230 },
        { 41, 0x02, 230 },
        { 56, 0x03, 230 },
        { 2, 0x02, 129 },
        { 9, 0x02, 129 },
        { 23, 0x02, 129 },
        { 40, 0x03, 129 },
        { 2, 0x02, 132 },
        { 9, 0x02, 132 },
        { 23, 0x02, 132 },
        { 40, 0x03, 132 },
    },
    /* 122 */
    {
        { 3, 0x02, 129 },
        { 6, 0x02, 129 },
        { 10, 0x02, 129 },
        { 15, 0x02, 129 },
        { 24, 0x02, 129 },
        { 31, 0x02, 129 },
        { 41, 0x02, 129 },
        { 56, 0x03, 129 },
        { 3, 0x02, 132 },
        { 6, 0x02, 132 },
        { 10, 0x02, 132 },
        { 15, 0x02, 132 },
        { 24, 0x02, 132 },
        { 31, 0x02, 132 },
        { 41, 0x02, 132 },
        { 56, 0x03, 132 },
    },
    /* 123 */
    {
        { 2, 0x02, 133 },
        { 9, 0x02, 133 },
        { 23, 0x02, 133 },
        { 40, 0x03, 133 },
        { 2, 0x02, 134 },
        { 9, 0x02, 134 },
        { 23, 0x02, 134 },
        { 40, 0x03, 134 },
        { 2, 0x02, 136 },
        { 9, 0x02, 136 },
        { 23, 0x02, 136 },
        { 40, 0x03, 136 },
        { 2, 0x02, 146 },
        { 9, 0x02, 146 },
        { 23, 0x02, 146 },
        { 40, 0x03, 146 },
    },
    /* 124 */
    {
        { 3, 0x02, 133 },
        { 6, 0x02, 133 },
        { 10, 0x02, 133 },
        { 15, 0x02, 133 },
        { 24, 0x02, 133 },
        { 31, 0x02, 133 },
        { 41, 0x02, 133 },
        { 56, 0x03, 133 },
        { 3, 0x02, 134 },
        { 6, 0x02, 134 },
        { 10, 0x02, 134 },
        { 15, 0x02, 134 },
        { 24, 0x02, 134 },
        { 31, 0x02, 134 },
        { 41, 0x02, 134 },
        { 56, 0x03, 134 },
    },
    /* 125 */
    {
        { 3, 0x02, 136 },
        { 6, 0x02, 136 },
        { 10, 0x02, 136 },
        { 15, 0x02, 136 },
        { 24, 0x02, 136 },
        { 31, 0x02, 136 },
        { 41, 0x02, 136 },
        { 56, 0x03, 136 },
        { 3, 0x02, 146 },
        { 6, 0x02, 146 },
        { 10, 0x02, 146 },
        { 15, 0x02, 146 },
        { 24, 0x02, 146 },
        { 31, 0x02, 146 },
        { 41, 0x02, 146 },
        { 56, 0x03, 146 },
    },
    /* 126 */
    {
        { 1, 0x02, 154 },
        { 22, 0x03, 154 },
        { 1, 0x02, 156 },
        { 22, 0x03, 156 },
        { 1, 0x02, 160 },
        { 22, 0x03, 160 },
        { 1, 0x02, 163 },
        { 22, 0x03, 163 },
        { 1, 0x02, 164 },
        { 22, 0x03, 164 },
        { 1, 0x02, 169 },
        { 22, 0x03, 169 },
        { 1, 0x02, 170 },
        { 22, 0x03, 170 },
        { 1, 0x02, 173 },
        { 22, 0x03, 173 },
    },
    /* 127 */
    {
        { 2, 0x02, 154 },
        { 9, 0x02, 154 },
        { 23, 0x02, 154 },
        { 40, 0x03, 154 },
        { 2, 0x02, 156 },
        { 9, 0x02, 156 },
        { 23, 0x02, 156 },
        { 40, 0x03, 156 },
        { 2, 0x02, 160 },
        { 9, 0x02, 160 },
        { 23, 0x02, 160 },
        { 40, 0x03, 160 },
        { 2, 0x02, 163 },
        { 9, 0x02, 163 },
        { 23, 0x02, 163 },
        { 40, 0x03, 163 },
    },
    /* 128 */
    {
        { 3, 0x02, 154 },
        { 6, 0x02, 154 },
        { 10, 0x02, 154 },
        { 15, 0x02, 154 },
        { 24, 0x02, 154 },
        { 31, 0x02, 154 },
        { 41, 0x02, 154 },
        { 56, 0x03, 154 },
        { 3, 0x02, 156 },
        { 6, 0x02, 156 },
        { 10, 0x02, 156 },
        { 15, 0x02, 156 },
        { 24, 0x02, 156 },
        { 31, 0x02, 156 },
        { 41, 0x02, 156 },
        { 56, 0x03, 156 },
    },
    /* 129 */
    {
        { 3, 0x02, 160 },
        { 6, 0x02, 160 },
        { 10, 0x02, 160 },
        { 15, 0x02, 160 },
        { 24, 0x02, 160 },
        { 31, 0x02, 160 },
        { 41, 0x02, 160 },
        { 56, 0x03, 160 },
        { 3, 0x02, 163 },
        { 6, 0x02, 163 },
        { 10, 0x02, 163 },
        { 15, 0x02, 163 },
        { 24, 0x02, 163 },
        { 31, 0x02, 163 },
        { 41, 0x02, 163 },
        { 56, 0x03, 163 },
    },
    /* 130 */
    {
        { 2, 0x02, 164 },
        { 9, 0x02, 164 },
        { 23, 0x02, 164 },
        { 40, 0x03, 164 },
        { 2, 0x02, 169 },
        { 9, 0x02, 169 },
        { 23, 0x02, 169 },
        { 40, 0x03, 169 },
        { 2, 0x02, 170 },
        { 9, 0x02, 170 },
        { 23, 0x02, 170 },
        { 40, 0x03, 170 },
        { 2, 0x02, 173 },
        { 9, 0x02, 173 },
        { 23, 0x02, 173 },
        { 40, 0x03, 173 },
    },
    /* 131 */
    {
        { 3, 0x02, 164 },
        { 6, 0x02, 164 },
        { 10, 0x02, 164 },
        { 15, 0x02, 164 },
        { 24, 0x02, 164 },
        { 31, 0x02, 164 },
        { 41, 0x02, 164 },
        { 56, 0x03, 164 },
        { 3, 0x02, 169 },
        { 6, 0x02, 169 },
        { 10, 0x02, 169 },
        { 15, 0x02, 169 },
        { 24, 0x02, 169 },
        { 31, 0x02, 169 },
        { 41, 0x02, 169 },
        { 56, 0x03, 169 },
    },
    /* 132 */
    {
        { 3, 0x02, 170 },
        { 6, 0x02, 170 },
        { 10, 0x02, 170 },
        { 15, 0x02, 170 },
        { 24, 0x02, 170 },
        { 31, 0x02, 170 },
        { 41, 0x02, 170 },
        { 56, 0x03, 170 },
        { 3, 0x02, 173 },
        { 6, 0x02, 173 },
        { 10, 0x02, 173 },
        { 15, 0x02, 173 },
        { 24, 0x02, 173 },
        { 31, 0x02, 173 },
        { 41, 0x02, 173 },
        { 56, 0x03, 173 },
    },
    /* 133 */
    {
        { 137, 0x00, 0 },
        { 138, 0x00, 0 },
        { 140, 0x00, 0 },
        { 141, 0x00, 0 },
        { 144, 0x00, 0 },
        { 145, 0x00, 0 },
        { 147, 0x00, 0 },
        { 150, 0x00, 0 },
        { 156, 0x00, 0 },
        { 159, 0x00, 0 },
        { 163, 0x00, 0 },
        { 166, 0x00, 0 },
        { 171, 0x00, 0 },
        { 174, 0x00, 0 },
        { 181, 0x00, 0 },
        { 190, 0x00, 0 },
    },
    /* 134 */
    {
        { 0, 0x03, 178 },
        { 0, 0x03, 181 },
        { 0, 0x03, 185 },
        { 0, 0x03, 186 },
        { 0, 0x03, 187 },
        { 0, 0x03, 189 },
        { 0, 0x03, 190 },
        { 0, 0x03, 196 },
        { 0, 0x03, 198 },
        { 0, 0x03, 228 },
        { 0, 0x03, 232 },
        { 0, 0x03, 233 },
        { 148, 0x00, 0 },
        { 149, 0x00, 0 },
        { 151, 0x00, 0 },
        { 152, 0x00, 0 },
    },
    /* 135 */
    {
        { 1, 0x02, 178 },
        { 22, 0x03, 178 },
        { 1, 0x02, 181 },
        { 22, 0x03, 181 },
        { 1, 0x02, 185 },
        { 22, 0x03, 185 },
        { 1, 0x02, 186 },
        { 22, 0x03, 186 },
        { 1, 0x02, 187 },
        { 22, 0x03, 187 },
        { 1, 0x02, 189 },
        { 22, 0x03, 189 },
        { 1, 0x02, 190 },
        { 22, 0x03, 190 },
        { 1, 0x02, 196 },
        { 22, 0x03, 196 },
    },
    /* 136 */
    {
        { 2, 0x02, 178 },
        { 9, 0x02, 178 },
        { 23, 0x02, 178 },
        { 40, 0x03, 178 },
        { 2, 0x02, 181 },
        { 9, 0x02, 181 },
        { 23, 0x02, 181 },
        { 40, 0x03, 181 },
        { 2, 0x02, 185 },
        { 9, 0x02, 185 },
        { 23, 0x02, 185 },
        { 40, 0x03, 185 },
        { 2, 0x02, 186 },
        { 9, 0x02, 186 },
        { 23, 0x02, 186 },
        { 40, 0x03, 186 },
    },
    /* 137 */
    {
        { 3, 0x02, 178 },
        { 6, 0x02, 178 },
        { 10, 0x02, 178 },
        { 15, 0x02, 178 },
        { 24, 0x02, 178 },
        { 31, 0x02, 178 },
        { 41, 0x02, 178 },
        { 56, 0x03, 178 },
        { 3, 0x02, 181 },
        { 6, 0x02, 181 },
        { 10, 0x02, 181 },
        { 15, 0x02, 181 },
        { 24, 0x02, 181 },
        { 31, 0x02, 181 },
        { 41, 0x02, 181 },
        { 56, 0x03, 181 },
    },
    /* 138 */
    {
        { 3, 0x02, 185 },
        { 6, 0x02, 185 },
        { 10, 0x02, 185 },
        { 15, 0x02, 185 },
        { 24, 0x02, 185 },
        { 31, 0x02, 185 },
        { 41, 0x02, 185 },
        { 56, 0x03, 185 },
        { 3, 0x02, 186 },
        { 6, 0x02, 186 },
        { 10, 0x02, 186 },
        { 15, 0x02, 186 },
        { 24, 0x02, 186 },
        { 31, 0x02, 186 },
        { 41, 0x02, 186 },
        { 56, 0x03, 186 },
    },
    /* 139 */
    {
        { 2, 0x02, 187 },
        { 9, 0x02, 187 },
        { 23, 0x02, 187 },
        { 40, 0x03, 187 },
        { 2, 0x02, 189 },
        { 9, 0x02, 189 },
        { 23, 0x02, 189 },
        { 40, 0x03, 189 },
        { 2, 0x02, 190 },
        { 9, 0x02, 190 },
        { 23, 0x02, 190 },
        { 40, 0x03, 190 },
        { 2, 0x02, 196 },
        { 9, 0x02, 196 },
        { 23, 0x02, 196 },
        { 40, 0x03, 196 },
    },
    /* 140 */
    {
        { 3, 0x02, 187 },
        { 6, 0x02, 187 },
        { 10, 0x02, 187 },
        { 15, 0x02, 187 },
        { 24, 0x02, 187 },
        { 31, 0x02, 187 },
        { 41, 0x02, 187 },
        { 56, 0x03, 187 },
        { 3, 0x02, 189 },
        { 6, 0x02, 189 },
        { 10, 0x02, 189 },
        { 15, 0x02, 189 },
        { 24, 0x02, 189 },
        { 31, 0x02, 189 },
        { 41, 0x02, 189 },
        { 56, 0x03, 189 },
    },
    /* 141 */
    {
        { 3, 0x02, 190 },
        { 6, 0x02, 190 },
        { 10, 0x02, 190 },
        { 15, 0x02, 190 },
        { 24, 0x02, 190 },
        { 31, 0x02, 190 },
        { 41, 0x02, 190 },
        { 56, 0x03, 190 },
        { 3, 0x02, 196 },
        { 6, 0x02, 196 },
        { 10, 0x02, 196 },
        { 15, 0x02, 196 },
        { 24, 0x02, 196 },
        { 31, 0x02, 196 },
        { 41, 0x02, 196 },
        { 56, 0x03, 196 },
    },
    /* 142 */
    {
        { 1, 0x02, 198 },
        { 22, 0x03, 198 },
        { 1, 0x02, 228 },
        { 22, 0x03, 228 },
        { 1, 0x02, 232 },
        { 22, 0x03, 232 },
        { 1, 0x02, 233 },
        { 22, 0x03, 233 },
        { 0, 0x03, 1 },
        { 0, 0x03, 135 },
        { 0, 0x03, 137 },
        { 0, 0x03, 138 },
        { 0, 0x03, 139 },
        { 0, 0x03, 140 },
        { 0, 0x03, 141 },
        { 0, 0x03, 143 },
    },
    /* 143 */
    {
        { 2, 0x02, 198 },
        { 9, 0x02, 198 },
        { 23, 0x02, 198 },
        { 40, 0x03, 198 },
        { 2, 0x02, 228 },
        { 9, 0x02, 228 },
        { 23, 0x02, 228 },
        { 40, 0x03, 228 },
        { 2, 0x02, 232 },
        { 9, 0x02, 232 },
        { 23, 0x02, 232 },
        { 40, 0x03, 232 },
        { 2, 0x02, 233 },
        { 9, 0x02, 233 },
        { 23, 0x02, 233 },
        { 40, 0x03, 233 },
    },
    /* 144 */
    {
        { 3, 0x02, 198 },
        { 6, 0x02, 198 },
        { 10, 0x02, 198 },
        { 15, 0x02, 198 },
        { 24, 0x02, 198 },
        { 31, 0x02, 198 },
        { 41, 0x02, 198 },
        { 56, 0x03, 198 },
        { 3, 0x02, 228 },
        { 6, 0x02, 228 },
        { 10, 0x02, 228 },
        { 15, 0x02, 228 },
        { 24, 0x02, 228 },
        { 31, 0x02, 228 },
        { 41, 0x02, 228 },
        { 56, 0x03, 228 },
    },
    /* 145 */
    {
        { 3, 0x02, 232 },
        { 6, 0x02, 232 },
        { 10, 0x02, 232 },
        { 15, 0x02, 232 },
        { 24, 0x02, 232 },
        { 31, 0x02, 232 },
        { 41, 0x02, 232 },
        { 56, 0x03, 232 },
        { 3, 0x02, 233 },
        { 6, 0x02, 233 },
        { 10, 0x02, 233 },
        { 15, 0x02, 233 },
        { 24, 0x02, 233 },
        { 31, 0x02, 233 },
        { 41, 0x02, 233 },
        { 56, 0x03, 233 },
    },
    /* 146 */
    {
        { 1, 0x02, 1 },
        { 22, 0x03, 1 },
        { 1, 0x02, 135 },
        { 22, 0x03, 135 },
        { 1, 0x02, 137 },
        { 22, 0x03, 137 },
        { 1, 0x02, 138 },
        { 22, 0x03, 138 },
        { 1, 0x02, 139 },
        { 22, 0x03, 139 },
        { 1, 0x02, 140 },
        { 22, 0x03, 140 },
        { 1, 0x02, 141 },
        { 22, 0x03, 141 },
        { 1, 0x02, 143 },
        { 22, 0x03, 143 },
    },
    /* 147 */
    {
        { 2, 0x02, 1 },
        { 9, 0x02, 1 },
        { 23, 0x02, 1 },
        { 40, 0x03, 1 },
        { 2, 0x02, 135 },
        { 9, 0x02, 135 },
        { 23, 0x02, 135 },
        { 40, 0x03, 135 },
        { 2, 0x02, 137 },
        { 9, 0x02, 137 },
        { 23, 0x02, 137 },
        { 40, 0x03, 137 },
        { 2, 0x02, 138 },
        { 9, 0x02, 138 },
        { 23, 0x02, 138 },
        { 40, 0x03, 138 },
    },
    /* 148 */
    {
        { 3, 0x02, 1 },
        { 6, 0x02, 1 },
        { 10, 0x02, 1 },
        { 15, 0x02, 1 },
        { 24, 0x02, 1 },
        { 31, 0x02, 1 },
        { 41, 0x02, 1 },
        { 56, 0x03, 1 },
        { 3, 0x02, 135 },
        { 6, 0x02, 135 },
        { 10, 0x02, 135 },
        { 15, 0x02, 135 },
        { 24, 0x02, 135 },
        { 31, 0x02, 135 },
        { 41, 0x02, 135 },
        { 56, 0x03, 135 },
    },
    /* 149 */
    {
        { 3, 0x02, 137 },
        { 6, 0x02, 137 },
        { 10, 0x02, 137 },
        { 15, 0x02, 137 },
        { 24, 0x02, 137 },
        { 31, 0x02, 137 },
        { 41, 0x02, 137 },
        { 56, 0x03, 137 },
        { 3, 0x02, 138 },
        { 6, 0x02, 138 },
        { 10, 0x02, 138 },
        { 15, 0x02, 138 },
        { 24, 0x02, 138 },
        { 31, 0x02, 138 },
        { 41, 0x02, 138 },
        { 56, 0x03, 138 },
    },
    /* 150 */
    {
        { 2, 0x02, 139 },
        { 9, 0x02, 139 },
        { 23, 0x02, 139 },
        { 40, 0x03, 139 },
        { 2, 0x02, 140 },
        { 9, 0x02, 140 },
        { 23, 0x02, 140 },
        { 40, 0x03, 140 },
        { 2, 0x02, 141 },
        { 9, 0x02, 141 },
        { 23, 0x02, 141 },
        { 40, 0x03, 141 },
        { 2, 0x02, 143 },
        { 9, 0x02, 143 },
        { 23, 0x02, 143 },
        { 40, 0x03, 143 },
    },
    /* 151 */
    {
        { 3, 0x02, 139 },
        { 6, 0x02, 139 },
        { 10, 0x02, 139 },
        { 15, 0x02, 139 },
        { 24, 0x02, 139 },
        { 31, 0x02, 139 },
        { 41, 0x02, 139 },
        { 56, 0x03, 139 },
        { 3, 0x02, 140 },
        { 6, 0x02, 140 },
        { 10, 0x02, 140 },
        { 15, 0x02, 140 },
        { 24, 0x02, 140 },
        { 31, 0x02, 140 },
        { 41, 0x02, 140 },
        { 56, 0x03, 140 },
    },
    /* 152 */
    {
        { 3, 0x02, 141 },
        { 6, 0x02, 141 },
        { 10, 0x02, 141 },
        { 15, 0x02, 141 },
        { 24, 0x02, 141 },
        { 31, 0x02, 141 },
        { 41, 0x02, 141 },
        { 56, 0x03, 141 },
        { 3, 0x02, 143 },
        { 6, 0x02, 143 },
        { 10, 0x02, 143 },
        { 15, 0x02, 143 },
        { 24, 0x02, 143 },
        { 31, 0x02, 143 },
        { 41, 0x02, 143 },
        { 56, 0x03, 143 },
    },
    /* 153 */
    {
        { 157, 0x00, 0 },
        { 158, 0x00, 0 },
        { 160, 0x00, 0 },
        { 161, 0x00, 0 },
        { 164, 0x00, 0 },
        { 165, 0x00, 0 },
        { 167, 0x00, 0 },
        { 168, 0x00, 0 },
        { 172, 0x00, 0 },
        { 173, 0x00, 0 },
        { 175, 0x00, 0 },
        { 177, 0x00, 0 },
        { 182, 0x00, 0 },
        { 185, 0x00, 0 },
        { 191, 0x00, 0 },
        { 207, 0x00, 0 },
    },
    /* 154 */
    {
        { 0, 0x03, 147 },
        { 0, 0x03, 149 },
        { 0, 0x03, 150 },
        { 0, 0x03, 151 },
        { 0, 0x03, 152 },
        { 0, 0x03, 155 },
        { 0, 0x03, 157 },
        { 0, 0x03, 158 },
        { 0, 0x03, 165 },
        { 0, 0x03, 166 },
        { 0, 0x03, 168 },
        { 0, 0x03, 174 },
        { 0, 0x03, 175 },
        { 0, 0x03, 180 },
        { 0, 0x03, 182 },
        { 0, 0x03, 183 },
    },
    /* 155 */
    {
        { 1, 0x02, 147 },
        { 22, 0x03, 147 },
        { 1, 0x02, 149 },
        { 22, 0x03, 149 },
        { 1, 0x02, 150 },
        { 22, 0x03, 150 },
        { 1, 0x02, 151 },
        { 22, 0x03, 151 },
        { 1, 0x02, 152 },
        { 22, 0x03, 152 },
        { 1, 0x02, 155 },
        { 22, 0x03, 155 },
        { 1, 0x02, 157 },
        { 22, 0x03, 157 },
        { 1, 0x02, 158 },
        { 22, 0x03, 158 },
    },
    /* 156 */
    {
        { 2, 0x02, 147 },
        { 9, 0x02, 147 },
        { 23, 0x02, 147 },
        { 40, 0x03, 147 },
        { 2, 0x02, 149 },
        { 9, 0x02, 149 },
        { 23, 0x02, 149 },
        { 40, 0x03, 149 },
        { 2, 0x02, 150 },
        { 9, 0x02, 150 },
        { 23, 0x02, 150 },
        { 40, 0x03, 150 },
        { 2, 0x02, 151 },
        { 9, 0x02, 151 },
        { 23, 0x02, 151 },
        { 40, 0x03, 151 },
    },
    /* 157 */
    {
        { 3, 0x02, 147 },
        { 6, 0x02, 147 },
        { 10, 0x02, 147 },
        { 15, 0x02, 147 },
        { 24, 0x02, 147 },
        { 31, 0x02, 147 },
        { 41, 0x02, 147 },
        { 56, 0x03, 147 },
        { 3, 0x02, 149 },
        { 6, 0x02, 149 },
        { 10, 0x02, 149 },
        { 15, 0x02, 149 },
        { 24, 0x02, 149 },
        { 31, 0x02, 149 },
        { 41, 0x02, 149 },
        { 56, 0x03, 149 },
    },
    /* 158 */
    {
        { 3, 0x02, 150 },
        { 6, 0x02, 150 },
        { 10, 0x02, 150 },
        { 15, 0x02, 150 },
        { 24, 0x02, 150 },
        { 31, 0x02, 150 },
        { 41, 0x02, 150 },
        { 56, 0x03, 150 },
        { 3, 0x02, 151 },
        { 6, 0x02, 151 },
        { 10, 0x02, 151 },
        { 15, 0x02, 151 },
        { 24, 0x02, 151 },
        { 31, 0x02, 151 },
        { 41, 0x02, 151 },
        { 56, 0x03, 151 },
    },
    /* 159 */
    {
        { 2, 0x02, 152 },
        { 9, 0x02, 152 },
        { 23, 0x02, 152 },
        { 40, 0x03, 152 },
        { 2, 0x02, 155 },
        { 9, 0x02, 155 },
        { 23, 0x02, 155 },
        { 40, 0x03, 155 },
        { 2, 0x02, 157 },
        { 9, 0x02, 157 },
        { 23, 0x02, 157 },
        { 40, 0x03, 157 },
        { 2, 0x02, 158 },
        { 9, 0x02, 158 },
        { 23, 0x02, 158 },
        { 40, 0x03, 158 },
    },
    /* 160 */
    {
        { 3, 0x02, 152 },
        { 6, 0x02, 152 },
        { 10, 0x02, 152 },
        { 15, 0x02, 152 },
        { 24, 0x02, 152 },
        { 31, 0x02, 152 },
        { 41, 0x02, 152 },
        { 56, 0x03, 152 },
        { 3, 0x02, 155 },
        { 6, 0x02, 155 },
        { 10, 0x02, 155 },
        { 15, 0x02, 155 },
        { 24, 0x02, 155 },
        { 31, 0x02, 155 },
        { 41, 0x02, 155 },
        { 56, 0x03, 155 },
    },
    /* 161 */
    {
        { 3, 0x02, 157 },
        { 6, 0x02, 157 },
        { 10, 0x02, 157 },
        { 15, 0x02, 157 },
        { 24, 0x02, 157 },
        { 31, 0x02, 157 },
        { 41, 0x02, 157 },
        { 56, 0x03, 157 },
        { 3, 0x02, 158 },
        { 6, 0x02, 158 },
        { 10, 0x02, 158 },
        { 15, 0x02, 158 },
        { 24, 0x02, 158 },
        { 31, 0x02, 158 },
        { 41, 0x02, 158 },
        { 56, 0x03, 158 },
    },
    /* 162 */
    {
        { 1, 0x02, 165 },
        { 22, 0x03, 165 },
        { 1, 0x02, 166 },
        { 22, 0x03, 166 },
        { 1, 0x02, 168 },
        { 22, 0x03, 168 },
        { 1, 0x02, 174 },
        { 22, 0x03, 174 },
        { 1, 0x02, 175 },
        { 22, 0x03, 175 },
        { 1, 0x02, 180 },
        { 22, 0x03, 180 },
        { 1, 0x02, 182 },
        { 22, 0x03, 182 },
        { 1, 0x02, 183 },
        { 22, 0x03, 183 },
    },
    /* 163 */
    {
        { 2, 0x02, 165 },
        { 9, 0x02, 165 },
        { 23, 0x02, 165 },
        { 40, 0x03, 165 },
        { 2, 0x02, 166 },
        { 9, 0x02, 166 },
        { 23, 0x02, 166 },
        { 40, 0x03, 166 },
        { 2, 0x02, 168 },
        { 9, 0x02, 168 },
        { 23, 0x02, 168 },
        { 40, 0x03, 168 },
        { 2, 0x02, 174 },
        { 9, 0x02, 174 },
        { 23, 0x02, 174 },
        { 40, 0x03, 174 },
    },
    /* 164 */
    {
        { 3, 0x02, 165 },
        { 6, 0x02, 165 },
        { 10, 0x02, 165 },
        { 15, 0x02, 165 },
        { 24, 0x02, 165 },
        { 31, 0x02, 165 },
        { 41, 0x02, 165 },
        { 56, 0x03, 165 },
        { 3, 0x02, 166 },
        { 6, 0x02, 166 },
        { 10, 0x02, 166 },
        { 15, 0x02, 166 },
        { 24, 0x02, 166 },
        { 31, 0x02, 166 },
        { 41, 0x02, 166 },
        { 56, 0x03, 166 },
    },
    /* 165 */
    {
        { 3, 0x02, 168 },
        { 6, 0x02, 168 },
        { 10, 0x02, 168 },
        { 15, 0x02, 168 },
        { 24, 0x02, 168 },
        { 31, 0x02, 168 },
        { 41, 0x02, 168 },
        { 56, 0x03, 168 },
        { 3, 0x02, 174 },
        { 6, 0x02, 174 },
        { 10, 0x02, 174 },
        { 15, 0x02, 174 },
        { 24, 0x02, 174 },
        { 31, 0x02, 174 },
        { 41, 0x02, 174 },
        { 56, 0x03, 174 },
    },
    /* 166 */
    {
        { 2, 0x02, 175 },
        { 9, 0x02, 175 },
        { 23, 0x02, 175 },
        { 40, 0x03, 175 },
        { 2, 0x02, 180 },
        { 9, 0x02, 180 },
        { 23, 0x02, 180 },
        { 40, 0x03, 180 },
        { 2, 0x02, 182 },
        { 9, 0x02, 182 },
        { 23, 0x02, 182 },
        { 40, 0x03, 182 },
        { 2, 0x02, 183 },
        { 9, 0x02, 183 },
        { 23, 0x02, 183 },
        { 40, 0x03, 183 },
    },
    /* 167 */
    {
        { 3, 0x02, 175 },
        { 6, 0x02, 175 },
        { 10, 0x02, 175 },
        { 15, 0x02, 175 },
        { 24, 0x02, 175 },
        { 31, 0x02, 175 },
        { 41, 0x02, 175 },
        { 56, 0x03, 175 },
        { 3, 0x02, 180 },
        { 6, 0x02, 180 },
        { 10, 0x02, 180 },
        { 15, 0x02, 180 },
        { 24, 0x02, 180 },
        { 31, 0x02, 180 },
        { 41, 0x02, 180 },
        { 56, 0x03, 180 },
    },
    /* 168 */
    {
        { 3, 0x02, 182 },
        { 6, 0x02, 182 },
        { 10, 0x02, 182 },
        { 15, 0x02, 182 },
        { 24, 0x02, 182 },
        { 31, 0x02, 182 },
        { 41, 0x02, 182 },
        { 56, 0x03, 182 },
        { 3, 0x02, 183 },
        { 6, 0x02, 183 },
        { 10, 0x02, 183 },
        { 15, 0x02, 183 },
        { 24, 0x02, 183 },
        { 31, 0x02, 183 },
        { 41, 0x02, 183 },
        { 56, 0x03, 183 },
    },
    /* 169 */
    {
        { 0, 0x03, 188 },
        { 0, 0x03, 191 },
        { 0, 0x03, 197 },
        { 0, 0x03, 231 },
        { 0, 0x03, 239 },
        { 176, 0x00, 0 },
        { 178, 0x00, 0 },
        { 179, 0x00, 0 },
        { 183, 0x00, 0 },
        { 184, 0x00, 0 },
        { 186, 0x00, 0 },
        { 187, 0x00, 0 },
        { 192, 0x00, 0 },
        { 199, 0x00, 0 },
        { 208, 0x00, 0 },
        { 223, 0x00, 0 },
    },
    /* 170 */
    {
        { 1, 0x02, 188 },
        { 22, 0x03, 188 },
        { 1, 0x02, 191 },
        { 22, 0x03, 191 },
        { 1, 0x02, 197 },
        { 22, 0x03, 197 },
        { 1, 0x02, 231 },
        { 22, 0x03, 231 },
        { 1, 0x02, 239 },
        { 22, 0x03, 239 },
        { 0, 0x03, 9 },
        { 0, 0x03, 142 },
        { 0, 0x03, 144 },
        { 0, 0x03, 145 },
        { 0, 0x03, 148 },
        { 0, 0x03, 159 },
    },
    /* 171 */
    {
        { 2, 0x02, 188 },
        { 9, 0x02, 188 },
        { 23, 0x02, 188 },
        { 40, 0x03, 188 },
        { 2, 0x02, 191 },
        { 9, 0x02, 191 },
        { 23, 0x02, 191 },
        { 40, 0x03, 191 },
        { 2, 0x02, 197 },
        { 9, 0x02, 197 },
        { 23, 0x02, 197 },
        { 40, 0x03, 197 },
        { 2, 0x02, 231 },
        { 9, 0x02, 231 },
        { 23, 0x02, 231 },
        { 40, 0x03, 231 },
    },
    /* 172 */
    {
        { 3, 0x02, 188 },
        { 6, 0x02, 188 },
        { 10, 0x02, 188 },
        { 15, 0x02, 188 },
        { 24, 0x02, 188 },
        { 31, 0x02, 188 },
        { 41, 0x02, 188 },
        { 56, 0x03, 188 },
        { 3, 0x02, 191 },
        { 6, 0x02, 191 },
        { 10, 0x02, 191 },
        { 15, 0x02, 191 },
        { 24, 0x02, 191 },
        { 31, 0x02, 191 },
        { 41, 0x02, 191 },
        { 56, 0x03, 191 },
    },
    /* 173 */
    {
        { 3, 0x02, 197 },
        { 6, 0x02, 197 },
        { 10, 0x02, 197 },
        { 15, 0x02, 197 },
        { 24, 0x02, 197 },
        { 31, 0x02, 197 },
        { 41, 0x02, 197 },
        { 56, 0x03, 197 },
        { 3, 0x02, 231 },
        { 6, 0x02, 231 },
        { 10, 0x02, 231 },
        { 15, 0x02, 231 },
        { 24, 0x02, 231 },
        { 31, 0x02, 231 },
        { 41, 0x02, 231 },
        { 56, 0x03, 231 },
    },
    /* 174 */
    {
        { 2, 0x02, 239 },
        { 9, 0x02, 239 },
        { 23, 0x02, 239 },
        { 40, 0x03, 239 },
        { 1, 0x02, 9 },
        { 22, 0x03, 9 },
        { 1, 0x02, 142 },
        { 22, 0x03, 142 },
        { 1, 0x02, 144 },
        { 22, 0x03, 144 },
        { 1, 0x02, 145 },
        { 22, 0x03, 145 },
        { 1, 0x02, 148 },
        { 22, 0x03, 148 },
        { 1, 0x02, 159 },
        { 22, 0x03, 159 },
    },
    /* 175 */
    {
        { 3, 0x02, 239 },
        { 6, 0x02, 239 },
        { 10, 0x02, 239 },
        { 15, 0x02, 239 },
        { 24, 0x02, 239 },
        { 31, 0x02, 239 },
        { 41, 0x02, 239 },
        { 56, 0x03, 239 },
        { 2, 0x02, 9 },
        { 9, 0x02, 9 },
        { 23, 0x02, 9 },
        { 40, 0x03, 9 },
        { 2, 0x02, 142 },
        { 9, 0x02, 142 },
        { 23, 0x02, 142 },
        { 40, 0x03, 142 },
    },
    /* 176 */
    {
        { 3, 0x02, 9 },
        { 6, 0x02, 9 },
        { 10, 0x02, 9 },
        { 15, 0x02, 9 },
        { 24, 0x02, 9 },
        { 31, 0x02, 9 },
        { 41, 0x02, 9 },
        { 56, 0x03, 9 },
        { 3, 0x02, 142 },
        { 6, 0x02, 142 },
        { 10, 0x02, 142 },
        { 15, 0x02, 142 },
        { 24, 0x02, 142 },
        { 31, 0x02, 142 },
        { 41, 0x02, 142 },
        { 56, 0x03, 142 },
    },
    /* 177 */
    {
        { 2, 0x02, 144 },
        { 9, 0x02, 144 },
        { 23, 0x02, 144 },
        { 40, 0x03, 144 },
        { 2, 0x02, 145 },
        { 9, 0x02, 145 },
        { 23, 0x02, 145 },
        { 40, 0x03, 145 },
        { 2, 0x02, 148 },
        { 9, 0x02, 148 },
        { 23, 0x02, 148 },
        { 40, 0x03, 148 },
        { 2, 0x02, 159 },
        { 9, 0x02, 159 },
        { 23, 0x02, 159 },
        { 40, 0x03, 159 },
    },
    /* 178 */
    {
        { 3, 0x02, 144 },
        { 6, 0x02, 144 },
        { 10, 0x02, 144 },
        { 15, 0x02, 144 },
        { 24, 0x02, 144 },
        { 31, 0x02, 144 },
        { 41, 0x02, 144 },
        { 56, 0x03, 144 },
        { 3, 0x02, 145 },
        { 6, 0x02, 145 },
        { 10, 0x02, 145 },
        { 15, 0x02, 145 },
        { 24, 0x02, 145 },
        { 31, 0x02, 145 },
        { 41, 0x02, 145 },
        { 56, 0x03, 145 },
    },
    /* 179 */
    {
        { 3, 0x02, 148 },
        { 6, 0x02, 148 },
        { 10, 0x02, 148 },
        { 15, 0x02, 148 },
        { 24, 0x02, 148 },
        { 31, 0x02, 148 },
        { 41, 0x02, 148 },
        { 56, 0x03, 148 },
        { 3, 0x02, 159 },
        { 6, 0x02, 159 },
        { 10, 0x02, 159 },
        { 15, 0x02, 159 },
        { 24, 0x02, 159 },
        { 31, 0x02, 159 },
        { 41, 0x02, 159 },
        { 56, 0x03, 159 },
    },
    /* 180 */
    {
        { 0, 0x03, 171 },
        { 0, 0x03, 206 },
        { 0, 0x03, 215 },
        { 0, 0x03, 225 },
        { 0, 0x03, 236 },
        { 0, 0x03, 237 },
        { 188, 0x00, 0 },
        { 189, 0x00, 0 },
        { 193, 0x00, 0 },
        { 196, 0x00, 0 },
        { 200, 0x00, 0 },
        { 203, 0x00, 0 },
        { 209, 0x00, 0 },
        { 216, 0x00, 0 },
        { 224, 0x00, 0 },
        { 238, 0x00, 0 },
    },
    /* 181 */
    {
        { 1, 0x02, 171 },
        { 22, 0x03, 171 },
        { 1, 0x02, 206 },
        { 22, 0x03, 206 },
        { 1, 0x02, 215 },
        { 22, 0x03, 215 },
        { 1, 0x02, 225 },
        { 22, 0x03, 225 },
        { 1, 0x02, 236 },
        { 22, 0x03, 236 },
        { 1, 0x02, 237 },
        { 22, 0x03, 237 },
        { 0, 0x03, 199 },
        { 0, 0x03, 207 },
        { 0, 0x03, 234 },
        { 0, 0x03, 235 },
    },
    /* 182 */
    {
        { 2, 0x02, 171 },
        { 9, 0x02, 171 },
        { 23, 0x02, 171 },
        { 40, 0x03, 171 },
        { 2, 0x02, 206 },
        { 9, 0x02, 206 },
        { 23, 0x02, 206 },
        { 40, 0x03, 206 },
        { 2, 0x02, 215 },
        { 9, 0x02, 215 },
        { 23, 0x02, 215 },
        { 40, 0x03, 215 },
        { 2, 0x02, 225 },
        { 9, 0x02, 225 },
        { 23, 0x02, 225 },
        { 40, 0x03, 225 },
    },
    /* 183 */
    {
        { 3, 0x02, 171 },
        { 6, 0x02, 171 },
        { 10, 0x02, 171 },
        { 15, 0x02, 171 },
        { 24, 0x02, 171 },
        { 31, 0x02, 171 },
        { 41, 0x02, 171 },
        { 56, 0x03, 171 },
        { 3, 0x02, 206 },
        { 6, 0x02, 206 },
        { 10, 0x02, 206 },
        { 15, 0x02, 206 },
        { 24, 0x02, 206 },
        { 31, 0x02, 206 },
        { 41, 0x02, 206 },
        { 56, 0x03, 206 },
    },
    /* 184 */
    {
        { 3, 0x02, 215 },
        { 6, 0x02, 215 },
        { 10, 0x02, 215 },
        { 15, 0x02, 215 },
        { 24, 0x02, 215 },
        { 31, 0x02, 215 },
        { 41, 0x02, 215 },
        { 56, 0x03, 215 },
        { 3, 0x02, 225 },
        { 6, 0x02, 225 },
        { 10, 0x02, 225 },
        { 15, 0x02, 225 },
        { 24, 0x02, 225 },
        { 31, 0x02, 225 },
        { 41, 0x02, 225 },
        { 56, 0x03, 225 },
    },
    /* 185 */
    {
        { 2, 0x02, 236 },
        { 9, 0x02, 236 },
        { 23, 0x02, 236 },
        { 40, 0x03, 236 },
        { 2, 0x02, 237 },
        { 9, 0x02, 237 },
        { 23, 0x02, 237 },
        { 40, 0x03, 237 },
        { 1, 0x02, 199 },
        { 22, 0x03, 199 },
        { 1, 0x02, 207 },
        { 22, 0x03, 207 },
        { 1, 0x02, 234 },
        { 22, 0x03, 234 },
        { 1, 0x02, 235 },
        { 22, 0x03, 235 },
    },
    /* 186 */
    {
        { 3, 0x02, 236 },
        { 6, 0x02, 236 },
        { 10, 0x02, 236 },
        { 15, 0x02, 236 },
        { 24, 0x02, 236 },
        { 31, 0x02, 236 },
        { 41, 0x02, 236 },
        { 56, 0x03, 236 },
        { 3, 0x02, 237 },
        { 6, 0x02, 237 },
        { 10, 0x02, 237 },
        { 15, 0x02, 237 },
        { 24, 0x02, 237 },
        { 31, 0x02, 237 },
        { 41, 0x02, 237 },
        { 56, 0x03, 237 },
    },
    /* 187 */
    {
        { 2, 0x02, 199 },
        { 9, 0x02, 199 },
        { 23, 0x02, 199 },
        { 40, 0x03, 199 },
        { 2, 0x02, 207 },
        { 9, 0x02, 207 },
        { 23, 0x02, 207 },
        { 40, 0x03, 207 },
        { 2, 0x02, 234 },
        { 9, 0x02, 234 },
        { 23, 0x02, 234 },
        { 40, 0x03, 234 },
        { 2, 0x02, 235 },
        { 9, 0x02, 235 },
        { 23, 0x02, 235 },
        { 40, 0x03, 235 },
    },
    /* 188 */
    {
        { 3, 0x02, 199 },
        { 6, 0x02, 199 },
        { 10, 0x02, 199 },
        { 15, 0x02, 199 },
        { 24, 0x02, 199 },
        { 31, 0x02, 199 },
        { 41, 0x02, 199 },
        { 56, 0x03, 199 },
        { 3, 0x02, 207 },
        { 6, 0x02, 207 },
        { 10, 0x02, 207 },
        { 15, 0x02, 207 },
        { 24, 0x02, 207 },
        { 31, 0x02, 207 },
        { 41, 0x02, 207 },
        { 56, 0x03, 207 },
    },
    /* 189 */
    {
        { 3, 0x02, 234 },
        { 6, 0x02, 234 },
        { 10, 0x02, 234 },
        { 15, 0x02, 234 },
        { 24, 0x02, 234 },
        { 31, 0x02, 234 },
        { 41, 0x02, 234 },
        { 56, 0x03, 234 },
        { 3, 0x02, 235 },
        { 6, 0x02, 235 },
        { 10, 0x02, 235 },
        { 15, 0x02, 235 },
        { 24, 0x02, 235 },
        { 31, 0x02, 235 },
        { 41, 0x02, 235 },
        { 56, 0x03, 235 },
    },
    /* 190 */
    {
        { 194, 0x00, 0 },
        { 195, 0x00, 0 },
        { 197, 0x00, 0 },
        { 198, 0x00, 0 },
        { 201, 0x00, 0 },
        { 202, 0x00, 0 },
        { 204, 0x00, 0 },
        { 205, 0x00, 0 },
        { 210, 0x00, 0 },
        { 213, 0x00, 0 },
        { 217, 0x00, 0 },
        { 220, 0x00, 0 },
        { 225, 0x00, 0 },
        { 231, 0x00, 0 },
        { 239, 0x00, 0 },
        { 246, 0x00, 0 },
    },
    /* 191 */
    {
        { 0, 0x03, 192 },
        { 0, 0x03, 193 },
        { 0, 0x03, 200 },
        { 0, 0x03, 201 },
        { 0, 0x03, 202 },
        { 0, 0x03, 205 },
        { 0, 0x03, 210 },
        { 0, 0x03, 213 },
        { 0, 0x03, 218 },
        { 0, 0x03, 219 },
        { 0, 0x03, 238 },
        { 0, 0x03, 240 },
        { 0, 0x03, 242 },
        { 0, 0x03, 243 },
        { 0, 0x03, 255 },
        { 206, 0x00, 0 },
    },
    /* 192 */
    {
        { 1, 0x02, 192 },
        { 22, 0x03, 192 },
        { 1, 0x02, 193 },
        { 22, 0x03, 193 },
        { 1, 0x02, 200 },
        { 22, 0x03, 200 },
        { 1, 0x02, 201 },
        { 22, 0x03, 201 },
        { 1, 0x02, 202 },
        { 22, 0x03, 202 },
        { 1, 0x02, 205 },
        { 22, 0x03, 205 },
        { 1, 0x02, 210 },
        { 22, 0x03, 210 },
        { 1, 0x02, 213 },
        { 22, 0x03, 213 },
    },
    /* 193 */
    {
        { 2, 0x02, 192 },
        { 9, 0x02, 192 },
        { 23, 0x02, 192 },
        { 40, 0x03, 192 },
        { 2, 0x02, 193 },
        { 9, 0x02, 193 },
        { 23, 0x02, 193 },
        { 40, 0x03, 193 },
        { 2, 0x02, 200 },
        { 9, 0x02, 200 },
        { 23, 0x02, 200 },
        { 40, 0x03, 200 },
        { 2, 0x02, 201 },
        { 9, 0x02, 201 },
        { 23, 0x02, 201 },
        { 40, 0x03, 201 },
    },
    /* 194 */
    {
        { 3, 0x02, 192 },
        { 6, 0x02, 192 },
        { 10, 0x02, 192 },
        { 15, 0x02, 192 },
        { 24, 0x02, 192 },
        { 31, 0x02, 192 },
        { 41, 0x02, 192 },
        { 56, 0x03, 192 },
        { 3, 0x02, 193 },
        { 6, 0x02, 193 },
        { 10, 0x02, 193 },
        { 15, 0x02, 193 },
        { 24, 0x02, 193 },
        { 31, 0x02, 193 },
        { 41, 0x02, 193 },
        { 56, 0x03, 193 },
    },
    /* 195 */
    {
        { 3, 0x02, 200 },
        { 6, 0x02, 200 },
        { 10, 0x02, 200 },
        { 15, 0x02, 200 },
        { 24, 0x02, 200 },
        { 31, 0x02, 200 },
        { 41, 0x02, 200 },
        { 56, 0x03, 200 },
        { 3, 0x02, 201 },
        { 6, 0x02, 201 },
        { 10, 0x02, 201 },
        { 15, 0x02, 201 },
        { 24, 0x02, 201 },
        { 31, 0x02, 201 },
        { 41, 0x02, 201 },
        { 56, 0x03, 201 },
    },
    /* 196 */
    {
        { 2, 0x02, 202 },
        { 9, 0x02, 202 },
        { 23, 0x02, 202 },
        { 40, 0x03, 202 },
        { 2, 0x02, 205 },
        { 9, 0x02, 205 },
        { 23, 0x02, 205 },
        { 40, 0x03, 205 },
        { 2, 0x02, 210 },
        { 9, 0x02, 210 },
        { 23, 0x02, 210 },
        { 40, 0x03, 210 },
        { 2, 0x02, 213 },
        { 9, 0x02, 213 },
        { 23, 0x02, 213 },
        { 40, 0x03, 213 },
    },
    /* 197 */
    {
        { 3, 0x02, 202 },
        { 6, 0x02, 202 },
        { 10, 0x02, 202 },
        { 15, 0x02, 202 },
        { 24, 0x02, 202 },
        { 31, 0x02, 202 },
        { 41, 0x02, 202 },
        { 56, 0x03, 202 },
        { 3, 0x02, 205 },
        { 6, 0x02, 205 },
        { 10, 0x02, 205 },
        { 15, 0x02, 205 },
        { 24, 0x02, 205 },
        { 31, 0x02, 205 },
        { 41, 0x02, 205 },
        { 56, 0x03, 205 },
    },
    /* 198 */
    {
        { 3, 0x02, 210 },
        { 6, 0x02, 210 },
        { 10, 0x02, 210 },
        { 15, 0x02, 210 },
        { 24, 0x02, 210 },
        { 31, 0x02, 210 },
        { 41, 0x02, 210 },
        { 56, 0x03, 210 },
        { 3, 0x02, 213 },
        { 6, 0x02, 213 },
        { 10, 0x02, 213 },
        { 15, 0x02, 213 },
        { 24, 0x02, 213 },
        { 31, 0x02, 213 },
        { 41, 0x02, 213 },
        { 56, 0x03, 213 },
    },
    /* 199 */
    {
        { 1, 0x02, 218 },
        { 22, 0x03, 218 },
        { 1, 0x02, 219 },
        { 22, 0x03, 219 },
        { 1, 0x02, 238 },
        { 22, 0x03, 238 },
        { 1, 0x02, 240 },
        { 22, 0x03, 240 },
        { 1, 0x02, 242 },
        { 22, 0x03, 242 },
        { 1, 0x02, 243 },
        { 22, 0x03, 243 },
        { 1, 0x02, 255 },
        { 22, 0x03, 255 },
        { 0, 0x03, 203 },
        { 0, 0x03, 204 },
    },
    /* 200 */
    {
        { 2, 0x02, 218 },
        { 9, 0x02, 218 },
        { 23, 0x02, 218 },
        { 40, 0x03, 218 },
        { 2, 0x02, 219 },
        { 9, 0x02, 219 },
        { 23, 0x02, 219 },
        { 40, 0x03, 219 },
        { 2, 0x02, 238 },
        { 9, 0x02, 238 },
        { 23, 0x02, 238 },
        { 40, 0x03, 238 },
        { 2, 0x02, 240 },
        { 9, 0x02, 240 },
        { 23, 0x02, 240 },
        { 40, 0x03, 240 },
    },
    /* 201 */
    {
        { 3, 0x02, 218 },
        { 6, 0x02, 218 },
        { 10, 0x02, 218 },
        { 15, 0x02, 218 },
        { 24, 0x02, 218 },
        { 31, 0x02, 218 },
        { 41, 0x02, 218 },
        { 56, 0x03, 218 },
        { 3, 0x02, 219 },
        { 6, 0x02, 219 },
        { 10, 0x02, 219 },
        { 15, 0x02, 219 },
        { 24, 0x02, 219 },
        { 31, 0x02, 219 },
        { 41, 0x02, 219 },
        { 56, 0x03, 219 },
    },
    /* 202 */
    {
        { 3, 0x02, 238 },
        { 6, 0x02, 238 },
        { 10, 0x02, 238 },
        { 15, 0x02, 238 },
        { 24, 0x02, 238 },
        { 31, 0x02, 238 },
        { 41, 0x02, 238 },
        { 56, 0x03, 238 },
        { 3, 0x02, 240 },
        { 6, 0x02, 240 },
        { 10, 0x02, 240 },
        { 15, 0x02, 240 },
        { 24, 0x02, 240 },
        { 31, 0x02, 240 },
        { 41, 0x02, 240 },
        { 56, 0x03, 240 },
    },
    /* 203 */
    {
        { 2, 0x02, 242 },
        { 9, 0x02, 242 },
        { 23, 0x02, 242 },
        { 40, 0x03, 242 },
        { 2, 0x02, 243 },
        { 9, 0x02, 243 },
        { 23, 0x02, 243 },
        { 40, 0x03, 243 },
        { 2, 0x02, 255 },
        { 9, 0x02, 255 },
        { 23, 0x02, 255 },
        { 40, 0x03, 255 },
        { 1, 0x02, 203 },
        { 22, 0x03, 203 },
        { 1, 0x02, 204 },
        { 22, 0x03, 204 },
    },
    /* 204 */
    {
        { 3, 0x02, 242 },
        { 6, 0x02, 242 },
        { 10, 0x02, 242 },
        { 15, 0x02, 242 },
        { 24, 0x02, 242 },
        { 31, 0x02, 242 },
        { 41, 0x02, 242 },
        { 56, 0x03, 242 },
        { 3, 0x02, 243 },
        { 6, 0x02, 243 },
        { 10, 0x02, 243 },
        { 15, 0x02, 243 },
        { 24, 0x02, 243 },
        { 31, 0x02, 243 },
        { 41, 0x02, 243 },
        { 56, 0x03, 243 },
    },
    /* 205 */
    {
        { 3, 0x02, 255 },
        { 6, 0x02, 255 },
        { 10, 0x02, 255 },
        { 15, 0x02, 255 },
        { 24, 0x02, 255 },
        { 31, 0x02, 255 },
        { 41, 0x02, 255 },
        { 56, 0x03, 255 },
        { 2, 0x02, 203 },
        { 9, 0x02, 203 },
        { 23, 0x02, 203 },
        { 40, 0x03, 203 },
        { 2, 0x02, 204 },
        { 9, 0x02, 204 },
        { 23, 0x02, 204 },
        { 40, 0x03, 204 },
    },
    /* 206 */
    {
        { 3, 0x02, 203 },
        { 6, 0x02, 203 },
        { 10, 0x02, 203 },
        { 15, 0x02, 203 },
        { 24, 0x02, 203 },
        { 31, 0x02, 203 },
        { 41, 0x02, 203 },
        { 56, 0x03, 203 },
        { 3, 0x02, 204 },
        { 6, 0x02, 204 },
        { 10, 0x02, 204 },
        { 15, 0x02, 204 },
        { 24, 0x02, 204 },
        { 31, 0x02, 204 },
        { 41, 0x02, 204 },
        { 56, 0x03, 204 },
    },
    /* 207 */
    {
        { 211, 0x00, 0 },
        { 212, 0x00, 0 },
        { 214, 0x00, 0 },
        { 215, 0x00, 0 },
        { 218, 0x00, 0 },
        { 219, 0x00, 0 },
        { 221, 0x00, 0 },
        { 222, 0x00, 0 },
        { 226, 0x00, 0 },
        { 228, 0x00, 0 },
        { 232, 0x00, 0 },
        { 235, 0x00, 0 },
        { 240, 0x00, 0 },
        { 243, 0x00, 0 },
        { 247, 0x00, 0 },
        { 250, 0x00, 0 },
    },
    /* 208 */
    {
        { 0, 0x03, 211 },
        { 0, 0x03, 212 },
        { 0, 0x03, 214 },
        { 0, 0x03, 221 },
        { 0, 0x03, 222 },
        { 0, 0x03, 223 },
        { 0, 0x03, 241 },
        { 0, 0x03, 244 },
        { 0, 0x03, 245 },
        { 0, 0x03, 246 },
        { 0, 0x03, 247 },
        { 0, 0x03, 248 },
        { 0, 0x03, 250 },
        { 0, 0x03, 251 },
        { 0, 0x03, 252 },
        { 0, 0x03, 253 },
    },
    /* 209 */
    {
        { 1, 0x02, 211 },
        { 22, 0x03, 211 },
        { 1, 0x02, 212 },
        { 22, 0x03, 212 },
        { 1, 0x02, 214 },
        { 22, 0x03, 214 },
        { 1, 0x02, 221 },
        { 22, 0x03, 221 },
        { 1, 0x02, 222 },
        { 22, 0x03, 222 },
        { 1, 0x02, 223 },
        { 22, 0x03, 223 },
        { 1, 0x02, 241 },
        { 22, 0x03, 241 },
        { 1, 0x02, 244 },
        { 22, 0x03, 244 },
    },
    /* 210 */
    {
        { 2, 0x02, 211 },
        { 9, 0x02, 211 },
        { 23, 0x02, 211 },
        { 40, 0x03, 211 },
        { 2, 0x02, 212 },
        { 9, 0x02, 212 },
        { 23, 0x02, 212 },
        { 40, 0x03, 212 },
        { 2, 0x02, 214 },
        { 9, 0x02, 214 },
        { 23, 0x02, 214 },
        { 40, 0x03, 214 },
        { 2, 0x02, 221 },
        { 9, 0x02, 221 },
        { 23, 0x02, 221 },
        { 40, 0x03, 221 },
    },
    /* 211 */
    {
        { 3, 0x02, 211 },
        { 6, 0x02, 211 },
        { 10, 0x02, 211 },
        { 15, 0x02, 211 },
        { 24, 0x02, 211 },
        { 31, 0x02, 211 },
        { 41, 0x02, 211 },
        { 56, 0x03, 211 },
        { 3, 0x02, 212 },
        { 6, 0x02, 212 },
        { 10, 0x02, 212 },
        { 15, 0x02, 212 },
        { 24, 0x02, 212 },
        { 31, 0x02, 212 },
        { 41, 0x02, 212 },
        { 56, 0x03, 212 },
    },
    /* 212 */
    {
        { 3, 0x02, 214 },
        { 6, 0x02, 214 },
        { 10, 0x02, 214 },
        { 15, 0x02, 214 },
        { 24, 0x02, 214 },
        { 31, 0x02, 214 },
        { 41, 0x02, 214 },
        { 56, 0x03, 214 },
        { 3, 0x02, 221 },
        { 6, 0x02, 221 },
        { 10, 0x02, 221 },
        { 15, 0x02, 221 },
        { 24, 0x02, 221 },
        { 31, 0x02, 221 },
        { 41, 0x02, 221 },
        { 56, 0x03, 221 },
    },
    /* 213 */
    {
        { 2, 0x02, 222 },
        { 9, 0x02, 222 },
        { 23, 0x02, 222 },
        { 40, 0x03, 222 },
        { 2, 0x02, 223 },
        { 9, 0x02, 223 },
        { 23, 0x02, 223 },
        { 40, 0x03, 223 },
        { 2, 0x02, 241 },
        { 9, 0x02, 241 },
        { 23, 0x02, 241 },
        { 40, 0x03, 241 },
        { 2, 0x02, 244 },
        { 9, 0x02, 244 },
        { 23, 0x02, 244 },
        { 40, 0x03, 244 },
    },
    /* 214 */
    {
        { 3, 0x02, 222 },
        { 6, 0x02, 222 },
        { 10, 0x02, 222 },
        { 15, 0x02, 222 },
        { 24, 0x02, 222 },
        { 31, 0x02, 222 },
        { 41, 0x02, 222 },
        { 56, 0x03, 222 },
        { 3, 0x02, 223 },
        { 6, 0x02, 223 },
        { 10, 0x02, 223 },
        { 15, 0x02, 223 },
        { 24, 0x02, 223 },
        { 31, 0x02, 223 },
        { 41, 0x02, 223 },
        { 56, 0x03, 223 },
    },
    /* 215 */
    {
        { 3, 0x02, 241 },
        { 6, 0x02, 241 },
        { 10, 0x02, 241 },
        { 15, 0x02, 241 },
        { 24, 0x02, 241 },
        { 31, 0x02, 241 },
        { 41, 0x02, 241 },
        { 56, 0x03, 241 },
        { 3, 0x02, 244 },
        { 6, 0x02, 244 },
        { 10, 0x02, 244 },
        { 15, 0x02, 244 },
        { 24, 0x02, 244 },
        { 31, 0x02, 244 },
        { 41, 0x02, 244 },
        { 56, 0x03, 244 },
    },
    /* 216 */
    {
        { 1, 0x02, 245 },
        { 22, 0x03, 245 },
        { 1, 0x02, 246 },
        { 22, 0x03, 246 },
        { 1, 0x02, 247 },
        { 22, 0x03, 247 },
        { 1, 0x02, 248 },
        { 22, 0x03, 248 },
        { 1, 0x02, 250 },
        { 22, 0x03, 250 },
        { 1, 0x02, 251 },
        { 22, 0x03, 251 },
        { 1, 0x02, 252 },
        { 22, 0x03, 252 },
        { 1, 0x02, 253 },
        { 22, 0x03, 253 },
    },
    /* 217 */
    {
        { 2, 0x02, 245 },
        { 9, 0x02, 245 },
        { 23, 0x02, 245 },
        { 40, 0x03, 245 },
        { 2, 0x02, 246 },
        { 9, 0x02, 246 },
        { 23, 0x02, 246 },
        { 40, 0x03, 246 },
        { 2, 0x02, 247 },
        { 9, 0x02, 247 },
        { 23, 0x02, 247 },
        { 40, 0x03, 247 },
        { 2, 0x02, 248 },
        { 9, 0x02, 248 },
        { 23, 0x02, 248 },
        { 40, 0x03, 248 },
    },
    /* 218 */
    {
        { 3, 0x02, 245 },
        { 6, 0x02, 245 },
        { 10, 0x02, 245 },
        { 15, 0x02, 245 },
        { 24, 0x02, 245 },
        { 31, 0x02, 245 },
        { 41, 0x02, 245 },
        { 56, 0x03, 245 },
        { 3, 0x02, 246 },
        { 6, 0x02, 246 },
        { 10, 0x02, 246 },
        { 15, 0x02, 246 },
        { 24, 0x02, 246 },
        { 31, 0x02, 246 },
        { 41, 0x02, 246 },
        { 56, 0x03, 246 },
    },
    /* 219 */
    {
        { 3, 0x02, 247 },
        { 6, 0x02, 247 },
        { 10, 0x02, 247 },
        { 15, 0x02, 247 },
        { 24, 0x02, 247 },
        { 31, 0x02, 247 },
        { 41, 0x02, 247 },
        { 56, 0x03, 247 },
        { 3, 0x02, 248 },
        { 6, 0x02, 248 },
        { 10, 0x02, 248 },
        { 15, 0x02, 248 },
        { 24, 0x02, 248 },
        { 31, 0x02, 248 },
        { 41, 0x02, 248 },
        { 56, 0x03, 248 },
    },
    /* 220 */
    {
        { 2, 0x02, 250 },
        { 9, 0x02, 250 },
        { 23, 0x02, 250 },
        { 40, 0x03, 250 },
        { 2, 0x02, 251 },
        { 9, 0x02, 251 },
        { 23, 0x02, 251 },
        { 40, 0x03, 251 },
        { 2, 0x02, 252 },
        { 9, 0x02, 252 },
        { 23, 0x02, 252 },
        { 40, 0x03, 252 },
        { 2, 0x02, 253 },
        { 9, 0x02, 253 },
        { 23, 0x02, 253 },
        { 40, 0x03, 253 },
    },
    /* 221 */
    {
        { 3, 0x02, 250 },
        { 6, 0x02, 250 },
        { 10, 0x02, 250 },
        { 15, 0x02, 250 },
        { 24, 0x02, 250 },
        { 31, 0x02, 250 },
        { 41, 0x02, 250 },
        { 56, 0x03, 250 },
        { 3, 0x02, 251 },
        { 6, 0x02, 251 },
        { 10, 0x02, 251 },
        { 15, 0x02, 251 },
        { 24, 0x02, 251 },
        { 31, 0x02, 251 },
        { 41, 0x02, 251 },
        { 56, 0x03, 251 },
    },
    /* 222 */
    {
        { 3, 0x02, 252 },
        { 6, 0x02, 252 },
        { 10, 0x02, 252 },
        { 15, 0x02, 252 },
        { 24, 0x02, 252 },
        { 31, 0x02, 252 },
        { 41, 0x02, 252 },
        { 56, 0x03, 252 },
        { 3, 0x02, 253 },
        { 6, 0x02, 253 },
        { 10, 0x02, 253 },
        { 15, 0x02, 253 },
        { 24, 0x02, 253 },
        { 31, 0x02, 253 },
        { 41, 0x02, 253 },
        { 56, 0x03, 253 },
    },
    /* 223 */
    {
        { 0, 0x03, 254 },
        { 227, 0x00, 0 },
        { 229, 0x00, 0 },
        { 230, 0x00, 0 },
        { 233, 0x00, 0 },
        { 234, 0x00, 0 },
        { 236, 0x00, 0 },
        { 237, 0x00, 0 },
        { 241, 0x00, 0 },
        { 242, 0x00, 0 },
        { 244, 0x00, 0 },
        { 245, 0x00, 0 },
        { 248, 0x00, 0 },
        { 249, 0x00, 0 },
        { 251, 0x00, 0 },
        { 252, 0x00, 0 },
    },
    /* 224 */
    {
        { 1, 0x02, 254 },
        { 22, 0x03, 254 },
        { 0, 0x03, 2 },
        { 0, 0x03, 3 },
        { 0, 0x03, 4 },
        { 0, 0x03, 5 },
        { 0, 0x03, 6 },
        { 0, 0x03, 7 },
        { 0, 0x03, 8 },
        { 0, 0x03, 11 },
        { 0, 0x03, 12 },
        { 0, 0x03, 14 },
        { 0, 0x03, 15 },
        { 0, 0x03, 16 },
        { 0, 0x03, 17 },
        { 0, 0x03, 18 },
    },
    /* 225 */
    {
        { 2, 0x02, 254 },
        { 9, 0x02, 254 },
        { 23, 0x02, 254 },
        { 40, 0x03, 254 },
        { 1, 0x02, 2 },
        { 22, 0x03, 2 },
        { 1, 0x02, 3 },
        { 22, 0x03, 3 },
        { 1, 0x02, 4 },
        { 22, 0x03, 4 },
        { 1, 0x02, 5 },
        { 22, 0x03, 5 },
        { 1, 0x02, 6 },
        { 22, 0x03, 6 },
        { 1, 0x02, 7 },
        { 22, 0x03, 7 },
    },
    /* 226 */
    {
        { 3, 0x02, 254 },
        { 6, 0x02, 254 },
        { 10, 0x02, 254 },
        { 15, 0x02, 254 },
        { 24, 0x02, 254 },
        { 31, 0x02, 254 },
        { 41, 0x02, 254 },
        { 56, 0x03, 254 },
        { 2, 0x02, 2 },
        { 9, 0x02, 2 },
        { 23, 0x02, 2 },
        { 40, 0x03, 2 },
        { 2, 0x02, 3 },
        { 9, 0x02, 3 },
        { 23, 0x02, 3 },
        { 40, 0x03, 3 },
    },
    /* 227 */
    {
        { 3, 0x02, 2 },
        { 6, 0x02, 2 },
        { 10, 0x02, 2 },
        { 15, 0x02, 2 },
        { 24, 0x02, 2 },
        { 31, 0x02, 2 },
        { 41, 0x02, 2 },
        { 56, 0x03, 2 },
        { 3, 0x02, 3 },
        { 6, 0x02, 3 },
        { 10, 0x02, 3 },
        { 15, 0x02, 3 },
        { 24, 0x02, 3 },
        { 31, 0x02, 3 },
        { 41, 0x02, 3 },
        { 56, 0x03, 3 },
    },
    /* 228 */
    {
        { 2, 0x02, 4 },
        { 9, 0x02, 4 },
        { 23, 0x02, 4 },
        { 40, 0x03, 4 },
        { 2, 0x02, 5 },
        { 9, 0x02, 5 },
        { 23, 0x02, 5 },
        { 40, 0x03, 5 },
        { 2, 0x02, 6 },
        { 9, 0x02, 6 },
        { 23, 0x02, 6 },
        { 40, 0x03, 6 },
        { 2, 0x02, 7 },
        { 9, 0x02, 7 },
        { 23, 0x02, 7 },
        { 40, 0x03, 7 },
    },
    /* 229 */
    {
        { 3, 0x02, 4 },
        { 6, 0x02, 4 },
        { 10, 0x02, 4 },
        { 15, 0x02, 4 },
        { 24, 0x02, 4 },
        { 31, 0x02, 4 },
        { 41, 0x02, 4 },
        { 56, 0x03, 4 },
        { 3, 0x02, 5 },
        { 6, 0x02, 5 },
        { 10, 0x02, 5 },
        { 15, 0x02, 5 },
        { 24, 0x02, 5 },
        { 31, 0x02, 5 },
        { 41, 0x02, 5 },
        { 56, 0x03, 5 },
    },
    /* 230 */
    {
        { 3, 0x02, 6 },
        { 6, 0x02, 6 },
        { 10, 0x02, 6 },
        { 15, 0x02, 6 },
        { 24, 0x02, 6 },
        { 31, 0x02, 6 },
        { 41, 0x02, 6 },
        { 56, 0x03, 6 },
        { 3, 0x02, 7 },
        { 6, 0x02, 7 },
        { 10, 0x02, 7 },
        { 15, 0x02, 7 },
        { 24, 0x02, 7 },
        { 31, 0x02, 7 },
        { 41, 0x02, 7 },
        { 56, 0x03, 7 },
    },
    /* 231 */
    {
        { 1, 0x02, 8 },
        { 22, 0x03, 8 },
        { 1, 0x02, 11 },
        { 22, 0x03, 11 },
        { 1, 0x02, 12 },
        { 22, 0x03, 12 },
        { 1, 0x02, 14 },
        { 22, 0x03, 14 },
        { 1, 0x02, 15 },
        { 22, 0x03, 15 },
        { 1, 0x02, 16 },
        { 22, 0x03, 16 },
        { 1, 0x02, 17 },
        { 22, 0x03, 17 },
        { 1, 0x02, 18 },
        { 22, 0x03, 18 },
    },
    /* 232 */
    {
        { 2, 0x02, 8 },
        { 9, 0x02, 8 },
        { 23, 0x02, 8 },
        { 40, 0x03, 8 },
        { 2, 0x02, 11 },
        { 9, 0x02, 11 },
        { 23, 0x02, 11 },
        { 40, 0x03, 11 },
        { 2, 0x02, 12 },
        { 9, 0x02, 12 },
        { 23, 0x02, 12 },
        { 40, 0x03, 12 },
        { 2, 0x02, 14 },
        { 9, 0x02, 14 },
        { 23, 0x02, 14 },
        { 40, 0x03, 14 },
    },
    /* 233 */
    {
        { 3, 0x02, 8 },
        { 6, 0x02, 8 },
        { 10, 0x02, 8 },
        { 15, 0x02, 8 },
        { 24, 0x02, 8 },
        { 31, 0x02, 8 },
        { 41, 0x02, 8 },
        { 56, 0x03, 8 },
        { 3, 0x02, 11 },
        { 6, 0x02, 11 },
        { 10, 0x02, 11 },
        { 15, 0x02, 11 },
        { 24, 0x02, 11 },
        { 31, 0x02, 11 },
        { 41, 0x02, 11 },
        { 56, 0x03, 11 },
    },
    /* 234 */
    {
        { 3, 0x02, 12 },
        { 6, 0x02, 12 },
        { 10, 0x02, 12 },
        { 15, 0x02, 12 },
        { 24, 0x02, 12 },
        { 31, 0x02, 12 },
        { 41, 0x02, 12 },
        { 56, 0x03, 12 },
        { 3, 0x02, 14 },
        { 6, 0x02, 14 },
        { 10, 0x02, 14 },
        { 15, 0x02, 14 },
        { 24, 0x02, 14 },
        { 31, 0x02, 14 },
        { 41, 0x02, 14 },
        { 56, 0x03, 14 },
    },
    /* 235 */
    {
        { 2, 0x02, 15 },
        { 9, 0x02, 15 },
        { 23, 0x02, 15 },
        { 40, 0x03, 15 },
        { 2, 0x02, 16 },
        { 9, 0x02, 16 },
        { 23, 0x02, 16 },
        { 40, 0x03, 16 },
        { 2, 0x02, 17 },
        { 9, 0x02, 17 },
        { 23, 0x02, 17 },
        { 40, 0x03, 17 },
        { 2, 0x02, 18 },
        { 9, 0x02, 18 },
        { 23, 0x02, 18 },
        { 40, 0x03, 18 },
    },
    /* 236 */
    {
        { 3, 0x02, 15 },
        { 6, 0x02, 15 },
        { 10, 0x02, 15 },
        { 15, 0x02, 15 },
        { 24, 0x02, 15 },
        { 31, 0x02, 15 },
        { 41, 0x02, 15 },
        { 56, 0x03, 15 },
        { 3, 0x02, 16 },
        { 6, 0x02, 16 },
        { 10, 0x02, 16 },
        { 15, 0x02, 16 },
        { 24, 0x02, 16 },
        { 31, 0x02, 16 },
        { 41, 0x02, 16 },
        { 56, 0x03, 16 },
    },
    /* 237 */
    {
        { 3, 0x02, 17 },
        { 6, 0x02, 17 },
        { 10, 0x02, 17 },
        { 15, 0x02, 17 },
        { 24, 0x02, 17 },
        { 31, 0x02, 17 },
        { 41, 0x02, 17 },
        { 56, 0x03, 17 },
        { 3, 0x02, 18 },
        { 6, 0x02, 18 },
        { 10, 0x02, 18 },
        { 15, 0x02, 18 },
        { 24, 0x02, 18 },
        { 31, 0x02, 18 },
        { 41, 0x02, 18 },
        { 56, 0x03, 18 },
    },
    /* 238 */
    {
        { 0, 0x03, 19 },
        { 0, 0x03, 20 },
        { 0, 0x03, 21 },
        { 0, 0x03, 23 },
        { 0, 0x03, 24 },
        { 0, 0x03, 25 },
        { 0, 0x03, 26 },
        { 0, 0x03, 27 },
        { 0, 0x03, 28 },
        { 0, 0x03, 29 },
        { 0, 0x03, 30 },
        { 0, 0x03, 31 },
        { 0, 0x03, 127 },
        { 0, 0x03, 220 },
        { 0, 0x03, 249 },
        { 253, 0x00, 0 },
    },
    /* 239 */
    {
        { 1, 0x02, 19 },
        { 22, 0x03, 19 },
        { 1, 0x02, 20 },
        { 22, 0x03, 20 },
        { 1, 0x02, 21 },
        { 22, 0x03, 21 },
        { 1, 0x02, 23 },
        { 22, 0x03, 23 },
        { 1, 0x02, 24 },
        { 22, 0x03, 24 },
        { 1, 0x02, 25 },
        { 22, 0x03, 25 },
        { 1, 0x02, 26 },
        { 22, 0x03, 26 },
        { 1, 0x02, 27 },
        { 22, 0x03, 27 },
    },
    /* 240 */
    {
        { 2, 0x02, 19 },
        { 9, 0x02, 19 },
        { 23, 0x02, 19 },
        { 40, 0x03, 19 },
        { 2, 0x02, 20 },
        { 9, 0x02, 20 },
        { 23, 0x02, 20 },
        { 40, 0x03, 20 },
        { 2, 0x02, 21 },
        { 9, 0x02, 21 },
        { 23, 0x02, 21 },
        { 40, 0x03, 21 },
        { 2, 0x02, 23 },
        { 9, 0x02, 23 },
        { 23, 0x02, 23 },
        { 40, 0x03, 23 },
    },
    /* 241 */
    {
        { 3, 0x02, 19 },
        { 6, 0x02, 19 },
        { 10, 0x02, 19 },
        { 15, 0x02, 19 },
        { 24, 0x02, 19 },
        { 31, 0x02, 19 },
        { 41, 0x02, 19 },
        { 56, 0x03, 19 },
        { 3, 0x02, 20 },
        { 6, 0x02, 20 },
        { 10, 0x02, 20 },
        { 15, 0x02, 20 },
        { 24, 0x02, 20 },
        { 31, 0x02, 20 },
        { 41, 0x02, 20 },
        { 56, 0x03, 20 },
    },
    /* 242 */
    {
        { 3, 0x02, 21 },
        { 6, 0x02, 21 },
        { 10, 0x02, 21 },
        { 15, 0x02, 21 },
        { 24, 0x02, 21 },
        { 31, 0x02, 21 },
        { 41, 0x02, 21 },
        { 56, 0x03, 21 },
        { 3, 0x02, 23 },
        { 6, 0x02, 23 },
        { 10, 0x02, 23 },
        { 15, 0x02, 23 },
        { 24, 0x02, 23 },
        { 31, 0x02, 23 },
        { 41, 0x02, 23 },
        { 56, 0x03, 23 },
    },
    /* 243 */
    {
        { 2, 0x02, 24 },
        { 9, 0x02, 24 },
        { 23, 0x02, 24 },
        { 40, 0x03, 24 },
        { 2, 0x02, 25 },
        { 9, 0x02, 25 },
        { 23, 0x02, 25 },
        { 40, 0x03, 25 },
        { 2, 0x02, 26 },
        { 9, 0x02, 26 },
        { 23, 0x02, 26 },
        { 40, 0x03, 26 },
        { 2, 0x02, 27 },
        { 9, 0x02, 27 },
        { 23, 0x02, 27 },
        { 40, 0x03, 27 },
    },
    /* 244 */
    {
        { 3, 0x02, 24 },
        { 6, 0x02, 24 },
        { 10, 0x02, 24 },
        { 15, 0x02, 24 },
        { 24, 0x02, 24 },
        { 31, 0x02, 24 },
        { 41, 0x02, 24 },
        { 56, 0x03, 24 },
        { 3, 0x02, 25 },
        { 6, 0x02, 25 },
        { 10, 0x02, 25 },
        { 15, 0x02, 25 },
        { 24, 0x02, 25 },
        { 31, 0x02, 25 },
        { 41, 0x02, 25 },
        { 56, 0x03, 25 },
    },
    /* 245 */
    {
        { 3, 0x02, 26 },
        { 6, 0x02, 26 },
        { 10, 0x02, 26 },
        { 15, 0x02, 26 },
        { 24, 0x02, 26 },
        { 31, 0x02, 26 },
        { 41, 0x02, 26 },
        { 56, 0x03, 26 },
        { 3, 0x02, 27 },
        { 6, 0x02, 27 },
        { 10, 0x02, 27 },
        { 15, 0x02, 27 },
        { 24, 0x02, 27 },
        { 31, 0x02, 27 },
        { 41, 0x02, 27 },
        { 56, 0x03, 27 },
    },
    /* 246 */
    {
        { 1, 0x02, 28 },
        { 22, 0x03, 28 },
        { 1, 0x02, 29 },
        { 22, 0x03, 29 },
        { 1, 0x02, 30 },
        { 22, 0x03, 30 },
        { 1, 0x02, 31 },
        { 22, 0x03, 31 },
        { 1, 0x02, 127 },
        { 22, 0x03, 127 },
        { 1, 0x02, 220 },
        { 22, 0x03, 220 },
        { 1, 0x02, 249 },
        { 22, 0x03, 249 },
        { 254, 0x00, 0 },
        { 255, 0x00, 0 },
    },
    /* 247 */
    {
        { 2, 0x02, 28 },
        { 9, 0x02, 28 },
        { 23, 0x02, 28 },
        { 40, 0x03, 28 },
        { 2, 0x02, 29 },
        { 9, 0x02, 29 },
        { 23, 0x02, 29 },
        { 40, 0x03, 29 },
        { 2, 0x02, 30 },
        { 9, 0x02, 30 },
        { 23, 0x02, 30 },
        { 40, 0x03, 30 },
        { 2, 0x02, 31 },
        { 9, 0x02, 31 },
        { 23, 0x02, 31 },
        { 40, 0x03, 31 },
    },
    /* 248 */
    {
        { 3, 0x02, 28 },
        { 6, 0x02, 28 },
        { 10, 0x02, 28 },
        { 15, 0x02, 28 },
        { 24, 0x02, 28 },
        { 31, 0x02, 28 },
        { 41, 0x02, 28 },
        { 56, 0x03, 28 },
        { 3, 0x02, 29 },
        { 6, 0x02, 29 },
        { 10, 0x02, 29 },
        { 15, 0x02, 29 },
        { 24, 0x02, 29 },
        { 31, 0x02, 29 },
        { 41, 0x02, 29 },
        { 56, 0x03, 29 },
    },
    /* 249 */
    {
        { 3, 0x02, 30 },
        { 6, 0x02, 30 },
        { 10, 0x02, 30 },
        { 15, 0x02, 30 },
        { 24, 0x02, 30 },
        { 31, 0x02, 30 },
        { 41, 0x02, 30 },
        { 56, 0x03, 30 },
        { 3, 0x02, 31 },
        { 6, 0x02, 31 },
        { 10, 0x02, 31 },
        { 15, 0x02, 31 },
        { 24, 0x02, 31 },
        { 31, 0x02, 31 },
        { 41, 0x02, 31 },
        { 56, 0x03, 31 },
    },
    /* 250 */
    {
        { 2, 0x02, 127 },
        { 9, 0x02, 127 },
        { 23, 0x02, 127 },
        { 40, 0x03, 127 },
        { 2, 0x02, 220 },
        { 9, 0x02, 220 },
        { 23, 0x02, 220 },
        { 40, 0x03, 220 },
        { 2, 0x02, 249 },
        { 9, 0x02, 249 },
        { 23, 0x02, 249 },
        { 40, 0x03, 249 },
        { 0, 0x03, 10 },
        { 0, 0x03, 13 },
        { 0, 0x03, 22 },
        { 0, 0x04, 0 },
    },
    /* 251 */
    {
        { 3, 0x02, 127 },
        { 6, 0x02, 127 },
        { 10, 0x02, 127 },
        { 15, 0x02, 127 },
        { 24, 0x02, 127 },
        { 31, 0x02, 127 },
        { 41, 0x02, 127 },
        { 56, 0x03, 127 },
        { 3, 0x02, 220 },
        { 6, 0x02, 220 },
        { 10, 0x02, 220 },
        { 15, 0x02, 220 },
        { 24, 0x02, 220 },
        { 31, 0x02, 220 },
        { 41, 0x02, 220 },
        { 56, 0x03, 220 },
    },
    /* 252 */
    {
        { 3, 0x02, 249 },
        { 6, 0x02, 249 },
        { 10, 0x02, 249 },
        { 15, 0x02, 249 },
        { 24, 0x02, 249 },
        { 31, 0x02, 249 },
        { 41, 0x02, 249 },
        { 56, 0x03, 249 },
        { 1, 0x02, 10 },
        { 22, 0x03, 10 },
        { 1, 0x02, 13 },
        { 22, 0x03, 13 },
        { 1, 0x02, 22 },
        { 22, 0x03, 22 },
        { 0, 0x04, 0 },
        { 0, 0x04, 0 },
    },
    /* 253 */
    {
        { 2, 0x02, 10 },
        { 9, 0x02, 10 },
        { 23, 0x02, 10 },
        { 40, 0x03, 10 },
        { 2, 0x02, 13 },
        { 9, 0x02, 13 },
        { 23, 0x02, 13 },
        { 40, 0x03, 13 },
        { 2, 0x02, 22 },
        { 9, 0x02, 22 },
        { 23, 0x02, 22 },
        { 40, 0x03, 22 },
        { 0, 0x04, 0 },
        { 0, 0x04, 0 },
        { 0, 0x04, 0 },
        { 0, 0x04, 0 },
    },
    /* 254 */
    {
        { 3, 0x02, 10 },
        { 6, 0x02, 10 },
        { 10, 0x02, 10 },
        { 15, 0x02, 10 },
        { 24, 0x02, 10 },
        { 31, 0x02, 10 },
        { 41, 0x02, 10 },
        { 56, 0x03, 10 },
        { 3, 0x02, 13 },
        { 6, 0x02, 13 },
        { 10, 0x02, 13 },
        { 15, 0x02, 13 },
        { 24, 0x02, 13 },
        { 31, 0x02, 13 },
        { 41, 0x02, 13 },
        { 56, 0x03, 13 },
    },
    /* 255 */
    {
        { 3, 0x02, 22 },
        { 6, 0x02, 22 },
        { 10, 0x02, 22 },
        { 15, 0x02, 22 },
        { 24, 0x02, 22 },
        { 31, 0x02, 22 },
        { 41, 0x02, 22 },
        { 56, 0x03, 22 },
        { 0, 0x04, 0 },
        { 0, 0x04, 0 },
        { 0, 0x04, 0 },
        { 0, 0x04, 0 },
        { 0, 0x04, 0 },
        { 0, 0x04, 0 },
        { 0, 0x04, 0 },
        { 0, 0x04, 0 },
    },
};

#define lshpack_arr_init(a) do {                                        \
    memset((a), 0, sizeof(*(a)));                                       \
} while (0)

#define lshpack_arr_cleanup(a) do {                                     \
    free((a)->els);                                                     \
    memset((a), 0, sizeof(*(a)));                                       \
} while (0)

#define lshpack_arr_get(a, i) (                                         \
    assert((i) < (a)->nelem),                                           \
    (a)->els[(a)->off + (i)]                                            \
)

#define lshpack_arr_shift(a) (                                          \
    assert((a)->nelem > 0),                                             \
    (a)->nelem -= 1,                                                    \
    (a)->els[(a)->off++]                                                \
)

#define lshpack_arr_pop(a) (                                            \
    assert((a)->nelem > 0),                                             \
    (a)->nelem -= 1,                                                    \
    (a)->els[(a)->off + (a)->nelem]                                     \
)

#define lshpack_arr_count(a) (+(a)->nelem)

static int
lshpack_arr_push (struct lshpack_arr *arr, uintptr_t val)
{
    uintptr_t *new_els;
    unsigned n;

    if (arr->off + arr->nelem < arr->nalloc)
    {
        arr->els[arr->off + arr->nelem] = val;
        ++arr->nelem;
        return 0;
    }

    if (arr->off > arr->nalloc / 2)
    {
        memmove(arr->els, arr->els + arr->off,
                                        sizeof(arr->els[0]) * arr->nelem);
        arr->off = 0;
        arr->els[arr->nelem] = val;
        ++arr->nelem;
        return 0;
    }

    if (arr->nalloc)
        n = arr->nalloc * 2;
    else
        n = 64;
    new_els = malloc(n * sizeof(arr->els[0]));
    if (!new_els)
        return -1;
    memcpy(new_els, arr->els + arr->off, sizeof(arr->els[0]) * arr->nelem);
    free(arr->els);
    arr->off = 0;
    arr->els = new_els;
    arr->nalloc = n;
    arr->els[arr->off + arr->nelem] = val;
    ++arr->nelem;
    return 0;
}

struct lshpack_double_enc_head
{
    struct lshpack_enc_head by_name;
    struct lshpack_enc_head by_nameval;
};

struct lshpack_enc_table_entry
{
    /* An entry always lives on all three lists */
    STAILQ_ENTRY(lshpack_enc_table_entry)
                                    ete_next_nameval,
                                    ete_next_name,
                                    ete_next_all;
    unsigned                        ete_id;
    unsigned                        ete_nameval_hash;
    unsigned                        ete_name_hash;
    unsigned                        ete_name_len;
    unsigned                        ete_val_len;
    char                            ete_buf[0];
};

#define ETE_NAME(ete) ((ete)->ete_buf)
#define ETE_VALUE(ete) (&(ete)->ete_buf[(ete)->ete_name_len])


#define N_BUCKETS(n_bits) (1U << (n_bits))
#define BUCKNO(n_bits, hash) ((hash) & (N_BUCKETS(n_bits) - 1))


/* We estimate average number of entries in the dynamic table to be 1/3
 * of the theoretical maximum.  This number is used to size the history
 * buffer: we want it large enough to cover recent entries, yet not too
 * large to cover entries that appear with a period larger than the
 * dynamic table.
 */
static unsigned
henc_hist_size (unsigned max_capacity)
{
    return max_capacity / DYNAMIC_ENTRY_OVERHEAD / 3;
}


int
lshpack_enc_init (struct lshpack_enc *enc)
{
    struct lshpack_double_enc_head *buckets;
    unsigned nbits = 2;
    unsigned i;

    buckets = malloc(sizeof(buckets[0]) * N_BUCKETS(nbits));
    if (!buckets)
        return -1;

    for (i = 0; i < N_BUCKETS(nbits); ++i)
    {
        STAILQ_INIT(&buckets[i].by_name);
        STAILQ_INIT(&buckets[i].by_nameval);
    }

    memset(enc, 0, sizeof(*enc));
    STAILQ_INIT(&enc->hpe_all_entries);
    enc->hpe_max_capacity = INITIAL_DYNAMIC_TABLE_SIZE;
    enc->hpe_buckets      = buckets;
    /* The initial value of the entry ID is completely arbitrary.  As long as
     * there are fewer than 2^32 dynamic table entries, the math to calculate
     * the entry ID works.  To prove to ourselves that the wraparound works
     * and to have the unit tests cover it, we initialize the next ID so that
     * it is just about to wrap around.
     */
    enc->hpe_next_id      = ~0 - 3;
    enc->hpe_nbits        = nbits;
    enc->hpe_nelem        = 0;
    return 0;
}


void
lshpack_enc_cleanup (struct lshpack_enc *enc)
{
    struct lshpack_enc_table_entry *entry, *next;
    for (entry = STAILQ_FIRST(&enc->hpe_all_entries); entry; entry = next)
    {
        next = STAILQ_NEXT(entry, ete_next_all);
        free(entry);
    }
    free(enc->hpe_hist_buf);
    free(enc->hpe_buckets);
}


static int
henc_use_hist (struct lshpack_enc *enc)
{
    unsigned hist_size;

    if (enc->hpe_hist_buf)
        return 0;

    hist_size = henc_hist_size(INITIAL_DYNAMIC_TABLE_SIZE);
    if (!hist_size)
        return 0;

    enc->hpe_hist_buf = malloc(sizeof(enc->hpe_hist_buf[0]) * (hist_size + 1));
    if (!enc->hpe_hist_buf)
        return -1;

    enc->hpe_hist_size = hist_size;
    enc->hpe_flags |= LSHPACK_ENC_USE_HIST;
    return 0;
}


int
lshpack_enc_use_hist (struct lshpack_enc *enc, int on)
{
    if (on)
        return henc_use_hist(enc);
    else
    {
        enc->hpe_flags &= ~LSHPACK_ENC_USE_HIST;
        free(enc->hpe_hist_buf);
        enc->hpe_hist_buf = NULL;
        enc->hpe_hist_size = 0;
        enc->hpe_hist_idx = 0;
        enc->hpe_hist_wrapped = 0;
        return 0;
    }
}


int
lshpack_enc_hist_used (const struct lshpack_enc *enc)
{
    return (enc->hpe_flags & LSHPACK_ENC_USE_HIST) != 0;
}


#define LSHPACK_XXH_SEED 0
#define XXH_NAMEVAL_WIDTH 9
#define XXH_NAMEVAL_SHIFT 2
#define XXH_NAME_WIDTH 9
#define XXH_NAME_SHIFT 9

static const unsigned char nameval2id[ 1 << XXH_NAMEVAL_WIDTH ] =
{
    [475]  =  2,   [361]  =  3,   [149]  =  4,   [219]  =  5,   [511]  =  6,
    [299]  =  7,   [298]  =  8,   [19]   =  9,   [370]  =  10,  [95]   =  11,
    [158]  =  12,  [281]  = 13,   [423]  =  14,  [501]  =  16,
};

static const unsigned char name2id[ 1 << XXH_NAME_WIDTH ] =
{
    [215]  =  1,   [207]  =  2,   [286]  =  4,   [321]  =  6,   [289]  =  8,
    [63]   =  15,  [122]  =  16,  [209]  =  17,  [194]  =  18,  [331]  =  19,
    [273]  =  20,  [278]  =  21,  [431]  =  22,  [232]  =  23,  [182]  =  24,
    [483]  =  25,  [59]   =  26,  [52]   =  27,  [422]  =  28,  [62]   =  29,
    [381]  =  30,  [233]  =  31,  [110]  =  32,  [15]   =  33,  [85]   =  34,
    [452]  =  35,  [28]   =  36,  [414]  =  37,  [82]   =  38,  [345]  =  39,
    [467]  =  40,  [501]  =  41,  [299]  =  42,  [168]  =  43,  [326]  =  44,
    [199]  =  45,  [360]  =  46,  [36]   =  47,  [349]  =  48,  [347]  =  49,
    [126]  =  50,  [208]  =  51,  [416]  =  52,  [373]  =  53,  [477]  =  54,
    [33]   =  55,  [316]  =  56,  [156]  =  57,  [386]  =  58,  [116]  =  59,
    [450]  =  60,  [297]  =  61,
};

//not find return 0, otherwise return the index
#if !LS_HPACK_EMIT_TEST_CODE
static
#endif
       unsigned
lshpack_enc_get_static_nameval (uint32_t nameval_hash, const char *name,
        unsigned name_len, const char *val, unsigned val_len)
{
    unsigned i;

    i = (nameval_hash >> XXH_NAMEVAL_SHIFT) & ((1 << XXH_NAMEVAL_WIDTH) - 1);
    if (nameval2id[i])
    {
        i = nameval2id[i] - 1;
        if (static_table[i].name_len == name_len
            && static_table[i].val_len == val_len
            && memcmp(name, static_table[i].name, name_len) == 0
            && memcmp(val, static_table[i].val, val_len) == 0)
        {
            return i + 1;
        }
    }

    return 0;
}

#if !LS_HPACK_EMIT_TEST_CODE
static
#endif
       unsigned
lshpack_enc_get_static_name (uint32_t name_hash, const char *name,
                                            unsigned name_len)
{
    unsigned i;

    i = (name_hash >> XXH_NAME_SHIFT) & ((1 << XXH_NAME_WIDTH) - 1);
    if (name2id[i])
    {
        i = name2id[i] - 1;
        if (static_table[i].name_len == name_len
            && memcmp(name, static_table[i].name, name_len) == 0)
        {
            return i + 1;
        }
    }

    return 0;
}


unsigned
lshpack_enc_get_stx_tab_id (const char *name, unsigned name_len,
                                    const char *val, unsigned val_len)
{
    uint32_t name_hash, nameval_hash;
    unsigned i;

    name_hash = XXH32(name, name_len, 0);
    nameval_hash = XXH32(val, val_len, name_hash);

    i = (nameval_hash >> XXH_NAMEVAL_SHIFT) & ((1 << XXH_NAMEVAL_WIDTH) - 1);
    if (nameval2id[i])
    {
        i = nameval2id[i] - 1;
        if (static_table[i].name_len == name_len
            && static_table[i].val_len == val_len
            && memcmp(name, static_table[i].name, name_len) == 0
            && memcmp(val, static_table[i].val, val_len) == 0)
        {
            return i + 1;
        }
    }

    i = (name_hash >> XXH_NAME_SHIFT) & ((1 << XXH_NAME_WIDTH) - 1);
    if (name2id[i])
    {
        i = name2id[i] - 1;
        if (static_table[i].name_len == name_len
            && memcmp(name, static_table[i].name, name_len) == 0)
        {
            return i + 1;
        }
    }

    return 0;
}


/* Given a dynamic entry, return its table ID */
static unsigned
henc_calc_table_id (const struct lshpack_enc *enc,
                                    const struct lshpack_enc_table_entry *entry)
{
    return HPACK_STATIC_TABLE_SIZE
         + (enc->hpe_next_id - entry->ete_id)
    ;
}


static unsigned
henc_find_table_id (struct lshpack_enc *enc, uint32_t name_hash,
        uint32_t nameval_hash, const char *name,
        unsigned name_len, const char *value,
        unsigned value_len, int *val_matched)
{
    struct lshpack_enc_table_entry *entry;
    unsigned buckno, static_table_id;

    /* First, look for a match in the static table: */
    static_table_id = lshpack_enc_get_static_nameval(nameval_hash, name,
                                            name_len, value, value_len);
    if (static_table_id > 0)
    {
        *val_matched = 1;
        return static_table_id;
    }

    buckno = BUCKNO(enc->hpe_nbits, nameval_hash);
    STAILQ_FOREACH(entry, &enc->hpe_buckets[buckno].by_nameval,
                                                        ete_next_nameval)
        if (nameval_hash == entry->ete_nameval_hash &&
            name_len == entry->ete_name_len &&
            value_len == entry->ete_val_len &&
            0 == memcmp(name, ETE_NAME(entry), name_len) &&
            0 == memcmp(value, ETE_VALUE(entry), value_len))
        {
            *val_matched = 1;
            return henc_calc_table_id(enc, entry);
        }

    static_table_id = lshpack_enc_get_static_name(name_hash, name, name_len);
    if (static_table_id > 0)
    {
        *val_matched = 0;
        return static_table_id;
    }

    /* Search by name only: */
    buckno = BUCKNO(enc->hpe_nbits, name_hash);
    STAILQ_FOREACH(entry, &enc->hpe_buckets[buckno].by_name, ete_next_name)
        if (name_hash == entry->ete_name_hash &&
            name_len == entry->ete_name_len &&
            0 == memcmp(name, ETE_NAME(entry), name_len))
        {
            *val_matched = 0;
            return henc_calc_table_id(enc, entry);
        }

    return 0;
}


#if !LS_HPACK_EMIT_TEST_CODE
static
#endif
       unsigned char *
lshpack_enc_enc_int (unsigned char *dst, unsigned char *const end,
                                        uint32_t value, uint8_t prefix_bits)
{
    unsigned char *const dst_orig = dst;

    /* This function assumes that at least one byte is available */
    assert(dst < end);
    if (value < (uint32_t)(1 << prefix_bits) - 1)
        *dst++ |= value;
    else
    {
        *dst++ |= (1 << prefix_bits) - 1;
        value -= (1 << prefix_bits) - 1;
        while (value >= 128)
        {
            if (dst < end)
            {
                *dst++ = (0x80 | value);
                value >>= 7;
            }
            else
                return dst_orig;
        }
        if (dst < end)
            *dst++ = value;
        else
            return dst_orig;
    }
    return dst;
}


int
lshpack_enc_huff_encode (const unsigned char *src,
    const unsigned char *const src_end, unsigned char *const dst, int dst_len)
{
    unsigned char *p_dst = dst;
    unsigned char *dst_end = p_dst + dst_len;
    uintptr_t bits;  /* OK not to initialize this variable */
    unsigned bits_used = 0, adj;
    struct encode_el cur_enc_code;

    while (src != src_end)
    {
        cur_enc_code = encode_table[*src++];
        if (bits_used + cur_enc_code.bits < sizeof(bits) * 8)
        {
            bits <<= cur_enc_code.bits;
            bits |= cur_enc_code.code;
            bits_used += cur_enc_code.bits;
            continue;
        }
        else if (p_dst + sizeof(bits) <= dst_end)
        {
            bits <<= sizeof(bits) * 8 - bits_used;
            bits_used = cur_enc_code.bits - (sizeof(bits) * 8 - bits_used);
            bits |= cur_enc_code.code >> bits_used;
#if UINTPTR_MAX == 18446744073709551615ull
            *p_dst++ = bits >> 56;
            *p_dst++ = bits >> 48;
            *p_dst++ = bits >> 40;
            *p_dst++ = bits >> 32;
#endif
            *p_dst++ = bits >> 24;
            *p_dst++ = bits >> 16;
            *p_dst++ = bits >> 8;
            *p_dst++ = bits;
            bits = cur_enc_code.code;   /* OK not to clear high bits */
        }
        else
            return -1;
    }

    adj = bits_used + (-bits_used & 7);     /* Round up to 8 */
    if (bits_used && p_dst + (adj >> 3) <= dst_end)
    {
        bits <<= -bits_used & 7;            /* Align to byte boundary */
        bits |= ((1 << (-bits_used & 7)) - 1);  /* EOF */
        switch (adj >> 3)
        {                               /* Write out */
#if UINTPTR_MAX == 18446744073709551615ull
        case 8: *p_dst++ = bits >> 56;
        case 7: *p_dst++ = bits >> 48;
        case 6: *p_dst++ = bits >> 40;
        case 5: *p_dst++ = bits >> 32;
#endif
        case 4: *p_dst++ = bits >> 24;
        case 3: *p_dst++ = bits >> 16;
        case 2: *p_dst++ = bits >> 8;
        default: *p_dst++ = bits;
        }
        return p_dst - dst;
    }
    else if (p_dst + (adj >> 3) <= dst_end)
        return p_dst - dst;
    else
        return -1;
}


#if !LS_HPACK_EMIT_TEST_CODE
static
#endif
       int
lshpack_enc_enc_str (unsigned char *const dst, size_t dst_len,
                        const unsigned char *str, unsigned str_len)
{
    unsigned char size_buf[4];
    unsigned char *p;
    unsigned size_len;
    int rc;

    if (dst_len > 1)
        /* We guess that the string size fits into a single byte -- meaning
         * compressed string of size 126 and smaller -- which is the normal
         * case.  Thus, we immediately write compressed string to the output
         * buffer.  If our guess is not correct, we fix it later.
         */
        rc = lshpack_enc_huff_encode(str, str + str_len, dst + 1, dst_len - 1);
    else if (dst_len == 1)
        /* Here, the call can only succeed if the string to encode is empty. */
        rc = 0;
    else
        return -1;

    /*
     * Check if need huffman encoding or not
     * Comment: (size_t)rc <= str_len   = means if same length, still use
     *                                                              Huffman
     *                     ^
     */
    if (rc > 0 && (size_t)rc <= str_len)
    {
        if (rc < 127)
        {
            *dst = 0x80 | rc;
            return 1 + rc;
        }
        size_buf[0] = 0x80;
        str_len = rc;
        str = dst + 1;
    }
    else if (str_len <= dst_len - 1)
    {
        if (str_len < 127)
        {
            *dst = (unsigned char) str_len;
            memcpy(dst + 1, str, str_len);
            return 1 + str_len;
        }
        size_buf[0] = 0x00;
    }
    else
        return -1;

    /* The guess of one-byte size was incorrect.  Perform necessary
     * adjustments.
     */
    p = lshpack_enc_enc_int(size_buf, size_buf + sizeof(size_buf), str_len, 7);
    if (p == size_buf)
        return -1;

    size_len = p - size_buf;
    assert(size_len > 1);

    /* Check if there is enough room in the output buffer for both
     * encoded size and the string.
     */
    if (size_len + str_len > dst_len)
        return -1;

    memmove(dst + size_len, str, str_len);
    memcpy(dst, size_buf, size_len);
    return size_len + str_len;
}


static void
henc_drop_oldest_entry (struct lshpack_enc *enc)
{
    struct lshpack_enc_table_entry *entry;
    unsigned buckno;

    entry = STAILQ_FIRST(&enc->hpe_all_entries);
    assert(entry);
    STAILQ_REMOVE_HEAD(&enc->hpe_all_entries, ete_next_all);
    buckno = BUCKNO(enc->hpe_nbits, entry->ete_nameval_hash);
    assert(entry == STAILQ_FIRST(&enc->hpe_buckets[buckno].by_nameval));
    STAILQ_REMOVE_HEAD(&enc->hpe_buckets[buckno].by_nameval, ete_next_nameval);
    buckno = BUCKNO(enc->hpe_nbits, entry->ete_name_hash);
    assert(entry == STAILQ_FIRST(&enc->hpe_buckets[buckno].by_name));
    STAILQ_REMOVE_HEAD(&enc->hpe_buckets[buckno].by_name, ete_next_name);

    enc->hpe_cur_capacity -= DYNAMIC_ENTRY_OVERHEAD + entry->ete_name_len
                                                        + entry->ete_val_len;
    --enc->hpe_nelem;
    free(entry);
}


static void
henc_remove_overflow_entries (struct lshpack_enc *enc)
{
    while (enc->hpe_cur_capacity > enc->hpe_max_capacity)
        henc_drop_oldest_entry(enc);
}


static int
henc_grow_tables (struct lshpack_enc *enc)
{
    struct lshpack_double_enc_head *new_buckets, *new[2];
    struct lshpack_enc_table_entry *entry;
    unsigned n, old_nbits;
    int idx;

    old_nbits = enc->hpe_nbits;
    new_buckets = malloc(sizeof(enc->hpe_buckets[0])
                                                * N_BUCKETS(old_nbits + 1));
    if (!new_buckets)
        return -1;

    for (n = 0; n < N_BUCKETS(old_nbits); ++n)
    {
        new[0] = &new_buckets[n];
        new[1] = &new_buckets[n + N_BUCKETS(old_nbits)];
        STAILQ_INIT(&new[0]->by_name);
        STAILQ_INIT(&new[1]->by_name);
        STAILQ_INIT(&new[0]->by_nameval);
        STAILQ_INIT(&new[1]->by_nameval);
        while ((entry = STAILQ_FIRST(&enc->hpe_buckets[n].by_name)))
        {
            STAILQ_REMOVE_HEAD(&enc->hpe_buckets[n].by_name, ete_next_name);
            idx = (BUCKNO(old_nbits + 1, entry->ete_name_hash)
                                                        >> old_nbits) & 1;
            STAILQ_INSERT_TAIL(&new[idx]->by_name, entry, ete_next_name);
        }
        while ((entry = STAILQ_FIRST(&enc->hpe_buckets[n].by_nameval)))
        {
            STAILQ_REMOVE_HEAD(&enc->hpe_buckets[n].by_nameval,
                                                        ete_next_nameval);
            idx = (BUCKNO(old_nbits + 1, entry->ete_nameval_hash)
                                                        >> old_nbits) & 1;
            STAILQ_INSERT_TAIL(&new[idx]->by_nameval, entry,
                                                        ete_next_nameval);
        }
    }

    free(enc->hpe_buckets);
    enc->hpe_nbits   = old_nbits + 1;
    enc->hpe_buckets = new_buckets;
    return 0;
}

#if !LS_HPACK_EMIT_TEST_CODE
static
#endif
       int
lshpack_enc_push_entry (struct lshpack_enc *enc, uint32_t name_hash,
        uint32_t nameval_hash, const char *name, unsigned name_len,
        const char *value, unsigned value_len)
{
    struct lshpack_enc_table_entry *entry;
    unsigned buckno;
    size_t size;

    if (enc->hpe_nelem >= N_BUCKETS(enc->hpe_nbits) / 2 &&
                                                0 != henc_grow_tables(enc))
        return -1;

    size = sizeof(*entry) + name_len + value_len;
    entry = malloc(size);
    if (!entry)
        return -1;

    entry->ete_name_hash = name_hash;
    entry->ete_nameval_hash = nameval_hash;
    entry->ete_name_len = name_len;
    entry->ete_val_len = value_len;
    entry->ete_id = enc->hpe_next_id++;
    memcpy(ETE_NAME(entry), name, name_len);
    memcpy(ETE_VALUE(entry), value, value_len);

    STAILQ_INSERT_TAIL(&enc->hpe_all_entries, entry, ete_next_all);
    buckno = BUCKNO(enc->hpe_nbits, nameval_hash);
    STAILQ_INSERT_TAIL(&enc->hpe_buckets[buckno].by_nameval, entry,
                                                        ete_next_nameval);
    buckno = BUCKNO(enc->hpe_nbits, name_hash);
    STAILQ_INSERT_TAIL(&enc->hpe_buckets[buckno].by_name, entry,
                                                        ete_next_name);

    enc->hpe_cur_capacity += DYNAMIC_ENTRY_OVERHEAD + name_len + value_len;
    ++enc->hpe_nelem;
    henc_remove_overflow_entries(enc);
    return 0;
}


static void
henc_resize_history (struct lshpack_enc *enc)
{
    uint32_t *hist_buf;
    unsigned hist_size, first, count, i, j;

    hist_size = henc_hist_size(enc->hpe_max_capacity);

    if (hist_size == enc->hpe_hist_size)
        return;

    if (hist_size == 0)
    {
        free(enc->hpe_hist_buf);
        enc->hpe_hist_buf = NULL;
        enc->hpe_hist_size = 0;
        enc->hpe_hist_idx = 0;
        enc->hpe_hist_wrapped = 0;
        return;
    }

    hist_buf = malloc(sizeof(hist_buf[0]) * (hist_size + 1));
    if (!hist_buf)
        return;

    if (enc->hpe_hist_wrapped)
    {
        first = (enc->hpe_hist_idx + 1) % enc->hpe_hist_size;
        count = enc->hpe_hist_size;
    }
    else
    {
        first = 0;
        count = enc->hpe_hist_idx;
    }
    for (i = 0, j = 0; count > 0 && j < hist_size; ++i, ++j, --count)
        hist_buf[j] = enc->hpe_hist_buf[ (first + i) % enc->hpe_hist_size ];
    enc->hpe_hist_size = hist_size;
    enc->hpe_hist_idx = j % hist_size;
    enc->hpe_hist_wrapped = enc->hpe_hist_idx == 0;
    free(enc->hpe_hist_buf);
    enc->hpe_hist_buf = hist_buf;
}


/* Returns true if `nameval_hash' was already in history, false otherwise. */
static int
henc_hist_add (struct lshpack_enc *enc, uint32_t nameval_hash)
{
    unsigned last;
    uint32_t *p;

    if (enc->hpe_hist_wrapped)
        last = enc->hpe_hist_size;
    else
        last = enc->hpe_hist_idx;

    enc->hpe_hist_buf[ last ] = nameval_hash;
    for (p = enc->hpe_hist_buf; *p != nameval_hash; ++p)
        ;
    enc->hpe_hist_buf[ enc->hpe_hist_idx ] = nameval_hash;
    enc->hpe_hist_idx = (enc->hpe_hist_idx + 1) % enc->hpe_hist_size;
    enc->hpe_hist_wrapped |= enc->hpe_hist_idx == 0;

    return p < enc->hpe_hist_buf + last;
}


unsigned char *
lshpack_enc_encode2 (struct lshpack_enc *enc, unsigned char *dst,
        unsigned char *dst_end, const char *name, unsigned name_len,
        const char *value, unsigned value_len, int indexed_type)
{
    //indexed_type: 0, Add, 1,: without, 2: never
    static const char indexed_prefix_number[] = {0x40, 0x00, 0x10};
    unsigned char *const dst_org = dst;
    uint32_t name_hash, nameval_hash;
    int val_matched, rc;
    unsigned table_id;
    XXH32_state_t hash_state;

    assert(indexed_type >= 0 && indexed_type <= 2);

    if (dst_end <= dst)
        return dst_org;

    XXH32_reset(&hash_state, LSHPACK_XXH_SEED);
    XXH32_update(&hash_state, name, name_len);
    name_hash = XXH32_digest(&hash_state);
    XXH32_update(&hash_state, value, value_len);
    nameval_hash = XXH32_digest(&hash_state);

    if (enc->hpe_hist_buf)
    {
        rc = henc_hist_add(enc, nameval_hash);
        if (!rc && enc->hpe_hist_wrapped && indexed_type == 0)
            indexed_type = 1;
    }

    table_id = henc_find_table_id(enc, name_hash, nameval_hash, name,
                                    name_len, value, value_len, &val_matched);
    if (table_id > 0)
    {
        if (val_matched)
        {
            *dst = 0x80;
            dst = lshpack_enc_enc_int(dst, dst_end, table_id, 7);
            /* No need to check return value: we pass it up as-is because
             * the behavior is the same.
             */
            return dst;
        }
        else
        {
            *dst = indexed_prefix_number[indexed_type];
            dst = lshpack_enc_enc_int(dst, dst_end, table_id,
                                            ((indexed_type == 0) ? 6 : 4));
            if (dst == dst_org)
                return dst_org;
        }
    }
    else
    {
        *dst++ = indexed_prefix_number[indexed_type];
        rc = lshpack_enc_enc_str(dst, dst_end - dst,
                                    (const unsigned char *)name, name_len);
        if (rc < 0)
            return dst_org; //Failed to enc this header, return unchanged ptr.
        dst += rc;
    }

    rc = lshpack_enc_enc_str(dst, dst_end - dst,
                                    (const unsigned char *)value, value_len);
    if (rc < 0)
        return dst_org; //Failed to enc this header, return unchanged ptr.
    dst += rc;

    if (indexed_type == 0)
    {
        rc = lshpack_enc_push_entry(enc, name_hash, nameval_hash, name,
                                                name_len, value, value_len);
        if (rc != 0)
            return dst_org; //Failed to enc this header, return unchanged ptr.
    }

    return dst;
}


unsigned char *
lshpack_enc_encode (struct lshpack_enc *henc, unsigned char *dst,
                     unsigned char *dst_end, int hpack_idx,
                     const lshpack_header_t *hdr, int indexed_type)
{
    return lshpack_enc_encode2(henc, dst, dst_end, hdr->name.iov_base,
        hdr->name.iov_len, hdr->value.iov_base, hdr->value.iov_len, indexed_type);
}


void
lshpack_enc_set_max_capacity (struct lshpack_enc *enc, unsigned max_capacity)
{
    enc->hpe_max_capacity = max_capacity;
    henc_remove_overflow_entries(enc);
    if (lshpack_enc_hist_used(enc))
        henc_resize_history(enc);
}

#if LS_HPACK_EMIT_TEST_CODE
void
lshpack_enc_iter_init (struct lshpack_enc *enc, void **iter)
{
    *iter = STAILQ_FIRST(&enc->hpe_all_entries);
}


/* Returns 0 if entry is found */
int
lshpack_enc_iter_next (struct lshpack_enc *enc, void **iter,
                                        struct enc_dyn_table_entry *retval)
{
    const struct lshpack_enc_table_entry *entry;

    entry = *iter;
    if (!entry)
        return -1;

    *iter = STAILQ_NEXT(entry, ete_next_all);

    retval->name = ETE_NAME(entry);
    retval->value = ETE_VALUE(entry);
    retval->name_len = entry->ete_name_len;
    retval->value_len = entry->ete_val_len;
    retval->entry_id = henc_calc_table_id(enc, entry);
    return 0;
}
#endif


/* Dynamic table entry: */
struct dec_table_entry
{
    unsigned    dte_name_len;
    unsigned    dte_val_len;
    uint8_t     dte_name_idx;
    char        dte_buf[0];     /* Contains both name and value */
};

#define DTE_NAME(dte) ((dte)->dte_buf)
#define DTE_VALUE(dte) (&(dte)->dte_buf[(dte)->dte_name_len])

enum
{
    HPACK_HUFFMAN_FLAG_ACCEPTED = 0x01,
    HPACK_HUFFMAN_FLAG_SYM = 0x02,
    HPACK_HUFFMAN_FLAG_FAIL = 0x04,
};

struct decode_status
{
    uint8_t state;
    uint8_t eos;
};


void
lshpack_dec_init (struct lshpack_dec *dec)
{
    memset(dec, 0, sizeof(*dec));
    dec->hpd_max_capacity = INITIAL_DYNAMIC_TABLE_SIZE;
    dec->hpd_cur_max_capacity = INITIAL_DYNAMIC_TABLE_SIZE;
    lshpack_arr_init(&dec->hpd_dyn_table);
}


void
lshpack_dec_cleanup (struct lshpack_dec *dec)
{
    uintptr_t val;

    while (lshpack_arr_count(&dec->hpd_dyn_table) > 0)
    {
        val = lshpack_arr_pop(&dec->hpd_dyn_table);
        free((struct dec_table_entry *) val);
    }
    lshpack_arr_cleanup(&dec->hpd_dyn_table);
}


/* Maximum number of bytes required to encode a 32-bit integer */
#define LSHPACK_UINT32_ENC_SZ 6


/* Assumption: we have at least one byte to work with */
#if !LS_HPACK_EMIT_TEST_CODE
static
#endif
       int
lshpack_dec_dec_int (const unsigned char **src_p, const unsigned char *src_end,
                                        unsigned prefix_bits, uint32_t *value_p)
{
    const unsigned char *const orig_src = *src_p;
    const unsigned char *src;
    unsigned prefix_max, M;
    uint32_t val, B;

    src = *src_p;

    prefix_max = (1 << prefix_bits) - 1;
    val = *src++;
    val &= prefix_max;

    if (val < prefix_max)
    {
        *src_p = src;
        *value_p = val;
        return 0;
    }

    M = 0;
    do
    {
        if (src < src_end)
        {
            B = *src++;
            val = val + ((B & 0x7f) << M);
            M += 7;
        }
        else if (src - orig_src < LSHPACK_UINT32_ENC_SZ)
            return -1;
        else
            return -2;
    }
    while (B & 0x80);

    if (M <= 28 || (M == 35 && src[-1] <= 0xF && val - (src[-1] << 28) < val))
    {
        *src_p = src;
        *value_p = val;
        return 0;
    }
    else
        return -2;
}


static void
hdec_drop_oldest_entry (struct lshpack_dec *dec)
{
    struct dec_table_entry *entry;
    entry = (void *) lshpack_arr_shift(&dec->hpd_dyn_table);
    dec->hpd_cur_capacity -= DYNAMIC_ENTRY_OVERHEAD + entry->dte_name_len
                                                        + entry->dte_val_len;
    free(entry);
}


static void
hdec_remove_overflow_entries (struct lshpack_dec *dec)
{
    while (dec->hpd_cur_capacity > dec->hpd_cur_max_capacity)
        hdec_drop_oldest_entry(dec);
}


static void
hdec_update_max_capacity (struct lshpack_dec *dec, uint32_t new_capacity)
{
    dec->hpd_cur_max_capacity = new_capacity;
    hdec_remove_overflow_entries(dec);
}


void
lshpack_dec_set_max_capacity (struct lshpack_dec *dec, unsigned max_capacity)
{
    dec->hpd_max_capacity = max_capacity;
    hdec_update_max_capacity(dec, max_capacity);
}


static unsigned char *
hdec_huff_dec4bits (uint8_t src_4bits, unsigned char *dst,
                                        struct decode_status *status)
{
    const struct decode_el cur_dec_code =
        decode_tables[status->state][src_4bits];
    if (cur_dec_code.flags & HPACK_HUFFMAN_FLAG_FAIL) {
        return NULL; //failed
    }
    if (cur_dec_code.flags & HPACK_HUFFMAN_FLAG_SYM)
    {
        *dst = cur_dec_code.sym;
        dst++;
    }

    status->state = cur_dec_code.state;
    status->eos = ((cur_dec_code.flags & HPACK_HUFFMAN_FLAG_ACCEPTED) != 0);
    return dst;
}


int
lshpack_dec_huff_decode_full (const unsigned char *src, int src_len,
                                            unsigned char *dst, int dst_len)
{
    const unsigned char *p_src = src;
    const unsigned char *const src_end = src + src_len;
    unsigned char *p_dst = dst;
    unsigned char *dst_end = dst + dst_len;
    struct decode_status status = { 0, 1 };

    while (p_src != src_end)
    {
        if (p_dst == dst_end)
            return -2;
        if ((p_dst = hdec_huff_dec4bits(*p_src >> 4, p_dst, &status))
                == NULL)
            return -1;
        if (p_dst == dst_end)
            return -2;
        if ((p_dst = hdec_huff_dec4bits(*p_src & 0xf, p_dst, &status))
                == NULL)
            return -1;
        ++p_src;
    }

    if (!status.eos)
        return -1;

    return p_dst - dst;
}


int
lshpack_dec_huff_decode (const unsigned char *src, int src_len,
                                            unsigned char *dst, int dst_len);


//reutrn the length in the dst, also update the src
#if !LS_HPACK_EMIT_TEST_CODE
static
#endif
       int
hdec_dec_str (unsigned char *dst, size_t dst_len, const unsigned char **src,
        const unsigned char *src_end)
{
    if ((*src) == src_end)
        return 0;

    int is_huffman = (*(*src) & 0x80);
    uint32_t len;
    if (0 != lshpack_dec_dec_int(src, src_end, 7, &len))
        return -2;  //wrong int

    int ret = 0;
    if ((uint32_t)(src_end - (*src)) < len) {
        return -2;  //wrong int
    }

    if (is_huffman)
    {
        ret = lshpack_dec_huff_decode(*src, len, dst, dst_len);
        if (ret < 0)
            return -3; //Wrong code

        (*src) += len;
    }
    else
    {
        if (dst_len < len)
            ret = -3;  //dst not enough space
        else
        {
            memcpy(dst, (*src), len);
            (*src) += len;
            ret = len;
        }
    }

    return ret;
}


/* hpd_dyn_table is a dynamic array.  New entries are pushed onto it,
 * while old entries are shifted from it.
 */
static struct dec_table_entry *
hdec_get_table_entry (struct lshpack_dec *dec, uint32_t index)
{
    uintptr_t val;

    index -= HPACK_STATIC_TABLE_SIZE;
    if (index == 0 || index > lshpack_arr_count(&dec->hpd_dyn_table))
        return NULL;

    index = lshpack_arr_count(&dec->hpd_dyn_table) - index;
    val = lshpack_arr_get(&dec->hpd_dyn_table, index);
    return (struct dec_table_entry *) val;
}


#if !LS_HPACK_EMIT_TEST_CODE
static
#endif
       int
lshpack_dec_push_entry (struct lshpack_dec *dec, uint8_t name_idx, const char *name,
                        unsigned name_len, const char *val, unsigned val_len)
{
    struct dec_table_entry *entry;
    size_t size;

    size = sizeof(*entry) + name_len + val_len;
    entry = malloc(size);
    if (!entry)
        return -1;

    if (0 != lshpack_arr_push(&dec->hpd_dyn_table, (uintptr_t) entry))
    {
        free(entry);
        return -1;
    }

    dec->hpd_cur_capacity += DYNAMIC_ENTRY_OVERHEAD + name_len + val_len;
    entry->dte_name_len = name_len;
    entry->dte_val_len = val_len;
    entry->dte_name_idx = name_idx;
    memcpy(DTE_NAME(entry), name, name_len);
    memcpy(DTE_VALUE(entry), val, val_len);
    return 0;
}


int
lshpack_dec_decode (struct lshpack_dec *dec,
    const unsigned char **src, const unsigned char *src_end,
    char *dst, char *const dst_end, unsigned *name_len, unsigned *val_len,
    uint32_t *name_idx)
{
    struct dec_table_entry *entry;
    uint32_t index, new_capacity;
    int indexed_type, len;

    if ((*src) == src_end)
        return -1;

    while ((*(*src) & 0xe0) == 0x20)    //001 xxxxx
    {
        if (0 != lshpack_dec_dec_int(src, src_end, 5, &new_capacity))
            return -1;
        if (new_capacity > dec->hpd_max_capacity)
            return -1;
        hdec_update_max_capacity(dec, new_capacity);
        if (*src == src_end)
            return -1;
    }

    /* lshpack_dec_dec_int() sets `index' and advances `src'.  If we do not
     * call it, we set `index' and advance `src' ourselves:
     */
    if (*(*src) & 0x80) //1 xxxxxxx
    {
        if (0 != lshpack_dec_dec_int(src, src_end, 7, &index))
            return -1;
        if (index == 0)
            return -1;
        indexed_type = 3; //need to parse value
    }
    else if (*(*src) > 0x40) //01 xxxxxx
    {
        if (0 != lshpack_dec_dec_int(src, src_end, 6, &index))
            return -1;

        indexed_type = 0;
    }
    else if (*(*src) == 0x40) //custmized //0100 0000
    {
        indexed_type = 0;
        index = 0;
        ++(*src);
    }

    //Never indexed
    else if (*(*src) == 0x10)  //00010000
    {
        indexed_type = 2;
        index = 0;
        ++(*src);
    }
    else if ((*(*src) & 0xf0) == 0x10)  //0001 xxxx
    {
        if (0 != lshpack_dec_dec_int(src, src_end, 4, &index))
            return -1;

        indexed_type = 2;
    }

    //without indexed
    else if (*(*src) == 0x00)  //0000 0000
    {
        indexed_type = 1;
        index = 0;
        ++(*src);
    }
    else // 0000 xxxx
    {
        if (0 != lshpack_dec_dec_int(src, src_end, 4, &index))
            return -1;

        indexed_type = 1;
    }
    *name_idx = index;

    char *const name = dst;
    if (index > 0)
    {
        if (index <= HPACK_STATIC_TABLE_SIZE) //static table
        {
            if (static_table[index - 1].name_len > dst_end - dst)
                return -1;
            *name_len = static_table[index - 1].name_len;
            memcpy(name, static_table[index - 1].name, *name_len);
            if (indexed_type == 3)
            {
                if (static_table[index - 1].name_len +
                    static_table[index - 1].val_len > dst_end - dst)
                    return -1;
                *val_len = static_table[index - 1].val_len;
                memcpy(name + *name_len, static_table[index - 1].val, *val_len);
                return 0;
            }
        }
        else
        {
            entry = hdec_get_table_entry(dec, index);
            if (entry == NULL)
                return -1;
            if (entry->dte_name_len > dst_end - dst)
                return -1;

            *name_len = entry->dte_name_len;
            memcpy(name, DTE_NAME(entry), *name_len);
            if (entry->dte_name_idx)
                *name_idx = entry->dte_name_idx;
            else
                *name_idx = 0;
            if (indexed_type == 3)
            {
                if (entry->dte_name_len + entry->dte_val_len > dst_end - dst)
                    return -1;
                *val_len = entry->dte_val_len;
                memcpy(name + *name_len, DTE_VALUE(entry), *val_len);
                return 0;
            }
        }
    }
    else
    {
        len = hdec_dec_str((unsigned char *)name, dst_end - dst, src, src_end);
        if (len < 0)
            return len; //error
        *name_len = len;
    }

    len = hdec_dec_str((unsigned char *)name + *name_len,
                                    dst_end - dst - *name_len, src, src_end);
    if (len < 0)
        return len; //error
    *val_len = len;

    if (indexed_type == 0)
    {
        if (index > HPACK_STATIC_TABLE_SIZE)
            index = 0;
        if (0 != lshpack_dec_push_entry(dec, index, name, *name_len,
                                            name + *name_len, *val_len))
            return -1;  //error
    }

    return 0;
}


static const struct hdec { uint8_t lens; uint8_t out[3]; } hdecs[] =
{
/*
          ,------------- Number of bits consumed by the index
          |      ,------ Number of bytes in output; 0 means long or invalid code
          |      |  ,--- Output, aligned to the left
          V      V  V                                                         */
/* 0 */ {(15<<2)|3,{48,48,48}},
/* 1 */ {(15<<2)|3,{48,48,48}},
/* 2 */ {(15<<2)|3,{48,48,49}},
/* 3 */ {(15<<2)|3,{48,48,49}},
/* 4 */ {(15<<2)|3,{48,48,50}},
/* 5 */ {(15<<2)|3,{48,48,50}},
/* 6 */ {(15<<2)|3,{48,48,97}},
/* 7 */ {(15<<2)|3,{48,48,97}},
/* 8 */ {(15<<2)|3,{48,48,99}},
/* 9 */ {(15<<2)|3,{48,48,99}},
/* 10 */ {(15<<2)|3,{48,48,101}},
/* 11 */ {(15<<2)|3,{48,48,101}},
/* 12 */ {(15<<2)|3,{48,48,105}},
/* 13 */ {(15<<2)|3,{48,48,105}},
/* 14 */ {(15<<2)|3,{48,48,111}},
/* 15 */ {(15<<2)|3,{48,48,111}},
/* 16 */ {(15<<2)|3,{48,48,115}},
/* 17 */ {(15<<2)|3,{48,48,115}},
/* 18 */ {(15<<2)|3,{48,48,116}},
/* 19 */ {(15<<2)|3,{48,48,116}},
/* 20 */ {(16<<2)|3,{48,48,32}},
/* 21 */ {(16<<2)|3,{48,48,37}},
/* 22 */ {(16<<2)|3,{48,48,45}},
/* 23 */ {(16<<2)|3,{48,48,46}},
/* 24 */ {(16<<2)|3,{48,48,47}},
/* 25 */ {(16<<2)|3,{48,48,51}},
/* 26 */ {(16<<2)|3,{48,48,52}},
/* 27 */ {(16<<2)|3,{48,48,53}},
/* 28 */ {(16<<2)|3,{48,48,54}},
/* 29 */ {(16<<2)|3,{48,48,55}},
/* 30 */ {(16<<2)|3,{48,48,56}},
/* 31 */ {(16<<2)|3,{48,48,57}},
/* 32 */ {(16<<2)|3,{48,48,61}},
/* 33 */ {(16<<2)|3,{48,48,65}},
/* 34 */ {(16<<2)|3,{48,48,95}},
/* 35 */ {(16<<2)|3,{48,48,98}},
/* 36 */ {(16<<2)|3,{48,48,100}},
/* 37 */ {(16<<2)|3,{48,48,102}},
/* 38 */ {(16<<2)|3,{48,48,103}},
/* 39 */ {(16<<2)|3,{48,48,104}},
/* 40 */ {(16<<2)|3,{48,48,108}},
/* 41 */ {(16<<2)|3,{48,48,109}},
/* 42 */ {(16<<2)|3,{48,48,110}},
/* 43 */ {(16<<2)|3,{48,48,112}},
/* 44 */ {(16<<2)|3,{48,48,114}},
/* 45 */ {(16<<2)|3,{48,48,117}},
/* 46 */ {(10<<2)|2,{48,48,0}},
/* 47 */ {(10<<2)|2,{48,48,0}},
/* 48 */ {(10<<2)|2,{48,48,0}},
/* 49 */ {(10<<2)|2,{48,48,0}},
/* 50 */ {(10<<2)|2,{48,48,0}},
/* 51 */ {(10<<2)|2,{48,48,0}},
/* 52 */ {(10<<2)|2,{48,48,0}},
/* 53 */ {(10<<2)|2,{48,48,0}},
/* 54 */ {(10<<2)|2,{48,48,0}},
/* 55 */ {(10<<2)|2,{48,48,0}},
/* 56 */ {(10<<2)|2,{48,48,0}},
/* 57 */ {(10<<2)|2,{48,48,0}},
/* 58 */ {(10<<2)|2,{48,48,0}},
/* 59 */ {(10<<2)|2,{48,48,0}},
/* 60 */ {(10<<2)|2,{48,48,0}},
/* 61 */ {(10<<2)|2,{48,48,0}},
/* 62 */ {(10<<2)|2,{48,48,0}},
/* 63 */ {(10<<2)|2,{48,48,0}},
/* 64 */ {(15<<2)|3,{48,49,48}},
/* 65 */ {(15<<2)|3,{48,49,48}},
/* 66 */ {(15<<2)|3,{48,49,49}},
/* 67 */ {(15<<2)|3,{48,49,49}},
/* 68 */ {(15<<2)|3,{48,49,50}},
/* 69 */ {(15<<2)|3,{48,49,50}},
/* 70 */ {(15<<2)|3,{48,49,97}},
/* 71 */ {(15<<2)|3,{48,49,97}},
/* 72 */ {(15<<2)|3,{48,49,99}},
/* 73 */ {(15<<2)|3,{48,49,99}},
/* 74 */ {(15<<2)|3,{48,49,101}},
/* 75 */ {(15<<2)|3,{48,49,101}},
/* 76 */ {(15<<2)|3,{48,49,105}},
/* 77 */ {(15<<2)|3,{48,49,105}},
/* 78 */ {(15<<2)|3,{48,49,111}},
/* 79 */ {(15<<2)|3,{48,49,111}},
/* 80 */ {(15<<2)|3,{48,49,115}},
/* 81 */ {(15<<2)|3,{48,49,115}},
/* 82 */ {(15<<2)|3,{48,49,116}},
/* 83 */ {(15<<2)|3,{48,49,116}},
/* 84 */ {(16<<2)|3,{48,49,32}},
/* 85 */ {(16<<2)|3,{48,49,37}},
/* 86 */ {(16<<2)|3,{48,49,45}},
/* 87 */ {(16<<2)|3,{48,49,46}},
/* 88 */ {(16<<2)|3,{48,49,47}},
/* 89 */ {(16<<2)|3,{48,49,51}},
/* 90 */ {(16<<2)|3,{48,49,52}},
/* 91 */ {(16<<2)|3,{48,49,53}},
/* 92 */ {(16<<2)|3,{48,49,54}},
/* 93 */ {(16<<2)|3,{48,49,55}},
/* 94 */ {(16<<2)|3,{48,49,56}},
/* 95 */ {(16<<2)|3,{48,49,57}},
/* 96 */ {(16<<2)|3,{48,49,61}},
/* 97 */ {(16<<2)|3,{48,49,65}},
/* 98 */ {(16<<2)|3,{48,49,95}},
/* 99 */ {(16<<2)|3,{48,49,98}},
/* 100 */ {(16<<2)|3,{48,49,100}},
/* 101 */ {(16<<2)|3,{48,49,102}},
/* 102 */ {(16<<2)|3,{48,49,103}},
/* 103 */ {(16<<2)|3,{48,49,104}},
/* 104 */ {(16<<2)|3,{48,49,108}},
/* 105 */ {(16<<2)|3,{48,49,109}},
/* 106 */ {(16<<2)|3,{48,49,110}},
/* 107 */ {(16<<2)|3,{48,49,112}},
/* 108 */ {(16<<2)|3,{48,49,114}},
/* 109 */ {(16<<2)|3,{48,49,117}},
/* 110 */ {(10<<2)|2,{48,49,0}},
/* 111 */ {(10<<2)|2,{48,49,0}},
/* 112 */ {(10<<2)|2,{48,49,0}},
/* 113 */ {(10<<2)|2,{48,49,0}},
/* 114 */ {(10<<2)|2,{48,49,0}},
/* 115 */ {(10<<2)|2,{48,49,0}},
/* 116 */ {(10<<2)|2,{48,49,0}},
/* 117 */ {(10<<2)|2,{48,49,0}},
/* 118 */ {(10<<2)|2,{48,49,0}},
/* 119 */ {(10<<2)|2,{48,49,0}},
/* 120 */ {(10<<2)|2,{48,49,0}},
/* 121 */ {(10<<2)|2,{48,49,0}},
/* 122 */ {(10<<2)|2,{48,49,0}},
/* 123 */ {(10<<2)|2,{48,49,0}},
/* 124 */ {(10<<2)|2,{48,49,0}},
/* 125 */ {(10<<2)|2,{48,49,0}},
/* 126 */ {(10<<2)|2,{48,49,0}},
/* 127 */ {(10<<2)|2,{48,49,0}},
/* 128 */ {(15<<2)|3,{48,50,48}},
/* 129 */ {(15<<2)|3,{48,50,48}},
/* 130 */ {(15<<2)|3,{48,50,49}},
/* 131 */ {(15<<2)|3,{48,50,49}},
/* 132 */ {(15<<2)|3,{48,50,50}},
/* 133 */ {(15<<2)|3,{48,50,50}},
/* 134 */ {(15<<2)|3,{48,50,97}},
/* 135 */ {(15<<2)|3,{48,50,97}},
/* 136 */ {(15<<2)|3,{48,50,99}},
/* 137 */ {(15<<2)|3,{48,50,99}},
/* 138 */ {(15<<2)|3,{48,50,101}},
/* 139 */ {(15<<2)|3,{48,50,101}},
/* 140 */ {(15<<2)|3,{48,50,105}},
/* 141 */ {(15<<2)|3,{48,50,105}},
/* 142 */ {(15<<2)|3,{48,50,111}},
/* 143 */ {(15<<2)|3,{48,50,111}},
/* 144 */ {(15<<2)|3,{48,50,115}},
/* 145 */ {(15<<2)|3,{48,50,115}},
/* 146 */ {(15<<2)|3,{48,50,116}},
/* 147 */ {(15<<2)|3,{48,50,116}},
/* 148 */ {(16<<2)|3,{48,50,32}},
/* 149 */ {(16<<2)|3,{48,50,37}},
/* 150 */ {(16<<2)|3,{48,50,45}},
/* 151 */ {(16<<2)|3,{48,50,46}},
/* 152 */ {(16<<2)|3,{48,50,47}},
/* 153 */ {(16<<2)|3,{48,50,51}},
/* 154 */ {(16<<2)|3,{48,50,52}},
/* 155 */ {(16<<2)|3,{48,50,53}},
/* 156 */ {(16<<2)|3,{48,50,54}},
/* 157 */ {(16<<2)|3,{48,50,55}},
/* 158 */ {(16<<2)|3,{48,50,56}},
/* 159 */ {(16<<2)|3,{48,50,57}},
/* 160 */ {(16<<2)|3,{48,50,61}},
/* 161 */ {(16<<2)|3,{48,50,65}},
/* 162 */ {(16<<2)|3,{48,50,95}},
/* 163 */ {(16<<2)|3,{48,50,98}},
/* 164 */ {(16<<2)|3,{48,50,100}},
/* 165 */ {(16<<2)|3,{48,50,102}},
/* 166 */ {(16<<2)|3,{48,50,103}},
/* 167 */ {(16<<2)|3,{48,50,104}},
/* 168 */ {(16<<2)|3,{48,50,108}},
/* 169 */ {(16<<2)|3,{48,50,109}},
/* 170 */ {(16<<2)|3,{48,50,110}},
/* 171 */ {(16<<2)|3,{48,50,112}},
/* 172 */ {(16<<2)|3,{48,50,114}},
/* 173 */ {(16<<2)|3,{48,50,117}},
/* 174 */ {(10<<2)|2,{48,50,0}},
/* 175 */ {(10<<2)|2,{48,50,0}},
/* 176 */ {(10<<2)|2,{48,50,0}},
/* 177 */ {(10<<2)|2,{48,50,0}},
/* 178 */ {(10<<2)|2,{48,50,0}},
/* 179 */ {(10<<2)|2,{48,50,0}},
/* 180 */ {(10<<2)|2,{48,50,0}},
/* 181 */ {(10<<2)|2,{48,50,0}},
/* 182 */ {(10<<2)|2,{48,50,0}},
/* 183 */ {(10<<2)|2,{48,50,0}},
/* 184 */ {(10<<2)|2,{48,50,0}},
/* 185 */ {(10<<2)|2,{48,50,0}},
/* 186 */ {(10<<2)|2,{48,50,0}},
/* 187 */ {(10<<2)|2,{48,50,0}},
/* 188 */ {(10<<2)|2,{48,50,0}},
/* 189 */ {(10<<2)|2,{48,50,0}},
/* 190 */ {(10<<2)|2,{48,50,0}},
/* 191 */ {(10<<2)|2,{48,50,0}},
/* 192 */ {(15<<2)|3,{48,97,48}},
/* 193 */ {(15<<2)|3,{48,97,48}},
/* 194 */ {(15<<2)|3,{48,97,49}},
/* 195 */ {(15<<2)|3,{48,97,49}},
/* 196 */ {(15<<2)|3,{48,97,50}},
/* 197 */ {(15<<2)|3,{48,97,50}},
/* 198 */ {(15<<2)|3,{48,97,97}},
/* 199 */ {(15<<2)|3,{48,97,97}},
/* 200 */ {(15<<2)|3,{48,97,99}},
/* 201 */ {(15<<2)|3,{48,97,99}},
/* 202 */ {(15<<2)|3,{48,97,101}},
/* 203 */ {(15<<2)|3,{48,97,101}},
/* 204 */ {(15<<2)|3,{48,97,105}},
/* 205 */ {(15<<2)|3,{48,97,105}},
/* 206 */ {(15<<2)|3,{48,97,111}},
/* 207 */ {(15<<2)|3,{48,97,111}},
/* 208 */ {(15<<2)|3,{48,97,115}},
/* 209 */ {(15<<2)|3,{48,97,115}},
/* 210 */ {(15<<2)|3,{48,97,116}},
/* 211 */ {(15<<2)|3,{48,97,116}},
/* 212 */ {(16<<2)|3,{48,97,32}},
/* 213 */ {(16<<2)|3,{48,97,37}},
/* 214 */ {(16<<2)|3,{48,97,45}},
/* 215 */ {(16<<2)|3,{48,97,46}},
/* 216 */ {(16<<2)|3,{48,97,47}},
/* 217 */ {(16<<2)|3,{48,97,51}},
/* 218 */ {(16<<2)|3,{48,97,52}},
/* 219 */ {(16<<2)|3,{48,97,53}},
/* 220 */ {(16<<2)|3,{48,97,54}},
/* 221 */ {(16<<2)|3,{48,97,55}},
/* 222 */ {(16<<2)|3,{48,97,56}},
/* 223 */ {(16<<2)|3,{48,97,57}},
/* 224 */ {(16<<2)|3,{48,97,61}},
/* 225 */ {(16<<2)|3,{48,97,65}},
/* 226 */ {(16<<2)|3,{48,97,95}},
/* 227 */ {(16<<2)|3,{48,97,98}},
/* 228 */ {(16<<2)|3,{48,97,100}},
/* 229 */ {(16<<2)|3,{48,97,102}},
/* 230 */ {(16<<2)|3,{48,97,103}},
/* 231 */ {(16<<2)|3,{48,97,104}},
/* 232 */ {(16<<2)|3,{48,97,108}},
/* 233 */ {(16<<2)|3,{48,97,109}},
/* 234 */ {(16<<2)|3,{48,97,110}},
/* 235 */ {(16<<2)|3,{48,97,112}},
/* 236 */ {(16<<2)|3,{48,97,114}},
/* 237 */ {(16<<2)|3,{48,97,117}},
/* 238 */ {(10<<2)|2,{48,97,0}},
/* 239 */ {(10<<2)|2,{48,97,0}},
/* 240 */ {(10<<2)|2,{48,97,0}},
/* 241 */ {(10<<2)|2,{48,97,0}},
/* 242 */ {(10<<2)|2,{48,97,0}},
/* 243 */ {(10<<2)|2,{48,97,0}},
/* 244 */ {(10<<2)|2,{48,97,0}},
/* 245 */ {(10<<2)|2,{48,97,0}},
/* 246 */ {(10<<2)|2,{48,97,0}},
/* 247 */ {(10<<2)|2,{48,97,0}},
/* 248 */ {(10<<2)|2,{48,97,0}},
/* 249 */ {(10<<2)|2,{48,97,0}},
/* 250 */ {(10<<2)|2,{48,97,0}},
/* 251 */ {(10<<2)|2,{48,97,0}},
/* 252 */ {(10<<2)|2,{48,97,0}},
/* 253 */ {(10<<2)|2,{48,97,0}},
/* 254 */ {(10<<2)|2,{48,97,0}},
/* 255 */ {(10<<2)|2,{48,97,0}},
/* 256 */ {(15<<2)|3,{48,99,48}},
/* 257 */ {(15<<2)|3,{48,99,48}},
/* 258 */ {(15<<2)|3,{48,99,49}},
/* 259 */ {(15<<2)|3,{48,99,49}},
/* 260 */ {(15<<2)|3,{48,99,50}},
/* 261 */ {(15<<2)|3,{48,99,50}},
/* 262 */ {(15<<2)|3,{48,99,97}},
/* 263 */ {(15<<2)|3,{48,99,97}},
/* 264 */ {(15<<2)|3,{48,99,99}},
/* 265 */ {(15<<2)|3,{48,99,99}},
/* 266 */ {(15<<2)|3,{48,99,101}},
/* 267 */ {(15<<2)|3,{48,99,101}},
/* 268 */ {(15<<2)|3,{48,99,105}},
/* 269 */ {(15<<2)|3,{48,99,105}},
/* 270 */ {(15<<2)|3,{48,99,111}},
/* 271 */ {(15<<2)|3,{48,99,111}},
/* 272 */ {(15<<2)|3,{48,99,115}},
/* 273 */ {(15<<2)|3,{48,99,115}},
/* 274 */ {(15<<2)|3,{48,99,116}},
/* 275 */ {(15<<2)|3,{48,99,116}},
/* 276 */ {(16<<2)|3,{48,99,32}},
/* 277 */ {(16<<2)|3,{48,99,37}},
/* 278 */ {(16<<2)|3,{48,99,45}},
/* 279 */ {(16<<2)|3,{48,99,46}},
/* 280 */ {(16<<2)|3,{48,99,47}},
/* 281 */ {(16<<2)|3,{48,99,51}},
/* 282 */ {(16<<2)|3,{48,99,52}},
/* 283 */ {(16<<2)|3,{48,99,53}},
/* 284 */ {(16<<2)|3,{48,99,54}},
/* 285 */ {(16<<2)|3,{48,99,55}},
/* 286 */ {(16<<2)|3,{48,99,56}},
/* 287 */ {(16<<2)|3,{48,99,57}},
/* 288 */ {(16<<2)|3,{48,99,61}},
/* 289 */ {(16<<2)|3,{48,99,65}},
/* 290 */ {(16<<2)|3,{48,99,95}},
/* 291 */ {(16<<2)|3,{48,99,98}},
/* 292 */ {(16<<2)|3,{48,99,100}},
/* 293 */ {(16<<2)|3,{48,99,102}},
/* 294 */ {(16<<2)|3,{48,99,103}},
/* 295 */ {(16<<2)|3,{48,99,104}},
/* 296 */ {(16<<2)|3,{48,99,108}},
/* 297 */ {(16<<2)|3,{48,99,109}},
/* 298 */ {(16<<2)|3,{48,99,110}},
/* 299 */ {(16<<2)|3,{48,99,112}},
/* 300 */ {(16<<2)|3,{48,99,114}},
/* 301 */ {(16<<2)|3,{48,99,117}},
/* 302 */ {(10<<2)|2,{48,99,0}},
/* 303 */ {(10<<2)|2,{48,99,0}},
/* 304 */ {(10<<2)|2,{48,99,0}},
/* 305 */ {(10<<2)|2,{48,99,0}},
/* 306 */ {(10<<2)|2,{48,99,0}},
/* 307 */ {(10<<2)|2,{48,99,0}},
/* 308 */ {(10<<2)|2,{48,99,0}},
/* 309 */ {(10<<2)|2,{48,99,0}},
/* 310 */ {(10<<2)|2,{48,99,0}},
/* 311 */ {(10<<2)|2,{48,99,0}},
/* 312 */ {(10<<2)|2,{48,99,0}},
/* 313 */ {(10<<2)|2,{48,99,0}},
/* 314 */ {(10<<2)|2,{48,99,0}},
/* 315 */ {(10<<2)|2,{48,99,0}},
/* 316 */ {(10<<2)|2,{48,99,0}},
/* 317 */ {(10<<2)|2,{48,99,0}},
/* 318 */ {(10<<2)|2,{48,99,0}},
/* 319 */ {(10<<2)|2,{48,99,0}},
/* 320 */ {(15<<2)|3,{48,101,48}},
/* 321 */ {(15<<2)|3,{48,101,48}},
/* 322 */ {(15<<2)|3,{48,101,49}},
/* 323 */ {(15<<2)|3,{48,101,49}},
/* 324 */ {(15<<2)|3,{48,101,50}},
/* 325 */ {(15<<2)|3,{48,101,50}},
/* 326 */ {(15<<2)|3,{48,101,97}},
/* 327 */ {(15<<2)|3,{48,101,97}},
/* 328 */ {(15<<2)|3,{48,101,99}},
/* 329 */ {(15<<2)|3,{48,101,99}},
/* 330 */ {(15<<2)|3,{48,101,101}},
/* 331 */ {(15<<2)|3,{48,101,101}},
/* 332 */ {(15<<2)|3,{48,101,105}},
/* 333 */ {(15<<2)|3,{48,101,105}},
/* 334 */ {(15<<2)|3,{48,101,111}},
/* 335 */ {(15<<2)|3,{48,101,111}},
/* 336 */ {(15<<2)|3,{48,101,115}},
/* 337 */ {(15<<2)|3,{48,101,115}},
/* 338 */ {(15<<2)|3,{48,101,116}},
/* 339 */ {(15<<2)|3,{48,101,116}},
/* 340 */ {(16<<2)|3,{48,101,32}},
/* 341 */ {(16<<2)|3,{48,101,37}},
/* 342 */ {(16<<2)|3,{48,101,45}},
/* 343 */ {(16<<2)|3,{48,101,46}},
/* 344 */ {(16<<2)|3,{48,101,47}},
/* 345 */ {(16<<2)|3,{48,101,51}},
/* 346 */ {(16<<2)|3,{48,101,52}},
/* 347 */ {(16<<2)|3,{48,101,53}},
/* 348 */ {(16<<2)|3,{48,101,54}},
/* 349 */ {(16<<2)|3,{48,101,55}},
/* 350 */ {(16<<2)|3,{48,101,56}},
/* 351 */ {(16<<2)|3,{48,101,57}},
/* 352 */ {(16<<2)|3,{48,101,61}},
/* 353 */ {(16<<2)|3,{48,101,65}},
/* 354 */ {(16<<2)|3,{48,101,95}},
/* 355 */ {(16<<2)|3,{48,101,98}},
/* 356 */ {(16<<2)|3,{48,101,100}},
/* 357 */ {(16<<2)|3,{48,101,102}},
/* 358 */ {(16<<2)|3,{48,101,103}},
/* 359 */ {(16<<2)|3,{48,101,104}},
/* 360 */ {(16<<2)|3,{48,101,108}},
/* 361 */ {(16<<2)|3,{48,101,109}},
/* 362 */ {(16<<2)|3,{48,101,110}},
/* 363 */ {(16<<2)|3,{48,101,112}},
/* 364 */ {(16<<2)|3,{48,101,114}},
/* 365 */ {(16<<2)|3,{48,101,117}},
/* 366 */ {(10<<2)|2,{48,101,0}},
/* 367 */ {(10<<2)|2,{48,101,0}},
/* 368 */ {(10<<2)|2,{48,101,0}},
/* 369 */ {(10<<2)|2,{48,101,0}},
/* 370 */ {(10<<2)|2,{48,101,0}},
/* 371 */ {(10<<2)|2,{48,101,0}},
/* 372 */ {(10<<2)|2,{48,101,0}},
/* 373 */ {(10<<2)|2,{48,101,0}},
/* 374 */ {(10<<2)|2,{48,101,0}},
/* 375 */ {(10<<2)|2,{48,101,0}},
/* 376 */ {(10<<2)|2,{48,101,0}},
/* 377 */ {(10<<2)|2,{48,101,0}},
/* 378 */ {(10<<2)|2,{48,101,0}},
/* 379 */ {(10<<2)|2,{48,101,0}},
/* 380 */ {(10<<2)|2,{48,101,0}},
/* 381 */ {(10<<2)|2,{48,101,0}},
/* 382 */ {(10<<2)|2,{48,101,0}},
/* 383 */ {(10<<2)|2,{48,101,0}},
/* 384 */ {(15<<2)|3,{48,105,48}},
/* 385 */ {(15<<2)|3,{48,105,48}},
/* 386 */ {(15<<2)|3,{48,105,49}},
/* 387 */ {(15<<2)|3,{48,105,49}},
/* 388 */ {(15<<2)|3,{48,105,50}},
/* 389 */ {(15<<2)|3,{48,105,50}},
/* 390 */ {(15<<2)|3,{48,105,97}},
/* 391 */ {(15<<2)|3,{48,105,97}},
/* 392 */ {(15<<2)|3,{48,105,99}},
/* 393 */ {(15<<2)|3,{48,105,99}},
/* 394 */ {(15<<2)|3,{48,105,101}},
/* 395 */ {(15<<2)|3,{48,105,101}},
/* 396 */ {(15<<2)|3,{48,105,105}},
/* 397 */ {(15<<2)|3,{48,105,105}},
/* 398 */ {(15<<2)|3,{48,105,111}},
/* 399 */ {(15<<2)|3,{48,105,111}},
/* 400 */ {(15<<2)|3,{48,105,115}},
/* 401 */ {(15<<2)|3,{48,105,115}},
/* 402 */ {(15<<2)|3,{48,105,116}},
/* 403 */ {(15<<2)|3,{48,105,116}},
/* 404 */ {(16<<2)|3,{48,105,32}},
/* 405 */ {(16<<2)|3,{48,105,37}},
/* 406 */ {(16<<2)|3,{48,105,45}},
/* 407 */ {(16<<2)|3,{48,105,46}},
/* 408 */ {(16<<2)|3,{48,105,47}},
/* 409 */ {(16<<2)|3,{48,105,51}},
/* 410 */ {(16<<2)|3,{48,105,52}},
/* 411 */ {(16<<2)|3,{48,105,53}},
/* 412 */ {(16<<2)|3,{48,105,54}},
/* 413 */ {(16<<2)|3,{48,105,55}},
/* 414 */ {(16<<2)|3,{48,105,56}},
/* 415 */ {(16<<2)|3,{48,105,57}},
/* 416 */ {(16<<2)|3,{48,105,61}},
/* 417 */ {(16<<2)|3,{48,105,65}},
/* 418 */ {(16<<2)|3,{48,105,95}},
/* 419 */ {(16<<2)|3,{48,105,98}},
/* 420 */ {(16<<2)|3,{48,105,100}},
/* 421 */ {(16<<2)|3,{48,105,102}},
/* 422 */ {(16<<2)|3,{48,105,103}},
/* 423 */ {(16<<2)|3,{48,105,104}},
/* 424 */ {(16<<2)|3,{48,105,108}},
/* 425 */ {(16<<2)|3,{48,105,109}},
/* 426 */ {(16<<2)|3,{48,105,110}},
/* 427 */ {(16<<2)|3,{48,105,112}},
/* 428 */ {(16<<2)|3,{48,105,114}},
/* 429 */ {(16<<2)|3,{48,105,117}},
/* 430 */ {(10<<2)|2,{48,105,0}},
/* 431 */ {(10<<2)|2,{48,105,0}},
/* 432 */ {(10<<2)|2,{48,105,0}},
/* 433 */ {(10<<2)|2,{48,105,0}},
/* 434 */ {(10<<2)|2,{48,105,0}},
/* 435 */ {(10<<2)|2,{48,105,0}},
/* 436 */ {(10<<2)|2,{48,105,0}},
/* 437 */ {(10<<2)|2,{48,105,0}},
/* 438 */ {(10<<2)|2,{48,105,0}},
/* 439 */ {(10<<2)|2,{48,105,0}},
/* 440 */ {(10<<2)|2,{48,105,0}},
/* 441 */ {(10<<2)|2,{48,105,0}},
/* 442 */ {(10<<2)|2,{48,105,0}},
/* 443 */ {(10<<2)|2,{48,105,0}},
/* 444 */ {(10<<2)|2,{48,105,0}},
/* 445 */ {(10<<2)|2,{48,105,0}},
/* 446 */ {(10<<2)|2,{48,105,0}},
/* 447 */ {(10<<2)|2,{48,105,0}},
/* 448 */ {(15<<2)|3,{48,111,48}},
/* 449 */ {(15<<2)|3,{48,111,48}},
/* 450 */ {(15<<2)|3,{48,111,49}},
/* 451 */ {(15<<2)|3,{48,111,49}},
/* 452 */ {(15<<2)|3,{48,111,50}},
/* 453 */ {(15<<2)|3,{48,111,50}},
/* 454 */ {(15<<2)|3,{48,111,97}},
/* 455 */ {(15<<2)|3,{48,111,97}},
/* 456 */ {(15<<2)|3,{48,111,99}},
/* 457 */ {(15<<2)|3,{48,111,99}},
/* 458 */ {(15<<2)|3,{48,111,101}},
/* 459 */ {(15<<2)|3,{48,111,101}},
/* 460 */ {(15<<2)|3,{48,111,105}},
/* 461 */ {(15<<2)|3,{48,111,105}},
/* 462 */ {(15<<2)|3,{48,111,111}},
/* 463 */ {(15<<2)|3,{48,111,111}},
/* 464 */ {(15<<2)|3,{48,111,115}},
/* 465 */ {(15<<2)|3,{48,111,115}},
/* 466 */ {(15<<2)|3,{48,111,116}},
/* 467 */ {(15<<2)|3,{48,111,116}},
/* 468 */ {(16<<2)|3,{48,111,32}},
/* 469 */ {(16<<2)|3,{48,111,37}},
/* 470 */ {(16<<2)|3,{48,111,45}},
/* 471 */ {(16<<2)|3,{48,111,46}},
/* 472 */ {(16<<2)|3,{48,111,47}},
/* 473 */ {(16<<2)|3,{48,111,51}},
/* 474 */ {(16<<2)|3,{48,111,52}},
/* 475 */ {(16<<2)|3,{48,111,53}},
/* 476 */ {(16<<2)|3,{48,111,54}},
/* 477 */ {(16<<2)|3,{48,111,55}},
/* 478 */ {(16<<2)|3,{48,111,56}},
/* 479 */ {(16<<2)|3,{48,111,57}},
/* 480 */ {(16<<2)|3,{48,111,61}},
/* 481 */ {(16<<2)|3,{48,111,65}},
/* 482 */ {(16<<2)|3,{48,111,95}},
/* 483 */ {(16<<2)|3,{48,111,98}},
/* 484 */ {(16<<2)|3,{48,111,100}},
/* 485 */ {(16<<2)|3,{48,111,102}},
/* 486 */ {(16<<2)|3,{48,111,103}},
/* 487 */ {(16<<2)|3,{48,111,104}},
/* 488 */ {(16<<2)|3,{48,111,108}},
/* 489 */ {(16<<2)|3,{48,111,109}},
/* 490 */ {(16<<2)|3,{48,111,110}},
/* 491 */ {(16<<2)|3,{48,111,112}},
/* 492 */ {(16<<2)|3,{48,111,114}},
/* 493 */ {(16<<2)|3,{48,111,117}},
/* 494 */ {(10<<2)|2,{48,111,0}},
/* 495 */ {(10<<2)|2,{48,111,0}},
/* 496 */ {(10<<2)|2,{48,111,0}},
/* 497 */ {(10<<2)|2,{48,111,0}},
/* 498 */ {(10<<2)|2,{48,111,0}},
/* 499 */ {(10<<2)|2,{48,111,0}},
/* 500 */ {(10<<2)|2,{48,111,0}},
/* 501 */ {(10<<2)|2,{48,111,0}},
/* 502 */ {(10<<2)|2,{48,111,0}},
/* 503 */ {(10<<2)|2,{48,111,0}},
/* 504 */ {(10<<2)|2,{48,111,0}},
/* 505 */ {(10<<2)|2,{48,111,0}},
/* 506 */ {(10<<2)|2,{48,111,0}},
/* 507 */ {(10<<2)|2,{48,111,0}},
/* 508 */ {(10<<2)|2,{48,111,0}},
/* 509 */ {(10<<2)|2,{48,111,0}},
/* 510 */ {(10<<2)|2,{48,111,0}},
/* 511 */ {(10<<2)|2,{48,111,0}},
/* 512 */ {(15<<2)|3,{48,115,48}},
/* 513 */ {(15<<2)|3,{48,115,48}},
/* 514 */ {(15<<2)|3,{48,115,49}},
/* 515 */ {(15<<2)|3,{48,115,49}},
/* 516 */ {(15<<2)|3,{48,115,50}},
/* 517 */ {(15<<2)|3,{48,115,50}},
/* 518 */ {(15<<2)|3,{48,115,97}},
/* 519 */ {(15<<2)|3,{48,115,97}},
/* 520 */ {(15<<2)|3,{48,115,99}},
/* 521 */ {(15<<2)|3,{48,115,99}},
/* 522 */ {(15<<2)|3,{48,115,101}},
/* 523 */ {(15<<2)|3,{48,115,101}},
/* 524 */ {(15<<2)|3,{48,115,105}},
/* 525 */ {(15<<2)|3,{48,115,105}},
/* 526 */ {(15<<2)|3,{48,115,111}},
/* 527 */ {(15<<2)|3,{48,115,111}},
/* 528 */ {(15<<2)|3,{48,115,115}},
/* 529 */ {(15<<2)|3,{48,115,115}},
/* 530 */ {(15<<2)|3,{48,115,116}},
/* 531 */ {(15<<2)|3,{48,115,116}},
/* 532 */ {(16<<2)|3,{48,115,32}},
/* 533 */ {(16<<2)|3,{48,115,37}},
/* 534 */ {(16<<2)|3,{48,115,45}},
/* 535 */ {(16<<2)|3,{48,115,46}},
/* 536 */ {(16<<2)|3,{48,115,47}},
/* 537 */ {(16<<2)|3,{48,115,51}},
/* 538 */ {(16<<2)|3,{48,115,52}},
/* 539 */ {(16<<2)|3,{48,115,53}},
/* 540 */ {(16<<2)|3,{48,115,54}},
/* 541 */ {(16<<2)|3,{48,115,55}},
/* 542 */ {(16<<2)|3,{48,115,56}},
/* 543 */ {(16<<2)|3,{48,115,57}},
/* 544 */ {(16<<2)|3,{48,115,61}},
/* 545 */ {(16<<2)|3,{48,115,65}},
/* 546 */ {(16<<2)|3,{48,115,95}},
/* 547 */ {(16<<2)|3,{48,115,98}},
/* 548 */ {(16<<2)|3,{48,115,100}},
/* 549 */ {(16<<2)|3,{48,115,102}},
/* 550 */ {(16<<2)|3,{48,115,103}},
/* 551 */ {(16<<2)|3,{48,115,104}},
/* 552 */ {(16<<2)|3,{48,115,108}},
/* 553 */ {(16<<2)|3,{48,115,109}},
/* 554 */ {(16<<2)|3,{48,115,110}},
/* 555 */ {(16<<2)|3,{48,115,112}},
/* 556 */ {(16<<2)|3,{48,115,114}},
/* 557 */ {(16<<2)|3,{48,115,117}},
/* 558 */ {(10<<2)|2,{48,115,0}},
/* 559 */ {(10<<2)|2,{48,115,0}},
/* 560 */ {(10<<2)|2,{48,115,0}},
/* 561 */ {(10<<2)|2,{48,115,0}},
/* 562 */ {(10<<2)|2,{48,115,0}},
/* 563 */ {(10<<2)|2,{48,115,0}},
/* 564 */ {(10<<2)|2,{48,115,0}},
/* 565 */ {(10<<2)|2,{48,115,0}},
/* 566 */ {(10<<2)|2,{48,115,0}},
/* 567 */ {(10<<2)|2,{48,115,0}},
/* 568 */ {(10<<2)|2,{48,115,0}},
/* 569 */ {(10<<2)|2,{48,115,0}},
/* 570 */ {(10<<2)|2,{48,115,0}},
/* 571 */ {(10<<2)|2,{48,115,0}},
/* 572 */ {(10<<2)|2,{48,115,0}},
/* 573 */ {(10<<2)|2,{48,115,0}},
/* 574 */ {(10<<2)|2,{48,115,0}},
/* 575 */ {(10<<2)|2,{48,115,0}},
/* 576 */ {(15<<2)|3,{48,116,48}},
/* 577 */ {(15<<2)|3,{48,116,48}},
/* 578 */ {(15<<2)|3,{48,116,49}},
/* 579 */ {(15<<2)|3,{48,116,49}},
/* 580 */ {(15<<2)|3,{48,116,50}},
/* 581 */ {(15<<2)|3,{48,116,50}},
/* 582 */ {(15<<2)|3,{48,116,97}},
/* 583 */ {(15<<2)|3,{48,116,97}},
/* 584 */ {(15<<2)|3,{48,116,99}},
/* 585 */ {(15<<2)|3,{48,116,99}},
/* 586 */ {(15<<2)|3,{48,116,101}},
/* 587 */ {(15<<2)|3,{48,116,101}},
/* 588 */ {(15<<2)|3,{48,116,105}},
/* 589 */ {(15<<2)|3,{48,116,105}},
/* 590 */ {(15<<2)|3,{48,116,111}},
/* 591 */ {(15<<2)|3,{48,116,111}},
/* 592 */ {(15<<2)|3,{48,116,115}},
/* 593 */ {(15<<2)|3,{48,116,115}},
/* 594 */ {(15<<2)|3,{48,116,116}},
/* 595 */ {(15<<2)|3,{48,116,116}},
/* 596 */ {(16<<2)|3,{48,116,32}},
/* 597 */ {(16<<2)|3,{48,116,37}},
/* 598 */ {(16<<2)|3,{48,116,45}},
/* 599 */ {(16<<2)|3,{48,116,46}},
/* 600 */ {(16<<2)|3,{48,116,47}},
/* 601 */ {(16<<2)|3,{48,116,51}},
/* 602 */ {(16<<2)|3,{48,116,52}},
/* 603 */ {(16<<2)|3,{48,116,53}},
/* 604 */ {(16<<2)|3,{48,116,54}},
/* 605 */ {(16<<2)|3,{48,116,55}},
/* 606 */ {(16<<2)|3,{48,116,56}},
/* 607 */ {(16<<2)|3,{48,116,57}},
/* 608 */ {(16<<2)|3,{48,116,61}},
/* 609 */ {(16<<2)|3,{48,116,65}},
/* 610 */ {(16<<2)|3,{48,116,95}},
/* 611 */ {(16<<2)|3,{48,116,98}},
/* 612 */ {(16<<2)|3,{48,116,100}},
/* 613 */ {(16<<2)|3,{48,116,102}},
/* 614 */ {(16<<2)|3,{48,116,103}},
/* 615 */ {(16<<2)|3,{48,116,104}},
/* 616 */ {(16<<2)|3,{48,116,108}},
/* 617 */ {(16<<2)|3,{48,116,109}},
/* 618 */ {(16<<2)|3,{48,116,110}},
/* 619 */ {(16<<2)|3,{48,116,112}},
/* 620 */ {(16<<2)|3,{48,116,114}},
/* 621 */ {(16<<2)|3,{48,116,117}},
/* 622 */ {(10<<2)|2,{48,116,0}},
/* 623 */ {(10<<2)|2,{48,116,0}},
/* 624 */ {(10<<2)|2,{48,116,0}},
/* 625 */ {(10<<2)|2,{48,116,0}},
/* 626 */ {(10<<2)|2,{48,116,0}},
/* 627 */ {(10<<2)|2,{48,116,0}},
/* 628 */ {(10<<2)|2,{48,116,0}},
/* 629 */ {(10<<2)|2,{48,116,0}},
/* 630 */ {(10<<2)|2,{48,116,0}},
/* 631 */ {(10<<2)|2,{48,116,0}},
/* 632 */ {(10<<2)|2,{48,116,0}},
/* 633 */ {(10<<2)|2,{48,116,0}},
/* 634 */ {(10<<2)|2,{48,116,0}},
/* 635 */ {(10<<2)|2,{48,116,0}},
/* 636 */ {(10<<2)|2,{48,116,0}},
/* 637 */ {(10<<2)|2,{48,116,0}},
/* 638 */ {(10<<2)|2,{48,116,0}},
/* 639 */ {(10<<2)|2,{48,116,0}},
/* 640 */ {(16<<2)|3,{48,32,48}},
/* 641 */ {(16<<2)|3,{48,32,49}},
/* 642 */ {(16<<2)|3,{48,32,50}},
/* 643 */ {(16<<2)|3,{48,32,97}},
/* 644 */ {(16<<2)|3,{48,32,99}},
/* 645 */ {(16<<2)|3,{48,32,101}},
/* 646 */ {(16<<2)|3,{48,32,105}},
/* 647 */ {(16<<2)|3,{48,32,111}},
/* 648 */ {(16<<2)|3,{48,32,115}},
/* 649 */ {(16<<2)|3,{48,32,116}},
/* 650 */ {(11<<2)|2,{48,32,0}},
/* 651 */ {(11<<2)|2,{48,32,0}},
/* 652 */ {(11<<2)|2,{48,32,0}},
/* 653 */ {(11<<2)|2,{48,32,0}},
/* 654 */ {(11<<2)|2,{48,32,0}},
/* 655 */ {(11<<2)|2,{48,32,0}},
/* 656 */ {(11<<2)|2,{48,32,0}},
/* 657 */ {(11<<2)|2,{48,32,0}},
/* 658 */ {(11<<2)|2,{48,32,0}},
/* 659 */ {(11<<2)|2,{48,32,0}},
/* 660 */ {(11<<2)|2,{48,32,0}},
/* 661 */ {(11<<2)|2,{48,32,0}},
/* 662 */ {(11<<2)|2,{48,32,0}},
/* 663 */ {(11<<2)|2,{48,32,0}},
/* 664 */ {(11<<2)|2,{48,32,0}},
/* 665 */ {(11<<2)|2,{48,32,0}},
/* 666 */ {(11<<2)|2,{48,32,0}},
/* 667 */ {(11<<2)|2,{48,32,0}},
/* 668 */ {(11<<2)|2,{48,32,0}},
/* 669 */ {(11<<2)|2,{48,32,0}},
/* 670 */ {(11<<2)|2,{48,32,0}},
/* 671 */ {(11<<2)|2,{48,32,0}},
/* 672 */ {(16<<2)|3,{48,37,48}},
/* 673 */ {(16<<2)|3,{48,37,49}},
/* 674 */ {(16<<2)|3,{48,37,50}},
/* 675 */ {(16<<2)|3,{48,37,97}},
/* 676 */ {(16<<2)|3,{48,37,99}},
/* 677 */ {(16<<2)|3,{48,37,101}},
/* 678 */ {(16<<2)|3,{48,37,105}},
/* 679 */ {(16<<2)|3,{48,37,111}},
/* 680 */ {(16<<2)|3,{48,37,115}},
/* 681 */ {(16<<2)|3,{48,37,116}},
/* 682 */ {(11<<2)|2,{48,37,0}},
/* 683 */ {(11<<2)|2,{48,37,0}},
/* 684 */ {(11<<2)|2,{48,37,0}},
/* 685 */ {(11<<2)|2,{48,37,0}},
/* 686 */ {(11<<2)|2,{48,37,0}},
/* 687 */ {(11<<2)|2,{48,37,0}},
/* 688 */ {(11<<2)|2,{48,37,0}},
/* 689 */ {(11<<2)|2,{48,37,0}},
/* 690 */ {(11<<2)|2,{48,37,0}},
/* 691 */ {(11<<2)|2,{48,37,0}},
/* 692 */ {(11<<2)|2,{48,37,0}},
/* 693 */ {(11<<2)|2,{48,37,0}},
/* 694 */ {(11<<2)|2,{48,37,0}},
/* 695 */ {(11<<2)|2,{48,37,0}},
/* 696 */ {(11<<2)|2,{48,37,0}},
/* 697 */ {(11<<2)|2,{48,37,0}},
/* 698 */ {(11<<2)|2,{48,37,0}},
/* 699 */ {(11<<2)|2,{48,37,0}},
/* 700 */ {(11<<2)|2,{48,37,0}},
/* 701 */ {(11<<2)|2,{48,37,0}},
/* 702 */ {(11<<2)|2,{48,37,0}},
/* 703 */ {(11<<2)|2,{48,37,0}},
/* 704 */ {(16<<2)|3,{48,45,48}},
/* 705 */ {(16<<2)|3,{48,45,49}},
/* 706 */ {(16<<2)|3,{48,45,50}},
/* 707 */ {(16<<2)|3,{48,45,97}},
/* 708 */ {(16<<2)|3,{48,45,99}},
/* 709 */ {(16<<2)|3,{48,45,101}},
/* 710 */ {(16<<2)|3,{48,45,105}},
/* 711 */ {(16<<2)|3,{48,45,111}},
/* 712 */ {(16<<2)|3,{48,45,115}},
/* 713 */ {(16<<2)|3,{48,45,116}},
/* 714 */ {(11<<2)|2,{48,45,0}},
/* 715 */ {(11<<2)|2,{48,45,0}},
/* 716 */ {(11<<2)|2,{48,45,0}},
/* 717 */ {(11<<2)|2,{48,45,0}},
/* 718 */ {(11<<2)|2,{48,45,0}},
/* 719 */ {(11<<2)|2,{48,45,0}},
/* 720 */ {(11<<2)|2,{48,45,0}},
/* 721 */ {(11<<2)|2,{48,45,0}},
/* 722 */ {(11<<2)|2,{48,45,0}},
/* 723 */ {(11<<2)|2,{48,45,0}},
/* 724 */ {(11<<2)|2,{48,45,0}},
/* 725 */ {(11<<2)|2,{48,45,0}},
/* 726 */ {(11<<2)|2,{48,45,0}},
/* 727 */ {(11<<2)|2,{48,45,0}},
/* 728 */ {(11<<2)|2,{48,45,0}},
/* 729 */ {(11<<2)|2,{48,45,0}},
/* 730 */ {(11<<2)|2,{48,45,0}},
/* 731 */ {(11<<2)|2,{48,45,0}},
/* 732 */ {(11<<2)|2,{48,45,0}},
/* 733 */ {(11<<2)|2,{48,45,0}},
/* 734 */ {(11<<2)|2,{48,45,0}},
/* 735 */ {(11<<2)|2,{48,45,0}},
/* 736 */ {(16<<2)|3,{48,46,48}},
/* 737 */ {(16<<2)|3,{48,46,49}},
/* 738 */ {(16<<2)|3,{48,46,50}},
/* 739 */ {(16<<2)|3,{48,46,97}},
/* 740 */ {(16<<2)|3,{48,46,99}},
/* 741 */ {(16<<2)|3,{48,46,101}},
/* 742 */ {(16<<2)|3,{48,46,105}},
/* 743 */ {(16<<2)|3,{48,46,111}},
/* 744 */ {(16<<2)|3,{48,46,115}},
/* 745 */ {(16<<2)|3,{48,46,116}},
/* 746 */ {(11<<2)|2,{48,46,0}},
/* 747 */ {(11<<2)|2,{48,46,0}},
/* 748 */ {(11<<2)|2,{48,46,0}},
/* 749 */ {(11<<2)|2,{48,46,0}},
/* 750 */ {(11<<2)|2,{48,46,0}},
/* 751 */ {(11<<2)|2,{48,46,0}},
/* 752 */ {(11<<2)|2,{48,46,0}},
/* 753 */ {(11<<2)|2,{48,46,0}},
/* 754 */ {(11<<2)|2,{48,46,0}},
/* 755 */ {(11<<2)|2,{48,46,0}},
/* 756 */ {(11<<2)|2,{48,46,0}},
/* 757 */ {(11<<2)|2,{48,46,0}},
/* 758 */ {(11<<2)|2,{48,46,0}},
/* 759 */ {(11<<2)|2,{48,46,0}},
/* 760 */ {(11<<2)|2,{48,46,0}},
/* 761 */ {(11<<2)|2,{48,46,0}},
/* 762 */ {(11<<2)|2,{48,46,0}},
/* 763 */ {(11<<2)|2,{48,46,0}},
/* 764 */ {(11<<2)|2,{48,46,0}},
/* 765 */ {(11<<2)|2,{48,46,0}},
/* 766 */ {(11<<2)|2,{48,46,0}},
/* 767 */ {(11<<2)|2,{48,46,0}},
/* 768 */ {(16<<2)|3,{48,47,48}},
/* 769 */ {(16<<2)|3,{48,47,49}},
/* 770 */ {(16<<2)|3,{48,47,50}},
/* 771 */ {(16<<2)|3,{48,47,97}},
/* 772 */ {(16<<2)|3,{48,47,99}},
/* 773 */ {(16<<2)|3,{48,47,101}},
/* 774 */ {(16<<2)|3,{48,47,105}},
/* 775 */ {(16<<2)|3,{48,47,111}},
/* 776 */ {(16<<2)|3,{48,47,115}},
/* 777 */ {(16<<2)|3,{48,47,116}},
/* 778 */ {(11<<2)|2,{48,47,0}},
/* 779 */ {(11<<2)|2,{48,47,0}},
/* 780 */ {(11<<2)|2,{48,47,0}},
/* 781 */ {(11<<2)|2,{48,47,0}},
/* 782 */ {(11<<2)|2,{48,47,0}},
/* 783 */ {(11<<2)|2,{48,47,0}},
/* 784 */ {(11<<2)|2,{48,47,0}},
/* 785 */ {(11<<2)|2,{48,47,0}},
/* 786 */ {(11<<2)|2,{48,47,0}},
/* 787 */ {(11<<2)|2,{48,47,0}},
/* 788 */ {(11<<2)|2,{48,47,0}},
/* 789 */ {(11<<2)|2,{48,47,0}},
/* 790 */ {(11<<2)|2,{48,47,0}},
/* 791 */ {(11<<2)|2,{48,47,0}},
/* 792 */ {(11<<2)|2,{48,47,0}},
/* 793 */ {(11<<2)|2,{48,47,0}},
/* 794 */ {(11<<2)|2,{48,47,0}},
/* 795 */ {(11<<2)|2,{48,47,0}},
/* 796 */ {(11<<2)|2,{48,47,0}},
/* 797 */ {(11<<2)|2,{48,47,0}},
/* 798 */ {(11<<2)|2,{48,47,0}},
/* 799 */ {(11<<2)|2,{48,47,0}},
/* 800 */ {(16<<2)|3,{48,51,48}},
/* 801 */ {(16<<2)|3,{48,51,49}},
/* 802 */ {(16<<2)|3,{48,51,50}},
/* 803 */ {(16<<2)|3,{48,51,97}},
/* 804 */ {(16<<2)|3,{48,51,99}},
/* 805 */ {(16<<2)|3,{48,51,101}},
/* 806 */ {(16<<2)|3,{48,51,105}},
/* 807 */ {(16<<2)|3,{48,51,111}},
/* 808 */ {(16<<2)|3,{48,51,115}},
/* 809 */ {(16<<2)|3,{48,51,116}},
/* 810 */ {(11<<2)|2,{48,51,0}},
/* 811 */ {(11<<2)|2,{48,51,0}},
/* 812 */ {(11<<2)|2,{48,51,0}},
/* 813 */ {(11<<2)|2,{48,51,0}},
/* 814 */ {(11<<2)|2,{48,51,0}},
/* 815 */ {(11<<2)|2,{48,51,0}},
/* 816 */ {(11<<2)|2,{48,51,0}},
/* 817 */ {(11<<2)|2,{48,51,0}},
/* 818 */ {(11<<2)|2,{48,51,0}},
/* 819 */ {(11<<2)|2,{48,51,0}},
/* 820 */ {(11<<2)|2,{48,51,0}},
/* 821 */ {(11<<2)|2,{48,51,0}},
/* 822 */ {(11<<2)|2,{48,51,0}},
/* 823 */ {(11<<2)|2,{48,51,0}},
/* 824 */ {(11<<2)|2,{48,51,0}},
/* 825 */ {(11<<2)|2,{48,51,0}},
/* 826 */ {(11<<2)|2,{48,51,0}},
/* 827 */ {(11<<2)|2,{48,51,0}},
/* 828 */ {(11<<2)|2,{48,51,0}},
/* 829 */ {(11<<2)|2,{48,51,0}},
/* 830 */ {(11<<2)|2,{48,51,0}},
/* 831 */ {(11<<2)|2,{48,51,0}},
/* 832 */ {(16<<2)|3,{48,52,48}},
/* 833 */ {(16<<2)|3,{48,52,49}},
/* 834 */ {(16<<2)|3,{48,52,50}},
/* 835 */ {(16<<2)|3,{48,52,97}},
/* 836 */ {(16<<2)|3,{48,52,99}},
/* 837 */ {(16<<2)|3,{48,52,101}},
/* 838 */ {(16<<2)|3,{48,52,105}},
/* 839 */ {(16<<2)|3,{48,52,111}},
/* 840 */ {(16<<2)|3,{48,52,115}},
/* 841 */ {(16<<2)|3,{48,52,116}},
/* 842 */ {(11<<2)|2,{48,52,0}},
/* 843 */ {(11<<2)|2,{48,52,0}},
/* 844 */ {(11<<2)|2,{48,52,0}},
/* 845 */ {(11<<2)|2,{48,52,0}},
/* 846 */ {(11<<2)|2,{48,52,0}},
/* 847 */ {(11<<2)|2,{48,52,0}},
/* 848 */ {(11<<2)|2,{48,52,0}},
/* 849 */ {(11<<2)|2,{48,52,0}},
/* 850 */ {(11<<2)|2,{48,52,0}},
/* 851 */ {(11<<2)|2,{48,52,0}},
/* 852 */ {(11<<2)|2,{48,52,0}},
/* 853 */ {(11<<2)|2,{48,52,0}},
/* 854 */ {(11<<2)|2,{48,52,0}},
/* 855 */ {(11<<2)|2,{48,52,0}},
/* 856 */ {(11<<2)|2,{48,52,0}},
/* 857 */ {(11<<2)|2,{48,52,0}},
/* 858 */ {(11<<2)|2,{48,52,0}},
/* 859 */ {(11<<2)|2,{48,52,0}},
/* 860 */ {(11<<2)|2,{48,52,0}},
/* 861 */ {(11<<2)|2,{48,52,0}},
/* 862 */ {(11<<2)|2,{48,52,0}},
/* 863 */ {(11<<2)|2,{48,52,0}},
/* 864 */ {(16<<2)|3,{48,53,48}},
/* 865 */ {(16<<2)|3,{48,53,49}},
/* 866 */ {(16<<2)|3,{48,53,50}},
/* 867 */ {(16<<2)|3,{48,53,97}},
/* 868 */ {(16<<2)|3,{48,53,99}},
/* 869 */ {(16<<2)|3,{48,53,101}},
/* 870 */ {(16<<2)|3,{48,53,105}},
/* 871 */ {(16<<2)|3,{48,53,111}},
/* 872 */ {(16<<2)|3,{48,53,115}},
/* 873 */ {(16<<2)|3,{48,53,116}},
/* 874 */ {(11<<2)|2,{48,53,0}},
/* 875 */ {(11<<2)|2,{48,53,0}},
/* 876 */ {(11<<2)|2,{48,53,0}},
/* 877 */ {(11<<2)|2,{48,53,0}},
/* 878 */ {(11<<2)|2,{48,53,0}},
/* 879 */ {(11<<2)|2,{48,53,0}},
/* 880 */ {(11<<2)|2,{48,53,0}},
/* 881 */ {(11<<2)|2,{48,53,0}},
/* 882 */ {(11<<2)|2,{48,53,0}},
/* 883 */ {(11<<2)|2,{48,53,0}},
/* 884 */ {(11<<2)|2,{48,53,0}},
/* 885 */ {(11<<2)|2,{48,53,0}},
/* 886 */ {(11<<2)|2,{48,53,0}},
/* 887 */ {(11<<2)|2,{48,53,0}},
/* 888 */ {(11<<2)|2,{48,53,0}},
/* 889 */ {(11<<2)|2,{48,53,0}},
/* 890 */ {(11<<2)|2,{48,53,0}},
/* 891 */ {(11<<2)|2,{48,53,0}},
/* 892 */ {(11<<2)|2,{48,53,0}},
/* 893 */ {(11<<2)|2,{48,53,0}},
/* 894 */ {(11<<2)|2,{48,53,0}},
/* 895 */ {(11<<2)|2,{48,53,0}},
/* 896 */ {(16<<2)|3,{48,54,48}},
/* 897 */ {(16<<2)|3,{48,54,49}},
/* 898 */ {(16<<2)|3,{48,54,50}},
/* 899 */ {(16<<2)|3,{48,54,97}},
/* 900 */ {(16<<2)|3,{48,54,99}},
/* 901 */ {(16<<2)|3,{48,54,101}},
/* 902 */ {(16<<2)|3,{48,54,105}},
/* 903 */ {(16<<2)|3,{48,54,111}},
/* 904 */ {(16<<2)|3,{48,54,115}},
/* 905 */ {(16<<2)|3,{48,54,116}},
/* 906 */ {(11<<2)|2,{48,54,0}},
/* 907 */ {(11<<2)|2,{48,54,0}},
/* 908 */ {(11<<2)|2,{48,54,0}},
/* 909 */ {(11<<2)|2,{48,54,0}},
/* 910 */ {(11<<2)|2,{48,54,0}},
/* 911 */ {(11<<2)|2,{48,54,0}},
/* 912 */ {(11<<2)|2,{48,54,0}},
/* 913 */ {(11<<2)|2,{48,54,0}},
/* 914 */ {(11<<2)|2,{48,54,0}},
/* 915 */ {(11<<2)|2,{48,54,0}},
/* 916 */ {(11<<2)|2,{48,54,0}},
/* 917 */ {(11<<2)|2,{48,54,0}},
/* 918 */ {(11<<2)|2,{48,54,0}},
/* 919 */ {(11<<2)|2,{48,54,0}},
/* 920 */ {(11<<2)|2,{48,54,0}},
/* 921 */ {(11<<2)|2,{48,54,0}},
/* 922 */ {(11<<2)|2,{48,54,0}},
/* 923 */ {(11<<2)|2,{48,54,0}},
/* 924 */ {(11<<2)|2,{48,54,0}},
/* 925 */ {(11<<2)|2,{48,54,0}},
/* 926 */ {(11<<2)|2,{48,54,0}},
/* 927 */ {(11<<2)|2,{48,54,0}},
/* 928 */ {(16<<2)|3,{48,55,48}},
/* 929 */ {(16<<2)|3,{48,55,49}},
/* 930 */ {(16<<2)|3,{48,55,50}},
/* 931 */ {(16<<2)|3,{48,55,97}},
/* 932 */ {(16<<2)|3,{48,55,99}},
/* 933 */ {(16<<2)|3,{48,55,101}},
/* 934 */ {(16<<2)|3,{48,55,105}},
/* 935 */ {(16<<2)|3,{48,55,111}},
/* 936 */ {(16<<2)|3,{48,55,115}},
/* 937 */ {(16<<2)|3,{48,55,116}},
/* 938 */ {(11<<2)|2,{48,55,0}},
/* 939 */ {(11<<2)|2,{48,55,0}},
/* 940 */ {(11<<2)|2,{48,55,0}},
/* 941 */ {(11<<2)|2,{48,55,0}},
/* 942 */ {(11<<2)|2,{48,55,0}},
/* 943 */ {(11<<2)|2,{48,55,0}},
/* 944 */ {(11<<2)|2,{48,55,0}},
/* 945 */ {(11<<2)|2,{48,55,0}},
/* 946 */ {(11<<2)|2,{48,55,0}},
/* 947 */ {(11<<2)|2,{48,55,0}},
/* 948 */ {(11<<2)|2,{48,55,0}},
/* 949 */ {(11<<2)|2,{48,55,0}},
/* 950 */ {(11<<2)|2,{48,55,0}},
/* 951 */ {(11<<2)|2,{48,55,0}},
/* 952 */ {(11<<2)|2,{48,55,0}},
/* 953 */ {(11<<2)|2,{48,55,0}},
/* 954 */ {(11<<2)|2,{48,55,0}},
/* 955 */ {(11<<2)|2,{48,55,0}},
/* 956 */ {(11<<2)|2,{48,55,0}},
/* 957 */ {(11<<2)|2,{48,55,0}},
/* 958 */ {(11<<2)|2,{48,55,0}},
/* 959 */ {(11<<2)|2,{48,55,0}},
/* 960 */ {(16<<2)|3,{48,56,48}},
/* 961 */ {(16<<2)|3,{48,56,49}},
/* 962 */ {(16<<2)|3,{48,56,50}},
/* 963 */ {(16<<2)|3,{48,56,97}},
/* 964 */ {(16<<2)|3,{48,56,99}},
/* 965 */ {(16<<2)|3,{48,56,101}},
/* 966 */ {(16<<2)|3,{48,56,105}},
/* 967 */ {(16<<2)|3,{48,56,111}},
/* 968 */ {(16<<2)|3,{48,56,115}},
/* 969 */ {(16<<2)|3,{48,56,116}},
/* 970 */ {(11<<2)|2,{48,56,0}},
/* 971 */ {(11<<2)|2,{48,56,0}},
/* 972 */ {(11<<2)|2,{48,56,0}},
/* 973 */ {(11<<2)|2,{48,56,0}},
/* 974 */ {(11<<2)|2,{48,56,0}},
/* 975 */ {(11<<2)|2,{48,56,0}},
/* 976 */ {(11<<2)|2,{48,56,0}},
/* 977 */ {(11<<2)|2,{48,56,0}},
/* 978 */ {(11<<2)|2,{48,56,0}},
/* 979 */ {(11<<2)|2,{48,56,0}},
/* 980 */ {(11<<2)|2,{48,56,0}},
/* 981 */ {(11<<2)|2,{48,56,0}},
/* 982 */ {(11<<2)|2,{48,56,0}},
/* 983 */ {(11<<2)|2,{48,56,0}},
/* 984 */ {(11<<2)|2,{48,56,0}},
/* 985 */ {(11<<2)|2,{48,56,0}},
/* 986 */ {(11<<2)|2,{48,56,0}},
/* 987 */ {(11<<2)|2,{48,56,0}},
/* 988 */ {(11<<2)|2,{48,56,0}},
/* 989 */ {(11<<2)|2,{48,56,0}},
/* 990 */ {(11<<2)|2,{48,56,0}},
/* 991 */ {(11<<2)|2,{48,56,0}},
/* 992 */ {(16<<2)|3,{48,57,48}},
/* 993 */ {(16<<2)|3,{48,57,49}},
/* 994 */ {(16<<2)|3,{48,57,50}},
/* 995 */ {(16<<2)|3,{48,57,97}},
/* 996 */ {(16<<2)|3,{48,57,99}},
/* 997 */ {(16<<2)|3,{48,57,101}},
/* 998 */ {(16<<2)|3,{48,57,105}},
/* 999 */ {(16<<2)|3,{48,57,111}},
/* 1000 */ {(16<<2)|3,{48,57,115}},
/* 1001 */ {(16<<2)|3,{48,57,116}},
/* 1002 */ {(11<<2)|2,{48,57,0}},
/* 1003 */ {(11<<2)|2,{48,57,0}},
/* 1004 */ {(11<<2)|2,{48,57,0}},
/* 1005 */ {(11<<2)|2,{48,57,0}},
/* 1006 */ {(11<<2)|2,{48,57,0}},
/* 1007 */ {(11<<2)|2,{48,57,0}},
/* 1008 */ {(11<<2)|2,{48,57,0}},
/* 1009 */ {(11<<2)|2,{48,57,0}},
/* 1010 */ {(11<<2)|2,{48,57,0}},
/* 1011 */ {(11<<2)|2,{48,57,0}},
/* 1012 */ {(11<<2)|2,{48,57,0}},
/* 1013 */ {(11<<2)|2,{48,57,0}},
/* 1014 */ {(11<<2)|2,{48,57,0}},
/* 1015 */ {(11<<2)|2,{48,57,0}},
/* 1016 */ {(11<<2)|2,{48,57,0}},
/* 1017 */ {(11<<2)|2,{48,57,0}},
/* 1018 */ {(11<<2)|2,{48,57,0}},
/* 1019 */ {(11<<2)|2,{48,57,0}},
/* 1020 */ {(11<<2)|2,{48,57,0}},
/* 1021 */ {(11<<2)|2,{48,57,0}},
/* 1022 */ {(11<<2)|2,{48,57,0}},
/* 1023 */ {(11<<2)|2,{48,57,0}},
/* 1024 */ {(16<<2)|3,{48,61,48}},
/* 1025 */ {(16<<2)|3,{48,61,49}},
/* 1026 */ {(16<<2)|3,{48,61,50}},
/* 1027 */ {(16<<2)|3,{48,61,97}},
/* 1028 */ {(16<<2)|3,{48,61,99}},
/* 1029 */ {(16<<2)|3,{48,61,101}},
/* 1030 */ {(16<<2)|3,{48,61,105}},
/* 1031 */ {(16<<2)|3,{48,61,111}},
/* 1032 */ {(16<<2)|3,{48,61,115}},
/* 1033 */ {(16<<2)|3,{48,61,116}},
/* 1034 */ {(11<<2)|2,{48,61,0}},
/* 1035 */ {(11<<2)|2,{48,61,0}},
/* 1036 */ {(11<<2)|2,{48,61,0}},
/* 1037 */ {(11<<2)|2,{48,61,0}},
/* 1038 */ {(11<<2)|2,{48,61,0}},
/* 1039 */ {(11<<2)|2,{48,61,0}},
/* 1040 */ {(11<<2)|2,{48,61,0}},
/* 1041 */ {(11<<2)|2,{48,61,0}},
/* 1042 */ {(11<<2)|2,{48,61,0}},
/* 1043 */ {(11<<2)|2,{48,61,0}},
/* 1044 */ {(11<<2)|2,{48,61,0}},
/* 1045 */ {(11<<2)|2,{48,61,0}},
/* 1046 */ {(11<<2)|2,{48,61,0}},
/* 1047 */ {(11<<2)|2,{48,61,0}},
/* 1048 */ {(11<<2)|2,{48,61,0}},
/* 1049 */ {(11<<2)|2,{48,61,0}},
/* 1050 */ {(11<<2)|2,{48,61,0}},
/* 1051 */ {(11<<2)|2,{48,61,0}},
/* 1052 */ {(11<<2)|2,{48,61,0}},
/* 1053 */ {(11<<2)|2,{48,61,0}},
/* 1054 */ {(11<<2)|2,{48,61,0}},
/* 1055 */ {(11<<2)|2,{48,61,0}},
/* 1056 */ {(16<<2)|3,{48,65,48}},
/* 1057 */ {(16<<2)|3,{48,65,49}},
/* 1058 */ {(16<<2)|3,{48,65,50}},
/* 1059 */ {(16<<2)|3,{48,65,97}},
/* 1060 */ {(16<<2)|3,{48,65,99}},
/* 1061 */ {(16<<2)|3,{48,65,101}},
/* 1062 */ {(16<<2)|3,{48,65,105}},
/* 1063 */ {(16<<2)|3,{48,65,111}},
/* 1064 */ {(16<<2)|3,{48,65,115}},
/* 1065 */ {(16<<2)|3,{48,65,116}},
/* 1066 */ {(11<<2)|2,{48,65,0}},
/* 1067 */ {(11<<2)|2,{48,65,0}},
/* 1068 */ {(11<<2)|2,{48,65,0}},
/* 1069 */ {(11<<2)|2,{48,65,0}},
/* 1070 */ {(11<<2)|2,{48,65,0}},
/* 1071 */ {(11<<2)|2,{48,65,0}},
/* 1072 */ {(11<<2)|2,{48,65,0}},
/* 1073 */ {(11<<2)|2,{48,65,0}},
/* 1074 */ {(11<<2)|2,{48,65,0}},
/* 1075 */ {(11<<2)|2,{48,65,0}},
/* 1076 */ {(11<<2)|2,{48,65,0}},
/* 1077 */ {(11<<2)|2,{48,65,0}},
/* 1078 */ {(11<<2)|2,{48,65,0}},
/* 1079 */ {(11<<2)|2,{48,65,0}},
/* 1080 */ {(11<<2)|2,{48,65,0}},
/* 1081 */ {(11<<2)|2,{48,65,0}},
/* 1082 */ {(11<<2)|2,{48,65,0}},
/* 1083 */ {(11<<2)|2,{48,65,0}},
/* 1084 */ {(11<<2)|2,{48,65,0}},
/* 1085 */ {(11<<2)|2,{48,65,0}},
/* 1086 */ {(11<<2)|2,{48,65,0}},
/* 1087 */ {(11<<2)|2,{48,65,0}},
/* 1088 */ {(16<<2)|3,{48,95,48}},
/* 1089 */ {(16<<2)|3,{48,95,49}},
/* 1090 */ {(16<<2)|3,{48,95,50}},
/* 1091 */ {(16<<2)|3,{48,95,97}},
/* 1092 */ {(16<<2)|3,{48,95,99}},
/* 1093 */ {(16<<2)|3,{48,95,101}},
/* 1094 */ {(16<<2)|3,{48,95,105}},
/* 1095 */ {(16<<2)|3,{48,95,111}},
/* 1096 */ {(16<<2)|3,{48,95,115}},
/* 1097 */ {(16<<2)|3,{48,95,116}},
/* 1098 */ {(11<<2)|2,{48,95,0}},
/* 1099 */ {(11<<2)|2,{48,95,0}},
/* 1100 */ {(11<<2)|2,{48,95,0}},
/* 1101 */ {(11<<2)|2,{48,95,0}},
/* 1102 */ {(11<<2)|2,{48,95,0}},
/* 1103 */ {(11<<2)|2,{48,95,0}},
/* 1104 */ {(11<<2)|2,{48,95,0}},
/* 1105 */ {(11<<2)|2,{48,95,0}},
/* 1106 */ {(11<<2)|2,{48,95,0}},
/* 1107 */ {(11<<2)|2,{48,95,0}},
/* 1108 */ {(11<<2)|2,{48,95,0}},
/* 1109 */ {(11<<2)|2,{48,95,0}},
/* 1110 */ {(11<<2)|2,{48,95,0}},
/* 1111 */ {(11<<2)|2,{48,95,0}},
/* 1112 */ {(11<<2)|2,{48,95,0}},
/* 1113 */ {(11<<2)|2,{48,95,0}},
/* 1114 */ {(11<<2)|2,{48,95,0}},
/* 1115 */ {(11<<2)|2,{48,95,0}},
/* 1116 */ {(11<<2)|2,{48,95,0}},
/* 1117 */ {(11<<2)|2,{48,95,0}},
/* 1118 */ {(11<<2)|2,{48,95,0}},
/* 1119 */ {(11<<2)|2,{48,95,0}},
/* 1120 */ {(16<<2)|3,{48,98,48}},
/* 1121 */ {(16<<2)|3,{48,98,49}},
/* 1122 */ {(16<<2)|3,{48,98,50}},
/* 1123 */ {(16<<2)|3,{48,98,97}},
/* 1124 */ {(16<<2)|3,{48,98,99}},
/* 1125 */ {(16<<2)|3,{48,98,101}},
/* 1126 */ {(16<<2)|3,{48,98,105}},
/* 1127 */ {(16<<2)|3,{48,98,111}},
/* 1128 */ {(16<<2)|3,{48,98,115}},
/* 1129 */ {(16<<2)|3,{48,98,116}},
/* 1130 */ {(11<<2)|2,{48,98,0}},
/* 1131 */ {(11<<2)|2,{48,98,0}},
/* 1132 */ {(11<<2)|2,{48,98,0}},
/* 1133 */ {(11<<2)|2,{48,98,0}},
/* 1134 */ {(11<<2)|2,{48,98,0}},
/* 1135 */ {(11<<2)|2,{48,98,0}},
/* 1136 */ {(11<<2)|2,{48,98,0}},
/* 1137 */ {(11<<2)|2,{48,98,0}},
/* 1138 */ {(11<<2)|2,{48,98,0}},
/* 1139 */ {(11<<2)|2,{48,98,0}},
/* 1140 */ {(11<<2)|2,{48,98,0}},
/* 1141 */ {(11<<2)|2,{48,98,0}},
/* 1142 */ {(11<<2)|2,{48,98,0}},
/* 1143 */ {(11<<2)|2,{48,98,0}},
/* 1144 */ {(11<<2)|2,{48,98,0}},
/* 1145 */ {(11<<2)|2,{48,98,0}},
/* 1146 */ {(11<<2)|2,{48,98,0}},
/* 1147 */ {(11<<2)|2,{48,98,0}},
/* 1148 */ {(11<<2)|2,{48,98,0}},
/* 1149 */ {(11<<2)|2,{48,98,0}},
/* 1150 */ {(11<<2)|2,{48,98,0}},
/* 1151 */ {(11<<2)|2,{48,98,0}},
/* 1152 */ {(16<<2)|3,{48,100,48}},
/* 1153 */ {(16<<2)|3,{48,100,49}},
/* 1154 */ {(16<<2)|3,{48,100,50}},
/* 1155 */ {(16<<2)|3,{48,100,97}},
/* 1156 */ {(16<<2)|3,{48,100,99}},
/* 1157 */ {(16<<2)|3,{48,100,101}},
/* 1158 */ {(16<<2)|3,{48,100,105}},
/* 1159 */ {(16<<2)|3,{48,100,111}},
/* 1160 */ {(16<<2)|3,{48,100,115}},
/* 1161 */ {(16<<2)|3,{48,100,116}},
/* 1162 */ {(11<<2)|2,{48,100,0}},
/* 1163 */ {(11<<2)|2,{48,100,0}},
/* 1164 */ {(11<<2)|2,{48,100,0}},
/* 1165 */ {(11<<2)|2,{48,100,0}},
/* 1166 */ {(11<<2)|2,{48,100,0}},
/* 1167 */ {(11<<2)|2,{48,100,0}},
/* 1168 */ {(11<<2)|2,{48,100,0}},
/* 1169 */ {(11<<2)|2,{48,100,0}},
/* 1170 */ {(11<<2)|2,{48,100,0}},
/* 1171 */ {(11<<2)|2,{48,100,0}},
/* 1172 */ {(11<<2)|2,{48,100,0}},
/* 1173 */ {(11<<2)|2,{48,100,0}},
/* 1174 */ {(11<<2)|2,{48,100,0}},
/* 1175 */ {(11<<2)|2,{48,100,0}},
/* 1176 */ {(11<<2)|2,{48,100,0}},
/* 1177 */ {(11<<2)|2,{48,100,0}},
/* 1178 */ {(11<<2)|2,{48,100,0}},
/* 1179 */ {(11<<2)|2,{48,100,0}},
/* 1180 */ {(11<<2)|2,{48,100,0}},
/* 1181 */ {(11<<2)|2,{48,100,0}},
/* 1182 */ {(11<<2)|2,{48,100,0}},
/* 1183 */ {(11<<2)|2,{48,100,0}},
/* 1184 */ {(16<<2)|3,{48,102,48}},
/* 1185 */ {(16<<2)|3,{48,102,49}},
/* 1186 */ {(16<<2)|3,{48,102,50}},
/* 1187 */ {(16<<2)|3,{48,102,97}},
/* 1188 */ {(16<<2)|3,{48,102,99}},
/* 1189 */ {(16<<2)|3,{48,102,101}},
/* 1190 */ {(16<<2)|3,{48,102,105}},
/* 1191 */ {(16<<2)|3,{48,102,111}},
/* 1192 */ {(16<<2)|3,{48,102,115}},
/* 1193 */ {(16<<2)|3,{48,102,116}},
/* 1194 */ {(11<<2)|2,{48,102,0}},
/* 1195 */ {(11<<2)|2,{48,102,0}},
/* 1196 */ {(11<<2)|2,{48,102,0}},
/* 1197 */ {(11<<2)|2,{48,102,0}},
/* 1198 */ {(11<<2)|2,{48,102,0}},
/* 1199 */ {(11<<2)|2,{48,102,0}},
/* 1200 */ {(11<<2)|2,{48,102,0}},
/* 1201 */ {(11<<2)|2,{48,102,0}},
/* 1202 */ {(11<<2)|2,{48,102,0}},
/* 1203 */ {(11<<2)|2,{48,102,0}},
/* 1204 */ {(11<<2)|2,{48,102,0}},
/* 1205 */ {(11<<2)|2,{48,102,0}},
/* 1206 */ {(11<<2)|2,{48,102,0}},
/* 1207 */ {(11<<2)|2,{48,102,0}},
/* 1208 */ {(11<<2)|2,{48,102,0}},
/* 1209 */ {(11<<2)|2,{48,102,0}},
/* 1210 */ {(11<<2)|2,{48,102,0}},
/* 1211 */ {(11<<2)|2,{48,102,0}},
/* 1212 */ {(11<<2)|2,{48,102,0}},
/* 1213 */ {(11<<2)|2,{48,102,0}},
/* 1214 */ {(11<<2)|2,{48,102,0}},
/* 1215 */ {(11<<2)|2,{48,102,0}},
/* 1216 */ {(16<<2)|3,{48,103,48}},
/* 1217 */ {(16<<2)|3,{48,103,49}},
/* 1218 */ {(16<<2)|3,{48,103,50}},
/* 1219 */ {(16<<2)|3,{48,103,97}},
/* 1220 */ {(16<<2)|3,{48,103,99}},
/* 1221 */ {(16<<2)|3,{48,103,101}},
/* 1222 */ {(16<<2)|3,{48,103,105}},
/* 1223 */ {(16<<2)|3,{48,103,111}},
/* 1224 */ {(16<<2)|3,{48,103,115}},
/* 1225 */ {(16<<2)|3,{48,103,116}},
/* 1226 */ {(11<<2)|2,{48,103,0}},
/* 1227 */ {(11<<2)|2,{48,103,0}},
/* 1228 */ {(11<<2)|2,{48,103,0}},
/* 1229 */ {(11<<2)|2,{48,103,0}},
/* 1230 */ {(11<<2)|2,{48,103,0}},
/* 1231 */ {(11<<2)|2,{48,103,0}},
/* 1232 */ {(11<<2)|2,{48,103,0}},
/* 1233 */ {(11<<2)|2,{48,103,0}},
/* 1234 */ {(11<<2)|2,{48,103,0}},
/* 1235 */ {(11<<2)|2,{48,103,0}},
/* 1236 */ {(11<<2)|2,{48,103,0}},
/* 1237 */ {(11<<2)|2,{48,103,0}},
/* 1238 */ {(11<<2)|2,{48,103,0}},
/* 1239 */ {(11<<2)|2,{48,103,0}},
/* 1240 */ {(11<<2)|2,{48,103,0}},
/* 1241 */ {(11<<2)|2,{48,103,0}},
/* 1242 */ {(11<<2)|2,{48,103,0}},
/* 1243 */ {(11<<2)|2,{48,103,0}},
/* 1244 */ {(11<<2)|2,{48,103,0}},
/* 1245 */ {(11<<2)|2,{48,103,0}},
/* 1246 */ {(11<<2)|2,{48,103,0}},
/* 1247 */ {(11<<2)|2,{48,103,0}},
/* 1248 */ {(16<<2)|3,{48,104,48}},
/* 1249 */ {(16<<2)|3,{48,104,49}},
/* 1250 */ {(16<<2)|3,{48,104,50}},
/* 1251 */ {(16<<2)|3,{48,104,97}},
/* 1252 */ {(16<<2)|3,{48,104,99}},
/* 1253 */ {(16<<2)|3,{48,104,101}},
/* 1254 */ {(16<<2)|3,{48,104,105}},
/* 1255 */ {(16<<2)|3,{48,104,111}},
/* 1256 */ {(16<<2)|3,{48,104,115}},
/* 1257 */ {(16<<2)|3,{48,104,116}},
/* 1258 */ {(11<<2)|2,{48,104,0}},
/* 1259 */ {(11<<2)|2,{48,104,0}},
/* 1260 */ {(11<<2)|2,{48,104,0}},
/* 1261 */ {(11<<2)|2,{48,104,0}},
/* 1262 */ {(11<<2)|2,{48,104,0}},
/* 1263 */ {(11<<2)|2,{48,104,0}},
/* 1264 */ {(11<<2)|2,{48,104,0}},
/* 1265 */ {(11<<2)|2,{48,104,0}},
/* 1266 */ {(11<<2)|2,{48,104,0}},
/* 1267 */ {(11<<2)|2,{48,104,0}},
/* 1268 */ {(11<<2)|2,{48,104,0}},
/* 1269 */ {(11<<2)|2,{48,104,0}},
/* 1270 */ {(11<<2)|2,{48,104,0}},
/* 1271 */ {(11<<2)|2,{48,104,0}},
/* 1272 */ {(11<<2)|2,{48,104,0}},
/* 1273 */ {(11<<2)|2,{48,104,0}},
/* 1274 */ {(11<<2)|2,{48,104,0}},
/* 1275 */ {(11<<2)|2,{48,104,0}},
/* 1276 */ {(11<<2)|2,{48,104,0}},
/* 1277 */ {(11<<2)|2,{48,104,0}},
/* 1278 */ {(11<<2)|2,{48,104,0}},
/* 1279 */ {(11<<2)|2,{48,104,0}},
/* 1280 */ {(16<<2)|3,{48,108,48}},
/* 1281 */ {(16<<2)|3,{48,108,49}},
/* 1282 */ {(16<<2)|3,{48,108,50}},
/* 1283 */ {(16<<2)|3,{48,108,97}},
/* 1284 */ {(16<<2)|3,{48,108,99}},
/* 1285 */ {(16<<2)|3,{48,108,101}},
/* 1286 */ {(16<<2)|3,{48,108,105}},
/* 1287 */ {(16<<2)|3,{48,108,111}},
/* 1288 */ {(16<<2)|3,{48,108,115}},
/* 1289 */ {(16<<2)|3,{48,108,116}},
/* 1290 */ {(11<<2)|2,{48,108,0}},
/* 1291 */ {(11<<2)|2,{48,108,0}},
/* 1292 */ {(11<<2)|2,{48,108,0}},
/* 1293 */ {(11<<2)|2,{48,108,0}},
/* 1294 */ {(11<<2)|2,{48,108,0}},
/* 1295 */ {(11<<2)|2,{48,108,0}},
/* 1296 */ {(11<<2)|2,{48,108,0}},
/* 1297 */ {(11<<2)|2,{48,108,0}},
/* 1298 */ {(11<<2)|2,{48,108,0}},
/* 1299 */ {(11<<2)|2,{48,108,0}},
/* 1300 */ {(11<<2)|2,{48,108,0}},
/* 1301 */ {(11<<2)|2,{48,108,0}},
/* 1302 */ {(11<<2)|2,{48,108,0}},
/* 1303 */ {(11<<2)|2,{48,108,0}},
/* 1304 */ {(11<<2)|2,{48,108,0}},
/* 1305 */ {(11<<2)|2,{48,108,0}},
/* 1306 */ {(11<<2)|2,{48,108,0}},
/* 1307 */ {(11<<2)|2,{48,108,0}},
/* 1308 */ {(11<<2)|2,{48,108,0}},
/* 1309 */ {(11<<2)|2,{48,108,0}},
/* 1310 */ {(11<<2)|2,{48,108,0}},
/* 1311 */ {(11<<2)|2,{48,108,0}},
/* 1312 */ {(16<<2)|3,{48,109,48}},
/* 1313 */ {(16<<2)|3,{48,109,49}},
/* 1314 */ {(16<<2)|3,{48,109,50}},
/* 1315 */ {(16<<2)|3,{48,109,97}},
/* 1316 */ {(16<<2)|3,{48,109,99}},
/* 1317 */ {(16<<2)|3,{48,109,101}},
/* 1318 */ {(16<<2)|3,{48,109,105}},
/* 1319 */ {(16<<2)|3,{48,109,111}},
/* 1320 */ {(16<<2)|3,{48,109,115}},
/* 1321 */ {(16<<2)|3,{48,109,116}},
/* 1322 */ {(11<<2)|2,{48,109,0}},
/* 1323 */ {(11<<2)|2,{48,109,0}},
/* 1324 */ {(11<<2)|2,{48,109,0}},
/* 1325 */ {(11<<2)|2,{48,109,0}},
/* 1326 */ {(11<<2)|2,{48,109,0}},
/* 1327 */ {(11<<2)|2,{48,109,0}},
/* 1328 */ {(11<<2)|2,{48,109,0}},
/* 1329 */ {(11<<2)|2,{48,109,0}},
/* 1330 */ {(11<<2)|2,{48,109,0}},
/* 1331 */ {(11<<2)|2,{48,109,0}},
/* 1332 */ {(11<<2)|2,{48,109,0}},
/* 1333 */ {(11<<2)|2,{48,109,0}},
/* 1334 */ {(11<<2)|2,{48,109,0}},
/* 1335 */ {(11<<2)|2,{48,109,0}},
/* 1336 */ {(11<<2)|2,{48,109,0}},
/* 1337 */ {(11<<2)|2,{48,109,0}},
/* 1338 */ {(11<<2)|2,{48,109,0}},
/* 1339 */ {(11<<2)|2,{48,109,0}},
/* 1340 */ {(11<<2)|2,{48,109,0}},
/* 1341 */ {(11<<2)|2,{48,109,0}},
/* 1342 */ {(11<<2)|2,{48,109,0}},
/* 1343 */ {(11<<2)|2,{48,109,0}},
/* 1344 */ {(16<<2)|3,{48,110,48}},
/* 1345 */ {(16<<2)|3,{48,110,49}},
/* 1346 */ {(16<<2)|3,{48,110,50}},
/* 1347 */ {(16<<2)|3,{48,110,97}},
/* 1348 */ {(16<<2)|3,{48,110,99}},
/* 1349 */ {(16<<2)|3,{48,110,101}},
/* 1350 */ {(16<<2)|3,{48,110,105}},
/* 1351 */ {(16<<2)|3,{48,110,111}},
/* 1352 */ {(16<<2)|3,{48,110,115}},
/* 1353 */ {(16<<2)|3,{48,110,116}},
/* 1354 */ {(11<<2)|2,{48,110,0}},
/* 1355 */ {(11<<2)|2,{48,110,0}},
/* 1356 */ {(11<<2)|2,{48,110,0}},
/* 1357 */ {(11<<2)|2,{48,110,0}},
/* 1358 */ {(11<<2)|2,{48,110,0}},
/* 1359 */ {(11<<2)|2,{48,110,0}},
/* 1360 */ {(11<<2)|2,{48,110,0}},
/* 1361 */ {(11<<2)|2,{48,110,0}},
/* 1362 */ {(11<<2)|2,{48,110,0}},
/* 1363 */ {(11<<2)|2,{48,110,0}},
/* 1364 */ {(11<<2)|2,{48,110,0}},
/* 1365 */ {(11<<2)|2,{48,110,0}},
/* 1366 */ {(11<<2)|2,{48,110,0}},
/* 1367 */ {(11<<2)|2,{48,110,0}},
/* 1368 */ {(11<<2)|2,{48,110,0}},
/* 1369 */ {(11<<2)|2,{48,110,0}},
/* 1370 */ {(11<<2)|2,{48,110,0}},
/* 1371 */ {(11<<2)|2,{48,110,0}},
/* 1372 */ {(11<<2)|2,{48,110,0}},
/* 1373 */ {(11<<2)|2,{48,110,0}},
/* 1374 */ {(11<<2)|2,{48,110,0}},
/* 1375 */ {(11<<2)|2,{48,110,0}},
/* 1376 */ {(16<<2)|3,{48,112,48}},
/* 1377 */ {(16<<2)|3,{48,112,49}},
/* 1378 */ {(16<<2)|3,{48,112,50}},
/* 1379 */ {(16<<2)|3,{48,112,97}},
/* 1380 */ {(16<<2)|3,{48,112,99}},
/* 1381 */ {(16<<2)|3,{48,112,101}},
/* 1382 */ {(16<<2)|3,{48,112,105}},
/* 1383 */ {(16<<2)|3,{48,112,111}},
/* 1384 */ {(16<<2)|3,{48,112,115}},
/* 1385 */ {(16<<2)|3,{48,112,116}},
/* 1386 */ {(11<<2)|2,{48,112,0}},
/* 1387 */ {(11<<2)|2,{48,112,0}},
/* 1388 */ {(11<<2)|2,{48,112,0}},
/* 1389 */ {(11<<2)|2,{48,112,0}},
/* 1390 */ {(11<<2)|2,{48,112,0}},
/* 1391 */ {(11<<2)|2,{48,112,0}},
/* 1392 */ {(11<<2)|2,{48,112,0}},
/* 1393 */ {(11<<2)|2,{48,112,0}},
/* 1394 */ {(11<<2)|2,{48,112,0}},
/* 1395 */ {(11<<2)|2,{48,112,0}},
/* 1396 */ {(11<<2)|2,{48,112,0}},
/* 1397 */ {(11<<2)|2,{48,112,0}},
/* 1398 */ {(11<<2)|2,{48,112,0}},
/* 1399 */ {(11<<2)|2,{48,112,0}},
/* 1400 */ {(11<<2)|2,{48,112,0}},
/* 1401 */ {(11<<2)|2,{48,112,0}},
/* 1402 */ {(11<<2)|2,{48,112,0}},
/* 1403 */ {(11<<2)|2,{48,112,0}},
/* 1404 */ {(11<<2)|2,{48,112,0}},
/* 1405 */ {(11<<2)|2,{48,112,0}},
/* 1406 */ {(11<<2)|2,{48,112,0}},
/* 1407 */ {(11<<2)|2,{48,112,0}},
/* 1408 */ {(16<<2)|3,{48,114,48}},
/* 1409 */ {(16<<2)|3,{48,114,49}},
/* 1410 */ {(16<<2)|3,{48,114,50}},
/* 1411 */ {(16<<2)|3,{48,114,97}},
/* 1412 */ {(16<<2)|3,{48,114,99}},
/* 1413 */ {(16<<2)|3,{48,114,101}},
/* 1414 */ {(16<<2)|3,{48,114,105}},
/* 1415 */ {(16<<2)|3,{48,114,111}},
/* 1416 */ {(16<<2)|3,{48,114,115}},
/* 1417 */ {(16<<2)|3,{48,114,116}},
/* 1418 */ {(11<<2)|2,{48,114,0}},
/* 1419 */ {(11<<2)|2,{48,114,0}},
/* 1420 */ {(11<<2)|2,{48,114,0}},
/* 1421 */ {(11<<2)|2,{48,114,0}},
/* 1422 */ {(11<<2)|2,{48,114,0}},
/* 1423 */ {(11<<2)|2,{48,114,0}},
/* 1424 */ {(11<<2)|2,{48,114,0}},
/* 1425 */ {(11<<2)|2,{48,114,0}},
/* 1426 */ {(11<<2)|2,{48,114,0}},
/* 1427 */ {(11<<2)|2,{48,114,0}},
/* 1428 */ {(11<<2)|2,{48,114,0}},
/* 1429 */ {(11<<2)|2,{48,114,0}},
/* 1430 */ {(11<<2)|2,{48,114,0}},
/* 1431 */ {(11<<2)|2,{48,114,0}},
/* 1432 */ {(11<<2)|2,{48,114,0}},
/* 1433 */ {(11<<2)|2,{48,114,0}},
/* 1434 */ {(11<<2)|2,{48,114,0}},
/* 1435 */ {(11<<2)|2,{48,114,0}},
/* 1436 */ {(11<<2)|2,{48,114,0}},
/* 1437 */ {(11<<2)|2,{48,114,0}},
/* 1438 */ {(11<<2)|2,{48,114,0}},
/* 1439 */ {(11<<2)|2,{48,114,0}},
/* 1440 */ {(16<<2)|3,{48,117,48}},
/* 1441 */ {(16<<2)|3,{48,117,49}},
/* 1442 */ {(16<<2)|3,{48,117,50}},
/* 1443 */ {(16<<2)|3,{48,117,97}},
/* 1444 */ {(16<<2)|3,{48,117,99}},
/* 1445 */ {(16<<2)|3,{48,117,101}},
/* 1446 */ {(16<<2)|3,{48,117,105}},
/* 1447 */ {(16<<2)|3,{48,117,111}},
/* 1448 */ {(16<<2)|3,{48,117,115}},
/* 1449 */ {(16<<2)|3,{48,117,116}},
/* 1450 */ {(11<<2)|2,{48,117,0}},
/* 1451 */ {(11<<2)|2,{48,117,0}},
/* 1452 */ {(11<<2)|2,{48,117,0}},
/* 1453 */ {(11<<2)|2,{48,117,0}},
/* 1454 */ {(11<<2)|2,{48,117,0}},
/* 1455 */ {(11<<2)|2,{48,117,0}},
/* 1456 */ {(11<<2)|2,{48,117,0}},
/* 1457 */ {(11<<2)|2,{48,117,0}},
/* 1458 */ {(11<<2)|2,{48,117,0}},
/* 1459 */ {(11<<2)|2,{48,117,0}},
/* 1460 */ {(11<<2)|2,{48,117,0}},
/* 1461 */ {(11<<2)|2,{48,117,0}},
/* 1462 */ {(11<<2)|2,{48,117,0}},
/* 1463 */ {(11<<2)|2,{48,117,0}},
/* 1464 */ {(11<<2)|2,{48,117,0}},
/* 1465 */ {(11<<2)|2,{48,117,0}},
/* 1466 */ {(11<<2)|2,{48,117,0}},
/* 1467 */ {(11<<2)|2,{48,117,0}},
/* 1468 */ {(11<<2)|2,{48,117,0}},
/* 1469 */ {(11<<2)|2,{48,117,0}},
/* 1470 */ {(11<<2)|2,{48,117,0}},
/* 1471 */ {(11<<2)|2,{48,117,0}},
/* 1472 */ {(12<<2)|2,{48,58,0}},
/* 1473 */ {(12<<2)|2,{48,58,0}},
/* 1474 */ {(12<<2)|2,{48,58,0}},
/* 1475 */ {(12<<2)|2,{48,58,0}},
/* 1476 */ {(12<<2)|2,{48,58,0}},
/* 1477 */ {(12<<2)|2,{48,58,0}},
/* 1478 */ {(12<<2)|2,{48,58,0}},
/* 1479 */ {(12<<2)|2,{48,58,0}},
/* 1480 */ {(12<<2)|2,{48,58,0}},
/* 1481 */ {(12<<2)|2,{48,58,0}},
/* 1482 */ {(12<<2)|2,{48,58,0}},
/* 1483 */ {(12<<2)|2,{48,58,0}},
/* 1484 */ {(12<<2)|2,{48,58,0}},
/* 1485 */ {(12<<2)|2,{48,58,0}},
/* 1486 */ {(12<<2)|2,{48,58,0}},
/* 1487 */ {(12<<2)|2,{48,58,0}},
/* 1488 */ {(12<<2)|2,{48,66,0}},
/* 1489 */ {(12<<2)|2,{48,66,0}},
/* 1490 */ {(12<<2)|2,{48,66,0}},
/* 1491 */ {(12<<2)|2,{48,66,0}},
/* 1492 */ {(12<<2)|2,{48,66,0}},
/* 1493 */ {(12<<2)|2,{48,66,0}},
/* 1494 */ {(12<<2)|2,{48,66,0}},
/* 1495 */ {(12<<2)|2,{48,66,0}},
/* 1496 */ {(12<<2)|2,{48,66,0}},
/* 1497 */ {(12<<2)|2,{48,66,0}},
/* 1498 */ {(12<<2)|2,{48,66,0}},
/* 1499 */ {(12<<2)|2,{48,66,0}},
/* 1500 */ {(12<<2)|2,{48,66,0}},
/* 1501 */ {(12<<2)|2,{48,66,0}},
/* 1502 */ {(12<<2)|2,{48,66,0}},
/* 1503 */ {(12<<2)|2,{48,66,0}},
/* 1504 */ {(12<<2)|2,{48,67,0}},
/* 1505 */ {(12<<2)|2,{48,67,0}},
/* 1506 */ {(12<<2)|2,{48,67,0}},
/* 1507 */ {(12<<2)|2,{48,67,0}},
/* 1508 */ {(12<<2)|2,{48,67,0}},
/* 1509 */ {(12<<2)|2,{48,67,0}},
/* 1510 */ {(12<<2)|2,{48,67,0}},
/* 1511 */ {(12<<2)|2,{48,67,0}},
/* 1512 */ {(12<<2)|2,{48,67,0}},
/* 1513 */ {(12<<2)|2,{48,67,0}},
/* 1514 */ {(12<<2)|2,{48,67,0}},
/* 1515 */ {(12<<2)|2,{48,67,0}},
/* 1516 */ {(12<<2)|2,{48,67,0}},
/* 1517 */ {(12<<2)|2,{48,67,0}},
/* 1518 */ {(12<<2)|2,{48,67,0}},
/* 1519 */ {(12<<2)|2,{48,67,0}},
/* 1520 */ {(12<<2)|2,{48,68,0}},
/* 1521 */ {(12<<2)|2,{48,68,0}},
/* 1522 */ {(12<<2)|2,{48,68,0}},
/* 1523 */ {(12<<2)|2,{48,68,0}},
/* 1524 */ {(12<<2)|2,{48,68,0}},
/* 1525 */ {(12<<2)|2,{48,68,0}},
/* 1526 */ {(12<<2)|2,{48,68,0}},
/* 1527 */ {(12<<2)|2,{48,68,0}},
/* 1528 */ {(12<<2)|2,{48,68,0}},
/* 1529 */ {(12<<2)|2,{48,68,0}},
/* 1530 */ {(12<<2)|2,{48,68,0}},
/* 1531 */ {(12<<2)|2,{48,68,0}},
/* 1532 */ {(12<<2)|2,{48,68,0}},
/* 1533 */ {(12<<2)|2,{48,68,0}},
/* 1534 */ {(12<<2)|2,{48,68,0}},
/* 1535 */ {(12<<2)|2,{48,68,0}},
/* 1536 */ {(12<<2)|2,{48,69,0}},
/* 1537 */ {(12<<2)|2,{48,69,0}},
/* 1538 */ {(12<<2)|2,{48,69,0}},
/* 1539 */ {(12<<2)|2,{48,69,0}},
/* 1540 */ {(12<<2)|2,{48,69,0}},
/* 1541 */ {(12<<2)|2,{48,69,0}},
/* 1542 */ {(12<<2)|2,{48,69,0}},
/* 1543 */ {(12<<2)|2,{48,69,0}},
/* 1544 */ {(12<<2)|2,{48,69,0}},
/* 1545 */ {(12<<2)|2,{48,69,0}},
/* 1546 */ {(12<<2)|2,{48,69,0}},
/* 1547 */ {(12<<2)|2,{48,69,0}},
/* 1548 */ {(12<<2)|2,{48,69,0}},
/* 1549 */ {(12<<2)|2,{48,69,0}},
/* 1550 */ {(12<<2)|2,{48,69,0}},
/* 1551 */ {(12<<2)|2,{48,69,0}},
/* 1552 */ {(12<<2)|2,{48,70,0}},
/* 1553 */ {(12<<2)|2,{48,70,0}},
/* 1554 */ {(12<<2)|2,{48,70,0}},
/* 1555 */ {(12<<2)|2,{48,70,0}},
/* 1556 */ {(12<<2)|2,{48,70,0}},
/* 1557 */ {(12<<2)|2,{48,70,0}},
/* 1558 */ {(12<<2)|2,{48,70,0}},
/* 1559 */ {(12<<2)|2,{48,70,0}},
/* 1560 */ {(12<<2)|2,{48,70,0}},
/* 1561 */ {(12<<2)|2,{48,70,0}},
/* 1562 */ {(12<<2)|2,{48,70,0}},
/* 1563 */ {(12<<2)|2,{48,70,0}},
/* 1564 */ {(12<<2)|2,{48,70,0}},
/* 1565 */ {(12<<2)|2,{48,70,0}},
/* 1566 */ {(12<<2)|2,{48,70,0}},
/* 1567 */ {(12<<2)|2,{48,70,0}},
/* 1568 */ {(12<<2)|2,{48,71,0}},
/* 1569 */ {(12<<2)|2,{48,71,0}},
/* 1570 */ {(12<<2)|2,{48,71,0}},
/* 1571 */ {(12<<2)|2,{48,71,0}},
/* 1572 */ {(12<<2)|2,{48,71,0}},
/* 1573 */ {(12<<2)|2,{48,71,0}},
/* 1574 */ {(12<<2)|2,{48,71,0}},
/* 1575 */ {(12<<2)|2,{48,71,0}},
/* 1576 */ {(12<<2)|2,{48,71,0}},
/* 1577 */ {(12<<2)|2,{48,71,0}},
/* 1578 */ {(12<<2)|2,{48,71,0}},
/* 1579 */ {(12<<2)|2,{48,71,0}},
/* 1580 */ {(12<<2)|2,{48,71,0}},
/* 1581 */ {(12<<2)|2,{48,71,0}},
/* 1582 */ {(12<<2)|2,{48,71,0}},
/* 1583 */ {(12<<2)|2,{48,71,0}},
/* 1584 */ {(12<<2)|2,{48,72,0}},
/* 1585 */ {(12<<2)|2,{48,72,0}},
/* 1586 */ {(12<<2)|2,{48,72,0}},
/* 1587 */ {(12<<2)|2,{48,72,0}},
/* 1588 */ {(12<<2)|2,{48,72,0}},
/* 1589 */ {(12<<2)|2,{48,72,0}},
/* 1590 */ {(12<<2)|2,{48,72,0}},
/* 1591 */ {(12<<2)|2,{48,72,0}},
/* 1592 */ {(12<<2)|2,{48,72,0}},
/* 1593 */ {(12<<2)|2,{48,72,0}},
/* 1594 */ {(12<<2)|2,{48,72,0}},
/* 1595 */ {(12<<2)|2,{48,72,0}},
/* 1596 */ {(12<<2)|2,{48,72,0}},
/* 1597 */ {(12<<2)|2,{48,72,0}},
/* 1598 */ {(12<<2)|2,{48,72,0}},
/* 1599 */ {(12<<2)|2,{48,72,0}},
/* 1600 */ {(12<<2)|2,{48,73,0}},
/* 1601 */ {(12<<2)|2,{48,73,0}},
/* 1602 */ {(12<<2)|2,{48,73,0}},
/* 1603 */ {(12<<2)|2,{48,73,0}},
/* 1604 */ {(12<<2)|2,{48,73,0}},
/* 1605 */ {(12<<2)|2,{48,73,0}},
/* 1606 */ {(12<<2)|2,{48,73,0}},
/* 1607 */ {(12<<2)|2,{48,73,0}},
/* 1608 */ {(12<<2)|2,{48,73,0}},
/* 1609 */ {(12<<2)|2,{48,73,0}},
/* 1610 */ {(12<<2)|2,{48,73,0}},
/* 1611 */ {(12<<2)|2,{48,73,0}},
/* 1612 */ {(12<<2)|2,{48,73,0}},
/* 1613 */ {(12<<2)|2,{48,73,0}},
/* 1614 */ {(12<<2)|2,{48,73,0}},
/* 1615 */ {(12<<2)|2,{48,73,0}},
/* 1616 */ {(12<<2)|2,{48,74,0}},
/* 1617 */ {(12<<2)|2,{48,74,0}},
/* 1618 */ {(12<<2)|2,{48,74,0}},
/* 1619 */ {(12<<2)|2,{48,74,0}},
/* 1620 */ {(12<<2)|2,{48,74,0}},
/* 1621 */ {(12<<2)|2,{48,74,0}},
/* 1622 */ {(12<<2)|2,{48,74,0}},
/* 1623 */ {(12<<2)|2,{48,74,0}},
/* 1624 */ {(12<<2)|2,{48,74,0}},
/* 1625 */ {(12<<2)|2,{48,74,0}},
/* 1626 */ {(12<<2)|2,{48,74,0}},
/* 1627 */ {(12<<2)|2,{48,74,0}},
/* 1628 */ {(12<<2)|2,{48,74,0}},
/* 1629 */ {(12<<2)|2,{48,74,0}},
/* 1630 */ {(12<<2)|2,{48,74,0}},
/* 1631 */ {(12<<2)|2,{48,74,0}},
/* 1632 */ {(12<<2)|2,{48,75,0}},
/* 1633 */ {(12<<2)|2,{48,75,0}},
/* 1634 */ {(12<<2)|2,{48,75,0}},
/* 1635 */ {(12<<2)|2,{48,75,0}},
/* 1636 */ {(12<<2)|2,{48,75,0}},
/* 1637 */ {(12<<2)|2,{48,75,0}},
/* 1638 */ {(12<<2)|2,{48,75,0}},
/* 1639 */ {(12<<2)|2,{48,75,0}},
/* 1640 */ {(12<<2)|2,{48,75,0}},
/* 1641 */ {(12<<2)|2,{48,75,0}},
/* 1642 */ {(12<<2)|2,{48,75,0}},
/* 1643 */ {(12<<2)|2,{48,75,0}},
/* 1644 */ {(12<<2)|2,{48,75,0}},
/* 1645 */ {(12<<2)|2,{48,75,0}},
/* 1646 */ {(12<<2)|2,{48,75,0}},
/* 1647 */ {(12<<2)|2,{48,75,0}},
/* 1648 */ {(12<<2)|2,{48,76,0}},
/* 1649 */ {(12<<2)|2,{48,76,0}},
/* 1650 */ {(12<<2)|2,{48,76,0}},
/* 1651 */ {(12<<2)|2,{48,76,0}},
/* 1652 */ {(12<<2)|2,{48,76,0}},
/* 1653 */ {(12<<2)|2,{48,76,0}},
/* 1654 */ {(12<<2)|2,{48,76,0}},
/* 1655 */ {(12<<2)|2,{48,76,0}},
/* 1656 */ {(12<<2)|2,{48,76,0}},
/* 1657 */ {(12<<2)|2,{48,76,0}},
/* 1658 */ {(12<<2)|2,{48,76,0}},
/* 1659 */ {(12<<2)|2,{48,76,0}},
/* 1660 */ {(12<<2)|2,{48,76,0}},
/* 1661 */ {(12<<2)|2,{48,76,0}},
/* 1662 */ {(12<<2)|2,{48,76,0}},
/* 1663 */ {(12<<2)|2,{48,76,0}},
/* 1664 */ {(12<<2)|2,{48,77,0}},
/* 1665 */ {(12<<2)|2,{48,77,0}},
/* 1666 */ {(12<<2)|2,{48,77,0}},
/* 1667 */ {(12<<2)|2,{48,77,0}},
/* 1668 */ {(12<<2)|2,{48,77,0}},
/* 1669 */ {(12<<2)|2,{48,77,0}},
/* 1670 */ {(12<<2)|2,{48,77,0}},
/* 1671 */ {(12<<2)|2,{48,77,0}},
/* 1672 */ {(12<<2)|2,{48,77,0}},
/* 1673 */ {(12<<2)|2,{48,77,0}},
/* 1674 */ {(12<<2)|2,{48,77,0}},
/* 1675 */ {(12<<2)|2,{48,77,0}},
/* 1676 */ {(12<<2)|2,{48,77,0}},
/* 1677 */ {(12<<2)|2,{48,77,0}},
/* 1678 */ {(12<<2)|2,{48,77,0}},
/* 1679 */ {(12<<2)|2,{48,77,0}},
/* 1680 */ {(12<<2)|2,{48,78,0}},
/* 1681 */ {(12<<2)|2,{48,78,0}},
/* 1682 */ {(12<<2)|2,{48,78,0}},
/* 1683 */ {(12<<2)|2,{48,78,0}},
/* 1684 */ {(12<<2)|2,{48,78,0}},
/* 1685 */ {(12<<2)|2,{48,78,0}},
/* 1686 */ {(12<<2)|2,{48,78,0}},
/* 1687 */ {(12<<2)|2,{48,78,0}},
/* 1688 */ {(12<<2)|2,{48,78,0}},
/* 1689 */ {(12<<2)|2,{48,78,0}},
/* 1690 */ {(12<<2)|2,{48,78,0}},
/* 1691 */ {(12<<2)|2,{48,78,0}},
/* 1692 */ {(12<<2)|2,{48,78,0}},
/* 1693 */ {(12<<2)|2,{48,78,0}},
/* 1694 */ {(12<<2)|2,{48,78,0}},
/* 1695 */ {(12<<2)|2,{48,78,0}},
/* 1696 */ {(12<<2)|2,{48,79,0}},
/* 1697 */ {(12<<2)|2,{48,79,0}},
/* 1698 */ {(12<<2)|2,{48,79,0}},
/* 1699 */ {(12<<2)|2,{48,79,0}},
/* 1700 */ {(12<<2)|2,{48,79,0}},
/* 1701 */ {(12<<2)|2,{48,79,0}},
/* 1702 */ {(12<<2)|2,{48,79,0}},
/* 1703 */ {(12<<2)|2,{48,79,0}},
/* 1704 */ {(12<<2)|2,{48,79,0}},
/* 1705 */ {(12<<2)|2,{48,79,0}},
/* 1706 */ {(12<<2)|2,{48,79,0}},
/* 1707 */ {(12<<2)|2,{48,79,0}},
/* 1708 */ {(12<<2)|2,{48,79,0}},
/* 1709 */ {(12<<2)|2,{48,79,0}},
/* 1710 */ {(12<<2)|2,{48,79,0}},
/* 1711 */ {(12<<2)|2,{48,79,0}},
/* 1712 */ {(12<<2)|2,{48,80,0}},
/* 1713 */ {(12<<2)|2,{48,80,0}},
/* 1714 */ {(12<<2)|2,{48,80,0}},
/* 1715 */ {(12<<2)|2,{48,80,0}},
/* 1716 */ {(12<<2)|2,{48,80,0}},
/* 1717 */ {(12<<2)|2,{48,80,0}},
/* 1718 */ {(12<<2)|2,{48,80,0}},
/* 1719 */ {(12<<2)|2,{48,80,0}},
/* 1720 */ {(12<<2)|2,{48,80,0}},
/* 1721 */ {(12<<2)|2,{48,80,0}},
/* 1722 */ {(12<<2)|2,{48,80,0}},
/* 1723 */ {(12<<2)|2,{48,80,0}},
/* 1724 */ {(12<<2)|2,{48,80,0}},
/* 1725 */ {(12<<2)|2,{48,80,0}},
/* 1726 */ {(12<<2)|2,{48,80,0}},
/* 1727 */ {(12<<2)|2,{48,80,0}},
/* 1728 */ {(12<<2)|2,{48,81,0}},
/* 1729 */ {(12<<2)|2,{48,81,0}},
/* 1730 */ {(12<<2)|2,{48,81,0}},
/* 1731 */ {(12<<2)|2,{48,81,0}},
/* 1732 */ {(12<<2)|2,{48,81,0}},
/* 1733 */ {(12<<2)|2,{48,81,0}},
/* 1734 */ {(12<<2)|2,{48,81,0}},
/* 1735 */ {(12<<2)|2,{48,81,0}},
/* 1736 */ {(12<<2)|2,{48,81,0}},
/* 1737 */ {(12<<2)|2,{48,81,0}},
/* 1738 */ {(12<<2)|2,{48,81,0}},
/* 1739 */ {(12<<2)|2,{48,81,0}},
/* 1740 */ {(12<<2)|2,{48,81,0}},
/* 1741 */ {(12<<2)|2,{48,81,0}},
/* 1742 */ {(12<<2)|2,{48,81,0}},
/* 1743 */ {(12<<2)|2,{48,81,0}},
/* 1744 */ {(12<<2)|2,{48,82,0}},
/* 1745 */ {(12<<2)|2,{48,82,0}},
/* 1746 */ {(12<<2)|2,{48,82,0}},
/* 1747 */ {(12<<2)|2,{48,82,0}},
/* 1748 */ {(12<<2)|2,{48,82,0}},
/* 1749 */ {(12<<2)|2,{48,82,0}},
/* 1750 */ {(12<<2)|2,{48,82,0}},
/* 1751 */ {(12<<2)|2,{48,82,0}},
/* 1752 */ {(12<<2)|2,{48,82,0}},
/* 1753 */ {(12<<2)|2,{48,82,0}},
/* 1754 */ {(12<<2)|2,{48,82,0}},
/* 1755 */ {(12<<2)|2,{48,82,0}},
/* 1756 */ {(12<<2)|2,{48,82,0}},
/* 1757 */ {(12<<2)|2,{48,82,0}},
/* 1758 */ {(12<<2)|2,{48,82,0}},
/* 1759 */ {(12<<2)|2,{48,82,0}},
/* 1760 */ {(12<<2)|2,{48,83,0}},
/* 1761 */ {(12<<2)|2,{48,83,0}},
/* 1762 */ {(12<<2)|2,{48,83,0}},
/* 1763 */ {(12<<2)|2,{48,83,0}},
/* 1764 */ {(12<<2)|2,{48,83,0}},
/* 1765 */ {(12<<2)|2,{48,83,0}},
/* 1766 */ {(12<<2)|2,{48,83,0}},
/* 1767 */ {(12<<2)|2,{48,83,0}},
/* 1768 */ {(12<<2)|2,{48,83,0}},
/* 1769 */ {(12<<2)|2,{48,83,0}},
/* 1770 */ {(12<<2)|2,{48,83,0}},
/* 1771 */ {(12<<2)|2,{48,83,0}},
/* 1772 */ {(12<<2)|2,{48,83,0}},
/* 1773 */ {(12<<2)|2,{48,83,0}},
/* 1774 */ {(12<<2)|2,{48,83,0}},
/* 1775 */ {(12<<2)|2,{48,83,0}},
/* 1776 */ {(12<<2)|2,{48,84,0}},
/* 1777 */ {(12<<2)|2,{48,84,0}},
/* 1778 */ {(12<<2)|2,{48,84,0}},
/* 1779 */ {(12<<2)|2,{48,84,0}},
/* 1780 */ {(12<<2)|2,{48,84,0}},
/* 1781 */ {(12<<2)|2,{48,84,0}},
/* 1782 */ {(12<<2)|2,{48,84,0}},
/* 1783 */ {(12<<2)|2,{48,84,0}},
/* 1784 */ {(12<<2)|2,{48,84,0}},
/* 1785 */ {(12<<2)|2,{48,84,0}},
/* 1786 */ {(12<<2)|2,{48,84,0}},
/* 1787 */ {(12<<2)|2,{48,84,0}},
/* 1788 */ {(12<<2)|2,{48,84,0}},
/* 1789 */ {(12<<2)|2,{48,84,0}},
/* 1790 */ {(12<<2)|2,{48,84,0}},
/* 1791 */ {(12<<2)|2,{48,84,0}},
/* 1792 */ {(12<<2)|2,{48,85,0}},
/* 1793 */ {(12<<2)|2,{48,85,0}},
/* 1794 */ {(12<<2)|2,{48,85,0}},
/* 1795 */ {(12<<2)|2,{48,85,0}},
/* 1796 */ {(12<<2)|2,{48,85,0}},
/* 1797 */ {(12<<2)|2,{48,85,0}},
/* 1798 */ {(12<<2)|2,{48,85,0}},
/* 1799 */ {(12<<2)|2,{48,85,0}},
/* 1800 */ {(12<<2)|2,{48,85,0}},
/* 1801 */ {(12<<2)|2,{48,85,0}},
/* 1802 */ {(12<<2)|2,{48,85,0}},
/* 1803 */ {(12<<2)|2,{48,85,0}},
/* 1804 */ {(12<<2)|2,{48,85,0}},
/* 1805 */ {(12<<2)|2,{48,85,0}},
/* 1806 */ {(12<<2)|2,{48,85,0}},
/* 1807 */ {(12<<2)|2,{48,85,0}},
/* 1808 */ {(12<<2)|2,{48,86,0}},
/* 1809 */ {(12<<2)|2,{48,86,0}},
/* 1810 */ {(12<<2)|2,{48,86,0}},
/* 1811 */ {(12<<2)|2,{48,86,0}},
/* 1812 */ {(12<<2)|2,{48,86,0}},
/* 1813 */ {(12<<2)|2,{48,86,0}},
/* 1814 */ {(12<<2)|2,{48,86,0}},
/* 1815 */ {(12<<2)|2,{48,86,0}},
/* 1816 */ {(12<<2)|2,{48,86,0}},
/* 1817 */ {(12<<2)|2,{48,86,0}},
/* 1818 */ {(12<<2)|2,{48,86,0}},
/* 1819 */ {(12<<2)|2,{48,86,0}},
/* 1820 */ {(12<<2)|2,{48,86,0}},
/* 1821 */ {(12<<2)|2,{48,86,0}},
/* 1822 */ {(12<<2)|2,{48,86,0}},
/* 1823 */ {(12<<2)|2,{48,86,0}},
/* 1824 */ {(12<<2)|2,{48,87,0}},
/* 1825 */ {(12<<2)|2,{48,87,0}},
/* 1826 */ {(12<<2)|2,{48,87,0}},
/* 1827 */ {(12<<2)|2,{48,87,0}},
/* 1828 */ {(12<<2)|2,{48,87,0}},
/* 1829 */ {(12<<2)|2,{48,87,0}},
/* 1830 */ {(12<<2)|2,{48,87,0}},
/* 1831 */ {(12<<2)|2,{48,87,0}},
/* 1832 */ {(12<<2)|2,{48,87,0}},
/* 1833 */ {(12<<2)|2,{48,87,0}},
/* 1834 */ {(12<<2)|2,{48,87,0}},
/* 1835 */ {(12<<2)|2,{48,87,0}},
/* 1836 */ {(12<<2)|2,{48,87,0}},
/* 1837 */ {(12<<2)|2,{48,87,0}},
/* 1838 */ {(12<<2)|2,{48,87,0}},
/* 1839 */ {(12<<2)|2,{48,87,0}},
/* 1840 */ {(12<<2)|2,{48,89,0}},
/* 1841 */ {(12<<2)|2,{48,89,0}},
/* 1842 */ {(12<<2)|2,{48,89,0}},
/* 1843 */ {(12<<2)|2,{48,89,0}},
/* 1844 */ {(12<<2)|2,{48,89,0}},
/* 1845 */ {(12<<2)|2,{48,89,0}},
/* 1846 */ {(12<<2)|2,{48,89,0}},
/* 1847 */ {(12<<2)|2,{48,89,0}},
/* 1848 */ {(12<<2)|2,{48,89,0}},
/* 1849 */ {(12<<2)|2,{48,89,0}},
/* 1850 */ {(12<<2)|2,{48,89,0}},
/* 1851 */ {(12<<2)|2,{48,89,0}},
/* 1852 */ {(12<<2)|2,{48,89,0}},
/* 1853 */ {(12<<2)|2,{48,89,0}},
/* 1854 */ {(12<<2)|2,{48,89,0}},
/* 1855 */ {(12<<2)|2,{48,89,0}},
/* 1856 */ {(12<<2)|2,{48,106,0}},
/* 1857 */ {(12<<2)|2,{48,106,0}},
/* 1858 */ {(12<<2)|2,{48,106,0}},
/* 1859 */ {(12<<2)|2,{48,106,0}},
/* 1860 */ {(12<<2)|2,{48,106,0}},
/* 1861 */ {(12<<2)|2,{48,106,0}},
/* 1862 */ {(12<<2)|2,{48,106,0}},
/* 1863 */ {(12<<2)|2,{48,106,0}},
/* 1864 */ {(12<<2)|2,{48,106,0}},
/* 1865 */ {(12<<2)|2,{48,106,0}},
/* 1866 */ {(12<<2)|2,{48,106,0}},
/* 1867 */ {(12<<2)|2,{48,106,0}},
/* 1868 */ {(12<<2)|2,{48,106,0}},
/* 1869 */ {(12<<2)|2,{48,106,0}},
/* 1870 */ {(12<<2)|2,{48,106,0}},
/* 1871 */ {(12<<2)|2,{48,106,0}},
/* 1872 */ {(12<<2)|2,{48,107,0}},
/* 1873 */ {(12<<2)|2,{48,107,0}},
/* 1874 */ {(12<<2)|2,{48,107,0}},
/* 1875 */ {(12<<2)|2,{48,107,0}},
/* 1876 */ {(12<<2)|2,{48,107,0}},
/* 1877 */ {(12<<2)|2,{48,107,0}},
/* 1878 */ {(12<<2)|2,{48,107,0}},
/* 1879 */ {(12<<2)|2,{48,107,0}},
/* 1880 */ {(12<<2)|2,{48,107,0}},
/* 1881 */ {(12<<2)|2,{48,107,0}},
/* 1882 */ {(12<<2)|2,{48,107,0}},
/* 1883 */ {(12<<2)|2,{48,107,0}},
/* 1884 */ {(12<<2)|2,{48,107,0}},
/* 1885 */ {(12<<2)|2,{48,107,0}},
/* 1886 */ {(12<<2)|2,{48,107,0}},
/* 1887 */ {(12<<2)|2,{48,107,0}},
/* 1888 */ {(12<<2)|2,{48,113,0}},
/* 1889 */ {(12<<2)|2,{48,113,0}},
/* 1890 */ {(12<<2)|2,{48,113,0}},
/* 1891 */ {(12<<2)|2,{48,113,0}},
/* 1892 */ {(12<<2)|2,{48,113,0}},
/* 1893 */ {(12<<2)|2,{48,113,0}},
/* 1894 */ {(12<<2)|2,{48,113,0}},
/* 1895 */ {(12<<2)|2,{48,113,0}},
/* 1896 */ {(12<<2)|2,{48,113,0}},
/* 1897 */ {(12<<2)|2,{48,113,0}},
/* 1898 */ {(12<<2)|2,{48,113,0}},
/* 1899 */ {(12<<2)|2,{48,113,0}},
/* 1900 */ {(12<<2)|2,{48,113,0}},
/* 1901 */ {(12<<2)|2,{48,113,0}},
/* 1902 */ {(12<<2)|2,{48,113,0}},
/* 1903 */ {(12<<2)|2,{48,113,0}},
/* 1904 */ {(12<<2)|2,{48,118,0}},
/* 1905 */ {(12<<2)|2,{48,118,0}},
/* 1906 */ {(12<<2)|2,{48,118,0}},
/* 1907 */ {(12<<2)|2,{48,118,0}},
/* 1908 */ {(12<<2)|2,{48,118,0}},
/* 1909 */ {(12<<2)|2,{48,118,0}},
/* 1910 */ {(12<<2)|2,{48,118,0}},
/* 1911 */ {(12<<2)|2,{48,118,0}},
/* 1912 */ {(12<<2)|2,{48,118,0}},
/* 1913 */ {(12<<2)|2,{48,118,0}},
/* 1914 */ {(12<<2)|2,{48,118,0}},
/* 1915 */ {(12<<2)|2,{48,118,0}},
/* 1916 */ {(12<<2)|2,{48,118,0}},
/* 1917 */ {(12<<2)|2,{48,118,0}},
/* 1918 */ {(12<<2)|2,{48,118,0}},
/* 1919 */ {(12<<2)|2,{48,118,0}},
/* 1920 */ {(12<<2)|2,{48,119,0}},
/* 1921 */ {(12<<2)|2,{48,119,0}},
/* 1922 */ {(12<<2)|2,{48,119,0}},
/* 1923 */ {(12<<2)|2,{48,119,0}},
/* 1924 */ {(12<<2)|2,{48,119,0}},
/* 1925 */ {(12<<2)|2,{48,119,0}},
/* 1926 */ {(12<<2)|2,{48,119,0}},
/* 1927 */ {(12<<2)|2,{48,119,0}},
/* 1928 */ {(12<<2)|2,{48,119,0}},
/* 1929 */ {(12<<2)|2,{48,119,0}},
/* 1930 */ {(12<<2)|2,{48,119,0}},
/* 1931 */ {(12<<2)|2,{48,119,0}},
/* 1932 */ {(12<<2)|2,{48,119,0}},
/* 1933 */ {(12<<2)|2,{48,119,0}},
/* 1934 */ {(12<<2)|2,{48,119,0}},
/* 1935 */ {(12<<2)|2,{48,119,0}},
/* 1936 */ {(12<<2)|2,{48,120,0}},
/* 1937 */ {(12<<2)|2,{48,120,0}},
/* 1938 */ {(12<<2)|2,{48,120,0}},
/* 1939 */ {(12<<2)|2,{48,120,0}},
/* 1940 */ {(12<<2)|2,{48,120,0}},
/* 1941 */ {(12<<2)|2,{48,120,0}},
/* 1942 */ {(12<<2)|2,{48,120,0}},
/* 1943 */ {(12<<2)|2,{48,120,0}},
/* 1944 */ {(12<<2)|2,{48,120,0}},
/* 1945 */ {(12<<2)|2,{48,120,0}},
/* 1946 */ {(12<<2)|2,{48,120,0}},
/* 1947 */ {(12<<2)|2,{48,120,0}},
/* 1948 */ {(12<<2)|2,{48,120,0}},
/* 1949 */ {(12<<2)|2,{48,120,0}},
/* 1950 */ {(12<<2)|2,{48,120,0}},
/* 1951 */ {(12<<2)|2,{48,120,0}},
/* 1952 */ {(12<<2)|2,{48,121,0}},
/* 1953 */ {(12<<2)|2,{48,121,0}},
/* 1954 */ {(12<<2)|2,{48,121,0}},
/* 1955 */ {(12<<2)|2,{48,121,0}},
/* 1956 */ {(12<<2)|2,{48,121,0}},
/* 1957 */ {(12<<2)|2,{48,121,0}},
/* 1958 */ {(12<<2)|2,{48,121,0}},
/* 1959 */ {(12<<2)|2,{48,121,0}},
/* 1960 */ {(12<<2)|2,{48,121,0}},
/* 1961 */ {(12<<2)|2,{48,121,0}},
/* 1962 */ {(12<<2)|2,{48,121,0}},
/* 1963 */ {(12<<2)|2,{48,121,0}},
/* 1964 */ {(12<<2)|2,{48,121,0}},
/* 1965 */ {(12<<2)|2,{48,121,0}},
/* 1966 */ {(12<<2)|2,{48,121,0}},
/* 1967 */ {(12<<2)|2,{48,121,0}},
/* 1968 */ {(12<<2)|2,{48,122,0}},
/* 1969 */ {(12<<2)|2,{48,122,0}},
/* 1970 */ {(12<<2)|2,{48,122,0}},
/* 1971 */ {(12<<2)|2,{48,122,0}},
/* 1972 */ {(12<<2)|2,{48,122,0}},
/* 1973 */ {(12<<2)|2,{48,122,0}},
/* 1974 */ {(12<<2)|2,{48,122,0}},
/* 1975 */ {(12<<2)|2,{48,122,0}},
/* 1976 */ {(12<<2)|2,{48,122,0}},
/* 1977 */ {(12<<2)|2,{48,122,0}},
/* 1978 */ {(12<<2)|2,{48,122,0}},
/* 1979 */ {(12<<2)|2,{48,122,0}},
/* 1980 */ {(12<<2)|2,{48,122,0}},
/* 1981 */ {(12<<2)|2,{48,122,0}},
/* 1982 */ {(12<<2)|2,{48,122,0}},
/* 1983 */ {(12<<2)|2,{48,122,0}},
/* 1984 */ {(13<<2)|2,{48,38,0}},
/* 1985 */ {(13<<2)|2,{48,38,0}},
/* 1986 */ {(13<<2)|2,{48,38,0}},
/* 1987 */ {(13<<2)|2,{48,38,0}},
/* 1988 */ {(13<<2)|2,{48,38,0}},
/* 1989 */ {(13<<2)|2,{48,38,0}},
/* 1990 */ {(13<<2)|2,{48,38,0}},
/* 1991 */ {(13<<2)|2,{48,38,0}},
/* 1992 */ {(13<<2)|2,{48,42,0}},
/* 1993 */ {(13<<2)|2,{48,42,0}},
/* 1994 */ {(13<<2)|2,{48,42,0}},
/* 1995 */ {(13<<2)|2,{48,42,0}},
/* 1996 */ {(13<<2)|2,{48,42,0}},
/* 1997 */ {(13<<2)|2,{48,42,0}},
/* 1998 */ {(13<<2)|2,{48,42,0}},
/* 1999 */ {(13<<2)|2,{48,42,0}},
/* 2000 */ {(13<<2)|2,{48,44,0}},
/* 2001 */ {(13<<2)|2,{48,44,0}},
/* 2002 */ {(13<<2)|2,{48,44,0}},
/* 2003 */ {(13<<2)|2,{48,44,0}},
/* 2004 */ {(13<<2)|2,{48,44,0}},
/* 2005 */ {(13<<2)|2,{48,44,0}},
/* 2006 */ {(13<<2)|2,{48,44,0}},
/* 2007 */ {(13<<2)|2,{48,44,0}},
/* 2008 */ {(13<<2)|2,{48,59,0}},
/* 2009 */ {(13<<2)|2,{48,59,0}},
/* 2010 */ {(13<<2)|2,{48,59,0}},
/* 2011 */ {(13<<2)|2,{48,59,0}},
/* 2012 */ {(13<<2)|2,{48,59,0}},
/* 2013 */ {(13<<2)|2,{48,59,0}},
/* 2014 */ {(13<<2)|2,{48,59,0}},
/* 2015 */ {(13<<2)|2,{48,59,0}},
/* 2016 */ {(13<<2)|2,{48,88,0}},
/* 2017 */ {(13<<2)|2,{48,88,0}},
/* 2018 */ {(13<<2)|2,{48,88,0}},
/* 2019 */ {(13<<2)|2,{48,88,0}},
/* 2020 */ {(13<<2)|2,{48,88,0}},
/* 2021 */ {(13<<2)|2,{48,88,0}},
/* 2022 */ {(13<<2)|2,{48,88,0}},
/* 2023 */ {(13<<2)|2,{48,88,0}},
/* 2024 */ {(13<<2)|2,{48,90,0}},
/* 2025 */ {(13<<2)|2,{48,90,0}},
/* 2026 */ {(13<<2)|2,{48,90,0}},
/* 2027 */ {(13<<2)|2,{48,90,0}},
/* 2028 */ {(13<<2)|2,{48,90,0}},
/* 2029 */ {(13<<2)|2,{48,90,0}},
/* 2030 */ {(13<<2)|2,{48,90,0}},
/* 2031 */ {(13<<2)|2,{48,90,0}},
/* 2032 */ {(15<<2)|2,{48,33,0}},
/* 2033 */ {(15<<2)|2,{48,33,0}},
/* 2034 */ {(15<<2)|2,{48,34,0}},
/* 2035 */ {(15<<2)|2,{48,34,0}},
/* 2036 */ {(15<<2)|2,{48,40,0}},
/* 2037 */ {(15<<2)|2,{48,40,0}},
/* 2038 */ {(15<<2)|2,{48,41,0}},
/* 2039 */ {(15<<2)|2,{48,41,0}},
/* 2040 */ {(15<<2)|2,{48,63,0}},
/* 2041 */ {(15<<2)|2,{48,63,0}},
/* 2042 */ {(16<<2)|2,{48,39,0}},
/* 2043 */ {(16<<2)|2,{48,43,0}},
/* 2044 */ {(16<<2)|2,{48,124,0}},
/* 2045 */ {(5<<2)|1,{48,0,0}},
/* 2046 */ {(5<<2)|1,{48,0,0}},
/* 2047 */ {(5<<2)|1,{48,0,0}},
/* 2048 */ {(15<<2)|3,{49,48,48}},
/* 2049 */ {(15<<2)|3,{49,48,48}},
/* 2050 */ {(15<<2)|3,{49,48,49}},
/* 2051 */ {(15<<2)|3,{49,48,49}},
/* 2052 */ {(15<<2)|3,{49,48,50}},
/* 2053 */ {(15<<2)|3,{49,48,50}},
/* 2054 */ {(15<<2)|3,{49,48,97}},
/* 2055 */ {(15<<2)|3,{49,48,97}},
/* 2056 */ {(15<<2)|3,{49,48,99}},
/* 2057 */ {(15<<2)|3,{49,48,99}},
/* 2058 */ {(15<<2)|3,{49,48,101}},
/* 2059 */ {(15<<2)|3,{49,48,101}},
/* 2060 */ {(15<<2)|3,{49,48,105}},
/* 2061 */ {(15<<2)|3,{49,48,105}},
/* 2062 */ {(15<<2)|3,{49,48,111}},
/* 2063 */ {(15<<2)|3,{49,48,111}},
/* 2064 */ {(15<<2)|3,{49,48,115}},
/* 2065 */ {(15<<2)|3,{49,48,115}},
/* 2066 */ {(15<<2)|3,{49,48,116}},
/* 2067 */ {(15<<2)|3,{49,48,116}},
/* 2068 */ {(16<<2)|3,{49,48,32}},
/* 2069 */ {(16<<2)|3,{49,48,37}},
/* 2070 */ {(16<<2)|3,{49,48,45}},
/* 2071 */ {(16<<2)|3,{49,48,46}},
/* 2072 */ {(16<<2)|3,{49,48,47}},
/* 2073 */ {(16<<2)|3,{49,48,51}},
/* 2074 */ {(16<<2)|3,{49,48,52}},
/* 2075 */ {(16<<2)|3,{49,48,53}},
/* 2076 */ {(16<<2)|3,{49,48,54}},
/* 2077 */ {(16<<2)|3,{49,48,55}},
/* 2078 */ {(16<<2)|3,{49,48,56}},
/* 2079 */ {(16<<2)|3,{49,48,57}},
/* 2080 */ {(16<<2)|3,{49,48,61}},
/* 2081 */ {(16<<2)|3,{49,48,65}},
/* 2082 */ {(16<<2)|3,{49,48,95}},
/* 2083 */ {(16<<2)|3,{49,48,98}},
/* 2084 */ {(16<<2)|3,{49,48,100}},
/* 2085 */ {(16<<2)|3,{49,48,102}},
/* 2086 */ {(16<<2)|3,{49,48,103}},
/* 2087 */ {(16<<2)|3,{49,48,104}},
/* 2088 */ {(16<<2)|3,{49,48,108}},
/* 2089 */ {(16<<2)|3,{49,48,109}},
/* 2090 */ {(16<<2)|3,{49,48,110}},
/* 2091 */ {(16<<2)|3,{49,48,112}},
/* 2092 */ {(16<<2)|3,{49,48,114}},
/* 2093 */ {(16<<2)|3,{49,48,117}},
/* 2094 */ {(10<<2)|2,{49,48,0}},
/* 2095 */ {(10<<2)|2,{49,48,0}},
/* 2096 */ {(10<<2)|2,{49,48,0}},
/* 2097 */ {(10<<2)|2,{49,48,0}},
/* 2098 */ {(10<<2)|2,{49,48,0}},
/* 2099 */ {(10<<2)|2,{49,48,0}},
/* 2100 */ {(10<<2)|2,{49,48,0}},
/* 2101 */ {(10<<2)|2,{49,48,0}},
/* 2102 */ {(10<<2)|2,{49,48,0}},
/* 2103 */ {(10<<2)|2,{49,48,0}},
/* 2104 */ {(10<<2)|2,{49,48,0}},
/* 2105 */ {(10<<2)|2,{49,48,0}},
/* 2106 */ {(10<<2)|2,{49,48,0}},
/* 2107 */ {(10<<2)|2,{49,48,0}},
/* 2108 */ {(10<<2)|2,{49,48,0}},
/* 2109 */ {(10<<2)|2,{49,48,0}},
/* 2110 */ {(10<<2)|2,{49,48,0}},
/* 2111 */ {(10<<2)|2,{49,48,0}},
/* 2112 */ {(15<<2)|3,{49,49,48}},
/* 2113 */ {(15<<2)|3,{49,49,48}},
/* 2114 */ {(15<<2)|3,{49,49,49}},
/* 2115 */ {(15<<2)|3,{49,49,49}},
/* 2116 */ {(15<<2)|3,{49,49,50}},
/* 2117 */ {(15<<2)|3,{49,49,50}},
/* 2118 */ {(15<<2)|3,{49,49,97}},
/* 2119 */ {(15<<2)|3,{49,49,97}},
/* 2120 */ {(15<<2)|3,{49,49,99}},
/* 2121 */ {(15<<2)|3,{49,49,99}},
/* 2122 */ {(15<<2)|3,{49,49,101}},
/* 2123 */ {(15<<2)|3,{49,49,101}},
/* 2124 */ {(15<<2)|3,{49,49,105}},
/* 2125 */ {(15<<2)|3,{49,49,105}},
/* 2126 */ {(15<<2)|3,{49,49,111}},
/* 2127 */ {(15<<2)|3,{49,49,111}},
/* 2128 */ {(15<<2)|3,{49,49,115}},
/* 2129 */ {(15<<2)|3,{49,49,115}},
/* 2130 */ {(15<<2)|3,{49,49,116}},
/* 2131 */ {(15<<2)|3,{49,49,116}},
/* 2132 */ {(16<<2)|3,{49,49,32}},
/* 2133 */ {(16<<2)|3,{49,49,37}},
/* 2134 */ {(16<<2)|3,{49,49,45}},
/* 2135 */ {(16<<2)|3,{49,49,46}},
/* 2136 */ {(16<<2)|3,{49,49,47}},
/* 2137 */ {(16<<2)|3,{49,49,51}},
/* 2138 */ {(16<<2)|3,{49,49,52}},
/* 2139 */ {(16<<2)|3,{49,49,53}},
/* 2140 */ {(16<<2)|3,{49,49,54}},
/* 2141 */ {(16<<2)|3,{49,49,55}},
/* 2142 */ {(16<<2)|3,{49,49,56}},
/* 2143 */ {(16<<2)|3,{49,49,57}},
/* 2144 */ {(16<<2)|3,{49,49,61}},
/* 2145 */ {(16<<2)|3,{49,49,65}},
/* 2146 */ {(16<<2)|3,{49,49,95}},
/* 2147 */ {(16<<2)|3,{49,49,98}},
/* 2148 */ {(16<<2)|3,{49,49,100}},
/* 2149 */ {(16<<2)|3,{49,49,102}},
/* 2150 */ {(16<<2)|3,{49,49,103}},
/* 2151 */ {(16<<2)|3,{49,49,104}},
/* 2152 */ {(16<<2)|3,{49,49,108}},
/* 2153 */ {(16<<2)|3,{49,49,109}},
/* 2154 */ {(16<<2)|3,{49,49,110}},
/* 2155 */ {(16<<2)|3,{49,49,112}},
/* 2156 */ {(16<<2)|3,{49,49,114}},
/* 2157 */ {(16<<2)|3,{49,49,117}},
/* 2158 */ {(10<<2)|2,{49,49,0}},
/* 2159 */ {(10<<2)|2,{49,49,0}},
/* 2160 */ {(10<<2)|2,{49,49,0}},
/* 2161 */ {(10<<2)|2,{49,49,0}},
/* 2162 */ {(10<<2)|2,{49,49,0}},
/* 2163 */ {(10<<2)|2,{49,49,0}},
/* 2164 */ {(10<<2)|2,{49,49,0}},
/* 2165 */ {(10<<2)|2,{49,49,0}},
/* 2166 */ {(10<<2)|2,{49,49,0}},
/* 2167 */ {(10<<2)|2,{49,49,0}},
/* 2168 */ {(10<<2)|2,{49,49,0}},
/* 2169 */ {(10<<2)|2,{49,49,0}},
/* 2170 */ {(10<<2)|2,{49,49,0}},
/* 2171 */ {(10<<2)|2,{49,49,0}},
/* 2172 */ {(10<<2)|2,{49,49,0}},
/* 2173 */ {(10<<2)|2,{49,49,0}},
/* 2174 */ {(10<<2)|2,{49,49,0}},
/* 2175 */ {(10<<2)|2,{49,49,0}},
/* 2176 */ {(15<<2)|3,{49,50,48}},
/* 2177 */ {(15<<2)|3,{49,50,48}},
/* 2178 */ {(15<<2)|3,{49,50,49}},
/* 2179 */ {(15<<2)|3,{49,50,49}},
/* 2180 */ {(15<<2)|3,{49,50,50}},
/* 2181 */ {(15<<2)|3,{49,50,50}},
/* 2182 */ {(15<<2)|3,{49,50,97}},
/* 2183 */ {(15<<2)|3,{49,50,97}},
/* 2184 */ {(15<<2)|3,{49,50,99}},
/* 2185 */ {(15<<2)|3,{49,50,99}},
/* 2186 */ {(15<<2)|3,{49,50,101}},
/* 2187 */ {(15<<2)|3,{49,50,101}},
/* 2188 */ {(15<<2)|3,{49,50,105}},
/* 2189 */ {(15<<2)|3,{49,50,105}},
/* 2190 */ {(15<<2)|3,{49,50,111}},
/* 2191 */ {(15<<2)|3,{49,50,111}},
/* 2192 */ {(15<<2)|3,{49,50,115}},
/* 2193 */ {(15<<2)|3,{49,50,115}},
/* 2194 */ {(15<<2)|3,{49,50,116}},
/* 2195 */ {(15<<2)|3,{49,50,116}},
/* 2196 */ {(16<<2)|3,{49,50,32}},
/* 2197 */ {(16<<2)|3,{49,50,37}},
/* 2198 */ {(16<<2)|3,{49,50,45}},
/* 2199 */ {(16<<2)|3,{49,50,46}},
/* 2200 */ {(16<<2)|3,{49,50,47}},
/* 2201 */ {(16<<2)|3,{49,50,51}},
/* 2202 */ {(16<<2)|3,{49,50,52}},
/* 2203 */ {(16<<2)|3,{49,50,53}},
/* 2204 */ {(16<<2)|3,{49,50,54}},
/* 2205 */ {(16<<2)|3,{49,50,55}},
/* 2206 */ {(16<<2)|3,{49,50,56}},
/* 2207 */ {(16<<2)|3,{49,50,57}},
/* 2208 */ {(16<<2)|3,{49,50,61}},
/* 2209 */ {(16<<2)|3,{49,50,65}},
/* 2210 */ {(16<<2)|3,{49,50,95}},
/* 2211 */ {(16<<2)|3,{49,50,98}},
/* 2212 */ {(16<<2)|3,{49,50,100}},
/* 2213 */ {(16<<2)|3,{49,50,102}},
/* 2214 */ {(16<<2)|3,{49,50,103}},
/* 2215 */ {(16<<2)|3,{49,50,104}},
/* 2216 */ {(16<<2)|3,{49,50,108}},
/* 2217 */ {(16<<2)|3,{49,50,109}},
/* 2218 */ {(16<<2)|3,{49,50,110}},
/* 2219 */ {(16<<2)|3,{49,50,112}},
/* 2220 */ {(16<<2)|3,{49,50,114}},
/* 2221 */ {(16<<2)|3,{49,50,117}},
/* 2222 */ {(10<<2)|2,{49,50,0}},
/* 2223 */ {(10<<2)|2,{49,50,0}},
/* 2224 */ {(10<<2)|2,{49,50,0}},
/* 2225 */ {(10<<2)|2,{49,50,0}},
/* 2226 */ {(10<<2)|2,{49,50,0}},
/* 2227 */ {(10<<2)|2,{49,50,0}},
/* 2228 */ {(10<<2)|2,{49,50,0}},
/* 2229 */ {(10<<2)|2,{49,50,0}},
/* 2230 */ {(10<<2)|2,{49,50,0}},
/* 2231 */ {(10<<2)|2,{49,50,0}},
/* 2232 */ {(10<<2)|2,{49,50,0}},
/* 2233 */ {(10<<2)|2,{49,50,0}},
/* 2234 */ {(10<<2)|2,{49,50,0}},
/* 2235 */ {(10<<2)|2,{49,50,0}},
/* 2236 */ {(10<<2)|2,{49,50,0}},
/* 2237 */ {(10<<2)|2,{49,50,0}},
/* 2238 */ {(10<<2)|2,{49,50,0}},
/* 2239 */ {(10<<2)|2,{49,50,0}},
/* 2240 */ {(15<<2)|3,{49,97,48}},
/* 2241 */ {(15<<2)|3,{49,97,48}},
/* 2242 */ {(15<<2)|3,{49,97,49}},
/* 2243 */ {(15<<2)|3,{49,97,49}},
/* 2244 */ {(15<<2)|3,{49,97,50}},
/* 2245 */ {(15<<2)|3,{49,97,50}},
/* 2246 */ {(15<<2)|3,{49,97,97}},
/* 2247 */ {(15<<2)|3,{49,97,97}},
/* 2248 */ {(15<<2)|3,{49,97,99}},
/* 2249 */ {(15<<2)|3,{49,97,99}},
/* 2250 */ {(15<<2)|3,{49,97,101}},
/* 2251 */ {(15<<2)|3,{49,97,101}},
/* 2252 */ {(15<<2)|3,{49,97,105}},
/* 2253 */ {(15<<2)|3,{49,97,105}},
/* 2254 */ {(15<<2)|3,{49,97,111}},
/* 2255 */ {(15<<2)|3,{49,97,111}},
/* 2256 */ {(15<<2)|3,{49,97,115}},
/* 2257 */ {(15<<2)|3,{49,97,115}},
/* 2258 */ {(15<<2)|3,{49,97,116}},
/* 2259 */ {(15<<2)|3,{49,97,116}},
/* 2260 */ {(16<<2)|3,{49,97,32}},
/* 2261 */ {(16<<2)|3,{49,97,37}},
/* 2262 */ {(16<<2)|3,{49,97,45}},
/* 2263 */ {(16<<2)|3,{49,97,46}},
/* 2264 */ {(16<<2)|3,{49,97,47}},
/* 2265 */ {(16<<2)|3,{49,97,51}},
/* 2266 */ {(16<<2)|3,{49,97,52}},
/* 2267 */ {(16<<2)|3,{49,97,53}},
/* 2268 */ {(16<<2)|3,{49,97,54}},
/* 2269 */ {(16<<2)|3,{49,97,55}},
/* 2270 */ {(16<<2)|3,{49,97,56}},
/* 2271 */ {(16<<2)|3,{49,97,57}},
/* 2272 */ {(16<<2)|3,{49,97,61}},
/* 2273 */ {(16<<2)|3,{49,97,65}},
/* 2274 */ {(16<<2)|3,{49,97,95}},
/* 2275 */ {(16<<2)|3,{49,97,98}},
/* 2276 */ {(16<<2)|3,{49,97,100}},
/* 2277 */ {(16<<2)|3,{49,97,102}},
/* 2278 */ {(16<<2)|3,{49,97,103}},
/* 2279 */ {(16<<2)|3,{49,97,104}},
/* 2280 */ {(16<<2)|3,{49,97,108}},
/* 2281 */ {(16<<2)|3,{49,97,109}},
/* 2282 */ {(16<<2)|3,{49,97,110}},
/* 2283 */ {(16<<2)|3,{49,97,112}},
/* 2284 */ {(16<<2)|3,{49,97,114}},
/* 2285 */ {(16<<2)|3,{49,97,117}},
/* 2286 */ {(10<<2)|2,{49,97,0}},
/* 2287 */ {(10<<2)|2,{49,97,0}},
/* 2288 */ {(10<<2)|2,{49,97,0}},
/* 2289 */ {(10<<2)|2,{49,97,0}},
/* 2290 */ {(10<<2)|2,{49,97,0}},
/* 2291 */ {(10<<2)|2,{49,97,0}},
/* 2292 */ {(10<<2)|2,{49,97,0}},
/* 2293 */ {(10<<2)|2,{49,97,0}},
/* 2294 */ {(10<<2)|2,{49,97,0}},
/* 2295 */ {(10<<2)|2,{49,97,0}},
/* 2296 */ {(10<<2)|2,{49,97,0}},
/* 2297 */ {(10<<2)|2,{49,97,0}},
/* 2298 */ {(10<<2)|2,{49,97,0}},
/* 2299 */ {(10<<2)|2,{49,97,0}},
/* 2300 */ {(10<<2)|2,{49,97,0}},
/* 2301 */ {(10<<2)|2,{49,97,0}},
/* 2302 */ {(10<<2)|2,{49,97,0}},
/* 2303 */ {(10<<2)|2,{49,97,0}},
/* 2304 */ {(15<<2)|3,{49,99,48}},
/* 2305 */ {(15<<2)|3,{49,99,48}},
/* 2306 */ {(15<<2)|3,{49,99,49}},
/* 2307 */ {(15<<2)|3,{49,99,49}},
/* 2308 */ {(15<<2)|3,{49,99,50}},
/* 2309 */ {(15<<2)|3,{49,99,50}},
/* 2310 */ {(15<<2)|3,{49,99,97}},
/* 2311 */ {(15<<2)|3,{49,99,97}},
/* 2312 */ {(15<<2)|3,{49,99,99}},
/* 2313 */ {(15<<2)|3,{49,99,99}},
/* 2314 */ {(15<<2)|3,{49,99,101}},
/* 2315 */ {(15<<2)|3,{49,99,101}},
/* 2316 */ {(15<<2)|3,{49,99,105}},
/* 2317 */ {(15<<2)|3,{49,99,105}},
/* 2318 */ {(15<<2)|3,{49,99,111}},
/* 2319 */ {(15<<2)|3,{49,99,111}},
/* 2320 */ {(15<<2)|3,{49,99,115}},
/* 2321 */ {(15<<2)|3,{49,99,115}},
/* 2322 */ {(15<<2)|3,{49,99,116}},
/* 2323 */ {(15<<2)|3,{49,99,116}},
/* 2324 */ {(16<<2)|3,{49,99,32}},
/* 2325 */ {(16<<2)|3,{49,99,37}},
/* 2326 */ {(16<<2)|3,{49,99,45}},
/* 2327 */ {(16<<2)|3,{49,99,46}},
/* 2328 */ {(16<<2)|3,{49,99,47}},
/* 2329 */ {(16<<2)|3,{49,99,51}},
/* 2330 */ {(16<<2)|3,{49,99,52}},
/* 2331 */ {(16<<2)|3,{49,99,53}},
/* 2332 */ {(16<<2)|3,{49,99,54}},
/* 2333 */ {(16<<2)|3,{49,99,55}},
/* 2334 */ {(16<<2)|3,{49,99,56}},
/* 2335 */ {(16<<2)|3,{49,99,57}},
/* 2336 */ {(16<<2)|3,{49,99,61}},
/* 2337 */ {(16<<2)|3,{49,99,65}},
/* 2338 */ {(16<<2)|3,{49,99,95}},
/* 2339 */ {(16<<2)|3,{49,99,98}},
/* 2340 */ {(16<<2)|3,{49,99,100}},
/* 2341 */ {(16<<2)|3,{49,99,102}},
/* 2342 */ {(16<<2)|3,{49,99,103}},
/* 2343 */ {(16<<2)|3,{49,99,104}},
/* 2344 */ {(16<<2)|3,{49,99,108}},
/* 2345 */ {(16<<2)|3,{49,99,109}},
/* 2346 */ {(16<<2)|3,{49,99,110}},
/* 2347 */ {(16<<2)|3,{49,99,112}},
/* 2348 */ {(16<<2)|3,{49,99,114}},
/* 2349 */ {(16<<2)|3,{49,99,117}},
/* 2350 */ {(10<<2)|2,{49,99,0}},
/* 2351 */ {(10<<2)|2,{49,99,0}},
/* 2352 */ {(10<<2)|2,{49,99,0}},
/* 2353 */ {(10<<2)|2,{49,99,0}},
/* 2354 */ {(10<<2)|2,{49,99,0}},
/* 2355 */ {(10<<2)|2,{49,99,0}},
/* 2356 */ {(10<<2)|2,{49,99,0}},
/* 2357 */ {(10<<2)|2,{49,99,0}},
/* 2358 */ {(10<<2)|2,{49,99,0}},
/* 2359 */ {(10<<2)|2,{49,99,0}},
/* 2360 */ {(10<<2)|2,{49,99,0}},
/* 2361 */ {(10<<2)|2,{49,99,0}},
/* 2362 */ {(10<<2)|2,{49,99,0}},
/* 2363 */ {(10<<2)|2,{49,99,0}},
/* 2364 */ {(10<<2)|2,{49,99,0}},
/* 2365 */ {(10<<2)|2,{49,99,0}},
/* 2366 */ {(10<<2)|2,{49,99,0}},
/* 2367 */ {(10<<2)|2,{49,99,0}},
/* 2368 */ {(15<<2)|3,{49,101,48}},
/* 2369 */ {(15<<2)|3,{49,101,48}},
/* 2370 */ {(15<<2)|3,{49,101,49}},
/* 2371 */ {(15<<2)|3,{49,101,49}},
/* 2372 */ {(15<<2)|3,{49,101,50}},
/* 2373 */ {(15<<2)|3,{49,101,50}},
/* 2374 */ {(15<<2)|3,{49,101,97}},
/* 2375 */ {(15<<2)|3,{49,101,97}},
/* 2376 */ {(15<<2)|3,{49,101,99}},
/* 2377 */ {(15<<2)|3,{49,101,99}},
/* 2378 */ {(15<<2)|3,{49,101,101}},
/* 2379 */ {(15<<2)|3,{49,101,101}},
/* 2380 */ {(15<<2)|3,{49,101,105}},
/* 2381 */ {(15<<2)|3,{49,101,105}},
/* 2382 */ {(15<<2)|3,{49,101,111}},
/* 2383 */ {(15<<2)|3,{49,101,111}},
/* 2384 */ {(15<<2)|3,{49,101,115}},
/* 2385 */ {(15<<2)|3,{49,101,115}},
/* 2386 */ {(15<<2)|3,{49,101,116}},
/* 2387 */ {(15<<2)|3,{49,101,116}},
/* 2388 */ {(16<<2)|3,{49,101,32}},
/* 2389 */ {(16<<2)|3,{49,101,37}},
/* 2390 */ {(16<<2)|3,{49,101,45}},
/* 2391 */ {(16<<2)|3,{49,101,46}},
/* 2392 */ {(16<<2)|3,{49,101,47}},
/* 2393 */ {(16<<2)|3,{49,101,51}},
/* 2394 */ {(16<<2)|3,{49,101,52}},
/* 2395 */ {(16<<2)|3,{49,101,53}},
/* 2396 */ {(16<<2)|3,{49,101,54}},
/* 2397 */ {(16<<2)|3,{49,101,55}},
/* 2398 */ {(16<<2)|3,{49,101,56}},
/* 2399 */ {(16<<2)|3,{49,101,57}},
/* 2400 */ {(16<<2)|3,{49,101,61}},
/* 2401 */ {(16<<2)|3,{49,101,65}},
/* 2402 */ {(16<<2)|3,{49,101,95}},
/* 2403 */ {(16<<2)|3,{49,101,98}},
/* 2404 */ {(16<<2)|3,{49,101,100}},
/* 2405 */ {(16<<2)|3,{49,101,102}},
/* 2406 */ {(16<<2)|3,{49,101,103}},
/* 2407 */ {(16<<2)|3,{49,101,104}},
/* 2408 */ {(16<<2)|3,{49,101,108}},
/* 2409 */ {(16<<2)|3,{49,101,109}},
/* 2410 */ {(16<<2)|3,{49,101,110}},
/* 2411 */ {(16<<2)|3,{49,101,112}},
/* 2412 */ {(16<<2)|3,{49,101,114}},
/* 2413 */ {(16<<2)|3,{49,101,117}},
/* 2414 */ {(10<<2)|2,{49,101,0}},
/* 2415 */ {(10<<2)|2,{49,101,0}},
/* 2416 */ {(10<<2)|2,{49,101,0}},
/* 2417 */ {(10<<2)|2,{49,101,0}},
/* 2418 */ {(10<<2)|2,{49,101,0}},
/* 2419 */ {(10<<2)|2,{49,101,0}},
/* 2420 */ {(10<<2)|2,{49,101,0}},
/* 2421 */ {(10<<2)|2,{49,101,0}},
/* 2422 */ {(10<<2)|2,{49,101,0}},
/* 2423 */ {(10<<2)|2,{49,101,0}},
/* 2424 */ {(10<<2)|2,{49,101,0}},
/* 2425 */ {(10<<2)|2,{49,101,0}},
/* 2426 */ {(10<<2)|2,{49,101,0}},
/* 2427 */ {(10<<2)|2,{49,101,0}},
/* 2428 */ {(10<<2)|2,{49,101,0}},
/* 2429 */ {(10<<2)|2,{49,101,0}},
/* 2430 */ {(10<<2)|2,{49,101,0}},
/* 2431 */ {(10<<2)|2,{49,101,0}},
/* 2432 */ {(15<<2)|3,{49,105,48}},
/* 2433 */ {(15<<2)|3,{49,105,48}},
/* 2434 */ {(15<<2)|3,{49,105,49}},
/* 2435 */ {(15<<2)|3,{49,105,49}},
/* 2436 */ {(15<<2)|3,{49,105,50}},
/* 2437 */ {(15<<2)|3,{49,105,50}},
/* 2438 */ {(15<<2)|3,{49,105,97}},
/* 2439 */ {(15<<2)|3,{49,105,97}},
/* 2440 */ {(15<<2)|3,{49,105,99}},
/* 2441 */ {(15<<2)|3,{49,105,99}},
/* 2442 */ {(15<<2)|3,{49,105,101}},
/* 2443 */ {(15<<2)|3,{49,105,101}},
/* 2444 */ {(15<<2)|3,{49,105,105}},
/* 2445 */ {(15<<2)|3,{49,105,105}},
/* 2446 */ {(15<<2)|3,{49,105,111}},
/* 2447 */ {(15<<2)|3,{49,105,111}},
/* 2448 */ {(15<<2)|3,{49,105,115}},
/* 2449 */ {(15<<2)|3,{49,105,115}},
/* 2450 */ {(15<<2)|3,{49,105,116}},
/* 2451 */ {(15<<2)|3,{49,105,116}},
/* 2452 */ {(16<<2)|3,{49,105,32}},
/* 2453 */ {(16<<2)|3,{49,105,37}},
/* 2454 */ {(16<<2)|3,{49,105,45}},
/* 2455 */ {(16<<2)|3,{49,105,46}},
/* 2456 */ {(16<<2)|3,{49,105,47}},
/* 2457 */ {(16<<2)|3,{49,105,51}},
/* 2458 */ {(16<<2)|3,{49,105,52}},
/* 2459 */ {(16<<2)|3,{49,105,53}},
/* 2460 */ {(16<<2)|3,{49,105,54}},
/* 2461 */ {(16<<2)|3,{49,105,55}},
/* 2462 */ {(16<<2)|3,{49,105,56}},
/* 2463 */ {(16<<2)|3,{49,105,57}},
/* 2464 */ {(16<<2)|3,{49,105,61}},
/* 2465 */ {(16<<2)|3,{49,105,65}},
/* 2466 */ {(16<<2)|3,{49,105,95}},
/* 2467 */ {(16<<2)|3,{49,105,98}},
/* 2468 */ {(16<<2)|3,{49,105,100}},
/* 2469 */ {(16<<2)|3,{49,105,102}},
/* 2470 */ {(16<<2)|3,{49,105,103}},
/* 2471 */ {(16<<2)|3,{49,105,104}},
/* 2472 */ {(16<<2)|3,{49,105,108}},
/* 2473 */ {(16<<2)|3,{49,105,109}},
/* 2474 */ {(16<<2)|3,{49,105,110}},
/* 2475 */ {(16<<2)|3,{49,105,112}},
/* 2476 */ {(16<<2)|3,{49,105,114}},
/* 2477 */ {(16<<2)|3,{49,105,117}},
/* 2478 */ {(10<<2)|2,{49,105,0}},
/* 2479 */ {(10<<2)|2,{49,105,0}},
/* 2480 */ {(10<<2)|2,{49,105,0}},
/* 2481 */ {(10<<2)|2,{49,105,0}},
/* 2482 */ {(10<<2)|2,{49,105,0}},
/* 2483 */ {(10<<2)|2,{49,105,0}},
/* 2484 */ {(10<<2)|2,{49,105,0}},
/* 2485 */ {(10<<2)|2,{49,105,0}},
/* 2486 */ {(10<<2)|2,{49,105,0}},
/* 2487 */ {(10<<2)|2,{49,105,0}},
/* 2488 */ {(10<<2)|2,{49,105,0}},
/* 2489 */ {(10<<2)|2,{49,105,0}},
/* 2490 */ {(10<<2)|2,{49,105,0}},
/* 2491 */ {(10<<2)|2,{49,105,0}},
/* 2492 */ {(10<<2)|2,{49,105,0}},
/* 2493 */ {(10<<2)|2,{49,105,0}},
/* 2494 */ {(10<<2)|2,{49,105,0}},
/* 2495 */ {(10<<2)|2,{49,105,0}},
/* 2496 */ {(15<<2)|3,{49,111,48}},
/* 2497 */ {(15<<2)|3,{49,111,48}},
/* 2498 */ {(15<<2)|3,{49,111,49}},
/* 2499 */ {(15<<2)|3,{49,111,49}},
/* 2500 */ {(15<<2)|3,{49,111,50}},
/* 2501 */ {(15<<2)|3,{49,111,50}},
/* 2502 */ {(15<<2)|3,{49,111,97}},
/* 2503 */ {(15<<2)|3,{49,111,97}},
/* 2504 */ {(15<<2)|3,{49,111,99}},
/* 2505 */ {(15<<2)|3,{49,111,99}},
/* 2506 */ {(15<<2)|3,{49,111,101}},
/* 2507 */ {(15<<2)|3,{49,111,101}},
/* 2508 */ {(15<<2)|3,{49,111,105}},
/* 2509 */ {(15<<2)|3,{49,111,105}},
/* 2510 */ {(15<<2)|3,{49,111,111}},
/* 2511 */ {(15<<2)|3,{49,111,111}},
/* 2512 */ {(15<<2)|3,{49,111,115}},
/* 2513 */ {(15<<2)|3,{49,111,115}},
/* 2514 */ {(15<<2)|3,{49,111,116}},
/* 2515 */ {(15<<2)|3,{49,111,116}},
/* 2516 */ {(16<<2)|3,{49,111,32}},
/* 2517 */ {(16<<2)|3,{49,111,37}},
/* 2518 */ {(16<<2)|3,{49,111,45}},
/* 2519 */ {(16<<2)|3,{49,111,46}},
/* 2520 */ {(16<<2)|3,{49,111,47}},
/* 2521 */ {(16<<2)|3,{49,111,51}},
/* 2522 */ {(16<<2)|3,{49,111,52}},
/* 2523 */ {(16<<2)|3,{49,111,53}},
/* 2524 */ {(16<<2)|3,{49,111,54}},
/* 2525 */ {(16<<2)|3,{49,111,55}},
/* 2526 */ {(16<<2)|3,{49,111,56}},
/* 2527 */ {(16<<2)|3,{49,111,57}},
/* 2528 */ {(16<<2)|3,{49,111,61}},
/* 2529 */ {(16<<2)|3,{49,111,65}},
/* 2530 */ {(16<<2)|3,{49,111,95}},
/* 2531 */ {(16<<2)|3,{49,111,98}},
/* 2532 */ {(16<<2)|3,{49,111,100}},
/* 2533 */ {(16<<2)|3,{49,111,102}},
/* 2534 */ {(16<<2)|3,{49,111,103}},
/* 2535 */ {(16<<2)|3,{49,111,104}},
/* 2536 */ {(16<<2)|3,{49,111,108}},
/* 2537 */ {(16<<2)|3,{49,111,109}},
/* 2538 */ {(16<<2)|3,{49,111,110}},
/* 2539 */ {(16<<2)|3,{49,111,112}},
/* 2540 */ {(16<<2)|3,{49,111,114}},
/* 2541 */ {(16<<2)|3,{49,111,117}},
/* 2542 */ {(10<<2)|2,{49,111,0}},
/* 2543 */ {(10<<2)|2,{49,111,0}},
/* 2544 */ {(10<<2)|2,{49,111,0}},
/* 2545 */ {(10<<2)|2,{49,111,0}},
/* 2546 */ {(10<<2)|2,{49,111,0}},
/* 2547 */ {(10<<2)|2,{49,111,0}},
/* 2548 */ {(10<<2)|2,{49,111,0}},
/* 2549 */ {(10<<2)|2,{49,111,0}},
/* 2550 */ {(10<<2)|2,{49,111,0}},
/* 2551 */ {(10<<2)|2,{49,111,0}},
/* 2552 */ {(10<<2)|2,{49,111,0}},
/* 2553 */ {(10<<2)|2,{49,111,0}},
/* 2554 */ {(10<<2)|2,{49,111,0}},
/* 2555 */ {(10<<2)|2,{49,111,0}},
/* 2556 */ {(10<<2)|2,{49,111,0}},
/* 2557 */ {(10<<2)|2,{49,111,0}},
/* 2558 */ {(10<<2)|2,{49,111,0}},
/* 2559 */ {(10<<2)|2,{49,111,0}},
/* 2560 */ {(15<<2)|3,{49,115,48}},
/* 2561 */ {(15<<2)|3,{49,115,48}},
/* 2562 */ {(15<<2)|3,{49,115,49}},
/* 2563 */ {(15<<2)|3,{49,115,49}},
/* 2564 */ {(15<<2)|3,{49,115,50}},
/* 2565 */ {(15<<2)|3,{49,115,50}},
/* 2566 */ {(15<<2)|3,{49,115,97}},
/* 2567 */ {(15<<2)|3,{49,115,97}},
/* 2568 */ {(15<<2)|3,{49,115,99}},
/* 2569 */ {(15<<2)|3,{49,115,99}},
/* 2570 */ {(15<<2)|3,{49,115,101}},
/* 2571 */ {(15<<2)|3,{49,115,101}},
/* 2572 */ {(15<<2)|3,{49,115,105}},
/* 2573 */ {(15<<2)|3,{49,115,105}},
/* 2574 */ {(15<<2)|3,{49,115,111}},
/* 2575 */ {(15<<2)|3,{49,115,111}},
/* 2576 */ {(15<<2)|3,{49,115,115}},
/* 2577 */ {(15<<2)|3,{49,115,115}},
/* 2578 */ {(15<<2)|3,{49,115,116}},
/* 2579 */ {(15<<2)|3,{49,115,116}},
/* 2580 */ {(16<<2)|3,{49,115,32}},
/* 2581 */ {(16<<2)|3,{49,115,37}},
/* 2582 */ {(16<<2)|3,{49,115,45}},
/* 2583 */ {(16<<2)|3,{49,115,46}},
/* 2584 */ {(16<<2)|3,{49,115,47}},
/* 2585 */ {(16<<2)|3,{49,115,51}},
/* 2586 */ {(16<<2)|3,{49,115,52}},
/* 2587 */ {(16<<2)|3,{49,115,53}},
/* 2588 */ {(16<<2)|3,{49,115,54}},
/* 2589 */ {(16<<2)|3,{49,115,55}},
/* 2590 */ {(16<<2)|3,{49,115,56}},
/* 2591 */ {(16<<2)|3,{49,115,57}},
/* 2592 */ {(16<<2)|3,{49,115,61}},
/* 2593 */ {(16<<2)|3,{49,115,65}},
/* 2594 */ {(16<<2)|3,{49,115,95}},
/* 2595 */ {(16<<2)|3,{49,115,98}},
/* 2596 */ {(16<<2)|3,{49,115,100}},
/* 2597 */ {(16<<2)|3,{49,115,102}},
/* 2598 */ {(16<<2)|3,{49,115,103}},
/* 2599 */ {(16<<2)|3,{49,115,104}},
/* 2600 */ {(16<<2)|3,{49,115,108}},
/* 2601 */ {(16<<2)|3,{49,115,109}},
/* 2602 */ {(16<<2)|3,{49,115,110}},
/* 2603 */ {(16<<2)|3,{49,115,112}},
/* 2604 */ {(16<<2)|3,{49,115,114}},
/* 2605 */ {(16<<2)|3,{49,115,117}},
/* 2606 */ {(10<<2)|2,{49,115,0}},
/* 2607 */ {(10<<2)|2,{49,115,0}},
/* 2608 */ {(10<<2)|2,{49,115,0}},
/* 2609 */ {(10<<2)|2,{49,115,0}},
/* 2610 */ {(10<<2)|2,{49,115,0}},
/* 2611 */ {(10<<2)|2,{49,115,0}},
/* 2612 */ {(10<<2)|2,{49,115,0}},
/* 2613 */ {(10<<2)|2,{49,115,0}},
/* 2614 */ {(10<<2)|2,{49,115,0}},
/* 2615 */ {(10<<2)|2,{49,115,0}},
/* 2616 */ {(10<<2)|2,{49,115,0}},
/* 2617 */ {(10<<2)|2,{49,115,0}},
/* 2618 */ {(10<<2)|2,{49,115,0}},
/* 2619 */ {(10<<2)|2,{49,115,0}},
/* 2620 */ {(10<<2)|2,{49,115,0}},
/* 2621 */ {(10<<2)|2,{49,115,0}},
/* 2622 */ {(10<<2)|2,{49,115,0}},
/* 2623 */ {(10<<2)|2,{49,115,0}},
/* 2624 */ {(15<<2)|3,{49,116,48}},
/* 2625 */ {(15<<2)|3,{49,116,48}},
/* 2626 */ {(15<<2)|3,{49,116,49}},
/* 2627 */ {(15<<2)|3,{49,116,49}},
/* 2628 */ {(15<<2)|3,{49,116,50}},
/* 2629 */ {(15<<2)|3,{49,116,50}},
/* 2630 */ {(15<<2)|3,{49,116,97}},
/* 2631 */ {(15<<2)|3,{49,116,97}},
/* 2632 */ {(15<<2)|3,{49,116,99}},
/* 2633 */ {(15<<2)|3,{49,116,99}},
/* 2634 */ {(15<<2)|3,{49,116,101}},
/* 2635 */ {(15<<2)|3,{49,116,101}},
/* 2636 */ {(15<<2)|3,{49,116,105}},
/* 2637 */ {(15<<2)|3,{49,116,105}},
/* 2638 */ {(15<<2)|3,{49,116,111}},
/* 2639 */ {(15<<2)|3,{49,116,111}},
/* 2640 */ {(15<<2)|3,{49,116,115}},
/* 2641 */ {(15<<2)|3,{49,116,115}},
/* 2642 */ {(15<<2)|3,{49,116,116}},
/* 2643 */ {(15<<2)|3,{49,116,116}},
/* 2644 */ {(16<<2)|3,{49,116,32}},
/* 2645 */ {(16<<2)|3,{49,116,37}},
/* 2646 */ {(16<<2)|3,{49,116,45}},
/* 2647 */ {(16<<2)|3,{49,116,46}},
/* 2648 */ {(16<<2)|3,{49,116,47}},
/* 2649 */ {(16<<2)|3,{49,116,51}},
/* 2650 */ {(16<<2)|3,{49,116,52}},
/* 2651 */ {(16<<2)|3,{49,116,53}},
/* 2652 */ {(16<<2)|3,{49,116,54}},
/* 2653 */ {(16<<2)|3,{49,116,55}},
/* 2654 */ {(16<<2)|3,{49,116,56}},
/* 2655 */ {(16<<2)|3,{49,116,57}},
/* 2656 */ {(16<<2)|3,{49,116,61}},
/* 2657 */ {(16<<2)|3,{49,116,65}},
/* 2658 */ {(16<<2)|3,{49,116,95}},
/* 2659 */ {(16<<2)|3,{49,116,98}},
/* 2660 */ {(16<<2)|3,{49,116,100}},
/* 2661 */ {(16<<2)|3,{49,116,102}},
/* 2662 */ {(16<<2)|3,{49,116,103}},
/* 2663 */ {(16<<2)|3,{49,116,104}},
/* 2664 */ {(16<<2)|3,{49,116,108}},
/* 2665 */ {(16<<2)|3,{49,116,109}},
/* 2666 */ {(16<<2)|3,{49,116,110}},
/* 2667 */ {(16<<2)|3,{49,116,112}},
/* 2668 */ {(16<<2)|3,{49,116,114}},
/* 2669 */ {(16<<2)|3,{49,116,117}},
/* 2670 */ {(10<<2)|2,{49,116,0}},
/* 2671 */ {(10<<2)|2,{49,116,0}},
/* 2672 */ {(10<<2)|2,{49,116,0}},
/* 2673 */ {(10<<2)|2,{49,116,0}},
/* 2674 */ {(10<<2)|2,{49,116,0}},
/* 2675 */ {(10<<2)|2,{49,116,0}},
/* 2676 */ {(10<<2)|2,{49,116,0}},
/* 2677 */ {(10<<2)|2,{49,116,0}},
/* 2678 */ {(10<<2)|2,{49,116,0}},
/* 2679 */ {(10<<2)|2,{49,116,0}},
/* 2680 */ {(10<<2)|2,{49,116,0}},
/* 2681 */ {(10<<2)|2,{49,116,0}},
/* 2682 */ {(10<<2)|2,{49,116,0}},
/* 2683 */ {(10<<2)|2,{49,116,0}},
/* 2684 */ {(10<<2)|2,{49,116,0}},
/* 2685 */ {(10<<2)|2,{49,116,0}},
/* 2686 */ {(10<<2)|2,{49,116,0}},
/* 2687 */ {(10<<2)|2,{49,116,0}},
/* 2688 */ {(16<<2)|3,{49,32,48}},
/* 2689 */ {(16<<2)|3,{49,32,49}},
/* 2690 */ {(16<<2)|3,{49,32,50}},
/* 2691 */ {(16<<2)|3,{49,32,97}},
/* 2692 */ {(16<<2)|3,{49,32,99}},
/* 2693 */ {(16<<2)|3,{49,32,101}},
/* 2694 */ {(16<<2)|3,{49,32,105}},
/* 2695 */ {(16<<2)|3,{49,32,111}},
/* 2696 */ {(16<<2)|3,{49,32,115}},
/* 2697 */ {(16<<2)|3,{49,32,116}},
/* 2698 */ {(11<<2)|2,{49,32,0}},
/* 2699 */ {(11<<2)|2,{49,32,0}},
/* 2700 */ {(11<<2)|2,{49,32,0}},
/* 2701 */ {(11<<2)|2,{49,32,0}},
/* 2702 */ {(11<<2)|2,{49,32,0}},
/* 2703 */ {(11<<2)|2,{49,32,0}},
/* 2704 */ {(11<<2)|2,{49,32,0}},
/* 2705 */ {(11<<2)|2,{49,32,0}},
/* 2706 */ {(11<<2)|2,{49,32,0}},
/* 2707 */ {(11<<2)|2,{49,32,0}},
/* 2708 */ {(11<<2)|2,{49,32,0}},
/* 2709 */ {(11<<2)|2,{49,32,0}},
/* 2710 */ {(11<<2)|2,{49,32,0}},
/* 2711 */ {(11<<2)|2,{49,32,0}},
/* 2712 */ {(11<<2)|2,{49,32,0}},
/* 2713 */ {(11<<2)|2,{49,32,0}},
/* 2714 */ {(11<<2)|2,{49,32,0}},
/* 2715 */ {(11<<2)|2,{49,32,0}},
/* 2716 */ {(11<<2)|2,{49,32,0}},
/* 2717 */ {(11<<2)|2,{49,32,0}},
/* 2718 */ {(11<<2)|2,{49,32,0}},
/* 2719 */ {(11<<2)|2,{49,32,0}},
/* 2720 */ {(16<<2)|3,{49,37,48}},
/* 2721 */ {(16<<2)|3,{49,37,49}},
/* 2722 */ {(16<<2)|3,{49,37,50}},
/* 2723 */ {(16<<2)|3,{49,37,97}},
/* 2724 */ {(16<<2)|3,{49,37,99}},
/* 2725 */ {(16<<2)|3,{49,37,101}},
/* 2726 */ {(16<<2)|3,{49,37,105}},
/* 2727 */ {(16<<2)|3,{49,37,111}},
/* 2728 */ {(16<<2)|3,{49,37,115}},
/* 2729 */ {(16<<2)|3,{49,37,116}},
/* 2730 */ {(11<<2)|2,{49,37,0}},
/* 2731 */ {(11<<2)|2,{49,37,0}},
/* 2732 */ {(11<<2)|2,{49,37,0}},
/* 2733 */ {(11<<2)|2,{49,37,0}},
/* 2734 */ {(11<<2)|2,{49,37,0}},
/* 2735 */ {(11<<2)|2,{49,37,0}},
/* 2736 */ {(11<<2)|2,{49,37,0}},
/* 2737 */ {(11<<2)|2,{49,37,0}},
/* 2738 */ {(11<<2)|2,{49,37,0}},
/* 2739 */ {(11<<2)|2,{49,37,0}},
/* 2740 */ {(11<<2)|2,{49,37,0}},
/* 2741 */ {(11<<2)|2,{49,37,0}},
/* 2742 */ {(11<<2)|2,{49,37,0}},
/* 2743 */ {(11<<2)|2,{49,37,0}},
/* 2744 */ {(11<<2)|2,{49,37,0}},
/* 2745 */ {(11<<2)|2,{49,37,0}},
/* 2746 */ {(11<<2)|2,{49,37,0}},
/* 2747 */ {(11<<2)|2,{49,37,0}},
/* 2748 */ {(11<<2)|2,{49,37,0}},
/* 2749 */ {(11<<2)|2,{49,37,0}},
/* 2750 */ {(11<<2)|2,{49,37,0}},
/* 2751 */ {(11<<2)|2,{49,37,0}},
/* 2752 */ {(16<<2)|3,{49,45,48}},
/* 2753 */ {(16<<2)|3,{49,45,49}},
/* 2754 */ {(16<<2)|3,{49,45,50}},
/* 2755 */ {(16<<2)|3,{49,45,97}},
/* 2756 */ {(16<<2)|3,{49,45,99}},
/* 2757 */ {(16<<2)|3,{49,45,101}},
/* 2758 */ {(16<<2)|3,{49,45,105}},
/* 2759 */ {(16<<2)|3,{49,45,111}},
/* 2760 */ {(16<<2)|3,{49,45,115}},
/* 2761 */ {(16<<2)|3,{49,45,116}},
/* 2762 */ {(11<<2)|2,{49,45,0}},
/* 2763 */ {(11<<2)|2,{49,45,0}},
/* 2764 */ {(11<<2)|2,{49,45,0}},
/* 2765 */ {(11<<2)|2,{49,45,0}},
/* 2766 */ {(11<<2)|2,{49,45,0}},
/* 2767 */ {(11<<2)|2,{49,45,0}},
/* 2768 */ {(11<<2)|2,{49,45,0}},
/* 2769 */ {(11<<2)|2,{49,45,0}},
/* 2770 */ {(11<<2)|2,{49,45,0}},
/* 2771 */ {(11<<2)|2,{49,45,0}},
/* 2772 */ {(11<<2)|2,{49,45,0}},
/* 2773 */ {(11<<2)|2,{49,45,0}},
/* 2774 */ {(11<<2)|2,{49,45,0}},
/* 2775 */ {(11<<2)|2,{49,45,0}},
/* 2776 */ {(11<<2)|2,{49,45,0}},
/* 2777 */ {(11<<2)|2,{49,45,0}},
/* 2778 */ {(11<<2)|2,{49,45,0}},
/* 2779 */ {(11<<2)|2,{49,45,0}},
/* 2780 */ {(11<<2)|2,{49,45,0}},
/* 2781 */ {(11<<2)|2,{49,45,0}},
/* 2782 */ {(11<<2)|2,{49,45,0}},
/* 2783 */ {(11<<2)|2,{49,45,0}},
/* 2784 */ {(16<<2)|3,{49,46,48}},
/* 2785 */ {(16<<2)|3,{49,46,49}},
/* 2786 */ {(16<<2)|3,{49,46,50}},
/* 2787 */ {(16<<2)|3,{49,46,97}},
/* 2788 */ {(16<<2)|3,{49,46,99}},
/* 2789 */ {(16<<2)|3,{49,46,101}},
/* 2790 */ {(16<<2)|3,{49,46,105}},
/* 2791 */ {(16<<2)|3,{49,46,111}},
/* 2792 */ {(16<<2)|3,{49,46,115}},
/* 2793 */ {(16<<2)|3,{49,46,116}},
/* 2794 */ {(11<<2)|2,{49,46,0}},
/* 2795 */ {(11<<2)|2,{49,46,0}},
/* 2796 */ {(11<<2)|2,{49,46,0}},
/* 2797 */ {(11<<2)|2,{49,46,0}},
/* 2798 */ {(11<<2)|2,{49,46,0}},
/* 2799 */ {(11<<2)|2,{49,46,0}},
/* 2800 */ {(11<<2)|2,{49,46,0}},
/* 2801 */ {(11<<2)|2,{49,46,0}},
/* 2802 */ {(11<<2)|2,{49,46,0}},
/* 2803 */ {(11<<2)|2,{49,46,0}},
/* 2804 */ {(11<<2)|2,{49,46,0}},
/* 2805 */ {(11<<2)|2,{49,46,0}},
/* 2806 */ {(11<<2)|2,{49,46,0}},
/* 2807 */ {(11<<2)|2,{49,46,0}},
/* 2808 */ {(11<<2)|2,{49,46,0}},
/* 2809 */ {(11<<2)|2,{49,46,0}},
/* 2810 */ {(11<<2)|2,{49,46,0}},
/* 2811 */ {(11<<2)|2,{49,46,0}},
/* 2812 */ {(11<<2)|2,{49,46,0}},
/* 2813 */ {(11<<2)|2,{49,46,0}},
/* 2814 */ {(11<<2)|2,{49,46,0}},
/* 2815 */ {(11<<2)|2,{49,46,0}},
/* 2816 */ {(16<<2)|3,{49,47,48}},
/* 2817 */ {(16<<2)|3,{49,47,49}},
/* 2818 */ {(16<<2)|3,{49,47,50}},
/* 2819 */ {(16<<2)|3,{49,47,97}},
/* 2820 */ {(16<<2)|3,{49,47,99}},
/* 2821 */ {(16<<2)|3,{49,47,101}},
/* 2822 */ {(16<<2)|3,{49,47,105}},
/* 2823 */ {(16<<2)|3,{49,47,111}},
/* 2824 */ {(16<<2)|3,{49,47,115}},
/* 2825 */ {(16<<2)|3,{49,47,116}},
/* 2826 */ {(11<<2)|2,{49,47,0}},
/* 2827 */ {(11<<2)|2,{49,47,0}},
/* 2828 */ {(11<<2)|2,{49,47,0}},
/* 2829 */ {(11<<2)|2,{49,47,0}},
/* 2830 */ {(11<<2)|2,{49,47,0}},
/* 2831 */ {(11<<2)|2,{49,47,0}},
/* 2832 */ {(11<<2)|2,{49,47,0}},
/* 2833 */ {(11<<2)|2,{49,47,0}},
/* 2834 */ {(11<<2)|2,{49,47,0}},
/* 2835 */ {(11<<2)|2,{49,47,0}},
/* 2836 */ {(11<<2)|2,{49,47,0}},
/* 2837 */ {(11<<2)|2,{49,47,0}},
/* 2838 */ {(11<<2)|2,{49,47,0}},
/* 2839 */ {(11<<2)|2,{49,47,0}},
/* 2840 */ {(11<<2)|2,{49,47,0}},
/* 2841 */ {(11<<2)|2,{49,47,0}},
/* 2842 */ {(11<<2)|2,{49,47,0}},
/* 2843 */ {(11<<2)|2,{49,47,0}},
/* 2844 */ {(11<<2)|2,{49,47,0}},
/* 2845 */ {(11<<2)|2,{49,47,0}},
/* 2846 */ {(11<<2)|2,{49,47,0}},
/* 2847 */ {(11<<2)|2,{49,47,0}},
/* 2848 */ {(16<<2)|3,{49,51,48}},
/* 2849 */ {(16<<2)|3,{49,51,49}},
/* 2850 */ {(16<<2)|3,{49,51,50}},
/* 2851 */ {(16<<2)|3,{49,51,97}},
/* 2852 */ {(16<<2)|3,{49,51,99}},
/* 2853 */ {(16<<2)|3,{49,51,101}},
/* 2854 */ {(16<<2)|3,{49,51,105}},
/* 2855 */ {(16<<2)|3,{49,51,111}},
/* 2856 */ {(16<<2)|3,{49,51,115}},
/* 2857 */ {(16<<2)|3,{49,51,116}},
/* 2858 */ {(11<<2)|2,{49,51,0}},
/* 2859 */ {(11<<2)|2,{49,51,0}},
/* 2860 */ {(11<<2)|2,{49,51,0}},
/* 2861 */ {(11<<2)|2,{49,51,0}},
/* 2862 */ {(11<<2)|2,{49,51,0}},
/* 2863 */ {(11<<2)|2,{49,51,0}},
/* 2864 */ {(11<<2)|2,{49,51,0}},
/* 2865 */ {(11<<2)|2,{49,51,0}},
/* 2866 */ {(11<<2)|2,{49,51,0}},
/* 2867 */ {(11<<2)|2,{49,51,0}},
/* 2868 */ {(11<<2)|2,{49,51,0}},
/* 2869 */ {(11<<2)|2,{49,51,0}},
/* 2870 */ {(11<<2)|2,{49,51,0}},
/* 2871 */ {(11<<2)|2,{49,51,0}},
/* 2872 */ {(11<<2)|2,{49,51,0}},
/* 2873 */ {(11<<2)|2,{49,51,0}},
/* 2874 */ {(11<<2)|2,{49,51,0}},
/* 2875 */ {(11<<2)|2,{49,51,0}},
/* 2876 */ {(11<<2)|2,{49,51,0}},
/* 2877 */ {(11<<2)|2,{49,51,0}},
/* 2878 */ {(11<<2)|2,{49,51,0}},
/* 2879 */ {(11<<2)|2,{49,51,0}},
/* 2880 */ {(16<<2)|3,{49,52,48}},
/* 2881 */ {(16<<2)|3,{49,52,49}},
/* 2882 */ {(16<<2)|3,{49,52,50}},
/* 2883 */ {(16<<2)|3,{49,52,97}},
/* 2884 */ {(16<<2)|3,{49,52,99}},
/* 2885 */ {(16<<2)|3,{49,52,101}},
/* 2886 */ {(16<<2)|3,{49,52,105}},
/* 2887 */ {(16<<2)|3,{49,52,111}},
/* 2888 */ {(16<<2)|3,{49,52,115}},
/* 2889 */ {(16<<2)|3,{49,52,116}},
/* 2890 */ {(11<<2)|2,{49,52,0}},
/* 2891 */ {(11<<2)|2,{49,52,0}},
/* 2892 */ {(11<<2)|2,{49,52,0}},
/* 2893 */ {(11<<2)|2,{49,52,0}},
/* 2894 */ {(11<<2)|2,{49,52,0}},
/* 2895 */ {(11<<2)|2,{49,52,0}},
/* 2896 */ {(11<<2)|2,{49,52,0}},
/* 2897 */ {(11<<2)|2,{49,52,0}},
/* 2898 */ {(11<<2)|2,{49,52,0}},
/* 2899 */ {(11<<2)|2,{49,52,0}},
/* 2900 */ {(11<<2)|2,{49,52,0}},
/* 2901 */ {(11<<2)|2,{49,52,0}},
/* 2902 */ {(11<<2)|2,{49,52,0}},
/* 2903 */ {(11<<2)|2,{49,52,0}},
/* 2904 */ {(11<<2)|2,{49,52,0}},
/* 2905 */ {(11<<2)|2,{49,52,0}},
/* 2906 */ {(11<<2)|2,{49,52,0}},
/* 2907 */ {(11<<2)|2,{49,52,0}},
/* 2908 */ {(11<<2)|2,{49,52,0}},
/* 2909 */ {(11<<2)|2,{49,52,0}},
/* 2910 */ {(11<<2)|2,{49,52,0}},
/* 2911 */ {(11<<2)|2,{49,52,0}},
/* 2912 */ {(16<<2)|3,{49,53,48}},
/* 2913 */ {(16<<2)|3,{49,53,49}},
/* 2914 */ {(16<<2)|3,{49,53,50}},
/* 2915 */ {(16<<2)|3,{49,53,97}},
/* 2916 */ {(16<<2)|3,{49,53,99}},
/* 2917 */ {(16<<2)|3,{49,53,101}},
/* 2918 */ {(16<<2)|3,{49,53,105}},
/* 2919 */ {(16<<2)|3,{49,53,111}},
/* 2920 */ {(16<<2)|3,{49,53,115}},
/* 2921 */ {(16<<2)|3,{49,53,116}},
/* 2922 */ {(11<<2)|2,{49,53,0}},
/* 2923 */ {(11<<2)|2,{49,53,0}},
/* 2924 */ {(11<<2)|2,{49,53,0}},
/* 2925 */ {(11<<2)|2,{49,53,0}},
/* 2926 */ {(11<<2)|2,{49,53,0}},
/* 2927 */ {(11<<2)|2,{49,53,0}},
/* 2928 */ {(11<<2)|2,{49,53,0}},
/* 2929 */ {(11<<2)|2,{49,53,0}},
/* 2930 */ {(11<<2)|2,{49,53,0}},
/* 2931 */ {(11<<2)|2,{49,53,0}},
/* 2932 */ {(11<<2)|2,{49,53,0}},
/* 2933 */ {(11<<2)|2,{49,53,0}},
/* 2934 */ {(11<<2)|2,{49,53,0}},
/* 2935 */ {(11<<2)|2,{49,53,0}},
/* 2936 */ {(11<<2)|2,{49,53,0}},
/* 2937 */ {(11<<2)|2,{49,53,0}},
/* 2938 */ {(11<<2)|2,{49,53,0}},
/* 2939 */ {(11<<2)|2,{49,53,0}},
/* 2940 */ {(11<<2)|2,{49,53,0}},
/* 2941 */ {(11<<2)|2,{49,53,0}},
/* 2942 */ {(11<<2)|2,{49,53,0}},
/* 2943 */ {(11<<2)|2,{49,53,0}},
/* 2944 */ {(16<<2)|3,{49,54,48}},
/* 2945 */ {(16<<2)|3,{49,54,49}},
/* 2946 */ {(16<<2)|3,{49,54,50}},
/* 2947 */ {(16<<2)|3,{49,54,97}},
/* 2948 */ {(16<<2)|3,{49,54,99}},
/* 2949 */ {(16<<2)|3,{49,54,101}},
/* 2950 */ {(16<<2)|3,{49,54,105}},
/* 2951 */ {(16<<2)|3,{49,54,111}},
/* 2952 */ {(16<<2)|3,{49,54,115}},
/* 2953 */ {(16<<2)|3,{49,54,116}},
/* 2954 */ {(11<<2)|2,{49,54,0}},
/* 2955 */ {(11<<2)|2,{49,54,0}},
/* 2956 */ {(11<<2)|2,{49,54,0}},
/* 2957 */ {(11<<2)|2,{49,54,0}},
/* 2958 */ {(11<<2)|2,{49,54,0}},
/* 2959 */ {(11<<2)|2,{49,54,0}},
/* 2960 */ {(11<<2)|2,{49,54,0}},
/* 2961 */ {(11<<2)|2,{49,54,0}},
/* 2962 */ {(11<<2)|2,{49,54,0}},
/* 2963 */ {(11<<2)|2,{49,54,0}},
/* 2964 */ {(11<<2)|2,{49,54,0}},
/* 2965 */ {(11<<2)|2,{49,54,0}},
/* 2966 */ {(11<<2)|2,{49,54,0}},
/* 2967 */ {(11<<2)|2,{49,54,0}},
/* 2968 */ {(11<<2)|2,{49,54,0}},
/* 2969 */ {(11<<2)|2,{49,54,0}},
/* 2970 */ {(11<<2)|2,{49,54,0}},
/* 2971 */ {(11<<2)|2,{49,54,0}},
/* 2972 */ {(11<<2)|2,{49,54,0}},
/* 2973 */ {(11<<2)|2,{49,54,0}},
/* 2974 */ {(11<<2)|2,{49,54,0}},
/* 2975 */ {(11<<2)|2,{49,54,0}},
/* 2976 */ {(16<<2)|3,{49,55,48}},
/* 2977 */ {(16<<2)|3,{49,55,49}},
/* 2978 */ {(16<<2)|3,{49,55,50}},
/* 2979 */ {(16<<2)|3,{49,55,97}},
/* 2980 */ {(16<<2)|3,{49,55,99}},
/* 2981 */ {(16<<2)|3,{49,55,101}},
/* 2982 */ {(16<<2)|3,{49,55,105}},
/* 2983 */ {(16<<2)|3,{49,55,111}},
/* 2984 */ {(16<<2)|3,{49,55,115}},
/* 2985 */ {(16<<2)|3,{49,55,116}},
/* 2986 */ {(11<<2)|2,{49,55,0}},
/* 2987 */ {(11<<2)|2,{49,55,0}},
/* 2988 */ {(11<<2)|2,{49,55,0}},
/* 2989 */ {(11<<2)|2,{49,55,0}},
/* 2990 */ {(11<<2)|2,{49,55,0}},
/* 2991 */ {(11<<2)|2,{49,55,0}},
/* 2992 */ {(11<<2)|2,{49,55,0}},
/* 2993 */ {(11<<2)|2,{49,55,0}},
/* 2994 */ {(11<<2)|2,{49,55,0}},
/* 2995 */ {(11<<2)|2,{49,55,0}},
/* 2996 */ {(11<<2)|2,{49,55,0}},
/* 2997 */ {(11<<2)|2,{49,55,0}},
/* 2998 */ {(11<<2)|2,{49,55,0}},
/* 2999 */ {(11<<2)|2,{49,55,0}},
/* 3000 */ {(11<<2)|2,{49,55,0}},
/* 3001 */ {(11<<2)|2,{49,55,0}},
/* 3002 */ {(11<<2)|2,{49,55,0}},
/* 3003 */ {(11<<2)|2,{49,55,0}},
/* 3004 */ {(11<<2)|2,{49,55,0}},
/* 3005 */ {(11<<2)|2,{49,55,0}},
/* 3006 */ {(11<<2)|2,{49,55,0}},
/* 3007 */ {(11<<2)|2,{49,55,0}},
/* 3008 */ {(16<<2)|3,{49,56,48}},
/* 3009 */ {(16<<2)|3,{49,56,49}},
/* 3010 */ {(16<<2)|3,{49,56,50}},
/* 3011 */ {(16<<2)|3,{49,56,97}},
/* 3012 */ {(16<<2)|3,{49,56,99}},
/* 3013 */ {(16<<2)|3,{49,56,101}},
/* 3014 */ {(16<<2)|3,{49,56,105}},
/* 3015 */ {(16<<2)|3,{49,56,111}},
/* 3016 */ {(16<<2)|3,{49,56,115}},
/* 3017 */ {(16<<2)|3,{49,56,116}},
/* 3018 */ {(11<<2)|2,{49,56,0}},
/* 3019 */ {(11<<2)|2,{49,56,0}},
/* 3020 */ {(11<<2)|2,{49,56,0}},
/* 3021 */ {(11<<2)|2,{49,56,0}},
/* 3022 */ {(11<<2)|2,{49,56,0}},
/* 3023 */ {(11<<2)|2,{49,56,0}},
/* 3024 */ {(11<<2)|2,{49,56,0}},
/* 3025 */ {(11<<2)|2,{49,56,0}},
/* 3026 */ {(11<<2)|2,{49,56,0}},
/* 3027 */ {(11<<2)|2,{49,56,0}},
/* 3028 */ {(11<<2)|2,{49,56,0}},
/* 3029 */ {(11<<2)|2,{49,56,0}},
/* 3030 */ {(11<<2)|2,{49,56,0}},
/* 3031 */ {(11<<2)|2,{49,56,0}},
/* 3032 */ {(11<<2)|2,{49,56,0}},
/* 3033 */ {(11<<2)|2,{49,56,0}},
/* 3034 */ {(11<<2)|2,{49,56,0}},
/* 3035 */ {(11<<2)|2,{49,56,0}},
/* 3036 */ {(11<<2)|2,{49,56,0}},
/* 3037 */ {(11<<2)|2,{49,56,0}},
/* 3038 */ {(11<<2)|2,{49,56,0}},
/* 3039 */ {(11<<2)|2,{49,56,0}},
/* 3040 */ {(16<<2)|3,{49,57,48}},
/* 3041 */ {(16<<2)|3,{49,57,49}},
/* 3042 */ {(16<<2)|3,{49,57,50}},
/* 3043 */ {(16<<2)|3,{49,57,97}},
/* 3044 */ {(16<<2)|3,{49,57,99}},
/* 3045 */ {(16<<2)|3,{49,57,101}},
/* 3046 */ {(16<<2)|3,{49,57,105}},
/* 3047 */ {(16<<2)|3,{49,57,111}},
/* 3048 */ {(16<<2)|3,{49,57,115}},
/* 3049 */ {(16<<2)|3,{49,57,116}},
/* 3050 */ {(11<<2)|2,{49,57,0}},
/* 3051 */ {(11<<2)|2,{49,57,0}},
/* 3052 */ {(11<<2)|2,{49,57,0}},
/* 3053 */ {(11<<2)|2,{49,57,0}},
/* 3054 */ {(11<<2)|2,{49,57,0}},
/* 3055 */ {(11<<2)|2,{49,57,0}},
/* 3056 */ {(11<<2)|2,{49,57,0}},
/* 3057 */ {(11<<2)|2,{49,57,0}},
/* 3058 */ {(11<<2)|2,{49,57,0}},
/* 3059 */ {(11<<2)|2,{49,57,0}},
/* 3060 */ {(11<<2)|2,{49,57,0}},
/* 3061 */ {(11<<2)|2,{49,57,0}},
/* 3062 */ {(11<<2)|2,{49,57,0}},
/* 3063 */ {(11<<2)|2,{49,57,0}},
/* 3064 */ {(11<<2)|2,{49,57,0}},
/* 3065 */ {(11<<2)|2,{49,57,0}},
/* 3066 */ {(11<<2)|2,{49,57,0}},
/* 3067 */ {(11<<2)|2,{49,57,0}},
/* 3068 */ {(11<<2)|2,{49,57,0}},
/* 3069 */ {(11<<2)|2,{49,57,0}},
/* 3070 */ {(11<<2)|2,{49,57,0}},
/* 3071 */ {(11<<2)|2,{49,57,0}},
/* 3072 */ {(16<<2)|3,{49,61,48}},
/* 3073 */ {(16<<2)|3,{49,61,49}},
/* 3074 */ {(16<<2)|3,{49,61,50}},
/* 3075 */ {(16<<2)|3,{49,61,97}},
/* 3076 */ {(16<<2)|3,{49,61,99}},
/* 3077 */ {(16<<2)|3,{49,61,101}},
/* 3078 */ {(16<<2)|3,{49,61,105}},
/* 3079 */ {(16<<2)|3,{49,61,111}},
/* 3080 */ {(16<<2)|3,{49,61,115}},
/* 3081 */ {(16<<2)|3,{49,61,116}},
/* 3082 */ {(11<<2)|2,{49,61,0}},
/* 3083 */ {(11<<2)|2,{49,61,0}},
/* 3084 */ {(11<<2)|2,{49,61,0}},
/* 3085 */ {(11<<2)|2,{49,61,0}},
/* 3086 */ {(11<<2)|2,{49,61,0}},
/* 3087 */ {(11<<2)|2,{49,61,0}},
/* 3088 */ {(11<<2)|2,{49,61,0}},
/* 3089 */ {(11<<2)|2,{49,61,0}},
/* 3090 */ {(11<<2)|2,{49,61,0}},
/* 3091 */ {(11<<2)|2,{49,61,0}},
/* 3092 */ {(11<<2)|2,{49,61,0}},
/* 3093 */ {(11<<2)|2,{49,61,0}},
/* 3094 */ {(11<<2)|2,{49,61,0}},
/* 3095 */ {(11<<2)|2,{49,61,0}},
/* 3096 */ {(11<<2)|2,{49,61,0}},
/* 3097 */ {(11<<2)|2,{49,61,0}},
/* 3098 */ {(11<<2)|2,{49,61,0}},
/* 3099 */ {(11<<2)|2,{49,61,0}},
/* 3100 */ {(11<<2)|2,{49,61,0}},
/* 3101 */ {(11<<2)|2,{49,61,0}},
/* 3102 */ {(11<<2)|2,{49,61,0}},
/* 3103 */ {(11<<2)|2,{49,61,0}},
/* 3104 */ {(16<<2)|3,{49,65,48}},
/* 3105 */ {(16<<2)|3,{49,65,49}},
/* 3106 */ {(16<<2)|3,{49,65,50}},
/* 3107 */ {(16<<2)|3,{49,65,97}},
/* 3108 */ {(16<<2)|3,{49,65,99}},
/* 3109 */ {(16<<2)|3,{49,65,101}},
/* 3110 */ {(16<<2)|3,{49,65,105}},
/* 3111 */ {(16<<2)|3,{49,65,111}},
/* 3112 */ {(16<<2)|3,{49,65,115}},
/* 3113 */ {(16<<2)|3,{49,65,116}},
/* 3114 */ {(11<<2)|2,{49,65,0}},
/* 3115 */ {(11<<2)|2,{49,65,0}},
/* 3116 */ {(11<<2)|2,{49,65,0}},
/* 3117 */ {(11<<2)|2,{49,65,0}},
/* 3118 */ {(11<<2)|2,{49,65,0}},
/* 3119 */ {(11<<2)|2,{49,65,0}},
/* 3120 */ {(11<<2)|2,{49,65,0}},
/* 3121 */ {(11<<2)|2,{49,65,0}},
/* 3122 */ {(11<<2)|2,{49,65,0}},
/* 3123 */ {(11<<2)|2,{49,65,0}},
/* 3124 */ {(11<<2)|2,{49,65,0}},
/* 3125 */ {(11<<2)|2,{49,65,0}},
/* 3126 */ {(11<<2)|2,{49,65,0}},
/* 3127 */ {(11<<2)|2,{49,65,0}},
/* 3128 */ {(11<<2)|2,{49,65,0}},
/* 3129 */ {(11<<2)|2,{49,65,0}},
/* 3130 */ {(11<<2)|2,{49,65,0}},
/* 3131 */ {(11<<2)|2,{49,65,0}},
/* 3132 */ {(11<<2)|2,{49,65,0}},
/* 3133 */ {(11<<2)|2,{49,65,0}},
/* 3134 */ {(11<<2)|2,{49,65,0}},
/* 3135 */ {(11<<2)|2,{49,65,0}},
/* 3136 */ {(16<<2)|3,{49,95,48}},
/* 3137 */ {(16<<2)|3,{49,95,49}},
/* 3138 */ {(16<<2)|3,{49,95,50}},
/* 3139 */ {(16<<2)|3,{49,95,97}},
/* 3140 */ {(16<<2)|3,{49,95,99}},
/* 3141 */ {(16<<2)|3,{49,95,101}},
/* 3142 */ {(16<<2)|3,{49,95,105}},
/* 3143 */ {(16<<2)|3,{49,95,111}},
/* 3144 */ {(16<<2)|3,{49,95,115}},
/* 3145 */ {(16<<2)|3,{49,95,116}},
/* 3146 */ {(11<<2)|2,{49,95,0}},
/* 3147 */ {(11<<2)|2,{49,95,0}},
/* 3148 */ {(11<<2)|2,{49,95,0}},
/* 3149 */ {(11<<2)|2,{49,95,0}},
/* 3150 */ {(11<<2)|2,{49,95,0}},
/* 3151 */ {(11<<2)|2,{49,95,0}},
/* 3152 */ {(11<<2)|2,{49,95,0}},
/* 3153 */ {(11<<2)|2,{49,95,0}},
/* 3154 */ {(11<<2)|2,{49,95,0}},
/* 3155 */ {(11<<2)|2,{49,95,0}},
/* 3156 */ {(11<<2)|2,{49,95,0}},
/* 3157 */ {(11<<2)|2,{49,95,0}},
/* 3158 */ {(11<<2)|2,{49,95,0}},
/* 3159 */ {(11<<2)|2,{49,95,0}},
/* 3160 */ {(11<<2)|2,{49,95,0}},
/* 3161 */ {(11<<2)|2,{49,95,0}},
/* 3162 */ {(11<<2)|2,{49,95,0}},
/* 3163 */ {(11<<2)|2,{49,95,0}},
/* 3164 */ {(11<<2)|2,{49,95,0}},
/* 3165 */ {(11<<2)|2,{49,95,0}},
/* 3166 */ {(11<<2)|2,{49,95,0}},
/* 3167 */ {(11<<2)|2,{49,95,0}},
/* 3168 */ {(16<<2)|3,{49,98,48}},
/* 3169 */ {(16<<2)|3,{49,98,49}},
/* 3170 */ {(16<<2)|3,{49,98,50}},
/* 3171 */ {(16<<2)|3,{49,98,97}},
/* 3172 */ {(16<<2)|3,{49,98,99}},
/* 3173 */ {(16<<2)|3,{49,98,101}},
/* 3174 */ {(16<<2)|3,{49,98,105}},
/* 3175 */ {(16<<2)|3,{49,98,111}},
/* 3176 */ {(16<<2)|3,{49,98,115}},
/* 3177 */ {(16<<2)|3,{49,98,116}},
/* 3178 */ {(11<<2)|2,{49,98,0}},
/* 3179 */ {(11<<2)|2,{49,98,0}},
/* 3180 */ {(11<<2)|2,{49,98,0}},
/* 3181 */ {(11<<2)|2,{49,98,0}},
/* 3182 */ {(11<<2)|2,{49,98,0}},
/* 3183 */ {(11<<2)|2,{49,98,0}},
/* 3184 */ {(11<<2)|2,{49,98,0}},
/* 3185 */ {(11<<2)|2,{49,98,0}},
/* 3186 */ {(11<<2)|2,{49,98,0}},
/* 3187 */ {(11<<2)|2,{49,98,0}},
/* 3188 */ {(11<<2)|2,{49,98,0}},
/* 3189 */ {(11<<2)|2,{49,98,0}},
/* 3190 */ {(11<<2)|2,{49,98,0}},
/* 3191 */ {(11<<2)|2,{49,98,0}},
/* 3192 */ {(11<<2)|2,{49,98,0}},
/* 3193 */ {(11<<2)|2,{49,98,0}},
/* 3194 */ {(11<<2)|2,{49,98,0}},
/* 3195 */ {(11<<2)|2,{49,98,0}},
/* 3196 */ {(11<<2)|2,{49,98,0}},
/* 3197 */ {(11<<2)|2,{49,98,0}},
/* 3198 */ {(11<<2)|2,{49,98,0}},
/* 3199 */ {(11<<2)|2,{49,98,0}},
/* 3200 */ {(16<<2)|3,{49,100,48}},
/* 3201 */ {(16<<2)|3,{49,100,49}},
/* 3202 */ {(16<<2)|3,{49,100,50}},
/* 3203 */ {(16<<2)|3,{49,100,97}},
/* 3204 */ {(16<<2)|3,{49,100,99}},
/* 3205 */ {(16<<2)|3,{49,100,101}},
/* 3206 */ {(16<<2)|3,{49,100,105}},
/* 3207 */ {(16<<2)|3,{49,100,111}},
/* 3208 */ {(16<<2)|3,{49,100,115}},
/* 3209 */ {(16<<2)|3,{49,100,116}},
/* 3210 */ {(11<<2)|2,{49,100,0}},
/* 3211 */ {(11<<2)|2,{49,100,0}},
/* 3212 */ {(11<<2)|2,{49,100,0}},
/* 3213 */ {(11<<2)|2,{49,100,0}},
/* 3214 */ {(11<<2)|2,{49,100,0}},
/* 3215 */ {(11<<2)|2,{49,100,0}},
/* 3216 */ {(11<<2)|2,{49,100,0}},
/* 3217 */ {(11<<2)|2,{49,100,0}},
/* 3218 */ {(11<<2)|2,{49,100,0}},
/* 3219 */ {(11<<2)|2,{49,100,0}},
/* 3220 */ {(11<<2)|2,{49,100,0}},
/* 3221 */ {(11<<2)|2,{49,100,0}},
/* 3222 */ {(11<<2)|2,{49,100,0}},
/* 3223 */ {(11<<2)|2,{49,100,0}},
/* 3224 */ {(11<<2)|2,{49,100,0}},
/* 3225 */ {(11<<2)|2,{49,100,0}},
/* 3226 */ {(11<<2)|2,{49,100,0}},
/* 3227 */ {(11<<2)|2,{49,100,0}},
/* 3228 */ {(11<<2)|2,{49,100,0}},
/* 3229 */ {(11<<2)|2,{49,100,0}},
/* 3230 */ {(11<<2)|2,{49,100,0}},
/* 3231 */ {(11<<2)|2,{49,100,0}},
/* 3232 */ {(16<<2)|3,{49,102,48}},
/* 3233 */ {(16<<2)|3,{49,102,49}},
/* 3234 */ {(16<<2)|3,{49,102,50}},
/* 3235 */ {(16<<2)|3,{49,102,97}},
/* 3236 */ {(16<<2)|3,{49,102,99}},
/* 3237 */ {(16<<2)|3,{49,102,101}},
/* 3238 */ {(16<<2)|3,{49,102,105}},
/* 3239 */ {(16<<2)|3,{49,102,111}},
/* 3240 */ {(16<<2)|3,{49,102,115}},
/* 3241 */ {(16<<2)|3,{49,102,116}},
/* 3242 */ {(11<<2)|2,{49,102,0}},
/* 3243 */ {(11<<2)|2,{49,102,0}},
/* 3244 */ {(11<<2)|2,{49,102,0}},
/* 3245 */ {(11<<2)|2,{49,102,0}},
/* 3246 */ {(11<<2)|2,{49,102,0}},
/* 3247 */ {(11<<2)|2,{49,102,0}},
/* 3248 */ {(11<<2)|2,{49,102,0}},
/* 3249 */ {(11<<2)|2,{49,102,0}},
/* 3250 */ {(11<<2)|2,{49,102,0}},
/* 3251 */ {(11<<2)|2,{49,102,0}},
/* 3252 */ {(11<<2)|2,{49,102,0}},
/* 3253 */ {(11<<2)|2,{49,102,0}},
/* 3254 */ {(11<<2)|2,{49,102,0}},
/* 3255 */ {(11<<2)|2,{49,102,0}},
/* 3256 */ {(11<<2)|2,{49,102,0}},
/* 3257 */ {(11<<2)|2,{49,102,0}},
/* 3258 */ {(11<<2)|2,{49,102,0}},
/* 3259 */ {(11<<2)|2,{49,102,0}},
/* 3260 */ {(11<<2)|2,{49,102,0}},
/* 3261 */ {(11<<2)|2,{49,102,0}},
/* 3262 */ {(11<<2)|2,{49,102,0}},
/* 3263 */ {(11<<2)|2,{49,102,0}},
/* 3264 */ {(16<<2)|3,{49,103,48}},
/* 3265 */ {(16<<2)|3,{49,103,49}},
/* 3266 */ {(16<<2)|3,{49,103,50}},
/* 3267 */ {(16<<2)|3,{49,103,97}},
/* 3268 */ {(16<<2)|3,{49,103,99}},
/* 3269 */ {(16<<2)|3,{49,103,101}},
/* 3270 */ {(16<<2)|3,{49,103,105}},
/* 3271 */ {(16<<2)|3,{49,103,111}},
/* 3272 */ {(16<<2)|3,{49,103,115}},
/* 3273 */ {(16<<2)|3,{49,103,116}},
/* 3274 */ {(11<<2)|2,{49,103,0}},
/* 3275 */ {(11<<2)|2,{49,103,0}},
/* 3276 */ {(11<<2)|2,{49,103,0}},
/* 3277 */ {(11<<2)|2,{49,103,0}},
/* 3278 */ {(11<<2)|2,{49,103,0}},
/* 3279 */ {(11<<2)|2,{49,103,0}},
/* 3280 */ {(11<<2)|2,{49,103,0}},
/* 3281 */ {(11<<2)|2,{49,103,0}},
/* 3282 */ {(11<<2)|2,{49,103,0}},
/* 3283 */ {(11<<2)|2,{49,103,0}},
/* 3284 */ {(11<<2)|2,{49,103,0}},
/* 3285 */ {(11<<2)|2,{49,103,0}},
/* 3286 */ {(11<<2)|2,{49,103,0}},
/* 3287 */ {(11<<2)|2,{49,103,0}},
/* 3288 */ {(11<<2)|2,{49,103,0}},
/* 3289 */ {(11<<2)|2,{49,103,0}},
/* 3290 */ {(11<<2)|2,{49,103,0}},
/* 3291 */ {(11<<2)|2,{49,103,0}},
/* 3292 */ {(11<<2)|2,{49,103,0}},
/* 3293 */ {(11<<2)|2,{49,103,0}},
/* 3294 */ {(11<<2)|2,{49,103,0}},
/* 3295 */ {(11<<2)|2,{49,103,0}},
/* 3296 */ {(16<<2)|3,{49,104,48}},
/* 3297 */ {(16<<2)|3,{49,104,49}},
/* 3298 */ {(16<<2)|3,{49,104,50}},
/* 3299 */ {(16<<2)|3,{49,104,97}},
/* 3300 */ {(16<<2)|3,{49,104,99}},
/* 3301 */ {(16<<2)|3,{49,104,101}},
/* 3302 */ {(16<<2)|3,{49,104,105}},
/* 3303 */ {(16<<2)|3,{49,104,111}},
/* 3304 */ {(16<<2)|3,{49,104,115}},
/* 3305 */ {(16<<2)|3,{49,104,116}},
/* 3306 */ {(11<<2)|2,{49,104,0}},
/* 3307 */ {(11<<2)|2,{49,104,0}},
/* 3308 */ {(11<<2)|2,{49,104,0}},
/* 3309 */ {(11<<2)|2,{49,104,0}},
/* 3310 */ {(11<<2)|2,{49,104,0}},
/* 3311 */ {(11<<2)|2,{49,104,0}},
/* 3312 */ {(11<<2)|2,{49,104,0}},
/* 3313 */ {(11<<2)|2,{49,104,0}},
/* 3314 */ {(11<<2)|2,{49,104,0}},
/* 3315 */ {(11<<2)|2,{49,104,0}},
/* 3316 */ {(11<<2)|2,{49,104,0}},
/* 3317 */ {(11<<2)|2,{49,104,0}},
/* 3318 */ {(11<<2)|2,{49,104,0}},
/* 3319 */ {(11<<2)|2,{49,104,0}},
/* 3320 */ {(11<<2)|2,{49,104,0}},
/* 3321 */ {(11<<2)|2,{49,104,0}},
/* 3322 */ {(11<<2)|2,{49,104,0}},
/* 3323 */ {(11<<2)|2,{49,104,0}},
/* 3324 */ {(11<<2)|2,{49,104,0}},
/* 3325 */ {(11<<2)|2,{49,104,0}},
/* 3326 */ {(11<<2)|2,{49,104,0}},
/* 3327 */ {(11<<2)|2,{49,104,0}},
/* 3328 */ {(16<<2)|3,{49,108,48}},
/* 3329 */ {(16<<2)|3,{49,108,49}},
/* 3330 */ {(16<<2)|3,{49,108,50}},
/* 3331 */ {(16<<2)|3,{49,108,97}},
/* 3332 */ {(16<<2)|3,{49,108,99}},
/* 3333 */ {(16<<2)|3,{49,108,101}},
/* 3334 */ {(16<<2)|3,{49,108,105}},
/* 3335 */ {(16<<2)|3,{49,108,111}},
/* 3336 */ {(16<<2)|3,{49,108,115}},
/* 3337 */ {(16<<2)|3,{49,108,116}},
/* 3338 */ {(11<<2)|2,{49,108,0}},
/* 3339 */ {(11<<2)|2,{49,108,0}},
/* 3340 */ {(11<<2)|2,{49,108,0}},
/* 3341 */ {(11<<2)|2,{49,108,0}},
/* 3342 */ {(11<<2)|2,{49,108,0}},
/* 3343 */ {(11<<2)|2,{49,108,0}},
/* 3344 */ {(11<<2)|2,{49,108,0}},
/* 3345 */ {(11<<2)|2,{49,108,0}},
/* 3346 */ {(11<<2)|2,{49,108,0}},
/* 3347 */ {(11<<2)|2,{49,108,0}},
/* 3348 */ {(11<<2)|2,{49,108,0}},
/* 3349 */ {(11<<2)|2,{49,108,0}},
/* 3350 */ {(11<<2)|2,{49,108,0}},
/* 3351 */ {(11<<2)|2,{49,108,0}},
/* 3352 */ {(11<<2)|2,{49,108,0}},
/* 3353 */ {(11<<2)|2,{49,108,0}},
/* 3354 */ {(11<<2)|2,{49,108,0}},
/* 3355 */ {(11<<2)|2,{49,108,0}},
/* 3356 */ {(11<<2)|2,{49,108,0}},
/* 3357 */ {(11<<2)|2,{49,108,0}},
/* 3358 */ {(11<<2)|2,{49,108,0}},
/* 3359 */ {(11<<2)|2,{49,108,0}},
/* 3360 */ {(16<<2)|3,{49,109,48}},
/* 3361 */ {(16<<2)|3,{49,109,49}},
/* 3362 */ {(16<<2)|3,{49,109,50}},
/* 3363 */ {(16<<2)|3,{49,109,97}},
/* 3364 */ {(16<<2)|3,{49,109,99}},
/* 3365 */ {(16<<2)|3,{49,109,101}},
/* 3366 */ {(16<<2)|3,{49,109,105}},
/* 3367 */ {(16<<2)|3,{49,109,111}},
/* 3368 */ {(16<<2)|3,{49,109,115}},
/* 3369 */ {(16<<2)|3,{49,109,116}},
/* 3370 */ {(11<<2)|2,{49,109,0}},
/* 3371 */ {(11<<2)|2,{49,109,0}},
/* 3372 */ {(11<<2)|2,{49,109,0}},
/* 3373 */ {(11<<2)|2,{49,109,0}},
/* 3374 */ {(11<<2)|2,{49,109,0}},
/* 3375 */ {(11<<2)|2,{49,109,0}},
/* 3376 */ {(11<<2)|2,{49,109,0}},
/* 3377 */ {(11<<2)|2,{49,109,0}},
/* 3378 */ {(11<<2)|2,{49,109,0}},
/* 3379 */ {(11<<2)|2,{49,109,0}},
/* 3380 */ {(11<<2)|2,{49,109,0}},
/* 3381 */ {(11<<2)|2,{49,109,0}},
/* 3382 */ {(11<<2)|2,{49,109,0}},
/* 3383 */ {(11<<2)|2,{49,109,0}},
/* 3384 */ {(11<<2)|2,{49,109,0}},
/* 3385 */ {(11<<2)|2,{49,109,0}},
/* 3386 */ {(11<<2)|2,{49,109,0}},
/* 3387 */ {(11<<2)|2,{49,109,0}},
/* 3388 */ {(11<<2)|2,{49,109,0}},
/* 3389 */ {(11<<2)|2,{49,109,0}},
/* 3390 */ {(11<<2)|2,{49,109,0}},
/* 3391 */ {(11<<2)|2,{49,109,0}},
/* 3392 */ {(16<<2)|3,{49,110,48}},
/* 3393 */ {(16<<2)|3,{49,110,49}},
/* 3394 */ {(16<<2)|3,{49,110,50}},
/* 3395 */ {(16<<2)|3,{49,110,97}},
/* 3396 */ {(16<<2)|3,{49,110,99}},
/* 3397 */ {(16<<2)|3,{49,110,101}},
/* 3398 */ {(16<<2)|3,{49,110,105}},
/* 3399 */ {(16<<2)|3,{49,110,111}},
/* 3400 */ {(16<<2)|3,{49,110,115}},
/* 3401 */ {(16<<2)|3,{49,110,116}},
/* 3402 */ {(11<<2)|2,{49,110,0}},
/* 3403 */ {(11<<2)|2,{49,110,0}},
/* 3404 */ {(11<<2)|2,{49,110,0}},
/* 3405 */ {(11<<2)|2,{49,110,0}},
/* 3406 */ {(11<<2)|2,{49,110,0}},
/* 3407 */ {(11<<2)|2,{49,110,0}},
/* 3408 */ {(11<<2)|2,{49,110,0}},
/* 3409 */ {(11<<2)|2,{49,110,0}},
/* 3410 */ {(11<<2)|2,{49,110,0}},
/* 3411 */ {(11<<2)|2,{49,110,0}},
/* 3412 */ {(11<<2)|2,{49,110,0}},
/* 3413 */ {(11<<2)|2,{49,110,0}},
/* 3414 */ {(11<<2)|2,{49,110,0}},
/* 3415 */ {(11<<2)|2,{49,110,0}},
/* 3416 */ {(11<<2)|2,{49,110,0}},
/* 3417 */ {(11<<2)|2,{49,110,0}},
/* 3418 */ {(11<<2)|2,{49,110,0}},
/* 3419 */ {(11<<2)|2,{49,110,0}},
/* 3420 */ {(11<<2)|2,{49,110,0}},
/* 3421 */ {(11<<2)|2,{49,110,0}},
/* 3422 */ {(11<<2)|2,{49,110,0}},
/* 3423 */ {(11<<2)|2,{49,110,0}},
/* 3424 */ {(16<<2)|3,{49,112,48}},
/* 3425 */ {(16<<2)|3,{49,112,49}},
/* 3426 */ {(16<<2)|3,{49,112,50}},
/* 3427 */ {(16<<2)|3,{49,112,97}},
/* 3428 */ {(16<<2)|3,{49,112,99}},
/* 3429 */ {(16<<2)|3,{49,112,101}},
/* 3430 */ {(16<<2)|3,{49,112,105}},
/* 3431 */ {(16<<2)|3,{49,112,111}},
/* 3432 */ {(16<<2)|3,{49,112,115}},
/* 3433 */ {(16<<2)|3,{49,112,116}},
/* 3434 */ {(11<<2)|2,{49,112,0}},
/* 3435 */ {(11<<2)|2,{49,112,0}},
/* 3436 */ {(11<<2)|2,{49,112,0}},
/* 3437 */ {(11<<2)|2,{49,112,0}},
/* 3438 */ {(11<<2)|2,{49,112,0}},
/* 3439 */ {(11<<2)|2,{49,112,0}},
/* 3440 */ {(11<<2)|2,{49,112,0}},
/* 3441 */ {(11<<2)|2,{49,112,0}},
/* 3442 */ {(11<<2)|2,{49,112,0}},
/* 3443 */ {(11<<2)|2,{49,112,0}},
/* 3444 */ {(11<<2)|2,{49,112,0}},
/* 3445 */ {(11<<2)|2,{49,112,0}},
/* 3446 */ {(11<<2)|2,{49,112,0}},
/* 3447 */ {(11<<2)|2,{49,112,0}},
/* 3448 */ {(11<<2)|2,{49,112,0}},
/* 3449 */ {(11<<2)|2,{49,112,0}},
/* 3450 */ {(11<<2)|2,{49,112,0}},
/* 3451 */ {(11<<2)|2,{49,112,0}},
/* 3452 */ {(11<<2)|2,{49,112,0}},
/* 3453 */ {(11<<2)|2,{49,112,0}},
/* 3454 */ {(11<<2)|2,{49,112,0}},
/* 3455 */ {(11<<2)|2,{49,112,0}},
/* 3456 */ {(16<<2)|3,{49,114,48}},
/* 3457 */ {(16<<2)|3,{49,114,49}},
/* 3458 */ {(16<<2)|3,{49,114,50}},
/* 3459 */ {(16<<2)|3,{49,114,97}},
/* 3460 */ {(16<<2)|3,{49,114,99}},
/* 3461 */ {(16<<2)|3,{49,114,101}},
/* 3462 */ {(16<<2)|3,{49,114,105}},
/* 3463 */ {(16<<2)|3,{49,114,111}},
/* 3464 */ {(16<<2)|3,{49,114,115}},
/* 3465 */ {(16<<2)|3,{49,114,116}},
/* 3466 */ {(11<<2)|2,{49,114,0}},
/* 3467 */ {(11<<2)|2,{49,114,0}},
/* 3468 */ {(11<<2)|2,{49,114,0}},
/* 3469 */ {(11<<2)|2,{49,114,0}},
/* 3470 */ {(11<<2)|2,{49,114,0}},
/* 3471 */ {(11<<2)|2,{49,114,0}},
/* 3472 */ {(11<<2)|2,{49,114,0}},
/* 3473 */ {(11<<2)|2,{49,114,0}},
/* 3474 */ {(11<<2)|2,{49,114,0}},
/* 3475 */ {(11<<2)|2,{49,114,0}},
/* 3476 */ {(11<<2)|2,{49,114,0}},
/* 3477 */ {(11<<2)|2,{49,114,0}},
/* 3478 */ {(11<<2)|2,{49,114,0}},
/* 3479 */ {(11<<2)|2,{49,114,0}},
/* 3480 */ {(11<<2)|2,{49,114,0}},
/* 3481 */ {(11<<2)|2,{49,114,0}},
/* 3482 */ {(11<<2)|2,{49,114,0}},
/* 3483 */ {(11<<2)|2,{49,114,0}},
/* 3484 */ {(11<<2)|2,{49,114,0}},
/* 3485 */ {(11<<2)|2,{49,114,0}},
/* 3486 */ {(11<<2)|2,{49,114,0}},
/* 3487 */ {(11<<2)|2,{49,114,0}},
/* 3488 */ {(16<<2)|3,{49,117,48}},
/* 3489 */ {(16<<2)|3,{49,117,49}},
/* 3490 */ {(16<<2)|3,{49,117,50}},
/* 3491 */ {(16<<2)|3,{49,117,97}},
/* 3492 */ {(16<<2)|3,{49,117,99}},
/* 3493 */ {(16<<2)|3,{49,117,101}},
/* 3494 */ {(16<<2)|3,{49,117,105}},
/* 3495 */ {(16<<2)|3,{49,117,111}},
/* 3496 */ {(16<<2)|3,{49,117,115}},
/* 3497 */ {(16<<2)|3,{49,117,116}},
/* 3498 */ {(11<<2)|2,{49,117,0}},
/* 3499 */ {(11<<2)|2,{49,117,0}},
/* 3500 */ {(11<<2)|2,{49,117,0}},
/* 3501 */ {(11<<2)|2,{49,117,0}},
/* 3502 */ {(11<<2)|2,{49,117,0}},
/* 3503 */ {(11<<2)|2,{49,117,0}},
/* 3504 */ {(11<<2)|2,{49,117,0}},
/* 3505 */ {(11<<2)|2,{49,117,0}},
/* 3506 */ {(11<<2)|2,{49,117,0}},
/* 3507 */ {(11<<2)|2,{49,117,0}},
/* 3508 */ {(11<<2)|2,{49,117,0}},
/* 3509 */ {(11<<2)|2,{49,117,0}},
/* 3510 */ {(11<<2)|2,{49,117,0}},
/* 3511 */ {(11<<2)|2,{49,117,0}},
/* 3512 */ {(11<<2)|2,{49,117,0}},
/* 3513 */ {(11<<2)|2,{49,117,0}},
/* 3514 */ {(11<<2)|2,{49,117,0}},
/* 3515 */ {(11<<2)|2,{49,117,0}},
/* 3516 */ {(11<<2)|2,{49,117,0}},
/* 3517 */ {(11<<2)|2,{49,117,0}},
/* 3518 */ {(11<<2)|2,{49,117,0}},
/* 3519 */ {(11<<2)|2,{49,117,0}},
/* 3520 */ {(12<<2)|2,{49,58,0}},
/* 3521 */ {(12<<2)|2,{49,58,0}},
/* 3522 */ {(12<<2)|2,{49,58,0}},
/* 3523 */ {(12<<2)|2,{49,58,0}},
/* 3524 */ {(12<<2)|2,{49,58,0}},
/* 3525 */ {(12<<2)|2,{49,58,0}},
/* 3526 */ {(12<<2)|2,{49,58,0}},
/* 3527 */ {(12<<2)|2,{49,58,0}},
/* 3528 */ {(12<<2)|2,{49,58,0}},
/* 3529 */ {(12<<2)|2,{49,58,0}},
/* 3530 */ {(12<<2)|2,{49,58,0}},
/* 3531 */ {(12<<2)|2,{49,58,0}},
/* 3532 */ {(12<<2)|2,{49,58,0}},
/* 3533 */ {(12<<2)|2,{49,58,0}},
/* 3534 */ {(12<<2)|2,{49,58,0}},
/* 3535 */ {(12<<2)|2,{49,58,0}},
/* 3536 */ {(12<<2)|2,{49,66,0}},
/* 3537 */ {(12<<2)|2,{49,66,0}},
/* 3538 */ {(12<<2)|2,{49,66,0}},
/* 3539 */ {(12<<2)|2,{49,66,0}},
/* 3540 */ {(12<<2)|2,{49,66,0}},
/* 3541 */ {(12<<2)|2,{49,66,0}},
/* 3542 */ {(12<<2)|2,{49,66,0}},
/* 3543 */ {(12<<2)|2,{49,66,0}},
/* 3544 */ {(12<<2)|2,{49,66,0}},
/* 3545 */ {(12<<2)|2,{49,66,0}},
/* 3546 */ {(12<<2)|2,{49,66,0}},
/* 3547 */ {(12<<2)|2,{49,66,0}},
/* 3548 */ {(12<<2)|2,{49,66,0}},
/* 3549 */ {(12<<2)|2,{49,66,0}},
/* 3550 */ {(12<<2)|2,{49,66,0}},
/* 3551 */ {(12<<2)|2,{49,66,0}},
/* 3552 */ {(12<<2)|2,{49,67,0}},
/* 3553 */ {(12<<2)|2,{49,67,0}},
/* 3554 */ {(12<<2)|2,{49,67,0}},
/* 3555 */ {(12<<2)|2,{49,67,0}},
/* 3556 */ {(12<<2)|2,{49,67,0}},
/* 3557 */ {(12<<2)|2,{49,67,0}},
/* 3558 */ {(12<<2)|2,{49,67,0}},
/* 3559 */ {(12<<2)|2,{49,67,0}},
/* 3560 */ {(12<<2)|2,{49,67,0}},
/* 3561 */ {(12<<2)|2,{49,67,0}},
/* 3562 */ {(12<<2)|2,{49,67,0}},
/* 3563 */ {(12<<2)|2,{49,67,0}},
/* 3564 */ {(12<<2)|2,{49,67,0}},
/* 3565 */ {(12<<2)|2,{49,67,0}},
/* 3566 */ {(12<<2)|2,{49,67,0}},
/* 3567 */ {(12<<2)|2,{49,67,0}},
/* 3568 */ {(12<<2)|2,{49,68,0}},
/* 3569 */ {(12<<2)|2,{49,68,0}},
/* 3570 */ {(12<<2)|2,{49,68,0}},
/* 3571 */ {(12<<2)|2,{49,68,0}},
/* 3572 */ {(12<<2)|2,{49,68,0}},
/* 3573 */ {(12<<2)|2,{49,68,0}},
/* 3574 */ {(12<<2)|2,{49,68,0}},
/* 3575 */ {(12<<2)|2,{49,68,0}},
/* 3576 */ {(12<<2)|2,{49,68,0}},
/* 3577 */ {(12<<2)|2,{49,68,0}},
/* 3578 */ {(12<<2)|2,{49,68,0}},
/* 3579 */ {(12<<2)|2,{49,68,0}},
/* 3580 */ {(12<<2)|2,{49,68,0}},
/* 3581 */ {(12<<2)|2,{49,68,0}},
/* 3582 */ {(12<<2)|2,{49,68,0}},
/* 3583 */ {(12<<2)|2,{49,68,0}},
/* 3584 */ {(12<<2)|2,{49,69,0}},
/* 3585 */ {(12<<2)|2,{49,69,0}},
/* 3586 */ {(12<<2)|2,{49,69,0}},
/* 3587 */ {(12<<2)|2,{49,69,0}},
/* 3588 */ {(12<<2)|2,{49,69,0}},
/* 3589 */ {(12<<2)|2,{49,69,0}},
/* 3590 */ {(12<<2)|2,{49,69,0}},
/* 3591 */ {(12<<2)|2,{49,69,0}},
/* 3592 */ {(12<<2)|2,{49,69,0}},
/* 3593 */ {(12<<2)|2,{49,69,0}},
/* 3594 */ {(12<<2)|2,{49,69,0}},
/* 3595 */ {(12<<2)|2,{49,69,0}},
/* 3596 */ {(12<<2)|2,{49,69,0}},
/* 3597 */ {(12<<2)|2,{49,69,0}},
/* 3598 */ {(12<<2)|2,{49,69,0}},
/* 3599 */ {(12<<2)|2,{49,69,0}},
/* 3600 */ {(12<<2)|2,{49,70,0}},
/* 3601 */ {(12<<2)|2,{49,70,0}},
/* 3602 */ {(12<<2)|2,{49,70,0}},
/* 3603 */ {(12<<2)|2,{49,70,0}},
/* 3604 */ {(12<<2)|2,{49,70,0}},
/* 3605 */ {(12<<2)|2,{49,70,0}},
/* 3606 */ {(12<<2)|2,{49,70,0}},
/* 3607 */ {(12<<2)|2,{49,70,0}},
/* 3608 */ {(12<<2)|2,{49,70,0}},
/* 3609 */ {(12<<2)|2,{49,70,0}},
/* 3610 */ {(12<<2)|2,{49,70,0}},
/* 3611 */ {(12<<2)|2,{49,70,0}},
/* 3612 */ {(12<<2)|2,{49,70,0}},
/* 3613 */ {(12<<2)|2,{49,70,0}},
/* 3614 */ {(12<<2)|2,{49,70,0}},
/* 3615 */ {(12<<2)|2,{49,70,0}},
/* 3616 */ {(12<<2)|2,{49,71,0}},
/* 3617 */ {(12<<2)|2,{49,71,0}},
/* 3618 */ {(12<<2)|2,{49,71,0}},
/* 3619 */ {(12<<2)|2,{49,71,0}},
/* 3620 */ {(12<<2)|2,{49,71,0}},
/* 3621 */ {(12<<2)|2,{49,71,0}},
/* 3622 */ {(12<<2)|2,{49,71,0}},
/* 3623 */ {(12<<2)|2,{49,71,0}},
/* 3624 */ {(12<<2)|2,{49,71,0}},
/* 3625 */ {(12<<2)|2,{49,71,0}},
/* 3626 */ {(12<<2)|2,{49,71,0}},
/* 3627 */ {(12<<2)|2,{49,71,0}},
/* 3628 */ {(12<<2)|2,{49,71,0}},
/* 3629 */ {(12<<2)|2,{49,71,0}},
/* 3630 */ {(12<<2)|2,{49,71,0}},
/* 3631 */ {(12<<2)|2,{49,71,0}},
/* 3632 */ {(12<<2)|2,{49,72,0}},
/* 3633 */ {(12<<2)|2,{49,72,0}},
/* 3634 */ {(12<<2)|2,{49,72,0}},
/* 3635 */ {(12<<2)|2,{49,72,0}},
/* 3636 */ {(12<<2)|2,{49,72,0}},
/* 3637 */ {(12<<2)|2,{49,72,0}},
/* 3638 */ {(12<<2)|2,{49,72,0}},
/* 3639 */ {(12<<2)|2,{49,72,0}},
/* 3640 */ {(12<<2)|2,{49,72,0}},
/* 3641 */ {(12<<2)|2,{49,72,0}},
/* 3642 */ {(12<<2)|2,{49,72,0}},
/* 3643 */ {(12<<2)|2,{49,72,0}},
/* 3644 */ {(12<<2)|2,{49,72,0}},
/* 3645 */ {(12<<2)|2,{49,72,0}},
/* 3646 */ {(12<<2)|2,{49,72,0}},
/* 3647 */ {(12<<2)|2,{49,72,0}},
/* 3648 */ {(12<<2)|2,{49,73,0}},
/* 3649 */ {(12<<2)|2,{49,73,0}},
/* 3650 */ {(12<<2)|2,{49,73,0}},
/* 3651 */ {(12<<2)|2,{49,73,0}},
/* 3652 */ {(12<<2)|2,{49,73,0}},
/* 3653 */ {(12<<2)|2,{49,73,0}},
/* 3654 */ {(12<<2)|2,{49,73,0}},
/* 3655 */ {(12<<2)|2,{49,73,0}},
/* 3656 */ {(12<<2)|2,{49,73,0}},
/* 3657 */ {(12<<2)|2,{49,73,0}},
/* 3658 */ {(12<<2)|2,{49,73,0}},
/* 3659 */ {(12<<2)|2,{49,73,0}},
/* 3660 */ {(12<<2)|2,{49,73,0}},
/* 3661 */ {(12<<2)|2,{49,73,0}},
/* 3662 */ {(12<<2)|2,{49,73,0}},
/* 3663 */ {(12<<2)|2,{49,73,0}},
/* 3664 */ {(12<<2)|2,{49,74,0}},
/* 3665 */ {(12<<2)|2,{49,74,0}},
/* 3666 */ {(12<<2)|2,{49,74,0}},
/* 3667 */ {(12<<2)|2,{49,74,0}},
/* 3668 */ {(12<<2)|2,{49,74,0}},
/* 3669 */ {(12<<2)|2,{49,74,0}},
/* 3670 */ {(12<<2)|2,{49,74,0}},
/* 3671 */ {(12<<2)|2,{49,74,0}},
/* 3672 */ {(12<<2)|2,{49,74,0}},
/* 3673 */ {(12<<2)|2,{49,74,0}},
/* 3674 */ {(12<<2)|2,{49,74,0}},
/* 3675 */ {(12<<2)|2,{49,74,0}},
/* 3676 */ {(12<<2)|2,{49,74,0}},
/* 3677 */ {(12<<2)|2,{49,74,0}},
/* 3678 */ {(12<<2)|2,{49,74,0}},
/* 3679 */ {(12<<2)|2,{49,74,0}},
/* 3680 */ {(12<<2)|2,{49,75,0}},
/* 3681 */ {(12<<2)|2,{49,75,0}},
/* 3682 */ {(12<<2)|2,{49,75,0}},
/* 3683 */ {(12<<2)|2,{49,75,0}},
/* 3684 */ {(12<<2)|2,{49,75,0}},
/* 3685 */ {(12<<2)|2,{49,75,0}},
/* 3686 */ {(12<<2)|2,{49,75,0}},
/* 3687 */ {(12<<2)|2,{49,75,0}},
/* 3688 */ {(12<<2)|2,{49,75,0}},
/* 3689 */ {(12<<2)|2,{49,75,0}},
/* 3690 */ {(12<<2)|2,{49,75,0}},
/* 3691 */ {(12<<2)|2,{49,75,0}},
/* 3692 */ {(12<<2)|2,{49,75,0}},
/* 3693 */ {(12<<2)|2,{49,75,0}},
/* 3694 */ {(12<<2)|2,{49,75,0}},
/* 3695 */ {(12<<2)|2,{49,75,0}},
/* 3696 */ {(12<<2)|2,{49,76,0}},
/* 3697 */ {(12<<2)|2,{49,76,0}},
/* 3698 */ {(12<<2)|2,{49,76,0}},
/* 3699 */ {(12<<2)|2,{49,76,0}},
/* 3700 */ {(12<<2)|2,{49,76,0}},
/* 3701 */ {(12<<2)|2,{49,76,0}},
/* 3702 */ {(12<<2)|2,{49,76,0}},
/* 3703 */ {(12<<2)|2,{49,76,0}},
/* 3704 */ {(12<<2)|2,{49,76,0}},
/* 3705 */ {(12<<2)|2,{49,76,0}},
/* 3706 */ {(12<<2)|2,{49,76,0}},
/* 3707 */ {(12<<2)|2,{49,76,0}},
/* 3708 */ {(12<<2)|2,{49,76,0}},
/* 3709 */ {(12<<2)|2,{49,76,0}},
/* 3710 */ {(12<<2)|2,{49,76,0}},
/* 3711 */ {(12<<2)|2,{49,76,0}},
/* 3712 */ {(12<<2)|2,{49,77,0}},
/* 3713 */ {(12<<2)|2,{49,77,0}},
/* 3714 */ {(12<<2)|2,{49,77,0}},
/* 3715 */ {(12<<2)|2,{49,77,0}},
/* 3716 */ {(12<<2)|2,{49,77,0}},
/* 3717 */ {(12<<2)|2,{49,77,0}},
/* 3718 */ {(12<<2)|2,{49,77,0}},
/* 3719 */ {(12<<2)|2,{49,77,0}},
/* 3720 */ {(12<<2)|2,{49,77,0}},
/* 3721 */ {(12<<2)|2,{49,77,0}},
/* 3722 */ {(12<<2)|2,{49,77,0}},
/* 3723 */ {(12<<2)|2,{49,77,0}},
/* 3724 */ {(12<<2)|2,{49,77,0}},
/* 3725 */ {(12<<2)|2,{49,77,0}},
/* 3726 */ {(12<<2)|2,{49,77,0}},
/* 3727 */ {(12<<2)|2,{49,77,0}},
/* 3728 */ {(12<<2)|2,{49,78,0}},
/* 3729 */ {(12<<2)|2,{49,78,0}},
/* 3730 */ {(12<<2)|2,{49,78,0}},
/* 3731 */ {(12<<2)|2,{49,78,0}},
/* 3732 */ {(12<<2)|2,{49,78,0}},
/* 3733 */ {(12<<2)|2,{49,78,0}},
/* 3734 */ {(12<<2)|2,{49,78,0}},
/* 3735 */ {(12<<2)|2,{49,78,0}},
/* 3736 */ {(12<<2)|2,{49,78,0}},
/* 3737 */ {(12<<2)|2,{49,78,0}},
/* 3738 */ {(12<<2)|2,{49,78,0}},
/* 3739 */ {(12<<2)|2,{49,78,0}},
/* 3740 */ {(12<<2)|2,{49,78,0}},
/* 3741 */ {(12<<2)|2,{49,78,0}},
/* 3742 */ {(12<<2)|2,{49,78,0}},
/* 3743 */ {(12<<2)|2,{49,78,0}},
/* 3744 */ {(12<<2)|2,{49,79,0}},
/* 3745 */ {(12<<2)|2,{49,79,0}},
/* 3746 */ {(12<<2)|2,{49,79,0}},
/* 3747 */ {(12<<2)|2,{49,79,0}},
/* 3748 */ {(12<<2)|2,{49,79,0}},
/* 3749 */ {(12<<2)|2,{49,79,0}},
/* 3750 */ {(12<<2)|2,{49,79,0}},
/* 3751 */ {(12<<2)|2,{49,79,0}},
/* 3752 */ {(12<<2)|2,{49,79,0}},
/* 3753 */ {(12<<2)|2,{49,79,0}},
/* 3754 */ {(12<<2)|2,{49,79,0}},
/* 3755 */ {(12<<2)|2,{49,79,0}},
/* 3756 */ {(12<<2)|2,{49,79,0}},
/* 3757 */ {(12<<2)|2,{49,79,0}},
/* 3758 */ {(12<<2)|2,{49,79,0}},
/* 3759 */ {(12<<2)|2,{49,79,0}},
/* 3760 */ {(12<<2)|2,{49,80,0}},
/* 3761 */ {(12<<2)|2,{49,80,0}},
/* 3762 */ {(12<<2)|2,{49,80,0}},
/* 3763 */ {(12<<2)|2,{49,80,0}},
/* 3764 */ {(12<<2)|2,{49,80,0}},
/* 3765 */ {(12<<2)|2,{49,80,0}},
/* 3766 */ {(12<<2)|2,{49,80,0}},
/* 3767 */ {(12<<2)|2,{49,80,0}},
/* 3768 */ {(12<<2)|2,{49,80,0}},
/* 3769 */ {(12<<2)|2,{49,80,0}},
/* 3770 */ {(12<<2)|2,{49,80,0}},
/* 3771 */ {(12<<2)|2,{49,80,0}},
/* 3772 */ {(12<<2)|2,{49,80,0}},
/* 3773 */ {(12<<2)|2,{49,80,0}},
/* 3774 */ {(12<<2)|2,{49,80,0}},
/* 3775 */ {(12<<2)|2,{49,80,0}},
/* 3776 */ {(12<<2)|2,{49,81,0}},
/* 3777 */ {(12<<2)|2,{49,81,0}},
/* 3778 */ {(12<<2)|2,{49,81,0}},
/* 3779 */ {(12<<2)|2,{49,81,0}},
/* 3780 */ {(12<<2)|2,{49,81,0}},
/* 3781 */ {(12<<2)|2,{49,81,0}},
/* 3782 */ {(12<<2)|2,{49,81,0}},
/* 3783 */ {(12<<2)|2,{49,81,0}},
/* 3784 */ {(12<<2)|2,{49,81,0}},
/* 3785 */ {(12<<2)|2,{49,81,0}},
/* 3786 */ {(12<<2)|2,{49,81,0}},
/* 3787 */ {(12<<2)|2,{49,81,0}},
/* 3788 */ {(12<<2)|2,{49,81,0}},
/* 3789 */ {(12<<2)|2,{49,81,0}},
/* 3790 */ {(12<<2)|2,{49,81,0}},
/* 3791 */ {(12<<2)|2,{49,81,0}},
/* 3792 */ {(12<<2)|2,{49,82,0}},
/* 3793 */ {(12<<2)|2,{49,82,0}},
/* 3794 */ {(12<<2)|2,{49,82,0}},
/* 3795 */ {(12<<2)|2,{49,82,0}},
/* 3796 */ {(12<<2)|2,{49,82,0}},
/* 3797 */ {(12<<2)|2,{49,82,0}},
/* 3798 */ {(12<<2)|2,{49,82,0}},
/* 3799 */ {(12<<2)|2,{49,82,0}},
/* 3800 */ {(12<<2)|2,{49,82,0}},
/* 3801 */ {(12<<2)|2,{49,82,0}},
/* 3802 */ {(12<<2)|2,{49,82,0}},
/* 3803 */ {(12<<2)|2,{49,82,0}},
/* 3804 */ {(12<<2)|2,{49,82,0}},
/* 3805 */ {(12<<2)|2,{49,82,0}},
/* 3806 */ {(12<<2)|2,{49,82,0}},
/* 3807 */ {(12<<2)|2,{49,82,0}},
/* 3808 */ {(12<<2)|2,{49,83,0}},
/* 3809 */ {(12<<2)|2,{49,83,0}},
/* 3810 */ {(12<<2)|2,{49,83,0}},
/* 3811 */ {(12<<2)|2,{49,83,0}},
/* 3812 */ {(12<<2)|2,{49,83,0}},
/* 3813 */ {(12<<2)|2,{49,83,0}},
/* 3814 */ {(12<<2)|2,{49,83,0}},
/* 3815 */ {(12<<2)|2,{49,83,0}},
/* 3816 */ {(12<<2)|2,{49,83,0}},
/* 3817 */ {(12<<2)|2,{49,83,0}},
/* 3818 */ {(12<<2)|2,{49,83,0}},
/* 3819 */ {(12<<2)|2,{49,83,0}},
/* 3820 */ {(12<<2)|2,{49,83,0}},
/* 3821 */ {(12<<2)|2,{49,83,0}},
/* 3822 */ {(12<<2)|2,{49,83,0}},
/* 3823 */ {(12<<2)|2,{49,83,0}},
/* 3824 */ {(12<<2)|2,{49,84,0}},
/* 3825 */ {(12<<2)|2,{49,84,0}},
/* 3826 */ {(12<<2)|2,{49,84,0}},
/* 3827 */ {(12<<2)|2,{49,84,0}},
/* 3828 */ {(12<<2)|2,{49,84,0}},
/* 3829 */ {(12<<2)|2,{49,84,0}},
/* 3830 */ {(12<<2)|2,{49,84,0}},
/* 3831 */ {(12<<2)|2,{49,84,0}},
/* 3832 */ {(12<<2)|2,{49,84,0}},
/* 3833 */ {(12<<2)|2,{49,84,0}},
/* 3834 */ {(12<<2)|2,{49,84,0}},
/* 3835 */ {(12<<2)|2,{49,84,0}},
/* 3836 */ {(12<<2)|2,{49,84,0}},
/* 3837 */ {(12<<2)|2,{49,84,0}},
/* 3838 */ {(12<<2)|2,{49,84,0}},
/* 3839 */ {(12<<2)|2,{49,84,0}},
/* 3840 */ {(12<<2)|2,{49,85,0}},
/* 3841 */ {(12<<2)|2,{49,85,0}},
/* 3842 */ {(12<<2)|2,{49,85,0}},
/* 3843 */ {(12<<2)|2,{49,85,0}},
/* 3844 */ {(12<<2)|2,{49,85,0}},
/* 3845 */ {(12<<2)|2,{49,85,0}},
/* 3846 */ {(12<<2)|2,{49,85,0}},
/* 3847 */ {(12<<2)|2,{49,85,0}},
/* 3848 */ {(12<<2)|2,{49,85,0}},
/* 3849 */ {(12<<2)|2,{49,85,0}},
/* 3850 */ {(12<<2)|2,{49,85,0}},
/* 3851 */ {(12<<2)|2,{49,85,0}},
/* 3852 */ {(12<<2)|2,{49,85,0}},
/* 3853 */ {(12<<2)|2,{49,85,0}},
/* 3854 */ {(12<<2)|2,{49,85,0}},
/* 3855 */ {(12<<2)|2,{49,85,0}},
/* 3856 */ {(12<<2)|2,{49,86,0}},
/* 3857 */ {(12<<2)|2,{49,86,0}},
/* 3858 */ {(12<<2)|2,{49,86,0}},
/* 3859 */ {(12<<2)|2,{49,86,0}},
/* 3860 */ {(12<<2)|2,{49,86,0}},
/* 3861 */ {(12<<2)|2,{49,86,0}},
/* 3862 */ {(12<<2)|2,{49,86,0}},
/* 3863 */ {(12<<2)|2,{49,86,0}},
/* 3864 */ {(12<<2)|2,{49,86,0}},
/* 3865 */ {(12<<2)|2,{49,86,0}},
/* 3866 */ {(12<<2)|2,{49,86,0}},
/* 3867 */ {(12<<2)|2,{49,86,0}},
/* 3868 */ {(12<<2)|2,{49,86,0}},
/* 3869 */ {(12<<2)|2,{49,86,0}},
/* 3870 */ {(12<<2)|2,{49,86,0}},
/* 3871 */ {(12<<2)|2,{49,86,0}},
/* 3872 */ {(12<<2)|2,{49,87,0}},
/* 3873 */ {(12<<2)|2,{49,87,0}},
/* 3874 */ {(12<<2)|2,{49,87,0}},
/* 3875 */ {(12<<2)|2,{49,87,0}},
/* 3876 */ {(12<<2)|2,{49,87,0}},
/* 3877 */ {(12<<2)|2,{49,87,0}},
/* 3878 */ {(12<<2)|2,{49,87,0}},
/* 3879 */ {(12<<2)|2,{49,87,0}},
/* 3880 */ {(12<<2)|2,{49,87,0}},
/* 3881 */ {(12<<2)|2,{49,87,0}},
/* 3882 */ {(12<<2)|2,{49,87,0}},
/* 3883 */ {(12<<2)|2,{49,87,0}},
/* 3884 */ {(12<<2)|2,{49,87,0}},
/* 3885 */ {(12<<2)|2,{49,87,0}},
/* 3886 */ {(12<<2)|2,{49,87,0}},
/* 3887 */ {(12<<2)|2,{49,87,0}},
/* 3888 */ {(12<<2)|2,{49,89,0}},
/* 3889 */ {(12<<2)|2,{49,89,0}},
/* 3890 */ {(12<<2)|2,{49,89,0}},
/* 3891 */ {(12<<2)|2,{49,89,0}},
/* 3892 */ {(12<<2)|2,{49,89,0}},
/* 3893 */ {(12<<2)|2,{49,89,0}},
/* 3894 */ {(12<<2)|2,{49,89,0}},
/* 3895 */ {(12<<2)|2,{49,89,0}},
/* 3896 */ {(12<<2)|2,{49,89,0}},
/* 3897 */ {(12<<2)|2,{49,89,0}},
/* 3898 */ {(12<<2)|2,{49,89,0}},
/* 3899 */ {(12<<2)|2,{49,89,0}},
/* 3900 */ {(12<<2)|2,{49,89,0}},
/* 3901 */ {(12<<2)|2,{49,89,0}},
/* 3902 */ {(12<<2)|2,{49,89,0}},
/* 3903 */ {(12<<2)|2,{49,89,0}},
/* 3904 */ {(12<<2)|2,{49,106,0}},
/* 3905 */ {(12<<2)|2,{49,106,0}},
/* 3906 */ {(12<<2)|2,{49,106,0}},
/* 3907 */ {(12<<2)|2,{49,106,0}},
/* 3908 */ {(12<<2)|2,{49,106,0}},
/* 3909 */ {(12<<2)|2,{49,106,0}},
/* 3910 */ {(12<<2)|2,{49,106,0}},
/* 3911 */ {(12<<2)|2,{49,106,0}},
/* 3912 */ {(12<<2)|2,{49,106,0}},
/* 3913 */ {(12<<2)|2,{49,106,0}},
/* 3914 */ {(12<<2)|2,{49,106,0}},
/* 3915 */ {(12<<2)|2,{49,106,0}},
/* 3916 */ {(12<<2)|2,{49,106,0}},
/* 3917 */ {(12<<2)|2,{49,106,0}},
/* 3918 */ {(12<<2)|2,{49,106,0}},
/* 3919 */ {(12<<2)|2,{49,106,0}},
/* 3920 */ {(12<<2)|2,{49,107,0}},
/* 3921 */ {(12<<2)|2,{49,107,0}},
/* 3922 */ {(12<<2)|2,{49,107,0}},
/* 3923 */ {(12<<2)|2,{49,107,0}},
/* 3924 */ {(12<<2)|2,{49,107,0}},
/* 3925 */ {(12<<2)|2,{49,107,0}},
/* 3926 */ {(12<<2)|2,{49,107,0}},
/* 3927 */ {(12<<2)|2,{49,107,0}},
/* 3928 */ {(12<<2)|2,{49,107,0}},
/* 3929 */ {(12<<2)|2,{49,107,0}},
/* 3930 */ {(12<<2)|2,{49,107,0}},
/* 3931 */ {(12<<2)|2,{49,107,0}},
/* 3932 */ {(12<<2)|2,{49,107,0}},
/* 3933 */ {(12<<2)|2,{49,107,0}},
/* 3934 */ {(12<<2)|2,{49,107,0}},
/* 3935 */ {(12<<2)|2,{49,107,0}},
/* 3936 */ {(12<<2)|2,{49,113,0}},
/* 3937 */ {(12<<2)|2,{49,113,0}},
/* 3938 */ {(12<<2)|2,{49,113,0}},
/* 3939 */ {(12<<2)|2,{49,113,0}},
/* 3940 */ {(12<<2)|2,{49,113,0}},
/* 3941 */ {(12<<2)|2,{49,113,0}},
/* 3942 */ {(12<<2)|2,{49,113,0}},
/* 3943 */ {(12<<2)|2,{49,113,0}},
/* 3944 */ {(12<<2)|2,{49,113,0}},
/* 3945 */ {(12<<2)|2,{49,113,0}},
/* 3946 */ {(12<<2)|2,{49,113,0}},
/* 3947 */ {(12<<2)|2,{49,113,0}},
/* 3948 */ {(12<<2)|2,{49,113,0}},
/* 3949 */ {(12<<2)|2,{49,113,0}},
/* 3950 */ {(12<<2)|2,{49,113,0}},
/* 3951 */ {(12<<2)|2,{49,113,0}},
/* 3952 */ {(12<<2)|2,{49,118,0}},
/* 3953 */ {(12<<2)|2,{49,118,0}},
/* 3954 */ {(12<<2)|2,{49,118,0}},
/* 3955 */ {(12<<2)|2,{49,118,0}},
/* 3956 */ {(12<<2)|2,{49,118,0}},
/* 3957 */ {(12<<2)|2,{49,118,0}},
/* 3958 */ {(12<<2)|2,{49,118,0}},
/* 3959 */ {(12<<2)|2,{49,118,0}},
/* 3960 */ {(12<<2)|2,{49,118,0}},
/* 3961 */ {(12<<2)|2,{49,118,0}},
/* 3962 */ {(12<<2)|2,{49,118,0}},
/* 3963 */ {(12<<2)|2,{49,118,0}},
/* 3964 */ {(12<<2)|2,{49,118,0}},
/* 3965 */ {(12<<2)|2,{49,118,0}},
/* 3966 */ {(12<<2)|2,{49,118,0}},
/* 3967 */ {(12<<2)|2,{49,118,0}},
/* 3968 */ {(12<<2)|2,{49,119,0}},
/* 3969 */ {(12<<2)|2,{49,119,0}},
/* 3970 */ {(12<<2)|2,{49,119,0}},
/* 3971 */ {(12<<2)|2,{49,119,0}},
/* 3972 */ {(12<<2)|2,{49,119,0}},
/* 3973 */ {(12<<2)|2,{49,119,0}},
/* 3974 */ {(12<<2)|2,{49,119,0}},
/* 3975 */ {(12<<2)|2,{49,119,0}},
/* 3976 */ {(12<<2)|2,{49,119,0}},
/* 3977 */ {(12<<2)|2,{49,119,0}},
/* 3978 */ {(12<<2)|2,{49,119,0}},
/* 3979 */ {(12<<2)|2,{49,119,0}},
/* 3980 */ {(12<<2)|2,{49,119,0}},
/* 3981 */ {(12<<2)|2,{49,119,0}},
/* 3982 */ {(12<<2)|2,{49,119,0}},
/* 3983 */ {(12<<2)|2,{49,119,0}},
/* 3984 */ {(12<<2)|2,{49,120,0}},
/* 3985 */ {(12<<2)|2,{49,120,0}},
/* 3986 */ {(12<<2)|2,{49,120,0}},
/* 3987 */ {(12<<2)|2,{49,120,0}},
/* 3988 */ {(12<<2)|2,{49,120,0}},
/* 3989 */ {(12<<2)|2,{49,120,0}},
/* 3990 */ {(12<<2)|2,{49,120,0}},
/* 3991 */ {(12<<2)|2,{49,120,0}},
/* 3992 */ {(12<<2)|2,{49,120,0}},
/* 3993 */ {(12<<2)|2,{49,120,0}},
/* 3994 */ {(12<<2)|2,{49,120,0}},
/* 3995 */ {(12<<2)|2,{49,120,0}},
/* 3996 */ {(12<<2)|2,{49,120,0}},
/* 3997 */ {(12<<2)|2,{49,120,0}},
/* 3998 */ {(12<<2)|2,{49,120,0}},
/* 3999 */ {(12<<2)|2,{49,120,0}},
/* 4000 */ {(12<<2)|2,{49,121,0}},
/* 4001 */ {(12<<2)|2,{49,121,0}},
/* 4002 */ {(12<<2)|2,{49,121,0}},
/* 4003 */ {(12<<2)|2,{49,121,0}},
/* 4004 */ {(12<<2)|2,{49,121,0}},
/* 4005 */ {(12<<2)|2,{49,121,0}},
/* 4006 */ {(12<<2)|2,{49,121,0}},
/* 4007 */ {(12<<2)|2,{49,121,0}},
/* 4008 */ {(12<<2)|2,{49,121,0}},
/* 4009 */ {(12<<2)|2,{49,121,0}},
/* 4010 */ {(12<<2)|2,{49,121,0}},
/* 4011 */ {(12<<2)|2,{49,121,0}},
/* 4012 */ {(12<<2)|2,{49,121,0}},
/* 4013 */ {(12<<2)|2,{49,121,0}},
/* 4014 */ {(12<<2)|2,{49,121,0}},
/* 4015 */ {(12<<2)|2,{49,121,0}},
/* 4016 */ {(12<<2)|2,{49,122,0}},
/* 4017 */ {(12<<2)|2,{49,122,0}},
/* 4018 */ {(12<<2)|2,{49,122,0}},
/* 4019 */ {(12<<2)|2,{49,122,0}},
/* 4020 */ {(12<<2)|2,{49,122,0}},
/* 4021 */ {(12<<2)|2,{49,122,0}},
/* 4022 */ {(12<<2)|2,{49,122,0}},
/* 4023 */ {(12<<2)|2,{49,122,0}},
/* 4024 */ {(12<<2)|2,{49,122,0}},
/* 4025 */ {(12<<2)|2,{49,122,0}},
/* 4026 */ {(12<<2)|2,{49,122,0}},
/* 4027 */ {(12<<2)|2,{49,122,0}},
/* 4028 */ {(12<<2)|2,{49,122,0}},
/* 4029 */ {(12<<2)|2,{49,122,0}},
/* 4030 */ {(12<<2)|2,{49,122,0}},
/* 4031 */ {(12<<2)|2,{49,122,0}},
/* 4032 */ {(13<<2)|2,{49,38,0}},
/* 4033 */ {(13<<2)|2,{49,38,0}},
/* 4034 */ {(13<<2)|2,{49,38,0}},
/* 4035 */ {(13<<2)|2,{49,38,0}},
/* 4036 */ {(13<<2)|2,{49,38,0}},
/* 4037 */ {(13<<2)|2,{49,38,0}},
/* 4038 */ {(13<<2)|2,{49,38,0}},
/* 4039 */ {(13<<2)|2,{49,38,0}},
/* 4040 */ {(13<<2)|2,{49,42,0}},
/* 4041 */ {(13<<2)|2,{49,42,0}},
/* 4042 */ {(13<<2)|2,{49,42,0}},
/* 4043 */ {(13<<2)|2,{49,42,0}},
/* 4044 */ {(13<<2)|2,{49,42,0}},
/* 4045 */ {(13<<2)|2,{49,42,0}},
/* 4046 */ {(13<<2)|2,{49,42,0}},
/* 4047 */ {(13<<2)|2,{49,42,0}},
/* 4048 */ {(13<<2)|2,{49,44,0}},
/* 4049 */ {(13<<2)|2,{49,44,0}},
/* 4050 */ {(13<<2)|2,{49,44,0}},
/* 4051 */ {(13<<2)|2,{49,44,0}},
/* 4052 */ {(13<<2)|2,{49,44,0}},
/* 4053 */ {(13<<2)|2,{49,44,0}},
/* 4054 */ {(13<<2)|2,{49,44,0}},
/* 4055 */ {(13<<2)|2,{49,44,0}},
/* 4056 */ {(13<<2)|2,{49,59,0}},
/* 4057 */ {(13<<2)|2,{49,59,0}},
/* 4058 */ {(13<<2)|2,{49,59,0}},
/* 4059 */ {(13<<2)|2,{49,59,0}},
/* 4060 */ {(13<<2)|2,{49,59,0}},
/* 4061 */ {(13<<2)|2,{49,59,0}},
/* 4062 */ {(13<<2)|2,{49,59,0}},
/* 4063 */ {(13<<2)|2,{49,59,0}},
/* 4064 */ {(13<<2)|2,{49,88,0}},
/* 4065 */ {(13<<2)|2,{49,88,0}},
/* 4066 */ {(13<<2)|2,{49,88,0}},
/* 4067 */ {(13<<2)|2,{49,88,0}},
/* 4068 */ {(13<<2)|2,{49,88,0}},
/* 4069 */ {(13<<2)|2,{49,88,0}},
/* 4070 */ {(13<<2)|2,{49,88,0}},
/* 4071 */ {(13<<2)|2,{49,88,0}},
/* 4072 */ {(13<<2)|2,{49,90,0}},
/* 4073 */ {(13<<2)|2,{49,90,0}},
/* 4074 */ {(13<<2)|2,{49,90,0}},
/* 4075 */ {(13<<2)|2,{49,90,0}},
/* 4076 */ {(13<<2)|2,{49,90,0}},
/* 4077 */ {(13<<2)|2,{49,90,0}},
/* 4078 */ {(13<<2)|2,{49,90,0}},
/* 4079 */ {(13<<2)|2,{49,90,0}},
/* 4080 */ {(15<<2)|2,{49,33,0}},
/* 4081 */ {(15<<2)|2,{49,33,0}},
/* 4082 */ {(15<<2)|2,{49,34,0}},
/* 4083 */ {(15<<2)|2,{49,34,0}},
/* 4084 */ {(15<<2)|2,{49,40,0}},
/* 4085 */ {(15<<2)|2,{49,40,0}},
/* 4086 */ {(15<<2)|2,{49,41,0}},
/* 4087 */ {(15<<2)|2,{49,41,0}},
/* 4088 */ {(15<<2)|2,{49,63,0}},
/* 4089 */ {(15<<2)|2,{49,63,0}},
/* 4090 */ {(16<<2)|2,{49,39,0}},
/* 4091 */ {(16<<2)|2,{49,43,0}},
/* 4092 */ {(16<<2)|2,{49,124,0}},
/* 4093 */ {(5<<2)|1,{49,0,0}},
/* 4094 */ {(5<<2)|1,{49,0,0}},
/* 4095 */ {(5<<2)|1,{49,0,0}},
/* 4096 */ {(15<<2)|3,{50,48,48}},
/* 4097 */ {(15<<2)|3,{50,48,48}},
/* 4098 */ {(15<<2)|3,{50,48,49}},
/* 4099 */ {(15<<2)|3,{50,48,49}},
/* 4100 */ {(15<<2)|3,{50,48,50}},
/* 4101 */ {(15<<2)|3,{50,48,50}},
/* 4102 */ {(15<<2)|3,{50,48,97}},
/* 4103 */ {(15<<2)|3,{50,48,97}},
/* 4104 */ {(15<<2)|3,{50,48,99}},
/* 4105 */ {(15<<2)|3,{50,48,99}},
/* 4106 */ {(15<<2)|3,{50,48,101}},
/* 4107 */ {(15<<2)|3,{50,48,101}},
/* 4108 */ {(15<<2)|3,{50,48,105}},
/* 4109 */ {(15<<2)|3,{50,48,105}},
/* 4110 */ {(15<<2)|3,{50,48,111}},
/* 4111 */ {(15<<2)|3,{50,48,111}},
/* 4112 */ {(15<<2)|3,{50,48,115}},
/* 4113 */ {(15<<2)|3,{50,48,115}},
/* 4114 */ {(15<<2)|3,{50,48,116}},
/* 4115 */ {(15<<2)|3,{50,48,116}},
/* 4116 */ {(16<<2)|3,{50,48,32}},
/* 4117 */ {(16<<2)|3,{50,48,37}},
/* 4118 */ {(16<<2)|3,{50,48,45}},
/* 4119 */ {(16<<2)|3,{50,48,46}},
/* 4120 */ {(16<<2)|3,{50,48,47}},
/* 4121 */ {(16<<2)|3,{50,48,51}},
/* 4122 */ {(16<<2)|3,{50,48,52}},
/* 4123 */ {(16<<2)|3,{50,48,53}},
/* 4124 */ {(16<<2)|3,{50,48,54}},
/* 4125 */ {(16<<2)|3,{50,48,55}},
/* 4126 */ {(16<<2)|3,{50,48,56}},
/* 4127 */ {(16<<2)|3,{50,48,57}},
/* 4128 */ {(16<<2)|3,{50,48,61}},
/* 4129 */ {(16<<2)|3,{50,48,65}},
/* 4130 */ {(16<<2)|3,{50,48,95}},
/* 4131 */ {(16<<2)|3,{50,48,98}},
/* 4132 */ {(16<<2)|3,{50,48,100}},
/* 4133 */ {(16<<2)|3,{50,48,102}},
/* 4134 */ {(16<<2)|3,{50,48,103}},
/* 4135 */ {(16<<2)|3,{50,48,104}},
/* 4136 */ {(16<<2)|3,{50,48,108}},
/* 4137 */ {(16<<2)|3,{50,48,109}},
/* 4138 */ {(16<<2)|3,{50,48,110}},
/* 4139 */ {(16<<2)|3,{50,48,112}},
/* 4140 */ {(16<<2)|3,{50,48,114}},
/* 4141 */ {(16<<2)|3,{50,48,117}},
/* 4142 */ {(10<<2)|2,{50,48,0}},
/* 4143 */ {(10<<2)|2,{50,48,0}},
/* 4144 */ {(10<<2)|2,{50,48,0}},
/* 4145 */ {(10<<2)|2,{50,48,0}},
/* 4146 */ {(10<<2)|2,{50,48,0}},
/* 4147 */ {(10<<2)|2,{50,48,0}},
/* 4148 */ {(10<<2)|2,{50,48,0}},
/* 4149 */ {(10<<2)|2,{50,48,0}},
/* 4150 */ {(10<<2)|2,{50,48,0}},
/* 4151 */ {(10<<2)|2,{50,48,0}},
/* 4152 */ {(10<<2)|2,{50,48,0}},
/* 4153 */ {(10<<2)|2,{50,48,0}},
/* 4154 */ {(10<<2)|2,{50,48,0}},
/* 4155 */ {(10<<2)|2,{50,48,0}},
/* 4156 */ {(10<<2)|2,{50,48,0}},
/* 4157 */ {(10<<2)|2,{50,48,0}},
/* 4158 */ {(10<<2)|2,{50,48,0}},
/* 4159 */ {(10<<2)|2,{50,48,0}},
/* 4160 */ {(15<<2)|3,{50,49,48}},
/* 4161 */ {(15<<2)|3,{50,49,48}},
/* 4162 */ {(15<<2)|3,{50,49,49}},
/* 4163 */ {(15<<2)|3,{50,49,49}},
/* 4164 */ {(15<<2)|3,{50,49,50}},
/* 4165 */ {(15<<2)|3,{50,49,50}},
/* 4166 */ {(15<<2)|3,{50,49,97}},
/* 4167 */ {(15<<2)|3,{50,49,97}},
/* 4168 */ {(15<<2)|3,{50,49,99}},
/* 4169 */ {(15<<2)|3,{50,49,99}},
/* 4170 */ {(15<<2)|3,{50,49,101}},
/* 4171 */ {(15<<2)|3,{50,49,101}},
/* 4172 */ {(15<<2)|3,{50,49,105}},
/* 4173 */ {(15<<2)|3,{50,49,105}},
/* 4174 */ {(15<<2)|3,{50,49,111}},
/* 4175 */ {(15<<2)|3,{50,49,111}},
/* 4176 */ {(15<<2)|3,{50,49,115}},
/* 4177 */ {(15<<2)|3,{50,49,115}},
/* 4178 */ {(15<<2)|3,{50,49,116}},
/* 4179 */ {(15<<2)|3,{50,49,116}},
/* 4180 */ {(16<<2)|3,{50,49,32}},
/* 4181 */ {(16<<2)|3,{50,49,37}},
/* 4182 */ {(16<<2)|3,{50,49,45}},
/* 4183 */ {(16<<2)|3,{50,49,46}},
/* 4184 */ {(16<<2)|3,{50,49,47}},
/* 4185 */ {(16<<2)|3,{50,49,51}},
/* 4186 */ {(16<<2)|3,{50,49,52}},
/* 4187 */ {(16<<2)|3,{50,49,53}},
/* 4188 */ {(16<<2)|3,{50,49,54}},
/* 4189 */ {(16<<2)|3,{50,49,55}},
/* 4190 */ {(16<<2)|3,{50,49,56}},
/* 4191 */ {(16<<2)|3,{50,49,57}},
/* 4192 */ {(16<<2)|3,{50,49,61}},
/* 4193 */ {(16<<2)|3,{50,49,65}},
/* 4194 */ {(16<<2)|3,{50,49,95}},
/* 4195 */ {(16<<2)|3,{50,49,98}},
/* 4196 */ {(16<<2)|3,{50,49,100}},
/* 4197 */ {(16<<2)|3,{50,49,102}},
/* 4198 */ {(16<<2)|3,{50,49,103}},
/* 4199 */ {(16<<2)|3,{50,49,104}},
/* 4200 */ {(16<<2)|3,{50,49,108}},
/* 4201 */ {(16<<2)|3,{50,49,109}},
/* 4202 */ {(16<<2)|3,{50,49,110}},
/* 4203 */ {(16<<2)|3,{50,49,112}},
/* 4204 */ {(16<<2)|3,{50,49,114}},
/* 4205 */ {(16<<2)|3,{50,49,117}},
/* 4206 */ {(10<<2)|2,{50,49,0}},
/* 4207 */ {(10<<2)|2,{50,49,0}},
/* 4208 */ {(10<<2)|2,{50,49,0}},
/* 4209 */ {(10<<2)|2,{50,49,0}},
/* 4210 */ {(10<<2)|2,{50,49,0}},
/* 4211 */ {(10<<2)|2,{50,49,0}},
/* 4212 */ {(10<<2)|2,{50,49,0}},
/* 4213 */ {(10<<2)|2,{50,49,0}},
/* 4214 */ {(10<<2)|2,{50,49,0}},
/* 4215 */ {(10<<2)|2,{50,49,0}},
/* 4216 */ {(10<<2)|2,{50,49,0}},
/* 4217 */ {(10<<2)|2,{50,49,0}},
/* 4218 */ {(10<<2)|2,{50,49,0}},
/* 4219 */ {(10<<2)|2,{50,49,0}},
/* 4220 */ {(10<<2)|2,{50,49,0}},
/* 4221 */ {(10<<2)|2,{50,49,0}},
/* 4222 */ {(10<<2)|2,{50,49,0}},
/* 4223 */ {(10<<2)|2,{50,49,0}},
/* 4224 */ {(15<<2)|3,{50,50,48}},
/* 4225 */ {(15<<2)|3,{50,50,48}},
/* 4226 */ {(15<<2)|3,{50,50,49}},
/* 4227 */ {(15<<2)|3,{50,50,49}},
/* 4228 */ {(15<<2)|3,{50,50,50}},
/* 4229 */ {(15<<2)|3,{50,50,50}},
/* 4230 */ {(15<<2)|3,{50,50,97}},
/* 4231 */ {(15<<2)|3,{50,50,97}},
/* 4232 */ {(15<<2)|3,{50,50,99}},
/* 4233 */ {(15<<2)|3,{50,50,99}},
/* 4234 */ {(15<<2)|3,{50,50,101}},
/* 4235 */ {(15<<2)|3,{50,50,101}},
/* 4236 */ {(15<<2)|3,{50,50,105}},
/* 4237 */ {(15<<2)|3,{50,50,105}},
/* 4238 */ {(15<<2)|3,{50,50,111}},
/* 4239 */ {(15<<2)|3,{50,50,111}},
/* 4240 */ {(15<<2)|3,{50,50,115}},
/* 4241 */ {(15<<2)|3,{50,50,115}},
/* 4242 */ {(15<<2)|3,{50,50,116}},
/* 4243 */ {(15<<2)|3,{50,50,116}},
/* 4244 */ {(16<<2)|3,{50,50,32}},
/* 4245 */ {(16<<2)|3,{50,50,37}},
/* 4246 */ {(16<<2)|3,{50,50,45}},
/* 4247 */ {(16<<2)|3,{50,50,46}},
/* 4248 */ {(16<<2)|3,{50,50,47}},
/* 4249 */ {(16<<2)|3,{50,50,51}},
/* 4250 */ {(16<<2)|3,{50,50,52}},
/* 4251 */ {(16<<2)|3,{50,50,53}},
/* 4252 */ {(16<<2)|3,{50,50,54}},
/* 4253 */ {(16<<2)|3,{50,50,55}},
/* 4254 */ {(16<<2)|3,{50,50,56}},
/* 4255 */ {(16<<2)|3,{50,50,57}},
/* 4256 */ {(16<<2)|3,{50,50,61}},
/* 4257 */ {(16<<2)|3,{50,50,65}},
/* 4258 */ {(16<<2)|3,{50,50,95}},
/* 4259 */ {(16<<2)|3,{50,50,98}},
/* 4260 */ {(16<<2)|3,{50,50,100}},
/* 4261 */ {(16<<2)|3,{50,50,102}},
/* 4262 */ {(16<<2)|3,{50,50,103}},
/* 4263 */ {(16<<2)|3,{50,50,104}},
/* 4264 */ {(16<<2)|3,{50,50,108}},
/* 4265 */ {(16<<2)|3,{50,50,109}},
/* 4266 */ {(16<<2)|3,{50,50,110}},
/* 4267 */ {(16<<2)|3,{50,50,112}},
/* 4268 */ {(16<<2)|3,{50,50,114}},
/* 4269 */ {(16<<2)|3,{50,50,117}},
/* 4270 */ {(10<<2)|2,{50,50,0}},
/* 4271 */ {(10<<2)|2,{50,50,0}},
/* 4272 */ {(10<<2)|2,{50,50,0}},
/* 4273 */ {(10<<2)|2,{50,50,0}},
/* 4274 */ {(10<<2)|2,{50,50,0}},
/* 4275 */ {(10<<2)|2,{50,50,0}},
/* 4276 */ {(10<<2)|2,{50,50,0}},
/* 4277 */ {(10<<2)|2,{50,50,0}},
/* 4278 */ {(10<<2)|2,{50,50,0}},
/* 4279 */ {(10<<2)|2,{50,50,0}},
/* 4280 */ {(10<<2)|2,{50,50,0}},
/* 4281 */ {(10<<2)|2,{50,50,0}},
/* 4282 */ {(10<<2)|2,{50,50,0}},
/* 4283 */ {(10<<2)|2,{50,50,0}},
/* 4284 */ {(10<<2)|2,{50,50,0}},
/* 4285 */ {(10<<2)|2,{50,50,0}},
/* 4286 */ {(10<<2)|2,{50,50,0}},
/* 4287 */ {(10<<2)|2,{50,50,0}},
/* 4288 */ {(15<<2)|3,{50,97,48}},
/* 4289 */ {(15<<2)|3,{50,97,48}},
/* 4290 */ {(15<<2)|3,{50,97,49}},
/* 4291 */ {(15<<2)|3,{50,97,49}},
/* 4292 */ {(15<<2)|3,{50,97,50}},
/* 4293 */ {(15<<2)|3,{50,97,50}},
/* 4294 */ {(15<<2)|3,{50,97,97}},
/* 4295 */ {(15<<2)|3,{50,97,97}},
/* 4296 */ {(15<<2)|3,{50,97,99}},
/* 4297 */ {(15<<2)|3,{50,97,99}},
/* 4298 */ {(15<<2)|3,{50,97,101}},
/* 4299 */ {(15<<2)|3,{50,97,101}},
/* 4300 */ {(15<<2)|3,{50,97,105}},
/* 4301 */ {(15<<2)|3,{50,97,105}},
/* 4302 */ {(15<<2)|3,{50,97,111}},
/* 4303 */ {(15<<2)|3,{50,97,111}},
/* 4304 */ {(15<<2)|3,{50,97,115}},
/* 4305 */ {(15<<2)|3,{50,97,115}},
/* 4306 */ {(15<<2)|3,{50,97,116}},
/* 4307 */ {(15<<2)|3,{50,97,116}},
/* 4308 */ {(16<<2)|3,{50,97,32}},
/* 4309 */ {(16<<2)|3,{50,97,37}},
/* 4310 */ {(16<<2)|3,{50,97,45}},
/* 4311 */ {(16<<2)|3,{50,97,46}},
/* 4312 */ {(16<<2)|3,{50,97,47}},
/* 4313 */ {(16<<2)|3,{50,97,51}},
/* 4314 */ {(16<<2)|3,{50,97,52}},
/* 4315 */ {(16<<2)|3,{50,97,53}},
/* 4316 */ {(16<<2)|3,{50,97,54}},
/* 4317 */ {(16<<2)|3,{50,97,55}},
/* 4318 */ {(16<<2)|3,{50,97,56}},
/* 4319 */ {(16<<2)|3,{50,97,57}},
/* 4320 */ {(16<<2)|3,{50,97,61}},
/* 4321 */ {(16<<2)|3,{50,97,65}},
/* 4322 */ {(16<<2)|3,{50,97,95}},
/* 4323 */ {(16<<2)|3,{50,97,98}},
/* 4324 */ {(16<<2)|3,{50,97,100}},
/* 4325 */ {(16<<2)|3,{50,97,102}},
/* 4326 */ {(16<<2)|3,{50,97,103}},
/* 4327 */ {(16<<2)|3,{50,97,104}},
/* 4328 */ {(16<<2)|3,{50,97,108}},
/* 4329 */ {(16<<2)|3,{50,97,109}},
/* 4330 */ {(16<<2)|3,{50,97,110}},
/* 4331 */ {(16<<2)|3,{50,97,112}},
/* 4332 */ {(16<<2)|3,{50,97,114}},
/* 4333 */ {(16<<2)|3,{50,97,117}},
/* 4334 */ {(10<<2)|2,{50,97,0}},
/* 4335 */ {(10<<2)|2,{50,97,0}},
/* 4336 */ {(10<<2)|2,{50,97,0}},
/* 4337 */ {(10<<2)|2,{50,97,0}},
/* 4338 */ {(10<<2)|2,{50,97,0}},
/* 4339 */ {(10<<2)|2,{50,97,0}},
/* 4340 */ {(10<<2)|2,{50,97,0}},
/* 4341 */ {(10<<2)|2,{50,97,0}},
/* 4342 */ {(10<<2)|2,{50,97,0}},
/* 4343 */ {(10<<2)|2,{50,97,0}},
/* 4344 */ {(10<<2)|2,{50,97,0}},
/* 4345 */ {(10<<2)|2,{50,97,0}},
/* 4346 */ {(10<<2)|2,{50,97,0}},
/* 4347 */ {(10<<2)|2,{50,97,0}},
/* 4348 */ {(10<<2)|2,{50,97,0}},
/* 4349 */ {(10<<2)|2,{50,97,0}},
/* 4350 */ {(10<<2)|2,{50,97,0}},
/* 4351 */ {(10<<2)|2,{50,97,0}},
/* 4352 */ {(15<<2)|3,{50,99,48}},
/* 4353 */ {(15<<2)|3,{50,99,48}},
/* 4354 */ {(15<<2)|3,{50,99,49}},
/* 4355 */ {(15<<2)|3,{50,99,49}},
/* 4356 */ {(15<<2)|3,{50,99,50}},
/* 4357 */ {(15<<2)|3,{50,99,50}},
/* 4358 */ {(15<<2)|3,{50,99,97}},
/* 4359 */ {(15<<2)|3,{50,99,97}},
/* 4360 */ {(15<<2)|3,{50,99,99}},
/* 4361 */ {(15<<2)|3,{50,99,99}},
/* 4362 */ {(15<<2)|3,{50,99,101}},
/* 4363 */ {(15<<2)|3,{50,99,101}},
/* 4364 */ {(15<<2)|3,{50,99,105}},
/* 4365 */ {(15<<2)|3,{50,99,105}},
/* 4366 */ {(15<<2)|3,{50,99,111}},
/* 4367 */ {(15<<2)|3,{50,99,111}},
/* 4368 */ {(15<<2)|3,{50,99,115}},
/* 4369 */ {(15<<2)|3,{50,99,115}},
/* 4370 */ {(15<<2)|3,{50,99,116}},
/* 4371 */ {(15<<2)|3,{50,99,116}},
/* 4372 */ {(16<<2)|3,{50,99,32}},
/* 4373 */ {(16<<2)|3,{50,99,37}},
/* 4374 */ {(16<<2)|3,{50,99,45}},
/* 4375 */ {(16<<2)|3,{50,99,46}},
/* 4376 */ {(16<<2)|3,{50,99,47}},
/* 4377 */ {(16<<2)|3,{50,99,51}},
/* 4378 */ {(16<<2)|3,{50,99,52}},
/* 4379 */ {(16<<2)|3,{50,99,53}},
/* 4380 */ {(16<<2)|3,{50,99,54}},
/* 4381 */ {(16<<2)|3,{50,99,55}},
/* 4382 */ {(16<<2)|3,{50,99,56}},
/* 4383 */ {(16<<2)|3,{50,99,57}},
/* 4384 */ {(16<<2)|3,{50,99,61}},
/* 4385 */ {(16<<2)|3,{50,99,65}},
/* 4386 */ {(16<<2)|3,{50,99,95}},
/* 4387 */ {(16<<2)|3,{50,99,98}},
/* 4388 */ {(16<<2)|3,{50,99,100}},
/* 4389 */ {(16<<2)|3,{50,99,102}},
/* 4390 */ {(16<<2)|3,{50,99,103}},
/* 4391 */ {(16<<2)|3,{50,99,104}},
/* 4392 */ {(16<<2)|3,{50,99,108}},
/* 4393 */ {(16<<2)|3,{50,99,109}},
/* 4394 */ {(16<<2)|3,{50,99,110}},
/* 4395 */ {(16<<2)|3,{50,99,112}},
/* 4396 */ {(16<<2)|3,{50,99,114}},
/* 4397 */ {(16<<2)|3,{50,99,117}},
/* 4398 */ {(10<<2)|2,{50,99,0}},
/* 4399 */ {(10<<2)|2,{50,99,0}},
/* 4400 */ {(10<<2)|2,{50,99,0}},
/* 4401 */ {(10<<2)|2,{50,99,0}},
/* 4402 */ {(10<<2)|2,{50,99,0}},
/* 4403 */ {(10<<2)|2,{50,99,0}},
/* 4404 */ {(10<<2)|2,{50,99,0}},
/* 4405 */ {(10<<2)|2,{50,99,0}},
/* 4406 */ {(10<<2)|2,{50,99,0}},
/* 4407 */ {(10<<2)|2,{50,99,0}},
/* 4408 */ {(10<<2)|2,{50,99,0}},
/* 4409 */ {(10<<2)|2,{50,99,0}},
/* 4410 */ {(10<<2)|2,{50,99,0}},
/* 4411 */ {(10<<2)|2,{50,99,0}},
/* 4412 */ {(10<<2)|2,{50,99,0}},
/* 4413 */ {(10<<2)|2,{50,99,0}},
/* 4414 */ {(10<<2)|2,{50,99,0}},
/* 4415 */ {(10<<2)|2,{50,99,0}},
/* 4416 */ {(15<<2)|3,{50,101,48}},
/* 4417 */ {(15<<2)|3,{50,101,48}},
/* 4418 */ {(15<<2)|3,{50,101,49}},
/* 4419 */ {(15<<2)|3,{50,101,49}},
/* 4420 */ {(15<<2)|3,{50,101,50}},
/* 4421 */ {(15<<2)|3,{50,101,50}},
/* 4422 */ {(15<<2)|3,{50,101,97}},
/* 4423 */ {(15<<2)|3,{50,101,97}},
/* 4424 */ {(15<<2)|3,{50,101,99}},
/* 4425 */ {(15<<2)|3,{50,101,99}},
/* 4426 */ {(15<<2)|3,{50,101,101}},
/* 4427 */ {(15<<2)|3,{50,101,101}},
/* 4428 */ {(15<<2)|3,{50,101,105}},
/* 4429 */ {(15<<2)|3,{50,101,105}},
/* 4430 */ {(15<<2)|3,{50,101,111}},
/* 4431 */ {(15<<2)|3,{50,101,111}},
/* 4432 */ {(15<<2)|3,{50,101,115}},
/* 4433 */ {(15<<2)|3,{50,101,115}},
/* 4434 */ {(15<<2)|3,{50,101,116}},
/* 4435 */ {(15<<2)|3,{50,101,116}},
/* 4436 */ {(16<<2)|3,{50,101,32}},
/* 4437 */ {(16<<2)|3,{50,101,37}},
/* 4438 */ {(16<<2)|3,{50,101,45}},
/* 4439 */ {(16<<2)|3,{50,101,46}},
/* 4440 */ {(16<<2)|3,{50,101,47}},
/* 4441 */ {(16<<2)|3,{50,101,51}},
/* 4442 */ {(16<<2)|3,{50,101,52}},
/* 4443 */ {(16<<2)|3,{50,101,53}},
/* 4444 */ {(16<<2)|3,{50,101,54}},
/* 4445 */ {(16<<2)|3,{50,101,55}},
/* 4446 */ {(16<<2)|3,{50,101,56}},
/* 4447 */ {(16<<2)|3,{50,101,57}},
/* 4448 */ {(16<<2)|3,{50,101,61}},
/* 4449 */ {(16<<2)|3,{50,101,65}},
/* 4450 */ {(16<<2)|3,{50,101,95}},
/* 4451 */ {(16<<2)|3,{50,101,98}},
/* 4452 */ {(16<<2)|3,{50,101,100}},
/* 4453 */ {(16<<2)|3,{50,101,102}},
/* 4454 */ {(16<<2)|3,{50,101,103}},
/* 4455 */ {(16<<2)|3,{50,101,104}},
/* 4456 */ {(16<<2)|3,{50,101,108}},
/* 4457 */ {(16<<2)|3,{50,101,109}},
/* 4458 */ {(16<<2)|3,{50,101,110}},
/* 4459 */ {(16<<2)|3,{50,101,112}},
/* 4460 */ {(16<<2)|3,{50,101,114}},
/* 4461 */ {(16<<2)|3,{50,101,117}},
/* 4462 */ {(10<<2)|2,{50,101,0}},
/* 4463 */ {(10<<2)|2,{50,101,0}},
/* 4464 */ {(10<<2)|2,{50,101,0}},
/* 4465 */ {(10<<2)|2,{50,101,0}},
/* 4466 */ {(10<<2)|2,{50,101,0}},
/* 4467 */ {(10<<2)|2,{50,101,0}},
/* 4468 */ {(10<<2)|2,{50,101,0}},
/* 4469 */ {(10<<2)|2,{50,101,0}},
/* 4470 */ {(10<<2)|2,{50,101,0}},
/* 4471 */ {(10<<2)|2,{50,101,0}},
/* 4472 */ {(10<<2)|2,{50,101,0}},
/* 4473 */ {(10<<2)|2,{50,101,0}},
/* 4474 */ {(10<<2)|2,{50,101,0}},
/* 4475 */ {(10<<2)|2,{50,101,0}},
/* 4476 */ {(10<<2)|2,{50,101,0}},
/* 4477 */ {(10<<2)|2,{50,101,0}},
/* 4478 */ {(10<<2)|2,{50,101,0}},
/* 4479 */ {(10<<2)|2,{50,101,0}},
/* 4480 */ {(15<<2)|3,{50,105,48}},
/* 4481 */ {(15<<2)|3,{50,105,48}},
/* 4482 */ {(15<<2)|3,{50,105,49}},
/* 4483 */ {(15<<2)|3,{50,105,49}},
/* 4484 */ {(15<<2)|3,{50,105,50}},
/* 4485 */ {(15<<2)|3,{50,105,50}},
/* 4486 */ {(15<<2)|3,{50,105,97}},
/* 4487 */ {(15<<2)|3,{50,105,97}},
/* 4488 */ {(15<<2)|3,{50,105,99}},
/* 4489 */ {(15<<2)|3,{50,105,99}},
/* 4490 */ {(15<<2)|3,{50,105,101}},
/* 4491 */ {(15<<2)|3,{50,105,101}},
/* 4492 */ {(15<<2)|3,{50,105,105}},
/* 4493 */ {(15<<2)|3,{50,105,105}},
/* 4494 */ {(15<<2)|3,{50,105,111}},
/* 4495 */ {(15<<2)|3,{50,105,111}},
/* 4496 */ {(15<<2)|3,{50,105,115}},
/* 4497 */ {(15<<2)|3,{50,105,115}},
/* 4498 */ {(15<<2)|3,{50,105,116}},
/* 4499 */ {(15<<2)|3,{50,105,116}},
/* 4500 */ {(16<<2)|3,{50,105,32}},
/* 4501 */ {(16<<2)|3,{50,105,37}},
/* 4502 */ {(16<<2)|3,{50,105,45}},
/* 4503 */ {(16<<2)|3,{50,105,46}},
/* 4504 */ {(16<<2)|3,{50,105,47}},
/* 4505 */ {(16<<2)|3,{50,105,51}},
/* 4506 */ {(16<<2)|3,{50,105,52}},
/* 4507 */ {(16<<2)|3,{50,105,53}},
/* 4508 */ {(16<<2)|3,{50,105,54}},
/* 4509 */ {(16<<2)|3,{50,105,55}},
/* 4510 */ {(16<<2)|3,{50,105,56}},
/* 4511 */ {(16<<2)|3,{50,105,57}},
/* 4512 */ {(16<<2)|3,{50,105,61}},
/* 4513 */ {(16<<2)|3,{50,105,65}},
/* 4514 */ {(16<<2)|3,{50,105,95}},
/* 4515 */ {(16<<2)|3,{50,105,98}},
/* 4516 */ {(16<<2)|3,{50,105,100}},
/* 4517 */ {(16<<2)|3,{50,105,102}},
/* 4518 */ {(16<<2)|3,{50,105,103}},
/* 4519 */ {(16<<2)|3,{50,105,104}},
/* 4520 */ {(16<<2)|3,{50,105,108}},
/* 4521 */ {(16<<2)|3,{50,105,109}},
/* 4522 */ {(16<<2)|3,{50,105,110}},
/* 4523 */ {(16<<2)|3,{50,105,112}},
/* 4524 */ {(16<<2)|3,{50,105,114}},
/* 4525 */ {(16<<2)|3,{50,105,117}},
/* 4526 */ {(10<<2)|2,{50,105,0}},
/* 4527 */ {(10<<2)|2,{50,105,0}},
/* 4528 */ {(10<<2)|2,{50,105,0}},
/* 4529 */ {(10<<2)|2,{50,105,0}},
/* 4530 */ {(10<<2)|2,{50,105,0}},
/* 4531 */ {(10<<2)|2,{50,105,0}},
/* 4532 */ {(10<<2)|2,{50,105,0}},
/* 4533 */ {(10<<2)|2,{50,105,0}},
/* 4534 */ {(10<<2)|2,{50,105,0}},
/* 4535 */ {(10<<2)|2,{50,105,0}},
/* 4536 */ {(10<<2)|2,{50,105,0}},
/* 4537 */ {(10<<2)|2,{50,105,0}},
/* 4538 */ {(10<<2)|2,{50,105,0}},
/* 4539 */ {(10<<2)|2,{50,105,0}},
/* 4540 */ {(10<<2)|2,{50,105,0}},
/* 4541 */ {(10<<2)|2,{50,105,0}},
/* 4542 */ {(10<<2)|2,{50,105,0}},
/* 4543 */ {(10<<2)|2,{50,105,0}},
/* 4544 */ {(15<<2)|3,{50,111,48}},
/* 4545 */ {(15<<2)|3,{50,111,48}},
/* 4546 */ {(15<<2)|3,{50,111,49}},
/* 4547 */ {(15<<2)|3,{50,111,49}},
/* 4548 */ {(15<<2)|3,{50,111,50}},
/* 4549 */ {(15<<2)|3,{50,111,50}},
/* 4550 */ {(15<<2)|3,{50,111,97}},
/* 4551 */ {(15<<2)|3,{50,111,97}},
/* 4552 */ {(15<<2)|3,{50,111,99}},
/* 4553 */ {(15<<2)|3,{50,111,99}},
/* 4554 */ {(15<<2)|3,{50,111,101}},
/* 4555 */ {(15<<2)|3,{50,111,101}},
/* 4556 */ {(15<<2)|3,{50,111,105}},
/* 4557 */ {(15<<2)|3,{50,111,105}},
/* 4558 */ {(15<<2)|3,{50,111,111}},
/* 4559 */ {(15<<2)|3,{50,111,111}},
/* 4560 */ {(15<<2)|3,{50,111,115}},
/* 4561 */ {(15<<2)|3,{50,111,115}},
/* 4562 */ {(15<<2)|3,{50,111,116}},
/* 4563 */ {(15<<2)|3,{50,111,116}},
/* 4564 */ {(16<<2)|3,{50,111,32}},
/* 4565 */ {(16<<2)|3,{50,111,37}},
/* 4566 */ {(16<<2)|3,{50,111,45}},
/* 4567 */ {(16<<2)|3,{50,111,46}},
/* 4568 */ {(16<<2)|3,{50,111,47}},
/* 4569 */ {(16<<2)|3,{50,111,51}},
/* 4570 */ {(16<<2)|3,{50,111,52}},
/* 4571 */ {(16<<2)|3,{50,111,53}},
/* 4572 */ {(16<<2)|3,{50,111,54}},
/* 4573 */ {(16<<2)|3,{50,111,55}},
/* 4574 */ {(16<<2)|3,{50,111,56}},
/* 4575 */ {(16<<2)|3,{50,111,57}},
/* 4576 */ {(16<<2)|3,{50,111,61}},
/* 4577 */ {(16<<2)|3,{50,111,65}},
/* 4578 */ {(16<<2)|3,{50,111,95}},
/* 4579 */ {(16<<2)|3,{50,111,98}},
/* 4580 */ {(16<<2)|3,{50,111,100}},
/* 4581 */ {(16<<2)|3,{50,111,102}},
/* 4582 */ {(16<<2)|3,{50,111,103}},
/* 4583 */ {(16<<2)|3,{50,111,104}},
/* 4584 */ {(16<<2)|3,{50,111,108}},
/* 4585 */ {(16<<2)|3,{50,111,109}},
/* 4586 */ {(16<<2)|3,{50,111,110}},
/* 4587 */ {(16<<2)|3,{50,111,112}},
/* 4588 */ {(16<<2)|3,{50,111,114}},
/* 4589 */ {(16<<2)|3,{50,111,117}},
/* 4590 */ {(10<<2)|2,{50,111,0}},
/* 4591 */ {(10<<2)|2,{50,111,0}},
/* 4592 */ {(10<<2)|2,{50,111,0}},
/* 4593 */ {(10<<2)|2,{50,111,0}},
/* 4594 */ {(10<<2)|2,{50,111,0}},
/* 4595 */ {(10<<2)|2,{50,111,0}},
/* 4596 */ {(10<<2)|2,{50,111,0}},
/* 4597 */ {(10<<2)|2,{50,111,0}},
/* 4598 */ {(10<<2)|2,{50,111,0}},
/* 4599 */ {(10<<2)|2,{50,111,0}},
/* 4600 */ {(10<<2)|2,{50,111,0}},
/* 4601 */ {(10<<2)|2,{50,111,0}},
/* 4602 */ {(10<<2)|2,{50,111,0}},
/* 4603 */ {(10<<2)|2,{50,111,0}},
/* 4604 */ {(10<<2)|2,{50,111,0}},
/* 4605 */ {(10<<2)|2,{50,111,0}},
/* 4606 */ {(10<<2)|2,{50,111,0}},
/* 4607 */ {(10<<2)|2,{50,111,0}},
/* 4608 */ {(15<<2)|3,{50,115,48}},
/* 4609 */ {(15<<2)|3,{50,115,48}},
/* 4610 */ {(15<<2)|3,{50,115,49}},
/* 4611 */ {(15<<2)|3,{50,115,49}},
/* 4612 */ {(15<<2)|3,{50,115,50}},
/* 4613 */ {(15<<2)|3,{50,115,50}},
/* 4614 */ {(15<<2)|3,{50,115,97}},
/* 4615 */ {(15<<2)|3,{50,115,97}},
/* 4616 */ {(15<<2)|3,{50,115,99}},
/* 4617 */ {(15<<2)|3,{50,115,99}},
/* 4618 */ {(15<<2)|3,{50,115,101}},
/* 4619 */ {(15<<2)|3,{50,115,101}},
/* 4620 */ {(15<<2)|3,{50,115,105}},
/* 4621 */ {(15<<2)|3,{50,115,105}},
/* 4622 */ {(15<<2)|3,{50,115,111}},
/* 4623 */ {(15<<2)|3,{50,115,111}},
/* 4624 */ {(15<<2)|3,{50,115,115}},
/* 4625 */ {(15<<2)|3,{50,115,115}},
/* 4626 */ {(15<<2)|3,{50,115,116}},
/* 4627 */ {(15<<2)|3,{50,115,116}},
/* 4628 */ {(16<<2)|3,{50,115,32}},
/* 4629 */ {(16<<2)|3,{50,115,37}},
/* 4630 */ {(16<<2)|3,{50,115,45}},
/* 4631 */ {(16<<2)|3,{50,115,46}},
/* 4632 */ {(16<<2)|3,{50,115,47}},
/* 4633 */ {(16<<2)|3,{50,115,51}},
/* 4634 */ {(16<<2)|3,{50,115,52}},
/* 4635 */ {(16<<2)|3,{50,115,53}},
/* 4636 */ {(16<<2)|3,{50,115,54}},
/* 4637 */ {(16<<2)|3,{50,115,55}},
/* 4638 */ {(16<<2)|3,{50,115,56}},
/* 4639 */ {(16<<2)|3,{50,115,57}},
/* 4640 */ {(16<<2)|3,{50,115,61}},
/* 4641 */ {(16<<2)|3,{50,115,65}},
/* 4642 */ {(16<<2)|3,{50,115,95}},
/* 4643 */ {(16<<2)|3,{50,115,98}},
/* 4644 */ {(16<<2)|3,{50,115,100}},
/* 4645 */ {(16<<2)|3,{50,115,102}},
/* 4646 */ {(16<<2)|3,{50,115,103}},
/* 4647 */ {(16<<2)|3,{50,115,104}},
/* 4648 */ {(16<<2)|3,{50,115,108}},
/* 4649 */ {(16<<2)|3,{50,115,109}},
/* 4650 */ {(16<<2)|3,{50,115,110}},
/* 4651 */ {(16<<2)|3,{50,115,112}},
/* 4652 */ {(16<<2)|3,{50,115,114}},
/* 4653 */ {(16<<2)|3,{50,115,117}},
/* 4654 */ {(10<<2)|2,{50,115,0}},
/* 4655 */ {(10<<2)|2,{50,115,0}},
/* 4656 */ {(10<<2)|2,{50,115,0}},
/* 4657 */ {(10<<2)|2,{50,115,0}},
/* 4658 */ {(10<<2)|2,{50,115,0}},
/* 4659 */ {(10<<2)|2,{50,115,0}},
/* 4660 */ {(10<<2)|2,{50,115,0}},
/* 4661 */ {(10<<2)|2,{50,115,0}},
/* 4662 */ {(10<<2)|2,{50,115,0}},
/* 4663 */ {(10<<2)|2,{50,115,0}},
/* 4664 */ {(10<<2)|2,{50,115,0}},
/* 4665 */ {(10<<2)|2,{50,115,0}},
/* 4666 */ {(10<<2)|2,{50,115,0}},
/* 4667 */ {(10<<2)|2,{50,115,0}},
/* 4668 */ {(10<<2)|2,{50,115,0}},
/* 4669 */ {(10<<2)|2,{50,115,0}},
/* 4670 */ {(10<<2)|2,{50,115,0}},
/* 4671 */ {(10<<2)|2,{50,115,0}},
/* 4672 */ {(15<<2)|3,{50,116,48}},
/* 4673 */ {(15<<2)|3,{50,116,48}},
/* 4674 */ {(15<<2)|3,{50,116,49}},
/* 4675 */ {(15<<2)|3,{50,116,49}},
/* 4676 */ {(15<<2)|3,{50,116,50}},
/* 4677 */ {(15<<2)|3,{50,116,50}},
/* 4678 */ {(15<<2)|3,{50,116,97}},
/* 4679 */ {(15<<2)|3,{50,116,97}},
/* 4680 */ {(15<<2)|3,{50,116,99}},
/* 4681 */ {(15<<2)|3,{50,116,99}},
/* 4682 */ {(15<<2)|3,{50,116,101}},
/* 4683 */ {(15<<2)|3,{50,116,101}},
/* 4684 */ {(15<<2)|3,{50,116,105}},
/* 4685 */ {(15<<2)|3,{50,116,105}},
/* 4686 */ {(15<<2)|3,{50,116,111}},
/* 4687 */ {(15<<2)|3,{50,116,111}},
/* 4688 */ {(15<<2)|3,{50,116,115}},
/* 4689 */ {(15<<2)|3,{50,116,115}},
/* 4690 */ {(15<<2)|3,{50,116,116}},
/* 4691 */ {(15<<2)|3,{50,116,116}},
/* 4692 */ {(16<<2)|3,{50,116,32}},
/* 4693 */ {(16<<2)|3,{50,116,37}},
/* 4694 */ {(16<<2)|3,{50,116,45}},
/* 4695 */ {(16<<2)|3,{50,116,46}},
/* 4696 */ {(16<<2)|3,{50,116,47}},
/* 4697 */ {(16<<2)|3,{50,116,51}},
/* 4698 */ {(16<<2)|3,{50,116,52}},
/* 4699 */ {(16<<2)|3,{50,116,53}},
/* 4700 */ {(16<<2)|3,{50,116,54}},
/* 4701 */ {(16<<2)|3,{50,116,55}},
/* 4702 */ {(16<<2)|3,{50,116,56}},
/* 4703 */ {(16<<2)|3,{50,116,57}},
/* 4704 */ {(16<<2)|3,{50,116,61}},
/* 4705 */ {(16<<2)|3,{50,116,65}},
/* 4706 */ {(16<<2)|3,{50,116,95}},
/* 4707 */ {(16<<2)|3,{50,116,98}},
/* 4708 */ {(16<<2)|3,{50,116,100}},
/* 4709 */ {(16<<2)|3,{50,116,102}},
/* 4710 */ {(16<<2)|3,{50,116,103}},
/* 4711 */ {(16<<2)|3,{50,116,104}},
/* 4712 */ {(16<<2)|3,{50,116,108}},
/* 4713 */ {(16<<2)|3,{50,116,109}},
/* 4714 */ {(16<<2)|3,{50,116,110}},
/* 4715 */ {(16<<2)|3,{50,116,112}},
/* 4716 */ {(16<<2)|3,{50,116,114}},
/* 4717 */ {(16<<2)|3,{50,116,117}},
/* 4718 */ {(10<<2)|2,{50,116,0}},
/* 4719 */ {(10<<2)|2,{50,116,0}},
/* 4720 */ {(10<<2)|2,{50,116,0}},
/* 4721 */ {(10<<2)|2,{50,116,0}},
/* 4722 */ {(10<<2)|2,{50,116,0}},
/* 4723 */ {(10<<2)|2,{50,116,0}},
/* 4724 */ {(10<<2)|2,{50,116,0}},
/* 4725 */ {(10<<2)|2,{50,116,0}},
/* 4726 */ {(10<<2)|2,{50,116,0}},
/* 4727 */ {(10<<2)|2,{50,116,0}},
/* 4728 */ {(10<<2)|2,{50,116,0}},
/* 4729 */ {(10<<2)|2,{50,116,0}},
/* 4730 */ {(10<<2)|2,{50,116,0}},
/* 4731 */ {(10<<2)|2,{50,116,0}},
/* 4732 */ {(10<<2)|2,{50,116,0}},
/* 4733 */ {(10<<2)|2,{50,116,0}},
/* 4734 */ {(10<<2)|2,{50,116,0}},
/* 4735 */ {(10<<2)|2,{50,116,0}},
/* 4736 */ {(16<<2)|3,{50,32,48}},
/* 4737 */ {(16<<2)|3,{50,32,49}},
/* 4738 */ {(16<<2)|3,{50,32,50}},
/* 4739 */ {(16<<2)|3,{50,32,97}},
/* 4740 */ {(16<<2)|3,{50,32,99}},
/* 4741 */ {(16<<2)|3,{50,32,101}},
/* 4742 */ {(16<<2)|3,{50,32,105}},
/* 4743 */ {(16<<2)|3,{50,32,111}},
/* 4744 */ {(16<<2)|3,{50,32,115}},
/* 4745 */ {(16<<2)|3,{50,32,116}},
/* 4746 */ {(11<<2)|2,{50,32,0}},
/* 4747 */ {(11<<2)|2,{50,32,0}},
/* 4748 */ {(11<<2)|2,{50,32,0}},
/* 4749 */ {(11<<2)|2,{50,32,0}},
/* 4750 */ {(11<<2)|2,{50,32,0}},
/* 4751 */ {(11<<2)|2,{50,32,0}},
/* 4752 */ {(11<<2)|2,{50,32,0}},
/* 4753 */ {(11<<2)|2,{50,32,0}},
/* 4754 */ {(11<<2)|2,{50,32,0}},
/* 4755 */ {(11<<2)|2,{50,32,0}},
/* 4756 */ {(11<<2)|2,{50,32,0}},
/* 4757 */ {(11<<2)|2,{50,32,0}},
/* 4758 */ {(11<<2)|2,{50,32,0}},
/* 4759 */ {(11<<2)|2,{50,32,0}},
/* 4760 */ {(11<<2)|2,{50,32,0}},
/* 4761 */ {(11<<2)|2,{50,32,0}},
/* 4762 */ {(11<<2)|2,{50,32,0}},
/* 4763 */ {(11<<2)|2,{50,32,0}},
/* 4764 */ {(11<<2)|2,{50,32,0}},
/* 4765 */ {(11<<2)|2,{50,32,0}},
/* 4766 */ {(11<<2)|2,{50,32,0}},
/* 4767 */ {(11<<2)|2,{50,32,0}},
/* 4768 */ {(16<<2)|3,{50,37,48}},
/* 4769 */ {(16<<2)|3,{50,37,49}},
/* 4770 */ {(16<<2)|3,{50,37,50}},
/* 4771 */ {(16<<2)|3,{50,37,97}},
/* 4772 */ {(16<<2)|3,{50,37,99}},
/* 4773 */ {(16<<2)|3,{50,37,101}},
/* 4774 */ {(16<<2)|3,{50,37,105}},
/* 4775 */ {(16<<2)|3,{50,37,111}},
/* 4776 */ {(16<<2)|3,{50,37,115}},
/* 4777 */ {(16<<2)|3,{50,37,116}},
/* 4778 */ {(11<<2)|2,{50,37,0}},
/* 4779 */ {(11<<2)|2,{50,37,0}},
/* 4780 */ {(11<<2)|2,{50,37,0}},
/* 4781 */ {(11<<2)|2,{50,37,0}},
/* 4782 */ {(11<<2)|2,{50,37,0}},
/* 4783 */ {(11<<2)|2,{50,37,0}},
/* 4784 */ {(11<<2)|2,{50,37,0}},
/* 4785 */ {(11<<2)|2,{50,37,0}},
/* 4786 */ {(11<<2)|2,{50,37,0}},
/* 4787 */ {(11<<2)|2,{50,37,0}},
/* 4788 */ {(11<<2)|2,{50,37,0}},
/* 4789 */ {(11<<2)|2,{50,37,0}},
/* 4790 */ {(11<<2)|2,{50,37,0}},
/* 4791 */ {(11<<2)|2,{50,37,0}},
/* 4792 */ {(11<<2)|2,{50,37,0}},
/* 4793 */ {(11<<2)|2,{50,37,0}},
/* 4794 */ {(11<<2)|2,{50,37,0}},
/* 4795 */ {(11<<2)|2,{50,37,0}},
/* 4796 */ {(11<<2)|2,{50,37,0}},
/* 4797 */ {(11<<2)|2,{50,37,0}},
/* 4798 */ {(11<<2)|2,{50,37,0}},
/* 4799 */ {(11<<2)|2,{50,37,0}},
/* 4800 */ {(16<<2)|3,{50,45,48}},
/* 4801 */ {(16<<2)|3,{50,45,49}},
/* 4802 */ {(16<<2)|3,{50,45,50}},
/* 4803 */ {(16<<2)|3,{50,45,97}},
/* 4804 */ {(16<<2)|3,{50,45,99}},
/* 4805 */ {(16<<2)|3,{50,45,101}},
/* 4806 */ {(16<<2)|3,{50,45,105}},
/* 4807 */ {(16<<2)|3,{50,45,111}},
/* 4808 */ {(16<<2)|3,{50,45,115}},
/* 4809 */ {(16<<2)|3,{50,45,116}},
/* 4810 */ {(11<<2)|2,{50,45,0}},
/* 4811 */ {(11<<2)|2,{50,45,0}},
/* 4812 */ {(11<<2)|2,{50,45,0}},
/* 4813 */ {(11<<2)|2,{50,45,0}},
/* 4814 */ {(11<<2)|2,{50,45,0}},
/* 4815 */ {(11<<2)|2,{50,45,0}},
/* 4816 */ {(11<<2)|2,{50,45,0}},
/* 4817 */ {(11<<2)|2,{50,45,0}},
/* 4818 */ {(11<<2)|2,{50,45,0}},
/* 4819 */ {(11<<2)|2,{50,45,0}},
/* 4820 */ {(11<<2)|2,{50,45,0}},
/* 4821 */ {(11<<2)|2,{50,45,0}},
/* 4822 */ {(11<<2)|2,{50,45,0}},
/* 4823 */ {(11<<2)|2,{50,45,0}},
/* 4824 */ {(11<<2)|2,{50,45,0}},
/* 4825 */ {(11<<2)|2,{50,45,0}},
/* 4826 */ {(11<<2)|2,{50,45,0}},
/* 4827 */ {(11<<2)|2,{50,45,0}},
/* 4828 */ {(11<<2)|2,{50,45,0}},
/* 4829 */ {(11<<2)|2,{50,45,0}},
/* 4830 */ {(11<<2)|2,{50,45,0}},
/* 4831 */ {(11<<2)|2,{50,45,0}},
/* 4832 */ {(16<<2)|3,{50,46,48}},
/* 4833 */ {(16<<2)|3,{50,46,49}},
/* 4834 */ {(16<<2)|3,{50,46,50}},
/* 4835 */ {(16<<2)|3,{50,46,97}},
/* 4836 */ {(16<<2)|3,{50,46,99}},
/* 4837 */ {(16<<2)|3,{50,46,101}},
/* 4838 */ {(16<<2)|3,{50,46,105}},
/* 4839 */ {(16<<2)|3,{50,46,111}},
/* 4840 */ {(16<<2)|3,{50,46,115}},
/* 4841 */ {(16<<2)|3,{50,46,116}},
/* 4842 */ {(11<<2)|2,{50,46,0}},
/* 4843 */ {(11<<2)|2,{50,46,0}},
/* 4844 */ {(11<<2)|2,{50,46,0}},
/* 4845 */ {(11<<2)|2,{50,46,0}},
/* 4846 */ {(11<<2)|2,{50,46,0}},
/* 4847 */ {(11<<2)|2,{50,46,0}},
/* 4848 */ {(11<<2)|2,{50,46,0}},
/* 4849 */ {(11<<2)|2,{50,46,0}},
/* 4850 */ {(11<<2)|2,{50,46,0}},
/* 4851 */ {(11<<2)|2,{50,46,0}},
/* 4852 */ {(11<<2)|2,{50,46,0}},
/* 4853 */ {(11<<2)|2,{50,46,0}},
/* 4854 */ {(11<<2)|2,{50,46,0}},
/* 4855 */ {(11<<2)|2,{50,46,0}},
/* 4856 */ {(11<<2)|2,{50,46,0}},
/* 4857 */ {(11<<2)|2,{50,46,0}},
/* 4858 */ {(11<<2)|2,{50,46,0}},
/* 4859 */ {(11<<2)|2,{50,46,0}},
/* 4860 */ {(11<<2)|2,{50,46,0}},
/* 4861 */ {(11<<2)|2,{50,46,0}},
/* 4862 */ {(11<<2)|2,{50,46,0}},
/* 4863 */ {(11<<2)|2,{50,46,0}},
/* 4864 */ {(16<<2)|3,{50,47,48}},
/* 4865 */ {(16<<2)|3,{50,47,49}},
/* 4866 */ {(16<<2)|3,{50,47,50}},
/* 4867 */ {(16<<2)|3,{50,47,97}},
/* 4868 */ {(16<<2)|3,{50,47,99}},
/* 4869 */ {(16<<2)|3,{50,47,101}},
/* 4870 */ {(16<<2)|3,{50,47,105}},
/* 4871 */ {(16<<2)|3,{50,47,111}},
/* 4872 */ {(16<<2)|3,{50,47,115}},
/* 4873 */ {(16<<2)|3,{50,47,116}},
/* 4874 */ {(11<<2)|2,{50,47,0}},
/* 4875 */ {(11<<2)|2,{50,47,0}},
/* 4876 */ {(11<<2)|2,{50,47,0}},
/* 4877 */ {(11<<2)|2,{50,47,0}},
/* 4878 */ {(11<<2)|2,{50,47,0}},
/* 4879 */ {(11<<2)|2,{50,47,0}},
/* 4880 */ {(11<<2)|2,{50,47,0}},
/* 4881 */ {(11<<2)|2,{50,47,0}},
/* 4882 */ {(11<<2)|2,{50,47,0}},
/* 4883 */ {(11<<2)|2,{50,47,0}},
/* 4884 */ {(11<<2)|2,{50,47,0}},
/* 4885 */ {(11<<2)|2,{50,47,0}},
/* 4886 */ {(11<<2)|2,{50,47,0}},
/* 4887 */ {(11<<2)|2,{50,47,0}},
/* 4888 */ {(11<<2)|2,{50,47,0}},
/* 4889 */ {(11<<2)|2,{50,47,0}},
/* 4890 */ {(11<<2)|2,{50,47,0}},
/* 4891 */ {(11<<2)|2,{50,47,0}},
/* 4892 */ {(11<<2)|2,{50,47,0}},
/* 4893 */ {(11<<2)|2,{50,47,0}},
/* 4894 */ {(11<<2)|2,{50,47,0}},
/* 4895 */ {(11<<2)|2,{50,47,0}},
/* 4896 */ {(16<<2)|3,{50,51,48}},
/* 4897 */ {(16<<2)|3,{50,51,49}},
/* 4898 */ {(16<<2)|3,{50,51,50}},
/* 4899 */ {(16<<2)|3,{50,51,97}},
/* 4900 */ {(16<<2)|3,{50,51,99}},
/* 4901 */ {(16<<2)|3,{50,51,101}},
/* 4902 */ {(16<<2)|3,{50,51,105}},
/* 4903 */ {(16<<2)|3,{50,51,111}},
/* 4904 */ {(16<<2)|3,{50,51,115}},
/* 4905 */ {(16<<2)|3,{50,51,116}},
/* 4906 */ {(11<<2)|2,{50,51,0}},
/* 4907 */ {(11<<2)|2,{50,51,0}},
/* 4908 */ {(11<<2)|2,{50,51,0}},
/* 4909 */ {(11<<2)|2,{50,51,0}},
/* 4910 */ {(11<<2)|2,{50,51,0}},
/* 4911 */ {(11<<2)|2,{50,51,0}},
/* 4912 */ {(11<<2)|2,{50,51,0}},
/* 4913 */ {(11<<2)|2,{50,51,0}},
/* 4914 */ {(11<<2)|2,{50,51,0}},
/* 4915 */ {(11<<2)|2,{50,51,0}},
/* 4916 */ {(11<<2)|2,{50,51,0}},
/* 4917 */ {(11<<2)|2,{50,51,0}},
/* 4918 */ {(11<<2)|2,{50,51,0}},
/* 4919 */ {(11<<2)|2,{50,51,0}},
/* 4920 */ {(11<<2)|2,{50,51,0}},
/* 4921 */ {(11<<2)|2,{50,51,0}},
/* 4922 */ {(11<<2)|2,{50,51,0}},
/* 4923 */ {(11<<2)|2,{50,51,0}},
/* 4924 */ {(11<<2)|2,{50,51,0}},
/* 4925 */ {(11<<2)|2,{50,51,0}},
/* 4926 */ {(11<<2)|2,{50,51,0}},
/* 4927 */ {(11<<2)|2,{50,51,0}},
/* 4928 */ {(16<<2)|3,{50,52,48}},
/* 4929 */ {(16<<2)|3,{50,52,49}},
/* 4930 */ {(16<<2)|3,{50,52,50}},
/* 4931 */ {(16<<2)|3,{50,52,97}},
/* 4932 */ {(16<<2)|3,{50,52,99}},
/* 4933 */ {(16<<2)|3,{50,52,101}},
/* 4934 */ {(16<<2)|3,{50,52,105}},
/* 4935 */ {(16<<2)|3,{50,52,111}},
/* 4936 */ {(16<<2)|3,{50,52,115}},
/* 4937 */ {(16<<2)|3,{50,52,116}},
/* 4938 */ {(11<<2)|2,{50,52,0}},
/* 4939 */ {(11<<2)|2,{50,52,0}},
/* 4940 */ {(11<<2)|2,{50,52,0}},
/* 4941 */ {(11<<2)|2,{50,52,0}},
/* 4942 */ {(11<<2)|2,{50,52,0}},
/* 4943 */ {(11<<2)|2,{50,52,0}},
/* 4944 */ {(11<<2)|2,{50,52,0}},
/* 4945 */ {(11<<2)|2,{50,52,0}},
/* 4946 */ {(11<<2)|2,{50,52,0}},
/* 4947 */ {(11<<2)|2,{50,52,0}},
/* 4948 */ {(11<<2)|2,{50,52,0}},
/* 4949 */ {(11<<2)|2,{50,52,0}},
/* 4950 */ {(11<<2)|2,{50,52,0}},
/* 4951 */ {(11<<2)|2,{50,52,0}},
/* 4952 */ {(11<<2)|2,{50,52,0}},
/* 4953 */ {(11<<2)|2,{50,52,0}},
/* 4954 */ {(11<<2)|2,{50,52,0}},
/* 4955 */ {(11<<2)|2,{50,52,0}},
/* 4956 */ {(11<<2)|2,{50,52,0}},
/* 4957 */ {(11<<2)|2,{50,52,0}},
/* 4958 */ {(11<<2)|2,{50,52,0}},
/* 4959 */ {(11<<2)|2,{50,52,0}},
/* 4960 */ {(16<<2)|3,{50,53,48}},
/* 4961 */ {(16<<2)|3,{50,53,49}},
/* 4962 */ {(16<<2)|3,{50,53,50}},
/* 4963 */ {(16<<2)|3,{50,53,97}},
/* 4964 */ {(16<<2)|3,{50,53,99}},
/* 4965 */ {(16<<2)|3,{50,53,101}},
/* 4966 */ {(16<<2)|3,{50,53,105}},
/* 4967 */ {(16<<2)|3,{50,53,111}},
/* 4968 */ {(16<<2)|3,{50,53,115}},
/* 4969 */ {(16<<2)|3,{50,53,116}},
/* 4970 */ {(11<<2)|2,{50,53,0}},
/* 4971 */ {(11<<2)|2,{50,53,0}},
/* 4972 */ {(11<<2)|2,{50,53,0}},
/* 4973 */ {(11<<2)|2,{50,53,0}},
/* 4974 */ {(11<<2)|2,{50,53,0}},
/* 4975 */ {(11<<2)|2,{50,53,0}},
/* 4976 */ {(11<<2)|2,{50,53,0}},
/* 4977 */ {(11<<2)|2,{50,53,0}},
/* 4978 */ {(11<<2)|2,{50,53,0}},
/* 4979 */ {(11<<2)|2,{50,53,0}},
/* 4980 */ {(11<<2)|2,{50,53,0}},
/* 4981 */ {(11<<2)|2,{50,53,0}},
/* 4982 */ {(11<<2)|2,{50,53,0}},
/* 4983 */ {(11<<2)|2,{50,53,0}},
/* 4984 */ {(11<<2)|2,{50,53,0}},
/* 4985 */ {(11<<2)|2,{50,53,0}},
/* 4986 */ {(11<<2)|2,{50,53,0}},
/* 4987 */ {(11<<2)|2,{50,53,0}},
/* 4988 */ {(11<<2)|2,{50,53,0}},
/* 4989 */ {(11<<2)|2,{50,53,0}},
/* 4990 */ {(11<<2)|2,{50,53,0}},
/* 4991 */ {(11<<2)|2,{50,53,0}},
/* 4992 */ {(16<<2)|3,{50,54,48}},
/* 4993 */ {(16<<2)|3,{50,54,49}},
/* 4994 */ {(16<<2)|3,{50,54,50}},
/* 4995 */ {(16<<2)|3,{50,54,97}},
/* 4996 */ {(16<<2)|3,{50,54,99}},
/* 4997 */ {(16<<2)|3,{50,54,101}},
/* 4998 */ {(16<<2)|3,{50,54,105}},
/* 4999 */ {(16<<2)|3,{50,54,111}},
/* 5000 */ {(16<<2)|3,{50,54,115}},
/* 5001 */ {(16<<2)|3,{50,54,116}},
/* 5002 */ {(11<<2)|2,{50,54,0}},
/* 5003 */ {(11<<2)|2,{50,54,0}},
/* 5004 */ {(11<<2)|2,{50,54,0}},
/* 5005 */ {(11<<2)|2,{50,54,0}},
/* 5006 */ {(11<<2)|2,{50,54,0}},
/* 5007 */ {(11<<2)|2,{50,54,0}},
/* 5008 */ {(11<<2)|2,{50,54,0}},
/* 5009 */ {(11<<2)|2,{50,54,0}},
/* 5010 */ {(11<<2)|2,{50,54,0}},
/* 5011 */ {(11<<2)|2,{50,54,0}},
/* 5012 */ {(11<<2)|2,{50,54,0}},
/* 5013 */ {(11<<2)|2,{50,54,0}},
/* 5014 */ {(11<<2)|2,{50,54,0}},
/* 5015 */ {(11<<2)|2,{50,54,0}},
/* 5016 */ {(11<<2)|2,{50,54,0}},
/* 5017 */ {(11<<2)|2,{50,54,0}},
/* 5018 */ {(11<<2)|2,{50,54,0}},
/* 5019 */ {(11<<2)|2,{50,54,0}},
/* 5020 */ {(11<<2)|2,{50,54,0}},
/* 5021 */ {(11<<2)|2,{50,54,0}},
/* 5022 */ {(11<<2)|2,{50,54,0}},
/* 5023 */ {(11<<2)|2,{50,54,0}},
/* 5024 */ {(16<<2)|3,{50,55,48}},
/* 5025 */ {(16<<2)|3,{50,55,49}},
/* 5026 */ {(16<<2)|3,{50,55,50}},
/* 5027 */ {(16<<2)|3,{50,55,97}},
/* 5028 */ {(16<<2)|3,{50,55,99}},
/* 5029 */ {(16<<2)|3,{50,55,101}},
/* 5030 */ {(16<<2)|3,{50,55,105}},
/* 5031 */ {(16<<2)|3,{50,55,111}},
/* 5032 */ {(16<<2)|3,{50,55,115}},
/* 5033 */ {(16<<2)|3,{50,55,116}},
/* 5034 */ {(11<<2)|2,{50,55,0}},
/* 5035 */ {(11<<2)|2,{50,55,0}},
/* 5036 */ {(11<<2)|2,{50,55,0}},
/* 5037 */ {(11<<2)|2,{50,55,0}},
/* 5038 */ {(11<<2)|2,{50,55,0}},
/* 5039 */ {(11<<2)|2,{50,55,0}},
/* 5040 */ {(11<<2)|2,{50,55,0}},
/* 5041 */ {(11<<2)|2,{50,55,0}},
/* 5042 */ {(11<<2)|2,{50,55,0}},
/* 5043 */ {(11<<2)|2,{50,55,0}},
/* 5044 */ {(11<<2)|2,{50,55,0}},
/* 5045 */ {(11<<2)|2,{50,55,0}},
/* 5046 */ {(11<<2)|2,{50,55,0}},
/* 5047 */ {(11<<2)|2,{50,55,0}},
/* 5048 */ {(11<<2)|2,{50,55,0}},
/* 5049 */ {(11<<2)|2,{50,55,0}},
/* 5050 */ {(11<<2)|2,{50,55,0}},
/* 5051 */ {(11<<2)|2,{50,55,0}},
/* 5052 */ {(11<<2)|2,{50,55,0}},
/* 5053 */ {(11<<2)|2,{50,55,0}},
/* 5054 */ {(11<<2)|2,{50,55,0}},
/* 5055 */ {(11<<2)|2,{50,55,0}},
/* 5056 */ {(16<<2)|3,{50,56,48}},
/* 5057 */ {(16<<2)|3,{50,56,49}},
/* 5058 */ {(16<<2)|3,{50,56,50}},
/* 5059 */ {(16<<2)|3,{50,56,97}},
/* 5060 */ {(16<<2)|3,{50,56,99}},
/* 5061 */ {(16<<2)|3,{50,56,101}},
/* 5062 */ {(16<<2)|3,{50,56,105}},
/* 5063 */ {(16<<2)|3,{50,56,111}},
/* 5064 */ {(16<<2)|3,{50,56,115}},
/* 5065 */ {(16<<2)|3,{50,56,116}},
/* 5066 */ {(11<<2)|2,{50,56,0}},
/* 5067 */ {(11<<2)|2,{50,56,0}},
/* 5068 */ {(11<<2)|2,{50,56,0}},
/* 5069 */ {(11<<2)|2,{50,56,0}},
/* 5070 */ {(11<<2)|2,{50,56,0}},
/* 5071 */ {(11<<2)|2,{50,56,0}},
/* 5072 */ {(11<<2)|2,{50,56,0}},
/* 5073 */ {(11<<2)|2,{50,56,0}},
/* 5074 */ {(11<<2)|2,{50,56,0}},
/* 5075 */ {(11<<2)|2,{50,56,0}},
/* 5076 */ {(11<<2)|2,{50,56,0}},
/* 5077 */ {(11<<2)|2,{50,56,0}},
/* 5078 */ {(11<<2)|2,{50,56,0}},
/* 5079 */ {(11<<2)|2,{50,56,0}},
/* 5080 */ {(11<<2)|2,{50,56,0}},
/* 5081 */ {(11<<2)|2,{50,56,0}},
/* 5082 */ {(11<<2)|2,{50,56,0}},
/* 5083 */ {(11<<2)|2,{50,56,0}},
/* 5084 */ {(11<<2)|2,{50,56,0}},
/* 5085 */ {(11<<2)|2,{50,56,0}},
/* 5086 */ {(11<<2)|2,{50,56,0}},
/* 5087 */ {(11<<2)|2,{50,56,0}},
/* 5088 */ {(16<<2)|3,{50,57,48}},
/* 5089 */ {(16<<2)|3,{50,57,49}},
/* 5090 */ {(16<<2)|3,{50,57,50}},
/* 5091 */ {(16<<2)|3,{50,57,97}},
/* 5092 */ {(16<<2)|3,{50,57,99}},
/* 5093 */ {(16<<2)|3,{50,57,101}},
/* 5094 */ {(16<<2)|3,{50,57,105}},
/* 5095 */ {(16<<2)|3,{50,57,111}},
/* 5096 */ {(16<<2)|3,{50,57,115}},
/* 5097 */ {(16<<2)|3,{50,57,116}},
/* 5098 */ {(11<<2)|2,{50,57,0}},
/* 5099 */ {(11<<2)|2,{50,57,0}},
/* 5100 */ {(11<<2)|2,{50,57,0}},
/* 5101 */ {(11<<2)|2,{50,57,0}},
/* 5102 */ {(11<<2)|2,{50,57,0}},
/* 5103 */ {(11<<2)|2,{50,57,0}},
/* 5104 */ {(11<<2)|2,{50,57,0}},
/* 5105 */ {(11<<2)|2,{50,57,0}},
/* 5106 */ {(11<<2)|2,{50,57,0}},
/* 5107 */ {(11<<2)|2,{50,57,0}},
/* 5108 */ {(11<<2)|2,{50,57,0}},
/* 5109 */ {(11<<2)|2,{50,57,0}},
/* 5110 */ {(11<<2)|2,{50,57,0}},
/* 5111 */ {(11<<2)|2,{50,57,0}},
/* 5112 */ {(11<<2)|2,{50,57,0}},
/* 5113 */ {(11<<2)|2,{50,57,0}},
/* 5114 */ {(11<<2)|2,{50,57,0}},
/* 5115 */ {(11<<2)|2,{50,57,0}},
/* 5116 */ {(11<<2)|2,{50,57,0}},
/* 5117 */ {(11<<2)|2,{50,57,0}},
/* 5118 */ {(11<<2)|2,{50,57,0}},
/* 5119 */ {(11<<2)|2,{50,57,0}},
/* 5120 */ {(16<<2)|3,{50,61,48}},
/* 5121 */ {(16<<2)|3,{50,61,49}},
/* 5122 */ {(16<<2)|3,{50,61,50}},
/* 5123 */ {(16<<2)|3,{50,61,97}},
/* 5124 */ {(16<<2)|3,{50,61,99}},
/* 5125 */ {(16<<2)|3,{50,61,101}},
/* 5126 */ {(16<<2)|3,{50,61,105}},
/* 5127 */ {(16<<2)|3,{50,61,111}},
/* 5128 */ {(16<<2)|3,{50,61,115}},
/* 5129 */ {(16<<2)|3,{50,61,116}},
/* 5130 */ {(11<<2)|2,{50,61,0}},
/* 5131 */ {(11<<2)|2,{50,61,0}},
/* 5132 */ {(11<<2)|2,{50,61,0}},
/* 5133 */ {(11<<2)|2,{50,61,0}},
/* 5134 */ {(11<<2)|2,{50,61,0}},
/* 5135 */ {(11<<2)|2,{50,61,0}},
/* 5136 */ {(11<<2)|2,{50,61,0}},
/* 5137 */ {(11<<2)|2,{50,61,0}},
/* 5138 */ {(11<<2)|2,{50,61,0}},
/* 5139 */ {(11<<2)|2,{50,61,0}},
/* 5140 */ {(11<<2)|2,{50,61,0}},
/* 5141 */ {(11<<2)|2,{50,61,0}},
/* 5142 */ {(11<<2)|2,{50,61,0}},
/* 5143 */ {(11<<2)|2,{50,61,0}},
/* 5144 */ {(11<<2)|2,{50,61,0}},
/* 5145 */ {(11<<2)|2,{50,61,0}},
/* 5146 */ {(11<<2)|2,{50,61,0}},
/* 5147 */ {(11<<2)|2,{50,61,0}},
/* 5148 */ {(11<<2)|2,{50,61,0}},
/* 5149 */ {(11<<2)|2,{50,61,0}},
/* 5150 */ {(11<<2)|2,{50,61,0}},
/* 5151 */ {(11<<2)|2,{50,61,0}},
/* 5152 */ {(16<<2)|3,{50,65,48}},
/* 5153 */ {(16<<2)|3,{50,65,49}},
/* 5154 */ {(16<<2)|3,{50,65,50}},
/* 5155 */ {(16<<2)|3,{50,65,97}},
/* 5156 */ {(16<<2)|3,{50,65,99}},
/* 5157 */ {(16<<2)|3,{50,65,101}},
/* 5158 */ {(16<<2)|3,{50,65,105}},
/* 5159 */ {(16<<2)|3,{50,65,111}},
/* 5160 */ {(16<<2)|3,{50,65,115}},
/* 5161 */ {(16<<2)|3,{50,65,116}},
/* 5162 */ {(11<<2)|2,{50,65,0}},
/* 5163 */ {(11<<2)|2,{50,65,0}},
/* 5164 */ {(11<<2)|2,{50,65,0}},
/* 5165 */ {(11<<2)|2,{50,65,0}},
/* 5166 */ {(11<<2)|2,{50,65,0}},
/* 5167 */ {(11<<2)|2,{50,65,0}},
/* 5168 */ {(11<<2)|2,{50,65,0}},
/* 5169 */ {(11<<2)|2,{50,65,0}},
/* 5170 */ {(11<<2)|2,{50,65,0}},
/* 5171 */ {(11<<2)|2,{50,65,0}},
/* 5172 */ {(11<<2)|2,{50,65,0}},
/* 5173 */ {(11<<2)|2,{50,65,0}},
/* 5174 */ {(11<<2)|2,{50,65,0}},
/* 5175 */ {(11<<2)|2,{50,65,0}},
/* 5176 */ {(11<<2)|2,{50,65,0}},
/* 5177 */ {(11<<2)|2,{50,65,0}},
/* 5178 */ {(11<<2)|2,{50,65,0}},
/* 5179 */ {(11<<2)|2,{50,65,0}},
/* 5180 */ {(11<<2)|2,{50,65,0}},
/* 5181 */ {(11<<2)|2,{50,65,0}},
/* 5182 */ {(11<<2)|2,{50,65,0}},
/* 5183 */ {(11<<2)|2,{50,65,0}},
/* 5184 */ {(16<<2)|3,{50,95,48}},
/* 5185 */ {(16<<2)|3,{50,95,49}},
/* 5186 */ {(16<<2)|3,{50,95,50}},
/* 5187 */ {(16<<2)|3,{50,95,97}},
/* 5188 */ {(16<<2)|3,{50,95,99}},
/* 5189 */ {(16<<2)|3,{50,95,101}},
/* 5190 */ {(16<<2)|3,{50,95,105}},
/* 5191 */ {(16<<2)|3,{50,95,111}},
/* 5192 */ {(16<<2)|3,{50,95,115}},
/* 5193 */ {(16<<2)|3,{50,95,116}},
/* 5194 */ {(11<<2)|2,{50,95,0}},
/* 5195 */ {(11<<2)|2,{50,95,0}},
/* 5196 */ {(11<<2)|2,{50,95,0}},
/* 5197 */ {(11<<2)|2,{50,95,0}},
/* 5198 */ {(11<<2)|2,{50,95,0}},
/* 5199 */ {(11<<2)|2,{50,95,0}},
/* 5200 */ {(11<<2)|2,{50,95,0}},
/* 5201 */ {(11<<2)|2,{50,95,0}},
/* 5202 */ {(11<<2)|2,{50,95,0}},
/* 5203 */ {(11<<2)|2,{50,95,0}},
/* 5204 */ {(11<<2)|2,{50,95,0}},
/* 5205 */ {(11<<2)|2,{50,95,0}},
/* 5206 */ {(11<<2)|2,{50,95,0}},
/* 5207 */ {(11<<2)|2,{50,95,0}},
/* 5208 */ {(11<<2)|2,{50,95,0}},
/* 5209 */ {(11<<2)|2,{50,95,0}},
/* 5210 */ {(11<<2)|2,{50,95,0}},
/* 5211 */ {(11<<2)|2,{50,95,0}},
/* 5212 */ {(11<<2)|2,{50,95,0}},
/* 5213 */ {(11<<2)|2,{50,95,0}},
/* 5214 */ {(11<<2)|2,{50,95,0}},
/* 5215 */ {(11<<2)|2,{50,95,0}},
/* 5216 */ {(16<<2)|3,{50,98,48}},
/* 5217 */ {(16<<2)|3,{50,98,49}},
/* 5218 */ {(16<<2)|3,{50,98,50}},
/* 5219 */ {(16<<2)|3,{50,98,97}},
/* 5220 */ {(16<<2)|3,{50,98,99}},
/* 5221 */ {(16<<2)|3,{50,98,101}},
/* 5222 */ {(16<<2)|3,{50,98,105}},
/* 5223 */ {(16<<2)|3,{50,98,111}},
/* 5224 */ {(16<<2)|3,{50,98,115}},
/* 5225 */ {(16<<2)|3,{50,98,116}},
/* 5226 */ {(11<<2)|2,{50,98,0}},
/* 5227 */ {(11<<2)|2,{50,98,0}},
/* 5228 */ {(11<<2)|2,{50,98,0}},
/* 5229 */ {(11<<2)|2,{50,98,0}},
/* 5230 */ {(11<<2)|2,{50,98,0}},
/* 5231 */ {(11<<2)|2,{50,98,0}},
/* 5232 */ {(11<<2)|2,{50,98,0}},
/* 5233 */ {(11<<2)|2,{50,98,0}},
/* 5234 */ {(11<<2)|2,{50,98,0}},
/* 5235 */ {(11<<2)|2,{50,98,0}},
/* 5236 */ {(11<<2)|2,{50,98,0}},
/* 5237 */ {(11<<2)|2,{50,98,0}},
/* 5238 */ {(11<<2)|2,{50,98,0}},
/* 5239 */ {(11<<2)|2,{50,98,0}},
/* 5240 */ {(11<<2)|2,{50,98,0}},
/* 5241 */ {(11<<2)|2,{50,98,0}},
/* 5242 */ {(11<<2)|2,{50,98,0}},
/* 5243 */ {(11<<2)|2,{50,98,0}},
/* 5244 */ {(11<<2)|2,{50,98,0}},
/* 5245 */ {(11<<2)|2,{50,98,0}},
/* 5246 */ {(11<<2)|2,{50,98,0}},
/* 5247 */ {(11<<2)|2,{50,98,0}},
/* 5248 */ {(16<<2)|3,{50,100,48}},
/* 5249 */ {(16<<2)|3,{50,100,49}},
/* 5250 */ {(16<<2)|3,{50,100,50}},
/* 5251 */ {(16<<2)|3,{50,100,97}},
/* 5252 */ {(16<<2)|3,{50,100,99}},
/* 5253 */ {(16<<2)|3,{50,100,101}},
/* 5254 */ {(16<<2)|3,{50,100,105}},
/* 5255 */ {(16<<2)|3,{50,100,111}},
/* 5256 */ {(16<<2)|3,{50,100,115}},
/* 5257 */ {(16<<2)|3,{50,100,116}},
/* 5258 */ {(11<<2)|2,{50,100,0}},
/* 5259 */ {(11<<2)|2,{50,100,0}},
/* 5260 */ {(11<<2)|2,{50,100,0}},
/* 5261 */ {(11<<2)|2,{50,100,0}},
/* 5262 */ {(11<<2)|2,{50,100,0}},
/* 5263 */ {(11<<2)|2,{50,100,0}},
/* 5264 */ {(11<<2)|2,{50,100,0}},
/* 5265 */ {(11<<2)|2,{50,100,0}},
/* 5266 */ {(11<<2)|2,{50,100,0}},
/* 5267 */ {(11<<2)|2,{50,100,0}},
/* 5268 */ {(11<<2)|2,{50,100,0}},
/* 5269 */ {(11<<2)|2,{50,100,0}},
/* 5270 */ {(11<<2)|2,{50,100,0}},
/* 5271 */ {(11<<2)|2,{50,100,0}},
/* 5272 */ {(11<<2)|2,{50,100,0}},
/* 5273 */ {(11<<2)|2,{50,100,0}},
/* 5274 */ {(11<<2)|2,{50,100,0}},
/* 5275 */ {(11<<2)|2,{50,100,0}},
/* 5276 */ {(11<<2)|2,{50,100,0}},
/* 5277 */ {(11<<2)|2,{50,100,0}},
/* 5278 */ {(11<<2)|2,{50,100,0}},
/* 5279 */ {(11<<2)|2,{50,100,0}},
/* 5280 */ {(16<<2)|3,{50,102,48}},
/* 5281 */ {(16<<2)|3,{50,102,49}},
/* 5282 */ {(16<<2)|3,{50,102,50}},
/* 5283 */ {(16<<2)|3,{50,102,97}},
/* 5284 */ {(16<<2)|3,{50,102,99}},
/* 5285 */ {(16<<2)|3,{50,102,101}},
/* 5286 */ {(16<<2)|3,{50,102,105}},
/* 5287 */ {(16<<2)|3,{50,102,111}},
/* 5288 */ {(16<<2)|3,{50,102,115}},
/* 5289 */ {(16<<2)|3,{50,102,116}},
/* 5290 */ {(11<<2)|2,{50,102,0}},
/* 5291 */ {(11<<2)|2,{50,102,0}},
/* 5292 */ {(11<<2)|2,{50,102,0}},
/* 5293 */ {(11<<2)|2,{50,102,0}},
/* 5294 */ {(11<<2)|2,{50,102,0}},
/* 5295 */ {(11<<2)|2,{50,102,0}},
/* 5296 */ {(11<<2)|2,{50,102,0}},
/* 5297 */ {(11<<2)|2,{50,102,0}},
/* 5298 */ {(11<<2)|2,{50,102,0}},
/* 5299 */ {(11<<2)|2,{50,102,0}},
/* 5300 */ {(11<<2)|2,{50,102,0}},
/* 5301 */ {(11<<2)|2,{50,102,0}},
/* 5302 */ {(11<<2)|2,{50,102,0}},
/* 5303 */ {(11<<2)|2,{50,102,0}},
/* 5304 */ {(11<<2)|2,{50,102,0}},
/* 5305 */ {(11<<2)|2,{50,102,0}},
/* 5306 */ {(11<<2)|2,{50,102,0}},
/* 5307 */ {(11<<2)|2,{50,102,0}},
/* 5308 */ {(11<<2)|2,{50,102,0}},
/* 5309 */ {(11<<2)|2,{50,102,0}},
/* 5310 */ {(11<<2)|2,{50,102,0}},
/* 5311 */ {(11<<2)|2,{50,102,0}},
/* 5312 */ {(16<<2)|3,{50,103,48}},
/* 5313 */ {(16<<2)|3,{50,103,49}},
/* 5314 */ {(16<<2)|3,{50,103,50}},
/* 5315 */ {(16<<2)|3,{50,103,97}},
/* 5316 */ {(16<<2)|3,{50,103,99}},
/* 5317 */ {(16<<2)|3,{50,103,101}},
/* 5318 */ {(16<<2)|3,{50,103,105}},
/* 5319 */ {(16<<2)|3,{50,103,111}},
/* 5320 */ {(16<<2)|3,{50,103,115}},
/* 5321 */ {(16<<2)|3,{50,103,116}},
/* 5322 */ {(11<<2)|2,{50,103,0}},
/* 5323 */ {(11<<2)|2,{50,103,0}},
/* 5324 */ {(11<<2)|2,{50,103,0}},
/* 5325 */ {(11<<2)|2,{50,103,0}},
/* 5326 */ {(11<<2)|2,{50,103,0}},
/* 5327 */ {(11<<2)|2,{50,103,0}},
/* 5328 */ {(11<<2)|2,{50,103,0}},
/* 5329 */ {(11<<2)|2,{50,103,0}},
/* 5330 */ {(11<<2)|2,{50,103,0}},
/* 5331 */ {(11<<2)|2,{50,103,0}},
/* 5332 */ {(11<<2)|2,{50,103,0}},
/* 5333 */ {(11<<2)|2,{50,103,0}},
/* 5334 */ {(11<<2)|2,{50,103,0}},
/* 5335 */ {(11<<2)|2,{50,103,0}},
/* 5336 */ {(11<<2)|2,{50,103,0}},
/* 5337 */ {(11<<2)|2,{50,103,0}},
/* 5338 */ {(11<<2)|2,{50,103,0}},
/* 5339 */ {(11<<2)|2,{50,103,0}},
/* 5340 */ {(11<<2)|2,{50,103,0}},
/* 5341 */ {(11<<2)|2,{50,103,0}},
/* 5342 */ {(11<<2)|2,{50,103,0}},
/* 5343 */ {(11<<2)|2,{50,103,0}},
/* 5344 */ {(16<<2)|3,{50,104,48}},
/* 5345 */ {(16<<2)|3,{50,104,49}},
/* 5346 */ {(16<<2)|3,{50,104,50}},
/* 5347 */ {(16<<2)|3,{50,104,97}},
/* 5348 */ {(16<<2)|3,{50,104,99}},
/* 5349 */ {(16<<2)|3,{50,104,101}},
/* 5350 */ {(16<<2)|3,{50,104,105}},
/* 5351 */ {(16<<2)|3,{50,104,111}},
/* 5352 */ {(16<<2)|3,{50,104,115}},
/* 5353 */ {(16<<2)|3,{50,104,116}},
/* 5354 */ {(11<<2)|2,{50,104,0}},
/* 5355 */ {(11<<2)|2,{50,104,0}},
/* 5356 */ {(11<<2)|2,{50,104,0}},
/* 5357 */ {(11<<2)|2,{50,104,0}},
/* 5358 */ {(11<<2)|2,{50,104,0}},
/* 5359 */ {(11<<2)|2,{50,104,0}},
/* 5360 */ {(11<<2)|2,{50,104,0}},
/* 5361 */ {(11<<2)|2,{50,104,0}},
/* 5362 */ {(11<<2)|2,{50,104,0}},
/* 5363 */ {(11<<2)|2,{50,104,0}},
/* 5364 */ {(11<<2)|2,{50,104,0}},
/* 5365 */ {(11<<2)|2,{50,104,0}},
/* 5366 */ {(11<<2)|2,{50,104,0}},
/* 5367 */ {(11<<2)|2,{50,104,0}},
/* 5368 */ {(11<<2)|2,{50,104,0}},
/* 5369 */ {(11<<2)|2,{50,104,0}},
/* 5370 */ {(11<<2)|2,{50,104,0}},
/* 5371 */ {(11<<2)|2,{50,104,0}},
/* 5372 */ {(11<<2)|2,{50,104,0}},
/* 5373 */ {(11<<2)|2,{50,104,0}},
/* 5374 */ {(11<<2)|2,{50,104,0}},
/* 5375 */ {(11<<2)|2,{50,104,0}},
/* 5376 */ {(16<<2)|3,{50,108,48}},
/* 5377 */ {(16<<2)|3,{50,108,49}},
/* 5378 */ {(16<<2)|3,{50,108,50}},
/* 5379 */ {(16<<2)|3,{50,108,97}},
/* 5380 */ {(16<<2)|3,{50,108,99}},
/* 5381 */ {(16<<2)|3,{50,108,101}},
/* 5382 */ {(16<<2)|3,{50,108,105}},
/* 5383 */ {(16<<2)|3,{50,108,111}},
/* 5384 */ {(16<<2)|3,{50,108,115}},
/* 5385 */ {(16<<2)|3,{50,108,116}},
/* 5386 */ {(11<<2)|2,{50,108,0}},
/* 5387 */ {(11<<2)|2,{50,108,0}},
/* 5388 */ {(11<<2)|2,{50,108,0}},
/* 5389 */ {(11<<2)|2,{50,108,0}},
/* 5390 */ {(11<<2)|2,{50,108,0}},
/* 5391 */ {(11<<2)|2,{50,108,0}},
/* 5392 */ {(11<<2)|2,{50,108,0}},
/* 5393 */ {(11<<2)|2,{50,108,0}},
/* 5394 */ {(11<<2)|2,{50,108,0}},
/* 5395 */ {(11<<2)|2,{50,108,0}},
/* 5396 */ {(11<<2)|2,{50,108,0}},
/* 5397 */ {(11<<2)|2,{50,108,0}},
/* 5398 */ {(11<<2)|2,{50,108,0}},
/* 5399 */ {(11<<2)|2,{50,108,0}},
/* 5400 */ {(11<<2)|2,{50,108,0}},
/* 5401 */ {(11<<2)|2,{50,108,0}},
/* 5402 */ {(11<<2)|2,{50,108,0}},
/* 5403 */ {(11<<2)|2,{50,108,0}},
/* 5404 */ {(11<<2)|2,{50,108,0}},
/* 5405 */ {(11<<2)|2,{50,108,0}},
/* 5406 */ {(11<<2)|2,{50,108,0}},
/* 5407 */ {(11<<2)|2,{50,108,0}},
/* 5408 */ {(16<<2)|3,{50,109,48}},
/* 5409 */ {(16<<2)|3,{50,109,49}},
/* 5410 */ {(16<<2)|3,{50,109,50}},
/* 5411 */ {(16<<2)|3,{50,109,97}},
/* 5412 */ {(16<<2)|3,{50,109,99}},
/* 5413 */ {(16<<2)|3,{50,109,101}},
/* 5414 */ {(16<<2)|3,{50,109,105}},
/* 5415 */ {(16<<2)|3,{50,109,111}},
/* 5416 */ {(16<<2)|3,{50,109,115}},
/* 5417 */ {(16<<2)|3,{50,109,116}},
/* 5418 */ {(11<<2)|2,{50,109,0}},
/* 5419 */ {(11<<2)|2,{50,109,0}},
/* 5420 */ {(11<<2)|2,{50,109,0}},
/* 5421 */ {(11<<2)|2,{50,109,0}},
/* 5422 */ {(11<<2)|2,{50,109,0}},
/* 5423 */ {(11<<2)|2,{50,109,0}},
/* 5424 */ {(11<<2)|2,{50,109,0}},
/* 5425 */ {(11<<2)|2,{50,109,0}},
/* 5426 */ {(11<<2)|2,{50,109,0}},
/* 5427 */ {(11<<2)|2,{50,109,0}},
/* 5428 */ {(11<<2)|2,{50,109,0}},
/* 5429 */ {(11<<2)|2,{50,109,0}},
/* 5430 */ {(11<<2)|2,{50,109,0}},
/* 5431 */ {(11<<2)|2,{50,109,0}},
/* 5432 */ {(11<<2)|2,{50,109,0}},
/* 5433 */ {(11<<2)|2,{50,109,0}},
/* 5434 */ {(11<<2)|2,{50,109,0}},
/* 5435 */ {(11<<2)|2,{50,109,0}},
/* 5436 */ {(11<<2)|2,{50,109,0}},
/* 5437 */ {(11<<2)|2,{50,109,0}},
/* 5438 */ {(11<<2)|2,{50,109,0}},
/* 5439 */ {(11<<2)|2,{50,109,0}},
/* 5440 */ {(16<<2)|3,{50,110,48}},
/* 5441 */ {(16<<2)|3,{50,110,49}},
/* 5442 */ {(16<<2)|3,{50,110,50}},
/* 5443 */ {(16<<2)|3,{50,110,97}},
/* 5444 */ {(16<<2)|3,{50,110,99}},
/* 5445 */ {(16<<2)|3,{50,110,101}},
/* 5446 */ {(16<<2)|3,{50,110,105}},
/* 5447 */ {(16<<2)|3,{50,110,111}},
/* 5448 */ {(16<<2)|3,{50,110,115}},
/* 5449 */ {(16<<2)|3,{50,110,116}},
/* 5450 */ {(11<<2)|2,{50,110,0}},
/* 5451 */ {(11<<2)|2,{50,110,0}},
/* 5452 */ {(11<<2)|2,{50,110,0}},
/* 5453 */ {(11<<2)|2,{50,110,0}},
/* 5454 */ {(11<<2)|2,{50,110,0}},
/* 5455 */ {(11<<2)|2,{50,110,0}},
/* 5456 */ {(11<<2)|2,{50,110,0}},
/* 5457 */ {(11<<2)|2,{50,110,0}},
/* 5458 */ {(11<<2)|2,{50,110,0}},
/* 5459 */ {(11<<2)|2,{50,110,0}},
/* 5460 */ {(11<<2)|2,{50,110,0}},
/* 5461 */ {(11<<2)|2,{50,110,0}},
/* 5462 */ {(11<<2)|2,{50,110,0}},
/* 5463 */ {(11<<2)|2,{50,110,0}},
/* 5464 */ {(11<<2)|2,{50,110,0}},
/* 5465 */ {(11<<2)|2,{50,110,0}},
/* 5466 */ {(11<<2)|2,{50,110,0}},
/* 5467 */ {(11<<2)|2,{50,110,0}},
/* 5468 */ {(11<<2)|2,{50,110,0}},
/* 5469 */ {(11<<2)|2,{50,110,0}},
/* 5470 */ {(11<<2)|2,{50,110,0}},
/* 5471 */ {(11<<2)|2,{50,110,0}},
/* 5472 */ {(16<<2)|3,{50,112,48}},
/* 5473 */ {(16<<2)|3,{50,112,49}},
/* 5474 */ {(16<<2)|3,{50,112,50}},
/* 5475 */ {(16<<2)|3,{50,112,97}},
/* 5476 */ {(16<<2)|3,{50,112,99}},
/* 5477 */ {(16<<2)|3,{50,112,101}},
/* 5478 */ {(16<<2)|3,{50,112,105}},
/* 5479 */ {(16<<2)|3,{50,112,111}},
/* 5480 */ {(16<<2)|3,{50,112,115}},
/* 5481 */ {(16<<2)|3,{50,112,116}},
/* 5482 */ {(11<<2)|2,{50,112,0}},
/* 5483 */ {(11<<2)|2,{50,112,0}},
/* 5484 */ {(11<<2)|2,{50,112,0}},
/* 5485 */ {(11<<2)|2,{50,112,0}},
/* 5486 */ {(11<<2)|2,{50,112,0}},
/* 5487 */ {(11<<2)|2,{50,112,0}},
/* 5488 */ {(11<<2)|2,{50,112,0}},
/* 5489 */ {(11<<2)|2,{50,112,0}},
/* 5490 */ {(11<<2)|2,{50,112,0}},
/* 5491 */ {(11<<2)|2,{50,112,0}},
/* 5492 */ {(11<<2)|2,{50,112,0}},
/* 5493 */ {(11<<2)|2,{50,112,0}},
/* 5494 */ {(11<<2)|2,{50,112,0}},
/* 5495 */ {(11<<2)|2,{50,112,0}},
/* 5496 */ {(11<<2)|2,{50,112,0}},
/* 5497 */ {(11<<2)|2,{50,112,0}},
/* 5498 */ {(11<<2)|2,{50,112,0}},
/* 5499 */ {(11<<2)|2,{50,112,0}},
/* 5500 */ {(11<<2)|2,{50,112,0}},
/* 5501 */ {(11<<2)|2,{50,112,0}},
/* 5502 */ {(11<<2)|2,{50,112,0}},
/* 5503 */ {(11<<2)|2,{50,112,0}},
/* 5504 */ {(16<<2)|3,{50,114,48}},
/* 5505 */ {(16<<2)|3,{50,114,49}},
/* 5506 */ {(16<<2)|3,{50,114,50}},
/* 5507 */ {(16<<2)|3,{50,114,97}},
/* 5508 */ {(16<<2)|3,{50,114,99}},
/* 5509 */ {(16<<2)|3,{50,114,101}},
/* 5510 */ {(16<<2)|3,{50,114,105}},
/* 5511 */ {(16<<2)|3,{50,114,111}},
/* 5512 */ {(16<<2)|3,{50,114,115}},
/* 5513 */ {(16<<2)|3,{50,114,116}},
/* 5514 */ {(11<<2)|2,{50,114,0}},
/* 5515 */ {(11<<2)|2,{50,114,0}},
/* 5516 */ {(11<<2)|2,{50,114,0}},
/* 5517 */ {(11<<2)|2,{50,114,0}},
/* 5518 */ {(11<<2)|2,{50,114,0}},
/* 5519 */ {(11<<2)|2,{50,114,0}},
/* 5520 */ {(11<<2)|2,{50,114,0}},
/* 5521 */ {(11<<2)|2,{50,114,0}},
/* 5522 */ {(11<<2)|2,{50,114,0}},
/* 5523 */ {(11<<2)|2,{50,114,0}},
/* 5524 */ {(11<<2)|2,{50,114,0}},
/* 5525 */ {(11<<2)|2,{50,114,0}},
/* 5526 */ {(11<<2)|2,{50,114,0}},
/* 5527 */ {(11<<2)|2,{50,114,0}},
/* 5528 */ {(11<<2)|2,{50,114,0}},
/* 5529 */ {(11<<2)|2,{50,114,0}},
/* 5530 */ {(11<<2)|2,{50,114,0}},
/* 5531 */ {(11<<2)|2,{50,114,0}},
/* 5532 */ {(11<<2)|2,{50,114,0}},
/* 5533 */ {(11<<2)|2,{50,114,0}},
/* 5534 */ {(11<<2)|2,{50,114,0}},
/* 5535 */ {(11<<2)|2,{50,114,0}},
/* 5536 */ {(16<<2)|3,{50,117,48}},
/* 5537 */ {(16<<2)|3,{50,117,49}},
/* 5538 */ {(16<<2)|3,{50,117,50}},
/* 5539 */ {(16<<2)|3,{50,117,97}},
/* 5540 */ {(16<<2)|3,{50,117,99}},
/* 5541 */ {(16<<2)|3,{50,117,101}},
/* 5542 */ {(16<<2)|3,{50,117,105}},
/* 5543 */ {(16<<2)|3,{50,117,111}},
/* 5544 */ {(16<<2)|3,{50,117,115}},
/* 5545 */ {(16<<2)|3,{50,117,116}},
/* 5546 */ {(11<<2)|2,{50,117,0}},
/* 5547 */ {(11<<2)|2,{50,117,0}},
/* 5548 */ {(11<<2)|2,{50,117,0}},
/* 5549 */ {(11<<2)|2,{50,117,0}},
/* 5550 */ {(11<<2)|2,{50,117,0}},
/* 5551 */ {(11<<2)|2,{50,117,0}},
/* 5552 */ {(11<<2)|2,{50,117,0}},
/* 5553 */ {(11<<2)|2,{50,117,0}},
/* 5554 */ {(11<<2)|2,{50,117,0}},
/* 5555 */ {(11<<2)|2,{50,117,0}},
/* 5556 */ {(11<<2)|2,{50,117,0}},
/* 5557 */ {(11<<2)|2,{50,117,0}},
/* 5558 */ {(11<<2)|2,{50,117,0}},
/* 5559 */ {(11<<2)|2,{50,117,0}},
/* 5560 */ {(11<<2)|2,{50,117,0}},
/* 5561 */ {(11<<2)|2,{50,117,0}},
/* 5562 */ {(11<<2)|2,{50,117,0}},
/* 5563 */ {(11<<2)|2,{50,117,0}},
/* 5564 */ {(11<<2)|2,{50,117,0}},
/* 5565 */ {(11<<2)|2,{50,117,0}},
/* 5566 */ {(11<<2)|2,{50,117,0}},
/* 5567 */ {(11<<2)|2,{50,117,0}},
/* 5568 */ {(12<<2)|2,{50,58,0}},
/* 5569 */ {(12<<2)|2,{50,58,0}},
/* 5570 */ {(12<<2)|2,{50,58,0}},
/* 5571 */ {(12<<2)|2,{50,58,0}},
/* 5572 */ {(12<<2)|2,{50,58,0}},
/* 5573 */ {(12<<2)|2,{50,58,0}},
/* 5574 */ {(12<<2)|2,{50,58,0}},
/* 5575 */ {(12<<2)|2,{50,58,0}},
/* 5576 */ {(12<<2)|2,{50,58,0}},
/* 5577 */ {(12<<2)|2,{50,58,0}},
/* 5578 */ {(12<<2)|2,{50,58,0}},
/* 5579 */ {(12<<2)|2,{50,58,0}},
/* 5580 */ {(12<<2)|2,{50,58,0}},
/* 5581 */ {(12<<2)|2,{50,58,0}},
/* 5582 */ {(12<<2)|2,{50,58,0}},
/* 5583 */ {(12<<2)|2,{50,58,0}},
/* 5584 */ {(12<<2)|2,{50,66,0}},
/* 5585 */ {(12<<2)|2,{50,66,0}},
/* 5586 */ {(12<<2)|2,{50,66,0}},
/* 5587 */ {(12<<2)|2,{50,66,0}},
/* 5588 */ {(12<<2)|2,{50,66,0}},
/* 5589 */ {(12<<2)|2,{50,66,0}},
/* 5590 */ {(12<<2)|2,{50,66,0}},
/* 5591 */ {(12<<2)|2,{50,66,0}},
/* 5592 */ {(12<<2)|2,{50,66,0}},
/* 5593 */ {(12<<2)|2,{50,66,0}},
/* 5594 */ {(12<<2)|2,{50,66,0}},
/* 5595 */ {(12<<2)|2,{50,66,0}},
/* 5596 */ {(12<<2)|2,{50,66,0}},
/* 5597 */ {(12<<2)|2,{50,66,0}},
/* 5598 */ {(12<<2)|2,{50,66,0}},
/* 5599 */ {(12<<2)|2,{50,66,0}},
/* 5600 */ {(12<<2)|2,{50,67,0}},
/* 5601 */ {(12<<2)|2,{50,67,0}},
/* 5602 */ {(12<<2)|2,{50,67,0}},
/* 5603 */ {(12<<2)|2,{50,67,0}},
/* 5604 */ {(12<<2)|2,{50,67,0}},
/* 5605 */ {(12<<2)|2,{50,67,0}},
/* 5606 */ {(12<<2)|2,{50,67,0}},
/* 5607 */ {(12<<2)|2,{50,67,0}},
/* 5608 */ {(12<<2)|2,{50,67,0}},
/* 5609 */ {(12<<2)|2,{50,67,0}},
/* 5610 */ {(12<<2)|2,{50,67,0}},
/* 5611 */ {(12<<2)|2,{50,67,0}},
/* 5612 */ {(12<<2)|2,{50,67,0}},
/* 5613 */ {(12<<2)|2,{50,67,0}},
/* 5614 */ {(12<<2)|2,{50,67,0}},
/* 5615 */ {(12<<2)|2,{50,67,0}},
/* 5616 */ {(12<<2)|2,{50,68,0}},
/* 5617 */ {(12<<2)|2,{50,68,0}},
/* 5618 */ {(12<<2)|2,{50,68,0}},
/* 5619 */ {(12<<2)|2,{50,68,0}},
/* 5620 */ {(12<<2)|2,{50,68,0}},
/* 5621 */ {(12<<2)|2,{50,68,0}},
/* 5622 */ {(12<<2)|2,{50,68,0}},
/* 5623 */ {(12<<2)|2,{50,68,0}},
/* 5624 */ {(12<<2)|2,{50,68,0}},
/* 5625 */ {(12<<2)|2,{50,68,0}},
/* 5626 */ {(12<<2)|2,{50,68,0}},
/* 5627 */ {(12<<2)|2,{50,68,0}},
/* 5628 */ {(12<<2)|2,{50,68,0}},
/* 5629 */ {(12<<2)|2,{50,68,0}},
/* 5630 */ {(12<<2)|2,{50,68,0}},
/* 5631 */ {(12<<2)|2,{50,68,0}},
/* 5632 */ {(12<<2)|2,{50,69,0}},
/* 5633 */ {(12<<2)|2,{50,69,0}},
/* 5634 */ {(12<<2)|2,{50,69,0}},
/* 5635 */ {(12<<2)|2,{50,69,0}},
/* 5636 */ {(12<<2)|2,{50,69,0}},
/* 5637 */ {(12<<2)|2,{50,69,0}},
/* 5638 */ {(12<<2)|2,{50,69,0}},
/* 5639 */ {(12<<2)|2,{50,69,0}},
/* 5640 */ {(12<<2)|2,{50,69,0}},
/* 5641 */ {(12<<2)|2,{50,69,0}},
/* 5642 */ {(12<<2)|2,{50,69,0}},
/* 5643 */ {(12<<2)|2,{50,69,0}},
/* 5644 */ {(12<<2)|2,{50,69,0}},
/* 5645 */ {(12<<2)|2,{50,69,0}},
/* 5646 */ {(12<<2)|2,{50,69,0}},
/* 5647 */ {(12<<2)|2,{50,69,0}},
/* 5648 */ {(12<<2)|2,{50,70,0}},
/* 5649 */ {(12<<2)|2,{50,70,0}},
/* 5650 */ {(12<<2)|2,{50,70,0}},
/* 5651 */ {(12<<2)|2,{50,70,0}},
/* 5652 */ {(12<<2)|2,{50,70,0}},
/* 5653 */ {(12<<2)|2,{50,70,0}},
/* 5654 */ {(12<<2)|2,{50,70,0}},
/* 5655 */ {(12<<2)|2,{50,70,0}},
/* 5656 */ {(12<<2)|2,{50,70,0}},
/* 5657 */ {(12<<2)|2,{50,70,0}},
/* 5658 */ {(12<<2)|2,{50,70,0}},
/* 5659 */ {(12<<2)|2,{50,70,0}},
/* 5660 */ {(12<<2)|2,{50,70,0}},
/* 5661 */ {(12<<2)|2,{50,70,0}},
/* 5662 */ {(12<<2)|2,{50,70,0}},
/* 5663 */ {(12<<2)|2,{50,70,0}},
/* 5664 */ {(12<<2)|2,{50,71,0}},
/* 5665 */ {(12<<2)|2,{50,71,0}},
/* 5666 */ {(12<<2)|2,{50,71,0}},
/* 5667 */ {(12<<2)|2,{50,71,0}},
/* 5668 */ {(12<<2)|2,{50,71,0}},
/* 5669 */ {(12<<2)|2,{50,71,0}},
/* 5670 */ {(12<<2)|2,{50,71,0}},
/* 5671 */ {(12<<2)|2,{50,71,0}},
/* 5672 */ {(12<<2)|2,{50,71,0}},
/* 5673 */ {(12<<2)|2,{50,71,0}},
/* 5674 */ {(12<<2)|2,{50,71,0}},
/* 5675 */ {(12<<2)|2,{50,71,0}},
/* 5676 */ {(12<<2)|2,{50,71,0}},
/* 5677 */ {(12<<2)|2,{50,71,0}},
/* 5678 */ {(12<<2)|2,{50,71,0}},
/* 5679 */ {(12<<2)|2,{50,71,0}},
/* 5680 */ {(12<<2)|2,{50,72,0}},
/* 5681 */ {(12<<2)|2,{50,72,0}},
/* 5682 */ {(12<<2)|2,{50,72,0}},
/* 5683 */ {(12<<2)|2,{50,72,0}},
/* 5684 */ {(12<<2)|2,{50,72,0}},
/* 5685 */ {(12<<2)|2,{50,72,0}},
/* 5686 */ {(12<<2)|2,{50,72,0}},
/* 5687 */ {(12<<2)|2,{50,72,0}},
/* 5688 */ {(12<<2)|2,{50,72,0}},
/* 5689 */ {(12<<2)|2,{50,72,0}},
/* 5690 */ {(12<<2)|2,{50,72,0}},
/* 5691 */ {(12<<2)|2,{50,72,0}},
/* 5692 */ {(12<<2)|2,{50,72,0}},
/* 5693 */ {(12<<2)|2,{50,72,0}},
/* 5694 */ {(12<<2)|2,{50,72,0}},
/* 5695 */ {(12<<2)|2,{50,72,0}},
/* 5696 */ {(12<<2)|2,{50,73,0}},
/* 5697 */ {(12<<2)|2,{50,73,0}},
/* 5698 */ {(12<<2)|2,{50,73,0}},
/* 5699 */ {(12<<2)|2,{50,73,0}},
/* 5700 */ {(12<<2)|2,{50,73,0}},
/* 5701 */ {(12<<2)|2,{50,73,0}},
/* 5702 */ {(12<<2)|2,{50,73,0}},
/* 5703 */ {(12<<2)|2,{50,73,0}},
/* 5704 */ {(12<<2)|2,{50,73,0}},
/* 5705 */ {(12<<2)|2,{50,73,0}},
/* 5706 */ {(12<<2)|2,{50,73,0}},
/* 5707 */ {(12<<2)|2,{50,73,0}},
/* 5708 */ {(12<<2)|2,{50,73,0}},
/* 5709 */ {(12<<2)|2,{50,73,0}},
/* 5710 */ {(12<<2)|2,{50,73,0}},
/* 5711 */ {(12<<2)|2,{50,73,0}},
/* 5712 */ {(12<<2)|2,{50,74,0}},
/* 5713 */ {(12<<2)|2,{50,74,0}},
/* 5714 */ {(12<<2)|2,{50,74,0}},
/* 5715 */ {(12<<2)|2,{50,74,0}},
/* 5716 */ {(12<<2)|2,{50,74,0}},
/* 5717 */ {(12<<2)|2,{50,74,0}},
/* 5718 */ {(12<<2)|2,{50,74,0}},
/* 5719 */ {(12<<2)|2,{50,74,0}},
/* 5720 */ {(12<<2)|2,{50,74,0}},
/* 5721 */ {(12<<2)|2,{50,74,0}},
/* 5722 */ {(12<<2)|2,{50,74,0}},
/* 5723 */ {(12<<2)|2,{50,74,0}},
/* 5724 */ {(12<<2)|2,{50,74,0}},
/* 5725 */ {(12<<2)|2,{50,74,0}},
/* 5726 */ {(12<<2)|2,{50,74,0}},
/* 5727 */ {(12<<2)|2,{50,74,0}},
/* 5728 */ {(12<<2)|2,{50,75,0}},
/* 5729 */ {(12<<2)|2,{50,75,0}},
/* 5730 */ {(12<<2)|2,{50,75,0}},
/* 5731 */ {(12<<2)|2,{50,75,0}},
/* 5732 */ {(12<<2)|2,{50,75,0}},
/* 5733 */ {(12<<2)|2,{50,75,0}},
/* 5734 */ {(12<<2)|2,{50,75,0}},
/* 5735 */ {(12<<2)|2,{50,75,0}},
/* 5736 */ {(12<<2)|2,{50,75,0}},
/* 5737 */ {(12<<2)|2,{50,75,0}},
/* 5738 */ {(12<<2)|2,{50,75,0}},
/* 5739 */ {(12<<2)|2,{50,75,0}},
/* 5740 */ {(12<<2)|2,{50,75,0}},
/* 5741 */ {(12<<2)|2,{50,75,0}},
/* 5742 */ {(12<<2)|2,{50,75,0}},
/* 5743 */ {(12<<2)|2,{50,75,0}},
/* 5744 */ {(12<<2)|2,{50,76,0}},
/* 5745 */ {(12<<2)|2,{50,76,0}},
/* 5746 */ {(12<<2)|2,{50,76,0}},
/* 5747 */ {(12<<2)|2,{50,76,0}},
/* 5748 */ {(12<<2)|2,{50,76,0}},
/* 5749 */ {(12<<2)|2,{50,76,0}},
/* 5750 */ {(12<<2)|2,{50,76,0}},
/* 5751 */ {(12<<2)|2,{50,76,0}},
/* 5752 */ {(12<<2)|2,{50,76,0}},
/* 5753 */ {(12<<2)|2,{50,76,0}},
/* 5754 */ {(12<<2)|2,{50,76,0}},
/* 5755 */ {(12<<2)|2,{50,76,0}},
/* 5756 */ {(12<<2)|2,{50,76,0}},
/* 5757 */ {(12<<2)|2,{50,76,0}},
/* 5758 */ {(12<<2)|2,{50,76,0}},
/* 5759 */ {(12<<2)|2,{50,76,0}},
/* 5760 */ {(12<<2)|2,{50,77,0}},
/* 5761 */ {(12<<2)|2,{50,77,0}},
/* 5762 */ {(12<<2)|2,{50,77,0}},
/* 5763 */ {(12<<2)|2,{50,77,0}},
/* 5764 */ {(12<<2)|2,{50,77,0}},
/* 5765 */ {(12<<2)|2,{50,77,0}},
/* 5766 */ {(12<<2)|2,{50,77,0}},
/* 5767 */ {(12<<2)|2,{50,77,0}},
/* 5768 */ {(12<<2)|2,{50,77,0}},
/* 5769 */ {(12<<2)|2,{50,77,0}},
/* 5770 */ {(12<<2)|2,{50,77,0}},
/* 5771 */ {(12<<2)|2,{50,77,0}},
/* 5772 */ {(12<<2)|2,{50,77,0}},
/* 5773 */ {(12<<2)|2,{50,77,0}},
/* 5774 */ {(12<<2)|2,{50,77,0}},
/* 5775 */ {(12<<2)|2,{50,77,0}},
/* 5776 */ {(12<<2)|2,{50,78,0}},
/* 5777 */ {(12<<2)|2,{50,78,0}},
/* 5778 */ {(12<<2)|2,{50,78,0}},
/* 5779 */ {(12<<2)|2,{50,78,0}},
/* 5780 */ {(12<<2)|2,{50,78,0}},
/* 5781 */ {(12<<2)|2,{50,78,0}},
/* 5782 */ {(12<<2)|2,{50,78,0}},
/* 5783 */ {(12<<2)|2,{50,78,0}},
/* 5784 */ {(12<<2)|2,{50,78,0}},
/* 5785 */ {(12<<2)|2,{50,78,0}},
/* 5786 */ {(12<<2)|2,{50,78,0}},
/* 5787 */ {(12<<2)|2,{50,78,0}},
/* 5788 */ {(12<<2)|2,{50,78,0}},
/* 5789 */ {(12<<2)|2,{50,78,0}},
/* 5790 */ {(12<<2)|2,{50,78,0}},
/* 5791 */ {(12<<2)|2,{50,78,0}},
/* 5792 */ {(12<<2)|2,{50,79,0}},
/* 5793 */ {(12<<2)|2,{50,79,0}},
/* 5794 */ {(12<<2)|2,{50,79,0}},
/* 5795 */ {(12<<2)|2,{50,79,0}},
/* 5796 */ {(12<<2)|2,{50,79,0}},
/* 5797 */ {(12<<2)|2,{50,79,0}},
/* 5798 */ {(12<<2)|2,{50,79,0}},
/* 5799 */ {(12<<2)|2,{50,79,0}},
/* 5800 */ {(12<<2)|2,{50,79,0}},
/* 5801 */ {(12<<2)|2,{50,79,0}},
/* 5802 */ {(12<<2)|2,{50,79,0}},
/* 5803 */ {(12<<2)|2,{50,79,0}},
/* 5804 */ {(12<<2)|2,{50,79,0}},
/* 5805 */ {(12<<2)|2,{50,79,0}},
/* 5806 */ {(12<<2)|2,{50,79,0}},
/* 5807 */ {(12<<2)|2,{50,79,0}},
/* 5808 */ {(12<<2)|2,{50,80,0}},
/* 5809 */ {(12<<2)|2,{50,80,0}},
/* 5810 */ {(12<<2)|2,{50,80,0}},
/* 5811 */ {(12<<2)|2,{50,80,0}},
/* 5812 */ {(12<<2)|2,{50,80,0}},
/* 5813 */ {(12<<2)|2,{50,80,0}},
/* 5814 */ {(12<<2)|2,{50,80,0}},
/* 5815 */ {(12<<2)|2,{50,80,0}},
/* 5816 */ {(12<<2)|2,{50,80,0}},
/* 5817 */ {(12<<2)|2,{50,80,0}},
/* 5818 */ {(12<<2)|2,{50,80,0}},
/* 5819 */ {(12<<2)|2,{50,80,0}},
/* 5820 */ {(12<<2)|2,{50,80,0}},
/* 5821 */ {(12<<2)|2,{50,80,0}},
/* 5822 */ {(12<<2)|2,{50,80,0}},
/* 5823 */ {(12<<2)|2,{50,80,0}},
/* 5824 */ {(12<<2)|2,{50,81,0}},
/* 5825 */ {(12<<2)|2,{50,81,0}},
/* 5826 */ {(12<<2)|2,{50,81,0}},
/* 5827 */ {(12<<2)|2,{50,81,0}},
/* 5828 */ {(12<<2)|2,{50,81,0}},
/* 5829 */ {(12<<2)|2,{50,81,0}},
/* 5830 */ {(12<<2)|2,{50,81,0}},
/* 5831 */ {(12<<2)|2,{50,81,0}},
/* 5832 */ {(12<<2)|2,{50,81,0}},
/* 5833 */ {(12<<2)|2,{50,81,0}},
/* 5834 */ {(12<<2)|2,{50,81,0}},
/* 5835 */ {(12<<2)|2,{50,81,0}},
/* 5836 */ {(12<<2)|2,{50,81,0}},
/* 5837 */ {(12<<2)|2,{50,81,0}},
/* 5838 */ {(12<<2)|2,{50,81,0}},
/* 5839 */ {(12<<2)|2,{50,81,0}},
/* 5840 */ {(12<<2)|2,{50,82,0}},
/* 5841 */ {(12<<2)|2,{50,82,0}},
/* 5842 */ {(12<<2)|2,{50,82,0}},
/* 5843 */ {(12<<2)|2,{50,82,0}},
/* 5844 */ {(12<<2)|2,{50,82,0}},
/* 5845 */ {(12<<2)|2,{50,82,0}},
/* 5846 */ {(12<<2)|2,{50,82,0}},
/* 5847 */ {(12<<2)|2,{50,82,0}},
/* 5848 */ {(12<<2)|2,{50,82,0}},
/* 5849 */ {(12<<2)|2,{50,82,0}},
/* 5850 */ {(12<<2)|2,{50,82,0}},
/* 5851 */ {(12<<2)|2,{50,82,0}},
/* 5852 */ {(12<<2)|2,{50,82,0}},
/* 5853 */ {(12<<2)|2,{50,82,0}},
/* 5854 */ {(12<<2)|2,{50,82,0}},
/* 5855 */ {(12<<2)|2,{50,82,0}},
/* 5856 */ {(12<<2)|2,{50,83,0}},
/* 5857 */ {(12<<2)|2,{50,83,0}},
/* 5858 */ {(12<<2)|2,{50,83,0}},
/* 5859 */ {(12<<2)|2,{50,83,0}},
/* 5860 */ {(12<<2)|2,{50,83,0}},
/* 5861 */ {(12<<2)|2,{50,83,0}},
/* 5862 */ {(12<<2)|2,{50,83,0}},
/* 5863 */ {(12<<2)|2,{50,83,0}},
/* 5864 */ {(12<<2)|2,{50,83,0}},
/* 5865 */ {(12<<2)|2,{50,83,0}},
/* 5866 */ {(12<<2)|2,{50,83,0}},
/* 5867 */ {(12<<2)|2,{50,83,0}},
/* 5868 */ {(12<<2)|2,{50,83,0}},
/* 5869 */ {(12<<2)|2,{50,83,0}},
/* 5870 */ {(12<<2)|2,{50,83,0}},
/* 5871 */ {(12<<2)|2,{50,83,0}},
/* 5872 */ {(12<<2)|2,{50,84,0}},
/* 5873 */ {(12<<2)|2,{50,84,0}},
/* 5874 */ {(12<<2)|2,{50,84,0}},
/* 5875 */ {(12<<2)|2,{50,84,0}},
/* 5876 */ {(12<<2)|2,{50,84,0}},
/* 5877 */ {(12<<2)|2,{50,84,0}},
/* 5878 */ {(12<<2)|2,{50,84,0}},
/* 5879 */ {(12<<2)|2,{50,84,0}},
/* 5880 */ {(12<<2)|2,{50,84,0}},
/* 5881 */ {(12<<2)|2,{50,84,0}},
/* 5882 */ {(12<<2)|2,{50,84,0}},
/* 5883 */ {(12<<2)|2,{50,84,0}},
/* 5884 */ {(12<<2)|2,{50,84,0}},
/* 5885 */ {(12<<2)|2,{50,84,0}},
/* 5886 */ {(12<<2)|2,{50,84,0}},
/* 5887 */ {(12<<2)|2,{50,84,0}},
/* 5888 */ {(12<<2)|2,{50,85,0}},
/* 5889 */ {(12<<2)|2,{50,85,0}},
/* 5890 */ {(12<<2)|2,{50,85,0}},
/* 5891 */ {(12<<2)|2,{50,85,0}},
/* 5892 */ {(12<<2)|2,{50,85,0}},
/* 5893 */ {(12<<2)|2,{50,85,0}},
/* 5894 */ {(12<<2)|2,{50,85,0}},
/* 5895 */ {(12<<2)|2,{50,85,0}},
/* 5896 */ {(12<<2)|2,{50,85,0}},
/* 5897 */ {(12<<2)|2,{50,85,0}},
/* 5898 */ {(12<<2)|2,{50,85,0}},
/* 5899 */ {(12<<2)|2,{50,85,0}},
/* 5900 */ {(12<<2)|2,{50,85,0}},
/* 5901 */ {(12<<2)|2,{50,85,0}},
/* 5902 */ {(12<<2)|2,{50,85,0}},
/* 5903 */ {(12<<2)|2,{50,85,0}},
/* 5904 */ {(12<<2)|2,{50,86,0}},
/* 5905 */ {(12<<2)|2,{50,86,0}},
/* 5906 */ {(12<<2)|2,{50,86,0}},
/* 5907 */ {(12<<2)|2,{50,86,0}},
/* 5908 */ {(12<<2)|2,{50,86,0}},
/* 5909 */ {(12<<2)|2,{50,86,0}},
/* 5910 */ {(12<<2)|2,{50,86,0}},
/* 5911 */ {(12<<2)|2,{50,86,0}},
/* 5912 */ {(12<<2)|2,{50,86,0}},
/* 5913 */ {(12<<2)|2,{50,86,0}},
/* 5914 */ {(12<<2)|2,{50,86,0}},
/* 5915 */ {(12<<2)|2,{50,86,0}},
/* 5916 */ {(12<<2)|2,{50,86,0}},
/* 5917 */ {(12<<2)|2,{50,86,0}},
/* 5918 */ {(12<<2)|2,{50,86,0}},
/* 5919 */ {(12<<2)|2,{50,86,0}},
/* 5920 */ {(12<<2)|2,{50,87,0}},
/* 5921 */ {(12<<2)|2,{50,87,0}},
/* 5922 */ {(12<<2)|2,{50,87,0}},
/* 5923 */ {(12<<2)|2,{50,87,0}},
/* 5924 */ {(12<<2)|2,{50,87,0}},
/* 5925 */ {(12<<2)|2,{50,87,0}},
/* 5926 */ {(12<<2)|2,{50,87,0}},
/* 5927 */ {(12<<2)|2,{50,87,0}},
/* 5928 */ {(12<<2)|2,{50,87,0}},
/* 5929 */ {(12<<2)|2,{50,87,0}},
/* 5930 */ {(12<<2)|2,{50,87,0}},
/* 5931 */ {(12<<2)|2,{50,87,0}},
/* 5932 */ {(12<<2)|2,{50,87,0}},
/* 5933 */ {(12<<2)|2,{50,87,0}},
/* 5934 */ {(12<<2)|2,{50,87,0}},
/* 5935 */ {(12<<2)|2,{50,87,0}},
/* 5936 */ {(12<<2)|2,{50,89,0}},
/* 5937 */ {(12<<2)|2,{50,89,0}},
/* 5938 */ {(12<<2)|2,{50,89,0}},
/* 5939 */ {(12<<2)|2,{50,89,0}},
/* 5940 */ {(12<<2)|2,{50,89,0}},
/* 5941 */ {(12<<2)|2,{50,89,0}},
/* 5942 */ {(12<<2)|2,{50,89,0}},
/* 5943 */ {(12<<2)|2,{50,89,0}},
/* 5944 */ {(12<<2)|2,{50,89,0}},
/* 5945 */ {(12<<2)|2,{50,89,0}},
/* 5946 */ {(12<<2)|2,{50,89,0}},
/* 5947 */ {(12<<2)|2,{50,89,0}},
/* 5948 */ {(12<<2)|2,{50,89,0}},
/* 5949 */ {(12<<2)|2,{50,89,0}},
/* 5950 */ {(12<<2)|2,{50,89,0}},
/* 5951 */ {(12<<2)|2,{50,89,0}},
/* 5952 */ {(12<<2)|2,{50,106,0}},
/* 5953 */ {(12<<2)|2,{50,106,0}},
/* 5954 */ {(12<<2)|2,{50,106,0}},
/* 5955 */ {(12<<2)|2,{50,106,0}},
/* 5956 */ {(12<<2)|2,{50,106,0}},
/* 5957 */ {(12<<2)|2,{50,106,0}},
/* 5958 */ {(12<<2)|2,{50,106,0}},
/* 5959 */ {(12<<2)|2,{50,106,0}},
/* 5960 */ {(12<<2)|2,{50,106,0}},
/* 5961 */ {(12<<2)|2,{50,106,0}},
/* 5962 */ {(12<<2)|2,{50,106,0}},
/* 5963 */ {(12<<2)|2,{50,106,0}},
/* 5964 */ {(12<<2)|2,{50,106,0}},
/* 5965 */ {(12<<2)|2,{50,106,0}},
/* 5966 */ {(12<<2)|2,{50,106,0}},
/* 5967 */ {(12<<2)|2,{50,106,0}},
/* 5968 */ {(12<<2)|2,{50,107,0}},
/* 5969 */ {(12<<2)|2,{50,107,0}},
/* 5970 */ {(12<<2)|2,{50,107,0}},
/* 5971 */ {(12<<2)|2,{50,107,0}},
/* 5972 */ {(12<<2)|2,{50,107,0}},
/* 5973 */ {(12<<2)|2,{50,107,0}},
/* 5974 */ {(12<<2)|2,{50,107,0}},
/* 5975 */ {(12<<2)|2,{50,107,0}},
/* 5976 */ {(12<<2)|2,{50,107,0}},
/* 5977 */ {(12<<2)|2,{50,107,0}},
/* 5978 */ {(12<<2)|2,{50,107,0}},
/* 5979 */ {(12<<2)|2,{50,107,0}},
/* 5980 */ {(12<<2)|2,{50,107,0}},
/* 5981 */ {(12<<2)|2,{50,107,0}},
/* 5982 */ {(12<<2)|2,{50,107,0}},
/* 5983 */ {(12<<2)|2,{50,107,0}},
/* 5984 */ {(12<<2)|2,{50,113,0}},
/* 5985 */ {(12<<2)|2,{50,113,0}},
/* 5986 */ {(12<<2)|2,{50,113,0}},
/* 5987 */ {(12<<2)|2,{50,113,0}},
/* 5988 */ {(12<<2)|2,{50,113,0}},
/* 5989 */ {(12<<2)|2,{50,113,0}},
/* 5990 */ {(12<<2)|2,{50,113,0}},
/* 5991 */ {(12<<2)|2,{50,113,0}},
/* 5992 */ {(12<<2)|2,{50,113,0}},
/* 5993 */ {(12<<2)|2,{50,113,0}},
/* 5994 */ {(12<<2)|2,{50,113,0}},
/* 5995 */ {(12<<2)|2,{50,113,0}},
/* 5996 */ {(12<<2)|2,{50,113,0}},
/* 5997 */ {(12<<2)|2,{50,113,0}},
/* 5998 */ {(12<<2)|2,{50,113,0}},
/* 5999 */ {(12<<2)|2,{50,113,0}},
/* 6000 */ {(12<<2)|2,{50,118,0}},
/* 6001 */ {(12<<2)|2,{50,118,0}},
/* 6002 */ {(12<<2)|2,{50,118,0}},
/* 6003 */ {(12<<2)|2,{50,118,0}},
/* 6004 */ {(12<<2)|2,{50,118,0}},
/* 6005 */ {(12<<2)|2,{50,118,0}},
/* 6006 */ {(12<<2)|2,{50,118,0}},
/* 6007 */ {(12<<2)|2,{50,118,0}},
/* 6008 */ {(12<<2)|2,{50,118,0}},
/* 6009 */ {(12<<2)|2,{50,118,0}},
/* 6010 */ {(12<<2)|2,{50,118,0}},
/* 6011 */ {(12<<2)|2,{50,118,0}},
/* 6012 */ {(12<<2)|2,{50,118,0}},
/* 6013 */ {(12<<2)|2,{50,118,0}},
/* 6014 */ {(12<<2)|2,{50,118,0}},
/* 6015 */ {(12<<2)|2,{50,118,0}},
/* 6016 */ {(12<<2)|2,{50,119,0}},
/* 6017 */ {(12<<2)|2,{50,119,0}},
/* 6018 */ {(12<<2)|2,{50,119,0}},
/* 6019 */ {(12<<2)|2,{50,119,0}},
/* 6020 */ {(12<<2)|2,{50,119,0}},
/* 6021 */ {(12<<2)|2,{50,119,0}},
/* 6022 */ {(12<<2)|2,{50,119,0}},
/* 6023 */ {(12<<2)|2,{50,119,0}},
/* 6024 */ {(12<<2)|2,{50,119,0}},
/* 6025 */ {(12<<2)|2,{50,119,0}},
/* 6026 */ {(12<<2)|2,{50,119,0}},
/* 6027 */ {(12<<2)|2,{50,119,0}},
/* 6028 */ {(12<<2)|2,{50,119,0}},
/* 6029 */ {(12<<2)|2,{50,119,0}},
/* 6030 */ {(12<<2)|2,{50,119,0}},
/* 6031 */ {(12<<2)|2,{50,119,0}},
/* 6032 */ {(12<<2)|2,{50,120,0}},
/* 6033 */ {(12<<2)|2,{50,120,0}},
/* 6034 */ {(12<<2)|2,{50,120,0}},
/* 6035 */ {(12<<2)|2,{50,120,0}},
/* 6036 */ {(12<<2)|2,{50,120,0}},
/* 6037 */ {(12<<2)|2,{50,120,0}},
/* 6038 */ {(12<<2)|2,{50,120,0}},
/* 6039 */ {(12<<2)|2,{50,120,0}},
/* 6040 */ {(12<<2)|2,{50,120,0}},
/* 6041 */ {(12<<2)|2,{50,120,0}},
/* 6042 */ {(12<<2)|2,{50,120,0}},
/* 6043 */ {(12<<2)|2,{50,120,0}},
/* 6044 */ {(12<<2)|2,{50,120,0}},
/* 6045 */ {(12<<2)|2,{50,120,0}},
/* 6046 */ {(12<<2)|2,{50,120,0}},
/* 6047 */ {(12<<2)|2,{50,120,0}},
/* 6048 */ {(12<<2)|2,{50,121,0}},
/* 6049 */ {(12<<2)|2,{50,121,0}},
/* 6050 */ {(12<<2)|2,{50,121,0}},
/* 6051 */ {(12<<2)|2,{50,121,0}},
/* 6052 */ {(12<<2)|2,{50,121,0}},
/* 6053 */ {(12<<2)|2,{50,121,0}},
/* 6054 */ {(12<<2)|2,{50,121,0}},
/* 6055 */ {(12<<2)|2,{50,121,0}},
/* 6056 */ {(12<<2)|2,{50,121,0}},
/* 6057 */ {(12<<2)|2,{50,121,0}},
/* 6058 */ {(12<<2)|2,{50,121,0}},
/* 6059 */ {(12<<2)|2,{50,121,0}},
/* 6060 */ {(12<<2)|2,{50,121,0}},
/* 6061 */ {(12<<2)|2,{50,121,0}},
/* 6062 */ {(12<<2)|2,{50,121,0}},
/* 6063 */ {(12<<2)|2,{50,121,0}},
/* 6064 */ {(12<<2)|2,{50,122,0}},
/* 6065 */ {(12<<2)|2,{50,122,0}},
/* 6066 */ {(12<<2)|2,{50,122,0}},
/* 6067 */ {(12<<2)|2,{50,122,0}},
/* 6068 */ {(12<<2)|2,{50,122,0}},
/* 6069 */ {(12<<2)|2,{50,122,0}},
/* 6070 */ {(12<<2)|2,{50,122,0}},
/* 6071 */ {(12<<2)|2,{50,122,0}},
/* 6072 */ {(12<<2)|2,{50,122,0}},
/* 6073 */ {(12<<2)|2,{50,122,0}},
/* 6074 */ {(12<<2)|2,{50,122,0}},
/* 6075 */ {(12<<2)|2,{50,122,0}},
/* 6076 */ {(12<<2)|2,{50,122,0}},
/* 6077 */ {(12<<2)|2,{50,122,0}},
/* 6078 */ {(12<<2)|2,{50,122,0}},
/* 6079 */ {(12<<2)|2,{50,122,0}},
/* 6080 */ {(13<<2)|2,{50,38,0}},
/* 6081 */ {(13<<2)|2,{50,38,0}},
/* 6082 */ {(13<<2)|2,{50,38,0}},
/* 6083 */ {(13<<2)|2,{50,38,0}},
/* 6084 */ {(13<<2)|2,{50,38,0}},
/* 6085 */ {(13<<2)|2,{50,38,0}},
/* 6086 */ {(13<<2)|2,{50,38,0}},
/* 6087 */ {(13<<2)|2,{50,38,0}},
/* 6088 */ {(13<<2)|2,{50,42,0}},
/* 6089 */ {(13<<2)|2,{50,42,0}},
/* 6090 */ {(13<<2)|2,{50,42,0}},
/* 6091 */ {(13<<2)|2,{50,42,0}},
/* 6092 */ {(13<<2)|2,{50,42,0}},
/* 6093 */ {(13<<2)|2,{50,42,0}},
/* 6094 */ {(13<<2)|2,{50,42,0}},
/* 6095 */ {(13<<2)|2,{50,42,0}},
/* 6096 */ {(13<<2)|2,{50,44,0}},
/* 6097 */ {(13<<2)|2,{50,44,0}},
/* 6098 */ {(13<<2)|2,{50,44,0}},
/* 6099 */ {(13<<2)|2,{50,44,0}},
/* 6100 */ {(13<<2)|2,{50,44,0}},
/* 6101 */ {(13<<2)|2,{50,44,0}},
/* 6102 */ {(13<<2)|2,{50,44,0}},
/* 6103 */ {(13<<2)|2,{50,44,0}},
/* 6104 */ {(13<<2)|2,{50,59,0}},
/* 6105 */ {(13<<2)|2,{50,59,0}},
/* 6106 */ {(13<<2)|2,{50,59,0}},
/* 6107 */ {(13<<2)|2,{50,59,0}},
/* 6108 */ {(13<<2)|2,{50,59,0}},
/* 6109 */ {(13<<2)|2,{50,59,0}},
/* 6110 */ {(13<<2)|2,{50,59,0}},
/* 6111 */ {(13<<2)|2,{50,59,0}},
/* 6112 */ {(13<<2)|2,{50,88,0}},
/* 6113 */ {(13<<2)|2,{50,88,0}},
/* 6114 */ {(13<<2)|2,{50,88,0}},
/* 6115 */ {(13<<2)|2,{50,88,0}},
/* 6116 */ {(13<<2)|2,{50,88,0}},
/* 6117 */ {(13<<2)|2,{50,88,0}},
/* 6118 */ {(13<<2)|2,{50,88,0}},
/* 6119 */ {(13<<2)|2,{50,88,0}},
/* 6120 */ {(13<<2)|2,{50,90,0}},
/* 6121 */ {(13<<2)|2,{50,90,0}},
/* 6122 */ {(13<<2)|2,{50,90,0}},
/* 6123 */ {(13<<2)|2,{50,90,0}},
/* 6124 */ {(13<<2)|2,{50,90,0}},
/* 6125 */ {(13<<2)|2,{50,90,0}},
/* 6126 */ {(13<<2)|2,{50,90,0}},
/* 6127 */ {(13<<2)|2,{50,90,0}},
/* 6128 */ {(15<<2)|2,{50,33,0}},
/* 6129 */ {(15<<2)|2,{50,33,0}},
/* 6130 */ {(15<<2)|2,{50,34,0}},
/* 6131 */ {(15<<2)|2,{50,34,0}},
/* 6132 */ {(15<<2)|2,{50,40,0}},
/* 6133 */ {(15<<2)|2,{50,40,0}},
/* 6134 */ {(15<<2)|2,{50,41,0}},
/* 6135 */ {(15<<2)|2,{50,41,0}},
/* 6136 */ {(15<<2)|2,{50,63,0}},
/* 6137 */ {(15<<2)|2,{50,63,0}},
/* 6138 */ {(16<<2)|2,{50,39,0}},
/* 6139 */ {(16<<2)|2,{50,43,0}},
/* 6140 */ {(16<<2)|2,{50,124,0}},
/* 6141 */ {(5<<2)|1,{50,0,0}},
/* 6142 */ {(5<<2)|1,{50,0,0}},
/* 6143 */ {(5<<2)|1,{50,0,0}},
/* 6144 */ {(15<<2)|3,{97,48,48}},
/* 6145 */ {(15<<2)|3,{97,48,48}},
/* 6146 */ {(15<<2)|3,{97,48,49}},
/* 6147 */ {(15<<2)|3,{97,48,49}},
/* 6148 */ {(15<<2)|3,{97,48,50}},
/* 6149 */ {(15<<2)|3,{97,48,50}},
/* 6150 */ {(15<<2)|3,{97,48,97}},
/* 6151 */ {(15<<2)|3,{97,48,97}},
/* 6152 */ {(15<<2)|3,{97,48,99}},
/* 6153 */ {(15<<2)|3,{97,48,99}},
/* 6154 */ {(15<<2)|3,{97,48,101}},
/* 6155 */ {(15<<2)|3,{97,48,101}},
/* 6156 */ {(15<<2)|3,{97,48,105}},
/* 6157 */ {(15<<2)|3,{97,48,105}},
/* 6158 */ {(15<<2)|3,{97,48,111}},
/* 6159 */ {(15<<2)|3,{97,48,111}},
/* 6160 */ {(15<<2)|3,{97,48,115}},
/* 6161 */ {(15<<2)|3,{97,48,115}},
/* 6162 */ {(15<<2)|3,{97,48,116}},
/* 6163 */ {(15<<2)|3,{97,48,116}},
/* 6164 */ {(16<<2)|3,{97,48,32}},
/* 6165 */ {(16<<2)|3,{97,48,37}},
/* 6166 */ {(16<<2)|3,{97,48,45}},
/* 6167 */ {(16<<2)|3,{97,48,46}},
/* 6168 */ {(16<<2)|3,{97,48,47}},
/* 6169 */ {(16<<2)|3,{97,48,51}},
/* 6170 */ {(16<<2)|3,{97,48,52}},
/* 6171 */ {(16<<2)|3,{97,48,53}},
/* 6172 */ {(16<<2)|3,{97,48,54}},
/* 6173 */ {(16<<2)|3,{97,48,55}},
/* 6174 */ {(16<<2)|3,{97,48,56}},
/* 6175 */ {(16<<2)|3,{97,48,57}},
/* 6176 */ {(16<<2)|3,{97,48,61}},
/* 6177 */ {(16<<2)|3,{97,48,65}},
/* 6178 */ {(16<<2)|3,{97,48,95}},
/* 6179 */ {(16<<2)|3,{97,48,98}},
/* 6180 */ {(16<<2)|3,{97,48,100}},
/* 6181 */ {(16<<2)|3,{97,48,102}},
/* 6182 */ {(16<<2)|3,{97,48,103}},
/* 6183 */ {(16<<2)|3,{97,48,104}},
/* 6184 */ {(16<<2)|3,{97,48,108}},
/* 6185 */ {(16<<2)|3,{97,48,109}},
/* 6186 */ {(16<<2)|3,{97,48,110}},
/* 6187 */ {(16<<2)|3,{97,48,112}},
/* 6188 */ {(16<<2)|3,{97,48,114}},
/* 6189 */ {(16<<2)|3,{97,48,117}},
/* 6190 */ {(10<<2)|2,{97,48,0}},
/* 6191 */ {(10<<2)|2,{97,48,0}},
/* 6192 */ {(10<<2)|2,{97,48,0}},
/* 6193 */ {(10<<2)|2,{97,48,0}},
/* 6194 */ {(10<<2)|2,{97,48,0}},
/* 6195 */ {(10<<2)|2,{97,48,0}},
/* 6196 */ {(10<<2)|2,{97,48,0}},
/* 6197 */ {(10<<2)|2,{97,48,0}},
/* 6198 */ {(10<<2)|2,{97,48,0}},
/* 6199 */ {(10<<2)|2,{97,48,0}},
/* 6200 */ {(10<<2)|2,{97,48,0}},
/* 6201 */ {(10<<2)|2,{97,48,0}},
/* 6202 */ {(10<<2)|2,{97,48,0}},
/* 6203 */ {(10<<2)|2,{97,48,0}},
/* 6204 */ {(10<<2)|2,{97,48,0}},
/* 6205 */ {(10<<2)|2,{97,48,0}},
/* 6206 */ {(10<<2)|2,{97,48,0}},
/* 6207 */ {(10<<2)|2,{97,48,0}},
/* 6208 */ {(15<<2)|3,{97,49,48}},
/* 6209 */ {(15<<2)|3,{97,49,48}},
/* 6210 */ {(15<<2)|3,{97,49,49}},
/* 6211 */ {(15<<2)|3,{97,49,49}},
/* 6212 */ {(15<<2)|3,{97,49,50}},
/* 6213 */ {(15<<2)|3,{97,49,50}},
/* 6214 */ {(15<<2)|3,{97,49,97}},
/* 6215 */ {(15<<2)|3,{97,49,97}},
/* 6216 */ {(15<<2)|3,{97,49,99}},
/* 6217 */ {(15<<2)|3,{97,49,99}},
/* 6218 */ {(15<<2)|3,{97,49,101}},
/* 6219 */ {(15<<2)|3,{97,49,101}},
/* 6220 */ {(15<<2)|3,{97,49,105}},
/* 6221 */ {(15<<2)|3,{97,49,105}},
/* 6222 */ {(15<<2)|3,{97,49,111}},
/* 6223 */ {(15<<2)|3,{97,49,111}},
/* 6224 */ {(15<<2)|3,{97,49,115}},
/* 6225 */ {(15<<2)|3,{97,49,115}},
/* 6226 */ {(15<<2)|3,{97,49,116}},
/* 6227 */ {(15<<2)|3,{97,49,116}},
/* 6228 */ {(16<<2)|3,{97,49,32}},
/* 6229 */ {(16<<2)|3,{97,49,37}},
/* 6230 */ {(16<<2)|3,{97,49,45}},
/* 6231 */ {(16<<2)|3,{97,49,46}},
/* 6232 */ {(16<<2)|3,{97,49,47}},
/* 6233 */ {(16<<2)|3,{97,49,51}},
/* 6234 */ {(16<<2)|3,{97,49,52}},
/* 6235 */ {(16<<2)|3,{97,49,53}},
/* 6236 */ {(16<<2)|3,{97,49,54}},
/* 6237 */ {(16<<2)|3,{97,49,55}},
/* 6238 */ {(16<<2)|3,{97,49,56}},
/* 6239 */ {(16<<2)|3,{97,49,57}},
/* 6240 */ {(16<<2)|3,{97,49,61}},
/* 6241 */ {(16<<2)|3,{97,49,65}},
/* 6242 */ {(16<<2)|3,{97,49,95}},
/* 6243 */ {(16<<2)|3,{97,49,98}},
/* 6244 */ {(16<<2)|3,{97,49,100}},
/* 6245 */ {(16<<2)|3,{97,49,102}},
/* 6246 */ {(16<<2)|3,{97,49,103}},
/* 6247 */ {(16<<2)|3,{97,49,104}},
/* 6248 */ {(16<<2)|3,{97,49,108}},
/* 6249 */ {(16<<2)|3,{97,49,109}},
/* 6250 */ {(16<<2)|3,{97,49,110}},
/* 6251 */ {(16<<2)|3,{97,49,112}},
/* 6252 */ {(16<<2)|3,{97,49,114}},
/* 6253 */ {(16<<2)|3,{97,49,117}},
/* 6254 */ {(10<<2)|2,{97,49,0}},
/* 6255 */ {(10<<2)|2,{97,49,0}},
/* 6256 */ {(10<<2)|2,{97,49,0}},
/* 6257 */ {(10<<2)|2,{97,49,0}},
/* 6258 */ {(10<<2)|2,{97,49,0}},
/* 6259 */ {(10<<2)|2,{97,49,0}},
/* 6260 */ {(10<<2)|2,{97,49,0}},
/* 6261 */ {(10<<2)|2,{97,49,0}},
/* 6262 */ {(10<<2)|2,{97,49,0}},
/* 6263 */ {(10<<2)|2,{97,49,0}},
/* 6264 */ {(10<<2)|2,{97,49,0}},
/* 6265 */ {(10<<2)|2,{97,49,0}},
/* 6266 */ {(10<<2)|2,{97,49,0}},
/* 6267 */ {(10<<2)|2,{97,49,0}},
/* 6268 */ {(10<<2)|2,{97,49,0}},
/* 6269 */ {(10<<2)|2,{97,49,0}},
/* 6270 */ {(10<<2)|2,{97,49,0}},
/* 6271 */ {(10<<2)|2,{97,49,0}},
/* 6272 */ {(15<<2)|3,{97,50,48}},
/* 6273 */ {(15<<2)|3,{97,50,48}},
/* 6274 */ {(15<<2)|3,{97,50,49}},
/* 6275 */ {(15<<2)|3,{97,50,49}},
/* 6276 */ {(15<<2)|3,{97,50,50}},
/* 6277 */ {(15<<2)|3,{97,50,50}},
/* 6278 */ {(15<<2)|3,{97,50,97}},
/* 6279 */ {(15<<2)|3,{97,50,97}},
/* 6280 */ {(15<<2)|3,{97,50,99}},
/* 6281 */ {(15<<2)|3,{97,50,99}},
/* 6282 */ {(15<<2)|3,{97,50,101}},
/* 6283 */ {(15<<2)|3,{97,50,101}},
/* 6284 */ {(15<<2)|3,{97,50,105}},
/* 6285 */ {(15<<2)|3,{97,50,105}},
/* 6286 */ {(15<<2)|3,{97,50,111}},
/* 6287 */ {(15<<2)|3,{97,50,111}},
/* 6288 */ {(15<<2)|3,{97,50,115}},
/* 6289 */ {(15<<2)|3,{97,50,115}},
/* 6290 */ {(15<<2)|3,{97,50,116}},
/* 6291 */ {(15<<2)|3,{97,50,116}},
/* 6292 */ {(16<<2)|3,{97,50,32}},
/* 6293 */ {(16<<2)|3,{97,50,37}},
/* 6294 */ {(16<<2)|3,{97,50,45}},
/* 6295 */ {(16<<2)|3,{97,50,46}},
/* 6296 */ {(16<<2)|3,{97,50,47}},
/* 6297 */ {(16<<2)|3,{97,50,51}},
/* 6298 */ {(16<<2)|3,{97,50,52}},
/* 6299 */ {(16<<2)|3,{97,50,53}},
/* 6300 */ {(16<<2)|3,{97,50,54}},
/* 6301 */ {(16<<2)|3,{97,50,55}},
/* 6302 */ {(16<<2)|3,{97,50,56}},
/* 6303 */ {(16<<2)|3,{97,50,57}},
/* 6304 */ {(16<<2)|3,{97,50,61}},
/* 6305 */ {(16<<2)|3,{97,50,65}},
/* 6306 */ {(16<<2)|3,{97,50,95}},
/* 6307 */ {(16<<2)|3,{97,50,98}},
/* 6308 */ {(16<<2)|3,{97,50,100}},
/* 6309 */ {(16<<2)|3,{97,50,102}},
/* 6310 */ {(16<<2)|3,{97,50,103}},
/* 6311 */ {(16<<2)|3,{97,50,104}},
/* 6312 */ {(16<<2)|3,{97,50,108}},
/* 6313 */ {(16<<2)|3,{97,50,109}},
/* 6314 */ {(16<<2)|3,{97,50,110}},
/* 6315 */ {(16<<2)|3,{97,50,112}},
/* 6316 */ {(16<<2)|3,{97,50,114}},
/* 6317 */ {(16<<2)|3,{97,50,117}},
/* 6318 */ {(10<<2)|2,{97,50,0}},
/* 6319 */ {(10<<2)|2,{97,50,0}},
/* 6320 */ {(10<<2)|2,{97,50,0}},
/* 6321 */ {(10<<2)|2,{97,50,0}},
/* 6322 */ {(10<<2)|2,{97,50,0}},
/* 6323 */ {(10<<2)|2,{97,50,0}},
/* 6324 */ {(10<<2)|2,{97,50,0}},
/* 6325 */ {(10<<2)|2,{97,50,0}},
/* 6326 */ {(10<<2)|2,{97,50,0}},
/* 6327 */ {(10<<2)|2,{97,50,0}},
/* 6328 */ {(10<<2)|2,{97,50,0}},
/* 6329 */ {(10<<2)|2,{97,50,0}},
/* 6330 */ {(10<<2)|2,{97,50,0}},
/* 6331 */ {(10<<2)|2,{97,50,0}},
/* 6332 */ {(10<<2)|2,{97,50,0}},
/* 6333 */ {(10<<2)|2,{97,50,0}},
/* 6334 */ {(10<<2)|2,{97,50,0}},
/* 6335 */ {(10<<2)|2,{97,50,0}},
/* 6336 */ {(15<<2)|3,{97,97,48}},
/* 6337 */ {(15<<2)|3,{97,97,48}},
/* 6338 */ {(15<<2)|3,{97,97,49}},
/* 6339 */ {(15<<2)|3,{97,97,49}},
/* 6340 */ {(15<<2)|3,{97,97,50}},
/* 6341 */ {(15<<2)|3,{97,97,50}},
/* 6342 */ {(15<<2)|3,{97,97,97}},
/* 6343 */ {(15<<2)|3,{97,97,97}},
/* 6344 */ {(15<<2)|3,{97,97,99}},
/* 6345 */ {(15<<2)|3,{97,97,99}},
/* 6346 */ {(15<<2)|3,{97,97,101}},
/* 6347 */ {(15<<2)|3,{97,97,101}},
/* 6348 */ {(15<<2)|3,{97,97,105}},
/* 6349 */ {(15<<2)|3,{97,97,105}},
/* 6350 */ {(15<<2)|3,{97,97,111}},
/* 6351 */ {(15<<2)|3,{97,97,111}},
/* 6352 */ {(15<<2)|3,{97,97,115}},
/* 6353 */ {(15<<2)|3,{97,97,115}},
/* 6354 */ {(15<<2)|3,{97,97,116}},
/* 6355 */ {(15<<2)|3,{97,97,116}},
/* 6356 */ {(16<<2)|3,{97,97,32}},
/* 6357 */ {(16<<2)|3,{97,97,37}},
/* 6358 */ {(16<<2)|3,{97,97,45}},
/* 6359 */ {(16<<2)|3,{97,97,46}},
/* 6360 */ {(16<<2)|3,{97,97,47}},
/* 6361 */ {(16<<2)|3,{97,97,51}},
/* 6362 */ {(16<<2)|3,{97,97,52}},
/* 6363 */ {(16<<2)|3,{97,97,53}},
/* 6364 */ {(16<<2)|3,{97,97,54}},
/* 6365 */ {(16<<2)|3,{97,97,55}},
/* 6366 */ {(16<<2)|3,{97,97,56}},
/* 6367 */ {(16<<2)|3,{97,97,57}},
/* 6368 */ {(16<<2)|3,{97,97,61}},
/* 6369 */ {(16<<2)|3,{97,97,65}},
/* 6370 */ {(16<<2)|3,{97,97,95}},
/* 6371 */ {(16<<2)|3,{97,97,98}},
/* 6372 */ {(16<<2)|3,{97,97,100}},
/* 6373 */ {(16<<2)|3,{97,97,102}},
/* 6374 */ {(16<<2)|3,{97,97,103}},
/* 6375 */ {(16<<2)|3,{97,97,104}},
/* 6376 */ {(16<<2)|3,{97,97,108}},
/* 6377 */ {(16<<2)|3,{97,97,109}},
/* 6378 */ {(16<<2)|3,{97,97,110}},
/* 6379 */ {(16<<2)|3,{97,97,112}},
/* 6380 */ {(16<<2)|3,{97,97,114}},
/* 6381 */ {(16<<2)|3,{97,97,117}},
/* 6382 */ {(10<<2)|2,{97,97,0}},
/* 6383 */ {(10<<2)|2,{97,97,0}},
/* 6384 */ {(10<<2)|2,{97,97,0}},
/* 6385 */ {(10<<2)|2,{97,97,0}},
/* 6386 */ {(10<<2)|2,{97,97,0}},
/* 6387 */ {(10<<2)|2,{97,97,0}},
/* 6388 */ {(10<<2)|2,{97,97,0}},
/* 6389 */ {(10<<2)|2,{97,97,0}},
/* 6390 */ {(10<<2)|2,{97,97,0}},
/* 6391 */ {(10<<2)|2,{97,97,0}},
/* 6392 */ {(10<<2)|2,{97,97,0}},
/* 6393 */ {(10<<2)|2,{97,97,0}},
/* 6394 */ {(10<<2)|2,{97,97,0}},
/* 6395 */ {(10<<2)|2,{97,97,0}},
/* 6396 */ {(10<<2)|2,{97,97,0}},
/* 6397 */ {(10<<2)|2,{97,97,0}},
/* 6398 */ {(10<<2)|2,{97,97,0}},
/* 6399 */ {(10<<2)|2,{97,97,0}},
/* 6400 */ {(15<<2)|3,{97,99,48}},
/* 6401 */ {(15<<2)|3,{97,99,48}},
/* 6402 */ {(15<<2)|3,{97,99,49}},
/* 6403 */ {(15<<2)|3,{97,99,49}},
/* 6404 */ {(15<<2)|3,{97,99,50}},
/* 6405 */ {(15<<2)|3,{97,99,50}},
/* 6406 */ {(15<<2)|3,{97,99,97}},
/* 6407 */ {(15<<2)|3,{97,99,97}},
/* 6408 */ {(15<<2)|3,{97,99,99}},
/* 6409 */ {(15<<2)|3,{97,99,99}},
/* 6410 */ {(15<<2)|3,{97,99,101}},
/* 6411 */ {(15<<2)|3,{97,99,101}},
/* 6412 */ {(15<<2)|3,{97,99,105}},
/* 6413 */ {(15<<2)|3,{97,99,105}},
/* 6414 */ {(15<<2)|3,{97,99,111}},
/* 6415 */ {(15<<2)|3,{97,99,111}},
/* 6416 */ {(15<<2)|3,{97,99,115}},
/* 6417 */ {(15<<2)|3,{97,99,115}},
/* 6418 */ {(15<<2)|3,{97,99,116}},
/* 6419 */ {(15<<2)|3,{97,99,116}},
/* 6420 */ {(16<<2)|3,{97,99,32}},
/* 6421 */ {(16<<2)|3,{97,99,37}},
/* 6422 */ {(16<<2)|3,{97,99,45}},
/* 6423 */ {(16<<2)|3,{97,99,46}},
/* 6424 */ {(16<<2)|3,{97,99,47}},
/* 6425 */ {(16<<2)|3,{97,99,51}},
/* 6426 */ {(16<<2)|3,{97,99,52}},
/* 6427 */ {(16<<2)|3,{97,99,53}},
/* 6428 */ {(16<<2)|3,{97,99,54}},
/* 6429 */ {(16<<2)|3,{97,99,55}},
/* 6430 */ {(16<<2)|3,{97,99,56}},
/* 6431 */ {(16<<2)|3,{97,99,57}},
/* 6432 */ {(16<<2)|3,{97,99,61}},
/* 6433 */ {(16<<2)|3,{97,99,65}},
/* 6434 */ {(16<<2)|3,{97,99,95}},
/* 6435 */ {(16<<2)|3,{97,99,98}},
/* 6436 */ {(16<<2)|3,{97,99,100}},
/* 6437 */ {(16<<2)|3,{97,99,102}},
/* 6438 */ {(16<<2)|3,{97,99,103}},
/* 6439 */ {(16<<2)|3,{97,99,104}},
/* 6440 */ {(16<<2)|3,{97,99,108}},
/* 6441 */ {(16<<2)|3,{97,99,109}},
/* 6442 */ {(16<<2)|3,{97,99,110}},
/* 6443 */ {(16<<2)|3,{97,99,112}},
/* 6444 */ {(16<<2)|3,{97,99,114}},
/* 6445 */ {(16<<2)|3,{97,99,117}},
/* 6446 */ {(10<<2)|2,{97,99,0}},
/* 6447 */ {(10<<2)|2,{97,99,0}},
/* 6448 */ {(10<<2)|2,{97,99,0}},
/* 6449 */ {(10<<2)|2,{97,99,0}},
/* 6450 */ {(10<<2)|2,{97,99,0}},
/* 6451 */ {(10<<2)|2,{97,99,0}},
/* 6452 */ {(10<<2)|2,{97,99,0}},
/* 6453 */ {(10<<2)|2,{97,99,0}},
/* 6454 */ {(10<<2)|2,{97,99,0}},
/* 6455 */ {(10<<2)|2,{97,99,0}},
/* 6456 */ {(10<<2)|2,{97,99,0}},
/* 6457 */ {(10<<2)|2,{97,99,0}},
/* 6458 */ {(10<<2)|2,{97,99,0}},
/* 6459 */ {(10<<2)|2,{97,99,0}},
/* 6460 */ {(10<<2)|2,{97,99,0}},
/* 6461 */ {(10<<2)|2,{97,99,0}},
/* 6462 */ {(10<<2)|2,{97,99,0}},
/* 6463 */ {(10<<2)|2,{97,99,0}},
/* 6464 */ {(15<<2)|3,{97,101,48}},
/* 6465 */ {(15<<2)|3,{97,101,48}},
/* 6466 */ {(15<<2)|3,{97,101,49}},
/* 6467 */ {(15<<2)|3,{97,101,49}},
/* 6468 */ {(15<<2)|3,{97,101,50}},
/* 6469 */ {(15<<2)|3,{97,101,50}},
/* 6470 */ {(15<<2)|3,{97,101,97}},
/* 6471 */ {(15<<2)|3,{97,101,97}},
/* 6472 */ {(15<<2)|3,{97,101,99}},
/* 6473 */ {(15<<2)|3,{97,101,99}},
/* 6474 */ {(15<<2)|3,{97,101,101}},
/* 6475 */ {(15<<2)|3,{97,101,101}},
/* 6476 */ {(15<<2)|3,{97,101,105}},
/* 6477 */ {(15<<2)|3,{97,101,105}},
/* 6478 */ {(15<<2)|3,{97,101,111}},
/* 6479 */ {(15<<2)|3,{97,101,111}},
/* 6480 */ {(15<<2)|3,{97,101,115}},
/* 6481 */ {(15<<2)|3,{97,101,115}},
/* 6482 */ {(15<<2)|3,{97,101,116}},
/* 6483 */ {(15<<2)|3,{97,101,116}},
/* 6484 */ {(16<<2)|3,{97,101,32}},
/* 6485 */ {(16<<2)|3,{97,101,37}},
/* 6486 */ {(16<<2)|3,{97,101,45}},
/* 6487 */ {(16<<2)|3,{97,101,46}},
/* 6488 */ {(16<<2)|3,{97,101,47}},
/* 6489 */ {(16<<2)|3,{97,101,51}},
/* 6490 */ {(16<<2)|3,{97,101,52}},
/* 6491 */ {(16<<2)|3,{97,101,53}},
/* 6492 */ {(16<<2)|3,{97,101,54}},
/* 6493 */ {(16<<2)|3,{97,101,55}},
/* 6494 */ {(16<<2)|3,{97,101,56}},
/* 6495 */ {(16<<2)|3,{97,101,57}},
/* 6496 */ {(16<<2)|3,{97,101,61}},
/* 6497 */ {(16<<2)|3,{97,101,65}},
/* 6498 */ {(16<<2)|3,{97,101,95}},
/* 6499 */ {(16<<2)|3,{97,101,98}},
/* 6500 */ {(16<<2)|3,{97,101,100}},
/* 6501 */ {(16<<2)|3,{97,101,102}},
/* 6502 */ {(16<<2)|3,{97,101,103}},
/* 6503 */ {(16<<2)|3,{97,101,104}},
/* 6504 */ {(16<<2)|3,{97,101,108}},
/* 6505 */ {(16<<2)|3,{97,101,109}},
/* 6506 */ {(16<<2)|3,{97,101,110}},
/* 6507 */ {(16<<2)|3,{97,101,112}},
/* 6508 */ {(16<<2)|3,{97,101,114}},
/* 6509 */ {(16<<2)|3,{97,101,117}},
/* 6510 */ {(10<<2)|2,{97,101,0}},
/* 6511 */ {(10<<2)|2,{97,101,0}},
/* 6512 */ {(10<<2)|2,{97,101,0}},
/* 6513 */ {(10<<2)|2,{97,101,0}},
/* 6514 */ {(10<<2)|2,{97,101,0}},
/* 6515 */ {(10<<2)|2,{97,101,0}},
/* 6516 */ {(10<<2)|2,{97,101,0}},
/* 6517 */ {(10<<2)|2,{97,101,0}},
/* 6518 */ {(10<<2)|2,{97,101,0}},
/* 6519 */ {(10<<2)|2,{97,101,0}},
/* 6520 */ {(10<<2)|2,{97,101,0}},
/* 6521 */ {(10<<2)|2,{97,101,0}},
/* 6522 */ {(10<<2)|2,{97,101,0}},
/* 6523 */ {(10<<2)|2,{97,101,0}},
/* 6524 */ {(10<<2)|2,{97,101,0}},
/* 6525 */ {(10<<2)|2,{97,101,0}},
/* 6526 */ {(10<<2)|2,{97,101,0}},
/* 6527 */ {(10<<2)|2,{97,101,0}},
/* 6528 */ {(15<<2)|3,{97,105,48}},
/* 6529 */ {(15<<2)|3,{97,105,48}},
/* 6530 */ {(15<<2)|3,{97,105,49}},
/* 6531 */ {(15<<2)|3,{97,105,49}},
/* 6532 */ {(15<<2)|3,{97,105,50}},
/* 6533 */ {(15<<2)|3,{97,105,50}},
/* 6534 */ {(15<<2)|3,{97,105,97}},
/* 6535 */ {(15<<2)|3,{97,105,97}},
/* 6536 */ {(15<<2)|3,{97,105,99}},
/* 6537 */ {(15<<2)|3,{97,105,99}},
/* 6538 */ {(15<<2)|3,{97,105,101}},
/* 6539 */ {(15<<2)|3,{97,105,101}},
/* 6540 */ {(15<<2)|3,{97,105,105}},
/* 6541 */ {(15<<2)|3,{97,105,105}},
/* 6542 */ {(15<<2)|3,{97,105,111}},
/* 6543 */ {(15<<2)|3,{97,105,111}},
/* 6544 */ {(15<<2)|3,{97,105,115}},
/* 6545 */ {(15<<2)|3,{97,105,115}},
/* 6546 */ {(15<<2)|3,{97,105,116}},
/* 6547 */ {(15<<2)|3,{97,105,116}},
/* 6548 */ {(16<<2)|3,{97,105,32}},
/* 6549 */ {(16<<2)|3,{97,105,37}},
/* 6550 */ {(16<<2)|3,{97,105,45}},
/* 6551 */ {(16<<2)|3,{97,105,46}},
/* 6552 */ {(16<<2)|3,{97,105,47}},
/* 6553 */ {(16<<2)|3,{97,105,51}},
/* 6554 */ {(16<<2)|3,{97,105,52}},
/* 6555 */ {(16<<2)|3,{97,105,53}},
/* 6556 */ {(16<<2)|3,{97,105,54}},
/* 6557 */ {(16<<2)|3,{97,105,55}},
/* 6558 */ {(16<<2)|3,{97,105,56}},
/* 6559 */ {(16<<2)|3,{97,105,57}},
/* 6560 */ {(16<<2)|3,{97,105,61}},
/* 6561 */ {(16<<2)|3,{97,105,65}},
/* 6562 */ {(16<<2)|3,{97,105,95}},
/* 6563 */ {(16<<2)|3,{97,105,98}},
/* 6564 */ {(16<<2)|3,{97,105,100}},
/* 6565 */ {(16<<2)|3,{97,105,102}},
/* 6566 */ {(16<<2)|3,{97,105,103}},
/* 6567 */ {(16<<2)|3,{97,105,104}},
/* 6568 */ {(16<<2)|3,{97,105,108}},
/* 6569 */ {(16<<2)|3,{97,105,109}},
/* 6570 */ {(16<<2)|3,{97,105,110}},
/* 6571 */ {(16<<2)|3,{97,105,112}},
/* 6572 */ {(16<<2)|3,{97,105,114}},
/* 6573 */ {(16<<2)|3,{97,105,117}},
/* 6574 */ {(10<<2)|2,{97,105,0}},
/* 6575 */ {(10<<2)|2,{97,105,0}},
/* 6576 */ {(10<<2)|2,{97,105,0}},
/* 6577 */ {(10<<2)|2,{97,105,0}},
/* 6578 */ {(10<<2)|2,{97,105,0}},
/* 6579 */ {(10<<2)|2,{97,105,0}},
/* 6580 */ {(10<<2)|2,{97,105,0}},
/* 6581 */ {(10<<2)|2,{97,105,0}},
/* 6582 */ {(10<<2)|2,{97,105,0}},
/* 6583 */ {(10<<2)|2,{97,105,0}},
/* 6584 */ {(10<<2)|2,{97,105,0}},
/* 6585 */ {(10<<2)|2,{97,105,0}},
/* 6586 */ {(10<<2)|2,{97,105,0}},
/* 6587 */ {(10<<2)|2,{97,105,0}},
/* 6588 */ {(10<<2)|2,{97,105,0}},
/* 6589 */ {(10<<2)|2,{97,105,0}},
/* 6590 */ {(10<<2)|2,{97,105,0}},
/* 6591 */ {(10<<2)|2,{97,105,0}},
/* 6592 */ {(15<<2)|3,{97,111,48}},
/* 6593 */ {(15<<2)|3,{97,111,48}},
/* 6594 */ {(15<<2)|3,{97,111,49}},
/* 6595 */ {(15<<2)|3,{97,111,49}},
/* 6596 */ {(15<<2)|3,{97,111,50}},
/* 6597 */ {(15<<2)|3,{97,111,50}},
/* 6598 */ {(15<<2)|3,{97,111,97}},
/* 6599 */ {(15<<2)|3,{97,111,97}},
/* 6600 */ {(15<<2)|3,{97,111,99}},
/* 6601 */ {(15<<2)|3,{97,111,99}},
/* 6602 */ {(15<<2)|3,{97,111,101}},
/* 6603 */ {(15<<2)|3,{97,111,101}},
/* 6604 */ {(15<<2)|3,{97,111,105}},
/* 6605 */ {(15<<2)|3,{97,111,105}},
/* 6606 */ {(15<<2)|3,{97,111,111}},
/* 6607 */ {(15<<2)|3,{97,111,111}},
/* 6608 */ {(15<<2)|3,{97,111,115}},
/* 6609 */ {(15<<2)|3,{97,111,115}},
/* 6610 */ {(15<<2)|3,{97,111,116}},
/* 6611 */ {(15<<2)|3,{97,111,116}},
/* 6612 */ {(16<<2)|3,{97,111,32}},
/* 6613 */ {(16<<2)|3,{97,111,37}},
/* 6614 */ {(16<<2)|3,{97,111,45}},
/* 6615 */ {(16<<2)|3,{97,111,46}},
/* 6616 */ {(16<<2)|3,{97,111,47}},
/* 6617 */ {(16<<2)|3,{97,111,51}},
/* 6618 */ {(16<<2)|3,{97,111,52}},
/* 6619 */ {(16<<2)|3,{97,111,53}},
/* 6620 */ {(16<<2)|3,{97,111,54}},
/* 6621 */ {(16<<2)|3,{97,111,55}},
/* 6622 */ {(16<<2)|3,{97,111,56}},
/* 6623 */ {(16<<2)|3,{97,111,57}},
/* 6624 */ {(16<<2)|3,{97,111,61}},
/* 6625 */ {(16<<2)|3,{97,111,65}},
/* 6626 */ {(16<<2)|3,{97,111,95}},
/* 6627 */ {(16<<2)|3,{97,111,98}},
/* 6628 */ {(16<<2)|3,{97,111,100}},
/* 6629 */ {(16<<2)|3,{97,111,102}},
/* 6630 */ {(16<<2)|3,{97,111,103}},
/* 6631 */ {(16<<2)|3,{97,111,104}},
/* 6632 */ {(16<<2)|3,{97,111,108}},
/* 6633 */ {(16<<2)|3,{97,111,109}},
/* 6634 */ {(16<<2)|3,{97,111,110}},
/* 6635 */ {(16<<2)|3,{97,111,112}},
/* 6636 */ {(16<<2)|3,{97,111,114}},
/* 6637 */ {(16<<2)|3,{97,111,117}},
/* 6638 */ {(10<<2)|2,{97,111,0}},
/* 6639 */ {(10<<2)|2,{97,111,0}},
/* 6640 */ {(10<<2)|2,{97,111,0}},
/* 6641 */ {(10<<2)|2,{97,111,0}},
/* 6642 */ {(10<<2)|2,{97,111,0}},
/* 6643 */ {(10<<2)|2,{97,111,0}},
/* 6644 */ {(10<<2)|2,{97,111,0}},
/* 6645 */ {(10<<2)|2,{97,111,0}},
/* 6646 */ {(10<<2)|2,{97,111,0}},
/* 6647 */ {(10<<2)|2,{97,111,0}},
/* 6648 */ {(10<<2)|2,{97,111,0}},
/* 6649 */ {(10<<2)|2,{97,111,0}},
/* 6650 */ {(10<<2)|2,{97,111,0}},
/* 6651 */ {(10<<2)|2,{97,111,0}},
/* 6652 */ {(10<<2)|2,{97,111,0}},
/* 6653 */ {(10<<2)|2,{97,111,0}},
/* 6654 */ {(10<<2)|2,{97,111,0}},
/* 6655 */ {(10<<2)|2,{97,111,0}},
/* 6656 */ {(15<<2)|3,{97,115,48}},
/* 6657 */ {(15<<2)|3,{97,115,48}},
/* 6658 */ {(15<<2)|3,{97,115,49}},
/* 6659 */ {(15<<2)|3,{97,115,49}},
/* 6660 */ {(15<<2)|3,{97,115,50}},
/* 6661 */ {(15<<2)|3,{97,115,50}},
/* 6662 */ {(15<<2)|3,{97,115,97}},
/* 6663 */ {(15<<2)|3,{97,115,97}},
/* 6664 */ {(15<<2)|3,{97,115,99}},
/* 6665 */ {(15<<2)|3,{97,115,99}},
/* 6666 */ {(15<<2)|3,{97,115,101}},
/* 6667 */ {(15<<2)|3,{97,115,101}},
/* 6668 */ {(15<<2)|3,{97,115,105}},
/* 6669 */ {(15<<2)|3,{97,115,105}},
/* 6670 */ {(15<<2)|3,{97,115,111}},
/* 6671 */ {(15<<2)|3,{97,115,111}},
/* 6672 */ {(15<<2)|3,{97,115,115}},
/* 6673 */ {(15<<2)|3,{97,115,115}},
/* 6674 */ {(15<<2)|3,{97,115,116}},
/* 6675 */ {(15<<2)|3,{97,115,116}},
/* 6676 */ {(16<<2)|3,{97,115,32}},
/* 6677 */ {(16<<2)|3,{97,115,37}},
/* 6678 */ {(16<<2)|3,{97,115,45}},
/* 6679 */ {(16<<2)|3,{97,115,46}},
/* 6680 */ {(16<<2)|3,{97,115,47}},
/* 6681 */ {(16<<2)|3,{97,115,51}},
/* 6682 */ {(16<<2)|3,{97,115,52}},
/* 6683 */ {(16<<2)|3,{97,115,53}},
/* 6684 */ {(16<<2)|3,{97,115,54}},
/* 6685 */ {(16<<2)|3,{97,115,55}},
/* 6686 */ {(16<<2)|3,{97,115,56}},
/* 6687 */ {(16<<2)|3,{97,115,57}},
/* 6688 */ {(16<<2)|3,{97,115,61}},
/* 6689 */ {(16<<2)|3,{97,115,65}},
/* 6690 */ {(16<<2)|3,{97,115,95}},
/* 6691 */ {(16<<2)|3,{97,115,98}},
/* 6692 */ {(16<<2)|3,{97,115,100}},
/* 6693 */ {(16<<2)|3,{97,115,102}},
/* 6694 */ {(16<<2)|3,{97,115,103}},
/* 6695 */ {(16<<2)|3,{97,115,104}},
/* 6696 */ {(16<<2)|3,{97,115,108}},
/* 6697 */ {(16<<2)|3,{97,115,109}},
/* 6698 */ {(16<<2)|3,{97,115,110}},
/* 6699 */ {(16<<2)|3,{97,115,112}},
/* 6700 */ {(16<<2)|3,{97,115,114}},
/* 6701 */ {(16<<2)|3,{97,115,117}},
/* 6702 */ {(10<<2)|2,{97,115,0}},
/* 6703 */ {(10<<2)|2,{97,115,0}},
/* 6704 */ {(10<<2)|2,{97,115,0}},
/* 6705 */ {(10<<2)|2,{97,115,0}},
/* 6706 */ {(10<<2)|2,{97,115,0}},
/* 6707 */ {(10<<2)|2,{97,115,0}},
/* 6708 */ {(10<<2)|2,{97,115,0}},
/* 6709 */ {(10<<2)|2,{97,115,0}},
/* 6710 */ {(10<<2)|2,{97,115,0}},
/* 6711 */ {(10<<2)|2,{97,115,0}},
/* 6712 */ {(10<<2)|2,{97,115,0}},
/* 6713 */ {(10<<2)|2,{97,115,0}},
/* 6714 */ {(10<<2)|2,{97,115,0}},
/* 6715 */ {(10<<2)|2,{97,115,0}},
/* 6716 */ {(10<<2)|2,{97,115,0}},
/* 6717 */ {(10<<2)|2,{97,115,0}},
/* 6718 */ {(10<<2)|2,{97,115,0}},
/* 6719 */ {(10<<2)|2,{97,115,0}},
/* 6720 */ {(15<<2)|3,{97,116,48}},
/* 6721 */ {(15<<2)|3,{97,116,48}},
/* 6722 */ {(15<<2)|3,{97,116,49}},
/* 6723 */ {(15<<2)|3,{97,116,49}},
/* 6724 */ {(15<<2)|3,{97,116,50}},
/* 6725 */ {(15<<2)|3,{97,116,50}},
/* 6726 */ {(15<<2)|3,{97,116,97}},
/* 6727 */ {(15<<2)|3,{97,116,97}},
/* 6728 */ {(15<<2)|3,{97,116,99}},
/* 6729 */ {(15<<2)|3,{97,116,99}},
/* 6730 */ {(15<<2)|3,{97,116,101}},
/* 6731 */ {(15<<2)|3,{97,116,101}},
/* 6732 */ {(15<<2)|3,{97,116,105}},
/* 6733 */ {(15<<2)|3,{97,116,105}},
/* 6734 */ {(15<<2)|3,{97,116,111}},
/* 6735 */ {(15<<2)|3,{97,116,111}},
/* 6736 */ {(15<<2)|3,{97,116,115}},
/* 6737 */ {(15<<2)|3,{97,116,115}},
/* 6738 */ {(15<<2)|3,{97,116,116}},
/* 6739 */ {(15<<2)|3,{97,116,116}},
/* 6740 */ {(16<<2)|3,{97,116,32}},
/* 6741 */ {(16<<2)|3,{97,116,37}},
/* 6742 */ {(16<<2)|3,{97,116,45}},
/* 6743 */ {(16<<2)|3,{97,116,46}},
/* 6744 */ {(16<<2)|3,{97,116,47}},
/* 6745 */ {(16<<2)|3,{97,116,51}},
/* 6746 */ {(16<<2)|3,{97,116,52}},
/* 6747 */ {(16<<2)|3,{97,116,53}},
/* 6748 */ {(16<<2)|3,{97,116,54}},
/* 6749 */ {(16<<2)|3,{97,116,55}},
/* 6750 */ {(16<<2)|3,{97,116,56}},
/* 6751 */ {(16<<2)|3,{97,116,57}},
/* 6752 */ {(16<<2)|3,{97,116,61}},
/* 6753 */ {(16<<2)|3,{97,116,65}},
/* 6754 */ {(16<<2)|3,{97,116,95}},
/* 6755 */ {(16<<2)|3,{97,116,98}},
/* 6756 */ {(16<<2)|3,{97,116,100}},
/* 6757 */ {(16<<2)|3,{97,116,102}},
/* 6758 */ {(16<<2)|3,{97,116,103}},
/* 6759 */ {(16<<2)|3,{97,116,104}},
/* 6760 */ {(16<<2)|3,{97,116,108}},
/* 6761 */ {(16<<2)|3,{97,116,109}},
/* 6762 */ {(16<<2)|3,{97,116,110}},
/* 6763 */ {(16<<2)|3,{97,116,112}},
/* 6764 */ {(16<<2)|3,{97,116,114}},
/* 6765 */ {(16<<2)|3,{97,116,117}},
/* 6766 */ {(10<<2)|2,{97,116,0}},
/* 6767 */ {(10<<2)|2,{97,116,0}},
/* 6768 */ {(10<<2)|2,{97,116,0}},
/* 6769 */ {(10<<2)|2,{97,116,0}},
/* 6770 */ {(10<<2)|2,{97,116,0}},
/* 6771 */ {(10<<2)|2,{97,116,0}},
/* 6772 */ {(10<<2)|2,{97,116,0}},
/* 6773 */ {(10<<2)|2,{97,116,0}},
/* 6774 */ {(10<<2)|2,{97,116,0}},
/* 6775 */ {(10<<2)|2,{97,116,0}},
/* 6776 */ {(10<<2)|2,{97,116,0}},
/* 6777 */ {(10<<2)|2,{97,116,0}},
/* 6778 */ {(10<<2)|2,{97,116,0}},
/* 6779 */ {(10<<2)|2,{97,116,0}},
/* 6780 */ {(10<<2)|2,{97,116,0}},
/* 6781 */ {(10<<2)|2,{97,116,0}},
/* 6782 */ {(10<<2)|2,{97,116,0}},
/* 6783 */ {(10<<2)|2,{97,116,0}},
/* 6784 */ {(16<<2)|3,{97,32,48}},
/* 6785 */ {(16<<2)|3,{97,32,49}},
/* 6786 */ {(16<<2)|3,{97,32,50}},
/* 6787 */ {(16<<2)|3,{97,32,97}},
/* 6788 */ {(16<<2)|3,{97,32,99}},
/* 6789 */ {(16<<2)|3,{97,32,101}},
/* 6790 */ {(16<<2)|3,{97,32,105}},
/* 6791 */ {(16<<2)|3,{97,32,111}},
/* 6792 */ {(16<<2)|3,{97,32,115}},
/* 6793 */ {(16<<2)|3,{97,32,116}},
/* 6794 */ {(11<<2)|2,{97,32,0}},
/* 6795 */ {(11<<2)|2,{97,32,0}},
/* 6796 */ {(11<<2)|2,{97,32,0}},
/* 6797 */ {(11<<2)|2,{97,32,0}},
/* 6798 */ {(11<<2)|2,{97,32,0}},
/* 6799 */ {(11<<2)|2,{97,32,0}},
/* 6800 */ {(11<<2)|2,{97,32,0}},
/* 6801 */ {(11<<2)|2,{97,32,0}},
/* 6802 */ {(11<<2)|2,{97,32,0}},
/* 6803 */ {(11<<2)|2,{97,32,0}},
/* 6804 */ {(11<<2)|2,{97,32,0}},
/* 6805 */ {(11<<2)|2,{97,32,0}},
/* 6806 */ {(11<<2)|2,{97,32,0}},
/* 6807 */ {(11<<2)|2,{97,32,0}},
/* 6808 */ {(11<<2)|2,{97,32,0}},
/* 6809 */ {(11<<2)|2,{97,32,0}},
/* 6810 */ {(11<<2)|2,{97,32,0}},
/* 6811 */ {(11<<2)|2,{97,32,0}},
/* 6812 */ {(11<<2)|2,{97,32,0}},
/* 6813 */ {(11<<2)|2,{97,32,0}},
/* 6814 */ {(11<<2)|2,{97,32,0}},
/* 6815 */ {(11<<2)|2,{97,32,0}},
/* 6816 */ {(16<<2)|3,{97,37,48}},
/* 6817 */ {(16<<2)|3,{97,37,49}},
/* 6818 */ {(16<<2)|3,{97,37,50}},
/* 6819 */ {(16<<2)|3,{97,37,97}},
/* 6820 */ {(16<<2)|3,{97,37,99}},
/* 6821 */ {(16<<2)|3,{97,37,101}},
/* 6822 */ {(16<<2)|3,{97,37,105}},
/* 6823 */ {(16<<2)|3,{97,37,111}},
/* 6824 */ {(16<<2)|3,{97,37,115}},
/* 6825 */ {(16<<2)|3,{97,37,116}},
/* 6826 */ {(11<<2)|2,{97,37,0}},
/* 6827 */ {(11<<2)|2,{97,37,0}},
/* 6828 */ {(11<<2)|2,{97,37,0}},
/* 6829 */ {(11<<2)|2,{97,37,0}},
/* 6830 */ {(11<<2)|2,{97,37,0}},
/* 6831 */ {(11<<2)|2,{97,37,0}},
/* 6832 */ {(11<<2)|2,{97,37,0}},
/* 6833 */ {(11<<2)|2,{97,37,0}},
/* 6834 */ {(11<<2)|2,{97,37,0}},
/* 6835 */ {(11<<2)|2,{97,37,0}},
/* 6836 */ {(11<<2)|2,{97,37,0}},
/* 6837 */ {(11<<2)|2,{97,37,0}},
/* 6838 */ {(11<<2)|2,{97,37,0}},
/* 6839 */ {(11<<2)|2,{97,37,0}},
/* 6840 */ {(11<<2)|2,{97,37,0}},
/* 6841 */ {(11<<2)|2,{97,37,0}},
/* 6842 */ {(11<<2)|2,{97,37,0}},
/* 6843 */ {(11<<2)|2,{97,37,0}},
/* 6844 */ {(11<<2)|2,{97,37,0}},
/* 6845 */ {(11<<2)|2,{97,37,0}},
/* 6846 */ {(11<<2)|2,{97,37,0}},
/* 6847 */ {(11<<2)|2,{97,37,0}},
/* 6848 */ {(16<<2)|3,{97,45,48}},
/* 6849 */ {(16<<2)|3,{97,45,49}},
/* 6850 */ {(16<<2)|3,{97,45,50}},
/* 6851 */ {(16<<2)|3,{97,45,97}},
/* 6852 */ {(16<<2)|3,{97,45,99}},
/* 6853 */ {(16<<2)|3,{97,45,101}},
/* 6854 */ {(16<<2)|3,{97,45,105}},
/* 6855 */ {(16<<2)|3,{97,45,111}},
/* 6856 */ {(16<<2)|3,{97,45,115}},
/* 6857 */ {(16<<2)|3,{97,45,116}},
/* 6858 */ {(11<<2)|2,{97,45,0}},
/* 6859 */ {(11<<2)|2,{97,45,0}},
/* 6860 */ {(11<<2)|2,{97,45,0}},
/* 6861 */ {(11<<2)|2,{97,45,0}},
/* 6862 */ {(11<<2)|2,{97,45,0}},
/* 6863 */ {(11<<2)|2,{97,45,0}},
/* 6864 */ {(11<<2)|2,{97,45,0}},
/* 6865 */ {(11<<2)|2,{97,45,0}},
/* 6866 */ {(11<<2)|2,{97,45,0}},
/* 6867 */ {(11<<2)|2,{97,45,0}},
/* 6868 */ {(11<<2)|2,{97,45,0}},
/* 6869 */ {(11<<2)|2,{97,45,0}},
/* 6870 */ {(11<<2)|2,{97,45,0}},
/* 6871 */ {(11<<2)|2,{97,45,0}},
/* 6872 */ {(11<<2)|2,{97,45,0}},
/* 6873 */ {(11<<2)|2,{97,45,0}},
/* 6874 */ {(11<<2)|2,{97,45,0}},
/* 6875 */ {(11<<2)|2,{97,45,0}},
/* 6876 */ {(11<<2)|2,{97,45,0}},
/* 6877 */ {(11<<2)|2,{97,45,0}},
/* 6878 */ {(11<<2)|2,{97,45,0}},
/* 6879 */ {(11<<2)|2,{97,45,0}},
/* 6880 */ {(16<<2)|3,{97,46,48}},
/* 6881 */ {(16<<2)|3,{97,46,49}},
/* 6882 */ {(16<<2)|3,{97,46,50}},
/* 6883 */ {(16<<2)|3,{97,46,97}},
/* 6884 */ {(16<<2)|3,{97,46,99}},
/* 6885 */ {(16<<2)|3,{97,46,101}},
/* 6886 */ {(16<<2)|3,{97,46,105}},
/* 6887 */ {(16<<2)|3,{97,46,111}},
/* 6888 */ {(16<<2)|3,{97,46,115}},
/* 6889 */ {(16<<2)|3,{97,46,116}},
/* 6890 */ {(11<<2)|2,{97,46,0}},
/* 6891 */ {(11<<2)|2,{97,46,0}},
/* 6892 */ {(11<<2)|2,{97,46,0}},
/* 6893 */ {(11<<2)|2,{97,46,0}},
/* 6894 */ {(11<<2)|2,{97,46,0}},
/* 6895 */ {(11<<2)|2,{97,46,0}},
/* 6896 */ {(11<<2)|2,{97,46,0}},
/* 6897 */ {(11<<2)|2,{97,46,0}},
/* 6898 */ {(11<<2)|2,{97,46,0}},
/* 6899 */ {(11<<2)|2,{97,46,0}},
/* 6900 */ {(11<<2)|2,{97,46,0}},
/* 6901 */ {(11<<2)|2,{97,46,0}},
/* 6902 */ {(11<<2)|2,{97,46,0}},
/* 6903 */ {(11<<2)|2,{97,46,0}},
/* 6904 */ {(11<<2)|2,{97,46,0}},
/* 6905 */ {(11<<2)|2,{97,46,0}},
/* 6906 */ {(11<<2)|2,{97,46,0}},
/* 6907 */ {(11<<2)|2,{97,46,0}},
/* 6908 */ {(11<<2)|2,{97,46,0}},
/* 6909 */ {(11<<2)|2,{97,46,0}},
/* 6910 */ {(11<<2)|2,{97,46,0}},
/* 6911 */ {(11<<2)|2,{97,46,0}},
/* 6912 */ {(16<<2)|3,{97,47,48}},
/* 6913 */ {(16<<2)|3,{97,47,49}},
/* 6914 */ {(16<<2)|3,{97,47,50}},
/* 6915 */ {(16<<2)|3,{97,47,97}},
/* 6916 */ {(16<<2)|3,{97,47,99}},
/* 6917 */ {(16<<2)|3,{97,47,101}},
/* 6918 */ {(16<<2)|3,{97,47,105}},
/* 6919 */ {(16<<2)|3,{97,47,111}},
/* 6920 */ {(16<<2)|3,{97,47,115}},
/* 6921 */ {(16<<2)|3,{97,47,116}},
/* 6922 */ {(11<<2)|2,{97,47,0}},
/* 6923 */ {(11<<2)|2,{97,47,0}},
/* 6924 */ {(11<<2)|2,{97,47,0}},
/* 6925 */ {(11<<2)|2,{97,47,0}},
/* 6926 */ {(11<<2)|2,{97,47,0}},
/* 6927 */ {(11<<2)|2,{97,47,0}},
/* 6928 */ {(11<<2)|2,{97,47,0}},
/* 6929 */ {(11<<2)|2,{97,47,0}},
/* 6930 */ {(11<<2)|2,{97,47,0}},
/* 6931 */ {(11<<2)|2,{97,47,0}},
/* 6932 */ {(11<<2)|2,{97,47,0}},
/* 6933 */ {(11<<2)|2,{97,47,0}},
/* 6934 */ {(11<<2)|2,{97,47,0}},
/* 6935 */ {(11<<2)|2,{97,47,0}},
/* 6936 */ {(11<<2)|2,{97,47,0}},
/* 6937 */ {(11<<2)|2,{97,47,0}},
/* 6938 */ {(11<<2)|2,{97,47,0}},
/* 6939 */ {(11<<2)|2,{97,47,0}},
/* 6940 */ {(11<<2)|2,{97,47,0}},
/* 6941 */ {(11<<2)|2,{97,47,0}},
/* 6942 */ {(11<<2)|2,{97,47,0}},
/* 6943 */ {(11<<2)|2,{97,47,0}},
/* 6944 */ {(16<<2)|3,{97,51,48}},
/* 6945 */ {(16<<2)|3,{97,51,49}},
/* 6946 */ {(16<<2)|3,{97,51,50}},
/* 6947 */ {(16<<2)|3,{97,51,97}},
/* 6948 */ {(16<<2)|3,{97,51,99}},
/* 6949 */ {(16<<2)|3,{97,51,101}},
/* 6950 */ {(16<<2)|3,{97,51,105}},
/* 6951 */ {(16<<2)|3,{97,51,111}},
/* 6952 */ {(16<<2)|3,{97,51,115}},
/* 6953 */ {(16<<2)|3,{97,51,116}},
/* 6954 */ {(11<<2)|2,{97,51,0}},
/* 6955 */ {(11<<2)|2,{97,51,0}},
/* 6956 */ {(11<<2)|2,{97,51,0}},
/* 6957 */ {(11<<2)|2,{97,51,0}},
/* 6958 */ {(11<<2)|2,{97,51,0}},
/* 6959 */ {(11<<2)|2,{97,51,0}},
/* 6960 */ {(11<<2)|2,{97,51,0}},
/* 6961 */ {(11<<2)|2,{97,51,0}},
/* 6962 */ {(11<<2)|2,{97,51,0}},
/* 6963 */ {(11<<2)|2,{97,51,0}},
/* 6964 */ {(11<<2)|2,{97,51,0}},
/* 6965 */ {(11<<2)|2,{97,51,0}},
/* 6966 */ {(11<<2)|2,{97,51,0}},
/* 6967 */ {(11<<2)|2,{97,51,0}},
/* 6968 */ {(11<<2)|2,{97,51,0}},
/* 6969 */ {(11<<2)|2,{97,51,0}},
/* 6970 */ {(11<<2)|2,{97,51,0}},
/* 6971 */ {(11<<2)|2,{97,51,0}},
/* 6972 */ {(11<<2)|2,{97,51,0}},
/* 6973 */ {(11<<2)|2,{97,51,0}},
/* 6974 */ {(11<<2)|2,{97,51,0}},
/* 6975 */ {(11<<2)|2,{97,51,0}},
/* 6976 */ {(16<<2)|3,{97,52,48}},
/* 6977 */ {(16<<2)|3,{97,52,49}},
/* 6978 */ {(16<<2)|3,{97,52,50}},
/* 6979 */ {(16<<2)|3,{97,52,97}},
/* 6980 */ {(16<<2)|3,{97,52,99}},
/* 6981 */ {(16<<2)|3,{97,52,101}},
/* 6982 */ {(16<<2)|3,{97,52,105}},
/* 6983 */ {(16<<2)|3,{97,52,111}},
/* 6984 */ {(16<<2)|3,{97,52,115}},
/* 6985 */ {(16<<2)|3,{97,52,116}},
/* 6986 */ {(11<<2)|2,{97,52,0}},
/* 6987 */ {(11<<2)|2,{97,52,0}},
/* 6988 */ {(11<<2)|2,{97,52,0}},
/* 6989 */ {(11<<2)|2,{97,52,0}},
/* 6990 */ {(11<<2)|2,{97,52,0}},
/* 6991 */ {(11<<2)|2,{97,52,0}},
/* 6992 */ {(11<<2)|2,{97,52,0}},
/* 6993 */ {(11<<2)|2,{97,52,0}},
/* 6994 */ {(11<<2)|2,{97,52,0}},
/* 6995 */ {(11<<2)|2,{97,52,0}},
/* 6996 */ {(11<<2)|2,{97,52,0}},
/* 6997 */ {(11<<2)|2,{97,52,0}},
/* 6998 */ {(11<<2)|2,{97,52,0}},
/* 6999 */ {(11<<2)|2,{97,52,0}},
/* 7000 */ {(11<<2)|2,{97,52,0}},
/* 7001 */ {(11<<2)|2,{97,52,0}},
/* 7002 */ {(11<<2)|2,{97,52,0}},
/* 7003 */ {(11<<2)|2,{97,52,0}},
/* 7004 */ {(11<<2)|2,{97,52,0}},
/* 7005 */ {(11<<2)|2,{97,52,0}},
/* 7006 */ {(11<<2)|2,{97,52,0}},
/* 7007 */ {(11<<2)|2,{97,52,0}},
/* 7008 */ {(16<<2)|3,{97,53,48}},
/* 7009 */ {(16<<2)|3,{97,53,49}},
/* 7010 */ {(16<<2)|3,{97,53,50}},
/* 7011 */ {(16<<2)|3,{97,53,97}},
/* 7012 */ {(16<<2)|3,{97,53,99}},
/* 7013 */ {(16<<2)|3,{97,53,101}},
/* 7014 */ {(16<<2)|3,{97,53,105}},
/* 7015 */ {(16<<2)|3,{97,53,111}},
/* 7016 */ {(16<<2)|3,{97,53,115}},
/* 7017 */ {(16<<2)|3,{97,53,116}},
/* 7018 */ {(11<<2)|2,{97,53,0}},
/* 7019 */ {(11<<2)|2,{97,53,0}},
/* 7020 */ {(11<<2)|2,{97,53,0}},
/* 7021 */ {(11<<2)|2,{97,53,0}},
/* 7022 */ {(11<<2)|2,{97,53,0}},
/* 7023 */ {(11<<2)|2,{97,53,0}},
/* 7024 */ {(11<<2)|2,{97,53,0}},
/* 7025 */ {(11<<2)|2,{97,53,0}},
/* 7026 */ {(11<<2)|2,{97,53,0}},
/* 7027 */ {(11<<2)|2,{97,53,0}},
/* 7028 */ {(11<<2)|2,{97,53,0}},
/* 7029 */ {(11<<2)|2,{97,53,0}},
/* 7030 */ {(11<<2)|2,{97,53,0}},
/* 7031 */ {(11<<2)|2,{97,53,0}},
/* 7032 */ {(11<<2)|2,{97,53,0}},
/* 7033 */ {(11<<2)|2,{97,53,0}},
/* 7034 */ {(11<<2)|2,{97,53,0}},
/* 7035 */ {(11<<2)|2,{97,53,0}},
/* 7036 */ {(11<<2)|2,{97,53,0}},
/* 7037 */ {(11<<2)|2,{97,53,0}},
/* 7038 */ {(11<<2)|2,{97,53,0}},
/* 7039 */ {(11<<2)|2,{97,53,0}},
/* 7040 */ {(16<<2)|3,{97,54,48}},
/* 7041 */ {(16<<2)|3,{97,54,49}},
/* 7042 */ {(16<<2)|3,{97,54,50}},
/* 7043 */ {(16<<2)|3,{97,54,97}},
/* 7044 */ {(16<<2)|3,{97,54,99}},
/* 7045 */ {(16<<2)|3,{97,54,101}},
/* 7046 */ {(16<<2)|3,{97,54,105}},
/* 7047 */ {(16<<2)|3,{97,54,111}},
/* 7048 */ {(16<<2)|3,{97,54,115}},
/* 7049 */ {(16<<2)|3,{97,54,116}},
/* 7050 */ {(11<<2)|2,{97,54,0}},
/* 7051 */ {(11<<2)|2,{97,54,0}},
/* 7052 */ {(11<<2)|2,{97,54,0}},
/* 7053 */ {(11<<2)|2,{97,54,0}},
/* 7054 */ {(11<<2)|2,{97,54,0}},
/* 7055 */ {(11<<2)|2,{97,54,0}},
/* 7056 */ {(11<<2)|2,{97,54,0}},
/* 7057 */ {(11<<2)|2,{97,54,0}},
/* 7058 */ {(11<<2)|2,{97,54,0}},
/* 7059 */ {(11<<2)|2,{97,54,0}},
/* 7060 */ {(11<<2)|2,{97,54,0}},
/* 7061 */ {(11<<2)|2,{97,54,0}},
/* 7062 */ {(11<<2)|2,{97,54,0}},
/* 7063 */ {(11<<2)|2,{97,54,0}},
/* 7064 */ {(11<<2)|2,{97,54,0}},
/* 7065 */ {(11<<2)|2,{97,54,0}},
/* 7066 */ {(11<<2)|2,{97,54,0}},
/* 7067 */ {(11<<2)|2,{97,54,0}},
/* 7068 */ {(11<<2)|2,{97,54,0}},
/* 7069 */ {(11<<2)|2,{97,54,0}},
/* 7070 */ {(11<<2)|2,{97,54,0}},
/* 7071 */ {(11<<2)|2,{97,54,0}},
/* 7072 */ {(16<<2)|3,{97,55,48}},
/* 7073 */ {(16<<2)|3,{97,55,49}},
/* 7074 */ {(16<<2)|3,{97,55,50}},
/* 7075 */ {(16<<2)|3,{97,55,97}},
/* 7076 */ {(16<<2)|3,{97,55,99}},
/* 7077 */ {(16<<2)|3,{97,55,101}},
/* 7078 */ {(16<<2)|3,{97,55,105}},
/* 7079 */ {(16<<2)|3,{97,55,111}},
/* 7080 */ {(16<<2)|3,{97,55,115}},
/* 7081 */ {(16<<2)|3,{97,55,116}},
/* 7082 */ {(11<<2)|2,{97,55,0}},
/* 7083 */ {(11<<2)|2,{97,55,0}},
/* 7084 */ {(11<<2)|2,{97,55,0}},
/* 7085 */ {(11<<2)|2,{97,55,0}},
/* 7086 */ {(11<<2)|2,{97,55,0}},
/* 7087 */ {(11<<2)|2,{97,55,0}},
/* 7088 */ {(11<<2)|2,{97,55,0}},
/* 7089 */ {(11<<2)|2,{97,55,0}},
/* 7090 */ {(11<<2)|2,{97,55,0}},
/* 7091 */ {(11<<2)|2,{97,55,0}},
/* 7092 */ {(11<<2)|2,{97,55,0}},
/* 7093 */ {(11<<2)|2,{97,55,0}},
/* 7094 */ {(11<<2)|2,{97,55,0}},
/* 7095 */ {(11<<2)|2,{97,55,0}},
/* 7096 */ {(11<<2)|2,{97,55,0}},
/* 7097 */ {(11<<2)|2,{97,55,0}},
/* 7098 */ {(11<<2)|2,{97,55,0}},
/* 7099 */ {(11<<2)|2,{97,55,0}},
/* 7100 */ {(11<<2)|2,{97,55,0}},
/* 7101 */ {(11<<2)|2,{97,55,0}},
/* 7102 */ {(11<<2)|2,{97,55,0}},
/* 7103 */ {(11<<2)|2,{97,55,0}},
/* 7104 */ {(16<<2)|3,{97,56,48}},
/* 7105 */ {(16<<2)|3,{97,56,49}},
/* 7106 */ {(16<<2)|3,{97,56,50}},
/* 7107 */ {(16<<2)|3,{97,56,97}},
/* 7108 */ {(16<<2)|3,{97,56,99}},
/* 7109 */ {(16<<2)|3,{97,56,101}},
/* 7110 */ {(16<<2)|3,{97,56,105}},
/* 7111 */ {(16<<2)|3,{97,56,111}},
/* 7112 */ {(16<<2)|3,{97,56,115}},
/* 7113 */ {(16<<2)|3,{97,56,116}},
/* 7114 */ {(11<<2)|2,{97,56,0}},
/* 7115 */ {(11<<2)|2,{97,56,0}},
/* 7116 */ {(11<<2)|2,{97,56,0}},
/* 7117 */ {(11<<2)|2,{97,56,0}},
/* 7118 */ {(11<<2)|2,{97,56,0}},
/* 7119 */ {(11<<2)|2,{97,56,0}},
/* 7120 */ {(11<<2)|2,{97,56,0}},
/* 7121 */ {(11<<2)|2,{97,56,0}},
/* 7122 */ {(11<<2)|2,{97,56,0}},
/* 7123 */ {(11<<2)|2,{97,56,0}},
/* 7124 */ {(11<<2)|2,{97,56,0}},
/* 7125 */ {(11<<2)|2,{97,56,0}},
/* 7126 */ {(11<<2)|2,{97,56,0}},
/* 7127 */ {(11<<2)|2,{97,56,0}},
/* 7128 */ {(11<<2)|2,{97,56,0}},
/* 7129 */ {(11<<2)|2,{97,56,0}},
/* 7130 */ {(11<<2)|2,{97,56,0}},
/* 7131 */ {(11<<2)|2,{97,56,0}},
/* 7132 */ {(11<<2)|2,{97,56,0}},
/* 7133 */ {(11<<2)|2,{97,56,0}},
/* 7134 */ {(11<<2)|2,{97,56,0}},
/* 7135 */ {(11<<2)|2,{97,56,0}},
/* 7136 */ {(16<<2)|3,{97,57,48}},
/* 7137 */ {(16<<2)|3,{97,57,49}},
/* 7138 */ {(16<<2)|3,{97,57,50}},
/* 7139 */ {(16<<2)|3,{97,57,97}},
/* 7140 */ {(16<<2)|3,{97,57,99}},
/* 7141 */ {(16<<2)|3,{97,57,101}},
/* 7142 */ {(16<<2)|3,{97,57,105}},
/* 7143 */ {(16<<2)|3,{97,57,111}},
/* 7144 */ {(16<<2)|3,{97,57,115}},
/* 7145 */ {(16<<2)|3,{97,57,116}},
/* 7146 */ {(11<<2)|2,{97,57,0}},
/* 7147 */ {(11<<2)|2,{97,57,0}},
/* 7148 */ {(11<<2)|2,{97,57,0}},
/* 7149 */ {(11<<2)|2,{97,57,0}},
/* 7150 */ {(11<<2)|2,{97,57,0}},
/* 7151 */ {(11<<2)|2,{97,57,0}},
/* 7152 */ {(11<<2)|2,{97,57,0}},
/* 7153 */ {(11<<2)|2,{97,57,0}},
/* 7154 */ {(11<<2)|2,{97,57,0}},
/* 7155 */ {(11<<2)|2,{97,57,0}},
/* 7156 */ {(11<<2)|2,{97,57,0}},
/* 7157 */ {(11<<2)|2,{97,57,0}},
/* 7158 */ {(11<<2)|2,{97,57,0}},
/* 7159 */ {(11<<2)|2,{97,57,0}},
/* 7160 */ {(11<<2)|2,{97,57,0}},
/* 7161 */ {(11<<2)|2,{97,57,0}},
/* 7162 */ {(11<<2)|2,{97,57,0}},
/* 7163 */ {(11<<2)|2,{97,57,0}},
/* 7164 */ {(11<<2)|2,{97,57,0}},
/* 7165 */ {(11<<2)|2,{97,57,0}},
/* 7166 */ {(11<<2)|2,{97,57,0}},
/* 7167 */ {(11<<2)|2,{97,57,0}},
/* 7168 */ {(16<<2)|3,{97,61,48}},
/* 7169 */ {(16<<2)|3,{97,61,49}},
/* 7170 */ {(16<<2)|3,{97,61,50}},
/* 7171 */ {(16<<2)|3,{97,61,97}},
/* 7172 */ {(16<<2)|3,{97,61,99}},
/* 7173 */ {(16<<2)|3,{97,61,101}},
/* 7174 */ {(16<<2)|3,{97,61,105}},
/* 7175 */ {(16<<2)|3,{97,61,111}},
/* 7176 */ {(16<<2)|3,{97,61,115}},
/* 7177 */ {(16<<2)|3,{97,61,116}},
/* 7178 */ {(11<<2)|2,{97,61,0}},
/* 7179 */ {(11<<2)|2,{97,61,0}},
/* 7180 */ {(11<<2)|2,{97,61,0}},
/* 7181 */ {(11<<2)|2,{97,61,0}},
/* 7182 */ {(11<<2)|2,{97,61,0}},
/* 7183 */ {(11<<2)|2,{97,61,0}},
/* 7184 */ {(11<<2)|2,{97,61,0}},
/* 7185 */ {(11<<2)|2,{97,61,0}},
/* 7186 */ {(11<<2)|2,{97,61,0}},
/* 7187 */ {(11<<2)|2,{97,61,0}},
/* 7188 */ {(11<<2)|2,{97,61,0}},
/* 7189 */ {(11<<2)|2,{97,61,0}},
/* 7190 */ {(11<<2)|2,{97,61,0}},
/* 7191 */ {(11<<2)|2,{97,61,0}},
/* 7192 */ {(11<<2)|2,{97,61,0}},
/* 7193 */ {(11<<2)|2,{97,61,0}},
/* 7194 */ {(11<<2)|2,{97,61,0}},
/* 7195 */ {(11<<2)|2,{97,61,0}},
/* 7196 */ {(11<<2)|2,{97,61,0}},
/* 7197 */ {(11<<2)|2,{97,61,0}},
/* 7198 */ {(11<<2)|2,{97,61,0}},
/* 7199 */ {(11<<2)|2,{97,61,0}},
/* 7200 */ {(16<<2)|3,{97,65,48}},
/* 7201 */ {(16<<2)|3,{97,65,49}},
/* 7202 */ {(16<<2)|3,{97,65,50}},
/* 7203 */ {(16<<2)|3,{97,65,97}},
/* 7204 */ {(16<<2)|3,{97,65,99}},
/* 7205 */ {(16<<2)|3,{97,65,101}},
/* 7206 */ {(16<<2)|3,{97,65,105}},
/* 7207 */ {(16<<2)|3,{97,65,111}},
/* 7208 */ {(16<<2)|3,{97,65,115}},
/* 7209 */ {(16<<2)|3,{97,65,116}},
/* 7210 */ {(11<<2)|2,{97,65,0}},
/* 7211 */ {(11<<2)|2,{97,65,0}},
/* 7212 */ {(11<<2)|2,{97,65,0}},
/* 7213 */ {(11<<2)|2,{97,65,0}},
/* 7214 */ {(11<<2)|2,{97,65,0}},
/* 7215 */ {(11<<2)|2,{97,65,0}},
/* 7216 */ {(11<<2)|2,{97,65,0}},
/* 7217 */ {(11<<2)|2,{97,65,0}},
/* 7218 */ {(11<<2)|2,{97,65,0}},
/* 7219 */ {(11<<2)|2,{97,65,0}},
/* 7220 */ {(11<<2)|2,{97,65,0}},
/* 7221 */ {(11<<2)|2,{97,65,0}},
/* 7222 */ {(11<<2)|2,{97,65,0}},
/* 7223 */ {(11<<2)|2,{97,65,0}},
/* 7224 */ {(11<<2)|2,{97,65,0}},
/* 7225 */ {(11<<2)|2,{97,65,0}},
/* 7226 */ {(11<<2)|2,{97,65,0}},
/* 7227 */ {(11<<2)|2,{97,65,0}},
/* 7228 */ {(11<<2)|2,{97,65,0}},
/* 7229 */ {(11<<2)|2,{97,65,0}},
/* 7230 */ {(11<<2)|2,{97,65,0}},
/* 7231 */ {(11<<2)|2,{97,65,0}},
/* 7232 */ {(16<<2)|3,{97,95,48}},
/* 7233 */ {(16<<2)|3,{97,95,49}},
/* 7234 */ {(16<<2)|3,{97,95,50}},
/* 7235 */ {(16<<2)|3,{97,95,97}},
/* 7236 */ {(16<<2)|3,{97,95,99}},
/* 7237 */ {(16<<2)|3,{97,95,101}},
/* 7238 */ {(16<<2)|3,{97,95,105}},
/* 7239 */ {(16<<2)|3,{97,95,111}},
/* 7240 */ {(16<<2)|3,{97,95,115}},
/* 7241 */ {(16<<2)|3,{97,95,116}},
/* 7242 */ {(11<<2)|2,{97,95,0}},
/* 7243 */ {(11<<2)|2,{97,95,0}},
/* 7244 */ {(11<<2)|2,{97,95,0}},
/* 7245 */ {(11<<2)|2,{97,95,0}},
/* 7246 */ {(11<<2)|2,{97,95,0}},
/* 7247 */ {(11<<2)|2,{97,95,0}},
/* 7248 */ {(11<<2)|2,{97,95,0}},
/* 7249 */ {(11<<2)|2,{97,95,0}},
/* 7250 */ {(11<<2)|2,{97,95,0}},
/* 7251 */ {(11<<2)|2,{97,95,0}},
/* 7252 */ {(11<<2)|2,{97,95,0}},
/* 7253 */ {(11<<2)|2,{97,95,0}},
/* 7254 */ {(11<<2)|2,{97,95,0}},
/* 7255 */ {(11<<2)|2,{97,95,0}},
/* 7256 */ {(11<<2)|2,{97,95,0}},
/* 7257 */ {(11<<2)|2,{97,95,0}},
/* 7258 */ {(11<<2)|2,{97,95,0}},
/* 7259 */ {(11<<2)|2,{97,95,0}},
/* 7260 */ {(11<<2)|2,{97,95,0}},
/* 7261 */ {(11<<2)|2,{97,95,0}},
/* 7262 */ {(11<<2)|2,{97,95,0}},
/* 7263 */ {(11<<2)|2,{97,95,0}},
/* 7264 */ {(16<<2)|3,{97,98,48}},
/* 7265 */ {(16<<2)|3,{97,98,49}},
/* 7266 */ {(16<<2)|3,{97,98,50}},
/* 7267 */ {(16<<2)|3,{97,98,97}},
/* 7268 */ {(16<<2)|3,{97,98,99}},
/* 7269 */ {(16<<2)|3,{97,98,101}},
/* 7270 */ {(16<<2)|3,{97,98,105}},
/* 7271 */ {(16<<2)|3,{97,98,111}},
/* 7272 */ {(16<<2)|3,{97,98,115}},
/* 7273 */ {(16<<2)|3,{97,98,116}},
/* 7274 */ {(11<<2)|2,{97,98,0}},
/* 7275 */ {(11<<2)|2,{97,98,0}},
/* 7276 */ {(11<<2)|2,{97,98,0}},
/* 7277 */ {(11<<2)|2,{97,98,0}},
/* 7278 */ {(11<<2)|2,{97,98,0}},
/* 7279 */ {(11<<2)|2,{97,98,0}},
/* 7280 */ {(11<<2)|2,{97,98,0}},
/* 7281 */ {(11<<2)|2,{97,98,0}},
/* 7282 */ {(11<<2)|2,{97,98,0}},
/* 7283 */ {(11<<2)|2,{97,98,0}},
/* 7284 */ {(11<<2)|2,{97,98,0}},
/* 7285 */ {(11<<2)|2,{97,98,0}},
/* 7286 */ {(11<<2)|2,{97,98,0}},
/* 7287 */ {(11<<2)|2,{97,98,0}},
/* 7288 */ {(11<<2)|2,{97,98,0}},
/* 7289 */ {(11<<2)|2,{97,98,0}},
/* 7290 */ {(11<<2)|2,{97,98,0}},
/* 7291 */ {(11<<2)|2,{97,98,0}},
/* 7292 */ {(11<<2)|2,{97,98,0}},
/* 7293 */ {(11<<2)|2,{97,98,0}},
/* 7294 */ {(11<<2)|2,{97,98,0}},
/* 7295 */ {(11<<2)|2,{97,98,0}},
/* 7296 */ {(16<<2)|3,{97,100,48}},
/* 7297 */ {(16<<2)|3,{97,100,49}},
/* 7298 */ {(16<<2)|3,{97,100,50}},
/* 7299 */ {(16<<2)|3,{97,100,97}},
/* 7300 */ {(16<<2)|3,{97,100,99}},
/* 7301 */ {(16<<2)|3,{97,100,101}},
/* 7302 */ {(16<<2)|3,{97,100,105}},
/* 7303 */ {(16<<2)|3,{97,100,111}},
/* 7304 */ {(16<<2)|3,{97,100,115}},
/* 7305 */ {(16<<2)|3,{97,100,116}},
/* 7306 */ {(11<<2)|2,{97,100,0}},
/* 7307 */ {(11<<2)|2,{97,100,0}},
/* 7308 */ {(11<<2)|2,{97,100,0}},
/* 7309 */ {(11<<2)|2,{97,100,0}},
/* 7310 */ {(11<<2)|2,{97,100,0}},
/* 7311 */ {(11<<2)|2,{97,100,0}},
/* 7312 */ {(11<<2)|2,{97,100,0}},
/* 7313 */ {(11<<2)|2,{97,100,0}},
/* 7314 */ {(11<<2)|2,{97,100,0}},
/* 7315 */ {(11<<2)|2,{97,100,0}},
/* 7316 */ {(11<<2)|2,{97,100,0}},
/* 7317 */ {(11<<2)|2,{97,100,0}},
/* 7318 */ {(11<<2)|2,{97,100,0}},
/* 7319 */ {(11<<2)|2,{97,100,0}},
/* 7320 */ {(11<<2)|2,{97,100,0}},
/* 7321 */ {(11<<2)|2,{97,100,0}},
/* 7322 */ {(11<<2)|2,{97,100,0}},
/* 7323 */ {(11<<2)|2,{97,100,0}},
/* 7324 */ {(11<<2)|2,{97,100,0}},
/* 7325 */ {(11<<2)|2,{97,100,0}},
/* 7326 */ {(11<<2)|2,{97,100,0}},
/* 7327 */ {(11<<2)|2,{97,100,0}},
/* 7328 */ {(16<<2)|3,{97,102,48}},
/* 7329 */ {(16<<2)|3,{97,102,49}},
/* 7330 */ {(16<<2)|3,{97,102,50}},
/* 7331 */ {(16<<2)|3,{97,102,97}},
/* 7332 */ {(16<<2)|3,{97,102,99}},
/* 7333 */ {(16<<2)|3,{97,102,101}},
/* 7334 */ {(16<<2)|3,{97,102,105}},
/* 7335 */ {(16<<2)|3,{97,102,111}},
/* 7336 */ {(16<<2)|3,{97,102,115}},
/* 7337 */ {(16<<2)|3,{97,102,116}},
/* 7338 */ {(11<<2)|2,{97,102,0}},
/* 7339 */ {(11<<2)|2,{97,102,0}},
/* 7340 */ {(11<<2)|2,{97,102,0}},
/* 7341 */ {(11<<2)|2,{97,102,0}},
/* 7342 */ {(11<<2)|2,{97,102,0}},
/* 7343 */ {(11<<2)|2,{97,102,0}},
/* 7344 */ {(11<<2)|2,{97,102,0}},
/* 7345 */ {(11<<2)|2,{97,102,0}},
/* 7346 */ {(11<<2)|2,{97,102,0}},
/* 7347 */ {(11<<2)|2,{97,102,0}},
/* 7348 */ {(11<<2)|2,{97,102,0}},
/* 7349 */ {(11<<2)|2,{97,102,0}},
/* 7350 */ {(11<<2)|2,{97,102,0}},
/* 7351 */ {(11<<2)|2,{97,102,0}},
/* 7352 */ {(11<<2)|2,{97,102,0}},
/* 7353 */ {(11<<2)|2,{97,102,0}},
/* 7354 */ {(11<<2)|2,{97,102,0}},
/* 7355 */ {(11<<2)|2,{97,102,0}},
/* 7356 */ {(11<<2)|2,{97,102,0}},
/* 7357 */ {(11<<2)|2,{97,102,0}},
/* 7358 */ {(11<<2)|2,{97,102,0}},
/* 7359 */ {(11<<2)|2,{97,102,0}},
/* 7360 */ {(16<<2)|3,{97,103,48}},
/* 7361 */ {(16<<2)|3,{97,103,49}},
/* 7362 */ {(16<<2)|3,{97,103,50}},
/* 7363 */ {(16<<2)|3,{97,103,97}},
/* 7364 */ {(16<<2)|3,{97,103,99}},
/* 7365 */ {(16<<2)|3,{97,103,101}},
/* 7366 */ {(16<<2)|3,{97,103,105}},
/* 7367 */ {(16<<2)|3,{97,103,111}},
/* 7368 */ {(16<<2)|3,{97,103,115}},
/* 7369 */ {(16<<2)|3,{97,103,116}},
/* 7370 */ {(11<<2)|2,{97,103,0}},
/* 7371 */ {(11<<2)|2,{97,103,0}},
/* 7372 */ {(11<<2)|2,{97,103,0}},
/* 7373 */ {(11<<2)|2,{97,103,0}},
/* 7374 */ {(11<<2)|2,{97,103,0}},
/* 7375 */ {(11<<2)|2,{97,103,0}},
/* 7376 */ {(11<<2)|2,{97,103,0}},
/* 7377 */ {(11<<2)|2,{97,103,0}},
/* 7378 */ {(11<<2)|2,{97,103,0}},
/* 7379 */ {(11<<2)|2,{97,103,0}},
/* 7380 */ {(11<<2)|2,{97,103,0}},
/* 7381 */ {(11<<2)|2,{97,103,0}},
/* 7382 */ {(11<<2)|2,{97,103,0}},
/* 7383 */ {(11<<2)|2,{97,103,0}},
/* 7384 */ {(11<<2)|2,{97,103,0}},
/* 7385 */ {(11<<2)|2,{97,103,0}},
/* 7386 */ {(11<<2)|2,{97,103,0}},
/* 7387 */ {(11<<2)|2,{97,103,0}},
/* 7388 */ {(11<<2)|2,{97,103,0}},
/* 7389 */ {(11<<2)|2,{97,103,0}},
/* 7390 */ {(11<<2)|2,{97,103,0}},
/* 7391 */ {(11<<2)|2,{97,103,0}},
/* 7392 */ {(16<<2)|3,{97,104,48}},
/* 7393 */ {(16<<2)|3,{97,104,49}},
/* 7394 */ {(16<<2)|3,{97,104,50}},
/* 7395 */ {(16<<2)|3,{97,104,97}},
/* 7396 */ {(16<<2)|3,{97,104,99}},
/* 7397 */ {(16<<2)|3,{97,104,101}},
/* 7398 */ {(16<<2)|3,{97,104,105}},
/* 7399 */ {(16<<2)|3,{97,104,111}},
/* 7400 */ {(16<<2)|3,{97,104,115}},
/* 7401 */ {(16<<2)|3,{97,104,116}},
/* 7402 */ {(11<<2)|2,{97,104,0}},
/* 7403 */ {(11<<2)|2,{97,104,0}},
/* 7404 */ {(11<<2)|2,{97,104,0}},
/* 7405 */ {(11<<2)|2,{97,104,0}},
/* 7406 */ {(11<<2)|2,{97,104,0}},
/* 7407 */ {(11<<2)|2,{97,104,0}},
/* 7408 */ {(11<<2)|2,{97,104,0}},
/* 7409 */ {(11<<2)|2,{97,104,0}},
/* 7410 */ {(11<<2)|2,{97,104,0}},
/* 7411 */ {(11<<2)|2,{97,104,0}},
/* 7412 */ {(11<<2)|2,{97,104,0}},
/* 7413 */ {(11<<2)|2,{97,104,0}},
/* 7414 */ {(11<<2)|2,{97,104,0}},
/* 7415 */ {(11<<2)|2,{97,104,0}},
/* 7416 */ {(11<<2)|2,{97,104,0}},
/* 7417 */ {(11<<2)|2,{97,104,0}},
/* 7418 */ {(11<<2)|2,{97,104,0}},
/* 7419 */ {(11<<2)|2,{97,104,0}},
/* 7420 */ {(11<<2)|2,{97,104,0}},
/* 7421 */ {(11<<2)|2,{97,104,0}},
/* 7422 */ {(11<<2)|2,{97,104,0}},
/* 7423 */ {(11<<2)|2,{97,104,0}},
/* 7424 */ {(16<<2)|3,{97,108,48}},
/* 7425 */ {(16<<2)|3,{97,108,49}},
/* 7426 */ {(16<<2)|3,{97,108,50}},
/* 7427 */ {(16<<2)|3,{97,108,97}},
/* 7428 */ {(16<<2)|3,{97,108,99}},
/* 7429 */ {(16<<2)|3,{97,108,101}},
/* 7430 */ {(16<<2)|3,{97,108,105}},
/* 7431 */ {(16<<2)|3,{97,108,111}},
/* 7432 */ {(16<<2)|3,{97,108,115}},
/* 7433 */ {(16<<2)|3,{97,108,116}},
/* 7434 */ {(11<<2)|2,{97,108,0}},
/* 7435 */ {(11<<2)|2,{97,108,0}},
/* 7436 */ {(11<<2)|2,{97,108,0}},
/* 7437 */ {(11<<2)|2,{97,108,0}},
/* 7438 */ {(11<<2)|2,{97,108,0}},
/* 7439 */ {(11<<2)|2,{97,108,0}},
/* 7440 */ {(11<<2)|2,{97,108,0}},
/* 7441 */ {(11<<2)|2,{97,108,0}},
/* 7442 */ {(11<<2)|2,{97,108,0}},
/* 7443 */ {(11<<2)|2,{97,108,0}},
/* 7444 */ {(11<<2)|2,{97,108,0}},
/* 7445 */ {(11<<2)|2,{97,108,0}},
/* 7446 */ {(11<<2)|2,{97,108,0}},
/* 7447 */ {(11<<2)|2,{97,108,0}},
/* 7448 */ {(11<<2)|2,{97,108,0}},
/* 7449 */ {(11<<2)|2,{97,108,0}},
/* 7450 */ {(11<<2)|2,{97,108,0}},
/* 7451 */ {(11<<2)|2,{97,108,0}},
/* 7452 */ {(11<<2)|2,{97,108,0}},
/* 7453 */ {(11<<2)|2,{97,108,0}},
/* 7454 */ {(11<<2)|2,{97,108,0}},
/* 7455 */ {(11<<2)|2,{97,108,0}},
/* 7456 */ {(16<<2)|3,{97,109,48}},
/* 7457 */ {(16<<2)|3,{97,109,49}},
/* 7458 */ {(16<<2)|3,{97,109,50}},
/* 7459 */ {(16<<2)|3,{97,109,97}},
/* 7460 */ {(16<<2)|3,{97,109,99}},
/* 7461 */ {(16<<2)|3,{97,109,101}},
/* 7462 */ {(16<<2)|3,{97,109,105}},
/* 7463 */ {(16<<2)|3,{97,109,111}},
/* 7464 */ {(16<<2)|3,{97,109,115}},
/* 7465 */ {(16<<2)|3,{97,109,116}},
/* 7466 */ {(11<<2)|2,{97,109,0}},
/* 7467 */ {(11<<2)|2,{97,109,0}},
/* 7468 */ {(11<<2)|2,{97,109,0}},
/* 7469 */ {(11<<2)|2,{97,109,0}},
/* 7470 */ {(11<<2)|2,{97,109,0}},
/* 7471 */ {(11<<2)|2,{97,109,0}},
/* 7472 */ {(11<<2)|2,{97,109,0}},
/* 7473 */ {(11<<2)|2,{97,109,0}},
/* 7474 */ {(11<<2)|2,{97,109,0}},
/* 7475 */ {(11<<2)|2,{97,109,0}},
/* 7476 */ {(11<<2)|2,{97,109,0}},
/* 7477 */ {(11<<2)|2,{97,109,0}},
/* 7478 */ {(11<<2)|2,{97,109,0}},
/* 7479 */ {(11<<2)|2,{97,109,0}},
/* 7480 */ {(11<<2)|2,{97,109,0}},
/* 7481 */ {(11<<2)|2,{97,109,0}},
/* 7482 */ {(11<<2)|2,{97,109,0}},
/* 7483 */ {(11<<2)|2,{97,109,0}},
/* 7484 */ {(11<<2)|2,{97,109,0}},
/* 7485 */ {(11<<2)|2,{97,109,0}},
/* 7486 */ {(11<<2)|2,{97,109,0}},
/* 7487 */ {(11<<2)|2,{97,109,0}},
/* 7488 */ {(16<<2)|3,{97,110,48}},
/* 7489 */ {(16<<2)|3,{97,110,49}},
/* 7490 */ {(16<<2)|3,{97,110,50}},
/* 7491 */ {(16<<2)|3,{97,110,97}},
/* 7492 */ {(16<<2)|3,{97,110,99}},
/* 7493 */ {(16<<2)|3,{97,110,101}},
/* 7494 */ {(16<<2)|3,{97,110,105}},
/* 7495 */ {(16<<2)|3,{97,110,111}},
/* 7496 */ {(16<<2)|3,{97,110,115}},
/* 7497 */ {(16<<2)|3,{97,110,116}},
/* 7498 */ {(11<<2)|2,{97,110,0}},
/* 7499 */ {(11<<2)|2,{97,110,0}},
/* 7500 */ {(11<<2)|2,{97,110,0}},
/* 7501 */ {(11<<2)|2,{97,110,0}},
/* 7502 */ {(11<<2)|2,{97,110,0}},
/* 7503 */ {(11<<2)|2,{97,110,0}},
/* 7504 */ {(11<<2)|2,{97,110,0}},
/* 7505 */ {(11<<2)|2,{97,110,0}},
/* 7506 */ {(11<<2)|2,{97,110,0}},
/* 7507 */ {(11<<2)|2,{97,110,0}},
/* 7508 */ {(11<<2)|2,{97,110,0}},
/* 7509 */ {(11<<2)|2,{97,110,0}},
/* 7510 */ {(11<<2)|2,{97,110,0}},
/* 7511 */ {(11<<2)|2,{97,110,0}},
/* 7512 */ {(11<<2)|2,{97,110,0}},
/* 7513 */ {(11<<2)|2,{97,110,0}},
/* 7514 */ {(11<<2)|2,{97,110,0}},
/* 7515 */ {(11<<2)|2,{97,110,0}},
/* 7516 */ {(11<<2)|2,{97,110,0}},
/* 7517 */ {(11<<2)|2,{97,110,0}},
/* 7518 */ {(11<<2)|2,{97,110,0}},
/* 7519 */ {(11<<2)|2,{97,110,0}},
/* 7520 */ {(16<<2)|3,{97,112,48}},
/* 7521 */ {(16<<2)|3,{97,112,49}},
/* 7522 */ {(16<<2)|3,{97,112,50}},
/* 7523 */ {(16<<2)|3,{97,112,97}},
/* 7524 */ {(16<<2)|3,{97,112,99}},
/* 7525 */ {(16<<2)|3,{97,112,101}},
/* 7526 */ {(16<<2)|3,{97,112,105}},
/* 7527 */ {(16<<2)|3,{97,112,111}},
/* 7528 */ {(16<<2)|3,{97,112,115}},
/* 7529 */ {(16<<2)|3,{97,112,116}},
/* 7530 */ {(11<<2)|2,{97,112,0}},
/* 7531 */ {(11<<2)|2,{97,112,0}},
/* 7532 */ {(11<<2)|2,{97,112,0}},
/* 7533 */ {(11<<2)|2,{97,112,0}},
/* 7534 */ {(11<<2)|2,{97,112,0}},
/* 7535 */ {(11<<2)|2,{97,112,0}},
/* 7536 */ {(11<<2)|2,{97,112,0}},
/* 7537 */ {(11<<2)|2,{97,112,0}},
/* 7538 */ {(11<<2)|2,{97,112,0}},
/* 7539 */ {(11<<2)|2,{97,112,0}},
/* 7540 */ {(11<<2)|2,{97,112,0}},
/* 7541 */ {(11<<2)|2,{97,112,0}},
/* 7542 */ {(11<<2)|2,{97,112,0}},
/* 7543 */ {(11<<2)|2,{97,112,0}},
/* 7544 */ {(11<<2)|2,{97,112,0}},
/* 7545 */ {(11<<2)|2,{97,112,0}},
/* 7546 */ {(11<<2)|2,{97,112,0}},
/* 7547 */ {(11<<2)|2,{97,112,0}},
/* 7548 */ {(11<<2)|2,{97,112,0}},
/* 7549 */ {(11<<2)|2,{97,112,0}},
/* 7550 */ {(11<<2)|2,{97,112,0}},
/* 7551 */ {(11<<2)|2,{97,112,0}},
/* 7552 */ {(16<<2)|3,{97,114,48}},
/* 7553 */ {(16<<2)|3,{97,114,49}},
/* 7554 */ {(16<<2)|3,{97,114,50}},
/* 7555 */ {(16<<2)|3,{97,114,97}},
/* 7556 */ {(16<<2)|3,{97,114,99}},
/* 7557 */ {(16<<2)|3,{97,114,101}},
/* 7558 */ {(16<<2)|3,{97,114,105}},
/* 7559 */ {(16<<2)|3,{97,114,111}},
/* 7560 */ {(16<<2)|3,{97,114,115}},
/* 7561 */ {(16<<2)|3,{97,114,116}},
/* 7562 */ {(11<<2)|2,{97,114,0}},
/* 7563 */ {(11<<2)|2,{97,114,0}},
/* 7564 */ {(11<<2)|2,{97,114,0}},
/* 7565 */ {(11<<2)|2,{97,114,0}},
/* 7566 */ {(11<<2)|2,{97,114,0}},
/* 7567 */ {(11<<2)|2,{97,114,0}},
/* 7568 */ {(11<<2)|2,{97,114,0}},
/* 7569 */ {(11<<2)|2,{97,114,0}},
/* 7570 */ {(11<<2)|2,{97,114,0}},
/* 7571 */ {(11<<2)|2,{97,114,0}},
/* 7572 */ {(11<<2)|2,{97,114,0}},
/* 7573 */ {(11<<2)|2,{97,114,0}},
/* 7574 */ {(11<<2)|2,{97,114,0}},
/* 7575 */ {(11<<2)|2,{97,114,0}},
/* 7576 */ {(11<<2)|2,{97,114,0}},
/* 7577 */ {(11<<2)|2,{97,114,0}},
/* 7578 */ {(11<<2)|2,{97,114,0}},
/* 7579 */ {(11<<2)|2,{97,114,0}},
/* 7580 */ {(11<<2)|2,{97,114,0}},
/* 7581 */ {(11<<2)|2,{97,114,0}},
/* 7582 */ {(11<<2)|2,{97,114,0}},
/* 7583 */ {(11<<2)|2,{97,114,0}},
/* 7584 */ {(16<<2)|3,{97,117,48}},
/* 7585 */ {(16<<2)|3,{97,117,49}},
/* 7586 */ {(16<<2)|3,{97,117,50}},
/* 7587 */ {(16<<2)|3,{97,117,97}},
/* 7588 */ {(16<<2)|3,{97,117,99}},
/* 7589 */ {(16<<2)|3,{97,117,101}},
/* 7590 */ {(16<<2)|3,{97,117,105}},
/* 7591 */ {(16<<2)|3,{97,117,111}},
/* 7592 */ {(16<<2)|3,{97,117,115}},
/* 7593 */ {(16<<2)|3,{97,117,116}},
/* 7594 */ {(11<<2)|2,{97,117,0}},
/* 7595 */ {(11<<2)|2,{97,117,0}},
/* 7596 */ {(11<<2)|2,{97,117,0}},
/* 7597 */ {(11<<2)|2,{97,117,0}},
/* 7598 */ {(11<<2)|2,{97,117,0}},
/* 7599 */ {(11<<2)|2,{97,117,0}},
/* 7600 */ {(11<<2)|2,{97,117,0}},
/* 7601 */ {(11<<2)|2,{97,117,0}},
/* 7602 */ {(11<<2)|2,{97,117,0}},
/* 7603 */ {(11<<2)|2,{97,117,0}},
/* 7604 */ {(11<<2)|2,{97,117,0}},
/* 7605 */ {(11<<2)|2,{97,117,0}},
/* 7606 */ {(11<<2)|2,{97,117,0}},
/* 7607 */ {(11<<2)|2,{97,117,0}},
/* 7608 */ {(11<<2)|2,{97,117,0}},
/* 7609 */ {(11<<2)|2,{97,117,0}},
/* 7610 */ {(11<<2)|2,{97,117,0}},
/* 7611 */ {(11<<2)|2,{97,117,0}},
/* 7612 */ {(11<<2)|2,{97,117,0}},
/* 7613 */ {(11<<2)|2,{97,117,0}},
/* 7614 */ {(11<<2)|2,{97,117,0}},
/* 7615 */ {(11<<2)|2,{97,117,0}},
/* 7616 */ {(12<<2)|2,{97,58,0}},
/* 7617 */ {(12<<2)|2,{97,58,0}},
/* 7618 */ {(12<<2)|2,{97,58,0}},
/* 7619 */ {(12<<2)|2,{97,58,0}},
/* 7620 */ {(12<<2)|2,{97,58,0}},
/* 7621 */ {(12<<2)|2,{97,58,0}},
/* 7622 */ {(12<<2)|2,{97,58,0}},
/* 7623 */ {(12<<2)|2,{97,58,0}},
/* 7624 */ {(12<<2)|2,{97,58,0}},
/* 7625 */ {(12<<2)|2,{97,58,0}},
/* 7626 */ {(12<<2)|2,{97,58,0}},
/* 7627 */ {(12<<2)|2,{97,58,0}},
/* 7628 */ {(12<<2)|2,{97,58,0}},
/* 7629 */ {(12<<2)|2,{97,58,0}},
/* 7630 */ {(12<<2)|2,{97,58,0}},
/* 7631 */ {(12<<2)|2,{97,58,0}},
/* 7632 */ {(12<<2)|2,{97,66,0}},
/* 7633 */ {(12<<2)|2,{97,66,0}},
/* 7634 */ {(12<<2)|2,{97,66,0}},
/* 7635 */ {(12<<2)|2,{97,66,0}},
/* 7636 */ {(12<<2)|2,{97,66,0}},
/* 7637 */ {(12<<2)|2,{97,66,0}},
/* 7638 */ {(12<<2)|2,{97,66,0}},
/* 7639 */ {(12<<2)|2,{97,66,0}},
/* 7640 */ {(12<<2)|2,{97,66,0}},
/* 7641 */ {(12<<2)|2,{97,66,0}},
/* 7642 */ {(12<<2)|2,{97,66,0}},
/* 7643 */ {(12<<2)|2,{97,66,0}},
/* 7644 */ {(12<<2)|2,{97,66,0}},
/* 7645 */ {(12<<2)|2,{97,66,0}},
/* 7646 */ {(12<<2)|2,{97,66,0}},
/* 7647 */ {(12<<2)|2,{97,66,0}},
/* 7648 */ {(12<<2)|2,{97,67,0}},
/* 7649 */ {(12<<2)|2,{97,67,0}},
/* 7650 */ {(12<<2)|2,{97,67,0}},
/* 7651 */ {(12<<2)|2,{97,67,0}},
/* 7652 */ {(12<<2)|2,{97,67,0}},
/* 7653 */ {(12<<2)|2,{97,67,0}},
/* 7654 */ {(12<<2)|2,{97,67,0}},
/* 7655 */ {(12<<2)|2,{97,67,0}},
/* 7656 */ {(12<<2)|2,{97,67,0}},
/* 7657 */ {(12<<2)|2,{97,67,0}},
/* 7658 */ {(12<<2)|2,{97,67,0}},
/* 7659 */ {(12<<2)|2,{97,67,0}},
/* 7660 */ {(12<<2)|2,{97,67,0}},
/* 7661 */ {(12<<2)|2,{97,67,0}},
/* 7662 */ {(12<<2)|2,{97,67,0}},
/* 7663 */ {(12<<2)|2,{97,67,0}},
/* 7664 */ {(12<<2)|2,{97,68,0}},
/* 7665 */ {(12<<2)|2,{97,68,0}},
/* 7666 */ {(12<<2)|2,{97,68,0}},
/* 7667 */ {(12<<2)|2,{97,68,0}},
/* 7668 */ {(12<<2)|2,{97,68,0}},
/* 7669 */ {(12<<2)|2,{97,68,0}},
/* 7670 */ {(12<<2)|2,{97,68,0}},
/* 7671 */ {(12<<2)|2,{97,68,0}},
/* 7672 */ {(12<<2)|2,{97,68,0}},
/* 7673 */ {(12<<2)|2,{97,68,0}},
/* 7674 */ {(12<<2)|2,{97,68,0}},
/* 7675 */ {(12<<2)|2,{97,68,0}},
/* 7676 */ {(12<<2)|2,{97,68,0}},
/* 7677 */ {(12<<2)|2,{97,68,0}},
/* 7678 */ {(12<<2)|2,{97,68,0}},
/* 7679 */ {(12<<2)|2,{97,68,0}},
/* 7680 */ {(12<<2)|2,{97,69,0}},
/* 7681 */ {(12<<2)|2,{97,69,0}},
/* 7682 */ {(12<<2)|2,{97,69,0}},
/* 7683 */ {(12<<2)|2,{97,69,0}},
/* 7684 */ {(12<<2)|2,{97,69,0}},
/* 7685 */ {(12<<2)|2,{97,69,0}},
/* 7686 */ {(12<<2)|2,{97,69,0}},
/* 7687 */ {(12<<2)|2,{97,69,0}},
/* 7688 */ {(12<<2)|2,{97,69,0}},
/* 7689 */ {(12<<2)|2,{97,69,0}},
/* 7690 */ {(12<<2)|2,{97,69,0}},
/* 7691 */ {(12<<2)|2,{97,69,0}},
/* 7692 */ {(12<<2)|2,{97,69,0}},
/* 7693 */ {(12<<2)|2,{97,69,0}},
/* 7694 */ {(12<<2)|2,{97,69,0}},
/* 7695 */ {(12<<2)|2,{97,69,0}},
/* 7696 */ {(12<<2)|2,{97,70,0}},
/* 7697 */ {(12<<2)|2,{97,70,0}},
/* 7698 */ {(12<<2)|2,{97,70,0}},
/* 7699 */ {(12<<2)|2,{97,70,0}},
/* 7700 */ {(12<<2)|2,{97,70,0}},
/* 7701 */ {(12<<2)|2,{97,70,0}},
/* 7702 */ {(12<<2)|2,{97,70,0}},
/* 7703 */ {(12<<2)|2,{97,70,0}},
/* 7704 */ {(12<<2)|2,{97,70,0}},
/* 7705 */ {(12<<2)|2,{97,70,0}},
/* 7706 */ {(12<<2)|2,{97,70,0}},
/* 7707 */ {(12<<2)|2,{97,70,0}},
/* 7708 */ {(12<<2)|2,{97,70,0}},
/* 7709 */ {(12<<2)|2,{97,70,0}},
/* 7710 */ {(12<<2)|2,{97,70,0}},
/* 7711 */ {(12<<2)|2,{97,70,0}},
/* 7712 */ {(12<<2)|2,{97,71,0}},
/* 7713 */ {(12<<2)|2,{97,71,0}},
/* 7714 */ {(12<<2)|2,{97,71,0}},
/* 7715 */ {(12<<2)|2,{97,71,0}},
/* 7716 */ {(12<<2)|2,{97,71,0}},
/* 7717 */ {(12<<2)|2,{97,71,0}},
/* 7718 */ {(12<<2)|2,{97,71,0}},
/* 7719 */ {(12<<2)|2,{97,71,0}},
/* 7720 */ {(12<<2)|2,{97,71,0}},
/* 7721 */ {(12<<2)|2,{97,71,0}},
/* 7722 */ {(12<<2)|2,{97,71,0}},
/* 7723 */ {(12<<2)|2,{97,71,0}},
/* 7724 */ {(12<<2)|2,{97,71,0}},
/* 7725 */ {(12<<2)|2,{97,71,0}},
/* 7726 */ {(12<<2)|2,{97,71,0}},
/* 7727 */ {(12<<2)|2,{97,71,0}},
/* 7728 */ {(12<<2)|2,{97,72,0}},
/* 7729 */ {(12<<2)|2,{97,72,0}},
/* 7730 */ {(12<<2)|2,{97,72,0}},
/* 7731 */ {(12<<2)|2,{97,72,0}},
/* 7732 */ {(12<<2)|2,{97,72,0}},
/* 7733 */ {(12<<2)|2,{97,72,0}},
/* 7734 */ {(12<<2)|2,{97,72,0}},
/* 7735 */ {(12<<2)|2,{97,72,0}},
/* 7736 */ {(12<<2)|2,{97,72,0}},
/* 7737 */ {(12<<2)|2,{97,72,0}},
/* 7738 */ {(12<<2)|2,{97,72,0}},
/* 7739 */ {(12<<2)|2,{97,72,0}},
/* 7740 */ {(12<<2)|2,{97,72,0}},
/* 7741 */ {(12<<2)|2,{97,72,0}},
/* 7742 */ {(12<<2)|2,{97,72,0}},
/* 7743 */ {(12<<2)|2,{97,72,0}},
/* 7744 */ {(12<<2)|2,{97,73,0}},
/* 7745 */ {(12<<2)|2,{97,73,0}},
/* 7746 */ {(12<<2)|2,{97,73,0}},
/* 7747 */ {(12<<2)|2,{97,73,0}},
/* 7748 */ {(12<<2)|2,{97,73,0}},
/* 7749 */ {(12<<2)|2,{97,73,0}},
/* 7750 */ {(12<<2)|2,{97,73,0}},
/* 7751 */ {(12<<2)|2,{97,73,0}},
/* 7752 */ {(12<<2)|2,{97,73,0}},
/* 7753 */ {(12<<2)|2,{97,73,0}},
/* 7754 */ {(12<<2)|2,{97,73,0}},
/* 7755 */ {(12<<2)|2,{97,73,0}},
/* 7756 */ {(12<<2)|2,{97,73,0}},
/* 7757 */ {(12<<2)|2,{97,73,0}},
/* 7758 */ {(12<<2)|2,{97,73,0}},
/* 7759 */ {(12<<2)|2,{97,73,0}},
/* 7760 */ {(12<<2)|2,{97,74,0}},
/* 7761 */ {(12<<2)|2,{97,74,0}},
/* 7762 */ {(12<<2)|2,{97,74,0}},
/* 7763 */ {(12<<2)|2,{97,74,0}},
/* 7764 */ {(12<<2)|2,{97,74,0}},
/* 7765 */ {(12<<2)|2,{97,74,0}},
/* 7766 */ {(12<<2)|2,{97,74,0}},
/* 7767 */ {(12<<2)|2,{97,74,0}},
/* 7768 */ {(12<<2)|2,{97,74,0}},
/* 7769 */ {(12<<2)|2,{97,74,0}},
/* 7770 */ {(12<<2)|2,{97,74,0}},
/* 7771 */ {(12<<2)|2,{97,74,0}},
/* 7772 */ {(12<<2)|2,{97,74,0}},
/* 7773 */ {(12<<2)|2,{97,74,0}},
/* 7774 */ {(12<<2)|2,{97,74,0}},
/* 7775 */ {(12<<2)|2,{97,74,0}},
/* 7776 */ {(12<<2)|2,{97,75,0}},
/* 7777 */ {(12<<2)|2,{97,75,0}},
/* 7778 */ {(12<<2)|2,{97,75,0}},
/* 7779 */ {(12<<2)|2,{97,75,0}},
/* 7780 */ {(12<<2)|2,{97,75,0}},
/* 7781 */ {(12<<2)|2,{97,75,0}},
/* 7782 */ {(12<<2)|2,{97,75,0}},
/* 7783 */ {(12<<2)|2,{97,75,0}},
/* 7784 */ {(12<<2)|2,{97,75,0}},
/* 7785 */ {(12<<2)|2,{97,75,0}},
/* 7786 */ {(12<<2)|2,{97,75,0}},
/* 7787 */ {(12<<2)|2,{97,75,0}},
/* 7788 */ {(12<<2)|2,{97,75,0}},
/* 7789 */ {(12<<2)|2,{97,75,0}},
/* 7790 */ {(12<<2)|2,{97,75,0}},
/* 7791 */ {(12<<2)|2,{97,75,0}},
/* 7792 */ {(12<<2)|2,{97,76,0}},
/* 7793 */ {(12<<2)|2,{97,76,0}},
/* 7794 */ {(12<<2)|2,{97,76,0}},
/* 7795 */ {(12<<2)|2,{97,76,0}},
/* 7796 */ {(12<<2)|2,{97,76,0}},
/* 7797 */ {(12<<2)|2,{97,76,0}},
/* 7798 */ {(12<<2)|2,{97,76,0}},
/* 7799 */ {(12<<2)|2,{97,76,0}},
/* 7800 */ {(12<<2)|2,{97,76,0}},
/* 7801 */ {(12<<2)|2,{97,76,0}},
/* 7802 */ {(12<<2)|2,{97,76,0}},
/* 7803 */ {(12<<2)|2,{97,76,0}},
/* 7804 */ {(12<<2)|2,{97,76,0}},
/* 7805 */ {(12<<2)|2,{97,76,0}},
/* 7806 */ {(12<<2)|2,{97,76,0}},
/* 7807 */ {(12<<2)|2,{97,76,0}},
/* 7808 */ {(12<<2)|2,{97,77,0}},
/* 7809 */ {(12<<2)|2,{97,77,0}},
/* 7810 */ {(12<<2)|2,{97,77,0}},
/* 7811 */ {(12<<2)|2,{97,77,0}},
/* 7812 */ {(12<<2)|2,{97,77,0}},
/* 7813 */ {(12<<2)|2,{97,77,0}},
/* 7814 */ {(12<<2)|2,{97,77,0}},
/* 7815 */ {(12<<2)|2,{97,77,0}},
/* 7816 */ {(12<<2)|2,{97,77,0}},
/* 7817 */ {(12<<2)|2,{97,77,0}},
/* 7818 */ {(12<<2)|2,{97,77,0}},
/* 7819 */ {(12<<2)|2,{97,77,0}},
/* 7820 */ {(12<<2)|2,{97,77,0}},
/* 7821 */ {(12<<2)|2,{97,77,0}},
/* 7822 */ {(12<<2)|2,{97,77,0}},
/* 7823 */ {(12<<2)|2,{97,77,0}},
/* 7824 */ {(12<<2)|2,{97,78,0}},
/* 7825 */ {(12<<2)|2,{97,78,0}},
/* 7826 */ {(12<<2)|2,{97,78,0}},
/* 7827 */ {(12<<2)|2,{97,78,0}},
/* 7828 */ {(12<<2)|2,{97,78,0}},
/* 7829 */ {(12<<2)|2,{97,78,0}},
/* 7830 */ {(12<<2)|2,{97,78,0}},
/* 7831 */ {(12<<2)|2,{97,78,0}},
/* 7832 */ {(12<<2)|2,{97,78,0}},
/* 7833 */ {(12<<2)|2,{97,78,0}},
/* 7834 */ {(12<<2)|2,{97,78,0}},
/* 7835 */ {(12<<2)|2,{97,78,0}},
/* 7836 */ {(12<<2)|2,{97,78,0}},
/* 7837 */ {(12<<2)|2,{97,78,0}},
/* 7838 */ {(12<<2)|2,{97,78,0}},
/* 7839 */ {(12<<2)|2,{97,78,0}},
/* 7840 */ {(12<<2)|2,{97,79,0}},
/* 7841 */ {(12<<2)|2,{97,79,0}},
/* 7842 */ {(12<<2)|2,{97,79,0}},
/* 7843 */ {(12<<2)|2,{97,79,0}},
/* 7844 */ {(12<<2)|2,{97,79,0}},
/* 7845 */ {(12<<2)|2,{97,79,0}},
/* 7846 */ {(12<<2)|2,{97,79,0}},
/* 7847 */ {(12<<2)|2,{97,79,0}},
/* 7848 */ {(12<<2)|2,{97,79,0}},
/* 7849 */ {(12<<2)|2,{97,79,0}},
/* 7850 */ {(12<<2)|2,{97,79,0}},
/* 7851 */ {(12<<2)|2,{97,79,0}},
/* 7852 */ {(12<<2)|2,{97,79,0}},
/* 7853 */ {(12<<2)|2,{97,79,0}},
/* 7854 */ {(12<<2)|2,{97,79,0}},
/* 7855 */ {(12<<2)|2,{97,79,0}},
/* 7856 */ {(12<<2)|2,{97,80,0}},
/* 7857 */ {(12<<2)|2,{97,80,0}},
/* 7858 */ {(12<<2)|2,{97,80,0}},
/* 7859 */ {(12<<2)|2,{97,80,0}},
/* 7860 */ {(12<<2)|2,{97,80,0}},
/* 7861 */ {(12<<2)|2,{97,80,0}},
/* 7862 */ {(12<<2)|2,{97,80,0}},
/* 7863 */ {(12<<2)|2,{97,80,0}},
/* 7864 */ {(12<<2)|2,{97,80,0}},
/* 7865 */ {(12<<2)|2,{97,80,0}},
/* 7866 */ {(12<<2)|2,{97,80,0}},
/* 7867 */ {(12<<2)|2,{97,80,0}},
/* 7868 */ {(12<<2)|2,{97,80,0}},
/* 7869 */ {(12<<2)|2,{97,80,0}},
/* 7870 */ {(12<<2)|2,{97,80,0}},
/* 7871 */ {(12<<2)|2,{97,80,0}},
/* 7872 */ {(12<<2)|2,{97,81,0}},
/* 7873 */ {(12<<2)|2,{97,81,0}},
/* 7874 */ {(12<<2)|2,{97,81,0}},
/* 7875 */ {(12<<2)|2,{97,81,0}},
/* 7876 */ {(12<<2)|2,{97,81,0}},
/* 7877 */ {(12<<2)|2,{97,81,0}},
/* 7878 */ {(12<<2)|2,{97,81,0}},
/* 7879 */ {(12<<2)|2,{97,81,0}},
/* 7880 */ {(12<<2)|2,{97,81,0}},
/* 7881 */ {(12<<2)|2,{97,81,0}},
/* 7882 */ {(12<<2)|2,{97,81,0}},
/* 7883 */ {(12<<2)|2,{97,81,0}},
/* 7884 */ {(12<<2)|2,{97,81,0}},
/* 7885 */ {(12<<2)|2,{97,81,0}},
/* 7886 */ {(12<<2)|2,{97,81,0}},
/* 7887 */ {(12<<2)|2,{97,81,0}},
/* 7888 */ {(12<<2)|2,{97,82,0}},
/* 7889 */ {(12<<2)|2,{97,82,0}},
/* 7890 */ {(12<<2)|2,{97,82,0}},
/* 7891 */ {(12<<2)|2,{97,82,0}},
/* 7892 */ {(12<<2)|2,{97,82,0}},
/* 7893 */ {(12<<2)|2,{97,82,0}},
/* 7894 */ {(12<<2)|2,{97,82,0}},
/* 7895 */ {(12<<2)|2,{97,82,0}},
/* 7896 */ {(12<<2)|2,{97,82,0}},
/* 7897 */ {(12<<2)|2,{97,82,0}},
/* 7898 */ {(12<<2)|2,{97,82,0}},
/* 7899 */ {(12<<2)|2,{97,82,0}},
/* 7900 */ {(12<<2)|2,{97,82,0}},
/* 7901 */ {(12<<2)|2,{97,82,0}},
/* 7902 */ {(12<<2)|2,{97,82,0}},
/* 7903 */ {(12<<2)|2,{97,82,0}},
/* 7904 */ {(12<<2)|2,{97,83,0}},
/* 7905 */ {(12<<2)|2,{97,83,0}},
/* 7906 */ {(12<<2)|2,{97,83,0}},
/* 7907 */ {(12<<2)|2,{97,83,0}},
/* 7908 */ {(12<<2)|2,{97,83,0}},
/* 7909 */ {(12<<2)|2,{97,83,0}},
/* 7910 */ {(12<<2)|2,{97,83,0}},
/* 7911 */ {(12<<2)|2,{97,83,0}},
/* 7912 */ {(12<<2)|2,{97,83,0}},
/* 7913 */ {(12<<2)|2,{97,83,0}},
/* 7914 */ {(12<<2)|2,{97,83,0}},
/* 7915 */ {(12<<2)|2,{97,83,0}},
/* 7916 */ {(12<<2)|2,{97,83,0}},
/* 7917 */ {(12<<2)|2,{97,83,0}},
/* 7918 */ {(12<<2)|2,{97,83,0}},
/* 7919 */ {(12<<2)|2,{97,83,0}},
/* 7920 */ {(12<<2)|2,{97,84,0}},
/* 7921 */ {(12<<2)|2,{97,84,0}},
/* 7922 */ {(12<<2)|2,{97,84,0}},
/* 7923 */ {(12<<2)|2,{97,84,0}},
/* 7924 */ {(12<<2)|2,{97,84,0}},
/* 7925 */ {(12<<2)|2,{97,84,0}},
/* 7926 */ {(12<<2)|2,{97,84,0}},
/* 7927 */ {(12<<2)|2,{97,84,0}},
/* 7928 */ {(12<<2)|2,{97,84,0}},
/* 7929 */ {(12<<2)|2,{97,84,0}},
/* 7930 */ {(12<<2)|2,{97,84,0}},
/* 7931 */ {(12<<2)|2,{97,84,0}},
/* 7932 */ {(12<<2)|2,{97,84,0}},
/* 7933 */ {(12<<2)|2,{97,84,0}},
/* 7934 */ {(12<<2)|2,{97,84,0}},
/* 7935 */ {(12<<2)|2,{97,84,0}},
/* 7936 */ {(12<<2)|2,{97,85,0}},
/* 7937 */ {(12<<2)|2,{97,85,0}},
/* 7938 */ {(12<<2)|2,{97,85,0}},
/* 7939 */ {(12<<2)|2,{97,85,0}},
/* 7940 */ {(12<<2)|2,{97,85,0}},
/* 7941 */ {(12<<2)|2,{97,85,0}},
/* 7942 */ {(12<<2)|2,{97,85,0}},
/* 7943 */ {(12<<2)|2,{97,85,0}},
/* 7944 */ {(12<<2)|2,{97,85,0}},
/* 7945 */ {(12<<2)|2,{97,85,0}},
/* 7946 */ {(12<<2)|2,{97,85,0}},
/* 7947 */ {(12<<2)|2,{97,85,0}},
/* 7948 */ {(12<<2)|2,{97,85,0}},
/* 7949 */ {(12<<2)|2,{97,85,0}},
/* 7950 */ {(12<<2)|2,{97,85,0}},
/* 7951 */ {(12<<2)|2,{97,85,0}},
/* 7952 */ {(12<<2)|2,{97,86,0}},
/* 7953 */ {(12<<2)|2,{97,86,0}},
/* 7954 */ {(12<<2)|2,{97,86,0}},
/* 7955 */ {(12<<2)|2,{97,86,0}},
/* 7956 */ {(12<<2)|2,{97,86,0}},
/* 7957 */ {(12<<2)|2,{97,86,0}},
/* 7958 */ {(12<<2)|2,{97,86,0}},
/* 7959 */ {(12<<2)|2,{97,86,0}},
/* 7960 */ {(12<<2)|2,{97,86,0}},
/* 7961 */ {(12<<2)|2,{97,86,0}},
/* 7962 */ {(12<<2)|2,{97,86,0}},
/* 7963 */ {(12<<2)|2,{97,86,0}},
/* 7964 */ {(12<<2)|2,{97,86,0}},
/* 7965 */ {(12<<2)|2,{97,86,0}},
/* 7966 */ {(12<<2)|2,{97,86,0}},
/* 7967 */ {(12<<2)|2,{97,86,0}},
/* 7968 */ {(12<<2)|2,{97,87,0}},
/* 7969 */ {(12<<2)|2,{97,87,0}},
/* 7970 */ {(12<<2)|2,{97,87,0}},
/* 7971 */ {(12<<2)|2,{97,87,0}},
/* 7972 */ {(12<<2)|2,{97,87,0}},
/* 7973 */ {(12<<2)|2,{97,87,0}},
/* 7974 */ {(12<<2)|2,{97,87,0}},
/* 7975 */ {(12<<2)|2,{97,87,0}},
/* 7976 */ {(12<<2)|2,{97,87,0}},
/* 7977 */ {(12<<2)|2,{97,87,0}},
/* 7978 */ {(12<<2)|2,{97,87,0}},
/* 7979 */ {(12<<2)|2,{97,87,0}},
/* 7980 */ {(12<<2)|2,{97,87,0}},
/* 7981 */ {(12<<2)|2,{97,87,0}},
/* 7982 */ {(12<<2)|2,{97,87,0}},
/* 7983 */ {(12<<2)|2,{97,87,0}},
/* 7984 */ {(12<<2)|2,{97,89,0}},
/* 7985 */ {(12<<2)|2,{97,89,0}},
/* 7986 */ {(12<<2)|2,{97,89,0}},
/* 7987 */ {(12<<2)|2,{97,89,0}},
/* 7988 */ {(12<<2)|2,{97,89,0}},
/* 7989 */ {(12<<2)|2,{97,89,0}},
/* 7990 */ {(12<<2)|2,{97,89,0}},
/* 7991 */ {(12<<2)|2,{97,89,0}},
/* 7992 */ {(12<<2)|2,{97,89,0}},
/* 7993 */ {(12<<2)|2,{97,89,0}},
/* 7994 */ {(12<<2)|2,{97,89,0}},
/* 7995 */ {(12<<2)|2,{97,89,0}},
/* 7996 */ {(12<<2)|2,{97,89,0}},
/* 7997 */ {(12<<2)|2,{97,89,0}},
/* 7998 */ {(12<<2)|2,{97,89,0}},
/* 7999 */ {(12<<2)|2,{97,89,0}},
/* 8000 */ {(12<<2)|2,{97,106,0}},
/* 8001 */ {(12<<2)|2,{97,106,0}},
/* 8002 */ {(12<<2)|2,{97,106,0}},
/* 8003 */ {(12<<2)|2,{97,106,0}},
/* 8004 */ {(12<<2)|2,{97,106,0}},
/* 8005 */ {(12<<2)|2,{97,106,0}},
/* 8006 */ {(12<<2)|2,{97,106,0}},
/* 8007 */ {(12<<2)|2,{97,106,0}},
/* 8008 */ {(12<<2)|2,{97,106,0}},
/* 8009 */ {(12<<2)|2,{97,106,0}},
/* 8010 */ {(12<<2)|2,{97,106,0}},
/* 8011 */ {(12<<2)|2,{97,106,0}},
/* 8012 */ {(12<<2)|2,{97,106,0}},
/* 8013 */ {(12<<2)|2,{97,106,0}},
/* 8014 */ {(12<<2)|2,{97,106,0}},
/* 8015 */ {(12<<2)|2,{97,106,0}},
/* 8016 */ {(12<<2)|2,{97,107,0}},
/* 8017 */ {(12<<2)|2,{97,107,0}},
/* 8018 */ {(12<<2)|2,{97,107,0}},
/* 8019 */ {(12<<2)|2,{97,107,0}},
/* 8020 */ {(12<<2)|2,{97,107,0}},
/* 8021 */ {(12<<2)|2,{97,107,0}},
/* 8022 */ {(12<<2)|2,{97,107,0}},
/* 8023 */ {(12<<2)|2,{97,107,0}},
/* 8024 */ {(12<<2)|2,{97,107,0}},
/* 8025 */ {(12<<2)|2,{97,107,0}},
/* 8026 */ {(12<<2)|2,{97,107,0}},
/* 8027 */ {(12<<2)|2,{97,107,0}},
/* 8028 */ {(12<<2)|2,{97,107,0}},
/* 8029 */ {(12<<2)|2,{97,107,0}},
/* 8030 */ {(12<<2)|2,{97,107,0}},
/* 8031 */ {(12<<2)|2,{97,107,0}},
/* 8032 */ {(12<<2)|2,{97,113,0}},
/* 8033 */ {(12<<2)|2,{97,113,0}},
/* 8034 */ {(12<<2)|2,{97,113,0}},
/* 8035 */ {(12<<2)|2,{97,113,0}},
/* 8036 */ {(12<<2)|2,{97,113,0}},
/* 8037 */ {(12<<2)|2,{97,113,0}},
/* 8038 */ {(12<<2)|2,{97,113,0}},
/* 8039 */ {(12<<2)|2,{97,113,0}},
/* 8040 */ {(12<<2)|2,{97,113,0}},
/* 8041 */ {(12<<2)|2,{97,113,0}},
/* 8042 */ {(12<<2)|2,{97,113,0}},
/* 8043 */ {(12<<2)|2,{97,113,0}},
/* 8044 */ {(12<<2)|2,{97,113,0}},
/* 8045 */ {(12<<2)|2,{97,113,0}},
/* 8046 */ {(12<<2)|2,{97,113,0}},
/* 8047 */ {(12<<2)|2,{97,113,0}},
/* 8048 */ {(12<<2)|2,{97,118,0}},
/* 8049 */ {(12<<2)|2,{97,118,0}},
/* 8050 */ {(12<<2)|2,{97,118,0}},
/* 8051 */ {(12<<2)|2,{97,118,0}},
/* 8052 */ {(12<<2)|2,{97,118,0}},
/* 8053 */ {(12<<2)|2,{97,118,0}},
/* 8054 */ {(12<<2)|2,{97,118,0}},
/* 8055 */ {(12<<2)|2,{97,118,0}},
/* 8056 */ {(12<<2)|2,{97,118,0}},
/* 8057 */ {(12<<2)|2,{97,118,0}},
/* 8058 */ {(12<<2)|2,{97,118,0}},
/* 8059 */ {(12<<2)|2,{97,118,0}},
/* 8060 */ {(12<<2)|2,{97,118,0}},
/* 8061 */ {(12<<2)|2,{97,118,0}},
/* 8062 */ {(12<<2)|2,{97,118,0}},
/* 8063 */ {(12<<2)|2,{97,118,0}},
/* 8064 */ {(12<<2)|2,{97,119,0}},
/* 8065 */ {(12<<2)|2,{97,119,0}},
/* 8066 */ {(12<<2)|2,{97,119,0}},
/* 8067 */ {(12<<2)|2,{97,119,0}},
/* 8068 */ {(12<<2)|2,{97,119,0}},
/* 8069 */ {(12<<2)|2,{97,119,0}},
/* 8070 */ {(12<<2)|2,{97,119,0}},
/* 8071 */ {(12<<2)|2,{97,119,0}},
/* 8072 */ {(12<<2)|2,{97,119,0}},
/* 8073 */ {(12<<2)|2,{97,119,0}},
/* 8074 */ {(12<<2)|2,{97,119,0}},
/* 8075 */ {(12<<2)|2,{97,119,0}},
/* 8076 */ {(12<<2)|2,{97,119,0}},
/* 8077 */ {(12<<2)|2,{97,119,0}},
/* 8078 */ {(12<<2)|2,{97,119,0}},
/* 8079 */ {(12<<2)|2,{97,119,0}},
/* 8080 */ {(12<<2)|2,{97,120,0}},
/* 8081 */ {(12<<2)|2,{97,120,0}},
/* 8082 */ {(12<<2)|2,{97,120,0}},
/* 8083 */ {(12<<2)|2,{97,120,0}},
/* 8084 */ {(12<<2)|2,{97,120,0}},
/* 8085 */ {(12<<2)|2,{97,120,0}},
/* 8086 */ {(12<<2)|2,{97,120,0}},
/* 8087 */ {(12<<2)|2,{97,120,0}},
/* 8088 */ {(12<<2)|2,{97,120,0}},
/* 8089 */ {(12<<2)|2,{97,120,0}},
/* 8090 */ {(12<<2)|2,{97,120,0}},
/* 8091 */ {(12<<2)|2,{97,120,0}},
/* 8092 */ {(12<<2)|2,{97,120,0}},
/* 8093 */ {(12<<2)|2,{97,120,0}},
/* 8094 */ {(12<<2)|2,{97,120,0}},
/* 8095 */ {(12<<2)|2,{97,120,0}},
/* 8096 */ {(12<<2)|2,{97,121,0}},
/* 8097 */ {(12<<2)|2,{97,121,0}},
/* 8098 */ {(12<<2)|2,{97,121,0}},
/* 8099 */ {(12<<2)|2,{97,121,0}},
/* 8100 */ {(12<<2)|2,{97,121,0}},
/* 8101 */ {(12<<2)|2,{97,121,0}},
/* 8102 */ {(12<<2)|2,{97,121,0}},
/* 8103 */ {(12<<2)|2,{97,121,0}},
/* 8104 */ {(12<<2)|2,{97,121,0}},
/* 8105 */ {(12<<2)|2,{97,121,0}},
/* 8106 */ {(12<<2)|2,{97,121,0}},
/* 8107 */ {(12<<2)|2,{97,121,0}},
/* 8108 */ {(12<<2)|2,{97,121,0}},
/* 8109 */ {(12<<2)|2,{97,121,0}},
/* 8110 */ {(12<<2)|2,{97,121,0}},
/* 8111 */ {(12<<2)|2,{97,121,0}},
/* 8112 */ {(12<<2)|2,{97,122,0}},
/* 8113 */ {(12<<2)|2,{97,122,0}},
/* 8114 */ {(12<<2)|2,{97,122,0}},
/* 8115 */ {(12<<2)|2,{97,122,0}},
/* 8116 */ {(12<<2)|2,{97,122,0}},
/* 8117 */ {(12<<2)|2,{97,122,0}},
/* 8118 */ {(12<<2)|2,{97,122,0}},
/* 8119 */ {(12<<2)|2,{97,122,0}},
/* 8120 */ {(12<<2)|2,{97,122,0}},
/* 8121 */ {(12<<2)|2,{97,122,0}},
/* 8122 */ {(12<<2)|2,{97,122,0}},
/* 8123 */ {(12<<2)|2,{97,122,0}},
/* 8124 */ {(12<<2)|2,{97,122,0}},
/* 8125 */ {(12<<2)|2,{97,122,0}},
/* 8126 */ {(12<<2)|2,{97,122,0}},
/* 8127 */ {(12<<2)|2,{97,122,0}},
/* 8128 */ {(13<<2)|2,{97,38,0}},
/* 8129 */ {(13<<2)|2,{97,38,0}},
/* 8130 */ {(13<<2)|2,{97,38,0}},
/* 8131 */ {(13<<2)|2,{97,38,0}},
/* 8132 */ {(13<<2)|2,{97,38,0}},
/* 8133 */ {(13<<2)|2,{97,38,0}},
/* 8134 */ {(13<<2)|2,{97,38,0}},
/* 8135 */ {(13<<2)|2,{97,38,0}},
/* 8136 */ {(13<<2)|2,{97,42,0}},
/* 8137 */ {(13<<2)|2,{97,42,0}},
/* 8138 */ {(13<<2)|2,{97,42,0}},
/* 8139 */ {(13<<2)|2,{97,42,0}},
/* 8140 */ {(13<<2)|2,{97,42,0}},
/* 8141 */ {(13<<2)|2,{97,42,0}},
/* 8142 */ {(13<<2)|2,{97,42,0}},
/* 8143 */ {(13<<2)|2,{97,42,0}},
/* 8144 */ {(13<<2)|2,{97,44,0}},
/* 8145 */ {(13<<2)|2,{97,44,0}},
/* 8146 */ {(13<<2)|2,{97,44,0}},
/* 8147 */ {(13<<2)|2,{97,44,0}},
/* 8148 */ {(13<<2)|2,{97,44,0}},
/* 8149 */ {(13<<2)|2,{97,44,0}},
/* 8150 */ {(13<<2)|2,{97,44,0}},
/* 8151 */ {(13<<2)|2,{97,44,0}},
/* 8152 */ {(13<<2)|2,{97,59,0}},
/* 8153 */ {(13<<2)|2,{97,59,0}},
/* 8154 */ {(13<<2)|2,{97,59,0}},
/* 8155 */ {(13<<2)|2,{97,59,0}},
/* 8156 */ {(13<<2)|2,{97,59,0}},
/* 8157 */ {(13<<2)|2,{97,59,0}},
/* 8158 */ {(13<<2)|2,{97,59,0}},
/* 8159 */ {(13<<2)|2,{97,59,0}},
/* 8160 */ {(13<<2)|2,{97,88,0}},
/* 8161 */ {(13<<2)|2,{97,88,0}},
/* 8162 */ {(13<<2)|2,{97,88,0}},
/* 8163 */ {(13<<2)|2,{97,88,0}},
/* 8164 */ {(13<<2)|2,{97,88,0}},
/* 8165 */ {(13<<2)|2,{97,88,0}},
/* 8166 */ {(13<<2)|2,{97,88,0}},
/* 8167 */ {(13<<2)|2,{97,88,0}},
/* 8168 */ {(13<<2)|2,{97,90,0}},
/* 8169 */ {(13<<2)|2,{97,90,0}},
/* 8170 */ {(13<<2)|2,{97,90,0}},
/* 8171 */ {(13<<2)|2,{97,90,0}},
/* 8172 */ {(13<<2)|2,{97,90,0}},
/* 8173 */ {(13<<2)|2,{97,90,0}},
/* 8174 */ {(13<<2)|2,{97,90,0}},
/* 8175 */ {(13<<2)|2,{97,90,0}},
/* 8176 */ {(15<<2)|2,{97,33,0}},
/* 8177 */ {(15<<2)|2,{97,33,0}},
/* 8178 */ {(15<<2)|2,{97,34,0}},
/* 8179 */ {(15<<2)|2,{97,34,0}},
/* 8180 */ {(15<<2)|2,{97,40,0}},
/* 8181 */ {(15<<2)|2,{97,40,0}},
/* 8182 */ {(15<<2)|2,{97,41,0}},
/* 8183 */ {(15<<2)|2,{97,41,0}},
/* 8184 */ {(15<<2)|2,{97,63,0}},
/* 8185 */ {(15<<2)|2,{97,63,0}},
/* 8186 */ {(16<<2)|2,{97,39,0}},
/* 8187 */ {(16<<2)|2,{97,43,0}},
/* 8188 */ {(16<<2)|2,{97,124,0}},
/* 8189 */ {(5<<2)|1,{97,0,0}},
/* 8190 */ {(5<<2)|1,{97,0,0}},
/* 8191 */ {(5<<2)|1,{97,0,0}},
/* 8192 */ {(15<<2)|3,{99,48,48}},
/* 8193 */ {(15<<2)|3,{99,48,48}},
/* 8194 */ {(15<<2)|3,{99,48,49}},
/* 8195 */ {(15<<2)|3,{99,48,49}},
/* 8196 */ {(15<<2)|3,{99,48,50}},
/* 8197 */ {(15<<2)|3,{99,48,50}},
/* 8198 */ {(15<<2)|3,{99,48,97}},
/* 8199 */ {(15<<2)|3,{99,48,97}},
/* 8200 */ {(15<<2)|3,{99,48,99}},
/* 8201 */ {(15<<2)|3,{99,48,99}},
/* 8202 */ {(15<<2)|3,{99,48,101}},
/* 8203 */ {(15<<2)|3,{99,48,101}},
/* 8204 */ {(15<<2)|3,{99,48,105}},
/* 8205 */ {(15<<2)|3,{99,48,105}},
/* 8206 */ {(15<<2)|3,{99,48,111}},
/* 8207 */ {(15<<2)|3,{99,48,111}},
/* 8208 */ {(15<<2)|3,{99,48,115}},
/* 8209 */ {(15<<2)|3,{99,48,115}},
/* 8210 */ {(15<<2)|3,{99,48,116}},
/* 8211 */ {(15<<2)|3,{99,48,116}},
/* 8212 */ {(16<<2)|3,{99,48,32}},
/* 8213 */ {(16<<2)|3,{99,48,37}},
/* 8214 */ {(16<<2)|3,{99,48,45}},
/* 8215 */ {(16<<2)|3,{99,48,46}},
/* 8216 */ {(16<<2)|3,{99,48,47}},
/* 8217 */ {(16<<2)|3,{99,48,51}},
/* 8218 */ {(16<<2)|3,{99,48,52}},
/* 8219 */ {(16<<2)|3,{99,48,53}},
/* 8220 */ {(16<<2)|3,{99,48,54}},
/* 8221 */ {(16<<2)|3,{99,48,55}},
/* 8222 */ {(16<<2)|3,{99,48,56}},
/* 8223 */ {(16<<2)|3,{99,48,57}},
/* 8224 */ {(16<<2)|3,{99,48,61}},
/* 8225 */ {(16<<2)|3,{99,48,65}},
/* 8226 */ {(16<<2)|3,{99,48,95}},
/* 8227 */ {(16<<2)|3,{99,48,98}},
/* 8228 */ {(16<<2)|3,{99,48,100}},
/* 8229 */ {(16<<2)|3,{99,48,102}},
/* 8230 */ {(16<<2)|3,{99,48,103}},
/* 8231 */ {(16<<2)|3,{99,48,104}},
/* 8232 */ {(16<<2)|3,{99,48,108}},
/* 8233 */ {(16<<2)|3,{99,48,109}},
/* 8234 */ {(16<<2)|3,{99,48,110}},
/* 8235 */ {(16<<2)|3,{99,48,112}},
/* 8236 */ {(16<<2)|3,{99,48,114}},
/* 8237 */ {(16<<2)|3,{99,48,117}},
/* 8238 */ {(10<<2)|2,{99,48,0}},
/* 8239 */ {(10<<2)|2,{99,48,0}},
/* 8240 */ {(10<<2)|2,{99,48,0}},
/* 8241 */ {(10<<2)|2,{99,48,0}},
/* 8242 */ {(10<<2)|2,{99,48,0}},
/* 8243 */ {(10<<2)|2,{99,48,0}},
/* 8244 */ {(10<<2)|2,{99,48,0}},
/* 8245 */ {(10<<2)|2,{99,48,0}},
/* 8246 */ {(10<<2)|2,{99,48,0}},
/* 8247 */ {(10<<2)|2,{99,48,0}},
/* 8248 */ {(10<<2)|2,{99,48,0}},
/* 8249 */ {(10<<2)|2,{99,48,0}},
/* 8250 */ {(10<<2)|2,{99,48,0}},
/* 8251 */ {(10<<2)|2,{99,48,0}},
/* 8252 */ {(10<<2)|2,{99,48,0}},
/* 8253 */ {(10<<2)|2,{99,48,0}},
/* 8254 */ {(10<<2)|2,{99,48,0}},
/* 8255 */ {(10<<2)|2,{99,48,0}},
/* 8256 */ {(15<<2)|3,{99,49,48}},
/* 8257 */ {(15<<2)|3,{99,49,48}},
/* 8258 */ {(15<<2)|3,{99,49,49}},
/* 8259 */ {(15<<2)|3,{99,49,49}},
/* 8260 */ {(15<<2)|3,{99,49,50}},
/* 8261 */ {(15<<2)|3,{99,49,50}},
/* 8262 */ {(15<<2)|3,{99,49,97}},
/* 8263 */ {(15<<2)|3,{99,49,97}},
/* 8264 */ {(15<<2)|3,{99,49,99}},
/* 8265 */ {(15<<2)|3,{99,49,99}},
/* 8266 */ {(15<<2)|3,{99,49,101}},
/* 8267 */ {(15<<2)|3,{99,49,101}},
/* 8268 */ {(15<<2)|3,{99,49,105}},
/* 8269 */ {(15<<2)|3,{99,49,105}},
/* 8270 */ {(15<<2)|3,{99,49,111}},
/* 8271 */ {(15<<2)|3,{99,49,111}},
/* 8272 */ {(15<<2)|3,{99,49,115}},
/* 8273 */ {(15<<2)|3,{99,49,115}},
/* 8274 */ {(15<<2)|3,{99,49,116}},
/* 8275 */ {(15<<2)|3,{99,49,116}},
/* 8276 */ {(16<<2)|3,{99,49,32}},
/* 8277 */ {(16<<2)|3,{99,49,37}},
/* 8278 */ {(16<<2)|3,{99,49,45}},
/* 8279 */ {(16<<2)|3,{99,49,46}},
/* 8280 */ {(16<<2)|3,{99,49,47}},
/* 8281 */ {(16<<2)|3,{99,49,51}},
/* 8282 */ {(16<<2)|3,{99,49,52}},
/* 8283 */ {(16<<2)|3,{99,49,53}},
/* 8284 */ {(16<<2)|3,{99,49,54}},
/* 8285 */ {(16<<2)|3,{99,49,55}},
/* 8286 */ {(16<<2)|3,{99,49,56}},
/* 8287 */ {(16<<2)|3,{99,49,57}},
/* 8288 */ {(16<<2)|3,{99,49,61}},
/* 8289 */ {(16<<2)|3,{99,49,65}},
/* 8290 */ {(16<<2)|3,{99,49,95}},
/* 8291 */ {(16<<2)|3,{99,49,98}},
/* 8292 */ {(16<<2)|3,{99,49,100}},
/* 8293 */ {(16<<2)|3,{99,49,102}},
/* 8294 */ {(16<<2)|3,{99,49,103}},
/* 8295 */ {(16<<2)|3,{99,49,104}},
/* 8296 */ {(16<<2)|3,{99,49,108}},
/* 8297 */ {(16<<2)|3,{99,49,109}},
/* 8298 */ {(16<<2)|3,{99,49,110}},
/* 8299 */ {(16<<2)|3,{99,49,112}},
/* 8300 */ {(16<<2)|3,{99,49,114}},
/* 8301 */ {(16<<2)|3,{99,49,117}},
/* 8302 */ {(10<<2)|2,{99,49,0}},
/* 8303 */ {(10<<2)|2,{99,49,0}},
/* 8304 */ {(10<<2)|2,{99,49,0}},
/* 8305 */ {(10<<2)|2,{99,49,0}},
/* 8306 */ {(10<<2)|2,{99,49,0}},
/* 8307 */ {(10<<2)|2,{99,49,0}},
/* 8308 */ {(10<<2)|2,{99,49,0}},
/* 8309 */ {(10<<2)|2,{99,49,0}},
/* 8310 */ {(10<<2)|2,{99,49,0}},
/* 8311 */ {(10<<2)|2,{99,49,0}},
/* 8312 */ {(10<<2)|2,{99,49,0}},
/* 8313 */ {(10<<2)|2,{99,49,0}},
/* 8314 */ {(10<<2)|2,{99,49,0}},
/* 8315 */ {(10<<2)|2,{99,49,0}},
/* 8316 */ {(10<<2)|2,{99,49,0}},
/* 8317 */ {(10<<2)|2,{99,49,0}},
/* 8318 */ {(10<<2)|2,{99,49,0}},
/* 8319 */ {(10<<2)|2,{99,49,0}},
/* 8320 */ {(15<<2)|3,{99,50,48}},
/* 8321 */ {(15<<2)|3,{99,50,48}},
/* 8322 */ {(15<<2)|3,{99,50,49}},
/* 8323 */ {(15<<2)|3,{99,50,49}},
/* 8324 */ {(15<<2)|3,{99,50,50}},
/* 8325 */ {(15<<2)|3,{99,50,50}},
/* 8326 */ {(15<<2)|3,{99,50,97}},
/* 8327 */ {(15<<2)|3,{99,50,97}},
/* 8328 */ {(15<<2)|3,{99,50,99}},
/* 8329 */ {(15<<2)|3,{99,50,99}},
/* 8330 */ {(15<<2)|3,{99,50,101}},
/* 8331 */ {(15<<2)|3,{99,50,101}},
/* 8332 */ {(15<<2)|3,{99,50,105}},
/* 8333 */ {(15<<2)|3,{99,50,105}},
/* 8334 */ {(15<<2)|3,{99,50,111}},
/* 8335 */ {(15<<2)|3,{99,50,111}},
/* 8336 */ {(15<<2)|3,{99,50,115}},
/* 8337 */ {(15<<2)|3,{99,50,115}},
/* 8338 */ {(15<<2)|3,{99,50,116}},
/* 8339 */ {(15<<2)|3,{99,50,116}},
/* 8340 */ {(16<<2)|3,{99,50,32}},
/* 8341 */ {(16<<2)|3,{99,50,37}},
/* 8342 */ {(16<<2)|3,{99,50,45}},
/* 8343 */ {(16<<2)|3,{99,50,46}},
/* 8344 */ {(16<<2)|3,{99,50,47}},
/* 8345 */ {(16<<2)|3,{99,50,51}},
/* 8346 */ {(16<<2)|3,{99,50,52}},
/* 8347 */ {(16<<2)|3,{99,50,53}},
/* 8348 */ {(16<<2)|3,{99,50,54}},
/* 8349 */ {(16<<2)|3,{99,50,55}},
/* 8350 */ {(16<<2)|3,{99,50,56}},
/* 8351 */ {(16<<2)|3,{99,50,57}},
/* 8352 */ {(16<<2)|3,{99,50,61}},
/* 8353 */ {(16<<2)|3,{99,50,65}},
/* 8354 */ {(16<<2)|3,{99,50,95}},
/* 8355 */ {(16<<2)|3,{99,50,98}},
/* 8356 */ {(16<<2)|3,{99,50,100}},
/* 8357 */ {(16<<2)|3,{99,50,102}},
/* 8358 */ {(16<<2)|3,{99,50,103}},
/* 8359 */ {(16<<2)|3,{99,50,104}},
/* 8360 */ {(16<<2)|3,{99,50,108}},
/* 8361 */ {(16<<2)|3,{99,50,109}},
/* 8362 */ {(16<<2)|3,{99,50,110}},
/* 8363 */ {(16<<2)|3,{99,50,112}},
/* 8364 */ {(16<<2)|3,{99,50,114}},
/* 8365 */ {(16<<2)|3,{99,50,117}},
/* 8366 */ {(10<<2)|2,{99,50,0}},
/* 8367 */ {(10<<2)|2,{99,50,0}},
/* 8368 */ {(10<<2)|2,{99,50,0}},
/* 8369 */ {(10<<2)|2,{99,50,0}},
/* 8370 */ {(10<<2)|2,{99,50,0}},
/* 8371 */ {(10<<2)|2,{99,50,0}},
/* 8372 */ {(10<<2)|2,{99,50,0}},
/* 8373 */ {(10<<2)|2,{99,50,0}},
/* 8374 */ {(10<<2)|2,{99,50,0}},
/* 8375 */ {(10<<2)|2,{99,50,0}},
/* 8376 */ {(10<<2)|2,{99,50,0}},
/* 8377 */ {(10<<2)|2,{99,50,0}},
/* 8378 */ {(10<<2)|2,{99,50,0}},
/* 8379 */ {(10<<2)|2,{99,50,0}},
/* 8380 */ {(10<<2)|2,{99,50,0}},
/* 8381 */ {(10<<2)|2,{99,50,0}},
/* 8382 */ {(10<<2)|2,{99,50,0}},
/* 8383 */ {(10<<2)|2,{99,50,0}},
/* 8384 */ {(15<<2)|3,{99,97,48}},
/* 8385 */ {(15<<2)|3,{99,97,48}},
/* 8386 */ {(15<<2)|3,{99,97,49}},
/* 8387 */ {(15<<2)|3,{99,97,49}},
/* 8388 */ {(15<<2)|3,{99,97,50}},
/* 8389 */ {(15<<2)|3,{99,97,50}},
/* 8390 */ {(15<<2)|3,{99,97,97}},
/* 8391 */ {(15<<2)|3,{99,97,97}},
/* 8392 */ {(15<<2)|3,{99,97,99}},
/* 8393 */ {(15<<2)|3,{99,97,99}},
/* 8394 */ {(15<<2)|3,{99,97,101}},
/* 8395 */ {(15<<2)|3,{99,97,101}},
/* 8396 */ {(15<<2)|3,{99,97,105}},
/* 8397 */ {(15<<2)|3,{99,97,105}},
/* 8398 */ {(15<<2)|3,{99,97,111}},
/* 8399 */ {(15<<2)|3,{99,97,111}},
/* 8400 */ {(15<<2)|3,{99,97,115}},
/* 8401 */ {(15<<2)|3,{99,97,115}},
/* 8402 */ {(15<<2)|3,{99,97,116}},
/* 8403 */ {(15<<2)|3,{99,97,116}},
/* 8404 */ {(16<<2)|3,{99,97,32}},
/* 8405 */ {(16<<2)|3,{99,97,37}},
/* 8406 */ {(16<<2)|3,{99,97,45}},
/* 8407 */ {(16<<2)|3,{99,97,46}},
/* 8408 */ {(16<<2)|3,{99,97,47}},
/* 8409 */ {(16<<2)|3,{99,97,51}},
/* 8410 */ {(16<<2)|3,{99,97,52}},
/* 8411 */ {(16<<2)|3,{99,97,53}},
/* 8412 */ {(16<<2)|3,{99,97,54}},
/* 8413 */ {(16<<2)|3,{99,97,55}},
/* 8414 */ {(16<<2)|3,{99,97,56}},
/* 8415 */ {(16<<2)|3,{99,97,57}},
/* 8416 */ {(16<<2)|3,{99,97,61}},
/* 8417 */ {(16<<2)|3,{99,97,65}},
/* 8418 */ {(16<<2)|3,{99,97,95}},
/* 8419 */ {(16<<2)|3,{99,97,98}},
/* 8420 */ {(16<<2)|3,{99,97,100}},
/* 8421 */ {(16<<2)|3,{99,97,102}},
/* 8422 */ {(16<<2)|3,{99,97,103}},
/* 8423 */ {(16<<2)|3,{99,97,104}},
/* 8424 */ {(16<<2)|3,{99,97,108}},
/* 8425 */ {(16<<2)|3,{99,97,109}},
/* 8426 */ {(16<<2)|3,{99,97,110}},
/* 8427 */ {(16<<2)|3,{99,97,112}},
/* 8428 */ {(16<<2)|3,{99,97,114}},
/* 8429 */ {(16<<2)|3,{99,97,117}},
/* 8430 */ {(10<<2)|2,{99,97,0}},
/* 8431 */ {(10<<2)|2,{99,97,0}},
/* 8432 */ {(10<<2)|2,{99,97,0}},
/* 8433 */ {(10<<2)|2,{99,97,0}},
/* 8434 */ {(10<<2)|2,{99,97,0}},
/* 8435 */ {(10<<2)|2,{99,97,0}},
/* 8436 */ {(10<<2)|2,{99,97,0}},
/* 8437 */ {(10<<2)|2,{99,97,0}},
/* 8438 */ {(10<<2)|2,{99,97,0}},
/* 8439 */ {(10<<2)|2,{99,97,0}},
/* 8440 */ {(10<<2)|2,{99,97,0}},
/* 8441 */ {(10<<2)|2,{99,97,0}},
/* 8442 */ {(10<<2)|2,{99,97,0}},
/* 8443 */ {(10<<2)|2,{99,97,0}},
/* 8444 */ {(10<<2)|2,{99,97,0}},
/* 8445 */ {(10<<2)|2,{99,97,0}},
/* 8446 */ {(10<<2)|2,{99,97,0}},
/* 8447 */ {(10<<2)|2,{99,97,0}},
/* 8448 */ {(15<<2)|3,{99,99,48}},
/* 8449 */ {(15<<2)|3,{99,99,48}},
/* 8450 */ {(15<<2)|3,{99,99,49}},
/* 8451 */ {(15<<2)|3,{99,99,49}},
/* 8452 */ {(15<<2)|3,{99,99,50}},
/* 8453 */ {(15<<2)|3,{99,99,50}},
/* 8454 */ {(15<<2)|3,{99,99,97}},
/* 8455 */ {(15<<2)|3,{99,99,97}},
/* 8456 */ {(15<<2)|3,{99,99,99}},
/* 8457 */ {(15<<2)|3,{99,99,99}},
/* 8458 */ {(15<<2)|3,{99,99,101}},
/* 8459 */ {(15<<2)|3,{99,99,101}},
/* 8460 */ {(15<<2)|3,{99,99,105}},
/* 8461 */ {(15<<2)|3,{99,99,105}},
/* 8462 */ {(15<<2)|3,{99,99,111}},
/* 8463 */ {(15<<2)|3,{99,99,111}},
/* 8464 */ {(15<<2)|3,{99,99,115}},
/* 8465 */ {(15<<2)|3,{99,99,115}},
/* 8466 */ {(15<<2)|3,{99,99,116}},
/* 8467 */ {(15<<2)|3,{99,99,116}},
/* 8468 */ {(16<<2)|3,{99,99,32}},
/* 8469 */ {(16<<2)|3,{99,99,37}},
/* 8470 */ {(16<<2)|3,{99,99,45}},
/* 8471 */ {(16<<2)|3,{99,99,46}},
/* 8472 */ {(16<<2)|3,{99,99,47}},
/* 8473 */ {(16<<2)|3,{99,99,51}},
/* 8474 */ {(16<<2)|3,{99,99,52}},
/* 8475 */ {(16<<2)|3,{99,99,53}},
/* 8476 */ {(16<<2)|3,{99,99,54}},
/* 8477 */ {(16<<2)|3,{99,99,55}},
/* 8478 */ {(16<<2)|3,{99,99,56}},
/* 8479 */ {(16<<2)|3,{99,99,57}},
/* 8480 */ {(16<<2)|3,{99,99,61}},
/* 8481 */ {(16<<2)|3,{99,99,65}},
/* 8482 */ {(16<<2)|3,{99,99,95}},
/* 8483 */ {(16<<2)|3,{99,99,98}},
/* 8484 */ {(16<<2)|3,{99,99,100}},
/* 8485 */ {(16<<2)|3,{99,99,102}},
/* 8486 */ {(16<<2)|3,{99,99,103}},
/* 8487 */ {(16<<2)|3,{99,99,104}},
/* 8488 */ {(16<<2)|3,{99,99,108}},
/* 8489 */ {(16<<2)|3,{99,99,109}},
/* 8490 */ {(16<<2)|3,{99,99,110}},
/* 8491 */ {(16<<2)|3,{99,99,112}},
/* 8492 */ {(16<<2)|3,{99,99,114}},
/* 8493 */ {(16<<2)|3,{99,99,117}},
/* 8494 */ {(10<<2)|2,{99,99,0}},
/* 8495 */ {(10<<2)|2,{99,99,0}},
/* 8496 */ {(10<<2)|2,{99,99,0}},
/* 8497 */ {(10<<2)|2,{99,99,0}},
/* 8498 */ {(10<<2)|2,{99,99,0}},
/* 8499 */ {(10<<2)|2,{99,99,0}},
/* 8500 */ {(10<<2)|2,{99,99,0}},
/* 8501 */ {(10<<2)|2,{99,99,0}},
/* 8502 */ {(10<<2)|2,{99,99,0}},
/* 8503 */ {(10<<2)|2,{99,99,0}},
/* 8504 */ {(10<<2)|2,{99,99,0}},
/* 8505 */ {(10<<2)|2,{99,99,0}},
/* 8506 */ {(10<<2)|2,{99,99,0}},
/* 8507 */ {(10<<2)|2,{99,99,0}},
/* 8508 */ {(10<<2)|2,{99,99,0}},
/* 8509 */ {(10<<2)|2,{99,99,0}},
/* 8510 */ {(10<<2)|2,{99,99,0}},
/* 8511 */ {(10<<2)|2,{99,99,0}},
/* 8512 */ {(15<<2)|3,{99,101,48}},
/* 8513 */ {(15<<2)|3,{99,101,48}},
/* 8514 */ {(15<<2)|3,{99,101,49}},
/* 8515 */ {(15<<2)|3,{99,101,49}},
/* 8516 */ {(15<<2)|3,{99,101,50}},
/* 8517 */ {(15<<2)|3,{99,101,50}},
/* 8518 */ {(15<<2)|3,{99,101,97}},
/* 8519 */ {(15<<2)|3,{99,101,97}},
/* 8520 */ {(15<<2)|3,{99,101,99}},
/* 8521 */ {(15<<2)|3,{99,101,99}},
/* 8522 */ {(15<<2)|3,{99,101,101}},
/* 8523 */ {(15<<2)|3,{99,101,101}},
/* 8524 */ {(15<<2)|3,{99,101,105}},
/* 8525 */ {(15<<2)|3,{99,101,105}},
/* 8526 */ {(15<<2)|3,{99,101,111}},
/* 8527 */ {(15<<2)|3,{99,101,111}},
/* 8528 */ {(15<<2)|3,{99,101,115}},
/* 8529 */ {(15<<2)|3,{99,101,115}},
/* 8530 */ {(15<<2)|3,{99,101,116}},
/* 8531 */ {(15<<2)|3,{99,101,116}},
/* 8532 */ {(16<<2)|3,{99,101,32}},
/* 8533 */ {(16<<2)|3,{99,101,37}},
/* 8534 */ {(16<<2)|3,{99,101,45}},
/* 8535 */ {(16<<2)|3,{99,101,46}},
/* 8536 */ {(16<<2)|3,{99,101,47}},
/* 8537 */ {(16<<2)|3,{99,101,51}},
/* 8538 */ {(16<<2)|3,{99,101,52}},
/* 8539 */ {(16<<2)|3,{99,101,53}},
/* 8540 */ {(16<<2)|3,{99,101,54}},
/* 8541 */ {(16<<2)|3,{99,101,55}},
/* 8542 */ {(16<<2)|3,{99,101,56}},
/* 8543 */ {(16<<2)|3,{99,101,57}},
/* 8544 */ {(16<<2)|3,{99,101,61}},
/* 8545 */ {(16<<2)|3,{99,101,65}},
/* 8546 */ {(16<<2)|3,{99,101,95}},
/* 8547 */ {(16<<2)|3,{99,101,98}},
/* 8548 */ {(16<<2)|3,{99,101,100}},
/* 8549 */ {(16<<2)|3,{99,101,102}},
/* 8550 */ {(16<<2)|3,{99,101,103}},
/* 8551 */ {(16<<2)|3,{99,101,104}},
/* 8552 */ {(16<<2)|3,{99,101,108}},
/* 8553 */ {(16<<2)|3,{99,101,109}},
/* 8554 */ {(16<<2)|3,{99,101,110}},
/* 8555 */ {(16<<2)|3,{99,101,112}},
/* 8556 */ {(16<<2)|3,{99,101,114}},
/* 8557 */ {(16<<2)|3,{99,101,117}},
/* 8558 */ {(10<<2)|2,{99,101,0}},
/* 8559 */ {(10<<2)|2,{99,101,0}},
/* 8560 */ {(10<<2)|2,{99,101,0}},
/* 8561 */ {(10<<2)|2,{99,101,0}},
/* 8562 */ {(10<<2)|2,{99,101,0}},
/* 8563 */ {(10<<2)|2,{99,101,0}},
/* 8564 */ {(10<<2)|2,{99,101,0}},
/* 8565 */ {(10<<2)|2,{99,101,0}},
/* 8566 */ {(10<<2)|2,{99,101,0}},
/* 8567 */ {(10<<2)|2,{99,101,0}},
/* 8568 */ {(10<<2)|2,{99,101,0}},
/* 8569 */ {(10<<2)|2,{99,101,0}},
/* 8570 */ {(10<<2)|2,{99,101,0}},
/* 8571 */ {(10<<2)|2,{99,101,0}},
/* 8572 */ {(10<<2)|2,{99,101,0}},
/* 8573 */ {(10<<2)|2,{99,101,0}},
/* 8574 */ {(10<<2)|2,{99,101,0}},
/* 8575 */ {(10<<2)|2,{99,101,0}},
/* 8576 */ {(15<<2)|3,{99,105,48}},
/* 8577 */ {(15<<2)|3,{99,105,48}},
/* 8578 */ {(15<<2)|3,{99,105,49}},
/* 8579 */ {(15<<2)|3,{99,105,49}},
/* 8580 */ {(15<<2)|3,{99,105,50}},
/* 8581 */ {(15<<2)|3,{99,105,50}},
/* 8582 */ {(15<<2)|3,{99,105,97}},
/* 8583 */ {(15<<2)|3,{99,105,97}},
/* 8584 */ {(15<<2)|3,{99,105,99}},
/* 8585 */ {(15<<2)|3,{99,105,99}},
/* 8586 */ {(15<<2)|3,{99,105,101}},
/* 8587 */ {(15<<2)|3,{99,105,101}},
/* 8588 */ {(15<<2)|3,{99,105,105}},
/* 8589 */ {(15<<2)|3,{99,105,105}},
/* 8590 */ {(15<<2)|3,{99,105,111}},
/* 8591 */ {(15<<2)|3,{99,105,111}},
/* 8592 */ {(15<<2)|3,{99,105,115}},
/* 8593 */ {(15<<2)|3,{99,105,115}},
/* 8594 */ {(15<<2)|3,{99,105,116}},
/* 8595 */ {(15<<2)|3,{99,105,116}},
/* 8596 */ {(16<<2)|3,{99,105,32}},
/* 8597 */ {(16<<2)|3,{99,105,37}},
/* 8598 */ {(16<<2)|3,{99,105,45}},
/* 8599 */ {(16<<2)|3,{99,105,46}},
/* 8600 */ {(16<<2)|3,{99,105,47}},
/* 8601 */ {(16<<2)|3,{99,105,51}},
/* 8602 */ {(16<<2)|3,{99,105,52}},
/* 8603 */ {(16<<2)|3,{99,105,53}},
/* 8604 */ {(16<<2)|3,{99,105,54}},
/* 8605 */ {(16<<2)|3,{99,105,55}},
/* 8606 */ {(16<<2)|3,{99,105,56}},
/* 8607 */ {(16<<2)|3,{99,105,57}},
/* 8608 */ {(16<<2)|3,{99,105,61}},
/* 8609 */ {(16<<2)|3,{99,105,65}},
/* 8610 */ {(16<<2)|3,{99,105,95}},
/* 8611 */ {(16<<2)|3,{99,105,98}},
/* 8612 */ {(16<<2)|3,{99,105,100}},
/* 8613 */ {(16<<2)|3,{99,105,102}},
/* 8614 */ {(16<<2)|3,{99,105,103}},
/* 8615 */ {(16<<2)|3,{99,105,104}},
/* 8616 */ {(16<<2)|3,{99,105,108}},
/* 8617 */ {(16<<2)|3,{99,105,109}},
/* 8618 */ {(16<<2)|3,{99,105,110}},
/* 8619 */ {(16<<2)|3,{99,105,112}},
/* 8620 */ {(16<<2)|3,{99,105,114}},
/* 8621 */ {(16<<2)|3,{99,105,117}},
/* 8622 */ {(10<<2)|2,{99,105,0}},
/* 8623 */ {(10<<2)|2,{99,105,0}},
/* 8624 */ {(10<<2)|2,{99,105,0}},
/* 8625 */ {(10<<2)|2,{99,105,0}},
/* 8626 */ {(10<<2)|2,{99,105,0}},
/* 8627 */ {(10<<2)|2,{99,105,0}},
/* 8628 */ {(10<<2)|2,{99,105,0}},
/* 8629 */ {(10<<2)|2,{99,105,0}},
/* 8630 */ {(10<<2)|2,{99,105,0}},
/* 8631 */ {(10<<2)|2,{99,105,0}},
/* 8632 */ {(10<<2)|2,{99,105,0}},
/* 8633 */ {(10<<2)|2,{99,105,0}},
/* 8634 */ {(10<<2)|2,{99,105,0}},
/* 8635 */ {(10<<2)|2,{99,105,0}},
/* 8636 */ {(10<<2)|2,{99,105,0}},
/* 8637 */ {(10<<2)|2,{99,105,0}},
/* 8638 */ {(10<<2)|2,{99,105,0}},
/* 8639 */ {(10<<2)|2,{99,105,0}},
/* 8640 */ {(15<<2)|3,{99,111,48}},
/* 8641 */ {(15<<2)|3,{99,111,48}},
/* 8642 */ {(15<<2)|3,{99,111,49}},
/* 8643 */ {(15<<2)|3,{99,111,49}},
/* 8644 */ {(15<<2)|3,{99,111,50}},
/* 8645 */ {(15<<2)|3,{99,111,50}},
/* 8646 */ {(15<<2)|3,{99,111,97}},
/* 8647 */ {(15<<2)|3,{99,111,97}},
/* 8648 */ {(15<<2)|3,{99,111,99}},
/* 8649 */ {(15<<2)|3,{99,111,99}},
/* 8650 */ {(15<<2)|3,{99,111,101}},
/* 8651 */ {(15<<2)|3,{99,111,101}},
/* 8652 */ {(15<<2)|3,{99,111,105}},
/* 8653 */ {(15<<2)|3,{99,111,105}},
/* 8654 */ {(15<<2)|3,{99,111,111}},
/* 8655 */ {(15<<2)|3,{99,111,111}},
/* 8656 */ {(15<<2)|3,{99,111,115}},
/* 8657 */ {(15<<2)|3,{99,111,115}},
/* 8658 */ {(15<<2)|3,{99,111,116}},
/* 8659 */ {(15<<2)|3,{99,111,116}},
/* 8660 */ {(16<<2)|3,{99,111,32}},
/* 8661 */ {(16<<2)|3,{99,111,37}},
/* 8662 */ {(16<<2)|3,{99,111,45}},
/* 8663 */ {(16<<2)|3,{99,111,46}},
/* 8664 */ {(16<<2)|3,{99,111,47}},
/* 8665 */ {(16<<2)|3,{99,111,51}},
/* 8666 */ {(16<<2)|3,{99,111,52}},
/* 8667 */ {(16<<2)|3,{99,111,53}},
/* 8668 */ {(16<<2)|3,{99,111,54}},
/* 8669 */ {(16<<2)|3,{99,111,55}},
/* 8670 */ {(16<<2)|3,{99,111,56}},
/* 8671 */ {(16<<2)|3,{99,111,57}},
/* 8672 */ {(16<<2)|3,{99,111,61}},
/* 8673 */ {(16<<2)|3,{99,111,65}},
/* 8674 */ {(16<<2)|3,{99,111,95}},
/* 8675 */ {(16<<2)|3,{99,111,98}},
/* 8676 */ {(16<<2)|3,{99,111,100}},
/* 8677 */ {(16<<2)|3,{99,111,102}},
/* 8678 */ {(16<<2)|3,{99,111,103}},
/* 8679 */ {(16<<2)|3,{99,111,104}},
/* 8680 */ {(16<<2)|3,{99,111,108}},
/* 8681 */ {(16<<2)|3,{99,111,109}},
/* 8682 */ {(16<<2)|3,{99,111,110}},
/* 8683 */ {(16<<2)|3,{99,111,112}},
/* 8684 */ {(16<<2)|3,{99,111,114}},
/* 8685 */ {(16<<2)|3,{99,111,117}},
/* 8686 */ {(10<<2)|2,{99,111,0}},
/* 8687 */ {(10<<2)|2,{99,111,0}},
/* 8688 */ {(10<<2)|2,{99,111,0}},
/* 8689 */ {(10<<2)|2,{99,111,0}},
/* 8690 */ {(10<<2)|2,{99,111,0}},
/* 8691 */ {(10<<2)|2,{99,111,0}},
/* 8692 */ {(10<<2)|2,{99,111,0}},
/* 8693 */ {(10<<2)|2,{99,111,0}},
/* 8694 */ {(10<<2)|2,{99,111,0}},
/* 8695 */ {(10<<2)|2,{99,111,0}},
/* 8696 */ {(10<<2)|2,{99,111,0}},
/* 8697 */ {(10<<2)|2,{99,111,0}},
/* 8698 */ {(10<<2)|2,{99,111,0}},
/* 8699 */ {(10<<2)|2,{99,111,0}},
/* 8700 */ {(10<<2)|2,{99,111,0}},
/* 8701 */ {(10<<2)|2,{99,111,0}},
/* 8702 */ {(10<<2)|2,{99,111,0}},
/* 8703 */ {(10<<2)|2,{99,111,0}},
/* 8704 */ {(15<<2)|3,{99,115,48}},
/* 8705 */ {(15<<2)|3,{99,115,48}},
/* 8706 */ {(15<<2)|3,{99,115,49}},
/* 8707 */ {(15<<2)|3,{99,115,49}},
/* 8708 */ {(15<<2)|3,{99,115,50}},
/* 8709 */ {(15<<2)|3,{99,115,50}},
/* 8710 */ {(15<<2)|3,{99,115,97}},
/* 8711 */ {(15<<2)|3,{99,115,97}},
/* 8712 */ {(15<<2)|3,{99,115,99}},
/* 8713 */ {(15<<2)|3,{99,115,99}},
/* 8714 */ {(15<<2)|3,{99,115,101}},
/* 8715 */ {(15<<2)|3,{99,115,101}},
/* 8716 */ {(15<<2)|3,{99,115,105}},
/* 8717 */ {(15<<2)|3,{99,115,105}},
/* 8718 */ {(15<<2)|3,{99,115,111}},
/* 8719 */ {(15<<2)|3,{99,115,111}},
/* 8720 */ {(15<<2)|3,{99,115,115}},
/* 8721 */ {(15<<2)|3,{99,115,115}},
/* 8722 */ {(15<<2)|3,{99,115,116}},
/* 8723 */ {(15<<2)|3,{99,115,116}},
/* 8724 */ {(16<<2)|3,{99,115,32}},
/* 8725 */ {(16<<2)|3,{99,115,37}},
/* 8726 */ {(16<<2)|3,{99,115,45}},
/* 8727 */ {(16<<2)|3,{99,115,46}},
/* 8728 */ {(16<<2)|3,{99,115,47}},
/* 8729 */ {(16<<2)|3,{99,115,51}},
/* 8730 */ {(16<<2)|3,{99,115,52}},
/* 8731 */ {(16<<2)|3,{99,115,53}},
/* 8732 */ {(16<<2)|3,{99,115,54}},
/* 8733 */ {(16<<2)|3,{99,115,55}},
/* 8734 */ {(16<<2)|3,{99,115,56}},
/* 8735 */ {(16<<2)|3,{99,115,57}},
/* 8736 */ {(16<<2)|3,{99,115,61}},
/* 8737 */ {(16<<2)|3,{99,115,65}},
/* 8738 */ {(16<<2)|3,{99,115,95}},
/* 8739 */ {(16<<2)|3,{99,115,98}},
/* 8740 */ {(16<<2)|3,{99,115,100}},
/* 8741 */ {(16<<2)|3,{99,115,102}},
/* 8742 */ {(16<<2)|3,{99,115,103}},
/* 8743 */ {(16<<2)|3,{99,115,104}},
/* 8744 */ {(16<<2)|3,{99,115,108}},
/* 8745 */ {(16<<2)|3,{99,115,109}},
/* 8746 */ {(16<<2)|3,{99,115,110}},
/* 8747 */ {(16<<2)|3,{99,115,112}},
/* 8748 */ {(16<<2)|3,{99,115,114}},
/* 8749 */ {(16<<2)|3,{99,115,117}},
/* 8750 */ {(10<<2)|2,{99,115,0}},
/* 8751 */ {(10<<2)|2,{99,115,0}},
/* 8752 */ {(10<<2)|2,{99,115,0}},
/* 8753 */ {(10<<2)|2,{99,115,0}},
/* 8754 */ {(10<<2)|2,{99,115,0}},
/* 8755 */ {(10<<2)|2,{99,115,0}},
/* 8756 */ {(10<<2)|2,{99,115,0}},
/* 8757 */ {(10<<2)|2,{99,115,0}},
/* 8758 */ {(10<<2)|2,{99,115,0}},
/* 8759 */ {(10<<2)|2,{99,115,0}},
/* 8760 */ {(10<<2)|2,{99,115,0}},
/* 8761 */ {(10<<2)|2,{99,115,0}},
/* 8762 */ {(10<<2)|2,{99,115,0}},
/* 8763 */ {(10<<2)|2,{99,115,0}},
/* 8764 */ {(10<<2)|2,{99,115,0}},
/* 8765 */ {(10<<2)|2,{99,115,0}},
/* 8766 */ {(10<<2)|2,{99,115,0}},
/* 8767 */ {(10<<2)|2,{99,115,0}},
/* 8768 */ {(15<<2)|3,{99,116,48}},
/* 8769 */ {(15<<2)|3,{99,116,48}},
/* 8770 */ {(15<<2)|3,{99,116,49}},
/* 8771 */ {(15<<2)|3,{99,116,49}},
/* 8772 */ {(15<<2)|3,{99,116,50}},
/* 8773 */ {(15<<2)|3,{99,116,50}},
/* 8774 */ {(15<<2)|3,{99,116,97}},
/* 8775 */ {(15<<2)|3,{99,116,97}},
/* 8776 */ {(15<<2)|3,{99,116,99}},
/* 8777 */ {(15<<2)|3,{99,116,99}},
/* 8778 */ {(15<<2)|3,{99,116,101}},
/* 8779 */ {(15<<2)|3,{99,116,101}},
/* 8780 */ {(15<<2)|3,{99,116,105}},
/* 8781 */ {(15<<2)|3,{99,116,105}},
/* 8782 */ {(15<<2)|3,{99,116,111}},
/* 8783 */ {(15<<2)|3,{99,116,111}},
/* 8784 */ {(15<<2)|3,{99,116,115}},
/* 8785 */ {(15<<2)|3,{99,116,115}},
/* 8786 */ {(15<<2)|3,{99,116,116}},
/* 8787 */ {(15<<2)|3,{99,116,116}},
/* 8788 */ {(16<<2)|3,{99,116,32}},
/* 8789 */ {(16<<2)|3,{99,116,37}},
/* 8790 */ {(16<<2)|3,{99,116,45}},
/* 8791 */ {(16<<2)|3,{99,116,46}},
/* 8792 */ {(16<<2)|3,{99,116,47}},
/* 8793 */ {(16<<2)|3,{99,116,51}},
/* 8794 */ {(16<<2)|3,{99,116,52}},
/* 8795 */ {(16<<2)|3,{99,116,53}},
/* 8796 */ {(16<<2)|3,{99,116,54}},
/* 8797 */ {(16<<2)|3,{99,116,55}},
/* 8798 */ {(16<<2)|3,{99,116,56}},
/* 8799 */ {(16<<2)|3,{99,116,57}},
/* 8800 */ {(16<<2)|3,{99,116,61}},
/* 8801 */ {(16<<2)|3,{99,116,65}},
/* 8802 */ {(16<<2)|3,{99,116,95}},
/* 8803 */ {(16<<2)|3,{99,116,98}},
/* 8804 */ {(16<<2)|3,{99,116,100}},
/* 8805 */ {(16<<2)|3,{99,116,102}},
/* 8806 */ {(16<<2)|3,{99,116,103}},
/* 8807 */ {(16<<2)|3,{99,116,104}},
/* 8808 */ {(16<<2)|3,{99,116,108}},
/* 8809 */ {(16<<2)|3,{99,116,109}},
/* 8810 */ {(16<<2)|3,{99,116,110}},
/* 8811 */ {(16<<2)|3,{99,116,112}},
/* 8812 */ {(16<<2)|3,{99,116,114}},
/* 8813 */ {(16<<2)|3,{99,116,117}},
/* 8814 */ {(10<<2)|2,{99,116,0}},
/* 8815 */ {(10<<2)|2,{99,116,0}},
/* 8816 */ {(10<<2)|2,{99,116,0}},
/* 8817 */ {(10<<2)|2,{99,116,0}},
/* 8818 */ {(10<<2)|2,{99,116,0}},
/* 8819 */ {(10<<2)|2,{99,116,0}},
/* 8820 */ {(10<<2)|2,{99,116,0}},
/* 8821 */ {(10<<2)|2,{99,116,0}},
/* 8822 */ {(10<<2)|2,{99,116,0}},
/* 8823 */ {(10<<2)|2,{99,116,0}},
/* 8824 */ {(10<<2)|2,{99,116,0}},
/* 8825 */ {(10<<2)|2,{99,116,0}},
/* 8826 */ {(10<<2)|2,{99,116,0}},
/* 8827 */ {(10<<2)|2,{99,116,0}},
/* 8828 */ {(10<<2)|2,{99,116,0}},
/* 8829 */ {(10<<2)|2,{99,116,0}},
/* 8830 */ {(10<<2)|2,{99,116,0}},
/* 8831 */ {(10<<2)|2,{99,116,0}},
/* 8832 */ {(16<<2)|3,{99,32,48}},
/* 8833 */ {(16<<2)|3,{99,32,49}},
/* 8834 */ {(16<<2)|3,{99,32,50}},
/* 8835 */ {(16<<2)|3,{99,32,97}},
/* 8836 */ {(16<<2)|3,{99,32,99}},
/* 8837 */ {(16<<2)|3,{99,32,101}},
/* 8838 */ {(16<<2)|3,{99,32,105}},
/* 8839 */ {(16<<2)|3,{99,32,111}},
/* 8840 */ {(16<<2)|3,{99,32,115}},
/* 8841 */ {(16<<2)|3,{99,32,116}},
/* 8842 */ {(11<<2)|2,{99,32,0}},
/* 8843 */ {(11<<2)|2,{99,32,0}},
/* 8844 */ {(11<<2)|2,{99,32,0}},
/* 8845 */ {(11<<2)|2,{99,32,0}},
/* 8846 */ {(11<<2)|2,{99,32,0}},
/* 8847 */ {(11<<2)|2,{99,32,0}},
/* 8848 */ {(11<<2)|2,{99,32,0}},
/* 8849 */ {(11<<2)|2,{99,32,0}},
/* 8850 */ {(11<<2)|2,{99,32,0}},
/* 8851 */ {(11<<2)|2,{99,32,0}},
/* 8852 */ {(11<<2)|2,{99,32,0}},
/* 8853 */ {(11<<2)|2,{99,32,0}},
/* 8854 */ {(11<<2)|2,{99,32,0}},
/* 8855 */ {(11<<2)|2,{99,32,0}},
/* 8856 */ {(11<<2)|2,{99,32,0}},
/* 8857 */ {(11<<2)|2,{99,32,0}},
/* 8858 */ {(11<<2)|2,{99,32,0}},
/* 8859 */ {(11<<2)|2,{99,32,0}},
/* 8860 */ {(11<<2)|2,{99,32,0}},
/* 8861 */ {(11<<2)|2,{99,32,0}},
/* 8862 */ {(11<<2)|2,{99,32,0}},
/* 8863 */ {(11<<2)|2,{99,32,0}},
/* 8864 */ {(16<<2)|3,{99,37,48}},
/* 8865 */ {(16<<2)|3,{99,37,49}},
/* 8866 */ {(16<<2)|3,{99,37,50}},
/* 8867 */ {(16<<2)|3,{99,37,97}},
/* 8868 */ {(16<<2)|3,{99,37,99}},
/* 8869 */ {(16<<2)|3,{99,37,101}},
/* 8870 */ {(16<<2)|3,{99,37,105}},
/* 8871 */ {(16<<2)|3,{99,37,111}},
/* 8872 */ {(16<<2)|3,{99,37,115}},
/* 8873 */ {(16<<2)|3,{99,37,116}},
/* 8874 */ {(11<<2)|2,{99,37,0}},
/* 8875 */ {(11<<2)|2,{99,37,0}},
/* 8876 */ {(11<<2)|2,{99,37,0}},
/* 8877 */ {(11<<2)|2,{99,37,0}},
/* 8878 */ {(11<<2)|2,{99,37,0}},
/* 8879 */ {(11<<2)|2,{99,37,0}},
/* 8880 */ {(11<<2)|2,{99,37,0}},
/* 8881 */ {(11<<2)|2,{99,37,0}},
/* 8882 */ {(11<<2)|2,{99,37,0}},
/* 8883 */ {(11<<2)|2,{99,37,0}},
/* 8884 */ {(11<<2)|2,{99,37,0}},
/* 8885 */ {(11<<2)|2,{99,37,0}},
/* 8886 */ {(11<<2)|2,{99,37,0}},
/* 8887 */ {(11<<2)|2,{99,37,0}},
/* 8888 */ {(11<<2)|2,{99,37,0}},
/* 8889 */ {(11<<2)|2,{99,37,0}},
/* 8890 */ {(11<<2)|2,{99,37,0}},
/* 8891 */ {(11<<2)|2,{99,37,0}},
/* 8892 */ {(11<<2)|2,{99,37,0}},
/* 8893 */ {(11<<2)|2,{99,37,0}},
/* 8894 */ {(11<<2)|2,{99,37,0}},
/* 8895 */ {(11<<2)|2,{99,37,0}},
/* 8896 */ {(16<<2)|3,{99,45,48}},
/* 8897 */ {(16<<2)|3,{99,45,49}},
/* 8898 */ {(16<<2)|3,{99,45,50}},
/* 8899 */ {(16<<2)|3,{99,45,97}},
/* 8900 */ {(16<<2)|3,{99,45,99}},
/* 8901 */ {(16<<2)|3,{99,45,101}},
/* 8902 */ {(16<<2)|3,{99,45,105}},
/* 8903 */ {(16<<2)|3,{99,45,111}},
/* 8904 */ {(16<<2)|3,{99,45,115}},
/* 8905 */ {(16<<2)|3,{99,45,116}},
/* 8906 */ {(11<<2)|2,{99,45,0}},
/* 8907 */ {(11<<2)|2,{99,45,0}},
/* 8908 */ {(11<<2)|2,{99,45,0}},
/* 8909 */ {(11<<2)|2,{99,45,0}},
/* 8910 */ {(11<<2)|2,{99,45,0}},
/* 8911 */ {(11<<2)|2,{99,45,0}},
/* 8912 */ {(11<<2)|2,{99,45,0}},
/* 8913 */ {(11<<2)|2,{99,45,0}},
/* 8914 */ {(11<<2)|2,{99,45,0}},
/* 8915 */ {(11<<2)|2,{99,45,0}},
/* 8916 */ {(11<<2)|2,{99,45,0}},
/* 8917 */ {(11<<2)|2,{99,45,0}},
/* 8918 */ {(11<<2)|2,{99,45,0}},
/* 8919 */ {(11<<2)|2,{99,45,0}},
/* 8920 */ {(11<<2)|2,{99,45,0}},
/* 8921 */ {(11<<2)|2,{99,45,0}},
/* 8922 */ {(11<<2)|2,{99,45,0}},
/* 8923 */ {(11<<2)|2,{99,45,0}},
/* 8924 */ {(11<<2)|2,{99,45,0}},
/* 8925 */ {(11<<2)|2,{99,45,0}},
/* 8926 */ {(11<<2)|2,{99,45,0}},
/* 8927 */ {(11<<2)|2,{99,45,0}},
/* 8928 */ {(16<<2)|3,{99,46,48}},
/* 8929 */ {(16<<2)|3,{99,46,49}},
/* 8930 */ {(16<<2)|3,{99,46,50}},
/* 8931 */ {(16<<2)|3,{99,46,97}},
/* 8932 */ {(16<<2)|3,{99,46,99}},
/* 8933 */ {(16<<2)|3,{99,46,101}},
/* 8934 */ {(16<<2)|3,{99,46,105}},
/* 8935 */ {(16<<2)|3,{99,46,111}},
/* 8936 */ {(16<<2)|3,{99,46,115}},
/* 8937 */ {(16<<2)|3,{99,46,116}},
/* 8938 */ {(11<<2)|2,{99,46,0}},
/* 8939 */ {(11<<2)|2,{99,46,0}},
/* 8940 */ {(11<<2)|2,{99,46,0}},
/* 8941 */ {(11<<2)|2,{99,46,0}},
/* 8942 */ {(11<<2)|2,{99,46,0}},
/* 8943 */ {(11<<2)|2,{99,46,0}},
/* 8944 */ {(11<<2)|2,{99,46,0}},
/* 8945 */ {(11<<2)|2,{99,46,0}},
/* 8946 */ {(11<<2)|2,{99,46,0}},
/* 8947 */ {(11<<2)|2,{99,46,0}},
/* 8948 */ {(11<<2)|2,{99,46,0}},
/* 8949 */ {(11<<2)|2,{99,46,0}},
/* 8950 */ {(11<<2)|2,{99,46,0}},
/* 8951 */ {(11<<2)|2,{99,46,0}},
/* 8952 */ {(11<<2)|2,{99,46,0}},
/* 8953 */ {(11<<2)|2,{99,46,0}},
/* 8954 */ {(11<<2)|2,{99,46,0}},
/* 8955 */ {(11<<2)|2,{99,46,0}},
/* 8956 */ {(11<<2)|2,{99,46,0}},
/* 8957 */ {(11<<2)|2,{99,46,0}},
/* 8958 */ {(11<<2)|2,{99,46,0}},
/* 8959 */ {(11<<2)|2,{99,46,0}},
/* 8960 */ {(16<<2)|3,{99,47,48}},
/* 8961 */ {(16<<2)|3,{99,47,49}},
/* 8962 */ {(16<<2)|3,{99,47,50}},
/* 8963 */ {(16<<2)|3,{99,47,97}},
/* 8964 */ {(16<<2)|3,{99,47,99}},
/* 8965 */ {(16<<2)|3,{99,47,101}},
/* 8966 */ {(16<<2)|3,{99,47,105}},
/* 8967 */ {(16<<2)|3,{99,47,111}},
/* 8968 */ {(16<<2)|3,{99,47,115}},
/* 8969 */ {(16<<2)|3,{99,47,116}},
/* 8970 */ {(11<<2)|2,{99,47,0}},
/* 8971 */ {(11<<2)|2,{99,47,0}},
/* 8972 */ {(11<<2)|2,{99,47,0}},
/* 8973 */ {(11<<2)|2,{99,47,0}},
/* 8974 */ {(11<<2)|2,{99,47,0}},
/* 8975 */ {(11<<2)|2,{99,47,0}},
/* 8976 */ {(11<<2)|2,{99,47,0}},
/* 8977 */ {(11<<2)|2,{99,47,0}},
/* 8978 */ {(11<<2)|2,{99,47,0}},
/* 8979 */ {(11<<2)|2,{99,47,0}},
/* 8980 */ {(11<<2)|2,{99,47,0}},
/* 8981 */ {(11<<2)|2,{99,47,0}},
/* 8982 */ {(11<<2)|2,{99,47,0}},
/* 8983 */ {(11<<2)|2,{99,47,0}},
/* 8984 */ {(11<<2)|2,{99,47,0}},
/* 8985 */ {(11<<2)|2,{99,47,0}},
/* 8986 */ {(11<<2)|2,{99,47,0}},
/* 8987 */ {(11<<2)|2,{99,47,0}},
/* 8988 */ {(11<<2)|2,{99,47,0}},
/* 8989 */ {(11<<2)|2,{99,47,0}},
/* 8990 */ {(11<<2)|2,{99,47,0}},
/* 8991 */ {(11<<2)|2,{99,47,0}},
/* 8992 */ {(16<<2)|3,{99,51,48}},
/* 8993 */ {(16<<2)|3,{99,51,49}},
/* 8994 */ {(16<<2)|3,{99,51,50}},
/* 8995 */ {(16<<2)|3,{99,51,97}},
/* 8996 */ {(16<<2)|3,{99,51,99}},
/* 8997 */ {(16<<2)|3,{99,51,101}},
/* 8998 */ {(16<<2)|3,{99,51,105}},
/* 8999 */ {(16<<2)|3,{99,51,111}},
/* 9000 */ {(16<<2)|3,{99,51,115}},
/* 9001 */ {(16<<2)|3,{99,51,116}},
/* 9002 */ {(11<<2)|2,{99,51,0}},
/* 9003 */ {(11<<2)|2,{99,51,0}},
/* 9004 */ {(11<<2)|2,{99,51,0}},
/* 9005 */ {(11<<2)|2,{99,51,0}},
/* 9006 */ {(11<<2)|2,{99,51,0}},
/* 9007 */ {(11<<2)|2,{99,51,0}},
/* 9008 */ {(11<<2)|2,{99,51,0}},
/* 9009 */ {(11<<2)|2,{99,51,0}},
/* 9010 */ {(11<<2)|2,{99,51,0}},
/* 9011 */ {(11<<2)|2,{99,51,0}},
/* 9012 */ {(11<<2)|2,{99,51,0}},
/* 9013 */ {(11<<2)|2,{99,51,0}},
/* 9014 */ {(11<<2)|2,{99,51,0}},
/* 9015 */ {(11<<2)|2,{99,51,0}},
/* 9016 */ {(11<<2)|2,{99,51,0}},
/* 9017 */ {(11<<2)|2,{99,51,0}},
/* 9018 */ {(11<<2)|2,{99,51,0}},
/* 9019 */ {(11<<2)|2,{99,51,0}},
/* 9020 */ {(11<<2)|2,{99,51,0}},
/* 9021 */ {(11<<2)|2,{99,51,0}},
/* 9022 */ {(11<<2)|2,{99,51,0}},
/* 9023 */ {(11<<2)|2,{99,51,0}},
/* 9024 */ {(16<<2)|3,{99,52,48}},
/* 9025 */ {(16<<2)|3,{99,52,49}},
/* 9026 */ {(16<<2)|3,{99,52,50}},
/* 9027 */ {(16<<2)|3,{99,52,97}},
/* 9028 */ {(16<<2)|3,{99,52,99}},
/* 9029 */ {(16<<2)|3,{99,52,101}},
/* 9030 */ {(16<<2)|3,{99,52,105}},
/* 9031 */ {(16<<2)|3,{99,52,111}},
/* 9032 */ {(16<<2)|3,{99,52,115}},
/* 9033 */ {(16<<2)|3,{99,52,116}},
/* 9034 */ {(11<<2)|2,{99,52,0}},
/* 9035 */ {(11<<2)|2,{99,52,0}},
/* 9036 */ {(11<<2)|2,{99,52,0}},
/* 9037 */ {(11<<2)|2,{99,52,0}},
/* 9038 */ {(11<<2)|2,{99,52,0}},
/* 9039 */ {(11<<2)|2,{99,52,0}},
/* 9040 */ {(11<<2)|2,{99,52,0}},
/* 9041 */ {(11<<2)|2,{99,52,0}},
/* 9042 */ {(11<<2)|2,{99,52,0}},
/* 9043 */ {(11<<2)|2,{99,52,0}},
/* 9044 */ {(11<<2)|2,{99,52,0}},
/* 9045 */ {(11<<2)|2,{99,52,0}},
/* 9046 */ {(11<<2)|2,{99,52,0}},
/* 9047 */ {(11<<2)|2,{99,52,0}},
/* 9048 */ {(11<<2)|2,{99,52,0}},
/* 9049 */ {(11<<2)|2,{99,52,0}},
/* 9050 */ {(11<<2)|2,{99,52,0}},
/* 9051 */ {(11<<2)|2,{99,52,0}},
/* 9052 */ {(11<<2)|2,{99,52,0}},
/* 9053 */ {(11<<2)|2,{99,52,0}},
/* 9054 */ {(11<<2)|2,{99,52,0}},
/* 9055 */ {(11<<2)|2,{99,52,0}},
/* 9056 */ {(16<<2)|3,{99,53,48}},
/* 9057 */ {(16<<2)|3,{99,53,49}},
/* 9058 */ {(16<<2)|3,{99,53,50}},
/* 9059 */ {(16<<2)|3,{99,53,97}},
/* 9060 */ {(16<<2)|3,{99,53,99}},
/* 9061 */ {(16<<2)|3,{99,53,101}},
/* 9062 */ {(16<<2)|3,{99,53,105}},
/* 9063 */ {(16<<2)|3,{99,53,111}},
/* 9064 */ {(16<<2)|3,{99,53,115}},
/* 9065 */ {(16<<2)|3,{99,53,116}},
/* 9066 */ {(11<<2)|2,{99,53,0}},
/* 9067 */ {(11<<2)|2,{99,53,0}},
/* 9068 */ {(11<<2)|2,{99,53,0}},
/* 9069 */ {(11<<2)|2,{99,53,0}},
/* 9070 */ {(11<<2)|2,{99,53,0}},
/* 9071 */ {(11<<2)|2,{99,53,0}},
/* 9072 */ {(11<<2)|2,{99,53,0}},
/* 9073 */ {(11<<2)|2,{99,53,0}},
/* 9074 */ {(11<<2)|2,{99,53,0}},
/* 9075 */ {(11<<2)|2,{99,53,0}},
/* 9076 */ {(11<<2)|2,{99,53,0}},
/* 9077 */ {(11<<2)|2,{99,53,0}},
/* 9078 */ {(11<<2)|2,{99,53,0}},
/* 9079 */ {(11<<2)|2,{99,53,0}},
/* 9080 */ {(11<<2)|2,{99,53,0}},
/* 9081 */ {(11<<2)|2,{99,53,0}},
/* 9082 */ {(11<<2)|2,{99,53,0}},
/* 9083 */ {(11<<2)|2,{99,53,0}},
/* 9084 */ {(11<<2)|2,{99,53,0}},
/* 9085 */ {(11<<2)|2,{99,53,0}},
/* 9086 */ {(11<<2)|2,{99,53,0}},
/* 9087 */ {(11<<2)|2,{99,53,0}},
/* 9088 */ {(16<<2)|3,{99,54,48}},
/* 9089 */ {(16<<2)|3,{99,54,49}},
/* 9090 */ {(16<<2)|3,{99,54,50}},
/* 9091 */ {(16<<2)|3,{99,54,97}},
/* 9092 */ {(16<<2)|3,{99,54,99}},
/* 9093 */ {(16<<2)|3,{99,54,101}},
/* 9094 */ {(16<<2)|3,{99,54,105}},
/* 9095 */ {(16<<2)|3,{99,54,111}},
/* 9096 */ {(16<<2)|3,{99,54,115}},
/* 9097 */ {(16<<2)|3,{99,54,116}},
/* 9098 */ {(11<<2)|2,{99,54,0}},
/* 9099 */ {(11<<2)|2,{99,54,0}},
/* 9100 */ {(11<<2)|2,{99,54,0}},
/* 9101 */ {(11<<2)|2,{99,54,0}},
/* 9102 */ {(11<<2)|2,{99,54,0}},
/* 9103 */ {(11<<2)|2,{99,54,0}},
/* 9104 */ {(11<<2)|2,{99,54,0}},
/* 9105 */ {(11<<2)|2,{99,54,0}},
/* 9106 */ {(11<<2)|2,{99,54,0}},
/* 9107 */ {(11<<2)|2,{99,54,0}},
/* 9108 */ {(11<<2)|2,{99,54,0}},
/* 9109 */ {(11<<2)|2,{99,54,0}},
/* 9110 */ {(11<<2)|2,{99,54,0}},
/* 9111 */ {(11<<2)|2,{99,54,0}},
/* 9112 */ {(11<<2)|2,{99,54,0}},
/* 9113 */ {(11<<2)|2,{99,54,0}},
/* 9114 */ {(11<<2)|2,{99,54,0}},
/* 9115 */ {(11<<2)|2,{99,54,0}},
/* 9116 */ {(11<<2)|2,{99,54,0}},
/* 9117 */ {(11<<2)|2,{99,54,0}},
/* 9118 */ {(11<<2)|2,{99,54,0}},
/* 9119 */ {(11<<2)|2,{99,54,0}},
/* 9120 */ {(16<<2)|3,{99,55,48}},
/* 9121 */ {(16<<2)|3,{99,55,49}},
/* 9122 */ {(16<<2)|3,{99,55,50}},
/* 9123 */ {(16<<2)|3,{99,55,97}},
/* 9124 */ {(16<<2)|3,{99,55,99}},
/* 9125 */ {(16<<2)|3,{99,55,101}},
/* 9126 */ {(16<<2)|3,{99,55,105}},
/* 9127 */ {(16<<2)|3,{99,55,111}},
/* 9128 */ {(16<<2)|3,{99,55,115}},
/* 9129 */ {(16<<2)|3,{99,55,116}},
/* 9130 */ {(11<<2)|2,{99,55,0}},
/* 9131 */ {(11<<2)|2,{99,55,0}},
/* 9132 */ {(11<<2)|2,{99,55,0}},
/* 9133 */ {(11<<2)|2,{99,55,0}},
/* 9134 */ {(11<<2)|2,{99,55,0}},
/* 9135 */ {(11<<2)|2,{99,55,0}},
/* 9136 */ {(11<<2)|2,{99,55,0}},
/* 9137 */ {(11<<2)|2,{99,55,0}},
/* 9138 */ {(11<<2)|2,{99,55,0}},
/* 9139 */ {(11<<2)|2,{99,55,0}},
/* 9140 */ {(11<<2)|2,{99,55,0}},
/* 9141 */ {(11<<2)|2,{99,55,0}},
/* 9142 */ {(11<<2)|2,{99,55,0}},
/* 9143 */ {(11<<2)|2,{99,55,0}},
/* 9144 */ {(11<<2)|2,{99,55,0}},
/* 9145 */ {(11<<2)|2,{99,55,0}},
/* 9146 */ {(11<<2)|2,{99,55,0}},
/* 9147 */ {(11<<2)|2,{99,55,0}},
/* 9148 */ {(11<<2)|2,{99,55,0}},
/* 9149 */ {(11<<2)|2,{99,55,0}},
/* 9150 */ {(11<<2)|2,{99,55,0}},
/* 9151 */ {(11<<2)|2,{99,55,0}},
/* 9152 */ {(16<<2)|3,{99,56,48}},
/* 9153 */ {(16<<2)|3,{99,56,49}},
/* 9154 */ {(16<<2)|3,{99,56,50}},
/* 9155 */ {(16<<2)|3,{99,56,97}},
/* 9156 */ {(16<<2)|3,{99,56,99}},
/* 9157 */ {(16<<2)|3,{99,56,101}},
/* 9158 */ {(16<<2)|3,{99,56,105}},
/* 9159 */ {(16<<2)|3,{99,56,111}},
/* 9160 */ {(16<<2)|3,{99,56,115}},
/* 9161 */ {(16<<2)|3,{99,56,116}},
/* 9162 */ {(11<<2)|2,{99,56,0}},
/* 9163 */ {(11<<2)|2,{99,56,0}},
/* 9164 */ {(11<<2)|2,{99,56,0}},
/* 9165 */ {(11<<2)|2,{99,56,0}},
/* 9166 */ {(11<<2)|2,{99,56,0}},
/* 9167 */ {(11<<2)|2,{99,56,0}},
/* 9168 */ {(11<<2)|2,{99,56,0}},
/* 9169 */ {(11<<2)|2,{99,56,0}},
/* 9170 */ {(11<<2)|2,{99,56,0}},
/* 9171 */ {(11<<2)|2,{99,56,0}},
/* 9172 */ {(11<<2)|2,{99,56,0}},
/* 9173 */ {(11<<2)|2,{99,56,0}},
/* 9174 */ {(11<<2)|2,{99,56,0}},
/* 9175 */ {(11<<2)|2,{99,56,0}},
/* 9176 */ {(11<<2)|2,{99,56,0}},
/* 9177 */ {(11<<2)|2,{99,56,0}},
/* 9178 */ {(11<<2)|2,{99,56,0}},
/* 9179 */ {(11<<2)|2,{99,56,0}},
/* 9180 */ {(11<<2)|2,{99,56,0}},
/* 9181 */ {(11<<2)|2,{99,56,0}},
/* 9182 */ {(11<<2)|2,{99,56,0}},
/* 9183 */ {(11<<2)|2,{99,56,0}},
/* 9184 */ {(16<<2)|3,{99,57,48}},
/* 9185 */ {(16<<2)|3,{99,57,49}},
/* 9186 */ {(16<<2)|3,{99,57,50}},
/* 9187 */ {(16<<2)|3,{99,57,97}},
/* 9188 */ {(16<<2)|3,{99,57,99}},
/* 9189 */ {(16<<2)|3,{99,57,101}},
/* 9190 */ {(16<<2)|3,{99,57,105}},
/* 9191 */ {(16<<2)|3,{99,57,111}},
/* 9192 */ {(16<<2)|3,{99,57,115}},
/* 9193 */ {(16<<2)|3,{99,57,116}},
/* 9194 */ {(11<<2)|2,{99,57,0}},
/* 9195 */ {(11<<2)|2,{99,57,0}},
/* 9196 */ {(11<<2)|2,{99,57,0}},
/* 9197 */ {(11<<2)|2,{99,57,0}},
/* 9198 */ {(11<<2)|2,{99,57,0}},
/* 9199 */ {(11<<2)|2,{99,57,0}},
/* 9200 */ {(11<<2)|2,{99,57,0}},
/* 9201 */ {(11<<2)|2,{99,57,0}},
/* 9202 */ {(11<<2)|2,{99,57,0}},
/* 9203 */ {(11<<2)|2,{99,57,0}},
/* 9204 */ {(11<<2)|2,{99,57,0}},
/* 9205 */ {(11<<2)|2,{99,57,0}},
/* 9206 */ {(11<<2)|2,{99,57,0}},
/* 9207 */ {(11<<2)|2,{99,57,0}},
/* 9208 */ {(11<<2)|2,{99,57,0}},
/* 9209 */ {(11<<2)|2,{99,57,0}},
/* 9210 */ {(11<<2)|2,{99,57,0}},
/* 9211 */ {(11<<2)|2,{99,57,0}},
/* 9212 */ {(11<<2)|2,{99,57,0}},
/* 9213 */ {(11<<2)|2,{99,57,0}},
/* 9214 */ {(11<<2)|2,{99,57,0}},
/* 9215 */ {(11<<2)|2,{99,57,0}},
/* 9216 */ {(16<<2)|3,{99,61,48}},
/* 9217 */ {(16<<2)|3,{99,61,49}},
/* 9218 */ {(16<<2)|3,{99,61,50}},
/* 9219 */ {(16<<2)|3,{99,61,97}},
/* 9220 */ {(16<<2)|3,{99,61,99}},
/* 9221 */ {(16<<2)|3,{99,61,101}},
/* 9222 */ {(16<<2)|3,{99,61,105}},
/* 9223 */ {(16<<2)|3,{99,61,111}},
/* 9224 */ {(16<<2)|3,{99,61,115}},
/* 9225 */ {(16<<2)|3,{99,61,116}},
/* 9226 */ {(11<<2)|2,{99,61,0}},
/* 9227 */ {(11<<2)|2,{99,61,0}},
/* 9228 */ {(11<<2)|2,{99,61,0}},
/* 9229 */ {(11<<2)|2,{99,61,0}},
/* 9230 */ {(11<<2)|2,{99,61,0}},
/* 9231 */ {(11<<2)|2,{99,61,0}},
/* 9232 */ {(11<<2)|2,{99,61,0}},
/* 9233 */ {(11<<2)|2,{99,61,0}},
/* 9234 */ {(11<<2)|2,{99,61,0}},
/* 9235 */ {(11<<2)|2,{99,61,0}},
/* 9236 */ {(11<<2)|2,{99,61,0}},
/* 9237 */ {(11<<2)|2,{99,61,0}},
/* 9238 */ {(11<<2)|2,{99,61,0}},
/* 9239 */ {(11<<2)|2,{99,61,0}},
/* 9240 */ {(11<<2)|2,{99,61,0}},
/* 9241 */ {(11<<2)|2,{99,61,0}},
/* 9242 */ {(11<<2)|2,{99,61,0}},
/* 9243 */ {(11<<2)|2,{99,61,0}},
/* 9244 */ {(11<<2)|2,{99,61,0}},
/* 9245 */ {(11<<2)|2,{99,61,0}},
/* 9246 */ {(11<<2)|2,{99,61,0}},
/* 9247 */ {(11<<2)|2,{99,61,0}},
/* 9248 */ {(16<<2)|3,{99,65,48}},
/* 9249 */ {(16<<2)|3,{99,65,49}},
/* 9250 */ {(16<<2)|3,{99,65,50}},
/* 9251 */ {(16<<2)|3,{99,65,97}},
/* 9252 */ {(16<<2)|3,{99,65,99}},
/* 9253 */ {(16<<2)|3,{99,65,101}},
/* 9254 */ {(16<<2)|3,{99,65,105}},
/* 9255 */ {(16<<2)|3,{99,65,111}},
/* 9256 */ {(16<<2)|3,{99,65,115}},
/* 9257 */ {(16<<2)|3,{99,65,116}},
/* 9258 */ {(11<<2)|2,{99,65,0}},
/* 9259 */ {(11<<2)|2,{99,65,0}},
/* 9260 */ {(11<<2)|2,{99,65,0}},
/* 9261 */ {(11<<2)|2,{99,65,0}},
/* 9262 */ {(11<<2)|2,{99,65,0}},
/* 9263 */ {(11<<2)|2,{99,65,0}},
/* 9264 */ {(11<<2)|2,{99,65,0}},
/* 9265 */ {(11<<2)|2,{99,65,0}},
/* 9266 */ {(11<<2)|2,{99,65,0}},
/* 9267 */ {(11<<2)|2,{99,65,0}},
/* 9268 */ {(11<<2)|2,{99,65,0}},
/* 9269 */ {(11<<2)|2,{99,65,0}},
/* 9270 */ {(11<<2)|2,{99,65,0}},
/* 9271 */ {(11<<2)|2,{99,65,0}},
/* 9272 */ {(11<<2)|2,{99,65,0}},
/* 9273 */ {(11<<2)|2,{99,65,0}},
/* 9274 */ {(11<<2)|2,{99,65,0}},
/* 9275 */ {(11<<2)|2,{99,65,0}},
/* 9276 */ {(11<<2)|2,{99,65,0}},
/* 9277 */ {(11<<2)|2,{99,65,0}},
/* 9278 */ {(11<<2)|2,{99,65,0}},
/* 9279 */ {(11<<2)|2,{99,65,0}},
/* 9280 */ {(16<<2)|3,{99,95,48}},
/* 9281 */ {(16<<2)|3,{99,95,49}},
/* 9282 */ {(16<<2)|3,{99,95,50}},
/* 9283 */ {(16<<2)|3,{99,95,97}},
/* 9284 */ {(16<<2)|3,{99,95,99}},
/* 9285 */ {(16<<2)|3,{99,95,101}},
/* 9286 */ {(16<<2)|3,{99,95,105}},
/* 9287 */ {(16<<2)|3,{99,95,111}},
/* 9288 */ {(16<<2)|3,{99,95,115}},
/* 9289 */ {(16<<2)|3,{99,95,116}},
/* 9290 */ {(11<<2)|2,{99,95,0}},
/* 9291 */ {(11<<2)|2,{99,95,0}},
/* 9292 */ {(11<<2)|2,{99,95,0}},
/* 9293 */ {(11<<2)|2,{99,95,0}},
/* 9294 */ {(11<<2)|2,{99,95,0}},
/* 9295 */ {(11<<2)|2,{99,95,0}},
/* 9296 */ {(11<<2)|2,{99,95,0}},
/* 9297 */ {(11<<2)|2,{99,95,0}},
/* 9298 */ {(11<<2)|2,{99,95,0}},
/* 9299 */ {(11<<2)|2,{99,95,0}},
/* 9300 */ {(11<<2)|2,{99,95,0}},
/* 9301 */ {(11<<2)|2,{99,95,0}},
/* 9302 */ {(11<<2)|2,{99,95,0}},
/* 9303 */ {(11<<2)|2,{99,95,0}},
/* 9304 */ {(11<<2)|2,{99,95,0}},
/* 9305 */ {(11<<2)|2,{99,95,0}},
/* 9306 */ {(11<<2)|2,{99,95,0}},
/* 9307 */ {(11<<2)|2,{99,95,0}},
/* 9308 */ {(11<<2)|2,{99,95,0}},
/* 9309 */ {(11<<2)|2,{99,95,0}},
/* 9310 */ {(11<<2)|2,{99,95,0}},
/* 9311 */ {(11<<2)|2,{99,95,0}},
/* 9312 */ {(16<<2)|3,{99,98,48}},
/* 9313 */ {(16<<2)|3,{99,98,49}},
/* 9314 */ {(16<<2)|3,{99,98,50}},
/* 9315 */ {(16<<2)|3,{99,98,97}},
/* 9316 */ {(16<<2)|3,{99,98,99}},
/* 9317 */ {(16<<2)|3,{99,98,101}},
/* 9318 */ {(16<<2)|3,{99,98,105}},
/* 9319 */ {(16<<2)|3,{99,98,111}},
/* 9320 */ {(16<<2)|3,{99,98,115}},
/* 9321 */ {(16<<2)|3,{99,98,116}},
/* 9322 */ {(11<<2)|2,{99,98,0}},
/* 9323 */ {(11<<2)|2,{99,98,0}},
/* 9324 */ {(11<<2)|2,{99,98,0}},
/* 9325 */ {(11<<2)|2,{99,98,0}},
/* 9326 */ {(11<<2)|2,{99,98,0}},
/* 9327 */ {(11<<2)|2,{99,98,0}},
/* 9328 */ {(11<<2)|2,{99,98,0}},
/* 9329 */ {(11<<2)|2,{99,98,0}},
/* 9330 */ {(11<<2)|2,{99,98,0}},
/* 9331 */ {(11<<2)|2,{99,98,0}},
/* 9332 */ {(11<<2)|2,{99,98,0}},
/* 9333 */ {(11<<2)|2,{99,98,0}},
/* 9334 */ {(11<<2)|2,{99,98,0}},
/* 9335 */ {(11<<2)|2,{99,98,0}},
/* 9336 */ {(11<<2)|2,{99,98,0}},
/* 9337 */ {(11<<2)|2,{99,98,0}},
/* 9338 */ {(11<<2)|2,{99,98,0}},
/* 9339 */ {(11<<2)|2,{99,98,0}},
/* 9340 */ {(11<<2)|2,{99,98,0}},
/* 9341 */ {(11<<2)|2,{99,98,0}},
/* 9342 */ {(11<<2)|2,{99,98,0}},
/* 9343 */ {(11<<2)|2,{99,98,0}},
/* 9344 */ {(16<<2)|3,{99,100,48}},
/* 9345 */ {(16<<2)|3,{99,100,49}},
/* 9346 */ {(16<<2)|3,{99,100,50}},
/* 9347 */ {(16<<2)|3,{99,100,97}},
/* 9348 */ {(16<<2)|3,{99,100,99}},
/* 9349 */ {(16<<2)|3,{99,100,101}},
/* 9350 */ {(16<<2)|3,{99,100,105}},
/* 9351 */ {(16<<2)|3,{99,100,111}},
/* 9352 */ {(16<<2)|3,{99,100,115}},
/* 9353 */ {(16<<2)|3,{99,100,116}},
/* 9354 */ {(11<<2)|2,{99,100,0}},
/* 9355 */ {(11<<2)|2,{99,100,0}},
/* 9356 */ {(11<<2)|2,{99,100,0}},
/* 9357 */ {(11<<2)|2,{99,100,0}},
/* 9358 */ {(11<<2)|2,{99,100,0}},
/* 9359 */ {(11<<2)|2,{99,100,0}},
/* 9360 */ {(11<<2)|2,{99,100,0}},
/* 9361 */ {(11<<2)|2,{99,100,0}},
/* 9362 */ {(11<<2)|2,{99,100,0}},
/* 9363 */ {(11<<2)|2,{99,100,0}},
/* 9364 */ {(11<<2)|2,{99,100,0}},
/* 9365 */ {(11<<2)|2,{99,100,0}},
/* 9366 */ {(11<<2)|2,{99,100,0}},
/* 9367 */ {(11<<2)|2,{99,100,0}},
/* 9368 */ {(11<<2)|2,{99,100,0}},
/* 9369 */ {(11<<2)|2,{99,100,0}},
/* 9370 */ {(11<<2)|2,{99,100,0}},
/* 9371 */ {(11<<2)|2,{99,100,0}},
/* 9372 */ {(11<<2)|2,{99,100,0}},
/* 9373 */ {(11<<2)|2,{99,100,0}},
/* 9374 */ {(11<<2)|2,{99,100,0}},
/* 9375 */ {(11<<2)|2,{99,100,0}},
/* 9376 */ {(16<<2)|3,{99,102,48}},
/* 9377 */ {(16<<2)|3,{99,102,49}},
/* 9378 */ {(16<<2)|3,{99,102,50}},
/* 9379 */ {(16<<2)|3,{99,102,97}},
/* 9380 */ {(16<<2)|3,{99,102,99}},
/* 9381 */ {(16<<2)|3,{99,102,101}},
/* 9382 */ {(16<<2)|3,{99,102,105}},
/* 9383 */ {(16<<2)|3,{99,102,111}},
/* 9384 */ {(16<<2)|3,{99,102,115}},
/* 9385 */ {(16<<2)|3,{99,102,116}},
/* 9386 */ {(11<<2)|2,{99,102,0}},
/* 9387 */ {(11<<2)|2,{99,102,0}},
/* 9388 */ {(11<<2)|2,{99,102,0}},
/* 9389 */ {(11<<2)|2,{99,102,0}},
/* 9390 */ {(11<<2)|2,{99,102,0}},
/* 9391 */ {(11<<2)|2,{99,102,0}},
/* 9392 */ {(11<<2)|2,{99,102,0}},
/* 9393 */ {(11<<2)|2,{99,102,0}},
/* 9394 */ {(11<<2)|2,{99,102,0}},
/* 9395 */ {(11<<2)|2,{99,102,0}},
/* 9396 */ {(11<<2)|2,{99,102,0}},
/* 9397 */ {(11<<2)|2,{99,102,0}},
/* 9398 */ {(11<<2)|2,{99,102,0}},
/* 9399 */ {(11<<2)|2,{99,102,0}},
/* 9400 */ {(11<<2)|2,{99,102,0}},
/* 9401 */ {(11<<2)|2,{99,102,0}},
/* 9402 */ {(11<<2)|2,{99,102,0}},
/* 9403 */ {(11<<2)|2,{99,102,0}},
/* 9404 */ {(11<<2)|2,{99,102,0}},
/* 9405 */ {(11<<2)|2,{99,102,0}},
/* 9406 */ {(11<<2)|2,{99,102,0}},
/* 9407 */ {(11<<2)|2,{99,102,0}},
/* 9408 */ {(16<<2)|3,{99,103,48}},
/* 9409 */ {(16<<2)|3,{99,103,49}},
/* 9410 */ {(16<<2)|3,{99,103,50}},
/* 9411 */ {(16<<2)|3,{99,103,97}},
/* 9412 */ {(16<<2)|3,{99,103,99}},
/* 9413 */ {(16<<2)|3,{99,103,101}},
/* 9414 */ {(16<<2)|3,{99,103,105}},
/* 9415 */ {(16<<2)|3,{99,103,111}},
/* 9416 */ {(16<<2)|3,{99,103,115}},
/* 9417 */ {(16<<2)|3,{99,103,116}},
/* 9418 */ {(11<<2)|2,{99,103,0}},
/* 9419 */ {(11<<2)|2,{99,103,0}},
/* 9420 */ {(11<<2)|2,{99,103,0}},
/* 9421 */ {(11<<2)|2,{99,103,0}},
/* 9422 */ {(11<<2)|2,{99,103,0}},
/* 9423 */ {(11<<2)|2,{99,103,0}},
/* 9424 */ {(11<<2)|2,{99,103,0}},
/* 9425 */ {(11<<2)|2,{99,103,0}},
/* 9426 */ {(11<<2)|2,{99,103,0}},
/* 9427 */ {(11<<2)|2,{99,103,0}},
/* 9428 */ {(11<<2)|2,{99,103,0}},
/* 9429 */ {(11<<2)|2,{99,103,0}},
/* 9430 */ {(11<<2)|2,{99,103,0}},
/* 9431 */ {(11<<2)|2,{99,103,0}},
/* 9432 */ {(11<<2)|2,{99,103,0}},
/* 9433 */ {(11<<2)|2,{99,103,0}},
/* 9434 */ {(11<<2)|2,{99,103,0}},
/* 9435 */ {(11<<2)|2,{99,103,0}},
/* 9436 */ {(11<<2)|2,{99,103,0}},
/* 9437 */ {(11<<2)|2,{99,103,0}},
/* 9438 */ {(11<<2)|2,{99,103,0}},
/* 9439 */ {(11<<2)|2,{99,103,0}},
/* 9440 */ {(16<<2)|3,{99,104,48}},
/* 9441 */ {(16<<2)|3,{99,104,49}},
/* 9442 */ {(16<<2)|3,{99,104,50}},
/* 9443 */ {(16<<2)|3,{99,104,97}},
/* 9444 */ {(16<<2)|3,{99,104,99}},
/* 9445 */ {(16<<2)|3,{99,104,101}},
/* 9446 */ {(16<<2)|3,{99,104,105}},
/* 9447 */ {(16<<2)|3,{99,104,111}},
/* 9448 */ {(16<<2)|3,{99,104,115}},
/* 9449 */ {(16<<2)|3,{99,104,116}},
/* 9450 */ {(11<<2)|2,{99,104,0}},
/* 9451 */ {(11<<2)|2,{99,104,0}},
/* 9452 */ {(11<<2)|2,{99,104,0}},
/* 9453 */ {(11<<2)|2,{99,104,0}},
/* 9454 */ {(11<<2)|2,{99,104,0}},
/* 9455 */ {(11<<2)|2,{99,104,0}},
/* 9456 */ {(11<<2)|2,{99,104,0}},
/* 9457 */ {(11<<2)|2,{99,104,0}},
/* 9458 */ {(11<<2)|2,{99,104,0}},
/* 9459 */ {(11<<2)|2,{99,104,0}},
/* 9460 */ {(11<<2)|2,{99,104,0}},
/* 9461 */ {(11<<2)|2,{99,104,0}},
/* 9462 */ {(11<<2)|2,{99,104,0}},
/* 9463 */ {(11<<2)|2,{99,104,0}},
/* 9464 */ {(11<<2)|2,{99,104,0}},
/* 9465 */ {(11<<2)|2,{99,104,0}},
/* 9466 */ {(11<<2)|2,{99,104,0}},
/* 9467 */ {(11<<2)|2,{99,104,0}},
/* 9468 */ {(11<<2)|2,{99,104,0}},
/* 9469 */ {(11<<2)|2,{99,104,0}},
/* 9470 */ {(11<<2)|2,{99,104,0}},
/* 9471 */ {(11<<2)|2,{99,104,0}},
/* 9472 */ {(16<<2)|3,{99,108,48}},
/* 9473 */ {(16<<2)|3,{99,108,49}},
/* 9474 */ {(16<<2)|3,{99,108,50}},
/* 9475 */ {(16<<2)|3,{99,108,97}},
/* 9476 */ {(16<<2)|3,{99,108,99}},
/* 9477 */ {(16<<2)|3,{99,108,101}},
/* 9478 */ {(16<<2)|3,{99,108,105}},
/* 9479 */ {(16<<2)|3,{99,108,111}},
/* 9480 */ {(16<<2)|3,{99,108,115}},
/* 9481 */ {(16<<2)|3,{99,108,116}},
/* 9482 */ {(11<<2)|2,{99,108,0}},
/* 9483 */ {(11<<2)|2,{99,108,0}},
/* 9484 */ {(11<<2)|2,{99,108,0}},
/* 9485 */ {(11<<2)|2,{99,108,0}},
/* 9486 */ {(11<<2)|2,{99,108,0}},
/* 9487 */ {(11<<2)|2,{99,108,0}},
/* 9488 */ {(11<<2)|2,{99,108,0}},
/* 9489 */ {(11<<2)|2,{99,108,0}},
/* 9490 */ {(11<<2)|2,{99,108,0}},
/* 9491 */ {(11<<2)|2,{99,108,0}},
/* 9492 */ {(11<<2)|2,{99,108,0}},
/* 9493 */ {(11<<2)|2,{99,108,0}},
/* 9494 */ {(11<<2)|2,{99,108,0}},
/* 9495 */ {(11<<2)|2,{99,108,0}},
/* 9496 */ {(11<<2)|2,{99,108,0}},
/* 9497 */ {(11<<2)|2,{99,108,0}},
/* 9498 */ {(11<<2)|2,{99,108,0}},
/* 9499 */ {(11<<2)|2,{99,108,0}},
/* 9500 */ {(11<<2)|2,{99,108,0}},
/* 9501 */ {(11<<2)|2,{99,108,0}},
/* 9502 */ {(11<<2)|2,{99,108,0}},
/* 9503 */ {(11<<2)|2,{99,108,0}},
/* 9504 */ {(16<<2)|3,{99,109,48}},
/* 9505 */ {(16<<2)|3,{99,109,49}},
/* 9506 */ {(16<<2)|3,{99,109,50}},
/* 9507 */ {(16<<2)|3,{99,109,97}},
/* 9508 */ {(16<<2)|3,{99,109,99}},
/* 9509 */ {(16<<2)|3,{99,109,101}},
/* 9510 */ {(16<<2)|3,{99,109,105}},
/* 9511 */ {(16<<2)|3,{99,109,111}},
/* 9512 */ {(16<<2)|3,{99,109,115}},
/* 9513 */ {(16<<2)|3,{99,109,116}},
/* 9514 */ {(11<<2)|2,{99,109,0}},
/* 9515 */ {(11<<2)|2,{99,109,0}},
/* 9516 */ {(11<<2)|2,{99,109,0}},
/* 9517 */ {(11<<2)|2,{99,109,0}},
/* 9518 */ {(11<<2)|2,{99,109,0}},
/* 9519 */ {(11<<2)|2,{99,109,0}},
/* 9520 */ {(11<<2)|2,{99,109,0}},
/* 9521 */ {(11<<2)|2,{99,109,0}},
/* 9522 */ {(11<<2)|2,{99,109,0}},
/* 9523 */ {(11<<2)|2,{99,109,0}},
/* 9524 */ {(11<<2)|2,{99,109,0}},
/* 9525 */ {(11<<2)|2,{99,109,0}},
/* 9526 */ {(11<<2)|2,{99,109,0}},
/* 9527 */ {(11<<2)|2,{99,109,0}},
/* 9528 */ {(11<<2)|2,{99,109,0}},
/* 9529 */ {(11<<2)|2,{99,109,0}},
/* 9530 */ {(11<<2)|2,{99,109,0}},
/* 9531 */ {(11<<2)|2,{99,109,0}},
/* 9532 */ {(11<<2)|2,{99,109,0}},
/* 9533 */ {(11<<2)|2,{99,109,0}},
/* 9534 */ {(11<<2)|2,{99,109,0}},
/* 9535 */ {(11<<2)|2,{99,109,0}},
/* 9536 */ {(16<<2)|3,{99,110,48}},
/* 9537 */ {(16<<2)|3,{99,110,49}},
/* 9538 */ {(16<<2)|3,{99,110,50}},
/* 9539 */ {(16<<2)|3,{99,110,97}},
/* 9540 */ {(16<<2)|3,{99,110,99}},
/* 9541 */ {(16<<2)|3,{99,110,101}},
/* 9542 */ {(16<<2)|3,{99,110,105}},
/* 9543 */ {(16<<2)|3,{99,110,111}},
/* 9544 */ {(16<<2)|3,{99,110,115}},
/* 9545 */ {(16<<2)|3,{99,110,116}},
/* 9546 */ {(11<<2)|2,{99,110,0}},
/* 9547 */ {(11<<2)|2,{99,110,0}},
/* 9548 */ {(11<<2)|2,{99,110,0}},
/* 9549 */ {(11<<2)|2,{99,110,0}},
/* 9550 */ {(11<<2)|2,{99,110,0}},
/* 9551 */ {(11<<2)|2,{99,110,0}},
/* 9552 */ {(11<<2)|2,{99,110,0}},
/* 9553 */ {(11<<2)|2,{99,110,0}},
/* 9554 */ {(11<<2)|2,{99,110,0}},
/* 9555 */ {(11<<2)|2,{99,110,0}},
/* 9556 */ {(11<<2)|2,{99,110,0}},
/* 9557 */ {(11<<2)|2,{99,110,0}},
/* 9558 */ {(11<<2)|2,{99,110,0}},
/* 9559 */ {(11<<2)|2,{99,110,0}},
/* 9560 */ {(11<<2)|2,{99,110,0}},
/* 9561 */ {(11<<2)|2,{99,110,0}},
/* 9562 */ {(11<<2)|2,{99,110,0}},
/* 9563 */ {(11<<2)|2,{99,110,0}},
/* 9564 */ {(11<<2)|2,{99,110,0}},
/* 9565 */ {(11<<2)|2,{99,110,0}},
/* 9566 */ {(11<<2)|2,{99,110,0}},
/* 9567 */ {(11<<2)|2,{99,110,0}},
/* 9568 */ {(16<<2)|3,{99,112,48}},
/* 9569 */ {(16<<2)|3,{99,112,49}},
/* 9570 */ {(16<<2)|3,{99,112,50}},
/* 9571 */ {(16<<2)|3,{99,112,97}},
/* 9572 */ {(16<<2)|3,{99,112,99}},
/* 9573 */ {(16<<2)|3,{99,112,101}},
/* 9574 */ {(16<<2)|3,{99,112,105}},
/* 9575 */ {(16<<2)|3,{99,112,111}},
/* 9576 */ {(16<<2)|3,{99,112,115}},
/* 9577 */ {(16<<2)|3,{99,112,116}},
/* 9578 */ {(11<<2)|2,{99,112,0}},
/* 9579 */ {(11<<2)|2,{99,112,0}},
/* 9580 */ {(11<<2)|2,{99,112,0}},
/* 9581 */ {(11<<2)|2,{99,112,0}},
/* 9582 */ {(11<<2)|2,{99,112,0}},
/* 9583 */ {(11<<2)|2,{99,112,0}},
/* 9584 */ {(11<<2)|2,{99,112,0}},
/* 9585 */ {(11<<2)|2,{99,112,0}},
/* 9586 */ {(11<<2)|2,{99,112,0}},
/* 9587 */ {(11<<2)|2,{99,112,0}},
/* 9588 */ {(11<<2)|2,{99,112,0}},
/* 9589 */ {(11<<2)|2,{99,112,0}},
/* 9590 */ {(11<<2)|2,{99,112,0}},
/* 9591 */ {(11<<2)|2,{99,112,0}},
/* 9592 */ {(11<<2)|2,{99,112,0}},
/* 9593 */ {(11<<2)|2,{99,112,0}},
/* 9594 */ {(11<<2)|2,{99,112,0}},
/* 9595 */ {(11<<2)|2,{99,112,0}},
/* 9596 */ {(11<<2)|2,{99,112,0}},
/* 9597 */ {(11<<2)|2,{99,112,0}},
/* 9598 */ {(11<<2)|2,{99,112,0}},
/* 9599 */ {(11<<2)|2,{99,112,0}},
/* 9600 */ {(16<<2)|3,{99,114,48}},
/* 9601 */ {(16<<2)|3,{99,114,49}},
/* 9602 */ {(16<<2)|3,{99,114,50}},
/* 9603 */ {(16<<2)|3,{99,114,97}},
/* 9604 */ {(16<<2)|3,{99,114,99}},
/* 9605 */ {(16<<2)|3,{99,114,101}},
/* 9606 */ {(16<<2)|3,{99,114,105}},
/* 9607 */ {(16<<2)|3,{99,114,111}},
/* 9608 */ {(16<<2)|3,{99,114,115}},
/* 9609 */ {(16<<2)|3,{99,114,116}},
/* 9610 */ {(11<<2)|2,{99,114,0}},
/* 9611 */ {(11<<2)|2,{99,114,0}},
/* 9612 */ {(11<<2)|2,{99,114,0}},
/* 9613 */ {(11<<2)|2,{99,114,0}},
/* 9614 */ {(11<<2)|2,{99,114,0}},
/* 9615 */ {(11<<2)|2,{99,114,0}},
/* 9616 */ {(11<<2)|2,{99,114,0}},
/* 9617 */ {(11<<2)|2,{99,114,0}},
/* 9618 */ {(11<<2)|2,{99,114,0}},
/* 9619 */ {(11<<2)|2,{99,114,0}},
/* 9620 */ {(11<<2)|2,{99,114,0}},
/* 9621 */ {(11<<2)|2,{99,114,0}},
/* 9622 */ {(11<<2)|2,{99,114,0}},
/* 9623 */ {(11<<2)|2,{99,114,0}},
/* 9624 */ {(11<<2)|2,{99,114,0}},
/* 9625 */ {(11<<2)|2,{99,114,0}},
/* 9626 */ {(11<<2)|2,{99,114,0}},
/* 9627 */ {(11<<2)|2,{99,114,0}},
/* 9628 */ {(11<<2)|2,{99,114,0}},
/* 9629 */ {(11<<2)|2,{99,114,0}},
/* 9630 */ {(11<<2)|2,{99,114,0}},
/* 9631 */ {(11<<2)|2,{99,114,0}},
/* 9632 */ {(16<<2)|3,{99,117,48}},
/* 9633 */ {(16<<2)|3,{99,117,49}},
/* 9634 */ {(16<<2)|3,{99,117,50}},
/* 9635 */ {(16<<2)|3,{99,117,97}},
/* 9636 */ {(16<<2)|3,{99,117,99}},
/* 9637 */ {(16<<2)|3,{99,117,101}},
/* 9638 */ {(16<<2)|3,{99,117,105}},
/* 9639 */ {(16<<2)|3,{99,117,111}},
/* 9640 */ {(16<<2)|3,{99,117,115}},
/* 9641 */ {(16<<2)|3,{99,117,116}},
/* 9642 */ {(11<<2)|2,{99,117,0}},
/* 9643 */ {(11<<2)|2,{99,117,0}},
/* 9644 */ {(11<<2)|2,{99,117,0}},
/* 9645 */ {(11<<2)|2,{99,117,0}},
/* 9646 */ {(11<<2)|2,{99,117,0}},
/* 9647 */ {(11<<2)|2,{99,117,0}},
/* 9648 */ {(11<<2)|2,{99,117,0}},
/* 9649 */ {(11<<2)|2,{99,117,0}},
/* 9650 */ {(11<<2)|2,{99,117,0}},
/* 9651 */ {(11<<2)|2,{99,117,0}},
/* 9652 */ {(11<<2)|2,{99,117,0}},
/* 9653 */ {(11<<2)|2,{99,117,0}},
/* 9654 */ {(11<<2)|2,{99,117,0}},
/* 9655 */ {(11<<2)|2,{99,117,0}},
/* 9656 */ {(11<<2)|2,{99,117,0}},
/* 9657 */ {(11<<2)|2,{99,117,0}},
/* 9658 */ {(11<<2)|2,{99,117,0}},
/* 9659 */ {(11<<2)|2,{99,117,0}},
/* 9660 */ {(11<<2)|2,{99,117,0}},
/* 9661 */ {(11<<2)|2,{99,117,0}},
/* 9662 */ {(11<<2)|2,{99,117,0}},
/* 9663 */ {(11<<2)|2,{99,117,0}},
/* 9664 */ {(12<<2)|2,{99,58,0}},
/* 9665 */ {(12<<2)|2,{99,58,0}},
/* 9666 */ {(12<<2)|2,{99,58,0}},
/* 9667 */ {(12<<2)|2,{99,58,0}},
/* 9668 */ {(12<<2)|2,{99,58,0}},
/* 9669 */ {(12<<2)|2,{99,58,0}},
/* 9670 */ {(12<<2)|2,{99,58,0}},
/* 9671 */ {(12<<2)|2,{99,58,0}},
/* 9672 */ {(12<<2)|2,{99,58,0}},
/* 9673 */ {(12<<2)|2,{99,58,0}},
/* 9674 */ {(12<<2)|2,{99,58,0}},
/* 9675 */ {(12<<2)|2,{99,58,0}},
/* 9676 */ {(12<<2)|2,{99,58,0}},
/* 9677 */ {(12<<2)|2,{99,58,0}},
/* 9678 */ {(12<<2)|2,{99,58,0}},
/* 9679 */ {(12<<2)|2,{99,58,0}},
/* 9680 */ {(12<<2)|2,{99,66,0}},
/* 9681 */ {(12<<2)|2,{99,66,0}},
/* 9682 */ {(12<<2)|2,{99,66,0}},
/* 9683 */ {(12<<2)|2,{99,66,0}},
/* 9684 */ {(12<<2)|2,{99,66,0}},
/* 9685 */ {(12<<2)|2,{99,66,0}},
/* 9686 */ {(12<<2)|2,{99,66,0}},
/* 9687 */ {(12<<2)|2,{99,66,0}},
/* 9688 */ {(12<<2)|2,{99,66,0}},
/* 9689 */ {(12<<2)|2,{99,66,0}},
/* 9690 */ {(12<<2)|2,{99,66,0}},
/* 9691 */ {(12<<2)|2,{99,66,0}},
/* 9692 */ {(12<<2)|2,{99,66,0}},
/* 9693 */ {(12<<2)|2,{99,66,0}},
/* 9694 */ {(12<<2)|2,{99,66,0}},
/* 9695 */ {(12<<2)|2,{99,66,0}},
/* 9696 */ {(12<<2)|2,{99,67,0}},
/* 9697 */ {(12<<2)|2,{99,67,0}},
/* 9698 */ {(12<<2)|2,{99,67,0}},
/* 9699 */ {(12<<2)|2,{99,67,0}},
/* 9700 */ {(12<<2)|2,{99,67,0}},
/* 9701 */ {(12<<2)|2,{99,67,0}},
/* 9702 */ {(12<<2)|2,{99,67,0}},
/* 9703 */ {(12<<2)|2,{99,67,0}},
/* 9704 */ {(12<<2)|2,{99,67,0}},
/* 9705 */ {(12<<2)|2,{99,67,0}},
/* 9706 */ {(12<<2)|2,{99,67,0}},
/* 9707 */ {(12<<2)|2,{99,67,0}},
/* 9708 */ {(12<<2)|2,{99,67,0}},
/* 9709 */ {(12<<2)|2,{99,67,0}},
/* 9710 */ {(12<<2)|2,{99,67,0}},
/* 9711 */ {(12<<2)|2,{99,67,0}},
/* 9712 */ {(12<<2)|2,{99,68,0}},
/* 9713 */ {(12<<2)|2,{99,68,0}},
/* 9714 */ {(12<<2)|2,{99,68,0}},
/* 9715 */ {(12<<2)|2,{99,68,0}},
/* 9716 */ {(12<<2)|2,{99,68,0}},
/* 9717 */ {(12<<2)|2,{99,68,0}},
/* 9718 */ {(12<<2)|2,{99,68,0}},
/* 9719 */ {(12<<2)|2,{99,68,0}},
/* 9720 */ {(12<<2)|2,{99,68,0}},
/* 9721 */ {(12<<2)|2,{99,68,0}},
/* 9722 */ {(12<<2)|2,{99,68,0}},
/* 9723 */ {(12<<2)|2,{99,68,0}},
/* 9724 */ {(12<<2)|2,{99,68,0}},
/* 9725 */ {(12<<2)|2,{99,68,0}},
/* 9726 */ {(12<<2)|2,{99,68,0}},
/* 9727 */ {(12<<2)|2,{99,68,0}},
/* 9728 */ {(12<<2)|2,{99,69,0}},
/* 9729 */ {(12<<2)|2,{99,69,0}},
/* 9730 */ {(12<<2)|2,{99,69,0}},
/* 9731 */ {(12<<2)|2,{99,69,0}},
/* 9732 */ {(12<<2)|2,{99,69,0}},
/* 9733 */ {(12<<2)|2,{99,69,0}},
/* 9734 */ {(12<<2)|2,{99,69,0}},
/* 9735 */ {(12<<2)|2,{99,69,0}},
/* 9736 */ {(12<<2)|2,{99,69,0}},
/* 9737 */ {(12<<2)|2,{99,69,0}},
/* 9738 */ {(12<<2)|2,{99,69,0}},
/* 9739 */ {(12<<2)|2,{99,69,0}},
/* 9740 */ {(12<<2)|2,{99,69,0}},
/* 9741 */ {(12<<2)|2,{99,69,0}},
/* 9742 */ {(12<<2)|2,{99,69,0}},
/* 9743 */ {(12<<2)|2,{99,69,0}},
/* 9744 */ {(12<<2)|2,{99,70,0}},
/* 9745 */ {(12<<2)|2,{99,70,0}},
/* 9746 */ {(12<<2)|2,{99,70,0}},
/* 9747 */ {(12<<2)|2,{99,70,0}},
/* 9748 */ {(12<<2)|2,{99,70,0}},
/* 9749 */ {(12<<2)|2,{99,70,0}},
/* 9750 */ {(12<<2)|2,{99,70,0}},
/* 9751 */ {(12<<2)|2,{99,70,0}},
/* 9752 */ {(12<<2)|2,{99,70,0}},
/* 9753 */ {(12<<2)|2,{99,70,0}},
/* 9754 */ {(12<<2)|2,{99,70,0}},
/* 9755 */ {(12<<2)|2,{99,70,0}},
/* 9756 */ {(12<<2)|2,{99,70,0}},
/* 9757 */ {(12<<2)|2,{99,70,0}},
/* 9758 */ {(12<<2)|2,{99,70,0}},
/* 9759 */ {(12<<2)|2,{99,70,0}},
/* 9760 */ {(12<<2)|2,{99,71,0}},
/* 9761 */ {(12<<2)|2,{99,71,0}},
/* 9762 */ {(12<<2)|2,{99,71,0}},
/* 9763 */ {(12<<2)|2,{99,71,0}},
/* 9764 */ {(12<<2)|2,{99,71,0}},
/* 9765 */ {(12<<2)|2,{99,71,0}},
/* 9766 */ {(12<<2)|2,{99,71,0}},
/* 9767 */ {(12<<2)|2,{99,71,0}},
/* 9768 */ {(12<<2)|2,{99,71,0}},
/* 9769 */ {(12<<2)|2,{99,71,0}},
/* 9770 */ {(12<<2)|2,{99,71,0}},
/* 9771 */ {(12<<2)|2,{99,71,0}},
/* 9772 */ {(12<<2)|2,{99,71,0}},
/* 9773 */ {(12<<2)|2,{99,71,0}},
/* 9774 */ {(12<<2)|2,{99,71,0}},
/* 9775 */ {(12<<2)|2,{99,71,0}},
/* 9776 */ {(12<<2)|2,{99,72,0}},
/* 9777 */ {(12<<2)|2,{99,72,0}},
/* 9778 */ {(12<<2)|2,{99,72,0}},
/* 9779 */ {(12<<2)|2,{99,72,0}},
/* 9780 */ {(12<<2)|2,{99,72,0}},
/* 9781 */ {(12<<2)|2,{99,72,0}},
/* 9782 */ {(12<<2)|2,{99,72,0}},
/* 9783 */ {(12<<2)|2,{99,72,0}},
/* 9784 */ {(12<<2)|2,{99,72,0}},
/* 9785 */ {(12<<2)|2,{99,72,0}},
/* 9786 */ {(12<<2)|2,{99,72,0}},
/* 9787 */ {(12<<2)|2,{99,72,0}},
/* 9788 */ {(12<<2)|2,{99,72,0}},
/* 9789 */ {(12<<2)|2,{99,72,0}},
/* 9790 */ {(12<<2)|2,{99,72,0}},
/* 9791 */ {(12<<2)|2,{99,72,0}},
/* 9792 */ {(12<<2)|2,{99,73,0}},
/* 9793 */ {(12<<2)|2,{99,73,0}},
/* 9794 */ {(12<<2)|2,{99,73,0}},
/* 9795 */ {(12<<2)|2,{99,73,0}},
/* 9796 */ {(12<<2)|2,{99,73,0}},
/* 9797 */ {(12<<2)|2,{99,73,0}},
/* 9798 */ {(12<<2)|2,{99,73,0}},
/* 9799 */ {(12<<2)|2,{99,73,0}},
/* 9800 */ {(12<<2)|2,{99,73,0}},
/* 9801 */ {(12<<2)|2,{99,73,0}},
/* 9802 */ {(12<<2)|2,{99,73,0}},
/* 9803 */ {(12<<2)|2,{99,73,0}},
/* 9804 */ {(12<<2)|2,{99,73,0}},
/* 9805 */ {(12<<2)|2,{99,73,0}},
/* 9806 */ {(12<<2)|2,{99,73,0}},
/* 9807 */ {(12<<2)|2,{99,73,0}},
/* 9808 */ {(12<<2)|2,{99,74,0}},
/* 9809 */ {(12<<2)|2,{99,74,0}},
/* 9810 */ {(12<<2)|2,{99,74,0}},
/* 9811 */ {(12<<2)|2,{99,74,0}},
/* 9812 */ {(12<<2)|2,{99,74,0}},
/* 9813 */ {(12<<2)|2,{99,74,0}},
/* 9814 */ {(12<<2)|2,{99,74,0}},
/* 9815 */ {(12<<2)|2,{99,74,0}},
/* 9816 */ {(12<<2)|2,{99,74,0}},
/* 9817 */ {(12<<2)|2,{99,74,0}},
/* 9818 */ {(12<<2)|2,{99,74,0}},
/* 9819 */ {(12<<2)|2,{99,74,0}},
/* 9820 */ {(12<<2)|2,{99,74,0}},
/* 9821 */ {(12<<2)|2,{99,74,0}},
/* 9822 */ {(12<<2)|2,{99,74,0}},
/* 9823 */ {(12<<2)|2,{99,74,0}},
/* 9824 */ {(12<<2)|2,{99,75,0}},
/* 9825 */ {(12<<2)|2,{99,75,0}},
/* 9826 */ {(12<<2)|2,{99,75,0}},
/* 9827 */ {(12<<2)|2,{99,75,0}},
/* 9828 */ {(12<<2)|2,{99,75,0}},
/* 9829 */ {(12<<2)|2,{99,75,0}},
/* 9830 */ {(12<<2)|2,{99,75,0}},
/* 9831 */ {(12<<2)|2,{99,75,0}},
/* 9832 */ {(12<<2)|2,{99,75,0}},
/* 9833 */ {(12<<2)|2,{99,75,0}},
/* 9834 */ {(12<<2)|2,{99,75,0}},
/* 9835 */ {(12<<2)|2,{99,75,0}},
/* 9836 */ {(12<<2)|2,{99,75,0}},
/* 9837 */ {(12<<2)|2,{99,75,0}},
/* 9838 */ {(12<<2)|2,{99,75,0}},
/* 9839 */ {(12<<2)|2,{99,75,0}},
/* 9840 */ {(12<<2)|2,{99,76,0}},
/* 9841 */ {(12<<2)|2,{99,76,0}},
/* 9842 */ {(12<<2)|2,{99,76,0}},
/* 9843 */ {(12<<2)|2,{99,76,0}},
/* 9844 */ {(12<<2)|2,{99,76,0}},
/* 9845 */ {(12<<2)|2,{99,76,0}},
/* 9846 */ {(12<<2)|2,{99,76,0}},
/* 9847 */ {(12<<2)|2,{99,76,0}},
/* 9848 */ {(12<<2)|2,{99,76,0}},
/* 9849 */ {(12<<2)|2,{99,76,0}},
/* 9850 */ {(12<<2)|2,{99,76,0}},
/* 9851 */ {(12<<2)|2,{99,76,0}},
/* 9852 */ {(12<<2)|2,{99,76,0}},
/* 9853 */ {(12<<2)|2,{99,76,0}},
/* 9854 */ {(12<<2)|2,{99,76,0}},
/* 9855 */ {(12<<2)|2,{99,76,0}},
/* 9856 */ {(12<<2)|2,{99,77,0}},
/* 9857 */ {(12<<2)|2,{99,77,0}},
/* 9858 */ {(12<<2)|2,{99,77,0}},
/* 9859 */ {(12<<2)|2,{99,77,0}},
/* 9860 */ {(12<<2)|2,{99,77,0}},
/* 9861 */ {(12<<2)|2,{99,77,0}},
/* 9862 */ {(12<<2)|2,{99,77,0}},
/* 9863 */ {(12<<2)|2,{99,77,0}},
/* 9864 */ {(12<<2)|2,{99,77,0}},
/* 9865 */ {(12<<2)|2,{99,77,0}},
/* 9866 */ {(12<<2)|2,{99,77,0}},
/* 9867 */ {(12<<2)|2,{99,77,0}},
/* 9868 */ {(12<<2)|2,{99,77,0}},
/* 9869 */ {(12<<2)|2,{99,77,0}},
/* 9870 */ {(12<<2)|2,{99,77,0}},
/* 9871 */ {(12<<2)|2,{99,77,0}},
/* 9872 */ {(12<<2)|2,{99,78,0}},
/* 9873 */ {(12<<2)|2,{99,78,0}},
/* 9874 */ {(12<<2)|2,{99,78,0}},
/* 9875 */ {(12<<2)|2,{99,78,0}},
/* 9876 */ {(12<<2)|2,{99,78,0}},
/* 9877 */ {(12<<2)|2,{99,78,0}},
/* 9878 */ {(12<<2)|2,{99,78,0}},
/* 9879 */ {(12<<2)|2,{99,78,0}},
/* 9880 */ {(12<<2)|2,{99,78,0}},
/* 9881 */ {(12<<2)|2,{99,78,0}},
/* 9882 */ {(12<<2)|2,{99,78,0}},
/* 9883 */ {(12<<2)|2,{99,78,0}},
/* 9884 */ {(12<<2)|2,{99,78,0}},
/* 9885 */ {(12<<2)|2,{99,78,0}},
/* 9886 */ {(12<<2)|2,{99,78,0}},
/* 9887 */ {(12<<2)|2,{99,78,0}},
/* 9888 */ {(12<<2)|2,{99,79,0}},
/* 9889 */ {(12<<2)|2,{99,79,0}},
/* 9890 */ {(12<<2)|2,{99,79,0}},
/* 9891 */ {(12<<2)|2,{99,79,0}},
/* 9892 */ {(12<<2)|2,{99,79,0}},
/* 9893 */ {(12<<2)|2,{99,79,0}},
/* 9894 */ {(12<<2)|2,{99,79,0}},
/* 9895 */ {(12<<2)|2,{99,79,0}},
/* 9896 */ {(12<<2)|2,{99,79,0}},
/* 9897 */ {(12<<2)|2,{99,79,0}},
/* 9898 */ {(12<<2)|2,{99,79,0}},
/* 9899 */ {(12<<2)|2,{99,79,0}},
/* 9900 */ {(12<<2)|2,{99,79,0}},
/* 9901 */ {(12<<2)|2,{99,79,0}},
/* 9902 */ {(12<<2)|2,{99,79,0}},
/* 9903 */ {(12<<2)|2,{99,79,0}},
/* 9904 */ {(12<<2)|2,{99,80,0}},
/* 9905 */ {(12<<2)|2,{99,80,0}},
/* 9906 */ {(12<<2)|2,{99,80,0}},
/* 9907 */ {(12<<2)|2,{99,80,0}},
/* 9908 */ {(12<<2)|2,{99,80,0}},
/* 9909 */ {(12<<2)|2,{99,80,0}},
/* 9910 */ {(12<<2)|2,{99,80,0}},
/* 9911 */ {(12<<2)|2,{99,80,0}},
/* 9912 */ {(12<<2)|2,{99,80,0}},
/* 9913 */ {(12<<2)|2,{99,80,0}},
/* 9914 */ {(12<<2)|2,{99,80,0}},
/* 9915 */ {(12<<2)|2,{99,80,0}},
/* 9916 */ {(12<<2)|2,{99,80,0}},
/* 9917 */ {(12<<2)|2,{99,80,0}},
/* 9918 */ {(12<<2)|2,{99,80,0}},
/* 9919 */ {(12<<2)|2,{99,80,0}},
/* 9920 */ {(12<<2)|2,{99,81,0}},
/* 9921 */ {(12<<2)|2,{99,81,0}},
/* 9922 */ {(12<<2)|2,{99,81,0}},
/* 9923 */ {(12<<2)|2,{99,81,0}},
/* 9924 */ {(12<<2)|2,{99,81,0}},
/* 9925 */ {(12<<2)|2,{99,81,0}},
/* 9926 */ {(12<<2)|2,{99,81,0}},
/* 9927 */ {(12<<2)|2,{99,81,0}},
/* 9928 */ {(12<<2)|2,{99,81,0}},
/* 9929 */ {(12<<2)|2,{99,81,0}},
/* 9930 */ {(12<<2)|2,{99,81,0}},
/* 9931 */ {(12<<2)|2,{99,81,0}},
/* 9932 */ {(12<<2)|2,{99,81,0}},
/* 9933 */ {(12<<2)|2,{99,81,0}},
/* 9934 */ {(12<<2)|2,{99,81,0}},
/* 9935 */ {(12<<2)|2,{99,81,0}},
/* 9936 */ {(12<<2)|2,{99,82,0}},
/* 9937 */ {(12<<2)|2,{99,82,0}},
/* 9938 */ {(12<<2)|2,{99,82,0}},
/* 9939 */ {(12<<2)|2,{99,82,0}},
/* 9940 */ {(12<<2)|2,{99,82,0}},
/* 9941 */ {(12<<2)|2,{99,82,0}},
/* 9942 */ {(12<<2)|2,{99,82,0}},
/* 9943 */ {(12<<2)|2,{99,82,0}},
/* 9944 */ {(12<<2)|2,{99,82,0}},
/* 9945 */ {(12<<2)|2,{99,82,0}},
/* 9946 */ {(12<<2)|2,{99,82,0}},
/* 9947 */ {(12<<2)|2,{99,82,0}},
/* 9948 */ {(12<<2)|2,{99,82,0}},
/* 9949 */ {(12<<2)|2,{99,82,0}},
/* 9950 */ {(12<<2)|2,{99,82,0}},
/* 9951 */ {(12<<2)|2,{99,82,0}},
/* 9952 */ {(12<<2)|2,{99,83,0}},
/* 9953 */ {(12<<2)|2,{99,83,0}},
/* 9954 */ {(12<<2)|2,{99,83,0}},
/* 9955 */ {(12<<2)|2,{99,83,0}},
/* 9956 */ {(12<<2)|2,{99,83,0}},
/* 9957 */ {(12<<2)|2,{99,83,0}},
/* 9958 */ {(12<<2)|2,{99,83,0}},
/* 9959 */ {(12<<2)|2,{99,83,0}},
/* 9960 */ {(12<<2)|2,{99,83,0}},
/* 9961 */ {(12<<2)|2,{99,83,0}},
/* 9962 */ {(12<<2)|2,{99,83,0}},
/* 9963 */ {(12<<2)|2,{99,83,0}},
/* 9964 */ {(12<<2)|2,{99,83,0}},
/* 9965 */ {(12<<2)|2,{99,83,0}},
/* 9966 */ {(12<<2)|2,{99,83,0}},
/* 9967 */ {(12<<2)|2,{99,83,0}},
/* 9968 */ {(12<<2)|2,{99,84,0}},
/* 9969 */ {(12<<2)|2,{99,84,0}},
/* 9970 */ {(12<<2)|2,{99,84,0}},
/* 9971 */ {(12<<2)|2,{99,84,0}},
/* 9972 */ {(12<<2)|2,{99,84,0}},
/* 9973 */ {(12<<2)|2,{99,84,0}},
/* 9974 */ {(12<<2)|2,{99,84,0}},
/* 9975 */ {(12<<2)|2,{99,84,0}},
/* 9976 */ {(12<<2)|2,{99,84,0}},
/* 9977 */ {(12<<2)|2,{99,84,0}},
/* 9978 */ {(12<<2)|2,{99,84,0}},
/* 9979 */ {(12<<2)|2,{99,84,0}},
/* 9980 */ {(12<<2)|2,{99,84,0}},
/* 9981 */ {(12<<2)|2,{99,84,0}},
/* 9982 */ {(12<<2)|2,{99,84,0}},
/* 9983 */ {(12<<2)|2,{99,84,0}},
/* 9984 */ {(12<<2)|2,{99,85,0}},
/* 9985 */ {(12<<2)|2,{99,85,0}},
/* 9986 */ {(12<<2)|2,{99,85,0}},
/* 9987 */ {(12<<2)|2,{99,85,0}},
/* 9988 */ {(12<<2)|2,{99,85,0}},
/* 9989 */ {(12<<2)|2,{99,85,0}},
/* 9990 */ {(12<<2)|2,{99,85,0}},
/* 9991 */ {(12<<2)|2,{99,85,0}},
/* 9992 */ {(12<<2)|2,{99,85,0}},
/* 9993 */ {(12<<2)|2,{99,85,0}},
/* 9994 */ {(12<<2)|2,{99,85,0}},
/* 9995 */ {(12<<2)|2,{99,85,0}},
/* 9996 */ {(12<<2)|2,{99,85,0}},
/* 9997 */ {(12<<2)|2,{99,85,0}},
/* 9998 */ {(12<<2)|2,{99,85,0}},
/* 9999 */ {(12<<2)|2,{99,85,0}},
/* 10000 */ {(12<<2)|2,{99,86,0}},
/* 10001 */ {(12<<2)|2,{99,86,0}},
/* 10002 */ {(12<<2)|2,{99,86,0}},
/* 10003 */ {(12<<2)|2,{99,86,0}},
/* 10004 */ {(12<<2)|2,{99,86,0}},
/* 10005 */ {(12<<2)|2,{99,86,0}},
/* 10006 */ {(12<<2)|2,{99,86,0}},
/* 10007 */ {(12<<2)|2,{99,86,0}},
/* 10008 */ {(12<<2)|2,{99,86,0}},
/* 10009 */ {(12<<2)|2,{99,86,0}},
/* 10010 */ {(12<<2)|2,{99,86,0}},
/* 10011 */ {(12<<2)|2,{99,86,0}},
/* 10012 */ {(12<<2)|2,{99,86,0}},
/* 10013 */ {(12<<2)|2,{99,86,0}},
/* 10014 */ {(12<<2)|2,{99,86,0}},
/* 10015 */ {(12<<2)|2,{99,86,0}},
/* 10016 */ {(12<<2)|2,{99,87,0}},
/* 10017 */ {(12<<2)|2,{99,87,0}},
/* 10018 */ {(12<<2)|2,{99,87,0}},
/* 10019 */ {(12<<2)|2,{99,87,0}},
/* 10020 */ {(12<<2)|2,{99,87,0}},
/* 10021 */ {(12<<2)|2,{99,87,0}},
/* 10022 */ {(12<<2)|2,{99,87,0}},
/* 10023 */ {(12<<2)|2,{99,87,0}},
/* 10024 */ {(12<<2)|2,{99,87,0}},
/* 10025 */ {(12<<2)|2,{99,87,0}},
/* 10026 */ {(12<<2)|2,{99,87,0}},
/* 10027 */ {(12<<2)|2,{99,87,0}},
/* 10028 */ {(12<<2)|2,{99,87,0}},
/* 10029 */ {(12<<2)|2,{99,87,0}},
/* 10030 */ {(12<<2)|2,{99,87,0}},
/* 10031 */ {(12<<2)|2,{99,87,0}},
/* 10032 */ {(12<<2)|2,{99,89,0}},
/* 10033 */ {(12<<2)|2,{99,89,0}},
/* 10034 */ {(12<<2)|2,{99,89,0}},
/* 10035 */ {(12<<2)|2,{99,89,0}},
/* 10036 */ {(12<<2)|2,{99,89,0}},
/* 10037 */ {(12<<2)|2,{99,89,0}},
/* 10038 */ {(12<<2)|2,{99,89,0}},
/* 10039 */ {(12<<2)|2,{99,89,0}},
/* 10040 */ {(12<<2)|2,{99,89,0}},
/* 10041 */ {(12<<2)|2,{99,89,0}},
/* 10042 */ {(12<<2)|2,{99,89,0}},
/* 10043 */ {(12<<2)|2,{99,89,0}},
/* 10044 */ {(12<<2)|2,{99,89,0}},
/* 10045 */ {(12<<2)|2,{99,89,0}},
/* 10046 */ {(12<<2)|2,{99,89,0}},
/* 10047 */ {(12<<2)|2,{99,89,0}},
/* 10048 */ {(12<<2)|2,{99,106,0}},
/* 10049 */ {(12<<2)|2,{99,106,0}},
/* 10050 */ {(12<<2)|2,{99,106,0}},
/* 10051 */ {(12<<2)|2,{99,106,0}},
/* 10052 */ {(12<<2)|2,{99,106,0}},
/* 10053 */ {(12<<2)|2,{99,106,0}},
/* 10054 */ {(12<<2)|2,{99,106,0}},
/* 10055 */ {(12<<2)|2,{99,106,0}},
/* 10056 */ {(12<<2)|2,{99,106,0}},
/* 10057 */ {(12<<2)|2,{99,106,0}},
/* 10058 */ {(12<<2)|2,{99,106,0}},
/* 10059 */ {(12<<2)|2,{99,106,0}},
/* 10060 */ {(12<<2)|2,{99,106,0}},
/* 10061 */ {(12<<2)|2,{99,106,0}},
/* 10062 */ {(12<<2)|2,{99,106,0}},
/* 10063 */ {(12<<2)|2,{99,106,0}},
/* 10064 */ {(12<<2)|2,{99,107,0}},
/* 10065 */ {(12<<2)|2,{99,107,0}},
/* 10066 */ {(12<<2)|2,{99,107,0}},
/* 10067 */ {(12<<2)|2,{99,107,0}},
/* 10068 */ {(12<<2)|2,{99,107,0}},
/* 10069 */ {(12<<2)|2,{99,107,0}},
/* 10070 */ {(12<<2)|2,{99,107,0}},
/* 10071 */ {(12<<2)|2,{99,107,0}},
/* 10072 */ {(12<<2)|2,{99,107,0}},
/* 10073 */ {(12<<2)|2,{99,107,0}},
/* 10074 */ {(12<<2)|2,{99,107,0}},
/* 10075 */ {(12<<2)|2,{99,107,0}},
/* 10076 */ {(12<<2)|2,{99,107,0}},
/* 10077 */ {(12<<2)|2,{99,107,0}},
/* 10078 */ {(12<<2)|2,{99,107,0}},
/* 10079 */ {(12<<2)|2,{99,107,0}},
/* 10080 */ {(12<<2)|2,{99,113,0}},
/* 10081 */ {(12<<2)|2,{99,113,0}},
/* 10082 */ {(12<<2)|2,{99,113,0}},
/* 10083 */ {(12<<2)|2,{99,113,0}},
/* 10084 */ {(12<<2)|2,{99,113,0}},
/* 10085 */ {(12<<2)|2,{99,113,0}},
/* 10086 */ {(12<<2)|2,{99,113,0}},
/* 10087 */ {(12<<2)|2,{99,113,0}},
/* 10088 */ {(12<<2)|2,{99,113,0}},
/* 10089 */ {(12<<2)|2,{99,113,0}},
/* 10090 */ {(12<<2)|2,{99,113,0}},
/* 10091 */ {(12<<2)|2,{99,113,0}},
/* 10092 */ {(12<<2)|2,{99,113,0}},
/* 10093 */ {(12<<2)|2,{99,113,0}},
/* 10094 */ {(12<<2)|2,{99,113,0}},
/* 10095 */ {(12<<2)|2,{99,113,0}},
/* 10096 */ {(12<<2)|2,{99,118,0}},
/* 10097 */ {(12<<2)|2,{99,118,0}},
/* 10098 */ {(12<<2)|2,{99,118,0}},
/* 10099 */ {(12<<2)|2,{99,118,0}},
/* 10100 */ {(12<<2)|2,{99,118,0}},
/* 10101 */ {(12<<2)|2,{99,118,0}},
/* 10102 */ {(12<<2)|2,{99,118,0}},
/* 10103 */ {(12<<2)|2,{99,118,0}},
/* 10104 */ {(12<<2)|2,{99,118,0}},
/* 10105 */ {(12<<2)|2,{99,118,0}},
/* 10106 */ {(12<<2)|2,{99,118,0}},
/* 10107 */ {(12<<2)|2,{99,118,0}},
/* 10108 */ {(12<<2)|2,{99,118,0}},
/* 10109 */ {(12<<2)|2,{99,118,0}},
/* 10110 */ {(12<<2)|2,{99,118,0}},
/* 10111 */ {(12<<2)|2,{99,118,0}},
/* 10112 */ {(12<<2)|2,{99,119,0}},
/* 10113 */ {(12<<2)|2,{99,119,0}},
/* 10114 */ {(12<<2)|2,{99,119,0}},
/* 10115 */ {(12<<2)|2,{99,119,0}},
/* 10116 */ {(12<<2)|2,{99,119,0}},
/* 10117 */ {(12<<2)|2,{99,119,0}},
/* 10118 */ {(12<<2)|2,{99,119,0}},
/* 10119 */ {(12<<2)|2,{99,119,0}},
/* 10120 */ {(12<<2)|2,{99,119,0}},
/* 10121 */ {(12<<2)|2,{99,119,0}},
/* 10122 */ {(12<<2)|2,{99,119,0}},
/* 10123 */ {(12<<2)|2,{99,119,0}},
/* 10124 */ {(12<<2)|2,{99,119,0}},
/* 10125 */ {(12<<2)|2,{99,119,0}},
/* 10126 */ {(12<<2)|2,{99,119,0}},
/* 10127 */ {(12<<2)|2,{99,119,0}},
/* 10128 */ {(12<<2)|2,{99,120,0}},
/* 10129 */ {(12<<2)|2,{99,120,0}},
/* 10130 */ {(12<<2)|2,{99,120,0}},
/* 10131 */ {(12<<2)|2,{99,120,0}},
/* 10132 */ {(12<<2)|2,{99,120,0}},
/* 10133 */ {(12<<2)|2,{99,120,0}},
/* 10134 */ {(12<<2)|2,{99,120,0}},
/* 10135 */ {(12<<2)|2,{99,120,0}},
/* 10136 */ {(12<<2)|2,{99,120,0}},
/* 10137 */ {(12<<2)|2,{99,120,0}},
/* 10138 */ {(12<<2)|2,{99,120,0}},
/* 10139 */ {(12<<2)|2,{99,120,0}},
/* 10140 */ {(12<<2)|2,{99,120,0}},
/* 10141 */ {(12<<2)|2,{99,120,0}},
/* 10142 */ {(12<<2)|2,{99,120,0}},
/* 10143 */ {(12<<2)|2,{99,120,0}},
/* 10144 */ {(12<<2)|2,{99,121,0}},
/* 10145 */ {(12<<2)|2,{99,121,0}},
/* 10146 */ {(12<<2)|2,{99,121,0}},
/* 10147 */ {(12<<2)|2,{99,121,0}},
/* 10148 */ {(12<<2)|2,{99,121,0}},
/* 10149 */ {(12<<2)|2,{99,121,0}},
/* 10150 */ {(12<<2)|2,{99,121,0}},
/* 10151 */ {(12<<2)|2,{99,121,0}},
/* 10152 */ {(12<<2)|2,{99,121,0}},
/* 10153 */ {(12<<2)|2,{99,121,0}},
/* 10154 */ {(12<<2)|2,{99,121,0}},
/* 10155 */ {(12<<2)|2,{99,121,0}},
/* 10156 */ {(12<<2)|2,{99,121,0}},
/* 10157 */ {(12<<2)|2,{99,121,0}},
/* 10158 */ {(12<<2)|2,{99,121,0}},
/* 10159 */ {(12<<2)|2,{99,121,0}},
/* 10160 */ {(12<<2)|2,{99,122,0}},
/* 10161 */ {(12<<2)|2,{99,122,0}},
/* 10162 */ {(12<<2)|2,{99,122,0}},
/* 10163 */ {(12<<2)|2,{99,122,0}},
/* 10164 */ {(12<<2)|2,{99,122,0}},
/* 10165 */ {(12<<2)|2,{99,122,0}},
/* 10166 */ {(12<<2)|2,{99,122,0}},
/* 10167 */ {(12<<2)|2,{99,122,0}},
/* 10168 */ {(12<<2)|2,{99,122,0}},
/* 10169 */ {(12<<2)|2,{99,122,0}},
/* 10170 */ {(12<<2)|2,{99,122,0}},
/* 10171 */ {(12<<2)|2,{99,122,0}},
/* 10172 */ {(12<<2)|2,{99,122,0}},
/* 10173 */ {(12<<2)|2,{99,122,0}},
/* 10174 */ {(12<<2)|2,{99,122,0}},
/* 10175 */ {(12<<2)|2,{99,122,0}},
/* 10176 */ {(13<<2)|2,{99,38,0}},
/* 10177 */ {(13<<2)|2,{99,38,0}},
/* 10178 */ {(13<<2)|2,{99,38,0}},
/* 10179 */ {(13<<2)|2,{99,38,0}},
/* 10180 */ {(13<<2)|2,{99,38,0}},
/* 10181 */ {(13<<2)|2,{99,38,0}},
/* 10182 */ {(13<<2)|2,{99,38,0}},
/* 10183 */ {(13<<2)|2,{99,38,0}},
/* 10184 */ {(13<<2)|2,{99,42,0}},
/* 10185 */ {(13<<2)|2,{99,42,0}},
/* 10186 */ {(13<<2)|2,{99,42,0}},
/* 10187 */ {(13<<2)|2,{99,42,0}},
/* 10188 */ {(13<<2)|2,{99,42,0}},
/* 10189 */ {(13<<2)|2,{99,42,0}},
/* 10190 */ {(13<<2)|2,{99,42,0}},
/* 10191 */ {(13<<2)|2,{99,42,0}},
/* 10192 */ {(13<<2)|2,{99,44,0}},
/* 10193 */ {(13<<2)|2,{99,44,0}},
/* 10194 */ {(13<<2)|2,{99,44,0}},
/* 10195 */ {(13<<2)|2,{99,44,0}},
/* 10196 */ {(13<<2)|2,{99,44,0}},
/* 10197 */ {(13<<2)|2,{99,44,0}},
/* 10198 */ {(13<<2)|2,{99,44,0}},
/* 10199 */ {(13<<2)|2,{99,44,0}},
/* 10200 */ {(13<<2)|2,{99,59,0}},
/* 10201 */ {(13<<2)|2,{99,59,0}},
/* 10202 */ {(13<<2)|2,{99,59,0}},
/* 10203 */ {(13<<2)|2,{99,59,0}},
/* 10204 */ {(13<<2)|2,{99,59,0}},
/* 10205 */ {(13<<2)|2,{99,59,0}},
/* 10206 */ {(13<<2)|2,{99,59,0}},
/* 10207 */ {(13<<2)|2,{99,59,0}},
/* 10208 */ {(13<<2)|2,{99,88,0}},
/* 10209 */ {(13<<2)|2,{99,88,0}},
/* 10210 */ {(13<<2)|2,{99,88,0}},
/* 10211 */ {(13<<2)|2,{99,88,0}},
/* 10212 */ {(13<<2)|2,{99,88,0}},
/* 10213 */ {(13<<2)|2,{99,88,0}},
/* 10214 */ {(13<<2)|2,{99,88,0}},
/* 10215 */ {(13<<2)|2,{99,88,0}},
/* 10216 */ {(13<<2)|2,{99,90,0}},
/* 10217 */ {(13<<2)|2,{99,90,0}},
/* 10218 */ {(13<<2)|2,{99,90,0}},
/* 10219 */ {(13<<2)|2,{99,90,0}},
/* 10220 */ {(13<<2)|2,{99,90,0}},
/* 10221 */ {(13<<2)|2,{99,90,0}},
/* 10222 */ {(13<<2)|2,{99,90,0}},
/* 10223 */ {(13<<2)|2,{99,90,0}},
/* 10224 */ {(15<<2)|2,{99,33,0}},
/* 10225 */ {(15<<2)|2,{99,33,0}},
/* 10226 */ {(15<<2)|2,{99,34,0}},
/* 10227 */ {(15<<2)|2,{99,34,0}},
/* 10228 */ {(15<<2)|2,{99,40,0}},
/* 10229 */ {(15<<2)|2,{99,40,0}},
/* 10230 */ {(15<<2)|2,{99,41,0}},
/* 10231 */ {(15<<2)|2,{99,41,0}},
/* 10232 */ {(15<<2)|2,{99,63,0}},
/* 10233 */ {(15<<2)|2,{99,63,0}},
/* 10234 */ {(16<<2)|2,{99,39,0}},
/* 10235 */ {(16<<2)|2,{99,43,0}},
/* 10236 */ {(16<<2)|2,{99,124,0}},
/* 10237 */ {(5<<2)|1,{99,0,0}},
/* 10238 */ {(5<<2)|1,{99,0,0}},
/* 10239 */ {(5<<2)|1,{99,0,0}},
/* 10240 */ {(15<<2)|3,{101,48,48}},
/* 10241 */ {(15<<2)|3,{101,48,48}},
/* 10242 */ {(15<<2)|3,{101,48,49}},
/* 10243 */ {(15<<2)|3,{101,48,49}},
/* 10244 */ {(15<<2)|3,{101,48,50}},
/* 10245 */ {(15<<2)|3,{101,48,50}},
/* 10246 */ {(15<<2)|3,{101,48,97}},
/* 10247 */ {(15<<2)|3,{101,48,97}},
/* 10248 */ {(15<<2)|3,{101,48,99}},
/* 10249 */ {(15<<2)|3,{101,48,99}},
/* 10250 */ {(15<<2)|3,{101,48,101}},
/* 10251 */ {(15<<2)|3,{101,48,101}},
/* 10252 */ {(15<<2)|3,{101,48,105}},
/* 10253 */ {(15<<2)|3,{101,48,105}},
/* 10254 */ {(15<<2)|3,{101,48,111}},
/* 10255 */ {(15<<2)|3,{101,48,111}},
/* 10256 */ {(15<<2)|3,{101,48,115}},
/* 10257 */ {(15<<2)|3,{101,48,115}},
/* 10258 */ {(15<<2)|3,{101,48,116}},
/* 10259 */ {(15<<2)|3,{101,48,116}},
/* 10260 */ {(16<<2)|3,{101,48,32}},
/* 10261 */ {(16<<2)|3,{101,48,37}},
/* 10262 */ {(16<<2)|3,{101,48,45}},
/* 10263 */ {(16<<2)|3,{101,48,46}},
/* 10264 */ {(16<<2)|3,{101,48,47}},
/* 10265 */ {(16<<2)|3,{101,48,51}},
/* 10266 */ {(16<<2)|3,{101,48,52}},
/* 10267 */ {(16<<2)|3,{101,48,53}},
/* 10268 */ {(16<<2)|3,{101,48,54}},
/* 10269 */ {(16<<2)|3,{101,48,55}},
/* 10270 */ {(16<<2)|3,{101,48,56}},
/* 10271 */ {(16<<2)|3,{101,48,57}},
/* 10272 */ {(16<<2)|3,{101,48,61}},
/* 10273 */ {(16<<2)|3,{101,48,65}},
/* 10274 */ {(16<<2)|3,{101,48,95}},
/* 10275 */ {(16<<2)|3,{101,48,98}},
/* 10276 */ {(16<<2)|3,{101,48,100}},
/* 10277 */ {(16<<2)|3,{101,48,102}},
/* 10278 */ {(16<<2)|3,{101,48,103}},
/* 10279 */ {(16<<2)|3,{101,48,104}},
/* 10280 */ {(16<<2)|3,{101,48,108}},
/* 10281 */ {(16<<2)|3,{101,48,109}},
/* 10282 */ {(16<<2)|3,{101,48,110}},
/* 10283 */ {(16<<2)|3,{101,48,112}},
/* 10284 */ {(16<<2)|3,{101,48,114}},
/* 10285 */ {(16<<2)|3,{101,48,117}},
/* 10286 */ {(10<<2)|2,{101,48,0}},
/* 10287 */ {(10<<2)|2,{101,48,0}},
/* 10288 */ {(10<<2)|2,{101,48,0}},
/* 10289 */ {(10<<2)|2,{101,48,0}},
/* 10290 */ {(10<<2)|2,{101,48,0}},
/* 10291 */ {(10<<2)|2,{101,48,0}},
/* 10292 */ {(10<<2)|2,{101,48,0}},
/* 10293 */ {(10<<2)|2,{101,48,0}},
/* 10294 */ {(10<<2)|2,{101,48,0}},
/* 10295 */ {(10<<2)|2,{101,48,0}},
/* 10296 */ {(10<<2)|2,{101,48,0}},
/* 10297 */ {(10<<2)|2,{101,48,0}},
/* 10298 */ {(10<<2)|2,{101,48,0}},
/* 10299 */ {(10<<2)|2,{101,48,0}},
/* 10300 */ {(10<<2)|2,{101,48,0}},
/* 10301 */ {(10<<2)|2,{101,48,0}},
/* 10302 */ {(10<<2)|2,{101,48,0}},
/* 10303 */ {(10<<2)|2,{101,48,0}},
/* 10304 */ {(15<<2)|3,{101,49,48}},
/* 10305 */ {(15<<2)|3,{101,49,48}},
/* 10306 */ {(15<<2)|3,{101,49,49}},
/* 10307 */ {(15<<2)|3,{101,49,49}},
/* 10308 */ {(15<<2)|3,{101,49,50}},
/* 10309 */ {(15<<2)|3,{101,49,50}},
/* 10310 */ {(15<<2)|3,{101,49,97}},
/* 10311 */ {(15<<2)|3,{101,49,97}},
/* 10312 */ {(15<<2)|3,{101,49,99}},
/* 10313 */ {(15<<2)|3,{101,49,99}},
/* 10314 */ {(15<<2)|3,{101,49,101}},
/* 10315 */ {(15<<2)|3,{101,49,101}},
/* 10316 */ {(15<<2)|3,{101,49,105}},
/* 10317 */ {(15<<2)|3,{101,49,105}},
/* 10318 */ {(15<<2)|3,{101,49,111}},
/* 10319 */ {(15<<2)|3,{101,49,111}},
/* 10320 */ {(15<<2)|3,{101,49,115}},
/* 10321 */ {(15<<2)|3,{101,49,115}},
/* 10322 */ {(15<<2)|3,{101,49,116}},
/* 10323 */ {(15<<2)|3,{101,49,116}},
/* 10324 */ {(16<<2)|3,{101,49,32}},
/* 10325 */ {(16<<2)|3,{101,49,37}},
/* 10326 */ {(16<<2)|3,{101,49,45}},
/* 10327 */ {(16<<2)|3,{101,49,46}},
/* 10328 */ {(16<<2)|3,{101,49,47}},
/* 10329 */ {(16<<2)|3,{101,49,51}},
/* 10330 */ {(16<<2)|3,{101,49,52}},
/* 10331 */ {(16<<2)|3,{101,49,53}},
/* 10332 */ {(16<<2)|3,{101,49,54}},
/* 10333 */ {(16<<2)|3,{101,49,55}},
/* 10334 */ {(16<<2)|3,{101,49,56}},
/* 10335 */ {(16<<2)|3,{101,49,57}},
/* 10336 */ {(16<<2)|3,{101,49,61}},
/* 10337 */ {(16<<2)|3,{101,49,65}},
/* 10338 */ {(16<<2)|3,{101,49,95}},
/* 10339 */ {(16<<2)|3,{101,49,98}},
/* 10340 */ {(16<<2)|3,{101,49,100}},
/* 10341 */ {(16<<2)|3,{101,49,102}},
/* 10342 */ {(16<<2)|3,{101,49,103}},
/* 10343 */ {(16<<2)|3,{101,49,104}},
/* 10344 */ {(16<<2)|3,{101,49,108}},
/* 10345 */ {(16<<2)|3,{101,49,109}},
/* 10346 */ {(16<<2)|3,{101,49,110}},
/* 10347 */ {(16<<2)|3,{101,49,112}},
/* 10348 */ {(16<<2)|3,{101,49,114}},
/* 10349 */ {(16<<2)|3,{101,49,117}},
/* 10350 */ {(10<<2)|2,{101,49,0}},
/* 10351 */ {(10<<2)|2,{101,49,0}},
/* 10352 */ {(10<<2)|2,{101,49,0}},
/* 10353 */ {(10<<2)|2,{101,49,0}},
/* 10354 */ {(10<<2)|2,{101,49,0}},
/* 10355 */ {(10<<2)|2,{101,49,0}},
/* 10356 */ {(10<<2)|2,{101,49,0}},
/* 10357 */ {(10<<2)|2,{101,49,0}},
/* 10358 */ {(10<<2)|2,{101,49,0}},
/* 10359 */ {(10<<2)|2,{101,49,0}},
/* 10360 */ {(10<<2)|2,{101,49,0}},
/* 10361 */ {(10<<2)|2,{101,49,0}},
/* 10362 */ {(10<<2)|2,{101,49,0}},
/* 10363 */ {(10<<2)|2,{101,49,0}},
/* 10364 */ {(10<<2)|2,{101,49,0}},
/* 10365 */ {(10<<2)|2,{101,49,0}},
/* 10366 */ {(10<<2)|2,{101,49,0}},
/* 10367 */ {(10<<2)|2,{101,49,0}},
/* 10368 */ {(15<<2)|3,{101,50,48}},
/* 10369 */ {(15<<2)|3,{101,50,48}},
/* 10370 */ {(15<<2)|3,{101,50,49}},
/* 10371 */ {(15<<2)|3,{101,50,49}},
/* 10372 */ {(15<<2)|3,{101,50,50}},
/* 10373 */ {(15<<2)|3,{101,50,50}},
/* 10374 */ {(15<<2)|3,{101,50,97}},
/* 10375 */ {(15<<2)|3,{101,50,97}},
/* 10376 */ {(15<<2)|3,{101,50,99}},
/* 10377 */ {(15<<2)|3,{101,50,99}},
/* 10378 */ {(15<<2)|3,{101,50,101}},
/* 10379 */ {(15<<2)|3,{101,50,101}},
/* 10380 */ {(15<<2)|3,{101,50,105}},
/* 10381 */ {(15<<2)|3,{101,50,105}},
/* 10382 */ {(15<<2)|3,{101,50,111}},
/* 10383 */ {(15<<2)|3,{101,50,111}},
/* 10384 */ {(15<<2)|3,{101,50,115}},
/* 10385 */ {(15<<2)|3,{101,50,115}},
/* 10386 */ {(15<<2)|3,{101,50,116}},
/* 10387 */ {(15<<2)|3,{101,50,116}},
/* 10388 */ {(16<<2)|3,{101,50,32}},
/* 10389 */ {(16<<2)|3,{101,50,37}},
/* 10390 */ {(16<<2)|3,{101,50,45}},
/* 10391 */ {(16<<2)|3,{101,50,46}},
/* 10392 */ {(16<<2)|3,{101,50,47}},
/* 10393 */ {(16<<2)|3,{101,50,51}},
/* 10394 */ {(16<<2)|3,{101,50,52}},
/* 10395 */ {(16<<2)|3,{101,50,53}},
/* 10396 */ {(16<<2)|3,{101,50,54}},
/* 10397 */ {(16<<2)|3,{101,50,55}},
/* 10398 */ {(16<<2)|3,{101,50,56}},
/* 10399 */ {(16<<2)|3,{101,50,57}},
/* 10400 */ {(16<<2)|3,{101,50,61}},
/* 10401 */ {(16<<2)|3,{101,50,65}},
/* 10402 */ {(16<<2)|3,{101,50,95}},
/* 10403 */ {(16<<2)|3,{101,50,98}},
/* 10404 */ {(16<<2)|3,{101,50,100}},
/* 10405 */ {(16<<2)|3,{101,50,102}},
/* 10406 */ {(16<<2)|3,{101,50,103}},
/* 10407 */ {(16<<2)|3,{101,50,104}},
/* 10408 */ {(16<<2)|3,{101,50,108}},
/* 10409 */ {(16<<2)|3,{101,50,109}},
/* 10410 */ {(16<<2)|3,{101,50,110}},
/* 10411 */ {(16<<2)|3,{101,50,112}},
/* 10412 */ {(16<<2)|3,{101,50,114}},
/* 10413 */ {(16<<2)|3,{101,50,117}},
/* 10414 */ {(10<<2)|2,{101,50,0}},
/* 10415 */ {(10<<2)|2,{101,50,0}},
/* 10416 */ {(10<<2)|2,{101,50,0}},
/* 10417 */ {(10<<2)|2,{101,50,0}},
/* 10418 */ {(10<<2)|2,{101,50,0}},
/* 10419 */ {(10<<2)|2,{101,50,0}},
/* 10420 */ {(10<<2)|2,{101,50,0}},
/* 10421 */ {(10<<2)|2,{101,50,0}},
/* 10422 */ {(10<<2)|2,{101,50,0}},
/* 10423 */ {(10<<2)|2,{101,50,0}},
/* 10424 */ {(10<<2)|2,{101,50,0}},
/* 10425 */ {(10<<2)|2,{101,50,0}},
/* 10426 */ {(10<<2)|2,{101,50,0}},
/* 10427 */ {(10<<2)|2,{101,50,0}},
/* 10428 */ {(10<<2)|2,{101,50,0}},
/* 10429 */ {(10<<2)|2,{101,50,0}},
/* 10430 */ {(10<<2)|2,{101,50,0}},
/* 10431 */ {(10<<2)|2,{101,50,0}},
/* 10432 */ {(15<<2)|3,{101,97,48}},
/* 10433 */ {(15<<2)|3,{101,97,48}},
/* 10434 */ {(15<<2)|3,{101,97,49}},
/* 10435 */ {(15<<2)|3,{101,97,49}},
/* 10436 */ {(15<<2)|3,{101,97,50}},
/* 10437 */ {(15<<2)|3,{101,97,50}},
/* 10438 */ {(15<<2)|3,{101,97,97}},
/* 10439 */ {(15<<2)|3,{101,97,97}},
/* 10440 */ {(15<<2)|3,{101,97,99}},
/* 10441 */ {(15<<2)|3,{101,97,99}},
/* 10442 */ {(15<<2)|3,{101,97,101}},
/* 10443 */ {(15<<2)|3,{101,97,101}},
/* 10444 */ {(15<<2)|3,{101,97,105}},
/* 10445 */ {(15<<2)|3,{101,97,105}},
/* 10446 */ {(15<<2)|3,{101,97,111}},
/* 10447 */ {(15<<2)|3,{101,97,111}},
/* 10448 */ {(15<<2)|3,{101,97,115}},
/* 10449 */ {(15<<2)|3,{101,97,115}},
/* 10450 */ {(15<<2)|3,{101,97,116}},
/* 10451 */ {(15<<2)|3,{101,97,116}},
/* 10452 */ {(16<<2)|3,{101,97,32}},
/* 10453 */ {(16<<2)|3,{101,97,37}},
/* 10454 */ {(16<<2)|3,{101,97,45}},
/* 10455 */ {(16<<2)|3,{101,97,46}},
/* 10456 */ {(16<<2)|3,{101,97,47}},
/* 10457 */ {(16<<2)|3,{101,97,51}},
/* 10458 */ {(16<<2)|3,{101,97,52}},
/* 10459 */ {(16<<2)|3,{101,97,53}},
/* 10460 */ {(16<<2)|3,{101,97,54}},
/* 10461 */ {(16<<2)|3,{101,97,55}},
/* 10462 */ {(16<<2)|3,{101,97,56}},
/* 10463 */ {(16<<2)|3,{101,97,57}},
/* 10464 */ {(16<<2)|3,{101,97,61}},
/* 10465 */ {(16<<2)|3,{101,97,65}},
/* 10466 */ {(16<<2)|3,{101,97,95}},
/* 10467 */ {(16<<2)|3,{101,97,98}},
/* 10468 */ {(16<<2)|3,{101,97,100}},
/* 10469 */ {(16<<2)|3,{101,97,102}},
/* 10470 */ {(16<<2)|3,{101,97,103}},
/* 10471 */ {(16<<2)|3,{101,97,104}},
/* 10472 */ {(16<<2)|3,{101,97,108}},
/* 10473 */ {(16<<2)|3,{101,97,109}},
/* 10474 */ {(16<<2)|3,{101,97,110}},
/* 10475 */ {(16<<2)|3,{101,97,112}},
/* 10476 */ {(16<<2)|3,{101,97,114}},
/* 10477 */ {(16<<2)|3,{101,97,117}},
/* 10478 */ {(10<<2)|2,{101,97,0}},
/* 10479 */ {(10<<2)|2,{101,97,0}},
/* 10480 */ {(10<<2)|2,{101,97,0}},
/* 10481 */ {(10<<2)|2,{101,97,0}},
/* 10482 */ {(10<<2)|2,{101,97,0}},
/* 10483 */ {(10<<2)|2,{101,97,0}},
/* 10484 */ {(10<<2)|2,{101,97,0}},
/* 10485 */ {(10<<2)|2,{101,97,0}},
/* 10486 */ {(10<<2)|2,{101,97,0}},
/* 10487 */ {(10<<2)|2,{101,97,0}},
/* 10488 */ {(10<<2)|2,{101,97,0}},
/* 10489 */ {(10<<2)|2,{101,97,0}},
/* 10490 */ {(10<<2)|2,{101,97,0}},
/* 10491 */ {(10<<2)|2,{101,97,0}},
/* 10492 */ {(10<<2)|2,{101,97,0}},
/* 10493 */ {(10<<2)|2,{101,97,0}},
/* 10494 */ {(10<<2)|2,{101,97,0}},
/* 10495 */ {(10<<2)|2,{101,97,0}},
/* 10496 */ {(15<<2)|3,{101,99,48}},
/* 10497 */ {(15<<2)|3,{101,99,48}},
/* 10498 */ {(15<<2)|3,{101,99,49}},
/* 10499 */ {(15<<2)|3,{101,99,49}},
/* 10500 */ {(15<<2)|3,{101,99,50}},
/* 10501 */ {(15<<2)|3,{101,99,50}},
/* 10502 */ {(15<<2)|3,{101,99,97}},
/* 10503 */ {(15<<2)|3,{101,99,97}},
/* 10504 */ {(15<<2)|3,{101,99,99}},
/* 10505 */ {(15<<2)|3,{101,99,99}},
/* 10506 */ {(15<<2)|3,{101,99,101}},
/* 10507 */ {(15<<2)|3,{101,99,101}},
/* 10508 */ {(15<<2)|3,{101,99,105}},
/* 10509 */ {(15<<2)|3,{101,99,105}},
/* 10510 */ {(15<<2)|3,{101,99,111}},
/* 10511 */ {(15<<2)|3,{101,99,111}},
/* 10512 */ {(15<<2)|3,{101,99,115}},
/* 10513 */ {(15<<2)|3,{101,99,115}},
/* 10514 */ {(15<<2)|3,{101,99,116}},
/* 10515 */ {(15<<2)|3,{101,99,116}},
/* 10516 */ {(16<<2)|3,{101,99,32}},
/* 10517 */ {(16<<2)|3,{101,99,37}},
/* 10518 */ {(16<<2)|3,{101,99,45}},
/* 10519 */ {(16<<2)|3,{101,99,46}},
/* 10520 */ {(16<<2)|3,{101,99,47}},
/* 10521 */ {(16<<2)|3,{101,99,51}},
/* 10522 */ {(16<<2)|3,{101,99,52}},
/* 10523 */ {(16<<2)|3,{101,99,53}},
/* 10524 */ {(16<<2)|3,{101,99,54}},
/* 10525 */ {(16<<2)|3,{101,99,55}},
/* 10526 */ {(16<<2)|3,{101,99,56}},
/* 10527 */ {(16<<2)|3,{101,99,57}},
/* 10528 */ {(16<<2)|3,{101,99,61}},
/* 10529 */ {(16<<2)|3,{101,99,65}},
/* 10530 */ {(16<<2)|3,{101,99,95}},
/* 10531 */ {(16<<2)|3,{101,99,98}},
/* 10532 */ {(16<<2)|3,{101,99,100}},
/* 10533 */ {(16<<2)|3,{101,99,102}},
/* 10534 */ {(16<<2)|3,{101,99,103}},
/* 10535 */ {(16<<2)|3,{101,99,104}},
/* 10536 */ {(16<<2)|3,{101,99,108}},
/* 10537 */ {(16<<2)|3,{101,99,109}},
/* 10538 */ {(16<<2)|3,{101,99,110}},
/* 10539 */ {(16<<2)|3,{101,99,112}},
/* 10540 */ {(16<<2)|3,{101,99,114}},
/* 10541 */ {(16<<2)|3,{101,99,117}},
/* 10542 */ {(10<<2)|2,{101,99,0}},
/* 10543 */ {(10<<2)|2,{101,99,0}},
/* 10544 */ {(10<<2)|2,{101,99,0}},
/* 10545 */ {(10<<2)|2,{101,99,0}},
/* 10546 */ {(10<<2)|2,{101,99,0}},
/* 10547 */ {(10<<2)|2,{101,99,0}},
/* 10548 */ {(10<<2)|2,{101,99,0}},
/* 10549 */ {(10<<2)|2,{101,99,0}},
/* 10550 */ {(10<<2)|2,{101,99,0}},
/* 10551 */ {(10<<2)|2,{101,99,0}},
/* 10552 */ {(10<<2)|2,{101,99,0}},
/* 10553 */ {(10<<2)|2,{101,99,0}},
/* 10554 */ {(10<<2)|2,{101,99,0}},
/* 10555 */ {(10<<2)|2,{101,99,0}},
/* 10556 */ {(10<<2)|2,{101,99,0}},
/* 10557 */ {(10<<2)|2,{101,99,0}},
/* 10558 */ {(10<<2)|2,{101,99,0}},
/* 10559 */ {(10<<2)|2,{101,99,0}},
/* 10560 */ {(15<<2)|3,{101,101,48}},
/* 10561 */ {(15<<2)|3,{101,101,48}},
/* 10562 */ {(15<<2)|3,{101,101,49}},
/* 10563 */ {(15<<2)|3,{101,101,49}},
/* 10564 */ {(15<<2)|3,{101,101,50}},
/* 10565 */ {(15<<2)|3,{101,101,50}},
/* 10566 */ {(15<<2)|3,{101,101,97}},
/* 10567 */ {(15<<2)|3,{101,101,97}},
/* 10568 */ {(15<<2)|3,{101,101,99}},
/* 10569 */ {(15<<2)|3,{101,101,99}},
/* 10570 */ {(15<<2)|3,{101,101,101}},
/* 10571 */ {(15<<2)|3,{101,101,101}},
/* 10572 */ {(15<<2)|3,{101,101,105}},
/* 10573 */ {(15<<2)|3,{101,101,105}},
/* 10574 */ {(15<<2)|3,{101,101,111}},
/* 10575 */ {(15<<2)|3,{101,101,111}},
/* 10576 */ {(15<<2)|3,{101,101,115}},
/* 10577 */ {(15<<2)|3,{101,101,115}},
/* 10578 */ {(15<<2)|3,{101,101,116}},
/* 10579 */ {(15<<2)|3,{101,101,116}},
/* 10580 */ {(16<<2)|3,{101,101,32}},
/* 10581 */ {(16<<2)|3,{101,101,37}},
/* 10582 */ {(16<<2)|3,{101,101,45}},
/* 10583 */ {(16<<2)|3,{101,101,46}},
/* 10584 */ {(16<<2)|3,{101,101,47}},
/* 10585 */ {(16<<2)|3,{101,101,51}},
/* 10586 */ {(16<<2)|3,{101,101,52}},
/* 10587 */ {(16<<2)|3,{101,101,53}},
/* 10588 */ {(16<<2)|3,{101,101,54}},
/* 10589 */ {(16<<2)|3,{101,101,55}},
/* 10590 */ {(16<<2)|3,{101,101,56}},
/* 10591 */ {(16<<2)|3,{101,101,57}},
/* 10592 */ {(16<<2)|3,{101,101,61}},
/* 10593 */ {(16<<2)|3,{101,101,65}},
/* 10594 */ {(16<<2)|3,{101,101,95}},
/* 10595 */ {(16<<2)|3,{101,101,98}},
/* 10596 */ {(16<<2)|3,{101,101,100}},
/* 10597 */ {(16<<2)|3,{101,101,102}},
/* 10598 */ {(16<<2)|3,{101,101,103}},
/* 10599 */ {(16<<2)|3,{101,101,104}},
/* 10600 */ {(16<<2)|3,{101,101,108}},
/* 10601 */ {(16<<2)|3,{101,101,109}},
/* 10602 */ {(16<<2)|3,{101,101,110}},
/* 10603 */ {(16<<2)|3,{101,101,112}},
/* 10604 */ {(16<<2)|3,{101,101,114}},
/* 10605 */ {(16<<2)|3,{101,101,117}},
/* 10606 */ {(10<<2)|2,{101,101,0}},
/* 10607 */ {(10<<2)|2,{101,101,0}},
/* 10608 */ {(10<<2)|2,{101,101,0}},
/* 10609 */ {(10<<2)|2,{101,101,0}},
/* 10610 */ {(10<<2)|2,{101,101,0}},
/* 10611 */ {(10<<2)|2,{101,101,0}},
/* 10612 */ {(10<<2)|2,{101,101,0}},
/* 10613 */ {(10<<2)|2,{101,101,0}},
/* 10614 */ {(10<<2)|2,{101,101,0}},
/* 10615 */ {(10<<2)|2,{101,101,0}},
/* 10616 */ {(10<<2)|2,{101,101,0}},
/* 10617 */ {(10<<2)|2,{101,101,0}},
/* 10618 */ {(10<<2)|2,{101,101,0}},
/* 10619 */ {(10<<2)|2,{101,101,0}},
/* 10620 */ {(10<<2)|2,{101,101,0}},
/* 10621 */ {(10<<2)|2,{101,101,0}},
/* 10622 */ {(10<<2)|2,{101,101,0}},
/* 10623 */ {(10<<2)|2,{101,101,0}},
/* 10624 */ {(15<<2)|3,{101,105,48}},
/* 10625 */ {(15<<2)|3,{101,105,48}},
/* 10626 */ {(15<<2)|3,{101,105,49}},
/* 10627 */ {(15<<2)|3,{101,105,49}},
/* 10628 */ {(15<<2)|3,{101,105,50}},
/* 10629 */ {(15<<2)|3,{101,105,50}},
/* 10630 */ {(15<<2)|3,{101,105,97}},
/* 10631 */ {(15<<2)|3,{101,105,97}},
/* 10632 */ {(15<<2)|3,{101,105,99}},
/* 10633 */ {(15<<2)|3,{101,105,99}},
/* 10634 */ {(15<<2)|3,{101,105,101}},
/* 10635 */ {(15<<2)|3,{101,105,101}},
/* 10636 */ {(15<<2)|3,{101,105,105}},
/* 10637 */ {(15<<2)|3,{101,105,105}},
/* 10638 */ {(15<<2)|3,{101,105,111}},
/* 10639 */ {(15<<2)|3,{101,105,111}},
/* 10640 */ {(15<<2)|3,{101,105,115}},
/* 10641 */ {(15<<2)|3,{101,105,115}},
/* 10642 */ {(15<<2)|3,{101,105,116}},
/* 10643 */ {(15<<2)|3,{101,105,116}},
/* 10644 */ {(16<<2)|3,{101,105,32}},
/* 10645 */ {(16<<2)|3,{101,105,37}},
/* 10646 */ {(16<<2)|3,{101,105,45}},
/* 10647 */ {(16<<2)|3,{101,105,46}},
/* 10648 */ {(16<<2)|3,{101,105,47}},
/* 10649 */ {(16<<2)|3,{101,105,51}},
/* 10650 */ {(16<<2)|3,{101,105,52}},
/* 10651 */ {(16<<2)|3,{101,105,53}},
/* 10652 */ {(16<<2)|3,{101,105,54}},
/* 10653 */ {(16<<2)|3,{101,105,55}},
/* 10654 */ {(16<<2)|3,{101,105,56}},
/* 10655 */ {(16<<2)|3,{101,105,57}},
/* 10656 */ {(16<<2)|3,{101,105,61}},
/* 10657 */ {(16<<2)|3,{101,105,65}},
/* 10658 */ {(16<<2)|3,{101,105,95}},
/* 10659 */ {(16<<2)|3,{101,105,98}},
/* 10660 */ {(16<<2)|3,{101,105,100}},
/* 10661 */ {(16<<2)|3,{101,105,102}},
/* 10662 */ {(16<<2)|3,{101,105,103}},
/* 10663 */ {(16<<2)|3,{101,105,104}},
/* 10664 */ {(16<<2)|3,{101,105,108}},
/* 10665 */ {(16<<2)|3,{101,105,109}},
/* 10666 */ {(16<<2)|3,{101,105,110}},
/* 10667 */ {(16<<2)|3,{101,105,112}},
/* 10668 */ {(16<<2)|3,{101,105,114}},
/* 10669 */ {(16<<2)|3,{101,105,117}},
/* 10670 */ {(10<<2)|2,{101,105,0}},
/* 10671 */ {(10<<2)|2,{101,105,0}},
/* 10672 */ {(10<<2)|2,{101,105,0}},
/* 10673 */ {(10<<2)|2,{101,105,0}},
/* 10674 */ {(10<<2)|2,{101,105,0}},
/* 10675 */ {(10<<2)|2,{101,105,0}},
/* 10676 */ {(10<<2)|2,{101,105,0}},
/* 10677 */ {(10<<2)|2,{101,105,0}},
/* 10678 */ {(10<<2)|2,{101,105,0}},
/* 10679 */ {(10<<2)|2,{101,105,0}},
/* 10680 */ {(10<<2)|2,{101,105,0}},
/* 10681 */ {(10<<2)|2,{101,105,0}},
/* 10682 */ {(10<<2)|2,{101,105,0}},
/* 10683 */ {(10<<2)|2,{101,105,0}},
/* 10684 */ {(10<<2)|2,{101,105,0}},
/* 10685 */ {(10<<2)|2,{101,105,0}},
/* 10686 */ {(10<<2)|2,{101,105,0}},
/* 10687 */ {(10<<2)|2,{101,105,0}},
/* 10688 */ {(15<<2)|3,{101,111,48}},
/* 10689 */ {(15<<2)|3,{101,111,48}},
/* 10690 */ {(15<<2)|3,{101,111,49}},
/* 10691 */ {(15<<2)|3,{101,111,49}},
/* 10692 */ {(15<<2)|3,{101,111,50}},
/* 10693 */ {(15<<2)|3,{101,111,50}},
/* 10694 */ {(15<<2)|3,{101,111,97}},
/* 10695 */ {(15<<2)|3,{101,111,97}},
/* 10696 */ {(15<<2)|3,{101,111,99}},
/* 10697 */ {(15<<2)|3,{101,111,99}},
/* 10698 */ {(15<<2)|3,{101,111,101}},
/* 10699 */ {(15<<2)|3,{101,111,101}},
/* 10700 */ {(15<<2)|3,{101,111,105}},
/* 10701 */ {(15<<2)|3,{101,111,105}},
/* 10702 */ {(15<<2)|3,{101,111,111}},
/* 10703 */ {(15<<2)|3,{101,111,111}},
/* 10704 */ {(15<<2)|3,{101,111,115}},
/* 10705 */ {(15<<2)|3,{101,111,115}},
/* 10706 */ {(15<<2)|3,{101,111,116}},
/* 10707 */ {(15<<2)|3,{101,111,116}},
/* 10708 */ {(16<<2)|3,{101,111,32}},
/* 10709 */ {(16<<2)|3,{101,111,37}},
/* 10710 */ {(16<<2)|3,{101,111,45}},
/* 10711 */ {(16<<2)|3,{101,111,46}},
/* 10712 */ {(16<<2)|3,{101,111,47}},
/* 10713 */ {(16<<2)|3,{101,111,51}},
/* 10714 */ {(16<<2)|3,{101,111,52}},
/* 10715 */ {(16<<2)|3,{101,111,53}},
/* 10716 */ {(16<<2)|3,{101,111,54}},
/* 10717 */ {(16<<2)|3,{101,111,55}},
/* 10718 */ {(16<<2)|3,{101,111,56}},
/* 10719 */ {(16<<2)|3,{101,111,57}},
/* 10720 */ {(16<<2)|3,{101,111,61}},
/* 10721 */ {(16<<2)|3,{101,111,65}},
/* 10722 */ {(16<<2)|3,{101,111,95}},
/* 10723 */ {(16<<2)|3,{101,111,98}},
/* 10724 */ {(16<<2)|3,{101,111,100}},
/* 10725 */ {(16<<2)|3,{101,111,102}},
/* 10726 */ {(16<<2)|3,{101,111,103}},
/* 10727 */ {(16<<2)|3,{101,111,104}},
/* 10728 */ {(16<<2)|3,{101,111,108}},
/* 10729 */ {(16<<2)|3,{101,111,109}},
/* 10730 */ {(16<<2)|3,{101,111,110}},
/* 10731 */ {(16<<2)|3,{101,111,112}},
/* 10732 */ {(16<<2)|3,{101,111,114}},
/* 10733 */ {(16<<2)|3,{101,111,117}},
/* 10734 */ {(10<<2)|2,{101,111,0}},
/* 10735 */ {(10<<2)|2,{101,111,0}},
/* 10736 */ {(10<<2)|2,{101,111,0}},
/* 10737 */ {(10<<2)|2,{101,111,0}},
/* 10738 */ {(10<<2)|2,{101,111,0}},
/* 10739 */ {(10<<2)|2,{101,111,0}},
/* 10740 */ {(10<<2)|2,{101,111,0}},
/* 10741 */ {(10<<2)|2,{101,111,0}},
/* 10742 */ {(10<<2)|2,{101,111,0}},
/* 10743 */ {(10<<2)|2,{101,111,0}},
/* 10744 */ {(10<<2)|2,{101,111,0}},
/* 10745 */ {(10<<2)|2,{101,111,0}},
/* 10746 */ {(10<<2)|2,{101,111,0}},
/* 10747 */ {(10<<2)|2,{101,111,0}},
/* 10748 */ {(10<<2)|2,{101,111,0}},
/* 10749 */ {(10<<2)|2,{101,111,0}},
/* 10750 */ {(10<<2)|2,{101,111,0}},
/* 10751 */ {(10<<2)|2,{101,111,0}},
/* 10752 */ {(15<<2)|3,{101,115,48}},
/* 10753 */ {(15<<2)|3,{101,115,48}},
/* 10754 */ {(15<<2)|3,{101,115,49}},
/* 10755 */ {(15<<2)|3,{101,115,49}},
/* 10756 */ {(15<<2)|3,{101,115,50}},
/* 10757 */ {(15<<2)|3,{101,115,50}},
/* 10758 */ {(15<<2)|3,{101,115,97}},
/* 10759 */ {(15<<2)|3,{101,115,97}},
/* 10760 */ {(15<<2)|3,{101,115,99}},
/* 10761 */ {(15<<2)|3,{101,115,99}},
/* 10762 */ {(15<<2)|3,{101,115,101}},
/* 10763 */ {(15<<2)|3,{101,115,101}},
/* 10764 */ {(15<<2)|3,{101,115,105}},
/* 10765 */ {(15<<2)|3,{101,115,105}},
/* 10766 */ {(15<<2)|3,{101,115,111}},
/* 10767 */ {(15<<2)|3,{101,115,111}},
/* 10768 */ {(15<<2)|3,{101,115,115}},
/* 10769 */ {(15<<2)|3,{101,115,115}},
/* 10770 */ {(15<<2)|3,{101,115,116}},
/* 10771 */ {(15<<2)|3,{101,115,116}},
/* 10772 */ {(16<<2)|3,{101,115,32}},
/* 10773 */ {(16<<2)|3,{101,115,37}},
/* 10774 */ {(16<<2)|3,{101,115,45}},
/* 10775 */ {(16<<2)|3,{101,115,46}},
/* 10776 */ {(16<<2)|3,{101,115,47}},
/* 10777 */ {(16<<2)|3,{101,115,51}},
/* 10778 */ {(16<<2)|3,{101,115,52}},
/* 10779 */ {(16<<2)|3,{101,115,53}},
/* 10780 */ {(16<<2)|3,{101,115,54}},
/* 10781 */ {(16<<2)|3,{101,115,55}},
/* 10782 */ {(16<<2)|3,{101,115,56}},
/* 10783 */ {(16<<2)|3,{101,115,57}},
/* 10784 */ {(16<<2)|3,{101,115,61}},
/* 10785 */ {(16<<2)|3,{101,115,65}},
/* 10786 */ {(16<<2)|3,{101,115,95}},
/* 10787 */ {(16<<2)|3,{101,115,98}},
/* 10788 */ {(16<<2)|3,{101,115,100}},
/* 10789 */ {(16<<2)|3,{101,115,102}},
/* 10790 */ {(16<<2)|3,{101,115,103}},
/* 10791 */ {(16<<2)|3,{101,115,104}},
/* 10792 */ {(16<<2)|3,{101,115,108}},
/* 10793 */ {(16<<2)|3,{101,115,109}},
/* 10794 */ {(16<<2)|3,{101,115,110}},
/* 10795 */ {(16<<2)|3,{101,115,112}},
/* 10796 */ {(16<<2)|3,{101,115,114}},
/* 10797 */ {(16<<2)|3,{101,115,117}},
/* 10798 */ {(10<<2)|2,{101,115,0}},
/* 10799 */ {(10<<2)|2,{101,115,0}},
/* 10800 */ {(10<<2)|2,{101,115,0}},
/* 10801 */ {(10<<2)|2,{101,115,0}},
/* 10802 */ {(10<<2)|2,{101,115,0}},
/* 10803 */ {(10<<2)|2,{101,115,0}},
/* 10804 */ {(10<<2)|2,{101,115,0}},
/* 10805 */ {(10<<2)|2,{101,115,0}},
/* 10806 */ {(10<<2)|2,{101,115,0}},
/* 10807 */ {(10<<2)|2,{101,115,0}},
/* 10808 */ {(10<<2)|2,{101,115,0}},
/* 10809 */ {(10<<2)|2,{101,115,0}},
/* 10810 */ {(10<<2)|2,{101,115,0}},
/* 10811 */ {(10<<2)|2,{101,115,0}},
/* 10812 */ {(10<<2)|2,{101,115,0}},
/* 10813 */ {(10<<2)|2,{101,115,0}},
/* 10814 */ {(10<<2)|2,{101,115,0}},
/* 10815 */ {(10<<2)|2,{101,115,0}},
/* 10816 */ {(15<<2)|3,{101,116,48}},
/* 10817 */ {(15<<2)|3,{101,116,48}},
/* 10818 */ {(15<<2)|3,{101,116,49}},
/* 10819 */ {(15<<2)|3,{101,116,49}},
/* 10820 */ {(15<<2)|3,{101,116,50}},
/* 10821 */ {(15<<2)|3,{101,116,50}},
/* 10822 */ {(15<<2)|3,{101,116,97}},
/* 10823 */ {(15<<2)|3,{101,116,97}},
/* 10824 */ {(15<<2)|3,{101,116,99}},
/* 10825 */ {(15<<2)|3,{101,116,99}},
/* 10826 */ {(15<<2)|3,{101,116,101}},
/* 10827 */ {(15<<2)|3,{101,116,101}},
/* 10828 */ {(15<<2)|3,{101,116,105}},
/* 10829 */ {(15<<2)|3,{101,116,105}},
/* 10830 */ {(15<<2)|3,{101,116,111}},
/* 10831 */ {(15<<2)|3,{101,116,111}},
/* 10832 */ {(15<<2)|3,{101,116,115}},
/* 10833 */ {(15<<2)|3,{101,116,115}},
/* 10834 */ {(15<<2)|3,{101,116,116}},
/* 10835 */ {(15<<2)|3,{101,116,116}},
/* 10836 */ {(16<<2)|3,{101,116,32}},
/* 10837 */ {(16<<2)|3,{101,116,37}},
/* 10838 */ {(16<<2)|3,{101,116,45}},
/* 10839 */ {(16<<2)|3,{101,116,46}},
/* 10840 */ {(16<<2)|3,{101,116,47}},
/* 10841 */ {(16<<2)|3,{101,116,51}},
/* 10842 */ {(16<<2)|3,{101,116,52}},
/* 10843 */ {(16<<2)|3,{101,116,53}},
/* 10844 */ {(16<<2)|3,{101,116,54}},
/* 10845 */ {(16<<2)|3,{101,116,55}},
/* 10846 */ {(16<<2)|3,{101,116,56}},
/* 10847 */ {(16<<2)|3,{101,116,57}},
/* 10848 */ {(16<<2)|3,{101,116,61}},
/* 10849 */ {(16<<2)|3,{101,116,65}},
/* 10850 */ {(16<<2)|3,{101,116,95}},
/* 10851 */ {(16<<2)|3,{101,116,98}},
/* 10852 */ {(16<<2)|3,{101,116,100}},
/* 10853 */ {(16<<2)|3,{101,116,102}},
/* 10854 */ {(16<<2)|3,{101,116,103}},
/* 10855 */ {(16<<2)|3,{101,116,104}},
/* 10856 */ {(16<<2)|3,{101,116,108}},
/* 10857 */ {(16<<2)|3,{101,116,109}},
/* 10858 */ {(16<<2)|3,{101,116,110}},
/* 10859 */ {(16<<2)|3,{101,116,112}},
/* 10860 */ {(16<<2)|3,{101,116,114}},
/* 10861 */ {(16<<2)|3,{101,116,117}},
/* 10862 */ {(10<<2)|2,{101,116,0}},
/* 10863 */ {(10<<2)|2,{101,116,0}},
/* 10864 */ {(10<<2)|2,{101,116,0}},
/* 10865 */ {(10<<2)|2,{101,116,0}},
/* 10866 */ {(10<<2)|2,{101,116,0}},
/* 10867 */ {(10<<2)|2,{101,116,0}},
/* 10868 */ {(10<<2)|2,{101,116,0}},
/* 10869 */ {(10<<2)|2,{101,116,0}},
/* 10870 */ {(10<<2)|2,{101,116,0}},
/* 10871 */ {(10<<2)|2,{101,116,0}},
/* 10872 */ {(10<<2)|2,{101,116,0}},
/* 10873 */ {(10<<2)|2,{101,116,0}},
/* 10874 */ {(10<<2)|2,{101,116,0}},
/* 10875 */ {(10<<2)|2,{101,116,0}},
/* 10876 */ {(10<<2)|2,{101,116,0}},
/* 10877 */ {(10<<2)|2,{101,116,0}},
/* 10878 */ {(10<<2)|2,{101,116,0}},
/* 10879 */ {(10<<2)|2,{101,116,0}},
/* 10880 */ {(16<<2)|3,{101,32,48}},
/* 10881 */ {(16<<2)|3,{101,32,49}},
/* 10882 */ {(16<<2)|3,{101,32,50}},
/* 10883 */ {(16<<2)|3,{101,32,97}},
/* 10884 */ {(16<<2)|3,{101,32,99}},
/* 10885 */ {(16<<2)|3,{101,32,101}},
/* 10886 */ {(16<<2)|3,{101,32,105}},
/* 10887 */ {(16<<2)|3,{101,32,111}},
/* 10888 */ {(16<<2)|3,{101,32,115}},
/* 10889 */ {(16<<2)|3,{101,32,116}},
/* 10890 */ {(11<<2)|2,{101,32,0}},
/* 10891 */ {(11<<2)|2,{101,32,0}},
/* 10892 */ {(11<<2)|2,{101,32,0}},
/* 10893 */ {(11<<2)|2,{101,32,0}},
/* 10894 */ {(11<<2)|2,{101,32,0}},
/* 10895 */ {(11<<2)|2,{101,32,0}},
/* 10896 */ {(11<<2)|2,{101,32,0}},
/* 10897 */ {(11<<2)|2,{101,32,0}},
/* 10898 */ {(11<<2)|2,{101,32,0}},
/* 10899 */ {(11<<2)|2,{101,32,0}},
/* 10900 */ {(11<<2)|2,{101,32,0}},
/* 10901 */ {(11<<2)|2,{101,32,0}},
/* 10902 */ {(11<<2)|2,{101,32,0}},
/* 10903 */ {(11<<2)|2,{101,32,0}},
/* 10904 */ {(11<<2)|2,{101,32,0}},
/* 10905 */ {(11<<2)|2,{101,32,0}},
/* 10906 */ {(11<<2)|2,{101,32,0}},
/* 10907 */ {(11<<2)|2,{101,32,0}},
/* 10908 */ {(11<<2)|2,{101,32,0}},
/* 10909 */ {(11<<2)|2,{101,32,0}},
/* 10910 */ {(11<<2)|2,{101,32,0}},
/* 10911 */ {(11<<2)|2,{101,32,0}},
/* 10912 */ {(16<<2)|3,{101,37,48}},
/* 10913 */ {(16<<2)|3,{101,37,49}},
/* 10914 */ {(16<<2)|3,{101,37,50}},
/* 10915 */ {(16<<2)|3,{101,37,97}},
/* 10916 */ {(16<<2)|3,{101,37,99}},
/* 10917 */ {(16<<2)|3,{101,37,101}},
/* 10918 */ {(16<<2)|3,{101,37,105}},
/* 10919 */ {(16<<2)|3,{101,37,111}},
/* 10920 */ {(16<<2)|3,{101,37,115}},
/* 10921 */ {(16<<2)|3,{101,37,116}},
/* 10922 */ {(11<<2)|2,{101,37,0}},
/* 10923 */ {(11<<2)|2,{101,37,0}},
/* 10924 */ {(11<<2)|2,{101,37,0}},
/* 10925 */ {(11<<2)|2,{101,37,0}},
/* 10926 */ {(11<<2)|2,{101,37,0}},
/* 10927 */ {(11<<2)|2,{101,37,0}},
/* 10928 */ {(11<<2)|2,{101,37,0}},
/* 10929 */ {(11<<2)|2,{101,37,0}},
/* 10930 */ {(11<<2)|2,{101,37,0}},
/* 10931 */ {(11<<2)|2,{101,37,0}},
/* 10932 */ {(11<<2)|2,{101,37,0}},
/* 10933 */ {(11<<2)|2,{101,37,0}},
/* 10934 */ {(11<<2)|2,{101,37,0}},
/* 10935 */ {(11<<2)|2,{101,37,0}},
/* 10936 */ {(11<<2)|2,{101,37,0}},
/* 10937 */ {(11<<2)|2,{101,37,0}},
/* 10938 */ {(11<<2)|2,{101,37,0}},
/* 10939 */ {(11<<2)|2,{101,37,0}},
/* 10940 */ {(11<<2)|2,{101,37,0}},
/* 10941 */ {(11<<2)|2,{101,37,0}},
/* 10942 */ {(11<<2)|2,{101,37,0}},
/* 10943 */ {(11<<2)|2,{101,37,0}},
/* 10944 */ {(16<<2)|3,{101,45,48}},
/* 10945 */ {(16<<2)|3,{101,45,49}},
/* 10946 */ {(16<<2)|3,{101,45,50}},
/* 10947 */ {(16<<2)|3,{101,45,97}},
/* 10948 */ {(16<<2)|3,{101,45,99}},
/* 10949 */ {(16<<2)|3,{101,45,101}},
/* 10950 */ {(16<<2)|3,{101,45,105}},
/* 10951 */ {(16<<2)|3,{101,45,111}},
/* 10952 */ {(16<<2)|3,{101,45,115}},
/* 10953 */ {(16<<2)|3,{101,45,116}},
/* 10954 */ {(11<<2)|2,{101,45,0}},
/* 10955 */ {(11<<2)|2,{101,45,0}},
/* 10956 */ {(11<<2)|2,{101,45,0}},
/* 10957 */ {(11<<2)|2,{101,45,0}},
/* 10958 */ {(11<<2)|2,{101,45,0}},
/* 10959 */ {(11<<2)|2,{101,45,0}},
/* 10960 */ {(11<<2)|2,{101,45,0}},
/* 10961 */ {(11<<2)|2,{101,45,0}},
/* 10962 */ {(11<<2)|2,{101,45,0}},
/* 10963 */ {(11<<2)|2,{101,45,0}},
/* 10964 */ {(11<<2)|2,{101,45,0}},
/* 10965 */ {(11<<2)|2,{101,45,0}},
/* 10966 */ {(11<<2)|2,{101,45,0}},
/* 10967 */ {(11<<2)|2,{101,45,0}},
/* 10968 */ {(11<<2)|2,{101,45,0}},
/* 10969 */ {(11<<2)|2,{101,45,0}},
/* 10970 */ {(11<<2)|2,{101,45,0}},
/* 10971 */ {(11<<2)|2,{101,45,0}},
/* 10972 */ {(11<<2)|2,{101,45,0}},
/* 10973 */ {(11<<2)|2,{101,45,0}},
/* 10974 */ {(11<<2)|2,{101,45,0}},
/* 10975 */ {(11<<2)|2,{101,45,0}},
/* 10976 */ {(16<<2)|3,{101,46,48}},
/* 10977 */ {(16<<2)|3,{101,46,49}},
/* 10978 */ {(16<<2)|3,{101,46,50}},
/* 10979 */ {(16<<2)|3,{101,46,97}},
/* 10980 */ {(16<<2)|3,{101,46,99}},
/* 10981 */ {(16<<2)|3,{101,46,101}},
/* 10982 */ {(16<<2)|3,{101,46,105}},
/* 10983 */ {(16<<2)|3,{101,46,111}},
/* 10984 */ {(16<<2)|3,{101,46,115}},
/* 10985 */ {(16<<2)|3,{101,46,116}},
/* 10986 */ {(11<<2)|2,{101,46,0}},
/* 10987 */ {(11<<2)|2,{101,46,0}},
/* 10988 */ {(11<<2)|2,{101,46,0}},
/* 10989 */ {(11<<2)|2,{101,46,0}},
/* 10990 */ {(11<<2)|2,{101,46,0}},
/* 10991 */ {(11<<2)|2,{101,46,0}},
/* 10992 */ {(11<<2)|2,{101,46,0}},
/* 10993 */ {(11<<2)|2,{101,46,0}},
/* 10994 */ {(11<<2)|2,{101,46,0}},
/* 10995 */ {(11<<2)|2,{101,46,0}},
/* 10996 */ {(11<<2)|2,{101,46,0}},
/* 10997 */ {(11<<2)|2,{101,46,0}},
/* 10998 */ {(11<<2)|2,{101,46,0}},
/* 10999 */ {(11<<2)|2,{101,46,0}},
/* 11000 */ {(11<<2)|2,{101,46,0}},
/* 11001 */ {(11<<2)|2,{101,46,0}},
/* 11002 */ {(11<<2)|2,{101,46,0}},
/* 11003 */ {(11<<2)|2,{101,46,0}},
/* 11004 */ {(11<<2)|2,{101,46,0}},
/* 11005 */ {(11<<2)|2,{101,46,0}},
/* 11006 */ {(11<<2)|2,{101,46,0}},
/* 11007 */ {(11<<2)|2,{101,46,0}},
/* 11008 */ {(16<<2)|3,{101,47,48}},
/* 11009 */ {(16<<2)|3,{101,47,49}},
/* 11010 */ {(16<<2)|3,{101,47,50}},
/* 11011 */ {(16<<2)|3,{101,47,97}},
/* 11012 */ {(16<<2)|3,{101,47,99}},
/* 11013 */ {(16<<2)|3,{101,47,101}},
/* 11014 */ {(16<<2)|3,{101,47,105}},
/* 11015 */ {(16<<2)|3,{101,47,111}},
/* 11016 */ {(16<<2)|3,{101,47,115}},
/* 11017 */ {(16<<2)|3,{101,47,116}},
/* 11018 */ {(11<<2)|2,{101,47,0}},
/* 11019 */ {(11<<2)|2,{101,47,0}},
/* 11020 */ {(11<<2)|2,{101,47,0}},
/* 11021 */ {(11<<2)|2,{101,47,0}},
/* 11022 */ {(11<<2)|2,{101,47,0}},
/* 11023 */ {(11<<2)|2,{101,47,0}},
/* 11024 */ {(11<<2)|2,{101,47,0}},
/* 11025 */ {(11<<2)|2,{101,47,0}},
/* 11026 */ {(11<<2)|2,{101,47,0}},
/* 11027 */ {(11<<2)|2,{101,47,0}},
/* 11028 */ {(11<<2)|2,{101,47,0}},
/* 11029 */ {(11<<2)|2,{101,47,0}},
/* 11030 */ {(11<<2)|2,{101,47,0}},
/* 11031 */ {(11<<2)|2,{101,47,0}},
/* 11032 */ {(11<<2)|2,{101,47,0}},
/* 11033 */ {(11<<2)|2,{101,47,0}},
/* 11034 */ {(11<<2)|2,{101,47,0}},
/* 11035 */ {(11<<2)|2,{101,47,0}},
/* 11036 */ {(11<<2)|2,{101,47,0}},
/* 11037 */ {(11<<2)|2,{101,47,0}},
/* 11038 */ {(11<<2)|2,{101,47,0}},
/* 11039 */ {(11<<2)|2,{101,47,0}},
/* 11040 */ {(16<<2)|3,{101,51,48}},
/* 11041 */ {(16<<2)|3,{101,51,49}},
/* 11042 */ {(16<<2)|3,{101,51,50}},
/* 11043 */ {(16<<2)|3,{101,51,97}},
/* 11044 */ {(16<<2)|3,{101,51,99}},
/* 11045 */ {(16<<2)|3,{101,51,101}},
/* 11046 */ {(16<<2)|3,{101,51,105}},
/* 11047 */ {(16<<2)|3,{101,51,111}},
/* 11048 */ {(16<<2)|3,{101,51,115}},
/* 11049 */ {(16<<2)|3,{101,51,116}},
/* 11050 */ {(11<<2)|2,{101,51,0}},
/* 11051 */ {(11<<2)|2,{101,51,0}},
/* 11052 */ {(11<<2)|2,{101,51,0}},
/* 11053 */ {(11<<2)|2,{101,51,0}},
/* 11054 */ {(11<<2)|2,{101,51,0}},
/* 11055 */ {(11<<2)|2,{101,51,0}},
/* 11056 */ {(11<<2)|2,{101,51,0}},
/* 11057 */ {(11<<2)|2,{101,51,0}},
/* 11058 */ {(11<<2)|2,{101,51,0}},
/* 11059 */ {(11<<2)|2,{101,51,0}},
/* 11060 */ {(11<<2)|2,{101,51,0}},
/* 11061 */ {(11<<2)|2,{101,51,0}},
/* 11062 */ {(11<<2)|2,{101,51,0}},
/* 11063 */ {(11<<2)|2,{101,51,0}},
/* 11064 */ {(11<<2)|2,{101,51,0}},
/* 11065 */ {(11<<2)|2,{101,51,0}},
/* 11066 */ {(11<<2)|2,{101,51,0}},
/* 11067 */ {(11<<2)|2,{101,51,0}},
/* 11068 */ {(11<<2)|2,{101,51,0}},
/* 11069 */ {(11<<2)|2,{101,51,0}},
/* 11070 */ {(11<<2)|2,{101,51,0}},
/* 11071 */ {(11<<2)|2,{101,51,0}},
/* 11072 */ {(16<<2)|3,{101,52,48}},
/* 11073 */ {(16<<2)|3,{101,52,49}},
/* 11074 */ {(16<<2)|3,{101,52,50}},
/* 11075 */ {(16<<2)|3,{101,52,97}},
/* 11076 */ {(16<<2)|3,{101,52,99}},
/* 11077 */ {(16<<2)|3,{101,52,101}},
/* 11078 */ {(16<<2)|3,{101,52,105}},
/* 11079 */ {(16<<2)|3,{101,52,111}},
/* 11080 */ {(16<<2)|3,{101,52,115}},
/* 11081 */ {(16<<2)|3,{101,52,116}},
/* 11082 */ {(11<<2)|2,{101,52,0}},
/* 11083 */ {(11<<2)|2,{101,52,0}},
/* 11084 */ {(11<<2)|2,{101,52,0}},
/* 11085 */ {(11<<2)|2,{101,52,0}},
/* 11086 */ {(11<<2)|2,{101,52,0}},
/* 11087 */ {(11<<2)|2,{101,52,0}},
/* 11088 */ {(11<<2)|2,{101,52,0}},
/* 11089 */ {(11<<2)|2,{101,52,0}},
/* 11090 */ {(11<<2)|2,{101,52,0}},
/* 11091 */ {(11<<2)|2,{101,52,0}},
/* 11092 */ {(11<<2)|2,{101,52,0}},
/* 11093 */ {(11<<2)|2,{101,52,0}},
/* 11094 */ {(11<<2)|2,{101,52,0}},
/* 11095 */ {(11<<2)|2,{101,52,0}},
/* 11096 */ {(11<<2)|2,{101,52,0}},
/* 11097 */ {(11<<2)|2,{101,52,0}},
/* 11098 */ {(11<<2)|2,{101,52,0}},
/* 11099 */ {(11<<2)|2,{101,52,0}},
/* 11100 */ {(11<<2)|2,{101,52,0}},
/* 11101 */ {(11<<2)|2,{101,52,0}},
/* 11102 */ {(11<<2)|2,{101,52,0}},
/* 11103 */ {(11<<2)|2,{101,52,0}},
/* 11104 */ {(16<<2)|3,{101,53,48}},
/* 11105 */ {(16<<2)|3,{101,53,49}},
/* 11106 */ {(16<<2)|3,{101,53,50}},
/* 11107 */ {(16<<2)|3,{101,53,97}},
/* 11108 */ {(16<<2)|3,{101,53,99}},
/* 11109 */ {(16<<2)|3,{101,53,101}},
/* 11110 */ {(16<<2)|3,{101,53,105}},
/* 11111 */ {(16<<2)|3,{101,53,111}},
/* 11112 */ {(16<<2)|3,{101,53,115}},
/* 11113 */ {(16<<2)|3,{101,53,116}},
/* 11114 */ {(11<<2)|2,{101,53,0}},
/* 11115 */ {(11<<2)|2,{101,53,0}},
/* 11116 */ {(11<<2)|2,{101,53,0}},
/* 11117 */ {(11<<2)|2,{101,53,0}},
/* 11118 */ {(11<<2)|2,{101,53,0}},
/* 11119 */ {(11<<2)|2,{101,53,0}},
/* 11120 */ {(11<<2)|2,{101,53,0}},
/* 11121 */ {(11<<2)|2,{101,53,0}},
/* 11122 */ {(11<<2)|2,{101,53,0}},
/* 11123 */ {(11<<2)|2,{101,53,0}},
/* 11124 */ {(11<<2)|2,{101,53,0}},
/* 11125 */ {(11<<2)|2,{101,53,0}},
/* 11126 */ {(11<<2)|2,{101,53,0}},
/* 11127 */ {(11<<2)|2,{101,53,0}},
/* 11128 */ {(11<<2)|2,{101,53,0}},
/* 11129 */ {(11<<2)|2,{101,53,0}},
/* 11130 */ {(11<<2)|2,{101,53,0}},
/* 11131 */ {(11<<2)|2,{101,53,0}},
/* 11132 */ {(11<<2)|2,{101,53,0}},
/* 11133 */ {(11<<2)|2,{101,53,0}},
/* 11134 */ {(11<<2)|2,{101,53,0}},
/* 11135 */ {(11<<2)|2,{101,53,0}},
/* 11136 */ {(16<<2)|3,{101,54,48}},
/* 11137 */ {(16<<2)|3,{101,54,49}},
/* 11138 */ {(16<<2)|3,{101,54,50}},
/* 11139 */ {(16<<2)|3,{101,54,97}},
/* 11140 */ {(16<<2)|3,{101,54,99}},
/* 11141 */ {(16<<2)|3,{101,54,101}},
/* 11142 */ {(16<<2)|3,{101,54,105}},
/* 11143 */ {(16<<2)|3,{101,54,111}},
/* 11144 */ {(16<<2)|3,{101,54,115}},
/* 11145 */ {(16<<2)|3,{101,54,116}},
/* 11146 */ {(11<<2)|2,{101,54,0}},
/* 11147 */ {(11<<2)|2,{101,54,0}},
/* 11148 */ {(11<<2)|2,{101,54,0}},
/* 11149 */ {(11<<2)|2,{101,54,0}},
/* 11150 */ {(11<<2)|2,{101,54,0}},
/* 11151 */ {(11<<2)|2,{101,54,0}},
/* 11152 */ {(11<<2)|2,{101,54,0}},
/* 11153 */ {(11<<2)|2,{101,54,0}},
/* 11154 */ {(11<<2)|2,{101,54,0}},
/* 11155 */ {(11<<2)|2,{101,54,0}},
/* 11156 */ {(11<<2)|2,{101,54,0}},
/* 11157 */ {(11<<2)|2,{101,54,0}},
/* 11158 */ {(11<<2)|2,{101,54,0}},
/* 11159 */ {(11<<2)|2,{101,54,0}},
/* 11160 */ {(11<<2)|2,{101,54,0}},
/* 11161 */ {(11<<2)|2,{101,54,0}},
/* 11162 */ {(11<<2)|2,{101,54,0}},
/* 11163 */ {(11<<2)|2,{101,54,0}},
/* 11164 */ {(11<<2)|2,{101,54,0}},
/* 11165 */ {(11<<2)|2,{101,54,0}},
/* 11166 */ {(11<<2)|2,{101,54,0}},
/* 11167 */ {(11<<2)|2,{101,54,0}},
/* 11168 */ {(16<<2)|3,{101,55,48}},
/* 11169 */ {(16<<2)|3,{101,55,49}},
/* 11170 */ {(16<<2)|3,{101,55,50}},
/* 11171 */ {(16<<2)|3,{101,55,97}},
/* 11172 */ {(16<<2)|3,{101,55,99}},
/* 11173 */ {(16<<2)|3,{101,55,101}},
/* 11174 */ {(16<<2)|3,{101,55,105}},
/* 11175 */ {(16<<2)|3,{101,55,111}},
/* 11176 */ {(16<<2)|3,{101,55,115}},
/* 11177 */ {(16<<2)|3,{101,55,116}},
/* 11178 */ {(11<<2)|2,{101,55,0}},
/* 11179 */ {(11<<2)|2,{101,55,0}},
/* 11180 */ {(11<<2)|2,{101,55,0}},
/* 11181 */ {(11<<2)|2,{101,55,0}},
/* 11182 */ {(11<<2)|2,{101,55,0}},
/* 11183 */ {(11<<2)|2,{101,55,0}},
/* 11184 */ {(11<<2)|2,{101,55,0}},
/* 11185 */ {(11<<2)|2,{101,55,0}},
/* 11186 */ {(11<<2)|2,{101,55,0}},
/* 11187 */ {(11<<2)|2,{101,55,0}},
/* 11188 */ {(11<<2)|2,{101,55,0}},
/* 11189 */ {(11<<2)|2,{101,55,0}},
/* 11190 */ {(11<<2)|2,{101,55,0}},
/* 11191 */ {(11<<2)|2,{101,55,0}},
/* 11192 */ {(11<<2)|2,{101,55,0}},
/* 11193 */ {(11<<2)|2,{101,55,0}},
/* 11194 */ {(11<<2)|2,{101,55,0}},
/* 11195 */ {(11<<2)|2,{101,55,0}},
/* 11196 */ {(11<<2)|2,{101,55,0}},
/* 11197 */ {(11<<2)|2,{101,55,0}},
/* 11198 */ {(11<<2)|2,{101,55,0}},
/* 11199 */ {(11<<2)|2,{101,55,0}},
/* 11200 */ {(16<<2)|3,{101,56,48}},
/* 11201 */ {(16<<2)|3,{101,56,49}},
/* 11202 */ {(16<<2)|3,{101,56,50}},
/* 11203 */ {(16<<2)|3,{101,56,97}},
/* 11204 */ {(16<<2)|3,{101,56,99}},
/* 11205 */ {(16<<2)|3,{101,56,101}},
/* 11206 */ {(16<<2)|3,{101,56,105}},
/* 11207 */ {(16<<2)|3,{101,56,111}},
/* 11208 */ {(16<<2)|3,{101,56,115}},
/* 11209 */ {(16<<2)|3,{101,56,116}},
/* 11210 */ {(11<<2)|2,{101,56,0}},
/* 11211 */ {(11<<2)|2,{101,56,0}},
/* 11212 */ {(11<<2)|2,{101,56,0}},
/* 11213 */ {(11<<2)|2,{101,56,0}},
/* 11214 */ {(11<<2)|2,{101,56,0}},
/* 11215 */ {(11<<2)|2,{101,56,0}},
/* 11216 */ {(11<<2)|2,{101,56,0}},
/* 11217 */ {(11<<2)|2,{101,56,0}},
/* 11218 */ {(11<<2)|2,{101,56,0}},
/* 11219 */ {(11<<2)|2,{101,56,0}},
/* 11220 */ {(11<<2)|2,{101,56,0}},
/* 11221 */ {(11<<2)|2,{101,56,0}},
/* 11222 */ {(11<<2)|2,{101,56,0}},
/* 11223 */ {(11<<2)|2,{101,56,0}},
/* 11224 */ {(11<<2)|2,{101,56,0}},
/* 11225 */ {(11<<2)|2,{101,56,0}},
/* 11226 */ {(11<<2)|2,{101,56,0}},
/* 11227 */ {(11<<2)|2,{101,56,0}},
/* 11228 */ {(11<<2)|2,{101,56,0}},
/* 11229 */ {(11<<2)|2,{101,56,0}},
/* 11230 */ {(11<<2)|2,{101,56,0}},
/* 11231 */ {(11<<2)|2,{101,56,0}},
/* 11232 */ {(16<<2)|3,{101,57,48}},
/* 11233 */ {(16<<2)|3,{101,57,49}},
/* 11234 */ {(16<<2)|3,{101,57,50}},
/* 11235 */ {(16<<2)|3,{101,57,97}},
/* 11236 */ {(16<<2)|3,{101,57,99}},
/* 11237 */ {(16<<2)|3,{101,57,101}},
/* 11238 */ {(16<<2)|3,{101,57,105}},
/* 11239 */ {(16<<2)|3,{101,57,111}},
/* 11240 */ {(16<<2)|3,{101,57,115}},
/* 11241 */ {(16<<2)|3,{101,57,116}},
/* 11242 */ {(11<<2)|2,{101,57,0}},
/* 11243 */ {(11<<2)|2,{101,57,0}},
/* 11244 */ {(11<<2)|2,{101,57,0}},
/* 11245 */ {(11<<2)|2,{101,57,0}},
/* 11246 */ {(11<<2)|2,{101,57,0}},
/* 11247 */ {(11<<2)|2,{101,57,0}},
/* 11248 */ {(11<<2)|2,{101,57,0}},
/* 11249 */ {(11<<2)|2,{101,57,0}},
/* 11250 */ {(11<<2)|2,{101,57,0}},
/* 11251 */ {(11<<2)|2,{101,57,0}},
/* 11252 */ {(11<<2)|2,{101,57,0}},
/* 11253 */ {(11<<2)|2,{101,57,0}},
/* 11254 */ {(11<<2)|2,{101,57,0}},
/* 11255 */ {(11<<2)|2,{101,57,0}},
/* 11256 */ {(11<<2)|2,{101,57,0}},
/* 11257 */ {(11<<2)|2,{101,57,0}},
/* 11258 */ {(11<<2)|2,{101,57,0}},
/* 11259 */ {(11<<2)|2,{101,57,0}},
/* 11260 */ {(11<<2)|2,{101,57,0}},
/* 11261 */ {(11<<2)|2,{101,57,0}},
/* 11262 */ {(11<<2)|2,{101,57,0}},
/* 11263 */ {(11<<2)|2,{101,57,0}},
/* 11264 */ {(16<<2)|3,{101,61,48}},
/* 11265 */ {(16<<2)|3,{101,61,49}},
/* 11266 */ {(16<<2)|3,{101,61,50}},
/* 11267 */ {(16<<2)|3,{101,61,97}},
/* 11268 */ {(16<<2)|3,{101,61,99}},
/* 11269 */ {(16<<2)|3,{101,61,101}},
/* 11270 */ {(16<<2)|3,{101,61,105}},
/* 11271 */ {(16<<2)|3,{101,61,111}},
/* 11272 */ {(16<<2)|3,{101,61,115}},
/* 11273 */ {(16<<2)|3,{101,61,116}},
/* 11274 */ {(11<<2)|2,{101,61,0}},
/* 11275 */ {(11<<2)|2,{101,61,0}},
/* 11276 */ {(11<<2)|2,{101,61,0}},
/* 11277 */ {(11<<2)|2,{101,61,0}},
/* 11278 */ {(11<<2)|2,{101,61,0}},
/* 11279 */ {(11<<2)|2,{101,61,0}},
/* 11280 */ {(11<<2)|2,{101,61,0}},
/* 11281 */ {(11<<2)|2,{101,61,0}},
/* 11282 */ {(11<<2)|2,{101,61,0}},
/* 11283 */ {(11<<2)|2,{101,61,0}},
/* 11284 */ {(11<<2)|2,{101,61,0}},
/* 11285 */ {(11<<2)|2,{101,61,0}},
/* 11286 */ {(11<<2)|2,{101,61,0}},
/* 11287 */ {(11<<2)|2,{101,61,0}},
/* 11288 */ {(11<<2)|2,{101,61,0}},
/* 11289 */ {(11<<2)|2,{101,61,0}},
/* 11290 */ {(11<<2)|2,{101,61,0}},
/* 11291 */ {(11<<2)|2,{101,61,0}},
/* 11292 */ {(11<<2)|2,{101,61,0}},
/* 11293 */ {(11<<2)|2,{101,61,0}},
/* 11294 */ {(11<<2)|2,{101,61,0}},
/* 11295 */ {(11<<2)|2,{101,61,0}},
/* 11296 */ {(16<<2)|3,{101,65,48}},
/* 11297 */ {(16<<2)|3,{101,65,49}},
/* 11298 */ {(16<<2)|3,{101,65,50}},
/* 11299 */ {(16<<2)|3,{101,65,97}},
/* 11300 */ {(16<<2)|3,{101,65,99}},
/* 11301 */ {(16<<2)|3,{101,65,101}},
/* 11302 */ {(16<<2)|3,{101,65,105}},
/* 11303 */ {(16<<2)|3,{101,65,111}},
/* 11304 */ {(16<<2)|3,{101,65,115}},
/* 11305 */ {(16<<2)|3,{101,65,116}},
/* 11306 */ {(11<<2)|2,{101,65,0}},
/* 11307 */ {(11<<2)|2,{101,65,0}},
/* 11308 */ {(11<<2)|2,{101,65,0}},
/* 11309 */ {(11<<2)|2,{101,65,0}},
/* 11310 */ {(11<<2)|2,{101,65,0}},
/* 11311 */ {(11<<2)|2,{101,65,0}},
/* 11312 */ {(11<<2)|2,{101,65,0}},
/* 11313 */ {(11<<2)|2,{101,65,0}},
/* 11314 */ {(11<<2)|2,{101,65,0}},
/* 11315 */ {(11<<2)|2,{101,65,0}},
/* 11316 */ {(11<<2)|2,{101,65,0}},
/* 11317 */ {(11<<2)|2,{101,65,0}},
/* 11318 */ {(11<<2)|2,{101,65,0}},
/* 11319 */ {(11<<2)|2,{101,65,0}},
/* 11320 */ {(11<<2)|2,{101,65,0}},
/* 11321 */ {(11<<2)|2,{101,65,0}},
/* 11322 */ {(11<<2)|2,{101,65,0}},
/* 11323 */ {(11<<2)|2,{101,65,0}},
/* 11324 */ {(11<<2)|2,{101,65,0}},
/* 11325 */ {(11<<2)|2,{101,65,0}},
/* 11326 */ {(11<<2)|2,{101,65,0}},
/* 11327 */ {(11<<2)|2,{101,65,0}},
/* 11328 */ {(16<<2)|3,{101,95,48}},
/* 11329 */ {(16<<2)|3,{101,95,49}},
/* 11330 */ {(16<<2)|3,{101,95,50}},
/* 11331 */ {(16<<2)|3,{101,95,97}},
/* 11332 */ {(16<<2)|3,{101,95,99}},
/* 11333 */ {(16<<2)|3,{101,95,101}},
/* 11334 */ {(16<<2)|3,{101,95,105}},
/* 11335 */ {(16<<2)|3,{101,95,111}},
/* 11336 */ {(16<<2)|3,{101,95,115}},
/* 11337 */ {(16<<2)|3,{101,95,116}},
/* 11338 */ {(11<<2)|2,{101,95,0}},
/* 11339 */ {(11<<2)|2,{101,95,0}},
/* 11340 */ {(11<<2)|2,{101,95,0}},
/* 11341 */ {(11<<2)|2,{101,95,0}},
/* 11342 */ {(11<<2)|2,{101,95,0}},
/* 11343 */ {(11<<2)|2,{101,95,0}},
/* 11344 */ {(11<<2)|2,{101,95,0}},
/* 11345 */ {(11<<2)|2,{101,95,0}},
/* 11346 */ {(11<<2)|2,{101,95,0}},
/* 11347 */ {(11<<2)|2,{101,95,0}},
/* 11348 */ {(11<<2)|2,{101,95,0}},
/* 11349 */ {(11<<2)|2,{101,95,0}},
/* 11350 */ {(11<<2)|2,{101,95,0}},
/* 11351 */ {(11<<2)|2,{101,95,0}},
/* 11352 */ {(11<<2)|2,{101,95,0}},
/* 11353 */ {(11<<2)|2,{101,95,0}},
/* 11354 */ {(11<<2)|2,{101,95,0}},
/* 11355 */ {(11<<2)|2,{101,95,0}},
/* 11356 */ {(11<<2)|2,{101,95,0}},
/* 11357 */ {(11<<2)|2,{101,95,0}},
/* 11358 */ {(11<<2)|2,{101,95,0}},
/* 11359 */ {(11<<2)|2,{101,95,0}},
/* 11360 */ {(16<<2)|3,{101,98,48}},
/* 11361 */ {(16<<2)|3,{101,98,49}},
/* 11362 */ {(16<<2)|3,{101,98,50}},
/* 11363 */ {(16<<2)|3,{101,98,97}},
/* 11364 */ {(16<<2)|3,{101,98,99}},
/* 11365 */ {(16<<2)|3,{101,98,101}},
/* 11366 */ {(16<<2)|3,{101,98,105}},
/* 11367 */ {(16<<2)|3,{101,98,111}},
/* 11368 */ {(16<<2)|3,{101,98,115}},
/* 11369 */ {(16<<2)|3,{101,98,116}},
/* 11370 */ {(11<<2)|2,{101,98,0}},
/* 11371 */ {(11<<2)|2,{101,98,0}},
/* 11372 */ {(11<<2)|2,{101,98,0}},
/* 11373 */ {(11<<2)|2,{101,98,0}},
/* 11374 */ {(11<<2)|2,{101,98,0}},
/* 11375 */ {(11<<2)|2,{101,98,0}},
/* 11376 */ {(11<<2)|2,{101,98,0}},
/* 11377 */ {(11<<2)|2,{101,98,0}},
/* 11378 */ {(11<<2)|2,{101,98,0}},
/* 11379 */ {(11<<2)|2,{101,98,0}},
/* 11380 */ {(11<<2)|2,{101,98,0}},
/* 11381 */ {(11<<2)|2,{101,98,0}},
/* 11382 */ {(11<<2)|2,{101,98,0}},
/* 11383 */ {(11<<2)|2,{101,98,0}},
/* 11384 */ {(11<<2)|2,{101,98,0}},
/* 11385 */ {(11<<2)|2,{101,98,0}},
/* 11386 */ {(11<<2)|2,{101,98,0}},
/* 11387 */ {(11<<2)|2,{101,98,0}},
/* 11388 */ {(11<<2)|2,{101,98,0}},
/* 11389 */ {(11<<2)|2,{101,98,0}},
/* 11390 */ {(11<<2)|2,{101,98,0}},
/* 11391 */ {(11<<2)|2,{101,98,0}},
/* 11392 */ {(16<<2)|3,{101,100,48}},
/* 11393 */ {(16<<2)|3,{101,100,49}},
/* 11394 */ {(16<<2)|3,{101,100,50}},
/* 11395 */ {(16<<2)|3,{101,100,97}},
/* 11396 */ {(16<<2)|3,{101,100,99}},
/* 11397 */ {(16<<2)|3,{101,100,101}},
/* 11398 */ {(16<<2)|3,{101,100,105}},
/* 11399 */ {(16<<2)|3,{101,100,111}},
/* 11400 */ {(16<<2)|3,{101,100,115}},
/* 11401 */ {(16<<2)|3,{101,100,116}},
/* 11402 */ {(11<<2)|2,{101,100,0}},
/* 11403 */ {(11<<2)|2,{101,100,0}},
/* 11404 */ {(11<<2)|2,{101,100,0}},
/* 11405 */ {(11<<2)|2,{101,100,0}},
/* 11406 */ {(11<<2)|2,{101,100,0}},
/* 11407 */ {(11<<2)|2,{101,100,0}},
/* 11408 */ {(11<<2)|2,{101,100,0}},
/* 11409 */ {(11<<2)|2,{101,100,0}},
/* 11410 */ {(11<<2)|2,{101,100,0}},
/* 11411 */ {(11<<2)|2,{101,100,0}},
/* 11412 */ {(11<<2)|2,{101,100,0}},
/* 11413 */ {(11<<2)|2,{101,100,0}},
/* 11414 */ {(11<<2)|2,{101,100,0}},
/* 11415 */ {(11<<2)|2,{101,100,0}},
/* 11416 */ {(11<<2)|2,{101,100,0}},
/* 11417 */ {(11<<2)|2,{101,100,0}},
/* 11418 */ {(11<<2)|2,{101,100,0}},
/* 11419 */ {(11<<2)|2,{101,100,0}},
/* 11420 */ {(11<<2)|2,{101,100,0}},
/* 11421 */ {(11<<2)|2,{101,100,0}},
/* 11422 */ {(11<<2)|2,{101,100,0}},
/* 11423 */ {(11<<2)|2,{101,100,0}},
/* 11424 */ {(16<<2)|3,{101,102,48}},
/* 11425 */ {(16<<2)|3,{101,102,49}},
/* 11426 */ {(16<<2)|3,{101,102,50}},
/* 11427 */ {(16<<2)|3,{101,102,97}},
/* 11428 */ {(16<<2)|3,{101,102,99}},
/* 11429 */ {(16<<2)|3,{101,102,101}},
/* 11430 */ {(16<<2)|3,{101,102,105}},
/* 11431 */ {(16<<2)|3,{101,102,111}},
/* 11432 */ {(16<<2)|3,{101,102,115}},
/* 11433 */ {(16<<2)|3,{101,102,116}},
/* 11434 */ {(11<<2)|2,{101,102,0}},
/* 11435 */ {(11<<2)|2,{101,102,0}},
/* 11436 */ {(11<<2)|2,{101,102,0}},
/* 11437 */ {(11<<2)|2,{101,102,0}},
/* 11438 */ {(11<<2)|2,{101,102,0}},
/* 11439 */ {(11<<2)|2,{101,102,0}},
/* 11440 */ {(11<<2)|2,{101,102,0}},
/* 11441 */ {(11<<2)|2,{101,102,0}},
/* 11442 */ {(11<<2)|2,{101,102,0}},
/* 11443 */ {(11<<2)|2,{101,102,0}},
/* 11444 */ {(11<<2)|2,{101,102,0}},
/* 11445 */ {(11<<2)|2,{101,102,0}},
/* 11446 */ {(11<<2)|2,{101,102,0}},
/* 11447 */ {(11<<2)|2,{101,102,0}},
/* 11448 */ {(11<<2)|2,{101,102,0}},
/* 11449 */ {(11<<2)|2,{101,102,0}},
/* 11450 */ {(11<<2)|2,{101,102,0}},
/* 11451 */ {(11<<2)|2,{101,102,0}},
/* 11452 */ {(11<<2)|2,{101,102,0}},
/* 11453 */ {(11<<2)|2,{101,102,0}},
/* 11454 */ {(11<<2)|2,{101,102,0}},
/* 11455 */ {(11<<2)|2,{101,102,0}},
/* 11456 */ {(16<<2)|3,{101,103,48}},
/* 11457 */ {(16<<2)|3,{101,103,49}},
/* 11458 */ {(16<<2)|3,{101,103,50}},
/* 11459 */ {(16<<2)|3,{101,103,97}},
/* 11460 */ {(16<<2)|3,{101,103,99}},
/* 11461 */ {(16<<2)|3,{101,103,101}},
/* 11462 */ {(16<<2)|3,{101,103,105}},
/* 11463 */ {(16<<2)|3,{101,103,111}},
/* 11464 */ {(16<<2)|3,{101,103,115}},
/* 11465 */ {(16<<2)|3,{101,103,116}},
/* 11466 */ {(11<<2)|2,{101,103,0}},
/* 11467 */ {(11<<2)|2,{101,103,0}},
/* 11468 */ {(11<<2)|2,{101,103,0}},
/* 11469 */ {(11<<2)|2,{101,103,0}},
/* 11470 */ {(11<<2)|2,{101,103,0}},
/* 11471 */ {(11<<2)|2,{101,103,0}},
/* 11472 */ {(11<<2)|2,{101,103,0}},
/* 11473 */ {(11<<2)|2,{101,103,0}},
/* 11474 */ {(11<<2)|2,{101,103,0}},
/* 11475 */ {(11<<2)|2,{101,103,0}},
/* 11476 */ {(11<<2)|2,{101,103,0}},
/* 11477 */ {(11<<2)|2,{101,103,0}},
/* 11478 */ {(11<<2)|2,{101,103,0}},
/* 11479 */ {(11<<2)|2,{101,103,0}},
/* 11480 */ {(11<<2)|2,{101,103,0}},
/* 11481 */ {(11<<2)|2,{101,103,0}},
/* 11482 */ {(11<<2)|2,{101,103,0}},
/* 11483 */ {(11<<2)|2,{101,103,0}},
/* 11484 */ {(11<<2)|2,{101,103,0}},
/* 11485 */ {(11<<2)|2,{101,103,0}},
/* 11486 */ {(11<<2)|2,{101,103,0}},
/* 11487 */ {(11<<2)|2,{101,103,0}},
/* 11488 */ {(16<<2)|3,{101,104,48}},
/* 11489 */ {(16<<2)|3,{101,104,49}},
/* 11490 */ {(16<<2)|3,{101,104,50}},
/* 11491 */ {(16<<2)|3,{101,104,97}},
/* 11492 */ {(16<<2)|3,{101,104,99}},
/* 11493 */ {(16<<2)|3,{101,104,101}},
/* 11494 */ {(16<<2)|3,{101,104,105}},
/* 11495 */ {(16<<2)|3,{101,104,111}},
/* 11496 */ {(16<<2)|3,{101,104,115}},
/* 11497 */ {(16<<2)|3,{101,104,116}},
/* 11498 */ {(11<<2)|2,{101,104,0}},
/* 11499 */ {(11<<2)|2,{101,104,0}},
/* 11500 */ {(11<<2)|2,{101,104,0}},
/* 11501 */ {(11<<2)|2,{101,104,0}},
/* 11502 */ {(11<<2)|2,{101,104,0}},
/* 11503 */ {(11<<2)|2,{101,104,0}},
/* 11504 */ {(11<<2)|2,{101,104,0}},
/* 11505 */ {(11<<2)|2,{101,104,0}},
/* 11506 */ {(11<<2)|2,{101,104,0}},
/* 11507 */ {(11<<2)|2,{101,104,0}},
/* 11508 */ {(11<<2)|2,{101,104,0}},
/* 11509 */ {(11<<2)|2,{101,104,0}},
/* 11510 */ {(11<<2)|2,{101,104,0}},
/* 11511 */ {(11<<2)|2,{101,104,0}},
/* 11512 */ {(11<<2)|2,{101,104,0}},
/* 11513 */ {(11<<2)|2,{101,104,0}},
/* 11514 */ {(11<<2)|2,{101,104,0}},
/* 11515 */ {(11<<2)|2,{101,104,0}},
/* 11516 */ {(11<<2)|2,{101,104,0}},
/* 11517 */ {(11<<2)|2,{101,104,0}},
/* 11518 */ {(11<<2)|2,{101,104,0}},
/* 11519 */ {(11<<2)|2,{101,104,0}},
/* 11520 */ {(16<<2)|3,{101,108,48}},
/* 11521 */ {(16<<2)|3,{101,108,49}},
/* 11522 */ {(16<<2)|3,{101,108,50}},
/* 11523 */ {(16<<2)|3,{101,108,97}},
/* 11524 */ {(16<<2)|3,{101,108,99}},
/* 11525 */ {(16<<2)|3,{101,108,101}},
/* 11526 */ {(16<<2)|3,{101,108,105}},
/* 11527 */ {(16<<2)|3,{101,108,111}},
/* 11528 */ {(16<<2)|3,{101,108,115}},
/* 11529 */ {(16<<2)|3,{101,108,116}},
/* 11530 */ {(11<<2)|2,{101,108,0}},
/* 11531 */ {(11<<2)|2,{101,108,0}},
/* 11532 */ {(11<<2)|2,{101,108,0}},
/* 11533 */ {(11<<2)|2,{101,108,0}},
/* 11534 */ {(11<<2)|2,{101,108,0}},
/* 11535 */ {(11<<2)|2,{101,108,0}},
/* 11536 */ {(11<<2)|2,{101,108,0}},
/* 11537 */ {(11<<2)|2,{101,108,0}},
/* 11538 */ {(11<<2)|2,{101,108,0}},
/* 11539 */ {(11<<2)|2,{101,108,0}},
/* 11540 */ {(11<<2)|2,{101,108,0}},
/* 11541 */ {(11<<2)|2,{101,108,0}},
/* 11542 */ {(11<<2)|2,{101,108,0}},
/* 11543 */ {(11<<2)|2,{101,108,0}},
/* 11544 */ {(11<<2)|2,{101,108,0}},
/* 11545 */ {(11<<2)|2,{101,108,0}},
/* 11546 */ {(11<<2)|2,{101,108,0}},
/* 11547 */ {(11<<2)|2,{101,108,0}},
/* 11548 */ {(11<<2)|2,{101,108,0}},
/* 11549 */ {(11<<2)|2,{101,108,0}},
/* 11550 */ {(11<<2)|2,{101,108,0}},
/* 11551 */ {(11<<2)|2,{101,108,0}},
/* 11552 */ {(16<<2)|3,{101,109,48}},
/* 11553 */ {(16<<2)|3,{101,109,49}},
/* 11554 */ {(16<<2)|3,{101,109,50}},
/* 11555 */ {(16<<2)|3,{101,109,97}},
/* 11556 */ {(16<<2)|3,{101,109,99}},
/* 11557 */ {(16<<2)|3,{101,109,101}},
/* 11558 */ {(16<<2)|3,{101,109,105}},
/* 11559 */ {(16<<2)|3,{101,109,111}},
/* 11560 */ {(16<<2)|3,{101,109,115}},
/* 11561 */ {(16<<2)|3,{101,109,116}},
/* 11562 */ {(11<<2)|2,{101,109,0}},
/* 11563 */ {(11<<2)|2,{101,109,0}},
/* 11564 */ {(11<<2)|2,{101,109,0}},
/* 11565 */ {(11<<2)|2,{101,109,0}},
/* 11566 */ {(11<<2)|2,{101,109,0}},
/* 11567 */ {(11<<2)|2,{101,109,0}},
/* 11568 */ {(11<<2)|2,{101,109,0}},
/* 11569 */ {(11<<2)|2,{101,109,0}},
/* 11570 */ {(11<<2)|2,{101,109,0}},
/* 11571 */ {(11<<2)|2,{101,109,0}},
/* 11572 */ {(11<<2)|2,{101,109,0}},
/* 11573 */ {(11<<2)|2,{101,109,0}},
/* 11574 */ {(11<<2)|2,{101,109,0}},
/* 11575 */ {(11<<2)|2,{101,109,0}},
/* 11576 */ {(11<<2)|2,{101,109,0}},
/* 11577 */ {(11<<2)|2,{101,109,0}},
/* 11578 */ {(11<<2)|2,{101,109,0}},
/* 11579 */ {(11<<2)|2,{101,109,0}},
/* 11580 */ {(11<<2)|2,{101,109,0}},
/* 11581 */ {(11<<2)|2,{101,109,0}},
/* 11582 */ {(11<<2)|2,{101,109,0}},
/* 11583 */ {(11<<2)|2,{101,109,0}},
/* 11584 */ {(16<<2)|3,{101,110,48}},
/* 11585 */ {(16<<2)|3,{101,110,49}},
/* 11586 */ {(16<<2)|3,{101,110,50}},
/* 11587 */ {(16<<2)|3,{101,110,97}},
/* 11588 */ {(16<<2)|3,{101,110,99}},
/* 11589 */ {(16<<2)|3,{101,110,101}},
/* 11590 */ {(16<<2)|3,{101,110,105}},
/* 11591 */ {(16<<2)|3,{101,110,111}},
/* 11592 */ {(16<<2)|3,{101,110,115}},
/* 11593 */ {(16<<2)|3,{101,110,116}},
/* 11594 */ {(11<<2)|2,{101,110,0}},
/* 11595 */ {(11<<2)|2,{101,110,0}},
/* 11596 */ {(11<<2)|2,{101,110,0}},
/* 11597 */ {(11<<2)|2,{101,110,0}},
/* 11598 */ {(11<<2)|2,{101,110,0}},
/* 11599 */ {(11<<2)|2,{101,110,0}},
/* 11600 */ {(11<<2)|2,{101,110,0}},
/* 11601 */ {(11<<2)|2,{101,110,0}},
/* 11602 */ {(11<<2)|2,{101,110,0}},
/* 11603 */ {(11<<2)|2,{101,110,0}},
/* 11604 */ {(11<<2)|2,{101,110,0}},
/* 11605 */ {(11<<2)|2,{101,110,0}},
/* 11606 */ {(11<<2)|2,{101,110,0}},
/* 11607 */ {(11<<2)|2,{101,110,0}},
/* 11608 */ {(11<<2)|2,{101,110,0}},
/* 11609 */ {(11<<2)|2,{101,110,0}},
/* 11610 */ {(11<<2)|2,{101,110,0}},
/* 11611 */ {(11<<2)|2,{101,110,0}},
/* 11612 */ {(11<<2)|2,{101,110,0}},
/* 11613 */ {(11<<2)|2,{101,110,0}},
/* 11614 */ {(11<<2)|2,{101,110,0}},
/* 11615 */ {(11<<2)|2,{101,110,0}},
/* 11616 */ {(16<<2)|3,{101,112,48}},
/* 11617 */ {(16<<2)|3,{101,112,49}},
/* 11618 */ {(16<<2)|3,{101,112,50}},
/* 11619 */ {(16<<2)|3,{101,112,97}},
/* 11620 */ {(16<<2)|3,{101,112,99}},
/* 11621 */ {(16<<2)|3,{101,112,101}},
/* 11622 */ {(16<<2)|3,{101,112,105}},
/* 11623 */ {(16<<2)|3,{101,112,111}},
/* 11624 */ {(16<<2)|3,{101,112,115}},
/* 11625 */ {(16<<2)|3,{101,112,116}},
/* 11626 */ {(11<<2)|2,{101,112,0}},
/* 11627 */ {(11<<2)|2,{101,112,0}},
/* 11628 */ {(11<<2)|2,{101,112,0}},
/* 11629 */ {(11<<2)|2,{101,112,0}},
/* 11630 */ {(11<<2)|2,{101,112,0}},
/* 11631 */ {(11<<2)|2,{101,112,0}},
/* 11632 */ {(11<<2)|2,{101,112,0}},
/* 11633 */ {(11<<2)|2,{101,112,0}},
/* 11634 */ {(11<<2)|2,{101,112,0}},
/* 11635 */ {(11<<2)|2,{101,112,0}},
/* 11636 */ {(11<<2)|2,{101,112,0}},
/* 11637 */ {(11<<2)|2,{101,112,0}},
/* 11638 */ {(11<<2)|2,{101,112,0}},
/* 11639 */ {(11<<2)|2,{101,112,0}},
/* 11640 */ {(11<<2)|2,{101,112,0}},
/* 11641 */ {(11<<2)|2,{101,112,0}},
/* 11642 */ {(11<<2)|2,{101,112,0}},
/* 11643 */ {(11<<2)|2,{101,112,0}},
/* 11644 */ {(11<<2)|2,{101,112,0}},
/* 11645 */ {(11<<2)|2,{101,112,0}},
/* 11646 */ {(11<<2)|2,{101,112,0}},
/* 11647 */ {(11<<2)|2,{101,112,0}},
/* 11648 */ {(16<<2)|3,{101,114,48}},
/* 11649 */ {(16<<2)|3,{101,114,49}},
/* 11650 */ {(16<<2)|3,{101,114,50}},
/* 11651 */ {(16<<2)|3,{101,114,97}},
/* 11652 */ {(16<<2)|3,{101,114,99}},
/* 11653 */ {(16<<2)|3,{101,114,101}},
/* 11654 */ {(16<<2)|3,{101,114,105}},
/* 11655 */ {(16<<2)|3,{101,114,111}},
/* 11656 */ {(16<<2)|3,{101,114,115}},
/* 11657 */ {(16<<2)|3,{101,114,116}},
/* 11658 */ {(11<<2)|2,{101,114,0}},
/* 11659 */ {(11<<2)|2,{101,114,0}},
/* 11660 */ {(11<<2)|2,{101,114,0}},
/* 11661 */ {(11<<2)|2,{101,114,0}},
/* 11662 */ {(11<<2)|2,{101,114,0}},
/* 11663 */ {(11<<2)|2,{101,114,0}},
/* 11664 */ {(11<<2)|2,{101,114,0}},
/* 11665 */ {(11<<2)|2,{101,114,0}},
/* 11666 */ {(11<<2)|2,{101,114,0}},
/* 11667 */ {(11<<2)|2,{101,114,0}},
/* 11668 */ {(11<<2)|2,{101,114,0}},
/* 11669 */ {(11<<2)|2,{101,114,0}},
/* 11670 */ {(11<<2)|2,{101,114,0}},
/* 11671 */ {(11<<2)|2,{101,114,0}},
/* 11672 */ {(11<<2)|2,{101,114,0}},
/* 11673 */ {(11<<2)|2,{101,114,0}},
/* 11674 */ {(11<<2)|2,{101,114,0}},
/* 11675 */ {(11<<2)|2,{101,114,0}},
/* 11676 */ {(11<<2)|2,{101,114,0}},
/* 11677 */ {(11<<2)|2,{101,114,0}},
/* 11678 */ {(11<<2)|2,{101,114,0}},
/* 11679 */ {(11<<2)|2,{101,114,0}},
/* 11680 */ {(16<<2)|3,{101,117,48}},
/* 11681 */ {(16<<2)|3,{101,117,49}},
/* 11682 */ {(16<<2)|3,{101,117,50}},
/* 11683 */ {(16<<2)|3,{101,117,97}},
/* 11684 */ {(16<<2)|3,{101,117,99}},
/* 11685 */ {(16<<2)|3,{101,117,101}},
/* 11686 */ {(16<<2)|3,{101,117,105}},
/* 11687 */ {(16<<2)|3,{101,117,111}},
/* 11688 */ {(16<<2)|3,{101,117,115}},
/* 11689 */ {(16<<2)|3,{101,117,116}},
/* 11690 */ {(11<<2)|2,{101,117,0}},
/* 11691 */ {(11<<2)|2,{101,117,0}},
/* 11692 */ {(11<<2)|2,{101,117,0}},
/* 11693 */ {(11<<2)|2,{101,117,0}},
/* 11694 */ {(11<<2)|2,{101,117,0}},
/* 11695 */ {(11<<2)|2,{101,117,0}},
/* 11696 */ {(11<<2)|2,{101,117,0}},
/* 11697 */ {(11<<2)|2,{101,117,0}},
/* 11698 */ {(11<<2)|2,{101,117,0}},
/* 11699 */ {(11<<2)|2,{101,117,0}},
/* 11700 */ {(11<<2)|2,{101,117,0}},
/* 11701 */ {(11<<2)|2,{101,117,0}},
/* 11702 */ {(11<<2)|2,{101,117,0}},
/* 11703 */ {(11<<2)|2,{101,117,0}},
/* 11704 */ {(11<<2)|2,{101,117,0}},
/* 11705 */ {(11<<2)|2,{101,117,0}},
/* 11706 */ {(11<<2)|2,{101,117,0}},
/* 11707 */ {(11<<2)|2,{101,117,0}},
/* 11708 */ {(11<<2)|2,{101,117,0}},
/* 11709 */ {(11<<2)|2,{101,117,0}},
/* 11710 */ {(11<<2)|2,{101,117,0}},
/* 11711 */ {(11<<2)|2,{101,117,0}},
/* 11712 */ {(12<<2)|2,{101,58,0}},
/* 11713 */ {(12<<2)|2,{101,58,0}},
/* 11714 */ {(12<<2)|2,{101,58,0}},
/* 11715 */ {(12<<2)|2,{101,58,0}},
/* 11716 */ {(12<<2)|2,{101,58,0}},
/* 11717 */ {(12<<2)|2,{101,58,0}},
/* 11718 */ {(12<<2)|2,{101,58,0}},
/* 11719 */ {(12<<2)|2,{101,58,0}},
/* 11720 */ {(12<<2)|2,{101,58,0}},
/* 11721 */ {(12<<2)|2,{101,58,0}},
/* 11722 */ {(12<<2)|2,{101,58,0}},
/* 11723 */ {(12<<2)|2,{101,58,0}},
/* 11724 */ {(12<<2)|2,{101,58,0}},
/* 11725 */ {(12<<2)|2,{101,58,0}},
/* 11726 */ {(12<<2)|2,{101,58,0}},
/* 11727 */ {(12<<2)|2,{101,58,0}},
/* 11728 */ {(12<<2)|2,{101,66,0}},
/* 11729 */ {(12<<2)|2,{101,66,0}},
/* 11730 */ {(12<<2)|2,{101,66,0}},
/* 11731 */ {(12<<2)|2,{101,66,0}},
/* 11732 */ {(12<<2)|2,{101,66,0}},
/* 11733 */ {(12<<2)|2,{101,66,0}},
/* 11734 */ {(12<<2)|2,{101,66,0}},
/* 11735 */ {(12<<2)|2,{101,66,0}},
/* 11736 */ {(12<<2)|2,{101,66,0}},
/* 11737 */ {(12<<2)|2,{101,66,0}},
/* 11738 */ {(12<<2)|2,{101,66,0}},
/* 11739 */ {(12<<2)|2,{101,66,0}},
/* 11740 */ {(12<<2)|2,{101,66,0}},
/* 11741 */ {(12<<2)|2,{101,66,0}},
/* 11742 */ {(12<<2)|2,{101,66,0}},
/* 11743 */ {(12<<2)|2,{101,66,0}},
/* 11744 */ {(12<<2)|2,{101,67,0}},
/* 11745 */ {(12<<2)|2,{101,67,0}},
/* 11746 */ {(12<<2)|2,{101,67,0}},
/* 11747 */ {(12<<2)|2,{101,67,0}},
/* 11748 */ {(12<<2)|2,{101,67,0}},
/* 11749 */ {(12<<2)|2,{101,67,0}},
/* 11750 */ {(12<<2)|2,{101,67,0}},
/* 11751 */ {(12<<2)|2,{101,67,0}},
/* 11752 */ {(12<<2)|2,{101,67,0}},
/* 11753 */ {(12<<2)|2,{101,67,0}},
/* 11754 */ {(12<<2)|2,{101,67,0}},
/* 11755 */ {(12<<2)|2,{101,67,0}},
/* 11756 */ {(12<<2)|2,{101,67,0}},
/* 11757 */ {(12<<2)|2,{101,67,0}},
/* 11758 */ {(12<<2)|2,{101,67,0}},
/* 11759 */ {(12<<2)|2,{101,67,0}},
/* 11760 */ {(12<<2)|2,{101,68,0}},
/* 11761 */ {(12<<2)|2,{101,68,0}},
/* 11762 */ {(12<<2)|2,{101,68,0}},
/* 11763 */ {(12<<2)|2,{101,68,0}},
/* 11764 */ {(12<<2)|2,{101,68,0}},
/* 11765 */ {(12<<2)|2,{101,68,0}},
/* 11766 */ {(12<<2)|2,{101,68,0}},
/* 11767 */ {(12<<2)|2,{101,68,0}},
/* 11768 */ {(12<<2)|2,{101,68,0}},
/* 11769 */ {(12<<2)|2,{101,68,0}},
/* 11770 */ {(12<<2)|2,{101,68,0}},
/* 11771 */ {(12<<2)|2,{101,68,0}},
/* 11772 */ {(12<<2)|2,{101,68,0}},
/* 11773 */ {(12<<2)|2,{101,68,0}},
/* 11774 */ {(12<<2)|2,{101,68,0}},
/* 11775 */ {(12<<2)|2,{101,68,0}},
/* 11776 */ {(12<<2)|2,{101,69,0}},
/* 11777 */ {(12<<2)|2,{101,69,0}},
/* 11778 */ {(12<<2)|2,{101,69,0}},
/* 11779 */ {(12<<2)|2,{101,69,0}},
/* 11780 */ {(12<<2)|2,{101,69,0}},
/* 11781 */ {(12<<2)|2,{101,69,0}},
/* 11782 */ {(12<<2)|2,{101,69,0}},
/* 11783 */ {(12<<2)|2,{101,69,0}},
/* 11784 */ {(12<<2)|2,{101,69,0}},
/* 11785 */ {(12<<2)|2,{101,69,0}},
/* 11786 */ {(12<<2)|2,{101,69,0}},
/* 11787 */ {(12<<2)|2,{101,69,0}},
/* 11788 */ {(12<<2)|2,{101,69,0}},
/* 11789 */ {(12<<2)|2,{101,69,0}},
/* 11790 */ {(12<<2)|2,{101,69,0}},
/* 11791 */ {(12<<2)|2,{101,69,0}},
/* 11792 */ {(12<<2)|2,{101,70,0}},
/* 11793 */ {(12<<2)|2,{101,70,0}},
/* 11794 */ {(12<<2)|2,{101,70,0}},
/* 11795 */ {(12<<2)|2,{101,70,0}},
/* 11796 */ {(12<<2)|2,{101,70,0}},
/* 11797 */ {(12<<2)|2,{101,70,0}},
/* 11798 */ {(12<<2)|2,{101,70,0}},
/* 11799 */ {(12<<2)|2,{101,70,0}},
/* 11800 */ {(12<<2)|2,{101,70,0}},
/* 11801 */ {(12<<2)|2,{101,70,0}},
/* 11802 */ {(12<<2)|2,{101,70,0}},
/* 11803 */ {(12<<2)|2,{101,70,0}},
/* 11804 */ {(12<<2)|2,{101,70,0}},
/* 11805 */ {(12<<2)|2,{101,70,0}},
/* 11806 */ {(12<<2)|2,{101,70,0}},
/* 11807 */ {(12<<2)|2,{101,70,0}},
/* 11808 */ {(12<<2)|2,{101,71,0}},
/* 11809 */ {(12<<2)|2,{101,71,0}},
/* 11810 */ {(12<<2)|2,{101,71,0}},
/* 11811 */ {(12<<2)|2,{101,71,0}},
/* 11812 */ {(12<<2)|2,{101,71,0}},
/* 11813 */ {(12<<2)|2,{101,71,0}},
/* 11814 */ {(12<<2)|2,{101,71,0}},
/* 11815 */ {(12<<2)|2,{101,71,0}},
/* 11816 */ {(12<<2)|2,{101,71,0}},
/* 11817 */ {(12<<2)|2,{101,71,0}},
/* 11818 */ {(12<<2)|2,{101,71,0}},
/* 11819 */ {(12<<2)|2,{101,71,0}},
/* 11820 */ {(12<<2)|2,{101,71,0}},
/* 11821 */ {(12<<2)|2,{101,71,0}},
/* 11822 */ {(12<<2)|2,{101,71,0}},
/* 11823 */ {(12<<2)|2,{101,71,0}},
/* 11824 */ {(12<<2)|2,{101,72,0}},
/* 11825 */ {(12<<2)|2,{101,72,0}},
/* 11826 */ {(12<<2)|2,{101,72,0}},
/* 11827 */ {(12<<2)|2,{101,72,0}},
/* 11828 */ {(12<<2)|2,{101,72,0}},
/* 11829 */ {(12<<2)|2,{101,72,0}},
/* 11830 */ {(12<<2)|2,{101,72,0}},
/* 11831 */ {(12<<2)|2,{101,72,0}},
/* 11832 */ {(12<<2)|2,{101,72,0}},
/* 11833 */ {(12<<2)|2,{101,72,0}},
/* 11834 */ {(12<<2)|2,{101,72,0}},
/* 11835 */ {(12<<2)|2,{101,72,0}},
/* 11836 */ {(12<<2)|2,{101,72,0}},
/* 11837 */ {(12<<2)|2,{101,72,0}},
/* 11838 */ {(12<<2)|2,{101,72,0}},
/* 11839 */ {(12<<2)|2,{101,72,0}},
/* 11840 */ {(12<<2)|2,{101,73,0}},
/* 11841 */ {(12<<2)|2,{101,73,0}},
/* 11842 */ {(12<<2)|2,{101,73,0}},
/* 11843 */ {(12<<2)|2,{101,73,0}},
/* 11844 */ {(12<<2)|2,{101,73,0}},
/* 11845 */ {(12<<2)|2,{101,73,0}},
/* 11846 */ {(12<<2)|2,{101,73,0}},
/* 11847 */ {(12<<2)|2,{101,73,0}},
/* 11848 */ {(12<<2)|2,{101,73,0}},
/* 11849 */ {(12<<2)|2,{101,73,0}},
/* 11850 */ {(12<<2)|2,{101,73,0}},
/* 11851 */ {(12<<2)|2,{101,73,0}},
/* 11852 */ {(12<<2)|2,{101,73,0}},
/* 11853 */ {(12<<2)|2,{101,73,0}},
/* 11854 */ {(12<<2)|2,{101,73,0}},
/* 11855 */ {(12<<2)|2,{101,73,0}},
/* 11856 */ {(12<<2)|2,{101,74,0}},
/* 11857 */ {(12<<2)|2,{101,74,0}},
/* 11858 */ {(12<<2)|2,{101,74,0}},
/* 11859 */ {(12<<2)|2,{101,74,0}},
/* 11860 */ {(12<<2)|2,{101,74,0}},
/* 11861 */ {(12<<2)|2,{101,74,0}},
/* 11862 */ {(12<<2)|2,{101,74,0}},
/* 11863 */ {(12<<2)|2,{101,74,0}},
/* 11864 */ {(12<<2)|2,{101,74,0}},
/* 11865 */ {(12<<2)|2,{101,74,0}},
/* 11866 */ {(12<<2)|2,{101,74,0}},
/* 11867 */ {(12<<2)|2,{101,74,0}},
/* 11868 */ {(12<<2)|2,{101,74,0}},
/* 11869 */ {(12<<2)|2,{101,74,0}},
/* 11870 */ {(12<<2)|2,{101,74,0}},
/* 11871 */ {(12<<2)|2,{101,74,0}},
/* 11872 */ {(12<<2)|2,{101,75,0}},
/* 11873 */ {(12<<2)|2,{101,75,0}},
/* 11874 */ {(12<<2)|2,{101,75,0}},
/* 11875 */ {(12<<2)|2,{101,75,0}},
/* 11876 */ {(12<<2)|2,{101,75,0}},
/* 11877 */ {(12<<2)|2,{101,75,0}},
/* 11878 */ {(12<<2)|2,{101,75,0}},
/* 11879 */ {(12<<2)|2,{101,75,0}},
/* 11880 */ {(12<<2)|2,{101,75,0}},
/* 11881 */ {(12<<2)|2,{101,75,0}},
/* 11882 */ {(12<<2)|2,{101,75,0}},
/* 11883 */ {(12<<2)|2,{101,75,0}},
/* 11884 */ {(12<<2)|2,{101,75,0}},
/* 11885 */ {(12<<2)|2,{101,75,0}},
/* 11886 */ {(12<<2)|2,{101,75,0}},
/* 11887 */ {(12<<2)|2,{101,75,0}},
/* 11888 */ {(12<<2)|2,{101,76,0}},
/* 11889 */ {(12<<2)|2,{101,76,0}},
/* 11890 */ {(12<<2)|2,{101,76,0}},
/* 11891 */ {(12<<2)|2,{101,76,0}},
/* 11892 */ {(12<<2)|2,{101,76,0}},
/* 11893 */ {(12<<2)|2,{101,76,0}},
/* 11894 */ {(12<<2)|2,{101,76,0}},
/* 11895 */ {(12<<2)|2,{101,76,0}},
/* 11896 */ {(12<<2)|2,{101,76,0}},
/* 11897 */ {(12<<2)|2,{101,76,0}},
/* 11898 */ {(12<<2)|2,{101,76,0}},
/* 11899 */ {(12<<2)|2,{101,76,0}},
/* 11900 */ {(12<<2)|2,{101,76,0}},
/* 11901 */ {(12<<2)|2,{101,76,0}},
/* 11902 */ {(12<<2)|2,{101,76,0}},
/* 11903 */ {(12<<2)|2,{101,76,0}},
/* 11904 */ {(12<<2)|2,{101,77,0}},
/* 11905 */ {(12<<2)|2,{101,77,0}},
/* 11906 */ {(12<<2)|2,{101,77,0}},
/* 11907 */ {(12<<2)|2,{101,77,0}},
/* 11908 */ {(12<<2)|2,{101,77,0}},
/* 11909 */ {(12<<2)|2,{101,77,0}},
/* 11910 */ {(12<<2)|2,{101,77,0}},
/* 11911 */ {(12<<2)|2,{101,77,0}},
/* 11912 */ {(12<<2)|2,{101,77,0}},
/* 11913 */ {(12<<2)|2,{101,77,0}},
/* 11914 */ {(12<<2)|2,{101,77,0}},
/* 11915 */ {(12<<2)|2,{101,77,0}},
/* 11916 */ {(12<<2)|2,{101,77,0}},
/* 11917 */ {(12<<2)|2,{101,77,0}},
/* 11918 */ {(12<<2)|2,{101,77,0}},
/* 11919 */ {(12<<2)|2,{101,77,0}},
/* 11920 */ {(12<<2)|2,{101,78,0}},
/* 11921 */ {(12<<2)|2,{101,78,0}},
/* 11922 */ {(12<<2)|2,{101,78,0}},
/* 11923 */ {(12<<2)|2,{101,78,0}},
/* 11924 */ {(12<<2)|2,{101,78,0}},
/* 11925 */ {(12<<2)|2,{101,78,0}},
/* 11926 */ {(12<<2)|2,{101,78,0}},
/* 11927 */ {(12<<2)|2,{101,78,0}},
/* 11928 */ {(12<<2)|2,{101,78,0}},
/* 11929 */ {(12<<2)|2,{101,78,0}},
/* 11930 */ {(12<<2)|2,{101,78,0}},
/* 11931 */ {(12<<2)|2,{101,78,0}},
/* 11932 */ {(12<<2)|2,{101,78,0}},
/* 11933 */ {(12<<2)|2,{101,78,0}},
/* 11934 */ {(12<<2)|2,{101,78,0}},
/* 11935 */ {(12<<2)|2,{101,78,0}},
/* 11936 */ {(12<<2)|2,{101,79,0}},
/* 11937 */ {(12<<2)|2,{101,79,0}},
/* 11938 */ {(12<<2)|2,{101,79,0}},
/* 11939 */ {(12<<2)|2,{101,79,0}},
/* 11940 */ {(12<<2)|2,{101,79,0}},
/* 11941 */ {(12<<2)|2,{101,79,0}},
/* 11942 */ {(12<<2)|2,{101,79,0}},
/* 11943 */ {(12<<2)|2,{101,79,0}},
/* 11944 */ {(12<<2)|2,{101,79,0}},
/* 11945 */ {(12<<2)|2,{101,79,0}},
/* 11946 */ {(12<<2)|2,{101,79,0}},
/* 11947 */ {(12<<2)|2,{101,79,0}},
/* 11948 */ {(12<<2)|2,{101,79,0}},
/* 11949 */ {(12<<2)|2,{101,79,0}},
/* 11950 */ {(12<<2)|2,{101,79,0}},
/* 11951 */ {(12<<2)|2,{101,79,0}},
/* 11952 */ {(12<<2)|2,{101,80,0}},
/* 11953 */ {(12<<2)|2,{101,80,0}},
/* 11954 */ {(12<<2)|2,{101,80,0}},
/* 11955 */ {(12<<2)|2,{101,80,0}},
/* 11956 */ {(12<<2)|2,{101,80,0}},
/* 11957 */ {(12<<2)|2,{101,80,0}},
/* 11958 */ {(12<<2)|2,{101,80,0}},
/* 11959 */ {(12<<2)|2,{101,80,0}},
/* 11960 */ {(12<<2)|2,{101,80,0}},
/* 11961 */ {(12<<2)|2,{101,80,0}},
/* 11962 */ {(12<<2)|2,{101,80,0}},
/* 11963 */ {(12<<2)|2,{101,80,0}},
/* 11964 */ {(12<<2)|2,{101,80,0}},
/* 11965 */ {(12<<2)|2,{101,80,0}},
/* 11966 */ {(12<<2)|2,{101,80,0}},
/* 11967 */ {(12<<2)|2,{101,80,0}},
/* 11968 */ {(12<<2)|2,{101,81,0}},
/* 11969 */ {(12<<2)|2,{101,81,0}},
/* 11970 */ {(12<<2)|2,{101,81,0}},
/* 11971 */ {(12<<2)|2,{101,81,0}},
/* 11972 */ {(12<<2)|2,{101,81,0}},
/* 11973 */ {(12<<2)|2,{101,81,0}},
/* 11974 */ {(12<<2)|2,{101,81,0}},
/* 11975 */ {(12<<2)|2,{101,81,0}},
/* 11976 */ {(12<<2)|2,{101,81,0}},
/* 11977 */ {(12<<2)|2,{101,81,0}},
/* 11978 */ {(12<<2)|2,{101,81,0}},
/* 11979 */ {(12<<2)|2,{101,81,0}},
/* 11980 */ {(12<<2)|2,{101,81,0}},
/* 11981 */ {(12<<2)|2,{101,81,0}},
/* 11982 */ {(12<<2)|2,{101,81,0}},
/* 11983 */ {(12<<2)|2,{101,81,0}},
/* 11984 */ {(12<<2)|2,{101,82,0}},
/* 11985 */ {(12<<2)|2,{101,82,0}},
/* 11986 */ {(12<<2)|2,{101,82,0}},
/* 11987 */ {(12<<2)|2,{101,82,0}},
/* 11988 */ {(12<<2)|2,{101,82,0}},
/* 11989 */ {(12<<2)|2,{101,82,0}},
/* 11990 */ {(12<<2)|2,{101,82,0}},
/* 11991 */ {(12<<2)|2,{101,82,0}},
/* 11992 */ {(12<<2)|2,{101,82,0}},
/* 11993 */ {(12<<2)|2,{101,82,0}},
/* 11994 */ {(12<<2)|2,{101,82,0}},
/* 11995 */ {(12<<2)|2,{101,82,0}},
/* 11996 */ {(12<<2)|2,{101,82,0}},
/* 11997 */ {(12<<2)|2,{101,82,0}},
/* 11998 */ {(12<<2)|2,{101,82,0}},
/* 11999 */ {(12<<2)|2,{101,82,0}},
/* 12000 */ {(12<<2)|2,{101,83,0}},
/* 12001 */ {(12<<2)|2,{101,83,0}},
/* 12002 */ {(12<<2)|2,{101,83,0}},
/* 12003 */ {(12<<2)|2,{101,83,0}},
/* 12004 */ {(12<<2)|2,{101,83,0}},
/* 12005 */ {(12<<2)|2,{101,83,0}},
/* 12006 */ {(12<<2)|2,{101,83,0}},
/* 12007 */ {(12<<2)|2,{101,83,0}},
/* 12008 */ {(12<<2)|2,{101,83,0}},
/* 12009 */ {(12<<2)|2,{101,83,0}},
/* 12010 */ {(12<<2)|2,{101,83,0}},
/* 12011 */ {(12<<2)|2,{101,83,0}},
/* 12012 */ {(12<<2)|2,{101,83,0}},
/* 12013 */ {(12<<2)|2,{101,83,0}},
/* 12014 */ {(12<<2)|2,{101,83,0}},
/* 12015 */ {(12<<2)|2,{101,83,0}},
/* 12016 */ {(12<<2)|2,{101,84,0}},
/* 12017 */ {(12<<2)|2,{101,84,0}},
/* 12018 */ {(12<<2)|2,{101,84,0}},
/* 12019 */ {(12<<2)|2,{101,84,0}},
/* 12020 */ {(12<<2)|2,{101,84,0}},
/* 12021 */ {(12<<2)|2,{101,84,0}},
/* 12022 */ {(12<<2)|2,{101,84,0}},
/* 12023 */ {(12<<2)|2,{101,84,0}},
/* 12024 */ {(12<<2)|2,{101,84,0}},
/* 12025 */ {(12<<2)|2,{101,84,0}},
/* 12026 */ {(12<<2)|2,{101,84,0}},
/* 12027 */ {(12<<2)|2,{101,84,0}},
/* 12028 */ {(12<<2)|2,{101,84,0}},
/* 12029 */ {(12<<2)|2,{101,84,0}},
/* 12030 */ {(12<<2)|2,{101,84,0}},
/* 12031 */ {(12<<2)|2,{101,84,0}},
/* 12032 */ {(12<<2)|2,{101,85,0}},
/* 12033 */ {(12<<2)|2,{101,85,0}},
/* 12034 */ {(12<<2)|2,{101,85,0}},
/* 12035 */ {(12<<2)|2,{101,85,0}},
/* 12036 */ {(12<<2)|2,{101,85,0}},
/* 12037 */ {(12<<2)|2,{101,85,0}},
/* 12038 */ {(12<<2)|2,{101,85,0}},
/* 12039 */ {(12<<2)|2,{101,85,0}},
/* 12040 */ {(12<<2)|2,{101,85,0}},
/* 12041 */ {(12<<2)|2,{101,85,0}},
/* 12042 */ {(12<<2)|2,{101,85,0}},
/* 12043 */ {(12<<2)|2,{101,85,0}},
/* 12044 */ {(12<<2)|2,{101,85,0}},
/* 12045 */ {(12<<2)|2,{101,85,0}},
/* 12046 */ {(12<<2)|2,{101,85,0}},
/* 12047 */ {(12<<2)|2,{101,85,0}},
/* 12048 */ {(12<<2)|2,{101,86,0}},
/* 12049 */ {(12<<2)|2,{101,86,0}},
/* 12050 */ {(12<<2)|2,{101,86,0}},
/* 12051 */ {(12<<2)|2,{101,86,0}},
/* 12052 */ {(12<<2)|2,{101,86,0}},
/* 12053 */ {(12<<2)|2,{101,86,0}},
/* 12054 */ {(12<<2)|2,{101,86,0}},
/* 12055 */ {(12<<2)|2,{101,86,0}},
/* 12056 */ {(12<<2)|2,{101,86,0}},
/* 12057 */ {(12<<2)|2,{101,86,0}},
/* 12058 */ {(12<<2)|2,{101,86,0}},
/* 12059 */ {(12<<2)|2,{101,86,0}},
/* 12060 */ {(12<<2)|2,{101,86,0}},
/* 12061 */ {(12<<2)|2,{101,86,0}},
/* 12062 */ {(12<<2)|2,{101,86,0}},
/* 12063 */ {(12<<2)|2,{101,86,0}},
/* 12064 */ {(12<<2)|2,{101,87,0}},
/* 12065 */ {(12<<2)|2,{101,87,0}},
/* 12066 */ {(12<<2)|2,{101,87,0}},
/* 12067 */ {(12<<2)|2,{101,87,0}},
/* 12068 */ {(12<<2)|2,{101,87,0}},
/* 12069 */ {(12<<2)|2,{101,87,0}},
/* 12070 */ {(12<<2)|2,{101,87,0}},
/* 12071 */ {(12<<2)|2,{101,87,0}},
/* 12072 */ {(12<<2)|2,{101,87,0}},
/* 12073 */ {(12<<2)|2,{101,87,0}},
/* 12074 */ {(12<<2)|2,{101,87,0}},
/* 12075 */ {(12<<2)|2,{101,87,0}},
/* 12076 */ {(12<<2)|2,{101,87,0}},
/* 12077 */ {(12<<2)|2,{101,87,0}},
/* 12078 */ {(12<<2)|2,{101,87,0}},
/* 12079 */ {(12<<2)|2,{101,87,0}},
/* 12080 */ {(12<<2)|2,{101,89,0}},
/* 12081 */ {(12<<2)|2,{101,89,0}},
/* 12082 */ {(12<<2)|2,{101,89,0}},
/* 12083 */ {(12<<2)|2,{101,89,0}},
/* 12084 */ {(12<<2)|2,{101,89,0}},
/* 12085 */ {(12<<2)|2,{101,89,0}},
/* 12086 */ {(12<<2)|2,{101,89,0}},
/* 12087 */ {(12<<2)|2,{101,89,0}},
/* 12088 */ {(12<<2)|2,{101,89,0}},
/* 12089 */ {(12<<2)|2,{101,89,0}},
/* 12090 */ {(12<<2)|2,{101,89,0}},
/* 12091 */ {(12<<2)|2,{101,89,0}},
/* 12092 */ {(12<<2)|2,{101,89,0}},
/* 12093 */ {(12<<2)|2,{101,89,0}},
/* 12094 */ {(12<<2)|2,{101,89,0}},
/* 12095 */ {(12<<2)|2,{101,89,0}},
/* 12096 */ {(12<<2)|2,{101,106,0}},
/* 12097 */ {(12<<2)|2,{101,106,0}},
/* 12098 */ {(12<<2)|2,{101,106,0}},
/* 12099 */ {(12<<2)|2,{101,106,0}},
/* 12100 */ {(12<<2)|2,{101,106,0}},
/* 12101 */ {(12<<2)|2,{101,106,0}},
/* 12102 */ {(12<<2)|2,{101,106,0}},
/* 12103 */ {(12<<2)|2,{101,106,0}},
/* 12104 */ {(12<<2)|2,{101,106,0}},
/* 12105 */ {(12<<2)|2,{101,106,0}},
/* 12106 */ {(12<<2)|2,{101,106,0}},
/* 12107 */ {(12<<2)|2,{101,106,0}},
/* 12108 */ {(12<<2)|2,{101,106,0}},
/* 12109 */ {(12<<2)|2,{101,106,0}},
/* 12110 */ {(12<<2)|2,{101,106,0}},
/* 12111 */ {(12<<2)|2,{101,106,0}},
/* 12112 */ {(12<<2)|2,{101,107,0}},
/* 12113 */ {(12<<2)|2,{101,107,0}},
/* 12114 */ {(12<<2)|2,{101,107,0}},
/* 12115 */ {(12<<2)|2,{101,107,0}},
/* 12116 */ {(12<<2)|2,{101,107,0}},
/* 12117 */ {(12<<2)|2,{101,107,0}},
/* 12118 */ {(12<<2)|2,{101,107,0}},
/* 12119 */ {(12<<2)|2,{101,107,0}},
/* 12120 */ {(12<<2)|2,{101,107,0}},
/* 12121 */ {(12<<2)|2,{101,107,0}},
/* 12122 */ {(12<<2)|2,{101,107,0}},
/* 12123 */ {(12<<2)|2,{101,107,0}},
/* 12124 */ {(12<<2)|2,{101,107,0}},
/* 12125 */ {(12<<2)|2,{101,107,0}},
/* 12126 */ {(12<<2)|2,{101,107,0}},
/* 12127 */ {(12<<2)|2,{101,107,0}},
/* 12128 */ {(12<<2)|2,{101,113,0}},
/* 12129 */ {(12<<2)|2,{101,113,0}},
/* 12130 */ {(12<<2)|2,{101,113,0}},
/* 12131 */ {(12<<2)|2,{101,113,0}},
/* 12132 */ {(12<<2)|2,{101,113,0}},
/* 12133 */ {(12<<2)|2,{101,113,0}},
/* 12134 */ {(12<<2)|2,{101,113,0}},
/* 12135 */ {(12<<2)|2,{101,113,0}},
/* 12136 */ {(12<<2)|2,{101,113,0}},
/* 12137 */ {(12<<2)|2,{101,113,0}},
/* 12138 */ {(12<<2)|2,{101,113,0}},
/* 12139 */ {(12<<2)|2,{101,113,0}},
/* 12140 */ {(12<<2)|2,{101,113,0}},
/* 12141 */ {(12<<2)|2,{101,113,0}},
/* 12142 */ {(12<<2)|2,{101,113,0}},
/* 12143 */ {(12<<2)|2,{101,113,0}},
/* 12144 */ {(12<<2)|2,{101,118,0}},
/* 12145 */ {(12<<2)|2,{101,118,0}},
/* 12146 */ {(12<<2)|2,{101,118,0}},
/* 12147 */ {(12<<2)|2,{101,118,0}},
/* 12148 */ {(12<<2)|2,{101,118,0}},
/* 12149 */ {(12<<2)|2,{101,118,0}},
/* 12150 */ {(12<<2)|2,{101,118,0}},
/* 12151 */ {(12<<2)|2,{101,118,0}},
/* 12152 */ {(12<<2)|2,{101,118,0}},
/* 12153 */ {(12<<2)|2,{101,118,0}},
/* 12154 */ {(12<<2)|2,{101,118,0}},
/* 12155 */ {(12<<2)|2,{101,118,0}},
/* 12156 */ {(12<<2)|2,{101,118,0}},
/* 12157 */ {(12<<2)|2,{101,118,0}},
/* 12158 */ {(12<<2)|2,{101,118,0}},
/* 12159 */ {(12<<2)|2,{101,118,0}},
/* 12160 */ {(12<<2)|2,{101,119,0}},
/* 12161 */ {(12<<2)|2,{101,119,0}},
/* 12162 */ {(12<<2)|2,{101,119,0}},
/* 12163 */ {(12<<2)|2,{101,119,0}},
/* 12164 */ {(12<<2)|2,{101,119,0}},
/* 12165 */ {(12<<2)|2,{101,119,0}},
/* 12166 */ {(12<<2)|2,{101,119,0}},
/* 12167 */ {(12<<2)|2,{101,119,0}},
/* 12168 */ {(12<<2)|2,{101,119,0}},
/* 12169 */ {(12<<2)|2,{101,119,0}},
/* 12170 */ {(12<<2)|2,{101,119,0}},
/* 12171 */ {(12<<2)|2,{101,119,0}},
/* 12172 */ {(12<<2)|2,{101,119,0}},
/* 12173 */ {(12<<2)|2,{101,119,0}},
/* 12174 */ {(12<<2)|2,{101,119,0}},
/* 12175 */ {(12<<2)|2,{101,119,0}},
/* 12176 */ {(12<<2)|2,{101,120,0}},
/* 12177 */ {(12<<2)|2,{101,120,0}},
/* 12178 */ {(12<<2)|2,{101,120,0}},
/* 12179 */ {(12<<2)|2,{101,120,0}},
/* 12180 */ {(12<<2)|2,{101,120,0}},
/* 12181 */ {(12<<2)|2,{101,120,0}},
/* 12182 */ {(12<<2)|2,{101,120,0}},
/* 12183 */ {(12<<2)|2,{101,120,0}},
/* 12184 */ {(12<<2)|2,{101,120,0}},
/* 12185 */ {(12<<2)|2,{101,120,0}},
/* 12186 */ {(12<<2)|2,{101,120,0}},
/* 12187 */ {(12<<2)|2,{101,120,0}},
/* 12188 */ {(12<<2)|2,{101,120,0}},
/* 12189 */ {(12<<2)|2,{101,120,0}},
/* 12190 */ {(12<<2)|2,{101,120,0}},
/* 12191 */ {(12<<2)|2,{101,120,0}},
/* 12192 */ {(12<<2)|2,{101,121,0}},
/* 12193 */ {(12<<2)|2,{101,121,0}},
/* 12194 */ {(12<<2)|2,{101,121,0}},
/* 12195 */ {(12<<2)|2,{101,121,0}},
/* 12196 */ {(12<<2)|2,{101,121,0}},
/* 12197 */ {(12<<2)|2,{101,121,0}},
/* 12198 */ {(12<<2)|2,{101,121,0}},
/* 12199 */ {(12<<2)|2,{101,121,0}},
/* 12200 */ {(12<<2)|2,{101,121,0}},
/* 12201 */ {(12<<2)|2,{101,121,0}},
/* 12202 */ {(12<<2)|2,{101,121,0}},
/* 12203 */ {(12<<2)|2,{101,121,0}},
/* 12204 */ {(12<<2)|2,{101,121,0}},
/* 12205 */ {(12<<2)|2,{101,121,0}},
/* 12206 */ {(12<<2)|2,{101,121,0}},
/* 12207 */ {(12<<2)|2,{101,121,0}},
/* 12208 */ {(12<<2)|2,{101,122,0}},
/* 12209 */ {(12<<2)|2,{101,122,0}},
/* 12210 */ {(12<<2)|2,{101,122,0}},
/* 12211 */ {(12<<2)|2,{101,122,0}},
/* 12212 */ {(12<<2)|2,{101,122,0}},
/* 12213 */ {(12<<2)|2,{101,122,0}},
/* 12214 */ {(12<<2)|2,{101,122,0}},
/* 12215 */ {(12<<2)|2,{101,122,0}},
/* 12216 */ {(12<<2)|2,{101,122,0}},
/* 12217 */ {(12<<2)|2,{101,122,0}},
/* 12218 */ {(12<<2)|2,{101,122,0}},
/* 12219 */ {(12<<2)|2,{101,122,0}},
/* 12220 */ {(12<<2)|2,{101,122,0}},
/* 12221 */ {(12<<2)|2,{101,122,0}},
/* 12222 */ {(12<<2)|2,{101,122,0}},
/* 12223 */ {(12<<2)|2,{101,122,0}},
/* 12224 */ {(13<<2)|2,{101,38,0}},
/* 12225 */ {(13<<2)|2,{101,38,0}},
/* 12226 */ {(13<<2)|2,{101,38,0}},
/* 12227 */ {(13<<2)|2,{101,38,0}},
/* 12228 */ {(13<<2)|2,{101,38,0}},
/* 12229 */ {(13<<2)|2,{101,38,0}},
/* 12230 */ {(13<<2)|2,{101,38,0}},
/* 12231 */ {(13<<2)|2,{101,38,0}},
/* 12232 */ {(13<<2)|2,{101,42,0}},
/* 12233 */ {(13<<2)|2,{101,42,0}},
/* 12234 */ {(13<<2)|2,{101,42,0}},
/* 12235 */ {(13<<2)|2,{101,42,0}},
/* 12236 */ {(13<<2)|2,{101,42,0}},
/* 12237 */ {(13<<2)|2,{101,42,0}},
/* 12238 */ {(13<<2)|2,{101,42,0}},
/* 12239 */ {(13<<2)|2,{101,42,0}},
/* 12240 */ {(13<<2)|2,{101,44,0}},
/* 12241 */ {(13<<2)|2,{101,44,0}},
/* 12242 */ {(13<<2)|2,{101,44,0}},
/* 12243 */ {(13<<2)|2,{101,44,0}},
/* 12244 */ {(13<<2)|2,{101,44,0}},
/* 12245 */ {(13<<2)|2,{101,44,0}},
/* 12246 */ {(13<<2)|2,{101,44,0}},
/* 12247 */ {(13<<2)|2,{101,44,0}},
/* 12248 */ {(13<<2)|2,{101,59,0}},
/* 12249 */ {(13<<2)|2,{101,59,0}},
/* 12250 */ {(13<<2)|2,{101,59,0}},
/* 12251 */ {(13<<2)|2,{101,59,0}},
/* 12252 */ {(13<<2)|2,{101,59,0}},
/* 12253 */ {(13<<2)|2,{101,59,0}},
/* 12254 */ {(13<<2)|2,{101,59,0}},
/* 12255 */ {(13<<2)|2,{101,59,0}},
/* 12256 */ {(13<<2)|2,{101,88,0}},
/* 12257 */ {(13<<2)|2,{101,88,0}},
/* 12258 */ {(13<<2)|2,{101,88,0}},
/* 12259 */ {(13<<2)|2,{101,88,0}},
/* 12260 */ {(13<<2)|2,{101,88,0}},
/* 12261 */ {(13<<2)|2,{101,88,0}},
/* 12262 */ {(13<<2)|2,{101,88,0}},
/* 12263 */ {(13<<2)|2,{101,88,0}},
/* 12264 */ {(13<<2)|2,{101,90,0}},
/* 12265 */ {(13<<2)|2,{101,90,0}},
/* 12266 */ {(13<<2)|2,{101,90,0}},
/* 12267 */ {(13<<2)|2,{101,90,0}},
/* 12268 */ {(13<<2)|2,{101,90,0}},
/* 12269 */ {(13<<2)|2,{101,90,0}},
/* 12270 */ {(13<<2)|2,{101,90,0}},
/* 12271 */ {(13<<2)|2,{101,90,0}},
/* 12272 */ {(15<<2)|2,{101,33,0}},
/* 12273 */ {(15<<2)|2,{101,33,0}},
/* 12274 */ {(15<<2)|2,{101,34,0}},
/* 12275 */ {(15<<2)|2,{101,34,0}},
/* 12276 */ {(15<<2)|2,{101,40,0}},
/* 12277 */ {(15<<2)|2,{101,40,0}},
/* 12278 */ {(15<<2)|2,{101,41,0}},
/* 12279 */ {(15<<2)|2,{101,41,0}},
/* 12280 */ {(15<<2)|2,{101,63,0}},
/* 12281 */ {(15<<2)|2,{101,63,0}},
/* 12282 */ {(16<<2)|2,{101,39,0}},
/* 12283 */ {(16<<2)|2,{101,43,0}},
/* 12284 */ {(16<<2)|2,{101,124,0}},
/* 12285 */ {(5<<2)|1,{101,0,0}},
/* 12286 */ {(5<<2)|1,{101,0,0}},
/* 12287 */ {(5<<2)|1,{101,0,0}},
/* 12288 */ {(15<<2)|3,{105,48,48}},
/* 12289 */ {(15<<2)|3,{105,48,48}},
/* 12290 */ {(15<<2)|3,{105,48,49}},
/* 12291 */ {(15<<2)|3,{105,48,49}},
/* 12292 */ {(15<<2)|3,{105,48,50}},
/* 12293 */ {(15<<2)|3,{105,48,50}},
/* 12294 */ {(15<<2)|3,{105,48,97}},
/* 12295 */ {(15<<2)|3,{105,48,97}},
/* 12296 */ {(15<<2)|3,{105,48,99}},
/* 12297 */ {(15<<2)|3,{105,48,99}},
/* 12298 */ {(15<<2)|3,{105,48,101}},
/* 12299 */ {(15<<2)|3,{105,48,101}},
/* 12300 */ {(15<<2)|3,{105,48,105}},
/* 12301 */ {(15<<2)|3,{105,48,105}},
/* 12302 */ {(15<<2)|3,{105,48,111}},
/* 12303 */ {(15<<2)|3,{105,48,111}},
/* 12304 */ {(15<<2)|3,{105,48,115}},
/* 12305 */ {(15<<2)|3,{105,48,115}},
/* 12306 */ {(15<<2)|3,{105,48,116}},
/* 12307 */ {(15<<2)|3,{105,48,116}},
/* 12308 */ {(16<<2)|3,{105,48,32}},
/* 12309 */ {(16<<2)|3,{105,48,37}},
/* 12310 */ {(16<<2)|3,{105,48,45}},
/* 12311 */ {(16<<2)|3,{105,48,46}},
/* 12312 */ {(16<<2)|3,{105,48,47}},
/* 12313 */ {(16<<2)|3,{105,48,51}},
/* 12314 */ {(16<<2)|3,{105,48,52}},
/* 12315 */ {(16<<2)|3,{105,48,53}},
/* 12316 */ {(16<<2)|3,{105,48,54}},
/* 12317 */ {(16<<2)|3,{105,48,55}},
/* 12318 */ {(16<<2)|3,{105,48,56}},
/* 12319 */ {(16<<2)|3,{105,48,57}},
/* 12320 */ {(16<<2)|3,{105,48,61}},
/* 12321 */ {(16<<2)|3,{105,48,65}},
/* 12322 */ {(16<<2)|3,{105,48,95}},
/* 12323 */ {(16<<2)|3,{105,48,98}},
/* 12324 */ {(16<<2)|3,{105,48,100}},
/* 12325 */ {(16<<2)|3,{105,48,102}},
/* 12326 */ {(16<<2)|3,{105,48,103}},
/* 12327 */ {(16<<2)|3,{105,48,104}},
/* 12328 */ {(16<<2)|3,{105,48,108}},
/* 12329 */ {(16<<2)|3,{105,48,109}},
/* 12330 */ {(16<<2)|3,{105,48,110}},
/* 12331 */ {(16<<2)|3,{105,48,112}},
/* 12332 */ {(16<<2)|3,{105,48,114}},
/* 12333 */ {(16<<2)|3,{105,48,117}},
/* 12334 */ {(10<<2)|2,{105,48,0}},
/* 12335 */ {(10<<2)|2,{105,48,0}},
/* 12336 */ {(10<<2)|2,{105,48,0}},
/* 12337 */ {(10<<2)|2,{105,48,0}},
/* 12338 */ {(10<<2)|2,{105,48,0}},
/* 12339 */ {(10<<2)|2,{105,48,0}},
/* 12340 */ {(10<<2)|2,{105,48,0}},
/* 12341 */ {(10<<2)|2,{105,48,0}},
/* 12342 */ {(10<<2)|2,{105,48,0}},
/* 12343 */ {(10<<2)|2,{105,48,0}},
/* 12344 */ {(10<<2)|2,{105,48,0}},
/* 12345 */ {(10<<2)|2,{105,48,0}},
/* 12346 */ {(10<<2)|2,{105,48,0}},
/* 12347 */ {(10<<2)|2,{105,48,0}},
/* 12348 */ {(10<<2)|2,{105,48,0}},
/* 12349 */ {(10<<2)|2,{105,48,0}},
/* 12350 */ {(10<<2)|2,{105,48,0}},
/* 12351 */ {(10<<2)|2,{105,48,0}},
/* 12352 */ {(15<<2)|3,{105,49,48}},
/* 12353 */ {(15<<2)|3,{105,49,48}},
/* 12354 */ {(15<<2)|3,{105,49,49}},
/* 12355 */ {(15<<2)|3,{105,49,49}},
/* 12356 */ {(15<<2)|3,{105,49,50}},
/* 12357 */ {(15<<2)|3,{105,49,50}},
/* 12358 */ {(15<<2)|3,{105,49,97}},
/* 12359 */ {(15<<2)|3,{105,49,97}},
/* 12360 */ {(15<<2)|3,{105,49,99}},
/* 12361 */ {(15<<2)|3,{105,49,99}},
/* 12362 */ {(15<<2)|3,{105,49,101}},
/* 12363 */ {(15<<2)|3,{105,49,101}},
/* 12364 */ {(15<<2)|3,{105,49,105}},
/* 12365 */ {(15<<2)|3,{105,49,105}},
/* 12366 */ {(15<<2)|3,{105,49,111}},
/* 12367 */ {(15<<2)|3,{105,49,111}},
/* 12368 */ {(15<<2)|3,{105,49,115}},
/* 12369 */ {(15<<2)|3,{105,49,115}},
/* 12370 */ {(15<<2)|3,{105,49,116}},
/* 12371 */ {(15<<2)|3,{105,49,116}},
/* 12372 */ {(16<<2)|3,{105,49,32}},
/* 12373 */ {(16<<2)|3,{105,49,37}},
/* 12374 */ {(16<<2)|3,{105,49,45}},
/* 12375 */ {(16<<2)|3,{105,49,46}},
/* 12376 */ {(16<<2)|3,{105,49,47}},
/* 12377 */ {(16<<2)|3,{105,49,51}},
/* 12378 */ {(16<<2)|3,{105,49,52}},
/* 12379 */ {(16<<2)|3,{105,49,53}},
/* 12380 */ {(16<<2)|3,{105,49,54}},
/* 12381 */ {(16<<2)|3,{105,49,55}},
/* 12382 */ {(16<<2)|3,{105,49,56}},
/* 12383 */ {(16<<2)|3,{105,49,57}},
/* 12384 */ {(16<<2)|3,{105,49,61}},
/* 12385 */ {(16<<2)|3,{105,49,65}},
/* 12386 */ {(16<<2)|3,{105,49,95}},
/* 12387 */ {(16<<2)|3,{105,49,98}},
/* 12388 */ {(16<<2)|3,{105,49,100}},
/* 12389 */ {(16<<2)|3,{105,49,102}},
/* 12390 */ {(16<<2)|3,{105,49,103}},
/* 12391 */ {(16<<2)|3,{105,49,104}},
/* 12392 */ {(16<<2)|3,{105,49,108}},
/* 12393 */ {(16<<2)|3,{105,49,109}},
/* 12394 */ {(16<<2)|3,{105,49,110}},
/* 12395 */ {(16<<2)|3,{105,49,112}},
/* 12396 */ {(16<<2)|3,{105,49,114}},
/* 12397 */ {(16<<2)|3,{105,49,117}},
/* 12398 */ {(10<<2)|2,{105,49,0}},
/* 12399 */ {(10<<2)|2,{105,49,0}},
/* 12400 */ {(10<<2)|2,{105,49,0}},
/* 12401 */ {(10<<2)|2,{105,49,0}},
/* 12402 */ {(10<<2)|2,{105,49,0}},
/* 12403 */ {(10<<2)|2,{105,49,0}},
/* 12404 */ {(10<<2)|2,{105,49,0}},
/* 12405 */ {(10<<2)|2,{105,49,0}},
/* 12406 */ {(10<<2)|2,{105,49,0}},
/* 12407 */ {(10<<2)|2,{105,49,0}},
/* 12408 */ {(10<<2)|2,{105,49,0}},
/* 12409 */ {(10<<2)|2,{105,49,0}},
/* 12410 */ {(10<<2)|2,{105,49,0}},
/* 12411 */ {(10<<2)|2,{105,49,0}},
/* 12412 */ {(10<<2)|2,{105,49,0}},
/* 12413 */ {(10<<2)|2,{105,49,0}},
/* 12414 */ {(10<<2)|2,{105,49,0}},
/* 12415 */ {(10<<2)|2,{105,49,0}},
/* 12416 */ {(15<<2)|3,{105,50,48}},
/* 12417 */ {(15<<2)|3,{105,50,48}},
/* 12418 */ {(15<<2)|3,{105,50,49}},
/* 12419 */ {(15<<2)|3,{105,50,49}},
/* 12420 */ {(15<<2)|3,{105,50,50}},
/* 12421 */ {(15<<2)|3,{105,50,50}},
/* 12422 */ {(15<<2)|3,{105,50,97}},
/* 12423 */ {(15<<2)|3,{105,50,97}},
/* 12424 */ {(15<<2)|3,{105,50,99}},
/* 12425 */ {(15<<2)|3,{105,50,99}},
/* 12426 */ {(15<<2)|3,{105,50,101}},
/* 12427 */ {(15<<2)|3,{105,50,101}},
/* 12428 */ {(15<<2)|3,{105,50,105}},
/* 12429 */ {(15<<2)|3,{105,50,105}},
/* 12430 */ {(15<<2)|3,{105,50,111}},
/* 12431 */ {(15<<2)|3,{105,50,111}},
/* 12432 */ {(15<<2)|3,{105,50,115}},
/* 12433 */ {(15<<2)|3,{105,50,115}},
/* 12434 */ {(15<<2)|3,{105,50,116}},
/* 12435 */ {(15<<2)|3,{105,50,116}},
/* 12436 */ {(16<<2)|3,{105,50,32}},
/* 12437 */ {(16<<2)|3,{105,50,37}},
/* 12438 */ {(16<<2)|3,{105,50,45}},
/* 12439 */ {(16<<2)|3,{105,50,46}},
/* 12440 */ {(16<<2)|3,{105,50,47}},
/* 12441 */ {(16<<2)|3,{105,50,51}},
/* 12442 */ {(16<<2)|3,{105,50,52}},
/* 12443 */ {(16<<2)|3,{105,50,53}},
/* 12444 */ {(16<<2)|3,{105,50,54}},
/* 12445 */ {(16<<2)|3,{105,50,55}},
/* 12446 */ {(16<<2)|3,{105,50,56}},
/* 12447 */ {(16<<2)|3,{105,50,57}},
/* 12448 */ {(16<<2)|3,{105,50,61}},
/* 12449 */ {(16<<2)|3,{105,50,65}},
/* 12450 */ {(16<<2)|3,{105,50,95}},
/* 12451 */ {(16<<2)|3,{105,50,98}},
/* 12452 */ {(16<<2)|3,{105,50,100}},
/* 12453 */ {(16<<2)|3,{105,50,102}},
/* 12454 */ {(16<<2)|3,{105,50,103}},
/* 12455 */ {(16<<2)|3,{105,50,104}},
/* 12456 */ {(16<<2)|3,{105,50,108}},
/* 12457 */ {(16<<2)|3,{105,50,109}},
/* 12458 */ {(16<<2)|3,{105,50,110}},
/* 12459 */ {(16<<2)|3,{105,50,112}},
/* 12460 */ {(16<<2)|3,{105,50,114}},
/* 12461 */ {(16<<2)|3,{105,50,117}},
/* 12462 */ {(10<<2)|2,{105,50,0}},
/* 12463 */ {(10<<2)|2,{105,50,0}},
/* 12464 */ {(10<<2)|2,{105,50,0}},
/* 12465 */ {(10<<2)|2,{105,50,0}},
/* 12466 */ {(10<<2)|2,{105,50,0}},
/* 12467 */ {(10<<2)|2,{105,50,0}},
/* 12468 */ {(10<<2)|2,{105,50,0}},
/* 12469 */ {(10<<2)|2,{105,50,0}},
/* 12470 */ {(10<<2)|2,{105,50,0}},
/* 12471 */ {(10<<2)|2,{105,50,0}},
/* 12472 */ {(10<<2)|2,{105,50,0}},
/* 12473 */ {(10<<2)|2,{105,50,0}},
/* 12474 */ {(10<<2)|2,{105,50,0}},
/* 12475 */ {(10<<2)|2,{105,50,0}},
/* 12476 */ {(10<<2)|2,{105,50,0}},
/* 12477 */ {(10<<2)|2,{105,50,0}},
/* 12478 */ {(10<<2)|2,{105,50,0}},
/* 12479 */ {(10<<2)|2,{105,50,0}},
/* 12480 */ {(15<<2)|3,{105,97,48}},
/* 12481 */ {(15<<2)|3,{105,97,48}},
/* 12482 */ {(15<<2)|3,{105,97,49}},
/* 12483 */ {(15<<2)|3,{105,97,49}},
/* 12484 */ {(15<<2)|3,{105,97,50}},
/* 12485 */ {(15<<2)|3,{105,97,50}},
/* 12486 */ {(15<<2)|3,{105,97,97}},
/* 12487 */ {(15<<2)|3,{105,97,97}},
/* 12488 */ {(15<<2)|3,{105,97,99}},
/* 12489 */ {(15<<2)|3,{105,97,99}},
/* 12490 */ {(15<<2)|3,{105,97,101}},
/* 12491 */ {(15<<2)|3,{105,97,101}},
/* 12492 */ {(15<<2)|3,{105,97,105}},
/* 12493 */ {(15<<2)|3,{105,97,105}},
/* 12494 */ {(15<<2)|3,{105,97,111}},
/* 12495 */ {(15<<2)|3,{105,97,111}},
/* 12496 */ {(15<<2)|3,{105,97,115}},
/* 12497 */ {(15<<2)|3,{105,97,115}},
/* 12498 */ {(15<<2)|3,{105,97,116}},
/* 12499 */ {(15<<2)|3,{105,97,116}},
/* 12500 */ {(16<<2)|3,{105,97,32}},
/* 12501 */ {(16<<2)|3,{105,97,37}},
/* 12502 */ {(16<<2)|3,{105,97,45}},
/* 12503 */ {(16<<2)|3,{105,97,46}},
/* 12504 */ {(16<<2)|3,{105,97,47}},
/* 12505 */ {(16<<2)|3,{105,97,51}},
/* 12506 */ {(16<<2)|3,{105,97,52}},
/* 12507 */ {(16<<2)|3,{105,97,53}},
/* 12508 */ {(16<<2)|3,{105,97,54}},
/* 12509 */ {(16<<2)|3,{105,97,55}},
/* 12510 */ {(16<<2)|3,{105,97,56}},
/* 12511 */ {(16<<2)|3,{105,97,57}},
/* 12512 */ {(16<<2)|3,{105,97,61}},
/* 12513 */ {(16<<2)|3,{105,97,65}},
/* 12514 */ {(16<<2)|3,{105,97,95}},
/* 12515 */ {(16<<2)|3,{105,97,98}},
/* 12516 */ {(16<<2)|3,{105,97,100}},
/* 12517 */ {(16<<2)|3,{105,97,102}},
/* 12518 */ {(16<<2)|3,{105,97,103}},
/* 12519 */ {(16<<2)|3,{105,97,104}},
/* 12520 */ {(16<<2)|3,{105,97,108}},
/* 12521 */ {(16<<2)|3,{105,97,109}},
/* 12522 */ {(16<<2)|3,{105,97,110}},
/* 12523 */ {(16<<2)|3,{105,97,112}},
/* 12524 */ {(16<<2)|3,{105,97,114}},
/* 12525 */ {(16<<2)|3,{105,97,117}},
/* 12526 */ {(10<<2)|2,{105,97,0}},
/* 12527 */ {(10<<2)|2,{105,97,0}},
/* 12528 */ {(10<<2)|2,{105,97,0}},
/* 12529 */ {(10<<2)|2,{105,97,0}},
/* 12530 */ {(10<<2)|2,{105,97,0}},
/* 12531 */ {(10<<2)|2,{105,97,0}},
/* 12532 */ {(10<<2)|2,{105,97,0}},
/* 12533 */ {(10<<2)|2,{105,97,0}},
/* 12534 */ {(10<<2)|2,{105,97,0}},
/* 12535 */ {(10<<2)|2,{105,97,0}},
/* 12536 */ {(10<<2)|2,{105,97,0}},
/* 12537 */ {(10<<2)|2,{105,97,0}},
/* 12538 */ {(10<<2)|2,{105,97,0}},
/* 12539 */ {(10<<2)|2,{105,97,0}},
/* 12540 */ {(10<<2)|2,{105,97,0}},
/* 12541 */ {(10<<2)|2,{105,97,0}},
/* 12542 */ {(10<<2)|2,{105,97,0}},
/* 12543 */ {(10<<2)|2,{105,97,0}},
/* 12544 */ {(15<<2)|3,{105,99,48}},
/* 12545 */ {(15<<2)|3,{105,99,48}},
/* 12546 */ {(15<<2)|3,{105,99,49}},
/* 12547 */ {(15<<2)|3,{105,99,49}},
/* 12548 */ {(15<<2)|3,{105,99,50}},
/* 12549 */ {(15<<2)|3,{105,99,50}},
/* 12550 */ {(15<<2)|3,{105,99,97}},
/* 12551 */ {(15<<2)|3,{105,99,97}},
/* 12552 */ {(15<<2)|3,{105,99,99}},
/* 12553 */ {(15<<2)|3,{105,99,99}},
/* 12554 */ {(15<<2)|3,{105,99,101}},
/* 12555 */ {(15<<2)|3,{105,99,101}},
/* 12556 */ {(15<<2)|3,{105,99,105}},
/* 12557 */ {(15<<2)|3,{105,99,105}},
/* 12558 */ {(15<<2)|3,{105,99,111}},
/* 12559 */ {(15<<2)|3,{105,99,111}},
/* 12560 */ {(15<<2)|3,{105,99,115}},
/* 12561 */ {(15<<2)|3,{105,99,115}},
/* 12562 */ {(15<<2)|3,{105,99,116}},
/* 12563 */ {(15<<2)|3,{105,99,116}},
/* 12564 */ {(16<<2)|3,{105,99,32}},
/* 12565 */ {(16<<2)|3,{105,99,37}},
/* 12566 */ {(16<<2)|3,{105,99,45}},
/* 12567 */ {(16<<2)|3,{105,99,46}},
/* 12568 */ {(16<<2)|3,{105,99,47}},
/* 12569 */ {(16<<2)|3,{105,99,51}},
/* 12570 */ {(16<<2)|3,{105,99,52}},
/* 12571 */ {(16<<2)|3,{105,99,53}},
/* 12572 */ {(16<<2)|3,{105,99,54}},
/* 12573 */ {(16<<2)|3,{105,99,55}},
/* 12574 */ {(16<<2)|3,{105,99,56}},
/* 12575 */ {(16<<2)|3,{105,99,57}},
/* 12576 */ {(16<<2)|3,{105,99,61}},
/* 12577 */ {(16<<2)|3,{105,99,65}},
/* 12578 */ {(16<<2)|3,{105,99,95}},
/* 12579 */ {(16<<2)|3,{105,99,98}},
/* 12580 */ {(16<<2)|3,{105,99,100}},
/* 12581 */ {(16<<2)|3,{105,99,102}},
/* 12582 */ {(16<<2)|3,{105,99,103}},
/* 12583 */ {(16<<2)|3,{105,99,104}},
/* 12584 */ {(16<<2)|3,{105,99,108}},
/* 12585 */ {(16<<2)|3,{105,99,109}},
/* 12586 */ {(16<<2)|3,{105,99,110}},
/* 12587 */ {(16<<2)|3,{105,99,112}},
/* 12588 */ {(16<<2)|3,{105,99,114}},
/* 12589 */ {(16<<2)|3,{105,99,117}},
/* 12590 */ {(10<<2)|2,{105,99,0}},
/* 12591 */ {(10<<2)|2,{105,99,0}},
/* 12592 */ {(10<<2)|2,{105,99,0}},
/* 12593 */ {(10<<2)|2,{105,99,0}},
/* 12594 */ {(10<<2)|2,{105,99,0}},
/* 12595 */ {(10<<2)|2,{105,99,0}},
/* 12596 */ {(10<<2)|2,{105,99,0}},
/* 12597 */ {(10<<2)|2,{105,99,0}},
/* 12598 */ {(10<<2)|2,{105,99,0}},
/* 12599 */ {(10<<2)|2,{105,99,0}},
/* 12600 */ {(10<<2)|2,{105,99,0}},
/* 12601 */ {(10<<2)|2,{105,99,0}},
/* 12602 */ {(10<<2)|2,{105,99,0}},
/* 12603 */ {(10<<2)|2,{105,99,0}},
/* 12604 */ {(10<<2)|2,{105,99,0}},
/* 12605 */ {(10<<2)|2,{105,99,0}},
/* 12606 */ {(10<<2)|2,{105,99,0}},
/* 12607 */ {(10<<2)|2,{105,99,0}},
/* 12608 */ {(15<<2)|3,{105,101,48}},
/* 12609 */ {(15<<2)|3,{105,101,48}},
/* 12610 */ {(15<<2)|3,{105,101,49}},
/* 12611 */ {(15<<2)|3,{105,101,49}},
/* 12612 */ {(15<<2)|3,{105,101,50}},
/* 12613 */ {(15<<2)|3,{105,101,50}},
/* 12614 */ {(15<<2)|3,{105,101,97}},
/* 12615 */ {(15<<2)|3,{105,101,97}},
/* 12616 */ {(15<<2)|3,{105,101,99}},
/* 12617 */ {(15<<2)|3,{105,101,99}},
/* 12618 */ {(15<<2)|3,{105,101,101}},
/* 12619 */ {(15<<2)|3,{105,101,101}},
/* 12620 */ {(15<<2)|3,{105,101,105}},
/* 12621 */ {(15<<2)|3,{105,101,105}},
/* 12622 */ {(15<<2)|3,{105,101,111}},
/* 12623 */ {(15<<2)|3,{105,101,111}},
/* 12624 */ {(15<<2)|3,{105,101,115}},
/* 12625 */ {(15<<2)|3,{105,101,115}},
/* 12626 */ {(15<<2)|3,{105,101,116}},
/* 12627 */ {(15<<2)|3,{105,101,116}},
/* 12628 */ {(16<<2)|3,{105,101,32}},
/* 12629 */ {(16<<2)|3,{105,101,37}},
/* 12630 */ {(16<<2)|3,{105,101,45}},
/* 12631 */ {(16<<2)|3,{105,101,46}},
/* 12632 */ {(16<<2)|3,{105,101,47}},
/* 12633 */ {(16<<2)|3,{105,101,51}},
/* 12634 */ {(16<<2)|3,{105,101,52}},
/* 12635 */ {(16<<2)|3,{105,101,53}},
/* 12636 */ {(16<<2)|3,{105,101,54}},
/* 12637 */ {(16<<2)|3,{105,101,55}},
/* 12638 */ {(16<<2)|3,{105,101,56}},
/* 12639 */ {(16<<2)|3,{105,101,57}},
/* 12640 */ {(16<<2)|3,{105,101,61}},
/* 12641 */ {(16<<2)|3,{105,101,65}},
/* 12642 */ {(16<<2)|3,{105,101,95}},
/* 12643 */ {(16<<2)|3,{105,101,98}},
/* 12644 */ {(16<<2)|3,{105,101,100}},
/* 12645 */ {(16<<2)|3,{105,101,102}},
/* 12646 */ {(16<<2)|3,{105,101,103}},
/* 12647 */ {(16<<2)|3,{105,101,104}},
/* 12648 */ {(16<<2)|3,{105,101,108}},
/* 12649 */ {(16<<2)|3,{105,101,109}},
/* 12650 */ {(16<<2)|3,{105,101,110}},
/* 12651 */ {(16<<2)|3,{105,101,112}},
/* 12652 */ {(16<<2)|3,{105,101,114}},
/* 12653 */ {(16<<2)|3,{105,101,117}},
/* 12654 */ {(10<<2)|2,{105,101,0}},
/* 12655 */ {(10<<2)|2,{105,101,0}},
/* 12656 */ {(10<<2)|2,{105,101,0}},
/* 12657 */ {(10<<2)|2,{105,101,0}},
/* 12658 */ {(10<<2)|2,{105,101,0}},
/* 12659 */ {(10<<2)|2,{105,101,0}},
/* 12660 */ {(10<<2)|2,{105,101,0}},
/* 12661 */ {(10<<2)|2,{105,101,0}},
/* 12662 */ {(10<<2)|2,{105,101,0}},
/* 12663 */ {(10<<2)|2,{105,101,0}},
/* 12664 */ {(10<<2)|2,{105,101,0}},
/* 12665 */ {(10<<2)|2,{105,101,0}},
/* 12666 */ {(10<<2)|2,{105,101,0}},
/* 12667 */ {(10<<2)|2,{105,101,0}},
/* 12668 */ {(10<<2)|2,{105,101,0}},
/* 12669 */ {(10<<2)|2,{105,101,0}},
/* 12670 */ {(10<<2)|2,{105,101,0}},
/* 12671 */ {(10<<2)|2,{105,101,0}},
/* 12672 */ {(15<<2)|3,{105,105,48}},
/* 12673 */ {(15<<2)|3,{105,105,48}},
/* 12674 */ {(15<<2)|3,{105,105,49}},
/* 12675 */ {(15<<2)|3,{105,105,49}},
/* 12676 */ {(15<<2)|3,{105,105,50}},
/* 12677 */ {(15<<2)|3,{105,105,50}},
/* 12678 */ {(15<<2)|3,{105,105,97}},
/* 12679 */ {(15<<2)|3,{105,105,97}},
/* 12680 */ {(15<<2)|3,{105,105,99}},
/* 12681 */ {(15<<2)|3,{105,105,99}},
/* 12682 */ {(15<<2)|3,{105,105,101}},
/* 12683 */ {(15<<2)|3,{105,105,101}},
/* 12684 */ {(15<<2)|3,{105,105,105}},
/* 12685 */ {(15<<2)|3,{105,105,105}},
/* 12686 */ {(15<<2)|3,{105,105,111}},
/* 12687 */ {(15<<2)|3,{105,105,111}},
/* 12688 */ {(15<<2)|3,{105,105,115}},
/* 12689 */ {(15<<2)|3,{105,105,115}},
/* 12690 */ {(15<<2)|3,{105,105,116}},
/* 12691 */ {(15<<2)|3,{105,105,116}},
/* 12692 */ {(16<<2)|3,{105,105,32}},
/* 12693 */ {(16<<2)|3,{105,105,37}},
/* 12694 */ {(16<<2)|3,{105,105,45}},
/* 12695 */ {(16<<2)|3,{105,105,46}},
/* 12696 */ {(16<<2)|3,{105,105,47}},
/* 12697 */ {(16<<2)|3,{105,105,51}},
/* 12698 */ {(16<<2)|3,{105,105,52}},
/* 12699 */ {(16<<2)|3,{105,105,53}},
/* 12700 */ {(16<<2)|3,{105,105,54}},
/* 12701 */ {(16<<2)|3,{105,105,55}},
/* 12702 */ {(16<<2)|3,{105,105,56}},
/* 12703 */ {(16<<2)|3,{105,105,57}},
/* 12704 */ {(16<<2)|3,{105,105,61}},
/* 12705 */ {(16<<2)|3,{105,105,65}},
/* 12706 */ {(16<<2)|3,{105,105,95}},
/* 12707 */ {(16<<2)|3,{105,105,98}},
/* 12708 */ {(16<<2)|3,{105,105,100}},
/* 12709 */ {(16<<2)|3,{105,105,102}},
/* 12710 */ {(16<<2)|3,{105,105,103}},
/* 12711 */ {(16<<2)|3,{105,105,104}},
/* 12712 */ {(16<<2)|3,{105,105,108}},
/* 12713 */ {(16<<2)|3,{105,105,109}},
/* 12714 */ {(16<<2)|3,{105,105,110}},
/* 12715 */ {(16<<2)|3,{105,105,112}},
/* 12716 */ {(16<<2)|3,{105,105,114}},
/* 12717 */ {(16<<2)|3,{105,105,117}},
/* 12718 */ {(10<<2)|2,{105,105,0}},
/* 12719 */ {(10<<2)|2,{105,105,0}},
/* 12720 */ {(10<<2)|2,{105,105,0}},
/* 12721 */ {(10<<2)|2,{105,105,0}},
/* 12722 */ {(10<<2)|2,{105,105,0}},
/* 12723 */ {(10<<2)|2,{105,105,0}},
/* 12724 */ {(10<<2)|2,{105,105,0}},
/* 12725 */ {(10<<2)|2,{105,105,0}},
/* 12726 */ {(10<<2)|2,{105,105,0}},
/* 12727 */ {(10<<2)|2,{105,105,0}},
/* 12728 */ {(10<<2)|2,{105,105,0}},
/* 12729 */ {(10<<2)|2,{105,105,0}},
/* 12730 */ {(10<<2)|2,{105,105,0}},
/* 12731 */ {(10<<2)|2,{105,105,0}},
/* 12732 */ {(10<<2)|2,{105,105,0}},
/* 12733 */ {(10<<2)|2,{105,105,0}},
/* 12734 */ {(10<<2)|2,{105,105,0}},
/* 12735 */ {(10<<2)|2,{105,105,0}},
/* 12736 */ {(15<<2)|3,{105,111,48}},
/* 12737 */ {(15<<2)|3,{105,111,48}},
/* 12738 */ {(15<<2)|3,{105,111,49}},
/* 12739 */ {(15<<2)|3,{105,111,49}},
/* 12740 */ {(15<<2)|3,{105,111,50}},
/* 12741 */ {(15<<2)|3,{105,111,50}},
/* 12742 */ {(15<<2)|3,{105,111,97}},
/* 12743 */ {(15<<2)|3,{105,111,97}},
/* 12744 */ {(15<<2)|3,{105,111,99}},
/* 12745 */ {(15<<2)|3,{105,111,99}},
/* 12746 */ {(15<<2)|3,{105,111,101}},
/* 12747 */ {(15<<2)|3,{105,111,101}},
/* 12748 */ {(15<<2)|3,{105,111,105}},
/* 12749 */ {(15<<2)|3,{105,111,105}},
/* 12750 */ {(15<<2)|3,{105,111,111}},
/* 12751 */ {(15<<2)|3,{105,111,111}},
/* 12752 */ {(15<<2)|3,{105,111,115}},
/* 12753 */ {(15<<2)|3,{105,111,115}},
/* 12754 */ {(15<<2)|3,{105,111,116}},
/* 12755 */ {(15<<2)|3,{105,111,116}},
/* 12756 */ {(16<<2)|3,{105,111,32}},
/* 12757 */ {(16<<2)|3,{105,111,37}},
/* 12758 */ {(16<<2)|3,{105,111,45}},
/* 12759 */ {(16<<2)|3,{105,111,46}},
/* 12760 */ {(16<<2)|3,{105,111,47}},
/* 12761 */ {(16<<2)|3,{105,111,51}},
/* 12762 */ {(16<<2)|3,{105,111,52}},
/* 12763 */ {(16<<2)|3,{105,111,53}},
/* 12764 */ {(16<<2)|3,{105,111,54}},
/* 12765 */ {(16<<2)|3,{105,111,55}},
/* 12766 */ {(16<<2)|3,{105,111,56}},
/* 12767 */ {(16<<2)|3,{105,111,57}},
/* 12768 */ {(16<<2)|3,{105,111,61}},
/* 12769 */ {(16<<2)|3,{105,111,65}},
/* 12770 */ {(16<<2)|3,{105,111,95}},
/* 12771 */ {(16<<2)|3,{105,111,98}},
/* 12772 */ {(16<<2)|3,{105,111,100}},
/* 12773 */ {(16<<2)|3,{105,111,102}},
/* 12774 */ {(16<<2)|3,{105,111,103}},
/* 12775 */ {(16<<2)|3,{105,111,104}},
/* 12776 */ {(16<<2)|3,{105,111,108}},
/* 12777 */ {(16<<2)|3,{105,111,109}},
/* 12778 */ {(16<<2)|3,{105,111,110}},
/* 12779 */ {(16<<2)|3,{105,111,112}},
/* 12780 */ {(16<<2)|3,{105,111,114}},
/* 12781 */ {(16<<2)|3,{105,111,117}},
/* 12782 */ {(10<<2)|2,{105,111,0}},
/* 12783 */ {(10<<2)|2,{105,111,0}},
/* 12784 */ {(10<<2)|2,{105,111,0}},
/* 12785 */ {(10<<2)|2,{105,111,0}},
/* 12786 */ {(10<<2)|2,{105,111,0}},
/* 12787 */ {(10<<2)|2,{105,111,0}},
/* 12788 */ {(10<<2)|2,{105,111,0}},
/* 12789 */ {(10<<2)|2,{105,111,0}},
/* 12790 */ {(10<<2)|2,{105,111,0}},
/* 12791 */ {(10<<2)|2,{105,111,0}},
/* 12792 */ {(10<<2)|2,{105,111,0}},
/* 12793 */ {(10<<2)|2,{105,111,0}},
/* 12794 */ {(10<<2)|2,{105,111,0}},
/* 12795 */ {(10<<2)|2,{105,111,0}},
/* 12796 */ {(10<<2)|2,{105,111,0}},
/* 12797 */ {(10<<2)|2,{105,111,0}},
/* 12798 */ {(10<<2)|2,{105,111,0}},
/* 12799 */ {(10<<2)|2,{105,111,0}},
/* 12800 */ {(15<<2)|3,{105,115,48}},
/* 12801 */ {(15<<2)|3,{105,115,48}},
/* 12802 */ {(15<<2)|3,{105,115,49}},
/* 12803 */ {(15<<2)|3,{105,115,49}},
/* 12804 */ {(15<<2)|3,{105,115,50}},
/* 12805 */ {(15<<2)|3,{105,115,50}},
/* 12806 */ {(15<<2)|3,{105,115,97}},
/* 12807 */ {(15<<2)|3,{105,115,97}},
/* 12808 */ {(15<<2)|3,{105,115,99}},
/* 12809 */ {(15<<2)|3,{105,115,99}},
/* 12810 */ {(15<<2)|3,{105,115,101}},
/* 12811 */ {(15<<2)|3,{105,115,101}},
/* 12812 */ {(15<<2)|3,{105,115,105}},
/* 12813 */ {(15<<2)|3,{105,115,105}},
/* 12814 */ {(15<<2)|3,{105,115,111}},
/* 12815 */ {(15<<2)|3,{105,115,111}},
/* 12816 */ {(15<<2)|3,{105,115,115}},
/* 12817 */ {(15<<2)|3,{105,115,115}},
/* 12818 */ {(15<<2)|3,{105,115,116}},
/* 12819 */ {(15<<2)|3,{105,115,116}},
/* 12820 */ {(16<<2)|3,{105,115,32}},
/* 12821 */ {(16<<2)|3,{105,115,37}},
/* 12822 */ {(16<<2)|3,{105,115,45}},
/* 12823 */ {(16<<2)|3,{105,115,46}},
/* 12824 */ {(16<<2)|3,{105,115,47}},
/* 12825 */ {(16<<2)|3,{105,115,51}},
/* 12826 */ {(16<<2)|3,{105,115,52}},
/* 12827 */ {(16<<2)|3,{105,115,53}},
/* 12828 */ {(16<<2)|3,{105,115,54}},
/* 12829 */ {(16<<2)|3,{105,115,55}},
/* 12830 */ {(16<<2)|3,{105,115,56}},
/* 12831 */ {(16<<2)|3,{105,115,57}},
/* 12832 */ {(16<<2)|3,{105,115,61}},
/* 12833 */ {(16<<2)|3,{105,115,65}},
/* 12834 */ {(16<<2)|3,{105,115,95}},
/* 12835 */ {(16<<2)|3,{105,115,98}},
/* 12836 */ {(16<<2)|3,{105,115,100}},
/* 12837 */ {(16<<2)|3,{105,115,102}},
/* 12838 */ {(16<<2)|3,{105,115,103}},
/* 12839 */ {(16<<2)|3,{105,115,104}},
/* 12840 */ {(16<<2)|3,{105,115,108}},
/* 12841 */ {(16<<2)|3,{105,115,109}},
/* 12842 */ {(16<<2)|3,{105,115,110}},
/* 12843 */ {(16<<2)|3,{105,115,112}},
/* 12844 */ {(16<<2)|3,{105,115,114}},
/* 12845 */ {(16<<2)|3,{105,115,117}},
/* 12846 */ {(10<<2)|2,{105,115,0}},
/* 12847 */ {(10<<2)|2,{105,115,0}},
/* 12848 */ {(10<<2)|2,{105,115,0}},
/* 12849 */ {(10<<2)|2,{105,115,0}},
/* 12850 */ {(10<<2)|2,{105,115,0}},
/* 12851 */ {(10<<2)|2,{105,115,0}},
/* 12852 */ {(10<<2)|2,{105,115,0}},
/* 12853 */ {(10<<2)|2,{105,115,0}},
/* 12854 */ {(10<<2)|2,{105,115,0}},
/* 12855 */ {(10<<2)|2,{105,115,0}},
/* 12856 */ {(10<<2)|2,{105,115,0}},
/* 12857 */ {(10<<2)|2,{105,115,0}},
/* 12858 */ {(10<<2)|2,{105,115,0}},
/* 12859 */ {(10<<2)|2,{105,115,0}},
/* 12860 */ {(10<<2)|2,{105,115,0}},
/* 12861 */ {(10<<2)|2,{105,115,0}},
/* 12862 */ {(10<<2)|2,{105,115,0}},
/* 12863 */ {(10<<2)|2,{105,115,0}},
/* 12864 */ {(15<<2)|3,{105,116,48}},
/* 12865 */ {(15<<2)|3,{105,116,48}},
/* 12866 */ {(15<<2)|3,{105,116,49}},
/* 12867 */ {(15<<2)|3,{105,116,49}},
/* 12868 */ {(15<<2)|3,{105,116,50}},
/* 12869 */ {(15<<2)|3,{105,116,50}},
/* 12870 */ {(15<<2)|3,{105,116,97}},
/* 12871 */ {(15<<2)|3,{105,116,97}},
/* 12872 */ {(15<<2)|3,{105,116,99}},
/* 12873 */ {(15<<2)|3,{105,116,99}},
/* 12874 */ {(15<<2)|3,{105,116,101}},
/* 12875 */ {(15<<2)|3,{105,116,101}},
/* 12876 */ {(15<<2)|3,{105,116,105}},
/* 12877 */ {(15<<2)|3,{105,116,105}},
/* 12878 */ {(15<<2)|3,{105,116,111}},
/* 12879 */ {(15<<2)|3,{105,116,111}},
/* 12880 */ {(15<<2)|3,{105,116,115}},
/* 12881 */ {(15<<2)|3,{105,116,115}},
/* 12882 */ {(15<<2)|3,{105,116,116}},
/* 12883 */ {(15<<2)|3,{105,116,116}},
/* 12884 */ {(16<<2)|3,{105,116,32}},
/* 12885 */ {(16<<2)|3,{105,116,37}},
/* 12886 */ {(16<<2)|3,{105,116,45}},
/* 12887 */ {(16<<2)|3,{105,116,46}},
/* 12888 */ {(16<<2)|3,{105,116,47}},
/* 12889 */ {(16<<2)|3,{105,116,51}},
/* 12890 */ {(16<<2)|3,{105,116,52}},
/* 12891 */ {(16<<2)|3,{105,116,53}},
/* 12892 */ {(16<<2)|3,{105,116,54}},
/* 12893 */ {(16<<2)|3,{105,116,55}},
/* 12894 */ {(16<<2)|3,{105,116,56}},
/* 12895 */ {(16<<2)|3,{105,116,57}},
/* 12896 */ {(16<<2)|3,{105,116,61}},
/* 12897 */ {(16<<2)|3,{105,116,65}},
/* 12898 */ {(16<<2)|3,{105,116,95}},
/* 12899 */ {(16<<2)|3,{105,116,98}},
/* 12900 */ {(16<<2)|3,{105,116,100}},
/* 12901 */ {(16<<2)|3,{105,116,102}},
/* 12902 */ {(16<<2)|3,{105,116,103}},
/* 12903 */ {(16<<2)|3,{105,116,104}},
/* 12904 */ {(16<<2)|3,{105,116,108}},
/* 12905 */ {(16<<2)|3,{105,116,109}},
/* 12906 */ {(16<<2)|3,{105,116,110}},
/* 12907 */ {(16<<2)|3,{105,116,112}},
/* 12908 */ {(16<<2)|3,{105,116,114}},
/* 12909 */ {(16<<2)|3,{105,116,117}},
/* 12910 */ {(10<<2)|2,{105,116,0}},
/* 12911 */ {(10<<2)|2,{105,116,0}},
/* 12912 */ {(10<<2)|2,{105,116,0}},
/* 12913 */ {(10<<2)|2,{105,116,0}},
/* 12914 */ {(10<<2)|2,{105,116,0}},
/* 12915 */ {(10<<2)|2,{105,116,0}},
/* 12916 */ {(10<<2)|2,{105,116,0}},
/* 12917 */ {(10<<2)|2,{105,116,0}},
/* 12918 */ {(10<<2)|2,{105,116,0}},
/* 12919 */ {(10<<2)|2,{105,116,0}},
/* 12920 */ {(10<<2)|2,{105,116,0}},
/* 12921 */ {(10<<2)|2,{105,116,0}},
/* 12922 */ {(10<<2)|2,{105,116,0}},
/* 12923 */ {(10<<2)|2,{105,116,0}},
/* 12924 */ {(10<<2)|2,{105,116,0}},
/* 12925 */ {(10<<2)|2,{105,116,0}},
/* 12926 */ {(10<<2)|2,{105,116,0}},
/* 12927 */ {(10<<2)|2,{105,116,0}},
/* 12928 */ {(16<<2)|3,{105,32,48}},
/* 12929 */ {(16<<2)|3,{105,32,49}},
/* 12930 */ {(16<<2)|3,{105,32,50}},
/* 12931 */ {(16<<2)|3,{105,32,97}},
/* 12932 */ {(16<<2)|3,{105,32,99}},
/* 12933 */ {(16<<2)|3,{105,32,101}},
/* 12934 */ {(16<<2)|3,{105,32,105}},
/* 12935 */ {(16<<2)|3,{105,32,111}},
/* 12936 */ {(16<<2)|3,{105,32,115}},
/* 12937 */ {(16<<2)|3,{105,32,116}},
/* 12938 */ {(11<<2)|2,{105,32,0}},
/* 12939 */ {(11<<2)|2,{105,32,0}},
/* 12940 */ {(11<<2)|2,{105,32,0}},
/* 12941 */ {(11<<2)|2,{105,32,0}},
/* 12942 */ {(11<<2)|2,{105,32,0}},
/* 12943 */ {(11<<2)|2,{105,32,0}},
/* 12944 */ {(11<<2)|2,{105,32,0}},
/* 12945 */ {(11<<2)|2,{105,32,0}},
/* 12946 */ {(11<<2)|2,{105,32,0}},
/* 12947 */ {(11<<2)|2,{105,32,0}},
/* 12948 */ {(11<<2)|2,{105,32,0}},
/* 12949 */ {(11<<2)|2,{105,32,0}},
/* 12950 */ {(11<<2)|2,{105,32,0}},
/* 12951 */ {(11<<2)|2,{105,32,0}},
/* 12952 */ {(11<<2)|2,{105,32,0}},
/* 12953 */ {(11<<2)|2,{105,32,0}},
/* 12954 */ {(11<<2)|2,{105,32,0}},
/* 12955 */ {(11<<2)|2,{105,32,0}},
/* 12956 */ {(11<<2)|2,{105,32,0}},
/* 12957 */ {(11<<2)|2,{105,32,0}},
/* 12958 */ {(11<<2)|2,{105,32,0}},
/* 12959 */ {(11<<2)|2,{105,32,0}},
/* 12960 */ {(16<<2)|3,{105,37,48}},
/* 12961 */ {(16<<2)|3,{105,37,49}},
/* 12962 */ {(16<<2)|3,{105,37,50}},
/* 12963 */ {(16<<2)|3,{105,37,97}},
/* 12964 */ {(16<<2)|3,{105,37,99}},
/* 12965 */ {(16<<2)|3,{105,37,101}},
/* 12966 */ {(16<<2)|3,{105,37,105}},
/* 12967 */ {(16<<2)|3,{105,37,111}},
/* 12968 */ {(16<<2)|3,{105,37,115}},
/* 12969 */ {(16<<2)|3,{105,37,116}},
/* 12970 */ {(11<<2)|2,{105,37,0}},
/* 12971 */ {(11<<2)|2,{105,37,0}},
/* 12972 */ {(11<<2)|2,{105,37,0}},
/* 12973 */ {(11<<2)|2,{105,37,0}},
/* 12974 */ {(11<<2)|2,{105,37,0}},
/* 12975 */ {(11<<2)|2,{105,37,0}},
/* 12976 */ {(11<<2)|2,{105,37,0}},
/* 12977 */ {(11<<2)|2,{105,37,0}},
/* 12978 */ {(11<<2)|2,{105,37,0}},
/* 12979 */ {(11<<2)|2,{105,37,0}},
/* 12980 */ {(11<<2)|2,{105,37,0}},
/* 12981 */ {(11<<2)|2,{105,37,0}},
/* 12982 */ {(11<<2)|2,{105,37,0}},
/* 12983 */ {(11<<2)|2,{105,37,0}},
/* 12984 */ {(11<<2)|2,{105,37,0}},
/* 12985 */ {(11<<2)|2,{105,37,0}},
/* 12986 */ {(11<<2)|2,{105,37,0}},
/* 12987 */ {(11<<2)|2,{105,37,0}},
/* 12988 */ {(11<<2)|2,{105,37,0}},
/* 12989 */ {(11<<2)|2,{105,37,0}},
/* 12990 */ {(11<<2)|2,{105,37,0}},
/* 12991 */ {(11<<2)|2,{105,37,0}},
/* 12992 */ {(16<<2)|3,{105,45,48}},
/* 12993 */ {(16<<2)|3,{105,45,49}},
/* 12994 */ {(16<<2)|3,{105,45,50}},
/* 12995 */ {(16<<2)|3,{105,45,97}},
/* 12996 */ {(16<<2)|3,{105,45,99}},
/* 12997 */ {(16<<2)|3,{105,45,101}},
/* 12998 */ {(16<<2)|3,{105,45,105}},
/* 12999 */ {(16<<2)|3,{105,45,111}},
/* 13000 */ {(16<<2)|3,{105,45,115}},
/* 13001 */ {(16<<2)|3,{105,45,116}},
/* 13002 */ {(11<<2)|2,{105,45,0}},
/* 13003 */ {(11<<2)|2,{105,45,0}},
/* 13004 */ {(11<<2)|2,{105,45,0}},
/* 13005 */ {(11<<2)|2,{105,45,0}},
/* 13006 */ {(11<<2)|2,{105,45,0}},
/* 13007 */ {(11<<2)|2,{105,45,0}},
/* 13008 */ {(11<<2)|2,{105,45,0}},
/* 13009 */ {(11<<2)|2,{105,45,0}},
/* 13010 */ {(11<<2)|2,{105,45,0}},
/* 13011 */ {(11<<2)|2,{105,45,0}},
/* 13012 */ {(11<<2)|2,{105,45,0}},
/* 13013 */ {(11<<2)|2,{105,45,0}},
/* 13014 */ {(11<<2)|2,{105,45,0}},
/* 13015 */ {(11<<2)|2,{105,45,0}},
/* 13016 */ {(11<<2)|2,{105,45,0}},
/* 13017 */ {(11<<2)|2,{105,45,0}},
/* 13018 */ {(11<<2)|2,{105,45,0}},
/* 13019 */ {(11<<2)|2,{105,45,0}},
/* 13020 */ {(11<<2)|2,{105,45,0}},
/* 13021 */ {(11<<2)|2,{105,45,0}},
/* 13022 */ {(11<<2)|2,{105,45,0}},
/* 13023 */ {(11<<2)|2,{105,45,0}},
/* 13024 */ {(16<<2)|3,{105,46,48}},
/* 13025 */ {(16<<2)|3,{105,46,49}},
/* 13026 */ {(16<<2)|3,{105,46,50}},
/* 13027 */ {(16<<2)|3,{105,46,97}},
/* 13028 */ {(16<<2)|3,{105,46,99}},
/* 13029 */ {(16<<2)|3,{105,46,101}},
/* 13030 */ {(16<<2)|3,{105,46,105}},
/* 13031 */ {(16<<2)|3,{105,46,111}},
/* 13032 */ {(16<<2)|3,{105,46,115}},
/* 13033 */ {(16<<2)|3,{105,46,116}},
/* 13034 */ {(11<<2)|2,{105,46,0}},
/* 13035 */ {(11<<2)|2,{105,46,0}},
/* 13036 */ {(11<<2)|2,{105,46,0}},
/* 13037 */ {(11<<2)|2,{105,46,0}},
/* 13038 */ {(11<<2)|2,{105,46,0}},
/* 13039 */ {(11<<2)|2,{105,46,0}},
/* 13040 */ {(11<<2)|2,{105,46,0}},
/* 13041 */ {(11<<2)|2,{105,46,0}},
/* 13042 */ {(11<<2)|2,{105,46,0}},
/* 13043 */ {(11<<2)|2,{105,46,0}},
/* 13044 */ {(11<<2)|2,{105,46,0}},
/* 13045 */ {(11<<2)|2,{105,46,0}},
/* 13046 */ {(11<<2)|2,{105,46,0}},
/* 13047 */ {(11<<2)|2,{105,46,0}},
/* 13048 */ {(11<<2)|2,{105,46,0}},
/* 13049 */ {(11<<2)|2,{105,46,0}},
/* 13050 */ {(11<<2)|2,{105,46,0}},
/* 13051 */ {(11<<2)|2,{105,46,0}},
/* 13052 */ {(11<<2)|2,{105,46,0}},
/* 13053 */ {(11<<2)|2,{105,46,0}},
/* 13054 */ {(11<<2)|2,{105,46,0}},
/* 13055 */ {(11<<2)|2,{105,46,0}},
/* 13056 */ {(16<<2)|3,{105,47,48}},
/* 13057 */ {(16<<2)|3,{105,47,49}},
/* 13058 */ {(16<<2)|3,{105,47,50}},
/* 13059 */ {(16<<2)|3,{105,47,97}},
/* 13060 */ {(16<<2)|3,{105,47,99}},
/* 13061 */ {(16<<2)|3,{105,47,101}},
/* 13062 */ {(16<<2)|3,{105,47,105}},
/* 13063 */ {(16<<2)|3,{105,47,111}},
/* 13064 */ {(16<<2)|3,{105,47,115}},
/* 13065 */ {(16<<2)|3,{105,47,116}},
/* 13066 */ {(11<<2)|2,{105,47,0}},
/* 13067 */ {(11<<2)|2,{105,47,0}},
/* 13068 */ {(11<<2)|2,{105,47,0}},
/* 13069 */ {(11<<2)|2,{105,47,0}},
/* 13070 */ {(11<<2)|2,{105,47,0}},
/* 13071 */ {(11<<2)|2,{105,47,0}},
/* 13072 */ {(11<<2)|2,{105,47,0}},
/* 13073 */ {(11<<2)|2,{105,47,0}},
/* 13074 */ {(11<<2)|2,{105,47,0}},
/* 13075 */ {(11<<2)|2,{105,47,0}},
/* 13076 */ {(11<<2)|2,{105,47,0}},
/* 13077 */ {(11<<2)|2,{105,47,0}},
/* 13078 */ {(11<<2)|2,{105,47,0}},
/* 13079 */ {(11<<2)|2,{105,47,0}},
/* 13080 */ {(11<<2)|2,{105,47,0}},
/* 13081 */ {(11<<2)|2,{105,47,0}},
/* 13082 */ {(11<<2)|2,{105,47,0}},
/* 13083 */ {(11<<2)|2,{105,47,0}},
/* 13084 */ {(11<<2)|2,{105,47,0}},
/* 13085 */ {(11<<2)|2,{105,47,0}},
/* 13086 */ {(11<<2)|2,{105,47,0}},
/* 13087 */ {(11<<2)|2,{105,47,0}},
/* 13088 */ {(16<<2)|3,{105,51,48}},
/* 13089 */ {(16<<2)|3,{105,51,49}},
/* 13090 */ {(16<<2)|3,{105,51,50}},
/* 13091 */ {(16<<2)|3,{105,51,97}},
/* 13092 */ {(16<<2)|3,{105,51,99}},
/* 13093 */ {(16<<2)|3,{105,51,101}},
/* 13094 */ {(16<<2)|3,{105,51,105}},
/* 13095 */ {(16<<2)|3,{105,51,111}},
/* 13096 */ {(16<<2)|3,{105,51,115}},
/* 13097 */ {(16<<2)|3,{105,51,116}},
/* 13098 */ {(11<<2)|2,{105,51,0}},
/* 13099 */ {(11<<2)|2,{105,51,0}},
/* 13100 */ {(11<<2)|2,{105,51,0}},
/* 13101 */ {(11<<2)|2,{105,51,0}},
/* 13102 */ {(11<<2)|2,{105,51,0}},
/* 13103 */ {(11<<2)|2,{105,51,0}},
/* 13104 */ {(11<<2)|2,{105,51,0}},
/* 13105 */ {(11<<2)|2,{105,51,0}},
/* 13106 */ {(11<<2)|2,{105,51,0}},
/* 13107 */ {(11<<2)|2,{105,51,0}},
/* 13108 */ {(11<<2)|2,{105,51,0}},
/* 13109 */ {(11<<2)|2,{105,51,0}},
/* 13110 */ {(11<<2)|2,{105,51,0}},
/* 13111 */ {(11<<2)|2,{105,51,0}},
/* 13112 */ {(11<<2)|2,{105,51,0}},
/* 13113 */ {(11<<2)|2,{105,51,0}},
/* 13114 */ {(11<<2)|2,{105,51,0}},
/* 13115 */ {(11<<2)|2,{105,51,0}},
/* 13116 */ {(11<<2)|2,{105,51,0}},
/* 13117 */ {(11<<2)|2,{105,51,0}},
/* 13118 */ {(11<<2)|2,{105,51,0}},
/* 13119 */ {(11<<2)|2,{105,51,0}},
/* 13120 */ {(16<<2)|3,{105,52,48}},
/* 13121 */ {(16<<2)|3,{105,52,49}},
/* 13122 */ {(16<<2)|3,{105,52,50}},
/* 13123 */ {(16<<2)|3,{105,52,97}},
/* 13124 */ {(16<<2)|3,{105,52,99}},
/* 13125 */ {(16<<2)|3,{105,52,101}},
/* 13126 */ {(16<<2)|3,{105,52,105}},
/* 13127 */ {(16<<2)|3,{105,52,111}},
/* 13128 */ {(16<<2)|3,{105,52,115}},
/* 13129 */ {(16<<2)|3,{105,52,116}},
/* 13130 */ {(11<<2)|2,{105,52,0}},
/* 13131 */ {(11<<2)|2,{105,52,0}},
/* 13132 */ {(11<<2)|2,{105,52,0}},
/* 13133 */ {(11<<2)|2,{105,52,0}},
/* 13134 */ {(11<<2)|2,{105,52,0}},
/* 13135 */ {(11<<2)|2,{105,52,0}},
/* 13136 */ {(11<<2)|2,{105,52,0}},
/* 13137 */ {(11<<2)|2,{105,52,0}},
/* 13138 */ {(11<<2)|2,{105,52,0}},
/* 13139 */ {(11<<2)|2,{105,52,0}},
/* 13140 */ {(11<<2)|2,{105,52,0}},
/* 13141 */ {(11<<2)|2,{105,52,0}},
/* 13142 */ {(11<<2)|2,{105,52,0}},
/* 13143 */ {(11<<2)|2,{105,52,0}},
/* 13144 */ {(11<<2)|2,{105,52,0}},
/* 13145 */ {(11<<2)|2,{105,52,0}},
/* 13146 */ {(11<<2)|2,{105,52,0}},
/* 13147 */ {(11<<2)|2,{105,52,0}},
/* 13148 */ {(11<<2)|2,{105,52,0}},
/* 13149 */ {(11<<2)|2,{105,52,0}},
/* 13150 */ {(11<<2)|2,{105,52,0}},
/* 13151 */ {(11<<2)|2,{105,52,0}},
/* 13152 */ {(16<<2)|3,{105,53,48}},
/* 13153 */ {(16<<2)|3,{105,53,49}},
/* 13154 */ {(16<<2)|3,{105,53,50}},
/* 13155 */ {(16<<2)|3,{105,53,97}},
/* 13156 */ {(16<<2)|3,{105,53,99}},
/* 13157 */ {(16<<2)|3,{105,53,101}},
/* 13158 */ {(16<<2)|3,{105,53,105}},
/* 13159 */ {(16<<2)|3,{105,53,111}},
/* 13160 */ {(16<<2)|3,{105,53,115}},
/* 13161 */ {(16<<2)|3,{105,53,116}},
/* 13162 */ {(11<<2)|2,{105,53,0}},
/* 13163 */ {(11<<2)|2,{105,53,0}},
/* 13164 */ {(11<<2)|2,{105,53,0}},
/* 13165 */ {(11<<2)|2,{105,53,0}},
/* 13166 */ {(11<<2)|2,{105,53,0}},
/* 13167 */ {(11<<2)|2,{105,53,0}},
/* 13168 */ {(11<<2)|2,{105,53,0}},
/* 13169 */ {(11<<2)|2,{105,53,0}},
/* 13170 */ {(11<<2)|2,{105,53,0}},
/* 13171 */ {(11<<2)|2,{105,53,0}},
/* 13172 */ {(11<<2)|2,{105,53,0}},
/* 13173 */ {(11<<2)|2,{105,53,0}},
/* 13174 */ {(11<<2)|2,{105,53,0}},
/* 13175 */ {(11<<2)|2,{105,53,0}},
/* 13176 */ {(11<<2)|2,{105,53,0}},
/* 13177 */ {(11<<2)|2,{105,53,0}},
/* 13178 */ {(11<<2)|2,{105,53,0}},
/* 13179 */ {(11<<2)|2,{105,53,0}},
/* 13180 */ {(11<<2)|2,{105,53,0}},
/* 13181 */ {(11<<2)|2,{105,53,0}},
/* 13182 */ {(11<<2)|2,{105,53,0}},
/* 13183 */ {(11<<2)|2,{105,53,0}},
/* 13184 */ {(16<<2)|3,{105,54,48}},
/* 13185 */ {(16<<2)|3,{105,54,49}},
/* 13186 */ {(16<<2)|3,{105,54,50}},
/* 13187 */ {(16<<2)|3,{105,54,97}},
/* 13188 */ {(16<<2)|3,{105,54,99}},
/* 13189 */ {(16<<2)|3,{105,54,101}},
/* 13190 */ {(16<<2)|3,{105,54,105}},
/* 13191 */ {(16<<2)|3,{105,54,111}},
/* 13192 */ {(16<<2)|3,{105,54,115}},
/* 13193 */ {(16<<2)|3,{105,54,116}},
/* 13194 */ {(11<<2)|2,{105,54,0}},
/* 13195 */ {(11<<2)|2,{105,54,0}},
/* 13196 */ {(11<<2)|2,{105,54,0}},
/* 13197 */ {(11<<2)|2,{105,54,0}},
/* 13198 */ {(11<<2)|2,{105,54,0}},
/* 13199 */ {(11<<2)|2,{105,54,0}},
/* 13200 */ {(11<<2)|2,{105,54,0}},
/* 13201 */ {(11<<2)|2,{105,54,0}},
/* 13202 */ {(11<<2)|2,{105,54,0}},
/* 13203 */ {(11<<2)|2,{105,54,0}},
/* 13204 */ {(11<<2)|2,{105,54,0}},
/* 13205 */ {(11<<2)|2,{105,54,0}},
/* 13206 */ {(11<<2)|2,{105,54,0}},
/* 13207 */ {(11<<2)|2,{105,54,0}},
/* 13208 */ {(11<<2)|2,{105,54,0}},
/* 13209 */ {(11<<2)|2,{105,54,0}},
/* 13210 */ {(11<<2)|2,{105,54,0}},
/* 13211 */ {(11<<2)|2,{105,54,0}},
/* 13212 */ {(11<<2)|2,{105,54,0}},
/* 13213 */ {(11<<2)|2,{105,54,0}},
/* 13214 */ {(11<<2)|2,{105,54,0}},
/* 13215 */ {(11<<2)|2,{105,54,0}},
/* 13216 */ {(16<<2)|3,{105,55,48}},
/* 13217 */ {(16<<2)|3,{105,55,49}},
/* 13218 */ {(16<<2)|3,{105,55,50}},
/* 13219 */ {(16<<2)|3,{105,55,97}},
/* 13220 */ {(16<<2)|3,{105,55,99}},
/* 13221 */ {(16<<2)|3,{105,55,101}},
/* 13222 */ {(16<<2)|3,{105,55,105}},
/* 13223 */ {(16<<2)|3,{105,55,111}},
/* 13224 */ {(16<<2)|3,{105,55,115}},
/* 13225 */ {(16<<2)|3,{105,55,116}},
/* 13226 */ {(11<<2)|2,{105,55,0}},
/* 13227 */ {(11<<2)|2,{105,55,0}},
/* 13228 */ {(11<<2)|2,{105,55,0}},
/* 13229 */ {(11<<2)|2,{105,55,0}},
/* 13230 */ {(11<<2)|2,{105,55,0}},
/* 13231 */ {(11<<2)|2,{105,55,0}},
/* 13232 */ {(11<<2)|2,{105,55,0}},
/* 13233 */ {(11<<2)|2,{105,55,0}},
/* 13234 */ {(11<<2)|2,{105,55,0}},
/* 13235 */ {(11<<2)|2,{105,55,0}},
/* 13236 */ {(11<<2)|2,{105,55,0}},
/* 13237 */ {(11<<2)|2,{105,55,0}},
/* 13238 */ {(11<<2)|2,{105,55,0}},
/* 13239 */ {(11<<2)|2,{105,55,0}},
/* 13240 */ {(11<<2)|2,{105,55,0}},
/* 13241 */ {(11<<2)|2,{105,55,0}},
/* 13242 */ {(11<<2)|2,{105,55,0}},
/* 13243 */ {(11<<2)|2,{105,55,0}},
/* 13244 */ {(11<<2)|2,{105,55,0}},
/* 13245 */ {(11<<2)|2,{105,55,0}},
/* 13246 */ {(11<<2)|2,{105,55,0}},
/* 13247 */ {(11<<2)|2,{105,55,0}},
/* 13248 */ {(16<<2)|3,{105,56,48}},
/* 13249 */ {(16<<2)|3,{105,56,49}},
/* 13250 */ {(16<<2)|3,{105,56,50}},
/* 13251 */ {(16<<2)|3,{105,56,97}},
/* 13252 */ {(16<<2)|3,{105,56,99}},
/* 13253 */ {(16<<2)|3,{105,56,101}},
/* 13254 */ {(16<<2)|3,{105,56,105}},
/* 13255 */ {(16<<2)|3,{105,56,111}},
/* 13256 */ {(16<<2)|3,{105,56,115}},
/* 13257 */ {(16<<2)|3,{105,56,116}},
/* 13258 */ {(11<<2)|2,{105,56,0}},
/* 13259 */ {(11<<2)|2,{105,56,0}},
/* 13260 */ {(11<<2)|2,{105,56,0}},
/* 13261 */ {(11<<2)|2,{105,56,0}},
/* 13262 */ {(11<<2)|2,{105,56,0}},
/* 13263 */ {(11<<2)|2,{105,56,0}},
/* 13264 */ {(11<<2)|2,{105,56,0}},
/* 13265 */ {(11<<2)|2,{105,56,0}},
/* 13266 */ {(11<<2)|2,{105,56,0}},
/* 13267 */ {(11<<2)|2,{105,56,0}},
/* 13268 */ {(11<<2)|2,{105,56,0}},
/* 13269 */ {(11<<2)|2,{105,56,0}},
/* 13270 */ {(11<<2)|2,{105,56,0}},
/* 13271 */ {(11<<2)|2,{105,56,0}},
/* 13272 */ {(11<<2)|2,{105,56,0}},
/* 13273 */ {(11<<2)|2,{105,56,0}},
/* 13274 */ {(11<<2)|2,{105,56,0}},
/* 13275 */ {(11<<2)|2,{105,56,0}},
/* 13276 */ {(11<<2)|2,{105,56,0}},
/* 13277 */ {(11<<2)|2,{105,56,0}},
/* 13278 */ {(11<<2)|2,{105,56,0}},
/* 13279 */ {(11<<2)|2,{105,56,0}},
/* 13280 */ {(16<<2)|3,{105,57,48}},
/* 13281 */ {(16<<2)|3,{105,57,49}},
/* 13282 */ {(16<<2)|3,{105,57,50}},
/* 13283 */ {(16<<2)|3,{105,57,97}},
/* 13284 */ {(16<<2)|3,{105,57,99}},
/* 13285 */ {(16<<2)|3,{105,57,101}},
/* 13286 */ {(16<<2)|3,{105,57,105}},
/* 13287 */ {(16<<2)|3,{105,57,111}},
/* 13288 */ {(16<<2)|3,{105,57,115}},
/* 13289 */ {(16<<2)|3,{105,57,116}},
/* 13290 */ {(11<<2)|2,{105,57,0}},
/* 13291 */ {(11<<2)|2,{105,57,0}},
/* 13292 */ {(11<<2)|2,{105,57,0}},
/* 13293 */ {(11<<2)|2,{105,57,0}},
/* 13294 */ {(11<<2)|2,{105,57,0}},
/* 13295 */ {(11<<2)|2,{105,57,0}},
/* 13296 */ {(11<<2)|2,{105,57,0}},
/* 13297 */ {(11<<2)|2,{105,57,0}},
/* 13298 */ {(11<<2)|2,{105,57,0}},
/* 13299 */ {(11<<2)|2,{105,57,0}},
/* 13300 */ {(11<<2)|2,{105,57,0}},
/* 13301 */ {(11<<2)|2,{105,57,0}},
/* 13302 */ {(11<<2)|2,{105,57,0}},
/* 13303 */ {(11<<2)|2,{105,57,0}},
/* 13304 */ {(11<<2)|2,{105,57,0}},
/* 13305 */ {(11<<2)|2,{105,57,0}},
/* 13306 */ {(11<<2)|2,{105,57,0}},
/* 13307 */ {(11<<2)|2,{105,57,0}},
/* 13308 */ {(11<<2)|2,{105,57,0}},
/* 13309 */ {(11<<2)|2,{105,57,0}},
/* 13310 */ {(11<<2)|2,{105,57,0}},
/* 13311 */ {(11<<2)|2,{105,57,0}},
/* 13312 */ {(16<<2)|3,{105,61,48}},
/* 13313 */ {(16<<2)|3,{105,61,49}},
/* 13314 */ {(16<<2)|3,{105,61,50}},
/* 13315 */ {(16<<2)|3,{105,61,97}},
/* 13316 */ {(16<<2)|3,{105,61,99}},
/* 13317 */ {(16<<2)|3,{105,61,101}},
/* 13318 */ {(16<<2)|3,{105,61,105}},
/* 13319 */ {(16<<2)|3,{105,61,111}},
/* 13320 */ {(16<<2)|3,{105,61,115}},
/* 13321 */ {(16<<2)|3,{105,61,116}},
/* 13322 */ {(11<<2)|2,{105,61,0}},
/* 13323 */ {(11<<2)|2,{105,61,0}},
/* 13324 */ {(11<<2)|2,{105,61,0}},
/* 13325 */ {(11<<2)|2,{105,61,0}},
/* 13326 */ {(11<<2)|2,{105,61,0}},
/* 13327 */ {(11<<2)|2,{105,61,0}},
/* 13328 */ {(11<<2)|2,{105,61,0}},
/* 13329 */ {(11<<2)|2,{105,61,0}},
/* 13330 */ {(11<<2)|2,{105,61,0}},
/* 13331 */ {(11<<2)|2,{105,61,0}},
/* 13332 */ {(11<<2)|2,{105,61,0}},
/* 13333 */ {(11<<2)|2,{105,61,0}},
/* 13334 */ {(11<<2)|2,{105,61,0}},
/* 13335 */ {(11<<2)|2,{105,61,0}},
/* 13336 */ {(11<<2)|2,{105,61,0}},
/* 13337 */ {(11<<2)|2,{105,61,0}},
/* 13338 */ {(11<<2)|2,{105,61,0}},
/* 13339 */ {(11<<2)|2,{105,61,0}},
/* 13340 */ {(11<<2)|2,{105,61,0}},
/* 13341 */ {(11<<2)|2,{105,61,0}},
/* 13342 */ {(11<<2)|2,{105,61,0}},
/* 13343 */ {(11<<2)|2,{105,61,0}},
/* 13344 */ {(16<<2)|3,{105,65,48}},
/* 13345 */ {(16<<2)|3,{105,65,49}},
/* 13346 */ {(16<<2)|3,{105,65,50}},
/* 13347 */ {(16<<2)|3,{105,65,97}},
/* 13348 */ {(16<<2)|3,{105,65,99}},
/* 13349 */ {(16<<2)|3,{105,65,101}},
/* 13350 */ {(16<<2)|3,{105,65,105}},
/* 13351 */ {(16<<2)|3,{105,65,111}},
/* 13352 */ {(16<<2)|3,{105,65,115}},
/* 13353 */ {(16<<2)|3,{105,65,116}},
/* 13354 */ {(11<<2)|2,{105,65,0}},
/* 13355 */ {(11<<2)|2,{105,65,0}},
/* 13356 */ {(11<<2)|2,{105,65,0}},
/* 13357 */ {(11<<2)|2,{105,65,0}},
/* 13358 */ {(11<<2)|2,{105,65,0}},
/* 13359 */ {(11<<2)|2,{105,65,0}},
/* 13360 */ {(11<<2)|2,{105,65,0}},
/* 13361 */ {(11<<2)|2,{105,65,0}},
/* 13362 */ {(11<<2)|2,{105,65,0}},
/* 13363 */ {(11<<2)|2,{105,65,0}},
/* 13364 */ {(11<<2)|2,{105,65,0}},
/* 13365 */ {(11<<2)|2,{105,65,0}},
/* 13366 */ {(11<<2)|2,{105,65,0}},
/* 13367 */ {(11<<2)|2,{105,65,0}},
/* 13368 */ {(11<<2)|2,{105,65,0}},
/* 13369 */ {(11<<2)|2,{105,65,0}},
/* 13370 */ {(11<<2)|2,{105,65,0}},
/* 13371 */ {(11<<2)|2,{105,65,0}},
/* 13372 */ {(11<<2)|2,{105,65,0}},
/* 13373 */ {(11<<2)|2,{105,65,0}},
/* 13374 */ {(11<<2)|2,{105,65,0}},
/* 13375 */ {(11<<2)|2,{105,65,0}},
/* 13376 */ {(16<<2)|3,{105,95,48}},
/* 13377 */ {(16<<2)|3,{105,95,49}},
/* 13378 */ {(16<<2)|3,{105,95,50}},
/* 13379 */ {(16<<2)|3,{105,95,97}},
/* 13380 */ {(16<<2)|3,{105,95,99}},
/* 13381 */ {(16<<2)|3,{105,95,101}},
/* 13382 */ {(16<<2)|3,{105,95,105}},
/* 13383 */ {(16<<2)|3,{105,95,111}},
/* 13384 */ {(16<<2)|3,{105,95,115}},
/* 13385 */ {(16<<2)|3,{105,95,116}},
/* 13386 */ {(11<<2)|2,{105,95,0}},
/* 13387 */ {(11<<2)|2,{105,95,0}},
/* 13388 */ {(11<<2)|2,{105,95,0}},
/* 13389 */ {(11<<2)|2,{105,95,0}},
/* 13390 */ {(11<<2)|2,{105,95,0}},
/* 13391 */ {(11<<2)|2,{105,95,0}},
/* 13392 */ {(11<<2)|2,{105,95,0}},
/* 13393 */ {(11<<2)|2,{105,95,0}},
/* 13394 */ {(11<<2)|2,{105,95,0}},
/* 13395 */ {(11<<2)|2,{105,95,0}},
/* 13396 */ {(11<<2)|2,{105,95,0}},
/* 13397 */ {(11<<2)|2,{105,95,0}},
/* 13398 */ {(11<<2)|2,{105,95,0}},
/* 13399 */ {(11<<2)|2,{105,95,0}},
/* 13400 */ {(11<<2)|2,{105,95,0}},
/* 13401 */ {(11<<2)|2,{105,95,0}},
/* 13402 */ {(11<<2)|2,{105,95,0}},
/* 13403 */ {(11<<2)|2,{105,95,0}},
/* 13404 */ {(11<<2)|2,{105,95,0}},
/* 13405 */ {(11<<2)|2,{105,95,0}},
/* 13406 */ {(11<<2)|2,{105,95,0}},
/* 13407 */ {(11<<2)|2,{105,95,0}},
/* 13408 */ {(16<<2)|3,{105,98,48}},
/* 13409 */ {(16<<2)|3,{105,98,49}},
/* 13410 */ {(16<<2)|3,{105,98,50}},
/* 13411 */ {(16<<2)|3,{105,98,97}},
/* 13412 */ {(16<<2)|3,{105,98,99}},
/* 13413 */ {(16<<2)|3,{105,98,101}},
/* 13414 */ {(16<<2)|3,{105,98,105}},
/* 13415 */ {(16<<2)|3,{105,98,111}},
/* 13416 */ {(16<<2)|3,{105,98,115}},
/* 13417 */ {(16<<2)|3,{105,98,116}},
/* 13418 */ {(11<<2)|2,{105,98,0}},
/* 13419 */ {(11<<2)|2,{105,98,0}},
/* 13420 */ {(11<<2)|2,{105,98,0}},
/* 13421 */ {(11<<2)|2,{105,98,0}},
/* 13422 */ {(11<<2)|2,{105,98,0}},
/* 13423 */ {(11<<2)|2,{105,98,0}},
/* 13424 */ {(11<<2)|2,{105,98,0}},
/* 13425 */ {(11<<2)|2,{105,98,0}},
/* 13426 */ {(11<<2)|2,{105,98,0}},
/* 13427 */ {(11<<2)|2,{105,98,0}},
/* 13428 */ {(11<<2)|2,{105,98,0}},
/* 13429 */ {(11<<2)|2,{105,98,0}},
/* 13430 */ {(11<<2)|2,{105,98,0}},
/* 13431 */ {(11<<2)|2,{105,98,0}},
/* 13432 */ {(11<<2)|2,{105,98,0}},
/* 13433 */ {(11<<2)|2,{105,98,0}},
/* 13434 */ {(11<<2)|2,{105,98,0}},
/* 13435 */ {(11<<2)|2,{105,98,0}},
/* 13436 */ {(11<<2)|2,{105,98,0}},
/* 13437 */ {(11<<2)|2,{105,98,0}},
/* 13438 */ {(11<<2)|2,{105,98,0}},
/* 13439 */ {(11<<2)|2,{105,98,0}},
/* 13440 */ {(16<<2)|3,{105,100,48}},
/* 13441 */ {(16<<2)|3,{105,100,49}},
/* 13442 */ {(16<<2)|3,{105,100,50}},
/* 13443 */ {(16<<2)|3,{105,100,97}},
/* 13444 */ {(16<<2)|3,{105,100,99}},
/* 13445 */ {(16<<2)|3,{105,100,101}},
/* 13446 */ {(16<<2)|3,{105,100,105}},
/* 13447 */ {(16<<2)|3,{105,100,111}},
/* 13448 */ {(16<<2)|3,{105,100,115}},
/* 13449 */ {(16<<2)|3,{105,100,116}},
/* 13450 */ {(11<<2)|2,{105,100,0}},
/* 13451 */ {(11<<2)|2,{105,100,0}},
/* 13452 */ {(11<<2)|2,{105,100,0}},
/* 13453 */ {(11<<2)|2,{105,100,0}},
/* 13454 */ {(11<<2)|2,{105,100,0}},
/* 13455 */ {(11<<2)|2,{105,100,0}},
/* 13456 */ {(11<<2)|2,{105,100,0}},
/* 13457 */ {(11<<2)|2,{105,100,0}},
/* 13458 */ {(11<<2)|2,{105,100,0}},
/* 13459 */ {(11<<2)|2,{105,100,0}},
/* 13460 */ {(11<<2)|2,{105,100,0}},
/* 13461 */ {(11<<2)|2,{105,100,0}},
/* 13462 */ {(11<<2)|2,{105,100,0}},
/* 13463 */ {(11<<2)|2,{105,100,0}},
/* 13464 */ {(11<<2)|2,{105,100,0}},
/* 13465 */ {(11<<2)|2,{105,100,0}},
/* 13466 */ {(11<<2)|2,{105,100,0}},
/* 13467 */ {(11<<2)|2,{105,100,0}},
/* 13468 */ {(11<<2)|2,{105,100,0}},
/* 13469 */ {(11<<2)|2,{105,100,0}},
/* 13470 */ {(11<<2)|2,{105,100,0}},
/* 13471 */ {(11<<2)|2,{105,100,0}},
/* 13472 */ {(16<<2)|3,{105,102,48}},
/* 13473 */ {(16<<2)|3,{105,102,49}},
/* 13474 */ {(16<<2)|3,{105,102,50}},
/* 13475 */ {(16<<2)|3,{105,102,97}},
/* 13476 */ {(16<<2)|3,{105,102,99}},
/* 13477 */ {(16<<2)|3,{105,102,101}},
/* 13478 */ {(16<<2)|3,{105,102,105}},
/* 13479 */ {(16<<2)|3,{105,102,111}},
/* 13480 */ {(16<<2)|3,{105,102,115}},
/* 13481 */ {(16<<2)|3,{105,102,116}},
/* 13482 */ {(11<<2)|2,{105,102,0}},
/* 13483 */ {(11<<2)|2,{105,102,0}},
/* 13484 */ {(11<<2)|2,{105,102,0}},
/* 13485 */ {(11<<2)|2,{105,102,0}},
/* 13486 */ {(11<<2)|2,{105,102,0}},
/* 13487 */ {(11<<2)|2,{105,102,0}},
/* 13488 */ {(11<<2)|2,{105,102,0}},
/* 13489 */ {(11<<2)|2,{105,102,0}},
/* 13490 */ {(11<<2)|2,{105,102,0}},
/* 13491 */ {(11<<2)|2,{105,102,0}},
/* 13492 */ {(11<<2)|2,{105,102,0}},
/* 13493 */ {(11<<2)|2,{105,102,0}},
/* 13494 */ {(11<<2)|2,{105,102,0}},
/* 13495 */ {(11<<2)|2,{105,102,0}},
/* 13496 */ {(11<<2)|2,{105,102,0}},
/* 13497 */ {(11<<2)|2,{105,102,0}},
/* 13498 */ {(11<<2)|2,{105,102,0}},
/* 13499 */ {(11<<2)|2,{105,102,0}},
/* 13500 */ {(11<<2)|2,{105,102,0}},
/* 13501 */ {(11<<2)|2,{105,102,0}},
/* 13502 */ {(11<<2)|2,{105,102,0}},
/* 13503 */ {(11<<2)|2,{105,102,0}},
/* 13504 */ {(16<<2)|3,{105,103,48}},
/* 13505 */ {(16<<2)|3,{105,103,49}},
/* 13506 */ {(16<<2)|3,{105,103,50}},
/* 13507 */ {(16<<2)|3,{105,103,97}},
/* 13508 */ {(16<<2)|3,{105,103,99}},
/* 13509 */ {(16<<2)|3,{105,103,101}},
/* 13510 */ {(16<<2)|3,{105,103,105}},
/* 13511 */ {(16<<2)|3,{105,103,111}},
/* 13512 */ {(16<<2)|3,{105,103,115}},
/* 13513 */ {(16<<2)|3,{105,103,116}},
/* 13514 */ {(11<<2)|2,{105,103,0}},
/* 13515 */ {(11<<2)|2,{105,103,0}},
/* 13516 */ {(11<<2)|2,{105,103,0}},
/* 13517 */ {(11<<2)|2,{105,103,0}},
/* 13518 */ {(11<<2)|2,{105,103,0}},
/* 13519 */ {(11<<2)|2,{105,103,0}},
/* 13520 */ {(11<<2)|2,{105,103,0}},
/* 13521 */ {(11<<2)|2,{105,103,0}},
/* 13522 */ {(11<<2)|2,{105,103,0}},
/* 13523 */ {(11<<2)|2,{105,103,0}},
/* 13524 */ {(11<<2)|2,{105,103,0}},
/* 13525 */ {(11<<2)|2,{105,103,0}},
/* 13526 */ {(11<<2)|2,{105,103,0}},
/* 13527 */ {(11<<2)|2,{105,103,0}},
/* 13528 */ {(11<<2)|2,{105,103,0}},
/* 13529 */ {(11<<2)|2,{105,103,0}},
/* 13530 */ {(11<<2)|2,{105,103,0}},
/* 13531 */ {(11<<2)|2,{105,103,0}},
/* 13532 */ {(11<<2)|2,{105,103,0}},
/* 13533 */ {(11<<2)|2,{105,103,0}},
/* 13534 */ {(11<<2)|2,{105,103,0}},
/* 13535 */ {(11<<2)|2,{105,103,0}},
/* 13536 */ {(16<<2)|3,{105,104,48}},
/* 13537 */ {(16<<2)|3,{105,104,49}},
/* 13538 */ {(16<<2)|3,{105,104,50}},
/* 13539 */ {(16<<2)|3,{105,104,97}},
/* 13540 */ {(16<<2)|3,{105,104,99}},
/* 13541 */ {(16<<2)|3,{105,104,101}},
/* 13542 */ {(16<<2)|3,{105,104,105}},
/* 13543 */ {(16<<2)|3,{105,104,111}},
/* 13544 */ {(16<<2)|3,{105,104,115}},
/* 13545 */ {(16<<2)|3,{105,104,116}},
/* 13546 */ {(11<<2)|2,{105,104,0}},
/* 13547 */ {(11<<2)|2,{105,104,0}},
/* 13548 */ {(11<<2)|2,{105,104,0}},
/* 13549 */ {(11<<2)|2,{105,104,0}},
/* 13550 */ {(11<<2)|2,{105,104,0}},
/* 13551 */ {(11<<2)|2,{105,104,0}},
/* 13552 */ {(11<<2)|2,{105,104,0}},
/* 13553 */ {(11<<2)|2,{105,104,0}},
/* 13554 */ {(11<<2)|2,{105,104,0}},
/* 13555 */ {(11<<2)|2,{105,104,0}},
/* 13556 */ {(11<<2)|2,{105,104,0}},
/* 13557 */ {(11<<2)|2,{105,104,0}},
/* 13558 */ {(11<<2)|2,{105,104,0}},
/* 13559 */ {(11<<2)|2,{105,104,0}},
/* 13560 */ {(11<<2)|2,{105,104,0}},
/* 13561 */ {(11<<2)|2,{105,104,0}},
/* 13562 */ {(11<<2)|2,{105,104,0}},
/* 13563 */ {(11<<2)|2,{105,104,0}},
/* 13564 */ {(11<<2)|2,{105,104,0}},
/* 13565 */ {(11<<2)|2,{105,104,0}},
/* 13566 */ {(11<<2)|2,{105,104,0}},
/* 13567 */ {(11<<2)|2,{105,104,0}},
/* 13568 */ {(16<<2)|3,{105,108,48}},
/* 13569 */ {(16<<2)|3,{105,108,49}},
/* 13570 */ {(16<<2)|3,{105,108,50}},
/* 13571 */ {(16<<2)|3,{105,108,97}},
/* 13572 */ {(16<<2)|3,{105,108,99}},
/* 13573 */ {(16<<2)|3,{105,108,101}},
/* 13574 */ {(16<<2)|3,{105,108,105}},
/* 13575 */ {(16<<2)|3,{105,108,111}},
/* 13576 */ {(16<<2)|3,{105,108,115}},
/* 13577 */ {(16<<2)|3,{105,108,116}},
/* 13578 */ {(11<<2)|2,{105,108,0}},
/* 13579 */ {(11<<2)|2,{105,108,0}},
/* 13580 */ {(11<<2)|2,{105,108,0}},
/* 13581 */ {(11<<2)|2,{105,108,0}},
/* 13582 */ {(11<<2)|2,{105,108,0}},
/* 13583 */ {(11<<2)|2,{105,108,0}},
/* 13584 */ {(11<<2)|2,{105,108,0}},
/* 13585 */ {(11<<2)|2,{105,108,0}},
/* 13586 */ {(11<<2)|2,{105,108,0}},
/* 13587 */ {(11<<2)|2,{105,108,0}},
/* 13588 */ {(11<<2)|2,{105,108,0}},
/* 13589 */ {(11<<2)|2,{105,108,0}},
/* 13590 */ {(11<<2)|2,{105,108,0}},
/* 13591 */ {(11<<2)|2,{105,108,0}},
/* 13592 */ {(11<<2)|2,{105,108,0}},
/* 13593 */ {(11<<2)|2,{105,108,0}},
/* 13594 */ {(11<<2)|2,{105,108,0}},
/* 13595 */ {(11<<2)|2,{105,108,0}},
/* 13596 */ {(11<<2)|2,{105,108,0}},
/* 13597 */ {(11<<2)|2,{105,108,0}},
/* 13598 */ {(11<<2)|2,{105,108,0}},
/* 13599 */ {(11<<2)|2,{105,108,0}},
/* 13600 */ {(16<<2)|3,{105,109,48}},
/* 13601 */ {(16<<2)|3,{105,109,49}},
/* 13602 */ {(16<<2)|3,{105,109,50}},
/* 13603 */ {(16<<2)|3,{105,109,97}},
/* 13604 */ {(16<<2)|3,{105,109,99}},
/* 13605 */ {(16<<2)|3,{105,109,101}},
/* 13606 */ {(16<<2)|3,{105,109,105}},
/* 13607 */ {(16<<2)|3,{105,109,111}},
/* 13608 */ {(16<<2)|3,{105,109,115}},
/* 13609 */ {(16<<2)|3,{105,109,116}},
/* 13610 */ {(11<<2)|2,{105,109,0}},
/* 13611 */ {(11<<2)|2,{105,109,0}},
/* 13612 */ {(11<<2)|2,{105,109,0}},
/* 13613 */ {(11<<2)|2,{105,109,0}},
/* 13614 */ {(11<<2)|2,{105,109,0}},
/* 13615 */ {(11<<2)|2,{105,109,0}},
/* 13616 */ {(11<<2)|2,{105,109,0}},
/* 13617 */ {(11<<2)|2,{105,109,0}},
/* 13618 */ {(11<<2)|2,{105,109,0}},
/* 13619 */ {(11<<2)|2,{105,109,0}},
/* 13620 */ {(11<<2)|2,{105,109,0}},
/* 13621 */ {(11<<2)|2,{105,109,0}},
/* 13622 */ {(11<<2)|2,{105,109,0}},
/* 13623 */ {(11<<2)|2,{105,109,0}},
/* 13624 */ {(11<<2)|2,{105,109,0}},
/* 13625 */ {(11<<2)|2,{105,109,0}},
/* 13626 */ {(11<<2)|2,{105,109,0}},
/* 13627 */ {(11<<2)|2,{105,109,0}},
/* 13628 */ {(11<<2)|2,{105,109,0}},
/* 13629 */ {(11<<2)|2,{105,109,0}},
/* 13630 */ {(11<<2)|2,{105,109,0}},
/* 13631 */ {(11<<2)|2,{105,109,0}},
/* 13632 */ {(16<<2)|3,{105,110,48}},
/* 13633 */ {(16<<2)|3,{105,110,49}},
/* 13634 */ {(16<<2)|3,{105,110,50}},
/* 13635 */ {(16<<2)|3,{105,110,97}},
/* 13636 */ {(16<<2)|3,{105,110,99}},
/* 13637 */ {(16<<2)|3,{105,110,101}},
/* 13638 */ {(16<<2)|3,{105,110,105}},
/* 13639 */ {(16<<2)|3,{105,110,111}},
/* 13640 */ {(16<<2)|3,{105,110,115}},
/* 13641 */ {(16<<2)|3,{105,110,116}},
/* 13642 */ {(11<<2)|2,{105,110,0}},
/* 13643 */ {(11<<2)|2,{105,110,0}},
/* 13644 */ {(11<<2)|2,{105,110,0}},
/* 13645 */ {(11<<2)|2,{105,110,0}},
/* 13646 */ {(11<<2)|2,{105,110,0}},
/* 13647 */ {(11<<2)|2,{105,110,0}},
/* 13648 */ {(11<<2)|2,{105,110,0}},
/* 13649 */ {(11<<2)|2,{105,110,0}},
/* 13650 */ {(11<<2)|2,{105,110,0}},
/* 13651 */ {(11<<2)|2,{105,110,0}},
/* 13652 */ {(11<<2)|2,{105,110,0}},
/* 13653 */ {(11<<2)|2,{105,110,0}},
/* 13654 */ {(11<<2)|2,{105,110,0}},
/* 13655 */ {(11<<2)|2,{105,110,0}},
/* 13656 */ {(11<<2)|2,{105,110,0}},
/* 13657 */ {(11<<2)|2,{105,110,0}},
/* 13658 */ {(11<<2)|2,{105,110,0}},
/* 13659 */ {(11<<2)|2,{105,110,0}},
/* 13660 */ {(11<<2)|2,{105,110,0}},
/* 13661 */ {(11<<2)|2,{105,110,0}},
/* 13662 */ {(11<<2)|2,{105,110,0}},
/* 13663 */ {(11<<2)|2,{105,110,0}},
/* 13664 */ {(16<<2)|3,{105,112,48}},
/* 13665 */ {(16<<2)|3,{105,112,49}},
/* 13666 */ {(16<<2)|3,{105,112,50}},
/* 13667 */ {(16<<2)|3,{105,112,97}},
/* 13668 */ {(16<<2)|3,{105,112,99}},
/* 13669 */ {(16<<2)|3,{105,112,101}},
/* 13670 */ {(16<<2)|3,{105,112,105}},
/* 13671 */ {(16<<2)|3,{105,112,111}},
/* 13672 */ {(16<<2)|3,{105,112,115}},
/* 13673 */ {(16<<2)|3,{105,112,116}},
/* 13674 */ {(11<<2)|2,{105,112,0}},
/* 13675 */ {(11<<2)|2,{105,112,0}},
/* 13676 */ {(11<<2)|2,{105,112,0}},
/* 13677 */ {(11<<2)|2,{105,112,0}},
/* 13678 */ {(11<<2)|2,{105,112,0}},
/* 13679 */ {(11<<2)|2,{105,112,0}},
/* 13680 */ {(11<<2)|2,{105,112,0}},
/* 13681 */ {(11<<2)|2,{105,112,0}},
/* 13682 */ {(11<<2)|2,{105,112,0}},
/* 13683 */ {(11<<2)|2,{105,112,0}},
/* 13684 */ {(11<<2)|2,{105,112,0}},
/* 13685 */ {(11<<2)|2,{105,112,0}},
/* 13686 */ {(11<<2)|2,{105,112,0}},
/* 13687 */ {(11<<2)|2,{105,112,0}},
/* 13688 */ {(11<<2)|2,{105,112,0}},
/* 13689 */ {(11<<2)|2,{105,112,0}},
/* 13690 */ {(11<<2)|2,{105,112,0}},
/* 13691 */ {(11<<2)|2,{105,112,0}},
/* 13692 */ {(11<<2)|2,{105,112,0}},
/* 13693 */ {(11<<2)|2,{105,112,0}},
/* 13694 */ {(11<<2)|2,{105,112,0}},
/* 13695 */ {(11<<2)|2,{105,112,0}},
/* 13696 */ {(16<<2)|3,{105,114,48}},
/* 13697 */ {(16<<2)|3,{105,114,49}},
/* 13698 */ {(16<<2)|3,{105,114,50}},
/* 13699 */ {(16<<2)|3,{105,114,97}},
/* 13700 */ {(16<<2)|3,{105,114,99}},
/* 13701 */ {(16<<2)|3,{105,114,101}},
/* 13702 */ {(16<<2)|3,{105,114,105}},
/* 13703 */ {(16<<2)|3,{105,114,111}},
/* 13704 */ {(16<<2)|3,{105,114,115}},
/* 13705 */ {(16<<2)|3,{105,114,116}},
/* 13706 */ {(11<<2)|2,{105,114,0}},
/* 13707 */ {(11<<2)|2,{105,114,0}},
/* 13708 */ {(11<<2)|2,{105,114,0}},
/* 13709 */ {(11<<2)|2,{105,114,0}},
/* 13710 */ {(11<<2)|2,{105,114,0}},
/* 13711 */ {(11<<2)|2,{105,114,0}},
/* 13712 */ {(11<<2)|2,{105,114,0}},
/* 13713 */ {(11<<2)|2,{105,114,0}},
/* 13714 */ {(11<<2)|2,{105,114,0}},
/* 13715 */ {(11<<2)|2,{105,114,0}},
/* 13716 */ {(11<<2)|2,{105,114,0}},
/* 13717 */ {(11<<2)|2,{105,114,0}},
/* 13718 */ {(11<<2)|2,{105,114,0}},
/* 13719 */ {(11<<2)|2,{105,114,0}},
/* 13720 */ {(11<<2)|2,{105,114,0}},
/* 13721 */ {(11<<2)|2,{105,114,0}},
/* 13722 */ {(11<<2)|2,{105,114,0}},
/* 13723 */ {(11<<2)|2,{105,114,0}},
/* 13724 */ {(11<<2)|2,{105,114,0}},
/* 13725 */ {(11<<2)|2,{105,114,0}},
/* 13726 */ {(11<<2)|2,{105,114,0}},
/* 13727 */ {(11<<2)|2,{105,114,0}},
/* 13728 */ {(16<<2)|3,{105,117,48}},
/* 13729 */ {(16<<2)|3,{105,117,49}},
/* 13730 */ {(16<<2)|3,{105,117,50}},
/* 13731 */ {(16<<2)|3,{105,117,97}},
/* 13732 */ {(16<<2)|3,{105,117,99}},
/* 13733 */ {(16<<2)|3,{105,117,101}},
/* 13734 */ {(16<<2)|3,{105,117,105}},
/* 13735 */ {(16<<2)|3,{105,117,111}},
/* 13736 */ {(16<<2)|3,{105,117,115}},
/* 13737 */ {(16<<2)|3,{105,117,116}},
/* 13738 */ {(11<<2)|2,{105,117,0}},
/* 13739 */ {(11<<2)|2,{105,117,0}},
/* 13740 */ {(11<<2)|2,{105,117,0}},
/* 13741 */ {(11<<2)|2,{105,117,0}},
/* 13742 */ {(11<<2)|2,{105,117,0}},
/* 13743 */ {(11<<2)|2,{105,117,0}},
/* 13744 */ {(11<<2)|2,{105,117,0}},
/* 13745 */ {(11<<2)|2,{105,117,0}},
/* 13746 */ {(11<<2)|2,{105,117,0}},
/* 13747 */ {(11<<2)|2,{105,117,0}},
/* 13748 */ {(11<<2)|2,{105,117,0}},
/* 13749 */ {(11<<2)|2,{105,117,0}},
/* 13750 */ {(11<<2)|2,{105,117,0}},
/* 13751 */ {(11<<2)|2,{105,117,0}},
/* 13752 */ {(11<<2)|2,{105,117,0}},
/* 13753 */ {(11<<2)|2,{105,117,0}},
/* 13754 */ {(11<<2)|2,{105,117,0}},
/* 13755 */ {(11<<2)|2,{105,117,0}},
/* 13756 */ {(11<<2)|2,{105,117,0}},
/* 13757 */ {(11<<2)|2,{105,117,0}},
/* 13758 */ {(11<<2)|2,{105,117,0}},
/* 13759 */ {(11<<2)|2,{105,117,0}},
/* 13760 */ {(12<<2)|2,{105,58,0}},
/* 13761 */ {(12<<2)|2,{105,58,0}},
/* 13762 */ {(12<<2)|2,{105,58,0}},
/* 13763 */ {(12<<2)|2,{105,58,0}},
/* 13764 */ {(12<<2)|2,{105,58,0}},
/* 13765 */ {(12<<2)|2,{105,58,0}},
/* 13766 */ {(12<<2)|2,{105,58,0}},
/* 13767 */ {(12<<2)|2,{105,58,0}},
/* 13768 */ {(12<<2)|2,{105,58,0}},
/* 13769 */ {(12<<2)|2,{105,58,0}},
/* 13770 */ {(12<<2)|2,{105,58,0}},
/* 13771 */ {(12<<2)|2,{105,58,0}},
/* 13772 */ {(12<<2)|2,{105,58,0}},
/* 13773 */ {(12<<2)|2,{105,58,0}},
/* 13774 */ {(12<<2)|2,{105,58,0}},
/* 13775 */ {(12<<2)|2,{105,58,0}},
/* 13776 */ {(12<<2)|2,{105,66,0}},
/* 13777 */ {(12<<2)|2,{105,66,0}},
/* 13778 */ {(12<<2)|2,{105,66,0}},
/* 13779 */ {(12<<2)|2,{105,66,0}},
/* 13780 */ {(12<<2)|2,{105,66,0}},
/* 13781 */ {(12<<2)|2,{105,66,0}},
/* 13782 */ {(12<<2)|2,{105,66,0}},
/* 13783 */ {(12<<2)|2,{105,66,0}},
/* 13784 */ {(12<<2)|2,{105,66,0}},
/* 13785 */ {(12<<2)|2,{105,66,0}},
/* 13786 */ {(12<<2)|2,{105,66,0}},
/* 13787 */ {(12<<2)|2,{105,66,0}},
/* 13788 */ {(12<<2)|2,{105,66,0}},
/* 13789 */ {(12<<2)|2,{105,66,0}},
/* 13790 */ {(12<<2)|2,{105,66,0}},
/* 13791 */ {(12<<2)|2,{105,66,0}},
/* 13792 */ {(12<<2)|2,{105,67,0}},
/* 13793 */ {(12<<2)|2,{105,67,0}},
/* 13794 */ {(12<<2)|2,{105,67,0}},
/* 13795 */ {(12<<2)|2,{105,67,0}},
/* 13796 */ {(12<<2)|2,{105,67,0}},
/* 13797 */ {(12<<2)|2,{105,67,0}},
/* 13798 */ {(12<<2)|2,{105,67,0}},
/* 13799 */ {(12<<2)|2,{105,67,0}},
/* 13800 */ {(12<<2)|2,{105,67,0}},
/* 13801 */ {(12<<2)|2,{105,67,0}},
/* 13802 */ {(12<<2)|2,{105,67,0}},
/* 13803 */ {(12<<2)|2,{105,67,0}},
/* 13804 */ {(12<<2)|2,{105,67,0}},
/* 13805 */ {(12<<2)|2,{105,67,0}},
/* 13806 */ {(12<<2)|2,{105,67,0}},
/* 13807 */ {(12<<2)|2,{105,67,0}},
/* 13808 */ {(12<<2)|2,{105,68,0}},
/* 13809 */ {(12<<2)|2,{105,68,0}},
/* 13810 */ {(12<<2)|2,{105,68,0}},
/* 13811 */ {(12<<2)|2,{105,68,0}},
/* 13812 */ {(12<<2)|2,{105,68,0}},
/* 13813 */ {(12<<2)|2,{105,68,0}},
/* 13814 */ {(12<<2)|2,{105,68,0}},
/* 13815 */ {(12<<2)|2,{105,68,0}},
/* 13816 */ {(12<<2)|2,{105,68,0}},
/* 13817 */ {(12<<2)|2,{105,68,0}},
/* 13818 */ {(12<<2)|2,{105,68,0}},
/* 13819 */ {(12<<2)|2,{105,68,0}},
/* 13820 */ {(12<<2)|2,{105,68,0}},
/* 13821 */ {(12<<2)|2,{105,68,0}},
/* 13822 */ {(12<<2)|2,{105,68,0}},
/* 13823 */ {(12<<2)|2,{105,68,0}},
/* 13824 */ {(12<<2)|2,{105,69,0}},
/* 13825 */ {(12<<2)|2,{105,69,0}},
/* 13826 */ {(12<<2)|2,{105,69,0}},
/* 13827 */ {(12<<2)|2,{105,69,0}},
/* 13828 */ {(12<<2)|2,{105,69,0}},
/* 13829 */ {(12<<2)|2,{105,69,0}},
/* 13830 */ {(12<<2)|2,{105,69,0}},
/* 13831 */ {(12<<2)|2,{105,69,0}},
/* 13832 */ {(12<<2)|2,{105,69,0}},
/* 13833 */ {(12<<2)|2,{105,69,0}},
/* 13834 */ {(12<<2)|2,{105,69,0}},
/* 13835 */ {(12<<2)|2,{105,69,0}},
/* 13836 */ {(12<<2)|2,{105,69,0}},
/* 13837 */ {(12<<2)|2,{105,69,0}},
/* 13838 */ {(12<<2)|2,{105,69,0}},
/* 13839 */ {(12<<2)|2,{105,69,0}},
/* 13840 */ {(12<<2)|2,{105,70,0}},
/* 13841 */ {(12<<2)|2,{105,70,0}},
/* 13842 */ {(12<<2)|2,{105,70,0}},
/* 13843 */ {(12<<2)|2,{105,70,0}},
/* 13844 */ {(12<<2)|2,{105,70,0}},
/* 13845 */ {(12<<2)|2,{105,70,0}},
/* 13846 */ {(12<<2)|2,{105,70,0}},
/* 13847 */ {(12<<2)|2,{105,70,0}},
/* 13848 */ {(12<<2)|2,{105,70,0}},
/* 13849 */ {(12<<2)|2,{105,70,0}},
/* 13850 */ {(12<<2)|2,{105,70,0}},
/* 13851 */ {(12<<2)|2,{105,70,0}},
/* 13852 */ {(12<<2)|2,{105,70,0}},
/* 13853 */ {(12<<2)|2,{105,70,0}},
/* 13854 */ {(12<<2)|2,{105,70,0}},
/* 13855 */ {(12<<2)|2,{105,70,0}},
/* 13856 */ {(12<<2)|2,{105,71,0}},
/* 13857 */ {(12<<2)|2,{105,71,0}},
/* 13858 */ {(12<<2)|2,{105,71,0}},
/* 13859 */ {(12<<2)|2,{105,71,0}},
/* 13860 */ {(12<<2)|2,{105,71,0}},
/* 13861 */ {(12<<2)|2,{105,71,0}},
/* 13862 */ {(12<<2)|2,{105,71,0}},
/* 13863 */ {(12<<2)|2,{105,71,0}},
/* 13864 */ {(12<<2)|2,{105,71,0}},
/* 13865 */ {(12<<2)|2,{105,71,0}},
/* 13866 */ {(12<<2)|2,{105,71,0}},
/* 13867 */ {(12<<2)|2,{105,71,0}},
/* 13868 */ {(12<<2)|2,{105,71,0}},
/* 13869 */ {(12<<2)|2,{105,71,0}},
/* 13870 */ {(12<<2)|2,{105,71,0}},
/* 13871 */ {(12<<2)|2,{105,71,0}},
/* 13872 */ {(12<<2)|2,{105,72,0}},
/* 13873 */ {(12<<2)|2,{105,72,0}},
/* 13874 */ {(12<<2)|2,{105,72,0}},
/* 13875 */ {(12<<2)|2,{105,72,0}},
/* 13876 */ {(12<<2)|2,{105,72,0}},
/* 13877 */ {(12<<2)|2,{105,72,0}},
/* 13878 */ {(12<<2)|2,{105,72,0}},
/* 13879 */ {(12<<2)|2,{105,72,0}},
/* 13880 */ {(12<<2)|2,{105,72,0}},
/* 13881 */ {(12<<2)|2,{105,72,0}},
/* 13882 */ {(12<<2)|2,{105,72,0}},
/* 13883 */ {(12<<2)|2,{105,72,0}},
/* 13884 */ {(12<<2)|2,{105,72,0}},
/* 13885 */ {(12<<2)|2,{105,72,0}},
/* 13886 */ {(12<<2)|2,{105,72,0}},
/* 13887 */ {(12<<2)|2,{105,72,0}},
/* 13888 */ {(12<<2)|2,{105,73,0}},
/* 13889 */ {(12<<2)|2,{105,73,0}},
/* 13890 */ {(12<<2)|2,{105,73,0}},
/* 13891 */ {(12<<2)|2,{105,73,0}},
/* 13892 */ {(12<<2)|2,{105,73,0}},
/* 13893 */ {(12<<2)|2,{105,73,0}},
/* 13894 */ {(12<<2)|2,{105,73,0}},
/* 13895 */ {(12<<2)|2,{105,73,0}},
/* 13896 */ {(12<<2)|2,{105,73,0}},
/* 13897 */ {(12<<2)|2,{105,73,0}},
/* 13898 */ {(12<<2)|2,{105,73,0}},
/* 13899 */ {(12<<2)|2,{105,73,0}},
/* 13900 */ {(12<<2)|2,{105,73,0}},
/* 13901 */ {(12<<2)|2,{105,73,0}},
/* 13902 */ {(12<<2)|2,{105,73,0}},
/* 13903 */ {(12<<2)|2,{105,73,0}},
/* 13904 */ {(12<<2)|2,{105,74,0}},
/* 13905 */ {(12<<2)|2,{105,74,0}},
/* 13906 */ {(12<<2)|2,{105,74,0}},
/* 13907 */ {(12<<2)|2,{105,74,0}},
/* 13908 */ {(12<<2)|2,{105,74,0}},
/* 13909 */ {(12<<2)|2,{105,74,0}},
/* 13910 */ {(12<<2)|2,{105,74,0}},
/* 13911 */ {(12<<2)|2,{105,74,0}},
/* 13912 */ {(12<<2)|2,{105,74,0}},
/* 13913 */ {(12<<2)|2,{105,74,0}},
/* 13914 */ {(12<<2)|2,{105,74,0}},
/* 13915 */ {(12<<2)|2,{105,74,0}},
/* 13916 */ {(12<<2)|2,{105,74,0}},
/* 13917 */ {(12<<2)|2,{105,74,0}},
/* 13918 */ {(12<<2)|2,{105,74,0}},
/* 13919 */ {(12<<2)|2,{105,74,0}},
/* 13920 */ {(12<<2)|2,{105,75,0}},
/* 13921 */ {(12<<2)|2,{105,75,0}},
/* 13922 */ {(12<<2)|2,{105,75,0}},
/* 13923 */ {(12<<2)|2,{105,75,0}},
/* 13924 */ {(12<<2)|2,{105,75,0}},
/* 13925 */ {(12<<2)|2,{105,75,0}},
/* 13926 */ {(12<<2)|2,{105,75,0}},
/* 13927 */ {(12<<2)|2,{105,75,0}},
/* 13928 */ {(12<<2)|2,{105,75,0}},
/* 13929 */ {(12<<2)|2,{105,75,0}},
/* 13930 */ {(12<<2)|2,{105,75,0}},
/* 13931 */ {(12<<2)|2,{105,75,0}},
/* 13932 */ {(12<<2)|2,{105,75,0}},
/* 13933 */ {(12<<2)|2,{105,75,0}},
/* 13934 */ {(12<<2)|2,{105,75,0}},
/* 13935 */ {(12<<2)|2,{105,75,0}},
/* 13936 */ {(12<<2)|2,{105,76,0}},
/* 13937 */ {(12<<2)|2,{105,76,0}},
/* 13938 */ {(12<<2)|2,{105,76,0}},
/* 13939 */ {(12<<2)|2,{105,76,0}},
/* 13940 */ {(12<<2)|2,{105,76,0}},
/* 13941 */ {(12<<2)|2,{105,76,0}},
/* 13942 */ {(12<<2)|2,{105,76,0}},
/* 13943 */ {(12<<2)|2,{105,76,0}},
/* 13944 */ {(12<<2)|2,{105,76,0}},
/* 13945 */ {(12<<2)|2,{105,76,0}},
/* 13946 */ {(12<<2)|2,{105,76,0}},
/* 13947 */ {(12<<2)|2,{105,76,0}},
/* 13948 */ {(12<<2)|2,{105,76,0}},
/* 13949 */ {(12<<2)|2,{105,76,0}},
/* 13950 */ {(12<<2)|2,{105,76,0}},
/* 13951 */ {(12<<2)|2,{105,76,0}},
/* 13952 */ {(12<<2)|2,{105,77,0}},
/* 13953 */ {(12<<2)|2,{105,77,0}},
/* 13954 */ {(12<<2)|2,{105,77,0}},
/* 13955 */ {(12<<2)|2,{105,77,0}},
/* 13956 */ {(12<<2)|2,{105,77,0}},
/* 13957 */ {(12<<2)|2,{105,77,0}},
/* 13958 */ {(12<<2)|2,{105,77,0}},
/* 13959 */ {(12<<2)|2,{105,77,0}},
/* 13960 */ {(12<<2)|2,{105,77,0}},
/* 13961 */ {(12<<2)|2,{105,77,0}},
/* 13962 */ {(12<<2)|2,{105,77,0}},
/* 13963 */ {(12<<2)|2,{105,77,0}},
/* 13964 */ {(12<<2)|2,{105,77,0}},
/* 13965 */ {(12<<2)|2,{105,77,0}},
/* 13966 */ {(12<<2)|2,{105,77,0}},
/* 13967 */ {(12<<2)|2,{105,77,0}},
/* 13968 */ {(12<<2)|2,{105,78,0}},
/* 13969 */ {(12<<2)|2,{105,78,0}},
/* 13970 */ {(12<<2)|2,{105,78,0}},
/* 13971 */ {(12<<2)|2,{105,78,0}},
/* 13972 */ {(12<<2)|2,{105,78,0}},
/* 13973 */ {(12<<2)|2,{105,78,0}},
/* 13974 */ {(12<<2)|2,{105,78,0}},
/* 13975 */ {(12<<2)|2,{105,78,0}},
/* 13976 */ {(12<<2)|2,{105,78,0}},
/* 13977 */ {(12<<2)|2,{105,78,0}},
/* 13978 */ {(12<<2)|2,{105,78,0}},
/* 13979 */ {(12<<2)|2,{105,78,0}},
/* 13980 */ {(12<<2)|2,{105,78,0}},
/* 13981 */ {(12<<2)|2,{105,78,0}},
/* 13982 */ {(12<<2)|2,{105,78,0}},
/* 13983 */ {(12<<2)|2,{105,78,0}},
/* 13984 */ {(12<<2)|2,{105,79,0}},
/* 13985 */ {(12<<2)|2,{105,79,0}},
/* 13986 */ {(12<<2)|2,{105,79,0}},
/* 13987 */ {(12<<2)|2,{105,79,0}},
/* 13988 */ {(12<<2)|2,{105,79,0}},
/* 13989 */ {(12<<2)|2,{105,79,0}},
/* 13990 */ {(12<<2)|2,{105,79,0}},
/* 13991 */ {(12<<2)|2,{105,79,0}},
/* 13992 */ {(12<<2)|2,{105,79,0}},
/* 13993 */ {(12<<2)|2,{105,79,0}},
/* 13994 */ {(12<<2)|2,{105,79,0}},
/* 13995 */ {(12<<2)|2,{105,79,0}},
/* 13996 */ {(12<<2)|2,{105,79,0}},
/* 13997 */ {(12<<2)|2,{105,79,0}},
/* 13998 */ {(12<<2)|2,{105,79,0}},
/* 13999 */ {(12<<2)|2,{105,79,0}},
/* 14000 */ {(12<<2)|2,{105,80,0}},
/* 14001 */ {(12<<2)|2,{105,80,0}},
/* 14002 */ {(12<<2)|2,{105,80,0}},
/* 14003 */ {(12<<2)|2,{105,80,0}},
/* 14004 */ {(12<<2)|2,{105,80,0}},
/* 14005 */ {(12<<2)|2,{105,80,0}},
/* 14006 */ {(12<<2)|2,{105,80,0}},
/* 14007 */ {(12<<2)|2,{105,80,0}},
/* 14008 */ {(12<<2)|2,{105,80,0}},
/* 14009 */ {(12<<2)|2,{105,80,0}},
/* 14010 */ {(12<<2)|2,{105,80,0}},
/* 14011 */ {(12<<2)|2,{105,80,0}},
/* 14012 */ {(12<<2)|2,{105,80,0}},
/* 14013 */ {(12<<2)|2,{105,80,0}},
/* 14014 */ {(12<<2)|2,{105,80,0}},
/* 14015 */ {(12<<2)|2,{105,80,0}},
/* 14016 */ {(12<<2)|2,{105,81,0}},
/* 14017 */ {(12<<2)|2,{105,81,0}},
/* 14018 */ {(12<<2)|2,{105,81,0}},
/* 14019 */ {(12<<2)|2,{105,81,0}},
/* 14020 */ {(12<<2)|2,{105,81,0}},
/* 14021 */ {(12<<2)|2,{105,81,0}},
/* 14022 */ {(12<<2)|2,{105,81,0}},
/* 14023 */ {(12<<2)|2,{105,81,0}},
/* 14024 */ {(12<<2)|2,{105,81,0}},
/* 14025 */ {(12<<2)|2,{105,81,0}},
/* 14026 */ {(12<<2)|2,{105,81,0}},
/* 14027 */ {(12<<2)|2,{105,81,0}},
/* 14028 */ {(12<<2)|2,{105,81,0}},
/* 14029 */ {(12<<2)|2,{105,81,0}},
/* 14030 */ {(12<<2)|2,{105,81,0}},
/* 14031 */ {(12<<2)|2,{105,81,0}},
/* 14032 */ {(12<<2)|2,{105,82,0}},
/* 14033 */ {(12<<2)|2,{105,82,0}},
/* 14034 */ {(12<<2)|2,{105,82,0}},
/* 14035 */ {(12<<2)|2,{105,82,0}},
/* 14036 */ {(12<<2)|2,{105,82,0}},
/* 14037 */ {(12<<2)|2,{105,82,0}},
/* 14038 */ {(12<<2)|2,{105,82,0}},
/* 14039 */ {(12<<2)|2,{105,82,0}},
/* 14040 */ {(12<<2)|2,{105,82,0}},
/* 14041 */ {(12<<2)|2,{105,82,0}},
/* 14042 */ {(12<<2)|2,{105,82,0}},
/* 14043 */ {(12<<2)|2,{105,82,0}},
/* 14044 */ {(12<<2)|2,{105,82,0}},
/* 14045 */ {(12<<2)|2,{105,82,0}},
/* 14046 */ {(12<<2)|2,{105,82,0}},
/* 14047 */ {(12<<2)|2,{105,82,0}},
/* 14048 */ {(12<<2)|2,{105,83,0}},
/* 14049 */ {(12<<2)|2,{105,83,0}},
/* 14050 */ {(12<<2)|2,{105,83,0}},
/* 14051 */ {(12<<2)|2,{105,83,0}},
/* 14052 */ {(12<<2)|2,{105,83,0}},
/* 14053 */ {(12<<2)|2,{105,83,0}},
/* 14054 */ {(12<<2)|2,{105,83,0}},
/* 14055 */ {(12<<2)|2,{105,83,0}},
/* 14056 */ {(12<<2)|2,{105,83,0}},
/* 14057 */ {(12<<2)|2,{105,83,0}},
/* 14058 */ {(12<<2)|2,{105,83,0}},
/* 14059 */ {(12<<2)|2,{105,83,0}},
/* 14060 */ {(12<<2)|2,{105,83,0}},
/* 14061 */ {(12<<2)|2,{105,83,0}},
/* 14062 */ {(12<<2)|2,{105,83,0}},
/* 14063 */ {(12<<2)|2,{105,83,0}},
/* 14064 */ {(12<<2)|2,{105,84,0}},
/* 14065 */ {(12<<2)|2,{105,84,0}},
/* 14066 */ {(12<<2)|2,{105,84,0}},
/* 14067 */ {(12<<2)|2,{105,84,0}},
/* 14068 */ {(12<<2)|2,{105,84,0}},
/* 14069 */ {(12<<2)|2,{105,84,0}},
/* 14070 */ {(12<<2)|2,{105,84,0}},
/* 14071 */ {(12<<2)|2,{105,84,0}},
/* 14072 */ {(12<<2)|2,{105,84,0}},
/* 14073 */ {(12<<2)|2,{105,84,0}},
/* 14074 */ {(12<<2)|2,{105,84,0}},
/* 14075 */ {(12<<2)|2,{105,84,0}},
/* 14076 */ {(12<<2)|2,{105,84,0}},
/* 14077 */ {(12<<2)|2,{105,84,0}},
/* 14078 */ {(12<<2)|2,{105,84,0}},
/* 14079 */ {(12<<2)|2,{105,84,0}},
/* 14080 */ {(12<<2)|2,{105,85,0}},
/* 14081 */ {(12<<2)|2,{105,85,0}},
/* 14082 */ {(12<<2)|2,{105,85,0}},
/* 14083 */ {(12<<2)|2,{105,85,0}},
/* 14084 */ {(12<<2)|2,{105,85,0}},
/* 14085 */ {(12<<2)|2,{105,85,0}},
/* 14086 */ {(12<<2)|2,{105,85,0}},
/* 14087 */ {(12<<2)|2,{105,85,0}},
/* 14088 */ {(12<<2)|2,{105,85,0}},
/* 14089 */ {(12<<2)|2,{105,85,0}},
/* 14090 */ {(12<<2)|2,{105,85,0}},
/* 14091 */ {(12<<2)|2,{105,85,0}},
/* 14092 */ {(12<<2)|2,{105,85,0}},
/* 14093 */ {(12<<2)|2,{105,85,0}},
/* 14094 */ {(12<<2)|2,{105,85,0}},
/* 14095 */ {(12<<2)|2,{105,85,0}},
/* 14096 */ {(12<<2)|2,{105,86,0}},
/* 14097 */ {(12<<2)|2,{105,86,0}},
/* 14098 */ {(12<<2)|2,{105,86,0}},
/* 14099 */ {(12<<2)|2,{105,86,0}},
/* 14100 */ {(12<<2)|2,{105,86,0}},
/* 14101 */ {(12<<2)|2,{105,86,0}},
/* 14102 */ {(12<<2)|2,{105,86,0}},
/* 14103 */ {(12<<2)|2,{105,86,0}},
/* 14104 */ {(12<<2)|2,{105,86,0}},
/* 14105 */ {(12<<2)|2,{105,86,0}},
/* 14106 */ {(12<<2)|2,{105,86,0}},
/* 14107 */ {(12<<2)|2,{105,86,0}},
/* 14108 */ {(12<<2)|2,{105,86,0}},
/* 14109 */ {(12<<2)|2,{105,86,0}},
/* 14110 */ {(12<<2)|2,{105,86,0}},
/* 14111 */ {(12<<2)|2,{105,86,0}},
/* 14112 */ {(12<<2)|2,{105,87,0}},
/* 14113 */ {(12<<2)|2,{105,87,0}},
/* 14114 */ {(12<<2)|2,{105,87,0}},
/* 14115 */ {(12<<2)|2,{105,87,0}},
/* 14116 */ {(12<<2)|2,{105,87,0}},
/* 14117 */ {(12<<2)|2,{105,87,0}},
/* 14118 */ {(12<<2)|2,{105,87,0}},
/* 14119 */ {(12<<2)|2,{105,87,0}},
/* 14120 */ {(12<<2)|2,{105,87,0}},
/* 14121 */ {(12<<2)|2,{105,87,0}},
/* 14122 */ {(12<<2)|2,{105,87,0}},
/* 14123 */ {(12<<2)|2,{105,87,0}},
/* 14124 */ {(12<<2)|2,{105,87,0}},
/* 14125 */ {(12<<2)|2,{105,87,0}},
/* 14126 */ {(12<<2)|2,{105,87,0}},
/* 14127 */ {(12<<2)|2,{105,87,0}},
/* 14128 */ {(12<<2)|2,{105,89,0}},
/* 14129 */ {(12<<2)|2,{105,89,0}},
/* 14130 */ {(12<<2)|2,{105,89,0}},
/* 14131 */ {(12<<2)|2,{105,89,0}},
/* 14132 */ {(12<<2)|2,{105,89,0}},
/* 14133 */ {(12<<2)|2,{105,89,0}},
/* 14134 */ {(12<<2)|2,{105,89,0}},
/* 14135 */ {(12<<2)|2,{105,89,0}},
/* 14136 */ {(12<<2)|2,{105,89,0}},
/* 14137 */ {(12<<2)|2,{105,89,0}},
/* 14138 */ {(12<<2)|2,{105,89,0}},
/* 14139 */ {(12<<2)|2,{105,89,0}},
/* 14140 */ {(12<<2)|2,{105,89,0}},
/* 14141 */ {(12<<2)|2,{105,89,0}},
/* 14142 */ {(12<<2)|2,{105,89,0}},
/* 14143 */ {(12<<2)|2,{105,89,0}},
/* 14144 */ {(12<<2)|2,{105,106,0}},
/* 14145 */ {(12<<2)|2,{105,106,0}},
/* 14146 */ {(12<<2)|2,{105,106,0}},
/* 14147 */ {(12<<2)|2,{105,106,0}},
/* 14148 */ {(12<<2)|2,{105,106,0}},
/* 14149 */ {(12<<2)|2,{105,106,0}},
/* 14150 */ {(12<<2)|2,{105,106,0}},
/* 14151 */ {(12<<2)|2,{105,106,0}},
/* 14152 */ {(12<<2)|2,{105,106,0}},
/* 14153 */ {(12<<2)|2,{105,106,0}},
/* 14154 */ {(12<<2)|2,{105,106,0}},
/* 14155 */ {(12<<2)|2,{105,106,0}},
/* 14156 */ {(12<<2)|2,{105,106,0}},
/* 14157 */ {(12<<2)|2,{105,106,0}},
/* 14158 */ {(12<<2)|2,{105,106,0}},
/* 14159 */ {(12<<2)|2,{105,106,0}},
/* 14160 */ {(12<<2)|2,{105,107,0}},
/* 14161 */ {(12<<2)|2,{105,107,0}},
/* 14162 */ {(12<<2)|2,{105,107,0}},
/* 14163 */ {(12<<2)|2,{105,107,0}},
/* 14164 */ {(12<<2)|2,{105,107,0}},
/* 14165 */ {(12<<2)|2,{105,107,0}},
/* 14166 */ {(12<<2)|2,{105,107,0}},
/* 14167 */ {(12<<2)|2,{105,107,0}},
/* 14168 */ {(12<<2)|2,{105,107,0}},
/* 14169 */ {(12<<2)|2,{105,107,0}},
/* 14170 */ {(12<<2)|2,{105,107,0}},
/* 14171 */ {(12<<2)|2,{105,107,0}},
/* 14172 */ {(12<<2)|2,{105,107,0}},
/* 14173 */ {(12<<2)|2,{105,107,0}},
/* 14174 */ {(12<<2)|2,{105,107,0}},
/* 14175 */ {(12<<2)|2,{105,107,0}},
/* 14176 */ {(12<<2)|2,{105,113,0}},
/* 14177 */ {(12<<2)|2,{105,113,0}},
/* 14178 */ {(12<<2)|2,{105,113,0}},
/* 14179 */ {(12<<2)|2,{105,113,0}},
/* 14180 */ {(12<<2)|2,{105,113,0}},
/* 14181 */ {(12<<2)|2,{105,113,0}},
/* 14182 */ {(12<<2)|2,{105,113,0}},
/* 14183 */ {(12<<2)|2,{105,113,0}},
/* 14184 */ {(12<<2)|2,{105,113,0}},
/* 14185 */ {(12<<2)|2,{105,113,0}},
/* 14186 */ {(12<<2)|2,{105,113,0}},
/* 14187 */ {(12<<2)|2,{105,113,0}},
/* 14188 */ {(12<<2)|2,{105,113,0}},
/* 14189 */ {(12<<2)|2,{105,113,0}},
/* 14190 */ {(12<<2)|2,{105,113,0}},
/* 14191 */ {(12<<2)|2,{105,113,0}},
/* 14192 */ {(12<<2)|2,{105,118,0}},
/* 14193 */ {(12<<2)|2,{105,118,0}},
/* 14194 */ {(12<<2)|2,{105,118,0}},
/* 14195 */ {(12<<2)|2,{105,118,0}},
/* 14196 */ {(12<<2)|2,{105,118,0}},
/* 14197 */ {(12<<2)|2,{105,118,0}},
/* 14198 */ {(12<<2)|2,{105,118,0}},
/* 14199 */ {(12<<2)|2,{105,118,0}},
/* 14200 */ {(12<<2)|2,{105,118,0}},
/* 14201 */ {(12<<2)|2,{105,118,0}},
/* 14202 */ {(12<<2)|2,{105,118,0}},
/* 14203 */ {(12<<2)|2,{105,118,0}},
/* 14204 */ {(12<<2)|2,{105,118,0}},
/* 14205 */ {(12<<2)|2,{105,118,0}},
/* 14206 */ {(12<<2)|2,{105,118,0}},
/* 14207 */ {(12<<2)|2,{105,118,0}},
/* 14208 */ {(12<<2)|2,{105,119,0}},
/* 14209 */ {(12<<2)|2,{105,119,0}},
/* 14210 */ {(12<<2)|2,{105,119,0}},
/* 14211 */ {(12<<2)|2,{105,119,0}},
/* 14212 */ {(12<<2)|2,{105,119,0}},
/* 14213 */ {(12<<2)|2,{105,119,0}},
/* 14214 */ {(12<<2)|2,{105,119,0}},
/* 14215 */ {(12<<2)|2,{105,119,0}},
/* 14216 */ {(12<<2)|2,{105,119,0}},
/* 14217 */ {(12<<2)|2,{105,119,0}},
/* 14218 */ {(12<<2)|2,{105,119,0}},
/* 14219 */ {(12<<2)|2,{105,119,0}},
/* 14220 */ {(12<<2)|2,{105,119,0}},
/* 14221 */ {(12<<2)|2,{105,119,0}},
/* 14222 */ {(12<<2)|2,{105,119,0}},
/* 14223 */ {(12<<2)|2,{105,119,0}},
/* 14224 */ {(12<<2)|2,{105,120,0}},
/* 14225 */ {(12<<2)|2,{105,120,0}},
/* 14226 */ {(12<<2)|2,{105,120,0}},
/* 14227 */ {(12<<2)|2,{105,120,0}},
/* 14228 */ {(12<<2)|2,{105,120,0}},
/* 14229 */ {(12<<2)|2,{105,120,0}},
/* 14230 */ {(12<<2)|2,{105,120,0}},
/* 14231 */ {(12<<2)|2,{105,120,0}},
/* 14232 */ {(12<<2)|2,{105,120,0}},
/* 14233 */ {(12<<2)|2,{105,120,0}},
/* 14234 */ {(12<<2)|2,{105,120,0}},
/* 14235 */ {(12<<2)|2,{105,120,0}},
/* 14236 */ {(12<<2)|2,{105,120,0}},
/* 14237 */ {(12<<2)|2,{105,120,0}},
/* 14238 */ {(12<<2)|2,{105,120,0}},
/* 14239 */ {(12<<2)|2,{105,120,0}},
/* 14240 */ {(12<<2)|2,{105,121,0}},
/* 14241 */ {(12<<2)|2,{105,121,0}},
/* 14242 */ {(12<<2)|2,{105,121,0}},
/* 14243 */ {(12<<2)|2,{105,121,0}},
/* 14244 */ {(12<<2)|2,{105,121,0}},
/* 14245 */ {(12<<2)|2,{105,121,0}},
/* 14246 */ {(12<<2)|2,{105,121,0}},
/* 14247 */ {(12<<2)|2,{105,121,0}},
/* 14248 */ {(12<<2)|2,{105,121,0}},
/* 14249 */ {(12<<2)|2,{105,121,0}},
/* 14250 */ {(12<<2)|2,{105,121,0}},
/* 14251 */ {(12<<2)|2,{105,121,0}},
/* 14252 */ {(12<<2)|2,{105,121,0}},
/* 14253 */ {(12<<2)|2,{105,121,0}},
/* 14254 */ {(12<<2)|2,{105,121,0}},
/* 14255 */ {(12<<2)|2,{105,121,0}},
/* 14256 */ {(12<<2)|2,{105,122,0}},
/* 14257 */ {(12<<2)|2,{105,122,0}},
/* 14258 */ {(12<<2)|2,{105,122,0}},
/* 14259 */ {(12<<2)|2,{105,122,0}},
/* 14260 */ {(12<<2)|2,{105,122,0}},
/* 14261 */ {(12<<2)|2,{105,122,0}},
/* 14262 */ {(12<<2)|2,{105,122,0}},
/* 14263 */ {(12<<2)|2,{105,122,0}},
/* 14264 */ {(12<<2)|2,{105,122,0}},
/* 14265 */ {(12<<2)|2,{105,122,0}},
/* 14266 */ {(12<<2)|2,{105,122,0}},
/* 14267 */ {(12<<2)|2,{105,122,0}},
/* 14268 */ {(12<<2)|2,{105,122,0}},
/* 14269 */ {(12<<2)|2,{105,122,0}},
/* 14270 */ {(12<<2)|2,{105,122,0}},
/* 14271 */ {(12<<2)|2,{105,122,0}},
/* 14272 */ {(13<<2)|2,{105,38,0}},
/* 14273 */ {(13<<2)|2,{105,38,0}},
/* 14274 */ {(13<<2)|2,{105,38,0}},
/* 14275 */ {(13<<2)|2,{105,38,0}},
/* 14276 */ {(13<<2)|2,{105,38,0}},
/* 14277 */ {(13<<2)|2,{105,38,0}},
/* 14278 */ {(13<<2)|2,{105,38,0}},
/* 14279 */ {(13<<2)|2,{105,38,0}},
/* 14280 */ {(13<<2)|2,{105,42,0}},
/* 14281 */ {(13<<2)|2,{105,42,0}},
/* 14282 */ {(13<<2)|2,{105,42,0}},
/* 14283 */ {(13<<2)|2,{105,42,0}},
/* 14284 */ {(13<<2)|2,{105,42,0}},
/* 14285 */ {(13<<2)|2,{105,42,0}},
/* 14286 */ {(13<<2)|2,{105,42,0}},
/* 14287 */ {(13<<2)|2,{105,42,0}},
/* 14288 */ {(13<<2)|2,{105,44,0}},
/* 14289 */ {(13<<2)|2,{105,44,0}},
/* 14290 */ {(13<<2)|2,{105,44,0}},
/* 14291 */ {(13<<2)|2,{105,44,0}},
/* 14292 */ {(13<<2)|2,{105,44,0}},
/* 14293 */ {(13<<2)|2,{105,44,0}},
/* 14294 */ {(13<<2)|2,{105,44,0}},
/* 14295 */ {(13<<2)|2,{105,44,0}},
/* 14296 */ {(13<<2)|2,{105,59,0}},
/* 14297 */ {(13<<2)|2,{105,59,0}},
/* 14298 */ {(13<<2)|2,{105,59,0}},
/* 14299 */ {(13<<2)|2,{105,59,0}},
/* 14300 */ {(13<<2)|2,{105,59,0}},
/* 14301 */ {(13<<2)|2,{105,59,0}},
/* 14302 */ {(13<<2)|2,{105,59,0}},
/* 14303 */ {(13<<2)|2,{105,59,0}},
/* 14304 */ {(13<<2)|2,{105,88,0}},
/* 14305 */ {(13<<2)|2,{105,88,0}},
/* 14306 */ {(13<<2)|2,{105,88,0}},
/* 14307 */ {(13<<2)|2,{105,88,0}},
/* 14308 */ {(13<<2)|2,{105,88,0}},
/* 14309 */ {(13<<2)|2,{105,88,0}},
/* 14310 */ {(13<<2)|2,{105,88,0}},
/* 14311 */ {(13<<2)|2,{105,88,0}},
/* 14312 */ {(13<<2)|2,{105,90,0}},
/* 14313 */ {(13<<2)|2,{105,90,0}},
/* 14314 */ {(13<<2)|2,{105,90,0}},
/* 14315 */ {(13<<2)|2,{105,90,0}},
/* 14316 */ {(13<<2)|2,{105,90,0}},
/* 14317 */ {(13<<2)|2,{105,90,0}},
/* 14318 */ {(13<<2)|2,{105,90,0}},
/* 14319 */ {(13<<2)|2,{105,90,0}},
/* 14320 */ {(15<<2)|2,{105,33,0}},
/* 14321 */ {(15<<2)|2,{105,33,0}},
/* 14322 */ {(15<<2)|2,{105,34,0}},
/* 14323 */ {(15<<2)|2,{105,34,0}},
/* 14324 */ {(15<<2)|2,{105,40,0}},
/* 14325 */ {(15<<2)|2,{105,40,0}},
/* 14326 */ {(15<<2)|2,{105,41,0}},
/* 14327 */ {(15<<2)|2,{105,41,0}},
/* 14328 */ {(15<<2)|2,{105,63,0}},
/* 14329 */ {(15<<2)|2,{105,63,0}},
/* 14330 */ {(16<<2)|2,{105,39,0}},
/* 14331 */ {(16<<2)|2,{105,43,0}},
/* 14332 */ {(16<<2)|2,{105,124,0}},
/* 14333 */ {(5<<2)|1,{105,0,0}},
/* 14334 */ {(5<<2)|1,{105,0,0}},
/* 14335 */ {(5<<2)|1,{105,0,0}},
/* 14336 */ {(15<<2)|3,{111,48,48}},
/* 14337 */ {(15<<2)|3,{111,48,48}},
/* 14338 */ {(15<<2)|3,{111,48,49}},
/* 14339 */ {(15<<2)|3,{111,48,49}},
/* 14340 */ {(15<<2)|3,{111,48,50}},
/* 14341 */ {(15<<2)|3,{111,48,50}},
/* 14342 */ {(15<<2)|3,{111,48,97}},
/* 14343 */ {(15<<2)|3,{111,48,97}},
/* 14344 */ {(15<<2)|3,{111,48,99}},
/* 14345 */ {(15<<2)|3,{111,48,99}},
/* 14346 */ {(15<<2)|3,{111,48,101}},
/* 14347 */ {(15<<2)|3,{111,48,101}},
/* 14348 */ {(15<<2)|3,{111,48,105}},
/* 14349 */ {(15<<2)|3,{111,48,105}},
/* 14350 */ {(15<<2)|3,{111,48,111}},
/* 14351 */ {(15<<2)|3,{111,48,111}},
/* 14352 */ {(15<<2)|3,{111,48,115}},
/* 14353 */ {(15<<2)|3,{111,48,115}},
/* 14354 */ {(15<<2)|3,{111,48,116}},
/* 14355 */ {(15<<2)|3,{111,48,116}},
/* 14356 */ {(16<<2)|3,{111,48,32}},
/* 14357 */ {(16<<2)|3,{111,48,37}},
/* 14358 */ {(16<<2)|3,{111,48,45}},
/* 14359 */ {(16<<2)|3,{111,48,46}},
/* 14360 */ {(16<<2)|3,{111,48,47}},
/* 14361 */ {(16<<2)|3,{111,48,51}},
/* 14362 */ {(16<<2)|3,{111,48,52}},
/* 14363 */ {(16<<2)|3,{111,48,53}},
/* 14364 */ {(16<<2)|3,{111,48,54}},
/* 14365 */ {(16<<2)|3,{111,48,55}},
/* 14366 */ {(16<<2)|3,{111,48,56}},
/* 14367 */ {(16<<2)|3,{111,48,57}},
/* 14368 */ {(16<<2)|3,{111,48,61}},
/* 14369 */ {(16<<2)|3,{111,48,65}},
/* 14370 */ {(16<<2)|3,{111,48,95}},
/* 14371 */ {(16<<2)|3,{111,48,98}},
/* 14372 */ {(16<<2)|3,{111,48,100}},
/* 14373 */ {(16<<2)|3,{111,48,102}},
/* 14374 */ {(16<<2)|3,{111,48,103}},
/* 14375 */ {(16<<2)|3,{111,48,104}},
/* 14376 */ {(16<<2)|3,{111,48,108}},
/* 14377 */ {(16<<2)|3,{111,48,109}},
/* 14378 */ {(16<<2)|3,{111,48,110}},
/* 14379 */ {(16<<2)|3,{111,48,112}},
/* 14380 */ {(16<<2)|3,{111,48,114}},
/* 14381 */ {(16<<2)|3,{111,48,117}},
/* 14382 */ {(10<<2)|2,{111,48,0}},
/* 14383 */ {(10<<2)|2,{111,48,0}},
/* 14384 */ {(10<<2)|2,{111,48,0}},
/* 14385 */ {(10<<2)|2,{111,48,0}},
/* 14386 */ {(10<<2)|2,{111,48,0}},
/* 14387 */ {(10<<2)|2,{111,48,0}},
/* 14388 */ {(10<<2)|2,{111,48,0}},
/* 14389 */ {(10<<2)|2,{111,48,0}},
/* 14390 */ {(10<<2)|2,{111,48,0}},
/* 14391 */ {(10<<2)|2,{111,48,0}},
/* 14392 */ {(10<<2)|2,{111,48,0}},
/* 14393 */ {(10<<2)|2,{111,48,0}},
/* 14394 */ {(10<<2)|2,{111,48,0}},
/* 14395 */ {(10<<2)|2,{111,48,0}},
/* 14396 */ {(10<<2)|2,{111,48,0}},
/* 14397 */ {(10<<2)|2,{111,48,0}},
/* 14398 */ {(10<<2)|2,{111,48,0}},
/* 14399 */ {(10<<2)|2,{111,48,0}},
/* 14400 */ {(15<<2)|3,{111,49,48}},
/* 14401 */ {(15<<2)|3,{111,49,48}},
/* 14402 */ {(15<<2)|3,{111,49,49}},
/* 14403 */ {(15<<2)|3,{111,49,49}},
/* 14404 */ {(15<<2)|3,{111,49,50}},
/* 14405 */ {(15<<2)|3,{111,49,50}},
/* 14406 */ {(15<<2)|3,{111,49,97}},
/* 14407 */ {(15<<2)|3,{111,49,97}},
/* 14408 */ {(15<<2)|3,{111,49,99}},
/* 14409 */ {(15<<2)|3,{111,49,99}},
/* 14410 */ {(15<<2)|3,{111,49,101}},
/* 14411 */ {(15<<2)|3,{111,49,101}},
/* 14412 */ {(15<<2)|3,{111,49,105}},
/* 14413 */ {(15<<2)|3,{111,49,105}},
/* 14414 */ {(15<<2)|3,{111,49,111}},
/* 14415 */ {(15<<2)|3,{111,49,111}},
/* 14416 */ {(15<<2)|3,{111,49,115}},
/* 14417 */ {(15<<2)|3,{111,49,115}},
/* 14418 */ {(15<<2)|3,{111,49,116}},
/* 14419 */ {(15<<2)|3,{111,49,116}},
/* 14420 */ {(16<<2)|3,{111,49,32}},
/* 14421 */ {(16<<2)|3,{111,49,37}},
/* 14422 */ {(16<<2)|3,{111,49,45}},
/* 14423 */ {(16<<2)|3,{111,49,46}},
/* 14424 */ {(16<<2)|3,{111,49,47}},
/* 14425 */ {(16<<2)|3,{111,49,51}},
/* 14426 */ {(16<<2)|3,{111,49,52}},
/* 14427 */ {(16<<2)|3,{111,49,53}},
/* 14428 */ {(16<<2)|3,{111,49,54}},
/* 14429 */ {(16<<2)|3,{111,49,55}},
/* 14430 */ {(16<<2)|3,{111,49,56}},
/* 14431 */ {(16<<2)|3,{111,49,57}},
/* 14432 */ {(16<<2)|3,{111,49,61}},
/* 14433 */ {(16<<2)|3,{111,49,65}},
/* 14434 */ {(16<<2)|3,{111,49,95}},
/* 14435 */ {(16<<2)|3,{111,49,98}},
/* 14436 */ {(16<<2)|3,{111,49,100}},
/* 14437 */ {(16<<2)|3,{111,49,102}},
/* 14438 */ {(16<<2)|3,{111,49,103}},
/* 14439 */ {(16<<2)|3,{111,49,104}},
/* 14440 */ {(16<<2)|3,{111,49,108}},
/* 14441 */ {(16<<2)|3,{111,49,109}},
/* 14442 */ {(16<<2)|3,{111,49,110}},
/* 14443 */ {(16<<2)|3,{111,49,112}},
/* 14444 */ {(16<<2)|3,{111,49,114}},
/* 14445 */ {(16<<2)|3,{111,49,117}},
/* 14446 */ {(10<<2)|2,{111,49,0}},
/* 14447 */ {(10<<2)|2,{111,49,0}},
/* 14448 */ {(10<<2)|2,{111,49,0}},
/* 14449 */ {(10<<2)|2,{111,49,0}},
/* 14450 */ {(10<<2)|2,{111,49,0}},
/* 14451 */ {(10<<2)|2,{111,49,0}},
/* 14452 */ {(10<<2)|2,{111,49,0}},
/* 14453 */ {(10<<2)|2,{111,49,0}},
/* 14454 */ {(10<<2)|2,{111,49,0}},
/* 14455 */ {(10<<2)|2,{111,49,0}},
/* 14456 */ {(10<<2)|2,{111,49,0}},
/* 14457 */ {(10<<2)|2,{111,49,0}},
/* 14458 */ {(10<<2)|2,{111,49,0}},
/* 14459 */ {(10<<2)|2,{111,49,0}},
/* 14460 */ {(10<<2)|2,{111,49,0}},
/* 14461 */ {(10<<2)|2,{111,49,0}},
/* 14462 */ {(10<<2)|2,{111,49,0}},
/* 14463 */ {(10<<2)|2,{111,49,0}},
/* 14464 */ {(15<<2)|3,{111,50,48}},
/* 14465 */ {(15<<2)|3,{111,50,48}},
/* 14466 */ {(15<<2)|3,{111,50,49}},
/* 14467 */ {(15<<2)|3,{111,50,49}},
/* 14468 */ {(15<<2)|3,{111,50,50}},
/* 14469 */ {(15<<2)|3,{111,50,50}},
/* 14470 */ {(15<<2)|3,{111,50,97}},
/* 14471 */ {(15<<2)|3,{111,50,97}},
/* 14472 */ {(15<<2)|3,{111,50,99}},
/* 14473 */ {(15<<2)|3,{111,50,99}},
/* 14474 */ {(15<<2)|3,{111,50,101}},
/* 14475 */ {(15<<2)|3,{111,50,101}},
/* 14476 */ {(15<<2)|3,{111,50,105}},
/* 14477 */ {(15<<2)|3,{111,50,105}},
/* 14478 */ {(15<<2)|3,{111,50,111}},
/* 14479 */ {(15<<2)|3,{111,50,111}},
/* 14480 */ {(15<<2)|3,{111,50,115}},
/* 14481 */ {(15<<2)|3,{111,50,115}},
/* 14482 */ {(15<<2)|3,{111,50,116}},
/* 14483 */ {(15<<2)|3,{111,50,116}},
/* 14484 */ {(16<<2)|3,{111,50,32}},
/* 14485 */ {(16<<2)|3,{111,50,37}},
/* 14486 */ {(16<<2)|3,{111,50,45}},
/* 14487 */ {(16<<2)|3,{111,50,46}},
/* 14488 */ {(16<<2)|3,{111,50,47}},
/* 14489 */ {(16<<2)|3,{111,50,51}},
/* 14490 */ {(16<<2)|3,{111,50,52}},
/* 14491 */ {(16<<2)|3,{111,50,53}},
/* 14492 */ {(16<<2)|3,{111,50,54}},
/* 14493 */ {(16<<2)|3,{111,50,55}},
/* 14494 */ {(16<<2)|3,{111,50,56}},
/* 14495 */ {(16<<2)|3,{111,50,57}},
/* 14496 */ {(16<<2)|3,{111,50,61}},
/* 14497 */ {(16<<2)|3,{111,50,65}},
/* 14498 */ {(16<<2)|3,{111,50,95}},
/* 14499 */ {(16<<2)|3,{111,50,98}},
/* 14500 */ {(16<<2)|3,{111,50,100}},
/* 14501 */ {(16<<2)|3,{111,50,102}},
/* 14502 */ {(16<<2)|3,{111,50,103}},
/* 14503 */ {(16<<2)|3,{111,50,104}},
/* 14504 */ {(16<<2)|3,{111,50,108}},
/* 14505 */ {(16<<2)|3,{111,50,109}},
/* 14506 */ {(16<<2)|3,{111,50,110}},
/* 14507 */ {(16<<2)|3,{111,50,112}},
/* 14508 */ {(16<<2)|3,{111,50,114}},
/* 14509 */ {(16<<2)|3,{111,50,117}},
/* 14510 */ {(10<<2)|2,{111,50,0}},
/* 14511 */ {(10<<2)|2,{111,50,0}},
/* 14512 */ {(10<<2)|2,{111,50,0}},
/* 14513 */ {(10<<2)|2,{111,50,0}},
/* 14514 */ {(10<<2)|2,{111,50,0}},
/* 14515 */ {(10<<2)|2,{111,50,0}},
/* 14516 */ {(10<<2)|2,{111,50,0}},
/* 14517 */ {(10<<2)|2,{111,50,0}},
/* 14518 */ {(10<<2)|2,{111,50,0}},
/* 14519 */ {(10<<2)|2,{111,50,0}},
/* 14520 */ {(10<<2)|2,{111,50,0}},
/* 14521 */ {(10<<2)|2,{111,50,0}},
/* 14522 */ {(10<<2)|2,{111,50,0}},
/* 14523 */ {(10<<2)|2,{111,50,0}},
/* 14524 */ {(10<<2)|2,{111,50,0}},
/* 14525 */ {(10<<2)|2,{111,50,0}},
/* 14526 */ {(10<<2)|2,{111,50,0}},
/* 14527 */ {(10<<2)|2,{111,50,0}},
/* 14528 */ {(15<<2)|3,{111,97,48}},
/* 14529 */ {(15<<2)|3,{111,97,48}},
/* 14530 */ {(15<<2)|3,{111,97,49}},
/* 14531 */ {(15<<2)|3,{111,97,49}},
/* 14532 */ {(15<<2)|3,{111,97,50}},
/* 14533 */ {(15<<2)|3,{111,97,50}},
/* 14534 */ {(15<<2)|3,{111,97,97}},
/* 14535 */ {(15<<2)|3,{111,97,97}},
/* 14536 */ {(15<<2)|3,{111,97,99}},
/* 14537 */ {(15<<2)|3,{111,97,99}},
/* 14538 */ {(15<<2)|3,{111,97,101}},
/* 14539 */ {(15<<2)|3,{111,97,101}},
/* 14540 */ {(15<<2)|3,{111,97,105}},
/* 14541 */ {(15<<2)|3,{111,97,105}},
/* 14542 */ {(15<<2)|3,{111,97,111}},
/* 14543 */ {(15<<2)|3,{111,97,111}},
/* 14544 */ {(15<<2)|3,{111,97,115}},
/* 14545 */ {(15<<2)|3,{111,97,115}},
/* 14546 */ {(15<<2)|3,{111,97,116}},
/* 14547 */ {(15<<2)|3,{111,97,116}},
/* 14548 */ {(16<<2)|3,{111,97,32}},
/* 14549 */ {(16<<2)|3,{111,97,37}},
/* 14550 */ {(16<<2)|3,{111,97,45}},
/* 14551 */ {(16<<2)|3,{111,97,46}},
/* 14552 */ {(16<<2)|3,{111,97,47}},
/* 14553 */ {(16<<2)|3,{111,97,51}},
/* 14554 */ {(16<<2)|3,{111,97,52}},
/* 14555 */ {(16<<2)|3,{111,97,53}},
/* 14556 */ {(16<<2)|3,{111,97,54}},
/* 14557 */ {(16<<2)|3,{111,97,55}},
/* 14558 */ {(16<<2)|3,{111,97,56}},
/* 14559 */ {(16<<2)|3,{111,97,57}},
/* 14560 */ {(16<<2)|3,{111,97,61}},
/* 14561 */ {(16<<2)|3,{111,97,65}},
/* 14562 */ {(16<<2)|3,{111,97,95}},
/* 14563 */ {(16<<2)|3,{111,97,98}},
/* 14564 */ {(16<<2)|3,{111,97,100}},
/* 14565 */ {(16<<2)|3,{111,97,102}},
/* 14566 */ {(16<<2)|3,{111,97,103}},
/* 14567 */ {(16<<2)|3,{111,97,104}},
/* 14568 */ {(16<<2)|3,{111,97,108}},
/* 14569 */ {(16<<2)|3,{111,97,109}},
/* 14570 */ {(16<<2)|3,{111,97,110}},
/* 14571 */ {(16<<2)|3,{111,97,112}},
/* 14572 */ {(16<<2)|3,{111,97,114}},
/* 14573 */ {(16<<2)|3,{111,97,117}},
/* 14574 */ {(10<<2)|2,{111,97,0}},
/* 14575 */ {(10<<2)|2,{111,97,0}},
/* 14576 */ {(10<<2)|2,{111,97,0}},
/* 14577 */ {(10<<2)|2,{111,97,0}},
/* 14578 */ {(10<<2)|2,{111,97,0}},
/* 14579 */ {(10<<2)|2,{111,97,0}},
/* 14580 */ {(10<<2)|2,{111,97,0}},
/* 14581 */ {(10<<2)|2,{111,97,0}},
/* 14582 */ {(10<<2)|2,{111,97,0}},
/* 14583 */ {(10<<2)|2,{111,97,0}},
/* 14584 */ {(10<<2)|2,{111,97,0}},
/* 14585 */ {(10<<2)|2,{111,97,0}},
/* 14586 */ {(10<<2)|2,{111,97,0}},
/* 14587 */ {(10<<2)|2,{111,97,0}},
/* 14588 */ {(10<<2)|2,{111,97,0}},
/* 14589 */ {(10<<2)|2,{111,97,0}},
/* 14590 */ {(10<<2)|2,{111,97,0}},
/* 14591 */ {(10<<2)|2,{111,97,0}},
/* 14592 */ {(15<<2)|3,{111,99,48}},
/* 14593 */ {(15<<2)|3,{111,99,48}},
/* 14594 */ {(15<<2)|3,{111,99,49}},
/* 14595 */ {(15<<2)|3,{111,99,49}},
/* 14596 */ {(15<<2)|3,{111,99,50}},
/* 14597 */ {(15<<2)|3,{111,99,50}},
/* 14598 */ {(15<<2)|3,{111,99,97}},
/* 14599 */ {(15<<2)|3,{111,99,97}},
/* 14600 */ {(15<<2)|3,{111,99,99}},
/* 14601 */ {(15<<2)|3,{111,99,99}},
/* 14602 */ {(15<<2)|3,{111,99,101}},
/* 14603 */ {(15<<2)|3,{111,99,101}},
/* 14604 */ {(15<<2)|3,{111,99,105}},
/* 14605 */ {(15<<2)|3,{111,99,105}},
/* 14606 */ {(15<<2)|3,{111,99,111}},
/* 14607 */ {(15<<2)|3,{111,99,111}},
/* 14608 */ {(15<<2)|3,{111,99,115}},
/* 14609 */ {(15<<2)|3,{111,99,115}},
/* 14610 */ {(15<<2)|3,{111,99,116}},
/* 14611 */ {(15<<2)|3,{111,99,116}},
/* 14612 */ {(16<<2)|3,{111,99,32}},
/* 14613 */ {(16<<2)|3,{111,99,37}},
/* 14614 */ {(16<<2)|3,{111,99,45}},
/* 14615 */ {(16<<2)|3,{111,99,46}},
/* 14616 */ {(16<<2)|3,{111,99,47}},
/* 14617 */ {(16<<2)|3,{111,99,51}},
/* 14618 */ {(16<<2)|3,{111,99,52}},
/* 14619 */ {(16<<2)|3,{111,99,53}},
/* 14620 */ {(16<<2)|3,{111,99,54}},
/* 14621 */ {(16<<2)|3,{111,99,55}},
/* 14622 */ {(16<<2)|3,{111,99,56}},
/* 14623 */ {(16<<2)|3,{111,99,57}},
/* 14624 */ {(16<<2)|3,{111,99,61}},
/* 14625 */ {(16<<2)|3,{111,99,65}},
/* 14626 */ {(16<<2)|3,{111,99,95}},
/* 14627 */ {(16<<2)|3,{111,99,98}},
/* 14628 */ {(16<<2)|3,{111,99,100}},
/* 14629 */ {(16<<2)|3,{111,99,102}},
/* 14630 */ {(16<<2)|3,{111,99,103}},
/* 14631 */ {(16<<2)|3,{111,99,104}},
/* 14632 */ {(16<<2)|3,{111,99,108}},
/* 14633 */ {(16<<2)|3,{111,99,109}},
/* 14634 */ {(16<<2)|3,{111,99,110}},
/* 14635 */ {(16<<2)|3,{111,99,112}},
/* 14636 */ {(16<<2)|3,{111,99,114}},
/* 14637 */ {(16<<2)|3,{111,99,117}},
/* 14638 */ {(10<<2)|2,{111,99,0}},
/* 14639 */ {(10<<2)|2,{111,99,0}},
/* 14640 */ {(10<<2)|2,{111,99,0}},
/* 14641 */ {(10<<2)|2,{111,99,0}},
/* 14642 */ {(10<<2)|2,{111,99,0}},
/* 14643 */ {(10<<2)|2,{111,99,0}},
/* 14644 */ {(10<<2)|2,{111,99,0}},
/* 14645 */ {(10<<2)|2,{111,99,0}},
/* 14646 */ {(10<<2)|2,{111,99,0}},
/* 14647 */ {(10<<2)|2,{111,99,0}},
/* 14648 */ {(10<<2)|2,{111,99,0}},
/* 14649 */ {(10<<2)|2,{111,99,0}},
/* 14650 */ {(10<<2)|2,{111,99,0}},
/* 14651 */ {(10<<2)|2,{111,99,0}},
/* 14652 */ {(10<<2)|2,{111,99,0}},
/* 14653 */ {(10<<2)|2,{111,99,0}},
/* 14654 */ {(10<<2)|2,{111,99,0}},
/* 14655 */ {(10<<2)|2,{111,99,0}},
/* 14656 */ {(15<<2)|3,{111,101,48}},
/* 14657 */ {(15<<2)|3,{111,101,48}},
/* 14658 */ {(15<<2)|3,{111,101,49}},
/* 14659 */ {(15<<2)|3,{111,101,49}},
/* 14660 */ {(15<<2)|3,{111,101,50}},
/* 14661 */ {(15<<2)|3,{111,101,50}},
/* 14662 */ {(15<<2)|3,{111,101,97}},
/* 14663 */ {(15<<2)|3,{111,101,97}},
/* 14664 */ {(15<<2)|3,{111,101,99}},
/* 14665 */ {(15<<2)|3,{111,101,99}},
/* 14666 */ {(15<<2)|3,{111,101,101}},
/* 14667 */ {(15<<2)|3,{111,101,101}},
/* 14668 */ {(15<<2)|3,{111,101,105}},
/* 14669 */ {(15<<2)|3,{111,101,105}},
/* 14670 */ {(15<<2)|3,{111,101,111}},
/* 14671 */ {(15<<2)|3,{111,101,111}},
/* 14672 */ {(15<<2)|3,{111,101,115}},
/* 14673 */ {(15<<2)|3,{111,101,115}},
/* 14674 */ {(15<<2)|3,{111,101,116}},
/* 14675 */ {(15<<2)|3,{111,101,116}},
/* 14676 */ {(16<<2)|3,{111,101,32}},
/* 14677 */ {(16<<2)|3,{111,101,37}},
/* 14678 */ {(16<<2)|3,{111,101,45}},
/* 14679 */ {(16<<2)|3,{111,101,46}},
/* 14680 */ {(16<<2)|3,{111,101,47}},
/* 14681 */ {(16<<2)|3,{111,101,51}},
/* 14682 */ {(16<<2)|3,{111,101,52}},
/* 14683 */ {(16<<2)|3,{111,101,53}},
/* 14684 */ {(16<<2)|3,{111,101,54}},
/* 14685 */ {(16<<2)|3,{111,101,55}},
/* 14686 */ {(16<<2)|3,{111,101,56}},
/* 14687 */ {(16<<2)|3,{111,101,57}},
/* 14688 */ {(16<<2)|3,{111,101,61}},
/* 14689 */ {(16<<2)|3,{111,101,65}},
/* 14690 */ {(16<<2)|3,{111,101,95}},
/* 14691 */ {(16<<2)|3,{111,101,98}},
/* 14692 */ {(16<<2)|3,{111,101,100}},
/* 14693 */ {(16<<2)|3,{111,101,102}},
/* 14694 */ {(16<<2)|3,{111,101,103}},
/* 14695 */ {(16<<2)|3,{111,101,104}},
/* 14696 */ {(16<<2)|3,{111,101,108}},
/* 14697 */ {(16<<2)|3,{111,101,109}},
/* 14698 */ {(16<<2)|3,{111,101,110}},
/* 14699 */ {(16<<2)|3,{111,101,112}},
/* 14700 */ {(16<<2)|3,{111,101,114}},
/* 14701 */ {(16<<2)|3,{111,101,117}},
/* 14702 */ {(10<<2)|2,{111,101,0}},
/* 14703 */ {(10<<2)|2,{111,101,0}},
/* 14704 */ {(10<<2)|2,{111,101,0}},
/* 14705 */ {(10<<2)|2,{111,101,0}},
/* 14706 */ {(10<<2)|2,{111,101,0}},
/* 14707 */ {(10<<2)|2,{111,101,0}},
/* 14708 */ {(10<<2)|2,{111,101,0}},
/* 14709 */ {(10<<2)|2,{111,101,0}},
/* 14710 */ {(10<<2)|2,{111,101,0}},
/* 14711 */ {(10<<2)|2,{111,101,0}},
/* 14712 */ {(10<<2)|2,{111,101,0}},
/* 14713 */ {(10<<2)|2,{111,101,0}},
/* 14714 */ {(10<<2)|2,{111,101,0}},
/* 14715 */ {(10<<2)|2,{111,101,0}},
/* 14716 */ {(10<<2)|2,{111,101,0}},
/* 14717 */ {(10<<2)|2,{111,101,0}},
/* 14718 */ {(10<<2)|2,{111,101,0}},
/* 14719 */ {(10<<2)|2,{111,101,0}},
/* 14720 */ {(15<<2)|3,{111,105,48}},
/* 14721 */ {(15<<2)|3,{111,105,48}},
/* 14722 */ {(15<<2)|3,{111,105,49}},
/* 14723 */ {(15<<2)|3,{111,105,49}},
/* 14724 */ {(15<<2)|3,{111,105,50}},
/* 14725 */ {(15<<2)|3,{111,105,50}},
/* 14726 */ {(15<<2)|3,{111,105,97}},
/* 14727 */ {(15<<2)|3,{111,105,97}},
/* 14728 */ {(15<<2)|3,{111,105,99}},
/* 14729 */ {(15<<2)|3,{111,105,99}},
/* 14730 */ {(15<<2)|3,{111,105,101}},
/* 14731 */ {(15<<2)|3,{111,105,101}},
/* 14732 */ {(15<<2)|3,{111,105,105}},
/* 14733 */ {(15<<2)|3,{111,105,105}},
/* 14734 */ {(15<<2)|3,{111,105,111}},
/* 14735 */ {(15<<2)|3,{111,105,111}},
/* 14736 */ {(15<<2)|3,{111,105,115}},
/* 14737 */ {(15<<2)|3,{111,105,115}},
/* 14738 */ {(15<<2)|3,{111,105,116}},
/* 14739 */ {(15<<2)|3,{111,105,116}},
/* 14740 */ {(16<<2)|3,{111,105,32}},
/* 14741 */ {(16<<2)|3,{111,105,37}},
/* 14742 */ {(16<<2)|3,{111,105,45}},
/* 14743 */ {(16<<2)|3,{111,105,46}},
/* 14744 */ {(16<<2)|3,{111,105,47}},
/* 14745 */ {(16<<2)|3,{111,105,51}},
/* 14746 */ {(16<<2)|3,{111,105,52}},
/* 14747 */ {(16<<2)|3,{111,105,53}},
/* 14748 */ {(16<<2)|3,{111,105,54}},
/* 14749 */ {(16<<2)|3,{111,105,55}},
/* 14750 */ {(16<<2)|3,{111,105,56}},
/* 14751 */ {(16<<2)|3,{111,105,57}},
/* 14752 */ {(16<<2)|3,{111,105,61}},
/* 14753 */ {(16<<2)|3,{111,105,65}},
/* 14754 */ {(16<<2)|3,{111,105,95}},
/* 14755 */ {(16<<2)|3,{111,105,98}},
/* 14756 */ {(16<<2)|3,{111,105,100}},
/* 14757 */ {(16<<2)|3,{111,105,102}},
/* 14758 */ {(16<<2)|3,{111,105,103}},
/* 14759 */ {(16<<2)|3,{111,105,104}},
/* 14760 */ {(16<<2)|3,{111,105,108}},
/* 14761 */ {(16<<2)|3,{111,105,109}},
/* 14762 */ {(16<<2)|3,{111,105,110}},
/* 14763 */ {(16<<2)|3,{111,105,112}},
/* 14764 */ {(16<<2)|3,{111,105,114}},
/* 14765 */ {(16<<2)|3,{111,105,117}},
/* 14766 */ {(10<<2)|2,{111,105,0}},
/* 14767 */ {(10<<2)|2,{111,105,0}},
/* 14768 */ {(10<<2)|2,{111,105,0}},
/* 14769 */ {(10<<2)|2,{111,105,0}},
/* 14770 */ {(10<<2)|2,{111,105,0}},
/* 14771 */ {(10<<2)|2,{111,105,0}},
/* 14772 */ {(10<<2)|2,{111,105,0}},
/* 14773 */ {(10<<2)|2,{111,105,0}},
/* 14774 */ {(10<<2)|2,{111,105,0}},
/* 14775 */ {(10<<2)|2,{111,105,0}},
/* 14776 */ {(10<<2)|2,{111,105,0}},
/* 14777 */ {(10<<2)|2,{111,105,0}},
/* 14778 */ {(10<<2)|2,{111,105,0}},
/* 14779 */ {(10<<2)|2,{111,105,0}},
/* 14780 */ {(10<<2)|2,{111,105,0}},
/* 14781 */ {(10<<2)|2,{111,105,0}},
/* 14782 */ {(10<<2)|2,{111,105,0}},
/* 14783 */ {(10<<2)|2,{111,105,0}},
/* 14784 */ {(15<<2)|3,{111,111,48}},
/* 14785 */ {(15<<2)|3,{111,111,48}},
/* 14786 */ {(15<<2)|3,{111,111,49}},
/* 14787 */ {(15<<2)|3,{111,111,49}},
/* 14788 */ {(15<<2)|3,{111,111,50}},
/* 14789 */ {(15<<2)|3,{111,111,50}},
/* 14790 */ {(15<<2)|3,{111,111,97}},
/* 14791 */ {(15<<2)|3,{111,111,97}},
/* 14792 */ {(15<<2)|3,{111,111,99}},
/* 14793 */ {(15<<2)|3,{111,111,99}},
/* 14794 */ {(15<<2)|3,{111,111,101}},
/* 14795 */ {(15<<2)|3,{111,111,101}},
/* 14796 */ {(15<<2)|3,{111,111,105}},
/* 14797 */ {(15<<2)|3,{111,111,105}},
/* 14798 */ {(15<<2)|3,{111,111,111}},
/* 14799 */ {(15<<2)|3,{111,111,111}},
/* 14800 */ {(15<<2)|3,{111,111,115}},
/* 14801 */ {(15<<2)|3,{111,111,115}},
/* 14802 */ {(15<<2)|3,{111,111,116}},
/* 14803 */ {(15<<2)|3,{111,111,116}},
/* 14804 */ {(16<<2)|3,{111,111,32}},
/* 14805 */ {(16<<2)|3,{111,111,37}},
/* 14806 */ {(16<<2)|3,{111,111,45}},
/* 14807 */ {(16<<2)|3,{111,111,46}},
/* 14808 */ {(16<<2)|3,{111,111,47}},
/* 14809 */ {(16<<2)|3,{111,111,51}},
/* 14810 */ {(16<<2)|3,{111,111,52}},
/* 14811 */ {(16<<2)|3,{111,111,53}},
/* 14812 */ {(16<<2)|3,{111,111,54}},
/* 14813 */ {(16<<2)|3,{111,111,55}},
/* 14814 */ {(16<<2)|3,{111,111,56}},
/* 14815 */ {(16<<2)|3,{111,111,57}},
/* 14816 */ {(16<<2)|3,{111,111,61}},
/* 14817 */ {(16<<2)|3,{111,111,65}},
/* 14818 */ {(16<<2)|3,{111,111,95}},
/* 14819 */ {(16<<2)|3,{111,111,98}},
/* 14820 */ {(16<<2)|3,{111,111,100}},
/* 14821 */ {(16<<2)|3,{111,111,102}},
/* 14822 */ {(16<<2)|3,{111,111,103}},
/* 14823 */ {(16<<2)|3,{111,111,104}},
/* 14824 */ {(16<<2)|3,{111,111,108}},
/* 14825 */ {(16<<2)|3,{111,111,109}},
/* 14826 */ {(16<<2)|3,{111,111,110}},
/* 14827 */ {(16<<2)|3,{111,111,112}},
/* 14828 */ {(16<<2)|3,{111,111,114}},
/* 14829 */ {(16<<2)|3,{111,111,117}},
/* 14830 */ {(10<<2)|2,{111,111,0}},
/* 14831 */ {(10<<2)|2,{111,111,0}},
/* 14832 */ {(10<<2)|2,{111,111,0}},
/* 14833 */ {(10<<2)|2,{111,111,0}},
/* 14834 */ {(10<<2)|2,{111,111,0}},
/* 14835 */ {(10<<2)|2,{111,111,0}},
/* 14836 */ {(10<<2)|2,{111,111,0}},
/* 14837 */ {(10<<2)|2,{111,111,0}},
/* 14838 */ {(10<<2)|2,{111,111,0}},
/* 14839 */ {(10<<2)|2,{111,111,0}},
/* 14840 */ {(10<<2)|2,{111,111,0}},
/* 14841 */ {(10<<2)|2,{111,111,0}},
/* 14842 */ {(10<<2)|2,{111,111,0}},
/* 14843 */ {(10<<2)|2,{111,111,0}},
/* 14844 */ {(10<<2)|2,{111,111,0}},
/* 14845 */ {(10<<2)|2,{111,111,0}},
/* 14846 */ {(10<<2)|2,{111,111,0}},
/* 14847 */ {(10<<2)|2,{111,111,0}},
/* 14848 */ {(15<<2)|3,{111,115,48}},
/* 14849 */ {(15<<2)|3,{111,115,48}},
/* 14850 */ {(15<<2)|3,{111,115,49}},
/* 14851 */ {(15<<2)|3,{111,115,49}},
/* 14852 */ {(15<<2)|3,{111,115,50}},
/* 14853 */ {(15<<2)|3,{111,115,50}},
/* 14854 */ {(15<<2)|3,{111,115,97}},
/* 14855 */ {(15<<2)|3,{111,115,97}},
/* 14856 */ {(15<<2)|3,{111,115,99}},
/* 14857 */ {(15<<2)|3,{111,115,99}},
/* 14858 */ {(15<<2)|3,{111,115,101}},
/* 14859 */ {(15<<2)|3,{111,115,101}},
/* 14860 */ {(15<<2)|3,{111,115,105}},
/* 14861 */ {(15<<2)|3,{111,115,105}},
/* 14862 */ {(15<<2)|3,{111,115,111}},
/* 14863 */ {(15<<2)|3,{111,115,111}},
/* 14864 */ {(15<<2)|3,{111,115,115}},
/* 14865 */ {(15<<2)|3,{111,115,115}},
/* 14866 */ {(15<<2)|3,{111,115,116}},
/* 14867 */ {(15<<2)|3,{111,115,116}},
/* 14868 */ {(16<<2)|3,{111,115,32}},
/* 14869 */ {(16<<2)|3,{111,115,37}},
/* 14870 */ {(16<<2)|3,{111,115,45}},
/* 14871 */ {(16<<2)|3,{111,115,46}},
/* 14872 */ {(16<<2)|3,{111,115,47}},
/* 14873 */ {(16<<2)|3,{111,115,51}},
/* 14874 */ {(16<<2)|3,{111,115,52}},
/* 14875 */ {(16<<2)|3,{111,115,53}},
/* 14876 */ {(16<<2)|3,{111,115,54}},
/* 14877 */ {(16<<2)|3,{111,115,55}},
/* 14878 */ {(16<<2)|3,{111,115,56}},
/* 14879 */ {(16<<2)|3,{111,115,57}},
/* 14880 */ {(16<<2)|3,{111,115,61}},
/* 14881 */ {(16<<2)|3,{111,115,65}},
/* 14882 */ {(16<<2)|3,{111,115,95}},
/* 14883 */ {(16<<2)|3,{111,115,98}},
/* 14884 */ {(16<<2)|3,{111,115,100}},
/* 14885 */ {(16<<2)|3,{111,115,102}},
/* 14886 */ {(16<<2)|3,{111,115,103}},
/* 14887 */ {(16<<2)|3,{111,115,104}},
/* 14888 */ {(16<<2)|3,{111,115,108}},
/* 14889 */ {(16<<2)|3,{111,115,109}},
/* 14890 */ {(16<<2)|3,{111,115,110}},
/* 14891 */ {(16<<2)|3,{111,115,112}},
/* 14892 */ {(16<<2)|3,{111,115,114}},
/* 14893 */ {(16<<2)|3,{111,115,117}},
/* 14894 */ {(10<<2)|2,{111,115,0}},
/* 14895 */ {(10<<2)|2,{111,115,0}},
/* 14896 */ {(10<<2)|2,{111,115,0}},
/* 14897 */ {(10<<2)|2,{111,115,0}},
/* 14898 */ {(10<<2)|2,{111,115,0}},
/* 14899 */ {(10<<2)|2,{111,115,0}},
/* 14900 */ {(10<<2)|2,{111,115,0}},
/* 14901 */ {(10<<2)|2,{111,115,0}},
/* 14902 */ {(10<<2)|2,{111,115,0}},
/* 14903 */ {(10<<2)|2,{111,115,0}},
/* 14904 */ {(10<<2)|2,{111,115,0}},
/* 14905 */ {(10<<2)|2,{111,115,0}},
/* 14906 */ {(10<<2)|2,{111,115,0}},
/* 14907 */ {(10<<2)|2,{111,115,0}},
/* 14908 */ {(10<<2)|2,{111,115,0}},
/* 14909 */ {(10<<2)|2,{111,115,0}},
/* 14910 */ {(10<<2)|2,{111,115,0}},
/* 14911 */ {(10<<2)|2,{111,115,0}},
/* 14912 */ {(15<<2)|3,{111,116,48}},
/* 14913 */ {(15<<2)|3,{111,116,48}},
/* 14914 */ {(15<<2)|3,{111,116,49}},
/* 14915 */ {(15<<2)|3,{111,116,49}},
/* 14916 */ {(15<<2)|3,{111,116,50}},
/* 14917 */ {(15<<2)|3,{111,116,50}},
/* 14918 */ {(15<<2)|3,{111,116,97}},
/* 14919 */ {(15<<2)|3,{111,116,97}},
/* 14920 */ {(15<<2)|3,{111,116,99}},
/* 14921 */ {(15<<2)|3,{111,116,99}},
/* 14922 */ {(15<<2)|3,{111,116,101}},
/* 14923 */ {(15<<2)|3,{111,116,101}},
/* 14924 */ {(15<<2)|3,{111,116,105}},
/* 14925 */ {(15<<2)|3,{111,116,105}},
/* 14926 */ {(15<<2)|3,{111,116,111}},
/* 14927 */ {(15<<2)|3,{111,116,111}},
/* 14928 */ {(15<<2)|3,{111,116,115}},
/* 14929 */ {(15<<2)|3,{111,116,115}},
/* 14930 */ {(15<<2)|3,{111,116,116}},
/* 14931 */ {(15<<2)|3,{111,116,116}},
/* 14932 */ {(16<<2)|3,{111,116,32}},
/* 14933 */ {(16<<2)|3,{111,116,37}},
/* 14934 */ {(16<<2)|3,{111,116,45}},
/* 14935 */ {(16<<2)|3,{111,116,46}},
/* 14936 */ {(16<<2)|3,{111,116,47}},
/* 14937 */ {(16<<2)|3,{111,116,51}},
/* 14938 */ {(16<<2)|3,{111,116,52}},
/* 14939 */ {(16<<2)|3,{111,116,53}},
/* 14940 */ {(16<<2)|3,{111,116,54}},
/* 14941 */ {(16<<2)|3,{111,116,55}},
/* 14942 */ {(16<<2)|3,{111,116,56}},
/* 14943 */ {(16<<2)|3,{111,116,57}},
/* 14944 */ {(16<<2)|3,{111,116,61}},
/* 14945 */ {(16<<2)|3,{111,116,65}},
/* 14946 */ {(16<<2)|3,{111,116,95}},
/* 14947 */ {(16<<2)|3,{111,116,98}},
/* 14948 */ {(16<<2)|3,{111,116,100}},
/* 14949 */ {(16<<2)|3,{111,116,102}},
/* 14950 */ {(16<<2)|3,{111,116,103}},
/* 14951 */ {(16<<2)|3,{111,116,104}},
/* 14952 */ {(16<<2)|3,{111,116,108}},
/* 14953 */ {(16<<2)|3,{111,116,109}},
/* 14954 */ {(16<<2)|3,{111,116,110}},
/* 14955 */ {(16<<2)|3,{111,116,112}},
/* 14956 */ {(16<<2)|3,{111,116,114}},
/* 14957 */ {(16<<2)|3,{111,116,117}},
/* 14958 */ {(10<<2)|2,{111,116,0}},
/* 14959 */ {(10<<2)|2,{111,116,0}},
/* 14960 */ {(10<<2)|2,{111,116,0}},
/* 14961 */ {(10<<2)|2,{111,116,0}},
/* 14962 */ {(10<<2)|2,{111,116,0}},
/* 14963 */ {(10<<2)|2,{111,116,0}},
/* 14964 */ {(10<<2)|2,{111,116,0}},
/* 14965 */ {(10<<2)|2,{111,116,0}},
/* 14966 */ {(10<<2)|2,{111,116,0}},
/* 14967 */ {(10<<2)|2,{111,116,0}},
/* 14968 */ {(10<<2)|2,{111,116,0}},
/* 14969 */ {(10<<2)|2,{111,116,0}},
/* 14970 */ {(10<<2)|2,{111,116,0}},
/* 14971 */ {(10<<2)|2,{111,116,0}},
/* 14972 */ {(10<<2)|2,{111,116,0}},
/* 14973 */ {(10<<2)|2,{111,116,0}},
/* 14974 */ {(10<<2)|2,{111,116,0}},
/* 14975 */ {(10<<2)|2,{111,116,0}},
/* 14976 */ {(16<<2)|3,{111,32,48}},
/* 14977 */ {(16<<2)|3,{111,32,49}},
/* 14978 */ {(16<<2)|3,{111,32,50}},
/* 14979 */ {(16<<2)|3,{111,32,97}},
/* 14980 */ {(16<<2)|3,{111,32,99}},
/* 14981 */ {(16<<2)|3,{111,32,101}},
/* 14982 */ {(16<<2)|3,{111,32,105}},
/* 14983 */ {(16<<2)|3,{111,32,111}},
/* 14984 */ {(16<<2)|3,{111,32,115}},
/* 14985 */ {(16<<2)|3,{111,32,116}},
/* 14986 */ {(11<<2)|2,{111,32,0}},
/* 14987 */ {(11<<2)|2,{111,32,0}},
/* 14988 */ {(11<<2)|2,{111,32,0}},
/* 14989 */ {(11<<2)|2,{111,32,0}},
/* 14990 */ {(11<<2)|2,{111,32,0}},
/* 14991 */ {(11<<2)|2,{111,32,0}},
/* 14992 */ {(11<<2)|2,{111,32,0}},
/* 14993 */ {(11<<2)|2,{111,32,0}},
/* 14994 */ {(11<<2)|2,{111,32,0}},
/* 14995 */ {(11<<2)|2,{111,32,0}},
/* 14996 */ {(11<<2)|2,{111,32,0}},
/* 14997 */ {(11<<2)|2,{111,32,0}},
/* 14998 */ {(11<<2)|2,{111,32,0}},
/* 14999 */ {(11<<2)|2,{111,32,0}},
/* 15000 */ {(11<<2)|2,{111,32,0}},
/* 15001 */ {(11<<2)|2,{111,32,0}},
/* 15002 */ {(11<<2)|2,{111,32,0}},
/* 15003 */ {(11<<2)|2,{111,32,0}},
/* 15004 */ {(11<<2)|2,{111,32,0}},
/* 15005 */ {(11<<2)|2,{111,32,0}},
/* 15006 */ {(11<<2)|2,{111,32,0}},
/* 15007 */ {(11<<2)|2,{111,32,0}},
/* 15008 */ {(16<<2)|3,{111,37,48}},
/* 15009 */ {(16<<2)|3,{111,37,49}},
/* 15010 */ {(16<<2)|3,{111,37,50}},
/* 15011 */ {(16<<2)|3,{111,37,97}},
/* 15012 */ {(16<<2)|3,{111,37,99}},
/* 15013 */ {(16<<2)|3,{111,37,101}},
/* 15014 */ {(16<<2)|3,{111,37,105}},
/* 15015 */ {(16<<2)|3,{111,37,111}},
/* 15016 */ {(16<<2)|3,{111,37,115}},
/* 15017 */ {(16<<2)|3,{111,37,116}},
/* 15018 */ {(11<<2)|2,{111,37,0}},
/* 15019 */ {(11<<2)|2,{111,37,0}},
/* 15020 */ {(11<<2)|2,{111,37,0}},
/* 15021 */ {(11<<2)|2,{111,37,0}},
/* 15022 */ {(11<<2)|2,{111,37,0}},
/* 15023 */ {(11<<2)|2,{111,37,0}},
/* 15024 */ {(11<<2)|2,{111,37,0}},
/* 15025 */ {(11<<2)|2,{111,37,0}},
/* 15026 */ {(11<<2)|2,{111,37,0}},
/* 15027 */ {(11<<2)|2,{111,37,0}},
/* 15028 */ {(11<<2)|2,{111,37,0}},
/* 15029 */ {(11<<2)|2,{111,37,0}},
/* 15030 */ {(11<<2)|2,{111,37,0}},
/* 15031 */ {(11<<2)|2,{111,37,0}},
/* 15032 */ {(11<<2)|2,{111,37,0}},
/* 15033 */ {(11<<2)|2,{111,37,0}},
/* 15034 */ {(11<<2)|2,{111,37,0}},
/* 15035 */ {(11<<2)|2,{111,37,0}},
/* 15036 */ {(11<<2)|2,{111,37,0}},
/* 15037 */ {(11<<2)|2,{111,37,0}},
/* 15038 */ {(11<<2)|2,{111,37,0}},
/* 15039 */ {(11<<2)|2,{111,37,0}},
/* 15040 */ {(16<<2)|3,{111,45,48}},
/* 15041 */ {(16<<2)|3,{111,45,49}},
/* 15042 */ {(16<<2)|3,{111,45,50}},
/* 15043 */ {(16<<2)|3,{111,45,97}},
/* 15044 */ {(16<<2)|3,{111,45,99}},
/* 15045 */ {(16<<2)|3,{111,45,101}},
/* 15046 */ {(16<<2)|3,{111,45,105}},
/* 15047 */ {(16<<2)|3,{111,45,111}},
/* 15048 */ {(16<<2)|3,{111,45,115}},
/* 15049 */ {(16<<2)|3,{111,45,116}},
/* 15050 */ {(11<<2)|2,{111,45,0}},
/* 15051 */ {(11<<2)|2,{111,45,0}},
/* 15052 */ {(11<<2)|2,{111,45,0}},
/* 15053 */ {(11<<2)|2,{111,45,0}},
/* 15054 */ {(11<<2)|2,{111,45,0}},
/* 15055 */ {(11<<2)|2,{111,45,0}},
/* 15056 */ {(11<<2)|2,{111,45,0}},
/* 15057 */ {(11<<2)|2,{111,45,0}},
/* 15058 */ {(11<<2)|2,{111,45,0}},
/* 15059 */ {(11<<2)|2,{111,45,0}},
/* 15060 */ {(11<<2)|2,{111,45,0}},
/* 15061 */ {(11<<2)|2,{111,45,0}},
/* 15062 */ {(11<<2)|2,{111,45,0}},
/* 15063 */ {(11<<2)|2,{111,45,0}},
/* 15064 */ {(11<<2)|2,{111,45,0}},
/* 15065 */ {(11<<2)|2,{111,45,0}},
/* 15066 */ {(11<<2)|2,{111,45,0}},
/* 15067 */ {(11<<2)|2,{111,45,0}},
/* 15068 */ {(11<<2)|2,{111,45,0}},
/* 15069 */ {(11<<2)|2,{111,45,0}},
/* 15070 */ {(11<<2)|2,{111,45,0}},
/* 15071 */ {(11<<2)|2,{111,45,0}},
/* 15072 */ {(16<<2)|3,{111,46,48}},
/* 15073 */ {(16<<2)|3,{111,46,49}},
/* 15074 */ {(16<<2)|3,{111,46,50}},
/* 15075 */ {(16<<2)|3,{111,46,97}},
/* 15076 */ {(16<<2)|3,{111,46,99}},
/* 15077 */ {(16<<2)|3,{111,46,101}},
/* 15078 */ {(16<<2)|3,{111,46,105}},
/* 15079 */ {(16<<2)|3,{111,46,111}},
/* 15080 */ {(16<<2)|3,{111,46,115}},
/* 15081 */ {(16<<2)|3,{111,46,116}},
/* 15082 */ {(11<<2)|2,{111,46,0}},
/* 15083 */ {(11<<2)|2,{111,46,0}},
/* 15084 */ {(11<<2)|2,{111,46,0}},
/* 15085 */ {(11<<2)|2,{111,46,0}},
/* 15086 */ {(11<<2)|2,{111,46,0}},
/* 15087 */ {(11<<2)|2,{111,46,0}},
/* 15088 */ {(11<<2)|2,{111,46,0}},
/* 15089 */ {(11<<2)|2,{111,46,0}},
/* 15090 */ {(11<<2)|2,{111,46,0}},
/* 15091 */ {(11<<2)|2,{111,46,0}},
/* 15092 */ {(11<<2)|2,{111,46,0}},
/* 15093 */ {(11<<2)|2,{111,46,0}},
/* 15094 */ {(11<<2)|2,{111,46,0}},
/* 15095 */ {(11<<2)|2,{111,46,0}},
/* 15096 */ {(11<<2)|2,{111,46,0}},
/* 15097 */ {(11<<2)|2,{111,46,0}},
/* 15098 */ {(11<<2)|2,{111,46,0}},
/* 15099 */ {(11<<2)|2,{111,46,0}},
/* 15100 */ {(11<<2)|2,{111,46,0}},
/* 15101 */ {(11<<2)|2,{111,46,0}},
/* 15102 */ {(11<<2)|2,{111,46,0}},
/* 15103 */ {(11<<2)|2,{111,46,0}},
/* 15104 */ {(16<<2)|3,{111,47,48}},
/* 15105 */ {(16<<2)|3,{111,47,49}},
/* 15106 */ {(16<<2)|3,{111,47,50}},
/* 15107 */ {(16<<2)|3,{111,47,97}},
/* 15108 */ {(16<<2)|3,{111,47,99}},
/* 15109 */ {(16<<2)|3,{111,47,101}},
/* 15110 */ {(16<<2)|3,{111,47,105}},
/* 15111 */ {(16<<2)|3,{111,47,111}},
/* 15112 */ {(16<<2)|3,{111,47,115}},
/* 15113 */ {(16<<2)|3,{111,47,116}},
/* 15114 */ {(11<<2)|2,{111,47,0}},
/* 15115 */ {(11<<2)|2,{111,47,0}},
/* 15116 */ {(11<<2)|2,{111,47,0}},
/* 15117 */ {(11<<2)|2,{111,47,0}},
/* 15118 */ {(11<<2)|2,{111,47,0}},
/* 15119 */ {(11<<2)|2,{111,47,0}},
/* 15120 */ {(11<<2)|2,{111,47,0}},
/* 15121 */ {(11<<2)|2,{111,47,0}},
/* 15122 */ {(11<<2)|2,{111,47,0}},
/* 15123 */ {(11<<2)|2,{111,47,0}},
/* 15124 */ {(11<<2)|2,{111,47,0}},
/* 15125 */ {(11<<2)|2,{111,47,0}},
/* 15126 */ {(11<<2)|2,{111,47,0}},
/* 15127 */ {(11<<2)|2,{111,47,0}},
/* 15128 */ {(11<<2)|2,{111,47,0}},
/* 15129 */ {(11<<2)|2,{111,47,0}},
/* 15130 */ {(11<<2)|2,{111,47,0}},
/* 15131 */ {(11<<2)|2,{111,47,0}},
/* 15132 */ {(11<<2)|2,{111,47,0}},
/* 15133 */ {(11<<2)|2,{111,47,0}},
/* 15134 */ {(11<<2)|2,{111,47,0}},
/* 15135 */ {(11<<2)|2,{111,47,0}},
/* 15136 */ {(16<<2)|3,{111,51,48}},
/* 15137 */ {(16<<2)|3,{111,51,49}},
/* 15138 */ {(16<<2)|3,{111,51,50}},
/* 15139 */ {(16<<2)|3,{111,51,97}},
/* 15140 */ {(16<<2)|3,{111,51,99}},
/* 15141 */ {(16<<2)|3,{111,51,101}},
/* 15142 */ {(16<<2)|3,{111,51,105}},
/* 15143 */ {(16<<2)|3,{111,51,111}},
/* 15144 */ {(16<<2)|3,{111,51,115}},
/* 15145 */ {(16<<2)|3,{111,51,116}},
/* 15146 */ {(11<<2)|2,{111,51,0}},
/* 15147 */ {(11<<2)|2,{111,51,0}},
/* 15148 */ {(11<<2)|2,{111,51,0}},
/* 15149 */ {(11<<2)|2,{111,51,0}},
/* 15150 */ {(11<<2)|2,{111,51,0}},
/* 15151 */ {(11<<2)|2,{111,51,0}},
/* 15152 */ {(11<<2)|2,{111,51,0}},
/* 15153 */ {(11<<2)|2,{111,51,0}},
/* 15154 */ {(11<<2)|2,{111,51,0}},
/* 15155 */ {(11<<2)|2,{111,51,0}},
/* 15156 */ {(11<<2)|2,{111,51,0}},
/* 15157 */ {(11<<2)|2,{111,51,0}},
/* 15158 */ {(11<<2)|2,{111,51,0}},
/* 15159 */ {(11<<2)|2,{111,51,0}},
/* 15160 */ {(11<<2)|2,{111,51,0}},
/* 15161 */ {(11<<2)|2,{111,51,0}},
/* 15162 */ {(11<<2)|2,{111,51,0}},
/* 15163 */ {(11<<2)|2,{111,51,0}},
/* 15164 */ {(11<<2)|2,{111,51,0}},
/* 15165 */ {(11<<2)|2,{111,51,0}},
/* 15166 */ {(11<<2)|2,{111,51,0}},
/* 15167 */ {(11<<2)|2,{111,51,0}},
/* 15168 */ {(16<<2)|3,{111,52,48}},
/* 15169 */ {(16<<2)|3,{111,52,49}},
/* 15170 */ {(16<<2)|3,{111,52,50}},
/* 15171 */ {(16<<2)|3,{111,52,97}},
/* 15172 */ {(16<<2)|3,{111,52,99}},
/* 15173 */ {(16<<2)|3,{111,52,101}},
/* 15174 */ {(16<<2)|3,{111,52,105}},
/* 15175 */ {(16<<2)|3,{111,52,111}},
/* 15176 */ {(16<<2)|3,{111,52,115}},
/* 15177 */ {(16<<2)|3,{111,52,116}},
/* 15178 */ {(11<<2)|2,{111,52,0}},
/* 15179 */ {(11<<2)|2,{111,52,0}},
/* 15180 */ {(11<<2)|2,{111,52,0}},
/* 15181 */ {(11<<2)|2,{111,52,0}},
/* 15182 */ {(11<<2)|2,{111,52,0}},
/* 15183 */ {(11<<2)|2,{111,52,0}},
/* 15184 */ {(11<<2)|2,{111,52,0}},
/* 15185 */ {(11<<2)|2,{111,52,0}},
/* 15186 */ {(11<<2)|2,{111,52,0}},
/* 15187 */ {(11<<2)|2,{111,52,0}},
/* 15188 */ {(11<<2)|2,{111,52,0}},
/* 15189 */ {(11<<2)|2,{111,52,0}},
/* 15190 */ {(11<<2)|2,{111,52,0}},
/* 15191 */ {(11<<2)|2,{111,52,0}},
/* 15192 */ {(11<<2)|2,{111,52,0}},
/* 15193 */ {(11<<2)|2,{111,52,0}},
/* 15194 */ {(11<<2)|2,{111,52,0}},
/* 15195 */ {(11<<2)|2,{111,52,0}},
/* 15196 */ {(11<<2)|2,{111,52,0}},
/* 15197 */ {(11<<2)|2,{111,52,0}},
/* 15198 */ {(11<<2)|2,{111,52,0}},
/* 15199 */ {(11<<2)|2,{111,52,0}},
/* 15200 */ {(16<<2)|3,{111,53,48}},
/* 15201 */ {(16<<2)|3,{111,53,49}},
/* 15202 */ {(16<<2)|3,{111,53,50}},
/* 15203 */ {(16<<2)|3,{111,53,97}},
/* 15204 */ {(16<<2)|3,{111,53,99}},
/* 15205 */ {(16<<2)|3,{111,53,101}},
/* 15206 */ {(16<<2)|3,{111,53,105}},
/* 15207 */ {(16<<2)|3,{111,53,111}},
/* 15208 */ {(16<<2)|3,{111,53,115}},
/* 15209 */ {(16<<2)|3,{111,53,116}},
/* 15210 */ {(11<<2)|2,{111,53,0}},
/* 15211 */ {(11<<2)|2,{111,53,0}},
/* 15212 */ {(11<<2)|2,{111,53,0}},
/* 15213 */ {(11<<2)|2,{111,53,0}},
/* 15214 */ {(11<<2)|2,{111,53,0}},
/* 15215 */ {(11<<2)|2,{111,53,0}},
/* 15216 */ {(11<<2)|2,{111,53,0}},
/* 15217 */ {(11<<2)|2,{111,53,0}},
/* 15218 */ {(11<<2)|2,{111,53,0}},
/* 15219 */ {(11<<2)|2,{111,53,0}},
/* 15220 */ {(11<<2)|2,{111,53,0}},
/* 15221 */ {(11<<2)|2,{111,53,0}},
/* 15222 */ {(11<<2)|2,{111,53,0}},
/* 15223 */ {(11<<2)|2,{111,53,0}},
/* 15224 */ {(11<<2)|2,{111,53,0}},
/* 15225 */ {(11<<2)|2,{111,53,0}},
/* 15226 */ {(11<<2)|2,{111,53,0}},
/* 15227 */ {(11<<2)|2,{111,53,0}},
/* 15228 */ {(11<<2)|2,{111,53,0}},
/* 15229 */ {(11<<2)|2,{111,53,0}},
/* 15230 */ {(11<<2)|2,{111,53,0}},
/* 15231 */ {(11<<2)|2,{111,53,0}},
/* 15232 */ {(16<<2)|3,{111,54,48}},
/* 15233 */ {(16<<2)|3,{111,54,49}},
/* 15234 */ {(16<<2)|3,{111,54,50}},
/* 15235 */ {(16<<2)|3,{111,54,97}},
/* 15236 */ {(16<<2)|3,{111,54,99}},
/* 15237 */ {(16<<2)|3,{111,54,101}},
/* 15238 */ {(16<<2)|3,{111,54,105}},
/* 15239 */ {(16<<2)|3,{111,54,111}},
/* 15240 */ {(16<<2)|3,{111,54,115}},
/* 15241 */ {(16<<2)|3,{111,54,116}},
/* 15242 */ {(11<<2)|2,{111,54,0}},
/* 15243 */ {(11<<2)|2,{111,54,0}},
/* 15244 */ {(11<<2)|2,{111,54,0}},
/* 15245 */ {(11<<2)|2,{111,54,0}},
/* 15246 */ {(11<<2)|2,{111,54,0}},
/* 15247 */ {(11<<2)|2,{111,54,0}},
/* 15248 */ {(11<<2)|2,{111,54,0}},
/* 15249 */ {(11<<2)|2,{111,54,0}},
/* 15250 */ {(11<<2)|2,{111,54,0}},
/* 15251 */ {(11<<2)|2,{111,54,0}},
/* 15252 */ {(11<<2)|2,{111,54,0}},
/* 15253 */ {(11<<2)|2,{111,54,0}},
/* 15254 */ {(11<<2)|2,{111,54,0}},
/* 15255 */ {(11<<2)|2,{111,54,0}},
/* 15256 */ {(11<<2)|2,{111,54,0}},
/* 15257 */ {(11<<2)|2,{111,54,0}},
/* 15258 */ {(11<<2)|2,{111,54,0}},
/* 15259 */ {(11<<2)|2,{111,54,0}},
/* 15260 */ {(11<<2)|2,{111,54,0}},
/* 15261 */ {(11<<2)|2,{111,54,0}},
/* 15262 */ {(11<<2)|2,{111,54,0}},
/* 15263 */ {(11<<2)|2,{111,54,0}},
/* 15264 */ {(16<<2)|3,{111,55,48}},
/* 15265 */ {(16<<2)|3,{111,55,49}},
/* 15266 */ {(16<<2)|3,{111,55,50}},
/* 15267 */ {(16<<2)|3,{111,55,97}},
/* 15268 */ {(16<<2)|3,{111,55,99}},
/* 15269 */ {(16<<2)|3,{111,55,101}},
/* 15270 */ {(16<<2)|3,{111,55,105}},
/* 15271 */ {(16<<2)|3,{111,55,111}},
/* 15272 */ {(16<<2)|3,{111,55,115}},
/* 15273 */ {(16<<2)|3,{111,55,116}},
/* 15274 */ {(11<<2)|2,{111,55,0}},
/* 15275 */ {(11<<2)|2,{111,55,0}},
/* 15276 */ {(11<<2)|2,{111,55,0}},
/* 15277 */ {(11<<2)|2,{111,55,0}},
/* 15278 */ {(11<<2)|2,{111,55,0}},
/* 15279 */ {(11<<2)|2,{111,55,0}},
/* 15280 */ {(11<<2)|2,{111,55,0}},
/* 15281 */ {(11<<2)|2,{111,55,0}},
/* 15282 */ {(11<<2)|2,{111,55,0}},
/* 15283 */ {(11<<2)|2,{111,55,0}},
/* 15284 */ {(11<<2)|2,{111,55,0}},
/* 15285 */ {(11<<2)|2,{111,55,0}},
/* 15286 */ {(11<<2)|2,{111,55,0}},
/* 15287 */ {(11<<2)|2,{111,55,0}},
/* 15288 */ {(11<<2)|2,{111,55,0}},
/* 15289 */ {(11<<2)|2,{111,55,0}},
/* 15290 */ {(11<<2)|2,{111,55,0}},
/* 15291 */ {(11<<2)|2,{111,55,0}},
/* 15292 */ {(11<<2)|2,{111,55,0}},
/* 15293 */ {(11<<2)|2,{111,55,0}},
/* 15294 */ {(11<<2)|2,{111,55,0}},
/* 15295 */ {(11<<2)|2,{111,55,0}},
/* 15296 */ {(16<<2)|3,{111,56,48}},
/* 15297 */ {(16<<2)|3,{111,56,49}},
/* 15298 */ {(16<<2)|3,{111,56,50}},
/* 15299 */ {(16<<2)|3,{111,56,97}},
/* 15300 */ {(16<<2)|3,{111,56,99}},
/* 15301 */ {(16<<2)|3,{111,56,101}},
/* 15302 */ {(16<<2)|3,{111,56,105}},
/* 15303 */ {(16<<2)|3,{111,56,111}},
/* 15304 */ {(16<<2)|3,{111,56,115}},
/* 15305 */ {(16<<2)|3,{111,56,116}},
/* 15306 */ {(11<<2)|2,{111,56,0}},
/* 15307 */ {(11<<2)|2,{111,56,0}},
/* 15308 */ {(11<<2)|2,{111,56,0}},
/* 15309 */ {(11<<2)|2,{111,56,0}},
/* 15310 */ {(11<<2)|2,{111,56,0}},
/* 15311 */ {(11<<2)|2,{111,56,0}},
/* 15312 */ {(11<<2)|2,{111,56,0}},
/* 15313 */ {(11<<2)|2,{111,56,0}},
/* 15314 */ {(11<<2)|2,{111,56,0}},
/* 15315 */ {(11<<2)|2,{111,56,0}},
/* 15316 */ {(11<<2)|2,{111,56,0}},
/* 15317 */ {(11<<2)|2,{111,56,0}},
/* 15318 */ {(11<<2)|2,{111,56,0}},
/* 15319 */ {(11<<2)|2,{111,56,0}},
/* 15320 */ {(11<<2)|2,{111,56,0}},
/* 15321 */ {(11<<2)|2,{111,56,0}},
/* 15322 */ {(11<<2)|2,{111,56,0}},
/* 15323 */ {(11<<2)|2,{111,56,0}},
/* 15324 */ {(11<<2)|2,{111,56,0}},
/* 15325 */ {(11<<2)|2,{111,56,0}},
/* 15326 */ {(11<<2)|2,{111,56,0}},
/* 15327 */ {(11<<2)|2,{111,56,0}},
/* 15328 */ {(16<<2)|3,{111,57,48}},
/* 15329 */ {(16<<2)|3,{111,57,49}},
/* 15330 */ {(16<<2)|3,{111,57,50}},
/* 15331 */ {(16<<2)|3,{111,57,97}},
/* 15332 */ {(16<<2)|3,{111,57,99}},
/* 15333 */ {(16<<2)|3,{111,57,101}},
/* 15334 */ {(16<<2)|3,{111,57,105}},
/* 15335 */ {(16<<2)|3,{111,57,111}},
/* 15336 */ {(16<<2)|3,{111,57,115}},
/* 15337 */ {(16<<2)|3,{111,57,116}},
/* 15338 */ {(11<<2)|2,{111,57,0}},
/* 15339 */ {(11<<2)|2,{111,57,0}},
/* 15340 */ {(11<<2)|2,{111,57,0}},
/* 15341 */ {(11<<2)|2,{111,57,0}},
/* 15342 */ {(11<<2)|2,{111,57,0}},
/* 15343 */ {(11<<2)|2,{111,57,0}},
/* 15344 */ {(11<<2)|2,{111,57,0}},
/* 15345 */ {(11<<2)|2,{111,57,0}},
/* 15346 */ {(11<<2)|2,{111,57,0}},
/* 15347 */ {(11<<2)|2,{111,57,0}},
/* 15348 */ {(11<<2)|2,{111,57,0}},
/* 15349 */ {(11<<2)|2,{111,57,0}},
/* 15350 */ {(11<<2)|2,{111,57,0}},
/* 15351 */ {(11<<2)|2,{111,57,0}},
/* 15352 */ {(11<<2)|2,{111,57,0}},
/* 15353 */ {(11<<2)|2,{111,57,0}},
/* 15354 */ {(11<<2)|2,{111,57,0}},
/* 15355 */ {(11<<2)|2,{111,57,0}},
/* 15356 */ {(11<<2)|2,{111,57,0}},
/* 15357 */ {(11<<2)|2,{111,57,0}},
/* 15358 */ {(11<<2)|2,{111,57,0}},
/* 15359 */ {(11<<2)|2,{111,57,0}},
/* 15360 */ {(16<<2)|3,{111,61,48}},
/* 15361 */ {(16<<2)|3,{111,61,49}},
/* 15362 */ {(16<<2)|3,{111,61,50}},
/* 15363 */ {(16<<2)|3,{111,61,97}},
/* 15364 */ {(16<<2)|3,{111,61,99}},
/* 15365 */ {(16<<2)|3,{111,61,101}},
/* 15366 */ {(16<<2)|3,{111,61,105}},
/* 15367 */ {(16<<2)|3,{111,61,111}},
/* 15368 */ {(16<<2)|3,{111,61,115}},
/* 15369 */ {(16<<2)|3,{111,61,116}},
/* 15370 */ {(11<<2)|2,{111,61,0}},
/* 15371 */ {(11<<2)|2,{111,61,0}},
/* 15372 */ {(11<<2)|2,{111,61,0}},
/* 15373 */ {(11<<2)|2,{111,61,0}},
/* 15374 */ {(11<<2)|2,{111,61,0}},
/* 15375 */ {(11<<2)|2,{111,61,0}},
/* 15376 */ {(11<<2)|2,{111,61,0}},
/* 15377 */ {(11<<2)|2,{111,61,0}},
/* 15378 */ {(11<<2)|2,{111,61,0}},
/* 15379 */ {(11<<2)|2,{111,61,0}},
/* 15380 */ {(11<<2)|2,{111,61,0}},
/* 15381 */ {(11<<2)|2,{111,61,0}},
/* 15382 */ {(11<<2)|2,{111,61,0}},
/* 15383 */ {(11<<2)|2,{111,61,0}},
/* 15384 */ {(11<<2)|2,{111,61,0}},
/* 15385 */ {(11<<2)|2,{111,61,0}},
/* 15386 */ {(11<<2)|2,{111,61,0}},
/* 15387 */ {(11<<2)|2,{111,61,0}},
/* 15388 */ {(11<<2)|2,{111,61,0}},
/* 15389 */ {(11<<2)|2,{111,61,0}},
/* 15390 */ {(11<<2)|2,{111,61,0}},
/* 15391 */ {(11<<2)|2,{111,61,0}},
/* 15392 */ {(16<<2)|3,{111,65,48}},
/* 15393 */ {(16<<2)|3,{111,65,49}},
/* 15394 */ {(16<<2)|3,{111,65,50}},
/* 15395 */ {(16<<2)|3,{111,65,97}},
/* 15396 */ {(16<<2)|3,{111,65,99}},
/* 15397 */ {(16<<2)|3,{111,65,101}},
/* 15398 */ {(16<<2)|3,{111,65,105}},
/* 15399 */ {(16<<2)|3,{111,65,111}},
/* 15400 */ {(16<<2)|3,{111,65,115}},
/* 15401 */ {(16<<2)|3,{111,65,116}},
/* 15402 */ {(11<<2)|2,{111,65,0}},
/* 15403 */ {(11<<2)|2,{111,65,0}},
/* 15404 */ {(11<<2)|2,{111,65,0}},
/* 15405 */ {(11<<2)|2,{111,65,0}},
/* 15406 */ {(11<<2)|2,{111,65,0}},
/* 15407 */ {(11<<2)|2,{111,65,0}},
/* 15408 */ {(11<<2)|2,{111,65,0}},
/* 15409 */ {(11<<2)|2,{111,65,0}},
/* 15410 */ {(11<<2)|2,{111,65,0}},
/* 15411 */ {(11<<2)|2,{111,65,0}},
/* 15412 */ {(11<<2)|2,{111,65,0}},
/* 15413 */ {(11<<2)|2,{111,65,0}},
/* 15414 */ {(11<<2)|2,{111,65,0}},
/* 15415 */ {(11<<2)|2,{111,65,0}},
/* 15416 */ {(11<<2)|2,{111,65,0}},
/* 15417 */ {(11<<2)|2,{111,65,0}},
/* 15418 */ {(11<<2)|2,{111,65,0}},
/* 15419 */ {(11<<2)|2,{111,65,0}},
/* 15420 */ {(11<<2)|2,{111,65,0}},
/* 15421 */ {(11<<2)|2,{111,65,0}},
/* 15422 */ {(11<<2)|2,{111,65,0}},
/* 15423 */ {(11<<2)|2,{111,65,0}},
/* 15424 */ {(16<<2)|3,{111,95,48}},
/* 15425 */ {(16<<2)|3,{111,95,49}},
/* 15426 */ {(16<<2)|3,{111,95,50}},
/* 15427 */ {(16<<2)|3,{111,95,97}},
/* 15428 */ {(16<<2)|3,{111,95,99}},
/* 15429 */ {(16<<2)|3,{111,95,101}},
/* 15430 */ {(16<<2)|3,{111,95,105}},
/* 15431 */ {(16<<2)|3,{111,95,111}},
/* 15432 */ {(16<<2)|3,{111,95,115}},
/* 15433 */ {(16<<2)|3,{111,95,116}},
/* 15434 */ {(11<<2)|2,{111,95,0}},
/* 15435 */ {(11<<2)|2,{111,95,0}},
/* 15436 */ {(11<<2)|2,{111,95,0}},
/* 15437 */ {(11<<2)|2,{111,95,0}},
/* 15438 */ {(11<<2)|2,{111,95,0}},
/* 15439 */ {(11<<2)|2,{111,95,0}},
/* 15440 */ {(11<<2)|2,{111,95,0}},
/* 15441 */ {(11<<2)|2,{111,95,0}},
/* 15442 */ {(11<<2)|2,{111,95,0}},
/* 15443 */ {(11<<2)|2,{111,95,0}},
/* 15444 */ {(11<<2)|2,{111,95,0}},
/* 15445 */ {(11<<2)|2,{111,95,0}},
/* 15446 */ {(11<<2)|2,{111,95,0}},
/* 15447 */ {(11<<2)|2,{111,95,0}},
/* 15448 */ {(11<<2)|2,{111,95,0}},
/* 15449 */ {(11<<2)|2,{111,95,0}},
/* 15450 */ {(11<<2)|2,{111,95,0}},
/* 15451 */ {(11<<2)|2,{111,95,0}},
/* 15452 */ {(11<<2)|2,{111,95,0}},
/* 15453 */ {(11<<2)|2,{111,95,0}},
/* 15454 */ {(11<<2)|2,{111,95,0}},
/* 15455 */ {(11<<2)|2,{111,95,0}},
/* 15456 */ {(16<<2)|3,{111,98,48}},
/* 15457 */ {(16<<2)|3,{111,98,49}},
/* 15458 */ {(16<<2)|3,{111,98,50}},
/* 15459 */ {(16<<2)|3,{111,98,97}},
/* 15460 */ {(16<<2)|3,{111,98,99}},
/* 15461 */ {(16<<2)|3,{111,98,101}},
/* 15462 */ {(16<<2)|3,{111,98,105}},
/* 15463 */ {(16<<2)|3,{111,98,111}},
/* 15464 */ {(16<<2)|3,{111,98,115}},
/* 15465 */ {(16<<2)|3,{111,98,116}},
/* 15466 */ {(11<<2)|2,{111,98,0}},
/* 15467 */ {(11<<2)|2,{111,98,0}},
/* 15468 */ {(11<<2)|2,{111,98,0}},
/* 15469 */ {(11<<2)|2,{111,98,0}},
/* 15470 */ {(11<<2)|2,{111,98,0}},
/* 15471 */ {(11<<2)|2,{111,98,0}},
/* 15472 */ {(11<<2)|2,{111,98,0}},
/* 15473 */ {(11<<2)|2,{111,98,0}},
/* 15474 */ {(11<<2)|2,{111,98,0}},
/* 15475 */ {(11<<2)|2,{111,98,0}},
/* 15476 */ {(11<<2)|2,{111,98,0}},
/* 15477 */ {(11<<2)|2,{111,98,0}},
/* 15478 */ {(11<<2)|2,{111,98,0}},
/* 15479 */ {(11<<2)|2,{111,98,0}},
/* 15480 */ {(11<<2)|2,{111,98,0}},
/* 15481 */ {(11<<2)|2,{111,98,0}},
/* 15482 */ {(11<<2)|2,{111,98,0}},
/* 15483 */ {(11<<2)|2,{111,98,0}},
/* 15484 */ {(11<<2)|2,{111,98,0}},
/* 15485 */ {(11<<2)|2,{111,98,0}},
/* 15486 */ {(11<<2)|2,{111,98,0}},
/* 15487 */ {(11<<2)|2,{111,98,0}},
/* 15488 */ {(16<<2)|3,{111,100,48}},
/* 15489 */ {(16<<2)|3,{111,100,49}},
/* 15490 */ {(16<<2)|3,{111,100,50}},
/* 15491 */ {(16<<2)|3,{111,100,97}},
/* 15492 */ {(16<<2)|3,{111,100,99}},
/* 15493 */ {(16<<2)|3,{111,100,101}},
/* 15494 */ {(16<<2)|3,{111,100,105}},
/* 15495 */ {(16<<2)|3,{111,100,111}},
/* 15496 */ {(16<<2)|3,{111,100,115}},
/* 15497 */ {(16<<2)|3,{111,100,116}},
/* 15498 */ {(11<<2)|2,{111,100,0}},
/* 15499 */ {(11<<2)|2,{111,100,0}},
/* 15500 */ {(11<<2)|2,{111,100,0}},
/* 15501 */ {(11<<2)|2,{111,100,0}},
/* 15502 */ {(11<<2)|2,{111,100,0}},
/* 15503 */ {(11<<2)|2,{111,100,0}},
/* 15504 */ {(11<<2)|2,{111,100,0}},
/* 15505 */ {(11<<2)|2,{111,100,0}},
/* 15506 */ {(11<<2)|2,{111,100,0}},
/* 15507 */ {(11<<2)|2,{111,100,0}},
/* 15508 */ {(11<<2)|2,{111,100,0}},
/* 15509 */ {(11<<2)|2,{111,100,0}},
/* 15510 */ {(11<<2)|2,{111,100,0}},
/* 15511 */ {(11<<2)|2,{111,100,0}},
/* 15512 */ {(11<<2)|2,{111,100,0}},
/* 15513 */ {(11<<2)|2,{111,100,0}},
/* 15514 */ {(11<<2)|2,{111,100,0}},
/* 15515 */ {(11<<2)|2,{111,100,0}},
/* 15516 */ {(11<<2)|2,{111,100,0}},
/* 15517 */ {(11<<2)|2,{111,100,0}},
/* 15518 */ {(11<<2)|2,{111,100,0}},
/* 15519 */ {(11<<2)|2,{111,100,0}},
/* 15520 */ {(16<<2)|3,{111,102,48}},
/* 15521 */ {(16<<2)|3,{111,102,49}},
/* 15522 */ {(16<<2)|3,{111,102,50}},
/* 15523 */ {(16<<2)|3,{111,102,97}},
/* 15524 */ {(16<<2)|3,{111,102,99}},
/* 15525 */ {(16<<2)|3,{111,102,101}},
/* 15526 */ {(16<<2)|3,{111,102,105}},
/* 15527 */ {(16<<2)|3,{111,102,111}},
/* 15528 */ {(16<<2)|3,{111,102,115}},
/* 15529 */ {(16<<2)|3,{111,102,116}},
/* 15530 */ {(11<<2)|2,{111,102,0}},
/* 15531 */ {(11<<2)|2,{111,102,0}},
/* 15532 */ {(11<<2)|2,{111,102,0}},
/* 15533 */ {(11<<2)|2,{111,102,0}},
/* 15534 */ {(11<<2)|2,{111,102,0}},
/* 15535 */ {(11<<2)|2,{111,102,0}},
/* 15536 */ {(11<<2)|2,{111,102,0}},
/* 15537 */ {(11<<2)|2,{111,102,0}},
/* 15538 */ {(11<<2)|2,{111,102,0}},
/* 15539 */ {(11<<2)|2,{111,102,0}},
/* 15540 */ {(11<<2)|2,{111,102,0}},
/* 15541 */ {(11<<2)|2,{111,102,0}},
/* 15542 */ {(11<<2)|2,{111,102,0}},
/* 15543 */ {(11<<2)|2,{111,102,0}},
/* 15544 */ {(11<<2)|2,{111,102,0}},
/* 15545 */ {(11<<2)|2,{111,102,0}},
/* 15546 */ {(11<<2)|2,{111,102,0}},
/* 15547 */ {(11<<2)|2,{111,102,0}},
/* 15548 */ {(11<<2)|2,{111,102,0}},
/* 15549 */ {(11<<2)|2,{111,102,0}},
/* 15550 */ {(11<<2)|2,{111,102,0}},
/* 15551 */ {(11<<2)|2,{111,102,0}},
/* 15552 */ {(16<<2)|3,{111,103,48}},
/* 15553 */ {(16<<2)|3,{111,103,49}},
/* 15554 */ {(16<<2)|3,{111,103,50}},
/* 15555 */ {(16<<2)|3,{111,103,97}},
/* 15556 */ {(16<<2)|3,{111,103,99}},
/* 15557 */ {(16<<2)|3,{111,103,101}},
/* 15558 */ {(16<<2)|3,{111,103,105}},
/* 15559 */ {(16<<2)|3,{111,103,111}},
/* 15560 */ {(16<<2)|3,{111,103,115}},
/* 15561 */ {(16<<2)|3,{111,103,116}},
/* 15562 */ {(11<<2)|2,{111,103,0}},
/* 15563 */ {(11<<2)|2,{111,103,0}},
/* 15564 */ {(11<<2)|2,{111,103,0}},
/* 15565 */ {(11<<2)|2,{111,103,0}},
/* 15566 */ {(11<<2)|2,{111,103,0}},
/* 15567 */ {(11<<2)|2,{111,103,0}},
/* 15568 */ {(11<<2)|2,{111,103,0}},
/* 15569 */ {(11<<2)|2,{111,103,0}},
/* 15570 */ {(11<<2)|2,{111,103,0}},
/* 15571 */ {(11<<2)|2,{111,103,0}},
/* 15572 */ {(11<<2)|2,{111,103,0}},
/* 15573 */ {(11<<2)|2,{111,103,0}},
/* 15574 */ {(11<<2)|2,{111,103,0}},
/* 15575 */ {(11<<2)|2,{111,103,0}},
/* 15576 */ {(11<<2)|2,{111,103,0}},
/* 15577 */ {(11<<2)|2,{111,103,0}},
/* 15578 */ {(11<<2)|2,{111,103,0}},
/* 15579 */ {(11<<2)|2,{111,103,0}},
/* 15580 */ {(11<<2)|2,{111,103,0}},
/* 15581 */ {(11<<2)|2,{111,103,0}},
/* 15582 */ {(11<<2)|2,{111,103,0}},
/* 15583 */ {(11<<2)|2,{111,103,0}},
/* 15584 */ {(16<<2)|3,{111,104,48}},
/* 15585 */ {(16<<2)|3,{111,104,49}},
/* 15586 */ {(16<<2)|3,{111,104,50}},
/* 15587 */ {(16<<2)|3,{111,104,97}},
/* 15588 */ {(16<<2)|3,{111,104,99}},
/* 15589 */ {(16<<2)|3,{111,104,101}},
/* 15590 */ {(16<<2)|3,{111,104,105}},
/* 15591 */ {(16<<2)|3,{111,104,111}},
/* 15592 */ {(16<<2)|3,{111,104,115}},
/* 15593 */ {(16<<2)|3,{111,104,116}},
/* 15594 */ {(11<<2)|2,{111,104,0}},
/* 15595 */ {(11<<2)|2,{111,104,0}},
/* 15596 */ {(11<<2)|2,{111,104,0}},
/* 15597 */ {(11<<2)|2,{111,104,0}},
/* 15598 */ {(11<<2)|2,{111,104,0}},
/* 15599 */ {(11<<2)|2,{111,104,0}},
/* 15600 */ {(11<<2)|2,{111,104,0}},
/* 15601 */ {(11<<2)|2,{111,104,0}},
/* 15602 */ {(11<<2)|2,{111,104,0}},
/* 15603 */ {(11<<2)|2,{111,104,0}},
/* 15604 */ {(11<<2)|2,{111,104,0}},
/* 15605 */ {(11<<2)|2,{111,104,0}},
/* 15606 */ {(11<<2)|2,{111,104,0}},
/* 15607 */ {(11<<2)|2,{111,104,0}},
/* 15608 */ {(11<<2)|2,{111,104,0}},
/* 15609 */ {(11<<2)|2,{111,104,0}},
/* 15610 */ {(11<<2)|2,{111,104,0}},
/* 15611 */ {(11<<2)|2,{111,104,0}},
/* 15612 */ {(11<<2)|2,{111,104,0}},
/* 15613 */ {(11<<2)|2,{111,104,0}},
/* 15614 */ {(11<<2)|2,{111,104,0}},
/* 15615 */ {(11<<2)|2,{111,104,0}},
/* 15616 */ {(16<<2)|3,{111,108,48}},
/* 15617 */ {(16<<2)|3,{111,108,49}},
/* 15618 */ {(16<<2)|3,{111,108,50}},
/* 15619 */ {(16<<2)|3,{111,108,97}},
/* 15620 */ {(16<<2)|3,{111,108,99}},
/* 15621 */ {(16<<2)|3,{111,108,101}},
/* 15622 */ {(16<<2)|3,{111,108,105}},
/* 15623 */ {(16<<2)|3,{111,108,111}},
/* 15624 */ {(16<<2)|3,{111,108,115}},
/* 15625 */ {(16<<2)|3,{111,108,116}},
/* 15626 */ {(11<<2)|2,{111,108,0}},
/* 15627 */ {(11<<2)|2,{111,108,0}},
/* 15628 */ {(11<<2)|2,{111,108,0}},
/* 15629 */ {(11<<2)|2,{111,108,0}},
/* 15630 */ {(11<<2)|2,{111,108,0}},
/* 15631 */ {(11<<2)|2,{111,108,0}},
/* 15632 */ {(11<<2)|2,{111,108,0}},
/* 15633 */ {(11<<2)|2,{111,108,0}},
/* 15634 */ {(11<<2)|2,{111,108,0}},
/* 15635 */ {(11<<2)|2,{111,108,0}},
/* 15636 */ {(11<<2)|2,{111,108,0}},
/* 15637 */ {(11<<2)|2,{111,108,0}},
/* 15638 */ {(11<<2)|2,{111,108,0}},
/* 15639 */ {(11<<2)|2,{111,108,0}},
/* 15640 */ {(11<<2)|2,{111,108,0}},
/* 15641 */ {(11<<2)|2,{111,108,0}},
/* 15642 */ {(11<<2)|2,{111,108,0}},
/* 15643 */ {(11<<2)|2,{111,108,0}},
/* 15644 */ {(11<<2)|2,{111,108,0}},
/* 15645 */ {(11<<2)|2,{111,108,0}},
/* 15646 */ {(11<<2)|2,{111,108,0}},
/* 15647 */ {(11<<2)|2,{111,108,0}},
/* 15648 */ {(16<<2)|3,{111,109,48}},
/* 15649 */ {(16<<2)|3,{111,109,49}},
/* 15650 */ {(16<<2)|3,{111,109,50}},
/* 15651 */ {(16<<2)|3,{111,109,97}},
/* 15652 */ {(16<<2)|3,{111,109,99}},
/* 15653 */ {(16<<2)|3,{111,109,101}},
/* 15654 */ {(16<<2)|3,{111,109,105}},
/* 15655 */ {(16<<2)|3,{111,109,111}},
/* 15656 */ {(16<<2)|3,{111,109,115}},
/* 15657 */ {(16<<2)|3,{111,109,116}},
/* 15658 */ {(11<<2)|2,{111,109,0}},
/* 15659 */ {(11<<2)|2,{111,109,0}},
/* 15660 */ {(11<<2)|2,{111,109,0}},
/* 15661 */ {(11<<2)|2,{111,109,0}},
/* 15662 */ {(11<<2)|2,{111,109,0}},
/* 15663 */ {(11<<2)|2,{111,109,0}},
/* 15664 */ {(11<<2)|2,{111,109,0}},
/* 15665 */ {(11<<2)|2,{111,109,0}},
/* 15666 */ {(11<<2)|2,{111,109,0}},
/* 15667 */ {(11<<2)|2,{111,109,0}},
/* 15668 */ {(11<<2)|2,{111,109,0}},
/* 15669 */ {(11<<2)|2,{111,109,0}},
/* 15670 */ {(11<<2)|2,{111,109,0}},
/* 15671 */ {(11<<2)|2,{111,109,0}},
/* 15672 */ {(11<<2)|2,{111,109,0}},
/* 15673 */ {(11<<2)|2,{111,109,0}},
/* 15674 */ {(11<<2)|2,{111,109,0}},
/* 15675 */ {(11<<2)|2,{111,109,0}},
/* 15676 */ {(11<<2)|2,{111,109,0}},
/* 15677 */ {(11<<2)|2,{111,109,0}},
/* 15678 */ {(11<<2)|2,{111,109,0}},
/* 15679 */ {(11<<2)|2,{111,109,0}},
/* 15680 */ {(16<<2)|3,{111,110,48}},
/* 15681 */ {(16<<2)|3,{111,110,49}},
/* 15682 */ {(16<<2)|3,{111,110,50}},
/* 15683 */ {(16<<2)|3,{111,110,97}},
/* 15684 */ {(16<<2)|3,{111,110,99}},
/* 15685 */ {(16<<2)|3,{111,110,101}},
/* 15686 */ {(16<<2)|3,{111,110,105}},
/* 15687 */ {(16<<2)|3,{111,110,111}},
/* 15688 */ {(16<<2)|3,{111,110,115}},
/* 15689 */ {(16<<2)|3,{111,110,116}},
/* 15690 */ {(11<<2)|2,{111,110,0}},
/* 15691 */ {(11<<2)|2,{111,110,0}},
/* 15692 */ {(11<<2)|2,{111,110,0}},
/* 15693 */ {(11<<2)|2,{111,110,0}},
/* 15694 */ {(11<<2)|2,{111,110,0}},
/* 15695 */ {(11<<2)|2,{111,110,0}},
/* 15696 */ {(11<<2)|2,{111,110,0}},
/* 15697 */ {(11<<2)|2,{111,110,0}},
/* 15698 */ {(11<<2)|2,{111,110,0}},
/* 15699 */ {(11<<2)|2,{111,110,0}},
/* 15700 */ {(11<<2)|2,{111,110,0}},
/* 15701 */ {(11<<2)|2,{111,110,0}},
/* 15702 */ {(11<<2)|2,{111,110,0}},
/* 15703 */ {(11<<2)|2,{111,110,0}},
/* 15704 */ {(11<<2)|2,{111,110,0}},
/* 15705 */ {(11<<2)|2,{111,110,0}},
/* 15706 */ {(11<<2)|2,{111,110,0}},
/* 15707 */ {(11<<2)|2,{111,110,0}},
/* 15708 */ {(11<<2)|2,{111,110,0}},
/* 15709 */ {(11<<2)|2,{111,110,0}},
/* 15710 */ {(11<<2)|2,{111,110,0}},
/* 15711 */ {(11<<2)|2,{111,110,0}},
/* 15712 */ {(16<<2)|3,{111,112,48}},
/* 15713 */ {(16<<2)|3,{111,112,49}},
/* 15714 */ {(16<<2)|3,{111,112,50}},
/* 15715 */ {(16<<2)|3,{111,112,97}},
/* 15716 */ {(16<<2)|3,{111,112,99}},
/* 15717 */ {(16<<2)|3,{111,112,101}},
/* 15718 */ {(16<<2)|3,{111,112,105}},
/* 15719 */ {(16<<2)|3,{111,112,111}},
/* 15720 */ {(16<<2)|3,{111,112,115}},
/* 15721 */ {(16<<2)|3,{111,112,116}},
/* 15722 */ {(11<<2)|2,{111,112,0}},
/* 15723 */ {(11<<2)|2,{111,112,0}},
/* 15724 */ {(11<<2)|2,{111,112,0}},
/* 15725 */ {(11<<2)|2,{111,112,0}},
/* 15726 */ {(11<<2)|2,{111,112,0}},
/* 15727 */ {(11<<2)|2,{111,112,0}},
/* 15728 */ {(11<<2)|2,{111,112,0}},
/* 15729 */ {(11<<2)|2,{111,112,0}},
/* 15730 */ {(11<<2)|2,{111,112,0}},
/* 15731 */ {(11<<2)|2,{111,112,0}},
/* 15732 */ {(11<<2)|2,{111,112,0}},
/* 15733 */ {(11<<2)|2,{111,112,0}},
/* 15734 */ {(11<<2)|2,{111,112,0}},
/* 15735 */ {(11<<2)|2,{111,112,0}},
/* 15736 */ {(11<<2)|2,{111,112,0}},
/* 15737 */ {(11<<2)|2,{111,112,0}},
/* 15738 */ {(11<<2)|2,{111,112,0}},
/* 15739 */ {(11<<2)|2,{111,112,0}},
/* 15740 */ {(11<<2)|2,{111,112,0}},
/* 15741 */ {(11<<2)|2,{111,112,0}},
/* 15742 */ {(11<<2)|2,{111,112,0}},
/* 15743 */ {(11<<2)|2,{111,112,0}},
/* 15744 */ {(16<<2)|3,{111,114,48}},
/* 15745 */ {(16<<2)|3,{111,114,49}},
/* 15746 */ {(16<<2)|3,{111,114,50}},
/* 15747 */ {(16<<2)|3,{111,114,97}},
/* 15748 */ {(16<<2)|3,{111,114,99}},
/* 15749 */ {(16<<2)|3,{111,114,101}},
/* 15750 */ {(16<<2)|3,{111,114,105}},
/* 15751 */ {(16<<2)|3,{111,114,111}},
/* 15752 */ {(16<<2)|3,{111,114,115}},
/* 15753 */ {(16<<2)|3,{111,114,116}},
/* 15754 */ {(11<<2)|2,{111,114,0}},
/* 15755 */ {(11<<2)|2,{111,114,0}},
/* 15756 */ {(11<<2)|2,{111,114,0}},
/* 15757 */ {(11<<2)|2,{111,114,0}},
/* 15758 */ {(11<<2)|2,{111,114,0}},
/* 15759 */ {(11<<2)|2,{111,114,0}},
/* 15760 */ {(11<<2)|2,{111,114,0}},
/* 15761 */ {(11<<2)|2,{111,114,0}},
/* 15762 */ {(11<<2)|2,{111,114,0}},
/* 15763 */ {(11<<2)|2,{111,114,0}},
/* 15764 */ {(11<<2)|2,{111,114,0}},
/* 15765 */ {(11<<2)|2,{111,114,0}},
/* 15766 */ {(11<<2)|2,{111,114,0}},
/* 15767 */ {(11<<2)|2,{111,114,0}},
/* 15768 */ {(11<<2)|2,{111,114,0}},
/* 15769 */ {(11<<2)|2,{111,114,0}},
/* 15770 */ {(11<<2)|2,{111,114,0}},
/* 15771 */ {(11<<2)|2,{111,114,0}},
/* 15772 */ {(11<<2)|2,{111,114,0}},
/* 15773 */ {(11<<2)|2,{111,114,0}},
/* 15774 */ {(11<<2)|2,{111,114,0}},
/* 15775 */ {(11<<2)|2,{111,114,0}},
/* 15776 */ {(16<<2)|3,{111,117,48}},
/* 15777 */ {(16<<2)|3,{111,117,49}},
/* 15778 */ {(16<<2)|3,{111,117,50}},
/* 15779 */ {(16<<2)|3,{111,117,97}},
/* 15780 */ {(16<<2)|3,{111,117,99}},
/* 15781 */ {(16<<2)|3,{111,117,101}},
/* 15782 */ {(16<<2)|3,{111,117,105}},
/* 15783 */ {(16<<2)|3,{111,117,111}},
/* 15784 */ {(16<<2)|3,{111,117,115}},
/* 15785 */ {(16<<2)|3,{111,117,116}},
/* 15786 */ {(11<<2)|2,{111,117,0}},
/* 15787 */ {(11<<2)|2,{111,117,0}},
/* 15788 */ {(11<<2)|2,{111,117,0}},
/* 15789 */ {(11<<2)|2,{111,117,0}},
/* 15790 */ {(11<<2)|2,{111,117,0}},
/* 15791 */ {(11<<2)|2,{111,117,0}},
/* 15792 */ {(11<<2)|2,{111,117,0}},
/* 15793 */ {(11<<2)|2,{111,117,0}},
/* 15794 */ {(11<<2)|2,{111,117,0}},
/* 15795 */ {(11<<2)|2,{111,117,0}},
/* 15796 */ {(11<<2)|2,{111,117,0}},
/* 15797 */ {(11<<2)|2,{111,117,0}},
/* 15798 */ {(11<<2)|2,{111,117,0}},
/* 15799 */ {(11<<2)|2,{111,117,0}},
/* 15800 */ {(11<<2)|2,{111,117,0}},
/* 15801 */ {(11<<2)|2,{111,117,0}},
/* 15802 */ {(11<<2)|2,{111,117,0}},
/* 15803 */ {(11<<2)|2,{111,117,0}},
/* 15804 */ {(11<<2)|2,{111,117,0}},
/* 15805 */ {(11<<2)|2,{111,117,0}},
/* 15806 */ {(11<<2)|2,{111,117,0}},
/* 15807 */ {(11<<2)|2,{111,117,0}},
/* 15808 */ {(12<<2)|2,{111,58,0}},
/* 15809 */ {(12<<2)|2,{111,58,0}},
/* 15810 */ {(12<<2)|2,{111,58,0}},
/* 15811 */ {(12<<2)|2,{111,58,0}},
/* 15812 */ {(12<<2)|2,{111,58,0}},
/* 15813 */ {(12<<2)|2,{111,58,0}},
/* 15814 */ {(12<<2)|2,{111,58,0}},
/* 15815 */ {(12<<2)|2,{111,58,0}},
/* 15816 */ {(12<<2)|2,{111,58,0}},
/* 15817 */ {(12<<2)|2,{111,58,0}},
/* 15818 */ {(12<<2)|2,{111,58,0}},
/* 15819 */ {(12<<2)|2,{111,58,0}},
/* 15820 */ {(12<<2)|2,{111,58,0}},
/* 15821 */ {(12<<2)|2,{111,58,0}},
/* 15822 */ {(12<<2)|2,{111,58,0}},
/* 15823 */ {(12<<2)|2,{111,58,0}},
/* 15824 */ {(12<<2)|2,{111,66,0}},
/* 15825 */ {(12<<2)|2,{111,66,0}},
/* 15826 */ {(12<<2)|2,{111,66,0}},
/* 15827 */ {(12<<2)|2,{111,66,0}},
/* 15828 */ {(12<<2)|2,{111,66,0}},
/* 15829 */ {(12<<2)|2,{111,66,0}},
/* 15830 */ {(12<<2)|2,{111,66,0}},
/* 15831 */ {(12<<2)|2,{111,66,0}},
/* 15832 */ {(12<<2)|2,{111,66,0}},
/* 15833 */ {(12<<2)|2,{111,66,0}},
/* 15834 */ {(12<<2)|2,{111,66,0}},
/* 15835 */ {(12<<2)|2,{111,66,0}},
/* 15836 */ {(12<<2)|2,{111,66,0}},
/* 15837 */ {(12<<2)|2,{111,66,0}},
/* 15838 */ {(12<<2)|2,{111,66,0}},
/* 15839 */ {(12<<2)|2,{111,66,0}},
/* 15840 */ {(12<<2)|2,{111,67,0}},
/* 15841 */ {(12<<2)|2,{111,67,0}},
/* 15842 */ {(12<<2)|2,{111,67,0}},
/* 15843 */ {(12<<2)|2,{111,67,0}},
/* 15844 */ {(12<<2)|2,{111,67,0}},
/* 15845 */ {(12<<2)|2,{111,67,0}},
/* 15846 */ {(12<<2)|2,{111,67,0}},
/* 15847 */ {(12<<2)|2,{111,67,0}},
/* 15848 */ {(12<<2)|2,{111,67,0}},
/* 15849 */ {(12<<2)|2,{111,67,0}},
/* 15850 */ {(12<<2)|2,{111,67,0}},
/* 15851 */ {(12<<2)|2,{111,67,0}},
/* 15852 */ {(12<<2)|2,{111,67,0}},
/* 15853 */ {(12<<2)|2,{111,67,0}},
/* 15854 */ {(12<<2)|2,{111,67,0}},
/* 15855 */ {(12<<2)|2,{111,67,0}},
/* 15856 */ {(12<<2)|2,{111,68,0}},
/* 15857 */ {(12<<2)|2,{111,68,0}},
/* 15858 */ {(12<<2)|2,{111,68,0}},
/* 15859 */ {(12<<2)|2,{111,68,0}},
/* 15860 */ {(12<<2)|2,{111,68,0}},
/* 15861 */ {(12<<2)|2,{111,68,0}},
/* 15862 */ {(12<<2)|2,{111,68,0}},
/* 15863 */ {(12<<2)|2,{111,68,0}},
/* 15864 */ {(12<<2)|2,{111,68,0}},
/* 15865 */ {(12<<2)|2,{111,68,0}},
/* 15866 */ {(12<<2)|2,{111,68,0}},
/* 15867 */ {(12<<2)|2,{111,68,0}},
/* 15868 */ {(12<<2)|2,{111,68,0}},
/* 15869 */ {(12<<2)|2,{111,68,0}},
/* 15870 */ {(12<<2)|2,{111,68,0}},
/* 15871 */ {(12<<2)|2,{111,68,0}},
/* 15872 */ {(12<<2)|2,{111,69,0}},
/* 15873 */ {(12<<2)|2,{111,69,0}},
/* 15874 */ {(12<<2)|2,{111,69,0}},
/* 15875 */ {(12<<2)|2,{111,69,0}},
/* 15876 */ {(12<<2)|2,{111,69,0}},
/* 15877 */ {(12<<2)|2,{111,69,0}},
/* 15878 */ {(12<<2)|2,{111,69,0}},
/* 15879 */ {(12<<2)|2,{111,69,0}},
/* 15880 */ {(12<<2)|2,{111,69,0}},
/* 15881 */ {(12<<2)|2,{111,69,0}},
/* 15882 */ {(12<<2)|2,{111,69,0}},
/* 15883 */ {(12<<2)|2,{111,69,0}},
/* 15884 */ {(12<<2)|2,{111,69,0}},
/* 15885 */ {(12<<2)|2,{111,69,0}},
/* 15886 */ {(12<<2)|2,{111,69,0}},
/* 15887 */ {(12<<2)|2,{111,69,0}},
/* 15888 */ {(12<<2)|2,{111,70,0}},
/* 15889 */ {(12<<2)|2,{111,70,0}},
/* 15890 */ {(12<<2)|2,{111,70,0}},
/* 15891 */ {(12<<2)|2,{111,70,0}},
/* 15892 */ {(12<<2)|2,{111,70,0}},
/* 15893 */ {(12<<2)|2,{111,70,0}},
/* 15894 */ {(12<<2)|2,{111,70,0}},
/* 15895 */ {(12<<2)|2,{111,70,0}},
/* 15896 */ {(12<<2)|2,{111,70,0}},
/* 15897 */ {(12<<2)|2,{111,70,0}},
/* 15898 */ {(12<<2)|2,{111,70,0}},
/* 15899 */ {(12<<2)|2,{111,70,0}},
/* 15900 */ {(12<<2)|2,{111,70,0}},
/* 15901 */ {(12<<2)|2,{111,70,0}},
/* 15902 */ {(12<<2)|2,{111,70,0}},
/* 15903 */ {(12<<2)|2,{111,70,0}},
/* 15904 */ {(12<<2)|2,{111,71,0}},
/* 15905 */ {(12<<2)|2,{111,71,0}},
/* 15906 */ {(12<<2)|2,{111,71,0}},
/* 15907 */ {(12<<2)|2,{111,71,0}},
/* 15908 */ {(12<<2)|2,{111,71,0}},
/* 15909 */ {(12<<2)|2,{111,71,0}},
/* 15910 */ {(12<<2)|2,{111,71,0}},
/* 15911 */ {(12<<2)|2,{111,71,0}},
/* 15912 */ {(12<<2)|2,{111,71,0}},
/* 15913 */ {(12<<2)|2,{111,71,0}},
/* 15914 */ {(12<<2)|2,{111,71,0}},
/* 15915 */ {(12<<2)|2,{111,71,0}},
/* 15916 */ {(12<<2)|2,{111,71,0}},
/* 15917 */ {(12<<2)|2,{111,71,0}},
/* 15918 */ {(12<<2)|2,{111,71,0}},
/* 15919 */ {(12<<2)|2,{111,71,0}},
/* 15920 */ {(12<<2)|2,{111,72,0}},
/* 15921 */ {(12<<2)|2,{111,72,0}},
/* 15922 */ {(12<<2)|2,{111,72,0}},
/* 15923 */ {(12<<2)|2,{111,72,0}},
/* 15924 */ {(12<<2)|2,{111,72,0}},
/* 15925 */ {(12<<2)|2,{111,72,0}},
/* 15926 */ {(12<<2)|2,{111,72,0}},
/* 15927 */ {(12<<2)|2,{111,72,0}},
/* 15928 */ {(12<<2)|2,{111,72,0}},
/* 15929 */ {(12<<2)|2,{111,72,0}},
/* 15930 */ {(12<<2)|2,{111,72,0}},
/* 15931 */ {(12<<2)|2,{111,72,0}},
/* 15932 */ {(12<<2)|2,{111,72,0}},
/* 15933 */ {(12<<2)|2,{111,72,0}},
/* 15934 */ {(12<<2)|2,{111,72,0}},
/* 15935 */ {(12<<2)|2,{111,72,0}},
/* 15936 */ {(12<<2)|2,{111,73,0}},
/* 15937 */ {(12<<2)|2,{111,73,0}},
/* 15938 */ {(12<<2)|2,{111,73,0}},
/* 15939 */ {(12<<2)|2,{111,73,0}},
/* 15940 */ {(12<<2)|2,{111,73,0}},
/* 15941 */ {(12<<2)|2,{111,73,0}},
/* 15942 */ {(12<<2)|2,{111,73,0}},
/* 15943 */ {(12<<2)|2,{111,73,0}},
/* 15944 */ {(12<<2)|2,{111,73,0}},
/* 15945 */ {(12<<2)|2,{111,73,0}},
/* 15946 */ {(12<<2)|2,{111,73,0}},
/* 15947 */ {(12<<2)|2,{111,73,0}},
/* 15948 */ {(12<<2)|2,{111,73,0}},
/* 15949 */ {(12<<2)|2,{111,73,0}},
/* 15950 */ {(12<<2)|2,{111,73,0}},
/* 15951 */ {(12<<2)|2,{111,73,0}},
/* 15952 */ {(12<<2)|2,{111,74,0}},
/* 15953 */ {(12<<2)|2,{111,74,0}},
/* 15954 */ {(12<<2)|2,{111,74,0}},
/* 15955 */ {(12<<2)|2,{111,74,0}},
/* 15956 */ {(12<<2)|2,{111,74,0}},
/* 15957 */ {(12<<2)|2,{111,74,0}},
/* 15958 */ {(12<<2)|2,{111,74,0}},
/* 15959 */ {(12<<2)|2,{111,74,0}},
/* 15960 */ {(12<<2)|2,{111,74,0}},
/* 15961 */ {(12<<2)|2,{111,74,0}},
/* 15962 */ {(12<<2)|2,{111,74,0}},
/* 15963 */ {(12<<2)|2,{111,74,0}},
/* 15964 */ {(12<<2)|2,{111,74,0}},
/* 15965 */ {(12<<2)|2,{111,74,0}},
/* 15966 */ {(12<<2)|2,{111,74,0}},
/* 15967 */ {(12<<2)|2,{111,74,0}},
/* 15968 */ {(12<<2)|2,{111,75,0}},
/* 15969 */ {(12<<2)|2,{111,75,0}},
/* 15970 */ {(12<<2)|2,{111,75,0}},
/* 15971 */ {(12<<2)|2,{111,75,0}},
/* 15972 */ {(12<<2)|2,{111,75,0}},
/* 15973 */ {(12<<2)|2,{111,75,0}},
/* 15974 */ {(12<<2)|2,{111,75,0}},
/* 15975 */ {(12<<2)|2,{111,75,0}},
/* 15976 */ {(12<<2)|2,{111,75,0}},
/* 15977 */ {(12<<2)|2,{111,75,0}},
/* 15978 */ {(12<<2)|2,{111,75,0}},
/* 15979 */ {(12<<2)|2,{111,75,0}},
/* 15980 */ {(12<<2)|2,{111,75,0}},
/* 15981 */ {(12<<2)|2,{111,75,0}},
/* 15982 */ {(12<<2)|2,{111,75,0}},
/* 15983 */ {(12<<2)|2,{111,75,0}},
/* 15984 */ {(12<<2)|2,{111,76,0}},
/* 15985 */ {(12<<2)|2,{111,76,0}},
/* 15986 */ {(12<<2)|2,{111,76,0}},
/* 15987 */ {(12<<2)|2,{111,76,0}},
/* 15988 */ {(12<<2)|2,{111,76,0}},
/* 15989 */ {(12<<2)|2,{111,76,0}},
/* 15990 */ {(12<<2)|2,{111,76,0}},
/* 15991 */ {(12<<2)|2,{111,76,0}},
/* 15992 */ {(12<<2)|2,{111,76,0}},
/* 15993 */ {(12<<2)|2,{111,76,0}},
/* 15994 */ {(12<<2)|2,{111,76,0}},
/* 15995 */ {(12<<2)|2,{111,76,0}},
/* 15996 */ {(12<<2)|2,{111,76,0}},
/* 15997 */ {(12<<2)|2,{111,76,0}},
/* 15998 */ {(12<<2)|2,{111,76,0}},
/* 15999 */ {(12<<2)|2,{111,76,0}},
/* 16000 */ {(12<<2)|2,{111,77,0}},
/* 16001 */ {(12<<2)|2,{111,77,0}},
/* 16002 */ {(12<<2)|2,{111,77,0}},
/* 16003 */ {(12<<2)|2,{111,77,0}},
/* 16004 */ {(12<<2)|2,{111,77,0}},
/* 16005 */ {(12<<2)|2,{111,77,0}},
/* 16006 */ {(12<<2)|2,{111,77,0}},
/* 16007 */ {(12<<2)|2,{111,77,0}},
/* 16008 */ {(12<<2)|2,{111,77,0}},
/* 16009 */ {(12<<2)|2,{111,77,0}},
/* 16010 */ {(12<<2)|2,{111,77,0}},
/* 16011 */ {(12<<2)|2,{111,77,0}},
/* 16012 */ {(12<<2)|2,{111,77,0}},
/* 16013 */ {(12<<2)|2,{111,77,0}},
/* 16014 */ {(12<<2)|2,{111,77,0}},
/* 16015 */ {(12<<2)|2,{111,77,0}},
/* 16016 */ {(12<<2)|2,{111,78,0}},
/* 16017 */ {(12<<2)|2,{111,78,0}},
/* 16018 */ {(12<<2)|2,{111,78,0}},
/* 16019 */ {(12<<2)|2,{111,78,0}},
/* 16020 */ {(12<<2)|2,{111,78,0}},
/* 16021 */ {(12<<2)|2,{111,78,0}},
/* 16022 */ {(12<<2)|2,{111,78,0}},
/* 16023 */ {(12<<2)|2,{111,78,0}},
/* 16024 */ {(12<<2)|2,{111,78,0}},
/* 16025 */ {(12<<2)|2,{111,78,0}},
/* 16026 */ {(12<<2)|2,{111,78,0}},
/* 16027 */ {(12<<2)|2,{111,78,0}},
/* 16028 */ {(12<<2)|2,{111,78,0}},
/* 16029 */ {(12<<2)|2,{111,78,0}},
/* 16030 */ {(12<<2)|2,{111,78,0}},
/* 16031 */ {(12<<2)|2,{111,78,0}},
/* 16032 */ {(12<<2)|2,{111,79,0}},
/* 16033 */ {(12<<2)|2,{111,79,0}},
/* 16034 */ {(12<<2)|2,{111,79,0}},
/* 16035 */ {(12<<2)|2,{111,79,0}},
/* 16036 */ {(12<<2)|2,{111,79,0}},
/* 16037 */ {(12<<2)|2,{111,79,0}},
/* 16038 */ {(12<<2)|2,{111,79,0}},
/* 16039 */ {(12<<2)|2,{111,79,0}},
/* 16040 */ {(12<<2)|2,{111,79,0}},
/* 16041 */ {(12<<2)|2,{111,79,0}},
/* 16042 */ {(12<<2)|2,{111,79,0}},
/* 16043 */ {(12<<2)|2,{111,79,0}},
/* 16044 */ {(12<<2)|2,{111,79,0}},
/* 16045 */ {(12<<2)|2,{111,79,0}},
/* 16046 */ {(12<<2)|2,{111,79,0}},
/* 16047 */ {(12<<2)|2,{111,79,0}},
/* 16048 */ {(12<<2)|2,{111,80,0}},
/* 16049 */ {(12<<2)|2,{111,80,0}},
/* 16050 */ {(12<<2)|2,{111,80,0}},
/* 16051 */ {(12<<2)|2,{111,80,0}},
/* 16052 */ {(12<<2)|2,{111,80,0}},
/* 16053 */ {(12<<2)|2,{111,80,0}},
/* 16054 */ {(12<<2)|2,{111,80,0}},
/* 16055 */ {(12<<2)|2,{111,80,0}},
/* 16056 */ {(12<<2)|2,{111,80,0}},
/* 16057 */ {(12<<2)|2,{111,80,0}},
/* 16058 */ {(12<<2)|2,{111,80,0}},
/* 16059 */ {(12<<2)|2,{111,80,0}},
/* 16060 */ {(12<<2)|2,{111,80,0}},
/* 16061 */ {(12<<2)|2,{111,80,0}},
/* 16062 */ {(12<<2)|2,{111,80,0}},
/* 16063 */ {(12<<2)|2,{111,80,0}},
/* 16064 */ {(12<<2)|2,{111,81,0}},
/* 16065 */ {(12<<2)|2,{111,81,0}},
/* 16066 */ {(12<<2)|2,{111,81,0}},
/* 16067 */ {(12<<2)|2,{111,81,0}},
/* 16068 */ {(12<<2)|2,{111,81,0}},
/* 16069 */ {(12<<2)|2,{111,81,0}},
/* 16070 */ {(12<<2)|2,{111,81,0}},
/* 16071 */ {(12<<2)|2,{111,81,0}},
/* 16072 */ {(12<<2)|2,{111,81,0}},
/* 16073 */ {(12<<2)|2,{111,81,0}},
/* 16074 */ {(12<<2)|2,{111,81,0}},
/* 16075 */ {(12<<2)|2,{111,81,0}},
/* 16076 */ {(12<<2)|2,{111,81,0}},
/* 16077 */ {(12<<2)|2,{111,81,0}},
/* 16078 */ {(12<<2)|2,{111,81,0}},
/* 16079 */ {(12<<2)|2,{111,81,0}},
/* 16080 */ {(12<<2)|2,{111,82,0}},
/* 16081 */ {(12<<2)|2,{111,82,0}},
/* 16082 */ {(12<<2)|2,{111,82,0}},
/* 16083 */ {(12<<2)|2,{111,82,0}},
/* 16084 */ {(12<<2)|2,{111,82,0}},
/* 16085 */ {(12<<2)|2,{111,82,0}},
/* 16086 */ {(12<<2)|2,{111,82,0}},
/* 16087 */ {(12<<2)|2,{111,82,0}},
/* 16088 */ {(12<<2)|2,{111,82,0}},
/* 16089 */ {(12<<2)|2,{111,82,0}},
/* 16090 */ {(12<<2)|2,{111,82,0}},
/* 16091 */ {(12<<2)|2,{111,82,0}},
/* 16092 */ {(12<<2)|2,{111,82,0}},
/* 16093 */ {(12<<2)|2,{111,82,0}},
/* 16094 */ {(12<<2)|2,{111,82,0}},
/* 16095 */ {(12<<2)|2,{111,82,0}},
/* 16096 */ {(12<<2)|2,{111,83,0}},
/* 16097 */ {(12<<2)|2,{111,83,0}},
/* 16098 */ {(12<<2)|2,{111,83,0}},
/* 16099 */ {(12<<2)|2,{111,83,0}},
/* 16100 */ {(12<<2)|2,{111,83,0}},
/* 16101 */ {(12<<2)|2,{111,83,0}},
/* 16102 */ {(12<<2)|2,{111,83,0}},
/* 16103 */ {(12<<2)|2,{111,83,0}},
/* 16104 */ {(12<<2)|2,{111,83,0}},
/* 16105 */ {(12<<2)|2,{111,83,0}},
/* 16106 */ {(12<<2)|2,{111,83,0}},
/* 16107 */ {(12<<2)|2,{111,83,0}},
/* 16108 */ {(12<<2)|2,{111,83,0}},
/* 16109 */ {(12<<2)|2,{111,83,0}},
/* 16110 */ {(12<<2)|2,{111,83,0}},
/* 16111 */ {(12<<2)|2,{111,83,0}},
/* 16112 */ {(12<<2)|2,{111,84,0}},
/* 16113 */ {(12<<2)|2,{111,84,0}},
/* 16114 */ {(12<<2)|2,{111,84,0}},
/* 16115 */ {(12<<2)|2,{111,84,0}},
/* 16116 */ {(12<<2)|2,{111,84,0}},
/* 16117 */ {(12<<2)|2,{111,84,0}},
/* 16118 */ {(12<<2)|2,{111,84,0}},
/* 16119 */ {(12<<2)|2,{111,84,0}},
/* 16120 */ {(12<<2)|2,{111,84,0}},
/* 16121 */ {(12<<2)|2,{111,84,0}},
/* 16122 */ {(12<<2)|2,{111,84,0}},
/* 16123 */ {(12<<2)|2,{111,84,0}},
/* 16124 */ {(12<<2)|2,{111,84,0}},
/* 16125 */ {(12<<2)|2,{111,84,0}},
/* 16126 */ {(12<<2)|2,{111,84,0}},
/* 16127 */ {(12<<2)|2,{111,84,0}},
/* 16128 */ {(12<<2)|2,{111,85,0}},
/* 16129 */ {(12<<2)|2,{111,85,0}},
/* 16130 */ {(12<<2)|2,{111,85,0}},
/* 16131 */ {(12<<2)|2,{111,85,0}},
/* 16132 */ {(12<<2)|2,{111,85,0}},
/* 16133 */ {(12<<2)|2,{111,85,0}},
/* 16134 */ {(12<<2)|2,{111,85,0}},
/* 16135 */ {(12<<2)|2,{111,85,0}},
/* 16136 */ {(12<<2)|2,{111,85,0}},
/* 16137 */ {(12<<2)|2,{111,85,0}},
/* 16138 */ {(12<<2)|2,{111,85,0}},
/* 16139 */ {(12<<2)|2,{111,85,0}},
/* 16140 */ {(12<<2)|2,{111,85,0}},
/* 16141 */ {(12<<2)|2,{111,85,0}},
/* 16142 */ {(12<<2)|2,{111,85,0}},
/* 16143 */ {(12<<2)|2,{111,85,0}},
/* 16144 */ {(12<<2)|2,{111,86,0}},
/* 16145 */ {(12<<2)|2,{111,86,0}},
/* 16146 */ {(12<<2)|2,{111,86,0}},
/* 16147 */ {(12<<2)|2,{111,86,0}},
/* 16148 */ {(12<<2)|2,{111,86,0}},
/* 16149 */ {(12<<2)|2,{111,86,0}},
/* 16150 */ {(12<<2)|2,{111,86,0}},
/* 16151 */ {(12<<2)|2,{111,86,0}},
/* 16152 */ {(12<<2)|2,{111,86,0}},
/* 16153 */ {(12<<2)|2,{111,86,0}},
/* 16154 */ {(12<<2)|2,{111,86,0}},
/* 16155 */ {(12<<2)|2,{111,86,0}},
/* 16156 */ {(12<<2)|2,{111,86,0}},
/* 16157 */ {(12<<2)|2,{111,86,0}},
/* 16158 */ {(12<<2)|2,{111,86,0}},
/* 16159 */ {(12<<2)|2,{111,86,0}},
/* 16160 */ {(12<<2)|2,{111,87,0}},
/* 16161 */ {(12<<2)|2,{111,87,0}},
/* 16162 */ {(12<<2)|2,{111,87,0}},
/* 16163 */ {(12<<2)|2,{111,87,0}},
/* 16164 */ {(12<<2)|2,{111,87,0}},
/* 16165 */ {(12<<2)|2,{111,87,0}},
/* 16166 */ {(12<<2)|2,{111,87,0}},
/* 16167 */ {(12<<2)|2,{111,87,0}},
/* 16168 */ {(12<<2)|2,{111,87,0}},
/* 16169 */ {(12<<2)|2,{111,87,0}},
/* 16170 */ {(12<<2)|2,{111,87,0}},
/* 16171 */ {(12<<2)|2,{111,87,0}},
/* 16172 */ {(12<<2)|2,{111,87,0}},
/* 16173 */ {(12<<2)|2,{111,87,0}},
/* 16174 */ {(12<<2)|2,{111,87,0}},
/* 16175 */ {(12<<2)|2,{111,87,0}},
/* 16176 */ {(12<<2)|2,{111,89,0}},
/* 16177 */ {(12<<2)|2,{111,89,0}},
/* 16178 */ {(12<<2)|2,{111,89,0}},
/* 16179 */ {(12<<2)|2,{111,89,0}},
/* 16180 */ {(12<<2)|2,{111,89,0}},
/* 16181 */ {(12<<2)|2,{111,89,0}},
/* 16182 */ {(12<<2)|2,{111,89,0}},
/* 16183 */ {(12<<2)|2,{111,89,0}},
/* 16184 */ {(12<<2)|2,{111,89,0}},
/* 16185 */ {(12<<2)|2,{111,89,0}},
/* 16186 */ {(12<<2)|2,{111,89,0}},
/* 16187 */ {(12<<2)|2,{111,89,0}},
/* 16188 */ {(12<<2)|2,{111,89,0}},
/* 16189 */ {(12<<2)|2,{111,89,0}},
/* 16190 */ {(12<<2)|2,{111,89,0}},
/* 16191 */ {(12<<2)|2,{111,89,0}},
/* 16192 */ {(12<<2)|2,{111,106,0}},
/* 16193 */ {(12<<2)|2,{111,106,0}},
/* 16194 */ {(12<<2)|2,{111,106,0}},
/* 16195 */ {(12<<2)|2,{111,106,0}},
/* 16196 */ {(12<<2)|2,{111,106,0}},
/* 16197 */ {(12<<2)|2,{111,106,0}},
/* 16198 */ {(12<<2)|2,{111,106,0}},
/* 16199 */ {(12<<2)|2,{111,106,0}},
/* 16200 */ {(12<<2)|2,{111,106,0}},
/* 16201 */ {(12<<2)|2,{111,106,0}},
/* 16202 */ {(12<<2)|2,{111,106,0}},
/* 16203 */ {(12<<2)|2,{111,106,0}},
/* 16204 */ {(12<<2)|2,{111,106,0}},
/* 16205 */ {(12<<2)|2,{111,106,0}},
/* 16206 */ {(12<<2)|2,{111,106,0}},
/* 16207 */ {(12<<2)|2,{111,106,0}},
/* 16208 */ {(12<<2)|2,{111,107,0}},
/* 16209 */ {(12<<2)|2,{111,107,0}},
/* 16210 */ {(12<<2)|2,{111,107,0}},
/* 16211 */ {(12<<2)|2,{111,107,0}},
/* 16212 */ {(12<<2)|2,{111,107,0}},
/* 16213 */ {(12<<2)|2,{111,107,0}},
/* 16214 */ {(12<<2)|2,{111,107,0}},
/* 16215 */ {(12<<2)|2,{111,107,0}},
/* 16216 */ {(12<<2)|2,{111,107,0}},
/* 16217 */ {(12<<2)|2,{111,107,0}},
/* 16218 */ {(12<<2)|2,{111,107,0}},
/* 16219 */ {(12<<2)|2,{111,107,0}},
/* 16220 */ {(12<<2)|2,{111,107,0}},
/* 16221 */ {(12<<2)|2,{111,107,0}},
/* 16222 */ {(12<<2)|2,{111,107,0}},
/* 16223 */ {(12<<2)|2,{111,107,0}},
/* 16224 */ {(12<<2)|2,{111,113,0}},
/* 16225 */ {(12<<2)|2,{111,113,0}},
/* 16226 */ {(12<<2)|2,{111,113,0}},
/* 16227 */ {(12<<2)|2,{111,113,0}},
/* 16228 */ {(12<<2)|2,{111,113,0}},
/* 16229 */ {(12<<2)|2,{111,113,0}},
/* 16230 */ {(12<<2)|2,{111,113,0}},
/* 16231 */ {(12<<2)|2,{111,113,0}},
/* 16232 */ {(12<<2)|2,{111,113,0}},
/* 16233 */ {(12<<2)|2,{111,113,0}},
/* 16234 */ {(12<<2)|2,{111,113,0}},
/* 16235 */ {(12<<2)|2,{111,113,0}},
/* 16236 */ {(12<<2)|2,{111,113,0}},
/* 16237 */ {(12<<2)|2,{111,113,0}},
/* 16238 */ {(12<<2)|2,{111,113,0}},
/* 16239 */ {(12<<2)|2,{111,113,0}},
/* 16240 */ {(12<<2)|2,{111,118,0}},
/* 16241 */ {(12<<2)|2,{111,118,0}},
/* 16242 */ {(12<<2)|2,{111,118,0}},
/* 16243 */ {(12<<2)|2,{111,118,0}},
/* 16244 */ {(12<<2)|2,{111,118,0}},
/* 16245 */ {(12<<2)|2,{111,118,0}},
/* 16246 */ {(12<<2)|2,{111,118,0}},
/* 16247 */ {(12<<2)|2,{111,118,0}},
/* 16248 */ {(12<<2)|2,{111,118,0}},
/* 16249 */ {(12<<2)|2,{111,118,0}},
/* 16250 */ {(12<<2)|2,{111,118,0}},
/* 16251 */ {(12<<2)|2,{111,118,0}},
/* 16252 */ {(12<<2)|2,{111,118,0}},
/* 16253 */ {(12<<2)|2,{111,118,0}},
/* 16254 */ {(12<<2)|2,{111,118,0}},
/* 16255 */ {(12<<2)|2,{111,118,0}},
/* 16256 */ {(12<<2)|2,{111,119,0}},
/* 16257 */ {(12<<2)|2,{111,119,0}},
/* 16258 */ {(12<<2)|2,{111,119,0}},
/* 16259 */ {(12<<2)|2,{111,119,0}},
/* 16260 */ {(12<<2)|2,{111,119,0}},
/* 16261 */ {(12<<2)|2,{111,119,0}},
/* 16262 */ {(12<<2)|2,{111,119,0}},
/* 16263 */ {(12<<2)|2,{111,119,0}},
/* 16264 */ {(12<<2)|2,{111,119,0}},
/* 16265 */ {(12<<2)|2,{111,119,0}},
/* 16266 */ {(12<<2)|2,{111,119,0}},
/* 16267 */ {(12<<2)|2,{111,119,0}},
/* 16268 */ {(12<<2)|2,{111,119,0}},
/* 16269 */ {(12<<2)|2,{111,119,0}},
/* 16270 */ {(12<<2)|2,{111,119,0}},
/* 16271 */ {(12<<2)|2,{111,119,0}},
/* 16272 */ {(12<<2)|2,{111,120,0}},
/* 16273 */ {(12<<2)|2,{111,120,0}},
/* 16274 */ {(12<<2)|2,{111,120,0}},
/* 16275 */ {(12<<2)|2,{111,120,0}},
/* 16276 */ {(12<<2)|2,{111,120,0}},
/* 16277 */ {(12<<2)|2,{111,120,0}},
/* 16278 */ {(12<<2)|2,{111,120,0}},
/* 16279 */ {(12<<2)|2,{111,120,0}},
/* 16280 */ {(12<<2)|2,{111,120,0}},
/* 16281 */ {(12<<2)|2,{111,120,0}},
/* 16282 */ {(12<<2)|2,{111,120,0}},
/* 16283 */ {(12<<2)|2,{111,120,0}},
/* 16284 */ {(12<<2)|2,{111,120,0}},
/* 16285 */ {(12<<2)|2,{111,120,0}},
/* 16286 */ {(12<<2)|2,{111,120,0}},
/* 16287 */ {(12<<2)|2,{111,120,0}},
/* 16288 */ {(12<<2)|2,{111,121,0}},
/* 16289 */ {(12<<2)|2,{111,121,0}},
/* 16290 */ {(12<<2)|2,{111,121,0}},
/* 16291 */ {(12<<2)|2,{111,121,0}},
/* 16292 */ {(12<<2)|2,{111,121,0}},
/* 16293 */ {(12<<2)|2,{111,121,0}},
/* 16294 */ {(12<<2)|2,{111,121,0}},
/* 16295 */ {(12<<2)|2,{111,121,0}},
/* 16296 */ {(12<<2)|2,{111,121,0}},
/* 16297 */ {(12<<2)|2,{111,121,0}},
/* 16298 */ {(12<<2)|2,{111,121,0}},
/* 16299 */ {(12<<2)|2,{111,121,0}},
/* 16300 */ {(12<<2)|2,{111,121,0}},
/* 16301 */ {(12<<2)|2,{111,121,0}},
/* 16302 */ {(12<<2)|2,{111,121,0}},
/* 16303 */ {(12<<2)|2,{111,121,0}},
/* 16304 */ {(12<<2)|2,{111,122,0}},
/* 16305 */ {(12<<2)|2,{111,122,0}},
/* 16306 */ {(12<<2)|2,{111,122,0}},
/* 16307 */ {(12<<2)|2,{111,122,0}},
/* 16308 */ {(12<<2)|2,{111,122,0}},
/* 16309 */ {(12<<2)|2,{111,122,0}},
/* 16310 */ {(12<<2)|2,{111,122,0}},
/* 16311 */ {(12<<2)|2,{111,122,0}},
/* 16312 */ {(12<<2)|2,{111,122,0}},
/* 16313 */ {(12<<2)|2,{111,122,0}},
/* 16314 */ {(12<<2)|2,{111,122,0}},
/* 16315 */ {(12<<2)|2,{111,122,0}},
/* 16316 */ {(12<<2)|2,{111,122,0}},
/* 16317 */ {(12<<2)|2,{111,122,0}},
/* 16318 */ {(12<<2)|2,{111,122,0}},
/* 16319 */ {(12<<2)|2,{111,122,0}},
/* 16320 */ {(13<<2)|2,{111,38,0}},
/* 16321 */ {(13<<2)|2,{111,38,0}},
/* 16322 */ {(13<<2)|2,{111,38,0}},
/* 16323 */ {(13<<2)|2,{111,38,0}},
/* 16324 */ {(13<<2)|2,{111,38,0}},
/* 16325 */ {(13<<2)|2,{111,38,0}},
/* 16326 */ {(13<<2)|2,{111,38,0}},
/* 16327 */ {(13<<2)|2,{111,38,0}},
/* 16328 */ {(13<<2)|2,{111,42,0}},
/* 16329 */ {(13<<2)|2,{111,42,0}},
/* 16330 */ {(13<<2)|2,{111,42,0}},
/* 16331 */ {(13<<2)|2,{111,42,0}},
/* 16332 */ {(13<<2)|2,{111,42,0}},
/* 16333 */ {(13<<2)|2,{111,42,0}},
/* 16334 */ {(13<<2)|2,{111,42,0}},
/* 16335 */ {(13<<2)|2,{111,42,0}},
/* 16336 */ {(13<<2)|2,{111,44,0}},
/* 16337 */ {(13<<2)|2,{111,44,0}},
/* 16338 */ {(13<<2)|2,{111,44,0}},
/* 16339 */ {(13<<2)|2,{111,44,0}},
/* 16340 */ {(13<<2)|2,{111,44,0}},
/* 16341 */ {(13<<2)|2,{111,44,0}},
/* 16342 */ {(13<<2)|2,{111,44,0}},
/* 16343 */ {(13<<2)|2,{111,44,0}},
/* 16344 */ {(13<<2)|2,{111,59,0}},
/* 16345 */ {(13<<2)|2,{111,59,0}},
/* 16346 */ {(13<<2)|2,{111,59,0}},
/* 16347 */ {(13<<2)|2,{111,59,0}},
/* 16348 */ {(13<<2)|2,{111,59,0}},
/* 16349 */ {(13<<2)|2,{111,59,0}},
/* 16350 */ {(13<<2)|2,{111,59,0}},
/* 16351 */ {(13<<2)|2,{111,59,0}},
/* 16352 */ {(13<<2)|2,{111,88,0}},
/* 16353 */ {(13<<2)|2,{111,88,0}},
/* 16354 */ {(13<<2)|2,{111,88,0}},
/* 16355 */ {(13<<2)|2,{111,88,0}},
/* 16356 */ {(13<<2)|2,{111,88,0}},
/* 16357 */ {(13<<2)|2,{111,88,0}},
/* 16358 */ {(13<<2)|2,{111,88,0}},
/* 16359 */ {(13<<2)|2,{111,88,0}},
/* 16360 */ {(13<<2)|2,{111,90,0}},
/* 16361 */ {(13<<2)|2,{111,90,0}},
/* 16362 */ {(13<<2)|2,{111,90,0}},
/* 16363 */ {(13<<2)|2,{111,90,0}},
/* 16364 */ {(13<<2)|2,{111,90,0}},
/* 16365 */ {(13<<2)|2,{111,90,0}},
/* 16366 */ {(13<<2)|2,{111,90,0}},
/* 16367 */ {(13<<2)|2,{111,90,0}},
/* 16368 */ {(15<<2)|2,{111,33,0}},
/* 16369 */ {(15<<2)|2,{111,33,0}},
/* 16370 */ {(15<<2)|2,{111,34,0}},
/* 16371 */ {(15<<2)|2,{111,34,0}},
/* 16372 */ {(15<<2)|2,{111,40,0}},
/* 16373 */ {(15<<2)|2,{111,40,0}},
/* 16374 */ {(15<<2)|2,{111,41,0}},
/* 16375 */ {(15<<2)|2,{111,41,0}},
/* 16376 */ {(15<<2)|2,{111,63,0}},
/* 16377 */ {(15<<2)|2,{111,63,0}},
/* 16378 */ {(16<<2)|2,{111,39,0}},
/* 16379 */ {(16<<2)|2,{111,43,0}},
/* 16380 */ {(16<<2)|2,{111,124,0}},
/* 16381 */ {(5<<2)|1,{111,0,0}},
/* 16382 */ {(5<<2)|1,{111,0,0}},
/* 16383 */ {(5<<2)|1,{111,0,0}},
/* 16384 */ {(15<<2)|3,{115,48,48}},
/* 16385 */ {(15<<2)|3,{115,48,48}},
/* 16386 */ {(15<<2)|3,{115,48,49}},
/* 16387 */ {(15<<2)|3,{115,48,49}},
/* 16388 */ {(15<<2)|3,{115,48,50}},
/* 16389 */ {(15<<2)|3,{115,48,50}},
/* 16390 */ {(15<<2)|3,{115,48,97}},
/* 16391 */ {(15<<2)|3,{115,48,97}},
/* 16392 */ {(15<<2)|3,{115,48,99}},
/* 16393 */ {(15<<2)|3,{115,48,99}},
/* 16394 */ {(15<<2)|3,{115,48,101}},
/* 16395 */ {(15<<2)|3,{115,48,101}},
/* 16396 */ {(15<<2)|3,{115,48,105}},
/* 16397 */ {(15<<2)|3,{115,48,105}},
/* 16398 */ {(15<<2)|3,{115,48,111}},
/* 16399 */ {(15<<2)|3,{115,48,111}},
/* 16400 */ {(15<<2)|3,{115,48,115}},
/* 16401 */ {(15<<2)|3,{115,48,115}},
/* 16402 */ {(15<<2)|3,{115,48,116}},
/* 16403 */ {(15<<2)|3,{115,48,116}},
/* 16404 */ {(16<<2)|3,{115,48,32}},
/* 16405 */ {(16<<2)|3,{115,48,37}},
/* 16406 */ {(16<<2)|3,{115,48,45}},
/* 16407 */ {(16<<2)|3,{115,48,46}},
/* 16408 */ {(16<<2)|3,{115,48,47}},
/* 16409 */ {(16<<2)|3,{115,48,51}},
/* 16410 */ {(16<<2)|3,{115,48,52}},
/* 16411 */ {(16<<2)|3,{115,48,53}},
/* 16412 */ {(16<<2)|3,{115,48,54}},
/* 16413 */ {(16<<2)|3,{115,48,55}},
/* 16414 */ {(16<<2)|3,{115,48,56}},
/* 16415 */ {(16<<2)|3,{115,48,57}},
/* 16416 */ {(16<<2)|3,{115,48,61}},
/* 16417 */ {(16<<2)|3,{115,48,65}},
/* 16418 */ {(16<<2)|3,{115,48,95}},
/* 16419 */ {(16<<2)|3,{115,48,98}},
/* 16420 */ {(16<<2)|3,{115,48,100}},
/* 16421 */ {(16<<2)|3,{115,48,102}},
/* 16422 */ {(16<<2)|3,{115,48,103}},
/* 16423 */ {(16<<2)|3,{115,48,104}},
/* 16424 */ {(16<<2)|3,{115,48,108}},
/* 16425 */ {(16<<2)|3,{115,48,109}},
/* 16426 */ {(16<<2)|3,{115,48,110}},
/* 16427 */ {(16<<2)|3,{115,48,112}},
/* 16428 */ {(16<<2)|3,{115,48,114}},
/* 16429 */ {(16<<2)|3,{115,48,117}},
/* 16430 */ {(10<<2)|2,{115,48,0}},
/* 16431 */ {(10<<2)|2,{115,48,0}},
/* 16432 */ {(10<<2)|2,{115,48,0}},
/* 16433 */ {(10<<2)|2,{115,48,0}},
/* 16434 */ {(10<<2)|2,{115,48,0}},
/* 16435 */ {(10<<2)|2,{115,48,0}},
/* 16436 */ {(10<<2)|2,{115,48,0}},
/* 16437 */ {(10<<2)|2,{115,48,0}},
/* 16438 */ {(10<<2)|2,{115,48,0}},
/* 16439 */ {(10<<2)|2,{115,48,0}},
/* 16440 */ {(10<<2)|2,{115,48,0}},
/* 16441 */ {(10<<2)|2,{115,48,0}},
/* 16442 */ {(10<<2)|2,{115,48,0}},
/* 16443 */ {(10<<2)|2,{115,48,0}},
/* 16444 */ {(10<<2)|2,{115,48,0}},
/* 16445 */ {(10<<2)|2,{115,48,0}},
/* 16446 */ {(10<<2)|2,{115,48,0}},
/* 16447 */ {(10<<2)|2,{115,48,0}},
/* 16448 */ {(15<<2)|3,{115,49,48}},
/* 16449 */ {(15<<2)|3,{115,49,48}},
/* 16450 */ {(15<<2)|3,{115,49,49}},
/* 16451 */ {(15<<2)|3,{115,49,49}},
/* 16452 */ {(15<<2)|3,{115,49,50}},
/* 16453 */ {(15<<2)|3,{115,49,50}},
/* 16454 */ {(15<<2)|3,{115,49,97}},
/* 16455 */ {(15<<2)|3,{115,49,97}},
/* 16456 */ {(15<<2)|3,{115,49,99}},
/* 16457 */ {(15<<2)|3,{115,49,99}},
/* 16458 */ {(15<<2)|3,{115,49,101}},
/* 16459 */ {(15<<2)|3,{115,49,101}},
/* 16460 */ {(15<<2)|3,{115,49,105}},
/* 16461 */ {(15<<2)|3,{115,49,105}},
/* 16462 */ {(15<<2)|3,{115,49,111}},
/* 16463 */ {(15<<2)|3,{115,49,111}},
/* 16464 */ {(15<<2)|3,{115,49,115}},
/* 16465 */ {(15<<2)|3,{115,49,115}},
/* 16466 */ {(15<<2)|3,{115,49,116}},
/* 16467 */ {(15<<2)|3,{115,49,116}},
/* 16468 */ {(16<<2)|3,{115,49,32}},
/* 16469 */ {(16<<2)|3,{115,49,37}},
/* 16470 */ {(16<<2)|3,{115,49,45}},
/* 16471 */ {(16<<2)|3,{115,49,46}},
/* 16472 */ {(16<<2)|3,{115,49,47}},
/* 16473 */ {(16<<2)|3,{115,49,51}},
/* 16474 */ {(16<<2)|3,{115,49,52}},
/* 16475 */ {(16<<2)|3,{115,49,53}},
/* 16476 */ {(16<<2)|3,{115,49,54}},
/* 16477 */ {(16<<2)|3,{115,49,55}},
/* 16478 */ {(16<<2)|3,{115,49,56}},
/* 16479 */ {(16<<2)|3,{115,49,57}},
/* 16480 */ {(16<<2)|3,{115,49,61}},
/* 16481 */ {(16<<2)|3,{115,49,65}},
/* 16482 */ {(16<<2)|3,{115,49,95}},
/* 16483 */ {(16<<2)|3,{115,49,98}},
/* 16484 */ {(16<<2)|3,{115,49,100}},
/* 16485 */ {(16<<2)|3,{115,49,102}},
/* 16486 */ {(16<<2)|3,{115,49,103}},
/* 16487 */ {(16<<2)|3,{115,49,104}},
/* 16488 */ {(16<<2)|3,{115,49,108}},
/* 16489 */ {(16<<2)|3,{115,49,109}},
/* 16490 */ {(16<<2)|3,{115,49,110}},
/* 16491 */ {(16<<2)|3,{115,49,112}},
/* 16492 */ {(16<<2)|3,{115,49,114}},
/* 16493 */ {(16<<2)|3,{115,49,117}},
/* 16494 */ {(10<<2)|2,{115,49,0}},
/* 16495 */ {(10<<2)|2,{115,49,0}},
/* 16496 */ {(10<<2)|2,{115,49,0}},
/* 16497 */ {(10<<2)|2,{115,49,0}},
/* 16498 */ {(10<<2)|2,{115,49,0}},
/* 16499 */ {(10<<2)|2,{115,49,0}},
/* 16500 */ {(10<<2)|2,{115,49,0}},
/* 16501 */ {(10<<2)|2,{115,49,0}},
/* 16502 */ {(10<<2)|2,{115,49,0}},
/* 16503 */ {(10<<2)|2,{115,49,0}},
/* 16504 */ {(10<<2)|2,{115,49,0}},
/* 16505 */ {(10<<2)|2,{115,49,0}},
/* 16506 */ {(10<<2)|2,{115,49,0}},
/* 16507 */ {(10<<2)|2,{115,49,0}},
/* 16508 */ {(10<<2)|2,{115,49,0}},
/* 16509 */ {(10<<2)|2,{115,49,0}},
/* 16510 */ {(10<<2)|2,{115,49,0}},
/* 16511 */ {(10<<2)|2,{115,49,0}},
/* 16512 */ {(15<<2)|3,{115,50,48}},
/* 16513 */ {(15<<2)|3,{115,50,48}},
/* 16514 */ {(15<<2)|3,{115,50,49}},
/* 16515 */ {(15<<2)|3,{115,50,49}},
/* 16516 */ {(15<<2)|3,{115,50,50}},
/* 16517 */ {(15<<2)|3,{115,50,50}},
/* 16518 */ {(15<<2)|3,{115,50,97}},
/* 16519 */ {(15<<2)|3,{115,50,97}},
/* 16520 */ {(15<<2)|3,{115,50,99}},
/* 16521 */ {(15<<2)|3,{115,50,99}},
/* 16522 */ {(15<<2)|3,{115,50,101}},
/* 16523 */ {(15<<2)|3,{115,50,101}},
/* 16524 */ {(15<<2)|3,{115,50,105}},
/* 16525 */ {(15<<2)|3,{115,50,105}},
/* 16526 */ {(15<<2)|3,{115,50,111}},
/* 16527 */ {(15<<2)|3,{115,50,111}},
/* 16528 */ {(15<<2)|3,{115,50,115}},
/* 16529 */ {(15<<2)|3,{115,50,115}},
/* 16530 */ {(15<<2)|3,{115,50,116}},
/* 16531 */ {(15<<2)|3,{115,50,116}},
/* 16532 */ {(16<<2)|3,{115,50,32}},
/* 16533 */ {(16<<2)|3,{115,50,37}},
/* 16534 */ {(16<<2)|3,{115,50,45}},
/* 16535 */ {(16<<2)|3,{115,50,46}},
/* 16536 */ {(16<<2)|3,{115,50,47}},
/* 16537 */ {(16<<2)|3,{115,50,51}},
/* 16538 */ {(16<<2)|3,{115,50,52}},
/* 16539 */ {(16<<2)|3,{115,50,53}},
/* 16540 */ {(16<<2)|3,{115,50,54}},
/* 16541 */ {(16<<2)|3,{115,50,55}},
/* 16542 */ {(16<<2)|3,{115,50,56}},
/* 16543 */ {(16<<2)|3,{115,50,57}},
/* 16544 */ {(16<<2)|3,{115,50,61}},
/* 16545 */ {(16<<2)|3,{115,50,65}},
/* 16546 */ {(16<<2)|3,{115,50,95}},
/* 16547 */ {(16<<2)|3,{115,50,98}},
/* 16548 */ {(16<<2)|3,{115,50,100}},
/* 16549 */ {(16<<2)|3,{115,50,102}},
/* 16550 */ {(16<<2)|3,{115,50,103}},
/* 16551 */ {(16<<2)|3,{115,50,104}},
/* 16552 */ {(16<<2)|3,{115,50,108}},
/* 16553 */ {(16<<2)|3,{115,50,109}},
/* 16554 */ {(16<<2)|3,{115,50,110}},
/* 16555 */ {(16<<2)|3,{115,50,112}},
/* 16556 */ {(16<<2)|3,{115,50,114}},
/* 16557 */ {(16<<2)|3,{115,50,117}},
/* 16558 */ {(10<<2)|2,{115,50,0}},
/* 16559 */ {(10<<2)|2,{115,50,0}},
/* 16560 */ {(10<<2)|2,{115,50,0}},
/* 16561 */ {(10<<2)|2,{115,50,0}},
/* 16562 */ {(10<<2)|2,{115,50,0}},
/* 16563 */ {(10<<2)|2,{115,50,0}},
/* 16564 */ {(10<<2)|2,{115,50,0}},
/* 16565 */ {(10<<2)|2,{115,50,0}},
/* 16566 */ {(10<<2)|2,{115,50,0}},
/* 16567 */ {(10<<2)|2,{115,50,0}},
/* 16568 */ {(10<<2)|2,{115,50,0}},
/* 16569 */ {(10<<2)|2,{115,50,0}},
/* 16570 */ {(10<<2)|2,{115,50,0}},
/* 16571 */ {(10<<2)|2,{115,50,0}},
/* 16572 */ {(10<<2)|2,{115,50,0}},
/* 16573 */ {(10<<2)|2,{115,50,0}},
/* 16574 */ {(10<<2)|2,{115,50,0}},
/* 16575 */ {(10<<2)|2,{115,50,0}},
/* 16576 */ {(15<<2)|3,{115,97,48}},
/* 16577 */ {(15<<2)|3,{115,97,48}},
/* 16578 */ {(15<<2)|3,{115,97,49}},
/* 16579 */ {(15<<2)|3,{115,97,49}},
/* 16580 */ {(15<<2)|3,{115,97,50}},
/* 16581 */ {(15<<2)|3,{115,97,50}},
/* 16582 */ {(15<<2)|3,{115,97,97}},
/* 16583 */ {(15<<2)|3,{115,97,97}},
/* 16584 */ {(15<<2)|3,{115,97,99}},
/* 16585 */ {(15<<2)|3,{115,97,99}},
/* 16586 */ {(15<<2)|3,{115,97,101}},
/* 16587 */ {(15<<2)|3,{115,97,101}},
/* 16588 */ {(15<<2)|3,{115,97,105}},
/* 16589 */ {(15<<2)|3,{115,97,105}},
/* 16590 */ {(15<<2)|3,{115,97,111}},
/* 16591 */ {(15<<2)|3,{115,97,111}},
/* 16592 */ {(15<<2)|3,{115,97,115}},
/* 16593 */ {(15<<2)|3,{115,97,115}},
/* 16594 */ {(15<<2)|3,{115,97,116}},
/* 16595 */ {(15<<2)|3,{115,97,116}},
/* 16596 */ {(16<<2)|3,{115,97,32}},
/* 16597 */ {(16<<2)|3,{115,97,37}},
/* 16598 */ {(16<<2)|3,{115,97,45}},
/* 16599 */ {(16<<2)|3,{115,97,46}},
/* 16600 */ {(16<<2)|3,{115,97,47}},
/* 16601 */ {(16<<2)|3,{115,97,51}},
/* 16602 */ {(16<<2)|3,{115,97,52}},
/* 16603 */ {(16<<2)|3,{115,97,53}},
/* 16604 */ {(16<<2)|3,{115,97,54}},
/* 16605 */ {(16<<2)|3,{115,97,55}},
/* 16606 */ {(16<<2)|3,{115,97,56}},
/* 16607 */ {(16<<2)|3,{115,97,57}},
/* 16608 */ {(16<<2)|3,{115,97,61}},
/* 16609 */ {(16<<2)|3,{115,97,65}},
/* 16610 */ {(16<<2)|3,{115,97,95}},
/* 16611 */ {(16<<2)|3,{115,97,98}},
/* 16612 */ {(16<<2)|3,{115,97,100}},
/* 16613 */ {(16<<2)|3,{115,97,102}},
/* 16614 */ {(16<<2)|3,{115,97,103}},
/* 16615 */ {(16<<2)|3,{115,97,104}},
/* 16616 */ {(16<<2)|3,{115,97,108}},
/* 16617 */ {(16<<2)|3,{115,97,109}},
/* 16618 */ {(16<<2)|3,{115,97,110}},
/* 16619 */ {(16<<2)|3,{115,97,112}},
/* 16620 */ {(16<<2)|3,{115,97,114}},
/* 16621 */ {(16<<2)|3,{115,97,117}},
/* 16622 */ {(10<<2)|2,{115,97,0}},
/* 16623 */ {(10<<2)|2,{115,97,0}},
/* 16624 */ {(10<<2)|2,{115,97,0}},
/* 16625 */ {(10<<2)|2,{115,97,0}},
/* 16626 */ {(10<<2)|2,{115,97,0}},
/* 16627 */ {(10<<2)|2,{115,97,0}},
/* 16628 */ {(10<<2)|2,{115,97,0}},
/* 16629 */ {(10<<2)|2,{115,97,0}},
/* 16630 */ {(10<<2)|2,{115,97,0}},
/* 16631 */ {(10<<2)|2,{115,97,0}},
/* 16632 */ {(10<<2)|2,{115,97,0}},
/* 16633 */ {(10<<2)|2,{115,97,0}},
/* 16634 */ {(10<<2)|2,{115,97,0}},
/* 16635 */ {(10<<2)|2,{115,97,0}},
/* 16636 */ {(10<<2)|2,{115,97,0}},
/* 16637 */ {(10<<2)|2,{115,97,0}},
/* 16638 */ {(10<<2)|2,{115,97,0}},
/* 16639 */ {(10<<2)|2,{115,97,0}},
/* 16640 */ {(15<<2)|3,{115,99,48}},
/* 16641 */ {(15<<2)|3,{115,99,48}},
/* 16642 */ {(15<<2)|3,{115,99,49}},
/* 16643 */ {(15<<2)|3,{115,99,49}},
/* 16644 */ {(15<<2)|3,{115,99,50}},
/* 16645 */ {(15<<2)|3,{115,99,50}},
/* 16646 */ {(15<<2)|3,{115,99,97}},
/* 16647 */ {(15<<2)|3,{115,99,97}},
/* 16648 */ {(15<<2)|3,{115,99,99}},
/* 16649 */ {(15<<2)|3,{115,99,99}},
/* 16650 */ {(15<<2)|3,{115,99,101}},
/* 16651 */ {(15<<2)|3,{115,99,101}},
/* 16652 */ {(15<<2)|3,{115,99,105}},
/* 16653 */ {(15<<2)|3,{115,99,105}},
/* 16654 */ {(15<<2)|3,{115,99,111}},
/* 16655 */ {(15<<2)|3,{115,99,111}},
/* 16656 */ {(15<<2)|3,{115,99,115}},
/* 16657 */ {(15<<2)|3,{115,99,115}},
/* 16658 */ {(15<<2)|3,{115,99,116}},
/* 16659 */ {(15<<2)|3,{115,99,116}},
/* 16660 */ {(16<<2)|3,{115,99,32}},
/* 16661 */ {(16<<2)|3,{115,99,37}},
/* 16662 */ {(16<<2)|3,{115,99,45}},
/* 16663 */ {(16<<2)|3,{115,99,46}},
/* 16664 */ {(16<<2)|3,{115,99,47}},
/* 16665 */ {(16<<2)|3,{115,99,51}},
/* 16666 */ {(16<<2)|3,{115,99,52}},
/* 16667 */ {(16<<2)|3,{115,99,53}},
/* 16668 */ {(16<<2)|3,{115,99,54}},
/* 16669 */ {(16<<2)|3,{115,99,55}},
/* 16670 */ {(16<<2)|3,{115,99,56}},
/* 16671 */ {(16<<2)|3,{115,99,57}},
/* 16672 */ {(16<<2)|3,{115,99,61}},
/* 16673 */ {(16<<2)|3,{115,99,65}},
/* 16674 */ {(16<<2)|3,{115,99,95}},
/* 16675 */ {(16<<2)|3,{115,99,98}},
/* 16676 */ {(16<<2)|3,{115,99,100}},
/* 16677 */ {(16<<2)|3,{115,99,102}},
/* 16678 */ {(16<<2)|3,{115,99,103}},
/* 16679 */ {(16<<2)|3,{115,99,104}},
/* 16680 */ {(16<<2)|3,{115,99,108}},
/* 16681 */ {(16<<2)|3,{115,99,109}},
/* 16682 */ {(16<<2)|3,{115,99,110}},
/* 16683 */ {(16<<2)|3,{115,99,112}},
/* 16684 */ {(16<<2)|3,{115,99,114}},
/* 16685 */ {(16<<2)|3,{115,99,117}},
/* 16686 */ {(10<<2)|2,{115,99,0}},
/* 16687 */ {(10<<2)|2,{115,99,0}},
/* 16688 */ {(10<<2)|2,{115,99,0}},
/* 16689 */ {(10<<2)|2,{115,99,0}},
/* 16690 */ {(10<<2)|2,{115,99,0}},
/* 16691 */ {(10<<2)|2,{115,99,0}},
/* 16692 */ {(10<<2)|2,{115,99,0}},
/* 16693 */ {(10<<2)|2,{115,99,0}},
/* 16694 */ {(10<<2)|2,{115,99,0}},
/* 16695 */ {(10<<2)|2,{115,99,0}},
/* 16696 */ {(10<<2)|2,{115,99,0}},
/* 16697 */ {(10<<2)|2,{115,99,0}},
/* 16698 */ {(10<<2)|2,{115,99,0}},
/* 16699 */ {(10<<2)|2,{115,99,0}},
/* 16700 */ {(10<<2)|2,{115,99,0}},
/* 16701 */ {(10<<2)|2,{115,99,0}},
/* 16702 */ {(10<<2)|2,{115,99,0}},
/* 16703 */ {(10<<2)|2,{115,99,0}},
/* 16704 */ {(15<<2)|3,{115,101,48}},
/* 16705 */ {(15<<2)|3,{115,101,48}},
/* 16706 */ {(15<<2)|3,{115,101,49}},
/* 16707 */ {(15<<2)|3,{115,101,49}},
/* 16708 */ {(15<<2)|3,{115,101,50}},
/* 16709 */ {(15<<2)|3,{115,101,50}},
/* 16710 */ {(15<<2)|3,{115,101,97}},
/* 16711 */ {(15<<2)|3,{115,101,97}},
/* 16712 */ {(15<<2)|3,{115,101,99}},
/* 16713 */ {(15<<2)|3,{115,101,99}},
/* 16714 */ {(15<<2)|3,{115,101,101}},
/* 16715 */ {(15<<2)|3,{115,101,101}},
/* 16716 */ {(15<<2)|3,{115,101,105}},
/* 16717 */ {(15<<2)|3,{115,101,105}},
/* 16718 */ {(15<<2)|3,{115,101,111}},
/* 16719 */ {(15<<2)|3,{115,101,111}},
/* 16720 */ {(15<<2)|3,{115,101,115}},
/* 16721 */ {(15<<2)|3,{115,101,115}},
/* 16722 */ {(15<<2)|3,{115,101,116}},
/* 16723 */ {(15<<2)|3,{115,101,116}},
/* 16724 */ {(16<<2)|3,{115,101,32}},
/* 16725 */ {(16<<2)|3,{115,101,37}},
/* 16726 */ {(16<<2)|3,{115,101,45}},
/* 16727 */ {(16<<2)|3,{115,101,46}},
/* 16728 */ {(16<<2)|3,{115,101,47}},
/* 16729 */ {(16<<2)|3,{115,101,51}},
/* 16730 */ {(16<<2)|3,{115,101,52}},
/* 16731 */ {(16<<2)|3,{115,101,53}},
/* 16732 */ {(16<<2)|3,{115,101,54}},
/* 16733 */ {(16<<2)|3,{115,101,55}},
/* 16734 */ {(16<<2)|3,{115,101,56}},
/* 16735 */ {(16<<2)|3,{115,101,57}},
/* 16736 */ {(16<<2)|3,{115,101,61}},
/* 16737 */ {(16<<2)|3,{115,101,65}},
/* 16738 */ {(16<<2)|3,{115,101,95}},
/* 16739 */ {(16<<2)|3,{115,101,98}},
/* 16740 */ {(16<<2)|3,{115,101,100}},
/* 16741 */ {(16<<2)|3,{115,101,102}},
/* 16742 */ {(16<<2)|3,{115,101,103}},
/* 16743 */ {(16<<2)|3,{115,101,104}},
/* 16744 */ {(16<<2)|3,{115,101,108}},
/* 16745 */ {(16<<2)|3,{115,101,109}},
/* 16746 */ {(16<<2)|3,{115,101,110}},
/* 16747 */ {(16<<2)|3,{115,101,112}},
/* 16748 */ {(16<<2)|3,{115,101,114}},
/* 16749 */ {(16<<2)|3,{115,101,117}},
/* 16750 */ {(10<<2)|2,{115,101,0}},
/* 16751 */ {(10<<2)|2,{115,101,0}},
/* 16752 */ {(10<<2)|2,{115,101,0}},
/* 16753 */ {(10<<2)|2,{115,101,0}},
/* 16754 */ {(10<<2)|2,{115,101,0}},
/* 16755 */ {(10<<2)|2,{115,101,0}},
/* 16756 */ {(10<<2)|2,{115,101,0}},
/* 16757 */ {(10<<2)|2,{115,101,0}},
/* 16758 */ {(10<<2)|2,{115,101,0}},
/* 16759 */ {(10<<2)|2,{115,101,0}},
/* 16760 */ {(10<<2)|2,{115,101,0}},
/* 16761 */ {(10<<2)|2,{115,101,0}},
/* 16762 */ {(10<<2)|2,{115,101,0}},
/* 16763 */ {(10<<2)|2,{115,101,0}},
/* 16764 */ {(10<<2)|2,{115,101,0}},
/* 16765 */ {(10<<2)|2,{115,101,0}},
/* 16766 */ {(10<<2)|2,{115,101,0}},
/* 16767 */ {(10<<2)|2,{115,101,0}},
/* 16768 */ {(15<<2)|3,{115,105,48}},
/* 16769 */ {(15<<2)|3,{115,105,48}},
/* 16770 */ {(15<<2)|3,{115,105,49}},
/* 16771 */ {(15<<2)|3,{115,105,49}},
/* 16772 */ {(15<<2)|3,{115,105,50}},
/* 16773 */ {(15<<2)|3,{115,105,50}},
/* 16774 */ {(15<<2)|3,{115,105,97}},
/* 16775 */ {(15<<2)|3,{115,105,97}},
/* 16776 */ {(15<<2)|3,{115,105,99}},
/* 16777 */ {(15<<2)|3,{115,105,99}},
/* 16778 */ {(15<<2)|3,{115,105,101}},
/* 16779 */ {(15<<2)|3,{115,105,101}},
/* 16780 */ {(15<<2)|3,{115,105,105}},
/* 16781 */ {(15<<2)|3,{115,105,105}},
/* 16782 */ {(15<<2)|3,{115,105,111}},
/* 16783 */ {(15<<2)|3,{115,105,111}},
/* 16784 */ {(15<<2)|3,{115,105,115}},
/* 16785 */ {(15<<2)|3,{115,105,115}},
/* 16786 */ {(15<<2)|3,{115,105,116}},
/* 16787 */ {(15<<2)|3,{115,105,116}},
/* 16788 */ {(16<<2)|3,{115,105,32}},
/* 16789 */ {(16<<2)|3,{115,105,37}},
/* 16790 */ {(16<<2)|3,{115,105,45}},
/* 16791 */ {(16<<2)|3,{115,105,46}},
/* 16792 */ {(16<<2)|3,{115,105,47}},
/* 16793 */ {(16<<2)|3,{115,105,51}},
/* 16794 */ {(16<<2)|3,{115,105,52}},
/* 16795 */ {(16<<2)|3,{115,105,53}},
/* 16796 */ {(16<<2)|3,{115,105,54}},
/* 16797 */ {(16<<2)|3,{115,105,55}},
/* 16798 */ {(16<<2)|3,{115,105,56}},
/* 16799 */ {(16<<2)|3,{115,105,57}},
/* 16800 */ {(16<<2)|3,{115,105,61}},
/* 16801 */ {(16<<2)|3,{115,105,65}},
/* 16802 */ {(16<<2)|3,{115,105,95}},
/* 16803 */ {(16<<2)|3,{115,105,98}},
/* 16804 */ {(16<<2)|3,{115,105,100}},
/* 16805 */ {(16<<2)|3,{115,105,102}},
/* 16806 */ {(16<<2)|3,{115,105,103}},
/* 16807 */ {(16<<2)|3,{115,105,104}},
/* 16808 */ {(16<<2)|3,{115,105,108}},
/* 16809 */ {(16<<2)|3,{115,105,109}},
/* 16810 */ {(16<<2)|3,{115,105,110}},
/* 16811 */ {(16<<2)|3,{115,105,112}},
/* 16812 */ {(16<<2)|3,{115,105,114}},
/* 16813 */ {(16<<2)|3,{115,105,117}},
/* 16814 */ {(10<<2)|2,{115,105,0}},
/* 16815 */ {(10<<2)|2,{115,105,0}},
/* 16816 */ {(10<<2)|2,{115,105,0}},
/* 16817 */ {(10<<2)|2,{115,105,0}},
/* 16818 */ {(10<<2)|2,{115,105,0}},
/* 16819 */ {(10<<2)|2,{115,105,0}},
/* 16820 */ {(10<<2)|2,{115,105,0}},
/* 16821 */ {(10<<2)|2,{115,105,0}},
/* 16822 */ {(10<<2)|2,{115,105,0}},
/* 16823 */ {(10<<2)|2,{115,105,0}},
/* 16824 */ {(10<<2)|2,{115,105,0}},
/* 16825 */ {(10<<2)|2,{115,105,0}},
/* 16826 */ {(10<<2)|2,{115,105,0}},
/* 16827 */ {(10<<2)|2,{115,105,0}},
/* 16828 */ {(10<<2)|2,{115,105,0}},
/* 16829 */ {(10<<2)|2,{115,105,0}},
/* 16830 */ {(10<<2)|2,{115,105,0}},
/* 16831 */ {(10<<2)|2,{115,105,0}},
/* 16832 */ {(15<<2)|3,{115,111,48}},
/* 16833 */ {(15<<2)|3,{115,111,48}},
/* 16834 */ {(15<<2)|3,{115,111,49}},
/* 16835 */ {(15<<2)|3,{115,111,49}},
/* 16836 */ {(15<<2)|3,{115,111,50}},
/* 16837 */ {(15<<2)|3,{115,111,50}},
/* 16838 */ {(15<<2)|3,{115,111,97}},
/* 16839 */ {(15<<2)|3,{115,111,97}},
/* 16840 */ {(15<<2)|3,{115,111,99}},
/* 16841 */ {(15<<2)|3,{115,111,99}},
/* 16842 */ {(15<<2)|3,{115,111,101}},
/* 16843 */ {(15<<2)|3,{115,111,101}},
/* 16844 */ {(15<<2)|3,{115,111,105}},
/* 16845 */ {(15<<2)|3,{115,111,105}},
/* 16846 */ {(15<<2)|3,{115,111,111}},
/* 16847 */ {(15<<2)|3,{115,111,111}},
/* 16848 */ {(15<<2)|3,{115,111,115}},
/* 16849 */ {(15<<2)|3,{115,111,115}},
/* 16850 */ {(15<<2)|3,{115,111,116}},
/* 16851 */ {(15<<2)|3,{115,111,116}},
/* 16852 */ {(16<<2)|3,{115,111,32}},
/* 16853 */ {(16<<2)|3,{115,111,37}},
/* 16854 */ {(16<<2)|3,{115,111,45}},
/* 16855 */ {(16<<2)|3,{115,111,46}},
/* 16856 */ {(16<<2)|3,{115,111,47}},
/* 16857 */ {(16<<2)|3,{115,111,51}},
/* 16858 */ {(16<<2)|3,{115,111,52}},
/* 16859 */ {(16<<2)|3,{115,111,53}},
/* 16860 */ {(16<<2)|3,{115,111,54}},
/* 16861 */ {(16<<2)|3,{115,111,55}},
/* 16862 */ {(16<<2)|3,{115,111,56}},
/* 16863 */ {(16<<2)|3,{115,111,57}},
/* 16864 */ {(16<<2)|3,{115,111,61}},
/* 16865 */ {(16<<2)|3,{115,111,65}},
/* 16866 */ {(16<<2)|3,{115,111,95}},
/* 16867 */ {(16<<2)|3,{115,111,98}},
/* 16868 */ {(16<<2)|3,{115,111,100}},
/* 16869 */ {(16<<2)|3,{115,111,102}},
/* 16870 */ {(16<<2)|3,{115,111,103}},
/* 16871 */ {(16<<2)|3,{115,111,104}},
/* 16872 */ {(16<<2)|3,{115,111,108}},
/* 16873 */ {(16<<2)|3,{115,111,109}},
/* 16874 */ {(16<<2)|3,{115,111,110}},
/* 16875 */ {(16<<2)|3,{115,111,112}},
/* 16876 */ {(16<<2)|3,{115,111,114}},
/* 16877 */ {(16<<2)|3,{115,111,117}},
/* 16878 */ {(10<<2)|2,{115,111,0}},
/* 16879 */ {(10<<2)|2,{115,111,0}},
/* 16880 */ {(10<<2)|2,{115,111,0}},
/* 16881 */ {(10<<2)|2,{115,111,0}},
/* 16882 */ {(10<<2)|2,{115,111,0}},
/* 16883 */ {(10<<2)|2,{115,111,0}},
/* 16884 */ {(10<<2)|2,{115,111,0}},
/* 16885 */ {(10<<2)|2,{115,111,0}},
/* 16886 */ {(10<<2)|2,{115,111,0}},
/* 16887 */ {(10<<2)|2,{115,111,0}},
/* 16888 */ {(10<<2)|2,{115,111,0}},
/* 16889 */ {(10<<2)|2,{115,111,0}},
/* 16890 */ {(10<<2)|2,{115,111,0}},
/* 16891 */ {(10<<2)|2,{115,111,0}},
/* 16892 */ {(10<<2)|2,{115,111,0}},
/* 16893 */ {(10<<2)|2,{115,111,0}},
/* 16894 */ {(10<<2)|2,{115,111,0}},
/* 16895 */ {(10<<2)|2,{115,111,0}},
/* 16896 */ {(15<<2)|3,{115,115,48}},
/* 16897 */ {(15<<2)|3,{115,115,48}},
/* 16898 */ {(15<<2)|3,{115,115,49}},
/* 16899 */ {(15<<2)|3,{115,115,49}},
/* 16900 */ {(15<<2)|3,{115,115,50}},
/* 16901 */ {(15<<2)|3,{115,115,50}},
/* 16902 */ {(15<<2)|3,{115,115,97}},
/* 16903 */ {(15<<2)|3,{115,115,97}},
/* 16904 */ {(15<<2)|3,{115,115,99}},
/* 16905 */ {(15<<2)|3,{115,115,99}},
/* 16906 */ {(15<<2)|3,{115,115,101}},
/* 16907 */ {(15<<2)|3,{115,115,101}},
/* 16908 */ {(15<<2)|3,{115,115,105}},
/* 16909 */ {(15<<2)|3,{115,115,105}},
/* 16910 */ {(15<<2)|3,{115,115,111}},
/* 16911 */ {(15<<2)|3,{115,115,111}},
/* 16912 */ {(15<<2)|3,{115,115,115}},
/* 16913 */ {(15<<2)|3,{115,115,115}},
/* 16914 */ {(15<<2)|3,{115,115,116}},
/* 16915 */ {(15<<2)|3,{115,115,116}},
/* 16916 */ {(16<<2)|3,{115,115,32}},
/* 16917 */ {(16<<2)|3,{115,115,37}},
/* 16918 */ {(16<<2)|3,{115,115,45}},
/* 16919 */ {(16<<2)|3,{115,115,46}},
/* 16920 */ {(16<<2)|3,{115,115,47}},
/* 16921 */ {(16<<2)|3,{115,115,51}},
/* 16922 */ {(16<<2)|3,{115,115,52}},
/* 16923 */ {(16<<2)|3,{115,115,53}},
/* 16924 */ {(16<<2)|3,{115,115,54}},
/* 16925 */ {(16<<2)|3,{115,115,55}},
/* 16926 */ {(16<<2)|3,{115,115,56}},
/* 16927 */ {(16<<2)|3,{115,115,57}},
/* 16928 */ {(16<<2)|3,{115,115,61}},
/* 16929 */ {(16<<2)|3,{115,115,65}},
/* 16930 */ {(16<<2)|3,{115,115,95}},
/* 16931 */ {(16<<2)|3,{115,115,98}},
/* 16932 */ {(16<<2)|3,{115,115,100}},
/* 16933 */ {(16<<2)|3,{115,115,102}},
/* 16934 */ {(16<<2)|3,{115,115,103}},
/* 16935 */ {(16<<2)|3,{115,115,104}},
/* 16936 */ {(16<<2)|3,{115,115,108}},
/* 16937 */ {(16<<2)|3,{115,115,109}},
/* 16938 */ {(16<<2)|3,{115,115,110}},
/* 16939 */ {(16<<2)|3,{115,115,112}},
/* 16940 */ {(16<<2)|3,{115,115,114}},
/* 16941 */ {(16<<2)|3,{115,115,117}},
/* 16942 */ {(10<<2)|2,{115,115,0}},
/* 16943 */ {(10<<2)|2,{115,115,0}},
/* 16944 */ {(10<<2)|2,{115,115,0}},
/* 16945 */ {(10<<2)|2,{115,115,0}},
/* 16946 */ {(10<<2)|2,{115,115,0}},
/* 16947 */ {(10<<2)|2,{115,115,0}},
/* 16948 */ {(10<<2)|2,{115,115,0}},
/* 16949 */ {(10<<2)|2,{115,115,0}},
/* 16950 */ {(10<<2)|2,{115,115,0}},
/* 16951 */ {(10<<2)|2,{115,115,0}},
/* 16952 */ {(10<<2)|2,{115,115,0}},
/* 16953 */ {(10<<2)|2,{115,115,0}},
/* 16954 */ {(10<<2)|2,{115,115,0}},
/* 16955 */ {(10<<2)|2,{115,115,0}},
/* 16956 */ {(10<<2)|2,{115,115,0}},
/* 16957 */ {(10<<2)|2,{115,115,0}},
/* 16958 */ {(10<<2)|2,{115,115,0}},
/* 16959 */ {(10<<2)|2,{115,115,0}},
/* 16960 */ {(15<<2)|3,{115,116,48}},
/* 16961 */ {(15<<2)|3,{115,116,48}},
/* 16962 */ {(15<<2)|3,{115,116,49}},
/* 16963 */ {(15<<2)|3,{115,116,49}},
/* 16964 */ {(15<<2)|3,{115,116,50}},
/* 16965 */ {(15<<2)|3,{115,116,50}},
/* 16966 */ {(15<<2)|3,{115,116,97}},
/* 16967 */ {(15<<2)|3,{115,116,97}},
/* 16968 */ {(15<<2)|3,{115,116,99}},
/* 16969 */ {(15<<2)|3,{115,116,99}},
/* 16970 */ {(15<<2)|3,{115,116,101}},
/* 16971 */ {(15<<2)|3,{115,116,101}},
/* 16972 */ {(15<<2)|3,{115,116,105}},
/* 16973 */ {(15<<2)|3,{115,116,105}},
/* 16974 */ {(15<<2)|3,{115,116,111}},
/* 16975 */ {(15<<2)|3,{115,116,111}},
/* 16976 */ {(15<<2)|3,{115,116,115}},
/* 16977 */ {(15<<2)|3,{115,116,115}},
/* 16978 */ {(15<<2)|3,{115,116,116}},
/* 16979 */ {(15<<2)|3,{115,116,116}},
/* 16980 */ {(16<<2)|3,{115,116,32}},
/* 16981 */ {(16<<2)|3,{115,116,37}},
/* 16982 */ {(16<<2)|3,{115,116,45}},
/* 16983 */ {(16<<2)|3,{115,116,46}},
/* 16984 */ {(16<<2)|3,{115,116,47}},
/* 16985 */ {(16<<2)|3,{115,116,51}},
/* 16986 */ {(16<<2)|3,{115,116,52}},
/* 16987 */ {(16<<2)|3,{115,116,53}},
/* 16988 */ {(16<<2)|3,{115,116,54}},
/* 16989 */ {(16<<2)|3,{115,116,55}},
/* 16990 */ {(16<<2)|3,{115,116,56}},
/* 16991 */ {(16<<2)|3,{115,116,57}},
/* 16992 */ {(16<<2)|3,{115,116,61}},
/* 16993 */ {(16<<2)|3,{115,116,65}},
/* 16994 */ {(16<<2)|3,{115,116,95}},
/* 16995 */ {(16<<2)|3,{115,116,98}},
/* 16996 */ {(16<<2)|3,{115,116,100}},
/* 16997 */ {(16<<2)|3,{115,116,102}},
/* 16998 */ {(16<<2)|3,{115,116,103}},
/* 16999 */ {(16<<2)|3,{115,116,104}},
/* 17000 */ {(16<<2)|3,{115,116,108}},
/* 17001 */ {(16<<2)|3,{115,116,109}},
/* 17002 */ {(16<<2)|3,{115,116,110}},
/* 17003 */ {(16<<2)|3,{115,116,112}},
/* 17004 */ {(16<<2)|3,{115,116,114}},
/* 17005 */ {(16<<2)|3,{115,116,117}},
/* 17006 */ {(10<<2)|2,{115,116,0}},
/* 17007 */ {(10<<2)|2,{115,116,0}},
/* 17008 */ {(10<<2)|2,{115,116,0}},
/* 17009 */ {(10<<2)|2,{115,116,0}},
/* 17010 */ {(10<<2)|2,{115,116,0}},
/* 17011 */ {(10<<2)|2,{115,116,0}},
/* 17012 */ {(10<<2)|2,{115,116,0}},
/* 17013 */ {(10<<2)|2,{115,116,0}},
/* 17014 */ {(10<<2)|2,{115,116,0}},
/* 17015 */ {(10<<2)|2,{115,116,0}},
/* 17016 */ {(10<<2)|2,{115,116,0}},
/* 17017 */ {(10<<2)|2,{115,116,0}},
/* 17018 */ {(10<<2)|2,{115,116,0}},
/* 17019 */ {(10<<2)|2,{115,116,0}},
/* 17020 */ {(10<<2)|2,{115,116,0}},
/* 17021 */ {(10<<2)|2,{115,116,0}},
/* 17022 */ {(10<<2)|2,{115,116,0}},
/* 17023 */ {(10<<2)|2,{115,116,0}},
/* 17024 */ {(16<<2)|3,{115,32,48}},
/* 17025 */ {(16<<2)|3,{115,32,49}},
/* 17026 */ {(16<<2)|3,{115,32,50}},
/* 17027 */ {(16<<2)|3,{115,32,97}},
/* 17028 */ {(16<<2)|3,{115,32,99}},
/* 17029 */ {(16<<2)|3,{115,32,101}},
/* 17030 */ {(16<<2)|3,{115,32,105}},
/* 17031 */ {(16<<2)|3,{115,32,111}},
/* 17032 */ {(16<<2)|3,{115,32,115}},
/* 17033 */ {(16<<2)|3,{115,32,116}},
/* 17034 */ {(11<<2)|2,{115,32,0}},
/* 17035 */ {(11<<2)|2,{115,32,0}},
/* 17036 */ {(11<<2)|2,{115,32,0}},
/* 17037 */ {(11<<2)|2,{115,32,0}},
/* 17038 */ {(11<<2)|2,{115,32,0}},
/* 17039 */ {(11<<2)|2,{115,32,0}},
/* 17040 */ {(11<<2)|2,{115,32,0}},
/* 17041 */ {(11<<2)|2,{115,32,0}},
/* 17042 */ {(11<<2)|2,{115,32,0}},
/* 17043 */ {(11<<2)|2,{115,32,0}},
/* 17044 */ {(11<<2)|2,{115,32,0}},
/* 17045 */ {(11<<2)|2,{115,32,0}},
/* 17046 */ {(11<<2)|2,{115,32,0}},
/* 17047 */ {(11<<2)|2,{115,32,0}},
/* 17048 */ {(11<<2)|2,{115,32,0}},
/* 17049 */ {(11<<2)|2,{115,32,0}},
/* 17050 */ {(11<<2)|2,{115,32,0}},
/* 17051 */ {(11<<2)|2,{115,32,0}},
/* 17052 */ {(11<<2)|2,{115,32,0}},
/* 17053 */ {(11<<2)|2,{115,32,0}},
/* 17054 */ {(11<<2)|2,{115,32,0}},
/* 17055 */ {(11<<2)|2,{115,32,0}},
/* 17056 */ {(16<<2)|3,{115,37,48}},
/* 17057 */ {(16<<2)|3,{115,37,49}},
/* 17058 */ {(16<<2)|3,{115,37,50}},
/* 17059 */ {(16<<2)|3,{115,37,97}},
/* 17060 */ {(16<<2)|3,{115,37,99}},
/* 17061 */ {(16<<2)|3,{115,37,101}},
/* 17062 */ {(16<<2)|3,{115,37,105}},
/* 17063 */ {(16<<2)|3,{115,37,111}},
/* 17064 */ {(16<<2)|3,{115,37,115}},
/* 17065 */ {(16<<2)|3,{115,37,116}},
/* 17066 */ {(11<<2)|2,{115,37,0}},
/* 17067 */ {(11<<2)|2,{115,37,0}},
/* 17068 */ {(11<<2)|2,{115,37,0}},
/* 17069 */ {(11<<2)|2,{115,37,0}},
/* 17070 */ {(11<<2)|2,{115,37,0}},
/* 17071 */ {(11<<2)|2,{115,37,0}},
/* 17072 */ {(11<<2)|2,{115,37,0}},
/* 17073 */ {(11<<2)|2,{115,37,0}},
/* 17074 */ {(11<<2)|2,{115,37,0}},
/* 17075 */ {(11<<2)|2,{115,37,0}},
/* 17076 */ {(11<<2)|2,{115,37,0}},
/* 17077 */ {(11<<2)|2,{115,37,0}},
/* 17078 */ {(11<<2)|2,{115,37,0}},
/* 17079 */ {(11<<2)|2,{115,37,0}},
/* 17080 */ {(11<<2)|2,{115,37,0}},
/* 17081 */ {(11<<2)|2,{115,37,0}},
/* 17082 */ {(11<<2)|2,{115,37,0}},
/* 17083 */ {(11<<2)|2,{115,37,0}},
/* 17084 */ {(11<<2)|2,{115,37,0}},
/* 17085 */ {(11<<2)|2,{115,37,0}},
/* 17086 */ {(11<<2)|2,{115,37,0}},
/* 17087 */ {(11<<2)|2,{115,37,0}},
/* 17088 */ {(16<<2)|3,{115,45,48}},
/* 17089 */ {(16<<2)|3,{115,45,49}},
/* 17090 */ {(16<<2)|3,{115,45,50}},
/* 17091 */ {(16<<2)|3,{115,45,97}},
/* 17092 */ {(16<<2)|3,{115,45,99}},
/* 17093 */ {(16<<2)|3,{115,45,101}},
/* 17094 */ {(16<<2)|3,{115,45,105}},
/* 17095 */ {(16<<2)|3,{115,45,111}},
/* 17096 */ {(16<<2)|3,{115,45,115}},
/* 17097 */ {(16<<2)|3,{115,45,116}},
/* 17098 */ {(11<<2)|2,{115,45,0}},
/* 17099 */ {(11<<2)|2,{115,45,0}},
/* 17100 */ {(11<<2)|2,{115,45,0}},
/* 17101 */ {(11<<2)|2,{115,45,0}},
/* 17102 */ {(11<<2)|2,{115,45,0}},
/* 17103 */ {(11<<2)|2,{115,45,0}},
/* 17104 */ {(11<<2)|2,{115,45,0}},
/* 17105 */ {(11<<2)|2,{115,45,0}},
/* 17106 */ {(11<<2)|2,{115,45,0}},
/* 17107 */ {(11<<2)|2,{115,45,0}},
/* 17108 */ {(11<<2)|2,{115,45,0}},
/* 17109 */ {(11<<2)|2,{115,45,0}},
/* 17110 */ {(11<<2)|2,{115,45,0}},
/* 17111 */ {(11<<2)|2,{115,45,0}},
/* 17112 */ {(11<<2)|2,{115,45,0}},
/* 17113 */ {(11<<2)|2,{115,45,0}},
/* 17114 */ {(11<<2)|2,{115,45,0}},
/* 17115 */ {(11<<2)|2,{115,45,0}},
/* 17116 */ {(11<<2)|2,{115,45,0}},
/* 17117 */ {(11<<2)|2,{115,45,0}},
/* 17118 */ {(11<<2)|2,{115,45,0}},
/* 17119 */ {(11<<2)|2,{115,45,0}},
/* 17120 */ {(16<<2)|3,{115,46,48}},
/* 17121 */ {(16<<2)|3,{115,46,49}},
/* 17122 */ {(16<<2)|3,{115,46,50}},
/* 17123 */ {(16<<2)|3,{115,46,97}},
/* 17124 */ {(16<<2)|3,{115,46,99}},
/* 17125 */ {(16<<2)|3,{115,46,101}},
/* 17126 */ {(16<<2)|3,{115,46,105}},
/* 17127 */ {(16<<2)|3,{115,46,111}},
/* 17128 */ {(16<<2)|3,{115,46,115}},
/* 17129 */ {(16<<2)|3,{115,46,116}},
/* 17130 */ {(11<<2)|2,{115,46,0}},
/* 17131 */ {(11<<2)|2,{115,46,0}},
/* 17132 */ {(11<<2)|2,{115,46,0}},
/* 17133 */ {(11<<2)|2,{115,46,0}},
/* 17134 */ {(11<<2)|2,{115,46,0}},
/* 17135 */ {(11<<2)|2,{115,46,0}},
/* 17136 */ {(11<<2)|2,{115,46,0}},
/* 17137 */ {(11<<2)|2,{115,46,0}},
/* 17138 */ {(11<<2)|2,{115,46,0}},
/* 17139 */ {(11<<2)|2,{115,46,0}},
/* 17140 */ {(11<<2)|2,{115,46,0}},
/* 17141 */ {(11<<2)|2,{115,46,0}},
/* 17142 */ {(11<<2)|2,{115,46,0}},
/* 17143 */ {(11<<2)|2,{115,46,0}},
/* 17144 */ {(11<<2)|2,{115,46,0}},
/* 17145 */ {(11<<2)|2,{115,46,0}},
/* 17146 */ {(11<<2)|2,{115,46,0}},
/* 17147 */ {(11<<2)|2,{115,46,0}},
/* 17148 */ {(11<<2)|2,{115,46,0}},
/* 17149 */ {(11<<2)|2,{115,46,0}},
/* 17150 */ {(11<<2)|2,{115,46,0}},
/* 17151 */ {(11<<2)|2,{115,46,0}},
/* 17152 */ {(16<<2)|3,{115,47,48}},
/* 17153 */ {(16<<2)|3,{115,47,49}},
/* 17154 */ {(16<<2)|3,{115,47,50}},
/* 17155 */ {(16<<2)|3,{115,47,97}},
/* 17156 */ {(16<<2)|3,{115,47,99}},
/* 17157 */ {(16<<2)|3,{115,47,101}},
/* 17158 */ {(16<<2)|3,{115,47,105}},
/* 17159 */ {(16<<2)|3,{115,47,111}},
/* 17160 */ {(16<<2)|3,{115,47,115}},
/* 17161 */ {(16<<2)|3,{115,47,116}},
/* 17162 */ {(11<<2)|2,{115,47,0}},
/* 17163 */ {(11<<2)|2,{115,47,0}},
/* 17164 */ {(11<<2)|2,{115,47,0}},
/* 17165 */ {(11<<2)|2,{115,47,0}},
/* 17166 */ {(11<<2)|2,{115,47,0}},
/* 17167 */ {(11<<2)|2,{115,47,0}},
/* 17168 */ {(11<<2)|2,{115,47,0}},
/* 17169 */ {(11<<2)|2,{115,47,0}},
/* 17170 */ {(11<<2)|2,{115,47,0}},
/* 17171 */ {(11<<2)|2,{115,47,0}},
/* 17172 */ {(11<<2)|2,{115,47,0}},
/* 17173 */ {(11<<2)|2,{115,47,0}},
/* 17174 */ {(11<<2)|2,{115,47,0}},
/* 17175 */ {(11<<2)|2,{115,47,0}},
/* 17176 */ {(11<<2)|2,{115,47,0}},
/* 17177 */ {(11<<2)|2,{115,47,0}},
/* 17178 */ {(11<<2)|2,{115,47,0}},
/* 17179 */ {(11<<2)|2,{115,47,0}},
/* 17180 */ {(11<<2)|2,{115,47,0}},
/* 17181 */ {(11<<2)|2,{115,47,0}},
/* 17182 */ {(11<<2)|2,{115,47,0}},
/* 17183 */ {(11<<2)|2,{115,47,0}},
/* 17184 */ {(16<<2)|3,{115,51,48}},
/* 17185 */ {(16<<2)|3,{115,51,49}},
/* 17186 */ {(16<<2)|3,{115,51,50}},
/* 17187 */ {(16<<2)|3,{115,51,97}},
/* 17188 */ {(16<<2)|3,{115,51,99}},
/* 17189 */ {(16<<2)|3,{115,51,101}},
/* 17190 */ {(16<<2)|3,{115,51,105}},
/* 17191 */ {(16<<2)|3,{115,51,111}},
/* 17192 */ {(16<<2)|3,{115,51,115}},
/* 17193 */ {(16<<2)|3,{115,51,116}},
/* 17194 */ {(11<<2)|2,{115,51,0}},
/* 17195 */ {(11<<2)|2,{115,51,0}},
/* 17196 */ {(11<<2)|2,{115,51,0}},
/* 17197 */ {(11<<2)|2,{115,51,0}},
/* 17198 */ {(11<<2)|2,{115,51,0}},
/* 17199 */ {(11<<2)|2,{115,51,0}},
/* 17200 */ {(11<<2)|2,{115,51,0}},
/* 17201 */ {(11<<2)|2,{115,51,0}},
/* 17202 */ {(11<<2)|2,{115,51,0}},
/* 17203 */ {(11<<2)|2,{115,51,0}},
/* 17204 */ {(11<<2)|2,{115,51,0}},
/* 17205 */ {(11<<2)|2,{115,51,0}},
/* 17206 */ {(11<<2)|2,{115,51,0}},
/* 17207 */ {(11<<2)|2,{115,51,0}},
/* 17208 */ {(11<<2)|2,{115,51,0}},
/* 17209 */ {(11<<2)|2,{115,51,0}},
/* 17210 */ {(11<<2)|2,{115,51,0}},
/* 17211 */ {(11<<2)|2,{115,51,0}},
/* 17212 */ {(11<<2)|2,{115,51,0}},
/* 17213 */ {(11<<2)|2,{115,51,0}},
/* 17214 */ {(11<<2)|2,{115,51,0}},
/* 17215 */ {(11<<2)|2,{115,51,0}},
/* 17216 */ {(16<<2)|3,{115,52,48}},
/* 17217 */ {(16<<2)|3,{115,52,49}},
/* 17218 */ {(16<<2)|3,{115,52,50}},
/* 17219 */ {(16<<2)|3,{115,52,97}},
/* 17220 */ {(16<<2)|3,{115,52,99}},
/* 17221 */ {(16<<2)|3,{115,52,101}},
/* 17222 */ {(16<<2)|3,{115,52,105}},
/* 17223 */ {(16<<2)|3,{115,52,111}},
/* 17224 */ {(16<<2)|3,{115,52,115}},
/* 17225 */ {(16<<2)|3,{115,52,116}},
/* 17226 */ {(11<<2)|2,{115,52,0}},
/* 17227 */ {(11<<2)|2,{115,52,0}},
/* 17228 */ {(11<<2)|2,{115,52,0}},
/* 17229 */ {(11<<2)|2,{115,52,0}},
/* 17230 */ {(11<<2)|2,{115,52,0}},
/* 17231 */ {(11<<2)|2,{115,52,0}},
/* 17232 */ {(11<<2)|2,{115,52,0}},
/* 17233 */ {(11<<2)|2,{115,52,0}},
/* 17234 */ {(11<<2)|2,{115,52,0}},
/* 17235 */ {(11<<2)|2,{115,52,0}},
/* 17236 */ {(11<<2)|2,{115,52,0}},
/* 17237 */ {(11<<2)|2,{115,52,0}},
/* 17238 */ {(11<<2)|2,{115,52,0}},
/* 17239 */ {(11<<2)|2,{115,52,0}},
/* 17240 */ {(11<<2)|2,{115,52,0}},
/* 17241 */ {(11<<2)|2,{115,52,0}},
/* 17242 */ {(11<<2)|2,{115,52,0}},
/* 17243 */ {(11<<2)|2,{115,52,0}},
/* 17244 */ {(11<<2)|2,{115,52,0}},
/* 17245 */ {(11<<2)|2,{115,52,0}},
/* 17246 */ {(11<<2)|2,{115,52,0}},
/* 17247 */ {(11<<2)|2,{115,52,0}},
/* 17248 */ {(16<<2)|3,{115,53,48}},
/* 17249 */ {(16<<2)|3,{115,53,49}},
/* 17250 */ {(16<<2)|3,{115,53,50}},
/* 17251 */ {(16<<2)|3,{115,53,97}},
/* 17252 */ {(16<<2)|3,{115,53,99}},
/* 17253 */ {(16<<2)|3,{115,53,101}},
/* 17254 */ {(16<<2)|3,{115,53,105}},
/* 17255 */ {(16<<2)|3,{115,53,111}},
/* 17256 */ {(16<<2)|3,{115,53,115}},
/* 17257 */ {(16<<2)|3,{115,53,116}},
/* 17258 */ {(11<<2)|2,{115,53,0}},
/* 17259 */ {(11<<2)|2,{115,53,0}},
/* 17260 */ {(11<<2)|2,{115,53,0}},
/* 17261 */ {(11<<2)|2,{115,53,0}},
/* 17262 */ {(11<<2)|2,{115,53,0}},
/* 17263 */ {(11<<2)|2,{115,53,0}},
/* 17264 */ {(11<<2)|2,{115,53,0}},
/* 17265 */ {(11<<2)|2,{115,53,0}},
/* 17266 */ {(11<<2)|2,{115,53,0}},
/* 17267 */ {(11<<2)|2,{115,53,0}},
/* 17268 */ {(11<<2)|2,{115,53,0}},
/* 17269 */ {(11<<2)|2,{115,53,0}},
/* 17270 */ {(11<<2)|2,{115,53,0}},
/* 17271 */ {(11<<2)|2,{115,53,0}},
/* 17272 */ {(11<<2)|2,{115,53,0}},
/* 17273 */ {(11<<2)|2,{115,53,0}},
/* 17274 */ {(11<<2)|2,{115,53,0}},
/* 17275 */ {(11<<2)|2,{115,53,0}},
/* 17276 */ {(11<<2)|2,{115,53,0}},
/* 17277 */ {(11<<2)|2,{115,53,0}},
/* 17278 */ {(11<<2)|2,{115,53,0}},
/* 17279 */ {(11<<2)|2,{115,53,0}},
/* 17280 */ {(16<<2)|3,{115,54,48}},
/* 17281 */ {(16<<2)|3,{115,54,49}},
/* 17282 */ {(16<<2)|3,{115,54,50}},
/* 17283 */ {(16<<2)|3,{115,54,97}},
/* 17284 */ {(16<<2)|3,{115,54,99}},
/* 17285 */ {(16<<2)|3,{115,54,101}},
/* 17286 */ {(16<<2)|3,{115,54,105}},
/* 17287 */ {(16<<2)|3,{115,54,111}},
/* 17288 */ {(16<<2)|3,{115,54,115}},
/* 17289 */ {(16<<2)|3,{115,54,116}},
/* 17290 */ {(11<<2)|2,{115,54,0}},
/* 17291 */ {(11<<2)|2,{115,54,0}},
/* 17292 */ {(11<<2)|2,{115,54,0}},
/* 17293 */ {(11<<2)|2,{115,54,0}},
/* 17294 */ {(11<<2)|2,{115,54,0}},
/* 17295 */ {(11<<2)|2,{115,54,0}},
/* 17296 */ {(11<<2)|2,{115,54,0}},
/* 17297 */ {(11<<2)|2,{115,54,0}},
/* 17298 */ {(11<<2)|2,{115,54,0}},
/* 17299 */ {(11<<2)|2,{115,54,0}},
/* 17300 */ {(11<<2)|2,{115,54,0}},
/* 17301 */ {(11<<2)|2,{115,54,0}},
/* 17302 */ {(11<<2)|2,{115,54,0}},
/* 17303 */ {(11<<2)|2,{115,54,0}},
/* 17304 */ {(11<<2)|2,{115,54,0}},
/* 17305 */ {(11<<2)|2,{115,54,0}},
/* 17306 */ {(11<<2)|2,{115,54,0}},
/* 17307 */ {(11<<2)|2,{115,54,0}},
/* 17308 */ {(11<<2)|2,{115,54,0}},
/* 17309 */ {(11<<2)|2,{115,54,0}},
/* 17310 */ {(11<<2)|2,{115,54,0}},
/* 17311 */ {(11<<2)|2,{115,54,0}},
/* 17312 */ {(16<<2)|3,{115,55,48}},
/* 17313 */ {(16<<2)|3,{115,55,49}},
/* 17314 */ {(16<<2)|3,{115,55,50}},
/* 17315 */ {(16<<2)|3,{115,55,97}},
/* 17316 */ {(16<<2)|3,{115,55,99}},
/* 17317 */ {(16<<2)|3,{115,55,101}},
/* 17318 */ {(16<<2)|3,{115,55,105}},
/* 17319 */ {(16<<2)|3,{115,55,111}},
/* 17320 */ {(16<<2)|3,{115,55,115}},
/* 17321 */ {(16<<2)|3,{115,55,116}},
/* 17322 */ {(11<<2)|2,{115,55,0}},
/* 17323 */ {(11<<2)|2,{115,55,0}},
/* 17324 */ {(11<<2)|2,{115,55,0}},
/* 17325 */ {(11<<2)|2,{115,55,0}},
/* 17326 */ {(11<<2)|2,{115,55,0}},
/* 17327 */ {(11<<2)|2,{115,55,0}},
/* 17328 */ {(11<<2)|2,{115,55,0}},
/* 17329 */ {(11<<2)|2,{115,55,0}},
/* 17330 */ {(11<<2)|2,{115,55,0}},
/* 17331 */ {(11<<2)|2,{115,55,0}},
/* 17332 */ {(11<<2)|2,{115,55,0}},
/* 17333 */ {(11<<2)|2,{115,55,0}},
/* 17334 */ {(11<<2)|2,{115,55,0}},
/* 17335 */ {(11<<2)|2,{115,55,0}},
/* 17336 */ {(11<<2)|2,{115,55,0}},
/* 17337 */ {(11<<2)|2,{115,55,0}},
/* 17338 */ {(11<<2)|2,{115,55,0}},
/* 17339 */ {(11<<2)|2,{115,55,0}},
/* 17340 */ {(11<<2)|2,{115,55,0}},
/* 17341 */ {(11<<2)|2,{115,55,0}},
/* 17342 */ {(11<<2)|2,{115,55,0}},
/* 17343 */ {(11<<2)|2,{115,55,0}},
/* 17344 */ {(16<<2)|3,{115,56,48}},
/* 17345 */ {(16<<2)|3,{115,56,49}},
/* 17346 */ {(16<<2)|3,{115,56,50}},
/* 17347 */ {(16<<2)|3,{115,56,97}},
/* 17348 */ {(16<<2)|3,{115,56,99}},
/* 17349 */ {(16<<2)|3,{115,56,101}},
/* 17350 */ {(16<<2)|3,{115,56,105}},
/* 17351 */ {(16<<2)|3,{115,56,111}},
/* 17352 */ {(16<<2)|3,{115,56,115}},
/* 17353 */ {(16<<2)|3,{115,56,116}},
/* 17354 */ {(11<<2)|2,{115,56,0}},
/* 17355 */ {(11<<2)|2,{115,56,0}},
/* 17356 */ {(11<<2)|2,{115,56,0}},
/* 17357 */ {(11<<2)|2,{115,56,0}},
/* 17358 */ {(11<<2)|2,{115,56,0}},
/* 17359 */ {(11<<2)|2,{115,56,0}},
/* 17360 */ {(11<<2)|2,{115,56,0}},
/* 17361 */ {(11<<2)|2,{115,56,0}},
/* 17362 */ {(11<<2)|2,{115,56,0}},
/* 17363 */ {(11<<2)|2,{115,56,0}},
/* 17364 */ {(11<<2)|2,{115,56,0}},
/* 17365 */ {(11<<2)|2,{115,56,0}},
/* 17366 */ {(11<<2)|2,{115,56,0}},
/* 17367 */ {(11<<2)|2,{115,56,0}},
/* 17368 */ {(11<<2)|2,{115,56,0}},
/* 17369 */ {(11<<2)|2,{115,56,0}},
/* 17370 */ {(11<<2)|2,{115,56,0}},
/* 17371 */ {(11<<2)|2,{115,56,0}},
/* 17372 */ {(11<<2)|2,{115,56,0}},
/* 17373 */ {(11<<2)|2,{115,56,0}},
/* 17374 */ {(11<<2)|2,{115,56,0}},
/* 17375 */ {(11<<2)|2,{115,56,0}},
/* 17376 */ {(16<<2)|3,{115,57,48}},
/* 17377 */ {(16<<2)|3,{115,57,49}},
/* 17378 */ {(16<<2)|3,{115,57,50}},
/* 17379 */ {(16<<2)|3,{115,57,97}},
/* 17380 */ {(16<<2)|3,{115,57,99}},
/* 17381 */ {(16<<2)|3,{115,57,101}},
/* 17382 */ {(16<<2)|3,{115,57,105}},
/* 17383 */ {(16<<2)|3,{115,57,111}},
/* 17384 */ {(16<<2)|3,{115,57,115}},
/* 17385 */ {(16<<2)|3,{115,57,116}},
/* 17386 */ {(11<<2)|2,{115,57,0}},
/* 17387 */ {(11<<2)|2,{115,57,0}},
/* 17388 */ {(11<<2)|2,{115,57,0}},
/* 17389 */ {(11<<2)|2,{115,57,0}},
/* 17390 */ {(11<<2)|2,{115,57,0}},
/* 17391 */ {(11<<2)|2,{115,57,0}},
/* 17392 */ {(11<<2)|2,{115,57,0}},
/* 17393 */ {(11<<2)|2,{115,57,0}},
/* 17394 */ {(11<<2)|2,{115,57,0}},
/* 17395 */ {(11<<2)|2,{115,57,0}},
/* 17396 */ {(11<<2)|2,{115,57,0}},
/* 17397 */ {(11<<2)|2,{115,57,0}},
/* 17398 */ {(11<<2)|2,{115,57,0}},
/* 17399 */ {(11<<2)|2,{115,57,0}},
/* 17400 */ {(11<<2)|2,{115,57,0}},
/* 17401 */ {(11<<2)|2,{115,57,0}},
/* 17402 */ {(11<<2)|2,{115,57,0}},
/* 17403 */ {(11<<2)|2,{115,57,0}},
/* 17404 */ {(11<<2)|2,{115,57,0}},
/* 17405 */ {(11<<2)|2,{115,57,0}},
/* 17406 */ {(11<<2)|2,{115,57,0}},
/* 17407 */ {(11<<2)|2,{115,57,0}},
/* 17408 */ {(16<<2)|3,{115,61,48}},
/* 17409 */ {(16<<2)|3,{115,61,49}},
/* 17410 */ {(16<<2)|3,{115,61,50}},
/* 17411 */ {(16<<2)|3,{115,61,97}},
/* 17412 */ {(16<<2)|3,{115,61,99}},
/* 17413 */ {(16<<2)|3,{115,61,101}},
/* 17414 */ {(16<<2)|3,{115,61,105}},
/* 17415 */ {(16<<2)|3,{115,61,111}},
/* 17416 */ {(16<<2)|3,{115,61,115}},
/* 17417 */ {(16<<2)|3,{115,61,116}},
/* 17418 */ {(11<<2)|2,{115,61,0}},
/* 17419 */ {(11<<2)|2,{115,61,0}},
/* 17420 */ {(11<<2)|2,{115,61,0}},
/* 17421 */ {(11<<2)|2,{115,61,0}},
/* 17422 */ {(11<<2)|2,{115,61,0}},
/* 17423 */ {(11<<2)|2,{115,61,0}},
/* 17424 */ {(11<<2)|2,{115,61,0}},
/* 17425 */ {(11<<2)|2,{115,61,0}},
/* 17426 */ {(11<<2)|2,{115,61,0}},
/* 17427 */ {(11<<2)|2,{115,61,0}},
/* 17428 */ {(11<<2)|2,{115,61,0}},
/* 17429 */ {(11<<2)|2,{115,61,0}},
/* 17430 */ {(11<<2)|2,{115,61,0}},
/* 17431 */ {(11<<2)|2,{115,61,0}},
/* 17432 */ {(11<<2)|2,{115,61,0}},
/* 17433 */ {(11<<2)|2,{115,61,0}},
/* 17434 */ {(11<<2)|2,{115,61,0}},
/* 17435 */ {(11<<2)|2,{115,61,0}},
/* 17436 */ {(11<<2)|2,{115,61,0}},
/* 17437 */ {(11<<2)|2,{115,61,0}},
/* 17438 */ {(11<<2)|2,{115,61,0}},
/* 17439 */ {(11<<2)|2,{115,61,0}},
/* 17440 */ {(16<<2)|3,{115,65,48}},
/* 17441 */ {(16<<2)|3,{115,65,49}},
/* 17442 */ {(16<<2)|3,{115,65,50}},
/* 17443 */ {(16<<2)|3,{115,65,97}},
/* 17444 */ {(16<<2)|3,{115,65,99}},
/* 17445 */ {(16<<2)|3,{115,65,101}},
/* 17446 */ {(16<<2)|3,{115,65,105}},
/* 17447 */ {(16<<2)|3,{115,65,111}},
/* 17448 */ {(16<<2)|3,{115,65,115}},
/* 17449 */ {(16<<2)|3,{115,65,116}},
/* 17450 */ {(11<<2)|2,{115,65,0}},
/* 17451 */ {(11<<2)|2,{115,65,0}},
/* 17452 */ {(11<<2)|2,{115,65,0}},
/* 17453 */ {(11<<2)|2,{115,65,0}},
/* 17454 */ {(11<<2)|2,{115,65,0}},
/* 17455 */ {(11<<2)|2,{115,65,0}},
/* 17456 */ {(11<<2)|2,{115,65,0}},
/* 17457 */ {(11<<2)|2,{115,65,0}},
/* 17458 */ {(11<<2)|2,{115,65,0}},
/* 17459 */ {(11<<2)|2,{115,65,0}},
/* 17460 */ {(11<<2)|2,{115,65,0}},
/* 17461 */ {(11<<2)|2,{115,65,0}},
/* 17462 */ {(11<<2)|2,{115,65,0}},
/* 17463 */ {(11<<2)|2,{115,65,0}},
/* 17464 */ {(11<<2)|2,{115,65,0}},
/* 17465 */ {(11<<2)|2,{115,65,0}},
/* 17466 */ {(11<<2)|2,{115,65,0}},
/* 17467 */ {(11<<2)|2,{115,65,0}},
/* 17468 */ {(11<<2)|2,{115,65,0}},
/* 17469 */ {(11<<2)|2,{115,65,0}},
/* 17470 */ {(11<<2)|2,{115,65,0}},
/* 17471 */ {(11<<2)|2,{115,65,0}},
/* 17472 */ {(16<<2)|3,{115,95,48}},
/* 17473 */ {(16<<2)|3,{115,95,49}},
/* 17474 */ {(16<<2)|3,{115,95,50}},
/* 17475 */ {(16<<2)|3,{115,95,97}},
/* 17476 */ {(16<<2)|3,{115,95,99}},
/* 17477 */ {(16<<2)|3,{115,95,101}},
/* 17478 */ {(16<<2)|3,{115,95,105}},
/* 17479 */ {(16<<2)|3,{115,95,111}},
/* 17480 */ {(16<<2)|3,{115,95,115}},
/* 17481 */ {(16<<2)|3,{115,95,116}},
/* 17482 */ {(11<<2)|2,{115,95,0}},
/* 17483 */ {(11<<2)|2,{115,95,0}},
/* 17484 */ {(11<<2)|2,{115,95,0}},
/* 17485 */ {(11<<2)|2,{115,95,0}},
/* 17486 */ {(11<<2)|2,{115,95,0}},
/* 17487 */ {(11<<2)|2,{115,95,0}},
/* 17488 */ {(11<<2)|2,{115,95,0}},
/* 17489 */ {(11<<2)|2,{115,95,0}},
/* 17490 */ {(11<<2)|2,{115,95,0}},
/* 17491 */ {(11<<2)|2,{115,95,0}},
/* 17492 */ {(11<<2)|2,{115,95,0}},
/* 17493 */ {(11<<2)|2,{115,95,0}},
/* 17494 */ {(11<<2)|2,{115,95,0}},
/* 17495 */ {(11<<2)|2,{115,95,0}},
/* 17496 */ {(11<<2)|2,{115,95,0}},
/* 17497 */ {(11<<2)|2,{115,95,0}},
/* 17498 */ {(11<<2)|2,{115,95,0}},
/* 17499 */ {(11<<2)|2,{115,95,0}},
/* 17500 */ {(11<<2)|2,{115,95,0}},
/* 17501 */ {(11<<2)|2,{115,95,0}},
/* 17502 */ {(11<<2)|2,{115,95,0}},
/* 17503 */ {(11<<2)|2,{115,95,0}},
/* 17504 */ {(16<<2)|3,{115,98,48}},
/* 17505 */ {(16<<2)|3,{115,98,49}},
/* 17506 */ {(16<<2)|3,{115,98,50}},
/* 17507 */ {(16<<2)|3,{115,98,97}},
/* 17508 */ {(16<<2)|3,{115,98,99}},
/* 17509 */ {(16<<2)|3,{115,98,101}},
/* 17510 */ {(16<<2)|3,{115,98,105}},
/* 17511 */ {(16<<2)|3,{115,98,111}},
/* 17512 */ {(16<<2)|3,{115,98,115}},
/* 17513 */ {(16<<2)|3,{115,98,116}},
/* 17514 */ {(11<<2)|2,{115,98,0}},
/* 17515 */ {(11<<2)|2,{115,98,0}},
/* 17516 */ {(11<<2)|2,{115,98,0}},
/* 17517 */ {(11<<2)|2,{115,98,0}},
/* 17518 */ {(11<<2)|2,{115,98,0}},
/* 17519 */ {(11<<2)|2,{115,98,0}},
/* 17520 */ {(11<<2)|2,{115,98,0}},
/* 17521 */ {(11<<2)|2,{115,98,0}},
/* 17522 */ {(11<<2)|2,{115,98,0}},
/* 17523 */ {(11<<2)|2,{115,98,0}},
/* 17524 */ {(11<<2)|2,{115,98,0}},
/* 17525 */ {(11<<2)|2,{115,98,0}},
/* 17526 */ {(11<<2)|2,{115,98,0}},
/* 17527 */ {(11<<2)|2,{115,98,0}},
/* 17528 */ {(11<<2)|2,{115,98,0}},
/* 17529 */ {(11<<2)|2,{115,98,0}},
/* 17530 */ {(11<<2)|2,{115,98,0}},
/* 17531 */ {(11<<2)|2,{115,98,0}},
/* 17532 */ {(11<<2)|2,{115,98,0}},
/* 17533 */ {(11<<2)|2,{115,98,0}},
/* 17534 */ {(11<<2)|2,{115,98,0}},
/* 17535 */ {(11<<2)|2,{115,98,0}},
/* 17536 */ {(16<<2)|3,{115,100,48}},
/* 17537 */ {(16<<2)|3,{115,100,49}},
/* 17538 */ {(16<<2)|3,{115,100,50}},
/* 17539 */ {(16<<2)|3,{115,100,97}},
/* 17540 */ {(16<<2)|3,{115,100,99}},
/* 17541 */ {(16<<2)|3,{115,100,101}},
/* 17542 */ {(16<<2)|3,{115,100,105}},
/* 17543 */ {(16<<2)|3,{115,100,111}},
/* 17544 */ {(16<<2)|3,{115,100,115}},
/* 17545 */ {(16<<2)|3,{115,100,116}},
/* 17546 */ {(11<<2)|2,{115,100,0}},
/* 17547 */ {(11<<2)|2,{115,100,0}},
/* 17548 */ {(11<<2)|2,{115,100,0}},
/* 17549 */ {(11<<2)|2,{115,100,0}},
/* 17550 */ {(11<<2)|2,{115,100,0}},
/* 17551 */ {(11<<2)|2,{115,100,0}},
/* 17552 */ {(11<<2)|2,{115,100,0}},
/* 17553 */ {(11<<2)|2,{115,100,0}},
/* 17554 */ {(11<<2)|2,{115,100,0}},
/* 17555 */ {(11<<2)|2,{115,100,0}},
/* 17556 */ {(11<<2)|2,{115,100,0}},
/* 17557 */ {(11<<2)|2,{115,100,0}},
/* 17558 */ {(11<<2)|2,{115,100,0}},
/* 17559 */ {(11<<2)|2,{115,100,0}},
/* 17560 */ {(11<<2)|2,{115,100,0}},
/* 17561 */ {(11<<2)|2,{115,100,0}},
/* 17562 */ {(11<<2)|2,{115,100,0}},
/* 17563 */ {(11<<2)|2,{115,100,0}},
/* 17564 */ {(11<<2)|2,{115,100,0}},
/* 17565 */ {(11<<2)|2,{115,100,0}},
/* 17566 */ {(11<<2)|2,{115,100,0}},
/* 17567 */ {(11<<2)|2,{115,100,0}},
/* 17568 */ {(16<<2)|3,{115,102,48}},
/* 17569 */ {(16<<2)|3,{115,102,49}},
/* 17570 */ {(16<<2)|3,{115,102,50}},
/* 17571 */ {(16<<2)|3,{115,102,97}},
/* 17572 */ {(16<<2)|3,{115,102,99}},
/* 17573 */ {(16<<2)|3,{115,102,101}},
/* 17574 */ {(16<<2)|3,{115,102,105}},
/* 17575 */ {(16<<2)|3,{115,102,111}},
/* 17576 */ {(16<<2)|3,{115,102,115}},
/* 17577 */ {(16<<2)|3,{115,102,116}},
/* 17578 */ {(11<<2)|2,{115,102,0}},
/* 17579 */ {(11<<2)|2,{115,102,0}},
/* 17580 */ {(11<<2)|2,{115,102,0}},
/* 17581 */ {(11<<2)|2,{115,102,0}},
/* 17582 */ {(11<<2)|2,{115,102,0}},
/* 17583 */ {(11<<2)|2,{115,102,0}},
/* 17584 */ {(11<<2)|2,{115,102,0}},
/* 17585 */ {(11<<2)|2,{115,102,0}},
/* 17586 */ {(11<<2)|2,{115,102,0}},
/* 17587 */ {(11<<2)|2,{115,102,0}},
/* 17588 */ {(11<<2)|2,{115,102,0}},
/* 17589 */ {(11<<2)|2,{115,102,0}},
/* 17590 */ {(11<<2)|2,{115,102,0}},
/* 17591 */ {(11<<2)|2,{115,102,0}},
/* 17592 */ {(11<<2)|2,{115,102,0}},
/* 17593 */ {(11<<2)|2,{115,102,0}},
/* 17594 */ {(11<<2)|2,{115,102,0}},
/* 17595 */ {(11<<2)|2,{115,102,0}},
/* 17596 */ {(11<<2)|2,{115,102,0}},
/* 17597 */ {(11<<2)|2,{115,102,0}},
/* 17598 */ {(11<<2)|2,{115,102,0}},
/* 17599 */ {(11<<2)|2,{115,102,0}},
/* 17600 */ {(16<<2)|3,{115,103,48}},
/* 17601 */ {(16<<2)|3,{115,103,49}},
/* 17602 */ {(16<<2)|3,{115,103,50}},
/* 17603 */ {(16<<2)|3,{115,103,97}},
/* 17604 */ {(16<<2)|3,{115,103,99}},
/* 17605 */ {(16<<2)|3,{115,103,101}},
/* 17606 */ {(16<<2)|3,{115,103,105}},
/* 17607 */ {(16<<2)|3,{115,103,111}},
/* 17608 */ {(16<<2)|3,{115,103,115}},
/* 17609 */ {(16<<2)|3,{115,103,116}},
/* 17610 */ {(11<<2)|2,{115,103,0}},
/* 17611 */ {(11<<2)|2,{115,103,0}},
/* 17612 */ {(11<<2)|2,{115,103,0}},
/* 17613 */ {(11<<2)|2,{115,103,0}},
/* 17614 */ {(11<<2)|2,{115,103,0}},
/* 17615 */ {(11<<2)|2,{115,103,0}},
/* 17616 */ {(11<<2)|2,{115,103,0}},
/* 17617 */ {(11<<2)|2,{115,103,0}},
/* 17618 */ {(11<<2)|2,{115,103,0}},
/* 17619 */ {(11<<2)|2,{115,103,0}},
/* 17620 */ {(11<<2)|2,{115,103,0}},
/* 17621 */ {(11<<2)|2,{115,103,0}},
/* 17622 */ {(11<<2)|2,{115,103,0}},
/* 17623 */ {(11<<2)|2,{115,103,0}},
/* 17624 */ {(11<<2)|2,{115,103,0}},
/* 17625 */ {(11<<2)|2,{115,103,0}},
/* 17626 */ {(11<<2)|2,{115,103,0}},
/* 17627 */ {(11<<2)|2,{115,103,0}},
/* 17628 */ {(11<<2)|2,{115,103,0}},
/* 17629 */ {(11<<2)|2,{115,103,0}},
/* 17630 */ {(11<<2)|2,{115,103,0}},
/* 17631 */ {(11<<2)|2,{115,103,0}},
/* 17632 */ {(16<<2)|3,{115,104,48}},
/* 17633 */ {(16<<2)|3,{115,104,49}},
/* 17634 */ {(16<<2)|3,{115,104,50}},
/* 17635 */ {(16<<2)|3,{115,104,97}},
/* 17636 */ {(16<<2)|3,{115,104,99}},
/* 17637 */ {(16<<2)|3,{115,104,101}},
/* 17638 */ {(16<<2)|3,{115,104,105}},
/* 17639 */ {(16<<2)|3,{115,104,111}},
/* 17640 */ {(16<<2)|3,{115,104,115}},
/* 17641 */ {(16<<2)|3,{115,104,116}},
/* 17642 */ {(11<<2)|2,{115,104,0}},
/* 17643 */ {(11<<2)|2,{115,104,0}},
/* 17644 */ {(11<<2)|2,{115,104,0}},
/* 17645 */ {(11<<2)|2,{115,104,0}},
/* 17646 */ {(11<<2)|2,{115,104,0}},
/* 17647 */ {(11<<2)|2,{115,104,0}},
/* 17648 */ {(11<<2)|2,{115,104,0}},
/* 17649 */ {(11<<2)|2,{115,104,0}},
/* 17650 */ {(11<<2)|2,{115,104,0}},
/* 17651 */ {(11<<2)|2,{115,104,0}},
/* 17652 */ {(11<<2)|2,{115,104,0}},
/* 17653 */ {(11<<2)|2,{115,104,0}},
/* 17654 */ {(11<<2)|2,{115,104,0}},
/* 17655 */ {(11<<2)|2,{115,104,0}},
/* 17656 */ {(11<<2)|2,{115,104,0}},
/* 17657 */ {(11<<2)|2,{115,104,0}},
/* 17658 */ {(11<<2)|2,{115,104,0}},
/* 17659 */ {(11<<2)|2,{115,104,0}},
/* 17660 */ {(11<<2)|2,{115,104,0}},
/* 17661 */ {(11<<2)|2,{115,104,0}},
/* 17662 */ {(11<<2)|2,{115,104,0}},
/* 17663 */ {(11<<2)|2,{115,104,0}},
/* 17664 */ {(16<<2)|3,{115,108,48}},
/* 17665 */ {(16<<2)|3,{115,108,49}},
/* 17666 */ {(16<<2)|3,{115,108,50}},
/* 17667 */ {(16<<2)|3,{115,108,97}},
/* 17668 */ {(16<<2)|3,{115,108,99}},
/* 17669 */ {(16<<2)|3,{115,108,101}},
/* 17670 */ {(16<<2)|3,{115,108,105}},
/* 17671 */ {(16<<2)|3,{115,108,111}},
/* 17672 */ {(16<<2)|3,{115,108,115}},
/* 17673 */ {(16<<2)|3,{115,108,116}},
/* 17674 */ {(11<<2)|2,{115,108,0}},
/* 17675 */ {(11<<2)|2,{115,108,0}},
/* 17676 */ {(11<<2)|2,{115,108,0}},
/* 17677 */ {(11<<2)|2,{115,108,0}},
/* 17678 */ {(11<<2)|2,{115,108,0}},
/* 17679 */ {(11<<2)|2,{115,108,0}},
/* 17680 */ {(11<<2)|2,{115,108,0}},
/* 17681 */ {(11<<2)|2,{115,108,0}},
/* 17682 */ {(11<<2)|2,{115,108,0}},
/* 17683 */ {(11<<2)|2,{115,108,0}},
/* 17684 */ {(11<<2)|2,{115,108,0}},
/* 17685 */ {(11<<2)|2,{115,108,0}},
/* 17686 */ {(11<<2)|2,{115,108,0}},
/* 17687 */ {(11<<2)|2,{115,108,0}},
/* 17688 */ {(11<<2)|2,{115,108,0}},
/* 17689 */ {(11<<2)|2,{115,108,0}},
/* 17690 */ {(11<<2)|2,{115,108,0}},
/* 17691 */ {(11<<2)|2,{115,108,0}},
/* 17692 */ {(11<<2)|2,{115,108,0}},
/* 17693 */ {(11<<2)|2,{115,108,0}},
/* 17694 */ {(11<<2)|2,{115,108,0}},
/* 17695 */ {(11<<2)|2,{115,108,0}},
/* 17696 */ {(16<<2)|3,{115,109,48}},
/* 17697 */ {(16<<2)|3,{115,109,49}},
/* 17698 */ {(16<<2)|3,{115,109,50}},
/* 17699 */ {(16<<2)|3,{115,109,97}},
/* 17700 */ {(16<<2)|3,{115,109,99}},
/* 17701 */ {(16<<2)|3,{115,109,101}},
/* 17702 */ {(16<<2)|3,{115,109,105}},
/* 17703 */ {(16<<2)|3,{115,109,111}},
/* 17704 */ {(16<<2)|3,{115,109,115}},
/* 17705 */ {(16<<2)|3,{115,109,116}},
/* 17706 */ {(11<<2)|2,{115,109,0}},
/* 17707 */ {(11<<2)|2,{115,109,0}},
/* 17708 */ {(11<<2)|2,{115,109,0}},
/* 17709 */ {(11<<2)|2,{115,109,0}},
/* 17710 */ {(11<<2)|2,{115,109,0}},
/* 17711 */ {(11<<2)|2,{115,109,0}},
/* 17712 */ {(11<<2)|2,{115,109,0}},
/* 17713 */ {(11<<2)|2,{115,109,0}},
/* 17714 */ {(11<<2)|2,{115,109,0}},
/* 17715 */ {(11<<2)|2,{115,109,0}},
/* 17716 */ {(11<<2)|2,{115,109,0}},
/* 17717 */ {(11<<2)|2,{115,109,0}},
/* 17718 */ {(11<<2)|2,{115,109,0}},
/* 17719 */ {(11<<2)|2,{115,109,0}},
/* 17720 */ {(11<<2)|2,{115,109,0}},
/* 17721 */ {(11<<2)|2,{115,109,0}},
/* 17722 */ {(11<<2)|2,{115,109,0}},
/* 17723 */ {(11<<2)|2,{115,109,0}},
/* 17724 */ {(11<<2)|2,{115,109,0}},
/* 17725 */ {(11<<2)|2,{115,109,0}},
/* 17726 */ {(11<<2)|2,{115,109,0}},
/* 17727 */ {(11<<2)|2,{115,109,0}},
/* 17728 */ {(16<<2)|3,{115,110,48}},
/* 17729 */ {(16<<2)|3,{115,110,49}},
/* 17730 */ {(16<<2)|3,{115,110,50}},
/* 17731 */ {(16<<2)|3,{115,110,97}},
/* 17732 */ {(16<<2)|3,{115,110,99}},
/* 17733 */ {(16<<2)|3,{115,110,101}},
/* 17734 */ {(16<<2)|3,{115,110,105}},
/* 17735 */ {(16<<2)|3,{115,110,111}},
/* 17736 */ {(16<<2)|3,{115,110,115}},
/* 17737 */ {(16<<2)|3,{115,110,116}},
/* 17738 */ {(11<<2)|2,{115,110,0}},
/* 17739 */ {(11<<2)|2,{115,110,0}},
/* 17740 */ {(11<<2)|2,{115,110,0}},
/* 17741 */ {(11<<2)|2,{115,110,0}},
/* 17742 */ {(11<<2)|2,{115,110,0}},
/* 17743 */ {(11<<2)|2,{115,110,0}},
/* 17744 */ {(11<<2)|2,{115,110,0}},
/* 17745 */ {(11<<2)|2,{115,110,0}},
/* 17746 */ {(11<<2)|2,{115,110,0}},
/* 17747 */ {(11<<2)|2,{115,110,0}},
/* 17748 */ {(11<<2)|2,{115,110,0}},
/* 17749 */ {(11<<2)|2,{115,110,0}},
/* 17750 */ {(11<<2)|2,{115,110,0}},
/* 17751 */ {(11<<2)|2,{115,110,0}},
/* 17752 */ {(11<<2)|2,{115,110,0}},
/* 17753 */ {(11<<2)|2,{115,110,0}},
/* 17754 */ {(11<<2)|2,{115,110,0}},
/* 17755 */ {(11<<2)|2,{115,110,0}},
/* 17756 */ {(11<<2)|2,{115,110,0}},
/* 17757 */ {(11<<2)|2,{115,110,0}},
/* 17758 */ {(11<<2)|2,{115,110,0}},
/* 17759 */ {(11<<2)|2,{115,110,0}},
/* 17760 */ {(16<<2)|3,{115,112,48}},
/* 17761 */ {(16<<2)|3,{115,112,49}},
/* 17762 */ {(16<<2)|3,{115,112,50}},
/* 17763 */ {(16<<2)|3,{115,112,97}},
/* 17764 */ {(16<<2)|3,{115,112,99}},
/* 17765 */ {(16<<2)|3,{115,112,101}},
/* 17766 */ {(16<<2)|3,{115,112,105}},
/* 17767 */ {(16<<2)|3,{115,112,111}},
/* 17768 */ {(16<<2)|3,{115,112,115}},
/* 17769 */ {(16<<2)|3,{115,112,116}},
/* 17770 */ {(11<<2)|2,{115,112,0}},
/* 17771 */ {(11<<2)|2,{115,112,0}},
/* 17772 */ {(11<<2)|2,{115,112,0}},
/* 17773 */ {(11<<2)|2,{115,112,0}},
/* 17774 */ {(11<<2)|2,{115,112,0}},
/* 17775 */ {(11<<2)|2,{115,112,0}},
/* 17776 */ {(11<<2)|2,{115,112,0}},
/* 17777 */ {(11<<2)|2,{115,112,0}},
/* 17778 */ {(11<<2)|2,{115,112,0}},
/* 17779 */ {(11<<2)|2,{115,112,0}},
/* 17780 */ {(11<<2)|2,{115,112,0}},
/* 17781 */ {(11<<2)|2,{115,112,0}},
/* 17782 */ {(11<<2)|2,{115,112,0}},
/* 17783 */ {(11<<2)|2,{115,112,0}},
/* 17784 */ {(11<<2)|2,{115,112,0}},
/* 17785 */ {(11<<2)|2,{115,112,0}},
/* 17786 */ {(11<<2)|2,{115,112,0}},
/* 17787 */ {(11<<2)|2,{115,112,0}},
/* 17788 */ {(11<<2)|2,{115,112,0}},
/* 17789 */ {(11<<2)|2,{115,112,0}},
/* 17790 */ {(11<<2)|2,{115,112,0}},
/* 17791 */ {(11<<2)|2,{115,112,0}},
/* 17792 */ {(16<<2)|3,{115,114,48}},
/* 17793 */ {(16<<2)|3,{115,114,49}},
/* 17794 */ {(16<<2)|3,{115,114,50}},
/* 17795 */ {(16<<2)|3,{115,114,97}},
/* 17796 */ {(16<<2)|3,{115,114,99}},
/* 17797 */ {(16<<2)|3,{115,114,101}},
/* 17798 */ {(16<<2)|3,{115,114,105}},
/* 17799 */ {(16<<2)|3,{115,114,111}},
/* 17800 */ {(16<<2)|3,{115,114,115}},
/* 17801 */ {(16<<2)|3,{115,114,116}},
/* 17802 */ {(11<<2)|2,{115,114,0}},
/* 17803 */ {(11<<2)|2,{115,114,0}},
/* 17804 */ {(11<<2)|2,{115,114,0}},
/* 17805 */ {(11<<2)|2,{115,114,0}},
/* 17806 */ {(11<<2)|2,{115,114,0}},
/* 17807 */ {(11<<2)|2,{115,114,0}},
/* 17808 */ {(11<<2)|2,{115,114,0}},
/* 17809 */ {(11<<2)|2,{115,114,0}},
/* 17810 */ {(11<<2)|2,{115,114,0}},
/* 17811 */ {(11<<2)|2,{115,114,0}},
/* 17812 */ {(11<<2)|2,{115,114,0}},
/* 17813 */ {(11<<2)|2,{115,114,0}},
/* 17814 */ {(11<<2)|2,{115,114,0}},
/* 17815 */ {(11<<2)|2,{115,114,0}},
/* 17816 */ {(11<<2)|2,{115,114,0}},
/* 17817 */ {(11<<2)|2,{115,114,0}},
/* 17818 */ {(11<<2)|2,{115,114,0}},
/* 17819 */ {(11<<2)|2,{115,114,0}},
/* 17820 */ {(11<<2)|2,{115,114,0}},
/* 17821 */ {(11<<2)|2,{115,114,0}},
/* 17822 */ {(11<<2)|2,{115,114,0}},
/* 17823 */ {(11<<2)|2,{115,114,0}},
/* 17824 */ {(16<<2)|3,{115,117,48}},
/* 17825 */ {(16<<2)|3,{115,117,49}},
/* 17826 */ {(16<<2)|3,{115,117,50}},
/* 17827 */ {(16<<2)|3,{115,117,97}},
/* 17828 */ {(16<<2)|3,{115,117,99}},
/* 17829 */ {(16<<2)|3,{115,117,101}},
/* 17830 */ {(16<<2)|3,{115,117,105}},
/* 17831 */ {(16<<2)|3,{115,117,111}},
/* 17832 */ {(16<<2)|3,{115,117,115}},
/* 17833 */ {(16<<2)|3,{115,117,116}},
/* 17834 */ {(11<<2)|2,{115,117,0}},
/* 17835 */ {(11<<2)|2,{115,117,0}},
/* 17836 */ {(11<<2)|2,{115,117,0}},
/* 17837 */ {(11<<2)|2,{115,117,0}},
/* 17838 */ {(11<<2)|2,{115,117,0}},
/* 17839 */ {(11<<2)|2,{115,117,0}},
/* 17840 */ {(11<<2)|2,{115,117,0}},
/* 17841 */ {(11<<2)|2,{115,117,0}},
/* 17842 */ {(11<<2)|2,{115,117,0}},
/* 17843 */ {(11<<2)|2,{115,117,0}},
/* 17844 */ {(11<<2)|2,{115,117,0}},
/* 17845 */ {(11<<2)|2,{115,117,0}},
/* 17846 */ {(11<<2)|2,{115,117,0}},
/* 17847 */ {(11<<2)|2,{115,117,0}},
/* 17848 */ {(11<<2)|2,{115,117,0}},
/* 17849 */ {(11<<2)|2,{115,117,0}},
/* 17850 */ {(11<<2)|2,{115,117,0}},
/* 17851 */ {(11<<2)|2,{115,117,0}},
/* 17852 */ {(11<<2)|2,{115,117,0}},
/* 17853 */ {(11<<2)|2,{115,117,0}},
/* 17854 */ {(11<<2)|2,{115,117,0}},
/* 17855 */ {(11<<2)|2,{115,117,0}},
/* 17856 */ {(12<<2)|2,{115,58,0}},
/* 17857 */ {(12<<2)|2,{115,58,0}},
/* 17858 */ {(12<<2)|2,{115,58,0}},
/* 17859 */ {(12<<2)|2,{115,58,0}},
/* 17860 */ {(12<<2)|2,{115,58,0}},
/* 17861 */ {(12<<2)|2,{115,58,0}},
/* 17862 */ {(12<<2)|2,{115,58,0}},
/* 17863 */ {(12<<2)|2,{115,58,0}},
/* 17864 */ {(12<<2)|2,{115,58,0}},
/* 17865 */ {(12<<2)|2,{115,58,0}},
/* 17866 */ {(12<<2)|2,{115,58,0}},
/* 17867 */ {(12<<2)|2,{115,58,0}},
/* 17868 */ {(12<<2)|2,{115,58,0}},
/* 17869 */ {(12<<2)|2,{115,58,0}},
/* 17870 */ {(12<<2)|2,{115,58,0}},
/* 17871 */ {(12<<2)|2,{115,58,0}},
/* 17872 */ {(12<<2)|2,{115,66,0}},
/* 17873 */ {(12<<2)|2,{115,66,0}},
/* 17874 */ {(12<<2)|2,{115,66,0}},
/* 17875 */ {(12<<2)|2,{115,66,0}},
/* 17876 */ {(12<<2)|2,{115,66,0}},
/* 17877 */ {(12<<2)|2,{115,66,0}},
/* 17878 */ {(12<<2)|2,{115,66,0}},
/* 17879 */ {(12<<2)|2,{115,66,0}},
/* 17880 */ {(12<<2)|2,{115,66,0}},
/* 17881 */ {(12<<2)|2,{115,66,0}},
/* 17882 */ {(12<<2)|2,{115,66,0}},
/* 17883 */ {(12<<2)|2,{115,66,0}},
/* 17884 */ {(12<<2)|2,{115,66,0}},
/* 17885 */ {(12<<2)|2,{115,66,0}},
/* 17886 */ {(12<<2)|2,{115,66,0}},
/* 17887 */ {(12<<2)|2,{115,66,0}},
/* 17888 */ {(12<<2)|2,{115,67,0}},
/* 17889 */ {(12<<2)|2,{115,67,0}},
/* 17890 */ {(12<<2)|2,{115,67,0}},
/* 17891 */ {(12<<2)|2,{115,67,0}},
/* 17892 */ {(12<<2)|2,{115,67,0}},
/* 17893 */ {(12<<2)|2,{115,67,0}},
/* 17894 */ {(12<<2)|2,{115,67,0}},
/* 17895 */ {(12<<2)|2,{115,67,0}},
/* 17896 */ {(12<<2)|2,{115,67,0}},
/* 17897 */ {(12<<2)|2,{115,67,0}},
/* 17898 */ {(12<<2)|2,{115,67,0}},
/* 17899 */ {(12<<2)|2,{115,67,0}},
/* 17900 */ {(12<<2)|2,{115,67,0}},
/* 17901 */ {(12<<2)|2,{115,67,0}},
/* 17902 */ {(12<<2)|2,{115,67,0}},
/* 17903 */ {(12<<2)|2,{115,67,0}},
/* 17904 */ {(12<<2)|2,{115,68,0}},
/* 17905 */ {(12<<2)|2,{115,68,0}},
/* 17906 */ {(12<<2)|2,{115,68,0}},
/* 17907 */ {(12<<2)|2,{115,68,0}},
/* 17908 */ {(12<<2)|2,{115,68,0}},
/* 17909 */ {(12<<2)|2,{115,68,0}},
/* 17910 */ {(12<<2)|2,{115,68,0}},
/* 17911 */ {(12<<2)|2,{115,68,0}},
/* 17912 */ {(12<<2)|2,{115,68,0}},
/* 17913 */ {(12<<2)|2,{115,68,0}},
/* 17914 */ {(12<<2)|2,{115,68,0}},
/* 17915 */ {(12<<2)|2,{115,68,0}},
/* 17916 */ {(12<<2)|2,{115,68,0}},
/* 17917 */ {(12<<2)|2,{115,68,0}},
/* 17918 */ {(12<<2)|2,{115,68,0}},
/* 17919 */ {(12<<2)|2,{115,68,0}},
/* 17920 */ {(12<<2)|2,{115,69,0}},
/* 17921 */ {(12<<2)|2,{115,69,0}},
/* 17922 */ {(12<<2)|2,{115,69,0}},
/* 17923 */ {(12<<2)|2,{115,69,0}},
/* 17924 */ {(12<<2)|2,{115,69,0}},
/* 17925 */ {(12<<2)|2,{115,69,0}},
/* 17926 */ {(12<<2)|2,{115,69,0}},
/* 17927 */ {(12<<2)|2,{115,69,0}},
/* 17928 */ {(12<<2)|2,{115,69,0}},
/* 17929 */ {(12<<2)|2,{115,69,0}},
/* 17930 */ {(12<<2)|2,{115,69,0}},
/* 17931 */ {(12<<2)|2,{115,69,0}},
/* 17932 */ {(12<<2)|2,{115,69,0}},
/* 17933 */ {(12<<2)|2,{115,69,0}},
/* 17934 */ {(12<<2)|2,{115,69,0}},
/* 17935 */ {(12<<2)|2,{115,69,0}},
/* 17936 */ {(12<<2)|2,{115,70,0}},
/* 17937 */ {(12<<2)|2,{115,70,0}},
/* 17938 */ {(12<<2)|2,{115,70,0}},
/* 17939 */ {(12<<2)|2,{115,70,0}},
/* 17940 */ {(12<<2)|2,{115,70,0}},
/* 17941 */ {(12<<2)|2,{115,70,0}},
/* 17942 */ {(12<<2)|2,{115,70,0}},
/* 17943 */ {(12<<2)|2,{115,70,0}},
/* 17944 */ {(12<<2)|2,{115,70,0}},
/* 17945 */ {(12<<2)|2,{115,70,0}},
/* 17946 */ {(12<<2)|2,{115,70,0}},
/* 17947 */ {(12<<2)|2,{115,70,0}},
/* 17948 */ {(12<<2)|2,{115,70,0}},
/* 17949 */ {(12<<2)|2,{115,70,0}},
/* 17950 */ {(12<<2)|2,{115,70,0}},
/* 17951 */ {(12<<2)|2,{115,70,0}},
/* 17952 */ {(12<<2)|2,{115,71,0}},
/* 17953 */ {(12<<2)|2,{115,71,0}},
/* 17954 */ {(12<<2)|2,{115,71,0}},
/* 17955 */ {(12<<2)|2,{115,71,0}},
/* 17956 */ {(12<<2)|2,{115,71,0}},
/* 17957 */ {(12<<2)|2,{115,71,0}},
/* 17958 */ {(12<<2)|2,{115,71,0}},
/* 17959 */ {(12<<2)|2,{115,71,0}},
/* 17960 */ {(12<<2)|2,{115,71,0}},
/* 17961 */ {(12<<2)|2,{115,71,0}},
/* 17962 */ {(12<<2)|2,{115,71,0}},
/* 17963 */ {(12<<2)|2,{115,71,0}},
/* 17964 */ {(12<<2)|2,{115,71,0}},
/* 17965 */ {(12<<2)|2,{115,71,0}},
/* 17966 */ {(12<<2)|2,{115,71,0}},
/* 17967 */ {(12<<2)|2,{115,71,0}},
/* 17968 */ {(12<<2)|2,{115,72,0}},
/* 17969 */ {(12<<2)|2,{115,72,0}},
/* 17970 */ {(12<<2)|2,{115,72,0}},
/* 17971 */ {(12<<2)|2,{115,72,0}},
/* 17972 */ {(12<<2)|2,{115,72,0}},
/* 17973 */ {(12<<2)|2,{115,72,0}},
/* 17974 */ {(12<<2)|2,{115,72,0}},
/* 17975 */ {(12<<2)|2,{115,72,0}},
/* 17976 */ {(12<<2)|2,{115,72,0}},
/* 17977 */ {(12<<2)|2,{115,72,0}},
/* 17978 */ {(12<<2)|2,{115,72,0}},
/* 17979 */ {(12<<2)|2,{115,72,0}},
/* 17980 */ {(12<<2)|2,{115,72,0}},
/* 17981 */ {(12<<2)|2,{115,72,0}},
/* 17982 */ {(12<<2)|2,{115,72,0}},
/* 17983 */ {(12<<2)|2,{115,72,0}},
/* 17984 */ {(12<<2)|2,{115,73,0}},
/* 17985 */ {(12<<2)|2,{115,73,0}},
/* 17986 */ {(12<<2)|2,{115,73,0}},
/* 17987 */ {(12<<2)|2,{115,73,0}},
/* 17988 */ {(12<<2)|2,{115,73,0}},
/* 17989 */ {(12<<2)|2,{115,73,0}},
/* 17990 */ {(12<<2)|2,{115,73,0}},
/* 17991 */ {(12<<2)|2,{115,73,0}},
/* 17992 */ {(12<<2)|2,{115,73,0}},
/* 17993 */ {(12<<2)|2,{115,73,0}},
/* 17994 */ {(12<<2)|2,{115,73,0}},
/* 17995 */ {(12<<2)|2,{115,73,0}},
/* 17996 */ {(12<<2)|2,{115,73,0}},
/* 17997 */ {(12<<2)|2,{115,73,0}},
/* 17998 */ {(12<<2)|2,{115,73,0}},
/* 17999 */ {(12<<2)|2,{115,73,0}},
/* 18000 */ {(12<<2)|2,{115,74,0}},
/* 18001 */ {(12<<2)|2,{115,74,0}},
/* 18002 */ {(12<<2)|2,{115,74,0}},
/* 18003 */ {(12<<2)|2,{115,74,0}},
/* 18004 */ {(12<<2)|2,{115,74,0}},
/* 18005 */ {(12<<2)|2,{115,74,0}},
/* 18006 */ {(12<<2)|2,{115,74,0}},
/* 18007 */ {(12<<2)|2,{115,74,0}},
/* 18008 */ {(12<<2)|2,{115,74,0}},
/* 18009 */ {(12<<2)|2,{115,74,0}},
/* 18010 */ {(12<<2)|2,{115,74,0}},
/* 18011 */ {(12<<2)|2,{115,74,0}},
/* 18012 */ {(12<<2)|2,{115,74,0}},
/* 18013 */ {(12<<2)|2,{115,74,0}},
/* 18014 */ {(12<<2)|2,{115,74,0}},
/* 18015 */ {(12<<2)|2,{115,74,0}},
/* 18016 */ {(12<<2)|2,{115,75,0}},
/* 18017 */ {(12<<2)|2,{115,75,0}},
/* 18018 */ {(12<<2)|2,{115,75,0}},
/* 18019 */ {(12<<2)|2,{115,75,0}},
/* 18020 */ {(12<<2)|2,{115,75,0}},
/* 18021 */ {(12<<2)|2,{115,75,0}},
/* 18022 */ {(12<<2)|2,{115,75,0}},
/* 18023 */ {(12<<2)|2,{115,75,0}},
/* 18024 */ {(12<<2)|2,{115,75,0}},
/* 18025 */ {(12<<2)|2,{115,75,0}},
/* 18026 */ {(12<<2)|2,{115,75,0}},
/* 18027 */ {(12<<2)|2,{115,75,0}},
/* 18028 */ {(12<<2)|2,{115,75,0}},
/* 18029 */ {(12<<2)|2,{115,75,0}},
/* 18030 */ {(12<<2)|2,{115,75,0}},
/* 18031 */ {(12<<2)|2,{115,75,0}},
/* 18032 */ {(12<<2)|2,{115,76,0}},
/* 18033 */ {(12<<2)|2,{115,76,0}},
/* 18034 */ {(12<<2)|2,{115,76,0}},
/* 18035 */ {(12<<2)|2,{115,76,0}},
/* 18036 */ {(12<<2)|2,{115,76,0}},
/* 18037 */ {(12<<2)|2,{115,76,0}},
/* 18038 */ {(12<<2)|2,{115,76,0}},
/* 18039 */ {(12<<2)|2,{115,76,0}},
/* 18040 */ {(12<<2)|2,{115,76,0}},
/* 18041 */ {(12<<2)|2,{115,76,0}},
/* 18042 */ {(12<<2)|2,{115,76,0}},
/* 18043 */ {(12<<2)|2,{115,76,0}},
/* 18044 */ {(12<<2)|2,{115,76,0}},
/* 18045 */ {(12<<2)|2,{115,76,0}},
/* 18046 */ {(12<<2)|2,{115,76,0}},
/* 18047 */ {(12<<2)|2,{115,76,0}},
/* 18048 */ {(12<<2)|2,{115,77,0}},
/* 18049 */ {(12<<2)|2,{115,77,0}},
/* 18050 */ {(12<<2)|2,{115,77,0}},
/* 18051 */ {(12<<2)|2,{115,77,0}},
/* 18052 */ {(12<<2)|2,{115,77,0}},
/* 18053 */ {(12<<2)|2,{115,77,0}},
/* 18054 */ {(12<<2)|2,{115,77,0}},
/* 18055 */ {(12<<2)|2,{115,77,0}},
/* 18056 */ {(12<<2)|2,{115,77,0}},
/* 18057 */ {(12<<2)|2,{115,77,0}},
/* 18058 */ {(12<<2)|2,{115,77,0}},
/* 18059 */ {(12<<2)|2,{115,77,0}},
/* 18060 */ {(12<<2)|2,{115,77,0}},
/* 18061 */ {(12<<2)|2,{115,77,0}},
/* 18062 */ {(12<<2)|2,{115,77,0}},
/* 18063 */ {(12<<2)|2,{115,77,0}},
/* 18064 */ {(12<<2)|2,{115,78,0}},
/* 18065 */ {(12<<2)|2,{115,78,0}},
/* 18066 */ {(12<<2)|2,{115,78,0}},
/* 18067 */ {(12<<2)|2,{115,78,0}},
/* 18068 */ {(12<<2)|2,{115,78,0}},
/* 18069 */ {(12<<2)|2,{115,78,0}},
/* 18070 */ {(12<<2)|2,{115,78,0}},
/* 18071 */ {(12<<2)|2,{115,78,0}},
/* 18072 */ {(12<<2)|2,{115,78,0}},
/* 18073 */ {(12<<2)|2,{115,78,0}},
/* 18074 */ {(12<<2)|2,{115,78,0}},
/* 18075 */ {(12<<2)|2,{115,78,0}},
/* 18076 */ {(12<<2)|2,{115,78,0}},
/* 18077 */ {(12<<2)|2,{115,78,0}},
/* 18078 */ {(12<<2)|2,{115,78,0}},
/* 18079 */ {(12<<2)|2,{115,78,0}},
/* 18080 */ {(12<<2)|2,{115,79,0}},
/* 18081 */ {(12<<2)|2,{115,79,0}},
/* 18082 */ {(12<<2)|2,{115,79,0}},
/* 18083 */ {(12<<2)|2,{115,79,0}},
/* 18084 */ {(12<<2)|2,{115,79,0}},
/* 18085 */ {(12<<2)|2,{115,79,0}},
/* 18086 */ {(12<<2)|2,{115,79,0}},
/* 18087 */ {(12<<2)|2,{115,79,0}},
/* 18088 */ {(12<<2)|2,{115,79,0}},
/* 18089 */ {(12<<2)|2,{115,79,0}},
/* 18090 */ {(12<<2)|2,{115,79,0}},
/* 18091 */ {(12<<2)|2,{115,79,0}},
/* 18092 */ {(12<<2)|2,{115,79,0}},
/* 18093 */ {(12<<2)|2,{115,79,0}},
/* 18094 */ {(12<<2)|2,{115,79,0}},
/* 18095 */ {(12<<2)|2,{115,79,0}},
/* 18096 */ {(12<<2)|2,{115,80,0}},
/* 18097 */ {(12<<2)|2,{115,80,0}},
/* 18098 */ {(12<<2)|2,{115,80,0}},
/* 18099 */ {(12<<2)|2,{115,80,0}},
/* 18100 */ {(12<<2)|2,{115,80,0}},
/* 18101 */ {(12<<2)|2,{115,80,0}},
/* 18102 */ {(12<<2)|2,{115,80,0}},
/* 18103 */ {(12<<2)|2,{115,80,0}},
/* 18104 */ {(12<<2)|2,{115,80,0}},
/* 18105 */ {(12<<2)|2,{115,80,0}},
/* 18106 */ {(12<<2)|2,{115,80,0}},
/* 18107 */ {(12<<2)|2,{115,80,0}},
/* 18108 */ {(12<<2)|2,{115,80,0}},
/* 18109 */ {(12<<2)|2,{115,80,0}},
/* 18110 */ {(12<<2)|2,{115,80,0}},
/* 18111 */ {(12<<2)|2,{115,80,0}},
/* 18112 */ {(12<<2)|2,{115,81,0}},
/* 18113 */ {(12<<2)|2,{115,81,0}},
/* 18114 */ {(12<<2)|2,{115,81,0}},
/* 18115 */ {(12<<2)|2,{115,81,0}},
/* 18116 */ {(12<<2)|2,{115,81,0}},
/* 18117 */ {(12<<2)|2,{115,81,0}},
/* 18118 */ {(12<<2)|2,{115,81,0}},
/* 18119 */ {(12<<2)|2,{115,81,0}},
/* 18120 */ {(12<<2)|2,{115,81,0}},
/* 18121 */ {(12<<2)|2,{115,81,0}},
/* 18122 */ {(12<<2)|2,{115,81,0}},
/* 18123 */ {(12<<2)|2,{115,81,0}},
/* 18124 */ {(12<<2)|2,{115,81,0}},
/* 18125 */ {(12<<2)|2,{115,81,0}},
/* 18126 */ {(12<<2)|2,{115,81,0}},
/* 18127 */ {(12<<2)|2,{115,81,0}},
/* 18128 */ {(12<<2)|2,{115,82,0}},
/* 18129 */ {(12<<2)|2,{115,82,0}},
/* 18130 */ {(12<<2)|2,{115,82,0}},
/* 18131 */ {(12<<2)|2,{115,82,0}},
/* 18132 */ {(12<<2)|2,{115,82,0}},
/* 18133 */ {(12<<2)|2,{115,82,0}},
/* 18134 */ {(12<<2)|2,{115,82,0}},
/* 18135 */ {(12<<2)|2,{115,82,0}},
/* 18136 */ {(12<<2)|2,{115,82,0}},
/* 18137 */ {(12<<2)|2,{115,82,0}},
/* 18138 */ {(12<<2)|2,{115,82,0}},
/* 18139 */ {(12<<2)|2,{115,82,0}},
/* 18140 */ {(12<<2)|2,{115,82,0}},
/* 18141 */ {(12<<2)|2,{115,82,0}},
/* 18142 */ {(12<<2)|2,{115,82,0}},
/* 18143 */ {(12<<2)|2,{115,82,0}},
/* 18144 */ {(12<<2)|2,{115,83,0}},
/* 18145 */ {(12<<2)|2,{115,83,0}},
/* 18146 */ {(12<<2)|2,{115,83,0}},
/* 18147 */ {(12<<2)|2,{115,83,0}},
/* 18148 */ {(12<<2)|2,{115,83,0}},
/* 18149 */ {(12<<2)|2,{115,83,0}},
/* 18150 */ {(12<<2)|2,{115,83,0}},
/* 18151 */ {(12<<2)|2,{115,83,0}},
/* 18152 */ {(12<<2)|2,{115,83,0}},
/* 18153 */ {(12<<2)|2,{115,83,0}},
/* 18154 */ {(12<<2)|2,{115,83,0}},
/* 18155 */ {(12<<2)|2,{115,83,0}},
/* 18156 */ {(12<<2)|2,{115,83,0}},
/* 18157 */ {(12<<2)|2,{115,83,0}},
/* 18158 */ {(12<<2)|2,{115,83,0}},
/* 18159 */ {(12<<2)|2,{115,83,0}},
/* 18160 */ {(12<<2)|2,{115,84,0}},
/* 18161 */ {(12<<2)|2,{115,84,0}},
/* 18162 */ {(12<<2)|2,{115,84,0}},
/* 18163 */ {(12<<2)|2,{115,84,0}},
/* 18164 */ {(12<<2)|2,{115,84,0}},
/* 18165 */ {(12<<2)|2,{115,84,0}},
/* 18166 */ {(12<<2)|2,{115,84,0}},
/* 18167 */ {(12<<2)|2,{115,84,0}},
/* 18168 */ {(12<<2)|2,{115,84,0}},
/* 18169 */ {(12<<2)|2,{115,84,0}},
/* 18170 */ {(12<<2)|2,{115,84,0}},
/* 18171 */ {(12<<2)|2,{115,84,0}},
/* 18172 */ {(12<<2)|2,{115,84,0}},
/* 18173 */ {(12<<2)|2,{115,84,0}},
/* 18174 */ {(12<<2)|2,{115,84,0}},
/* 18175 */ {(12<<2)|2,{115,84,0}},
/* 18176 */ {(12<<2)|2,{115,85,0}},
/* 18177 */ {(12<<2)|2,{115,85,0}},
/* 18178 */ {(12<<2)|2,{115,85,0}},
/* 18179 */ {(12<<2)|2,{115,85,0}},
/* 18180 */ {(12<<2)|2,{115,85,0}},
/* 18181 */ {(12<<2)|2,{115,85,0}},
/* 18182 */ {(12<<2)|2,{115,85,0}},
/* 18183 */ {(12<<2)|2,{115,85,0}},
/* 18184 */ {(12<<2)|2,{115,85,0}},
/* 18185 */ {(12<<2)|2,{115,85,0}},
/* 18186 */ {(12<<2)|2,{115,85,0}},
/* 18187 */ {(12<<2)|2,{115,85,0}},
/* 18188 */ {(12<<2)|2,{115,85,0}},
/* 18189 */ {(12<<2)|2,{115,85,0}},
/* 18190 */ {(12<<2)|2,{115,85,0}},
/* 18191 */ {(12<<2)|2,{115,85,0}},
/* 18192 */ {(12<<2)|2,{115,86,0}},
/* 18193 */ {(12<<2)|2,{115,86,0}},
/* 18194 */ {(12<<2)|2,{115,86,0}},
/* 18195 */ {(12<<2)|2,{115,86,0}},
/* 18196 */ {(12<<2)|2,{115,86,0}},
/* 18197 */ {(12<<2)|2,{115,86,0}},
/* 18198 */ {(12<<2)|2,{115,86,0}},
/* 18199 */ {(12<<2)|2,{115,86,0}},
/* 18200 */ {(12<<2)|2,{115,86,0}},
/* 18201 */ {(12<<2)|2,{115,86,0}},
/* 18202 */ {(12<<2)|2,{115,86,0}},
/* 18203 */ {(12<<2)|2,{115,86,0}},
/* 18204 */ {(12<<2)|2,{115,86,0}},
/* 18205 */ {(12<<2)|2,{115,86,0}},
/* 18206 */ {(12<<2)|2,{115,86,0}},
/* 18207 */ {(12<<2)|2,{115,86,0}},
/* 18208 */ {(12<<2)|2,{115,87,0}},
/* 18209 */ {(12<<2)|2,{115,87,0}},
/* 18210 */ {(12<<2)|2,{115,87,0}},
/* 18211 */ {(12<<2)|2,{115,87,0}},
/* 18212 */ {(12<<2)|2,{115,87,0}},
/* 18213 */ {(12<<2)|2,{115,87,0}},
/* 18214 */ {(12<<2)|2,{115,87,0}},
/* 18215 */ {(12<<2)|2,{115,87,0}},
/* 18216 */ {(12<<2)|2,{115,87,0}},
/* 18217 */ {(12<<2)|2,{115,87,0}},
/* 18218 */ {(12<<2)|2,{115,87,0}},
/* 18219 */ {(12<<2)|2,{115,87,0}},
/* 18220 */ {(12<<2)|2,{115,87,0}},
/* 18221 */ {(12<<2)|2,{115,87,0}},
/* 18222 */ {(12<<2)|2,{115,87,0}},
/* 18223 */ {(12<<2)|2,{115,87,0}},
/* 18224 */ {(12<<2)|2,{115,89,0}},
/* 18225 */ {(12<<2)|2,{115,89,0}},
/* 18226 */ {(12<<2)|2,{115,89,0}},
/* 18227 */ {(12<<2)|2,{115,89,0}},
/* 18228 */ {(12<<2)|2,{115,89,0}},
/* 18229 */ {(12<<2)|2,{115,89,0}},
/* 18230 */ {(12<<2)|2,{115,89,0}},
/* 18231 */ {(12<<2)|2,{115,89,0}},
/* 18232 */ {(12<<2)|2,{115,89,0}},
/* 18233 */ {(12<<2)|2,{115,89,0}},
/* 18234 */ {(12<<2)|2,{115,89,0}},
/* 18235 */ {(12<<2)|2,{115,89,0}},
/* 18236 */ {(12<<2)|2,{115,89,0}},
/* 18237 */ {(12<<2)|2,{115,89,0}},
/* 18238 */ {(12<<2)|2,{115,89,0}},
/* 18239 */ {(12<<2)|2,{115,89,0}},
/* 18240 */ {(12<<2)|2,{115,106,0}},
/* 18241 */ {(12<<2)|2,{115,106,0}},
/* 18242 */ {(12<<2)|2,{115,106,0}},
/* 18243 */ {(12<<2)|2,{115,106,0}},
/* 18244 */ {(12<<2)|2,{115,106,0}},
/* 18245 */ {(12<<2)|2,{115,106,0}},
/* 18246 */ {(12<<2)|2,{115,106,0}},
/* 18247 */ {(12<<2)|2,{115,106,0}},
/* 18248 */ {(12<<2)|2,{115,106,0}},
/* 18249 */ {(12<<2)|2,{115,106,0}},
/* 18250 */ {(12<<2)|2,{115,106,0}},
/* 18251 */ {(12<<2)|2,{115,106,0}},
/* 18252 */ {(12<<2)|2,{115,106,0}},
/* 18253 */ {(12<<2)|2,{115,106,0}},
/* 18254 */ {(12<<2)|2,{115,106,0}},
/* 18255 */ {(12<<2)|2,{115,106,0}},
/* 18256 */ {(12<<2)|2,{115,107,0}},
/* 18257 */ {(12<<2)|2,{115,107,0}},
/* 18258 */ {(12<<2)|2,{115,107,0}},
/* 18259 */ {(12<<2)|2,{115,107,0}},
/* 18260 */ {(12<<2)|2,{115,107,0}},
/* 18261 */ {(12<<2)|2,{115,107,0}},
/* 18262 */ {(12<<2)|2,{115,107,0}},
/* 18263 */ {(12<<2)|2,{115,107,0}},
/* 18264 */ {(12<<2)|2,{115,107,0}},
/* 18265 */ {(12<<2)|2,{115,107,0}},
/* 18266 */ {(12<<2)|2,{115,107,0}},
/* 18267 */ {(12<<2)|2,{115,107,0}},
/* 18268 */ {(12<<2)|2,{115,107,0}},
/* 18269 */ {(12<<2)|2,{115,107,0}},
/* 18270 */ {(12<<2)|2,{115,107,0}},
/* 18271 */ {(12<<2)|2,{115,107,0}},
/* 18272 */ {(12<<2)|2,{115,113,0}},
/* 18273 */ {(12<<2)|2,{115,113,0}},
/* 18274 */ {(12<<2)|2,{115,113,0}},
/* 18275 */ {(12<<2)|2,{115,113,0}},
/* 18276 */ {(12<<2)|2,{115,113,0}},
/* 18277 */ {(12<<2)|2,{115,113,0}},
/* 18278 */ {(12<<2)|2,{115,113,0}},
/* 18279 */ {(12<<2)|2,{115,113,0}},
/* 18280 */ {(12<<2)|2,{115,113,0}},
/* 18281 */ {(12<<2)|2,{115,113,0}},
/* 18282 */ {(12<<2)|2,{115,113,0}},
/* 18283 */ {(12<<2)|2,{115,113,0}},
/* 18284 */ {(12<<2)|2,{115,113,0}},
/* 18285 */ {(12<<2)|2,{115,113,0}},
/* 18286 */ {(12<<2)|2,{115,113,0}},
/* 18287 */ {(12<<2)|2,{115,113,0}},
/* 18288 */ {(12<<2)|2,{115,118,0}},
/* 18289 */ {(12<<2)|2,{115,118,0}},
/* 18290 */ {(12<<2)|2,{115,118,0}},
/* 18291 */ {(12<<2)|2,{115,118,0}},
/* 18292 */ {(12<<2)|2,{115,118,0}},
/* 18293 */ {(12<<2)|2,{115,118,0}},
/* 18294 */ {(12<<2)|2,{115,118,0}},
/* 18295 */ {(12<<2)|2,{115,118,0}},
/* 18296 */ {(12<<2)|2,{115,118,0}},
/* 18297 */ {(12<<2)|2,{115,118,0}},
/* 18298 */ {(12<<2)|2,{115,118,0}},
/* 18299 */ {(12<<2)|2,{115,118,0}},
/* 18300 */ {(12<<2)|2,{115,118,0}},
/* 18301 */ {(12<<2)|2,{115,118,0}},
/* 18302 */ {(12<<2)|2,{115,118,0}},
/* 18303 */ {(12<<2)|2,{115,118,0}},
/* 18304 */ {(12<<2)|2,{115,119,0}},
/* 18305 */ {(12<<2)|2,{115,119,0}},
/* 18306 */ {(12<<2)|2,{115,119,0}},
/* 18307 */ {(12<<2)|2,{115,119,0}},
/* 18308 */ {(12<<2)|2,{115,119,0}},
/* 18309 */ {(12<<2)|2,{115,119,0}},
/* 18310 */ {(12<<2)|2,{115,119,0}},
/* 18311 */ {(12<<2)|2,{115,119,0}},
/* 18312 */ {(12<<2)|2,{115,119,0}},
/* 18313 */ {(12<<2)|2,{115,119,0}},
/* 18314 */ {(12<<2)|2,{115,119,0}},
/* 18315 */ {(12<<2)|2,{115,119,0}},
/* 18316 */ {(12<<2)|2,{115,119,0}},
/* 18317 */ {(12<<2)|2,{115,119,0}},
/* 18318 */ {(12<<2)|2,{115,119,0}},
/* 18319 */ {(12<<2)|2,{115,119,0}},
/* 18320 */ {(12<<2)|2,{115,120,0}},
/* 18321 */ {(12<<2)|2,{115,120,0}},
/* 18322 */ {(12<<2)|2,{115,120,0}},
/* 18323 */ {(12<<2)|2,{115,120,0}},
/* 18324 */ {(12<<2)|2,{115,120,0}},
/* 18325 */ {(12<<2)|2,{115,120,0}},
/* 18326 */ {(12<<2)|2,{115,120,0}},
/* 18327 */ {(12<<2)|2,{115,120,0}},
/* 18328 */ {(12<<2)|2,{115,120,0}},
/* 18329 */ {(12<<2)|2,{115,120,0}},
/* 18330 */ {(12<<2)|2,{115,120,0}},
/* 18331 */ {(12<<2)|2,{115,120,0}},
/* 18332 */ {(12<<2)|2,{115,120,0}},
/* 18333 */ {(12<<2)|2,{115,120,0}},
/* 18334 */ {(12<<2)|2,{115,120,0}},
/* 18335 */ {(12<<2)|2,{115,120,0}},
/* 18336 */ {(12<<2)|2,{115,121,0}},
/* 18337 */ {(12<<2)|2,{115,121,0}},
/* 18338 */ {(12<<2)|2,{115,121,0}},
/* 18339 */ {(12<<2)|2,{115,121,0}},
/* 18340 */ {(12<<2)|2,{115,121,0}},
/* 18341 */ {(12<<2)|2,{115,121,0}},
/* 18342 */ {(12<<2)|2,{115,121,0}},
/* 18343 */ {(12<<2)|2,{115,121,0}},
/* 18344 */ {(12<<2)|2,{115,121,0}},
/* 18345 */ {(12<<2)|2,{115,121,0}},
/* 18346 */ {(12<<2)|2,{115,121,0}},
/* 18347 */ {(12<<2)|2,{115,121,0}},
/* 18348 */ {(12<<2)|2,{115,121,0}},
/* 18349 */ {(12<<2)|2,{115,121,0}},
/* 18350 */ {(12<<2)|2,{115,121,0}},
/* 18351 */ {(12<<2)|2,{115,121,0}},
/* 18352 */ {(12<<2)|2,{115,122,0}},
/* 18353 */ {(12<<2)|2,{115,122,0}},
/* 18354 */ {(12<<2)|2,{115,122,0}},
/* 18355 */ {(12<<2)|2,{115,122,0}},
/* 18356 */ {(12<<2)|2,{115,122,0}},
/* 18357 */ {(12<<2)|2,{115,122,0}},
/* 18358 */ {(12<<2)|2,{115,122,0}},
/* 18359 */ {(12<<2)|2,{115,122,0}},
/* 18360 */ {(12<<2)|2,{115,122,0}},
/* 18361 */ {(12<<2)|2,{115,122,0}},
/* 18362 */ {(12<<2)|2,{115,122,0}},
/* 18363 */ {(12<<2)|2,{115,122,0}},
/* 18364 */ {(12<<2)|2,{115,122,0}},
/* 18365 */ {(12<<2)|2,{115,122,0}},
/* 18366 */ {(12<<2)|2,{115,122,0}},
/* 18367 */ {(12<<2)|2,{115,122,0}},
/* 18368 */ {(13<<2)|2,{115,38,0}},
/* 18369 */ {(13<<2)|2,{115,38,0}},
/* 18370 */ {(13<<2)|2,{115,38,0}},
/* 18371 */ {(13<<2)|2,{115,38,0}},
/* 18372 */ {(13<<2)|2,{115,38,0}},
/* 18373 */ {(13<<2)|2,{115,38,0}},
/* 18374 */ {(13<<2)|2,{115,38,0}},
/* 18375 */ {(13<<2)|2,{115,38,0}},
/* 18376 */ {(13<<2)|2,{115,42,0}},
/* 18377 */ {(13<<2)|2,{115,42,0}},
/* 18378 */ {(13<<2)|2,{115,42,0}},
/* 18379 */ {(13<<2)|2,{115,42,0}},
/* 18380 */ {(13<<2)|2,{115,42,0}},
/* 18381 */ {(13<<2)|2,{115,42,0}},
/* 18382 */ {(13<<2)|2,{115,42,0}},
/* 18383 */ {(13<<2)|2,{115,42,0}},
/* 18384 */ {(13<<2)|2,{115,44,0}},
/* 18385 */ {(13<<2)|2,{115,44,0}},
/* 18386 */ {(13<<2)|2,{115,44,0}},
/* 18387 */ {(13<<2)|2,{115,44,0}},
/* 18388 */ {(13<<2)|2,{115,44,0}},
/* 18389 */ {(13<<2)|2,{115,44,0}},
/* 18390 */ {(13<<2)|2,{115,44,0}},
/* 18391 */ {(13<<2)|2,{115,44,0}},
/* 18392 */ {(13<<2)|2,{115,59,0}},
/* 18393 */ {(13<<2)|2,{115,59,0}},
/* 18394 */ {(13<<2)|2,{115,59,0}},
/* 18395 */ {(13<<2)|2,{115,59,0}},
/* 18396 */ {(13<<2)|2,{115,59,0}},
/* 18397 */ {(13<<2)|2,{115,59,0}},
/* 18398 */ {(13<<2)|2,{115,59,0}},
/* 18399 */ {(13<<2)|2,{115,59,0}},
/* 18400 */ {(13<<2)|2,{115,88,0}},
/* 18401 */ {(13<<2)|2,{115,88,0}},
/* 18402 */ {(13<<2)|2,{115,88,0}},
/* 18403 */ {(13<<2)|2,{115,88,0}},
/* 18404 */ {(13<<2)|2,{115,88,0}},
/* 18405 */ {(13<<2)|2,{115,88,0}},
/* 18406 */ {(13<<2)|2,{115,88,0}},
/* 18407 */ {(13<<2)|2,{115,88,0}},
/* 18408 */ {(13<<2)|2,{115,90,0}},
/* 18409 */ {(13<<2)|2,{115,90,0}},
/* 18410 */ {(13<<2)|2,{115,90,0}},
/* 18411 */ {(13<<2)|2,{115,90,0}},
/* 18412 */ {(13<<2)|2,{115,90,0}},
/* 18413 */ {(13<<2)|2,{115,90,0}},
/* 18414 */ {(13<<2)|2,{115,90,0}},
/* 18415 */ {(13<<2)|2,{115,90,0}},
/* 18416 */ {(15<<2)|2,{115,33,0}},
/* 18417 */ {(15<<2)|2,{115,33,0}},
/* 18418 */ {(15<<2)|2,{115,34,0}},
/* 18419 */ {(15<<2)|2,{115,34,0}},
/* 18420 */ {(15<<2)|2,{115,40,0}},
/* 18421 */ {(15<<2)|2,{115,40,0}},
/* 18422 */ {(15<<2)|2,{115,41,0}},
/* 18423 */ {(15<<2)|2,{115,41,0}},
/* 18424 */ {(15<<2)|2,{115,63,0}},
/* 18425 */ {(15<<2)|2,{115,63,0}},
/* 18426 */ {(16<<2)|2,{115,39,0}},
/* 18427 */ {(16<<2)|2,{115,43,0}},
/* 18428 */ {(16<<2)|2,{115,124,0}},
/* 18429 */ {(5<<2)|1,{115,0,0}},
/* 18430 */ {(5<<2)|1,{115,0,0}},
/* 18431 */ {(5<<2)|1,{115,0,0}},
/* 18432 */ {(15<<2)|3,{116,48,48}},
/* 18433 */ {(15<<2)|3,{116,48,48}},
/* 18434 */ {(15<<2)|3,{116,48,49}},
/* 18435 */ {(15<<2)|3,{116,48,49}},
/* 18436 */ {(15<<2)|3,{116,48,50}},
/* 18437 */ {(15<<2)|3,{116,48,50}},
/* 18438 */ {(15<<2)|3,{116,48,97}},
/* 18439 */ {(15<<2)|3,{116,48,97}},
/* 18440 */ {(15<<2)|3,{116,48,99}},
/* 18441 */ {(15<<2)|3,{116,48,99}},
/* 18442 */ {(15<<2)|3,{116,48,101}},
/* 18443 */ {(15<<2)|3,{116,48,101}},
/* 18444 */ {(15<<2)|3,{116,48,105}},
/* 18445 */ {(15<<2)|3,{116,48,105}},
/* 18446 */ {(15<<2)|3,{116,48,111}},
/* 18447 */ {(15<<2)|3,{116,48,111}},
/* 18448 */ {(15<<2)|3,{116,48,115}},
/* 18449 */ {(15<<2)|3,{116,48,115}},
/* 18450 */ {(15<<2)|3,{116,48,116}},
/* 18451 */ {(15<<2)|3,{116,48,116}},
/* 18452 */ {(16<<2)|3,{116,48,32}},
/* 18453 */ {(16<<2)|3,{116,48,37}},
/* 18454 */ {(16<<2)|3,{116,48,45}},
/* 18455 */ {(16<<2)|3,{116,48,46}},
/* 18456 */ {(16<<2)|3,{116,48,47}},
/* 18457 */ {(16<<2)|3,{116,48,51}},
/* 18458 */ {(16<<2)|3,{116,48,52}},
/* 18459 */ {(16<<2)|3,{116,48,53}},
/* 18460 */ {(16<<2)|3,{116,48,54}},
/* 18461 */ {(16<<2)|3,{116,48,55}},
/* 18462 */ {(16<<2)|3,{116,48,56}},
/* 18463 */ {(16<<2)|3,{116,48,57}},
/* 18464 */ {(16<<2)|3,{116,48,61}},
/* 18465 */ {(16<<2)|3,{116,48,65}},
/* 18466 */ {(16<<2)|3,{116,48,95}},
/* 18467 */ {(16<<2)|3,{116,48,98}},
/* 18468 */ {(16<<2)|3,{116,48,100}},
/* 18469 */ {(16<<2)|3,{116,48,102}},
/* 18470 */ {(16<<2)|3,{116,48,103}},
/* 18471 */ {(16<<2)|3,{116,48,104}},
/* 18472 */ {(16<<2)|3,{116,48,108}},
/* 18473 */ {(16<<2)|3,{116,48,109}},
/* 18474 */ {(16<<2)|3,{116,48,110}},
/* 18475 */ {(16<<2)|3,{116,48,112}},
/* 18476 */ {(16<<2)|3,{116,48,114}},
/* 18477 */ {(16<<2)|3,{116,48,117}},
/* 18478 */ {(10<<2)|2,{116,48,0}},
/* 18479 */ {(10<<2)|2,{116,48,0}},
/* 18480 */ {(10<<2)|2,{116,48,0}},
/* 18481 */ {(10<<2)|2,{116,48,0}},
/* 18482 */ {(10<<2)|2,{116,48,0}},
/* 18483 */ {(10<<2)|2,{116,48,0}},
/* 18484 */ {(10<<2)|2,{116,48,0}},
/* 18485 */ {(10<<2)|2,{116,48,0}},
/* 18486 */ {(10<<2)|2,{116,48,0}},
/* 18487 */ {(10<<2)|2,{116,48,0}},
/* 18488 */ {(10<<2)|2,{116,48,0}},
/* 18489 */ {(10<<2)|2,{116,48,0}},
/* 18490 */ {(10<<2)|2,{116,48,0}},
/* 18491 */ {(10<<2)|2,{116,48,0}},
/* 18492 */ {(10<<2)|2,{116,48,0}},
/* 18493 */ {(10<<2)|2,{116,48,0}},
/* 18494 */ {(10<<2)|2,{116,48,0}},
/* 18495 */ {(10<<2)|2,{116,48,0}},
/* 18496 */ {(15<<2)|3,{116,49,48}},
/* 18497 */ {(15<<2)|3,{116,49,48}},
/* 18498 */ {(15<<2)|3,{116,49,49}},
/* 18499 */ {(15<<2)|3,{116,49,49}},
/* 18500 */ {(15<<2)|3,{116,49,50}},
/* 18501 */ {(15<<2)|3,{116,49,50}},
/* 18502 */ {(15<<2)|3,{116,49,97}},
/* 18503 */ {(15<<2)|3,{116,49,97}},
/* 18504 */ {(15<<2)|3,{116,49,99}},
/* 18505 */ {(15<<2)|3,{116,49,99}},
/* 18506 */ {(15<<2)|3,{116,49,101}},
/* 18507 */ {(15<<2)|3,{116,49,101}},
/* 18508 */ {(15<<2)|3,{116,49,105}},
/* 18509 */ {(15<<2)|3,{116,49,105}},
/* 18510 */ {(15<<2)|3,{116,49,111}},
/* 18511 */ {(15<<2)|3,{116,49,111}},
/* 18512 */ {(15<<2)|3,{116,49,115}},
/* 18513 */ {(15<<2)|3,{116,49,115}},
/* 18514 */ {(15<<2)|3,{116,49,116}},
/* 18515 */ {(15<<2)|3,{116,49,116}},
/* 18516 */ {(16<<2)|3,{116,49,32}},
/* 18517 */ {(16<<2)|3,{116,49,37}},
/* 18518 */ {(16<<2)|3,{116,49,45}},
/* 18519 */ {(16<<2)|3,{116,49,46}},
/* 18520 */ {(16<<2)|3,{116,49,47}},
/* 18521 */ {(16<<2)|3,{116,49,51}},
/* 18522 */ {(16<<2)|3,{116,49,52}},
/* 18523 */ {(16<<2)|3,{116,49,53}},
/* 18524 */ {(16<<2)|3,{116,49,54}},
/* 18525 */ {(16<<2)|3,{116,49,55}},
/* 18526 */ {(16<<2)|3,{116,49,56}},
/* 18527 */ {(16<<2)|3,{116,49,57}},
/* 18528 */ {(16<<2)|3,{116,49,61}},
/* 18529 */ {(16<<2)|3,{116,49,65}},
/* 18530 */ {(16<<2)|3,{116,49,95}},
/* 18531 */ {(16<<2)|3,{116,49,98}},
/* 18532 */ {(16<<2)|3,{116,49,100}},
/* 18533 */ {(16<<2)|3,{116,49,102}},
/* 18534 */ {(16<<2)|3,{116,49,103}},
/* 18535 */ {(16<<2)|3,{116,49,104}},
/* 18536 */ {(16<<2)|3,{116,49,108}},
/* 18537 */ {(16<<2)|3,{116,49,109}},
/* 18538 */ {(16<<2)|3,{116,49,110}},
/* 18539 */ {(16<<2)|3,{116,49,112}},
/* 18540 */ {(16<<2)|3,{116,49,114}},
/* 18541 */ {(16<<2)|3,{116,49,117}},
/* 18542 */ {(10<<2)|2,{116,49,0}},
/* 18543 */ {(10<<2)|2,{116,49,0}},
/* 18544 */ {(10<<2)|2,{116,49,0}},
/* 18545 */ {(10<<2)|2,{116,49,0}},
/* 18546 */ {(10<<2)|2,{116,49,0}},
/* 18547 */ {(10<<2)|2,{116,49,0}},
/* 18548 */ {(10<<2)|2,{116,49,0}},
/* 18549 */ {(10<<2)|2,{116,49,0}},
/* 18550 */ {(10<<2)|2,{116,49,0}},
/* 18551 */ {(10<<2)|2,{116,49,0}},
/* 18552 */ {(10<<2)|2,{116,49,0}},
/* 18553 */ {(10<<2)|2,{116,49,0}},
/* 18554 */ {(10<<2)|2,{116,49,0}},
/* 18555 */ {(10<<2)|2,{116,49,0}},
/* 18556 */ {(10<<2)|2,{116,49,0}},
/* 18557 */ {(10<<2)|2,{116,49,0}},
/* 18558 */ {(10<<2)|2,{116,49,0}},
/* 18559 */ {(10<<2)|2,{116,49,0}},
/* 18560 */ {(15<<2)|3,{116,50,48}},
/* 18561 */ {(15<<2)|3,{116,50,48}},
/* 18562 */ {(15<<2)|3,{116,50,49}},
/* 18563 */ {(15<<2)|3,{116,50,49}},
/* 18564 */ {(15<<2)|3,{116,50,50}},
/* 18565 */ {(15<<2)|3,{116,50,50}},
/* 18566 */ {(15<<2)|3,{116,50,97}},
/* 18567 */ {(15<<2)|3,{116,50,97}},
/* 18568 */ {(15<<2)|3,{116,50,99}},
/* 18569 */ {(15<<2)|3,{116,50,99}},
/* 18570 */ {(15<<2)|3,{116,50,101}},
/* 18571 */ {(15<<2)|3,{116,50,101}},
/* 18572 */ {(15<<2)|3,{116,50,105}},
/* 18573 */ {(15<<2)|3,{116,50,105}},
/* 18574 */ {(15<<2)|3,{116,50,111}},
/* 18575 */ {(15<<2)|3,{116,50,111}},
/* 18576 */ {(15<<2)|3,{116,50,115}},
/* 18577 */ {(15<<2)|3,{116,50,115}},
/* 18578 */ {(15<<2)|3,{116,50,116}},
/* 18579 */ {(15<<2)|3,{116,50,116}},
/* 18580 */ {(16<<2)|3,{116,50,32}},
/* 18581 */ {(16<<2)|3,{116,50,37}},
/* 18582 */ {(16<<2)|3,{116,50,45}},
/* 18583 */ {(16<<2)|3,{116,50,46}},
/* 18584 */ {(16<<2)|3,{116,50,47}},
/* 18585 */ {(16<<2)|3,{116,50,51}},
/* 18586 */ {(16<<2)|3,{116,50,52}},
/* 18587 */ {(16<<2)|3,{116,50,53}},
/* 18588 */ {(16<<2)|3,{116,50,54}},
/* 18589 */ {(16<<2)|3,{116,50,55}},
/* 18590 */ {(16<<2)|3,{116,50,56}},
/* 18591 */ {(16<<2)|3,{116,50,57}},
/* 18592 */ {(16<<2)|3,{116,50,61}},
/* 18593 */ {(16<<2)|3,{116,50,65}},
/* 18594 */ {(16<<2)|3,{116,50,95}},
/* 18595 */ {(16<<2)|3,{116,50,98}},
/* 18596 */ {(16<<2)|3,{116,50,100}},
/* 18597 */ {(16<<2)|3,{116,50,102}},
/* 18598 */ {(16<<2)|3,{116,50,103}},
/* 18599 */ {(16<<2)|3,{116,50,104}},
/* 18600 */ {(16<<2)|3,{116,50,108}},
/* 18601 */ {(16<<2)|3,{116,50,109}},
/* 18602 */ {(16<<2)|3,{116,50,110}},
/* 18603 */ {(16<<2)|3,{116,50,112}},
/* 18604 */ {(16<<2)|3,{116,50,114}},
/* 18605 */ {(16<<2)|3,{116,50,117}},
/* 18606 */ {(10<<2)|2,{116,50,0}},
/* 18607 */ {(10<<2)|2,{116,50,0}},
/* 18608 */ {(10<<2)|2,{116,50,0}},
/* 18609 */ {(10<<2)|2,{116,50,0}},
/* 18610 */ {(10<<2)|2,{116,50,0}},
/* 18611 */ {(10<<2)|2,{116,50,0}},
/* 18612 */ {(10<<2)|2,{116,50,0}},
/* 18613 */ {(10<<2)|2,{116,50,0}},
/* 18614 */ {(10<<2)|2,{116,50,0}},
/* 18615 */ {(10<<2)|2,{116,50,0}},
/* 18616 */ {(10<<2)|2,{116,50,0}},
/* 18617 */ {(10<<2)|2,{116,50,0}},
/* 18618 */ {(10<<2)|2,{116,50,0}},
/* 18619 */ {(10<<2)|2,{116,50,0}},
/* 18620 */ {(10<<2)|2,{116,50,0}},
/* 18621 */ {(10<<2)|2,{116,50,0}},
/* 18622 */ {(10<<2)|2,{116,50,0}},
/* 18623 */ {(10<<2)|2,{116,50,0}},
/* 18624 */ {(15<<2)|3,{116,97,48}},
/* 18625 */ {(15<<2)|3,{116,97,48}},
/* 18626 */ {(15<<2)|3,{116,97,49}},
/* 18627 */ {(15<<2)|3,{116,97,49}},
/* 18628 */ {(15<<2)|3,{116,97,50}},
/* 18629 */ {(15<<2)|3,{116,97,50}},
/* 18630 */ {(15<<2)|3,{116,97,97}},
/* 18631 */ {(15<<2)|3,{116,97,97}},
/* 18632 */ {(15<<2)|3,{116,97,99}},
/* 18633 */ {(15<<2)|3,{116,97,99}},
/* 18634 */ {(15<<2)|3,{116,97,101}},
/* 18635 */ {(15<<2)|3,{116,97,101}},
/* 18636 */ {(15<<2)|3,{116,97,105}},
/* 18637 */ {(15<<2)|3,{116,97,105}},
/* 18638 */ {(15<<2)|3,{116,97,111}},
/* 18639 */ {(15<<2)|3,{116,97,111}},
/* 18640 */ {(15<<2)|3,{116,97,115}},
/* 18641 */ {(15<<2)|3,{116,97,115}},
/* 18642 */ {(15<<2)|3,{116,97,116}},
/* 18643 */ {(15<<2)|3,{116,97,116}},
/* 18644 */ {(16<<2)|3,{116,97,32}},
/* 18645 */ {(16<<2)|3,{116,97,37}},
/* 18646 */ {(16<<2)|3,{116,97,45}},
/* 18647 */ {(16<<2)|3,{116,97,46}},
/* 18648 */ {(16<<2)|3,{116,97,47}},
/* 18649 */ {(16<<2)|3,{116,97,51}},
/* 18650 */ {(16<<2)|3,{116,97,52}},
/* 18651 */ {(16<<2)|3,{116,97,53}},
/* 18652 */ {(16<<2)|3,{116,97,54}},
/* 18653 */ {(16<<2)|3,{116,97,55}},
/* 18654 */ {(16<<2)|3,{116,97,56}},
/* 18655 */ {(16<<2)|3,{116,97,57}},
/* 18656 */ {(16<<2)|3,{116,97,61}},
/* 18657 */ {(16<<2)|3,{116,97,65}},
/* 18658 */ {(16<<2)|3,{116,97,95}},
/* 18659 */ {(16<<2)|3,{116,97,98}},
/* 18660 */ {(16<<2)|3,{116,97,100}},
/* 18661 */ {(16<<2)|3,{116,97,102}},
/* 18662 */ {(16<<2)|3,{116,97,103}},
/* 18663 */ {(16<<2)|3,{116,97,104}},
/* 18664 */ {(16<<2)|3,{116,97,108}},
/* 18665 */ {(16<<2)|3,{116,97,109}},
/* 18666 */ {(16<<2)|3,{116,97,110}},
/* 18667 */ {(16<<2)|3,{116,97,112}},
/* 18668 */ {(16<<2)|3,{116,97,114}},
/* 18669 */ {(16<<2)|3,{116,97,117}},
/* 18670 */ {(10<<2)|2,{116,97,0}},
/* 18671 */ {(10<<2)|2,{116,97,0}},
/* 18672 */ {(10<<2)|2,{116,97,0}},
/* 18673 */ {(10<<2)|2,{116,97,0}},
/* 18674 */ {(10<<2)|2,{116,97,0}},
/* 18675 */ {(10<<2)|2,{116,97,0}},
/* 18676 */ {(10<<2)|2,{116,97,0}},
/* 18677 */ {(10<<2)|2,{116,97,0}},
/* 18678 */ {(10<<2)|2,{116,97,0}},
/* 18679 */ {(10<<2)|2,{116,97,0}},
/* 18680 */ {(10<<2)|2,{116,97,0}},
/* 18681 */ {(10<<2)|2,{116,97,0}},
/* 18682 */ {(10<<2)|2,{116,97,0}},
/* 18683 */ {(10<<2)|2,{116,97,0}},
/* 18684 */ {(10<<2)|2,{116,97,0}},
/* 18685 */ {(10<<2)|2,{116,97,0}},
/* 18686 */ {(10<<2)|2,{116,97,0}},
/* 18687 */ {(10<<2)|2,{116,97,0}},
/* 18688 */ {(15<<2)|3,{116,99,48}},
/* 18689 */ {(15<<2)|3,{116,99,48}},
/* 18690 */ {(15<<2)|3,{116,99,49}},
/* 18691 */ {(15<<2)|3,{116,99,49}},
/* 18692 */ {(15<<2)|3,{116,99,50}},
/* 18693 */ {(15<<2)|3,{116,99,50}},
/* 18694 */ {(15<<2)|3,{116,99,97}},
/* 18695 */ {(15<<2)|3,{116,99,97}},
/* 18696 */ {(15<<2)|3,{116,99,99}},
/* 18697 */ {(15<<2)|3,{116,99,99}},
/* 18698 */ {(15<<2)|3,{116,99,101}},
/* 18699 */ {(15<<2)|3,{116,99,101}},
/* 18700 */ {(15<<2)|3,{116,99,105}},
/* 18701 */ {(15<<2)|3,{116,99,105}},
/* 18702 */ {(15<<2)|3,{116,99,111}},
/* 18703 */ {(15<<2)|3,{116,99,111}},
/* 18704 */ {(15<<2)|3,{116,99,115}},
/* 18705 */ {(15<<2)|3,{116,99,115}},
/* 18706 */ {(15<<2)|3,{116,99,116}},
/* 18707 */ {(15<<2)|3,{116,99,116}},
/* 18708 */ {(16<<2)|3,{116,99,32}},
/* 18709 */ {(16<<2)|3,{116,99,37}},
/* 18710 */ {(16<<2)|3,{116,99,45}},
/* 18711 */ {(16<<2)|3,{116,99,46}},
/* 18712 */ {(16<<2)|3,{116,99,47}},
/* 18713 */ {(16<<2)|3,{116,99,51}},
/* 18714 */ {(16<<2)|3,{116,99,52}},
/* 18715 */ {(16<<2)|3,{116,99,53}},
/* 18716 */ {(16<<2)|3,{116,99,54}},
/* 18717 */ {(16<<2)|3,{116,99,55}},
/* 18718 */ {(16<<2)|3,{116,99,56}},
/* 18719 */ {(16<<2)|3,{116,99,57}},
/* 18720 */ {(16<<2)|3,{116,99,61}},
/* 18721 */ {(16<<2)|3,{116,99,65}},
/* 18722 */ {(16<<2)|3,{116,99,95}},
/* 18723 */ {(16<<2)|3,{116,99,98}},
/* 18724 */ {(16<<2)|3,{116,99,100}},
/* 18725 */ {(16<<2)|3,{116,99,102}},
/* 18726 */ {(16<<2)|3,{116,99,103}},
/* 18727 */ {(16<<2)|3,{116,99,104}},
/* 18728 */ {(16<<2)|3,{116,99,108}},
/* 18729 */ {(16<<2)|3,{116,99,109}},
/* 18730 */ {(16<<2)|3,{116,99,110}},
/* 18731 */ {(16<<2)|3,{116,99,112}},
/* 18732 */ {(16<<2)|3,{116,99,114}},
/* 18733 */ {(16<<2)|3,{116,99,117}},
/* 18734 */ {(10<<2)|2,{116,99,0}},
/* 18735 */ {(10<<2)|2,{116,99,0}},
/* 18736 */ {(10<<2)|2,{116,99,0}},
/* 18737 */ {(10<<2)|2,{116,99,0}},
/* 18738 */ {(10<<2)|2,{116,99,0}},
/* 18739 */ {(10<<2)|2,{116,99,0}},
/* 18740 */ {(10<<2)|2,{116,99,0}},
/* 18741 */ {(10<<2)|2,{116,99,0}},
/* 18742 */ {(10<<2)|2,{116,99,0}},
/* 18743 */ {(10<<2)|2,{116,99,0}},
/* 18744 */ {(10<<2)|2,{116,99,0}},
/* 18745 */ {(10<<2)|2,{116,99,0}},
/* 18746 */ {(10<<2)|2,{116,99,0}},
/* 18747 */ {(10<<2)|2,{116,99,0}},
/* 18748 */ {(10<<2)|2,{116,99,0}},
/* 18749 */ {(10<<2)|2,{116,99,0}},
/* 18750 */ {(10<<2)|2,{116,99,0}},
/* 18751 */ {(10<<2)|2,{116,99,0}},
/* 18752 */ {(15<<2)|3,{116,101,48}},
/* 18753 */ {(15<<2)|3,{116,101,48}},
/* 18754 */ {(15<<2)|3,{116,101,49}},
/* 18755 */ {(15<<2)|3,{116,101,49}},
/* 18756 */ {(15<<2)|3,{116,101,50}},
/* 18757 */ {(15<<2)|3,{116,101,50}},
/* 18758 */ {(15<<2)|3,{116,101,97}},
/* 18759 */ {(15<<2)|3,{116,101,97}},
/* 18760 */ {(15<<2)|3,{116,101,99}},
/* 18761 */ {(15<<2)|3,{116,101,99}},
/* 18762 */ {(15<<2)|3,{116,101,101}},
/* 18763 */ {(15<<2)|3,{116,101,101}},
/* 18764 */ {(15<<2)|3,{116,101,105}},
/* 18765 */ {(15<<2)|3,{116,101,105}},
/* 18766 */ {(15<<2)|3,{116,101,111}},
/* 18767 */ {(15<<2)|3,{116,101,111}},
/* 18768 */ {(15<<2)|3,{116,101,115}},
/* 18769 */ {(15<<2)|3,{116,101,115}},
/* 18770 */ {(15<<2)|3,{116,101,116}},
/* 18771 */ {(15<<2)|3,{116,101,116}},
/* 18772 */ {(16<<2)|3,{116,101,32}},
/* 18773 */ {(16<<2)|3,{116,101,37}},
/* 18774 */ {(16<<2)|3,{116,101,45}},
/* 18775 */ {(16<<2)|3,{116,101,46}},
/* 18776 */ {(16<<2)|3,{116,101,47}},
/* 18777 */ {(16<<2)|3,{116,101,51}},
/* 18778 */ {(16<<2)|3,{116,101,52}},
/* 18779 */ {(16<<2)|3,{116,101,53}},
/* 18780 */ {(16<<2)|3,{116,101,54}},
/* 18781 */ {(16<<2)|3,{116,101,55}},
/* 18782 */ {(16<<2)|3,{116,101,56}},
/* 18783 */ {(16<<2)|3,{116,101,57}},
/* 18784 */ {(16<<2)|3,{116,101,61}},
/* 18785 */ {(16<<2)|3,{116,101,65}},
/* 18786 */ {(16<<2)|3,{116,101,95}},
/* 18787 */ {(16<<2)|3,{116,101,98}},
/* 18788 */ {(16<<2)|3,{116,101,100}},
/* 18789 */ {(16<<2)|3,{116,101,102}},
/* 18790 */ {(16<<2)|3,{116,101,103}},
/* 18791 */ {(16<<2)|3,{116,101,104}},
/* 18792 */ {(16<<2)|3,{116,101,108}},
/* 18793 */ {(16<<2)|3,{116,101,109}},
/* 18794 */ {(16<<2)|3,{116,101,110}},
/* 18795 */ {(16<<2)|3,{116,101,112}},
/* 18796 */ {(16<<2)|3,{116,101,114}},
/* 18797 */ {(16<<2)|3,{116,101,117}},
/* 18798 */ {(10<<2)|2,{116,101,0}},
/* 18799 */ {(10<<2)|2,{116,101,0}},
/* 18800 */ {(10<<2)|2,{116,101,0}},
/* 18801 */ {(10<<2)|2,{116,101,0}},
/* 18802 */ {(10<<2)|2,{116,101,0}},
/* 18803 */ {(10<<2)|2,{116,101,0}},
/* 18804 */ {(10<<2)|2,{116,101,0}},
/* 18805 */ {(10<<2)|2,{116,101,0}},
/* 18806 */ {(10<<2)|2,{116,101,0}},
/* 18807 */ {(10<<2)|2,{116,101,0}},
/* 18808 */ {(10<<2)|2,{116,101,0}},
/* 18809 */ {(10<<2)|2,{116,101,0}},
/* 18810 */ {(10<<2)|2,{116,101,0}},
/* 18811 */ {(10<<2)|2,{116,101,0}},
/* 18812 */ {(10<<2)|2,{116,101,0}},
/* 18813 */ {(10<<2)|2,{116,101,0}},
/* 18814 */ {(10<<2)|2,{116,101,0}},
/* 18815 */ {(10<<2)|2,{116,101,0}},
/* 18816 */ {(15<<2)|3,{116,105,48}},
/* 18817 */ {(15<<2)|3,{116,105,48}},
/* 18818 */ {(15<<2)|3,{116,105,49}},
/* 18819 */ {(15<<2)|3,{116,105,49}},
/* 18820 */ {(15<<2)|3,{116,105,50}},
/* 18821 */ {(15<<2)|3,{116,105,50}},
/* 18822 */ {(15<<2)|3,{116,105,97}},
/* 18823 */ {(15<<2)|3,{116,105,97}},
/* 18824 */ {(15<<2)|3,{116,105,99}},
/* 18825 */ {(15<<2)|3,{116,105,99}},
/* 18826 */ {(15<<2)|3,{116,105,101}},
/* 18827 */ {(15<<2)|3,{116,105,101}},
/* 18828 */ {(15<<2)|3,{116,105,105}},
/* 18829 */ {(15<<2)|3,{116,105,105}},
/* 18830 */ {(15<<2)|3,{116,105,111}},
/* 18831 */ {(15<<2)|3,{116,105,111}},
/* 18832 */ {(15<<2)|3,{116,105,115}},
/* 18833 */ {(15<<2)|3,{116,105,115}},
/* 18834 */ {(15<<2)|3,{116,105,116}},
/* 18835 */ {(15<<2)|3,{116,105,116}},
/* 18836 */ {(16<<2)|3,{116,105,32}},
/* 18837 */ {(16<<2)|3,{116,105,37}},
/* 18838 */ {(16<<2)|3,{116,105,45}},
/* 18839 */ {(16<<2)|3,{116,105,46}},
/* 18840 */ {(16<<2)|3,{116,105,47}},
/* 18841 */ {(16<<2)|3,{116,105,51}},
/* 18842 */ {(16<<2)|3,{116,105,52}},
/* 18843 */ {(16<<2)|3,{116,105,53}},
/* 18844 */ {(16<<2)|3,{116,105,54}},
/* 18845 */ {(16<<2)|3,{116,105,55}},
/* 18846 */ {(16<<2)|3,{116,105,56}},
/* 18847 */ {(16<<2)|3,{116,105,57}},
/* 18848 */ {(16<<2)|3,{116,105,61}},
/* 18849 */ {(16<<2)|3,{116,105,65}},
/* 18850 */ {(16<<2)|3,{116,105,95}},
/* 18851 */ {(16<<2)|3,{116,105,98}},
/* 18852 */ {(16<<2)|3,{116,105,100}},
/* 18853 */ {(16<<2)|3,{116,105,102}},
/* 18854 */ {(16<<2)|3,{116,105,103}},
/* 18855 */ {(16<<2)|3,{116,105,104}},
/* 18856 */ {(16<<2)|3,{116,105,108}},
/* 18857 */ {(16<<2)|3,{116,105,109}},
/* 18858 */ {(16<<2)|3,{116,105,110}},
/* 18859 */ {(16<<2)|3,{116,105,112}},
/* 18860 */ {(16<<2)|3,{116,105,114}},
/* 18861 */ {(16<<2)|3,{116,105,117}},
/* 18862 */ {(10<<2)|2,{116,105,0}},
/* 18863 */ {(10<<2)|2,{116,105,0}},
/* 18864 */ {(10<<2)|2,{116,105,0}},
/* 18865 */ {(10<<2)|2,{116,105,0}},
/* 18866 */ {(10<<2)|2,{116,105,0}},
/* 18867 */ {(10<<2)|2,{116,105,0}},
/* 18868 */ {(10<<2)|2,{116,105,0}},
/* 18869 */ {(10<<2)|2,{116,105,0}},
/* 18870 */ {(10<<2)|2,{116,105,0}},
/* 18871 */ {(10<<2)|2,{116,105,0}},
/* 18872 */ {(10<<2)|2,{116,105,0}},
/* 18873 */ {(10<<2)|2,{116,105,0}},
/* 18874 */ {(10<<2)|2,{116,105,0}},
/* 18875 */ {(10<<2)|2,{116,105,0}},
/* 18876 */ {(10<<2)|2,{116,105,0}},
/* 18877 */ {(10<<2)|2,{116,105,0}},
/* 18878 */ {(10<<2)|2,{116,105,0}},
/* 18879 */ {(10<<2)|2,{116,105,0}},
/* 18880 */ {(15<<2)|3,{116,111,48}},
/* 18881 */ {(15<<2)|3,{116,111,48}},
/* 18882 */ {(15<<2)|3,{116,111,49}},
/* 18883 */ {(15<<2)|3,{116,111,49}},
/* 18884 */ {(15<<2)|3,{116,111,50}},
/* 18885 */ {(15<<2)|3,{116,111,50}},
/* 18886 */ {(15<<2)|3,{116,111,97}},
/* 18887 */ {(15<<2)|3,{116,111,97}},
/* 18888 */ {(15<<2)|3,{116,111,99}},
/* 18889 */ {(15<<2)|3,{116,111,99}},
/* 18890 */ {(15<<2)|3,{116,111,101}},
/* 18891 */ {(15<<2)|3,{116,111,101}},
/* 18892 */ {(15<<2)|3,{116,111,105}},
/* 18893 */ {(15<<2)|3,{116,111,105}},
/* 18894 */ {(15<<2)|3,{116,111,111}},
/* 18895 */ {(15<<2)|3,{116,111,111}},
/* 18896 */ {(15<<2)|3,{116,111,115}},
/* 18897 */ {(15<<2)|3,{116,111,115}},
/* 18898 */ {(15<<2)|3,{116,111,116}},
/* 18899 */ {(15<<2)|3,{116,111,116}},
/* 18900 */ {(16<<2)|3,{116,111,32}},
/* 18901 */ {(16<<2)|3,{116,111,37}},
/* 18902 */ {(16<<2)|3,{116,111,45}},
/* 18903 */ {(16<<2)|3,{116,111,46}},
/* 18904 */ {(16<<2)|3,{116,111,47}},
/* 18905 */ {(16<<2)|3,{116,111,51}},
/* 18906 */ {(16<<2)|3,{116,111,52}},
/* 18907 */ {(16<<2)|3,{116,111,53}},
/* 18908 */ {(16<<2)|3,{116,111,54}},
/* 18909 */ {(16<<2)|3,{116,111,55}},
/* 18910 */ {(16<<2)|3,{116,111,56}},
/* 18911 */ {(16<<2)|3,{116,111,57}},
/* 18912 */ {(16<<2)|3,{116,111,61}},
/* 18913 */ {(16<<2)|3,{116,111,65}},
/* 18914 */ {(16<<2)|3,{116,111,95}},
/* 18915 */ {(16<<2)|3,{116,111,98}},
/* 18916 */ {(16<<2)|3,{116,111,100}},
/* 18917 */ {(16<<2)|3,{116,111,102}},
/* 18918 */ {(16<<2)|3,{116,111,103}},
/* 18919 */ {(16<<2)|3,{116,111,104}},
/* 18920 */ {(16<<2)|3,{116,111,108}},
/* 18921 */ {(16<<2)|3,{116,111,109}},
/* 18922 */ {(16<<2)|3,{116,111,110}},
/* 18923 */ {(16<<2)|3,{116,111,112}},
/* 18924 */ {(16<<2)|3,{116,111,114}},
/* 18925 */ {(16<<2)|3,{116,111,117}},
/* 18926 */ {(10<<2)|2,{116,111,0}},
/* 18927 */ {(10<<2)|2,{116,111,0}},
/* 18928 */ {(10<<2)|2,{116,111,0}},
/* 18929 */ {(10<<2)|2,{116,111,0}},
/* 18930 */ {(10<<2)|2,{116,111,0}},
/* 18931 */ {(10<<2)|2,{116,111,0}},
/* 18932 */ {(10<<2)|2,{116,111,0}},
/* 18933 */ {(10<<2)|2,{116,111,0}},
/* 18934 */ {(10<<2)|2,{116,111,0}},
/* 18935 */ {(10<<2)|2,{116,111,0}},
/* 18936 */ {(10<<2)|2,{116,111,0}},
/* 18937 */ {(10<<2)|2,{116,111,0}},
/* 18938 */ {(10<<2)|2,{116,111,0}},
/* 18939 */ {(10<<2)|2,{116,111,0}},
/* 18940 */ {(10<<2)|2,{116,111,0}},
/* 18941 */ {(10<<2)|2,{116,111,0}},
/* 18942 */ {(10<<2)|2,{116,111,0}},
/* 18943 */ {(10<<2)|2,{116,111,0}},
/* 18944 */ {(15<<2)|3,{116,115,48}},
/* 18945 */ {(15<<2)|3,{116,115,48}},
/* 18946 */ {(15<<2)|3,{116,115,49}},
/* 18947 */ {(15<<2)|3,{116,115,49}},
/* 18948 */ {(15<<2)|3,{116,115,50}},
/* 18949 */ {(15<<2)|3,{116,115,50}},
/* 18950 */ {(15<<2)|3,{116,115,97}},
/* 18951 */ {(15<<2)|3,{116,115,97}},
/* 18952 */ {(15<<2)|3,{116,115,99}},
/* 18953 */ {(15<<2)|3,{116,115,99}},
/* 18954 */ {(15<<2)|3,{116,115,101}},
/* 18955 */ {(15<<2)|3,{116,115,101}},
/* 18956 */ {(15<<2)|3,{116,115,105}},
/* 18957 */ {(15<<2)|3,{116,115,105}},
/* 18958 */ {(15<<2)|3,{116,115,111}},
/* 18959 */ {(15<<2)|3,{116,115,111}},
/* 18960 */ {(15<<2)|3,{116,115,115}},
/* 18961 */ {(15<<2)|3,{116,115,115}},
/* 18962 */ {(15<<2)|3,{116,115,116}},
/* 18963 */ {(15<<2)|3,{116,115,116}},
/* 18964 */ {(16<<2)|3,{116,115,32}},
/* 18965 */ {(16<<2)|3,{116,115,37}},
/* 18966 */ {(16<<2)|3,{116,115,45}},
/* 18967 */ {(16<<2)|3,{116,115,46}},
/* 18968 */ {(16<<2)|3,{116,115,47}},
/* 18969 */ {(16<<2)|3,{116,115,51}},
/* 18970 */ {(16<<2)|3,{116,115,52}},
/* 18971 */ {(16<<2)|3,{116,115,53}},
/* 18972 */ {(16<<2)|3,{116,115,54}},
/* 18973 */ {(16<<2)|3,{116,115,55}},
/* 18974 */ {(16<<2)|3,{116,115,56}},
/* 18975 */ {(16<<2)|3,{116,115,57}},
/* 18976 */ {(16<<2)|3,{116,115,61}},
/* 18977 */ {(16<<2)|3,{116,115,65}},
/* 18978 */ {(16<<2)|3,{116,115,95}},
/* 18979 */ {(16<<2)|3,{116,115,98}},
/* 18980 */ {(16<<2)|3,{116,115,100}},
/* 18981 */ {(16<<2)|3,{116,115,102}},
/* 18982 */ {(16<<2)|3,{116,115,103}},
/* 18983 */ {(16<<2)|3,{116,115,104}},
/* 18984 */ {(16<<2)|3,{116,115,108}},
/* 18985 */ {(16<<2)|3,{116,115,109}},
/* 18986 */ {(16<<2)|3,{116,115,110}},
/* 18987 */ {(16<<2)|3,{116,115,112}},
/* 18988 */ {(16<<2)|3,{116,115,114}},
/* 18989 */ {(16<<2)|3,{116,115,117}},
/* 18990 */ {(10<<2)|2,{116,115,0}},
/* 18991 */ {(10<<2)|2,{116,115,0}},
/* 18992 */ {(10<<2)|2,{116,115,0}},
/* 18993 */ {(10<<2)|2,{116,115,0}},
/* 18994 */ {(10<<2)|2,{116,115,0}},
/* 18995 */ {(10<<2)|2,{116,115,0}},
/* 18996 */ {(10<<2)|2,{116,115,0}},
/* 18997 */ {(10<<2)|2,{116,115,0}},
/* 18998 */ {(10<<2)|2,{116,115,0}},
/* 18999 */ {(10<<2)|2,{116,115,0}},
/* 19000 */ {(10<<2)|2,{116,115,0}},
/* 19001 */ {(10<<2)|2,{116,115,0}},
/* 19002 */ {(10<<2)|2,{116,115,0}},
/* 19003 */ {(10<<2)|2,{116,115,0}},
/* 19004 */ {(10<<2)|2,{116,115,0}},
/* 19005 */ {(10<<2)|2,{116,115,0}},
/* 19006 */ {(10<<2)|2,{116,115,0}},
/* 19007 */ {(10<<2)|2,{116,115,0}},
/* 19008 */ {(15<<2)|3,{116,116,48}},
/* 19009 */ {(15<<2)|3,{116,116,48}},
/* 19010 */ {(15<<2)|3,{116,116,49}},
/* 19011 */ {(15<<2)|3,{116,116,49}},
/* 19012 */ {(15<<2)|3,{116,116,50}},
/* 19013 */ {(15<<2)|3,{116,116,50}},
/* 19014 */ {(15<<2)|3,{116,116,97}},
/* 19015 */ {(15<<2)|3,{116,116,97}},
/* 19016 */ {(15<<2)|3,{116,116,99}},
/* 19017 */ {(15<<2)|3,{116,116,99}},
/* 19018 */ {(15<<2)|3,{116,116,101}},
/* 19019 */ {(15<<2)|3,{116,116,101}},
/* 19020 */ {(15<<2)|3,{116,116,105}},
/* 19021 */ {(15<<2)|3,{116,116,105}},
/* 19022 */ {(15<<2)|3,{116,116,111}},
/* 19023 */ {(15<<2)|3,{116,116,111}},
/* 19024 */ {(15<<2)|3,{116,116,115}},
/* 19025 */ {(15<<2)|3,{116,116,115}},
/* 19026 */ {(15<<2)|3,{116,116,116}},
/* 19027 */ {(15<<2)|3,{116,116,116}},
/* 19028 */ {(16<<2)|3,{116,116,32}},
/* 19029 */ {(16<<2)|3,{116,116,37}},
/* 19030 */ {(16<<2)|3,{116,116,45}},
/* 19031 */ {(16<<2)|3,{116,116,46}},
/* 19032 */ {(16<<2)|3,{116,116,47}},
/* 19033 */ {(16<<2)|3,{116,116,51}},
/* 19034 */ {(16<<2)|3,{116,116,52}},
/* 19035 */ {(16<<2)|3,{116,116,53}},
/* 19036 */ {(16<<2)|3,{116,116,54}},
/* 19037 */ {(16<<2)|3,{116,116,55}},
/* 19038 */ {(16<<2)|3,{116,116,56}},
/* 19039 */ {(16<<2)|3,{116,116,57}},
/* 19040 */ {(16<<2)|3,{116,116,61}},
/* 19041 */ {(16<<2)|3,{116,116,65}},
/* 19042 */ {(16<<2)|3,{116,116,95}},
/* 19043 */ {(16<<2)|3,{116,116,98}},
/* 19044 */ {(16<<2)|3,{116,116,100}},
/* 19045 */ {(16<<2)|3,{116,116,102}},
/* 19046 */ {(16<<2)|3,{116,116,103}},
/* 19047 */ {(16<<2)|3,{116,116,104}},
/* 19048 */ {(16<<2)|3,{116,116,108}},
/* 19049 */ {(16<<2)|3,{116,116,109}},
/* 19050 */ {(16<<2)|3,{116,116,110}},
/* 19051 */ {(16<<2)|3,{116,116,112}},
/* 19052 */ {(16<<2)|3,{116,116,114}},
/* 19053 */ {(16<<2)|3,{116,116,117}},
/* 19054 */ {(10<<2)|2,{116,116,0}},
/* 19055 */ {(10<<2)|2,{116,116,0}},
/* 19056 */ {(10<<2)|2,{116,116,0}},
/* 19057 */ {(10<<2)|2,{116,116,0}},
/* 19058 */ {(10<<2)|2,{116,116,0}},
/* 19059 */ {(10<<2)|2,{116,116,0}},
/* 19060 */ {(10<<2)|2,{116,116,0}},
/* 19061 */ {(10<<2)|2,{116,116,0}},
/* 19062 */ {(10<<2)|2,{116,116,0}},
/* 19063 */ {(10<<2)|2,{116,116,0}},
/* 19064 */ {(10<<2)|2,{116,116,0}},
/* 19065 */ {(10<<2)|2,{116,116,0}},
/* 19066 */ {(10<<2)|2,{116,116,0}},
/* 19067 */ {(10<<2)|2,{116,116,0}},
/* 19068 */ {(10<<2)|2,{116,116,0}},
/* 19069 */ {(10<<2)|2,{116,116,0}},
/* 19070 */ {(10<<2)|2,{116,116,0}},
/* 19071 */ {(10<<2)|2,{116,116,0}},
/* 19072 */ {(16<<2)|3,{116,32,48}},
/* 19073 */ {(16<<2)|3,{116,32,49}},
/* 19074 */ {(16<<2)|3,{116,32,50}},
/* 19075 */ {(16<<2)|3,{116,32,97}},
/* 19076 */ {(16<<2)|3,{116,32,99}},
/* 19077 */ {(16<<2)|3,{116,32,101}},
/* 19078 */ {(16<<2)|3,{116,32,105}},
/* 19079 */ {(16<<2)|3,{116,32,111}},
/* 19080 */ {(16<<2)|3,{116,32,115}},
/* 19081 */ {(16<<2)|3,{116,32,116}},
/* 19082 */ {(11<<2)|2,{116,32,0}},
/* 19083 */ {(11<<2)|2,{116,32,0}},
/* 19084 */ {(11<<2)|2,{116,32,0}},
/* 19085 */ {(11<<2)|2,{116,32,0}},
/* 19086 */ {(11<<2)|2,{116,32,0}},
/* 19087 */ {(11<<2)|2,{116,32,0}},
/* 19088 */ {(11<<2)|2,{116,32,0}},
/* 19089 */ {(11<<2)|2,{116,32,0}},
/* 19090 */ {(11<<2)|2,{116,32,0}},
/* 19091 */ {(11<<2)|2,{116,32,0}},
/* 19092 */ {(11<<2)|2,{116,32,0}},
/* 19093 */ {(11<<2)|2,{116,32,0}},
/* 19094 */ {(11<<2)|2,{116,32,0}},
/* 19095 */ {(11<<2)|2,{116,32,0}},
/* 19096 */ {(11<<2)|2,{116,32,0}},
/* 19097 */ {(11<<2)|2,{116,32,0}},
/* 19098 */ {(11<<2)|2,{116,32,0}},
/* 19099 */ {(11<<2)|2,{116,32,0}},
/* 19100 */ {(11<<2)|2,{116,32,0}},
/* 19101 */ {(11<<2)|2,{116,32,0}},
/* 19102 */ {(11<<2)|2,{116,32,0}},
/* 19103 */ {(11<<2)|2,{116,32,0}},
/* 19104 */ {(16<<2)|3,{116,37,48}},
/* 19105 */ {(16<<2)|3,{116,37,49}},
/* 19106 */ {(16<<2)|3,{116,37,50}},
/* 19107 */ {(16<<2)|3,{116,37,97}},
/* 19108 */ {(16<<2)|3,{116,37,99}},
/* 19109 */ {(16<<2)|3,{116,37,101}},
/* 19110 */ {(16<<2)|3,{116,37,105}},
/* 19111 */ {(16<<2)|3,{116,37,111}},
/* 19112 */ {(16<<2)|3,{116,37,115}},
/* 19113 */ {(16<<2)|3,{116,37,116}},
/* 19114 */ {(11<<2)|2,{116,37,0}},
/* 19115 */ {(11<<2)|2,{116,37,0}},
/* 19116 */ {(11<<2)|2,{116,37,0}},
/* 19117 */ {(11<<2)|2,{116,37,0}},
/* 19118 */ {(11<<2)|2,{116,37,0}},
/* 19119 */ {(11<<2)|2,{116,37,0}},
/* 19120 */ {(11<<2)|2,{116,37,0}},
/* 19121 */ {(11<<2)|2,{116,37,0}},
/* 19122 */ {(11<<2)|2,{116,37,0}},
/* 19123 */ {(11<<2)|2,{116,37,0}},
/* 19124 */ {(11<<2)|2,{116,37,0}},
/* 19125 */ {(11<<2)|2,{116,37,0}},
/* 19126 */ {(11<<2)|2,{116,37,0}},
/* 19127 */ {(11<<2)|2,{116,37,0}},
/* 19128 */ {(11<<2)|2,{116,37,0}},
/* 19129 */ {(11<<2)|2,{116,37,0}},
/* 19130 */ {(11<<2)|2,{116,37,0}},
/* 19131 */ {(11<<2)|2,{116,37,0}},
/* 19132 */ {(11<<2)|2,{116,37,0}},
/* 19133 */ {(11<<2)|2,{116,37,0}},
/* 19134 */ {(11<<2)|2,{116,37,0}},
/* 19135 */ {(11<<2)|2,{116,37,0}},
/* 19136 */ {(16<<2)|3,{116,45,48}},
/* 19137 */ {(16<<2)|3,{116,45,49}},
/* 19138 */ {(16<<2)|3,{116,45,50}},
/* 19139 */ {(16<<2)|3,{116,45,97}},
/* 19140 */ {(16<<2)|3,{116,45,99}},
/* 19141 */ {(16<<2)|3,{116,45,101}},
/* 19142 */ {(16<<2)|3,{116,45,105}},
/* 19143 */ {(16<<2)|3,{116,45,111}},
/* 19144 */ {(16<<2)|3,{116,45,115}},
/* 19145 */ {(16<<2)|3,{116,45,116}},
/* 19146 */ {(11<<2)|2,{116,45,0}},
/* 19147 */ {(11<<2)|2,{116,45,0}},
/* 19148 */ {(11<<2)|2,{116,45,0}},
/* 19149 */ {(11<<2)|2,{116,45,0}},
/* 19150 */ {(11<<2)|2,{116,45,0}},
/* 19151 */ {(11<<2)|2,{116,45,0}},
/* 19152 */ {(11<<2)|2,{116,45,0}},
/* 19153 */ {(11<<2)|2,{116,45,0}},
/* 19154 */ {(11<<2)|2,{116,45,0}},
/* 19155 */ {(11<<2)|2,{116,45,0}},
/* 19156 */ {(11<<2)|2,{116,45,0}},
/* 19157 */ {(11<<2)|2,{116,45,0}},
/* 19158 */ {(11<<2)|2,{116,45,0}},
/* 19159 */ {(11<<2)|2,{116,45,0}},
/* 19160 */ {(11<<2)|2,{116,45,0}},
/* 19161 */ {(11<<2)|2,{116,45,0}},
/* 19162 */ {(11<<2)|2,{116,45,0}},
/* 19163 */ {(11<<2)|2,{116,45,0}},
/* 19164 */ {(11<<2)|2,{116,45,0}},
/* 19165 */ {(11<<2)|2,{116,45,0}},
/* 19166 */ {(11<<2)|2,{116,45,0}},
/* 19167 */ {(11<<2)|2,{116,45,0}},
/* 19168 */ {(16<<2)|3,{116,46,48}},
/* 19169 */ {(16<<2)|3,{116,46,49}},
/* 19170 */ {(16<<2)|3,{116,46,50}},
/* 19171 */ {(16<<2)|3,{116,46,97}},
/* 19172 */ {(16<<2)|3,{116,46,99}},
/* 19173 */ {(16<<2)|3,{116,46,101}},
/* 19174 */ {(16<<2)|3,{116,46,105}},
/* 19175 */ {(16<<2)|3,{116,46,111}},
/* 19176 */ {(16<<2)|3,{116,46,115}},
/* 19177 */ {(16<<2)|3,{116,46,116}},
/* 19178 */ {(11<<2)|2,{116,46,0}},
/* 19179 */ {(11<<2)|2,{116,46,0}},
/* 19180 */ {(11<<2)|2,{116,46,0}},
/* 19181 */ {(11<<2)|2,{116,46,0}},
/* 19182 */ {(11<<2)|2,{116,46,0}},
/* 19183 */ {(11<<2)|2,{116,46,0}},
/* 19184 */ {(11<<2)|2,{116,46,0}},
/* 19185 */ {(11<<2)|2,{116,46,0}},
/* 19186 */ {(11<<2)|2,{116,46,0}},
/* 19187 */ {(11<<2)|2,{116,46,0}},
/* 19188 */ {(11<<2)|2,{116,46,0}},
/* 19189 */ {(11<<2)|2,{116,46,0}},
/* 19190 */ {(11<<2)|2,{116,46,0}},
/* 19191 */ {(11<<2)|2,{116,46,0}},
/* 19192 */ {(11<<2)|2,{116,46,0}},
/* 19193 */ {(11<<2)|2,{116,46,0}},
/* 19194 */ {(11<<2)|2,{116,46,0}},
/* 19195 */ {(11<<2)|2,{116,46,0}},
/* 19196 */ {(11<<2)|2,{116,46,0}},
/* 19197 */ {(11<<2)|2,{116,46,0}},
/* 19198 */ {(11<<2)|2,{116,46,0}},
/* 19199 */ {(11<<2)|2,{116,46,0}},
/* 19200 */ {(16<<2)|3,{116,47,48}},
/* 19201 */ {(16<<2)|3,{116,47,49}},
/* 19202 */ {(16<<2)|3,{116,47,50}},
/* 19203 */ {(16<<2)|3,{116,47,97}},
/* 19204 */ {(16<<2)|3,{116,47,99}},
/* 19205 */ {(16<<2)|3,{116,47,101}},
/* 19206 */ {(16<<2)|3,{116,47,105}},
/* 19207 */ {(16<<2)|3,{116,47,111}},
/* 19208 */ {(16<<2)|3,{116,47,115}},
/* 19209 */ {(16<<2)|3,{116,47,116}},
/* 19210 */ {(11<<2)|2,{116,47,0}},
/* 19211 */ {(11<<2)|2,{116,47,0}},
/* 19212 */ {(11<<2)|2,{116,47,0}},
/* 19213 */ {(11<<2)|2,{116,47,0}},
/* 19214 */ {(11<<2)|2,{116,47,0}},
/* 19215 */ {(11<<2)|2,{116,47,0}},
/* 19216 */ {(11<<2)|2,{116,47,0}},
/* 19217 */ {(11<<2)|2,{116,47,0}},
/* 19218 */ {(11<<2)|2,{116,47,0}},
/* 19219 */ {(11<<2)|2,{116,47,0}},
/* 19220 */ {(11<<2)|2,{116,47,0}},
/* 19221 */ {(11<<2)|2,{116,47,0}},
/* 19222 */ {(11<<2)|2,{116,47,0}},
/* 19223 */ {(11<<2)|2,{116,47,0}},
/* 19224 */ {(11<<2)|2,{116,47,0}},
/* 19225 */ {(11<<2)|2,{116,47,0}},
/* 19226 */ {(11<<2)|2,{116,47,0}},
/* 19227 */ {(11<<2)|2,{116,47,0}},
/* 19228 */ {(11<<2)|2,{116,47,0}},
/* 19229 */ {(11<<2)|2,{116,47,0}},
/* 19230 */ {(11<<2)|2,{116,47,0}},
/* 19231 */ {(11<<2)|2,{116,47,0}},
/* 19232 */ {(16<<2)|3,{116,51,48}},
/* 19233 */ {(16<<2)|3,{116,51,49}},
/* 19234 */ {(16<<2)|3,{116,51,50}},
/* 19235 */ {(16<<2)|3,{116,51,97}},
/* 19236 */ {(16<<2)|3,{116,51,99}},
/* 19237 */ {(16<<2)|3,{116,51,101}},
/* 19238 */ {(16<<2)|3,{116,51,105}},
/* 19239 */ {(16<<2)|3,{116,51,111}},
/* 19240 */ {(16<<2)|3,{116,51,115}},
/* 19241 */ {(16<<2)|3,{116,51,116}},
/* 19242 */ {(11<<2)|2,{116,51,0}},
/* 19243 */ {(11<<2)|2,{116,51,0}},
/* 19244 */ {(11<<2)|2,{116,51,0}},
/* 19245 */ {(11<<2)|2,{116,51,0}},
/* 19246 */ {(11<<2)|2,{116,51,0}},
/* 19247 */ {(11<<2)|2,{116,51,0}},
/* 19248 */ {(11<<2)|2,{116,51,0}},
/* 19249 */ {(11<<2)|2,{116,51,0}},
/* 19250 */ {(11<<2)|2,{116,51,0}},
/* 19251 */ {(11<<2)|2,{116,51,0}},
/* 19252 */ {(11<<2)|2,{116,51,0}},
/* 19253 */ {(11<<2)|2,{116,51,0}},
/* 19254 */ {(11<<2)|2,{116,51,0}},
/* 19255 */ {(11<<2)|2,{116,51,0}},
/* 19256 */ {(11<<2)|2,{116,51,0}},
/* 19257 */ {(11<<2)|2,{116,51,0}},
/* 19258 */ {(11<<2)|2,{116,51,0}},
/* 19259 */ {(11<<2)|2,{116,51,0}},
/* 19260 */ {(11<<2)|2,{116,51,0}},
/* 19261 */ {(11<<2)|2,{116,51,0}},
/* 19262 */ {(11<<2)|2,{116,51,0}},
/* 19263 */ {(11<<2)|2,{116,51,0}},
/* 19264 */ {(16<<2)|3,{116,52,48}},
/* 19265 */ {(16<<2)|3,{116,52,49}},
/* 19266 */ {(16<<2)|3,{116,52,50}},
/* 19267 */ {(16<<2)|3,{116,52,97}},
/* 19268 */ {(16<<2)|3,{116,52,99}},
/* 19269 */ {(16<<2)|3,{116,52,101}},
/* 19270 */ {(16<<2)|3,{116,52,105}},
/* 19271 */ {(16<<2)|3,{116,52,111}},
/* 19272 */ {(16<<2)|3,{116,52,115}},
/* 19273 */ {(16<<2)|3,{116,52,116}},
/* 19274 */ {(11<<2)|2,{116,52,0}},
/* 19275 */ {(11<<2)|2,{116,52,0}},
/* 19276 */ {(11<<2)|2,{116,52,0}},
/* 19277 */ {(11<<2)|2,{116,52,0}},
/* 19278 */ {(11<<2)|2,{116,52,0}},
/* 19279 */ {(11<<2)|2,{116,52,0}},
/* 19280 */ {(11<<2)|2,{116,52,0}},
/* 19281 */ {(11<<2)|2,{116,52,0}},
/* 19282 */ {(11<<2)|2,{116,52,0}},
/* 19283 */ {(11<<2)|2,{116,52,0}},
/* 19284 */ {(11<<2)|2,{116,52,0}},
/* 19285 */ {(11<<2)|2,{116,52,0}},
/* 19286 */ {(11<<2)|2,{116,52,0}},
/* 19287 */ {(11<<2)|2,{116,52,0}},
/* 19288 */ {(11<<2)|2,{116,52,0}},
/* 19289 */ {(11<<2)|2,{116,52,0}},
/* 19290 */ {(11<<2)|2,{116,52,0}},
/* 19291 */ {(11<<2)|2,{116,52,0}},
/* 19292 */ {(11<<2)|2,{116,52,0}},
/* 19293 */ {(11<<2)|2,{116,52,0}},
/* 19294 */ {(11<<2)|2,{116,52,0}},
/* 19295 */ {(11<<2)|2,{116,52,0}},
/* 19296 */ {(16<<2)|3,{116,53,48}},
/* 19297 */ {(16<<2)|3,{116,53,49}},
/* 19298 */ {(16<<2)|3,{116,53,50}},
/* 19299 */ {(16<<2)|3,{116,53,97}},
/* 19300 */ {(16<<2)|3,{116,53,99}},
/* 19301 */ {(16<<2)|3,{116,53,101}},
/* 19302 */ {(16<<2)|3,{116,53,105}},
/* 19303 */ {(16<<2)|3,{116,53,111}},
/* 19304 */ {(16<<2)|3,{116,53,115}},
/* 19305 */ {(16<<2)|3,{116,53,116}},
/* 19306 */ {(11<<2)|2,{116,53,0}},
/* 19307 */ {(11<<2)|2,{116,53,0}},
/* 19308 */ {(11<<2)|2,{116,53,0}},
/* 19309 */ {(11<<2)|2,{116,53,0}},
/* 19310 */ {(11<<2)|2,{116,53,0}},
/* 19311 */ {(11<<2)|2,{116,53,0}},
/* 19312 */ {(11<<2)|2,{116,53,0}},
/* 19313 */ {(11<<2)|2,{116,53,0}},
/* 19314 */ {(11<<2)|2,{116,53,0}},
/* 19315 */ {(11<<2)|2,{116,53,0}},
/* 19316 */ {(11<<2)|2,{116,53,0}},
/* 19317 */ {(11<<2)|2,{116,53,0}},
/* 19318 */ {(11<<2)|2,{116,53,0}},
/* 19319 */ {(11<<2)|2,{116,53,0}},
/* 19320 */ {(11<<2)|2,{116,53,0}},
/* 19321 */ {(11<<2)|2,{116,53,0}},
/* 19322 */ {(11<<2)|2,{116,53,0}},
/* 19323 */ {(11<<2)|2,{116,53,0}},
/* 19324 */ {(11<<2)|2,{116,53,0}},
/* 19325 */ {(11<<2)|2,{116,53,0}},
/* 19326 */ {(11<<2)|2,{116,53,0}},
/* 19327 */ {(11<<2)|2,{116,53,0}},
/* 19328 */ {(16<<2)|3,{116,54,48}},
/* 19329 */ {(16<<2)|3,{116,54,49}},
/* 19330 */ {(16<<2)|3,{116,54,50}},
/* 19331 */ {(16<<2)|3,{116,54,97}},
/* 19332 */ {(16<<2)|3,{116,54,99}},
/* 19333 */ {(16<<2)|3,{116,54,101}},
/* 19334 */ {(16<<2)|3,{116,54,105}},
/* 19335 */ {(16<<2)|3,{116,54,111}},
/* 19336 */ {(16<<2)|3,{116,54,115}},
/* 19337 */ {(16<<2)|3,{116,54,116}},
/* 19338 */ {(11<<2)|2,{116,54,0}},
/* 19339 */ {(11<<2)|2,{116,54,0}},
/* 19340 */ {(11<<2)|2,{116,54,0}},
/* 19341 */ {(11<<2)|2,{116,54,0}},
/* 19342 */ {(11<<2)|2,{116,54,0}},
/* 19343 */ {(11<<2)|2,{116,54,0}},
/* 19344 */ {(11<<2)|2,{116,54,0}},
/* 19345 */ {(11<<2)|2,{116,54,0}},
/* 19346 */ {(11<<2)|2,{116,54,0}},
/* 19347 */ {(11<<2)|2,{116,54,0}},
/* 19348 */ {(11<<2)|2,{116,54,0}},
/* 19349 */ {(11<<2)|2,{116,54,0}},
/* 19350 */ {(11<<2)|2,{116,54,0}},
/* 19351 */ {(11<<2)|2,{116,54,0}},
/* 19352 */ {(11<<2)|2,{116,54,0}},
/* 19353 */ {(11<<2)|2,{116,54,0}},
/* 19354 */ {(11<<2)|2,{116,54,0}},
/* 19355 */ {(11<<2)|2,{116,54,0}},
/* 19356 */ {(11<<2)|2,{116,54,0}},
/* 19357 */ {(11<<2)|2,{116,54,0}},
/* 19358 */ {(11<<2)|2,{116,54,0}},
/* 19359 */ {(11<<2)|2,{116,54,0}},
/* 19360 */ {(16<<2)|3,{116,55,48}},
/* 19361 */ {(16<<2)|3,{116,55,49}},
/* 19362 */ {(16<<2)|3,{116,55,50}},
/* 19363 */ {(16<<2)|3,{116,55,97}},
/* 19364 */ {(16<<2)|3,{116,55,99}},
/* 19365 */ {(16<<2)|3,{116,55,101}},
/* 19366 */ {(16<<2)|3,{116,55,105}},
/* 19367 */ {(16<<2)|3,{116,55,111}},
/* 19368 */ {(16<<2)|3,{116,55,115}},
/* 19369 */ {(16<<2)|3,{116,55,116}},
/* 19370 */ {(11<<2)|2,{116,55,0}},
/* 19371 */ {(11<<2)|2,{116,55,0}},
/* 19372 */ {(11<<2)|2,{116,55,0}},
/* 19373 */ {(11<<2)|2,{116,55,0}},
/* 19374 */ {(11<<2)|2,{116,55,0}},
/* 19375 */ {(11<<2)|2,{116,55,0}},
/* 19376 */ {(11<<2)|2,{116,55,0}},
/* 19377 */ {(11<<2)|2,{116,55,0}},
/* 19378 */ {(11<<2)|2,{116,55,0}},
/* 19379 */ {(11<<2)|2,{116,55,0}},
/* 19380 */ {(11<<2)|2,{116,55,0}},
/* 19381 */ {(11<<2)|2,{116,55,0}},
/* 19382 */ {(11<<2)|2,{116,55,0}},
/* 19383 */ {(11<<2)|2,{116,55,0}},
/* 19384 */ {(11<<2)|2,{116,55,0}},
/* 19385 */ {(11<<2)|2,{116,55,0}},
/* 19386 */ {(11<<2)|2,{116,55,0}},
/* 19387 */ {(11<<2)|2,{116,55,0}},
/* 19388 */ {(11<<2)|2,{116,55,0}},
/* 19389 */ {(11<<2)|2,{116,55,0}},
/* 19390 */ {(11<<2)|2,{116,55,0}},
/* 19391 */ {(11<<2)|2,{116,55,0}},
/* 19392 */ {(16<<2)|3,{116,56,48}},
/* 19393 */ {(16<<2)|3,{116,56,49}},
/* 19394 */ {(16<<2)|3,{116,56,50}},
/* 19395 */ {(16<<2)|3,{116,56,97}},
/* 19396 */ {(16<<2)|3,{116,56,99}},
/* 19397 */ {(16<<2)|3,{116,56,101}},
/* 19398 */ {(16<<2)|3,{116,56,105}},
/* 19399 */ {(16<<2)|3,{116,56,111}},
/* 19400 */ {(16<<2)|3,{116,56,115}},
/* 19401 */ {(16<<2)|3,{116,56,116}},
/* 19402 */ {(11<<2)|2,{116,56,0}},
/* 19403 */ {(11<<2)|2,{116,56,0}},
/* 19404 */ {(11<<2)|2,{116,56,0}},
/* 19405 */ {(11<<2)|2,{116,56,0}},
/* 19406 */ {(11<<2)|2,{116,56,0}},
/* 19407 */ {(11<<2)|2,{116,56,0}},
/* 19408 */ {(11<<2)|2,{116,56,0}},
/* 19409 */ {(11<<2)|2,{116,56,0}},
/* 19410 */ {(11<<2)|2,{116,56,0}},
/* 19411 */ {(11<<2)|2,{116,56,0}},
/* 19412 */ {(11<<2)|2,{116,56,0}},
/* 19413 */ {(11<<2)|2,{116,56,0}},
/* 19414 */ {(11<<2)|2,{116,56,0}},
/* 19415 */ {(11<<2)|2,{116,56,0}},
/* 19416 */ {(11<<2)|2,{116,56,0}},
/* 19417 */ {(11<<2)|2,{116,56,0}},
/* 19418 */ {(11<<2)|2,{116,56,0}},
/* 19419 */ {(11<<2)|2,{116,56,0}},
/* 19420 */ {(11<<2)|2,{116,56,0}},
/* 19421 */ {(11<<2)|2,{116,56,0}},
/* 19422 */ {(11<<2)|2,{116,56,0}},
/* 19423 */ {(11<<2)|2,{116,56,0}},
/* 19424 */ {(16<<2)|3,{116,57,48}},
/* 19425 */ {(16<<2)|3,{116,57,49}},
/* 19426 */ {(16<<2)|3,{116,57,50}},
/* 19427 */ {(16<<2)|3,{116,57,97}},
/* 19428 */ {(16<<2)|3,{116,57,99}},
/* 19429 */ {(16<<2)|3,{116,57,101}},
/* 19430 */ {(16<<2)|3,{116,57,105}},
/* 19431 */ {(16<<2)|3,{116,57,111}},
/* 19432 */ {(16<<2)|3,{116,57,115}},
/* 19433 */ {(16<<2)|3,{116,57,116}},
/* 19434 */ {(11<<2)|2,{116,57,0}},
/* 19435 */ {(11<<2)|2,{116,57,0}},
/* 19436 */ {(11<<2)|2,{116,57,0}},
/* 19437 */ {(11<<2)|2,{116,57,0}},
/* 19438 */ {(11<<2)|2,{116,57,0}},
/* 19439 */ {(11<<2)|2,{116,57,0}},
/* 19440 */ {(11<<2)|2,{116,57,0}},
/* 19441 */ {(11<<2)|2,{116,57,0}},
/* 19442 */ {(11<<2)|2,{116,57,0}},
/* 19443 */ {(11<<2)|2,{116,57,0}},
/* 19444 */ {(11<<2)|2,{116,57,0}},
/* 19445 */ {(11<<2)|2,{116,57,0}},
/* 19446 */ {(11<<2)|2,{116,57,0}},
/* 19447 */ {(11<<2)|2,{116,57,0}},
/* 19448 */ {(11<<2)|2,{116,57,0}},
/* 19449 */ {(11<<2)|2,{116,57,0}},
/* 19450 */ {(11<<2)|2,{116,57,0}},
/* 19451 */ {(11<<2)|2,{116,57,0}},
/* 19452 */ {(11<<2)|2,{116,57,0}},
/* 19453 */ {(11<<2)|2,{116,57,0}},
/* 19454 */ {(11<<2)|2,{116,57,0}},
/* 19455 */ {(11<<2)|2,{116,57,0}},
/* 19456 */ {(16<<2)|3,{116,61,48}},
/* 19457 */ {(16<<2)|3,{116,61,49}},
/* 19458 */ {(16<<2)|3,{116,61,50}},
/* 19459 */ {(16<<2)|3,{116,61,97}},
/* 19460 */ {(16<<2)|3,{116,61,99}},
/* 19461 */ {(16<<2)|3,{116,61,101}},
/* 19462 */ {(16<<2)|3,{116,61,105}},
/* 19463 */ {(16<<2)|3,{116,61,111}},
/* 19464 */ {(16<<2)|3,{116,61,115}},
/* 19465 */ {(16<<2)|3,{116,61,116}},
/* 19466 */ {(11<<2)|2,{116,61,0}},
/* 19467 */ {(11<<2)|2,{116,61,0}},
/* 19468 */ {(11<<2)|2,{116,61,0}},
/* 19469 */ {(11<<2)|2,{116,61,0}},
/* 19470 */ {(11<<2)|2,{116,61,0}},
/* 19471 */ {(11<<2)|2,{116,61,0}},
/* 19472 */ {(11<<2)|2,{116,61,0}},
/* 19473 */ {(11<<2)|2,{116,61,0}},
/* 19474 */ {(11<<2)|2,{116,61,0}},
/* 19475 */ {(11<<2)|2,{116,61,0}},
/* 19476 */ {(11<<2)|2,{116,61,0}},
/* 19477 */ {(11<<2)|2,{116,61,0}},
/* 19478 */ {(11<<2)|2,{116,61,0}},
/* 19479 */ {(11<<2)|2,{116,61,0}},
/* 19480 */ {(11<<2)|2,{116,61,0}},
/* 19481 */ {(11<<2)|2,{116,61,0}},
/* 19482 */ {(11<<2)|2,{116,61,0}},
/* 19483 */ {(11<<2)|2,{116,61,0}},
/* 19484 */ {(11<<2)|2,{116,61,0}},
/* 19485 */ {(11<<2)|2,{116,61,0}},
/* 19486 */ {(11<<2)|2,{116,61,0}},
/* 19487 */ {(11<<2)|2,{116,61,0}},
/* 19488 */ {(16<<2)|3,{116,65,48}},
/* 19489 */ {(16<<2)|3,{116,65,49}},
/* 19490 */ {(16<<2)|3,{116,65,50}},
/* 19491 */ {(16<<2)|3,{116,65,97}},
/* 19492 */ {(16<<2)|3,{116,65,99}},
/* 19493 */ {(16<<2)|3,{116,65,101}},
/* 19494 */ {(16<<2)|3,{116,65,105}},
/* 19495 */ {(16<<2)|3,{116,65,111}},
/* 19496 */ {(16<<2)|3,{116,65,115}},
/* 19497 */ {(16<<2)|3,{116,65,116}},
/* 19498 */ {(11<<2)|2,{116,65,0}},
/* 19499 */ {(11<<2)|2,{116,65,0}},
/* 19500 */ {(11<<2)|2,{116,65,0}},
/* 19501 */ {(11<<2)|2,{116,65,0}},
/* 19502 */ {(11<<2)|2,{116,65,0}},
/* 19503 */ {(11<<2)|2,{116,65,0}},
/* 19504 */ {(11<<2)|2,{116,65,0}},
/* 19505 */ {(11<<2)|2,{116,65,0}},
/* 19506 */ {(11<<2)|2,{116,65,0}},
/* 19507 */ {(11<<2)|2,{116,65,0}},
/* 19508 */ {(11<<2)|2,{116,65,0}},
/* 19509 */ {(11<<2)|2,{116,65,0}},
/* 19510 */ {(11<<2)|2,{116,65,0}},
/* 19511 */ {(11<<2)|2,{116,65,0}},
/* 19512 */ {(11<<2)|2,{116,65,0}},
/* 19513 */ {(11<<2)|2,{116,65,0}},
/* 19514 */ {(11<<2)|2,{116,65,0}},
/* 19515 */ {(11<<2)|2,{116,65,0}},
/* 19516 */ {(11<<2)|2,{116,65,0}},
/* 19517 */ {(11<<2)|2,{116,65,0}},
/* 19518 */ {(11<<2)|2,{116,65,0}},
/* 19519 */ {(11<<2)|2,{116,65,0}},
/* 19520 */ {(16<<2)|3,{116,95,48}},
/* 19521 */ {(16<<2)|3,{116,95,49}},
/* 19522 */ {(16<<2)|3,{116,95,50}},
/* 19523 */ {(16<<2)|3,{116,95,97}},
/* 19524 */ {(16<<2)|3,{116,95,99}},
/* 19525 */ {(16<<2)|3,{116,95,101}},
/* 19526 */ {(16<<2)|3,{116,95,105}},
/* 19527 */ {(16<<2)|3,{116,95,111}},
/* 19528 */ {(16<<2)|3,{116,95,115}},
/* 19529 */ {(16<<2)|3,{116,95,116}},
/* 19530 */ {(11<<2)|2,{116,95,0}},
/* 19531 */ {(11<<2)|2,{116,95,0}},
/* 19532 */ {(11<<2)|2,{116,95,0}},
/* 19533 */ {(11<<2)|2,{116,95,0}},
/* 19534 */ {(11<<2)|2,{116,95,0}},
/* 19535 */ {(11<<2)|2,{116,95,0}},
/* 19536 */ {(11<<2)|2,{116,95,0}},
/* 19537 */ {(11<<2)|2,{116,95,0}},
/* 19538 */ {(11<<2)|2,{116,95,0}},
/* 19539 */ {(11<<2)|2,{116,95,0}},
/* 19540 */ {(11<<2)|2,{116,95,0}},
/* 19541 */ {(11<<2)|2,{116,95,0}},
/* 19542 */ {(11<<2)|2,{116,95,0}},
/* 19543 */ {(11<<2)|2,{116,95,0}},
/* 19544 */ {(11<<2)|2,{116,95,0}},
/* 19545 */ {(11<<2)|2,{116,95,0}},
/* 19546 */ {(11<<2)|2,{116,95,0}},
/* 19547 */ {(11<<2)|2,{116,95,0}},
/* 19548 */ {(11<<2)|2,{116,95,0}},
/* 19549 */ {(11<<2)|2,{116,95,0}},
/* 19550 */ {(11<<2)|2,{116,95,0}},
/* 19551 */ {(11<<2)|2,{116,95,0}},
/* 19552 */ {(16<<2)|3,{116,98,48}},
/* 19553 */ {(16<<2)|3,{116,98,49}},
/* 19554 */ {(16<<2)|3,{116,98,50}},
/* 19555 */ {(16<<2)|3,{116,98,97}},
/* 19556 */ {(16<<2)|3,{116,98,99}},
/* 19557 */ {(16<<2)|3,{116,98,101}},
/* 19558 */ {(16<<2)|3,{116,98,105}},
/* 19559 */ {(16<<2)|3,{116,98,111}},
/* 19560 */ {(16<<2)|3,{116,98,115}},
/* 19561 */ {(16<<2)|3,{116,98,116}},
/* 19562 */ {(11<<2)|2,{116,98,0}},
/* 19563 */ {(11<<2)|2,{116,98,0}},
/* 19564 */ {(11<<2)|2,{116,98,0}},
/* 19565 */ {(11<<2)|2,{116,98,0}},
/* 19566 */ {(11<<2)|2,{116,98,0}},
/* 19567 */ {(11<<2)|2,{116,98,0}},
/* 19568 */ {(11<<2)|2,{116,98,0}},
/* 19569 */ {(11<<2)|2,{116,98,0}},
/* 19570 */ {(11<<2)|2,{116,98,0}},
/* 19571 */ {(11<<2)|2,{116,98,0}},
/* 19572 */ {(11<<2)|2,{116,98,0}},
/* 19573 */ {(11<<2)|2,{116,98,0}},
/* 19574 */ {(11<<2)|2,{116,98,0}},
/* 19575 */ {(11<<2)|2,{116,98,0}},
/* 19576 */ {(11<<2)|2,{116,98,0}},
/* 19577 */ {(11<<2)|2,{116,98,0}},
/* 19578 */ {(11<<2)|2,{116,98,0}},
/* 19579 */ {(11<<2)|2,{116,98,0}},
/* 19580 */ {(11<<2)|2,{116,98,0}},
/* 19581 */ {(11<<2)|2,{116,98,0}},
/* 19582 */ {(11<<2)|2,{116,98,0}},
/* 19583 */ {(11<<2)|2,{116,98,0}},
/* 19584 */ {(16<<2)|3,{116,100,48}},
/* 19585 */ {(16<<2)|3,{116,100,49}},
/* 19586 */ {(16<<2)|3,{116,100,50}},
/* 19587 */ {(16<<2)|3,{116,100,97}},
/* 19588 */ {(16<<2)|3,{116,100,99}},
/* 19589 */ {(16<<2)|3,{116,100,101}},
/* 19590 */ {(16<<2)|3,{116,100,105}},
/* 19591 */ {(16<<2)|3,{116,100,111}},
/* 19592 */ {(16<<2)|3,{116,100,115}},
/* 19593 */ {(16<<2)|3,{116,100,116}},
/* 19594 */ {(11<<2)|2,{116,100,0}},
/* 19595 */ {(11<<2)|2,{116,100,0}},
/* 19596 */ {(11<<2)|2,{116,100,0}},
/* 19597 */ {(11<<2)|2,{116,100,0}},
/* 19598 */ {(11<<2)|2,{116,100,0}},
/* 19599 */ {(11<<2)|2,{116,100,0}},
/* 19600 */ {(11<<2)|2,{116,100,0}},
/* 19601 */ {(11<<2)|2,{116,100,0}},
/* 19602 */ {(11<<2)|2,{116,100,0}},
/* 19603 */ {(11<<2)|2,{116,100,0}},
/* 19604 */ {(11<<2)|2,{116,100,0}},
/* 19605 */ {(11<<2)|2,{116,100,0}},
/* 19606 */ {(11<<2)|2,{116,100,0}},
/* 19607 */ {(11<<2)|2,{116,100,0}},
/* 19608 */ {(11<<2)|2,{116,100,0}},
/* 19609 */ {(11<<2)|2,{116,100,0}},
/* 19610 */ {(11<<2)|2,{116,100,0}},
/* 19611 */ {(11<<2)|2,{116,100,0}},
/* 19612 */ {(11<<2)|2,{116,100,0}},
/* 19613 */ {(11<<2)|2,{116,100,0}},
/* 19614 */ {(11<<2)|2,{116,100,0}},
/* 19615 */ {(11<<2)|2,{116,100,0}},
/* 19616 */ {(16<<2)|3,{116,102,48}},
/* 19617 */ {(16<<2)|3,{116,102,49}},
/* 19618 */ {(16<<2)|3,{116,102,50}},
/* 19619 */ {(16<<2)|3,{116,102,97}},
/* 19620 */ {(16<<2)|3,{116,102,99}},
/* 19621 */ {(16<<2)|3,{116,102,101}},
/* 19622 */ {(16<<2)|3,{116,102,105}},
/* 19623 */ {(16<<2)|3,{116,102,111}},
/* 19624 */ {(16<<2)|3,{116,102,115}},
/* 19625 */ {(16<<2)|3,{116,102,116}},
/* 19626 */ {(11<<2)|2,{116,102,0}},
/* 19627 */ {(11<<2)|2,{116,102,0}},
/* 19628 */ {(11<<2)|2,{116,102,0}},
/* 19629 */ {(11<<2)|2,{116,102,0}},
/* 19630 */ {(11<<2)|2,{116,102,0}},
/* 19631 */ {(11<<2)|2,{116,102,0}},
/* 19632 */ {(11<<2)|2,{116,102,0}},
/* 19633 */ {(11<<2)|2,{116,102,0}},
/* 19634 */ {(11<<2)|2,{116,102,0}},
/* 19635 */ {(11<<2)|2,{116,102,0}},
/* 19636 */ {(11<<2)|2,{116,102,0}},
/* 19637 */ {(11<<2)|2,{116,102,0}},
/* 19638 */ {(11<<2)|2,{116,102,0}},
/* 19639 */ {(11<<2)|2,{116,102,0}},
/* 19640 */ {(11<<2)|2,{116,102,0}},
/* 19641 */ {(11<<2)|2,{116,102,0}},
/* 19642 */ {(11<<2)|2,{116,102,0}},
/* 19643 */ {(11<<2)|2,{116,102,0}},
/* 19644 */ {(11<<2)|2,{116,102,0}},
/* 19645 */ {(11<<2)|2,{116,102,0}},
/* 19646 */ {(11<<2)|2,{116,102,0}},
/* 19647 */ {(11<<2)|2,{116,102,0}},
/* 19648 */ {(16<<2)|3,{116,103,48}},
/* 19649 */ {(16<<2)|3,{116,103,49}},
/* 19650 */ {(16<<2)|3,{116,103,50}},
/* 19651 */ {(16<<2)|3,{116,103,97}},
/* 19652 */ {(16<<2)|3,{116,103,99}},
/* 19653 */ {(16<<2)|3,{116,103,101}},
/* 19654 */ {(16<<2)|3,{116,103,105}},
/* 19655 */ {(16<<2)|3,{116,103,111}},
/* 19656 */ {(16<<2)|3,{116,103,115}},
/* 19657 */ {(16<<2)|3,{116,103,116}},
/* 19658 */ {(11<<2)|2,{116,103,0}},
/* 19659 */ {(11<<2)|2,{116,103,0}},
/* 19660 */ {(11<<2)|2,{116,103,0}},
/* 19661 */ {(11<<2)|2,{116,103,0}},
/* 19662 */ {(11<<2)|2,{116,103,0}},
/* 19663 */ {(11<<2)|2,{116,103,0}},
/* 19664 */ {(11<<2)|2,{116,103,0}},
/* 19665 */ {(11<<2)|2,{116,103,0}},
/* 19666 */ {(11<<2)|2,{116,103,0}},
/* 19667 */ {(11<<2)|2,{116,103,0}},
/* 19668 */ {(11<<2)|2,{116,103,0}},
/* 19669 */ {(11<<2)|2,{116,103,0}},
/* 19670 */ {(11<<2)|2,{116,103,0}},
/* 19671 */ {(11<<2)|2,{116,103,0}},
/* 19672 */ {(11<<2)|2,{116,103,0}},
/* 19673 */ {(11<<2)|2,{116,103,0}},
/* 19674 */ {(11<<2)|2,{116,103,0}},
/* 19675 */ {(11<<2)|2,{116,103,0}},
/* 19676 */ {(11<<2)|2,{116,103,0}},
/* 19677 */ {(11<<2)|2,{116,103,0}},
/* 19678 */ {(11<<2)|2,{116,103,0}},
/* 19679 */ {(11<<2)|2,{116,103,0}},
/* 19680 */ {(16<<2)|3,{116,104,48}},
/* 19681 */ {(16<<2)|3,{116,104,49}},
/* 19682 */ {(16<<2)|3,{116,104,50}},
/* 19683 */ {(16<<2)|3,{116,104,97}},
/* 19684 */ {(16<<2)|3,{116,104,99}},
/* 19685 */ {(16<<2)|3,{116,104,101}},
/* 19686 */ {(16<<2)|3,{116,104,105}},
/* 19687 */ {(16<<2)|3,{116,104,111}},
/* 19688 */ {(16<<2)|3,{116,104,115}},
/* 19689 */ {(16<<2)|3,{116,104,116}},
/* 19690 */ {(11<<2)|2,{116,104,0}},
/* 19691 */ {(11<<2)|2,{116,104,0}},
/* 19692 */ {(11<<2)|2,{116,104,0}},
/* 19693 */ {(11<<2)|2,{116,104,0}},
/* 19694 */ {(11<<2)|2,{116,104,0}},
/* 19695 */ {(11<<2)|2,{116,104,0}},
/* 19696 */ {(11<<2)|2,{116,104,0}},
/* 19697 */ {(11<<2)|2,{116,104,0}},
/* 19698 */ {(11<<2)|2,{116,104,0}},
/* 19699 */ {(11<<2)|2,{116,104,0}},
/* 19700 */ {(11<<2)|2,{116,104,0}},
/* 19701 */ {(11<<2)|2,{116,104,0}},
/* 19702 */ {(11<<2)|2,{116,104,0}},
/* 19703 */ {(11<<2)|2,{116,104,0}},
/* 19704 */ {(11<<2)|2,{116,104,0}},
/* 19705 */ {(11<<2)|2,{116,104,0}},
/* 19706 */ {(11<<2)|2,{116,104,0}},
/* 19707 */ {(11<<2)|2,{116,104,0}},
/* 19708 */ {(11<<2)|2,{116,104,0}},
/* 19709 */ {(11<<2)|2,{116,104,0}},
/* 19710 */ {(11<<2)|2,{116,104,0}},
/* 19711 */ {(11<<2)|2,{116,104,0}},
/* 19712 */ {(16<<2)|3,{116,108,48}},
/* 19713 */ {(16<<2)|3,{116,108,49}},
/* 19714 */ {(16<<2)|3,{116,108,50}},
/* 19715 */ {(16<<2)|3,{116,108,97}},
/* 19716 */ {(16<<2)|3,{116,108,99}},
/* 19717 */ {(16<<2)|3,{116,108,101}},
/* 19718 */ {(16<<2)|3,{116,108,105}},
/* 19719 */ {(16<<2)|3,{116,108,111}},
/* 19720 */ {(16<<2)|3,{116,108,115}},
/* 19721 */ {(16<<2)|3,{116,108,116}},
/* 19722 */ {(11<<2)|2,{116,108,0}},
/* 19723 */ {(11<<2)|2,{116,108,0}},
/* 19724 */ {(11<<2)|2,{116,108,0}},
/* 19725 */ {(11<<2)|2,{116,108,0}},
/* 19726 */ {(11<<2)|2,{116,108,0}},
/* 19727 */ {(11<<2)|2,{116,108,0}},
/* 19728 */ {(11<<2)|2,{116,108,0}},
/* 19729 */ {(11<<2)|2,{116,108,0}},
/* 19730 */ {(11<<2)|2,{116,108,0}},
/* 19731 */ {(11<<2)|2,{116,108,0}},
/* 19732 */ {(11<<2)|2,{116,108,0}},
/* 19733 */ {(11<<2)|2,{116,108,0}},
/* 19734 */ {(11<<2)|2,{116,108,0}},
/* 19735 */ {(11<<2)|2,{116,108,0}},
/* 19736 */ {(11<<2)|2,{116,108,0}},
/* 19737 */ {(11<<2)|2,{116,108,0}},
/* 19738 */ {(11<<2)|2,{116,108,0}},
/* 19739 */ {(11<<2)|2,{116,108,0}},
/* 19740 */ {(11<<2)|2,{116,108,0}},
/* 19741 */ {(11<<2)|2,{116,108,0}},
/* 19742 */ {(11<<2)|2,{116,108,0}},
/* 19743 */ {(11<<2)|2,{116,108,0}},
/* 19744 */ {(16<<2)|3,{116,109,48}},
/* 19745 */ {(16<<2)|3,{116,109,49}},
/* 19746 */ {(16<<2)|3,{116,109,50}},
/* 19747 */ {(16<<2)|3,{116,109,97}},
/* 19748 */ {(16<<2)|3,{116,109,99}},
/* 19749 */ {(16<<2)|3,{116,109,101}},
/* 19750 */ {(16<<2)|3,{116,109,105}},
/* 19751 */ {(16<<2)|3,{116,109,111}},
/* 19752 */ {(16<<2)|3,{116,109,115}},
/* 19753 */ {(16<<2)|3,{116,109,116}},
/* 19754 */ {(11<<2)|2,{116,109,0}},
/* 19755 */ {(11<<2)|2,{116,109,0}},
/* 19756 */ {(11<<2)|2,{116,109,0}},
/* 19757 */ {(11<<2)|2,{116,109,0}},
/* 19758 */ {(11<<2)|2,{116,109,0}},
/* 19759 */ {(11<<2)|2,{116,109,0}},
/* 19760 */ {(11<<2)|2,{116,109,0}},
/* 19761 */ {(11<<2)|2,{116,109,0}},
/* 19762 */ {(11<<2)|2,{116,109,0}},
/* 19763 */ {(11<<2)|2,{116,109,0}},
/* 19764 */ {(11<<2)|2,{116,109,0}},
/* 19765 */ {(11<<2)|2,{116,109,0}},
/* 19766 */ {(11<<2)|2,{116,109,0}},
/* 19767 */ {(11<<2)|2,{116,109,0}},
/* 19768 */ {(11<<2)|2,{116,109,0}},
/* 19769 */ {(11<<2)|2,{116,109,0}},
/* 19770 */ {(11<<2)|2,{116,109,0}},
/* 19771 */ {(11<<2)|2,{116,109,0}},
/* 19772 */ {(11<<2)|2,{116,109,0}},
/* 19773 */ {(11<<2)|2,{116,109,0}},
/* 19774 */ {(11<<2)|2,{116,109,0}},
/* 19775 */ {(11<<2)|2,{116,109,0}},
/* 19776 */ {(16<<2)|3,{116,110,48}},
/* 19777 */ {(16<<2)|3,{116,110,49}},
/* 19778 */ {(16<<2)|3,{116,110,50}},
/* 19779 */ {(16<<2)|3,{116,110,97}},
/* 19780 */ {(16<<2)|3,{116,110,99}},
/* 19781 */ {(16<<2)|3,{116,110,101}},
/* 19782 */ {(16<<2)|3,{116,110,105}},
/* 19783 */ {(16<<2)|3,{116,110,111}},
/* 19784 */ {(16<<2)|3,{116,110,115}},
/* 19785 */ {(16<<2)|3,{116,110,116}},
/* 19786 */ {(11<<2)|2,{116,110,0}},
/* 19787 */ {(11<<2)|2,{116,110,0}},
/* 19788 */ {(11<<2)|2,{116,110,0}},
/* 19789 */ {(11<<2)|2,{116,110,0}},
/* 19790 */ {(11<<2)|2,{116,110,0}},
/* 19791 */ {(11<<2)|2,{116,110,0}},
/* 19792 */ {(11<<2)|2,{116,110,0}},
/* 19793 */ {(11<<2)|2,{116,110,0}},
/* 19794 */ {(11<<2)|2,{116,110,0}},
/* 19795 */ {(11<<2)|2,{116,110,0}},
/* 19796 */ {(11<<2)|2,{116,110,0}},
/* 19797 */ {(11<<2)|2,{116,110,0}},
/* 19798 */ {(11<<2)|2,{116,110,0}},
/* 19799 */ {(11<<2)|2,{116,110,0}},
/* 19800 */ {(11<<2)|2,{116,110,0}},
/* 19801 */ {(11<<2)|2,{116,110,0}},
/* 19802 */ {(11<<2)|2,{116,110,0}},
/* 19803 */ {(11<<2)|2,{116,110,0}},
/* 19804 */ {(11<<2)|2,{116,110,0}},
/* 19805 */ {(11<<2)|2,{116,110,0}},
/* 19806 */ {(11<<2)|2,{116,110,0}},
/* 19807 */ {(11<<2)|2,{116,110,0}},
/* 19808 */ {(16<<2)|3,{116,112,48}},
/* 19809 */ {(16<<2)|3,{116,112,49}},
/* 19810 */ {(16<<2)|3,{116,112,50}},
/* 19811 */ {(16<<2)|3,{116,112,97}},
/* 19812 */ {(16<<2)|3,{116,112,99}},
/* 19813 */ {(16<<2)|3,{116,112,101}},
/* 19814 */ {(16<<2)|3,{116,112,105}},
/* 19815 */ {(16<<2)|3,{116,112,111}},
/* 19816 */ {(16<<2)|3,{116,112,115}},
/* 19817 */ {(16<<2)|3,{116,112,116}},
/* 19818 */ {(11<<2)|2,{116,112,0}},
/* 19819 */ {(11<<2)|2,{116,112,0}},
/* 19820 */ {(11<<2)|2,{116,112,0}},
/* 19821 */ {(11<<2)|2,{116,112,0}},
/* 19822 */ {(11<<2)|2,{116,112,0}},
/* 19823 */ {(11<<2)|2,{116,112,0}},
/* 19824 */ {(11<<2)|2,{116,112,0}},
/* 19825 */ {(11<<2)|2,{116,112,0}},
/* 19826 */ {(11<<2)|2,{116,112,0}},
/* 19827 */ {(11<<2)|2,{116,112,0}},
/* 19828 */ {(11<<2)|2,{116,112,0}},
/* 19829 */ {(11<<2)|2,{116,112,0}},
/* 19830 */ {(11<<2)|2,{116,112,0}},
/* 19831 */ {(11<<2)|2,{116,112,0}},
/* 19832 */ {(11<<2)|2,{116,112,0}},
/* 19833 */ {(11<<2)|2,{116,112,0}},
/* 19834 */ {(11<<2)|2,{116,112,0}},
/* 19835 */ {(11<<2)|2,{116,112,0}},
/* 19836 */ {(11<<2)|2,{116,112,0}},
/* 19837 */ {(11<<2)|2,{116,112,0}},
/* 19838 */ {(11<<2)|2,{116,112,0}},
/* 19839 */ {(11<<2)|2,{116,112,0}},
/* 19840 */ {(16<<2)|3,{116,114,48}},
/* 19841 */ {(16<<2)|3,{116,114,49}},
/* 19842 */ {(16<<2)|3,{116,114,50}},
/* 19843 */ {(16<<2)|3,{116,114,97}},
/* 19844 */ {(16<<2)|3,{116,114,99}},
/* 19845 */ {(16<<2)|3,{116,114,101}},
/* 19846 */ {(16<<2)|3,{116,114,105}},
/* 19847 */ {(16<<2)|3,{116,114,111}},
/* 19848 */ {(16<<2)|3,{116,114,115}},
/* 19849 */ {(16<<2)|3,{116,114,116}},
/* 19850 */ {(11<<2)|2,{116,114,0}},
/* 19851 */ {(11<<2)|2,{116,114,0}},
/* 19852 */ {(11<<2)|2,{116,114,0}},
/* 19853 */ {(11<<2)|2,{116,114,0}},
/* 19854 */ {(11<<2)|2,{116,114,0}},
/* 19855 */ {(11<<2)|2,{116,114,0}},
/* 19856 */ {(11<<2)|2,{116,114,0}},
/* 19857 */ {(11<<2)|2,{116,114,0}},
/* 19858 */ {(11<<2)|2,{116,114,0}},
/* 19859 */ {(11<<2)|2,{116,114,0}},
/* 19860 */ {(11<<2)|2,{116,114,0}},
/* 19861 */ {(11<<2)|2,{116,114,0}},
/* 19862 */ {(11<<2)|2,{116,114,0}},
/* 19863 */ {(11<<2)|2,{116,114,0}},
/* 19864 */ {(11<<2)|2,{116,114,0}},
/* 19865 */ {(11<<2)|2,{116,114,0}},
/* 19866 */ {(11<<2)|2,{116,114,0}},
/* 19867 */ {(11<<2)|2,{116,114,0}},
/* 19868 */ {(11<<2)|2,{116,114,0}},
/* 19869 */ {(11<<2)|2,{116,114,0}},
/* 19870 */ {(11<<2)|2,{116,114,0}},
/* 19871 */ {(11<<2)|2,{116,114,0}},
/* 19872 */ {(16<<2)|3,{116,117,48}},
/* 19873 */ {(16<<2)|3,{116,117,49}},
/* 19874 */ {(16<<2)|3,{116,117,50}},
/* 19875 */ {(16<<2)|3,{116,117,97}},
/* 19876 */ {(16<<2)|3,{116,117,99}},
/* 19877 */ {(16<<2)|3,{116,117,101}},
/* 19878 */ {(16<<2)|3,{116,117,105}},
/* 19879 */ {(16<<2)|3,{116,117,111}},
/* 19880 */ {(16<<2)|3,{116,117,115}},
/* 19881 */ {(16<<2)|3,{116,117,116}},
/* 19882 */ {(11<<2)|2,{116,117,0}},
/* 19883 */ {(11<<2)|2,{116,117,0}},
/* 19884 */ {(11<<2)|2,{116,117,0}},
/* 19885 */ {(11<<2)|2,{116,117,0}},
/* 19886 */ {(11<<2)|2,{116,117,0}},
/* 19887 */ {(11<<2)|2,{116,117,0}},
/* 19888 */ {(11<<2)|2,{116,117,0}},
/* 19889 */ {(11<<2)|2,{116,117,0}},
/* 19890 */ {(11<<2)|2,{116,117,0}},
/* 19891 */ {(11<<2)|2,{116,117,0}},
/* 19892 */ {(11<<2)|2,{116,117,0}},
/* 19893 */ {(11<<2)|2,{116,117,0}},
/* 19894 */ {(11<<2)|2,{116,117,0}},
/* 19895 */ {(11<<2)|2,{116,117,0}},
/* 19896 */ {(11<<2)|2,{116,117,0}},
/* 19897 */ {(11<<2)|2,{116,117,0}},
/* 19898 */ {(11<<2)|2,{116,117,0}},
/* 19899 */ {(11<<2)|2,{116,117,0}},
/* 19900 */ {(11<<2)|2,{116,117,0}},
/* 19901 */ {(11<<2)|2,{116,117,0}},
/* 19902 */ {(11<<2)|2,{116,117,0}},
/* 19903 */ {(11<<2)|2,{116,117,0}},
/* 19904 */ {(12<<2)|2,{116,58,0}},
/* 19905 */ {(12<<2)|2,{116,58,0}},
/* 19906 */ {(12<<2)|2,{116,58,0}},
/* 19907 */ {(12<<2)|2,{116,58,0}},
/* 19908 */ {(12<<2)|2,{116,58,0}},
/* 19909 */ {(12<<2)|2,{116,58,0}},
/* 19910 */ {(12<<2)|2,{116,58,0}},
/* 19911 */ {(12<<2)|2,{116,58,0}},
/* 19912 */ {(12<<2)|2,{116,58,0}},
/* 19913 */ {(12<<2)|2,{116,58,0}},
/* 19914 */ {(12<<2)|2,{116,58,0}},
/* 19915 */ {(12<<2)|2,{116,58,0}},
/* 19916 */ {(12<<2)|2,{116,58,0}},
/* 19917 */ {(12<<2)|2,{116,58,0}},
/* 19918 */ {(12<<2)|2,{116,58,0}},
/* 19919 */ {(12<<2)|2,{116,58,0}},
/* 19920 */ {(12<<2)|2,{116,66,0}},
/* 19921 */ {(12<<2)|2,{116,66,0}},
/* 19922 */ {(12<<2)|2,{116,66,0}},
/* 19923 */ {(12<<2)|2,{116,66,0}},
/* 19924 */ {(12<<2)|2,{116,66,0}},
/* 19925 */ {(12<<2)|2,{116,66,0}},
/* 19926 */ {(12<<2)|2,{116,66,0}},
/* 19927 */ {(12<<2)|2,{116,66,0}},
/* 19928 */ {(12<<2)|2,{116,66,0}},
/* 19929 */ {(12<<2)|2,{116,66,0}},
/* 19930 */ {(12<<2)|2,{116,66,0}},
/* 19931 */ {(12<<2)|2,{116,66,0}},
/* 19932 */ {(12<<2)|2,{116,66,0}},
/* 19933 */ {(12<<2)|2,{116,66,0}},
/* 19934 */ {(12<<2)|2,{116,66,0}},
/* 19935 */ {(12<<2)|2,{116,66,0}},
/* 19936 */ {(12<<2)|2,{116,67,0}},
/* 19937 */ {(12<<2)|2,{116,67,0}},
/* 19938 */ {(12<<2)|2,{116,67,0}},
/* 19939 */ {(12<<2)|2,{116,67,0}},
/* 19940 */ {(12<<2)|2,{116,67,0}},
/* 19941 */ {(12<<2)|2,{116,67,0}},
/* 19942 */ {(12<<2)|2,{116,67,0}},
/* 19943 */ {(12<<2)|2,{116,67,0}},
/* 19944 */ {(12<<2)|2,{116,67,0}},
/* 19945 */ {(12<<2)|2,{116,67,0}},
/* 19946 */ {(12<<2)|2,{116,67,0}},
/* 19947 */ {(12<<2)|2,{116,67,0}},
/* 19948 */ {(12<<2)|2,{116,67,0}},
/* 19949 */ {(12<<2)|2,{116,67,0}},
/* 19950 */ {(12<<2)|2,{116,67,0}},
/* 19951 */ {(12<<2)|2,{116,67,0}},
/* 19952 */ {(12<<2)|2,{116,68,0}},
/* 19953 */ {(12<<2)|2,{116,68,0}},
/* 19954 */ {(12<<2)|2,{116,68,0}},
/* 19955 */ {(12<<2)|2,{116,68,0}},
/* 19956 */ {(12<<2)|2,{116,68,0}},
/* 19957 */ {(12<<2)|2,{116,68,0}},
/* 19958 */ {(12<<2)|2,{116,68,0}},
/* 19959 */ {(12<<2)|2,{116,68,0}},
/* 19960 */ {(12<<2)|2,{116,68,0}},
/* 19961 */ {(12<<2)|2,{116,68,0}},
/* 19962 */ {(12<<2)|2,{116,68,0}},
/* 19963 */ {(12<<2)|2,{116,68,0}},
/* 19964 */ {(12<<2)|2,{116,68,0}},
/* 19965 */ {(12<<2)|2,{116,68,0}},
/* 19966 */ {(12<<2)|2,{116,68,0}},
/* 19967 */ {(12<<2)|2,{116,68,0}},
/* 19968 */ {(12<<2)|2,{116,69,0}},
/* 19969 */ {(12<<2)|2,{116,69,0}},
/* 19970 */ {(12<<2)|2,{116,69,0}},
/* 19971 */ {(12<<2)|2,{116,69,0}},
/* 19972 */ {(12<<2)|2,{116,69,0}},
/* 19973 */ {(12<<2)|2,{116,69,0}},
/* 19974 */ {(12<<2)|2,{116,69,0}},
/* 19975 */ {(12<<2)|2,{116,69,0}},
/* 19976 */ {(12<<2)|2,{116,69,0}},
/* 19977 */ {(12<<2)|2,{116,69,0}},
/* 19978 */ {(12<<2)|2,{116,69,0}},
/* 19979 */ {(12<<2)|2,{116,69,0}},
/* 19980 */ {(12<<2)|2,{116,69,0}},
/* 19981 */ {(12<<2)|2,{116,69,0}},
/* 19982 */ {(12<<2)|2,{116,69,0}},
/* 19983 */ {(12<<2)|2,{116,69,0}},
/* 19984 */ {(12<<2)|2,{116,70,0}},
/* 19985 */ {(12<<2)|2,{116,70,0}},
/* 19986 */ {(12<<2)|2,{116,70,0}},
/* 19987 */ {(12<<2)|2,{116,70,0}},
/* 19988 */ {(12<<2)|2,{116,70,0}},
/* 19989 */ {(12<<2)|2,{116,70,0}},
/* 19990 */ {(12<<2)|2,{116,70,0}},
/* 19991 */ {(12<<2)|2,{116,70,0}},
/* 19992 */ {(12<<2)|2,{116,70,0}},
/* 19993 */ {(12<<2)|2,{116,70,0}},
/* 19994 */ {(12<<2)|2,{116,70,0}},
/* 19995 */ {(12<<2)|2,{116,70,0}},
/* 19996 */ {(12<<2)|2,{116,70,0}},
/* 19997 */ {(12<<2)|2,{116,70,0}},
/* 19998 */ {(12<<2)|2,{116,70,0}},
/* 19999 */ {(12<<2)|2,{116,70,0}},
/* 20000 */ {(12<<2)|2,{116,71,0}},
/* 20001 */ {(12<<2)|2,{116,71,0}},
/* 20002 */ {(12<<2)|2,{116,71,0}},
/* 20003 */ {(12<<2)|2,{116,71,0}},
/* 20004 */ {(12<<2)|2,{116,71,0}},
/* 20005 */ {(12<<2)|2,{116,71,0}},
/* 20006 */ {(12<<2)|2,{116,71,0}},
/* 20007 */ {(12<<2)|2,{116,71,0}},
/* 20008 */ {(12<<2)|2,{116,71,0}},
/* 20009 */ {(12<<2)|2,{116,71,0}},
/* 20010 */ {(12<<2)|2,{116,71,0}},
/* 20011 */ {(12<<2)|2,{116,71,0}},
/* 20012 */ {(12<<2)|2,{116,71,0}},
/* 20013 */ {(12<<2)|2,{116,71,0}},
/* 20014 */ {(12<<2)|2,{116,71,0}},
/* 20015 */ {(12<<2)|2,{116,71,0}},
/* 20016 */ {(12<<2)|2,{116,72,0}},
/* 20017 */ {(12<<2)|2,{116,72,0}},
/* 20018 */ {(12<<2)|2,{116,72,0}},
/* 20019 */ {(12<<2)|2,{116,72,0}},
/* 20020 */ {(12<<2)|2,{116,72,0}},
/* 20021 */ {(12<<2)|2,{116,72,0}},
/* 20022 */ {(12<<2)|2,{116,72,0}},
/* 20023 */ {(12<<2)|2,{116,72,0}},
/* 20024 */ {(12<<2)|2,{116,72,0}},
/* 20025 */ {(12<<2)|2,{116,72,0}},
/* 20026 */ {(12<<2)|2,{116,72,0}},
/* 20027 */ {(12<<2)|2,{116,72,0}},
/* 20028 */ {(12<<2)|2,{116,72,0}},
/* 20029 */ {(12<<2)|2,{116,72,0}},
/* 20030 */ {(12<<2)|2,{116,72,0}},
/* 20031 */ {(12<<2)|2,{116,72,0}},
/* 20032 */ {(12<<2)|2,{116,73,0}},
/* 20033 */ {(12<<2)|2,{116,73,0}},
/* 20034 */ {(12<<2)|2,{116,73,0}},
/* 20035 */ {(12<<2)|2,{116,73,0}},
/* 20036 */ {(12<<2)|2,{116,73,0}},
/* 20037 */ {(12<<2)|2,{116,73,0}},
/* 20038 */ {(12<<2)|2,{116,73,0}},
/* 20039 */ {(12<<2)|2,{116,73,0}},
/* 20040 */ {(12<<2)|2,{116,73,0}},
/* 20041 */ {(12<<2)|2,{116,73,0}},
/* 20042 */ {(12<<2)|2,{116,73,0}},
/* 20043 */ {(12<<2)|2,{116,73,0}},
/* 20044 */ {(12<<2)|2,{116,73,0}},
/* 20045 */ {(12<<2)|2,{116,73,0}},
/* 20046 */ {(12<<2)|2,{116,73,0}},
/* 20047 */ {(12<<2)|2,{116,73,0}},
/* 20048 */ {(12<<2)|2,{116,74,0}},
/* 20049 */ {(12<<2)|2,{116,74,0}},
/* 20050 */ {(12<<2)|2,{116,74,0}},
/* 20051 */ {(12<<2)|2,{116,74,0}},
/* 20052 */ {(12<<2)|2,{116,74,0}},
/* 20053 */ {(12<<2)|2,{116,74,0}},
/* 20054 */ {(12<<2)|2,{116,74,0}},
/* 20055 */ {(12<<2)|2,{116,74,0}},
/* 20056 */ {(12<<2)|2,{116,74,0}},
/* 20057 */ {(12<<2)|2,{116,74,0}},
/* 20058 */ {(12<<2)|2,{116,74,0}},
/* 20059 */ {(12<<2)|2,{116,74,0}},
/* 20060 */ {(12<<2)|2,{116,74,0}},
/* 20061 */ {(12<<2)|2,{116,74,0}},
/* 20062 */ {(12<<2)|2,{116,74,0}},
/* 20063 */ {(12<<2)|2,{116,74,0}},
/* 20064 */ {(12<<2)|2,{116,75,0}},
/* 20065 */ {(12<<2)|2,{116,75,0}},
/* 20066 */ {(12<<2)|2,{116,75,0}},
/* 20067 */ {(12<<2)|2,{116,75,0}},
/* 20068 */ {(12<<2)|2,{116,75,0}},
/* 20069 */ {(12<<2)|2,{116,75,0}},
/* 20070 */ {(12<<2)|2,{116,75,0}},
/* 20071 */ {(12<<2)|2,{116,75,0}},
/* 20072 */ {(12<<2)|2,{116,75,0}},
/* 20073 */ {(12<<2)|2,{116,75,0}},
/* 20074 */ {(12<<2)|2,{116,75,0}},
/* 20075 */ {(12<<2)|2,{116,75,0}},
/* 20076 */ {(12<<2)|2,{116,75,0}},
/* 20077 */ {(12<<2)|2,{116,75,0}},
/* 20078 */ {(12<<2)|2,{116,75,0}},
/* 20079 */ {(12<<2)|2,{116,75,0}},
/* 20080 */ {(12<<2)|2,{116,76,0}},
/* 20081 */ {(12<<2)|2,{116,76,0}},
/* 20082 */ {(12<<2)|2,{116,76,0}},
/* 20083 */ {(12<<2)|2,{116,76,0}},
/* 20084 */ {(12<<2)|2,{116,76,0}},
/* 20085 */ {(12<<2)|2,{116,76,0}},
/* 20086 */ {(12<<2)|2,{116,76,0}},
/* 20087 */ {(12<<2)|2,{116,76,0}},
/* 20088 */ {(12<<2)|2,{116,76,0}},
/* 20089 */ {(12<<2)|2,{116,76,0}},
/* 20090 */ {(12<<2)|2,{116,76,0}},
/* 20091 */ {(12<<2)|2,{116,76,0}},
/* 20092 */ {(12<<2)|2,{116,76,0}},
/* 20093 */ {(12<<2)|2,{116,76,0}},
/* 20094 */ {(12<<2)|2,{116,76,0}},
/* 20095 */ {(12<<2)|2,{116,76,0}},
/* 20096 */ {(12<<2)|2,{116,77,0}},
/* 20097 */ {(12<<2)|2,{116,77,0}},
/* 20098 */ {(12<<2)|2,{116,77,0}},
/* 20099 */ {(12<<2)|2,{116,77,0}},
/* 20100 */ {(12<<2)|2,{116,77,0}},
/* 20101 */ {(12<<2)|2,{116,77,0}},
/* 20102 */ {(12<<2)|2,{116,77,0}},
/* 20103 */ {(12<<2)|2,{116,77,0}},
/* 20104 */ {(12<<2)|2,{116,77,0}},
/* 20105 */ {(12<<2)|2,{116,77,0}},
/* 20106 */ {(12<<2)|2,{116,77,0}},
/* 20107 */ {(12<<2)|2,{116,77,0}},
/* 20108 */ {(12<<2)|2,{116,77,0}},
/* 20109 */ {(12<<2)|2,{116,77,0}},
/* 20110 */ {(12<<2)|2,{116,77,0}},
/* 20111 */ {(12<<2)|2,{116,77,0}},
/* 20112 */ {(12<<2)|2,{116,78,0}},
/* 20113 */ {(12<<2)|2,{116,78,0}},
/* 20114 */ {(12<<2)|2,{116,78,0}},
/* 20115 */ {(12<<2)|2,{116,78,0}},
/* 20116 */ {(12<<2)|2,{116,78,0}},
/* 20117 */ {(12<<2)|2,{116,78,0}},
/* 20118 */ {(12<<2)|2,{116,78,0}},
/* 20119 */ {(12<<2)|2,{116,78,0}},
/* 20120 */ {(12<<2)|2,{116,78,0}},
/* 20121 */ {(12<<2)|2,{116,78,0}},
/* 20122 */ {(12<<2)|2,{116,78,0}},
/* 20123 */ {(12<<2)|2,{116,78,0}},
/* 20124 */ {(12<<2)|2,{116,78,0}},
/* 20125 */ {(12<<2)|2,{116,78,0}},
/* 20126 */ {(12<<2)|2,{116,78,0}},
/* 20127 */ {(12<<2)|2,{116,78,0}},
/* 20128 */ {(12<<2)|2,{116,79,0}},
/* 20129 */ {(12<<2)|2,{116,79,0}},
/* 20130 */ {(12<<2)|2,{116,79,0}},
/* 20131 */ {(12<<2)|2,{116,79,0}},
/* 20132 */ {(12<<2)|2,{116,79,0}},
/* 20133 */ {(12<<2)|2,{116,79,0}},
/* 20134 */ {(12<<2)|2,{116,79,0}},
/* 20135 */ {(12<<2)|2,{116,79,0}},
/* 20136 */ {(12<<2)|2,{116,79,0}},
/* 20137 */ {(12<<2)|2,{116,79,0}},
/* 20138 */ {(12<<2)|2,{116,79,0}},
/* 20139 */ {(12<<2)|2,{116,79,0}},
/* 20140 */ {(12<<2)|2,{116,79,0}},
/* 20141 */ {(12<<2)|2,{116,79,0}},
/* 20142 */ {(12<<2)|2,{116,79,0}},
/* 20143 */ {(12<<2)|2,{116,79,0}},
/* 20144 */ {(12<<2)|2,{116,80,0}},
/* 20145 */ {(12<<2)|2,{116,80,0}},
/* 20146 */ {(12<<2)|2,{116,80,0}},
/* 20147 */ {(12<<2)|2,{116,80,0}},
/* 20148 */ {(12<<2)|2,{116,80,0}},
/* 20149 */ {(12<<2)|2,{116,80,0}},
/* 20150 */ {(12<<2)|2,{116,80,0}},
/* 20151 */ {(12<<2)|2,{116,80,0}},
/* 20152 */ {(12<<2)|2,{116,80,0}},
/* 20153 */ {(12<<2)|2,{116,80,0}},
/* 20154 */ {(12<<2)|2,{116,80,0}},
/* 20155 */ {(12<<2)|2,{116,80,0}},
/* 20156 */ {(12<<2)|2,{116,80,0}},
/* 20157 */ {(12<<2)|2,{116,80,0}},
/* 20158 */ {(12<<2)|2,{116,80,0}},
/* 20159 */ {(12<<2)|2,{116,80,0}},
/* 20160 */ {(12<<2)|2,{116,81,0}},
/* 20161 */ {(12<<2)|2,{116,81,0}},
/* 20162 */ {(12<<2)|2,{116,81,0}},
/* 20163 */ {(12<<2)|2,{116,81,0}},
/* 20164 */ {(12<<2)|2,{116,81,0}},
/* 20165 */ {(12<<2)|2,{116,81,0}},
/* 20166 */ {(12<<2)|2,{116,81,0}},
/* 20167 */ {(12<<2)|2,{116,81,0}},
/* 20168 */ {(12<<2)|2,{116,81,0}},
/* 20169 */ {(12<<2)|2,{116,81,0}},
/* 20170 */ {(12<<2)|2,{116,81,0}},
/* 20171 */ {(12<<2)|2,{116,81,0}},
/* 20172 */ {(12<<2)|2,{116,81,0}},
/* 20173 */ {(12<<2)|2,{116,81,0}},
/* 20174 */ {(12<<2)|2,{116,81,0}},
/* 20175 */ {(12<<2)|2,{116,81,0}},
/* 20176 */ {(12<<2)|2,{116,82,0}},
/* 20177 */ {(12<<2)|2,{116,82,0}},
/* 20178 */ {(12<<2)|2,{116,82,0}},
/* 20179 */ {(12<<2)|2,{116,82,0}},
/* 20180 */ {(12<<2)|2,{116,82,0}},
/* 20181 */ {(12<<2)|2,{116,82,0}},
/* 20182 */ {(12<<2)|2,{116,82,0}},
/* 20183 */ {(12<<2)|2,{116,82,0}},
/* 20184 */ {(12<<2)|2,{116,82,0}},
/* 20185 */ {(12<<2)|2,{116,82,0}},
/* 20186 */ {(12<<2)|2,{116,82,0}},
/* 20187 */ {(12<<2)|2,{116,82,0}},
/* 20188 */ {(12<<2)|2,{116,82,0}},
/* 20189 */ {(12<<2)|2,{116,82,0}},
/* 20190 */ {(12<<2)|2,{116,82,0}},
/* 20191 */ {(12<<2)|2,{116,82,0}},
/* 20192 */ {(12<<2)|2,{116,83,0}},
/* 20193 */ {(12<<2)|2,{116,83,0}},
/* 20194 */ {(12<<2)|2,{116,83,0}},
/* 20195 */ {(12<<2)|2,{116,83,0}},
/* 20196 */ {(12<<2)|2,{116,83,0}},
/* 20197 */ {(12<<2)|2,{116,83,0}},
/* 20198 */ {(12<<2)|2,{116,83,0}},
/* 20199 */ {(12<<2)|2,{116,83,0}},
/* 20200 */ {(12<<2)|2,{116,83,0}},
/* 20201 */ {(12<<2)|2,{116,83,0}},
/* 20202 */ {(12<<2)|2,{116,83,0}},
/* 20203 */ {(12<<2)|2,{116,83,0}},
/* 20204 */ {(12<<2)|2,{116,83,0}},
/* 20205 */ {(12<<2)|2,{116,83,0}},
/* 20206 */ {(12<<2)|2,{116,83,0}},
/* 20207 */ {(12<<2)|2,{116,83,0}},
/* 20208 */ {(12<<2)|2,{116,84,0}},
/* 20209 */ {(12<<2)|2,{116,84,0}},
/* 20210 */ {(12<<2)|2,{116,84,0}},
/* 20211 */ {(12<<2)|2,{116,84,0}},
/* 20212 */ {(12<<2)|2,{116,84,0}},
/* 20213 */ {(12<<2)|2,{116,84,0}},
/* 20214 */ {(12<<2)|2,{116,84,0}},
/* 20215 */ {(12<<2)|2,{116,84,0}},
/* 20216 */ {(12<<2)|2,{116,84,0}},
/* 20217 */ {(12<<2)|2,{116,84,0}},
/* 20218 */ {(12<<2)|2,{116,84,0}},
/* 20219 */ {(12<<2)|2,{116,84,0}},
/* 20220 */ {(12<<2)|2,{116,84,0}},
/* 20221 */ {(12<<2)|2,{116,84,0}},
/* 20222 */ {(12<<2)|2,{116,84,0}},
/* 20223 */ {(12<<2)|2,{116,84,0}},
/* 20224 */ {(12<<2)|2,{116,85,0}},
/* 20225 */ {(12<<2)|2,{116,85,0}},
/* 20226 */ {(12<<2)|2,{116,85,0}},
/* 20227 */ {(12<<2)|2,{116,85,0}},
/* 20228 */ {(12<<2)|2,{116,85,0}},
/* 20229 */ {(12<<2)|2,{116,85,0}},
/* 20230 */ {(12<<2)|2,{116,85,0}},
/* 20231 */ {(12<<2)|2,{116,85,0}},
/* 20232 */ {(12<<2)|2,{116,85,0}},
/* 20233 */ {(12<<2)|2,{116,85,0}},
/* 20234 */ {(12<<2)|2,{116,85,0}},
/* 20235 */ {(12<<2)|2,{116,85,0}},
/* 20236 */ {(12<<2)|2,{116,85,0}},
/* 20237 */ {(12<<2)|2,{116,85,0}},
/* 20238 */ {(12<<2)|2,{116,85,0}},
/* 20239 */ {(12<<2)|2,{116,85,0}},
/* 20240 */ {(12<<2)|2,{116,86,0}},
/* 20241 */ {(12<<2)|2,{116,86,0}},
/* 20242 */ {(12<<2)|2,{116,86,0}},
/* 20243 */ {(12<<2)|2,{116,86,0}},
/* 20244 */ {(12<<2)|2,{116,86,0}},
/* 20245 */ {(12<<2)|2,{116,86,0}},
/* 20246 */ {(12<<2)|2,{116,86,0}},
/* 20247 */ {(12<<2)|2,{116,86,0}},
/* 20248 */ {(12<<2)|2,{116,86,0}},
/* 20249 */ {(12<<2)|2,{116,86,0}},
/* 20250 */ {(12<<2)|2,{116,86,0}},
/* 20251 */ {(12<<2)|2,{116,86,0}},
/* 20252 */ {(12<<2)|2,{116,86,0}},
/* 20253 */ {(12<<2)|2,{116,86,0}},
/* 20254 */ {(12<<2)|2,{116,86,0}},
/* 20255 */ {(12<<2)|2,{116,86,0}},
/* 20256 */ {(12<<2)|2,{116,87,0}},
/* 20257 */ {(12<<2)|2,{116,87,0}},
/* 20258 */ {(12<<2)|2,{116,87,0}},
/* 20259 */ {(12<<2)|2,{116,87,0}},
/* 20260 */ {(12<<2)|2,{116,87,0}},
/* 20261 */ {(12<<2)|2,{116,87,0}},
/* 20262 */ {(12<<2)|2,{116,87,0}},
/* 20263 */ {(12<<2)|2,{116,87,0}},
/* 20264 */ {(12<<2)|2,{116,87,0}},
/* 20265 */ {(12<<2)|2,{116,87,0}},
/* 20266 */ {(12<<2)|2,{116,87,0}},
/* 20267 */ {(12<<2)|2,{116,87,0}},
/* 20268 */ {(12<<2)|2,{116,87,0}},
/* 20269 */ {(12<<2)|2,{116,87,0}},
/* 20270 */ {(12<<2)|2,{116,87,0}},
/* 20271 */ {(12<<2)|2,{116,87,0}},
/* 20272 */ {(12<<2)|2,{116,89,0}},
/* 20273 */ {(12<<2)|2,{116,89,0}},
/* 20274 */ {(12<<2)|2,{116,89,0}},
/* 20275 */ {(12<<2)|2,{116,89,0}},
/* 20276 */ {(12<<2)|2,{116,89,0}},
/* 20277 */ {(12<<2)|2,{116,89,0}},
/* 20278 */ {(12<<2)|2,{116,89,0}},
/* 20279 */ {(12<<2)|2,{116,89,0}},
/* 20280 */ {(12<<2)|2,{116,89,0}},
/* 20281 */ {(12<<2)|2,{116,89,0}},
/* 20282 */ {(12<<2)|2,{116,89,0}},
/* 20283 */ {(12<<2)|2,{116,89,0}},
/* 20284 */ {(12<<2)|2,{116,89,0}},
/* 20285 */ {(12<<2)|2,{116,89,0}},
/* 20286 */ {(12<<2)|2,{116,89,0}},
/* 20287 */ {(12<<2)|2,{116,89,0}},
/* 20288 */ {(12<<2)|2,{116,106,0}},
/* 20289 */ {(12<<2)|2,{116,106,0}},
/* 20290 */ {(12<<2)|2,{116,106,0}},
/* 20291 */ {(12<<2)|2,{116,106,0}},
/* 20292 */ {(12<<2)|2,{116,106,0}},
/* 20293 */ {(12<<2)|2,{116,106,0}},
/* 20294 */ {(12<<2)|2,{116,106,0}},
/* 20295 */ {(12<<2)|2,{116,106,0}},
/* 20296 */ {(12<<2)|2,{116,106,0}},
/* 20297 */ {(12<<2)|2,{116,106,0}},
/* 20298 */ {(12<<2)|2,{116,106,0}},
/* 20299 */ {(12<<2)|2,{116,106,0}},
/* 20300 */ {(12<<2)|2,{116,106,0}},
/* 20301 */ {(12<<2)|2,{116,106,0}},
/* 20302 */ {(12<<2)|2,{116,106,0}},
/* 20303 */ {(12<<2)|2,{116,106,0}},
/* 20304 */ {(12<<2)|2,{116,107,0}},
/* 20305 */ {(12<<2)|2,{116,107,0}},
/* 20306 */ {(12<<2)|2,{116,107,0}},
/* 20307 */ {(12<<2)|2,{116,107,0}},
/* 20308 */ {(12<<2)|2,{116,107,0}},
/* 20309 */ {(12<<2)|2,{116,107,0}},
/* 20310 */ {(12<<2)|2,{116,107,0}},
/* 20311 */ {(12<<2)|2,{116,107,0}},
/* 20312 */ {(12<<2)|2,{116,107,0}},
/* 20313 */ {(12<<2)|2,{116,107,0}},
/* 20314 */ {(12<<2)|2,{116,107,0}},
/* 20315 */ {(12<<2)|2,{116,107,0}},
/* 20316 */ {(12<<2)|2,{116,107,0}},
/* 20317 */ {(12<<2)|2,{116,107,0}},
/* 20318 */ {(12<<2)|2,{116,107,0}},
/* 20319 */ {(12<<2)|2,{116,107,0}},
/* 20320 */ {(12<<2)|2,{116,113,0}},
/* 20321 */ {(12<<2)|2,{116,113,0}},
/* 20322 */ {(12<<2)|2,{116,113,0}},
/* 20323 */ {(12<<2)|2,{116,113,0}},
/* 20324 */ {(12<<2)|2,{116,113,0}},
/* 20325 */ {(12<<2)|2,{116,113,0}},
/* 20326 */ {(12<<2)|2,{116,113,0}},
/* 20327 */ {(12<<2)|2,{116,113,0}},
/* 20328 */ {(12<<2)|2,{116,113,0}},
/* 20329 */ {(12<<2)|2,{116,113,0}},
/* 20330 */ {(12<<2)|2,{116,113,0}},
/* 20331 */ {(12<<2)|2,{116,113,0}},
/* 20332 */ {(12<<2)|2,{116,113,0}},
/* 20333 */ {(12<<2)|2,{116,113,0}},
/* 20334 */ {(12<<2)|2,{116,113,0}},
/* 20335 */ {(12<<2)|2,{116,113,0}},
/* 20336 */ {(12<<2)|2,{116,118,0}},
/* 20337 */ {(12<<2)|2,{116,118,0}},
/* 20338 */ {(12<<2)|2,{116,118,0}},
/* 20339 */ {(12<<2)|2,{116,118,0}},
/* 20340 */ {(12<<2)|2,{116,118,0}},
/* 20341 */ {(12<<2)|2,{116,118,0}},
/* 20342 */ {(12<<2)|2,{116,118,0}},
/* 20343 */ {(12<<2)|2,{116,118,0}},
/* 20344 */ {(12<<2)|2,{116,118,0}},
/* 20345 */ {(12<<2)|2,{116,118,0}},
/* 20346 */ {(12<<2)|2,{116,118,0}},
/* 20347 */ {(12<<2)|2,{116,118,0}},
/* 20348 */ {(12<<2)|2,{116,118,0}},
/* 20349 */ {(12<<2)|2,{116,118,0}},
/* 20350 */ {(12<<2)|2,{116,118,0}},
/* 20351 */ {(12<<2)|2,{116,118,0}},
/* 20352 */ {(12<<2)|2,{116,119,0}},
/* 20353 */ {(12<<2)|2,{116,119,0}},
/* 20354 */ {(12<<2)|2,{116,119,0}},
/* 20355 */ {(12<<2)|2,{116,119,0}},
/* 20356 */ {(12<<2)|2,{116,119,0}},
/* 20357 */ {(12<<2)|2,{116,119,0}},
/* 20358 */ {(12<<2)|2,{116,119,0}},
/* 20359 */ {(12<<2)|2,{116,119,0}},
/* 20360 */ {(12<<2)|2,{116,119,0}},
/* 20361 */ {(12<<2)|2,{116,119,0}},
/* 20362 */ {(12<<2)|2,{116,119,0}},
/* 20363 */ {(12<<2)|2,{116,119,0}},
/* 20364 */ {(12<<2)|2,{116,119,0}},
/* 20365 */ {(12<<2)|2,{116,119,0}},
/* 20366 */ {(12<<2)|2,{116,119,0}},
/* 20367 */ {(12<<2)|2,{116,119,0}},
/* 20368 */ {(12<<2)|2,{116,120,0}},
/* 20369 */ {(12<<2)|2,{116,120,0}},
/* 20370 */ {(12<<2)|2,{116,120,0}},
/* 20371 */ {(12<<2)|2,{116,120,0}},
/* 20372 */ {(12<<2)|2,{116,120,0}},
/* 20373 */ {(12<<2)|2,{116,120,0}},
/* 20374 */ {(12<<2)|2,{116,120,0}},
/* 20375 */ {(12<<2)|2,{116,120,0}},
/* 20376 */ {(12<<2)|2,{116,120,0}},
/* 20377 */ {(12<<2)|2,{116,120,0}},
/* 20378 */ {(12<<2)|2,{116,120,0}},
/* 20379 */ {(12<<2)|2,{116,120,0}},
/* 20380 */ {(12<<2)|2,{116,120,0}},
/* 20381 */ {(12<<2)|2,{116,120,0}},
/* 20382 */ {(12<<2)|2,{116,120,0}},
/* 20383 */ {(12<<2)|2,{116,120,0}},
/* 20384 */ {(12<<2)|2,{116,121,0}},
/* 20385 */ {(12<<2)|2,{116,121,0}},
/* 20386 */ {(12<<2)|2,{116,121,0}},
/* 20387 */ {(12<<2)|2,{116,121,0}},
/* 20388 */ {(12<<2)|2,{116,121,0}},
/* 20389 */ {(12<<2)|2,{116,121,0}},
/* 20390 */ {(12<<2)|2,{116,121,0}},
/* 20391 */ {(12<<2)|2,{116,121,0}},
/* 20392 */ {(12<<2)|2,{116,121,0}},
/* 20393 */ {(12<<2)|2,{116,121,0}},
/* 20394 */ {(12<<2)|2,{116,121,0}},
/* 20395 */ {(12<<2)|2,{116,121,0}},
/* 20396 */ {(12<<2)|2,{116,121,0}},
/* 20397 */ {(12<<2)|2,{116,121,0}},
/* 20398 */ {(12<<2)|2,{116,121,0}},
/* 20399 */ {(12<<2)|2,{116,121,0}},
/* 20400 */ {(12<<2)|2,{116,122,0}},
/* 20401 */ {(12<<2)|2,{116,122,0}},
/* 20402 */ {(12<<2)|2,{116,122,0}},
/* 20403 */ {(12<<2)|2,{116,122,0}},
/* 20404 */ {(12<<2)|2,{116,122,0}},
/* 20405 */ {(12<<2)|2,{116,122,0}},
/* 20406 */ {(12<<2)|2,{116,122,0}},
/* 20407 */ {(12<<2)|2,{116,122,0}},
/* 20408 */ {(12<<2)|2,{116,122,0}},
/* 20409 */ {(12<<2)|2,{116,122,0}},
/* 20410 */ {(12<<2)|2,{116,122,0}},
/* 20411 */ {(12<<2)|2,{116,122,0}},
/* 20412 */ {(12<<2)|2,{116,122,0}},
/* 20413 */ {(12<<2)|2,{116,122,0}},
/* 20414 */ {(12<<2)|2,{116,122,0}},
/* 20415 */ {(12<<2)|2,{116,122,0}},
/* 20416 */ {(13<<2)|2,{116,38,0}},
/* 20417 */ {(13<<2)|2,{116,38,0}},
/* 20418 */ {(13<<2)|2,{116,38,0}},
/* 20419 */ {(13<<2)|2,{116,38,0}},
/* 20420 */ {(13<<2)|2,{116,38,0}},
/* 20421 */ {(13<<2)|2,{116,38,0}},
/* 20422 */ {(13<<2)|2,{116,38,0}},
/* 20423 */ {(13<<2)|2,{116,38,0}},
/* 20424 */ {(13<<2)|2,{116,42,0}},
/* 20425 */ {(13<<2)|2,{116,42,0}},
/* 20426 */ {(13<<2)|2,{116,42,0}},
/* 20427 */ {(13<<2)|2,{116,42,0}},
/* 20428 */ {(13<<2)|2,{116,42,0}},
/* 20429 */ {(13<<2)|2,{116,42,0}},
/* 20430 */ {(13<<2)|2,{116,42,0}},
/* 20431 */ {(13<<2)|2,{116,42,0}},
/* 20432 */ {(13<<2)|2,{116,44,0}},
/* 20433 */ {(13<<2)|2,{116,44,0}},
/* 20434 */ {(13<<2)|2,{116,44,0}},
/* 20435 */ {(13<<2)|2,{116,44,0}},
/* 20436 */ {(13<<2)|2,{116,44,0}},
/* 20437 */ {(13<<2)|2,{116,44,0}},
/* 20438 */ {(13<<2)|2,{116,44,0}},
/* 20439 */ {(13<<2)|2,{116,44,0}},
/* 20440 */ {(13<<2)|2,{116,59,0}},
/* 20441 */ {(13<<2)|2,{116,59,0}},
/* 20442 */ {(13<<2)|2,{116,59,0}},
/* 20443 */ {(13<<2)|2,{116,59,0}},
/* 20444 */ {(13<<2)|2,{116,59,0}},
/* 20445 */ {(13<<2)|2,{116,59,0}},
/* 20446 */ {(13<<2)|2,{116,59,0}},
/* 20447 */ {(13<<2)|2,{116,59,0}},
/* 20448 */ {(13<<2)|2,{116,88,0}},
/* 20449 */ {(13<<2)|2,{116,88,0}},
/* 20450 */ {(13<<2)|2,{116,88,0}},
/* 20451 */ {(13<<2)|2,{116,88,0}},
/* 20452 */ {(13<<2)|2,{116,88,0}},
/* 20453 */ {(13<<2)|2,{116,88,0}},
/* 20454 */ {(13<<2)|2,{116,88,0}},
/* 20455 */ {(13<<2)|2,{116,88,0}},
/* 20456 */ {(13<<2)|2,{116,90,0}},
/* 20457 */ {(13<<2)|2,{116,90,0}},
/* 20458 */ {(13<<2)|2,{116,90,0}},
/* 20459 */ {(13<<2)|2,{116,90,0}},
/* 20460 */ {(13<<2)|2,{116,90,0}},
/* 20461 */ {(13<<2)|2,{116,90,0}},
/* 20462 */ {(13<<2)|2,{116,90,0}},
/* 20463 */ {(13<<2)|2,{116,90,0}},
/* 20464 */ {(15<<2)|2,{116,33,0}},
/* 20465 */ {(15<<2)|2,{116,33,0}},
/* 20466 */ {(15<<2)|2,{116,34,0}},
/* 20467 */ {(15<<2)|2,{116,34,0}},
/* 20468 */ {(15<<2)|2,{116,40,0}},
/* 20469 */ {(15<<2)|2,{116,40,0}},
/* 20470 */ {(15<<2)|2,{116,41,0}},
/* 20471 */ {(15<<2)|2,{116,41,0}},
/* 20472 */ {(15<<2)|2,{116,63,0}},
/* 20473 */ {(15<<2)|2,{116,63,0}},
/* 20474 */ {(16<<2)|2,{116,39,0}},
/* 20475 */ {(16<<2)|2,{116,43,0}},
/* 20476 */ {(16<<2)|2,{116,124,0}},
/* 20477 */ {(5<<2)|1,{116,0,0}},
/* 20478 */ {(5<<2)|1,{116,0,0}},
/* 20479 */ {(5<<2)|1,{116,0,0}},
/* 20480 */ {(16<<2)|3,{32,48,48}},
/* 20481 */ {(16<<2)|3,{32,48,49}},
/* 20482 */ {(16<<2)|3,{32,48,50}},
/* 20483 */ {(16<<2)|3,{32,48,97}},
/* 20484 */ {(16<<2)|3,{32,48,99}},
/* 20485 */ {(16<<2)|3,{32,48,101}},
/* 20486 */ {(16<<2)|3,{32,48,105}},
/* 20487 */ {(16<<2)|3,{32,48,111}},
/* 20488 */ {(16<<2)|3,{32,48,115}},
/* 20489 */ {(16<<2)|3,{32,48,116}},
/* 20490 */ {(11<<2)|2,{32,48,0}},
/* 20491 */ {(11<<2)|2,{32,48,0}},
/* 20492 */ {(11<<2)|2,{32,48,0}},
/* 20493 */ {(11<<2)|2,{32,48,0}},
/* 20494 */ {(11<<2)|2,{32,48,0}},
/* 20495 */ {(11<<2)|2,{32,48,0}},
/* 20496 */ {(11<<2)|2,{32,48,0}},
/* 20497 */ {(11<<2)|2,{32,48,0}},
/* 20498 */ {(11<<2)|2,{32,48,0}},
/* 20499 */ {(11<<2)|2,{32,48,0}},
/* 20500 */ {(11<<2)|2,{32,48,0}},
/* 20501 */ {(11<<2)|2,{32,48,0}},
/* 20502 */ {(11<<2)|2,{32,48,0}},
/* 20503 */ {(11<<2)|2,{32,48,0}},
/* 20504 */ {(11<<2)|2,{32,48,0}},
/* 20505 */ {(11<<2)|2,{32,48,0}},
/* 20506 */ {(11<<2)|2,{32,48,0}},
/* 20507 */ {(11<<2)|2,{32,48,0}},
/* 20508 */ {(11<<2)|2,{32,48,0}},
/* 20509 */ {(11<<2)|2,{32,48,0}},
/* 20510 */ {(11<<2)|2,{32,48,0}},
/* 20511 */ {(11<<2)|2,{32,48,0}},
/* 20512 */ {(16<<2)|3,{32,49,48}},
/* 20513 */ {(16<<2)|3,{32,49,49}},
/* 20514 */ {(16<<2)|3,{32,49,50}},
/* 20515 */ {(16<<2)|3,{32,49,97}},
/* 20516 */ {(16<<2)|3,{32,49,99}},
/* 20517 */ {(16<<2)|3,{32,49,101}},
/* 20518 */ {(16<<2)|3,{32,49,105}},
/* 20519 */ {(16<<2)|3,{32,49,111}},
/* 20520 */ {(16<<2)|3,{32,49,115}},
/* 20521 */ {(16<<2)|3,{32,49,116}},
/* 20522 */ {(11<<2)|2,{32,49,0}},
/* 20523 */ {(11<<2)|2,{32,49,0}},
/* 20524 */ {(11<<2)|2,{32,49,0}},
/* 20525 */ {(11<<2)|2,{32,49,0}},
/* 20526 */ {(11<<2)|2,{32,49,0}},
/* 20527 */ {(11<<2)|2,{32,49,0}},
/* 20528 */ {(11<<2)|2,{32,49,0}},
/* 20529 */ {(11<<2)|2,{32,49,0}},
/* 20530 */ {(11<<2)|2,{32,49,0}},
/* 20531 */ {(11<<2)|2,{32,49,0}},
/* 20532 */ {(11<<2)|2,{32,49,0}},
/* 20533 */ {(11<<2)|2,{32,49,0}},
/* 20534 */ {(11<<2)|2,{32,49,0}},
/* 20535 */ {(11<<2)|2,{32,49,0}},
/* 20536 */ {(11<<2)|2,{32,49,0}},
/* 20537 */ {(11<<2)|2,{32,49,0}},
/* 20538 */ {(11<<2)|2,{32,49,0}},
/* 20539 */ {(11<<2)|2,{32,49,0}},
/* 20540 */ {(11<<2)|2,{32,49,0}},
/* 20541 */ {(11<<2)|2,{32,49,0}},
/* 20542 */ {(11<<2)|2,{32,49,0}},
/* 20543 */ {(11<<2)|2,{32,49,0}},
/* 20544 */ {(16<<2)|3,{32,50,48}},
/* 20545 */ {(16<<2)|3,{32,50,49}},
/* 20546 */ {(16<<2)|3,{32,50,50}},
/* 20547 */ {(16<<2)|3,{32,50,97}},
/* 20548 */ {(16<<2)|3,{32,50,99}},
/* 20549 */ {(16<<2)|3,{32,50,101}},
/* 20550 */ {(16<<2)|3,{32,50,105}},
/* 20551 */ {(16<<2)|3,{32,50,111}},
/* 20552 */ {(16<<2)|3,{32,50,115}},
/* 20553 */ {(16<<2)|3,{32,50,116}},
/* 20554 */ {(11<<2)|2,{32,50,0}},
/* 20555 */ {(11<<2)|2,{32,50,0}},
/* 20556 */ {(11<<2)|2,{32,50,0}},
/* 20557 */ {(11<<2)|2,{32,50,0}},
/* 20558 */ {(11<<2)|2,{32,50,0}},
/* 20559 */ {(11<<2)|2,{32,50,0}},
/* 20560 */ {(11<<2)|2,{32,50,0}},
/* 20561 */ {(11<<2)|2,{32,50,0}},
/* 20562 */ {(11<<2)|2,{32,50,0}},
/* 20563 */ {(11<<2)|2,{32,50,0}},
/* 20564 */ {(11<<2)|2,{32,50,0}},
/* 20565 */ {(11<<2)|2,{32,50,0}},
/* 20566 */ {(11<<2)|2,{32,50,0}},
/* 20567 */ {(11<<2)|2,{32,50,0}},
/* 20568 */ {(11<<2)|2,{32,50,0}},
/* 20569 */ {(11<<2)|2,{32,50,0}},
/* 20570 */ {(11<<2)|2,{32,50,0}},
/* 20571 */ {(11<<2)|2,{32,50,0}},
/* 20572 */ {(11<<2)|2,{32,50,0}},
/* 20573 */ {(11<<2)|2,{32,50,0}},
/* 20574 */ {(11<<2)|2,{32,50,0}},
/* 20575 */ {(11<<2)|2,{32,50,0}},
/* 20576 */ {(16<<2)|3,{32,97,48}},
/* 20577 */ {(16<<2)|3,{32,97,49}},
/* 20578 */ {(16<<2)|3,{32,97,50}},
/* 20579 */ {(16<<2)|3,{32,97,97}},
/* 20580 */ {(16<<2)|3,{32,97,99}},
/* 20581 */ {(16<<2)|3,{32,97,101}},
/* 20582 */ {(16<<2)|3,{32,97,105}},
/* 20583 */ {(16<<2)|3,{32,97,111}},
/* 20584 */ {(16<<2)|3,{32,97,115}},
/* 20585 */ {(16<<2)|3,{32,97,116}},
/* 20586 */ {(11<<2)|2,{32,97,0}},
/* 20587 */ {(11<<2)|2,{32,97,0}},
/* 20588 */ {(11<<2)|2,{32,97,0}},
/* 20589 */ {(11<<2)|2,{32,97,0}},
/* 20590 */ {(11<<2)|2,{32,97,0}},
/* 20591 */ {(11<<2)|2,{32,97,0}},
/* 20592 */ {(11<<2)|2,{32,97,0}},
/* 20593 */ {(11<<2)|2,{32,97,0}},
/* 20594 */ {(11<<2)|2,{32,97,0}},
/* 20595 */ {(11<<2)|2,{32,97,0}},
/* 20596 */ {(11<<2)|2,{32,97,0}},
/* 20597 */ {(11<<2)|2,{32,97,0}},
/* 20598 */ {(11<<2)|2,{32,97,0}},
/* 20599 */ {(11<<2)|2,{32,97,0}},
/* 20600 */ {(11<<2)|2,{32,97,0}},
/* 20601 */ {(11<<2)|2,{32,97,0}},
/* 20602 */ {(11<<2)|2,{32,97,0}},
/* 20603 */ {(11<<2)|2,{32,97,0}},
/* 20604 */ {(11<<2)|2,{32,97,0}},
/* 20605 */ {(11<<2)|2,{32,97,0}},
/* 20606 */ {(11<<2)|2,{32,97,0}},
/* 20607 */ {(11<<2)|2,{32,97,0}},
/* 20608 */ {(16<<2)|3,{32,99,48}},
/* 20609 */ {(16<<2)|3,{32,99,49}},
/* 20610 */ {(16<<2)|3,{32,99,50}},
/* 20611 */ {(16<<2)|3,{32,99,97}},
/* 20612 */ {(16<<2)|3,{32,99,99}},
/* 20613 */ {(16<<2)|3,{32,99,101}},
/* 20614 */ {(16<<2)|3,{32,99,105}},
/* 20615 */ {(16<<2)|3,{32,99,111}},
/* 20616 */ {(16<<2)|3,{32,99,115}},
/* 20617 */ {(16<<2)|3,{32,99,116}},
/* 20618 */ {(11<<2)|2,{32,99,0}},
/* 20619 */ {(11<<2)|2,{32,99,0}},
/* 20620 */ {(11<<2)|2,{32,99,0}},
/* 20621 */ {(11<<2)|2,{32,99,0}},
/* 20622 */ {(11<<2)|2,{32,99,0}},
/* 20623 */ {(11<<2)|2,{32,99,0}},
/* 20624 */ {(11<<2)|2,{32,99,0}},
/* 20625 */ {(11<<2)|2,{32,99,0}},
/* 20626 */ {(11<<2)|2,{32,99,0}},
/* 20627 */ {(11<<2)|2,{32,99,0}},
/* 20628 */ {(11<<2)|2,{32,99,0}},
/* 20629 */ {(11<<2)|2,{32,99,0}},
/* 20630 */ {(11<<2)|2,{32,99,0}},
/* 20631 */ {(11<<2)|2,{32,99,0}},
/* 20632 */ {(11<<2)|2,{32,99,0}},
/* 20633 */ {(11<<2)|2,{32,99,0}},
/* 20634 */ {(11<<2)|2,{32,99,0}},
/* 20635 */ {(11<<2)|2,{32,99,0}},
/* 20636 */ {(11<<2)|2,{32,99,0}},
/* 20637 */ {(11<<2)|2,{32,99,0}},
/* 20638 */ {(11<<2)|2,{32,99,0}},
/* 20639 */ {(11<<2)|2,{32,99,0}},
/* 20640 */ {(16<<2)|3,{32,101,48}},
/* 20641 */ {(16<<2)|3,{32,101,49}},
/* 20642 */ {(16<<2)|3,{32,101,50}},
/* 20643 */ {(16<<2)|3,{32,101,97}},
/* 20644 */ {(16<<2)|3,{32,101,99}},
/* 20645 */ {(16<<2)|3,{32,101,101}},
/* 20646 */ {(16<<2)|3,{32,101,105}},
/* 20647 */ {(16<<2)|3,{32,101,111}},
/* 20648 */ {(16<<2)|3,{32,101,115}},
/* 20649 */ {(16<<2)|3,{32,101,116}},
/* 20650 */ {(11<<2)|2,{32,101,0}},
/* 20651 */ {(11<<2)|2,{32,101,0}},
/* 20652 */ {(11<<2)|2,{32,101,0}},
/* 20653 */ {(11<<2)|2,{32,101,0}},
/* 20654 */ {(11<<2)|2,{32,101,0}},
/* 20655 */ {(11<<2)|2,{32,101,0}},
/* 20656 */ {(11<<2)|2,{32,101,0}},
/* 20657 */ {(11<<2)|2,{32,101,0}},
/* 20658 */ {(11<<2)|2,{32,101,0}},
/* 20659 */ {(11<<2)|2,{32,101,0}},
/* 20660 */ {(11<<2)|2,{32,101,0}},
/* 20661 */ {(11<<2)|2,{32,101,0}},
/* 20662 */ {(11<<2)|2,{32,101,0}},
/* 20663 */ {(11<<2)|2,{32,101,0}},
/* 20664 */ {(11<<2)|2,{32,101,0}},
/* 20665 */ {(11<<2)|2,{32,101,0}},
/* 20666 */ {(11<<2)|2,{32,101,0}},
/* 20667 */ {(11<<2)|2,{32,101,0}},
/* 20668 */ {(11<<2)|2,{32,101,0}},
/* 20669 */ {(11<<2)|2,{32,101,0}},
/* 20670 */ {(11<<2)|2,{32,101,0}},
/* 20671 */ {(11<<2)|2,{32,101,0}},
/* 20672 */ {(16<<2)|3,{32,105,48}},
/* 20673 */ {(16<<2)|3,{32,105,49}},
/* 20674 */ {(16<<2)|3,{32,105,50}},
/* 20675 */ {(16<<2)|3,{32,105,97}},
/* 20676 */ {(16<<2)|3,{32,105,99}},
/* 20677 */ {(16<<2)|3,{32,105,101}},
/* 20678 */ {(16<<2)|3,{32,105,105}},
/* 20679 */ {(16<<2)|3,{32,105,111}},
/* 20680 */ {(16<<2)|3,{32,105,115}},
/* 20681 */ {(16<<2)|3,{32,105,116}},
/* 20682 */ {(11<<2)|2,{32,105,0}},
/* 20683 */ {(11<<2)|2,{32,105,0}},
/* 20684 */ {(11<<2)|2,{32,105,0}},
/* 20685 */ {(11<<2)|2,{32,105,0}},
/* 20686 */ {(11<<2)|2,{32,105,0}},
/* 20687 */ {(11<<2)|2,{32,105,0}},
/* 20688 */ {(11<<2)|2,{32,105,0}},
/* 20689 */ {(11<<2)|2,{32,105,0}},
/* 20690 */ {(11<<2)|2,{32,105,0}},
/* 20691 */ {(11<<2)|2,{32,105,0}},
/* 20692 */ {(11<<2)|2,{32,105,0}},
/* 20693 */ {(11<<2)|2,{32,105,0}},
/* 20694 */ {(11<<2)|2,{32,105,0}},
/* 20695 */ {(11<<2)|2,{32,105,0}},
/* 20696 */ {(11<<2)|2,{32,105,0}},
/* 20697 */ {(11<<2)|2,{32,105,0}},
/* 20698 */ {(11<<2)|2,{32,105,0}},
/* 20699 */ {(11<<2)|2,{32,105,0}},
/* 20700 */ {(11<<2)|2,{32,105,0}},
/* 20701 */ {(11<<2)|2,{32,105,0}},
/* 20702 */ {(11<<2)|2,{32,105,0}},
/* 20703 */ {(11<<2)|2,{32,105,0}},
/* 20704 */ {(16<<2)|3,{32,111,48}},
/* 20705 */ {(16<<2)|3,{32,111,49}},
/* 20706 */ {(16<<2)|3,{32,111,50}},
/* 20707 */ {(16<<2)|3,{32,111,97}},
/* 20708 */ {(16<<2)|3,{32,111,99}},
/* 20709 */ {(16<<2)|3,{32,111,101}},
/* 20710 */ {(16<<2)|3,{32,111,105}},
/* 20711 */ {(16<<2)|3,{32,111,111}},
/* 20712 */ {(16<<2)|3,{32,111,115}},
/* 20713 */ {(16<<2)|3,{32,111,116}},
/* 20714 */ {(11<<2)|2,{32,111,0}},
/* 20715 */ {(11<<2)|2,{32,111,0}},
/* 20716 */ {(11<<2)|2,{32,111,0}},
/* 20717 */ {(11<<2)|2,{32,111,0}},
/* 20718 */ {(11<<2)|2,{32,111,0}},
/* 20719 */ {(11<<2)|2,{32,111,0}},
/* 20720 */ {(11<<2)|2,{32,111,0}},
/* 20721 */ {(11<<2)|2,{32,111,0}},
/* 20722 */ {(11<<2)|2,{32,111,0}},
/* 20723 */ {(11<<2)|2,{32,111,0}},
/* 20724 */ {(11<<2)|2,{32,111,0}},
/* 20725 */ {(11<<2)|2,{32,111,0}},
/* 20726 */ {(11<<2)|2,{32,111,0}},
/* 20727 */ {(11<<2)|2,{32,111,0}},
/* 20728 */ {(11<<2)|2,{32,111,0}},
/* 20729 */ {(11<<2)|2,{32,111,0}},
/* 20730 */ {(11<<2)|2,{32,111,0}},
/* 20731 */ {(11<<2)|2,{32,111,0}},
/* 20732 */ {(11<<2)|2,{32,111,0}},
/* 20733 */ {(11<<2)|2,{32,111,0}},
/* 20734 */ {(11<<2)|2,{32,111,0}},
/* 20735 */ {(11<<2)|2,{32,111,0}},
/* 20736 */ {(16<<2)|3,{32,115,48}},
/* 20737 */ {(16<<2)|3,{32,115,49}},
/* 20738 */ {(16<<2)|3,{32,115,50}},
/* 20739 */ {(16<<2)|3,{32,115,97}},
/* 20740 */ {(16<<2)|3,{32,115,99}},
/* 20741 */ {(16<<2)|3,{32,115,101}},
/* 20742 */ {(16<<2)|3,{32,115,105}},
/* 20743 */ {(16<<2)|3,{32,115,111}},
/* 20744 */ {(16<<2)|3,{32,115,115}},
/* 20745 */ {(16<<2)|3,{32,115,116}},
/* 20746 */ {(11<<2)|2,{32,115,0}},
/* 20747 */ {(11<<2)|2,{32,115,0}},
/* 20748 */ {(11<<2)|2,{32,115,0}},
/* 20749 */ {(11<<2)|2,{32,115,0}},
/* 20750 */ {(11<<2)|2,{32,115,0}},
/* 20751 */ {(11<<2)|2,{32,115,0}},
/* 20752 */ {(11<<2)|2,{32,115,0}},
/* 20753 */ {(11<<2)|2,{32,115,0}},
/* 20754 */ {(11<<2)|2,{32,115,0}},
/* 20755 */ {(11<<2)|2,{32,115,0}},
/* 20756 */ {(11<<2)|2,{32,115,0}},
/* 20757 */ {(11<<2)|2,{32,115,0}},
/* 20758 */ {(11<<2)|2,{32,115,0}},
/* 20759 */ {(11<<2)|2,{32,115,0}},
/* 20760 */ {(11<<2)|2,{32,115,0}},
/* 20761 */ {(11<<2)|2,{32,115,0}},
/* 20762 */ {(11<<2)|2,{32,115,0}},
/* 20763 */ {(11<<2)|2,{32,115,0}},
/* 20764 */ {(11<<2)|2,{32,115,0}},
/* 20765 */ {(11<<2)|2,{32,115,0}},
/* 20766 */ {(11<<2)|2,{32,115,0}},
/* 20767 */ {(11<<2)|2,{32,115,0}},
/* 20768 */ {(16<<2)|3,{32,116,48}},
/* 20769 */ {(16<<2)|3,{32,116,49}},
/* 20770 */ {(16<<2)|3,{32,116,50}},
/* 20771 */ {(16<<2)|3,{32,116,97}},
/* 20772 */ {(16<<2)|3,{32,116,99}},
/* 20773 */ {(16<<2)|3,{32,116,101}},
/* 20774 */ {(16<<2)|3,{32,116,105}},
/* 20775 */ {(16<<2)|3,{32,116,111}},
/* 20776 */ {(16<<2)|3,{32,116,115}},
/* 20777 */ {(16<<2)|3,{32,116,116}},
/* 20778 */ {(11<<2)|2,{32,116,0}},
/* 20779 */ {(11<<2)|2,{32,116,0}},
/* 20780 */ {(11<<2)|2,{32,116,0}},
/* 20781 */ {(11<<2)|2,{32,116,0}},
/* 20782 */ {(11<<2)|2,{32,116,0}},
/* 20783 */ {(11<<2)|2,{32,116,0}},
/* 20784 */ {(11<<2)|2,{32,116,0}},
/* 20785 */ {(11<<2)|2,{32,116,0}},
/* 20786 */ {(11<<2)|2,{32,116,0}},
/* 20787 */ {(11<<2)|2,{32,116,0}},
/* 20788 */ {(11<<2)|2,{32,116,0}},
/* 20789 */ {(11<<2)|2,{32,116,0}},
/* 20790 */ {(11<<2)|2,{32,116,0}},
/* 20791 */ {(11<<2)|2,{32,116,0}},
/* 20792 */ {(11<<2)|2,{32,116,0}},
/* 20793 */ {(11<<2)|2,{32,116,0}},
/* 20794 */ {(11<<2)|2,{32,116,0}},
/* 20795 */ {(11<<2)|2,{32,116,0}},
/* 20796 */ {(11<<2)|2,{32,116,0}},
/* 20797 */ {(11<<2)|2,{32,116,0}},
/* 20798 */ {(11<<2)|2,{32,116,0}},
/* 20799 */ {(11<<2)|2,{32,116,0}},
/* 20800 */ {(12<<2)|2,{32,32,0}},
/* 20801 */ {(12<<2)|2,{32,32,0}},
/* 20802 */ {(12<<2)|2,{32,32,0}},
/* 20803 */ {(12<<2)|2,{32,32,0}},
/* 20804 */ {(12<<2)|2,{32,32,0}},
/* 20805 */ {(12<<2)|2,{32,32,0}},
/* 20806 */ {(12<<2)|2,{32,32,0}},
/* 20807 */ {(12<<2)|2,{32,32,0}},
/* 20808 */ {(12<<2)|2,{32,32,0}},
/* 20809 */ {(12<<2)|2,{32,32,0}},
/* 20810 */ {(12<<2)|2,{32,32,0}},
/* 20811 */ {(12<<2)|2,{32,32,0}},
/* 20812 */ {(12<<2)|2,{32,32,0}},
/* 20813 */ {(12<<2)|2,{32,32,0}},
/* 20814 */ {(12<<2)|2,{32,32,0}},
/* 20815 */ {(12<<2)|2,{32,32,0}},
/* 20816 */ {(12<<2)|2,{32,37,0}},
/* 20817 */ {(12<<2)|2,{32,37,0}},
/* 20818 */ {(12<<2)|2,{32,37,0}},
/* 20819 */ {(12<<2)|2,{32,37,0}},
/* 20820 */ {(12<<2)|2,{32,37,0}},
/* 20821 */ {(12<<2)|2,{32,37,0}},
/* 20822 */ {(12<<2)|2,{32,37,0}},
/* 20823 */ {(12<<2)|2,{32,37,0}},
/* 20824 */ {(12<<2)|2,{32,37,0}},
/* 20825 */ {(12<<2)|2,{32,37,0}},
/* 20826 */ {(12<<2)|2,{32,37,0}},
/* 20827 */ {(12<<2)|2,{32,37,0}},
/* 20828 */ {(12<<2)|2,{32,37,0}},
/* 20829 */ {(12<<2)|2,{32,37,0}},
/* 20830 */ {(12<<2)|2,{32,37,0}},
/* 20831 */ {(12<<2)|2,{32,37,0}},
/* 20832 */ {(12<<2)|2,{32,45,0}},
/* 20833 */ {(12<<2)|2,{32,45,0}},
/* 20834 */ {(12<<2)|2,{32,45,0}},
/* 20835 */ {(12<<2)|2,{32,45,0}},
/* 20836 */ {(12<<2)|2,{32,45,0}},
/* 20837 */ {(12<<2)|2,{32,45,0}},
/* 20838 */ {(12<<2)|2,{32,45,0}},
/* 20839 */ {(12<<2)|2,{32,45,0}},
/* 20840 */ {(12<<2)|2,{32,45,0}},
/* 20841 */ {(12<<2)|2,{32,45,0}},
/* 20842 */ {(12<<2)|2,{32,45,0}},
/* 20843 */ {(12<<2)|2,{32,45,0}},
/* 20844 */ {(12<<2)|2,{32,45,0}},
/* 20845 */ {(12<<2)|2,{32,45,0}},
/* 20846 */ {(12<<2)|2,{32,45,0}},
/* 20847 */ {(12<<2)|2,{32,45,0}},
/* 20848 */ {(12<<2)|2,{32,46,0}},
/* 20849 */ {(12<<2)|2,{32,46,0}},
/* 20850 */ {(12<<2)|2,{32,46,0}},
/* 20851 */ {(12<<2)|2,{32,46,0}},
/* 20852 */ {(12<<2)|2,{32,46,0}},
/* 20853 */ {(12<<2)|2,{32,46,0}},
/* 20854 */ {(12<<2)|2,{32,46,0}},
/* 20855 */ {(12<<2)|2,{32,46,0}},
/* 20856 */ {(12<<2)|2,{32,46,0}},
/* 20857 */ {(12<<2)|2,{32,46,0}},
/* 20858 */ {(12<<2)|2,{32,46,0}},
/* 20859 */ {(12<<2)|2,{32,46,0}},
/* 20860 */ {(12<<2)|2,{32,46,0}},
/* 20861 */ {(12<<2)|2,{32,46,0}},
/* 20862 */ {(12<<2)|2,{32,46,0}},
/* 20863 */ {(12<<2)|2,{32,46,0}},
/* 20864 */ {(12<<2)|2,{32,47,0}},
/* 20865 */ {(12<<2)|2,{32,47,0}},
/* 20866 */ {(12<<2)|2,{32,47,0}},
/* 20867 */ {(12<<2)|2,{32,47,0}},
/* 20868 */ {(12<<2)|2,{32,47,0}},
/* 20869 */ {(12<<2)|2,{32,47,0}},
/* 20870 */ {(12<<2)|2,{32,47,0}},
/* 20871 */ {(12<<2)|2,{32,47,0}},
/* 20872 */ {(12<<2)|2,{32,47,0}},
/* 20873 */ {(12<<2)|2,{32,47,0}},
/* 20874 */ {(12<<2)|2,{32,47,0}},
/* 20875 */ {(12<<2)|2,{32,47,0}},
/* 20876 */ {(12<<2)|2,{32,47,0}},
/* 20877 */ {(12<<2)|2,{32,47,0}},
/* 20878 */ {(12<<2)|2,{32,47,0}},
/* 20879 */ {(12<<2)|2,{32,47,0}},
/* 20880 */ {(12<<2)|2,{32,51,0}},
/* 20881 */ {(12<<2)|2,{32,51,0}},
/* 20882 */ {(12<<2)|2,{32,51,0}},
/* 20883 */ {(12<<2)|2,{32,51,0}},
/* 20884 */ {(12<<2)|2,{32,51,0}},
/* 20885 */ {(12<<2)|2,{32,51,0}},
/* 20886 */ {(12<<2)|2,{32,51,0}},
/* 20887 */ {(12<<2)|2,{32,51,0}},
/* 20888 */ {(12<<2)|2,{32,51,0}},
/* 20889 */ {(12<<2)|2,{32,51,0}},
/* 20890 */ {(12<<2)|2,{32,51,0}},
/* 20891 */ {(12<<2)|2,{32,51,0}},
/* 20892 */ {(12<<2)|2,{32,51,0}},
/* 20893 */ {(12<<2)|2,{32,51,0}},
/* 20894 */ {(12<<2)|2,{32,51,0}},
/* 20895 */ {(12<<2)|2,{32,51,0}},
/* 20896 */ {(12<<2)|2,{32,52,0}},
/* 20897 */ {(12<<2)|2,{32,52,0}},
/* 20898 */ {(12<<2)|2,{32,52,0}},
/* 20899 */ {(12<<2)|2,{32,52,0}},
/* 20900 */ {(12<<2)|2,{32,52,0}},
/* 20901 */ {(12<<2)|2,{32,52,0}},
/* 20902 */ {(12<<2)|2,{32,52,0}},
/* 20903 */ {(12<<2)|2,{32,52,0}},
/* 20904 */ {(12<<2)|2,{32,52,0}},
/* 20905 */ {(12<<2)|2,{32,52,0}},
/* 20906 */ {(12<<2)|2,{32,52,0}},
/* 20907 */ {(12<<2)|2,{32,52,0}},
/* 20908 */ {(12<<2)|2,{32,52,0}},
/* 20909 */ {(12<<2)|2,{32,52,0}},
/* 20910 */ {(12<<2)|2,{32,52,0}},
/* 20911 */ {(12<<2)|2,{32,52,0}},
/* 20912 */ {(12<<2)|2,{32,53,0}},
/* 20913 */ {(12<<2)|2,{32,53,0}},
/* 20914 */ {(12<<2)|2,{32,53,0}},
/* 20915 */ {(12<<2)|2,{32,53,0}},
/* 20916 */ {(12<<2)|2,{32,53,0}},
/* 20917 */ {(12<<2)|2,{32,53,0}},
/* 20918 */ {(12<<2)|2,{32,53,0}},
/* 20919 */ {(12<<2)|2,{32,53,0}},
/* 20920 */ {(12<<2)|2,{32,53,0}},
/* 20921 */ {(12<<2)|2,{32,53,0}},
/* 20922 */ {(12<<2)|2,{32,53,0}},
/* 20923 */ {(12<<2)|2,{32,53,0}},
/* 20924 */ {(12<<2)|2,{32,53,0}},
/* 20925 */ {(12<<2)|2,{32,53,0}},
/* 20926 */ {(12<<2)|2,{32,53,0}},
/* 20927 */ {(12<<2)|2,{32,53,0}},
/* 20928 */ {(12<<2)|2,{32,54,0}},
/* 20929 */ {(12<<2)|2,{32,54,0}},
/* 20930 */ {(12<<2)|2,{32,54,0}},
/* 20931 */ {(12<<2)|2,{32,54,0}},
/* 20932 */ {(12<<2)|2,{32,54,0}},
/* 20933 */ {(12<<2)|2,{32,54,0}},
/* 20934 */ {(12<<2)|2,{32,54,0}},
/* 20935 */ {(12<<2)|2,{32,54,0}},
/* 20936 */ {(12<<2)|2,{32,54,0}},
/* 20937 */ {(12<<2)|2,{32,54,0}},
/* 20938 */ {(12<<2)|2,{32,54,0}},
/* 20939 */ {(12<<2)|2,{32,54,0}},
/* 20940 */ {(12<<2)|2,{32,54,0}},
/* 20941 */ {(12<<2)|2,{32,54,0}},
/* 20942 */ {(12<<2)|2,{32,54,0}},
/* 20943 */ {(12<<2)|2,{32,54,0}},
/* 20944 */ {(12<<2)|2,{32,55,0}},
/* 20945 */ {(12<<2)|2,{32,55,0}},
/* 20946 */ {(12<<2)|2,{32,55,0}},
/* 20947 */ {(12<<2)|2,{32,55,0}},
/* 20948 */ {(12<<2)|2,{32,55,0}},
/* 20949 */ {(12<<2)|2,{32,55,0}},
/* 20950 */ {(12<<2)|2,{32,55,0}},
/* 20951 */ {(12<<2)|2,{32,55,0}},
/* 20952 */ {(12<<2)|2,{32,55,0}},
/* 20953 */ {(12<<2)|2,{32,55,0}},
/* 20954 */ {(12<<2)|2,{32,55,0}},
/* 20955 */ {(12<<2)|2,{32,55,0}},
/* 20956 */ {(12<<2)|2,{32,55,0}},
/* 20957 */ {(12<<2)|2,{32,55,0}},
/* 20958 */ {(12<<2)|2,{32,55,0}},
/* 20959 */ {(12<<2)|2,{32,55,0}},
/* 20960 */ {(12<<2)|2,{32,56,0}},
/* 20961 */ {(12<<2)|2,{32,56,0}},
/* 20962 */ {(12<<2)|2,{32,56,0}},
/* 20963 */ {(12<<2)|2,{32,56,0}},
/* 20964 */ {(12<<2)|2,{32,56,0}},
/* 20965 */ {(12<<2)|2,{32,56,0}},
/* 20966 */ {(12<<2)|2,{32,56,0}},
/* 20967 */ {(12<<2)|2,{32,56,0}},
/* 20968 */ {(12<<2)|2,{32,56,0}},
/* 20969 */ {(12<<2)|2,{32,56,0}},
/* 20970 */ {(12<<2)|2,{32,56,0}},
/* 20971 */ {(12<<2)|2,{32,56,0}},
/* 20972 */ {(12<<2)|2,{32,56,0}},
/* 20973 */ {(12<<2)|2,{32,56,0}},
/* 20974 */ {(12<<2)|2,{32,56,0}},
/* 20975 */ {(12<<2)|2,{32,56,0}},
/* 20976 */ {(12<<2)|2,{32,57,0}},
/* 20977 */ {(12<<2)|2,{32,57,0}},
/* 20978 */ {(12<<2)|2,{32,57,0}},
/* 20979 */ {(12<<2)|2,{32,57,0}},
/* 20980 */ {(12<<2)|2,{32,57,0}},
/* 20981 */ {(12<<2)|2,{32,57,0}},
/* 20982 */ {(12<<2)|2,{32,57,0}},
/* 20983 */ {(12<<2)|2,{32,57,0}},
/* 20984 */ {(12<<2)|2,{32,57,0}},
/* 20985 */ {(12<<2)|2,{32,57,0}},
/* 20986 */ {(12<<2)|2,{32,57,0}},
/* 20987 */ {(12<<2)|2,{32,57,0}},
/* 20988 */ {(12<<2)|2,{32,57,0}},
/* 20989 */ {(12<<2)|2,{32,57,0}},
/* 20990 */ {(12<<2)|2,{32,57,0}},
/* 20991 */ {(12<<2)|2,{32,57,0}},
/* 20992 */ {(12<<2)|2,{32,61,0}},
/* 20993 */ {(12<<2)|2,{32,61,0}},
/* 20994 */ {(12<<2)|2,{32,61,0}},
/* 20995 */ {(12<<2)|2,{32,61,0}},
/* 20996 */ {(12<<2)|2,{32,61,0}},
/* 20997 */ {(12<<2)|2,{32,61,0}},
/* 20998 */ {(12<<2)|2,{32,61,0}},
/* 20999 */ {(12<<2)|2,{32,61,0}},
/* 21000 */ {(12<<2)|2,{32,61,0}},
/* 21001 */ {(12<<2)|2,{32,61,0}},
/* 21002 */ {(12<<2)|2,{32,61,0}},
/* 21003 */ {(12<<2)|2,{32,61,0}},
/* 21004 */ {(12<<2)|2,{32,61,0}},
/* 21005 */ {(12<<2)|2,{32,61,0}},
/* 21006 */ {(12<<2)|2,{32,61,0}},
/* 21007 */ {(12<<2)|2,{32,61,0}},
/* 21008 */ {(12<<2)|2,{32,65,0}},
/* 21009 */ {(12<<2)|2,{32,65,0}},
/* 21010 */ {(12<<2)|2,{32,65,0}},
/* 21011 */ {(12<<2)|2,{32,65,0}},
/* 21012 */ {(12<<2)|2,{32,65,0}},
/* 21013 */ {(12<<2)|2,{32,65,0}},
/* 21014 */ {(12<<2)|2,{32,65,0}},
/* 21015 */ {(12<<2)|2,{32,65,0}},
/* 21016 */ {(12<<2)|2,{32,65,0}},
/* 21017 */ {(12<<2)|2,{32,65,0}},
/* 21018 */ {(12<<2)|2,{32,65,0}},
/* 21019 */ {(12<<2)|2,{32,65,0}},
/* 21020 */ {(12<<2)|2,{32,65,0}},
/* 21021 */ {(12<<2)|2,{32,65,0}},
/* 21022 */ {(12<<2)|2,{32,65,0}},
/* 21023 */ {(12<<2)|2,{32,65,0}},
/* 21024 */ {(12<<2)|2,{32,95,0}},
/* 21025 */ {(12<<2)|2,{32,95,0}},
/* 21026 */ {(12<<2)|2,{32,95,0}},
/* 21027 */ {(12<<2)|2,{32,95,0}},
/* 21028 */ {(12<<2)|2,{32,95,0}},
/* 21029 */ {(12<<2)|2,{32,95,0}},
/* 21030 */ {(12<<2)|2,{32,95,0}},
/* 21031 */ {(12<<2)|2,{32,95,0}},
/* 21032 */ {(12<<2)|2,{32,95,0}},
/* 21033 */ {(12<<2)|2,{32,95,0}},
/* 21034 */ {(12<<2)|2,{32,95,0}},
/* 21035 */ {(12<<2)|2,{32,95,0}},
/* 21036 */ {(12<<2)|2,{32,95,0}},
/* 21037 */ {(12<<2)|2,{32,95,0}},
/* 21038 */ {(12<<2)|2,{32,95,0}},
/* 21039 */ {(12<<2)|2,{32,95,0}},
/* 21040 */ {(12<<2)|2,{32,98,0}},
/* 21041 */ {(12<<2)|2,{32,98,0}},
/* 21042 */ {(12<<2)|2,{32,98,0}},
/* 21043 */ {(12<<2)|2,{32,98,0}},
/* 21044 */ {(12<<2)|2,{32,98,0}},
/* 21045 */ {(12<<2)|2,{32,98,0}},
/* 21046 */ {(12<<2)|2,{32,98,0}},
/* 21047 */ {(12<<2)|2,{32,98,0}},
/* 21048 */ {(12<<2)|2,{32,98,0}},
/* 21049 */ {(12<<2)|2,{32,98,0}},
/* 21050 */ {(12<<2)|2,{32,98,0}},
/* 21051 */ {(12<<2)|2,{32,98,0}},
/* 21052 */ {(12<<2)|2,{32,98,0}},
/* 21053 */ {(12<<2)|2,{32,98,0}},
/* 21054 */ {(12<<2)|2,{32,98,0}},
/* 21055 */ {(12<<2)|2,{32,98,0}},
/* 21056 */ {(12<<2)|2,{32,100,0}},
/* 21057 */ {(12<<2)|2,{32,100,0}},
/* 21058 */ {(12<<2)|2,{32,100,0}},
/* 21059 */ {(12<<2)|2,{32,100,0}},
/* 21060 */ {(12<<2)|2,{32,100,0}},
/* 21061 */ {(12<<2)|2,{32,100,0}},
/* 21062 */ {(12<<2)|2,{32,100,0}},
/* 21063 */ {(12<<2)|2,{32,100,0}},
/* 21064 */ {(12<<2)|2,{32,100,0}},
/* 21065 */ {(12<<2)|2,{32,100,0}},
/* 21066 */ {(12<<2)|2,{32,100,0}},
/* 21067 */ {(12<<2)|2,{32,100,0}},
/* 21068 */ {(12<<2)|2,{32,100,0}},
/* 21069 */ {(12<<2)|2,{32,100,0}},
/* 21070 */ {(12<<2)|2,{32,100,0}},
/* 21071 */ {(12<<2)|2,{32,100,0}},
/* 21072 */ {(12<<2)|2,{32,102,0}},
/* 21073 */ {(12<<2)|2,{32,102,0}},
/* 21074 */ {(12<<2)|2,{32,102,0}},
/* 21075 */ {(12<<2)|2,{32,102,0}},
/* 21076 */ {(12<<2)|2,{32,102,0}},
/* 21077 */ {(12<<2)|2,{32,102,0}},
/* 21078 */ {(12<<2)|2,{32,102,0}},
/* 21079 */ {(12<<2)|2,{32,102,0}},
/* 21080 */ {(12<<2)|2,{32,102,0}},
/* 21081 */ {(12<<2)|2,{32,102,0}},
/* 21082 */ {(12<<2)|2,{32,102,0}},
/* 21083 */ {(12<<2)|2,{32,102,0}},
/* 21084 */ {(12<<2)|2,{32,102,0}},
/* 21085 */ {(12<<2)|2,{32,102,0}},
/* 21086 */ {(12<<2)|2,{32,102,0}},
/* 21087 */ {(12<<2)|2,{32,102,0}},
/* 21088 */ {(12<<2)|2,{32,103,0}},
/* 21089 */ {(12<<2)|2,{32,103,0}},
/* 21090 */ {(12<<2)|2,{32,103,0}},
/* 21091 */ {(12<<2)|2,{32,103,0}},
/* 21092 */ {(12<<2)|2,{32,103,0}},
/* 21093 */ {(12<<2)|2,{32,103,0}},
/* 21094 */ {(12<<2)|2,{32,103,0}},
/* 21095 */ {(12<<2)|2,{32,103,0}},
/* 21096 */ {(12<<2)|2,{32,103,0}},
/* 21097 */ {(12<<2)|2,{32,103,0}},
/* 21098 */ {(12<<2)|2,{32,103,0}},
/* 21099 */ {(12<<2)|2,{32,103,0}},
/* 21100 */ {(12<<2)|2,{32,103,0}},
/* 21101 */ {(12<<2)|2,{32,103,0}},
/* 21102 */ {(12<<2)|2,{32,103,0}},
/* 21103 */ {(12<<2)|2,{32,103,0}},
/* 21104 */ {(12<<2)|2,{32,104,0}},
/* 21105 */ {(12<<2)|2,{32,104,0}},
/* 21106 */ {(12<<2)|2,{32,104,0}},
/* 21107 */ {(12<<2)|2,{32,104,0}},
/* 21108 */ {(12<<2)|2,{32,104,0}},
/* 21109 */ {(12<<2)|2,{32,104,0}},
/* 21110 */ {(12<<2)|2,{32,104,0}},
/* 21111 */ {(12<<2)|2,{32,104,0}},
/* 21112 */ {(12<<2)|2,{32,104,0}},
/* 21113 */ {(12<<2)|2,{32,104,0}},
/* 21114 */ {(12<<2)|2,{32,104,0}},
/* 21115 */ {(12<<2)|2,{32,104,0}},
/* 21116 */ {(12<<2)|2,{32,104,0}},
/* 21117 */ {(12<<2)|2,{32,104,0}},
/* 21118 */ {(12<<2)|2,{32,104,0}},
/* 21119 */ {(12<<2)|2,{32,104,0}},
/* 21120 */ {(12<<2)|2,{32,108,0}},
/* 21121 */ {(12<<2)|2,{32,108,0}},
/* 21122 */ {(12<<2)|2,{32,108,0}},
/* 21123 */ {(12<<2)|2,{32,108,0}},
/* 21124 */ {(12<<2)|2,{32,108,0}},
/* 21125 */ {(12<<2)|2,{32,108,0}},
/* 21126 */ {(12<<2)|2,{32,108,0}},
/* 21127 */ {(12<<2)|2,{32,108,0}},
/* 21128 */ {(12<<2)|2,{32,108,0}},
/* 21129 */ {(12<<2)|2,{32,108,0}},
/* 21130 */ {(12<<2)|2,{32,108,0}},
/* 21131 */ {(12<<2)|2,{32,108,0}},
/* 21132 */ {(12<<2)|2,{32,108,0}},
/* 21133 */ {(12<<2)|2,{32,108,0}},
/* 21134 */ {(12<<2)|2,{32,108,0}},
/* 21135 */ {(12<<2)|2,{32,108,0}},
/* 21136 */ {(12<<2)|2,{32,109,0}},
/* 21137 */ {(12<<2)|2,{32,109,0}},
/* 21138 */ {(12<<2)|2,{32,109,0}},
/* 21139 */ {(12<<2)|2,{32,109,0}},
/* 21140 */ {(12<<2)|2,{32,109,0}},
/* 21141 */ {(12<<2)|2,{32,109,0}},
/* 21142 */ {(12<<2)|2,{32,109,0}},
/* 21143 */ {(12<<2)|2,{32,109,0}},
/* 21144 */ {(12<<2)|2,{32,109,0}},
/* 21145 */ {(12<<2)|2,{32,109,0}},
/* 21146 */ {(12<<2)|2,{32,109,0}},
/* 21147 */ {(12<<2)|2,{32,109,0}},
/* 21148 */ {(12<<2)|2,{32,109,0}},
/* 21149 */ {(12<<2)|2,{32,109,0}},
/* 21150 */ {(12<<2)|2,{32,109,0}},
/* 21151 */ {(12<<2)|2,{32,109,0}},
/* 21152 */ {(12<<2)|2,{32,110,0}},
/* 21153 */ {(12<<2)|2,{32,110,0}},
/* 21154 */ {(12<<2)|2,{32,110,0}},
/* 21155 */ {(12<<2)|2,{32,110,0}},
/* 21156 */ {(12<<2)|2,{32,110,0}},
/* 21157 */ {(12<<2)|2,{32,110,0}},
/* 21158 */ {(12<<2)|2,{32,110,0}},
/* 21159 */ {(12<<2)|2,{32,110,0}},
/* 21160 */ {(12<<2)|2,{32,110,0}},
/* 21161 */ {(12<<2)|2,{32,110,0}},
/* 21162 */ {(12<<2)|2,{32,110,0}},
/* 21163 */ {(12<<2)|2,{32,110,0}},
/* 21164 */ {(12<<2)|2,{32,110,0}},
/* 21165 */ {(12<<2)|2,{32,110,0}},
/* 21166 */ {(12<<2)|2,{32,110,0}},
/* 21167 */ {(12<<2)|2,{32,110,0}},
/* 21168 */ {(12<<2)|2,{32,112,0}},
/* 21169 */ {(12<<2)|2,{32,112,0}},
/* 21170 */ {(12<<2)|2,{32,112,0}},
/* 21171 */ {(12<<2)|2,{32,112,0}},
/* 21172 */ {(12<<2)|2,{32,112,0}},
/* 21173 */ {(12<<2)|2,{32,112,0}},
/* 21174 */ {(12<<2)|2,{32,112,0}},
/* 21175 */ {(12<<2)|2,{32,112,0}},
/* 21176 */ {(12<<2)|2,{32,112,0}},
/* 21177 */ {(12<<2)|2,{32,112,0}},
/* 21178 */ {(12<<2)|2,{32,112,0}},
/* 21179 */ {(12<<2)|2,{32,112,0}},
/* 21180 */ {(12<<2)|2,{32,112,0}},
/* 21181 */ {(12<<2)|2,{32,112,0}},
/* 21182 */ {(12<<2)|2,{32,112,0}},
/* 21183 */ {(12<<2)|2,{32,112,0}},
/* 21184 */ {(12<<2)|2,{32,114,0}},
/* 21185 */ {(12<<2)|2,{32,114,0}},
/* 21186 */ {(12<<2)|2,{32,114,0}},
/* 21187 */ {(12<<2)|2,{32,114,0}},
/* 21188 */ {(12<<2)|2,{32,114,0}},
/* 21189 */ {(12<<2)|2,{32,114,0}},
/* 21190 */ {(12<<2)|2,{32,114,0}},
/* 21191 */ {(12<<2)|2,{32,114,0}},
/* 21192 */ {(12<<2)|2,{32,114,0}},
/* 21193 */ {(12<<2)|2,{32,114,0}},
/* 21194 */ {(12<<2)|2,{32,114,0}},
/* 21195 */ {(12<<2)|2,{32,114,0}},
/* 21196 */ {(12<<2)|2,{32,114,0}},
/* 21197 */ {(12<<2)|2,{32,114,0}},
/* 21198 */ {(12<<2)|2,{32,114,0}},
/* 21199 */ {(12<<2)|2,{32,114,0}},
/* 21200 */ {(12<<2)|2,{32,117,0}},
/* 21201 */ {(12<<2)|2,{32,117,0}},
/* 21202 */ {(12<<2)|2,{32,117,0}},
/* 21203 */ {(12<<2)|2,{32,117,0}},
/* 21204 */ {(12<<2)|2,{32,117,0}},
/* 21205 */ {(12<<2)|2,{32,117,0}},
/* 21206 */ {(12<<2)|2,{32,117,0}},
/* 21207 */ {(12<<2)|2,{32,117,0}},
/* 21208 */ {(12<<2)|2,{32,117,0}},
/* 21209 */ {(12<<2)|2,{32,117,0}},
/* 21210 */ {(12<<2)|2,{32,117,0}},
/* 21211 */ {(12<<2)|2,{32,117,0}},
/* 21212 */ {(12<<2)|2,{32,117,0}},
/* 21213 */ {(12<<2)|2,{32,117,0}},
/* 21214 */ {(12<<2)|2,{32,117,0}},
/* 21215 */ {(12<<2)|2,{32,117,0}},
/* 21216 */ {(13<<2)|2,{32,58,0}},
/* 21217 */ {(13<<2)|2,{32,58,0}},
/* 21218 */ {(13<<2)|2,{32,58,0}},
/* 21219 */ {(13<<2)|2,{32,58,0}},
/* 21220 */ {(13<<2)|2,{32,58,0}},
/* 21221 */ {(13<<2)|2,{32,58,0}},
/* 21222 */ {(13<<2)|2,{32,58,0}},
/* 21223 */ {(13<<2)|2,{32,58,0}},
/* 21224 */ {(13<<2)|2,{32,66,0}},
/* 21225 */ {(13<<2)|2,{32,66,0}},
/* 21226 */ {(13<<2)|2,{32,66,0}},
/* 21227 */ {(13<<2)|2,{32,66,0}},
/* 21228 */ {(13<<2)|2,{32,66,0}},
/* 21229 */ {(13<<2)|2,{32,66,0}},
/* 21230 */ {(13<<2)|2,{32,66,0}},
/* 21231 */ {(13<<2)|2,{32,66,0}},
/* 21232 */ {(13<<2)|2,{32,67,0}},
/* 21233 */ {(13<<2)|2,{32,67,0}},
/* 21234 */ {(13<<2)|2,{32,67,0}},
/* 21235 */ {(13<<2)|2,{32,67,0}},
/* 21236 */ {(13<<2)|2,{32,67,0}},
/* 21237 */ {(13<<2)|2,{32,67,0}},
/* 21238 */ {(13<<2)|2,{32,67,0}},
/* 21239 */ {(13<<2)|2,{32,67,0}},
/* 21240 */ {(13<<2)|2,{32,68,0}},
/* 21241 */ {(13<<2)|2,{32,68,0}},
/* 21242 */ {(13<<2)|2,{32,68,0}},
/* 21243 */ {(13<<2)|2,{32,68,0}},
/* 21244 */ {(13<<2)|2,{32,68,0}},
/* 21245 */ {(13<<2)|2,{32,68,0}},
/* 21246 */ {(13<<2)|2,{32,68,0}},
/* 21247 */ {(13<<2)|2,{32,68,0}},
/* 21248 */ {(13<<2)|2,{32,69,0}},
/* 21249 */ {(13<<2)|2,{32,69,0}},
/* 21250 */ {(13<<2)|2,{32,69,0}},
/* 21251 */ {(13<<2)|2,{32,69,0}},
/* 21252 */ {(13<<2)|2,{32,69,0}},
/* 21253 */ {(13<<2)|2,{32,69,0}},
/* 21254 */ {(13<<2)|2,{32,69,0}},
/* 21255 */ {(13<<2)|2,{32,69,0}},
/* 21256 */ {(13<<2)|2,{32,70,0}},
/* 21257 */ {(13<<2)|2,{32,70,0}},
/* 21258 */ {(13<<2)|2,{32,70,0}},
/* 21259 */ {(13<<2)|2,{32,70,0}},
/* 21260 */ {(13<<2)|2,{32,70,0}},
/* 21261 */ {(13<<2)|2,{32,70,0}},
/* 21262 */ {(13<<2)|2,{32,70,0}},
/* 21263 */ {(13<<2)|2,{32,70,0}},
/* 21264 */ {(13<<2)|2,{32,71,0}},
/* 21265 */ {(13<<2)|2,{32,71,0}},
/* 21266 */ {(13<<2)|2,{32,71,0}},
/* 21267 */ {(13<<2)|2,{32,71,0}},
/* 21268 */ {(13<<2)|2,{32,71,0}},
/* 21269 */ {(13<<2)|2,{32,71,0}},
/* 21270 */ {(13<<2)|2,{32,71,0}},
/* 21271 */ {(13<<2)|2,{32,71,0}},
/* 21272 */ {(13<<2)|2,{32,72,0}},
/* 21273 */ {(13<<2)|2,{32,72,0}},
/* 21274 */ {(13<<2)|2,{32,72,0}},
/* 21275 */ {(13<<2)|2,{32,72,0}},
/* 21276 */ {(13<<2)|2,{32,72,0}},
/* 21277 */ {(13<<2)|2,{32,72,0}},
/* 21278 */ {(13<<2)|2,{32,72,0}},
/* 21279 */ {(13<<2)|2,{32,72,0}},
/* 21280 */ {(13<<2)|2,{32,73,0}},
/* 21281 */ {(13<<2)|2,{32,73,0}},
/* 21282 */ {(13<<2)|2,{32,73,0}},
/* 21283 */ {(13<<2)|2,{32,73,0}},
/* 21284 */ {(13<<2)|2,{32,73,0}},
/* 21285 */ {(13<<2)|2,{32,73,0}},
/* 21286 */ {(13<<2)|2,{32,73,0}},
/* 21287 */ {(13<<2)|2,{32,73,0}},
/* 21288 */ {(13<<2)|2,{32,74,0}},
/* 21289 */ {(13<<2)|2,{32,74,0}},
/* 21290 */ {(13<<2)|2,{32,74,0}},
/* 21291 */ {(13<<2)|2,{32,74,0}},
/* 21292 */ {(13<<2)|2,{32,74,0}},
/* 21293 */ {(13<<2)|2,{32,74,0}},
/* 21294 */ {(13<<2)|2,{32,74,0}},
/* 21295 */ {(13<<2)|2,{32,74,0}},
/* 21296 */ {(13<<2)|2,{32,75,0}},
/* 21297 */ {(13<<2)|2,{32,75,0}},
/* 21298 */ {(13<<2)|2,{32,75,0}},
/* 21299 */ {(13<<2)|2,{32,75,0}},
/* 21300 */ {(13<<2)|2,{32,75,0}},
/* 21301 */ {(13<<2)|2,{32,75,0}},
/* 21302 */ {(13<<2)|2,{32,75,0}},
/* 21303 */ {(13<<2)|2,{32,75,0}},
/* 21304 */ {(13<<2)|2,{32,76,0}},
/* 21305 */ {(13<<2)|2,{32,76,0}},
/* 21306 */ {(13<<2)|2,{32,76,0}},
/* 21307 */ {(13<<2)|2,{32,76,0}},
/* 21308 */ {(13<<2)|2,{32,76,0}},
/* 21309 */ {(13<<2)|2,{32,76,0}},
/* 21310 */ {(13<<2)|2,{32,76,0}},
/* 21311 */ {(13<<2)|2,{32,76,0}},
/* 21312 */ {(13<<2)|2,{32,77,0}},
/* 21313 */ {(13<<2)|2,{32,77,0}},
/* 21314 */ {(13<<2)|2,{32,77,0}},
/* 21315 */ {(13<<2)|2,{32,77,0}},
/* 21316 */ {(13<<2)|2,{32,77,0}},
/* 21317 */ {(13<<2)|2,{32,77,0}},
/* 21318 */ {(13<<2)|2,{32,77,0}},
/* 21319 */ {(13<<2)|2,{32,77,0}},
/* 21320 */ {(13<<2)|2,{32,78,0}},
/* 21321 */ {(13<<2)|2,{32,78,0}},
/* 21322 */ {(13<<2)|2,{32,78,0}},
/* 21323 */ {(13<<2)|2,{32,78,0}},
/* 21324 */ {(13<<2)|2,{32,78,0}},
/* 21325 */ {(13<<2)|2,{32,78,0}},
/* 21326 */ {(13<<2)|2,{32,78,0}},
/* 21327 */ {(13<<2)|2,{32,78,0}},
/* 21328 */ {(13<<2)|2,{32,79,0}},
/* 21329 */ {(13<<2)|2,{32,79,0}},
/* 21330 */ {(13<<2)|2,{32,79,0}},
/* 21331 */ {(13<<2)|2,{32,79,0}},
/* 21332 */ {(13<<2)|2,{32,79,0}},
/* 21333 */ {(13<<2)|2,{32,79,0}},
/* 21334 */ {(13<<2)|2,{32,79,0}},
/* 21335 */ {(13<<2)|2,{32,79,0}},
/* 21336 */ {(13<<2)|2,{32,80,0}},
/* 21337 */ {(13<<2)|2,{32,80,0}},
/* 21338 */ {(13<<2)|2,{32,80,0}},
/* 21339 */ {(13<<2)|2,{32,80,0}},
/* 21340 */ {(13<<2)|2,{32,80,0}},
/* 21341 */ {(13<<2)|2,{32,80,0}},
/* 21342 */ {(13<<2)|2,{32,80,0}},
/* 21343 */ {(13<<2)|2,{32,80,0}},
/* 21344 */ {(13<<2)|2,{32,81,0}},
/* 21345 */ {(13<<2)|2,{32,81,0}},
/* 21346 */ {(13<<2)|2,{32,81,0}},
/* 21347 */ {(13<<2)|2,{32,81,0}},
/* 21348 */ {(13<<2)|2,{32,81,0}},
/* 21349 */ {(13<<2)|2,{32,81,0}},
/* 21350 */ {(13<<2)|2,{32,81,0}},
/* 21351 */ {(13<<2)|2,{32,81,0}},
/* 21352 */ {(13<<2)|2,{32,82,0}},
/* 21353 */ {(13<<2)|2,{32,82,0}},
/* 21354 */ {(13<<2)|2,{32,82,0}},
/* 21355 */ {(13<<2)|2,{32,82,0}},
/* 21356 */ {(13<<2)|2,{32,82,0}},
/* 21357 */ {(13<<2)|2,{32,82,0}},
/* 21358 */ {(13<<2)|2,{32,82,0}},
/* 21359 */ {(13<<2)|2,{32,82,0}},
/* 21360 */ {(13<<2)|2,{32,83,0}},
/* 21361 */ {(13<<2)|2,{32,83,0}},
/* 21362 */ {(13<<2)|2,{32,83,0}},
/* 21363 */ {(13<<2)|2,{32,83,0}},
/* 21364 */ {(13<<2)|2,{32,83,0}},
/* 21365 */ {(13<<2)|2,{32,83,0}},
/* 21366 */ {(13<<2)|2,{32,83,0}},
/* 21367 */ {(13<<2)|2,{32,83,0}},
/* 21368 */ {(13<<2)|2,{32,84,0}},
/* 21369 */ {(13<<2)|2,{32,84,0}},
/* 21370 */ {(13<<2)|2,{32,84,0}},
/* 21371 */ {(13<<2)|2,{32,84,0}},
/* 21372 */ {(13<<2)|2,{32,84,0}},
/* 21373 */ {(13<<2)|2,{32,84,0}},
/* 21374 */ {(13<<2)|2,{32,84,0}},
/* 21375 */ {(13<<2)|2,{32,84,0}},
/* 21376 */ {(13<<2)|2,{32,85,0}},
/* 21377 */ {(13<<2)|2,{32,85,0}},
/* 21378 */ {(13<<2)|2,{32,85,0}},
/* 21379 */ {(13<<2)|2,{32,85,0}},
/* 21380 */ {(13<<2)|2,{32,85,0}},
/* 21381 */ {(13<<2)|2,{32,85,0}},
/* 21382 */ {(13<<2)|2,{32,85,0}},
/* 21383 */ {(13<<2)|2,{32,85,0}},
/* 21384 */ {(13<<2)|2,{32,86,0}},
/* 21385 */ {(13<<2)|2,{32,86,0}},
/* 21386 */ {(13<<2)|2,{32,86,0}},
/* 21387 */ {(13<<2)|2,{32,86,0}},
/* 21388 */ {(13<<2)|2,{32,86,0}},
/* 21389 */ {(13<<2)|2,{32,86,0}},
/* 21390 */ {(13<<2)|2,{32,86,0}},
/* 21391 */ {(13<<2)|2,{32,86,0}},
/* 21392 */ {(13<<2)|2,{32,87,0}},
/* 21393 */ {(13<<2)|2,{32,87,0}},
/* 21394 */ {(13<<2)|2,{32,87,0}},
/* 21395 */ {(13<<2)|2,{32,87,0}},
/* 21396 */ {(13<<2)|2,{32,87,0}},
/* 21397 */ {(13<<2)|2,{32,87,0}},
/* 21398 */ {(13<<2)|2,{32,87,0}},
/* 21399 */ {(13<<2)|2,{32,87,0}},
/* 21400 */ {(13<<2)|2,{32,89,0}},
/* 21401 */ {(13<<2)|2,{32,89,0}},
/* 21402 */ {(13<<2)|2,{32,89,0}},
/* 21403 */ {(13<<2)|2,{32,89,0}},
/* 21404 */ {(13<<2)|2,{32,89,0}},
/* 21405 */ {(13<<2)|2,{32,89,0}},
/* 21406 */ {(13<<2)|2,{32,89,0}},
/* 21407 */ {(13<<2)|2,{32,89,0}},
/* 21408 */ {(13<<2)|2,{32,106,0}},
/* 21409 */ {(13<<2)|2,{32,106,0}},
/* 21410 */ {(13<<2)|2,{32,106,0}},
/* 21411 */ {(13<<2)|2,{32,106,0}},
/* 21412 */ {(13<<2)|2,{32,106,0}},
/* 21413 */ {(13<<2)|2,{32,106,0}},
/* 21414 */ {(13<<2)|2,{32,106,0}},
/* 21415 */ {(13<<2)|2,{32,106,0}},
/* 21416 */ {(13<<2)|2,{32,107,0}},
/* 21417 */ {(13<<2)|2,{32,107,0}},
/* 21418 */ {(13<<2)|2,{32,107,0}},
/* 21419 */ {(13<<2)|2,{32,107,0}},
/* 21420 */ {(13<<2)|2,{32,107,0}},
/* 21421 */ {(13<<2)|2,{32,107,0}},
/* 21422 */ {(13<<2)|2,{32,107,0}},
/* 21423 */ {(13<<2)|2,{32,107,0}},
/* 21424 */ {(13<<2)|2,{32,113,0}},
/* 21425 */ {(13<<2)|2,{32,113,0}},
/* 21426 */ {(13<<2)|2,{32,113,0}},
/* 21427 */ {(13<<2)|2,{32,113,0}},
/* 21428 */ {(13<<2)|2,{32,113,0}},
/* 21429 */ {(13<<2)|2,{32,113,0}},
/* 21430 */ {(13<<2)|2,{32,113,0}},
/* 21431 */ {(13<<2)|2,{32,113,0}},
/* 21432 */ {(13<<2)|2,{32,118,0}},
/* 21433 */ {(13<<2)|2,{32,118,0}},
/* 21434 */ {(13<<2)|2,{32,118,0}},
/* 21435 */ {(13<<2)|2,{32,118,0}},
/* 21436 */ {(13<<2)|2,{32,118,0}},
/* 21437 */ {(13<<2)|2,{32,118,0}},
/* 21438 */ {(13<<2)|2,{32,118,0}},
/* 21439 */ {(13<<2)|2,{32,118,0}},
/* 21440 */ {(13<<2)|2,{32,119,0}},
/* 21441 */ {(13<<2)|2,{32,119,0}},
/* 21442 */ {(13<<2)|2,{32,119,0}},
/* 21443 */ {(13<<2)|2,{32,119,0}},
/* 21444 */ {(13<<2)|2,{32,119,0}},
/* 21445 */ {(13<<2)|2,{32,119,0}},
/* 21446 */ {(13<<2)|2,{32,119,0}},
/* 21447 */ {(13<<2)|2,{32,119,0}},
/* 21448 */ {(13<<2)|2,{32,120,0}},
/* 21449 */ {(13<<2)|2,{32,120,0}},
/* 21450 */ {(13<<2)|2,{32,120,0}},
/* 21451 */ {(13<<2)|2,{32,120,0}},
/* 21452 */ {(13<<2)|2,{32,120,0}},
/* 21453 */ {(13<<2)|2,{32,120,0}},
/* 21454 */ {(13<<2)|2,{32,120,0}},
/* 21455 */ {(13<<2)|2,{32,120,0}},
/* 21456 */ {(13<<2)|2,{32,121,0}},
/* 21457 */ {(13<<2)|2,{32,121,0}},
/* 21458 */ {(13<<2)|2,{32,121,0}},
/* 21459 */ {(13<<2)|2,{32,121,0}},
/* 21460 */ {(13<<2)|2,{32,121,0}},
/* 21461 */ {(13<<2)|2,{32,121,0}},
/* 21462 */ {(13<<2)|2,{32,121,0}},
/* 21463 */ {(13<<2)|2,{32,121,0}},
/* 21464 */ {(13<<2)|2,{32,122,0}},
/* 21465 */ {(13<<2)|2,{32,122,0}},
/* 21466 */ {(13<<2)|2,{32,122,0}},
/* 21467 */ {(13<<2)|2,{32,122,0}},
/* 21468 */ {(13<<2)|2,{32,122,0}},
/* 21469 */ {(13<<2)|2,{32,122,0}},
/* 21470 */ {(13<<2)|2,{32,122,0}},
/* 21471 */ {(13<<2)|2,{32,122,0}},
/* 21472 */ {(14<<2)|2,{32,38,0}},
/* 21473 */ {(14<<2)|2,{32,38,0}},
/* 21474 */ {(14<<2)|2,{32,38,0}},
/* 21475 */ {(14<<2)|2,{32,38,0}},
/* 21476 */ {(14<<2)|2,{32,42,0}},
/* 21477 */ {(14<<2)|2,{32,42,0}},
/* 21478 */ {(14<<2)|2,{32,42,0}},
/* 21479 */ {(14<<2)|2,{32,42,0}},
/* 21480 */ {(14<<2)|2,{32,44,0}},
/* 21481 */ {(14<<2)|2,{32,44,0}},
/* 21482 */ {(14<<2)|2,{32,44,0}},
/* 21483 */ {(14<<2)|2,{32,44,0}},
/* 21484 */ {(14<<2)|2,{32,59,0}},
/* 21485 */ {(14<<2)|2,{32,59,0}},
/* 21486 */ {(14<<2)|2,{32,59,0}},
/* 21487 */ {(14<<2)|2,{32,59,0}},
/* 21488 */ {(14<<2)|2,{32,88,0}},
/* 21489 */ {(14<<2)|2,{32,88,0}},
/* 21490 */ {(14<<2)|2,{32,88,0}},
/* 21491 */ {(14<<2)|2,{32,88,0}},
/* 21492 */ {(14<<2)|2,{32,90,0}},
/* 21493 */ {(14<<2)|2,{32,90,0}},
/* 21494 */ {(14<<2)|2,{32,90,0}},
/* 21495 */ {(14<<2)|2,{32,90,0}},
/* 21496 */ {(16<<2)|2,{32,33,0}},
/* 21497 */ {(16<<2)|2,{32,34,0}},
/* 21498 */ {(16<<2)|2,{32,40,0}},
/* 21499 */ {(16<<2)|2,{32,41,0}},
/* 21500 */ {(16<<2)|2,{32,63,0}},
/* 21501 */ {(6<<2)|1,{32,0,0}},
/* 21502 */ {(6<<2)|1,{32,0,0}},
/* 21503 */ {(6<<2)|1,{32,0,0}},
/* 21504 */ {(16<<2)|3,{37,48,48}},
/* 21505 */ {(16<<2)|3,{37,48,49}},
/* 21506 */ {(16<<2)|3,{37,48,50}},
/* 21507 */ {(16<<2)|3,{37,48,97}},
/* 21508 */ {(16<<2)|3,{37,48,99}},
/* 21509 */ {(16<<2)|3,{37,48,101}},
/* 21510 */ {(16<<2)|3,{37,48,105}},
/* 21511 */ {(16<<2)|3,{37,48,111}},
/* 21512 */ {(16<<2)|3,{37,48,115}},
/* 21513 */ {(16<<2)|3,{37,48,116}},
/* 21514 */ {(11<<2)|2,{37,48,0}},
/* 21515 */ {(11<<2)|2,{37,48,0}},
/* 21516 */ {(11<<2)|2,{37,48,0}},
/* 21517 */ {(11<<2)|2,{37,48,0}},
/* 21518 */ {(11<<2)|2,{37,48,0}},
/* 21519 */ {(11<<2)|2,{37,48,0}},
/* 21520 */ {(11<<2)|2,{37,48,0}},
/* 21521 */ {(11<<2)|2,{37,48,0}},
/* 21522 */ {(11<<2)|2,{37,48,0}},
/* 21523 */ {(11<<2)|2,{37,48,0}},
/* 21524 */ {(11<<2)|2,{37,48,0}},
/* 21525 */ {(11<<2)|2,{37,48,0}},
/* 21526 */ {(11<<2)|2,{37,48,0}},
/* 21527 */ {(11<<2)|2,{37,48,0}},
/* 21528 */ {(11<<2)|2,{37,48,0}},
/* 21529 */ {(11<<2)|2,{37,48,0}},
/* 21530 */ {(11<<2)|2,{37,48,0}},
/* 21531 */ {(11<<2)|2,{37,48,0}},
/* 21532 */ {(11<<2)|2,{37,48,0}},
/* 21533 */ {(11<<2)|2,{37,48,0}},
/* 21534 */ {(11<<2)|2,{37,48,0}},
/* 21535 */ {(11<<2)|2,{37,48,0}},
/* 21536 */ {(16<<2)|3,{37,49,48}},
/* 21537 */ {(16<<2)|3,{37,49,49}},
/* 21538 */ {(16<<2)|3,{37,49,50}},
/* 21539 */ {(16<<2)|3,{37,49,97}},
/* 21540 */ {(16<<2)|3,{37,49,99}},
/* 21541 */ {(16<<2)|3,{37,49,101}},
/* 21542 */ {(16<<2)|3,{37,49,105}},
/* 21543 */ {(16<<2)|3,{37,49,111}},
/* 21544 */ {(16<<2)|3,{37,49,115}},
/* 21545 */ {(16<<2)|3,{37,49,116}},
/* 21546 */ {(11<<2)|2,{37,49,0}},
/* 21547 */ {(11<<2)|2,{37,49,0}},
/* 21548 */ {(11<<2)|2,{37,49,0}},
/* 21549 */ {(11<<2)|2,{37,49,0}},
/* 21550 */ {(11<<2)|2,{37,49,0}},
/* 21551 */ {(11<<2)|2,{37,49,0}},
/* 21552 */ {(11<<2)|2,{37,49,0}},
/* 21553 */ {(11<<2)|2,{37,49,0}},
/* 21554 */ {(11<<2)|2,{37,49,0}},
/* 21555 */ {(11<<2)|2,{37,49,0}},
/* 21556 */ {(11<<2)|2,{37,49,0}},
/* 21557 */ {(11<<2)|2,{37,49,0}},
/* 21558 */ {(11<<2)|2,{37,49,0}},
/* 21559 */ {(11<<2)|2,{37,49,0}},
/* 21560 */ {(11<<2)|2,{37,49,0}},
/* 21561 */ {(11<<2)|2,{37,49,0}},
/* 21562 */ {(11<<2)|2,{37,49,0}},
/* 21563 */ {(11<<2)|2,{37,49,0}},
/* 21564 */ {(11<<2)|2,{37,49,0}},
/* 21565 */ {(11<<2)|2,{37,49,0}},
/* 21566 */ {(11<<2)|2,{37,49,0}},
/* 21567 */ {(11<<2)|2,{37,49,0}},
/* 21568 */ {(16<<2)|3,{37,50,48}},
/* 21569 */ {(16<<2)|3,{37,50,49}},
/* 21570 */ {(16<<2)|3,{37,50,50}},
/* 21571 */ {(16<<2)|3,{37,50,97}},
/* 21572 */ {(16<<2)|3,{37,50,99}},
/* 21573 */ {(16<<2)|3,{37,50,101}},
/* 21574 */ {(16<<2)|3,{37,50,105}},
/* 21575 */ {(16<<2)|3,{37,50,111}},
/* 21576 */ {(16<<2)|3,{37,50,115}},
/* 21577 */ {(16<<2)|3,{37,50,116}},
/* 21578 */ {(11<<2)|2,{37,50,0}},
/* 21579 */ {(11<<2)|2,{37,50,0}},
/* 21580 */ {(11<<2)|2,{37,50,0}},
/* 21581 */ {(11<<2)|2,{37,50,0}},
/* 21582 */ {(11<<2)|2,{37,50,0}},
/* 21583 */ {(11<<2)|2,{37,50,0}},
/* 21584 */ {(11<<2)|2,{37,50,0}},
/* 21585 */ {(11<<2)|2,{37,50,0}},
/* 21586 */ {(11<<2)|2,{37,50,0}},
/* 21587 */ {(11<<2)|2,{37,50,0}},
/* 21588 */ {(11<<2)|2,{37,50,0}},
/* 21589 */ {(11<<2)|2,{37,50,0}},
/* 21590 */ {(11<<2)|2,{37,50,0}},
/* 21591 */ {(11<<2)|2,{37,50,0}},
/* 21592 */ {(11<<2)|2,{37,50,0}},
/* 21593 */ {(11<<2)|2,{37,50,0}},
/* 21594 */ {(11<<2)|2,{37,50,0}},
/* 21595 */ {(11<<2)|2,{37,50,0}},
/* 21596 */ {(11<<2)|2,{37,50,0}},
/* 21597 */ {(11<<2)|2,{37,50,0}},
/* 21598 */ {(11<<2)|2,{37,50,0}},
/* 21599 */ {(11<<2)|2,{37,50,0}},
/* 21600 */ {(16<<2)|3,{37,97,48}},
/* 21601 */ {(16<<2)|3,{37,97,49}},
/* 21602 */ {(16<<2)|3,{37,97,50}},
/* 21603 */ {(16<<2)|3,{37,97,97}},
/* 21604 */ {(16<<2)|3,{37,97,99}},
/* 21605 */ {(16<<2)|3,{37,97,101}},
/* 21606 */ {(16<<2)|3,{37,97,105}},
/* 21607 */ {(16<<2)|3,{37,97,111}},
/* 21608 */ {(16<<2)|3,{37,97,115}},
/* 21609 */ {(16<<2)|3,{37,97,116}},
/* 21610 */ {(11<<2)|2,{37,97,0}},
/* 21611 */ {(11<<2)|2,{37,97,0}},
/* 21612 */ {(11<<2)|2,{37,97,0}},
/* 21613 */ {(11<<2)|2,{37,97,0}},
/* 21614 */ {(11<<2)|2,{37,97,0}},
/* 21615 */ {(11<<2)|2,{37,97,0}},
/* 21616 */ {(11<<2)|2,{37,97,0}},
/* 21617 */ {(11<<2)|2,{37,97,0}},
/* 21618 */ {(11<<2)|2,{37,97,0}},
/* 21619 */ {(11<<2)|2,{37,97,0}},
/* 21620 */ {(11<<2)|2,{37,97,0}},
/* 21621 */ {(11<<2)|2,{37,97,0}},
/* 21622 */ {(11<<2)|2,{37,97,0}},
/* 21623 */ {(11<<2)|2,{37,97,0}},
/* 21624 */ {(11<<2)|2,{37,97,0}},
/* 21625 */ {(11<<2)|2,{37,97,0}},
/* 21626 */ {(11<<2)|2,{37,97,0}},
/* 21627 */ {(11<<2)|2,{37,97,0}},
/* 21628 */ {(11<<2)|2,{37,97,0}},
/* 21629 */ {(11<<2)|2,{37,97,0}},
/* 21630 */ {(11<<2)|2,{37,97,0}},
/* 21631 */ {(11<<2)|2,{37,97,0}},
/* 21632 */ {(16<<2)|3,{37,99,48}},
/* 21633 */ {(16<<2)|3,{37,99,49}},
/* 21634 */ {(16<<2)|3,{37,99,50}},
/* 21635 */ {(16<<2)|3,{37,99,97}},
/* 21636 */ {(16<<2)|3,{37,99,99}},
/* 21637 */ {(16<<2)|3,{37,99,101}},
/* 21638 */ {(16<<2)|3,{37,99,105}},
/* 21639 */ {(16<<2)|3,{37,99,111}},
/* 21640 */ {(16<<2)|3,{37,99,115}},
/* 21641 */ {(16<<2)|3,{37,99,116}},
/* 21642 */ {(11<<2)|2,{37,99,0}},
/* 21643 */ {(11<<2)|2,{37,99,0}},
/* 21644 */ {(11<<2)|2,{37,99,0}},
/* 21645 */ {(11<<2)|2,{37,99,0}},
/* 21646 */ {(11<<2)|2,{37,99,0}},
/* 21647 */ {(11<<2)|2,{37,99,0}},
/* 21648 */ {(11<<2)|2,{37,99,0}},
/* 21649 */ {(11<<2)|2,{37,99,0}},
/* 21650 */ {(11<<2)|2,{37,99,0}},
/* 21651 */ {(11<<2)|2,{37,99,0}},
/* 21652 */ {(11<<2)|2,{37,99,0}},
/* 21653 */ {(11<<2)|2,{37,99,0}},
/* 21654 */ {(11<<2)|2,{37,99,0}},
/* 21655 */ {(11<<2)|2,{37,99,0}},
/* 21656 */ {(11<<2)|2,{37,99,0}},
/* 21657 */ {(11<<2)|2,{37,99,0}},
/* 21658 */ {(11<<2)|2,{37,99,0}},
/* 21659 */ {(11<<2)|2,{37,99,0}},
/* 21660 */ {(11<<2)|2,{37,99,0}},
/* 21661 */ {(11<<2)|2,{37,99,0}},
/* 21662 */ {(11<<2)|2,{37,99,0}},
/* 21663 */ {(11<<2)|2,{37,99,0}},
/* 21664 */ {(16<<2)|3,{37,101,48}},
/* 21665 */ {(16<<2)|3,{37,101,49}},
/* 21666 */ {(16<<2)|3,{37,101,50}},
/* 21667 */ {(16<<2)|3,{37,101,97}},
/* 21668 */ {(16<<2)|3,{37,101,99}},
/* 21669 */ {(16<<2)|3,{37,101,101}},
/* 21670 */ {(16<<2)|3,{37,101,105}},
/* 21671 */ {(16<<2)|3,{37,101,111}},
/* 21672 */ {(16<<2)|3,{37,101,115}},
/* 21673 */ {(16<<2)|3,{37,101,116}},
/* 21674 */ {(11<<2)|2,{37,101,0}},
/* 21675 */ {(11<<2)|2,{37,101,0}},
/* 21676 */ {(11<<2)|2,{37,101,0}},
/* 21677 */ {(11<<2)|2,{37,101,0}},
/* 21678 */ {(11<<2)|2,{37,101,0}},
/* 21679 */ {(11<<2)|2,{37,101,0}},
/* 21680 */ {(11<<2)|2,{37,101,0}},
/* 21681 */ {(11<<2)|2,{37,101,0}},
/* 21682 */ {(11<<2)|2,{37,101,0}},
/* 21683 */ {(11<<2)|2,{37,101,0}},
/* 21684 */ {(11<<2)|2,{37,101,0}},
/* 21685 */ {(11<<2)|2,{37,101,0}},
/* 21686 */ {(11<<2)|2,{37,101,0}},
/* 21687 */ {(11<<2)|2,{37,101,0}},
/* 21688 */ {(11<<2)|2,{37,101,0}},
/* 21689 */ {(11<<2)|2,{37,101,0}},
/* 21690 */ {(11<<2)|2,{37,101,0}},
/* 21691 */ {(11<<2)|2,{37,101,0}},
/* 21692 */ {(11<<2)|2,{37,101,0}},
/* 21693 */ {(11<<2)|2,{37,101,0}},
/* 21694 */ {(11<<2)|2,{37,101,0}},
/* 21695 */ {(11<<2)|2,{37,101,0}},
/* 21696 */ {(16<<2)|3,{37,105,48}},
/* 21697 */ {(16<<2)|3,{37,105,49}},
/* 21698 */ {(16<<2)|3,{37,105,50}},
/* 21699 */ {(16<<2)|3,{37,105,97}},
/* 21700 */ {(16<<2)|3,{37,105,99}},
/* 21701 */ {(16<<2)|3,{37,105,101}},
/* 21702 */ {(16<<2)|3,{37,105,105}},
/* 21703 */ {(16<<2)|3,{37,105,111}},
/* 21704 */ {(16<<2)|3,{37,105,115}},
/* 21705 */ {(16<<2)|3,{37,105,116}},
/* 21706 */ {(11<<2)|2,{37,105,0}},
/* 21707 */ {(11<<2)|2,{37,105,0}},
/* 21708 */ {(11<<2)|2,{37,105,0}},
/* 21709 */ {(11<<2)|2,{37,105,0}},
/* 21710 */ {(11<<2)|2,{37,105,0}},
/* 21711 */ {(11<<2)|2,{37,105,0}},
/* 21712 */ {(11<<2)|2,{37,105,0}},
/* 21713 */ {(11<<2)|2,{37,105,0}},
/* 21714 */ {(11<<2)|2,{37,105,0}},
/* 21715 */ {(11<<2)|2,{37,105,0}},
/* 21716 */ {(11<<2)|2,{37,105,0}},
/* 21717 */ {(11<<2)|2,{37,105,0}},
/* 21718 */ {(11<<2)|2,{37,105,0}},
/* 21719 */ {(11<<2)|2,{37,105,0}},
/* 21720 */ {(11<<2)|2,{37,105,0}},
/* 21721 */ {(11<<2)|2,{37,105,0}},
/* 21722 */ {(11<<2)|2,{37,105,0}},
/* 21723 */ {(11<<2)|2,{37,105,0}},
/* 21724 */ {(11<<2)|2,{37,105,0}},
/* 21725 */ {(11<<2)|2,{37,105,0}},
/* 21726 */ {(11<<2)|2,{37,105,0}},
/* 21727 */ {(11<<2)|2,{37,105,0}},
/* 21728 */ {(16<<2)|3,{37,111,48}},
/* 21729 */ {(16<<2)|3,{37,111,49}},
/* 21730 */ {(16<<2)|3,{37,111,50}},
/* 21731 */ {(16<<2)|3,{37,111,97}},
/* 21732 */ {(16<<2)|3,{37,111,99}},
/* 21733 */ {(16<<2)|3,{37,111,101}},
/* 21734 */ {(16<<2)|3,{37,111,105}},
/* 21735 */ {(16<<2)|3,{37,111,111}},
/* 21736 */ {(16<<2)|3,{37,111,115}},
/* 21737 */ {(16<<2)|3,{37,111,116}},
/* 21738 */ {(11<<2)|2,{37,111,0}},
/* 21739 */ {(11<<2)|2,{37,111,0}},
/* 21740 */ {(11<<2)|2,{37,111,0}},
/* 21741 */ {(11<<2)|2,{37,111,0}},
/* 21742 */ {(11<<2)|2,{37,111,0}},
/* 21743 */ {(11<<2)|2,{37,111,0}},
/* 21744 */ {(11<<2)|2,{37,111,0}},
/* 21745 */ {(11<<2)|2,{37,111,0}},
/* 21746 */ {(11<<2)|2,{37,111,0}},
/* 21747 */ {(11<<2)|2,{37,111,0}},
/* 21748 */ {(11<<2)|2,{37,111,0}},
/* 21749 */ {(11<<2)|2,{37,111,0}},
/* 21750 */ {(11<<2)|2,{37,111,0}},
/* 21751 */ {(11<<2)|2,{37,111,0}},
/* 21752 */ {(11<<2)|2,{37,111,0}},
/* 21753 */ {(11<<2)|2,{37,111,0}},
/* 21754 */ {(11<<2)|2,{37,111,0}},
/* 21755 */ {(11<<2)|2,{37,111,0}},
/* 21756 */ {(11<<2)|2,{37,111,0}},
/* 21757 */ {(11<<2)|2,{37,111,0}},
/* 21758 */ {(11<<2)|2,{37,111,0}},
/* 21759 */ {(11<<2)|2,{37,111,0}},
/* 21760 */ {(16<<2)|3,{37,115,48}},
/* 21761 */ {(16<<2)|3,{37,115,49}},
/* 21762 */ {(16<<2)|3,{37,115,50}},
/* 21763 */ {(16<<2)|3,{37,115,97}},
/* 21764 */ {(16<<2)|3,{37,115,99}},
/* 21765 */ {(16<<2)|3,{37,115,101}},
/* 21766 */ {(16<<2)|3,{37,115,105}},
/* 21767 */ {(16<<2)|3,{37,115,111}},
/* 21768 */ {(16<<2)|3,{37,115,115}},
/* 21769 */ {(16<<2)|3,{37,115,116}},
/* 21770 */ {(11<<2)|2,{37,115,0}},
/* 21771 */ {(11<<2)|2,{37,115,0}},
/* 21772 */ {(11<<2)|2,{37,115,0}},
/* 21773 */ {(11<<2)|2,{37,115,0}},
/* 21774 */ {(11<<2)|2,{37,115,0}},
/* 21775 */ {(11<<2)|2,{37,115,0}},
/* 21776 */ {(11<<2)|2,{37,115,0}},
/* 21777 */ {(11<<2)|2,{37,115,0}},
/* 21778 */ {(11<<2)|2,{37,115,0}},
/* 21779 */ {(11<<2)|2,{37,115,0}},
/* 21780 */ {(11<<2)|2,{37,115,0}},
/* 21781 */ {(11<<2)|2,{37,115,0}},
/* 21782 */ {(11<<2)|2,{37,115,0}},
/* 21783 */ {(11<<2)|2,{37,115,0}},
/* 21784 */ {(11<<2)|2,{37,115,0}},
/* 21785 */ {(11<<2)|2,{37,115,0}},
/* 21786 */ {(11<<2)|2,{37,115,0}},
/* 21787 */ {(11<<2)|2,{37,115,0}},
/* 21788 */ {(11<<2)|2,{37,115,0}},
/* 21789 */ {(11<<2)|2,{37,115,0}},
/* 21790 */ {(11<<2)|2,{37,115,0}},
/* 21791 */ {(11<<2)|2,{37,115,0}},
/* 21792 */ {(16<<2)|3,{37,116,48}},
/* 21793 */ {(16<<2)|3,{37,116,49}},
/* 21794 */ {(16<<2)|3,{37,116,50}},
/* 21795 */ {(16<<2)|3,{37,116,97}},
/* 21796 */ {(16<<2)|3,{37,116,99}},
/* 21797 */ {(16<<2)|3,{37,116,101}},
/* 21798 */ {(16<<2)|3,{37,116,105}},
/* 21799 */ {(16<<2)|3,{37,116,111}},
/* 21800 */ {(16<<2)|3,{37,116,115}},
/* 21801 */ {(16<<2)|3,{37,116,116}},
/* 21802 */ {(11<<2)|2,{37,116,0}},
/* 21803 */ {(11<<2)|2,{37,116,0}},
/* 21804 */ {(11<<2)|2,{37,116,0}},
/* 21805 */ {(11<<2)|2,{37,116,0}},
/* 21806 */ {(11<<2)|2,{37,116,0}},
/* 21807 */ {(11<<2)|2,{37,116,0}},
/* 21808 */ {(11<<2)|2,{37,116,0}},
/* 21809 */ {(11<<2)|2,{37,116,0}},
/* 21810 */ {(11<<2)|2,{37,116,0}},
/* 21811 */ {(11<<2)|2,{37,116,0}},
/* 21812 */ {(11<<2)|2,{37,116,0}},
/* 21813 */ {(11<<2)|2,{37,116,0}},
/* 21814 */ {(11<<2)|2,{37,116,0}},
/* 21815 */ {(11<<2)|2,{37,116,0}},
/* 21816 */ {(11<<2)|2,{37,116,0}},
/* 21817 */ {(11<<2)|2,{37,116,0}},
/* 21818 */ {(11<<2)|2,{37,116,0}},
/* 21819 */ {(11<<2)|2,{37,116,0}},
/* 21820 */ {(11<<2)|2,{37,116,0}},
/* 21821 */ {(11<<2)|2,{37,116,0}},
/* 21822 */ {(11<<2)|2,{37,116,0}},
/* 21823 */ {(11<<2)|2,{37,116,0}},
/* 21824 */ {(12<<2)|2,{37,32,0}},
/* 21825 */ {(12<<2)|2,{37,32,0}},
/* 21826 */ {(12<<2)|2,{37,32,0}},
/* 21827 */ {(12<<2)|2,{37,32,0}},
/* 21828 */ {(12<<2)|2,{37,32,0}},
/* 21829 */ {(12<<2)|2,{37,32,0}},
/* 21830 */ {(12<<2)|2,{37,32,0}},
/* 21831 */ {(12<<2)|2,{37,32,0}},
/* 21832 */ {(12<<2)|2,{37,32,0}},
/* 21833 */ {(12<<2)|2,{37,32,0}},
/* 21834 */ {(12<<2)|2,{37,32,0}},
/* 21835 */ {(12<<2)|2,{37,32,0}},
/* 21836 */ {(12<<2)|2,{37,32,0}},
/* 21837 */ {(12<<2)|2,{37,32,0}},
/* 21838 */ {(12<<2)|2,{37,32,0}},
/* 21839 */ {(12<<2)|2,{37,32,0}},
/* 21840 */ {(12<<2)|2,{37,37,0}},
/* 21841 */ {(12<<2)|2,{37,37,0}},
/* 21842 */ {(12<<2)|2,{37,37,0}},
/* 21843 */ {(12<<2)|2,{37,37,0}},
/* 21844 */ {(12<<2)|2,{37,37,0}},
/* 21845 */ {(12<<2)|2,{37,37,0}},
/* 21846 */ {(12<<2)|2,{37,37,0}},
/* 21847 */ {(12<<2)|2,{37,37,0}},
/* 21848 */ {(12<<2)|2,{37,37,0}},
/* 21849 */ {(12<<2)|2,{37,37,0}},
/* 21850 */ {(12<<2)|2,{37,37,0}},
/* 21851 */ {(12<<2)|2,{37,37,0}},
/* 21852 */ {(12<<2)|2,{37,37,0}},
/* 21853 */ {(12<<2)|2,{37,37,0}},
/* 21854 */ {(12<<2)|2,{37,37,0}},
/* 21855 */ {(12<<2)|2,{37,37,0}},
/* 21856 */ {(12<<2)|2,{37,45,0}},
/* 21857 */ {(12<<2)|2,{37,45,0}},
/* 21858 */ {(12<<2)|2,{37,45,0}},
/* 21859 */ {(12<<2)|2,{37,45,0}},
/* 21860 */ {(12<<2)|2,{37,45,0}},
/* 21861 */ {(12<<2)|2,{37,45,0}},
/* 21862 */ {(12<<2)|2,{37,45,0}},
/* 21863 */ {(12<<2)|2,{37,45,0}},
/* 21864 */ {(12<<2)|2,{37,45,0}},
/* 21865 */ {(12<<2)|2,{37,45,0}},
/* 21866 */ {(12<<2)|2,{37,45,0}},
/* 21867 */ {(12<<2)|2,{37,45,0}},
/* 21868 */ {(12<<2)|2,{37,45,0}},
/* 21869 */ {(12<<2)|2,{37,45,0}},
/* 21870 */ {(12<<2)|2,{37,45,0}},
/* 21871 */ {(12<<2)|2,{37,45,0}},
/* 21872 */ {(12<<2)|2,{37,46,0}},
/* 21873 */ {(12<<2)|2,{37,46,0}},
/* 21874 */ {(12<<2)|2,{37,46,0}},
/* 21875 */ {(12<<2)|2,{37,46,0}},
/* 21876 */ {(12<<2)|2,{37,46,0}},
/* 21877 */ {(12<<2)|2,{37,46,0}},
/* 21878 */ {(12<<2)|2,{37,46,0}},
/* 21879 */ {(12<<2)|2,{37,46,0}},
/* 21880 */ {(12<<2)|2,{37,46,0}},
/* 21881 */ {(12<<2)|2,{37,46,0}},
/* 21882 */ {(12<<2)|2,{37,46,0}},
/* 21883 */ {(12<<2)|2,{37,46,0}},
/* 21884 */ {(12<<2)|2,{37,46,0}},
/* 21885 */ {(12<<2)|2,{37,46,0}},
/* 21886 */ {(12<<2)|2,{37,46,0}},
/* 21887 */ {(12<<2)|2,{37,46,0}},
/* 21888 */ {(12<<2)|2,{37,47,0}},
/* 21889 */ {(12<<2)|2,{37,47,0}},
/* 21890 */ {(12<<2)|2,{37,47,0}},
/* 21891 */ {(12<<2)|2,{37,47,0}},
/* 21892 */ {(12<<2)|2,{37,47,0}},
/* 21893 */ {(12<<2)|2,{37,47,0}},
/* 21894 */ {(12<<2)|2,{37,47,0}},
/* 21895 */ {(12<<2)|2,{37,47,0}},
/* 21896 */ {(12<<2)|2,{37,47,0}},
/* 21897 */ {(12<<2)|2,{37,47,0}},
/* 21898 */ {(12<<2)|2,{37,47,0}},
/* 21899 */ {(12<<2)|2,{37,47,0}},
/* 21900 */ {(12<<2)|2,{37,47,0}},
/* 21901 */ {(12<<2)|2,{37,47,0}},
/* 21902 */ {(12<<2)|2,{37,47,0}},
/* 21903 */ {(12<<2)|2,{37,47,0}},
/* 21904 */ {(12<<2)|2,{37,51,0}},
/* 21905 */ {(12<<2)|2,{37,51,0}},
/* 21906 */ {(12<<2)|2,{37,51,0}},
/* 21907 */ {(12<<2)|2,{37,51,0}},
/* 21908 */ {(12<<2)|2,{37,51,0}},
/* 21909 */ {(12<<2)|2,{37,51,0}},
/* 21910 */ {(12<<2)|2,{37,51,0}},
/* 21911 */ {(12<<2)|2,{37,51,0}},
/* 21912 */ {(12<<2)|2,{37,51,0}},
/* 21913 */ {(12<<2)|2,{37,51,0}},
/* 21914 */ {(12<<2)|2,{37,51,0}},
/* 21915 */ {(12<<2)|2,{37,51,0}},
/* 21916 */ {(12<<2)|2,{37,51,0}},
/* 21917 */ {(12<<2)|2,{37,51,0}},
/* 21918 */ {(12<<2)|2,{37,51,0}},
/* 21919 */ {(12<<2)|2,{37,51,0}},
/* 21920 */ {(12<<2)|2,{37,52,0}},
/* 21921 */ {(12<<2)|2,{37,52,0}},
/* 21922 */ {(12<<2)|2,{37,52,0}},
/* 21923 */ {(12<<2)|2,{37,52,0}},
/* 21924 */ {(12<<2)|2,{37,52,0}},
/* 21925 */ {(12<<2)|2,{37,52,0}},
/* 21926 */ {(12<<2)|2,{37,52,0}},
/* 21927 */ {(12<<2)|2,{37,52,0}},
/* 21928 */ {(12<<2)|2,{37,52,0}},
/* 21929 */ {(12<<2)|2,{37,52,0}},
/* 21930 */ {(12<<2)|2,{37,52,0}},
/* 21931 */ {(12<<2)|2,{37,52,0}},
/* 21932 */ {(12<<2)|2,{37,52,0}},
/* 21933 */ {(12<<2)|2,{37,52,0}},
/* 21934 */ {(12<<2)|2,{37,52,0}},
/* 21935 */ {(12<<2)|2,{37,52,0}},
/* 21936 */ {(12<<2)|2,{37,53,0}},
/* 21937 */ {(12<<2)|2,{37,53,0}},
/* 21938 */ {(12<<2)|2,{37,53,0}},
/* 21939 */ {(12<<2)|2,{37,53,0}},
/* 21940 */ {(12<<2)|2,{37,53,0}},
/* 21941 */ {(12<<2)|2,{37,53,0}},
/* 21942 */ {(12<<2)|2,{37,53,0}},
/* 21943 */ {(12<<2)|2,{37,53,0}},
/* 21944 */ {(12<<2)|2,{37,53,0}},
/* 21945 */ {(12<<2)|2,{37,53,0}},
/* 21946 */ {(12<<2)|2,{37,53,0}},
/* 21947 */ {(12<<2)|2,{37,53,0}},
/* 21948 */ {(12<<2)|2,{37,53,0}},
/* 21949 */ {(12<<2)|2,{37,53,0}},
/* 21950 */ {(12<<2)|2,{37,53,0}},
/* 21951 */ {(12<<2)|2,{37,53,0}},
/* 21952 */ {(12<<2)|2,{37,54,0}},
/* 21953 */ {(12<<2)|2,{37,54,0}},
/* 21954 */ {(12<<2)|2,{37,54,0}},
/* 21955 */ {(12<<2)|2,{37,54,0}},
/* 21956 */ {(12<<2)|2,{37,54,0}},
/* 21957 */ {(12<<2)|2,{37,54,0}},
/* 21958 */ {(12<<2)|2,{37,54,0}},
/* 21959 */ {(12<<2)|2,{37,54,0}},
/* 21960 */ {(12<<2)|2,{37,54,0}},
/* 21961 */ {(12<<2)|2,{37,54,0}},
/* 21962 */ {(12<<2)|2,{37,54,0}},
/* 21963 */ {(12<<2)|2,{37,54,0}},
/* 21964 */ {(12<<2)|2,{37,54,0}},
/* 21965 */ {(12<<2)|2,{37,54,0}},
/* 21966 */ {(12<<2)|2,{37,54,0}},
/* 21967 */ {(12<<2)|2,{37,54,0}},
/* 21968 */ {(12<<2)|2,{37,55,0}},
/* 21969 */ {(12<<2)|2,{37,55,0}},
/* 21970 */ {(12<<2)|2,{37,55,0}},
/* 21971 */ {(12<<2)|2,{37,55,0}},
/* 21972 */ {(12<<2)|2,{37,55,0}},
/* 21973 */ {(12<<2)|2,{37,55,0}},
/* 21974 */ {(12<<2)|2,{37,55,0}},
/* 21975 */ {(12<<2)|2,{37,55,0}},
/* 21976 */ {(12<<2)|2,{37,55,0}},
/* 21977 */ {(12<<2)|2,{37,55,0}},
/* 21978 */ {(12<<2)|2,{37,55,0}},
/* 21979 */ {(12<<2)|2,{37,55,0}},
/* 21980 */ {(12<<2)|2,{37,55,0}},
/* 21981 */ {(12<<2)|2,{37,55,0}},
/* 21982 */ {(12<<2)|2,{37,55,0}},
/* 21983 */ {(12<<2)|2,{37,55,0}},
/* 21984 */ {(12<<2)|2,{37,56,0}},
/* 21985 */ {(12<<2)|2,{37,56,0}},
/* 21986 */ {(12<<2)|2,{37,56,0}},
/* 21987 */ {(12<<2)|2,{37,56,0}},
/* 21988 */ {(12<<2)|2,{37,56,0}},
/* 21989 */ {(12<<2)|2,{37,56,0}},
/* 21990 */ {(12<<2)|2,{37,56,0}},
/* 21991 */ {(12<<2)|2,{37,56,0}},
/* 21992 */ {(12<<2)|2,{37,56,0}},
/* 21993 */ {(12<<2)|2,{37,56,0}},
/* 21994 */ {(12<<2)|2,{37,56,0}},
/* 21995 */ {(12<<2)|2,{37,56,0}},
/* 21996 */ {(12<<2)|2,{37,56,0}},
/* 21997 */ {(12<<2)|2,{37,56,0}},
/* 21998 */ {(12<<2)|2,{37,56,0}},
/* 21999 */ {(12<<2)|2,{37,56,0}},
/* 22000 */ {(12<<2)|2,{37,57,0}},
/* 22001 */ {(12<<2)|2,{37,57,0}},
/* 22002 */ {(12<<2)|2,{37,57,0}},
/* 22003 */ {(12<<2)|2,{37,57,0}},
/* 22004 */ {(12<<2)|2,{37,57,0}},
/* 22005 */ {(12<<2)|2,{37,57,0}},
/* 22006 */ {(12<<2)|2,{37,57,0}},
/* 22007 */ {(12<<2)|2,{37,57,0}},
/* 22008 */ {(12<<2)|2,{37,57,0}},
/* 22009 */ {(12<<2)|2,{37,57,0}},
/* 22010 */ {(12<<2)|2,{37,57,0}},
/* 22011 */ {(12<<2)|2,{37,57,0}},
/* 22012 */ {(12<<2)|2,{37,57,0}},
/* 22013 */ {(12<<2)|2,{37,57,0}},
/* 22014 */ {(12<<2)|2,{37,57,0}},
/* 22015 */ {(12<<2)|2,{37,57,0}},
/* 22016 */ {(12<<2)|2,{37,61,0}},
/* 22017 */ {(12<<2)|2,{37,61,0}},
/* 22018 */ {(12<<2)|2,{37,61,0}},
/* 22019 */ {(12<<2)|2,{37,61,0}},
/* 22020 */ {(12<<2)|2,{37,61,0}},
/* 22021 */ {(12<<2)|2,{37,61,0}},
/* 22022 */ {(12<<2)|2,{37,61,0}},
/* 22023 */ {(12<<2)|2,{37,61,0}},
/* 22024 */ {(12<<2)|2,{37,61,0}},
/* 22025 */ {(12<<2)|2,{37,61,0}},
/* 22026 */ {(12<<2)|2,{37,61,0}},
/* 22027 */ {(12<<2)|2,{37,61,0}},
/* 22028 */ {(12<<2)|2,{37,61,0}},
/* 22029 */ {(12<<2)|2,{37,61,0}},
/* 22030 */ {(12<<2)|2,{37,61,0}},
/* 22031 */ {(12<<2)|2,{37,61,0}},
/* 22032 */ {(12<<2)|2,{37,65,0}},
/* 22033 */ {(12<<2)|2,{37,65,0}},
/* 22034 */ {(12<<2)|2,{37,65,0}},
/* 22035 */ {(12<<2)|2,{37,65,0}},
/* 22036 */ {(12<<2)|2,{37,65,0}},
/* 22037 */ {(12<<2)|2,{37,65,0}},
/* 22038 */ {(12<<2)|2,{37,65,0}},
/* 22039 */ {(12<<2)|2,{37,65,0}},
/* 22040 */ {(12<<2)|2,{37,65,0}},
/* 22041 */ {(12<<2)|2,{37,65,0}},
/* 22042 */ {(12<<2)|2,{37,65,0}},
/* 22043 */ {(12<<2)|2,{37,65,0}},
/* 22044 */ {(12<<2)|2,{37,65,0}},
/* 22045 */ {(12<<2)|2,{37,65,0}},
/* 22046 */ {(12<<2)|2,{37,65,0}},
/* 22047 */ {(12<<2)|2,{37,65,0}},
/* 22048 */ {(12<<2)|2,{37,95,0}},
/* 22049 */ {(12<<2)|2,{37,95,0}},
/* 22050 */ {(12<<2)|2,{37,95,0}},
/* 22051 */ {(12<<2)|2,{37,95,0}},
/* 22052 */ {(12<<2)|2,{37,95,0}},
/* 22053 */ {(12<<2)|2,{37,95,0}},
/* 22054 */ {(12<<2)|2,{37,95,0}},
/* 22055 */ {(12<<2)|2,{37,95,0}},
/* 22056 */ {(12<<2)|2,{37,95,0}},
/* 22057 */ {(12<<2)|2,{37,95,0}},
/* 22058 */ {(12<<2)|2,{37,95,0}},
/* 22059 */ {(12<<2)|2,{37,95,0}},
/* 22060 */ {(12<<2)|2,{37,95,0}},
/* 22061 */ {(12<<2)|2,{37,95,0}},
/* 22062 */ {(12<<2)|2,{37,95,0}},
/* 22063 */ {(12<<2)|2,{37,95,0}},
/* 22064 */ {(12<<2)|2,{37,98,0}},
/* 22065 */ {(12<<2)|2,{37,98,0}},
/* 22066 */ {(12<<2)|2,{37,98,0}},
/* 22067 */ {(12<<2)|2,{37,98,0}},
/* 22068 */ {(12<<2)|2,{37,98,0}},
/* 22069 */ {(12<<2)|2,{37,98,0}},
/* 22070 */ {(12<<2)|2,{37,98,0}},
/* 22071 */ {(12<<2)|2,{37,98,0}},
/* 22072 */ {(12<<2)|2,{37,98,0}},
/* 22073 */ {(12<<2)|2,{37,98,0}},
/* 22074 */ {(12<<2)|2,{37,98,0}},
/* 22075 */ {(12<<2)|2,{37,98,0}},
/* 22076 */ {(12<<2)|2,{37,98,0}},
/* 22077 */ {(12<<2)|2,{37,98,0}},
/* 22078 */ {(12<<2)|2,{37,98,0}},
/* 22079 */ {(12<<2)|2,{37,98,0}},
/* 22080 */ {(12<<2)|2,{37,100,0}},
/* 22081 */ {(12<<2)|2,{37,100,0}},
/* 22082 */ {(12<<2)|2,{37,100,0}},
/* 22083 */ {(12<<2)|2,{37,100,0}},
/* 22084 */ {(12<<2)|2,{37,100,0}},
/* 22085 */ {(12<<2)|2,{37,100,0}},
/* 22086 */ {(12<<2)|2,{37,100,0}},
/* 22087 */ {(12<<2)|2,{37,100,0}},
/* 22088 */ {(12<<2)|2,{37,100,0}},
/* 22089 */ {(12<<2)|2,{37,100,0}},
/* 22090 */ {(12<<2)|2,{37,100,0}},
/* 22091 */ {(12<<2)|2,{37,100,0}},
/* 22092 */ {(12<<2)|2,{37,100,0}},
/* 22093 */ {(12<<2)|2,{37,100,0}},
/* 22094 */ {(12<<2)|2,{37,100,0}},
/* 22095 */ {(12<<2)|2,{37,100,0}},
/* 22096 */ {(12<<2)|2,{37,102,0}},
/* 22097 */ {(12<<2)|2,{37,102,0}},
/* 22098 */ {(12<<2)|2,{37,102,0}},
/* 22099 */ {(12<<2)|2,{37,102,0}},
/* 22100 */ {(12<<2)|2,{37,102,0}},
/* 22101 */ {(12<<2)|2,{37,102,0}},
/* 22102 */ {(12<<2)|2,{37,102,0}},
/* 22103 */ {(12<<2)|2,{37,102,0}},
/* 22104 */ {(12<<2)|2,{37,102,0}},
/* 22105 */ {(12<<2)|2,{37,102,0}},
/* 22106 */ {(12<<2)|2,{37,102,0}},
/* 22107 */ {(12<<2)|2,{37,102,0}},
/* 22108 */ {(12<<2)|2,{37,102,0}},
/* 22109 */ {(12<<2)|2,{37,102,0}},
/* 22110 */ {(12<<2)|2,{37,102,0}},
/* 22111 */ {(12<<2)|2,{37,102,0}},
/* 22112 */ {(12<<2)|2,{37,103,0}},
/* 22113 */ {(12<<2)|2,{37,103,0}},
/* 22114 */ {(12<<2)|2,{37,103,0}},
/* 22115 */ {(12<<2)|2,{37,103,0}},
/* 22116 */ {(12<<2)|2,{37,103,0}},
/* 22117 */ {(12<<2)|2,{37,103,0}},
/* 22118 */ {(12<<2)|2,{37,103,0}},
/* 22119 */ {(12<<2)|2,{37,103,0}},
/* 22120 */ {(12<<2)|2,{37,103,0}},
/* 22121 */ {(12<<2)|2,{37,103,0}},
/* 22122 */ {(12<<2)|2,{37,103,0}},
/* 22123 */ {(12<<2)|2,{37,103,0}},
/* 22124 */ {(12<<2)|2,{37,103,0}},
/* 22125 */ {(12<<2)|2,{37,103,0}},
/* 22126 */ {(12<<2)|2,{37,103,0}},
/* 22127 */ {(12<<2)|2,{37,103,0}},
/* 22128 */ {(12<<2)|2,{37,104,0}},
/* 22129 */ {(12<<2)|2,{37,104,0}},
/* 22130 */ {(12<<2)|2,{37,104,0}},
/* 22131 */ {(12<<2)|2,{37,104,0}},
/* 22132 */ {(12<<2)|2,{37,104,0}},
/* 22133 */ {(12<<2)|2,{37,104,0}},
/* 22134 */ {(12<<2)|2,{37,104,0}},
/* 22135 */ {(12<<2)|2,{37,104,0}},
/* 22136 */ {(12<<2)|2,{37,104,0}},
/* 22137 */ {(12<<2)|2,{37,104,0}},
/* 22138 */ {(12<<2)|2,{37,104,0}},
/* 22139 */ {(12<<2)|2,{37,104,0}},
/* 22140 */ {(12<<2)|2,{37,104,0}},
/* 22141 */ {(12<<2)|2,{37,104,0}},
/* 22142 */ {(12<<2)|2,{37,104,0}},
/* 22143 */ {(12<<2)|2,{37,104,0}},
/* 22144 */ {(12<<2)|2,{37,108,0}},
/* 22145 */ {(12<<2)|2,{37,108,0}},
/* 22146 */ {(12<<2)|2,{37,108,0}},
/* 22147 */ {(12<<2)|2,{37,108,0}},
/* 22148 */ {(12<<2)|2,{37,108,0}},
/* 22149 */ {(12<<2)|2,{37,108,0}},
/* 22150 */ {(12<<2)|2,{37,108,0}},
/* 22151 */ {(12<<2)|2,{37,108,0}},
/* 22152 */ {(12<<2)|2,{37,108,0}},
/* 22153 */ {(12<<2)|2,{37,108,0}},
/* 22154 */ {(12<<2)|2,{37,108,0}},
/* 22155 */ {(12<<2)|2,{37,108,0}},
/* 22156 */ {(12<<2)|2,{37,108,0}},
/* 22157 */ {(12<<2)|2,{37,108,0}},
/* 22158 */ {(12<<2)|2,{37,108,0}},
/* 22159 */ {(12<<2)|2,{37,108,0}},
/* 22160 */ {(12<<2)|2,{37,109,0}},
/* 22161 */ {(12<<2)|2,{37,109,0}},
/* 22162 */ {(12<<2)|2,{37,109,0}},
/* 22163 */ {(12<<2)|2,{37,109,0}},
/* 22164 */ {(12<<2)|2,{37,109,0}},
/* 22165 */ {(12<<2)|2,{37,109,0}},
/* 22166 */ {(12<<2)|2,{37,109,0}},
/* 22167 */ {(12<<2)|2,{37,109,0}},
/* 22168 */ {(12<<2)|2,{37,109,0}},
/* 22169 */ {(12<<2)|2,{37,109,0}},
/* 22170 */ {(12<<2)|2,{37,109,0}},
/* 22171 */ {(12<<2)|2,{37,109,0}},
/* 22172 */ {(12<<2)|2,{37,109,0}},
/* 22173 */ {(12<<2)|2,{37,109,0}},
/* 22174 */ {(12<<2)|2,{37,109,0}},
/* 22175 */ {(12<<2)|2,{37,109,0}},
/* 22176 */ {(12<<2)|2,{37,110,0}},
/* 22177 */ {(12<<2)|2,{37,110,0}},
/* 22178 */ {(12<<2)|2,{37,110,0}},
/* 22179 */ {(12<<2)|2,{37,110,0}},
/* 22180 */ {(12<<2)|2,{37,110,0}},
/* 22181 */ {(12<<2)|2,{37,110,0}},
/* 22182 */ {(12<<2)|2,{37,110,0}},
/* 22183 */ {(12<<2)|2,{37,110,0}},
/* 22184 */ {(12<<2)|2,{37,110,0}},
/* 22185 */ {(12<<2)|2,{37,110,0}},
/* 22186 */ {(12<<2)|2,{37,110,0}},
/* 22187 */ {(12<<2)|2,{37,110,0}},
/* 22188 */ {(12<<2)|2,{37,110,0}},
/* 22189 */ {(12<<2)|2,{37,110,0}},
/* 22190 */ {(12<<2)|2,{37,110,0}},
/* 22191 */ {(12<<2)|2,{37,110,0}},
/* 22192 */ {(12<<2)|2,{37,112,0}},
/* 22193 */ {(12<<2)|2,{37,112,0}},
/* 22194 */ {(12<<2)|2,{37,112,0}},
/* 22195 */ {(12<<2)|2,{37,112,0}},
/* 22196 */ {(12<<2)|2,{37,112,0}},
/* 22197 */ {(12<<2)|2,{37,112,0}},
/* 22198 */ {(12<<2)|2,{37,112,0}},
/* 22199 */ {(12<<2)|2,{37,112,0}},
/* 22200 */ {(12<<2)|2,{37,112,0}},
/* 22201 */ {(12<<2)|2,{37,112,0}},
/* 22202 */ {(12<<2)|2,{37,112,0}},
/* 22203 */ {(12<<2)|2,{37,112,0}},
/* 22204 */ {(12<<2)|2,{37,112,0}},
/* 22205 */ {(12<<2)|2,{37,112,0}},
/* 22206 */ {(12<<2)|2,{37,112,0}},
/* 22207 */ {(12<<2)|2,{37,112,0}},
/* 22208 */ {(12<<2)|2,{37,114,0}},
/* 22209 */ {(12<<2)|2,{37,114,0}},
/* 22210 */ {(12<<2)|2,{37,114,0}},
/* 22211 */ {(12<<2)|2,{37,114,0}},
/* 22212 */ {(12<<2)|2,{37,114,0}},
/* 22213 */ {(12<<2)|2,{37,114,0}},
/* 22214 */ {(12<<2)|2,{37,114,0}},
/* 22215 */ {(12<<2)|2,{37,114,0}},
/* 22216 */ {(12<<2)|2,{37,114,0}},
/* 22217 */ {(12<<2)|2,{37,114,0}},
/* 22218 */ {(12<<2)|2,{37,114,0}},
/* 22219 */ {(12<<2)|2,{37,114,0}},
/* 22220 */ {(12<<2)|2,{37,114,0}},
/* 22221 */ {(12<<2)|2,{37,114,0}},
/* 22222 */ {(12<<2)|2,{37,114,0}},
/* 22223 */ {(12<<2)|2,{37,114,0}},
/* 22224 */ {(12<<2)|2,{37,117,0}},
/* 22225 */ {(12<<2)|2,{37,117,0}},
/* 22226 */ {(12<<2)|2,{37,117,0}},
/* 22227 */ {(12<<2)|2,{37,117,0}},
/* 22228 */ {(12<<2)|2,{37,117,0}},
/* 22229 */ {(12<<2)|2,{37,117,0}},
/* 22230 */ {(12<<2)|2,{37,117,0}},
/* 22231 */ {(12<<2)|2,{37,117,0}},
/* 22232 */ {(12<<2)|2,{37,117,0}},
/* 22233 */ {(12<<2)|2,{37,117,0}},
/* 22234 */ {(12<<2)|2,{37,117,0}},
/* 22235 */ {(12<<2)|2,{37,117,0}},
/* 22236 */ {(12<<2)|2,{37,117,0}},
/* 22237 */ {(12<<2)|2,{37,117,0}},
/* 22238 */ {(12<<2)|2,{37,117,0}},
/* 22239 */ {(12<<2)|2,{37,117,0}},
/* 22240 */ {(13<<2)|2,{37,58,0}},
/* 22241 */ {(13<<2)|2,{37,58,0}},
/* 22242 */ {(13<<2)|2,{37,58,0}},
/* 22243 */ {(13<<2)|2,{37,58,0}},
/* 22244 */ {(13<<2)|2,{37,58,0}},
/* 22245 */ {(13<<2)|2,{37,58,0}},
/* 22246 */ {(13<<2)|2,{37,58,0}},
/* 22247 */ {(13<<2)|2,{37,58,0}},
/* 22248 */ {(13<<2)|2,{37,66,0}},
/* 22249 */ {(13<<2)|2,{37,66,0}},
/* 22250 */ {(13<<2)|2,{37,66,0}},
/* 22251 */ {(13<<2)|2,{37,66,0}},
/* 22252 */ {(13<<2)|2,{37,66,0}},
/* 22253 */ {(13<<2)|2,{37,66,0}},
/* 22254 */ {(13<<2)|2,{37,66,0}},
/* 22255 */ {(13<<2)|2,{37,66,0}},
/* 22256 */ {(13<<2)|2,{37,67,0}},
/* 22257 */ {(13<<2)|2,{37,67,0}},
/* 22258 */ {(13<<2)|2,{37,67,0}},
/* 22259 */ {(13<<2)|2,{37,67,0}},
/* 22260 */ {(13<<2)|2,{37,67,0}},
/* 22261 */ {(13<<2)|2,{37,67,0}},
/* 22262 */ {(13<<2)|2,{37,67,0}},
/* 22263 */ {(13<<2)|2,{37,67,0}},
/* 22264 */ {(13<<2)|2,{37,68,0}},
/* 22265 */ {(13<<2)|2,{37,68,0}},
/* 22266 */ {(13<<2)|2,{37,68,0}},
/* 22267 */ {(13<<2)|2,{37,68,0}},
/* 22268 */ {(13<<2)|2,{37,68,0}},
/* 22269 */ {(13<<2)|2,{37,68,0}},
/* 22270 */ {(13<<2)|2,{37,68,0}},
/* 22271 */ {(13<<2)|2,{37,68,0}},
/* 22272 */ {(13<<2)|2,{37,69,0}},
/* 22273 */ {(13<<2)|2,{37,69,0}},
/* 22274 */ {(13<<2)|2,{37,69,0}},
/* 22275 */ {(13<<2)|2,{37,69,0}},
/* 22276 */ {(13<<2)|2,{37,69,0}},
/* 22277 */ {(13<<2)|2,{37,69,0}},
/* 22278 */ {(13<<2)|2,{37,69,0}},
/* 22279 */ {(13<<2)|2,{37,69,0}},
/* 22280 */ {(13<<2)|2,{37,70,0}},
/* 22281 */ {(13<<2)|2,{37,70,0}},
/* 22282 */ {(13<<2)|2,{37,70,0}},
/* 22283 */ {(13<<2)|2,{37,70,0}},
/* 22284 */ {(13<<2)|2,{37,70,0}},
/* 22285 */ {(13<<2)|2,{37,70,0}},
/* 22286 */ {(13<<2)|2,{37,70,0}},
/* 22287 */ {(13<<2)|2,{37,70,0}},
/* 22288 */ {(13<<2)|2,{37,71,0}},
/* 22289 */ {(13<<2)|2,{37,71,0}},
/* 22290 */ {(13<<2)|2,{37,71,0}},
/* 22291 */ {(13<<2)|2,{37,71,0}},
/* 22292 */ {(13<<2)|2,{37,71,0}},
/* 22293 */ {(13<<2)|2,{37,71,0}},
/* 22294 */ {(13<<2)|2,{37,71,0}},
/* 22295 */ {(13<<2)|2,{37,71,0}},
/* 22296 */ {(13<<2)|2,{37,72,0}},
/* 22297 */ {(13<<2)|2,{37,72,0}},
/* 22298 */ {(13<<2)|2,{37,72,0}},
/* 22299 */ {(13<<2)|2,{37,72,0}},
/* 22300 */ {(13<<2)|2,{37,72,0}},
/* 22301 */ {(13<<2)|2,{37,72,0}},
/* 22302 */ {(13<<2)|2,{37,72,0}},
/* 22303 */ {(13<<2)|2,{37,72,0}},
/* 22304 */ {(13<<2)|2,{37,73,0}},
/* 22305 */ {(13<<2)|2,{37,73,0}},
/* 22306 */ {(13<<2)|2,{37,73,0}},
/* 22307 */ {(13<<2)|2,{37,73,0}},
/* 22308 */ {(13<<2)|2,{37,73,0}},
/* 22309 */ {(13<<2)|2,{37,73,0}},
/* 22310 */ {(13<<2)|2,{37,73,0}},
/* 22311 */ {(13<<2)|2,{37,73,0}},
/* 22312 */ {(13<<2)|2,{37,74,0}},
/* 22313 */ {(13<<2)|2,{37,74,0}},
/* 22314 */ {(13<<2)|2,{37,74,0}},
/* 22315 */ {(13<<2)|2,{37,74,0}},
/* 22316 */ {(13<<2)|2,{37,74,0}},
/* 22317 */ {(13<<2)|2,{37,74,0}},
/* 22318 */ {(13<<2)|2,{37,74,0}},
/* 22319 */ {(13<<2)|2,{37,74,0}},
/* 22320 */ {(13<<2)|2,{37,75,0}},
/* 22321 */ {(13<<2)|2,{37,75,0}},
/* 22322 */ {(13<<2)|2,{37,75,0}},
/* 22323 */ {(13<<2)|2,{37,75,0}},
/* 22324 */ {(13<<2)|2,{37,75,0}},
/* 22325 */ {(13<<2)|2,{37,75,0}},
/* 22326 */ {(13<<2)|2,{37,75,0}},
/* 22327 */ {(13<<2)|2,{37,75,0}},
/* 22328 */ {(13<<2)|2,{37,76,0}},
/* 22329 */ {(13<<2)|2,{37,76,0}},
/* 22330 */ {(13<<2)|2,{37,76,0}},
/* 22331 */ {(13<<2)|2,{37,76,0}},
/* 22332 */ {(13<<2)|2,{37,76,0}},
/* 22333 */ {(13<<2)|2,{37,76,0}},
/* 22334 */ {(13<<2)|2,{37,76,0}},
/* 22335 */ {(13<<2)|2,{37,76,0}},
/* 22336 */ {(13<<2)|2,{37,77,0}},
/* 22337 */ {(13<<2)|2,{37,77,0}},
/* 22338 */ {(13<<2)|2,{37,77,0}},
/* 22339 */ {(13<<2)|2,{37,77,0}},
/* 22340 */ {(13<<2)|2,{37,77,0}},
/* 22341 */ {(13<<2)|2,{37,77,0}},
/* 22342 */ {(13<<2)|2,{37,77,0}},
/* 22343 */ {(13<<2)|2,{37,77,0}},
/* 22344 */ {(13<<2)|2,{37,78,0}},
/* 22345 */ {(13<<2)|2,{37,78,0}},
/* 22346 */ {(13<<2)|2,{37,78,0}},
/* 22347 */ {(13<<2)|2,{37,78,0}},
/* 22348 */ {(13<<2)|2,{37,78,0}},
/* 22349 */ {(13<<2)|2,{37,78,0}},
/* 22350 */ {(13<<2)|2,{37,78,0}},
/* 22351 */ {(13<<2)|2,{37,78,0}},
/* 22352 */ {(13<<2)|2,{37,79,0}},
/* 22353 */ {(13<<2)|2,{37,79,0}},
/* 22354 */ {(13<<2)|2,{37,79,0}},
/* 22355 */ {(13<<2)|2,{37,79,0}},
/* 22356 */ {(13<<2)|2,{37,79,0}},
/* 22357 */ {(13<<2)|2,{37,79,0}},
/* 22358 */ {(13<<2)|2,{37,79,0}},
/* 22359 */ {(13<<2)|2,{37,79,0}},
/* 22360 */ {(13<<2)|2,{37,80,0}},
/* 22361 */ {(13<<2)|2,{37,80,0}},
/* 22362 */ {(13<<2)|2,{37,80,0}},
/* 22363 */ {(13<<2)|2,{37,80,0}},
/* 22364 */ {(13<<2)|2,{37,80,0}},
/* 22365 */ {(13<<2)|2,{37,80,0}},
/* 22366 */ {(13<<2)|2,{37,80,0}},
/* 22367 */ {(13<<2)|2,{37,80,0}},
/* 22368 */ {(13<<2)|2,{37,81,0}},
/* 22369 */ {(13<<2)|2,{37,81,0}},
/* 22370 */ {(13<<2)|2,{37,81,0}},
/* 22371 */ {(13<<2)|2,{37,81,0}},
/* 22372 */ {(13<<2)|2,{37,81,0}},
/* 22373 */ {(13<<2)|2,{37,81,0}},
/* 22374 */ {(13<<2)|2,{37,81,0}},
/* 22375 */ {(13<<2)|2,{37,81,0}},
/* 22376 */ {(13<<2)|2,{37,82,0}},
/* 22377 */ {(13<<2)|2,{37,82,0}},
/* 22378 */ {(13<<2)|2,{37,82,0}},
/* 22379 */ {(13<<2)|2,{37,82,0}},
/* 22380 */ {(13<<2)|2,{37,82,0}},
/* 22381 */ {(13<<2)|2,{37,82,0}},
/* 22382 */ {(13<<2)|2,{37,82,0}},
/* 22383 */ {(13<<2)|2,{37,82,0}},
/* 22384 */ {(13<<2)|2,{37,83,0}},
/* 22385 */ {(13<<2)|2,{37,83,0}},
/* 22386 */ {(13<<2)|2,{37,83,0}},
/* 22387 */ {(13<<2)|2,{37,83,0}},
/* 22388 */ {(13<<2)|2,{37,83,0}},
/* 22389 */ {(13<<2)|2,{37,83,0}},
/* 22390 */ {(13<<2)|2,{37,83,0}},
/* 22391 */ {(13<<2)|2,{37,83,0}},
/* 22392 */ {(13<<2)|2,{37,84,0}},
/* 22393 */ {(13<<2)|2,{37,84,0}},
/* 22394 */ {(13<<2)|2,{37,84,0}},
/* 22395 */ {(13<<2)|2,{37,84,0}},
/* 22396 */ {(13<<2)|2,{37,84,0}},
/* 22397 */ {(13<<2)|2,{37,84,0}},
/* 22398 */ {(13<<2)|2,{37,84,0}},
/* 22399 */ {(13<<2)|2,{37,84,0}},
/* 22400 */ {(13<<2)|2,{37,85,0}},
/* 22401 */ {(13<<2)|2,{37,85,0}},
/* 22402 */ {(13<<2)|2,{37,85,0}},
/* 22403 */ {(13<<2)|2,{37,85,0}},
/* 22404 */ {(13<<2)|2,{37,85,0}},
/* 22405 */ {(13<<2)|2,{37,85,0}},
/* 22406 */ {(13<<2)|2,{37,85,0}},
/* 22407 */ {(13<<2)|2,{37,85,0}},
/* 22408 */ {(13<<2)|2,{37,86,0}},
/* 22409 */ {(13<<2)|2,{37,86,0}},
/* 22410 */ {(13<<2)|2,{37,86,0}},
/* 22411 */ {(13<<2)|2,{37,86,0}},
/* 22412 */ {(13<<2)|2,{37,86,0}},
/* 22413 */ {(13<<2)|2,{37,86,0}},
/* 22414 */ {(13<<2)|2,{37,86,0}},
/* 22415 */ {(13<<2)|2,{37,86,0}},
/* 22416 */ {(13<<2)|2,{37,87,0}},
/* 22417 */ {(13<<2)|2,{37,87,0}},
/* 22418 */ {(13<<2)|2,{37,87,0}},
/* 22419 */ {(13<<2)|2,{37,87,0}},
/* 22420 */ {(13<<2)|2,{37,87,0}},
/* 22421 */ {(13<<2)|2,{37,87,0}},
/* 22422 */ {(13<<2)|2,{37,87,0}},
/* 22423 */ {(13<<2)|2,{37,87,0}},
/* 22424 */ {(13<<2)|2,{37,89,0}},
/* 22425 */ {(13<<2)|2,{37,89,0}},
/* 22426 */ {(13<<2)|2,{37,89,0}},
/* 22427 */ {(13<<2)|2,{37,89,0}},
/* 22428 */ {(13<<2)|2,{37,89,0}},
/* 22429 */ {(13<<2)|2,{37,89,0}},
/* 22430 */ {(13<<2)|2,{37,89,0}},
/* 22431 */ {(13<<2)|2,{37,89,0}},
/* 22432 */ {(13<<2)|2,{37,106,0}},
/* 22433 */ {(13<<2)|2,{37,106,0}},
/* 22434 */ {(13<<2)|2,{37,106,0}},
/* 22435 */ {(13<<2)|2,{37,106,0}},
/* 22436 */ {(13<<2)|2,{37,106,0}},
/* 22437 */ {(13<<2)|2,{37,106,0}},
/* 22438 */ {(13<<2)|2,{37,106,0}},
/* 22439 */ {(13<<2)|2,{37,106,0}},
/* 22440 */ {(13<<2)|2,{37,107,0}},
/* 22441 */ {(13<<2)|2,{37,107,0}},
/* 22442 */ {(13<<2)|2,{37,107,0}},
/* 22443 */ {(13<<2)|2,{37,107,0}},
/* 22444 */ {(13<<2)|2,{37,107,0}},
/* 22445 */ {(13<<2)|2,{37,107,0}},
/* 22446 */ {(13<<2)|2,{37,107,0}},
/* 22447 */ {(13<<2)|2,{37,107,0}},
/* 22448 */ {(13<<2)|2,{37,113,0}},
/* 22449 */ {(13<<2)|2,{37,113,0}},
/* 22450 */ {(13<<2)|2,{37,113,0}},
/* 22451 */ {(13<<2)|2,{37,113,0}},
/* 22452 */ {(13<<2)|2,{37,113,0}},
/* 22453 */ {(13<<2)|2,{37,113,0}},
/* 22454 */ {(13<<2)|2,{37,113,0}},
/* 22455 */ {(13<<2)|2,{37,113,0}},
/* 22456 */ {(13<<2)|2,{37,118,0}},
/* 22457 */ {(13<<2)|2,{37,118,0}},
/* 22458 */ {(13<<2)|2,{37,118,0}},
/* 22459 */ {(13<<2)|2,{37,118,0}},
/* 22460 */ {(13<<2)|2,{37,118,0}},
/* 22461 */ {(13<<2)|2,{37,118,0}},
/* 22462 */ {(13<<2)|2,{37,118,0}},
/* 22463 */ {(13<<2)|2,{37,118,0}},
/* 22464 */ {(13<<2)|2,{37,119,0}},
/* 22465 */ {(13<<2)|2,{37,119,0}},
/* 22466 */ {(13<<2)|2,{37,119,0}},
/* 22467 */ {(13<<2)|2,{37,119,0}},
/* 22468 */ {(13<<2)|2,{37,119,0}},
/* 22469 */ {(13<<2)|2,{37,119,0}},
/* 22470 */ {(13<<2)|2,{37,119,0}},
/* 22471 */ {(13<<2)|2,{37,119,0}},
/* 22472 */ {(13<<2)|2,{37,120,0}},
/* 22473 */ {(13<<2)|2,{37,120,0}},
/* 22474 */ {(13<<2)|2,{37,120,0}},
/* 22475 */ {(13<<2)|2,{37,120,0}},
/* 22476 */ {(13<<2)|2,{37,120,0}},
/* 22477 */ {(13<<2)|2,{37,120,0}},
/* 22478 */ {(13<<2)|2,{37,120,0}},
/* 22479 */ {(13<<2)|2,{37,120,0}},
/* 22480 */ {(13<<2)|2,{37,121,0}},
/* 22481 */ {(13<<2)|2,{37,121,0}},
/* 22482 */ {(13<<2)|2,{37,121,0}},
/* 22483 */ {(13<<2)|2,{37,121,0}},
/* 22484 */ {(13<<2)|2,{37,121,0}},
/* 22485 */ {(13<<2)|2,{37,121,0}},
/* 22486 */ {(13<<2)|2,{37,121,0}},
/* 22487 */ {(13<<2)|2,{37,121,0}},
/* 22488 */ {(13<<2)|2,{37,122,0}},
/* 22489 */ {(13<<2)|2,{37,122,0}},
/* 22490 */ {(13<<2)|2,{37,122,0}},
/* 22491 */ {(13<<2)|2,{37,122,0}},
/* 22492 */ {(13<<2)|2,{37,122,0}},
/* 22493 */ {(13<<2)|2,{37,122,0}},
/* 22494 */ {(13<<2)|2,{37,122,0}},
/* 22495 */ {(13<<2)|2,{37,122,0}},
/* 22496 */ {(14<<2)|2,{37,38,0}},
/* 22497 */ {(14<<2)|2,{37,38,0}},
/* 22498 */ {(14<<2)|2,{37,38,0}},
/* 22499 */ {(14<<2)|2,{37,38,0}},
/* 22500 */ {(14<<2)|2,{37,42,0}},
/* 22501 */ {(14<<2)|2,{37,42,0}},
/* 22502 */ {(14<<2)|2,{37,42,0}},
/* 22503 */ {(14<<2)|2,{37,42,0}},
/* 22504 */ {(14<<2)|2,{37,44,0}},
/* 22505 */ {(14<<2)|2,{37,44,0}},
/* 22506 */ {(14<<2)|2,{37,44,0}},
/* 22507 */ {(14<<2)|2,{37,44,0}},
/* 22508 */ {(14<<2)|2,{37,59,0}},
/* 22509 */ {(14<<2)|2,{37,59,0}},
/* 22510 */ {(14<<2)|2,{37,59,0}},
/* 22511 */ {(14<<2)|2,{37,59,0}},
/* 22512 */ {(14<<2)|2,{37,88,0}},
/* 22513 */ {(14<<2)|2,{37,88,0}},
/* 22514 */ {(14<<2)|2,{37,88,0}},
/* 22515 */ {(14<<2)|2,{37,88,0}},
/* 22516 */ {(14<<2)|2,{37,90,0}},
/* 22517 */ {(14<<2)|2,{37,90,0}},
/* 22518 */ {(14<<2)|2,{37,90,0}},
/* 22519 */ {(14<<2)|2,{37,90,0}},
/* 22520 */ {(16<<2)|2,{37,33,0}},
/* 22521 */ {(16<<2)|2,{37,34,0}},
/* 22522 */ {(16<<2)|2,{37,40,0}},
/* 22523 */ {(16<<2)|2,{37,41,0}},
/* 22524 */ {(16<<2)|2,{37,63,0}},
/* 22525 */ {(6<<2)|1,{37,0,0}},
/* 22526 */ {(6<<2)|1,{37,0,0}},
/* 22527 */ {(6<<2)|1,{37,0,0}},
/* 22528 */ {(16<<2)|3,{45,48,48}},
/* 22529 */ {(16<<2)|3,{45,48,49}},
/* 22530 */ {(16<<2)|3,{45,48,50}},
/* 22531 */ {(16<<2)|3,{45,48,97}},
/* 22532 */ {(16<<2)|3,{45,48,99}},
/* 22533 */ {(16<<2)|3,{45,48,101}},
/* 22534 */ {(16<<2)|3,{45,48,105}},
/* 22535 */ {(16<<2)|3,{45,48,111}},
/* 22536 */ {(16<<2)|3,{45,48,115}},
/* 22537 */ {(16<<2)|3,{45,48,116}},
/* 22538 */ {(11<<2)|2,{45,48,0}},
/* 22539 */ {(11<<2)|2,{45,48,0}},
/* 22540 */ {(11<<2)|2,{45,48,0}},
/* 22541 */ {(11<<2)|2,{45,48,0}},
/* 22542 */ {(11<<2)|2,{45,48,0}},
/* 22543 */ {(11<<2)|2,{45,48,0}},
/* 22544 */ {(11<<2)|2,{45,48,0}},
/* 22545 */ {(11<<2)|2,{45,48,0}},
/* 22546 */ {(11<<2)|2,{45,48,0}},
/* 22547 */ {(11<<2)|2,{45,48,0}},
/* 22548 */ {(11<<2)|2,{45,48,0}},
/* 22549 */ {(11<<2)|2,{45,48,0}},
/* 22550 */ {(11<<2)|2,{45,48,0}},
/* 22551 */ {(11<<2)|2,{45,48,0}},
/* 22552 */ {(11<<2)|2,{45,48,0}},
/* 22553 */ {(11<<2)|2,{45,48,0}},
/* 22554 */ {(11<<2)|2,{45,48,0}},
/* 22555 */ {(11<<2)|2,{45,48,0}},
/* 22556 */ {(11<<2)|2,{45,48,0}},
/* 22557 */ {(11<<2)|2,{45,48,0}},
/* 22558 */ {(11<<2)|2,{45,48,0}},
/* 22559 */ {(11<<2)|2,{45,48,0}},
/* 22560 */ {(16<<2)|3,{45,49,48}},
/* 22561 */ {(16<<2)|3,{45,49,49}},
/* 22562 */ {(16<<2)|3,{45,49,50}},
/* 22563 */ {(16<<2)|3,{45,49,97}},
/* 22564 */ {(16<<2)|3,{45,49,99}},
/* 22565 */ {(16<<2)|3,{45,49,101}},
/* 22566 */ {(16<<2)|3,{45,49,105}},
/* 22567 */ {(16<<2)|3,{45,49,111}},
/* 22568 */ {(16<<2)|3,{45,49,115}},
/* 22569 */ {(16<<2)|3,{45,49,116}},
/* 22570 */ {(11<<2)|2,{45,49,0}},
/* 22571 */ {(11<<2)|2,{45,49,0}},
/* 22572 */ {(11<<2)|2,{45,49,0}},
/* 22573 */ {(11<<2)|2,{45,49,0}},
/* 22574 */ {(11<<2)|2,{45,49,0}},
/* 22575 */ {(11<<2)|2,{45,49,0}},
/* 22576 */ {(11<<2)|2,{45,49,0}},
/* 22577 */ {(11<<2)|2,{45,49,0}},
/* 22578 */ {(11<<2)|2,{45,49,0}},
/* 22579 */ {(11<<2)|2,{45,49,0}},
/* 22580 */ {(11<<2)|2,{45,49,0}},
/* 22581 */ {(11<<2)|2,{45,49,0}},
/* 22582 */ {(11<<2)|2,{45,49,0}},
/* 22583 */ {(11<<2)|2,{45,49,0}},
/* 22584 */ {(11<<2)|2,{45,49,0}},
/* 22585 */ {(11<<2)|2,{45,49,0}},
/* 22586 */ {(11<<2)|2,{45,49,0}},
/* 22587 */ {(11<<2)|2,{45,49,0}},
/* 22588 */ {(11<<2)|2,{45,49,0}},
/* 22589 */ {(11<<2)|2,{45,49,0}},
/* 22590 */ {(11<<2)|2,{45,49,0}},
/* 22591 */ {(11<<2)|2,{45,49,0}},
/* 22592 */ {(16<<2)|3,{45,50,48}},
/* 22593 */ {(16<<2)|3,{45,50,49}},
/* 22594 */ {(16<<2)|3,{45,50,50}},
/* 22595 */ {(16<<2)|3,{45,50,97}},
/* 22596 */ {(16<<2)|3,{45,50,99}},
/* 22597 */ {(16<<2)|3,{45,50,101}},
/* 22598 */ {(16<<2)|3,{45,50,105}},
/* 22599 */ {(16<<2)|3,{45,50,111}},
/* 22600 */ {(16<<2)|3,{45,50,115}},
/* 22601 */ {(16<<2)|3,{45,50,116}},
/* 22602 */ {(11<<2)|2,{45,50,0}},
/* 22603 */ {(11<<2)|2,{45,50,0}},
/* 22604 */ {(11<<2)|2,{45,50,0}},
/* 22605 */ {(11<<2)|2,{45,50,0}},
/* 22606 */ {(11<<2)|2,{45,50,0}},
/* 22607 */ {(11<<2)|2,{45,50,0}},
/* 22608 */ {(11<<2)|2,{45,50,0}},
/* 22609 */ {(11<<2)|2,{45,50,0}},
/* 22610 */ {(11<<2)|2,{45,50,0}},
/* 22611 */ {(11<<2)|2,{45,50,0}},
/* 22612 */ {(11<<2)|2,{45,50,0}},
/* 22613 */ {(11<<2)|2,{45,50,0}},
/* 22614 */ {(11<<2)|2,{45,50,0}},
/* 22615 */ {(11<<2)|2,{45,50,0}},
/* 22616 */ {(11<<2)|2,{45,50,0}},
/* 22617 */ {(11<<2)|2,{45,50,0}},
/* 22618 */ {(11<<2)|2,{45,50,0}},
/* 22619 */ {(11<<2)|2,{45,50,0}},
/* 22620 */ {(11<<2)|2,{45,50,0}},
/* 22621 */ {(11<<2)|2,{45,50,0}},
/* 22622 */ {(11<<2)|2,{45,50,0}},
/* 22623 */ {(11<<2)|2,{45,50,0}},
/* 22624 */ {(16<<2)|3,{45,97,48}},
/* 22625 */ {(16<<2)|3,{45,97,49}},
/* 22626 */ {(16<<2)|3,{45,97,50}},
/* 22627 */ {(16<<2)|3,{45,97,97}},
/* 22628 */ {(16<<2)|3,{45,97,99}},
/* 22629 */ {(16<<2)|3,{45,97,101}},
/* 22630 */ {(16<<2)|3,{45,97,105}},
/* 22631 */ {(16<<2)|3,{45,97,111}},
/* 22632 */ {(16<<2)|3,{45,97,115}},
/* 22633 */ {(16<<2)|3,{45,97,116}},
/* 22634 */ {(11<<2)|2,{45,97,0}},
/* 22635 */ {(11<<2)|2,{45,97,0}},
/* 22636 */ {(11<<2)|2,{45,97,0}},
/* 22637 */ {(11<<2)|2,{45,97,0}},
/* 22638 */ {(11<<2)|2,{45,97,0}},
/* 22639 */ {(11<<2)|2,{45,97,0}},
/* 22640 */ {(11<<2)|2,{45,97,0}},
/* 22641 */ {(11<<2)|2,{45,97,0}},
/* 22642 */ {(11<<2)|2,{45,97,0}},
/* 22643 */ {(11<<2)|2,{45,97,0}},
/* 22644 */ {(11<<2)|2,{45,97,0}},
/* 22645 */ {(11<<2)|2,{45,97,0}},
/* 22646 */ {(11<<2)|2,{45,97,0}},
/* 22647 */ {(11<<2)|2,{45,97,0}},
/* 22648 */ {(11<<2)|2,{45,97,0}},
/* 22649 */ {(11<<2)|2,{45,97,0}},
/* 22650 */ {(11<<2)|2,{45,97,0}},
/* 22651 */ {(11<<2)|2,{45,97,0}},
/* 22652 */ {(11<<2)|2,{45,97,0}},
/* 22653 */ {(11<<2)|2,{45,97,0}},
/* 22654 */ {(11<<2)|2,{45,97,0}},
/* 22655 */ {(11<<2)|2,{45,97,0}},
/* 22656 */ {(16<<2)|3,{45,99,48}},
/* 22657 */ {(16<<2)|3,{45,99,49}},
/* 22658 */ {(16<<2)|3,{45,99,50}},
/* 22659 */ {(16<<2)|3,{45,99,97}},
/* 22660 */ {(16<<2)|3,{45,99,99}},
/* 22661 */ {(16<<2)|3,{45,99,101}},
/* 22662 */ {(16<<2)|3,{45,99,105}},
/* 22663 */ {(16<<2)|3,{45,99,111}},
/* 22664 */ {(16<<2)|3,{45,99,115}},
/* 22665 */ {(16<<2)|3,{45,99,116}},
/* 22666 */ {(11<<2)|2,{45,99,0}},
/* 22667 */ {(11<<2)|2,{45,99,0}},
/* 22668 */ {(11<<2)|2,{45,99,0}},
/* 22669 */ {(11<<2)|2,{45,99,0}},
/* 22670 */ {(11<<2)|2,{45,99,0}},
/* 22671 */ {(11<<2)|2,{45,99,0}},
/* 22672 */ {(11<<2)|2,{45,99,0}},
/* 22673 */ {(11<<2)|2,{45,99,0}},
/* 22674 */ {(11<<2)|2,{45,99,0}},
/* 22675 */ {(11<<2)|2,{45,99,0}},
/* 22676 */ {(11<<2)|2,{45,99,0}},
/* 22677 */ {(11<<2)|2,{45,99,0}},
/* 22678 */ {(11<<2)|2,{45,99,0}},
/* 22679 */ {(11<<2)|2,{45,99,0}},
/* 22680 */ {(11<<2)|2,{45,99,0}},
/* 22681 */ {(11<<2)|2,{45,99,0}},
/* 22682 */ {(11<<2)|2,{45,99,0}},
/* 22683 */ {(11<<2)|2,{45,99,0}},
/* 22684 */ {(11<<2)|2,{45,99,0}},
/* 22685 */ {(11<<2)|2,{45,99,0}},
/* 22686 */ {(11<<2)|2,{45,99,0}},
/* 22687 */ {(11<<2)|2,{45,99,0}},
/* 22688 */ {(16<<2)|3,{45,101,48}},
/* 22689 */ {(16<<2)|3,{45,101,49}},
/* 22690 */ {(16<<2)|3,{45,101,50}},
/* 22691 */ {(16<<2)|3,{45,101,97}},
/* 22692 */ {(16<<2)|3,{45,101,99}},
/* 22693 */ {(16<<2)|3,{45,101,101}},
/* 22694 */ {(16<<2)|3,{45,101,105}},
/* 22695 */ {(16<<2)|3,{45,101,111}},
/* 22696 */ {(16<<2)|3,{45,101,115}},
/* 22697 */ {(16<<2)|3,{45,101,116}},
/* 22698 */ {(11<<2)|2,{45,101,0}},
/* 22699 */ {(11<<2)|2,{45,101,0}},
/* 22700 */ {(11<<2)|2,{45,101,0}},
/* 22701 */ {(11<<2)|2,{45,101,0}},
/* 22702 */ {(11<<2)|2,{45,101,0}},
/* 22703 */ {(11<<2)|2,{45,101,0}},
/* 22704 */ {(11<<2)|2,{45,101,0}},
/* 22705 */ {(11<<2)|2,{45,101,0}},
/* 22706 */ {(11<<2)|2,{45,101,0}},
/* 22707 */ {(11<<2)|2,{45,101,0}},
/* 22708 */ {(11<<2)|2,{45,101,0}},
/* 22709 */ {(11<<2)|2,{45,101,0}},
/* 22710 */ {(11<<2)|2,{45,101,0}},
/* 22711 */ {(11<<2)|2,{45,101,0}},
/* 22712 */ {(11<<2)|2,{45,101,0}},
/* 22713 */ {(11<<2)|2,{45,101,0}},
/* 22714 */ {(11<<2)|2,{45,101,0}},
/* 22715 */ {(11<<2)|2,{45,101,0}},
/* 22716 */ {(11<<2)|2,{45,101,0}},
/* 22717 */ {(11<<2)|2,{45,101,0}},
/* 22718 */ {(11<<2)|2,{45,101,0}},
/* 22719 */ {(11<<2)|2,{45,101,0}},
/* 22720 */ {(16<<2)|3,{45,105,48}},
/* 22721 */ {(16<<2)|3,{45,105,49}},
/* 22722 */ {(16<<2)|3,{45,105,50}},
/* 22723 */ {(16<<2)|3,{45,105,97}},
/* 22724 */ {(16<<2)|3,{45,105,99}},
/* 22725 */ {(16<<2)|3,{45,105,101}},
/* 22726 */ {(16<<2)|3,{45,105,105}},
/* 22727 */ {(16<<2)|3,{45,105,111}},
/* 22728 */ {(16<<2)|3,{45,105,115}},
/* 22729 */ {(16<<2)|3,{45,105,116}},
/* 22730 */ {(11<<2)|2,{45,105,0}},
/* 22731 */ {(11<<2)|2,{45,105,0}},
/* 22732 */ {(11<<2)|2,{45,105,0}},
/* 22733 */ {(11<<2)|2,{45,105,0}},
/* 22734 */ {(11<<2)|2,{45,105,0}},
/* 22735 */ {(11<<2)|2,{45,105,0}},
/* 22736 */ {(11<<2)|2,{45,105,0}},
/* 22737 */ {(11<<2)|2,{45,105,0}},
/* 22738 */ {(11<<2)|2,{45,105,0}},
/* 22739 */ {(11<<2)|2,{45,105,0}},
/* 22740 */ {(11<<2)|2,{45,105,0}},
/* 22741 */ {(11<<2)|2,{45,105,0}},
/* 22742 */ {(11<<2)|2,{45,105,0}},
/* 22743 */ {(11<<2)|2,{45,105,0}},
/* 22744 */ {(11<<2)|2,{45,105,0}},
/* 22745 */ {(11<<2)|2,{45,105,0}},
/* 22746 */ {(11<<2)|2,{45,105,0}},
/* 22747 */ {(11<<2)|2,{45,105,0}},
/* 22748 */ {(11<<2)|2,{45,105,0}},
/* 22749 */ {(11<<2)|2,{45,105,0}},
/* 22750 */ {(11<<2)|2,{45,105,0}},
/* 22751 */ {(11<<2)|2,{45,105,0}},
/* 22752 */ {(16<<2)|3,{45,111,48}},
/* 22753 */ {(16<<2)|3,{45,111,49}},
/* 22754 */ {(16<<2)|3,{45,111,50}},
/* 22755 */ {(16<<2)|3,{45,111,97}},
/* 22756 */ {(16<<2)|3,{45,111,99}},
/* 22757 */ {(16<<2)|3,{45,111,101}},
/* 22758 */ {(16<<2)|3,{45,111,105}},
/* 22759 */ {(16<<2)|3,{45,111,111}},
/* 22760 */ {(16<<2)|3,{45,111,115}},
/* 22761 */ {(16<<2)|3,{45,111,116}},
/* 22762 */ {(11<<2)|2,{45,111,0}},
/* 22763 */ {(11<<2)|2,{45,111,0}},
/* 22764 */ {(11<<2)|2,{45,111,0}},
/* 22765 */ {(11<<2)|2,{45,111,0}},
/* 22766 */ {(11<<2)|2,{45,111,0}},
/* 22767 */ {(11<<2)|2,{45,111,0}},
/* 22768 */ {(11<<2)|2,{45,111,0}},
/* 22769 */ {(11<<2)|2,{45,111,0}},
/* 22770 */ {(11<<2)|2,{45,111,0}},
/* 22771 */ {(11<<2)|2,{45,111,0}},
/* 22772 */ {(11<<2)|2,{45,111,0}},
/* 22773 */ {(11<<2)|2,{45,111,0}},
/* 22774 */ {(11<<2)|2,{45,111,0}},
/* 22775 */ {(11<<2)|2,{45,111,0}},
/* 22776 */ {(11<<2)|2,{45,111,0}},
/* 22777 */ {(11<<2)|2,{45,111,0}},
/* 22778 */ {(11<<2)|2,{45,111,0}},
/* 22779 */ {(11<<2)|2,{45,111,0}},
/* 22780 */ {(11<<2)|2,{45,111,0}},
/* 22781 */ {(11<<2)|2,{45,111,0}},
/* 22782 */ {(11<<2)|2,{45,111,0}},
/* 22783 */ {(11<<2)|2,{45,111,0}},
/* 22784 */ {(16<<2)|3,{45,115,48}},
/* 22785 */ {(16<<2)|3,{45,115,49}},
/* 22786 */ {(16<<2)|3,{45,115,50}},
/* 22787 */ {(16<<2)|3,{45,115,97}},
/* 22788 */ {(16<<2)|3,{45,115,99}},
/* 22789 */ {(16<<2)|3,{45,115,101}},
/* 22790 */ {(16<<2)|3,{45,115,105}},
/* 22791 */ {(16<<2)|3,{45,115,111}},
/* 22792 */ {(16<<2)|3,{45,115,115}},
/* 22793 */ {(16<<2)|3,{45,115,116}},
/* 22794 */ {(11<<2)|2,{45,115,0}},
/* 22795 */ {(11<<2)|2,{45,115,0}},
/* 22796 */ {(11<<2)|2,{45,115,0}},
/* 22797 */ {(11<<2)|2,{45,115,0}},
/* 22798 */ {(11<<2)|2,{45,115,0}},
/* 22799 */ {(11<<2)|2,{45,115,0}},
/* 22800 */ {(11<<2)|2,{45,115,0}},
/* 22801 */ {(11<<2)|2,{45,115,0}},
/* 22802 */ {(11<<2)|2,{45,115,0}},
/* 22803 */ {(11<<2)|2,{45,115,0}},
/* 22804 */ {(11<<2)|2,{45,115,0}},
/* 22805 */ {(11<<2)|2,{45,115,0}},
/* 22806 */ {(11<<2)|2,{45,115,0}},
/* 22807 */ {(11<<2)|2,{45,115,0}},
/* 22808 */ {(11<<2)|2,{45,115,0}},
/* 22809 */ {(11<<2)|2,{45,115,0}},
/* 22810 */ {(11<<2)|2,{45,115,0}},
/* 22811 */ {(11<<2)|2,{45,115,0}},
/* 22812 */ {(11<<2)|2,{45,115,0}},
/* 22813 */ {(11<<2)|2,{45,115,0}},
/* 22814 */ {(11<<2)|2,{45,115,0}},
/* 22815 */ {(11<<2)|2,{45,115,0}},
/* 22816 */ {(16<<2)|3,{45,116,48}},
/* 22817 */ {(16<<2)|3,{45,116,49}},
/* 22818 */ {(16<<2)|3,{45,116,50}},
/* 22819 */ {(16<<2)|3,{45,116,97}},
/* 22820 */ {(16<<2)|3,{45,116,99}},
/* 22821 */ {(16<<2)|3,{45,116,101}},
/* 22822 */ {(16<<2)|3,{45,116,105}},
/* 22823 */ {(16<<2)|3,{45,116,111}},
/* 22824 */ {(16<<2)|3,{45,116,115}},
/* 22825 */ {(16<<2)|3,{45,116,116}},
/* 22826 */ {(11<<2)|2,{45,116,0}},
/* 22827 */ {(11<<2)|2,{45,116,0}},
/* 22828 */ {(11<<2)|2,{45,116,0}},
/* 22829 */ {(11<<2)|2,{45,116,0}},
/* 22830 */ {(11<<2)|2,{45,116,0}},
/* 22831 */ {(11<<2)|2,{45,116,0}},
/* 22832 */ {(11<<2)|2,{45,116,0}},
/* 22833 */ {(11<<2)|2,{45,116,0}},
/* 22834 */ {(11<<2)|2,{45,116,0}},
/* 22835 */ {(11<<2)|2,{45,116,0}},
/* 22836 */ {(11<<2)|2,{45,116,0}},
/* 22837 */ {(11<<2)|2,{45,116,0}},
/* 22838 */ {(11<<2)|2,{45,116,0}},
/* 22839 */ {(11<<2)|2,{45,116,0}},
/* 22840 */ {(11<<2)|2,{45,116,0}},
/* 22841 */ {(11<<2)|2,{45,116,0}},
/* 22842 */ {(11<<2)|2,{45,116,0}},
/* 22843 */ {(11<<2)|2,{45,116,0}},
/* 22844 */ {(11<<2)|2,{45,116,0}},
/* 22845 */ {(11<<2)|2,{45,116,0}},
/* 22846 */ {(11<<2)|2,{45,116,0}},
/* 22847 */ {(11<<2)|2,{45,116,0}},
/* 22848 */ {(12<<2)|2,{45,32,0}},
/* 22849 */ {(12<<2)|2,{45,32,0}},
/* 22850 */ {(12<<2)|2,{45,32,0}},
/* 22851 */ {(12<<2)|2,{45,32,0}},
/* 22852 */ {(12<<2)|2,{45,32,0}},
/* 22853 */ {(12<<2)|2,{45,32,0}},
/* 22854 */ {(12<<2)|2,{45,32,0}},
/* 22855 */ {(12<<2)|2,{45,32,0}},
/* 22856 */ {(12<<2)|2,{45,32,0}},
/* 22857 */ {(12<<2)|2,{45,32,0}},
/* 22858 */ {(12<<2)|2,{45,32,0}},
/* 22859 */ {(12<<2)|2,{45,32,0}},
/* 22860 */ {(12<<2)|2,{45,32,0}},
/* 22861 */ {(12<<2)|2,{45,32,0}},
/* 22862 */ {(12<<2)|2,{45,32,0}},
/* 22863 */ {(12<<2)|2,{45,32,0}},
/* 22864 */ {(12<<2)|2,{45,37,0}},
/* 22865 */ {(12<<2)|2,{45,37,0}},
/* 22866 */ {(12<<2)|2,{45,37,0}},
/* 22867 */ {(12<<2)|2,{45,37,0}},
/* 22868 */ {(12<<2)|2,{45,37,0}},
/* 22869 */ {(12<<2)|2,{45,37,0}},
/* 22870 */ {(12<<2)|2,{45,37,0}},
/* 22871 */ {(12<<2)|2,{45,37,0}},
/* 22872 */ {(12<<2)|2,{45,37,0}},
/* 22873 */ {(12<<2)|2,{45,37,0}},
/* 22874 */ {(12<<2)|2,{45,37,0}},
/* 22875 */ {(12<<2)|2,{45,37,0}},
/* 22876 */ {(12<<2)|2,{45,37,0}},
/* 22877 */ {(12<<2)|2,{45,37,0}},
/* 22878 */ {(12<<2)|2,{45,37,0}},
/* 22879 */ {(12<<2)|2,{45,37,0}},
/* 22880 */ {(12<<2)|2,{45,45,0}},
/* 22881 */ {(12<<2)|2,{45,45,0}},
/* 22882 */ {(12<<2)|2,{45,45,0}},
/* 22883 */ {(12<<2)|2,{45,45,0}},
/* 22884 */ {(12<<2)|2,{45,45,0}},
/* 22885 */ {(12<<2)|2,{45,45,0}},
/* 22886 */ {(12<<2)|2,{45,45,0}},
/* 22887 */ {(12<<2)|2,{45,45,0}},
/* 22888 */ {(12<<2)|2,{45,45,0}},
/* 22889 */ {(12<<2)|2,{45,45,0}},
/* 22890 */ {(12<<2)|2,{45,45,0}},
/* 22891 */ {(12<<2)|2,{45,45,0}},
/* 22892 */ {(12<<2)|2,{45,45,0}},
/* 22893 */ {(12<<2)|2,{45,45,0}},
/* 22894 */ {(12<<2)|2,{45,45,0}},
/* 22895 */ {(12<<2)|2,{45,45,0}},
/* 22896 */ {(12<<2)|2,{45,46,0}},
/* 22897 */ {(12<<2)|2,{45,46,0}},
/* 22898 */ {(12<<2)|2,{45,46,0}},
/* 22899 */ {(12<<2)|2,{45,46,0}},
/* 22900 */ {(12<<2)|2,{45,46,0}},
/* 22901 */ {(12<<2)|2,{45,46,0}},
/* 22902 */ {(12<<2)|2,{45,46,0}},
/* 22903 */ {(12<<2)|2,{45,46,0}},
/* 22904 */ {(12<<2)|2,{45,46,0}},
/* 22905 */ {(12<<2)|2,{45,46,0}},
/* 22906 */ {(12<<2)|2,{45,46,0}},
/* 22907 */ {(12<<2)|2,{45,46,0}},
/* 22908 */ {(12<<2)|2,{45,46,0}},
/* 22909 */ {(12<<2)|2,{45,46,0}},
/* 22910 */ {(12<<2)|2,{45,46,0}},
/* 22911 */ {(12<<2)|2,{45,46,0}},
/* 22912 */ {(12<<2)|2,{45,47,0}},
/* 22913 */ {(12<<2)|2,{45,47,0}},
/* 22914 */ {(12<<2)|2,{45,47,0}},
/* 22915 */ {(12<<2)|2,{45,47,0}},
/* 22916 */ {(12<<2)|2,{45,47,0}},
/* 22917 */ {(12<<2)|2,{45,47,0}},
/* 22918 */ {(12<<2)|2,{45,47,0}},
/* 22919 */ {(12<<2)|2,{45,47,0}},
/* 22920 */ {(12<<2)|2,{45,47,0}},
/* 22921 */ {(12<<2)|2,{45,47,0}},
/* 22922 */ {(12<<2)|2,{45,47,0}},
/* 22923 */ {(12<<2)|2,{45,47,0}},
/* 22924 */ {(12<<2)|2,{45,47,0}},
/* 22925 */ {(12<<2)|2,{45,47,0}},
/* 22926 */ {(12<<2)|2,{45,47,0}},
/* 22927 */ {(12<<2)|2,{45,47,0}},
/* 22928 */ {(12<<2)|2,{45,51,0}},
/* 22929 */ {(12<<2)|2,{45,51,0}},
/* 22930 */ {(12<<2)|2,{45,51,0}},
/* 22931 */ {(12<<2)|2,{45,51,0}},
/* 22932 */ {(12<<2)|2,{45,51,0}},
/* 22933 */ {(12<<2)|2,{45,51,0}},
/* 22934 */ {(12<<2)|2,{45,51,0}},
/* 22935 */ {(12<<2)|2,{45,51,0}},
/* 22936 */ {(12<<2)|2,{45,51,0}},
/* 22937 */ {(12<<2)|2,{45,51,0}},
/* 22938 */ {(12<<2)|2,{45,51,0}},
/* 22939 */ {(12<<2)|2,{45,51,0}},
/* 22940 */ {(12<<2)|2,{45,51,0}},
/* 22941 */ {(12<<2)|2,{45,51,0}},
/* 22942 */ {(12<<2)|2,{45,51,0}},
/* 22943 */ {(12<<2)|2,{45,51,0}},
/* 22944 */ {(12<<2)|2,{45,52,0}},
/* 22945 */ {(12<<2)|2,{45,52,0}},
/* 22946 */ {(12<<2)|2,{45,52,0}},
/* 22947 */ {(12<<2)|2,{45,52,0}},
/* 22948 */ {(12<<2)|2,{45,52,0}},
/* 22949 */ {(12<<2)|2,{45,52,0}},
/* 22950 */ {(12<<2)|2,{45,52,0}},
/* 22951 */ {(12<<2)|2,{45,52,0}},
/* 22952 */ {(12<<2)|2,{45,52,0}},
/* 22953 */ {(12<<2)|2,{45,52,0}},
/* 22954 */ {(12<<2)|2,{45,52,0}},
/* 22955 */ {(12<<2)|2,{45,52,0}},
/* 22956 */ {(12<<2)|2,{45,52,0}},
/* 22957 */ {(12<<2)|2,{45,52,0}},
/* 22958 */ {(12<<2)|2,{45,52,0}},
/* 22959 */ {(12<<2)|2,{45,52,0}},
/* 22960 */ {(12<<2)|2,{45,53,0}},
/* 22961 */ {(12<<2)|2,{45,53,0}},
/* 22962 */ {(12<<2)|2,{45,53,0}},
/* 22963 */ {(12<<2)|2,{45,53,0}},
/* 22964 */ {(12<<2)|2,{45,53,0}},
/* 22965 */ {(12<<2)|2,{45,53,0}},
/* 22966 */ {(12<<2)|2,{45,53,0}},
/* 22967 */ {(12<<2)|2,{45,53,0}},
/* 22968 */ {(12<<2)|2,{45,53,0}},
/* 22969 */ {(12<<2)|2,{45,53,0}},
/* 22970 */ {(12<<2)|2,{45,53,0}},
/* 22971 */ {(12<<2)|2,{45,53,0}},
/* 22972 */ {(12<<2)|2,{45,53,0}},
/* 22973 */ {(12<<2)|2,{45,53,0}},
/* 22974 */ {(12<<2)|2,{45,53,0}},
/* 22975 */ {(12<<2)|2,{45,53,0}},
/* 22976 */ {(12<<2)|2,{45,54,0}},
/* 22977 */ {(12<<2)|2,{45,54,0}},
/* 22978 */ {(12<<2)|2,{45,54,0}},
/* 22979 */ {(12<<2)|2,{45,54,0}},
/* 22980 */ {(12<<2)|2,{45,54,0}},
/* 22981 */ {(12<<2)|2,{45,54,0}},
/* 22982 */ {(12<<2)|2,{45,54,0}},
/* 22983 */ {(12<<2)|2,{45,54,0}},
/* 22984 */ {(12<<2)|2,{45,54,0}},
/* 22985 */ {(12<<2)|2,{45,54,0}},
/* 22986 */ {(12<<2)|2,{45,54,0}},
/* 22987 */ {(12<<2)|2,{45,54,0}},
/* 22988 */ {(12<<2)|2,{45,54,0}},
/* 22989 */ {(12<<2)|2,{45,54,0}},
/* 22990 */ {(12<<2)|2,{45,54,0}},
/* 22991 */ {(12<<2)|2,{45,54,0}},
/* 22992 */ {(12<<2)|2,{45,55,0}},
/* 22993 */ {(12<<2)|2,{45,55,0}},
/* 22994 */ {(12<<2)|2,{45,55,0}},
/* 22995 */ {(12<<2)|2,{45,55,0}},
/* 22996 */ {(12<<2)|2,{45,55,0}},
/* 22997 */ {(12<<2)|2,{45,55,0}},
/* 22998 */ {(12<<2)|2,{45,55,0}},
/* 22999 */ {(12<<2)|2,{45,55,0}},
/* 23000 */ {(12<<2)|2,{45,55,0}},
/* 23001 */ {(12<<2)|2,{45,55,0}},
/* 23002 */ {(12<<2)|2,{45,55,0}},
/* 23003 */ {(12<<2)|2,{45,55,0}},
/* 23004 */ {(12<<2)|2,{45,55,0}},
/* 23005 */ {(12<<2)|2,{45,55,0}},
/* 23006 */ {(12<<2)|2,{45,55,0}},
/* 23007 */ {(12<<2)|2,{45,55,0}},
/* 23008 */ {(12<<2)|2,{45,56,0}},
/* 23009 */ {(12<<2)|2,{45,56,0}},
/* 23010 */ {(12<<2)|2,{45,56,0}},
/* 23011 */ {(12<<2)|2,{45,56,0}},
/* 23012 */ {(12<<2)|2,{45,56,0}},
/* 23013 */ {(12<<2)|2,{45,56,0}},
/* 23014 */ {(12<<2)|2,{45,56,0}},
/* 23015 */ {(12<<2)|2,{45,56,0}},
/* 23016 */ {(12<<2)|2,{45,56,0}},
/* 23017 */ {(12<<2)|2,{45,56,0}},
/* 23018 */ {(12<<2)|2,{45,56,0}},
/* 23019 */ {(12<<2)|2,{45,56,0}},
/* 23020 */ {(12<<2)|2,{45,56,0}},
/* 23021 */ {(12<<2)|2,{45,56,0}},
/* 23022 */ {(12<<2)|2,{45,56,0}},
/* 23023 */ {(12<<2)|2,{45,56,0}},
/* 23024 */ {(12<<2)|2,{45,57,0}},
/* 23025 */ {(12<<2)|2,{45,57,0}},
/* 23026 */ {(12<<2)|2,{45,57,0}},
/* 23027 */ {(12<<2)|2,{45,57,0}},
/* 23028 */ {(12<<2)|2,{45,57,0}},
/* 23029 */ {(12<<2)|2,{45,57,0}},
/* 23030 */ {(12<<2)|2,{45,57,0}},
/* 23031 */ {(12<<2)|2,{45,57,0}},
/* 23032 */ {(12<<2)|2,{45,57,0}},
/* 23033 */ {(12<<2)|2,{45,57,0}},
/* 23034 */ {(12<<2)|2,{45,57,0}},
/* 23035 */ {(12<<2)|2,{45,57,0}},
/* 23036 */ {(12<<2)|2,{45,57,0}},
/* 23037 */ {(12<<2)|2,{45,57,0}},
/* 23038 */ {(12<<2)|2,{45,57,0}},
/* 23039 */ {(12<<2)|2,{45,57,0}},
/* 23040 */ {(12<<2)|2,{45,61,0}},
/* 23041 */ {(12<<2)|2,{45,61,0}},
/* 23042 */ {(12<<2)|2,{45,61,0}},
/* 23043 */ {(12<<2)|2,{45,61,0}},
/* 23044 */ {(12<<2)|2,{45,61,0}},
/* 23045 */ {(12<<2)|2,{45,61,0}},
/* 23046 */ {(12<<2)|2,{45,61,0}},
/* 23047 */ {(12<<2)|2,{45,61,0}},
/* 23048 */ {(12<<2)|2,{45,61,0}},
/* 23049 */ {(12<<2)|2,{45,61,0}},
/* 23050 */ {(12<<2)|2,{45,61,0}},
/* 23051 */ {(12<<2)|2,{45,61,0}},
/* 23052 */ {(12<<2)|2,{45,61,0}},
/* 23053 */ {(12<<2)|2,{45,61,0}},
/* 23054 */ {(12<<2)|2,{45,61,0}},
/* 23055 */ {(12<<2)|2,{45,61,0}},
/* 23056 */ {(12<<2)|2,{45,65,0}},
/* 23057 */ {(12<<2)|2,{45,65,0}},
/* 23058 */ {(12<<2)|2,{45,65,0}},
/* 23059 */ {(12<<2)|2,{45,65,0}},
/* 23060 */ {(12<<2)|2,{45,65,0}},
/* 23061 */ {(12<<2)|2,{45,65,0}},
/* 23062 */ {(12<<2)|2,{45,65,0}},
/* 23063 */ {(12<<2)|2,{45,65,0}},
/* 23064 */ {(12<<2)|2,{45,65,0}},
/* 23065 */ {(12<<2)|2,{45,65,0}},
/* 23066 */ {(12<<2)|2,{45,65,0}},
/* 23067 */ {(12<<2)|2,{45,65,0}},
/* 23068 */ {(12<<2)|2,{45,65,0}},
/* 23069 */ {(12<<2)|2,{45,65,0}},
/* 23070 */ {(12<<2)|2,{45,65,0}},
/* 23071 */ {(12<<2)|2,{45,65,0}},
/* 23072 */ {(12<<2)|2,{45,95,0}},
/* 23073 */ {(12<<2)|2,{45,95,0}},
/* 23074 */ {(12<<2)|2,{45,95,0}},
/* 23075 */ {(12<<2)|2,{45,95,0}},
/* 23076 */ {(12<<2)|2,{45,95,0}},
/* 23077 */ {(12<<2)|2,{45,95,0}},
/* 23078 */ {(12<<2)|2,{45,95,0}},
/* 23079 */ {(12<<2)|2,{45,95,0}},
/* 23080 */ {(12<<2)|2,{45,95,0}},
/* 23081 */ {(12<<2)|2,{45,95,0}},
/* 23082 */ {(12<<2)|2,{45,95,0}},
/* 23083 */ {(12<<2)|2,{45,95,0}},
/* 23084 */ {(12<<2)|2,{45,95,0}},
/* 23085 */ {(12<<2)|2,{45,95,0}},
/* 23086 */ {(12<<2)|2,{45,95,0}},
/* 23087 */ {(12<<2)|2,{45,95,0}},
/* 23088 */ {(12<<2)|2,{45,98,0}},
/* 23089 */ {(12<<2)|2,{45,98,0}},
/* 23090 */ {(12<<2)|2,{45,98,0}},
/* 23091 */ {(12<<2)|2,{45,98,0}},
/* 23092 */ {(12<<2)|2,{45,98,0}},
/* 23093 */ {(12<<2)|2,{45,98,0}},
/* 23094 */ {(12<<2)|2,{45,98,0}},
/* 23095 */ {(12<<2)|2,{45,98,0}},
/* 23096 */ {(12<<2)|2,{45,98,0}},
/* 23097 */ {(12<<2)|2,{45,98,0}},
/* 23098 */ {(12<<2)|2,{45,98,0}},
/* 23099 */ {(12<<2)|2,{45,98,0}},
/* 23100 */ {(12<<2)|2,{45,98,0}},
/* 23101 */ {(12<<2)|2,{45,98,0}},
/* 23102 */ {(12<<2)|2,{45,98,0}},
/* 23103 */ {(12<<2)|2,{45,98,0}},
/* 23104 */ {(12<<2)|2,{45,100,0}},
/* 23105 */ {(12<<2)|2,{45,100,0}},
/* 23106 */ {(12<<2)|2,{45,100,0}},
/* 23107 */ {(12<<2)|2,{45,100,0}},
/* 23108 */ {(12<<2)|2,{45,100,0}},
/* 23109 */ {(12<<2)|2,{45,100,0}},
/* 23110 */ {(12<<2)|2,{45,100,0}},
/* 23111 */ {(12<<2)|2,{45,100,0}},
/* 23112 */ {(12<<2)|2,{45,100,0}},
/* 23113 */ {(12<<2)|2,{45,100,0}},
/* 23114 */ {(12<<2)|2,{45,100,0}},
/* 23115 */ {(12<<2)|2,{45,100,0}},
/* 23116 */ {(12<<2)|2,{45,100,0}},
/* 23117 */ {(12<<2)|2,{45,100,0}},
/* 23118 */ {(12<<2)|2,{45,100,0}},
/* 23119 */ {(12<<2)|2,{45,100,0}},
/* 23120 */ {(12<<2)|2,{45,102,0}},
/* 23121 */ {(12<<2)|2,{45,102,0}},
/* 23122 */ {(12<<2)|2,{45,102,0}},
/* 23123 */ {(12<<2)|2,{45,102,0}},
/* 23124 */ {(12<<2)|2,{45,102,0}},
/* 23125 */ {(12<<2)|2,{45,102,0}},
/* 23126 */ {(12<<2)|2,{45,102,0}},
/* 23127 */ {(12<<2)|2,{45,102,0}},
/* 23128 */ {(12<<2)|2,{45,102,0}},
/* 23129 */ {(12<<2)|2,{45,102,0}},
/* 23130 */ {(12<<2)|2,{45,102,0}},
/* 23131 */ {(12<<2)|2,{45,102,0}},
/* 23132 */ {(12<<2)|2,{45,102,0}},
/* 23133 */ {(12<<2)|2,{45,102,0}},
/* 23134 */ {(12<<2)|2,{45,102,0}},
/* 23135 */ {(12<<2)|2,{45,102,0}},
/* 23136 */ {(12<<2)|2,{45,103,0}},
/* 23137 */ {(12<<2)|2,{45,103,0}},
/* 23138 */ {(12<<2)|2,{45,103,0}},
/* 23139 */ {(12<<2)|2,{45,103,0}},
/* 23140 */ {(12<<2)|2,{45,103,0}},
/* 23141 */ {(12<<2)|2,{45,103,0}},
/* 23142 */ {(12<<2)|2,{45,103,0}},
/* 23143 */ {(12<<2)|2,{45,103,0}},
/* 23144 */ {(12<<2)|2,{45,103,0}},
/* 23145 */ {(12<<2)|2,{45,103,0}},
/* 23146 */ {(12<<2)|2,{45,103,0}},
/* 23147 */ {(12<<2)|2,{45,103,0}},
/* 23148 */ {(12<<2)|2,{45,103,0}},
/* 23149 */ {(12<<2)|2,{45,103,0}},
/* 23150 */ {(12<<2)|2,{45,103,0}},
/* 23151 */ {(12<<2)|2,{45,103,0}},
/* 23152 */ {(12<<2)|2,{45,104,0}},
/* 23153 */ {(12<<2)|2,{45,104,0}},
/* 23154 */ {(12<<2)|2,{45,104,0}},
/* 23155 */ {(12<<2)|2,{45,104,0}},
/* 23156 */ {(12<<2)|2,{45,104,0}},
/* 23157 */ {(12<<2)|2,{45,104,0}},
/* 23158 */ {(12<<2)|2,{45,104,0}},
/* 23159 */ {(12<<2)|2,{45,104,0}},
/* 23160 */ {(12<<2)|2,{45,104,0}},
/* 23161 */ {(12<<2)|2,{45,104,0}},
/* 23162 */ {(12<<2)|2,{45,104,0}},
/* 23163 */ {(12<<2)|2,{45,104,0}},
/* 23164 */ {(12<<2)|2,{45,104,0}},
/* 23165 */ {(12<<2)|2,{45,104,0}},
/* 23166 */ {(12<<2)|2,{45,104,0}},
/* 23167 */ {(12<<2)|2,{45,104,0}},
/* 23168 */ {(12<<2)|2,{45,108,0}},
/* 23169 */ {(12<<2)|2,{45,108,0}},
/* 23170 */ {(12<<2)|2,{45,108,0}},
/* 23171 */ {(12<<2)|2,{45,108,0}},
/* 23172 */ {(12<<2)|2,{45,108,0}},
/* 23173 */ {(12<<2)|2,{45,108,0}},
/* 23174 */ {(12<<2)|2,{45,108,0}},
/* 23175 */ {(12<<2)|2,{45,108,0}},
/* 23176 */ {(12<<2)|2,{45,108,0}},
/* 23177 */ {(12<<2)|2,{45,108,0}},
/* 23178 */ {(12<<2)|2,{45,108,0}},
/* 23179 */ {(12<<2)|2,{45,108,0}},
/* 23180 */ {(12<<2)|2,{45,108,0}},
/* 23181 */ {(12<<2)|2,{45,108,0}},
/* 23182 */ {(12<<2)|2,{45,108,0}},
/* 23183 */ {(12<<2)|2,{45,108,0}},
/* 23184 */ {(12<<2)|2,{45,109,0}},
/* 23185 */ {(12<<2)|2,{45,109,0}},
/* 23186 */ {(12<<2)|2,{45,109,0}},
/* 23187 */ {(12<<2)|2,{45,109,0}},
/* 23188 */ {(12<<2)|2,{45,109,0}},
/* 23189 */ {(12<<2)|2,{45,109,0}},
/* 23190 */ {(12<<2)|2,{45,109,0}},
/* 23191 */ {(12<<2)|2,{45,109,0}},
/* 23192 */ {(12<<2)|2,{45,109,0}},
/* 23193 */ {(12<<2)|2,{45,109,0}},
/* 23194 */ {(12<<2)|2,{45,109,0}},
/* 23195 */ {(12<<2)|2,{45,109,0}},
/* 23196 */ {(12<<2)|2,{45,109,0}},
/* 23197 */ {(12<<2)|2,{45,109,0}},
/* 23198 */ {(12<<2)|2,{45,109,0}},
/* 23199 */ {(12<<2)|2,{45,109,0}},
/* 23200 */ {(12<<2)|2,{45,110,0}},
/* 23201 */ {(12<<2)|2,{45,110,0}},
/* 23202 */ {(12<<2)|2,{45,110,0}},
/* 23203 */ {(12<<2)|2,{45,110,0}},
/* 23204 */ {(12<<2)|2,{45,110,0}},
/* 23205 */ {(12<<2)|2,{45,110,0}},
/* 23206 */ {(12<<2)|2,{45,110,0}},
/* 23207 */ {(12<<2)|2,{45,110,0}},
/* 23208 */ {(12<<2)|2,{45,110,0}},
/* 23209 */ {(12<<2)|2,{45,110,0}},
/* 23210 */ {(12<<2)|2,{45,110,0}},
/* 23211 */ {(12<<2)|2,{45,110,0}},
/* 23212 */ {(12<<2)|2,{45,110,0}},
/* 23213 */ {(12<<2)|2,{45,110,0}},
/* 23214 */ {(12<<2)|2,{45,110,0}},
/* 23215 */ {(12<<2)|2,{45,110,0}},
/* 23216 */ {(12<<2)|2,{45,112,0}},
/* 23217 */ {(12<<2)|2,{45,112,0}},
/* 23218 */ {(12<<2)|2,{45,112,0}},
/* 23219 */ {(12<<2)|2,{45,112,0}},
/* 23220 */ {(12<<2)|2,{45,112,0}},
/* 23221 */ {(12<<2)|2,{45,112,0}},
/* 23222 */ {(12<<2)|2,{45,112,0}},
/* 23223 */ {(12<<2)|2,{45,112,0}},
/* 23224 */ {(12<<2)|2,{45,112,0}},
/* 23225 */ {(12<<2)|2,{45,112,0}},
/* 23226 */ {(12<<2)|2,{45,112,0}},
/* 23227 */ {(12<<2)|2,{45,112,0}},
/* 23228 */ {(12<<2)|2,{45,112,0}},
/* 23229 */ {(12<<2)|2,{45,112,0}},
/* 23230 */ {(12<<2)|2,{45,112,0}},
/* 23231 */ {(12<<2)|2,{45,112,0}},
/* 23232 */ {(12<<2)|2,{45,114,0}},
/* 23233 */ {(12<<2)|2,{45,114,0}},
/* 23234 */ {(12<<2)|2,{45,114,0}},
/* 23235 */ {(12<<2)|2,{45,114,0}},
/* 23236 */ {(12<<2)|2,{45,114,0}},
/* 23237 */ {(12<<2)|2,{45,114,0}},
/* 23238 */ {(12<<2)|2,{45,114,0}},
/* 23239 */ {(12<<2)|2,{45,114,0}},
/* 23240 */ {(12<<2)|2,{45,114,0}},
/* 23241 */ {(12<<2)|2,{45,114,0}},
/* 23242 */ {(12<<2)|2,{45,114,0}},
/* 23243 */ {(12<<2)|2,{45,114,0}},
/* 23244 */ {(12<<2)|2,{45,114,0}},
/* 23245 */ {(12<<2)|2,{45,114,0}},
/* 23246 */ {(12<<2)|2,{45,114,0}},
/* 23247 */ {(12<<2)|2,{45,114,0}},
/* 23248 */ {(12<<2)|2,{45,117,0}},
/* 23249 */ {(12<<2)|2,{45,117,0}},
/* 23250 */ {(12<<2)|2,{45,117,0}},
/* 23251 */ {(12<<2)|2,{45,117,0}},
/* 23252 */ {(12<<2)|2,{45,117,0}},
/* 23253 */ {(12<<2)|2,{45,117,0}},
/* 23254 */ {(12<<2)|2,{45,117,0}},
/* 23255 */ {(12<<2)|2,{45,117,0}},
/* 23256 */ {(12<<2)|2,{45,117,0}},
/* 23257 */ {(12<<2)|2,{45,117,0}},
/* 23258 */ {(12<<2)|2,{45,117,0}},
/* 23259 */ {(12<<2)|2,{45,117,0}},
/* 23260 */ {(12<<2)|2,{45,117,0}},
/* 23261 */ {(12<<2)|2,{45,117,0}},
/* 23262 */ {(12<<2)|2,{45,117,0}},
/* 23263 */ {(12<<2)|2,{45,117,0}},
/* 23264 */ {(13<<2)|2,{45,58,0}},
/* 23265 */ {(13<<2)|2,{45,58,0}},
/* 23266 */ {(13<<2)|2,{45,58,0}},
/* 23267 */ {(13<<2)|2,{45,58,0}},
/* 23268 */ {(13<<2)|2,{45,58,0}},
/* 23269 */ {(13<<2)|2,{45,58,0}},
/* 23270 */ {(13<<2)|2,{45,58,0}},
/* 23271 */ {(13<<2)|2,{45,58,0}},
/* 23272 */ {(13<<2)|2,{45,66,0}},
/* 23273 */ {(13<<2)|2,{45,66,0}},
/* 23274 */ {(13<<2)|2,{45,66,0}},
/* 23275 */ {(13<<2)|2,{45,66,0}},
/* 23276 */ {(13<<2)|2,{45,66,0}},
/* 23277 */ {(13<<2)|2,{45,66,0}},
/* 23278 */ {(13<<2)|2,{45,66,0}},
/* 23279 */ {(13<<2)|2,{45,66,0}},
/* 23280 */ {(13<<2)|2,{45,67,0}},
/* 23281 */ {(13<<2)|2,{45,67,0}},
/* 23282 */ {(13<<2)|2,{45,67,0}},
/* 23283 */ {(13<<2)|2,{45,67,0}},
/* 23284 */ {(13<<2)|2,{45,67,0}},
/* 23285 */ {(13<<2)|2,{45,67,0}},
/* 23286 */ {(13<<2)|2,{45,67,0}},
/* 23287 */ {(13<<2)|2,{45,67,0}},
/* 23288 */ {(13<<2)|2,{45,68,0}},
/* 23289 */ {(13<<2)|2,{45,68,0}},
/* 23290 */ {(13<<2)|2,{45,68,0}},
/* 23291 */ {(13<<2)|2,{45,68,0}},
/* 23292 */ {(13<<2)|2,{45,68,0}},
/* 23293 */ {(13<<2)|2,{45,68,0}},
/* 23294 */ {(13<<2)|2,{45,68,0}},
/* 23295 */ {(13<<2)|2,{45,68,0}},
/* 23296 */ {(13<<2)|2,{45,69,0}},
/* 23297 */ {(13<<2)|2,{45,69,0}},
/* 23298 */ {(13<<2)|2,{45,69,0}},
/* 23299 */ {(13<<2)|2,{45,69,0}},
/* 23300 */ {(13<<2)|2,{45,69,0}},
/* 23301 */ {(13<<2)|2,{45,69,0}},
/* 23302 */ {(13<<2)|2,{45,69,0}},
/* 23303 */ {(13<<2)|2,{45,69,0}},
/* 23304 */ {(13<<2)|2,{45,70,0}},
/* 23305 */ {(13<<2)|2,{45,70,0}},
/* 23306 */ {(13<<2)|2,{45,70,0}},
/* 23307 */ {(13<<2)|2,{45,70,0}},
/* 23308 */ {(13<<2)|2,{45,70,0}},
/* 23309 */ {(13<<2)|2,{45,70,0}},
/* 23310 */ {(13<<2)|2,{45,70,0}},
/* 23311 */ {(13<<2)|2,{45,70,0}},
/* 23312 */ {(13<<2)|2,{45,71,0}},
/* 23313 */ {(13<<2)|2,{45,71,0}},
/* 23314 */ {(13<<2)|2,{45,71,0}},
/* 23315 */ {(13<<2)|2,{45,71,0}},
/* 23316 */ {(13<<2)|2,{45,71,0}},
/* 23317 */ {(13<<2)|2,{45,71,0}},
/* 23318 */ {(13<<2)|2,{45,71,0}},
/* 23319 */ {(13<<2)|2,{45,71,0}},
/* 23320 */ {(13<<2)|2,{45,72,0}},
/* 23321 */ {(13<<2)|2,{45,72,0}},
/* 23322 */ {(13<<2)|2,{45,72,0}},
/* 23323 */ {(13<<2)|2,{45,72,0}},
/* 23324 */ {(13<<2)|2,{45,72,0}},
/* 23325 */ {(13<<2)|2,{45,72,0}},
/* 23326 */ {(13<<2)|2,{45,72,0}},
/* 23327 */ {(13<<2)|2,{45,72,0}},
/* 23328 */ {(13<<2)|2,{45,73,0}},
/* 23329 */ {(13<<2)|2,{45,73,0}},
/* 23330 */ {(13<<2)|2,{45,73,0}},
/* 23331 */ {(13<<2)|2,{45,73,0}},
/* 23332 */ {(13<<2)|2,{45,73,0}},
/* 23333 */ {(13<<2)|2,{45,73,0}},
/* 23334 */ {(13<<2)|2,{45,73,0}},
/* 23335 */ {(13<<2)|2,{45,73,0}},
/* 23336 */ {(13<<2)|2,{45,74,0}},
/* 23337 */ {(13<<2)|2,{45,74,0}},
/* 23338 */ {(13<<2)|2,{45,74,0}},
/* 23339 */ {(13<<2)|2,{45,74,0}},
/* 23340 */ {(13<<2)|2,{45,74,0}},
/* 23341 */ {(13<<2)|2,{45,74,0}},
/* 23342 */ {(13<<2)|2,{45,74,0}},
/* 23343 */ {(13<<2)|2,{45,74,0}},
/* 23344 */ {(13<<2)|2,{45,75,0}},
/* 23345 */ {(13<<2)|2,{45,75,0}},
/* 23346 */ {(13<<2)|2,{45,75,0}},
/* 23347 */ {(13<<2)|2,{45,75,0}},
/* 23348 */ {(13<<2)|2,{45,75,0}},
/* 23349 */ {(13<<2)|2,{45,75,0}},
/* 23350 */ {(13<<2)|2,{45,75,0}},
/* 23351 */ {(13<<2)|2,{45,75,0}},
/* 23352 */ {(13<<2)|2,{45,76,0}},
/* 23353 */ {(13<<2)|2,{45,76,0}},
/* 23354 */ {(13<<2)|2,{45,76,0}},
/* 23355 */ {(13<<2)|2,{45,76,0}},
/* 23356 */ {(13<<2)|2,{45,76,0}},
/* 23357 */ {(13<<2)|2,{45,76,0}},
/* 23358 */ {(13<<2)|2,{45,76,0}},
/* 23359 */ {(13<<2)|2,{45,76,0}},
/* 23360 */ {(13<<2)|2,{45,77,0}},
/* 23361 */ {(13<<2)|2,{45,77,0}},
/* 23362 */ {(13<<2)|2,{45,77,0}},
/* 23363 */ {(13<<2)|2,{45,77,0}},
/* 23364 */ {(13<<2)|2,{45,77,0}},
/* 23365 */ {(13<<2)|2,{45,77,0}},
/* 23366 */ {(13<<2)|2,{45,77,0}},
/* 23367 */ {(13<<2)|2,{45,77,0}},
/* 23368 */ {(13<<2)|2,{45,78,0}},
/* 23369 */ {(13<<2)|2,{45,78,0}},
/* 23370 */ {(13<<2)|2,{45,78,0}},
/* 23371 */ {(13<<2)|2,{45,78,0}},
/* 23372 */ {(13<<2)|2,{45,78,0}},
/* 23373 */ {(13<<2)|2,{45,78,0}},
/* 23374 */ {(13<<2)|2,{45,78,0}},
/* 23375 */ {(13<<2)|2,{45,78,0}},
/* 23376 */ {(13<<2)|2,{45,79,0}},
/* 23377 */ {(13<<2)|2,{45,79,0}},
/* 23378 */ {(13<<2)|2,{45,79,0}},
/* 23379 */ {(13<<2)|2,{45,79,0}},
/* 23380 */ {(13<<2)|2,{45,79,0}},
/* 23381 */ {(13<<2)|2,{45,79,0}},
/* 23382 */ {(13<<2)|2,{45,79,0}},
/* 23383 */ {(13<<2)|2,{45,79,0}},
/* 23384 */ {(13<<2)|2,{45,80,0}},
/* 23385 */ {(13<<2)|2,{45,80,0}},
/* 23386 */ {(13<<2)|2,{45,80,0}},
/* 23387 */ {(13<<2)|2,{45,80,0}},
/* 23388 */ {(13<<2)|2,{45,80,0}},
/* 23389 */ {(13<<2)|2,{45,80,0}},
/* 23390 */ {(13<<2)|2,{45,80,0}},
/* 23391 */ {(13<<2)|2,{45,80,0}},
/* 23392 */ {(13<<2)|2,{45,81,0}},
/* 23393 */ {(13<<2)|2,{45,81,0}},
/* 23394 */ {(13<<2)|2,{45,81,0}},
/* 23395 */ {(13<<2)|2,{45,81,0}},
/* 23396 */ {(13<<2)|2,{45,81,0}},
/* 23397 */ {(13<<2)|2,{45,81,0}},
/* 23398 */ {(13<<2)|2,{45,81,0}},
/* 23399 */ {(13<<2)|2,{45,81,0}},
/* 23400 */ {(13<<2)|2,{45,82,0}},
/* 23401 */ {(13<<2)|2,{45,82,0}},
/* 23402 */ {(13<<2)|2,{45,82,0}},
/* 23403 */ {(13<<2)|2,{45,82,0}},
/* 23404 */ {(13<<2)|2,{45,82,0}},
/* 23405 */ {(13<<2)|2,{45,82,0}},
/* 23406 */ {(13<<2)|2,{45,82,0}},
/* 23407 */ {(13<<2)|2,{45,82,0}},
/* 23408 */ {(13<<2)|2,{45,83,0}},
/* 23409 */ {(13<<2)|2,{45,83,0}},
/* 23410 */ {(13<<2)|2,{45,83,0}},
/* 23411 */ {(13<<2)|2,{45,83,0}},
/* 23412 */ {(13<<2)|2,{45,83,0}},
/* 23413 */ {(13<<2)|2,{45,83,0}},
/* 23414 */ {(13<<2)|2,{45,83,0}},
/* 23415 */ {(13<<2)|2,{45,83,0}},
/* 23416 */ {(13<<2)|2,{45,84,0}},
/* 23417 */ {(13<<2)|2,{45,84,0}},
/* 23418 */ {(13<<2)|2,{45,84,0}},
/* 23419 */ {(13<<2)|2,{45,84,0}},
/* 23420 */ {(13<<2)|2,{45,84,0}},
/* 23421 */ {(13<<2)|2,{45,84,0}},
/* 23422 */ {(13<<2)|2,{45,84,0}},
/* 23423 */ {(13<<2)|2,{45,84,0}},
/* 23424 */ {(13<<2)|2,{45,85,0}},
/* 23425 */ {(13<<2)|2,{45,85,0}},
/* 23426 */ {(13<<2)|2,{45,85,0}},
/* 23427 */ {(13<<2)|2,{45,85,0}},
/* 23428 */ {(13<<2)|2,{45,85,0}},
/* 23429 */ {(13<<2)|2,{45,85,0}},
/* 23430 */ {(13<<2)|2,{45,85,0}},
/* 23431 */ {(13<<2)|2,{45,85,0}},
/* 23432 */ {(13<<2)|2,{45,86,0}},
/* 23433 */ {(13<<2)|2,{45,86,0}},
/* 23434 */ {(13<<2)|2,{45,86,0}},
/* 23435 */ {(13<<2)|2,{45,86,0}},
/* 23436 */ {(13<<2)|2,{45,86,0}},
/* 23437 */ {(13<<2)|2,{45,86,0}},
/* 23438 */ {(13<<2)|2,{45,86,0}},
/* 23439 */ {(13<<2)|2,{45,86,0}},
/* 23440 */ {(13<<2)|2,{45,87,0}},
/* 23441 */ {(13<<2)|2,{45,87,0}},
/* 23442 */ {(13<<2)|2,{45,87,0}},
/* 23443 */ {(13<<2)|2,{45,87,0}},
/* 23444 */ {(13<<2)|2,{45,87,0}},
/* 23445 */ {(13<<2)|2,{45,87,0}},
/* 23446 */ {(13<<2)|2,{45,87,0}},
/* 23447 */ {(13<<2)|2,{45,87,0}},
/* 23448 */ {(13<<2)|2,{45,89,0}},
/* 23449 */ {(13<<2)|2,{45,89,0}},
/* 23450 */ {(13<<2)|2,{45,89,0}},
/* 23451 */ {(13<<2)|2,{45,89,0}},
/* 23452 */ {(13<<2)|2,{45,89,0}},
/* 23453 */ {(13<<2)|2,{45,89,0}},
/* 23454 */ {(13<<2)|2,{45,89,0}},
/* 23455 */ {(13<<2)|2,{45,89,0}},
/* 23456 */ {(13<<2)|2,{45,106,0}},
/* 23457 */ {(13<<2)|2,{45,106,0}},
/* 23458 */ {(13<<2)|2,{45,106,0}},
/* 23459 */ {(13<<2)|2,{45,106,0}},
/* 23460 */ {(13<<2)|2,{45,106,0}},
/* 23461 */ {(13<<2)|2,{45,106,0}},
/* 23462 */ {(13<<2)|2,{45,106,0}},
/* 23463 */ {(13<<2)|2,{45,106,0}},
/* 23464 */ {(13<<2)|2,{45,107,0}},
/* 23465 */ {(13<<2)|2,{45,107,0}},
/* 23466 */ {(13<<2)|2,{45,107,0}},
/* 23467 */ {(13<<2)|2,{45,107,0}},
/* 23468 */ {(13<<2)|2,{45,107,0}},
/* 23469 */ {(13<<2)|2,{45,107,0}},
/* 23470 */ {(13<<2)|2,{45,107,0}},
/* 23471 */ {(13<<2)|2,{45,107,0}},
/* 23472 */ {(13<<2)|2,{45,113,0}},
/* 23473 */ {(13<<2)|2,{45,113,0}},
/* 23474 */ {(13<<2)|2,{45,113,0}},
/* 23475 */ {(13<<2)|2,{45,113,0}},
/* 23476 */ {(13<<2)|2,{45,113,0}},
/* 23477 */ {(13<<2)|2,{45,113,0}},
/* 23478 */ {(13<<2)|2,{45,113,0}},
/* 23479 */ {(13<<2)|2,{45,113,0}},
/* 23480 */ {(13<<2)|2,{45,118,0}},
/* 23481 */ {(13<<2)|2,{45,118,0}},
/* 23482 */ {(13<<2)|2,{45,118,0}},
/* 23483 */ {(13<<2)|2,{45,118,0}},
/* 23484 */ {(13<<2)|2,{45,118,0}},
/* 23485 */ {(13<<2)|2,{45,118,0}},
/* 23486 */ {(13<<2)|2,{45,118,0}},
/* 23487 */ {(13<<2)|2,{45,118,0}},
/* 23488 */ {(13<<2)|2,{45,119,0}},
/* 23489 */ {(13<<2)|2,{45,119,0}},
/* 23490 */ {(13<<2)|2,{45,119,0}},
/* 23491 */ {(13<<2)|2,{45,119,0}},
/* 23492 */ {(13<<2)|2,{45,119,0}},
/* 23493 */ {(13<<2)|2,{45,119,0}},
/* 23494 */ {(13<<2)|2,{45,119,0}},
/* 23495 */ {(13<<2)|2,{45,119,0}},
/* 23496 */ {(13<<2)|2,{45,120,0}},
/* 23497 */ {(13<<2)|2,{45,120,0}},
/* 23498 */ {(13<<2)|2,{45,120,0}},
/* 23499 */ {(13<<2)|2,{45,120,0}},
/* 23500 */ {(13<<2)|2,{45,120,0}},
/* 23501 */ {(13<<2)|2,{45,120,0}},
/* 23502 */ {(13<<2)|2,{45,120,0}},
/* 23503 */ {(13<<2)|2,{45,120,0}},
/* 23504 */ {(13<<2)|2,{45,121,0}},
/* 23505 */ {(13<<2)|2,{45,121,0}},
/* 23506 */ {(13<<2)|2,{45,121,0}},
/* 23507 */ {(13<<2)|2,{45,121,0}},
/* 23508 */ {(13<<2)|2,{45,121,0}},
/* 23509 */ {(13<<2)|2,{45,121,0}},
/* 23510 */ {(13<<2)|2,{45,121,0}},
/* 23511 */ {(13<<2)|2,{45,121,0}},
/* 23512 */ {(13<<2)|2,{45,122,0}},
/* 23513 */ {(13<<2)|2,{45,122,0}},
/* 23514 */ {(13<<2)|2,{45,122,0}},
/* 23515 */ {(13<<2)|2,{45,122,0}},
/* 23516 */ {(13<<2)|2,{45,122,0}},
/* 23517 */ {(13<<2)|2,{45,122,0}},
/* 23518 */ {(13<<2)|2,{45,122,0}},
/* 23519 */ {(13<<2)|2,{45,122,0}},
/* 23520 */ {(14<<2)|2,{45,38,0}},
/* 23521 */ {(14<<2)|2,{45,38,0}},
/* 23522 */ {(14<<2)|2,{45,38,0}},
/* 23523 */ {(14<<2)|2,{45,38,0}},
/* 23524 */ {(14<<2)|2,{45,42,0}},
/* 23525 */ {(14<<2)|2,{45,42,0}},
/* 23526 */ {(14<<2)|2,{45,42,0}},
/* 23527 */ {(14<<2)|2,{45,42,0}},
/* 23528 */ {(14<<2)|2,{45,44,0}},
/* 23529 */ {(14<<2)|2,{45,44,0}},
/* 23530 */ {(14<<2)|2,{45,44,0}},
/* 23531 */ {(14<<2)|2,{45,44,0}},
/* 23532 */ {(14<<2)|2,{45,59,0}},
/* 23533 */ {(14<<2)|2,{45,59,0}},
/* 23534 */ {(14<<2)|2,{45,59,0}},
/* 23535 */ {(14<<2)|2,{45,59,0}},
/* 23536 */ {(14<<2)|2,{45,88,0}},
/* 23537 */ {(14<<2)|2,{45,88,0}},
/* 23538 */ {(14<<2)|2,{45,88,0}},
/* 23539 */ {(14<<2)|2,{45,88,0}},
/* 23540 */ {(14<<2)|2,{45,90,0}},
/* 23541 */ {(14<<2)|2,{45,90,0}},
/* 23542 */ {(14<<2)|2,{45,90,0}},
/* 23543 */ {(14<<2)|2,{45,90,0}},
/* 23544 */ {(16<<2)|2,{45,33,0}},
/* 23545 */ {(16<<2)|2,{45,34,0}},
/* 23546 */ {(16<<2)|2,{45,40,0}},
/* 23547 */ {(16<<2)|2,{45,41,0}},
/* 23548 */ {(16<<2)|2,{45,63,0}},
/* 23549 */ {(6<<2)|1,{45,0,0}},
/* 23550 */ {(6<<2)|1,{45,0,0}},
/* 23551 */ {(6<<2)|1,{45,0,0}},
/* 23552 */ {(16<<2)|3,{46,48,48}},
/* 23553 */ {(16<<2)|3,{46,48,49}},
/* 23554 */ {(16<<2)|3,{46,48,50}},
/* 23555 */ {(16<<2)|3,{46,48,97}},
/* 23556 */ {(16<<2)|3,{46,48,99}},
/* 23557 */ {(16<<2)|3,{46,48,101}},
/* 23558 */ {(16<<2)|3,{46,48,105}},
/* 23559 */ {(16<<2)|3,{46,48,111}},
/* 23560 */ {(16<<2)|3,{46,48,115}},
/* 23561 */ {(16<<2)|3,{46,48,116}},
/* 23562 */ {(11<<2)|2,{46,48,0}},
/* 23563 */ {(11<<2)|2,{46,48,0}},
/* 23564 */ {(11<<2)|2,{46,48,0}},
/* 23565 */ {(11<<2)|2,{46,48,0}},
/* 23566 */ {(11<<2)|2,{46,48,0}},
/* 23567 */ {(11<<2)|2,{46,48,0}},
/* 23568 */ {(11<<2)|2,{46,48,0}},
/* 23569 */ {(11<<2)|2,{46,48,0}},
/* 23570 */ {(11<<2)|2,{46,48,0}},
/* 23571 */ {(11<<2)|2,{46,48,0}},
/* 23572 */ {(11<<2)|2,{46,48,0}},
/* 23573 */ {(11<<2)|2,{46,48,0}},
/* 23574 */ {(11<<2)|2,{46,48,0}},
/* 23575 */ {(11<<2)|2,{46,48,0}},
/* 23576 */ {(11<<2)|2,{46,48,0}},
/* 23577 */ {(11<<2)|2,{46,48,0}},
/* 23578 */ {(11<<2)|2,{46,48,0}},
/* 23579 */ {(11<<2)|2,{46,48,0}},
/* 23580 */ {(11<<2)|2,{46,48,0}},
/* 23581 */ {(11<<2)|2,{46,48,0}},
/* 23582 */ {(11<<2)|2,{46,48,0}},
/* 23583 */ {(11<<2)|2,{46,48,0}},
/* 23584 */ {(16<<2)|3,{46,49,48}},
/* 23585 */ {(16<<2)|3,{46,49,49}},
/* 23586 */ {(16<<2)|3,{46,49,50}},
/* 23587 */ {(16<<2)|3,{46,49,97}},
/* 23588 */ {(16<<2)|3,{46,49,99}},
/* 23589 */ {(16<<2)|3,{46,49,101}},
/* 23590 */ {(16<<2)|3,{46,49,105}},
/* 23591 */ {(16<<2)|3,{46,49,111}},
/* 23592 */ {(16<<2)|3,{46,49,115}},
/* 23593 */ {(16<<2)|3,{46,49,116}},
/* 23594 */ {(11<<2)|2,{46,49,0}},
/* 23595 */ {(11<<2)|2,{46,49,0}},
/* 23596 */ {(11<<2)|2,{46,49,0}},
/* 23597 */ {(11<<2)|2,{46,49,0}},
/* 23598 */ {(11<<2)|2,{46,49,0}},
/* 23599 */ {(11<<2)|2,{46,49,0}},
/* 23600 */ {(11<<2)|2,{46,49,0}},
/* 23601 */ {(11<<2)|2,{46,49,0}},
/* 23602 */ {(11<<2)|2,{46,49,0}},
/* 23603 */ {(11<<2)|2,{46,49,0}},
/* 23604 */ {(11<<2)|2,{46,49,0}},
/* 23605 */ {(11<<2)|2,{46,49,0}},
/* 23606 */ {(11<<2)|2,{46,49,0}},
/* 23607 */ {(11<<2)|2,{46,49,0}},
/* 23608 */ {(11<<2)|2,{46,49,0}},
/* 23609 */ {(11<<2)|2,{46,49,0}},
/* 23610 */ {(11<<2)|2,{46,49,0}},
/* 23611 */ {(11<<2)|2,{46,49,0}},
/* 23612 */ {(11<<2)|2,{46,49,0}},
/* 23613 */ {(11<<2)|2,{46,49,0}},
/* 23614 */ {(11<<2)|2,{46,49,0}},
/* 23615 */ {(11<<2)|2,{46,49,0}},
/* 23616 */ {(16<<2)|3,{46,50,48}},
/* 23617 */ {(16<<2)|3,{46,50,49}},
/* 23618 */ {(16<<2)|3,{46,50,50}},
/* 23619 */ {(16<<2)|3,{46,50,97}},
/* 23620 */ {(16<<2)|3,{46,50,99}},
/* 23621 */ {(16<<2)|3,{46,50,101}},
/* 23622 */ {(16<<2)|3,{46,50,105}},
/* 23623 */ {(16<<2)|3,{46,50,111}},
/* 23624 */ {(16<<2)|3,{46,50,115}},
/* 23625 */ {(16<<2)|3,{46,50,116}},
/* 23626 */ {(11<<2)|2,{46,50,0}},
/* 23627 */ {(11<<2)|2,{46,50,0}},
/* 23628 */ {(11<<2)|2,{46,50,0}},
/* 23629 */ {(11<<2)|2,{46,50,0}},
/* 23630 */ {(11<<2)|2,{46,50,0}},
/* 23631 */ {(11<<2)|2,{46,50,0}},
/* 23632 */ {(11<<2)|2,{46,50,0}},
/* 23633 */ {(11<<2)|2,{46,50,0}},
/* 23634 */ {(11<<2)|2,{46,50,0}},
/* 23635 */ {(11<<2)|2,{46,50,0}},
/* 23636 */ {(11<<2)|2,{46,50,0}},
/* 23637 */ {(11<<2)|2,{46,50,0}},
/* 23638 */ {(11<<2)|2,{46,50,0}},
/* 23639 */ {(11<<2)|2,{46,50,0}},
/* 23640 */ {(11<<2)|2,{46,50,0}},
/* 23641 */ {(11<<2)|2,{46,50,0}},
/* 23642 */ {(11<<2)|2,{46,50,0}},
/* 23643 */ {(11<<2)|2,{46,50,0}},
/* 23644 */ {(11<<2)|2,{46,50,0}},
/* 23645 */ {(11<<2)|2,{46,50,0}},
/* 23646 */ {(11<<2)|2,{46,50,0}},
/* 23647 */ {(11<<2)|2,{46,50,0}},
/* 23648 */ {(16<<2)|3,{46,97,48}},
/* 23649 */ {(16<<2)|3,{46,97,49}},
/* 23650 */ {(16<<2)|3,{46,97,50}},
/* 23651 */ {(16<<2)|3,{46,97,97}},
/* 23652 */ {(16<<2)|3,{46,97,99}},
/* 23653 */ {(16<<2)|3,{46,97,101}},
/* 23654 */ {(16<<2)|3,{46,97,105}},
/* 23655 */ {(16<<2)|3,{46,97,111}},
/* 23656 */ {(16<<2)|3,{46,97,115}},
/* 23657 */ {(16<<2)|3,{46,97,116}},
/* 23658 */ {(11<<2)|2,{46,97,0}},
/* 23659 */ {(11<<2)|2,{46,97,0}},
/* 23660 */ {(11<<2)|2,{46,97,0}},
/* 23661 */ {(11<<2)|2,{46,97,0}},
/* 23662 */ {(11<<2)|2,{46,97,0}},
/* 23663 */ {(11<<2)|2,{46,97,0}},
/* 23664 */ {(11<<2)|2,{46,97,0}},
/* 23665 */ {(11<<2)|2,{46,97,0}},
/* 23666 */ {(11<<2)|2,{46,97,0}},
/* 23667 */ {(11<<2)|2,{46,97,0}},
/* 23668 */ {(11<<2)|2,{46,97,0}},
/* 23669 */ {(11<<2)|2,{46,97,0}},
/* 23670 */ {(11<<2)|2,{46,97,0}},
/* 23671 */ {(11<<2)|2,{46,97,0}},
/* 23672 */ {(11<<2)|2,{46,97,0}},
/* 23673 */ {(11<<2)|2,{46,97,0}},
/* 23674 */ {(11<<2)|2,{46,97,0}},
/* 23675 */ {(11<<2)|2,{46,97,0}},
/* 23676 */ {(11<<2)|2,{46,97,0}},
/* 23677 */ {(11<<2)|2,{46,97,0}},
/* 23678 */ {(11<<2)|2,{46,97,0}},
/* 23679 */ {(11<<2)|2,{46,97,0}},
/* 23680 */ {(16<<2)|3,{46,99,48}},
/* 23681 */ {(16<<2)|3,{46,99,49}},
/* 23682 */ {(16<<2)|3,{46,99,50}},
/* 23683 */ {(16<<2)|3,{46,99,97}},
/* 23684 */ {(16<<2)|3,{46,99,99}},
/* 23685 */ {(16<<2)|3,{46,99,101}},
/* 23686 */ {(16<<2)|3,{46,99,105}},
/* 23687 */ {(16<<2)|3,{46,99,111}},
/* 23688 */ {(16<<2)|3,{46,99,115}},
/* 23689 */ {(16<<2)|3,{46,99,116}},
/* 23690 */ {(11<<2)|2,{46,99,0}},
/* 23691 */ {(11<<2)|2,{46,99,0}},
/* 23692 */ {(11<<2)|2,{46,99,0}},
/* 23693 */ {(11<<2)|2,{46,99,0}},
/* 23694 */ {(11<<2)|2,{46,99,0}},
/* 23695 */ {(11<<2)|2,{46,99,0}},
/* 23696 */ {(11<<2)|2,{46,99,0}},
/* 23697 */ {(11<<2)|2,{46,99,0}},
/* 23698 */ {(11<<2)|2,{46,99,0}},
/* 23699 */ {(11<<2)|2,{46,99,0}},
/* 23700 */ {(11<<2)|2,{46,99,0}},
/* 23701 */ {(11<<2)|2,{46,99,0}},
/* 23702 */ {(11<<2)|2,{46,99,0}},
/* 23703 */ {(11<<2)|2,{46,99,0}},
/* 23704 */ {(11<<2)|2,{46,99,0}},
/* 23705 */ {(11<<2)|2,{46,99,0}},
/* 23706 */ {(11<<2)|2,{46,99,0}},
/* 23707 */ {(11<<2)|2,{46,99,0}},
/* 23708 */ {(11<<2)|2,{46,99,0}},
/* 23709 */ {(11<<2)|2,{46,99,0}},
/* 23710 */ {(11<<2)|2,{46,99,0}},
/* 23711 */ {(11<<2)|2,{46,99,0}},
/* 23712 */ {(16<<2)|3,{46,101,48}},
/* 23713 */ {(16<<2)|3,{46,101,49}},
/* 23714 */ {(16<<2)|3,{46,101,50}},
/* 23715 */ {(16<<2)|3,{46,101,97}},
/* 23716 */ {(16<<2)|3,{46,101,99}},
/* 23717 */ {(16<<2)|3,{46,101,101}},
/* 23718 */ {(16<<2)|3,{46,101,105}},
/* 23719 */ {(16<<2)|3,{46,101,111}},
/* 23720 */ {(16<<2)|3,{46,101,115}},
/* 23721 */ {(16<<2)|3,{46,101,116}},
/* 23722 */ {(11<<2)|2,{46,101,0}},
/* 23723 */ {(11<<2)|2,{46,101,0}},
/* 23724 */ {(11<<2)|2,{46,101,0}},
/* 23725 */ {(11<<2)|2,{46,101,0}},
/* 23726 */ {(11<<2)|2,{46,101,0}},
/* 23727 */ {(11<<2)|2,{46,101,0}},
/* 23728 */ {(11<<2)|2,{46,101,0}},
/* 23729 */ {(11<<2)|2,{46,101,0}},
/* 23730 */ {(11<<2)|2,{46,101,0}},
/* 23731 */ {(11<<2)|2,{46,101,0}},
/* 23732 */ {(11<<2)|2,{46,101,0}},
/* 23733 */ {(11<<2)|2,{46,101,0}},
/* 23734 */ {(11<<2)|2,{46,101,0}},
/* 23735 */ {(11<<2)|2,{46,101,0}},
/* 23736 */ {(11<<2)|2,{46,101,0}},
/* 23737 */ {(11<<2)|2,{46,101,0}},
/* 23738 */ {(11<<2)|2,{46,101,0}},
/* 23739 */ {(11<<2)|2,{46,101,0}},
/* 23740 */ {(11<<2)|2,{46,101,0}},
/* 23741 */ {(11<<2)|2,{46,101,0}},
/* 23742 */ {(11<<2)|2,{46,101,0}},
/* 23743 */ {(11<<2)|2,{46,101,0}},
/* 23744 */ {(16<<2)|3,{46,105,48}},
/* 23745 */ {(16<<2)|3,{46,105,49}},
/* 23746 */ {(16<<2)|3,{46,105,50}},
/* 23747 */ {(16<<2)|3,{46,105,97}},
/* 23748 */ {(16<<2)|3,{46,105,99}},
/* 23749 */ {(16<<2)|3,{46,105,101}},
/* 23750 */ {(16<<2)|3,{46,105,105}},
/* 23751 */ {(16<<2)|3,{46,105,111}},
/* 23752 */ {(16<<2)|3,{46,105,115}},
/* 23753 */ {(16<<2)|3,{46,105,116}},
/* 23754 */ {(11<<2)|2,{46,105,0}},
/* 23755 */ {(11<<2)|2,{46,105,0}},
/* 23756 */ {(11<<2)|2,{46,105,0}},
/* 23757 */ {(11<<2)|2,{46,105,0}},
/* 23758 */ {(11<<2)|2,{46,105,0}},
/* 23759 */ {(11<<2)|2,{46,105,0}},
/* 23760 */ {(11<<2)|2,{46,105,0}},
/* 23761 */ {(11<<2)|2,{46,105,0}},
/* 23762 */ {(11<<2)|2,{46,105,0}},
/* 23763 */ {(11<<2)|2,{46,105,0}},
/* 23764 */ {(11<<2)|2,{46,105,0}},
/* 23765 */ {(11<<2)|2,{46,105,0}},
/* 23766 */ {(11<<2)|2,{46,105,0}},
/* 23767 */ {(11<<2)|2,{46,105,0}},
/* 23768 */ {(11<<2)|2,{46,105,0}},
/* 23769 */ {(11<<2)|2,{46,105,0}},
/* 23770 */ {(11<<2)|2,{46,105,0}},
/* 23771 */ {(11<<2)|2,{46,105,0}},
/* 23772 */ {(11<<2)|2,{46,105,0}},
/* 23773 */ {(11<<2)|2,{46,105,0}},
/* 23774 */ {(11<<2)|2,{46,105,0}},
/* 23775 */ {(11<<2)|2,{46,105,0}},
/* 23776 */ {(16<<2)|3,{46,111,48}},
/* 23777 */ {(16<<2)|3,{46,111,49}},
/* 23778 */ {(16<<2)|3,{46,111,50}},
/* 23779 */ {(16<<2)|3,{46,111,97}},
/* 23780 */ {(16<<2)|3,{46,111,99}},
/* 23781 */ {(16<<2)|3,{46,111,101}},
/* 23782 */ {(16<<2)|3,{46,111,105}},
/* 23783 */ {(16<<2)|3,{46,111,111}},
/* 23784 */ {(16<<2)|3,{46,111,115}},
/* 23785 */ {(16<<2)|3,{46,111,116}},
/* 23786 */ {(11<<2)|2,{46,111,0}},
/* 23787 */ {(11<<2)|2,{46,111,0}},
/* 23788 */ {(11<<2)|2,{46,111,0}},
/* 23789 */ {(11<<2)|2,{46,111,0}},
/* 23790 */ {(11<<2)|2,{46,111,0}},
/* 23791 */ {(11<<2)|2,{46,111,0}},
/* 23792 */ {(11<<2)|2,{46,111,0}},
/* 23793 */ {(11<<2)|2,{46,111,0}},
/* 23794 */ {(11<<2)|2,{46,111,0}},
/* 23795 */ {(11<<2)|2,{46,111,0}},
/* 23796 */ {(11<<2)|2,{46,111,0}},
/* 23797 */ {(11<<2)|2,{46,111,0}},
/* 23798 */ {(11<<2)|2,{46,111,0}},
/* 23799 */ {(11<<2)|2,{46,111,0}},
/* 23800 */ {(11<<2)|2,{46,111,0}},
/* 23801 */ {(11<<2)|2,{46,111,0}},
/* 23802 */ {(11<<2)|2,{46,111,0}},
/* 23803 */ {(11<<2)|2,{46,111,0}},
/* 23804 */ {(11<<2)|2,{46,111,0}},
/* 23805 */ {(11<<2)|2,{46,111,0}},
/* 23806 */ {(11<<2)|2,{46,111,0}},
/* 23807 */ {(11<<2)|2,{46,111,0}},
/* 23808 */ {(16<<2)|3,{46,115,48}},
/* 23809 */ {(16<<2)|3,{46,115,49}},
/* 23810 */ {(16<<2)|3,{46,115,50}},
/* 23811 */ {(16<<2)|3,{46,115,97}},
/* 23812 */ {(16<<2)|3,{46,115,99}},
/* 23813 */ {(16<<2)|3,{46,115,101}},
/* 23814 */ {(16<<2)|3,{46,115,105}},
/* 23815 */ {(16<<2)|3,{46,115,111}},
/* 23816 */ {(16<<2)|3,{46,115,115}},
/* 23817 */ {(16<<2)|3,{46,115,116}},
/* 23818 */ {(11<<2)|2,{46,115,0}},
/* 23819 */ {(11<<2)|2,{46,115,0}},
/* 23820 */ {(11<<2)|2,{46,115,0}},
/* 23821 */ {(11<<2)|2,{46,115,0}},
/* 23822 */ {(11<<2)|2,{46,115,0}},
/* 23823 */ {(11<<2)|2,{46,115,0}},
/* 23824 */ {(11<<2)|2,{46,115,0}},
/* 23825 */ {(11<<2)|2,{46,115,0}},
/* 23826 */ {(11<<2)|2,{46,115,0}},
/* 23827 */ {(11<<2)|2,{46,115,0}},
/* 23828 */ {(11<<2)|2,{46,115,0}},
/* 23829 */ {(11<<2)|2,{46,115,0}},
/* 23830 */ {(11<<2)|2,{46,115,0}},
/* 23831 */ {(11<<2)|2,{46,115,0}},
/* 23832 */ {(11<<2)|2,{46,115,0}},
/* 23833 */ {(11<<2)|2,{46,115,0}},
/* 23834 */ {(11<<2)|2,{46,115,0}},
/* 23835 */ {(11<<2)|2,{46,115,0}},
/* 23836 */ {(11<<2)|2,{46,115,0}},
/* 23837 */ {(11<<2)|2,{46,115,0}},
/* 23838 */ {(11<<2)|2,{46,115,0}},
/* 23839 */ {(11<<2)|2,{46,115,0}},
/* 23840 */ {(16<<2)|3,{46,116,48}},
/* 23841 */ {(16<<2)|3,{46,116,49}},
/* 23842 */ {(16<<2)|3,{46,116,50}},
/* 23843 */ {(16<<2)|3,{46,116,97}},
/* 23844 */ {(16<<2)|3,{46,116,99}},
/* 23845 */ {(16<<2)|3,{46,116,101}},
/* 23846 */ {(16<<2)|3,{46,116,105}},
/* 23847 */ {(16<<2)|3,{46,116,111}},
/* 23848 */ {(16<<2)|3,{46,116,115}},
/* 23849 */ {(16<<2)|3,{46,116,116}},
/* 23850 */ {(11<<2)|2,{46,116,0}},
/* 23851 */ {(11<<2)|2,{46,116,0}},
/* 23852 */ {(11<<2)|2,{46,116,0}},
/* 23853 */ {(11<<2)|2,{46,116,0}},
/* 23854 */ {(11<<2)|2,{46,116,0}},
/* 23855 */ {(11<<2)|2,{46,116,0}},
/* 23856 */ {(11<<2)|2,{46,116,0}},
/* 23857 */ {(11<<2)|2,{46,116,0}},
/* 23858 */ {(11<<2)|2,{46,116,0}},
/* 23859 */ {(11<<2)|2,{46,116,0}},
/* 23860 */ {(11<<2)|2,{46,116,0}},
/* 23861 */ {(11<<2)|2,{46,116,0}},
/* 23862 */ {(11<<2)|2,{46,116,0}},
/* 23863 */ {(11<<2)|2,{46,116,0}},
/* 23864 */ {(11<<2)|2,{46,116,0}},
/* 23865 */ {(11<<2)|2,{46,116,0}},
/* 23866 */ {(11<<2)|2,{46,116,0}},
/* 23867 */ {(11<<2)|2,{46,116,0}},
/* 23868 */ {(11<<2)|2,{46,116,0}},
/* 23869 */ {(11<<2)|2,{46,116,0}},
/* 23870 */ {(11<<2)|2,{46,116,0}},
/* 23871 */ {(11<<2)|2,{46,116,0}},
/* 23872 */ {(12<<2)|2,{46,32,0}},
/* 23873 */ {(12<<2)|2,{46,32,0}},
/* 23874 */ {(12<<2)|2,{46,32,0}},
/* 23875 */ {(12<<2)|2,{46,32,0}},
/* 23876 */ {(12<<2)|2,{46,32,0}},
/* 23877 */ {(12<<2)|2,{46,32,0}},
/* 23878 */ {(12<<2)|2,{46,32,0}},
/* 23879 */ {(12<<2)|2,{46,32,0}},
/* 23880 */ {(12<<2)|2,{46,32,0}},
/* 23881 */ {(12<<2)|2,{46,32,0}},
/* 23882 */ {(12<<2)|2,{46,32,0}},
/* 23883 */ {(12<<2)|2,{46,32,0}},
/* 23884 */ {(12<<2)|2,{46,32,0}},
/* 23885 */ {(12<<2)|2,{46,32,0}},
/* 23886 */ {(12<<2)|2,{46,32,0}},
/* 23887 */ {(12<<2)|2,{46,32,0}},
/* 23888 */ {(12<<2)|2,{46,37,0}},
/* 23889 */ {(12<<2)|2,{46,37,0}},
/* 23890 */ {(12<<2)|2,{46,37,0}},
/* 23891 */ {(12<<2)|2,{46,37,0}},
/* 23892 */ {(12<<2)|2,{46,37,0}},
/* 23893 */ {(12<<2)|2,{46,37,0}},
/* 23894 */ {(12<<2)|2,{46,37,0}},
/* 23895 */ {(12<<2)|2,{46,37,0}},
/* 23896 */ {(12<<2)|2,{46,37,0}},
/* 23897 */ {(12<<2)|2,{46,37,0}},
/* 23898 */ {(12<<2)|2,{46,37,0}},
/* 23899 */ {(12<<2)|2,{46,37,0}},
/* 23900 */ {(12<<2)|2,{46,37,0}},
/* 23901 */ {(12<<2)|2,{46,37,0}},
/* 23902 */ {(12<<2)|2,{46,37,0}},
/* 23903 */ {(12<<2)|2,{46,37,0}},
/* 23904 */ {(12<<2)|2,{46,45,0}},
/* 23905 */ {(12<<2)|2,{46,45,0}},
/* 23906 */ {(12<<2)|2,{46,45,0}},
/* 23907 */ {(12<<2)|2,{46,45,0}},
/* 23908 */ {(12<<2)|2,{46,45,0}},
/* 23909 */ {(12<<2)|2,{46,45,0}},
/* 23910 */ {(12<<2)|2,{46,45,0}},
/* 23911 */ {(12<<2)|2,{46,45,0}},
/* 23912 */ {(12<<2)|2,{46,45,0}},
/* 23913 */ {(12<<2)|2,{46,45,0}},
/* 23914 */ {(12<<2)|2,{46,45,0}},
/* 23915 */ {(12<<2)|2,{46,45,0}},
/* 23916 */ {(12<<2)|2,{46,45,0}},
/* 23917 */ {(12<<2)|2,{46,45,0}},
/* 23918 */ {(12<<2)|2,{46,45,0}},
/* 23919 */ {(12<<2)|2,{46,45,0}},
/* 23920 */ {(12<<2)|2,{46,46,0}},
/* 23921 */ {(12<<2)|2,{46,46,0}},
/* 23922 */ {(12<<2)|2,{46,46,0}},
/* 23923 */ {(12<<2)|2,{46,46,0}},
/* 23924 */ {(12<<2)|2,{46,46,0}},
/* 23925 */ {(12<<2)|2,{46,46,0}},
/* 23926 */ {(12<<2)|2,{46,46,0}},
/* 23927 */ {(12<<2)|2,{46,46,0}},
/* 23928 */ {(12<<2)|2,{46,46,0}},
/* 23929 */ {(12<<2)|2,{46,46,0}},
/* 23930 */ {(12<<2)|2,{46,46,0}},
/* 23931 */ {(12<<2)|2,{46,46,0}},
/* 23932 */ {(12<<2)|2,{46,46,0}},
/* 23933 */ {(12<<2)|2,{46,46,0}},
/* 23934 */ {(12<<2)|2,{46,46,0}},
/* 23935 */ {(12<<2)|2,{46,46,0}},
/* 23936 */ {(12<<2)|2,{46,47,0}},
/* 23937 */ {(12<<2)|2,{46,47,0}},
/* 23938 */ {(12<<2)|2,{46,47,0}},
/* 23939 */ {(12<<2)|2,{46,47,0}},
/* 23940 */ {(12<<2)|2,{46,47,0}},
/* 23941 */ {(12<<2)|2,{46,47,0}},
/* 23942 */ {(12<<2)|2,{46,47,0}},
/* 23943 */ {(12<<2)|2,{46,47,0}},
/* 23944 */ {(12<<2)|2,{46,47,0}},
/* 23945 */ {(12<<2)|2,{46,47,0}},
/* 23946 */ {(12<<2)|2,{46,47,0}},
/* 23947 */ {(12<<2)|2,{46,47,0}},
/* 23948 */ {(12<<2)|2,{46,47,0}},
/* 23949 */ {(12<<2)|2,{46,47,0}},
/* 23950 */ {(12<<2)|2,{46,47,0}},
/* 23951 */ {(12<<2)|2,{46,47,0}},
/* 23952 */ {(12<<2)|2,{46,51,0}},
/* 23953 */ {(12<<2)|2,{46,51,0}},
/* 23954 */ {(12<<2)|2,{46,51,0}},
/* 23955 */ {(12<<2)|2,{46,51,0}},
/* 23956 */ {(12<<2)|2,{46,51,0}},
/* 23957 */ {(12<<2)|2,{46,51,0}},
/* 23958 */ {(12<<2)|2,{46,51,0}},
/* 23959 */ {(12<<2)|2,{46,51,0}},
/* 23960 */ {(12<<2)|2,{46,51,0}},
/* 23961 */ {(12<<2)|2,{46,51,0}},
/* 23962 */ {(12<<2)|2,{46,51,0}},
/* 23963 */ {(12<<2)|2,{46,51,0}},
/* 23964 */ {(12<<2)|2,{46,51,0}},
/* 23965 */ {(12<<2)|2,{46,51,0}},
/* 23966 */ {(12<<2)|2,{46,51,0}},
/* 23967 */ {(12<<2)|2,{46,51,0}},
/* 23968 */ {(12<<2)|2,{46,52,0}},
/* 23969 */ {(12<<2)|2,{46,52,0}},
/* 23970 */ {(12<<2)|2,{46,52,0}},
/* 23971 */ {(12<<2)|2,{46,52,0}},
/* 23972 */ {(12<<2)|2,{46,52,0}},
/* 23973 */ {(12<<2)|2,{46,52,0}},
/* 23974 */ {(12<<2)|2,{46,52,0}},
/* 23975 */ {(12<<2)|2,{46,52,0}},
/* 23976 */ {(12<<2)|2,{46,52,0}},
/* 23977 */ {(12<<2)|2,{46,52,0}},
/* 23978 */ {(12<<2)|2,{46,52,0}},
/* 23979 */ {(12<<2)|2,{46,52,0}},
/* 23980 */ {(12<<2)|2,{46,52,0}},
/* 23981 */ {(12<<2)|2,{46,52,0}},
/* 23982 */ {(12<<2)|2,{46,52,0}},
/* 23983 */ {(12<<2)|2,{46,52,0}},
/* 23984 */ {(12<<2)|2,{46,53,0}},
/* 23985 */ {(12<<2)|2,{46,53,0}},
/* 23986 */ {(12<<2)|2,{46,53,0}},
/* 23987 */ {(12<<2)|2,{46,53,0}},
/* 23988 */ {(12<<2)|2,{46,53,0}},
/* 23989 */ {(12<<2)|2,{46,53,0}},
/* 23990 */ {(12<<2)|2,{46,53,0}},
/* 23991 */ {(12<<2)|2,{46,53,0}},
/* 23992 */ {(12<<2)|2,{46,53,0}},
/* 23993 */ {(12<<2)|2,{46,53,0}},
/* 23994 */ {(12<<2)|2,{46,53,0}},
/* 23995 */ {(12<<2)|2,{46,53,0}},
/* 23996 */ {(12<<2)|2,{46,53,0}},
/* 23997 */ {(12<<2)|2,{46,53,0}},
/* 23998 */ {(12<<2)|2,{46,53,0}},
/* 23999 */ {(12<<2)|2,{46,53,0}},
/* 24000 */ {(12<<2)|2,{46,54,0}},
/* 24001 */ {(12<<2)|2,{46,54,0}},
/* 24002 */ {(12<<2)|2,{46,54,0}},
/* 24003 */ {(12<<2)|2,{46,54,0}},
/* 24004 */ {(12<<2)|2,{46,54,0}},
/* 24005 */ {(12<<2)|2,{46,54,0}},
/* 24006 */ {(12<<2)|2,{46,54,0}},
/* 24007 */ {(12<<2)|2,{46,54,0}},
/* 24008 */ {(12<<2)|2,{46,54,0}},
/* 24009 */ {(12<<2)|2,{46,54,0}},
/* 24010 */ {(12<<2)|2,{46,54,0}},
/* 24011 */ {(12<<2)|2,{46,54,0}},
/* 24012 */ {(12<<2)|2,{46,54,0}},
/* 24013 */ {(12<<2)|2,{46,54,0}},
/* 24014 */ {(12<<2)|2,{46,54,0}},
/* 24015 */ {(12<<2)|2,{46,54,0}},
/* 24016 */ {(12<<2)|2,{46,55,0}},
/* 24017 */ {(12<<2)|2,{46,55,0}},
/* 24018 */ {(12<<2)|2,{46,55,0}},
/* 24019 */ {(12<<2)|2,{46,55,0}},
/* 24020 */ {(12<<2)|2,{46,55,0}},
/* 24021 */ {(12<<2)|2,{46,55,0}},
/* 24022 */ {(12<<2)|2,{46,55,0}},
/* 24023 */ {(12<<2)|2,{46,55,0}},
/* 24024 */ {(12<<2)|2,{46,55,0}},
/* 24025 */ {(12<<2)|2,{46,55,0}},
/* 24026 */ {(12<<2)|2,{46,55,0}},
/* 24027 */ {(12<<2)|2,{46,55,0}},
/* 24028 */ {(12<<2)|2,{46,55,0}},
/* 24029 */ {(12<<2)|2,{46,55,0}},
/* 24030 */ {(12<<2)|2,{46,55,0}},
/* 24031 */ {(12<<2)|2,{46,55,0}},
/* 24032 */ {(12<<2)|2,{46,56,0}},
/* 24033 */ {(12<<2)|2,{46,56,0}},
/* 24034 */ {(12<<2)|2,{46,56,0}},
/* 24035 */ {(12<<2)|2,{46,56,0}},
/* 24036 */ {(12<<2)|2,{46,56,0}},
/* 24037 */ {(12<<2)|2,{46,56,0}},
/* 24038 */ {(12<<2)|2,{46,56,0}},
/* 24039 */ {(12<<2)|2,{46,56,0}},
/* 24040 */ {(12<<2)|2,{46,56,0}},
/* 24041 */ {(12<<2)|2,{46,56,0}},
/* 24042 */ {(12<<2)|2,{46,56,0}},
/* 24043 */ {(12<<2)|2,{46,56,0}},
/* 24044 */ {(12<<2)|2,{46,56,0}},
/* 24045 */ {(12<<2)|2,{46,56,0}},
/* 24046 */ {(12<<2)|2,{46,56,0}},
/* 24047 */ {(12<<2)|2,{46,56,0}},
/* 24048 */ {(12<<2)|2,{46,57,0}},
/* 24049 */ {(12<<2)|2,{46,57,0}},
/* 24050 */ {(12<<2)|2,{46,57,0}},
/* 24051 */ {(12<<2)|2,{46,57,0}},
/* 24052 */ {(12<<2)|2,{46,57,0}},
/* 24053 */ {(12<<2)|2,{46,57,0}},
/* 24054 */ {(12<<2)|2,{46,57,0}},
/* 24055 */ {(12<<2)|2,{46,57,0}},
/* 24056 */ {(12<<2)|2,{46,57,0}},
/* 24057 */ {(12<<2)|2,{46,57,0}},
/* 24058 */ {(12<<2)|2,{46,57,0}},
/* 24059 */ {(12<<2)|2,{46,57,0}},
/* 24060 */ {(12<<2)|2,{46,57,0}},
/* 24061 */ {(12<<2)|2,{46,57,0}},
/* 24062 */ {(12<<2)|2,{46,57,0}},
/* 24063 */ {(12<<2)|2,{46,57,0}},
/* 24064 */ {(12<<2)|2,{46,61,0}},
/* 24065 */ {(12<<2)|2,{46,61,0}},
/* 24066 */ {(12<<2)|2,{46,61,0}},
/* 24067 */ {(12<<2)|2,{46,61,0}},
/* 24068 */ {(12<<2)|2,{46,61,0}},
/* 24069 */ {(12<<2)|2,{46,61,0}},
/* 24070 */ {(12<<2)|2,{46,61,0}},
/* 24071 */ {(12<<2)|2,{46,61,0}},
/* 24072 */ {(12<<2)|2,{46,61,0}},
/* 24073 */ {(12<<2)|2,{46,61,0}},
/* 24074 */ {(12<<2)|2,{46,61,0}},
/* 24075 */ {(12<<2)|2,{46,61,0}},
/* 24076 */ {(12<<2)|2,{46,61,0}},
/* 24077 */ {(12<<2)|2,{46,61,0}},
/* 24078 */ {(12<<2)|2,{46,61,0}},
/* 24079 */ {(12<<2)|2,{46,61,0}},
/* 24080 */ {(12<<2)|2,{46,65,0}},
/* 24081 */ {(12<<2)|2,{46,65,0}},
/* 24082 */ {(12<<2)|2,{46,65,0}},
/* 24083 */ {(12<<2)|2,{46,65,0}},
/* 24084 */ {(12<<2)|2,{46,65,0}},
/* 24085 */ {(12<<2)|2,{46,65,0}},
/* 24086 */ {(12<<2)|2,{46,65,0}},
/* 24087 */ {(12<<2)|2,{46,65,0}},
/* 24088 */ {(12<<2)|2,{46,65,0}},
/* 24089 */ {(12<<2)|2,{46,65,0}},
/* 24090 */ {(12<<2)|2,{46,65,0}},
/* 24091 */ {(12<<2)|2,{46,65,0}},
/* 24092 */ {(12<<2)|2,{46,65,0}},
/* 24093 */ {(12<<2)|2,{46,65,0}},
/* 24094 */ {(12<<2)|2,{46,65,0}},
/* 24095 */ {(12<<2)|2,{46,65,0}},
/* 24096 */ {(12<<2)|2,{46,95,0}},
/* 24097 */ {(12<<2)|2,{46,95,0}},
/* 24098 */ {(12<<2)|2,{46,95,0}},
/* 24099 */ {(12<<2)|2,{46,95,0}},
/* 24100 */ {(12<<2)|2,{46,95,0}},
/* 24101 */ {(12<<2)|2,{46,95,0}},
/* 24102 */ {(12<<2)|2,{46,95,0}},
/* 24103 */ {(12<<2)|2,{46,95,0}},
/* 24104 */ {(12<<2)|2,{46,95,0}},
/* 24105 */ {(12<<2)|2,{46,95,0}},
/* 24106 */ {(12<<2)|2,{46,95,0}},
/* 24107 */ {(12<<2)|2,{46,95,0}},
/* 24108 */ {(12<<2)|2,{46,95,0}},
/* 24109 */ {(12<<2)|2,{46,95,0}},
/* 24110 */ {(12<<2)|2,{46,95,0}},
/* 24111 */ {(12<<2)|2,{46,95,0}},
/* 24112 */ {(12<<2)|2,{46,98,0}},
/* 24113 */ {(12<<2)|2,{46,98,0}},
/* 24114 */ {(12<<2)|2,{46,98,0}},
/* 24115 */ {(12<<2)|2,{46,98,0}},
/* 24116 */ {(12<<2)|2,{46,98,0}},
/* 24117 */ {(12<<2)|2,{46,98,0}},
/* 24118 */ {(12<<2)|2,{46,98,0}},
/* 24119 */ {(12<<2)|2,{46,98,0}},
/* 24120 */ {(12<<2)|2,{46,98,0}},
/* 24121 */ {(12<<2)|2,{46,98,0}},
/* 24122 */ {(12<<2)|2,{46,98,0}},
/* 24123 */ {(12<<2)|2,{46,98,0}},
/* 24124 */ {(12<<2)|2,{46,98,0}},
/* 24125 */ {(12<<2)|2,{46,98,0}},
/* 24126 */ {(12<<2)|2,{46,98,0}},
/* 24127 */ {(12<<2)|2,{46,98,0}},
/* 24128 */ {(12<<2)|2,{46,100,0}},
/* 24129 */ {(12<<2)|2,{46,100,0}},
/* 24130 */ {(12<<2)|2,{46,100,0}},
/* 24131 */ {(12<<2)|2,{46,100,0}},
/* 24132 */ {(12<<2)|2,{46,100,0}},
/* 24133 */ {(12<<2)|2,{46,100,0}},
/* 24134 */ {(12<<2)|2,{46,100,0}},
/* 24135 */ {(12<<2)|2,{46,100,0}},
/* 24136 */ {(12<<2)|2,{46,100,0}},
/* 24137 */ {(12<<2)|2,{46,100,0}},
/* 24138 */ {(12<<2)|2,{46,100,0}},
/* 24139 */ {(12<<2)|2,{46,100,0}},
/* 24140 */ {(12<<2)|2,{46,100,0}},
/* 24141 */ {(12<<2)|2,{46,100,0}},
/* 24142 */ {(12<<2)|2,{46,100,0}},
/* 24143 */ {(12<<2)|2,{46,100,0}},
/* 24144 */ {(12<<2)|2,{46,102,0}},
/* 24145 */ {(12<<2)|2,{46,102,0}},
/* 24146 */ {(12<<2)|2,{46,102,0}},
/* 24147 */ {(12<<2)|2,{46,102,0}},
/* 24148 */ {(12<<2)|2,{46,102,0}},
/* 24149 */ {(12<<2)|2,{46,102,0}},
/* 24150 */ {(12<<2)|2,{46,102,0}},
/* 24151 */ {(12<<2)|2,{46,102,0}},
/* 24152 */ {(12<<2)|2,{46,102,0}},
/* 24153 */ {(12<<2)|2,{46,102,0}},
/* 24154 */ {(12<<2)|2,{46,102,0}},
/* 24155 */ {(12<<2)|2,{46,102,0}},
/* 24156 */ {(12<<2)|2,{46,102,0}},
/* 24157 */ {(12<<2)|2,{46,102,0}},
/* 24158 */ {(12<<2)|2,{46,102,0}},
/* 24159 */ {(12<<2)|2,{46,102,0}},
/* 24160 */ {(12<<2)|2,{46,103,0}},
/* 24161 */ {(12<<2)|2,{46,103,0}},
/* 24162 */ {(12<<2)|2,{46,103,0}},
/* 24163 */ {(12<<2)|2,{46,103,0}},
/* 24164 */ {(12<<2)|2,{46,103,0}},
/* 24165 */ {(12<<2)|2,{46,103,0}},
/* 24166 */ {(12<<2)|2,{46,103,0}},
/* 24167 */ {(12<<2)|2,{46,103,0}},
/* 24168 */ {(12<<2)|2,{46,103,0}},
/* 24169 */ {(12<<2)|2,{46,103,0}},
/* 24170 */ {(12<<2)|2,{46,103,0}},
/* 24171 */ {(12<<2)|2,{46,103,0}},
/* 24172 */ {(12<<2)|2,{46,103,0}},
/* 24173 */ {(12<<2)|2,{46,103,0}},
/* 24174 */ {(12<<2)|2,{46,103,0}},
/* 24175 */ {(12<<2)|2,{46,103,0}},
/* 24176 */ {(12<<2)|2,{46,104,0}},
/* 24177 */ {(12<<2)|2,{46,104,0}},
/* 24178 */ {(12<<2)|2,{46,104,0}},
/* 24179 */ {(12<<2)|2,{46,104,0}},
/* 24180 */ {(12<<2)|2,{46,104,0}},
/* 24181 */ {(12<<2)|2,{46,104,0}},
/* 24182 */ {(12<<2)|2,{46,104,0}},
/* 24183 */ {(12<<2)|2,{46,104,0}},
/* 24184 */ {(12<<2)|2,{46,104,0}},
/* 24185 */ {(12<<2)|2,{46,104,0}},
/* 24186 */ {(12<<2)|2,{46,104,0}},
/* 24187 */ {(12<<2)|2,{46,104,0}},
/* 24188 */ {(12<<2)|2,{46,104,0}},
/* 24189 */ {(12<<2)|2,{46,104,0}},
/* 24190 */ {(12<<2)|2,{46,104,0}},
/* 24191 */ {(12<<2)|2,{46,104,0}},
/* 24192 */ {(12<<2)|2,{46,108,0}},
/* 24193 */ {(12<<2)|2,{46,108,0}},
/* 24194 */ {(12<<2)|2,{46,108,0}},
/* 24195 */ {(12<<2)|2,{46,108,0}},
/* 24196 */ {(12<<2)|2,{46,108,0}},
/* 24197 */ {(12<<2)|2,{46,108,0}},
/* 24198 */ {(12<<2)|2,{46,108,0}},
/* 24199 */ {(12<<2)|2,{46,108,0}},
/* 24200 */ {(12<<2)|2,{46,108,0}},
/* 24201 */ {(12<<2)|2,{46,108,0}},
/* 24202 */ {(12<<2)|2,{46,108,0}},
/* 24203 */ {(12<<2)|2,{46,108,0}},
/* 24204 */ {(12<<2)|2,{46,108,0}},
/* 24205 */ {(12<<2)|2,{46,108,0}},
/* 24206 */ {(12<<2)|2,{46,108,0}},
/* 24207 */ {(12<<2)|2,{46,108,0}},
/* 24208 */ {(12<<2)|2,{46,109,0}},
/* 24209 */ {(12<<2)|2,{46,109,0}},
/* 24210 */ {(12<<2)|2,{46,109,0}},
/* 24211 */ {(12<<2)|2,{46,109,0}},
/* 24212 */ {(12<<2)|2,{46,109,0}},
/* 24213 */ {(12<<2)|2,{46,109,0}},
/* 24214 */ {(12<<2)|2,{46,109,0}},
/* 24215 */ {(12<<2)|2,{46,109,0}},
/* 24216 */ {(12<<2)|2,{46,109,0}},
/* 24217 */ {(12<<2)|2,{46,109,0}},
/* 24218 */ {(12<<2)|2,{46,109,0}},
/* 24219 */ {(12<<2)|2,{46,109,0}},
/* 24220 */ {(12<<2)|2,{46,109,0}},
/* 24221 */ {(12<<2)|2,{46,109,0}},
/* 24222 */ {(12<<2)|2,{46,109,0}},
/* 24223 */ {(12<<2)|2,{46,109,0}},
/* 24224 */ {(12<<2)|2,{46,110,0}},
/* 24225 */ {(12<<2)|2,{46,110,0}},
/* 24226 */ {(12<<2)|2,{46,110,0}},
/* 24227 */ {(12<<2)|2,{46,110,0}},
/* 24228 */ {(12<<2)|2,{46,110,0}},
/* 24229 */ {(12<<2)|2,{46,110,0}},
/* 24230 */ {(12<<2)|2,{46,110,0}},
/* 24231 */ {(12<<2)|2,{46,110,0}},
/* 24232 */ {(12<<2)|2,{46,110,0}},
/* 24233 */ {(12<<2)|2,{46,110,0}},
/* 24234 */ {(12<<2)|2,{46,110,0}},
/* 24235 */ {(12<<2)|2,{46,110,0}},
/* 24236 */ {(12<<2)|2,{46,110,0}},
/* 24237 */ {(12<<2)|2,{46,110,0}},
/* 24238 */ {(12<<2)|2,{46,110,0}},
/* 24239 */ {(12<<2)|2,{46,110,0}},
/* 24240 */ {(12<<2)|2,{46,112,0}},
/* 24241 */ {(12<<2)|2,{46,112,0}},
/* 24242 */ {(12<<2)|2,{46,112,0}},
/* 24243 */ {(12<<2)|2,{46,112,0}},
/* 24244 */ {(12<<2)|2,{46,112,0}},
/* 24245 */ {(12<<2)|2,{46,112,0}},
/* 24246 */ {(12<<2)|2,{46,112,0}},
/* 24247 */ {(12<<2)|2,{46,112,0}},
/* 24248 */ {(12<<2)|2,{46,112,0}},
/* 24249 */ {(12<<2)|2,{46,112,0}},
/* 24250 */ {(12<<2)|2,{46,112,0}},
/* 24251 */ {(12<<2)|2,{46,112,0}},
/* 24252 */ {(12<<2)|2,{46,112,0}},
/* 24253 */ {(12<<2)|2,{46,112,0}},
/* 24254 */ {(12<<2)|2,{46,112,0}},
/* 24255 */ {(12<<2)|2,{46,112,0}},
/* 24256 */ {(12<<2)|2,{46,114,0}},
/* 24257 */ {(12<<2)|2,{46,114,0}},
/* 24258 */ {(12<<2)|2,{46,114,0}},
/* 24259 */ {(12<<2)|2,{46,114,0}},
/* 24260 */ {(12<<2)|2,{46,114,0}},
/* 24261 */ {(12<<2)|2,{46,114,0}},
/* 24262 */ {(12<<2)|2,{46,114,0}},
/* 24263 */ {(12<<2)|2,{46,114,0}},
/* 24264 */ {(12<<2)|2,{46,114,0}},
/* 24265 */ {(12<<2)|2,{46,114,0}},
/* 24266 */ {(12<<2)|2,{46,114,0}},
/* 24267 */ {(12<<2)|2,{46,114,0}},
/* 24268 */ {(12<<2)|2,{46,114,0}},
/* 24269 */ {(12<<2)|2,{46,114,0}},
/* 24270 */ {(12<<2)|2,{46,114,0}},
/* 24271 */ {(12<<2)|2,{46,114,0}},
/* 24272 */ {(12<<2)|2,{46,117,0}},
/* 24273 */ {(12<<2)|2,{46,117,0}},
/* 24274 */ {(12<<2)|2,{46,117,0}},
/* 24275 */ {(12<<2)|2,{46,117,0}},
/* 24276 */ {(12<<2)|2,{46,117,0}},
/* 24277 */ {(12<<2)|2,{46,117,0}},
/* 24278 */ {(12<<2)|2,{46,117,0}},
/* 24279 */ {(12<<2)|2,{46,117,0}},
/* 24280 */ {(12<<2)|2,{46,117,0}},
/* 24281 */ {(12<<2)|2,{46,117,0}},
/* 24282 */ {(12<<2)|2,{46,117,0}},
/* 24283 */ {(12<<2)|2,{46,117,0}},
/* 24284 */ {(12<<2)|2,{46,117,0}},
/* 24285 */ {(12<<2)|2,{46,117,0}},
/* 24286 */ {(12<<2)|2,{46,117,0}},
/* 24287 */ {(12<<2)|2,{46,117,0}},
/* 24288 */ {(13<<2)|2,{46,58,0}},
/* 24289 */ {(13<<2)|2,{46,58,0}},
/* 24290 */ {(13<<2)|2,{46,58,0}},
/* 24291 */ {(13<<2)|2,{46,58,0}},
/* 24292 */ {(13<<2)|2,{46,58,0}},
/* 24293 */ {(13<<2)|2,{46,58,0}},
/* 24294 */ {(13<<2)|2,{46,58,0}},
/* 24295 */ {(13<<2)|2,{46,58,0}},
/* 24296 */ {(13<<2)|2,{46,66,0}},
/* 24297 */ {(13<<2)|2,{46,66,0}},
/* 24298 */ {(13<<2)|2,{46,66,0}},
/* 24299 */ {(13<<2)|2,{46,66,0}},
/* 24300 */ {(13<<2)|2,{46,66,0}},
/* 24301 */ {(13<<2)|2,{46,66,0}},
/* 24302 */ {(13<<2)|2,{46,66,0}},
/* 24303 */ {(13<<2)|2,{46,66,0}},
/* 24304 */ {(13<<2)|2,{46,67,0}},
/* 24305 */ {(13<<2)|2,{46,67,0}},
/* 24306 */ {(13<<2)|2,{46,67,0}},
/* 24307 */ {(13<<2)|2,{46,67,0}},
/* 24308 */ {(13<<2)|2,{46,67,0}},
/* 24309 */ {(13<<2)|2,{46,67,0}},
/* 24310 */ {(13<<2)|2,{46,67,0}},
/* 24311 */ {(13<<2)|2,{46,67,0}},
/* 24312 */ {(13<<2)|2,{46,68,0}},
/* 24313 */ {(13<<2)|2,{46,68,0}},
/* 24314 */ {(13<<2)|2,{46,68,0}},
/* 24315 */ {(13<<2)|2,{46,68,0}},
/* 24316 */ {(13<<2)|2,{46,68,0}},
/* 24317 */ {(13<<2)|2,{46,68,0}},
/* 24318 */ {(13<<2)|2,{46,68,0}},
/* 24319 */ {(13<<2)|2,{46,68,0}},
/* 24320 */ {(13<<2)|2,{46,69,0}},
/* 24321 */ {(13<<2)|2,{46,69,0}},
/* 24322 */ {(13<<2)|2,{46,69,0}},
/* 24323 */ {(13<<2)|2,{46,69,0}},
/* 24324 */ {(13<<2)|2,{46,69,0}},
/* 24325 */ {(13<<2)|2,{46,69,0}},
/* 24326 */ {(13<<2)|2,{46,69,0}},
/* 24327 */ {(13<<2)|2,{46,69,0}},
/* 24328 */ {(13<<2)|2,{46,70,0}},
/* 24329 */ {(13<<2)|2,{46,70,0}},
/* 24330 */ {(13<<2)|2,{46,70,0}},
/* 24331 */ {(13<<2)|2,{46,70,0}},
/* 24332 */ {(13<<2)|2,{46,70,0}},
/* 24333 */ {(13<<2)|2,{46,70,0}},
/* 24334 */ {(13<<2)|2,{46,70,0}},
/* 24335 */ {(13<<2)|2,{46,70,0}},
/* 24336 */ {(13<<2)|2,{46,71,0}},
/* 24337 */ {(13<<2)|2,{46,71,0}},
/* 24338 */ {(13<<2)|2,{46,71,0}},
/* 24339 */ {(13<<2)|2,{46,71,0}},
/* 24340 */ {(13<<2)|2,{46,71,0}},
/* 24341 */ {(13<<2)|2,{46,71,0}},
/* 24342 */ {(13<<2)|2,{46,71,0}},
/* 24343 */ {(13<<2)|2,{46,71,0}},
/* 24344 */ {(13<<2)|2,{46,72,0}},
/* 24345 */ {(13<<2)|2,{46,72,0}},
/* 24346 */ {(13<<2)|2,{46,72,0}},
/* 24347 */ {(13<<2)|2,{46,72,0}},
/* 24348 */ {(13<<2)|2,{46,72,0}},
/* 24349 */ {(13<<2)|2,{46,72,0}},
/* 24350 */ {(13<<2)|2,{46,72,0}},
/* 24351 */ {(13<<2)|2,{46,72,0}},
/* 24352 */ {(13<<2)|2,{46,73,0}},
/* 24353 */ {(13<<2)|2,{46,73,0}},
/* 24354 */ {(13<<2)|2,{46,73,0}},
/* 24355 */ {(13<<2)|2,{46,73,0}},
/* 24356 */ {(13<<2)|2,{46,73,0}},
/* 24357 */ {(13<<2)|2,{46,73,0}},
/* 24358 */ {(13<<2)|2,{46,73,0}},
/* 24359 */ {(13<<2)|2,{46,73,0}},
/* 24360 */ {(13<<2)|2,{46,74,0}},
/* 24361 */ {(13<<2)|2,{46,74,0}},
/* 24362 */ {(13<<2)|2,{46,74,0}},
/* 24363 */ {(13<<2)|2,{46,74,0}},
/* 24364 */ {(13<<2)|2,{46,74,0}},
/* 24365 */ {(13<<2)|2,{46,74,0}},
/* 24366 */ {(13<<2)|2,{46,74,0}},
/* 24367 */ {(13<<2)|2,{46,74,0}},
/* 24368 */ {(13<<2)|2,{46,75,0}},
/* 24369 */ {(13<<2)|2,{46,75,0}},
/* 24370 */ {(13<<2)|2,{46,75,0}},
/* 24371 */ {(13<<2)|2,{46,75,0}},
/* 24372 */ {(13<<2)|2,{46,75,0}},
/* 24373 */ {(13<<2)|2,{46,75,0}},
/* 24374 */ {(13<<2)|2,{46,75,0}},
/* 24375 */ {(13<<2)|2,{46,75,0}},
/* 24376 */ {(13<<2)|2,{46,76,0}},
/* 24377 */ {(13<<2)|2,{46,76,0}},
/* 24378 */ {(13<<2)|2,{46,76,0}},
/* 24379 */ {(13<<2)|2,{46,76,0}},
/* 24380 */ {(13<<2)|2,{46,76,0}},
/* 24381 */ {(13<<2)|2,{46,76,0}},
/* 24382 */ {(13<<2)|2,{46,76,0}},
/* 24383 */ {(13<<2)|2,{46,76,0}},
/* 24384 */ {(13<<2)|2,{46,77,0}},
/* 24385 */ {(13<<2)|2,{46,77,0}},
/* 24386 */ {(13<<2)|2,{46,77,0}},
/* 24387 */ {(13<<2)|2,{46,77,0}},
/* 24388 */ {(13<<2)|2,{46,77,0}},
/* 24389 */ {(13<<2)|2,{46,77,0}},
/* 24390 */ {(13<<2)|2,{46,77,0}},
/* 24391 */ {(13<<2)|2,{46,77,0}},
/* 24392 */ {(13<<2)|2,{46,78,0}},
/* 24393 */ {(13<<2)|2,{46,78,0}},
/* 24394 */ {(13<<2)|2,{46,78,0}},
/* 24395 */ {(13<<2)|2,{46,78,0}},
/* 24396 */ {(13<<2)|2,{46,78,0}},
/* 24397 */ {(13<<2)|2,{46,78,0}},
/* 24398 */ {(13<<2)|2,{46,78,0}},
/* 24399 */ {(13<<2)|2,{46,78,0}},
/* 24400 */ {(13<<2)|2,{46,79,0}},
/* 24401 */ {(13<<2)|2,{46,79,0}},
/* 24402 */ {(13<<2)|2,{46,79,0}},
/* 24403 */ {(13<<2)|2,{46,79,0}},
/* 24404 */ {(13<<2)|2,{46,79,0}},
/* 24405 */ {(13<<2)|2,{46,79,0}},
/* 24406 */ {(13<<2)|2,{46,79,0}},
/* 24407 */ {(13<<2)|2,{46,79,0}},
/* 24408 */ {(13<<2)|2,{46,80,0}},
/* 24409 */ {(13<<2)|2,{46,80,0}},
/* 24410 */ {(13<<2)|2,{46,80,0}},
/* 24411 */ {(13<<2)|2,{46,80,0}},
/* 24412 */ {(13<<2)|2,{46,80,0}},
/* 24413 */ {(13<<2)|2,{46,80,0}},
/* 24414 */ {(13<<2)|2,{46,80,0}},
/* 24415 */ {(13<<2)|2,{46,80,0}},
/* 24416 */ {(13<<2)|2,{46,81,0}},
/* 24417 */ {(13<<2)|2,{46,81,0}},
/* 24418 */ {(13<<2)|2,{46,81,0}},
/* 24419 */ {(13<<2)|2,{46,81,0}},
/* 24420 */ {(13<<2)|2,{46,81,0}},
/* 24421 */ {(13<<2)|2,{46,81,0}},
/* 24422 */ {(13<<2)|2,{46,81,0}},
/* 24423 */ {(13<<2)|2,{46,81,0}},
/* 24424 */ {(13<<2)|2,{46,82,0}},
/* 24425 */ {(13<<2)|2,{46,82,0}},
/* 24426 */ {(13<<2)|2,{46,82,0}},
/* 24427 */ {(13<<2)|2,{46,82,0}},
/* 24428 */ {(13<<2)|2,{46,82,0}},
/* 24429 */ {(13<<2)|2,{46,82,0}},
/* 24430 */ {(13<<2)|2,{46,82,0}},
/* 24431 */ {(13<<2)|2,{46,82,0}},
/* 24432 */ {(13<<2)|2,{46,83,0}},
/* 24433 */ {(13<<2)|2,{46,83,0}},
/* 24434 */ {(13<<2)|2,{46,83,0}},
/* 24435 */ {(13<<2)|2,{46,83,0}},
/* 24436 */ {(13<<2)|2,{46,83,0}},
/* 24437 */ {(13<<2)|2,{46,83,0}},
/* 24438 */ {(13<<2)|2,{46,83,0}},
/* 24439 */ {(13<<2)|2,{46,83,0}},
/* 24440 */ {(13<<2)|2,{46,84,0}},
/* 24441 */ {(13<<2)|2,{46,84,0}},
/* 24442 */ {(13<<2)|2,{46,84,0}},
/* 24443 */ {(13<<2)|2,{46,84,0}},
/* 24444 */ {(13<<2)|2,{46,84,0}},
/* 24445 */ {(13<<2)|2,{46,84,0}},
/* 24446 */ {(13<<2)|2,{46,84,0}},
/* 24447 */ {(13<<2)|2,{46,84,0}},
/* 24448 */ {(13<<2)|2,{46,85,0}},
/* 24449 */ {(13<<2)|2,{46,85,0}},
/* 24450 */ {(13<<2)|2,{46,85,0}},
/* 24451 */ {(13<<2)|2,{46,85,0}},
/* 24452 */ {(13<<2)|2,{46,85,0}},
/* 24453 */ {(13<<2)|2,{46,85,0}},
/* 24454 */ {(13<<2)|2,{46,85,0}},
/* 24455 */ {(13<<2)|2,{46,85,0}},
/* 24456 */ {(13<<2)|2,{46,86,0}},
/* 24457 */ {(13<<2)|2,{46,86,0}},
/* 24458 */ {(13<<2)|2,{46,86,0}},
/* 24459 */ {(13<<2)|2,{46,86,0}},
/* 24460 */ {(13<<2)|2,{46,86,0}},
/* 24461 */ {(13<<2)|2,{46,86,0}},
/* 24462 */ {(13<<2)|2,{46,86,0}},
/* 24463 */ {(13<<2)|2,{46,86,0}},
/* 24464 */ {(13<<2)|2,{46,87,0}},
/* 24465 */ {(13<<2)|2,{46,87,0}},
/* 24466 */ {(13<<2)|2,{46,87,0}},
/* 24467 */ {(13<<2)|2,{46,87,0}},
/* 24468 */ {(13<<2)|2,{46,87,0}},
/* 24469 */ {(13<<2)|2,{46,87,0}},
/* 24470 */ {(13<<2)|2,{46,87,0}},
/* 24471 */ {(13<<2)|2,{46,87,0}},
/* 24472 */ {(13<<2)|2,{46,89,0}},
/* 24473 */ {(13<<2)|2,{46,89,0}},
/* 24474 */ {(13<<2)|2,{46,89,0}},
/* 24475 */ {(13<<2)|2,{46,89,0}},
/* 24476 */ {(13<<2)|2,{46,89,0}},
/* 24477 */ {(13<<2)|2,{46,89,0}},
/* 24478 */ {(13<<2)|2,{46,89,0}},
/* 24479 */ {(13<<2)|2,{46,89,0}},
/* 24480 */ {(13<<2)|2,{46,106,0}},
/* 24481 */ {(13<<2)|2,{46,106,0}},
/* 24482 */ {(13<<2)|2,{46,106,0}},
/* 24483 */ {(13<<2)|2,{46,106,0}},
/* 24484 */ {(13<<2)|2,{46,106,0}},
/* 24485 */ {(13<<2)|2,{46,106,0}},
/* 24486 */ {(13<<2)|2,{46,106,0}},
/* 24487 */ {(13<<2)|2,{46,106,0}},
/* 24488 */ {(13<<2)|2,{46,107,0}},
/* 24489 */ {(13<<2)|2,{46,107,0}},
/* 24490 */ {(13<<2)|2,{46,107,0}},
/* 24491 */ {(13<<2)|2,{46,107,0}},
/* 24492 */ {(13<<2)|2,{46,107,0}},
/* 24493 */ {(13<<2)|2,{46,107,0}},
/* 24494 */ {(13<<2)|2,{46,107,0}},
/* 24495 */ {(13<<2)|2,{46,107,0}},
/* 24496 */ {(13<<2)|2,{46,113,0}},
/* 24497 */ {(13<<2)|2,{46,113,0}},
/* 24498 */ {(13<<2)|2,{46,113,0}},
/* 24499 */ {(13<<2)|2,{46,113,0}},
/* 24500 */ {(13<<2)|2,{46,113,0}},
/* 24501 */ {(13<<2)|2,{46,113,0}},
/* 24502 */ {(13<<2)|2,{46,113,0}},
/* 24503 */ {(13<<2)|2,{46,113,0}},
/* 24504 */ {(13<<2)|2,{46,118,0}},
/* 24505 */ {(13<<2)|2,{46,118,0}},
/* 24506 */ {(13<<2)|2,{46,118,0}},
/* 24507 */ {(13<<2)|2,{46,118,0}},
/* 24508 */ {(13<<2)|2,{46,118,0}},
/* 24509 */ {(13<<2)|2,{46,118,0}},
/* 24510 */ {(13<<2)|2,{46,118,0}},
/* 24511 */ {(13<<2)|2,{46,118,0}},
/* 24512 */ {(13<<2)|2,{46,119,0}},
/* 24513 */ {(13<<2)|2,{46,119,0}},
/* 24514 */ {(13<<2)|2,{46,119,0}},
/* 24515 */ {(13<<2)|2,{46,119,0}},
/* 24516 */ {(13<<2)|2,{46,119,0}},
/* 24517 */ {(13<<2)|2,{46,119,0}},
/* 24518 */ {(13<<2)|2,{46,119,0}},
/* 24519 */ {(13<<2)|2,{46,119,0}},
/* 24520 */ {(13<<2)|2,{46,120,0}},
/* 24521 */ {(13<<2)|2,{46,120,0}},
/* 24522 */ {(13<<2)|2,{46,120,0}},
/* 24523 */ {(13<<2)|2,{46,120,0}},
/* 24524 */ {(13<<2)|2,{46,120,0}},
/* 24525 */ {(13<<2)|2,{46,120,0}},
/* 24526 */ {(13<<2)|2,{46,120,0}},
/* 24527 */ {(13<<2)|2,{46,120,0}},
/* 24528 */ {(13<<2)|2,{46,121,0}},
/* 24529 */ {(13<<2)|2,{46,121,0}},
/* 24530 */ {(13<<2)|2,{46,121,0}},
/* 24531 */ {(13<<2)|2,{46,121,0}},
/* 24532 */ {(13<<2)|2,{46,121,0}},
/* 24533 */ {(13<<2)|2,{46,121,0}},
/* 24534 */ {(13<<2)|2,{46,121,0}},
/* 24535 */ {(13<<2)|2,{46,121,0}},
/* 24536 */ {(13<<2)|2,{46,122,0}},
/* 24537 */ {(13<<2)|2,{46,122,0}},
/* 24538 */ {(13<<2)|2,{46,122,0}},
/* 24539 */ {(13<<2)|2,{46,122,0}},
/* 24540 */ {(13<<2)|2,{46,122,0}},
/* 24541 */ {(13<<2)|2,{46,122,0}},
/* 24542 */ {(13<<2)|2,{46,122,0}},
/* 24543 */ {(13<<2)|2,{46,122,0}},
/* 24544 */ {(14<<2)|2,{46,38,0}},
/* 24545 */ {(14<<2)|2,{46,38,0}},
/* 24546 */ {(14<<2)|2,{46,38,0}},
/* 24547 */ {(14<<2)|2,{46,38,0}},
/* 24548 */ {(14<<2)|2,{46,42,0}},
/* 24549 */ {(14<<2)|2,{46,42,0}},
/* 24550 */ {(14<<2)|2,{46,42,0}},
/* 24551 */ {(14<<2)|2,{46,42,0}},
/* 24552 */ {(14<<2)|2,{46,44,0}},
/* 24553 */ {(14<<2)|2,{46,44,0}},
/* 24554 */ {(14<<2)|2,{46,44,0}},
/* 24555 */ {(14<<2)|2,{46,44,0}},
/* 24556 */ {(14<<2)|2,{46,59,0}},
/* 24557 */ {(14<<2)|2,{46,59,0}},
/* 24558 */ {(14<<2)|2,{46,59,0}},
/* 24559 */ {(14<<2)|2,{46,59,0}},
/* 24560 */ {(14<<2)|2,{46,88,0}},
/* 24561 */ {(14<<2)|2,{46,88,0}},
/* 24562 */ {(14<<2)|2,{46,88,0}},
/* 24563 */ {(14<<2)|2,{46,88,0}},
/* 24564 */ {(14<<2)|2,{46,90,0}},
/* 24565 */ {(14<<2)|2,{46,90,0}},
/* 24566 */ {(14<<2)|2,{46,90,0}},
/* 24567 */ {(14<<2)|2,{46,90,0}},
/* 24568 */ {(16<<2)|2,{46,33,0}},
/* 24569 */ {(16<<2)|2,{46,34,0}},
/* 24570 */ {(16<<2)|2,{46,40,0}},
/* 24571 */ {(16<<2)|2,{46,41,0}},
/* 24572 */ {(16<<2)|2,{46,63,0}},
/* 24573 */ {(6<<2)|1,{46,0,0}},
/* 24574 */ {(6<<2)|1,{46,0,0}},
/* 24575 */ {(6<<2)|1,{46,0,0}},
/* 24576 */ {(16<<2)|3,{47,48,48}},
/* 24577 */ {(16<<2)|3,{47,48,49}},
/* 24578 */ {(16<<2)|3,{47,48,50}},
/* 24579 */ {(16<<2)|3,{47,48,97}},
/* 24580 */ {(16<<2)|3,{47,48,99}},
/* 24581 */ {(16<<2)|3,{47,48,101}},
/* 24582 */ {(16<<2)|3,{47,48,105}},
/* 24583 */ {(16<<2)|3,{47,48,111}},
/* 24584 */ {(16<<2)|3,{47,48,115}},
/* 24585 */ {(16<<2)|3,{47,48,116}},
/* 24586 */ {(11<<2)|2,{47,48,0}},
/* 24587 */ {(11<<2)|2,{47,48,0}},
/* 24588 */ {(11<<2)|2,{47,48,0}},
/* 24589 */ {(11<<2)|2,{47,48,0}},
/* 24590 */ {(11<<2)|2,{47,48,0}},
/* 24591 */ {(11<<2)|2,{47,48,0}},
/* 24592 */ {(11<<2)|2,{47,48,0}},
/* 24593 */ {(11<<2)|2,{47,48,0}},
/* 24594 */ {(11<<2)|2,{47,48,0}},
/* 24595 */ {(11<<2)|2,{47,48,0}},
/* 24596 */ {(11<<2)|2,{47,48,0}},
/* 24597 */ {(11<<2)|2,{47,48,0}},
/* 24598 */ {(11<<2)|2,{47,48,0}},
/* 24599 */ {(11<<2)|2,{47,48,0}},
/* 24600 */ {(11<<2)|2,{47,48,0}},
/* 24601 */ {(11<<2)|2,{47,48,0}},
/* 24602 */ {(11<<2)|2,{47,48,0}},
/* 24603 */ {(11<<2)|2,{47,48,0}},
/* 24604 */ {(11<<2)|2,{47,48,0}},
/* 24605 */ {(11<<2)|2,{47,48,0}},
/* 24606 */ {(11<<2)|2,{47,48,0}},
/* 24607 */ {(11<<2)|2,{47,48,0}},
/* 24608 */ {(16<<2)|3,{47,49,48}},
/* 24609 */ {(16<<2)|3,{47,49,49}},
/* 24610 */ {(16<<2)|3,{47,49,50}},
/* 24611 */ {(16<<2)|3,{47,49,97}},
/* 24612 */ {(16<<2)|3,{47,49,99}},
/* 24613 */ {(16<<2)|3,{47,49,101}},
/* 24614 */ {(16<<2)|3,{47,49,105}},
/* 24615 */ {(16<<2)|3,{47,49,111}},
/* 24616 */ {(16<<2)|3,{47,49,115}},
/* 24617 */ {(16<<2)|3,{47,49,116}},
/* 24618 */ {(11<<2)|2,{47,49,0}},
/* 24619 */ {(11<<2)|2,{47,49,0}},
/* 24620 */ {(11<<2)|2,{47,49,0}},
/* 24621 */ {(11<<2)|2,{47,49,0}},
/* 24622 */ {(11<<2)|2,{47,49,0}},
/* 24623 */ {(11<<2)|2,{47,49,0}},
/* 24624 */ {(11<<2)|2,{47,49,0}},
/* 24625 */ {(11<<2)|2,{47,49,0}},
/* 24626 */ {(11<<2)|2,{47,49,0}},
/* 24627 */ {(11<<2)|2,{47,49,0}},
/* 24628 */ {(11<<2)|2,{47,49,0}},
/* 24629 */ {(11<<2)|2,{47,49,0}},
/* 24630 */ {(11<<2)|2,{47,49,0}},
/* 24631 */ {(11<<2)|2,{47,49,0}},
/* 24632 */ {(11<<2)|2,{47,49,0}},
/* 24633 */ {(11<<2)|2,{47,49,0}},
/* 24634 */ {(11<<2)|2,{47,49,0}},
/* 24635 */ {(11<<2)|2,{47,49,0}},
/* 24636 */ {(11<<2)|2,{47,49,0}},
/* 24637 */ {(11<<2)|2,{47,49,0}},
/* 24638 */ {(11<<2)|2,{47,49,0}},
/* 24639 */ {(11<<2)|2,{47,49,0}},
/* 24640 */ {(16<<2)|3,{47,50,48}},
/* 24641 */ {(16<<2)|3,{47,50,49}},
/* 24642 */ {(16<<2)|3,{47,50,50}},
/* 24643 */ {(16<<2)|3,{47,50,97}},
/* 24644 */ {(16<<2)|3,{47,50,99}},
/* 24645 */ {(16<<2)|3,{47,50,101}},
/* 24646 */ {(16<<2)|3,{47,50,105}},
/* 24647 */ {(16<<2)|3,{47,50,111}},
/* 24648 */ {(16<<2)|3,{47,50,115}},
/* 24649 */ {(16<<2)|3,{47,50,116}},
/* 24650 */ {(11<<2)|2,{47,50,0}},
/* 24651 */ {(11<<2)|2,{47,50,0}},
/* 24652 */ {(11<<2)|2,{47,50,0}},
/* 24653 */ {(11<<2)|2,{47,50,0}},
/* 24654 */ {(11<<2)|2,{47,50,0}},
/* 24655 */ {(11<<2)|2,{47,50,0}},
/* 24656 */ {(11<<2)|2,{47,50,0}},
/* 24657 */ {(11<<2)|2,{47,50,0}},
/* 24658 */ {(11<<2)|2,{47,50,0}},
/* 24659 */ {(11<<2)|2,{47,50,0}},
/* 24660 */ {(11<<2)|2,{47,50,0}},
/* 24661 */ {(11<<2)|2,{47,50,0}},
/* 24662 */ {(11<<2)|2,{47,50,0}},
/* 24663 */ {(11<<2)|2,{47,50,0}},
/* 24664 */ {(11<<2)|2,{47,50,0}},
/* 24665 */ {(11<<2)|2,{47,50,0}},
/* 24666 */ {(11<<2)|2,{47,50,0}},
/* 24667 */ {(11<<2)|2,{47,50,0}},
/* 24668 */ {(11<<2)|2,{47,50,0}},
/* 24669 */ {(11<<2)|2,{47,50,0}},
/* 24670 */ {(11<<2)|2,{47,50,0}},
/* 24671 */ {(11<<2)|2,{47,50,0}},
/* 24672 */ {(16<<2)|3,{47,97,48}},
/* 24673 */ {(16<<2)|3,{47,97,49}},
/* 24674 */ {(16<<2)|3,{47,97,50}},
/* 24675 */ {(16<<2)|3,{47,97,97}},
/* 24676 */ {(16<<2)|3,{47,97,99}},
/* 24677 */ {(16<<2)|3,{47,97,101}},
/* 24678 */ {(16<<2)|3,{47,97,105}},
/* 24679 */ {(16<<2)|3,{47,97,111}},
/* 24680 */ {(16<<2)|3,{47,97,115}},
/* 24681 */ {(16<<2)|3,{47,97,116}},
/* 24682 */ {(11<<2)|2,{47,97,0}},
/* 24683 */ {(11<<2)|2,{47,97,0}},
/* 24684 */ {(11<<2)|2,{47,97,0}},
/* 24685 */ {(11<<2)|2,{47,97,0}},
/* 24686 */ {(11<<2)|2,{47,97,0}},
/* 24687 */ {(11<<2)|2,{47,97,0}},
/* 24688 */ {(11<<2)|2,{47,97,0}},
/* 24689 */ {(11<<2)|2,{47,97,0}},
/* 24690 */ {(11<<2)|2,{47,97,0}},
/* 24691 */ {(11<<2)|2,{47,97,0}},
/* 24692 */ {(11<<2)|2,{47,97,0}},
/* 24693 */ {(11<<2)|2,{47,97,0}},
/* 24694 */ {(11<<2)|2,{47,97,0}},
/* 24695 */ {(11<<2)|2,{47,97,0}},
/* 24696 */ {(11<<2)|2,{47,97,0}},
/* 24697 */ {(11<<2)|2,{47,97,0}},
/* 24698 */ {(11<<2)|2,{47,97,0}},
/* 24699 */ {(11<<2)|2,{47,97,0}},
/* 24700 */ {(11<<2)|2,{47,97,0}},
/* 24701 */ {(11<<2)|2,{47,97,0}},
/* 24702 */ {(11<<2)|2,{47,97,0}},
/* 24703 */ {(11<<2)|2,{47,97,0}},
/* 24704 */ {(16<<2)|3,{47,99,48}},
/* 24705 */ {(16<<2)|3,{47,99,49}},
/* 24706 */ {(16<<2)|3,{47,99,50}},
/* 24707 */ {(16<<2)|3,{47,99,97}},
/* 24708 */ {(16<<2)|3,{47,99,99}},
/* 24709 */ {(16<<2)|3,{47,99,101}},
/* 24710 */ {(16<<2)|3,{47,99,105}},
/* 24711 */ {(16<<2)|3,{47,99,111}},
/* 24712 */ {(16<<2)|3,{47,99,115}},
/* 24713 */ {(16<<2)|3,{47,99,116}},
/* 24714 */ {(11<<2)|2,{47,99,0}},
/* 24715 */ {(11<<2)|2,{47,99,0}},
/* 24716 */ {(11<<2)|2,{47,99,0}},
/* 24717 */ {(11<<2)|2,{47,99,0}},
/* 24718 */ {(11<<2)|2,{47,99,0}},
/* 24719 */ {(11<<2)|2,{47,99,0}},
/* 24720 */ {(11<<2)|2,{47,99,0}},
/* 24721 */ {(11<<2)|2,{47,99,0}},
/* 24722 */ {(11<<2)|2,{47,99,0}},
/* 24723 */ {(11<<2)|2,{47,99,0}},
/* 24724 */ {(11<<2)|2,{47,99,0}},
/* 24725 */ {(11<<2)|2,{47,99,0}},
/* 24726 */ {(11<<2)|2,{47,99,0}},
/* 24727 */ {(11<<2)|2,{47,99,0}},
/* 24728 */ {(11<<2)|2,{47,99,0}},
/* 24729 */ {(11<<2)|2,{47,99,0}},
/* 24730 */ {(11<<2)|2,{47,99,0}},
/* 24731 */ {(11<<2)|2,{47,99,0}},
/* 24732 */ {(11<<2)|2,{47,99,0}},
/* 24733 */ {(11<<2)|2,{47,99,0}},
/* 24734 */ {(11<<2)|2,{47,99,0}},
/* 24735 */ {(11<<2)|2,{47,99,0}},
/* 24736 */ {(16<<2)|3,{47,101,48}},
/* 24737 */ {(16<<2)|3,{47,101,49}},
/* 24738 */ {(16<<2)|3,{47,101,50}},
/* 24739 */ {(16<<2)|3,{47,101,97}},
/* 24740 */ {(16<<2)|3,{47,101,99}},
/* 24741 */ {(16<<2)|3,{47,101,101}},
/* 24742 */ {(16<<2)|3,{47,101,105}},
/* 24743 */ {(16<<2)|3,{47,101,111}},
/* 24744 */ {(16<<2)|3,{47,101,115}},
/* 24745 */ {(16<<2)|3,{47,101,116}},
/* 24746 */ {(11<<2)|2,{47,101,0}},
/* 24747 */ {(11<<2)|2,{47,101,0}},
/* 24748 */ {(11<<2)|2,{47,101,0}},
/* 24749 */ {(11<<2)|2,{47,101,0}},
/* 24750 */ {(11<<2)|2,{47,101,0}},
/* 24751 */ {(11<<2)|2,{47,101,0}},
/* 24752 */ {(11<<2)|2,{47,101,0}},
/* 24753 */ {(11<<2)|2,{47,101,0}},
/* 24754 */ {(11<<2)|2,{47,101,0}},
/* 24755 */ {(11<<2)|2,{47,101,0}},
/* 24756 */ {(11<<2)|2,{47,101,0}},
/* 24757 */ {(11<<2)|2,{47,101,0}},
/* 24758 */ {(11<<2)|2,{47,101,0}},
/* 24759 */ {(11<<2)|2,{47,101,0}},
/* 24760 */ {(11<<2)|2,{47,101,0}},
/* 24761 */ {(11<<2)|2,{47,101,0}},
/* 24762 */ {(11<<2)|2,{47,101,0}},
/* 24763 */ {(11<<2)|2,{47,101,0}},
/* 24764 */ {(11<<2)|2,{47,101,0}},
/* 24765 */ {(11<<2)|2,{47,101,0}},
/* 24766 */ {(11<<2)|2,{47,101,0}},
/* 24767 */ {(11<<2)|2,{47,101,0}},
/* 24768 */ {(16<<2)|3,{47,105,48}},
/* 24769 */ {(16<<2)|3,{47,105,49}},
/* 24770 */ {(16<<2)|3,{47,105,50}},
/* 24771 */ {(16<<2)|3,{47,105,97}},
/* 24772 */ {(16<<2)|3,{47,105,99}},
/* 24773 */ {(16<<2)|3,{47,105,101}},
/* 24774 */ {(16<<2)|3,{47,105,105}},
/* 24775 */ {(16<<2)|3,{47,105,111}},
/* 24776 */ {(16<<2)|3,{47,105,115}},
/* 24777 */ {(16<<2)|3,{47,105,116}},
/* 24778 */ {(11<<2)|2,{47,105,0}},
/* 24779 */ {(11<<2)|2,{47,105,0}},
/* 24780 */ {(11<<2)|2,{47,105,0}},
/* 24781 */ {(11<<2)|2,{47,105,0}},
/* 24782 */ {(11<<2)|2,{47,105,0}},
/* 24783 */ {(11<<2)|2,{47,105,0}},
/* 24784 */ {(11<<2)|2,{47,105,0}},
/* 24785 */ {(11<<2)|2,{47,105,0}},
/* 24786 */ {(11<<2)|2,{47,105,0}},
/* 24787 */ {(11<<2)|2,{47,105,0}},
/* 24788 */ {(11<<2)|2,{47,105,0}},
/* 24789 */ {(11<<2)|2,{47,105,0}},
/* 24790 */ {(11<<2)|2,{47,105,0}},
/* 24791 */ {(11<<2)|2,{47,105,0}},
/* 24792 */ {(11<<2)|2,{47,105,0}},
/* 24793 */ {(11<<2)|2,{47,105,0}},
/* 24794 */ {(11<<2)|2,{47,105,0}},
/* 24795 */ {(11<<2)|2,{47,105,0}},
/* 24796 */ {(11<<2)|2,{47,105,0}},
/* 24797 */ {(11<<2)|2,{47,105,0}},
/* 24798 */ {(11<<2)|2,{47,105,0}},
/* 24799 */ {(11<<2)|2,{47,105,0}},
/* 24800 */ {(16<<2)|3,{47,111,48}},
/* 24801 */ {(16<<2)|3,{47,111,49}},
/* 24802 */ {(16<<2)|3,{47,111,50}},
/* 24803 */ {(16<<2)|3,{47,111,97}},
/* 24804 */ {(16<<2)|3,{47,111,99}},
/* 24805 */ {(16<<2)|3,{47,111,101}},
/* 24806 */ {(16<<2)|3,{47,111,105}},
/* 24807 */ {(16<<2)|3,{47,111,111}},
/* 24808 */ {(16<<2)|3,{47,111,115}},
/* 24809 */ {(16<<2)|3,{47,111,116}},
/* 24810 */ {(11<<2)|2,{47,111,0}},
/* 24811 */ {(11<<2)|2,{47,111,0}},
/* 24812 */ {(11<<2)|2,{47,111,0}},
/* 24813 */ {(11<<2)|2,{47,111,0}},
/* 24814 */ {(11<<2)|2,{47,111,0}},
/* 24815 */ {(11<<2)|2,{47,111,0}},
/* 24816 */ {(11<<2)|2,{47,111,0}},
/* 24817 */ {(11<<2)|2,{47,111,0}},
/* 24818 */ {(11<<2)|2,{47,111,0}},
/* 24819 */ {(11<<2)|2,{47,111,0}},
/* 24820 */ {(11<<2)|2,{47,111,0}},
/* 24821 */ {(11<<2)|2,{47,111,0}},
/* 24822 */ {(11<<2)|2,{47,111,0}},
/* 24823 */ {(11<<2)|2,{47,111,0}},
/* 24824 */ {(11<<2)|2,{47,111,0}},
/* 24825 */ {(11<<2)|2,{47,111,0}},
/* 24826 */ {(11<<2)|2,{47,111,0}},
/* 24827 */ {(11<<2)|2,{47,111,0}},
/* 24828 */ {(11<<2)|2,{47,111,0}},
/* 24829 */ {(11<<2)|2,{47,111,0}},
/* 24830 */ {(11<<2)|2,{47,111,0}},
/* 24831 */ {(11<<2)|2,{47,111,0}},
/* 24832 */ {(16<<2)|3,{47,115,48}},
/* 24833 */ {(16<<2)|3,{47,115,49}},
/* 24834 */ {(16<<2)|3,{47,115,50}},
/* 24835 */ {(16<<2)|3,{47,115,97}},
/* 24836 */ {(16<<2)|3,{47,115,99}},
/* 24837 */ {(16<<2)|3,{47,115,101}},
/* 24838 */ {(16<<2)|3,{47,115,105}},
/* 24839 */ {(16<<2)|3,{47,115,111}},
/* 24840 */ {(16<<2)|3,{47,115,115}},
/* 24841 */ {(16<<2)|3,{47,115,116}},
/* 24842 */ {(11<<2)|2,{47,115,0}},
/* 24843 */ {(11<<2)|2,{47,115,0}},
/* 24844 */ {(11<<2)|2,{47,115,0}},
/* 24845 */ {(11<<2)|2,{47,115,0}},
/* 24846 */ {(11<<2)|2,{47,115,0}},
/* 24847 */ {(11<<2)|2,{47,115,0}},
/* 24848 */ {(11<<2)|2,{47,115,0}},
/* 24849 */ {(11<<2)|2,{47,115,0}},
/* 24850 */ {(11<<2)|2,{47,115,0}},
/* 24851 */ {(11<<2)|2,{47,115,0}},
/* 24852 */ {(11<<2)|2,{47,115,0}},
/* 24853 */ {(11<<2)|2,{47,115,0}},
/* 24854 */ {(11<<2)|2,{47,115,0}},
/* 24855 */ {(11<<2)|2,{47,115,0}},
/* 24856 */ {(11<<2)|2,{47,115,0}},
/* 24857 */ {(11<<2)|2,{47,115,0}},
/* 24858 */ {(11<<2)|2,{47,115,0}},
/* 24859 */ {(11<<2)|2,{47,115,0}},
/* 24860 */ {(11<<2)|2,{47,115,0}},
/* 24861 */ {(11<<2)|2,{47,115,0}},
/* 24862 */ {(11<<2)|2,{47,115,0}},
/* 24863 */ {(11<<2)|2,{47,115,0}},
/* 24864 */ {(16<<2)|3,{47,116,48}},
/* 24865 */ {(16<<2)|3,{47,116,49}},
/* 24866 */ {(16<<2)|3,{47,116,50}},
/* 24867 */ {(16<<2)|3,{47,116,97}},
/* 24868 */ {(16<<2)|3,{47,116,99}},
/* 24869 */ {(16<<2)|3,{47,116,101}},
/* 24870 */ {(16<<2)|3,{47,116,105}},
/* 24871 */ {(16<<2)|3,{47,116,111}},
/* 24872 */ {(16<<2)|3,{47,116,115}},
/* 24873 */ {(16<<2)|3,{47,116,116}},
/* 24874 */ {(11<<2)|2,{47,116,0}},
/* 24875 */ {(11<<2)|2,{47,116,0}},
/* 24876 */ {(11<<2)|2,{47,116,0}},
/* 24877 */ {(11<<2)|2,{47,116,0}},
/* 24878 */ {(11<<2)|2,{47,116,0}},
/* 24879 */ {(11<<2)|2,{47,116,0}},
/* 24880 */ {(11<<2)|2,{47,116,0}},
/* 24881 */ {(11<<2)|2,{47,116,0}},
/* 24882 */ {(11<<2)|2,{47,116,0}},
/* 24883 */ {(11<<2)|2,{47,116,0}},
/* 24884 */ {(11<<2)|2,{47,116,0}},
/* 24885 */ {(11<<2)|2,{47,116,0}},
/* 24886 */ {(11<<2)|2,{47,116,0}},
/* 24887 */ {(11<<2)|2,{47,116,0}},
/* 24888 */ {(11<<2)|2,{47,116,0}},
/* 24889 */ {(11<<2)|2,{47,116,0}},
/* 24890 */ {(11<<2)|2,{47,116,0}},
/* 24891 */ {(11<<2)|2,{47,116,0}},
/* 24892 */ {(11<<2)|2,{47,116,0}},
/* 24893 */ {(11<<2)|2,{47,116,0}},
/* 24894 */ {(11<<2)|2,{47,116,0}},
/* 24895 */ {(11<<2)|2,{47,116,0}},
/* 24896 */ {(12<<2)|2,{47,32,0}},
/* 24897 */ {(12<<2)|2,{47,32,0}},
/* 24898 */ {(12<<2)|2,{47,32,0}},
/* 24899 */ {(12<<2)|2,{47,32,0}},
/* 24900 */ {(12<<2)|2,{47,32,0}},
/* 24901 */ {(12<<2)|2,{47,32,0}},
/* 24902 */ {(12<<2)|2,{47,32,0}},
/* 24903 */ {(12<<2)|2,{47,32,0}},
/* 24904 */ {(12<<2)|2,{47,32,0}},
/* 24905 */ {(12<<2)|2,{47,32,0}},
/* 24906 */ {(12<<2)|2,{47,32,0}},
/* 24907 */ {(12<<2)|2,{47,32,0}},
/* 24908 */ {(12<<2)|2,{47,32,0}},
/* 24909 */ {(12<<2)|2,{47,32,0}},
/* 24910 */ {(12<<2)|2,{47,32,0}},
/* 24911 */ {(12<<2)|2,{47,32,0}},
/* 24912 */ {(12<<2)|2,{47,37,0}},
/* 24913 */ {(12<<2)|2,{47,37,0}},
/* 24914 */ {(12<<2)|2,{47,37,0}},
/* 24915 */ {(12<<2)|2,{47,37,0}},
/* 24916 */ {(12<<2)|2,{47,37,0}},
/* 24917 */ {(12<<2)|2,{47,37,0}},
/* 24918 */ {(12<<2)|2,{47,37,0}},
/* 24919 */ {(12<<2)|2,{47,37,0}},
/* 24920 */ {(12<<2)|2,{47,37,0}},
/* 24921 */ {(12<<2)|2,{47,37,0}},
/* 24922 */ {(12<<2)|2,{47,37,0}},
/* 24923 */ {(12<<2)|2,{47,37,0}},
/* 24924 */ {(12<<2)|2,{47,37,0}},
/* 24925 */ {(12<<2)|2,{47,37,0}},
/* 24926 */ {(12<<2)|2,{47,37,0}},
/* 24927 */ {(12<<2)|2,{47,37,0}},
/* 24928 */ {(12<<2)|2,{47,45,0}},
/* 24929 */ {(12<<2)|2,{47,45,0}},
/* 24930 */ {(12<<2)|2,{47,45,0}},
/* 24931 */ {(12<<2)|2,{47,45,0}},
/* 24932 */ {(12<<2)|2,{47,45,0}},
/* 24933 */ {(12<<2)|2,{47,45,0}},
/* 24934 */ {(12<<2)|2,{47,45,0}},
/* 24935 */ {(12<<2)|2,{47,45,0}},
/* 24936 */ {(12<<2)|2,{47,45,0}},
/* 24937 */ {(12<<2)|2,{47,45,0}},
/* 24938 */ {(12<<2)|2,{47,45,0}},
/* 24939 */ {(12<<2)|2,{47,45,0}},
/* 24940 */ {(12<<2)|2,{47,45,0}},
/* 24941 */ {(12<<2)|2,{47,45,0}},
/* 24942 */ {(12<<2)|2,{47,45,0}},
/* 24943 */ {(12<<2)|2,{47,45,0}},
/* 24944 */ {(12<<2)|2,{47,46,0}},
/* 24945 */ {(12<<2)|2,{47,46,0}},
/* 24946 */ {(12<<2)|2,{47,46,0}},
/* 24947 */ {(12<<2)|2,{47,46,0}},
/* 24948 */ {(12<<2)|2,{47,46,0}},
/* 24949 */ {(12<<2)|2,{47,46,0}},
/* 24950 */ {(12<<2)|2,{47,46,0}},
/* 24951 */ {(12<<2)|2,{47,46,0}},
/* 24952 */ {(12<<2)|2,{47,46,0}},
/* 24953 */ {(12<<2)|2,{47,46,0}},
/* 24954 */ {(12<<2)|2,{47,46,0}},
/* 24955 */ {(12<<2)|2,{47,46,0}},
/* 24956 */ {(12<<2)|2,{47,46,0}},
/* 24957 */ {(12<<2)|2,{47,46,0}},
/* 24958 */ {(12<<2)|2,{47,46,0}},
/* 24959 */ {(12<<2)|2,{47,46,0}},
/* 24960 */ {(12<<2)|2,{47,47,0}},
/* 24961 */ {(12<<2)|2,{47,47,0}},
/* 24962 */ {(12<<2)|2,{47,47,0}},
/* 24963 */ {(12<<2)|2,{47,47,0}},
/* 24964 */ {(12<<2)|2,{47,47,0}},
/* 24965 */ {(12<<2)|2,{47,47,0}},
/* 24966 */ {(12<<2)|2,{47,47,0}},
/* 24967 */ {(12<<2)|2,{47,47,0}},
/* 24968 */ {(12<<2)|2,{47,47,0}},
/* 24969 */ {(12<<2)|2,{47,47,0}},
/* 24970 */ {(12<<2)|2,{47,47,0}},
/* 24971 */ {(12<<2)|2,{47,47,0}},
/* 24972 */ {(12<<2)|2,{47,47,0}},
/* 24973 */ {(12<<2)|2,{47,47,0}},
/* 24974 */ {(12<<2)|2,{47,47,0}},
/* 24975 */ {(12<<2)|2,{47,47,0}},
/* 24976 */ {(12<<2)|2,{47,51,0}},
/* 24977 */ {(12<<2)|2,{47,51,0}},
/* 24978 */ {(12<<2)|2,{47,51,0}},
/* 24979 */ {(12<<2)|2,{47,51,0}},
/* 24980 */ {(12<<2)|2,{47,51,0}},
/* 24981 */ {(12<<2)|2,{47,51,0}},
/* 24982 */ {(12<<2)|2,{47,51,0}},
/* 24983 */ {(12<<2)|2,{47,51,0}},
/* 24984 */ {(12<<2)|2,{47,51,0}},
/* 24985 */ {(12<<2)|2,{47,51,0}},
/* 24986 */ {(12<<2)|2,{47,51,0}},
/* 24987 */ {(12<<2)|2,{47,51,0}},
/* 24988 */ {(12<<2)|2,{47,51,0}},
/* 24989 */ {(12<<2)|2,{47,51,0}},
/* 24990 */ {(12<<2)|2,{47,51,0}},
/* 24991 */ {(12<<2)|2,{47,51,0}},
/* 24992 */ {(12<<2)|2,{47,52,0}},
/* 24993 */ {(12<<2)|2,{47,52,0}},
/* 24994 */ {(12<<2)|2,{47,52,0}},
/* 24995 */ {(12<<2)|2,{47,52,0}},
/* 24996 */ {(12<<2)|2,{47,52,0}},
/* 24997 */ {(12<<2)|2,{47,52,0}},
/* 24998 */ {(12<<2)|2,{47,52,0}},
/* 24999 */ {(12<<2)|2,{47,52,0}},
/* 25000 */ {(12<<2)|2,{47,52,0}},
/* 25001 */ {(12<<2)|2,{47,52,0}},
/* 25002 */ {(12<<2)|2,{47,52,0}},
/* 25003 */ {(12<<2)|2,{47,52,0}},
/* 25004 */ {(12<<2)|2,{47,52,0}},
/* 25005 */ {(12<<2)|2,{47,52,0}},
/* 25006 */ {(12<<2)|2,{47,52,0}},
/* 25007 */ {(12<<2)|2,{47,52,0}},
/* 25008 */ {(12<<2)|2,{47,53,0}},
/* 25009 */ {(12<<2)|2,{47,53,0}},
/* 25010 */ {(12<<2)|2,{47,53,0}},
/* 25011 */ {(12<<2)|2,{47,53,0}},
/* 25012 */ {(12<<2)|2,{47,53,0}},
/* 25013 */ {(12<<2)|2,{47,53,0}},
/* 25014 */ {(12<<2)|2,{47,53,0}},
/* 25015 */ {(12<<2)|2,{47,53,0}},
/* 25016 */ {(12<<2)|2,{47,53,0}},
/* 25017 */ {(12<<2)|2,{47,53,0}},
/* 25018 */ {(12<<2)|2,{47,53,0}},
/* 25019 */ {(12<<2)|2,{47,53,0}},
/* 25020 */ {(12<<2)|2,{47,53,0}},
/* 25021 */ {(12<<2)|2,{47,53,0}},
/* 25022 */ {(12<<2)|2,{47,53,0}},
/* 25023 */ {(12<<2)|2,{47,53,0}},
/* 25024 */ {(12<<2)|2,{47,54,0}},
/* 25025 */ {(12<<2)|2,{47,54,0}},
/* 25026 */ {(12<<2)|2,{47,54,0}},
/* 25027 */ {(12<<2)|2,{47,54,0}},
/* 25028 */ {(12<<2)|2,{47,54,0}},
/* 25029 */ {(12<<2)|2,{47,54,0}},
/* 25030 */ {(12<<2)|2,{47,54,0}},
/* 25031 */ {(12<<2)|2,{47,54,0}},
/* 25032 */ {(12<<2)|2,{47,54,0}},
/* 25033 */ {(12<<2)|2,{47,54,0}},
/* 25034 */ {(12<<2)|2,{47,54,0}},
/* 25035 */ {(12<<2)|2,{47,54,0}},
/* 25036 */ {(12<<2)|2,{47,54,0}},
/* 25037 */ {(12<<2)|2,{47,54,0}},
/* 25038 */ {(12<<2)|2,{47,54,0}},
/* 25039 */ {(12<<2)|2,{47,54,0}},
/* 25040 */ {(12<<2)|2,{47,55,0}},
/* 25041 */ {(12<<2)|2,{47,55,0}},
/* 25042 */ {(12<<2)|2,{47,55,0}},
/* 25043 */ {(12<<2)|2,{47,55,0}},
/* 25044 */ {(12<<2)|2,{47,55,0}},
/* 25045 */ {(12<<2)|2,{47,55,0}},
/* 25046 */ {(12<<2)|2,{47,55,0}},
/* 25047 */ {(12<<2)|2,{47,55,0}},
/* 25048 */ {(12<<2)|2,{47,55,0}},
/* 25049 */ {(12<<2)|2,{47,55,0}},
/* 25050 */ {(12<<2)|2,{47,55,0}},
/* 25051 */ {(12<<2)|2,{47,55,0}},
/* 25052 */ {(12<<2)|2,{47,55,0}},
/* 25053 */ {(12<<2)|2,{47,55,0}},
/* 25054 */ {(12<<2)|2,{47,55,0}},
/* 25055 */ {(12<<2)|2,{47,55,0}},
/* 25056 */ {(12<<2)|2,{47,56,0}},
/* 25057 */ {(12<<2)|2,{47,56,0}},
/* 25058 */ {(12<<2)|2,{47,56,0}},
/* 25059 */ {(12<<2)|2,{47,56,0}},
/* 25060 */ {(12<<2)|2,{47,56,0}},
/* 25061 */ {(12<<2)|2,{47,56,0}},
/* 25062 */ {(12<<2)|2,{47,56,0}},
/* 25063 */ {(12<<2)|2,{47,56,0}},
/* 25064 */ {(12<<2)|2,{47,56,0}},
/* 25065 */ {(12<<2)|2,{47,56,0}},
/* 25066 */ {(12<<2)|2,{47,56,0}},
/* 25067 */ {(12<<2)|2,{47,56,0}},
/* 25068 */ {(12<<2)|2,{47,56,0}},
/* 25069 */ {(12<<2)|2,{47,56,0}},
/* 25070 */ {(12<<2)|2,{47,56,0}},
/* 25071 */ {(12<<2)|2,{47,56,0}},
/* 25072 */ {(12<<2)|2,{47,57,0}},
/* 25073 */ {(12<<2)|2,{47,57,0}},
/* 25074 */ {(12<<2)|2,{47,57,0}},
/* 25075 */ {(12<<2)|2,{47,57,0}},
/* 25076 */ {(12<<2)|2,{47,57,0}},
/* 25077 */ {(12<<2)|2,{47,57,0}},
/* 25078 */ {(12<<2)|2,{47,57,0}},
/* 25079 */ {(12<<2)|2,{47,57,0}},
/* 25080 */ {(12<<2)|2,{47,57,0}},
/* 25081 */ {(12<<2)|2,{47,57,0}},
/* 25082 */ {(12<<2)|2,{47,57,0}},
/* 25083 */ {(12<<2)|2,{47,57,0}},
/* 25084 */ {(12<<2)|2,{47,57,0}},
/* 25085 */ {(12<<2)|2,{47,57,0}},
/* 25086 */ {(12<<2)|2,{47,57,0}},
/* 25087 */ {(12<<2)|2,{47,57,0}},
/* 25088 */ {(12<<2)|2,{47,61,0}},
/* 25089 */ {(12<<2)|2,{47,61,0}},
/* 25090 */ {(12<<2)|2,{47,61,0}},
/* 25091 */ {(12<<2)|2,{47,61,0}},
/* 25092 */ {(12<<2)|2,{47,61,0}},
/* 25093 */ {(12<<2)|2,{47,61,0}},
/* 25094 */ {(12<<2)|2,{47,61,0}},
/* 25095 */ {(12<<2)|2,{47,61,0}},
/* 25096 */ {(12<<2)|2,{47,61,0}},
/* 25097 */ {(12<<2)|2,{47,61,0}},
/* 25098 */ {(12<<2)|2,{47,61,0}},
/* 25099 */ {(12<<2)|2,{47,61,0}},
/* 25100 */ {(12<<2)|2,{47,61,0}},
/* 25101 */ {(12<<2)|2,{47,61,0}},
/* 25102 */ {(12<<2)|2,{47,61,0}},
/* 25103 */ {(12<<2)|2,{47,61,0}},
/* 25104 */ {(12<<2)|2,{47,65,0}},
/* 25105 */ {(12<<2)|2,{47,65,0}},
/* 25106 */ {(12<<2)|2,{47,65,0}},
/* 25107 */ {(12<<2)|2,{47,65,0}},
/* 25108 */ {(12<<2)|2,{47,65,0}},
/* 25109 */ {(12<<2)|2,{47,65,0}},
/* 25110 */ {(12<<2)|2,{47,65,0}},
/* 25111 */ {(12<<2)|2,{47,65,0}},
/* 25112 */ {(12<<2)|2,{47,65,0}},
/* 25113 */ {(12<<2)|2,{47,65,0}},
/* 25114 */ {(12<<2)|2,{47,65,0}},
/* 25115 */ {(12<<2)|2,{47,65,0}},
/* 25116 */ {(12<<2)|2,{47,65,0}},
/* 25117 */ {(12<<2)|2,{47,65,0}},
/* 25118 */ {(12<<2)|2,{47,65,0}},
/* 25119 */ {(12<<2)|2,{47,65,0}},
/* 25120 */ {(12<<2)|2,{47,95,0}},
/* 25121 */ {(12<<2)|2,{47,95,0}},
/* 25122 */ {(12<<2)|2,{47,95,0}},
/* 25123 */ {(12<<2)|2,{47,95,0}},
/* 25124 */ {(12<<2)|2,{47,95,0}},
/* 25125 */ {(12<<2)|2,{47,95,0}},
/* 25126 */ {(12<<2)|2,{47,95,0}},
/* 25127 */ {(12<<2)|2,{47,95,0}},
/* 25128 */ {(12<<2)|2,{47,95,0}},
/* 25129 */ {(12<<2)|2,{47,95,0}},
/* 25130 */ {(12<<2)|2,{47,95,0}},
/* 25131 */ {(12<<2)|2,{47,95,0}},
/* 25132 */ {(12<<2)|2,{47,95,0}},
/* 25133 */ {(12<<2)|2,{47,95,0}},
/* 25134 */ {(12<<2)|2,{47,95,0}},
/* 25135 */ {(12<<2)|2,{47,95,0}},
/* 25136 */ {(12<<2)|2,{47,98,0}},
/* 25137 */ {(12<<2)|2,{47,98,0}},
/* 25138 */ {(12<<2)|2,{47,98,0}},
/* 25139 */ {(12<<2)|2,{47,98,0}},
/* 25140 */ {(12<<2)|2,{47,98,0}},
/* 25141 */ {(12<<2)|2,{47,98,0}},
/* 25142 */ {(12<<2)|2,{47,98,0}},
/* 25143 */ {(12<<2)|2,{47,98,0}},
/* 25144 */ {(12<<2)|2,{47,98,0}},
/* 25145 */ {(12<<2)|2,{47,98,0}},
/* 25146 */ {(12<<2)|2,{47,98,0}},
/* 25147 */ {(12<<2)|2,{47,98,0}},
/* 25148 */ {(12<<2)|2,{47,98,0}},
/* 25149 */ {(12<<2)|2,{47,98,0}},
/* 25150 */ {(12<<2)|2,{47,98,0}},
/* 25151 */ {(12<<2)|2,{47,98,0}},
/* 25152 */ {(12<<2)|2,{47,100,0}},
/* 25153 */ {(12<<2)|2,{47,100,0}},
/* 25154 */ {(12<<2)|2,{47,100,0}},
/* 25155 */ {(12<<2)|2,{47,100,0}},
/* 25156 */ {(12<<2)|2,{47,100,0}},
/* 25157 */ {(12<<2)|2,{47,100,0}},
/* 25158 */ {(12<<2)|2,{47,100,0}},
/* 25159 */ {(12<<2)|2,{47,100,0}},
/* 25160 */ {(12<<2)|2,{47,100,0}},
/* 25161 */ {(12<<2)|2,{47,100,0}},
/* 25162 */ {(12<<2)|2,{47,100,0}},
/* 25163 */ {(12<<2)|2,{47,100,0}},
/* 25164 */ {(12<<2)|2,{47,100,0}},
/* 25165 */ {(12<<2)|2,{47,100,0}},
/* 25166 */ {(12<<2)|2,{47,100,0}},
/* 25167 */ {(12<<2)|2,{47,100,0}},
/* 25168 */ {(12<<2)|2,{47,102,0}},
/* 25169 */ {(12<<2)|2,{47,102,0}},
/* 25170 */ {(12<<2)|2,{47,102,0}},
/* 25171 */ {(12<<2)|2,{47,102,0}},
/* 25172 */ {(12<<2)|2,{47,102,0}},
/* 25173 */ {(12<<2)|2,{47,102,0}},
/* 25174 */ {(12<<2)|2,{47,102,0}},
/* 25175 */ {(12<<2)|2,{47,102,0}},
/* 25176 */ {(12<<2)|2,{47,102,0}},
/* 25177 */ {(12<<2)|2,{47,102,0}},
/* 25178 */ {(12<<2)|2,{47,102,0}},
/* 25179 */ {(12<<2)|2,{47,102,0}},
/* 25180 */ {(12<<2)|2,{47,102,0}},
/* 25181 */ {(12<<2)|2,{47,102,0}},
/* 25182 */ {(12<<2)|2,{47,102,0}},
/* 25183 */ {(12<<2)|2,{47,102,0}},
/* 25184 */ {(12<<2)|2,{47,103,0}},
/* 25185 */ {(12<<2)|2,{47,103,0}},
/* 25186 */ {(12<<2)|2,{47,103,0}},
/* 25187 */ {(12<<2)|2,{47,103,0}},
/* 25188 */ {(12<<2)|2,{47,103,0}},
/* 25189 */ {(12<<2)|2,{47,103,0}},
/* 25190 */ {(12<<2)|2,{47,103,0}},
/* 25191 */ {(12<<2)|2,{47,103,0}},
/* 25192 */ {(12<<2)|2,{47,103,0}},
/* 25193 */ {(12<<2)|2,{47,103,0}},
/* 25194 */ {(12<<2)|2,{47,103,0}},
/* 25195 */ {(12<<2)|2,{47,103,0}},
/* 25196 */ {(12<<2)|2,{47,103,0}},
/* 25197 */ {(12<<2)|2,{47,103,0}},
/* 25198 */ {(12<<2)|2,{47,103,0}},
/* 25199 */ {(12<<2)|2,{47,103,0}},
/* 25200 */ {(12<<2)|2,{47,104,0}},
/* 25201 */ {(12<<2)|2,{47,104,0}},
/* 25202 */ {(12<<2)|2,{47,104,0}},
/* 25203 */ {(12<<2)|2,{47,104,0}},
/* 25204 */ {(12<<2)|2,{47,104,0}},
/* 25205 */ {(12<<2)|2,{47,104,0}},
/* 25206 */ {(12<<2)|2,{47,104,0}},
/* 25207 */ {(12<<2)|2,{47,104,0}},
/* 25208 */ {(12<<2)|2,{47,104,0}},
/* 25209 */ {(12<<2)|2,{47,104,0}},
/* 25210 */ {(12<<2)|2,{47,104,0}},
/* 25211 */ {(12<<2)|2,{47,104,0}},
/* 25212 */ {(12<<2)|2,{47,104,0}},
/* 25213 */ {(12<<2)|2,{47,104,0}},
/* 25214 */ {(12<<2)|2,{47,104,0}},
/* 25215 */ {(12<<2)|2,{47,104,0}},
/* 25216 */ {(12<<2)|2,{47,108,0}},
/* 25217 */ {(12<<2)|2,{47,108,0}},
/* 25218 */ {(12<<2)|2,{47,108,0}},
/* 25219 */ {(12<<2)|2,{47,108,0}},
/* 25220 */ {(12<<2)|2,{47,108,0}},
/* 25221 */ {(12<<2)|2,{47,108,0}},
/* 25222 */ {(12<<2)|2,{47,108,0}},
/* 25223 */ {(12<<2)|2,{47,108,0}},
/* 25224 */ {(12<<2)|2,{47,108,0}},
/* 25225 */ {(12<<2)|2,{47,108,0}},
/* 25226 */ {(12<<2)|2,{47,108,0}},
/* 25227 */ {(12<<2)|2,{47,108,0}},
/* 25228 */ {(12<<2)|2,{47,108,0}},
/* 25229 */ {(12<<2)|2,{47,108,0}},
/* 25230 */ {(12<<2)|2,{47,108,0}},
/* 25231 */ {(12<<2)|2,{47,108,0}},
/* 25232 */ {(12<<2)|2,{47,109,0}},
/* 25233 */ {(12<<2)|2,{47,109,0}},
/* 25234 */ {(12<<2)|2,{47,109,0}},
/* 25235 */ {(12<<2)|2,{47,109,0}},
/* 25236 */ {(12<<2)|2,{47,109,0}},
/* 25237 */ {(12<<2)|2,{47,109,0}},
/* 25238 */ {(12<<2)|2,{47,109,0}},
/* 25239 */ {(12<<2)|2,{47,109,0}},
/* 25240 */ {(12<<2)|2,{47,109,0}},
/* 25241 */ {(12<<2)|2,{47,109,0}},
/* 25242 */ {(12<<2)|2,{47,109,0}},
/* 25243 */ {(12<<2)|2,{47,109,0}},
/* 25244 */ {(12<<2)|2,{47,109,0}},
/* 25245 */ {(12<<2)|2,{47,109,0}},
/* 25246 */ {(12<<2)|2,{47,109,0}},
/* 25247 */ {(12<<2)|2,{47,109,0}},
/* 25248 */ {(12<<2)|2,{47,110,0}},
/* 25249 */ {(12<<2)|2,{47,110,0}},
/* 25250 */ {(12<<2)|2,{47,110,0}},
/* 25251 */ {(12<<2)|2,{47,110,0}},
/* 25252 */ {(12<<2)|2,{47,110,0}},
/* 25253 */ {(12<<2)|2,{47,110,0}},
/* 25254 */ {(12<<2)|2,{47,110,0}},
/* 25255 */ {(12<<2)|2,{47,110,0}},
/* 25256 */ {(12<<2)|2,{47,110,0}},
/* 25257 */ {(12<<2)|2,{47,110,0}},
/* 25258 */ {(12<<2)|2,{47,110,0}},
/* 25259 */ {(12<<2)|2,{47,110,0}},
/* 25260 */ {(12<<2)|2,{47,110,0}},
/* 25261 */ {(12<<2)|2,{47,110,0}},
/* 25262 */ {(12<<2)|2,{47,110,0}},
/* 25263 */ {(12<<2)|2,{47,110,0}},
/* 25264 */ {(12<<2)|2,{47,112,0}},
/* 25265 */ {(12<<2)|2,{47,112,0}},
/* 25266 */ {(12<<2)|2,{47,112,0}},
/* 25267 */ {(12<<2)|2,{47,112,0}},
/* 25268 */ {(12<<2)|2,{47,112,0}},
/* 25269 */ {(12<<2)|2,{47,112,0}},
/* 25270 */ {(12<<2)|2,{47,112,0}},
/* 25271 */ {(12<<2)|2,{47,112,0}},
/* 25272 */ {(12<<2)|2,{47,112,0}},
/* 25273 */ {(12<<2)|2,{47,112,0}},
/* 25274 */ {(12<<2)|2,{47,112,0}},
/* 25275 */ {(12<<2)|2,{47,112,0}},
/* 25276 */ {(12<<2)|2,{47,112,0}},
/* 25277 */ {(12<<2)|2,{47,112,0}},
/* 25278 */ {(12<<2)|2,{47,112,0}},
/* 25279 */ {(12<<2)|2,{47,112,0}},
/* 25280 */ {(12<<2)|2,{47,114,0}},
/* 25281 */ {(12<<2)|2,{47,114,0}},
/* 25282 */ {(12<<2)|2,{47,114,0}},
/* 25283 */ {(12<<2)|2,{47,114,0}},
/* 25284 */ {(12<<2)|2,{47,114,0}},
/* 25285 */ {(12<<2)|2,{47,114,0}},
/* 25286 */ {(12<<2)|2,{47,114,0}},
/* 25287 */ {(12<<2)|2,{47,114,0}},
/* 25288 */ {(12<<2)|2,{47,114,0}},
/* 25289 */ {(12<<2)|2,{47,114,0}},
/* 25290 */ {(12<<2)|2,{47,114,0}},
/* 25291 */ {(12<<2)|2,{47,114,0}},
/* 25292 */ {(12<<2)|2,{47,114,0}},
/* 25293 */ {(12<<2)|2,{47,114,0}},
/* 25294 */ {(12<<2)|2,{47,114,0}},
/* 25295 */ {(12<<2)|2,{47,114,0}},
/* 25296 */ {(12<<2)|2,{47,117,0}},
/* 25297 */ {(12<<2)|2,{47,117,0}},
/* 25298 */ {(12<<2)|2,{47,117,0}},
/* 25299 */ {(12<<2)|2,{47,117,0}},
/* 25300 */ {(12<<2)|2,{47,117,0}},
/* 25301 */ {(12<<2)|2,{47,117,0}},
/* 25302 */ {(12<<2)|2,{47,117,0}},
/* 25303 */ {(12<<2)|2,{47,117,0}},
/* 25304 */ {(12<<2)|2,{47,117,0}},
/* 25305 */ {(12<<2)|2,{47,117,0}},
/* 25306 */ {(12<<2)|2,{47,117,0}},
/* 25307 */ {(12<<2)|2,{47,117,0}},
/* 25308 */ {(12<<2)|2,{47,117,0}},
/* 25309 */ {(12<<2)|2,{47,117,0}},
/* 25310 */ {(12<<2)|2,{47,117,0}},
/* 25311 */ {(12<<2)|2,{47,117,0}},
/* 25312 */ {(13<<2)|2,{47,58,0}},
/* 25313 */ {(13<<2)|2,{47,58,0}},
/* 25314 */ {(13<<2)|2,{47,58,0}},
/* 25315 */ {(13<<2)|2,{47,58,0}},
/* 25316 */ {(13<<2)|2,{47,58,0}},
/* 25317 */ {(13<<2)|2,{47,58,0}},
/* 25318 */ {(13<<2)|2,{47,58,0}},
/* 25319 */ {(13<<2)|2,{47,58,0}},
/* 25320 */ {(13<<2)|2,{47,66,0}},
/* 25321 */ {(13<<2)|2,{47,66,0}},
/* 25322 */ {(13<<2)|2,{47,66,0}},
/* 25323 */ {(13<<2)|2,{47,66,0}},
/* 25324 */ {(13<<2)|2,{47,66,0}},
/* 25325 */ {(13<<2)|2,{47,66,0}},
/* 25326 */ {(13<<2)|2,{47,66,0}},
/* 25327 */ {(13<<2)|2,{47,66,0}},
/* 25328 */ {(13<<2)|2,{47,67,0}},
/* 25329 */ {(13<<2)|2,{47,67,0}},
/* 25330 */ {(13<<2)|2,{47,67,0}},
/* 25331 */ {(13<<2)|2,{47,67,0}},
/* 25332 */ {(13<<2)|2,{47,67,0}},
/* 25333 */ {(13<<2)|2,{47,67,0}},
/* 25334 */ {(13<<2)|2,{47,67,0}},
/* 25335 */ {(13<<2)|2,{47,67,0}},
/* 25336 */ {(13<<2)|2,{47,68,0}},
/* 25337 */ {(13<<2)|2,{47,68,0}},
/* 25338 */ {(13<<2)|2,{47,68,0}},
/* 25339 */ {(13<<2)|2,{47,68,0}},
/* 25340 */ {(13<<2)|2,{47,68,0}},
/* 25341 */ {(13<<2)|2,{47,68,0}},
/* 25342 */ {(13<<2)|2,{47,68,0}},
/* 25343 */ {(13<<2)|2,{47,68,0}},
/* 25344 */ {(13<<2)|2,{47,69,0}},
/* 25345 */ {(13<<2)|2,{47,69,0}},
/* 25346 */ {(13<<2)|2,{47,69,0}},
/* 25347 */ {(13<<2)|2,{47,69,0}},
/* 25348 */ {(13<<2)|2,{47,69,0}},
/* 25349 */ {(13<<2)|2,{47,69,0}},
/* 25350 */ {(13<<2)|2,{47,69,0}},
/* 25351 */ {(13<<2)|2,{47,69,0}},
/* 25352 */ {(13<<2)|2,{47,70,0}},
/* 25353 */ {(13<<2)|2,{47,70,0}},
/* 25354 */ {(13<<2)|2,{47,70,0}},
/* 25355 */ {(13<<2)|2,{47,70,0}},
/* 25356 */ {(13<<2)|2,{47,70,0}},
/* 25357 */ {(13<<2)|2,{47,70,0}},
/* 25358 */ {(13<<2)|2,{47,70,0}},
/* 25359 */ {(13<<2)|2,{47,70,0}},
/* 25360 */ {(13<<2)|2,{47,71,0}},
/* 25361 */ {(13<<2)|2,{47,71,0}},
/* 25362 */ {(13<<2)|2,{47,71,0}},
/* 25363 */ {(13<<2)|2,{47,71,0}},
/* 25364 */ {(13<<2)|2,{47,71,0}},
/* 25365 */ {(13<<2)|2,{47,71,0}},
/* 25366 */ {(13<<2)|2,{47,71,0}},
/* 25367 */ {(13<<2)|2,{47,71,0}},
/* 25368 */ {(13<<2)|2,{47,72,0}},
/* 25369 */ {(13<<2)|2,{47,72,0}},
/* 25370 */ {(13<<2)|2,{47,72,0}},
/* 25371 */ {(13<<2)|2,{47,72,0}},
/* 25372 */ {(13<<2)|2,{47,72,0}},
/* 25373 */ {(13<<2)|2,{47,72,0}},
/* 25374 */ {(13<<2)|2,{47,72,0}},
/* 25375 */ {(13<<2)|2,{47,72,0}},
/* 25376 */ {(13<<2)|2,{47,73,0}},
/* 25377 */ {(13<<2)|2,{47,73,0}},
/* 25378 */ {(13<<2)|2,{47,73,0}},
/* 25379 */ {(13<<2)|2,{47,73,0}},
/* 25380 */ {(13<<2)|2,{47,73,0}},
/* 25381 */ {(13<<2)|2,{47,73,0}},
/* 25382 */ {(13<<2)|2,{47,73,0}},
/* 25383 */ {(13<<2)|2,{47,73,0}},
/* 25384 */ {(13<<2)|2,{47,74,0}},
/* 25385 */ {(13<<2)|2,{47,74,0}},
/* 25386 */ {(13<<2)|2,{47,74,0}},
/* 25387 */ {(13<<2)|2,{47,74,0}},
/* 25388 */ {(13<<2)|2,{47,74,0}},
/* 25389 */ {(13<<2)|2,{47,74,0}},
/* 25390 */ {(13<<2)|2,{47,74,0}},
/* 25391 */ {(13<<2)|2,{47,74,0}},
/* 25392 */ {(13<<2)|2,{47,75,0}},
/* 25393 */ {(13<<2)|2,{47,75,0}},
/* 25394 */ {(13<<2)|2,{47,75,0}},
/* 25395 */ {(13<<2)|2,{47,75,0}},
/* 25396 */ {(13<<2)|2,{47,75,0}},
/* 25397 */ {(13<<2)|2,{47,75,0}},
/* 25398 */ {(13<<2)|2,{47,75,0}},
/* 25399 */ {(13<<2)|2,{47,75,0}},
/* 25400 */ {(13<<2)|2,{47,76,0}},
/* 25401 */ {(13<<2)|2,{47,76,0}},
/* 25402 */ {(13<<2)|2,{47,76,0}},
/* 25403 */ {(13<<2)|2,{47,76,0}},
/* 25404 */ {(13<<2)|2,{47,76,0}},
/* 25405 */ {(13<<2)|2,{47,76,0}},
/* 25406 */ {(13<<2)|2,{47,76,0}},
/* 25407 */ {(13<<2)|2,{47,76,0}},
/* 25408 */ {(13<<2)|2,{47,77,0}},
/* 25409 */ {(13<<2)|2,{47,77,0}},
/* 25410 */ {(13<<2)|2,{47,77,0}},
/* 25411 */ {(13<<2)|2,{47,77,0}},
/* 25412 */ {(13<<2)|2,{47,77,0}},
/* 25413 */ {(13<<2)|2,{47,77,0}},
/* 25414 */ {(13<<2)|2,{47,77,0}},
/* 25415 */ {(13<<2)|2,{47,77,0}},
/* 25416 */ {(13<<2)|2,{47,78,0}},
/* 25417 */ {(13<<2)|2,{47,78,0}},
/* 25418 */ {(13<<2)|2,{47,78,0}},
/* 25419 */ {(13<<2)|2,{47,78,0}},
/* 25420 */ {(13<<2)|2,{47,78,0}},
/* 25421 */ {(13<<2)|2,{47,78,0}},
/* 25422 */ {(13<<2)|2,{47,78,0}},
/* 25423 */ {(13<<2)|2,{47,78,0}},
/* 25424 */ {(13<<2)|2,{47,79,0}},
/* 25425 */ {(13<<2)|2,{47,79,0}},
/* 25426 */ {(13<<2)|2,{47,79,0}},
/* 25427 */ {(13<<2)|2,{47,79,0}},
/* 25428 */ {(13<<2)|2,{47,79,0}},
/* 25429 */ {(13<<2)|2,{47,79,0}},
/* 25430 */ {(13<<2)|2,{47,79,0}},
/* 25431 */ {(13<<2)|2,{47,79,0}},
/* 25432 */ {(13<<2)|2,{47,80,0}},
/* 25433 */ {(13<<2)|2,{47,80,0}},
/* 25434 */ {(13<<2)|2,{47,80,0}},
/* 25435 */ {(13<<2)|2,{47,80,0}},
/* 25436 */ {(13<<2)|2,{47,80,0}},
/* 25437 */ {(13<<2)|2,{47,80,0}},
/* 25438 */ {(13<<2)|2,{47,80,0}},
/* 25439 */ {(13<<2)|2,{47,80,0}},
/* 25440 */ {(13<<2)|2,{47,81,0}},
/* 25441 */ {(13<<2)|2,{47,81,0}},
/* 25442 */ {(13<<2)|2,{47,81,0}},
/* 25443 */ {(13<<2)|2,{47,81,0}},
/* 25444 */ {(13<<2)|2,{47,81,0}},
/* 25445 */ {(13<<2)|2,{47,81,0}},
/* 25446 */ {(13<<2)|2,{47,81,0}},
/* 25447 */ {(13<<2)|2,{47,81,0}},
/* 25448 */ {(13<<2)|2,{47,82,0}},
/* 25449 */ {(13<<2)|2,{47,82,0}},
/* 25450 */ {(13<<2)|2,{47,82,0}},
/* 25451 */ {(13<<2)|2,{47,82,0}},
/* 25452 */ {(13<<2)|2,{47,82,0}},
/* 25453 */ {(13<<2)|2,{47,82,0}},
/* 25454 */ {(13<<2)|2,{47,82,0}},
/* 25455 */ {(13<<2)|2,{47,82,0}},
/* 25456 */ {(13<<2)|2,{47,83,0}},
/* 25457 */ {(13<<2)|2,{47,83,0}},
/* 25458 */ {(13<<2)|2,{47,83,0}},
/* 25459 */ {(13<<2)|2,{47,83,0}},
/* 25460 */ {(13<<2)|2,{47,83,0}},
/* 25461 */ {(13<<2)|2,{47,83,0}},
/* 25462 */ {(13<<2)|2,{47,83,0}},
/* 25463 */ {(13<<2)|2,{47,83,0}},
/* 25464 */ {(13<<2)|2,{47,84,0}},
/* 25465 */ {(13<<2)|2,{47,84,0}},
/* 25466 */ {(13<<2)|2,{47,84,0}},
/* 25467 */ {(13<<2)|2,{47,84,0}},
/* 25468 */ {(13<<2)|2,{47,84,0}},
/* 25469 */ {(13<<2)|2,{47,84,0}},
/* 25470 */ {(13<<2)|2,{47,84,0}},
/* 25471 */ {(13<<2)|2,{47,84,0}},
/* 25472 */ {(13<<2)|2,{47,85,0}},
/* 25473 */ {(13<<2)|2,{47,85,0}},
/* 25474 */ {(13<<2)|2,{47,85,0}},
/* 25475 */ {(13<<2)|2,{47,85,0}},
/* 25476 */ {(13<<2)|2,{47,85,0}},
/* 25477 */ {(13<<2)|2,{47,85,0}},
/* 25478 */ {(13<<2)|2,{47,85,0}},
/* 25479 */ {(13<<2)|2,{47,85,0}},
/* 25480 */ {(13<<2)|2,{47,86,0}},
/* 25481 */ {(13<<2)|2,{47,86,0}},
/* 25482 */ {(13<<2)|2,{47,86,0}},
/* 25483 */ {(13<<2)|2,{47,86,0}},
/* 25484 */ {(13<<2)|2,{47,86,0}},
/* 25485 */ {(13<<2)|2,{47,86,0}},
/* 25486 */ {(13<<2)|2,{47,86,0}},
/* 25487 */ {(13<<2)|2,{47,86,0}},
/* 25488 */ {(13<<2)|2,{47,87,0}},
/* 25489 */ {(13<<2)|2,{47,87,0}},
/* 25490 */ {(13<<2)|2,{47,87,0}},
/* 25491 */ {(13<<2)|2,{47,87,0}},
/* 25492 */ {(13<<2)|2,{47,87,0}},
/* 25493 */ {(13<<2)|2,{47,87,0}},
/* 25494 */ {(13<<2)|2,{47,87,0}},
/* 25495 */ {(13<<2)|2,{47,87,0}},
/* 25496 */ {(13<<2)|2,{47,89,0}},
/* 25497 */ {(13<<2)|2,{47,89,0}},
/* 25498 */ {(13<<2)|2,{47,89,0}},
/* 25499 */ {(13<<2)|2,{47,89,0}},
/* 25500 */ {(13<<2)|2,{47,89,0}},
/* 25501 */ {(13<<2)|2,{47,89,0}},
/* 25502 */ {(13<<2)|2,{47,89,0}},
/* 25503 */ {(13<<2)|2,{47,89,0}},
/* 25504 */ {(13<<2)|2,{47,106,0}},
/* 25505 */ {(13<<2)|2,{47,106,0}},
/* 25506 */ {(13<<2)|2,{47,106,0}},
/* 25507 */ {(13<<2)|2,{47,106,0}},
/* 25508 */ {(13<<2)|2,{47,106,0}},
/* 25509 */ {(13<<2)|2,{47,106,0}},
/* 25510 */ {(13<<2)|2,{47,106,0}},
/* 25511 */ {(13<<2)|2,{47,106,0}},
/* 25512 */ {(13<<2)|2,{47,107,0}},
/* 25513 */ {(13<<2)|2,{47,107,0}},
/* 25514 */ {(13<<2)|2,{47,107,0}},
/* 25515 */ {(13<<2)|2,{47,107,0}},
/* 25516 */ {(13<<2)|2,{47,107,0}},
/* 25517 */ {(13<<2)|2,{47,107,0}},
/* 25518 */ {(13<<2)|2,{47,107,0}},
/* 25519 */ {(13<<2)|2,{47,107,0}},
/* 25520 */ {(13<<2)|2,{47,113,0}},
/* 25521 */ {(13<<2)|2,{47,113,0}},
/* 25522 */ {(13<<2)|2,{47,113,0}},
/* 25523 */ {(13<<2)|2,{47,113,0}},
/* 25524 */ {(13<<2)|2,{47,113,0}},
/* 25525 */ {(13<<2)|2,{47,113,0}},
/* 25526 */ {(13<<2)|2,{47,113,0}},
/* 25527 */ {(13<<2)|2,{47,113,0}},
/* 25528 */ {(13<<2)|2,{47,118,0}},
/* 25529 */ {(13<<2)|2,{47,118,0}},
/* 25530 */ {(13<<2)|2,{47,118,0}},
/* 25531 */ {(13<<2)|2,{47,118,0}},
/* 25532 */ {(13<<2)|2,{47,118,0}},
/* 25533 */ {(13<<2)|2,{47,118,0}},
/* 25534 */ {(13<<2)|2,{47,118,0}},
/* 25535 */ {(13<<2)|2,{47,118,0}},
/* 25536 */ {(13<<2)|2,{47,119,0}},
/* 25537 */ {(13<<2)|2,{47,119,0}},
/* 25538 */ {(13<<2)|2,{47,119,0}},
/* 25539 */ {(13<<2)|2,{47,119,0}},
/* 25540 */ {(13<<2)|2,{47,119,0}},
/* 25541 */ {(13<<2)|2,{47,119,0}},
/* 25542 */ {(13<<2)|2,{47,119,0}},
/* 25543 */ {(13<<2)|2,{47,119,0}},
/* 25544 */ {(13<<2)|2,{47,120,0}},
/* 25545 */ {(13<<2)|2,{47,120,0}},
/* 25546 */ {(13<<2)|2,{47,120,0}},
/* 25547 */ {(13<<2)|2,{47,120,0}},
/* 25548 */ {(13<<2)|2,{47,120,0}},
/* 25549 */ {(13<<2)|2,{47,120,0}},
/* 25550 */ {(13<<2)|2,{47,120,0}},
/* 25551 */ {(13<<2)|2,{47,120,0}},
/* 25552 */ {(13<<2)|2,{47,121,0}},
/* 25553 */ {(13<<2)|2,{47,121,0}},
/* 25554 */ {(13<<2)|2,{47,121,0}},
/* 25555 */ {(13<<2)|2,{47,121,0}},
/* 25556 */ {(13<<2)|2,{47,121,0}},
/* 25557 */ {(13<<2)|2,{47,121,0}},
/* 25558 */ {(13<<2)|2,{47,121,0}},
/* 25559 */ {(13<<2)|2,{47,121,0}},
/* 25560 */ {(13<<2)|2,{47,122,0}},
/* 25561 */ {(13<<2)|2,{47,122,0}},
/* 25562 */ {(13<<2)|2,{47,122,0}},
/* 25563 */ {(13<<2)|2,{47,122,0}},
/* 25564 */ {(13<<2)|2,{47,122,0}},
/* 25565 */ {(13<<2)|2,{47,122,0}},
/* 25566 */ {(13<<2)|2,{47,122,0}},
/* 25567 */ {(13<<2)|2,{47,122,0}},
/* 25568 */ {(14<<2)|2,{47,38,0}},
/* 25569 */ {(14<<2)|2,{47,38,0}},
/* 25570 */ {(14<<2)|2,{47,38,0}},
/* 25571 */ {(14<<2)|2,{47,38,0}},
/* 25572 */ {(14<<2)|2,{47,42,0}},
/* 25573 */ {(14<<2)|2,{47,42,0}},
/* 25574 */ {(14<<2)|2,{47,42,0}},
/* 25575 */ {(14<<2)|2,{47,42,0}},
/* 25576 */ {(14<<2)|2,{47,44,0}},
/* 25577 */ {(14<<2)|2,{47,44,0}},
/* 25578 */ {(14<<2)|2,{47,44,0}},
/* 25579 */ {(14<<2)|2,{47,44,0}},
/* 25580 */ {(14<<2)|2,{47,59,0}},
/* 25581 */ {(14<<2)|2,{47,59,0}},
/* 25582 */ {(14<<2)|2,{47,59,0}},
/* 25583 */ {(14<<2)|2,{47,59,0}},
/* 25584 */ {(14<<2)|2,{47,88,0}},
/* 25585 */ {(14<<2)|2,{47,88,0}},
/* 25586 */ {(14<<2)|2,{47,88,0}},
/* 25587 */ {(14<<2)|2,{47,88,0}},
/* 25588 */ {(14<<2)|2,{47,90,0}},
/* 25589 */ {(14<<2)|2,{47,90,0}},
/* 25590 */ {(14<<2)|2,{47,90,0}},
/* 25591 */ {(14<<2)|2,{47,90,0}},
/* 25592 */ {(16<<2)|2,{47,33,0}},
/* 25593 */ {(16<<2)|2,{47,34,0}},
/* 25594 */ {(16<<2)|2,{47,40,0}},
/* 25595 */ {(16<<2)|2,{47,41,0}},
/* 25596 */ {(16<<2)|2,{47,63,0}},
/* 25597 */ {(6<<2)|1,{47,0,0}},
/* 25598 */ {(6<<2)|1,{47,0,0}},
/* 25599 */ {(6<<2)|1,{47,0,0}},
/* 25600 */ {(16<<2)|3,{51,48,48}},
/* 25601 */ {(16<<2)|3,{51,48,49}},
/* 25602 */ {(16<<2)|3,{51,48,50}},
/* 25603 */ {(16<<2)|3,{51,48,97}},
/* 25604 */ {(16<<2)|3,{51,48,99}},
/* 25605 */ {(16<<2)|3,{51,48,101}},
/* 25606 */ {(16<<2)|3,{51,48,105}},
/* 25607 */ {(16<<2)|3,{51,48,111}},
/* 25608 */ {(16<<2)|3,{51,48,115}},
/* 25609 */ {(16<<2)|3,{51,48,116}},
/* 25610 */ {(11<<2)|2,{51,48,0}},
/* 25611 */ {(11<<2)|2,{51,48,0}},
/* 25612 */ {(11<<2)|2,{51,48,0}},
/* 25613 */ {(11<<2)|2,{51,48,0}},
/* 25614 */ {(11<<2)|2,{51,48,0}},
/* 25615 */ {(11<<2)|2,{51,48,0}},
/* 25616 */ {(11<<2)|2,{51,48,0}},
/* 25617 */ {(11<<2)|2,{51,48,0}},
/* 25618 */ {(11<<2)|2,{51,48,0}},
/* 25619 */ {(11<<2)|2,{51,48,0}},
/* 25620 */ {(11<<2)|2,{51,48,0}},
/* 25621 */ {(11<<2)|2,{51,48,0}},
/* 25622 */ {(11<<2)|2,{51,48,0}},
/* 25623 */ {(11<<2)|2,{51,48,0}},
/* 25624 */ {(11<<2)|2,{51,48,0}},
/* 25625 */ {(11<<2)|2,{51,48,0}},
/* 25626 */ {(11<<2)|2,{51,48,0}},
/* 25627 */ {(11<<2)|2,{51,48,0}},
/* 25628 */ {(11<<2)|2,{51,48,0}},
/* 25629 */ {(11<<2)|2,{51,48,0}},
/* 25630 */ {(11<<2)|2,{51,48,0}},
/* 25631 */ {(11<<2)|2,{51,48,0}},
/* 25632 */ {(16<<2)|3,{51,49,48}},
/* 25633 */ {(16<<2)|3,{51,49,49}},
/* 25634 */ {(16<<2)|3,{51,49,50}},
/* 25635 */ {(16<<2)|3,{51,49,97}},
/* 25636 */ {(16<<2)|3,{51,49,99}},
/* 25637 */ {(16<<2)|3,{51,49,101}},
/* 25638 */ {(16<<2)|3,{51,49,105}},
/* 25639 */ {(16<<2)|3,{51,49,111}},
/* 25640 */ {(16<<2)|3,{51,49,115}},
/* 25641 */ {(16<<2)|3,{51,49,116}},
/* 25642 */ {(11<<2)|2,{51,49,0}},
/* 25643 */ {(11<<2)|2,{51,49,0}},
/* 25644 */ {(11<<2)|2,{51,49,0}},
/* 25645 */ {(11<<2)|2,{51,49,0}},
/* 25646 */ {(11<<2)|2,{51,49,0}},
/* 25647 */ {(11<<2)|2,{51,49,0}},
/* 25648 */ {(11<<2)|2,{51,49,0}},
/* 25649 */ {(11<<2)|2,{51,49,0}},
/* 25650 */ {(11<<2)|2,{51,49,0}},
/* 25651 */ {(11<<2)|2,{51,49,0}},
/* 25652 */ {(11<<2)|2,{51,49,0}},
/* 25653 */ {(11<<2)|2,{51,49,0}},
/* 25654 */ {(11<<2)|2,{51,49,0}},
/* 25655 */ {(11<<2)|2,{51,49,0}},
/* 25656 */ {(11<<2)|2,{51,49,0}},
/* 25657 */ {(11<<2)|2,{51,49,0}},
/* 25658 */ {(11<<2)|2,{51,49,0}},
/* 25659 */ {(11<<2)|2,{51,49,0}},
/* 25660 */ {(11<<2)|2,{51,49,0}},
/* 25661 */ {(11<<2)|2,{51,49,0}},
/* 25662 */ {(11<<2)|2,{51,49,0}},
/* 25663 */ {(11<<2)|2,{51,49,0}},
/* 25664 */ {(16<<2)|3,{51,50,48}},
/* 25665 */ {(16<<2)|3,{51,50,49}},
/* 25666 */ {(16<<2)|3,{51,50,50}},
/* 25667 */ {(16<<2)|3,{51,50,97}},
/* 25668 */ {(16<<2)|3,{51,50,99}},
/* 25669 */ {(16<<2)|3,{51,50,101}},
/* 25670 */ {(16<<2)|3,{51,50,105}},
/* 25671 */ {(16<<2)|3,{51,50,111}},
/* 25672 */ {(16<<2)|3,{51,50,115}},
/* 25673 */ {(16<<2)|3,{51,50,116}},
/* 25674 */ {(11<<2)|2,{51,50,0}},
/* 25675 */ {(11<<2)|2,{51,50,0}},
/* 25676 */ {(11<<2)|2,{51,50,0}},
/* 25677 */ {(11<<2)|2,{51,50,0}},
/* 25678 */ {(11<<2)|2,{51,50,0}},
/* 25679 */ {(11<<2)|2,{51,50,0}},
/* 25680 */ {(11<<2)|2,{51,50,0}},
/* 25681 */ {(11<<2)|2,{51,50,0}},
/* 25682 */ {(11<<2)|2,{51,50,0}},
/* 25683 */ {(11<<2)|2,{51,50,0}},
/* 25684 */ {(11<<2)|2,{51,50,0}},
/* 25685 */ {(11<<2)|2,{51,50,0}},
/* 25686 */ {(11<<2)|2,{51,50,0}},
/* 25687 */ {(11<<2)|2,{51,50,0}},
/* 25688 */ {(11<<2)|2,{51,50,0}},
/* 25689 */ {(11<<2)|2,{51,50,0}},
/* 25690 */ {(11<<2)|2,{51,50,0}},
/* 25691 */ {(11<<2)|2,{51,50,0}},
/* 25692 */ {(11<<2)|2,{51,50,0}},
/* 25693 */ {(11<<2)|2,{51,50,0}},
/* 25694 */ {(11<<2)|2,{51,50,0}},
/* 25695 */ {(11<<2)|2,{51,50,0}},
/* 25696 */ {(16<<2)|3,{51,97,48}},
/* 25697 */ {(16<<2)|3,{51,97,49}},
/* 25698 */ {(16<<2)|3,{51,97,50}},
/* 25699 */ {(16<<2)|3,{51,97,97}},
/* 25700 */ {(16<<2)|3,{51,97,99}},
/* 25701 */ {(16<<2)|3,{51,97,101}},
/* 25702 */ {(16<<2)|3,{51,97,105}},
/* 25703 */ {(16<<2)|3,{51,97,111}},
/* 25704 */ {(16<<2)|3,{51,97,115}},
/* 25705 */ {(16<<2)|3,{51,97,116}},
/* 25706 */ {(11<<2)|2,{51,97,0}},
/* 25707 */ {(11<<2)|2,{51,97,0}},
/* 25708 */ {(11<<2)|2,{51,97,0}},
/* 25709 */ {(11<<2)|2,{51,97,0}},
/* 25710 */ {(11<<2)|2,{51,97,0}},
/* 25711 */ {(11<<2)|2,{51,97,0}},
/* 25712 */ {(11<<2)|2,{51,97,0}},
/* 25713 */ {(11<<2)|2,{51,97,0}},
/* 25714 */ {(11<<2)|2,{51,97,0}},
/* 25715 */ {(11<<2)|2,{51,97,0}},
/* 25716 */ {(11<<2)|2,{51,97,0}},
/* 25717 */ {(11<<2)|2,{51,97,0}},
/* 25718 */ {(11<<2)|2,{51,97,0}},
/* 25719 */ {(11<<2)|2,{51,97,0}},
/* 25720 */ {(11<<2)|2,{51,97,0}},
/* 25721 */ {(11<<2)|2,{51,97,0}},
/* 25722 */ {(11<<2)|2,{51,97,0}},
/* 25723 */ {(11<<2)|2,{51,97,0}},
/* 25724 */ {(11<<2)|2,{51,97,0}},
/* 25725 */ {(11<<2)|2,{51,97,0}},
/* 25726 */ {(11<<2)|2,{51,97,0}},
/* 25727 */ {(11<<2)|2,{51,97,0}},
/* 25728 */ {(16<<2)|3,{51,99,48}},
/* 25729 */ {(16<<2)|3,{51,99,49}},
/* 25730 */ {(16<<2)|3,{51,99,50}},
/* 25731 */ {(16<<2)|3,{51,99,97}},
/* 25732 */ {(16<<2)|3,{51,99,99}},
/* 25733 */ {(16<<2)|3,{51,99,101}},
/* 25734 */ {(16<<2)|3,{51,99,105}},
/* 25735 */ {(16<<2)|3,{51,99,111}},
/* 25736 */ {(16<<2)|3,{51,99,115}},
/* 25737 */ {(16<<2)|3,{51,99,116}},
/* 25738 */ {(11<<2)|2,{51,99,0}},
/* 25739 */ {(11<<2)|2,{51,99,0}},
/* 25740 */ {(11<<2)|2,{51,99,0}},
/* 25741 */ {(11<<2)|2,{51,99,0}},
/* 25742 */ {(11<<2)|2,{51,99,0}},
/* 25743 */ {(11<<2)|2,{51,99,0}},
/* 25744 */ {(11<<2)|2,{51,99,0}},
/* 25745 */ {(11<<2)|2,{51,99,0}},
/* 25746 */ {(11<<2)|2,{51,99,0}},
/* 25747 */ {(11<<2)|2,{51,99,0}},
/* 25748 */ {(11<<2)|2,{51,99,0}},
/* 25749 */ {(11<<2)|2,{51,99,0}},
/* 25750 */ {(11<<2)|2,{51,99,0}},
/* 25751 */ {(11<<2)|2,{51,99,0}},
/* 25752 */ {(11<<2)|2,{51,99,0}},
/* 25753 */ {(11<<2)|2,{51,99,0}},
/* 25754 */ {(11<<2)|2,{51,99,0}},
/* 25755 */ {(11<<2)|2,{51,99,0}},
/* 25756 */ {(11<<2)|2,{51,99,0}},
/* 25757 */ {(11<<2)|2,{51,99,0}},
/* 25758 */ {(11<<2)|2,{51,99,0}},
/* 25759 */ {(11<<2)|2,{51,99,0}},
/* 25760 */ {(16<<2)|3,{51,101,48}},
/* 25761 */ {(16<<2)|3,{51,101,49}},
/* 25762 */ {(16<<2)|3,{51,101,50}},
/* 25763 */ {(16<<2)|3,{51,101,97}},
/* 25764 */ {(16<<2)|3,{51,101,99}},
/* 25765 */ {(16<<2)|3,{51,101,101}},
/* 25766 */ {(16<<2)|3,{51,101,105}},
/* 25767 */ {(16<<2)|3,{51,101,111}},
/* 25768 */ {(16<<2)|3,{51,101,115}},
/* 25769 */ {(16<<2)|3,{51,101,116}},
/* 25770 */ {(11<<2)|2,{51,101,0}},
/* 25771 */ {(11<<2)|2,{51,101,0}},
/* 25772 */ {(11<<2)|2,{51,101,0}},
/* 25773 */ {(11<<2)|2,{51,101,0}},
/* 25774 */ {(11<<2)|2,{51,101,0}},
/* 25775 */ {(11<<2)|2,{51,101,0}},
/* 25776 */ {(11<<2)|2,{51,101,0}},
/* 25777 */ {(11<<2)|2,{51,101,0}},
/* 25778 */ {(11<<2)|2,{51,101,0}},
/* 25779 */ {(11<<2)|2,{51,101,0}},
/* 25780 */ {(11<<2)|2,{51,101,0}},
/* 25781 */ {(11<<2)|2,{51,101,0}},
/* 25782 */ {(11<<2)|2,{51,101,0}},
/* 25783 */ {(11<<2)|2,{51,101,0}},
/* 25784 */ {(11<<2)|2,{51,101,0}},
/* 25785 */ {(11<<2)|2,{51,101,0}},
/* 25786 */ {(11<<2)|2,{51,101,0}},
/* 25787 */ {(11<<2)|2,{51,101,0}},
/* 25788 */ {(11<<2)|2,{51,101,0}},
/* 25789 */ {(11<<2)|2,{51,101,0}},
/* 25790 */ {(11<<2)|2,{51,101,0}},
/* 25791 */ {(11<<2)|2,{51,101,0}},
/* 25792 */ {(16<<2)|3,{51,105,48}},
/* 25793 */ {(16<<2)|3,{51,105,49}},
/* 25794 */ {(16<<2)|3,{51,105,50}},
/* 25795 */ {(16<<2)|3,{51,105,97}},
/* 25796 */ {(16<<2)|3,{51,105,99}},
/* 25797 */ {(16<<2)|3,{51,105,101}},
/* 25798 */ {(16<<2)|3,{51,105,105}},
/* 25799 */ {(16<<2)|3,{51,105,111}},
/* 25800 */ {(16<<2)|3,{51,105,115}},
/* 25801 */ {(16<<2)|3,{51,105,116}},
/* 25802 */ {(11<<2)|2,{51,105,0}},
/* 25803 */ {(11<<2)|2,{51,105,0}},
/* 25804 */ {(11<<2)|2,{51,105,0}},
/* 25805 */ {(11<<2)|2,{51,105,0}},
/* 25806 */ {(11<<2)|2,{51,105,0}},
/* 25807 */ {(11<<2)|2,{51,105,0}},
/* 25808 */ {(11<<2)|2,{51,105,0}},
/* 25809 */ {(11<<2)|2,{51,105,0}},
/* 25810 */ {(11<<2)|2,{51,105,0}},
/* 25811 */ {(11<<2)|2,{51,105,0}},
/* 25812 */ {(11<<2)|2,{51,105,0}},
/* 25813 */ {(11<<2)|2,{51,105,0}},
/* 25814 */ {(11<<2)|2,{51,105,0}},
/* 25815 */ {(11<<2)|2,{51,105,0}},
/* 25816 */ {(11<<2)|2,{51,105,0}},
/* 25817 */ {(11<<2)|2,{51,105,0}},
/* 25818 */ {(11<<2)|2,{51,105,0}},
/* 25819 */ {(11<<2)|2,{51,105,0}},
/* 25820 */ {(11<<2)|2,{51,105,0}},
/* 25821 */ {(11<<2)|2,{51,105,0}},
/* 25822 */ {(11<<2)|2,{51,105,0}},
/* 25823 */ {(11<<2)|2,{51,105,0}},
/* 25824 */ {(16<<2)|3,{51,111,48}},
/* 25825 */ {(16<<2)|3,{51,111,49}},
/* 25826 */ {(16<<2)|3,{51,111,50}},
/* 25827 */ {(16<<2)|3,{51,111,97}},
/* 25828 */ {(16<<2)|3,{51,111,99}},
/* 25829 */ {(16<<2)|3,{51,111,101}},
/* 25830 */ {(16<<2)|3,{51,111,105}},
/* 25831 */ {(16<<2)|3,{51,111,111}},
/* 25832 */ {(16<<2)|3,{51,111,115}},
/* 25833 */ {(16<<2)|3,{51,111,116}},
/* 25834 */ {(11<<2)|2,{51,111,0}},
/* 25835 */ {(11<<2)|2,{51,111,0}},
/* 25836 */ {(11<<2)|2,{51,111,0}},
/* 25837 */ {(11<<2)|2,{51,111,0}},
/* 25838 */ {(11<<2)|2,{51,111,0}},
/* 25839 */ {(11<<2)|2,{51,111,0}},
/* 25840 */ {(11<<2)|2,{51,111,0}},
/* 25841 */ {(11<<2)|2,{51,111,0}},
/* 25842 */ {(11<<2)|2,{51,111,0}},
/* 25843 */ {(11<<2)|2,{51,111,0}},
/* 25844 */ {(11<<2)|2,{51,111,0}},
/* 25845 */ {(11<<2)|2,{51,111,0}},
/* 25846 */ {(11<<2)|2,{51,111,0}},
/* 25847 */ {(11<<2)|2,{51,111,0}},
/* 25848 */ {(11<<2)|2,{51,111,0}},
/* 25849 */ {(11<<2)|2,{51,111,0}},
/* 25850 */ {(11<<2)|2,{51,111,0}},
/* 25851 */ {(11<<2)|2,{51,111,0}},
/* 25852 */ {(11<<2)|2,{51,111,0}},
/* 25853 */ {(11<<2)|2,{51,111,0}},
/* 25854 */ {(11<<2)|2,{51,111,0}},
/* 25855 */ {(11<<2)|2,{51,111,0}},
/* 25856 */ {(16<<2)|3,{51,115,48}},
/* 25857 */ {(16<<2)|3,{51,115,49}},
/* 25858 */ {(16<<2)|3,{51,115,50}},
/* 25859 */ {(16<<2)|3,{51,115,97}},
/* 25860 */ {(16<<2)|3,{51,115,99}},
/* 25861 */ {(16<<2)|3,{51,115,101}},
/* 25862 */ {(16<<2)|3,{51,115,105}},
/* 25863 */ {(16<<2)|3,{51,115,111}},
/* 25864 */ {(16<<2)|3,{51,115,115}},
/* 25865 */ {(16<<2)|3,{51,115,116}},
/* 25866 */ {(11<<2)|2,{51,115,0}},
/* 25867 */ {(11<<2)|2,{51,115,0}},
/* 25868 */ {(11<<2)|2,{51,115,0}},
/* 25869 */ {(11<<2)|2,{51,115,0}},
/* 25870 */ {(11<<2)|2,{51,115,0}},
/* 25871 */ {(11<<2)|2,{51,115,0}},
/* 25872 */ {(11<<2)|2,{51,115,0}},
/* 25873 */ {(11<<2)|2,{51,115,0}},
/* 25874 */ {(11<<2)|2,{51,115,0}},
/* 25875 */ {(11<<2)|2,{51,115,0}},
/* 25876 */ {(11<<2)|2,{51,115,0}},
/* 25877 */ {(11<<2)|2,{51,115,0}},
/* 25878 */ {(11<<2)|2,{51,115,0}},
/* 25879 */ {(11<<2)|2,{51,115,0}},
/* 25880 */ {(11<<2)|2,{51,115,0}},
/* 25881 */ {(11<<2)|2,{51,115,0}},
/* 25882 */ {(11<<2)|2,{51,115,0}},
/* 25883 */ {(11<<2)|2,{51,115,0}},
/* 25884 */ {(11<<2)|2,{51,115,0}},
/* 25885 */ {(11<<2)|2,{51,115,0}},
/* 25886 */ {(11<<2)|2,{51,115,0}},
/* 25887 */ {(11<<2)|2,{51,115,0}},
/* 25888 */ {(16<<2)|3,{51,116,48}},
/* 25889 */ {(16<<2)|3,{51,116,49}},
/* 25890 */ {(16<<2)|3,{51,116,50}},
/* 25891 */ {(16<<2)|3,{51,116,97}},
/* 25892 */ {(16<<2)|3,{51,116,99}},
/* 25893 */ {(16<<2)|3,{51,116,101}},
/* 25894 */ {(16<<2)|3,{51,116,105}},
/* 25895 */ {(16<<2)|3,{51,116,111}},
/* 25896 */ {(16<<2)|3,{51,116,115}},
/* 25897 */ {(16<<2)|3,{51,116,116}},
/* 25898 */ {(11<<2)|2,{51,116,0}},
/* 25899 */ {(11<<2)|2,{51,116,0}},
/* 25900 */ {(11<<2)|2,{51,116,0}},
/* 25901 */ {(11<<2)|2,{51,116,0}},
/* 25902 */ {(11<<2)|2,{51,116,0}},
/* 25903 */ {(11<<2)|2,{51,116,0}},
/* 25904 */ {(11<<2)|2,{51,116,0}},
/* 25905 */ {(11<<2)|2,{51,116,0}},
/* 25906 */ {(11<<2)|2,{51,116,0}},
/* 25907 */ {(11<<2)|2,{51,116,0}},
/* 25908 */ {(11<<2)|2,{51,116,0}},
/* 25909 */ {(11<<2)|2,{51,116,0}},
/* 25910 */ {(11<<2)|2,{51,116,0}},
/* 25911 */ {(11<<2)|2,{51,116,0}},
/* 25912 */ {(11<<2)|2,{51,116,0}},
/* 25913 */ {(11<<2)|2,{51,116,0}},
/* 25914 */ {(11<<2)|2,{51,116,0}},
/* 25915 */ {(11<<2)|2,{51,116,0}},
/* 25916 */ {(11<<2)|2,{51,116,0}},
/* 25917 */ {(11<<2)|2,{51,116,0}},
/* 25918 */ {(11<<2)|2,{51,116,0}},
/* 25919 */ {(11<<2)|2,{51,116,0}},
/* 25920 */ {(12<<2)|2,{51,32,0}},
/* 25921 */ {(12<<2)|2,{51,32,0}},
/* 25922 */ {(12<<2)|2,{51,32,0}},
/* 25923 */ {(12<<2)|2,{51,32,0}},
/* 25924 */ {(12<<2)|2,{51,32,0}},
/* 25925 */ {(12<<2)|2,{51,32,0}},
/* 25926 */ {(12<<2)|2,{51,32,0}},
/* 25927 */ {(12<<2)|2,{51,32,0}},
/* 25928 */ {(12<<2)|2,{51,32,0}},
/* 25929 */ {(12<<2)|2,{51,32,0}},
/* 25930 */ {(12<<2)|2,{51,32,0}},
/* 25931 */ {(12<<2)|2,{51,32,0}},
/* 25932 */ {(12<<2)|2,{51,32,0}},
/* 25933 */ {(12<<2)|2,{51,32,0}},
/* 25934 */ {(12<<2)|2,{51,32,0}},
/* 25935 */ {(12<<2)|2,{51,32,0}},
/* 25936 */ {(12<<2)|2,{51,37,0}},
/* 25937 */ {(12<<2)|2,{51,37,0}},
/* 25938 */ {(12<<2)|2,{51,37,0}},
/* 25939 */ {(12<<2)|2,{51,37,0}},
/* 25940 */ {(12<<2)|2,{51,37,0}},
/* 25941 */ {(12<<2)|2,{51,37,0}},
/* 25942 */ {(12<<2)|2,{51,37,0}},
/* 25943 */ {(12<<2)|2,{51,37,0}},
/* 25944 */ {(12<<2)|2,{51,37,0}},
/* 25945 */ {(12<<2)|2,{51,37,0}},
/* 25946 */ {(12<<2)|2,{51,37,0}},
/* 25947 */ {(12<<2)|2,{51,37,0}},
/* 25948 */ {(12<<2)|2,{51,37,0}},
/* 25949 */ {(12<<2)|2,{51,37,0}},
/* 25950 */ {(12<<2)|2,{51,37,0}},
/* 25951 */ {(12<<2)|2,{51,37,0}},
/* 25952 */ {(12<<2)|2,{51,45,0}},
/* 25953 */ {(12<<2)|2,{51,45,0}},
/* 25954 */ {(12<<2)|2,{51,45,0}},
/* 25955 */ {(12<<2)|2,{51,45,0}},
/* 25956 */ {(12<<2)|2,{51,45,0}},
/* 25957 */ {(12<<2)|2,{51,45,0}},
/* 25958 */ {(12<<2)|2,{51,45,0}},
/* 25959 */ {(12<<2)|2,{51,45,0}},
/* 25960 */ {(12<<2)|2,{51,45,0}},
/* 25961 */ {(12<<2)|2,{51,45,0}},
/* 25962 */ {(12<<2)|2,{51,45,0}},
/* 25963 */ {(12<<2)|2,{51,45,0}},
/* 25964 */ {(12<<2)|2,{51,45,0}},
/* 25965 */ {(12<<2)|2,{51,45,0}},
/* 25966 */ {(12<<2)|2,{51,45,0}},
/* 25967 */ {(12<<2)|2,{51,45,0}},
/* 25968 */ {(12<<2)|2,{51,46,0}},
/* 25969 */ {(12<<2)|2,{51,46,0}},
/* 25970 */ {(12<<2)|2,{51,46,0}},
/* 25971 */ {(12<<2)|2,{51,46,0}},
/* 25972 */ {(12<<2)|2,{51,46,0}},
/* 25973 */ {(12<<2)|2,{51,46,0}},
/* 25974 */ {(12<<2)|2,{51,46,0}},
/* 25975 */ {(12<<2)|2,{51,46,0}},
/* 25976 */ {(12<<2)|2,{51,46,0}},
/* 25977 */ {(12<<2)|2,{51,46,0}},
/* 25978 */ {(12<<2)|2,{51,46,0}},
/* 25979 */ {(12<<2)|2,{51,46,0}},
/* 25980 */ {(12<<2)|2,{51,46,0}},
/* 25981 */ {(12<<2)|2,{51,46,0}},
/* 25982 */ {(12<<2)|2,{51,46,0}},
/* 25983 */ {(12<<2)|2,{51,46,0}},
/* 25984 */ {(12<<2)|2,{51,47,0}},
/* 25985 */ {(12<<2)|2,{51,47,0}},
/* 25986 */ {(12<<2)|2,{51,47,0}},
/* 25987 */ {(12<<2)|2,{51,47,0}},
/* 25988 */ {(12<<2)|2,{51,47,0}},
/* 25989 */ {(12<<2)|2,{51,47,0}},
/* 25990 */ {(12<<2)|2,{51,47,0}},
/* 25991 */ {(12<<2)|2,{51,47,0}},
/* 25992 */ {(12<<2)|2,{51,47,0}},
/* 25993 */ {(12<<2)|2,{51,47,0}},
/* 25994 */ {(12<<2)|2,{51,47,0}},
/* 25995 */ {(12<<2)|2,{51,47,0}},
/* 25996 */ {(12<<2)|2,{51,47,0}},
/* 25997 */ {(12<<2)|2,{51,47,0}},
/* 25998 */ {(12<<2)|2,{51,47,0}},
/* 25999 */ {(12<<2)|2,{51,47,0}},
/* 26000 */ {(12<<2)|2,{51,51,0}},
/* 26001 */ {(12<<2)|2,{51,51,0}},
/* 26002 */ {(12<<2)|2,{51,51,0}},
/* 26003 */ {(12<<2)|2,{51,51,0}},
/* 26004 */ {(12<<2)|2,{51,51,0}},
/* 26005 */ {(12<<2)|2,{51,51,0}},
/* 26006 */ {(12<<2)|2,{51,51,0}},
/* 26007 */ {(12<<2)|2,{51,51,0}},
/* 26008 */ {(12<<2)|2,{51,51,0}},
/* 26009 */ {(12<<2)|2,{51,51,0}},
/* 26010 */ {(12<<2)|2,{51,51,0}},
/* 26011 */ {(12<<2)|2,{51,51,0}},
/* 26012 */ {(12<<2)|2,{51,51,0}},
/* 26013 */ {(12<<2)|2,{51,51,0}},
/* 26014 */ {(12<<2)|2,{51,51,0}},
/* 26015 */ {(12<<2)|2,{51,51,0}},
/* 26016 */ {(12<<2)|2,{51,52,0}},
/* 26017 */ {(12<<2)|2,{51,52,0}},
/* 26018 */ {(12<<2)|2,{51,52,0}},
/* 26019 */ {(12<<2)|2,{51,52,0}},
/* 26020 */ {(12<<2)|2,{51,52,0}},
/* 26021 */ {(12<<2)|2,{51,52,0}},
/* 26022 */ {(12<<2)|2,{51,52,0}},
/* 26023 */ {(12<<2)|2,{51,52,0}},
/* 26024 */ {(12<<2)|2,{51,52,0}},
/* 26025 */ {(12<<2)|2,{51,52,0}},
/* 26026 */ {(12<<2)|2,{51,52,0}},
/* 26027 */ {(12<<2)|2,{51,52,0}},
/* 26028 */ {(12<<2)|2,{51,52,0}},
/* 26029 */ {(12<<2)|2,{51,52,0}},
/* 26030 */ {(12<<2)|2,{51,52,0}},
/* 26031 */ {(12<<2)|2,{51,52,0}},
/* 26032 */ {(12<<2)|2,{51,53,0}},
/* 26033 */ {(12<<2)|2,{51,53,0}},
/* 26034 */ {(12<<2)|2,{51,53,0}},
/* 26035 */ {(12<<2)|2,{51,53,0}},
/* 26036 */ {(12<<2)|2,{51,53,0}},
/* 26037 */ {(12<<2)|2,{51,53,0}},
/* 26038 */ {(12<<2)|2,{51,53,0}},
/* 26039 */ {(12<<2)|2,{51,53,0}},
/* 26040 */ {(12<<2)|2,{51,53,0}},
/* 26041 */ {(12<<2)|2,{51,53,0}},
/* 26042 */ {(12<<2)|2,{51,53,0}},
/* 26043 */ {(12<<2)|2,{51,53,0}},
/* 26044 */ {(12<<2)|2,{51,53,0}},
/* 26045 */ {(12<<2)|2,{51,53,0}},
/* 26046 */ {(12<<2)|2,{51,53,0}},
/* 26047 */ {(12<<2)|2,{51,53,0}},
/* 26048 */ {(12<<2)|2,{51,54,0}},
/* 26049 */ {(12<<2)|2,{51,54,0}},
/* 26050 */ {(12<<2)|2,{51,54,0}},
/* 26051 */ {(12<<2)|2,{51,54,0}},
/* 26052 */ {(12<<2)|2,{51,54,0}},
/* 26053 */ {(12<<2)|2,{51,54,0}},
/* 26054 */ {(12<<2)|2,{51,54,0}},
/* 26055 */ {(12<<2)|2,{51,54,0}},
/* 26056 */ {(12<<2)|2,{51,54,0}},
/* 26057 */ {(12<<2)|2,{51,54,0}},
/* 26058 */ {(12<<2)|2,{51,54,0}},
/* 26059 */ {(12<<2)|2,{51,54,0}},
/* 26060 */ {(12<<2)|2,{51,54,0}},
/* 26061 */ {(12<<2)|2,{51,54,0}},
/* 26062 */ {(12<<2)|2,{51,54,0}},
/* 26063 */ {(12<<2)|2,{51,54,0}},
/* 26064 */ {(12<<2)|2,{51,55,0}},
/* 26065 */ {(12<<2)|2,{51,55,0}},
/* 26066 */ {(12<<2)|2,{51,55,0}},
/* 26067 */ {(12<<2)|2,{51,55,0}},
/* 26068 */ {(12<<2)|2,{51,55,0}},
/* 26069 */ {(12<<2)|2,{51,55,0}},
/* 26070 */ {(12<<2)|2,{51,55,0}},
/* 26071 */ {(12<<2)|2,{51,55,0}},
/* 26072 */ {(12<<2)|2,{51,55,0}},
/* 26073 */ {(12<<2)|2,{51,55,0}},
/* 26074 */ {(12<<2)|2,{51,55,0}},
/* 26075 */ {(12<<2)|2,{51,55,0}},
/* 26076 */ {(12<<2)|2,{51,55,0}},
/* 26077 */ {(12<<2)|2,{51,55,0}},
/* 26078 */ {(12<<2)|2,{51,55,0}},
/* 26079 */ {(12<<2)|2,{51,55,0}},
/* 26080 */ {(12<<2)|2,{51,56,0}},
/* 26081 */ {(12<<2)|2,{51,56,0}},
/* 26082 */ {(12<<2)|2,{51,56,0}},
/* 26083 */ {(12<<2)|2,{51,56,0}},
/* 26084 */ {(12<<2)|2,{51,56,0}},
/* 26085 */ {(12<<2)|2,{51,56,0}},
/* 26086 */ {(12<<2)|2,{51,56,0}},
/* 26087 */ {(12<<2)|2,{51,56,0}},
/* 26088 */ {(12<<2)|2,{51,56,0}},
/* 26089 */ {(12<<2)|2,{51,56,0}},
/* 26090 */ {(12<<2)|2,{51,56,0}},
/* 26091 */ {(12<<2)|2,{51,56,0}},
/* 26092 */ {(12<<2)|2,{51,56,0}},
/* 26093 */ {(12<<2)|2,{51,56,0}},
/* 26094 */ {(12<<2)|2,{51,56,0}},
/* 26095 */ {(12<<2)|2,{51,56,0}},
/* 26096 */ {(12<<2)|2,{51,57,0}},
/* 26097 */ {(12<<2)|2,{51,57,0}},
/* 26098 */ {(12<<2)|2,{51,57,0}},
/* 26099 */ {(12<<2)|2,{51,57,0}},
/* 26100 */ {(12<<2)|2,{51,57,0}},
/* 26101 */ {(12<<2)|2,{51,57,0}},
/* 26102 */ {(12<<2)|2,{51,57,0}},
/* 26103 */ {(12<<2)|2,{51,57,0}},
/* 26104 */ {(12<<2)|2,{51,57,0}},
/* 26105 */ {(12<<2)|2,{51,57,0}},
/* 26106 */ {(12<<2)|2,{51,57,0}},
/* 26107 */ {(12<<2)|2,{51,57,0}},
/* 26108 */ {(12<<2)|2,{51,57,0}},
/* 26109 */ {(12<<2)|2,{51,57,0}},
/* 26110 */ {(12<<2)|2,{51,57,0}},
/* 26111 */ {(12<<2)|2,{51,57,0}},
/* 26112 */ {(12<<2)|2,{51,61,0}},
/* 26113 */ {(12<<2)|2,{51,61,0}},
/* 26114 */ {(12<<2)|2,{51,61,0}},
/* 26115 */ {(12<<2)|2,{51,61,0}},
/* 26116 */ {(12<<2)|2,{51,61,0}},
/* 26117 */ {(12<<2)|2,{51,61,0}},
/* 26118 */ {(12<<2)|2,{51,61,0}},
/* 26119 */ {(12<<2)|2,{51,61,0}},
/* 26120 */ {(12<<2)|2,{51,61,0}},
/* 26121 */ {(12<<2)|2,{51,61,0}},
/* 26122 */ {(12<<2)|2,{51,61,0}},
/* 26123 */ {(12<<2)|2,{51,61,0}},
/* 26124 */ {(12<<2)|2,{51,61,0}},
/* 26125 */ {(12<<2)|2,{51,61,0}},
/* 26126 */ {(12<<2)|2,{51,61,0}},
/* 26127 */ {(12<<2)|2,{51,61,0}},
/* 26128 */ {(12<<2)|2,{51,65,0}},
/* 26129 */ {(12<<2)|2,{51,65,0}},
/* 26130 */ {(12<<2)|2,{51,65,0}},
/* 26131 */ {(12<<2)|2,{51,65,0}},
/* 26132 */ {(12<<2)|2,{51,65,0}},
/* 26133 */ {(12<<2)|2,{51,65,0}},
/* 26134 */ {(12<<2)|2,{51,65,0}},
/* 26135 */ {(12<<2)|2,{51,65,0}},
/* 26136 */ {(12<<2)|2,{51,65,0}},
/* 26137 */ {(12<<2)|2,{51,65,0}},
/* 26138 */ {(12<<2)|2,{51,65,0}},
/* 26139 */ {(12<<2)|2,{51,65,0}},
/* 26140 */ {(12<<2)|2,{51,65,0}},
/* 26141 */ {(12<<2)|2,{51,65,0}},
/* 26142 */ {(12<<2)|2,{51,65,0}},
/* 26143 */ {(12<<2)|2,{51,65,0}},
/* 26144 */ {(12<<2)|2,{51,95,0}},
/* 26145 */ {(12<<2)|2,{51,95,0}},
/* 26146 */ {(12<<2)|2,{51,95,0}},
/* 26147 */ {(12<<2)|2,{51,95,0}},
/* 26148 */ {(12<<2)|2,{51,95,0}},
/* 26149 */ {(12<<2)|2,{51,95,0}},
/* 26150 */ {(12<<2)|2,{51,95,0}},
/* 26151 */ {(12<<2)|2,{51,95,0}},
/* 26152 */ {(12<<2)|2,{51,95,0}},
/* 26153 */ {(12<<2)|2,{51,95,0}},
/* 26154 */ {(12<<2)|2,{51,95,0}},
/* 26155 */ {(12<<2)|2,{51,95,0}},
/* 26156 */ {(12<<2)|2,{51,95,0}},
/* 26157 */ {(12<<2)|2,{51,95,0}},
/* 26158 */ {(12<<2)|2,{51,95,0}},
/* 26159 */ {(12<<2)|2,{51,95,0}},
/* 26160 */ {(12<<2)|2,{51,98,0}},
/* 26161 */ {(12<<2)|2,{51,98,0}},
/* 26162 */ {(12<<2)|2,{51,98,0}},
/* 26163 */ {(12<<2)|2,{51,98,0}},
/* 26164 */ {(12<<2)|2,{51,98,0}},
/* 26165 */ {(12<<2)|2,{51,98,0}},
/* 26166 */ {(12<<2)|2,{51,98,0}},
/* 26167 */ {(12<<2)|2,{51,98,0}},
/* 26168 */ {(12<<2)|2,{51,98,0}},
/* 26169 */ {(12<<2)|2,{51,98,0}},
/* 26170 */ {(12<<2)|2,{51,98,0}},
/* 26171 */ {(12<<2)|2,{51,98,0}},
/* 26172 */ {(12<<2)|2,{51,98,0}},
/* 26173 */ {(12<<2)|2,{51,98,0}},
/* 26174 */ {(12<<2)|2,{51,98,0}},
/* 26175 */ {(12<<2)|2,{51,98,0}},
/* 26176 */ {(12<<2)|2,{51,100,0}},
/* 26177 */ {(12<<2)|2,{51,100,0}},
/* 26178 */ {(12<<2)|2,{51,100,0}},
/* 26179 */ {(12<<2)|2,{51,100,0}},
/* 26180 */ {(12<<2)|2,{51,100,0}},
/* 26181 */ {(12<<2)|2,{51,100,0}},
/* 26182 */ {(12<<2)|2,{51,100,0}},
/* 26183 */ {(12<<2)|2,{51,100,0}},
/* 26184 */ {(12<<2)|2,{51,100,0}},
/* 26185 */ {(12<<2)|2,{51,100,0}},
/* 26186 */ {(12<<2)|2,{51,100,0}},
/* 26187 */ {(12<<2)|2,{51,100,0}},
/* 26188 */ {(12<<2)|2,{51,100,0}},
/* 26189 */ {(12<<2)|2,{51,100,0}},
/* 26190 */ {(12<<2)|2,{51,100,0}},
/* 26191 */ {(12<<2)|2,{51,100,0}},
/* 26192 */ {(12<<2)|2,{51,102,0}},
/* 26193 */ {(12<<2)|2,{51,102,0}},
/* 26194 */ {(12<<2)|2,{51,102,0}},
/* 26195 */ {(12<<2)|2,{51,102,0}},
/* 26196 */ {(12<<2)|2,{51,102,0}},
/* 26197 */ {(12<<2)|2,{51,102,0}},
/* 26198 */ {(12<<2)|2,{51,102,0}},
/* 26199 */ {(12<<2)|2,{51,102,0}},
/* 26200 */ {(12<<2)|2,{51,102,0}},
/* 26201 */ {(12<<2)|2,{51,102,0}},
/* 26202 */ {(12<<2)|2,{51,102,0}},
/* 26203 */ {(12<<2)|2,{51,102,0}},
/* 26204 */ {(12<<2)|2,{51,102,0}},
/* 26205 */ {(12<<2)|2,{51,102,0}},
/* 26206 */ {(12<<2)|2,{51,102,0}},
/* 26207 */ {(12<<2)|2,{51,102,0}},
/* 26208 */ {(12<<2)|2,{51,103,0}},
/* 26209 */ {(12<<2)|2,{51,103,0}},
/* 26210 */ {(12<<2)|2,{51,103,0}},
/* 26211 */ {(12<<2)|2,{51,103,0}},
/* 26212 */ {(12<<2)|2,{51,103,0}},
/* 26213 */ {(12<<2)|2,{51,103,0}},
/* 26214 */ {(12<<2)|2,{51,103,0}},
/* 26215 */ {(12<<2)|2,{51,103,0}},
/* 26216 */ {(12<<2)|2,{51,103,0}},
/* 26217 */ {(12<<2)|2,{51,103,0}},
/* 26218 */ {(12<<2)|2,{51,103,0}},
/* 26219 */ {(12<<2)|2,{51,103,0}},
/* 26220 */ {(12<<2)|2,{51,103,0}},
/* 26221 */ {(12<<2)|2,{51,103,0}},
/* 26222 */ {(12<<2)|2,{51,103,0}},
/* 26223 */ {(12<<2)|2,{51,103,0}},
/* 26224 */ {(12<<2)|2,{51,104,0}},
/* 26225 */ {(12<<2)|2,{51,104,0}},
/* 26226 */ {(12<<2)|2,{51,104,0}},
/* 26227 */ {(12<<2)|2,{51,104,0}},
/* 26228 */ {(12<<2)|2,{51,104,0}},
/* 26229 */ {(12<<2)|2,{51,104,0}},
/* 26230 */ {(12<<2)|2,{51,104,0}},
/* 26231 */ {(12<<2)|2,{51,104,0}},
/* 26232 */ {(12<<2)|2,{51,104,0}},
/* 26233 */ {(12<<2)|2,{51,104,0}},
/* 26234 */ {(12<<2)|2,{51,104,0}},
/* 26235 */ {(12<<2)|2,{51,104,0}},
/* 26236 */ {(12<<2)|2,{51,104,0}},
/* 26237 */ {(12<<2)|2,{51,104,0}},
/* 26238 */ {(12<<2)|2,{51,104,0}},
/* 26239 */ {(12<<2)|2,{51,104,0}},
/* 26240 */ {(12<<2)|2,{51,108,0}},
/* 26241 */ {(12<<2)|2,{51,108,0}},
/* 26242 */ {(12<<2)|2,{51,108,0}},
/* 26243 */ {(12<<2)|2,{51,108,0}},
/* 26244 */ {(12<<2)|2,{51,108,0}},
/* 26245 */ {(12<<2)|2,{51,108,0}},
/* 26246 */ {(12<<2)|2,{51,108,0}},
/* 26247 */ {(12<<2)|2,{51,108,0}},
/* 26248 */ {(12<<2)|2,{51,108,0}},
/* 26249 */ {(12<<2)|2,{51,108,0}},
/* 26250 */ {(12<<2)|2,{51,108,0}},
/* 26251 */ {(12<<2)|2,{51,108,0}},
/* 26252 */ {(12<<2)|2,{51,108,0}},
/* 26253 */ {(12<<2)|2,{51,108,0}},
/* 26254 */ {(12<<2)|2,{51,108,0}},
/* 26255 */ {(12<<2)|2,{51,108,0}},
/* 26256 */ {(12<<2)|2,{51,109,0}},
/* 26257 */ {(12<<2)|2,{51,109,0}},
/* 26258 */ {(12<<2)|2,{51,109,0}},
/* 26259 */ {(12<<2)|2,{51,109,0}},
/* 26260 */ {(12<<2)|2,{51,109,0}},
/* 26261 */ {(12<<2)|2,{51,109,0}},
/* 26262 */ {(12<<2)|2,{51,109,0}},
/* 26263 */ {(12<<2)|2,{51,109,0}},
/* 26264 */ {(12<<2)|2,{51,109,0}},
/* 26265 */ {(12<<2)|2,{51,109,0}},
/* 26266 */ {(12<<2)|2,{51,109,0}},
/* 26267 */ {(12<<2)|2,{51,109,0}},
/* 26268 */ {(12<<2)|2,{51,109,0}},
/* 26269 */ {(12<<2)|2,{51,109,0}},
/* 26270 */ {(12<<2)|2,{51,109,0}},
/* 26271 */ {(12<<2)|2,{51,109,0}},
/* 26272 */ {(12<<2)|2,{51,110,0}},
/* 26273 */ {(12<<2)|2,{51,110,0}},
/* 26274 */ {(12<<2)|2,{51,110,0}},
/* 26275 */ {(12<<2)|2,{51,110,0}},
/* 26276 */ {(12<<2)|2,{51,110,0}},
/* 26277 */ {(12<<2)|2,{51,110,0}},
/* 26278 */ {(12<<2)|2,{51,110,0}},
/* 26279 */ {(12<<2)|2,{51,110,0}},
/* 26280 */ {(12<<2)|2,{51,110,0}},
/* 26281 */ {(12<<2)|2,{51,110,0}},
/* 26282 */ {(12<<2)|2,{51,110,0}},
/* 26283 */ {(12<<2)|2,{51,110,0}},
/* 26284 */ {(12<<2)|2,{51,110,0}},
/* 26285 */ {(12<<2)|2,{51,110,0}},
/* 26286 */ {(12<<2)|2,{51,110,0}},
/* 26287 */ {(12<<2)|2,{51,110,0}},
/* 26288 */ {(12<<2)|2,{51,112,0}},
/* 26289 */ {(12<<2)|2,{51,112,0}},
/* 26290 */ {(12<<2)|2,{51,112,0}},
/* 26291 */ {(12<<2)|2,{51,112,0}},
/* 26292 */ {(12<<2)|2,{51,112,0}},
/* 26293 */ {(12<<2)|2,{51,112,0}},
/* 26294 */ {(12<<2)|2,{51,112,0}},
/* 26295 */ {(12<<2)|2,{51,112,0}},
/* 26296 */ {(12<<2)|2,{51,112,0}},
/* 26297 */ {(12<<2)|2,{51,112,0}},
/* 26298 */ {(12<<2)|2,{51,112,0}},
/* 26299 */ {(12<<2)|2,{51,112,0}},
/* 26300 */ {(12<<2)|2,{51,112,0}},
/* 26301 */ {(12<<2)|2,{51,112,0}},
/* 26302 */ {(12<<2)|2,{51,112,0}},
/* 26303 */ {(12<<2)|2,{51,112,0}},
/* 26304 */ {(12<<2)|2,{51,114,0}},
/* 26305 */ {(12<<2)|2,{51,114,0}},
/* 26306 */ {(12<<2)|2,{51,114,0}},
/* 26307 */ {(12<<2)|2,{51,114,0}},
/* 26308 */ {(12<<2)|2,{51,114,0}},
/* 26309 */ {(12<<2)|2,{51,114,0}},
/* 26310 */ {(12<<2)|2,{51,114,0}},
/* 26311 */ {(12<<2)|2,{51,114,0}},
/* 26312 */ {(12<<2)|2,{51,114,0}},
/* 26313 */ {(12<<2)|2,{51,114,0}},
/* 26314 */ {(12<<2)|2,{51,114,0}},
/* 26315 */ {(12<<2)|2,{51,114,0}},
/* 26316 */ {(12<<2)|2,{51,114,0}},
/* 26317 */ {(12<<2)|2,{51,114,0}},
/* 26318 */ {(12<<2)|2,{51,114,0}},
/* 26319 */ {(12<<2)|2,{51,114,0}},
/* 26320 */ {(12<<2)|2,{51,117,0}},
/* 26321 */ {(12<<2)|2,{51,117,0}},
/* 26322 */ {(12<<2)|2,{51,117,0}},
/* 26323 */ {(12<<2)|2,{51,117,0}},
/* 26324 */ {(12<<2)|2,{51,117,0}},
/* 26325 */ {(12<<2)|2,{51,117,0}},
/* 26326 */ {(12<<2)|2,{51,117,0}},
/* 26327 */ {(12<<2)|2,{51,117,0}},
/* 26328 */ {(12<<2)|2,{51,117,0}},
/* 26329 */ {(12<<2)|2,{51,117,0}},
/* 26330 */ {(12<<2)|2,{51,117,0}},
/* 26331 */ {(12<<2)|2,{51,117,0}},
/* 26332 */ {(12<<2)|2,{51,117,0}},
/* 26333 */ {(12<<2)|2,{51,117,0}},
/* 26334 */ {(12<<2)|2,{51,117,0}},
/* 26335 */ {(12<<2)|2,{51,117,0}},
/* 26336 */ {(13<<2)|2,{51,58,0}},
/* 26337 */ {(13<<2)|2,{51,58,0}},
/* 26338 */ {(13<<2)|2,{51,58,0}},
/* 26339 */ {(13<<2)|2,{51,58,0}},
/* 26340 */ {(13<<2)|2,{51,58,0}},
/* 26341 */ {(13<<2)|2,{51,58,0}},
/* 26342 */ {(13<<2)|2,{51,58,0}},
/* 26343 */ {(13<<2)|2,{51,58,0}},
/* 26344 */ {(13<<2)|2,{51,66,0}},
/* 26345 */ {(13<<2)|2,{51,66,0}},
/* 26346 */ {(13<<2)|2,{51,66,0}},
/* 26347 */ {(13<<2)|2,{51,66,0}},
/* 26348 */ {(13<<2)|2,{51,66,0}},
/* 26349 */ {(13<<2)|2,{51,66,0}},
/* 26350 */ {(13<<2)|2,{51,66,0}},
/* 26351 */ {(13<<2)|2,{51,66,0}},
/* 26352 */ {(13<<2)|2,{51,67,0}},
/* 26353 */ {(13<<2)|2,{51,67,0}},
/* 26354 */ {(13<<2)|2,{51,67,0}},
/* 26355 */ {(13<<2)|2,{51,67,0}},
/* 26356 */ {(13<<2)|2,{51,67,0}},
/* 26357 */ {(13<<2)|2,{51,67,0}},
/* 26358 */ {(13<<2)|2,{51,67,0}},
/* 26359 */ {(13<<2)|2,{51,67,0}},
/* 26360 */ {(13<<2)|2,{51,68,0}},
/* 26361 */ {(13<<2)|2,{51,68,0}},
/* 26362 */ {(13<<2)|2,{51,68,0}},
/* 26363 */ {(13<<2)|2,{51,68,0}},
/* 26364 */ {(13<<2)|2,{51,68,0}},
/* 26365 */ {(13<<2)|2,{51,68,0}},
/* 26366 */ {(13<<2)|2,{51,68,0}},
/* 26367 */ {(13<<2)|2,{51,68,0}},
/* 26368 */ {(13<<2)|2,{51,69,0}},
/* 26369 */ {(13<<2)|2,{51,69,0}},
/* 26370 */ {(13<<2)|2,{51,69,0}},
/* 26371 */ {(13<<2)|2,{51,69,0}},
/* 26372 */ {(13<<2)|2,{51,69,0}},
/* 26373 */ {(13<<2)|2,{51,69,0}},
/* 26374 */ {(13<<2)|2,{51,69,0}},
/* 26375 */ {(13<<2)|2,{51,69,0}},
/* 26376 */ {(13<<2)|2,{51,70,0}},
/* 26377 */ {(13<<2)|2,{51,70,0}},
/* 26378 */ {(13<<2)|2,{51,70,0}},
/* 26379 */ {(13<<2)|2,{51,70,0}},
/* 26380 */ {(13<<2)|2,{51,70,0}},
/* 26381 */ {(13<<2)|2,{51,70,0}},
/* 26382 */ {(13<<2)|2,{51,70,0}},
/* 26383 */ {(13<<2)|2,{51,70,0}},
/* 26384 */ {(13<<2)|2,{51,71,0}},
/* 26385 */ {(13<<2)|2,{51,71,0}},
/* 26386 */ {(13<<2)|2,{51,71,0}},
/* 26387 */ {(13<<2)|2,{51,71,0}},
/* 26388 */ {(13<<2)|2,{51,71,0}},
/* 26389 */ {(13<<2)|2,{51,71,0}},
/* 26390 */ {(13<<2)|2,{51,71,0}},
/* 26391 */ {(13<<2)|2,{51,71,0}},
/* 26392 */ {(13<<2)|2,{51,72,0}},
/* 26393 */ {(13<<2)|2,{51,72,0}},
/* 26394 */ {(13<<2)|2,{51,72,0}},
/* 26395 */ {(13<<2)|2,{51,72,0}},
/* 26396 */ {(13<<2)|2,{51,72,0}},
/* 26397 */ {(13<<2)|2,{51,72,0}},
/* 26398 */ {(13<<2)|2,{51,72,0}},
/* 26399 */ {(13<<2)|2,{51,72,0}},
/* 26400 */ {(13<<2)|2,{51,73,0}},
/* 26401 */ {(13<<2)|2,{51,73,0}},
/* 26402 */ {(13<<2)|2,{51,73,0}},
/* 26403 */ {(13<<2)|2,{51,73,0}},
/* 26404 */ {(13<<2)|2,{51,73,0}},
/* 26405 */ {(13<<2)|2,{51,73,0}},
/* 26406 */ {(13<<2)|2,{51,73,0}},
/* 26407 */ {(13<<2)|2,{51,73,0}},
/* 26408 */ {(13<<2)|2,{51,74,0}},
/* 26409 */ {(13<<2)|2,{51,74,0}},
/* 26410 */ {(13<<2)|2,{51,74,0}},
/* 26411 */ {(13<<2)|2,{51,74,0}},
/* 26412 */ {(13<<2)|2,{51,74,0}},
/* 26413 */ {(13<<2)|2,{51,74,0}},
/* 26414 */ {(13<<2)|2,{51,74,0}},
/* 26415 */ {(13<<2)|2,{51,74,0}},
/* 26416 */ {(13<<2)|2,{51,75,0}},
/* 26417 */ {(13<<2)|2,{51,75,0}},
/* 26418 */ {(13<<2)|2,{51,75,0}},
/* 26419 */ {(13<<2)|2,{51,75,0}},
/* 26420 */ {(13<<2)|2,{51,75,0}},
/* 26421 */ {(13<<2)|2,{51,75,0}},
/* 26422 */ {(13<<2)|2,{51,75,0}},
/* 26423 */ {(13<<2)|2,{51,75,0}},
/* 26424 */ {(13<<2)|2,{51,76,0}},
/* 26425 */ {(13<<2)|2,{51,76,0}},
/* 26426 */ {(13<<2)|2,{51,76,0}},
/* 26427 */ {(13<<2)|2,{51,76,0}},
/* 26428 */ {(13<<2)|2,{51,76,0}},
/* 26429 */ {(13<<2)|2,{51,76,0}},
/* 26430 */ {(13<<2)|2,{51,76,0}},
/* 26431 */ {(13<<2)|2,{51,76,0}},
/* 26432 */ {(13<<2)|2,{51,77,0}},
/* 26433 */ {(13<<2)|2,{51,77,0}},
/* 26434 */ {(13<<2)|2,{51,77,0}},
/* 26435 */ {(13<<2)|2,{51,77,0}},
/* 26436 */ {(13<<2)|2,{51,77,0}},
/* 26437 */ {(13<<2)|2,{51,77,0}},
/* 26438 */ {(13<<2)|2,{51,77,0}},
/* 26439 */ {(13<<2)|2,{51,77,0}},
/* 26440 */ {(13<<2)|2,{51,78,0}},
/* 26441 */ {(13<<2)|2,{51,78,0}},
/* 26442 */ {(13<<2)|2,{51,78,0}},
/* 26443 */ {(13<<2)|2,{51,78,0}},
/* 26444 */ {(13<<2)|2,{51,78,0}},
/* 26445 */ {(13<<2)|2,{51,78,0}},
/* 26446 */ {(13<<2)|2,{51,78,0}},
/* 26447 */ {(13<<2)|2,{51,78,0}},
/* 26448 */ {(13<<2)|2,{51,79,0}},
/* 26449 */ {(13<<2)|2,{51,79,0}},
/* 26450 */ {(13<<2)|2,{51,79,0}},
/* 26451 */ {(13<<2)|2,{51,79,0}},
/* 26452 */ {(13<<2)|2,{51,79,0}},
/* 26453 */ {(13<<2)|2,{51,79,0}},
/* 26454 */ {(13<<2)|2,{51,79,0}},
/* 26455 */ {(13<<2)|2,{51,79,0}},
/* 26456 */ {(13<<2)|2,{51,80,0}},
/* 26457 */ {(13<<2)|2,{51,80,0}},
/* 26458 */ {(13<<2)|2,{51,80,0}},
/* 26459 */ {(13<<2)|2,{51,80,0}},
/* 26460 */ {(13<<2)|2,{51,80,0}},
/* 26461 */ {(13<<2)|2,{51,80,0}},
/* 26462 */ {(13<<2)|2,{51,80,0}},
/* 26463 */ {(13<<2)|2,{51,80,0}},
/* 26464 */ {(13<<2)|2,{51,81,0}},
/* 26465 */ {(13<<2)|2,{51,81,0}},
/* 26466 */ {(13<<2)|2,{51,81,0}},
/* 26467 */ {(13<<2)|2,{51,81,0}},
/* 26468 */ {(13<<2)|2,{51,81,0}},
/* 26469 */ {(13<<2)|2,{51,81,0}},
/* 26470 */ {(13<<2)|2,{51,81,0}},
/* 26471 */ {(13<<2)|2,{51,81,0}},
/* 26472 */ {(13<<2)|2,{51,82,0}},
/* 26473 */ {(13<<2)|2,{51,82,0}},
/* 26474 */ {(13<<2)|2,{51,82,0}},
/* 26475 */ {(13<<2)|2,{51,82,0}},
/* 26476 */ {(13<<2)|2,{51,82,0}},
/* 26477 */ {(13<<2)|2,{51,82,0}},
/* 26478 */ {(13<<2)|2,{51,82,0}},
/* 26479 */ {(13<<2)|2,{51,82,0}},
/* 26480 */ {(13<<2)|2,{51,83,0}},
/* 26481 */ {(13<<2)|2,{51,83,0}},
/* 26482 */ {(13<<2)|2,{51,83,0}},
/* 26483 */ {(13<<2)|2,{51,83,0}},
/* 26484 */ {(13<<2)|2,{51,83,0}},
/* 26485 */ {(13<<2)|2,{51,83,0}},
/* 26486 */ {(13<<2)|2,{51,83,0}},
/* 26487 */ {(13<<2)|2,{51,83,0}},
/* 26488 */ {(13<<2)|2,{51,84,0}},
/* 26489 */ {(13<<2)|2,{51,84,0}},
/* 26490 */ {(13<<2)|2,{51,84,0}},
/* 26491 */ {(13<<2)|2,{51,84,0}},
/* 26492 */ {(13<<2)|2,{51,84,0}},
/* 26493 */ {(13<<2)|2,{51,84,0}},
/* 26494 */ {(13<<2)|2,{51,84,0}},
/* 26495 */ {(13<<2)|2,{51,84,0}},
/* 26496 */ {(13<<2)|2,{51,85,0}},
/* 26497 */ {(13<<2)|2,{51,85,0}},
/* 26498 */ {(13<<2)|2,{51,85,0}},
/* 26499 */ {(13<<2)|2,{51,85,0}},
/* 26500 */ {(13<<2)|2,{51,85,0}},
/* 26501 */ {(13<<2)|2,{51,85,0}},
/* 26502 */ {(13<<2)|2,{51,85,0}},
/* 26503 */ {(13<<2)|2,{51,85,0}},
/* 26504 */ {(13<<2)|2,{51,86,0}},
/* 26505 */ {(13<<2)|2,{51,86,0}},
/* 26506 */ {(13<<2)|2,{51,86,0}},
/* 26507 */ {(13<<2)|2,{51,86,0}},
/* 26508 */ {(13<<2)|2,{51,86,0}},
/* 26509 */ {(13<<2)|2,{51,86,0}},
/* 26510 */ {(13<<2)|2,{51,86,0}},
/* 26511 */ {(13<<2)|2,{51,86,0}},
/* 26512 */ {(13<<2)|2,{51,87,0}},
/* 26513 */ {(13<<2)|2,{51,87,0}},
/* 26514 */ {(13<<2)|2,{51,87,0}},
/* 26515 */ {(13<<2)|2,{51,87,0}},
/* 26516 */ {(13<<2)|2,{51,87,0}},
/* 26517 */ {(13<<2)|2,{51,87,0}},
/* 26518 */ {(13<<2)|2,{51,87,0}},
/* 26519 */ {(13<<2)|2,{51,87,0}},
/* 26520 */ {(13<<2)|2,{51,89,0}},
/* 26521 */ {(13<<2)|2,{51,89,0}},
/* 26522 */ {(13<<2)|2,{51,89,0}},
/* 26523 */ {(13<<2)|2,{51,89,0}},
/* 26524 */ {(13<<2)|2,{51,89,0}},
/* 26525 */ {(13<<2)|2,{51,89,0}},
/* 26526 */ {(13<<2)|2,{51,89,0}},
/* 26527 */ {(13<<2)|2,{51,89,0}},
/* 26528 */ {(13<<2)|2,{51,106,0}},
/* 26529 */ {(13<<2)|2,{51,106,0}},
/* 26530 */ {(13<<2)|2,{51,106,0}},
/* 26531 */ {(13<<2)|2,{51,106,0}},
/* 26532 */ {(13<<2)|2,{51,106,0}},
/* 26533 */ {(13<<2)|2,{51,106,0}},
/* 26534 */ {(13<<2)|2,{51,106,0}},
/* 26535 */ {(13<<2)|2,{51,106,0}},
/* 26536 */ {(13<<2)|2,{51,107,0}},
/* 26537 */ {(13<<2)|2,{51,107,0}},
/* 26538 */ {(13<<2)|2,{51,107,0}},
/* 26539 */ {(13<<2)|2,{51,107,0}},
/* 26540 */ {(13<<2)|2,{51,107,0}},
/* 26541 */ {(13<<2)|2,{51,107,0}},
/* 26542 */ {(13<<2)|2,{51,107,0}},
/* 26543 */ {(13<<2)|2,{51,107,0}},
/* 26544 */ {(13<<2)|2,{51,113,0}},
/* 26545 */ {(13<<2)|2,{51,113,0}},
/* 26546 */ {(13<<2)|2,{51,113,0}},
/* 26547 */ {(13<<2)|2,{51,113,0}},
/* 26548 */ {(13<<2)|2,{51,113,0}},
/* 26549 */ {(13<<2)|2,{51,113,0}},
/* 26550 */ {(13<<2)|2,{51,113,0}},
/* 26551 */ {(13<<2)|2,{51,113,0}},
/* 26552 */ {(13<<2)|2,{51,118,0}},
/* 26553 */ {(13<<2)|2,{51,118,0}},
/* 26554 */ {(13<<2)|2,{51,118,0}},
/* 26555 */ {(13<<2)|2,{51,118,0}},
/* 26556 */ {(13<<2)|2,{51,118,0}},
/* 26557 */ {(13<<2)|2,{51,118,0}},
/* 26558 */ {(13<<2)|2,{51,118,0}},
/* 26559 */ {(13<<2)|2,{51,118,0}},
/* 26560 */ {(13<<2)|2,{51,119,0}},
/* 26561 */ {(13<<2)|2,{51,119,0}},
/* 26562 */ {(13<<2)|2,{51,119,0}},
/* 26563 */ {(13<<2)|2,{51,119,0}},
/* 26564 */ {(13<<2)|2,{51,119,0}},
/* 26565 */ {(13<<2)|2,{51,119,0}},
/* 26566 */ {(13<<2)|2,{51,119,0}},
/* 26567 */ {(13<<2)|2,{51,119,0}},
/* 26568 */ {(13<<2)|2,{51,120,0}},
/* 26569 */ {(13<<2)|2,{51,120,0}},
/* 26570 */ {(13<<2)|2,{51,120,0}},
/* 26571 */ {(13<<2)|2,{51,120,0}},
/* 26572 */ {(13<<2)|2,{51,120,0}},
/* 26573 */ {(13<<2)|2,{51,120,0}},
/* 26574 */ {(13<<2)|2,{51,120,0}},
/* 26575 */ {(13<<2)|2,{51,120,0}},
/* 26576 */ {(13<<2)|2,{51,121,0}},
/* 26577 */ {(13<<2)|2,{51,121,0}},
/* 26578 */ {(13<<2)|2,{51,121,0}},
/* 26579 */ {(13<<2)|2,{51,121,0}},
/* 26580 */ {(13<<2)|2,{51,121,0}},
/* 26581 */ {(13<<2)|2,{51,121,0}},
/* 26582 */ {(13<<2)|2,{51,121,0}},
/* 26583 */ {(13<<2)|2,{51,121,0}},
/* 26584 */ {(13<<2)|2,{51,122,0}},
/* 26585 */ {(13<<2)|2,{51,122,0}},
/* 26586 */ {(13<<2)|2,{51,122,0}},
/* 26587 */ {(13<<2)|2,{51,122,0}},
/* 26588 */ {(13<<2)|2,{51,122,0}},
/* 26589 */ {(13<<2)|2,{51,122,0}},
/* 26590 */ {(13<<2)|2,{51,122,0}},
/* 26591 */ {(13<<2)|2,{51,122,0}},
/* 26592 */ {(14<<2)|2,{51,38,0}},
/* 26593 */ {(14<<2)|2,{51,38,0}},
/* 26594 */ {(14<<2)|2,{51,38,0}},
/* 26595 */ {(14<<2)|2,{51,38,0}},
/* 26596 */ {(14<<2)|2,{51,42,0}},
/* 26597 */ {(14<<2)|2,{51,42,0}},
/* 26598 */ {(14<<2)|2,{51,42,0}},
/* 26599 */ {(14<<2)|2,{51,42,0}},
/* 26600 */ {(14<<2)|2,{51,44,0}},
/* 26601 */ {(14<<2)|2,{51,44,0}},
/* 26602 */ {(14<<2)|2,{51,44,0}},
/* 26603 */ {(14<<2)|2,{51,44,0}},
/* 26604 */ {(14<<2)|2,{51,59,0}},
/* 26605 */ {(14<<2)|2,{51,59,0}},
/* 26606 */ {(14<<2)|2,{51,59,0}},
/* 26607 */ {(14<<2)|2,{51,59,0}},
/* 26608 */ {(14<<2)|2,{51,88,0}},
/* 26609 */ {(14<<2)|2,{51,88,0}},
/* 26610 */ {(14<<2)|2,{51,88,0}},
/* 26611 */ {(14<<2)|2,{51,88,0}},
/* 26612 */ {(14<<2)|2,{51,90,0}},
/* 26613 */ {(14<<2)|2,{51,90,0}},
/* 26614 */ {(14<<2)|2,{51,90,0}},
/* 26615 */ {(14<<2)|2,{51,90,0}},
/* 26616 */ {(16<<2)|2,{51,33,0}},
/* 26617 */ {(16<<2)|2,{51,34,0}},
/* 26618 */ {(16<<2)|2,{51,40,0}},
/* 26619 */ {(16<<2)|2,{51,41,0}},
/* 26620 */ {(16<<2)|2,{51,63,0}},
/* 26621 */ {(6<<2)|1,{51,0,0}},
/* 26622 */ {(6<<2)|1,{51,0,0}},
/* 26623 */ {(6<<2)|1,{51,0,0}},
/* 26624 */ {(16<<2)|3,{52,48,48}},
/* 26625 */ {(16<<2)|3,{52,48,49}},
/* 26626 */ {(16<<2)|3,{52,48,50}},
/* 26627 */ {(16<<2)|3,{52,48,97}},
/* 26628 */ {(16<<2)|3,{52,48,99}},
/* 26629 */ {(16<<2)|3,{52,48,101}},
/* 26630 */ {(16<<2)|3,{52,48,105}},
/* 26631 */ {(16<<2)|3,{52,48,111}},
/* 26632 */ {(16<<2)|3,{52,48,115}},
/* 26633 */ {(16<<2)|3,{52,48,116}},
/* 26634 */ {(11<<2)|2,{52,48,0}},
/* 26635 */ {(11<<2)|2,{52,48,0}},
/* 26636 */ {(11<<2)|2,{52,48,0}},
/* 26637 */ {(11<<2)|2,{52,48,0}},
/* 26638 */ {(11<<2)|2,{52,48,0}},
/* 26639 */ {(11<<2)|2,{52,48,0}},
/* 26640 */ {(11<<2)|2,{52,48,0}},
/* 26641 */ {(11<<2)|2,{52,48,0}},
/* 26642 */ {(11<<2)|2,{52,48,0}},
/* 26643 */ {(11<<2)|2,{52,48,0}},
/* 26644 */ {(11<<2)|2,{52,48,0}},
/* 26645 */ {(11<<2)|2,{52,48,0}},
/* 26646 */ {(11<<2)|2,{52,48,0}},
/* 26647 */ {(11<<2)|2,{52,48,0}},
/* 26648 */ {(11<<2)|2,{52,48,0}},
/* 26649 */ {(11<<2)|2,{52,48,0}},
/* 26650 */ {(11<<2)|2,{52,48,0}},
/* 26651 */ {(11<<2)|2,{52,48,0}},
/* 26652 */ {(11<<2)|2,{52,48,0}},
/* 26653 */ {(11<<2)|2,{52,48,0}},
/* 26654 */ {(11<<2)|2,{52,48,0}},
/* 26655 */ {(11<<2)|2,{52,48,0}},
/* 26656 */ {(16<<2)|3,{52,49,48}},
/* 26657 */ {(16<<2)|3,{52,49,49}},
/* 26658 */ {(16<<2)|3,{52,49,50}},
/* 26659 */ {(16<<2)|3,{52,49,97}},
/* 26660 */ {(16<<2)|3,{52,49,99}},
/* 26661 */ {(16<<2)|3,{52,49,101}},
/* 26662 */ {(16<<2)|3,{52,49,105}},
/* 26663 */ {(16<<2)|3,{52,49,111}},
/* 26664 */ {(16<<2)|3,{52,49,115}},
/* 26665 */ {(16<<2)|3,{52,49,116}},
/* 26666 */ {(11<<2)|2,{52,49,0}},
/* 26667 */ {(11<<2)|2,{52,49,0}},
/* 26668 */ {(11<<2)|2,{52,49,0}},
/* 26669 */ {(11<<2)|2,{52,49,0}},
/* 26670 */ {(11<<2)|2,{52,49,0}},
/* 26671 */ {(11<<2)|2,{52,49,0}},
/* 26672 */ {(11<<2)|2,{52,49,0}},
/* 26673 */ {(11<<2)|2,{52,49,0}},
/* 26674 */ {(11<<2)|2,{52,49,0}},
/* 26675 */ {(11<<2)|2,{52,49,0}},
/* 26676 */ {(11<<2)|2,{52,49,0}},
/* 26677 */ {(11<<2)|2,{52,49,0}},
/* 26678 */ {(11<<2)|2,{52,49,0}},
/* 26679 */ {(11<<2)|2,{52,49,0}},
/* 26680 */ {(11<<2)|2,{52,49,0}},
/* 26681 */ {(11<<2)|2,{52,49,0}},
/* 26682 */ {(11<<2)|2,{52,49,0}},
/* 26683 */ {(11<<2)|2,{52,49,0}},
/* 26684 */ {(11<<2)|2,{52,49,0}},
/* 26685 */ {(11<<2)|2,{52,49,0}},
/* 26686 */ {(11<<2)|2,{52,49,0}},
/* 26687 */ {(11<<2)|2,{52,49,0}},
/* 26688 */ {(16<<2)|3,{52,50,48}},
/* 26689 */ {(16<<2)|3,{52,50,49}},
/* 26690 */ {(16<<2)|3,{52,50,50}},
/* 26691 */ {(16<<2)|3,{52,50,97}},
/* 26692 */ {(16<<2)|3,{52,50,99}},
/* 26693 */ {(16<<2)|3,{52,50,101}},
/* 26694 */ {(16<<2)|3,{52,50,105}},
/* 26695 */ {(16<<2)|3,{52,50,111}},
/* 26696 */ {(16<<2)|3,{52,50,115}},
/* 26697 */ {(16<<2)|3,{52,50,116}},
/* 26698 */ {(11<<2)|2,{52,50,0}},
/* 26699 */ {(11<<2)|2,{52,50,0}},
/* 26700 */ {(11<<2)|2,{52,50,0}},
/* 26701 */ {(11<<2)|2,{52,50,0}},
/* 26702 */ {(11<<2)|2,{52,50,0}},
/* 26703 */ {(11<<2)|2,{52,50,0}},
/* 26704 */ {(11<<2)|2,{52,50,0}},
/* 26705 */ {(11<<2)|2,{52,50,0}},
/* 26706 */ {(11<<2)|2,{52,50,0}},
/* 26707 */ {(11<<2)|2,{52,50,0}},
/* 26708 */ {(11<<2)|2,{52,50,0}},
/* 26709 */ {(11<<2)|2,{52,50,0}},
/* 26710 */ {(11<<2)|2,{52,50,0}},
/* 26711 */ {(11<<2)|2,{52,50,0}},
/* 26712 */ {(11<<2)|2,{52,50,0}},
/* 26713 */ {(11<<2)|2,{52,50,0}},
/* 26714 */ {(11<<2)|2,{52,50,0}},
/* 26715 */ {(11<<2)|2,{52,50,0}},
/* 26716 */ {(11<<2)|2,{52,50,0}},
/* 26717 */ {(11<<2)|2,{52,50,0}},
/* 26718 */ {(11<<2)|2,{52,50,0}},
/* 26719 */ {(11<<2)|2,{52,50,0}},
/* 26720 */ {(16<<2)|3,{52,97,48}},
/* 26721 */ {(16<<2)|3,{52,97,49}},
/* 26722 */ {(16<<2)|3,{52,97,50}},
/* 26723 */ {(16<<2)|3,{52,97,97}},
/* 26724 */ {(16<<2)|3,{52,97,99}},
/* 26725 */ {(16<<2)|3,{52,97,101}},
/* 26726 */ {(16<<2)|3,{52,97,105}},
/* 26727 */ {(16<<2)|3,{52,97,111}},
/* 26728 */ {(16<<2)|3,{52,97,115}},
/* 26729 */ {(16<<2)|3,{52,97,116}},
/* 26730 */ {(11<<2)|2,{52,97,0}},
/* 26731 */ {(11<<2)|2,{52,97,0}},
/* 26732 */ {(11<<2)|2,{52,97,0}},
/* 26733 */ {(11<<2)|2,{52,97,0}},
/* 26734 */ {(11<<2)|2,{52,97,0}},
/* 26735 */ {(11<<2)|2,{52,97,0}},
/* 26736 */ {(11<<2)|2,{52,97,0}},
/* 26737 */ {(11<<2)|2,{52,97,0}},
/* 26738 */ {(11<<2)|2,{52,97,0}},
/* 26739 */ {(11<<2)|2,{52,97,0}},
/* 26740 */ {(11<<2)|2,{52,97,0}},
/* 26741 */ {(11<<2)|2,{52,97,0}},
/* 26742 */ {(11<<2)|2,{52,97,0}},
/* 26743 */ {(11<<2)|2,{52,97,0}},
/* 26744 */ {(11<<2)|2,{52,97,0}},
/* 26745 */ {(11<<2)|2,{52,97,0}},
/* 26746 */ {(11<<2)|2,{52,97,0}},
/* 26747 */ {(11<<2)|2,{52,97,0}},
/* 26748 */ {(11<<2)|2,{52,97,0}},
/* 26749 */ {(11<<2)|2,{52,97,0}},
/* 26750 */ {(11<<2)|2,{52,97,0}},
/* 26751 */ {(11<<2)|2,{52,97,0}},
/* 26752 */ {(16<<2)|3,{52,99,48}},
/* 26753 */ {(16<<2)|3,{52,99,49}},
/* 26754 */ {(16<<2)|3,{52,99,50}},
/* 26755 */ {(16<<2)|3,{52,99,97}},
/* 26756 */ {(16<<2)|3,{52,99,99}},
/* 26757 */ {(16<<2)|3,{52,99,101}},
/* 26758 */ {(16<<2)|3,{52,99,105}},
/* 26759 */ {(16<<2)|3,{52,99,111}},
/* 26760 */ {(16<<2)|3,{52,99,115}},
/* 26761 */ {(16<<2)|3,{52,99,116}},
/* 26762 */ {(11<<2)|2,{52,99,0}},
/* 26763 */ {(11<<2)|2,{52,99,0}},
/* 26764 */ {(11<<2)|2,{52,99,0}},
/* 26765 */ {(11<<2)|2,{52,99,0}},
/* 26766 */ {(11<<2)|2,{52,99,0}},
/* 26767 */ {(11<<2)|2,{52,99,0}},
/* 26768 */ {(11<<2)|2,{52,99,0}},
/* 26769 */ {(11<<2)|2,{52,99,0}},
/* 26770 */ {(11<<2)|2,{52,99,0}},
/* 26771 */ {(11<<2)|2,{52,99,0}},
/* 26772 */ {(11<<2)|2,{52,99,0}},
/* 26773 */ {(11<<2)|2,{52,99,0}},
/* 26774 */ {(11<<2)|2,{52,99,0}},
/* 26775 */ {(11<<2)|2,{52,99,0}},
/* 26776 */ {(11<<2)|2,{52,99,0}},
/* 26777 */ {(11<<2)|2,{52,99,0}},
/* 26778 */ {(11<<2)|2,{52,99,0}},
/* 26779 */ {(11<<2)|2,{52,99,0}},
/* 26780 */ {(11<<2)|2,{52,99,0}},
/* 26781 */ {(11<<2)|2,{52,99,0}},
/* 26782 */ {(11<<2)|2,{52,99,0}},
/* 26783 */ {(11<<2)|2,{52,99,0}},
/* 26784 */ {(16<<2)|3,{52,101,48}},
/* 26785 */ {(16<<2)|3,{52,101,49}},
/* 26786 */ {(16<<2)|3,{52,101,50}},
/* 26787 */ {(16<<2)|3,{52,101,97}},
/* 26788 */ {(16<<2)|3,{52,101,99}},
/* 26789 */ {(16<<2)|3,{52,101,101}},
/* 26790 */ {(16<<2)|3,{52,101,105}},
/* 26791 */ {(16<<2)|3,{52,101,111}},
/* 26792 */ {(16<<2)|3,{52,101,115}},
/* 26793 */ {(16<<2)|3,{52,101,116}},
/* 26794 */ {(11<<2)|2,{52,101,0}},
/* 26795 */ {(11<<2)|2,{52,101,0}},
/* 26796 */ {(11<<2)|2,{52,101,0}},
/* 26797 */ {(11<<2)|2,{52,101,0}},
/* 26798 */ {(11<<2)|2,{52,101,0}},
/* 26799 */ {(11<<2)|2,{52,101,0}},
/* 26800 */ {(11<<2)|2,{52,101,0}},
/* 26801 */ {(11<<2)|2,{52,101,0}},
/* 26802 */ {(11<<2)|2,{52,101,0}},
/* 26803 */ {(11<<2)|2,{52,101,0}},
/* 26804 */ {(11<<2)|2,{52,101,0}},
/* 26805 */ {(11<<2)|2,{52,101,0}},
/* 26806 */ {(11<<2)|2,{52,101,0}},
/* 26807 */ {(11<<2)|2,{52,101,0}},
/* 26808 */ {(11<<2)|2,{52,101,0}},
/* 26809 */ {(11<<2)|2,{52,101,0}},
/* 26810 */ {(11<<2)|2,{52,101,0}},
/* 26811 */ {(11<<2)|2,{52,101,0}},
/* 26812 */ {(11<<2)|2,{52,101,0}},
/* 26813 */ {(11<<2)|2,{52,101,0}},
/* 26814 */ {(11<<2)|2,{52,101,0}},
/* 26815 */ {(11<<2)|2,{52,101,0}},
/* 26816 */ {(16<<2)|3,{52,105,48}},
/* 26817 */ {(16<<2)|3,{52,105,49}},
/* 26818 */ {(16<<2)|3,{52,105,50}},
/* 26819 */ {(16<<2)|3,{52,105,97}},
/* 26820 */ {(16<<2)|3,{52,105,99}},
/* 26821 */ {(16<<2)|3,{52,105,101}},
/* 26822 */ {(16<<2)|3,{52,105,105}},
/* 26823 */ {(16<<2)|3,{52,105,111}},
/* 26824 */ {(16<<2)|3,{52,105,115}},
/* 26825 */ {(16<<2)|3,{52,105,116}},
/* 26826 */ {(11<<2)|2,{52,105,0}},
/* 26827 */ {(11<<2)|2,{52,105,0}},
/* 26828 */ {(11<<2)|2,{52,105,0}},
/* 26829 */ {(11<<2)|2,{52,105,0}},
/* 26830 */ {(11<<2)|2,{52,105,0}},
/* 26831 */ {(11<<2)|2,{52,105,0}},
/* 26832 */ {(11<<2)|2,{52,105,0}},
/* 26833 */ {(11<<2)|2,{52,105,0}},
/* 26834 */ {(11<<2)|2,{52,105,0}},
/* 26835 */ {(11<<2)|2,{52,105,0}},
/* 26836 */ {(11<<2)|2,{52,105,0}},
/* 26837 */ {(11<<2)|2,{52,105,0}},
/* 26838 */ {(11<<2)|2,{52,105,0}},
/* 26839 */ {(11<<2)|2,{52,105,0}},
/* 26840 */ {(11<<2)|2,{52,105,0}},
/* 26841 */ {(11<<2)|2,{52,105,0}},
/* 26842 */ {(11<<2)|2,{52,105,0}},
/* 26843 */ {(11<<2)|2,{52,105,0}},
/* 26844 */ {(11<<2)|2,{52,105,0}},
/* 26845 */ {(11<<2)|2,{52,105,0}},
/* 26846 */ {(11<<2)|2,{52,105,0}},
/* 26847 */ {(11<<2)|2,{52,105,0}},
/* 26848 */ {(16<<2)|3,{52,111,48}},
/* 26849 */ {(16<<2)|3,{52,111,49}},
/* 26850 */ {(16<<2)|3,{52,111,50}},
/* 26851 */ {(16<<2)|3,{52,111,97}},
/* 26852 */ {(16<<2)|3,{52,111,99}},
/* 26853 */ {(16<<2)|3,{52,111,101}},
/* 26854 */ {(16<<2)|3,{52,111,105}},
/* 26855 */ {(16<<2)|3,{52,111,111}},
/* 26856 */ {(16<<2)|3,{52,111,115}},
/* 26857 */ {(16<<2)|3,{52,111,116}},
/* 26858 */ {(11<<2)|2,{52,111,0}},
/* 26859 */ {(11<<2)|2,{52,111,0}},
/* 26860 */ {(11<<2)|2,{52,111,0}},
/* 26861 */ {(11<<2)|2,{52,111,0}},
/* 26862 */ {(11<<2)|2,{52,111,0}},
/* 26863 */ {(11<<2)|2,{52,111,0}},
/* 26864 */ {(11<<2)|2,{52,111,0}},
/* 26865 */ {(11<<2)|2,{52,111,0}},
/* 26866 */ {(11<<2)|2,{52,111,0}},
/* 26867 */ {(11<<2)|2,{52,111,0}},
/* 26868 */ {(11<<2)|2,{52,111,0}},
/* 26869 */ {(11<<2)|2,{52,111,0}},
/* 26870 */ {(11<<2)|2,{52,111,0}},
/* 26871 */ {(11<<2)|2,{52,111,0}},
/* 26872 */ {(11<<2)|2,{52,111,0}},
/* 26873 */ {(11<<2)|2,{52,111,0}},
/* 26874 */ {(11<<2)|2,{52,111,0}},
/* 26875 */ {(11<<2)|2,{52,111,0}},
/* 26876 */ {(11<<2)|2,{52,111,0}},
/* 26877 */ {(11<<2)|2,{52,111,0}},
/* 26878 */ {(11<<2)|2,{52,111,0}},
/* 26879 */ {(11<<2)|2,{52,111,0}},
/* 26880 */ {(16<<2)|3,{52,115,48}},
/* 26881 */ {(16<<2)|3,{52,115,49}},
/* 26882 */ {(16<<2)|3,{52,115,50}},
/* 26883 */ {(16<<2)|3,{52,115,97}},
/* 26884 */ {(16<<2)|3,{52,115,99}},
/* 26885 */ {(16<<2)|3,{52,115,101}},
/* 26886 */ {(16<<2)|3,{52,115,105}},
/* 26887 */ {(16<<2)|3,{52,115,111}},
/* 26888 */ {(16<<2)|3,{52,115,115}},
/* 26889 */ {(16<<2)|3,{52,115,116}},
/* 26890 */ {(11<<2)|2,{52,115,0}},
/* 26891 */ {(11<<2)|2,{52,115,0}},
/* 26892 */ {(11<<2)|2,{52,115,0}},
/* 26893 */ {(11<<2)|2,{52,115,0}},
/* 26894 */ {(11<<2)|2,{52,115,0}},
/* 26895 */ {(11<<2)|2,{52,115,0}},
/* 26896 */ {(11<<2)|2,{52,115,0}},
/* 26897 */ {(11<<2)|2,{52,115,0}},
/* 26898 */ {(11<<2)|2,{52,115,0}},
/* 26899 */ {(11<<2)|2,{52,115,0}},
/* 26900 */ {(11<<2)|2,{52,115,0}},
/* 26901 */ {(11<<2)|2,{52,115,0}},
/* 26902 */ {(11<<2)|2,{52,115,0}},
/* 26903 */ {(11<<2)|2,{52,115,0}},
/* 26904 */ {(11<<2)|2,{52,115,0}},
/* 26905 */ {(11<<2)|2,{52,115,0}},
/* 26906 */ {(11<<2)|2,{52,115,0}},
/* 26907 */ {(11<<2)|2,{52,115,0}},
/* 26908 */ {(11<<2)|2,{52,115,0}},
/* 26909 */ {(11<<2)|2,{52,115,0}},
/* 26910 */ {(11<<2)|2,{52,115,0}},
/* 26911 */ {(11<<2)|2,{52,115,0}},
/* 26912 */ {(16<<2)|3,{52,116,48}},
/* 26913 */ {(16<<2)|3,{52,116,49}},
/* 26914 */ {(16<<2)|3,{52,116,50}},
/* 26915 */ {(16<<2)|3,{52,116,97}},
/* 26916 */ {(16<<2)|3,{52,116,99}},
/* 26917 */ {(16<<2)|3,{52,116,101}},
/* 26918 */ {(16<<2)|3,{52,116,105}},
/* 26919 */ {(16<<2)|3,{52,116,111}},
/* 26920 */ {(16<<2)|3,{52,116,115}},
/* 26921 */ {(16<<2)|3,{52,116,116}},
/* 26922 */ {(11<<2)|2,{52,116,0}},
/* 26923 */ {(11<<2)|2,{52,116,0}},
/* 26924 */ {(11<<2)|2,{52,116,0}},
/* 26925 */ {(11<<2)|2,{52,116,0}},
/* 26926 */ {(11<<2)|2,{52,116,0}},
/* 26927 */ {(11<<2)|2,{52,116,0}},
/* 26928 */ {(11<<2)|2,{52,116,0}},
/* 26929 */ {(11<<2)|2,{52,116,0}},
/* 26930 */ {(11<<2)|2,{52,116,0}},
/* 26931 */ {(11<<2)|2,{52,116,0}},
/* 26932 */ {(11<<2)|2,{52,116,0}},
/* 26933 */ {(11<<2)|2,{52,116,0}},
/* 26934 */ {(11<<2)|2,{52,116,0}},
/* 26935 */ {(11<<2)|2,{52,116,0}},
/* 26936 */ {(11<<2)|2,{52,116,0}},
/* 26937 */ {(11<<2)|2,{52,116,0}},
/* 26938 */ {(11<<2)|2,{52,116,0}},
/* 26939 */ {(11<<2)|2,{52,116,0}},
/* 26940 */ {(11<<2)|2,{52,116,0}},
/* 26941 */ {(11<<2)|2,{52,116,0}},
/* 26942 */ {(11<<2)|2,{52,116,0}},
/* 26943 */ {(11<<2)|2,{52,116,0}},
/* 26944 */ {(12<<2)|2,{52,32,0}},
/* 26945 */ {(12<<2)|2,{52,32,0}},
/* 26946 */ {(12<<2)|2,{52,32,0}},
/* 26947 */ {(12<<2)|2,{52,32,0}},
/* 26948 */ {(12<<2)|2,{52,32,0}},
/* 26949 */ {(12<<2)|2,{52,32,0}},
/* 26950 */ {(12<<2)|2,{52,32,0}},
/* 26951 */ {(12<<2)|2,{52,32,0}},
/* 26952 */ {(12<<2)|2,{52,32,0}},
/* 26953 */ {(12<<2)|2,{52,32,0}},
/* 26954 */ {(12<<2)|2,{52,32,0}},
/* 26955 */ {(12<<2)|2,{52,32,0}},
/* 26956 */ {(12<<2)|2,{52,32,0}},
/* 26957 */ {(12<<2)|2,{52,32,0}},
/* 26958 */ {(12<<2)|2,{52,32,0}},
/* 26959 */ {(12<<2)|2,{52,32,0}},
/* 26960 */ {(12<<2)|2,{52,37,0}},
/* 26961 */ {(12<<2)|2,{52,37,0}},
/* 26962 */ {(12<<2)|2,{52,37,0}},
/* 26963 */ {(12<<2)|2,{52,37,0}},
/* 26964 */ {(12<<2)|2,{52,37,0}},
/* 26965 */ {(12<<2)|2,{52,37,0}},
/* 26966 */ {(12<<2)|2,{52,37,0}},
/* 26967 */ {(12<<2)|2,{52,37,0}},
/* 26968 */ {(12<<2)|2,{52,37,0}},
/* 26969 */ {(12<<2)|2,{52,37,0}},
/* 26970 */ {(12<<2)|2,{52,37,0}},
/* 26971 */ {(12<<2)|2,{52,37,0}},
/* 26972 */ {(12<<2)|2,{52,37,0}},
/* 26973 */ {(12<<2)|2,{52,37,0}},
/* 26974 */ {(12<<2)|2,{52,37,0}},
/* 26975 */ {(12<<2)|2,{52,37,0}},
/* 26976 */ {(12<<2)|2,{52,45,0}},
/* 26977 */ {(12<<2)|2,{52,45,0}},
/* 26978 */ {(12<<2)|2,{52,45,0}},
/* 26979 */ {(12<<2)|2,{52,45,0}},
/* 26980 */ {(12<<2)|2,{52,45,0}},
/* 26981 */ {(12<<2)|2,{52,45,0}},
/* 26982 */ {(12<<2)|2,{52,45,0}},
/* 26983 */ {(12<<2)|2,{52,45,0}},
/* 26984 */ {(12<<2)|2,{52,45,0}},
/* 26985 */ {(12<<2)|2,{52,45,0}},
/* 26986 */ {(12<<2)|2,{52,45,0}},
/* 26987 */ {(12<<2)|2,{52,45,0}},
/* 26988 */ {(12<<2)|2,{52,45,0}},
/* 26989 */ {(12<<2)|2,{52,45,0}},
/* 26990 */ {(12<<2)|2,{52,45,0}},
/* 26991 */ {(12<<2)|2,{52,45,0}},
/* 26992 */ {(12<<2)|2,{52,46,0}},
/* 26993 */ {(12<<2)|2,{52,46,0}},
/* 26994 */ {(12<<2)|2,{52,46,0}},
/* 26995 */ {(12<<2)|2,{52,46,0}},
/* 26996 */ {(12<<2)|2,{52,46,0}},
/* 26997 */ {(12<<2)|2,{52,46,0}},
/* 26998 */ {(12<<2)|2,{52,46,0}},
/* 26999 */ {(12<<2)|2,{52,46,0}},
/* 27000 */ {(12<<2)|2,{52,46,0}},
/* 27001 */ {(12<<2)|2,{52,46,0}},
/* 27002 */ {(12<<2)|2,{52,46,0}},
/* 27003 */ {(12<<2)|2,{52,46,0}},
/* 27004 */ {(12<<2)|2,{52,46,0}},
/* 27005 */ {(12<<2)|2,{52,46,0}},
/* 27006 */ {(12<<2)|2,{52,46,0}},
/* 27007 */ {(12<<2)|2,{52,46,0}},
/* 27008 */ {(12<<2)|2,{52,47,0}},
/* 27009 */ {(12<<2)|2,{52,47,0}},
/* 27010 */ {(12<<2)|2,{52,47,0}},
/* 27011 */ {(12<<2)|2,{52,47,0}},
/* 27012 */ {(12<<2)|2,{52,47,0}},
/* 27013 */ {(12<<2)|2,{52,47,0}},
/* 27014 */ {(12<<2)|2,{52,47,0}},
/* 27015 */ {(12<<2)|2,{52,47,0}},
/* 27016 */ {(12<<2)|2,{52,47,0}},
/* 27017 */ {(12<<2)|2,{52,47,0}},
/* 27018 */ {(12<<2)|2,{52,47,0}},
/* 27019 */ {(12<<2)|2,{52,47,0}},
/* 27020 */ {(12<<2)|2,{52,47,0}},
/* 27021 */ {(12<<2)|2,{52,47,0}},
/* 27022 */ {(12<<2)|2,{52,47,0}},
/* 27023 */ {(12<<2)|2,{52,47,0}},
/* 27024 */ {(12<<2)|2,{52,51,0}},
/* 27025 */ {(12<<2)|2,{52,51,0}},
/* 27026 */ {(12<<2)|2,{52,51,0}},
/* 27027 */ {(12<<2)|2,{52,51,0}},
/* 27028 */ {(12<<2)|2,{52,51,0}},
/* 27029 */ {(12<<2)|2,{52,51,0}},
/* 27030 */ {(12<<2)|2,{52,51,0}},
/* 27031 */ {(12<<2)|2,{52,51,0}},
/* 27032 */ {(12<<2)|2,{52,51,0}},
/* 27033 */ {(12<<2)|2,{52,51,0}},
/* 27034 */ {(12<<2)|2,{52,51,0}},
/* 27035 */ {(12<<2)|2,{52,51,0}},
/* 27036 */ {(12<<2)|2,{52,51,0}},
/* 27037 */ {(12<<2)|2,{52,51,0}},
/* 27038 */ {(12<<2)|2,{52,51,0}},
/* 27039 */ {(12<<2)|2,{52,51,0}},
/* 27040 */ {(12<<2)|2,{52,52,0}},
/* 27041 */ {(12<<2)|2,{52,52,0}},
/* 27042 */ {(12<<2)|2,{52,52,0}},
/* 27043 */ {(12<<2)|2,{52,52,0}},
/* 27044 */ {(12<<2)|2,{52,52,0}},
/* 27045 */ {(12<<2)|2,{52,52,0}},
/* 27046 */ {(12<<2)|2,{52,52,0}},
/* 27047 */ {(12<<2)|2,{52,52,0}},
/* 27048 */ {(12<<2)|2,{52,52,0}},
/* 27049 */ {(12<<2)|2,{52,52,0}},
/* 27050 */ {(12<<2)|2,{52,52,0}},
/* 27051 */ {(12<<2)|2,{52,52,0}},
/* 27052 */ {(12<<2)|2,{52,52,0}},
/* 27053 */ {(12<<2)|2,{52,52,0}},
/* 27054 */ {(12<<2)|2,{52,52,0}},
/* 27055 */ {(12<<2)|2,{52,52,0}},
/* 27056 */ {(12<<2)|2,{52,53,0}},
/* 27057 */ {(12<<2)|2,{52,53,0}},
/* 27058 */ {(12<<2)|2,{52,53,0}},
/* 27059 */ {(12<<2)|2,{52,53,0}},
/* 27060 */ {(12<<2)|2,{52,53,0}},
/* 27061 */ {(12<<2)|2,{52,53,0}},
/* 27062 */ {(12<<2)|2,{52,53,0}},
/* 27063 */ {(12<<2)|2,{52,53,0}},
/* 27064 */ {(12<<2)|2,{52,53,0}},
/* 27065 */ {(12<<2)|2,{52,53,0}},
/* 27066 */ {(12<<2)|2,{52,53,0}},
/* 27067 */ {(12<<2)|2,{52,53,0}},
/* 27068 */ {(12<<2)|2,{52,53,0}},
/* 27069 */ {(12<<2)|2,{52,53,0}},
/* 27070 */ {(12<<2)|2,{52,53,0}},
/* 27071 */ {(12<<2)|2,{52,53,0}},
/* 27072 */ {(12<<2)|2,{52,54,0}},
/* 27073 */ {(12<<2)|2,{52,54,0}},
/* 27074 */ {(12<<2)|2,{52,54,0}},
/* 27075 */ {(12<<2)|2,{52,54,0}},
/* 27076 */ {(12<<2)|2,{52,54,0}},
/* 27077 */ {(12<<2)|2,{52,54,0}},
/* 27078 */ {(12<<2)|2,{52,54,0}},
/* 27079 */ {(12<<2)|2,{52,54,0}},
/* 27080 */ {(12<<2)|2,{52,54,0}},
/* 27081 */ {(12<<2)|2,{52,54,0}},
/* 27082 */ {(12<<2)|2,{52,54,0}},
/* 27083 */ {(12<<2)|2,{52,54,0}},
/* 27084 */ {(12<<2)|2,{52,54,0}},
/* 27085 */ {(12<<2)|2,{52,54,0}},
/* 27086 */ {(12<<2)|2,{52,54,0}},
/* 27087 */ {(12<<2)|2,{52,54,0}},
/* 27088 */ {(12<<2)|2,{52,55,0}},
/* 27089 */ {(12<<2)|2,{52,55,0}},
/* 27090 */ {(12<<2)|2,{52,55,0}},
/* 27091 */ {(12<<2)|2,{52,55,0}},
/* 27092 */ {(12<<2)|2,{52,55,0}},
/* 27093 */ {(12<<2)|2,{52,55,0}},
/* 27094 */ {(12<<2)|2,{52,55,0}},
/* 27095 */ {(12<<2)|2,{52,55,0}},
/* 27096 */ {(12<<2)|2,{52,55,0}},
/* 27097 */ {(12<<2)|2,{52,55,0}},
/* 27098 */ {(12<<2)|2,{52,55,0}},
/* 27099 */ {(12<<2)|2,{52,55,0}},
/* 27100 */ {(12<<2)|2,{52,55,0}},
/* 27101 */ {(12<<2)|2,{52,55,0}},
/* 27102 */ {(12<<2)|2,{52,55,0}},
/* 27103 */ {(12<<2)|2,{52,55,0}},
/* 27104 */ {(12<<2)|2,{52,56,0}},
/* 27105 */ {(12<<2)|2,{52,56,0}},
/* 27106 */ {(12<<2)|2,{52,56,0}},
/* 27107 */ {(12<<2)|2,{52,56,0}},
/* 27108 */ {(12<<2)|2,{52,56,0}},
/* 27109 */ {(12<<2)|2,{52,56,0}},
/* 27110 */ {(12<<2)|2,{52,56,0}},
/* 27111 */ {(12<<2)|2,{52,56,0}},
/* 27112 */ {(12<<2)|2,{52,56,0}},
/* 27113 */ {(12<<2)|2,{52,56,0}},
/* 27114 */ {(12<<2)|2,{52,56,0}},
/* 27115 */ {(12<<2)|2,{52,56,0}},
/* 27116 */ {(12<<2)|2,{52,56,0}},
/* 27117 */ {(12<<2)|2,{52,56,0}},
/* 27118 */ {(12<<2)|2,{52,56,0}},
/* 27119 */ {(12<<2)|2,{52,56,0}},
/* 27120 */ {(12<<2)|2,{52,57,0}},
/* 27121 */ {(12<<2)|2,{52,57,0}},
/* 27122 */ {(12<<2)|2,{52,57,0}},
/* 27123 */ {(12<<2)|2,{52,57,0}},
/* 27124 */ {(12<<2)|2,{52,57,0}},
/* 27125 */ {(12<<2)|2,{52,57,0}},
/* 27126 */ {(12<<2)|2,{52,57,0}},
/* 27127 */ {(12<<2)|2,{52,57,0}},
/* 27128 */ {(12<<2)|2,{52,57,0}},
/* 27129 */ {(12<<2)|2,{52,57,0}},
/* 27130 */ {(12<<2)|2,{52,57,0}},
/* 27131 */ {(12<<2)|2,{52,57,0}},
/* 27132 */ {(12<<2)|2,{52,57,0}},
/* 27133 */ {(12<<2)|2,{52,57,0}},
/* 27134 */ {(12<<2)|2,{52,57,0}},
/* 27135 */ {(12<<2)|2,{52,57,0}},
/* 27136 */ {(12<<2)|2,{52,61,0}},
/* 27137 */ {(12<<2)|2,{52,61,0}},
/* 27138 */ {(12<<2)|2,{52,61,0}},
/* 27139 */ {(12<<2)|2,{52,61,0}},
/* 27140 */ {(12<<2)|2,{52,61,0}},
/* 27141 */ {(12<<2)|2,{52,61,0}},
/* 27142 */ {(12<<2)|2,{52,61,0}},
/* 27143 */ {(12<<2)|2,{52,61,0}},
/* 27144 */ {(12<<2)|2,{52,61,0}},
/* 27145 */ {(12<<2)|2,{52,61,0}},
/* 27146 */ {(12<<2)|2,{52,61,0}},
/* 27147 */ {(12<<2)|2,{52,61,0}},
/* 27148 */ {(12<<2)|2,{52,61,0}},
/* 27149 */ {(12<<2)|2,{52,61,0}},
/* 27150 */ {(12<<2)|2,{52,61,0}},
/* 27151 */ {(12<<2)|2,{52,61,0}},
/* 27152 */ {(12<<2)|2,{52,65,0}},
/* 27153 */ {(12<<2)|2,{52,65,0}},
/* 27154 */ {(12<<2)|2,{52,65,0}},
/* 27155 */ {(12<<2)|2,{52,65,0}},
/* 27156 */ {(12<<2)|2,{52,65,0}},
/* 27157 */ {(12<<2)|2,{52,65,0}},
/* 27158 */ {(12<<2)|2,{52,65,0}},
/* 27159 */ {(12<<2)|2,{52,65,0}},
/* 27160 */ {(12<<2)|2,{52,65,0}},
/* 27161 */ {(12<<2)|2,{52,65,0}},
/* 27162 */ {(12<<2)|2,{52,65,0}},
/* 27163 */ {(12<<2)|2,{52,65,0}},
/* 27164 */ {(12<<2)|2,{52,65,0}},
/* 27165 */ {(12<<2)|2,{52,65,0}},
/* 27166 */ {(12<<2)|2,{52,65,0}},
/* 27167 */ {(12<<2)|2,{52,65,0}},
/* 27168 */ {(12<<2)|2,{52,95,0}},
/* 27169 */ {(12<<2)|2,{52,95,0}},
/* 27170 */ {(12<<2)|2,{52,95,0}},
/* 27171 */ {(12<<2)|2,{52,95,0}},
/* 27172 */ {(12<<2)|2,{52,95,0}},
/* 27173 */ {(12<<2)|2,{52,95,0}},
/* 27174 */ {(12<<2)|2,{52,95,0}},
/* 27175 */ {(12<<2)|2,{52,95,0}},
/* 27176 */ {(12<<2)|2,{52,95,0}},
/* 27177 */ {(12<<2)|2,{52,95,0}},
/* 27178 */ {(12<<2)|2,{52,95,0}},
/* 27179 */ {(12<<2)|2,{52,95,0}},
/* 27180 */ {(12<<2)|2,{52,95,0}},
/* 27181 */ {(12<<2)|2,{52,95,0}},
/* 27182 */ {(12<<2)|2,{52,95,0}},
/* 27183 */ {(12<<2)|2,{52,95,0}},
/* 27184 */ {(12<<2)|2,{52,98,0}},
/* 27185 */ {(12<<2)|2,{52,98,0}},
/* 27186 */ {(12<<2)|2,{52,98,0}},
/* 27187 */ {(12<<2)|2,{52,98,0}},
/* 27188 */ {(12<<2)|2,{52,98,0}},
/* 27189 */ {(12<<2)|2,{52,98,0}},
/* 27190 */ {(12<<2)|2,{52,98,0}},
/* 27191 */ {(12<<2)|2,{52,98,0}},
/* 27192 */ {(12<<2)|2,{52,98,0}},
/* 27193 */ {(12<<2)|2,{52,98,0}},
/* 27194 */ {(12<<2)|2,{52,98,0}},
/* 27195 */ {(12<<2)|2,{52,98,0}},
/* 27196 */ {(12<<2)|2,{52,98,0}},
/* 27197 */ {(12<<2)|2,{52,98,0}},
/* 27198 */ {(12<<2)|2,{52,98,0}},
/* 27199 */ {(12<<2)|2,{52,98,0}},
/* 27200 */ {(12<<2)|2,{52,100,0}},
/* 27201 */ {(12<<2)|2,{52,100,0}},
/* 27202 */ {(12<<2)|2,{52,100,0}},
/* 27203 */ {(12<<2)|2,{52,100,0}},
/* 27204 */ {(12<<2)|2,{52,100,0}},
/* 27205 */ {(12<<2)|2,{52,100,0}},
/* 27206 */ {(12<<2)|2,{52,100,0}},
/* 27207 */ {(12<<2)|2,{52,100,0}},
/* 27208 */ {(12<<2)|2,{52,100,0}},
/* 27209 */ {(12<<2)|2,{52,100,0}},
/* 27210 */ {(12<<2)|2,{52,100,0}},
/* 27211 */ {(12<<2)|2,{52,100,0}},
/* 27212 */ {(12<<2)|2,{52,100,0}},
/* 27213 */ {(12<<2)|2,{52,100,0}},
/* 27214 */ {(12<<2)|2,{52,100,0}},
/* 27215 */ {(12<<2)|2,{52,100,0}},
/* 27216 */ {(12<<2)|2,{52,102,0}},
/* 27217 */ {(12<<2)|2,{52,102,0}},
/* 27218 */ {(12<<2)|2,{52,102,0}},
/* 27219 */ {(12<<2)|2,{52,102,0}},
/* 27220 */ {(12<<2)|2,{52,102,0}},
/* 27221 */ {(12<<2)|2,{52,102,0}},
/* 27222 */ {(12<<2)|2,{52,102,0}},
/* 27223 */ {(12<<2)|2,{52,102,0}},
/* 27224 */ {(12<<2)|2,{52,102,0}},
/* 27225 */ {(12<<2)|2,{52,102,0}},
/* 27226 */ {(12<<2)|2,{52,102,0}},
/* 27227 */ {(12<<2)|2,{52,102,0}},
/* 27228 */ {(12<<2)|2,{52,102,0}},
/* 27229 */ {(12<<2)|2,{52,102,0}},
/* 27230 */ {(12<<2)|2,{52,102,0}},
/* 27231 */ {(12<<2)|2,{52,102,0}},
/* 27232 */ {(12<<2)|2,{52,103,0}},
/* 27233 */ {(12<<2)|2,{52,103,0}},
/* 27234 */ {(12<<2)|2,{52,103,0}},
/* 27235 */ {(12<<2)|2,{52,103,0}},
/* 27236 */ {(12<<2)|2,{52,103,0}},
/* 27237 */ {(12<<2)|2,{52,103,0}},
/* 27238 */ {(12<<2)|2,{52,103,0}},
/* 27239 */ {(12<<2)|2,{52,103,0}},
/* 27240 */ {(12<<2)|2,{52,103,0}},
/* 27241 */ {(12<<2)|2,{52,103,0}},
/* 27242 */ {(12<<2)|2,{52,103,0}},
/* 27243 */ {(12<<2)|2,{52,103,0}},
/* 27244 */ {(12<<2)|2,{52,103,0}},
/* 27245 */ {(12<<2)|2,{52,103,0}},
/* 27246 */ {(12<<2)|2,{52,103,0}},
/* 27247 */ {(12<<2)|2,{52,103,0}},
/* 27248 */ {(12<<2)|2,{52,104,0}},
/* 27249 */ {(12<<2)|2,{52,104,0}},
/* 27250 */ {(12<<2)|2,{52,104,0}},
/* 27251 */ {(12<<2)|2,{52,104,0}},
/* 27252 */ {(12<<2)|2,{52,104,0}},
/* 27253 */ {(12<<2)|2,{52,104,0}},
/* 27254 */ {(12<<2)|2,{52,104,0}},
/* 27255 */ {(12<<2)|2,{52,104,0}},
/* 27256 */ {(12<<2)|2,{52,104,0}},
/* 27257 */ {(12<<2)|2,{52,104,0}},
/* 27258 */ {(12<<2)|2,{52,104,0}},
/* 27259 */ {(12<<2)|2,{52,104,0}},
/* 27260 */ {(12<<2)|2,{52,104,0}},
/* 27261 */ {(12<<2)|2,{52,104,0}},
/* 27262 */ {(12<<2)|2,{52,104,0}},
/* 27263 */ {(12<<2)|2,{52,104,0}},
/* 27264 */ {(12<<2)|2,{52,108,0}},
/* 27265 */ {(12<<2)|2,{52,108,0}},
/* 27266 */ {(12<<2)|2,{52,108,0}},
/* 27267 */ {(12<<2)|2,{52,108,0}},
/* 27268 */ {(12<<2)|2,{52,108,0}},
/* 27269 */ {(12<<2)|2,{52,108,0}},
/* 27270 */ {(12<<2)|2,{52,108,0}},
/* 27271 */ {(12<<2)|2,{52,108,0}},
/* 27272 */ {(12<<2)|2,{52,108,0}},
/* 27273 */ {(12<<2)|2,{52,108,0}},
/* 27274 */ {(12<<2)|2,{52,108,0}},
/* 27275 */ {(12<<2)|2,{52,108,0}},
/* 27276 */ {(12<<2)|2,{52,108,0}},
/* 27277 */ {(12<<2)|2,{52,108,0}},
/* 27278 */ {(12<<2)|2,{52,108,0}},
/* 27279 */ {(12<<2)|2,{52,108,0}},
/* 27280 */ {(12<<2)|2,{52,109,0}},
/* 27281 */ {(12<<2)|2,{52,109,0}},
/* 27282 */ {(12<<2)|2,{52,109,0}},
/* 27283 */ {(12<<2)|2,{52,109,0}},
/* 27284 */ {(12<<2)|2,{52,109,0}},
/* 27285 */ {(12<<2)|2,{52,109,0}},
/* 27286 */ {(12<<2)|2,{52,109,0}},
/* 27287 */ {(12<<2)|2,{52,109,0}},
/* 27288 */ {(12<<2)|2,{52,109,0}},
/* 27289 */ {(12<<2)|2,{52,109,0}},
/* 27290 */ {(12<<2)|2,{52,109,0}},
/* 27291 */ {(12<<2)|2,{52,109,0}},
/* 27292 */ {(12<<2)|2,{52,109,0}},
/* 27293 */ {(12<<2)|2,{52,109,0}},
/* 27294 */ {(12<<2)|2,{52,109,0}},
/* 27295 */ {(12<<2)|2,{52,109,0}},
/* 27296 */ {(12<<2)|2,{52,110,0}},
/* 27297 */ {(12<<2)|2,{52,110,0}},
/* 27298 */ {(12<<2)|2,{52,110,0}},
/* 27299 */ {(12<<2)|2,{52,110,0}},
/* 27300 */ {(12<<2)|2,{52,110,0}},
/* 27301 */ {(12<<2)|2,{52,110,0}},
/* 27302 */ {(12<<2)|2,{52,110,0}},
/* 27303 */ {(12<<2)|2,{52,110,0}},
/* 27304 */ {(12<<2)|2,{52,110,0}},
/* 27305 */ {(12<<2)|2,{52,110,0}},
/* 27306 */ {(12<<2)|2,{52,110,0}},
/* 27307 */ {(12<<2)|2,{52,110,0}},
/* 27308 */ {(12<<2)|2,{52,110,0}},
/* 27309 */ {(12<<2)|2,{52,110,0}},
/* 27310 */ {(12<<2)|2,{52,110,0}},
/* 27311 */ {(12<<2)|2,{52,110,0}},
/* 27312 */ {(12<<2)|2,{52,112,0}},
/* 27313 */ {(12<<2)|2,{52,112,0}},
/* 27314 */ {(12<<2)|2,{52,112,0}},
/* 27315 */ {(12<<2)|2,{52,112,0}},
/* 27316 */ {(12<<2)|2,{52,112,0}},
/* 27317 */ {(12<<2)|2,{52,112,0}},
/* 27318 */ {(12<<2)|2,{52,112,0}},
/* 27319 */ {(12<<2)|2,{52,112,0}},
/* 27320 */ {(12<<2)|2,{52,112,0}},
/* 27321 */ {(12<<2)|2,{52,112,0}},
/* 27322 */ {(12<<2)|2,{52,112,0}},
/* 27323 */ {(12<<2)|2,{52,112,0}},
/* 27324 */ {(12<<2)|2,{52,112,0}},
/* 27325 */ {(12<<2)|2,{52,112,0}},
/* 27326 */ {(12<<2)|2,{52,112,0}},
/* 27327 */ {(12<<2)|2,{52,112,0}},
/* 27328 */ {(12<<2)|2,{52,114,0}},
/* 27329 */ {(12<<2)|2,{52,114,0}},
/* 27330 */ {(12<<2)|2,{52,114,0}},
/* 27331 */ {(12<<2)|2,{52,114,0}},
/* 27332 */ {(12<<2)|2,{52,114,0}},
/* 27333 */ {(12<<2)|2,{52,114,0}},
/* 27334 */ {(12<<2)|2,{52,114,0}},
/* 27335 */ {(12<<2)|2,{52,114,0}},
/* 27336 */ {(12<<2)|2,{52,114,0}},
/* 27337 */ {(12<<2)|2,{52,114,0}},
/* 27338 */ {(12<<2)|2,{52,114,0}},
/* 27339 */ {(12<<2)|2,{52,114,0}},
/* 27340 */ {(12<<2)|2,{52,114,0}},
/* 27341 */ {(12<<2)|2,{52,114,0}},
/* 27342 */ {(12<<2)|2,{52,114,0}},
/* 27343 */ {(12<<2)|2,{52,114,0}},
/* 27344 */ {(12<<2)|2,{52,117,0}},
/* 27345 */ {(12<<2)|2,{52,117,0}},
/* 27346 */ {(12<<2)|2,{52,117,0}},
/* 27347 */ {(12<<2)|2,{52,117,0}},
/* 27348 */ {(12<<2)|2,{52,117,0}},
/* 27349 */ {(12<<2)|2,{52,117,0}},
/* 27350 */ {(12<<2)|2,{52,117,0}},
/* 27351 */ {(12<<2)|2,{52,117,0}},
/* 27352 */ {(12<<2)|2,{52,117,0}},
/* 27353 */ {(12<<2)|2,{52,117,0}},
/* 27354 */ {(12<<2)|2,{52,117,0}},
/* 27355 */ {(12<<2)|2,{52,117,0}},
/* 27356 */ {(12<<2)|2,{52,117,0}},
/* 27357 */ {(12<<2)|2,{52,117,0}},
/* 27358 */ {(12<<2)|2,{52,117,0}},
/* 27359 */ {(12<<2)|2,{52,117,0}},
/* 27360 */ {(13<<2)|2,{52,58,0}},
/* 27361 */ {(13<<2)|2,{52,58,0}},
/* 27362 */ {(13<<2)|2,{52,58,0}},
/* 27363 */ {(13<<2)|2,{52,58,0}},
/* 27364 */ {(13<<2)|2,{52,58,0}},
/* 27365 */ {(13<<2)|2,{52,58,0}},
/* 27366 */ {(13<<2)|2,{52,58,0}},
/* 27367 */ {(13<<2)|2,{52,58,0}},
/* 27368 */ {(13<<2)|2,{52,66,0}},
/* 27369 */ {(13<<2)|2,{52,66,0}},
/* 27370 */ {(13<<2)|2,{52,66,0}},
/* 27371 */ {(13<<2)|2,{52,66,0}},
/* 27372 */ {(13<<2)|2,{52,66,0}},
/* 27373 */ {(13<<2)|2,{52,66,0}},
/* 27374 */ {(13<<2)|2,{52,66,0}},
/* 27375 */ {(13<<2)|2,{52,66,0}},
/* 27376 */ {(13<<2)|2,{52,67,0}},
/* 27377 */ {(13<<2)|2,{52,67,0}},
/* 27378 */ {(13<<2)|2,{52,67,0}},
/* 27379 */ {(13<<2)|2,{52,67,0}},
/* 27380 */ {(13<<2)|2,{52,67,0}},
/* 27381 */ {(13<<2)|2,{52,67,0}},
/* 27382 */ {(13<<2)|2,{52,67,0}},
/* 27383 */ {(13<<2)|2,{52,67,0}},
/* 27384 */ {(13<<2)|2,{52,68,0}},
/* 27385 */ {(13<<2)|2,{52,68,0}},
/* 27386 */ {(13<<2)|2,{52,68,0}},
/* 27387 */ {(13<<2)|2,{52,68,0}},
/* 27388 */ {(13<<2)|2,{52,68,0}},
/* 27389 */ {(13<<2)|2,{52,68,0}},
/* 27390 */ {(13<<2)|2,{52,68,0}},
/* 27391 */ {(13<<2)|2,{52,68,0}},
/* 27392 */ {(13<<2)|2,{52,69,0}},
/* 27393 */ {(13<<2)|2,{52,69,0}},
/* 27394 */ {(13<<2)|2,{52,69,0}},
/* 27395 */ {(13<<2)|2,{52,69,0}},
/* 27396 */ {(13<<2)|2,{52,69,0}},
/* 27397 */ {(13<<2)|2,{52,69,0}},
/* 27398 */ {(13<<2)|2,{52,69,0}},
/* 27399 */ {(13<<2)|2,{52,69,0}},
/* 27400 */ {(13<<2)|2,{52,70,0}},
/* 27401 */ {(13<<2)|2,{52,70,0}},
/* 27402 */ {(13<<2)|2,{52,70,0}},
/* 27403 */ {(13<<2)|2,{52,70,0}},
/* 27404 */ {(13<<2)|2,{52,70,0}},
/* 27405 */ {(13<<2)|2,{52,70,0}},
/* 27406 */ {(13<<2)|2,{52,70,0}},
/* 27407 */ {(13<<2)|2,{52,70,0}},
/* 27408 */ {(13<<2)|2,{52,71,0}},
/* 27409 */ {(13<<2)|2,{52,71,0}},
/* 27410 */ {(13<<2)|2,{52,71,0}},
/* 27411 */ {(13<<2)|2,{52,71,0}},
/* 27412 */ {(13<<2)|2,{52,71,0}},
/* 27413 */ {(13<<2)|2,{52,71,0}},
/* 27414 */ {(13<<2)|2,{52,71,0}},
/* 27415 */ {(13<<2)|2,{52,71,0}},
/* 27416 */ {(13<<2)|2,{52,72,0}},
/* 27417 */ {(13<<2)|2,{52,72,0}},
/* 27418 */ {(13<<2)|2,{52,72,0}},
/* 27419 */ {(13<<2)|2,{52,72,0}},
/* 27420 */ {(13<<2)|2,{52,72,0}},
/* 27421 */ {(13<<2)|2,{52,72,0}},
/* 27422 */ {(13<<2)|2,{52,72,0}},
/* 27423 */ {(13<<2)|2,{52,72,0}},
/* 27424 */ {(13<<2)|2,{52,73,0}},
/* 27425 */ {(13<<2)|2,{52,73,0}},
/* 27426 */ {(13<<2)|2,{52,73,0}},
/* 27427 */ {(13<<2)|2,{52,73,0}},
/* 27428 */ {(13<<2)|2,{52,73,0}},
/* 27429 */ {(13<<2)|2,{52,73,0}},
/* 27430 */ {(13<<2)|2,{52,73,0}},
/* 27431 */ {(13<<2)|2,{52,73,0}},
/* 27432 */ {(13<<2)|2,{52,74,0}},
/* 27433 */ {(13<<2)|2,{52,74,0}},
/* 27434 */ {(13<<2)|2,{52,74,0}},
/* 27435 */ {(13<<2)|2,{52,74,0}},
/* 27436 */ {(13<<2)|2,{52,74,0}},
/* 27437 */ {(13<<2)|2,{52,74,0}},
/* 27438 */ {(13<<2)|2,{52,74,0}},
/* 27439 */ {(13<<2)|2,{52,74,0}},
/* 27440 */ {(13<<2)|2,{52,75,0}},
/* 27441 */ {(13<<2)|2,{52,75,0}},
/* 27442 */ {(13<<2)|2,{52,75,0}},
/* 27443 */ {(13<<2)|2,{52,75,0}},
/* 27444 */ {(13<<2)|2,{52,75,0}},
/* 27445 */ {(13<<2)|2,{52,75,0}},
/* 27446 */ {(13<<2)|2,{52,75,0}},
/* 27447 */ {(13<<2)|2,{52,75,0}},
/* 27448 */ {(13<<2)|2,{52,76,0}},
/* 27449 */ {(13<<2)|2,{52,76,0}},
/* 27450 */ {(13<<2)|2,{52,76,0}},
/* 27451 */ {(13<<2)|2,{52,76,0}},
/* 27452 */ {(13<<2)|2,{52,76,0}},
/* 27453 */ {(13<<2)|2,{52,76,0}},
/* 27454 */ {(13<<2)|2,{52,76,0}},
/* 27455 */ {(13<<2)|2,{52,76,0}},
/* 27456 */ {(13<<2)|2,{52,77,0}},
/* 27457 */ {(13<<2)|2,{52,77,0}},
/* 27458 */ {(13<<2)|2,{52,77,0}},
/* 27459 */ {(13<<2)|2,{52,77,0}},
/* 27460 */ {(13<<2)|2,{52,77,0}},
/* 27461 */ {(13<<2)|2,{52,77,0}},
/* 27462 */ {(13<<2)|2,{52,77,0}},
/* 27463 */ {(13<<2)|2,{52,77,0}},
/* 27464 */ {(13<<2)|2,{52,78,0}},
/* 27465 */ {(13<<2)|2,{52,78,0}},
/* 27466 */ {(13<<2)|2,{52,78,0}},
/* 27467 */ {(13<<2)|2,{52,78,0}},
/* 27468 */ {(13<<2)|2,{52,78,0}},
/* 27469 */ {(13<<2)|2,{52,78,0}},
/* 27470 */ {(13<<2)|2,{52,78,0}},
/* 27471 */ {(13<<2)|2,{52,78,0}},
/* 27472 */ {(13<<2)|2,{52,79,0}},
/* 27473 */ {(13<<2)|2,{52,79,0}},
/* 27474 */ {(13<<2)|2,{52,79,0}},
/* 27475 */ {(13<<2)|2,{52,79,0}},
/* 27476 */ {(13<<2)|2,{52,79,0}},
/* 27477 */ {(13<<2)|2,{52,79,0}},
/* 27478 */ {(13<<2)|2,{52,79,0}},
/* 27479 */ {(13<<2)|2,{52,79,0}},
/* 27480 */ {(13<<2)|2,{52,80,0}},
/* 27481 */ {(13<<2)|2,{52,80,0}},
/* 27482 */ {(13<<2)|2,{52,80,0}},
/* 27483 */ {(13<<2)|2,{52,80,0}},
/* 27484 */ {(13<<2)|2,{52,80,0}},
/* 27485 */ {(13<<2)|2,{52,80,0}},
/* 27486 */ {(13<<2)|2,{52,80,0}},
/* 27487 */ {(13<<2)|2,{52,80,0}},
/* 27488 */ {(13<<2)|2,{52,81,0}},
/* 27489 */ {(13<<2)|2,{52,81,0}},
/* 27490 */ {(13<<2)|2,{52,81,0}},
/* 27491 */ {(13<<2)|2,{52,81,0}},
/* 27492 */ {(13<<2)|2,{52,81,0}},
/* 27493 */ {(13<<2)|2,{52,81,0}},
/* 27494 */ {(13<<2)|2,{52,81,0}},
/* 27495 */ {(13<<2)|2,{52,81,0}},
/* 27496 */ {(13<<2)|2,{52,82,0}},
/* 27497 */ {(13<<2)|2,{52,82,0}},
/* 27498 */ {(13<<2)|2,{52,82,0}},
/* 27499 */ {(13<<2)|2,{52,82,0}},
/* 27500 */ {(13<<2)|2,{52,82,0}},
/* 27501 */ {(13<<2)|2,{52,82,0}},
/* 27502 */ {(13<<2)|2,{52,82,0}},
/* 27503 */ {(13<<2)|2,{52,82,0}},
/* 27504 */ {(13<<2)|2,{52,83,0}},
/* 27505 */ {(13<<2)|2,{52,83,0}},
/* 27506 */ {(13<<2)|2,{52,83,0}},
/* 27507 */ {(13<<2)|2,{52,83,0}},
/* 27508 */ {(13<<2)|2,{52,83,0}},
/* 27509 */ {(13<<2)|2,{52,83,0}},
/* 27510 */ {(13<<2)|2,{52,83,0}},
/* 27511 */ {(13<<2)|2,{52,83,0}},
/* 27512 */ {(13<<2)|2,{52,84,0}},
/* 27513 */ {(13<<2)|2,{52,84,0}},
/* 27514 */ {(13<<2)|2,{52,84,0}},
/* 27515 */ {(13<<2)|2,{52,84,0}},
/* 27516 */ {(13<<2)|2,{52,84,0}},
/* 27517 */ {(13<<2)|2,{52,84,0}},
/* 27518 */ {(13<<2)|2,{52,84,0}},
/* 27519 */ {(13<<2)|2,{52,84,0}},
/* 27520 */ {(13<<2)|2,{52,85,0}},
/* 27521 */ {(13<<2)|2,{52,85,0}},
/* 27522 */ {(13<<2)|2,{52,85,0}},
/* 27523 */ {(13<<2)|2,{52,85,0}},
/* 27524 */ {(13<<2)|2,{52,85,0}},
/* 27525 */ {(13<<2)|2,{52,85,0}},
/* 27526 */ {(13<<2)|2,{52,85,0}},
/* 27527 */ {(13<<2)|2,{52,85,0}},
/* 27528 */ {(13<<2)|2,{52,86,0}},
/* 27529 */ {(13<<2)|2,{52,86,0}},
/* 27530 */ {(13<<2)|2,{52,86,0}},
/* 27531 */ {(13<<2)|2,{52,86,0}},
/* 27532 */ {(13<<2)|2,{52,86,0}},
/* 27533 */ {(13<<2)|2,{52,86,0}},
/* 27534 */ {(13<<2)|2,{52,86,0}},
/* 27535 */ {(13<<2)|2,{52,86,0}},
/* 27536 */ {(13<<2)|2,{52,87,0}},
/* 27537 */ {(13<<2)|2,{52,87,0}},
/* 27538 */ {(13<<2)|2,{52,87,0}},
/* 27539 */ {(13<<2)|2,{52,87,0}},
/* 27540 */ {(13<<2)|2,{52,87,0}},
/* 27541 */ {(13<<2)|2,{52,87,0}},
/* 27542 */ {(13<<2)|2,{52,87,0}},
/* 27543 */ {(13<<2)|2,{52,87,0}},
/* 27544 */ {(13<<2)|2,{52,89,0}},
/* 27545 */ {(13<<2)|2,{52,89,0}},
/* 27546 */ {(13<<2)|2,{52,89,0}},
/* 27547 */ {(13<<2)|2,{52,89,0}},
/* 27548 */ {(13<<2)|2,{52,89,0}},
/* 27549 */ {(13<<2)|2,{52,89,0}},
/* 27550 */ {(13<<2)|2,{52,89,0}},
/* 27551 */ {(13<<2)|2,{52,89,0}},
/* 27552 */ {(13<<2)|2,{52,106,0}},
/* 27553 */ {(13<<2)|2,{52,106,0}},
/* 27554 */ {(13<<2)|2,{52,106,0}},
/* 27555 */ {(13<<2)|2,{52,106,0}},
/* 27556 */ {(13<<2)|2,{52,106,0}},
/* 27557 */ {(13<<2)|2,{52,106,0}},
/* 27558 */ {(13<<2)|2,{52,106,0}},
/* 27559 */ {(13<<2)|2,{52,106,0}},
/* 27560 */ {(13<<2)|2,{52,107,0}},
/* 27561 */ {(13<<2)|2,{52,107,0}},
/* 27562 */ {(13<<2)|2,{52,107,0}},
/* 27563 */ {(13<<2)|2,{52,107,0}},
/* 27564 */ {(13<<2)|2,{52,107,0}},
/* 27565 */ {(13<<2)|2,{52,107,0}},
/* 27566 */ {(13<<2)|2,{52,107,0}},
/* 27567 */ {(13<<2)|2,{52,107,0}},
/* 27568 */ {(13<<2)|2,{52,113,0}},
/* 27569 */ {(13<<2)|2,{52,113,0}},
/* 27570 */ {(13<<2)|2,{52,113,0}},
/* 27571 */ {(13<<2)|2,{52,113,0}},
/* 27572 */ {(13<<2)|2,{52,113,0}},
/* 27573 */ {(13<<2)|2,{52,113,0}},
/* 27574 */ {(13<<2)|2,{52,113,0}},
/* 27575 */ {(13<<2)|2,{52,113,0}},
/* 27576 */ {(13<<2)|2,{52,118,0}},
/* 27577 */ {(13<<2)|2,{52,118,0}},
/* 27578 */ {(13<<2)|2,{52,118,0}},
/* 27579 */ {(13<<2)|2,{52,118,0}},
/* 27580 */ {(13<<2)|2,{52,118,0}},
/* 27581 */ {(13<<2)|2,{52,118,0}},
/* 27582 */ {(13<<2)|2,{52,118,0}},
/* 27583 */ {(13<<2)|2,{52,118,0}},
/* 27584 */ {(13<<2)|2,{52,119,0}},
/* 27585 */ {(13<<2)|2,{52,119,0}},
/* 27586 */ {(13<<2)|2,{52,119,0}},
/* 27587 */ {(13<<2)|2,{52,119,0}},
/* 27588 */ {(13<<2)|2,{52,119,0}},
/* 27589 */ {(13<<2)|2,{52,119,0}},
/* 27590 */ {(13<<2)|2,{52,119,0}},
/* 27591 */ {(13<<2)|2,{52,119,0}},
/* 27592 */ {(13<<2)|2,{52,120,0}},
/* 27593 */ {(13<<2)|2,{52,120,0}},
/* 27594 */ {(13<<2)|2,{52,120,0}},
/* 27595 */ {(13<<2)|2,{52,120,0}},
/* 27596 */ {(13<<2)|2,{52,120,0}},
/* 27597 */ {(13<<2)|2,{52,120,0}},
/* 27598 */ {(13<<2)|2,{52,120,0}},
/* 27599 */ {(13<<2)|2,{52,120,0}},
/* 27600 */ {(13<<2)|2,{52,121,0}},
/* 27601 */ {(13<<2)|2,{52,121,0}},
/* 27602 */ {(13<<2)|2,{52,121,0}},
/* 27603 */ {(13<<2)|2,{52,121,0}},
/* 27604 */ {(13<<2)|2,{52,121,0}},
/* 27605 */ {(13<<2)|2,{52,121,0}},
/* 27606 */ {(13<<2)|2,{52,121,0}},
/* 27607 */ {(13<<2)|2,{52,121,0}},
/* 27608 */ {(13<<2)|2,{52,122,0}},
/* 27609 */ {(13<<2)|2,{52,122,0}},
/* 27610 */ {(13<<2)|2,{52,122,0}},
/* 27611 */ {(13<<2)|2,{52,122,0}},
/* 27612 */ {(13<<2)|2,{52,122,0}},
/* 27613 */ {(13<<2)|2,{52,122,0}},
/* 27614 */ {(13<<2)|2,{52,122,0}},
/* 27615 */ {(13<<2)|2,{52,122,0}},
/* 27616 */ {(14<<2)|2,{52,38,0}},
/* 27617 */ {(14<<2)|2,{52,38,0}},
/* 27618 */ {(14<<2)|2,{52,38,0}},
/* 27619 */ {(14<<2)|2,{52,38,0}},
/* 27620 */ {(14<<2)|2,{52,42,0}},
/* 27621 */ {(14<<2)|2,{52,42,0}},
/* 27622 */ {(14<<2)|2,{52,42,0}},
/* 27623 */ {(14<<2)|2,{52,42,0}},
/* 27624 */ {(14<<2)|2,{52,44,0}},
/* 27625 */ {(14<<2)|2,{52,44,0}},
/* 27626 */ {(14<<2)|2,{52,44,0}},
/* 27627 */ {(14<<2)|2,{52,44,0}},
/* 27628 */ {(14<<2)|2,{52,59,0}},
/* 27629 */ {(14<<2)|2,{52,59,0}},
/* 27630 */ {(14<<2)|2,{52,59,0}},
/* 27631 */ {(14<<2)|2,{52,59,0}},
/* 27632 */ {(14<<2)|2,{52,88,0}},
/* 27633 */ {(14<<2)|2,{52,88,0}},
/* 27634 */ {(14<<2)|2,{52,88,0}},
/* 27635 */ {(14<<2)|2,{52,88,0}},
/* 27636 */ {(14<<2)|2,{52,90,0}},
/* 27637 */ {(14<<2)|2,{52,90,0}},
/* 27638 */ {(14<<2)|2,{52,90,0}},
/* 27639 */ {(14<<2)|2,{52,90,0}},
/* 27640 */ {(16<<2)|2,{52,33,0}},
/* 27641 */ {(16<<2)|2,{52,34,0}},
/* 27642 */ {(16<<2)|2,{52,40,0}},
/* 27643 */ {(16<<2)|2,{52,41,0}},
/* 27644 */ {(16<<2)|2,{52,63,0}},
/* 27645 */ {(6<<2)|1,{52,0,0}},
/* 27646 */ {(6<<2)|1,{52,0,0}},
/* 27647 */ {(6<<2)|1,{52,0,0}},
/* 27648 */ {(16<<2)|3,{53,48,48}},
/* 27649 */ {(16<<2)|3,{53,48,49}},
/* 27650 */ {(16<<2)|3,{53,48,50}},
/* 27651 */ {(16<<2)|3,{53,48,97}},
/* 27652 */ {(16<<2)|3,{53,48,99}},
/* 27653 */ {(16<<2)|3,{53,48,101}},
/* 27654 */ {(16<<2)|3,{53,48,105}},
/* 27655 */ {(16<<2)|3,{53,48,111}},
/* 27656 */ {(16<<2)|3,{53,48,115}},
/* 27657 */ {(16<<2)|3,{53,48,116}},
/* 27658 */ {(11<<2)|2,{53,48,0}},
/* 27659 */ {(11<<2)|2,{53,48,0}},
/* 27660 */ {(11<<2)|2,{53,48,0}},
/* 27661 */ {(11<<2)|2,{53,48,0}},
/* 27662 */ {(11<<2)|2,{53,48,0}},
/* 27663 */ {(11<<2)|2,{53,48,0}},
/* 27664 */ {(11<<2)|2,{53,48,0}},
/* 27665 */ {(11<<2)|2,{53,48,0}},
/* 27666 */ {(11<<2)|2,{53,48,0}},
/* 27667 */ {(11<<2)|2,{53,48,0}},
/* 27668 */ {(11<<2)|2,{53,48,0}},
/* 27669 */ {(11<<2)|2,{53,48,0}},
/* 27670 */ {(11<<2)|2,{53,48,0}},
/* 27671 */ {(11<<2)|2,{53,48,0}},
/* 27672 */ {(11<<2)|2,{53,48,0}},
/* 27673 */ {(11<<2)|2,{53,48,0}},
/* 27674 */ {(11<<2)|2,{53,48,0}},
/* 27675 */ {(11<<2)|2,{53,48,0}},
/* 27676 */ {(11<<2)|2,{53,48,0}},
/* 27677 */ {(11<<2)|2,{53,48,0}},
/* 27678 */ {(11<<2)|2,{53,48,0}},
/* 27679 */ {(11<<2)|2,{53,48,0}},
/* 27680 */ {(16<<2)|3,{53,49,48}},
/* 27681 */ {(16<<2)|3,{53,49,49}},
/* 27682 */ {(16<<2)|3,{53,49,50}},
/* 27683 */ {(16<<2)|3,{53,49,97}},
/* 27684 */ {(16<<2)|3,{53,49,99}},
/* 27685 */ {(16<<2)|3,{53,49,101}},
/* 27686 */ {(16<<2)|3,{53,49,105}},
/* 27687 */ {(16<<2)|3,{53,49,111}},
/* 27688 */ {(16<<2)|3,{53,49,115}},
/* 27689 */ {(16<<2)|3,{53,49,116}},
/* 27690 */ {(11<<2)|2,{53,49,0}},
/* 27691 */ {(11<<2)|2,{53,49,0}},
/* 27692 */ {(11<<2)|2,{53,49,0}},
/* 27693 */ {(11<<2)|2,{53,49,0}},
/* 27694 */ {(11<<2)|2,{53,49,0}},
/* 27695 */ {(11<<2)|2,{53,49,0}},
/* 27696 */ {(11<<2)|2,{53,49,0}},
/* 27697 */ {(11<<2)|2,{53,49,0}},
/* 27698 */ {(11<<2)|2,{53,49,0}},
/* 27699 */ {(11<<2)|2,{53,49,0}},
/* 27700 */ {(11<<2)|2,{53,49,0}},
/* 27701 */ {(11<<2)|2,{53,49,0}},
/* 27702 */ {(11<<2)|2,{53,49,0}},
/* 27703 */ {(11<<2)|2,{53,49,0}},
/* 27704 */ {(11<<2)|2,{53,49,0}},
/* 27705 */ {(11<<2)|2,{53,49,0}},
/* 27706 */ {(11<<2)|2,{53,49,0}},
/* 27707 */ {(11<<2)|2,{53,49,0}},
/* 27708 */ {(11<<2)|2,{53,49,0}},
/* 27709 */ {(11<<2)|2,{53,49,0}},
/* 27710 */ {(11<<2)|2,{53,49,0}},
/* 27711 */ {(11<<2)|2,{53,49,0}},
/* 27712 */ {(16<<2)|3,{53,50,48}},
/* 27713 */ {(16<<2)|3,{53,50,49}},
/* 27714 */ {(16<<2)|3,{53,50,50}},
/* 27715 */ {(16<<2)|3,{53,50,97}},
/* 27716 */ {(16<<2)|3,{53,50,99}},
/* 27717 */ {(16<<2)|3,{53,50,101}},
/* 27718 */ {(16<<2)|3,{53,50,105}},
/* 27719 */ {(16<<2)|3,{53,50,111}},
/* 27720 */ {(16<<2)|3,{53,50,115}},
/* 27721 */ {(16<<2)|3,{53,50,116}},
/* 27722 */ {(11<<2)|2,{53,50,0}},
/* 27723 */ {(11<<2)|2,{53,50,0}},
/* 27724 */ {(11<<2)|2,{53,50,0}},
/* 27725 */ {(11<<2)|2,{53,50,0}},
/* 27726 */ {(11<<2)|2,{53,50,0}},
/* 27727 */ {(11<<2)|2,{53,50,0}},
/* 27728 */ {(11<<2)|2,{53,50,0}},
/* 27729 */ {(11<<2)|2,{53,50,0}},
/* 27730 */ {(11<<2)|2,{53,50,0}},
/* 27731 */ {(11<<2)|2,{53,50,0}},
/* 27732 */ {(11<<2)|2,{53,50,0}},
/* 27733 */ {(11<<2)|2,{53,50,0}},
/* 27734 */ {(11<<2)|2,{53,50,0}},
/* 27735 */ {(11<<2)|2,{53,50,0}},
/* 27736 */ {(11<<2)|2,{53,50,0}},
/* 27737 */ {(11<<2)|2,{53,50,0}},
/* 27738 */ {(11<<2)|2,{53,50,0}},
/* 27739 */ {(11<<2)|2,{53,50,0}},
/* 27740 */ {(11<<2)|2,{53,50,0}},
/* 27741 */ {(11<<2)|2,{53,50,0}},
/* 27742 */ {(11<<2)|2,{53,50,0}},
/* 27743 */ {(11<<2)|2,{53,50,0}},
/* 27744 */ {(16<<2)|3,{53,97,48}},
/* 27745 */ {(16<<2)|3,{53,97,49}},
/* 27746 */ {(16<<2)|3,{53,97,50}},
/* 27747 */ {(16<<2)|3,{53,97,97}},
/* 27748 */ {(16<<2)|3,{53,97,99}},
/* 27749 */ {(16<<2)|3,{53,97,101}},
/* 27750 */ {(16<<2)|3,{53,97,105}},
/* 27751 */ {(16<<2)|3,{53,97,111}},
/* 27752 */ {(16<<2)|3,{53,97,115}},
/* 27753 */ {(16<<2)|3,{53,97,116}},
/* 27754 */ {(11<<2)|2,{53,97,0}},
/* 27755 */ {(11<<2)|2,{53,97,0}},
/* 27756 */ {(11<<2)|2,{53,97,0}},
/* 27757 */ {(11<<2)|2,{53,97,0}},
/* 27758 */ {(11<<2)|2,{53,97,0}},
/* 27759 */ {(11<<2)|2,{53,97,0}},
/* 27760 */ {(11<<2)|2,{53,97,0}},
/* 27761 */ {(11<<2)|2,{53,97,0}},
/* 27762 */ {(11<<2)|2,{53,97,0}},
/* 27763 */ {(11<<2)|2,{53,97,0}},
/* 27764 */ {(11<<2)|2,{53,97,0}},
/* 27765 */ {(11<<2)|2,{53,97,0}},
/* 27766 */ {(11<<2)|2,{53,97,0}},
/* 27767 */ {(11<<2)|2,{53,97,0}},
/* 27768 */ {(11<<2)|2,{53,97,0}},
/* 27769 */ {(11<<2)|2,{53,97,0}},
/* 27770 */ {(11<<2)|2,{53,97,0}},
/* 27771 */ {(11<<2)|2,{53,97,0}},
/* 27772 */ {(11<<2)|2,{53,97,0}},
/* 27773 */ {(11<<2)|2,{53,97,0}},
/* 27774 */ {(11<<2)|2,{53,97,0}},
/* 27775 */ {(11<<2)|2,{53,97,0}},
/* 27776 */ {(16<<2)|3,{53,99,48}},
/* 27777 */ {(16<<2)|3,{53,99,49}},
/* 27778 */ {(16<<2)|3,{53,99,50}},
/* 27779 */ {(16<<2)|3,{53,99,97}},
/* 27780 */ {(16<<2)|3,{53,99,99}},
/* 27781 */ {(16<<2)|3,{53,99,101}},
/* 27782 */ {(16<<2)|3,{53,99,105}},
/* 27783 */ {(16<<2)|3,{53,99,111}},
/* 27784 */ {(16<<2)|3,{53,99,115}},
/* 27785 */ {(16<<2)|3,{53,99,116}},
/* 27786 */ {(11<<2)|2,{53,99,0}},
/* 27787 */ {(11<<2)|2,{53,99,0}},
/* 27788 */ {(11<<2)|2,{53,99,0}},
/* 27789 */ {(11<<2)|2,{53,99,0}},
/* 27790 */ {(11<<2)|2,{53,99,0}},
/* 27791 */ {(11<<2)|2,{53,99,0}},
/* 27792 */ {(11<<2)|2,{53,99,0}},
/* 27793 */ {(11<<2)|2,{53,99,0}},
/* 27794 */ {(11<<2)|2,{53,99,0}},
/* 27795 */ {(11<<2)|2,{53,99,0}},
/* 27796 */ {(11<<2)|2,{53,99,0}},
/* 27797 */ {(11<<2)|2,{53,99,0}},
/* 27798 */ {(11<<2)|2,{53,99,0}},
/* 27799 */ {(11<<2)|2,{53,99,0}},
/* 27800 */ {(11<<2)|2,{53,99,0}},
/* 27801 */ {(11<<2)|2,{53,99,0}},
/* 27802 */ {(11<<2)|2,{53,99,0}},
/* 27803 */ {(11<<2)|2,{53,99,0}},
/* 27804 */ {(11<<2)|2,{53,99,0}},
/* 27805 */ {(11<<2)|2,{53,99,0}},
/* 27806 */ {(11<<2)|2,{53,99,0}},
/* 27807 */ {(11<<2)|2,{53,99,0}},
/* 27808 */ {(16<<2)|3,{53,101,48}},
/* 27809 */ {(16<<2)|3,{53,101,49}},
/* 27810 */ {(16<<2)|3,{53,101,50}},
/* 27811 */ {(16<<2)|3,{53,101,97}},
/* 27812 */ {(16<<2)|3,{53,101,99}},
/* 27813 */ {(16<<2)|3,{53,101,101}},
/* 27814 */ {(16<<2)|3,{53,101,105}},
/* 27815 */ {(16<<2)|3,{53,101,111}},
/* 27816 */ {(16<<2)|3,{53,101,115}},
/* 27817 */ {(16<<2)|3,{53,101,116}},
/* 27818 */ {(11<<2)|2,{53,101,0}},
/* 27819 */ {(11<<2)|2,{53,101,0}},
/* 27820 */ {(11<<2)|2,{53,101,0}},
/* 27821 */ {(11<<2)|2,{53,101,0}},
/* 27822 */ {(11<<2)|2,{53,101,0}},
/* 27823 */ {(11<<2)|2,{53,101,0}},
/* 27824 */ {(11<<2)|2,{53,101,0}},
/* 27825 */ {(11<<2)|2,{53,101,0}},
/* 27826 */ {(11<<2)|2,{53,101,0}},
/* 27827 */ {(11<<2)|2,{53,101,0}},
/* 27828 */ {(11<<2)|2,{53,101,0}},
/* 27829 */ {(11<<2)|2,{53,101,0}},
/* 27830 */ {(11<<2)|2,{53,101,0}},
/* 27831 */ {(11<<2)|2,{53,101,0}},
/* 27832 */ {(11<<2)|2,{53,101,0}},
/* 27833 */ {(11<<2)|2,{53,101,0}},
/* 27834 */ {(11<<2)|2,{53,101,0}},
/* 27835 */ {(11<<2)|2,{53,101,0}},
/* 27836 */ {(11<<2)|2,{53,101,0}},
/* 27837 */ {(11<<2)|2,{53,101,0}},
/* 27838 */ {(11<<2)|2,{53,101,0}},
/* 27839 */ {(11<<2)|2,{53,101,0}},
/* 27840 */ {(16<<2)|3,{53,105,48}},
/* 27841 */ {(16<<2)|3,{53,105,49}},
/* 27842 */ {(16<<2)|3,{53,105,50}},
/* 27843 */ {(16<<2)|3,{53,105,97}},
/* 27844 */ {(16<<2)|3,{53,105,99}},
/* 27845 */ {(16<<2)|3,{53,105,101}},
/* 27846 */ {(16<<2)|3,{53,105,105}},
/* 27847 */ {(16<<2)|3,{53,105,111}},
/* 27848 */ {(16<<2)|3,{53,105,115}},
/* 27849 */ {(16<<2)|3,{53,105,116}},
/* 27850 */ {(11<<2)|2,{53,105,0}},
/* 27851 */ {(11<<2)|2,{53,105,0}},
/* 27852 */ {(11<<2)|2,{53,105,0}},
/* 27853 */ {(11<<2)|2,{53,105,0}},
/* 27854 */ {(11<<2)|2,{53,105,0}},
/* 27855 */ {(11<<2)|2,{53,105,0}},
/* 27856 */ {(11<<2)|2,{53,105,0}},
/* 27857 */ {(11<<2)|2,{53,105,0}},
/* 27858 */ {(11<<2)|2,{53,105,0}},
/* 27859 */ {(11<<2)|2,{53,105,0}},
/* 27860 */ {(11<<2)|2,{53,105,0}},
/* 27861 */ {(11<<2)|2,{53,105,0}},
/* 27862 */ {(11<<2)|2,{53,105,0}},
/* 27863 */ {(11<<2)|2,{53,105,0}},
/* 27864 */ {(11<<2)|2,{53,105,0}},
/* 27865 */ {(11<<2)|2,{53,105,0}},
/* 27866 */ {(11<<2)|2,{53,105,0}},
/* 27867 */ {(11<<2)|2,{53,105,0}},
/* 27868 */ {(11<<2)|2,{53,105,0}},
/* 27869 */ {(11<<2)|2,{53,105,0}},
/* 27870 */ {(11<<2)|2,{53,105,0}},
/* 27871 */ {(11<<2)|2,{53,105,0}},
/* 27872 */ {(16<<2)|3,{53,111,48}},
/* 27873 */ {(16<<2)|3,{53,111,49}},
/* 27874 */ {(16<<2)|3,{53,111,50}},
/* 27875 */ {(16<<2)|3,{53,111,97}},
/* 27876 */ {(16<<2)|3,{53,111,99}},
/* 27877 */ {(16<<2)|3,{53,111,101}},
/* 27878 */ {(16<<2)|3,{53,111,105}},
/* 27879 */ {(16<<2)|3,{53,111,111}},
/* 27880 */ {(16<<2)|3,{53,111,115}},
/* 27881 */ {(16<<2)|3,{53,111,116}},
/* 27882 */ {(11<<2)|2,{53,111,0}},
/* 27883 */ {(11<<2)|2,{53,111,0}},
/* 27884 */ {(11<<2)|2,{53,111,0}},
/* 27885 */ {(11<<2)|2,{53,111,0}},
/* 27886 */ {(11<<2)|2,{53,111,0}},
/* 27887 */ {(11<<2)|2,{53,111,0}},
/* 27888 */ {(11<<2)|2,{53,111,0}},
/* 27889 */ {(11<<2)|2,{53,111,0}},
/* 27890 */ {(11<<2)|2,{53,111,0}},
/* 27891 */ {(11<<2)|2,{53,111,0}},
/* 27892 */ {(11<<2)|2,{53,111,0}},
/* 27893 */ {(11<<2)|2,{53,111,0}},
/* 27894 */ {(11<<2)|2,{53,111,0}},
/* 27895 */ {(11<<2)|2,{53,111,0}},
/* 27896 */ {(11<<2)|2,{53,111,0}},
/* 27897 */ {(11<<2)|2,{53,111,0}},
/* 27898 */ {(11<<2)|2,{53,111,0}},
/* 27899 */ {(11<<2)|2,{53,111,0}},
/* 27900 */ {(11<<2)|2,{53,111,0}},
/* 27901 */ {(11<<2)|2,{53,111,0}},
/* 27902 */ {(11<<2)|2,{53,111,0}},
/* 27903 */ {(11<<2)|2,{53,111,0}},
/* 27904 */ {(16<<2)|3,{53,115,48}},
/* 27905 */ {(16<<2)|3,{53,115,49}},
/* 27906 */ {(16<<2)|3,{53,115,50}},
/* 27907 */ {(16<<2)|3,{53,115,97}},
/* 27908 */ {(16<<2)|3,{53,115,99}},
/* 27909 */ {(16<<2)|3,{53,115,101}},
/* 27910 */ {(16<<2)|3,{53,115,105}},
/* 27911 */ {(16<<2)|3,{53,115,111}},
/* 27912 */ {(16<<2)|3,{53,115,115}},
/* 27913 */ {(16<<2)|3,{53,115,116}},
/* 27914 */ {(11<<2)|2,{53,115,0}},
/* 27915 */ {(11<<2)|2,{53,115,0}},
/* 27916 */ {(11<<2)|2,{53,115,0}},
/* 27917 */ {(11<<2)|2,{53,115,0}},
/* 27918 */ {(11<<2)|2,{53,115,0}},
/* 27919 */ {(11<<2)|2,{53,115,0}},
/* 27920 */ {(11<<2)|2,{53,115,0}},
/* 27921 */ {(11<<2)|2,{53,115,0}},
/* 27922 */ {(11<<2)|2,{53,115,0}},
/* 27923 */ {(11<<2)|2,{53,115,0}},
/* 27924 */ {(11<<2)|2,{53,115,0}},
/* 27925 */ {(11<<2)|2,{53,115,0}},
/* 27926 */ {(11<<2)|2,{53,115,0}},
/* 27927 */ {(11<<2)|2,{53,115,0}},
/* 27928 */ {(11<<2)|2,{53,115,0}},
/* 27929 */ {(11<<2)|2,{53,115,0}},
/* 27930 */ {(11<<2)|2,{53,115,0}},
/* 27931 */ {(11<<2)|2,{53,115,0}},
/* 27932 */ {(11<<2)|2,{53,115,0}},
/* 27933 */ {(11<<2)|2,{53,115,0}},
/* 27934 */ {(11<<2)|2,{53,115,0}},
/* 27935 */ {(11<<2)|2,{53,115,0}},
/* 27936 */ {(16<<2)|3,{53,116,48}},
/* 27937 */ {(16<<2)|3,{53,116,49}},
/* 27938 */ {(16<<2)|3,{53,116,50}},
/* 27939 */ {(16<<2)|3,{53,116,97}},
/* 27940 */ {(16<<2)|3,{53,116,99}},
/* 27941 */ {(16<<2)|3,{53,116,101}},
/* 27942 */ {(16<<2)|3,{53,116,105}},
/* 27943 */ {(16<<2)|3,{53,116,111}},
/* 27944 */ {(16<<2)|3,{53,116,115}},
/* 27945 */ {(16<<2)|3,{53,116,116}},
/* 27946 */ {(11<<2)|2,{53,116,0}},
/* 27947 */ {(11<<2)|2,{53,116,0}},
/* 27948 */ {(11<<2)|2,{53,116,0}},
/* 27949 */ {(11<<2)|2,{53,116,0}},
/* 27950 */ {(11<<2)|2,{53,116,0}},
/* 27951 */ {(11<<2)|2,{53,116,0}},
/* 27952 */ {(11<<2)|2,{53,116,0}},
/* 27953 */ {(11<<2)|2,{53,116,0}},
/* 27954 */ {(11<<2)|2,{53,116,0}},
/* 27955 */ {(11<<2)|2,{53,116,0}},
/* 27956 */ {(11<<2)|2,{53,116,0}},
/* 27957 */ {(11<<2)|2,{53,116,0}},
/* 27958 */ {(11<<2)|2,{53,116,0}},
/* 27959 */ {(11<<2)|2,{53,116,0}},
/* 27960 */ {(11<<2)|2,{53,116,0}},
/* 27961 */ {(11<<2)|2,{53,116,0}},
/* 27962 */ {(11<<2)|2,{53,116,0}},
/* 27963 */ {(11<<2)|2,{53,116,0}},
/* 27964 */ {(11<<2)|2,{53,116,0}},
/* 27965 */ {(11<<2)|2,{53,116,0}},
/* 27966 */ {(11<<2)|2,{53,116,0}},
/* 27967 */ {(11<<2)|2,{53,116,0}},
/* 27968 */ {(12<<2)|2,{53,32,0}},
/* 27969 */ {(12<<2)|2,{53,32,0}},
/* 27970 */ {(12<<2)|2,{53,32,0}},
/* 27971 */ {(12<<2)|2,{53,32,0}},
/* 27972 */ {(12<<2)|2,{53,32,0}},
/* 27973 */ {(12<<2)|2,{53,32,0}},
/* 27974 */ {(12<<2)|2,{53,32,0}},
/* 27975 */ {(12<<2)|2,{53,32,0}},
/* 27976 */ {(12<<2)|2,{53,32,0}},
/* 27977 */ {(12<<2)|2,{53,32,0}},
/* 27978 */ {(12<<2)|2,{53,32,0}},
/* 27979 */ {(12<<2)|2,{53,32,0}},
/* 27980 */ {(12<<2)|2,{53,32,0}},
/* 27981 */ {(12<<2)|2,{53,32,0}},
/* 27982 */ {(12<<2)|2,{53,32,0}},
/* 27983 */ {(12<<2)|2,{53,32,0}},
/* 27984 */ {(12<<2)|2,{53,37,0}},
/* 27985 */ {(12<<2)|2,{53,37,0}},
/* 27986 */ {(12<<2)|2,{53,37,0}},
/* 27987 */ {(12<<2)|2,{53,37,0}},
/* 27988 */ {(12<<2)|2,{53,37,0}},
/* 27989 */ {(12<<2)|2,{53,37,0}},
/* 27990 */ {(12<<2)|2,{53,37,0}},
/* 27991 */ {(12<<2)|2,{53,37,0}},
/* 27992 */ {(12<<2)|2,{53,37,0}},
/* 27993 */ {(12<<2)|2,{53,37,0}},
/* 27994 */ {(12<<2)|2,{53,37,0}},
/* 27995 */ {(12<<2)|2,{53,37,0}},
/* 27996 */ {(12<<2)|2,{53,37,0}},
/* 27997 */ {(12<<2)|2,{53,37,0}},
/* 27998 */ {(12<<2)|2,{53,37,0}},
/* 27999 */ {(12<<2)|2,{53,37,0}},
/* 28000 */ {(12<<2)|2,{53,45,0}},
/* 28001 */ {(12<<2)|2,{53,45,0}},
/* 28002 */ {(12<<2)|2,{53,45,0}},
/* 28003 */ {(12<<2)|2,{53,45,0}},
/* 28004 */ {(12<<2)|2,{53,45,0}},
/* 28005 */ {(12<<2)|2,{53,45,0}},
/* 28006 */ {(12<<2)|2,{53,45,0}},
/* 28007 */ {(12<<2)|2,{53,45,0}},
/* 28008 */ {(12<<2)|2,{53,45,0}},
/* 28009 */ {(12<<2)|2,{53,45,0}},
/* 28010 */ {(12<<2)|2,{53,45,0}},
/* 28011 */ {(12<<2)|2,{53,45,0}},
/* 28012 */ {(12<<2)|2,{53,45,0}},
/* 28013 */ {(12<<2)|2,{53,45,0}},
/* 28014 */ {(12<<2)|2,{53,45,0}},
/* 28015 */ {(12<<2)|2,{53,45,0}},
/* 28016 */ {(12<<2)|2,{53,46,0}},
/* 28017 */ {(12<<2)|2,{53,46,0}},
/* 28018 */ {(12<<2)|2,{53,46,0}},
/* 28019 */ {(12<<2)|2,{53,46,0}},
/* 28020 */ {(12<<2)|2,{53,46,0}},
/* 28021 */ {(12<<2)|2,{53,46,0}},
/* 28022 */ {(12<<2)|2,{53,46,0}},
/* 28023 */ {(12<<2)|2,{53,46,0}},
/* 28024 */ {(12<<2)|2,{53,46,0}},
/* 28025 */ {(12<<2)|2,{53,46,0}},
/* 28026 */ {(12<<2)|2,{53,46,0}},
/* 28027 */ {(12<<2)|2,{53,46,0}},
/* 28028 */ {(12<<2)|2,{53,46,0}},
/* 28029 */ {(12<<2)|2,{53,46,0}},
/* 28030 */ {(12<<2)|2,{53,46,0}},
/* 28031 */ {(12<<2)|2,{53,46,0}},
/* 28032 */ {(12<<2)|2,{53,47,0}},
/* 28033 */ {(12<<2)|2,{53,47,0}},
/* 28034 */ {(12<<2)|2,{53,47,0}},
/* 28035 */ {(12<<2)|2,{53,47,0}},
/* 28036 */ {(12<<2)|2,{53,47,0}},
/* 28037 */ {(12<<2)|2,{53,47,0}},
/* 28038 */ {(12<<2)|2,{53,47,0}},
/* 28039 */ {(12<<2)|2,{53,47,0}},
/* 28040 */ {(12<<2)|2,{53,47,0}},
/* 28041 */ {(12<<2)|2,{53,47,0}},
/* 28042 */ {(12<<2)|2,{53,47,0}},
/* 28043 */ {(12<<2)|2,{53,47,0}},
/* 28044 */ {(12<<2)|2,{53,47,0}},
/* 28045 */ {(12<<2)|2,{53,47,0}},
/* 28046 */ {(12<<2)|2,{53,47,0}},
/* 28047 */ {(12<<2)|2,{53,47,0}},
/* 28048 */ {(12<<2)|2,{53,51,0}},
/* 28049 */ {(12<<2)|2,{53,51,0}},
/* 28050 */ {(12<<2)|2,{53,51,0}},
/* 28051 */ {(12<<2)|2,{53,51,0}},
/* 28052 */ {(12<<2)|2,{53,51,0}},
/* 28053 */ {(12<<2)|2,{53,51,0}},
/* 28054 */ {(12<<2)|2,{53,51,0}},
/* 28055 */ {(12<<2)|2,{53,51,0}},
/* 28056 */ {(12<<2)|2,{53,51,0}},
/* 28057 */ {(12<<2)|2,{53,51,0}},
/* 28058 */ {(12<<2)|2,{53,51,0}},
/* 28059 */ {(12<<2)|2,{53,51,0}},
/* 28060 */ {(12<<2)|2,{53,51,0}},
/* 28061 */ {(12<<2)|2,{53,51,0}},
/* 28062 */ {(12<<2)|2,{53,51,0}},
/* 28063 */ {(12<<2)|2,{53,51,0}},
/* 28064 */ {(12<<2)|2,{53,52,0}},
/* 28065 */ {(12<<2)|2,{53,52,0}},
/* 28066 */ {(12<<2)|2,{53,52,0}},
/* 28067 */ {(12<<2)|2,{53,52,0}},
/* 28068 */ {(12<<2)|2,{53,52,0}},
/* 28069 */ {(12<<2)|2,{53,52,0}},
/* 28070 */ {(12<<2)|2,{53,52,0}},
/* 28071 */ {(12<<2)|2,{53,52,0}},
/* 28072 */ {(12<<2)|2,{53,52,0}},
/* 28073 */ {(12<<2)|2,{53,52,0}},
/* 28074 */ {(12<<2)|2,{53,52,0}},
/* 28075 */ {(12<<2)|2,{53,52,0}},
/* 28076 */ {(12<<2)|2,{53,52,0}},
/* 28077 */ {(12<<2)|2,{53,52,0}},
/* 28078 */ {(12<<2)|2,{53,52,0}},
/* 28079 */ {(12<<2)|2,{53,52,0}},
/* 28080 */ {(12<<2)|2,{53,53,0}},
/* 28081 */ {(12<<2)|2,{53,53,0}},
/* 28082 */ {(12<<2)|2,{53,53,0}},
/* 28083 */ {(12<<2)|2,{53,53,0}},
/* 28084 */ {(12<<2)|2,{53,53,0}},
/* 28085 */ {(12<<2)|2,{53,53,0}},
/* 28086 */ {(12<<2)|2,{53,53,0}},
/* 28087 */ {(12<<2)|2,{53,53,0}},
/* 28088 */ {(12<<2)|2,{53,53,0}},
/* 28089 */ {(12<<2)|2,{53,53,0}},
/* 28090 */ {(12<<2)|2,{53,53,0}},
/* 28091 */ {(12<<2)|2,{53,53,0}},
/* 28092 */ {(12<<2)|2,{53,53,0}},
/* 28093 */ {(12<<2)|2,{53,53,0}},
/* 28094 */ {(12<<2)|2,{53,53,0}},
/* 28095 */ {(12<<2)|2,{53,53,0}},
/* 28096 */ {(12<<2)|2,{53,54,0}},
/* 28097 */ {(12<<2)|2,{53,54,0}},
/* 28098 */ {(12<<2)|2,{53,54,0}},
/* 28099 */ {(12<<2)|2,{53,54,0}},
/* 28100 */ {(12<<2)|2,{53,54,0}},
/* 28101 */ {(12<<2)|2,{53,54,0}},
/* 28102 */ {(12<<2)|2,{53,54,0}},
/* 28103 */ {(12<<2)|2,{53,54,0}},
/* 28104 */ {(12<<2)|2,{53,54,0}},
/* 28105 */ {(12<<2)|2,{53,54,0}},
/* 28106 */ {(12<<2)|2,{53,54,0}},
/* 28107 */ {(12<<2)|2,{53,54,0}},
/* 28108 */ {(12<<2)|2,{53,54,0}},
/* 28109 */ {(12<<2)|2,{53,54,0}},
/* 28110 */ {(12<<2)|2,{53,54,0}},
/* 28111 */ {(12<<2)|2,{53,54,0}},
/* 28112 */ {(12<<2)|2,{53,55,0}},
/* 28113 */ {(12<<2)|2,{53,55,0}},
/* 28114 */ {(12<<2)|2,{53,55,0}},
/* 28115 */ {(12<<2)|2,{53,55,0}},
/* 28116 */ {(12<<2)|2,{53,55,0}},
/* 28117 */ {(12<<2)|2,{53,55,0}},
/* 28118 */ {(12<<2)|2,{53,55,0}},
/* 28119 */ {(12<<2)|2,{53,55,0}},
/* 28120 */ {(12<<2)|2,{53,55,0}},
/* 28121 */ {(12<<2)|2,{53,55,0}},
/* 28122 */ {(12<<2)|2,{53,55,0}},
/* 28123 */ {(12<<2)|2,{53,55,0}},
/* 28124 */ {(12<<2)|2,{53,55,0}},
/* 28125 */ {(12<<2)|2,{53,55,0}},
/* 28126 */ {(12<<2)|2,{53,55,0}},
/* 28127 */ {(12<<2)|2,{53,55,0}},
/* 28128 */ {(12<<2)|2,{53,56,0}},
/* 28129 */ {(12<<2)|2,{53,56,0}},
/* 28130 */ {(12<<2)|2,{53,56,0}},
/* 28131 */ {(12<<2)|2,{53,56,0}},
/* 28132 */ {(12<<2)|2,{53,56,0}},
/* 28133 */ {(12<<2)|2,{53,56,0}},
/* 28134 */ {(12<<2)|2,{53,56,0}},
/* 28135 */ {(12<<2)|2,{53,56,0}},
/* 28136 */ {(12<<2)|2,{53,56,0}},
/* 28137 */ {(12<<2)|2,{53,56,0}},
/* 28138 */ {(12<<2)|2,{53,56,0}},
/* 28139 */ {(12<<2)|2,{53,56,0}},
/* 28140 */ {(12<<2)|2,{53,56,0}},
/* 28141 */ {(12<<2)|2,{53,56,0}},
/* 28142 */ {(12<<2)|2,{53,56,0}},
/* 28143 */ {(12<<2)|2,{53,56,0}},
/* 28144 */ {(12<<2)|2,{53,57,0}},
/* 28145 */ {(12<<2)|2,{53,57,0}},
/* 28146 */ {(12<<2)|2,{53,57,0}},
/* 28147 */ {(12<<2)|2,{53,57,0}},
/* 28148 */ {(12<<2)|2,{53,57,0}},
/* 28149 */ {(12<<2)|2,{53,57,0}},
/* 28150 */ {(12<<2)|2,{53,57,0}},
/* 28151 */ {(12<<2)|2,{53,57,0}},
/* 28152 */ {(12<<2)|2,{53,57,0}},
/* 28153 */ {(12<<2)|2,{53,57,0}},
/* 28154 */ {(12<<2)|2,{53,57,0}},
/* 28155 */ {(12<<2)|2,{53,57,0}},
/* 28156 */ {(12<<2)|2,{53,57,0}},
/* 28157 */ {(12<<2)|2,{53,57,0}},
/* 28158 */ {(12<<2)|2,{53,57,0}},
/* 28159 */ {(12<<2)|2,{53,57,0}},
/* 28160 */ {(12<<2)|2,{53,61,0}},
/* 28161 */ {(12<<2)|2,{53,61,0}},
/* 28162 */ {(12<<2)|2,{53,61,0}},
/* 28163 */ {(12<<2)|2,{53,61,0}},
/* 28164 */ {(12<<2)|2,{53,61,0}},
/* 28165 */ {(12<<2)|2,{53,61,0}},
/* 28166 */ {(12<<2)|2,{53,61,0}},
/* 28167 */ {(12<<2)|2,{53,61,0}},
/* 28168 */ {(12<<2)|2,{53,61,0}},
/* 28169 */ {(12<<2)|2,{53,61,0}},
/* 28170 */ {(12<<2)|2,{53,61,0}},
/* 28171 */ {(12<<2)|2,{53,61,0}},
/* 28172 */ {(12<<2)|2,{53,61,0}},
/* 28173 */ {(12<<2)|2,{53,61,0}},
/* 28174 */ {(12<<2)|2,{53,61,0}},
/* 28175 */ {(12<<2)|2,{53,61,0}},
/* 28176 */ {(12<<2)|2,{53,65,0}},
/* 28177 */ {(12<<2)|2,{53,65,0}},
/* 28178 */ {(12<<2)|2,{53,65,0}},
/* 28179 */ {(12<<2)|2,{53,65,0}},
/* 28180 */ {(12<<2)|2,{53,65,0}},
/* 28181 */ {(12<<2)|2,{53,65,0}},
/* 28182 */ {(12<<2)|2,{53,65,0}},
/* 28183 */ {(12<<2)|2,{53,65,0}},
/* 28184 */ {(12<<2)|2,{53,65,0}},
/* 28185 */ {(12<<2)|2,{53,65,0}},
/* 28186 */ {(12<<2)|2,{53,65,0}},
/* 28187 */ {(12<<2)|2,{53,65,0}},
/* 28188 */ {(12<<2)|2,{53,65,0}},
/* 28189 */ {(12<<2)|2,{53,65,0}},
/* 28190 */ {(12<<2)|2,{53,65,0}},
/* 28191 */ {(12<<2)|2,{53,65,0}},
/* 28192 */ {(12<<2)|2,{53,95,0}},
/* 28193 */ {(12<<2)|2,{53,95,0}},
/* 28194 */ {(12<<2)|2,{53,95,0}},
/* 28195 */ {(12<<2)|2,{53,95,0}},
/* 28196 */ {(12<<2)|2,{53,95,0}},
/* 28197 */ {(12<<2)|2,{53,95,0}},
/* 28198 */ {(12<<2)|2,{53,95,0}},
/* 28199 */ {(12<<2)|2,{53,95,0}},
/* 28200 */ {(12<<2)|2,{53,95,0}},
/* 28201 */ {(12<<2)|2,{53,95,0}},
/* 28202 */ {(12<<2)|2,{53,95,0}},
/* 28203 */ {(12<<2)|2,{53,95,0}},
/* 28204 */ {(12<<2)|2,{53,95,0}},
/* 28205 */ {(12<<2)|2,{53,95,0}},
/* 28206 */ {(12<<2)|2,{53,95,0}},
/* 28207 */ {(12<<2)|2,{53,95,0}},
/* 28208 */ {(12<<2)|2,{53,98,0}},
/* 28209 */ {(12<<2)|2,{53,98,0}},
/* 28210 */ {(12<<2)|2,{53,98,0}},
/* 28211 */ {(12<<2)|2,{53,98,0}},
/* 28212 */ {(12<<2)|2,{53,98,0}},
/* 28213 */ {(12<<2)|2,{53,98,0}},
/* 28214 */ {(12<<2)|2,{53,98,0}},
/* 28215 */ {(12<<2)|2,{53,98,0}},
/* 28216 */ {(12<<2)|2,{53,98,0}},
/* 28217 */ {(12<<2)|2,{53,98,0}},
/* 28218 */ {(12<<2)|2,{53,98,0}},
/* 28219 */ {(12<<2)|2,{53,98,0}},
/* 28220 */ {(12<<2)|2,{53,98,0}},
/* 28221 */ {(12<<2)|2,{53,98,0}},
/* 28222 */ {(12<<2)|2,{53,98,0}},
/* 28223 */ {(12<<2)|2,{53,98,0}},
/* 28224 */ {(12<<2)|2,{53,100,0}},
/* 28225 */ {(12<<2)|2,{53,100,0}},
/* 28226 */ {(12<<2)|2,{53,100,0}},
/* 28227 */ {(12<<2)|2,{53,100,0}},
/* 28228 */ {(12<<2)|2,{53,100,0}},
/* 28229 */ {(12<<2)|2,{53,100,0}},
/* 28230 */ {(12<<2)|2,{53,100,0}},
/* 28231 */ {(12<<2)|2,{53,100,0}},
/* 28232 */ {(12<<2)|2,{53,100,0}},
/* 28233 */ {(12<<2)|2,{53,100,0}},
/* 28234 */ {(12<<2)|2,{53,100,0}},
/* 28235 */ {(12<<2)|2,{53,100,0}},
/* 28236 */ {(12<<2)|2,{53,100,0}},
/* 28237 */ {(12<<2)|2,{53,100,0}},
/* 28238 */ {(12<<2)|2,{53,100,0}},
/* 28239 */ {(12<<2)|2,{53,100,0}},
/* 28240 */ {(12<<2)|2,{53,102,0}},
/* 28241 */ {(12<<2)|2,{53,102,0}},
/* 28242 */ {(12<<2)|2,{53,102,0}},
/* 28243 */ {(12<<2)|2,{53,102,0}},
/* 28244 */ {(12<<2)|2,{53,102,0}},
/* 28245 */ {(12<<2)|2,{53,102,0}},
/* 28246 */ {(12<<2)|2,{53,102,0}},
/* 28247 */ {(12<<2)|2,{53,102,0}},
/* 28248 */ {(12<<2)|2,{53,102,0}},
/* 28249 */ {(12<<2)|2,{53,102,0}},
/* 28250 */ {(12<<2)|2,{53,102,0}},
/* 28251 */ {(12<<2)|2,{53,102,0}},
/* 28252 */ {(12<<2)|2,{53,102,0}},
/* 28253 */ {(12<<2)|2,{53,102,0}},
/* 28254 */ {(12<<2)|2,{53,102,0}},
/* 28255 */ {(12<<2)|2,{53,102,0}},
/* 28256 */ {(12<<2)|2,{53,103,0}},
/* 28257 */ {(12<<2)|2,{53,103,0}},
/* 28258 */ {(12<<2)|2,{53,103,0}},
/* 28259 */ {(12<<2)|2,{53,103,0}},
/* 28260 */ {(12<<2)|2,{53,103,0}},
/* 28261 */ {(12<<2)|2,{53,103,0}},
/* 28262 */ {(12<<2)|2,{53,103,0}},
/* 28263 */ {(12<<2)|2,{53,103,0}},
/* 28264 */ {(12<<2)|2,{53,103,0}},
/* 28265 */ {(12<<2)|2,{53,103,0}},
/* 28266 */ {(12<<2)|2,{53,103,0}},
/* 28267 */ {(12<<2)|2,{53,103,0}},
/* 28268 */ {(12<<2)|2,{53,103,0}},
/* 28269 */ {(12<<2)|2,{53,103,0}},
/* 28270 */ {(12<<2)|2,{53,103,0}},
/* 28271 */ {(12<<2)|2,{53,103,0}},
/* 28272 */ {(12<<2)|2,{53,104,0}},
/* 28273 */ {(12<<2)|2,{53,104,0}},
/* 28274 */ {(12<<2)|2,{53,104,0}},
/* 28275 */ {(12<<2)|2,{53,104,0}},
/* 28276 */ {(12<<2)|2,{53,104,0}},
/* 28277 */ {(12<<2)|2,{53,104,0}},
/* 28278 */ {(12<<2)|2,{53,104,0}},
/* 28279 */ {(12<<2)|2,{53,104,0}},
/* 28280 */ {(12<<2)|2,{53,104,0}},
/* 28281 */ {(12<<2)|2,{53,104,0}},
/* 28282 */ {(12<<2)|2,{53,104,0}},
/* 28283 */ {(12<<2)|2,{53,104,0}},
/* 28284 */ {(12<<2)|2,{53,104,0}},
/* 28285 */ {(12<<2)|2,{53,104,0}},
/* 28286 */ {(12<<2)|2,{53,104,0}},
/* 28287 */ {(12<<2)|2,{53,104,0}},
/* 28288 */ {(12<<2)|2,{53,108,0}},
/* 28289 */ {(12<<2)|2,{53,108,0}},
/* 28290 */ {(12<<2)|2,{53,108,0}},
/* 28291 */ {(12<<2)|2,{53,108,0}},
/* 28292 */ {(12<<2)|2,{53,108,0}},
/* 28293 */ {(12<<2)|2,{53,108,0}},
/* 28294 */ {(12<<2)|2,{53,108,0}},
/* 28295 */ {(12<<2)|2,{53,108,0}},
/* 28296 */ {(12<<2)|2,{53,108,0}},
/* 28297 */ {(12<<2)|2,{53,108,0}},
/* 28298 */ {(12<<2)|2,{53,108,0}},
/* 28299 */ {(12<<2)|2,{53,108,0}},
/* 28300 */ {(12<<2)|2,{53,108,0}},
/* 28301 */ {(12<<2)|2,{53,108,0}},
/* 28302 */ {(12<<2)|2,{53,108,0}},
/* 28303 */ {(12<<2)|2,{53,108,0}},
/* 28304 */ {(12<<2)|2,{53,109,0}},
/* 28305 */ {(12<<2)|2,{53,109,0}},
/* 28306 */ {(12<<2)|2,{53,109,0}},
/* 28307 */ {(12<<2)|2,{53,109,0}},
/* 28308 */ {(12<<2)|2,{53,109,0}},
/* 28309 */ {(12<<2)|2,{53,109,0}},
/* 28310 */ {(12<<2)|2,{53,109,0}},
/* 28311 */ {(12<<2)|2,{53,109,0}},
/* 28312 */ {(12<<2)|2,{53,109,0}},
/* 28313 */ {(12<<2)|2,{53,109,0}},
/* 28314 */ {(12<<2)|2,{53,109,0}},
/* 28315 */ {(12<<2)|2,{53,109,0}},
/* 28316 */ {(12<<2)|2,{53,109,0}},
/* 28317 */ {(12<<2)|2,{53,109,0}},
/* 28318 */ {(12<<2)|2,{53,109,0}},
/* 28319 */ {(12<<2)|2,{53,109,0}},
/* 28320 */ {(12<<2)|2,{53,110,0}},
/* 28321 */ {(12<<2)|2,{53,110,0}},
/* 28322 */ {(12<<2)|2,{53,110,0}},
/* 28323 */ {(12<<2)|2,{53,110,0}},
/* 28324 */ {(12<<2)|2,{53,110,0}},
/* 28325 */ {(12<<2)|2,{53,110,0}},
/* 28326 */ {(12<<2)|2,{53,110,0}},
/* 28327 */ {(12<<2)|2,{53,110,0}},
/* 28328 */ {(12<<2)|2,{53,110,0}},
/* 28329 */ {(12<<2)|2,{53,110,0}},
/* 28330 */ {(12<<2)|2,{53,110,0}},
/* 28331 */ {(12<<2)|2,{53,110,0}},
/* 28332 */ {(12<<2)|2,{53,110,0}},
/* 28333 */ {(12<<2)|2,{53,110,0}},
/* 28334 */ {(12<<2)|2,{53,110,0}},
/* 28335 */ {(12<<2)|2,{53,110,0}},
/* 28336 */ {(12<<2)|2,{53,112,0}},
/* 28337 */ {(12<<2)|2,{53,112,0}},
/* 28338 */ {(12<<2)|2,{53,112,0}},
/* 28339 */ {(12<<2)|2,{53,112,0}},
/* 28340 */ {(12<<2)|2,{53,112,0}},
/* 28341 */ {(12<<2)|2,{53,112,0}},
/* 28342 */ {(12<<2)|2,{53,112,0}},
/* 28343 */ {(12<<2)|2,{53,112,0}},
/* 28344 */ {(12<<2)|2,{53,112,0}},
/* 28345 */ {(12<<2)|2,{53,112,0}},
/* 28346 */ {(12<<2)|2,{53,112,0}},
/* 28347 */ {(12<<2)|2,{53,112,0}},
/* 28348 */ {(12<<2)|2,{53,112,0}},
/* 28349 */ {(12<<2)|2,{53,112,0}},
/* 28350 */ {(12<<2)|2,{53,112,0}},
/* 28351 */ {(12<<2)|2,{53,112,0}},
/* 28352 */ {(12<<2)|2,{53,114,0}},
/* 28353 */ {(12<<2)|2,{53,114,0}},
/* 28354 */ {(12<<2)|2,{53,114,0}},
/* 28355 */ {(12<<2)|2,{53,114,0}},
/* 28356 */ {(12<<2)|2,{53,114,0}},
/* 28357 */ {(12<<2)|2,{53,114,0}},
/* 28358 */ {(12<<2)|2,{53,114,0}},
/* 28359 */ {(12<<2)|2,{53,114,0}},
/* 28360 */ {(12<<2)|2,{53,114,0}},
/* 28361 */ {(12<<2)|2,{53,114,0}},
/* 28362 */ {(12<<2)|2,{53,114,0}},
/* 28363 */ {(12<<2)|2,{53,114,0}},
/* 28364 */ {(12<<2)|2,{53,114,0}},
/* 28365 */ {(12<<2)|2,{53,114,0}},
/* 28366 */ {(12<<2)|2,{53,114,0}},
/* 28367 */ {(12<<2)|2,{53,114,0}},
/* 28368 */ {(12<<2)|2,{53,117,0}},
/* 28369 */ {(12<<2)|2,{53,117,0}},
/* 28370 */ {(12<<2)|2,{53,117,0}},
/* 28371 */ {(12<<2)|2,{53,117,0}},
/* 28372 */ {(12<<2)|2,{53,117,0}},
/* 28373 */ {(12<<2)|2,{53,117,0}},
/* 28374 */ {(12<<2)|2,{53,117,0}},
/* 28375 */ {(12<<2)|2,{53,117,0}},
/* 28376 */ {(12<<2)|2,{53,117,0}},
/* 28377 */ {(12<<2)|2,{53,117,0}},
/* 28378 */ {(12<<2)|2,{53,117,0}},
/* 28379 */ {(12<<2)|2,{53,117,0}},
/* 28380 */ {(12<<2)|2,{53,117,0}},
/* 28381 */ {(12<<2)|2,{53,117,0}},
/* 28382 */ {(12<<2)|2,{53,117,0}},
/* 28383 */ {(12<<2)|2,{53,117,0}},
/* 28384 */ {(13<<2)|2,{53,58,0}},
/* 28385 */ {(13<<2)|2,{53,58,0}},
/* 28386 */ {(13<<2)|2,{53,58,0}},
/* 28387 */ {(13<<2)|2,{53,58,0}},
/* 28388 */ {(13<<2)|2,{53,58,0}},
/* 28389 */ {(13<<2)|2,{53,58,0}},
/* 28390 */ {(13<<2)|2,{53,58,0}},
/* 28391 */ {(13<<2)|2,{53,58,0}},
/* 28392 */ {(13<<2)|2,{53,66,0}},
/* 28393 */ {(13<<2)|2,{53,66,0}},
/* 28394 */ {(13<<2)|2,{53,66,0}},
/* 28395 */ {(13<<2)|2,{53,66,0}},
/* 28396 */ {(13<<2)|2,{53,66,0}},
/* 28397 */ {(13<<2)|2,{53,66,0}},
/* 28398 */ {(13<<2)|2,{53,66,0}},
/* 28399 */ {(13<<2)|2,{53,66,0}},
/* 28400 */ {(13<<2)|2,{53,67,0}},
/* 28401 */ {(13<<2)|2,{53,67,0}},
/* 28402 */ {(13<<2)|2,{53,67,0}},
/* 28403 */ {(13<<2)|2,{53,67,0}},
/* 28404 */ {(13<<2)|2,{53,67,0}},
/* 28405 */ {(13<<2)|2,{53,67,0}},
/* 28406 */ {(13<<2)|2,{53,67,0}},
/* 28407 */ {(13<<2)|2,{53,67,0}},
/* 28408 */ {(13<<2)|2,{53,68,0}},
/* 28409 */ {(13<<2)|2,{53,68,0}},
/* 28410 */ {(13<<2)|2,{53,68,0}},
/* 28411 */ {(13<<2)|2,{53,68,0}},
/* 28412 */ {(13<<2)|2,{53,68,0}},
/* 28413 */ {(13<<2)|2,{53,68,0}},
/* 28414 */ {(13<<2)|2,{53,68,0}},
/* 28415 */ {(13<<2)|2,{53,68,0}},
/* 28416 */ {(13<<2)|2,{53,69,0}},
/* 28417 */ {(13<<2)|2,{53,69,0}},
/* 28418 */ {(13<<2)|2,{53,69,0}},
/* 28419 */ {(13<<2)|2,{53,69,0}},
/* 28420 */ {(13<<2)|2,{53,69,0}},
/* 28421 */ {(13<<2)|2,{53,69,0}},
/* 28422 */ {(13<<2)|2,{53,69,0}},
/* 28423 */ {(13<<2)|2,{53,69,0}},
/* 28424 */ {(13<<2)|2,{53,70,0}},
/* 28425 */ {(13<<2)|2,{53,70,0}},
/* 28426 */ {(13<<2)|2,{53,70,0}},
/* 28427 */ {(13<<2)|2,{53,70,0}},
/* 28428 */ {(13<<2)|2,{53,70,0}},
/* 28429 */ {(13<<2)|2,{53,70,0}},
/* 28430 */ {(13<<2)|2,{53,70,0}},
/* 28431 */ {(13<<2)|2,{53,70,0}},
/* 28432 */ {(13<<2)|2,{53,71,0}},
/* 28433 */ {(13<<2)|2,{53,71,0}},
/* 28434 */ {(13<<2)|2,{53,71,0}},
/* 28435 */ {(13<<2)|2,{53,71,0}},
/* 28436 */ {(13<<2)|2,{53,71,0}},
/* 28437 */ {(13<<2)|2,{53,71,0}},
/* 28438 */ {(13<<2)|2,{53,71,0}},
/* 28439 */ {(13<<2)|2,{53,71,0}},
/* 28440 */ {(13<<2)|2,{53,72,0}},
/* 28441 */ {(13<<2)|2,{53,72,0}},
/* 28442 */ {(13<<2)|2,{53,72,0}},
/* 28443 */ {(13<<2)|2,{53,72,0}},
/* 28444 */ {(13<<2)|2,{53,72,0}},
/* 28445 */ {(13<<2)|2,{53,72,0}},
/* 28446 */ {(13<<2)|2,{53,72,0}},
/* 28447 */ {(13<<2)|2,{53,72,0}},
/* 28448 */ {(13<<2)|2,{53,73,0}},
/* 28449 */ {(13<<2)|2,{53,73,0}},
/* 28450 */ {(13<<2)|2,{53,73,0}},
/* 28451 */ {(13<<2)|2,{53,73,0}},
/* 28452 */ {(13<<2)|2,{53,73,0}},
/* 28453 */ {(13<<2)|2,{53,73,0}},
/* 28454 */ {(13<<2)|2,{53,73,0}},
/* 28455 */ {(13<<2)|2,{53,73,0}},
/* 28456 */ {(13<<2)|2,{53,74,0}},
/* 28457 */ {(13<<2)|2,{53,74,0}},
/* 28458 */ {(13<<2)|2,{53,74,0}},
/* 28459 */ {(13<<2)|2,{53,74,0}},
/* 28460 */ {(13<<2)|2,{53,74,0}},
/* 28461 */ {(13<<2)|2,{53,74,0}},
/* 28462 */ {(13<<2)|2,{53,74,0}},
/* 28463 */ {(13<<2)|2,{53,74,0}},
/* 28464 */ {(13<<2)|2,{53,75,0}},
/* 28465 */ {(13<<2)|2,{53,75,0}},
/* 28466 */ {(13<<2)|2,{53,75,0}},
/* 28467 */ {(13<<2)|2,{53,75,0}},
/* 28468 */ {(13<<2)|2,{53,75,0}},
/* 28469 */ {(13<<2)|2,{53,75,0}},
/* 28470 */ {(13<<2)|2,{53,75,0}},
/* 28471 */ {(13<<2)|2,{53,75,0}},
/* 28472 */ {(13<<2)|2,{53,76,0}},
/* 28473 */ {(13<<2)|2,{53,76,0}},
/* 28474 */ {(13<<2)|2,{53,76,0}},
/* 28475 */ {(13<<2)|2,{53,76,0}},
/* 28476 */ {(13<<2)|2,{53,76,0}},
/* 28477 */ {(13<<2)|2,{53,76,0}},
/* 28478 */ {(13<<2)|2,{53,76,0}},
/* 28479 */ {(13<<2)|2,{53,76,0}},
/* 28480 */ {(13<<2)|2,{53,77,0}},
/* 28481 */ {(13<<2)|2,{53,77,0}},
/* 28482 */ {(13<<2)|2,{53,77,0}},
/* 28483 */ {(13<<2)|2,{53,77,0}},
/* 28484 */ {(13<<2)|2,{53,77,0}},
/* 28485 */ {(13<<2)|2,{53,77,0}},
/* 28486 */ {(13<<2)|2,{53,77,0}},
/* 28487 */ {(13<<2)|2,{53,77,0}},
/* 28488 */ {(13<<2)|2,{53,78,0}},
/* 28489 */ {(13<<2)|2,{53,78,0}},
/* 28490 */ {(13<<2)|2,{53,78,0}},
/* 28491 */ {(13<<2)|2,{53,78,0}},
/* 28492 */ {(13<<2)|2,{53,78,0}},
/* 28493 */ {(13<<2)|2,{53,78,0}},
/* 28494 */ {(13<<2)|2,{53,78,0}},
/* 28495 */ {(13<<2)|2,{53,78,0}},
/* 28496 */ {(13<<2)|2,{53,79,0}},
/* 28497 */ {(13<<2)|2,{53,79,0}},
/* 28498 */ {(13<<2)|2,{53,79,0}},
/* 28499 */ {(13<<2)|2,{53,79,0}},
/* 28500 */ {(13<<2)|2,{53,79,0}},
/* 28501 */ {(13<<2)|2,{53,79,0}},
/* 28502 */ {(13<<2)|2,{53,79,0}},
/* 28503 */ {(13<<2)|2,{53,79,0}},
/* 28504 */ {(13<<2)|2,{53,80,0}},
/* 28505 */ {(13<<2)|2,{53,80,0}},
/* 28506 */ {(13<<2)|2,{53,80,0}},
/* 28507 */ {(13<<2)|2,{53,80,0}},
/* 28508 */ {(13<<2)|2,{53,80,0}},
/* 28509 */ {(13<<2)|2,{53,80,0}},
/* 28510 */ {(13<<2)|2,{53,80,0}},
/* 28511 */ {(13<<2)|2,{53,80,0}},
/* 28512 */ {(13<<2)|2,{53,81,0}},
/* 28513 */ {(13<<2)|2,{53,81,0}},
/* 28514 */ {(13<<2)|2,{53,81,0}},
/* 28515 */ {(13<<2)|2,{53,81,0}},
/* 28516 */ {(13<<2)|2,{53,81,0}},
/* 28517 */ {(13<<2)|2,{53,81,0}},
/* 28518 */ {(13<<2)|2,{53,81,0}},
/* 28519 */ {(13<<2)|2,{53,81,0}},
/* 28520 */ {(13<<2)|2,{53,82,0}},
/* 28521 */ {(13<<2)|2,{53,82,0}},
/* 28522 */ {(13<<2)|2,{53,82,0}},
/* 28523 */ {(13<<2)|2,{53,82,0}},
/* 28524 */ {(13<<2)|2,{53,82,0}},
/* 28525 */ {(13<<2)|2,{53,82,0}},
/* 28526 */ {(13<<2)|2,{53,82,0}},
/* 28527 */ {(13<<2)|2,{53,82,0}},
/* 28528 */ {(13<<2)|2,{53,83,0}},
/* 28529 */ {(13<<2)|2,{53,83,0}},
/* 28530 */ {(13<<2)|2,{53,83,0}},
/* 28531 */ {(13<<2)|2,{53,83,0}},
/* 28532 */ {(13<<2)|2,{53,83,0}},
/* 28533 */ {(13<<2)|2,{53,83,0}},
/* 28534 */ {(13<<2)|2,{53,83,0}},
/* 28535 */ {(13<<2)|2,{53,83,0}},
/* 28536 */ {(13<<2)|2,{53,84,0}},
/* 28537 */ {(13<<2)|2,{53,84,0}},
/* 28538 */ {(13<<2)|2,{53,84,0}},
/* 28539 */ {(13<<2)|2,{53,84,0}},
/* 28540 */ {(13<<2)|2,{53,84,0}},
/* 28541 */ {(13<<2)|2,{53,84,0}},
/* 28542 */ {(13<<2)|2,{53,84,0}},
/* 28543 */ {(13<<2)|2,{53,84,0}},
/* 28544 */ {(13<<2)|2,{53,85,0}},
/* 28545 */ {(13<<2)|2,{53,85,0}},
/* 28546 */ {(13<<2)|2,{53,85,0}},
/* 28547 */ {(13<<2)|2,{53,85,0}},
/* 28548 */ {(13<<2)|2,{53,85,0}},
/* 28549 */ {(13<<2)|2,{53,85,0}},
/* 28550 */ {(13<<2)|2,{53,85,0}},
/* 28551 */ {(13<<2)|2,{53,85,0}},
/* 28552 */ {(13<<2)|2,{53,86,0}},
/* 28553 */ {(13<<2)|2,{53,86,0}},
/* 28554 */ {(13<<2)|2,{53,86,0}},
/* 28555 */ {(13<<2)|2,{53,86,0}},
/* 28556 */ {(13<<2)|2,{53,86,0}},
/* 28557 */ {(13<<2)|2,{53,86,0}},
/* 28558 */ {(13<<2)|2,{53,86,0}},
/* 28559 */ {(13<<2)|2,{53,86,0}},
/* 28560 */ {(13<<2)|2,{53,87,0}},
/* 28561 */ {(13<<2)|2,{53,87,0}},
/* 28562 */ {(13<<2)|2,{53,87,0}},
/* 28563 */ {(13<<2)|2,{53,87,0}},
/* 28564 */ {(13<<2)|2,{53,87,0}},
/* 28565 */ {(13<<2)|2,{53,87,0}},
/* 28566 */ {(13<<2)|2,{53,87,0}},
/* 28567 */ {(13<<2)|2,{53,87,0}},
/* 28568 */ {(13<<2)|2,{53,89,0}},
/* 28569 */ {(13<<2)|2,{53,89,0}},
/* 28570 */ {(13<<2)|2,{53,89,0}},
/* 28571 */ {(13<<2)|2,{53,89,0}},
/* 28572 */ {(13<<2)|2,{53,89,0}},
/* 28573 */ {(13<<2)|2,{53,89,0}},
/* 28574 */ {(13<<2)|2,{53,89,0}},
/* 28575 */ {(13<<2)|2,{53,89,0}},
/* 28576 */ {(13<<2)|2,{53,106,0}},
/* 28577 */ {(13<<2)|2,{53,106,0}},
/* 28578 */ {(13<<2)|2,{53,106,0}},
/* 28579 */ {(13<<2)|2,{53,106,0}},
/* 28580 */ {(13<<2)|2,{53,106,0}},
/* 28581 */ {(13<<2)|2,{53,106,0}},
/* 28582 */ {(13<<2)|2,{53,106,0}},
/* 28583 */ {(13<<2)|2,{53,106,0}},
/* 28584 */ {(13<<2)|2,{53,107,0}},
/* 28585 */ {(13<<2)|2,{53,107,0}},
/* 28586 */ {(13<<2)|2,{53,107,0}},
/* 28587 */ {(13<<2)|2,{53,107,0}},
/* 28588 */ {(13<<2)|2,{53,107,0}},
/* 28589 */ {(13<<2)|2,{53,107,0}},
/* 28590 */ {(13<<2)|2,{53,107,0}},
/* 28591 */ {(13<<2)|2,{53,107,0}},
/* 28592 */ {(13<<2)|2,{53,113,0}},
/* 28593 */ {(13<<2)|2,{53,113,0}},
/* 28594 */ {(13<<2)|2,{53,113,0}},
/* 28595 */ {(13<<2)|2,{53,113,0}},
/* 28596 */ {(13<<2)|2,{53,113,0}},
/* 28597 */ {(13<<2)|2,{53,113,0}},
/* 28598 */ {(13<<2)|2,{53,113,0}},
/* 28599 */ {(13<<2)|2,{53,113,0}},
/* 28600 */ {(13<<2)|2,{53,118,0}},
/* 28601 */ {(13<<2)|2,{53,118,0}},
/* 28602 */ {(13<<2)|2,{53,118,0}},
/* 28603 */ {(13<<2)|2,{53,118,0}},
/* 28604 */ {(13<<2)|2,{53,118,0}},
/* 28605 */ {(13<<2)|2,{53,118,0}},
/* 28606 */ {(13<<2)|2,{53,118,0}},
/* 28607 */ {(13<<2)|2,{53,118,0}},
/* 28608 */ {(13<<2)|2,{53,119,0}},
/* 28609 */ {(13<<2)|2,{53,119,0}},
/* 28610 */ {(13<<2)|2,{53,119,0}},
/* 28611 */ {(13<<2)|2,{53,119,0}},
/* 28612 */ {(13<<2)|2,{53,119,0}},
/* 28613 */ {(13<<2)|2,{53,119,0}},
/* 28614 */ {(13<<2)|2,{53,119,0}},
/* 28615 */ {(13<<2)|2,{53,119,0}},
/* 28616 */ {(13<<2)|2,{53,120,0}},
/* 28617 */ {(13<<2)|2,{53,120,0}},
/* 28618 */ {(13<<2)|2,{53,120,0}},
/* 28619 */ {(13<<2)|2,{53,120,0}},
/* 28620 */ {(13<<2)|2,{53,120,0}},
/* 28621 */ {(13<<2)|2,{53,120,0}},
/* 28622 */ {(13<<2)|2,{53,120,0}},
/* 28623 */ {(13<<2)|2,{53,120,0}},
/* 28624 */ {(13<<2)|2,{53,121,0}},
/* 28625 */ {(13<<2)|2,{53,121,0}},
/* 28626 */ {(13<<2)|2,{53,121,0}},
/* 28627 */ {(13<<2)|2,{53,121,0}},
/* 28628 */ {(13<<2)|2,{53,121,0}},
/* 28629 */ {(13<<2)|2,{53,121,0}},
/* 28630 */ {(13<<2)|2,{53,121,0}},
/* 28631 */ {(13<<2)|2,{53,121,0}},
/* 28632 */ {(13<<2)|2,{53,122,0}},
/* 28633 */ {(13<<2)|2,{53,122,0}},
/* 28634 */ {(13<<2)|2,{53,122,0}},
/* 28635 */ {(13<<2)|2,{53,122,0}},
/* 28636 */ {(13<<2)|2,{53,122,0}},
/* 28637 */ {(13<<2)|2,{53,122,0}},
/* 28638 */ {(13<<2)|2,{53,122,0}},
/* 28639 */ {(13<<2)|2,{53,122,0}},
/* 28640 */ {(14<<2)|2,{53,38,0}},
/* 28641 */ {(14<<2)|2,{53,38,0}},
/* 28642 */ {(14<<2)|2,{53,38,0}},
/* 28643 */ {(14<<2)|2,{53,38,0}},
/* 28644 */ {(14<<2)|2,{53,42,0}},
/* 28645 */ {(14<<2)|2,{53,42,0}},
/* 28646 */ {(14<<2)|2,{53,42,0}},
/* 28647 */ {(14<<2)|2,{53,42,0}},
/* 28648 */ {(14<<2)|2,{53,44,0}},
/* 28649 */ {(14<<2)|2,{53,44,0}},
/* 28650 */ {(14<<2)|2,{53,44,0}},
/* 28651 */ {(14<<2)|2,{53,44,0}},
/* 28652 */ {(14<<2)|2,{53,59,0}},
/* 28653 */ {(14<<2)|2,{53,59,0}},
/* 28654 */ {(14<<2)|2,{53,59,0}},
/* 28655 */ {(14<<2)|2,{53,59,0}},
/* 28656 */ {(14<<2)|2,{53,88,0}},
/* 28657 */ {(14<<2)|2,{53,88,0}},
/* 28658 */ {(14<<2)|2,{53,88,0}},
/* 28659 */ {(14<<2)|2,{53,88,0}},
/* 28660 */ {(14<<2)|2,{53,90,0}},
/* 28661 */ {(14<<2)|2,{53,90,0}},
/* 28662 */ {(14<<2)|2,{53,90,0}},
/* 28663 */ {(14<<2)|2,{53,90,0}},
/* 28664 */ {(16<<2)|2,{53,33,0}},
/* 28665 */ {(16<<2)|2,{53,34,0}},
/* 28666 */ {(16<<2)|2,{53,40,0}},
/* 28667 */ {(16<<2)|2,{53,41,0}},
/* 28668 */ {(16<<2)|2,{53,63,0}},
/* 28669 */ {(6<<2)|1,{53,0,0}},
/* 28670 */ {(6<<2)|1,{53,0,0}},
/* 28671 */ {(6<<2)|1,{53,0,0}},
/* 28672 */ {(16<<2)|3,{54,48,48}},
/* 28673 */ {(16<<2)|3,{54,48,49}},
/* 28674 */ {(16<<2)|3,{54,48,50}},
/* 28675 */ {(16<<2)|3,{54,48,97}},
/* 28676 */ {(16<<2)|3,{54,48,99}},
/* 28677 */ {(16<<2)|3,{54,48,101}},
/* 28678 */ {(16<<2)|3,{54,48,105}},
/* 28679 */ {(16<<2)|3,{54,48,111}},
/* 28680 */ {(16<<2)|3,{54,48,115}},
/* 28681 */ {(16<<2)|3,{54,48,116}},
/* 28682 */ {(11<<2)|2,{54,48,0}},
/* 28683 */ {(11<<2)|2,{54,48,0}},
/* 28684 */ {(11<<2)|2,{54,48,0}},
/* 28685 */ {(11<<2)|2,{54,48,0}},
/* 28686 */ {(11<<2)|2,{54,48,0}},
/* 28687 */ {(11<<2)|2,{54,48,0}},
/* 28688 */ {(11<<2)|2,{54,48,0}},
/* 28689 */ {(11<<2)|2,{54,48,0}},
/* 28690 */ {(11<<2)|2,{54,48,0}},
/* 28691 */ {(11<<2)|2,{54,48,0}},
/* 28692 */ {(11<<2)|2,{54,48,0}},
/* 28693 */ {(11<<2)|2,{54,48,0}},
/* 28694 */ {(11<<2)|2,{54,48,0}},
/* 28695 */ {(11<<2)|2,{54,48,0}},
/* 28696 */ {(11<<2)|2,{54,48,0}},
/* 28697 */ {(11<<2)|2,{54,48,0}},
/* 28698 */ {(11<<2)|2,{54,48,0}},
/* 28699 */ {(11<<2)|2,{54,48,0}},
/* 28700 */ {(11<<2)|2,{54,48,0}},
/* 28701 */ {(11<<2)|2,{54,48,0}},
/* 28702 */ {(11<<2)|2,{54,48,0}},
/* 28703 */ {(11<<2)|2,{54,48,0}},
/* 28704 */ {(16<<2)|3,{54,49,48}},
/* 28705 */ {(16<<2)|3,{54,49,49}},
/* 28706 */ {(16<<2)|3,{54,49,50}},
/* 28707 */ {(16<<2)|3,{54,49,97}},
/* 28708 */ {(16<<2)|3,{54,49,99}},
/* 28709 */ {(16<<2)|3,{54,49,101}},
/* 28710 */ {(16<<2)|3,{54,49,105}},
/* 28711 */ {(16<<2)|3,{54,49,111}},
/* 28712 */ {(16<<2)|3,{54,49,115}},
/* 28713 */ {(16<<2)|3,{54,49,116}},
/* 28714 */ {(11<<2)|2,{54,49,0}},
/* 28715 */ {(11<<2)|2,{54,49,0}},
/* 28716 */ {(11<<2)|2,{54,49,0}},
/* 28717 */ {(11<<2)|2,{54,49,0}},
/* 28718 */ {(11<<2)|2,{54,49,0}},
/* 28719 */ {(11<<2)|2,{54,49,0}},
/* 28720 */ {(11<<2)|2,{54,49,0}},
/* 28721 */ {(11<<2)|2,{54,49,0}},
/* 28722 */ {(11<<2)|2,{54,49,0}},
/* 28723 */ {(11<<2)|2,{54,49,0}},
/* 28724 */ {(11<<2)|2,{54,49,0}},
/* 28725 */ {(11<<2)|2,{54,49,0}},
/* 28726 */ {(11<<2)|2,{54,49,0}},
/* 28727 */ {(11<<2)|2,{54,49,0}},
/* 28728 */ {(11<<2)|2,{54,49,0}},
/* 28729 */ {(11<<2)|2,{54,49,0}},
/* 28730 */ {(11<<2)|2,{54,49,0}},
/* 28731 */ {(11<<2)|2,{54,49,0}},
/* 28732 */ {(11<<2)|2,{54,49,0}},
/* 28733 */ {(11<<2)|2,{54,49,0}},
/* 28734 */ {(11<<2)|2,{54,49,0}},
/* 28735 */ {(11<<2)|2,{54,49,0}},
/* 28736 */ {(16<<2)|3,{54,50,48}},
/* 28737 */ {(16<<2)|3,{54,50,49}},
/* 28738 */ {(16<<2)|3,{54,50,50}},
/* 28739 */ {(16<<2)|3,{54,50,97}},
/* 28740 */ {(16<<2)|3,{54,50,99}},
/* 28741 */ {(16<<2)|3,{54,50,101}},
/* 28742 */ {(16<<2)|3,{54,50,105}},
/* 28743 */ {(16<<2)|3,{54,50,111}},
/* 28744 */ {(16<<2)|3,{54,50,115}},
/* 28745 */ {(16<<2)|3,{54,50,116}},
/* 28746 */ {(11<<2)|2,{54,50,0}},
/* 28747 */ {(11<<2)|2,{54,50,0}},
/* 28748 */ {(11<<2)|2,{54,50,0}},
/* 28749 */ {(11<<2)|2,{54,50,0}},
/* 28750 */ {(11<<2)|2,{54,50,0}},
/* 28751 */ {(11<<2)|2,{54,50,0}},
/* 28752 */ {(11<<2)|2,{54,50,0}},
/* 28753 */ {(11<<2)|2,{54,50,0}},
/* 28754 */ {(11<<2)|2,{54,50,0}},
/* 28755 */ {(11<<2)|2,{54,50,0}},
/* 28756 */ {(11<<2)|2,{54,50,0}},
/* 28757 */ {(11<<2)|2,{54,50,0}},
/* 28758 */ {(11<<2)|2,{54,50,0}},
/* 28759 */ {(11<<2)|2,{54,50,0}},
/* 28760 */ {(11<<2)|2,{54,50,0}},
/* 28761 */ {(11<<2)|2,{54,50,0}},
/* 28762 */ {(11<<2)|2,{54,50,0}},
/* 28763 */ {(11<<2)|2,{54,50,0}},
/* 28764 */ {(11<<2)|2,{54,50,0}},
/* 28765 */ {(11<<2)|2,{54,50,0}},
/* 28766 */ {(11<<2)|2,{54,50,0}},
/* 28767 */ {(11<<2)|2,{54,50,0}},
/* 28768 */ {(16<<2)|3,{54,97,48}},
/* 28769 */ {(16<<2)|3,{54,97,49}},
/* 28770 */ {(16<<2)|3,{54,97,50}},
/* 28771 */ {(16<<2)|3,{54,97,97}},
/* 28772 */ {(16<<2)|3,{54,97,99}},
/* 28773 */ {(16<<2)|3,{54,97,101}},
/* 28774 */ {(16<<2)|3,{54,97,105}},
/* 28775 */ {(16<<2)|3,{54,97,111}},
/* 28776 */ {(16<<2)|3,{54,97,115}},
/* 28777 */ {(16<<2)|3,{54,97,116}},
/* 28778 */ {(11<<2)|2,{54,97,0}},
/* 28779 */ {(11<<2)|2,{54,97,0}},
/* 28780 */ {(11<<2)|2,{54,97,0}},
/* 28781 */ {(11<<2)|2,{54,97,0}},
/* 28782 */ {(11<<2)|2,{54,97,0}},
/* 28783 */ {(11<<2)|2,{54,97,0}},
/* 28784 */ {(11<<2)|2,{54,97,0}},
/* 28785 */ {(11<<2)|2,{54,97,0}},
/* 28786 */ {(11<<2)|2,{54,97,0}},
/* 28787 */ {(11<<2)|2,{54,97,0}},
/* 28788 */ {(11<<2)|2,{54,97,0}},
/* 28789 */ {(11<<2)|2,{54,97,0}},
/* 28790 */ {(11<<2)|2,{54,97,0}},
/* 28791 */ {(11<<2)|2,{54,97,0}},
/* 28792 */ {(11<<2)|2,{54,97,0}},
/* 28793 */ {(11<<2)|2,{54,97,0}},
/* 28794 */ {(11<<2)|2,{54,97,0}},
/* 28795 */ {(11<<2)|2,{54,97,0}},
/* 28796 */ {(11<<2)|2,{54,97,0}},
/* 28797 */ {(11<<2)|2,{54,97,0}},
/* 28798 */ {(11<<2)|2,{54,97,0}},
/* 28799 */ {(11<<2)|2,{54,97,0}},
/* 28800 */ {(16<<2)|3,{54,99,48}},
/* 28801 */ {(16<<2)|3,{54,99,49}},
/* 28802 */ {(16<<2)|3,{54,99,50}},
/* 28803 */ {(16<<2)|3,{54,99,97}},
/* 28804 */ {(16<<2)|3,{54,99,99}},
/* 28805 */ {(16<<2)|3,{54,99,101}},
/* 28806 */ {(16<<2)|3,{54,99,105}},
/* 28807 */ {(16<<2)|3,{54,99,111}},
/* 28808 */ {(16<<2)|3,{54,99,115}},
/* 28809 */ {(16<<2)|3,{54,99,116}},
/* 28810 */ {(11<<2)|2,{54,99,0}},
/* 28811 */ {(11<<2)|2,{54,99,0}},
/* 28812 */ {(11<<2)|2,{54,99,0}},
/* 28813 */ {(11<<2)|2,{54,99,0}},
/* 28814 */ {(11<<2)|2,{54,99,0}},
/* 28815 */ {(11<<2)|2,{54,99,0}},
/* 28816 */ {(11<<2)|2,{54,99,0}},
/* 28817 */ {(11<<2)|2,{54,99,0}},
/* 28818 */ {(11<<2)|2,{54,99,0}},
/* 28819 */ {(11<<2)|2,{54,99,0}},
/* 28820 */ {(11<<2)|2,{54,99,0}},
/* 28821 */ {(11<<2)|2,{54,99,0}},
/* 28822 */ {(11<<2)|2,{54,99,0}},
/* 28823 */ {(11<<2)|2,{54,99,0}},
/* 28824 */ {(11<<2)|2,{54,99,0}},
/* 28825 */ {(11<<2)|2,{54,99,0}},
/* 28826 */ {(11<<2)|2,{54,99,0}},
/* 28827 */ {(11<<2)|2,{54,99,0}},
/* 28828 */ {(11<<2)|2,{54,99,0}},
/* 28829 */ {(11<<2)|2,{54,99,0}},
/* 28830 */ {(11<<2)|2,{54,99,0}},
/* 28831 */ {(11<<2)|2,{54,99,0}},
/* 28832 */ {(16<<2)|3,{54,101,48}},
/* 28833 */ {(16<<2)|3,{54,101,49}},
/* 28834 */ {(16<<2)|3,{54,101,50}},
/* 28835 */ {(16<<2)|3,{54,101,97}},
/* 28836 */ {(16<<2)|3,{54,101,99}},
/* 28837 */ {(16<<2)|3,{54,101,101}},
/* 28838 */ {(16<<2)|3,{54,101,105}},
/* 28839 */ {(16<<2)|3,{54,101,111}},
/* 28840 */ {(16<<2)|3,{54,101,115}},
/* 28841 */ {(16<<2)|3,{54,101,116}},
/* 28842 */ {(11<<2)|2,{54,101,0}},
/* 28843 */ {(11<<2)|2,{54,101,0}},
/* 28844 */ {(11<<2)|2,{54,101,0}},
/* 28845 */ {(11<<2)|2,{54,101,0}},
/* 28846 */ {(11<<2)|2,{54,101,0}},
/* 28847 */ {(11<<2)|2,{54,101,0}},
/* 28848 */ {(11<<2)|2,{54,101,0}},
/* 28849 */ {(11<<2)|2,{54,101,0}},
/* 28850 */ {(11<<2)|2,{54,101,0}},
/* 28851 */ {(11<<2)|2,{54,101,0}},
/* 28852 */ {(11<<2)|2,{54,101,0}},
/* 28853 */ {(11<<2)|2,{54,101,0}},
/* 28854 */ {(11<<2)|2,{54,101,0}},
/* 28855 */ {(11<<2)|2,{54,101,0}},
/* 28856 */ {(11<<2)|2,{54,101,0}},
/* 28857 */ {(11<<2)|2,{54,101,0}},
/* 28858 */ {(11<<2)|2,{54,101,0}},
/* 28859 */ {(11<<2)|2,{54,101,0}},
/* 28860 */ {(11<<2)|2,{54,101,0}},
/* 28861 */ {(11<<2)|2,{54,101,0}},
/* 28862 */ {(11<<2)|2,{54,101,0}},
/* 28863 */ {(11<<2)|2,{54,101,0}},
/* 28864 */ {(16<<2)|3,{54,105,48}},
/* 28865 */ {(16<<2)|3,{54,105,49}},
/* 28866 */ {(16<<2)|3,{54,105,50}},
/* 28867 */ {(16<<2)|3,{54,105,97}},
/* 28868 */ {(16<<2)|3,{54,105,99}},
/* 28869 */ {(16<<2)|3,{54,105,101}},
/* 28870 */ {(16<<2)|3,{54,105,105}},
/* 28871 */ {(16<<2)|3,{54,105,111}},
/* 28872 */ {(16<<2)|3,{54,105,115}},
/* 28873 */ {(16<<2)|3,{54,105,116}},
/* 28874 */ {(11<<2)|2,{54,105,0}},
/* 28875 */ {(11<<2)|2,{54,105,0}},
/* 28876 */ {(11<<2)|2,{54,105,0}},
/* 28877 */ {(11<<2)|2,{54,105,0}},
/* 28878 */ {(11<<2)|2,{54,105,0}},
/* 28879 */ {(11<<2)|2,{54,105,0}},
/* 28880 */ {(11<<2)|2,{54,105,0}},
/* 28881 */ {(11<<2)|2,{54,105,0}},
/* 28882 */ {(11<<2)|2,{54,105,0}},
/* 28883 */ {(11<<2)|2,{54,105,0}},
/* 28884 */ {(11<<2)|2,{54,105,0}},
/* 28885 */ {(11<<2)|2,{54,105,0}},
/* 28886 */ {(11<<2)|2,{54,105,0}},
/* 28887 */ {(11<<2)|2,{54,105,0}},
/* 28888 */ {(11<<2)|2,{54,105,0}},
/* 28889 */ {(11<<2)|2,{54,105,0}},
/* 28890 */ {(11<<2)|2,{54,105,0}},
/* 28891 */ {(11<<2)|2,{54,105,0}},
/* 28892 */ {(11<<2)|2,{54,105,0}},
/* 28893 */ {(11<<2)|2,{54,105,0}},
/* 28894 */ {(11<<2)|2,{54,105,0}},
/* 28895 */ {(11<<2)|2,{54,105,0}},
/* 28896 */ {(16<<2)|3,{54,111,48}},
/* 28897 */ {(16<<2)|3,{54,111,49}},
/* 28898 */ {(16<<2)|3,{54,111,50}},
/* 28899 */ {(16<<2)|3,{54,111,97}},
/* 28900 */ {(16<<2)|3,{54,111,99}},
/* 28901 */ {(16<<2)|3,{54,111,101}},
/* 28902 */ {(16<<2)|3,{54,111,105}},
/* 28903 */ {(16<<2)|3,{54,111,111}},
/* 28904 */ {(16<<2)|3,{54,111,115}},
/* 28905 */ {(16<<2)|3,{54,111,116}},
/* 28906 */ {(11<<2)|2,{54,111,0}},
/* 28907 */ {(11<<2)|2,{54,111,0}},
/* 28908 */ {(11<<2)|2,{54,111,0}},
/* 28909 */ {(11<<2)|2,{54,111,0}},
/* 28910 */ {(11<<2)|2,{54,111,0}},
/* 28911 */ {(11<<2)|2,{54,111,0}},
/* 28912 */ {(11<<2)|2,{54,111,0}},
/* 28913 */ {(11<<2)|2,{54,111,0}},
/* 28914 */ {(11<<2)|2,{54,111,0}},
/* 28915 */ {(11<<2)|2,{54,111,0}},
/* 28916 */ {(11<<2)|2,{54,111,0}},
/* 28917 */ {(11<<2)|2,{54,111,0}},
/* 28918 */ {(11<<2)|2,{54,111,0}},
/* 28919 */ {(11<<2)|2,{54,111,0}},
/* 28920 */ {(11<<2)|2,{54,111,0}},
/* 28921 */ {(11<<2)|2,{54,111,0}},
/* 28922 */ {(11<<2)|2,{54,111,0}},
/* 28923 */ {(11<<2)|2,{54,111,0}},
/* 28924 */ {(11<<2)|2,{54,111,0}},
/* 28925 */ {(11<<2)|2,{54,111,0}},
/* 28926 */ {(11<<2)|2,{54,111,0}},
/* 28927 */ {(11<<2)|2,{54,111,0}},
/* 28928 */ {(16<<2)|3,{54,115,48}},
/* 28929 */ {(16<<2)|3,{54,115,49}},
/* 28930 */ {(16<<2)|3,{54,115,50}},
/* 28931 */ {(16<<2)|3,{54,115,97}},
/* 28932 */ {(16<<2)|3,{54,115,99}},
/* 28933 */ {(16<<2)|3,{54,115,101}},
/* 28934 */ {(16<<2)|3,{54,115,105}},
/* 28935 */ {(16<<2)|3,{54,115,111}},
/* 28936 */ {(16<<2)|3,{54,115,115}},
/* 28937 */ {(16<<2)|3,{54,115,116}},
/* 28938 */ {(11<<2)|2,{54,115,0}},
/* 28939 */ {(11<<2)|2,{54,115,0}},
/* 28940 */ {(11<<2)|2,{54,115,0}},
/* 28941 */ {(11<<2)|2,{54,115,0}},
/* 28942 */ {(11<<2)|2,{54,115,0}},
/* 28943 */ {(11<<2)|2,{54,115,0}},
/* 28944 */ {(11<<2)|2,{54,115,0}},
/* 28945 */ {(11<<2)|2,{54,115,0}},
/* 28946 */ {(11<<2)|2,{54,115,0}},
/* 28947 */ {(11<<2)|2,{54,115,0}},
/* 28948 */ {(11<<2)|2,{54,115,0}},
/* 28949 */ {(11<<2)|2,{54,115,0}},
/* 28950 */ {(11<<2)|2,{54,115,0}},
/* 28951 */ {(11<<2)|2,{54,115,0}},
/* 28952 */ {(11<<2)|2,{54,115,0}},
/* 28953 */ {(11<<2)|2,{54,115,0}},
/* 28954 */ {(11<<2)|2,{54,115,0}},
/* 28955 */ {(11<<2)|2,{54,115,0}},
/* 28956 */ {(11<<2)|2,{54,115,0}},
/* 28957 */ {(11<<2)|2,{54,115,0}},
/* 28958 */ {(11<<2)|2,{54,115,0}},
/* 28959 */ {(11<<2)|2,{54,115,0}},
/* 28960 */ {(16<<2)|3,{54,116,48}},
/* 28961 */ {(16<<2)|3,{54,116,49}},
/* 28962 */ {(16<<2)|3,{54,116,50}},
/* 28963 */ {(16<<2)|3,{54,116,97}},
/* 28964 */ {(16<<2)|3,{54,116,99}},
/* 28965 */ {(16<<2)|3,{54,116,101}},
/* 28966 */ {(16<<2)|3,{54,116,105}},
/* 28967 */ {(16<<2)|3,{54,116,111}},
/* 28968 */ {(16<<2)|3,{54,116,115}},
/* 28969 */ {(16<<2)|3,{54,116,116}},
/* 28970 */ {(11<<2)|2,{54,116,0}},
/* 28971 */ {(11<<2)|2,{54,116,0}},
/* 28972 */ {(11<<2)|2,{54,116,0}},
/* 28973 */ {(11<<2)|2,{54,116,0}},
/* 28974 */ {(11<<2)|2,{54,116,0}},
/* 28975 */ {(11<<2)|2,{54,116,0}},
/* 28976 */ {(11<<2)|2,{54,116,0}},
/* 28977 */ {(11<<2)|2,{54,116,0}},
/* 28978 */ {(11<<2)|2,{54,116,0}},
/* 28979 */ {(11<<2)|2,{54,116,0}},
/* 28980 */ {(11<<2)|2,{54,116,0}},
/* 28981 */ {(11<<2)|2,{54,116,0}},
/* 28982 */ {(11<<2)|2,{54,116,0}},
/* 28983 */ {(11<<2)|2,{54,116,0}},
/* 28984 */ {(11<<2)|2,{54,116,0}},
/* 28985 */ {(11<<2)|2,{54,116,0}},
/* 28986 */ {(11<<2)|2,{54,116,0}},
/* 28987 */ {(11<<2)|2,{54,116,0}},
/* 28988 */ {(11<<2)|2,{54,116,0}},
/* 28989 */ {(11<<2)|2,{54,116,0}},
/* 28990 */ {(11<<2)|2,{54,116,0}},
/* 28991 */ {(11<<2)|2,{54,116,0}},
/* 28992 */ {(12<<2)|2,{54,32,0}},
/* 28993 */ {(12<<2)|2,{54,32,0}},
/* 28994 */ {(12<<2)|2,{54,32,0}},
/* 28995 */ {(12<<2)|2,{54,32,0}},
/* 28996 */ {(12<<2)|2,{54,32,0}},
/* 28997 */ {(12<<2)|2,{54,32,0}},
/* 28998 */ {(12<<2)|2,{54,32,0}},
/* 28999 */ {(12<<2)|2,{54,32,0}},
/* 29000 */ {(12<<2)|2,{54,32,0}},
/* 29001 */ {(12<<2)|2,{54,32,0}},
/* 29002 */ {(12<<2)|2,{54,32,0}},
/* 29003 */ {(12<<2)|2,{54,32,0}},
/* 29004 */ {(12<<2)|2,{54,32,0}},
/* 29005 */ {(12<<2)|2,{54,32,0}},
/* 29006 */ {(12<<2)|2,{54,32,0}},
/* 29007 */ {(12<<2)|2,{54,32,0}},
/* 29008 */ {(12<<2)|2,{54,37,0}},
/* 29009 */ {(12<<2)|2,{54,37,0}},
/* 29010 */ {(12<<2)|2,{54,37,0}},
/* 29011 */ {(12<<2)|2,{54,37,0}},
/* 29012 */ {(12<<2)|2,{54,37,0}},
/* 29013 */ {(12<<2)|2,{54,37,0}},
/* 29014 */ {(12<<2)|2,{54,37,0}},
/* 29015 */ {(12<<2)|2,{54,37,0}},
/* 29016 */ {(12<<2)|2,{54,37,0}},
/* 29017 */ {(12<<2)|2,{54,37,0}},
/* 29018 */ {(12<<2)|2,{54,37,0}},
/* 29019 */ {(12<<2)|2,{54,37,0}},
/* 29020 */ {(12<<2)|2,{54,37,0}},
/* 29021 */ {(12<<2)|2,{54,37,0}},
/* 29022 */ {(12<<2)|2,{54,37,0}},
/* 29023 */ {(12<<2)|2,{54,37,0}},
/* 29024 */ {(12<<2)|2,{54,45,0}},
/* 29025 */ {(12<<2)|2,{54,45,0}},
/* 29026 */ {(12<<2)|2,{54,45,0}},
/* 29027 */ {(12<<2)|2,{54,45,0}},
/* 29028 */ {(12<<2)|2,{54,45,0}},
/* 29029 */ {(12<<2)|2,{54,45,0}},
/* 29030 */ {(12<<2)|2,{54,45,0}},
/* 29031 */ {(12<<2)|2,{54,45,0}},
/* 29032 */ {(12<<2)|2,{54,45,0}},
/* 29033 */ {(12<<2)|2,{54,45,0}},
/* 29034 */ {(12<<2)|2,{54,45,0}},
/* 29035 */ {(12<<2)|2,{54,45,0}},
/* 29036 */ {(12<<2)|2,{54,45,0}},
/* 29037 */ {(12<<2)|2,{54,45,0}},
/* 29038 */ {(12<<2)|2,{54,45,0}},
/* 29039 */ {(12<<2)|2,{54,45,0}},
/* 29040 */ {(12<<2)|2,{54,46,0}},
/* 29041 */ {(12<<2)|2,{54,46,0}},
/* 29042 */ {(12<<2)|2,{54,46,0}},
/* 29043 */ {(12<<2)|2,{54,46,0}},
/* 29044 */ {(12<<2)|2,{54,46,0}},
/* 29045 */ {(12<<2)|2,{54,46,0}},
/* 29046 */ {(12<<2)|2,{54,46,0}},
/* 29047 */ {(12<<2)|2,{54,46,0}},
/* 29048 */ {(12<<2)|2,{54,46,0}},
/* 29049 */ {(12<<2)|2,{54,46,0}},
/* 29050 */ {(12<<2)|2,{54,46,0}},
/* 29051 */ {(12<<2)|2,{54,46,0}},
/* 29052 */ {(12<<2)|2,{54,46,0}},
/* 29053 */ {(12<<2)|2,{54,46,0}},
/* 29054 */ {(12<<2)|2,{54,46,0}},
/* 29055 */ {(12<<2)|2,{54,46,0}},
/* 29056 */ {(12<<2)|2,{54,47,0}},
/* 29057 */ {(12<<2)|2,{54,47,0}},
/* 29058 */ {(12<<2)|2,{54,47,0}},
/* 29059 */ {(12<<2)|2,{54,47,0}},
/* 29060 */ {(12<<2)|2,{54,47,0}},
/* 29061 */ {(12<<2)|2,{54,47,0}},
/* 29062 */ {(12<<2)|2,{54,47,0}},
/* 29063 */ {(12<<2)|2,{54,47,0}},
/* 29064 */ {(12<<2)|2,{54,47,0}},
/* 29065 */ {(12<<2)|2,{54,47,0}},
/* 29066 */ {(12<<2)|2,{54,47,0}},
/* 29067 */ {(12<<2)|2,{54,47,0}},
/* 29068 */ {(12<<2)|2,{54,47,0}},
/* 29069 */ {(12<<2)|2,{54,47,0}},
/* 29070 */ {(12<<2)|2,{54,47,0}},
/* 29071 */ {(12<<2)|2,{54,47,0}},
/* 29072 */ {(12<<2)|2,{54,51,0}},
/* 29073 */ {(12<<2)|2,{54,51,0}},
/* 29074 */ {(12<<2)|2,{54,51,0}},
/* 29075 */ {(12<<2)|2,{54,51,0}},
/* 29076 */ {(12<<2)|2,{54,51,0}},
/* 29077 */ {(12<<2)|2,{54,51,0}},
/* 29078 */ {(12<<2)|2,{54,51,0}},
/* 29079 */ {(12<<2)|2,{54,51,0}},
/* 29080 */ {(12<<2)|2,{54,51,0}},
/* 29081 */ {(12<<2)|2,{54,51,0}},
/* 29082 */ {(12<<2)|2,{54,51,0}},
/* 29083 */ {(12<<2)|2,{54,51,0}},
/* 29084 */ {(12<<2)|2,{54,51,0}},
/* 29085 */ {(12<<2)|2,{54,51,0}},
/* 29086 */ {(12<<2)|2,{54,51,0}},
/* 29087 */ {(12<<2)|2,{54,51,0}},
/* 29088 */ {(12<<2)|2,{54,52,0}},
/* 29089 */ {(12<<2)|2,{54,52,0}},
/* 29090 */ {(12<<2)|2,{54,52,0}},
/* 29091 */ {(12<<2)|2,{54,52,0}},
/* 29092 */ {(12<<2)|2,{54,52,0}},
/* 29093 */ {(12<<2)|2,{54,52,0}},
/* 29094 */ {(12<<2)|2,{54,52,0}},
/* 29095 */ {(12<<2)|2,{54,52,0}},
/* 29096 */ {(12<<2)|2,{54,52,0}},
/* 29097 */ {(12<<2)|2,{54,52,0}},
/* 29098 */ {(12<<2)|2,{54,52,0}},
/* 29099 */ {(12<<2)|2,{54,52,0}},
/* 29100 */ {(12<<2)|2,{54,52,0}},
/* 29101 */ {(12<<2)|2,{54,52,0}},
/* 29102 */ {(12<<2)|2,{54,52,0}},
/* 29103 */ {(12<<2)|2,{54,52,0}},
/* 29104 */ {(12<<2)|2,{54,53,0}},
/* 29105 */ {(12<<2)|2,{54,53,0}},
/* 29106 */ {(12<<2)|2,{54,53,0}},
/* 29107 */ {(12<<2)|2,{54,53,0}},
/* 29108 */ {(12<<2)|2,{54,53,0}},
/* 29109 */ {(12<<2)|2,{54,53,0}},
/* 29110 */ {(12<<2)|2,{54,53,0}},
/* 29111 */ {(12<<2)|2,{54,53,0}},
/* 29112 */ {(12<<2)|2,{54,53,0}},
/* 29113 */ {(12<<2)|2,{54,53,0}},
/* 29114 */ {(12<<2)|2,{54,53,0}},
/* 29115 */ {(12<<2)|2,{54,53,0}},
/* 29116 */ {(12<<2)|2,{54,53,0}},
/* 29117 */ {(12<<2)|2,{54,53,0}},
/* 29118 */ {(12<<2)|2,{54,53,0}},
/* 29119 */ {(12<<2)|2,{54,53,0}},
/* 29120 */ {(12<<2)|2,{54,54,0}},
/* 29121 */ {(12<<2)|2,{54,54,0}},
/* 29122 */ {(12<<2)|2,{54,54,0}},
/* 29123 */ {(12<<2)|2,{54,54,0}},
/* 29124 */ {(12<<2)|2,{54,54,0}},
/* 29125 */ {(12<<2)|2,{54,54,0}},
/* 29126 */ {(12<<2)|2,{54,54,0}},
/* 29127 */ {(12<<2)|2,{54,54,0}},
/* 29128 */ {(12<<2)|2,{54,54,0}},
/* 29129 */ {(12<<2)|2,{54,54,0}},
/* 29130 */ {(12<<2)|2,{54,54,0}},
/* 29131 */ {(12<<2)|2,{54,54,0}},
/* 29132 */ {(12<<2)|2,{54,54,0}},
/* 29133 */ {(12<<2)|2,{54,54,0}},
/* 29134 */ {(12<<2)|2,{54,54,0}},
/* 29135 */ {(12<<2)|2,{54,54,0}},
/* 29136 */ {(12<<2)|2,{54,55,0}},
/* 29137 */ {(12<<2)|2,{54,55,0}},
/* 29138 */ {(12<<2)|2,{54,55,0}},
/* 29139 */ {(12<<2)|2,{54,55,0}},
/* 29140 */ {(12<<2)|2,{54,55,0}},
/* 29141 */ {(12<<2)|2,{54,55,0}},
/* 29142 */ {(12<<2)|2,{54,55,0}},
/* 29143 */ {(12<<2)|2,{54,55,0}},
/* 29144 */ {(12<<2)|2,{54,55,0}},
/* 29145 */ {(12<<2)|2,{54,55,0}},
/* 29146 */ {(12<<2)|2,{54,55,0}},
/* 29147 */ {(12<<2)|2,{54,55,0}},
/* 29148 */ {(12<<2)|2,{54,55,0}},
/* 29149 */ {(12<<2)|2,{54,55,0}},
/* 29150 */ {(12<<2)|2,{54,55,0}},
/* 29151 */ {(12<<2)|2,{54,55,0}},
/* 29152 */ {(12<<2)|2,{54,56,0}},
/* 29153 */ {(12<<2)|2,{54,56,0}},
/* 29154 */ {(12<<2)|2,{54,56,0}},
/* 29155 */ {(12<<2)|2,{54,56,0}},
/* 29156 */ {(12<<2)|2,{54,56,0}},
/* 29157 */ {(12<<2)|2,{54,56,0}},
/* 29158 */ {(12<<2)|2,{54,56,0}},
/* 29159 */ {(12<<2)|2,{54,56,0}},
/* 29160 */ {(12<<2)|2,{54,56,0}},
/* 29161 */ {(12<<2)|2,{54,56,0}},
/* 29162 */ {(12<<2)|2,{54,56,0}},
/* 29163 */ {(12<<2)|2,{54,56,0}},
/* 29164 */ {(12<<2)|2,{54,56,0}},
/* 29165 */ {(12<<2)|2,{54,56,0}},
/* 29166 */ {(12<<2)|2,{54,56,0}},
/* 29167 */ {(12<<2)|2,{54,56,0}},
/* 29168 */ {(12<<2)|2,{54,57,0}},
/* 29169 */ {(12<<2)|2,{54,57,0}},
/* 29170 */ {(12<<2)|2,{54,57,0}},
/* 29171 */ {(12<<2)|2,{54,57,0}},
/* 29172 */ {(12<<2)|2,{54,57,0}},
/* 29173 */ {(12<<2)|2,{54,57,0}},
/* 29174 */ {(12<<2)|2,{54,57,0}},
/* 29175 */ {(12<<2)|2,{54,57,0}},
/* 29176 */ {(12<<2)|2,{54,57,0}},
/* 29177 */ {(12<<2)|2,{54,57,0}},
/* 29178 */ {(12<<2)|2,{54,57,0}},
/* 29179 */ {(12<<2)|2,{54,57,0}},
/* 29180 */ {(12<<2)|2,{54,57,0}},
/* 29181 */ {(12<<2)|2,{54,57,0}},
/* 29182 */ {(12<<2)|2,{54,57,0}},
/* 29183 */ {(12<<2)|2,{54,57,0}},
/* 29184 */ {(12<<2)|2,{54,61,0}},
/* 29185 */ {(12<<2)|2,{54,61,0}},
/* 29186 */ {(12<<2)|2,{54,61,0}},
/* 29187 */ {(12<<2)|2,{54,61,0}},
/* 29188 */ {(12<<2)|2,{54,61,0}},
/* 29189 */ {(12<<2)|2,{54,61,0}},
/* 29190 */ {(12<<2)|2,{54,61,0}},
/* 29191 */ {(12<<2)|2,{54,61,0}},
/* 29192 */ {(12<<2)|2,{54,61,0}},
/* 29193 */ {(12<<2)|2,{54,61,0}},
/* 29194 */ {(12<<2)|2,{54,61,0}},
/* 29195 */ {(12<<2)|2,{54,61,0}},
/* 29196 */ {(12<<2)|2,{54,61,0}},
/* 29197 */ {(12<<2)|2,{54,61,0}},
/* 29198 */ {(12<<2)|2,{54,61,0}},
/* 29199 */ {(12<<2)|2,{54,61,0}},
/* 29200 */ {(12<<2)|2,{54,65,0}},
/* 29201 */ {(12<<2)|2,{54,65,0}},
/* 29202 */ {(12<<2)|2,{54,65,0}},
/* 29203 */ {(12<<2)|2,{54,65,0}},
/* 29204 */ {(12<<2)|2,{54,65,0}},
/* 29205 */ {(12<<2)|2,{54,65,0}},
/* 29206 */ {(12<<2)|2,{54,65,0}},
/* 29207 */ {(12<<2)|2,{54,65,0}},
/* 29208 */ {(12<<2)|2,{54,65,0}},
/* 29209 */ {(12<<2)|2,{54,65,0}},
/* 29210 */ {(12<<2)|2,{54,65,0}},
/* 29211 */ {(12<<2)|2,{54,65,0}},
/* 29212 */ {(12<<2)|2,{54,65,0}},
/* 29213 */ {(12<<2)|2,{54,65,0}},
/* 29214 */ {(12<<2)|2,{54,65,0}},
/* 29215 */ {(12<<2)|2,{54,65,0}},
/* 29216 */ {(12<<2)|2,{54,95,0}},
/* 29217 */ {(12<<2)|2,{54,95,0}},
/* 29218 */ {(12<<2)|2,{54,95,0}},
/* 29219 */ {(12<<2)|2,{54,95,0}},
/* 29220 */ {(12<<2)|2,{54,95,0}},
/* 29221 */ {(12<<2)|2,{54,95,0}},
/* 29222 */ {(12<<2)|2,{54,95,0}},
/* 29223 */ {(12<<2)|2,{54,95,0}},
/* 29224 */ {(12<<2)|2,{54,95,0}},
/* 29225 */ {(12<<2)|2,{54,95,0}},
/* 29226 */ {(12<<2)|2,{54,95,0}},
/* 29227 */ {(12<<2)|2,{54,95,0}},
/* 29228 */ {(12<<2)|2,{54,95,0}},
/* 29229 */ {(12<<2)|2,{54,95,0}},
/* 29230 */ {(12<<2)|2,{54,95,0}},
/* 29231 */ {(12<<2)|2,{54,95,0}},
/* 29232 */ {(12<<2)|2,{54,98,0}},
/* 29233 */ {(12<<2)|2,{54,98,0}},
/* 29234 */ {(12<<2)|2,{54,98,0}},
/* 29235 */ {(12<<2)|2,{54,98,0}},
/* 29236 */ {(12<<2)|2,{54,98,0}},
/* 29237 */ {(12<<2)|2,{54,98,0}},
/* 29238 */ {(12<<2)|2,{54,98,0}},
/* 29239 */ {(12<<2)|2,{54,98,0}},
/* 29240 */ {(12<<2)|2,{54,98,0}},
/* 29241 */ {(12<<2)|2,{54,98,0}},
/* 29242 */ {(12<<2)|2,{54,98,0}},
/* 29243 */ {(12<<2)|2,{54,98,0}},
/* 29244 */ {(12<<2)|2,{54,98,0}},
/* 29245 */ {(12<<2)|2,{54,98,0}},
/* 29246 */ {(12<<2)|2,{54,98,0}},
/* 29247 */ {(12<<2)|2,{54,98,0}},
/* 29248 */ {(12<<2)|2,{54,100,0}},
/* 29249 */ {(12<<2)|2,{54,100,0}},
/* 29250 */ {(12<<2)|2,{54,100,0}},
/* 29251 */ {(12<<2)|2,{54,100,0}},
/* 29252 */ {(12<<2)|2,{54,100,0}},
/* 29253 */ {(12<<2)|2,{54,100,0}},
/* 29254 */ {(12<<2)|2,{54,100,0}},
/* 29255 */ {(12<<2)|2,{54,100,0}},
/* 29256 */ {(12<<2)|2,{54,100,0}},
/* 29257 */ {(12<<2)|2,{54,100,0}},
/* 29258 */ {(12<<2)|2,{54,100,0}},
/* 29259 */ {(12<<2)|2,{54,100,0}},
/* 29260 */ {(12<<2)|2,{54,100,0}},
/* 29261 */ {(12<<2)|2,{54,100,0}},
/* 29262 */ {(12<<2)|2,{54,100,0}},
/* 29263 */ {(12<<2)|2,{54,100,0}},
/* 29264 */ {(12<<2)|2,{54,102,0}},
/* 29265 */ {(12<<2)|2,{54,102,0}},
/* 29266 */ {(12<<2)|2,{54,102,0}},
/* 29267 */ {(12<<2)|2,{54,102,0}},
/* 29268 */ {(12<<2)|2,{54,102,0}},
/* 29269 */ {(12<<2)|2,{54,102,0}},
/* 29270 */ {(12<<2)|2,{54,102,0}},
/* 29271 */ {(12<<2)|2,{54,102,0}},
/* 29272 */ {(12<<2)|2,{54,102,0}},
/* 29273 */ {(12<<2)|2,{54,102,0}},
/* 29274 */ {(12<<2)|2,{54,102,0}},
/* 29275 */ {(12<<2)|2,{54,102,0}},
/* 29276 */ {(12<<2)|2,{54,102,0}},
/* 29277 */ {(12<<2)|2,{54,102,0}},
/* 29278 */ {(12<<2)|2,{54,102,0}},
/* 29279 */ {(12<<2)|2,{54,102,0}},
/* 29280 */ {(12<<2)|2,{54,103,0}},
/* 29281 */ {(12<<2)|2,{54,103,0}},
/* 29282 */ {(12<<2)|2,{54,103,0}},
/* 29283 */ {(12<<2)|2,{54,103,0}},
/* 29284 */ {(12<<2)|2,{54,103,0}},
/* 29285 */ {(12<<2)|2,{54,103,0}},
/* 29286 */ {(12<<2)|2,{54,103,0}},
/* 29287 */ {(12<<2)|2,{54,103,0}},
/* 29288 */ {(12<<2)|2,{54,103,0}},
/* 29289 */ {(12<<2)|2,{54,103,0}},
/* 29290 */ {(12<<2)|2,{54,103,0}},
/* 29291 */ {(12<<2)|2,{54,103,0}},
/* 29292 */ {(12<<2)|2,{54,103,0}},
/* 29293 */ {(12<<2)|2,{54,103,0}},
/* 29294 */ {(12<<2)|2,{54,103,0}},
/* 29295 */ {(12<<2)|2,{54,103,0}},
/* 29296 */ {(12<<2)|2,{54,104,0}},
/* 29297 */ {(12<<2)|2,{54,104,0}},
/* 29298 */ {(12<<2)|2,{54,104,0}},
/* 29299 */ {(12<<2)|2,{54,104,0}},
/* 29300 */ {(12<<2)|2,{54,104,0}},
/* 29301 */ {(12<<2)|2,{54,104,0}},
/* 29302 */ {(12<<2)|2,{54,104,0}},
/* 29303 */ {(12<<2)|2,{54,104,0}},
/* 29304 */ {(12<<2)|2,{54,104,0}},
/* 29305 */ {(12<<2)|2,{54,104,0}},
/* 29306 */ {(12<<2)|2,{54,104,0}},
/* 29307 */ {(12<<2)|2,{54,104,0}},
/* 29308 */ {(12<<2)|2,{54,104,0}},
/* 29309 */ {(12<<2)|2,{54,104,0}},
/* 29310 */ {(12<<2)|2,{54,104,0}},
/* 29311 */ {(12<<2)|2,{54,104,0}},
/* 29312 */ {(12<<2)|2,{54,108,0}},
/* 29313 */ {(12<<2)|2,{54,108,0}},
/* 29314 */ {(12<<2)|2,{54,108,0}},
/* 29315 */ {(12<<2)|2,{54,108,0}},
/* 29316 */ {(12<<2)|2,{54,108,0}},
/* 29317 */ {(12<<2)|2,{54,108,0}},
/* 29318 */ {(12<<2)|2,{54,108,0}},
/* 29319 */ {(12<<2)|2,{54,108,0}},
/* 29320 */ {(12<<2)|2,{54,108,0}},
/* 29321 */ {(12<<2)|2,{54,108,0}},
/* 29322 */ {(12<<2)|2,{54,108,0}},
/* 29323 */ {(12<<2)|2,{54,108,0}},
/* 29324 */ {(12<<2)|2,{54,108,0}},
/* 29325 */ {(12<<2)|2,{54,108,0}},
/* 29326 */ {(12<<2)|2,{54,108,0}},
/* 29327 */ {(12<<2)|2,{54,108,0}},
/* 29328 */ {(12<<2)|2,{54,109,0}},
/* 29329 */ {(12<<2)|2,{54,109,0}},
/* 29330 */ {(12<<2)|2,{54,109,0}},
/* 29331 */ {(12<<2)|2,{54,109,0}},
/* 29332 */ {(12<<2)|2,{54,109,0}},
/* 29333 */ {(12<<2)|2,{54,109,0}},
/* 29334 */ {(12<<2)|2,{54,109,0}},
/* 29335 */ {(12<<2)|2,{54,109,0}},
/* 29336 */ {(12<<2)|2,{54,109,0}},
/* 29337 */ {(12<<2)|2,{54,109,0}},
/* 29338 */ {(12<<2)|2,{54,109,0}},
/* 29339 */ {(12<<2)|2,{54,109,0}},
/* 29340 */ {(12<<2)|2,{54,109,0}},
/* 29341 */ {(12<<2)|2,{54,109,0}},
/* 29342 */ {(12<<2)|2,{54,109,0}},
/* 29343 */ {(12<<2)|2,{54,109,0}},
/* 29344 */ {(12<<2)|2,{54,110,0}},
/* 29345 */ {(12<<2)|2,{54,110,0}},
/* 29346 */ {(12<<2)|2,{54,110,0}},
/* 29347 */ {(12<<2)|2,{54,110,0}},
/* 29348 */ {(12<<2)|2,{54,110,0}},
/* 29349 */ {(12<<2)|2,{54,110,0}},
/* 29350 */ {(12<<2)|2,{54,110,0}},
/* 29351 */ {(12<<2)|2,{54,110,0}},
/* 29352 */ {(12<<2)|2,{54,110,0}},
/* 29353 */ {(12<<2)|2,{54,110,0}},
/* 29354 */ {(12<<2)|2,{54,110,0}},
/* 29355 */ {(12<<2)|2,{54,110,0}},
/* 29356 */ {(12<<2)|2,{54,110,0}},
/* 29357 */ {(12<<2)|2,{54,110,0}},
/* 29358 */ {(12<<2)|2,{54,110,0}},
/* 29359 */ {(12<<2)|2,{54,110,0}},
/* 29360 */ {(12<<2)|2,{54,112,0}},
/* 29361 */ {(12<<2)|2,{54,112,0}},
/* 29362 */ {(12<<2)|2,{54,112,0}},
/* 29363 */ {(12<<2)|2,{54,112,0}},
/* 29364 */ {(12<<2)|2,{54,112,0}},
/* 29365 */ {(12<<2)|2,{54,112,0}},
/* 29366 */ {(12<<2)|2,{54,112,0}},
/* 29367 */ {(12<<2)|2,{54,112,0}},
/* 29368 */ {(12<<2)|2,{54,112,0}},
/* 29369 */ {(12<<2)|2,{54,112,0}},
/* 29370 */ {(12<<2)|2,{54,112,0}},
/* 29371 */ {(12<<2)|2,{54,112,0}},
/* 29372 */ {(12<<2)|2,{54,112,0}},
/* 29373 */ {(12<<2)|2,{54,112,0}},
/* 29374 */ {(12<<2)|2,{54,112,0}},
/* 29375 */ {(12<<2)|2,{54,112,0}},
/* 29376 */ {(12<<2)|2,{54,114,0}},
/* 29377 */ {(12<<2)|2,{54,114,0}},
/* 29378 */ {(12<<2)|2,{54,114,0}},
/* 29379 */ {(12<<2)|2,{54,114,0}},
/* 29380 */ {(12<<2)|2,{54,114,0}},
/* 29381 */ {(12<<2)|2,{54,114,0}},
/* 29382 */ {(12<<2)|2,{54,114,0}},
/* 29383 */ {(12<<2)|2,{54,114,0}},
/* 29384 */ {(12<<2)|2,{54,114,0}},
/* 29385 */ {(12<<2)|2,{54,114,0}},
/* 29386 */ {(12<<2)|2,{54,114,0}},
/* 29387 */ {(12<<2)|2,{54,114,0}},
/* 29388 */ {(12<<2)|2,{54,114,0}},
/* 29389 */ {(12<<2)|2,{54,114,0}},
/* 29390 */ {(12<<2)|2,{54,114,0}},
/* 29391 */ {(12<<2)|2,{54,114,0}},
/* 29392 */ {(12<<2)|2,{54,117,0}},
/* 29393 */ {(12<<2)|2,{54,117,0}},
/* 29394 */ {(12<<2)|2,{54,117,0}},
/* 29395 */ {(12<<2)|2,{54,117,0}},
/* 29396 */ {(12<<2)|2,{54,117,0}},
/* 29397 */ {(12<<2)|2,{54,117,0}},
/* 29398 */ {(12<<2)|2,{54,117,0}},
/* 29399 */ {(12<<2)|2,{54,117,0}},
/* 29400 */ {(12<<2)|2,{54,117,0}},
/* 29401 */ {(12<<2)|2,{54,117,0}},
/* 29402 */ {(12<<2)|2,{54,117,0}},
/* 29403 */ {(12<<2)|2,{54,117,0}},
/* 29404 */ {(12<<2)|2,{54,117,0}},
/* 29405 */ {(12<<2)|2,{54,117,0}},
/* 29406 */ {(12<<2)|2,{54,117,0}},
/* 29407 */ {(12<<2)|2,{54,117,0}},
/* 29408 */ {(13<<2)|2,{54,58,0}},
/* 29409 */ {(13<<2)|2,{54,58,0}},
/* 29410 */ {(13<<2)|2,{54,58,0}},
/* 29411 */ {(13<<2)|2,{54,58,0}},
/* 29412 */ {(13<<2)|2,{54,58,0}},
/* 29413 */ {(13<<2)|2,{54,58,0}},
/* 29414 */ {(13<<2)|2,{54,58,0}},
/* 29415 */ {(13<<2)|2,{54,58,0}},
/* 29416 */ {(13<<2)|2,{54,66,0}},
/* 29417 */ {(13<<2)|2,{54,66,0}},
/* 29418 */ {(13<<2)|2,{54,66,0}},
/* 29419 */ {(13<<2)|2,{54,66,0}},
/* 29420 */ {(13<<2)|2,{54,66,0}},
/* 29421 */ {(13<<2)|2,{54,66,0}},
/* 29422 */ {(13<<2)|2,{54,66,0}},
/* 29423 */ {(13<<2)|2,{54,66,0}},
/* 29424 */ {(13<<2)|2,{54,67,0}},
/* 29425 */ {(13<<2)|2,{54,67,0}},
/* 29426 */ {(13<<2)|2,{54,67,0}},
/* 29427 */ {(13<<2)|2,{54,67,0}},
/* 29428 */ {(13<<2)|2,{54,67,0}},
/* 29429 */ {(13<<2)|2,{54,67,0}},
/* 29430 */ {(13<<2)|2,{54,67,0}},
/* 29431 */ {(13<<2)|2,{54,67,0}},
/* 29432 */ {(13<<2)|2,{54,68,0}},
/* 29433 */ {(13<<2)|2,{54,68,0}},
/* 29434 */ {(13<<2)|2,{54,68,0}},
/* 29435 */ {(13<<2)|2,{54,68,0}},
/* 29436 */ {(13<<2)|2,{54,68,0}},
/* 29437 */ {(13<<2)|2,{54,68,0}},
/* 29438 */ {(13<<2)|2,{54,68,0}},
/* 29439 */ {(13<<2)|2,{54,68,0}},
/* 29440 */ {(13<<2)|2,{54,69,0}},
/* 29441 */ {(13<<2)|2,{54,69,0}},
/* 29442 */ {(13<<2)|2,{54,69,0}},
/* 29443 */ {(13<<2)|2,{54,69,0}},
/* 29444 */ {(13<<2)|2,{54,69,0}},
/* 29445 */ {(13<<2)|2,{54,69,0}},
/* 29446 */ {(13<<2)|2,{54,69,0}},
/* 29447 */ {(13<<2)|2,{54,69,0}},
/* 29448 */ {(13<<2)|2,{54,70,0}},
/* 29449 */ {(13<<2)|2,{54,70,0}},
/* 29450 */ {(13<<2)|2,{54,70,0}},
/* 29451 */ {(13<<2)|2,{54,70,0}},
/* 29452 */ {(13<<2)|2,{54,70,0}},
/* 29453 */ {(13<<2)|2,{54,70,0}},
/* 29454 */ {(13<<2)|2,{54,70,0}},
/* 29455 */ {(13<<2)|2,{54,70,0}},
/* 29456 */ {(13<<2)|2,{54,71,0}},
/* 29457 */ {(13<<2)|2,{54,71,0}},
/* 29458 */ {(13<<2)|2,{54,71,0}},
/* 29459 */ {(13<<2)|2,{54,71,0}},
/* 29460 */ {(13<<2)|2,{54,71,0}},
/* 29461 */ {(13<<2)|2,{54,71,0}},
/* 29462 */ {(13<<2)|2,{54,71,0}},
/* 29463 */ {(13<<2)|2,{54,71,0}},
/* 29464 */ {(13<<2)|2,{54,72,0}},
/* 29465 */ {(13<<2)|2,{54,72,0}},
/* 29466 */ {(13<<2)|2,{54,72,0}},
/* 29467 */ {(13<<2)|2,{54,72,0}},
/* 29468 */ {(13<<2)|2,{54,72,0}},
/* 29469 */ {(13<<2)|2,{54,72,0}},
/* 29470 */ {(13<<2)|2,{54,72,0}},
/* 29471 */ {(13<<2)|2,{54,72,0}},
/* 29472 */ {(13<<2)|2,{54,73,0}},
/* 29473 */ {(13<<2)|2,{54,73,0}},
/* 29474 */ {(13<<2)|2,{54,73,0}},
/* 29475 */ {(13<<2)|2,{54,73,0}},
/* 29476 */ {(13<<2)|2,{54,73,0}},
/* 29477 */ {(13<<2)|2,{54,73,0}},
/* 29478 */ {(13<<2)|2,{54,73,0}},
/* 29479 */ {(13<<2)|2,{54,73,0}},
/* 29480 */ {(13<<2)|2,{54,74,0}},
/* 29481 */ {(13<<2)|2,{54,74,0}},
/* 29482 */ {(13<<2)|2,{54,74,0}},
/* 29483 */ {(13<<2)|2,{54,74,0}},
/* 29484 */ {(13<<2)|2,{54,74,0}},
/* 29485 */ {(13<<2)|2,{54,74,0}},
/* 29486 */ {(13<<2)|2,{54,74,0}},
/* 29487 */ {(13<<2)|2,{54,74,0}},
/* 29488 */ {(13<<2)|2,{54,75,0}},
/* 29489 */ {(13<<2)|2,{54,75,0}},
/* 29490 */ {(13<<2)|2,{54,75,0}},
/* 29491 */ {(13<<2)|2,{54,75,0}},
/* 29492 */ {(13<<2)|2,{54,75,0}},
/* 29493 */ {(13<<2)|2,{54,75,0}},
/* 29494 */ {(13<<2)|2,{54,75,0}},
/* 29495 */ {(13<<2)|2,{54,75,0}},
/* 29496 */ {(13<<2)|2,{54,76,0}},
/* 29497 */ {(13<<2)|2,{54,76,0}},
/* 29498 */ {(13<<2)|2,{54,76,0}},
/* 29499 */ {(13<<2)|2,{54,76,0}},
/* 29500 */ {(13<<2)|2,{54,76,0}},
/* 29501 */ {(13<<2)|2,{54,76,0}},
/* 29502 */ {(13<<2)|2,{54,76,0}},
/* 29503 */ {(13<<2)|2,{54,76,0}},
/* 29504 */ {(13<<2)|2,{54,77,0}},
/* 29505 */ {(13<<2)|2,{54,77,0}},
/* 29506 */ {(13<<2)|2,{54,77,0}},
/* 29507 */ {(13<<2)|2,{54,77,0}},
/* 29508 */ {(13<<2)|2,{54,77,0}},
/* 29509 */ {(13<<2)|2,{54,77,0}},
/* 29510 */ {(13<<2)|2,{54,77,0}},
/* 29511 */ {(13<<2)|2,{54,77,0}},
/* 29512 */ {(13<<2)|2,{54,78,0}},
/* 29513 */ {(13<<2)|2,{54,78,0}},
/* 29514 */ {(13<<2)|2,{54,78,0}},
/* 29515 */ {(13<<2)|2,{54,78,0}},
/* 29516 */ {(13<<2)|2,{54,78,0}},
/* 29517 */ {(13<<2)|2,{54,78,0}},
/* 29518 */ {(13<<2)|2,{54,78,0}},
/* 29519 */ {(13<<2)|2,{54,78,0}},
/* 29520 */ {(13<<2)|2,{54,79,0}},
/* 29521 */ {(13<<2)|2,{54,79,0}},
/* 29522 */ {(13<<2)|2,{54,79,0}},
/* 29523 */ {(13<<2)|2,{54,79,0}},
/* 29524 */ {(13<<2)|2,{54,79,0}},
/* 29525 */ {(13<<2)|2,{54,79,0}},
/* 29526 */ {(13<<2)|2,{54,79,0}},
/* 29527 */ {(13<<2)|2,{54,79,0}},
/* 29528 */ {(13<<2)|2,{54,80,0}},
/* 29529 */ {(13<<2)|2,{54,80,0}},
/* 29530 */ {(13<<2)|2,{54,80,0}},
/* 29531 */ {(13<<2)|2,{54,80,0}},
/* 29532 */ {(13<<2)|2,{54,80,0}},
/* 29533 */ {(13<<2)|2,{54,80,0}},
/* 29534 */ {(13<<2)|2,{54,80,0}},
/* 29535 */ {(13<<2)|2,{54,80,0}},
/* 29536 */ {(13<<2)|2,{54,81,0}},
/* 29537 */ {(13<<2)|2,{54,81,0}},
/* 29538 */ {(13<<2)|2,{54,81,0}},
/* 29539 */ {(13<<2)|2,{54,81,0}},
/* 29540 */ {(13<<2)|2,{54,81,0}},
/* 29541 */ {(13<<2)|2,{54,81,0}},
/* 29542 */ {(13<<2)|2,{54,81,0}},
/* 29543 */ {(13<<2)|2,{54,81,0}},
/* 29544 */ {(13<<2)|2,{54,82,0}},
/* 29545 */ {(13<<2)|2,{54,82,0}},
/* 29546 */ {(13<<2)|2,{54,82,0}},
/* 29547 */ {(13<<2)|2,{54,82,0}},
/* 29548 */ {(13<<2)|2,{54,82,0}},
/* 29549 */ {(13<<2)|2,{54,82,0}},
/* 29550 */ {(13<<2)|2,{54,82,0}},
/* 29551 */ {(13<<2)|2,{54,82,0}},
/* 29552 */ {(13<<2)|2,{54,83,0}},
/* 29553 */ {(13<<2)|2,{54,83,0}},
/* 29554 */ {(13<<2)|2,{54,83,0}},
/* 29555 */ {(13<<2)|2,{54,83,0}},
/* 29556 */ {(13<<2)|2,{54,83,0}},
/* 29557 */ {(13<<2)|2,{54,83,0}},
/* 29558 */ {(13<<2)|2,{54,83,0}},
/* 29559 */ {(13<<2)|2,{54,83,0}},
/* 29560 */ {(13<<2)|2,{54,84,0}},
/* 29561 */ {(13<<2)|2,{54,84,0}},
/* 29562 */ {(13<<2)|2,{54,84,0}},
/* 29563 */ {(13<<2)|2,{54,84,0}},
/* 29564 */ {(13<<2)|2,{54,84,0}},
/* 29565 */ {(13<<2)|2,{54,84,0}},
/* 29566 */ {(13<<2)|2,{54,84,0}},
/* 29567 */ {(13<<2)|2,{54,84,0}},
/* 29568 */ {(13<<2)|2,{54,85,0}},
/* 29569 */ {(13<<2)|2,{54,85,0}},
/* 29570 */ {(13<<2)|2,{54,85,0}},
/* 29571 */ {(13<<2)|2,{54,85,0}},
/* 29572 */ {(13<<2)|2,{54,85,0}},
/* 29573 */ {(13<<2)|2,{54,85,0}},
/* 29574 */ {(13<<2)|2,{54,85,0}},
/* 29575 */ {(13<<2)|2,{54,85,0}},
/* 29576 */ {(13<<2)|2,{54,86,0}},
/* 29577 */ {(13<<2)|2,{54,86,0}},
/* 29578 */ {(13<<2)|2,{54,86,0}},
/* 29579 */ {(13<<2)|2,{54,86,0}},
/* 29580 */ {(13<<2)|2,{54,86,0}},
/* 29581 */ {(13<<2)|2,{54,86,0}},
/* 29582 */ {(13<<2)|2,{54,86,0}},
/* 29583 */ {(13<<2)|2,{54,86,0}},
/* 29584 */ {(13<<2)|2,{54,87,0}},
/* 29585 */ {(13<<2)|2,{54,87,0}},
/* 29586 */ {(13<<2)|2,{54,87,0}},
/* 29587 */ {(13<<2)|2,{54,87,0}},
/* 29588 */ {(13<<2)|2,{54,87,0}},
/* 29589 */ {(13<<2)|2,{54,87,0}},
/* 29590 */ {(13<<2)|2,{54,87,0}},
/* 29591 */ {(13<<2)|2,{54,87,0}},
/* 29592 */ {(13<<2)|2,{54,89,0}},
/* 29593 */ {(13<<2)|2,{54,89,0}},
/* 29594 */ {(13<<2)|2,{54,89,0}},
/* 29595 */ {(13<<2)|2,{54,89,0}},
/* 29596 */ {(13<<2)|2,{54,89,0}},
/* 29597 */ {(13<<2)|2,{54,89,0}},
/* 29598 */ {(13<<2)|2,{54,89,0}},
/* 29599 */ {(13<<2)|2,{54,89,0}},
/* 29600 */ {(13<<2)|2,{54,106,0}},
/* 29601 */ {(13<<2)|2,{54,106,0}},
/* 29602 */ {(13<<2)|2,{54,106,0}},
/* 29603 */ {(13<<2)|2,{54,106,0}},
/* 29604 */ {(13<<2)|2,{54,106,0}},
/* 29605 */ {(13<<2)|2,{54,106,0}},
/* 29606 */ {(13<<2)|2,{54,106,0}},
/* 29607 */ {(13<<2)|2,{54,106,0}},
/* 29608 */ {(13<<2)|2,{54,107,0}},
/* 29609 */ {(13<<2)|2,{54,107,0}},
/* 29610 */ {(13<<2)|2,{54,107,0}},
/* 29611 */ {(13<<2)|2,{54,107,0}},
/* 29612 */ {(13<<2)|2,{54,107,0}},
/* 29613 */ {(13<<2)|2,{54,107,0}},
/* 29614 */ {(13<<2)|2,{54,107,0}},
/* 29615 */ {(13<<2)|2,{54,107,0}},
/* 29616 */ {(13<<2)|2,{54,113,0}},
/* 29617 */ {(13<<2)|2,{54,113,0}},
/* 29618 */ {(13<<2)|2,{54,113,0}},
/* 29619 */ {(13<<2)|2,{54,113,0}},
/* 29620 */ {(13<<2)|2,{54,113,0}},
/* 29621 */ {(13<<2)|2,{54,113,0}},
/* 29622 */ {(13<<2)|2,{54,113,0}},
/* 29623 */ {(13<<2)|2,{54,113,0}},
/* 29624 */ {(13<<2)|2,{54,118,0}},
/* 29625 */ {(13<<2)|2,{54,118,0}},
/* 29626 */ {(13<<2)|2,{54,118,0}},
/* 29627 */ {(13<<2)|2,{54,118,0}},
/* 29628 */ {(13<<2)|2,{54,118,0}},
/* 29629 */ {(13<<2)|2,{54,118,0}},
/* 29630 */ {(13<<2)|2,{54,118,0}},
/* 29631 */ {(13<<2)|2,{54,118,0}},
/* 29632 */ {(13<<2)|2,{54,119,0}},
/* 29633 */ {(13<<2)|2,{54,119,0}},
/* 29634 */ {(13<<2)|2,{54,119,0}},
/* 29635 */ {(13<<2)|2,{54,119,0}},
/* 29636 */ {(13<<2)|2,{54,119,0}},
/* 29637 */ {(13<<2)|2,{54,119,0}},
/* 29638 */ {(13<<2)|2,{54,119,0}},
/* 29639 */ {(13<<2)|2,{54,119,0}},
/* 29640 */ {(13<<2)|2,{54,120,0}},
/* 29641 */ {(13<<2)|2,{54,120,0}},
/* 29642 */ {(13<<2)|2,{54,120,0}},
/* 29643 */ {(13<<2)|2,{54,120,0}},
/* 29644 */ {(13<<2)|2,{54,120,0}},
/* 29645 */ {(13<<2)|2,{54,120,0}},
/* 29646 */ {(13<<2)|2,{54,120,0}},
/* 29647 */ {(13<<2)|2,{54,120,0}},
/* 29648 */ {(13<<2)|2,{54,121,0}},
/* 29649 */ {(13<<2)|2,{54,121,0}},
/* 29650 */ {(13<<2)|2,{54,121,0}},
/* 29651 */ {(13<<2)|2,{54,121,0}},
/* 29652 */ {(13<<2)|2,{54,121,0}},
/* 29653 */ {(13<<2)|2,{54,121,0}},
/* 29654 */ {(13<<2)|2,{54,121,0}},
/* 29655 */ {(13<<2)|2,{54,121,0}},
/* 29656 */ {(13<<2)|2,{54,122,0}},
/* 29657 */ {(13<<2)|2,{54,122,0}},
/* 29658 */ {(13<<2)|2,{54,122,0}},
/* 29659 */ {(13<<2)|2,{54,122,0}},
/* 29660 */ {(13<<2)|2,{54,122,0}},
/* 29661 */ {(13<<2)|2,{54,122,0}},
/* 29662 */ {(13<<2)|2,{54,122,0}},
/* 29663 */ {(13<<2)|2,{54,122,0}},
/* 29664 */ {(14<<2)|2,{54,38,0}},
/* 29665 */ {(14<<2)|2,{54,38,0}},
/* 29666 */ {(14<<2)|2,{54,38,0}},
/* 29667 */ {(14<<2)|2,{54,38,0}},
/* 29668 */ {(14<<2)|2,{54,42,0}},
/* 29669 */ {(14<<2)|2,{54,42,0}},
/* 29670 */ {(14<<2)|2,{54,42,0}},
/* 29671 */ {(14<<2)|2,{54,42,0}},
/* 29672 */ {(14<<2)|2,{54,44,0}},
/* 29673 */ {(14<<2)|2,{54,44,0}},
/* 29674 */ {(14<<2)|2,{54,44,0}},
/* 29675 */ {(14<<2)|2,{54,44,0}},
/* 29676 */ {(14<<2)|2,{54,59,0}},
/* 29677 */ {(14<<2)|2,{54,59,0}},
/* 29678 */ {(14<<2)|2,{54,59,0}},
/* 29679 */ {(14<<2)|2,{54,59,0}},
/* 29680 */ {(14<<2)|2,{54,88,0}},
/* 29681 */ {(14<<2)|2,{54,88,0}},
/* 29682 */ {(14<<2)|2,{54,88,0}},
/* 29683 */ {(14<<2)|2,{54,88,0}},
/* 29684 */ {(14<<2)|2,{54,90,0}},
/* 29685 */ {(14<<2)|2,{54,90,0}},
/* 29686 */ {(14<<2)|2,{54,90,0}},
/* 29687 */ {(14<<2)|2,{54,90,0}},
/* 29688 */ {(16<<2)|2,{54,33,0}},
/* 29689 */ {(16<<2)|2,{54,34,0}},
/* 29690 */ {(16<<2)|2,{54,40,0}},
/* 29691 */ {(16<<2)|2,{54,41,0}},
/* 29692 */ {(16<<2)|2,{54,63,0}},
/* 29693 */ {(6<<2)|1,{54,0,0}},
/* 29694 */ {(6<<2)|1,{54,0,0}},
/* 29695 */ {(6<<2)|1,{54,0,0}},
/* 29696 */ {(16<<2)|3,{55,48,48}},
/* 29697 */ {(16<<2)|3,{55,48,49}},
/* 29698 */ {(16<<2)|3,{55,48,50}},
/* 29699 */ {(16<<2)|3,{55,48,97}},
/* 29700 */ {(16<<2)|3,{55,48,99}},
/* 29701 */ {(16<<2)|3,{55,48,101}},
/* 29702 */ {(16<<2)|3,{55,48,105}},
/* 29703 */ {(16<<2)|3,{55,48,111}},
/* 29704 */ {(16<<2)|3,{55,48,115}},
/* 29705 */ {(16<<2)|3,{55,48,116}},
/* 29706 */ {(11<<2)|2,{55,48,0}},
/* 29707 */ {(11<<2)|2,{55,48,0}},
/* 29708 */ {(11<<2)|2,{55,48,0}},
/* 29709 */ {(11<<2)|2,{55,48,0}},
/* 29710 */ {(11<<2)|2,{55,48,0}},
/* 29711 */ {(11<<2)|2,{55,48,0}},
/* 29712 */ {(11<<2)|2,{55,48,0}},
/* 29713 */ {(11<<2)|2,{55,48,0}},
/* 29714 */ {(11<<2)|2,{55,48,0}},
/* 29715 */ {(11<<2)|2,{55,48,0}},
/* 29716 */ {(11<<2)|2,{55,48,0}},
/* 29717 */ {(11<<2)|2,{55,48,0}},
/* 29718 */ {(11<<2)|2,{55,48,0}},
/* 29719 */ {(11<<2)|2,{55,48,0}},
/* 29720 */ {(11<<2)|2,{55,48,0}},
/* 29721 */ {(11<<2)|2,{55,48,0}},
/* 29722 */ {(11<<2)|2,{55,48,0}},
/* 29723 */ {(11<<2)|2,{55,48,0}},
/* 29724 */ {(11<<2)|2,{55,48,0}},
/* 29725 */ {(11<<2)|2,{55,48,0}},
/* 29726 */ {(11<<2)|2,{55,48,0}},
/* 29727 */ {(11<<2)|2,{55,48,0}},
/* 29728 */ {(16<<2)|3,{55,49,48}},
/* 29729 */ {(16<<2)|3,{55,49,49}},
/* 29730 */ {(16<<2)|3,{55,49,50}},
/* 29731 */ {(16<<2)|3,{55,49,97}},
/* 29732 */ {(16<<2)|3,{55,49,99}},
/* 29733 */ {(16<<2)|3,{55,49,101}},
/* 29734 */ {(16<<2)|3,{55,49,105}},
/* 29735 */ {(16<<2)|3,{55,49,111}},
/* 29736 */ {(16<<2)|3,{55,49,115}},
/* 29737 */ {(16<<2)|3,{55,49,116}},
/* 29738 */ {(11<<2)|2,{55,49,0}},
/* 29739 */ {(11<<2)|2,{55,49,0}},
/* 29740 */ {(11<<2)|2,{55,49,0}},
/* 29741 */ {(11<<2)|2,{55,49,0}},
/* 29742 */ {(11<<2)|2,{55,49,0}},
/* 29743 */ {(11<<2)|2,{55,49,0}},
/* 29744 */ {(11<<2)|2,{55,49,0}},
/* 29745 */ {(11<<2)|2,{55,49,0}},
/* 29746 */ {(11<<2)|2,{55,49,0}},
/* 29747 */ {(11<<2)|2,{55,49,0}},
/* 29748 */ {(11<<2)|2,{55,49,0}},
/* 29749 */ {(11<<2)|2,{55,49,0}},
/* 29750 */ {(11<<2)|2,{55,49,0}},
/* 29751 */ {(11<<2)|2,{55,49,0}},
/* 29752 */ {(11<<2)|2,{55,49,0}},
/* 29753 */ {(11<<2)|2,{55,49,0}},
/* 29754 */ {(11<<2)|2,{55,49,0}},
/* 29755 */ {(11<<2)|2,{55,49,0}},
/* 29756 */ {(11<<2)|2,{55,49,0}},
/* 29757 */ {(11<<2)|2,{55,49,0}},
/* 29758 */ {(11<<2)|2,{55,49,0}},
/* 29759 */ {(11<<2)|2,{55,49,0}},
/* 29760 */ {(16<<2)|3,{55,50,48}},
/* 29761 */ {(16<<2)|3,{55,50,49}},
/* 29762 */ {(16<<2)|3,{55,50,50}},
/* 29763 */ {(16<<2)|3,{55,50,97}},
/* 29764 */ {(16<<2)|3,{55,50,99}},
/* 29765 */ {(16<<2)|3,{55,50,101}},
/* 29766 */ {(16<<2)|3,{55,50,105}},
/* 29767 */ {(16<<2)|3,{55,50,111}},
/* 29768 */ {(16<<2)|3,{55,50,115}},
/* 29769 */ {(16<<2)|3,{55,50,116}},
/* 29770 */ {(11<<2)|2,{55,50,0}},
/* 29771 */ {(11<<2)|2,{55,50,0}},
/* 29772 */ {(11<<2)|2,{55,50,0}},
/* 29773 */ {(11<<2)|2,{55,50,0}},
/* 29774 */ {(11<<2)|2,{55,50,0}},
/* 29775 */ {(11<<2)|2,{55,50,0}},
/* 29776 */ {(11<<2)|2,{55,50,0}},
/* 29777 */ {(11<<2)|2,{55,50,0}},
/* 29778 */ {(11<<2)|2,{55,50,0}},
/* 29779 */ {(11<<2)|2,{55,50,0}},
/* 29780 */ {(11<<2)|2,{55,50,0}},
/* 29781 */ {(11<<2)|2,{55,50,0}},
/* 29782 */ {(11<<2)|2,{55,50,0}},
/* 29783 */ {(11<<2)|2,{55,50,0}},
/* 29784 */ {(11<<2)|2,{55,50,0}},
/* 29785 */ {(11<<2)|2,{55,50,0}},
/* 29786 */ {(11<<2)|2,{55,50,0}},
/* 29787 */ {(11<<2)|2,{55,50,0}},
/* 29788 */ {(11<<2)|2,{55,50,0}},
/* 29789 */ {(11<<2)|2,{55,50,0}},
/* 29790 */ {(11<<2)|2,{55,50,0}},
/* 29791 */ {(11<<2)|2,{55,50,0}},
/* 29792 */ {(16<<2)|3,{55,97,48}},
/* 29793 */ {(16<<2)|3,{55,97,49}},
/* 29794 */ {(16<<2)|3,{55,97,50}},
/* 29795 */ {(16<<2)|3,{55,97,97}},
/* 29796 */ {(16<<2)|3,{55,97,99}},
/* 29797 */ {(16<<2)|3,{55,97,101}},
/* 29798 */ {(16<<2)|3,{55,97,105}},
/* 29799 */ {(16<<2)|3,{55,97,111}},
/* 29800 */ {(16<<2)|3,{55,97,115}},
/* 29801 */ {(16<<2)|3,{55,97,116}},
/* 29802 */ {(11<<2)|2,{55,97,0}},
/* 29803 */ {(11<<2)|2,{55,97,0}},
/* 29804 */ {(11<<2)|2,{55,97,0}},
/* 29805 */ {(11<<2)|2,{55,97,0}},
/* 29806 */ {(11<<2)|2,{55,97,0}},
/* 29807 */ {(11<<2)|2,{55,97,0}},
/* 29808 */ {(11<<2)|2,{55,97,0}},
/* 29809 */ {(11<<2)|2,{55,97,0}},
/* 29810 */ {(11<<2)|2,{55,97,0}},
/* 29811 */ {(11<<2)|2,{55,97,0}},
/* 29812 */ {(11<<2)|2,{55,97,0}},
/* 29813 */ {(11<<2)|2,{55,97,0}},
/* 29814 */ {(11<<2)|2,{55,97,0}},
/* 29815 */ {(11<<2)|2,{55,97,0}},
/* 29816 */ {(11<<2)|2,{55,97,0}},
/* 29817 */ {(11<<2)|2,{55,97,0}},
/* 29818 */ {(11<<2)|2,{55,97,0}},
/* 29819 */ {(11<<2)|2,{55,97,0}},
/* 29820 */ {(11<<2)|2,{55,97,0}},
/* 29821 */ {(11<<2)|2,{55,97,0}},
/* 29822 */ {(11<<2)|2,{55,97,0}},
/* 29823 */ {(11<<2)|2,{55,97,0}},
/* 29824 */ {(16<<2)|3,{55,99,48}},
/* 29825 */ {(16<<2)|3,{55,99,49}},
/* 29826 */ {(16<<2)|3,{55,99,50}},
/* 29827 */ {(16<<2)|3,{55,99,97}},
/* 29828 */ {(16<<2)|3,{55,99,99}},
/* 29829 */ {(16<<2)|3,{55,99,101}},
/* 29830 */ {(16<<2)|3,{55,99,105}},
/* 29831 */ {(16<<2)|3,{55,99,111}},
/* 29832 */ {(16<<2)|3,{55,99,115}},
/* 29833 */ {(16<<2)|3,{55,99,116}},
/* 29834 */ {(11<<2)|2,{55,99,0}},
/* 29835 */ {(11<<2)|2,{55,99,0}},
/* 29836 */ {(11<<2)|2,{55,99,0}},
/* 29837 */ {(11<<2)|2,{55,99,0}},
/* 29838 */ {(11<<2)|2,{55,99,0}},
/* 29839 */ {(11<<2)|2,{55,99,0}},
/* 29840 */ {(11<<2)|2,{55,99,0}},
/* 29841 */ {(11<<2)|2,{55,99,0}},
/* 29842 */ {(11<<2)|2,{55,99,0}},
/* 29843 */ {(11<<2)|2,{55,99,0}},
/* 29844 */ {(11<<2)|2,{55,99,0}},
/* 29845 */ {(11<<2)|2,{55,99,0}},
/* 29846 */ {(11<<2)|2,{55,99,0}},
/* 29847 */ {(11<<2)|2,{55,99,0}},
/* 29848 */ {(11<<2)|2,{55,99,0}},
/* 29849 */ {(11<<2)|2,{55,99,0}},
/* 29850 */ {(11<<2)|2,{55,99,0}},
/* 29851 */ {(11<<2)|2,{55,99,0}},
/* 29852 */ {(11<<2)|2,{55,99,0}},
/* 29853 */ {(11<<2)|2,{55,99,0}},
/* 29854 */ {(11<<2)|2,{55,99,0}},
/* 29855 */ {(11<<2)|2,{55,99,0}},
/* 29856 */ {(16<<2)|3,{55,101,48}},
/* 29857 */ {(16<<2)|3,{55,101,49}},
/* 29858 */ {(16<<2)|3,{55,101,50}},
/* 29859 */ {(16<<2)|3,{55,101,97}},
/* 29860 */ {(16<<2)|3,{55,101,99}},
/* 29861 */ {(16<<2)|3,{55,101,101}},
/* 29862 */ {(16<<2)|3,{55,101,105}},
/* 29863 */ {(16<<2)|3,{55,101,111}},
/* 29864 */ {(16<<2)|3,{55,101,115}},
/* 29865 */ {(16<<2)|3,{55,101,116}},
/* 29866 */ {(11<<2)|2,{55,101,0}},
/* 29867 */ {(11<<2)|2,{55,101,0}},
/* 29868 */ {(11<<2)|2,{55,101,0}},
/* 29869 */ {(11<<2)|2,{55,101,0}},
/* 29870 */ {(11<<2)|2,{55,101,0}},
/* 29871 */ {(11<<2)|2,{55,101,0}},
/* 29872 */ {(11<<2)|2,{55,101,0}},
/* 29873 */ {(11<<2)|2,{55,101,0}},
/* 29874 */ {(11<<2)|2,{55,101,0}},
/* 29875 */ {(11<<2)|2,{55,101,0}},
/* 29876 */ {(11<<2)|2,{55,101,0}},
/* 29877 */ {(11<<2)|2,{55,101,0}},
/* 29878 */ {(11<<2)|2,{55,101,0}},
/* 29879 */ {(11<<2)|2,{55,101,0}},
/* 29880 */ {(11<<2)|2,{55,101,0}},
/* 29881 */ {(11<<2)|2,{55,101,0}},
/* 29882 */ {(11<<2)|2,{55,101,0}},
/* 29883 */ {(11<<2)|2,{55,101,0}},
/* 29884 */ {(11<<2)|2,{55,101,0}},
/* 29885 */ {(11<<2)|2,{55,101,0}},
/* 29886 */ {(11<<2)|2,{55,101,0}},
/* 29887 */ {(11<<2)|2,{55,101,0}},
/* 29888 */ {(16<<2)|3,{55,105,48}},
/* 29889 */ {(16<<2)|3,{55,105,49}},
/* 29890 */ {(16<<2)|3,{55,105,50}},
/* 29891 */ {(16<<2)|3,{55,105,97}},
/* 29892 */ {(16<<2)|3,{55,105,99}},
/* 29893 */ {(16<<2)|3,{55,105,101}},
/* 29894 */ {(16<<2)|3,{55,105,105}},
/* 29895 */ {(16<<2)|3,{55,105,111}},
/* 29896 */ {(16<<2)|3,{55,105,115}},
/* 29897 */ {(16<<2)|3,{55,105,116}},
/* 29898 */ {(11<<2)|2,{55,105,0}},
/* 29899 */ {(11<<2)|2,{55,105,0}},
/* 29900 */ {(11<<2)|2,{55,105,0}},
/* 29901 */ {(11<<2)|2,{55,105,0}},
/* 29902 */ {(11<<2)|2,{55,105,0}},
/* 29903 */ {(11<<2)|2,{55,105,0}},
/* 29904 */ {(11<<2)|2,{55,105,0}},
/* 29905 */ {(11<<2)|2,{55,105,0}},
/* 29906 */ {(11<<2)|2,{55,105,0}},
/* 29907 */ {(11<<2)|2,{55,105,0}},
/* 29908 */ {(11<<2)|2,{55,105,0}},
/* 29909 */ {(11<<2)|2,{55,105,0}},
/* 29910 */ {(11<<2)|2,{55,105,0}},
/* 29911 */ {(11<<2)|2,{55,105,0}},
/* 29912 */ {(11<<2)|2,{55,105,0}},
/* 29913 */ {(11<<2)|2,{55,105,0}},
/* 29914 */ {(11<<2)|2,{55,105,0}},
/* 29915 */ {(11<<2)|2,{55,105,0}},
/* 29916 */ {(11<<2)|2,{55,105,0}},
/* 29917 */ {(11<<2)|2,{55,105,0}},
/* 29918 */ {(11<<2)|2,{55,105,0}},
/* 29919 */ {(11<<2)|2,{55,105,0}},
/* 29920 */ {(16<<2)|3,{55,111,48}},
/* 29921 */ {(16<<2)|3,{55,111,49}},
/* 29922 */ {(16<<2)|3,{55,111,50}},
/* 29923 */ {(16<<2)|3,{55,111,97}},
/* 29924 */ {(16<<2)|3,{55,111,99}},
/* 29925 */ {(16<<2)|3,{55,111,101}},
/* 29926 */ {(16<<2)|3,{55,111,105}},
/* 29927 */ {(16<<2)|3,{55,111,111}},
/* 29928 */ {(16<<2)|3,{55,111,115}},
/* 29929 */ {(16<<2)|3,{55,111,116}},
/* 29930 */ {(11<<2)|2,{55,111,0}},
/* 29931 */ {(11<<2)|2,{55,111,0}},
/* 29932 */ {(11<<2)|2,{55,111,0}},
/* 29933 */ {(11<<2)|2,{55,111,0}},
/* 29934 */ {(11<<2)|2,{55,111,0}},
/* 29935 */ {(11<<2)|2,{55,111,0}},
/* 29936 */ {(11<<2)|2,{55,111,0}},
/* 29937 */ {(11<<2)|2,{55,111,0}},
/* 29938 */ {(11<<2)|2,{55,111,0}},
/* 29939 */ {(11<<2)|2,{55,111,0}},
/* 29940 */ {(11<<2)|2,{55,111,0}},
/* 29941 */ {(11<<2)|2,{55,111,0}},
/* 29942 */ {(11<<2)|2,{55,111,0}},
/* 29943 */ {(11<<2)|2,{55,111,0}},
/* 29944 */ {(11<<2)|2,{55,111,0}},
/* 29945 */ {(11<<2)|2,{55,111,0}},
/* 29946 */ {(11<<2)|2,{55,111,0}},
/* 29947 */ {(11<<2)|2,{55,111,0}},
/* 29948 */ {(11<<2)|2,{55,111,0}},
/* 29949 */ {(11<<2)|2,{55,111,0}},
/* 29950 */ {(11<<2)|2,{55,111,0}},
/* 29951 */ {(11<<2)|2,{55,111,0}},
/* 29952 */ {(16<<2)|3,{55,115,48}},
/* 29953 */ {(16<<2)|3,{55,115,49}},
/* 29954 */ {(16<<2)|3,{55,115,50}},
/* 29955 */ {(16<<2)|3,{55,115,97}},
/* 29956 */ {(16<<2)|3,{55,115,99}},
/* 29957 */ {(16<<2)|3,{55,115,101}},
/* 29958 */ {(16<<2)|3,{55,115,105}},
/* 29959 */ {(16<<2)|3,{55,115,111}},
/* 29960 */ {(16<<2)|3,{55,115,115}},
/* 29961 */ {(16<<2)|3,{55,115,116}},
/* 29962 */ {(11<<2)|2,{55,115,0}},
/* 29963 */ {(11<<2)|2,{55,115,0}},
/* 29964 */ {(11<<2)|2,{55,115,0}},
/* 29965 */ {(11<<2)|2,{55,115,0}},
/* 29966 */ {(11<<2)|2,{55,115,0}},
/* 29967 */ {(11<<2)|2,{55,115,0}},
/* 29968 */ {(11<<2)|2,{55,115,0}},
/* 29969 */ {(11<<2)|2,{55,115,0}},
/* 29970 */ {(11<<2)|2,{55,115,0}},
/* 29971 */ {(11<<2)|2,{55,115,0}},
/* 29972 */ {(11<<2)|2,{55,115,0}},
/* 29973 */ {(11<<2)|2,{55,115,0}},
/* 29974 */ {(11<<2)|2,{55,115,0}},
/* 29975 */ {(11<<2)|2,{55,115,0}},
/* 29976 */ {(11<<2)|2,{55,115,0}},
/* 29977 */ {(11<<2)|2,{55,115,0}},
/* 29978 */ {(11<<2)|2,{55,115,0}},
/* 29979 */ {(11<<2)|2,{55,115,0}},
/* 29980 */ {(11<<2)|2,{55,115,0}},
/* 29981 */ {(11<<2)|2,{55,115,0}},
/* 29982 */ {(11<<2)|2,{55,115,0}},
/* 29983 */ {(11<<2)|2,{55,115,0}},
/* 29984 */ {(16<<2)|3,{55,116,48}},
/* 29985 */ {(16<<2)|3,{55,116,49}},
/* 29986 */ {(16<<2)|3,{55,116,50}},
/* 29987 */ {(16<<2)|3,{55,116,97}},
/* 29988 */ {(16<<2)|3,{55,116,99}},
/* 29989 */ {(16<<2)|3,{55,116,101}},
/* 29990 */ {(16<<2)|3,{55,116,105}},
/* 29991 */ {(16<<2)|3,{55,116,111}},
/* 29992 */ {(16<<2)|3,{55,116,115}},
/* 29993 */ {(16<<2)|3,{55,116,116}},
/* 29994 */ {(11<<2)|2,{55,116,0}},
/* 29995 */ {(11<<2)|2,{55,116,0}},
/* 29996 */ {(11<<2)|2,{55,116,0}},
/* 29997 */ {(11<<2)|2,{55,116,0}},
/* 29998 */ {(11<<2)|2,{55,116,0}},
/* 29999 */ {(11<<2)|2,{55,116,0}},
/* 30000 */ {(11<<2)|2,{55,116,0}},
/* 30001 */ {(11<<2)|2,{55,116,0}},
/* 30002 */ {(11<<2)|2,{55,116,0}},
/* 30003 */ {(11<<2)|2,{55,116,0}},
/* 30004 */ {(11<<2)|2,{55,116,0}},
/* 30005 */ {(11<<2)|2,{55,116,0}},
/* 30006 */ {(11<<2)|2,{55,116,0}},
/* 30007 */ {(11<<2)|2,{55,116,0}},
/* 30008 */ {(11<<2)|2,{55,116,0}},
/* 30009 */ {(11<<2)|2,{55,116,0}},
/* 30010 */ {(11<<2)|2,{55,116,0}},
/* 30011 */ {(11<<2)|2,{55,116,0}},
/* 30012 */ {(11<<2)|2,{55,116,0}},
/* 30013 */ {(11<<2)|2,{55,116,0}},
/* 30014 */ {(11<<2)|2,{55,116,0}},
/* 30015 */ {(11<<2)|2,{55,116,0}},
/* 30016 */ {(12<<2)|2,{55,32,0}},
/* 30017 */ {(12<<2)|2,{55,32,0}},
/* 30018 */ {(12<<2)|2,{55,32,0}},
/* 30019 */ {(12<<2)|2,{55,32,0}},
/* 30020 */ {(12<<2)|2,{55,32,0}},
/* 30021 */ {(12<<2)|2,{55,32,0}},
/* 30022 */ {(12<<2)|2,{55,32,0}},
/* 30023 */ {(12<<2)|2,{55,32,0}},
/* 30024 */ {(12<<2)|2,{55,32,0}},
/* 30025 */ {(12<<2)|2,{55,32,0}},
/* 30026 */ {(12<<2)|2,{55,32,0}},
/* 30027 */ {(12<<2)|2,{55,32,0}},
/* 30028 */ {(12<<2)|2,{55,32,0}},
/* 30029 */ {(12<<2)|2,{55,32,0}},
/* 30030 */ {(12<<2)|2,{55,32,0}},
/* 30031 */ {(12<<2)|2,{55,32,0}},
/* 30032 */ {(12<<2)|2,{55,37,0}},
/* 30033 */ {(12<<2)|2,{55,37,0}},
/* 30034 */ {(12<<2)|2,{55,37,0}},
/* 30035 */ {(12<<2)|2,{55,37,0}},
/* 30036 */ {(12<<2)|2,{55,37,0}},
/* 30037 */ {(12<<2)|2,{55,37,0}},
/* 30038 */ {(12<<2)|2,{55,37,0}},
/* 30039 */ {(12<<2)|2,{55,37,0}},
/* 30040 */ {(12<<2)|2,{55,37,0}},
/* 30041 */ {(12<<2)|2,{55,37,0}},
/* 30042 */ {(12<<2)|2,{55,37,0}},
/* 30043 */ {(12<<2)|2,{55,37,0}},
/* 30044 */ {(12<<2)|2,{55,37,0}},
/* 30045 */ {(12<<2)|2,{55,37,0}},
/* 30046 */ {(12<<2)|2,{55,37,0}},
/* 30047 */ {(12<<2)|2,{55,37,0}},
/* 30048 */ {(12<<2)|2,{55,45,0}},
/* 30049 */ {(12<<2)|2,{55,45,0}},
/* 30050 */ {(12<<2)|2,{55,45,0}},
/* 30051 */ {(12<<2)|2,{55,45,0}},
/* 30052 */ {(12<<2)|2,{55,45,0}},
/* 30053 */ {(12<<2)|2,{55,45,0}},
/* 30054 */ {(12<<2)|2,{55,45,0}},
/* 30055 */ {(12<<2)|2,{55,45,0}},
/* 30056 */ {(12<<2)|2,{55,45,0}},
/* 30057 */ {(12<<2)|2,{55,45,0}},
/* 30058 */ {(12<<2)|2,{55,45,0}},
/* 30059 */ {(12<<2)|2,{55,45,0}},
/* 30060 */ {(12<<2)|2,{55,45,0}},
/* 30061 */ {(12<<2)|2,{55,45,0}},
/* 30062 */ {(12<<2)|2,{55,45,0}},
/* 30063 */ {(12<<2)|2,{55,45,0}},
/* 30064 */ {(12<<2)|2,{55,46,0}},
/* 30065 */ {(12<<2)|2,{55,46,0}},
/* 30066 */ {(12<<2)|2,{55,46,0}},
/* 30067 */ {(12<<2)|2,{55,46,0}},
/* 30068 */ {(12<<2)|2,{55,46,0}},
/* 30069 */ {(12<<2)|2,{55,46,0}},
/* 30070 */ {(12<<2)|2,{55,46,0}},
/* 30071 */ {(12<<2)|2,{55,46,0}},
/* 30072 */ {(12<<2)|2,{55,46,0}},
/* 30073 */ {(12<<2)|2,{55,46,0}},
/* 30074 */ {(12<<2)|2,{55,46,0}},
/* 30075 */ {(12<<2)|2,{55,46,0}},
/* 30076 */ {(12<<2)|2,{55,46,0}},
/* 30077 */ {(12<<2)|2,{55,46,0}},
/* 30078 */ {(12<<2)|2,{55,46,0}},
/* 30079 */ {(12<<2)|2,{55,46,0}},
/* 30080 */ {(12<<2)|2,{55,47,0}},
/* 30081 */ {(12<<2)|2,{55,47,0}},
/* 30082 */ {(12<<2)|2,{55,47,0}},
/* 30083 */ {(12<<2)|2,{55,47,0}},
/* 30084 */ {(12<<2)|2,{55,47,0}},
/* 30085 */ {(12<<2)|2,{55,47,0}},
/* 30086 */ {(12<<2)|2,{55,47,0}},
/* 30087 */ {(12<<2)|2,{55,47,0}},
/* 30088 */ {(12<<2)|2,{55,47,0}},
/* 30089 */ {(12<<2)|2,{55,47,0}},
/* 30090 */ {(12<<2)|2,{55,47,0}},
/* 30091 */ {(12<<2)|2,{55,47,0}},
/* 30092 */ {(12<<2)|2,{55,47,0}},
/* 30093 */ {(12<<2)|2,{55,47,0}},
/* 30094 */ {(12<<2)|2,{55,47,0}},
/* 30095 */ {(12<<2)|2,{55,47,0}},
/* 30096 */ {(12<<2)|2,{55,51,0}},
/* 30097 */ {(12<<2)|2,{55,51,0}},
/* 30098 */ {(12<<2)|2,{55,51,0}},
/* 30099 */ {(12<<2)|2,{55,51,0}},
/* 30100 */ {(12<<2)|2,{55,51,0}},
/* 30101 */ {(12<<2)|2,{55,51,0}},
/* 30102 */ {(12<<2)|2,{55,51,0}},
/* 30103 */ {(12<<2)|2,{55,51,0}},
/* 30104 */ {(12<<2)|2,{55,51,0}},
/* 30105 */ {(12<<2)|2,{55,51,0}},
/* 30106 */ {(12<<2)|2,{55,51,0}},
/* 30107 */ {(12<<2)|2,{55,51,0}},
/* 30108 */ {(12<<2)|2,{55,51,0}},
/* 30109 */ {(12<<2)|2,{55,51,0}},
/* 30110 */ {(12<<2)|2,{55,51,0}},
/* 30111 */ {(12<<2)|2,{55,51,0}},
/* 30112 */ {(12<<2)|2,{55,52,0}},
/* 30113 */ {(12<<2)|2,{55,52,0}},
/* 30114 */ {(12<<2)|2,{55,52,0}},
/* 30115 */ {(12<<2)|2,{55,52,0}},
/* 30116 */ {(12<<2)|2,{55,52,0}},
/* 30117 */ {(12<<2)|2,{55,52,0}},
/* 30118 */ {(12<<2)|2,{55,52,0}},
/* 30119 */ {(12<<2)|2,{55,52,0}},
/* 30120 */ {(12<<2)|2,{55,52,0}},
/* 30121 */ {(12<<2)|2,{55,52,0}},
/* 30122 */ {(12<<2)|2,{55,52,0}},
/* 30123 */ {(12<<2)|2,{55,52,0}},
/* 30124 */ {(12<<2)|2,{55,52,0}},
/* 30125 */ {(12<<2)|2,{55,52,0}},
/* 30126 */ {(12<<2)|2,{55,52,0}},
/* 30127 */ {(12<<2)|2,{55,52,0}},
/* 30128 */ {(12<<2)|2,{55,53,0}},
/* 30129 */ {(12<<2)|2,{55,53,0}},
/* 30130 */ {(12<<2)|2,{55,53,0}},
/* 30131 */ {(12<<2)|2,{55,53,0}},
/* 30132 */ {(12<<2)|2,{55,53,0}},
/* 30133 */ {(12<<2)|2,{55,53,0}},
/* 30134 */ {(12<<2)|2,{55,53,0}},
/* 30135 */ {(12<<2)|2,{55,53,0}},
/* 30136 */ {(12<<2)|2,{55,53,0}},
/* 30137 */ {(12<<2)|2,{55,53,0}},
/* 30138 */ {(12<<2)|2,{55,53,0}},
/* 30139 */ {(12<<2)|2,{55,53,0}},
/* 30140 */ {(12<<2)|2,{55,53,0}},
/* 30141 */ {(12<<2)|2,{55,53,0}},
/* 30142 */ {(12<<2)|2,{55,53,0}},
/* 30143 */ {(12<<2)|2,{55,53,0}},
/* 30144 */ {(12<<2)|2,{55,54,0}},
/* 30145 */ {(12<<2)|2,{55,54,0}},
/* 30146 */ {(12<<2)|2,{55,54,0}},
/* 30147 */ {(12<<2)|2,{55,54,0}},
/* 30148 */ {(12<<2)|2,{55,54,0}},
/* 30149 */ {(12<<2)|2,{55,54,0}},
/* 30150 */ {(12<<2)|2,{55,54,0}},
/* 30151 */ {(12<<2)|2,{55,54,0}},
/* 30152 */ {(12<<2)|2,{55,54,0}},
/* 30153 */ {(12<<2)|2,{55,54,0}},
/* 30154 */ {(12<<2)|2,{55,54,0}},
/* 30155 */ {(12<<2)|2,{55,54,0}},
/* 30156 */ {(12<<2)|2,{55,54,0}},
/* 30157 */ {(12<<2)|2,{55,54,0}},
/* 30158 */ {(12<<2)|2,{55,54,0}},
/* 30159 */ {(12<<2)|2,{55,54,0}},
/* 30160 */ {(12<<2)|2,{55,55,0}},
/* 30161 */ {(12<<2)|2,{55,55,0}},
/* 30162 */ {(12<<2)|2,{55,55,0}},
/* 30163 */ {(12<<2)|2,{55,55,0}},
/* 30164 */ {(12<<2)|2,{55,55,0}},
/* 30165 */ {(12<<2)|2,{55,55,0}},
/* 30166 */ {(12<<2)|2,{55,55,0}},
/* 30167 */ {(12<<2)|2,{55,55,0}},
/* 30168 */ {(12<<2)|2,{55,55,0}},
/* 30169 */ {(12<<2)|2,{55,55,0}},
/* 30170 */ {(12<<2)|2,{55,55,0}},
/* 30171 */ {(12<<2)|2,{55,55,0}},
/* 30172 */ {(12<<2)|2,{55,55,0}},
/* 30173 */ {(12<<2)|2,{55,55,0}},
/* 30174 */ {(12<<2)|2,{55,55,0}},
/* 30175 */ {(12<<2)|2,{55,55,0}},
/* 30176 */ {(12<<2)|2,{55,56,0}},
/* 30177 */ {(12<<2)|2,{55,56,0}},
/* 30178 */ {(12<<2)|2,{55,56,0}},
/* 30179 */ {(12<<2)|2,{55,56,0}},
/* 30180 */ {(12<<2)|2,{55,56,0}},
/* 30181 */ {(12<<2)|2,{55,56,0}},
/* 30182 */ {(12<<2)|2,{55,56,0}},
/* 30183 */ {(12<<2)|2,{55,56,0}},
/* 30184 */ {(12<<2)|2,{55,56,0}},
/* 30185 */ {(12<<2)|2,{55,56,0}},
/* 30186 */ {(12<<2)|2,{55,56,0}},
/* 30187 */ {(12<<2)|2,{55,56,0}},
/* 30188 */ {(12<<2)|2,{55,56,0}},
/* 30189 */ {(12<<2)|2,{55,56,0}},
/* 30190 */ {(12<<2)|2,{55,56,0}},
/* 30191 */ {(12<<2)|2,{55,56,0}},
/* 30192 */ {(12<<2)|2,{55,57,0}},
/* 30193 */ {(12<<2)|2,{55,57,0}},
/* 30194 */ {(12<<2)|2,{55,57,0}},
/* 30195 */ {(12<<2)|2,{55,57,0}},
/* 30196 */ {(12<<2)|2,{55,57,0}},
/* 30197 */ {(12<<2)|2,{55,57,0}},
/* 30198 */ {(12<<2)|2,{55,57,0}},
/* 30199 */ {(12<<2)|2,{55,57,0}},
/* 30200 */ {(12<<2)|2,{55,57,0}},
/* 30201 */ {(12<<2)|2,{55,57,0}},
/* 30202 */ {(12<<2)|2,{55,57,0}},
/* 30203 */ {(12<<2)|2,{55,57,0}},
/* 30204 */ {(12<<2)|2,{55,57,0}},
/* 30205 */ {(12<<2)|2,{55,57,0}},
/* 30206 */ {(12<<2)|2,{55,57,0}},
/* 30207 */ {(12<<2)|2,{55,57,0}},
/* 30208 */ {(12<<2)|2,{55,61,0}},
/* 30209 */ {(12<<2)|2,{55,61,0}},
/* 30210 */ {(12<<2)|2,{55,61,0}},
/* 30211 */ {(12<<2)|2,{55,61,0}},
/* 30212 */ {(12<<2)|2,{55,61,0}},
/* 30213 */ {(12<<2)|2,{55,61,0}},
/* 30214 */ {(12<<2)|2,{55,61,0}},
/* 30215 */ {(12<<2)|2,{55,61,0}},
/* 30216 */ {(12<<2)|2,{55,61,0}},
/* 30217 */ {(12<<2)|2,{55,61,0}},
/* 30218 */ {(12<<2)|2,{55,61,0}},
/* 30219 */ {(12<<2)|2,{55,61,0}},
/* 30220 */ {(12<<2)|2,{55,61,0}},
/* 30221 */ {(12<<2)|2,{55,61,0}},
/* 30222 */ {(12<<2)|2,{55,61,0}},
/* 30223 */ {(12<<2)|2,{55,61,0}},
/* 30224 */ {(12<<2)|2,{55,65,0}},
/* 30225 */ {(12<<2)|2,{55,65,0}},
/* 30226 */ {(12<<2)|2,{55,65,0}},
/* 30227 */ {(12<<2)|2,{55,65,0}},
/* 30228 */ {(12<<2)|2,{55,65,0}},
/* 30229 */ {(12<<2)|2,{55,65,0}},
/* 30230 */ {(12<<2)|2,{55,65,0}},
/* 30231 */ {(12<<2)|2,{55,65,0}},
/* 30232 */ {(12<<2)|2,{55,65,0}},
/* 30233 */ {(12<<2)|2,{55,65,0}},
/* 30234 */ {(12<<2)|2,{55,65,0}},
/* 30235 */ {(12<<2)|2,{55,65,0}},
/* 30236 */ {(12<<2)|2,{55,65,0}},
/* 30237 */ {(12<<2)|2,{55,65,0}},
/* 30238 */ {(12<<2)|2,{55,65,0}},
/* 30239 */ {(12<<2)|2,{55,65,0}},
/* 30240 */ {(12<<2)|2,{55,95,0}},
/* 30241 */ {(12<<2)|2,{55,95,0}},
/* 30242 */ {(12<<2)|2,{55,95,0}},
/* 30243 */ {(12<<2)|2,{55,95,0}},
/* 30244 */ {(12<<2)|2,{55,95,0}},
/* 30245 */ {(12<<2)|2,{55,95,0}},
/* 30246 */ {(12<<2)|2,{55,95,0}},
/* 30247 */ {(12<<2)|2,{55,95,0}},
/* 30248 */ {(12<<2)|2,{55,95,0}},
/* 30249 */ {(12<<2)|2,{55,95,0}},
/* 30250 */ {(12<<2)|2,{55,95,0}},
/* 30251 */ {(12<<2)|2,{55,95,0}},
/* 30252 */ {(12<<2)|2,{55,95,0}},
/* 30253 */ {(12<<2)|2,{55,95,0}},
/* 30254 */ {(12<<2)|2,{55,95,0}},
/* 30255 */ {(12<<2)|2,{55,95,0}},
/* 30256 */ {(12<<2)|2,{55,98,0}},
/* 30257 */ {(12<<2)|2,{55,98,0}},
/* 30258 */ {(12<<2)|2,{55,98,0}},
/* 30259 */ {(12<<2)|2,{55,98,0}},
/* 30260 */ {(12<<2)|2,{55,98,0}},
/* 30261 */ {(12<<2)|2,{55,98,0}},
/* 30262 */ {(12<<2)|2,{55,98,0}},
/* 30263 */ {(12<<2)|2,{55,98,0}},
/* 30264 */ {(12<<2)|2,{55,98,0}},
/* 30265 */ {(12<<2)|2,{55,98,0}},
/* 30266 */ {(12<<2)|2,{55,98,0}},
/* 30267 */ {(12<<2)|2,{55,98,0}},
/* 30268 */ {(12<<2)|2,{55,98,0}},
/* 30269 */ {(12<<2)|2,{55,98,0}},
/* 30270 */ {(12<<2)|2,{55,98,0}},
/* 30271 */ {(12<<2)|2,{55,98,0}},
/* 30272 */ {(12<<2)|2,{55,100,0}},
/* 30273 */ {(12<<2)|2,{55,100,0}},
/* 30274 */ {(12<<2)|2,{55,100,0}},
/* 30275 */ {(12<<2)|2,{55,100,0}},
/* 30276 */ {(12<<2)|2,{55,100,0}},
/* 30277 */ {(12<<2)|2,{55,100,0}},
/* 30278 */ {(12<<2)|2,{55,100,0}},
/* 30279 */ {(12<<2)|2,{55,100,0}},
/* 30280 */ {(12<<2)|2,{55,100,0}},
/* 30281 */ {(12<<2)|2,{55,100,0}},
/* 30282 */ {(12<<2)|2,{55,100,0}},
/* 30283 */ {(12<<2)|2,{55,100,0}},
/* 30284 */ {(12<<2)|2,{55,100,0}},
/* 30285 */ {(12<<2)|2,{55,100,0}},
/* 30286 */ {(12<<2)|2,{55,100,0}},
/* 30287 */ {(12<<2)|2,{55,100,0}},
/* 30288 */ {(12<<2)|2,{55,102,0}},
/* 30289 */ {(12<<2)|2,{55,102,0}},
/* 30290 */ {(12<<2)|2,{55,102,0}},
/* 30291 */ {(12<<2)|2,{55,102,0}},
/* 30292 */ {(12<<2)|2,{55,102,0}},
/* 30293 */ {(12<<2)|2,{55,102,0}},
/* 30294 */ {(12<<2)|2,{55,102,0}},
/* 30295 */ {(12<<2)|2,{55,102,0}},
/* 30296 */ {(12<<2)|2,{55,102,0}},
/* 30297 */ {(12<<2)|2,{55,102,0}},
/* 30298 */ {(12<<2)|2,{55,102,0}},
/* 30299 */ {(12<<2)|2,{55,102,0}},
/* 30300 */ {(12<<2)|2,{55,102,0}},
/* 30301 */ {(12<<2)|2,{55,102,0}},
/* 30302 */ {(12<<2)|2,{55,102,0}},
/* 30303 */ {(12<<2)|2,{55,102,0}},
/* 30304 */ {(12<<2)|2,{55,103,0}},
/* 30305 */ {(12<<2)|2,{55,103,0}},
/* 30306 */ {(12<<2)|2,{55,103,0}},
/* 30307 */ {(12<<2)|2,{55,103,0}},
/* 30308 */ {(12<<2)|2,{55,103,0}},
/* 30309 */ {(12<<2)|2,{55,103,0}},
/* 30310 */ {(12<<2)|2,{55,103,0}},
/* 30311 */ {(12<<2)|2,{55,103,0}},
/* 30312 */ {(12<<2)|2,{55,103,0}},
/* 30313 */ {(12<<2)|2,{55,103,0}},
/* 30314 */ {(12<<2)|2,{55,103,0}},
/* 30315 */ {(12<<2)|2,{55,103,0}},
/* 30316 */ {(12<<2)|2,{55,103,0}},
/* 30317 */ {(12<<2)|2,{55,103,0}},
/* 30318 */ {(12<<2)|2,{55,103,0}},
/* 30319 */ {(12<<2)|2,{55,103,0}},
/* 30320 */ {(12<<2)|2,{55,104,0}},
/* 30321 */ {(12<<2)|2,{55,104,0}},
/* 30322 */ {(12<<2)|2,{55,104,0}},
/* 30323 */ {(12<<2)|2,{55,104,0}},
/* 30324 */ {(12<<2)|2,{55,104,0}},
/* 30325 */ {(12<<2)|2,{55,104,0}},
/* 30326 */ {(12<<2)|2,{55,104,0}},
/* 30327 */ {(12<<2)|2,{55,104,0}},
/* 30328 */ {(12<<2)|2,{55,104,0}},
/* 30329 */ {(12<<2)|2,{55,104,0}},
/* 30330 */ {(12<<2)|2,{55,104,0}},
/* 30331 */ {(12<<2)|2,{55,104,0}},
/* 30332 */ {(12<<2)|2,{55,104,0}},
/* 30333 */ {(12<<2)|2,{55,104,0}},
/* 30334 */ {(12<<2)|2,{55,104,0}},
/* 30335 */ {(12<<2)|2,{55,104,0}},
/* 30336 */ {(12<<2)|2,{55,108,0}},
/* 30337 */ {(12<<2)|2,{55,108,0}},
/* 30338 */ {(12<<2)|2,{55,108,0}},
/* 30339 */ {(12<<2)|2,{55,108,0}},
/* 30340 */ {(12<<2)|2,{55,108,0}},
/* 30341 */ {(12<<2)|2,{55,108,0}},
/* 30342 */ {(12<<2)|2,{55,108,0}},
/* 30343 */ {(12<<2)|2,{55,108,0}},
/* 30344 */ {(12<<2)|2,{55,108,0}},
/* 30345 */ {(12<<2)|2,{55,108,0}},
/* 30346 */ {(12<<2)|2,{55,108,0}},
/* 30347 */ {(12<<2)|2,{55,108,0}},
/* 30348 */ {(12<<2)|2,{55,108,0}},
/* 30349 */ {(12<<2)|2,{55,108,0}},
/* 30350 */ {(12<<2)|2,{55,108,0}},
/* 30351 */ {(12<<2)|2,{55,108,0}},
/* 30352 */ {(12<<2)|2,{55,109,0}},
/* 30353 */ {(12<<2)|2,{55,109,0}},
/* 30354 */ {(12<<2)|2,{55,109,0}},
/* 30355 */ {(12<<2)|2,{55,109,0}},
/* 30356 */ {(12<<2)|2,{55,109,0}},
/* 30357 */ {(12<<2)|2,{55,109,0}},
/* 30358 */ {(12<<2)|2,{55,109,0}},
/* 30359 */ {(12<<2)|2,{55,109,0}},
/* 30360 */ {(12<<2)|2,{55,109,0}},
/* 30361 */ {(12<<2)|2,{55,109,0}},
/* 30362 */ {(12<<2)|2,{55,109,0}},
/* 30363 */ {(12<<2)|2,{55,109,0}},
/* 30364 */ {(12<<2)|2,{55,109,0}},
/* 30365 */ {(12<<2)|2,{55,109,0}},
/* 30366 */ {(12<<2)|2,{55,109,0}},
/* 30367 */ {(12<<2)|2,{55,109,0}},
/* 30368 */ {(12<<2)|2,{55,110,0}},
/* 30369 */ {(12<<2)|2,{55,110,0}},
/* 30370 */ {(12<<2)|2,{55,110,0}},
/* 30371 */ {(12<<2)|2,{55,110,0}},
/* 30372 */ {(12<<2)|2,{55,110,0}},
/* 30373 */ {(12<<2)|2,{55,110,0}},
/* 30374 */ {(12<<2)|2,{55,110,0}},
/* 30375 */ {(12<<2)|2,{55,110,0}},
/* 30376 */ {(12<<2)|2,{55,110,0}},
/* 30377 */ {(12<<2)|2,{55,110,0}},
/* 30378 */ {(12<<2)|2,{55,110,0}},
/* 30379 */ {(12<<2)|2,{55,110,0}},
/* 30380 */ {(12<<2)|2,{55,110,0}},
/* 30381 */ {(12<<2)|2,{55,110,0}},
/* 30382 */ {(12<<2)|2,{55,110,0}},
/* 30383 */ {(12<<2)|2,{55,110,0}},
/* 30384 */ {(12<<2)|2,{55,112,0}},
/* 30385 */ {(12<<2)|2,{55,112,0}},
/* 30386 */ {(12<<2)|2,{55,112,0}},
/* 30387 */ {(12<<2)|2,{55,112,0}},
/* 30388 */ {(12<<2)|2,{55,112,0}},
/* 30389 */ {(12<<2)|2,{55,112,0}},
/* 30390 */ {(12<<2)|2,{55,112,0}},
/* 30391 */ {(12<<2)|2,{55,112,0}},
/* 30392 */ {(12<<2)|2,{55,112,0}},
/* 30393 */ {(12<<2)|2,{55,112,0}},
/* 30394 */ {(12<<2)|2,{55,112,0}},
/* 30395 */ {(12<<2)|2,{55,112,0}},
/* 30396 */ {(12<<2)|2,{55,112,0}},
/* 30397 */ {(12<<2)|2,{55,112,0}},
/* 30398 */ {(12<<2)|2,{55,112,0}},
/* 30399 */ {(12<<2)|2,{55,112,0}},
/* 30400 */ {(12<<2)|2,{55,114,0}},
/* 30401 */ {(12<<2)|2,{55,114,0}},
/* 30402 */ {(12<<2)|2,{55,114,0}},
/* 30403 */ {(12<<2)|2,{55,114,0}},
/* 30404 */ {(12<<2)|2,{55,114,0}},
/* 30405 */ {(12<<2)|2,{55,114,0}},
/* 30406 */ {(12<<2)|2,{55,114,0}},
/* 30407 */ {(12<<2)|2,{55,114,0}},
/* 30408 */ {(12<<2)|2,{55,114,0}},
/* 30409 */ {(12<<2)|2,{55,114,0}},
/* 30410 */ {(12<<2)|2,{55,114,0}},
/* 30411 */ {(12<<2)|2,{55,114,0}},
/* 30412 */ {(12<<2)|2,{55,114,0}},
/* 30413 */ {(12<<2)|2,{55,114,0}},
/* 30414 */ {(12<<2)|2,{55,114,0}},
/* 30415 */ {(12<<2)|2,{55,114,0}},
/* 30416 */ {(12<<2)|2,{55,117,0}},
/* 30417 */ {(12<<2)|2,{55,117,0}},
/* 30418 */ {(12<<2)|2,{55,117,0}},
/* 30419 */ {(12<<2)|2,{55,117,0}},
/* 30420 */ {(12<<2)|2,{55,117,0}},
/* 30421 */ {(12<<2)|2,{55,117,0}},
/* 30422 */ {(12<<2)|2,{55,117,0}},
/* 30423 */ {(12<<2)|2,{55,117,0}},
/* 30424 */ {(12<<2)|2,{55,117,0}},
/* 30425 */ {(12<<2)|2,{55,117,0}},
/* 30426 */ {(12<<2)|2,{55,117,0}},
/* 30427 */ {(12<<2)|2,{55,117,0}},
/* 30428 */ {(12<<2)|2,{55,117,0}},
/* 30429 */ {(12<<2)|2,{55,117,0}},
/* 30430 */ {(12<<2)|2,{55,117,0}},
/* 30431 */ {(12<<2)|2,{55,117,0}},
/* 30432 */ {(13<<2)|2,{55,58,0}},
/* 30433 */ {(13<<2)|2,{55,58,0}},
/* 30434 */ {(13<<2)|2,{55,58,0}},
/* 30435 */ {(13<<2)|2,{55,58,0}},
/* 30436 */ {(13<<2)|2,{55,58,0}},
/* 30437 */ {(13<<2)|2,{55,58,0}},
/* 30438 */ {(13<<2)|2,{55,58,0}},
/* 30439 */ {(13<<2)|2,{55,58,0}},
/* 30440 */ {(13<<2)|2,{55,66,0}},
/* 30441 */ {(13<<2)|2,{55,66,0}},
/* 30442 */ {(13<<2)|2,{55,66,0}},
/* 30443 */ {(13<<2)|2,{55,66,0}},
/* 30444 */ {(13<<2)|2,{55,66,0}},
/* 30445 */ {(13<<2)|2,{55,66,0}},
/* 30446 */ {(13<<2)|2,{55,66,0}},
/* 30447 */ {(13<<2)|2,{55,66,0}},
/* 30448 */ {(13<<2)|2,{55,67,0}},
/* 30449 */ {(13<<2)|2,{55,67,0}},
/* 30450 */ {(13<<2)|2,{55,67,0}},
/* 30451 */ {(13<<2)|2,{55,67,0}},
/* 30452 */ {(13<<2)|2,{55,67,0}},
/* 30453 */ {(13<<2)|2,{55,67,0}},
/* 30454 */ {(13<<2)|2,{55,67,0}},
/* 30455 */ {(13<<2)|2,{55,67,0}},
/* 30456 */ {(13<<2)|2,{55,68,0}},
/* 30457 */ {(13<<2)|2,{55,68,0}},
/* 30458 */ {(13<<2)|2,{55,68,0}},
/* 30459 */ {(13<<2)|2,{55,68,0}},
/* 30460 */ {(13<<2)|2,{55,68,0}},
/* 30461 */ {(13<<2)|2,{55,68,0}},
/* 30462 */ {(13<<2)|2,{55,68,0}},
/* 30463 */ {(13<<2)|2,{55,68,0}},
/* 30464 */ {(13<<2)|2,{55,69,0}},
/* 30465 */ {(13<<2)|2,{55,69,0}},
/* 30466 */ {(13<<2)|2,{55,69,0}},
/* 30467 */ {(13<<2)|2,{55,69,0}},
/* 30468 */ {(13<<2)|2,{55,69,0}},
/* 30469 */ {(13<<2)|2,{55,69,0}},
/* 30470 */ {(13<<2)|2,{55,69,0}},
/* 30471 */ {(13<<2)|2,{55,69,0}},
/* 30472 */ {(13<<2)|2,{55,70,0}},
/* 30473 */ {(13<<2)|2,{55,70,0}},
/* 30474 */ {(13<<2)|2,{55,70,0}},
/* 30475 */ {(13<<2)|2,{55,70,0}},
/* 30476 */ {(13<<2)|2,{55,70,0}},
/* 30477 */ {(13<<2)|2,{55,70,0}},
/* 30478 */ {(13<<2)|2,{55,70,0}},
/* 30479 */ {(13<<2)|2,{55,70,0}},
/* 30480 */ {(13<<2)|2,{55,71,0}},
/* 30481 */ {(13<<2)|2,{55,71,0}},
/* 30482 */ {(13<<2)|2,{55,71,0}},
/* 30483 */ {(13<<2)|2,{55,71,0}},
/* 30484 */ {(13<<2)|2,{55,71,0}},
/* 30485 */ {(13<<2)|2,{55,71,0}},
/* 30486 */ {(13<<2)|2,{55,71,0}},
/* 30487 */ {(13<<2)|2,{55,71,0}},
/* 30488 */ {(13<<2)|2,{55,72,0}},
/* 30489 */ {(13<<2)|2,{55,72,0}},
/* 30490 */ {(13<<2)|2,{55,72,0}},
/* 30491 */ {(13<<2)|2,{55,72,0}},
/* 30492 */ {(13<<2)|2,{55,72,0}},
/* 30493 */ {(13<<2)|2,{55,72,0}},
/* 30494 */ {(13<<2)|2,{55,72,0}},
/* 30495 */ {(13<<2)|2,{55,72,0}},
/* 30496 */ {(13<<2)|2,{55,73,0}},
/* 30497 */ {(13<<2)|2,{55,73,0}},
/* 30498 */ {(13<<2)|2,{55,73,0}},
/* 30499 */ {(13<<2)|2,{55,73,0}},
/* 30500 */ {(13<<2)|2,{55,73,0}},
/* 30501 */ {(13<<2)|2,{55,73,0}},
/* 30502 */ {(13<<2)|2,{55,73,0}},
/* 30503 */ {(13<<2)|2,{55,73,0}},
/* 30504 */ {(13<<2)|2,{55,74,0}},
/* 30505 */ {(13<<2)|2,{55,74,0}},
/* 30506 */ {(13<<2)|2,{55,74,0}},
/* 30507 */ {(13<<2)|2,{55,74,0}},
/* 30508 */ {(13<<2)|2,{55,74,0}},
/* 30509 */ {(13<<2)|2,{55,74,0}},
/* 30510 */ {(13<<2)|2,{55,74,0}},
/* 30511 */ {(13<<2)|2,{55,74,0}},
/* 30512 */ {(13<<2)|2,{55,75,0}},
/* 30513 */ {(13<<2)|2,{55,75,0}},
/* 30514 */ {(13<<2)|2,{55,75,0}},
/* 30515 */ {(13<<2)|2,{55,75,0}},
/* 30516 */ {(13<<2)|2,{55,75,0}},
/* 30517 */ {(13<<2)|2,{55,75,0}},
/* 30518 */ {(13<<2)|2,{55,75,0}},
/* 30519 */ {(13<<2)|2,{55,75,0}},
/* 30520 */ {(13<<2)|2,{55,76,0}},
/* 30521 */ {(13<<2)|2,{55,76,0}},
/* 30522 */ {(13<<2)|2,{55,76,0}},
/* 30523 */ {(13<<2)|2,{55,76,0}},
/* 30524 */ {(13<<2)|2,{55,76,0}},
/* 30525 */ {(13<<2)|2,{55,76,0}},
/* 30526 */ {(13<<2)|2,{55,76,0}},
/* 30527 */ {(13<<2)|2,{55,76,0}},
/* 30528 */ {(13<<2)|2,{55,77,0}},
/* 30529 */ {(13<<2)|2,{55,77,0}},
/* 30530 */ {(13<<2)|2,{55,77,0}},
/* 30531 */ {(13<<2)|2,{55,77,0}},
/* 30532 */ {(13<<2)|2,{55,77,0}},
/* 30533 */ {(13<<2)|2,{55,77,0}},
/* 30534 */ {(13<<2)|2,{55,77,0}},
/* 30535 */ {(13<<2)|2,{55,77,0}},
/* 30536 */ {(13<<2)|2,{55,78,0}},
/* 30537 */ {(13<<2)|2,{55,78,0}},
/* 30538 */ {(13<<2)|2,{55,78,0}},
/* 30539 */ {(13<<2)|2,{55,78,0}},
/* 30540 */ {(13<<2)|2,{55,78,0}},
/* 30541 */ {(13<<2)|2,{55,78,0}},
/* 30542 */ {(13<<2)|2,{55,78,0}},
/* 30543 */ {(13<<2)|2,{55,78,0}},
/* 30544 */ {(13<<2)|2,{55,79,0}},
/* 30545 */ {(13<<2)|2,{55,79,0}},
/* 30546 */ {(13<<2)|2,{55,79,0}},
/* 30547 */ {(13<<2)|2,{55,79,0}},
/* 30548 */ {(13<<2)|2,{55,79,0}},
/* 30549 */ {(13<<2)|2,{55,79,0}},
/* 30550 */ {(13<<2)|2,{55,79,0}},
/* 30551 */ {(13<<2)|2,{55,79,0}},
/* 30552 */ {(13<<2)|2,{55,80,0}},
/* 30553 */ {(13<<2)|2,{55,80,0}},
/* 30554 */ {(13<<2)|2,{55,80,0}},
/* 30555 */ {(13<<2)|2,{55,80,0}},
/* 30556 */ {(13<<2)|2,{55,80,0}},
/* 30557 */ {(13<<2)|2,{55,80,0}},
/* 30558 */ {(13<<2)|2,{55,80,0}},
/* 30559 */ {(13<<2)|2,{55,80,0}},
/* 30560 */ {(13<<2)|2,{55,81,0}},
/* 30561 */ {(13<<2)|2,{55,81,0}},
/* 30562 */ {(13<<2)|2,{55,81,0}},
/* 30563 */ {(13<<2)|2,{55,81,0}},
/* 30564 */ {(13<<2)|2,{55,81,0}},
/* 30565 */ {(13<<2)|2,{55,81,0}},
/* 30566 */ {(13<<2)|2,{55,81,0}},
/* 30567 */ {(13<<2)|2,{55,81,0}},
/* 30568 */ {(13<<2)|2,{55,82,0}},
/* 30569 */ {(13<<2)|2,{55,82,0}},
/* 30570 */ {(13<<2)|2,{55,82,0}},
/* 30571 */ {(13<<2)|2,{55,82,0}},
/* 30572 */ {(13<<2)|2,{55,82,0}},
/* 30573 */ {(13<<2)|2,{55,82,0}},
/* 30574 */ {(13<<2)|2,{55,82,0}},
/* 30575 */ {(13<<2)|2,{55,82,0}},
/* 30576 */ {(13<<2)|2,{55,83,0}},
/* 30577 */ {(13<<2)|2,{55,83,0}},
/* 30578 */ {(13<<2)|2,{55,83,0}},
/* 30579 */ {(13<<2)|2,{55,83,0}},
/* 30580 */ {(13<<2)|2,{55,83,0}},
/* 30581 */ {(13<<2)|2,{55,83,0}},
/* 30582 */ {(13<<2)|2,{55,83,0}},
/* 30583 */ {(13<<2)|2,{55,83,0}},
/* 30584 */ {(13<<2)|2,{55,84,0}},
/* 30585 */ {(13<<2)|2,{55,84,0}},
/* 30586 */ {(13<<2)|2,{55,84,0}},
/* 30587 */ {(13<<2)|2,{55,84,0}},
/* 30588 */ {(13<<2)|2,{55,84,0}},
/* 30589 */ {(13<<2)|2,{55,84,0}},
/* 30590 */ {(13<<2)|2,{55,84,0}},
/* 30591 */ {(13<<2)|2,{55,84,0}},
/* 30592 */ {(13<<2)|2,{55,85,0}},
/* 30593 */ {(13<<2)|2,{55,85,0}},
/* 30594 */ {(13<<2)|2,{55,85,0}},
/* 30595 */ {(13<<2)|2,{55,85,0}},
/* 30596 */ {(13<<2)|2,{55,85,0}},
/* 30597 */ {(13<<2)|2,{55,85,0}},
/* 30598 */ {(13<<2)|2,{55,85,0}},
/* 30599 */ {(13<<2)|2,{55,85,0}},
/* 30600 */ {(13<<2)|2,{55,86,0}},
/* 30601 */ {(13<<2)|2,{55,86,0}},
/* 30602 */ {(13<<2)|2,{55,86,0}},
/* 30603 */ {(13<<2)|2,{55,86,0}},
/* 30604 */ {(13<<2)|2,{55,86,0}},
/* 30605 */ {(13<<2)|2,{55,86,0}},
/* 30606 */ {(13<<2)|2,{55,86,0}},
/* 30607 */ {(13<<2)|2,{55,86,0}},
/* 30608 */ {(13<<2)|2,{55,87,0}},
/* 30609 */ {(13<<2)|2,{55,87,0}},
/* 30610 */ {(13<<2)|2,{55,87,0}},
/* 30611 */ {(13<<2)|2,{55,87,0}},
/* 30612 */ {(13<<2)|2,{55,87,0}},
/* 30613 */ {(13<<2)|2,{55,87,0}},
/* 30614 */ {(13<<2)|2,{55,87,0}},
/* 30615 */ {(13<<2)|2,{55,87,0}},
/* 30616 */ {(13<<2)|2,{55,89,0}},
/* 30617 */ {(13<<2)|2,{55,89,0}},
/* 30618 */ {(13<<2)|2,{55,89,0}},
/* 30619 */ {(13<<2)|2,{55,89,0}},
/* 30620 */ {(13<<2)|2,{55,89,0}},
/* 30621 */ {(13<<2)|2,{55,89,0}},
/* 30622 */ {(13<<2)|2,{55,89,0}},
/* 30623 */ {(13<<2)|2,{55,89,0}},
/* 30624 */ {(13<<2)|2,{55,106,0}},
/* 30625 */ {(13<<2)|2,{55,106,0}},
/* 30626 */ {(13<<2)|2,{55,106,0}},
/* 30627 */ {(13<<2)|2,{55,106,0}},
/* 30628 */ {(13<<2)|2,{55,106,0}},
/* 30629 */ {(13<<2)|2,{55,106,0}},
/* 30630 */ {(13<<2)|2,{55,106,0}},
/* 30631 */ {(13<<2)|2,{55,106,0}},
/* 30632 */ {(13<<2)|2,{55,107,0}},
/* 30633 */ {(13<<2)|2,{55,107,0}},
/* 30634 */ {(13<<2)|2,{55,107,0}},
/* 30635 */ {(13<<2)|2,{55,107,0}},
/* 30636 */ {(13<<2)|2,{55,107,0}},
/* 30637 */ {(13<<2)|2,{55,107,0}},
/* 30638 */ {(13<<2)|2,{55,107,0}},
/* 30639 */ {(13<<2)|2,{55,107,0}},
/* 30640 */ {(13<<2)|2,{55,113,0}},
/* 30641 */ {(13<<2)|2,{55,113,0}},
/* 30642 */ {(13<<2)|2,{55,113,0}},
/* 30643 */ {(13<<2)|2,{55,113,0}},
/* 30644 */ {(13<<2)|2,{55,113,0}},
/* 30645 */ {(13<<2)|2,{55,113,0}},
/* 30646 */ {(13<<2)|2,{55,113,0}},
/* 30647 */ {(13<<2)|2,{55,113,0}},
/* 30648 */ {(13<<2)|2,{55,118,0}},
/* 30649 */ {(13<<2)|2,{55,118,0}},
/* 30650 */ {(13<<2)|2,{55,118,0}},
/* 30651 */ {(13<<2)|2,{55,118,0}},
/* 30652 */ {(13<<2)|2,{55,118,0}},
/* 30653 */ {(13<<2)|2,{55,118,0}},
/* 30654 */ {(13<<2)|2,{55,118,0}},
/* 30655 */ {(13<<2)|2,{55,118,0}},
/* 30656 */ {(13<<2)|2,{55,119,0}},
/* 30657 */ {(13<<2)|2,{55,119,0}},
/* 30658 */ {(13<<2)|2,{55,119,0}},
/* 30659 */ {(13<<2)|2,{55,119,0}},
/* 30660 */ {(13<<2)|2,{55,119,0}},
/* 30661 */ {(13<<2)|2,{55,119,0}},
/* 30662 */ {(13<<2)|2,{55,119,0}},
/* 30663 */ {(13<<2)|2,{55,119,0}},
/* 30664 */ {(13<<2)|2,{55,120,0}},
/* 30665 */ {(13<<2)|2,{55,120,0}},
/* 30666 */ {(13<<2)|2,{55,120,0}},
/* 30667 */ {(13<<2)|2,{55,120,0}},
/* 30668 */ {(13<<2)|2,{55,120,0}},
/* 30669 */ {(13<<2)|2,{55,120,0}},
/* 30670 */ {(13<<2)|2,{55,120,0}},
/* 30671 */ {(13<<2)|2,{55,120,0}},
/* 30672 */ {(13<<2)|2,{55,121,0}},
/* 30673 */ {(13<<2)|2,{55,121,0}},
/* 30674 */ {(13<<2)|2,{55,121,0}},
/* 30675 */ {(13<<2)|2,{55,121,0}},
/* 30676 */ {(13<<2)|2,{55,121,0}},
/* 30677 */ {(13<<2)|2,{55,121,0}},
/* 30678 */ {(13<<2)|2,{55,121,0}},
/* 30679 */ {(13<<2)|2,{55,121,0}},
/* 30680 */ {(13<<2)|2,{55,122,0}},
/* 30681 */ {(13<<2)|2,{55,122,0}},
/* 30682 */ {(13<<2)|2,{55,122,0}},
/* 30683 */ {(13<<2)|2,{55,122,0}},
/* 30684 */ {(13<<2)|2,{55,122,0}},
/* 30685 */ {(13<<2)|2,{55,122,0}},
/* 30686 */ {(13<<2)|2,{55,122,0}},
/* 30687 */ {(13<<2)|2,{55,122,0}},
/* 30688 */ {(14<<2)|2,{55,38,0}},
/* 30689 */ {(14<<2)|2,{55,38,0}},
/* 30690 */ {(14<<2)|2,{55,38,0}},
/* 30691 */ {(14<<2)|2,{55,38,0}},
/* 30692 */ {(14<<2)|2,{55,42,0}},
/* 30693 */ {(14<<2)|2,{55,42,0}},
/* 30694 */ {(14<<2)|2,{55,42,0}},
/* 30695 */ {(14<<2)|2,{55,42,0}},
/* 30696 */ {(14<<2)|2,{55,44,0}},
/* 30697 */ {(14<<2)|2,{55,44,0}},
/* 30698 */ {(14<<2)|2,{55,44,0}},
/* 30699 */ {(14<<2)|2,{55,44,0}},
/* 30700 */ {(14<<2)|2,{55,59,0}},
/* 30701 */ {(14<<2)|2,{55,59,0}},
/* 30702 */ {(14<<2)|2,{55,59,0}},
/* 30703 */ {(14<<2)|2,{55,59,0}},
/* 30704 */ {(14<<2)|2,{55,88,0}},
/* 30705 */ {(14<<2)|2,{55,88,0}},
/* 30706 */ {(14<<2)|2,{55,88,0}},
/* 30707 */ {(14<<2)|2,{55,88,0}},
/* 30708 */ {(14<<2)|2,{55,90,0}},
/* 30709 */ {(14<<2)|2,{55,90,0}},
/* 30710 */ {(14<<2)|2,{55,90,0}},
/* 30711 */ {(14<<2)|2,{55,90,0}},
/* 30712 */ {(16<<2)|2,{55,33,0}},
/* 30713 */ {(16<<2)|2,{55,34,0}},
/* 30714 */ {(16<<2)|2,{55,40,0}},
/* 30715 */ {(16<<2)|2,{55,41,0}},
/* 30716 */ {(16<<2)|2,{55,63,0}},
/* 30717 */ {(6<<2)|1,{55,0,0}},
/* 30718 */ {(6<<2)|1,{55,0,0}},
/* 30719 */ {(6<<2)|1,{55,0,0}},
/* 30720 */ {(16<<2)|3,{56,48,48}},
/* 30721 */ {(16<<2)|3,{56,48,49}},
/* 30722 */ {(16<<2)|3,{56,48,50}},
/* 30723 */ {(16<<2)|3,{56,48,97}},
/* 30724 */ {(16<<2)|3,{56,48,99}},
/* 30725 */ {(16<<2)|3,{56,48,101}},
/* 30726 */ {(16<<2)|3,{56,48,105}},
/* 30727 */ {(16<<2)|3,{56,48,111}},
/* 30728 */ {(16<<2)|3,{56,48,115}},
/* 30729 */ {(16<<2)|3,{56,48,116}},
/* 30730 */ {(11<<2)|2,{56,48,0}},
/* 30731 */ {(11<<2)|2,{56,48,0}},
/* 30732 */ {(11<<2)|2,{56,48,0}},
/* 30733 */ {(11<<2)|2,{56,48,0}},
/* 30734 */ {(11<<2)|2,{56,48,0}},
/* 30735 */ {(11<<2)|2,{56,48,0}},
/* 30736 */ {(11<<2)|2,{56,48,0}},
/* 30737 */ {(11<<2)|2,{56,48,0}},
/* 30738 */ {(11<<2)|2,{56,48,0}},
/* 30739 */ {(11<<2)|2,{56,48,0}},
/* 30740 */ {(11<<2)|2,{56,48,0}},
/* 30741 */ {(11<<2)|2,{56,48,0}},
/* 30742 */ {(11<<2)|2,{56,48,0}},
/* 30743 */ {(11<<2)|2,{56,48,0}},
/* 30744 */ {(11<<2)|2,{56,48,0}},
/* 30745 */ {(11<<2)|2,{56,48,0}},
/* 30746 */ {(11<<2)|2,{56,48,0}},
/* 30747 */ {(11<<2)|2,{56,48,0}},
/* 30748 */ {(11<<2)|2,{56,48,0}},
/* 30749 */ {(11<<2)|2,{56,48,0}},
/* 30750 */ {(11<<2)|2,{56,48,0}},
/* 30751 */ {(11<<2)|2,{56,48,0}},
/* 30752 */ {(16<<2)|3,{56,49,48}},
/* 30753 */ {(16<<2)|3,{56,49,49}},
/* 30754 */ {(16<<2)|3,{56,49,50}},
/* 30755 */ {(16<<2)|3,{56,49,97}},
/* 30756 */ {(16<<2)|3,{56,49,99}},
/* 30757 */ {(16<<2)|3,{56,49,101}},
/* 30758 */ {(16<<2)|3,{56,49,105}},
/* 30759 */ {(16<<2)|3,{56,49,111}},
/* 30760 */ {(16<<2)|3,{56,49,115}},
/* 30761 */ {(16<<2)|3,{56,49,116}},
/* 30762 */ {(11<<2)|2,{56,49,0}},
/* 30763 */ {(11<<2)|2,{56,49,0}},
/* 30764 */ {(11<<2)|2,{56,49,0}},
/* 30765 */ {(11<<2)|2,{56,49,0}},
/* 30766 */ {(11<<2)|2,{56,49,0}},
/* 30767 */ {(11<<2)|2,{56,49,0}},
/* 30768 */ {(11<<2)|2,{56,49,0}},
/* 30769 */ {(11<<2)|2,{56,49,0}},
/* 30770 */ {(11<<2)|2,{56,49,0}},
/* 30771 */ {(11<<2)|2,{56,49,0}},
/* 30772 */ {(11<<2)|2,{56,49,0}},
/* 30773 */ {(11<<2)|2,{56,49,0}},
/* 30774 */ {(11<<2)|2,{56,49,0}},
/* 30775 */ {(11<<2)|2,{56,49,0}},
/* 30776 */ {(11<<2)|2,{56,49,0}},
/* 30777 */ {(11<<2)|2,{56,49,0}},
/* 30778 */ {(11<<2)|2,{56,49,0}},
/* 30779 */ {(11<<2)|2,{56,49,0}},
/* 30780 */ {(11<<2)|2,{56,49,0}},
/* 30781 */ {(11<<2)|2,{56,49,0}},
/* 30782 */ {(11<<2)|2,{56,49,0}},
/* 30783 */ {(11<<2)|2,{56,49,0}},
/* 30784 */ {(16<<2)|3,{56,50,48}},
/* 30785 */ {(16<<2)|3,{56,50,49}},
/* 30786 */ {(16<<2)|3,{56,50,50}},
/* 30787 */ {(16<<2)|3,{56,50,97}},
/* 30788 */ {(16<<2)|3,{56,50,99}},
/* 30789 */ {(16<<2)|3,{56,50,101}},
/* 30790 */ {(16<<2)|3,{56,50,105}},
/* 30791 */ {(16<<2)|3,{56,50,111}},
/* 30792 */ {(16<<2)|3,{56,50,115}},
/* 30793 */ {(16<<2)|3,{56,50,116}},
/* 30794 */ {(11<<2)|2,{56,50,0}},
/* 30795 */ {(11<<2)|2,{56,50,0}},
/* 30796 */ {(11<<2)|2,{56,50,0}},
/* 30797 */ {(11<<2)|2,{56,50,0}},
/* 30798 */ {(11<<2)|2,{56,50,0}},
/* 30799 */ {(11<<2)|2,{56,50,0}},
/* 30800 */ {(11<<2)|2,{56,50,0}},
/* 30801 */ {(11<<2)|2,{56,50,0}},
/* 30802 */ {(11<<2)|2,{56,50,0}},
/* 30803 */ {(11<<2)|2,{56,50,0}},
/* 30804 */ {(11<<2)|2,{56,50,0}},
/* 30805 */ {(11<<2)|2,{56,50,0}},
/* 30806 */ {(11<<2)|2,{56,50,0}},
/* 30807 */ {(11<<2)|2,{56,50,0}},
/* 30808 */ {(11<<2)|2,{56,50,0}},
/* 30809 */ {(11<<2)|2,{56,50,0}},
/* 30810 */ {(11<<2)|2,{56,50,0}},
/* 30811 */ {(11<<2)|2,{56,50,0}},
/* 30812 */ {(11<<2)|2,{56,50,0}},
/* 30813 */ {(11<<2)|2,{56,50,0}},
/* 30814 */ {(11<<2)|2,{56,50,0}},
/* 30815 */ {(11<<2)|2,{56,50,0}},
/* 30816 */ {(16<<2)|3,{56,97,48}},
/* 30817 */ {(16<<2)|3,{56,97,49}},
/* 30818 */ {(16<<2)|3,{56,97,50}},
/* 30819 */ {(16<<2)|3,{56,97,97}},
/* 30820 */ {(16<<2)|3,{56,97,99}},
/* 30821 */ {(16<<2)|3,{56,97,101}},
/* 30822 */ {(16<<2)|3,{56,97,105}},
/* 30823 */ {(16<<2)|3,{56,97,111}},
/* 30824 */ {(16<<2)|3,{56,97,115}},
/* 30825 */ {(16<<2)|3,{56,97,116}},
/* 30826 */ {(11<<2)|2,{56,97,0}},
/* 30827 */ {(11<<2)|2,{56,97,0}},
/* 30828 */ {(11<<2)|2,{56,97,0}},
/* 30829 */ {(11<<2)|2,{56,97,0}},
/* 30830 */ {(11<<2)|2,{56,97,0}},
/* 30831 */ {(11<<2)|2,{56,97,0}},
/* 30832 */ {(11<<2)|2,{56,97,0}},
/* 30833 */ {(11<<2)|2,{56,97,0}},
/* 30834 */ {(11<<2)|2,{56,97,0}},
/* 30835 */ {(11<<2)|2,{56,97,0}},
/* 30836 */ {(11<<2)|2,{56,97,0}},
/* 30837 */ {(11<<2)|2,{56,97,0}},
/* 30838 */ {(11<<2)|2,{56,97,0}},
/* 30839 */ {(11<<2)|2,{56,97,0}},
/* 30840 */ {(11<<2)|2,{56,97,0}},
/* 30841 */ {(11<<2)|2,{56,97,0}},
/* 30842 */ {(11<<2)|2,{56,97,0}},
/* 30843 */ {(11<<2)|2,{56,97,0}},
/* 30844 */ {(11<<2)|2,{56,97,0}},
/* 30845 */ {(11<<2)|2,{56,97,0}},
/* 30846 */ {(11<<2)|2,{56,97,0}},
/* 30847 */ {(11<<2)|2,{56,97,0}},
/* 30848 */ {(16<<2)|3,{56,99,48}},
/* 30849 */ {(16<<2)|3,{56,99,49}},
/* 30850 */ {(16<<2)|3,{56,99,50}},
/* 30851 */ {(16<<2)|3,{56,99,97}},
/* 30852 */ {(16<<2)|3,{56,99,99}},
/* 30853 */ {(16<<2)|3,{56,99,101}},
/* 30854 */ {(16<<2)|3,{56,99,105}},
/* 30855 */ {(16<<2)|3,{56,99,111}},
/* 30856 */ {(16<<2)|3,{56,99,115}},
/* 30857 */ {(16<<2)|3,{56,99,116}},
/* 30858 */ {(11<<2)|2,{56,99,0}},
/* 30859 */ {(11<<2)|2,{56,99,0}},
/* 30860 */ {(11<<2)|2,{56,99,0}},
/* 30861 */ {(11<<2)|2,{56,99,0}},
/* 30862 */ {(11<<2)|2,{56,99,0}},
/* 30863 */ {(11<<2)|2,{56,99,0}},
/* 30864 */ {(11<<2)|2,{56,99,0}},
/* 30865 */ {(11<<2)|2,{56,99,0}},
/* 30866 */ {(11<<2)|2,{56,99,0}},
/* 30867 */ {(11<<2)|2,{56,99,0}},
/* 30868 */ {(11<<2)|2,{56,99,0}},
/* 30869 */ {(11<<2)|2,{56,99,0}},
/* 30870 */ {(11<<2)|2,{56,99,0}},
/* 30871 */ {(11<<2)|2,{56,99,0}},
/* 30872 */ {(11<<2)|2,{56,99,0}},
/* 30873 */ {(11<<2)|2,{56,99,0}},
/* 30874 */ {(11<<2)|2,{56,99,0}},
/* 30875 */ {(11<<2)|2,{56,99,0}},
/* 30876 */ {(11<<2)|2,{56,99,0}},
/* 30877 */ {(11<<2)|2,{56,99,0}},
/* 30878 */ {(11<<2)|2,{56,99,0}},
/* 30879 */ {(11<<2)|2,{56,99,0}},
/* 30880 */ {(16<<2)|3,{56,101,48}},
/* 30881 */ {(16<<2)|3,{56,101,49}},
/* 30882 */ {(16<<2)|3,{56,101,50}},
/* 30883 */ {(16<<2)|3,{56,101,97}},
/* 30884 */ {(16<<2)|3,{56,101,99}},
/* 30885 */ {(16<<2)|3,{56,101,101}},
/* 30886 */ {(16<<2)|3,{56,101,105}},
/* 30887 */ {(16<<2)|3,{56,101,111}},
/* 30888 */ {(16<<2)|3,{56,101,115}},
/* 30889 */ {(16<<2)|3,{56,101,116}},
/* 30890 */ {(11<<2)|2,{56,101,0}},
/* 30891 */ {(11<<2)|2,{56,101,0}},
/* 30892 */ {(11<<2)|2,{56,101,0}},
/* 30893 */ {(11<<2)|2,{56,101,0}},
/* 30894 */ {(11<<2)|2,{56,101,0}},
/* 30895 */ {(11<<2)|2,{56,101,0}},
/* 30896 */ {(11<<2)|2,{56,101,0}},
/* 30897 */ {(11<<2)|2,{56,101,0}},
/* 30898 */ {(11<<2)|2,{56,101,0}},
/* 30899 */ {(11<<2)|2,{56,101,0}},
/* 30900 */ {(11<<2)|2,{56,101,0}},
/* 30901 */ {(11<<2)|2,{56,101,0}},
/* 30902 */ {(11<<2)|2,{56,101,0}},
/* 30903 */ {(11<<2)|2,{56,101,0}},
/* 30904 */ {(11<<2)|2,{56,101,0}},
/* 30905 */ {(11<<2)|2,{56,101,0}},
/* 30906 */ {(11<<2)|2,{56,101,0}},
/* 30907 */ {(11<<2)|2,{56,101,0}},
/* 30908 */ {(11<<2)|2,{56,101,0}},
/* 30909 */ {(11<<2)|2,{56,101,0}},
/* 30910 */ {(11<<2)|2,{56,101,0}},
/* 30911 */ {(11<<2)|2,{56,101,0}},
/* 30912 */ {(16<<2)|3,{56,105,48}},
/* 30913 */ {(16<<2)|3,{56,105,49}},
/* 30914 */ {(16<<2)|3,{56,105,50}},
/* 30915 */ {(16<<2)|3,{56,105,97}},
/* 30916 */ {(16<<2)|3,{56,105,99}},
/* 30917 */ {(16<<2)|3,{56,105,101}},
/* 30918 */ {(16<<2)|3,{56,105,105}},
/* 30919 */ {(16<<2)|3,{56,105,111}},
/* 30920 */ {(16<<2)|3,{56,105,115}},
/* 30921 */ {(16<<2)|3,{56,105,116}},
/* 30922 */ {(11<<2)|2,{56,105,0}},
/* 30923 */ {(11<<2)|2,{56,105,0}},
/* 30924 */ {(11<<2)|2,{56,105,0}},
/* 30925 */ {(11<<2)|2,{56,105,0}},
/* 30926 */ {(11<<2)|2,{56,105,0}},
/* 30927 */ {(11<<2)|2,{56,105,0}},
/* 30928 */ {(11<<2)|2,{56,105,0}},
/* 30929 */ {(11<<2)|2,{56,105,0}},
/* 30930 */ {(11<<2)|2,{56,105,0}},
/* 30931 */ {(11<<2)|2,{56,105,0}},
/* 30932 */ {(11<<2)|2,{56,105,0}},
/* 30933 */ {(11<<2)|2,{56,105,0}},
/* 30934 */ {(11<<2)|2,{56,105,0}},
/* 30935 */ {(11<<2)|2,{56,105,0}},
/* 30936 */ {(11<<2)|2,{56,105,0}},
/* 30937 */ {(11<<2)|2,{56,105,0}},
/* 30938 */ {(11<<2)|2,{56,105,0}},
/* 30939 */ {(11<<2)|2,{56,105,0}},
/* 30940 */ {(11<<2)|2,{56,105,0}},
/* 30941 */ {(11<<2)|2,{56,105,0}},
/* 30942 */ {(11<<2)|2,{56,105,0}},
/* 30943 */ {(11<<2)|2,{56,105,0}},
/* 30944 */ {(16<<2)|3,{56,111,48}},
/* 30945 */ {(16<<2)|3,{56,111,49}},
/* 30946 */ {(16<<2)|3,{56,111,50}},
/* 30947 */ {(16<<2)|3,{56,111,97}},
/* 30948 */ {(16<<2)|3,{56,111,99}},
/* 30949 */ {(16<<2)|3,{56,111,101}},
/* 30950 */ {(16<<2)|3,{56,111,105}},
/* 30951 */ {(16<<2)|3,{56,111,111}},
/* 30952 */ {(16<<2)|3,{56,111,115}},
/* 30953 */ {(16<<2)|3,{56,111,116}},
/* 30954 */ {(11<<2)|2,{56,111,0}},
/* 30955 */ {(11<<2)|2,{56,111,0}},
/* 30956 */ {(11<<2)|2,{56,111,0}},
/* 30957 */ {(11<<2)|2,{56,111,0}},
/* 30958 */ {(11<<2)|2,{56,111,0}},
/* 30959 */ {(11<<2)|2,{56,111,0}},
/* 30960 */ {(11<<2)|2,{56,111,0}},
/* 30961 */ {(11<<2)|2,{56,111,0}},
/* 30962 */ {(11<<2)|2,{56,111,0}},
/* 30963 */ {(11<<2)|2,{56,111,0}},
/* 30964 */ {(11<<2)|2,{56,111,0}},
/* 30965 */ {(11<<2)|2,{56,111,0}},
/* 30966 */ {(11<<2)|2,{56,111,0}},
/* 30967 */ {(11<<2)|2,{56,111,0}},
/* 30968 */ {(11<<2)|2,{56,111,0}},
/* 30969 */ {(11<<2)|2,{56,111,0}},
/* 30970 */ {(11<<2)|2,{56,111,0}},
/* 30971 */ {(11<<2)|2,{56,111,0}},
/* 30972 */ {(11<<2)|2,{56,111,0}},
/* 30973 */ {(11<<2)|2,{56,111,0}},
/* 30974 */ {(11<<2)|2,{56,111,0}},
/* 30975 */ {(11<<2)|2,{56,111,0}},
/* 30976 */ {(16<<2)|3,{56,115,48}},
/* 30977 */ {(16<<2)|3,{56,115,49}},
/* 30978 */ {(16<<2)|3,{56,115,50}},
/* 30979 */ {(16<<2)|3,{56,115,97}},
/* 30980 */ {(16<<2)|3,{56,115,99}},
/* 30981 */ {(16<<2)|3,{56,115,101}},
/* 30982 */ {(16<<2)|3,{56,115,105}},
/* 30983 */ {(16<<2)|3,{56,115,111}},
/* 30984 */ {(16<<2)|3,{56,115,115}},
/* 30985 */ {(16<<2)|3,{56,115,116}},
/* 30986 */ {(11<<2)|2,{56,115,0}},
/* 30987 */ {(11<<2)|2,{56,115,0}},
/* 30988 */ {(11<<2)|2,{56,115,0}},
/* 30989 */ {(11<<2)|2,{56,115,0}},
/* 30990 */ {(11<<2)|2,{56,115,0}},
/* 30991 */ {(11<<2)|2,{56,115,0}},
/* 30992 */ {(11<<2)|2,{56,115,0}},
/* 30993 */ {(11<<2)|2,{56,115,0}},
/* 30994 */ {(11<<2)|2,{56,115,0}},
/* 30995 */ {(11<<2)|2,{56,115,0}},
/* 30996 */ {(11<<2)|2,{56,115,0}},
/* 30997 */ {(11<<2)|2,{56,115,0}},
/* 30998 */ {(11<<2)|2,{56,115,0}},
/* 30999 */ {(11<<2)|2,{56,115,0}},
/* 31000 */ {(11<<2)|2,{56,115,0}},
/* 31001 */ {(11<<2)|2,{56,115,0}},
/* 31002 */ {(11<<2)|2,{56,115,0}},
/* 31003 */ {(11<<2)|2,{56,115,0}},
/* 31004 */ {(11<<2)|2,{56,115,0}},
/* 31005 */ {(11<<2)|2,{56,115,0}},
/* 31006 */ {(11<<2)|2,{56,115,0}},
/* 31007 */ {(11<<2)|2,{56,115,0}},
/* 31008 */ {(16<<2)|3,{56,116,48}},
/* 31009 */ {(16<<2)|3,{56,116,49}},
/* 31010 */ {(16<<2)|3,{56,116,50}},
/* 31011 */ {(16<<2)|3,{56,116,97}},
/* 31012 */ {(16<<2)|3,{56,116,99}},
/* 31013 */ {(16<<2)|3,{56,116,101}},
/* 31014 */ {(16<<2)|3,{56,116,105}},
/* 31015 */ {(16<<2)|3,{56,116,111}},
/* 31016 */ {(16<<2)|3,{56,116,115}},
/* 31017 */ {(16<<2)|3,{56,116,116}},
/* 31018 */ {(11<<2)|2,{56,116,0}},
/* 31019 */ {(11<<2)|2,{56,116,0}},
/* 31020 */ {(11<<2)|2,{56,116,0}},
/* 31021 */ {(11<<2)|2,{56,116,0}},
/* 31022 */ {(11<<2)|2,{56,116,0}},
/* 31023 */ {(11<<2)|2,{56,116,0}},
/* 31024 */ {(11<<2)|2,{56,116,0}},
/* 31025 */ {(11<<2)|2,{56,116,0}},
/* 31026 */ {(11<<2)|2,{56,116,0}},
/* 31027 */ {(11<<2)|2,{56,116,0}},
/* 31028 */ {(11<<2)|2,{56,116,0}},
/* 31029 */ {(11<<2)|2,{56,116,0}},
/* 31030 */ {(11<<2)|2,{56,116,0}},
/* 31031 */ {(11<<2)|2,{56,116,0}},
/* 31032 */ {(11<<2)|2,{56,116,0}},
/* 31033 */ {(11<<2)|2,{56,116,0}},
/* 31034 */ {(11<<2)|2,{56,116,0}},
/* 31035 */ {(11<<2)|2,{56,116,0}},
/* 31036 */ {(11<<2)|2,{56,116,0}},
/* 31037 */ {(11<<2)|2,{56,116,0}},
/* 31038 */ {(11<<2)|2,{56,116,0}},
/* 31039 */ {(11<<2)|2,{56,116,0}},
/* 31040 */ {(12<<2)|2,{56,32,0}},
/* 31041 */ {(12<<2)|2,{56,32,0}},
/* 31042 */ {(12<<2)|2,{56,32,0}},
/* 31043 */ {(12<<2)|2,{56,32,0}},
/* 31044 */ {(12<<2)|2,{56,32,0}},
/* 31045 */ {(12<<2)|2,{56,32,0}},
/* 31046 */ {(12<<2)|2,{56,32,0}},
/* 31047 */ {(12<<2)|2,{56,32,0}},
/* 31048 */ {(12<<2)|2,{56,32,0}},
/* 31049 */ {(12<<2)|2,{56,32,0}},
/* 31050 */ {(12<<2)|2,{56,32,0}},
/* 31051 */ {(12<<2)|2,{56,32,0}},
/* 31052 */ {(12<<2)|2,{56,32,0}},
/* 31053 */ {(12<<2)|2,{56,32,0}},
/* 31054 */ {(12<<2)|2,{56,32,0}},
/* 31055 */ {(12<<2)|2,{56,32,0}},
/* 31056 */ {(12<<2)|2,{56,37,0}},
/* 31057 */ {(12<<2)|2,{56,37,0}},
/* 31058 */ {(12<<2)|2,{56,37,0}},
/* 31059 */ {(12<<2)|2,{56,37,0}},
/* 31060 */ {(12<<2)|2,{56,37,0}},
/* 31061 */ {(12<<2)|2,{56,37,0}},
/* 31062 */ {(12<<2)|2,{56,37,0}},
/* 31063 */ {(12<<2)|2,{56,37,0}},
/* 31064 */ {(12<<2)|2,{56,37,0}},
/* 31065 */ {(12<<2)|2,{56,37,0}},
/* 31066 */ {(12<<2)|2,{56,37,0}},
/* 31067 */ {(12<<2)|2,{56,37,0}},
/* 31068 */ {(12<<2)|2,{56,37,0}},
/* 31069 */ {(12<<2)|2,{56,37,0}},
/* 31070 */ {(12<<2)|2,{56,37,0}},
/* 31071 */ {(12<<2)|2,{56,37,0}},
/* 31072 */ {(12<<2)|2,{56,45,0}},
/* 31073 */ {(12<<2)|2,{56,45,0}},
/* 31074 */ {(12<<2)|2,{56,45,0}},
/* 31075 */ {(12<<2)|2,{56,45,0}},
/* 31076 */ {(12<<2)|2,{56,45,0}},
/* 31077 */ {(12<<2)|2,{56,45,0}},
/* 31078 */ {(12<<2)|2,{56,45,0}},
/* 31079 */ {(12<<2)|2,{56,45,0}},
/* 31080 */ {(12<<2)|2,{56,45,0}},
/* 31081 */ {(12<<2)|2,{56,45,0}},
/* 31082 */ {(12<<2)|2,{56,45,0}},
/* 31083 */ {(12<<2)|2,{56,45,0}},
/* 31084 */ {(12<<2)|2,{56,45,0}},
/* 31085 */ {(12<<2)|2,{56,45,0}},
/* 31086 */ {(12<<2)|2,{56,45,0}},
/* 31087 */ {(12<<2)|2,{56,45,0}},
/* 31088 */ {(12<<2)|2,{56,46,0}},
/* 31089 */ {(12<<2)|2,{56,46,0}},
/* 31090 */ {(12<<2)|2,{56,46,0}},
/* 31091 */ {(12<<2)|2,{56,46,0}},
/* 31092 */ {(12<<2)|2,{56,46,0}},
/* 31093 */ {(12<<2)|2,{56,46,0}},
/* 31094 */ {(12<<2)|2,{56,46,0}},
/* 31095 */ {(12<<2)|2,{56,46,0}},
/* 31096 */ {(12<<2)|2,{56,46,0}},
/* 31097 */ {(12<<2)|2,{56,46,0}},
/* 31098 */ {(12<<2)|2,{56,46,0}},
/* 31099 */ {(12<<2)|2,{56,46,0}},
/* 31100 */ {(12<<2)|2,{56,46,0}},
/* 31101 */ {(12<<2)|2,{56,46,0}},
/* 31102 */ {(12<<2)|2,{56,46,0}},
/* 31103 */ {(12<<2)|2,{56,46,0}},
/* 31104 */ {(12<<2)|2,{56,47,0}},
/* 31105 */ {(12<<2)|2,{56,47,0}},
/* 31106 */ {(12<<2)|2,{56,47,0}},
/* 31107 */ {(12<<2)|2,{56,47,0}},
/* 31108 */ {(12<<2)|2,{56,47,0}},
/* 31109 */ {(12<<2)|2,{56,47,0}},
/* 31110 */ {(12<<2)|2,{56,47,0}},
/* 31111 */ {(12<<2)|2,{56,47,0}},
/* 31112 */ {(12<<2)|2,{56,47,0}},
/* 31113 */ {(12<<2)|2,{56,47,0}},
/* 31114 */ {(12<<2)|2,{56,47,0}},
/* 31115 */ {(12<<2)|2,{56,47,0}},
/* 31116 */ {(12<<2)|2,{56,47,0}},
/* 31117 */ {(12<<2)|2,{56,47,0}},
/* 31118 */ {(12<<2)|2,{56,47,0}},
/* 31119 */ {(12<<2)|2,{56,47,0}},
/* 31120 */ {(12<<2)|2,{56,51,0}},
/* 31121 */ {(12<<2)|2,{56,51,0}},
/* 31122 */ {(12<<2)|2,{56,51,0}},
/* 31123 */ {(12<<2)|2,{56,51,0}},
/* 31124 */ {(12<<2)|2,{56,51,0}},
/* 31125 */ {(12<<2)|2,{56,51,0}},
/* 31126 */ {(12<<2)|2,{56,51,0}},
/* 31127 */ {(12<<2)|2,{56,51,0}},
/* 31128 */ {(12<<2)|2,{56,51,0}},
/* 31129 */ {(12<<2)|2,{56,51,0}},
/* 31130 */ {(12<<2)|2,{56,51,0}},
/* 31131 */ {(12<<2)|2,{56,51,0}},
/* 31132 */ {(12<<2)|2,{56,51,0}},
/* 31133 */ {(12<<2)|2,{56,51,0}},
/* 31134 */ {(12<<2)|2,{56,51,0}},
/* 31135 */ {(12<<2)|2,{56,51,0}},
/* 31136 */ {(12<<2)|2,{56,52,0}},
/* 31137 */ {(12<<2)|2,{56,52,0}},
/* 31138 */ {(12<<2)|2,{56,52,0}},
/* 31139 */ {(12<<2)|2,{56,52,0}},
/* 31140 */ {(12<<2)|2,{56,52,0}},
/* 31141 */ {(12<<2)|2,{56,52,0}},
/* 31142 */ {(12<<2)|2,{56,52,0}},
/* 31143 */ {(12<<2)|2,{56,52,0}},
/* 31144 */ {(12<<2)|2,{56,52,0}},
/* 31145 */ {(12<<2)|2,{56,52,0}},
/* 31146 */ {(12<<2)|2,{56,52,0}},
/* 31147 */ {(12<<2)|2,{56,52,0}},
/* 31148 */ {(12<<2)|2,{56,52,0}},
/* 31149 */ {(12<<2)|2,{56,52,0}},
/* 31150 */ {(12<<2)|2,{56,52,0}},
/* 31151 */ {(12<<2)|2,{56,52,0}},
/* 31152 */ {(12<<2)|2,{56,53,0}},
/* 31153 */ {(12<<2)|2,{56,53,0}},
/* 31154 */ {(12<<2)|2,{56,53,0}},
/* 31155 */ {(12<<2)|2,{56,53,0}},
/* 31156 */ {(12<<2)|2,{56,53,0}},
/* 31157 */ {(12<<2)|2,{56,53,0}},
/* 31158 */ {(12<<2)|2,{56,53,0}},
/* 31159 */ {(12<<2)|2,{56,53,0}},
/* 31160 */ {(12<<2)|2,{56,53,0}},
/* 31161 */ {(12<<2)|2,{56,53,0}},
/* 31162 */ {(12<<2)|2,{56,53,0}},
/* 31163 */ {(12<<2)|2,{56,53,0}},
/* 31164 */ {(12<<2)|2,{56,53,0}},
/* 31165 */ {(12<<2)|2,{56,53,0}},
/* 31166 */ {(12<<2)|2,{56,53,0}},
/* 31167 */ {(12<<2)|2,{56,53,0}},
/* 31168 */ {(12<<2)|2,{56,54,0}},
/* 31169 */ {(12<<2)|2,{56,54,0}},
/* 31170 */ {(12<<2)|2,{56,54,0}},
/* 31171 */ {(12<<2)|2,{56,54,0}},
/* 31172 */ {(12<<2)|2,{56,54,0}},
/* 31173 */ {(12<<2)|2,{56,54,0}},
/* 31174 */ {(12<<2)|2,{56,54,0}},
/* 31175 */ {(12<<2)|2,{56,54,0}},
/* 31176 */ {(12<<2)|2,{56,54,0}},
/* 31177 */ {(12<<2)|2,{56,54,0}},
/* 31178 */ {(12<<2)|2,{56,54,0}},
/* 31179 */ {(12<<2)|2,{56,54,0}},
/* 31180 */ {(12<<2)|2,{56,54,0}},
/* 31181 */ {(12<<2)|2,{56,54,0}},
/* 31182 */ {(12<<2)|2,{56,54,0}},
/* 31183 */ {(12<<2)|2,{56,54,0}},
/* 31184 */ {(12<<2)|2,{56,55,0}},
/* 31185 */ {(12<<2)|2,{56,55,0}},
/* 31186 */ {(12<<2)|2,{56,55,0}},
/* 31187 */ {(12<<2)|2,{56,55,0}},
/* 31188 */ {(12<<2)|2,{56,55,0}},
/* 31189 */ {(12<<2)|2,{56,55,0}},
/* 31190 */ {(12<<2)|2,{56,55,0}},
/* 31191 */ {(12<<2)|2,{56,55,0}},
/* 31192 */ {(12<<2)|2,{56,55,0}},
/* 31193 */ {(12<<2)|2,{56,55,0}},
/* 31194 */ {(12<<2)|2,{56,55,0}},
/* 31195 */ {(12<<2)|2,{56,55,0}},
/* 31196 */ {(12<<2)|2,{56,55,0}},
/* 31197 */ {(12<<2)|2,{56,55,0}},
/* 31198 */ {(12<<2)|2,{56,55,0}},
/* 31199 */ {(12<<2)|2,{56,55,0}},
/* 31200 */ {(12<<2)|2,{56,56,0}},
/* 31201 */ {(12<<2)|2,{56,56,0}},
/* 31202 */ {(12<<2)|2,{56,56,0}},
/* 31203 */ {(12<<2)|2,{56,56,0}},
/* 31204 */ {(12<<2)|2,{56,56,0}},
/* 31205 */ {(12<<2)|2,{56,56,0}},
/* 31206 */ {(12<<2)|2,{56,56,0}},
/* 31207 */ {(12<<2)|2,{56,56,0}},
/* 31208 */ {(12<<2)|2,{56,56,0}},
/* 31209 */ {(12<<2)|2,{56,56,0}},
/* 31210 */ {(12<<2)|2,{56,56,0}},
/* 31211 */ {(12<<2)|2,{56,56,0}},
/* 31212 */ {(12<<2)|2,{56,56,0}},
/* 31213 */ {(12<<2)|2,{56,56,0}},
/* 31214 */ {(12<<2)|2,{56,56,0}},
/* 31215 */ {(12<<2)|2,{56,56,0}},
/* 31216 */ {(12<<2)|2,{56,57,0}},
/* 31217 */ {(12<<2)|2,{56,57,0}},
/* 31218 */ {(12<<2)|2,{56,57,0}},
/* 31219 */ {(12<<2)|2,{56,57,0}},
/* 31220 */ {(12<<2)|2,{56,57,0}},
/* 31221 */ {(12<<2)|2,{56,57,0}},
/* 31222 */ {(12<<2)|2,{56,57,0}},
/* 31223 */ {(12<<2)|2,{56,57,0}},
/* 31224 */ {(12<<2)|2,{56,57,0}},
/* 31225 */ {(12<<2)|2,{56,57,0}},
/* 31226 */ {(12<<2)|2,{56,57,0}},
/* 31227 */ {(12<<2)|2,{56,57,0}},
/* 31228 */ {(12<<2)|2,{56,57,0}},
/* 31229 */ {(12<<2)|2,{56,57,0}},
/* 31230 */ {(12<<2)|2,{56,57,0}},
/* 31231 */ {(12<<2)|2,{56,57,0}},
/* 31232 */ {(12<<2)|2,{56,61,0}},
/* 31233 */ {(12<<2)|2,{56,61,0}},
/* 31234 */ {(12<<2)|2,{56,61,0}},
/* 31235 */ {(12<<2)|2,{56,61,0}},
/* 31236 */ {(12<<2)|2,{56,61,0}},
/* 31237 */ {(12<<2)|2,{56,61,0}},
/* 31238 */ {(12<<2)|2,{56,61,0}},
/* 31239 */ {(12<<2)|2,{56,61,0}},
/* 31240 */ {(12<<2)|2,{56,61,0}},
/* 31241 */ {(12<<2)|2,{56,61,0}},
/* 31242 */ {(12<<2)|2,{56,61,0}},
/* 31243 */ {(12<<2)|2,{56,61,0}},
/* 31244 */ {(12<<2)|2,{56,61,0}},
/* 31245 */ {(12<<2)|2,{56,61,0}},
/* 31246 */ {(12<<2)|2,{56,61,0}},
/* 31247 */ {(12<<2)|2,{56,61,0}},
/* 31248 */ {(12<<2)|2,{56,65,0}},
/* 31249 */ {(12<<2)|2,{56,65,0}},
/* 31250 */ {(12<<2)|2,{56,65,0}},
/* 31251 */ {(12<<2)|2,{56,65,0}},
/* 31252 */ {(12<<2)|2,{56,65,0}},
/* 31253 */ {(12<<2)|2,{56,65,0}},
/* 31254 */ {(12<<2)|2,{56,65,0}},
/* 31255 */ {(12<<2)|2,{56,65,0}},
/* 31256 */ {(12<<2)|2,{56,65,0}},
/* 31257 */ {(12<<2)|2,{56,65,0}},
/* 31258 */ {(12<<2)|2,{56,65,0}},
/* 31259 */ {(12<<2)|2,{56,65,0}},
/* 31260 */ {(12<<2)|2,{56,65,0}},
/* 31261 */ {(12<<2)|2,{56,65,0}},
/* 31262 */ {(12<<2)|2,{56,65,0}},
/* 31263 */ {(12<<2)|2,{56,65,0}},
/* 31264 */ {(12<<2)|2,{56,95,0}},
/* 31265 */ {(12<<2)|2,{56,95,0}},
/* 31266 */ {(12<<2)|2,{56,95,0}},
/* 31267 */ {(12<<2)|2,{56,95,0}},
/* 31268 */ {(12<<2)|2,{56,95,0}},
/* 31269 */ {(12<<2)|2,{56,95,0}},
/* 31270 */ {(12<<2)|2,{56,95,0}},
/* 31271 */ {(12<<2)|2,{56,95,0}},
/* 31272 */ {(12<<2)|2,{56,95,0}},
/* 31273 */ {(12<<2)|2,{56,95,0}},
/* 31274 */ {(12<<2)|2,{56,95,0}},
/* 31275 */ {(12<<2)|2,{56,95,0}},
/* 31276 */ {(12<<2)|2,{56,95,0}},
/* 31277 */ {(12<<2)|2,{56,95,0}},
/* 31278 */ {(12<<2)|2,{56,95,0}},
/* 31279 */ {(12<<2)|2,{56,95,0}},
/* 31280 */ {(12<<2)|2,{56,98,0}},
/* 31281 */ {(12<<2)|2,{56,98,0}},
/* 31282 */ {(12<<2)|2,{56,98,0}},
/* 31283 */ {(12<<2)|2,{56,98,0}},
/* 31284 */ {(12<<2)|2,{56,98,0}},
/* 31285 */ {(12<<2)|2,{56,98,0}},
/* 31286 */ {(12<<2)|2,{56,98,0}},
/* 31287 */ {(12<<2)|2,{56,98,0}},
/* 31288 */ {(12<<2)|2,{56,98,0}},
/* 31289 */ {(12<<2)|2,{56,98,0}},
/* 31290 */ {(12<<2)|2,{56,98,0}},
/* 31291 */ {(12<<2)|2,{56,98,0}},
/* 31292 */ {(12<<2)|2,{56,98,0}},
/* 31293 */ {(12<<2)|2,{56,98,0}},
/* 31294 */ {(12<<2)|2,{56,98,0}},
/* 31295 */ {(12<<2)|2,{56,98,0}},
/* 31296 */ {(12<<2)|2,{56,100,0}},
/* 31297 */ {(12<<2)|2,{56,100,0}},
/* 31298 */ {(12<<2)|2,{56,100,0}},
/* 31299 */ {(12<<2)|2,{56,100,0}},
/* 31300 */ {(12<<2)|2,{56,100,0}},
/* 31301 */ {(12<<2)|2,{56,100,0}},
/* 31302 */ {(12<<2)|2,{56,100,0}},
/* 31303 */ {(12<<2)|2,{56,100,0}},
/* 31304 */ {(12<<2)|2,{56,100,0}},
/* 31305 */ {(12<<2)|2,{56,100,0}},
/* 31306 */ {(12<<2)|2,{56,100,0}},
/* 31307 */ {(12<<2)|2,{56,100,0}},
/* 31308 */ {(12<<2)|2,{56,100,0}},
/* 31309 */ {(12<<2)|2,{56,100,0}},
/* 31310 */ {(12<<2)|2,{56,100,0}},
/* 31311 */ {(12<<2)|2,{56,100,0}},
/* 31312 */ {(12<<2)|2,{56,102,0}},
/* 31313 */ {(12<<2)|2,{56,102,0}},
/* 31314 */ {(12<<2)|2,{56,102,0}},
/* 31315 */ {(12<<2)|2,{56,102,0}},
/* 31316 */ {(12<<2)|2,{56,102,0}},
/* 31317 */ {(12<<2)|2,{56,102,0}},
/* 31318 */ {(12<<2)|2,{56,102,0}},
/* 31319 */ {(12<<2)|2,{56,102,0}},
/* 31320 */ {(12<<2)|2,{56,102,0}},
/* 31321 */ {(12<<2)|2,{56,102,0}},
/* 31322 */ {(12<<2)|2,{56,102,0}},
/* 31323 */ {(12<<2)|2,{56,102,0}},
/* 31324 */ {(12<<2)|2,{56,102,0}},
/* 31325 */ {(12<<2)|2,{56,102,0}},
/* 31326 */ {(12<<2)|2,{56,102,0}},
/* 31327 */ {(12<<2)|2,{56,102,0}},
/* 31328 */ {(12<<2)|2,{56,103,0}},
/* 31329 */ {(12<<2)|2,{56,103,0}},
/* 31330 */ {(12<<2)|2,{56,103,0}},
/* 31331 */ {(12<<2)|2,{56,103,0}},
/* 31332 */ {(12<<2)|2,{56,103,0}},
/* 31333 */ {(12<<2)|2,{56,103,0}},
/* 31334 */ {(12<<2)|2,{56,103,0}},
/* 31335 */ {(12<<2)|2,{56,103,0}},
/* 31336 */ {(12<<2)|2,{56,103,0}},
/* 31337 */ {(12<<2)|2,{56,103,0}},
/* 31338 */ {(12<<2)|2,{56,103,0}},
/* 31339 */ {(12<<2)|2,{56,103,0}},
/* 31340 */ {(12<<2)|2,{56,103,0}},
/* 31341 */ {(12<<2)|2,{56,103,0}},
/* 31342 */ {(12<<2)|2,{56,103,0}},
/* 31343 */ {(12<<2)|2,{56,103,0}},
/* 31344 */ {(12<<2)|2,{56,104,0}},
/* 31345 */ {(12<<2)|2,{56,104,0}},
/* 31346 */ {(12<<2)|2,{56,104,0}},
/* 31347 */ {(12<<2)|2,{56,104,0}},
/* 31348 */ {(12<<2)|2,{56,104,0}},
/* 31349 */ {(12<<2)|2,{56,104,0}},
/* 31350 */ {(12<<2)|2,{56,104,0}},
/* 31351 */ {(12<<2)|2,{56,104,0}},
/* 31352 */ {(12<<2)|2,{56,104,0}},
/* 31353 */ {(12<<2)|2,{56,104,0}},
/* 31354 */ {(12<<2)|2,{56,104,0}},
/* 31355 */ {(12<<2)|2,{56,104,0}},
/* 31356 */ {(12<<2)|2,{56,104,0}},
/* 31357 */ {(12<<2)|2,{56,104,0}},
/* 31358 */ {(12<<2)|2,{56,104,0}},
/* 31359 */ {(12<<2)|2,{56,104,0}},
/* 31360 */ {(12<<2)|2,{56,108,0}},
/* 31361 */ {(12<<2)|2,{56,108,0}},
/* 31362 */ {(12<<2)|2,{56,108,0}},
/* 31363 */ {(12<<2)|2,{56,108,0}},
/* 31364 */ {(12<<2)|2,{56,108,0}},
/* 31365 */ {(12<<2)|2,{56,108,0}},
/* 31366 */ {(12<<2)|2,{56,108,0}},
/* 31367 */ {(12<<2)|2,{56,108,0}},
/* 31368 */ {(12<<2)|2,{56,108,0}},
/* 31369 */ {(12<<2)|2,{56,108,0}},
/* 31370 */ {(12<<2)|2,{56,108,0}},
/* 31371 */ {(12<<2)|2,{56,108,0}},
/* 31372 */ {(12<<2)|2,{56,108,0}},
/* 31373 */ {(12<<2)|2,{56,108,0}},
/* 31374 */ {(12<<2)|2,{56,108,0}},
/* 31375 */ {(12<<2)|2,{56,108,0}},
/* 31376 */ {(12<<2)|2,{56,109,0}},
/* 31377 */ {(12<<2)|2,{56,109,0}},
/* 31378 */ {(12<<2)|2,{56,109,0}},
/* 31379 */ {(12<<2)|2,{56,109,0}},
/* 31380 */ {(12<<2)|2,{56,109,0}},
/* 31381 */ {(12<<2)|2,{56,109,0}},
/* 31382 */ {(12<<2)|2,{56,109,0}},
/* 31383 */ {(12<<2)|2,{56,109,0}},
/* 31384 */ {(12<<2)|2,{56,109,0}},
/* 31385 */ {(12<<2)|2,{56,109,0}},
/* 31386 */ {(12<<2)|2,{56,109,0}},
/* 31387 */ {(12<<2)|2,{56,109,0}},
/* 31388 */ {(12<<2)|2,{56,109,0}},
/* 31389 */ {(12<<2)|2,{56,109,0}},
/* 31390 */ {(12<<2)|2,{56,109,0}},
/* 31391 */ {(12<<2)|2,{56,109,0}},
/* 31392 */ {(12<<2)|2,{56,110,0}},
/* 31393 */ {(12<<2)|2,{56,110,0}},
/* 31394 */ {(12<<2)|2,{56,110,0}},
/* 31395 */ {(12<<2)|2,{56,110,0}},
/* 31396 */ {(12<<2)|2,{56,110,0}},
/* 31397 */ {(12<<2)|2,{56,110,0}},
/* 31398 */ {(12<<2)|2,{56,110,0}},
/* 31399 */ {(12<<2)|2,{56,110,0}},
/* 31400 */ {(12<<2)|2,{56,110,0}},
/* 31401 */ {(12<<2)|2,{56,110,0}},
/* 31402 */ {(12<<2)|2,{56,110,0}},
/* 31403 */ {(12<<2)|2,{56,110,0}},
/* 31404 */ {(12<<2)|2,{56,110,0}},
/* 31405 */ {(12<<2)|2,{56,110,0}},
/* 31406 */ {(12<<2)|2,{56,110,0}},
/* 31407 */ {(12<<2)|2,{56,110,0}},
/* 31408 */ {(12<<2)|2,{56,112,0}},
/* 31409 */ {(12<<2)|2,{56,112,0}},
/* 31410 */ {(12<<2)|2,{56,112,0}},
/* 31411 */ {(12<<2)|2,{56,112,0}},
/* 31412 */ {(12<<2)|2,{56,112,0}},
/* 31413 */ {(12<<2)|2,{56,112,0}},
/* 31414 */ {(12<<2)|2,{56,112,0}},
/* 31415 */ {(12<<2)|2,{56,112,0}},
/* 31416 */ {(12<<2)|2,{56,112,0}},
/* 31417 */ {(12<<2)|2,{56,112,0}},
/* 31418 */ {(12<<2)|2,{56,112,0}},
/* 31419 */ {(12<<2)|2,{56,112,0}},
/* 31420 */ {(12<<2)|2,{56,112,0}},
/* 31421 */ {(12<<2)|2,{56,112,0}},
/* 31422 */ {(12<<2)|2,{56,112,0}},
/* 31423 */ {(12<<2)|2,{56,112,0}},
/* 31424 */ {(12<<2)|2,{56,114,0}},
/* 31425 */ {(12<<2)|2,{56,114,0}},
/* 31426 */ {(12<<2)|2,{56,114,0}},
/* 31427 */ {(12<<2)|2,{56,114,0}},
/* 31428 */ {(12<<2)|2,{56,114,0}},
/* 31429 */ {(12<<2)|2,{56,114,0}},
/* 31430 */ {(12<<2)|2,{56,114,0}},
/* 31431 */ {(12<<2)|2,{56,114,0}},
/* 31432 */ {(12<<2)|2,{56,114,0}},
/* 31433 */ {(12<<2)|2,{56,114,0}},
/* 31434 */ {(12<<2)|2,{56,114,0}},
/* 31435 */ {(12<<2)|2,{56,114,0}},
/* 31436 */ {(12<<2)|2,{56,114,0}},
/* 31437 */ {(12<<2)|2,{56,114,0}},
/* 31438 */ {(12<<2)|2,{56,114,0}},
/* 31439 */ {(12<<2)|2,{56,114,0}},
/* 31440 */ {(12<<2)|2,{56,117,0}},
/* 31441 */ {(12<<2)|2,{56,117,0}},
/* 31442 */ {(12<<2)|2,{56,117,0}},
/* 31443 */ {(12<<2)|2,{56,117,0}},
/* 31444 */ {(12<<2)|2,{56,117,0}},
/* 31445 */ {(12<<2)|2,{56,117,0}},
/* 31446 */ {(12<<2)|2,{56,117,0}},
/* 31447 */ {(12<<2)|2,{56,117,0}},
/* 31448 */ {(12<<2)|2,{56,117,0}},
/* 31449 */ {(12<<2)|2,{56,117,0}},
/* 31450 */ {(12<<2)|2,{56,117,0}},
/* 31451 */ {(12<<2)|2,{56,117,0}},
/* 31452 */ {(12<<2)|2,{56,117,0}},
/* 31453 */ {(12<<2)|2,{56,117,0}},
/* 31454 */ {(12<<2)|2,{56,117,0}},
/* 31455 */ {(12<<2)|2,{56,117,0}},
/* 31456 */ {(13<<2)|2,{56,58,0}},
/* 31457 */ {(13<<2)|2,{56,58,0}},
/* 31458 */ {(13<<2)|2,{56,58,0}},
/* 31459 */ {(13<<2)|2,{56,58,0}},
/* 31460 */ {(13<<2)|2,{56,58,0}},
/* 31461 */ {(13<<2)|2,{56,58,0}},
/* 31462 */ {(13<<2)|2,{56,58,0}},
/* 31463 */ {(13<<2)|2,{56,58,0}},
/* 31464 */ {(13<<2)|2,{56,66,0}},
/* 31465 */ {(13<<2)|2,{56,66,0}},
/* 31466 */ {(13<<2)|2,{56,66,0}},
/* 31467 */ {(13<<2)|2,{56,66,0}},
/* 31468 */ {(13<<2)|2,{56,66,0}},
/* 31469 */ {(13<<2)|2,{56,66,0}},
/* 31470 */ {(13<<2)|2,{56,66,0}},
/* 31471 */ {(13<<2)|2,{56,66,0}},
/* 31472 */ {(13<<2)|2,{56,67,0}},
/* 31473 */ {(13<<2)|2,{56,67,0}},
/* 31474 */ {(13<<2)|2,{56,67,0}},
/* 31475 */ {(13<<2)|2,{56,67,0}},
/* 31476 */ {(13<<2)|2,{56,67,0}},
/* 31477 */ {(13<<2)|2,{56,67,0}},
/* 31478 */ {(13<<2)|2,{56,67,0}},
/* 31479 */ {(13<<2)|2,{56,67,0}},
/* 31480 */ {(13<<2)|2,{56,68,0}},
/* 31481 */ {(13<<2)|2,{56,68,0}},
/* 31482 */ {(13<<2)|2,{56,68,0}},
/* 31483 */ {(13<<2)|2,{56,68,0}},
/* 31484 */ {(13<<2)|2,{56,68,0}},
/* 31485 */ {(13<<2)|2,{56,68,0}},
/* 31486 */ {(13<<2)|2,{56,68,0}},
/* 31487 */ {(13<<2)|2,{56,68,0}},
/* 31488 */ {(13<<2)|2,{56,69,0}},
/* 31489 */ {(13<<2)|2,{56,69,0}},
/* 31490 */ {(13<<2)|2,{56,69,0}},
/* 31491 */ {(13<<2)|2,{56,69,0}},
/* 31492 */ {(13<<2)|2,{56,69,0}},
/* 31493 */ {(13<<2)|2,{56,69,0}},
/* 31494 */ {(13<<2)|2,{56,69,0}},
/* 31495 */ {(13<<2)|2,{56,69,0}},
/* 31496 */ {(13<<2)|2,{56,70,0}},
/* 31497 */ {(13<<2)|2,{56,70,0}},
/* 31498 */ {(13<<2)|2,{56,70,0}},
/* 31499 */ {(13<<2)|2,{56,70,0}},
/* 31500 */ {(13<<2)|2,{56,70,0}},
/* 31501 */ {(13<<2)|2,{56,70,0}},
/* 31502 */ {(13<<2)|2,{56,70,0}},
/* 31503 */ {(13<<2)|2,{56,70,0}},
/* 31504 */ {(13<<2)|2,{56,71,0}},
/* 31505 */ {(13<<2)|2,{56,71,0}},
/* 31506 */ {(13<<2)|2,{56,71,0}},
/* 31507 */ {(13<<2)|2,{56,71,0}},
/* 31508 */ {(13<<2)|2,{56,71,0}},
/* 31509 */ {(13<<2)|2,{56,71,0}},
/* 31510 */ {(13<<2)|2,{56,71,0}},
/* 31511 */ {(13<<2)|2,{56,71,0}},
/* 31512 */ {(13<<2)|2,{56,72,0}},
/* 31513 */ {(13<<2)|2,{56,72,0}},
/* 31514 */ {(13<<2)|2,{56,72,0}},
/* 31515 */ {(13<<2)|2,{56,72,0}},
/* 31516 */ {(13<<2)|2,{56,72,0}},
/* 31517 */ {(13<<2)|2,{56,72,0}},
/* 31518 */ {(13<<2)|2,{56,72,0}},
/* 31519 */ {(13<<2)|2,{56,72,0}},
/* 31520 */ {(13<<2)|2,{56,73,0}},
/* 31521 */ {(13<<2)|2,{56,73,0}},
/* 31522 */ {(13<<2)|2,{56,73,0}},
/* 31523 */ {(13<<2)|2,{56,73,0}},
/* 31524 */ {(13<<2)|2,{56,73,0}},
/* 31525 */ {(13<<2)|2,{56,73,0}},
/* 31526 */ {(13<<2)|2,{56,73,0}},
/* 31527 */ {(13<<2)|2,{56,73,0}},
/* 31528 */ {(13<<2)|2,{56,74,0}},
/* 31529 */ {(13<<2)|2,{56,74,0}},
/* 31530 */ {(13<<2)|2,{56,74,0}},
/* 31531 */ {(13<<2)|2,{56,74,0}},
/* 31532 */ {(13<<2)|2,{56,74,0}},
/* 31533 */ {(13<<2)|2,{56,74,0}},
/* 31534 */ {(13<<2)|2,{56,74,0}},
/* 31535 */ {(13<<2)|2,{56,74,0}},
/* 31536 */ {(13<<2)|2,{56,75,0}},
/* 31537 */ {(13<<2)|2,{56,75,0}},
/* 31538 */ {(13<<2)|2,{56,75,0}},
/* 31539 */ {(13<<2)|2,{56,75,0}},
/* 31540 */ {(13<<2)|2,{56,75,0}},
/* 31541 */ {(13<<2)|2,{56,75,0}},
/* 31542 */ {(13<<2)|2,{56,75,0}},
/* 31543 */ {(13<<2)|2,{56,75,0}},
/* 31544 */ {(13<<2)|2,{56,76,0}},
/* 31545 */ {(13<<2)|2,{56,76,0}},
/* 31546 */ {(13<<2)|2,{56,76,0}},
/* 31547 */ {(13<<2)|2,{56,76,0}},
/* 31548 */ {(13<<2)|2,{56,76,0}},
/* 31549 */ {(13<<2)|2,{56,76,0}},
/* 31550 */ {(13<<2)|2,{56,76,0}},
/* 31551 */ {(13<<2)|2,{56,76,0}},
/* 31552 */ {(13<<2)|2,{56,77,0}},
/* 31553 */ {(13<<2)|2,{56,77,0}},
/* 31554 */ {(13<<2)|2,{56,77,0}},
/* 31555 */ {(13<<2)|2,{56,77,0}},
/* 31556 */ {(13<<2)|2,{56,77,0}},
/* 31557 */ {(13<<2)|2,{56,77,0}},
/* 31558 */ {(13<<2)|2,{56,77,0}},
/* 31559 */ {(13<<2)|2,{56,77,0}},
/* 31560 */ {(13<<2)|2,{56,78,0}},
/* 31561 */ {(13<<2)|2,{56,78,0}},
/* 31562 */ {(13<<2)|2,{56,78,0}},
/* 31563 */ {(13<<2)|2,{56,78,0}},
/* 31564 */ {(13<<2)|2,{56,78,0}},
/* 31565 */ {(13<<2)|2,{56,78,0}},
/* 31566 */ {(13<<2)|2,{56,78,0}},
/* 31567 */ {(13<<2)|2,{56,78,0}},
/* 31568 */ {(13<<2)|2,{56,79,0}},
/* 31569 */ {(13<<2)|2,{56,79,0}},
/* 31570 */ {(13<<2)|2,{56,79,0}},
/* 31571 */ {(13<<2)|2,{56,79,0}},
/* 31572 */ {(13<<2)|2,{56,79,0}},
/* 31573 */ {(13<<2)|2,{56,79,0}},
/* 31574 */ {(13<<2)|2,{56,79,0}},
/* 31575 */ {(13<<2)|2,{56,79,0}},
/* 31576 */ {(13<<2)|2,{56,80,0}},
/* 31577 */ {(13<<2)|2,{56,80,0}},
/* 31578 */ {(13<<2)|2,{56,80,0}},
/* 31579 */ {(13<<2)|2,{56,80,0}},
/* 31580 */ {(13<<2)|2,{56,80,0}},
/* 31581 */ {(13<<2)|2,{56,80,0}},
/* 31582 */ {(13<<2)|2,{56,80,0}},
/* 31583 */ {(13<<2)|2,{56,80,0}},
/* 31584 */ {(13<<2)|2,{56,81,0}},
/* 31585 */ {(13<<2)|2,{56,81,0}},
/* 31586 */ {(13<<2)|2,{56,81,0}},
/* 31587 */ {(13<<2)|2,{56,81,0}},
/* 31588 */ {(13<<2)|2,{56,81,0}},
/* 31589 */ {(13<<2)|2,{56,81,0}},
/* 31590 */ {(13<<2)|2,{56,81,0}},
/* 31591 */ {(13<<2)|2,{56,81,0}},
/* 31592 */ {(13<<2)|2,{56,82,0}},
/* 31593 */ {(13<<2)|2,{56,82,0}},
/* 31594 */ {(13<<2)|2,{56,82,0}},
/* 31595 */ {(13<<2)|2,{56,82,0}},
/* 31596 */ {(13<<2)|2,{56,82,0}},
/* 31597 */ {(13<<2)|2,{56,82,0}},
/* 31598 */ {(13<<2)|2,{56,82,0}},
/* 31599 */ {(13<<2)|2,{56,82,0}},
/* 31600 */ {(13<<2)|2,{56,83,0}},
/* 31601 */ {(13<<2)|2,{56,83,0}},
/* 31602 */ {(13<<2)|2,{56,83,0}},
/* 31603 */ {(13<<2)|2,{56,83,0}},
/* 31604 */ {(13<<2)|2,{56,83,0}},
/* 31605 */ {(13<<2)|2,{56,83,0}},
/* 31606 */ {(13<<2)|2,{56,83,0}},
/* 31607 */ {(13<<2)|2,{56,83,0}},
/* 31608 */ {(13<<2)|2,{56,84,0}},
/* 31609 */ {(13<<2)|2,{56,84,0}},
/* 31610 */ {(13<<2)|2,{56,84,0}},
/* 31611 */ {(13<<2)|2,{56,84,0}},
/* 31612 */ {(13<<2)|2,{56,84,0}},
/* 31613 */ {(13<<2)|2,{56,84,0}},
/* 31614 */ {(13<<2)|2,{56,84,0}},
/* 31615 */ {(13<<2)|2,{56,84,0}},
/* 31616 */ {(13<<2)|2,{56,85,0}},
/* 31617 */ {(13<<2)|2,{56,85,0}},
/* 31618 */ {(13<<2)|2,{56,85,0}},
/* 31619 */ {(13<<2)|2,{56,85,0}},
/* 31620 */ {(13<<2)|2,{56,85,0}},
/* 31621 */ {(13<<2)|2,{56,85,0}},
/* 31622 */ {(13<<2)|2,{56,85,0}},
/* 31623 */ {(13<<2)|2,{56,85,0}},
/* 31624 */ {(13<<2)|2,{56,86,0}},
/* 31625 */ {(13<<2)|2,{56,86,0}},
/* 31626 */ {(13<<2)|2,{56,86,0}},
/* 31627 */ {(13<<2)|2,{56,86,0}},
/* 31628 */ {(13<<2)|2,{56,86,0}},
/* 31629 */ {(13<<2)|2,{56,86,0}},
/* 31630 */ {(13<<2)|2,{56,86,0}},
/* 31631 */ {(13<<2)|2,{56,86,0}},
/* 31632 */ {(13<<2)|2,{56,87,0}},
/* 31633 */ {(13<<2)|2,{56,87,0}},
/* 31634 */ {(13<<2)|2,{56,87,0}},
/* 31635 */ {(13<<2)|2,{56,87,0}},
/* 31636 */ {(13<<2)|2,{56,87,0}},
/* 31637 */ {(13<<2)|2,{56,87,0}},
/* 31638 */ {(13<<2)|2,{56,87,0}},
/* 31639 */ {(13<<2)|2,{56,87,0}},
/* 31640 */ {(13<<2)|2,{56,89,0}},
/* 31641 */ {(13<<2)|2,{56,89,0}},
/* 31642 */ {(13<<2)|2,{56,89,0}},
/* 31643 */ {(13<<2)|2,{56,89,0}},
/* 31644 */ {(13<<2)|2,{56,89,0}},
/* 31645 */ {(13<<2)|2,{56,89,0}},
/* 31646 */ {(13<<2)|2,{56,89,0}},
/* 31647 */ {(13<<2)|2,{56,89,0}},
/* 31648 */ {(13<<2)|2,{56,106,0}},
/* 31649 */ {(13<<2)|2,{56,106,0}},
/* 31650 */ {(13<<2)|2,{56,106,0}},
/* 31651 */ {(13<<2)|2,{56,106,0}},
/* 31652 */ {(13<<2)|2,{56,106,0}},
/* 31653 */ {(13<<2)|2,{56,106,0}},
/* 31654 */ {(13<<2)|2,{56,106,0}},
/* 31655 */ {(13<<2)|2,{56,106,0}},
/* 31656 */ {(13<<2)|2,{56,107,0}},
/* 31657 */ {(13<<2)|2,{56,107,0}},
/* 31658 */ {(13<<2)|2,{56,107,0}},
/* 31659 */ {(13<<2)|2,{56,107,0}},
/* 31660 */ {(13<<2)|2,{56,107,0}},
/* 31661 */ {(13<<2)|2,{56,107,0}},
/* 31662 */ {(13<<2)|2,{56,107,0}},
/* 31663 */ {(13<<2)|2,{56,107,0}},
/* 31664 */ {(13<<2)|2,{56,113,0}},
/* 31665 */ {(13<<2)|2,{56,113,0}},
/* 31666 */ {(13<<2)|2,{56,113,0}},
/* 31667 */ {(13<<2)|2,{56,113,0}},
/* 31668 */ {(13<<2)|2,{56,113,0}},
/* 31669 */ {(13<<2)|2,{56,113,0}},
/* 31670 */ {(13<<2)|2,{56,113,0}},
/* 31671 */ {(13<<2)|2,{56,113,0}},
/* 31672 */ {(13<<2)|2,{56,118,0}},
/* 31673 */ {(13<<2)|2,{56,118,0}},
/* 31674 */ {(13<<2)|2,{56,118,0}},
/* 31675 */ {(13<<2)|2,{56,118,0}},
/* 31676 */ {(13<<2)|2,{56,118,0}},
/* 31677 */ {(13<<2)|2,{56,118,0}},
/* 31678 */ {(13<<2)|2,{56,118,0}},
/* 31679 */ {(13<<2)|2,{56,118,0}},
/* 31680 */ {(13<<2)|2,{56,119,0}},
/* 31681 */ {(13<<2)|2,{56,119,0}},
/* 31682 */ {(13<<2)|2,{56,119,0}},
/* 31683 */ {(13<<2)|2,{56,119,0}},
/* 31684 */ {(13<<2)|2,{56,119,0}},
/* 31685 */ {(13<<2)|2,{56,119,0}},
/* 31686 */ {(13<<2)|2,{56,119,0}},
/* 31687 */ {(13<<2)|2,{56,119,0}},
/* 31688 */ {(13<<2)|2,{56,120,0}},
/* 31689 */ {(13<<2)|2,{56,120,0}},
/* 31690 */ {(13<<2)|2,{56,120,0}},
/* 31691 */ {(13<<2)|2,{56,120,0}},
/* 31692 */ {(13<<2)|2,{56,120,0}},
/* 31693 */ {(13<<2)|2,{56,120,0}},
/* 31694 */ {(13<<2)|2,{56,120,0}},
/* 31695 */ {(13<<2)|2,{56,120,0}},
/* 31696 */ {(13<<2)|2,{56,121,0}},
/* 31697 */ {(13<<2)|2,{56,121,0}},
/* 31698 */ {(13<<2)|2,{56,121,0}},
/* 31699 */ {(13<<2)|2,{56,121,0}},
/* 31700 */ {(13<<2)|2,{56,121,0}},
/* 31701 */ {(13<<2)|2,{56,121,0}},
/* 31702 */ {(13<<2)|2,{56,121,0}},
/* 31703 */ {(13<<2)|2,{56,121,0}},
/* 31704 */ {(13<<2)|2,{56,122,0}},
/* 31705 */ {(13<<2)|2,{56,122,0}},
/* 31706 */ {(13<<2)|2,{56,122,0}},
/* 31707 */ {(13<<2)|2,{56,122,0}},
/* 31708 */ {(13<<2)|2,{56,122,0}},
/* 31709 */ {(13<<2)|2,{56,122,0}},
/* 31710 */ {(13<<2)|2,{56,122,0}},
/* 31711 */ {(13<<2)|2,{56,122,0}},
/* 31712 */ {(14<<2)|2,{56,38,0}},
/* 31713 */ {(14<<2)|2,{56,38,0}},
/* 31714 */ {(14<<2)|2,{56,38,0}},
/* 31715 */ {(14<<2)|2,{56,38,0}},
/* 31716 */ {(14<<2)|2,{56,42,0}},
/* 31717 */ {(14<<2)|2,{56,42,0}},
/* 31718 */ {(14<<2)|2,{56,42,0}},
/* 31719 */ {(14<<2)|2,{56,42,0}},
/* 31720 */ {(14<<2)|2,{56,44,0}},
/* 31721 */ {(14<<2)|2,{56,44,0}},
/* 31722 */ {(14<<2)|2,{56,44,0}},
/* 31723 */ {(14<<2)|2,{56,44,0}},
/* 31724 */ {(14<<2)|2,{56,59,0}},
/* 31725 */ {(14<<2)|2,{56,59,0}},
/* 31726 */ {(14<<2)|2,{56,59,0}},
/* 31727 */ {(14<<2)|2,{56,59,0}},
/* 31728 */ {(14<<2)|2,{56,88,0}},
/* 31729 */ {(14<<2)|2,{56,88,0}},
/* 31730 */ {(14<<2)|2,{56,88,0}},
/* 31731 */ {(14<<2)|2,{56,88,0}},
/* 31732 */ {(14<<2)|2,{56,90,0}},
/* 31733 */ {(14<<2)|2,{56,90,0}},
/* 31734 */ {(14<<2)|2,{56,90,0}},
/* 31735 */ {(14<<2)|2,{56,90,0}},
/* 31736 */ {(16<<2)|2,{56,33,0}},
/* 31737 */ {(16<<2)|2,{56,34,0}},
/* 31738 */ {(16<<2)|2,{56,40,0}},
/* 31739 */ {(16<<2)|2,{56,41,0}},
/* 31740 */ {(16<<2)|2,{56,63,0}},
/* 31741 */ {(6<<2)|1,{56,0,0}},
/* 31742 */ {(6<<2)|1,{56,0,0}},
/* 31743 */ {(6<<2)|1,{56,0,0}},
/* 31744 */ {(16<<2)|3,{57,48,48}},
/* 31745 */ {(16<<2)|3,{57,48,49}},
/* 31746 */ {(16<<2)|3,{57,48,50}},
/* 31747 */ {(16<<2)|3,{57,48,97}},
/* 31748 */ {(16<<2)|3,{57,48,99}},
/* 31749 */ {(16<<2)|3,{57,48,101}},
/* 31750 */ {(16<<2)|3,{57,48,105}},
/* 31751 */ {(16<<2)|3,{57,48,111}},
/* 31752 */ {(16<<2)|3,{57,48,115}},
/* 31753 */ {(16<<2)|3,{57,48,116}},
/* 31754 */ {(11<<2)|2,{57,48,0}},
/* 31755 */ {(11<<2)|2,{57,48,0}},
/* 31756 */ {(11<<2)|2,{57,48,0}},
/* 31757 */ {(11<<2)|2,{57,48,0}},
/* 31758 */ {(11<<2)|2,{57,48,0}},
/* 31759 */ {(11<<2)|2,{57,48,0}},
/* 31760 */ {(11<<2)|2,{57,48,0}},
/* 31761 */ {(11<<2)|2,{57,48,0}},
/* 31762 */ {(11<<2)|2,{57,48,0}},
/* 31763 */ {(11<<2)|2,{57,48,0}},
/* 31764 */ {(11<<2)|2,{57,48,0}},
/* 31765 */ {(11<<2)|2,{57,48,0}},
/* 31766 */ {(11<<2)|2,{57,48,0}},
/* 31767 */ {(11<<2)|2,{57,48,0}},
/* 31768 */ {(11<<2)|2,{57,48,0}},
/* 31769 */ {(11<<2)|2,{57,48,0}},
/* 31770 */ {(11<<2)|2,{57,48,0}},
/* 31771 */ {(11<<2)|2,{57,48,0}},
/* 31772 */ {(11<<2)|2,{57,48,0}},
/* 31773 */ {(11<<2)|2,{57,48,0}},
/* 31774 */ {(11<<2)|2,{57,48,0}},
/* 31775 */ {(11<<2)|2,{57,48,0}},
/* 31776 */ {(16<<2)|3,{57,49,48}},
/* 31777 */ {(16<<2)|3,{57,49,49}},
/* 31778 */ {(16<<2)|3,{57,49,50}},
/* 31779 */ {(16<<2)|3,{57,49,97}},
/* 31780 */ {(16<<2)|3,{57,49,99}},
/* 31781 */ {(16<<2)|3,{57,49,101}},
/* 31782 */ {(16<<2)|3,{57,49,105}},
/* 31783 */ {(16<<2)|3,{57,49,111}},
/* 31784 */ {(16<<2)|3,{57,49,115}},
/* 31785 */ {(16<<2)|3,{57,49,116}},
/* 31786 */ {(11<<2)|2,{57,49,0}},
/* 31787 */ {(11<<2)|2,{57,49,0}},
/* 31788 */ {(11<<2)|2,{57,49,0}},
/* 31789 */ {(11<<2)|2,{57,49,0}},
/* 31790 */ {(11<<2)|2,{57,49,0}},
/* 31791 */ {(11<<2)|2,{57,49,0}},
/* 31792 */ {(11<<2)|2,{57,49,0}},
/* 31793 */ {(11<<2)|2,{57,49,0}},
/* 31794 */ {(11<<2)|2,{57,49,0}},
/* 31795 */ {(11<<2)|2,{57,49,0}},
/* 31796 */ {(11<<2)|2,{57,49,0}},
/* 31797 */ {(11<<2)|2,{57,49,0}},
/* 31798 */ {(11<<2)|2,{57,49,0}},
/* 31799 */ {(11<<2)|2,{57,49,0}},
/* 31800 */ {(11<<2)|2,{57,49,0}},
/* 31801 */ {(11<<2)|2,{57,49,0}},
/* 31802 */ {(11<<2)|2,{57,49,0}},
/* 31803 */ {(11<<2)|2,{57,49,0}},
/* 31804 */ {(11<<2)|2,{57,49,0}},
/* 31805 */ {(11<<2)|2,{57,49,0}},
/* 31806 */ {(11<<2)|2,{57,49,0}},
/* 31807 */ {(11<<2)|2,{57,49,0}},
/* 31808 */ {(16<<2)|3,{57,50,48}},
/* 31809 */ {(16<<2)|3,{57,50,49}},
/* 31810 */ {(16<<2)|3,{57,50,50}},
/* 31811 */ {(16<<2)|3,{57,50,97}},
/* 31812 */ {(16<<2)|3,{57,50,99}},
/* 31813 */ {(16<<2)|3,{57,50,101}},
/* 31814 */ {(16<<2)|3,{57,50,105}},
/* 31815 */ {(16<<2)|3,{57,50,111}},
/* 31816 */ {(16<<2)|3,{57,50,115}},
/* 31817 */ {(16<<2)|3,{57,50,116}},
/* 31818 */ {(11<<2)|2,{57,50,0}},
/* 31819 */ {(11<<2)|2,{57,50,0}},
/* 31820 */ {(11<<2)|2,{57,50,0}},
/* 31821 */ {(11<<2)|2,{57,50,0}},
/* 31822 */ {(11<<2)|2,{57,50,0}},
/* 31823 */ {(11<<2)|2,{57,50,0}},
/* 31824 */ {(11<<2)|2,{57,50,0}},
/* 31825 */ {(11<<2)|2,{57,50,0}},
/* 31826 */ {(11<<2)|2,{57,50,0}},
/* 31827 */ {(11<<2)|2,{57,50,0}},
/* 31828 */ {(11<<2)|2,{57,50,0}},
/* 31829 */ {(11<<2)|2,{57,50,0}},
/* 31830 */ {(11<<2)|2,{57,50,0}},
/* 31831 */ {(11<<2)|2,{57,50,0}},
/* 31832 */ {(11<<2)|2,{57,50,0}},
/* 31833 */ {(11<<2)|2,{57,50,0}},
/* 31834 */ {(11<<2)|2,{57,50,0}},
/* 31835 */ {(11<<2)|2,{57,50,0}},
/* 31836 */ {(11<<2)|2,{57,50,0}},
/* 31837 */ {(11<<2)|2,{57,50,0}},
/* 31838 */ {(11<<2)|2,{57,50,0}},
/* 31839 */ {(11<<2)|2,{57,50,0}},
/* 31840 */ {(16<<2)|3,{57,97,48}},
/* 31841 */ {(16<<2)|3,{57,97,49}},
/* 31842 */ {(16<<2)|3,{57,97,50}},
/* 31843 */ {(16<<2)|3,{57,97,97}},
/* 31844 */ {(16<<2)|3,{57,97,99}},
/* 31845 */ {(16<<2)|3,{57,97,101}},
/* 31846 */ {(16<<2)|3,{57,97,105}},
/* 31847 */ {(16<<2)|3,{57,97,111}},
/* 31848 */ {(16<<2)|3,{57,97,115}},
/* 31849 */ {(16<<2)|3,{57,97,116}},
/* 31850 */ {(11<<2)|2,{57,97,0}},
/* 31851 */ {(11<<2)|2,{57,97,0}},
/* 31852 */ {(11<<2)|2,{57,97,0}},
/* 31853 */ {(11<<2)|2,{57,97,0}},
/* 31854 */ {(11<<2)|2,{57,97,0}},
/* 31855 */ {(11<<2)|2,{57,97,0}},
/* 31856 */ {(11<<2)|2,{57,97,0}},
/* 31857 */ {(11<<2)|2,{57,97,0}},
/* 31858 */ {(11<<2)|2,{57,97,0}},
/* 31859 */ {(11<<2)|2,{57,97,0}},
/* 31860 */ {(11<<2)|2,{57,97,0}},
/* 31861 */ {(11<<2)|2,{57,97,0}},
/* 31862 */ {(11<<2)|2,{57,97,0}},
/* 31863 */ {(11<<2)|2,{57,97,0}},
/* 31864 */ {(11<<2)|2,{57,97,0}},
/* 31865 */ {(11<<2)|2,{57,97,0}},
/* 31866 */ {(11<<2)|2,{57,97,0}},
/* 31867 */ {(11<<2)|2,{57,97,0}},
/* 31868 */ {(11<<2)|2,{57,97,0}},
/* 31869 */ {(11<<2)|2,{57,97,0}},
/* 31870 */ {(11<<2)|2,{57,97,0}},
/* 31871 */ {(11<<2)|2,{57,97,0}},
/* 31872 */ {(16<<2)|3,{57,99,48}},
/* 31873 */ {(16<<2)|3,{57,99,49}},
/* 31874 */ {(16<<2)|3,{57,99,50}},
/* 31875 */ {(16<<2)|3,{57,99,97}},
/* 31876 */ {(16<<2)|3,{57,99,99}},
/* 31877 */ {(16<<2)|3,{57,99,101}},
/* 31878 */ {(16<<2)|3,{57,99,105}},
/* 31879 */ {(16<<2)|3,{57,99,111}},
/* 31880 */ {(16<<2)|3,{57,99,115}},
/* 31881 */ {(16<<2)|3,{57,99,116}},
/* 31882 */ {(11<<2)|2,{57,99,0}},
/* 31883 */ {(11<<2)|2,{57,99,0}},
/* 31884 */ {(11<<2)|2,{57,99,0}},
/* 31885 */ {(11<<2)|2,{57,99,0}},
/* 31886 */ {(11<<2)|2,{57,99,0}},
/* 31887 */ {(11<<2)|2,{57,99,0}},
/* 31888 */ {(11<<2)|2,{57,99,0}},
/* 31889 */ {(11<<2)|2,{57,99,0}},
/* 31890 */ {(11<<2)|2,{57,99,0}},
/* 31891 */ {(11<<2)|2,{57,99,0}},
/* 31892 */ {(11<<2)|2,{57,99,0}},
/* 31893 */ {(11<<2)|2,{57,99,0}},
/* 31894 */ {(11<<2)|2,{57,99,0}},
/* 31895 */ {(11<<2)|2,{57,99,0}},
/* 31896 */ {(11<<2)|2,{57,99,0}},
/* 31897 */ {(11<<2)|2,{57,99,0}},
/* 31898 */ {(11<<2)|2,{57,99,0}},
/* 31899 */ {(11<<2)|2,{57,99,0}},
/* 31900 */ {(11<<2)|2,{57,99,0}},
/* 31901 */ {(11<<2)|2,{57,99,0}},
/* 31902 */ {(11<<2)|2,{57,99,0}},
/* 31903 */ {(11<<2)|2,{57,99,0}},
/* 31904 */ {(16<<2)|3,{57,101,48}},
/* 31905 */ {(16<<2)|3,{57,101,49}},
/* 31906 */ {(16<<2)|3,{57,101,50}},
/* 31907 */ {(16<<2)|3,{57,101,97}},
/* 31908 */ {(16<<2)|3,{57,101,99}},
/* 31909 */ {(16<<2)|3,{57,101,101}},
/* 31910 */ {(16<<2)|3,{57,101,105}},
/* 31911 */ {(16<<2)|3,{57,101,111}},
/* 31912 */ {(16<<2)|3,{57,101,115}},
/* 31913 */ {(16<<2)|3,{57,101,116}},
/* 31914 */ {(11<<2)|2,{57,101,0}},
/* 31915 */ {(11<<2)|2,{57,101,0}},
/* 31916 */ {(11<<2)|2,{57,101,0}},
/* 31917 */ {(11<<2)|2,{57,101,0}},
/* 31918 */ {(11<<2)|2,{57,101,0}},
/* 31919 */ {(11<<2)|2,{57,101,0}},
/* 31920 */ {(11<<2)|2,{57,101,0}},
/* 31921 */ {(11<<2)|2,{57,101,0}},
/* 31922 */ {(11<<2)|2,{57,101,0}},
/* 31923 */ {(11<<2)|2,{57,101,0}},
/* 31924 */ {(11<<2)|2,{57,101,0}},
/* 31925 */ {(11<<2)|2,{57,101,0}},
/* 31926 */ {(11<<2)|2,{57,101,0}},
/* 31927 */ {(11<<2)|2,{57,101,0}},
/* 31928 */ {(11<<2)|2,{57,101,0}},
/* 31929 */ {(11<<2)|2,{57,101,0}},
/* 31930 */ {(11<<2)|2,{57,101,0}},
/* 31931 */ {(11<<2)|2,{57,101,0}},
/* 31932 */ {(11<<2)|2,{57,101,0}},
/* 31933 */ {(11<<2)|2,{57,101,0}},
/* 31934 */ {(11<<2)|2,{57,101,0}},
/* 31935 */ {(11<<2)|2,{57,101,0}},
/* 31936 */ {(16<<2)|3,{57,105,48}},
/* 31937 */ {(16<<2)|3,{57,105,49}},
/* 31938 */ {(16<<2)|3,{57,105,50}},
/* 31939 */ {(16<<2)|3,{57,105,97}},
/* 31940 */ {(16<<2)|3,{57,105,99}},
/* 31941 */ {(16<<2)|3,{57,105,101}},
/* 31942 */ {(16<<2)|3,{57,105,105}},
/* 31943 */ {(16<<2)|3,{57,105,111}},
/* 31944 */ {(16<<2)|3,{57,105,115}},
/* 31945 */ {(16<<2)|3,{57,105,116}},
/* 31946 */ {(11<<2)|2,{57,105,0}},
/* 31947 */ {(11<<2)|2,{57,105,0}},
/* 31948 */ {(11<<2)|2,{57,105,0}},
/* 31949 */ {(11<<2)|2,{57,105,0}},
/* 31950 */ {(11<<2)|2,{57,105,0}},
/* 31951 */ {(11<<2)|2,{57,105,0}},
/* 31952 */ {(11<<2)|2,{57,105,0}},
/* 31953 */ {(11<<2)|2,{57,105,0}},
/* 31954 */ {(11<<2)|2,{57,105,0}},
/* 31955 */ {(11<<2)|2,{57,105,0}},
/* 31956 */ {(11<<2)|2,{57,105,0}},
/* 31957 */ {(11<<2)|2,{57,105,0}},
/* 31958 */ {(11<<2)|2,{57,105,0}},
/* 31959 */ {(11<<2)|2,{57,105,0}},
/* 31960 */ {(11<<2)|2,{57,105,0}},
/* 31961 */ {(11<<2)|2,{57,105,0}},
/* 31962 */ {(11<<2)|2,{57,105,0}},
/* 31963 */ {(11<<2)|2,{57,105,0}},
/* 31964 */ {(11<<2)|2,{57,105,0}},
/* 31965 */ {(11<<2)|2,{57,105,0}},
/* 31966 */ {(11<<2)|2,{57,105,0}},
/* 31967 */ {(11<<2)|2,{57,105,0}},
/* 31968 */ {(16<<2)|3,{57,111,48}},
/* 31969 */ {(16<<2)|3,{57,111,49}},
/* 31970 */ {(16<<2)|3,{57,111,50}},
/* 31971 */ {(16<<2)|3,{57,111,97}},
/* 31972 */ {(16<<2)|3,{57,111,99}},
/* 31973 */ {(16<<2)|3,{57,111,101}},
/* 31974 */ {(16<<2)|3,{57,111,105}},
/* 31975 */ {(16<<2)|3,{57,111,111}},
/* 31976 */ {(16<<2)|3,{57,111,115}},
/* 31977 */ {(16<<2)|3,{57,111,116}},
/* 31978 */ {(11<<2)|2,{57,111,0}},
/* 31979 */ {(11<<2)|2,{57,111,0}},
/* 31980 */ {(11<<2)|2,{57,111,0}},
/* 31981 */ {(11<<2)|2,{57,111,0}},
/* 31982 */ {(11<<2)|2,{57,111,0}},
/* 31983 */ {(11<<2)|2,{57,111,0}},
/* 31984 */ {(11<<2)|2,{57,111,0}},
/* 31985 */ {(11<<2)|2,{57,111,0}},
/* 31986 */ {(11<<2)|2,{57,111,0}},
/* 31987 */ {(11<<2)|2,{57,111,0}},
/* 31988 */ {(11<<2)|2,{57,111,0}},
/* 31989 */ {(11<<2)|2,{57,111,0}},
/* 31990 */ {(11<<2)|2,{57,111,0}},
/* 31991 */ {(11<<2)|2,{57,111,0}},
/* 31992 */ {(11<<2)|2,{57,111,0}},
/* 31993 */ {(11<<2)|2,{57,111,0}},
/* 31994 */ {(11<<2)|2,{57,111,0}},
/* 31995 */ {(11<<2)|2,{57,111,0}},
/* 31996 */ {(11<<2)|2,{57,111,0}},
/* 31997 */ {(11<<2)|2,{57,111,0}},
/* 31998 */ {(11<<2)|2,{57,111,0}},
/* 31999 */ {(11<<2)|2,{57,111,0}},
/* 32000 */ {(16<<2)|3,{57,115,48}},
/* 32001 */ {(16<<2)|3,{57,115,49}},
/* 32002 */ {(16<<2)|3,{57,115,50}},
/* 32003 */ {(16<<2)|3,{57,115,97}},
/* 32004 */ {(16<<2)|3,{57,115,99}},
/* 32005 */ {(16<<2)|3,{57,115,101}},
/* 32006 */ {(16<<2)|3,{57,115,105}},
/* 32007 */ {(16<<2)|3,{57,115,111}},
/* 32008 */ {(16<<2)|3,{57,115,115}},
/* 32009 */ {(16<<2)|3,{57,115,116}},
/* 32010 */ {(11<<2)|2,{57,115,0}},
/* 32011 */ {(11<<2)|2,{57,115,0}},
/* 32012 */ {(11<<2)|2,{57,115,0}},
/* 32013 */ {(11<<2)|2,{57,115,0}},
/* 32014 */ {(11<<2)|2,{57,115,0}},
/* 32015 */ {(11<<2)|2,{57,115,0}},
/* 32016 */ {(11<<2)|2,{57,115,0}},
/* 32017 */ {(11<<2)|2,{57,115,0}},
/* 32018 */ {(11<<2)|2,{57,115,0}},
/* 32019 */ {(11<<2)|2,{57,115,0}},
/* 32020 */ {(11<<2)|2,{57,115,0}},
/* 32021 */ {(11<<2)|2,{57,115,0}},
/* 32022 */ {(11<<2)|2,{57,115,0}},
/* 32023 */ {(11<<2)|2,{57,115,0}},
/* 32024 */ {(11<<2)|2,{57,115,0}},
/* 32025 */ {(11<<2)|2,{57,115,0}},
/* 32026 */ {(11<<2)|2,{57,115,0}},
/* 32027 */ {(11<<2)|2,{57,115,0}},
/* 32028 */ {(11<<2)|2,{57,115,0}},
/* 32029 */ {(11<<2)|2,{57,115,0}},
/* 32030 */ {(11<<2)|2,{57,115,0}},
/* 32031 */ {(11<<2)|2,{57,115,0}},
/* 32032 */ {(16<<2)|3,{57,116,48}},
/* 32033 */ {(16<<2)|3,{57,116,49}},
/* 32034 */ {(16<<2)|3,{57,116,50}},
/* 32035 */ {(16<<2)|3,{57,116,97}},
/* 32036 */ {(16<<2)|3,{57,116,99}},
/* 32037 */ {(16<<2)|3,{57,116,101}},
/* 32038 */ {(16<<2)|3,{57,116,105}},
/* 32039 */ {(16<<2)|3,{57,116,111}},
/* 32040 */ {(16<<2)|3,{57,116,115}},
/* 32041 */ {(16<<2)|3,{57,116,116}},
/* 32042 */ {(11<<2)|2,{57,116,0}},
/* 32043 */ {(11<<2)|2,{57,116,0}},
/* 32044 */ {(11<<2)|2,{57,116,0}},
/* 32045 */ {(11<<2)|2,{57,116,0}},
/* 32046 */ {(11<<2)|2,{57,116,0}},
/* 32047 */ {(11<<2)|2,{57,116,0}},
/* 32048 */ {(11<<2)|2,{57,116,0}},
/* 32049 */ {(11<<2)|2,{57,116,0}},
/* 32050 */ {(11<<2)|2,{57,116,0}},
/* 32051 */ {(11<<2)|2,{57,116,0}},
/* 32052 */ {(11<<2)|2,{57,116,0}},
/* 32053 */ {(11<<2)|2,{57,116,0}},
/* 32054 */ {(11<<2)|2,{57,116,0}},
/* 32055 */ {(11<<2)|2,{57,116,0}},
/* 32056 */ {(11<<2)|2,{57,116,0}},
/* 32057 */ {(11<<2)|2,{57,116,0}},
/* 32058 */ {(11<<2)|2,{57,116,0}},
/* 32059 */ {(11<<2)|2,{57,116,0}},
/* 32060 */ {(11<<2)|2,{57,116,0}},
/* 32061 */ {(11<<2)|2,{57,116,0}},
/* 32062 */ {(11<<2)|2,{57,116,0}},
/* 32063 */ {(11<<2)|2,{57,116,0}},
/* 32064 */ {(12<<2)|2,{57,32,0}},
/* 32065 */ {(12<<2)|2,{57,32,0}},
/* 32066 */ {(12<<2)|2,{57,32,0}},
/* 32067 */ {(12<<2)|2,{57,32,0}},
/* 32068 */ {(12<<2)|2,{57,32,0}},
/* 32069 */ {(12<<2)|2,{57,32,0}},
/* 32070 */ {(12<<2)|2,{57,32,0}},
/* 32071 */ {(12<<2)|2,{57,32,0}},
/* 32072 */ {(12<<2)|2,{57,32,0}},
/* 32073 */ {(12<<2)|2,{57,32,0}},
/* 32074 */ {(12<<2)|2,{57,32,0}},
/* 32075 */ {(12<<2)|2,{57,32,0}},
/* 32076 */ {(12<<2)|2,{57,32,0}},
/* 32077 */ {(12<<2)|2,{57,32,0}},
/* 32078 */ {(12<<2)|2,{57,32,0}},
/* 32079 */ {(12<<2)|2,{57,32,0}},
/* 32080 */ {(12<<2)|2,{57,37,0}},
/* 32081 */ {(12<<2)|2,{57,37,0}},
/* 32082 */ {(12<<2)|2,{57,37,0}},
/* 32083 */ {(12<<2)|2,{57,37,0}},
/* 32084 */ {(12<<2)|2,{57,37,0}},
/* 32085 */ {(12<<2)|2,{57,37,0}},
/* 32086 */ {(12<<2)|2,{57,37,0}},
/* 32087 */ {(12<<2)|2,{57,37,0}},
/* 32088 */ {(12<<2)|2,{57,37,0}},
/* 32089 */ {(12<<2)|2,{57,37,0}},
/* 32090 */ {(12<<2)|2,{57,37,0}},
/* 32091 */ {(12<<2)|2,{57,37,0}},
/* 32092 */ {(12<<2)|2,{57,37,0}},
/* 32093 */ {(12<<2)|2,{57,37,0}},
/* 32094 */ {(12<<2)|2,{57,37,0}},
/* 32095 */ {(12<<2)|2,{57,37,0}},
/* 32096 */ {(12<<2)|2,{57,45,0}},
/* 32097 */ {(12<<2)|2,{57,45,0}},
/* 32098 */ {(12<<2)|2,{57,45,0}},
/* 32099 */ {(12<<2)|2,{57,45,0}},
/* 32100 */ {(12<<2)|2,{57,45,0}},
/* 32101 */ {(12<<2)|2,{57,45,0}},
/* 32102 */ {(12<<2)|2,{57,45,0}},
/* 32103 */ {(12<<2)|2,{57,45,0}},
/* 32104 */ {(12<<2)|2,{57,45,0}},
/* 32105 */ {(12<<2)|2,{57,45,0}},
/* 32106 */ {(12<<2)|2,{57,45,0}},
/* 32107 */ {(12<<2)|2,{57,45,0}},
/* 32108 */ {(12<<2)|2,{57,45,0}},
/* 32109 */ {(12<<2)|2,{57,45,0}},
/* 32110 */ {(12<<2)|2,{57,45,0}},
/* 32111 */ {(12<<2)|2,{57,45,0}},
/* 32112 */ {(12<<2)|2,{57,46,0}},
/* 32113 */ {(12<<2)|2,{57,46,0}},
/* 32114 */ {(12<<2)|2,{57,46,0}},
/* 32115 */ {(12<<2)|2,{57,46,0}},
/* 32116 */ {(12<<2)|2,{57,46,0}},
/* 32117 */ {(12<<2)|2,{57,46,0}},
/* 32118 */ {(12<<2)|2,{57,46,0}},
/* 32119 */ {(12<<2)|2,{57,46,0}},
/* 32120 */ {(12<<2)|2,{57,46,0}},
/* 32121 */ {(12<<2)|2,{57,46,0}},
/* 32122 */ {(12<<2)|2,{57,46,0}},
/* 32123 */ {(12<<2)|2,{57,46,0}},
/* 32124 */ {(12<<2)|2,{57,46,0}},
/* 32125 */ {(12<<2)|2,{57,46,0}},
/* 32126 */ {(12<<2)|2,{57,46,0}},
/* 32127 */ {(12<<2)|2,{57,46,0}},
/* 32128 */ {(12<<2)|2,{57,47,0}},
/* 32129 */ {(12<<2)|2,{57,47,0}},
/* 32130 */ {(12<<2)|2,{57,47,0}},
/* 32131 */ {(12<<2)|2,{57,47,0}},
/* 32132 */ {(12<<2)|2,{57,47,0}},
/* 32133 */ {(12<<2)|2,{57,47,0}},
/* 32134 */ {(12<<2)|2,{57,47,0}},
/* 32135 */ {(12<<2)|2,{57,47,0}},
/* 32136 */ {(12<<2)|2,{57,47,0}},
/* 32137 */ {(12<<2)|2,{57,47,0}},
/* 32138 */ {(12<<2)|2,{57,47,0}},
/* 32139 */ {(12<<2)|2,{57,47,0}},
/* 32140 */ {(12<<2)|2,{57,47,0}},
/* 32141 */ {(12<<2)|2,{57,47,0}},
/* 32142 */ {(12<<2)|2,{57,47,0}},
/* 32143 */ {(12<<2)|2,{57,47,0}},
/* 32144 */ {(12<<2)|2,{57,51,0}},
/* 32145 */ {(12<<2)|2,{57,51,0}},
/* 32146 */ {(12<<2)|2,{57,51,0}},
/* 32147 */ {(12<<2)|2,{57,51,0}},
/* 32148 */ {(12<<2)|2,{57,51,0}},
/* 32149 */ {(12<<2)|2,{57,51,0}},
/* 32150 */ {(12<<2)|2,{57,51,0}},
/* 32151 */ {(12<<2)|2,{57,51,0}},
/* 32152 */ {(12<<2)|2,{57,51,0}},
/* 32153 */ {(12<<2)|2,{57,51,0}},
/* 32154 */ {(12<<2)|2,{57,51,0}},
/* 32155 */ {(12<<2)|2,{57,51,0}},
/* 32156 */ {(12<<2)|2,{57,51,0}},
/* 32157 */ {(12<<2)|2,{57,51,0}},
/* 32158 */ {(12<<2)|2,{57,51,0}},
/* 32159 */ {(12<<2)|2,{57,51,0}},
/* 32160 */ {(12<<2)|2,{57,52,0}},
/* 32161 */ {(12<<2)|2,{57,52,0}},
/* 32162 */ {(12<<2)|2,{57,52,0}},
/* 32163 */ {(12<<2)|2,{57,52,0}},
/* 32164 */ {(12<<2)|2,{57,52,0}},
/* 32165 */ {(12<<2)|2,{57,52,0}},
/* 32166 */ {(12<<2)|2,{57,52,0}},
/* 32167 */ {(12<<2)|2,{57,52,0}},
/* 32168 */ {(12<<2)|2,{57,52,0}},
/* 32169 */ {(12<<2)|2,{57,52,0}},
/* 32170 */ {(12<<2)|2,{57,52,0}},
/* 32171 */ {(12<<2)|2,{57,52,0}},
/* 32172 */ {(12<<2)|2,{57,52,0}},
/* 32173 */ {(12<<2)|2,{57,52,0}},
/* 32174 */ {(12<<2)|2,{57,52,0}},
/* 32175 */ {(12<<2)|2,{57,52,0}},
/* 32176 */ {(12<<2)|2,{57,53,0}},
/* 32177 */ {(12<<2)|2,{57,53,0}},
/* 32178 */ {(12<<2)|2,{57,53,0}},
/* 32179 */ {(12<<2)|2,{57,53,0}},
/* 32180 */ {(12<<2)|2,{57,53,0}},
/* 32181 */ {(12<<2)|2,{57,53,0}},
/* 32182 */ {(12<<2)|2,{57,53,0}},
/* 32183 */ {(12<<2)|2,{57,53,0}},
/* 32184 */ {(12<<2)|2,{57,53,0}},
/* 32185 */ {(12<<2)|2,{57,53,0}},
/* 32186 */ {(12<<2)|2,{57,53,0}},
/* 32187 */ {(12<<2)|2,{57,53,0}},
/* 32188 */ {(12<<2)|2,{57,53,0}},
/* 32189 */ {(12<<2)|2,{57,53,0}},
/* 32190 */ {(12<<2)|2,{57,53,0}},
/* 32191 */ {(12<<2)|2,{57,53,0}},
/* 32192 */ {(12<<2)|2,{57,54,0}},
/* 32193 */ {(12<<2)|2,{57,54,0}},
/* 32194 */ {(12<<2)|2,{57,54,0}},
/* 32195 */ {(12<<2)|2,{57,54,0}},
/* 32196 */ {(12<<2)|2,{57,54,0}},
/* 32197 */ {(12<<2)|2,{57,54,0}},
/* 32198 */ {(12<<2)|2,{57,54,0}},
/* 32199 */ {(12<<2)|2,{57,54,0}},
/* 32200 */ {(12<<2)|2,{57,54,0}},
/* 32201 */ {(12<<2)|2,{57,54,0}},
/* 32202 */ {(12<<2)|2,{57,54,0}},
/* 32203 */ {(12<<2)|2,{57,54,0}},
/* 32204 */ {(12<<2)|2,{57,54,0}},
/* 32205 */ {(12<<2)|2,{57,54,0}},
/* 32206 */ {(12<<2)|2,{57,54,0}},
/* 32207 */ {(12<<2)|2,{57,54,0}},
/* 32208 */ {(12<<2)|2,{57,55,0}},
/* 32209 */ {(12<<2)|2,{57,55,0}},
/* 32210 */ {(12<<2)|2,{57,55,0}},
/* 32211 */ {(12<<2)|2,{57,55,0}},
/* 32212 */ {(12<<2)|2,{57,55,0}},
/* 32213 */ {(12<<2)|2,{57,55,0}},
/* 32214 */ {(12<<2)|2,{57,55,0}},
/* 32215 */ {(12<<2)|2,{57,55,0}},
/* 32216 */ {(12<<2)|2,{57,55,0}},
/* 32217 */ {(12<<2)|2,{57,55,0}},
/* 32218 */ {(12<<2)|2,{57,55,0}},
/* 32219 */ {(12<<2)|2,{57,55,0}},
/* 32220 */ {(12<<2)|2,{57,55,0}},
/* 32221 */ {(12<<2)|2,{57,55,0}},
/* 32222 */ {(12<<2)|2,{57,55,0}},
/* 32223 */ {(12<<2)|2,{57,55,0}},
/* 32224 */ {(12<<2)|2,{57,56,0}},
/* 32225 */ {(12<<2)|2,{57,56,0}},
/* 32226 */ {(12<<2)|2,{57,56,0}},
/* 32227 */ {(12<<2)|2,{57,56,0}},
/* 32228 */ {(12<<2)|2,{57,56,0}},
/* 32229 */ {(12<<2)|2,{57,56,0}},
/* 32230 */ {(12<<2)|2,{57,56,0}},
/* 32231 */ {(12<<2)|2,{57,56,0}},
/* 32232 */ {(12<<2)|2,{57,56,0}},
/* 32233 */ {(12<<2)|2,{57,56,0}},
/* 32234 */ {(12<<2)|2,{57,56,0}},
/* 32235 */ {(12<<2)|2,{57,56,0}},
/* 32236 */ {(12<<2)|2,{57,56,0}},
/* 32237 */ {(12<<2)|2,{57,56,0}},
/* 32238 */ {(12<<2)|2,{57,56,0}},
/* 32239 */ {(12<<2)|2,{57,56,0}},
/* 32240 */ {(12<<2)|2,{57,57,0}},
/* 32241 */ {(12<<2)|2,{57,57,0}},
/* 32242 */ {(12<<2)|2,{57,57,0}},
/* 32243 */ {(12<<2)|2,{57,57,0}},
/* 32244 */ {(12<<2)|2,{57,57,0}},
/* 32245 */ {(12<<2)|2,{57,57,0}},
/* 32246 */ {(12<<2)|2,{57,57,0}},
/* 32247 */ {(12<<2)|2,{57,57,0}},
/* 32248 */ {(12<<2)|2,{57,57,0}},
/* 32249 */ {(12<<2)|2,{57,57,0}},
/* 32250 */ {(12<<2)|2,{57,57,0}},
/* 32251 */ {(12<<2)|2,{57,57,0}},
/* 32252 */ {(12<<2)|2,{57,57,0}},
/* 32253 */ {(12<<2)|2,{57,57,0}},
/* 32254 */ {(12<<2)|2,{57,57,0}},
/* 32255 */ {(12<<2)|2,{57,57,0}},
/* 32256 */ {(12<<2)|2,{57,61,0}},
/* 32257 */ {(12<<2)|2,{57,61,0}},
/* 32258 */ {(12<<2)|2,{57,61,0}},
/* 32259 */ {(12<<2)|2,{57,61,0}},
/* 32260 */ {(12<<2)|2,{57,61,0}},
/* 32261 */ {(12<<2)|2,{57,61,0}},
/* 32262 */ {(12<<2)|2,{57,61,0}},
/* 32263 */ {(12<<2)|2,{57,61,0}},
/* 32264 */ {(12<<2)|2,{57,61,0}},
/* 32265 */ {(12<<2)|2,{57,61,0}},
/* 32266 */ {(12<<2)|2,{57,61,0}},
/* 32267 */ {(12<<2)|2,{57,61,0}},
/* 32268 */ {(12<<2)|2,{57,61,0}},
/* 32269 */ {(12<<2)|2,{57,61,0}},
/* 32270 */ {(12<<2)|2,{57,61,0}},
/* 32271 */ {(12<<2)|2,{57,61,0}},
/* 32272 */ {(12<<2)|2,{57,65,0}},
/* 32273 */ {(12<<2)|2,{57,65,0}},
/* 32274 */ {(12<<2)|2,{57,65,0}},
/* 32275 */ {(12<<2)|2,{57,65,0}},
/* 32276 */ {(12<<2)|2,{57,65,0}},
/* 32277 */ {(12<<2)|2,{57,65,0}},
/* 32278 */ {(12<<2)|2,{57,65,0}},
/* 32279 */ {(12<<2)|2,{57,65,0}},
/* 32280 */ {(12<<2)|2,{57,65,0}},
/* 32281 */ {(12<<2)|2,{57,65,0}},
/* 32282 */ {(12<<2)|2,{57,65,0}},
/* 32283 */ {(12<<2)|2,{57,65,0}},
/* 32284 */ {(12<<2)|2,{57,65,0}},
/* 32285 */ {(12<<2)|2,{57,65,0}},
/* 32286 */ {(12<<2)|2,{57,65,0}},
/* 32287 */ {(12<<2)|2,{57,65,0}},
/* 32288 */ {(12<<2)|2,{57,95,0}},
/* 32289 */ {(12<<2)|2,{57,95,0}},
/* 32290 */ {(12<<2)|2,{57,95,0}},
/* 32291 */ {(12<<2)|2,{57,95,0}},
/* 32292 */ {(12<<2)|2,{57,95,0}},
/* 32293 */ {(12<<2)|2,{57,95,0}},
/* 32294 */ {(12<<2)|2,{57,95,0}},
/* 32295 */ {(12<<2)|2,{57,95,0}},
/* 32296 */ {(12<<2)|2,{57,95,0}},
/* 32297 */ {(12<<2)|2,{57,95,0}},
/* 32298 */ {(12<<2)|2,{57,95,0}},
/* 32299 */ {(12<<2)|2,{57,95,0}},
/* 32300 */ {(12<<2)|2,{57,95,0}},
/* 32301 */ {(12<<2)|2,{57,95,0}},
/* 32302 */ {(12<<2)|2,{57,95,0}},
/* 32303 */ {(12<<2)|2,{57,95,0}},
/* 32304 */ {(12<<2)|2,{57,98,0}},
/* 32305 */ {(12<<2)|2,{57,98,0}},
/* 32306 */ {(12<<2)|2,{57,98,0}},
/* 32307 */ {(12<<2)|2,{57,98,0}},
/* 32308 */ {(12<<2)|2,{57,98,0}},
/* 32309 */ {(12<<2)|2,{57,98,0}},
/* 32310 */ {(12<<2)|2,{57,98,0}},
/* 32311 */ {(12<<2)|2,{57,98,0}},
/* 32312 */ {(12<<2)|2,{57,98,0}},
/* 32313 */ {(12<<2)|2,{57,98,0}},
/* 32314 */ {(12<<2)|2,{57,98,0}},
/* 32315 */ {(12<<2)|2,{57,98,0}},
/* 32316 */ {(12<<2)|2,{57,98,0}},
/* 32317 */ {(12<<2)|2,{57,98,0}},
/* 32318 */ {(12<<2)|2,{57,98,0}},
/* 32319 */ {(12<<2)|2,{57,98,0}},
/* 32320 */ {(12<<2)|2,{57,100,0}},
/* 32321 */ {(12<<2)|2,{57,100,0}},
/* 32322 */ {(12<<2)|2,{57,100,0}},
/* 32323 */ {(12<<2)|2,{57,100,0}},
/* 32324 */ {(12<<2)|2,{57,100,0}},
/* 32325 */ {(12<<2)|2,{57,100,0}},
/* 32326 */ {(12<<2)|2,{57,100,0}},
/* 32327 */ {(12<<2)|2,{57,100,0}},
/* 32328 */ {(12<<2)|2,{57,100,0}},
/* 32329 */ {(12<<2)|2,{57,100,0}},
/* 32330 */ {(12<<2)|2,{57,100,0}},
/* 32331 */ {(12<<2)|2,{57,100,0}},
/* 32332 */ {(12<<2)|2,{57,100,0}},
/* 32333 */ {(12<<2)|2,{57,100,0}},
/* 32334 */ {(12<<2)|2,{57,100,0}},
/* 32335 */ {(12<<2)|2,{57,100,0}},
/* 32336 */ {(12<<2)|2,{57,102,0}},
/* 32337 */ {(12<<2)|2,{57,102,0}},
/* 32338 */ {(12<<2)|2,{57,102,0}},
/* 32339 */ {(12<<2)|2,{57,102,0}},
/* 32340 */ {(12<<2)|2,{57,102,0}},
/* 32341 */ {(12<<2)|2,{57,102,0}},
/* 32342 */ {(12<<2)|2,{57,102,0}},
/* 32343 */ {(12<<2)|2,{57,102,0}},
/* 32344 */ {(12<<2)|2,{57,102,0}},
/* 32345 */ {(12<<2)|2,{57,102,0}},
/* 32346 */ {(12<<2)|2,{57,102,0}},
/* 32347 */ {(12<<2)|2,{57,102,0}},
/* 32348 */ {(12<<2)|2,{57,102,0}},
/* 32349 */ {(12<<2)|2,{57,102,0}},
/* 32350 */ {(12<<2)|2,{57,102,0}},
/* 32351 */ {(12<<2)|2,{57,102,0}},
/* 32352 */ {(12<<2)|2,{57,103,0}},
/* 32353 */ {(12<<2)|2,{57,103,0}},
/* 32354 */ {(12<<2)|2,{57,103,0}},
/* 32355 */ {(12<<2)|2,{57,103,0}},
/* 32356 */ {(12<<2)|2,{57,103,0}},
/* 32357 */ {(12<<2)|2,{57,103,0}},
/* 32358 */ {(12<<2)|2,{57,103,0}},
/* 32359 */ {(12<<2)|2,{57,103,0}},
/* 32360 */ {(12<<2)|2,{57,103,0}},
/* 32361 */ {(12<<2)|2,{57,103,0}},
/* 32362 */ {(12<<2)|2,{57,103,0}},
/* 32363 */ {(12<<2)|2,{57,103,0}},
/* 32364 */ {(12<<2)|2,{57,103,0}},
/* 32365 */ {(12<<2)|2,{57,103,0}},
/* 32366 */ {(12<<2)|2,{57,103,0}},
/* 32367 */ {(12<<2)|2,{57,103,0}},
/* 32368 */ {(12<<2)|2,{57,104,0}},
/* 32369 */ {(12<<2)|2,{57,104,0}},
/* 32370 */ {(12<<2)|2,{57,104,0}},
/* 32371 */ {(12<<2)|2,{57,104,0}},
/* 32372 */ {(12<<2)|2,{57,104,0}},
/* 32373 */ {(12<<2)|2,{57,104,0}},
/* 32374 */ {(12<<2)|2,{57,104,0}},
/* 32375 */ {(12<<2)|2,{57,104,0}},
/* 32376 */ {(12<<2)|2,{57,104,0}},
/* 32377 */ {(12<<2)|2,{57,104,0}},
/* 32378 */ {(12<<2)|2,{57,104,0}},
/* 32379 */ {(12<<2)|2,{57,104,0}},
/* 32380 */ {(12<<2)|2,{57,104,0}},
/* 32381 */ {(12<<2)|2,{57,104,0}},
/* 32382 */ {(12<<2)|2,{57,104,0}},
/* 32383 */ {(12<<2)|2,{57,104,0}},
/* 32384 */ {(12<<2)|2,{57,108,0}},
/* 32385 */ {(12<<2)|2,{57,108,0}},
/* 32386 */ {(12<<2)|2,{57,108,0}},
/* 32387 */ {(12<<2)|2,{57,108,0}},
/* 32388 */ {(12<<2)|2,{57,108,0}},
/* 32389 */ {(12<<2)|2,{57,108,0}},
/* 32390 */ {(12<<2)|2,{57,108,0}},
/* 32391 */ {(12<<2)|2,{57,108,0}},
/* 32392 */ {(12<<2)|2,{57,108,0}},
/* 32393 */ {(12<<2)|2,{57,108,0}},
/* 32394 */ {(12<<2)|2,{57,108,0}},
/* 32395 */ {(12<<2)|2,{57,108,0}},
/* 32396 */ {(12<<2)|2,{57,108,0}},
/* 32397 */ {(12<<2)|2,{57,108,0}},
/* 32398 */ {(12<<2)|2,{57,108,0}},
/* 32399 */ {(12<<2)|2,{57,108,0}},
/* 32400 */ {(12<<2)|2,{57,109,0}},
/* 32401 */ {(12<<2)|2,{57,109,0}},
/* 32402 */ {(12<<2)|2,{57,109,0}},
/* 32403 */ {(12<<2)|2,{57,109,0}},
/* 32404 */ {(12<<2)|2,{57,109,0}},
/* 32405 */ {(12<<2)|2,{57,109,0}},
/* 32406 */ {(12<<2)|2,{57,109,0}},
/* 32407 */ {(12<<2)|2,{57,109,0}},
/* 32408 */ {(12<<2)|2,{57,109,0}},
/* 32409 */ {(12<<2)|2,{57,109,0}},
/* 32410 */ {(12<<2)|2,{57,109,0}},
/* 32411 */ {(12<<2)|2,{57,109,0}},
/* 32412 */ {(12<<2)|2,{57,109,0}},
/* 32413 */ {(12<<2)|2,{57,109,0}},
/* 32414 */ {(12<<2)|2,{57,109,0}},
/* 32415 */ {(12<<2)|2,{57,109,0}},
/* 32416 */ {(12<<2)|2,{57,110,0}},
/* 32417 */ {(12<<2)|2,{57,110,0}},
/* 32418 */ {(12<<2)|2,{57,110,0}},
/* 32419 */ {(12<<2)|2,{57,110,0}},
/* 32420 */ {(12<<2)|2,{57,110,0}},
/* 32421 */ {(12<<2)|2,{57,110,0}},
/* 32422 */ {(12<<2)|2,{57,110,0}},
/* 32423 */ {(12<<2)|2,{57,110,0}},
/* 32424 */ {(12<<2)|2,{57,110,0}},
/* 32425 */ {(12<<2)|2,{57,110,0}},
/* 32426 */ {(12<<2)|2,{57,110,0}},
/* 32427 */ {(12<<2)|2,{57,110,0}},
/* 32428 */ {(12<<2)|2,{57,110,0}},
/* 32429 */ {(12<<2)|2,{57,110,0}},
/* 32430 */ {(12<<2)|2,{57,110,0}},
/* 32431 */ {(12<<2)|2,{57,110,0}},
/* 32432 */ {(12<<2)|2,{57,112,0}},
/* 32433 */ {(12<<2)|2,{57,112,0}},
/* 32434 */ {(12<<2)|2,{57,112,0}},
/* 32435 */ {(12<<2)|2,{57,112,0}},
/* 32436 */ {(12<<2)|2,{57,112,0}},
/* 32437 */ {(12<<2)|2,{57,112,0}},
/* 32438 */ {(12<<2)|2,{57,112,0}},
/* 32439 */ {(12<<2)|2,{57,112,0}},
/* 32440 */ {(12<<2)|2,{57,112,0}},
/* 32441 */ {(12<<2)|2,{57,112,0}},
/* 32442 */ {(12<<2)|2,{57,112,0}},
/* 32443 */ {(12<<2)|2,{57,112,0}},
/* 32444 */ {(12<<2)|2,{57,112,0}},
/* 32445 */ {(12<<2)|2,{57,112,0}},
/* 32446 */ {(12<<2)|2,{57,112,0}},
/* 32447 */ {(12<<2)|2,{57,112,0}},
/* 32448 */ {(12<<2)|2,{57,114,0}},
/* 32449 */ {(12<<2)|2,{57,114,0}},
/* 32450 */ {(12<<2)|2,{57,114,0}},
/* 32451 */ {(12<<2)|2,{57,114,0}},
/* 32452 */ {(12<<2)|2,{57,114,0}},
/* 32453 */ {(12<<2)|2,{57,114,0}},
/* 32454 */ {(12<<2)|2,{57,114,0}},
/* 32455 */ {(12<<2)|2,{57,114,0}},
/* 32456 */ {(12<<2)|2,{57,114,0}},
/* 32457 */ {(12<<2)|2,{57,114,0}},
/* 32458 */ {(12<<2)|2,{57,114,0}},
/* 32459 */ {(12<<2)|2,{57,114,0}},
/* 32460 */ {(12<<2)|2,{57,114,0}},
/* 32461 */ {(12<<2)|2,{57,114,0}},
/* 32462 */ {(12<<2)|2,{57,114,0}},
/* 32463 */ {(12<<2)|2,{57,114,0}},
/* 32464 */ {(12<<2)|2,{57,117,0}},
/* 32465 */ {(12<<2)|2,{57,117,0}},
/* 32466 */ {(12<<2)|2,{57,117,0}},
/* 32467 */ {(12<<2)|2,{57,117,0}},
/* 32468 */ {(12<<2)|2,{57,117,0}},
/* 32469 */ {(12<<2)|2,{57,117,0}},
/* 32470 */ {(12<<2)|2,{57,117,0}},
/* 32471 */ {(12<<2)|2,{57,117,0}},
/* 32472 */ {(12<<2)|2,{57,117,0}},
/* 32473 */ {(12<<2)|2,{57,117,0}},
/* 32474 */ {(12<<2)|2,{57,117,0}},
/* 32475 */ {(12<<2)|2,{57,117,0}},
/* 32476 */ {(12<<2)|2,{57,117,0}},
/* 32477 */ {(12<<2)|2,{57,117,0}},
/* 32478 */ {(12<<2)|2,{57,117,0}},
/* 32479 */ {(12<<2)|2,{57,117,0}},
/* 32480 */ {(13<<2)|2,{57,58,0}},
/* 32481 */ {(13<<2)|2,{57,58,0}},
/* 32482 */ {(13<<2)|2,{57,58,0}},
/* 32483 */ {(13<<2)|2,{57,58,0}},
/* 32484 */ {(13<<2)|2,{57,58,0}},
/* 32485 */ {(13<<2)|2,{57,58,0}},
/* 32486 */ {(13<<2)|2,{57,58,0}},
/* 32487 */ {(13<<2)|2,{57,58,0}},
/* 32488 */ {(13<<2)|2,{57,66,0}},
/* 32489 */ {(13<<2)|2,{57,66,0}},
/* 32490 */ {(13<<2)|2,{57,66,0}},
/* 32491 */ {(13<<2)|2,{57,66,0}},
/* 32492 */ {(13<<2)|2,{57,66,0}},
/* 32493 */ {(13<<2)|2,{57,66,0}},
/* 32494 */ {(13<<2)|2,{57,66,0}},
/* 32495 */ {(13<<2)|2,{57,66,0}},
/* 32496 */ {(13<<2)|2,{57,67,0}},
/* 32497 */ {(13<<2)|2,{57,67,0}},
/* 32498 */ {(13<<2)|2,{57,67,0}},
/* 32499 */ {(13<<2)|2,{57,67,0}},
/* 32500 */ {(13<<2)|2,{57,67,0}},
/* 32501 */ {(13<<2)|2,{57,67,0}},
/* 32502 */ {(13<<2)|2,{57,67,0}},
/* 32503 */ {(13<<2)|2,{57,67,0}},
/* 32504 */ {(13<<2)|2,{57,68,0}},
/* 32505 */ {(13<<2)|2,{57,68,0}},
/* 32506 */ {(13<<2)|2,{57,68,0}},
/* 32507 */ {(13<<2)|2,{57,68,0}},
/* 32508 */ {(13<<2)|2,{57,68,0}},
/* 32509 */ {(13<<2)|2,{57,68,0}},
/* 32510 */ {(13<<2)|2,{57,68,0}},
/* 32511 */ {(13<<2)|2,{57,68,0}},
/* 32512 */ {(13<<2)|2,{57,69,0}},
/* 32513 */ {(13<<2)|2,{57,69,0}},
/* 32514 */ {(13<<2)|2,{57,69,0}},
/* 32515 */ {(13<<2)|2,{57,69,0}},
/* 32516 */ {(13<<2)|2,{57,69,0}},
/* 32517 */ {(13<<2)|2,{57,69,0}},
/* 32518 */ {(13<<2)|2,{57,69,0}},
/* 32519 */ {(13<<2)|2,{57,69,0}},
/* 32520 */ {(13<<2)|2,{57,70,0}},
/* 32521 */ {(13<<2)|2,{57,70,0}},
/* 32522 */ {(13<<2)|2,{57,70,0}},
/* 32523 */ {(13<<2)|2,{57,70,0}},
/* 32524 */ {(13<<2)|2,{57,70,0}},
/* 32525 */ {(13<<2)|2,{57,70,0}},
/* 32526 */ {(13<<2)|2,{57,70,0}},
/* 32527 */ {(13<<2)|2,{57,70,0}},
/* 32528 */ {(13<<2)|2,{57,71,0}},
/* 32529 */ {(13<<2)|2,{57,71,0}},
/* 32530 */ {(13<<2)|2,{57,71,0}},
/* 32531 */ {(13<<2)|2,{57,71,0}},
/* 32532 */ {(13<<2)|2,{57,71,0}},
/* 32533 */ {(13<<2)|2,{57,71,0}},
/* 32534 */ {(13<<2)|2,{57,71,0}},
/* 32535 */ {(13<<2)|2,{57,71,0}},
/* 32536 */ {(13<<2)|2,{57,72,0}},
/* 32537 */ {(13<<2)|2,{57,72,0}},
/* 32538 */ {(13<<2)|2,{57,72,0}},
/* 32539 */ {(13<<2)|2,{57,72,0}},
/* 32540 */ {(13<<2)|2,{57,72,0}},
/* 32541 */ {(13<<2)|2,{57,72,0}},
/* 32542 */ {(13<<2)|2,{57,72,0}},
/* 32543 */ {(13<<2)|2,{57,72,0}},
/* 32544 */ {(13<<2)|2,{57,73,0}},
/* 32545 */ {(13<<2)|2,{57,73,0}},
/* 32546 */ {(13<<2)|2,{57,73,0}},
/* 32547 */ {(13<<2)|2,{57,73,0}},
/* 32548 */ {(13<<2)|2,{57,73,0}},
/* 32549 */ {(13<<2)|2,{57,73,0}},
/* 32550 */ {(13<<2)|2,{57,73,0}},
/* 32551 */ {(13<<2)|2,{57,73,0}},
/* 32552 */ {(13<<2)|2,{57,74,0}},
/* 32553 */ {(13<<2)|2,{57,74,0}},
/* 32554 */ {(13<<2)|2,{57,74,0}},
/* 32555 */ {(13<<2)|2,{57,74,0}},
/* 32556 */ {(13<<2)|2,{57,74,0}},
/* 32557 */ {(13<<2)|2,{57,74,0}},
/* 32558 */ {(13<<2)|2,{57,74,0}},
/* 32559 */ {(13<<2)|2,{57,74,0}},
/* 32560 */ {(13<<2)|2,{57,75,0}},
/* 32561 */ {(13<<2)|2,{57,75,0}},
/* 32562 */ {(13<<2)|2,{57,75,0}},
/* 32563 */ {(13<<2)|2,{57,75,0}},
/* 32564 */ {(13<<2)|2,{57,75,0}},
/* 32565 */ {(13<<2)|2,{57,75,0}},
/* 32566 */ {(13<<2)|2,{57,75,0}},
/* 32567 */ {(13<<2)|2,{57,75,0}},
/* 32568 */ {(13<<2)|2,{57,76,0}},
/* 32569 */ {(13<<2)|2,{57,76,0}},
/* 32570 */ {(13<<2)|2,{57,76,0}},
/* 32571 */ {(13<<2)|2,{57,76,0}},
/* 32572 */ {(13<<2)|2,{57,76,0}},
/* 32573 */ {(13<<2)|2,{57,76,0}},
/* 32574 */ {(13<<2)|2,{57,76,0}},
/* 32575 */ {(13<<2)|2,{57,76,0}},
/* 32576 */ {(13<<2)|2,{57,77,0}},
/* 32577 */ {(13<<2)|2,{57,77,0}},
/* 32578 */ {(13<<2)|2,{57,77,0}},
/* 32579 */ {(13<<2)|2,{57,77,0}},
/* 32580 */ {(13<<2)|2,{57,77,0}},
/* 32581 */ {(13<<2)|2,{57,77,0}},
/* 32582 */ {(13<<2)|2,{57,77,0}},
/* 32583 */ {(13<<2)|2,{57,77,0}},
/* 32584 */ {(13<<2)|2,{57,78,0}},
/* 32585 */ {(13<<2)|2,{57,78,0}},
/* 32586 */ {(13<<2)|2,{57,78,0}},
/* 32587 */ {(13<<2)|2,{57,78,0}},
/* 32588 */ {(13<<2)|2,{57,78,0}},
/* 32589 */ {(13<<2)|2,{57,78,0}},
/* 32590 */ {(13<<2)|2,{57,78,0}},
/* 32591 */ {(13<<2)|2,{57,78,0}},
/* 32592 */ {(13<<2)|2,{57,79,0}},
/* 32593 */ {(13<<2)|2,{57,79,0}},
/* 32594 */ {(13<<2)|2,{57,79,0}},
/* 32595 */ {(13<<2)|2,{57,79,0}},
/* 32596 */ {(13<<2)|2,{57,79,0}},
/* 32597 */ {(13<<2)|2,{57,79,0}},
/* 32598 */ {(13<<2)|2,{57,79,0}},
/* 32599 */ {(13<<2)|2,{57,79,0}},
/* 32600 */ {(13<<2)|2,{57,80,0}},
/* 32601 */ {(13<<2)|2,{57,80,0}},
/* 32602 */ {(13<<2)|2,{57,80,0}},
/* 32603 */ {(13<<2)|2,{57,80,0}},
/* 32604 */ {(13<<2)|2,{57,80,0}},
/* 32605 */ {(13<<2)|2,{57,80,0}},
/* 32606 */ {(13<<2)|2,{57,80,0}},
/* 32607 */ {(13<<2)|2,{57,80,0}},
/* 32608 */ {(13<<2)|2,{57,81,0}},
/* 32609 */ {(13<<2)|2,{57,81,0}},
/* 32610 */ {(13<<2)|2,{57,81,0}},
/* 32611 */ {(13<<2)|2,{57,81,0}},
/* 32612 */ {(13<<2)|2,{57,81,0}},
/* 32613 */ {(13<<2)|2,{57,81,0}},
/* 32614 */ {(13<<2)|2,{57,81,0}},
/* 32615 */ {(13<<2)|2,{57,81,0}},
/* 32616 */ {(13<<2)|2,{57,82,0}},
/* 32617 */ {(13<<2)|2,{57,82,0}},
/* 32618 */ {(13<<2)|2,{57,82,0}},
/* 32619 */ {(13<<2)|2,{57,82,0}},
/* 32620 */ {(13<<2)|2,{57,82,0}},
/* 32621 */ {(13<<2)|2,{57,82,0}},
/* 32622 */ {(13<<2)|2,{57,82,0}},
/* 32623 */ {(13<<2)|2,{57,82,0}},
/* 32624 */ {(13<<2)|2,{57,83,0}},
/* 32625 */ {(13<<2)|2,{57,83,0}},
/* 32626 */ {(13<<2)|2,{57,83,0}},
/* 32627 */ {(13<<2)|2,{57,83,0}},
/* 32628 */ {(13<<2)|2,{57,83,0}},
/* 32629 */ {(13<<2)|2,{57,83,0}},
/* 32630 */ {(13<<2)|2,{57,83,0}},
/* 32631 */ {(13<<2)|2,{57,83,0}},
/* 32632 */ {(13<<2)|2,{57,84,0}},
/* 32633 */ {(13<<2)|2,{57,84,0}},
/* 32634 */ {(13<<2)|2,{57,84,0}},
/* 32635 */ {(13<<2)|2,{57,84,0}},
/* 32636 */ {(13<<2)|2,{57,84,0}},
/* 32637 */ {(13<<2)|2,{57,84,0}},
/* 32638 */ {(13<<2)|2,{57,84,0}},
/* 32639 */ {(13<<2)|2,{57,84,0}},
/* 32640 */ {(13<<2)|2,{57,85,0}},
/* 32641 */ {(13<<2)|2,{57,85,0}},
/* 32642 */ {(13<<2)|2,{57,85,0}},
/* 32643 */ {(13<<2)|2,{57,85,0}},
/* 32644 */ {(13<<2)|2,{57,85,0}},
/* 32645 */ {(13<<2)|2,{57,85,0}},
/* 32646 */ {(13<<2)|2,{57,85,0}},
/* 32647 */ {(13<<2)|2,{57,85,0}},
/* 32648 */ {(13<<2)|2,{57,86,0}},
/* 32649 */ {(13<<2)|2,{57,86,0}},
/* 32650 */ {(13<<2)|2,{57,86,0}},
/* 32651 */ {(13<<2)|2,{57,86,0}},
/* 32652 */ {(13<<2)|2,{57,86,0}},
/* 32653 */ {(13<<2)|2,{57,86,0}},
/* 32654 */ {(13<<2)|2,{57,86,0}},
/* 32655 */ {(13<<2)|2,{57,86,0}},
/* 32656 */ {(13<<2)|2,{57,87,0}},
/* 32657 */ {(13<<2)|2,{57,87,0}},
/* 32658 */ {(13<<2)|2,{57,87,0}},
/* 32659 */ {(13<<2)|2,{57,87,0}},
/* 32660 */ {(13<<2)|2,{57,87,0}},
/* 32661 */ {(13<<2)|2,{57,87,0}},
/* 32662 */ {(13<<2)|2,{57,87,0}},
/* 32663 */ {(13<<2)|2,{57,87,0}},
/* 32664 */ {(13<<2)|2,{57,89,0}},
/* 32665 */ {(13<<2)|2,{57,89,0}},
/* 32666 */ {(13<<2)|2,{57,89,0}},
/* 32667 */ {(13<<2)|2,{57,89,0}},
/* 32668 */ {(13<<2)|2,{57,89,0}},
/* 32669 */ {(13<<2)|2,{57,89,0}},
/* 32670 */ {(13<<2)|2,{57,89,0}},
/* 32671 */ {(13<<2)|2,{57,89,0}},
/* 32672 */ {(13<<2)|2,{57,106,0}},
/* 32673 */ {(13<<2)|2,{57,106,0}},
/* 32674 */ {(13<<2)|2,{57,106,0}},
/* 32675 */ {(13<<2)|2,{57,106,0}},
/* 32676 */ {(13<<2)|2,{57,106,0}},
/* 32677 */ {(13<<2)|2,{57,106,0}},
/* 32678 */ {(13<<2)|2,{57,106,0}},
/* 32679 */ {(13<<2)|2,{57,106,0}},
/* 32680 */ {(13<<2)|2,{57,107,0}},
/* 32681 */ {(13<<2)|2,{57,107,0}},
/* 32682 */ {(13<<2)|2,{57,107,0}},
/* 32683 */ {(13<<2)|2,{57,107,0}},
/* 32684 */ {(13<<2)|2,{57,107,0}},
/* 32685 */ {(13<<2)|2,{57,107,0}},
/* 32686 */ {(13<<2)|2,{57,107,0}},
/* 32687 */ {(13<<2)|2,{57,107,0}},
/* 32688 */ {(13<<2)|2,{57,113,0}},
/* 32689 */ {(13<<2)|2,{57,113,0}},
/* 32690 */ {(13<<2)|2,{57,113,0}},
/* 32691 */ {(13<<2)|2,{57,113,0}},
/* 32692 */ {(13<<2)|2,{57,113,0}},
/* 32693 */ {(13<<2)|2,{57,113,0}},
/* 32694 */ {(13<<2)|2,{57,113,0}},
/* 32695 */ {(13<<2)|2,{57,113,0}},
/* 32696 */ {(13<<2)|2,{57,118,0}},
/* 32697 */ {(13<<2)|2,{57,118,0}},
/* 32698 */ {(13<<2)|2,{57,118,0}},
/* 32699 */ {(13<<2)|2,{57,118,0}},
/* 32700 */ {(13<<2)|2,{57,118,0}},
/* 32701 */ {(13<<2)|2,{57,118,0}},
/* 32702 */ {(13<<2)|2,{57,118,0}},
/* 32703 */ {(13<<2)|2,{57,118,0}},
/* 32704 */ {(13<<2)|2,{57,119,0}},
/* 32705 */ {(13<<2)|2,{57,119,0}},
/* 32706 */ {(13<<2)|2,{57,119,0}},
/* 32707 */ {(13<<2)|2,{57,119,0}},
/* 32708 */ {(13<<2)|2,{57,119,0}},
/* 32709 */ {(13<<2)|2,{57,119,0}},
/* 32710 */ {(13<<2)|2,{57,119,0}},
/* 32711 */ {(13<<2)|2,{57,119,0}},
/* 32712 */ {(13<<2)|2,{57,120,0}},
/* 32713 */ {(13<<2)|2,{57,120,0}},
/* 32714 */ {(13<<2)|2,{57,120,0}},
/* 32715 */ {(13<<2)|2,{57,120,0}},
/* 32716 */ {(13<<2)|2,{57,120,0}},
/* 32717 */ {(13<<2)|2,{57,120,0}},
/* 32718 */ {(13<<2)|2,{57,120,0}},
/* 32719 */ {(13<<2)|2,{57,120,0}},
/* 32720 */ {(13<<2)|2,{57,121,0}},
/* 32721 */ {(13<<2)|2,{57,121,0}},
/* 32722 */ {(13<<2)|2,{57,121,0}},
/* 32723 */ {(13<<2)|2,{57,121,0}},
/* 32724 */ {(13<<2)|2,{57,121,0}},
/* 32725 */ {(13<<2)|2,{57,121,0}},
/* 32726 */ {(13<<2)|2,{57,121,0}},
/* 32727 */ {(13<<2)|2,{57,121,0}},
/* 32728 */ {(13<<2)|2,{57,122,0}},
/* 32729 */ {(13<<2)|2,{57,122,0}},
/* 32730 */ {(13<<2)|2,{57,122,0}},
/* 32731 */ {(13<<2)|2,{57,122,0}},
/* 32732 */ {(13<<2)|2,{57,122,0}},
/* 32733 */ {(13<<2)|2,{57,122,0}},
/* 32734 */ {(13<<2)|2,{57,122,0}},
/* 32735 */ {(13<<2)|2,{57,122,0}},
/* 32736 */ {(14<<2)|2,{57,38,0}},
/* 32737 */ {(14<<2)|2,{57,38,0}},
/* 32738 */ {(14<<2)|2,{57,38,0}},
/* 32739 */ {(14<<2)|2,{57,38,0}},
/* 32740 */ {(14<<2)|2,{57,42,0}},
/* 32741 */ {(14<<2)|2,{57,42,0}},
/* 32742 */ {(14<<2)|2,{57,42,0}},
/* 32743 */ {(14<<2)|2,{57,42,0}},
/* 32744 */ {(14<<2)|2,{57,44,0}},
/* 32745 */ {(14<<2)|2,{57,44,0}},
/* 32746 */ {(14<<2)|2,{57,44,0}},
/* 32747 */ {(14<<2)|2,{57,44,0}},
/* 32748 */ {(14<<2)|2,{57,59,0}},
/* 32749 */ {(14<<2)|2,{57,59,0}},
/* 32750 */ {(14<<2)|2,{57,59,0}},
/* 32751 */ {(14<<2)|2,{57,59,0}},
/* 32752 */ {(14<<2)|2,{57,88,0}},
/* 32753 */ {(14<<2)|2,{57,88,0}},
/* 32754 */ {(14<<2)|2,{57,88,0}},
/* 32755 */ {(14<<2)|2,{57,88,0}},
/* 32756 */ {(14<<2)|2,{57,90,0}},
/* 32757 */ {(14<<2)|2,{57,90,0}},
/* 32758 */ {(14<<2)|2,{57,90,0}},
/* 32759 */ {(14<<2)|2,{57,90,0}},
/* 32760 */ {(16<<2)|2,{57,33,0}},
/* 32761 */ {(16<<2)|2,{57,34,0}},
/* 32762 */ {(16<<2)|2,{57,40,0}},
/* 32763 */ {(16<<2)|2,{57,41,0}},
/* 32764 */ {(16<<2)|2,{57,63,0}},
/* 32765 */ {(6<<2)|1,{57,0,0}},
/* 32766 */ {(6<<2)|1,{57,0,0}},
/* 32767 */ {(6<<2)|1,{57,0,0}},
/* 32768 */ {(16<<2)|3,{61,48,48}},
/* 32769 */ {(16<<2)|3,{61,48,49}},
/* 32770 */ {(16<<2)|3,{61,48,50}},
/* 32771 */ {(16<<2)|3,{61,48,97}},
/* 32772 */ {(16<<2)|3,{61,48,99}},
/* 32773 */ {(16<<2)|3,{61,48,101}},
/* 32774 */ {(16<<2)|3,{61,48,105}},
/* 32775 */ {(16<<2)|3,{61,48,111}},
/* 32776 */ {(16<<2)|3,{61,48,115}},
/* 32777 */ {(16<<2)|3,{61,48,116}},
/* 32778 */ {(11<<2)|2,{61,48,0}},
/* 32779 */ {(11<<2)|2,{61,48,0}},
/* 32780 */ {(11<<2)|2,{61,48,0}},
/* 32781 */ {(11<<2)|2,{61,48,0}},
/* 32782 */ {(11<<2)|2,{61,48,0}},
/* 32783 */ {(11<<2)|2,{61,48,0}},
/* 32784 */ {(11<<2)|2,{61,48,0}},
/* 32785 */ {(11<<2)|2,{61,48,0}},
/* 32786 */ {(11<<2)|2,{61,48,0}},
/* 32787 */ {(11<<2)|2,{61,48,0}},
/* 32788 */ {(11<<2)|2,{61,48,0}},
/* 32789 */ {(11<<2)|2,{61,48,0}},
/* 32790 */ {(11<<2)|2,{61,48,0}},
/* 32791 */ {(11<<2)|2,{61,48,0}},
/* 32792 */ {(11<<2)|2,{61,48,0}},
/* 32793 */ {(11<<2)|2,{61,48,0}},
/* 32794 */ {(11<<2)|2,{61,48,0}},
/* 32795 */ {(11<<2)|2,{61,48,0}},
/* 32796 */ {(11<<2)|2,{61,48,0}},
/* 32797 */ {(11<<2)|2,{61,48,0}},
/* 32798 */ {(11<<2)|2,{61,48,0}},
/* 32799 */ {(11<<2)|2,{61,48,0}},
/* 32800 */ {(16<<2)|3,{61,49,48}},
/* 32801 */ {(16<<2)|3,{61,49,49}},
/* 32802 */ {(16<<2)|3,{61,49,50}},
/* 32803 */ {(16<<2)|3,{61,49,97}},
/* 32804 */ {(16<<2)|3,{61,49,99}},
/* 32805 */ {(16<<2)|3,{61,49,101}},
/* 32806 */ {(16<<2)|3,{61,49,105}},
/* 32807 */ {(16<<2)|3,{61,49,111}},
/* 32808 */ {(16<<2)|3,{61,49,115}},
/* 32809 */ {(16<<2)|3,{61,49,116}},
/* 32810 */ {(11<<2)|2,{61,49,0}},
/* 32811 */ {(11<<2)|2,{61,49,0}},
/* 32812 */ {(11<<2)|2,{61,49,0}},
/* 32813 */ {(11<<2)|2,{61,49,0}},
/* 32814 */ {(11<<2)|2,{61,49,0}},
/* 32815 */ {(11<<2)|2,{61,49,0}},
/* 32816 */ {(11<<2)|2,{61,49,0}},
/* 32817 */ {(11<<2)|2,{61,49,0}},
/* 32818 */ {(11<<2)|2,{61,49,0}},
/* 32819 */ {(11<<2)|2,{61,49,0}},
/* 32820 */ {(11<<2)|2,{61,49,0}},
/* 32821 */ {(11<<2)|2,{61,49,0}},
/* 32822 */ {(11<<2)|2,{61,49,0}},
/* 32823 */ {(11<<2)|2,{61,49,0}},
/* 32824 */ {(11<<2)|2,{61,49,0}},
/* 32825 */ {(11<<2)|2,{61,49,0}},
/* 32826 */ {(11<<2)|2,{61,49,0}},
/* 32827 */ {(11<<2)|2,{61,49,0}},
/* 32828 */ {(11<<2)|2,{61,49,0}},
/* 32829 */ {(11<<2)|2,{61,49,0}},
/* 32830 */ {(11<<2)|2,{61,49,0}},
/* 32831 */ {(11<<2)|2,{61,49,0}},
/* 32832 */ {(16<<2)|3,{61,50,48}},
/* 32833 */ {(16<<2)|3,{61,50,49}},
/* 32834 */ {(16<<2)|3,{61,50,50}},
/* 32835 */ {(16<<2)|3,{61,50,97}},
/* 32836 */ {(16<<2)|3,{61,50,99}},
/* 32837 */ {(16<<2)|3,{61,50,101}},
/* 32838 */ {(16<<2)|3,{61,50,105}},
/* 32839 */ {(16<<2)|3,{61,50,111}},
/* 32840 */ {(16<<2)|3,{61,50,115}},
/* 32841 */ {(16<<2)|3,{61,50,116}},
/* 32842 */ {(11<<2)|2,{61,50,0}},
/* 32843 */ {(11<<2)|2,{61,50,0}},
/* 32844 */ {(11<<2)|2,{61,50,0}},
/* 32845 */ {(11<<2)|2,{61,50,0}},
/* 32846 */ {(11<<2)|2,{61,50,0}},
/* 32847 */ {(11<<2)|2,{61,50,0}},
/* 32848 */ {(11<<2)|2,{61,50,0}},
/* 32849 */ {(11<<2)|2,{61,50,0}},
/* 32850 */ {(11<<2)|2,{61,50,0}},
/* 32851 */ {(11<<2)|2,{61,50,0}},
/* 32852 */ {(11<<2)|2,{61,50,0}},
/* 32853 */ {(11<<2)|2,{61,50,0}},
/* 32854 */ {(11<<2)|2,{61,50,0}},
/* 32855 */ {(11<<2)|2,{61,50,0}},
/* 32856 */ {(11<<2)|2,{61,50,0}},
/* 32857 */ {(11<<2)|2,{61,50,0}},
/* 32858 */ {(11<<2)|2,{61,50,0}},
/* 32859 */ {(11<<2)|2,{61,50,0}},
/* 32860 */ {(11<<2)|2,{61,50,0}},
/* 32861 */ {(11<<2)|2,{61,50,0}},
/* 32862 */ {(11<<2)|2,{61,50,0}},
/* 32863 */ {(11<<2)|2,{61,50,0}},
/* 32864 */ {(16<<2)|3,{61,97,48}},
/* 32865 */ {(16<<2)|3,{61,97,49}},
/* 32866 */ {(16<<2)|3,{61,97,50}},
/* 32867 */ {(16<<2)|3,{61,97,97}},
/* 32868 */ {(16<<2)|3,{61,97,99}},
/* 32869 */ {(16<<2)|3,{61,97,101}},
/* 32870 */ {(16<<2)|3,{61,97,105}},
/* 32871 */ {(16<<2)|3,{61,97,111}},
/* 32872 */ {(16<<2)|3,{61,97,115}},
/* 32873 */ {(16<<2)|3,{61,97,116}},
/* 32874 */ {(11<<2)|2,{61,97,0}},
/* 32875 */ {(11<<2)|2,{61,97,0}},
/* 32876 */ {(11<<2)|2,{61,97,0}},
/* 32877 */ {(11<<2)|2,{61,97,0}},
/* 32878 */ {(11<<2)|2,{61,97,0}},
/* 32879 */ {(11<<2)|2,{61,97,0}},
/* 32880 */ {(11<<2)|2,{61,97,0}},
/* 32881 */ {(11<<2)|2,{61,97,0}},
/* 32882 */ {(11<<2)|2,{61,97,0}},
/* 32883 */ {(11<<2)|2,{61,97,0}},
/* 32884 */ {(11<<2)|2,{61,97,0}},
/* 32885 */ {(11<<2)|2,{61,97,0}},
/* 32886 */ {(11<<2)|2,{61,97,0}},
/* 32887 */ {(11<<2)|2,{61,97,0}},
/* 32888 */ {(11<<2)|2,{61,97,0}},
/* 32889 */ {(11<<2)|2,{61,97,0}},
/* 32890 */ {(11<<2)|2,{61,97,0}},
/* 32891 */ {(11<<2)|2,{61,97,0}},
/* 32892 */ {(11<<2)|2,{61,97,0}},
/* 32893 */ {(11<<2)|2,{61,97,0}},
/* 32894 */ {(11<<2)|2,{61,97,0}},
/* 32895 */ {(11<<2)|2,{61,97,0}},
/* 32896 */ {(16<<2)|3,{61,99,48}},
/* 32897 */ {(16<<2)|3,{61,99,49}},
/* 32898 */ {(16<<2)|3,{61,99,50}},
/* 32899 */ {(16<<2)|3,{61,99,97}},
/* 32900 */ {(16<<2)|3,{61,99,99}},
/* 32901 */ {(16<<2)|3,{61,99,101}},
/* 32902 */ {(16<<2)|3,{61,99,105}},
/* 32903 */ {(16<<2)|3,{61,99,111}},
/* 32904 */ {(16<<2)|3,{61,99,115}},
/* 32905 */ {(16<<2)|3,{61,99,116}},
/* 32906 */ {(11<<2)|2,{61,99,0}},
/* 32907 */ {(11<<2)|2,{61,99,0}},
/* 32908 */ {(11<<2)|2,{61,99,0}},
/* 32909 */ {(11<<2)|2,{61,99,0}},
/* 32910 */ {(11<<2)|2,{61,99,0}},
/* 32911 */ {(11<<2)|2,{61,99,0}},
/* 32912 */ {(11<<2)|2,{61,99,0}},
/* 32913 */ {(11<<2)|2,{61,99,0}},
/* 32914 */ {(11<<2)|2,{61,99,0}},
/* 32915 */ {(11<<2)|2,{61,99,0}},
/* 32916 */ {(11<<2)|2,{61,99,0}},
/* 32917 */ {(11<<2)|2,{61,99,0}},
/* 32918 */ {(11<<2)|2,{61,99,0}},
/* 32919 */ {(11<<2)|2,{61,99,0}},
/* 32920 */ {(11<<2)|2,{61,99,0}},
/* 32921 */ {(11<<2)|2,{61,99,0}},
/* 32922 */ {(11<<2)|2,{61,99,0}},
/* 32923 */ {(11<<2)|2,{61,99,0}},
/* 32924 */ {(11<<2)|2,{61,99,0}},
/* 32925 */ {(11<<2)|2,{61,99,0}},
/* 32926 */ {(11<<2)|2,{61,99,0}},
/* 32927 */ {(11<<2)|2,{61,99,0}},
/* 32928 */ {(16<<2)|3,{61,101,48}},
/* 32929 */ {(16<<2)|3,{61,101,49}},
/* 32930 */ {(16<<2)|3,{61,101,50}},
/* 32931 */ {(16<<2)|3,{61,101,97}},
/* 32932 */ {(16<<2)|3,{61,101,99}},
/* 32933 */ {(16<<2)|3,{61,101,101}},
/* 32934 */ {(16<<2)|3,{61,101,105}},
/* 32935 */ {(16<<2)|3,{61,101,111}},
/* 32936 */ {(16<<2)|3,{61,101,115}},
/* 32937 */ {(16<<2)|3,{61,101,116}},
/* 32938 */ {(11<<2)|2,{61,101,0}},
/* 32939 */ {(11<<2)|2,{61,101,0}},
/* 32940 */ {(11<<2)|2,{61,101,0}},
/* 32941 */ {(11<<2)|2,{61,101,0}},
/* 32942 */ {(11<<2)|2,{61,101,0}},
/* 32943 */ {(11<<2)|2,{61,101,0}},
/* 32944 */ {(11<<2)|2,{61,101,0}},
/* 32945 */ {(11<<2)|2,{61,101,0}},
/* 32946 */ {(11<<2)|2,{61,101,0}},
/* 32947 */ {(11<<2)|2,{61,101,0}},
/* 32948 */ {(11<<2)|2,{61,101,0}},
/* 32949 */ {(11<<2)|2,{61,101,0}},
/* 32950 */ {(11<<2)|2,{61,101,0}},
/* 32951 */ {(11<<2)|2,{61,101,0}},
/* 32952 */ {(11<<2)|2,{61,101,0}},
/* 32953 */ {(11<<2)|2,{61,101,0}},
/* 32954 */ {(11<<2)|2,{61,101,0}},
/* 32955 */ {(11<<2)|2,{61,101,0}},
/* 32956 */ {(11<<2)|2,{61,101,0}},
/* 32957 */ {(11<<2)|2,{61,101,0}},
/* 32958 */ {(11<<2)|2,{61,101,0}},
/* 32959 */ {(11<<2)|2,{61,101,0}},
/* 32960 */ {(16<<2)|3,{61,105,48}},
/* 32961 */ {(16<<2)|3,{61,105,49}},
/* 32962 */ {(16<<2)|3,{61,105,50}},
/* 32963 */ {(16<<2)|3,{61,105,97}},
/* 32964 */ {(16<<2)|3,{61,105,99}},
/* 32965 */ {(16<<2)|3,{61,105,101}},
/* 32966 */ {(16<<2)|3,{61,105,105}},
/* 32967 */ {(16<<2)|3,{61,105,111}},
/* 32968 */ {(16<<2)|3,{61,105,115}},
/* 32969 */ {(16<<2)|3,{61,105,116}},
/* 32970 */ {(11<<2)|2,{61,105,0}},
/* 32971 */ {(11<<2)|2,{61,105,0}},
/* 32972 */ {(11<<2)|2,{61,105,0}},
/* 32973 */ {(11<<2)|2,{61,105,0}},
/* 32974 */ {(11<<2)|2,{61,105,0}},
/* 32975 */ {(11<<2)|2,{61,105,0}},
/* 32976 */ {(11<<2)|2,{61,105,0}},
/* 32977 */ {(11<<2)|2,{61,105,0}},
/* 32978 */ {(11<<2)|2,{61,105,0}},
/* 32979 */ {(11<<2)|2,{61,105,0}},
/* 32980 */ {(11<<2)|2,{61,105,0}},
/* 32981 */ {(11<<2)|2,{61,105,0}},
/* 32982 */ {(11<<2)|2,{61,105,0}},
/* 32983 */ {(11<<2)|2,{61,105,0}},
/* 32984 */ {(11<<2)|2,{61,105,0}},
/* 32985 */ {(11<<2)|2,{61,105,0}},
/* 32986 */ {(11<<2)|2,{61,105,0}},
/* 32987 */ {(11<<2)|2,{61,105,0}},
/* 32988 */ {(11<<2)|2,{61,105,0}},
/* 32989 */ {(11<<2)|2,{61,105,0}},
/* 32990 */ {(11<<2)|2,{61,105,0}},
/* 32991 */ {(11<<2)|2,{61,105,0}},
/* 32992 */ {(16<<2)|3,{61,111,48}},
/* 32993 */ {(16<<2)|3,{61,111,49}},
/* 32994 */ {(16<<2)|3,{61,111,50}},
/* 32995 */ {(16<<2)|3,{61,111,97}},
/* 32996 */ {(16<<2)|3,{61,111,99}},
/* 32997 */ {(16<<2)|3,{61,111,101}},
/* 32998 */ {(16<<2)|3,{61,111,105}},
/* 32999 */ {(16<<2)|3,{61,111,111}},
/* 33000 */ {(16<<2)|3,{61,111,115}},
/* 33001 */ {(16<<2)|3,{61,111,116}},
/* 33002 */ {(11<<2)|2,{61,111,0}},
/* 33003 */ {(11<<2)|2,{61,111,0}},
/* 33004 */ {(11<<2)|2,{61,111,0}},
/* 33005 */ {(11<<2)|2,{61,111,0}},
/* 33006 */ {(11<<2)|2,{61,111,0}},
/* 33007 */ {(11<<2)|2,{61,111,0}},
/* 33008 */ {(11<<2)|2,{61,111,0}},
/* 33009 */ {(11<<2)|2,{61,111,0}},
/* 33010 */ {(11<<2)|2,{61,111,0}},
/* 33011 */ {(11<<2)|2,{61,111,0}},
/* 33012 */ {(11<<2)|2,{61,111,0}},
/* 33013 */ {(11<<2)|2,{61,111,0}},
/* 33014 */ {(11<<2)|2,{61,111,0}},
/* 33015 */ {(11<<2)|2,{61,111,0}},
/* 33016 */ {(11<<2)|2,{61,111,0}},
/* 33017 */ {(11<<2)|2,{61,111,0}},
/* 33018 */ {(11<<2)|2,{61,111,0}},
/* 33019 */ {(11<<2)|2,{61,111,0}},
/* 33020 */ {(11<<2)|2,{61,111,0}},
/* 33021 */ {(11<<2)|2,{61,111,0}},
/* 33022 */ {(11<<2)|2,{61,111,0}},
/* 33023 */ {(11<<2)|2,{61,111,0}},
/* 33024 */ {(16<<2)|3,{61,115,48}},
/* 33025 */ {(16<<2)|3,{61,115,49}},
/* 33026 */ {(16<<2)|3,{61,115,50}},
/* 33027 */ {(16<<2)|3,{61,115,97}},
/* 33028 */ {(16<<2)|3,{61,115,99}},
/* 33029 */ {(16<<2)|3,{61,115,101}},
/* 33030 */ {(16<<2)|3,{61,115,105}},
/* 33031 */ {(16<<2)|3,{61,115,111}},
/* 33032 */ {(16<<2)|3,{61,115,115}},
/* 33033 */ {(16<<2)|3,{61,115,116}},
/* 33034 */ {(11<<2)|2,{61,115,0}},
/* 33035 */ {(11<<2)|2,{61,115,0}},
/* 33036 */ {(11<<2)|2,{61,115,0}},
/* 33037 */ {(11<<2)|2,{61,115,0}},
/* 33038 */ {(11<<2)|2,{61,115,0}},
/* 33039 */ {(11<<2)|2,{61,115,0}},
/* 33040 */ {(11<<2)|2,{61,115,0}},
/* 33041 */ {(11<<2)|2,{61,115,0}},
/* 33042 */ {(11<<2)|2,{61,115,0}},
/* 33043 */ {(11<<2)|2,{61,115,0}},
/* 33044 */ {(11<<2)|2,{61,115,0}},
/* 33045 */ {(11<<2)|2,{61,115,0}},
/* 33046 */ {(11<<2)|2,{61,115,0}},
/* 33047 */ {(11<<2)|2,{61,115,0}},
/* 33048 */ {(11<<2)|2,{61,115,0}},
/* 33049 */ {(11<<2)|2,{61,115,0}},
/* 33050 */ {(11<<2)|2,{61,115,0}},
/* 33051 */ {(11<<2)|2,{61,115,0}},
/* 33052 */ {(11<<2)|2,{61,115,0}},
/* 33053 */ {(11<<2)|2,{61,115,0}},
/* 33054 */ {(11<<2)|2,{61,115,0}},
/* 33055 */ {(11<<2)|2,{61,115,0}},
/* 33056 */ {(16<<2)|3,{61,116,48}},
/* 33057 */ {(16<<2)|3,{61,116,49}},
/* 33058 */ {(16<<2)|3,{61,116,50}},
/* 33059 */ {(16<<2)|3,{61,116,97}},
/* 33060 */ {(16<<2)|3,{61,116,99}},
/* 33061 */ {(16<<2)|3,{61,116,101}},
/* 33062 */ {(16<<2)|3,{61,116,105}},
/* 33063 */ {(16<<2)|3,{61,116,111}},
/* 33064 */ {(16<<2)|3,{61,116,115}},
/* 33065 */ {(16<<2)|3,{61,116,116}},
/* 33066 */ {(11<<2)|2,{61,116,0}},
/* 33067 */ {(11<<2)|2,{61,116,0}},
/* 33068 */ {(11<<2)|2,{61,116,0}},
/* 33069 */ {(11<<2)|2,{61,116,0}},
/* 33070 */ {(11<<2)|2,{61,116,0}},
/* 33071 */ {(11<<2)|2,{61,116,0}},
/* 33072 */ {(11<<2)|2,{61,116,0}},
/* 33073 */ {(11<<2)|2,{61,116,0}},
/* 33074 */ {(11<<2)|2,{61,116,0}},
/* 33075 */ {(11<<2)|2,{61,116,0}},
/* 33076 */ {(11<<2)|2,{61,116,0}},
/* 33077 */ {(11<<2)|2,{61,116,0}},
/* 33078 */ {(11<<2)|2,{61,116,0}},
/* 33079 */ {(11<<2)|2,{61,116,0}},
/* 33080 */ {(11<<2)|2,{61,116,0}},
/* 33081 */ {(11<<2)|2,{61,116,0}},
/* 33082 */ {(11<<2)|2,{61,116,0}},
/* 33083 */ {(11<<2)|2,{61,116,0}},
/* 33084 */ {(11<<2)|2,{61,116,0}},
/* 33085 */ {(11<<2)|2,{61,116,0}},
/* 33086 */ {(11<<2)|2,{61,116,0}},
/* 33087 */ {(11<<2)|2,{61,116,0}},
/* 33088 */ {(12<<2)|2,{61,32,0}},
/* 33089 */ {(12<<2)|2,{61,32,0}},
/* 33090 */ {(12<<2)|2,{61,32,0}},
/* 33091 */ {(12<<2)|2,{61,32,0}},
/* 33092 */ {(12<<2)|2,{61,32,0}},
/* 33093 */ {(12<<2)|2,{61,32,0}},
/* 33094 */ {(12<<2)|2,{61,32,0}},
/* 33095 */ {(12<<2)|2,{61,32,0}},
/* 33096 */ {(12<<2)|2,{61,32,0}},
/* 33097 */ {(12<<2)|2,{61,32,0}},
/* 33098 */ {(12<<2)|2,{61,32,0}},
/* 33099 */ {(12<<2)|2,{61,32,0}},
/* 33100 */ {(12<<2)|2,{61,32,0}},
/* 33101 */ {(12<<2)|2,{61,32,0}},
/* 33102 */ {(12<<2)|2,{61,32,0}},
/* 33103 */ {(12<<2)|2,{61,32,0}},
/* 33104 */ {(12<<2)|2,{61,37,0}},
/* 33105 */ {(12<<2)|2,{61,37,0}},
/* 33106 */ {(12<<2)|2,{61,37,0}},
/* 33107 */ {(12<<2)|2,{61,37,0}},
/* 33108 */ {(12<<2)|2,{61,37,0}},
/* 33109 */ {(12<<2)|2,{61,37,0}},
/* 33110 */ {(12<<2)|2,{61,37,0}},
/* 33111 */ {(12<<2)|2,{61,37,0}},
/* 33112 */ {(12<<2)|2,{61,37,0}},
/* 33113 */ {(12<<2)|2,{61,37,0}},
/* 33114 */ {(12<<2)|2,{61,37,0}},
/* 33115 */ {(12<<2)|2,{61,37,0}},
/* 33116 */ {(12<<2)|2,{61,37,0}},
/* 33117 */ {(12<<2)|2,{61,37,0}},
/* 33118 */ {(12<<2)|2,{61,37,0}},
/* 33119 */ {(12<<2)|2,{61,37,0}},
/* 33120 */ {(12<<2)|2,{61,45,0}},
/* 33121 */ {(12<<2)|2,{61,45,0}},
/* 33122 */ {(12<<2)|2,{61,45,0}},
/* 33123 */ {(12<<2)|2,{61,45,0}},
/* 33124 */ {(12<<2)|2,{61,45,0}},
/* 33125 */ {(12<<2)|2,{61,45,0}},
/* 33126 */ {(12<<2)|2,{61,45,0}},
/* 33127 */ {(12<<2)|2,{61,45,0}},
/* 33128 */ {(12<<2)|2,{61,45,0}},
/* 33129 */ {(12<<2)|2,{61,45,0}},
/* 33130 */ {(12<<2)|2,{61,45,0}},
/* 33131 */ {(12<<2)|2,{61,45,0}},
/* 33132 */ {(12<<2)|2,{61,45,0}},
/* 33133 */ {(12<<2)|2,{61,45,0}},
/* 33134 */ {(12<<2)|2,{61,45,0}},
/* 33135 */ {(12<<2)|2,{61,45,0}},
/* 33136 */ {(12<<2)|2,{61,46,0}},
/* 33137 */ {(12<<2)|2,{61,46,0}},
/* 33138 */ {(12<<2)|2,{61,46,0}},
/* 33139 */ {(12<<2)|2,{61,46,0}},
/* 33140 */ {(12<<2)|2,{61,46,0}},
/* 33141 */ {(12<<2)|2,{61,46,0}},
/* 33142 */ {(12<<2)|2,{61,46,0}},
/* 33143 */ {(12<<2)|2,{61,46,0}},
/* 33144 */ {(12<<2)|2,{61,46,0}},
/* 33145 */ {(12<<2)|2,{61,46,0}},
/* 33146 */ {(12<<2)|2,{61,46,0}},
/* 33147 */ {(12<<2)|2,{61,46,0}},
/* 33148 */ {(12<<2)|2,{61,46,0}},
/* 33149 */ {(12<<2)|2,{61,46,0}},
/* 33150 */ {(12<<2)|2,{61,46,0}},
/* 33151 */ {(12<<2)|2,{61,46,0}},
/* 33152 */ {(12<<2)|2,{61,47,0}},
/* 33153 */ {(12<<2)|2,{61,47,0}},
/* 33154 */ {(12<<2)|2,{61,47,0}},
/* 33155 */ {(12<<2)|2,{61,47,0}},
/* 33156 */ {(12<<2)|2,{61,47,0}},
/* 33157 */ {(12<<2)|2,{61,47,0}},
/* 33158 */ {(12<<2)|2,{61,47,0}},
/* 33159 */ {(12<<2)|2,{61,47,0}},
/* 33160 */ {(12<<2)|2,{61,47,0}},
/* 33161 */ {(12<<2)|2,{61,47,0}},
/* 33162 */ {(12<<2)|2,{61,47,0}},
/* 33163 */ {(12<<2)|2,{61,47,0}},
/* 33164 */ {(12<<2)|2,{61,47,0}},
/* 33165 */ {(12<<2)|2,{61,47,0}},
/* 33166 */ {(12<<2)|2,{61,47,0}},
/* 33167 */ {(12<<2)|2,{61,47,0}},
/* 33168 */ {(12<<2)|2,{61,51,0}},
/* 33169 */ {(12<<2)|2,{61,51,0}},
/* 33170 */ {(12<<2)|2,{61,51,0}},
/* 33171 */ {(12<<2)|2,{61,51,0}},
/* 33172 */ {(12<<2)|2,{61,51,0}},
/* 33173 */ {(12<<2)|2,{61,51,0}},
/* 33174 */ {(12<<2)|2,{61,51,0}},
/* 33175 */ {(12<<2)|2,{61,51,0}},
/* 33176 */ {(12<<2)|2,{61,51,0}},
/* 33177 */ {(12<<2)|2,{61,51,0}},
/* 33178 */ {(12<<2)|2,{61,51,0}},
/* 33179 */ {(12<<2)|2,{61,51,0}},
/* 33180 */ {(12<<2)|2,{61,51,0}},
/* 33181 */ {(12<<2)|2,{61,51,0}},
/* 33182 */ {(12<<2)|2,{61,51,0}},
/* 33183 */ {(12<<2)|2,{61,51,0}},
/* 33184 */ {(12<<2)|2,{61,52,0}},
/* 33185 */ {(12<<2)|2,{61,52,0}},
/* 33186 */ {(12<<2)|2,{61,52,0}},
/* 33187 */ {(12<<2)|2,{61,52,0}},
/* 33188 */ {(12<<2)|2,{61,52,0}},
/* 33189 */ {(12<<2)|2,{61,52,0}},
/* 33190 */ {(12<<2)|2,{61,52,0}},
/* 33191 */ {(12<<2)|2,{61,52,0}},
/* 33192 */ {(12<<2)|2,{61,52,0}},
/* 33193 */ {(12<<2)|2,{61,52,0}},
/* 33194 */ {(12<<2)|2,{61,52,0}},
/* 33195 */ {(12<<2)|2,{61,52,0}},
/* 33196 */ {(12<<2)|2,{61,52,0}},
/* 33197 */ {(12<<2)|2,{61,52,0}},
/* 33198 */ {(12<<2)|2,{61,52,0}},
/* 33199 */ {(12<<2)|2,{61,52,0}},
/* 33200 */ {(12<<2)|2,{61,53,0}},
/* 33201 */ {(12<<2)|2,{61,53,0}},
/* 33202 */ {(12<<2)|2,{61,53,0}},
/* 33203 */ {(12<<2)|2,{61,53,0}},
/* 33204 */ {(12<<2)|2,{61,53,0}},
/* 33205 */ {(12<<2)|2,{61,53,0}},
/* 33206 */ {(12<<2)|2,{61,53,0}},
/* 33207 */ {(12<<2)|2,{61,53,0}},
/* 33208 */ {(12<<2)|2,{61,53,0}},
/* 33209 */ {(12<<2)|2,{61,53,0}},
/* 33210 */ {(12<<2)|2,{61,53,0}},
/* 33211 */ {(12<<2)|2,{61,53,0}},
/* 33212 */ {(12<<2)|2,{61,53,0}},
/* 33213 */ {(12<<2)|2,{61,53,0}},
/* 33214 */ {(12<<2)|2,{61,53,0}},
/* 33215 */ {(12<<2)|2,{61,53,0}},
/* 33216 */ {(12<<2)|2,{61,54,0}},
/* 33217 */ {(12<<2)|2,{61,54,0}},
/* 33218 */ {(12<<2)|2,{61,54,0}},
/* 33219 */ {(12<<2)|2,{61,54,0}},
/* 33220 */ {(12<<2)|2,{61,54,0}},
/* 33221 */ {(12<<2)|2,{61,54,0}},
/* 33222 */ {(12<<2)|2,{61,54,0}},
/* 33223 */ {(12<<2)|2,{61,54,0}},
/* 33224 */ {(12<<2)|2,{61,54,0}},
/* 33225 */ {(12<<2)|2,{61,54,0}},
/* 33226 */ {(12<<2)|2,{61,54,0}},
/* 33227 */ {(12<<2)|2,{61,54,0}},
/* 33228 */ {(12<<2)|2,{61,54,0}},
/* 33229 */ {(12<<2)|2,{61,54,0}},
/* 33230 */ {(12<<2)|2,{61,54,0}},
/* 33231 */ {(12<<2)|2,{61,54,0}},
/* 33232 */ {(12<<2)|2,{61,55,0}},
/* 33233 */ {(12<<2)|2,{61,55,0}},
/* 33234 */ {(12<<2)|2,{61,55,0}},
/* 33235 */ {(12<<2)|2,{61,55,0}},
/* 33236 */ {(12<<2)|2,{61,55,0}},
/* 33237 */ {(12<<2)|2,{61,55,0}},
/* 33238 */ {(12<<2)|2,{61,55,0}},
/* 33239 */ {(12<<2)|2,{61,55,0}},
/* 33240 */ {(12<<2)|2,{61,55,0}},
/* 33241 */ {(12<<2)|2,{61,55,0}},
/* 33242 */ {(12<<2)|2,{61,55,0}},
/* 33243 */ {(12<<2)|2,{61,55,0}},
/* 33244 */ {(12<<2)|2,{61,55,0}},
/* 33245 */ {(12<<2)|2,{61,55,0}},
/* 33246 */ {(12<<2)|2,{61,55,0}},
/* 33247 */ {(12<<2)|2,{61,55,0}},
/* 33248 */ {(12<<2)|2,{61,56,0}},
/* 33249 */ {(12<<2)|2,{61,56,0}},
/* 33250 */ {(12<<2)|2,{61,56,0}},
/* 33251 */ {(12<<2)|2,{61,56,0}},
/* 33252 */ {(12<<2)|2,{61,56,0}},
/* 33253 */ {(12<<2)|2,{61,56,0}},
/* 33254 */ {(12<<2)|2,{61,56,0}},
/* 33255 */ {(12<<2)|2,{61,56,0}},
/* 33256 */ {(12<<2)|2,{61,56,0}},
/* 33257 */ {(12<<2)|2,{61,56,0}},
/* 33258 */ {(12<<2)|2,{61,56,0}},
/* 33259 */ {(12<<2)|2,{61,56,0}},
/* 33260 */ {(12<<2)|2,{61,56,0}},
/* 33261 */ {(12<<2)|2,{61,56,0}},
/* 33262 */ {(12<<2)|2,{61,56,0}},
/* 33263 */ {(12<<2)|2,{61,56,0}},
/* 33264 */ {(12<<2)|2,{61,57,0}},
/* 33265 */ {(12<<2)|2,{61,57,0}},
/* 33266 */ {(12<<2)|2,{61,57,0}},
/* 33267 */ {(12<<2)|2,{61,57,0}},
/* 33268 */ {(12<<2)|2,{61,57,0}},
/* 33269 */ {(12<<2)|2,{61,57,0}},
/* 33270 */ {(12<<2)|2,{61,57,0}},
/* 33271 */ {(12<<2)|2,{61,57,0}},
/* 33272 */ {(12<<2)|2,{61,57,0}},
/* 33273 */ {(12<<2)|2,{61,57,0}},
/* 33274 */ {(12<<2)|2,{61,57,0}},
/* 33275 */ {(12<<2)|2,{61,57,0}},
/* 33276 */ {(12<<2)|2,{61,57,0}},
/* 33277 */ {(12<<2)|2,{61,57,0}},
/* 33278 */ {(12<<2)|2,{61,57,0}},
/* 33279 */ {(12<<2)|2,{61,57,0}},
/* 33280 */ {(12<<2)|2,{61,61,0}},
/* 33281 */ {(12<<2)|2,{61,61,0}},
/* 33282 */ {(12<<2)|2,{61,61,0}},
/* 33283 */ {(12<<2)|2,{61,61,0}},
/* 33284 */ {(12<<2)|2,{61,61,0}},
/* 33285 */ {(12<<2)|2,{61,61,0}},
/* 33286 */ {(12<<2)|2,{61,61,0}},
/* 33287 */ {(12<<2)|2,{61,61,0}},
/* 33288 */ {(12<<2)|2,{61,61,0}},
/* 33289 */ {(12<<2)|2,{61,61,0}},
/* 33290 */ {(12<<2)|2,{61,61,0}},
/* 33291 */ {(12<<2)|2,{61,61,0}},
/* 33292 */ {(12<<2)|2,{61,61,0}},
/* 33293 */ {(12<<2)|2,{61,61,0}},
/* 33294 */ {(12<<2)|2,{61,61,0}},
/* 33295 */ {(12<<2)|2,{61,61,0}},
/* 33296 */ {(12<<2)|2,{61,65,0}},
/* 33297 */ {(12<<2)|2,{61,65,0}},
/* 33298 */ {(12<<2)|2,{61,65,0}},
/* 33299 */ {(12<<2)|2,{61,65,0}},
/* 33300 */ {(12<<2)|2,{61,65,0}},
/* 33301 */ {(12<<2)|2,{61,65,0}},
/* 33302 */ {(12<<2)|2,{61,65,0}},
/* 33303 */ {(12<<2)|2,{61,65,0}},
/* 33304 */ {(12<<2)|2,{61,65,0}},
/* 33305 */ {(12<<2)|2,{61,65,0}},
/* 33306 */ {(12<<2)|2,{61,65,0}},
/* 33307 */ {(12<<2)|2,{61,65,0}},
/* 33308 */ {(12<<2)|2,{61,65,0}},
/* 33309 */ {(12<<2)|2,{61,65,0}},
/* 33310 */ {(12<<2)|2,{61,65,0}},
/* 33311 */ {(12<<2)|2,{61,65,0}},
/* 33312 */ {(12<<2)|2,{61,95,0}},
/* 33313 */ {(12<<2)|2,{61,95,0}},
/* 33314 */ {(12<<2)|2,{61,95,0}},
/* 33315 */ {(12<<2)|2,{61,95,0}},
/* 33316 */ {(12<<2)|2,{61,95,0}},
/* 33317 */ {(12<<2)|2,{61,95,0}},
/* 33318 */ {(12<<2)|2,{61,95,0}},
/* 33319 */ {(12<<2)|2,{61,95,0}},
/* 33320 */ {(12<<2)|2,{61,95,0}},
/* 33321 */ {(12<<2)|2,{61,95,0}},
/* 33322 */ {(12<<2)|2,{61,95,0}},
/* 33323 */ {(12<<2)|2,{61,95,0}},
/* 33324 */ {(12<<2)|2,{61,95,0}},
/* 33325 */ {(12<<2)|2,{61,95,0}},
/* 33326 */ {(12<<2)|2,{61,95,0}},
/* 33327 */ {(12<<2)|2,{61,95,0}},
/* 33328 */ {(12<<2)|2,{61,98,0}},
/* 33329 */ {(12<<2)|2,{61,98,0}},
/* 33330 */ {(12<<2)|2,{61,98,0}},
/* 33331 */ {(12<<2)|2,{61,98,0}},
/* 33332 */ {(12<<2)|2,{61,98,0}},
/* 33333 */ {(12<<2)|2,{61,98,0}},
/* 33334 */ {(12<<2)|2,{61,98,0}},
/* 33335 */ {(12<<2)|2,{61,98,0}},
/* 33336 */ {(12<<2)|2,{61,98,0}},
/* 33337 */ {(12<<2)|2,{61,98,0}},
/* 33338 */ {(12<<2)|2,{61,98,0}},
/* 33339 */ {(12<<2)|2,{61,98,0}},
/* 33340 */ {(12<<2)|2,{61,98,0}},
/* 33341 */ {(12<<2)|2,{61,98,0}},
/* 33342 */ {(12<<2)|2,{61,98,0}},
/* 33343 */ {(12<<2)|2,{61,98,0}},
/* 33344 */ {(12<<2)|2,{61,100,0}},
/* 33345 */ {(12<<2)|2,{61,100,0}},
/* 33346 */ {(12<<2)|2,{61,100,0}},
/* 33347 */ {(12<<2)|2,{61,100,0}},
/* 33348 */ {(12<<2)|2,{61,100,0}},
/* 33349 */ {(12<<2)|2,{61,100,0}},
/* 33350 */ {(12<<2)|2,{61,100,0}},
/* 33351 */ {(12<<2)|2,{61,100,0}},
/* 33352 */ {(12<<2)|2,{61,100,0}},
/* 33353 */ {(12<<2)|2,{61,100,0}},
/* 33354 */ {(12<<2)|2,{61,100,0}},
/* 33355 */ {(12<<2)|2,{61,100,0}},
/* 33356 */ {(12<<2)|2,{61,100,0}},
/* 33357 */ {(12<<2)|2,{61,100,0}},
/* 33358 */ {(12<<2)|2,{61,100,0}},
/* 33359 */ {(12<<2)|2,{61,100,0}},
/* 33360 */ {(12<<2)|2,{61,102,0}},
/* 33361 */ {(12<<2)|2,{61,102,0}},
/* 33362 */ {(12<<2)|2,{61,102,0}},
/* 33363 */ {(12<<2)|2,{61,102,0}},
/* 33364 */ {(12<<2)|2,{61,102,0}},
/* 33365 */ {(12<<2)|2,{61,102,0}},
/* 33366 */ {(12<<2)|2,{61,102,0}},
/* 33367 */ {(12<<2)|2,{61,102,0}},
/* 33368 */ {(12<<2)|2,{61,102,0}},
/* 33369 */ {(12<<2)|2,{61,102,0}},
/* 33370 */ {(12<<2)|2,{61,102,0}},
/* 33371 */ {(12<<2)|2,{61,102,0}},
/* 33372 */ {(12<<2)|2,{61,102,0}},
/* 33373 */ {(12<<2)|2,{61,102,0}},
/* 33374 */ {(12<<2)|2,{61,102,0}},
/* 33375 */ {(12<<2)|2,{61,102,0}},
/* 33376 */ {(12<<2)|2,{61,103,0}},
/* 33377 */ {(12<<2)|2,{61,103,0}},
/* 33378 */ {(12<<2)|2,{61,103,0}},
/* 33379 */ {(12<<2)|2,{61,103,0}},
/* 33380 */ {(12<<2)|2,{61,103,0}},
/* 33381 */ {(12<<2)|2,{61,103,0}},
/* 33382 */ {(12<<2)|2,{61,103,0}},
/* 33383 */ {(12<<2)|2,{61,103,0}},
/* 33384 */ {(12<<2)|2,{61,103,0}},
/* 33385 */ {(12<<2)|2,{61,103,0}},
/* 33386 */ {(12<<2)|2,{61,103,0}},
/* 33387 */ {(12<<2)|2,{61,103,0}},
/* 33388 */ {(12<<2)|2,{61,103,0}},
/* 33389 */ {(12<<2)|2,{61,103,0}},
/* 33390 */ {(12<<2)|2,{61,103,0}},
/* 33391 */ {(12<<2)|2,{61,103,0}},
/* 33392 */ {(12<<2)|2,{61,104,0}},
/* 33393 */ {(12<<2)|2,{61,104,0}},
/* 33394 */ {(12<<2)|2,{61,104,0}},
/* 33395 */ {(12<<2)|2,{61,104,0}},
/* 33396 */ {(12<<2)|2,{61,104,0}},
/* 33397 */ {(12<<2)|2,{61,104,0}},
/* 33398 */ {(12<<2)|2,{61,104,0}},
/* 33399 */ {(12<<2)|2,{61,104,0}},
/* 33400 */ {(12<<2)|2,{61,104,0}},
/* 33401 */ {(12<<2)|2,{61,104,0}},
/* 33402 */ {(12<<2)|2,{61,104,0}},
/* 33403 */ {(12<<2)|2,{61,104,0}},
/* 33404 */ {(12<<2)|2,{61,104,0}},
/* 33405 */ {(12<<2)|2,{61,104,0}},
/* 33406 */ {(12<<2)|2,{61,104,0}},
/* 33407 */ {(12<<2)|2,{61,104,0}},
/* 33408 */ {(12<<2)|2,{61,108,0}},
/* 33409 */ {(12<<2)|2,{61,108,0}},
/* 33410 */ {(12<<2)|2,{61,108,0}},
/* 33411 */ {(12<<2)|2,{61,108,0}},
/* 33412 */ {(12<<2)|2,{61,108,0}},
/* 33413 */ {(12<<2)|2,{61,108,0}},
/* 33414 */ {(12<<2)|2,{61,108,0}},
/* 33415 */ {(12<<2)|2,{61,108,0}},
/* 33416 */ {(12<<2)|2,{61,108,0}},
/* 33417 */ {(12<<2)|2,{61,108,0}},
/* 33418 */ {(12<<2)|2,{61,108,0}},
/* 33419 */ {(12<<2)|2,{61,108,0}},
/* 33420 */ {(12<<2)|2,{61,108,0}},
/* 33421 */ {(12<<2)|2,{61,108,0}},
/* 33422 */ {(12<<2)|2,{61,108,0}},
/* 33423 */ {(12<<2)|2,{61,108,0}},
/* 33424 */ {(12<<2)|2,{61,109,0}},
/* 33425 */ {(12<<2)|2,{61,109,0}},
/* 33426 */ {(12<<2)|2,{61,109,0}},
/* 33427 */ {(12<<2)|2,{61,109,0}},
/* 33428 */ {(12<<2)|2,{61,109,0}},
/* 33429 */ {(12<<2)|2,{61,109,0}},
/* 33430 */ {(12<<2)|2,{61,109,0}},
/* 33431 */ {(12<<2)|2,{61,109,0}},
/* 33432 */ {(12<<2)|2,{61,109,0}},
/* 33433 */ {(12<<2)|2,{61,109,0}},
/* 33434 */ {(12<<2)|2,{61,109,0}},
/* 33435 */ {(12<<2)|2,{61,109,0}},
/* 33436 */ {(12<<2)|2,{61,109,0}},
/* 33437 */ {(12<<2)|2,{61,109,0}},
/* 33438 */ {(12<<2)|2,{61,109,0}},
/* 33439 */ {(12<<2)|2,{61,109,0}},
/* 33440 */ {(12<<2)|2,{61,110,0}},
/* 33441 */ {(12<<2)|2,{61,110,0}},
/* 33442 */ {(12<<2)|2,{61,110,0}},
/* 33443 */ {(12<<2)|2,{61,110,0}},
/* 33444 */ {(12<<2)|2,{61,110,0}},
/* 33445 */ {(12<<2)|2,{61,110,0}},
/* 33446 */ {(12<<2)|2,{61,110,0}},
/* 33447 */ {(12<<2)|2,{61,110,0}},
/* 33448 */ {(12<<2)|2,{61,110,0}},
/* 33449 */ {(12<<2)|2,{61,110,0}},
/* 33450 */ {(12<<2)|2,{61,110,0}},
/* 33451 */ {(12<<2)|2,{61,110,0}},
/* 33452 */ {(12<<2)|2,{61,110,0}},
/* 33453 */ {(12<<2)|2,{61,110,0}},
/* 33454 */ {(12<<2)|2,{61,110,0}},
/* 33455 */ {(12<<2)|2,{61,110,0}},
/* 33456 */ {(12<<2)|2,{61,112,0}},
/* 33457 */ {(12<<2)|2,{61,112,0}},
/* 33458 */ {(12<<2)|2,{61,112,0}},
/* 33459 */ {(12<<2)|2,{61,112,0}},
/* 33460 */ {(12<<2)|2,{61,112,0}},
/* 33461 */ {(12<<2)|2,{61,112,0}},
/* 33462 */ {(12<<2)|2,{61,112,0}},
/* 33463 */ {(12<<2)|2,{61,112,0}},
/* 33464 */ {(12<<2)|2,{61,112,0}},
/* 33465 */ {(12<<2)|2,{61,112,0}},
/* 33466 */ {(12<<2)|2,{61,112,0}},
/* 33467 */ {(12<<2)|2,{61,112,0}},
/* 33468 */ {(12<<2)|2,{61,112,0}},
/* 33469 */ {(12<<2)|2,{61,112,0}},
/* 33470 */ {(12<<2)|2,{61,112,0}},
/* 33471 */ {(12<<2)|2,{61,112,0}},
/* 33472 */ {(12<<2)|2,{61,114,0}},
/* 33473 */ {(12<<2)|2,{61,114,0}},
/* 33474 */ {(12<<2)|2,{61,114,0}},
/* 33475 */ {(12<<2)|2,{61,114,0}},
/* 33476 */ {(12<<2)|2,{61,114,0}},
/* 33477 */ {(12<<2)|2,{61,114,0}},
/* 33478 */ {(12<<2)|2,{61,114,0}},
/* 33479 */ {(12<<2)|2,{61,114,0}},
/* 33480 */ {(12<<2)|2,{61,114,0}},
/* 33481 */ {(12<<2)|2,{61,114,0}},
/* 33482 */ {(12<<2)|2,{61,114,0}},
/* 33483 */ {(12<<2)|2,{61,114,0}},
/* 33484 */ {(12<<2)|2,{61,114,0}},
/* 33485 */ {(12<<2)|2,{61,114,0}},
/* 33486 */ {(12<<2)|2,{61,114,0}},
/* 33487 */ {(12<<2)|2,{61,114,0}},
/* 33488 */ {(12<<2)|2,{61,117,0}},
/* 33489 */ {(12<<2)|2,{61,117,0}},
/* 33490 */ {(12<<2)|2,{61,117,0}},
/* 33491 */ {(12<<2)|2,{61,117,0}},
/* 33492 */ {(12<<2)|2,{61,117,0}},
/* 33493 */ {(12<<2)|2,{61,117,0}},
/* 33494 */ {(12<<2)|2,{61,117,0}},
/* 33495 */ {(12<<2)|2,{61,117,0}},
/* 33496 */ {(12<<2)|2,{61,117,0}},
/* 33497 */ {(12<<2)|2,{61,117,0}},
/* 33498 */ {(12<<2)|2,{61,117,0}},
/* 33499 */ {(12<<2)|2,{61,117,0}},
/* 33500 */ {(12<<2)|2,{61,117,0}},
/* 33501 */ {(12<<2)|2,{61,117,0}},
/* 33502 */ {(12<<2)|2,{61,117,0}},
/* 33503 */ {(12<<2)|2,{61,117,0}},
/* 33504 */ {(13<<2)|2,{61,58,0}},
/* 33505 */ {(13<<2)|2,{61,58,0}},
/* 33506 */ {(13<<2)|2,{61,58,0}},
/* 33507 */ {(13<<2)|2,{61,58,0}},
/* 33508 */ {(13<<2)|2,{61,58,0}},
/* 33509 */ {(13<<2)|2,{61,58,0}},
/* 33510 */ {(13<<2)|2,{61,58,0}},
/* 33511 */ {(13<<2)|2,{61,58,0}},
/* 33512 */ {(13<<2)|2,{61,66,0}},
/* 33513 */ {(13<<2)|2,{61,66,0}},
/* 33514 */ {(13<<2)|2,{61,66,0}},
/* 33515 */ {(13<<2)|2,{61,66,0}},
/* 33516 */ {(13<<2)|2,{61,66,0}},
/* 33517 */ {(13<<2)|2,{61,66,0}},
/* 33518 */ {(13<<2)|2,{61,66,0}},
/* 33519 */ {(13<<2)|2,{61,66,0}},
/* 33520 */ {(13<<2)|2,{61,67,0}},
/* 33521 */ {(13<<2)|2,{61,67,0}},
/* 33522 */ {(13<<2)|2,{61,67,0}},
/* 33523 */ {(13<<2)|2,{61,67,0}},
/* 33524 */ {(13<<2)|2,{61,67,0}},
/* 33525 */ {(13<<2)|2,{61,67,0}},
/* 33526 */ {(13<<2)|2,{61,67,0}},
/* 33527 */ {(13<<2)|2,{61,67,0}},
/* 33528 */ {(13<<2)|2,{61,68,0}},
/* 33529 */ {(13<<2)|2,{61,68,0}},
/* 33530 */ {(13<<2)|2,{61,68,0}},
/* 33531 */ {(13<<2)|2,{61,68,0}},
/* 33532 */ {(13<<2)|2,{61,68,0}},
/* 33533 */ {(13<<2)|2,{61,68,0}},
/* 33534 */ {(13<<2)|2,{61,68,0}},
/* 33535 */ {(13<<2)|2,{61,68,0}},
/* 33536 */ {(13<<2)|2,{61,69,0}},
/* 33537 */ {(13<<2)|2,{61,69,0}},
/* 33538 */ {(13<<2)|2,{61,69,0}},
/* 33539 */ {(13<<2)|2,{61,69,0}},
/* 33540 */ {(13<<2)|2,{61,69,0}},
/* 33541 */ {(13<<2)|2,{61,69,0}},
/* 33542 */ {(13<<2)|2,{61,69,0}},
/* 33543 */ {(13<<2)|2,{61,69,0}},
/* 33544 */ {(13<<2)|2,{61,70,0}},
/* 33545 */ {(13<<2)|2,{61,70,0}},
/* 33546 */ {(13<<2)|2,{61,70,0}},
/* 33547 */ {(13<<2)|2,{61,70,0}},
/* 33548 */ {(13<<2)|2,{61,70,0}},
/* 33549 */ {(13<<2)|2,{61,70,0}},
/* 33550 */ {(13<<2)|2,{61,70,0}},
/* 33551 */ {(13<<2)|2,{61,70,0}},
/* 33552 */ {(13<<2)|2,{61,71,0}},
/* 33553 */ {(13<<2)|2,{61,71,0}},
/* 33554 */ {(13<<2)|2,{61,71,0}},
/* 33555 */ {(13<<2)|2,{61,71,0}},
/* 33556 */ {(13<<2)|2,{61,71,0}},
/* 33557 */ {(13<<2)|2,{61,71,0}},
/* 33558 */ {(13<<2)|2,{61,71,0}},
/* 33559 */ {(13<<2)|2,{61,71,0}},
/* 33560 */ {(13<<2)|2,{61,72,0}},
/* 33561 */ {(13<<2)|2,{61,72,0}},
/* 33562 */ {(13<<2)|2,{61,72,0}},
/* 33563 */ {(13<<2)|2,{61,72,0}},
/* 33564 */ {(13<<2)|2,{61,72,0}},
/* 33565 */ {(13<<2)|2,{61,72,0}},
/* 33566 */ {(13<<2)|2,{61,72,0}},
/* 33567 */ {(13<<2)|2,{61,72,0}},
/* 33568 */ {(13<<2)|2,{61,73,0}},
/* 33569 */ {(13<<2)|2,{61,73,0}},
/* 33570 */ {(13<<2)|2,{61,73,0}},
/* 33571 */ {(13<<2)|2,{61,73,0}},
/* 33572 */ {(13<<2)|2,{61,73,0}},
/* 33573 */ {(13<<2)|2,{61,73,0}},
/* 33574 */ {(13<<2)|2,{61,73,0}},
/* 33575 */ {(13<<2)|2,{61,73,0}},
/* 33576 */ {(13<<2)|2,{61,74,0}},
/* 33577 */ {(13<<2)|2,{61,74,0}},
/* 33578 */ {(13<<2)|2,{61,74,0}},
/* 33579 */ {(13<<2)|2,{61,74,0}},
/* 33580 */ {(13<<2)|2,{61,74,0}},
/* 33581 */ {(13<<2)|2,{61,74,0}},
/* 33582 */ {(13<<2)|2,{61,74,0}},
/* 33583 */ {(13<<2)|2,{61,74,0}},
/* 33584 */ {(13<<2)|2,{61,75,0}},
/* 33585 */ {(13<<2)|2,{61,75,0}},
/* 33586 */ {(13<<2)|2,{61,75,0}},
/* 33587 */ {(13<<2)|2,{61,75,0}},
/* 33588 */ {(13<<2)|2,{61,75,0}},
/* 33589 */ {(13<<2)|2,{61,75,0}},
/* 33590 */ {(13<<2)|2,{61,75,0}},
/* 33591 */ {(13<<2)|2,{61,75,0}},
/* 33592 */ {(13<<2)|2,{61,76,0}},
/* 33593 */ {(13<<2)|2,{61,76,0}},
/* 33594 */ {(13<<2)|2,{61,76,0}},
/* 33595 */ {(13<<2)|2,{61,76,0}},
/* 33596 */ {(13<<2)|2,{61,76,0}},
/* 33597 */ {(13<<2)|2,{61,76,0}},
/* 33598 */ {(13<<2)|2,{61,76,0}},
/* 33599 */ {(13<<2)|2,{61,76,0}},
/* 33600 */ {(13<<2)|2,{61,77,0}},
/* 33601 */ {(13<<2)|2,{61,77,0}},
/* 33602 */ {(13<<2)|2,{61,77,0}},
/* 33603 */ {(13<<2)|2,{61,77,0}},
/* 33604 */ {(13<<2)|2,{61,77,0}},
/* 33605 */ {(13<<2)|2,{61,77,0}},
/* 33606 */ {(13<<2)|2,{61,77,0}},
/* 33607 */ {(13<<2)|2,{61,77,0}},
/* 33608 */ {(13<<2)|2,{61,78,0}},
/* 33609 */ {(13<<2)|2,{61,78,0}},
/* 33610 */ {(13<<2)|2,{61,78,0}},
/* 33611 */ {(13<<2)|2,{61,78,0}},
/* 33612 */ {(13<<2)|2,{61,78,0}},
/* 33613 */ {(13<<2)|2,{61,78,0}},
/* 33614 */ {(13<<2)|2,{61,78,0}},
/* 33615 */ {(13<<2)|2,{61,78,0}},
/* 33616 */ {(13<<2)|2,{61,79,0}},
/* 33617 */ {(13<<2)|2,{61,79,0}},
/* 33618 */ {(13<<2)|2,{61,79,0}},
/* 33619 */ {(13<<2)|2,{61,79,0}},
/* 33620 */ {(13<<2)|2,{61,79,0}},
/* 33621 */ {(13<<2)|2,{61,79,0}},
/* 33622 */ {(13<<2)|2,{61,79,0}},
/* 33623 */ {(13<<2)|2,{61,79,0}},
/* 33624 */ {(13<<2)|2,{61,80,0}},
/* 33625 */ {(13<<2)|2,{61,80,0}},
/* 33626 */ {(13<<2)|2,{61,80,0}},
/* 33627 */ {(13<<2)|2,{61,80,0}},
/* 33628 */ {(13<<2)|2,{61,80,0}},
/* 33629 */ {(13<<2)|2,{61,80,0}},
/* 33630 */ {(13<<2)|2,{61,80,0}},
/* 33631 */ {(13<<2)|2,{61,80,0}},
/* 33632 */ {(13<<2)|2,{61,81,0}},
/* 33633 */ {(13<<2)|2,{61,81,0}},
/* 33634 */ {(13<<2)|2,{61,81,0}},
/* 33635 */ {(13<<2)|2,{61,81,0}},
/* 33636 */ {(13<<2)|2,{61,81,0}},
/* 33637 */ {(13<<2)|2,{61,81,0}},
/* 33638 */ {(13<<2)|2,{61,81,0}},
/* 33639 */ {(13<<2)|2,{61,81,0}},
/* 33640 */ {(13<<2)|2,{61,82,0}},
/* 33641 */ {(13<<2)|2,{61,82,0}},
/* 33642 */ {(13<<2)|2,{61,82,0}},
/* 33643 */ {(13<<2)|2,{61,82,0}},
/* 33644 */ {(13<<2)|2,{61,82,0}},
/* 33645 */ {(13<<2)|2,{61,82,0}},
/* 33646 */ {(13<<2)|2,{61,82,0}},
/* 33647 */ {(13<<2)|2,{61,82,0}},
/* 33648 */ {(13<<2)|2,{61,83,0}},
/* 33649 */ {(13<<2)|2,{61,83,0}},
/* 33650 */ {(13<<2)|2,{61,83,0}},
/* 33651 */ {(13<<2)|2,{61,83,0}},
/* 33652 */ {(13<<2)|2,{61,83,0}},
/* 33653 */ {(13<<2)|2,{61,83,0}},
/* 33654 */ {(13<<2)|2,{61,83,0}},
/* 33655 */ {(13<<2)|2,{61,83,0}},
/* 33656 */ {(13<<2)|2,{61,84,0}},
/* 33657 */ {(13<<2)|2,{61,84,0}},
/* 33658 */ {(13<<2)|2,{61,84,0}},
/* 33659 */ {(13<<2)|2,{61,84,0}},
/* 33660 */ {(13<<2)|2,{61,84,0}},
/* 33661 */ {(13<<2)|2,{61,84,0}},
/* 33662 */ {(13<<2)|2,{61,84,0}},
/* 33663 */ {(13<<2)|2,{61,84,0}},
/* 33664 */ {(13<<2)|2,{61,85,0}},
/* 33665 */ {(13<<2)|2,{61,85,0}},
/* 33666 */ {(13<<2)|2,{61,85,0}},
/* 33667 */ {(13<<2)|2,{61,85,0}},
/* 33668 */ {(13<<2)|2,{61,85,0}},
/* 33669 */ {(13<<2)|2,{61,85,0}},
/* 33670 */ {(13<<2)|2,{61,85,0}},
/* 33671 */ {(13<<2)|2,{61,85,0}},
/* 33672 */ {(13<<2)|2,{61,86,0}},
/* 33673 */ {(13<<2)|2,{61,86,0}},
/* 33674 */ {(13<<2)|2,{61,86,0}},
/* 33675 */ {(13<<2)|2,{61,86,0}},
/* 33676 */ {(13<<2)|2,{61,86,0}},
/* 33677 */ {(13<<2)|2,{61,86,0}},
/* 33678 */ {(13<<2)|2,{61,86,0}},
/* 33679 */ {(13<<2)|2,{61,86,0}},
/* 33680 */ {(13<<2)|2,{61,87,0}},
/* 33681 */ {(13<<2)|2,{61,87,0}},
/* 33682 */ {(13<<2)|2,{61,87,0}},
/* 33683 */ {(13<<2)|2,{61,87,0}},
/* 33684 */ {(13<<2)|2,{61,87,0}},
/* 33685 */ {(13<<2)|2,{61,87,0}},
/* 33686 */ {(13<<2)|2,{61,87,0}},
/* 33687 */ {(13<<2)|2,{61,87,0}},
/* 33688 */ {(13<<2)|2,{61,89,0}},
/* 33689 */ {(13<<2)|2,{61,89,0}},
/* 33690 */ {(13<<2)|2,{61,89,0}},
/* 33691 */ {(13<<2)|2,{61,89,0}},
/* 33692 */ {(13<<2)|2,{61,89,0}},
/* 33693 */ {(13<<2)|2,{61,89,0}},
/* 33694 */ {(13<<2)|2,{61,89,0}},
/* 33695 */ {(13<<2)|2,{61,89,0}},
/* 33696 */ {(13<<2)|2,{61,106,0}},
/* 33697 */ {(13<<2)|2,{61,106,0}},
/* 33698 */ {(13<<2)|2,{61,106,0}},
/* 33699 */ {(13<<2)|2,{61,106,0}},
/* 33700 */ {(13<<2)|2,{61,106,0}},
/* 33701 */ {(13<<2)|2,{61,106,0}},
/* 33702 */ {(13<<2)|2,{61,106,0}},
/* 33703 */ {(13<<2)|2,{61,106,0}},
/* 33704 */ {(13<<2)|2,{61,107,0}},
/* 33705 */ {(13<<2)|2,{61,107,0}},
/* 33706 */ {(13<<2)|2,{61,107,0}},
/* 33707 */ {(13<<2)|2,{61,107,0}},
/* 33708 */ {(13<<2)|2,{61,107,0}},
/* 33709 */ {(13<<2)|2,{61,107,0}},
/* 33710 */ {(13<<2)|2,{61,107,0}},
/* 33711 */ {(13<<2)|2,{61,107,0}},
/* 33712 */ {(13<<2)|2,{61,113,0}},
/* 33713 */ {(13<<2)|2,{61,113,0}},
/* 33714 */ {(13<<2)|2,{61,113,0}},
/* 33715 */ {(13<<2)|2,{61,113,0}},
/* 33716 */ {(13<<2)|2,{61,113,0}},
/* 33717 */ {(13<<2)|2,{61,113,0}},
/* 33718 */ {(13<<2)|2,{61,113,0}},
/* 33719 */ {(13<<2)|2,{61,113,0}},
/* 33720 */ {(13<<2)|2,{61,118,0}},
/* 33721 */ {(13<<2)|2,{61,118,0}},
/* 33722 */ {(13<<2)|2,{61,118,0}},
/* 33723 */ {(13<<2)|2,{61,118,0}},
/* 33724 */ {(13<<2)|2,{61,118,0}},
/* 33725 */ {(13<<2)|2,{61,118,0}},
/* 33726 */ {(13<<2)|2,{61,118,0}},
/* 33727 */ {(13<<2)|2,{61,118,0}},
/* 33728 */ {(13<<2)|2,{61,119,0}},
/* 33729 */ {(13<<2)|2,{61,119,0}},
/* 33730 */ {(13<<2)|2,{61,119,0}},
/* 33731 */ {(13<<2)|2,{61,119,0}},
/* 33732 */ {(13<<2)|2,{61,119,0}},
/* 33733 */ {(13<<2)|2,{61,119,0}},
/* 33734 */ {(13<<2)|2,{61,119,0}},
/* 33735 */ {(13<<2)|2,{61,119,0}},
/* 33736 */ {(13<<2)|2,{61,120,0}},
/* 33737 */ {(13<<2)|2,{61,120,0}},
/* 33738 */ {(13<<2)|2,{61,120,0}},
/* 33739 */ {(13<<2)|2,{61,120,0}},
/* 33740 */ {(13<<2)|2,{61,120,0}},
/* 33741 */ {(13<<2)|2,{61,120,0}},
/* 33742 */ {(13<<2)|2,{61,120,0}},
/* 33743 */ {(13<<2)|2,{61,120,0}},
/* 33744 */ {(13<<2)|2,{61,121,0}},
/* 33745 */ {(13<<2)|2,{61,121,0}},
/* 33746 */ {(13<<2)|2,{61,121,0}},
/* 33747 */ {(13<<2)|2,{61,121,0}},
/* 33748 */ {(13<<2)|2,{61,121,0}},
/* 33749 */ {(13<<2)|2,{61,121,0}},
/* 33750 */ {(13<<2)|2,{61,121,0}},
/* 33751 */ {(13<<2)|2,{61,121,0}},
/* 33752 */ {(13<<2)|2,{61,122,0}},
/* 33753 */ {(13<<2)|2,{61,122,0}},
/* 33754 */ {(13<<2)|2,{61,122,0}},
/* 33755 */ {(13<<2)|2,{61,122,0}},
/* 33756 */ {(13<<2)|2,{61,122,0}},
/* 33757 */ {(13<<2)|2,{61,122,0}},
/* 33758 */ {(13<<2)|2,{61,122,0}},
/* 33759 */ {(13<<2)|2,{61,122,0}},
/* 33760 */ {(14<<2)|2,{61,38,0}},
/* 33761 */ {(14<<2)|2,{61,38,0}},
/* 33762 */ {(14<<2)|2,{61,38,0}},
/* 33763 */ {(14<<2)|2,{61,38,0}},
/* 33764 */ {(14<<2)|2,{61,42,0}},
/* 33765 */ {(14<<2)|2,{61,42,0}},
/* 33766 */ {(14<<2)|2,{61,42,0}},
/* 33767 */ {(14<<2)|2,{61,42,0}},
/* 33768 */ {(14<<2)|2,{61,44,0}},
/* 33769 */ {(14<<2)|2,{61,44,0}},
/* 33770 */ {(14<<2)|2,{61,44,0}},
/* 33771 */ {(14<<2)|2,{61,44,0}},
/* 33772 */ {(14<<2)|2,{61,59,0}},
/* 33773 */ {(14<<2)|2,{61,59,0}},
/* 33774 */ {(14<<2)|2,{61,59,0}},
/* 33775 */ {(14<<2)|2,{61,59,0}},
/* 33776 */ {(14<<2)|2,{61,88,0}},
/* 33777 */ {(14<<2)|2,{61,88,0}},
/* 33778 */ {(14<<2)|2,{61,88,0}},
/* 33779 */ {(14<<2)|2,{61,88,0}},
/* 33780 */ {(14<<2)|2,{61,90,0}},
/* 33781 */ {(14<<2)|2,{61,90,0}},
/* 33782 */ {(14<<2)|2,{61,90,0}},
/* 33783 */ {(14<<2)|2,{61,90,0}},
/* 33784 */ {(16<<2)|2,{61,33,0}},
/* 33785 */ {(16<<2)|2,{61,34,0}},
/* 33786 */ {(16<<2)|2,{61,40,0}},
/* 33787 */ {(16<<2)|2,{61,41,0}},
/* 33788 */ {(16<<2)|2,{61,63,0}},
/* 33789 */ {(6<<2)|1,{61,0,0}},
/* 33790 */ {(6<<2)|1,{61,0,0}},
/* 33791 */ {(6<<2)|1,{61,0,0}},
/* 33792 */ {(16<<2)|3,{65,48,48}},
/* 33793 */ {(16<<2)|3,{65,48,49}},
/* 33794 */ {(16<<2)|3,{65,48,50}},
/* 33795 */ {(16<<2)|3,{65,48,97}},
/* 33796 */ {(16<<2)|3,{65,48,99}},
/* 33797 */ {(16<<2)|3,{65,48,101}},
/* 33798 */ {(16<<2)|3,{65,48,105}},
/* 33799 */ {(16<<2)|3,{65,48,111}},
/* 33800 */ {(16<<2)|3,{65,48,115}},
/* 33801 */ {(16<<2)|3,{65,48,116}},
/* 33802 */ {(11<<2)|2,{65,48,0}},
/* 33803 */ {(11<<2)|2,{65,48,0}},
/* 33804 */ {(11<<2)|2,{65,48,0}},
/* 33805 */ {(11<<2)|2,{65,48,0}},
/* 33806 */ {(11<<2)|2,{65,48,0}},
/* 33807 */ {(11<<2)|2,{65,48,0}},
/* 33808 */ {(11<<2)|2,{65,48,0}},
/* 33809 */ {(11<<2)|2,{65,48,0}},
/* 33810 */ {(11<<2)|2,{65,48,0}},
/* 33811 */ {(11<<2)|2,{65,48,0}},
/* 33812 */ {(11<<2)|2,{65,48,0}},
/* 33813 */ {(11<<2)|2,{65,48,0}},
/* 33814 */ {(11<<2)|2,{65,48,0}},
/* 33815 */ {(11<<2)|2,{65,48,0}},
/* 33816 */ {(11<<2)|2,{65,48,0}},
/* 33817 */ {(11<<2)|2,{65,48,0}},
/* 33818 */ {(11<<2)|2,{65,48,0}},
/* 33819 */ {(11<<2)|2,{65,48,0}},
/* 33820 */ {(11<<2)|2,{65,48,0}},
/* 33821 */ {(11<<2)|2,{65,48,0}},
/* 33822 */ {(11<<2)|2,{65,48,0}},
/* 33823 */ {(11<<2)|2,{65,48,0}},
/* 33824 */ {(16<<2)|3,{65,49,48}},
/* 33825 */ {(16<<2)|3,{65,49,49}},
/* 33826 */ {(16<<2)|3,{65,49,50}},
/* 33827 */ {(16<<2)|3,{65,49,97}},
/* 33828 */ {(16<<2)|3,{65,49,99}},
/* 33829 */ {(16<<2)|3,{65,49,101}},
/* 33830 */ {(16<<2)|3,{65,49,105}},
/* 33831 */ {(16<<2)|3,{65,49,111}},
/* 33832 */ {(16<<2)|3,{65,49,115}},
/* 33833 */ {(16<<2)|3,{65,49,116}},
/* 33834 */ {(11<<2)|2,{65,49,0}},
/* 33835 */ {(11<<2)|2,{65,49,0}},
/* 33836 */ {(11<<2)|2,{65,49,0}},
/* 33837 */ {(11<<2)|2,{65,49,0}},
/* 33838 */ {(11<<2)|2,{65,49,0}},
/* 33839 */ {(11<<2)|2,{65,49,0}},
/* 33840 */ {(11<<2)|2,{65,49,0}},
/* 33841 */ {(11<<2)|2,{65,49,0}},
/* 33842 */ {(11<<2)|2,{65,49,0}},
/* 33843 */ {(11<<2)|2,{65,49,0}},
/* 33844 */ {(11<<2)|2,{65,49,0}},
/* 33845 */ {(11<<2)|2,{65,49,0}},
/* 33846 */ {(11<<2)|2,{65,49,0}},
/* 33847 */ {(11<<2)|2,{65,49,0}},
/* 33848 */ {(11<<2)|2,{65,49,0}},
/* 33849 */ {(11<<2)|2,{65,49,0}},
/* 33850 */ {(11<<2)|2,{65,49,0}},
/* 33851 */ {(11<<2)|2,{65,49,0}},
/* 33852 */ {(11<<2)|2,{65,49,0}},
/* 33853 */ {(11<<2)|2,{65,49,0}},
/* 33854 */ {(11<<2)|2,{65,49,0}},
/* 33855 */ {(11<<2)|2,{65,49,0}},
/* 33856 */ {(16<<2)|3,{65,50,48}},
/* 33857 */ {(16<<2)|3,{65,50,49}},
/* 33858 */ {(16<<2)|3,{65,50,50}},
/* 33859 */ {(16<<2)|3,{65,50,97}},
/* 33860 */ {(16<<2)|3,{65,50,99}},
/* 33861 */ {(16<<2)|3,{65,50,101}},
/* 33862 */ {(16<<2)|3,{65,50,105}},
/* 33863 */ {(16<<2)|3,{65,50,111}},
/* 33864 */ {(16<<2)|3,{65,50,115}},
/* 33865 */ {(16<<2)|3,{65,50,116}},
/* 33866 */ {(11<<2)|2,{65,50,0}},
/* 33867 */ {(11<<2)|2,{65,50,0}},
/* 33868 */ {(11<<2)|2,{65,50,0}},
/* 33869 */ {(11<<2)|2,{65,50,0}},
/* 33870 */ {(11<<2)|2,{65,50,0}},
/* 33871 */ {(11<<2)|2,{65,50,0}},
/* 33872 */ {(11<<2)|2,{65,50,0}},
/* 33873 */ {(11<<2)|2,{65,50,0}},
/* 33874 */ {(11<<2)|2,{65,50,0}},
/* 33875 */ {(11<<2)|2,{65,50,0}},
/* 33876 */ {(11<<2)|2,{65,50,0}},
/* 33877 */ {(11<<2)|2,{65,50,0}},
/* 33878 */ {(11<<2)|2,{65,50,0}},
/* 33879 */ {(11<<2)|2,{65,50,0}},
/* 33880 */ {(11<<2)|2,{65,50,0}},
/* 33881 */ {(11<<2)|2,{65,50,0}},
/* 33882 */ {(11<<2)|2,{65,50,0}},
/* 33883 */ {(11<<2)|2,{65,50,0}},
/* 33884 */ {(11<<2)|2,{65,50,0}},
/* 33885 */ {(11<<2)|2,{65,50,0}},
/* 33886 */ {(11<<2)|2,{65,50,0}},
/* 33887 */ {(11<<2)|2,{65,50,0}},
/* 33888 */ {(16<<2)|3,{65,97,48}},
/* 33889 */ {(16<<2)|3,{65,97,49}},
/* 33890 */ {(16<<2)|3,{65,97,50}},
/* 33891 */ {(16<<2)|3,{65,97,97}},
/* 33892 */ {(16<<2)|3,{65,97,99}},
/* 33893 */ {(16<<2)|3,{65,97,101}},
/* 33894 */ {(16<<2)|3,{65,97,105}},
/* 33895 */ {(16<<2)|3,{65,97,111}},
/* 33896 */ {(16<<2)|3,{65,97,115}},
/* 33897 */ {(16<<2)|3,{65,97,116}},
/* 33898 */ {(11<<2)|2,{65,97,0}},
/* 33899 */ {(11<<2)|2,{65,97,0}},
/* 33900 */ {(11<<2)|2,{65,97,0}},
/* 33901 */ {(11<<2)|2,{65,97,0}},
/* 33902 */ {(11<<2)|2,{65,97,0}},
/* 33903 */ {(11<<2)|2,{65,97,0}},
/* 33904 */ {(11<<2)|2,{65,97,0}},
/* 33905 */ {(11<<2)|2,{65,97,0}},
/* 33906 */ {(11<<2)|2,{65,97,0}},
/* 33907 */ {(11<<2)|2,{65,97,0}},
/* 33908 */ {(11<<2)|2,{65,97,0}},
/* 33909 */ {(11<<2)|2,{65,97,0}},
/* 33910 */ {(11<<2)|2,{65,97,0}},
/* 33911 */ {(11<<2)|2,{65,97,0}},
/* 33912 */ {(11<<2)|2,{65,97,0}},
/* 33913 */ {(11<<2)|2,{65,97,0}},
/* 33914 */ {(11<<2)|2,{65,97,0}},
/* 33915 */ {(11<<2)|2,{65,97,0}},
/* 33916 */ {(11<<2)|2,{65,97,0}},
/* 33917 */ {(11<<2)|2,{65,97,0}},
/* 33918 */ {(11<<2)|2,{65,97,0}},
/* 33919 */ {(11<<2)|2,{65,97,0}},
/* 33920 */ {(16<<2)|3,{65,99,48}},
/* 33921 */ {(16<<2)|3,{65,99,49}},
/* 33922 */ {(16<<2)|3,{65,99,50}},
/* 33923 */ {(16<<2)|3,{65,99,97}},
/* 33924 */ {(16<<2)|3,{65,99,99}},
/* 33925 */ {(16<<2)|3,{65,99,101}},
/* 33926 */ {(16<<2)|3,{65,99,105}},
/* 33927 */ {(16<<2)|3,{65,99,111}},
/* 33928 */ {(16<<2)|3,{65,99,115}},
/* 33929 */ {(16<<2)|3,{65,99,116}},
/* 33930 */ {(11<<2)|2,{65,99,0}},
/* 33931 */ {(11<<2)|2,{65,99,0}},
/* 33932 */ {(11<<2)|2,{65,99,0}},
/* 33933 */ {(11<<2)|2,{65,99,0}},
/* 33934 */ {(11<<2)|2,{65,99,0}},
/* 33935 */ {(11<<2)|2,{65,99,0}},
/* 33936 */ {(11<<2)|2,{65,99,0}},
/* 33937 */ {(11<<2)|2,{65,99,0}},
/* 33938 */ {(11<<2)|2,{65,99,0}},
/* 33939 */ {(11<<2)|2,{65,99,0}},
/* 33940 */ {(11<<2)|2,{65,99,0}},
/* 33941 */ {(11<<2)|2,{65,99,0}},
/* 33942 */ {(11<<2)|2,{65,99,0}},
/* 33943 */ {(11<<2)|2,{65,99,0}},
/* 33944 */ {(11<<2)|2,{65,99,0}},
/* 33945 */ {(11<<2)|2,{65,99,0}},
/* 33946 */ {(11<<2)|2,{65,99,0}},
/* 33947 */ {(11<<2)|2,{65,99,0}},
/* 33948 */ {(11<<2)|2,{65,99,0}},
/* 33949 */ {(11<<2)|2,{65,99,0}},
/* 33950 */ {(11<<2)|2,{65,99,0}},
/* 33951 */ {(11<<2)|2,{65,99,0}},
/* 33952 */ {(16<<2)|3,{65,101,48}},
/* 33953 */ {(16<<2)|3,{65,101,49}},
/* 33954 */ {(16<<2)|3,{65,101,50}},
/* 33955 */ {(16<<2)|3,{65,101,97}},
/* 33956 */ {(16<<2)|3,{65,101,99}},
/* 33957 */ {(16<<2)|3,{65,101,101}},
/* 33958 */ {(16<<2)|3,{65,101,105}},
/* 33959 */ {(16<<2)|3,{65,101,111}},
/* 33960 */ {(16<<2)|3,{65,101,115}},
/* 33961 */ {(16<<2)|3,{65,101,116}},
/* 33962 */ {(11<<2)|2,{65,101,0}},
/* 33963 */ {(11<<2)|2,{65,101,0}},
/* 33964 */ {(11<<2)|2,{65,101,0}},
/* 33965 */ {(11<<2)|2,{65,101,0}},
/* 33966 */ {(11<<2)|2,{65,101,0}},
/* 33967 */ {(11<<2)|2,{65,101,0}},
/* 33968 */ {(11<<2)|2,{65,101,0}},
/* 33969 */ {(11<<2)|2,{65,101,0}},
/* 33970 */ {(11<<2)|2,{65,101,0}},
/* 33971 */ {(11<<2)|2,{65,101,0}},
/* 33972 */ {(11<<2)|2,{65,101,0}},
/* 33973 */ {(11<<2)|2,{65,101,0}},
/* 33974 */ {(11<<2)|2,{65,101,0}},
/* 33975 */ {(11<<2)|2,{65,101,0}},
/* 33976 */ {(11<<2)|2,{65,101,0}},
/* 33977 */ {(11<<2)|2,{65,101,0}},
/* 33978 */ {(11<<2)|2,{65,101,0}},
/* 33979 */ {(11<<2)|2,{65,101,0}},
/* 33980 */ {(11<<2)|2,{65,101,0}},
/* 33981 */ {(11<<2)|2,{65,101,0}},
/* 33982 */ {(11<<2)|2,{65,101,0}},
/* 33983 */ {(11<<2)|2,{65,101,0}},
/* 33984 */ {(16<<2)|3,{65,105,48}},
/* 33985 */ {(16<<2)|3,{65,105,49}},
/* 33986 */ {(16<<2)|3,{65,105,50}},
/* 33987 */ {(16<<2)|3,{65,105,97}},
/* 33988 */ {(16<<2)|3,{65,105,99}},
/* 33989 */ {(16<<2)|3,{65,105,101}},
/* 33990 */ {(16<<2)|3,{65,105,105}},
/* 33991 */ {(16<<2)|3,{65,105,111}},
/* 33992 */ {(16<<2)|3,{65,105,115}},
/* 33993 */ {(16<<2)|3,{65,105,116}},
/* 33994 */ {(11<<2)|2,{65,105,0}},
/* 33995 */ {(11<<2)|2,{65,105,0}},
/* 33996 */ {(11<<2)|2,{65,105,0}},
/* 33997 */ {(11<<2)|2,{65,105,0}},
/* 33998 */ {(11<<2)|2,{65,105,0}},
/* 33999 */ {(11<<2)|2,{65,105,0}},
/* 34000 */ {(11<<2)|2,{65,105,0}},
/* 34001 */ {(11<<2)|2,{65,105,0}},
/* 34002 */ {(11<<2)|2,{65,105,0}},
/* 34003 */ {(11<<2)|2,{65,105,0}},
/* 34004 */ {(11<<2)|2,{65,105,0}},
/* 34005 */ {(11<<2)|2,{65,105,0}},
/* 34006 */ {(11<<2)|2,{65,105,0}},
/* 34007 */ {(11<<2)|2,{65,105,0}},
/* 34008 */ {(11<<2)|2,{65,105,0}},
/* 34009 */ {(11<<2)|2,{65,105,0}},
/* 34010 */ {(11<<2)|2,{65,105,0}},
/* 34011 */ {(11<<2)|2,{65,105,0}},
/* 34012 */ {(11<<2)|2,{65,105,0}},
/* 34013 */ {(11<<2)|2,{65,105,0}},
/* 34014 */ {(11<<2)|2,{65,105,0}},
/* 34015 */ {(11<<2)|2,{65,105,0}},
/* 34016 */ {(16<<2)|3,{65,111,48}},
/* 34017 */ {(16<<2)|3,{65,111,49}},
/* 34018 */ {(16<<2)|3,{65,111,50}},
/* 34019 */ {(16<<2)|3,{65,111,97}},
/* 34020 */ {(16<<2)|3,{65,111,99}},
/* 34021 */ {(16<<2)|3,{65,111,101}},
/* 34022 */ {(16<<2)|3,{65,111,105}},
/* 34023 */ {(16<<2)|3,{65,111,111}},
/* 34024 */ {(16<<2)|3,{65,111,115}},
/* 34025 */ {(16<<2)|3,{65,111,116}},
/* 34026 */ {(11<<2)|2,{65,111,0}},
/* 34027 */ {(11<<2)|2,{65,111,0}},
/* 34028 */ {(11<<2)|2,{65,111,0}},
/* 34029 */ {(11<<2)|2,{65,111,0}},
/* 34030 */ {(11<<2)|2,{65,111,0}},
/* 34031 */ {(11<<2)|2,{65,111,0}},
/* 34032 */ {(11<<2)|2,{65,111,0}},
/* 34033 */ {(11<<2)|2,{65,111,0}},
/* 34034 */ {(11<<2)|2,{65,111,0}},
/* 34035 */ {(11<<2)|2,{65,111,0}},
/* 34036 */ {(11<<2)|2,{65,111,0}},
/* 34037 */ {(11<<2)|2,{65,111,0}},
/* 34038 */ {(11<<2)|2,{65,111,0}},
/* 34039 */ {(11<<2)|2,{65,111,0}},
/* 34040 */ {(11<<2)|2,{65,111,0}},
/* 34041 */ {(11<<2)|2,{65,111,0}},
/* 34042 */ {(11<<2)|2,{65,111,0}},
/* 34043 */ {(11<<2)|2,{65,111,0}},
/* 34044 */ {(11<<2)|2,{65,111,0}},
/* 34045 */ {(11<<2)|2,{65,111,0}},
/* 34046 */ {(11<<2)|2,{65,111,0}},
/* 34047 */ {(11<<2)|2,{65,111,0}},
/* 34048 */ {(16<<2)|3,{65,115,48}},
/* 34049 */ {(16<<2)|3,{65,115,49}},
/* 34050 */ {(16<<2)|3,{65,115,50}},
/* 34051 */ {(16<<2)|3,{65,115,97}},
/* 34052 */ {(16<<2)|3,{65,115,99}},
/* 34053 */ {(16<<2)|3,{65,115,101}},
/* 34054 */ {(16<<2)|3,{65,115,105}},
/* 34055 */ {(16<<2)|3,{65,115,111}},
/* 34056 */ {(16<<2)|3,{65,115,115}},
/* 34057 */ {(16<<2)|3,{65,115,116}},
/* 34058 */ {(11<<2)|2,{65,115,0}},
/* 34059 */ {(11<<2)|2,{65,115,0}},
/* 34060 */ {(11<<2)|2,{65,115,0}},
/* 34061 */ {(11<<2)|2,{65,115,0}},
/* 34062 */ {(11<<2)|2,{65,115,0}},
/* 34063 */ {(11<<2)|2,{65,115,0}},
/* 34064 */ {(11<<2)|2,{65,115,0}},
/* 34065 */ {(11<<2)|2,{65,115,0}},
/* 34066 */ {(11<<2)|2,{65,115,0}},
/* 34067 */ {(11<<2)|2,{65,115,0}},
/* 34068 */ {(11<<2)|2,{65,115,0}},
/* 34069 */ {(11<<2)|2,{65,115,0}},
/* 34070 */ {(11<<2)|2,{65,115,0}},
/* 34071 */ {(11<<2)|2,{65,115,0}},
/* 34072 */ {(11<<2)|2,{65,115,0}},
/* 34073 */ {(11<<2)|2,{65,115,0}},
/* 34074 */ {(11<<2)|2,{65,115,0}},
/* 34075 */ {(11<<2)|2,{65,115,0}},
/* 34076 */ {(11<<2)|2,{65,115,0}},
/* 34077 */ {(11<<2)|2,{65,115,0}},
/* 34078 */ {(11<<2)|2,{65,115,0}},
/* 34079 */ {(11<<2)|2,{65,115,0}},
/* 34080 */ {(16<<2)|3,{65,116,48}},
/* 34081 */ {(16<<2)|3,{65,116,49}},
/* 34082 */ {(16<<2)|3,{65,116,50}},
/* 34083 */ {(16<<2)|3,{65,116,97}},
/* 34084 */ {(16<<2)|3,{65,116,99}},
/* 34085 */ {(16<<2)|3,{65,116,101}},
/* 34086 */ {(16<<2)|3,{65,116,105}},
/* 34087 */ {(16<<2)|3,{65,116,111}},
/* 34088 */ {(16<<2)|3,{65,116,115}},
/* 34089 */ {(16<<2)|3,{65,116,116}},
/* 34090 */ {(11<<2)|2,{65,116,0}},
/* 34091 */ {(11<<2)|2,{65,116,0}},
/* 34092 */ {(11<<2)|2,{65,116,0}},
/* 34093 */ {(11<<2)|2,{65,116,0}},
/* 34094 */ {(11<<2)|2,{65,116,0}},
/* 34095 */ {(11<<2)|2,{65,116,0}},
/* 34096 */ {(11<<2)|2,{65,116,0}},
/* 34097 */ {(11<<2)|2,{65,116,0}},
/* 34098 */ {(11<<2)|2,{65,116,0}},
/* 34099 */ {(11<<2)|2,{65,116,0}},
/* 34100 */ {(11<<2)|2,{65,116,0}},
/* 34101 */ {(11<<2)|2,{65,116,0}},
/* 34102 */ {(11<<2)|2,{65,116,0}},
/* 34103 */ {(11<<2)|2,{65,116,0}},
/* 34104 */ {(11<<2)|2,{65,116,0}},
/* 34105 */ {(11<<2)|2,{65,116,0}},
/* 34106 */ {(11<<2)|2,{65,116,0}},
/* 34107 */ {(11<<2)|2,{65,116,0}},
/* 34108 */ {(11<<2)|2,{65,116,0}},
/* 34109 */ {(11<<2)|2,{65,116,0}},
/* 34110 */ {(11<<2)|2,{65,116,0}},
/* 34111 */ {(11<<2)|2,{65,116,0}},
/* 34112 */ {(12<<2)|2,{65,32,0}},
/* 34113 */ {(12<<2)|2,{65,32,0}},
/* 34114 */ {(12<<2)|2,{65,32,0}},
/* 34115 */ {(12<<2)|2,{65,32,0}},
/* 34116 */ {(12<<2)|2,{65,32,0}},
/* 34117 */ {(12<<2)|2,{65,32,0}},
/* 34118 */ {(12<<2)|2,{65,32,0}},
/* 34119 */ {(12<<2)|2,{65,32,0}},
/* 34120 */ {(12<<2)|2,{65,32,0}},
/* 34121 */ {(12<<2)|2,{65,32,0}},
/* 34122 */ {(12<<2)|2,{65,32,0}},
/* 34123 */ {(12<<2)|2,{65,32,0}},
/* 34124 */ {(12<<2)|2,{65,32,0}},
/* 34125 */ {(12<<2)|2,{65,32,0}},
/* 34126 */ {(12<<2)|2,{65,32,0}},
/* 34127 */ {(12<<2)|2,{65,32,0}},
/* 34128 */ {(12<<2)|2,{65,37,0}},
/* 34129 */ {(12<<2)|2,{65,37,0}},
/* 34130 */ {(12<<2)|2,{65,37,0}},
/* 34131 */ {(12<<2)|2,{65,37,0}},
/* 34132 */ {(12<<2)|2,{65,37,0}},
/* 34133 */ {(12<<2)|2,{65,37,0}},
/* 34134 */ {(12<<2)|2,{65,37,0}},
/* 34135 */ {(12<<2)|2,{65,37,0}},
/* 34136 */ {(12<<2)|2,{65,37,0}},
/* 34137 */ {(12<<2)|2,{65,37,0}},
/* 34138 */ {(12<<2)|2,{65,37,0}},
/* 34139 */ {(12<<2)|2,{65,37,0}},
/* 34140 */ {(12<<2)|2,{65,37,0}},
/* 34141 */ {(12<<2)|2,{65,37,0}},
/* 34142 */ {(12<<2)|2,{65,37,0}},
/* 34143 */ {(12<<2)|2,{65,37,0}},
/* 34144 */ {(12<<2)|2,{65,45,0}},
/* 34145 */ {(12<<2)|2,{65,45,0}},
/* 34146 */ {(12<<2)|2,{65,45,0}},
/* 34147 */ {(12<<2)|2,{65,45,0}},
/* 34148 */ {(12<<2)|2,{65,45,0}},
/* 34149 */ {(12<<2)|2,{65,45,0}},
/* 34150 */ {(12<<2)|2,{65,45,0}},
/* 34151 */ {(12<<2)|2,{65,45,0}},
/* 34152 */ {(12<<2)|2,{65,45,0}},
/* 34153 */ {(12<<2)|2,{65,45,0}},
/* 34154 */ {(12<<2)|2,{65,45,0}},
/* 34155 */ {(12<<2)|2,{65,45,0}},
/* 34156 */ {(12<<2)|2,{65,45,0}},
/* 34157 */ {(12<<2)|2,{65,45,0}},
/* 34158 */ {(12<<2)|2,{65,45,0}},
/* 34159 */ {(12<<2)|2,{65,45,0}},
/* 34160 */ {(12<<2)|2,{65,46,0}},
/* 34161 */ {(12<<2)|2,{65,46,0}},
/* 34162 */ {(12<<2)|2,{65,46,0}},
/* 34163 */ {(12<<2)|2,{65,46,0}},
/* 34164 */ {(12<<2)|2,{65,46,0}},
/* 34165 */ {(12<<2)|2,{65,46,0}},
/* 34166 */ {(12<<2)|2,{65,46,0}},
/* 34167 */ {(12<<2)|2,{65,46,0}},
/* 34168 */ {(12<<2)|2,{65,46,0}},
/* 34169 */ {(12<<2)|2,{65,46,0}},
/* 34170 */ {(12<<2)|2,{65,46,0}},
/* 34171 */ {(12<<2)|2,{65,46,0}},
/* 34172 */ {(12<<2)|2,{65,46,0}},
/* 34173 */ {(12<<2)|2,{65,46,0}},
/* 34174 */ {(12<<2)|2,{65,46,0}},
/* 34175 */ {(12<<2)|2,{65,46,0}},
/* 34176 */ {(12<<2)|2,{65,47,0}},
/* 34177 */ {(12<<2)|2,{65,47,0}},
/* 34178 */ {(12<<2)|2,{65,47,0}},
/* 34179 */ {(12<<2)|2,{65,47,0}},
/* 34180 */ {(12<<2)|2,{65,47,0}},
/* 34181 */ {(12<<2)|2,{65,47,0}},
/* 34182 */ {(12<<2)|2,{65,47,0}},
/* 34183 */ {(12<<2)|2,{65,47,0}},
/* 34184 */ {(12<<2)|2,{65,47,0}},
/* 34185 */ {(12<<2)|2,{65,47,0}},
/* 34186 */ {(12<<2)|2,{65,47,0}},
/* 34187 */ {(12<<2)|2,{65,47,0}},
/* 34188 */ {(12<<2)|2,{65,47,0}},
/* 34189 */ {(12<<2)|2,{65,47,0}},
/* 34190 */ {(12<<2)|2,{65,47,0}},
/* 34191 */ {(12<<2)|2,{65,47,0}},
/* 34192 */ {(12<<2)|2,{65,51,0}},
/* 34193 */ {(12<<2)|2,{65,51,0}},
/* 34194 */ {(12<<2)|2,{65,51,0}},
/* 34195 */ {(12<<2)|2,{65,51,0}},
/* 34196 */ {(12<<2)|2,{65,51,0}},
/* 34197 */ {(12<<2)|2,{65,51,0}},
/* 34198 */ {(12<<2)|2,{65,51,0}},
/* 34199 */ {(12<<2)|2,{65,51,0}},
/* 34200 */ {(12<<2)|2,{65,51,0}},
/* 34201 */ {(12<<2)|2,{65,51,0}},
/* 34202 */ {(12<<2)|2,{65,51,0}},
/* 34203 */ {(12<<2)|2,{65,51,0}},
/* 34204 */ {(12<<2)|2,{65,51,0}},
/* 34205 */ {(12<<2)|2,{65,51,0}},
/* 34206 */ {(12<<2)|2,{65,51,0}},
/* 34207 */ {(12<<2)|2,{65,51,0}},
/* 34208 */ {(12<<2)|2,{65,52,0}},
/* 34209 */ {(12<<2)|2,{65,52,0}},
/* 34210 */ {(12<<2)|2,{65,52,0}},
/* 34211 */ {(12<<2)|2,{65,52,0}},
/* 34212 */ {(12<<2)|2,{65,52,0}},
/* 34213 */ {(12<<2)|2,{65,52,0}},
/* 34214 */ {(12<<2)|2,{65,52,0}},
/* 34215 */ {(12<<2)|2,{65,52,0}},
/* 34216 */ {(12<<2)|2,{65,52,0}},
/* 34217 */ {(12<<2)|2,{65,52,0}},
/* 34218 */ {(12<<2)|2,{65,52,0}},
/* 34219 */ {(12<<2)|2,{65,52,0}},
/* 34220 */ {(12<<2)|2,{65,52,0}},
/* 34221 */ {(12<<2)|2,{65,52,0}},
/* 34222 */ {(12<<2)|2,{65,52,0}},
/* 34223 */ {(12<<2)|2,{65,52,0}},
/* 34224 */ {(12<<2)|2,{65,53,0}},
/* 34225 */ {(12<<2)|2,{65,53,0}},
/* 34226 */ {(12<<2)|2,{65,53,0}},
/* 34227 */ {(12<<2)|2,{65,53,0}},
/* 34228 */ {(12<<2)|2,{65,53,0}},
/* 34229 */ {(12<<2)|2,{65,53,0}},
/* 34230 */ {(12<<2)|2,{65,53,0}},
/* 34231 */ {(12<<2)|2,{65,53,0}},
/* 34232 */ {(12<<2)|2,{65,53,0}},
/* 34233 */ {(12<<2)|2,{65,53,0}},
/* 34234 */ {(12<<2)|2,{65,53,0}},
/* 34235 */ {(12<<2)|2,{65,53,0}},
/* 34236 */ {(12<<2)|2,{65,53,0}},
/* 34237 */ {(12<<2)|2,{65,53,0}},
/* 34238 */ {(12<<2)|2,{65,53,0}},
/* 34239 */ {(12<<2)|2,{65,53,0}},
/* 34240 */ {(12<<2)|2,{65,54,0}},
/* 34241 */ {(12<<2)|2,{65,54,0}},
/* 34242 */ {(12<<2)|2,{65,54,0}},
/* 34243 */ {(12<<2)|2,{65,54,0}},
/* 34244 */ {(12<<2)|2,{65,54,0}},
/* 34245 */ {(12<<2)|2,{65,54,0}},
/* 34246 */ {(12<<2)|2,{65,54,0}},
/* 34247 */ {(12<<2)|2,{65,54,0}},
/* 34248 */ {(12<<2)|2,{65,54,0}},
/* 34249 */ {(12<<2)|2,{65,54,0}},
/* 34250 */ {(12<<2)|2,{65,54,0}},
/* 34251 */ {(12<<2)|2,{65,54,0}},
/* 34252 */ {(12<<2)|2,{65,54,0}},
/* 34253 */ {(12<<2)|2,{65,54,0}},
/* 34254 */ {(12<<2)|2,{65,54,0}},
/* 34255 */ {(12<<2)|2,{65,54,0}},
/* 34256 */ {(12<<2)|2,{65,55,0}},
/* 34257 */ {(12<<2)|2,{65,55,0}},
/* 34258 */ {(12<<2)|2,{65,55,0}},
/* 34259 */ {(12<<2)|2,{65,55,0}},
/* 34260 */ {(12<<2)|2,{65,55,0}},
/* 34261 */ {(12<<2)|2,{65,55,0}},
/* 34262 */ {(12<<2)|2,{65,55,0}},
/* 34263 */ {(12<<2)|2,{65,55,0}},
/* 34264 */ {(12<<2)|2,{65,55,0}},
/* 34265 */ {(12<<2)|2,{65,55,0}},
/* 34266 */ {(12<<2)|2,{65,55,0}},
/* 34267 */ {(12<<2)|2,{65,55,0}},
/* 34268 */ {(12<<2)|2,{65,55,0}},
/* 34269 */ {(12<<2)|2,{65,55,0}},
/* 34270 */ {(12<<2)|2,{65,55,0}},
/* 34271 */ {(12<<2)|2,{65,55,0}},
/* 34272 */ {(12<<2)|2,{65,56,0}},
/* 34273 */ {(12<<2)|2,{65,56,0}},
/* 34274 */ {(12<<2)|2,{65,56,0}},
/* 34275 */ {(12<<2)|2,{65,56,0}},
/* 34276 */ {(12<<2)|2,{65,56,0}},
/* 34277 */ {(12<<2)|2,{65,56,0}},
/* 34278 */ {(12<<2)|2,{65,56,0}},
/* 34279 */ {(12<<2)|2,{65,56,0}},
/* 34280 */ {(12<<2)|2,{65,56,0}},
/* 34281 */ {(12<<2)|2,{65,56,0}},
/* 34282 */ {(12<<2)|2,{65,56,0}},
/* 34283 */ {(12<<2)|2,{65,56,0}},
/* 34284 */ {(12<<2)|2,{65,56,0}},
/* 34285 */ {(12<<2)|2,{65,56,0}},
/* 34286 */ {(12<<2)|2,{65,56,0}},
/* 34287 */ {(12<<2)|2,{65,56,0}},
/* 34288 */ {(12<<2)|2,{65,57,0}},
/* 34289 */ {(12<<2)|2,{65,57,0}},
/* 34290 */ {(12<<2)|2,{65,57,0}},
/* 34291 */ {(12<<2)|2,{65,57,0}},
/* 34292 */ {(12<<2)|2,{65,57,0}},
/* 34293 */ {(12<<2)|2,{65,57,0}},
/* 34294 */ {(12<<2)|2,{65,57,0}},
/* 34295 */ {(12<<2)|2,{65,57,0}},
/* 34296 */ {(12<<2)|2,{65,57,0}},
/* 34297 */ {(12<<2)|2,{65,57,0}},
/* 34298 */ {(12<<2)|2,{65,57,0}},
/* 34299 */ {(12<<2)|2,{65,57,0}},
/* 34300 */ {(12<<2)|2,{65,57,0}},
/* 34301 */ {(12<<2)|2,{65,57,0}},
/* 34302 */ {(12<<2)|2,{65,57,0}},
/* 34303 */ {(12<<2)|2,{65,57,0}},
/* 34304 */ {(12<<2)|2,{65,61,0}},
/* 34305 */ {(12<<2)|2,{65,61,0}},
/* 34306 */ {(12<<2)|2,{65,61,0}},
/* 34307 */ {(12<<2)|2,{65,61,0}},
/* 34308 */ {(12<<2)|2,{65,61,0}},
/* 34309 */ {(12<<2)|2,{65,61,0}},
/* 34310 */ {(12<<2)|2,{65,61,0}},
/* 34311 */ {(12<<2)|2,{65,61,0}},
/* 34312 */ {(12<<2)|2,{65,61,0}},
/* 34313 */ {(12<<2)|2,{65,61,0}},
/* 34314 */ {(12<<2)|2,{65,61,0}},
/* 34315 */ {(12<<2)|2,{65,61,0}},
/* 34316 */ {(12<<2)|2,{65,61,0}},
/* 34317 */ {(12<<2)|2,{65,61,0}},
/* 34318 */ {(12<<2)|2,{65,61,0}},
/* 34319 */ {(12<<2)|2,{65,61,0}},
/* 34320 */ {(12<<2)|2,{65,65,0}},
/* 34321 */ {(12<<2)|2,{65,65,0}},
/* 34322 */ {(12<<2)|2,{65,65,0}},
/* 34323 */ {(12<<2)|2,{65,65,0}},
/* 34324 */ {(12<<2)|2,{65,65,0}},
/* 34325 */ {(12<<2)|2,{65,65,0}},
/* 34326 */ {(12<<2)|2,{65,65,0}},
/* 34327 */ {(12<<2)|2,{65,65,0}},
/* 34328 */ {(12<<2)|2,{65,65,0}},
/* 34329 */ {(12<<2)|2,{65,65,0}},
/* 34330 */ {(12<<2)|2,{65,65,0}},
/* 34331 */ {(12<<2)|2,{65,65,0}},
/* 34332 */ {(12<<2)|2,{65,65,0}},
/* 34333 */ {(12<<2)|2,{65,65,0}},
/* 34334 */ {(12<<2)|2,{65,65,0}},
/* 34335 */ {(12<<2)|2,{65,65,0}},
/* 34336 */ {(12<<2)|2,{65,95,0}},
/* 34337 */ {(12<<2)|2,{65,95,0}},
/* 34338 */ {(12<<2)|2,{65,95,0}},
/* 34339 */ {(12<<2)|2,{65,95,0}},
/* 34340 */ {(12<<2)|2,{65,95,0}},
/* 34341 */ {(12<<2)|2,{65,95,0}},
/* 34342 */ {(12<<2)|2,{65,95,0}},
/* 34343 */ {(12<<2)|2,{65,95,0}},
/* 34344 */ {(12<<2)|2,{65,95,0}},
/* 34345 */ {(12<<2)|2,{65,95,0}},
/* 34346 */ {(12<<2)|2,{65,95,0}},
/* 34347 */ {(12<<2)|2,{65,95,0}},
/* 34348 */ {(12<<2)|2,{65,95,0}},
/* 34349 */ {(12<<2)|2,{65,95,0}},
/* 34350 */ {(12<<2)|2,{65,95,0}},
/* 34351 */ {(12<<2)|2,{65,95,0}},
/* 34352 */ {(12<<2)|2,{65,98,0}},
/* 34353 */ {(12<<2)|2,{65,98,0}},
/* 34354 */ {(12<<2)|2,{65,98,0}},
/* 34355 */ {(12<<2)|2,{65,98,0}},
/* 34356 */ {(12<<2)|2,{65,98,0}},
/* 34357 */ {(12<<2)|2,{65,98,0}},
/* 34358 */ {(12<<2)|2,{65,98,0}},
/* 34359 */ {(12<<2)|2,{65,98,0}},
/* 34360 */ {(12<<2)|2,{65,98,0}},
/* 34361 */ {(12<<2)|2,{65,98,0}},
/* 34362 */ {(12<<2)|2,{65,98,0}},
/* 34363 */ {(12<<2)|2,{65,98,0}},
/* 34364 */ {(12<<2)|2,{65,98,0}},
/* 34365 */ {(12<<2)|2,{65,98,0}},
/* 34366 */ {(12<<2)|2,{65,98,0}},
/* 34367 */ {(12<<2)|2,{65,98,0}},
/* 34368 */ {(12<<2)|2,{65,100,0}},
/* 34369 */ {(12<<2)|2,{65,100,0}},
/* 34370 */ {(12<<2)|2,{65,100,0}},
/* 34371 */ {(12<<2)|2,{65,100,0}},
/* 34372 */ {(12<<2)|2,{65,100,0}},
/* 34373 */ {(12<<2)|2,{65,100,0}},
/* 34374 */ {(12<<2)|2,{65,100,0}},
/* 34375 */ {(12<<2)|2,{65,100,0}},
/* 34376 */ {(12<<2)|2,{65,100,0}},
/* 34377 */ {(12<<2)|2,{65,100,0}},
/* 34378 */ {(12<<2)|2,{65,100,0}},
/* 34379 */ {(12<<2)|2,{65,100,0}},
/* 34380 */ {(12<<2)|2,{65,100,0}},
/* 34381 */ {(12<<2)|2,{65,100,0}},
/* 34382 */ {(12<<2)|2,{65,100,0}},
/* 34383 */ {(12<<2)|2,{65,100,0}},
/* 34384 */ {(12<<2)|2,{65,102,0}},
/* 34385 */ {(12<<2)|2,{65,102,0}},
/* 34386 */ {(12<<2)|2,{65,102,0}},
/* 34387 */ {(12<<2)|2,{65,102,0}},
/* 34388 */ {(12<<2)|2,{65,102,0}},
/* 34389 */ {(12<<2)|2,{65,102,0}},
/* 34390 */ {(12<<2)|2,{65,102,0}},
/* 34391 */ {(12<<2)|2,{65,102,0}},
/* 34392 */ {(12<<2)|2,{65,102,0}},
/* 34393 */ {(12<<2)|2,{65,102,0}},
/* 34394 */ {(12<<2)|2,{65,102,0}},
/* 34395 */ {(12<<2)|2,{65,102,0}},
/* 34396 */ {(12<<2)|2,{65,102,0}},
/* 34397 */ {(12<<2)|2,{65,102,0}},
/* 34398 */ {(12<<2)|2,{65,102,0}},
/* 34399 */ {(12<<2)|2,{65,102,0}},
/* 34400 */ {(12<<2)|2,{65,103,0}},
/* 34401 */ {(12<<2)|2,{65,103,0}},
/* 34402 */ {(12<<2)|2,{65,103,0}},
/* 34403 */ {(12<<2)|2,{65,103,0}},
/* 34404 */ {(12<<2)|2,{65,103,0}},
/* 34405 */ {(12<<2)|2,{65,103,0}},
/* 34406 */ {(12<<2)|2,{65,103,0}},
/* 34407 */ {(12<<2)|2,{65,103,0}},
/* 34408 */ {(12<<2)|2,{65,103,0}},
/* 34409 */ {(12<<2)|2,{65,103,0}},
/* 34410 */ {(12<<2)|2,{65,103,0}},
/* 34411 */ {(12<<2)|2,{65,103,0}},
/* 34412 */ {(12<<2)|2,{65,103,0}},
/* 34413 */ {(12<<2)|2,{65,103,0}},
/* 34414 */ {(12<<2)|2,{65,103,0}},
/* 34415 */ {(12<<2)|2,{65,103,0}},
/* 34416 */ {(12<<2)|2,{65,104,0}},
/* 34417 */ {(12<<2)|2,{65,104,0}},
/* 34418 */ {(12<<2)|2,{65,104,0}},
/* 34419 */ {(12<<2)|2,{65,104,0}},
/* 34420 */ {(12<<2)|2,{65,104,0}},
/* 34421 */ {(12<<2)|2,{65,104,0}},
/* 34422 */ {(12<<2)|2,{65,104,0}},
/* 34423 */ {(12<<2)|2,{65,104,0}},
/* 34424 */ {(12<<2)|2,{65,104,0}},
/* 34425 */ {(12<<2)|2,{65,104,0}},
/* 34426 */ {(12<<2)|2,{65,104,0}},
/* 34427 */ {(12<<2)|2,{65,104,0}},
/* 34428 */ {(12<<2)|2,{65,104,0}},
/* 34429 */ {(12<<2)|2,{65,104,0}},
/* 34430 */ {(12<<2)|2,{65,104,0}},
/* 34431 */ {(12<<2)|2,{65,104,0}},
/* 34432 */ {(12<<2)|2,{65,108,0}},
/* 34433 */ {(12<<2)|2,{65,108,0}},
/* 34434 */ {(12<<2)|2,{65,108,0}},
/* 34435 */ {(12<<2)|2,{65,108,0}},
/* 34436 */ {(12<<2)|2,{65,108,0}},
/* 34437 */ {(12<<2)|2,{65,108,0}},
/* 34438 */ {(12<<2)|2,{65,108,0}},
/* 34439 */ {(12<<2)|2,{65,108,0}},
/* 34440 */ {(12<<2)|2,{65,108,0}},
/* 34441 */ {(12<<2)|2,{65,108,0}},
/* 34442 */ {(12<<2)|2,{65,108,0}},
/* 34443 */ {(12<<2)|2,{65,108,0}},
/* 34444 */ {(12<<2)|2,{65,108,0}},
/* 34445 */ {(12<<2)|2,{65,108,0}},
/* 34446 */ {(12<<2)|2,{65,108,0}},
/* 34447 */ {(12<<2)|2,{65,108,0}},
/* 34448 */ {(12<<2)|2,{65,109,0}},
/* 34449 */ {(12<<2)|2,{65,109,0}},
/* 34450 */ {(12<<2)|2,{65,109,0}},
/* 34451 */ {(12<<2)|2,{65,109,0}},
/* 34452 */ {(12<<2)|2,{65,109,0}},
/* 34453 */ {(12<<2)|2,{65,109,0}},
/* 34454 */ {(12<<2)|2,{65,109,0}},
/* 34455 */ {(12<<2)|2,{65,109,0}},
/* 34456 */ {(12<<2)|2,{65,109,0}},
/* 34457 */ {(12<<2)|2,{65,109,0}},
/* 34458 */ {(12<<2)|2,{65,109,0}},
/* 34459 */ {(12<<2)|2,{65,109,0}},
/* 34460 */ {(12<<2)|2,{65,109,0}},
/* 34461 */ {(12<<2)|2,{65,109,0}},
/* 34462 */ {(12<<2)|2,{65,109,0}},
/* 34463 */ {(12<<2)|2,{65,109,0}},
/* 34464 */ {(12<<2)|2,{65,110,0}},
/* 34465 */ {(12<<2)|2,{65,110,0}},
/* 34466 */ {(12<<2)|2,{65,110,0}},
/* 34467 */ {(12<<2)|2,{65,110,0}},
/* 34468 */ {(12<<2)|2,{65,110,0}},
/* 34469 */ {(12<<2)|2,{65,110,0}},
/* 34470 */ {(12<<2)|2,{65,110,0}},
/* 34471 */ {(12<<2)|2,{65,110,0}},
/* 34472 */ {(12<<2)|2,{65,110,0}},
/* 34473 */ {(12<<2)|2,{65,110,0}},
/* 34474 */ {(12<<2)|2,{65,110,0}},
/* 34475 */ {(12<<2)|2,{65,110,0}},
/* 34476 */ {(12<<2)|2,{65,110,0}},
/* 34477 */ {(12<<2)|2,{65,110,0}},
/* 34478 */ {(12<<2)|2,{65,110,0}},
/* 34479 */ {(12<<2)|2,{65,110,0}},
/* 34480 */ {(12<<2)|2,{65,112,0}},
/* 34481 */ {(12<<2)|2,{65,112,0}},
/* 34482 */ {(12<<2)|2,{65,112,0}},
/* 34483 */ {(12<<2)|2,{65,112,0}},
/* 34484 */ {(12<<2)|2,{65,112,0}},
/* 34485 */ {(12<<2)|2,{65,112,0}},
/* 34486 */ {(12<<2)|2,{65,112,0}},
/* 34487 */ {(12<<2)|2,{65,112,0}},
/* 34488 */ {(12<<2)|2,{65,112,0}},
/* 34489 */ {(12<<2)|2,{65,112,0}},
/* 34490 */ {(12<<2)|2,{65,112,0}},
/* 34491 */ {(12<<2)|2,{65,112,0}},
/* 34492 */ {(12<<2)|2,{65,112,0}},
/* 34493 */ {(12<<2)|2,{65,112,0}},
/* 34494 */ {(12<<2)|2,{65,112,0}},
/* 34495 */ {(12<<2)|2,{65,112,0}},
/* 34496 */ {(12<<2)|2,{65,114,0}},
/* 34497 */ {(12<<2)|2,{65,114,0}},
/* 34498 */ {(12<<2)|2,{65,114,0}},
/* 34499 */ {(12<<2)|2,{65,114,0}},
/* 34500 */ {(12<<2)|2,{65,114,0}},
/* 34501 */ {(12<<2)|2,{65,114,0}},
/* 34502 */ {(12<<2)|2,{65,114,0}},
/* 34503 */ {(12<<2)|2,{65,114,0}},
/* 34504 */ {(12<<2)|2,{65,114,0}},
/* 34505 */ {(12<<2)|2,{65,114,0}},
/* 34506 */ {(12<<2)|2,{65,114,0}},
/* 34507 */ {(12<<2)|2,{65,114,0}},
/* 34508 */ {(12<<2)|2,{65,114,0}},
/* 34509 */ {(12<<2)|2,{65,114,0}},
/* 34510 */ {(12<<2)|2,{65,114,0}},
/* 34511 */ {(12<<2)|2,{65,114,0}},
/* 34512 */ {(12<<2)|2,{65,117,0}},
/* 34513 */ {(12<<2)|2,{65,117,0}},
/* 34514 */ {(12<<2)|2,{65,117,0}},
/* 34515 */ {(12<<2)|2,{65,117,0}},
/* 34516 */ {(12<<2)|2,{65,117,0}},
/* 34517 */ {(12<<2)|2,{65,117,0}},
/* 34518 */ {(12<<2)|2,{65,117,0}},
/* 34519 */ {(12<<2)|2,{65,117,0}},
/* 34520 */ {(12<<2)|2,{65,117,0}},
/* 34521 */ {(12<<2)|2,{65,117,0}},
/* 34522 */ {(12<<2)|2,{65,117,0}},
/* 34523 */ {(12<<2)|2,{65,117,0}},
/* 34524 */ {(12<<2)|2,{65,117,0}},
/* 34525 */ {(12<<2)|2,{65,117,0}},
/* 34526 */ {(12<<2)|2,{65,117,0}},
/* 34527 */ {(12<<2)|2,{65,117,0}},
/* 34528 */ {(13<<2)|2,{65,58,0}},
/* 34529 */ {(13<<2)|2,{65,58,0}},
/* 34530 */ {(13<<2)|2,{65,58,0}},
/* 34531 */ {(13<<2)|2,{65,58,0}},
/* 34532 */ {(13<<2)|2,{65,58,0}},
/* 34533 */ {(13<<2)|2,{65,58,0}},
/* 34534 */ {(13<<2)|2,{65,58,0}},
/* 34535 */ {(13<<2)|2,{65,58,0}},
/* 34536 */ {(13<<2)|2,{65,66,0}},
/* 34537 */ {(13<<2)|2,{65,66,0}},
/* 34538 */ {(13<<2)|2,{65,66,0}},
/* 34539 */ {(13<<2)|2,{65,66,0}},
/* 34540 */ {(13<<2)|2,{65,66,0}},
/* 34541 */ {(13<<2)|2,{65,66,0}},
/* 34542 */ {(13<<2)|2,{65,66,0}},
/* 34543 */ {(13<<2)|2,{65,66,0}},
/* 34544 */ {(13<<2)|2,{65,67,0}},
/* 34545 */ {(13<<2)|2,{65,67,0}},
/* 34546 */ {(13<<2)|2,{65,67,0}},
/* 34547 */ {(13<<2)|2,{65,67,0}},
/* 34548 */ {(13<<2)|2,{65,67,0}},
/* 34549 */ {(13<<2)|2,{65,67,0}},
/* 34550 */ {(13<<2)|2,{65,67,0}},
/* 34551 */ {(13<<2)|2,{65,67,0}},
/* 34552 */ {(13<<2)|2,{65,68,0}},
/* 34553 */ {(13<<2)|2,{65,68,0}},
/* 34554 */ {(13<<2)|2,{65,68,0}},
/* 34555 */ {(13<<2)|2,{65,68,0}},
/* 34556 */ {(13<<2)|2,{65,68,0}},
/* 34557 */ {(13<<2)|2,{65,68,0}},
/* 34558 */ {(13<<2)|2,{65,68,0}},
/* 34559 */ {(13<<2)|2,{65,68,0}},
/* 34560 */ {(13<<2)|2,{65,69,0}},
/* 34561 */ {(13<<2)|2,{65,69,0}},
/* 34562 */ {(13<<2)|2,{65,69,0}},
/* 34563 */ {(13<<2)|2,{65,69,0}},
/* 34564 */ {(13<<2)|2,{65,69,0}},
/* 34565 */ {(13<<2)|2,{65,69,0}},
/* 34566 */ {(13<<2)|2,{65,69,0}},
/* 34567 */ {(13<<2)|2,{65,69,0}},
/* 34568 */ {(13<<2)|2,{65,70,0}},
/* 34569 */ {(13<<2)|2,{65,70,0}},
/* 34570 */ {(13<<2)|2,{65,70,0}},
/* 34571 */ {(13<<2)|2,{65,70,0}},
/* 34572 */ {(13<<2)|2,{65,70,0}},
/* 34573 */ {(13<<2)|2,{65,70,0}},
/* 34574 */ {(13<<2)|2,{65,70,0}},
/* 34575 */ {(13<<2)|2,{65,70,0}},
/* 34576 */ {(13<<2)|2,{65,71,0}},
/* 34577 */ {(13<<2)|2,{65,71,0}},
/* 34578 */ {(13<<2)|2,{65,71,0}},
/* 34579 */ {(13<<2)|2,{65,71,0}},
/* 34580 */ {(13<<2)|2,{65,71,0}},
/* 34581 */ {(13<<2)|2,{65,71,0}},
/* 34582 */ {(13<<2)|2,{65,71,0}},
/* 34583 */ {(13<<2)|2,{65,71,0}},
/* 34584 */ {(13<<2)|2,{65,72,0}},
/* 34585 */ {(13<<2)|2,{65,72,0}},
/* 34586 */ {(13<<2)|2,{65,72,0}},
/* 34587 */ {(13<<2)|2,{65,72,0}},
/* 34588 */ {(13<<2)|2,{65,72,0}},
/* 34589 */ {(13<<2)|2,{65,72,0}},
/* 34590 */ {(13<<2)|2,{65,72,0}},
/* 34591 */ {(13<<2)|2,{65,72,0}},
/* 34592 */ {(13<<2)|2,{65,73,0}},
/* 34593 */ {(13<<2)|2,{65,73,0}},
/* 34594 */ {(13<<2)|2,{65,73,0}},
/* 34595 */ {(13<<2)|2,{65,73,0}},
/* 34596 */ {(13<<2)|2,{65,73,0}},
/* 34597 */ {(13<<2)|2,{65,73,0}},
/* 34598 */ {(13<<2)|2,{65,73,0}},
/* 34599 */ {(13<<2)|2,{65,73,0}},
/* 34600 */ {(13<<2)|2,{65,74,0}},
/* 34601 */ {(13<<2)|2,{65,74,0}},
/* 34602 */ {(13<<2)|2,{65,74,0}},
/* 34603 */ {(13<<2)|2,{65,74,0}},
/* 34604 */ {(13<<2)|2,{65,74,0}},
/* 34605 */ {(13<<2)|2,{65,74,0}},
/* 34606 */ {(13<<2)|2,{65,74,0}},
/* 34607 */ {(13<<2)|2,{65,74,0}},
/* 34608 */ {(13<<2)|2,{65,75,0}},
/* 34609 */ {(13<<2)|2,{65,75,0}},
/* 34610 */ {(13<<2)|2,{65,75,0}},
/* 34611 */ {(13<<2)|2,{65,75,0}},
/* 34612 */ {(13<<2)|2,{65,75,0}},
/* 34613 */ {(13<<2)|2,{65,75,0}},
/* 34614 */ {(13<<2)|2,{65,75,0}},
/* 34615 */ {(13<<2)|2,{65,75,0}},
/* 34616 */ {(13<<2)|2,{65,76,0}},
/* 34617 */ {(13<<2)|2,{65,76,0}},
/* 34618 */ {(13<<2)|2,{65,76,0}},
/* 34619 */ {(13<<2)|2,{65,76,0}},
/* 34620 */ {(13<<2)|2,{65,76,0}},
/* 34621 */ {(13<<2)|2,{65,76,0}},
/* 34622 */ {(13<<2)|2,{65,76,0}},
/* 34623 */ {(13<<2)|2,{65,76,0}},
/* 34624 */ {(13<<2)|2,{65,77,0}},
/* 34625 */ {(13<<2)|2,{65,77,0}},
/* 34626 */ {(13<<2)|2,{65,77,0}},
/* 34627 */ {(13<<2)|2,{65,77,0}},
/* 34628 */ {(13<<2)|2,{65,77,0}},
/* 34629 */ {(13<<2)|2,{65,77,0}},
/* 34630 */ {(13<<2)|2,{65,77,0}},
/* 34631 */ {(13<<2)|2,{65,77,0}},
/* 34632 */ {(13<<2)|2,{65,78,0}},
/* 34633 */ {(13<<2)|2,{65,78,0}},
/* 34634 */ {(13<<2)|2,{65,78,0}},
/* 34635 */ {(13<<2)|2,{65,78,0}},
/* 34636 */ {(13<<2)|2,{65,78,0}},
/* 34637 */ {(13<<2)|2,{65,78,0}},
/* 34638 */ {(13<<2)|2,{65,78,0}},
/* 34639 */ {(13<<2)|2,{65,78,0}},
/* 34640 */ {(13<<2)|2,{65,79,0}},
/* 34641 */ {(13<<2)|2,{65,79,0}},
/* 34642 */ {(13<<2)|2,{65,79,0}},
/* 34643 */ {(13<<2)|2,{65,79,0}},
/* 34644 */ {(13<<2)|2,{65,79,0}},
/* 34645 */ {(13<<2)|2,{65,79,0}},
/* 34646 */ {(13<<2)|2,{65,79,0}},
/* 34647 */ {(13<<2)|2,{65,79,0}},
/* 34648 */ {(13<<2)|2,{65,80,0}},
/* 34649 */ {(13<<2)|2,{65,80,0}},
/* 34650 */ {(13<<2)|2,{65,80,0}},
/* 34651 */ {(13<<2)|2,{65,80,0}},
/* 34652 */ {(13<<2)|2,{65,80,0}},
/* 34653 */ {(13<<2)|2,{65,80,0}},
/* 34654 */ {(13<<2)|2,{65,80,0}},
/* 34655 */ {(13<<2)|2,{65,80,0}},
/* 34656 */ {(13<<2)|2,{65,81,0}},
/* 34657 */ {(13<<2)|2,{65,81,0}},
/* 34658 */ {(13<<2)|2,{65,81,0}},
/* 34659 */ {(13<<2)|2,{65,81,0}},
/* 34660 */ {(13<<2)|2,{65,81,0}},
/* 34661 */ {(13<<2)|2,{65,81,0}},
/* 34662 */ {(13<<2)|2,{65,81,0}},
/* 34663 */ {(13<<2)|2,{65,81,0}},
/* 34664 */ {(13<<2)|2,{65,82,0}},
/* 34665 */ {(13<<2)|2,{65,82,0}},
/* 34666 */ {(13<<2)|2,{65,82,0}},
/* 34667 */ {(13<<2)|2,{65,82,0}},
/* 34668 */ {(13<<2)|2,{65,82,0}},
/* 34669 */ {(13<<2)|2,{65,82,0}},
/* 34670 */ {(13<<2)|2,{65,82,0}},
/* 34671 */ {(13<<2)|2,{65,82,0}},
/* 34672 */ {(13<<2)|2,{65,83,0}},
/* 34673 */ {(13<<2)|2,{65,83,0}},
/* 34674 */ {(13<<2)|2,{65,83,0}},
/* 34675 */ {(13<<2)|2,{65,83,0}},
/* 34676 */ {(13<<2)|2,{65,83,0}},
/* 34677 */ {(13<<2)|2,{65,83,0}},
/* 34678 */ {(13<<2)|2,{65,83,0}},
/* 34679 */ {(13<<2)|2,{65,83,0}},
/* 34680 */ {(13<<2)|2,{65,84,0}},
/* 34681 */ {(13<<2)|2,{65,84,0}},
/* 34682 */ {(13<<2)|2,{65,84,0}},
/* 34683 */ {(13<<2)|2,{65,84,0}},
/* 34684 */ {(13<<2)|2,{65,84,0}},
/* 34685 */ {(13<<2)|2,{65,84,0}},
/* 34686 */ {(13<<2)|2,{65,84,0}},
/* 34687 */ {(13<<2)|2,{65,84,0}},
/* 34688 */ {(13<<2)|2,{65,85,0}},
/* 34689 */ {(13<<2)|2,{65,85,0}},
/* 34690 */ {(13<<2)|2,{65,85,0}},
/* 34691 */ {(13<<2)|2,{65,85,0}},
/* 34692 */ {(13<<2)|2,{65,85,0}},
/* 34693 */ {(13<<2)|2,{65,85,0}},
/* 34694 */ {(13<<2)|2,{65,85,0}},
/* 34695 */ {(13<<2)|2,{65,85,0}},
/* 34696 */ {(13<<2)|2,{65,86,0}},
/* 34697 */ {(13<<2)|2,{65,86,0}},
/* 34698 */ {(13<<2)|2,{65,86,0}},
/* 34699 */ {(13<<2)|2,{65,86,0}},
/* 34700 */ {(13<<2)|2,{65,86,0}},
/* 34701 */ {(13<<2)|2,{65,86,0}},
/* 34702 */ {(13<<2)|2,{65,86,0}},
/* 34703 */ {(13<<2)|2,{65,86,0}},
/* 34704 */ {(13<<2)|2,{65,87,0}},
/* 34705 */ {(13<<2)|2,{65,87,0}},
/* 34706 */ {(13<<2)|2,{65,87,0}},
/* 34707 */ {(13<<2)|2,{65,87,0}},
/* 34708 */ {(13<<2)|2,{65,87,0}},
/* 34709 */ {(13<<2)|2,{65,87,0}},
/* 34710 */ {(13<<2)|2,{65,87,0}},
/* 34711 */ {(13<<2)|2,{65,87,0}},
/* 34712 */ {(13<<2)|2,{65,89,0}},
/* 34713 */ {(13<<2)|2,{65,89,0}},
/* 34714 */ {(13<<2)|2,{65,89,0}},
/* 34715 */ {(13<<2)|2,{65,89,0}},
/* 34716 */ {(13<<2)|2,{65,89,0}},
/* 34717 */ {(13<<2)|2,{65,89,0}},
/* 34718 */ {(13<<2)|2,{65,89,0}},
/* 34719 */ {(13<<2)|2,{65,89,0}},
/* 34720 */ {(13<<2)|2,{65,106,0}},
/* 34721 */ {(13<<2)|2,{65,106,0}},
/* 34722 */ {(13<<2)|2,{65,106,0}},
/* 34723 */ {(13<<2)|2,{65,106,0}},
/* 34724 */ {(13<<2)|2,{65,106,0}},
/* 34725 */ {(13<<2)|2,{65,106,0}},
/* 34726 */ {(13<<2)|2,{65,106,0}},
/* 34727 */ {(13<<2)|2,{65,106,0}},
/* 34728 */ {(13<<2)|2,{65,107,0}},
/* 34729 */ {(13<<2)|2,{65,107,0}},
/* 34730 */ {(13<<2)|2,{65,107,0}},
/* 34731 */ {(13<<2)|2,{65,107,0}},
/* 34732 */ {(13<<2)|2,{65,107,0}},
/* 34733 */ {(13<<2)|2,{65,107,0}},
/* 34734 */ {(13<<2)|2,{65,107,0}},
/* 34735 */ {(13<<2)|2,{65,107,0}},
/* 34736 */ {(13<<2)|2,{65,113,0}},
/* 34737 */ {(13<<2)|2,{65,113,0}},
/* 34738 */ {(13<<2)|2,{65,113,0}},
/* 34739 */ {(13<<2)|2,{65,113,0}},
/* 34740 */ {(13<<2)|2,{65,113,0}},
/* 34741 */ {(13<<2)|2,{65,113,0}},
/* 34742 */ {(13<<2)|2,{65,113,0}},
/* 34743 */ {(13<<2)|2,{65,113,0}},
/* 34744 */ {(13<<2)|2,{65,118,0}},
/* 34745 */ {(13<<2)|2,{65,118,0}},
/* 34746 */ {(13<<2)|2,{65,118,0}},
/* 34747 */ {(13<<2)|2,{65,118,0}},
/* 34748 */ {(13<<2)|2,{65,118,0}},
/* 34749 */ {(13<<2)|2,{65,118,0}},
/* 34750 */ {(13<<2)|2,{65,118,0}},
/* 34751 */ {(13<<2)|2,{65,118,0}},
/* 34752 */ {(13<<2)|2,{65,119,0}},
/* 34753 */ {(13<<2)|2,{65,119,0}},
/* 34754 */ {(13<<2)|2,{65,119,0}},
/* 34755 */ {(13<<2)|2,{65,119,0}},
/* 34756 */ {(13<<2)|2,{65,119,0}},
/* 34757 */ {(13<<2)|2,{65,119,0}},
/* 34758 */ {(13<<2)|2,{65,119,0}},
/* 34759 */ {(13<<2)|2,{65,119,0}},
/* 34760 */ {(13<<2)|2,{65,120,0}},
/* 34761 */ {(13<<2)|2,{65,120,0}},
/* 34762 */ {(13<<2)|2,{65,120,0}},
/* 34763 */ {(13<<2)|2,{65,120,0}},
/* 34764 */ {(13<<2)|2,{65,120,0}},
/* 34765 */ {(13<<2)|2,{65,120,0}},
/* 34766 */ {(13<<2)|2,{65,120,0}},
/* 34767 */ {(13<<2)|2,{65,120,0}},
/* 34768 */ {(13<<2)|2,{65,121,0}},
/* 34769 */ {(13<<2)|2,{65,121,0}},
/* 34770 */ {(13<<2)|2,{65,121,0}},
/* 34771 */ {(13<<2)|2,{65,121,0}},
/* 34772 */ {(13<<2)|2,{65,121,0}},
/* 34773 */ {(13<<2)|2,{65,121,0}},
/* 34774 */ {(13<<2)|2,{65,121,0}},
/* 34775 */ {(13<<2)|2,{65,121,0}},
/* 34776 */ {(13<<2)|2,{65,122,0}},
/* 34777 */ {(13<<2)|2,{65,122,0}},
/* 34778 */ {(13<<2)|2,{65,122,0}},
/* 34779 */ {(13<<2)|2,{65,122,0}},
/* 34780 */ {(13<<2)|2,{65,122,0}},
/* 34781 */ {(13<<2)|2,{65,122,0}},
/* 34782 */ {(13<<2)|2,{65,122,0}},
/* 34783 */ {(13<<2)|2,{65,122,0}},
/* 34784 */ {(14<<2)|2,{65,38,0}},
/* 34785 */ {(14<<2)|2,{65,38,0}},
/* 34786 */ {(14<<2)|2,{65,38,0}},
/* 34787 */ {(14<<2)|2,{65,38,0}},
/* 34788 */ {(14<<2)|2,{65,42,0}},
/* 34789 */ {(14<<2)|2,{65,42,0}},
/* 34790 */ {(14<<2)|2,{65,42,0}},
/* 34791 */ {(14<<2)|2,{65,42,0}},
/* 34792 */ {(14<<2)|2,{65,44,0}},
/* 34793 */ {(14<<2)|2,{65,44,0}},
/* 34794 */ {(14<<2)|2,{65,44,0}},
/* 34795 */ {(14<<2)|2,{65,44,0}},
/* 34796 */ {(14<<2)|2,{65,59,0}},
/* 34797 */ {(14<<2)|2,{65,59,0}},
/* 34798 */ {(14<<2)|2,{65,59,0}},
/* 34799 */ {(14<<2)|2,{65,59,0}},
/* 34800 */ {(14<<2)|2,{65,88,0}},
/* 34801 */ {(14<<2)|2,{65,88,0}},
/* 34802 */ {(14<<2)|2,{65,88,0}},
/* 34803 */ {(14<<2)|2,{65,88,0}},
/* 34804 */ {(14<<2)|2,{65,90,0}},
/* 34805 */ {(14<<2)|2,{65,90,0}},
/* 34806 */ {(14<<2)|2,{65,90,0}},
/* 34807 */ {(14<<2)|2,{65,90,0}},
/* 34808 */ {(16<<2)|2,{65,33,0}},
/* 34809 */ {(16<<2)|2,{65,34,0}},
/* 34810 */ {(16<<2)|2,{65,40,0}},
/* 34811 */ {(16<<2)|2,{65,41,0}},
/* 34812 */ {(16<<2)|2,{65,63,0}},
/* 34813 */ {(6<<2)|1,{65,0,0}},
/* 34814 */ {(6<<2)|1,{65,0,0}},
/* 34815 */ {(6<<2)|1,{65,0,0}},
/* 34816 */ {(16<<2)|3,{95,48,48}},
/* 34817 */ {(16<<2)|3,{95,48,49}},
/* 34818 */ {(16<<2)|3,{95,48,50}},
/* 34819 */ {(16<<2)|3,{95,48,97}},
/* 34820 */ {(16<<2)|3,{95,48,99}},
/* 34821 */ {(16<<2)|3,{95,48,101}},
/* 34822 */ {(16<<2)|3,{95,48,105}},
/* 34823 */ {(16<<2)|3,{95,48,111}},
/* 34824 */ {(16<<2)|3,{95,48,115}},
/* 34825 */ {(16<<2)|3,{95,48,116}},
/* 34826 */ {(11<<2)|2,{95,48,0}},
/* 34827 */ {(11<<2)|2,{95,48,0}},
/* 34828 */ {(11<<2)|2,{95,48,0}},
/* 34829 */ {(11<<2)|2,{95,48,0}},
/* 34830 */ {(11<<2)|2,{95,48,0}},
/* 34831 */ {(11<<2)|2,{95,48,0}},
/* 34832 */ {(11<<2)|2,{95,48,0}},
/* 34833 */ {(11<<2)|2,{95,48,0}},
/* 34834 */ {(11<<2)|2,{95,48,0}},
/* 34835 */ {(11<<2)|2,{95,48,0}},
/* 34836 */ {(11<<2)|2,{95,48,0}},
/* 34837 */ {(11<<2)|2,{95,48,0}},
/* 34838 */ {(11<<2)|2,{95,48,0}},
/* 34839 */ {(11<<2)|2,{95,48,0}},
/* 34840 */ {(11<<2)|2,{95,48,0}},
/* 34841 */ {(11<<2)|2,{95,48,0}},
/* 34842 */ {(11<<2)|2,{95,48,0}},
/* 34843 */ {(11<<2)|2,{95,48,0}},
/* 34844 */ {(11<<2)|2,{95,48,0}},
/* 34845 */ {(11<<2)|2,{95,48,0}},
/* 34846 */ {(11<<2)|2,{95,48,0}},
/* 34847 */ {(11<<2)|2,{95,48,0}},
/* 34848 */ {(16<<2)|3,{95,49,48}},
/* 34849 */ {(16<<2)|3,{95,49,49}},
/* 34850 */ {(16<<2)|3,{95,49,50}},
/* 34851 */ {(16<<2)|3,{95,49,97}},
/* 34852 */ {(16<<2)|3,{95,49,99}},
/* 34853 */ {(16<<2)|3,{95,49,101}},
/* 34854 */ {(16<<2)|3,{95,49,105}},
/* 34855 */ {(16<<2)|3,{95,49,111}},
/* 34856 */ {(16<<2)|3,{95,49,115}},
/* 34857 */ {(16<<2)|3,{95,49,116}},
/* 34858 */ {(11<<2)|2,{95,49,0}},
/* 34859 */ {(11<<2)|2,{95,49,0}},
/* 34860 */ {(11<<2)|2,{95,49,0}},
/* 34861 */ {(11<<2)|2,{95,49,0}},
/* 34862 */ {(11<<2)|2,{95,49,0}},
/* 34863 */ {(11<<2)|2,{95,49,0}},
/* 34864 */ {(11<<2)|2,{95,49,0}},
/* 34865 */ {(11<<2)|2,{95,49,0}},
/* 34866 */ {(11<<2)|2,{95,49,0}},
/* 34867 */ {(11<<2)|2,{95,49,0}},
/* 34868 */ {(11<<2)|2,{95,49,0}},
/* 34869 */ {(11<<2)|2,{95,49,0}},
/* 34870 */ {(11<<2)|2,{95,49,0}},
/* 34871 */ {(11<<2)|2,{95,49,0}},
/* 34872 */ {(11<<2)|2,{95,49,0}},
/* 34873 */ {(11<<2)|2,{95,49,0}},
/* 34874 */ {(11<<2)|2,{95,49,0}},
/* 34875 */ {(11<<2)|2,{95,49,0}},
/* 34876 */ {(11<<2)|2,{95,49,0}},
/* 34877 */ {(11<<2)|2,{95,49,0}},
/* 34878 */ {(11<<2)|2,{95,49,0}},
/* 34879 */ {(11<<2)|2,{95,49,0}},
/* 34880 */ {(16<<2)|3,{95,50,48}},
/* 34881 */ {(16<<2)|3,{95,50,49}},
/* 34882 */ {(16<<2)|3,{95,50,50}},
/* 34883 */ {(16<<2)|3,{95,50,97}},
/* 34884 */ {(16<<2)|3,{95,50,99}},
/* 34885 */ {(16<<2)|3,{95,50,101}},
/* 34886 */ {(16<<2)|3,{95,50,105}},
/* 34887 */ {(16<<2)|3,{95,50,111}},
/* 34888 */ {(16<<2)|3,{95,50,115}},
/* 34889 */ {(16<<2)|3,{95,50,116}},
/* 34890 */ {(11<<2)|2,{95,50,0}},
/* 34891 */ {(11<<2)|2,{95,50,0}},
/* 34892 */ {(11<<2)|2,{95,50,0}},
/* 34893 */ {(11<<2)|2,{95,50,0}},
/* 34894 */ {(11<<2)|2,{95,50,0}},
/* 34895 */ {(11<<2)|2,{95,50,0}},
/* 34896 */ {(11<<2)|2,{95,50,0}},
/* 34897 */ {(11<<2)|2,{95,50,0}},
/* 34898 */ {(11<<2)|2,{95,50,0}},
/* 34899 */ {(11<<2)|2,{95,50,0}},
/* 34900 */ {(11<<2)|2,{95,50,0}},
/* 34901 */ {(11<<2)|2,{95,50,0}},
/* 34902 */ {(11<<2)|2,{95,50,0}},
/* 34903 */ {(11<<2)|2,{95,50,0}},
/* 34904 */ {(11<<2)|2,{95,50,0}},
/* 34905 */ {(11<<2)|2,{95,50,0}},
/* 34906 */ {(11<<2)|2,{95,50,0}},
/* 34907 */ {(11<<2)|2,{95,50,0}},
/* 34908 */ {(11<<2)|2,{95,50,0}},
/* 34909 */ {(11<<2)|2,{95,50,0}},
/* 34910 */ {(11<<2)|2,{95,50,0}},
/* 34911 */ {(11<<2)|2,{95,50,0}},
/* 34912 */ {(16<<2)|3,{95,97,48}},
/* 34913 */ {(16<<2)|3,{95,97,49}},
/* 34914 */ {(16<<2)|3,{95,97,50}},
/* 34915 */ {(16<<2)|3,{95,97,97}},
/* 34916 */ {(16<<2)|3,{95,97,99}},
/* 34917 */ {(16<<2)|3,{95,97,101}},
/* 34918 */ {(16<<2)|3,{95,97,105}},
/* 34919 */ {(16<<2)|3,{95,97,111}},
/* 34920 */ {(16<<2)|3,{95,97,115}},
/* 34921 */ {(16<<2)|3,{95,97,116}},
/* 34922 */ {(11<<2)|2,{95,97,0}},
/* 34923 */ {(11<<2)|2,{95,97,0}},
/* 34924 */ {(11<<2)|2,{95,97,0}},
/* 34925 */ {(11<<2)|2,{95,97,0}},
/* 34926 */ {(11<<2)|2,{95,97,0}},
/* 34927 */ {(11<<2)|2,{95,97,0}},
/* 34928 */ {(11<<2)|2,{95,97,0}},
/* 34929 */ {(11<<2)|2,{95,97,0}},
/* 34930 */ {(11<<2)|2,{95,97,0}},
/* 34931 */ {(11<<2)|2,{95,97,0}},
/* 34932 */ {(11<<2)|2,{95,97,0}},
/* 34933 */ {(11<<2)|2,{95,97,0}},
/* 34934 */ {(11<<2)|2,{95,97,0}},
/* 34935 */ {(11<<2)|2,{95,97,0}},
/* 34936 */ {(11<<2)|2,{95,97,0}},
/* 34937 */ {(11<<2)|2,{95,97,0}},
/* 34938 */ {(11<<2)|2,{95,97,0}},
/* 34939 */ {(11<<2)|2,{95,97,0}},
/* 34940 */ {(11<<2)|2,{95,97,0}},
/* 34941 */ {(11<<2)|2,{95,97,0}},
/* 34942 */ {(11<<2)|2,{95,97,0}},
/* 34943 */ {(11<<2)|2,{95,97,0}},
/* 34944 */ {(16<<2)|3,{95,99,48}},
/* 34945 */ {(16<<2)|3,{95,99,49}},
/* 34946 */ {(16<<2)|3,{95,99,50}},
/* 34947 */ {(16<<2)|3,{95,99,97}},
/* 34948 */ {(16<<2)|3,{95,99,99}},
/* 34949 */ {(16<<2)|3,{95,99,101}},
/* 34950 */ {(16<<2)|3,{95,99,105}},
/* 34951 */ {(16<<2)|3,{95,99,111}},
/* 34952 */ {(16<<2)|3,{95,99,115}},
/* 34953 */ {(16<<2)|3,{95,99,116}},
/* 34954 */ {(11<<2)|2,{95,99,0}},
/* 34955 */ {(11<<2)|2,{95,99,0}},
/* 34956 */ {(11<<2)|2,{95,99,0}},
/* 34957 */ {(11<<2)|2,{95,99,0}},
/* 34958 */ {(11<<2)|2,{95,99,0}},
/* 34959 */ {(11<<2)|2,{95,99,0}},
/* 34960 */ {(11<<2)|2,{95,99,0}},
/* 34961 */ {(11<<2)|2,{95,99,0}},
/* 34962 */ {(11<<2)|2,{95,99,0}},
/* 34963 */ {(11<<2)|2,{95,99,0}},
/* 34964 */ {(11<<2)|2,{95,99,0}},
/* 34965 */ {(11<<2)|2,{95,99,0}},
/* 34966 */ {(11<<2)|2,{95,99,0}},
/* 34967 */ {(11<<2)|2,{95,99,0}},
/* 34968 */ {(11<<2)|2,{95,99,0}},
/* 34969 */ {(11<<2)|2,{95,99,0}},
/* 34970 */ {(11<<2)|2,{95,99,0}},
/* 34971 */ {(11<<2)|2,{95,99,0}},
/* 34972 */ {(11<<2)|2,{95,99,0}},
/* 34973 */ {(11<<2)|2,{95,99,0}},
/* 34974 */ {(11<<2)|2,{95,99,0}},
/* 34975 */ {(11<<2)|2,{95,99,0}},
/* 34976 */ {(16<<2)|3,{95,101,48}},
/* 34977 */ {(16<<2)|3,{95,101,49}},
/* 34978 */ {(16<<2)|3,{95,101,50}},
/* 34979 */ {(16<<2)|3,{95,101,97}},
/* 34980 */ {(16<<2)|3,{95,101,99}},
/* 34981 */ {(16<<2)|3,{95,101,101}},
/* 34982 */ {(16<<2)|3,{95,101,105}},
/* 34983 */ {(16<<2)|3,{95,101,111}},
/* 34984 */ {(16<<2)|3,{95,101,115}},
/* 34985 */ {(16<<2)|3,{95,101,116}},
/* 34986 */ {(11<<2)|2,{95,101,0}},
/* 34987 */ {(11<<2)|2,{95,101,0}},
/* 34988 */ {(11<<2)|2,{95,101,0}},
/* 34989 */ {(11<<2)|2,{95,101,0}},
/* 34990 */ {(11<<2)|2,{95,101,0}},
/* 34991 */ {(11<<2)|2,{95,101,0}},
/* 34992 */ {(11<<2)|2,{95,101,0}},
/* 34993 */ {(11<<2)|2,{95,101,0}},
/* 34994 */ {(11<<2)|2,{95,101,0}},
/* 34995 */ {(11<<2)|2,{95,101,0}},
/* 34996 */ {(11<<2)|2,{95,101,0}},
/* 34997 */ {(11<<2)|2,{95,101,0}},
/* 34998 */ {(11<<2)|2,{95,101,0}},
/* 34999 */ {(11<<2)|2,{95,101,0}},
/* 35000 */ {(11<<2)|2,{95,101,0}},
/* 35001 */ {(11<<2)|2,{95,101,0}},
/* 35002 */ {(11<<2)|2,{95,101,0}},
/* 35003 */ {(11<<2)|2,{95,101,0}},
/* 35004 */ {(11<<2)|2,{95,101,0}},
/* 35005 */ {(11<<2)|2,{95,101,0}},
/* 35006 */ {(11<<2)|2,{95,101,0}},
/* 35007 */ {(11<<2)|2,{95,101,0}},
/* 35008 */ {(16<<2)|3,{95,105,48}},
/* 35009 */ {(16<<2)|3,{95,105,49}},
/* 35010 */ {(16<<2)|3,{95,105,50}},
/* 35011 */ {(16<<2)|3,{95,105,97}},
/* 35012 */ {(16<<2)|3,{95,105,99}},
/* 35013 */ {(16<<2)|3,{95,105,101}},
/* 35014 */ {(16<<2)|3,{95,105,105}},
/* 35015 */ {(16<<2)|3,{95,105,111}},
/* 35016 */ {(16<<2)|3,{95,105,115}},
/* 35017 */ {(16<<2)|3,{95,105,116}},
/* 35018 */ {(11<<2)|2,{95,105,0}},
/* 35019 */ {(11<<2)|2,{95,105,0}},
/* 35020 */ {(11<<2)|2,{95,105,0}},
/* 35021 */ {(11<<2)|2,{95,105,0}},
/* 35022 */ {(11<<2)|2,{95,105,0}},
/* 35023 */ {(11<<2)|2,{95,105,0}},
/* 35024 */ {(11<<2)|2,{95,105,0}},
/* 35025 */ {(11<<2)|2,{95,105,0}},
/* 35026 */ {(11<<2)|2,{95,105,0}},
/* 35027 */ {(11<<2)|2,{95,105,0}},
/* 35028 */ {(11<<2)|2,{95,105,0}},
/* 35029 */ {(11<<2)|2,{95,105,0}},
/* 35030 */ {(11<<2)|2,{95,105,0}},
/* 35031 */ {(11<<2)|2,{95,105,0}},
/* 35032 */ {(11<<2)|2,{95,105,0}},
/* 35033 */ {(11<<2)|2,{95,105,0}},
/* 35034 */ {(11<<2)|2,{95,105,0}},
/* 35035 */ {(11<<2)|2,{95,105,0}},
/* 35036 */ {(11<<2)|2,{95,105,0}},
/* 35037 */ {(11<<2)|2,{95,105,0}},
/* 35038 */ {(11<<2)|2,{95,105,0}},
/* 35039 */ {(11<<2)|2,{95,105,0}},
/* 35040 */ {(16<<2)|3,{95,111,48}},
/* 35041 */ {(16<<2)|3,{95,111,49}},
/* 35042 */ {(16<<2)|3,{95,111,50}},
/* 35043 */ {(16<<2)|3,{95,111,97}},
/* 35044 */ {(16<<2)|3,{95,111,99}},
/* 35045 */ {(16<<2)|3,{95,111,101}},
/* 35046 */ {(16<<2)|3,{95,111,105}},
/* 35047 */ {(16<<2)|3,{95,111,111}},
/* 35048 */ {(16<<2)|3,{95,111,115}},
/* 35049 */ {(16<<2)|3,{95,111,116}},
/* 35050 */ {(11<<2)|2,{95,111,0}},
/* 35051 */ {(11<<2)|2,{95,111,0}},
/* 35052 */ {(11<<2)|2,{95,111,0}},
/* 35053 */ {(11<<2)|2,{95,111,0}},
/* 35054 */ {(11<<2)|2,{95,111,0}},
/* 35055 */ {(11<<2)|2,{95,111,0}},
/* 35056 */ {(11<<2)|2,{95,111,0}},
/* 35057 */ {(11<<2)|2,{95,111,0}},
/* 35058 */ {(11<<2)|2,{95,111,0}},
/* 35059 */ {(11<<2)|2,{95,111,0}},
/* 35060 */ {(11<<2)|2,{95,111,0}},
/* 35061 */ {(11<<2)|2,{95,111,0}},
/* 35062 */ {(11<<2)|2,{95,111,0}},
/* 35063 */ {(11<<2)|2,{95,111,0}},
/* 35064 */ {(11<<2)|2,{95,111,0}},
/* 35065 */ {(11<<2)|2,{95,111,0}},
/* 35066 */ {(11<<2)|2,{95,111,0}},
/* 35067 */ {(11<<2)|2,{95,111,0}},
/* 35068 */ {(11<<2)|2,{95,111,0}},
/* 35069 */ {(11<<2)|2,{95,111,0}},
/* 35070 */ {(11<<2)|2,{95,111,0}},
/* 35071 */ {(11<<2)|2,{95,111,0}},
/* 35072 */ {(16<<2)|3,{95,115,48}},
/* 35073 */ {(16<<2)|3,{95,115,49}},
/* 35074 */ {(16<<2)|3,{95,115,50}},
/* 35075 */ {(16<<2)|3,{95,115,97}},
/* 35076 */ {(16<<2)|3,{95,115,99}},
/* 35077 */ {(16<<2)|3,{95,115,101}},
/* 35078 */ {(16<<2)|3,{95,115,105}},
/* 35079 */ {(16<<2)|3,{95,115,111}},
/* 35080 */ {(16<<2)|3,{95,115,115}},
/* 35081 */ {(16<<2)|3,{95,115,116}},
/* 35082 */ {(11<<2)|2,{95,115,0}},
/* 35083 */ {(11<<2)|2,{95,115,0}},
/* 35084 */ {(11<<2)|2,{95,115,0}},
/* 35085 */ {(11<<2)|2,{95,115,0}},
/* 35086 */ {(11<<2)|2,{95,115,0}},
/* 35087 */ {(11<<2)|2,{95,115,0}},
/* 35088 */ {(11<<2)|2,{95,115,0}},
/* 35089 */ {(11<<2)|2,{95,115,0}},
/* 35090 */ {(11<<2)|2,{95,115,0}},
/* 35091 */ {(11<<2)|2,{95,115,0}},
/* 35092 */ {(11<<2)|2,{95,115,0}},
/* 35093 */ {(11<<2)|2,{95,115,0}},
/* 35094 */ {(11<<2)|2,{95,115,0}},
/* 35095 */ {(11<<2)|2,{95,115,0}},
/* 35096 */ {(11<<2)|2,{95,115,0}},
/* 35097 */ {(11<<2)|2,{95,115,0}},
/* 35098 */ {(11<<2)|2,{95,115,0}},
/* 35099 */ {(11<<2)|2,{95,115,0}},
/* 35100 */ {(11<<2)|2,{95,115,0}},
/* 35101 */ {(11<<2)|2,{95,115,0}},
/* 35102 */ {(11<<2)|2,{95,115,0}},
/* 35103 */ {(11<<2)|2,{95,115,0}},
/* 35104 */ {(16<<2)|3,{95,116,48}},
/* 35105 */ {(16<<2)|3,{95,116,49}},
/* 35106 */ {(16<<2)|3,{95,116,50}},
/* 35107 */ {(16<<2)|3,{95,116,97}},
/* 35108 */ {(16<<2)|3,{95,116,99}},
/* 35109 */ {(16<<2)|3,{95,116,101}},
/* 35110 */ {(16<<2)|3,{95,116,105}},
/* 35111 */ {(16<<2)|3,{95,116,111}},
/* 35112 */ {(16<<2)|3,{95,116,115}},
/* 35113 */ {(16<<2)|3,{95,116,116}},
/* 35114 */ {(11<<2)|2,{95,116,0}},
/* 35115 */ {(11<<2)|2,{95,116,0}},
/* 35116 */ {(11<<2)|2,{95,116,0}},
/* 35117 */ {(11<<2)|2,{95,116,0}},
/* 35118 */ {(11<<2)|2,{95,116,0}},
/* 35119 */ {(11<<2)|2,{95,116,0}},
/* 35120 */ {(11<<2)|2,{95,116,0}},
/* 35121 */ {(11<<2)|2,{95,116,0}},
/* 35122 */ {(11<<2)|2,{95,116,0}},
/* 35123 */ {(11<<2)|2,{95,116,0}},
/* 35124 */ {(11<<2)|2,{95,116,0}},
/* 35125 */ {(11<<2)|2,{95,116,0}},
/* 35126 */ {(11<<2)|2,{95,116,0}},
/* 35127 */ {(11<<2)|2,{95,116,0}},
/* 35128 */ {(11<<2)|2,{95,116,0}},
/* 35129 */ {(11<<2)|2,{95,116,0}},
/* 35130 */ {(11<<2)|2,{95,116,0}},
/* 35131 */ {(11<<2)|2,{95,116,0}},
/* 35132 */ {(11<<2)|2,{95,116,0}},
/* 35133 */ {(11<<2)|2,{95,116,0}},
/* 35134 */ {(11<<2)|2,{95,116,0}},
/* 35135 */ {(11<<2)|2,{95,116,0}},
/* 35136 */ {(12<<2)|2,{95,32,0}},
/* 35137 */ {(12<<2)|2,{95,32,0}},
/* 35138 */ {(12<<2)|2,{95,32,0}},
/* 35139 */ {(12<<2)|2,{95,32,0}},
/* 35140 */ {(12<<2)|2,{95,32,0}},
/* 35141 */ {(12<<2)|2,{95,32,0}},
/* 35142 */ {(12<<2)|2,{95,32,0}},
/* 35143 */ {(12<<2)|2,{95,32,0}},
/* 35144 */ {(12<<2)|2,{95,32,0}},
/* 35145 */ {(12<<2)|2,{95,32,0}},
/* 35146 */ {(12<<2)|2,{95,32,0}},
/* 35147 */ {(12<<2)|2,{95,32,0}},
/* 35148 */ {(12<<2)|2,{95,32,0}},
/* 35149 */ {(12<<2)|2,{95,32,0}},
/* 35150 */ {(12<<2)|2,{95,32,0}},
/* 35151 */ {(12<<2)|2,{95,32,0}},
/* 35152 */ {(12<<2)|2,{95,37,0}},
/* 35153 */ {(12<<2)|2,{95,37,0}},
/* 35154 */ {(12<<2)|2,{95,37,0}},
/* 35155 */ {(12<<2)|2,{95,37,0}},
/* 35156 */ {(12<<2)|2,{95,37,0}},
/* 35157 */ {(12<<2)|2,{95,37,0}},
/* 35158 */ {(12<<2)|2,{95,37,0}},
/* 35159 */ {(12<<2)|2,{95,37,0}},
/* 35160 */ {(12<<2)|2,{95,37,0}},
/* 35161 */ {(12<<2)|2,{95,37,0}},
/* 35162 */ {(12<<2)|2,{95,37,0}},
/* 35163 */ {(12<<2)|2,{95,37,0}},
/* 35164 */ {(12<<2)|2,{95,37,0}},
/* 35165 */ {(12<<2)|2,{95,37,0}},
/* 35166 */ {(12<<2)|2,{95,37,0}},
/* 35167 */ {(12<<2)|2,{95,37,0}},
/* 35168 */ {(12<<2)|2,{95,45,0}},
/* 35169 */ {(12<<2)|2,{95,45,0}},
/* 35170 */ {(12<<2)|2,{95,45,0}},
/* 35171 */ {(12<<2)|2,{95,45,0}},
/* 35172 */ {(12<<2)|2,{95,45,0}},
/* 35173 */ {(12<<2)|2,{95,45,0}},
/* 35174 */ {(12<<2)|2,{95,45,0}},
/* 35175 */ {(12<<2)|2,{95,45,0}},
/* 35176 */ {(12<<2)|2,{95,45,0}},
/* 35177 */ {(12<<2)|2,{95,45,0}},
/* 35178 */ {(12<<2)|2,{95,45,0}},
/* 35179 */ {(12<<2)|2,{95,45,0}},
/* 35180 */ {(12<<2)|2,{95,45,0}},
/* 35181 */ {(12<<2)|2,{95,45,0}},
/* 35182 */ {(12<<2)|2,{95,45,0}},
/* 35183 */ {(12<<2)|2,{95,45,0}},
/* 35184 */ {(12<<2)|2,{95,46,0}},
/* 35185 */ {(12<<2)|2,{95,46,0}},
/* 35186 */ {(12<<2)|2,{95,46,0}},
/* 35187 */ {(12<<2)|2,{95,46,0}},
/* 35188 */ {(12<<2)|2,{95,46,0}},
/* 35189 */ {(12<<2)|2,{95,46,0}},
/* 35190 */ {(12<<2)|2,{95,46,0}},
/* 35191 */ {(12<<2)|2,{95,46,0}},
/* 35192 */ {(12<<2)|2,{95,46,0}},
/* 35193 */ {(12<<2)|2,{95,46,0}},
/* 35194 */ {(12<<2)|2,{95,46,0}},
/* 35195 */ {(12<<2)|2,{95,46,0}},
/* 35196 */ {(12<<2)|2,{95,46,0}},
/* 35197 */ {(12<<2)|2,{95,46,0}},
/* 35198 */ {(12<<2)|2,{95,46,0}},
/* 35199 */ {(12<<2)|2,{95,46,0}},
/* 35200 */ {(12<<2)|2,{95,47,0}},
/* 35201 */ {(12<<2)|2,{95,47,0}},
/* 35202 */ {(12<<2)|2,{95,47,0}},
/* 35203 */ {(12<<2)|2,{95,47,0}},
/* 35204 */ {(12<<2)|2,{95,47,0}},
/* 35205 */ {(12<<2)|2,{95,47,0}},
/* 35206 */ {(12<<2)|2,{95,47,0}},
/* 35207 */ {(12<<2)|2,{95,47,0}},
/* 35208 */ {(12<<2)|2,{95,47,0}},
/* 35209 */ {(12<<2)|2,{95,47,0}},
/* 35210 */ {(12<<2)|2,{95,47,0}},
/* 35211 */ {(12<<2)|2,{95,47,0}},
/* 35212 */ {(12<<2)|2,{95,47,0}},
/* 35213 */ {(12<<2)|2,{95,47,0}},
/* 35214 */ {(12<<2)|2,{95,47,0}},
/* 35215 */ {(12<<2)|2,{95,47,0}},
/* 35216 */ {(12<<2)|2,{95,51,0}},
/* 35217 */ {(12<<2)|2,{95,51,0}},
/* 35218 */ {(12<<2)|2,{95,51,0}},
/* 35219 */ {(12<<2)|2,{95,51,0}},
/* 35220 */ {(12<<2)|2,{95,51,0}},
/* 35221 */ {(12<<2)|2,{95,51,0}},
/* 35222 */ {(12<<2)|2,{95,51,0}},
/* 35223 */ {(12<<2)|2,{95,51,0}},
/* 35224 */ {(12<<2)|2,{95,51,0}},
/* 35225 */ {(12<<2)|2,{95,51,0}},
/* 35226 */ {(12<<2)|2,{95,51,0}},
/* 35227 */ {(12<<2)|2,{95,51,0}},
/* 35228 */ {(12<<2)|2,{95,51,0}},
/* 35229 */ {(12<<2)|2,{95,51,0}},
/* 35230 */ {(12<<2)|2,{95,51,0}},
/* 35231 */ {(12<<2)|2,{95,51,0}},
/* 35232 */ {(12<<2)|2,{95,52,0}},
/* 35233 */ {(12<<2)|2,{95,52,0}},
/* 35234 */ {(12<<2)|2,{95,52,0}},
/* 35235 */ {(12<<2)|2,{95,52,0}},
/* 35236 */ {(12<<2)|2,{95,52,0}},
/* 35237 */ {(12<<2)|2,{95,52,0}},
/* 35238 */ {(12<<2)|2,{95,52,0}},
/* 35239 */ {(12<<2)|2,{95,52,0}},
/* 35240 */ {(12<<2)|2,{95,52,0}},
/* 35241 */ {(12<<2)|2,{95,52,0}},
/* 35242 */ {(12<<2)|2,{95,52,0}},
/* 35243 */ {(12<<2)|2,{95,52,0}},
/* 35244 */ {(12<<2)|2,{95,52,0}},
/* 35245 */ {(12<<2)|2,{95,52,0}},
/* 35246 */ {(12<<2)|2,{95,52,0}},
/* 35247 */ {(12<<2)|2,{95,52,0}},
/* 35248 */ {(12<<2)|2,{95,53,0}},
/* 35249 */ {(12<<2)|2,{95,53,0}},
/* 35250 */ {(12<<2)|2,{95,53,0}},
/* 35251 */ {(12<<2)|2,{95,53,0}},
/* 35252 */ {(12<<2)|2,{95,53,0}},
/* 35253 */ {(12<<2)|2,{95,53,0}},
/* 35254 */ {(12<<2)|2,{95,53,0}},
/* 35255 */ {(12<<2)|2,{95,53,0}},
/* 35256 */ {(12<<2)|2,{95,53,0}},
/* 35257 */ {(12<<2)|2,{95,53,0}},
/* 35258 */ {(12<<2)|2,{95,53,0}},
/* 35259 */ {(12<<2)|2,{95,53,0}},
/* 35260 */ {(12<<2)|2,{95,53,0}},
/* 35261 */ {(12<<2)|2,{95,53,0}},
/* 35262 */ {(12<<2)|2,{95,53,0}},
/* 35263 */ {(12<<2)|2,{95,53,0}},
/* 35264 */ {(12<<2)|2,{95,54,0}},
/* 35265 */ {(12<<2)|2,{95,54,0}},
/* 35266 */ {(12<<2)|2,{95,54,0}},
/* 35267 */ {(12<<2)|2,{95,54,0}},
/* 35268 */ {(12<<2)|2,{95,54,0}},
/* 35269 */ {(12<<2)|2,{95,54,0}},
/* 35270 */ {(12<<2)|2,{95,54,0}},
/* 35271 */ {(12<<2)|2,{95,54,0}},
/* 35272 */ {(12<<2)|2,{95,54,0}},
/* 35273 */ {(12<<2)|2,{95,54,0}},
/* 35274 */ {(12<<2)|2,{95,54,0}},
/* 35275 */ {(12<<2)|2,{95,54,0}},
/* 35276 */ {(12<<2)|2,{95,54,0}},
/* 35277 */ {(12<<2)|2,{95,54,0}},
/* 35278 */ {(12<<2)|2,{95,54,0}},
/* 35279 */ {(12<<2)|2,{95,54,0}},
/* 35280 */ {(12<<2)|2,{95,55,0}},
/* 35281 */ {(12<<2)|2,{95,55,0}},
/* 35282 */ {(12<<2)|2,{95,55,0}},
/* 35283 */ {(12<<2)|2,{95,55,0}},
/* 35284 */ {(12<<2)|2,{95,55,0}},
/* 35285 */ {(12<<2)|2,{95,55,0}},
/* 35286 */ {(12<<2)|2,{95,55,0}},
/* 35287 */ {(12<<2)|2,{95,55,0}},
/* 35288 */ {(12<<2)|2,{95,55,0}},
/* 35289 */ {(12<<2)|2,{95,55,0}},
/* 35290 */ {(12<<2)|2,{95,55,0}},
/* 35291 */ {(12<<2)|2,{95,55,0}},
/* 35292 */ {(12<<2)|2,{95,55,0}},
/* 35293 */ {(12<<2)|2,{95,55,0}},
/* 35294 */ {(12<<2)|2,{95,55,0}},
/* 35295 */ {(12<<2)|2,{95,55,0}},
/* 35296 */ {(12<<2)|2,{95,56,0}},
/* 35297 */ {(12<<2)|2,{95,56,0}},
/* 35298 */ {(12<<2)|2,{95,56,0}},
/* 35299 */ {(12<<2)|2,{95,56,0}},
/* 35300 */ {(12<<2)|2,{95,56,0}},
/* 35301 */ {(12<<2)|2,{95,56,0}},
/* 35302 */ {(12<<2)|2,{95,56,0}},
/* 35303 */ {(12<<2)|2,{95,56,0}},
/* 35304 */ {(12<<2)|2,{95,56,0}},
/* 35305 */ {(12<<2)|2,{95,56,0}},
/* 35306 */ {(12<<2)|2,{95,56,0}},
/* 35307 */ {(12<<2)|2,{95,56,0}},
/* 35308 */ {(12<<2)|2,{95,56,0}},
/* 35309 */ {(12<<2)|2,{95,56,0}},
/* 35310 */ {(12<<2)|2,{95,56,0}},
/* 35311 */ {(12<<2)|2,{95,56,0}},
/* 35312 */ {(12<<2)|2,{95,57,0}},
/* 35313 */ {(12<<2)|2,{95,57,0}},
/* 35314 */ {(12<<2)|2,{95,57,0}},
/* 35315 */ {(12<<2)|2,{95,57,0}},
/* 35316 */ {(12<<2)|2,{95,57,0}},
/* 35317 */ {(12<<2)|2,{95,57,0}},
/* 35318 */ {(12<<2)|2,{95,57,0}},
/* 35319 */ {(12<<2)|2,{95,57,0}},
/* 35320 */ {(12<<2)|2,{95,57,0}},
/* 35321 */ {(12<<2)|2,{95,57,0}},
/* 35322 */ {(12<<2)|2,{95,57,0}},
/* 35323 */ {(12<<2)|2,{95,57,0}},
/* 35324 */ {(12<<2)|2,{95,57,0}},
/* 35325 */ {(12<<2)|2,{95,57,0}},
/* 35326 */ {(12<<2)|2,{95,57,0}},
/* 35327 */ {(12<<2)|2,{95,57,0}},
/* 35328 */ {(12<<2)|2,{95,61,0}},
/* 35329 */ {(12<<2)|2,{95,61,0}},
/* 35330 */ {(12<<2)|2,{95,61,0}},
/* 35331 */ {(12<<2)|2,{95,61,0}},
/* 35332 */ {(12<<2)|2,{95,61,0}},
/* 35333 */ {(12<<2)|2,{95,61,0}},
/* 35334 */ {(12<<2)|2,{95,61,0}},
/* 35335 */ {(12<<2)|2,{95,61,0}},
/* 35336 */ {(12<<2)|2,{95,61,0}},
/* 35337 */ {(12<<2)|2,{95,61,0}},
/* 35338 */ {(12<<2)|2,{95,61,0}},
/* 35339 */ {(12<<2)|2,{95,61,0}},
/* 35340 */ {(12<<2)|2,{95,61,0}},
/* 35341 */ {(12<<2)|2,{95,61,0}},
/* 35342 */ {(12<<2)|2,{95,61,0}},
/* 35343 */ {(12<<2)|2,{95,61,0}},
/* 35344 */ {(12<<2)|2,{95,65,0}},
/* 35345 */ {(12<<2)|2,{95,65,0}},
/* 35346 */ {(12<<2)|2,{95,65,0}},
/* 35347 */ {(12<<2)|2,{95,65,0}},
/* 35348 */ {(12<<2)|2,{95,65,0}},
/* 35349 */ {(12<<2)|2,{95,65,0}},
/* 35350 */ {(12<<2)|2,{95,65,0}},
/* 35351 */ {(12<<2)|2,{95,65,0}},
/* 35352 */ {(12<<2)|2,{95,65,0}},
/* 35353 */ {(12<<2)|2,{95,65,0}},
/* 35354 */ {(12<<2)|2,{95,65,0}},
/* 35355 */ {(12<<2)|2,{95,65,0}},
/* 35356 */ {(12<<2)|2,{95,65,0}},
/* 35357 */ {(12<<2)|2,{95,65,0}},
/* 35358 */ {(12<<2)|2,{95,65,0}},
/* 35359 */ {(12<<2)|2,{95,65,0}},
/* 35360 */ {(12<<2)|2,{95,95,0}},
/* 35361 */ {(12<<2)|2,{95,95,0}},
/* 35362 */ {(12<<2)|2,{95,95,0}},
/* 35363 */ {(12<<2)|2,{95,95,0}},
/* 35364 */ {(12<<2)|2,{95,95,0}},
/* 35365 */ {(12<<2)|2,{95,95,0}},
/* 35366 */ {(12<<2)|2,{95,95,0}},
/* 35367 */ {(12<<2)|2,{95,95,0}},
/* 35368 */ {(12<<2)|2,{95,95,0}},
/* 35369 */ {(12<<2)|2,{95,95,0}},
/* 35370 */ {(12<<2)|2,{95,95,0}},
/* 35371 */ {(12<<2)|2,{95,95,0}},
/* 35372 */ {(12<<2)|2,{95,95,0}},
/* 35373 */ {(12<<2)|2,{95,95,0}},
/* 35374 */ {(12<<2)|2,{95,95,0}},
/* 35375 */ {(12<<2)|2,{95,95,0}},
/* 35376 */ {(12<<2)|2,{95,98,0}},
/* 35377 */ {(12<<2)|2,{95,98,0}},
/* 35378 */ {(12<<2)|2,{95,98,0}},
/* 35379 */ {(12<<2)|2,{95,98,0}},
/* 35380 */ {(12<<2)|2,{95,98,0}},
/* 35381 */ {(12<<2)|2,{95,98,0}},
/* 35382 */ {(12<<2)|2,{95,98,0}},
/* 35383 */ {(12<<2)|2,{95,98,0}},
/* 35384 */ {(12<<2)|2,{95,98,0}},
/* 35385 */ {(12<<2)|2,{95,98,0}},
/* 35386 */ {(12<<2)|2,{95,98,0}},
/* 35387 */ {(12<<2)|2,{95,98,0}},
/* 35388 */ {(12<<2)|2,{95,98,0}},
/* 35389 */ {(12<<2)|2,{95,98,0}},
/* 35390 */ {(12<<2)|2,{95,98,0}},
/* 35391 */ {(12<<2)|2,{95,98,0}},
/* 35392 */ {(12<<2)|2,{95,100,0}},
/* 35393 */ {(12<<2)|2,{95,100,0}},
/* 35394 */ {(12<<2)|2,{95,100,0}},
/* 35395 */ {(12<<2)|2,{95,100,0}},
/* 35396 */ {(12<<2)|2,{95,100,0}},
/* 35397 */ {(12<<2)|2,{95,100,0}},
/* 35398 */ {(12<<2)|2,{95,100,0}},
/* 35399 */ {(12<<2)|2,{95,100,0}},
/* 35400 */ {(12<<2)|2,{95,100,0}},
/* 35401 */ {(12<<2)|2,{95,100,0}},
/* 35402 */ {(12<<2)|2,{95,100,0}},
/* 35403 */ {(12<<2)|2,{95,100,0}},
/* 35404 */ {(12<<2)|2,{95,100,0}},
/* 35405 */ {(12<<2)|2,{95,100,0}},
/* 35406 */ {(12<<2)|2,{95,100,0}},
/* 35407 */ {(12<<2)|2,{95,100,0}},
/* 35408 */ {(12<<2)|2,{95,102,0}},
/* 35409 */ {(12<<2)|2,{95,102,0}},
/* 35410 */ {(12<<2)|2,{95,102,0}},
/* 35411 */ {(12<<2)|2,{95,102,0}},
/* 35412 */ {(12<<2)|2,{95,102,0}},
/* 35413 */ {(12<<2)|2,{95,102,0}},
/* 35414 */ {(12<<2)|2,{95,102,0}},
/* 35415 */ {(12<<2)|2,{95,102,0}},
/* 35416 */ {(12<<2)|2,{95,102,0}},
/* 35417 */ {(12<<2)|2,{95,102,0}},
/* 35418 */ {(12<<2)|2,{95,102,0}},
/* 35419 */ {(12<<2)|2,{95,102,0}},
/* 35420 */ {(12<<2)|2,{95,102,0}},
/* 35421 */ {(12<<2)|2,{95,102,0}},
/* 35422 */ {(12<<2)|2,{95,102,0}},
/* 35423 */ {(12<<2)|2,{95,102,0}},
/* 35424 */ {(12<<2)|2,{95,103,0}},
/* 35425 */ {(12<<2)|2,{95,103,0}},
/* 35426 */ {(12<<2)|2,{95,103,0}},
/* 35427 */ {(12<<2)|2,{95,103,0}},
/* 35428 */ {(12<<2)|2,{95,103,0}},
/* 35429 */ {(12<<2)|2,{95,103,0}},
/* 35430 */ {(12<<2)|2,{95,103,0}},
/* 35431 */ {(12<<2)|2,{95,103,0}},
/* 35432 */ {(12<<2)|2,{95,103,0}},
/* 35433 */ {(12<<2)|2,{95,103,0}},
/* 35434 */ {(12<<2)|2,{95,103,0}},
/* 35435 */ {(12<<2)|2,{95,103,0}},
/* 35436 */ {(12<<2)|2,{95,103,0}},
/* 35437 */ {(12<<2)|2,{95,103,0}},
/* 35438 */ {(12<<2)|2,{95,103,0}},
/* 35439 */ {(12<<2)|2,{95,103,0}},
/* 35440 */ {(12<<2)|2,{95,104,0}},
/* 35441 */ {(12<<2)|2,{95,104,0}},
/* 35442 */ {(12<<2)|2,{95,104,0}},
/* 35443 */ {(12<<2)|2,{95,104,0}},
/* 35444 */ {(12<<2)|2,{95,104,0}},
/* 35445 */ {(12<<2)|2,{95,104,0}},
/* 35446 */ {(12<<2)|2,{95,104,0}},
/* 35447 */ {(12<<2)|2,{95,104,0}},
/* 35448 */ {(12<<2)|2,{95,104,0}},
/* 35449 */ {(12<<2)|2,{95,104,0}},
/* 35450 */ {(12<<2)|2,{95,104,0}},
/* 35451 */ {(12<<2)|2,{95,104,0}},
/* 35452 */ {(12<<2)|2,{95,104,0}},
/* 35453 */ {(12<<2)|2,{95,104,0}},
/* 35454 */ {(12<<2)|2,{95,104,0}},
/* 35455 */ {(12<<2)|2,{95,104,0}},
/* 35456 */ {(12<<2)|2,{95,108,0}},
/* 35457 */ {(12<<2)|2,{95,108,0}},
/* 35458 */ {(12<<2)|2,{95,108,0}},
/* 35459 */ {(12<<2)|2,{95,108,0}},
/* 35460 */ {(12<<2)|2,{95,108,0}},
/* 35461 */ {(12<<2)|2,{95,108,0}},
/* 35462 */ {(12<<2)|2,{95,108,0}},
/* 35463 */ {(12<<2)|2,{95,108,0}},
/* 35464 */ {(12<<2)|2,{95,108,0}},
/* 35465 */ {(12<<2)|2,{95,108,0}},
/* 35466 */ {(12<<2)|2,{95,108,0}},
/* 35467 */ {(12<<2)|2,{95,108,0}},
/* 35468 */ {(12<<2)|2,{95,108,0}},
/* 35469 */ {(12<<2)|2,{95,108,0}},
/* 35470 */ {(12<<2)|2,{95,108,0}},
/* 35471 */ {(12<<2)|2,{95,108,0}},
/* 35472 */ {(12<<2)|2,{95,109,0}},
/* 35473 */ {(12<<2)|2,{95,109,0}},
/* 35474 */ {(12<<2)|2,{95,109,0}},
/* 35475 */ {(12<<2)|2,{95,109,0}},
/* 35476 */ {(12<<2)|2,{95,109,0}},
/* 35477 */ {(12<<2)|2,{95,109,0}},
/* 35478 */ {(12<<2)|2,{95,109,0}},
/* 35479 */ {(12<<2)|2,{95,109,0}},
/* 35480 */ {(12<<2)|2,{95,109,0}},
/* 35481 */ {(12<<2)|2,{95,109,0}},
/* 35482 */ {(12<<2)|2,{95,109,0}},
/* 35483 */ {(12<<2)|2,{95,109,0}},
/* 35484 */ {(12<<2)|2,{95,109,0}},
/* 35485 */ {(12<<2)|2,{95,109,0}},
/* 35486 */ {(12<<2)|2,{95,109,0}},
/* 35487 */ {(12<<2)|2,{95,109,0}},
/* 35488 */ {(12<<2)|2,{95,110,0}},
/* 35489 */ {(12<<2)|2,{95,110,0}},
/* 35490 */ {(12<<2)|2,{95,110,0}},
/* 35491 */ {(12<<2)|2,{95,110,0}},
/* 35492 */ {(12<<2)|2,{95,110,0}},
/* 35493 */ {(12<<2)|2,{95,110,0}},
/* 35494 */ {(12<<2)|2,{95,110,0}},
/* 35495 */ {(12<<2)|2,{95,110,0}},
/* 35496 */ {(12<<2)|2,{95,110,0}},
/* 35497 */ {(12<<2)|2,{95,110,0}},
/* 35498 */ {(12<<2)|2,{95,110,0}},
/* 35499 */ {(12<<2)|2,{95,110,0}},
/* 35500 */ {(12<<2)|2,{95,110,0}},
/* 35501 */ {(12<<2)|2,{95,110,0}},
/* 35502 */ {(12<<2)|2,{95,110,0}},
/* 35503 */ {(12<<2)|2,{95,110,0}},
/* 35504 */ {(12<<2)|2,{95,112,0}},
/* 35505 */ {(12<<2)|2,{95,112,0}},
/* 35506 */ {(12<<2)|2,{95,112,0}},
/* 35507 */ {(12<<2)|2,{95,112,0}},
/* 35508 */ {(12<<2)|2,{95,112,0}},
/* 35509 */ {(12<<2)|2,{95,112,0}},
/* 35510 */ {(12<<2)|2,{95,112,0}},
/* 35511 */ {(12<<2)|2,{95,112,0}},
/* 35512 */ {(12<<2)|2,{95,112,0}},
/* 35513 */ {(12<<2)|2,{95,112,0}},
/* 35514 */ {(12<<2)|2,{95,112,0}},
/* 35515 */ {(12<<2)|2,{95,112,0}},
/* 35516 */ {(12<<2)|2,{95,112,0}},
/* 35517 */ {(12<<2)|2,{95,112,0}},
/* 35518 */ {(12<<2)|2,{95,112,0}},
/* 35519 */ {(12<<2)|2,{95,112,0}},
/* 35520 */ {(12<<2)|2,{95,114,0}},
/* 35521 */ {(12<<2)|2,{95,114,0}},
/* 35522 */ {(12<<2)|2,{95,114,0}},
/* 35523 */ {(12<<2)|2,{95,114,0}},
/* 35524 */ {(12<<2)|2,{95,114,0}},
/* 35525 */ {(12<<2)|2,{95,114,0}},
/* 35526 */ {(12<<2)|2,{95,114,0}},
/* 35527 */ {(12<<2)|2,{95,114,0}},
/* 35528 */ {(12<<2)|2,{95,114,0}},
/* 35529 */ {(12<<2)|2,{95,114,0}},
/* 35530 */ {(12<<2)|2,{95,114,0}},
/* 35531 */ {(12<<2)|2,{95,114,0}},
/* 35532 */ {(12<<2)|2,{95,114,0}},
/* 35533 */ {(12<<2)|2,{95,114,0}},
/* 35534 */ {(12<<2)|2,{95,114,0}},
/* 35535 */ {(12<<2)|2,{95,114,0}},
/* 35536 */ {(12<<2)|2,{95,117,0}},
/* 35537 */ {(12<<2)|2,{95,117,0}},
/* 35538 */ {(12<<2)|2,{95,117,0}},
/* 35539 */ {(12<<2)|2,{95,117,0}},
/* 35540 */ {(12<<2)|2,{95,117,0}},
/* 35541 */ {(12<<2)|2,{95,117,0}},
/* 35542 */ {(12<<2)|2,{95,117,0}},
/* 35543 */ {(12<<2)|2,{95,117,0}},
/* 35544 */ {(12<<2)|2,{95,117,0}},
/* 35545 */ {(12<<2)|2,{95,117,0}},
/* 35546 */ {(12<<2)|2,{95,117,0}},
/* 35547 */ {(12<<2)|2,{95,117,0}},
/* 35548 */ {(12<<2)|2,{95,117,0}},
/* 35549 */ {(12<<2)|2,{95,117,0}},
/* 35550 */ {(12<<2)|2,{95,117,0}},
/* 35551 */ {(12<<2)|2,{95,117,0}},
/* 35552 */ {(13<<2)|2,{95,58,0}},
/* 35553 */ {(13<<2)|2,{95,58,0}},
/* 35554 */ {(13<<2)|2,{95,58,0}},
/* 35555 */ {(13<<2)|2,{95,58,0}},
/* 35556 */ {(13<<2)|2,{95,58,0}},
/* 35557 */ {(13<<2)|2,{95,58,0}},
/* 35558 */ {(13<<2)|2,{95,58,0}},
/* 35559 */ {(13<<2)|2,{95,58,0}},
/* 35560 */ {(13<<2)|2,{95,66,0}},
/* 35561 */ {(13<<2)|2,{95,66,0}},
/* 35562 */ {(13<<2)|2,{95,66,0}},
/* 35563 */ {(13<<2)|2,{95,66,0}},
/* 35564 */ {(13<<2)|2,{95,66,0}},
/* 35565 */ {(13<<2)|2,{95,66,0}},
/* 35566 */ {(13<<2)|2,{95,66,0}},
/* 35567 */ {(13<<2)|2,{95,66,0}},
/* 35568 */ {(13<<2)|2,{95,67,0}},
/* 35569 */ {(13<<2)|2,{95,67,0}},
/* 35570 */ {(13<<2)|2,{95,67,0}},
/* 35571 */ {(13<<2)|2,{95,67,0}},
/* 35572 */ {(13<<2)|2,{95,67,0}},
/* 35573 */ {(13<<2)|2,{95,67,0}},
/* 35574 */ {(13<<2)|2,{95,67,0}},
/* 35575 */ {(13<<2)|2,{95,67,0}},
/* 35576 */ {(13<<2)|2,{95,68,0}},
/* 35577 */ {(13<<2)|2,{95,68,0}},
/* 35578 */ {(13<<2)|2,{95,68,0}},
/* 35579 */ {(13<<2)|2,{95,68,0}},
/* 35580 */ {(13<<2)|2,{95,68,0}},
/* 35581 */ {(13<<2)|2,{95,68,0}},
/* 35582 */ {(13<<2)|2,{95,68,0}},
/* 35583 */ {(13<<2)|2,{95,68,0}},
/* 35584 */ {(13<<2)|2,{95,69,0}},
/* 35585 */ {(13<<2)|2,{95,69,0}},
/* 35586 */ {(13<<2)|2,{95,69,0}},
/* 35587 */ {(13<<2)|2,{95,69,0}},
/* 35588 */ {(13<<2)|2,{95,69,0}},
/* 35589 */ {(13<<2)|2,{95,69,0}},
/* 35590 */ {(13<<2)|2,{95,69,0}},
/* 35591 */ {(13<<2)|2,{95,69,0}},
/* 35592 */ {(13<<2)|2,{95,70,0}},
/* 35593 */ {(13<<2)|2,{95,70,0}},
/* 35594 */ {(13<<2)|2,{95,70,0}},
/* 35595 */ {(13<<2)|2,{95,70,0}},
/* 35596 */ {(13<<2)|2,{95,70,0}},
/* 35597 */ {(13<<2)|2,{95,70,0}},
/* 35598 */ {(13<<2)|2,{95,70,0}},
/* 35599 */ {(13<<2)|2,{95,70,0}},
/* 35600 */ {(13<<2)|2,{95,71,0}},
/* 35601 */ {(13<<2)|2,{95,71,0}},
/* 35602 */ {(13<<2)|2,{95,71,0}},
/* 35603 */ {(13<<2)|2,{95,71,0}},
/* 35604 */ {(13<<2)|2,{95,71,0}},
/* 35605 */ {(13<<2)|2,{95,71,0}},
/* 35606 */ {(13<<2)|2,{95,71,0}},
/* 35607 */ {(13<<2)|2,{95,71,0}},
/* 35608 */ {(13<<2)|2,{95,72,0}},
/* 35609 */ {(13<<2)|2,{95,72,0}},
/* 35610 */ {(13<<2)|2,{95,72,0}},
/* 35611 */ {(13<<2)|2,{95,72,0}},
/* 35612 */ {(13<<2)|2,{95,72,0}},
/* 35613 */ {(13<<2)|2,{95,72,0}},
/* 35614 */ {(13<<2)|2,{95,72,0}},
/* 35615 */ {(13<<2)|2,{95,72,0}},
/* 35616 */ {(13<<2)|2,{95,73,0}},
/* 35617 */ {(13<<2)|2,{95,73,0}},
/* 35618 */ {(13<<2)|2,{95,73,0}},
/* 35619 */ {(13<<2)|2,{95,73,0}},
/* 35620 */ {(13<<2)|2,{95,73,0}},
/* 35621 */ {(13<<2)|2,{95,73,0}},
/* 35622 */ {(13<<2)|2,{95,73,0}},
/* 35623 */ {(13<<2)|2,{95,73,0}},
/* 35624 */ {(13<<2)|2,{95,74,0}},
/* 35625 */ {(13<<2)|2,{95,74,0}},
/* 35626 */ {(13<<2)|2,{95,74,0}},
/* 35627 */ {(13<<2)|2,{95,74,0}},
/* 35628 */ {(13<<2)|2,{95,74,0}},
/* 35629 */ {(13<<2)|2,{95,74,0}},
/* 35630 */ {(13<<2)|2,{95,74,0}},
/* 35631 */ {(13<<2)|2,{95,74,0}},
/* 35632 */ {(13<<2)|2,{95,75,0}},
/* 35633 */ {(13<<2)|2,{95,75,0}},
/* 35634 */ {(13<<2)|2,{95,75,0}},
/* 35635 */ {(13<<2)|2,{95,75,0}},
/* 35636 */ {(13<<2)|2,{95,75,0}},
/* 35637 */ {(13<<2)|2,{95,75,0}},
/* 35638 */ {(13<<2)|2,{95,75,0}},
/* 35639 */ {(13<<2)|2,{95,75,0}},
/* 35640 */ {(13<<2)|2,{95,76,0}},
/* 35641 */ {(13<<2)|2,{95,76,0}},
/* 35642 */ {(13<<2)|2,{95,76,0}},
/* 35643 */ {(13<<2)|2,{95,76,0}},
/* 35644 */ {(13<<2)|2,{95,76,0}},
/* 35645 */ {(13<<2)|2,{95,76,0}},
/* 35646 */ {(13<<2)|2,{95,76,0}},
/* 35647 */ {(13<<2)|2,{95,76,0}},
/* 35648 */ {(13<<2)|2,{95,77,0}},
/* 35649 */ {(13<<2)|2,{95,77,0}},
/* 35650 */ {(13<<2)|2,{95,77,0}},
/* 35651 */ {(13<<2)|2,{95,77,0}},
/* 35652 */ {(13<<2)|2,{95,77,0}},
/* 35653 */ {(13<<2)|2,{95,77,0}},
/* 35654 */ {(13<<2)|2,{95,77,0}},
/* 35655 */ {(13<<2)|2,{95,77,0}},
/* 35656 */ {(13<<2)|2,{95,78,0}},
/* 35657 */ {(13<<2)|2,{95,78,0}},
/* 35658 */ {(13<<2)|2,{95,78,0}},
/* 35659 */ {(13<<2)|2,{95,78,0}},
/* 35660 */ {(13<<2)|2,{95,78,0}},
/* 35661 */ {(13<<2)|2,{95,78,0}},
/* 35662 */ {(13<<2)|2,{95,78,0}},
/* 35663 */ {(13<<2)|2,{95,78,0}},
/* 35664 */ {(13<<2)|2,{95,79,0}},
/* 35665 */ {(13<<2)|2,{95,79,0}},
/* 35666 */ {(13<<2)|2,{95,79,0}},
/* 35667 */ {(13<<2)|2,{95,79,0}},
/* 35668 */ {(13<<2)|2,{95,79,0}},
/* 35669 */ {(13<<2)|2,{95,79,0}},
/* 35670 */ {(13<<2)|2,{95,79,0}},
/* 35671 */ {(13<<2)|2,{95,79,0}},
/* 35672 */ {(13<<2)|2,{95,80,0}},
/* 35673 */ {(13<<2)|2,{95,80,0}},
/* 35674 */ {(13<<2)|2,{95,80,0}},
/* 35675 */ {(13<<2)|2,{95,80,0}},
/* 35676 */ {(13<<2)|2,{95,80,0}},
/* 35677 */ {(13<<2)|2,{95,80,0}},
/* 35678 */ {(13<<2)|2,{95,80,0}},
/* 35679 */ {(13<<2)|2,{95,80,0}},
/* 35680 */ {(13<<2)|2,{95,81,0}},
/* 35681 */ {(13<<2)|2,{95,81,0}},
/* 35682 */ {(13<<2)|2,{95,81,0}},
/* 35683 */ {(13<<2)|2,{95,81,0}},
/* 35684 */ {(13<<2)|2,{95,81,0}},
/* 35685 */ {(13<<2)|2,{95,81,0}},
/* 35686 */ {(13<<2)|2,{95,81,0}},
/* 35687 */ {(13<<2)|2,{95,81,0}},
/* 35688 */ {(13<<2)|2,{95,82,0}},
/* 35689 */ {(13<<2)|2,{95,82,0}},
/* 35690 */ {(13<<2)|2,{95,82,0}},
/* 35691 */ {(13<<2)|2,{95,82,0}},
/* 35692 */ {(13<<2)|2,{95,82,0}},
/* 35693 */ {(13<<2)|2,{95,82,0}},
/* 35694 */ {(13<<2)|2,{95,82,0}},
/* 35695 */ {(13<<2)|2,{95,82,0}},
/* 35696 */ {(13<<2)|2,{95,83,0}},
/* 35697 */ {(13<<2)|2,{95,83,0}},
/* 35698 */ {(13<<2)|2,{95,83,0}},
/* 35699 */ {(13<<2)|2,{95,83,0}},
/* 35700 */ {(13<<2)|2,{95,83,0}},
/* 35701 */ {(13<<2)|2,{95,83,0}},
/* 35702 */ {(13<<2)|2,{95,83,0}},
/* 35703 */ {(13<<2)|2,{95,83,0}},
/* 35704 */ {(13<<2)|2,{95,84,0}},
/* 35705 */ {(13<<2)|2,{95,84,0}},
/* 35706 */ {(13<<2)|2,{95,84,0}},
/* 35707 */ {(13<<2)|2,{95,84,0}},
/* 35708 */ {(13<<2)|2,{95,84,0}},
/* 35709 */ {(13<<2)|2,{95,84,0}},
/* 35710 */ {(13<<2)|2,{95,84,0}},
/* 35711 */ {(13<<2)|2,{95,84,0}},
/* 35712 */ {(13<<2)|2,{95,85,0}},
/* 35713 */ {(13<<2)|2,{95,85,0}},
/* 35714 */ {(13<<2)|2,{95,85,0}},
/* 35715 */ {(13<<2)|2,{95,85,0}},
/* 35716 */ {(13<<2)|2,{95,85,0}},
/* 35717 */ {(13<<2)|2,{95,85,0}},
/* 35718 */ {(13<<2)|2,{95,85,0}},
/* 35719 */ {(13<<2)|2,{95,85,0}},
/* 35720 */ {(13<<2)|2,{95,86,0}},
/* 35721 */ {(13<<2)|2,{95,86,0}},
/* 35722 */ {(13<<2)|2,{95,86,0}},
/* 35723 */ {(13<<2)|2,{95,86,0}},
/* 35724 */ {(13<<2)|2,{95,86,0}},
/* 35725 */ {(13<<2)|2,{95,86,0}},
/* 35726 */ {(13<<2)|2,{95,86,0}},
/* 35727 */ {(13<<2)|2,{95,86,0}},
/* 35728 */ {(13<<2)|2,{95,87,0}},
/* 35729 */ {(13<<2)|2,{95,87,0}},
/* 35730 */ {(13<<2)|2,{95,87,0}},
/* 35731 */ {(13<<2)|2,{95,87,0}},
/* 35732 */ {(13<<2)|2,{95,87,0}},
/* 35733 */ {(13<<2)|2,{95,87,0}},
/* 35734 */ {(13<<2)|2,{95,87,0}},
/* 35735 */ {(13<<2)|2,{95,87,0}},
/* 35736 */ {(13<<2)|2,{95,89,0}},
/* 35737 */ {(13<<2)|2,{95,89,0}},
/* 35738 */ {(13<<2)|2,{95,89,0}},
/* 35739 */ {(13<<2)|2,{95,89,0}},
/* 35740 */ {(13<<2)|2,{95,89,0}},
/* 35741 */ {(13<<2)|2,{95,89,0}},
/* 35742 */ {(13<<2)|2,{95,89,0}},
/* 35743 */ {(13<<2)|2,{95,89,0}},
/* 35744 */ {(13<<2)|2,{95,106,0}},
/* 35745 */ {(13<<2)|2,{95,106,0}},
/* 35746 */ {(13<<2)|2,{95,106,0}},
/* 35747 */ {(13<<2)|2,{95,106,0}},
/* 35748 */ {(13<<2)|2,{95,106,0}},
/* 35749 */ {(13<<2)|2,{95,106,0}},
/* 35750 */ {(13<<2)|2,{95,106,0}},
/* 35751 */ {(13<<2)|2,{95,106,0}},
/* 35752 */ {(13<<2)|2,{95,107,0}},
/* 35753 */ {(13<<2)|2,{95,107,0}},
/* 35754 */ {(13<<2)|2,{95,107,0}},
/* 35755 */ {(13<<2)|2,{95,107,0}},
/* 35756 */ {(13<<2)|2,{95,107,0}},
/* 35757 */ {(13<<2)|2,{95,107,0}},
/* 35758 */ {(13<<2)|2,{95,107,0}},
/* 35759 */ {(13<<2)|2,{95,107,0}},
/* 35760 */ {(13<<2)|2,{95,113,0}},
/* 35761 */ {(13<<2)|2,{95,113,0}},
/* 35762 */ {(13<<2)|2,{95,113,0}},
/* 35763 */ {(13<<2)|2,{95,113,0}},
/* 35764 */ {(13<<2)|2,{95,113,0}},
/* 35765 */ {(13<<2)|2,{95,113,0}},
/* 35766 */ {(13<<2)|2,{95,113,0}},
/* 35767 */ {(13<<2)|2,{95,113,0}},
/* 35768 */ {(13<<2)|2,{95,118,0}},
/* 35769 */ {(13<<2)|2,{95,118,0}},
/* 35770 */ {(13<<2)|2,{95,118,0}},
/* 35771 */ {(13<<2)|2,{95,118,0}},
/* 35772 */ {(13<<2)|2,{95,118,0}},
/* 35773 */ {(13<<2)|2,{95,118,0}},
/* 35774 */ {(13<<2)|2,{95,118,0}},
/* 35775 */ {(13<<2)|2,{95,118,0}},
/* 35776 */ {(13<<2)|2,{95,119,0}},
/* 35777 */ {(13<<2)|2,{95,119,0}},
/* 35778 */ {(13<<2)|2,{95,119,0}},
/* 35779 */ {(13<<2)|2,{95,119,0}},
/* 35780 */ {(13<<2)|2,{95,119,0}},
/* 35781 */ {(13<<2)|2,{95,119,0}},
/* 35782 */ {(13<<2)|2,{95,119,0}},
/* 35783 */ {(13<<2)|2,{95,119,0}},
/* 35784 */ {(13<<2)|2,{95,120,0}},
/* 35785 */ {(13<<2)|2,{95,120,0}},
/* 35786 */ {(13<<2)|2,{95,120,0}},
/* 35787 */ {(13<<2)|2,{95,120,0}},
/* 35788 */ {(13<<2)|2,{95,120,0}},
/* 35789 */ {(13<<2)|2,{95,120,0}},
/* 35790 */ {(13<<2)|2,{95,120,0}},
/* 35791 */ {(13<<2)|2,{95,120,0}},
/* 35792 */ {(13<<2)|2,{95,121,0}},
/* 35793 */ {(13<<2)|2,{95,121,0}},
/* 35794 */ {(13<<2)|2,{95,121,0}},
/* 35795 */ {(13<<2)|2,{95,121,0}},
/* 35796 */ {(13<<2)|2,{95,121,0}},
/* 35797 */ {(13<<2)|2,{95,121,0}},
/* 35798 */ {(13<<2)|2,{95,121,0}},
/* 35799 */ {(13<<2)|2,{95,121,0}},
/* 35800 */ {(13<<2)|2,{95,122,0}},
/* 35801 */ {(13<<2)|2,{95,122,0}},
/* 35802 */ {(13<<2)|2,{95,122,0}},
/* 35803 */ {(13<<2)|2,{95,122,0}},
/* 35804 */ {(13<<2)|2,{95,122,0}},
/* 35805 */ {(13<<2)|2,{95,122,0}},
/* 35806 */ {(13<<2)|2,{95,122,0}},
/* 35807 */ {(13<<2)|2,{95,122,0}},
/* 35808 */ {(14<<2)|2,{95,38,0}},
/* 35809 */ {(14<<2)|2,{95,38,0}},
/* 35810 */ {(14<<2)|2,{95,38,0}},
/* 35811 */ {(14<<2)|2,{95,38,0}},
/* 35812 */ {(14<<2)|2,{95,42,0}},
/* 35813 */ {(14<<2)|2,{95,42,0}},
/* 35814 */ {(14<<2)|2,{95,42,0}},
/* 35815 */ {(14<<2)|2,{95,42,0}},
/* 35816 */ {(14<<2)|2,{95,44,0}},
/* 35817 */ {(14<<2)|2,{95,44,0}},
/* 35818 */ {(14<<2)|2,{95,44,0}},
/* 35819 */ {(14<<2)|2,{95,44,0}},
/* 35820 */ {(14<<2)|2,{95,59,0}},
/* 35821 */ {(14<<2)|2,{95,59,0}},
/* 35822 */ {(14<<2)|2,{95,59,0}},
/* 35823 */ {(14<<2)|2,{95,59,0}},
/* 35824 */ {(14<<2)|2,{95,88,0}},
/* 35825 */ {(14<<2)|2,{95,88,0}},
/* 35826 */ {(14<<2)|2,{95,88,0}},
/* 35827 */ {(14<<2)|2,{95,88,0}},
/* 35828 */ {(14<<2)|2,{95,90,0}},
/* 35829 */ {(14<<2)|2,{95,90,0}},
/* 35830 */ {(14<<2)|2,{95,90,0}},
/* 35831 */ {(14<<2)|2,{95,90,0}},
/* 35832 */ {(16<<2)|2,{95,33,0}},
/* 35833 */ {(16<<2)|2,{95,34,0}},
/* 35834 */ {(16<<2)|2,{95,40,0}},
/* 35835 */ {(16<<2)|2,{95,41,0}},
/* 35836 */ {(16<<2)|2,{95,63,0}},
/* 35837 */ {(6<<2)|1,{95,0,0}},
/* 35838 */ {(6<<2)|1,{95,0,0}},
/* 35839 */ {(6<<2)|1,{95,0,0}},
/* 35840 */ {(16<<2)|3,{98,48,48}},
/* 35841 */ {(16<<2)|3,{98,48,49}},
/* 35842 */ {(16<<2)|3,{98,48,50}},
/* 35843 */ {(16<<2)|3,{98,48,97}},
/* 35844 */ {(16<<2)|3,{98,48,99}},
/* 35845 */ {(16<<2)|3,{98,48,101}},
/* 35846 */ {(16<<2)|3,{98,48,105}},
/* 35847 */ {(16<<2)|3,{98,48,111}},
/* 35848 */ {(16<<2)|3,{98,48,115}},
/* 35849 */ {(16<<2)|3,{98,48,116}},
/* 35850 */ {(11<<2)|2,{98,48,0}},
/* 35851 */ {(11<<2)|2,{98,48,0}},
/* 35852 */ {(11<<2)|2,{98,48,0}},
/* 35853 */ {(11<<2)|2,{98,48,0}},
/* 35854 */ {(11<<2)|2,{98,48,0}},
/* 35855 */ {(11<<2)|2,{98,48,0}},
/* 35856 */ {(11<<2)|2,{98,48,0}},
/* 35857 */ {(11<<2)|2,{98,48,0}},
/* 35858 */ {(11<<2)|2,{98,48,0}},
/* 35859 */ {(11<<2)|2,{98,48,0}},
/* 35860 */ {(11<<2)|2,{98,48,0}},
/* 35861 */ {(11<<2)|2,{98,48,0}},
/* 35862 */ {(11<<2)|2,{98,48,0}},
/* 35863 */ {(11<<2)|2,{98,48,0}},
/* 35864 */ {(11<<2)|2,{98,48,0}},
/* 35865 */ {(11<<2)|2,{98,48,0}},
/* 35866 */ {(11<<2)|2,{98,48,0}},
/* 35867 */ {(11<<2)|2,{98,48,0}},
/* 35868 */ {(11<<2)|2,{98,48,0}},
/* 35869 */ {(11<<2)|2,{98,48,0}},
/* 35870 */ {(11<<2)|2,{98,48,0}},
/* 35871 */ {(11<<2)|2,{98,48,0}},
/* 35872 */ {(16<<2)|3,{98,49,48}},
/* 35873 */ {(16<<2)|3,{98,49,49}},
/* 35874 */ {(16<<2)|3,{98,49,50}},
/* 35875 */ {(16<<2)|3,{98,49,97}},
/* 35876 */ {(16<<2)|3,{98,49,99}},
/* 35877 */ {(16<<2)|3,{98,49,101}},
/* 35878 */ {(16<<2)|3,{98,49,105}},
/* 35879 */ {(16<<2)|3,{98,49,111}},
/* 35880 */ {(16<<2)|3,{98,49,115}},
/* 35881 */ {(16<<2)|3,{98,49,116}},
/* 35882 */ {(11<<2)|2,{98,49,0}},
/* 35883 */ {(11<<2)|2,{98,49,0}},
/* 35884 */ {(11<<2)|2,{98,49,0}},
/* 35885 */ {(11<<2)|2,{98,49,0}},
/* 35886 */ {(11<<2)|2,{98,49,0}},
/* 35887 */ {(11<<2)|2,{98,49,0}},
/* 35888 */ {(11<<2)|2,{98,49,0}},
/* 35889 */ {(11<<2)|2,{98,49,0}},
/* 35890 */ {(11<<2)|2,{98,49,0}},
/* 35891 */ {(11<<2)|2,{98,49,0}},
/* 35892 */ {(11<<2)|2,{98,49,0}},
/* 35893 */ {(11<<2)|2,{98,49,0}},
/* 35894 */ {(11<<2)|2,{98,49,0}},
/* 35895 */ {(11<<2)|2,{98,49,0}},
/* 35896 */ {(11<<2)|2,{98,49,0}},
/* 35897 */ {(11<<2)|2,{98,49,0}},
/* 35898 */ {(11<<2)|2,{98,49,0}},
/* 35899 */ {(11<<2)|2,{98,49,0}},
/* 35900 */ {(11<<2)|2,{98,49,0}},
/* 35901 */ {(11<<2)|2,{98,49,0}},
/* 35902 */ {(11<<2)|2,{98,49,0}},
/* 35903 */ {(11<<2)|2,{98,49,0}},
/* 35904 */ {(16<<2)|3,{98,50,48}},
/* 35905 */ {(16<<2)|3,{98,50,49}},
/* 35906 */ {(16<<2)|3,{98,50,50}},
/* 35907 */ {(16<<2)|3,{98,50,97}},
/* 35908 */ {(16<<2)|3,{98,50,99}},
/* 35909 */ {(16<<2)|3,{98,50,101}},
/* 35910 */ {(16<<2)|3,{98,50,105}},
/* 35911 */ {(16<<2)|3,{98,50,111}},
/* 35912 */ {(16<<2)|3,{98,50,115}},
/* 35913 */ {(16<<2)|3,{98,50,116}},
/* 35914 */ {(11<<2)|2,{98,50,0}},
/* 35915 */ {(11<<2)|2,{98,50,0}},
/* 35916 */ {(11<<2)|2,{98,50,0}},
/* 35917 */ {(11<<2)|2,{98,50,0}},
/* 35918 */ {(11<<2)|2,{98,50,0}},
/* 35919 */ {(11<<2)|2,{98,50,0}},
/* 35920 */ {(11<<2)|2,{98,50,0}},
/* 35921 */ {(11<<2)|2,{98,50,0}},
/* 35922 */ {(11<<2)|2,{98,50,0}},
/* 35923 */ {(11<<2)|2,{98,50,0}},
/* 35924 */ {(11<<2)|2,{98,50,0}},
/* 35925 */ {(11<<2)|2,{98,50,0}},
/* 35926 */ {(11<<2)|2,{98,50,0}},
/* 35927 */ {(11<<2)|2,{98,50,0}},
/* 35928 */ {(11<<2)|2,{98,50,0}},
/* 35929 */ {(11<<2)|2,{98,50,0}},
/* 35930 */ {(11<<2)|2,{98,50,0}},
/* 35931 */ {(11<<2)|2,{98,50,0}},
/* 35932 */ {(11<<2)|2,{98,50,0}},
/* 35933 */ {(11<<2)|2,{98,50,0}},
/* 35934 */ {(11<<2)|2,{98,50,0}},
/* 35935 */ {(11<<2)|2,{98,50,0}},
/* 35936 */ {(16<<2)|3,{98,97,48}},
/* 35937 */ {(16<<2)|3,{98,97,49}},
/* 35938 */ {(16<<2)|3,{98,97,50}},
/* 35939 */ {(16<<2)|3,{98,97,97}},
/* 35940 */ {(16<<2)|3,{98,97,99}},
/* 35941 */ {(16<<2)|3,{98,97,101}},
/* 35942 */ {(16<<2)|3,{98,97,105}},
/* 35943 */ {(16<<2)|3,{98,97,111}},
/* 35944 */ {(16<<2)|3,{98,97,115}},
/* 35945 */ {(16<<2)|3,{98,97,116}},
/* 35946 */ {(11<<2)|2,{98,97,0}},
/* 35947 */ {(11<<2)|2,{98,97,0}},
/* 35948 */ {(11<<2)|2,{98,97,0}},
/* 35949 */ {(11<<2)|2,{98,97,0}},
/* 35950 */ {(11<<2)|2,{98,97,0}},
/* 35951 */ {(11<<2)|2,{98,97,0}},
/* 35952 */ {(11<<2)|2,{98,97,0}},
/* 35953 */ {(11<<2)|2,{98,97,0}},
/* 35954 */ {(11<<2)|2,{98,97,0}},
/* 35955 */ {(11<<2)|2,{98,97,0}},
/* 35956 */ {(11<<2)|2,{98,97,0}},
/* 35957 */ {(11<<2)|2,{98,97,0}},
/* 35958 */ {(11<<2)|2,{98,97,0}},
/* 35959 */ {(11<<2)|2,{98,97,0}},
/* 35960 */ {(11<<2)|2,{98,97,0}},
/* 35961 */ {(11<<2)|2,{98,97,0}},
/* 35962 */ {(11<<2)|2,{98,97,0}},
/* 35963 */ {(11<<2)|2,{98,97,0}},
/* 35964 */ {(11<<2)|2,{98,97,0}},
/* 35965 */ {(11<<2)|2,{98,97,0}},
/* 35966 */ {(11<<2)|2,{98,97,0}},
/* 35967 */ {(11<<2)|2,{98,97,0}},
/* 35968 */ {(16<<2)|3,{98,99,48}},
/* 35969 */ {(16<<2)|3,{98,99,49}},
/* 35970 */ {(16<<2)|3,{98,99,50}},
/* 35971 */ {(16<<2)|3,{98,99,97}},
/* 35972 */ {(16<<2)|3,{98,99,99}},
/* 35973 */ {(16<<2)|3,{98,99,101}},
/* 35974 */ {(16<<2)|3,{98,99,105}},
/* 35975 */ {(16<<2)|3,{98,99,111}},
/* 35976 */ {(16<<2)|3,{98,99,115}},
/* 35977 */ {(16<<2)|3,{98,99,116}},
/* 35978 */ {(11<<2)|2,{98,99,0}},
/* 35979 */ {(11<<2)|2,{98,99,0}},
/* 35980 */ {(11<<2)|2,{98,99,0}},
/* 35981 */ {(11<<2)|2,{98,99,0}},
/* 35982 */ {(11<<2)|2,{98,99,0}},
/* 35983 */ {(11<<2)|2,{98,99,0}},
/* 35984 */ {(11<<2)|2,{98,99,0}},
/* 35985 */ {(11<<2)|2,{98,99,0}},
/* 35986 */ {(11<<2)|2,{98,99,0}},
/* 35987 */ {(11<<2)|2,{98,99,0}},
/* 35988 */ {(11<<2)|2,{98,99,0}},
/* 35989 */ {(11<<2)|2,{98,99,0}},
/* 35990 */ {(11<<2)|2,{98,99,0}},
/* 35991 */ {(11<<2)|2,{98,99,0}},
/* 35992 */ {(11<<2)|2,{98,99,0}},
/* 35993 */ {(11<<2)|2,{98,99,0}},
/* 35994 */ {(11<<2)|2,{98,99,0}},
/* 35995 */ {(11<<2)|2,{98,99,0}},
/* 35996 */ {(11<<2)|2,{98,99,0}},
/* 35997 */ {(11<<2)|2,{98,99,0}},
/* 35998 */ {(11<<2)|2,{98,99,0}},
/* 35999 */ {(11<<2)|2,{98,99,0}},
/* 36000 */ {(16<<2)|3,{98,101,48}},
/* 36001 */ {(16<<2)|3,{98,101,49}},
/* 36002 */ {(16<<2)|3,{98,101,50}},
/* 36003 */ {(16<<2)|3,{98,101,97}},
/* 36004 */ {(16<<2)|3,{98,101,99}},
/* 36005 */ {(16<<2)|3,{98,101,101}},
/* 36006 */ {(16<<2)|3,{98,101,105}},
/* 36007 */ {(16<<2)|3,{98,101,111}},
/* 36008 */ {(16<<2)|3,{98,101,115}},
/* 36009 */ {(16<<2)|3,{98,101,116}},
/* 36010 */ {(11<<2)|2,{98,101,0}},
/* 36011 */ {(11<<2)|2,{98,101,0}},
/* 36012 */ {(11<<2)|2,{98,101,0}},
/* 36013 */ {(11<<2)|2,{98,101,0}},
/* 36014 */ {(11<<2)|2,{98,101,0}},
/* 36015 */ {(11<<2)|2,{98,101,0}},
/* 36016 */ {(11<<2)|2,{98,101,0}},
/* 36017 */ {(11<<2)|2,{98,101,0}},
/* 36018 */ {(11<<2)|2,{98,101,0}},
/* 36019 */ {(11<<2)|2,{98,101,0}},
/* 36020 */ {(11<<2)|2,{98,101,0}},
/* 36021 */ {(11<<2)|2,{98,101,0}},
/* 36022 */ {(11<<2)|2,{98,101,0}},
/* 36023 */ {(11<<2)|2,{98,101,0}},
/* 36024 */ {(11<<2)|2,{98,101,0}},
/* 36025 */ {(11<<2)|2,{98,101,0}},
/* 36026 */ {(11<<2)|2,{98,101,0}},
/* 36027 */ {(11<<2)|2,{98,101,0}},
/* 36028 */ {(11<<2)|2,{98,101,0}},
/* 36029 */ {(11<<2)|2,{98,101,0}},
/* 36030 */ {(11<<2)|2,{98,101,0}},
/* 36031 */ {(11<<2)|2,{98,101,0}},
/* 36032 */ {(16<<2)|3,{98,105,48}},
/* 36033 */ {(16<<2)|3,{98,105,49}},
/* 36034 */ {(16<<2)|3,{98,105,50}},
/* 36035 */ {(16<<2)|3,{98,105,97}},
/* 36036 */ {(16<<2)|3,{98,105,99}},
/* 36037 */ {(16<<2)|3,{98,105,101}},
/* 36038 */ {(16<<2)|3,{98,105,105}},
/* 36039 */ {(16<<2)|3,{98,105,111}},
/* 36040 */ {(16<<2)|3,{98,105,115}},
/* 36041 */ {(16<<2)|3,{98,105,116}},
/* 36042 */ {(11<<2)|2,{98,105,0}},
/* 36043 */ {(11<<2)|2,{98,105,0}},
/* 36044 */ {(11<<2)|2,{98,105,0}},
/* 36045 */ {(11<<2)|2,{98,105,0}},
/* 36046 */ {(11<<2)|2,{98,105,0}},
/* 36047 */ {(11<<2)|2,{98,105,0}},
/* 36048 */ {(11<<2)|2,{98,105,0}},
/* 36049 */ {(11<<2)|2,{98,105,0}},
/* 36050 */ {(11<<2)|2,{98,105,0}},
/* 36051 */ {(11<<2)|2,{98,105,0}},
/* 36052 */ {(11<<2)|2,{98,105,0}},
/* 36053 */ {(11<<2)|2,{98,105,0}},
/* 36054 */ {(11<<2)|2,{98,105,0}},
/* 36055 */ {(11<<2)|2,{98,105,0}},
/* 36056 */ {(11<<2)|2,{98,105,0}},
/* 36057 */ {(11<<2)|2,{98,105,0}},
/* 36058 */ {(11<<2)|2,{98,105,0}},
/* 36059 */ {(11<<2)|2,{98,105,0}},
/* 36060 */ {(11<<2)|2,{98,105,0}},
/* 36061 */ {(11<<2)|2,{98,105,0}},
/* 36062 */ {(11<<2)|2,{98,105,0}},
/* 36063 */ {(11<<2)|2,{98,105,0}},
/* 36064 */ {(16<<2)|3,{98,111,48}},
/* 36065 */ {(16<<2)|3,{98,111,49}},
/* 36066 */ {(16<<2)|3,{98,111,50}},
/* 36067 */ {(16<<2)|3,{98,111,97}},
/* 36068 */ {(16<<2)|3,{98,111,99}},
/* 36069 */ {(16<<2)|3,{98,111,101}},
/* 36070 */ {(16<<2)|3,{98,111,105}},
/* 36071 */ {(16<<2)|3,{98,111,111}},
/* 36072 */ {(16<<2)|3,{98,111,115}},
/* 36073 */ {(16<<2)|3,{98,111,116}},
/* 36074 */ {(11<<2)|2,{98,111,0}},
/* 36075 */ {(11<<2)|2,{98,111,0}},
/* 36076 */ {(11<<2)|2,{98,111,0}},
/* 36077 */ {(11<<2)|2,{98,111,0}},
/* 36078 */ {(11<<2)|2,{98,111,0}},
/* 36079 */ {(11<<2)|2,{98,111,0}},
/* 36080 */ {(11<<2)|2,{98,111,0}},
/* 36081 */ {(11<<2)|2,{98,111,0}},
/* 36082 */ {(11<<2)|2,{98,111,0}},
/* 36083 */ {(11<<2)|2,{98,111,0}},
/* 36084 */ {(11<<2)|2,{98,111,0}},
/* 36085 */ {(11<<2)|2,{98,111,0}},
/* 36086 */ {(11<<2)|2,{98,111,0}},
/* 36087 */ {(11<<2)|2,{98,111,0}},
/* 36088 */ {(11<<2)|2,{98,111,0}},
/* 36089 */ {(11<<2)|2,{98,111,0}},
/* 36090 */ {(11<<2)|2,{98,111,0}},
/* 36091 */ {(11<<2)|2,{98,111,0}},
/* 36092 */ {(11<<2)|2,{98,111,0}},
/* 36093 */ {(11<<2)|2,{98,111,0}},
/* 36094 */ {(11<<2)|2,{98,111,0}},
/* 36095 */ {(11<<2)|2,{98,111,0}},
/* 36096 */ {(16<<2)|3,{98,115,48}},
/* 36097 */ {(16<<2)|3,{98,115,49}},
/* 36098 */ {(16<<2)|3,{98,115,50}},
/* 36099 */ {(16<<2)|3,{98,115,97}},
/* 36100 */ {(16<<2)|3,{98,115,99}},
/* 36101 */ {(16<<2)|3,{98,115,101}},
/* 36102 */ {(16<<2)|3,{98,115,105}},
/* 36103 */ {(16<<2)|3,{98,115,111}},
/* 36104 */ {(16<<2)|3,{98,115,115}},
/* 36105 */ {(16<<2)|3,{98,115,116}},
/* 36106 */ {(11<<2)|2,{98,115,0}},
/* 36107 */ {(11<<2)|2,{98,115,0}},
/* 36108 */ {(11<<2)|2,{98,115,0}},
/* 36109 */ {(11<<2)|2,{98,115,0}},
/* 36110 */ {(11<<2)|2,{98,115,0}},
/* 36111 */ {(11<<2)|2,{98,115,0}},
/* 36112 */ {(11<<2)|2,{98,115,0}},
/* 36113 */ {(11<<2)|2,{98,115,0}},
/* 36114 */ {(11<<2)|2,{98,115,0}},
/* 36115 */ {(11<<2)|2,{98,115,0}},
/* 36116 */ {(11<<2)|2,{98,115,0}},
/* 36117 */ {(11<<2)|2,{98,115,0}},
/* 36118 */ {(11<<2)|2,{98,115,0}},
/* 36119 */ {(11<<2)|2,{98,115,0}},
/* 36120 */ {(11<<2)|2,{98,115,0}},
/* 36121 */ {(11<<2)|2,{98,115,0}},
/* 36122 */ {(11<<2)|2,{98,115,0}},
/* 36123 */ {(11<<2)|2,{98,115,0}},
/* 36124 */ {(11<<2)|2,{98,115,0}},
/* 36125 */ {(11<<2)|2,{98,115,0}},
/* 36126 */ {(11<<2)|2,{98,115,0}},
/* 36127 */ {(11<<2)|2,{98,115,0}},
/* 36128 */ {(16<<2)|3,{98,116,48}},
/* 36129 */ {(16<<2)|3,{98,116,49}},
/* 36130 */ {(16<<2)|3,{98,116,50}},
/* 36131 */ {(16<<2)|3,{98,116,97}},
/* 36132 */ {(16<<2)|3,{98,116,99}},
/* 36133 */ {(16<<2)|3,{98,116,101}},
/* 36134 */ {(16<<2)|3,{98,116,105}},
/* 36135 */ {(16<<2)|3,{98,116,111}},
/* 36136 */ {(16<<2)|3,{98,116,115}},
/* 36137 */ {(16<<2)|3,{98,116,116}},
/* 36138 */ {(11<<2)|2,{98,116,0}},
/* 36139 */ {(11<<2)|2,{98,116,0}},
/* 36140 */ {(11<<2)|2,{98,116,0}},
/* 36141 */ {(11<<2)|2,{98,116,0}},
/* 36142 */ {(11<<2)|2,{98,116,0}},
/* 36143 */ {(11<<2)|2,{98,116,0}},
/* 36144 */ {(11<<2)|2,{98,116,0}},
/* 36145 */ {(11<<2)|2,{98,116,0}},
/* 36146 */ {(11<<2)|2,{98,116,0}},
/* 36147 */ {(11<<2)|2,{98,116,0}},
/* 36148 */ {(11<<2)|2,{98,116,0}},
/* 36149 */ {(11<<2)|2,{98,116,0}},
/* 36150 */ {(11<<2)|2,{98,116,0}},
/* 36151 */ {(11<<2)|2,{98,116,0}},
/* 36152 */ {(11<<2)|2,{98,116,0}},
/* 36153 */ {(11<<2)|2,{98,116,0}},
/* 36154 */ {(11<<2)|2,{98,116,0}},
/* 36155 */ {(11<<2)|2,{98,116,0}},
/* 36156 */ {(11<<2)|2,{98,116,0}},
/* 36157 */ {(11<<2)|2,{98,116,0}},
/* 36158 */ {(11<<2)|2,{98,116,0}},
/* 36159 */ {(11<<2)|2,{98,116,0}},
/* 36160 */ {(12<<2)|2,{98,32,0}},
/* 36161 */ {(12<<2)|2,{98,32,0}},
/* 36162 */ {(12<<2)|2,{98,32,0}},
/* 36163 */ {(12<<2)|2,{98,32,0}},
/* 36164 */ {(12<<2)|2,{98,32,0}},
/* 36165 */ {(12<<2)|2,{98,32,0}},
/* 36166 */ {(12<<2)|2,{98,32,0}},
/* 36167 */ {(12<<2)|2,{98,32,0}},
/* 36168 */ {(12<<2)|2,{98,32,0}},
/* 36169 */ {(12<<2)|2,{98,32,0}},
/* 36170 */ {(12<<2)|2,{98,32,0}},
/* 36171 */ {(12<<2)|2,{98,32,0}},
/* 36172 */ {(12<<2)|2,{98,32,0}},
/* 36173 */ {(12<<2)|2,{98,32,0}},
/* 36174 */ {(12<<2)|2,{98,32,0}},
/* 36175 */ {(12<<2)|2,{98,32,0}},
/* 36176 */ {(12<<2)|2,{98,37,0}},
/* 36177 */ {(12<<2)|2,{98,37,0}},
/* 36178 */ {(12<<2)|2,{98,37,0}},
/* 36179 */ {(12<<2)|2,{98,37,0}},
/* 36180 */ {(12<<2)|2,{98,37,0}},
/* 36181 */ {(12<<2)|2,{98,37,0}},
/* 36182 */ {(12<<2)|2,{98,37,0}},
/* 36183 */ {(12<<2)|2,{98,37,0}},
/* 36184 */ {(12<<2)|2,{98,37,0}},
/* 36185 */ {(12<<2)|2,{98,37,0}},
/* 36186 */ {(12<<2)|2,{98,37,0}},
/* 36187 */ {(12<<2)|2,{98,37,0}},
/* 36188 */ {(12<<2)|2,{98,37,0}},
/* 36189 */ {(12<<2)|2,{98,37,0}},
/* 36190 */ {(12<<2)|2,{98,37,0}},
/* 36191 */ {(12<<2)|2,{98,37,0}},
/* 36192 */ {(12<<2)|2,{98,45,0}},
/* 36193 */ {(12<<2)|2,{98,45,0}},
/* 36194 */ {(12<<2)|2,{98,45,0}},
/* 36195 */ {(12<<2)|2,{98,45,0}},
/* 36196 */ {(12<<2)|2,{98,45,0}},
/* 36197 */ {(12<<2)|2,{98,45,0}},
/* 36198 */ {(12<<2)|2,{98,45,0}},
/* 36199 */ {(12<<2)|2,{98,45,0}},
/* 36200 */ {(12<<2)|2,{98,45,0}},
/* 36201 */ {(12<<2)|2,{98,45,0}},
/* 36202 */ {(12<<2)|2,{98,45,0}},
/* 36203 */ {(12<<2)|2,{98,45,0}},
/* 36204 */ {(12<<2)|2,{98,45,0}},
/* 36205 */ {(12<<2)|2,{98,45,0}},
/* 36206 */ {(12<<2)|2,{98,45,0}},
/* 36207 */ {(12<<2)|2,{98,45,0}},
/* 36208 */ {(12<<2)|2,{98,46,0}},
/* 36209 */ {(12<<2)|2,{98,46,0}},
/* 36210 */ {(12<<2)|2,{98,46,0}},
/* 36211 */ {(12<<2)|2,{98,46,0}},
/* 36212 */ {(12<<2)|2,{98,46,0}},
/* 36213 */ {(12<<2)|2,{98,46,0}},
/* 36214 */ {(12<<2)|2,{98,46,0}},
/* 36215 */ {(12<<2)|2,{98,46,0}},
/* 36216 */ {(12<<2)|2,{98,46,0}},
/* 36217 */ {(12<<2)|2,{98,46,0}},
/* 36218 */ {(12<<2)|2,{98,46,0}},
/* 36219 */ {(12<<2)|2,{98,46,0}},
/* 36220 */ {(12<<2)|2,{98,46,0}},
/* 36221 */ {(12<<2)|2,{98,46,0}},
/* 36222 */ {(12<<2)|2,{98,46,0}},
/* 36223 */ {(12<<2)|2,{98,46,0}},
/* 36224 */ {(12<<2)|2,{98,47,0}},
/* 36225 */ {(12<<2)|2,{98,47,0}},
/* 36226 */ {(12<<2)|2,{98,47,0}},
/* 36227 */ {(12<<2)|2,{98,47,0}},
/* 36228 */ {(12<<2)|2,{98,47,0}},
/* 36229 */ {(12<<2)|2,{98,47,0}},
/* 36230 */ {(12<<2)|2,{98,47,0}},
/* 36231 */ {(12<<2)|2,{98,47,0}},
/* 36232 */ {(12<<2)|2,{98,47,0}},
/* 36233 */ {(12<<2)|2,{98,47,0}},
/* 36234 */ {(12<<2)|2,{98,47,0}},
/* 36235 */ {(12<<2)|2,{98,47,0}},
/* 36236 */ {(12<<2)|2,{98,47,0}},
/* 36237 */ {(12<<2)|2,{98,47,0}},
/* 36238 */ {(12<<2)|2,{98,47,0}},
/* 36239 */ {(12<<2)|2,{98,47,0}},
/* 36240 */ {(12<<2)|2,{98,51,0}},
/* 36241 */ {(12<<2)|2,{98,51,0}},
/* 36242 */ {(12<<2)|2,{98,51,0}},
/* 36243 */ {(12<<2)|2,{98,51,0}},
/* 36244 */ {(12<<2)|2,{98,51,0}},
/* 36245 */ {(12<<2)|2,{98,51,0}},
/* 36246 */ {(12<<2)|2,{98,51,0}},
/* 36247 */ {(12<<2)|2,{98,51,0}},
/* 36248 */ {(12<<2)|2,{98,51,0}},
/* 36249 */ {(12<<2)|2,{98,51,0}},
/* 36250 */ {(12<<2)|2,{98,51,0}},
/* 36251 */ {(12<<2)|2,{98,51,0}},
/* 36252 */ {(12<<2)|2,{98,51,0}},
/* 36253 */ {(12<<2)|2,{98,51,0}},
/* 36254 */ {(12<<2)|2,{98,51,0}},
/* 36255 */ {(12<<2)|2,{98,51,0}},
/* 36256 */ {(12<<2)|2,{98,52,0}},
/* 36257 */ {(12<<2)|2,{98,52,0}},
/* 36258 */ {(12<<2)|2,{98,52,0}},
/* 36259 */ {(12<<2)|2,{98,52,0}},
/* 36260 */ {(12<<2)|2,{98,52,0}},
/* 36261 */ {(12<<2)|2,{98,52,0}},
/* 36262 */ {(12<<2)|2,{98,52,0}},
/* 36263 */ {(12<<2)|2,{98,52,0}},
/* 36264 */ {(12<<2)|2,{98,52,0}},
/* 36265 */ {(12<<2)|2,{98,52,0}},
/* 36266 */ {(12<<2)|2,{98,52,0}},
/* 36267 */ {(12<<2)|2,{98,52,0}},
/* 36268 */ {(12<<2)|2,{98,52,0}},
/* 36269 */ {(12<<2)|2,{98,52,0}},
/* 36270 */ {(12<<2)|2,{98,52,0}},
/* 36271 */ {(12<<2)|2,{98,52,0}},
/* 36272 */ {(12<<2)|2,{98,53,0}},
/* 36273 */ {(12<<2)|2,{98,53,0}},
/* 36274 */ {(12<<2)|2,{98,53,0}},
/* 36275 */ {(12<<2)|2,{98,53,0}},
/* 36276 */ {(12<<2)|2,{98,53,0}},
/* 36277 */ {(12<<2)|2,{98,53,0}},
/* 36278 */ {(12<<2)|2,{98,53,0}},
/* 36279 */ {(12<<2)|2,{98,53,0}},
/* 36280 */ {(12<<2)|2,{98,53,0}},
/* 36281 */ {(12<<2)|2,{98,53,0}},
/* 36282 */ {(12<<2)|2,{98,53,0}},
/* 36283 */ {(12<<2)|2,{98,53,0}},
/* 36284 */ {(12<<2)|2,{98,53,0}},
/* 36285 */ {(12<<2)|2,{98,53,0}},
/* 36286 */ {(12<<2)|2,{98,53,0}},
/* 36287 */ {(12<<2)|2,{98,53,0}},
/* 36288 */ {(12<<2)|2,{98,54,0}},
/* 36289 */ {(12<<2)|2,{98,54,0}},
/* 36290 */ {(12<<2)|2,{98,54,0}},
/* 36291 */ {(12<<2)|2,{98,54,0}},
/* 36292 */ {(12<<2)|2,{98,54,0}},
/* 36293 */ {(12<<2)|2,{98,54,0}},
/* 36294 */ {(12<<2)|2,{98,54,0}},
/* 36295 */ {(12<<2)|2,{98,54,0}},
/* 36296 */ {(12<<2)|2,{98,54,0}},
/* 36297 */ {(12<<2)|2,{98,54,0}},
/* 36298 */ {(12<<2)|2,{98,54,0}},
/* 36299 */ {(12<<2)|2,{98,54,0}},
/* 36300 */ {(12<<2)|2,{98,54,0}},
/* 36301 */ {(12<<2)|2,{98,54,0}},
/* 36302 */ {(12<<2)|2,{98,54,0}},
/* 36303 */ {(12<<2)|2,{98,54,0}},
/* 36304 */ {(12<<2)|2,{98,55,0}},
/* 36305 */ {(12<<2)|2,{98,55,0}},
/* 36306 */ {(12<<2)|2,{98,55,0}},
/* 36307 */ {(12<<2)|2,{98,55,0}},
/* 36308 */ {(12<<2)|2,{98,55,0}},
/* 36309 */ {(12<<2)|2,{98,55,0}},
/* 36310 */ {(12<<2)|2,{98,55,0}},
/* 36311 */ {(12<<2)|2,{98,55,0}},
/* 36312 */ {(12<<2)|2,{98,55,0}},
/* 36313 */ {(12<<2)|2,{98,55,0}},
/* 36314 */ {(12<<2)|2,{98,55,0}},
/* 36315 */ {(12<<2)|2,{98,55,0}},
/* 36316 */ {(12<<2)|2,{98,55,0}},
/* 36317 */ {(12<<2)|2,{98,55,0}},
/* 36318 */ {(12<<2)|2,{98,55,0}},
/* 36319 */ {(12<<2)|2,{98,55,0}},
/* 36320 */ {(12<<2)|2,{98,56,0}},
/* 36321 */ {(12<<2)|2,{98,56,0}},
/* 36322 */ {(12<<2)|2,{98,56,0}},
/* 36323 */ {(12<<2)|2,{98,56,0}},
/* 36324 */ {(12<<2)|2,{98,56,0}},
/* 36325 */ {(12<<2)|2,{98,56,0}},
/* 36326 */ {(12<<2)|2,{98,56,0}},
/* 36327 */ {(12<<2)|2,{98,56,0}},
/* 36328 */ {(12<<2)|2,{98,56,0}},
/* 36329 */ {(12<<2)|2,{98,56,0}},
/* 36330 */ {(12<<2)|2,{98,56,0}},
/* 36331 */ {(12<<2)|2,{98,56,0}},
/* 36332 */ {(12<<2)|2,{98,56,0}},
/* 36333 */ {(12<<2)|2,{98,56,0}},
/* 36334 */ {(12<<2)|2,{98,56,0}},
/* 36335 */ {(12<<2)|2,{98,56,0}},
/* 36336 */ {(12<<2)|2,{98,57,0}},
/* 36337 */ {(12<<2)|2,{98,57,0}},
/* 36338 */ {(12<<2)|2,{98,57,0}},
/* 36339 */ {(12<<2)|2,{98,57,0}},
/* 36340 */ {(12<<2)|2,{98,57,0}},
/* 36341 */ {(12<<2)|2,{98,57,0}},
/* 36342 */ {(12<<2)|2,{98,57,0}},
/* 36343 */ {(12<<2)|2,{98,57,0}},
/* 36344 */ {(12<<2)|2,{98,57,0}},
/* 36345 */ {(12<<2)|2,{98,57,0}},
/* 36346 */ {(12<<2)|2,{98,57,0}},
/* 36347 */ {(12<<2)|2,{98,57,0}},
/* 36348 */ {(12<<2)|2,{98,57,0}},
/* 36349 */ {(12<<2)|2,{98,57,0}},
/* 36350 */ {(12<<2)|2,{98,57,0}},
/* 36351 */ {(12<<2)|2,{98,57,0}},
/* 36352 */ {(12<<2)|2,{98,61,0}},
/* 36353 */ {(12<<2)|2,{98,61,0}},
/* 36354 */ {(12<<2)|2,{98,61,0}},
/* 36355 */ {(12<<2)|2,{98,61,0}},
/* 36356 */ {(12<<2)|2,{98,61,0}},
/* 36357 */ {(12<<2)|2,{98,61,0}},
/* 36358 */ {(12<<2)|2,{98,61,0}},
/* 36359 */ {(12<<2)|2,{98,61,0}},
/* 36360 */ {(12<<2)|2,{98,61,0}},
/* 36361 */ {(12<<2)|2,{98,61,0}},
/* 36362 */ {(12<<2)|2,{98,61,0}},
/* 36363 */ {(12<<2)|2,{98,61,0}},
/* 36364 */ {(12<<2)|2,{98,61,0}},
/* 36365 */ {(12<<2)|2,{98,61,0}},
/* 36366 */ {(12<<2)|2,{98,61,0}},
/* 36367 */ {(12<<2)|2,{98,61,0}},
/* 36368 */ {(12<<2)|2,{98,65,0}},
/* 36369 */ {(12<<2)|2,{98,65,0}},
/* 36370 */ {(12<<2)|2,{98,65,0}},
/* 36371 */ {(12<<2)|2,{98,65,0}},
/* 36372 */ {(12<<2)|2,{98,65,0}},
/* 36373 */ {(12<<2)|2,{98,65,0}},
/* 36374 */ {(12<<2)|2,{98,65,0}},
/* 36375 */ {(12<<2)|2,{98,65,0}},
/* 36376 */ {(12<<2)|2,{98,65,0}},
/* 36377 */ {(12<<2)|2,{98,65,0}},
/* 36378 */ {(12<<2)|2,{98,65,0}},
/* 36379 */ {(12<<2)|2,{98,65,0}},
/* 36380 */ {(12<<2)|2,{98,65,0}},
/* 36381 */ {(12<<2)|2,{98,65,0}},
/* 36382 */ {(12<<2)|2,{98,65,0}},
/* 36383 */ {(12<<2)|2,{98,65,0}},
/* 36384 */ {(12<<2)|2,{98,95,0}},
/* 36385 */ {(12<<2)|2,{98,95,0}},
/* 36386 */ {(12<<2)|2,{98,95,0}},
/* 36387 */ {(12<<2)|2,{98,95,0}},
/* 36388 */ {(12<<2)|2,{98,95,0}},
/* 36389 */ {(12<<2)|2,{98,95,0}},
/* 36390 */ {(12<<2)|2,{98,95,0}},
/* 36391 */ {(12<<2)|2,{98,95,0}},
/* 36392 */ {(12<<2)|2,{98,95,0}},
/* 36393 */ {(12<<2)|2,{98,95,0}},
/* 36394 */ {(12<<2)|2,{98,95,0}},
/* 36395 */ {(12<<2)|2,{98,95,0}},
/* 36396 */ {(12<<2)|2,{98,95,0}},
/* 36397 */ {(12<<2)|2,{98,95,0}},
/* 36398 */ {(12<<2)|2,{98,95,0}},
/* 36399 */ {(12<<2)|2,{98,95,0}},
/* 36400 */ {(12<<2)|2,{98,98,0}},
/* 36401 */ {(12<<2)|2,{98,98,0}},
/* 36402 */ {(12<<2)|2,{98,98,0}},
/* 36403 */ {(12<<2)|2,{98,98,0}},
/* 36404 */ {(12<<2)|2,{98,98,0}},
/* 36405 */ {(12<<2)|2,{98,98,0}},
/* 36406 */ {(12<<2)|2,{98,98,0}},
/* 36407 */ {(12<<2)|2,{98,98,0}},
/* 36408 */ {(12<<2)|2,{98,98,0}},
/* 36409 */ {(12<<2)|2,{98,98,0}},
/* 36410 */ {(12<<2)|2,{98,98,0}},
/* 36411 */ {(12<<2)|2,{98,98,0}},
/* 36412 */ {(12<<2)|2,{98,98,0}},
/* 36413 */ {(12<<2)|2,{98,98,0}},
/* 36414 */ {(12<<2)|2,{98,98,0}},
/* 36415 */ {(12<<2)|2,{98,98,0}},
/* 36416 */ {(12<<2)|2,{98,100,0}},
/* 36417 */ {(12<<2)|2,{98,100,0}},
/* 36418 */ {(12<<2)|2,{98,100,0}},
/* 36419 */ {(12<<2)|2,{98,100,0}},
/* 36420 */ {(12<<2)|2,{98,100,0}},
/* 36421 */ {(12<<2)|2,{98,100,0}},
/* 36422 */ {(12<<2)|2,{98,100,0}},
/* 36423 */ {(12<<2)|2,{98,100,0}},
/* 36424 */ {(12<<2)|2,{98,100,0}},
/* 36425 */ {(12<<2)|2,{98,100,0}},
/* 36426 */ {(12<<2)|2,{98,100,0}},
/* 36427 */ {(12<<2)|2,{98,100,0}},
/* 36428 */ {(12<<2)|2,{98,100,0}},
/* 36429 */ {(12<<2)|2,{98,100,0}},
/* 36430 */ {(12<<2)|2,{98,100,0}},
/* 36431 */ {(12<<2)|2,{98,100,0}},
/* 36432 */ {(12<<2)|2,{98,102,0}},
/* 36433 */ {(12<<2)|2,{98,102,0}},
/* 36434 */ {(12<<2)|2,{98,102,0}},
/* 36435 */ {(12<<2)|2,{98,102,0}},
/* 36436 */ {(12<<2)|2,{98,102,0}},
/* 36437 */ {(12<<2)|2,{98,102,0}},
/* 36438 */ {(12<<2)|2,{98,102,0}},
/* 36439 */ {(12<<2)|2,{98,102,0}},
/* 36440 */ {(12<<2)|2,{98,102,0}},
/* 36441 */ {(12<<2)|2,{98,102,0}},
/* 36442 */ {(12<<2)|2,{98,102,0}},
/* 36443 */ {(12<<2)|2,{98,102,0}},
/* 36444 */ {(12<<2)|2,{98,102,0}},
/* 36445 */ {(12<<2)|2,{98,102,0}},
/* 36446 */ {(12<<2)|2,{98,102,0}},
/* 36447 */ {(12<<2)|2,{98,102,0}},
/* 36448 */ {(12<<2)|2,{98,103,0}},
/* 36449 */ {(12<<2)|2,{98,103,0}},
/* 36450 */ {(12<<2)|2,{98,103,0}},
/* 36451 */ {(12<<2)|2,{98,103,0}},
/* 36452 */ {(12<<2)|2,{98,103,0}},
/* 36453 */ {(12<<2)|2,{98,103,0}},
/* 36454 */ {(12<<2)|2,{98,103,0}},
/* 36455 */ {(12<<2)|2,{98,103,0}},
/* 36456 */ {(12<<2)|2,{98,103,0}},
/* 36457 */ {(12<<2)|2,{98,103,0}},
/* 36458 */ {(12<<2)|2,{98,103,0}},
/* 36459 */ {(12<<2)|2,{98,103,0}},
/* 36460 */ {(12<<2)|2,{98,103,0}},
/* 36461 */ {(12<<2)|2,{98,103,0}},
/* 36462 */ {(12<<2)|2,{98,103,0}},
/* 36463 */ {(12<<2)|2,{98,103,0}},
/* 36464 */ {(12<<2)|2,{98,104,0}},
/* 36465 */ {(12<<2)|2,{98,104,0}},
/* 36466 */ {(12<<2)|2,{98,104,0}},
/* 36467 */ {(12<<2)|2,{98,104,0}},
/* 36468 */ {(12<<2)|2,{98,104,0}},
/* 36469 */ {(12<<2)|2,{98,104,0}},
/* 36470 */ {(12<<2)|2,{98,104,0}},
/* 36471 */ {(12<<2)|2,{98,104,0}},
/* 36472 */ {(12<<2)|2,{98,104,0}},
/* 36473 */ {(12<<2)|2,{98,104,0}},
/* 36474 */ {(12<<2)|2,{98,104,0}},
/* 36475 */ {(12<<2)|2,{98,104,0}},
/* 36476 */ {(12<<2)|2,{98,104,0}},
/* 36477 */ {(12<<2)|2,{98,104,0}},
/* 36478 */ {(12<<2)|2,{98,104,0}},
/* 36479 */ {(12<<2)|2,{98,104,0}},
/* 36480 */ {(12<<2)|2,{98,108,0}},
/* 36481 */ {(12<<2)|2,{98,108,0}},
/* 36482 */ {(12<<2)|2,{98,108,0}},
/* 36483 */ {(12<<2)|2,{98,108,0}},
/* 36484 */ {(12<<2)|2,{98,108,0}},
/* 36485 */ {(12<<2)|2,{98,108,0}},
/* 36486 */ {(12<<2)|2,{98,108,0}},
/* 36487 */ {(12<<2)|2,{98,108,0}},
/* 36488 */ {(12<<2)|2,{98,108,0}},
/* 36489 */ {(12<<2)|2,{98,108,0}},
/* 36490 */ {(12<<2)|2,{98,108,0}},
/* 36491 */ {(12<<2)|2,{98,108,0}},
/* 36492 */ {(12<<2)|2,{98,108,0}},
/* 36493 */ {(12<<2)|2,{98,108,0}},
/* 36494 */ {(12<<2)|2,{98,108,0}},
/* 36495 */ {(12<<2)|2,{98,108,0}},
/* 36496 */ {(12<<2)|2,{98,109,0}},
/* 36497 */ {(12<<2)|2,{98,109,0}},
/* 36498 */ {(12<<2)|2,{98,109,0}},
/* 36499 */ {(12<<2)|2,{98,109,0}},
/* 36500 */ {(12<<2)|2,{98,109,0}},
/* 36501 */ {(12<<2)|2,{98,109,0}},
/* 36502 */ {(12<<2)|2,{98,109,0}},
/* 36503 */ {(12<<2)|2,{98,109,0}},
/* 36504 */ {(12<<2)|2,{98,109,0}},
/* 36505 */ {(12<<2)|2,{98,109,0}},
/* 36506 */ {(12<<2)|2,{98,109,0}},
/* 36507 */ {(12<<2)|2,{98,109,0}},
/* 36508 */ {(12<<2)|2,{98,109,0}},
/* 36509 */ {(12<<2)|2,{98,109,0}},
/* 36510 */ {(12<<2)|2,{98,109,0}},
/* 36511 */ {(12<<2)|2,{98,109,0}},
/* 36512 */ {(12<<2)|2,{98,110,0}},
/* 36513 */ {(12<<2)|2,{98,110,0}},
/* 36514 */ {(12<<2)|2,{98,110,0}},
/* 36515 */ {(12<<2)|2,{98,110,0}},
/* 36516 */ {(12<<2)|2,{98,110,0}},
/* 36517 */ {(12<<2)|2,{98,110,0}},
/* 36518 */ {(12<<2)|2,{98,110,0}},
/* 36519 */ {(12<<2)|2,{98,110,0}},
/* 36520 */ {(12<<2)|2,{98,110,0}},
/* 36521 */ {(12<<2)|2,{98,110,0}},
/* 36522 */ {(12<<2)|2,{98,110,0}},
/* 36523 */ {(12<<2)|2,{98,110,0}},
/* 36524 */ {(12<<2)|2,{98,110,0}},
/* 36525 */ {(12<<2)|2,{98,110,0}},
/* 36526 */ {(12<<2)|2,{98,110,0}},
/* 36527 */ {(12<<2)|2,{98,110,0}},
/* 36528 */ {(12<<2)|2,{98,112,0}},
/* 36529 */ {(12<<2)|2,{98,112,0}},
/* 36530 */ {(12<<2)|2,{98,112,0}},
/* 36531 */ {(12<<2)|2,{98,112,0}},
/* 36532 */ {(12<<2)|2,{98,112,0}},
/* 36533 */ {(12<<2)|2,{98,112,0}},
/* 36534 */ {(12<<2)|2,{98,112,0}},
/* 36535 */ {(12<<2)|2,{98,112,0}},
/* 36536 */ {(12<<2)|2,{98,112,0}},
/* 36537 */ {(12<<2)|2,{98,112,0}},
/* 36538 */ {(12<<2)|2,{98,112,0}},
/* 36539 */ {(12<<2)|2,{98,112,0}},
/* 36540 */ {(12<<2)|2,{98,112,0}},
/* 36541 */ {(12<<2)|2,{98,112,0}},
/* 36542 */ {(12<<2)|2,{98,112,0}},
/* 36543 */ {(12<<2)|2,{98,112,0}},
/* 36544 */ {(12<<2)|2,{98,114,0}},
/* 36545 */ {(12<<2)|2,{98,114,0}},
/* 36546 */ {(12<<2)|2,{98,114,0}},
/* 36547 */ {(12<<2)|2,{98,114,0}},
/* 36548 */ {(12<<2)|2,{98,114,0}},
/* 36549 */ {(12<<2)|2,{98,114,0}},
/* 36550 */ {(12<<2)|2,{98,114,0}},
/* 36551 */ {(12<<2)|2,{98,114,0}},
/* 36552 */ {(12<<2)|2,{98,114,0}},
/* 36553 */ {(12<<2)|2,{98,114,0}},
/* 36554 */ {(12<<2)|2,{98,114,0}},
/* 36555 */ {(12<<2)|2,{98,114,0}},
/* 36556 */ {(12<<2)|2,{98,114,0}},
/* 36557 */ {(12<<2)|2,{98,114,0}},
/* 36558 */ {(12<<2)|2,{98,114,0}},
/* 36559 */ {(12<<2)|2,{98,114,0}},
/* 36560 */ {(12<<2)|2,{98,117,0}},
/* 36561 */ {(12<<2)|2,{98,117,0}},
/* 36562 */ {(12<<2)|2,{98,117,0}},
/* 36563 */ {(12<<2)|2,{98,117,0}},
/* 36564 */ {(12<<2)|2,{98,117,0}},
/* 36565 */ {(12<<2)|2,{98,117,0}},
/* 36566 */ {(12<<2)|2,{98,117,0}},
/* 36567 */ {(12<<2)|2,{98,117,0}},
/* 36568 */ {(12<<2)|2,{98,117,0}},
/* 36569 */ {(12<<2)|2,{98,117,0}},
/* 36570 */ {(12<<2)|2,{98,117,0}},
/* 36571 */ {(12<<2)|2,{98,117,0}},
/* 36572 */ {(12<<2)|2,{98,117,0}},
/* 36573 */ {(12<<2)|2,{98,117,0}},
/* 36574 */ {(12<<2)|2,{98,117,0}},
/* 36575 */ {(12<<2)|2,{98,117,0}},
/* 36576 */ {(13<<2)|2,{98,58,0}},
/* 36577 */ {(13<<2)|2,{98,58,0}},
/* 36578 */ {(13<<2)|2,{98,58,0}},
/* 36579 */ {(13<<2)|2,{98,58,0}},
/* 36580 */ {(13<<2)|2,{98,58,0}},
/* 36581 */ {(13<<2)|2,{98,58,0}},
/* 36582 */ {(13<<2)|2,{98,58,0}},
/* 36583 */ {(13<<2)|2,{98,58,0}},
/* 36584 */ {(13<<2)|2,{98,66,0}},
/* 36585 */ {(13<<2)|2,{98,66,0}},
/* 36586 */ {(13<<2)|2,{98,66,0}},
/* 36587 */ {(13<<2)|2,{98,66,0}},
/* 36588 */ {(13<<2)|2,{98,66,0}},
/* 36589 */ {(13<<2)|2,{98,66,0}},
/* 36590 */ {(13<<2)|2,{98,66,0}},
/* 36591 */ {(13<<2)|2,{98,66,0}},
/* 36592 */ {(13<<2)|2,{98,67,0}},
/* 36593 */ {(13<<2)|2,{98,67,0}},
/* 36594 */ {(13<<2)|2,{98,67,0}},
/* 36595 */ {(13<<2)|2,{98,67,0}},
/* 36596 */ {(13<<2)|2,{98,67,0}},
/* 36597 */ {(13<<2)|2,{98,67,0}},
/* 36598 */ {(13<<2)|2,{98,67,0}},
/* 36599 */ {(13<<2)|2,{98,67,0}},
/* 36600 */ {(13<<2)|2,{98,68,0}},
/* 36601 */ {(13<<2)|2,{98,68,0}},
/* 36602 */ {(13<<2)|2,{98,68,0}},
/* 36603 */ {(13<<2)|2,{98,68,0}},
/* 36604 */ {(13<<2)|2,{98,68,0}},
/* 36605 */ {(13<<2)|2,{98,68,0}},
/* 36606 */ {(13<<2)|2,{98,68,0}},
/* 36607 */ {(13<<2)|2,{98,68,0}},
/* 36608 */ {(13<<2)|2,{98,69,0}},
/* 36609 */ {(13<<2)|2,{98,69,0}},
/* 36610 */ {(13<<2)|2,{98,69,0}},
/* 36611 */ {(13<<2)|2,{98,69,0}},
/* 36612 */ {(13<<2)|2,{98,69,0}},
/* 36613 */ {(13<<2)|2,{98,69,0}},
/* 36614 */ {(13<<2)|2,{98,69,0}},
/* 36615 */ {(13<<2)|2,{98,69,0}},
/* 36616 */ {(13<<2)|2,{98,70,0}},
/* 36617 */ {(13<<2)|2,{98,70,0}},
/* 36618 */ {(13<<2)|2,{98,70,0}},
/* 36619 */ {(13<<2)|2,{98,70,0}},
/* 36620 */ {(13<<2)|2,{98,70,0}},
/* 36621 */ {(13<<2)|2,{98,70,0}},
/* 36622 */ {(13<<2)|2,{98,70,0}},
/* 36623 */ {(13<<2)|2,{98,70,0}},
/* 36624 */ {(13<<2)|2,{98,71,0}},
/* 36625 */ {(13<<2)|2,{98,71,0}},
/* 36626 */ {(13<<2)|2,{98,71,0}},
/* 36627 */ {(13<<2)|2,{98,71,0}},
/* 36628 */ {(13<<2)|2,{98,71,0}},
/* 36629 */ {(13<<2)|2,{98,71,0}},
/* 36630 */ {(13<<2)|2,{98,71,0}},
/* 36631 */ {(13<<2)|2,{98,71,0}},
/* 36632 */ {(13<<2)|2,{98,72,0}},
/* 36633 */ {(13<<2)|2,{98,72,0}},
/* 36634 */ {(13<<2)|2,{98,72,0}},
/* 36635 */ {(13<<2)|2,{98,72,0}},
/* 36636 */ {(13<<2)|2,{98,72,0}},
/* 36637 */ {(13<<2)|2,{98,72,0}},
/* 36638 */ {(13<<2)|2,{98,72,0}},
/* 36639 */ {(13<<2)|2,{98,72,0}},
/* 36640 */ {(13<<2)|2,{98,73,0}},
/* 36641 */ {(13<<2)|2,{98,73,0}},
/* 36642 */ {(13<<2)|2,{98,73,0}},
/* 36643 */ {(13<<2)|2,{98,73,0}},
/* 36644 */ {(13<<2)|2,{98,73,0}},
/* 36645 */ {(13<<2)|2,{98,73,0}},
/* 36646 */ {(13<<2)|2,{98,73,0}},
/* 36647 */ {(13<<2)|2,{98,73,0}},
/* 36648 */ {(13<<2)|2,{98,74,0}},
/* 36649 */ {(13<<2)|2,{98,74,0}},
/* 36650 */ {(13<<2)|2,{98,74,0}},
/* 36651 */ {(13<<2)|2,{98,74,0}},
/* 36652 */ {(13<<2)|2,{98,74,0}},
/* 36653 */ {(13<<2)|2,{98,74,0}},
/* 36654 */ {(13<<2)|2,{98,74,0}},
/* 36655 */ {(13<<2)|2,{98,74,0}},
/* 36656 */ {(13<<2)|2,{98,75,0}},
/* 36657 */ {(13<<2)|2,{98,75,0}},
/* 36658 */ {(13<<2)|2,{98,75,0}},
/* 36659 */ {(13<<2)|2,{98,75,0}},
/* 36660 */ {(13<<2)|2,{98,75,0}},
/* 36661 */ {(13<<2)|2,{98,75,0}},
/* 36662 */ {(13<<2)|2,{98,75,0}},
/* 36663 */ {(13<<2)|2,{98,75,0}},
/* 36664 */ {(13<<2)|2,{98,76,0}},
/* 36665 */ {(13<<2)|2,{98,76,0}},
/* 36666 */ {(13<<2)|2,{98,76,0}},
/* 36667 */ {(13<<2)|2,{98,76,0}},
/* 36668 */ {(13<<2)|2,{98,76,0}},
/* 36669 */ {(13<<2)|2,{98,76,0}},
/* 36670 */ {(13<<2)|2,{98,76,0}},
/* 36671 */ {(13<<2)|2,{98,76,0}},
/* 36672 */ {(13<<2)|2,{98,77,0}},
/* 36673 */ {(13<<2)|2,{98,77,0}},
/* 36674 */ {(13<<2)|2,{98,77,0}},
/* 36675 */ {(13<<2)|2,{98,77,0}},
/* 36676 */ {(13<<2)|2,{98,77,0}},
/* 36677 */ {(13<<2)|2,{98,77,0}},
/* 36678 */ {(13<<2)|2,{98,77,0}},
/* 36679 */ {(13<<2)|2,{98,77,0}},
/* 36680 */ {(13<<2)|2,{98,78,0}},
/* 36681 */ {(13<<2)|2,{98,78,0}},
/* 36682 */ {(13<<2)|2,{98,78,0}},
/* 36683 */ {(13<<2)|2,{98,78,0}},
/* 36684 */ {(13<<2)|2,{98,78,0}},
/* 36685 */ {(13<<2)|2,{98,78,0}},
/* 36686 */ {(13<<2)|2,{98,78,0}},
/* 36687 */ {(13<<2)|2,{98,78,0}},
/* 36688 */ {(13<<2)|2,{98,79,0}},
/* 36689 */ {(13<<2)|2,{98,79,0}},
/* 36690 */ {(13<<2)|2,{98,79,0}},
/* 36691 */ {(13<<2)|2,{98,79,0}},
/* 36692 */ {(13<<2)|2,{98,79,0}},
/* 36693 */ {(13<<2)|2,{98,79,0}},
/* 36694 */ {(13<<2)|2,{98,79,0}},
/* 36695 */ {(13<<2)|2,{98,79,0}},
/* 36696 */ {(13<<2)|2,{98,80,0}},
/* 36697 */ {(13<<2)|2,{98,80,0}},
/* 36698 */ {(13<<2)|2,{98,80,0}},
/* 36699 */ {(13<<2)|2,{98,80,0}},
/* 36700 */ {(13<<2)|2,{98,80,0}},
/* 36701 */ {(13<<2)|2,{98,80,0}},
/* 36702 */ {(13<<2)|2,{98,80,0}},
/* 36703 */ {(13<<2)|2,{98,80,0}},
/* 36704 */ {(13<<2)|2,{98,81,0}},
/* 36705 */ {(13<<2)|2,{98,81,0}},
/* 36706 */ {(13<<2)|2,{98,81,0}},
/* 36707 */ {(13<<2)|2,{98,81,0}},
/* 36708 */ {(13<<2)|2,{98,81,0}},
/* 36709 */ {(13<<2)|2,{98,81,0}},
/* 36710 */ {(13<<2)|2,{98,81,0}},
/* 36711 */ {(13<<2)|2,{98,81,0}},
/* 36712 */ {(13<<2)|2,{98,82,0}},
/* 36713 */ {(13<<2)|2,{98,82,0}},
/* 36714 */ {(13<<2)|2,{98,82,0}},
/* 36715 */ {(13<<2)|2,{98,82,0}},
/* 36716 */ {(13<<2)|2,{98,82,0}},
/* 36717 */ {(13<<2)|2,{98,82,0}},
/* 36718 */ {(13<<2)|2,{98,82,0}},
/* 36719 */ {(13<<2)|2,{98,82,0}},
/* 36720 */ {(13<<2)|2,{98,83,0}},
/* 36721 */ {(13<<2)|2,{98,83,0}},
/* 36722 */ {(13<<2)|2,{98,83,0}},
/* 36723 */ {(13<<2)|2,{98,83,0}},
/* 36724 */ {(13<<2)|2,{98,83,0}},
/* 36725 */ {(13<<2)|2,{98,83,0}},
/* 36726 */ {(13<<2)|2,{98,83,0}},
/* 36727 */ {(13<<2)|2,{98,83,0}},
/* 36728 */ {(13<<2)|2,{98,84,0}},
/* 36729 */ {(13<<2)|2,{98,84,0}},
/* 36730 */ {(13<<2)|2,{98,84,0}},
/* 36731 */ {(13<<2)|2,{98,84,0}},
/* 36732 */ {(13<<2)|2,{98,84,0}},
/* 36733 */ {(13<<2)|2,{98,84,0}},
/* 36734 */ {(13<<2)|2,{98,84,0}},
/* 36735 */ {(13<<2)|2,{98,84,0}},
/* 36736 */ {(13<<2)|2,{98,85,0}},
/* 36737 */ {(13<<2)|2,{98,85,0}},
/* 36738 */ {(13<<2)|2,{98,85,0}},
/* 36739 */ {(13<<2)|2,{98,85,0}},
/* 36740 */ {(13<<2)|2,{98,85,0}},
/* 36741 */ {(13<<2)|2,{98,85,0}},
/* 36742 */ {(13<<2)|2,{98,85,0}},
/* 36743 */ {(13<<2)|2,{98,85,0}},
/* 36744 */ {(13<<2)|2,{98,86,0}},
/* 36745 */ {(13<<2)|2,{98,86,0}},
/* 36746 */ {(13<<2)|2,{98,86,0}},
/* 36747 */ {(13<<2)|2,{98,86,0}},
/* 36748 */ {(13<<2)|2,{98,86,0}},
/* 36749 */ {(13<<2)|2,{98,86,0}},
/* 36750 */ {(13<<2)|2,{98,86,0}},
/* 36751 */ {(13<<2)|2,{98,86,0}},
/* 36752 */ {(13<<2)|2,{98,87,0}},
/* 36753 */ {(13<<2)|2,{98,87,0}},
/* 36754 */ {(13<<2)|2,{98,87,0}},
/* 36755 */ {(13<<2)|2,{98,87,0}},
/* 36756 */ {(13<<2)|2,{98,87,0}},
/* 36757 */ {(13<<2)|2,{98,87,0}},
/* 36758 */ {(13<<2)|2,{98,87,0}},
/* 36759 */ {(13<<2)|2,{98,87,0}},
/* 36760 */ {(13<<2)|2,{98,89,0}},
/* 36761 */ {(13<<2)|2,{98,89,0}},
/* 36762 */ {(13<<2)|2,{98,89,0}},
/* 36763 */ {(13<<2)|2,{98,89,0}},
/* 36764 */ {(13<<2)|2,{98,89,0}},
/* 36765 */ {(13<<2)|2,{98,89,0}},
/* 36766 */ {(13<<2)|2,{98,89,0}},
/* 36767 */ {(13<<2)|2,{98,89,0}},
/* 36768 */ {(13<<2)|2,{98,106,0}},
/* 36769 */ {(13<<2)|2,{98,106,0}},
/* 36770 */ {(13<<2)|2,{98,106,0}},
/* 36771 */ {(13<<2)|2,{98,106,0}},
/* 36772 */ {(13<<2)|2,{98,106,0}},
/* 36773 */ {(13<<2)|2,{98,106,0}},
/* 36774 */ {(13<<2)|2,{98,106,0}},
/* 36775 */ {(13<<2)|2,{98,106,0}},
/* 36776 */ {(13<<2)|2,{98,107,0}},
/* 36777 */ {(13<<2)|2,{98,107,0}},
/* 36778 */ {(13<<2)|2,{98,107,0}},
/* 36779 */ {(13<<2)|2,{98,107,0}},
/* 36780 */ {(13<<2)|2,{98,107,0}},
/* 36781 */ {(13<<2)|2,{98,107,0}},
/* 36782 */ {(13<<2)|2,{98,107,0}},
/* 36783 */ {(13<<2)|2,{98,107,0}},
/* 36784 */ {(13<<2)|2,{98,113,0}},
/* 36785 */ {(13<<2)|2,{98,113,0}},
/* 36786 */ {(13<<2)|2,{98,113,0}},
/* 36787 */ {(13<<2)|2,{98,113,0}},
/* 36788 */ {(13<<2)|2,{98,113,0}},
/* 36789 */ {(13<<2)|2,{98,113,0}},
/* 36790 */ {(13<<2)|2,{98,113,0}},
/* 36791 */ {(13<<2)|2,{98,113,0}},
/* 36792 */ {(13<<2)|2,{98,118,0}},
/* 36793 */ {(13<<2)|2,{98,118,0}},
/* 36794 */ {(13<<2)|2,{98,118,0}},
/* 36795 */ {(13<<2)|2,{98,118,0}},
/* 36796 */ {(13<<2)|2,{98,118,0}},
/* 36797 */ {(13<<2)|2,{98,118,0}},
/* 36798 */ {(13<<2)|2,{98,118,0}},
/* 36799 */ {(13<<2)|2,{98,118,0}},
/* 36800 */ {(13<<2)|2,{98,119,0}},
/* 36801 */ {(13<<2)|2,{98,119,0}},
/* 36802 */ {(13<<2)|2,{98,119,0}},
/* 36803 */ {(13<<2)|2,{98,119,0}},
/* 36804 */ {(13<<2)|2,{98,119,0}},
/* 36805 */ {(13<<2)|2,{98,119,0}},
/* 36806 */ {(13<<2)|2,{98,119,0}},
/* 36807 */ {(13<<2)|2,{98,119,0}},
/* 36808 */ {(13<<2)|2,{98,120,0}},
/* 36809 */ {(13<<2)|2,{98,120,0}},
/* 36810 */ {(13<<2)|2,{98,120,0}},
/* 36811 */ {(13<<2)|2,{98,120,0}},
/* 36812 */ {(13<<2)|2,{98,120,0}},
/* 36813 */ {(13<<2)|2,{98,120,0}},
/* 36814 */ {(13<<2)|2,{98,120,0}},
/* 36815 */ {(13<<2)|2,{98,120,0}},
/* 36816 */ {(13<<2)|2,{98,121,0}},
/* 36817 */ {(13<<2)|2,{98,121,0}},
/* 36818 */ {(13<<2)|2,{98,121,0}},
/* 36819 */ {(13<<2)|2,{98,121,0}},
/* 36820 */ {(13<<2)|2,{98,121,0}},
/* 36821 */ {(13<<2)|2,{98,121,0}},
/* 36822 */ {(13<<2)|2,{98,121,0}},
/* 36823 */ {(13<<2)|2,{98,121,0}},
/* 36824 */ {(13<<2)|2,{98,122,0}},
/* 36825 */ {(13<<2)|2,{98,122,0}},
/* 36826 */ {(13<<2)|2,{98,122,0}},
/* 36827 */ {(13<<2)|2,{98,122,0}},
/* 36828 */ {(13<<2)|2,{98,122,0}},
/* 36829 */ {(13<<2)|2,{98,122,0}},
/* 36830 */ {(13<<2)|2,{98,122,0}},
/* 36831 */ {(13<<2)|2,{98,122,0}},
/* 36832 */ {(14<<2)|2,{98,38,0}},
/* 36833 */ {(14<<2)|2,{98,38,0}},
/* 36834 */ {(14<<2)|2,{98,38,0}},
/* 36835 */ {(14<<2)|2,{98,38,0}},
/* 36836 */ {(14<<2)|2,{98,42,0}},
/* 36837 */ {(14<<2)|2,{98,42,0}},
/* 36838 */ {(14<<2)|2,{98,42,0}},
/* 36839 */ {(14<<2)|2,{98,42,0}},
/* 36840 */ {(14<<2)|2,{98,44,0}},
/* 36841 */ {(14<<2)|2,{98,44,0}},
/* 36842 */ {(14<<2)|2,{98,44,0}},
/* 36843 */ {(14<<2)|2,{98,44,0}},
/* 36844 */ {(14<<2)|2,{98,59,0}},
/* 36845 */ {(14<<2)|2,{98,59,0}},
/* 36846 */ {(14<<2)|2,{98,59,0}},
/* 36847 */ {(14<<2)|2,{98,59,0}},
/* 36848 */ {(14<<2)|2,{98,88,0}},
/* 36849 */ {(14<<2)|2,{98,88,0}},
/* 36850 */ {(14<<2)|2,{98,88,0}},
/* 36851 */ {(14<<2)|2,{98,88,0}},
/* 36852 */ {(14<<2)|2,{98,90,0}},
/* 36853 */ {(14<<2)|2,{98,90,0}},
/* 36854 */ {(14<<2)|2,{98,90,0}},
/* 36855 */ {(14<<2)|2,{98,90,0}},
/* 36856 */ {(16<<2)|2,{98,33,0}},
/* 36857 */ {(16<<2)|2,{98,34,0}},
/* 36858 */ {(16<<2)|2,{98,40,0}},
/* 36859 */ {(16<<2)|2,{98,41,0}},
/* 36860 */ {(16<<2)|2,{98,63,0}},
/* 36861 */ {(6<<2)|1,{98,0,0}},
/* 36862 */ {(6<<2)|1,{98,0,0}},
/* 36863 */ {(6<<2)|1,{98,0,0}},
/* 36864 */ {(16<<2)|3,{100,48,48}},
/* 36865 */ {(16<<2)|3,{100,48,49}},
/* 36866 */ {(16<<2)|3,{100,48,50}},
/* 36867 */ {(16<<2)|3,{100,48,97}},
/* 36868 */ {(16<<2)|3,{100,48,99}},
/* 36869 */ {(16<<2)|3,{100,48,101}},
/* 36870 */ {(16<<2)|3,{100,48,105}},
/* 36871 */ {(16<<2)|3,{100,48,111}},
/* 36872 */ {(16<<2)|3,{100,48,115}},
/* 36873 */ {(16<<2)|3,{100,48,116}},
/* 36874 */ {(11<<2)|2,{100,48,0}},
/* 36875 */ {(11<<2)|2,{100,48,0}},
/* 36876 */ {(11<<2)|2,{100,48,0}},
/* 36877 */ {(11<<2)|2,{100,48,0}},
/* 36878 */ {(11<<2)|2,{100,48,0}},
/* 36879 */ {(11<<2)|2,{100,48,0}},
/* 36880 */ {(11<<2)|2,{100,48,0}},
/* 36881 */ {(11<<2)|2,{100,48,0}},
/* 36882 */ {(11<<2)|2,{100,48,0}},
/* 36883 */ {(11<<2)|2,{100,48,0}},
/* 36884 */ {(11<<2)|2,{100,48,0}},
/* 36885 */ {(11<<2)|2,{100,48,0}},
/* 36886 */ {(11<<2)|2,{100,48,0}},
/* 36887 */ {(11<<2)|2,{100,48,0}},
/* 36888 */ {(11<<2)|2,{100,48,0}},
/* 36889 */ {(11<<2)|2,{100,48,0}},
/* 36890 */ {(11<<2)|2,{100,48,0}},
/* 36891 */ {(11<<2)|2,{100,48,0}},
/* 36892 */ {(11<<2)|2,{100,48,0}},
/* 36893 */ {(11<<2)|2,{100,48,0}},
/* 36894 */ {(11<<2)|2,{100,48,0}},
/* 36895 */ {(11<<2)|2,{100,48,0}},
/* 36896 */ {(16<<2)|3,{100,49,48}},
/* 36897 */ {(16<<2)|3,{100,49,49}},
/* 36898 */ {(16<<2)|3,{100,49,50}},
/* 36899 */ {(16<<2)|3,{100,49,97}},
/* 36900 */ {(16<<2)|3,{100,49,99}},
/* 36901 */ {(16<<2)|3,{100,49,101}},
/* 36902 */ {(16<<2)|3,{100,49,105}},
/* 36903 */ {(16<<2)|3,{100,49,111}},
/* 36904 */ {(16<<2)|3,{100,49,115}},
/* 36905 */ {(16<<2)|3,{100,49,116}},
/* 36906 */ {(11<<2)|2,{100,49,0}},
/* 36907 */ {(11<<2)|2,{100,49,0}},
/* 36908 */ {(11<<2)|2,{100,49,0}},
/* 36909 */ {(11<<2)|2,{100,49,0}},
/* 36910 */ {(11<<2)|2,{100,49,0}},
/* 36911 */ {(11<<2)|2,{100,49,0}},
/* 36912 */ {(11<<2)|2,{100,49,0}},
/* 36913 */ {(11<<2)|2,{100,49,0}},
/* 36914 */ {(11<<2)|2,{100,49,0}},
/* 36915 */ {(11<<2)|2,{100,49,0}},
/* 36916 */ {(11<<2)|2,{100,49,0}},
/* 36917 */ {(11<<2)|2,{100,49,0}},
/* 36918 */ {(11<<2)|2,{100,49,0}},
/* 36919 */ {(11<<2)|2,{100,49,0}},
/* 36920 */ {(11<<2)|2,{100,49,0}},
/* 36921 */ {(11<<2)|2,{100,49,0}},
/* 36922 */ {(11<<2)|2,{100,49,0}},
/* 36923 */ {(11<<2)|2,{100,49,0}},
/* 36924 */ {(11<<2)|2,{100,49,0}},
/* 36925 */ {(11<<2)|2,{100,49,0}},
/* 36926 */ {(11<<2)|2,{100,49,0}},
/* 36927 */ {(11<<2)|2,{100,49,0}},
/* 36928 */ {(16<<2)|3,{100,50,48}},
/* 36929 */ {(16<<2)|3,{100,50,49}},
/* 36930 */ {(16<<2)|3,{100,50,50}},
/* 36931 */ {(16<<2)|3,{100,50,97}},
/* 36932 */ {(16<<2)|3,{100,50,99}},
/* 36933 */ {(16<<2)|3,{100,50,101}},
/* 36934 */ {(16<<2)|3,{100,50,105}},
/* 36935 */ {(16<<2)|3,{100,50,111}},
/* 36936 */ {(16<<2)|3,{100,50,115}},
/* 36937 */ {(16<<2)|3,{100,50,116}},
/* 36938 */ {(11<<2)|2,{100,50,0}},
/* 36939 */ {(11<<2)|2,{100,50,0}},
/* 36940 */ {(11<<2)|2,{100,50,0}},
/* 36941 */ {(11<<2)|2,{100,50,0}},
/* 36942 */ {(11<<2)|2,{100,50,0}},
/* 36943 */ {(11<<2)|2,{100,50,0}},
/* 36944 */ {(11<<2)|2,{100,50,0}},
/* 36945 */ {(11<<2)|2,{100,50,0}},
/* 36946 */ {(11<<2)|2,{100,50,0}},
/* 36947 */ {(11<<2)|2,{100,50,0}},
/* 36948 */ {(11<<2)|2,{100,50,0}},
/* 36949 */ {(11<<2)|2,{100,50,0}},
/* 36950 */ {(11<<2)|2,{100,50,0}},
/* 36951 */ {(11<<2)|2,{100,50,0}},
/* 36952 */ {(11<<2)|2,{100,50,0}},
/* 36953 */ {(11<<2)|2,{100,50,0}},
/* 36954 */ {(11<<2)|2,{100,50,0}},
/* 36955 */ {(11<<2)|2,{100,50,0}},
/* 36956 */ {(11<<2)|2,{100,50,0}},
/* 36957 */ {(11<<2)|2,{100,50,0}},
/* 36958 */ {(11<<2)|2,{100,50,0}},
/* 36959 */ {(11<<2)|2,{100,50,0}},
/* 36960 */ {(16<<2)|3,{100,97,48}},
/* 36961 */ {(16<<2)|3,{100,97,49}},
/* 36962 */ {(16<<2)|3,{100,97,50}},
/* 36963 */ {(16<<2)|3,{100,97,97}},
/* 36964 */ {(16<<2)|3,{100,97,99}},
/* 36965 */ {(16<<2)|3,{100,97,101}},
/* 36966 */ {(16<<2)|3,{100,97,105}},
/* 36967 */ {(16<<2)|3,{100,97,111}},
/* 36968 */ {(16<<2)|3,{100,97,115}},
/* 36969 */ {(16<<2)|3,{100,97,116}},
/* 36970 */ {(11<<2)|2,{100,97,0}},
/* 36971 */ {(11<<2)|2,{100,97,0}},
/* 36972 */ {(11<<2)|2,{100,97,0}},
/* 36973 */ {(11<<2)|2,{100,97,0}},
/* 36974 */ {(11<<2)|2,{100,97,0}},
/* 36975 */ {(11<<2)|2,{100,97,0}},
/* 36976 */ {(11<<2)|2,{100,97,0}},
/* 36977 */ {(11<<2)|2,{100,97,0}},
/* 36978 */ {(11<<2)|2,{100,97,0}},
/* 36979 */ {(11<<2)|2,{100,97,0}},
/* 36980 */ {(11<<2)|2,{100,97,0}},
/* 36981 */ {(11<<2)|2,{100,97,0}},
/* 36982 */ {(11<<2)|2,{100,97,0}},
/* 36983 */ {(11<<2)|2,{100,97,0}},
/* 36984 */ {(11<<2)|2,{100,97,0}},
/* 36985 */ {(11<<2)|2,{100,97,0}},
/* 36986 */ {(11<<2)|2,{100,97,0}},
/* 36987 */ {(11<<2)|2,{100,97,0}},
/* 36988 */ {(11<<2)|2,{100,97,0}},
/* 36989 */ {(11<<2)|2,{100,97,0}},
/* 36990 */ {(11<<2)|2,{100,97,0}},
/* 36991 */ {(11<<2)|2,{100,97,0}},
/* 36992 */ {(16<<2)|3,{100,99,48}},
/* 36993 */ {(16<<2)|3,{100,99,49}},
/* 36994 */ {(16<<2)|3,{100,99,50}},
/* 36995 */ {(16<<2)|3,{100,99,97}},
/* 36996 */ {(16<<2)|3,{100,99,99}},
/* 36997 */ {(16<<2)|3,{100,99,101}},
/* 36998 */ {(16<<2)|3,{100,99,105}},
/* 36999 */ {(16<<2)|3,{100,99,111}},
/* 37000 */ {(16<<2)|3,{100,99,115}},
/* 37001 */ {(16<<2)|3,{100,99,116}},
/* 37002 */ {(11<<2)|2,{100,99,0}},
/* 37003 */ {(11<<2)|2,{100,99,0}},
/* 37004 */ {(11<<2)|2,{100,99,0}},
/* 37005 */ {(11<<2)|2,{100,99,0}},
/* 37006 */ {(11<<2)|2,{100,99,0}},
/* 37007 */ {(11<<2)|2,{100,99,0}},
/* 37008 */ {(11<<2)|2,{100,99,0}},
/* 37009 */ {(11<<2)|2,{100,99,0}},
/* 37010 */ {(11<<2)|2,{100,99,0}},
/* 37011 */ {(11<<2)|2,{100,99,0}},
/* 37012 */ {(11<<2)|2,{100,99,0}},
/* 37013 */ {(11<<2)|2,{100,99,0}},
/* 37014 */ {(11<<2)|2,{100,99,0}},
/* 37015 */ {(11<<2)|2,{100,99,0}},
/* 37016 */ {(11<<2)|2,{100,99,0}},
/* 37017 */ {(11<<2)|2,{100,99,0}},
/* 37018 */ {(11<<2)|2,{100,99,0}},
/* 37019 */ {(11<<2)|2,{100,99,0}},
/* 37020 */ {(11<<2)|2,{100,99,0}},
/* 37021 */ {(11<<2)|2,{100,99,0}},
/* 37022 */ {(11<<2)|2,{100,99,0}},
/* 37023 */ {(11<<2)|2,{100,99,0}},
/* 37024 */ {(16<<2)|3,{100,101,48}},
/* 37025 */ {(16<<2)|3,{100,101,49}},
/* 37026 */ {(16<<2)|3,{100,101,50}},
/* 37027 */ {(16<<2)|3,{100,101,97}},
/* 37028 */ {(16<<2)|3,{100,101,99}},
/* 37029 */ {(16<<2)|3,{100,101,101}},
/* 37030 */ {(16<<2)|3,{100,101,105}},
/* 37031 */ {(16<<2)|3,{100,101,111}},
/* 37032 */ {(16<<2)|3,{100,101,115}},
/* 37033 */ {(16<<2)|3,{100,101,116}},
/* 37034 */ {(11<<2)|2,{100,101,0}},
/* 37035 */ {(11<<2)|2,{100,101,0}},
/* 37036 */ {(11<<2)|2,{100,101,0}},
/* 37037 */ {(11<<2)|2,{100,101,0}},
/* 37038 */ {(11<<2)|2,{100,101,0}},
/* 37039 */ {(11<<2)|2,{100,101,0}},
/* 37040 */ {(11<<2)|2,{100,101,0}},
/* 37041 */ {(11<<2)|2,{100,101,0}},
/* 37042 */ {(11<<2)|2,{100,101,0}},
/* 37043 */ {(11<<2)|2,{100,101,0}},
/* 37044 */ {(11<<2)|2,{100,101,0}},
/* 37045 */ {(11<<2)|2,{100,101,0}},
/* 37046 */ {(11<<2)|2,{100,101,0}},
/* 37047 */ {(11<<2)|2,{100,101,0}},
/* 37048 */ {(11<<2)|2,{100,101,0}},
/* 37049 */ {(11<<2)|2,{100,101,0}},
/* 37050 */ {(11<<2)|2,{100,101,0}},
/* 37051 */ {(11<<2)|2,{100,101,0}},
/* 37052 */ {(11<<2)|2,{100,101,0}},
/* 37053 */ {(11<<2)|2,{100,101,0}},
/* 37054 */ {(11<<2)|2,{100,101,0}},
/* 37055 */ {(11<<2)|2,{100,101,0}},
/* 37056 */ {(16<<2)|3,{100,105,48}},
/* 37057 */ {(16<<2)|3,{100,105,49}},
/* 37058 */ {(16<<2)|3,{100,105,50}},
/* 37059 */ {(16<<2)|3,{100,105,97}},
/* 37060 */ {(16<<2)|3,{100,105,99}},
/* 37061 */ {(16<<2)|3,{100,105,101}},
/* 37062 */ {(16<<2)|3,{100,105,105}},
/* 37063 */ {(16<<2)|3,{100,105,111}},
/* 37064 */ {(16<<2)|3,{100,105,115}},
/* 37065 */ {(16<<2)|3,{100,105,116}},
/* 37066 */ {(11<<2)|2,{100,105,0}},
/* 37067 */ {(11<<2)|2,{100,105,0}},
/* 37068 */ {(11<<2)|2,{100,105,0}},
/* 37069 */ {(11<<2)|2,{100,105,0}},
/* 37070 */ {(11<<2)|2,{100,105,0}},
/* 37071 */ {(11<<2)|2,{100,105,0}},
/* 37072 */ {(11<<2)|2,{100,105,0}},
/* 37073 */ {(11<<2)|2,{100,105,0}},
/* 37074 */ {(11<<2)|2,{100,105,0}},
/* 37075 */ {(11<<2)|2,{100,105,0}},
/* 37076 */ {(11<<2)|2,{100,105,0}},
/* 37077 */ {(11<<2)|2,{100,105,0}},
/* 37078 */ {(11<<2)|2,{100,105,0}},
/* 37079 */ {(11<<2)|2,{100,105,0}},
/* 37080 */ {(11<<2)|2,{100,105,0}},
/* 37081 */ {(11<<2)|2,{100,105,0}},
/* 37082 */ {(11<<2)|2,{100,105,0}},
/* 37083 */ {(11<<2)|2,{100,105,0}},
/* 37084 */ {(11<<2)|2,{100,105,0}},
/* 37085 */ {(11<<2)|2,{100,105,0}},
/* 37086 */ {(11<<2)|2,{100,105,0}},
/* 37087 */ {(11<<2)|2,{100,105,0}},
/* 37088 */ {(16<<2)|3,{100,111,48}},
/* 37089 */ {(16<<2)|3,{100,111,49}},
/* 37090 */ {(16<<2)|3,{100,111,50}},
/* 37091 */ {(16<<2)|3,{100,111,97}},
/* 37092 */ {(16<<2)|3,{100,111,99}},
/* 37093 */ {(16<<2)|3,{100,111,101}},
/* 37094 */ {(16<<2)|3,{100,111,105}},
/* 37095 */ {(16<<2)|3,{100,111,111}},
/* 37096 */ {(16<<2)|3,{100,111,115}},
/* 37097 */ {(16<<2)|3,{100,111,116}},
/* 37098 */ {(11<<2)|2,{100,111,0}},
/* 37099 */ {(11<<2)|2,{100,111,0}},
/* 37100 */ {(11<<2)|2,{100,111,0}},
/* 37101 */ {(11<<2)|2,{100,111,0}},
/* 37102 */ {(11<<2)|2,{100,111,0}},
/* 37103 */ {(11<<2)|2,{100,111,0}},
/* 37104 */ {(11<<2)|2,{100,111,0}},
/* 37105 */ {(11<<2)|2,{100,111,0}},
/* 37106 */ {(11<<2)|2,{100,111,0}},
/* 37107 */ {(11<<2)|2,{100,111,0}},
/* 37108 */ {(11<<2)|2,{100,111,0}},
/* 37109 */ {(11<<2)|2,{100,111,0}},
/* 37110 */ {(11<<2)|2,{100,111,0}},
/* 37111 */ {(11<<2)|2,{100,111,0}},
/* 37112 */ {(11<<2)|2,{100,111,0}},
/* 37113 */ {(11<<2)|2,{100,111,0}},
/* 37114 */ {(11<<2)|2,{100,111,0}},
/* 37115 */ {(11<<2)|2,{100,111,0}},
/* 37116 */ {(11<<2)|2,{100,111,0}},
/* 37117 */ {(11<<2)|2,{100,111,0}},
/* 37118 */ {(11<<2)|2,{100,111,0}},
/* 37119 */ {(11<<2)|2,{100,111,0}},
/* 37120 */ {(16<<2)|3,{100,115,48}},
/* 37121 */ {(16<<2)|3,{100,115,49}},
/* 37122 */ {(16<<2)|3,{100,115,50}},
/* 37123 */ {(16<<2)|3,{100,115,97}},
/* 37124 */ {(16<<2)|3,{100,115,99}},
/* 37125 */ {(16<<2)|3,{100,115,101}},
/* 37126 */ {(16<<2)|3,{100,115,105}},
/* 37127 */ {(16<<2)|3,{100,115,111}},
/* 37128 */ {(16<<2)|3,{100,115,115}},
/* 37129 */ {(16<<2)|3,{100,115,116}},
/* 37130 */ {(11<<2)|2,{100,115,0}},
/* 37131 */ {(11<<2)|2,{100,115,0}},
/* 37132 */ {(11<<2)|2,{100,115,0}},
/* 37133 */ {(11<<2)|2,{100,115,0}},
/* 37134 */ {(11<<2)|2,{100,115,0}},
/* 37135 */ {(11<<2)|2,{100,115,0}},
/* 37136 */ {(11<<2)|2,{100,115,0}},
/* 37137 */ {(11<<2)|2,{100,115,0}},
/* 37138 */ {(11<<2)|2,{100,115,0}},
/* 37139 */ {(11<<2)|2,{100,115,0}},
/* 37140 */ {(11<<2)|2,{100,115,0}},
/* 37141 */ {(11<<2)|2,{100,115,0}},
/* 37142 */ {(11<<2)|2,{100,115,0}},
/* 37143 */ {(11<<2)|2,{100,115,0}},
/* 37144 */ {(11<<2)|2,{100,115,0}},
/* 37145 */ {(11<<2)|2,{100,115,0}},
/* 37146 */ {(11<<2)|2,{100,115,0}},
/* 37147 */ {(11<<2)|2,{100,115,0}},
/* 37148 */ {(11<<2)|2,{100,115,0}},
/* 37149 */ {(11<<2)|2,{100,115,0}},
/* 37150 */ {(11<<2)|2,{100,115,0}},
/* 37151 */ {(11<<2)|2,{100,115,0}},
/* 37152 */ {(16<<2)|3,{100,116,48}},
/* 37153 */ {(16<<2)|3,{100,116,49}},
/* 37154 */ {(16<<2)|3,{100,116,50}},
/* 37155 */ {(16<<2)|3,{100,116,97}},
/* 37156 */ {(16<<2)|3,{100,116,99}},
/* 37157 */ {(16<<2)|3,{100,116,101}},
/* 37158 */ {(16<<2)|3,{100,116,105}},
/* 37159 */ {(16<<2)|3,{100,116,111}},
/* 37160 */ {(16<<2)|3,{100,116,115}},
/* 37161 */ {(16<<2)|3,{100,116,116}},
/* 37162 */ {(11<<2)|2,{100,116,0}},
/* 37163 */ {(11<<2)|2,{100,116,0}},
/* 37164 */ {(11<<2)|2,{100,116,0}},
/* 37165 */ {(11<<2)|2,{100,116,0}},
/* 37166 */ {(11<<2)|2,{100,116,0}},
/* 37167 */ {(11<<2)|2,{100,116,0}},
/* 37168 */ {(11<<2)|2,{100,116,0}},
/* 37169 */ {(11<<2)|2,{100,116,0}},
/* 37170 */ {(11<<2)|2,{100,116,0}},
/* 37171 */ {(11<<2)|2,{100,116,0}},
/* 37172 */ {(11<<2)|2,{100,116,0}},
/* 37173 */ {(11<<2)|2,{100,116,0}},
/* 37174 */ {(11<<2)|2,{100,116,0}},
/* 37175 */ {(11<<2)|2,{100,116,0}},
/* 37176 */ {(11<<2)|2,{100,116,0}},
/* 37177 */ {(11<<2)|2,{100,116,0}},
/* 37178 */ {(11<<2)|2,{100,116,0}},
/* 37179 */ {(11<<2)|2,{100,116,0}},
/* 37180 */ {(11<<2)|2,{100,116,0}},
/* 37181 */ {(11<<2)|2,{100,116,0}},
/* 37182 */ {(11<<2)|2,{100,116,0}},
/* 37183 */ {(11<<2)|2,{100,116,0}},
/* 37184 */ {(12<<2)|2,{100,32,0}},
/* 37185 */ {(12<<2)|2,{100,32,0}},
/* 37186 */ {(12<<2)|2,{100,32,0}},
/* 37187 */ {(12<<2)|2,{100,32,0}},
/* 37188 */ {(12<<2)|2,{100,32,0}},
/* 37189 */ {(12<<2)|2,{100,32,0}},
/* 37190 */ {(12<<2)|2,{100,32,0}},
/* 37191 */ {(12<<2)|2,{100,32,0}},
/* 37192 */ {(12<<2)|2,{100,32,0}},
/* 37193 */ {(12<<2)|2,{100,32,0}},
/* 37194 */ {(12<<2)|2,{100,32,0}},
/* 37195 */ {(12<<2)|2,{100,32,0}},
/* 37196 */ {(12<<2)|2,{100,32,0}},
/* 37197 */ {(12<<2)|2,{100,32,0}},
/* 37198 */ {(12<<2)|2,{100,32,0}},
/* 37199 */ {(12<<2)|2,{100,32,0}},
/* 37200 */ {(12<<2)|2,{100,37,0}},
/* 37201 */ {(12<<2)|2,{100,37,0}},
/* 37202 */ {(12<<2)|2,{100,37,0}},
/* 37203 */ {(12<<2)|2,{100,37,0}},
/* 37204 */ {(12<<2)|2,{100,37,0}},
/* 37205 */ {(12<<2)|2,{100,37,0}},
/* 37206 */ {(12<<2)|2,{100,37,0}},
/* 37207 */ {(12<<2)|2,{100,37,0}},
/* 37208 */ {(12<<2)|2,{100,37,0}},
/* 37209 */ {(12<<2)|2,{100,37,0}},
/* 37210 */ {(12<<2)|2,{100,37,0}},
/* 37211 */ {(12<<2)|2,{100,37,0}},
/* 37212 */ {(12<<2)|2,{100,37,0}},
/* 37213 */ {(12<<2)|2,{100,37,0}},
/* 37214 */ {(12<<2)|2,{100,37,0}},
/* 37215 */ {(12<<2)|2,{100,37,0}},
/* 37216 */ {(12<<2)|2,{100,45,0}},
/* 37217 */ {(12<<2)|2,{100,45,0}},
/* 37218 */ {(12<<2)|2,{100,45,0}},
/* 37219 */ {(12<<2)|2,{100,45,0}},
/* 37220 */ {(12<<2)|2,{100,45,0}},
/* 37221 */ {(12<<2)|2,{100,45,0}},
/* 37222 */ {(12<<2)|2,{100,45,0}},
/* 37223 */ {(12<<2)|2,{100,45,0}},
/* 37224 */ {(12<<2)|2,{100,45,0}},
/* 37225 */ {(12<<2)|2,{100,45,0}},
/* 37226 */ {(12<<2)|2,{100,45,0}},
/* 37227 */ {(12<<2)|2,{100,45,0}},
/* 37228 */ {(12<<2)|2,{100,45,0}},
/* 37229 */ {(12<<2)|2,{100,45,0}},
/* 37230 */ {(12<<2)|2,{100,45,0}},
/* 37231 */ {(12<<2)|2,{100,45,0}},
/* 37232 */ {(12<<2)|2,{100,46,0}},
/* 37233 */ {(12<<2)|2,{100,46,0}},
/* 37234 */ {(12<<2)|2,{100,46,0}},
/* 37235 */ {(12<<2)|2,{100,46,0}},
/* 37236 */ {(12<<2)|2,{100,46,0}},
/* 37237 */ {(12<<2)|2,{100,46,0}},
/* 37238 */ {(12<<2)|2,{100,46,0}},
/* 37239 */ {(12<<2)|2,{100,46,0}},
/* 37240 */ {(12<<2)|2,{100,46,0}},
/* 37241 */ {(12<<2)|2,{100,46,0}},
/* 37242 */ {(12<<2)|2,{100,46,0}},
/* 37243 */ {(12<<2)|2,{100,46,0}},
/* 37244 */ {(12<<2)|2,{100,46,0}},
/* 37245 */ {(12<<2)|2,{100,46,0}},
/* 37246 */ {(12<<2)|2,{100,46,0}},
/* 37247 */ {(12<<2)|2,{100,46,0}},
/* 37248 */ {(12<<2)|2,{100,47,0}},
/* 37249 */ {(12<<2)|2,{100,47,0}},
/* 37250 */ {(12<<2)|2,{100,47,0}},
/* 37251 */ {(12<<2)|2,{100,47,0}},
/* 37252 */ {(12<<2)|2,{100,47,0}},
/* 37253 */ {(12<<2)|2,{100,47,0}},
/* 37254 */ {(12<<2)|2,{100,47,0}},
/* 37255 */ {(12<<2)|2,{100,47,0}},
/* 37256 */ {(12<<2)|2,{100,47,0}},
/* 37257 */ {(12<<2)|2,{100,47,0}},
/* 37258 */ {(12<<2)|2,{100,47,0}},
/* 37259 */ {(12<<2)|2,{100,47,0}},
/* 37260 */ {(12<<2)|2,{100,47,0}},
/* 37261 */ {(12<<2)|2,{100,47,0}},
/* 37262 */ {(12<<2)|2,{100,47,0}},
/* 37263 */ {(12<<2)|2,{100,47,0}},
/* 37264 */ {(12<<2)|2,{100,51,0}},
/* 37265 */ {(12<<2)|2,{100,51,0}},
/* 37266 */ {(12<<2)|2,{100,51,0}},
/* 37267 */ {(12<<2)|2,{100,51,0}},
/* 37268 */ {(12<<2)|2,{100,51,0}},
/* 37269 */ {(12<<2)|2,{100,51,0}},
/* 37270 */ {(12<<2)|2,{100,51,0}},
/* 37271 */ {(12<<2)|2,{100,51,0}},
/* 37272 */ {(12<<2)|2,{100,51,0}},
/* 37273 */ {(12<<2)|2,{100,51,0}},
/* 37274 */ {(12<<2)|2,{100,51,0}},
/* 37275 */ {(12<<2)|2,{100,51,0}},
/* 37276 */ {(12<<2)|2,{100,51,0}},
/* 37277 */ {(12<<2)|2,{100,51,0}},
/* 37278 */ {(12<<2)|2,{100,51,0}},
/* 37279 */ {(12<<2)|2,{100,51,0}},
/* 37280 */ {(12<<2)|2,{100,52,0}},
/* 37281 */ {(12<<2)|2,{100,52,0}},
/* 37282 */ {(12<<2)|2,{100,52,0}},
/* 37283 */ {(12<<2)|2,{100,52,0}},
/* 37284 */ {(12<<2)|2,{100,52,0}},
/* 37285 */ {(12<<2)|2,{100,52,0}},
/* 37286 */ {(12<<2)|2,{100,52,0}},
/* 37287 */ {(12<<2)|2,{100,52,0}},
/* 37288 */ {(12<<2)|2,{100,52,0}},
/* 37289 */ {(12<<2)|2,{100,52,0}},
/* 37290 */ {(12<<2)|2,{100,52,0}},
/* 37291 */ {(12<<2)|2,{100,52,0}},
/* 37292 */ {(12<<2)|2,{100,52,0}},
/* 37293 */ {(12<<2)|2,{100,52,0}},
/* 37294 */ {(12<<2)|2,{100,52,0}},
/* 37295 */ {(12<<2)|2,{100,52,0}},
/* 37296 */ {(12<<2)|2,{100,53,0}},
/* 37297 */ {(12<<2)|2,{100,53,0}},
/* 37298 */ {(12<<2)|2,{100,53,0}},
/* 37299 */ {(12<<2)|2,{100,53,0}},
/* 37300 */ {(12<<2)|2,{100,53,0}},
/* 37301 */ {(12<<2)|2,{100,53,0}},
/* 37302 */ {(12<<2)|2,{100,53,0}},
/* 37303 */ {(12<<2)|2,{100,53,0}},
/* 37304 */ {(12<<2)|2,{100,53,0}},
/* 37305 */ {(12<<2)|2,{100,53,0}},
/* 37306 */ {(12<<2)|2,{100,53,0}},
/* 37307 */ {(12<<2)|2,{100,53,0}},
/* 37308 */ {(12<<2)|2,{100,53,0}},
/* 37309 */ {(12<<2)|2,{100,53,0}},
/* 37310 */ {(12<<2)|2,{100,53,0}},
/* 37311 */ {(12<<2)|2,{100,53,0}},
/* 37312 */ {(12<<2)|2,{100,54,0}},
/* 37313 */ {(12<<2)|2,{100,54,0}},
/* 37314 */ {(12<<2)|2,{100,54,0}},
/* 37315 */ {(12<<2)|2,{100,54,0}},
/* 37316 */ {(12<<2)|2,{100,54,0}},
/* 37317 */ {(12<<2)|2,{100,54,0}},
/* 37318 */ {(12<<2)|2,{100,54,0}},
/* 37319 */ {(12<<2)|2,{100,54,0}},
/* 37320 */ {(12<<2)|2,{100,54,0}},
/* 37321 */ {(12<<2)|2,{100,54,0}},
/* 37322 */ {(12<<2)|2,{100,54,0}},
/* 37323 */ {(12<<2)|2,{100,54,0}},
/* 37324 */ {(12<<2)|2,{100,54,0}},
/* 37325 */ {(12<<2)|2,{100,54,0}},
/* 37326 */ {(12<<2)|2,{100,54,0}},
/* 37327 */ {(12<<2)|2,{100,54,0}},
/* 37328 */ {(12<<2)|2,{100,55,0}},
/* 37329 */ {(12<<2)|2,{100,55,0}},
/* 37330 */ {(12<<2)|2,{100,55,0}},
/* 37331 */ {(12<<2)|2,{100,55,0}},
/* 37332 */ {(12<<2)|2,{100,55,0}},
/* 37333 */ {(12<<2)|2,{100,55,0}},
/* 37334 */ {(12<<2)|2,{100,55,0}},
/* 37335 */ {(12<<2)|2,{100,55,0}},
/* 37336 */ {(12<<2)|2,{100,55,0}},
/* 37337 */ {(12<<2)|2,{100,55,0}},
/* 37338 */ {(12<<2)|2,{100,55,0}},
/* 37339 */ {(12<<2)|2,{100,55,0}},
/* 37340 */ {(12<<2)|2,{100,55,0}},
/* 37341 */ {(12<<2)|2,{100,55,0}},
/* 37342 */ {(12<<2)|2,{100,55,0}},
/* 37343 */ {(12<<2)|2,{100,55,0}},
/* 37344 */ {(12<<2)|2,{100,56,0}},
/* 37345 */ {(12<<2)|2,{100,56,0}},
/* 37346 */ {(12<<2)|2,{100,56,0}},
/* 37347 */ {(12<<2)|2,{100,56,0}},
/* 37348 */ {(12<<2)|2,{100,56,0}},
/* 37349 */ {(12<<2)|2,{100,56,0}},
/* 37350 */ {(12<<2)|2,{100,56,0}},
/* 37351 */ {(12<<2)|2,{100,56,0}},
/* 37352 */ {(12<<2)|2,{100,56,0}},
/* 37353 */ {(12<<2)|2,{100,56,0}},
/* 37354 */ {(12<<2)|2,{100,56,0}},
/* 37355 */ {(12<<2)|2,{100,56,0}},
/* 37356 */ {(12<<2)|2,{100,56,0}},
/* 37357 */ {(12<<2)|2,{100,56,0}},
/* 37358 */ {(12<<2)|2,{100,56,0}},
/* 37359 */ {(12<<2)|2,{100,56,0}},
/* 37360 */ {(12<<2)|2,{100,57,0}},
/* 37361 */ {(12<<2)|2,{100,57,0}},
/* 37362 */ {(12<<2)|2,{100,57,0}},
/* 37363 */ {(12<<2)|2,{100,57,0}},
/* 37364 */ {(12<<2)|2,{100,57,0}},
/* 37365 */ {(12<<2)|2,{100,57,0}},
/* 37366 */ {(12<<2)|2,{100,57,0}},
/* 37367 */ {(12<<2)|2,{100,57,0}},
/* 37368 */ {(12<<2)|2,{100,57,0}},
/* 37369 */ {(12<<2)|2,{100,57,0}},
/* 37370 */ {(12<<2)|2,{100,57,0}},
/* 37371 */ {(12<<2)|2,{100,57,0}},
/* 37372 */ {(12<<2)|2,{100,57,0}},
/* 37373 */ {(12<<2)|2,{100,57,0}},
/* 37374 */ {(12<<2)|2,{100,57,0}},
/* 37375 */ {(12<<2)|2,{100,57,0}},
/* 37376 */ {(12<<2)|2,{100,61,0}},
/* 37377 */ {(12<<2)|2,{100,61,0}},
/* 37378 */ {(12<<2)|2,{100,61,0}},
/* 37379 */ {(12<<2)|2,{100,61,0}},
/* 37380 */ {(12<<2)|2,{100,61,0}},
/* 37381 */ {(12<<2)|2,{100,61,0}},
/* 37382 */ {(12<<2)|2,{100,61,0}},
/* 37383 */ {(12<<2)|2,{100,61,0}},
/* 37384 */ {(12<<2)|2,{100,61,0}},
/* 37385 */ {(12<<2)|2,{100,61,0}},
/* 37386 */ {(12<<2)|2,{100,61,0}},
/* 37387 */ {(12<<2)|2,{100,61,0}},
/* 37388 */ {(12<<2)|2,{100,61,0}},
/* 37389 */ {(12<<2)|2,{100,61,0}},
/* 37390 */ {(12<<2)|2,{100,61,0}},
/* 37391 */ {(12<<2)|2,{100,61,0}},
/* 37392 */ {(12<<2)|2,{100,65,0}},
/* 37393 */ {(12<<2)|2,{100,65,0}},
/* 37394 */ {(12<<2)|2,{100,65,0}},
/* 37395 */ {(12<<2)|2,{100,65,0}},
/* 37396 */ {(12<<2)|2,{100,65,0}},
/* 37397 */ {(12<<2)|2,{100,65,0}},
/* 37398 */ {(12<<2)|2,{100,65,0}},
/* 37399 */ {(12<<2)|2,{100,65,0}},
/* 37400 */ {(12<<2)|2,{100,65,0}},
/* 37401 */ {(12<<2)|2,{100,65,0}},
/* 37402 */ {(12<<2)|2,{100,65,0}},
/* 37403 */ {(12<<2)|2,{100,65,0}},
/* 37404 */ {(12<<2)|2,{100,65,0}},
/* 37405 */ {(12<<2)|2,{100,65,0}},
/* 37406 */ {(12<<2)|2,{100,65,0}},
/* 37407 */ {(12<<2)|2,{100,65,0}},
/* 37408 */ {(12<<2)|2,{100,95,0}},
/* 37409 */ {(12<<2)|2,{100,95,0}},
/* 37410 */ {(12<<2)|2,{100,95,0}},
/* 37411 */ {(12<<2)|2,{100,95,0}},
/* 37412 */ {(12<<2)|2,{100,95,0}},
/* 37413 */ {(12<<2)|2,{100,95,0}},
/* 37414 */ {(12<<2)|2,{100,95,0}},
/* 37415 */ {(12<<2)|2,{100,95,0}},
/* 37416 */ {(12<<2)|2,{100,95,0}},
/* 37417 */ {(12<<2)|2,{100,95,0}},
/* 37418 */ {(12<<2)|2,{100,95,0}},
/* 37419 */ {(12<<2)|2,{100,95,0}},
/* 37420 */ {(12<<2)|2,{100,95,0}},
/* 37421 */ {(12<<2)|2,{100,95,0}},
/* 37422 */ {(12<<2)|2,{100,95,0}},
/* 37423 */ {(12<<2)|2,{100,95,0}},
/* 37424 */ {(12<<2)|2,{100,98,0}},
/* 37425 */ {(12<<2)|2,{100,98,0}},
/* 37426 */ {(12<<2)|2,{100,98,0}},
/* 37427 */ {(12<<2)|2,{100,98,0}},
/* 37428 */ {(12<<2)|2,{100,98,0}},
/* 37429 */ {(12<<2)|2,{100,98,0}},
/* 37430 */ {(12<<2)|2,{100,98,0}},
/* 37431 */ {(12<<2)|2,{100,98,0}},
/* 37432 */ {(12<<2)|2,{100,98,0}},
/* 37433 */ {(12<<2)|2,{100,98,0}},
/* 37434 */ {(12<<2)|2,{100,98,0}},
/* 37435 */ {(12<<2)|2,{100,98,0}},
/* 37436 */ {(12<<2)|2,{100,98,0}},
/* 37437 */ {(12<<2)|2,{100,98,0}},
/* 37438 */ {(12<<2)|2,{100,98,0}},
/* 37439 */ {(12<<2)|2,{100,98,0}},
/* 37440 */ {(12<<2)|2,{100,100,0}},
/* 37441 */ {(12<<2)|2,{100,100,0}},
/* 37442 */ {(12<<2)|2,{100,100,0}},
/* 37443 */ {(12<<2)|2,{100,100,0}},
/* 37444 */ {(12<<2)|2,{100,100,0}},
/* 37445 */ {(12<<2)|2,{100,100,0}},
/* 37446 */ {(12<<2)|2,{100,100,0}},
/* 37447 */ {(12<<2)|2,{100,100,0}},
/* 37448 */ {(12<<2)|2,{100,100,0}},
/* 37449 */ {(12<<2)|2,{100,100,0}},
/* 37450 */ {(12<<2)|2,{100,100,0}},
/* 37451 */ {(12<<2)|2,{100,100,0}},
/* 37452 */ {(12<<2)|2,{100,100,0}},
/* 37453 */ {(12<<2)|2,{100,100,0}},
/* 37454 */ {(12<<2)|2,{100,100,0}},
/* 37455 */ {(12<<2)|2,{100,100,0}},
/* 37456 */ {(12<<2)|2,{100,102,0}},
/* 37457 */ {(12<<2)|2,{100,102,0}},
/* 37458 */ {(12<<2)|2,{100,102,0}},
/* 37459 */ {(12<<2)|2,{100,102,0}},
/* 37460 */ {(12<<2)|2,{100,102,0}},
/* 37461 */ {(12<<2)|2,{100,102,0}},
/* 37462 */ {(12<<2)|2,{100,102,0}},
/* 37463 */ {(12<<2)|2,{100,102,0}},
/* 37464 */ {(12<<2)|2,{100,102,0}},
/* 37465 */ {(12<<2)|2,{100,102,0}},
/* 37466 */ {(12<<2)|2,{100,102,0}},
/* 37467 */ {(12<<2)|2,{100,102,0}},
/* 37468 */ {(12<<2)|2,{100,102,0}},
/* 37469 */ {(12<<2)|2,{100,102,0}},
/* 37470 */ {(12<<2)|2,{100,102,0}},
/* 37471 */ {(12<<2)|2,{100,102,0}},
/* 37472 */ {(12<<2)|2,{100,103,0}},
/* 37473 */ {(12<<2)|2,{100,103,0}},
/* 37474 */ {(12<<2)|2,{100,103,0}},
/* 37475 */ {(12<<2)|2,{100,103,0}},
/* 37476 */ {(12<<2)|2,{100,103,0}},
/* 37477 */ {(12<<2)|2,{100,103,0}},
/* 37478 */ {(12<<2)|2,{100,103,0}},
/* 37479 */ {(12<<2)|2,{100,103,0}},
/* 37480 */ {(12<<2)|2,{100,103,0}},
/* 37481 */ {(12<<2)|2,{100,103,0}},
/* 37482 */ {(12<<2)|2,{100,103,0}},
/* 37483 */ {(12<<2)|2,{100,103,0}},
/* 37484 */ {(12<<2)|2,{100,103,0}},
/* 37485 */ {(12<<2)|2,{100,103,0}},
/* 37486 */ {(12<<2)|2,{100,103,0}},
/* 37487 */ {(12<<2)|2,{100,103,0}},
/* 37488 */ {(12<<2)|2,{100,104,0}},
/* 37489 */ {(12<<2)|2,{100,104,0}},
/* 37490 */ {(12<<2)|2,{100,104,0}},
/* 37491 */ {(12<<2)|2,{100,104,0}},
/* 37492 */ {(12<<2)|2,{100,104,0}},
/* 37493 */ {(12<<2)|2,{100,104,0}},
/* 37494 */ {(12<<2)|2,{100,104,0}},
/* 37495 */ {(12<<2)|2,{100,104,0}},
/* 37496 */ {(12<<2)|2,{100,104,0}},
/* 37497 */ {(12<<2)|2,{100,104,0}},
/* 37498 */ {(12<<2)|2,{100,104,0}},
/* 37499 */ {(12<<2)|2,{100,104,0}},
/* 37500 */ {(12<<2)|2,{100,104,0}},
/* 37501 */ {(12<<2)|2,{100,104,0}},
/* 37502 */ {(12<<2)|2,{100,104,0}},
/* 37503 */ {(12<<2)|2,{100,104,0}},
/* 37504 */ {(12<<2)|2,{100,108,0}},
/* 37505 */ {(12<<2)|2,{100,108,0}},
/* 37506 */ {(12<<2)|2,{100,108,0}},
/* 37507 */ {(12<<2)|2,{100,108,0}},
/* 37508 */ {(12<<2)|2,{100,108,0}},
/* 37509 */ {(12<<2)|2,{100,108,0}},
/* 37510 */ {(12<<2)|2,{100,108,0}},
/* 37511 */ {(12<<2)|2,{100,108,0}},
/* 37512 */ {(12<<2)|2,{100,108,0}},
/* 37513 */ {(12<<2)|2,{100,108,0}},
/* 37514 */ {(12<<2)|2,{100,108,0}},
/* 37515 */ {(12<<2)|2,{100,108,0}},
/* 37516 */ {(12<<2)|2,{100,108,0}},
/* 37517 */ {(12<<2)|2,{100,108,0}},
/* 37518 */ {(12<<2)|2,{100,108,0}},
/* 37519 */ {(12<<2)|2,{100,108,0}},
/* 37520 */ {(12<<2)|2,{100,109,0}},
/* 37521 */ {(12<<2)|2,{100,109,0}},
/* 37522 */ {(12<<2)|2,{100,109,0}},
/* 37523 */ {(12<<2)|2,{100,109,0}},
/* 37524 */ {(12<<2)|2,{100,109,0}},
/* 37525 */ {(12<<2)|2,{100,109,0}},
/* 37526 */ {(12<<2)|2,{100,109,0}},
/* 37527 */ {(12<<2)|2,{100,109,0}},
/* 37528 */ {(12<<2)|2,{100,109,0}},
/* 37529 */ {(12<<2)|2,{100,109,0}},
/* 37530 */ {(12<<2)|2,{100,109,0}},
/* 37531 */ {(12<<2)|2,{100,109,0}},
/* 37532 */ {(12<<2)|2,{100,109,0}},
/* 37533 */ {(12<<2)|2,{100,109,0}},
/* 37534 */ {(12<<2)|2,{100,109,0}},
/* 37535 */ {(12<<2)|2,{100,109,0}},
/* 37536 */ {(12<<2)|2,{100,110,0}},
/* 37537 */ {(12<<2)|2,{100,110,0}},
/* 37538 */ {(12<<2)|2,{100,110,0}},
/* 37539 */ {(12<<2)|2,{100,110,0}},
/* 37540 */ {(12<<2)|2,{100,110,0}},
/* 37541 */ {(12<<2)|2,{100,110,0}},
/* 37542 */ {(12<<2)|2,{100,110,0}},
/* 37543 */ {(12<<2)|2,{100,110,0}},
/* 37544 */ {(12<<2)|2,{100,110,0}},
/* 37545 */ {(12<<2)|2,{100,110,0}},
/* 37546 */ {(12<<2)|2,{100,110,0}},
/* 37547 */ {(12<<2)|2,{100,110,0}},
/* 37548 */ {(12<<2)|2,{100,110,0}},
/* 37549 */ {(12<<2)|2,{100,110,0}},
/* 37550 */ {(12<<2)|2,{100,110,0}},
/* 37551 */ {(12<<2)|2,{100,110,0}},
/* 37552 */ {(12<<2)|2,{100,112,0}},
/* 37553 */ {(12<<2)|2,{100,112,0}},
/* 37554 */ {(12<<2)|2,{100,112,0}},
/* 37555 */ {(12<<2)|2,{100,112,0}},
/* 37556 */ {(12<<2)|2,{100,112,0}},
/* 37557 */ {(12<<2)|2,{100,112,0}},
/* 37558 */ {(12<<2)|2,{100,112,0}},
/* 37559 */ {(12<<2)|2,{100,112,0}},
/* 37560 */ {(12<<2)|2,{100,112,0}},
/* 37561 */ {(12<<2)|2,{100,112,0}},
/* 37562 */ {(12<<2)|2,{100,112,0}},
/* 37563 */ {(12<<2)|2,{100,112,0}},
/* 37564 */ {(12<<2)|2,{100,112,0}},
/* 37565 */ {(12<<2)|2,{100,112,0}},
/* 37566 */ {(12<<2)|2,{100,112,0}},
/* 37567 */ {(12<<2)|2,{100,112,0}},
/* 37568 */ {(12<<2)|2,{100,114,0}},
/* 37569 */ {(12<<2)|2,{100,114,0}},
/* 37570 */ {(12<<2)|2,{100,114,0}},
/* 37571 */ {(12<<2)|2,{100,114,0}},
/* 37572 */ {(12<<2)|2,{100,114,0}},
/* 37573 */ {(12<<2)|2,{100,114,0}},
/* 37574 */ {(12<<2)|2,{100,114,0}},
/* 37575 */ {(12<<2)|2,{100,114,0}},
/* 37576 */ {(12<<2)|2,{100,114,0}},
/* 37577 */ {(12<<2)|2,{100,114,0}},
/* 37578 */ {(12<<2)|2,{100,114,0}},
/* 37579 */ {(12<<2)|2,{100,114,0}},
/* 37580 */ {(12<<2)|2,{100,114,0}},
/* 37581 */ {(12<<2)|2,{100,114,0}},
/* 37582 */ {(12<<2)|2,{100,114,0}},
/* 37583 */ {(12<<2)|2,{100,114,0}},
/* 37584 */ {(12<<2)|2,{100,117,0}},
/* 37585 */ {(12<<2)|2,{100,117,0}},
/* 37586 */ {(12<<2)|2,{100,117,0}},
/* 37587 */ {(12<<2)|2,{100,117,0}},
/* 37588 */ {(12<<2)|2,{100,117,0}},
/* 37589 */ {(12<<2)|2,{100,117,0}},
/* 37590 */ {(12<<2)|2,{100,117,0}},
/* 37591 */ {(12<<2)|2,{100,117,0}},
/* 37592 */ {(12<<2)|2,{100,117,0}},
/* 37593 */ {(12<<2)|2,{100,117,0}},
/* 37594 */ {(12<<2)|2,{100,117,0}},
/* 37595 */ {(12<<2)|2,{100,117,0}},
/* 37596 */ {(12<<2)|2,{100,117,0}},
/* 37597 */ {(12<<2)|2,{100,117,0}},
/* 37598 */ {(12<<2)|2,{100,117,0}},
/* 37599 */ {(12<<2)|2,{100,117,0}},
/* 37600 */ {(13<<2)|2,{100,58,0}},
/* 37601 */ {(13<<2)|2,{100,58,0}},
/* 37602 */ {(13<<2)|2,{100,58,0}},
/* 37603 */ {(13<<2)|2,{100,58,0}},
/* 37604 */ {(13<<2)|2,{100,58,0}},
/* 37605 */ {(13<<2)|2,{100,58,0}},
/* 37606 */ {(13<<2)|2,{100,58,0}},
/* 37607 */ {(13<<2)|2,{100,58,0}},
/* 37608 */ {(13<<2)|2,{100,66,0}},
/* 37609 */ {(13<<2)|2,{100,66,0}},
/* 37610 */ {(13<<2)|2,{100,66,0}},
/* 37611 */ {(13<<2)|2,{100,66,0}},
/* 37612 */ {(13<<2)|2,{100,66,0}},
/* 37613 */ {(13<<2)|2,{100,66,0}},
/* 37614 */ {(13<<2)|2,{100,66,0}},
/* 37615 */ {(13<<2)|2,{100,66,0}},
/* 37616 */ {(13<<2)|2,{100,67,0}},
/* 37617 */ {(13<<2)|2,{100,67,0}},
/* 37618 */ {(13<<2)|2,{100,67,0}},
/* 37619 */ {(13<<2)|2,{100,67,0}},
/* 37620 */ {(13<<2)|2,{100,67,0}},
/* 37621 */ {(13<<2)|2,{100,67,0}},
/* 37622 */ {(13<<2)|2,{100,67,0}},
/* 37623 */ {(13<<2)|2,{100,67,0}},
/* 37624 */ {(13<<2)|2,{100,68,0}},
/* 37625 */ {(13<<2)|2,{100,68,0}},
/* 37626 */ {(13<<2)|2,{100,68,0}},
/* 37627 */ {(13<<2)|2,{100,68,0}},
/* 37628 */ {(13<<2)|2,{100,68,0}},
/* 37629 */ {(13<<2)|2,{100,68,0}},
/* 37630 */ {(13<<2)|2,{100,68,0}},
/* 37631 */ {(13<<2)|2,{100,68,0}},
/* 37632 */ {(13<<2)|2,{100,69,0}},
/* 37633 */ {(13<<2)|2,{100,69,0}},
/* 37634 */ {(13<<2)|2,{100,69,0}},
/* 37635 */ {(13<<2)|2,{100,69,0}},
/* 37636 */ {(13<<2)|2,{100,69,0}},
/* 37637 */ {(13<<2)|2,{100,69,0}},
/* 37638 */ {(13<<2)|2,{100,69,0}},
/* 37639 */ {(13<<2)|2,{100,69,0}},
/* 37640 */ {(13<<2)|2,{100,70,0}},
/* 37641 */ {(13<<2)|2,{100,70,0}},
/* 37642 */ {(13<<2)|2,{100,70,0}},
/* 37643 */ {(13<<2)|2,{100,70,0}},
/* 37644 */ {(13<<2)|2,{100,70,0}},
/* 37645 */ {(13<<2)|2,{100,70,0}},
/* 37646 */ {(13<<2)|2,{100,70,0}},
/* 37647 */ {(13<<2)|2,{100,70,0}},
/* 37648 */ {(13<<2)|2,{100,71,0}},
/* 37649 */ {(13<<2)|2,{100,71,0}},
/* 37650 */ {(13<<2)|2,{100,71,0}},
/* 37651 */ {(13<<2)|2,{100,71,0}},
/* 37652 */ {(13<<2)|2,{100,71,0}},
/* 37653 */ {(13<<2)|2,{100,71,0}},
/* 37654 */ {(13<<2)|2,{100,71,0}},
/* 37655 */ {(13<<2)|2,{100,71,0}},
/* 37656 */ {(13<<2)|2,{100,72,0}},
/* 37657 */ {(13<<2)|2,{100,72,0}},
/* 37658 */ {(13<<2)|2,{100,72,0}},
/* 37659 */ {(13<<2)|2,{100,72,0}},
/* 37660 */ {(13<<2)|2,{100,72,0}},
/* 37661 */ {(13<<2)|2,{100,72,0}},
/* 37662 */ {(13<<2)|2,{100,72,0}},
/* 37663 */ {(13<<2)|2,{100,72,0}},
/* 37664 */ {(13<<2)|2,{100,73,0}},
/* 37665 */ {(13<<2)|2,{100,73,0}},
/* 37666 */ {(13<<2)|2,{100,73,0}},
/* 37667 */ {(13<<2)|2,{100,73,0}},
/* 37668 */ {(13<<2)|2,{100,73,0}},
/* 37669 */ {(13<<2)|2,{100,73,0}},
/* 37670 */ {(13<<2)|2,{100,73,0}},
/* 37671 */ {(13<<2)|2,{100,73,0}},
/* 37672 */ {(13<<2)|2,{100,74,0}},
/* 37673 */ {(13<<2)|2,{100,74,0}},
/* 37674 */ {(13<<2)|2,{100,74,0}},
/* 37675 */ {(13<<2)|2,{100,74,0}},
/* 37676 */ {(13<<2)|2,{100,74,0}},
/* 37677 */ {(13<<2)|2,{100,74,0}},
/* 37678 */ {(13<<2)|2,{100,74,0}},
/* 37679 */ {(13<<2)|2,{100,74,0}},
/* 37680 */ {(13<<2)|2,{100,75,0}},
/* 37681 */ {(13<<2)|2,{100,75,0}},
/* 37682 */ {(13<<2)|2,{100,75,0}},
/* 37683 */ {(13<<2)|2,{100,75,0}},
/* 37684 */ {(13<<2)|2,{100,75,0}},
/* 37685 */ {(13<<2)|2,{100,75,0}},
/* 37686 */ {(13<<2)|2,{100,75,0}},
/* 37687 */ {(13<<2)|2,{100,75,0}},
/* 37688 */ {(13<<2)|2,{100,76,0}},
/* 37689 */ {(13<<2)|2,{100,76,0}},
/* 37690 */ {(13<<2)|2,{100,76,0}},
/* 37691 */ {(13<<2)|2,{100,76,0}},
/* 37692 */ {(13<<2)|2,{100,76,0}},
/* 37693 */ {(13<<2)|2,{100,76,0}},
/* 37694 */ {(13<<2)|2,{100,76,0}},
/* 37695 */ {(13<<2)|2,{100,76,0}},
/* 37696 */ {(13<<2)|2,{100,77,0}},
/* 37697 */ {(13<<2)|2,{100,77,0}},
/* 37698 */ {(13<<2)|2,{100,77,0}},
/* 37699 */ {(13<<2)|2,{100,77,0}},
/* 37700 */ {(13<<2)|2,{100,77,0}},
/* 37701 */ {(13<<2)|2,{100,77,0}},
/* 37702 */ {(13<<2)|2,{100,77,0}},
/* 37703 */ {(13<<2)|2,{100,77,0}},
/* 37704 */ {(13<<2)|2,{100,78,0}},
/* 37705 */ {(13<<2)|2,{100,78,0}},
/* 37706 */ {(13<<2)|2,{100,78,0}},
/* 37707 */ {(13<<2)|2,{100,78,0}},
/* 37708 */ {(13<<2)|2,{100,78,0}},
/* 37709 */ {(13<<2)|2,{100,78,0}},
/* 37710 */ {(13<<2)|2,{100,78,0}},
/* 37711 */ {(13<<2)|2,{100,78,0}},
/* 37712 */ {(13<<2)|2,{100,79,0}},
/* 37713 */ {(13<<2)|2,{100,79,0}},
/* 37714 */ {(13<<2)|2,{100,79,0}},
/* 37715 */ {(13<<2)|2,{100,79,0}},
/* 37716 */ {(13<<2)|2,{100,79,0}},
/* 37717 */ {(13<<2)|2,{100,79,0}},
/* 37718 */ {(13<<2)|2,{100,79,0}},
/* 37719 */ {(13<<2)|2,{100,79,0}},
/* 37720 */ {(13<<2)|2,{100,80,0}},
/* 37721 */ {(13<<2)|2,{100,80,0}},
/* 37722 */ {(13<<2)|2,{100,80,0}},
/* 37723 */ {(13<<2)|2,{100,80,0}},
/* 37724 */ {(13<<2)|2,{100,80,0}},
/* 37725 */ {(13<<2)|2,{100,80,0}},
/* 37726 */ {(13<<2)|2,{100,80,0}},
/* 37727 */ {(13<<2)|2,{100,80,0}},
/* 37728 */ {(13<<2)|2,{100,81,0}},
/* 37729 */ {(13<<2)|2,{100,81,0}},
/* 37730 */ {(13<<2)|2,{100,81,0}},
/* 37731 */ {(13<<2)|2,{100,81,0}},
/* 37732 */ {(13<<2)|2,{100,81,0}},
/* 37733 */ {(13<<2)|2,{100,81,0}},
/* 37734 */ {(13<<2)|2,{100,81,0}},
/* 37735 */ {(13<<2)|2,{100,81,0}},
/* 37736 */ {(13<<2)|2,{100,82,0}},
/* 37737 */ {(13<<2)|2,{100,82,0}},
/* 37738 */ {(13<<2)|2,{100,82,0}},
/* 37739 */ {(13<<2)|2,{100,82,0}},
/* 37740 */ {(13<<2)|2,{100,82,0}},
/* 37741 */ {(13<<2)|2,{100,82,0}},
/* 37742 */ {(13<<2)|2,{100,82,0}},
/* 37743 */ {(13<<2)|2,{100,82,0}},
/* 37744 */ {(13<<2)|2,{100,83,0}},
/* 37745 */ {(13<<2)|2,{100,83,0}},
/* 37746 */ {(13<<2)|2,{100,83,0}},
/* 37747 */ {(13<<2)|2,{100,83,0}},
/* 37748 */ {(13<<2)|2,{100,83,0}},
/* 37749 */ {(13<<2)|2,{100,83,0}},
/* 37750 */ {(13<<2)|2,{100,83,0}},
/* 37751 */ {(13<<2)|2,{100,83,0}},
/* 37752 */ {(13<<2)|2,{100,84,0}},
/* 37753 */ {(13<<2)|2,{100,84,0}},
/* 37754 */ {(13<<2)|2,{100,84,0}},
/* 37755 */ {(13<<2)|2,{100,84,0}},
/* 37756 */ {(13<<2)|2,{100,84,0}},
/* 37757 */ {(13<<2)|2,{100,84,0}},
/* 37758 */ {(13<<2)|2,{100,84,0}},
/* 37759 */ {(13<<2)|2,{100,84,0}},
/* 37760 */ {(13<<2)|2,{100,85,0}},
/* 37761 */ {(13<<2)|2,{100,85,0}},
/* 37762 */ {(13<<2)|2,{100,85,0}},
/* 37763 */ {(13<<2)|2,{100,85,0}},
/* 37764 */ {(13<<2)|2,{100,85,0}},
/* 37765 */ {(13<<2)|2,{100,85,0}},
/* 37766 */ {(13<<2)|2,{100,85,0}},
/* 37767 */ {(13<<2)|2,{100,85,0}},
/* 37768 */ {(13<<2)|2,{100,86,0}},
/* 37769 */ {(13<<2)|2,{100,86,0}},
/* 37770 */ {(13<<2)|2,{100,86,0}},
/* 37771 */ {(13<<2)|2,{100,86,0}},
/* 37772 */ {(13<<2)|2,{100,86,0}},
/* 37773 */ {(13<<2)|2,{100,86,0}},
/* 37774 */ {(13<<2)|2,{100,86,0}},
/* 37775 */ {(13<<2)|2,{100,86,0}},
/* 37776 */ {(13<<2)|2,{100,87,0}},
/* 37777 */ {(13<<2)|2,{100,87,0}},
/* 37778 */ {(13<<2)|2,{100,87,0}},
/* 37779 */ {(13<<2)|2,{100,87,0}},
/* 37780 */ {(13<<2)|2,{100,87,0}},
/* 37781 */ {(13<<2)|2,{100,87,0}},
/* 37782 */ {(13<<2)|2,{100,87,0}},
/* 37783 */ {(13<<2)|2,{100,87,0}},
/* 37784 */ {(13<<2)|2,{100,89,0}},
/* 37785 */ {(13<<2)|2,{100,89,0}},
/* 37786 */ {(13<<2)|2,{100,89,0}},
/* 37787 */ {(13<<2)|2,{100,89,0}},
/* 37788 */ {(13<<2)|2,{100,89,0}},
/* 37789 */ {(13<<2)|2,{100,89,0}},
/* 37790 */ {(13<<2)|2,{100,89,0}},
/* 37791 */ {(13<<2)|2,{100,89,0}},
/* 37792 */ {(13<<2)|2,{100,106,0}},
/* 37793 */ {(13<<2)|2,{100,106,0}},
/* 37794 */ {(13<<2)|2,{100,106,0}},
/* 37795 */ {(13<<2)|2,{100,106,0}},
/* 37796 */ {(13<<2)|2,{100,106,0}},
/* 37797 */ {(13<<2)|2,{100,106,0}},
/* 37798 */ {(13<<2)|2,{100,106,0}},
/* 37799 */ {(13<<2)|2,{100,106,0}},
/* 37800 */ {(13<<2)|2,{100,107,0}},
/* 37801 */ {(13<<2)|2,{100,107,0}},
/* 37802 */ {(13<<2)|2,{100,107,0}},
/* 37803 */ {(13<<2)|2,{100,107,0}},
/* 37804 */ {(13<<2)|2,{100,107,0}},
/* 37805 */ {(13<<2)|2,{100,107,0}},
/* 37806 */ {(13<<2)|2,{100,107,0}},
/* 37807 */ {(13<<2)|2,{100,107,0}},
/* 37808 */ {(13<<2)|2,{100,113,0}},
/* 37809 */ {(13<<2)|2,{100,113,0}},
/* 37810 */ {(13<<2)|2,{100,113,0}},
/* 37811 */ {(13<<2)|2,{100,113,0}},
/* 37812 */ {(13<<2)|2,{100,113,0}},
/* 37813 */ {(13<<2)|2,{100,113,0}},
/* 37814 */ {(13<<2)|2,{100,113,0}},
/* 37815 */ {(13<<2)|2,{100,113,0}},
/* 37816 */ {(13<<2)|2,{100,118,0}},
/* 37817 */ {(13<<2)|2,{100,118,0}},
/* 37818 */ {(13<<2)|2,{100,118,0}},
/* 37819 */ {(13<<2)|2,{100,118,0}},
/* 37820 */ {(13<<2)|2,{100,118,0}},
/* 37821 */ {(13<<2)|2,{100,118,0}},
/* 37822 */ {(13<<2)|2,{100,118,0}},
/* 37823 */ {(13<<2)|2,{100,118,0}},
/* 37824 */ {(13<<2)|2,{100,119,0}},
/* 37825 */ {(13<<2)|2,{100,119,0}},
/* 37826 */ {(13<<2)|2,{100,119,0}},
/* 37827 */ {(13<<2)|2,{100,119,0}},
/* 37828 */ {(13<<2)|2,{100,119,0}},
/* 37829 */ {(13<<2)|2,{100,119,0}},
/* 37830 */ {(13<<2)|2,{100,119,0}},
/* 37831 */ {(13<<2)|2,{100,119,0}},
/* 37832 */ {(13<<2)|2,{100,120,0}},
/* 37833 */ {(13<<2)|2,{100,120,0}},
/* 37834 */ {(13<<2)|2,{100,120,0}},
/* 37835 */ {(13<<2)|2,{100,120,0}},
/* 37836 */ {(13<<2)|2,{100,120,0}},
/* 37837 */ {(13<<2)|2,{100,120,0}},
/* 37838 */ {(13<<2)|2,{100,120,0}},
/* 37839 */ {(13<<2)|2,{100,120,0}},
/* 37840 */ {(13<<2)|2,{100,121,0}},
/* 37841 */ {(13<<2)|2,{100,121,0}},
/* 37842 */ {(13<<2)|2,{100,121,0}},
/* 37843 */ {(13<<2)|2,{100,121,0}},
/* 37844 */ {(13<<2)|2,{100,121,0}},
/* 37845 */ {(13<<2)|2,{100,121,0}},
/* 37846 */ {(13<<2)|2,{100,121,0}},
/* 37847 */ {(13<<2)|2,{100,121,0}},
/* 37848 */ {(13<<2)|2,{100,122,0}},
/* 37849 */ {(13<<2)|2,{100,122,0}},
/* 37850 */ {(13<<2)|2,{100,122,0}},
/* 37851 */ {(13<<2)|2,{100,122,0}},
/* 37852 */ {(13<<2)|2,{100,122,0}},
/* 37853 */ {(13<<2)|2,{100,122,0}},
/* 37854 */ {(13<<2)|2,{100,122,0}},
/* 37855 */ {(13<<2)|2,{100,122,0}},
/* 37856 */ {(14<<2)|2,{100,38,0}},
/* 37857 */ {(14<<2)|2,{100,38,0}},
/* 37858 */ {(14<<2)|2,{100,38,0}},
/* 37859 */ {(14<<2)|2,{100,38,0}},
/* 37860 */ {(14<<2)|2,{100,42,0}},
/* 37861 */ {(14<<2)|2,{100,42,0}},
/* 37862 */ {(14<<2)|2,{100,42,0}},
/* 37863 */ {(14<<2)|2,{100,42,0}},
/* 37864 */ {(14<<2)|2,{100,44,0}},
/* 37865 */ {(14<<2)|2,{100,44,0}},
/* 37866 */ {(14<<2)|2,{100,44,0}},
/* 37867 */ {(14<<2)|2,{100,44,0}},
/* 37868 */ {(14<<2)|2,{100,59,0}},
/* 37869 */ {(14<<2)|2,{100,59,0}},
/* 37870 */ {(14<<2)|2,{100,59,0}},
/* 37871 */ {(14<<2)|2,{100,59,0}},
/* 37872 */ {(14<<2)|2,{100,88,0}},
/* 37873 */ {(14<<2)|2,{100,88,0}},
/* 37874 */ {(14<<2)|2,{100,88,0}},
/* 37875 */ {(14<<2)|2,{100,88,0}},
/* 37876 */ {(14<<2)|2,{100,90,0}},
/* 37877 */ {(14<<2)|2,{100,90,0}},
/* 37878 */ {(14<<2)|2,{100,90,0}},
/* 37879 */ {(14<<2)|2,{100,90,0}},
/* 37880 */ {(16<<2)|2,{100,33,0}},
/* 37881 */ {(16<<2)|2,{100,34,0}},
/* 37882 */ {(16<<2)|2,{100,40,0}},
/* 37883 */ {(16<<2)|2,{100,41,0}},
/* 37884 */ {(16<<2)|2,{100,63,0}},
/* 37885 */ {(6<<2)|1,{100,0,0}},
/* 37886 */ {(6<<2)|1,{100,0,0}},
/* 37887 */ {(6<<2)|1,{100,0,0}},
/* 37888 */ {(16<<2)|3,{102,48,48}},
/* 37889 */ {(16<<2)|3,{102,48,49}},
/* 37890 */ {(16<<2)|3,{102,48,50}},
/* 37891 */ {(16<<2)|3,{102,48,97}},
/* 37892 */ {(16<<2)|3,{102,48,99}},
/* 37893 */ {(16<<2)|3,{102,48,101}},
/* 37894 */ {(16<<2)|3,{102,48,105}},
/* 37895 */ {(16<<2)|3,{102,48,111}},
/* 37896 */ {(16<<2)|3,{102,48,115}},
/* 37897 */ {(16<<2)|3,{102,48,116}},
/* 37898 */ {(11<<2)|2,{102,48,0}},
/* 37899 */ {(11<<2)|2,{102,48,0}},
/* 37900 */ {(11<<2)|2,{102,48,0}},
/* 37901 */ {(11<<2)|2,{102,48,0}},
/* 37902 */ {(11<<2)|2,{102,48,0}},
/* 37903 */ {(11<<2)|2,{102,48,0}},
/* 37904 */ {(11<<2)|2,{102,48,0}},
/* 37905 */ {(11<<2)|2,{102,48,0}},
/* 37906 */ {(11<<2)|2,{102,48,0}},
/* 37907 */ {(11<<2)|2,{102,48,0}},
/* 37908 */ {(11<<2)|2,{102,48,0}},
/* 37909 */ {(11<<2)|2,{102,48,0}},
/* 37910 */ {(11<<2)|2,{102,48,0}},
/* 37911 */ {(11<<2)|2,{102,48,0}},
/* 37912 */ {(11<<2)|2,{102,48,0}},
/* 37913 */ {(11<<2)|2,{102,48,0}},
/* 37914 */ {(11<<2)|2,{102,48,0}},
/* 37915 */ {(11<<2)|2,{102,48,0}},
/* 37916 */ {(11<<2)|2,{102,48,0}},
/* 37917 */ {(11<<2)|2,{102,48,0}},
/* 37918 */ {(11<<2)|2,{102,48,0}},
/* 37919 */ {(11<<2)|2,{102,48,0}},
/* 37920 */ {(16<<2)|3,{102,49,48}},
/* 37921 */ {(16<<2)|3,{102,49,49}},
/* 37922 */ {(16<<2)|3,{102,49,50}},
/* 37923 */ {(16<<2)|3,{102,49,97}},
/* 37924 */ {(16<<2)|3,{102,49,99}},
/* 37925 */ {(16<<2)|3,{102,49,101}},
/* 37926 */ {(16<<2)|3,{102,49,105}},
/* 37927 */ {(16<<2)|3,{102,49,111}},
/* 37928 */ {(16<<2)|3,{102,49,115}},
/* 37929 */ {(16<<2)|3,{102,49,116}},
/* 37930 */ {(11<<2)|2,{102,49,0}},
/* 37931 */ {(11<<2)|2,{102,49,0}},
/* 37932 */ {(11<<2)|2,{102,49,0}},
/* 37933 */ {(11<<2)|2,{102,49,0}},
/* 37934 */ {(11<<2)|2,{102,49,0}},
/* 37935 */ {(11<<2)|2,{102,49,0}},
/* 37936 */ {(11<<2)|2,{102,49,0}},
/* 37937 */ {(11<<2)|2,{102,49,0}},
/* 37938 */ {(11<<2)|2,{102,49,0}},
/* 37939 */ {(11<<2)|2,{102,49,0}},
/* 37940 */ {(11<<2)|2,{102,49,0}},
/* 37941 */ {(11<<2)|2,{102,49,0}},
/* 37942 */ {(11<<2)|2,{102,49,0}},
/* 37943 */ {(11<<2)|2,{102,49,0}},
/* 37944 */ {(11<<2)|2,{102,49,0}},
/* 37945 */ {(11<<2)|2,{102,49,0}},
/* 37946 */ {(11<<2)|2,{102,49,0}},
/* 37947 */ {(11<<2)|2,{102,49,0}},
/* 37948 */ {(11<<2)|2,{102,49,0}},
/* 37949 */ {(11<<2)|2,{102,49,0}},
/* 37950 */ {(11<<2)|2,{102,49,0}},
/* 37951 */ {(11<<2)|2,{102,49,0}},
/* 37952 */ {(16<<2)|3,{102,50,48}},
/* 37953 */ {(16<<2)|3,{102,50,49}},
/* 37954 */ {(16<<2)|3,{102,50,50}},
/* 37955 */ {(16<<2)|3,{102,50,97}},
/* 37956 */ {(16<<2)|3,{102,50,99}},
/* 37957 */ {(16<<2)|3,{102,50,101}},
/* 37958 */ {(16<<2)|3,{102,50,105}},
/* 37959 */ {(16<<2)|3,{102,50,111}},
/* 37960 */ {(16<<2)|3,{102,50,115}},
/* 37961 */ {(16<<2)|3,{102,50,116}},
/* 37962 */ {(11<<2)|2,{102,50,0}},
/* 37963 */ {(11<<2)|2,{102,50,0}},
/* 37964 */ {(11<<2)|2,{102,50,0}},
/* 37965 */ {(11<<2)|2,{102,50,0}},
/* 37966 */ {(11<<2)|2,{102,50,0}},
/* 37967 */ {(11<<2)|2,{102,50,0}},
/* 37968 */ {(11<<2)|2,{102,50,0}},
/* 37969 */ {(11<<2)|2,{102,50,0}},
/* 37970 */ {(11<<2)|2,{102,50,0}},
/* 37971 */ {(11<<2)|2,{102,50,0}},
/* 37972 */ {(11<<2)|2,{102,50,0}},
/* 37973 */ {(11<<2)|2,{102,50,0}},
/* 37974 */ {(11<<2)|2,{102,50,0}},
/* 37975 */ {(11<<2)|2,{102,50,0}},
/* 37976 */ {(11<<2)|2,{102,50,0}},
/* 37977 */ {(11<<2)|2,{102,50,0}},
/* 37978 */ {(11<<2)|2,{102,50,0}},
/* 37979 */ {(11<<2)|2,{102,50,0}},
/* 37980 */ {(11<<2)|2,{102,50,0}},
/* 37981 */ {(11<<2)|2,{102,50,0}},
/* 37982 */ {(11<<2)|2,{102,50,0}},
/* 37983 */ {(11<<2)|2,{102,50,0}},
/* 37984 */ {(16<<2)|3,{102,97,48}},
/* 37985 */ {(16<<2)|3,{102,97,49}},
/* 37986 */ {(16<<2)|3,{102,97,50}},
/* 37987 */ {(16<<2)|3,{102,97,97}},
/* 37988 */ {(16<<2)|3,{102,97,99}},
/* 37989 */ {(16<<2)|3,{102,97,101}},
/* 37990 */ {(16<<2)|3,{102,97,105}},
/* 37991 */ {(16<<2)|3,{102,97,111}},
/* 37992 */ {(16<<2)|3,{102,97,115}},
/* 37993 */ {(16<<2)|3,{102,97,116}},
/* 37994 */ {(11<<2)|2,{102,97,0}},
/* 37995 */ {(11<<2)|2,{102,97,0}},
/* 37996 */ {(11<<2)|2,{102,97,0}},
/* 37997 */ {(11<<2)|2,{102,97,0}},
/* 37998 */ {(11<<2)|2,{102,97,0}},
/* 37999 */ {(11<<2)|2,{102,97,0}},
/* 38000 */ {(11<<2)|2,{102,97,0}},
/* 38001 */ {(11<<2)|2,{102,97,0}},
/* 38002 */ {(11<<2)|2,{102,97,0}},
/* 38003 */ {(11<<2)|2,{102,97,0}},
/* 38004 */ {(11<<2)|2,{102,97,0}},
/* 38005 */ {(11<<2)|2,{102,97,0}},
/* 38006 */ {(11<<2)|2,{102,97,0}},
/* 38007 */ {(11<<2)|2,{102,97,0}},
/* 38008 */ {(11<<2)|2,{102,97,0}},
/* 38009 */ {(11<<2)|2,{102,97,0}},
/* 38010 */ {(11<<2)|2,{102,97,0}},
/* 38011 */ {(11<<2)|2,{102,97,0}},
/* 38012 */ {(11<<2)|2,{102,97,0}},
/* 38013 */ {(11<<2)|2,{102,97,0}},
/* 38014 */ {(11<<2)|2,{102,97,0}},
/* 38015 */ {(11<<2)|2,{102,97,0}},
/* 38016 */ {(16<<2)|3,{102,99,48}},
/* 38017 */ {(16<<2)|3,{102,99,49}},
/* 38018 */ {(16<<2)|3,{102,99,50}},
/* 38019 */ {(16<<2)|3,{102,99,97}},
/* 38020 */ {(16<<2)|3,{102,99,99}},
/* 38021 */ {(16<<2)|3,{102,99,101}},
/* 38022 */ {(16<<2)|3,{102,99,105}},
/* 38023 */ {(16<<2)|3,{102,99,111}},
/* 38024 */ {(16<<2)|3,{102,99,115}},
/* 38025 */ {(16<<2)|3,{102,99,116}},
/* 38026 */ {(11<<2)|2,{102,99,0}},
/* 38027 */ {(11<<2)|2,{102,99,0}},
/* 38028 */ {(11<<2)|2,{102,99,0}},
/* 38029 */ {(11<<2)|2,{102,99,0}},
/* 38030 */ {(11<<2)|2,{102,99,0}},
/* 38031 */ {(11<<2)|2,{102,99,0}},
/* 38032 */ {(11<<2)|2,{102,99,0}},
/* 38033 */ {(11<<2)|2,{102,99,0}},
/* 38034 */ {(11<<2)|2,{102,99,0}},
/* 38035 */ {(11<<2)|2,{102,99,0}},
/* 38036 */ {(11<<2)|2,{102,99,0}},
/* 38037 */ {(11<<2)|2,{102,99,0}},
/* 38038 */ {(11<<2)|2,{102,99,0}},
/* 38039 */ {(11<<2)|2,{102,99,0}},
/* 38040 */ {(11<<2)|2,{102,99,0}},
/* 38041 */ {(11<<2)|2,{102,99,0}},
/* 38042 */ {(11<<2)|2,{102,99,0}},
/* 38043 */ {(11<<2)|2,{102,99,0}},
/* 38044 */ {(11<<2)|2,{102,99,0}},
/* 38045 */ {(11<<2)|2,{102,99,0}},
/* 38046 */ {(11<<2)|2,{102,99,0}},
/* 38047 */ {(11<<2)|2,{102,99,0}},
/* 38048 */ {(16<<2)|3,{102,101,48}},
/* 38049 */ {(16<<2)|3,{102,101,49}},
/* 38050 */ {(16<<2)|3,{102,101,50}},
/* 38051 */ {(16<<2)|3,{102,101,97}},
/* 38052 */ {(16<<2)|3,{102,101,99}},
/* 38053 */ {(16<<2)|3,{102,101,101}},
/* 38054 */ {(16<<2)|3,{102,101,105}},
/* 38055 */ {(16<<2)|3,{102,101,111}},
/* 38056 */ {(16<<2)|3,{102,101,115}},
/* 38057 */ {(16<<2)|3,{102,101,116}},
/* 38058 */ {(11<<2)|2,{102,101,0}},
/* 38059 */ {(11<<2)|2,{102,101,0}},
/* 38060 */ {(11<<2)|2,{102,101,0}},
/* 38061 */ {(11<<2)|2,{102,101,0}},
/* 38062 */ {(11<<2)|2,{102,101,0}},
/* 38063 */ {(11<<2)|2,{102,101,0}},
/* 38064 */ {(11<<2)|2,{102,101,0}},
/* 38065 */ {(11<<2)|2,{102,101,0}},
/* 38066 */ {(11<<2)|2,{102,101,0}},
/* 38067 */ {(11<<2)|2,{102,101,0}},
/* 38068 */ {(11<<2)|2,{102,101,0}},
/* 38069 */ {(11<<2)|2,{102,101,0}},
/* 38070 */ {(11<<2)|2,{102,101,0}},
/* 38071 */ {(11<<2)|2,{102,101,0}},
/* 38072 */ {(11<<2)|2,{102,101,0}},
/* 38073 */ {(11<<2)|2,{102,101,0}},
/* 38074 */ {(11<<2)|2,{102,101,0}},
/* 38075 */ {(11<<2)|2,{102,101,0}},
/* 38076 */ {(11<<2)|2,{102,101,0}},
/* 38077 */ {(11<<2)|2,{102,101,0}},
/* 38078 */ {(11<<2)|2,{102,101,0}},
/* 38079 */ {(11<<2)|2,{102,101,0}},
/* 38080 */ {(16<<2)|3,{102,105,48}},
/* 38081 */ {(16<<2)|3,{102,105,49}},
/* 38082 */ {(16<<2)|3,{102,105,50}},
/* 38083 */ {(16<<2)|3,{102,105,97}},
/* 38084 */ {(16<<2)|3,{102,105,99}},
/* 38085 */ {(16<<2)|3,{102,105,101}},
/* 38086 */ {(16<<2)|3,{102,105,105}},
/* 38087 */ {(16<<2)|3,{102,105,111}},
/* 38088 */ {(16<<2)|3,{102,105,115}},
/* 38089 */ {(16<<2)|3,{102,105,116}},
/* 38090 */ {(11<<2)|2,{102,105,0}},
/* 38091 */ {(11<<2)|2,{102,105,0}},
/* 38092 */ {(11<<2)|2,{102,105,0}},
/* 38093 */ {(11<<2)|2,{102,105,0}},
/* 38094 */ {(11<<2)|2,{102,105,0}},
/* 38095 */ {(11<<2)|2,{102,105,0}},
/* 38096 */ {(11<<2)|2,{102,105,0}},
/* 38097 */ {(11<<2)|2,{102,105,0}},
/* 38098 */ {(11<<2)|2,{102,105,0}},
/* 38099 */ {(11<<2)|2,{102,105,0}},
/* 38100 */ {(11<<2)|2,{102,105,0}},
/* 38101 */ {(11<<2)|2,{102,105,0}},
/* 38102 */ {(11<<2)|2,{102,105,0}},
/* 38103 */ {(11<<2)|2,{102,105,0}},
/* 38104 */ {(11<<2)|2,{102,105,0}},
/* 38105 */ {(11<<2)|2,{102,105,0}},
/* 38106 */ {(11<<2)|2,{102,105,0}},
/* 38107 */ {(11<<2)|2,{102,105,0}},
/* 38108 */ {(11<<2)|2,{102,105,0}},
/* 38109 */ {(11<<2)|2,{102,105,0}},
/* 38110 */ {(11<<2)|2,{102,105,0}},
/* 38111 */ {(11<<2)|2,{102,105,0}},
/* 38112 */ {(16<<2)|3,{102,111,48}},
/* 38113 */ {(16<<2)|3,{102,111,49}},
/* 38114 */ {(16<<2)|3,{102,111,50}},
/* 38115 */ {(16<<2)|3,{102,111,97}},
/* 38116 */ {(16<<2)|3,{102,111,99}},
/* 38117 */ {(16<<2)|3,{102,111,101}},
/* 38118 */ {(16<<2)|3,{102,111,105}},
/* 38119 */ {(16<<2)|3,{102,111,111}},
/* 38120 */ {(16<<2)|3,{102,111,115}},
/* 38121 */ {(16<<2)|3,{102,111,116}},
/* 38122 */ {(11<<2)|2,{102,111,0}},
/* 38123 */ {(11<<2)|2,{102,111,0}},
/* 38124 */ {(11<<2)|2,{102,111,0}},
/* 38125 */ {(11<<2)|2,{102,111,0}},
/* 38126 */ {(11<<2)|2,{102,111,0}},
/* 38127 */ {(11<<2)|2,{102,111,0}},
/* 38128 */ {(11<<2)|2,{102,111,0}},
/* 38129 */ {(11<<2)|2,{102,111,0}},
/* 38130 */ {(11<<2)|2,{102,111,0}},
/* 38131 */ {(11<<2)|2,{102,111,0}},
/* 38132 */ {(11<<2)|2,{102,111,0}},
/* 38133 */ {(11<<2)|2,{102,111,0}},
/* 38134 */ {(11<<2)|2,{102,111,0}},
/* 38135 */ {(11<<2)|2,{102,111,0}},
/* 38136 */ {(11<<2)|2,{102,111,0}},
/* 38137 */ {(11<<2)|2,{102,111,0}},
/* 38138 */ {(11<<2)|2,{102,111,0}},
/* 38139 */ {(11<<2)|2,{102,111,0}},
/* 38140 */ {(11<<2)|2,{102,111,0}},
/* 38141 */ {(11<<2)|2,{102,111,0}},
/* 38142 */ {(11<<2)|2,{102,111,0}},
/* 38143 */ {(11<<2)|2,{102,111,0}},
/* 38144 */ {(16<<2)|3,{102,115,48}},
/* 38145 */ {(16<<2)|3,{102,115,49}},
/* 38146 */ {(16<<2)|3,{102,115,50}},
/* 38147 */ {(16<<2)|3,{102,115,97}},
/* 38148 */ {(16<<2)|3,{102,115,99}},
/* 38149 */ {(16<<2)|3,{102,115,101}},
/* 38150 */ {(16<<2)|3,{102,115,105}},
/* 38151 */ {(16<<2)|3,{102,115,111}},
/* 38152 */ {(16<<2)|3,{102,115,115}},
/* 38153 */ {(16<<2)|3,{102,115,116}},
/* 38154 */ {(11<<2)|2,{102,115,0}},
/* 38155 */ {(11<<2)|2,{102,115,0}},
/* 38156 */ {(11<<2)|2,{102,115,0}},
/* 38157 */ {(11<<2)|2,{102,115,0}},
/* 38158 */ {(11<<2)|2,{102,115,0}},
/* 38159 */ {(11<<2)|2,{102,115,0}},
/* 38160 */ {(11<<2)|2,{102,115,0}},
/* 38161 */ {(11<<2)|2,{102,115,0}},
/* 38162 */ {(11<<2)|2,{102,115,0}},
/* 38163 */ {(11<<2)|2,{102,115,0}},
/* 38164 */ {(11<<2)|2,{102,115,0}},
/* 38165 */ {(11<<2)|2,{102,115,0}},
/* 38166 */ {(11<<2)|2,{102,115,0}},
/* 38167 */ {(11<<2)|2,{102,115,0}},
/* 38168 */ {(11<<2)|2,{102,115,0}},
/* 38169 */ {(11<<2)|2,{102,115,0}},
/* 38170 */ {(11<<2)|2,{102,115,0}},
/* 38171 */ {(11<<2)|2,{102,115,0}},
/* 38172 */ {(11<<2)|2,{102,115,0}},
/* 38173 */ {(11<<2)|2,{102,115,0}},
/* 38174 */ {(11<<2)|2,{102,115,0}},
/* 38175 */ {(11<<2)|2,{102,115,0}},
/* 38176 */ {(16<<2)|3,{102,116,48}},
/* 38177 */ {(16<<2)|3,{102,116,49}},
/* 38178 */ {(16<<2)|3,{102,116,50}},
/* 38179 */ {(16<<2)|3,{102,116,97}},
/* 38180 */ {(16<<2)|3,{102,116,99}},
/* 38181 */ {(16<<2)|3,{102,116,101}},
/* 38182 */ {(16<<2)|3,{102,116,105}},
/* 38183 */ {(16<<2)|3,{102,116,111}},
/* 38184 */ {(16<<2)|3,{102,116,115}},
/* 38185 */ {(16<<2)|3,{102,116,116}},
/* 38186 */ {(11<<2)|2,{102,116,0}},
/* 38187 */ {(11<<2)|2,{102,116,0}},
/* 38188 */ {(11<<2)|2,{102,116,0}},
/* 38189 */ {(11<<2)|2,{102,116,0}},
/* 38190 */ {(11<<2)|2,{102,116,0}},
/* 38191 */ {(11<<2)|2,{102,116,0}},
/* 38192 */ {(11<<2)|2,{102,116,0}},
/* 38193 */ {(11<<2)|2,{102,116,0}},
/* 38194 */ {(11<<2)|2,{102,116,0}},
/* 38195 */ {(11<<2)|2,{102,116,0}},
/* 38196 */ {(11<<2)|2,{102,116,0}},
/* 38197 */ {(11<<2)|2,{102,116,0}},
/* 38198 */ {(11<<2)|2,{102,116,0}},
/* 38199 */ {(11<<2)|2,{102,116,0}},
/* 38200 */ {(11<<2)|2,{102,116,0}},
/* 38201 */ {(11<<2)|2,{102,116,0}},
/* 38202 */ {(11<<2)|2,{102,116,0}},
/* 38203 */ {(11<<2)|2,{102,116,0}},
/* 38204 */ {(11<<2)|2,{102,116,0}},
/* 38205 */ {(11<<2)|2,{102,116,0}},
/* 38206 */ {(11<<2)|2,{102,116,0}},
/* 38207 */ {(11<<2)|2,{102,116,0}},
/* 38208 */ {(12<<2)|2,{102,32,0}},
/* 38209 */ {(12<<2)|2,{102,32,0}},
/* 38210 */ {(12<<2)|2,{102,32,0}},
/* 38211 */ {(12<<2)|2,{102,32,0}},
/* 38212 */ {(12<<2)|2,{102,32,0}},
/* 38213 */ {(12<<2)|2,{102,32,0}},
/* 38214 */ {(12<<2)|2,{102,32,0}},
/* 38215 */ {(12<<2)|2,{102,32,0}},
/* 38216 */ {(12<<2)|2,{102,32,0}},
/* 38217 */ {(12<<2)|2,{102,32,0}},
/* 38218 */ {(12<<2)|2,{102,32,0}},
/* 38219 */ {(12<<2)|2,{102,32,0}},
/* 38220 */ {(12<<2)|2,{102,32,0}},
/* 38221 */ {(12<<2)|2,{102,32,0}},
/* 38222 */ {(12<<2)|2,{102,32,0}},
/* 38223 */ {(12<<2)|2,{102,32,0}},
/* 38224 */ {(12<<2)|2,{102,37,0}},
/* 38225 */ {(12<<2)|2,{102,37,0}},
/* 38226 */ {(12<<2)|2,{102,37,0}},
/* 38227 */ {(12<<2)|2,{102,37,0}},
/* 38228 */ {(12<<2)|2,{102,37,0}},
/* 38229 */ {(12<<2)|2,{102,37,0}},
/* 38230 */ {(12<<2)|2,{102,37,0}},
/* 38231 */ {(12<<2)|2,{102,37,0}},
/* 38232 */ {(12<<2)|2,{102,37,0}},
/* 38233 */ {(12<<2)|2,{102,37,0}},
/* 38234 */ {(12<<2)|2,{102,37,0}},
/* 38235 */ {(12<<2)|2,{102,37,0}},
/* 38236 */ {(12<<2)|2,{102,37,0}},
/* 38237 */ {(12<<2)|2,{102,37,0}},
/* 38238 */ {(12<<2)|2,{102,37,0}},
/* 38239 */ {(12<<2)|2,{102,37,0}},
/* 38240 */ {(12<<2)|2,{102,45,0}},
/* 38241 */ {(12<<2)|2,{102,45,0}},
/* 38242 */ {(12<<2)|2,{102,45,0}},
/* 38243 */ {(12<<2)|2,{102,45,0}},
/* 38244 */ {(12<<2)|2,{102,45,0}},
/* 38245 */ {(12<<2)|2,{102,45,0}},
/* 38246 */ {(12<<2)|2,{102,45,0}},
/* 38247 */ {(12<<2)|2,{102,45,0}},
/* 38248 */ {(12<<2)|2,{102,45,0}},
/* 38249 */ {(12<<2)|2,{102,45,0}},
/* 38250 */ {(12<<2)|2,{102,45,0}},
/* 38251 */ {(12<<2)|2,{102,45,0}},
/* 38252 */ {(12<<2)|2,{102,45,0}},
/* 38253 */ {(12<<2)|2,{102,45,0}},
/* 38254 */ {(12<<2)|2,{102,45,0}},
/* 38255 */ {(12<<2)|2,{102,45,0}},
/* 38256 */ {(12<<2)|2,{102,46,0}},
/* 38257 */ {(12<<2)|2,{102,46,0}},
/* 38258 */ {(12<<2)|2,{102,46,0}},
/* 38259 */ {(12<<2)|2,{102,46,0}},
/* 38260 */ {(12<<2)|2,{102,46,0}},
/* 38261 */ {(12<<2)|2,{102,46,0}},
/* 38262 */ {(12<<2)|2,{102,46,0}},
/* 38263 */ {(12<<2)|2,{102,46,0}},
/* 38264 */ {(12<<2)|2,{102,46,0}},
/* 38265 */ {(12<<2)|2,{102,46,0}},
/* 38266 */ {(12<<2)|2,{102,46,0}},
/* 38267 */ {(12<<2)|2,{102,46,0}},
/* 38268 */ {(12<<2)|2,{102,46,0}},
/* 38269 */ {(12<<2)|2,{102,46,0}},
/* 38270 */ {(12<<2)|2,{102,46,0}},
/* 38271 */ {(12<<2)|2,{102,46,0}},
/* 38272 */ {(12<<2)|2,{102,47,0}},
/* 38273 */ {(12<<2)|2,{102,47,0}},
/* 38274 */ {(12<<2)|2,{102,47,0}},
/* 38275 */ {(12<<2)|2,{102,47,0}},
/* 38276 */ {(12<<2)|2,{102,47,0}},
/* 38277 */ {(12<<2)|2,{102,47,0}},
/* 38278 */ {(12<<2)|2,{102,47,0}},
/* 38279 */ {(12<<2)|2,{102,47,0}},
/* 38280 */ {(12<<2)|2,{102,47,0}},
/* 38281 */ {(12<<2)|2,{102,47,0}},
/* 38282 */ {(12<<2)|2,{102,47,0}},
/* 38283 */ {(12<<2)|2,{102,47,0}},
/* 38284 */ {(12<<2)|2,{102,47,0}},
/* 38285 */ {(12<<2)|2,{102,47,0}},
/* 38286 */ {(12<<2)|2,{102,47,0}},
/* 38287 */ {(12<<2)|2,{102,47,0}},
/* 38288 */ {(12<<2)|2,{102,51,0}},
/* 38289 */ {(12<<2)|2,{102,51,0}},
/* 38290 */ {(12<<2)|2,{102,51,0}},
/* 38291 */ {(12<<2)|2,{102,51,0}},
/* 38292 */ {(12<<2)|2,{102,51,0}},
/* 38293 */ {(12<<2)|2,{102,51,0}},
/* 38294 */ {(12<<2)|2,{102,51,0}},
/* 38295 */ {(12<<2)|2,{102,51,0}},
/* 38296 */ {(12<<2)|2,{102,51,0}},
/* 38297 */ {(12<<2)|2,{102,51,0}},
/* 38298 */ {(12<<2)|2,{102,51,0}},
/* 38299 */ {(12<<2)|2,{102,51,0}},
/* 38300 */ {(12<<2)|2,{102,51,0}},
/* 38301 */ {(12<<2)|2,{102,51,0}},
/* 38302 */ {(12<<2)|2,{102,51,0}},
/* 38303 */ {(12<<2)|2,{102,51,0}},
/* 38304 */ {(12<<2)|2,{102,52,0}},
/* 38305 */ {(12<<2)|2,{102,52,0}},
/* 38306 */ {(12<<2)|2,{102,52,0}},
/* 38307 */ {(12<<2)|2,{102,52,0}},
/* 38308 */ {(12<<2)|2,{102,52,0}},
/* 38309 */ {(12<<2)|2,{102,52,0}},
/* 38310 */ {(12<<2)|2,{102,52,0}},
/* 38311 */ {(12<<2)|2,{102,52,0}},
/* 38312 */ {(12<<2)|2,{102,52,0}},
/* 38313 */ {(12<<2)|2,{102,52,0}},
/* 38314 */ {(12<<2)|2,{102,52,0}},
/* 38315 */ {(12<<2)|2,{102,52,0}},
/* 38316 */ {(12<<2)|2,{102,52,0}},
/* 38317 */ {(12<<2)|2,{102,52,0}},
/* 38318 */ {(12<<2)|2,{102,52,0}},
/* 38319 */ {(12<<2)|2,{102,52,0}},
/* 38320 */ {(12<<2)|2,{102,53,0}},
/* 38321 */ {(12<<2)|2,{102,53,0}},
/* 38322 */ {(12<<2)|2,{102,53,0}},
/* 38323 */ {(12<<2)|2,{102,53,0}},
/* 38324 */ {(12<<2)|2,{102,53,0}},
/* 38325 */ {(12<<2)|2,{102,53,0}},
/* 38326 */ {(12<<2)|2,{102,53,0}},
/* 38327 */ {(12<<2)|2,{102,53,0}},
/* 38328 */ {(12<<2)|2,{102,53,0}},
/* 38329 */ {(12<<2)|2,{102,53,0}},
/* 38330 */ {(12<<2)|2,{102,53,0}},
/* 38331 */ {(12<<2)|2,{102,53,0}},
/* 38332 */ {(12<<2)|2,{102,53,0}},
/* 38333 */ {(12<<2)|2,{102,53,0}},
/* 38334 */ {(12<<2)|2,{102,53,0}},
/* 38335 */ {(12<<2)|2,{102,53,0}},
/* 38336 */ {(12<<2)|2,{102,54,0}},
/* 38337 */ {(12<<2)|2,{102,54,0}},
/* 38338 */ {(12<<2)|2,{102,54,0}},
/* 38339 */ {(12<<2)|2,{102,54,0}},
/* 38340 */ {(12<<2)|2,{102,54,0}},
/* 38341 */ {(12<<2)|2,{102,54,0}},
/* 38342 */ {(12<<2)|2,{102,54,0}},
/* 38343 */ {(12<<2)|2,{102,54,0}},
/* 38344 */ {(12<<2)|2,{102,54,0}},
/* 38345 */ {(12<<2)|2,{102,54,0}},
/* 38346 */ {(12<<2)|2,{102,54,0}},
/* 38347 */ {(12<<2)|2,{102,54,0}},
/* 38348 */ {(12<<2)|2,{102,54,0}},
/* 38349 */ {(12<<2)|2,{102,54,0}},
/* 38350 */ {(12<<2)|2,{102,54,0}},
/* 38351 */ {(12<<2)|2,{102,54,0}},
/* 38352 */ {(12<<2)|2,{102,55,0}},
/* 38353 */ {(12<<2)|2,{102,55,0}},
/* 38354 */ {(12<<2)|2,{102,55,0}},
/* 38355 */ {(12<<2)|2,{102,55,0}},
/* 38356 */ {(12<<2)|2,{102,55,0}},
/* 38357 */ {(12<<2)|2,{102,55,0}},
/* 38358 */ {(12<<2)|2,{102,55,0}},
/* 38359 */ {(12<<2)|2,{102,55,0}},
/* 38360 */ {(12<<2)|2,{102,55,0}},
/* 38361 */ {(12<<2)|2,{102,55,0}},
/* 38362 */ {(12<<2)|2,{102,55,0}},
/* 38363 */ {(12<<2)|2,{102,55,0}},
/* 38364 */ {(12<<2)|2,{102,55,0}},
/* 38365 */ {(12<<2)|2,{102,55,0}},
/* 38366 */ {(12<<2)|2,{102,55,0}},
/* 38367 */ {(12<<2)|2,{102,55,0}},
/* 38368 */ {(12<<2)|2,{102,56,0}},
/* 38369 */ {(12<<2)|2,{102,56,0}},
/* 38370 */ {(12<<2)|2,{102,56,0}},
/* 38371 */ {(12<<2)|2,{102,56,0}},
/* 38372 */ {(12<<2)|2,{102,56,0}},
/* 38373 */ {(12<<2)|2,{102,56,0}},
/* 38374 */ {(12<<2)|2,{102,56,0}},
/* 38375 */ {(12<<2)|2,{102,56,0}},
/* 38376 */ {(12<<2)|2,{102,56,0}},
/* 38377 */ {(12<<2)|2,{102,56,0}},
/* 38378 */ {(12<<2)|2,{102,56,0}},
/* 38379 */ {(12<<2)|2,{102,56,0}},
/* 38380 */ {(12<<2)|2,{102,56,0}},
/* 38381 */ {(12<<2)|2,{102,56,0}},
/* 38382 */ {(12<<2)|2,{102,56,0}},
/* 38383 */ {(12<<2)|2,{102,56,0}},
/* 38384 */ {(12<<2)|2,{102,57,0}},
/* 38385 */ {(12<<2)|2,{102,57,0}},
/* 38386 */ {(12<<2)|2,{102,57,0}},
/* 38387 */ {(12<<2)|2,{102,57,0}},
/* 38388 */ {(12<<2)|2,{102,57,0}},
/* 38389 */ {(12<<2)|2,{102,57,0}},
/* 38390 */ {(12<<2)|2,{102,57,0}},
/* 38391 */ {(12<<2)|2,{102,57,0}},
/* 38392 */ {(12<<2)|2,{102,57,0}},
/* 38393 */ {(12<<2)|2,{102,57,0}},
/* 38394 */ {(12<<2)|2,{102,57,0}},
/* 38395 */ {(12<<2)|2,{102,57,0}},
/* 38396 */ {(12<<2)|2,{102,57,0}},
/* 38397 */ {(12<<2)|2,{102,57,0}},
/* 38398 */ {(12<<2)|2,{102,57,0}},
/* 38399 */ {(12<<2)|2,{102,57,0}},
/* 38400 */ {(12<<2)|2,{102,61,0}},
/* 38401 */ {(12<<2)|2,{102,61,0}},
/* 38402 */ {(12<<2)|2,{102,61,0}},
/* 38403 */ {(12<<2)|2,{102,61,0}},
/* 38404 */ {(12<<2)|2,{102,61,0}},
/* 38405 */ {(12<<2)|2,{102,61,0}},
/* 38406 */ {(12<<2)|2,{102,61,0}},
/* 38407 */ {(12<<2)|2,{102,61,0}},
/* 38408 */ {(12<<2)|2,{102,61,0}},
/* 38409 */ {(12<<2)|2,{102,61,0}},
/* 38410 */ {(12<<2)|2,{102,61,0}},
/* 38411 */ {(12<<2)|2,{102,61,0}},
/* 38412 */ {(12<<2)|2,{102,61,0}},
/* 38413 */ {(12<<2)|2,{102,61,0}},
/* 38414 */ {(12<<2)|2,{102,61,0}},
/* 38415 */ {(12<<2)|2,{102,61,0}},
/* 38416 */ {(12<<2)|2,{102,65,0}},
/* 38417 */ {(12<<2)|2,{102,65,0}},
/* 38418 */ {(12<<2)|2,{102,65,0}},
/* 38419 */ {(12<<2)|2,{102,65,0}},
/* 38420 */ {(12<<2)|2,{102,65,0}},
/* 38421 */ {(12<<2)|2,{102,65,0}},
/* 38422 */ {(12<<2)|2,{102,65,0}},
/* 38423 */ {(12<<2)|2,{102,65,0}},
/* 38424 */ {(12<<2)|2,{102,65,0}},
/* 38425 */ {(12<<2)|2,{102,65,0}},
/* 38426 */ {(12<<2)|2,{102,65,0}},
/* 38427 */ {(12<<2)|2,{102,65,0}},
/* 38428 */ {(12<<2)|2,{102,65,0}},
/* 38429 */ {(12<<2)|2,{102,65,0}},
/* 38430 */ {(12<<2)|2,{102,65,0}},
/* 38431 */ {(12<<2)|2,{102,65,0}},
/* 38432 */ {(12<<2)|2,{102,95,0}},
/* 38433 */ {(12<<2)|2,{102,95,0}},
/* 38434 */ {(12<<2)|2,{102,95,0}},
/* 38435 */ {(12<<2)|2,{102,95,0}},
/* 38436 */ {(12<<2)|2,{102,95,0}},
/* 38437 */ {(12<<2)|2,{102,95,0}},
/* 38438 */ {(12<<2)|2,{102,95,0}},
/* 38439 */ {(12<<2)|2,{102,95,0}},
/* 38440 */ {(12<<2)|2,{102,95,0}},
/* 38441 */ {(12<<2)|2,{102,95,0}},
/* 38442 */ {(12<<2)|2,{102,95,0}},
/* 38443 */ {(12<<2)|2,{102,95,0}},
/* 38444 */ {(12<<2)|2,{102,95,0}},
/* 38445 */ {(12<<2)|2,{102,95,0}},
/* 38446 */ {(12<<2)|2,{102,95,0}},
/* 38447 */ {(12<<2)|2,{102,95,0}},
/* 38448 */ {(12<<2)|2,{102,98,0}},
/* 38449 */ {(12<<2)|2,{102,98,0}},
/* 38450 */ {(12<<2)|2,{102,98,0}},
/* 38451 */ {(12<<2)|2,{102,98,0}},
/* 38452 */ {(12<<2)|2,{102,98,0}},
/* 38453 */ {(12<<2)|2,{102,98,0}},
/* 38454 */ {(12<<2)|2,{102,98,0}},
/* 38455 */ {(12<<2)|2,{102,98,0}},
/* 38456 */ {(12<<2)|2,{102,98,0}},
/* 38457 */ {(12<<2)|2,{102,98,0}},
/* 38458 */ {(12<<2)|2,{102,98,0}},
/* 38459 */ {(12<<2)|2,{102,98,0}},
/* 38460 */ {(12<<2)|2,{102,98,0}},
/* 38461 */ {(12<<2)|2,{102,98,0}},
/* 38462 */ {(12<<2)|2,{102,98,0}},
/* 38463 */ {(12<<2)|2,{102,98,0}},
/* 38464 */ {(12<<2)|2,{102,100,0}},
/* 38465 */ {(12<<2)|2,{102,100,0}},
/* 38466 */ {(12<<2)|2,{102,100,0}},
/* 38467 */ {(12<<2)|2,{102,100,0}},
/* 38468 */ {(12<<2)|2,{102,100,0}},
/* 38469 */ {(12<<2)|2,{102,100,0}},
/* 38470 */ {(12<<2)|2,{102,100,0}},
/* 38471 */ {(12<<2)|2,{102,100,0}},
/* 38472 */ {(12<<2)|2,{102,100,0}},
/* 38473 */ {(12<<2)|2,{102,100,0}},
/* 38474 */ {(12<<2)|2,{102,100,0}},
/* 38475 */ {(12<<2)|2,{102,100,0}},
/* 38476 */ {(12<<2)|2,{102,100,0}},
/* 38477 */ {(12<<2)|2,{102,100,0}},
/* 38478 */ {(12<<2)|2,{102,100,0}},
/* 38479 */ {(12<<2)|2,{102,100,0}},
/* 38480 */ {(12<<2)|2,{102,102,0}},
/* 38481 */ {(12<<2)|2,{102,102,0}},
/* 38482 */ {(12<<2)|2,{102,102,0}},
/* 38483 */ {(12<<2)|2,{102,102,0}},
/* 38484 */ {(12<<2)|2,{102,102,0}},
/* 38485 */ {(12<<2)|2,{102,102,0}},
/* 38486 */ {(12<<2)|2,{102,102,0}},
/* 38487 */ {(12<<2)|2,{102,102,0}},
/* 38488 */ {(12<<2)|2,{102,102,0}},
/* 38489 */ {(12<<2)|2,{102,102,0}},
/* 38490 */ {(12<<2)|2,{102,102,0}},
/* 38491 */ {(12<<2)|2,{102,102,0}},
/* 38492 */ {(12<<2)|2,{102,102,0}},
/* 38493 */ {(12<<2)|2,{102,102,0}},
/* 38494 */ {(12<<2)|2,{102,102,0}},
/* 38495 */ {(12<<2)|2,{102,102,0}},
/* 38496 */ {(12<<2)|2,{102,103,0}},
/* 38497 */ {(12<<2)|2,{102,103,0}},
/* 38498 */ {(12<<2)|2,{102,103,0}},
/* 38499 */ {(12<<2)|2,{102,103,0}},
/* 38500 */ {(12<<2)|2,{102,103,0}},
/* 38501 */ {(12<<2)|2,{102,103,0}},
/* 38502 */ {(12<<2)|2,{102,103,0}},
/* 38503 */ {(12<<2)|2,{102,103,0}},
/* 38504 */ {(12<<2)|2,{102,103,0}},
/* 38505 */ {(12<<2)|2,{102,103,0}},
/* 38506 */ {(12<<2)|2,{102,103,0}},
/* 38507 */ {(12<<2)|2,{102,103,0}},
/* 38508 */ {(12<<2)|2,{102,103,0}},
/* 38509 */ {(12<<2)|2,{102,103,0}},
/* 38510 */ {(12<<2)|2,{102,103,0}},
/* 38511 */ {(12<<2)|2,{102,103,0}},
/* 38512 */ {(12<<2)|2,{102,104,0}},
/* 38513 */ {(12<<2)|2,{102,104,0}},
/* 38514 */ {(12<<2)|2,{102,104,0}},
/* 38515 */ {(12<<2)|2,{102,104,0}},
/* 38516 */ {(12<<2)|2,{102,104,0}},
/* 38517 */ {(12<<2)|2,{102,104,0}},
/* 38518 */ {(12<<2)|2,{102,104,0}},
/* 38519 */ {(12<<2)|2,{102,104,0}},
/* 38520 */ {(12<<2)|2,{102,104,0}},
/* 38521 */ {(12<<2)|2,{102,104,0}},
/* 38522 */ {(12<<2)|2,{102,104,0}},
/* 38523 */ {(12<<2)|2,{102,104,0}},
/* 38524 */ {(12<<2)|2,{102,104,0}},
/* 38525 */ {(12<<2)|2,{102,104,0}},
/* 38526 */ {(12<<2)|2,{102,104,0}},
/* 38527 */ {(12<<2)|2,{102,104,0}},
/* 38528 */ {(12<<2)|2,{102,108,0}},
/* 38529 */ {(12<<2)|2,{102,108,0}},
/* 38530 */ {(12<<2)|2,{102,108,0}},
/* 38531 */ {(12<<2)|2,{102,108,0}},
/* 38532 */ {(12<<2)|2,{102,108,0}},
/* 38533 */ {(12<<2)|2,{102,108,0}},
/* 38534 */ {(12<<2)|2,{102,108,0}},
/* 38535 */ {(12<<2)|2,{102,108,0}},
/* 38536 */ {(12<<2)|2,{102,108,0}},
/* 38537 */ {(12<<2)|2,{102,108,0}},
/* 38538 */ {(12<<2)|2,{102,108,0}},
/* 38539 */ {(12<<2)|2,{102,108,0}},
/* 38540 */ {(12<<2)|2,{102,108,0}},
/* 38541 */ {(12<<2)|2,{102,108,0}},
/* 38542 */ {(12<<2)|2,{102,108,0}},
/* 38543 */ {(12<<2)|2,{102,108,0}},
/* 38544 */ {(12<<2)|2,{102,109,0}},
/* 38545 */ {(12<<2)|2,{102,109,0}},
/* 38546 */ {(12<<2)|2,{102,109,0}},
/* 38547 */ {(12<<2)|2,{102,109,0}},
/* 38548 */ {(12<<2)|2,{102,109,0}},
/* 38549 */ {(12<<2)|2,{102,109,0}},
/* 38550 */ {(12<<2)|2,{102,109,0}},
/* 38551 */ {(12<<2)|2,{102,109,0}},
/* 38552 */ {(12<<2)|2,{102,109,0}},
/* 38553 */ {(12<<2)|2,{102,109,0}},
/* 38554 */ {(12<<2)|2,{102,109,0}},
/* 38555 */ {(12<<2)|2,{102,109,0}},
/* 38556 */ {(12<<2)|2,{102,109,0}},
/* 38557 */ {(12<<2)|2,{102,109,0}},
/* 38558 */ {(12<<2)|2,{102,109,0}},
/* 38559 */ {(12<<2)|2,{102,109,0}},
/* 38560 */ {(12<<2)|2,{102,110,0}},
/* 38561 */ {(12<<2)|2,{102,110,0}},
/* 38562 */ {(12<<2)|2,{102,110,0}},
/* 38563 */ {(12<<2)|2,{102,110,0}},
/* 38564 */ {(12<<2)|2,{102,110,0}},
/* 38565 */ {(12<<2)|2,{102,110,0}},
/* 38566 */ {(12<<2)|2,{102,110,0}},
/* 38567 */ {(12<<2)|2,{102,110,0}},
/* 38568 */ {(12<<2)|2,{102,110,0}},
/* 38569 */ {(12<<2)|2,{102,110,0}},
/* 38570 */ {(12<<2)|2,{102,110,0}},
/* 38571 */ {(12<<2)|2,{102,110,0}},
/* 38572 */ {(12<<2)|2,{102,110,0}},
/* 38573 */ {(12<<2)|2,{102,110,0}},
/* 38574 */ {(12<<2)|2,{102,110,0}},
/* 38575 */ {(12<<2)|2,{102,110,0}},
/* 38576 */ {(12<<2)|2,{102,112,0}},
/* 38577 */ {(12<<2)|2,{102,112,0}},
/* 38578 */ {(12<<2)|2,{102,112,0}},
/* 38579 */ {(12<<2)|2,{102,112,0}},
/* 38580 */ {(12<<2)|2,{102,112,0}},
/* 38581 */ {(12<<2)|2,{102,112,0}},
/* 38582 */ {(12<<2)|2,{102,112,0}},
/* 38583 */ {(12<<2)|2,{102,112,0}},
/* 38584 */ {(12<<2)|2,{102,112,0}},
/* 38585 */ {(12<<2)|2,{102,112,0}},
/* 38586 */ {(12<<2)|2,{102,112,0}},
/* 38587 */ {(12<<2)|2,{102,112,0}},
/* 38588 */ {(12<<2)|2,{102,112,0}},
/* 38589 */ {(12<<2)|2,{102,112,0}},
/* 38590 */ {(12<<2)|2,{102,112,0}},
/* 38591 */ {(12<<2)|2,{102,112,0}},
/* 38592 */ {(12<<2)|2,{102,114,0}},
/* 38593 */ {(12<<2)|2,{102,114,0}},
/* 38594 */ {(12<<2)|2,{102,114,0}},
/* 38595 */ {(12<<2)|2,{102,114,0}},
/* 38596 */ {(12<<2)|2,{102,114,0}},
/* 38597 */ {(12<<2)|2,{102,114,0}},
/* 38598 */ {(12<<2)|2,{102,114,0}},
/* 38599 */ {(12<<2)|2,{102,114,0}},
/* 38600 */ {(12<<2)|2,{102,114,0}},
/* 38601 */ {(12<<2)|2,{102,114,0}},
/* 38602 */ {(12<<2)|2,{102,114,0}},
/* 38603 */ {(12<<2)|2,{102,114,0}},
/* 38604 */ {(12<<2)|2,{102,114,0}},
/* 38605 */ {(12<<2)|2,{102,114,0}},
/* 38606 */ {(12<<2)|2,{102,114,0}},
/* 38607 */ {(12<<2)|2,{102,114,0}},
/* 38608 */ {(12<<2)|2,{102,117,0}},
/* 38609 */ {(12<<2)|2,{102,117,0}},
/* 38610 */ {(12<<2)|2,{102,117,0}},
/* 38611 */ {(12<<2)|2,{102,117,0}},
/* 38612 */ {(12<<2)|2,{102,117,0}},
/* 38613 */ {(12<<2)|2,{102,117,0}},
/* 38614 */ {(12<<2)|2,{102,117,0}},
/* 38615 */ {(12<<2)|2,{102,117,0}},
/* 38616 */ {(12<<2)|2,{102,117,0}},
/* 38617 */ {(12<<2)|2,{102,117,0}},
/* 38618 */ {(12<<2)|2,{102,117,0}},
/* 38619 */ {(12<<2)|2,{102,117,0}},
/* 38620 */ {(12<<2)|2,{102,117,0}},
/* 38621 */ {(12<<2)|2,{102,117,0}},
/* 38622 */ {(12<<2)|2,{102,117,0}},
/* 38623 */ {(12<<2)|2,{102,117,0}},
/* 38624 */ {(13<<2)|2,{102,58,0}},
/* 38625 */ {(13<<2)|2,{102,58,0}},
/* 38626 */ {(13<<2)|2,{102,58,0}},
/* 38627 */ {(13<<2)|2,{102,58,0}},
/* 38628 */ {(13<<2)|2,{102,58,0}},
/* 38629 */ {(13<<2)|2,{102,58,0}},
/* 38630 */ {(13<<2)|2,{102,58,0}},
/* 38631 */ {(13<<2)|2,{102,58,0}},
/* 38632 */ {(13<<2)|2,{102,66,0}},
/* 38633 */ {(13<<2)|2,{102,66,0}},
/* 38634 */ {(13<<2)|2,{102,66,0}},
/* 38635 */ {(13<<2)|2,{102,66,0}},
/* 38636 */ {(13<<2)|2,{102,66,0}},
/* 38637 */ {(13<<2)|2,{102,66,0}},
/* 38638 */ {(13<<2)|2,{102,66,0}},
/* 38639 */ {(13<<2)|2,{102,66,0}},
/* 38640 */ {(13<<2)|2,{102,67,0}},
/* 38641 */ {(13<<2)|2,{102,67,0}},
/* 38642 */ {(13<<2)|2,{102,67,0}},
/* 38643 */ {(13<<2)|2,{102,67,0}},
/* 38644 */ {(13<<2)|2,{102,67,0}},
/* 38645 */ {(13<<2)|2,{102,67,0}},
/* 38646 */ {(13<<2)|2,{102,67,0}},
/* 38647 */ {(13<<2)|2,{102,67,0}},
/* 38648 */ {(13<<2)|2,{102,68,0}},
/* 38649 */ {(13<<2)|2,{102,68,0}},
/* 38650 */ {(13<<2)|2,{102,68,0}},
/* 38651 */ {(13<<2)|2,{102,68,0}},
/* 38652 */ {(13<<2)|2,{102,68,0}},
/* 38653 */ {(13<<2)|2,{102,68,0}},
/* 38654 */ {(13<<2)|2,{102,68,0}},
/* 38655 */ {(13<<2)|2,{102,68,0}},
/* 38656 */ {(13<<2)|2,{102,69,0}},
/* 38657 */ {(13<<2)|2,{102,69,0}},
/* 38658 */ {(13<<2)|2,{102,69,0}},
/* 38659 */ {(13<<2)|2,{102,69,0}},
/* 38660 */ {(13<<2)|2,{102,69,0}},
/* 38661 */ {(13<<2)|2,{102,69,0}},
/* 38662 */ {(13<<2)|2,{102,69,0}},
/* 38663 */ {(13<<2)|2,{102,69,0}},
/* 38664 */ {(13<<2)|2,{102,70,0}},
/* 38665 */ {(13<<2)|2,{102,70,0}},
/* 38666 */ {(13<<2)|2,{102,70,0}},
/* 38667 */ {(13<<2)|2,{102,70,0}},
/* 38668 */ {(13<<2)|2,{102,70,0}},
/* 38669 */ {(13<<2)|2,{102,70,0}},
/* 38670 */ {(13<<2)|2,{102,70,0}},
/* 38671 */ {(13<<2)|2,{102,70,0}},
/* 38672 */ {(13<<2)|2,{102,71,0}},
/* 38673 */ {(13<<2)|2,{102,71,0}},
/* 38674 */ {(13<<2)|2,{102,71,0}},
/* 38675 */ {(13<<2)|2,{102,71,0}},
/* 38676 */ {(13<<2)|2,{102,71,0}},
/* 38677 */ {(13<<2)|2,{102,71,0}},
/* 38678 */ {(13<<2)|2,{102,71,0}},
/* 38679 */ {(13<<2)|2,{102,71,0}},
/* 38680 */ {(13<<2)|2,{102,72,0}},
/* 38681 */ {(13<<2)|2,{102,72,0}},
/* 38682 */ {(13<<2)|2,{102,72,0}},
/* 38683 */ {(13<<2)|2,{102,72,0}},
/* 38684 */ {(13<<2)|2,{102,72,0}},
/* 38685 */ {(13<<2)|2,{102,72,0}},
/* 38686 */ {(13<<2)|2,{102,72,0}},
/* 38687 */ {(13<<2)|2,{102,72,0}},
/* 38688 */ {(13<<2)|2,{102,73,0}},
/* 38689 */ {(13<<2)|2,{102,73,0}},
/* 38690 */ {(13<<2)|2,{102,73,0}},
/* 38691 */ {(13<<2)|2,{102,73,0}},
/* 38692 */ {(13<<2)|2,{102,73,0}},
/* 38693 */ {(13<<2)|2,{102,73,0}},
/* 38694 */ {(13<<2)|2,{102,73,0}},
/* 38695 */ {(13<<2)|2,{102,73,0}},
/* 38696 */ {(13<<2)|2,{102,74,0}},
/* 38697 */ {(13<<2)|2,{102,74,0}},
/* 38698 */ {(13<<2)|2,{102,74,0}},
/* 38699 */ {(13<<2)|2,{102,74,0}},
/* 38700 */ {(13<<2)|2,{102,74,0}},
/* 38701 */ {(13<<2)|2,{102,74,0}},
/* 38702 */ {(13<<2)|2,{102,74,0}},
/* 38703 */ {(13<<2)|2,{102,74,0}},
/* 38704 */ {(13<<2)|2,{102,75,0}},
/* 38705 */ {(13<<2)|2,{102,75,0}},
/* 38706 */ {(13<<2)|2,{102,75,0}},
/* 38707 */ {(13<<2)|2,{102,75,0}},
/* 38708 */ {(13<<2)|2,{102,75,0}},
/* 38709 */ {(13<<2)|2,{102,75,0}},
/* 38710 */ {(13<<2)|2,{102,75,0}},
/* 38711 */ {(13<<2)|2,{102,75,0}},
/* 38712 */ {(13<<2)|2,{102,76,0}},
/* 38713 */ {(13<<2)|2,{102,76,0}},
/* 38714 */ {(13<<2)|2,{102,76,0}},
/* 38715 */ {(13<<2)|2,{102,76,0}},
/* 38716 */ {(13<<2)|2,{102,76,0}},
/* 38717 */ {(13<<2)|2,{102,76,0}},
/* 38718 */ {(13<<2)|2,{102,76,0}},
/* 38719 */ {(13<<2)|2,{102,76,0}},
/* 38720 */ {(13<<2)|2,{102,77,0}},
/* 38721 */ {(13<<2)|2,{102,77,0}},
/* 38722 */ {(13<<2)|2,{102,77,0}},
/* 38723 */ {(13<<2)|2,{102,77,0}},
/* 38724 */ {(13<<2)|2,{102,77,0}},
/* 38725 */ {(13<<2)|2,{102,77,0}},
/* 38726 */ {(13<<2)|2,{102,77,0}},
/* 38727 */ {(13<<2)|2,{102,77,0}},
/* 38728 */ {(13<<2)|2,{102,78,0}},
/* 38729 */ {(13<<2)|2,{102,78,0}},
/* 38730 */ {(13<<2)|2,{102,78,0}},
/* 38731 */ {(13<<2)|2,{102,78,0}},
/* 38732 */ {(13<<2)|2,{102,78,0}},
/* 38733 */ {(13<<2)|2,{102,78,0}},
/* 38734 */ {(13<<2)|2,{102,78,0}},
/* 38735 */ {(13<<2)|2,{102,78,0}},
/* 38736 */ {(13<<2)|2,{102,79,0}},
/* 38737 */ {(13<<2)|2,{102,79,0}},
/* 38738 */ {(13<<2)|2,{102,79,0}},
/* 38739 */ {(13<<2)|2,{102,79,0}},
/* 38740 */ {(13<<2)|2,{102,79,0}},
/* 38741 */ {(13<<2)|2,{102,79,0}},
/* 38742 */ {(13<<2)|2,{102,79,0}},
/* 38743 */ {(13<<2)|2,{102,79,0}},
/* 38744 */ {(13<<2)|2,{102,80,0}},
/* 38745 */ {(13<<2)|2,{102,80,0}},
/* 38746 */ {(13<<2)|2,{102,80,0}},
/* 38747 */ {(13<<2)|2,{102,80,0}},
/* 38748 */ {(13<<2)|2,{102,80,0}},
/* 38749 */ {(13<<2)|2,{102,80,0}},
/* 38750 */ {(13<<2)|2,{102,80,0}},
/* 38751 */ {(13<<2)|2,{102,80,0}},
/* 38752 */ {(13<<2)|2,{102,81,0}},
/* 38753 */ {(13<<2)|2,{102,81,0}},
/* 38754 */ {(13<<2)|2,{102,81,0}},
/* 38755 */ {(13<<2)|2,{102,81,0}},
/* 38756 */ {(13<<2)|2,{102,81,0}},
/* 38757 */ {(13<<2)|2,{102,81,0}},
/* 38758 */ {(13<<2)|2,{102,81,0}},
/* 38759 */ {(13<<2)|2,{102,81,0}},
/* 38760 */ {(13<<2)|2,{102,82,0}},
/* 38761 */ {(13<<2)|2,{102,82,0}},
/* 38762 */ {(13<<2)|2,{102,82,0}},
/* 38763 */ {(13<<2)|2,{102,82,0}},
/* 38764 */ {(13<<2)|2,{102,82,0}},
/* 38765 */ {(13<<2)|2,{102,82,0}},
/* 38766 */ {(13<<2)|2,{102,82,0}},
/* 38767 */ {(13<<2)|2,{102,82,0}},
/* 38768 */ {(13<<2)|2,{102,83,0}},
/* 38769 */ {(13<<2)|2,{102,83,0}},
/* 38770 */ {(13<<2)|2,{102,83,0}},
/* 38771 */ {(13<<2)|2,{102,83,0}},
/* 38772 */ {(13<<2)|2,{102,83,0}},
/* 38773 */ {(13<<2)|2,{102,83,0}},
/* 38774 */ {(13<<2)|2,{102,83,0}},
/* 38775 */ {(13<<2)|2,{102,83,0}},
/* 38776 */ {(13<<2)|2,{102,84,0}},
/* 38777 */ {(13<<2)|2,{102,84,0}},
/* 38778 */ {(13<<2)|2,{102,84,0}},
/* 38779 */ {(13<<2)|2,{102,84,0}},
/* 38780 */ {(13<<2)|2,{102,84,0}},
/* 38781 */ {(13<<2)|2,{102,84,0}},
/* 38782 */ {(13<<2)|2,{102,84,0}},
/* 38783 */ {(13<<2)|2,{102,84,0}},
/* 38784 */ {(13<<2)|2,{102,85,0}},
/* 38785 */ {(13<<2)|2,{102,85,0}},
/* 38786 */ {(13<<2)|2,{102,85,0}},
/* 38787 */ {(13<<2)|2,{102,85,0}},
/* 38788 */ {(13<<2)|2,{102,85,0}},
/* 38789 */ {(13<<2)|2,{102,85,0}},
/* 38790 */ {(13<<2)|2,{102,85,0}},
/* 38791 */ {(13<<2)|2,{102,85,0}},
/* 38792 */ {(13<<2)|2,{102,86,0}},
/* 38793 */ {(13<<2)|2,{102,86,0}},
/* 38794 */ {(13<<2)|2,{102,86,0}},
/* 38795 */ {(13<<2)|2,{102,86,0}},
/* 38796 */ {(13<<2)|2,{102,86,0}},
/* 38797 */ {(13<<2)|2,{102,86,0}},
/* 38798 */ {(13<<2)|2,{102,86,0}},
/* 38799 */ {(13<<2)|2,{102,86,0}},
/* 38800 */ {(13<<2)|2,{102,87,0}},
/* 38801 */ {(13<<2)|2,{102,87,0}},
/* 38802 */ {(13<<2)|2,{102,87,0}},
/* 38803 */ {(13<<2)|2,{102,87,0}},
/* 38804 */ {(13<<2)|2,{102,87,0}},
/* 38805 */ {(13<<2)|2,{102,87,0}},
/* 38806 */ {(13<<2)|2,{102,87,0}},
/* 38807 */ {(13<<2)|2,{102,87,0}},
/* 38808 */ {(13<<2)|2,{102,89,0}},
/* 38809 */ {(13<<2)|2,{102,89,0}},
/* 38810 */ {(13<<2)|2,{102,89,0}},
/* 38811 */ {(13<<2)|2,{102,89,0}},
/* 38812 */ {(13<<2)|2,{102,89,0}},
/* 38813 */ {(13<<2)|2,{102,89,0}},
/* 38814 */ {(13<<2)|2,{102,89,0}},
/* 38815 */ {(13<<2)|2,{102,89,0}},
/* 38816 */ {(13<<2)|2,{102,106,0}},
/* 38817 */ {(13<<2)|2,{102,106,0}},
/* 38818 */ {(13<<2)|2,{102,106,0}},
/* 38819 */ {(13<<2)|2,{102,106,0}},
/* 38820 */ {(13<<2)|2,{102,106,0}},
/* 38821 */ {(13<<2)|2,{102,106,0}},
/* 38822 */ {(13<<2)|2,{102,106,0}},
/* 38823 */ {(13<<2)|2,{102,106,0}},
/* 38824 */ {(13<<2)|2,{102,107,0}},
/* 38825 */ {(13<<2)|2,{102,107,0}},
/* 38826 */ {(13<<2)|2,{102,107,0}},
/* 38827 */ {(13<<2)|2,{102,107,0}},
/* 38828 */ {(13<<2)|2,{102,107,0}},
/* 38829 */ {(13<<2)|2,{102,107,0}},
/* 38830 */ {(13<<2)|2,{102,107,0}},
/* 38831 */ {(13<<2)|2,{102,107,0}},
/* 38832 */ {(13<<2)|2,{102,113,0}},
/* 38833 */ {(13<<2)|2,{102,113,0}},
/* 38834 */ {(13<<2)|2,{102,113,0}},
/* 38835 */ {(13<<2)|2,{102,113,0}},
/* 38836 */ {(13<<2)|2,{102,113,0}},
/* 38837 */ {(13<<2)|2,{102,113,0}},
/* 38838 */ {(13<<2)|2,{102,113,0}},
/* 38839 */ {(13<<2)|2,{102,113,0}},
/* 38840 */ {(13<<2)|2,{102,118,0}},
/* 38841 */ {(13<<2)|2,{102,118,0}},
/* 38842 */ {(13<<2)|2,{102,118,0}},
/* 38843 */ {(13<<2)|2,{102,118,0}},
/* 38844 */ {(13<<2)|2,{102,118,0}},
/* 38845 */ {(13<<2)|2,{102,118,0}},
/* 38846 */ {(13<<2)|2,{102,118,0}},
/* 38847 */ {(13<<2)|2,{102,118,0}},
/* 38848 */ {(13<<2)|2,{102,119,0}},
/* 38849 */ {(13<<2)|2,{102,119,0}},
/* 38850 */ {(13<<2)|2,{102,119,0}},
/* 38851 */ {(13<<2)|2,{102,119,0}},
/* 38852 */ {(13<<2)|2,{102,119,0}},
/* 38853 */ {(13<<2)|2,{102,119,0}},
/* 38854 */ {(13<<2)|2,{102,119,0}},
/* 38855 */ {(13<<2)|2,{102,119,0}},
/* 38856 */ {(13<<2)|2,{102,120,0}},
/* 38857 */ {(13<<2)|2,{102,120,0}},
/* 38858 */ {(13<<2)|2,{102,120,0}},
/* 38859 */ {(13<<2)|2,{102,120,0}},
/* 38860 */ {(13<<2)|2,{102,120,0}},
/* 38861 */ {(13<<2)|2,{102,120,0}},
/* 38862 */ {(13<<2)|2,{102,120,0}},
/* 38863 */ {(13<<2)|2,{102,120,0}},
/* 38864 */ {(13<<2)|2,{102,121,0}},
/* 38865 */ {(13<<2)|2,{102,121,0}},
/* 38866 */ {(13<<2)|2,{102,121,0}},
/* 38867 */ {(13<<2)|2,{102,121,0}},
/* 38868 */ {(13<<2)|2,{102,121,0}},
/* 38869 */ {(13<<2)|2,{102,121,0}},
/* 38870 */ {(13<<2)|2,{102,121,0}},
/* 38871 */ {(13<<2)|2,{102,121,0}},
/* 38872 */ {(13<<2)|2,{102,122,0}},
/* 38873 */ {(13<<2)|2,{102,122,0}},
/* 38874 */ {(13<<2)|2,{102,122,0}},
/* 38875 */ {(13<<2)|2,{102,122,0}},
/* 38876 */ {(13<<2)|2,{102,122,0}},
/* 38877 */ {(13<<2)|2,{102,122,0}},
/* 38878 */ {(13<<2)|2,{102,122,0}},
/* 38879 */ {(13<<2)|2,{102,122,0}},
/* 38880 */ {(14<<2)|2,{102,38,0}},
/* 38881 */ {(14<<2)|2,{102,38,0}},
/* 38882 */ {(14<<2)|2,{102,38,0}},
/* 38883 */ {(14<<2)|2,{102,38,0}},
/* 38884 */ {(14<<2)|2,{102,42,0}},
/* 38885 */ {(14<<2)|2,{102,42,0}},
/* 38886 */ {(14<<2)|2,{102,42,0}},
/* 38887 */ {(14<<2)|2,{102,42,0}},
/* 38888 */ {(14<<2)|2,{102,44,0}},
/* 38889 */ {(14<<2)|2,{102,44,0}},
/* 38890 */ {(14<<2)|2,{102,44,0}},
/* 38891 */ {(14<<2)|2,{102,44,0}},
/* 38892 */ {(14<<2)|2,{102,59,0}},
/* 38893 */ {(14<<2)|2,{102,59,0}},
/* 38894 */ {(14<<2)|2,{102,59,0}},
/* 38895 */ {(14<<2)|2,{102,59,0}},
/* 38896 */ {(14<<2)|2,{102,88,0}},
/* 38897 */ {(14<<2)|2,{102,88,0}},
/* 38898 */ {(14<<2)|2,{102,88,0}},
/* 38899 */ {(14<<2)|2,{102,88,0}},
/* 38900 */ {(14<<2)|2,{102,90,0}},
/* 38901 */ {(14<<2)|2,{102,90,0}},
/* 38902 */ {(14<<2)|2,{102,90,0}},
/* 38903 */ {(14<<2)|2,{102,90,0}},
/* 38904 */ {(16<<2)|2,{102,33,0}},
/* 38905 */ {(16<<2)|2,{102,34,0}},
/* 38906 */ {(16<<2)|2,{102,40,0}},
/* 38907 */ {(16<<2)|2,{102,41,0}},
/* 38908 */ {(16<<2)|2,{102,63,0}},
/* 38909 */ {(6<<2)|1,{102,0,0}},
/* 38910 */ {(6<<2)|1,{102,0,0}},
/* 38911 */ {(6<<2)|1,{102,0,0}},
/* 38912 */ {(16<<2)|3,{103,48,48}},
/* 38913 */ {(16<<2)|3,{103,48,49}},
/* 38914 */ {(16<<2)|3,{103,48,50}},
/* 38915 */ {(16<<2)|3,{103,48,97}},
/* 38916 */ {(16<<2)|3,{103,48,99}},
/* 38917 */ {(16<<2)|3,{103,48,101}},
/* 38918 */ {(16<<2)|3,{103,48,105}},
/* 38919 */ {(16<<2)|3,{103,48,111}},
/* 38920 */ {(16<<2)|3,{103,48,115}},
/* 38921 */ {(16<<2)|3,{103,48,116}},
/* 38922 */ {(11<<2)|2,{103,48,0}},
/* 38923 */ {(11<<2)|2,{103,48,0}},
/* 38924 */ {(11<<2)|2,{103,48,0}},
/* 38925 */ {(11<<2)|2,{103,48,0}},
/* 38926 */ {(11<<2)|2,{103,48,0}},
/* 38927 */ {(11<<2)|2,{103,48,0}},
/* 38928 */ {(11<<2)|2,{103,48,0}},
/* 38929 */ {(11<<2)|2,{103,48,0}},
/* 38930 */ {(11<<2)|2,{103,48,0}},
/* 38931 */ {(11<<2)|2,{103,48,0}},
/* 38932 */ {(11<<2)|2,{103,48,0}},
/* 38933 */ {(11<<2)|2,{103,48,0}},
/* 38934 */ {(11<<2)|2,{103,48,0}},
/* 38935 */ {(11<<2)|2,{103,48,0}},
/* 38936 */ {(11<<2)|2,{103,48,0}},
/* 38937 */ {(11<<2)|2,{103,48,0}},
/* 38938 */ {(11<<2)|2,{103,48,0}},
/* 38939 */ {(11<<2)|2,{103,48,0}},
/* 38940 */ {(11<<2)|2,{103,48,0}},
/* 38941 */ {(11<<2)|2,{103,48,0}},
/* 38942 */ {(11<<2)|2,{103,48,0}},
/* 38943 */ {(11<<2)|2,{103,48,0}},
/* 38944 */ {(16<<2)|3,{103,49,48}},
/* 38945 */ {(16<<2)|3,{103,49,49}},
/* 38946 */ {(16<<2)|3,{103,49,50}},
/* 38947 */ {(16<<2)|3,{103,49,97}},
/* 38948 */ {(16<<2)|3,{103,49,99}},
/* 38949 */ {(16<<2)|3,{103,49,101}},
/* 38950 */ {(16<<2)|3,{103,49,105}},
/* 38951 */ {(16<<2)|3,{103,49,111}},
/* 38952 */ {(16<<2)|3,{103,49,115}},
/* 38953 */ {(16<<2)|3,{103,49,116}},
/* 38954 */ {(11<<2)|2,{103,49,0}},
/* 38955 */ {(11<<2)|2,{103,49,0}},
/* 38956 */ {(11<<2)|2,{103,49,0}},
/* 38957 */ {(11<<2)|2,{103,49,0}},
/* 38958 */ {(11<<2)|2,{103,49,0}},
/* 38959 */ {(11<<2)|2,{103,49,0}},
/* 38960 */ {(11<<2)|2,{103,49,0}},
/* 38961 */ {(11<<2)|2,{103,49,0}},
/* 38962 */ {(11<<2)|2,{103,49,0}},
/* 38963 */ {(11<<2)|2,{103,49,0}},
/* 38964 */ {(11<<2)|2,{103,49,0}},
/* 38965 */ {(11<<2)|2,{103,49,0}},
/* 38966 */ {(11<<2)|2,{103,49,0}},
/* 38967 */ {(11<<2)|2,{103,49,0}},
/* 38968 */ {(11<<2)|2,{103,49,0}},
/* 38969 */ {(11<<2)|2,{103,49,0}},
/* 38970 */ {(11<<2)|2,{103,49,0}},
/* 38971 */ {(11<<2)|2,{103,49,0}},
/* 38972 */ {(11<<2)|2,{103,49,0}},
/* 38973 */ {(11<<2)|2,{103,49,0}},
/* 38974 */ {(11<<2)|2,{103,49,0}},
/* 38975 */ {(11<<2)|2,{103,49,0}},
/* 38976 */ {(16<<2)|3,{103,50,48}},
/* 38977 */ {(16<<2)|3,{103,50,49}},
/* 38978 */ {(16<<2)|3,{103,50,50}},
/* 38979 */ {(16<<2)|3,{103,50,97}},
/* 38980 */ {(16<<2)|3,{103,50,99}},
/* 38981 */ {(16<<2)|3,{103,50,101}},
/* 38982 */ {(16<<2)|3,{103,50,105}},
/* 38983 */ {(16<<2)|3,{103,50,111}},
/* 38984 */ {(16<<2)|3,{103,50,115}},
/* 38985 */ {(16<<2)|3,{103,50,116}},
/* 38986 */ {(11<<2)|2,{103,50,0}},
/* 38987 */ {(11<<2)|2,{103,50,0}},
/* 38988 */ {(11<<2)|2,{103,50,0}},
/* 38989 */ {(11<<2)|2,{103,50,0}},
/* 38990 */ {(11<<2)|2,{103,50,0}},
/* 38991 */ {(11<<2)|2,{103,50,0}},
/* 38992 */ {(11<<2)|2,{103,50,0}},
/* 38993 */ {(11<<2)|2,{103,50,0}},
/* 38994 */ {(11<<2)|2,{103,50,0}},
/* 38995 */ {(11<<2)|2,{103,50,0}},
/* 38996 */ {(11<<2)|2,{103,50,0}},
/* 38997 */ {(11<<2)|2,{103,50,0}},
/* 38998 */ {(11<<2)|2,{103,50,0}},
/* 38999 */ {(11<<2)|2,{103,50,0}},
/* 39000 */ {(11<<2)|2,{103,50,0}},
/* 39001 */ {(11<<2)|2,{103,50,0}},
/* 39002 */ {(11<<2)|2,{103,50,0}},
/* 39003 */ {(11<<2)|2,{103,50,0}},
/* 39004 */ {(11<<2)|2,{103,50,0}},
/* 39005 */ {(11<<2)|2,{103,50,0}},
/* 39006 */ {(11<<2)|2,{103,50,0}},
/* 39007 */ {(11<<2)|2,{103,50,0}},
/* 39008 */ {(16<<2)|3,{103,97,48}},
/* 39009 */ {(16<<2)|3,{103,97,49}},
/* 39010 */ {(16<<2)|3,{103,97,50}},
/* 39011 */ {(16<<2)|3,{103,97,97}},
/* 39012 */ {(16<<2)|3,{103,97,99}},
/* 39013 */ {(16<<2)|3,{103,97,101}},
/* 39014 */ {(16<<2)|3,{103,97,105}},
/* 39015 */ {(16<<2)|3,{103,97,111}},
/* 39016 */ {(16<<2)|3,{103,97,115}},
/* 39017 */ {(16<<2)|3,{103,97,116}},
/* 39018 */ {(11<<2)|2,{103,97,0}},
/* 39019 */ {(11<<2)|2,{103,97,0}},
/* 39020 */ {(11<<2)|2,{103,97,0}},
/* 39021 */ {(11<<2)|2,{103,97,0}},
/* 39022 */ {(11<<2)|2,{103,97,0}},
/* 39023 */ {(11<<2)|2,{103,97,0}},
/* 39024 */ {(11<<2)|2,{103,97,0}},
/* 39025 */ {(11<<2)|2,{103,97,0}},
/* 39026 */ {(11<<2)|2,{103,97,0}},
/* 39027 */ {(11<<2)|2,{103,97,0}},
/* 39028 */ {(11<<2)|2,{103,97,0}},
/* 39029 */ {(11<<2)|2,{103,97,0}},
/* 39030 */ {(11<<2)|2,{103,97,0}},
/* 39031 */ {(11<<2)|2,{103,97,0}},
/* 39032 */ {(11<<2)|2,{103,97,0}},
/* 39033 */ {(11<<2)|2,{103,97,0}},
/* 39034 */ {(11<<2)|2,{103,97,0}},
/* 39035 */ {(11<<2)|2,{103,97,0}},
/* 39036 */ {(11<<2)|2,{103,97,0}},
/* 39037 */ {(11<<2)|2,{103,97,0}},
/* 39038 */ {(11<<2)|2,{103,97,0}},
/* 39039 */ {(11<<2)|2,{103,97,0}},
/* 39040 */ {(16<<2)|3,{103,99,48}},
/* 39041 */ {(16<<2)|3,{103,99,49}},
/* 39042 */ {(16<<2)|3,{103,99,50}},
/* 39043 */ {(16<<2)|3,{103,99,97}},
/* 39044 */ {(16<<2)|3,{103,99,99}},
/* 39045 */ {(16<<2)|3,{103,99,101}},
/* 39046 */ {(16<<2)|3,{103,99,105}},
/* 39047 */ {(16<<2)|3,{103,99,111}},
/* 39048 */ {(16<<2)|3,{103,99,115}},
/* 39049 */ {(16<<2)|3,{103,99,116}},
/* 39050 */ {(11<<2)|2,{103,99,0}},
/* 39051 */ {(11<<2)|2,{103,99,0}},
/* 39052 */ {(11<<2)|2,{103,99,0}},
/* 39053 */ {(11<<2)|2,{103,99,0}},
/* 39054 */ {(11<<2)|2,{103,99,0}},
/* 39055 */ {(11<<2)|2,{103,99,0}},
/* 39056 */ {(11<<2)|2,{103,99,0}},
/* 39057 */ {(11<<2)|2,{103,99,0}},
/* 39058 */ {(11<<2)|2,{103,99,0}},
/* 39059 */ {(11<<2)|2,{103,99,0}},
/* 39060 */ {(11<<2)|2,{103,99,0}},
/* 39061 */ {(11<<2)|2,{103,99,0}},
/* 39062 */ {(11<<2)|2,{103,99,0}},
/* 39063 */ {(11<<2)|2,{103,99,0}},
/* 39064 */ {(11<<2)|2,{103,99,0}},
/* 39065 */ {(11<<2)|2,{103,99,0}},
/* 39066 */ {(11<<2)|2,{103,99,0}},
/* 39067 */ {(11<<2)|2,{103,99,0}},
/* 39068 */ {(11<<2)|2,{103,99,0}},
/* 39069 */ {(11<<2)|2,{103,99,0}},
/* 39070 */ {(11<<2)|2,{103,99,0}},
/* 39071 */ {(11<<2)|2,{103,99,0}},
/* 39072 */ {(16<<2)|3,{103,101,48}},
/* 39073 */ {(16<<2)|3,{103,101,49}},
/* 39074 */ {(16<<2)|3,{103,101,50}},
/* 39075 */ {(16<<2)|3,{103,101,97}},
/* 39076 */ {(16<<2)|3,{103,101,99}},
/* 39077 */ {(16<<2)|3,{103,101,101}},
/* 39078 */ {(16<<2)|3,{103,101,105}},
/* 39079 */ {(16<<2)|3,{103,101,111}},
/* 39080 */ {(16<<2)|3,{103,101,115}},
/* 39081 */ {(16<<2)|3,{103,101,116}},
/* 39082 */ {(11<<2)|2,{103,101,0}},
/* 39083 */ {(11<<2)|2,{103,101,0}},
/* 39084 */ {(11<<2)|2,{103,101,0}},
/* 39085 */ {(11<<2)|2,{103,101,0}},
/* 39086 */ {(11<<2)|2,{103,101,0}},
/* 39087 */ {(11<<2)|2,{103,101,0}},
/* 39088 */ {(11<<2)|2,{103,101,0}},
/* 39089 */ {(11<<2)|2,{103,101,0}},
/* 39090 */ {(11<<2)|2,{103,101,0}},
/* 39091 */ {(11<<2)|2,{103,101,0}},
/* 39092 */ {(11<<2)|2,{103,101,0}},
/* 39093 */ {(11<<2)|2,{103,101,0}},
/* 39094 */ {(11<<2)|2,{103,101,0}},
/* 39095 */ {(11<<2)|2,{103,101,0}},
/* 39096 */ {(11<<2)|2,{103,101,0}},
/* 39097 */ {(11<<2)|2,{103,101,0}},
/* 39098 */ {(11<<2)|2,{103,101,0}},
/* 39099 */ {(11<<2)|2,{103,101,0}},
/* 39100 */ {(11<<2)|2,{103,101,0}},
/* 39101 */ {(11<<2)|2,{103,101,0}},
/* 39102 */ {(11<<2)|2,{103,101,0}},
/* 39103 */ {(11<<2)|2,{103,101,0}},
/* 39104 */ {(16<<2)|3,{103,105,48}},
/* 39105 */ {(16<<2)|3,{103,105,49}},
/* 39106 */ {(16<<2)|3,{103,105,50}},
/* 39107 */ {(16<<2)|3,{103,105,97}},
/* 39108 */ {(16<<2)|3,{103,105,99}},
/* 39109 */ {(16<<2)|3,{103,105,101}},
/* 39110 */ {(16<<2)|3,{103,105,105}},
/* 39111 */ {(16<<2)|3,{103,105,111}},
/* 39112 */ {(16<<2)|3,{103,105,115}},
/* 39113 */ {(16<<2)|3,{103,105,116}},
/* 39114 */ {(11<<2)|2,{103,105,0}},
/* 39115 */ {(11<<2)|2,{103,105,0}},
/* 39116 */ {(11<<2)|2,{103,105,0}},
/* 39117 */ {(11<<2)|2,{103,105,0}},
/* 39118 */ {(11<<2)|2,{103,105,0}},
/* 39119 */ {(11<<2)|2,{103,105,0}},
/* 39120 */ {(11<<2)|2,{103,105,0}},
/* 39121 */ {(11<<2)|2,{103,105,0}},
/* 39122 */ {(11<<2)|2,{103,105,0}},
/* 39123 */ {(11<<2)|2,{103,105,0}},
/* 39124 */ {(11<<2)|2,{103,105,0}},
/* 39125 */ {(11<<2)|2,{103,105,0}},
/* 39126 */ {(11<<2)|2,{103,105,0}},
/* 39127 */ {(11<<2)|2,{103,105,0}},
/* 39128 */ {(11<<2)|2,{103,105,0}},
/* 39129 */ {(11<<2)|2,{103,105,0}},
/* 39130 */ {(11<<2)|2,{103,105,0}},
/* 39131 */ {(11<<2)|2,{103,105,0}},
/* 39132 */ {(11<<2)|2,{103,105,0}},
/* 39133 */ {(11<<2)|2,{103,105,0}},
/* 39134 */ {(11<<2)|2,{103,105,0}},
/* 39135 */ {(11<<2)|2,{103,105,0}},
/* 39136 */ {(16<<2)|3,{103,111,48}},
/* 39137 */ {(16<<2)|3,{103,111,49}},
/* 39138 */ {(16<<2)|3,{103,111,50}},
/* 39139 */ {(16<<2)|3,{103,111,97}},
/* 39140 */ {(16<<2)|3,{103,111,99}},
/* 39141 */ {(16<<2)|3,{103,111,101}},
/* 39142 */ {(16<<2)|3,{103,111,105}},
/* 39143 */ {(16<<2)|3,{103,111,111}},
/* 39144 */ {(16<<2)|3,{103,111,115}},
/* 39145 */ {(16<<2)|3,{103,111,116}},
/* 39146 */ {(11<<2)|2,{103,111,0}},
/* 39147 */ {(11<<2)|2,{103,111,0}},
/* 39148 */ {(11<<2)|2,{103,111,0}},
/* 39149 */ {(11<<2)|2,{103,111,0}},
/* 39150 */ {(11<<2)|2,{103,111,0}},
/* 39151 */ {(11<<2)|2,{103,111,0}},
/* 39152 */ {(11<<2)|2,{103,111,0}},
/* 39153 */ {(11<<2)|2,{103,111,0}},
/* 39154 */ {(11<<2)|2,{103,111,0}},
/* 39155 */ {(11<<2)|2,{103,111,0}},
/* 39156 */ {(11<<2)|2,{103,111,0}},
/* 39157 */ {(11<<2)|2,{103,111,0}},
/* 39158 */ {(11<<2)|2,{103,111,0}},
/* 39159 */ {(11<<2)|2,{103,111,0}},
/* 39160 */ {(11<<2)|2,{103,111,0}},
/* 39161 */ {(11<<2)|2,{103,111,0}},
/* 39162 */ {(11<<2)|2,{103,111,0}},
/* 39163 */ {(11<<2)|2,{103,111,0}},
/* 39164 */ {(11<<2)|2,{103,111,0}},
/* 39165 */ {(11<<2)|2,{103,111,0}},
/* 39166 */ {(11<<2)|2,{103,111,0}},
/* 39167 */ {(11<<2)|2,{103,111,0}},
/* 39168 */ {(16<<2)|3,{103,115,48}},
/* 39169 */ {(16<<2)|3,{103,115,49}},
/* 39170 */ {(16<<2)|3,{103,115,50}},
/* 39171 */ {(16<<2)|3,{103,115,97}},
/* 39172 */ {(16<<2)|3,{103,115,99}},
/* 39173 */ {(16<<2)|3,{103,115,101}},
/* 39174 */ {(16<<2)|3,{103,115,105}},
/* 39175 */ {(16<<2)|3,{103,115,111}},
/* 39176 */ {(16<<2)|3,{103,115,115}},
/* 39177 */ {(16<<2)|3,{103,115,116}},
/* 39178 */ {(11<<2)|2,{103,115,0}},
/* 39179 */ {(11<<2)|2,{103,115,0}},
/* 39180 */ {(11<<2)|2,{103,115,0}},
/* 39181 */ {(11<<2)|2,{103,115,0}},
/* 39182 */ {(11<<2)|2,{103,115,0}},
/* 39183 */ {(11<<2)|2,{103,115,0}},
/* 39184 */ {(11<<2)|2,{103,115,0}},
/* 39185 */ {(11<<2)|2,{103,115,0}},
/* 39186 */ {(11<<2)|2,{103,115,0}},
/* 39187 */ {(11<<2)|2,{103,115,0}},
/* 39188 */ {(11<<2)|2,{103,115,0}},
/* 39189 */ {(11<<2)|2,{103,115,0}},
/* 39190 */ {(11<<2)|2,{103,115,0}},
/* 39191 */ {(11<<2)|2,{103,115,0}},
/* 39192 */ {(11<<2)|2,{103,115,0}},
/* 39193 */ {(11<<2)|2,{103,115,0}},
/* 39194 */ {(11<<2)|2,{103,115,0}},
/* 39195 */ {(11<<2)|2,{103,115,0}},
/* 39196 */ {(11<<2)|2,{103,115,0}},
/* 39197 */ {(11<<2)|2,{103,115,0}},
/* 39198 */ {(11<<2)|2,{103,115,0}},
/* 39199 */ {(11<<2)|2,{103,115,0}},
/* 39200 */ {(16<<2)|3,{103,116,48}},
/* 39201 */ {(16<<2)|3,{103,116,49}},
/* 39202 */ {(16<<2)|3,{103,116,50}},
/* 39203 */ {(16<<2)|3,{103,116,97}},
/* 39204 */ {(16<<2)|3,{103,116,99}},
/* 39205 */ {(16<<2)|3,{103,116,101}},
/* 39206 */ {(16<<2)|3,{103,116,105}},
/* 39207 */ {(16<<2)|3,{103,116,111}},
/* 39208 */ {(16<<2)|3,{103,116,115}},
/* 39209 */ {(16<<2)|3,{103,116,116}},
/* 39210 */ {(11<<2)|2,{103,116,0}},
/* 39211 */ {(11<<2)|2,{103,116,0}},
/* 39212 */ {(11<<2)|2,{103,116,0}},
/* 39213 */ {(11<<2)|2,{103,116,0}},
/* 39214 */ {(11<<2)|2,{103,116,0}},
/* 39215 */ {(11<<2)|2,{103,116,0}},
/* 39216 */ {(11<<2)|2,{103,116,0}},
/* 39217 */ {(11<<2)|2,{103,116,0}},
/* 39218 */ {(11<<2)|2,{103,116,0}},
/* 39219 */ {(11<<2)|2,{103,116,0}},
/* 39220 */ {(11<<2)|2,{103,116,0}},
/* 39221 */ {(11<<2)|2,{103,116,0}},
/* 39222 */ {(11<<2)|2,{103,116,0}},
/* 39223 */ {(11<<2)|2,{103,116,0}},
/* 39224 */ {(11<<2)|2,{103,116,0}},
/* 39225 */ {(11<<2)|2,{103,116,0}},
/* 39226 */ {(11<<2)|2,{103,116,0}},
/* 39227 */ {(11<<2)|2,{103,116,0}},
/* 39228 */ {(11<<2)|2,{103,116,0}},
/* 39229 */ {(11<<2)|2,{103,116,0}},
/* 39230 */ {(11<<2)|2,{103,116,0}},
/* 39231 */ {(11<<2)|2,{103,116,0}},
/* 39232 */ {(12<<2)|2,{103,32,0}},
/* 39233 */ {(12<<2)|2,{103,32,0}},
/* 39234 */ {(12<<2)|2,{103,32,0}},
/* 39235 */ {(12<<2)|2,{103,32,0}},
/* 39236 */ {(12<<2)|2,{103,32,0}},
/* 39237 */ {(12<<2)|2,{103,32,0}},
/* 39238 */ {(12<<2)|2,{103,32,0}},
/* 39239 */ {(12<<2)|2,{103,32,0}},
/* 39240 */ {(12<<2)|2,{103,32,0}},
/* 39241 */ {(12<<2)|2,{103,32,0}},
/* 39242 */ {(12<<2)|2,{103,32,0}},
/* 39243 */ {(12<<2)|2,{103,32,0}},
/* 39244 */ {(12<<2)|2,{103,32,0}},
/* 39245 */ {(12<<2)|2,{103,32,0}},
/* 39246 */ {(12<<2)|2,{103,32,0}},
/* 39247 */ {(12<<2)|2,{103,32,0}},
/* 39248 */ {(12<<2)|2,{103,37,0}},
/* 39249 */ {(12<<2)|2,{103,37,0}},
/* 39250 */ {(12<<2)|2,{103,37,0}},
/* 39251 */ {(12<<2)|2,{103,37,0}},
/* 39252 */ {(12<<2)|2,{103,37,0}},
/* 39253 */ {(12<<2)|2,{103,37,0}},
/* 39254 */ {(12<<2)|2,{103,37,0}},
/* 39255 */ {(12<<2)|2,{103,37,0}},
/* 39256 */ {(12<<2)|2,{103,37,0}},
/* 39257 */ {(12<<2)|2,{103,37,0}},
/* 39258 */ {(12<<2)|2,{103,37,0}},
/* 39259 */ {(12<<2)|2,{103,37,0}},
/* 39260 */ {(12<<2)|2,{103,37,0}},
/* 39261 */ {(12<<2)|2,{103,37,0}},
/* 39262 */ {(12<<2)|2,{103,37,0}},
/* 39263 */ {(12<<2)|2,{103,37,0}},
/* 39264 */ {(12<<2)|2,{103,45,0}},
/* 39265 */ {(12<<2)|2,{103,45,0}},
/* 39266 */ {(12<<2)|2,{103,45,0}},
/* 39267 */ {(12<<2)|2,{103,45,0}},
/* 39268 */ {(12<<2)|2,{103,45,0}},
/* 39269 */ {(12<<2)|2,{103,45,0}},
/* 39270 */ {(12<<2)|2,{103,45,0}},
/* 39271 */ {(12<<2)|2,{103,45,0}},
/* 39272 */ {(12<<2)|2,{103,45,0}},
/* 39273 */ {(12<<2)|2,{103,45,0}},
/* 39274 */ {(12<<2)|2,{103,45,0}},
/* 39275 */ {(12<<2)|2,{103,45,0}},
/* 39276 */ {(12<<2)|2,{103,45,0}},
/* 39277 */ {(12<<2)|2,{103,45,0}},
/* 39278 */ {(12<<2)|2,{103,45,0}},
/* 39279 */ {(12<<2)|2,{103,45,0}},
/* 39280 */ {(12<<2)|2,{103,46,0}},
/* 39281 */ {(12<<2)|2,{103,46,0}},
/* 39282 */ {(12<<2)|2,{103,46,0}},
/* 39283 */ {(12<<2)|2,{103,46,0}},
/* 39284 */ {(12<<2)|2,{103,46,0}},
/* 39285 */ {(12<<2)|2,{103,46,0}},
/* 39286 */ {(12<<2)|2,{103,46,0}},
/* 39287 */ {(12<<2)|2,{103,46,0}},
/* 39288 */ {(12<<2)|2,{103,46,0}},
/* 39289 */ {(12<<2)|2,{103,46,0}},
/* 39290 */ {(12<<2)|2,{103,46,0}},
/* 39291 */ {(12<<2)|2,{103,46,0}},
/* 39292 */ {(12<<2)|2,{103,46,0}},
/* 39293 */ {(12<<2)|2,{103,46,0}},
/* 39294 */ {(12<<2)|2,{103,46,0}},
/* 39295 */ {(12<<2)|2,{103,46,0}},
/* 39296 */ {(12<<2)|2,{103,47,0}},
/* 39297 */ {(12<<2)|2,{103,47,0}},
/* 39298 */ {(12<<2)|2,{103,47,0}},
/* 39299 */ {(12<<2)|2,{103,47,0}},
/* 39300 */ {(12<<2)|2,{103,47,0}},
/* 39301 */ {(12<<2)|2,{103,47,0}},
/* 39302 */ {(12<<2)|2,{103,47,0}},
/* 39303 */ {(12<<2)|2,{103,47,0}},
/* 39304 */ {(12<<2)|2,{103,47,0}},
/* 39305 */ {(12<<2)|2,{103,47,0}},
/* 39306 */ {(12<<2)|2,{103,47,0}},
/* 39307 */ {(12<<2)|2,{103,47,0}},
/* 39308 */ {(12<<2)|2,{103,47,0}},
/* 39309 */ {(12<<2)|2,{103,47,0}},
/* 39310 */ {(12<<2)|2,{103,47,0}},
/* 39311 */ {(12<<2)|2,{103,47,0}},
/* 39312 */ {(12<<2)|2,{103,51,0}},
/* 39313 */ {(12<<2)|2,{103,51,0}},
/* 39314 */ {(12<<2)|2,{103,51,0}},
/* 39315 */ {(12<<2)|2,{103,51,0}},
/* 39316 */ {(12<<2)|2,{103,51,0}},
/* 39317 */ {(12<<2)|2,{103,51,0}},
/* 39318 */ {(12<<2)|2,{103,51,0}},
/* 39319 */ {(12<<2)|2,{103,51,0}},
/* 39320 */ {(12<<2)|2,{103,51,0}},
/* 39321 */ {(12<<2)|2,{103,51,0}},
/* 39322 */ {(12<<2)|2,{103,51,0}},
/* 39323 */ {(12<<2)|2,{103,51,0}},
/* 39324 */ {(12<<2)|2,{103,51,0}},
/* 39325 */ {(12<<2)|2,{103,51,0}},
/* 39326 */ {(12<<2)|2,{103,51,0}},
/* 39327 */ {(12<<2)|2,{103,51,0}},
/* 39328 */ {(12<<2)|2,{103,52,0}},
/* 39329 */ {(12<<2)|2,{103,52,0}},
/* 39330 */ {(12<<2)|2,{103,52,0}},
/* 39331 */ {(12<<2)|2,{103,52,0}},
/* 39332 */ {(12<<2)|2,{103,52,0}},
/* 39333 */ {(12<<2)|2,{103,52,0}},
/* 39334 */ {(12<<2)|2,{103,52,0}},
/* 39335 */ {(12<<2)|2,{103,52,0}},
/* 39336 */ {(12<<2)|2,{103,52,0}},
/* 39337 */ {(12<<2)|2,{103,52,0}},
/* 39338 */ {(12<<2)|2,{103,52,0}},
/* 39339 */ {(12<<2)|2,{103,52,0}},
/* 39340 */ {(12<<2)|2,{103,52,0}},
/* 39341 */ {(12<<2)|2,{103,52,0}},
/* 39342 */ {(12<<2)|2,{103,52,0}},
/* 39343 */ {(12<<2)|2,{103,52,0}},
/* 39344 */ {(12<<2)|2,{103,53,0}},
/* 39345 */ {(12<<2)|2,{103,53,0}},
/* 39346 */ {(12<<2)|2,{103,53,0}},
/* 39347 */ {(12<<2)|2,{103,53,0}},
/* 39348 */ {(12<<2)|2,{103,53,0}},
/* 39349 */ {(12<<2)|2,{103,53,0}},
/* 39350 */ {(12<<2)|2,{103,53,0}},
/* 39351 */ {(12<<2)|2,{103,53,0}},
/* 39352 */ {(12<<2)|2,{103,53,0}},
/* 39353 */ {(12<<2)|2,{103,53,0}},
/* 39354 */ {(12<<2)|2,{103,53,0}},
/* 39355 */ {(12<<2)|2,{103,53,0}},
/* 39356 */ {(12<<2)|2,{103,53,0}},
/* 39357 */ {(12<<2)|2,{103,53,0}},
/* 39358 */ {(12<<2)|2,{103,53,0}},
/* 39359 */ {(12<<2)|2,{103,53,0}},
/* 39360 */ {(12<<2)|2,{103,54,0}},
/* 39361 */ {(12<<2)|2,{103,54,0}},
/* 39362 */ {(12<<2)|2,{103,54,0}},
/* 39363 */ {(12<<2)|2,{103,54,0}},
/* 39364 */ {(12<<2)|2,{103,54,0}},
/* 39365 */ {(12<<2)|2,{103,54,0}},
/* 39366 */ {(12<<2)|2,{103,54,0}},
/* 39367 */ {(12<<2)|2,{103,54,0}},
/* 39368 */ {(12<<2)|2,{103,54,0}},
/* 39369 */ {(12<<2)|2,{103,54,0}},
/* 39370 */ {(12<<2)|2,{103,54,0}},
/* 39371 */ {(12<<2)|2,{103,54,0}},
/* 39372 */ {(12<<2)|2,{103,54,0}},
/* 39373 */ {(12<<2)|2,{103,54,0}},
/* 39374 */ {(12<<2)|2,{103,54,0}},
/* 39375 */ {(12<<2)|2,{103,54,0}},
/* 39376 */ {(12<<2)|2,{103,55,0}},
/* 39377 */ {(12<<2)|2,{103,55,0}},
/* 39378 */ {(12<<2)|2,{103,55,0}},
/* 39379 */ {(12<<2)|2,{103,55,0}},
/* 39380 */ {(12<<2)|2,{103,55,0}},
/* 39381 */ {(12<<2)|2,{103,55,0}},
/* 39382 */ {(12<<2)|2,{103,55,0}},
/* 39383 */ {(12<<2)|2,{103,55,0}},
/* 39384 */ {(12<<2)|2,{103,55,0}},
/* 39385 */ {(12<<2)|2,{103,55,0}},
/* 39386 */ {(12<<2)|2,{103,55,0}},
/* 39387 */ {(12<<2)|2,{103,55,0}},
/* 39388 */ {(12<<2)|2,{103,55,0}},
/* 39389 */ {(12<<2)|2,{103,55,0}},
/* 39390 */ {(12<<2)|2,{103,55,0}},
/* 39391 */ {(12<<2)|2,{103,55,0}},
/* 39392 */ {(12<<2)|2,{103,56,0}},
/* 39393 */ {(12<<2)|2,{103,56,0}},
/* 39394 */ {(12<<2)|2,{103,56,0}},
/* 39395 */ {(12<<2)|2,{103,56,0}},
/* 39396 */ {(12<<2)|2,{103,56,0}},
/* 39397 */ {(12<<2)|2,{103,56,0}},
/* 39398 */ {(12<<2)|2,{103,56,0}},
/* 39399 */ {(12<<2)|2,{103,56,0}},
/* 39400 */ {(12<<2)|2,{103,56,0}},
/* 39401 */ {(12<<2)|2,{103,56,0}},
/* 39402 */ {(12<<2)|2,{103,56,0}},
/* 39403 */ {(12<<2)|2,{103,56,0}},
/* 39404 */ {(12<<2)|2,{103,56,0}},
/* 39405 */ {(12<<2)|2,{103,56,0}},
/* 39406 */ {(12<<2)|2,{103,56,0}},
/* 39407 */ {(12<<2)|2,{103,56,0}},
/* 39408 */ {(12<<2)|2,{103,57,0}},
/* 39409 */ {(12<<2)|2,{103,57,0}},
/* 39410 */ {(12<<2)|2,{103,57,0}},
/* 39411 */ {(12<<2)|2,{103,57,0}},
/* 39412 */ {(12<<2)|2,{103,57,0}},
/* 39413 */ {(12<<2)|2,{103,57,0}},
/* 39414 */ {(12<<2)|2,{103,57,0}},
/* 39415 */ {(12<<2)|2,{103,57,0}},
/* 39416 */ {(12<<2)|2,{103,57,0}},
/* 39417 */ {(12<<2)|2,{103,57,0}},
/* 39418 */ {(12<<2)|2,{103,57,0}},
/* 39419 */ {(12<<2)|2,{103,57,0}},
/* 39420 */ {(12<<2)|2,{103,57,0}},
/* 39421 */ {(12<<2)|2,{103,57,0}},
/* 39422 */ {(12<<2)|2,{103,57,0}},
/* 39423 */ {(12<<2)|2,{103,57,0}},
/* 39424 */ {(12<<2)|2,{103,61,0}},
/* 39425 */ {(12<<2)|2,{103,61,0}},
/* 39426 */ {(12<<2)|2,{103,61,0}},
/* 39427 */ {(12<<2)|2,{103,61,0}},
/* 39428 */ {(12<<2)|2,{103,61,0}},
/* 39429 */ {(12<<2)|2,{103,61,0}},
/* 39430 */ {(12<<2)|2,{103,61,0}},
/* 39431 */ {(12<<2)|2,{103,61,0}},
/* 39432 */ {(12<<2)|2,{103,61,0}},
/* 39433 */ {(12<<2)|2,{103,61,0}},
/* 39434 */ {(12<<2)|2,{103,61,0}},
/* 39435 */ {(12<<2)|2,{103,61,0}},
/* 39436 */ {(12<<2)|2,{103,61,0}},
/* 39437 */ {(12<<2)|2,{103,61,0}},
/* 39438 */ {(12<<2)|2,{103,61,0}},
/* 39439 */ {(12<<2)|2,{103,61,0}},
/* 39440 */ {(12<<2)|2,{103,65,0}},
/* 39441 */ {(12<<2)|2,{103,65,0}},
/* 39442 */ {(12<<2)|2,{103,65,0}},
/* 39443 */ {(12<<2)|2,{103,65,0}},
/* 39444 */ {(12<<2)|2,{103,65,0}},
/* 39445 */ {(12<<2)|2,{103,65,0}},
/* 39446 */ {(12<<2)|2,{103,65,0}},
/* 39447 */ {(12<<2)|2,{103,65,0}},
/* 39448 */ {(12<<2)|2,{103,65,0}},
/* 39449 */ {(12<<2)|2,{103,65,0}},
/* 39450 */ {(12<<2)|2,{103,65,0}},
/* 39451 */ {(12<<2)|2,{103,65,0}},
/* 39452 */ {(12<<2)|2,{103,65,0}},
/* 39453 */ {(12<<2)|2,{103,65,0}},
/* 39454 */ {(12<<2)|2,{103,65,0}},
/* 39455 */ {(12<<2)|2,{103,65,0}},
/* 39456 */ {(12<<2)|2,{103,95,0}},
/* 39457 */ {(12<<2)|2,{103,95,0}},
/* 39458 */ {(12<<2)|2,{103,95,0}},
/* 39459 */ {(12<<2)|2,{103,95,0}},
/* 39460 */ {(12<<2)|2,{103,95,0}},
/* 39461 */ {(12<<2)|2,{103,95,0}},
/* 39462 */ {(12<<2)|2,{103,95,0}},
/* 39463 */ {(12<<2)|2,{103,95,0}},
/* 39464 */ {(12<<2)|2,{103,95,0}},
/* 39465 */ {(12<<2)|2,{103,95,0}},
/* 39466 */ {(12<<2)|2,{103,95,0}},
/* 39467 */ {(12<<2)|2,{103,95,0}},
/* 39468 */ {(12<<2)|2,{103,95,0}},
/* 39469 */ {(12<<2)|2,{103,95,0}},
/* 39470 */ {(12<<2)|2,{103,95,0}},
/* 39471 */ {(12<<2)|2,{103,95,0}},
/* 39472 */ {(12<<2)|2,{103,98,0}},
/* 39473 */ {(12<<2)|2,{103,98,0}},
/* 39474 */ {(12<<2)|2,{103,98,0}},
/* 39475 */ {(12<<2)|2,{103,98,0}},
/* 39476 */ {(12<<2)|2,{103,98,0}},
/* 39477 */ {(12<<2)|2,{103,98,0}},
/* 39478 */ {(12<<2)|2,{103,98,0}},
/* 39479 */ {(12<<2)|2,{103,98,0}},
/* 39480 */ {(12<<2)|2,{103,98,0}},
/* 39481 */ {(12<<2)|2,{103,98,0}},
/* 39482 */ {(12<<2)|2,{103,98,0}},
/* 39483 */ {(12<<2)|2,{103,98,0}},
/* 39484 */ {(12<<2)|2,{103,98,0}},
/* 39485 */ {(12<<2)|2,{103,98,0}},
/* 39486 */ {(12<<2)|2,{103,98,0}},
/* 39487 */ {(12<<2)|2,{103,98,0}},
/* 39488 */ {(12<<2)|2,{103,100,0}},
/* 39489 */ {(12<<2)|2,{103,100,0}},
/* 39490 */ {(12<<2)|2,{103,100,0}},
/* 39491 */ {(12<<2)|2,{103,100,0}},
/* 39492 */ {(12<<2)|2,{103,100,0}},
/* 39493 */ {(12<<2)|2,{103,100,0}},
/* 39494 */ {(12<<2)|2,{103,100,0}},
/* 39495 */ {(12<<2)|2,{103,100,0}},
/* 39496 */ {(12<<2)|2,{103,100,0}},
/* 39497 */ {(12<<2)|2,{103,100,0}},
/* 39498 */ {(12<<2)|2,{103,100,0}},
/* 39499 */ {(12<<2)|2,{103,100,0}},
/* 39500 */ {(12<<2)|2,{103,100,0}},
/* 39501 */ {(12<<2)|2,{103,100,0}},
/* 39502 */ {(12<<2)|2,{103,100,0}},
/* 39503 */ {(12<<2)|2,{103,100,0}},
/* 39504 */ {(12<<2)|2,{103,102,0}},
/* 39505 */ {(12<<2)|2,{103,102,0}},
/* 39506 */ {(12<<2)|2,{103,102,0}},
/* 39507 */ {(12<<2)|2,{103,102,0}},
/* 39508 */ {(12<<2)|2,{103,102,0}},
/* 39509 */ {(12<<2)|2,{103,102,0}},
/* 39510 */ {(12<<2)|2,{103,102,0}},
/* 39511 */ {(12<<2)|2,{103,102,0}},
/* 39512 */ {(12<<2)|2,{103,102,0}},
/* 39513 */ {(12<<2)|2,{103,102,0}},
/* 39514 */ {(12<<2)|2,{103,102,0}},
/* 39515 */ {(12<<2)|2,{103,102,0}},
/* 39516 */ {(12<<2)|2,{103,102,0}},
/* 39517 */ {(12<<2)|2,{103,102,0}},
/* 39518 */ {(12<<2)|2,{103,102,0}},
/* 39519 */ {(12<<2)|2,{103,102,0}},
/* 39520 */ {(12<<2)|2,{103,103,0}},
/* 39521 */ {(12<<2)|2,{103,103,0}},
/* 39522 */ {(12<<2)|2,{103,103,0}},
/* 39523 */ {(12<<2)|2,{103,103,0}},
/* 39524 */ {(12<<2)|2,{103,103,0}},
/* 39525 */ {(12<<2)|2,{103,103,0}},
/* 39526 */ {(12<<2)|2,{103,103,0}},
/* 39527 */ {(12<<2)|2,{103,103,0}},
/* 39528 */ {(12<<2)|2,{103,103,0}},
/* 39529 */ {(12<<2)|2,{103,103,0}},
/* 39530 */ {(12<<2)|2,{103,103,0}},
/* 39531 */ {(12<<2)|2,{103,103,0}},
/* 39532 */ {(12<<2)|2,{103,103,0}},
/* 39533 */ {(12<<2)|2,{103,103,0}},
/* 39534 */ {(12<<2)|2,{103,103,0}},
/* 39535 */ {(12<<2)|2,{103,103,0}},
/* 39536 */ {(12<<2)|2,{103,104,0}},
/* 39537 */ {(12<<2)|2,{103,104,0}},
/* 39538 */ {(12<<2)|2,{103,104,0}},
/* 39539 */ {(12<<2)|2,{103,104,0}},
/* 39540 */ {(12<<2)|2,{103,104,0}},
/* 39541 */ {(12<<2)|2,{103,104,0}},
/* 39542 */ {(12<<2)|2,{103,104,0}},
/* 39543 */ {(12<<2)|2,{103,104,0}},
/* 39544 */ {(12<<2)|2,{103,104,0}},
/* 39545 */ {(12<<2)|2,{103,104,0}},
/* 39546 */ {(12<<2)|2,{103,104,0}},
/* 39547 */ {(12<<2)|2,{103,104,0}},
/* 39548 */ {(12<<2)|2,{103,104,0}},
/* 39549 */ {(12<<2)|2,{103,104,0}},
/* 39550 */ {(12<<2)|2,{103,104,0}},
/* 39551 */ {(12<<2)|2,{103,104,0}},
/* 39552 */ {(12<<2)|2,{103,108,0}},
/* 39553 */ {(12<<2)|2,{103,108,0}},
/* 39554 */ {(12<<2)|2,{103,108,0}},
/* 39555 */ {(12<<2)|2,{103,108,0}},
/* 39556 */ {(12<<2)|2,{103,108,0}},
/* 39557 */ {(12<<2)|2,{103,108,0}},
/* 39558 */ {(12<<2)|2,{103,108,0}},
/* 39559 */ {(12<<2)|2,{103,108,0}},
/* 39560 */ {(12<<2)|2,{103,108,0}},
/* 39561 */ {(12<<2)|2,{103,108,0}},
/* 39562 */ {(12<<2)|2,{103,108,0}},
/* 39563 */ {(12<<2)|2,{103,108,0}},
/* 39564 */ {(12<<2)|2,{103,108,0}},
/* 39565 */ {(12<<2)|2,{103,108,0}},
/* 39566 */ {(12<<2)|2,{103,108,0}},
/* 39567 */ {(12<<2)|2,{103,108,0}},
/* 39568 */ {(12<<2)|2,{103,109,0}},
/* 39569 */ {(12<<2)|2,{103,109,0}},
/* 39570 */ {(12<<2)|2,{103,109,0}},
/* 39571 */ {(12<<2)|2,{103,109,0}},
/* 39572 */ {(12<<2)|2,{103,109,0}},
/* 39573 */ {(12<<2)|2,{103,109,0}},
/* 39574 */ {(12<<2)|2,{103,109,0}},
/* 39575 */ {(12<<2)|2,{103,109,0}},
/* 39576 */ {(12<<2)|2,{103,109,0}},
/* 39577 */ {(12<<2)|2,{103,109,0}},
/* 39578 */ {(12<<2)|2,{103,109,0}},
/* 39579 */ {(12<<2)|2,{103,109,0}},
/* 39580 */ {(12<<2)|2,{103,109,0}},
/* 39581 */ {(12<<2)|2,{103,109,0}},
/* 39582 */ {(12<<2)|2,{103,109,0}},
/* 39583 */ {(12<<2)|2,{103,109,0}},
/* 39584 */ {(12<<2)|2,{103,110,0}},
/* 39585 */ {(12<<2)|2,{103,110,0}},
/* 39586 */ {(12<<2)|2,{103,110,0}},
/* 39587 */ {(12<<2)|2,{103,110,0}},
/* 39588 */ {(12<<2)|2,{103,110,0}},
/* 39589 */ {(12<<2)|2,{103,110,0}},
/* 39590 */ {(12<<2)|2,{103,110,0}},
/* 39591 */ {(12<<2)|2,{103,110,0}},
/* 39592 */ {(12<<2)|2,{103,110,0}},
/* 39593 */ {(12<<2)|2,{103,110,0}},
/* 39594 */ {(12<<2)|2,{103,110,0}},
/* 39595 */ {(12<<2)|2,{103,110,0}},
/* 39596 */ {(12<<2)|2,{103,110,0}},
/* 39597 */ {(12<<2)|2,{103,110,0}},
/* 39598 */ {(12<<2)|2,{103,110,0}},
/* 39599 */ {(12<<2)|2,{103,110,0}},
/* 39600 */ {(12<<2)|2,{103,112,0}},
/* 39601 */ {(12<<2)|2,{103,112,0}},
/* 39602 */ {(12<<2)|2,{103,112,0}},
/* 39603 */ {(12<<2)|2,{103,112,0}},
/* 39604 */ {(12<<2)|2,{103,112,0}},
/* 39605 */ {(12<<2)|2,{103,112,0}},
/* 39606 */ {(12<<2)|2,{103,112,0}},
/* 39607 */ {(12<<2)|2,{103,112,0}},
/* 39608 */ {(12<<2)|2,{103,112,0}},
/* 39609 */ {(12<<2)|2,{103,112,0}},
/* 39610 */ {(12<<2)|2,{103,112,0}},
/* 39611 */ {(12<<2)|2,{103,112,0}},
/* 39612 */ {(12<<2)|2,{103,112,0}},
/* 39613 */ {(12<<2)|2,{103,112,0}},
/* 39614 */ {(12<<2)|2,{103,112,0}},
/* 39615 */ {(12<<2)|2,{103,112,0}},
/* 39616 */ {(12<<2)|2,{103,114,0}},
/* 39617 */ {(12<<2)|2,{103,114,0}},
/* 39618 */ {(12<<2)|2,{103,114,0}},
/* 39619 */ {(12<<2)|2,{103,114,0}},
/* 39620 */ {(12<<2)|2,{103,114,0}},
/* 39621 */ {(12<<2)|2,{103,114,0}},
/* 39622 */ {(12<<2)|2,{103,114,0}},
/* 39623 */ {(12<<2)|2,{103,114,0}},
/* 39624 */ {(12<<2)|2,{103,114,0}},
/* 39625 */ {(12<<2)|2,{103,114,0}},
/* 39626 */ {(12<<2)|2,{103,114,0}},
/* 39627 */ {(12<<2)|2,{103,114,0}},
/* 39628 */ {(12<<2)|2,{103,114,0}},
/* 39629 */ {(12<<2)|2,{103,114,0}},
/* 39630 */ {(12<<2)|2,{103,114,0}},
/* 39631 */ {(12<<2)|2,{103,114,0}},
/* 39632 */ {(12<<2)|2,{103,117,0}},
/* 39633 */ {(12<<2)|2,{103,117,0}},
/* 39634 */ {(12<<2)|2,{103,117,0}},
/* 39635 */ {(12<<2)|2,{103,117,0}},
/* 39636 */ {(12<<2)|2,{103,117,0}},
/* 39637 */ {(12<<2)|2,{103,117,0}},
/* 39638 */ {(12<<2)|2,{103,117,0}},
/* 39639 */ {(12<<2)|2,{103,117,0}},
/* 39640 */ {(12<<2)|2,{103,117,0}},
/* 39641 */ {(12<<2)|2,{103,117,0}},
/* 39642 */ {(12<<2)|2,{103,117,0}},
/* 39643 */ {(12<<2)|2,{103,117,0}},
/* 39644 */ {(12<<2)|2,{103,117,0}},
/* 39645 */ {(12<<2)|2,{103,117,0}},
/* 39646 */ {(12<<2)|2,{103,117,0}},
/* 39647 */ {(12<<2)|2,{103,117,0}},
/* 39648 */ {(13<<2)|2,{103,58,0}},
/* 39649 */ {(13<<2)|2,{103,58,0}},
/* 39650 */ {(13<<2)|2,{103,58,0}},
/* 39651 */ {(13<<2)|2,{103,58,0}},
/* 39652 */ {(13<<2)|2,{103,58,0}},
/* 39653 */ {(13<<2)|2,{103,58,0}},
/* 39654 */ {(13<<2)|2,{103,58,0}},
/* 39655 */ {(13<<2)|2,{103,58,0}},
/* 39656 */ {(13<<2)|2,{103,66,0}},
/* 39657 */ {(13<<2)|2,{103,66,0}},
/* 39658 */ {(13<<2)|2,{103,66,0}},
/* 39659 */ {(13<<2)|2,{103,66,0}},
/* 39660 */ {(13<<2)|2,{103,66,0}},
/* 39661 */ {(13<<2)|2,{103,66,0}},
/* 39662 */ {(13<<2)|2,{103,66,0}},
/* 39663 */ {(13<<2)|2,{103,66,0}},
/* 39664 */ {(13<<2)|2,{103,67,0}},
/* 39665 */ {(13<<2)|2,{103,67,0}},
/* 39666 */ {(13<<2)|2,{103,67,0}},
/* 39667 */ {(13<<2)|2,{103,67,0}},
/* 39668 */ {(13<<2)|2,{103,67,0}},
/* 39669 */ {(13<<2)|2,{103,67,0}},
/* 39670 */ {(13<<2)|2,{103,67,0}},
/* 39671 */ {(13<<2)|2,{103,67,0}},
/* 39672 */ {(13<<2)|2,{103,68,0}},
/* 39673 */ {(13<<2)|2,{103,68,0}},
/* 39674 */ {(13<<2)|2,{103,68,0}},
/* 39675 */ {(13<<2)|2,{103,68,0}},
/* 39676 */ {(13<<2)|2,{103,68,0}},
/* 39677 */ {(13<<2)|2,{103,68,0}},
/* 39678 */ {(13<<2)|2,{103,68,0}},
/* 39679 */ {(13<<2)|2,{103,68,0}},
/* 39680 */ {(13<<2)|2,{103,69,0}},
/* 39681 */ {(13<<2)|2,{103,69,0}},
/* 39682 */ {(13<<2)|2,{103,69,0}},
/* 39683 */ {(13<<2)|2,{103,69,0}},
/* 39684 */ {(13<<2)|2,{103,69,0}},
/* 39685 */ {(13<<2)|2,{103,69,0}},
/* 39686 */ {(13<<2)|2,{103,69,0}},
/* 39687 */ {(13<<2)|2,{103,69,0}},
/* 39688 */ {(13<<2)|2,{103,70,0}},
/* 39689 */ {(13<<2)|2,{103,70,0}},
/* 39690 */ {(13<<2)|2,{103,70,0}},
/* 39691 */ {(13<<2)|2,{103,70,0}},
/* 39692 */ {(13<<2)|2,{103,70,0}},
/* 39693 */ {(13<<2)|2,{103,70,0}},
/* 39694 */ {(13<<2)|2,{103,70,0}},
/* 39695 */ {(13<<2)|2,{103,70,0}},
/* 39696 */ {(13<<2)|2,{103,71,0}},
/* 39697 */ {(13<<2)|2,{103,71,0}},
/* 39698 */ {(13<<2)|2,{103,71,0}},
/* 39699 */ {(13<<2)|2,{103,71,0}},
/* 39700 */ {(13<<2)|2,{103,71,0}},
/* 39701 */ {(13<<2)|2,{103,71,0}},
/* 39702 */ {(13<<2)|2,{103,71,0}},
/* 39703 */ {(13<<2)|2,{103,71,0}},
/* 39704 */ {(13<<2)|2,{103,72,0}},
/* 39705 */ {(13<<2)|2,{103,72,0}},
/* 39706 */ {(13<<2)|2,{103,72,0}},
/* 39707 */ {(13<<2)|2,{103,72,0}},
/* 39708 */ {(13<<2)|2,{103,72,0}},
/* 39709 */ {(13<<2)|2,{103,72,0}},
/* 39710 */ {(13<<2)|2,{103,72,0}},
/* 39711 */ {(13<<2)|2,{103,72,0}},
/* 39712 */ {(13<<2)|2,{103,73,0}},
/* 39713 */ {(13<<2)|2,{103,73,0}},
/* 39714 */ {(13<<2)|2,{103,73,0}},
/* 39715 */ {(13<<2)|2,{103,73,0}},
/* 39716 */ {(13<<2)|2,{103,73,0}},
/* 39717 */ {(13<<2)|2,{103,73,0}},
/* 39718 */ {(13<<2)|2,{103,73,0}},
/* 39719 */ {(13<<2)|2,{103,73,0}},
/* 39720 */ {(13<<2)|2,{103,74,0}},
/* 39721 */ {(13<<2)|2,{103,74,0}},
/* 39722 */ {(13<<2)|2,{103,74,0}},
/* 39723 */ {(13<<2)|2,{103,74,0}},
/* 39724 */ {(13<<2)|2,{103,74,0}},
/* 39725 */ {(13<<2)|2,{103,74,0}},
/* 39726 */ {(13<<2)|2,{103,74,0}},
/* 39727 */ {(13<<2)|2,{103,74,0}},
/* 39728 */ {(13<<2)|2,{103,75,0}},
/* 39729 */ {(13<<2)|2,{103,75,0}},
/* 39730 */ {(13<<2)|2,{103,75,0}},
/* 39731 */ {(13<<2)|2,{103,75,0}},
/* 39732 */ {(13<<2)|2,{103,75,0}},
/* 39733 */ {(13<<2)|2,{103,75,0}},
/* 39734 */ {(13<<2)|2,{103,75,0}},
/* 39735 */ {(13<<2)|2,{103,75,0}},
/* 39736 */ {(13<<2)|2,{103,76,0}},
/* 39737 */ {(13<<2)|2,{103,76,0}},
/* 39738 */ {(13<<2)|2,{103,76,0}},
/* 39739 */ {(13<<2)|2,{103,76,0}},
/* 39740 */ {(13<<2)|2,{103,76,0}},
/* 39741 */ {(13<<2)|2,{103,76,0}},
/* 39742 */ {(13<<2)|2,{103,76,0}},
/* 39743 */ {(13<<2)|2,{103,76,0}},
/* 39744 */ {(13<<2)|2,{103,77,0}},
/* 39745 */ {(13<<2)|2,{103,77,0}},
/* 39746 */ {(13<<2)|2,{103,77,0}},
/* 39747 */ {(13<<2)|2,{103,77,0}},
/* 39748 */ {(13<<2)|2,{103,77,0}},
/* 39749 */ {(13<<2)|2,{103,77,0}},
/* 39750 */ {(13<<2)|2,{103,77,0}},
/* 39751 */ {(13<<2)|2,{103,77,0}},
/* 39752 */ {(13<<2)|2,{103,78,0}},
/* 39753 */ {(13<<2)|2,{103,78,0}},
/* 39754 */ {(13<<2)|2,{103,78,0}},
/* 39755 */ {(13<<2)|2,{103,78,0}},
/* 39756 */ {(13<<2)|2,{103,78,0}},
/* 39757 */ {(13<<2)|2,{103,78,0}},
/* 39758 */ {(13<<2)|2,{103,78,0}},
/* 39759 */ {(13<<2)|2,{103,78,0}},
/* 39760 */ {(13<<2)|2,{103,79,0}},
/* 39761 */ {(13<<2)|2,{103,79,0}},
/* 39762 */ {(13<<2)|2,{103,79,0}},
/* 39763 */ {(13<<2)|2,{103,79,0}},
/* 39764 */ {(13<<2)|2,{103,79,0}},
/* 39765 */ {(13<<2)|2,{103,79,0}},
/* 39766 */ {(13<<2)|2,{103,79,0}},
/* 39767 */ {(13<<2)|2,{103,79,0}},
/* 39768 */ {(13<<2)|2,{103,80,0}},
/* 39769 */ {(13<<2)|2,{103,80,0}},
/* 39770 */ {(13<<2)|2,{103,80,0}},
/* 39771 */ {(13<<2)|2,{103,80,0}},
/* 39772 */ {(13<<2)|2,{103,80,0}},
/* 39773 */ {(13<<2)|2,{103,80,0}},
/* 39774 */ {(13<<2)|2,{103,80,0}},
/* 39775 */ {(13<<2)|2,{103,80,0}},
/* 39776 */ {(13<<2)|2,{103,81,0}},
/* 39777 */ {(13<<2)|2,{103,81,0}},
/* 39778 */ {(13<<2)|2,{103,81,0}},
/* 39779 */ {(13<<2)|2,{103,81,0}},
/* 39780 */ {(13<<2)|2,{103,81,0}},
/* 39781 */ {(13<<2)|2,{103,81,0}},
/* 39782 */ {(13<<2)|2,{103,81,0}},
/* 39783 */ {(13<<2)|2,{103,81,0}},
/* 39784 */ {(13<<2)|2,{103,82,0}},
/* 39785 */ {(13<<2)|2,{103,82,0}},
/* 39786 */ {(13<<2)|2,{103,82,0}},
/* 39787 */ {(13<<2)|2,{103,82,0}},
/* 39788 */ {(13<<2)|2,{103,82,0}},
/* 39789 */ {(13<<2)|2,{103,82,0}},
/* 39790 */ {(13<<2)|2,{103,82,0}},
/* 39791 */ {(13<<2)|2,{103,82,0}},
/* 39792 */ {(13<<2)|2,{103,83,0}},
/* 39793 */ {(13<<2)|2,{103,83,0}},
/* 39794 */ {(13<<2)|2,{103,83,0}},
/* 39795 */ {(13<<2)|2,{103,83,0}},
/* 39796 */ {(13<<2)|2,{103,83,0}},
/* 39797 */ {(13<<2)|2,{103,83,0}},
/* 39798 */ {(13<<2)|2,{103,83,0}},
/* 39799 */ {(13<<2)|2,{103,83,0}},
/* 39800 */ {(13<<2)|2,{103,84,0}},
/* 39801 */ {(13<<2)|2,{103,84,0}},
/* 39802 */ {(13<<2)|2,{103,84,0}},
/* 39803 */ {(13<<2)|2,{103,84,0}},
/* 39804 */ {(13<<2)|2,{103,84,0}},
/* 39805 */ {(13<<2)|2,{103,84,0}},
/* 39806 */ {(13<<2)|2,{103,84,0}},
/* 39807 */ {(13<<2)|2,{103,84,0}},
/* 39808 */ {(13<<2)|2,{103,85,0}},
/* 39809 */ {(13<<2)|2,{103,85,0}},
/* 39810 */ {(13<<2)|2,{103,85,0}},
/* 39811 */ {(13<<2)|2,{103,85,0}},
/* 39812 */ {(13<<2)|2,{103,85,0}},
/* 39813 */ {(13<<2)|2,{103,85,0}},
/* 39814 */ {(13<<2)|2,{103,85,0}},
/* 39815 */ {(13<<2)|2,{103,85,0}},
/* 39816 */ {(13<<2)|2,{103,86,0}},
/* 39817 */ {(13<<2)|2,{103,86,0}},
/* 39818 */ {(13<<2)|2,{103,86,0}},
/* 39819 */ {(13<<2)|2,{103,86,0}},
/* 39820 */ {(13<<2)|2,{103,86,0}},
/* 39821 */ {(13<<2)|2,{103,86,0}},
/* 39822 */ {(13<<2)|2,{103,86,0}},
/* 39823 */ {(13<<2)|2,{103,86,0}},
/* 39824 */ {(13<<2)|2,{103,87,0}},
/* 39825 */ {(13<<2)|2,{103,87,0}},
/* 39826 */ {(13<<2)|2,{103,87,0}},
/* 39827 */ {(13<<2)|2,{103,87,0}},
/* 39828 */ {(13<<2)|2,{103,87,0}},
/* 39829 */ {(13<<2)|2,{103,87,0}},
/* 39830 */ {(13<<2)|2,{103,87,0}},
/* 39831 */ {(13<<2)|2,{103,87,0}},
/* 39832 */ {(13<<2)|2,{103,89,0}},
/* 39833 */ {(13<<2)|2,{103,89,0}},
/* 39834 */ {(13<<2)|2,{103,89,0}},
/* 39835 */ {(13<<2)|2,{103,89,0}},
/* 39836 */ {(13<<2)|2,{103,89,0}},
/* 39837 */ {(13<<2)|2,{103,89,0}},
/* 39838 */ {(13<<2)|2,{103,89,0}},
/* 39839 */ {(13<<2)|2,{103,89,0}},
/* 39840 */ {(13<<2)|2,{103,106,0}},
/* 39841 */ {(13<<2)|2,{103,106,0}},
/* 39842 */ {(13<<2)|2,{103,106,0}},
/* 39843 */ {(13<<2)|2,{103,106,0}},
/* 39844 */ {(13<<2)|2,{103,106,0}},
/* 39845 */ {(13<<2)|2,{103,106,0}},
/* 39846 */ {(13<<2)|2,{103,106,0}},
/* 39847 */ {(13<<2)|2,{103,106,0}},
/* 39848 */ {(13<<2)|2,{103,107,0}},
/* 39849 */ {(13<<2)|2,{103,107,0}},
/* 39850 */ {(13<<2)|2,{103,107,0}},
/* 39851 */ {(13<<2)|2,{103,107,0}},
/* 39852 */ {(13<<2)|2,{103,107,0}},
/* 39853 */ {(13<<2)|2,{103,107,0}},
/* 39854 */ {(13<<2)|2,{103,107,0}},
/* 39855 */ {(13<<2)|2,{103,107,0}},
/* 39856 */ {(13<<2)|2,{103,113,0}},
/* 39857 */ {(13<<2)|2,{103,113,0}},
/* 39858 */ {(13<<2)|2,{103,113,0}},
/* 39859 */ {(13<<2)|2,{103,113,0}},
/* 39860 */ {(13<<2)|2,{103,113,0}},
/* 39861 */ {(13<<2)|2,{103,113,0}},
/* 39862 */ {(13<<2)|2,{103,113,0}},
/* 39863 */ {(13<<2)|2,{103,113,0}},
/* 39864 */ {(13<<2)|2,{103,118,0}},
/* 39865 */ {(13<<2)|2,{103,118,0}},
/* 39866 */ {(13<<2)|2,{103,118,0}},
/* 39867 */ {(13<<2)|2,{103,118,0}},
/* 39868 */ {(13<<2)|2,{103,118,0}},
/* 39869 */ {(13<<2)|2,{103,118,0}},
/* 39870 */ {(13<<2)|2,{103,118,0}},
/* 39871 */ {(13<<2)|2,{103,118,0}},
/* 39872 */ {(13<<2)|2,{103,119,0}},
/* 39873 */ {(13<<2)|2,{103,119,0}},
/* 39874 */ {(13<<2)|2,{103,119,0}},
/* 39875 */ {(13<<2)|2,{103,119,0}},
/* 39876 */ {(13<<2)|2,{103,119,0}},
/* 39877 */ {(13<<2)|2,{103,119,0}},
/* 39878 */ {(13<<2)|2,{103,119,0}},
/* 39879 */ {(13<<2)|2,{103,119,0}},
/* 39880 */ {(13<<2)|2,{103,120,0}},
/* 39881 */ {(13<<2)|2,{103,120,0}},
/* 39882 */ {(13<<2)|2,{103,120,0}},
/* 39883 */ {(13<<2)|2,{103,120,0}},
/* 39884 */ {(13<<2)|2,{103,120,0}},
/* 39885 */ {(13<<2)|2,{103,120,0}},
/* 39886 */ {(13<<2)|2,{103,120,0}},
/* 39887 */ {(13<<2)|2,{103,120,0}},
/* 39888 */ {(13<<2)|2,{103,121,0}},
/* 39889 */ {(13<<2)|2,{103,121,0}},
/* 39890 */ {(13<<2)|2,{103,121,0}},
/* 39891 */ {(13<<2)|2,{103,121,0}},
/* 39892 */ {(13<<2)|2,{103,121,0}},
/* 39893 */ {(13<<2)|2,{103,121,0}},
/* 39894 */ {(13<<2)|2,{103,121,0}},
/* 39895 */ {(13<<2)|2,{103,121,0}},
/* 39896 */ {(13<<2)|2,{103,122,0}},
/* 39897 */ {(13<<2)|2,{103,122,0}},
/* 39898 */ {(13<<2)|2,{103,122,0}},
/* 39899 */ {(13<<2)|2,{103,122,0}},
/* 39900 */ {(13<<2)|2,{103,122,0}},
/* 39901 */ {(13<<2)|2,{103,122,0}},
/* 39902 */ {(13<<2)|2,{103,122,0}},
/* 39903 */ {(13<<2)|2,{103,122,0}},
/* 39904 */ {(14<<2)|2,{103,38,0}},
/* 39905 */ {(14<<2)|2,{103,38,0}},
/* 39906 */ {(14<<2)|2,{103,38,0}},
/* 39907 */ {(14<<2)|2,{103,38,0}},
/* 39908 */ {(14<<2)|2,{103,42,0}},
/* 39909 */ {(14<<2)|2,{103,42,0}},
/* 39910 */ {(14<<2)|2,{103,42,0}},
/* 39911 */ {(14<<2)|2,{103,42,0}},
/* 39912 */ {(14<<2)|2,{103,44,0}},
/* 39913 */ {(14<<2)|2,{103,44,0}},
/* 39914 */ {(14<<2)|2,{103,44,0}},
/* 39915 */ {(14<<2)|2,{103,44,0}},
/* 39916 */ {(14<<2)|2,{103,59,0}},
/* 39917 */ {(14<<2)|2,{103,59,0}},
/* 39918 */ {(14<<2)|2,{103,59,0}},
/* 39919 */ {(14<<2)|2,{103,59,0}},
/* 39920 */ {(14<<2)|2,{103,88,0}},
/* 39921 */ {(14<<2)|2,{103,88,0}},
/* 39922 */ {(14<<2)|2,{103,88,0}},
/* 39923 */ {(14<<2)|2,{103,88,0}},
/* 39924 */ {(14<<2)|2,{103,90,0}},
/* 39925 */ {(14<<2)|2,{103,90,0}},
/* 39926 */ {(14<<2)|2,{103,90,0}},
/* 39927 */ {(14<<2)|2,{103,90,0}},
/* 39928 */ {(16<<2)|2,{103,33,0}},
/* 39929 */ {(16<<2)|2,{103,34,0}},
/* 39930 */ {(16<<2)|2,{103,40,0}},
/* 39931 */ {(16<<2)|2,{103,41,0}},
/* 39932 */ {(16<<2)|2,{103,63,0}},
/* 39933 */ {(6<<2)|1,{103,0,0}},
/* 39934 */ {(6<<2)|1,{103,0,0}},
/* 39935 */ {(6<<2)|1,{103,0,0}},
/* 39936 */ {(16<<2)|3,{104,48,48}},
/* 39937 */ {(16<<2)|3,{104,48,49}},
/* 39938 */ {(16<<2)|3,{104,48,50}},
/* 39939 */ {(16<<2)|3,{104,48,97}},
/* 39940 */ {(16<<2)|3,{104,48,99}},
/* 39941 */ {(16<<2)|3,{104,48,101}},
/* 39942 */ {(16<<2)|3,{104,48,105}},
/* 39943 */ {(16<<2)|3,{104,48,111}},
/* 39944 */ {(16<<2)|3,{104,48,115}},
/* 39945 */ {(16<<2)|3,{104,48,116}},
/* 39946 */ {(11<<2)|2,{104,48,0}},
/* 39947 */ {(11<<2)|2,{104,48,0}},
/* 39948 */ {(11<<2)|2,{104,48,0}},
/* 39949 */ {(11<<2)|2,{104,48,0}},
/* 39950 */ {(11<<2)|2,{104,48,0}},
/* 39951 */ {(11<<2)|2,{104,48,0}},
/* 39952 */ {(11<<2)|2,{104,48,0}},
/* 39953 */ {(11<<2)|2,{104,48,0}},
/* 39954 */ {(11<<2)|2,{104,48,0}},
/* 39955 */ {(11<<2)|2,{104,48,0}},
/* 39956 */ {(11<<2)|2,{104,48,0}},
/* 39957 */ {(11<<2)|2,{104,48,0}},
/* 39958 */ {(11<<2)|2,{104,48,0}},
/* 39959 */ {(11<<2)|2,{104,48,0}},
/* 39960 */ {(11<<2)|2,{104,48,0}},
/* 39961 */ {(11<<2)|2,{104,48,0}},
/* 39962 */ {(11<<2)|2,{104,48,0}},
/* 39963 */ {(11<<2)|2,{104,48,0}},
/* 39964 */ {(11<<2)|2,{104,48,0}},
/* 39965 */ {(11<<2)|2,{104,48,0}},
/* 39966 */ {(11<<2)|2,{104,48,0}},
/* 39967 */ {(11<<2)|2,{104,48,0}},
/* 39968 */ {(16<<2)|3,{104,49,48}},
/* 39969 */ {(16<<2)|3,{104,49,49}},
/* 39970 */ {(16<<2)|3,{104,49,50}},
/* 39971 */ {(16<<2)|3,{104,49,97}},
/* 39972 */ {(16<<2)|3,{104,49,99}},
/* 39973 */ {(16<<2)|3,{104,49,101}},
/* 39974 */ {(16<<2)|3,{104,49,105}},
/* 39975 */ {(16<<2)|3,{104,49,111}},
/* 39976 */ {(16<<2)|3,{104,49,115}},
/* 39977 */ {(16<<2)|3,{104,49,116}},
/* 39978 */ {(11<<2)|2,{104,49,0}},
/* 39979 */ {(11<<2)|2,{104,49,0}},
/* 39980 */ {(11<<2)|2,{104,49,0}},
/* 39981 */ {(11<<2)|2,{104,49,0}},
/* 39982 */ {(11<<2)|2,{104,49,0}},
/* 39983 */ {(11<<2)|2,{104,49,0}},
/* 39984 */ {(11<<2)|2,{104,49,0}},
/* 39985 */ {(11<<2)|2,{104,49,0}},
/* 39986 */ {(11<<2)|2,{104,49,0}},
/* 39987 */ {(11<<2)|2,{104,49,0}},
/* 39988 */ {(11<<2)|2,{104,49,0}},
/* 39989 */ {(11<<2)|2,{104,49,0}},
/* 39990 */ {(11<<2)|2,{104,49,0}},
/* 39991 */ {(11<<2)|2,{104,49,0}},
/* 39992 */ {(11<<2)|2,{104,49,0}},
/* 39993 */ {(11<<2)|2,{104,49,0}},
/* 39994 */ {(11<<2)|2,{104,49,0}},
/* 39995 */ {(11<<2)|2,{104,49,0}},
/* 39996 */ {(11<<2)|2,{104,49,0}},
/* 39997 */ {(11<<2)|2,{104,49,0}},
/* 39998 */ {(11<<2)|2,{104,49,0}},
/* 39999 */ {(11<<2)|2,{104,49,0}},
/* 40000 */ {(16<<2)|3,{104,50,48}},
/* 40001 */ {(16<<2)|3,{104,50,49}},
/* 40002 */ {(16<<2)|3,{104,50,50}},
/* 40003 */ {(16<<2)|3,{104,50,97}},
/* 40004 */ {(16<<2)|3,{104,50,99}},
/* 40005 */ {(16<<2)|3,{104,50,101}},
/* 40006 */ {(16<<2)|3,{104,50,105}},
/* 40007 */ {(16<<2)|3,{104,50,111}},
/* 40008 */ {(16<<2)|3,{104,50,115}},
/* 40009 */ {(16<<2)|3,{104,50,116}},
/* 40010 */ {(11<<2)|2,{104,50,0}},
/* 40011 */ {(11<<2)|2,{104,50,0}},
/* 40012 */ {(11<<2)|2,{104,50,0}},
/* 40013 */ {(11<<2)|2,{104,50,0}},
/* 40014 */ {(11<<2)|2,{104,50,0}},
/* 40015 */ {(11<<2)|2,{104,50,0}},
/* 40016 */ {(11<<2)|2,{104,50,0}},
/* 40017 */ {(11<<2)|2,{104,50,0}},
/* 40018 */ {(11<<2)|2,{104,50,0}},
/* 40019 */ {(11<<2)|2,{104,50,0}},
/* 40020 */ {(11<<2)|2,{104,50,0}},
/* 40021 */ {(11<<2)|2,{104,50,0}},
/* 40022 */ {(11<<2)|2,{104,50,0}},
/* 40023 */ {(11<<2)|2,{104,50,0}},
/* 40024 */ {(11<<2)|2,{104,50,0}},
/* 40025 */ {(11<<2)|2,{104,50,0}},
/* 40026 */ {(11<<2)|2,{104,50,0}},
/* 40027 */ {(11<<2)|2,{104,50,0}},
/* 40028 */ {(11<<2)|2,{104,50,0}},
/* 40029 */ {(11<<2)|2,{104,50,0}},
/* 40030 */ {(11<<2)|2,{104,50,0}},
/* 40031 */ {(11<<2)|2,{104,50,0}},
/* 40032 */ {(16<<2)|3,{104,97,48}},
/* 40033 */ {(16<<2)|3,{104,97,49}},
/* 40034 */ {(16<<2)|3,{104,97,50}},
/* 40035 */ {(16<<2)|3,{104,97,97}},
/* 40036 */ {(16<<2)|3,{104,97,99}},
/* 40037 */ {(16<<2)|3,{104,97,101}},
/* 40038 */ {(16<<2)|3,{104,97,105}},
/* 40039 */ {(16<<2)|3,{104,97,111}},
/* 40040 */ {(16<<2)|3,{104,97,115}},
/* 40041 */ {(16<<2)|3,{104,97,116}},
/* 40042 */ {(11<<2)|2,{104,97,0}},
/* 40043 */ {(11<<2)|2,{104,97,0}},
/* 40044 */ {(11<<2)|2,{104,97,0}},
/* 40045 */ {(11<<2)|2,{104,97,0}},
/* 40046 */ {(11<<2)|2,{104,97,0}},
/* 40047 */ {(11<<2)|2,{104,97,0}},
/* 40048 */ {(11<<2)|2,{104,97,0}},
/* 40049 */ {(11<<2)|2,{104,97,0}},
/* 40050 */ {(11<<2)|2,{104,97,0}},
/* 40051 */ {(11<<2)|2,{104,97,0}},
/* 40052 */ {(11<<2)|2,{104,97,0}},
/* 40053 */ {(11<<2)|2,{104,97,0}},
/* 40054 */ {(11<<2)|2,{104,97,0}},
/* 40055 */ {(11<<2)|2,{104,97,0}},
/* 40056 */ {(11<<2)|2,{104,97,0}},
/* 40057 */ {(11<<2)|2,{104,97,0}},
/* 40058 */ {(11<<2)|2,{104,97,0}},
/* 40059 */ {(11<<2)|2,{104,97,0}},
/* 40060 */ {(11<<2)|2,{104,97,0}},
/* 40061 */ {(11<<2)|2,{104,97,0}},
/* 40062 */ {(11<<2)|2,{104,97,0}},
/* 40063 */ {(11<<2)|2,{104,97,0}},
/* 40064 */ {(16<<2)|3,{104,99,48}},
/* 40065 */ {(16<<2)|3,{104,99,49}},
/* 40066 */ {(16<<2)|3,{104,99,50}},
/* 40067 */ {(16<<2)|3,{104,99,97}},
/* 40068 */ {(16<<2)|3,{104,99,99}},
/* 40069 */ {(16<<2)|3,{104,99,101}},
/* 40070 */ {(16<<2)|3,{104,99,105}},
/* 40071 */ {(16<<2)|3,{104,99,111}},
/* 40072 */ {(16<<2)|3,{104,99,115}},
/* 40073 */ {(16<<2)|3,{104,99,116}},
/* 40074 */ {(11<<2)|2,{104,99,0}},
/* 40075 */ {(11<<2)|2,{104,99,0}},
/* 40076 */ {(11<<2)|2,{104,99,0}},
/* 40077 */ {(11<<2)|2,{104,99,0}},
/* 40078 */ {(11<<2)|2,{104,99,0}},
/* 40079 */ {(11<<2)|2,{104,99,0}},
/* 40080 */ {(11<<2)|2,{104,99,0}},
/* 40081 */ {(11<<2)|2,{104,99,0}},
/* 40082 */ {(11<<2)|2,{104,99,0}},
/* 40083 */ {(11<<2)|2,{104,99,0}},
/* 40084 */ {(11<<2)|2,{104,99,0}},
/* 40085 */ {(11<<2)|2,{104,99,0}},
/* 40086 */ {(11<<2)|2,{104,99,0}},
/* 40087 */ {(11<<2)|2,{104,99,0}},
/* 40088 */ {(11<<2)|2,{104,99,0}},
/* 40089 */ {(11<<2)|2,{104,99,0}},
/* 40090 */ {(11<<2)|2,{104,99,0}},
/* 40091 */ {(11<<2)|2,{104,99,0}},
/* 40092 */ {(11<<2)|2,{104,99,0}},
/* 40093 */ {(11<<2)|2,{104,99,0}},
/* 40094 */ {(11<<2)|2,{104,99,0}},
/* 40095 */ {(11<<2)|2,{104,99,0}},
/* 40096 */ {(16<<2)|3,{104,101,48}},
/* 40097 */ {(16<<2)|3,{104,101,49}},
/* 40098 */ {(16<<2)|3,{104,101,50}},
/* 40099 */ {(16<<2)|3,{104,101,97}},
/* 40100 */ {(16<<2)|3,{104,101,99}},
/* 40101 */ {(16<<2)|3,{104,101,101}},
/* 40102 */ {(16<<2)|3,{104,101,105}},
/* 40103 */ {(16<<2)|3,{104,101,111}},
/* 40104 */ {(16<<2)|3,{104,101,115}},
/* 40105 */ {(16<<2)|3,{104,101,116}},
/* 40106 */ {(11<<2)|2,{104,101,0}},
/* 40107 */ {(11<<2)|2,{104,101,0}},
/* 40108 */ {(11<<2)|2,{104,101,0}},
/* 40109 */ {(11<<2)|2,{104,101,0}},
/* 40110 */ {(11<<2)|2,{104,101,0}},
/* 40111 */ {(11<<2)|2,{104,101,0}},
/* 40112 */ {(11<<2)|2,{104,101,0}},
/* 40113 */ {(11<<2)|2,{104,101,0}},
/* 40114 */ {(11<<2)|2,{104,101,0}},
/* 40115 */ {(11<<2)|2,{104,101,0}},
/* 40116 */ {(11<<2)|2,{104,101,0}},
/* 40117 */ {(11<<2)|2,{104,101,0}},
/* 40118 */ {(11<<2)|2,{104,101,0}},
/* 40119 */ {(11<<2)|2,{104,101,0}},
/* 40120 */ {(11<<2)|2,{104,101,0}},
/* 40121 */ {(11<<2)|2,{104,101,0}},
/* 40122 */ {(11<<2)|2,{104,101,0}},
/* 40123 */ {(11<<2)|2,{104,101,0}},
/* 40124 */ {(11<<2)|2,{104,101,0}},
/* 40125 */ {(11<<2)|2,{104,101,0}},
/* 40126 */ {(11<<2)|2,{104,101,0}},
/* 40127 */ {(11<<2)|2,{104,101,0}},
/* 40128 */ {(16<<2)|3,{104,105,48}},
/* 40129 */ {(16<<2)|3,{104,105,49}},
/* 40130 */ {(16<<2)|3,{104,105,50}},
/* 40131 */ {(16<<2)|3,{104,105,97}},
/* 40132 */ {(16<<2)|3,{104,105,99}},
/* 40133 */ {(16<<2)|3,{104,105,101}},
/* 40134 */ {(16<<2)|3,{104,105,105}},
/* 40135 */ {(16<<2)|3,{104,105,111}},
/* 40136 */ {(16<<2)|3,{104,105,115}},
/* 40137 */ {(16<<2)|3,{104,105,116}},
/* 40138 */ {(11<<2)|2,{104,105,0}},
/* 40139 */ {(11<<2)|2,{104,105,0}},
/* 40140 */ {(11<<2)|2,{104,105,0}},
/* 40141 */ {(11<<2)|2,{104,105,0}},
/* 40142 */ {(11<<2)|2,{104,105,0}},
/* 40143 */ {(11<<2)|2,{104,105,0}},
/* 40144 */ {(11<<2)|2,{104,105,0}},
/* 40145 */ {(11<<2)|2,{104,105,0}},
/* 40146 */ {(11<<2)|2,{104,105,0}},
/* 40147 */ {(11<<2)|2,{104,105,0}},
/* 40148 */ {(11<<2)|2,{104,105,0}},
/* 40149 */ {(11<<2)|2,{104,105,0}},
/* 40150 */ {(11<<2)|2,{104,105,0}},
/* 40151 */ {(11<<2)|2,{104,105,0}},
/* 40152 */ {(11<<2)|2,{104,105,0}},
/* 40153 */ {(11<<2)|2,{104,105,0}},
/* 40154 */ {(11<<2)|2,{104,105,0}},
/* 40155 */ {(11<<2)|2,{104,105,0}},
/* 40156 */ {(11<<2)|2,{104,105,0}},
/* 40157 */ {(11<<2)|2,{104,105,0}},
/* 40158 */ {(11<<2)|2,{104,105,0}},
/* 40159 */ {(11<<2)|2,{104,105,0}},
/* 40160 */ {(16<<2)|3,{104,111,48}},
/* 40161 */ {(16<<2)|3,{104,111,49}},
/* 40162 */ {(16<<2)|3,{104,111,50}},
/* 40163 */ {(16<<2)|3,{104,111,97}},
/* 40164 */ {(16<<2)|3,{104,111,99}},
/* 40165 */ {(16<<2)|3,{104,111,101}},
/* 40166 */ {(16<<2)|3,{104,111,105}},
/* 40167 */ {(16<<2)|3,{104,111,111}},
/* 40168 */ {(16<<2)|3,{104,111,115}},
/* 40169 */ {(16<<2)|3,{104,111,116}},
/* 40170 */ {(11<<2)|2,{104,111,0}},
/* 40171 */ {(11<<2)|2,{104,111,0}},
/* 40172 */ {(11<<2)|2,{104,111,0}},
/* 40173 */ {(11<<2)|2,{104,111,0}},
/* 40174 */ {(11<<2)|2,{104,111,0}},
/* 40175 */ {(11<<2)|2,{104,111,0}},
/* 40176 */ {(11<<2)|2,{104,111,0}},
/* 40177 */ {(11<<2)|2,{104,111,0}},
/* 40178 */ {(11<<2)|2,{104,111,0}},
/* 40179 */ {(11<<2)|2,{104,111,0}},
/* 40180 */ {(11<<2)|2,{104,111,0}},
/* 40181 */ {(11<<2)|2,{104,111,0}},
/* 40182 */ {(11<<2)|2,{104,111,0}},
/* 40183 */ {(11<<2)|2,{104,111,0}},
/* 40184 */ {(11<<2)|2,{104,111,0}},
/* 40185 */ {(11<<2)|2,{104,111,0}},
/* 40186 */ {(11<<2)|2,{104,111,0}},
/* 40187 */ {(11<<2)|2,{104,111,0}},
/* 40188 */ {(11<<2)|2,{104,111,0}},
/* 40189 */ {(11<<2)|2,{104,111,0}},
/* 40190 */ {(11<<2)|2,{104,111,0}},
/* 40191 */ {(11<<2)|2,{104,111,0}},
/* 40192 */ {(16<<2)|3,{104,115,48}},
/* 40193 */ {(16<<2)|3,{104,115,49}},
/* 40194 */ {(16<<2)|3,{104,115,50}},
/* 40195 */ {(16<<2)|3,{104,115,97}},
/* 40196 */ {(16<<2)|3,{104,115,99}},
/* 40197 */ {(16<<2)|3,{104,115,101}},
/* 40198 */ {(16<<2)|3,{104,115,105}},
/* 40199 */ {(16<<2)|3,{104,115,111}},
/* 40200 */ {(16<<2)|3,{104,115,115}},
/* 40201 */ {(16<<2)|3,{104,115,116}},
/* 40202 */ {(11<<2)|2,{104,115,0}},
/* 40203 */ {(11<<2)|2,{104,115,0}},
/* 40204 */ {(11<<2)|2,{104,115,0}},
/* 40205 */ {(11<<2)|2,{104,115,0}},
/* 40206 */ {(11<<2)|2,{104,115,0}},
/* 40207 */ {(11<<2)|2,{104,115,0}},
/* 40208 */ {(11<<2)|2,{104,115,0}},
/* 40209 */ {(11<<2)|2,{104,115,0}},
/* 40210 */ {(11<<2)|2,{104,115,0}},
/* 40211 */ {(11<<2)|2,{104,115,0}},
/* 40212 */ {(11<<2)|2,{104,115,0}},
/* 40213 */ {(11<<2)|2,{104,115,0}},
/* 40214 */ {(11<<2)|2,{104,115,0}},
/* 40215 */ {(11<<2)|2,{104,115,0}},
/* 40216 */ {(11<<2)|2,{104,115,0}},
/* 40217 */ {(11<<2)|2,{104,115,0}},
/* 40218 */ {(11<<2)|2,{104,115,0}},
/* 40219 */ {(11<<2)|2,{104,115,0}},
/* 40220 */ {(11<<2)|2,{104,115,0}},
/* 40221 */ {(11<<2)|2,{104,115,0}},
/* 40222 */ {(11<<2)|2,{104,115,0}},
/* 40223 */ {(11<<2)|2,{104,115,0}},
/* 40224 */ {(16<<2)|3,{104,116,48}},
/* 40225 */ {(16<<2)|3,{104,116,49}},
/* 40226 */ {(16<<2)|3,{104,116,50}},
/* 40227 */ {(16<<2)|3,{104,116,97}},
/* 40228 */ {(16<<2)|3,{104,116,99}},
/* 40229 */ {(16<<2)|3,{104,116,101}},
/* 40230 */ {(16<<2)|3,{104,116,105}},
/* 40231 */ {(16<<2)|3,{104,116,111}},
/* 40232 */ {(16<<2)|3,{104,116,115}},
/* 40233 */ {(16<<2)|3,{104,116,116}},
/* 40234 */ {(11<<2)|2,{104,116,0}},
/* 40235 */ {(11<<2)|2,{104,116,0}},
/* 40236 */ {(11<<2)|2,{104,116,0}},
/* 40237 */ {(11<<2)|2,{104,116,0}},
/* 40238 */ {(11<<2)|2,{104,116,0}},
/* 40239 */ {(11<<2)|2,{104,116,0}},
/* 40240 */ {(11<<2)|2,{104,116,0}},
/* 40241 */ {(11<<2)|2,{104,116,0}},
/* 40242 */ {(11<<2)|2,{104,116,0}},
/* 40243 */ {(11<<2)|2,{104,116,0}},
/* 40244 */ {(11<<2)|2,{104,116,0}},
/* 40245 */ {(11<<2)|2,{104,116,0}},
/* 40246 */ {(11<<2)|2,{104,116,0}},
/* 40247 */ {(11<<2)|2,{104,116,0}},
/* 40248 */ {(11<<2)|2,{104,116,0}},
/* 40249 */ {(11<<2)|2,{104,116,0}},
/* 40250 */ {(11<<2)|2,{104,116,0}},
/* 40251 */ {(11<<2)|2,{104,116,0}},
/* 40252 */ {(11<<2)|2,{104,116,0}},
/* 40253 */ {(11<<2)|2,{104,116,0}},
/* 40254 */ {(11<<2)|2,{104,116,0}},
/* 40255 */ {(11<<2)|2,{104,116,0}},
/* 40256 */ {(12<<2)|2,{104,32,0}},
/* 40257 */ {(12<<2)|2,{104,32,0}},
/* 40258 */ {(12<<2)|2,{104,32,0}},
/* 40259 */ {(12<<2)|2,{104,32,0}},
/* 40260 */ {(12<<2)|2,{104,32,0}},
/* 40261 */ {(12<<2)|2,{104,32,0}},
/* 40262 */ {(12<<2)|2,{104,32,0}},
/* 40263 */ {(12<<2)|2,{104,32,0}},
/* 40264 */ {(12<<2)|2,{104,32,0}},
/* 40265 */ {(12<<2)|2,{104,32,0}},
/* 40266 */ {(12<<2)|2,{104,32,0}},
/* 40267 */ {(12<<2)|2,{104,32,0}},
/* 40268 */ {(12<<2)|2,{104,32,0}},
/* 40269 */ {(12<<2)|2,{104,32,0}},
/* 40270 */ {(12<<2)|2,{104,32,0}},
/* 40271 */ {(12<<2)|2,{104,32,0}},
/* 40272 */ {(12<<2)|2,{104,37,0}},
/* 40273 */ {(12<<2)|2,{104,37,0}},
/* 40274 */ {(12<<2)|2,{104,37,0}},
/* 40275 */ {(12<<2)|2,{104,37,0}},
/* 40276 */ {(12<<2)|2,{104,37,0}},
/* 40277 */ {(12<<2)|2,{104,37,0}},
/* 40278 */ {(12<<2)|2,{104,37,0}},
/* 40279 */ {(12<<2)|2,{104,37,0}},
/* 40280 */ {(12<<2)|2,{104,37,0}},
/* 40281 */ {(12<<2)|2,{104,37,0}},
/* 40282 */ {(12<<2)|2,{104,37,0}},
/* 40283 */ {(12<<2)|2,{104,37,0}},
/* 40284 */ {(12<<2)|2,{104,37,0}},
/* 40285 */ {(12<<2)|2,{104,37,0}},
/* 40286 */ {(12<<2)|2,{104,37,0}},
/* 40287 */ {(12<<2)|2,{104,37,0}},
/* 40288 */ {(12<<2)|2,{104,45,0}},
/* 40289 */ {(12<<2)|2,{104,45,0}},
/* 40290 */ {(12<<2)|2,{104,45,0}},
/* 40291 */ {(12<<2)|2,{104,45,0}},
/* 40292 */ {(12<<2)|2,{104,45,0}},
/* 40293 */ {(12<<2)|2,{104,45,0}},
/* 40294 */ {(12<<2)|2,{104,45,0}},
/* 40295 */ {(12<<2)|2,{104,45,0}},
/* 40296 */ {(12<<2)|2,{104,45,0}},
/* 40297 */ {(12<<2)|2,{104,45,0}},
/* 40298 */ {(12<<2)|2,{104,45,0}},
/* 40299 */ {(12<<2)|2,{104,45,0}},
/* 40300 */ {(12<<2)|2,{104,45,0}},
/* 40301 */ {(12<<2)|2,{104,45,0}},
/* 40302 */ {(12<<2)|2,{104,45,0}},
/* 40303 */ {(12<<2)|2,{104,45,0}},
/* 40304 */ {(12<<2)|2,{104,46,0}},
/* 40305 */ {(12<<2)|2,{104,46,0}},
/* 40306 */ {(12<<2)|2,{104,46,0}},
/* 40307 */ {(12<<2)|2,{104,46,0}},
/* 40308 */ {(12<<2)|2,{104,46,0}},
/* 40309 */ {(12<<2)|2,{104,46,0}},
/* 40310 */ {(12<<2)|2,{104,46,0}},
/* 40311 */ {(12<<2)|2,{104,46,0}},
/* 40312 */ {(12<<2)|2,{104,46,0}},
/* 40313 */ {(12<<2)|2,{104,46,0}},
/* 40314 */ {(12<<2)|2,{104,46,0}},
/* 40315 */ {(12<<2)|2,{104,46,0}},
/* 40316 */ {(12<<2)|2,{104,46,0}},
/* 40317 */ {(12<<2)|2,{104,46,0}},
/* 40318 */ {(12<<2)|2,{104,46,0}},
/* 40319 */ {(12<<2)|2,{104,46,0}},
/* 40320 */ {(12<<2)|2,{104,47,0}},
/* 40321 */ {(12<<2)|2,{104,47,0}},
/* 40322 */ {(12<<2)|2,{104,47,0}},
/* 40323 */ {(12<<2)|2,{104,47,0}},
/* 40324 */ {(12<<2)|2,{104,47,0}},
/* 40325 */ {(12<<2)|2,{104,47,0}},
/* 40326 */ {(12<<2)|2,{104,47,0}},
/* 40327 */ {(12<<2)|2,{104,47,0}},
/* 40328 */ {(12<<2)|2,{104,47,0}},
/* 40329 */ {(12<<2)|2,{104,47,0}},
/* 40330 */ {(12<<2)|2,{104,47,0}},
/* 40331 */ {(12<<2)|2,{104,47,0}},
/* 40332 */ {(12<<2)|2,{104,47,0}},
/* 40333 */ {(12<<2)|2,{104,47,0}},
/* 40334 */ {(12<<2)|2,{104,47,0}},
/* 40335 */ {(12<<2)|2,{104,47,0}},
/* 40336 */ {(12<<2)|2,{104,51,0}},
/* 40337 */ {(12<<2)|2,{104,51,0}},
/* 40338 */ {(12<<2)|2,{104,51,0}},
/* 40339 */ {(12<<2)|2,{104,51,0}},
/* 40340 */ {(12<<2)|2,{104,51,0}},
/* 40341 */ {(12<<2)|2,{104,51,0}},
/* 40342 */ {(12<<2)|2,{104,51,0}},
/* 40343 */ {(12<<2)|2,{104,51,0}},
/* 40344 */ {(12<<2)|2,{104,51,0}},
/* 40345 */ {(12<<2)|2,{104,51,0}},
/* 40346 */ {(12<<2)|2,{104,51,0}},
/* 40347 */ {(12<<2)|2,{104,51,0}},
/* 40348 */ {(12<<2)|2,{104,51,0}},
/* 40349 */ {(12<<2)|2,{104,51,0}},
/* 40350 */ {(12<<2)|2,{104,51,0}},
/* 40351 */ {(12<<2)|2,{104,51,0}},
/* 40352 */ {(12<<2)|2,{104,52,0}},
/* 40353 */ {(12<<2)|2,{104,52,0}},
/* 40354 */ {(12<<2)|2,{104,52,0}},
/* 40355 */ {(12<<2)|2,{104,52,0}},
/* 40356 */ {(12<<2)|2,{104,52,0}},
/* 40357 */ {(12<<2)|2,{104,52,0}},
/* 40358 */ {(12<<2)|2,{104,52,0}},
/* 40359 */ {(12<<2)|2,{104,52,0}},
/* 40360 */ {(12<<2)|2,{104,52,0}},
/* 40361 */ {(12<<2)|2,{104,52,0}},
/* 40362 */ {(12<<2)|2,{104,52,0}},
/* 40363 */ {(12<<2)|2,{104,52,0}},
/* 40364 */ {(12<<2)|2,{104,52,0}},
/* 40365 */ {(12<<2)|2,{104,52,0}},
/* 40366 */ {(12<<2)|2,{104,52,0}},
/* 40367 */ {(12<<2)|2,{104,52,0}},
/* 40368 */ {(12<<2)|2,{104,53,0}},
/* 40369 */ {(12<<2)|2,{104,53,0}},
/* 40370 */ {(12<<2)|2,{104,53,0}},
/* 40371 */ {(12<<2)|2,{104,53,0}},
/* 40372 */ {(12<<2)|2,{104,53,0}},
/* 40373 */ {(12<<2)|2,{104,53,0}},
/* 40374 */ {(12<<2)|2,{104,53,0}},
/* 40375 */ {(12<<2)|2,{104,53,0}},
/* 40376 */ {(12<<2)|2,{104,53,0}},
/* 40377 */ {(12<<2)|2,{104,53,0}},
/* 40378 */ {(12<<2)|2,{104,53,0}},
/* 40379 */ {(12<<2)|2,{104,53,0}},
/* 40380 */ {(12<<2)|2,{104,53,0}},
/* 40381 */ {(12<<2)|2,{104,53,0}},
/* 40382 */ {(12<<2)|2,{104,53,0}},
/* 40383 */ {(12<<2)|2,{104,53,0}},
/* 40384 */ {(12<<2)|2,{104,54,0}},
/* 40385 */ {(12<<2)|2,{104,54,0}},
/* 40386 */ {(12<<2)|2,{104,54,0}},
/* 40387 */ {(12<<2)|2,{104,54,0}},
/* 40388 */ {(12<<2)|2,{104,54,0}},
/* 40389 */ {(12<<2)|2,{104,54,0}},
/* 40390 */ {(12<<2)|2,{104,54,0}},
/* 40391 */ {(12<<2)|2,{104,54,0}},
/* 40392 */ {(12<<2)|2,{104,54,0}},
/* 40393 */ {(12<<2)|2,{104,54,0}},
/* 40394 */ {(12<<2)|2,{104,54,0}},
/* 40395 */ {(12<<2)|2,{104,54,0}},
/* 40396 */ {(12<<2)|2,{104,54,0}},
/* 40397 */ {(12<<2)|2,{104,54,0}},
/* 40398 */ {(12<<2)|2,{104,54,0}},
/* 40399 */ {(12<<2)|2,{104,54,0}},
/* 40400 */ {(12<<2)|2,{104,55,0}},
/* 40401 */ {(12<<2)|2,{104,55,0}},
/* 40402 */ {(12<<2)|2,{104,55,0}},
/* 40403 */ {(12<<2)|2,{104,55,0}},
/* 40404 */ {(12<<2)|2,{104,55,0}},
/* 40405 */ {(12<<2)|2,{104,55,0}},
/* 40406 */ {(12<<2)|2,{104,55,0}},
/* 40407 */ {(12<<2)|2,{104,55,0}},
/* 40408 */ {(12<<2)|2,{104,55,0}},
/* 40409 */ {(12<<2)|2,{104,55,0}},
/* 40410 */ {(12<<2)|2,{104,55,0}},
/* 40411 */ {(12<<2)|2,{104,55,0}},
/* 40412 */ {(12<<2)|2,{104,55,0}},
/* 40413 */ {(12<<2)|2,{104,55,0}},
/* 40414 */ {(12<<2)|2,{104,55,0}},
/* 40415 */ {(12<<2)|2,{104,55,0}},
/* 40416 */ {(12<<2)|2,{104,56,0}},
/* 40417 */ {(12<<2)|2,{104,56,0}},
/* 40418 */ {(12<<2)|2,{104,56,0}},
/* 40419 */ {(12<<2)|2,{104,56,0}},
/* 40420 */ {(12<<2)|2,{104,56,0}},
/* 40421 */ {(12<<2)|2,{104,56,0}},
/* 40422 */ {(12<<2)|2,{104,56,0}},
/* 40423 */ {(12<<2)|2,{104,56,0}},
/* 40424 */ {(12<<2)|2,{104,56,0}},
/* 40425 */ {(12<<2)|2,{104,56,0}},
/* 40426 */ {(12<<2)|2,{104,56,0}},
/* 40427 */ {(12<<2)|2,{104,56,0}},
/* 40428 */ {(12<<2)|2,{104,56,0}},
/* 40429 */ {(12<<2)|2,{104,56,0}},
/* 40430 */ {(12<<2)|2,{104,56,0}},
/* 40431 */ {(12<<2)|2,{104,56,0}},
/* 40432 */ {(12<<2)|2,{104,57,0}},
/* 40433 */ {(12<<2)|2,{104,57,0}},
/* 40434 */ {(12<<2)|2,{104,57,0}},
/* 40435 */ {(12<<2)|2,{104,57,0}},
/* 40436 */ {(12<<2)|2,{104,57,0}},
/* 40437 */ {(12<<2)|2,{104,57,0}},
/* 40438 */ {(12<<2)|2,{104,57,0}},
/* 40439 */ {(12<<2)|2,{104,57,0}},
/* 40440 */ {(12<<2)|2,{104,57,0}},
/* 40441 */ {(12<<2)|2,{104,57,0}},
/* 40442 */ {(12<<2)|2,{104,57,0}},
/* 40443 */ {(12<<2)|2,{104,57,0}},
/* 40444 */ {(12<<2)|2,{104,57,0}},
/* 40445 */ {(12<<2)|2,{104,57,0}},
/* 40446 */ {(12<<2)|2,{104,57,0}},
/* 40447 */ {(12<<2)|2,{104,57,0}},
/* 40448 */ {(12<<2)|2,{104,61,0}},
/* 40449 */ {(12<<2)|2,{104,61,0}},
/* 40450 */ {(12<<2)|2,{104,61,0}},
/* 40451 */ {(12<<2)|2,{104,61,0}},
/* 40452 */ {(12<<2)|2,{104,61,0}},
/* 40453 */ {(12<<2)|2,{104,61,0}},
/* 40454 */ {(12<<2)|2,{104,61,0}},
/* 40455 */ {(12<<2)|2,{104,61,0}},
/* 40456 */ {(12<<2)|2,{104,61,0}},
/* 40457 */ {(12<<2)|2,{104,61,0}},
/* 40458 */ {(12<<2)|2,{104,61,0}},
/* 40459 */ {(12<<2)|2,{104,61,0}},
/* 40460 */ {(12<<2)|2,{104,61,0}},
/* 40461 */ {(12<<2)|2,{104,61,0}},
/* 40462 */ {(12<<2)|2,{104,61,0}},
/* 40463 */ {(12<<2)|2,{104,61,0}},
/* 40464 */ {(12<<2)|2,{104,65,0}},
/* 40465 */ {(12<<2)|2,{104,65,0}},
/* 40466 */ {(12<<2)|2,{104,65,0}},
/* 40467 */ {(12<<2)|2,{104,65,0}},
/* 40468 */ {(12<<2)|2,{104,65,0}},
/* 40469 */ {(12<<2)|2,{104,65,0}},
/* 40470 */ {(12<<2)|2,{104,65,0}},
/* 40471 */ {(12<<2)|2,{104,65,0}},
/* 40472 */ {(12<<2)|2,{104,65,0}},
/* 40473 */ {(12<<2)|2,{104,65,0}},
/* 40474 */ {(12<<2)|2,{104,65,0}},
/* 40475 */ {(12<<2)|2,{104,65,0}},
/* 40476 */ {(12<<2)|2,{104,65,0}},
/* 40477 */ {(12<<2)|2,{104,65,0}},
/* 40478 */ {(12<<2)|2,{104,65,0}},
/* 40479 */ {(12<<2)|2,{104,65,0}},
/* 40480 */ {(12<<2)|2,{104,95,0}},
/* 40481 */ {(12<<2)|2,{104,95,0}},
/* 40482 */ {(12<<2)|2,{104,95,0}},
/* 40483 */ {(12<<2)|2,{104,95,0}},
/* 40484 */ {(12<<2)|2,{104,95,0}},
/* 40485 */ {(12<<2)|2,{104,95,0}},
/* 40486 */ {(12<<2)|2,{104,95,0}},
/* 40487 */ {(12<<2)|2,{104,95,0}},
/* 40488 */ {(12<<2)|2,{104,95,0}},
/* 40489 */ {(12<<2)|2,{104,95,0}},
/* 40490 */ {(12<<2)|2,{104,95,0}},
/* 40491 */ {(12<<2)|2,{104,95,0}},
/* 40492 */ {(12<<2)|2,{104,95,0}},
/* 40493 */ {(12<<2)|2,{104,95,0}},
/* 40494 */ {(12<<2)|2,{104,95,0}},
/* 40495 */ {(12<<2)|2,{104,95,0}},
/* 40496 */ {(12<<2)|2,{104,98,0}},
/* 40497 */ {(12<<2)|2,{104,98,0}},
/* 40498 */ {(12<<2)|2,{104,98,0}},
/* 40499 */ {(12<<2)|2,{104,98,0}},
/* 40500 */ {(12<<2)|2,{104,98,0}},
/* 40501 */ {(12<<2)|2,{104,98,0}},
/* 40502 */ {(12<<2)|2,{104,98,0}},
/* 40503 */ {(12<<2)|2,{104,98,0}},
/* 40504 */ {(12<<2)|2,{104,98,0}},
/* 40505 */ {(12<<2)|2,{104,98,0}},
/* 40506 */ {(12<<2)|2,{104,98,0}},
/* 40507 */ {(12<<2)|2,{104,98,0}},
/* 40508 */ {(12<<2)|2,{104,98,0}},
/* 40509 */ {(12<<2)|2,{104,98,0}},
/* 40510 */ {(12<<2)|2,{104,98,0}},
/* 40511 */ {(12<<2)|2,{104,98,0}},
/* 40512 */ {(12<<2)|2,{104,100,0}},
/* 40513 */ {(12<<2)|2,{104,100,0}},
/* 40514 */ {(12<<2)|2,{104,100,0}},
/* 40515 */ {(12<<2)|2,{104,100,0}},
/* 40516 */ {(12<<2)|2,{104,100,0}},
/* 40517 */ {(12<<2)|2,{104,100,0}},
/* 40518 */ {(12<<2)|2,{104,100,0}},
/* 40519 */ {(12<<2)|2,{104,100,0}},
/* 40520 */ {(12<<2)|2,{104,100,0}},
/* 40521 */ {(12<<2)|2,{104,100,0}},
/* 40522 */ {(12<<2)|2,{104,100,0}},
/* 40523 */ {(12<<2)|2,{104,100,0}},
/* 40524 */ {(12<<2)|2,{104,100,0}},
/* 40525 */ {(12<<2)|2,{104,100,0}},
/* 40526 */ {(12<<2)|2,{104,100,0}},
/* 40527 */ {(12<<2)|2,{104,100,0}},
/* 40528 */ {(12<<2)|2,{104,102,0}},
/* 40529 */ {(12<<2)|2,{104,102,0}},
/* 40530 */ {(12<<2)|2,{104,102,0}},
/* 40531 */ {(12<<2)|2,{104,102,0}},
/* 40532 */ {(12<<2)|2,{104,102,0}},
/* 40533 */ {(12<<2)|2,{104,102,0}},
/* 40534 */ {(12<<2)|2,{104,102,0}},
/* 40535 */ {(12<<2)|2,{104,102,0}},
/* 40536 */ {(12<<2)|2,{104,102,0}},
/* 40537 */ {(12<<2)|2,{104,102,0}},
/* 40538 */ {(12<<2)|2,{104,102,0}},
/* 40539 */ {(12<<2)|2,{104,102,0}},
/* 40540 */ {(12<<2)|2,{104,102,0}},
/* 40541 */ {(12<<2)|2,{104,102,0}},
/* 40542 */ {(12<<2)|2,{104,102,0}},
/* 40543 */ {(12<<2)|2,{104,102,0}},
/* 40544 */ {(12<<2)|2,{104,103,0}},
/* 40545 */ {(12<<2)|2,{104,103,0}},
/* 40546 */ {(12<<2)|2,{104,103,0}},
/* 40547 */ {(12<<2)|2,{104,103,0}},
/* 40548 */ {(12<<2)|2,{104,103,0}},
/* 40549 */ {(12<<2)|2,{104,103,0}},
/* 40550 */ {(12<<2)|2,{104,103,0}},
/* 40551 */ {(12<<2)|2,{104,103,0}},
/* 40552 */ {(12<<2)|2,{104,103,0}},
/* 40553 */ {(12<<2)|2,{104,103,0}},
/* 40554 */ {(12<<2)|2,{104,103,0}},
/* 40555 */ {(12<<2)|2,{104,103,0}},
/* 40556 */ {(12<<2)|2,{104,103,0}},
/* 40557 */ {(12<<2)|2,{104,103,0}},
/* 40558 */ {(12<<2)|2,{104,103,0}},
/* 40559 */ {(12<<2)|2,{104,103,0}},
/* 40560 */ {(12<<2)|2,{104,104,0}},
/* 40561 */ {(12<<2)|2,{104,104,0}},
/* 40562 */ {(12<<2)|2,{104,104,0}},
/* 40563 */ {(12<<2)|2,{104,104,0}},
/* 40564 */ {(12<<2)|2,{104,104,0}},
/* 40565 */ {(12<<2)|2,{104,104,0}},
/* 40566 */ {(12<<2)|2,{104,104,0}},
/* 40567 */ {(12<<2)|2,{104,104,0}},
/* 40568 */ {(12<<2)|2,{104,104,0}},
/* 40569 */ {(12<<2)|2,{104,104,0}},
/* 40570 */ {(12<<2)|2,{104,104,0}},
/* 40571 */ {(12<<2)|2,{104,104,0}},
/* 40572 */ {(12<<2)|2,{104,104,0}},
/* 40573 */ {(12<<2)|2,{104,104,0}},
/* 40574 */ {(12<<2)|2,{104,104,0}},
/* 40575 */ {(12<<2)|2,{104,104,0}},
/* 40576 */ {(12<<2)|2,{104,108,0}},
/* 40577 */ {(12<<2)|2,{104,108,0}},
/* 40578 */ {(12<<2)|2,{104,108,0}},
/* 40579 */ {(12<<2)|2,{104,108,0}},
/* 40580 */ {(12<<2)|2,{104,108,0}},
/* 40581 */ {(12<<2)|2,{104,108,0}},
/* 40582 */ {(12<<2)|2,{104,108,0}},
/* 40583 */ {(12<<2)|2,{104,108,0}},
/* 40584 */ {(12<<2)|2,{104,108,0}},
/* 40585 */ {(12<<2)|2,{104,108,0}},
/* 40586 */ {(12<<2)|2,{104,108,0}},
/* 40587 */ {(12<<2)|2,{104,108,0}},
/* 40588 */ {(12<<2)|2,{104,108,0}},
/* 40589 */ {(12<<2)|2,{104,108,0}},
/* 40590 */ {(12<<2)|2,{104,108,0}},
/* 40591 */ {(12<<2)|2,{104,108,0}},
/* 40592 */ {(12<<2)|2,{104,109,0}},
/* 40593 */ {(12<<2)|2,{104,109,0}},
/* 40594 */ {(12<<2)|2,{104,109,0}},
/* 40595 */ {(12<<2)|2,{104,109,0}},
/* 40596 */ {(12<<2)|2,{104,109,0}},
/* 40597 */ {(12<<2)|2,{104,109,0}},
/* 40598 */ {(12<<2)|2,{104,109,0}},
/* 40599 */ {(12<<2)|2,{104,109,0}},
/* 40600 */ {(12<<2)|2,{104,109,0}},
/* 40601 */ {(12<<2)|2,{104,109,0}},
/* 40602 */ {(12<<2)|2,{104,109,0}},
/* 40603 */ {(12<<2)|2,{104,109,0}},
/* 40604 */ {(12<<2)|2,{104,109,0}},
/* 40605 */ {(12<<2)|2,{104,109,0}},
/* 40606 */ {(12<<2)|2,{104,109,0}},
/* 40607 */ {(12<<2)|2,{104,109,0}},
/* 40608 */ {(12<<2)|2,{104,110,0}},
/* 40609 */ {(12<<2)|2,{104,110,0}},
/* 40610 */ {(12<<2)|2,{104,110,0}},
/* 40611 */ {(12<<2)|2,{104,110,0}},
/* 40612 */ {(12<<2)|2,{104,110,0}},
/* 40613 */ {(12<<2)|2,{104,110,0}},
/* 40614 */ {(12<<2)|2,{104,110,0}},
/* 40615 */ {(12<<2)|2,{104,110,0}},
/* 40616 */ {(12<<2)|2,{104,110,0}},
/* 40617 */ {(12<<2)|2,{104,110,0}},
/* 40618 */ {(12<<2)|2,{104,110,0}},
/* 40619 */ {(12<<2)|2,{104,110,0}},
/* 40620 */ {(12<<2)|2,{104,110,0}},
/* 40621 */ {(12<<2)|2,{104,110,0}},
/* 40622 */ {(12<<2)|2,{104,110,0}},
/* 40623 */ {(12<<2)|2,{104,110,0}},
/* 40624 */ {(12<<2)|2,{104,112,0}},
/* 40625 */ {(12<<2)|2,{104,112,0}},
/* 40626 */ {(12<<2)|2,{104,112,0}},
/* 40627 */ {(12<<2)|2,{104,112,0}},
/* 40628 */ {(12<<2)|2,{104,112,0}},
/* 40629 */ {(12<<2)|2,{104,112,0}},
/* 40630 */ {(12<<2)|2,{104,112,0}},
/* 40631 */ {(12<<2)|2,{104,112,0}},
/* 40632 */ {(12<<2)|2,{104,112,0}},
/* 40633 */ {(12<<2)|2,{104,112,0}},
/* 40634 */ {(12<<2)|2,{104,112,0}},
/* 40635 */ {(12<<2)|2,{104,112,0}},
/* 40636 */ {(12<<2)|2,{104,112,0}},
/* 40637 */ {(12<<2)|2,{104,112,0}},
/* 40638 */ {(12<<2)|2,{104,112,0}},
/* 40639 */ {(12<<2)|2,{104,112,0}},
/* 40640 */ {(12<<2)|2,{104,114,0}},
/* 40641 */ {(12<<2)|2,{104,114,0}},
/* 40642 */ {(12<<2)|2,{104,114,0}},
/* 40643 */ {(12<<2)|2,{104,114,0}},
/* 40644 */ {(12<<2)|2,{104,114,0}},
/* 40645 */ {(12<<2)|2,{104,114,0}},
/* 40646 */ {(12<<2)|2,{104,114,0}},
/* 40647 */ {(12<<2)|2,{104,114,0}},
/* 40648 */ {(12<<2)|2,{104,114,0}},
/* 40649 */ {(12<<2)|2,{104,114,0}},
/* 40650 */ {(12<<2)|2,{104,114,0}},
/* 40651 */ {(12<<2)|2,{104,114,0}},
/* 40652 */ {(12<<2)|2,{104,114,0}},
/* 40653 */ {(12<<2)|2,{104,114,0}},
/* 40654 */ {(12<<2)|2,{104,114,0}},
/* 40655 */ {(12<<2)|2,{104,114,0}},
/* 40656 */ {(12<<2)|2,{104,117,0}},
/* 40657 */ {(12<<2)|2,{104,117,0}},
/* 40658 */ {(12<<2)|2,{104,117,0}},
/* 40659 */ {(12<<2)|2,{104,117,0}},
/* 40660 */ {(12<<2)|2,{104,117,0}},
/* 40661 */ {(12<<2)|2,{104,117,0}},
/* 40662 */ {(12<<2)|2,{104,117,0}},
/* 40663 */ {(12<<2)|2,{104,117,0}},
/* 40664 */ {(12<<2)|2,{104,117,0}},
/* 40665 */ {(12<<2)|2,{104,117,0}},
/* 40666 */ {(12<<2)|2,{104,117,0}},
/* 40667 */ {(12<<2)|2,{104,117,0}},
/* 40668 */ {(12<<2)|2,{104,117,0}},
/* 40669 */ {(12<<2)|2,{104,117,0}},
/* 40670 */ {(12<<2)|2,{104,117,0}},
/* 40671 */ {(12<<2)|2,{104,117,0}},
/* 40672 */ {(13<<2)|2,{104,58,0}},
/* 40673 */ {(13<<2)|2,{104,58,0}},
/* 40674 */ {(13<<2)|2,{104,58,0}},
/* 40675 */ {(13<<2)|2,{104,58,0}},
/* 40676 */ {(13<<2)|2,{104,58,0}},
/* 40677 */ {(13<<2)|2,{104,58,0}},
/* 40678 */ {(13<<2)|2,{104,58,0}},
/* 40679 */ {(13<<2)|2,{104,58,0}},
/* 40680 */ {(13<<2)|2,{104,66,0}},
/* 40681 */ {(13<<2)|2,{104,66,0}},
/* 40682 */ {(13<<2)|2,{104,66,0}},
/* 40683 */ {(13<<2)|2,{104,66,0}},
/* 40684 */ {(13<<2)|2,{104,66,0}},
/* 40685 */ {(13<<2)|2,{104,66,0}},
/* 40686 */ {(13<<2)|2,{104,66,0}},
/* 40687 */ {(13<<2)|2,{104,66,0}},
/* 40688 */ {(13<<2)|2,{104,67,0}},
/* 40689 */ {(13<<2)|2,{104,67,0}},
/* 40690 */ {(13<<2)|2,{104,67,0}},
/* 40691 */ {(13<<2)|2,{104,67,0}},
/* 40692 */ {(13<<2)|2,{104,67,0}},
/* 40693 */ {(13<<2)|2,{104,67,0}},
/* 40694 */ {(13<<2)|2,{104,67,0}},
/* 40695 */ {(13<<2)|2,{104,67,0}},
/* 40696 */ {(13<<2)|2,{104,68,0}},
/* 40697 */ {(13<<2)|2,{104,68,0}},
/* 40698 */ {(13<<2)|2,{104,68,0}},
/* 40699 */ {(13<<2)|2,{104,68,0}},
/* 40700 */ {(13<<2)|2,{104,68,0}},
/* 40701 */ {(13<<2)|2,{104,68,0}},
/* 40702 */ {(13<<2)|2,{104,68,0}},
/* 40703 */ {(13<<2)|2,{104,68,0}},
/* 40704 */ {(13<<2)|2,{104,69,0}},
/* 40705 */ {(13<<2)|2,{104,69,0}},
/* 40706 */ {(13<<2)|2,{104,69,0}},
/* 40707 */ {(13<<2)|2,{104,69,0}},
/* 40708 */ {(13<<2)|2,{104,69,0}},
/* 40709 */ {(13<<2)|2,{104,69,0}},
/* 40710 */ {(13<<2)|2,{104,69,0}},
/* 40711 */ {(13<<2)|2,{104,69,0}},
/* 40712 */ {(13<<2)|2,{104,70,0}},
/* 40713 */ {(13<<2)|2,{104,70,0}},
/* 40714 */ {(13<<2)|2,{104,70,0}},
/* 40715 */ {(13<<2)|2,{104,70,0}},
/* 40716 */ {(13<<2)|2,{104,70,0}},
/* 40717 */ {(13<<2)|2,{104,70,0}},
/* 40718 */ {(13<<2)|2,{104,70,0}},
/* 40719 */ {(13<<2)|2,{104,70,0}},
/* 40720 */ {(13<<2)|2,{104,71,0}},
/* 40721 */ {(13<<2)|2,{104,71,0}},
/* 40722 */ {(13<<2)|2,{104,71,0}},
/* 40723 */ {(13<<2)|2,{104,71,0}},
/* 40724 */ {(13<<2)|2,{104,71,0}},
/* 40725 */ {(13<<2)|2,{104,71,0}},
/* 40726 */ {(13<<2)|2,{104,71,0}},
/* 40727 */ {(13<<2)|2,{104,71,0}},
/* 40728 */ {(13<<2)|2,{104,72,0}},
/* 40729 */ {(13<<2)|2,{104,72,0}},
/* 40730 */ {(13<<2)|2,{104,72,0}},
/* 40731 */ {(13<<2)|2,{104,72,0}},
/* 40732 */ {(13<<2)|2,{104,72,0}},
/* 40733 */ {(13<<2)|2,{104,72,0}},
/* 40734 */ {(13<<2)|2,{104,72,0}},
/* 40735 */ {(13<<2)|2,{104,72,0}},
/* 40736 */ {(13<<2)|2,{104,73,0}},
/* 40737 */ {(13<<2)|2,{104,73,0}},
/* 40738 */ {(13<<2)|2,{104,73,0}},
/* 40739 */ {(13<<2)|2,{104,73,0}},
/* 40740 */ {(13<<2)|2,{104,73,0}},
/* 40741 */ {(13<<2)|2,{104,73,0}},
/* 40742 */ {(13<<2)|2,{104,73,0}},
/* 40743 */ {(13<<2)|2,{104,73,0}},
/* 40744 */ {(13<<2)|2,{104,74,0}},
/* 40745 */ {(13<<2)|2,{104,74,0}},
/* 40746 */ {(13<<2)|2,{104,74,0}},
/* 40747 */ {(13<<2)|2,{104,74,0}},
/* 40748 */ {(13<<2)|2,{104,74,0}},
/* 40749 */ {(13<<2)|2,{104,74,0}},
/* 40750 */ {(13<<2)|2,{104,74,0}},
/* 40751 */ {(13<<2)|2,{104,74,0}},
/* 40752 */ {(13<<2)|2,{104,75,0}},
/* 40753 */ {(13<<2)|2,{104,75,0}},
/* 40754 */ {(13<<2)|2,{104,75,0}},
/* 40755 */ {(13<<2)|2,{104,75,0}},
/* 40756 */ {(13<<2)|2,{104,75,0}},
/* 40757 */ {(13<<2)|2,{104,75,0}},
/* 40758 */ {(13<<2)|2,{104,75,0}},
/* 40759 */ {(13<<2)|2,{104,75,0}},
/* 40760 */ {(13<<2)|2,{104,76,0}},
/* 40761 */ {(13<<2)|2,{104,76,0}},
/* 40762 */ {(13<<2)|2,{104,76,0}},
/* 40763 */ {(13<<2)|2,{104,76,0}},
/* 40764 */ {(13<<2)|2,{104,76,0}},
/* 40765 */ {(13<<2)|2,{104,76,0}},
/* 40766 */ {(13<<2)|2,{104,76,0}},
/* 40767 */ {(13<<2)|2,{104,76,0}},
/* 40768 */ {(13<<2)|2,{104,77,0}},
/* 40769 */ {(13<<2)|2,{104,77,0}},
/* 40770 */ {(13<<2)|2,{104,77,0}},
/* 40771 */ {(13<<2)|2,{104,77,0}},
/* 40772 */ {(13<<2)|2,{104,77,0}},
/* 40773 */ {(13<<2)|2,{104,77,0}},
/* 40774 */ {(13<<2)|2,{104,77,0}},
/* 40775 */ {(13<<2)|2,{104,77,0}},
/* 40776 */ {(13<<2)|2,{104,78,0}},
/* 40777 */ {(13<<2)|2,{104,78,0}},
/* 40778 */ {(13<<2)|2,{104,78,0}},
/* 40779 */ {(13<<2)|2,{104,78,0}},
/* 40780 */ {(13<<2)|2,{104,78,0}},
/* 40781 */ {(13<<2)|2,{104,78,0}},
/* 40782 */ {(13<<2)|2,{104,78,0}},
/* 40783 */ {(13<<2)|2,{104,78,0}},
/* 40784 */ {(13<<2)|2,{104,79,0}},
/* 40785 */ {(13<<2)|2,{104,79,0}},
/* 40786 */ {(13<<2)|2,{104,79,0}},
/* 40787 */ {(13<<2)|2,{104,79,0}},
/* 40788 */ {(13<<2)|2,{104,79,0}},
/* 40789 */ {(13<<2)|2,{104,79,0}},
/* 40790 */ {(13<<2)|2,{104,79,0}},
/* 40791 */ {(13<<2)|2,{104,79,0}},
/* 40792 */ {(13<<2)|2,{104,80,0}},
/* 40793 */ {(13<<2)|2,{104,80,0}},
/* 40794 */ {(13<<2)|2,{104,80,0}},
/* 40795 */ {(13<<2)|2,{104,80,0}},
/* 40796 */ {(13<<2)|2,{104,80,0}},
/* 40797 */ {(13<<2)|2,{104,80,0}},
/* 40798 */ {(13<<2)|2,{104,80,0}},
/* 40799 */ {(13<<2)|2,{104,80,0}},
/* 40800 */ {(13<<2)|2,{104,81,0}},
/* 40801 */ {(13<<2)|2,{104,81,0}},
/* 40802 */ {(13<<2)|2,{104,81,0}},
/* 40803 */ {(13<<2)|2,{104,81,0}},
/* 40804 */ {(13<<2)|2,{104,81,0}},
/* 40805 */ {(13<<2)|2,{104,81,0}},
/* 40806 */ {(13<<2)|2,{104,81,0}},
/* 40807 */ {(13<<2)|2,{104,81,0}},
/* 40808 */ {(13<<2)|2,{104,82,0}},
/* 40809 */ {(13<<2)|2,{104,82,0}},
/* 40810 */ {(13<<2)|2,{104,82,0}},
/* 40811 */ {(13<<2)|2,{104,82,0}},
/* 40812 */ {(13<<2)|2,{104,82,0}},
/* 40813 */ {(13<<2)|2,{104,82,0}},
/* 40814 */ {(13<<2)|2,{104,82,0}},
/* 40815 */ {(13<<2)|2,{104,82,0}},
/* 40816 */ {(13<<2)|2,{104,83,0}},
/* 40817 */ {(13<<2)|2,{104,83,0}},
/* 40818 */ {(13<<2)|2,{104,83,0}},
/* 40819 */ {(13<<2)|2,{104,83,0}},
/* 40820 */ {(13<<2)|2,{104,83,0}},
/* 40821 */ {(13<<2)|2,{104,83,0}},
/* 40822 */ {(13<<2)|2,{104,83,0}},
/* 40823 */ {(13<<2)|2,{104,83,0}},
/* 40824 */ {(13<<2)|2,{104,84,0}},
/* 40825 */ {(13<<2)|2,{104,84,0}},
/* 40826 */ {(13<<2)|2,{104,84,0}},
/* 40827 */ {(13<<2)|2,{104,84,0}},
/* 40828 */ {(13<<2)|2,{104,84,0}},
/* 40829 */ {(13<<2)|2,{104,84,0}},
/* 40830 */ {(13<<2)|2,{104,84,0}},
/* 40831 */ {(13<<2)|2,{104,84,0}},
/* 40832 */ {(13<<2)|2,{104,85,0}},
/* 40833 */ {(13<<2)|2,{104,85,0}},
/* 40834 */ {(13<<2)|2,{104,85,0}},
/* 40835 */ {(13<<2)|2,{104,85,0}},
/* 40836 */ {(13<<2)|2,{104,85,0}},
/* 40837 */ {(13<<2)|2,{104,85,0}},
/* 40838 */ {(13<<2)|2,{104,85,0}},
/* 40839 */ {(13<<2)|2,{104,85,0}},
/* 40840 */ {(13<<2)|2,{104,86,0}},
/* 40841 */ {(13<<2)|2,{104,86,0}},
/* 40842 */ {(13<<2)|2,{104,86,0}},
/* 40843 */ {(13<<2)|2,{104,86,0}},
/* 40844 */ {(13<<2)|2,{104,86,0}},
/* 40845 */ {(13<<2)|2,{104,86,0}},
/* 40846 */ {(13<<2)|2,{104,86,0}},
/* 40847 */ {(13<<2)|2,{104,86,0}},
/* 40848 */ {(13<<2)|2,{104,87,0}},
/* 40849 */ {(13<<2)|2,{104,87,0}},
/* 40850 */ {(13<<2)|2,{104,87,0}},
/* 40851 */ {(13<<2)|2,{104,87,0}},
/* 40852 */ {(13<<2)|2,{104,87,0}},
/* 40853 */ {(13<<2)|2,{104,87,0}},
/* 40854 */ {(13<<2)|2,{104,87,0}},
/* 40855 */ {(13<<2)|2,{104,87,0}},
/* 40856 */ {(13<<2)|2,{104,89,0}},
/* 40857 */ {(13<<2)|2,{104,89,0}},
/* 40858 */ {(13<<2)|2,{104,89,0}},
/* 40859 */ {(13<<2)|2,{104,89,0}},
/* 40860 */ {(13<<2)|2,{104,89,0}},
/* 40861 */ {(13<<2)|2,{104,89,0}},
/* 40862 */ {(13<<2)|2,{104,89,0}},
/* 40863 */ {(13<<2)|2,{104,89,0}},
/* 40864 */ {(13<<2)|2,{104,106,0}},
/* 40865 */ {(13<<2)|2,{104,106,0}},
/* 40866 */ {(13<<2)|2,{104,106,0}},
/* 40867 */ {(13<<2)|2,{104,106,0}},
/* 40868 */ {(13<<2)|2,{104,106,0}},
/* 40869 */ {(13<<2)|2,{104,106,0}},
/* 40870 */ {(13<<2)|2,{104,106,0}},
/* 40871 */ {(13<<2)|2,{104,106,0}},
/* 40872 */ {(13<<2)|2,{104,107,0}},
/* 40873 */ {(13<<2)|2,{104,107,0}},
/* 40874 */ {(13<<2)|2,{104,107,0}},
/* 40875 */ {(13<<2)|2,{104,107,0}},
/* 40876 */ {(13<<2)|2,{104,107,0}},
/* 40877 */ {(13<<2)|2,{104,107,0}},
/* 40878 */ {(13<<2)|2,{104,107,0}},
/* 40879 */ {(13<<2)|2,{104,107,0}},
/* 40880 */ {(13<<2)|2,{104,113,0}},
/* 40881 */ {(13<<2)|2,{104,113,0}},
/* 40882 */ {(13<<2)|2,{104,113,0}},
/* 40883 */ {(13<<2)|2,{104,113,0}},
/* 40884 */ {(13<<2)|2,{104,113,0}},
/* 40885 */ {(13<<2)|2,{104,113,0}},
/* 40886 */ {(13<<2)|2,{104,113,0}},
/* 40887 */ {(13<<2)|2,{104,113,0}},
/* 40888 */ {(13<<2)|2,{104,118,0}},
/* 40889 */ {(13<<2)|2,{104,118,0}},
/* 40890 */ {(13<<2)|2,{104,118,0}},
/* 40891 */ {(13<<2)|2,{104,118,0}},
/* 40892 */ {(13<<2)|2,{104,118,0}},
/* 40893 */ {(13<<2)|2,{104,118,0}},
/* 40894 */ {(13<<2)|2,{104,118,0}},
/* 40895 */ {(13<<2)|2,{104,118,0}},
/* 40896 */ {(13<<2)|2,{104,119,0}},
/* 40897 */ {(13<<2)|2,{104,119,0}},
/* 40898 */ {(13<<2)|2,{104,119,0}},
/* 40899 */ {(13<<2)|2,{104,119,0}},
/* 40900 */ {(13<<2)|2,{104,119,0}},
/* 40901 */ {(13<<2)|2,{104,119,0}},
/* 40902 */ {(13<<2)|2,{104,119,0}},
/* 40903 */ {(13<<2)|2,{104,119,0}},
/* 40904 */ {(13<<2)|2,{104,120,0}},
/* 40905 */ {(13<<2)|2,{104,120,0}},
/* 40906 */ {(13<<2)|2,{104,120,0}},
/* 40907 */ {(13<<2)|2,{104,120,0}},
/* 40908 */ {(13<<2)|2,{104,120,0}},
/* 40909 */ {(13<<2)|2,{104,120,0}},
/* 40910 */ {(13<<2)|2,{104,120,0}},
/* 40911 */ {(13<<2)|2,{104,120,0}},
/* 40912 */ {(13<<2)|2,{104,121,0}},
/* 40913 */ {(13<<2)|2,{104,121,0}},
/* 40914 */ {(13<<2)|2,{104,121,0}},
/* 40915 */ {(13<<2)|2,{104,121,0}},
/* 40916 */ {(13<<2)|2,{104,121,0}},
/* 40917 */ {(13<<2)|2,{104,121,0}},
/* 40918 */ {(13<<2)|2,{104,121,0}},
/* 40919 */ {(13<<2)|2,{104,121,0}},
/* 40920 */ {(13<<2)|2,{104,122,0}},
/* 40921 */ {(13<<2)|2,{104,122,0}},
/* 40922 */ {(13<<2)|2,{104,122,0}},
/* 40923 */ {(13<<2)|2,{104,122,0}},
/* 40924 */ {(13<<2)|2,{104,122,0}},
/* 40925 */ {(13<<2)|2,{104,122,0}},
/* 40926 */ {(13<<2)|2,{104,122,0}},
/* 40927 */ {(13<<2)|2,{104,122,0}},
/* 40928 */ {(14<<2)|2,{104,38,0}},
/* 40929 */ {(14<<2)|2,{104,38,0}},
/* 40930 */ {(14<<2)|2,{104,38,0}},
/* 40931 */ {(14<<2)|2,{104,38,0}},
/* 40932 */ {(14<<2)|2,{104,42,0}},
/* 40933 */ {(14<<2)|2,{104,42,0}},
/* 40934 */ {(14<<2)|2,{104,42,0}},
/* 40935 */ {(14<<2)|2,{104,42,0}},
/* 40936 */ {(14<<2)|2,{104,44,0}},
/* 40937 */ {(14<<2)|2,{104,44,0}},
/* 40938 */ {(14<<2)|2,{104,44,0}},
/* 40939 */ {(14<<2)|2,{104,44,0}},
/* 40940 */ {(14<<2)|2,{104,59,0}},
/* 40941 */ {(14<<2)|2,{104,59,0}},
/* 40942 */ {(14<<2)|2,{104,59,0}},
/* 40943 */ {(14<<2)|2,{104,59,0}},
/* 40944 */ {(14<<2)|2,{104,88,0}},
/* 40945 */ {(14<<2)|2,{104,88,0}},
/* 40946 */ {(14<<2)|2,{104,88,0}},
/* 40947 */ {(14<<2)|2,{104,88,0}},
/* 40948 */ {(14<<2)|2,{104,90,0}},
/* 40949 */ {(14<<2)|2,{104,90,0}},
/* 40950 */ {(14<<2)|2,{104,90,0}},
/* 40951 */ {(14<<2)|2,{104,90,0}},
/* 40952 */ {(16<<2)|2,{104,33,0}},
/* 40953 */ {(16<<2)|2,{104,34,0}},
/* 40954 */ {(16<<2)|2,{104,40,0}},
/* 40955 */ {(16<<2)|2,{104,41,0}},
/* 40956 */ {(16<<2)|2,{104,63,0}},
/* 40957 */ {(6<<2)|1,{104,0,0}},
/* 40958 */ {(6<<2)|1,{104,0,0}},
/* 40959 */ {(6<<2)|1,{104,0,0}},
/* 40960 */ {(16<<2)|3,{108,48,48}},
/* 40961 */ {(16<<2)|3,{108,48,49}},
/* 40962 */ {(16<<2)|3,{108,48,50}},
/* 40963 */ {(16<<2)|3,{108,48,97}},
/* 40964 */ {(16<<2)|3,{108,48,99}},
/* 40965 */ {(16<<2)|3,{108,48,101}},
/* 40966 */ {(16<<2)|3,{108,48,105}},
/* 40967 */ {(16<<2)|3,{108,48,111}},
/* 40968 */ {(16<<2)|3,{108,48,115}},
/* 40969 */ {(16<<2)|3,{108,48,116}},
/* 40970 */ {(11<<2)|2,{108,48,0}},
/* 40971 */ {(11<<2)|2,{108,48,0}},
/* 40972 */ {(11<<2)|2,{108,48,0}},
/* 40973 */ {(11<<2)|2,{108,48,0}},
/* 40974 */ {(11<<2)|2,{108,48,0}},
/* 40975 */ {(11<<2)|2,{108,48,0}},
/* 40976 */ {(11<<2)|2,{108,48,0}},
/* 40977 */ {(11<<2)|2,{108,48,0}},
/* 40978 */ {(11<<2)|2,{108,48,0}},
/* 40979 */ {(11<<2)|2,{108,48,0}},
/* 40980 */ {(11<<2)|2,{108,48,0}},
/* 40981 */ {(11<<2)|2,{108,48,0}},
/* 40982 */ {(11<<2)|2,{108,48,0}},
/* 40983 */ {(11<<2)|2,{108,48,0}},
/* 40984 */ {(11<<2)|2,{108,48,0}},
/* 40985 */ {(11<<2)|2,{108,48,0}},
/* 40986 */ {(11<<2)|2,{108,48,0}},
/* 40987 */ {(11<<2)|2,{108,48,0}},
/* 40988 */ {(11<<2)|2,{108,48,0}},
/* 40989 */ {(11<<2)|2,{108,48,0}},
/* 40990 */ {(11<<2)|2,{108,48,0}},
/* 40991 */ {(11<<2)|2,{108,48,0}},
/* 40992 */ {(16<<2)|3,{108,49,48}},
/* 40993 */ {(16<<2)|3,{108,49,49}},
/* 40994 */ {(16<<2)|3,{108,49,50}},
/* 40995 */ {(16<<2)|3,{108,49,97}},
/* 40996 */ {(16<<2)|3,{108,49,99}},
/* 40997 */ {(16<<2)|3,{108,49,101}},
/* 40998 */ {(16<<2)|3,{108,49,105}},
/* 40999 */ {(16<<2)|3,{108,49,111}},
/* 41000 */ {(16<<2)|3,{108,49,115}},
/* 41001 */ {(16<<2)|3,{108,49,116}},
/* 41002 */ {(11<<2)|2,{108,49,0}},
/* 41003 */ {(11<<2)|2,{108,49,0}},
/* 41004 */ {(11<<2)|2,{108,49,0}},
/* 41005 */ {(11<<2)|2,{108,49,0}},
/* 41006 */ {(11<<2)|2,{108,49,0}},
/* 41007 */ {(11<<2)|2,{108,49,0}},
/* 41008 */ {(11<<2)|2,{108,49,0}},
/* 41009 */ {(11<<2)|2,{108,49,0}},
/* 41010 */ {(11<<2)|2,{108,49,0}},
/* 41011 */ {(11<<2)|2,{108,49,0}},
/* 41012 */ {(11<<2)|2,{108,49,0}},
/* 41013 */ {(11<<2)|2,{108,49,0}},
/* 41014 */ {(11<<2)|2,{108,49,0}},
/* 41015 */ {(11<<2)|2,{108,49,0}},
/* 41016 */ {(11<<2)|2,{108,49,0}},
/* 41017 */ {(11<<2)|2,{108,49,0}},
/* 41018 */ {(11<<2)|2,{108,49,0}},
/* 41019 */ {(11<<2)|2,{108,49,0}},
/* 41020 */ {(11<<2)|2,{108,49,0}},
/* 41021 */ {(11<<2)|2,{108,49,0}},
/* 41022 */ {(11<<2)|2,{108,49,0}},
/* 41023 */ {(11<<2)|2,{108,49,0}},
/* 41024 */ {(16<<2)|3,{108,50,48}},
/* 41025 */ {(16<<2)|3,{108,50,49}},
/* 41026 */ {(16<<2)|3,{108,50,50}},
/* 41027 */ {(16<<2)|3,{108,50,97}},
/* 41028 */ {(16<<2)|3,{108,50,99}},
/* 41029 */ {(16<<2)|3,{108,50,101}},
/* 41030 */ {(16<<2)|3,{108,50,105}},
/* 41031 */ {(16<<2)|3,{108,50,111}},
/* 41032 */ {(16<<2)|3,{108,50,115}},
/* 41033 */ {(16<<2)|3,{108,50,116}},
/* 41034 */ {(11<<2)|2,{108,50,0}},
/* 41035 */ {(11<<2)|2,{108,50,0}},
/* 41036 */ {(11<<2)|2,{108,50,0}},
/* 41037 */ {(11<<2)|2,{108,50,0}},
/* 41038 */ {(11<<2)|2,{108,50,0}},
/* 41039 */ {(11<<2)|2,{108,50,0}},
/* 41040 */ {(11<<2)|2,{108,50,0}},
/* 41041 */ {(11<<2)|2,{108,50,0}},
/* 41042 */ {(11<<2)|2,{108,50,0}},
/* 41043 */ {(11<<2)|2,{108,50,0}},
/* 41044 */ {(11<<2)|2,{108,50,0}},
/* 41045 */ {(11<<2)|2,{108,50,0}},
/* 41046 */ {(11<<2)|2,{108,50,0}},
/* 41047 */ {(11<<2)|2,{108,50,0}},
/* 41048 */ {(11<<2)|2,{108,50,0}},
/* 41049 */ {(11<<2)|2,{108,50,0}},
/* 41050 */ {(11<<2)|2,{108,50,0}},
/* 41051 */ {(11<<2)|2,{108,50,0}},
/* 41052 */ {(11<<2)|2,{108,50,0}},
/* 41053 */ {(11<<2)|2,{108,50,0}},
/* 41054 */ {(11<<2)|2,{108,50,0}},
/* 41055 */ {(11<<2)|2,{108,50,0}},
/* 41056 */ {(16<<2)|3,{108,97,48}},
/* 41057 */ {(16<<2)|3,{108,97,49}},
/* 41058 */ {(16<<2)|3,{108,97,50}},
/* 41059 */ {(16<<2)|3,{108,97,97}},
/* 41060 */ {(16<<2)|3,{108,97,99}},
/* 41061 */ {(16<<2)|3,{108,97,101}},
/* 41062 */ {(16<<2)|3,{108,97,105}},
/* 41063 */ {(16<<2)|3,{108,97,111}},
/* 41064 */ {(16<<2)|3,{108,97,115}},
/* 41065 */ {(16<<2)|3,{108,97,116}},
/* 41066 */ {(11<<2)|2,{108,97,0}},
/* 41067 */ {(11<<2)|2,{108,97,0}},
/* 41068 */ {(11<<2)|2,{108,97,0}},
/* 41069 */ {(11<<2)|2,{108,97,0}},
/* 41070 */ {(11<<2)|2,{108,97,0}},
/* 41071 */ {(11<<2)|2,{108,97,0}},
/* 41072 */ {(11<<2)|2,{108,97,0}},
/* 41073 */ {(11<<2)|2,{108,97,0}},
/* 41074 */ {(11<<2)|2,{108,97,0}},
/* 41075 */ {(11<<2)|2,{108,97,0}},
/* 41076 */ {(11<<2)|2,{108,97,0}},
/* 41077 */ {(11<<2)|2,{108,97,0}},
/* 41078 */ {(11<<2)|2,{108,97,0}},
/* 41079 */ {(11<<2)|2,{108,97,0}},
/* 41080 */ {(11<<2)|2,{108,97,0}},
/* 41081 */ {(11<<2)|2,{108,97,0}},
/* 41082 */ {(11<<2)|2,{108,97,0}},
/* 41083 */ {(11<<2)|2,{108,97,0}},
/* 41084 */ {(11<<2)|2,{108,97,0}},
/* 41085 */ {(11<<2)|2,{108,97,0}},
/* 41086 */ {(11<<2)|2,{108,97,0}},
/* 41087 */ {(11<<2)|2,{108,97,0}},
/* 41088 */ {(16<<2)|3,{108,99,48}},
/* 41089 */ {(16<<2)|3,{108,99,49}},
/* 41090 */ {(16<<2)|3,{108,99,50}},
/* 41091 */ {(16<<2)|3,{108,99,97}},
/* 41092 */ {(16<<2)|3,{108,99,99}},
/* 41093 */ {(16<<2)|3,{108,99,101}},
/* 41094 */ {(16<<2)|3,{108,99,105}},
/* 41095 */ {(16<<2)|3,{108,99,111}},
/* 41096 */ {(16<<2)|3,{108,99,115}},
/* 41097 */ {(16<<2)|3,{108,99,116}},
/* 41098 */ {(11<<2)|2,{108,99,0}},
/* 41099 */ {(11<<2)|2,{108,99,0}},
/* 41100 */ {(11<<2)|2,{108,99,0}},
/* 41101 */ {(11<<2)|2,{108,99,0}},
/* 41102 */ {(11<<2)|2,{108,99,0}},
/* 41103 */ {(11<<2)|2,{108,99,0}},
/* 41104 */ {(11<<2)|2,{108,99,0}},
/* 41105 */ {(11<<2)|2,{108,99,0}},
/* 41106 */ {(11<<2)|2,{108,99,0}},
/* 41107 */ {(11<<2)|2,{108,99,0}},
/* 41108 */ {(11<<2)|2,{108,99,0}},
/* 41109 */ {(11<<2)|2,{108,99,0}},
/* 41110 */ {(11<<2)|2,{108,99,0}},
/* 41111 */ {(11<<2)|2,{108,99,0}},
/* 41112 */ {(11<<2)|2,{108,99,0}},
/* 41113 */ {(11<<2)|2,{108,99,0}},
/* 41114 */ {(11<<2)|2,{108,99,0}},
/* 41115 */ {(11<<2)|2,{108,99,0}},
/* 41116 */ {(11<<2)|2,{108,99,0}},
/* 41117 */ {(11<<2)|2,{108,99,0}},
/* 41118 */ {(11<<2)|2,{108,99,0}},
/* 41119 */ {(11<<2)|2,{108,99,0}},
/* 41120 */ {(16<<2)|3,{108,101,48}},
/* 41121 */ {(16<<2)|3,{108,101,49}},
/* 41122 */ {(16<<2)|3,{108,101,50}},
/* 41123 */ {(16<<2)|3,{108,101,97}},
/* 41124 */ {(16<<2)|3,{108,101,99}},
/* 41125 */ {(16<<2)|3,{108,101,101}},
/* 41126 */ {(16<<2)|3,{108,101,105}},
/* 41127 */ {(16<<2)|3,{108,101,111}},
/* 41128 */ {(16<<2)|3,{108,101,115}},
/* 41129 */ {(16<<2)|3,{108,101,116}},
/* 41130 */ {(11<<2)|2,{108,101,0}},
/* 41131 */ {(11<<2)|2,{108,101,0}},
/* 41132 */ {(11<<2)|2,{108,101,0}},
/* 41133 */ {(11<<2)|2,{108,101,0}},
/* 41134 */ {(11<<2)|2,{108,101,0}},
/* 41135 */ {(11<<2)|2,{108,101,0}},
/* 41136 */ {(11<<2)|2,{108,101,0}},
/* 41137 */ {(11<<2)|2,{108,101,0}},
/* 41138 */ {(11<<2)|2,{108,101,0}},
/* 41139 */ {(11<<2)|2,{108,101,0}},
/* 41140 */ {(11<<2)|2,{108,101,0}},
/* 41141 */ {(11<<2)|2,{108,101,0}},
/* 41142 */ {(11<<2)|2,{108,101,0}},
/* 41143 */ {(11<<2)|2,{108,101,0}},
/* 41144 */ {(11<<2)|2,{108,101,0}},
/* 41145 */ {(11<<2)|2,{108,101,0}},
/* 41146 */ {(11<<2)|2,{108,101,0}},
/* 41147 */ {(11<<2)|2,{108,101,0}},
/* 41148 */ {(11<<2)|2,{108,101,0}},
/* 41149 */ {(11<<2)|2,{108,101,0}},
/* 41150 */ {(11<<2)|2,{108,101,0}},
/* 41151 */ {(11<<2)|2,{108,101,0}},
/* 41152 */ {(16<<2)|3,{108,105,48}},
/* 41153 */ {(16<<2)|3,{108,105,49}},
/* 41154 */ {(16<<2)|3,{108,105,50}},
/* 41155 */ {(16<<2)|3,{108,105,97}},
/* 41156 */ {(16<<2)|3,{108,105,99}},
/* 41157 */ {(16<<2)|3,{108,105,101}},
/* 41158 */ {(16<<2)|3,{108,105,105}},
/* 41159 */ {(16<<2)|3,{108,105,111}},
/* 41160 */ {(16<<2)|3,{108,105,115}},
/* 41161 */ {(16<<2)|3,{108,105,116}},
/* 41162 */ {(11<<2)|2,{108,105,0}},
/* 41163 */ {(11<<2)|2,{108,105,0}},
/* 41164 */ {(11<<2)|2,{108,105,0}},
/* 41165 */ {(11<<2)|2,{108,105,0}},
/* 41166 */ {(11<<2)|2,{108,105,0}},
/* 41167 */ {(11<<2)|2,{108,105,0}},
/* 41168 */ {(11<<2)|2,{108,105,0}},
/* 41169 */ {(11<<2)|2,{108,105,0}},
/* 41170 */ {(11<<2)|2,{108,105,0}},
/* 41171 */ {(11<<2)|2,{108,105,0}},
/* 41172 */ {(11<<2)|2,{108,105,0}},
/* 41173 */ {(11<<2)|2,{108,105,0}},
/* 41174 */ {(11<<2)|2,{108,105,0}},
/* 41175 */ {(11<<2)|2,{108,105,0}},
/* 41176 */ {(11<<2)|2,{108,105,0}},
/* 41177 */ {(11<<2)|2,{108,105,0}},
/* 41178 */ {(11<<2)|2,{108,105,0}},
/* 41179 */ {(11<<2)|2,{108,105,0}},
/* 41180 */ {(11<<2)|2,{108,105,0}},
/* 41181 */ {(11<<2)|2,{108,105,0}},
/* 41182 */ {(11<<2)|2,{108,105,0}},
/* 41183 */ {(11<<2)|2,{108,105,0}},
/* 41184 */ {(16<<2)|3,{108,111,48}},
/* 41185 */ {(16<<2)|3,{108,111,49}},
/* 41186 */ {(16<<2)|3,{108,111,50}},
/* 41187 */ {(16<<2)|3,{108,111,97}},
/* 41188 */ {(16<<2)|3,{108,111,99}},
/* 41189 */ {(16<<2)|3,{108,111,101}},
/* 41190 */ {(16<<2)|3,{108,111,105}},
/* 41191 */ {(16<<2)|3,{108,111,111}},
/* 41192 */ {(16<<2)|3,{108,111,115}},
/* 41193 */ {(16<<2)|3,{108,111,116}},
/* 41194 */ {(11<<2)|2,{108,111,0}},
/* 41195 */ {(11<<2)|2,{108,111,0}},
/* 41196 */ {(11<<2)|2,{108,111,0}},
/* 41197 */ {(11<<2)|2,{108,111,0}},
/* 41198 */ {(11<<2)|2,{108,111,0}},
/* 41199 */ {(11<<2)|2,{108,111,0}},
/* 41200 */ {(11<<2)|2,{108,111,0}},
/* 41201 */ {(11<<2)|2,{108,111,0}},
/* 41202 */ {(11<<2)|2,{108,111,0}},
/* 41203 */ {(11<<2)|2,{108,111,0}},
/* 41204 */ {(11<<2)|2,{108,111,0}},
/* 41205 */ {(11<<2)|2,{108,111,0}},
/* 41206 */ {(11<<2)|2,{108,111,0}},
/* 41207 */ {(11<<2)|2,{108,111,0}},
/* 41208 */ {(11<<2)|2,{108,111,0}},
/* 41209 */ {(11<<2)|2,{108,111,0}},
/* 41210 */ {(11<<2)|2,{108,111,0}},
/* 41211 */ {(11<<2)|2,{108,111,0}},
/* 41212 */ {(11<<2)|2,{108,111,0}},
/* 41213 */ {(11<<2)|2,{108,111,0}},
/* 41214 */ {(11<<2)|2,{108,111,0}},
/* 41215 */ {(11<<2)|2,{108,111,0}},
/* 41216 */ {(16<<2)|3,{108,115,48}},
/* 41217 */ {(16<<2)|3,{108,115,49}},
/* 41218 */ {(16<<2)|3,{108,115,50}},
/* 41219 */ {(16<<2)|3,{108,115,97}},
/* 41220 */ {(16<<2)|3,{108,115,99}},
/* 41221 */ {(16<<2)|3,{108,115,101}},
/* 41222 */ {(16<<2)|3,{108,115,105}},
/* 41223 */ {(16<<2)|3,{108,115,111}},
/* 41224 */ {(16<<2)|3,{108,115,115}},
/* 41225 */ {(16<<2)|3,{108,115,116}},
/* 41226 */ {(11<<2)|2,{108,115,0}},
/* 41227 */ {(11<<2)|2,{108,115,0}},
/* 41228 */ {(11<<2)|2,{108,115,0}},
/* 41229 */ {(11<<2)|2,{108,115,0}},
/* 41230 */ {(11<<2)|2,{108,115,0}},
/* 41231 */ {(11<<2)|2,{108,115,0}},
/* 41232 */ {(11<<2)|2,{108,115,0}},
/* 41233 */ {(11<<2)|2,{108,115,0}},
/* 41234 */ {(11<<2)|2,{108,115,0}},
/* 41235 */ {(11<<2)|2,{108,115,0}},
/* 41236 */ {(11<<2)|2,{108,115,0}},
/* 41237 */ {(11<<2)|2,{108,115,0}},
/* 41238 */ {(11<<2)|2,{108,115,0}},
/* 41239 */ {(11<<2)|2,{108,115,0}},
/* 41240 */ {(11<<2)|2,{108,115,0}},
/* 41241 */ {(11<<2)|2,{108,115,0}},
/* 41242 */ {(11<<2)|2,{108,115,0}},
/* 41243 */ {(11<<2)|2,{108,115,0}},
/* 41244 */ {(11<<2)|2,{108,115,0}},
/* 41245 */ {(11<<2)|2,{108,115,0}},
/* 41246 */ {(11<<2)|2,{108,115,0}},
/* 41247 */ {(11<<2)|2,{108,115,0}},
/* 41248 */ {(16<<2)|3,{108,116,48}},
/* 41249 */ {(16<<2)|3,{108,116,49}},
/* 41250 */ {(16<<2)|3,{108,116,50}},
/* 41251 */ {(16<<2)|3,{108,116,97}},
/* 41252 */ {(16<<2)|3,{108,116,99}},
/* 41253 */ {(16<<2)|3,{108,116,101}},
/* 41254 */ {(16<<2)|3,{108,116,105}},
/* 41255 */ {(16<<2)|3,{108,116,111}},
/* 41256 */ {(16<<2)|3,{108,116,115}},
/* 41257 */ {(16<<2)|3,{108,116,116}},
/* 41258 */ {(11<<2)|2,{108,116,0}},
/* 41259 */ {(11<<2)|2,{108,116,0}},
/* 41260 */ {(11<<2)|2,{108,116,0}},
/* 41261 */ {(11<<2)|2,{108,116,0}},
/* 41262 */ {(11<<2)|2,{108,116,0}},
/* 41263 */ {(11<<2)|2,{108,116,0}},
/* 41264 */ {(11<<2)|2,{108,116,0}},
/* 41265 */ {(11<<2)|2,{108,116,0}},
/* 41266 */ {(11<<2)|2,{108,116,0}},
/* 41267 */ {(11<<2)|2,{108,116,0}},
/* 41268 */ {(11<<2)|2,{108,116,0}},
/* 41269 */ {(11<<2)|2,{108,116,0}},
/* 41270 */ {(11<<2)|2,{108,116,0}},
/* 41271 */ {(11<<2)|2,{108,116,0}},
/* 41272 */ {(11<<2)|2,{108,116,0}},
/* 41273 */ {(11<<2)|2,{108,116,0}},
/* 41274 */ {(11<<2)|2,{108,116,0}},
/* 41275 */ {(11<<2)|2,{108,116,0}},
/* 41276 */ {(11<<2)|2,{108,116,0}},
/* 41277 */ {(11<<2)|2,{108,116,0}},
/* 41278 */ {(11<<2)|2,{108,116,0}},
/* 41279 */ {(11<<2)|2,{108,116,0}},
/* 41280 */ {(12<<2)|2,{108,32,0}},
/* 41281 */ {(12<<2)|2,{108,32,0}},
/* 41282 */ {(12<<2)|2,{108,32,0}},
/* 41283 */ {(12<<2)|2,{108,32,0}},
/* 41284 */ {(12<<2)|2,{108,32,0}},
/* 41285 */ {(12<<2)|2,{108,32,0}},
/* 41286 */ {(12<<2)|2,{108,32,0}},
/* 41287 */ {(12<<2)|2,{108,32,0}},
/* 41288 */ {(12<<2)|2,{108,32,0}},
/* 41289 */ {(12<<2)|2,{108,32,0}},
/* 41290 */ {(12<<2)|2,{108,32,0}},
/* 41291 */ {(12<<2)|2,{108,32,0}},
/* 41292 */ {(12<<2)|2,{108,32,0}},
/* 41293 */ {(12<<2)|2,{108,32,0}},
/* 41294 */ {(12<<2)|2,{108,32,0}},
/* 41295 */ {(12<<2)|2,{108,32,0}},
/* 41296 */ {(12<<2)|2,{108,37,0}},
/* 41297 */ {(12<<2)|2,{108,37,0}},
/* 41298 */ {(12<<2)|2,{108,37,0}},
/* 41299 */ {(12<<2)|2,{108,37,0}},
/* 41300 */ {(12<<2)|2,{108,37,0}},
/* 41301 */ {(12<<2)|2,{108,37,0}},
/* 41302 */ {(12<<2)|2,{108,37,0}},
/* 41303 */ {(12<<2)|2,{108,37,0}},
/* 41304 */ {(12<<2)|2,{108,37,0}},
/* 41305 */ {(12<<2)|2,{108,37,0}},
/* 41306 */ {(12<<2)|2,{108,37,0}},
/* 41307 */ {(12<<2)|2,{108,37,0}},
/* 41308 */ {(12<<2)|2,{108,37,0}},
/* 41309 */ {(12<<2)|2,{108,37,0}},
/* 41310 */ {(12<<2)|2,{108,37,0}},
/* 41311 */ {(12<<2)|2,{108,37,0}},
/* 41312 */ {(12<<2)|2,{108,45,0}},
/* 41313 */ {(12<<2)|2,{108,45,0}},
/* 41314 */ {(12<<2)|2,{108,45,0}},
/* 41315 */ {(12<<2)|2,{108,45,0}},
/* 41316 */ {(12<<2)|2,{108,45,0}},
/* 41317 */ {(12<<2)|2,{108,45,0}},
/* 41318 */ {(12<<2)|2,{108,45,0}},
/* 41319 */ {(12<<2)|2,{108,45,0}},
/* 41320 */ {(12<<2)|2,{108,45,0}},
/* 41321 */ {(12<<2)|2,{108,45,0}},
/* 41322 */ {(12<<2)|2,{108,45,0}},
/* 41323 */ {(12<<2)|2,{108,45,0}},
/* 41324 */ {(12<<2)|2,{108,45,0}},
/* 41325 */ {(12<<2)|2,{108,45,0}},
/* 41326 */ {(12<<2)|2,{108,45,0}},
/* 41327 */ {(12<<2)|2,{108,45,0}},
/* 41328 */ {(12<<2)|2,{108,46,0}},
/* 41329 */ {(12<<2)|2,{108,46,0}},
/* 41330 */ {(12<<2)|2,{108,46,0}},
/* 41331 */ {(12<<2)|2,{108,46,0}},
/* 41332 */ {(12<<2)|2,{108,46,0}},
/* 41333 */ {(12<<2)|2,{108,46,0}},
/* 41334 */ {(12<<2)|2,{108,46,0}},
/* 41335 */ {(12<<2)|2,{108,46,0}},
/* 41336 */ {(12<<2)|2,{108,46,0}},
/* 41337 */ {(12<<2)|2,{108,46,0}},
/* 41338 */ {(12<<2)|2,{108,46,0}},
/* 41339 */ {(12<<2)|2,{108,46,0}},
/* 41340 */ {(12<<2)|2,{108,46,0}},
/* 41341 */ {(12<<2)|2,{108,46,0}},
/* 41342 */ {(12<<2)|2,{108,46,0}},
/* 41343 */ {(12<<2)|2,{108,46,0}},
/* 41344 */ {(12<<2)|2,{108,47,0}},
/* 41345 */ {(12<<2)|2,{108,47,0}},
/* 41346 */ {(12<<2)|2,{108,47,0}},
/* 41347 */ {(12<<2)|2,{108,47,0}},
/* 41348 */ {(12<<2)|2,{108,47,0}},
/* 41349 */ {(12<<2)|2,{108,47,0}},
/* 41350 */ {(12<<2)|2,{108,47,0}},
/* 41351 */ {(12<<2)|2,{108,47,0}},
/* 41352 */ {(12<<2)|2,{108,47,0}},
/* 41353 */ {(12<<2)|2,{108,47,0}},
/* 41354 */ {(12<<2)|2,{108,47,0}},
/* 41355 */ {(12<<2)|2,{108,47,0}},
/* 41356 */ {(12<<2)|2,{108,47,0}},
/* 41357 */ {(12<<2)|2,{108,47,0}},
/* 41358 */ {(12<<2)|2,{108,47,0}},
/* 41359 */ {(12<<2)|2,{108,47,0}},
/* 41360 */ {(12<<2)|2,{108,51,0}},
/* 41361 */ {(12<<2)|2,{108,51,0}},
/* 41362 */ {(12<<2)|2,{108,51,0}},
/* 41363 */ {(12<<2)|2,{108,51,0}},
/* 41364 */ {(12<<2)|2,{108,51,0}},
/* 41365 */ {(12<<2)|2,{108,51,0}},
/* 41366 */ {(12<<2)|2,{108,51,0}},
/* 41367 */ {(12<<2)|2,{108,51,0}},
/* 41368 */ {(12<<2)|2,{108,51,0}},
/* 41369 */ {(12<<2)|2,{108,51,0}},
/* 41370 */ {(12<<2)|2,{108,51,0}},
/* 41371 */ {(12<<2)|2,{108,51,0}},
/* 41372 */ {(12<<2)|2,{108,51,0}},
/* 41373 */ {(12<<2)|2,{108,51,0}},
/* 41374 */ {(12<<2)|2,{108,51,0}},
/* 41375 */ {(12<<2)|2,{108,51,0}},
/* 41376 */ {(12<<2)|2,{108,52,0}},
/* 41377 */ {(12<<2)|2,{108,52,0}},
/* 41378 */ {(12<<2)|2,{108,52,0}},
/* 41379 */ {(12<<2)|2,{108,52,0}},
/* 41380 */ {(12<<2)|2,{108,52,0}},
/* 41381 */ {(12<<2)|2,{108,52,0}},
/* 41382 */ {(12<<2)|2,{108,52,0}},
/* 41383 */ {(12<<2)|2,{108,52,0}},
/* 41384 */ {(12<<2)|2,{108,52,0}},
/* 41385 */ {(12<<2)|2,{108,52,0}},
/* 41386 */ {(12<<2)|2,{108,52,0}},
/* 41387 */ {(12<<2)|2,{108,52,0}},
/* 41388 */ {(12<<2)|2,{108,52,0}},
/* 41389 */ {(12<<2)|2,{108,52,0}},
/* 41390 */ {(12<<2)|2,{108,52,0}},
/* 41391 */ {(12<<2)|2,{108,52,0}},
/* 41392 */ {(12<<2)|2,{108,53,0}},
/* 41393 */ {(12<<2)|2,{108,53,0}},
/* 41394 */ {(12<<2)|2,{108,53,0}},
/* 41395 */ {(12<<2)|2,{108,53,0}},
/* 41396 */ {(12<<2)|2,{108,53,0}},
/* 41397 */ {(12<<2)|2,{108,53,0}},
/* 41398 */ {(12<<2)|2,{108,53,0}},
/* 41399 */ {(12<<2)|2,{108,53,0}},
/* 41400 */ {(12<<2)|2,{108,53,0}},
/* 41401 */ {(12<<2)|2,{108,53,0}},
/* 41402 */ {(12<<2)|2,{108,53,0}},
/* 41403 */ {(12<<2)|2,{108,53,0}},
/* 41404 */ {(12<<2)|2,{108,53,0}},
/* 41405 */ {(12<<2)|2,{108,53,0}},
/* 41406 */ {(12<<2)|2,{108,53,0}},
/* 41407 */ {(12<<2)|2,{108,53,0}},
/* 41408 */ {(12<<2)|2,{108,54,0}},
/* 41409 */ {(12<<2)|2,{108,54,0}},
/* 41410 */ {(12<<2)|2,{108,54,0}},
/* 41411 */ {(12<<2)|2,{108,54,0}},
/* 41412 */ {(12<<2)|2,{108,54,0}},
/* 41413 */ {(12<<2)|2,{108,54,0}},
/* 41414 */ {(12<<2)|2,{108,54,0}},
/* 41415 */ {(12<<2)|2,{108,54,0}},
/* 41416 */ {(12<<2)|2,{108,54,0}},
/* 41417 */ {(12<<2)|2,{108,54,0}},
/* 41418 */ {(12<<2)|2,{108,54,0}},
/* 41419 */ {(12<<2)|2,{108,54,0}},
/* 41420 */ {(12<<2)|2,{108,54,0}},
/* 41421 */ {(12<<2)|2,{108,54,0}},
/* 41422 */ {(12<<2)|2,{108,54,0}},
/* 41423 */ {(12<<2)|2,{108,54,0}},
/* 41424 */ {(12<<2)|2,{108,55,0}},
/* 41425 */ {(12<<2)|2,{108,55,0}},
/* 41426 */ {(12<<2)|2,{108,55,0}},
/* 41427 */ {(12<<2)|2,{108,55,0}},
/* 41428 */ {(12<<2)|2,{108,55,0}},
/* 41429 */ {(12<<2)|2,{108,55,0}},
/* 41430 */ {(12<<2)|2,{108,55,0}},
/* 41431 */ {(12<<2)|2,{108,55,0}},
/* 41432 */ {(12<<2)|2,{108,55,0}},
/* 41433 */ {(12<<2)|2,{108,55,0}},
/* 41434 */ {(12<<2)|2,{108,55,0}},
/* 41435 */ {(12<<2)|2,{108,55,0}},
/* 41436 */ {(12<<2)|2,{108,55,0}},
/* 41437 */ {(12<<2)|2,{108,55,0}},
/* 41438 */ {(12<<2)|2,{108,55,0}},
/* 41439 */ {(12<<2)|2,{108,55,0}},
/* 41440 */ {(12<<2)|2,{108,56,0}},
/* 41441 */ {(12<<2)|2,{108,56,0}},
/* 41442 */ {(12<<2)|2,{108,56,0}},
/* 41443 */ {(12<<2)|2,{108,56,0}},
/* 41444 */ {(12<<2)|2,{108,56,0}},
/* 41445 */ {(12<<2)|2,{108,56,0}},
/* 41446 */ {(12<<2)|2,{108,56,0}},
/* 41447 */ {(12<<2)|2,{108,56,0}},
/* 41448 */ {(12<<2)|2,{108,56,0}},
/* 41449 */ {(12<<2)|2,{108,56,0}},
/* 41450 */ {(12<<2)|2,{108,56,0}},
/* 41451 */ {(12<<2)|2,{108,56,0}},
/* 41452 */ {(12<<2)|2,{108,56,0}},
/* 41453 */ {(12<<2)|2,{108,56,0}},
/* 41454 */ {(12<<2)|2,{108,56,0}},
/* 41455 */ {(12<<2)|2,{108,56,0}},
/* 41456 */ {(12<<2)|2,{108,57,0}},
/* 41457 */ {(12<<2)|2,{108,57,0}},
/* 41458 */ {(12<<2)|2,{108,57,0}},
/* 41459 */ {(12<<2)|2,{108,57,0}},
/* 41460 */ {(12<<2)|2,{108,57,0}},
/* 41461 */ {(12<<2)|2,{108,57,0}},
/* 41462 */ {(12<<2)|2,{108,57,0}},
/* 41463 */ {(12<<2)|2,{108,57,0}},
/* 41464 */ {(12<<2)|2,{108,57,0}},
/* 41465 */ {(12<<2)|2,{108,57,0}},
/* 41466 */ {(12<<2)|2,{108,57,0}},
/* 41467 */ {(12<<2)|2,{108,57,0}},
/* 41468 */ {(12<<2)|2,{108,57,0}},
/* 41469 */ {(12<<2)|2,{108,57,0}},
/* 41470 */ {(12<<2)|2,{108,57,0}},
/* 41471 */ {(12<<2)|2,{108,57,0}},
/* 41472 */ {(12<<2)|2,{108,61,0}},
/* 41473 */ {(12<<2)|2,{108,61,0}},
/* 41474 */ {(12<<2)|2,{108,61,0}},
/* 41475 */ {(12<<2)|2,{108,61,0}},
/* 41476 */ {(12<<2)|2,{108,61,0}},
/* 41477 */ {(12<<2)|2,{108,61,0}},
/* 41478 */ {(12<<2)|2,{108,61,0}},
/* 41479 */ {(12<<2)|2,{108,61,0}},
/* 41480 */ {(12<<2)|2,{108,61,0}},
/* 41481 */ {(12<<2)|2,{108,61,0}},
/* 41482 */ {(12<<2)|2,{108,61,0}},
/* 41483 */ {(12<<2)|2,{108,61,0}},
/* 41484 */ {(12<<2)|2,{108,61,0}},
/* 41485 */ {(12<<2)|2,{108,61,0}},
/* 41486 */ {(12<<2)|2,{108,61,0}},
/* 41487 */ {(12<<2)|2,{108,61,0}},
/* 41488 */ {(12<<2)|2,{108,65,0}},
/* 41489 */ {(12<<2)|2,{108,65,0}},
/* 41490 */ {(12<<2)|2,{108,65,0}},
/* 41491 */ {(12<<2)|2,{108,65,0}},
/* 41492 */ {(12<<2)|2,{108,65,0}},
/* 41493 */ {(12<<2)|2,{108,65,0}},
/* 41494 */ {(12<<2)|2,{108,65,0}},
/* 41495 */ {(12<<2)|2,{108,65,0}},
/* 41496 */ {(12<<2)|2,{108,65,0}},
/* 41497 */ {(12<<2)|2,{108,65,0}},
/* 41498 */ {(12<<2)|2,{108,65,0}},
/* 41499 */ {(12<<2)|2,{108,65,0}},
/* 41500 */ {(12<<2)|2,{108,65,0}},
/* 41501 */ {(12<<2)|2,{108,65,0}},
/* 41502 */ {(12<<2)|2,{108,65,0}},
/* 41503 */ {(12<<2)|2,{108,65,0}},
/* 41504 */ {(12<<2)|2,{108,95,0}},
/* 41505 */ {(12<<2)|2,{108,95,0}},
/* 41506 */ {(12<<2)|2,{108,95,0}},
/* 41507 */ {(12<<2)|2,{108,95,0}},
/* 41508 */ {(12<<2)|2,{108,95,0}},
/* 41509 */ {(12<<2)|2,{108,95,0}},
/* 41510 */ {(12<<2)|2,{108,95,0}},
/* 41511 */ {(12<<2)|2,{108,95,0}},
/* 41512 */ {(12<<2)|2,{108,95,0}},
/* 41513 */ {(12<<2)|2,{108,95,0}},
/* 41514 */ {(12<<2)|2,{108,95,0}},
/* 41515 */ {(12<<2)|2,{108,95,0}},
/* 41516 */ {(12<<2)|2,{108,95,0}},
/* 41517 */ {(12<<2)|2,{108,95,0}},
/* 41518 */ {(12<<2)|2,{108,95,0}},
/* 41519 */ {(12<<2)|2,{108,95,0}},
/* 41520 */ {(12<<2)|2,{108,98,0}},
/* 41521 */ {(12<<2)|2,{108,98,0}},
/* 41522 */ {(12<<2)|2,{108,98,0}},
/* 41523 */ {(12<<2)|2,{108,98,0}},
/* 41524 */ {(12<<2)|2,{108,98,0}},
/* 41525 */ {(12<<2)|2,{108,98,0}},
/* 41526 */ {(12<<2)|2,{108,98,0}},
/* 41527 */ {(12<<2)|2,{108,98,0}},
/* 41528 */ {(12<<2)|2,{108,98,0}},
/* 41529 */ {(12<<2)|2,{108,98,0}},
/* 41530 */ {(12<<2)|2,{108,98,0}},
/* 41531 */ {(12<<2)|2,{108,98,0}},
/* 41532 */ {(12<<2)|2,{108,98,0}},
/* 41533 */ {(12<<2)|2,{108,98,0}},
/* 41534 */ {(12<<2)|2,{108,98,0}},
/* 41535 */ {(12<<2)|2,{108,98,0}},
/* 41536 */ {(12<<2)|2,{108,100,0}},
/* 41537 */ {(12<<2)|2,{108,100,0}},
/* 41538 */ {(12<<2)|2,{108,100,0}},
/* 41539 */ {(12<<2)|2,{108,100,0}},
/* 41540 */ {(12<<2)|2,{108,100,0}},
/* 41541 */ {(12<<2)|2,{108,100,0}},
/* 41542 */ {(12<<2)|2,{108,100,0}},
/* 41543 */ {(12<<2)|2,{108,100,0}},
/* 41544 */ {(12<<2)|2,{108,100,0}},
/* 41545 */ {(12<<2)|2,{108,100,0}},
/* 41546 */ {(12<<2)|2,{108,100,0}},
/* 41547 */ {(12<<2)|2,{108,100,0}},
/* 41548 */ {(12<<2)|2,{108,100,0}},
/* 41549 */ {(12<<2)|2,{108,100,0}},
/* 41550 */ {(12<<2)|2,{108,100,0}},
/* 41551 */ {(12<<2)|2,{108,100,0}},
/* 41552 */ {(12<<2)|2,{108,102,0}},
/* 41553 */ {(12<<2)|2,{108,102,0}},
/* 41554 */ {(12<<2)|2,{108,102,0}},
/* 41555 */ {(12<<2)|2,{108,102,0}},
/* 41556 */ {(12<<2)|2,{108,102,0}},
/* 41557 */ {(12<<2)|2,{108,102,0}},
/* 41558 */ {(12<<2)|2,{108,102,0}},
/* 41559 */ {(12<<2)|2,{108,102,0}},
/* 41560 */ {(12<<2)|2,{108,102,0}},
/* 41561 */ {(12<<2)|2,{108,102,0}},
/* 41562 */ {(12<<2)|2,{108,102,0}},
/* 41563 */ {(12<<2)|2,{108,102,0}},
/* 41564 */ {(12<<2)|2,{108,102,0}},
/* 41565 */ {(12<<2)|2,{108,102,0}},
/* 41566 */ {(12<<2)|2,{108,102,0}},
/* 41567 */ {(12<<2)|2,{108,102,0}},
/* 41568 */ {(12<<2)|2,{108,103,0}},
/* 41569 */ {(12<<2)|2,{108,103,0}},
/* 41570 */ {(12<<2)|2,{108,103,0}},
/* 41571 */ {(12<<2)|2,{108,103,0}},
/* 41572 */ {(12<<2)|2,{108,103,0}},
/* 41573 */ {(12<<2)|2,{108,103,0}},
/* 41574 */ {(12<<2)|2,{108,103,0}},
/* 41575 */ {(12<<2)|2,{108,103,0}},
/* 41576 */ {(12<<2)|2,{108,103,0}},
/* 41577 */ {(12<<2)|2,{108,103,0}},
/* 41578 */ {(12<<2)|2,{108,103,0}},
/* 41579 */ {(12<<2)|2,{108,103,0}},
/* 41580 */ {(12<<2)|2,{108,103,0}},
/* 41581 */ {(12<<2)|2,{108,103,0}},
/* 41582 */ {(12<<2)|2,{108,103,0}},
/* 41583 */ {(12<<2)|2,{108,103,0}},
/* 41584 */ {(12<<2)|2,{108,104,0}},
/* 41585 */ {(12<<2)|2,{108,104,0}},
/* 41586 */ {(12<<2)|2,{108,104,0}},
/* 41587 */ {(12<<2)|2,{108,104,0}},
/* 41588 */ {(12<<2)|2,{108,104,0}},
/* 41589 */ {(12<<2)|2,{108,104,0}},
/* 41590 */ {(12<<2)|2,{108,104,0}},
/* 41591 */ {(12<<2)|2,{108,104,0}},
/* 41592 */ {(12<<2)|2,{108,104,0}},
/* 41593 */ {(12<<2)|2,{108,104,0}},
/* 41594 */ {(12<<2)|2,{108,104,0}},
/* 41595 */ {(12<<2)|2,{108,104,0}},
/* 41596 */ {(12<<2)|2,{108,104,0}},
/* 41597 */ {(12<<2)|2,{108,104,0}},
/* 41598 */ {(12<<2)|2,{108,104,0}},
/* 41599 */ {(12<<2)|2,{108,104,0}},
/* 41600 */ {(12<<2)|2,{108,108,0}},
/* 41601 */ {(12<<2)|2,{108,108,0}},
/* 41602 */ {(12<<2)|2,{108,108,0}},
/* 41603 */ {(12<<2)|2,{108,108,0}},
/* 41604 */ {(12<<2)|2,{108,108,0}},
/* 41605 */ {(12<<2)|2,{108,108,0}},
/* 41606 */ {(12<<2)|2,{108,108,0}},
/* 41607 */ {(12<<2)|2,{108,108,0}},
/* 41608 */ {(12<<2)|2,{108,108,0}},
/* 41609 */ {(12<<2)|2,{108,108,0}},
/* 41610 */ {(12<<2)|2,{108,108,0}},
/* 41611 */ {(12<<2)|2,{108,108,0}},
/* 41612 */ {(12<<2)|2,{108,108,0}},
/* 41613 */ {(12<<2)|2,{108,108,0}},
/* 41614 */ {(12<<2)|2,{108,108,0}},
/* 41615 */ {(12<<2)|2,{108,108,0}},
/* 41616 */ {(12<<2)|2,{108,109,0}},
/* 41617 */ {(12<<2)|2,{108,109,0}},
/* 41618 */ {(12<<2)|2,{108,109,0}},
/* 41619 */ {(12<<2)|2,{108,109,0}},
/* 41620 */ {(12<<2)|2,{108,109,0}},
/* 41621 */ {(12<<2)|2,{108,109,0}},
/* 41622 */ {(12<<2)|2,{108,109,0}},
/* 41623 */ {(12<<2)|2,{108,109,0}},
/* 41624 */ {(12<<2)|2,{108,109,0}},
/* 41625 */ {(12<<2)|2,{108,109,0}},
/* 41626 */ {(12<<2)|2,{108,109,0}},
/* 41627 */ {(12<<2)|2,{108,109,0}},
/* 41628 */ {(12<<2)|2,{108,109,0}},
/* 41629 */ {(12<<2)|2,{108,109,0}},
/* 41630 */ {(12<<2)|2,{108,109,0}},
/* 41631 */ {(12<<2)|2,{108,109,0}},
/* 41632 */ {(12<<2)|2,{108,110,0}},
/* 41633 */ {(12<<2)|2,{108,110,0}},
/* 41634 */ {(12<<2)|2,{108,110,0}},
/* 41635 */ {(12<<2)|2,{108,110,0}},
/* 41636 */ {(12<<2)|2,{108,110,0}},
/* 41637 */ {(12<<2)|2,{108,110,0}},
/* 41638 */ {(12<<2)|2,{108,110,0}},
/* 41639 */ {(12<<2)|2,{108,110,0}},
/* 41640 */ {(12<<2)|2,{108,110,0}},
/* 41641 */ {(12<<2)|2,{108,110,0}},
/* 41642 */ {(12<<2)|2,{108,110,0}},
/* 41643 */ {(12<<2)|2,{108,110,0}},
/* 41644 */ {(12<<2)|2,{108,110,0}},
/* 41645 */ {(12<<2)|2,{108,110,0}},
/* 41646 */ {(12<<2)|2,{108,110,0}},
/* 41647 */ {(12<<2)|2,{108,110,0}},
/* 41648 */ {(12<<2)|2,{108,112,0}},
/* 41649 */ {(12<<2)|2,{108,112,0}},
/* 41650 */ {(12<<2)|2,{108,112,0}},
/* 41651 */ {(12<<2)|2,{108,112,0}},
/* 41652 */ {(12<<2)|2,{108,112,0}},
/* 41653 */ {(12<<2)|2,{108,112,0}},
/* 41654 */ {(12<<2)|2,{108,112,0}},
/* 41655 */ {(12<<2)|2,{108,112,0}},
/* 41656 */ {(12<<2)|2,{108,112,0}},
/* 41657 */ {(12<<2)|2,{108,112,0}},
/* 41658 */ {(12<<2)|2,{108,112,0}},
/* 41659 */ {(12<<2)|2,{108,112,0}},
/* 41660 */ {(12<<2)|2,{108,112,0}},
/* 41661 */ {(12<<2)|2,{108,112,0}},
/* 41662 */ {(12<<2)|2,{108,112,0}},
/* 41663 */ {(12<<2)|2,{108,112,0}},
/* 41664 */ {(12<<2)|2,{108,114,0}},
/* 41665 */ {(12<<2)|2,{108,114,0}},
/* 41666 */ {(12<<2)|2,{108,114,0}},
/* 41667 */ {(12<<2)|2,{108,114,0}},
/* 41668 */ {(12<<2)|2,{108,114,0}},
/* 41669 */ {(12<<2)|2,{108,114,0}},
/* 41670 */ {(12<<2)|2,{108,114,0}},
/* 41671 */ {(12<<2)|2,{108,114,0}},
/* 41672 */ {(12<<2)|2,{108,114,0}},
/* 41673 */ {(12<<2)|2,{108,114,0}},
/* 41674 */ {(12<<2)|2,{108,114,0}},
/* 41675 */ {(12<<2)|2,{108,114,0}},
/* 41676 */ {(12<<2)|2,{108,114,0}},
/* 41677 */ {(12<<2)|2,{108,114,0}},
/* 41678 */ {(12<<2)|2,{108,114,0}},
/* 41679 */ {(12<<2)|2,{108,114,0}},
/* 41680 */ {(12<<2)|2,{108,117,0}},
/* 41681 */ {(12<<2)|2,{108,117,0}},
/* 41682 */ {(12<<2)|2,{108,117,0}},
/* 41683 */ {(12<<2)|2,{108,117,0}},
/* 41684 */ {(12<<2)|2,{108,117,0}},
/* 41685 */ {(12<<2)|2,{108,117,0}},
/* 41686 */ {(12<<2)|2,{108,117,0}},
/* 41687 */ {(12<<2)|2,{108,117,0}},
/* 41688 */ {(12<<2)|2,{108,117,0}},
/* 41689 */ {(12<<2)|2,{108,117,0}},
/* 41690 */ {(12<<2)|2,{108,117,0}},
/* 41691 */ {(12<<2)|2,{108,117,0}},
/* 41692 */ {(12<<2)|2,{108,117,0}},
/* 41693 */ {(12<<2)|2,{108,117,0}},
/* 41694 */ {(12<<2)|2,{108,117,0}},
/* 41695 */ {(12<<2)|2,{108,117,0}},
/* 41696 */ {(13<<2)|2,{108,58,0}},
/* 41697 */ {(13<<2)|2,{108,58,0}},
/* 41698 */ {(13<<2)|2,{108,58,0}},
/* 41699 */ {(13<<2)|2,{108,58,0}},
/* 41700 */ {(13<<2)|2,{108,58,0}},
/* 41701 */ {(13<<2)|2,{108,58,0}},
/* 41702 */ {(13<<2)|2,{108,58,0}},
/* 41703 */ {(13<<2)|2,{108,58,0}},
/* 41704 */ {(13<<2)|2,{108,66,0}},
/* 41705 */ {(13<<2)|2,{108,66,0}},
/* 41706 */ {(13<<2)|2,{108,66,0}},
/* 41707 */ {(13<<2)|2,{108,66,0}},
/* 41708 */ {(13<<2)|2,{108,66,0}},
/* 41709 */ {(13<<2)|2,{108,66,0}},
/* 41710 */ {(13<<2)|2,{108,66,0}},
/* 41711 */ {(13<<2)|2,{108,66,0}},
/* 41712 */ {(13<<2)|2,{108,67,0}},
/* 41713 */ {(13<<2)|2,{108,67,0}},
/* 41714 */ {(13<<2)|2,{108,67,0}},
/* 41715 */ {(13<<2)|2,{108,67,0}},
/* 41716 */ {(13<<2)|2,{108,67,0}},
/* 41717 */ {(13<<2)|2,{108,67,0}},
/* 41718 */ {(13<<2)|2,{108,67,0}},
/* 41719 */ {(13<<2)|2,{108,67,0}},
/* 41720 */ {(13<<2)|2,{108,68,0}},
/* 41721 */ {(13<<2)|2,{108,68,0}},
/* 41722 */ {(13<<2)|2,{108,68,0}},
/* 41723 */ {(13<<2)|2,{108,68,0}},
/* 41724 */ {(13<<2)|2,{108,68,0}},
/* 41725 */ {(13<<2)|2,{108,68,0}},
/* 41726 */ {(13<<2)|2,{108,68,0}},
/* 41727 */ {(13<<2)|2,{108,68,0}},
/* 41728 */ {(13<<2)|2,{108,69,0}},
/* 41729 */ {(13<<2)|2,{108,69,0}},
/* 41730 */ {(13<<2)|2,{108,69,0}},
/* 41731 */ {(13<<2)|2,{108,69,0}},
/* 41732 */ {(13<<2)|2,{108,69,0}},
/* 41733 */ {(13<<2)|2,{108,69,0}},
/* 41734 */ {(13<<2)|2,{108,69,0}},
/* 41735 */ {(13<<2)|2,{108,69,0}},
/* 41736 */ {(13<<2)|2,{108,70,0}},
/* 41737 */ {(13<<2)|2,{108,70,0}},
/* 41738 */ {(13<<2)|2,{108,70,0}},
/* 41739 */ {(13<<2)|2,{108,70,0}},
/* 41740 */ {(13<<2)|2,{108,70,0}},
/* 41741 */ {(13<<2)|2,{108,70,0}},
/* 41742 */ {(13<<2)|2,{108,70,0}},
/* 41743 */ {(13<<2)|2,{108,70,0}},
/* 41744 */ {(13<<2)|2,{108,71,0}},
/* 41745 */ {(13<<2)|2,{108,71,0}},
/* 41746 */ {(13<<2)|2,{108,71,0}},
/* 41747 */ {(13<<2)|2,{108,71,0}},
/* 41748 */ {(13<<2)|2,{108,71,0}},
/* 41749 */ {(13<<2)|2,{108,71,0}},
/* 41750 */ {(13<<2)|2,{108,71,0}},
/* 41751 */ {(13<<2)|2,{108,71,0}},
/* 41752 */ {(13<<2)|2,{108,72,0}},
/* 41753 */ {(13<<2)|2,{108,72,0}},
/* 41754 */ {(13<<2)|2,{108,72,0}},
/* 41755 */ {(13<<2)|2,{108,72,0}},
/* 41756 */ {(13<<2)|2,{108,72,0}},
/* 41757 */ {(13<<2)|2,{108,72,0}},
/* 41758 */ {(13<<2)|2,{108,72,0}},
/* 41759 */ {(13<<2)|2,{108,72,0}},
/* 41760 */ {(13<<2)|2,{108,73,0}},
/* 41761 */ {(13<<2)|2,{108,73,0}},
/* 41762 */ {(13<<2)|2,{108,73,0}},
/* 41763 */ {(13<<2)|2,{108,73,0}},
/* 41764 */ {(13<<2)|2,{108,73,0}},
/* 41765 */ {(13<<2)|2,{108,73,0}},
/* 41766 */ {(13<<2)|2,{108,73,0}},
/* 41767 */ {(13<<2)|2,{108,73,0}},
/* 41768 */ {(13<<2)|2,{108,74,0}},
/* 41769 */ {(13<<2)|2,{108,74,0}},
/* 41770 */ {(13<<2)|2,{108,74,0}},
/* 41771 */ {(13<<2)|2,{108,74,0}},
/* 41772 */ {(13<<2)|2,{108,74,0}},
/* 41773 */ {(13<<2)|2,{108,74,0}},
/* 41774 */ {(13<<2)|2,{108,74,0}},
/* 41775 */ {(13<<2)|2,{108,74,0}},
/* 41776 */ {(13<<2)|2,{108,75,0}},
/* 41777 */ {(13<<2)|2,{108,75,0}},
/* 41778 */ {(13<<2)|2,{108,75,0}},
/* 41779 */ {(13<<2)|2,{108,75,0}},
/* 41780 */ {(13<<2)|2,{108,75,0}},
/* 41781 */ {(13<<2)|2,{108,75,0}},
/* 41782 */ {(13<<2)|2,{108,75,0}},
/* 41783 */ {(13<<2)|2,{108,75,0}},
/* 41784 */ {(13<<2)|2,{108,76,0}},
/* 41785 */ {(13<<2)|2,{108,76,0}},
/* 41786 */ {(13<<2)|2,{108,76,0}},
/* 41787 */ {(13<<2)|2,{108,76,0}},
/* 41788 */ {(13<<2)|2,{108,76,0}},
/* 41789 */ {(13<<2)|2,{108,76,0}},
/* 41790 */ {(13<<2)|2,{108,76,0}},
/* 41791 */ {(13<<2)|2,{108,76,0}},
/* 41792 */ {(13<<2)|2,{108,77,0}},
/* 41793 */ {(13<<2)|2,{108,77,0}},
/* 41794 */ {(13<<2)|2,{108,77,0}},
/* 41795 */ {(13<<2)|2,{108,77,0}},
/* 41796 */ {(13<<2)|2,{108,77,0}},
/* 41797 */ {(13<<2)|2,{108,77,0}},
/* 41798 */ {(13<<2)|2,{108,77,0}},
/* 41799 */ {(13<<2)|2,{108,77,0}},
/* 41800 */ {(13<<2)|2,{108,78,0}},
/* 41801 */ {(13<<2)|2,{108,78,0}},
/* 41802 */ {(13<<2)|2,{108,78,0}},
/* 41803 */ {(13<<2)|2,{108,78,0}},
/* 41804 */ {(13<<2)|2,{108,78,0}},
/* 41805 */ {(13<<2)|2,{108,78,0}},
/* 41806 */ {(13<<2)|2,{108,78,0}},
/* 41807 */ {(13<<2)|2,{108,78,0}},
/* 41808 */ {(13<<2)|2,{108,79,0}},
/* 41809 */ {(13<<2)|2,{108,79,0}},
/* 41810 */ {(13<<2)|2,{108,79,0}},
/* 41811 */ {(13<<2)|2,{108,79,0}},
/* 41812 */ {(13<<2)|2,{108,79,0}},
/* 41813 */ {(13<<2)|2,{108,79,0}},
/* 41814 */ {(13<<2)|2,{108,79,0}},
/* 41815 */ {(13<<2)|2,{108,79,0}},
/* 41816 */ {(13<<2)|2,{108,80,0}},
/* 41817 */ {(13<<2)|2,{108,80,0}},
/* 41818 */ {(13<<2)|2,{108,80,0}},
/* 41819 */ {(13<<2)|2,{108,80,0}},
/* 41820 */ {(13<<2)|2,{108,80,0}},
/* 41821 */ {(13<<2)|2,{108,80,0}},
/* 41822 */ {(13<<2)|2,{108,80,0}},
/* 41823 */ {(13<<2)|2,{108,80,0}},
/* 41824 */ {(13<<2)|2,{108,81,0}},
/* 41825 */ {(13<<2)|2,{108,81,0}},
/* 41826 */ {(13<<2)|2,{108,81,0}},
/* 41827 */ {(13<<2)|2,{108,81,0}},
/* 41828 */ {(13<<2)|2,{108,81,0}},
/* 41829 */ {(13<<2)|2,{108,81,0}},
/* 41830 */ {(13<<2)|2,{108,81,0}},
/* 41831 */ {(13<<2)|2,{108,81,0}},
/* 41832 */ {(13<<2)|2,{108,82,0}},
/* 41833 */ {(13<<2)|2,{108,82,0}},
/* 41834 */ {(13<<2)|2,{108,82,0}},
/* 41835 */ {(13<<2)|2,{108,82,0}},
/* 41836 */ {(13<<2)|2,{108,82,0}},
/* 41837 */ {(13<<2)|2,{108,82,0}},
/* 41838 */ {(13<<2)|2,{108,82,0}},
/* 41839 */ {(13<<2)|2,{108,82,0}},
/* 41840 */ {(13<<2)|2,{108,83,0}},
/* 41841 */ {(13<<2)|2,{108,83,0}},
/* 41842 */ {(13<<2)|2,{108,83,0}},
/* 41843 */ {(13<<2)|2,{108,83,0}},
/* 41844 */ {(13<<2)|2,{108,83,0}},
/* 41845 */ {(13<<2)|2,{108,83,0}},
/* 41846 */ {(13<<2)|2,{108,83,0}},
/* 41847 */ {(13<<2)|2,{108,83,0}},
/* 41848 */ {(13<<2)|2,{108,84,0}},
/* 41849 */ {(13<<2)|2,{108,84,0}},
/* 41850 */ {(13<<2)|2,{108,84,0}},
/* 41851 */ {(13<<2)|2,{108,84,0}},
/* 41852 */ {(13<<2)|2,{108,84,0}},
/* 41853 */ {(13<<2)|2,{108,84,0}},
/* 41854 */ {(13<<2)|2,{108,84,0}},
/* 41855 */ {(13<<2)|2,{108,84,0}},
/* 41856 */ {(13<<2)|2,{108,85,0}},
/* 41857 */ {(13<<2)|2,{108,85,0}},
/* 41858 */ {(13<<2)|2,{108,85,0}},
/* 41859 */ {(13<<2)|2,{108,85,0}},
/* 41860 */ {(13<<2)|2,{108,85,0}},
/* 41861 */ {(13<<2)|2,{108,85,0}},
/* 41862 */ {(13<<2)|2,{108,85,0}},
/* 41863 */ {(13<<2)|2,{108,85,0}},
/* 41864 */ {(13<<2)|2,{108,86,0}},
/* 41865 */ {(13<<2)|2,{108,86,0}},
/* 41866 */ {(13<<2)|2,{108,86,0}},
/* 41867 */ {(13<<2)|2,{108,86,0}},
/* 41868 */ {(13<<2)|2,{108,86,0}},
/* 41869 */ {(13<<2)|2,{108,86,0}},
/* 41870 */ {(13<<2)|2,{108,86,0}},
/* 41871 */ {(13<<2)|2,{108,86,0}},
/* 41872 */ {(13<<2)|2,{108,87,0}},
/* 41873 */ {(13<<2)|2,{108,87,0}},
/* 41874 */ {(13<<2)|2,{108,87,0}},
/* 41875 */ {(13<<2)|2,{108,87,0}},
/* 41876 */ {(13<<2)|2,{108,87,0}},
/* 41877 */ {(13<<2)|2,{108,87,0}},
/* 41878 */ {(13<<2)|2,{108,87,0}},
/* 41879 */ {(13<<2)|2,{108,87,0}},
/* 41880 */ {(13<<2)|2,{108,89,0}},
/* 41881 */ {(13<<2)|2,{108,89,0}},
/* 41882 */ {(13<<2)|2,{108,89,0}},
/* 41883 */ {(13<<2)|2,{108,89,0}},
/* 41884 */ {(13<<2)|2,{108,89,0}},
/* 41885 */ {(13<<2)|2,{108,89,0}},
/* 41886 */ {(13<<2)|2,{108,89,0}},
/* 41887 */ {(13<<2)|2,{108,89,0}},
/* 41888 */ {(13<<2)|2,{108,106,0}},
/* 41889 */ {(13<<2)|2,{108,106,0}},
/* 41890 */ {(13<<2)|2,{108,106,0}},
/* 41891 */ {(13<<2)|2,{108,106,0}},
/* 41892 */ {(13<<2)|2,{108,106,0}},
/* 41893 */ {(13<<2)|2,{108,106,0}},
/* 41894 */ {(13<<2)|2,{108,106,0}},
/* 41895 */ {(13<<2)|2,{108,106,0}},
/* 41896 */ {(13<<2)|2,{108,107,0}},
/* 41897 */ {(13<<2)|2,{108,107,0}},
/* 41898 */ {(13<<2)|2,{108,107,0}},
/* 41899 */ {(13<<2)|2,{108,107,0}},
/* 41900 */ {(13<<2)|2,{108,107,0}},
/* 41901 */ {(13<<2)|2,{108,107,0}},
/* 41902 */ {(13<<2)|2,{108,107,0}},
/* 41903 */ {(13<<2)|2,{108,107,0}},
/* 41904 */ {(13<<2)|2,{108,113,0}},
/* 41905 */ {(13<<2)|2,{108,113,0}},
/* 41906 */ {(13<<2)|2,{108,113,0}},
/* 41907 */ {(13<<2)|2,{108,113,0}},
/* 41908 */ {(13<<2)|2,{108,113,0}},
/* 41909 */ {(13<<2)|2,{108,113,0}},
/* 41910 */ {(13<<2)|2,{108,113,0}},
/* 41911 */ {(13<<2)|2,{108,113,0}},
/* 41912 */ {(13<<2)|2,{108,118,0}},
/* 41913 */ {(13<<2)|2,{108,118,0}},
/* 41914 */ {(13<<2)|2,{108,118,0}},
/* 41915 */ {(13<<2)|2,{108,118,0}},
/* 41916 */ {(13<<2)|2,{108,118,0}},
/* 41917 */ {(13<<2)|2,{108,118,0}},
/* 41918 */ {(13<<2)|2,{108,118,0}},
/* 41919 */ {(13<<2)|2,{108,118,0}},
/* 41920 */ {(13<<2)|2,{108,119,0}},
/* 41921 */ {(13<<2)|2,{108,119,0}},
/* 41922 */ {(13<<2)|2,{108,119,0}},
/* 41923 */ {(13<<2)|2,{108,119,0}},
/* 41924 */ {(13<<2)|2,{108,119,0}},
/* 41925 */ {(13<<2)|2,{108,119,0}},
/* 41926 */ {(13<<2)|2,{108,119,0}},
/* 41927 */ {(13<<2)|2,{108,119,0}},
/* 41928 */ {(13<<2)|2,{108,120,0}},
/* 41929 */ {(13<<2)|2,{108,120,0}},
/* 41930 */ {(13<<2)|2,{108,120,0}},
/* 41931 */ {(13<<2)|2,{108,120,0}},
/* 41932 */ {(13<<2)|2,{108,120,0}},
/* 41933 */ {(13<<2)|2,{108,120,0}},
/* 41934 */ {(13<<2)|2,{108,120,0}},
/* 41935 */ {(13<<2)|2,{108,120,0}},
/* 41936 */ {(13<<2)|2,{108,121,0}},
/* 41937 */ {(13<<2)|2,{108,121,0}},
/* 41938 */ {(13<<2)|2,{108,121,0}},
/* 41939 */ {(13<<2)|2,{108,121,0}},
/* 41940 */ {(13<<2)|2,{108,121,0}},
/* 41941 */ {(13<<2)|2,{108,121,0}},
/* 41942 */ {(13<<2)|2,{108,121,0}},
/* 41943 */ {(13<<2)|2,{108,121,0}},
/* 41944 */ {(13<<2)|2,{108,122,0}},
/* 41945 */ {(13<<2)|2,{108,122,0}},
/* 41946 */ {(13<<2)|2,{108,122,0}},
/* 41947 */ {(13<<2)|2,{108,122,0}},
/* 41948 */ {(13<<2)|2,{108,122,0}},
/* 41949 */ {(13<<2)|2,{108,122,0}},
/* 41950 */ {(13<<2)|2,{108,122,0}},
/* 41951 */ {(13<<2)|2,{108,122,0}},
/* 41952 */ {(14<<2)|2,{108,38,0}},
/* 41953 */ {(14<<2)|2,{108,38,0}},
/* 41954 */ {(14<<2)|2,{108,38,0}},
/* 41955 */ {(14<<2)|2,{108,38,0}},
/* 41956 */ {(14<<2)|2,{108,42,0}},
/* 41957 */ {(14<<2)|2,{108,42,0}},
/* 41958 */ {(14<<2)|2,{108,42,0}},
/* 41959 */ {(14<<2)|2,{108,42,0}},
/* 41960 */ {(14<<2)|2,{108,44,0}},
/* 41961 */ {(14<<2)|2,{108,44,0}},
/* 41962 */ {(14<<2)|2,{108,44,0}},
/* 41963 */ {(14<<2)|2,{108,44,0}},
/* 41964 */ {(14<<2)|2,{108,59,0}},
/* 41965 */ {(14<<2)|2,{108,59,0}},
/* 41966 */ {(14<<2)|2,{108,59,0}},
/* 41967 */ {(14<<2)|2,{108,59,0}},
/* 41968 */ {(14<<2)|2,{108,88,0}},
/* 41969 */ {(14<<2)|2,{108,88,0}},
/* 41970 */ {(14<<2)|2,{108,88,0}},
/* 41971 */ {(14<<2)|2,{108,88,0}},
/* 41972 */ {(14<<2)|2,{108,90,0}},
/* 41973 */ {(14<<2)|2,{108,90,0}},
/* 41974 */ {(14<<2)|2,{108,90,0}},
/* 41975 */ {(14<<2)|2,{108,90,0}},
/* 41976 */ {(16<<2)|2,{108,33,0}},
/* 41977 */ {(16<<2)|2,{108,34,0}},
/* 41978 */ {(16<<2)|2,{108,40,0}},
/* 41979 */ {(16<<2)|2,{108,41,0}},
/* 41980 */ {(16<<2)|2,{108,63,0}},
/* 41981 */ {(6<<2)|1,{108,0,0}},
/* 41982 */ {(6<<2)|1,{108,0,0}},
/* 41983 */ {(6<<2)|1,{108,0,0}},
/* 41984 */ {(16<<2)|3,{109,48,48}},
/* 41985 */ {(16<<2)|3,{109,48,49}},
/* 41986 */ {(16<<2)|3,{109,48,50}},
/* 41987 */ {(16<<2)|3,{109,48,97}},
/* 41988 */ {(16<<2)|3,{109,48,99}},
/* 41989 */ {(16<<2)|3,{109,48,101}},
/* 41990 */ {(16<<2)|3,{109,48,105}},
/* 41991 */ {(16<<2)|3,{109,48,111}},
/* 41992 */ {(16<<2)|3,{109,48,115}},
/* 41993 */ {(16<<2)|3,{109,48,116}},
/* 41994 */ {(11<<2)|2,{109,48,0}},
/* 41995 */ {(11<<2)|2,{109,48,0}},
/* 41996 */ {(11<<2)|2,{109,48,0}},
/* 41997 */ {(11<<2)|2,{109,48,0}},
/* 41998 */ {(11<<2)|2,{109,48,0}},
/* 41999 */ {(11<<2)|2,{109,48,0}},
/* 42000 */ {(11<<2)|2,{109,48,0}},
/* 42001 */ {(11<<2)|2,{109,48,0}},
/* 42002 */ {(11<<2)|2,{109,48,0}},
/* 42003 */ {(11<<2)|2,{109,48,0}},
/* 42004 */ {(11<<2)|2,{109,48,0}},
/* 42005 */ {(11<<2)|2,{109,48,0}},
/* 42006 */ {(11<<2)|2,{109,48,0}},
/* 42007 */ {(11<<2)|2,{109,48,0}},
/* 42008 */ {(11<<2)|2,{109,48,0}},
/* 42009 */ {(11<<2)|2,{109,48,0}},
/* 42010 */ {(11<<2)|2,{109,48,0}},
/* 42011 */ {(11<<2)|2,{109,48,0}},
/* 42012 */ {(11<<2)|2,{109,48,0}},
/* 42013 */ {(11<<2)|2,{109,48,0}},
/* 42014 */ {(11<<2)|2,{109,48,0}},
/* 42015 */ {(11<<2)|2,{109,48,0}},
/* 42016 */ {(16<<2)|3,{109,49,48}},
/* 42017 */ {(16<<2)|3,{109,49,49}},
/* 42018 */ {(16<<2)|3,{109,49,50}},
/* 42019 */ {(16<<2)|3,{109,49,97}},
/* 42020 */ {(16<<2)|3,{109,49,99}},
/* 42021 */ {(16<<2)|3,{109,49,101}},
/* 42022 */ {(16<<2)|3,{109,49,105}},
/* 42023 */ {(16<<2)|3,{109,49,111}},
/* 42024 */ {(16<<2)|3,{109,49,115}},
/* 42025 */ {(16<<2)|3,{109,49,116}},
/* 42026 */ {(11<<2)|2,{109,49,0}},
/* 42027 */ {(11<<2)|2,{109,49,0}},
/* 42028 */ {(11<<2)|2,{109,49,0}},
/* 42029 */ {(11<<2)|2,{109,49,0}},
/* 42030 */ {(11<<2)|2,{109,49,0}},
/* 42031 */ {(11<<2)|2,{109,49,0}},
/* 42032 */ {(11<<2)|2,{109,49,0}},
/* 42033 */ {(11<<2)|2,{109,49,0}},
/* 42034 */ {(11<<2)|2,{109,49,0}},
/* 42035 */ {(11<<2)|2,{109,49,0}},
/* 42036 */ {(11<<2)|2,{109,49,0}},
/* 42037 */ {(11<<2)|2,{109,49,0}},
/* 42038 */ {(11<<2)|2,{109,49,0}},
/* 42039 */ {(11<<2)|2,{109,49,0}},
/* 42040 */ {(11<<2)|2,{109,49,0}},
/* 42041 */ {(11<<2)|2,{109,49,0}},
/* 42042 */ {(11<<2)|2,{109,49,0}},
/* 42043 */ {(11<<2)|2,{109,49,0}},
/* 42044 */ {(11<<2)|2,{109,49,0}},
/* 42045 */ {(11<<2)|2,{109,49,0}},
/* 42046 */ {(11<<2)|2,{109,49,0}},
/* 42047 */ {(11<<2)|2,{109,49,0}},
/* 42048 */ {(16<<2)|3,{109,50,48}},
/* 42049 */ {(16<<2)|3,{109,50,49}},
/* 42050 */ {(16<<2)|3,{109,50,50}},
/* 42051 */ {(16<<2)|3,{109,50,97}},
/* 42052 */ {(16<<2)|3,{109,50,99}},
/* 42053 */ {(16<<2)|3,{109,50,101}},
/* 42054 */ {(16<<2)|3,{109,50,105}},
/* 42055 */ {(16<<2)|3,{109,50,111}},
/* 42056 */ {(16<<2)|3,{109,50,115}},
/* 42057 */ {(16<<2)|3,{109,50,116}},
/* 42058 */ {(11<<2)|2,{109,50,0}},
/* 42059 */ {(11<<2)|2,{109,50,0}},
/* 42060 */ {(11<<2)|2,{109,50,0}},
/* 42061 */ {(11<<2)|2,{109,50,0}},
/* 42062 */ {(11<<2)|2,{109,50,0}},
/* 42063 */ {(11<<2)|2,{109,50,0}},
/* 42064 */ {(11<<2)|2,{109,50,0}},
/* 42065 */ {(11<<2)|2,{109,50,0}},
/* 42066 */ {(11<<2)|2,{109,50,0}},
/* 42067 */ {(11<<2)|2,{109,50,0}},
/* 42068 */ {(11<<2)|2,{109,50,0}},
/* 42069 */ {(11<<2)|2,{109,50,0}},
/* 42070 */ {(11<<2)|2,{109,50,0}},
/* 42071 */ {(11<<2)|2,{109,50,0}},
/* 42072 */ {(11<<2)|2,{109,50,0}},
/* 42073 */ {(11<<2)|2,{109,50,0}},
/* 42074 */ {(11<<2)|2,{109,50,0}},
/* 42075 */ {(11<<2)|2,{109,50,0}},
/* 42076 */ {(11<<2)|2,{109,50,0}},
/* 42077 */ {(11<<2)|2,{109,50,0}},
/* 42078 */ {(11<<2)|2,{109,50,0}},
/* 42079 */ {(11<<2)|2,{109,50,0}},
/* 42080 */ {(16<<2)|3,{109,97,48}},
/* 42081 */ {(16<<2)|3,{109,97,49}},
/* 42082 */ {(16<<2)|3,{109,97,50}},
/* 42083 */ {(16<<2)|3,{109,97,97}},
/* 42084 */ {(16<<2)|3,{109,97,99}},
/* 42085 */ {(16<<2)|3,{109,97,101}},
/* 42086 */ {(16<<2)|3,{109,97,105}},
/* 42087 */ {(16<<2)|3,{109,97,111}},
/* 42088 */ {(16<<2)|3,{109,97,115}},
/* 42089 */ {(16<<2)|3,{109,97,116}},
/* 42090 */ {(11<<2)|2,{109,97,0}},
/* 42091 */ {(11<<2)|2,{109,97,0}},
/* 42092 */ {(11<<2)|2,{109,97,0}},
/* 42093 */ {(11<<2)|2,{109,97,0}},
/* 42094 */ {(11<<2)|2,{109,97,0}},
/* 42095 */ {(11<<2)|2,{109,97,0}},
/* 42096 */ {(11<<2)|2,{109,97,0}},
/* 42097 */ {(11<<2)|2,{109,97,0}},
/* 42098 */ {(11<<2)|2,{109,97,0}},
/* 42099 */ {(11<<2)|2,{109,97,0}},
/* 42100 */ {(11<<2)|2,{109,97,0}},
/* 42101 */ {(11<<2)|2,{109,97,0}},
/* 42102 */ {(11<<2)|2,{109,97,0}},
/* 42103 */ {(11<<2)|2,{109,97,0}},
/* 42104 */ {(11<<2)|2,{109,97,0}},
/* 42105 */ {(11<<2)|2,{109,97,0}},
/* 42106 */ {(11<<2)|2,{109,97,0}},
/* 42107 */ {(11<<2)|2,{109,97,0}},
/* 42108 */ {(11<<2)|2,{109,97,0}},
/* 42109 */ {(11<<2)|2,{109,97,0}},
/* 42110 */ {(11<<2)|2,{109,97,0}},
/* 42111 */ {(11<<2)|2,{109,97,0}},
/* 42112 */ {(16<<2)|3,{109,99,48}},
/* 42113 */ {(16<<2)|3,{109,99,49}},
/* 42114 */ {(16<<2)|3,{109,99,50}},
/* 42115 */ {(16<<2)|3,{109,99,97}},
/* 42116 */ {(16<<2)|3,{109,99,99}},
/* 42117 */ {(16<<2)|3,{109,99,101}},
/* 42118 */ {(16<<2)|3,{109,99,105}},
/* 42119 */ {(16<<2)|3,{109,99,111}},
/* 42120 */ {(16<<2)|3,{109,99,115}},
/* 42121 */ {(16<<2)|3,{109,99,116}},
/* 42122 */ {(11<<2)|2,{109,99,0}},
/* 42123 */ {(11<<2)|2,{109,99,0}},
/* 42124 */ {(11<<2)|2,{109,99,0}},
/* 42125 */ {(11<<2)|2,{109,99,0}},
/* 42126 */ {(11<<2)|2,{109,99,0}},
/* 42127 */ {(11<<2)|2,{109,99,0}},
/* 42128 */ {(11<<2)|2,{109,99,0}},
/* 42129 */ {(11<<2)|2,{109,99,0}},
/* 42130 */ {(11<<2)|2,{109,99,0}},
/* 42131 */ {(11<<2)|2,{109,99,0}},
/* 42132 */ {(11<<2)|2,{109,99,0}},
/* 42133 */ {(11<<2)|2,{109,99,0}},
/* 42134 */ {(11<<2)|2,{109,99,0}},
/* 42135 */ {(11<<2)|2,{109,99,0}},
/* 42136 */ {(11<<2)|2,{109,99,0}},
/* 42137 */ {(11<<2)|2,{109,99,0}},
/* 42138 */ {(11<<2)|2,{109,99,0}},
/* 42139 */ {(11<<2)|2,{109,99,0}},
/* 42140 */ {(11<<2)|2,{109,99,0}},
/* 42141 */ {(11<<2)|2,{109,99,0}},
/* 42142 */ {(11<<2)|2,{109,99,0}},
/* 42143 */ {(11<<2)|2,{109,99,0}},
/* 42144 */ {(16<<2)|3,{109,101,48}},
/* 42145 */ {(16<<2)|3,{109,101,49}},
/* 42146 */ {(16<<2)|3,{109,101,50}},
/* 42147 */ {(16<<2)|3,{109,101,97}},
/* 42148 */ {(16<<2)|3,{109,101,99}},
/* 42149 */ {(16<<2)|3,{109,101,101}},
/* 42150 */ {(16<<2)|3,{109,101,105}},
/* 42151 */ {(16<<2)|3,{109,101,111}},
/* 42152 */ {(16<<2)|3,{109,101,115}},
/* 42153 */ {(16<<2)|3,{109,101,116}},
/* 42154 */ {(11<<2)|2,{109,101,0}},
/* 42155 */ {(11<<2)|2,{109,101,0}},
/* 42156 */ {(11<<2)|2,{109,101,0}},
/* 42157 */ {(11<<2)|2,{109,101,0}},
/* 42158 */ {(11<<2)|2,{109,101,0}},
/* 42159 */ {(11<<2)|2,{109,101,0}},
/* 42160 */ {(11<<2)|2,{109,101,0}},
/* 42161 */ {(11<<2)|2,{109,101,0}},
/* 42162 */ {(11<<2)|2,{109,101,0}},
/* 42163 */ {(11<<2)|2,{109,101,0}},
/* 42164 */ {(11<<2)|2,{109,101,0}},
/* 42165 */ {(11<<2)|2,{109,101,0}},
/* 42166 */ {(11<<2)|2,{109,101,0}},
/* 42167 */ {(11<<2)|2,{109,101,0}},
/* 42168 */ {(11<<2)|2,{109,101,0}},
/* 42169 */ {(11<<2)|2,{109,101,0}},
/* 42170 */ {(11<<2)|2,{109,101,0}},
/* 42171 */ {(11<<2)|2,{109,101,0}},
/* 42172 */ {(11<<2)|2,{109,101,0}},
/* 42173 */ {(11<<2)|2,{109,101,0}},
/* 42174 */ {(11<<2)|2,{109,101,0}},
/* 42175 */ {(11<<2)|2,{109,101,0}},
/* 42176 */ {(16<<2)|3,{109,105,48}},
/* 42177 */ {(16<<2)|3,{109,105,49}},
/* 42178 */ {(16<<2)|3,{109,105,50}},
/* 42179 */ {(16<<2)|3,{109,105,97}},
/* 42180 */ {(16<<2)|3,{109,105,99}},
/* 42181 */ {(16<<2)|3,{109,105,101}},
/* 42182 */ {(16<<2)|3,{109,105,105}},
/* 42183 */ {(16<<2)|3,{109,105,111}},
/* 42184 */ {(16<<2)|3,{109,105,115}},
/* 42185 */ {(16<<2)|3,{109,105,116}},
/* 42186 */ {(11<<2)|2,{109,105,0}},
/* 42187 */ {(11<<2)|2,{109,105,0}},
/* 42188 */ {(11<<2)|2,{109,105,0}},
/* 42189 */ {(11<<2)|2,{109,105,0}},
/* 42190 */ {(11<<2)|2,{109,105,0}},
/* 42191 */ {(11<<2)|2,{109,105,0}},
/* 42192 */ {(11<<2)|2,{109,105,0}},
/* 42193 */ {(11<<2)|2,{109,105,0}},
/* 42194 */ {(11<<2)|2,{109,105,0}},
/* 42195 */ {(11<<2)|2,{109,105,0}},
/* 42196 */ {(11<<2)|2,{109,105,0}},
/* 42197 */ {(11<<2)|2,{109,105,0}},
/* 42198 */ {(11<<2)|2,{109,105,0}},
/* 42199 */ {(11<<2)|2,{109,105,0}},
/* 42200 */ {(11<<2)|2,{109,105,0}},
/* 42201 */ {(11<<2)|2,{109,105,0}},
/* 42202 */ {(11<<2)|2,{109,105,0}},
/* 42203 */ {(11<<2)|2,{109,105,0}},
/* 42204 */ {(11<<2)|2,{109,105,0}},
/* 42205 */ {(11<<2)|2,{109,105,0}},
/* 42206 */ {(11<<2)|2,{109,105,0}},
/* 42207 */ {(11<<2)|2,{109,105,0}},
/* 42208 */ {(16<<2)|3,{109,111,48}},
/* 42209 */ {(16<<2)|3,{109,111,49}},
/* 42210 */ {(16<<2)|3,{109,111,50}},
/* 42211 */ {(16<<2)|3,{109,111,97}},
/* 42212 */ {(16<<2)|3,{109,111,99}},
/* 42213 */ {(16<<2)|3,{109,111,101}},
/* 42214 */ {(16<<2)|3,{109,111,105}},
/* 42215 */ {(16<<2)|3,{109,111,111}},
/* 42216 */ {(16<<2)|3,{109,111,115}},
/* 42217 */ {(16<<2)|3,{109,111,116}},
/* 42218 */ {(11<<2)|2,{109,111,0}},
/* 42219 */ {(11<<2)|2,{109,111,0}},
/* 42220 */ {(11<<2)|2,{109,111,0}},
/* 42221 */ {(11<<2)|2,{109,111,0}},
/* 42222 */ {(11<<2)|2,{109,111,0}},
/* 42223 */ {(11<<2)|2,{109,111,0}},
/* 42224 */ {(11<<2)|2,{109,111,0}},
/* 42225 */ {(11<<2)|2,{109,111,0}},
/* 42226 */ {(11<<2)|2,{109,111,0}},
/* 42227 */ {(11<<2)|2,{109,111,0}},
/* 42228 */ {(11<<2)|2,{109,111,0}},
/* 42229 */ {(11<<2)|2,{109,111,0}},
/* 42230 */ {(11<<2)|2,{109,111,0}},
/* 42231 */ {(11<<2)|2,{109,111,0}},
/* 42232 */ {(11<<2)|2,{109,111,0}},
/* 42233 */ {(11<<2)|2,{109,111,0}},
/* 42234 */ {(11<<2)|2,{109,111,0}},
/* 42235 */ {(11<<2)|2,{109,111,0}},
/* 42236 */ {(11<<2)|2,{109,111,0}},
/* 42237 */ {(11<<2)|2,{109,111,0}},
/* 42238 */ {(11<<2)|2,{109,111,0}},
/* 42239 */ {(11<<2)|2,{109,111,0}},
/* 42240 */ {(16<<2)|3,{109,115,48}},
/* 42241 */ {(16<<2)|3,{109,115,49}},
/* 42242 */ {(16<<2)|3,{109,115,50}},
/* 42243 */ {(16<<2)|3,{109,115,97}},
/* 42244 */ {(16<<2)|3,{109,115,99}},
/* 42245 */ {(16<<2)|3,{109,115,101}},
/* 42246 */ {(16<<2)|3,{109,115,105}},
/* 42247 */ {(16<<2)|3,{109,115,111}},
/* 42248 */ {(16<<2)|3,{109,115,115}},
/* 42249 */ {(16<<2)|3,{109,115,116}},
/* 42250 */ {(11<<2)|2,{109,115,0}},
/* 42251 */ {(11<<2)|2,{109,115,0}},
/* 42252 */ {(11<<2)|2,{109,115,0}},
/* 42253 */ {(11<<2)|2,{109,115,0}},
/* 42254 */ {(11<<2)|2,{109,115,0}},
/* 42255 */ {(11<<2)|2,{109,115,0}},
/* 42256 */ {(11<<2)|2,{109,115,0}},
/* 42257 */ {(11<<2)|2,{109,115,0}},
/* 42258 */ {(11<<2)|2,{109,115,0}},
/* 42259 */ {(11<<2)|2,{109,115,0}},
/* 42260 */ {(11<<2)|2,{109,115,0}},
/* 42261 */ {(11<<2)|2,{109,115,0}},
/* 42262 */ {(11<<2)|2,{109,115,0}},
/* 42263 */ {(11<<2)|2,{109,115,0}},
/* 42264 */ {(11<<2)|2,{109,115,0}},
/* 42265 */ {(11<<2)|2,{109,115,0}},
/* 42266 */ {(11<<2)|2,{109,115,0}},
/* 42267 */ {(11<<2)|2,{109,115,0}},
/* 42268 */ {(11<<2)|2,{109,115,0}},
/* 42269 */ {(11<<2)|2,{109,115,0}},
/* 42270 */ {(11<<2)|2,{109,115,0}},
/* 42271 */ {(11<<2)|2,{109,115,0}},
/* 42272 */ {(16<<2)|3,{109,116,48}},
/* 42273 */ {(16<<2)|3,{109,116,49}},
/* 42274 */ {(16<<2)|3,{109,116,50}},
/* 42275 */ {(16<<2)|3,{109,116,97}},
/* 42276 */ {(16<<2)|3,{109,116,99}},
/* 42277 */ {(16<<2)|3,{109,116,101}},
/* 42278 */ {(16<<2)|3,{109,116,105}},
/* 42279 */ {(16<<2)|3,{109,116,111}},
/* 42280 */ {(16<<2)|3,{109,116,115}},
/* 42281 */ {(16<<2)|3,{109,116,116}},
/* 42282 */ {(11<<2)|2,{109,116,0}},
/* 42283 */ {(11<<2)|2,{109,116,0}},
/* 42284 */ {(11<<2)|2,{109,116,0}},
/* 42285 */ {(11<<2)|2,{109,116,0}},
/* 42286 */ {(11<<2)|2,{109,116,0}},
/* 42287 */ {(11<<2)|2,{109,116,0}},
/* 42288 */ {(11<<2)|2,{109,116,0}},
/* 42289 */ {(11<<2)|2,{109,116,0}},
/* 42290 */ {(11<<2)|2,{109,116,0}},
/* 42291 */ {(11<<2)|2,{109,116,0}},
/* 42292 */ {(11<<2)|2,{109,116,0}},
/* 42293 */ {(11<<2)|2,{109,116,0}},
/* 42294 */ {(11<<2)|2,{109,116,0}},
/* 42295 */ {(11<<2)|2,{109,116,0}},
/* 42296 */ {(11<<2)|2,{109,116,0}},
/* 42297 */ {(11<<2)|2,{109,116,0}},
/* 42298 */ {(11<<2)|2,{109,116,0}},
/* 42299 */ {(11<<2)|2,{109,116,0}},
/* 42300 */ {(11<<2)|2,{109,116,0}},
/* 42301 */ {(11<<2)|2,{109,116,0}},
/* 42302 */ {(11<<2)|2,{109,116,0}},
/* 42303 */ {(11<<2)|2,{109,116,0}},
/* 42304 */ {(12<<2)|2,{109,32,0}},
/* 42305 */ {(12<<2)|2,{109,32,0}},
/* 42306 */ {(12<<2)|2,{109,32,0}},
/* 42307 */ {(12<<2)|2,{109,32,0}},
/* 42308 */ {(12<<2)|2,{109,32,0}},
/* 42309 */ {(12<<2)|2,{109,32,0}},
/* 42310 */ {(12<<2)|2,{109,32,0}},
/* 42311 */ {(12<<2)|2,{109,32,0}},
/* 42312 */ {(12<<2)|2,{109,32,0}},
/* 42313 */ {(12<<2)|2,{109,32,0}},
/* 42314 */ {(12<<2)|2,{109,32,0}},
/* 42315 */ {(12<<2)|2,{109,32,0}},
/* 42316 */ {(12<<2)|2,{109,32,0}},
/* 42317 */ {(12<<2)|2,{109,32,0}},
/* 42318 */ {(12<<2)|2,{109,32,0}},
/* 42319 */ {(12<<2)|2,{109,32,0}},
/* 42320 */ {(12<<2)|2,{109,37,0}},
/* 42321 */ {(12<<2)|2,{109,37,0}},
/* 42322 */ {(12<<2)|2,{109,37,0}},
/* 42323 */ {(12<<2)|2,{109,37,0}},
/* 42324 */ {(12<<2)|2,{109,37,0}},
/* 42325 */ {(12<<2)|2,{109,37,0}},
/* 42326 */ {(12<<2)|2,{109,37,0}},
/* 42327 */ {(12<<2)|2,{109,37,0}},
/* 42328 */ {(12<<2)|2,{109,37,0}},
/* 42329 */ {(12<<2)|2,{109,37,0}},
/* 42330 */ {(12<<2)|2,{109,37,0}},
/* 42331 */ {(12<<2)|2,{109,37,0}},
/* 42332 */ {(12<<2)|2,{109,37,0}},
/* 42333 */ {(12<<2)|2,{109,37,0}},
/* 42334 */ {(12<<2)|2,{109,37,0}},
/* 42335 */ {(12<<2)|2,{109,37,0}},
/* 42336 */ {(12<<2)|2,{109,45,0}},
/* 42337 */ {(12<<2)|2,{109,45,0}},
/* 42338 */ {(12<<2)|2,{109,45,0}},
/* 42339 */ {(12<<2)|2,{109,45,0}},
/* 42340 */ {(12<<2)|2,{109,45,0}},
/* 42341 */ {(12<<2)|2,{109,45,0}},
/* 42342 */ {(12<<2)|2,{109,45,0}},
/* 42343 */ {(12<<2)|2,{109,45,0}},
/* 42344 */ {(12<<2)|2,{109,45,0}},
/* 42345 */ {(12<<2)|2,{109,45,0}},
/* 42346 */ {(12<<2)|2,{109,45,0}},
/* 42347 */ {(12<<2)|2,{109,45,0}},
/* 42348 */ {(12<<2)|2,{109,45,0}},
/* 42349 */ {(12<<2)|2,{109,45,0}},
/* 42350 */ {(12<<2)|2,{109,45,0}},
/* 42351 */ {(12<<2)|2,{109,45,0}},
/* 42352 */ {(12<<2)|2,{109,46,0}},
/* 42353 */ {(12<<2)|2,{109,46,0}},
/* 42354 */ {(12<<2)|2,{109,46,0}},
/* 42355 */ {(12<<2)|2,{109,46,0}},
/* 42356 */ {(12<<2)|2,{109,46,0}},
/* 42357 */ {(12<<2)|2,{109,46,0}},
/* 42358 */ {(12<<2)|2,{109,46,0}},
/* 42359 */ {(12<<2)|2,{109,46,0}},
/* 42360 */ {(12<<2)|2,{109,46,0}},
/* 42361 */ {(12<<2)|2,{109,46,0}},
/* 42362 */ {(12<<2)|2,{109,46,0}},
/* 42363 */ {(12<<2)|2,{109,46,0}},
/* 42364 */ {(12<<2)|2,{109,46,0}},
/* 42365 */ {(12<<2)|2,{109,46,0}},
/* 42366 */ {(12<<2)|2,{109,46,0}},
/* 42367 */ {(12<<2)|2,{109,46,0}},
/* 42368 */ {(12<<2)|2,{109,47,0}},
/* 42369 */ {(12<<2)|2,{109,47,0}},
/* 42370 */ {(12<<2)|2,{109,47,0}},
/* 42371 */ {(12<<2)|2,{109,47,0}},
/* 42372 */ {(12<<2)|2,{109,47,0}},
/* 42373 */ {(12<<2)|2,{109,47,0}},
/* 42374 */ {(12<<2)|2,{109,47,0}},
/* 42375 */ {(12<<2)|2,{109,47,0}},
/* 42376 */ {(12<<2)|2,{109,47,0}},
/* 42377 */ {(12<<2)|2,{109,47,0}},
/* 42378 */ {(12<<2)|2,{109,47,0}},
/* 42379 */ {(12<<2)|2,{109,47,0}},
/* 42380 */ {(12<<2)|2,{109,47,0}},
/* 42381 */ {(12<<2)|2,{109,47,0}},
/* 42382 */ {(12<<2)|2,{109,47,0}},
/* 42383 */ {(12<<2)|2,{109,47,0}},
/* 42384 */ {(12<<2)|2,{109,51,0}},
/* 42385 */ {(12<<2)|2,{109,51,0}},
/* 42386 */ {(12<<2)|2,{109,51,0}},
/* 42387 */ {(12<<2)|2,{109,51,0}},
/* 42388 */ {(12<<2)|2,{109,51,0}},
/* 42389 */ {(12<<2)|2,{109,51,0}},
/* 42390 */ {(12<<2)|2,{109,51,0}},
/* 42391 */ {(12<<2)|2,{109,51,0}},
/* 42392 */ {(12<<2)|2,{109,51,0}},
/* 42393 */ {(12<<2)|2,{109,51,0}},
/* 42394 */ {(12<<2)|2,{109,51,0}},
/* 42395 */ {(12<<2)|2,{109,51,0}},
/* 42396 */ {(12<<2)|2,{109,51,0}},
/* 42397 */ {(12<<2)|2,{109,51,0}},
/* 42398 */ {(12<<2)|2,{109,51,0}},
/* 42399 */ {(12<<2)|2,{109,51,0}},
/* 42400 */ {(12<<2)|2,{109,52,0}},
/* 42401 */ {(12<<2)|2,{109,52,0}},
/* 42402 */ {(12<<2)|2,{109,52,0}},
/* 42403 */ {(12<<2)|2,{109,52,0}},
/* 42404 */ {(12<<2)|2,{109,52,0}},
/* 42405 */ {(12<<2)|2,{109,52,0}},
/* 42406 */ {(12<<2)|2,{109,52,0}},
/* 42407 */ {(12<<2)|2,{109,52,0}},
/* 42408 */ {(12<<2)|2,{109,52,0}},
/* 42409 */ {(12<<2)|2,{109,52,0}},
/* 42410 */ {(12<<2)|2,{109,52,0}},
/* 42411 */ {(12<<2)|2,{109,52,0}},
/* 42412 */ {(12<<2)|2,{109,52,0}},
/* 42413 */ {(12<<2)|2,{109,52,0}},
/* 42414 */ {(12<<2)|2,{109,52,0}},
/* 42415 */ {(12<<2)|2,{109,52,0}},
/* 42416 */ {(12<<2)|2,{109,53,0}},
/* 42417 */ {(12<<2)|2,{109,53,0}},
/* 42418 */ {(12<<2)|2,{109,53,0}},
/* 42419 */ {(12<<2)|2,{109,53,0}},
/* 42420 */ {(12<<2)|2,{109,53,0}},
/* 42421 */ {(12<<2)|2,{109,53,0}},
/* 42422 */ {(12<<2)|2,{109,53,0}},
/* 42423 */ {(12<<2)|2,{109,53,0}},
/* 42424 */ {(12<<2)|2,{109,53,0}},
/* 42425 */ {(12<<2)|2,{109,53,0}},
/* 42426 */ {(12<<2)|2,{109,53,0}},
/* 42427 */ {(12<<2)|2,{109,53,0}},
/* 42428 */ {(12<<2)|2,{109,53,0}},
/* 42429 */ {(12<<2)|2,{109,53,0}},
/* 42430 */ {(12<<2)|2,{109,53,0}},
/* 42431 */ {(12<<2)|2,{109,53,0}},
/* 42432 */ {(12<<2)|2,{109,54,0}},
/* 42433 */ {(12<<2)|2,{109,54,0}},
/* 42434 */ {(12<<2)|2,{109,54,0}},
/* 42435 */ {(12<<2)|2,{109,54,0}},
/* 42436 */ {(12<<2)|2,{109,54,0}},
/* 42437 */ {(12<<2)|2,{109,54,0}},
/* 42438 */ {(12<<2)|2,{109,54,0}},
/* 42439 */ {(12<<2)|2,{109,54,0}},
/* 42440 */ {(12<<2)|2,{109,54,0}},
/* 42441 */ {(12<<2)|2,{109,54,0}},
/* 42442 */ {(12<<2)|2,{109,54,0}},
/* 42443 */ {(12<<2)|2,{109,54,0}},
/* 42444 */ {(12<<2)|2,{109,54,0}},
/* 42445 */ {(12<<2)|2,{109,54,0}},
/* 42446 */ {(12<<2)|2,{109,54,0}},
/* 42447 */ {(12<<2)|2,{109,54,0}},
/* 42448 */ {(12<<2)|2,{109,55,0}},
/* 42449 */ {(12<<2)|2,{109,55,0}},
/* 42450 */ {(12<<2)|2,{109,55,0}},
/* 42451 */ {(12<<2)|2,{109,55,0}},
/* 42452 */ {(12<<2)|2,{109,55,0}},
/* 42453 */ {(12<<2)|2,{109,55,0}},
/* 42454 */ {(12<<2)|2,{109,55,0}},
/* 42455 */ {(12<<2)|2,{109,55,0}},
/* 42456 */ {(12<<2)|2,{109,55,0}},
/* 42457 */ {(12<<2)|2,{109,55,0}},
/* 42458 */ {(12<<2)|2,{109,55,0}},
/* 42459 */ {(12<<2)|2,{109,55,0}},
/* 42460 */ {(12<<2)|2,{109,55,0}},
/* 42461 */ {(12<<2)|2,{109,55,0}},
/* 42462 */ {(12<<2)|2,{109,55,0}},
/* 42463 */ {(12<<2)|2,{109,55,0}},
/* 42464 */ {(12<<2)|2,{109,56,0}},
/* 42465 */ {(12<<2)|2,{109,56,0}},
/* 42466 */ {(12<<2)|2,{109,56,0}},
/* 42467 */ {(12<<2)|2,{109,56,0}},
/* 42468 */ {(12<<2)|2,{109,56,0}},
/* 42469 */ {(12<<2)|2,{109,56,0}},
/* 42470 */ {(12<<2)|2,{109,56,0}},
/* 42471 */ {(12<<2)|2,{109,56,0}},
/* 42472 */ {(12<<2)|2,{109,56,0}},
/* 42473 */ {(12<<2)|2,{109,56,0}},
/* 42474 */ {(12<<2)|2,{109,56,0}},
/* 42475 */ {(12<<2)|2,{109,56,0}},
/* 42476 */ {(12<<2)|2,{109,56,0}},
/* 42477 */ {(12<<2)|2,{109,56,0}},
/* 42478 */ {(12<<2)|2,{109,56,0}},
/* 42479 */ {(12<<2)|2,{109,56,0}},
/* 42480 */ {(12<<2)|2,{109,57,0}},
/* 42481 */ {(12<<2)|2,{109,57,0}},
/* 42482 */ {(12<<2)|2,{109,57,0}},
/* 42483 */ {(12<<2)|2,{109,57,0}},
/* 42484 */ {(12<<2)|2,{109,57,0}},
/* 42485 */ {(12<<2)|2,{109,57,0}},
/* 42486 */ {(12<<2)|2,{109,57,0}},
/* 42487 */ {(12<<2)|2,{109,57,0}},
/* 42488 */ {(12<<2)|2,{109,57,0}},
/* 42489 */ {(12<<2)|2,{109,57,0}},
/* 42490 */ {(12<<2)|2,{109,57,0}},
/* 42491 */ {(12<<2)|2,{109,57,0}},
/* 42492 */ {(12<<2)|2,{109,57,0}},
/* 42493 */ {(12<<2)|2,{109,57,0}},
/* 42494 */ {(12<<2)|2,{109,57,0}},
/* 42495 */ {(12<<2)|2,{109,57,0}},
/* 42496 */ {(12<<2)|2,{109,61,0}},
/* 42497 */ {(12<<2)|2,{109,61,0}},
/* 42498 */ {(12<<2)|2,{109,61,0}},
/* 42499 */ {(12<<2)|2,{109,61,0}},
/* 42500 */ {(12<<2)|2,{109,61,0}},
/* 42501 */ {(12<<2)|2,{109,61,0}},
/* 42502 */ {(12<<2)|2,{109,61,0}},
/* 42503 */ {(12<<2)|2,{109,61,0}},
/* 42504 */ {(12<<2)|2,{109,61,0}},
/* 42505 */ {(12<<2)|2,{109,61,0}},
/* 42506 */ {(12<<2)|2,{109,61,0}},
/* 42507 */ {(12<<2)|2,{109,61,0}},
/* 42508 */ {(12<<2)|2,{109,61,0}},
/* 42509 */ {(12<<2)|2,{109,61,0}},
/* 42510 */ {(12<<2)|2,{109,61,0}},
/* 42511 */ {(12<<2)|2,{109,61,0}},
/* 42512 */ {(12<<2)|2,{109,65,0}},
/* 42513 */ {(12<<2)|2,{109,65,0}},
/* 42514 */ {(12<<2)|2,{109,65,0}},
/* 42515 */ {(12<<2)|2,{109,65,0}},
/* 42516 */ {(12<<2)|2,{109,65,0}},
/* 42517 */ {(12<<2)|2,{109,65,0}},
/* 42518 */ {(12<<2)|2,{109,65,0}},
/* 42519 */ {(12<<2)|2,{109,65,0}},
/* 42520 */ {(12<<2)|2,{109,65,0}},
/* 42521 */ {(12<<2)|2,{109,65,0}},
/* 42522 */ {(12<<2)|2,{109,65,0}},
/* 42523 */ {(12<<2)|2,{109,65,0}},
/* 42524 */ {(12<<2)|2,{109,65,0}},
/* 42525 */ {(12<<2)|2,{109,65,0}},
/* 42526 */ {(12<<2)|2,{109,65,0}},
/* 42527 */ {(12<<2)|2,{109,65,0}},
/* 42528 */ {(12<<2)|2,{109,95,0}},
/* 42529 */ {(12<<2)|2,{109,95,0}},
/* 42530 */ {(12<<2)|2,{109,95,0}},
/* 42531 */ {(12<<2)|2,{109,95,0}},
/* 42532 */ {(12<<2)|2,{109,95,0}},
/* 42533 */ {(12<<2)|2,{109,95,0}},
/* 42534 */ {(12<<2)|2,{109,95,0}},
/* 42535 */ {(12<<2)|2,{109,95,0}},
/* 42536 */ {(12<<2)|2,{109,95,0}},
/* 42537 */ {(12<<2)|2,{109,95,0}},
/* 42538 */ {(12<<2)|2,{109,95,0}},
/* 42539 */ {(12<<2)|2,{109,95,0}},
/* 42540 */ {(12<<2)|2,{109,95,0}},
/* 42541 */ {(12<<2)|2,{109,95,0}},
/* 42542 */ {(12<<2)|2,{109,95,0}},
/* 42543 */ {(12<<2)|2,{109,95,0}},
/* 42544 */ {(12<<2)|2,{109,98,0}},
/* 42545 */ {(12<<2)|2,{109,98,0}},
/* 42546 */ {(12<<2)|2,{109,98,0}},
/* 42547 */ {(12<<2)|2,{109,98,0}},
/* 42548 */ {(12<<2)|2,{109,98,0}},
/* 42549 */ {(12<<2)|2,{109,98,0}},
/* 42550 */ {(12<<2)|2,{109,98,0}},
/* 42551 */ {(12<<2)|2,{109,98,0}},
/* 42552 */ {(12<<2)|2,{109,98,0}},
/* 42553 */ {(12<<2)|2,{109,98,0}},
/* 42554 */ {(12<<2)|2,{109,98,0}},
/* 42555 */ {(12<<2)|2,{109,98,0}},
/* 42556 */ {(12<<2)|2,{109,98,0}},
/* 42557 */ {(12<<2)|2,{109,98,0}},
/* 42558 */ {(12<<2)|2,{109,98,0}},
/* 42559 */ {(12<<2)|2,{109,98,0}},
/* 42560 */ {(12<<2)|2,{109,100,0}},
/* 42561 */ {(12<<2)|2,{109,100,0}},
/* 42562 */ {(12<<2)|2,{109,100,0}},
/* 42563 */ {(12<<2)|2,{109,100,0}},
/* 42564 */ {(12<<2)|2,{109,100,0}},
/* 42565 */ {(12<<2)|2,{109,100,0}},
/* 42566 */ {(12<<2)|2,{109,100,0}},
/* 42567 */ {(12<<2)|2,{109,100,0}},
/* 42568 */ {(12<<2)|2,{109,100,0}},
/* 42569 */ {(12<<2)|2,{109,100,0}},
/* 42570 */ {(12<<2)|2,{109,100,0}},
/* 42571 */ {(12<<2)|2,{109,100,0}},
/* 42572 */ {(12<<2)|2,{109,100,0}},
/* 42573 */ {(12<<2)|2,{109,100,0}},
/* 42574 */ {(12<<2)|2,{109,100,0}},
/* 42575 */ {(12<<2)|2,{109,100,0}},
/* 42576 */ {(12<<2)|2,{109,102,0}},
/* 42577 */ {(12<<2)|2,{109,102,0}},
/* 42578 */ {(12<<2)|2,{109,102,0}},
/* 42579 */ {(12<<2)|2,{109,102,0}},
/* 42580 */ {(12<<2)|2,{109,102,0}},
/* 42581 */ {(12<<2)|2,{109,102,0}},
/* 42582 */ {(12<<2)|2,{109,102,0}},
/* 42583 */ {(12<<2)|2,{109,102,0}},
/* 42584 */ {(12<<2)|2,{109,102,0}},
/* 42585 */ {(12<<2)|2,{109,102,0}},
/* 42586 */ {(12<<2)|2,{109,102,0}},
/* 42587 */ {(12<<2)|2,{109,102,0}},
/* 42588 */ {(12<<2)|2,{109,102,0}},
/* 42589 */ {(12<<2)|2,{109,102,0}},
/* 42590 */ {(12<<2)|2,{109,102,0}},
/* 42591 */ {(12<<2)|2,{109,102,0}},
/* 42592 */ {(12<<2)|2,{109,103,0}},
/* 42593 */ {(12<<2)|2,{109,103,0}},
/* 42594 */ {(12<<2)|2,{109,103,0}},
/* 42595 */ {(12<<2)|2,{109,103,0}},
/* 42596 */ {(12<<2)|2,{109,103,0}},
/* 42597 */ {(12<<2)|2,{109,103,0}},
/* 42598 */ {(12<<2)|2,{109,103,0}},
/* 42599 */ {(12<<2)|2,{109,103,0}},
/* 42600 */ {(12<<2)|2,{109,103,0}},
/* 42601 */ {(12<<2)|2,{109,103,0}},
/* 42602 */ {(12<<2)|2,{109,103,0}},
/* 42603 */ {(12<<2)|2,{109,103,0}},
/* 42604 */ {(12<<2)|2,{109,103,0}},
/* 42605 */ {(12<<2)|2,{109,103,0}},
/* 42606 */ {(12<<2)|2,{109,103,0}},
/* 42607 */ {(12<<2)|2,{109,103,0}},
/* 42608 */ {(12<<2)|2,{109,104,0}},
/* 42609 */ {(12<<2)|2,{109,104,0}},
/* 42610 */ {(12<<2)|2,{109,104,0}},
/* 42611 */ {(12<<2)|2,{109,104,0}},
/* 42612 */ {(12<<2)|2,{109,104,0}},
/* 42613 */ {(12<<2)|2,{109,104,0}},
/* 42614 */ {(12<<2)|2,{109,104,0}},
/* 42615 */ {(12<<2)|2,{109,104,0}},
/* 42616 */ {(12<<2)|2,{109,104,0}},
/* 42617 */ {(12<<2)|2,{109,104,0}},
/* 42618 */ {(12<<2)|2,{109,104,0}},
/* 42619 */ {(12<<2)|2,{109,104,0}},
/* 42620 */ {(12<<2)|2,{109,104,0}},
/* 42621 */ {(12<<2)|2,{109,104,0}},
/* 42622 */ {(12<<2)|2,{109,104,0}},
/* 42623 */ {(12<<2)|2,{109,104,0}},
/* 42624 */ {(12<<2)|2,{109,108,0}},
/* 42625 */ {(12<<2)|2,{109,108,0}},
/* 42626 */ {(12<<2)|2,{109,108,0}},
/* 42627 */ {(12<<2)|2,{109,108,0}},
/* 42628 */ {(12<<2)|2,{109,108,0}},
/* 42629 */ {(12<<2)|2,{109,108,0}},
/* 42630 */ {(12<<2)|2,{109,108,0}},
/* 42631 */ {(12<<2)|2,{109,108,0}},
/* 42632 */ {(12<<2)|2,{109,108,0}},
/* 42633 */ {(12<<2)|2,{109,108,0}},
/* 42634 */ {(12<<2)|2,{109,108,0}},
/* 42635 */ {(12<<2)|2,{109,108,0}},
/* 42636 */ {(12<<2)|2,{109,108,0}},
/* 42637 */ {(12<<2)|2,{109,108,0}},
/* 42638 */ {(12<<2)|2,{109,108,0}},
/* 42639 */ {(12<<2)|2,{109,108,0}},
/* 42640 */ {(12<<2)|2,{109,109,0}},
/* 42641 */ {(12<<2)|2,{109,109,0}},
/* 42642 */ {(12<<2)|2,{109,109,0}},
/* 42643 */ {(12<<2)|2,{109,109,0}},
/* 42644 */ {(12<<2)|2,{109,109,0}},
/* 42645 */ {(12<<2)|2,{109,109,0}},
/* 42646 */ {(12<<2)|2,{109,109,0}},
/* 42647 */ {(12<<2)|2,{109,109,0}},
/* 42648 */ {(12<<2)|2,{109,109,0}},
/* 42649 */ {(12<<2)|2,{109,109,0}},
/* 42650 */ {(12<<2)|2,{109,109,0}},
/* 42651 */ {(12<<2)|2,{109,109,0}},
/* 42652 */ {(12<<2)|2,{109,109,0}},
/* 42653 */ {(12<<2)|2,{109,109,0}},
/* 42654 */ {(12<<2)|2,{109,109,0}},
/* 42655 */ {(12<<2)|2,{109,109,0}},
/* 42656 */ {(12<<2)|2,{109,110,0}},
/* 42657 */ {(12<<2)|2,{109,110,0}},
/* 42658 */ {(12<<2)|2,{109,110,0}},
/* 42659 */ {(12<<2)|2,{109,110,0}},
/* 42660 */ {(12<<2)|2,{109,110,0}},
/* 42661 */ {(12<<2)|2,{109,110,0}},
/* 42662 */ {(12<<2)|2,{109,110,0}},
/* 42663 */ {(12<<2)|2,{109,110,0}},
/* 42664 */ {(12<<2)|2,{109,110,0}},
/* 42665 */ {(12<<2)|2,{109,110,0}},
/* 42666 */ {(12<<2)|2,{109,110,0}},
/* 42667 */ {(12<<2)|2,{109,110,0}},
/* 42668 */ {(12<<2)|2,{109,110,0}},
/* 42669 */ {(12<<2)|2,{109,110,0}},
/* 42670 */ {(12<<2)|2,{109,110,0}},
/* 42671 */ {(12<<2)|2,{109,110,0}},
/* 42672 */ {(12<<2)|2,{109,112,0}},
/* 42673 */ {(12<<2)|2,{109,112,0}},
/* 42674 */ {(12<<2)|2,{109,112,0}},
/* 42675 */ {(12<<2)|2,{109,112,0}},
/* 42676 */ {(12<<2)|2,{109,112,0}},
/* 42677 */ {(12<<2)|2,{109,112,0}},
/* 42678 */ {(12<<2)|2,{109,112,0}},
/* 42679 */ {(12<<2)|2,{109,112,0}},
/* 42680 */ {(12<<2)|2,{109,112,0}},
/* 42681 */ {(12<<2)|2,{109,112,0}},
/* 42682 */ {(12<<2)|2,{109,112,0}},
/* 42683 */ {(12<<2)|2,{109,112,0}},
/* 42684 */ {(12<<2)|2,{109,112,0}},
/* 42685 */ {(12<<2)|2,{109,112,0}},
/* 42686 */ {(12<<2)|2,{109,112,0}},
/* 42687 */ {(12<<2)|2,{109,112,0}},
/* 42688 */ {(12<<2)|2,{109,114,0}},
/* 42689 */ {(12<<2)|2,{109,114,0}},
/* 42690 */ {(12<<2)|2,{109,114,0}},
/* 42691 */ {(12<<2)|2,{109,114,0}},
/* 42692 */ {(12<<2)|2,{109,114,0}},
/* 42693 */ {(12<<2)|2,{109,114,0}},
/* 42694 */ {(12<<2)|2,{109,114,0}},
/* 42695 */ {(12<<2)|2,{109,114,0}},
/* 42696 */ {(12<<2)|2,{109,114,0}},
/* 42697 */ {(12<<2)|2,{109,114,0}},
/* 42698 */ {(12<<2)|2,{109,114,0}},
/* 42699 */ {(12<<2)|2,{109,114,0}},
/* 42700 */ {(12<<2)|2,{109,114,0}},
/* 42701 */ {(12<<2)|2,{109,114,0}},
/* 42702 */ {(12<<2)|2,{109,114,0}},
/* 42703 */ {(12<<2)|2,{109,114,0}},
/* 42704 */ {(12<<2)|2,{109,117,0}},
/* 42705 */ {(12<<2)|2,{109,117,0}},
/* 42706 */ {(12<<2)|2,{109,117,0}},
/* 42707 */ {(12<<2)|2,{109,117,0}},
/* 42708 */ {(12<<2)|2,{109,117,0}},
/* 42709 */ {(12<<2)|2,{109,117,0}},
/* 42710 */ {(12<<2)|2,{109,117,0}},
/* 42711 */ {(12<<2)|2,{109,117,0}},
/* 42712 */ {(12<<2)|2,{109,117,0}},
/* 42713 */ {(12<<2)|2,{109,117,0}},
/* 42714 */ {(12<<2)|2,{109,117,0}},
/* 42715 */ {(12<<2)|2,{109,117,0}},
/* 42716 */ {(12<<2)|2,{109,117,0}},
/* 42717 */ {(12<<2)|2,{109,117,0}},
/* 42718 */ {(12<<2)|2,{109,117,0}},
/* 42719 */ {(12<<2)|2,{109,117,0}},
/* 42720 */ {(13<<2)|2,{109,58,0}},
/* 42721 */ {(13<<2)|2,{109,58,0}},
/* 42722 */ {(13<<2)|2,{109,58,0}},
/* 42723 */ {(13<<2)|2,{109,58,0}},
/* 42724 */ {(13<<2)|2,{109,58,0}},
/* 42725 */ {(13<<2)|2,{109,58,0}},
/* 42726 */ {(13<<2)|2,{109,58,0}},
/* 42727 */ {(13<<2)|2,{109,58,0}},
/* 42728 */ {(13<<2)|2,{109,66,0}},
/* 42729 */ {(13<<2)|2,{109,66,0}},
/* 42730 */ {(13<<2)|2,{109,66,0}},
/* 42731 */ {(13<<2)|2,{109,66,0}},
/* 42732 */ {(13<<2)|2,{109,66,0}},
/* 42733 */ {(13<<2)|2,{109,66,0}},
/* 42734 */ {(13<<2)|2,{109,66,0}},
/* 42735 */ {(13<<2)|2,{109,66,0}},
/* 42736 */ {(13<<2)|2,{109,67,0}},
/* 42737 */ {(13<<2)|2,{109,67,0}},
/* 42738 */ {(13<<2)|2,{109,67,0}},
/* 42739 */ {(13<<2)|2,{109,67,0}},
/* 42740 */ {(13<<2)|2,{109,67,0}},
/* 42741 */ {(13<<2)|2,{109,67,0}},
/* 42742 */ {(13<<2)|2,{109,67,0}},
/* 42743 */ {(13<<2)|2,{109,67,0}},
/* 42744 */ {(13<<2)|2,{109,68,0}},
/* 42745 */ {(13<<2)|2,{109,68,0}},
/* 42746 */ {(13<<2)|2,{109,68,0}},
/* 42747 */ {(13<<2)|2,{109,68,0}},
/* 42748 */ {(13<<2)|2,{109,68,0}},
/* 42749 */ {(13<<2)|2,{109,68,0}},
/* 42750 */ {(13<<2)|2,{109,68,0}},
/* 42751 */ {(13<<2)|2,{109,68,0}},
/* 42752 */ {(13<<2)|2,{109,69,0}},
/* 42753 */ {(13<<2)|2,{109,69,0}},
/* 42754 */ {(13<<2)|2,{109,69,0}},
/* 42755 */ {(13<<2)|2,{109,69,0}},
/* 42756 */ {(13<<2)|2,{109,69,0}},
/* 42757 */ {(13<<2)|2,{109,69,0}},
/* 42758 */ {(13<<2)|2,{109,69,0}},
/* 42759 */ {(13<<2)|2,{109,69,0}},
/* 42760 */ {(13<<2)|2,{109,70,0}},
/* 42761 */ {(13<<2)|2,{109,70,0}},
/* 42762 */ {(13<<2)|2,{109,70,0}},
/* 42763 */ {(13<<2)|2,{109,70,0}},
/* 42764 */ {(13<<2)|2,{109,70,0}},
/* 42765 */ {(13<<2)|2,{109,70,0}},
/* 42766 */ {(13<<2)|2,{109,70,0}},
/* 42767 */ {(13<<2)|2,{109,70,0}},
/* 42768 */ {(13<<2)|2,{109,71,0}},
/* 42769 */ {(13<<2)|2,{109,71,0}},
/* 42770 */ {(13<<2)|2,{109,71,0}},
/* 42771 */ {(13<<2)|2,{109,71,0}},
/* 42772 */ {(13<<2)|2,{109,71,0}},
/* 42773 */ {(13<<2)|2,{109,71,0}},
/* 42774 */ {(13<<2)|2,{109,71,0}},
/* 42775 */ {(13<<2)|2,{109,71,0}},
/* 42776 */ {(13<<2)|2,{109,72,0}},
/* 42777 */ {(13<<2)|2,{109,72,0}},
/* 42778 */ {(13<<2)|2,{109,72,0}},
/* 42779 */ {(13<<2)|2,{109,72,0}},
/* 42780 */ {(13<<2)|2,{109,72,0}},
/* 42781 */ {(13<<2)|2,{109,72,0}},
/* 42782 */ {(13<<2)|2,{109,72,0}},
/* 42783 */ {(13<<2)|2,{109,72,0}},
/* 42784 */ {(13<<2)|2,{109,73,0}},
/* 42785 */ {(13<<2)|2,{109,73,0}},
/* 42786 */ {(13<<2)|2,{109,73,0}},
/* 42787 */ {(13<<2)|2,{109,73,0}},
/* 42788 */ {(13<<2)|2,{109,73,0}},
/* 42789 */ {(13<<2)|2,{109,73,0}},
/* 42790 */ {(13<<2)|2,{109,73,0}},
/* 42791 */ {(13<<2)|2,{109,73,0}},
/* 42792 */ {(13<<2)|2,{109,74,0}},
/* 42793 */ {(13<<2)|2,{109,74,0}},
/* 42794 */ {(13<<2)|2,{109,74,0}},
/* 42795 */ {(13<<2)|2,{109,74,0}},
/* 42796 */ {(13<<2)|2,{109,74,0}},
/* 42797 */ {(13<<2)|2,{109,74,0}},
/* 42798 */ {(13<<2)|2,{109,74,0}},
/* 42799 */ {(13<<2)|2,{109,74,0}},
/* 42800 */ {(13<<2)|2,{109,75,0}},
/* 42801 */ {(13<<2)|2,{109,75,0}},
/* 42802 */ {(13<<2)|2,{109,75,0}},
/* 42803 */ {(13<<2)|2,{109,75,0}},
/* 42804 */ {(13<<2)|2,{109,75,0}},
/* 42805 */ {(13<<2)|2,{109,75,0}},
/* 42806 */ {(13<<2)|2,{109,75,0}},
/* 42807 */ {(13<<2)|2,{109,75,0}},
/* 42808 */ {(13<<2)|2,{109,76,0}},
/* 42809 */ {(13<<2)|2,{109,76,0}},
/* 42810 */ {(13<<2)|2,{109,76,0}},
/* 42811 */ {(13<<2)|2,{109,76,0}},
/* 42812 */ {(13<<2)|2,{109,76,0}},
/* 42813 */ {(13<<2)|2,{109,76,0}},
/* 42814 */ {(13<<2)|2,{109,76,0}},
/* 42815 */ {(13<<2)|2,{109,76,0}},
/* 42816 */ {(13<<2)|2,{109,77,0}},
/* 42817 */ {(13<<2)|2,{109,77,0}},
/* 42818 */ {(13<<2)|2,{109,77,0}},
/* 42819 */ {(13<<2)|2,{109,77,0}},
/* 42820 */ {(13<<2)|2,{109,77,0}},
/* 42821 */ {(13<<2)|2,{109,77,0}},
/* 42822 */ {(13<<2)|2,{109,77,0}},
/* 42823 */ {(13<<2)|2,{109,77,0}},
/* 42824 */ {(13<<2)|2,{109,78,0}},
/* 42825 */ {(13<<2)|2,{109,78,0}},
/* 42826 */ {(13<<2)|2,{109,78,0}},
/* 42827 */ {(13<<2)|2,{109,78,0}},
/* 42828 */ {(13<<2)|2,{109,78,0}},
/* 42829 */ {(13<<2)|2,{109,78,0}},
/* 42830 */ {(13<<2)|2,{109,78,0}},
/* 42831 */ {(13<<2)|2,{109,78,0}},
/* 42832 */ {(13<<2)|2,{109,79,0}},
/* 42833 */ {(13<<2)|2,{109,79,0}},
/* 42834 */ {(13<<2)|2,{109,79,0}},
/* 42835 */ {(13<<2)|2,{109,79,0}},
/* 42836 */ {(13<<2)|2,{109,79,0}},
/* 42837 */ {(13<<2)|2,{109,79,0}},
/* 42838 */ {(13<<2)|2,{109,79,0}},
/* 42839 */ {(13<<2)|2,{109,79,0}},
/* 42840 */ {(13<<2)|2,{109,80,0}},
/* 42841 */ {(13<<2)|2,{109,80,0}},
/* 42842 */ {(13<<2)|2,{109,80,0}},
/* 42843 */ {(13<<2)|2,{109,80,0}},
/* 42844 */ {(13<<2)|2,{109,80,0}},
/* 42845 */ {(13<<2)|2,{109,80,0}},
/* 42846 */ {(13<<2)|2,{109,80,0}},
/* 42847 */ {(13<<2)|2,{109,80,0}},
/* 42848 */ {(13<<2)|2,{109,81,0}},
/* 42849 */ {(13<<2)|2,{109,81,0}},
/* 42850 */ {(13<<2)|2,{109,81,0}},
/* 42851 */ {(13<<2)|2,{109,81,0}},
/* 42852 */ {(13<<2)|2,{109,81,0}},
/* 42853 */ {(13<<2)|2,{109,81,0}},
/* 42854 */ {(13<<2)|2,{109,81,0}},
/* 42855 */ {(13<<2)|2,{109,81,0}},
/* 42856 */ {(13<<2)|2,{109,82,0}},
/* 42857 */ {(13<<2)|2,{109,82,0}},
/* 42858 */ {(13<<2)|2,{109,82,0}},
/* 42859 */ {(13<<2)|2,{109,82,0}},
/* 42860 */ {(13<<2)|2,{109,82,0}},
/* 42861 */ {(13<<2)|2,{109,82,0}},
/* 42862 */ {(13<<2)|2,{109,82,0}},
/* 42863 */ {(13<<2)|2,{109,82,0}},
/* 42864 */ {(13<<2)|2,{109,83,0}},
/* 42865 */ {(13<<2)|2,{109,83,0}},
/* 42866 */ {(13<<2)|2,{109,83,0}},
/* 42867 */ {(13<<2)|2,{109,83,0}},
/* 42868 */ {(13<<2)|2,{109,83,0}},
/* 42869 */ {(13<<2)|2,{109,83,0}},
/* 42870 */ {(13<<2)|2,{109,83,0}},
/* 42871 */ {(13<<2)|2,{109,83,0}},
/* 42872 */ {(13<<2)|2,{109,84,0}},
/* 42873 */ {(13<<2)|2,{109,84,0}},
/* 42874 */ {(13<<2)|2,{109,84,0}},
/* 42875 */ {(13<<2)|2,{109,84,0}},
/* 42876 */ {(13<<2)|2,{109,84,0}},
/* 42877 */ {(13<<2)|2,{109,84,0}},
/* 42878 */ {(13<<2)|2,{109,84,0}},
/* 42879 */ {(13<<2)|2,{109,84,0}},
/* 42880 */ {(13<<2)|2,{109,85,0}},
/* 42881 */ {(13<<2)|2,{109,85,0}},
/* 42882 */ {(13<<2)|2,{109,85,0}},
/* 42883 */ {(13<<2)|2,{109,85,0}},
/* 42884 */ {(13<<2)|2,{109,85,0}},
/* 42885 */ {(13<<2)|2,{109,85,0}},
/* 42886 */ {(13<<2)|2,{109,85,0}},
/* 42887 */ {(13<<2)|2,{109,85,0}},
/* 42888 */ {(13<<2)|2,{109,86,0}},
/* 42889 */ {(13<<2)|2,{109,86,0}},
/* 42890 */ {(13<<2)|2,{109,86,0}},
/* 42891 */ {(13<<2)|2,{109,86,0}},
/* 42892 */ {(13<<2)|2,{109,86,0}},
/* 42893 */ {(13<<2)|2,{109,86,0}},
/* 42894 */ {(13<<2)|2,{109,86,0}},
/* 42895 */ {(13<<2)|2,{109,86,0}},
/* 42896 */ {(13<<2)|2,{109,87,0}},
/* 42897 */ {(13<<2)|2,{109,87,0}},
/* 42898 */ {(13<<2)|2,{109,87,0}},
/* 42899 */ {(13<<2)|2,{109,87,0}},
/* 42900 */ {(13<<2)|2,{109,87,0}},
/* 42901 */ {(13<<2)|2,{109,87,0}},
/* 42902 */ {(13<<2)|2,{109,87,0}},
/* 42903 */ {(13<<2)|2,{109,87,0}},
/* 42904 */ {(13<<2)|2,{109,89,0}},
/* 42905 */ {(13<<2)|2,{109,89,0}},
/* 42906 */ {(13<<2)|2,{109,89,0}},
/* 42907 */ {(13<<2)|2,{109,89,0}},
/* 42908 */ {(13<<2)|2,{109,89,0}},
/* 42909 */ {(13<<2)|2,{109,89,0}},
/* 42910 */ {(13<<2)|2,{109,89,0}},
/* 42911 */ {(13<<2)|2,{109,89,0}},
/* 42912 */ {(13<<2)|2,{109,106,0}},
/* 42913 */ {(13<<2)|2,{109,106,0}},
/* 42914 */ {(13<<2)|2,{109,106,0}},
/* 42915 */ {(13<<2)|2,{109,106,0}},
/* 42916 */ {(13<<2)|2,{109,106,0}},
/* 42917 */ {(13<<2)|2,{109,106,0}},
/* 42918 */ {(13<<2)|2,{109,106,0}},
/* 42919 */ {(13<<2)|2,{109,106,0}},
/* 42920 */ {(13<<2)|2,{109,107,0}},
/* 42921 */ {(13<<2)|2,{109,107,0}},
/* 42922 */ {(13<<2)|2,{109,107,0}},
/* 42923 */ {(13<<2)|2,{109,107,0}},
/* 42924 */ {(13<<2)|2,{109,107,0}},
/* 42925 */ {(13<<2)|2,{109,107,0}},
/* 42926 */ {(13<<2)|2,{109,107,0}},
/* 42927 */ {(13<<2)|2,{109,107,0}},
/* 42928 */ {(13<<2)|2,{109,113,0}},
/* 42929 */ {(13<<2)|2,{109,113,0}},
/* 42930 */ {(13<<2)|2,{109,113,0}},
/* 42931 */ {(13<<2)|2,{109,113,0}},
/* 42932 */ {(13<<2)|2,{109,113,0}},
/* 42933 */ {(13<<2)|2,{109,113,0}},
/* 42934 */ {(13<<2)|2,{109,113,0}},
/* 42935 */ {(13<<2)|2,{109,113,0}},
/* 42936 */ {(13<<2)|2,{109,118,0}},
/* 42937 */ {(13<<2)|2,{109,118,0}},
/* 42938 */ {(13<<2)|2,{109,118,0}},
/* 42939 */ {(13<<2)|2,{109,118,0}},
/* 42940 */ {(13<<2)|2,{109,118,0}},
/* 42941 */ {(13<<2)|2,{109,118,0}},
/* 42942 */ {(13<<2)|2,{109,118,0}},
/* 42943 */ {(13<<2)|2,{109,118,0}},
/* 42944 */ {(13<<2)|2,{109,119,0}},
/* 42945 */ {(13<<2)|2,{109,119,0}},
/* 42946 */ {(13<<2)|2,{109,119,0}},
/* 42947 */ {(13<<2)|2,{109,119,0}},
/* 42948 */ {(13<<2)|2,{109,119,0}},
/* 42949 */ {(13<<2)|2,{109,119,0}},
/* 42950 */ {(13<<2)|2,{109,119,0}},
/* 42951 */ {(13<<2)|2,{109,119,0}},
/* 42952 */ {(13<<2)|2,{109,120,0}},
/* 42953 */ {(13<<2)|2,{109,120,0}},
/* 42954 */ {(13<<2)|2,{109,120,0}},
/* 42955 */ {(13<<2)|2,{109,120,0}},
/* 42956 */ {(13<<2)|2,{109,120,0}},
/* 42957 */ {(13<<2)|2,{109,120,0}},
/* 42958 */ {(13<<2)|2,{109,120,0}},
/* 42959 */ {(13<<2)|2,{109,120,0}},
/* 42960 */ {(13<<2)|2,{109,121,0}},
/* 42961 */ {(13<<2)|2,{109,121,0}},
/* 42962 */ {(13<<2)|2,{109,121,0}},
/* 42963 */ {(13<<2)|2,{109,121,0}},
/* 42964 */ {(13<<2)|2,{109,121,0}},
/* 42965 */ {(13<<2)|2,{109,121,0}},
/* 42966 */ {(13<<2)|2,{109,121,0}},
/* 42967 */ {(13<<2)|2,{109,121,0}},
/* 42968 */ {(13<<2)|2,{109,122,0}},
/* 42969 */ {(13<<2)|2,{109,122,0}},
/* 42970 */ {(13<<2)|2,{109,122,0}},
/* 42971 */ {(13<<2)|2,{109,122,0}},
/* 42972 */ {(13<<2)|2,{109,122,0}},
/* 42973 */ {(13<<2)|2,{109,122,0}},
/* 42974 */ {(13<<2)|2,{109,122,0}},
/* 42975 */ {(13<<2)|2,{109,122,0}},
/* 42976 */ {(14<<2)|2,{109,38,0}},
/* 42977 */ {(14<<2)|2,{109,38,0}},
/* 42978 */ {(14<<2)|2,{109,38,0}},
/* 42979 */ {(14<<2)|2,{109,38,0}},
/* 42980 */ {(14<<2)|2,{109,42,0}},
/* 42981 */ {(14<<2)|2,{109,42,0}},
/* 42982 */ {(14<<2)|2,{109,42,0}},
/* 42983 */ {(14<<2)|2,{109,42,0}},
/* 42984 */ {(14<<2)|2,{109,44,0}},
/* 42985 */ {(14<<2)|2,{109,44,0}},
/* 42986 */ {(14<<2)|2,{109,44,0}},
/* 42987 */ {(14<<2)|2,{109,44,0}},
/* 42988 */ {(14<<2)|2,{109,59,0}},
/* 42989 */ {(14<<2)|2,{109,59,0}},
/* 42990 */ {(14<<2)|2,{109,59,0}},
/* 42991 */ {(14<<2)|2,{109,59,0}},
/* 42992 */ {(14<<2)|2,{109,88,0}},
/* 42993 */ {(14<<2)|2,{109,88,0}},
/* 42994 */ {(14<<2)|2,{109,88,0}},
/* 42995 */ {(14<<2)|2,{109,88,0}},
/* 42996 */ {(14<<2)|2,{109,90,0}},
/* 42997 */ {(14<<2)|2,{109,90,0}},
/* 42998 */ {(14<<2)|2,{109,90,0}},
/* 42999 */ {(14<<2)|2,{109,90,0}},
/* 43000 */ {(16<<2)|2,{109,33,0}},
/* 43001 */ {(16<<2)|2,{109,34,0}},
/* 43002 */ {(16<<2)|2,{109,40,0}},
/* 43003 */ {(16<<2)|2,{109,41,0}},
/* 43004 */ {(16<<2)|2,{109,63,0}},
/* 43005 */ {(6<<2)|1,{109,0,0}},
/* 43006 */ {(6<<2)|1,{109,0,0}},
/* 43007 */ {(6<<2)|1,{109,0,0}},
/* 43008 */ {(16<<2)|3,{110,48,48}},
/* 43009 */ {(16<<2)|3,{110,48,49}},
/* 43010 */ {(16<<2)|3,{110,48,50}},
/* 43011 */ {(16<<2)|3,{110,48,97}},
/* 43012 */ {(16<<2)|3,{110,48,99}},
/* 43013 */ {(16<<2)|3,{110,48,101}},
/* 43014 */ {(16<<2)|3,{110,48,105}},
/* 43015 */ {(16<<2)|3,{110,48,111}},
/* 43016 */ {(16<<2)|3,{110,48,115}},
/* 43017 */ {(16<<2)|3,{110,48,116}},
/* 43018 */ {(11<<2)|2,{110,48,0}},
/* 43019 */ {(11<<2)|2,{110,48,0}},
/* 43020 */ {(11<<2)|2,{110,48,0}},
/* 43021 */ {(11<<2)|2,{110,48,0}},
/* 43022 */ {(11<<2)|2,{110,48,0}},
/* 43023 */ {(11<<2)|2,{110,48,0}},
/* 43024 */ {(11<<2)|2,{110,48,0}},
/* 43025 */ {(11<<2)|2,{110,48,0}},
/* 43026 */ {(11<<2)|2,{110,48,0}},
/* 43027 */ {(11<<2)|2,{110,48,0}},
/* 43028 */ {(11<<2)|2,{110,48,0}},
/* 43029 */ {(11<<2)|2,{110,48,0}},
/* 43030 */ {(11<<2)|2,{110,48,0}},
/* 43031 */ {(11<<2)|2,{110,48,0}},
/* 43032 */ {(11<<2)|2,{110,48,0}},
/* 43033 */ {(11<<2)|2,{110,48,0}},
/* 43034 */ {(11<<2)|2,{110,48,0}},
/* 43035 */ {(11<<2)|2,{110,48,0}},
/* 43036 */ {(11<<2)|2,{110,48,0}},
/* 43037 */ {(11<<2)|2,{110,48,0}},
/* 43038 */ {(11<<2)|2,{110,48,0}},
/* 43039 */ {(11<<2)|2,{110,48,0}},
/* 43040 */ {(16<<2)|3,{110,49,48}},
/* 43041 */ {(16<<2)|3,{110,49,49}},
/* 43042 */ {(16<<2)|3,{110,49,50}},
/* 43043 */ {(16<<2)|3,{110,49,97}},
/* 43044 */ {(16<<2)|3,{110,49,99}},
/* 43045 */ {(16<<2)|3,{110,49,101}},
/* 43046 */ {(16<<2)|3,{110,49,105}},
/* 43047 */ {(16<<2)|3,{110,49,111}},
/* 43048 */ {(16<<2)|3,{110,49,115}},
/* 43049 */ {(16<<2)|3,{110,49,116}},
/* 43050 */ {(11<<2)|2,{110,49,0}},
/* 43051 */ {(11<<2)|2,{110,49,0}},
/* 43052 */ {(11<<2)|2,{110,49,0}},
/* 43053 */ {(11<<2)|2,{110,49,0}},
/* 43054 */ {(11<<2)|2,{110,49,0}},
/* 43055 */ {(11<<2)|2,{110,49,0}},
/* 43056 */ {(11<<2)|2,{110,49,0}},
/* 43057 */ {(11<<2)|2,{110,49,0}},
/* 43058 */ {(11<<2)|2,{110,49,0}},
/* 43059 */ {(11<<2)|2,{110,49,0}},
/* 43060 */ {(11<<2)|2,{110,49,0}},
/* 43061 */ {(11<<2)|2,{110,49,0}},
/* 43062 */ {(11<<2)|2,{110,49,0}},
/* 43063 */ {(11<<2)|2,{110,49,0}},
/* 43064 */ {(11<<2)|2,{110,49,0}},
/* 43065 */ {(11<<2)|2,{110,49,0}},
/* 43066 */ {(11<<2)|2,{110,49,0}},
/* 43067 */ {(11<<2)|2,{110,49,0}},
/* 43068 */ {(11<<2)|2,{110,49,0}},
/* 43069 */ {(11<<2)|2,{110,49,0}},
/* 43070 */ {(11<<2)|2,{110,49,0}},
/* 43071 */ {(11<<2)|2,{110,49,0}},
/* 43072 */ {(16<<2)|3,{110,50,48}},
/* 43073 */ {(16<<2)|3,{110,50,49}},
/* 43074 */ {(16<<2)|3,{110,50,50}},
/* 43075 */ {(16<<2)|3,{110,50,97}},
/* 43076 */ {(16<<2)|3,{110,50,99}},
/* 43077 */ {(16<<2)|3,{110,50,101}},
/* 43078 */ {(16<<2)|3,{110,50,105}},
/* 43079 */ {(16<<2)|3,{110,50,111}},
/* 43080 */ {(16<<2)|3,{110,50,115}},
/* 43081 */ {(16<<2)|3,{110,50,116}},
/* 43082 */ {(11<<2)|2,{110,50,0}},
/* 43083 */ {(11<<2)|2,{110,50,0}},
/* 43084 */ {(11<<2)|2,{110,50,0}},
/* 43085 */ {(11<<2)|2,{110,50,0}},
/* 43086 */ {(11<<2)|2,{110,50,0}},
/* 43087 */ {(11<<2)|2,{110,50,0}},
/* 43088 */ {(11<<2)|2,{110,50,0}},
/* 43089 */ {(11<<2)|2,{110,50,0}},
/* 43090 */ {(11<<2)|2,{110,50,0}},
/* 43091 */ {(11<<2)|2,{110,50,0}},
/* 43092 */ {(11<<2)|2,{110,50,0}},
/* 43093 */ {(11<<2)|2,{110,50,0}},
/* 43094 */ {(11<<2)|2,{110,50,0}},
/* 43095 */ {(11<<2)|2,{110,50,0}},
/* 43096 */ {(11<<2)|2,{110,50,0}},
/* 43097 */ {(11<<2)|2,{110,50,0}},
/* 43098 */ {(11<<2)|2,{110,50,0}},
/* 43099 */ {(11<<2)|2,{110,50,0}},
/* 43100 */ {(11<<2)|2,{110,50,0}},
/* 43101 */ {(11<<2)|2,{110,50,0}},
/* 43102 */ {(11<<2)|2,{110,50,0}},
/* 43103 */ {(11<<2)|2,{110,50,0}},
/* 43104 */ {(16<<2)|3,{110,97,48}},
/* 43105 */ {(16<<2)|3,{110,97,49}},
/* 43106 */ {(16<<2)|3,{110,97,50}},
/* 43107 */ {(16<<2)|3,{110,97,97}},
/* 43108 */ {(16<<2)|3,{110,97,99}},
/* 43109 */ {(16<<2)|3,{110,97,101}},
/* 43110 */ {(16<<2)|3,{110,97,105}},
/* 43111 */ {(16<<2)|3,{110,97,111}},
/* 43112 */ {(16<<2)|3,{110,97,115}},
/* 43113 */ {(16<<2)|3,{110,97,116}},
/* 43114 */ {(11<<2)|2,{110,97,0}},
/* 43115 */ {(11<<2)|2,{110,97,0}},
/* 43116 */ {(11<<2)|2,{110,97,0}},
/* 43117 */ {(11<<2)|2,{110,97,0}},
/* 43118 */ {(11<<2)|2,{110,97,0}},
/* 43119 */ {(11<<2)|2,{110,97,0}},
/* 43120 */ {(11<<2)|2,{110,97,0}},
/* 43121 */ {(11<<2)|2,{110,97,0}},
/* 43122 */ {(11<<2)|2,{110,97,0}},
/* 43123 */ {(11<<2)|2,{110,97,0}},
/* 43124 */ {(11<<2)|2,{110,97,0}},
/* 43125 */ {(11<<2)|2,{110,97,0}},
/* 43126 */ {(11<<2)|2,{110,97,0}},
/* 43127 */ {(11<<2)|2,{110,97,0}},
/* 43128 */ {(11<<2)|2,{110,97,0}},
/* 43129 */ {(11<<2)|2,{110,97,0}},
/* 43130 */ {(11<<2)|2,{110,97,0}},
/* 43131 */ {(11<<2)|2,{110,97,0}},
/* 43132 */ {(11<<2)|2,{110,97,0}},
/* 43133 */ {(11<<2)|2,{110,97,0}},
/* 43134 */ {(11<<2)|2,{110,97,0}},
/* 43135 */ {(11<<2)|2,{110,97,0}},
/* 43136 */ {(16<<2)|3,{110,99,48}},
/* 43137 */ {(16<<2)|3,{110,99,49}},
/* 43138 */ {(16<<2)|3,{110,99,50}},
/* 43139 */ {(16<<2)|3,{110,99,97}},
/* 43140 */ {(16<<2)|3,{110,99,99}},
/* 43141 */ {(16<<2)|3,{110,99,101}},
/* 43142 */ {(16<<2)|3,{110,99,105}},
/* 43143 */ {(16<<2)|3,{110,99,111}},
/* 43144 */ {(16<<2)|3,{110,99,115}},
/* 43145 */ {(16<<2)|3,{110,99,116}},
/* 43146 */ {(11<<2)|2,{110,99,0}},
/* 43147 */ {(11<<2)|2,{110,99,0}},
/* 43148 */ {(11<<2)|2,{110,99,0}},
/* 43149 */ {(11<<2)|2,{110,99,0}},
/* 43150 */ {(11<<2)|2,{110,99,0}},
/* 43151 */ {(11<<2)|2,{110,99,0}},
/* 43152 */ {(11<<2)|2,{110,99,0}},
/* 43153 */ {(11<<2)|2,{110,99,0}},
/* 43154 */ {(11<<2)|2,{110,99,0}},
/* 43155 */ {(11<<2)|2,{110,99,0}},
/* 43156 */ {(11<<2)|2,{110,99,0}},
/* 43157 */ {(11<<2)|2,{110,99,0}},
/* 43158 */ {(11<<2)|2,{110,99,0}},
/* 43159 */ {(11<<2)|2,{110,99,0}},
/* 43160 */ {(11<<2)|2,{110,99,0}},
/* 43161 */ {(11<<2)|2,{110,99,0}},
/* 43162 */ {(11<<2)|2,{110,99,0}},
/* 43163 */ {(11<<2)|2,{110,99,0}},
/* 43164 */ {(11<<2)|2,{110,99,0}},
/* 43165 */ {(11<<2)|2,{110,99,0}},
/* 43166 */ {(11<<2)|2,{110,99,0}},
/* 43167 */ {(11<<2)|2,{110,99,0}},
/* 43168 */ {(16<<2)|3,{110,101,48}},
/* 43169 */ {(16<<2)|3,{110,101,49}},
/* 43170 */ {(16<<2)|3,{110,101,50}},
/* 43171 */ {(16<<2)|3,{110,101,97}},
/* 43172 */ {(16<<2)|3,{110,101,99}},
/* 43173 */ {(16<<2)|3,{110,101,101}},
/* 43174 */ {(16<<2)|3,{110,101,105}},
/* 43175 */ {(16<<2)|3,{110,101,111}},
/* 43176 */ {(16<<2)|3,{110,101,115}},
/* 43177 */ {(16<<2)|3,{110,101,116}},
/* 43178 */ {(11<<2)|2,{110,101,0}},
/* 43179 */ {(11<<2)|2,{110,101,0}},
/* 43180 */ {(11<<2)|2,{110,101,0}},
/* 43181 */ {(11<<2)|2,{110,101,0}},
/* 43182 */ {(11<<2)|2,{110,101,0}},
/* 43183 */ {(11<<2)|2,{110,101,0}},
/* 43184 */ {(11<<2)|2,{110,101,0}},
/* 43185 */ {(11<<2)|2,{110,101,0}},
/* 43186 */ {(11<<2)|2,{110,101,0}},
/* 43187 */ {(11<<2)|2,{110,101,0}},
/* 43188 */ {(11<<2)|2,{110,101,0}},
/* 43189 */ {(11<<2)|2,{110,101,0}},
/* 43190 */ {(11<<2)|2,{110,101,0}},
/* 43191 */ {(11<<2)|2,{110,101,0}},
/* 43192 */ {(11<<2)|2,{110,101,0}},
/* 43193 */ {(11<<2)|2,{110,101,0}},
/* 43194 */ {(11<<2)|2,{110,101,0}},
/* 43195 */ {(11<<2)|2,{110,101,0}},
/* 43196 */ {(11<<2)|2,{110,101,0}},
/* 43197 */ {(11<<2)|2,{110,101,0}},
/* 43198 */ {(11<<2)|2,{110,101,0}},
/* 43199 */ {(11<<2)|2,{110,101,0}},
/* 43200 */ {(16<<2)|3,{110,105,48}},
/* 43201 */ {(16<<2)|3,{110,105,49}},
/* 43202 */ {(16<<2)|3,{110,105,50}},
/* 43203 */ {(16<<2)|3,{110,105,97}},
/* 43204 */ {(16<<2)|3,{110,105,99}},
/* 43205 */ {(16<<2)|3,{110,105,101}},
/* 43206 */ {(16<<2)|3,{110,105,105}},
/* 43207 */ {(16<<2)|3,{110,105,111}},
/* 43208 */ {(16<<2)|3,{110,105,115}},
/* 43209 */ {(16<<2)|3,{110,105,116}},
/* 43210 */ {(11<<2)|2,{110,105,0}},
/* 43211 */ {(11<<2)|2,{110,105,0}},
/* 43212 */ {(11<<2)|2,{110,105,0}},
/* 43213 */ {(11<<2)|2,{110,105,0}},
/* 43214 */ {(11<<2)|2,{110,105,0}},
/* 43215 */ {(11<<2)|2,{110,105,0}},
/* 43216 */ {(11<<2)|2,{110,105,0}},
/* 43217 */ {(11<<2)|2,{110,105,0}},
/* 43218 */ {(11<<2)|2,{110,105,0}},
/* 43219 */ {(11<<2)|2,{110,105,0}},
/* 43220 */ {(11<<2)|2,{110,105,0}},
/* 43221 */ {(11<<2)|2,{110,105,0}},
/* 43222 */ {(11<<2)|2,{110,105,0}},
/* 43223 */ {(11<<2)|2,{110,105,0}},
/* 43224 */ {(11<<2)|2,{110,105,0}},
/* 43225 */ {(11<<2)|2,{110,105,0}},
/* 43226 */ {(11<<2)|2,{110,105,0}},
/* 43227 */ {(11<<2)|2,{110,105,0}},
/* 43228 */ {(11<<2)|2,{110,105,0}},
/* 43229 */ {(11<<2)|2,{110,105,0}},
/* 43230 */ {(11<<2)|2,{110,105,0}},
/* 43231 */ {(11<<2)|2,{110,105,0}},
/* 43232 */ {(16<<2)|3,{110,111,48}},
/* 43233 */ {(16<<2)|3,{110,111,49}},
/* 43234 */ {(16<<2)|3,{110,111,50}},
/* 43235 */ {(16<<2)|3,{110,111,97}},
/* 43236 */ {(16<<2)|3,{110,111,99}},
/* 43237 */ {(16<<2)|3,{110,111,101}},
/* 43238 */ {(16<<2)|3,{110,111,105}},
/* 43239 */ {(16<<2)|3,{110,111,111}},
/* 43240 */ {(16<<2)|3,{110,111,115}},
/* 43241 */ {(16<<2)|3,{110,111,116}},
/* 43242 */ {(11<<2)|2,{110,111,0}},
/* 43243 */ {(11<<2)|2,{110,111,0}},
/* 43244 */ {(11<<2)|2,{110,111,0}},
/* 43245 */ {(11<<2)|2,{110,111,0}},
/* 43246 */ {(11<<2)|2,{110,111,0}},
/* 43247 */ {(11<<2)|2,{110,111,0}},
/* 43248 */ {(11<<2)|2,{110,111,0}},
/* 43249 */ {(11<<2)|2,{110,111,0}},
/* 43250 */ {(11<<2)|2,{110,111,0}},
/* 43251 */ {(11<<2)|2,{110,111,0}},
/* 43252 */ {(11<<2)|2,{110,111,0}},
/* 43253 */ {(11<<2)|2,{110,111,0}},
/* 43254 */ {(11<<2)|2,{110,111,0}},
/* 43255 */ {(11<<2)|2,{110,111,0}},
/* 43256 */ {(11<<2)|2,{110,111,0}},
/* 43257 */ {(11<<2)|2,{110,111,0}},
/* 43258 */ {(11<<2)|2,{110,111,0}},
/* 43259 */ {(11<<2)|2,{110,111,0}},
/* 43260 */ {(11<<2)|2,{110,111,0}},
/* 43261 */ {(11<<2)|2,{110,111,0}},
/* 43262 */ {(11<<2)|2,{110,111,0}},
/* 43263 */ {(11<<2)|2,{110,111,0}},
/* 43264 */ {(16<<2)|3,{110,115,48}},
/* 43265 */ {(16<<2)|3,{110,115,49}},
/* 43266 */ {(16<<2)|3,{110,115,50}},
/* 43267 */ {(16<<2)|3,{110,115,97}},
/* 43268 */ {(16<<2)|3,{110,115,99}},
/* 43269 */ {(16<<2)|3,{110,115,101}},
/* 43270 */ {(16<<2)|3,{110,115,105}},
/* 43271 */ {(16<<2)|3,{110,115,111}},
/* 43272 */ {(16<<2)|3,{110,115,115}},
/* 43273 */ {(16<<2)|3,{110,115,116}},
/* 43274 */ {(11<<2)|2,{110,115,0}},
/* 43275 */ {(11<<2)|2,{110,115,0}},
/* 43276 */ {(11<<2)|2,{110,115,0}},
/* 43277 */ {(11<<2)|2,{110,115,0}},
/* 43278 */ {(11<<2)|2,{110,115,0}},
/* 43279 */ {(11<<2)|2,{110,115,0}},
/* 43280 */ {(11<<2)|2,{110,115,0}},
/* 43281 */ {(11<<2)|2,{110,115,0}},
/* 43282 */ {(11<<2)|2,{110,115,0}},
/* 43283 */ {(11<<2)|2,{110,115,0}},
/* 43284 */ {(11<<2)|2,{110,115,0}},
/* 43285 */ {(11<<2)|2,{110,115,0}},
/* 43286 */ {(11<<2)|2,{110,115,0}},
/* 43287 */ {(11<<2)|2,{110,115,0}},
/* 43288 */ {(11<<2)|2,{110,115,0}},
/* 43289 */ {(11<<2)|2,{110,115,0}},
/* 43290 */ {(11<<2)|2,{110,115,0}},
/* 43291 */ {(11<<2)|2,{110,115,0}},
/* 43292 */ {(11<<2)|2,{110,115,0}},
/* 43293 */ {(11<<2)|2,{110,115,0}},
/* 43294 */ {(11<<2)|2,{110,115,0}},
/* 43295 */ {(11<<2)|2,{110,115,0}},
/* 43296 */ {(16<<2)|3,{110,116,48}},
/* 43297 */ {(16<<2)|3,{110,116,49}},
/* 43298 */ {(16<<2)|3,{110,116,50}},
/* 43299 */ {(16<<2)|3,{110,116,97}},
/* 43300 */ {(16<<2)|3,{110,116,99}},
/* 43301 */ {(16<<2)|3,{110,116,101}},
/* 43302 */ {(16<<2)|3,{110,116,105}},
/* 43303 */ {(16<<2)|3,{110,116,111}},
/* 43304 */ {(16<<2)|3,{110,116,115}},
/* 43305 */ {(16<<2)|3,{110,116,116}},
/* 43306 */ {(11<<2)|2,{110,116,0}},
/* 43307 */ {(11<<2)|2,{110,116,0}},
/* 43308 */ {(11<<2)|2,{110,116,0}},
/* 43309 */ {(11<<2)|2,{110,116,0}},
/* 43310 */ {(11<<2)|2,{110,116,0}},
/* 43311 */ {(11<<2)|2,{110,116,0}},
/* 43312 */ {(11<<2)|2,{110,116,0}},
/* 43313 */ {(11<<2)|2,{110,116,0}},
/* 43314 */ {(11<<2)|2,{110,116,0}},
/* 43315 */ {(11<<2)|2,{110,116,0}},
/* 43316 */ {(11<<2)|2,{110,116,0}},
/* 43317 */ {(11<<2)|2,{110,116,0}},
/* 43318 */ {(11<<2)|2,{110,116,0}},
/* 43319 */ {(11<<2)|2,{110,116,0}},
/* 43320 */ {(11<<2)|2,{110,116,0}},
/* 43321 */ {(11<<2)|2,{110,116,0}},
/* 43322 */ {(11<<2)|2,{110,116,0}},
/* 43323 */ {(11<<2)|2,{110,116,0}},
/* 43324 */ {(11<<2)|2,{110,116,0}},
/* 43325 */ {(11<<2)|2,{110,116,0}},
/* 43326 */ {(11<<2)|2,{110,116,0}},
/* 43327 */ {(11<<2)|2,{110,116,0}},
/* 43328 */ {(12<<2)|2,{110,32,0}},
/* 43329 */ {(12<<2)|2,{110,32,0}},
/* 43330 */ {(12<<2)|2,{110,32,0}},
/* 43331 */ {(12<<2)|2,{110,32,0}},
/* 43332 */ {(12<<2)|2,{110,32,0}},
/* 43333 */ {(12<<2)|2,{110,32,0}},
/* 43334 */ {(12<<2)|2,{110,32,0}},
/* 43335 */ {(12<<2)|2,{110,32,0}},
/* 43336 */ {(12<<2)|2,{110,32,0}},
/* 43337 */ {(12<<2)|2,{110,32,0}},
/* 43338 */ {(12<<2)|2,{110,32,0}},
/* 43339 */ {(12<<2)|2,{110,32,0}},
/* 43340 */ {(12<<2)|2,{110,32,0}},
/* 43341 */ {(12<<2)|2,{110,32,0}},
/* 43342 */ {(12<<2)|2,{110,32,0}},
/* 43343 */ {(12<<2)|2,{110,32,0}},
/* 43344 */ {(12<<2)|2,{110,37,0}},
/* 43345 */ {(12<<2)|2,{110,37,0}},
/* 43346 */ {(12<<2)|2,{110,37,0}},
/* 43347 */ {(12<<2)|2,{110,37,0}},
/* 43348 */ {(12<<2)|2,{110,37,0}},
/* 43349 */ {(12<<2)|2,{110,37,0}},
/* 43350 */ {(12<<2)|2,{110,37,0}},
/* 43351 */ {(12<<2)|2,{110,37,0}},
/* 43352 */ {(12<<2)|2,{110,37,0}},
/* 43353 */ {(12<<2)|2,{110,37,0}},
/* 43354 */ {(12<<2)|2,{110,37,0}},
/* 43355 */ {(12<<2)|2,{110,37,0}},
/* 43356 */ {(12<<2)|2,{110,37,0}},
/* 43357 */ {(12<<2)|2,{110,37,0}},
/* 43358 */ {(12<<2)|2,{110,37,0}},
/* 43359 */ {(12<<2)|2,{110,37,0}},
/* 43360 */ {(12<<2)|2,{110,45,0}},
/* 43361 */ {(12<<2)|2,{110,45,0}},
/* 43362 */ {(12<<2)|2,{110,45,0}},
/* 43363 */ {(12<<2)|2,{110,45,0}},
/* 43364 */ {(12<<2)|2,{110,45,0}},
/* 43365 */ {(12<<2)|2,{110,45,0}},
/* 43366 */ {(12<<2)|2,{110,45,0}},
/* 43367 */ {(12<<2)|2,{110,45,0}},
/* 43368 */ {(12<<2)|2,{110,45,0}},
/* 43369 */ {(12<<2)|2,{110,45,0}},
/* 43370 */ {(12<<2)|2,{110,45,0}},
/* 43371 */ {(12<<2)|2,{110,45,0}},
/* 43372 */ {(12<<2)|2,{110,45,0}},
/* 43373 */ {(12<<2)|2,{110,45,0}},
/* 43374 */ {(12<<2)|2,{110,45,0}},
/* 43375 */ {(12<<2)|2,{110,45,0}},
/* 43376 */ {(12<<2)|2,{110,46,0}},
/* 43377 */ {(12<<2)|2,{110,46,0}},
/* 43378 */ {(12<<2)|2,{110,46,0}},
/* 43379 */ {(12<<2)|2,{110,46,0}},
/* 43380 */ {(12<<2)|2,{110,46,0}},
/* 43381 */ {(12<<2)|2,{110,46,0}},
/* 43382 */ {(12<<2)|2,{110,46,0}},
/* 43383 */ {(12<<2)|2,{110,46,0}},
/* 43384 */ {(12<<2)|2,{110,46,0}},
/* 43385 */ {(12<<2)|2,{110,46,0}},
/* 43386 */ {(12<<2)|2,{110,46,0}},
/* 43387 */ {(12<<2)|2,{110,46,0}},
/* 43388 */ {(12<<2)|2,{110,46,0}},
/* 43389 */ {(12<<2)|2,{110,46,0}},
/* 43390 */ {(12<<2)|2,{110,46,0}},
/* 43391 */ {(12<<2)|2,{110,46,0}},
/* 43392 */ {(12<<2)|2,{110,47,0}},
/* 43393 */ {(12<<2)|2,{110,47,0}},
/* 43394 */ {(12<<2)|2,{110,47,0}},
/* 43395 */ {(12<<2)|2,{110,47,0}},
/* 43396 */ {(12<<2)|2,{110,47,0}},
/* 43397 */ {(12<<2)|2,{110,47,0}},
/* 43398 */ {(12<<2)|2,{110,47,0}},
/* 43399 */ {(12<<2)|2,{110,47,0}},
/* 43400 */ {(12<<2)|2,{110,47,0}},
/* 43401 */ {(12<<2)|2,{110,47,0}},
/* 43402 */ {(12<<2)|2,{110,47,0}},
/* 43403 */ {(12<<2)|2,{110,47,0}},
/* 43404 */ {(12<<2)|2,{110,47,0}},
/* 43405 */ {(12<<2)|2,{110,47,0}},
/* 43406 */ {(12<<2)|2,{110,47,0}},
/* 43407 */ {(12<<2)|2,{110,47,0}},
/* 43408 */ {(12<<2)|2,{110,51,0}},
/* 43409 */ {(12<<2)|2,{110,51,0}},
/* 43410 */ {(12<<2)|2,{110,51,0}},
/* 43411 */ {(12<<2)|2,{110,51,0}},
/* 43412 */ {(12<<2)|2,{110,51,0}},
/* 43413 */ {(12<<2)|2,{110,51,0}},
/* 43414 */ {(12<<2)|2,{110,51,0}},
/* 43415 */ {(12<<2)|2,{110,51,0}},
/* 43416 */ {(12<<2)|2,{110,51,0}},
/* 43417 */ {(12<<2)|2,{110,51,0}},
/* 43418 */ {(12<<2)|2,{110,51,0}},
/* 43419 */ {(12<<2)|2,{110,51,0}},
/* 43420 */ {(12<<2)|2,{110,51,0}},
/* 43421 */ {(12<<2)|2,{110,51,0}},
/* 43422 */ {(12<<2)|2,{110,51,0}},
/* 43423 */ {(12<<2)|2,{110,51,0}},
/* 43424 */ {(12<<2)|2,{110,52,0}},
/* 43425 */ {(12<<2)|2,{110,52,0}},
/* 43426 */ {(12<<2)|2,{110,52,0}},
/* 43427 */ {(12<<2)|2,{110,52,0}},
/* 43428 */ {(12<<2)|2,{110,52,0}},
/* 43429 */ {(12<<2)|2,{110,52,0}},
/* 43430 */ {(12<<2)|2,{110,52,0}},
/* 43431 */ {(12<<2)|2,{110,52,0}},
/* 43432 */ {(12<<2)|2,{110,52,0}},
/* 43433 */ {(12<<2)|2,{110,52,0}},
/* 43434 */ {(12<<2)|2,{110,52,0}},
/* 43435 */ {(12<<2)|2,{110,52,0}},
/* 43436 */ {(12<<2)|2,{110,52,0}},
/* 43437 */ {(12<<2)|2,{110,52,0}},
/* 43438 */ {(12<<2)|2,{110,52,0}},
/* 43439 */ {(12<<2)|2,{110,52,0}},
/* 43440 */ {(12<<2)|2,{110,53,0}},
/* 43441 */ {(12<<2)|2,{110,53,0}},
/* 43442 */ {(12<<2)|2,{110,53,0}},
/* 43443 */ {(12<<2)|2,{110,53,0}},
/* 43444 */ {(12<<2)|2,{110,53,0}},
/* 43445 */ {(12<<2)|2,{110,53,0}},
/* 43446 */ {(12<<2)|2,{110,53,0}},
/* 43447 */ {(12<<2)|2,{110,53,0}},
/* 43448 */ {(12<<2)|2,{110,53,0}},
/* 43449 */ {(12<<2)|2,{110,53,0}},
/* 43450 */ {(12<<2)|2,{110,53,0}},
/* 43451 */ {(12<<2)|2,{110,53,0}},
/* 43452 */ {(12<<2)|2,{110,53,0}},
/* 43453 */ {(12<<2)|2,{110,53,0}},
/* 43454 */ {(12<<2)|2,{110,53,0}},
/* 43455 */ {(12<<2)|2,{110,53,0}},
/* 43456 */ {(12<<2)|2,{110,54,0}},
/* 43457 */ {(12<<2)|2,{110,54,0}},
/* 43458 */ {(12<<2)|2,{110,54,0}},
/* 43459 */ {(12<<2)|2,{110,54,0}},
/* 43460 */ {(12<<2)|2,{110,54,0}},
/* 43461 */ {(12<<2)|2,{110,54,0}},
/* 43462 */ {(12<<2)|2,{110,54,0}},
/* 43463 */ {(12<<2)|2,{110,54,0}},
/* 43464 */ {(12<<2)|2,{110,54,0}},
/* 43465 */ {(12<<2)|2,{110,54,0}},
/* 43466 */ {(12<<2)|2,{110,54,0}},
/* 43467 */ {(12<<2)|2,{110,54,0}},
/* 43468 */ {(12<<2)|2,{110,54,0}},
/* 43469 */ {(12<<2)|2,{110,54,0}},
/* 43470 */ {(12<<2)|2,{110,54,0}},
/* 43471 */ {(12<<2)|2,{110,54,0}},
/* 43472 */ {(12<<2)|2,{110,55,0}},
/* 43473 */ {(12<<2)|2,{110,55,0}},
/* 43474 */ {(12<<2)|2,{110,55,0}},
/* 43475 */ {(12<<2)|2,{110,55,0}},
/* 43476 */ {(12<<2)|2,{110,55,0}},
/* 43477 */ {(12<<2)|2,{110,55,0}},
/* 43478 */ {(12<<2)|2,{110,55,0}},
/* 43479 */ {(12<<2)|2,{110,55,0}},
/* 43480 */ {(12<<2)|2,{110,55,0}},
/* 43481 */ {(12<<2)|2,{110,55,0}},
/* 43482 */ {(12<<2)|2,{110,55,0}},
/* 43483 */ {(12<<2)|2,{110,55,0}},
/* 43484 */ {(12<<2)|2,{110,55,0}},
/* 43485 */ {(12<<2)|2,{110,55,0}},
/* 43486 */ {(12<<2)|2,{110,55,0}},
/* 43487 */ {(12<<2)|2,{110,55,0}},
/* 43488 */ {(12<<2)|2,{110,56,0}},
/* 43489 */ {(12<<2)|2,{110,56,0}},
/* 43490 */ {(12<<2)|2,{110,56,0}},
/* 43491 */ {(12<<2)|2,{110,56,0}},
/* 43492 */ {(12<<2)|2,{110,56,0}},
/* 43493 */ {(12<<2)|2,{110,56,0}},
/* 43494 */ {(12<<2)|2,{110,56,0}},
/* 43495 */ {(12<<2)|2,{110,56,0}},
/* 43496 */ {(12<<2)|2,{110,56,0}},
/* 43497 */ {(12<<2)|2,{110,56,0}},
/* 43498 */ {(12<<2)|2,{110,56,0}},
/* 43499 */ {(12<<2)|2,{110,56,0}},
/* 43500 */ {(12<<2)|2,{110,56,0}},
/* 43501 */ {(12<<2)|2,{110,56,0}},
/* 43502 */ {(12<<2)|2,{110,56,0}},
/* 43503 */ {(12<<2)|2,{110,56,0}},
/* 43504 */ {(12<<2)|2,{110,57,0}},
/* 43505 */ {(12<<2)|2,{110,57,0}},
/* 43506 */ {(12<<2)|2,{110,57,0}},
/* 43507 */ {(12<<2)|2,{110,57,0}},
/* 43508 */ {(12<<2)|2,{110,57,0}},
/* 43509 */ {(12<<2)|2,{110,57,0}},
/* 43510 */ {(12<<2)|2,{110,57,0}},
/* 43511 */ {(12<<2)|2,{110,57,0}},
/* 43512 */ {(12<<2)|2,{110,57,0}},
/* 43513 */ {(12<<2)|2,{110,57,0}},
/* 43514 */ {(12<<2)|2,{110,57,0}},
/* 43515 */ {(12<<2)|2,{110,57,0}},
/* 43516 */ {(12<<2)|2,{110,57,0}},
/* 43517 */ {(12<<2)|2,{110,57,0}},
/* 43518 */ {(12<<2)|2,{110,57,0}},
/* 43519 */ {(12<<2)|2,{110,57,0}},
/* 43520 */ {(12<<2)|2,{110,61,0}},
/* 43521 */ {(12<<2)|2,{110,61,0}},
/* 43522 */ {(12<<2)|2,{110,61,0}},
/* 43523 */ {(12<<2)|2,{110,61,0}},
/* 43524 */ {(12<<2)|2,{110,61,0}},
/* 43525 */ {(12<<2)|2,{110,61,0}},
/* 43526 */ {(12<<2)|2,{110,61,0}},
/* 43527 */ {(12<<2)|2,{110,61,0}},
/* 43528 */ {(12<<2)|2,{110,61,0}},
/* 43529 */ {(12<<2)|2,{110,61,0}},
/* 43530 */ {(12<<2)|2,{110,61,0}},
/* 43531 */ {(12<<2)|2,{110,61,0}},
/* 43532 */ {(12<<2)|2,{110,61,0}},
/* 43533 */ {(12<<2)|2,{110,61,0}},
/* 43534 */ {(12<<2)|2,{110,61,0}},
/* 43535 */ {(12<<2)|2,{110,61,0}},
/* 43536 */ {(12<<2)|2,{110,65,0}},
/* 43537 */ {(12<<2)|2,{110,65,0}},
/* 43538 */ {(12<<2)|2,{110,65,0}},
/* 43539 */ {(12<<2)|2,{110,65,0}},
/* 43540 */ {(12<<2)|2,{110,65,0}},
/* 43541 */ {(12<<2)|2,{110,65,0}},
/* 43542 */ {(12<<2)|2,{110,65,0}},
/* 43543 */ {(12<<2)|2,{110,65,0}},
/* 43544 */ {(12<<2)|2,{110,65,0}},
/* 43545 */ {(12<<2)|2,{110,65,0}},
/* 43546 */ {(12<<2)|2,{110,65,0}},
/* 43547 */ {(12<<2)|2,{110,65,0}},
/* 43548 */ {(12<<2)|2,{110,65,0}},
/* 43549 */ {(12<<2)|2,{110,65,0}},
/* 43550 */ {(12<<2)|2,{110,65,0}},
/* 43551 */ {(12<<2)|2,{110,65,0}},
/* 43552 */ {(12<<2)|2,{110,95,0}},
/* 43553 */ {(12<<2)|2,{110,95,0}},
/* 43554 */ {(12<<2)|2,{110,95,0}},
/* 43555 */ {(12<<2)|2,{110,95,0}},
/* 43556 */ {(12<<2)|2,{110,95,0}},
/* 43557 */ {(12<<2)|2,{110,95,0}},
/* 43558 */ {(12<<2)|2,{110,95,0}},
/* 43559 */ {(12<<2)|2,{110,95,0}},
/* 43560 */ {(12<<2)|2,{110,95,0}},
/* 43561 */ {(12<<2)|2,{110,95,0}},
/* 43562 */ {(12<<2)|2,{110,95,0}},
/* 43563 */ {(12<<2)|2,{110,95,0}},
/* 43564 */ {(12<<2)|2,{110,95,0}},
/* 43565 */ {(12<<2)|2,{110,95,0}},
/* 43566 */ {(12<<2)|2,{110,95,0}},
/* 43567 */ {(12<<2)|2,{110,95,0}},
/* 43568 */ {(12<<2)|2,{110,98,0}},
/* 43569 */ {(12<<2)|2,{110,98,0}},
/* 43570 */ {(12<<2)|2,{110,98,0}},
/* 43571 */ {(12<<2)|2,{110,98,0}},
/* 43572 */ {(12<<2)|2,{110,98,0}},
/* 43573 */ {(12<<2)|2,{110,98,0}},
/* 43574 */ {(12<<2)|2,{110,98,0}},
/* 43575 */ {(12<<2)|2,{110,98,0}},
/* 43576 */ {(12<<2)|2,{110,98,0}},
/* 43577 */ {(12<<2)|2,{110,98,0}},
/* 43578 */ {(12<<2)|2,{110,98,0}},
/* 43579 */ {(12<<2)|2,{110,98,0}},
/* 43580 */ {(12<<2)|2,{110,98,0}},
/* 43581 */ {(12<<2)|2,{110,98,0}},
/* 43582 */ {(12<<2)|2,{110,98,0}},
/* 43583 */ {(12<<2)|2,{110,98,0}},
/* 43584 */ {(12<<2)|2,{110,100,0}},
/* 43585 */ {(12<<2)|2,{110,100,0}},
/* 43586 */ {(12<<2)|2,{110,100,0}},
/* 43587 */ {(12<<2)|2,{110,100,0}},
/* 43588 */ {(12<<2)|2,{110,100,0}},
/* 43589 */ {(12<<2)|2,{110,100,0}},
/* 43590 */ {(12<<2)|2,{110,100,0}},
/* 43591 */ {(12<<2)|2,{110,100,0}},
/* 43592 */ {(12<<2)|2,{110,100,0}},
/* 43593 */ {(12<<2)|2,{110,100,0}},
/* 43594 */ {(12<<2)|2,{110,100,0}},
/* 43595 */ {(12<<2)|2,{110,100,0}},
/* 43596 */ {(12<<2)|2,{110,100,0}},
/* 43597 */ {(12<<2)|2,{110,100,0}},
/* 43598 */ {(12<<2)|2,{110,100,0}},
/* 43599 */ {(12<<2)|2,{110,100,0}},
/* 43600 */ {(12<<2)|2,{110,102,0}},
/* 43601 */ {(12<<2)|2,{110,102,0}},
/* 43602 */ {(12<<2)|2,{110,102,0}},
/* 43603 */ {(12<<2)|2,{110,102,0}},
/* 43604 */ {(12<<2)|2,{110,102,0}},
/* 43605 */ {(12<<2)|2,{110,102,0}},
/* 43606 */ {(12<<2)|2,{110,102,0}},
/* 43607 */ {(12<<2)|2,{110,102,0}},
/* 43608 */ {(12<<2)|2,{110,102,0}},
/* 43609 */ {(12<<2)|2,{110,102,0}},
/* 43610 */ {(12<<2)|2,{110,102,0}},
/* 43611 */ {(12<<2)|2,{110,102,0}},
/* 43612 */ {(12<<2)|2,{110,102,0}},
/* 43613 */ {(12<<2)|2,{110,102,0}},
/* 43614 */ {(12<<2)|2,{110,102,0}},
/* 43615 */ {(12<<2)|2,{110,102,0}},
/* 43616 */ {(12<<2)|2,{110,103,0}},
/* 43617 */ {(12<<2)|2,{110,103,0}},
/* 43618 */ {(12<<2)|2,{110,103,0}},
/* 43619 */ {(12<<2)|2,{110,103,0}},
/* 43620 */ {(12<<2)|2,{110,103,0}},
/* 43621 */ {(12<<2)|2,{110,103,0}},
/* 43622 */ {(12<<2)|2,{110,103,0}},
/* 43623 */ {(12<<2)|2,{110,103,0}},
/* 43624 */ {(12<<2)|2,{110,103,0}},
/* 43625 */ {(12<<2)|2,{110,103,0}},
/* 43626 */ {(12<<2)|2,{110,103,0}},
/* 43627 */ {(12<<2)|2,{110,103,0}},
/* 43628 */ {(12<<2)|2,{110,103,0}},
/* 43629 */ {(12<<2)|2,{110,103,0}},
/* 43630 */ {(12<<2)|2,{110,103,0}},
/* 43631 */ {(12<<2)|2,{110,103,0}},
/* 43632 */ {(12<<2)|2,{110,104,0}},
/* 43633 */ {(12<<2)|2,{110,104,0}},
/* 43634 */ {(12<<2)|2,{110,104,0}},
/* 43635 */ {(12<<2)|2,{110,104,0}},
/* 43636 */ {(12<<2)|2,{110,104,0}},
/* 43637 */ {(12<<2)|2,{110,104,0}},
/* 43638 */ {(12<<2)|2,{110,104,0}},
/* 43639 */ {(12<<2)|2,{110,104,0}},
/* 43640 */ {(12<<2)|2,{110,104,0}},
/* 43641 */ {(12<<2)|2,{110,104,0}},
/* 43642 */ {(12<<2)|2,{110,104,0}},
/* 43643 */ {(12<<2)|2,{110,104,0}},
/* 43644 */ {(12<<2)|2,{110,104,0}},
/* 43645 */ {(12<<2)|2,{110,104,0}},
/* 43646 */ {(12<<2)|2,{110,104,0}},
/* 43647 */ {(12<<2)|2,{110,104,0}},
/* 43648 */ {(12<<2)|2,{110,108,0}},
/* 43649 */ {(12<<2)|2,{110,108,0}},
/* 43650 */ {(12<<2)|2,{110,108,0}},
/* 43651 */ {(12<<2)|2,{110,108,0}},
/* 43652 */ {(12<<2)|2,{110,108,0}},
/* 43653 */ {(12<<2)|2,{110,108,0}},
/* 43654 */ {(12<<2)|2,{110,108,0}},
/* 43655 */ {(12<<2)|2,{110,108,0}},
/* 43656 */ {(12<<2)|2,{110,108,0}},
/* 43657 */ {(12<<2)|2,{110,108,0}},
/* 43658 */ {(12<<2)|2,{110,108,0}},
/* 43659 */ {(12<<2)|2,{110,108,0}},
/* 43660 */ {(12<<2)|2,{110,108,0}},
/* 43661 */ {(12<<2)|2,{110,108,0}},
/* 43662 */ {(12<<2)|2,{110,108,0}},
/* 43663 */ {(12<<2)|2,{110,108,0}},
/* 43664 */ {(12<<2)|2,{110,109,0}},
/* 43665 */ {(12<<2)|2,{110,109,0}},
/* 43666 */ {(12<<2)|2,{110,109,0}},
/* 43667 */ {(12<<2)|2,{110,109,0}},
/* 43668 */ {(12<<2)|2,{110,109,0}},
/* 43669 */ {(12<<2)|2,{110,109,0}},
/* 43670 */ {(12<<2)|2,{110,109,0}},
/* 43671 */ {(12<<2)|2,{110,109,0}},
/* 43672 */ {(12<<2)|2,{110,109,0}},
/* 43673 */ {(12<<2)|2,{110,109,0}},
/* 43674 */ {(12<<2)|2,{110,109,0}},
/* 43675 */ {(12<<2)|2,{110,109,0}},
/* 43676 */ {(12<<2)|2,{110,109,0}},
/* 43677 */ {(12<<2)|2,{110,109,0}},
/* 43678 */ {(12<<2)|2,{110,109,0}},
/* 43679 */ {(12<<2)|2,{110,109,0}},
/* 43680 */ {(12<<2)|2,{110,110,0}},
/* 43681 */ {(12<<2)|2,{110,110,0}},
/* 43682 */ {(12<<2)|2,{110,110,0}},
/* 43683 */ {(12<<2)|2,{110,110,0}},
/* 43684 */ {(12<<2)|2,{110,110,0}},
/* 43685 */ {(12<<2)|2,{110,110,0}},
/* 43686 */ {(12<<2)|2,{110,110,0}},
/* 43687 */ {(12<<2)|2,{110,110,0}},
/* 43688 */ {(12<<2)|2,{110,110,0}},
/* 43689 */ {(12<<2)|2,{110,110,0}},
/* 43690 */ {(12<<2)|2,{110,110,0}},
/* 43691 */ {(12<<2)|2,{110,110,0}},
/* 43692 */ {(12<<2)|2,{110,110,0}},
/* 43693 */ {(12<<2)|2,{110,110,0}},
/* 43694 */ {(12<<2)|2,{110,110,0}},
/* 43695 */ {(12<<2)|2,{110,110,0}},
/* 43696 */ {(12<<2)|2,{110,112,0}},
/* 43697 */ {(12<<2)|2,{110,112,0}},
/* 43698 */ {(12<<2)|2,{110,112,0}},
/* 43699 */ {(12<<2)|2,{110,112,0}},
/* 43700 */ {(12<<2)|2,{110,112,0}},
/* 43701 */ {(12<<2)|2,{110,112,0}},
/* 43702 */ {(12<<2)|2,{110,112,0}},
/* 43703 */ {(12<<2)|2,{110,112,0}},
/* 43704 */ {(12<<2)|2,{110,112,0}},
/* 43705 */ {(12<<2)|2,{110,112,0}},
/* 43706 */ {(12<<2)|2,{110,112,0}},
/* 43707 */ {(12<<2)|2,{110,112,0}},
/* 43708 */ {(12<<2)|2,{110,112,0}},
/* 43709 */ {(12<<2)|2,{110,112,0}},
/* 43710 */ {(12<<2)|2,{110,112,0}},
/* 43711 */ {(12<<2)|2,{110,112,0}},
/* 43712 */ {(12<<2)|2,{110,114,0}},
/* 43713 */ {(12<<2)|2,{110,114,0}},
/* 43714 */ {(12<<2)|2,{110,114,0}},
/* 43715 */ {(12<<2)|2,{110,114,0}},
/* 43716 */ {(12<<2)|2,{110,114,0}},
/* 43717 */ {(12<<2)|2,{110,114,0}},
/* 43718 */ {(12<<2)|2,{110,114,0}},
/* 43719 */ {(12<<2)|2,{110,114,0}},
/* 43720 */ {(12<<2)|2,{110,114,0}},
/* 43721 */ {(12<<2)|2,{110,114,0}},
/* 43722 */ {(12<<2)|2,{110,114,0}},
/* 43723 */ {(12<<2)|2,{110,114,0}},
/* 43724 */ {(12<<2)|2,{110,114,0}},
/* 43725 */ {(12<<2)|2,{110,114,0}},
/* 43726 */ {(12<<2)|2,{110,114,0}},
/* 43727 */ {(12<<2)|2,{110,114,0}},
/* 43728 */ {(12<<2)|2,{110,117,0}},
/* 43729 */ {(12<<2)|2,{110,117,0}},
/* 43730 */ {(12<<2)|2,{110,117,0}},
/* 43731 */ {(12<<2)|2,{110,117,0}},
/* 43732 */ {(12<<2)|2,{110,117,0}},
/* 43733 */ {(12<<2)|2,{110,117,0}},
/* 43734 */ {(12<<2)|2,{110,117,0}},
/* 43735 */ {(12<<2)|2,{110,117,0}},
/* 43736 */ {(12<<2)|2,{110,117,0}},
/* 43737 */ {(12<<2)|2,{110,117,0}},
/* 43738 */ {(12<<2)|2,{110,117,0}},
/* 43739 */ {(12<<2)|2,{110,117,0}},
/* 43740 */ {(12<<2)|2,{110,117,0}},
/* 43741 */ {(12<<2)|2,{110,117,0}},
/* 43742 */ {(12<<2)|2,{110,117,0}},
/* 43743 */ {(12<<2)|2,{110,117,0}},
/* 43744 */ {(13<<2)|2,{110,58,0}},
/* 43745 */ {(13<<2)|2,{110,58,0}},
/* 43746 */ {(13<<2)|2,{110,58,0}},
/* 43747 */ {(13<<2)|2,{110,58,0}},
/* 43748 */ {(13<<2)|2,{110,58,0}},
/* 43749 */ {(13<<2)|2,{110,58,0}},
/* 43750 */ {(13<<2)|2,{110,58,0}},
/* 43751 */ {(13<<2)|2,{110,58,0}},
/* 43752 */ {(13<<2)|2,{110,66,0}},
/* 43753 */ {(13<<2)|2,{110,66,0}},
/* 43754 */ {(13<<2)|2,{110,66,0}},
/* 43755 */ {(13<<2)|2,{110,66,0}},
/* 43756 */ {(13<<2)|2,{110,66,0}},
/* 43757 */ {(13<<2)|2,{110,66,0}},
/* 43758 */ {(13<<2)|2,{110,66,0}},
/* 43759 */ {(13<<2)|2,{110,66,0}},
/* 43760 */ {(13<<2)|2,{110,67,0}},
/* 43761 */ {(13<<2)|2,{110,67,0}},
/* 43762 */ {(13<<2)|2,{110,67,0}},
/* 43763 */ {(13<<2)|2,{110,67,0}},
/* 43764 */ {(13<<2)|2,{110,67,0}},
/* 43765 */ {(13<<2)|2,{110,67,0}},
/* 43766 */ {(13<<2)|2,{110,67,0}},
/* 43767 */ {(13<<2)|2,{110,67,0}},
/* 43768 */ {(13<<2)|2,{110,68,0}},
/* 43769 */ {(13<<2)|2,{110,68,0}},
/* 43770 */ {(13<<2)|2,{110,68,0}},
/* 43771 */ {(13<<2)|2,{110,68,0}},
/* 43772 */ {(13<<2)|2,{110,68,0}},
/* 43773 */ {(13<<2)|2,{110,68,0}},
/* 43774 */ {(13<<2)|2,{110,68,0}},
/* 43775 */ {(13<<2)|2,{110,68,0}},
/* 43776 */ {(13<<2)|2,{110,69,0}},
/* 43777 */ {(13<<2)|2,{110,69,0}},
/* 43778 */ {(13<<2)|2,{110,69,0}},
/* 43779 */ {(13<<2)|2,{110,69,0}},
/* 43780 */ {(13<<2)|2,{110,69,0}},
/* 43781 */ {(13<<2)|2,{110,69,0}},
/* 43782 */ {(13<<2)|2,{110,69,0}},
/* 43783 */ {(13<<2)|2,{110,69,0}},
/* 43784 */ {(13<<2)|2,{110,70,0}},
/* 43785 */ {(13<<2)|2,{110,70,0}},
/* 43786 */ {(13<<2)|2,{110,70,0}},
/* 43787 */ {(13<<2)|2,{110,70,0}},
/* 43788 */ {(13<<2)|2,{110,70,0}},
/* 43789 */ {(13<<2)|2,{110,70,0}},
/* 43790 */ {(13<<2)|2,{110,70,0}},
/* 43791 */ {(13<<2)|2,{110,70,0}},
/* 43792 */ {(13<<2)|2,{110,71,0}},
/* 43793 */ {(13<<2)|2,{110,71,0}},
/* 43794 */ {(13<<2)|2,{110,71,0}},
/* 43795 */ {(13<<2)|2,{110,71,0}},
/* 43796 */ {(13<<2)|2,{110,71,0}},
/* 43797 */ {(13<<2)|2,{110,71,0}},
/* 43798 */ {(13<<2)|2,{110,71,0}},
/* 43799 */ {(13<<2)|2,{110,71,0}},
/* 43800 */ {(13<<2)|2,{110,72,0}},
/* 43801 */ {(13<<2)|2,{110,72,0}},
/* 43802 */ {(13<<2)|2,{110,72,0}},
/* 43803 */ {(13<<2)|2,{110,72,0}},
/* 43804 */ {(13<<2)|2,{110,72,0}},
/* 43805 */ {(13<<2)|2,{110,72,0}},
/* 43806 */ {(13<<2)|2,{110,72,0}},
/* 43807 */ {(13<<2)|2,{110,72,0}},
/* 43808 */ {(13<<2)|2,{110,73,0}},
/* 43809 */ {(13<<2)|2,{110,73,0}},
/* 43810 */ {(13<<2)|2,{110,73,0}},
/* 43811 */ {(13<<2)|2,{110,73,0}},
/* 43812 */ {(13<<2)|2,{110,73,0}},
/* 43813 */ {(13<<2)|2,{110,73,0}},
/* 43814 */ {(13<<2)|2,{110,73,0}},
/* 43815 */ {(13<<2)|2,{110,73,0}},
/* 43816 */ {(13<<2)|2,{110,74,0}},
/* 43817 */ {(13<<2)|2,{110,74,0}},
/* 43818 */ {(13<<2)|2,{110,74,0}},
/* 43819 */ {(13<<2)|2,{110,74,0}},
/* 43820 */ {(13<<2)|2,{110,74,0}},
/* 43821 */ {(13<<2)|2,{110,74,0}},
/* 43822 */ {(13<<2)|2,{110,74,0}},
/* 43823 */ {(13<<2)|2,{110,74,0}},
/* 43824 */ {(13<<2)|2,{110,75,0}},
/* 43825 */ {(13<<2)|2,{110,75,0}},
/* 43826 */ {(13<<2)|2,{110,75,0}},
/* 43827 */ {(13<<2)|2,{110,75,0}},
/* 43828 */ {(13<<2)|2,{110,75,0}},
/* 43829 */ {(13<<2)|2,{110,75,0}},
/* 43830 */ {(13<<2)|2,{110,75,0}},
/* 43831 */ {(13<<2)|2,{110,75,0}},
/* 43832 */ {(13<<2)|2,{110,76,0}},
/* 43833 */ {(13<<2)|2,{110,76,0}},
/* 43834 */ {(13<<2)|2,{110,76,0}},
/* 43835 */ {(13<<2)|2,{110,76,0}},
/* 43836 */ {(13<<2)|2,{110,76,0}},
/* 43837 */ {(13<<2)|2,{110,76,0}},
/* 43838 */ {(13<<2)|2,{110,76,0}},
/* 43839 */ {(13<<2)|2,{110,76,0}},
/* 43840 */ {(13<<2)|2,{110,77,0}},
/* 43841 */ {(13<<2)|2,{110,77,0}},
/* 43842 */ {(13<<2)|2,{110,77,0}},
/* 43843 */ {(13<<2)|2,{110,77,0}},
/* 43844 */ {(13<<2)|2,{110,77,0}},
/* 43845 */ {(13<<2)|2,{110,77,0}},
/* 43846 */ {(13<<2)|2,{110,77,0}},
/* 43847 */ {(13<<2)|2,{110,77,0}},
/* 43848 */ {(13<<2)|2,{110,78,0}},
/* 43849 */ {(13<<2)|2,{110,78,0}},
/* 43850 */ {(13<<2)|2,{110,78,0}},
/* 43851 */ {(13<<2)|2,{110,78,0}},
/* 43852 */ {(13<<2)|2,{110,78,0}},
/* 43853 */ {(13<<2)|2,{110,78,0}},
/* 43854 */ {(13<<2)|2,{110,78,0}},
/* 43855 */ {(13<<2)|2,{110,78,0}},
/* 43856 */ {(13<<2)|2,{110,79,0}},
/* 43857 */ {(13<<2)|2,{110,79,0}},
/* 43858 */ {(13<<2)|2,{110,79,0}},
/* 43859 */ {(13<<2)|2,{110,79,0}},
/* 43860 */ {(13<<2)|2,{110,79,0}},
/* 43861 */ {(13<<2)|2,{110,79,0}},
/* 43862 */ {(13<<2)|2,{110,79,0}},
/* 43863 */ {(13<<2)|2,{110,79,0}},
/* 43864 */ {(13<<2)|2,{110,80,0}},
/* 43865 */ {(13<<2)|2,{110,80,0}},
/* 43866 */ {(13<<2)|2,{110,80,0}},
/* 43867 */ {(13<<2)|2,{110,80,0}},
/* 43868 */ {(13<<2)|2,{110,80,0}},
/* 43869 */ {(13<<2)|2,{110,80,0}},
/* 43870 */ {(13<<2)|2,{110,80,0}},
/* 43871 */ {(13<<2)|2,{110,80,0}},
/* 43872 */ {(13<<2)|2,{110,81,0}},
/* 43873 */ {(13<<2)|2,{110,81,0}},
/* 43874 */ {(13<<2)|2,{110,81,0}},
/* 43875 */ {(13<<2)|2,{110,81,0}},
/* 43876 */ {(13<<2)|2,{110,81,0}},
/* 43877 */ {(13<<2)|2,{110,81,0}},
/* 43878 */ {(13<<2)|2,{110,81,0}},
/* 43879 */ {(13<<2)|2,{110,81,0}},
/* 43880 */ {(13<<2)|2,{110,82,0}},
/* 43881 */ {(13<<2)|2,{110,82,0}},
/* 43882 */ {(13<<2)|2,{110,82,0}},
/* 43883 */ {(13<<2)|2,{110,82,0}},
/* 43884 */ {(13<<2)|2,{110,82,0}},
/* 43885 */ {(13<<2)|2,{110,82,0}},
/* 43886 */ {(13<<2)|2,{110,82,0}},
/* 43887 */ {(13<<2)|2,{110,82,0}},
/* 43888 */ {(13<<2)|2,{110,83,0}},
/* 43889 */ {(13<<2)|2,{110,83,0}},
/* 43890 */ {(13<<2)|2,{110,83,0}},
/* 43891 */ {(13<<2)|2,{110,83,0}},
/* 43892 */ {(13<<2)|2,{110,83,0}},
/* 43893 */ {(13<<2)|2,{110,83,0}},
/* 43894 */ {(13<<2)|2,{110,83,0}},
/* 43895 */ {(13<<2)|2,{110,83,0}},
/* 43896 */ {(13<<2)|2,{110,84,0}},
/* 43897 */ {(13<<2)|2,{110,84,0}},
/* 43898 */ {(13<<2)|2,{110,84,0}},
/* 43899 */ {(13<<2)|2,{110,84,0}},
/* 43900 */ {(13<<2)|2,{110,84,0}},
/* 43901 */ {(13<<2)|2,{110,84,0}},
/* 43902 */ {(13<<2)|2,{110,84,0}},
/* 43903 */ {(13<<2)|2,{110,84,0}},
/* 43904 */ {(13<<2)|2,{110,85,0}},
/* 43905 */ {(13<<2)|2,{110,85,0}},
/* 43906 */ {(13<<2)|2,{110,85,0}},
/* 43907 */ {(13<<2)|2,{110,85,0}},
/* 43908 */ {(13<<2)|2,{110,85,0}},
/* 43909 */ {(13<<2)|2,{110,85,0}},
/* 43910 */ {(13<<2)|2,{110,85,0}},
/* 43911 */ {(13<<2)|2,{110,85,0}},
/* 43912 */ {(13<<2)|2,{110,86,0}},
/* 43913 */ {(13<<2)|2,{110,86,0}},
/* 43914 */ {(13<<2)|2,{110,86,0}},
/* 43915 */ {(13<<2)|2,{110,86,0}},
/* 43916 */ {(13<<2)|2,{110,86,0}},
/* 43917 */ {(13<<2)|2,{110,86,0}},
/* 43918 */ {(13<<2)|2,{110,86,0}},
/* 43919 */ {(13<<2)|2,{110,86,0}},
/* 43920 */ {(13<<2)|2,{110,87,0}},
/* 43921 */ {(13<<2)|2,{110,87,0}},
/* 43922 */ {(13<<2)|2,{110,87,0}},
/* 43923 */ {(13<<2)|2,{110,87,0}},
/* 43924 */ {(13<<2)|2,{110,87,0}},
/* 43925 */ {(13<<2)|2,{110,87,0}},
/* 43926 */ {(13<<2)|2,{110,87,0}},
/* 43927 */ {(13<<2)|2,{110,87,0}},
/* 43928 */ {(13<<2)|2,{110,89,0}},
/* 43929 */ {(13<<2)|2,{110,89,0}},
/* 43930 */ {(13<<2)|2,{110,89,0}},
/* 43931 */ {(13<<2)|2,{110,89,0}},
/* 43932 */ {(13<<2)|2,{110,89,0}},
/* 43933 */ {(13<<2)|2,{110,89,0}},
/* 43934 */ {(13<<2)|2,{110,89,0}},
/* 43935 */ {(13<<2)|2,{110,89,0}},
/* 43936 */ {(13<<2)|2,{110,106,0}},
/* 43937 */ {(13<<2)|2,{110,106,0}},
/* 43938 */ {(13<<2)|2,{110,106,0}},
/* 43939 */ {(13<<2)|2,{110,106,0}},
/* 43940 */ {(13<<2)|2,{110,106,0}},
/* 43941 */ {(13<<2)|2,{110,106,0}},
/* 43942 */ {(13<<2)|2,{110,106,0}},
/* 43943 */ {(13<<2)|2,{110,106,0}},
/* 43944 */ {(13<<2)|2,{110,107,0}},
/* 43945 */ {(13<<2)|2,{110,107,0}},
/* 43946 */ {(13<<2)|2,{110,107,0}},
/* 43947 */ {(13<<2)|2,{110,107,0}},
/* 43948 */ {(13<<2)|2,{110,107,0}},
/* 43949 */ {(13<<2)|2,{110,107,0}},
/* 43950 */ {(13<<2)|2,{110,107,0}},
/* 43951 */ {(13<<2)|2,{110,107,0}},
/* 43952 */ {(13<<2)|2,{110,113,0}},
/* 43953 */ {(13<<2)|2,{110,113,0}},
/* 43954 */ {(13<<2)|2,{110,113,0}},
/* 43955 */ {(13<<2)|2,{110,113,0}},
/* 43956 */ {(13<<2)|2,{110,113,0}},
/* 43957 */ {(13<<2)|2,{110,113,0}},
/* 43958 */ {(13<<2)|2,{110,113,0}},
/* 43959 */ {(13<<2)|2,{110,113,0}},
/* 43960 */ {(13<<2)|2,{110,118,0}},
/* 43961 */ {(13<<2)|2,{110,118,0}},
/* 43962 */ {(13<<2)|2,{110,118,0}},
/* 43963 */ {(13<<2)|2,{110,118,0}},
/* 43964 */ {(13<<2)|2,{110,118,0}},
/* 43965 */ {(13<<2)|2,{110,118,0}},
/* 43966 */ {(13<<2)|2,{110,118,0}},
/* 43967 */ {(13<<2)|2,{110,118,0}},
/* 43968 */ {(13<<2)|2,{110,119,0}},
/* 43969 */ {(13<<2)|2,{110,119,0}},
/* 43970 */ {(13<<2)|2,{110,119,0}},
/* 43971 */ {(13<<2)|2,{110,119,0}},
/* 43972 */ {(13<<2)|2,{110,119,0}},
/* 43973 */ {(13<<2)|2,{110,119,0}},
/* 43974 */ {(13<<2)|2,{110,119,0}},
/* 43975 */ {(13<<2)|2,{110,119,0}},
/* 43976 */ {(13<<2)|2,{110,120,0}},
/* 43977 */ {(13<<2)|2,{110,120,0}},
/* 43978 */ {(13<<2)|2,{110,120,0}},
/* 43979 */ {(13<<2)|2,{110,120,0}},
/* 43980 */ {(13<<2)|2,{110,120,0}},
/* 43981 */ {(13<<2)|2,{110,120,0}},
/* 43982 */ {(13<<2)|2,{110,120,0}},
/* 43983 */ {(13<<2)|2,{110,120,0}},
/* 43984 */ {(13<<2)|2,{110,121,0}},
/* 43985 */ {(13<<2)|2,{110,121,0}},
/* 43986 */ {(13<<2)|2,{110,121,0}},
/* 43987 */ {(13<<2)|2,{110,121,0}},
/* 43988 */ {(13<<2)|2,{110,121,0}},
/* 43989 */ {(13<<2)|2,{110,121,0}},
/* 43990 */ {(13<<2)|2,{110,121,0}},
/* 43991 */ {(13<<2)|2,{110,121,0}},
/* 43992 */ {(13<<2)|2,{110,122,0}},
/* 43993 */ {(13<<2)|2,{110,122,0}},
/* 43994 */ {(13<<2)|2,{110,122,0}},
/* 43995 */ {(13<<2)|2,{110,122,0}},
/* 43996 */ {(13<<2)|2,{110,122,0}},
/* 43997 */ {(13<<2)|2,{110,122,0}},
/* 43998 */ {(13<<2)|2,{110,122,0}},
/* 43999 */ {(13<<2)|2,{110,122,0}},
/* 44000 */ {(14<<2)|2,{110,38,0}},
/* 44001 */ {(14<<2)|2,{110,38,0}},
/* 44002 */ {(14<<2)|2,{110,38,0}},
/* 44003 */ {(14<<2)|2,{110,38,0}},
/* 44004 */ {(14<<2)|2,{110,42,0}},
/* 44005 */ {(14<<2)|2,{110,42,0}},
/* 44006 */ {(14<<2)|2,{110,42,0}},
/* 44007 */ {(14<<2)|2,{110,42,0}},
/* 44008 */ {(14<<2)|2,{110,44,0}},
/* 44009 */ {(14<<2)|2,{110,44,0}},
/* 44010 */ {(14<<2)|2,{110,44,0}},
/* 44011 */ {(14<<2)|2,{110,44,0}},
/* 44012 */ {(14<<2)|2,{110,59,0}},
/* 44013 */ {(14<<2)|2,{110,59,0}},
/* 44014 */ {(14<<2)|2,{110,59,0}},
/* 44015 */ {(14<<2)|2,{110,59,0}},
/* 44016 */ {(14<<2)|2,{110,88,0}},
/* 44017 */ {(14<<2)|2,{110,88,0}},
/* 44018 */ {(14<<2)|2,{110,88,0}},
/* 44019 */ {(14<<2)|2,{110,88,0}},
/* 44020 */ {(14<<2)|2,{110,90,0}},
/* 44021 */ {(14<<2)|2,{110,90,0}},
/* 44022 */ {(14<<2)|2,{110,90,0}},
/* 44023 */ {(14<<2)|2,{110,90,0}},
/* 44024 */ {(16<<2)|2,{110,33,0}},
/* 44025 */ {(16<<2)|2,{110,34,0}},
/* 44026 */ {(16<<2)|2,{110,40,0}},
/* 44027 */ {(16<<2)|2,{110,41,0}},
/* 44028 */ {(16<<2)|2,{110,63,0}},
/* 44029 */ {(6<<2)|1,{110,0,0}},
/* 44030 */ {(6<<2)|1,{110,0,0}},
/* 44031 */ {(6<<2)|1,{110,0,0}},
/* 44032 */ {(16<<2)|3,{112,48,48}},
/* 44033 */ {(16<<2)|3,{112,48,49}},
/* 44034 */ {(16<<2)|3,{112,48,50}},
/* 44035 */ {(16<<2)|3,{112,48,97}},
/* 44036 */ {(16<<2)|3,{112,48,99}},
/* 44037 */ {(16<<2)|3,{112,48,101}},
/* 44038 */ {(16<<2)|3,{112,48,105}},
/* 44039 */ {(16<<2)|3,{112,48,111}},
/* 44040 */ {(16<<2)|3,{112,48,115}},
/* 44041 */ {(16<<2)|3,{112,48,116}},
/* 44042 */ {(11<<2)|2,{112,48,0}},
/* 44043 */ {(11<<2)|2,{112,48,0}},
/* 44044 */ {(11<<2)|2,{112,48,0}},
/* 44045 */ {(11<<2)|2,{112,48,0}},
/* 44046 */ {(11<<2)|2,{112,48,0}},
/* 44047 */ {(11<<2)|2,{112,48,0}},
/* 44048 */ {(11<<2)|2,{112,48,0}},
/* 44049 */ {(11<<2)|2,{112,48,0}},
/* 44050 */ {(11<<2)|2,{112,48,0}},
/* 44051 */ {(11<<2)|2,{112,48,0}},
/* 44052 */ {(11<<2)|2,{112,48,0}},
/* 44053 */ {(11<<2)|2,{112,48,0}},
/* 44054 */ {(11<<2)|2,{112,48,0}},
/* 44055 */ {(11<<2)|2,{112,48,0}},
/* 44056 */ {(11<<2)|2,{112,48,0}},
/* 44057 */ {(11<<2)|2,{112,48,0}},
/* 44058 */ {(11<<2)|2,{112,48,0}},
/* 44059 */ {(11<<2)|2,{112,48,0}},
/* 44060 */ {(11<<2)|2,{112,48,0}},
/* 44061 */ {(11<<2)|2,{112,48,0}},
/* 44062 */ {(11<<2)|2,{112,48,0}},
/* 44063 */ {(11<<2)|2,{112,48,0}},
/* 44064 */ {(16<<2)|3,{112,49,48}},
/* 44065 */ {(16<<2)|3,{112,49,49}},
/* 44066 */ {(16<<2)|3,{112,49,50}},
/* 44067 */ {(16<<2)|3,{112,49,97}},
/* 44068 */ {(16<<2)|3,{112,49,99}},
/* 44069 */ {(16<<2)|3,{112,49,101}},
/* 44070 */ {(16<<2)|3,{112,49,105}},
/* 44071 */ {(16<<2)|3,{112,49,111}},
/* 44072 */ {(16<<2)|3,{112,49,115}},
/* 44073 */ {(16<<2)|3,{112,49,116}},
/* 44074 */ {(11<<2)|2,{112,49,0}},
/* 44075 */ {(11<<2)|2,{112,49,0}},
/* 44076 */ {(11<<2)|2,{112,49,0}},
/* 44077 */ {(11<<2)|2,{112,49,0}},
/* 44078 */ {(11<<2)|2,{112,49,0}},
/* 44079 */ {(11<<2)|2,{112,49,0}},
/* 44080 */ {(11<<2)|2,{112,49,0}},
/* 44081 */ {(11<<2)|2,{112,49,0}},
/* 44082 */ {(11<<2)|2,{112,49,0}},
/* 44083 */ {(11<<2)|2,{112,49,0}},
/* 44084 */ {(11<<2)|2,{112,49,0}},
/* 44085 */ {(11<<2)|2,{112,49,0}},
/* 44086 */ {(11<<2)|2,{112,49,0}},
/* 44087 */ {(11<<2)|2,{112,49,0}},
/* 44088 */ {(11<<2)|2,{112,49,0}},
/* 44089 */ {(11<<2)|2,{112,49,0}},
/* 44090 */ {(11<<2)|2,{112,49,0}},
/* 44091 */ {(11<<2)|2,{112,49,0}},
/* 44092 */ {(11<<2)|2,{112,49,0}},
/* 44093 */ {(11<<2)|2,{112,49,0}},
/* 44094 */ {(11<<2)|2,{112,49,0}},
/* 44095 */ {(11<<2)|2,{112,49,0}},
/* 44096 */ {(16<<2)|3,{112,50,48}},
/* 44097 */ {(16<<2)|3,{112,50,49}},
/* 44098 */ {(16<<2)|3,{112,50,50}},
/* 44099 */ {(16<<2)|3,{112,50,97}},
/* 44100 */ {(16<<2)|3,{112,50,99}},
/* 44101 */ {(16<<2)|3,{112,50,101}},
/* 44102 */ {(16<<2)|3,{112,50,105}},
/* 44103 */ {(16<<2)|3,{112,50,111}},
/* 44104 */ {(16<<2)|3,{112,50,115}},
/* 44105 */ {(16<<2)|3,{112,50,116}},
/* 44106 */ {(11<<2)|2,{112,50,0}},
/* 44107 */ {(11<<2)|2,{112,50,0}},
/* 44108 */ {(11<<2)|2,{112,50,0}},
/* 44109 */ {(11<<2)|2,{112,50,0}},
/* 44110 */ {(11<<2)|2,{112,50,0}},
/* 44111 */ {(11<<2)|2,{112,50,0}},
/* 44112 */ {(11<<2)|2,{112,50,0}},
/* 44113 */ {(11<<2)|2,{112,50,0}},
/* 44114 */ {(11<<2)|2,{112,50,0}},
/* 44115 */ {(11<<2)|2,{112,50,0}},
/* 44116 */ {(11<<2)|2,{112,50,0}},
/* 44117 */ {(11<<2)|2,{112,50,0}},
/* 44118 */ {(11<<2)|2,{112,50,0}},
/* 44119 */ {(11<<2)|2,{112,50,0}},
/* 44120 */ {(11<<2)|2,{112,50,0}},
/* 44121 */ {(11<<2)|2,{112,50,0}},
/* 44122 */ {(11<<2)|2,{112,50,0}},
/* 44123 */ {(11<<2)|2,{112,50,0}},
/* 44124 */ {(11<<2)|2,{112,50,0}},
/* 44125 */ {(11<<2)|2,{112,50,0}},
/* 44126 */ {(11<<2)|2,{112,50,0}},
/* 44127 */ {(11<<2)|2,{112,50,0}},
/* 44128 */ {(16<<2)|3,{112,97,48}},
/* 44129 */ {(16<<2)|3,{112,97,49}},
/* 44130 */ {(16<<2)|3,{112,97,50}},
/* 44131 */ {(16<<2)|3,{112,97,97}},
/* 44132 */ {(16<<2)|3,{112,97,99}},
/* 44133 */ {(16<<2)|3,{112,97,101}},
/* 44134 */ {(16<<2)|3,{112,97,105}},
/* 44135 */ {(16<<2)|3,{112,97,111}},
/* 44136 */ {(16<<2)|3,{112,97,115}},
/* 44137 */ {(16<<2)|3,{112,97,116}},
/* 44138 */ {(11<<2)|2,{112,97,0}},
/* 44139 */ {(11<<2)|2,{112,97,0}},
/* 44140 */ {(11<<2)|2,{112,97,0}},
/* 44141 */ {(11<<2)|2,{112,97,0}},
/* 44142 */ {(11<<2)|2,{112,97,0}},
/* 44143 */ {(11<<2)|2,{112,97,0}},
/* 44144 */ {(11<<2)|2,{112,97,0}},
/* 44145 */ {(11<<2)|2,{112,97,0}},
/* 44146 */ {(11<<2)|2,{112,97,0}},
/* 44147 */ {(11<<2)|2,{112,97,0}},
/* 44148 */ {(11<<2)|2,{112,97,0}},
/* 44149 */ {(11<<2)|2,{112,97,0}},
/* 44150 */ {(11<<2)|2,{112,97,0}},
/* 44151 */ {(11<<2)|2,{112,97,0}},
/* 44152 */ {(11<<2)|2,{112,97,0}},
/* 44153 */ {(11<<2)|2,{112,97,0}},
/* 44154 */ {(11<<2)|2,{112,97,0}},
/* 44155 */ {(11<<2)|2,{112,97,0}},
/* 44156 */ {(11<<2)|2,{112,97,0}},
/* 44157 */ {(11<<2)|2,{112,97,0}},
/* 44158 */ {(11<<2)|2,{112,97,0}},
/* 44159 */ {(11<<2)|2,{112,97,0}},
/* 44160 */ {(16<<2)|3,{112,99,48}},
/* 44161 */ {(16<<2)|3,{112,99,49}},
/* 44162 */ {(16<<2)|3,{112,99,50}},
/* 44163 */ {(16<<2)|3,{112,99,97}},
/* 44164 */ {(16<<2)|3,{112,99,99}},
/* 44165 */ {(16<<2)|3,{112,99,101}},
/* 44166 */ {(16<<2)|3,{112,99,105}},
/* 44167 */ {(16<<2)|3,{112,99,111}},
/* 44168 */ {(16<<2)|3,{112,99,115}},
/* 44169 */ {(16<<2)|3,{112,99,116}},
/* 44170 */ {(11<<2)|2,{112,99,0}},
/* 44171 */ {(11<<2)|2,{112,99,0}},
/* 44172 */ {(11<<2)|2,{112,99,0}},
/* 44173 */ {(11<<2)|2,{112,99,0}},
/* 44174 */ {(11<<2)|2,{112,99,0}},
/* 44175 */ {(11<<2)|2,{112,99,0}},
/* 44176 */ {(11<<2)|2,{112,99,0}},
/* 44177 */ {(11<<2)|2,{112,99,0}},
/* 44178 */ {(11<<2)|2,{112,99,0}},
/* 44179 */ {(11<<2)|2,{112,99,0}},
/* 44180 */ {(11<<2)|2,{112,99,0}},
/* 44181 */ {(11<<2)|2,{112,99,0}},
/* 44182 */ {(11<<2)|2,{112,99,0}},
/* 44183 */ {(11<<2)|2,{112,99,0}},
/* 44184 */ {(11<<2)|2,{112,99,0}},
/* 44185 */ {(11<<2)|2,{112,99,0}},
/* 44186 */ {(11<<2)|2,{112,99,0}},
/* 44187 */ {(11<<2)|2,{112,99,0}},
/* 44188 */ {(11<<2)|2,{112,99,0}},
/* 44189 */ {(11<<2)|2,{112,99,0}},
/* 44190 */ {(11<<2)|2,{112,99,0}},
/* 44191 */ {(11<<2)|2,{112,99,0}},
/* 44192 */ {(16<<2)|3,{112,101,48}},
/* 44193 */ {(16<<2)|3,{112,101,49}},
/* 44194 */ {(16<<2)|3,{112,101,50}},
/* 44195 */ {(16<<2)|3,{112,101,97}},
/* 44196 */ {(16<<2)|3,{112,101,99}},
/* 44197 */ {(16<<2)|3,{112,101,101}},
/* 44198 */ {(16<<2)|3,{112,101,105}},
/* 44199 */ {(16<<2)|3,{112,101,111}},
/* 44200 */ {(16<<2)|3,{112,101,115}},
/* 44201 */ {(16<<2)|3,{112,101,116}},
/* 44202 */ {(11<<2)|2,{112,101,0}},
/* 44203 */ {(11<<2)|2,{112,101,0}},
/* 44204 */ {(11<<2)|2,{112,101,0}},
/* 44205 */ {(11<<2)|2,{112,101,0}},
/* 44206 */ {(11<<2)|2,{112,101,0}},
/* 44207 */ {(11<<2)|2,{112,101,0}},
/* 44208 */ {(11<<2)|2,{112,101,0}},
/* 44209 */ {(11<<2)|2,{112,101,0}},
/* 44210 */ {(11<<2)|2,{112,101,0}},
/* 44211 */ {(11<<2)|2,{112,101,0}},
/* 44212 */ {(11<<2)|2,{112,101,0}},
/* 44213 */ {(11<<2)|2,{112,101,0}},
/* 44214 */ {(11<<2)|2,{112,101,0}},
/* 44215 */ {(11<<2)|2,{112,101,0}},
/* 44216 */ {(11<<2)|2,{112,101,0}},
/* 44217 */ {(11<<2)|2,{112,101,0}},
/* 44218 */ {(11<<2)|2,{112,101,0}},
/* 44219 */ {(11<<2)|2,{112,101,0}},
/* 44220 */ {(11<<2)|2,{112,101,0}},
/* 44221 */ {(11<<2)|2,{112,101,0}},
/* 44222 */ {(11<<2)|2,{112,101,0}},
/* 44223 */ {(11<<2)|2,{112,101,0}},
/* 44224 */ {(16<<2)|3,{112,105,48}},
/* 44225 */ {(16<<2)|3,{112,105,49}},
/* 44226 */ {(16<<2)|3,{112,105,50}},
/* 44227 */ {(16<<2)|3,{112,105,97}},
/* 44228 */ {(16<<2)|3,{112,105,99}},
/* 44229 */ {(16<<2)|3,{112,105,101}},
/* 44230 */ {(16<<2)|3,{112,105,105}},
/* 44231 */ {(16<<2)|3,{112,105,111}},
/* 44232 */ {(16<<2)|3,{112,105,115}},
/* 44233 */ {(16<<2)|3,{112,105,116}},
/* 44234 */ {(11<<2)|2,{112,105,0}},
/* 44235 */ {(11<<2)|2,{112,105,0}},
/* 44236 */ {(11<<2)|2,{112,105,0}},
/* 44237 */ {(11<<2)|2,{112,105,0}},
/* 44238 */ {(11<<2)|2,{112,105,0}},
/* 44239 */ {(11<<2)|2,{112,105,0}},
/* 44240 */ {(11<<2)|2,{112,105,0}},
/* 44241 */ {(11<<2)|2,{112,105,0}},
/* 44242 */ {(11<<2)|2,{112,105,0}},
/* 44243 */ {(11<<2)|2,{112,105,0}},
/* 44244 */ {(11<<2)|2,{112,105,0}},
/* 44245 */ {(11<<2)|2,{112,105,0}},
/* 44246 */ {(11<<2)|2,{112,105,0}},
/* 44247 */ {(11<<2)|2,{112,105,0}},
/* 44248 */ {(11<<2)|2,{112,105,0}},
/* 44249 */ {(11<<2)|2,{112,105,0}},
/* 44250 */ {(11<<2)|2,{112,105,0}},
/* 44251 */ {(11<<2)|2,{112,105,0}},
/* 44252 */ {(11<<2)|2,{112,105,0}},
/* 44253 */ {(11<<2)|2,{112,105,0}},
/* 44254 */ {(11<<2)|2,{112,105,0}},
/* 44255 */ {(11<<2)|2,{112,105,0}},
/* 44256 */ {(16<<2)|3,{112,111,48}},
/* 44257 */ {(16<<2)|3,{112,111,49}},
/* 44258 */ {(16<<2)|3,{112,111,50}},
/* 44259 */ {(16<<2)|3,{112,111,97}},
/* 44260 */ {(16<<2)|3,{112,111,99}},
/* 44261 */ {(16<<2)|3,{112,111,101}},
/* 44262 */ {(16<<2)|3,{112,111,105}},
/* 44263 */ {(16<<2)|3,{112,111,111}},
/* 44264 */ {(16<<2)|3,{112,111,115}},
/* 44265 */ {(16<<2)|3,{112,111,116}},
/* 44266 */ {(11<<2)|2,{112,111,0}},
/* 44267 */ {(11<<2)|2,{112,111,0}},
/* 44268 */ {(11<<2)|2,{112,111,0}},
/* 44269 */ {(11<<2)|2,{112,111,0}},
/* 44270 */ {(11<<2)|2,{112,111,0}},
/* 44271 */ {(11<<2)|2,{112,111,0}},
/* 44272 */ {(11<<2)|2,{112,111,0}},
/* 44273 */ {(11<<2)|2,{112,111,0}},
/* 44274 */ {(11<<2)|2,{112,111,0}},
/* 44275 */ {(11<<2)|2,{112,111,0}},
/* 44276 */ {(11<<2)|2,{112,111,0}},
/* 44277 */ {(11<<2)|2,{112,111,0}},
/* 44278 */ {(11<<2)|2,{112,111,0}},
/* 44279 */ {(11<<2)|2,{112,111,0}},
/* 44280 */ {(11<<2)|2,{112,111,0}},
/* 44281 */ {(11<<2)|2,{112,111,0}},
/* 44282 */ {(11<<2)|2,{112,111,0}},
/* 44283 */ {(11<<2)|2,{112,111,0}},
/* 44284 */ {(11<<2)|2,{112,111,0}},
/* 44285 */ {(11<<2)|2,{112,111,0}},
/* 44286 */ {(11<<2)|2,{112,111,0}},
/* 44287 */ {(11<<2)|2,{112,111,0}},
/* 44288 */ {(16<<2)|3,{112,115,48}},
/* 44289 */ {(16<<2)|3,{112,115,49}},
/* 44290 */ {(16<<2)|3,{112,115,50}},
/* 44291 */ {(16<<2)|3,{112,115,97}},
/* 44292 */ {(16<<2)|3,{112,115,99}},
/* 44293 */ {(16<<2)|3,{112,115,101}},
/* 44294 */ {(16<<2)|3,{112,115,105}},
/* 44295 */ {(16<<2)|3,{112,115,111}},
/* 44296 */ {(16<<2)|3,{112,115,115}},
/* 44297 */ {(16<<2)|3,{112,115,116}},
/* 44298 */ {(11<<2)|2,{112,115,0}},
/* 44299 */ {(11<<2)|2,{112,115,0}},
/* 44300 */ {(11<<2)|2,{112,115,0}},
/* 44301 */ {(11<<2)|2,{112,115,0}},
/* 44302 */ {(11<<2)|2,{112,115,0}},
/* 44303 */ {(11<<2)|2,{112,115,0}},
/* 44304 */ {(11<<2)|2,{112,115,0}},
/* 44305 */ {(11<<2)|2,{112,115,0}},
/* 44306 */ {(11<<2)|2,{112,115,0}},
/* 44307 */ {(11<<2)|2,{112,115,0}},
/* 44308 */ {(11<<2)|2,{112,115,0}},
/* 44309 */ {(11<<2)|2,{112,115,0}},
/* 44310 */ {(11<<2)|2,{112,115,0}},
/* 44311 */ {(11<<2)|2,{112,115,0}},
/* 44312 */ {(11<<2)|2,{112,115,0}},
/* 44313 */ {(11<<2)|2,{112,115,0}},
/* 44314 */ {(11<<2)|2,{112,115,0}},
/* 44315 */ {(11<<2)|2,{112,115,0}},
/* 44316 */ {(11<<2)|2,{112,115,0}},
/* 44317 */ {(11<<2)|2,{112,115,0}},
/* 44318 */ {(11<<2)|2,{112,115,0}},
/* 44319 */ {(11<<2)|2,{112,115,0}},
/* 44320 */ {(16<<2)|3,{112,116,48}},
/* 44321 */ {(16<<2)|3,{112,116,49}},
/* 44322 */ {(16<<2)|3,{112,116,50}},
/* 44323 */ {(16<<2)|3,{112,116,97}},
/* 44324 */ {(16<<2)|3,{112,116,99}},
/* 44325 */ {(16<<2)|3,{112,116,101}},
/* 44326 */ {(16<<2)|3,{112,116,105}},
/* 44327 */ {(16<<2)|3,{112,116,111}},
/* 44328 */ {(16<<2)|3,{112,116,115}},
/* 44329 */ {(16<<2)|3,{112,116,116}},
/* 44330 */ {(11<<2)|2,{112,116,0}},
/* 44331 */ {(11<<2)|2,{112,116,0}},
/* 44332 */ {(11<<2)|2,{112,116,0}},
/* 44333 */ {(11<<2)|2,{112,116,0}},
/* 44334 */ {(11<<2)|2,{112,116,0}},
/* 44335 */ {(11<<2)|2,{112,116,0}},
/* 44336 */ {(11<<2)|2,{112,116,0}},
/* 44337 */ {(11<<2)|2,{112,116,0}},
/* 44338 */ {(11<<2)|2,{112,116,0}},
/* 44339 */ {(11<<2)|2,{112,116,0}},
/* 44340 */ {(11<<2)|2,{112,116,0}},
/* 44341 */ {(11<<2)|2,{112,116,0}},
/* 44342 */ {(11<<2)|2,{112,116,0}},
/* 44343 */ {(11<<2)|2,{112,116,0}},
/* 44344 */ {(11<<2)|2,{112,116,0}},
/* 44345 */ {(11<<2)|2,{112,116,0}},
/* 44346 */ {(11<<2)|2,{112,116,0}},
/* 44347 */ {(11<<2)|2,{112,116,0}},
/* 44348 */ {(11<<2)|2,{112,116,0}},
/* 44349 */ {(11<<2)|2,{112,116,0}},
/* 44350 */ {(11<<2)|2,{112,116,0}},
/* 44351 */ {(11<<2)|2,{112,116,0}},
/* 44352 */ {(12<<2)|2,{112,32,0}},
/* 44353 */ {(12<<2)|2,{112,32,0}},
/* 44354 */ {(12<<2)|2,{112,32,0}},
/* 44355 */ {(12<<2)|2,{112,32,0}},
/* 44356 */ {(12<<2)|2,{112,32,0}},
/* 44357 */ {(12<<2)|2,{112,32,0}},
/* 44358 */ {(12<<2)|2,{112,32,0}},
/* 44359 */ {(12<<2)|2,{112,32,0}},
/* 44360 */ {(12<<2)|2,{112,32,0}},
/* 44361 */ {(12<<2)|2,{112,32,0}},
/* 44362 */ {(12<<2)|2,{112,32,0}},
/* 44363 */ {(12<<2)|2,{112,32,0}},
/* 44364 */ {(12<<2)|2,{112,32,0}},
/* 44365 */ {(12<<2)|2,{112,32,0}},
/* 44366 */ {(12<<2)|2,{112,32,0}},
/* 44367 */ {(12<<2)|2,{112,32,0}},
/* 44368 */ {(12<<2)|2,{112,37,0}},
/* 44369 */ {(12<<2)|2,{112,37,0}},
/* 44370 */ {(12<<2)|2,{112,37,0}},
/* 44371 */ {(12<<2)|2,{112,37,0}},
/* 44372 */ {(12<<2)|2,{112,37,0}},
/* 44373 */ {(12<<2)|2,{112,37,0}},
/* 44374 */ {(12<<2)|2,{112,37,0}},
/* 44375 */ {(12<<2)|2,{112,37,0}},
/* 44376 */ {(12<<2)|2,{112,37,0}},
/* 44377 */ {(12<<2)|2,{112,37,0}},
/* 44378 */ {(12<<2)|2,{112,37,0}},
/* 44379 */ {(12<<2)|2,{112,37,0}},
/* 44380 */ {(12<<2)|2,{112,37,0}},
/* 44381 */ {(12<<2)|2,{112,37,0}},
/* 44382 */ {(12<<2)|2,{112,37,0}},
/* 44383 */ {(12<<2)|2,{112,37,0}},
/* 44384 */ {(12<<2)|2,{112,45,0}},
/* 44385 */ {(12<<2)|2,{112,45,0}},
/* 44386 */ {(12<<2)|2,{112,45,0}},
/* 44387 */ {(12<<2)|2,{112,45,0}},
/* 44388 */ {(12<<2)|2,{112,45,0}},
/* 44389 */ {(12<<2)|2,{112,45,0}},
/* 44390 */ {(12<<2)|2,{112,45,0}},
/* 44391 */ {(12<<2)|2,{112,45,0}},
/* 44392 */ {(12<<2)|2,{112,45,0}},
/* 44393 */ {(12<<2)|2,{112,45,0}},
/* 44394 */ {(12<<2)|2,{112,45,0}},
/* 44395 */ {(12<<2)|2,{112,45,0}},
/* 44396 */ {(12<<2)|2,{112,45,0}},
/* 44397 */ {(12<<2)|2,{112,45,0}},
/* 44398 */ {(12<<2)|2,{112,45,0}},
/* 44399 */ {(12<<2)|2,{112,45,0}},
/* 44400 */ {(12<<2)|2,{112,46,0}},
/* 44401 */ {(12<<2)|2,{112,46,0}},
/* 44402 */ {(12<<2)|2,{112,46,0}},
/* 44403 */ {(12<<2)|2,{112,46,0}},
/* 44404 */ {(12<<2)|2,{112,46,0}},
/* 44405 */ {(12<<2)|2,{112,46,0}},
/* 44406 */ {(12<<2)|2,{112,46,0}},
/* 44407 */ {(12<<2)|2,{112,46,0}},
/* 44408 */ {(12<<2)|2,{112,46,0}},
/* 44409 */ {(12<<2)|2,{112,46,0}},
/* 44410 */ {(12<<2)|2,{112,46,0}},
/* 44411 */ {(12<<2)|2,{112,46,0}},
/* 44412 */ {(12<<2)|2,{112,46,0}},
/* 44413 */ {(12<<2)|2,{112,46,0}},
/* 44414 */ {(12<<2)|2,{112,46,0}},
/* 44415 */ {(12<<2)|2,{112,46,0}},
/* 44416 */ {(12<<2)|2,{112,47,0}},
/* 44417 */ {(12<<2)|2,{112,47,0}},
/* 44418 */ {(12<<2)|2,{112,47,0}},
/* 44419 */ {(12<<2)|2,{112,47,0}},
/* 44420 */ {(12<<2)|2,{112,47,0}},
/* 44421 */ {(12<<2)|2,{112,47,0}},
/* 44422 */ {(12<<2)|2,{112,47,0}},
/* 44423 */ {(12<<2)|2,{112,47,0}},
/* 44424 */ {(12<<2)|2,{112,47,0}},
/* 44425 */ {(12<<2)|2,{112,47,0}},
/* 44426 */ {(12<<2)|2,{112,47,0}},
/* 44427 */ {(12<<2)|2,{112,47,0}},
/* 44428 */ {(12<<2)|2,{112,47,0}},
/* 44429 */ {(12<<2)|2,{112,47,0}},
/* 44430 */ {(12<<2)|2,{112,47,0}},
/* 44431 */ {(12<<2)|2,{112,47,0}},
/* 44432 */ {(12<<2)|2,{112,51,0}},
/* 44433 */ {(12<<2)|2,{112,51,0}},
/* 44434 */ {(12<<2)|2,{112,51,0}},
/* 44435 */ {(12<<2)|2,{112,51,0}},
/* 44436 */ {(12<<2)|2,{112,51,0}},
/* 44437 */ {(12<<2)|2,{112,51,0}},
/* 44438 */ {(12<<2)|2,{112,51,0}},
/* 44439 */ {(12<<2)|2,{112,51,0}},
/* 44440 */ {(12<<2)|2,{112,51,0}},
/* 44441 */ {(12<<2)|2,{112,51,0}},
/* 44442 */ {(12<<2)|2,{112,51,0}},
/* 44443 */ {(12<<2)|2,{112,51,0}},
/* 44444 */ {(12<<2)|2,{112,51,0}},
/* 44445 */ {(12<<2)|2,{112,51,0}},
/* 44446 */ {(12<<2)|2,{112,51,0}},
/* 44447 */ {(12<<2)|2,{112,51,0}},
/* 44448 */ {(12<<2)|2,{112,52,0}},
/* 44449 */ {(12<<2)|2,{112,52,0}},
/* 44450 */ {(12<<2)|2,{112,52,0}},
/* 44451 */ {(12<<2)|2,{112,52,0}},
/* 44452 */ {(12<<2)|2,{112,52,0}},
/* 44453 */ {(12<<2)|2,{112,52,0}},
/* 44454 */ {(12<<2)|2,{112,52,0}},
/* 44455 */ {(12<<2)|2,{112,52,0}},
/* 44456 */ {(12<<2)|2,{112,52,0}},
/* 44457 */ {(12<<2)|2,{112,52,0}},
/* 44458 */ {(12<<2)|2,{112,52,0}},
/* 44459 */ {(12<<2)|2,{112,52,0}},
/* 44460 */ {(12<<2)|2,{112,52,0}},
/* 44461 */ {(12<<2)|2,{112,52,0}},
/* 44462 */ {(12<<2)|2,{112,52,0}},
/* 44463 */ {(12<<2)|2,{112,52,0}},
/* 44464 */ {(12<<2)|2,{112,53,0}},
/* 44465 */ {(12<<2)|2,{112,53,0}},
/* 44466 */ {(12<<2)|2,{112,53,0}},
/* 44467 */ {(12<<2)|2,{112,53,0}},
/* 44468 */ {(12<<2)|2,{112,53,0}},
/* 44469 */ {(12<<2)|2,{112,53,0}},
/* 44470 */ {(12<<2)|2,{112,53,0}},
/* 44471 */ {(12<<2)|2,{112,53,0}},
/* 44472 */ {(12<<2)|2,{112,53,0}},
/* 44473 */ {(12<<2)|2,{112,53,0}},
/* 44474 */ {(12<<2)|2,{112,53,0}},
/* 44475 */ {(12<<2)|2,{112,53,0}},
/* 44476 */ {(12<<2)|2,{112,53,0}},
/* 44477 */ {(12<<2)|2,{112,53,0}},
/* 44478 */ {(12<<2)|2,{112,53,0}},
/* 44479 */ {(12<<2)|2,{112,53,0}},
/* 44480 */ {(12<<2)|2,{112,54,0}},
/* 44481 */ {(12<<2)|2,{112,54,0}},
/* 44482 */ {(12<<2)|2,{112,54,0}},
/* 44483 */ {(12<<2)|2,{112,54,0}},
/* 44484 */ {(12<<2)|2,{112,54,0}},
/* 44485 */ {(12<<2)|2,{112,54,0}},
/* 44486 */ {(12<<2)|2,{112,54,0}},
/* 44487 */ {(12<<2)|2,{112,54,0}},
/* 44488 */ {(12<<2)|2,{112,54,0}},
/* 44489 */ {(12<<2)|2,{112,54,0}},
/* 44490 */ {(12<<2)|2,{112,54,0}},
/* 44491 */ {(12<<2)|2,{112,54,0}},
/* 44492 */ {(12<<2)|2,{112,54,0}},
/* 44493 */ {(12<<2)|2,{112,54,0}},
/* 44494 */ {(12<<2)|2,{112,54,0}},
/* 44495 */ {(12<<2)|2,{112,54,0}},
/* 44496 */ {(12<<2)|2,{112,55,0}},
/* 44497 */ {(12<<2)|2,{112,55,0}},
/* 44498 */ {(12<<2)|2,{112,55,0}},
/* 44499 */ {(12<<2)|2,{112,55,0}},
/* 44500 */ {(12<<2)|2,{112,55,0}},
/* 44501 */ {(12<<2)|2,{112,55,0}},
/* 44502 */ {(12<<2)|2,{112,55,0}},
/* 44503 */ {(12<<2)|2,{112,55,0}},
/* 44504 */ {(12<<2)|2,{112,55,0}},
/* 44505 */ {(12<<2)|2,{112,55,0}},
/* 44506 */ {(12<<2)|2,{112,55,0}},
/* 44507 */ {(12<<2)|2,{112,55,0}},
/* 44508 */ {(12<<2)|2,{112,55,0}},
/* 44509 */ {(12<<2)|2,{112,55,0}},
/* 44510 */ {(12<<2)|2,{112,55,0}},
/* 44511 */ {(12<<2)|2,{112,55,0}},
/* 44512 */ {(12<<2)|2,{112,56,0}},
/* 44513 */ {(12<<2)|2,{112,56,0}},
/* 44514 */ {(12<<2)|2,{112,56,0}},
/* 44515 */ {(12<<2)|2,{112,56,0}},
/* 44516 */ {(12<<2)|2,{112,56,0}},
/* 44517 */ {(12<<2)|2,{112,56,0}},
/* 44518 */ {(12<<2)|2,{112,56,0}},
/* 44519 */ {(12<<2)|2,{112,56,0}},
/* 44520 */ {(12<<2)|2,{112,56,0}},
/* 44521 */ {(12<<2)|2,{112,56,0}},
/* 44522 */ {(12<<2)|2,{112,56,0}},
/* 44523 */ {(12<<2)|2,{112,56,0}},
/* 44524 */ {(12<<2)|2,{112,56,0}},
/* 44525 */ {(12<<2)|2,{112,56,0}},
/* 44526 */ {(12<<2)|2,{112,56,0}},
/* 44527 */ {(12<<2)|2,{112,56,0}},
/* 44528 */ {(12<<2)|2,{112,57,0}},
/* 44529 */ {(12<<2)|2,{112,57,0}},
/* 44530 */ {(12<<2)|2,{112,57,0}},
/* 44531 */ {(12<<2)|2,{112,57,0}},
/* 44532 */ {(12<<2)|2,{112,57,0}},
/* 44533 */ {(12<<2)|2,{112,57,0}},
/* 44534 */ {(12<<2)|2,{112,57,0}},
/* 44535 */ {(12<<2)|2,{112,57,0}},
/* 44536 */ {(12<<2)|2,{112,57,0}},
/* 44537 */ {(12<<2)|2,{112,57,0}},
/* 44538 */ {(12<<2)|2,{112,57,0}},
/* 44539 */ {(12<<2)|2,{112,57,0}},
/* 44540 */ {(12<<2)|2,{112,57,0}},
/* 44541 */ {(12<<2)|2,{112,57,0}},
/* 44542 */ {(12<<2)|2,{112,57,0}},
/* 44543 */ {(12<<2)|2,{112,57,0}},
/* 44544 */ {(12<<2)|2,{112,61,0}},
/* 44545 */ {(12<<2)|2,{112,61,0}},
/* 44546 */ {(12<<2)|2,{112,61,0}},
/* 44547 */ {(12<<2)|2,{112,61,0}},
/* 44548 */ {(12<<2)|2,{112,61,0}},
/* 44549 */ {(12<<2)|2,{112,61,0}},
/* 44550 */ {(12<<2)|2,{112,61,0}},
/* 44551 */ {(12<<2)|2,{112,61,0}},
/* 44552 */ {(12<<2)|2,{112,61,0}},
/* 44553 */ {(12<<2)|2,{112,61,0}},
/* 44554 */ {(12<<2)|2,{112,61,0}},
/* 44555 */ {(12<<2)|2,{112,61,0}},
/* 44556 */ {(12<<2)|2,{112,61,0}},
/* 44557 */ {(12<<2)|2,{112,61,0}},
/* 44558 */ {(12<<2)|2,{112,61,0}},
/* 44559 */ {(12<<2)|2,{112,61,0}},
/* 44560 */ {(12<<2)|2,{112,65,0}},
/* 44561 */ {(12<<2)|2,{112,65,0}},
/* 44562 */ {(12<<2)|2,{112,65,0}},
/* 44563 */ {(12<<2)|2,{112,65,0}},
/* 44564 */ {(12<<2)|2,{112,65,0}},
/* 44565 */ {(12<<2)|2,{112,65,0}},
/* 44566 */ {(12<<2)|2,{112,65,0}},
/* 44567 */ {(12<<2)|2,{112,65,0}},
/* 44568 */ {(12<<2)|2,{112,65,0}},
/* 44569 */ {(12<<2)|2,{112,65,0}},
/* 44570 */ {(12<<2)|2,{112,65,0}},
/* 44571 */ {(12<<2)|2,{112,65,0}},
/* 44572 */ {(12<<2)|2,{112,65,0}},
/* 44573 */ {(12<<2)|2,{112,65,0}},
/* 44574 */ {(12<<2)|2,{112,65,0}},
/* 44575 */ {(12<<2)|2,{112,65,0}},
/* 44576 */ {(12<<2)|2,{112,95,0}},
/* 44577 */ {(12<<2)|2,{112,95,0}},
/* 44578 */ {(12<<2)|2,{112,95,0}},
/* 44579 */ {(12<<2)|2,{112,95,0}},
/* 44580 */ {(12<<2)|2,{112,95,0}},
/* 44581 */ {(12<<2)|2,{112,95,0}},
/* 44582 */ {(12<<2)|2,{112,95,0}},
/* 44583 */ {(12<<2)|2,{112,95,0}},
/* 44584 */ {(12<<2)|2,{112,95,0}},
/* 44585 */ {(12<<2)|2,{112,95,0}},
/* 44586 */ {(12<<2)|2,{112,95,0}},
/* 44587 */ {(12<<2)|2,{112,95,0}},
/* 44588 */ {(12<<2)|2,{112,95,0}},
/* 44589 */ {(12<<2)|2,{112,95,0}},
/* 44590 */ {(12<<2)|2,{112,95,0}},
/* 44591 */ {(12<<2)|2,{112,95,0}},
/* 44592 */ {(12<<2)|2,{112,98,0}},
/* 44593 */ {(12<<2)|2,{112,98,0}},
/* 44594 */ {(12<<2)|2,{112,98,0}},
/* 44595 */ {(12<<2)|2,{112,98,0}},
/* 44596 */ {(12<<2)|2,{112,98,0}},
/* 44597 */ {(12<<2)|2,{112,98,0}},
/* 44598 */ {(12<<2)|2,{112,98,0}},
/* 44599 */ {(12<<2)|2,{112,98,0}},
/* 44600 */ {(12<<2)|2,{112,98,0}},
/* 44601 */ {(12<<2)|2,{112,98,0}},
/* 44602 */ {(12<<2)|2,{112,98,0}},
/* 44603 */ {(12<<2)|2,{112,98,0}},
/* 44604 */ {(12<<2)|2,{112,98,0}},
/* 44605 */ {(12<<2)|2,{112,98,0}},
/* 44606 */ {(12<<2)|2,{112,98,0}},
/* 44607 */ {(12<<2)|2,{112,98,0}},
/* 44608 */ {(12<<2)|2,{112,100,0}},
/* 44609 */ {(12<<2)|2,{112,100,0}},
/* 44610 */ {(12<<2)|2,{112,100,0}},
/* 44611 */ {(12<<2)|2,{112,100,0}},
/* 44612 */ {(12<<2)|2,{112,100,0}},
/* 44613 */ {(12<<2)|2,{112,100,0}},
/* 44614 */ {(12<<2)|2,{112,100,0}},
/* 44615 */ {(12<<2)|2,{112,100,0}},
/* 44616 */ {(12<<2)|2,{112,100,0}},
/* 44617 */ {(12<<2)|2,{112,100,0}},
/* 44618 */ {(12<<2)|2,{112,100,0}},
/* 44619 */ {(12<<2)|2,{112,100,0}},
/* 44620 */ {(12<<2)|2,{112,100,0}},
/* 44621 */ {(12<<2)|2,{112,100,0}},
/* 44622 */ {(12<<2)|2,{112,100,0}},
/* 44623 */ {(12<<2)|2,{112,100,0}},
/* 44624 */ {(12<<2)|2,{112,102,0}},
/* 44625 */ {(12<<2)|2,{112,102,0}},
/* 44626 */ {(12<<2)|2,{112,102,0}},
/* 44627 */ {(12<<2)|2,{112,102,0}},
/* 44628 */ {(12<<2)|2,{112,102,0}},
/* 44629 */ {(12<<2)|2,{112,102,0}},
/* 44630 */ {(12<<2)|2,{112,102,0}},
/* 44631 */ {(12<<2)|2,{112,102,0}},
/* 44632 */ {(12<<2)|2,{112,102,0}},
/* 44633 */ {(12<<2)|2,{112,102,0}},
/* 44634 */ {(12<<2)|2,{112,102,0}},
/* 44635 */ {(12<<2)|2,{112,102,0}},
/* 44636 */ {(12<<2)|2,{112,102,0}},
/* 44637 */ {(12<<2)|2,{112,102,0}},
/* 44638 */ {(12<<2)|2,{112,102,0}},
/* 44639 */ {(12<<2)|2,{112,102,0}},
/* 44640 */ {(12<<2)|2,{112,103,0}},
/* 44641 */ {(12<<2)|2,{112,103,0}},
/* 44642 */ {(12<<2)|2,{112,103,0}},
/* 44643 */ {(12<<2)|2,{112,103,0}},
/* 44644 */ {(12<<2)|2,{112,103,0}},
/* 44645 */ {(12<<2)|2,{112,103,0}},
/* 44646 */ {(12<<2)|2,{112,103,0}},
/* 44647 */ {(12<<2)|2,{112,103,0}},
/* 44648 */ {(12<<2)|2,{112,103,0}},
/* 44649 */ {(12<<2)|2,{112,103,0}},
/* 44650 */ {(12<<2)|2,{112,103,0}},
/* 44651 */ {(12<<2)|2,{112,103,0}},
/* 44652 */ {(12<<2)|2,{112,103,0}},
/* 44653 */ {(12<<2)|2,{112,103,0}},
/* 44654 */ {(12<<2)|2,{112,103,0}},
/* 44655 */ {(12<<2)|2,{112,103,0}},
/* 44656 */ {(12<<2)|2,{112,104,0}},
/* 44657 */ {(12<<2)|2,{112,104,0}},
/* 44658 */ {(12<<2)|2,{112,104,0}},
/* 44659 */ {(12<<2)|2,{112,104,0}},
/* 44660 */ {(12<<2)|2,{112,104,0}},
/* 44661 */ {(12<<2)|2,{112,104,0}},
/* 44662 */ {(12<<2)|2,{112,104,0}},
/* 44663 */ {(12<<2)|2,{112,104,0}},
/* 44664 */ {(12<<2)|2,{112,104,0}},
/* 44665 */ {(12<<2)|2,{112,104,0}},
/* 44666 */ {(12<<2)|2,{112,104,0}},
/* 44667 */ {(12<<2)|2,{112,104,0}},
/* 44668 */ {(12<<2)|2,{112,104,0}},
/* 44669 */ {(12<<2)|2,{112,104,0}},
/* 44670 */ {(12<<2)|2,{112,104,0}},
/* 44671 */ {(12<<2)|2,{112,104,0}},
/* 44672 */ {(12<<2)|2,{112,108,0}},
/* 44673 */ {(12<<2)|2,{112,108,0}},
/* 44674 */ {(12<<2)|2,{112,108,0}},
/* 44675 */ {(12<<2)|2,{112,108,0}},
/* 44676 */ {(12<<2)|2,{112,108,0}},
/* 44677 */ {(12<<2)|2,{112,108,0}},
/* 44678 */ {(12<<2)|2,{112,108,0}},
/* 44679 */ {(12<<2)|2,{112,108,0}},
/* 44680 */ {(12<<2)|2,{112,108,0}},
/* 44681 */ {(12<<2)|2,{112,108,0}},
/* 44682 */ {(12<<2)|2,{112,108,0}},
/* 44683 */ {(12<<2)|2,{112,108,0}},
/* 44684 */ {(12<<2)|2,{112,108,0}},
/* 44685 */ {(12<<2)|2,{112,108,0}},
/* 44686 */ {(12<<2)|2,{112,108,0}},
/* 44687 */ {(12<<2)|2,{112,108,0}},
/* 44688 */ {(12<<2)|2,{112,109,0}},
/* 44689 */ {(12<<2)|2,{112,109,0}},
/* 44690 */ {(12<<2)|2,{112,109,0}},
/* 44691 */ {(12<<2)|2,{112,109,0}},
/* 44692 */ {(12<<2)|2,{112,109,0}},
/* 44693 */ {(12<<2)|2,{112,109,0}},
/* 44694 */ {(12<<2)|2,{112,109,0}},
/* 44695 */ {(12<<2)|2,{112,109,0}},
/* 44696 */ {(12<<2)|2,{112,109,0}},
/* 44697 */ {(12<<2)|2,{112,109,0}},
/* 44698 */ {(12<<2)|2,{112,109,0}},
/* 44699 */ {(12<<2)|2,{112,109,0}},
/* 44700 */ {(12<<2)|2,{112,109,0}},
/* 44701 */ {(12<<2)|2,{112,109,0}},
/* 44702 */ {(12<<2)|2,{112,109,0}},
/* 44703 */ {(12<<2)|2,{112,109,0}},
/* 44704 */ {(12<<2)|2,{112,110,0}},
/* 44705 */ {(12<<2)|2,{112,110,0}},
/* 44706 */ {(12<<2)|2,{112,110,0}},
/* 44707 */ {(12<<2)|2,{112,110,0}},
/* 44708 */ {(12<<2)|2,{112,110,0}},
/* 44709 */ {(12<<2)|2,{112,110,0}},
/* 44710 */ {(12<<2)|2,{112,110,0}},
/* 44711 */ {(12<<2)|2,{112,110,0}},
/* 44712 */ {(12<<2)|2,{112,110,0}},
/* 44713 */ {(12<<2)|2,{112,110,0}},
/* 44714 */ {(12<<2)|2,{112,110,0}},
/* 44715 */ {(12<<2)|2,{112,110,0}},
/* 44716 */ {(12<<2)|2,{112,110,0}},
/* 44717 */ {(12<<2)|2,{112,110,0}},
/* 44718 */ {(12<<2)|2,{112,110,0}},
/* 44719 */ {(12<<2)|2,{112,110,0}},
/* 44720 */ {(12<<2)|2,{112,112,0}},
/* 44721 */ {(12<<2)|2,{112,112,0}},
/* 44722 */ {(12<<2)|2,{112,112,0}},
/* 44723 */ {(12<<2)|2,{112,112,0}},
/* 44724 */ {(12<<2)|2,{112,112,0}},
/* 44725 */ {(12<<2)|2,{112,112,0}},
/* 44726 */ {(12<<2)|2,{112,112,0}},
/* 44727 */ {(12<<2)|2,{112,112,0}},
/* 44728 */ {(12<<2)|2,{112,112,0}},
/* 44729 */ {(12<<2)|2,{112,112,0}},
/* 44730 */ {(12<<2)|2,{112,112,0}},
/* 44731 */ {(12<<2)|2,{112,112,0}},
/* 44732 */ {(12<<2)|2,{112,112,0}},
/* 44733 */ {(12<<2)|2,{112,112,0}},
/* 44734 */ {(12<<2)|2,{112,112,0}},
/* 44735 */ {(12<<2)|2,{112,112,0}},
/* 44736 */ {(12<<2)|2,{112,114,0}},
/* 44737 */ {(12<<2)|2,{112,114,0}},
/* 44738 */ {(12<<2)|2,{112,114,0}},
/* 44739 */ {(12<<2)|2,{112,114,0}},
/* 44740 */ {(12<<2)|2,{112,114,0}},
/* 44741 */ {(12<<2)|2,{112,114,0}},
/* 44742 */ {(12<<2)|2,{112,114,0}},
/* 44743 */ {(12<<2)|2,{112,114,0}},
/* 44744 */ {(12<<2)|2,{112,114,0}},
/* 44745 */ {(12<<2)|2,{112,114,0}},
/* 44746 */ {(12<<2)|2,{112,114,0}},
/* 44747 */ {(12<<2)|2,{112,114,0}},
/* 44748 */ {(12<<2)|2,{112,114,0}},
/* 44749 */ {(12<<2)|2,{112,114,0}},
/* 44750 */ {(12<<2)|2,{112,114,0}},
/* 44751 */ {(12<<2)|2,{112,114,0}},
/* 44752 */ {(12<<2)|2,{112,117,0}},
/* 44753 */ {(12<<2)|2,{112,117,0}},
/* 44754 */ {(12<<2)|2,{112,117,0}},
/* 44755 */ {(12<<2)|2,{112,117,0}},
/* 44756 */ {(12<<2)|2,{112,117,0}},
/* 44757 */ {(12<<2)|2,{112,117,0}},
/* 44758 */ {(12<<2)|2,{112,117,0}},
/* 44759 */ {(12<<2)|2,{112,117,0}},
/* 44760 */ {(12<<2)|2,{112,117,0}},
/* 44761 */ {(12<<2)|2,{112,117,0}},
/* 44762 */ {(12<<2)|2,{112,117,0}},
/* 44763 */ {(12<<2)|2,{112,117,0}},
/* 44764 */ {(12<<2)|2,{112,117,0}},
/* 44765 */ {(12<<2)|2,{112,117,0}},
/* 44766 */ {(12<<2)|2,{112,117,0}},
/* 44767 */ {(12<<2)|2,{112,117,0}},
/* 44768 */ {(13<<2)|2,{112,58,0}},
/* 44769 */ {(13<<2)|2,{112,58,0}},
/* 44770 */ {(13<<2)|2,{112,58,0}},
/* 44771 */ {(13<<2)|2,{112,58,0}},
/* 44772 */ {(13<<2)|2,{112,58,0}},
/* 44773 */ {(13<<2)|2,{112,58,0}},
/* 44774 */ {(13<<2)|2,{112,58,0}},
/* 44775 */ {(13<<2)|2,{112,58,0}},
/* 44776 */ {(13<<2)|2,{112,66,0}},
/* 44777 */ {(13<<2)|2,{112,66,0}},
/* 44778 */ {(13<<2)|2,{112,66,0}},
/* 44779 */ {(13<<2)|2,{112,66,0}},
/* 44780 */ {(13<<2)|2,{112,66,0}},
/* 44781 */ {(13<<2)|2,{112,66,0}},
/* 44782 */ {(13<<2)|2,{112,66,0}},
/* 44783 */ {(13<<2)|2,{112,66,0}},
/* 44784 */ {(13<<2)|2,{112,67,0}},
/* 44785 */ {(13<<2)|2,{112,67,0}},
/* 44786 */ {(13<<2)|2,{112,67,0}},
/* 44787 */ {(13<<2)|2,{112,67,0}},
/* 44788 */ {(13<<2)|2,{112,67,0}},
/* 44789 */ {(13<<2)|2,{112,67,0}},
/* 44790 */ {(13<<2)|2,{112,67,0}},
/* 44791 */ {(13<<2)|2,{112,67,0}},
/* 44792 */ {(13<<2)|2,{112,68,0}},
/* 44793 */ {(13<<2)|2,{112,68,0}},
/* 44794 */ {(13<<2)|2,{112,68,0}},
/* 44795 */ {(13<<2)|2,{112,68,0}},
/* 44796 */ {(13<<2)|2,{112,68,0}},
/* 44797 */ {(13<<2)|2,{112,68,0}},
/* 44798 */ {(13<<2)|2,{112,68,0}},
/* 44799 */ {(13<<2)|2,{112,68,0}},
/* 44800 */ {(13<<2)|2,{112,69,0}},
/* 44801 */ {(13<<2)|2,{112,69,0}},
/* 44802 */ {(13<<2)|2,{112,69,0}},
/* 44803 */ {(13<<2)|2,{112,69,0}},
/* 44804 */ {(13<<2)|2,{112,69,0}},
/* 44805 */ {(13<<2)|2,{112,69,0}},
/* 44806 */ {(13<<2)|2,{112,69,0}},
/* 44807 */ {(13<<2)|2,{112,69,0}},
/* 44808 */ {(13<<2)|2,{112,70,0}},
/* 44809 */ {(13<<2)|2,{112,70,0}},
/* 44810 */ {(13<<2)|2,{112,70,0}},
/* 44811 */ {(13<<2)|2,{112,70,0}},
/* 44812 */ {(13<<2)|2,{112,70,0}},
/* 44813 */ {(13<<2)|2,{112,70,0}},
/* 44814 */ {(13<<2)|2,{112,70,0}},
/* 44815 */ {(13<<2)|2,{112,70,0}},
/* 44816 */ {(13<<2)|2,{112,71,0}},
/* 44817 */ {(13<<2)|2,{112,71,0}},
/* 44818 */ {(13<<2)|2,{112,71,0}},
/* 44819 */ {(13<<2)|2,{112,71,0}},
/* 44820 */ {(13<<2)|2,{112,71,0}},
/* 44821 */ {(13<<2)|2,{112,71,0}},
/* 44822 */ {(13<<2)|2,{112,71,0}},
/* 44823 */ {(13<<2)|2,{112,71,0}},
/* 44824 */ {(13<<2)|2,{112,72,0}},
/* 44825 */ {(13<<2)|2,{112,72,0}},
/* 44826 */ {(13<<2)|2,{112,72,0}},
/* 44827 */ {(13<<2)|2,{112,72,0}},
/* 44828 */ {(13<<2)|2,{112,72,0}},
/* 44829 */ {(13<<2)|2,{112,72,0}},
/* 44830 */ {(13<<2)|2,{112,72,0}},
/* 44831 */ {(13<<2)|2,{112,72,0}},
/* 44832 */ {(13<<2)|2,{112,73,0}},
/* 44833 */ {(13<<2)|2,{112,73,0}},
/* 44834 */ {(13<<2)|2,{112,73,0}},
/* 44835 */ {(13<<2)|2,{112,73,0}},
/* 44836 */ {(13<<2)|2,{112,73,0}},
/* 44837 */ {(13<<2)|2,{112,73,0}},
/* 44838 */ {(13<<2)|2,{112,73,0}},
/* 44839 */ {(13<<2)|2,{112,73,0}},
/* 44840 */ {(13<<2)|2,{112,74,0}},
/* 44841 */ {(13<<2)|2,{112,74,0}},
/* 44842 */ {(13<<2)|2,{112,74,0}},
/* 44843 */ {(13<<2)|2,{112,74,0}},
/* 44844 */ {(13<<2)|2,{112,74,0}},
/* 44845 */ {(13<<2)|2,{112,74,0}},
/* 44846 */ {(13<<2)|2,{112,74,0}},
/* 44847 */ {(13<<2)|2,{112,74,0}},
/* 44848 */ {(13<<2)|2,{112,75,0}},
/* 44849 */ {(13<<2)|2,{112,75,0}},
/* 44850 */ {(13<<2)|2,{112,75,0}},
/* 44851 */ {(13<<2)|2,{112,75,0}},
/* 44852 */ {(13<<2)|2,{112,75,0}},
/* 44853 */ {(13<<2)|2,{112,75,0}},
/* 44854 */ {(13<<2)|2,{112,75,0}},
/* 44855 */ {(13<<2)|2,{112,75,0}},
/* 44856 */ {(13<<2)|2,{112,76,0}},
/* 44857 */ {(13<<2)|2,{112,76,0}},
/* 44858 */ {(13<<2)|2,{112,76,0}},
/* 44859 */ {(13<<2)|2,{112,76,0}},
/* 44860 */ {(13<<2)|2,{112,76,0}},
/* 44861 */ {(13<<2)|2,{112,76,0}},
/* 44862 */ {(13<<2)|2,{112,76,0}},
/* 44863 */ {(13<<2)|2,{112,76,0}},
/* 44864 */ {(13<<2)|2,{112,77,0}},
/* 44865 */ {(13<<2)|2,{112,77,0}},
/* 44866 */ {(13<<2)|2,{112,77,0}},
/* 44867 */ {(13<<2)|2,{112,77,0}},
/* 44868 */ {(13<<2)|2,{112,77,0}},
/* 44869 */ {(13<<2)|2,{112,77,0}},
/* 44870 */ {(13<<2)|2,{112,77,0}},
/* 44871 */ {(13<<2)|2,{112,77,0}},
/* 44872 */ {(13<<2)|2,{112,78,0}},
/* 44873 */ {(13<<2)|2,{112,78,0}},
/* 44874 */ {(13<<2)|2,{112,78,0}},
/* 44875 */ {(13<<2)|2,{112,78,0}},
/* 44876 */ {(13<<2)|2,{112,78,0}},
/* 44877 */ {(13<<2)|2,{112,78,0}},
/* 44878 */ {(13<<2)|2,{112,78,0}},
/* 44879 */ {(13<<2)|2,{112,78,0}},
/* 44880 */ {(13<<2)|2,{112,79,0}},
/* 44881 */ {(13<<2)|2,{112,79,0}},
/* 44882 */ {(13<<2)|2,{112,79,0}},
/* 44883 */ {(13<<2)|2,{112,79,0}},
/* 44884 */ {(13<<2)|2,{112,79,0}},
/* 44885 */ {(13<<2)|2,{112,79,0}},
/* 44886 */ {(13<<2)|2,{112,79,0}},
/* 44887 */ {(13<<2)|2,{112,79,0}},
/* 44888 */ {(13<<2)|2,{112,80,0}},
/* 44889 */ {(13<<2)|2,{112,80,0}},
/* 44890 */ {(13<<2)|2,{112,80,0}},
/* 44891 */ {(13<<2)|2,{112,80,0}},
/* 44892 */ {(13<<2)|2,{112,80,0}},
/* 44893 */ {(13<<2)|2,{112,80,0}},
/* 44894 */ {(13<<2)|2,{112,80,0}},
/* 44895 */ {(13<<2)|2,{112,80,0}},
/* 44896 */ {(13<<2)|2,{112,81,0}},
/* 44897 */ {(13<<2)|2,{112,81,0}},
/* 44898 */ {(13<<2)|2,{112,81,0}},
/* 44899 */ {(13<<2)|2,{112,81,0}},
/* 44900 */ {(13<<2)|2,{112,81,0}},
/* 44901 */ {(13<<2)|2,{112,81,0}},
/* 44902 */ {(13<<2)|2,{112,81,0}},
/* 44903 */ {(13<<2)|2,{112,81,0}},
/* 44904 */ {(13<<2)|2,{112,82,0}},
/* 44905 */ {(13<<2)|2,{112,82,0}},
/* 44906 */ {(13<<2)|2,{112,82,0}},
/* 44907 */ {(13<<2)|2,{112,82,0}},
/* 44908 */ {(13<<2)|2,{112,82,0}},
/* 44909 */ {(13<<2)|2,{112,82,0}},
/* 44910 */ {(13<<2)|2,{112,82,0}},
/* 44911 */ {(13<<2)|2,{112,82,0}},
/* 44912 */ {(13<<2)|2,{112,83,0}},
/* 44913 */ {(13<<2)|2,{112,83,0}},
/* 44914 */ {(13<<2)|2,{112,83,0}},
/* 44915 */ {(13<<2)|2,{112,83,0}},
/* 44916 */ {(13<<2)|2,{112,83,0}},
/* 44917 */ {(13<<2)|2,{112,83,0}},
/* 44918 */ {(13<<2)|2,{112,83,0}},
/* 44919 */ {(13<<2)|2,{112,83,0}},
/* 44920 */ {(13<<2)|2,{112,84,0}},
/* 44921 */ {(13<<2)|2,{112,84,0}},
/* 44922 */ {(13<<2)|2,{112,84,0}},
/* 44923 */ {(13<<2)|2,{112,84,0}},
/* 44924 */ {(13<<2)|2,{112,84,0}},
/* 44925 */ {(13<<2)|2,{112,84,0}},
/* 44926 */ {(13<<2)|2,{112,84,0}},
/* 44927 */ {(13<<2)|2,{112,84,0}},
/* 44928 */ {(13<<2)|2,{112,85,0}},
/* 44929 */ {(13<<2)|2,{112,85,0}},
/* 44930 */ {(13<<2)|2,{112,85,0}},
/* 44931 */ {(13<<2)|2,{112,85,0}},
/* 44932 */ {(13<<2)|2,{112,85,0}},
/* 44933 */ {(13<<2)|2,{112,85,0}},
/* 44934 */ {(13<<2)|2,{112,85,0}},
/* 44935 */ {(13<<2)|2,{112,85,0}},
/* 44936 */ {(13<<2)|2,{112,86,0}},
/* 44937 */ {(13<<2)|2,{112,86,0}},
/* 44938 */ {(13<<2)|2,{112,86,0}},
/* 44939 */ {(13<<2)|2,{112,86,0}},
/* 44940 */ {(13<<2)|2,{112,86,0}},
/* 44941 */ {(13<<2)|2,{112,86,0}},
/* 44942 */ {(13<<2)|2,{112,86,0}},
/* 44943 */ {(13<<2)|2,{112,86,0}},
/* 44944 */ {(13<<2)|2,{112,87,0}},
/* 44945 */ {(13<<2)|2,{112,87,0}},
/* 44946 */ {(13<<2)|2,{112,87,0}},
/* 44947 */ {(13<<2)|2,{112,87,0}},
/* 44948 */ {(13<<2)|2,{112,87,0}},
/* 44949 */ {(13<<2)|2,{112,87,0}},
/* 44950 */ {(13<<2)|2,{112,87,0}},
/* 44951 */ {(13<<2)|2,{112,87,0}},
/* 44952 */ {(13<<2)|2,{112,89,0}},
/* 44953 */ {(13<<2)|2,{112,89,0}},
/* 44954 */ {(13<<2)|2,{112,89,0}},
/* 44955 */ {(13<<2)|2,{112,89,0}},
/* 44956 */ {(13<<2)|2,{112,89,0}},
/* 44957 */ {(13<<2)|2,{112,89,0}},
/* 44958 */ {(13<<2)|2,{112,89,0}},
/* 44959 */ {(13<<2)|2,{112,89,0}},
/* 44960 */ {(13<<2)|2,{112,106,0}},
/* 44961 */ {(13<<2)|2,{112,106,0}},
/* 44962 */ {(13<<2)|2,{112,106,0}},
/* 44963 */ {(13<<2)|2,{112,106,0}},
/* 44964 */ {(13<<2)|2,{112,106,0}},
/* 44965 */ {(13<<2)|2,{112,106,0}},
/* 44966 */ {(13<<2)|2,{112,106,0}},
/* 44967 */ {(13<<2)|2,{112,106,0}},
/* 44968 */ {(13<<2)|2,{112,107,0}},
/* 44969 */ {(13<<2)|2,{112,107,0}},
/* 44970 */ {(13<<2)|2,{112,107,0}},
/* 44971 */ {(13<<2)|2,{112,107,0}},
/* 44972 */ {(13<<2)|2,{112,107,0}},
/* 44973 */ {(13<<2)|2,{112,107,0}},
/* 44974 */ {(13<<2)|2,{112,107,0}},
/* 44975 */ {(13<<2)|2,{112,107,0}},
/* 44976 */ {(13<<2)|2,{112,113,0}},
/* 44977 */ {(13<<2)|2,{112,113,0}},
/* 44978 */ {(13<<2)|2,{112,113,0}},
/* 44979 */ {(13<<2)|2,{112,113,0}},
/* 44980 */ {(13<<2)|2,{112,113,0}},
/* 44981 */ {(13<<2)|2,{112,113,0}},
/* 44982 */ {(13<<2)|2,{112,113,0}},
/* 44983 */ {(13<<2)|2,{112,113,0}},
/* 44984 */ {(13<<2)|2,{112,118,0}},
/* 44985 */ {(13<<2)|2,{112,118,0}},
/* 44986 */ {(13<<2)|2,{112,118,0}},
/* 44987 */ {(13<<2)|2,{112,118,0}},
/* 44988 */ {(13<<2)|2,{112,118,0}},
/* 44989 */ {(13<<2)|2,{112,118,0}},
/* 44990 */ {(13<<2)|2,{112,118,0}},
/* 44991 */ {(13<<2)|2,{112,118,0}},
/* 44992 */ {(13<<2)|2,{112,119,0}},
/* 44993 */ {(13<<2)|2,{112,119,0}},
/* 44994 */ {(13<<2)|2,{112,119,0}},
/* 44995 */ {(13<<2)|2,{112,119,0}},
/* 44996 */ {(13<<2)|2,{112,119,0}},
/* 44997 */ {(13<<2)|2,{112,119,0}},
/* 44998 */ {(13<<2)|2,{112,119,0}},
/* 44999 */ {(13<<2)|2,{112,119,0}},
/* 45000 */ {(13<<2)|2,{112,120,0}},
/* 45001 */ {(13<<2)|2,{112,120,0}},
/* 45002 */ {(13<<2)|2,{112,120,0}},
/* 45003 */ {(13<<2)|2,{112,120,0}},
/* 45004 */ {(13<<2)|2,{112,120,0}},
/* 45005 */ {(13<<2)|2,{112,120,0}},
/* 45006 */ {(13<<2)|2,{112,120,0}},
/* 45007 */ {(13<<2)|2,{112,120,0}},
/* 45008 */ {(13<<2)|2,{112,121,0}},
/* 45009 */ {(13<<2)|2,{112,121,0}},
/* 45010 */ {(13<<2)|2,{112,121,0}},
/* 45011 */ {(13<<2)|2,{112,121,0}},
/* 45012 */ {(13<<2)|2,{112,121,0}},
/* 45013 */ {(13<<2)|2,{112,121,0}},
/* 45014 */ {(13<<2)|2,{112,121,0}},
/* 45015 */ {(13<<2)|2,{112,121,0}},
/* 45016 */ {(13<<2)|2,{112,122,0}},
/* 45017 */ {(13<<2)|2,{112,122,0}},
/* 45018 */ {(13<<2)|2,{112,122,0}},
/* 45019 */ {(13<<2)|2,{112,122,0}},
/* 45020 */ {(13<<2)|2,{112,122,0}},
/* 45021 */ {(13<<2)|2,{112,122,0}},
/* 45022 */ {(13<<2)|2,{112,122,0}},
/* 45023 */ {(13<<2)|2,{112,122,0}},
/* 45024 */ {(14<<2)|2,{112,38,0}},
/* 45025 */ {(14<<2)|2,{112,38,0}},
/* 45026 */ {(14<<2)|2,{112,38,0}},
/* 45027 */ {(14<<2)|2,{112,38,0}},
/* 45028 */ {(14<<2)|2,{112,42,0}},
/* 45029 */ {(14<<2)|2,{112,42,0}},
/* 45030 */ {(14<<2)|2,{112,42,0}},
/* 45031 */ {(14<<2)|2,{112,42,0}},
/* 45032 */ {(14<<2)|2,{112,44,0}},
/* 45033 */ {(14<<2)|2,{112,44,0}},
/* 45034 */ {(14<<2)|2,{112,44,0}},
/* 45035 */ {(14<<2)|2,{112,44,0}},
/* 45036 */ {(14<<2)|2,{112,59,0}},
/* 45037 */ {(14<<2)|2,{112,59,0}},
/* 45038 */ {(14<<2)|2,{112,59,0}},
/* 45039 */ {(14<<2)|2,{112,59,0}},
/* 45040 */ {(14<<2)|2,{112,88,0}},
/* 45041 */ {(14<<2)|2,{112,88,0}},
/* 45042 */ {(14<<2)|2,{112,88,0}},
/* 45043 */ {(14<<2)|2,{112,88,0}},
/* 45044 */ {(14<<2)|2,{112,90,0}},
/* 45045 */ {(14<<2)|2,{112,90,0}},
/* 45046 */ {(14<<2)|2,{112,90,0}},
/* 45047 */ {(14<<2)|2,{112,90,0}},
/* 45048 */ {(16<<2)|2,{112,33,0}},
/* 45049 */ {(16<<2)|2,{112,34,0}},
/* 45050 */ {(16<<2)|2,{112,40,0}},
/* 45051 */ {(16<<2)|2,{112,41,0}},
/* 45052 */ {(16<<2)|2,{112,63,0}},
/* 45053 */ {(6<<2)|1,{112,0,0}},
/* 45054 */ {(6<<2)|1,{112,0,0}},
/* 45055 */ {(6<<2)|1,{112,0,0}},
/* 45056 */ {(16<<2)|3,{114,48,48}},
/* 45057 */ {(16<<2)|3,{114,48,49}},
/* 45058 */ {(16<<2)|3,{114,48,50}},
/* 45059 */ {(16<<2)|3,{114,48,97}},
/* 45060 */ {(16<<2)|3,{114,48,99}},
/* 45061 */ {(16<<2)|3,{114,48,101}},
/* 45062 */ {(16<<2)|3,{114,48,105}},
/* 45063 */ {(16<<2)|3,{114,48,111}},
/* 45064 */ {(16<<2)|3,{114,48,115}},
/* 45065 */ {(16<<2)|3,{114,48,116}},
/* 45066 */ {(11<<2)|2,{114,48,0}},
/* 45067 */ {(11<<2)|2,{114,48,0}},
/* 45068 */ {(11<<2)|2,{114,48,0}},
/* 45069 */ {(11<<2)|2,{114,48,0}},
/* 45070 */ {(11<<2)|2,{114,48,0}},
/* 45071 */ {(11<<2)|2,{114,48,0}},
/* 45072 */ {(11<<2)|2,{114,48,0}},
/* 45073 */ {(11<<2)|2,{114,48,0}},
/* 45074 */ {(11<<2)|2,{114,48,0}},
/* 45075 */ {(11<<2)|2,{114,48,0}},
/* 45076 */ {(11<<2)|2,{114,48,0}},
/* 45077 */ {(11<<2)|2,{114,48,0}},
/* 45078 */ {(11<<2)|2,{114,48,0}},
/* 45079 */ {(11<<2)|2,{114,48,0}},
/* 45080 */ {(11<<2)|2,{114,48,0}},
/* 45081 */ {(11<<2)|2,{114,48,0}},
/* 45082 */ {(11<<2)|2,{114,48,0}},
/* 45083 */ {(11<<2)|2,{114,48,0}},
/* 45084 */ {(11<<2)|2,{114,48,0}},
/* 45085 */ {(11<<2)|2,{114,48,0}},
/* 45086 */ {(11<<2)|2,{114,48,0}},
/* 45087 */ {(11<<2)|2,{114,48,0}},
/* 45088 */ {(16<<2)|3,{114,49,48}},
/* 45089 */ {(16<<2)|3,{114,49,49}},
/* 45090 */ {(16<<2)|3,{114,49,50}},
/* 45091 */ {(16<<2)|3,{114,49,97}},
/* 45092 */ {(16<<2)|3,{114,49,99}},
/* 45093 */ {(16<<2)|3,{114,49,101}},
/* 45094 */ {(16<<2)|3,{114,49,105}},
/* 45095 */ {(16<<2)|3,{114,49,111}},
/* 45096 */ {(16<<2)|3,{114,49,115}},
/* 45097 */ {(16<<2)|3,{114,49,116}},
/* 45098 */ {(11<<2)|2,{114,49,0}},
/* 45099 */ {(11<<2)|2,{114,49,0}},
/* 45100 */ {(11<<2)|2,{114,49,0}},
/* 45101 */ {(11<<2)|2,{114,49,0}},
/* 45102 */ {(11<<2)|2,{114,49,0}},
/* 45103 */ {(11<<2)|2,{114,49,0}},
/* 45104 */ {(11<<2)|2,{114,49,0}},
/* 45105 */ {(11<<2)|2,{114,49,0}},
/* 45106 */ {(11<<2)|2,{114,49,0}},
/* 45107 */ {(11<<2)|2,{114,49,0}},
/* 45108 */ {(11<<2)|2,{114,49,0}},
/* 45109 */ {(11<<2)|2,{114,49,0}},
/* 45110 */ {(11<<2)|2,{114,49,0}},
/* 45111 */ {(11<<2)|2,{114,49,0}},
/* 45112 */ {(11<<2)|2,{114,49,0}},
/* 45113 */ {(11<<2)|2,{114,49,0}},
/* 45114 */ {(11<<2)|2,{114,49,0}},
/* 45115 */ {(11<<2)|2,{114,49,0}},
/* 45116 */ {(11<<2)|2,{114,49,0}},
/* 45117 */ {(11<<2)|2,{114,49,0}},
/* 45118 */ {(11<<2)|2,{114,49,0}},
/* 45119 */ {(11<<2)|2,{114,49,0}},
/* 45120 */ {(16<<2)|3,{114,50,48}},
/* 45121 */ {(16<<2)|3,{114,50,49}},
/* 45122 */ {(16<<2)|3,{114,50,50}},
/* 45123 */ {(16<<2)|3,{114,50,97}},
/* 45124 */ {(16<<2)|3,{114,50,99}},
/* 45125 */ {(16<<2)|3,{114,50,101}},
/* 45126 */ {(16<<2)|3,{114,50,105}},
/* 45127 */ {(16<<2)|3,{114,50,111}},
/* 45128 */ {(16<<2)|3,{114,50,115}},
/* 45129 */ {(16<<2)|3,{114,50,116}},
/* 45130 */ {(11<<2)|2,{114,50,0}},
/* 45131 */ {(11<<2)|2,{114,50,0}},
/* 45132 */ {(11<<2)|2,{114,50,0}},
/* 45133 */ {(11<<2)|2,{114,50,0}},
/* 45134 */ {(11<<2)|2,{114,50,0}},
/* 45135 */ {(11<<2)|2,{114,50,0}},
/* 45136 */ {(11<<2)|2,{114,50,0}},
/* 45137 */ {(11<<2)|2,{114,50,0}},
/* 45138 */ {(11<<2)|2,{114,50,0}},
/* 45139 */ {(11<<2)|2,{114,50,0}},
/* 45140 */ {(11<<2)|2,{114,50,0}},
/* 45141 */ {(11<<2)|2,{114,50,0}},
/* 45142 */ {(11<<2)|2,{114,50,0}},
/* 45143 */ {(11<<2)|2,{114,50,0}},
/* 45144 */ {(11<<2)|2,{114,50,0}},
/* 45145 */ {(11<<2)|2,{114,50,0}},
/* 45146 */ {(11<<2)|2,{114,50,0}},
/* 45147 */ {(11<<2)|2,{114,50,0}},
/* 45148 */ {(11<<2)|2,{114,50,0}},
/* 45149 */ {(11<<2)|2,{114,50,0}},
/* 45150 */ {(11<<2)|2,{114,50,0}},
/* 45151 */ {(11<<2)|2,{114,50,0}},
/* 45152 */ {(16<<2)|3,{114,97,48}},
/* 45153 */ {(16<<2)|3,{114,97,49}},
/* 45154 */ {(16<<2)|3,{114,97,50}},
/* 45155 */ {(16<<2)|3,{114,97,97}},
/* 45156 */ {(16<<2)|3,{114,97,99}},
/* 45157 */ {(16<<2)|3,{114,97,101}},
/* 45158 */ {(16<<2)|3,{114,97,105}},
/* 45159 */ {(16<<2)|3,{114,97,111}},
/* 45160 */ {(16<<2)|3,{114,97,115}},
/* 45161 */ {(16<<2)|3,{114,97,116}},
/* 45162 */ {(11<<2)|2,{114,97,0}},
/* 45163 */ {(11<<2)|2,{114,97,0}},
/* 45164 */ {(11<<2)|2,{114,97,0}},
/* 45165 */ {(11<<2)|2,{114,97,0}},
/* 45166 */ {(11<<2)|2,{114,97,0}},
/* 45167 */ {(11<<2)|2,{114,97,0}},
/* 45168 */ {(11<<2)|2,{114,97,0}},
/* 45169 */ {(11<<2)|2,{114,97,0}},
/* 45170 */ {(11<<2)|2,{114,97,0}},
/* 45171 */ {(11<<2)|2,{114,97,0}},
/* 45172 */ {(11<<2)|2,{114,97,0}},
/* 45173 */ {(11<<2)|2,{114,97,0}},
/* 45174 */ {(11<<2)|2,{114,97,0}},
/* 45175 */ {(11<<2)|2,{114,97,0}},
/* 45176 */ {(11<<2)|2,{114,97,0}},
/* 45177 */ {(11<<2)|2,{114,97,0}},
/* 45178 */ {(11<<2)|2,{114,97,0}},
/* 45179 */ {(11<<2)|2,{114,97,0}},
/* 45180 */ {(11<<2)|2,{114,97,0}},
/* 45181 */ {(11<<2)|2,{114,97,0}},
/* 45182 */ {(11<<2)|2,{114,97,0}},
/* 45183 */ {(11<<2)|2,{114,97,0}},
/* 45184 */ {(16<<2)|3,{114,99,48}},
/* 45185 */ {(16<<2)|3,{114,99,49}},
/* 45186 */ {(16<<2)|3,{114,99,50}},
/* 45187 */ {(16<<2)|3,{114,99,97}},
/* 45188 */ {(16<<2)|3,{114,99,99}},
/* 45189 */ {(16<<2)|3,{114,99,101}},
/* 45190 */ {(16<<2)|3,{114,99,105}},
/* 45191 */ {(16<<2)|3,{114,99,111}},
/* 45192 */ {(16<<2)|3,{114,99,115}},
/* 45193 */ {(16<<2)|3,{114,99,116}},
/* 45194 */ {(11<<2)|2,{114,99,0}},
/* 45195 */ {(11<<2)|2,{114,99,0}},
/* 45196 */ {(11<<2)|2,{114,99,0}},
/* 45197 */ {(11<<2)|2,{114,99,0}},
/* 45198 */ {(11<<2)|2,{114,99,0}},
/* 45199 */ {(11<<2)|2,{114,99,0}},
/* 45200 */ {(11<<2)|2,{114,99,0}},
/* 45201 */ {(11<<2)|2,{114,99,0}},
/* 45202 */ {(11<<2)|2,{114,99,0}},
/* 45203 */ {(11<<2)|2,{114,99,0}},
/* 45204 */ {(11<<2)|2,{114,99,0}},
/* 45205 */ {(11<<2)|2,{114,99,0}},
/* 45206 */ {(11<<2)|2,{114,99,0}},
/* 45207 */ {(11<<2)|2,{114,99,0}},
/* 45208 */ {(11<<2)|2,{114,99,0}},
/* 45209 */ {(11<<2)|2,{114,99,0}},
/* 45210 */ {(11<<2)|2,{114,99,0}},
/* 45211 */ {(11<<2)|2,{114,99,0}},
/* 45212 */ {(11<<2)|2,{114,99,0}},
/* 45213 */ {(11<<2)|2,{114,99,0}},
/* 45214 */ {(11<<2)|2,{114,99,0}},
/* 45215 */ {(11<<2)|2,{114,99,0}},
/* 45216 */ {(16<<2)|3,{114,101,48}},
/* 45217 */ {(16<<2)|3,{114,101,49}},
/* 45218 */ {(16<<2)|3,{114,101,50}},
/* 45219 */ {(16<<2)|3,{114,101,97}},
/* 45220 */ {(16<<2)|3,{114,101,99}},
/* 45221 */ {(16<<2)|3,{114,101,101}},
/* 45222 */ {(16<<2)|3,{114,101,105}},
/* 45223 */ {(16<<2)|3,{114,101,111}},
/* 45224 */ {(16<<2)|3,{114,101,115}},
/* 45225 */ {(16<<2)|3,{114,101,116}},
/* 45226 */ {(11<<2)|2,{114,101,0}},
/* 45227 */ {(11<<2)|2,{114,101,0}},
/* 45228 */ {(11<<2)|2,{114,101,0}},
/* 45229 */ {(11<<2)|2,{114,101,0}},
/* 45230 */ {(11<<2)|2,{114,101,0}},
/* 45231 */ {(11<<2)|2,{114,101,0}},
/* 45232 */ {(11<<2)|2,{114,101,0}},
/* 45233 */ {(11<<2)|2,{114,101,0}},
/* 45234 */ {(11<<2)|2,{114,101,0}},
/* 45235 */ {(11<<2)|2,{114,101,0}},
/* 45236 */ {(11<<2)|2,{114,101,0}},
/* 45237 */ {(11<<2)|2,{114,101,0}},
/* 45238 */ {(11<<2)|2,{114,101,0}},
/* 45239 */ {(11<<2)|2,{114,101,0}},
/* 45240 */ {(11<<2)|2,{114,101,0}},
/* 45241 */ {(11<<2)|2,{114,101,0}},
/* 45242 */ {(11<<2)|2,{114,101,0}},
/* 45243 */ {(11<<2)|2,{114,101,0}},
/* 45244 */ {(11<<2)|2,{114,101,0}},
/* 45245 */ {(11<<2)|2,{114,101,0}},
/* 45246 */ {(11<<2)|2,{114,101,0}},
/* 45247 */ {(11<<2)|2,{114,101,0}},
/* 45248 */ {(16<<2)|3,{114,105,48}},
/* 45249 */ {(16<<2)|3,{114,105,49}},
/* 45250 */ {(16<<2)|3,{114,105,50}},
/* 45251 */ {(16<<2)|3,{114,105,97}},
/* 45252 */ {(16<<2)|3,{114,105,99}},
/* 45253 */ {(16<<2)|3,{114,105,101}},
/* 45254 */ {(16<<2)|3,{114,105,105}},
/* 45255 */ {(16<<2)|3,{114,105,111}},
/* 45256 */ {(16<<2)|3,{114,105,115}},
/* 45257 */ {(16<<2)|3,{114,105,116}},
/* 45258 */ {(11<<2)|2,{114,105,0}},
/* 45259 */ {(11<<2)|2,{114,105,0}},
/* 45260 */ {(11<<2)|2,{114,105,0}},
/* 45261 */ {(11<<2)|2,{114,105,0}},
/* 45262 */ {(11<<2)|2,{114,105,0}},
/* 45263 */ {(11<<2)|2,{114,105,0}},
/* 45264 */ {(11<<2)|2,{114,105,0}},
/* 45265 */ {(11<<2)|2,{114,105,0}},
/* 45266 */ {(11<<2)|2,{114,105,0}},
/* 45267 */ {(11<<2)|2,{114,105,0}},
/* 45268 */ {(11<<2)|2,{114,105,0}},
/* 45269 */ {(11<<2)|2,{114,105,0}},
/* 45270 */ {(11<<2)|2,{114,105,0}},
/* 45271 */ {(11<<2)|2,{114,105,0}},
/* 45272 */ {(11<<2)|2,{114,105,0}},
/* 45273 */ {(11<<2)|2,{114,105,0}},
/* 45274 */ {(11<<2)|2,{114,105,0}},
/* 45275 */ {(11<<2)|2,{114,105,0}},
/* 45276 */ {(11<<2)|2,{114,105,0}},
/* 45277 */ {(11<<2)|2,{114,105,0}},
/* 45278 */ {(11<<2)|2,{114,105,0}},
/* 45279 */ {(11<<2)|2,{114,105,0}},
/* 45280 */ {(16<<2)|3,{114,111,48}},
/* 45281 */ {(16<<2)|3,{114,111,49}},
/* 45282 */ {(16<<2)|3,{114,111,50}},
/* 45283 */ {(16<<2)|3,{114,111,97}},
/* 45284 */ {(16<<2)|3,{114,111,99}},
/* 45285 */ {(16<<2)|3,{114,111,101}},
/* 45286 */ {(16<<2)|3,{114,111,105}},
/* 45287 */ {(16<<2)|3,{114,111,111}},
/* 45288 */ {(16<<2)|3,{114,111,115}},
/* 45289 */ {(16<<2)|3,{114,111,116}},
/* 45290 */ {(11<<2)|2,{114,111,0}},
/* 45291 */ {(11<<2)|2,{114,111,0}},
/* 45292 */ {(11<<2)|2,{114,111,0}},
/* 45293 */ {(11<<2)|2,{114,111,0}},
/* 45294 */ {(11<<2)|2,{114,111,0}},
/* 45295 */ {(11<<2)|2,{114,111,0}},
/* 45296 */ {(11<<2)|2,{114,111,0}},
/* 45297 */ {(11<<2)|2,{114,111,0}},
/* 45298 */ {(11<<2)|2,{114,111,0}},
/* 45299 */ {(11<<2)|2,{114,111,0}},
/* 45300 */ {(11<<2)|2,{114,111,0}},
/* 45301 */ {(11<<2)|2,{114,111,0}},
/* 45302 */ {(11<<2)|2,{114,111,0}},
/* 45303 */ {(11<<2)|2,{114,111,0}},
/* 45304 */ {(11<<2)|2,{114,111,0}},
/* 45305 */ {(11<<2)|2,{114,111,0}},
/* 45306 */ {(11<<2)|2,{114,111,0}},
/* 45307 */ {(11<<2)|2,{114,111,0}},
/* 45308 */ {(11<<2)|2,{114,111,0}},
/* 45309 */ {(11<<2)|2,{114,111,0}},
/* 45310 */ {(11<<2)|2,{114,111,0}},
/* 45311 */ {(11<<2)|2,{114,111,0}},
/* 45312 */ {(16<<2)|3,{114,115,48}},
/* 45313 */ {(16<<2)|3,{114,115,49}},
/* 45314 */ {(16<<2)|3,{114,115,50}},
/* 45315 */ {(16<<2)|3,{114,115,97}},
/* 45316 */ {(16<<2)|3,{114,115,99}},
/* 45317 */ {(16<<2)|3,{114,115,101}},
/* 45318 */ {(16<<2)|3,{114,115,105}},
/* 45319 */ {(16<<2)|3,{114,115,111}},
/* 45320 */ {(16<<2)|3,{114,115,115}},
/* 45321 */ {(16<<2)|3,{114,115,116}},
/* 45322 */ {(11<<2)|2,{114,115,0}},
/* 45323 */ {(11<<2)|2,{114,115,0}},
/* 45324 */ {(11<<2)|2,{114,115,0}},
/* 45325 */ {(11<<2)|2,{114,115,0}},
/* 45326 */ {(11<<2)|2,{114,115,0}},
/* 45327 */ {(11<<2)|2,{114,115,0}},
/* 45328 */ {(11<<2)|2,{114,115,0}},
/* 45329 */ {(11<<2)|2,{114,115,0}},
/* 45330 */ {(11<<2)|2,{114,115,0}},
/* 45331 */ {(11<<2)|2,{114,115,0}},
/* 45332 */ {(11<<2)|2,{114,115,0}},
/* 45333 */ {(11<<2)|2,{114,115,0}},
/* 45334 */ {(11<<2)|2,{114,115,0}},
/* 45335 */ {(11<<2)|2,{114,115,0}},
/* 45336 */ {(11<<2)|2,{114,115,0}},
/* 45337 */ {(11<<2)|2,{114,115,0}},
/* 45338 */ {(11<<2)|2,{114,115,0}},
/* 45339 */ {(11<<2)|2,{114,115,0}},
/* 45340 */ {(11<<2)|2,{114,115,0}},
/* 45341 */ {(11<<2)|2,{114,115,0}},
/* 45342 */ {(11<<2)|2,{114,115,0}},
/* 45343 */ {(11<<2)|2,{114,115,0}},
/* 45344 */ {(16<<2)|3,{114,116,48}},
/* 45345 */ {(16<<2)|3,{114,116,49}},
/* 45346 */ {(16<<2)|3,{114,116,50}},
/* 45347 */ {(16<<2)|3,{114,116,97}},
/* 45348 */ {(16<<2)|3,{114,116,99}},
/* 45349 */ {(16<<2)|3,{114,116,101}},
/* 45350 */ {(16<<2)|3,{114,116,105}},
/* 45351 */ {(16<<2)|3,{114,116,111}},
/* 45352 */ {(16<<2)|3,{114,116,115}},
/* 45353 */ {(16<<2)|3,{114,116,116}},
/* 45354 */ {(11<<2)|2,{114,116,0}},
/* 45355 */ {(11<<2)|2,{114,116,0}},
/* 45356 */ {(11<<2)|2,{114,116,0}},
/* 45357 */ {(11<<2)|2,{114,116,0}},
/* 45358 */ {(11<<2)|2,{114,116,0}},
/* 45359 */ {(11<<2)|2,{114,116,0}},
/* 45360 */ {(11<<2)|2,{114,116,0}},
/* 45361 */ {(11<<2)|2,{114,116,0}},
/* 45362 */ {(11<<2)|2,{114,116,0}},
/* 45363 */ {(11<<2)|2,{114,116,0}},
/* 45364 */ {(11<<2)|2,{114,116,0}},
/* 45365 */ {(11<<2)|2,{114,116,0}},
/* 45366 */ {(11<<2)|2,{114,116,0}},
/* 45367 */ {(11<<2)|2,{114,116,0}},
/* 45368 */ {(11<<2)|2,{114,116,0}},
/* 45369 */ {(11<<2)|2,{114,116,0}},
/* 45370 */ {(11<<2)|2,{114,116,0}},
/* 45371 */ {(11<<2)|2,{114,116,0}},
/* 45372 */ {(11<<2)|2,{114,116,0}},
/* 45373 */ {(11<<2)|2,{114,116,0}},
/* 45374 */ {(11<<2)|2,{114,116,0}},
/* 45375 */ {(11<<2)|2,{114,116,0}},
/* 45376 */ {(12<<2)|2,{114,32,0}},
/* 45377 */ {(12<<2)|2,{114,32,0}},
/* 45378 */ {(12<<2)|2,{114,32,0}},
/* 45379 */ {(12<<2)|2,{114,32,0}},
/* 45380 */ {(12<<2)|2,{114,32,0}},
/* 45381 */ {(12<<2)|2,{114,32,0}},
/* 45382 */ {(12<<2)|2,{114,32,0}},
/* 45383 */ {(12<<2)|2,{114,32,0}},
/* 45384 */ {(12<<2)|2,{114,32,0}},
/* 45385 */ {(12<<2)|2,{114,32,0}},
/* 45386 */ {(12<<2)|2,{114,32,0}},
/* 45387 */ {(12<<2)|2,{114,32,0}},
/* 45388 */ {(12<<2)|2,{114,32,0}},
/* 45389 */ {(12<<2)|2,{114,32,0}},
/* 45390 */ {(12<<2)|2,{114,32,0}},
/* 45391 */ {(12<<2)|2,{114,32,0}},
/* 45392 */ {(12<<2)|2,{114,37,0}},
/* 45393 */ {(12<<2)|2,{114,37,0}},
/* 45394 */ {(12<<2)|2,{114,37,0}},
/* 45395 */ {(12<<2)|2,{114,37,0}},
/* 45396 */ {(12<<2)|2,{114,37,0}},
/* 45397 */ {(12<<2)|2,{114,37,0}},
/* 45398 */ {(12<<2)|2,{114,37,0}},
/* 45399 */ {(12<<2)|2,{114,37,0}},
/* 45400 */ {(12<<2)|2,{114,37,0}},
/* 45401 */ {(12<<2)|2,{114,37,0}},
/* 45402 */ {(12<<2)|2,{114,37,0}},
/* 45403 */ {(12<<2)|2,{114,37,0}},
/* 45404 */ {(12<<2)|2,{114,37,0}},
/* 45405 */ {(12<<2)|2,{114,37,0}},
/* 45406 */ {(12<<2)|2,{114,37,0}},
/* 45407 */ {(12<<2)|2,{114,37,0}},
/* 45408 */ {(12<<2)|2,{114,45,0}},
/* 45409 */ {(12<<2)|2,{114,45,0}},
/* 45410 */ {(12<<2)|2,{114,45,0}},
/* 45411 */ {(12<<2)|2,{114,45,0}},
/* 45412 */ {(12<<2)|2,{114,45,0}},
/* 45413 */ {(12<<2)|2,{114,45,0}},
/* 45414 */ {(12<<2)|2,{114,45,0}},
/* 45415 */ {(12<<2)|2,{114,45,0}},
/* 45416 */ {(12<<2)|2,{114,45,0}},
/* 45417 */ {(12<<2)|2,{114,45,0}},
/* 45418 */ {(12<<2)|2,{114,45,0}},
/* 45419 */ {(12<<2)|2,{114,45,0}},
/* 45420 */ {(12<<2)|2,{114,45,0}},
/* 45421 */ {(12<<2)|2,{114,45,0}},
/* 45422 */ {(12<<2)|2,{114,45,0}},
/* 45423 */ {(12<<2)|2,{114,45,0}},
/* 45424 */ {(12<<2)|2,{114,46,0}},
/* 45425 */ {(12<<2)|2,{114,46,0}},
/* 45426 */ {(12<<2)|2,{114,46,0}},
/* 45427 */ {(12<<2)|2,{114,46,0}},
/* 45428 */ {(12<<2)|2,{114,46,0}},
/* 45429 */ {(12<<2)|2,{114,46,0}},
/* 45430 */ {(12<<2)|2,{114,46,0}},
/* 45431 */ {(12<<2)|2,{114,46,0}},
/* 45432 */ {(12<<2)|2,{114,46,0}},
/* 45433 */ {(12<<2)|2,{114,46,0}},
/* 45434 */ {(12<<2)|2,{114,46,0}},
/* 45435 */ {(12<<2)|2,{114,46,0}},
/* 45436 */ {(12<<2)|2,{114,46,0}},
/* 45437 */ {(12<<2)|2,{114,46,0}},
/* 45438 */ {(12<<2)|2,{114,46,0}},
/* 45439 */ {(12<<2)|2,{114,46,0}},
/* 45440 */ {(12<<2)|2,{114,47,0}},
/* 45441 */ {(12<<2)|2,{114,47,0}},
/* 45442 */ {(12<<2)|2,{114,47,0}},
/* 45443 */ {(12<<2)|2,{114,47,0}},
/* 45444 */ {(12<<2)|2,{114,47,0}},
/* 45445 */ {(12<<2)|2,{114,47,0}},
/* 45446 */ {(12<<2)|2,{114,47,0}},
/* 45447 */ {(12<<2)|2,{114,47,0}},
/* 45448 */ {(12<<2)|2,{114,47,0}},
/* 45449 */ {(12<<2)|2,{114,47,0}},
/* 45450 */ {(12<<2)|2,{114,47,0}},
/* 45451 */ {(12<<2)|2,{114,47,0}},
/* 45452 */ {(12<<2)|2,{114,47,0}},
/* 45453 */ {(12<<2)|2,{114,47,0}},
/* 45454 */ {(12<<2)|2,{114,47,0}},
/* 45455 */ {(12<<2)|2,{114,47,0}},
/* 45456 */ {(12<<2)|2,{114,51,0}},
/* 45457 */ {(12<<2)|2,{114,51,0}},
/* 45458 */ {(12<<2)|2,{114,51,0}},
/* 45459 */ {(12<<2)|2,{114,51,0}},
/* 45460 */ {(12<<2)|2,{114,51,0}},
/* 45461 */ {(12<<2)|2,{114,51,0}},
/* 45462 */ {(12<<2)|2,{114,51,0}},
/* 45463 */ {(12<<2)|2,{114,51,0}},
/* 45464 */ {(12<<2)|2,{114,51,0}},
/* 45465 */ {(12<<2)|2,{114,51,0}},
/* 45466 */ {(12<<2)|2,{114,51,0}},
/* 45467 */ {(12<<2)|2,{114,51,0}},
/* 45468 */ {(12<<2)|2,{114,51,0}},
/* 45469 */ {(12<<2)|2,{114,51,0}},
/* 45470 */ {(12<<2)|2,{114,51,0}},
/* 45471 */ {(12<<2)|2,{114,51,0}},
/* 45472 */ {(12<<2)|2,{114,52,0}},
/* 45473 */ {(12<<2)|2,{114,52,0}},
/* 45474 */ {(12<<2)|2,{114,52,0}},
/* 45475 */ {(12<<2)|2,{114,52,0}},
/* 45476 */ {(12<<2)|2,{114,52,0}},
/* 45477 */ {(12<<2)|2,{114,52,0}},
/* 45478 */ {(12<<2)|2,{114,52,0}},
/* 45479 */ {(12<<2)|2,{114,52,0}},
/* 45480 */ {(12<<2)|2,{114,52,0}},
/* 45481 */ {(12<<2)|2,{114,52,0}},
/* 45482 */ {(12<<2)|2,{114,52,0}},
/* 45483 */ {(12<<2)|2,{114,52,0}},
/* 45484 */ {(12<<2)|2,{114,52,0}},
/* 45485 */ {(12<<2)|2,{114,52,0}},
/* 45486 */ {(12<<2)|2,{114,52,0}},
/* 45487 */ {(12<<2)|2,{114,52,0}},
/* 45488 */ {(12<<2)|2,{114,53,0}},
/* 45489 */ {(12<<2)|2,{114,53,0}},
/* 45490 */ {(12<<2)|2,{114,53,0}},
/* 45491 */ {(12<<2)|2,{114,53,0}},
/* 45492 */ {(12<<2)|2,{114,53,0}},
/* 45493 */ {(12<<2)|2,{114,53,0}},
/* 45494 */ {(12<<2)|2,{114,53,0}},
/* 45495 */ {(12<<2)|2,{114,53,0}},
/* 45496 */ {(12<<2)|2,{114,53,0}},
/* 45497 */ {(12<<2)|2,{114,53,0}},
/* 45498 */ {(12<<2)|2,{114,53,0}},
/* 45499 */ {(12<<2)|2,{114,53,0}},
/* 45500 */ {(12<<2)|2,{114,53,0}},
/* 45501 */ {(12<<2)|2,{114,53,0}},
/* 45502 */ {(12<<2)|2,{114,53,0}},
/* 45503 */ {(12<<2)|2,{114,53,0}},
/* 45504 */ {(12<<2)|2,{114,54,0}},
/* 45505 */ {(12<<2)|2,{114,54,0}},
/* 45506 */ {(12<<2)|2,{114,54,0}},
/* 45507 */ {(12<<2)|2,{114,54,0}},
/* 45508 */ {(12<<2)|2,{114,54,0}},
/* 45509 */ {(12<<2)|2,{114,54,0}},
/* 45510 */ {(12<<2)|2,{114,54,0}},
/* 45511 */ {(12<<2)|2,{114,54,0}},
/* 45512 */ {(12<<2)|2,{114,54,0}},
/* 45513 */ {(12<<2)|2,{114,54,0}},
/* 45514 */ {(12<<2)|2,{114,54,0}},
/* 45515 */ {(12<<2)|2,{114,54,0}},
/* 45516 */ {(12<<2)|2,{114,54,0}},
/* 45517 */ {(12<<2)|2,{114,54,0}},
/* 45518 */ {(12<<2)|2,{114,54,0}},
/* 45519 */ {(12<<2)|2,{114,54,0}},
/* 45520 */ {(12<<2)|2,{114,55,0}},
/* 45521 */ {(12<<2)|2,{114,55,0}},
/* 45522 */ {(12<<2)|2,{114,55,0}},
/* 45523 */ {(12<<2)|2,{114,55,0}},
/* 45524 */ {(12<<2)|2,{114,55,0}},
/* 45525 */ {(12<<2)|2,{114,55,0}},
/* 45526 */ {(12<<2)|2,{114,55,0}},
/* 45527 */ {(12<<2)|2,{114,55,0}},
/* 45528 */ {(12<<2)|2,{114,55,0}},
/* 45529 */ {(12<<2)|2,{114,55,0}},
/* 45530 */ {(12<<2)|2,{114,55,0}},
/* 45531 */ {(12<<2)|2,{114,55,0}},
/* 45532 */ {(12<<2)|2,{114,55,0}},
/* 45533 */ {(12<<2)|2,{114,55,0}},
/* 45534 */ {(12<<2)|2,{114,55,0}},
/* 45535 */ {(12<<2)|2,{114,55,0}},
/* 45536 */ {(12<<2)|2,{114,56,0}},
/* 45537 */ {(12<<2)|2,{114,56,0}},
/* 45538 */ {(12<<2)|2,{114,56,0}},
/* 45539 */ {(12<<2)|2,{114,56,0}},
/* 45540 */ {(12<<2)|2,{114,56,0}},
/* 45541 */ {(12<<2)|2,{114,56,0}},
/* 45542 */ {(12<<2)|2,{114,56,0}},
/* 45543 */ {(12<<2)|2,{114,56,0}},
/* 45544 */ {(12<<2)|2,{114,56,0}},
/* 45545 */ {(12<<2)|2,{114,56,0}},
/* 45546 */ {(12<<2)|2,{114,56,0}},
/* 45547 */ {(12<<2)|2,{114,56,0}},
/* 45548 */ {(12<<2)|2,{114,56,0}},
/* 45549 */ {(12<<2)|2,{114,56,0}},
/* 45550 */ {(12<<2)|2,{114,56,0}},
/* 45551 */ {(12<<2)|2,{114,56,0}},
/* 45552 */ {(12<<2)|2,{114,57,0}},
/* 45553 */ {(12<<2)|2,{114,57,0}},
/* 45554 */ {(12<<2)|2,{114,57,0}},
/* 45555 */ {(12<<2)|2,{114,57,0}},
/* 45556 */ {(12<<2)|2,{114,57,0}},
/* 45557 */ {(12<<2)|2,{114,57,0}},
/* 45558 */ {(12<<2)|2,{114,57,0}},
/* 45559 */ {(12<<2)|2,{114,57,0}},
/* 45560 */ {(12<<2)|2,{114,57,0}},
/* 45561 */ {(12<<2)|2,{114,57,0}},
/* 45562 */ {(12<<2)|2,{114,57,0}},
/* 45563 */ {(12<<2)|2,{114,57,0}},
/* 45564 */ {(12<<2)|2,{114,57,0}},
/* 45565 */ {(12<<2)|2,{114,57,0}},
/* 45566 */ {(12<<2)|2,{114,57,0}},
/* 45567 */ {(12<<2)|2,{114,57,0}},
/* 45568 */ {(12<<2)|2,{114,61,0}},
/* 45569 */ {(12<<2)|2,{114,61,0}},
/* 45570 */ {(12<<2)|2,{114,61,0}},
/* 45571 */ {(12<<2)|2,{114,61,0}},
/* 45572 */ {(12<<2)|2,{114,61,0}},
/* 45573 */ {(12<<2)|2,{114,61,0}},
/* 45574 */ {(12<<2)|2,{114,61,0}},
/* 45575 */ {(12<<2)|2,{114,61,0}},
/* 45576 */ {(12<<2)|2,{114,61,0}},
/* 45577 */ {(12<<2)|2,{114,61,0}},
/* 45578 */ {(12<<2)|2,{114,61,0}},
/* 45579 */ {(12<<2)|2,{114,61,0}},
/* 45580 */ {(12<<2)|2,{114,61,0}},
/* 45581 */ {(12<<2)|2,{114,61,0}},
/* 45582 */ {(12<<2)|2,{114,61,0}},
/* 45583 */ {(12<<2)|2,{114,61,0}},
/* 45584 */ {(12<<2)|2,{114,65,0}},
/* 45585 */ {(12<<2)|2,{114,65,0}},
/* 45586 */ {(12<<2)|2,{114,65,0}},
/* 45587 */ {(12<<2)|2,{114,65,0}},
/* 45588 */ {(12<<2)|2,{114,65,0}},
/* 45589 */ {(12<<2)|2,{114,65,0}},
/* 45590 */ {(12<<2)|2,{114,65,0}},
/* 45591 */ {(12<<2)|2,{114,65,0}},
/* 45592 */ {(12<<2)|2,{114,65,0}},
/* 45593 */ {(12<<2)|2,{114,65,0}},
/* 45594 */ {(12<<2)|2,{114,65,0}},
/* 45595 */ {(12<<2)|2,{114,65,0}},
/* 45596 */ {(12<<2)|2,{114,65,0}},
/* 45597 */ {(12<<2)|2,{114,65,0}},
/* 45598 */ {(12<<2)|2,{114,65,0}},
/* 45599 */ {(12<<2)|2,{114,65,0}},
/* 45600 */ {(12<<2)|2,{114,95,0}},
/* 45601 */ {(12<<2)|2,{114,95,0}},
/* 45602 */ {(12<<2)|2,{114,95,0}},
/* 45603 */ {(12<<2)|2,{114,95,0}},
/* 45604 */ {(12<<2)|2,{114,95,0}},
/* 45605 */ {(12<<2)|2,{114,95,0}},
/* 45606 */ {(12<<2)|2,{114,95,0}},
/* 45607 */ {(12<<2)|2,{114,95,0}},
/* 45608 */ {(12<<2)|2,{114,95,0}},
/* 45609 */ {(12<<2)|2,{114,95,0}},
/* 45610 */ {(12<<2)|2,{114,95,0}},
/* 45611 */ {(12<<2)|2,{114,95,0}},
/* 45612 */ {(12<<2)|2,{114,95,0}},
/* 45613 */ {(12<<2)|2,{114,95,0}},
/* 45614 */ {(12<<2)|2,{114,95,0}},
/* 45615 */ {(12<<2)|2,{114,95,0}},
/* 45616 */ {(12<<2)|2,{114,98,0}},
/* 45617 */ {(12<<2)|2,{114,98,0}},
/* 45618 */ {(12<<2)|2,{114,98,0}},
/* 45619 */ {(12<<2)|2,{114,98,0}},
/* 45620 */ {(12<<2)|2,{114,98,0}},
/* 45621 */ {(12<<2)|2,{114,98,0}},
/* 45622 */ {(12<<2)|2,{114,98,0}},
/* 45623 */ {(12<<2)|2,{114,98,0}},
/* 45624 */ {(12<<2)|2,{114,98,0}},
/* 45625 */ {(12<<2)|2,{114,98,0}},
/* 45626 */ {(12<<2)|2,{114,98,0}},
/* 45627 */ {(12<<2)|2,{114,98,0}},
/* 45628 */ {(12<<2)|2,{114,98,0}},
/* 45629 */ {(12<<2)|2,{114,98,0}},
/* 45630 */ {(12<<2)|2,{114,98,0}},
/* 45631 */ {(12<<2)|2,{114,98,0}},
/* 45632 */ {(12<<2)|2,{114,100,0}},
/* 45633 */ {(12<<2)|2,{114,100,0}},
/* 45634 */ {(12<<2)|2,{114,100,0}},
/* 45635 */ {(12<<2)|2,{114,100,0}},
/* 45636 */ {(12<<2)|2,{114,100,0}},
/* 45637 */ {(12<<2)|2,{114,100,0}},
/* 45638 */ {(12<<2)|2,{114,100,0}},
/* 45639 */ {(12<<2)|2,{114,100,0}},
/* 45640 */ {(12<<2)|2,{114,100,0}},
/* 45641 */ {(12<<2)|2,{114,100,0}},
/* 45642 */ {(12<<2)|2,{114,100,0}},
/* 45643 */ {(12<<2)|2,{114,100,0}},
/* 45644 */ {(12<<2)|2,{114,100,0}},
/* 45645 */ {(12<<2)|2,{114,100,0}},
/* 45646 */ {(12<<2)|2,{114,100,0}},
/* 45647 */ {(12<<2)|2,{114,100,0}},
/* 45648 */ {(12<<2)|2,{114,102,0}},
/* 45649 */ {(12<<2)|2,{114,102,0}},
/* 45650 */ {(12<<2)|2,{114,102,0}},
/* 45651 */ {(12<<2)|2,{114,102,0}},
/* 45652 */ {(12<<2)|2,{114,102,0}},
/* 45653 */ {(12<<2)|2,{114,102,0}},
/* 45654 */ {(12<<2)|2,{114,102,0}},
/* 45655 */ {(12<<2)|2,{114,102,0}},
/* 45656 */ {(12<<2)|2,{114,102,0}},
/* 45657 */ {(12<<2)|2,{114,102,0}},
/* 45658 */ {(12<<2)|2,{114,102,0}},
/* 45659 */ {(12<<2)|2,{114,102,0}},
/* 45660 */ {(12<<2)|2,{114,102,0}},
/* 45661 */ {(12<<2)|2,{114,102,0}},
/* 45662 */ {(12<<2)|2,{114,102,0}},
/* 45663 */ {(12<<2)|2,{114,102,0}},
/* 45664 */ {(12<<2)|2,{114,103,0}},
/* 45665 */ {(12<<2)|2,{114,103,0}},
/* 45666 */ {(12<<2)|2,{114,103,0}},
/* 45667 */ {(12<<2)|2,{114,103,0}},
/* 45668 */ {(12<<2)|2,{114,103,0}},
/* 45669 */ {(12<<2)|2,{114,103,0}},
/* 45670 */ {(12<<2)|2,{114,103,0}},
/* 45671 */ {(12<<2)|2,{114,103,0}},
/* 45672 */ {(12<<2)|2,{114,103,0}},
/* 45673 */ {(12<<2)|2,{114,103,0}},
/* 45674 */ {(12<<2)|2,{114,103,0}},
/* 45675 */ {(12<<2)|2,{114,103,0}},
/* 45676 */ {(12<<2)|2,{114,103,0}},
/* 45677 */ {(12<<2)|2,{114,103,0}},
/* 45678 */ {(12<<2)|2,{114,103,0}},
/* 45679 */ {(12<<2)|2,{114,103,0}},
/* 45680 */ {(12<<2)|2,{114,104,0}},
/* 45681 */ {(12<<2)|2,{114,104,0}},
/* 45682 */ {(12<<2)|2,{114,104,0}},
/* 45683 */ {(12<<2)|2,{114,104,0}},
/* 45684 */ {(12<<2)|2,{114,104,0}},
/* 45685 */ {(12<<2)|2,{114,104,0}},
/* 45686 */ {(12<<2)|2,{114,104,0}},
/* 45687 */ {(12<<2)|2,{114,104,0}},
/* 45688 */ {(12<<2)|2,{114,104,0}},
/* 45689 */ {(12<<2)|2,{114,104,0}},
/* 45690 */ {(12<<2)|2,{114,104,0}},
/* 45691 */ {(12<<2)|2,{114,104,0}},
/* 45692 */ {(12<<2)|2,{114,104,0}},
/* 45693 */ {(12<<2)|2,{114,104,0}},
/* 45694 */ {(12<<2)|2,{114,104,0}},
/* 45695 */ {(12<<2)|2,{114,104,0}},
/* 45696 */ {(12<<2)|2,{114,108,0}},
/* 45697 */ {(12<<2)|2,{114,108,0}},
/* 45698 */ {(12<<2)|2,{114,108,0}},
/* 45699 */ {(12<<2)|2,{114,108,0}},
/* 45700 */ {(12<<2)|2,{114,108,0}},
/* 45701 */ {(12<<2)|2,{114,108,0}},
/* 45702 */ {(12<<2)|2,{114,108,0}},
/* 45703 */ {(12<<2)|2,{114,108,0}},
/* 45704 */ {(12<<2)|2,{114,108,0}},
/* 45705 */ {(12<<2)|2,{114,108,0}},
/* 45706 */ {(12<<2)|2,{114,108,0}},
/* 45707 */ {(12<<2)|2,{114,108,0}},
/* 45708 */ {(12<<2)|2,{114,108,0}},
/* 45709 */ {(12<<2)|2,{114,108,0}},
/* 45710 */ {(12<<2)|2,{114,108,0}},
/* 45711 */ {(12<<2)|2,{114,108,0}},
/* 45712 */ {(12<<2)|2,{114,109,0}},
/* 45713 */ {(12<<2)|2,{114,109,0}},
/* 45714 */ {(12<<2)|2,{114,109,0}},
/* 45715 */ {(12<<2)|2,{114,109,0}},
/* 45716 */ {(12<<2)|2,{114,109,0}},
/* 45717 */ {(12<<2)|2,{114,109,0}},
/* 45718 */ {(12<<2)|2,{114,109,0}},
/* 45719 */ {(12<<2)|2,{114,109,0}},
/* 45720 */ {(12<<2)|2,{114,109,0}},
/* 45721 */ {(12<<2)|2,{114,109,0}},
/* 45722 */ {(12<<2)|2,{114,109,0}},
/* 45723 */ {(12<<2)|2,{114,109,0}},
/* 45724 */ {(12<<2)|2,{114,109,0}},
/* 45725 */ {(12<<2)|2,{114,109,0}},
/* 45726 */ {(12<<2)|2,{114,109,0}},
/* 45727 */ {(12<<2)|2,{114,109,0}},
/* 45728 */ {(12<<2)|2,{114,110,0}},
/* 45729 */ {(12<<2)|2,{114,110,0}},
/* 45730 */ {(12<<2)|2,{114,110,0}},
/* 45731 */ {(12<<2)|2,{114,110,0}},
/* 45732 */ {(12<<2)|2,{114,110,0}},
/* 45733 */ {(12<<2)|2,{114,110,0}},
/* 45734 */ {(12<<2)|2,{114,110,0}},
/* 45735 */ {(12<<2)|2,{114,110,0}},
/* 45736 */ {(12<<2)|2,{114,110,0}},
/* 45737 */ {(12<<2)|2,{114,110,0}},
/* 45738 */ {(12<<2)|2,{114,110,0}},
/* 45739 */ {(12<<2)|2,{114,110,0}},
/* 45740 */ {(12<<2)|2,{114,110,0}},
/* 45741 */ {(12<<2)|2,{114,110,0}},
/* 45742 */ {(12<<2)|2,{114,110,0}},
/* 45743 */ {(12<<2)|2,{114,110,0}},
/* 45744 */ {(12<<2)|2,{114,112,0}},
/* 45745 */ {(12<<2)|2,{114,112,0}},
/* 45746 */ {(12<<2)|2,{114,112,0}},
/* 45747 */ {(12<<2)|2,{114,112,0}},
/* 45748 */ {(12<<2)|2,{114,112,0}},
/* 45749 */ {(12<<2)|2,{114,112,0}},
/* 45750 */ {(12<<2)|2,{114,112,0}},
/* 45751 */ {(12<<2)|2,{114,112,0}},
/* 45752 */ {(12<<2)|2,{114,112,0}},
/* 45753 */ {(12<<2)|2,{114,112,0}},
/* 45754 */ {(12<<2)|2,{114,112,0}},
/* 45755 */ {(12<<2)|2,{114,112,0}},
/* 45756 */ {(12<<2)|2,{114,112,0}},
/* 45757 */ {(12<<2)|2,{114,112,0}},
/* 45758 */ {(12<<2)|2,{114,112,0}},
/* 45759 */ {(12<<2)|2,{114,112,0}},
/* 45760 */ {(12<<2)|2,{114,114,0}},
/* 45761 */ {(12<<2)|2,{114,114,0}},
/* 45762 */ {(12<<2)|2,{114,114,0}},
/* 45763 */ {(12<<2)|2,{114,114,0}},
/* 45764 */ {(12<<2)|2,{114,114,0}},
/* 45765 */ {(12<<2)|2,{114,114,0}},
/* 45766 */ {(12<<2)|2,{114,114,0}},
/* 45767 */ {(12<<2)|2,{114,114,0}},
/* 45768 */ {(12<<2)|2,{114,114,0}},
/* 45769 */ {(12<<2)|2,{114,114,0}},
/* 45770 */ {(12<<2)|2,{114,114,0}},
/* 45771 */ {(12<<2)|2,{114,114,0}},
/* 45772 */ {(12<<2)|2,{114,114,0}},
/* 45773 */ {(12<<2)|2,{114,114,0}},
/* 45774 */ {(12<<2)|2,{114,114,0}},
/* 45775 */ {(12<<2)|2,{114,114,0}},
/* 45776 */ {(12<<2)|2,{114,117,0}},
/* 45777 */ {(12<<2)|2,{114,117,0}},
/* 45778 */ {(12<<2)|2,{114,117,0}},
/* 45779 */ {(12<<2)|2,{114,117,0}},
/* 45780 */ {(12<<2)|2,{114,117,0}},
/* 45781 */ {(12<<2)|2,{114,117,0}},
/* 45782 */ {(12<<2)|2,{114,117,0}},
/* 45783 */ {(12<<2)|2,{114,117,0}},
/* 45784 */ {(12<<2)|2,{114,117,0}},
/* 45785 */ {(12<<2)|2,{114,117,0}},
/* 45786 */ {(12<<2)|2,{114,117,0}},
/* 45787 */ {(12<<2)|2,{114,117,0}},
/* 45788 */ {(12<<2)|2,{114,117,0}},
/* 45789 */ {(12<<2)|2,{114,117,0}},
/* 45790 */ {(12<<2)|2,{114,117,0}},
/* 45791 */ {(12<<2)|2,{114,117,0}},
/* 45792 */ {(13<<2)|2,{114,58,0}},
/* 45793 */ {(13<<2)|2,{114,58,0}},
/* 45794 */ {(13<<2)|2,{114,58,0}},
/* 45795 */ {(13<<2)|2,{114,58,0}},
/* 45796 */ {(13<<2)|2,{114,58,0}},
/* 45797 */ {(13<<2)|2,{114,58,0}},
/* 45798 */ {(13<<2)|2,{114,58,0}},
/* 45799 */ {(13<<2)|2,{114,58,0}},
/* 45800 */ {(13<<2)|2,{114,66,0}},
/* 45801 */ {(13<<2)|2,{114,66,0}},
/* 45802 */ {(13<<2)|2,{114,66,0}},
/* 45803 */ {(13<<2)|2,{114,66,0}},
/* 45804 */ {(13<<2)|2,{114,66,0}},
/* 45805 */ {(13<<2)|2,{114,66,0}},
/* 45806 */ {(13<<2)|2,{114,66,0}},
/* 45807 */ {(13<<2)|2,{114,66,0}},
/* 45808 */ {(13<<2)|2,{114,67,0}},
/* 45809 */ {(13<<2)|2,{114,67,0}},
/* 45810 */ {(13<<2)|2,{114,67,0}},
/* 45811 */ {(13<<2)|2,{114,67,0}},
/* 45812 */ {(13<<2)|2,{114,67,0}},
/* 45813 */ {(13<<2)|2,{114,67,0}},
/* 45814 */ {(13<<2)|2,{114,67,0}},
/* 45815 */ {(13<<2)|2,{114,67,0}},
/* 45816 */ {(13<<2)|2,{114,68,0}},
/* 45817 */ {(13<<2)|2,{114,68,0}},
/* 45818 */ {(13<<2)|2,{114,68,0}},
/* 45819 */ {(13<<2)|2,{114,68,0}},
/* 45820 */ {(13<<2)|2,{114,68,0}},
/* 45821 */ {(13<<2)|2,{114,68,0}},
/* 45822 */ {(13<<2)|2,{114,68,0}},
/* 45823 */ {(13<<2)|2,{114,68,0}},
/* 45824 */ {(13<<2)|2,{114,69,0}},
/* 45825 */ {(13<<2)|2,{114,69,0}},
/* 45826 */ {(13<<2)|2,{114,69,0}},
/* 45827 */ {(13<<2)|2,{114,69,0}},
/* 45828 */ {(13<<2)|2,{114,69,0}},
/* 45829 */ {(13<<2)|2,{114,69,0}},
/* 45830 */ {(13<<2)|2,{114,69,0}},
/* 45831 */ {(13<<2)|2,{114,69,0}},
/* 45832 */ {(13<<2)|2,{114,70,0}},
/* 45833 */ {(13<<2)|2,{114,70,0}},
/* 45834 */ {(13<<2)|2,{114,70,0}},
/* 45835 */ {(13<<2)|2,{114,70,0}},
/* 45836 */ {(13<<2)|2,{114,70,0}},
/* 45837 */ {(13<<2)|2,{114,70,0}},
/* 45838 */ {(13<<2)|2,{114,70,0}},
/* 45839 */ {(13<<2)|2,{114,70,0}},
/* 45840 */ {(13<<2)|2,{114,71,0}},
/* 45841 */ {(13<<2)|2,{114,71,0}},
/* 45842 */ {(13<<2)|2,{114,71,0}},
/* 45843 */ {(13<<2)|2,{114,71,0}},
/* 45844 */ {(13<<2)|2,{114,71,0}},
/* 45845 */ {(13<<2)|2,{114,71,0}},
/* 45846 */ {(13<<2)|2,{114,71,0}},
/* 45847 */ {(13<<2)|2,{114,71,0}},
/* 45848 */ {(13<<2)|2,{114,72,0}},
/* 45849 */ {(13<<2)|2,{114,72,0}},
/* 45850 */ {(13<<2)|2,{114,72,0}},
/* 45851 */ {(13<<2)|2,{114,72,0}},
/* 45852 */ {(13<<2)|2,{114,72,0}},
/* 45853 */ {(13<<2)|2,{114,72,0}},
/* 45854 */ {(13<<2)|2,{114,72,0}},
/* 45855 */ {(13<<2)|2,{114,72,0}},
/* 45856 */ {(13<<2)|2,{114,73,0}},
/* 45857 */ {(13<<2)|2,{114,73,0}},
/* 45858 */ {(13<<2)|2,{114,73,0}},
/* 45859 */ {(13<<2)|2,{114,73,0}},
/* 45860 */ {(13<<2)|2,{114,73,0}},
/* 45861 */ {(13<<2)|2,{114,73,0}},
/* 45862 */ {(13<<2)|2,{114,73,0}},
/* 45863 */ {(13<<2)|2,{114,73,0}},
/* 45864 */ {(13<<2)|2,{114,74,0}},
/* 45865 */ {(13<<2)|2,{114,74,0}},
/* 45866 */ {(13<<2)|2,{114,74,0}},
/* 45867 */ {(13<<2)|2,{114,74,0}},
/* 45868 */ {(13<<2)|2,{114,74,0}},
/* 45869 */ {(13<<2)|2,{114,74,0}},
/* 45870 */ {(13<<2)|2,{114,74,0}},
/* 45871 */ {(13<<2)|2,{114,74,0}},
/* 45872 */ {(13<<2)|2,{114,75,0}},
/* 45873 */ {(13<<2)|2,{114,75,0}},
/* 45874 */ {(13<<2)|2,{114,75,0}},
/* 45875 */ {(13<<2)|2,{114,75,0}},
/* 45876 */ {(13<<2)|2,{114,75,0}},
/* 45877 */ {(13<<2)|2,{114,75,0}},
/* 45878 */ {(13<<2)|2,{114,75,0}},
/* 45879 */ {(13<<2)|2,{114,75,0}},
/* 45880 */ {(13<<2)|2,{114,76,0}},
/* 45881 */ {(13<<2)|2,{114,76,0}},
/* 45882 */ {(13<<2)|2,{114,76,0}},
/* 45883 */ {(13<<2)|2,{114,76,0}},
/* 45884 */ {(13<<2)|2,{114,76,0}},
/* 45885 */ {(13<<2)|2,{114,76,0}},
/* 45886 */ {(13<<2)|2,{114,76,0}},
/* 45887 */ {(13<<2)|2,{114,76,0}},
/* 45888 */ {(13<<2)|2,{114,77,0}},
/* 45889 */ {(13<<2)|2,{114,77,0}},
/* 45890 */ {(13<<2)|2,{114,77,0}},
/* 45891 */ {(13<<2)|2,{114,77,0}},
/* 45892 */ {(13<<2)|2,{114,77,0}},
/* 45893 */ {(13<<2)|2,{114,77,0}},
/* 45894 */ {(13<<2)|2,{114,77,0}},
/* 45895 */ {(13<<2)|2,{114,77,0}},
/* 45896 */ {(13<<2)|2,{114,78,0}},
/* 45897 */ {(13<<2)|2,{114,78,0}},
/* 45898 */ {(13<<2)|2,{114,78,0}},
/* 45899 */ {(13<<2)|2,{114,78,0}},
/* 45900 */ {(13<<2)|2,{114,78,0}},
/* 45901 */ {(13<<2)|2,{114,78,0}},
/* 45902 */ {(13<<2)|2,{114,78,0}},
/* 45903 */ {(13<<2)|2,{114,78,0}},
/* 45904 */ {(13<<2)|2,{114,79,0}},
/* 45905 */ {(13<<2)|2,{114,79,0}},
/* 45906 */ {(13<<2)|2,{114,79,0}},
/* 45907 */ {(13<<2)|2,{114,79,0}},
/* 45908 */ {(13<<2)|2,{114,79,0}},
/* 45909 */ {(13<<2)|2,{114,79,0}},
/* 45910 */ {(13<<2)|2,{114,79,0}},
/* 45911 */ {(13<<2)|2,{114,79,0}},
/* 45912 */ {(13<<2)|2,{114,80,0}},
/* 45913 */ {(13<<2)|2,{114,80,0}},
/* 45914 */ {(13<<2)|2,{114,80,0}},
/* 45915 */ {(13<<2)|2,{114,80,0}},
/* 45916 */ {(13<<2)|2,{114,80,0}},
/* 45917 */ {(13<<2)|2,{114,80,0}},
/* 45918 */ {(13<<2)|2,{114,80,0}},
/* 45919 */ {(13<<2)|2,{114,80,0}},
/* 45920 */ {(13<<2)|2,{114,81,0}},
/* 45921 */ {(13<<2)|2,{114,81,0}},
/* 45922 */ {(13<<2)|2,{114,81,0}},
/* 45923 */ {(13<<2)|2,{114,81,0}},
/* 45924 */ {(13<<2)|2,{114,81,0}},
/* 45925 */ {(13<<2)|2,{114,81,0}},
/* 45926 */ {(13<<2)|2,{114,81,0}},
/* 45927 */ {(13<<2)|2,{114,81,0}},
/* 45928 */ {(13<<2)|2,{114,82,0}},
/* 45929 */ {(13<<2)|2,{114,82,0}},
/* 45930 */ {(13<<2)|2,{114,82,0}},
/* 45931 */ {(13<<2)|2,{114,82,0}},
/* 45932 */ {(13<<2)|2,{114,82,0}},
/* 45933 */ {(13<<2)|2,{114,82,0}},
/* 45934 */ {(13<<2)|2,{114,82,0}},
/* 45935 */ {(13<<2)|2,{114,82,0}},
/* 45936 */ {(13<<2)|2,{114,83,0}},
/* 45937 */ {(13<<2)|2,{114,83,0}},
/* 45938 */ {(13<<2)|2,{114,83,0}},
/* 45939 */ {(13<<2)|2,{114,83,0}},
/* 45940 */ {(13<<2)|2,{114,83,0}},
/* 45941 */ {(13<<2)|2,{114,83,0}},
/* 45942 */ {(13<<2)|2,{114,83,0}},
/* 45943 */ {(13<<2)|2,{114,83,0}},
/* 45944 */ {(13<<2)|2,{114,84,0}},
/* 45945 */ {(13<<2)|2,{114,84,0}},
/* 45946 */ {(13<<2)|2,{114,84,0}},
/* 45947 */ {(13<<2)|2,{114,84,0}},
/* 45948 */ {(13<<2)|2,{114,84,0}},
/* 45949 */ {(13<<2)|2,{114,84,0}},
/* 45950 */ {(13<<2)|2,{114,84,0}},
/* 45951 */ {(13<<2)|2,{114,84,0}},
/* 45952 */ {(13<<2)|2,{114,85,0}},
/* 45953 */ {(13<<2)|2,{114,85,0}},
/* 45954 */ {(13<<2)|2,{114,85,0}},
/* 45955 */ {(13<<2)|2,{114,85,0}},
/* 45956 */ {(13<<2)|2,{114,85,0}},
/* 45957 */ {(13<<2)|2,{114,85,0}},
/* 45958 */ {(13<<2)|2,{114,85,0}},
/* 45959 */ {(13<<2)|2,{114,85,0}},
/* 45960 */ {(13<<2)|2,{114,86,0}},
/* 45961 */ {(13<<2)|2,{114,86,0}},
/* 45962 */ {(13<<2)|2,{114,86,0}},
/* 45963 */ {(13<<2)|2,{114,86,0}},
/* 45964 */ {(13<<2)|2,{114,86,0}},
/* 45965 */ {(13<<2)|2,{114,86,0}},
/* 45966 */ {(13<<2)|2,{114,86,0}},
/* 45967 */ {(13<<2)|2,{114,86,0}},
/* 45968 */ {(13<<2)|2,{114,87,0}},
/* 45969 */ {(13<<2)|2,{114,87,0}},
/* 45970 */ {(13<<2)|2,{114,87,0}},
/* 45971 */ {(13<<2)|2,{114,87,0}},
/* 45972 */ {(13<<2)|2,{114,87,0}},
/* 45973 */ {(13<<2)|2,{114,87,0}},
/* 45974 */ {(13<<2)|2,{114,87,0}},
/* 45975 */ {(13<<2)|2,{114,87,0}},
/* 45976 */ {(13<<2)|2,{114,89,0}},
/* 45977 */ {(13<<2)|2,{114,89,0}},
/* 45978 */ {(13<<2)|2,{114,89,0}},
/* 45979 */ {(13<<2)|2,{114,89,0}},
/* 45980 */ {(13<<2)|2,{114,89,0}},
/* 45981 */ {(13<<2)|2,{114,89,0}},
/* 45982 */ {(13<<2)|2,{114,89,0}},
/* 45983 */ {(13<<2)|2,{114,89,0}},
/* 45984 */ {(13<<2)|2,{114,106,0}},
/* 45985 */ {(13<<2)|2,{114,106,0}},
/* 45986 */ {(13<<2)|2,{114,106,0}},
/* 45987 */ {(13<<2)|2,{114,106,0}},
/* 45988 */ {(13<<2)|2,{114,106,0}},
/* 45989 */ {(13<<2)|2,{114,106,0}},
/* 45990 */ {(13<<2)|2,{114,106,0}},
/* 45991 */ {(13<<2)|2,{114,106,0}},
/* 45992 */ {(13<<2)|2,{114,107,0}},
/* 45993 */ {(13<<2)|2,{114,107,0}},
/* 45994 */ {(13<<2)|2,{114,107,0}},
/* 45995 */ {(13<<2)|2,{114,107,0}},
/* 45996 */ {(13<<2)|2,{114,107,0}},
/* 45997 */ {(13<<2)|2,{114,107,0}},
/* 45998 */ {(13<<2)|2,{114,107,0}},
/* 45999 */ {(13<<2)|2,{114,107,0}},
/* 46000 */ {(13<<2)|2,{114,113,0}},
/* 46001 */ {(13<<2)|2,{114,113,0}},
/* 46002 */ {(13<<2)|2,{114,113,0}},
/* 46003 */ {(13<<2)|2,{114,113,0}},
/* 46004 */ {(13<<2)|2,{114,113,0}},
/* 46005 */ {(13<<2)|2,{114,113,0}},
/* 46006 */ {(13<<2)|2,{114,113,0}},
/* 46007 */ {(13<<2)|2,{114,113,0}},
/* 46008 */ {(13<<2)|2,{114,118,0}},
/* 46009 */ {(13<<2)|2,{114,118,0}},
/* 46010 */ {(13<<2)|2,{114,118,0}},
/* 46011 */ {(13<<2)|2,{114,118,0}},
/* 46012 */ {(13<<2)|2,{114,118,0}},
/* 46013 */ {(13<<2)|2,{114,118,0}},
/* 46014 */ {(13<<2)|2,{114,118,0}},
/* 46015 */ {(13<<2)|2,{114,118,0}},
/* 46016 */ {(13<<2)|2,{114,119,0}},
/* 46017 */ {(13<<2)|2,{114,119,0}},
/* 46018 */ {(13<<2)|2,{114,119,0}},
/* 46019 */ {(13<<2)|2,{114,119,0}},
/* 46020 */ {(13<<2)|2,{114,119,0}},
/* 46021 */ {(13<<2)|2,{114,119,0}},
/* 46022 */ {(13<<2)|2,{114,119,0}},
/* 46023 */ {(13<<2)|2,{114,119,0}},
/* 46024 */ {(13<<2)|2,{114,120,0}},
/* 46025 */ {(13<<2)|2,{114,120,0}},
/* 46026 */ {(13<<2)|2,{114,120,0}},
/* 46027 */ {(13<<2)|2,{114,120,0}},
/* 46028 */ {(13<<2)|2,{114,120,0}},
/* 46029 */ {(13<<2)|2,{114,120,0}},
/* 46030 */ {(13<<2)|2,{114,120,0}},
/* 46031 */ {(13<<2)|2,{114,120,0}},
/* 46032 */ {(13<<2)|2,{114,121,0}},
/* 46033 */ {(13<<2)|2,{114,121,0}},
/* 46034 */ {(13<<2)|2,{114,121,0}},
/* 46035 */ {(13<<2)|2,{114,121,0}},
/* 46036 */ {(13<<2)|2,{114,121,0}},
/* 46037 */ {(13<<2)|2,{114,121,0}},
/* 46038 */ {(13<<2)|2,{114,121,0}},
/* 46039 */ {(13<<2)|2,{114,121,0}},
/* 46040 */ {(13<<2)|2,{114,122,0}},
/* 46041 */ {(13<<2)|2,{114,122,0}},
/* 46042 */ {(13<<2)|2,{114,122,0}},
/* 46043 */ {(13<<2)|2,{114,122,0}},
/* 46044 */ {(13<<2)|2,{114,122,0}},
/* 46045 */ {(13<<2)|2,{114,122,0}},
/* 46046 */ {(13<<2)|2,{114,122,0}},
/* 46047 */ {(13<<2)|2,{114,122,0}},
/* 46048 */ {(14<<2)|2,{114,38,0}},
/* 46049 */ {(14<<2)|2,{114,38,0}},
/* 46050 */ {(14<<2)|2,{114,38,0}},
/* 46051 */ {(14<<2)|2,{114,38,0}},
/* 46052 */ {(14<<2)|2,{114,42,0}},
/* 46053 */ {(14<<2)|2,{114,42,0}},
/* 46054 */ {(14<<2)|2,{114,42,0}},
/* 46055 */ {(14<<2)|2,{114,42,0}},
/* 46056 */ {(14<<2)|2,{114,44,0}},
/* 46057 */ {(14<<2)|2,{114,44,0}},
/* 46058 */ {(14<<2)|2,{114,44,0}},
/* 46059 */ {(14<<2)|2,{114,44,0}},
/* 46060 */ {(14<<2)|2,{114,59,0}},
/* 46061 */ {(14<<2)|2,{114,59,0}},
/* 46062 */ {(14<<2)|2,{114,59,0}},
/* 46063 */ {(14<<2)|2,{114,59,0}},
/* 46064 */ {(14<<2)|2,{114,88,0}},
/* 46065 */ {(14<<2)|2,{114,88,0}},
/* 46066 */ {(14<<2)|2,{114,88,0}},
/* 46067 */ {(14<<2)|2,{114,88,0}},
/* 46068 */ {(14<<2)|2,{114,90,0}},
/* 46069 */ {(14<<2)|2,{114,90,0}},
/* 46070 */ {(14<<2)|2,{114,90,0}},
/* 46071 */ {(14<<2)|2,{114,90,0}},
/* 46072 */ {(16<<2)|2,{114,33,0}},
/* 46073 */ {(16<<2)|2,{114,34,0}},
/* 46074 */ {(16<<2)|2,{114,40,0}},
/* 46075 */ {(16<<2)|2,{114,41,0}},
/* 46076 */ {(16<<2)|2,{114,63,0}},
/* 46077 */ {(6<<2)|1,{114,0,0}},
/* 46078 */ {(6<<2)|1,{114,0,0}},
/* 46079 */ {(6<<2)|1,{114,0,0}},
/* 46080 */ {(16<<2)|3,{117,48,48}},
/* 46081 */ {(16<<2)|3,{117,48,49}},
/* 46082 */ {(16<<2)|3,{117,48,50}},
/* 46083 */ {(16<<2)|3,{117,48,97}},
/* 46084 */ {(16<<2)|3,{117,48,99}},
/* 46085 */ {(16<<2)|3,{117,48,101}},
/* 46086 */ {(16<<2)|3,{117,48,105}},
/* 46087 */ {(16<<2)|3,{117,48,111}},
/* 46088 */ {(16<<2)|3,{117,48,115}},
/* 46089 */ {(16<<2)|3,{117,48,116}},
/* 46090 */ {(11<<2)|2,{117,48,0}},
/* 46091 */ {(11<<2)|2,{117,48,0}},
/* 46092 */ {(11<<2)|2,{117,48,0}},
/* 46093 */ {(11<<2)|2,{117,48,0}},
/* 46094 */ {(11<<2)|2,{117,48,0}},
/* 46095 */ {(11<<2)|2,{117,48,0}},
/* 46096 */ {(11<<2)|2,{117,48,0}},
/* 46097 */ {(11<<2)|2,{117,48,0}},
/* 46098 */ {(11<<2)|2,{117,48,0}},
/* 46099 */ {(11<<2)|2,{117,48,0}},
/* 46100 */ {(11<<2)|2,{117,48,0}},
/* 46101 */ {(11<<2)|2,{117,48,0}},
/* 46102 */ {(11<<2)|2,{117,48,0}},
/* 46103 */ {(11<<2)|2,{117,48,0}},
/* 46104 */ {(11<<2)|2,{117,48,0}},
/* 46105 */ {(11<<2)|2,{117,48,0}},
/* 46106 */ {(11<<2)|2,{117,48,0}},
/* 46107 */ {(11<<2)|2,{117,48,0}},
/* 46108 */ {(11<<2)|2,{117,48,0}},
/* 46109 */ {(11<<2)|2,{117,48,0}},
/* 46110 */ {(11<<2)|2,{117,48,0}},
/* 46111 */ {(11<<2)|2,{117,48,0}},
/* 46112 */ {(16<<2)|3,{117,49,48}},
/* 46113 */ {(16<<2)|3,{117,49,49}},
/* 46114 */ {(16<<2)|3,{117,49,50}},
/* 46115 */ {(16<<2)|3,{117,49,97}},
/* 46116 */ {(16<<2)|3,{117,49,99}},
/* 46117 */ {(16<<2)|3,{117,49,101}},
/* 46118 */ {(16<<2)|3,{117,49,105}},
/* 46119 */ {(16<<2)|3,{117,49,111}},
/* 46120 */ {(16<<2)|3,{117,49,115}},
/* 46121 */ {(16<<2)|3,{117,49,116}},
/* 46122 */ {(11<<2)|2,{117,49,0}},
/* 46123 */ {(11<<2)|2,{117,49,0}},
/* 46124 */ {(11<<2)|2,{117,49,0}},
/* 46125 */ {(11<<2)|2,{117,49,0}},
/* 46126 */ {(11<<2)|2,{117,49,0}},
/* 46127 */ {(11<<2)|2,{117,49,0}},
/* 46128 */ {(11<<2)|2,{117,49,0}},
/* 46129 */ {(11<<2)|2,{117,49,0}},
/* 46130 */ {(11<<2)|2,{117,49,0}},
/* 46131 */ {(11<<2)|2,{117,49,0}},
/* 46132 */ {(11<<2)|2,{117,49,0}},
/* 46133 */ {(11<<2)|2,{117,49,0}},
/* 46134 */ {(11<<2)|2,{117,49,0}},
/* 46135 */ {(11<<2)|2,{117,49,0}},
/* 46136 */ {(11<<2)|2,{117,49,0}},
/* 46137 */ {(11<<2)|2,{117,49,0}},
/* 46138 */ {(11<<2)|2,{117,49,0}},
/* 46139 */ {(11<<2)|2,{117,49,0}},
/* 46140 */ {(11<<2)|2,{117,49,0}},
/* 46141 */ {(11<<2)|2,{117,49,0}},
/* 46142 */ {(11<<2)|2,{117,49,0}},
/* 46143 */ {(11<<2)|2,{117,49,0}},
/* 46144 */ {(16<<2)|3,{117,50,48}},
/* 46145 */ {(16<<2)|3,{117,50,49}},
/* 46146 */ {(16<<2)|3,{117,50,50}},
/* 46147 */ {(16<<2)|3,{117,50,97}},
/* 46148 */ {(16<<2)|3,{117,50,99}},
/* 46149 */ {(16<<2)|3,{117,50,101}},
/* 46150 */ {(16<<2)|3,{117,50,105}},
/* 46151 */ {(16<<2)|3,{117,50,111}},
/* 46152 */ {(16<<2)|3,{117,50,115}},
/* 46153 */ {(16<<2)|3,{117,50,116}},
/* 46154 */ {(11<<2)|2,{117,50,0}},
/* 46155 */ {(11<<2)|2,{117,50,0}},
/* 46156 */ {(11<<2)|2,{117,50,0}},
/* 46157 */ {(11<<2)|2,{117,50,0}},
/* 46158 */ {(11<<2)|2,{117,50,0}},
/* 46159 */ {(11<<2)|2,{117,50,0}},
/* 46160 */ {(11<<2)|2,{117,50,0}},
/* 46161 */ {(11<<2)|2,{117,50,0}},
/* 46162 */ {(11<<2)|2,{117,50,0}},
/* 46163 */ {(11<<2)|2,{117,50,0}},
/* 46164 */ {(11<<2)|2,{117,50,0}},
/* 46165 */ {(11<<2)|2,{117,50,0}},
/* 46166 */ {(11<<2)|2,{117,50,0}},
/* 46167 */ {(11<<2)|2,{117,50,0}},
/* 46168 */ {(11<<2)|2,{117,50,0}},
/* 46169 */ {(11<<2)|2,{117,50,0}},
/* 46170 */ {(11<<2)|2,{117,50,0}},
/* 46171 */ {(11<<2)|2,{117,50,0}},
/* 46172 */ {(11<<2)|2,{117,50,0}},
/* 46173 */ {(11<<2)|2,{117,50,0}},
/* 46174 */ {(11<<2)|2,{117,50,0}},
/* 46175 */ {(11<<2)|2,{117,50,0}},
/* 46176 */ {(16<<2)|3,{117,97,48}},
/* 46177 */ {(16<<2)|3,{117,97,49}},
/* 46178 */ {(16<<2)|3,{117,97,50}},
/* 46179 */ {(16<<2)|3,{117,97,97}},
/* 46180 */ {(16<<2)|3,{117,97,99}},
/* 46181 */ {(16<<2)|3,{117,97,101}},
/* 46182 */ {(16<<2)|3,{117,97,105}},
/* 46183 */ {(16<<2)|3,{117,97,111}},
/* 46184 */ {(16<<2)|3,{117,97,115}},
/* 46185 */ {(16<<2)|3,{117,97,116}},
/* 46186 */ {(11<<2)|2,{117,97,0}},
/* 46187 */ {(11<<2)|2,{117,97,0}},
/* 46188 */ {(11<<2)|2,{117,97,0}},
/* 46189 */ {(11<<2)|2,{117,97,0}},
/* 46190 */ {(11<<2)|2,{117,97,0}},
/* 46191 */ {(11<<2)|2,{117,97,0}},
/* 46192 */ {(11<<2)|2,{117,97,0}},
/* 46193 */ {(11<<2)|2,{117,97,0}},
/* 46194 */ {(11<<2)|2,{117,97,0}},
/* 46195 */ {(11<<2)|2,{117,97,0}},
/* 46196 */ {(11<<2)|2,{117,97,0}},
/* 46197 */ {(11<<2)|2,{117,97,0}},
/* 46198 */ {(11<<2)|2,{117,97,0}},
/* 46199 */ {(11<<2)|2,{117,97,0}},
/* 46200 */ {(11<<2)|2,{117,97,0}},
/* 46201 */ {(11<<2)|2,{117,97,0}},
/* 46202 */ {(11<<2)|2,{117,97,0}},
/* 46203 */ {(11<<2)|2,{117,97,0}},
/* 46204 */ {(11<<2)|2,{117,97,0}},
/* 46205 */ {(11<<2)|2,{117,97,0}},
/* 46206 */ {(11<<2)|2,{117,97,0}},
/* 46207 */ {(11<<2)|2,{117,97,0}},
/* 46208 */ {(16<<2)|3,{117,99,48}},
/* 46209 */ {(16<<2)|3,{117,99,49}},
/* 46210 */ {(16<<2)|3,{117,99,50}},
/* 46211 */ {(16<<2)|3,{117,99,97}},
/* 46212 */ {(16<<2)|3,{117,99,99}},
/* 46213 */ {(16<<2)|3,{117,99,101}},
/* 46214 */ {(16<<2)|3,{117,99,105}},
/* 46215 */ {(16<<2)|3,{117,99,111}},
/* 46216 */ {(16<<2)|3,{117,99,115}},
/* 46217 */ {(16<<2)|3,{117,99,116}},
/* 46218 */ {(11<<2)|2,{117,99,0}},
/* 46219 */ {(11<<2)|2,{117,99,0}},
/* 46220 */ {(11<<2)|2,{117,99,0}},
/* 46221 */ {(11<<2)|2,{117,99,0}},
/* 46222 */ {(11<<2)|2,{117,99,0}},
/* 46223 */ {(11<<2)|2,{117,99,0}},
/* 46224 */ {(11<<2)|2,{117,99,0}},
/* 46225 */ {(11<<2)|2,{117,99,0}},
/* 46226 */ {(11<<2)|2,{117,99,0}},
/* 46227 */ {(11<<2)|2,{117,99,0}},
/* 46228 */ {(11<<2)|2,{117,99,0}},
/* 46229 */ {(11<<2)|2,{117,99,0}},
/* 46230 */ {(11<<2)|2,{117,99,0}},
/* 46231 */ {(11<<2)|2,{117,99,0}},
/* 46232 */ {(11<<2)|2,{117,99,0}},
/* 46233 */ {(11<<2)|2,{117,99,0}},
/* 46234 */ {(11<<2)|2,{117,99,0}},
/* 46235 */ {(11<<2)|2,{117,99,0}},
/* 46236 */ {(11<<2)|2,{117,99,0}},
/* 46237 */ {(11<<2)|2,{117,99,0}},
/* 46238 */ {(11<<2)|2,{117,99,0}},
/* 46239 */ {(11<<2)|2,{117,99,0}},
/* 46240 */ {(16<<2)|3,{117,101,48}},
/* 46241 */ {(16<<2)|3,{117,101,49}},
/* 46242 */ {(16<<2)|3,{117,101,50}},
/* 46243 */ {(16<<2)|3,{117,101,97}},
/* 46244 */ {(16<<2)|3,{117,101,99}},
/* 46245 */ {(16<<2)|3,{117,101,101}},
/* 46246 */ {(16<<2)|3,{117,101,105}},
/* 46247 */ {(16<<2)|3,{117,101,111}},
/* 46248 */ {(16<<2)|3,{117,101,115}},
/* 46249 */ {(16<<2)|3,{117,101,116}},
/* 46250 */ {(11<<2)|2,{117,101,0}},
/* 46251 */ {(11<<2)|2,{117,101,0}},
/* 46252 */ {(11<<2)|2,{117,101,0}},
/* 46253 */ {(11<<2)|2,{117,101,0}},
/* 46254 */ {(11<<2)|2,{117,101,0}},
/* 46255 */ {(11<<2)|2,{117,101,0}},
/* 46256 */ {(11<<2)|2,{117,101,0}},
/* 46257 */ {(11<<2)|2,{117,101,0}},
/* 46258 */ {(11<<2)|2,{117,101,0}},
/* 46259 */ {(11<<2)|2,{117,101,0}},
/* 46260 */ {(11<<2)|2,{117,101,0}},
/* 46261 */ {(11<<2)|2,{117,101,0}},
/* 46262 */ {(11<<2)|2,{117,101,0}},
/* 46263 */ {(11<<2)|2,{117,101,0}},
/* 46264 */ {(11<<2)|2,{117,101,0}},
/* 46265 */ {(11<<2)|2,{117,101,0}},
/* 46266 */ {(11<<2)|2,{117,101,0}},
/* 46267 */ {(11<<2)|2,{117,101,0}},
/* 46268 */ {(11<<2)|2,{117,101,0}},
/* 46269 */ {(11<<2)|2,{117,101,0}},
/* 46270 */ {(11<<2)|2,{117,101,0}},
/* 46271 */ {(11<<2)|2,{117,101,0}},
/* 46272 */ {(16<<2)|3,{117,105,48}},
/* 46273 */ {(16<<2)|3,{117,105,49}},
/* 46274 */ {(16<<2)|3,{117,105,50}},
/* 46275 */ {(16<<2)|3,{117,105,97}},
/* 46276 */ {(16<<2)|3,{117,105,99}},
/* 46277 */ {(16<<2)|3,{117,105,101}},
/* 46278 */ {(16<<2)|3,{117,105,105}},
/* 46279 */ {(16<<2)|3,{117,105,111}},
/* 46280 */ {(16<<2)|3,{117,105,115}},
/* 46281 */ {(16<<2)|3,{117,105,116}},
/* 46282 */ {(11<<2)|2,{117,105,0}},
/* 46283 */ {(11<<2)|2,{117,105,0}},
/* 46284 */ {(11<<2)|2,{117,105,0}},
/* 46285 */ {(11<<2)|2,{117,105,0}},
/* 46286 */ {(11<<2)|2,{117,105,0}},
/* 46287 */ {(11<<2)|2,{117,105,0}},
/* 46288 */ {(11<<2)|2,{117,105,0}},
/* 46289 */ {(11<<2)|2,{117,105,0}},
/* 46290 */ {(11<<2)|2,{117,105,0}},
/* 46291 */ {(11<<2)|2,{117,105,0}},
/* 46292 */ {(11<<2)|2,{117,105,0}},
/* 46293 */ {(11<<2)|2,{117,105,0}},
/* 46294 */ {(11<<2)|2,{117,105,0}},
/* 46295 */ {(11<<2)|2,{117,105,0}},
/* 46296 */ {(11<<2)|2,{117,105,0}},
/* 46297 */ {(11<<2)|2,{117,105,0}},
/* 46298 */ {(11<<2)|2,{117,105,0}},
/* 46299 */ {(11<<2)|2,{117,105,0}},
/* 46300 */ {(11<<2)|2,{117,105,0}},
/* 46301 */ {(11<<2)|2,{117,105,0}},
/* 46302 */ {(11<<2)|2,{117,105,0}},
/* 46303 */ {(11<<2)|2,{117,105,0}},
/* 46304 */ {(16<<2)|3,{117,111,48}},
/* 46305 */ {(16<<2)|3,{117,111,49}},
/* 46306 */ {(16<<2)|3,{117,111,50}},
/* 46307 */ {(16<<2)|3,{117,111,97}},
/* 46308 */ {(16<<2)|3,{117,111,99}},
/* 46309 */ {(16<<2)|3,{117,111,101}},
/* 46310 */ {(16<<2)|3,{117,111,105}},
/* 46311 */ {(16<<2)|3,{117,111,111}},
/* 46312 */ {(16<<2)|3,{117,111,115}},
/* 46313 */ {(16<<2)|3,{117,111,116}},
/* 46314 */ {(11<<2)|2,{117,111,0}},
/* 46315 */ {(11<<2)|2,{117,111,0}},
/* 46316 */ {(11<<2)|2,{117,111,0}},
/* 46317 */ {(11<<2)|2,{117,111,0}},
/* 46318 */ {(11<<2)|2,{117,111,0}},
/* 46319 */ {(11<<2)|2,{117,111,0}},
/* 46320 */ {(11<<2)|2,{117,111,0}},
/* 46321 */ {(11<<2)|2,{117,111,0}},
/* 46322 */ {(11<<2)|2,{117,111,0}},
/* 46323 */ {(11<<2)|2,{117,111,0}},
/* 46324 */ {(11<<2)|2,{117,111,0}},
/* 46325 */ {(11<<2)|2,{117,111,0}},
/* 46326 */ {(11<<2)|2,{117,111,0}},
/* 46327 */ {(11<<2)|2,{117,111,0}},
/* 46328 */ {(11<<2)|2,{117,111,0}},
/* 46329 */ {(11<<2)|2,{117,111,0}},
/* 46330 */ {(11<<2)|2,{117,111,0}},
/* 46331 */ {(11<<2)|2,{117,111,0}},
/* 46332 */ {(11<<2)|2,{117,111,0}},
/* 46333 */ {(11<<2)|2,{117,111,0}},
/* 46334 */ {(11<<2)|2,{117,111,0}},
/* 46335 */ {(11<<2)|2,{117,111,0}},
/* 46336 */ {(16<<2)|3,{117,115,48}},
/* 46337 */ {(16<<2)|3,{117,115,49}},
/* 46338 */ {(16<<2)|3,{117,115,50}},
/* 46339 */ {(16<<2)|3,{117,115,97}},
/* 46340 */ {(16<<2)|3,{117,115,99}},
/* 46341 */ {(16<<2)|3,{117,115,101}},
/* 46342 */ {(16<<2)|3,{117,115,105}},
/* 46343 */ {(16<<2)|3,{117,115,111}},
/* 46344 */ {(16<<2)|3,{117,115,115}},
/* 46345 */ {(16<<2)|3,{117,115,116}},
/* 46346 */ {(11<<2)|2,{117,115,0}},
/* 46347 */ {(11<<2)|2,{117,115,0}},
/* 46348 */ {(11<<2)|2,{117,115,0}},
/* 46349 */ {(11<<2)|2,{117,115,0}},
/* 46350 */ {(11<<2)|2,{117,115,0}},
/* 46351 */ {(11<<2)|2,{117,115,0}},
/* 46352 */ {(11<<2)|2,{117,115,0}},
/* 46353 */ {(11<<2)|2,{117,115,0}},
/* 46354 */ {(11<<2)|2,{117,115,0}},
/* 46355 */ {(11<<2)|2,{117,115,0}},
/* 46356 */ {(11<<2)|2,{117,115,0}},
/* 46357 */ {(11<<2)|2,{117,115,0}},
/* 46358 */ {(11<<2)|2,{117,115,0}},
/* 46359 */ {(11<<2)|2,{117,115,0}},
/* 46360 */ {(11<<2)|2,{117,115,0}},
/* 46361 */ {(11<<2)|2,{117,115,0}},
/* 46362 */ {(11<<2)|2,{117,115,0}},
/* 46363 */ {(11<<2)|2,{117,115,0}},
/* 46364 */ {(11<<2)|2,{117,115,0}},
/* 46365 */ {(11<<2)|2,{117,115,0}},
/* 46366 */ {(11<<2)|2,{117,115,0}},
/* 46367 */ {(11<<2)|2,{117,115,0}},
/* 46368 */ {(16<<2)|3,{117,116,48}},
/* 46369 */ {(16<<2)|3,{117,116,49}},
/* 46370 */ {(16<<2)|3,{117,116,50}},
/* 46371 */ {(16<<2)|3,{117,116,97}},
/* 46372 */ {(16<<2)|3,{117,116,99}},
/* 46373 */ {(16<<2)|3,{117,116,101}},
/* 46374 */ {(16<<2)|3,{117,116,105}},
/* 46375 */ {(16<<2)|3,{117,116,111}},
/* 46376 */ {(16<<2)|3,{117,116,115}},
/* 46377 */ {(16<<2)|3,{117,116,116}},
/* 46378 */ {(11<<2)|2,{117,116,0}},
/* 46379 */ {(11<<2)|2,{117,116,0}},
/* 46380 */ {(11<<2)|2,{117,116,0}},
/* 46381 */ {(11<<2)|2,{117,116,0}},
/* 46382 */ {(11<<2)|2,{117,116,0}},
/* 46383 */ {(11<<2)|2,{117,116,0}},
/* 46384 */ {(11<<2)|2,{117,116,0}},
/* 46385 */ {(11<<2)|2,{117,116,0}},
/* 46386 */ {(11<<2)|2,{117,116,0}},
/* 46387 */ {(11<<2)|2,{117,116,0}},
/* 46388 */ {(11<<2)|2,{117,116,0}},
/* 46389 */ {(11<<2)|2,{117,116,0}},
/* 46390 */ {(11<<2)|2,{117,116,0}},
/* 46391 */ {(11<<2)|2,{117,116,0}},
/* 46392 */ {(11<<2)|2,{117,116,0}},
/* 46393 */ {(11<<2)|2,{117,116,0}},
/* 46394 */ {(11<<2)|2,{117,116,0}},
/* 46395 */ {(11<<2)|2,{117,116,0}},
/* 46396 */ {(11<<2)|2,{117,116,0}},
/* 46397 */ {(11<<2)|2,{117,116,0}},
/* 46398 */ {(11<<2)|2,{117,116,0}},
/* 46399 */ {(11<<2)|2,{117,116,0}},
/* 46400 */ {(12<<2)|2,{117,32,0}},
/* 46401 */ {(12<<2)|2,{117,32,0}},
/* 46402 */ {(12<<2)|2,{117,32,0}},
/* 46403 */ {(12<<2)|2,{117,32,0}},
/* 46404 */ {(12<<2)|2,{117,32,0}},
/* 46405 */ {(12<<2)|2,{117,32,0}},
/* 46406 */ {(12<<2)|2,{117,32,0}},
/* 46407 */ {(12<<2)|2,{117,32,0}},
/* 46408 */ {(12<<2)|2,{117,32,0}},
/* 46409 */ {(12<<2)|2,{117,32,0}},
/* 46410 */ {(12<<2)|2,{117,32,0}},
/* 46411 */ {(12<<2)|2,{117,32,0}},
/* 46412 */ {(12<<2)|2,{117,32,0}},
/* 46413 */ {(12<<2)|2,{117,32,0}},
/* 46414 */ {(12<<2)|2,{117,32,0}},
/* 46415 */ {(12<<2)|2,{117,32,0}},
/* 46416 */ {(12<<2)|2,{117,37,0}},
/* 46417 */ {(12<<2)|2,{117,37,0}},
/* 46418 */ {(12<<2)|2,{117,37,0}},
/* 46419 */ {(12<<2)|2,{117,37,0}},
/* 46420 */ {(12<<2)|2,{117,37,0}},
/* 46421 */ {(12<<2)|2,{117,37,0}},
/* 46422 */ {(12<<2)|2,{117,37,0}},
/* 46423 */ {(12<<2)|2,{117,37,0}},
/* 46424 */ {(12<<2)|2,{117,37,0}},
/* 46425 */ {(12<<2)|2,{117,37,0}},
/* 46426 */ {(12<<2)|2,{117,37,0}},
/* 46427 */ {(12<<2)|2,{117,37,0}},
/* 46428 */ {(12<<2)|2,{117,37,0}},
/* 46429 */ {(12<<2)|2,{117,37,0}},
/* 46430 */ {(12<<2)|2,{117,37,0}},
/* 46431 */ {(12<<2)|2,{117,37,0}},
/* 46432 */ {(12<<2)|2,{117,45,0}},
/* 46433 */ {(12<<2)|2,{117,45,0}},
/* 46434 */ {(12<<2)|2,{117,45,0}},
/* 46435 */ {(12<<2)|2,{117,45,0}},
/* 46436 */ {(12<<2)|2,{117,45,0}},
/* 46437 */ {(12<<2)|2,{117,45,0}},
/* 46438 */ {(12<<2)|2,{117,45,0}},
/* 46439 */ {(12<<2)|2,{117,45,0}},
/* 46440 */ {(12<<2)|2,{117,45,0}},
/* 46441 */ {(12<<2)|2,{117,45,0}},
/* 46442 */ {(12<<2)|2,{117,45,0}},
/* 46443 */ {(12<<2)|2,{117,45,0}},
/* 46444 */ {(12<<2)|2,{117,45,0}},
/* 46445 */ {(12<<2)|2,{117,45,0}},
/* 46446 */ {(12<<2)|2,{117,45,0}},
/* 46447 */ {(12<<2)|2,{117,45,0}},
/* 46448 */ {(12<<2)|2,{117,46,0}},
/* 46449 */ {(12<<2)|2,{117,46,0}},
/* 46450 */ {(12<<2)|2,{117,46,0}},
/* 46451 */ {(12<<2)|2,{117,46,0}},
/* 46452 */ {(12<<2)|2,{117,46,0}},
/* 46453 */ {(12<<2)|2,{117,46,0}},
/* 46454 */ {(12<<2)|2,{117,46,0}},
/* 46455 */ {(12<<2)|2,{117,46,0}},
/* 46456 */ {(12<<2)|2,{117,46,0}},
/* 46457 */ {(12<<2)|2,{117,46,0}},
/* 46458 */ {(12<<2)|2,{117,46,0}},
/* 46459 */ {(12<<2)|2,{117,46,0}},
/* 46460 */ {(12<<2)|2,{117,46,0}},
/* 46461 */ {(12<<2)|2,{117,46,0}},
/* 46462 */ {(12<<2)|2,{117,46,0}},
/* 46463 */ {(12<<2)|2,{117,46,0}},
/* 46464 */ {(12<<2)|2,{117,47,0}},
/* 46465 */ {(12<<2)|2,{117,47,0}},
/* 46466 */ {(12<<2)|2,{117,47,0}},
/* 46467 */ {(12<<2)|2,{117,47,0}},
/* 46468 */ {(12<<2)|2,{117,47,0}},
/* 46469 */ {(12<<2)|2,{117,47,0}},
/* 46470 */ {(12<<2)|2,{117,47,0}},
/* 46471 */ {(12<<2)|2,{117,47,0}},
/* 46472 */ {(12<<2)|2,{117,47,0}},
/* 46473 */ {(12<<2)|2,{117,47,0}},
/* 46474 */ {(12<<2)|2,{117,47,0}},
/* 46475 */ {(12<<2)|2,{117,47,0}},
/* 46476 */ {(12<<2)|2,{117,47,0}},
/* 46477 */ {(12<<2)|2,{117,47,0}},
/* 46478 */ {(12<<2)|2,{117,47,0}},
/* 46479 */ {(12<<2)|2,{117,47,0}},
/* 46480 */ {(12<<2)|2,{117,51,0}},
/* 46481 */ {(12<<2)|2,{117,51,0}},
/* 46482 */ {(12<<2)|2,{117,51,0}},
/* 46483 */ {(12<<2)|2,{117,51,0}},
/* 46484 */ {(12<<2)|2,{117,51,0}},
/* 46485 */ {(12<<2)|2,{117,51,0}},
/* 46486 */ {(12<<2)|2,{117,51,0}},
/* 46487 */ {(12<<2)|2,{117,51,0}},
/* 46488 */ {(12<<2)|2,{117,51,0}},
/* 46489 */ {(12<<2)|2,{117,51,0}},
/* 46490 */ {(12<<2)|2,{117,51,0}},
/* 46491 */ {(12<<2)|2,{117,51,0}},
/* 46492 */ {(12<<2)|2,{117,51,0}},
/* 46493 */ {(12<<2)|2,{117,51,0}},
/* 46494 */ {(12<<2)|2,{117,51,0}},
/* 46495 */ {(12<<2)|2,{117,51,0}},
/* 46496 */ {(12<<2)|2,{117,52,0}},
/* 46497 */ {(12<<2)|2,{117,52,0}},
/* 46498 */ {(12<<2)|2,{117,52,0}},
/* 46499 */ {(12<<2)|2,{117,52,0}},
/* 46500 */ {(12<<2)|2,{117,52,0}},
/* 46501 */ {(12<<2)|2,{117,52,0}},
/* 46502 */ {(12<<2)|2,{117,52,0}},
/* 46503 */ {(12<<2)|2,{117,52,0}},
/* 46504 */ {(12<<2)|2,{117,52,0}},
/* 46505 */ {(12<<2)|2,{117,52,0}},
/* 46506 */ {(12<<2)|2,{117,52,0}},
/* 46507 */ {(12<<2)|2,{117,52,0}},
/* 46508 */ {(12<<2)|2,{117,52,0}},
/* 46509 */ {(12<<2)|2,{117,52,0}},
/* 46510 */ {(12<<2)|2,{117,52,0}},
/* 46511 */ {(12<<2)|2,{117,52,0}},
/* 46512 */ {(12<<2)|2,{117,53,0}},
/* 46513 */ {(12<<2)|2,{117,53,0}},
/* 46514 */ {(12<<2)|2,{117,53,0}},
/* 46515 */ {(12<<2)|2,{117,53,0}},
/* 46516 */ {(12<<2)|2,{117,53,0}},
/* 46517 */ {(12<<2)|2,{117,53,0}},
/* 46518 */ {(12<<2)|2,{117,53,0}},
/* 46519 */ {(12<<2)|2,{117,53,0}},
/* 46520 */ {(12<<2)|2,{117,53,0}},
/* 46521 */ {(12<<2)|2,{117,53,0}},
/* 46522 */ {(12<<2)|2,{117,53,0}},
/* 46523 */ {(12<<2)|2,{117,53,0}},
/* 46524 */ {(12<<2)|2,{117,53,0}},
/* 46525 */ {(12<<2)|2,{117,53,0}},
/* 46526 */ {(12<<2)|2,{117,53,0}},
/* 46527 */ {(12<<2)|2,{117,53,0}},
/* 46528 */ {(12<<2)|2,{117,54,0}},
/* 46529 */ {(12<<2)|2,{117,54,0}},
/* 46530 */ {(12<<2)|2,{117,54,0}},
/* 46531 */ {(12<<2)|2,{117,54,0}},
/* 46532 */ {(12<<2)|2,{117,54,0}},
/* 46533 */ {(12<<2)|2,{117,54,0}},
/* 46534 */ {(12<<2)|2,{117,54,0}},
/* 46535 */ {(12<<2)|2,{117,54,0}},
/* 46536 */ {(12<<2)|2,{117,54,0}},
/* 46537 */ {(12<<2)|2,{117,54,0}},
/* 46538 */ {(12<<2)|2,{117,54,0}},
/* 46539 */ {(12<<2)|2,{117,54,0}},
/* 46540 */ {(12<<2)|2,{117,54,0}},
/* 46541 */ {(12<<2)|2,{117,54,0}},
/* 46542 */ {(12<<2)|2,{117,54,0}},
/* 46543 */ {(12<<2)|2,{117,54,0}},
/* 46544 */ {(12<<2)|2,{117,55,0}},
/* 46545 */ {(12<<2)|2,{117,55,0}},
/* 46546 */ {(12<<2)|2,{117,55,0}},
/* 46547 */ {(12<<2)|2,{117,55,0}},
/* 46548 */ {(12<<2)|2,{117,55,0}},
/* 46549 */ {(12<<2)|2,{117,55,0}},
/* 46550 */ {(12<<2)|2,{117,55,0}},
/* 46551 */ {(12<<2)|2,{117,55,0}},
/* 46552 */ {(12<<2)|2,{117,55,0}},
/* 46553 */ {(12<<2)|2,{117,55,0}},
/* 46554 */ {(12<<2)|2,{117,55,0}},
/* 46555 */ {(12<<2)|2,{117,55,0}},
/* 46556 */ {(12<<2)|2,{117,55,0}},
/* 46557 */ {(12<<2)|2,{117,55,0}},
/* 46558 */ {(12<<2)|2,{117,55,0}},
/* 46559 */ {(12<<2)|2,{117,55,0}},
/* 46560 */ {(12<<2)|2,{117,56,0}},
/* 46561 */ {(12<<2)|2,{117,56,0}},
/* 46562 */ {(12<<2)|2,{117,56,0}},
/* 46563 */ {(12<<2)|2,{117,56,0}},
/* 46564 */ {(12<<2)|2,{117,56,0}},
/* 46565 */ {(12<<2)|2,{117,56,0}},
/* 46566 */ {(12<<2)|2,{117,56,0}},
/* 46567 */ {(12<<2)|2,{117,56,0}},
/* 46568 */ {(12<<2)|2,{117,56,0}},
/* 46569 */ {(12<<2)|2,{117,56,0}},
/* 46570 */ {(12<<2)|2,{117,56,0}},
/* 46571 */ {(12<<2)|2,{117,56,0}},
/* 46572 */ {(12<<2)|2,{117,56,0}},
/* 46573 */ {(12<<2)|2,{117,56,0}},
/* 46574 */ {(12<<2)|2,{117,56,0}},
/* 46575 */ {(12<<2)|2,{117,56,0}},
/* 46576 */ {(12<<2)|2,{117,57,0}},
/* 46577 */ {(12<<2)|2,{117,57,0}},
/* 46578 */ {(12<<2)|2,{117,57,0}},
/* 46579 */ {(12<<2)|2,{117,57,0}},
/* 46580 */ {(12<<2)|2,{117,57,0}},
/* 46581 */ {(12<<2)|2,{117,57,0}},
/* 46582 */ {(12<<2)|2,{117,57,0}},
/* 46583 */ {(12<<2)|2,{117,57,0}},
/* 46584 */ {(12<<2)|2,{117,57,0}},
/* 46585 */ {(12<<2)|2,{117,57,0}},
/* 46586 */ {(12<<2)|2,{117,57,0}},
/* 46587 */ {(12<<2)|2,{117,57,0}},
/* 46588 */ {(12<<2)|2,{117,57,0}},
/* 46589 */ {(12<<2)|2,{117,57,0}},
/* 46590 */ {(12<<2)|2,{117,57,0}},
/* 46591 */ {(12<<2)|2,{117,57,0}},
/* 46592 */ {(12<<2)|2,{117,61,0}},
/* 46593 */ {(12<<2)|2,{117,61,0}},
/* 46594 */ {(12<<2)|2,{117,61,0}},
/* 46595 */ {(12<<2)|2,{117,61,0}},
/* 46596 */ {(12<<2)|2,{117,61,0}},
/* 46597 */ {(12<<2)|2,{117,61,0}},
/* 46598 */ {(12<<2)|2,{117,61,0}},
/* 46599 */ {(12<<2)|2,{117,61,0}},
/* 46600 */ {(12<<2)|2,{117,61,0}},
/* 46601 */ {(12<<2)|2,{117,61,0}},
/* 46602 */ {(12<<2)|2,{117,61,0}},
/* 46603 */ {(12<<2)|2,{117,61,0}},
/* 46604 */ {(12<<2)|2,{117,61,0}},
/* 46605 */ {(12<<2)|2,{117,61,0}},
/* 46606 */ {(12<<2)|2,{117,61,0}},
/* 46607 */ {(12<<2)|2,{117,61,0}},
/* 46608 */ {(12<<2)|2,{117,65,0}},
/* 46609 */ {(12<<2)|2,{117,65,0}},
/* 46610 */ {(12<<2)|2,{117,65,0}},
/* 46611 */ {(12<<2)|2,{117,65,0}},
/* 46612 */ {(12<<2)|2,{117,65,0}},
/* 46613 */ {(12<<2)|2,{117,65,0}},
/* 46614 */ {(12<<2)|2,{117,65,0}},
/* 46615 */ {(12<<2)|2,{117,65,0}},
/* 46616 */ {(12<<2)|2,{117,65,0}},
/* 46617 */ {(12<<2)|2,{117,65,0}},
/* 46618 */ {(12<<2)|2,{117,65,0}},
/* 46619 */ {(12<<2)|2,{117,65,0}},
/* 46620 */ {(12<<2)|2,{117,65,0}},
/* 46621 */ {(12<<2)|2,{117,65,0}},
/* 46622 */ {(12<<2)|2,{117,65,0}},
/* 46623 */ {(12<<2)|2,{117,65,0}},
/* 46624 */ {(12<<2)|2,{117,95,0}},
/* 46625 */ {(12<<2)|2,{117,95,0}},
/* 46626 */ {(12<<2)|2,{117,95,0}},
/* 46627 */ {(12<<2)|2,{117,95,0}},
/* 46628 */ {(12<<2)|2,{117,95,0}},
/* 46629 */ {(12<<2)|2,{117,95,0}},
/* 46630 */ {(12<<2)|2,{117,95,0}},
/* 46631 */ {(12<<2)|2,{117,95,0}},
/* 46632 */ {(12<<2)|2,{117,95,0}},
/* 46633 */ {(12<<2)|2,{117,95,0}},
/* 46634 */ {(12<<2)|2,{117,95,0}},
/* 46635 */ {(12<<2)|2,{117,95,0}},
/* 46636 */ {(12<<2)|2,{117,95,0}},
/* 46637 */ {(12<<2)|2,{117,95,0}},
/* 46638 */ {(12<<2)|2,{117,95,0}},
/* 46639 */ {(12<<2)|2,{117,95,0}},
/* 46640 */ {(12<<2)|2,{117,98,0}},
/* 46641 */ {(12<<2)|2,{117,98,0}},
/* 46642 */ {(12<<2)|2,{117,98,0}},
/* 46643 */ {(12<<2)|2,{117,98,0}},
/* 46644 */ {(12<<2)|2,{117,98,0}},
/* 46645 */ {(12<<2)|2,{117,98,0}},
/* 46646 */ {(12<<2)|2,{117,98,0}},
/* 46647 */ {(12<<2)|2,{117,98,0}},
/* 46648 */ {(12<<2)|2,{117,98,0}},
/* 46649 */ {(12<<2)|2,{117,98,0}},
/* 46650 */ {(12<<2)|2,{117,98,0}},
/* 46651 */ {(12<<2)|2,{117,98,0}},
/* 46652 */ {(12<<2)|2,{117,98,0}},
/* 46653 */ {(12<<2)|2,{117,98,0}},
/* 46654 */ {(12<<2)|2,{117,98,0}},
/* 46655 */ {(12<<2)|2,{117,98,0}},
/* 46656 */ {(12<<2)|2,{117,100,0}},
/* 46657 */ {(12<<2)|2,{117,100,0}},
/* 46658 */ {(12<<2)|2,{117,100,0}},
/* 46659 */ {(12<<2)|2,{117,100,0}},
/* 46660 */ {(12<<2)|2,{117,100,0}},
/* 46661 */ {(12<<2)|2,{117,100,0}},
/* 46662 */ {(12<<2)|2,{117,100,0}},
/* 46663 */ {(12<<2)|2,{117,100,0}},
/* 46664 */ {(12<<2)|2,{117,100,0}},
/* 46665 */ {(12<<2)|2,{117,100,0}},
/* 46666 */ {(12<<2)|2,{117,100,0}},
/* 46667 */ {(12<<2)|2,{117,100,0}},
/* 46668 */ {(12<<2)|2,{117,100,0}},
/* 46669 */ {(12<<2)|2,{117,100,0}},
/* 46670 */ {(12<<2)|2,{117,100,0}},
/* 46671 */ {(12<<2)|2,{117,100,0}},
/* 46672 */ {(12<<2)|2,{117,102,0}},
/* 46673 */ {(12<<2)|2,{117,102,0}},
/* 46674 */ {(12<<2)|2,{117,102,0}},
/* 46675 */ {(12<<2)|2,{117,102,0}},
/* 46676 */ {(12<<2)|2,{117,102,0}},
/* 46677 */ {(12<<2)|2,{117,102,0}},
/* 46678 */ {(12<<2)|2,{117,102,0}},
/* 46679 */ {(12<<2)|2,{117,102,0}},
/* 46680 */ {(12<<2)|2,{117,102,0}},
/* 46681 */ {(12<<2)|2,{117,102,0}},
/* 46682 */ {(12<<2)|2,{117,102,0}},
/* 46683 */ {(12<<2)|2,{117,102,0}},
/* 46684 */ {(12<<2)|2,{117,102,0}},
/* 46685 */ {(12<<2)|2,{117,102,0}},
/* 46686 */ {(12<<2)|2,{117,102,0}},
/* 46687 */ {(12<<2)|2,{117,102,0}},
/* 46688 */ {(12<<2)|2,{117,103,0}},
/* 46689 */ {(12<<2)|2,{117,103,0}},
/* 46690 */ {(12<<2)|2,{117,103,0}},
/* 46691 */ {(12<<2)|2,{117,103,0}},
/* 46692 */ {(12<<2)|2,{117,103,0}},
/* 46693 */ {(12<<2)|2,{117,103,0}},
/* 46694 */ {(12<<2)|2,{117,103,0}},
/* 46695 */ {(12<<2)|2,{117,103,0}},
/* 46696 */ {(12<<2)|2,{117,103,0}},
/* 46697 */ {(12<<2)|2,{117,103,0}},
/* 46698 */ {(12<<2)|2,{117,103,0}},
/* 46699 */ {(12<<2)|2,{117,103,0}},
/* 46700 */ {(12<<2)|2,{117,103,0}},
/* 46701 */ {(12<<2)|2,{117,103,0}},
/* 46702 */ {(12<<2)|2,{117,103,0}},
/* 46703 */ {(12<<2)|2,{117,103,0}},
/* 46704 */ {(12<<2)|2,{117,104,0}},
/* 46705 */ {(12<<2)|2,{117,104,0}},
/* 46706 */ {(12<<2)|2,{117,104,0}},
/* 46707 */ {(12<<2)|2,{117,104,0}},
/* 46708 */ {(12<<2)|2,{117,104,0}},
/* 46709 */ {(12<<2)|2,{117,104,0}},
/* 46710 */ {(12<<2)|2,{117,104,0}},
/* 46711 */ {(12<<2)|2,{117,104,0}},
/* 46712 */ {(12<<2)|2,{117,104,0}},
/* 46713 */ {(12<<2)|2,{117,104,0}},
/* 46714 */ {(12<<2)|2,{117,104,0}},
/* 46715 */ {(12<<2)|2,{117,104,0}},
/* 46716 */ {(12<<2)|2,{117,104,0}},
/* 46717 */ {(12<<2)|2,{117,104,0}},
/* 46718 */ {(12<<2)|2,{117,104,0}},
/* 46719 */ {(12<<2)|2,{117,104,0}},
/* 46720 */ {(12<<2)|2,{117,108,0}},
/* 46721 */ {(12<<2)|2,{117,108,0}},
/* 46722 */ {(12<<2)|2,{117,108,0}},
/* 46723 */ {(12<<2)|2,{117,108,0}},
/* 46724 */ {(12<<2)|2,{117,108,0}},
/* 46725 */ {(12<<2)|2,{117,108,0}},
/* 46726 */ {(12<<2)|2,{117,108,0}},
/* 46727 */ {(12<<2)|2,{117,108,0}},
/* 46728 */ {(12<<2)|2,{117,108,0}},
/* 46729 */ {(12<<2)|2,{117,108,0}},
/* 46730 */ {(12<<2)|2,{117,108,0}},
/* 46731 */ {(12<<2)|2,{117,108,0}},
/* 46732 */ {(12<<2)|2,{117,108,0}},
/* 46733 */ {(12<<2)|2,{117,108,0}},
/* 46734 */ {(12<<2)|2,{117,108,0}},
/* 46735 */ {(12<<2)|2,{117,108,0}},
/* 46736 */ {(12<<2)|2,{117,109,0}},
/* 46737 */ {(12<<2)|2,{117,109,0}},
/* 46738 */ {(12<<2)|2,{117,109,0}},
/* 46739 */ {(12<<2)|2,{117,109,0}},
/* 46740 */ {(12<<2)|2,{117,109,0}},
/* 46741 */ {(12<<2)|2,{117,109,0}},
/* 46742 */ {(12<<2)|2,{117,109,0}},
/* 46743 */ {(12<<2)|2,{117,109,0}},
/* 46744 */ {(12<<2)|2,{117,109,0}},
/* 46745 */ {(12<<2)|2,{117,109,0}},
/* 46746 */ {(12<<2)|2,{117,109,0}},
/* 46747 */ {(12<<2)|2,{117,109,0}},
/* 46748 */ {(12<<2)|2,{117,109,0}},
/* 46749 */ {(12<<2)|2,{117,109,0}},
/* 46750 */ {(12<<2)|2,{117,109,0}},
/* 46751 */ {(12<<2)|2,{117,109,0}},
/* 46752 */ {(12<<2)|2,{117,110,0}},
/* 46753 */ {(12<<2)|2,{117,110,0}},
/* 46754 */ {(12<<2)|2,{117,110,0}},
/* 46755 */ {(12<<2)|2,{117,110,0}},
/* 46756 */ {(12<<2)|2,{117,110,0}},
/* 46757 */ {(12<<2)|2,{117,110,0}},
/* 46758 */ {(12<<2)|2,{117,110,0}},
/* 46759 */ {(12<<2)|2,{117,110,0}},
/* 46760 */ {(12<<2)|2,{117,110,0}},
/* 46761 */ {(12<<2)|2,{117,110,0}},
/* 46762 */ {(12<<2)|2,{117,110,0}},
/* 46763 */ {(12<<2)|2,{117,110,0}},
/* 46764 */ {(12<<2)|2,{117,110,0}},
/* 46765 */ {(12<<2)|2,{117,110,0}},
/* 46766 */ {(12<<2)|2,{117,110,0}},
/* 46767 */ {(12<<2)|2,{117,110,0}},
/* 46768 */ {(12<<2)|2,{117,112,0}},
/* 46769 */ {(12<<2)|2,{117,112,0}},
/* 46770 */ {(12<<2)|2,{117,112,0}},
/* 46771 */ {(12<<2)|2,{117,112,0}},
/* 46772 */ {(12<<2)|2,{117,112,0}},
/* 46773 */ {(12<<2)|2,{117,112,0}},
/* 46774 */ {(12<<2)|2,{117,112,0}},
/* 46775 */ {(12<<2)|2,{117,112,0}},
/* 46776 */ {(12<<2)|2,{117,112,0}},
/* 46777 */ {(12<<2)|2,{117,112,0}},
/* 46778 */ {(12<<2)|2,{117,112,0}},
/* 46779 */ {(12<<2)|2,{117,112,0}},
/* 46780 */ {(12<<2)|2,{117,112,0}},
/* 46781 */ {(12<<2)|2,{117,112,0}},
/* 46782 */ {(12<<2)|2,{117,112,0}},
/* 46783 */ {(12<<2)|2,{117,112,0}},
/* 46784 */ {(12<<2)|2,{117,114,0}},
/* 46785 */ {(12<<2)|2,{117,114,0}},
/* 46786 */ {(12<<2)|2,{117,114,0}},
/* 46787 */ {(12<<2)|2,{117,114,0}},
/* 46788 */ {(12<<2)|2,{117,114,0}},
/* 46789 */ {(12<<2)|2,{117,114,0}},
/* 46790 */ {(12<<2)|2,{117,114,0}},
/* 46791 */ {(12<<2)|2,{117,114,0}},
/* 46792 */ {(12<<2)|2,{117,114,0}},
/* 46793 */ {(12<<2)|2,{117,114,0}},
/* 46794 */ {(12<<2)|2,{117,114,0}},
/* 46795 */ {(12<<2)|2,{117,114,0}},
/* 46796 */ {(12<<2)|2,{117,114,0}},
/* 46797 */ {(12<<2)|2,{117,114,0}},
/* 46798 */ {(12<<2)|2,{117,114,0}},
/* 46799 */ {(12<<2)|2,{117,114,0}},
/* 46800 */ {(12<<2)|2,{117,117,0}},
/* 46801 */ {(12<<2)|2,{117,117,0}},
/* 46802 */ {(12<<2)|2,{117,117,0}},
/* 46803 */ {(12<<2)|2,{117,117,0}},
/* 46804 */ {(12<<2)|2,{117,117,0}},
/* 46805 */ {(12<<2)|2,{117,117,0}},
/* 46806 */ {(12<<2)|2,{117,117,0}},
/* 46807 */ {(12<<2)|2,{117,117,0}},
/* 46808 */ {(12<<2)|2,{117,117,0}},
/* 46809 */ {(12<<2)|2,{117,117,0}},
/* 46810 */ {(12<<2)|2,{117,117,0}},
/* 46811 */ {(12<<2)|2,{117,117,0}},
/* 46812 */ {(12<<2)|2,{117,117,0}},
/* 46813 */ {(12<<2)|2,{117,117,0}},
/* 46814 */ {(12<<2)|2,{117,117,0}},
/* 46815 */ {(12<<2)|2,{117,117,0}},
/* 46816 */ {(13<<2)|2,{117,58,0}},
/* 46817 */ {(13<<2)|2,{117,58,0}},
/* 46818 */ {(13<<2)|2,{117,58,0}},
/* 46819 */ {(13<<2)|2,{117,58,0}},
/* 46820 */ {(13<<2)|2,{117,58,0}},
/* 46821 */ {(13<<2)|2,{117,58,0}},
/* 46822 */ {(13<<2)|2,{117,58,0}},
/* 46823 */ {(13<<2)|2,{117,58,0}},
/* 46824 */ {(13<<2)|2,{117,66,0}},
/* 46825 */ {(13<<2)|2,{117,66,0}},
/* 46826 */ {(13<<2)|2,{117,66,0}},
/* 46827 */ {(13<<2)|2,{117,66,0}},
/* 46828 */ {(13<<2)|2,{117,66,0}},
/* 46829 */ {(13<<2)|2,{117,66,0}},
/* 46830 */ {(13<<2)|2,{117,66,0}},
/* 46831 */ {(13<<2)|2,{117,66,0}},
/* 46832 */ {(13<<2)|2,{117,67,0}},
/* 46833 */ {(13<<2)|2,{117,67,0}},
/* 46834 */ {(13<<2)|2,{117,67,0}},
/* 46835 */ {(13<<2)|2,{117,67,0}},
/* 46836 */ {(13<<2)|2,{117,67,0}},
/* 46837 */ {(13<<2)|2,{117,67,0}},
/* 46838 */ {(13<<2)|2,{117,67,0}},
/* 46839 */ {(13<<2)|2,{117,67,0}},
/* 46840 */ {(13<<2)|2,{117,68,0}},
/* 46841 */ {(13<<2)|2,{117,68,0}},
/* 46842 */ {(13<<2)|2,{117,68,0}},
/* 46843 */ {(13<<2)|2,{117,68,0}},
/* 46844 */ {(13<<2)|2,{117,68,0}},
/* 46845 */ {(13<<2)|2,{117,68,0}},
/* 46846 */ {(13<<2)|2,{117,68,0}},
/* 46847 */ {(13<<2)|2,{117,68,0}},
/* 46848 */ {(13<<2)|2,{117,69,0}},
/* 46849 */ {(13<<2)|2,{117,69,0}},
/* 46850 */ {(13<<2)|2,{117,69,0}},
/* 46851 */ {(13<<2)|2,{117,69,0}},
/* 46852 */ {(13<<2)|2,{117,69,0}},
/* 46853 */ {(13<<2)|2,{117,69,0}},
/* 46854 */ {(13<<2)|2,{117,69,0}},
/* 46855 */ {(13<<2)|2,{117,69,0}},
/* 46856 */ {(13<<2)|2,{117,70,0}},
/* 46857 */ {(13<<2)|2,{117,70,0}},
/* 46858 */ {(13<<2)|2,{117,70,0}},
/* 46859 */ {(13<<2)|2,{117,70,0}},
/* 46860 */ {(13<<2)|2,{117,70,0}},
/* 46861 */ {(13<<2)|2,{117,70,0}},
/* 46862 */ {(13<<2)|2,{117,70,0}},
/* 46863 */ {(13<<2)|2,{117,70,0}},
/* 46864 */ {(13<<2)|2,{117,71,0}},
/* 46865 */ {(13<<2)|2,{117,71,0}},
/* 46866 */ {(13<<2)|2,{117,71,0}},
/* 46867 */ {(13<<2)|2,{117,71,0}},
/* 46868 */ {(13<<2)|2,{117,71,0}},
/* 46869 */ {(13<<2)|2,{117,71,0}},
/* 46870 */ {(13<<2)|2,{117,71,0}},
/* 46871 */ {(13<<2)|2,{117,71,0}},
/* 46872 */ {(13<<2)|2,{117,72,0}},
/* 46873 */ {(13<<2)|2,{117,72,0}},
/* 46874 */ {(13<<2)|2,{117,72,0}},
/* 46875 */ {(13<<2)|2,{117,72,0}},
/* 46876 */ {(13<<2)|2,{117,72,0}},
/* 46877 */ {(13<<2)|2,{117,72,0}},
/* 46878 */ {(13<<2)|2,{117,72,0}},
/* 46879 */ {(13<<2)|2,{117,72,0}},
/* 46880 */ {(13<<2)|2,{117,73,0}},
/* 46881 */ {(13<<2)|2,{117,73,0}},
/* 46882 */ {(13<<2)|2,{117,73,0}},
/* 46883 */ {(13<<2)|2,{117,73,0}},
/* 46884 */ {(13<<2)|2,{117,73,0}},
/* 46885 */ {(13<<2)|2,{117,73,0}},
/* 46886 */ {(13<<2)|2,{117,73,0}},
/* 46887 */ {(13<<2)|2,{117,73,0}},
/* 46888 */ {(13<<2)|2,{117,74,0}},
/* 46889 */ {(13<<2)|2,{117,74,0}},
/* 46890 */ {(13<<2)|2,{117,74,0}},
/* 46891 */ {(13<<2)|2,{117,74,0}},
/* 46892 */ {(13<<2)|2,{117,74,0}},
/* 46893 */ {(13<<2)|2,{117,74,0}},
/* 46894 */ {(13<<2)|2,{117,74,0}},
/* 46895 */ {(13<<2)|2,{117,74,0}},
/* 46896 */ {(13<<2)|2,{117,75,0}},
/* 46897 */ {(13<<2)|2,{117,75,0}},
/* 46898 */ {(13<<2)|2,{117,75,0}},
/* 46899 */ {(13<<2)|2,{117,75,0}},
/* 46900 */ {(13<<2)|2,{117,75,0}},
/* 46901 */ {(13<<2)|2,{117,75,0}},
/* 46902 */ {(13<<2)|2,{117,75,0}},
/* 46903 */ {(13<<2)|2,{117,75,0}},
/* 46904 */ {(13<<2)|2,{117,76,0}},
/* 46905 */ {(13<<2)|2,{117,76,0}},
/* 46906 */ {(13<<2)|2,{117,76,0}},
/* 46907 */ {(13<<2)|2,{117,76,0}},
/* 46908 */ {(13<<2)|2,{117,76,0}},
/* 46909 */ {(13<<2)|2,{117,76,0}},
/* 46910 */ {(13<<2)|2,{117,76,0}},
/* 46911 */ {(13<<2)|2,{117,76,0}},
/* 46912 */ {(13<<2)|2,{117,77,0}},
/* 46913 */ {(13<<2)|2,{117,77,0}},
/* 46914 */ {(13<<2)|2,{117,77,0}},
/* 46915 */ {(13<<2)|2,{117,77,0}},
/* 46916 */ {(13<<2)|2,{117,77,0}},
/* 46917 */ {(13<<2)|2,{117,77,0}},
/* 46918 */ {(13<<2)|2,{117,77,0}},
/* 46919 */ {(13<<2)|2,{117,77,0}},
/* 46920 */ {(13<<2)|2,{117,78,0}},
/* 46921 */ {(13<<2)|2,{117,78,0}},
/* 46922 */ {(13<<2)|2,{117,78,0}},
/* 46923 */ {(13<<2)|2,{117,78,0}},
/* 46924 */ {(13<<2)|2,{117,78,0}},
/* 46925 */ {(13<<2)|2,{117,78,0}},
/* 46926 */ {(13<<2)|2,{117,78,0}},
/* 46927 */ {(13<<2)|2,{117,78,0}},
/* 46928 */ {(13<<2)|2,{117,79,0}},
/* 46929 */ {(13<<2)|2,{117,79,0}},
/* 46930 */ {(13<<2)|2,{117,79,0}},
/* 46931 */ {(13<<2)|2,{117,79,0}},
/* 46932 */ {(13<<2)|2,{117,79,0}},
/* 46933 */ {(13<<2)|2,{117,79,0}},
/* 46934 */ {(13<<2)|2,{117,79,0}},
/* 46935 */ {(13<<2)|2,{117,79,0}},
/* 46936 */ {(13<<2)|2,{117,80,0}},
/* 46937 */ {(13<<2)|2,{117,80,0}},
/* 46938 */ {(13<<2)|2,{117,80,0}},
/* 46939 */ {(13<<2)|2,{117,80,0}},
/* 46940 */ {(13<<2)|2,{117,80,0}},
/* 46941 */ {(13<<2)|2,{117,80,0}},
/* 46942 */ {(13<<2)|2,{117,80,0}},
/* 46943 */ {(13<<2)|2,{117,80,0}},
/* 46944 */ {(13<<2)|2,{117,81,0}},
/* 46945 */ {(13<<2)|2,{117,81,0}},
/* 46946 */ {(13<<2)|2,{117,81,0}},
/* 46947 */ {(13<<2)|2,{117,81,0}},
/* 46948 */ {(13<<2)|2,{117,81,0}},
/* 46949 */ {(13<<2)|2,{117,81,0}},
/* 46950 */ {(13<<2)|2,{117,81,0}},
/* 46951 */ {(13<<2)|2,{117,81,0}},
/* 46952 */ {(13<<2)|2,{117,82,0}},
/* 46953 */ {(13<<2)|2,{117,82,0}},
/* 46954 */ {(13<<2)|2,{117,82,0}},
/* 46955 */ {(13<<2)|2,{117,82,0}},
/* 46956 */ {(13<<2)|2,{117,82,0}},
/* 46957 */ {(13<<2)|2,{117,82,0}},
/* 46958 */ {(13<<2)|2,{117,82,0}},
/* 46959 */ {(13<<2)|2,{117,82,0}},
/* 46960 */ {(13<<2)|2,{117,83,0}},
/* 46961 */ {(13<<2)|2,{117,83,0}},
/* 46962 */ {(13<<2)|2,{117,83,0}},
/* 46963 */ {(13<<2)|2,{117,83,0}},
/* 46964 */ {(13<<2)|2,{117,83,0}},
/* 46965 */ {(13<<2)|2,{117,83,0}},
/* 46966 */ {(13<<2)|2,{117,83,0}},
/* 46967 */ {(13<<2)|2,{117,83,0}},
/* 46968 */ {(13<<2)|2,{117,84,0}},
/* 46969 */ {(13<<2)|2,{117,84,0}},
/* 46970 */ {(13<<2)|2,{117,84,0}},
/* 46971 */ {(13<<2)|2,{117,84,0}},
/* 46972 */ {(13<<2)|2,{117,84,0}},
/* 46973 */ {(13<<2)|2,{117,84,0}},
/* 46974 */ {(13<<2)|2,{117,84,0}},
/* 46975 */ {(13<<2)|2,{117,84,0}},
/* 46976 */ {(13<<2)|2,{117,85,0}},
/* 46977 */ {(13<<2)|2,{117,85,0}},
/* 46978 */ {(13<<2)|2,{117,85,0}},
/* 46979 */ {(13<<2)|2,{117,85,0}},
/* 46980 */ {(13<<2)|2,{117,85,0}},
/* 46981 */ {(13<<2)|2,{117,85,0}},
/* 46982 */ {(13<<2)|2,{117,85,0}},
/* 46983 */ {(13<<2)|2,{117,85,0}},
/* 46984 */ {(13<<2)|2,{117,86,0}},
/* 46985 */ {(13<<2)|2,{117,86,0}},
/* 46986 */ {(13<<2)|2,{117,86,0}},
/* 46987 */ {(13<<2)|2,{117,86,0}},
/* 46988 */ {(13<<2)|2,{117,86,0}},
/* 46989 */ {(13<<2)|2,{117,86,0}},
/* 46990 */ {(13<<2)|2,{117,86,0}},
/* 46991 */ {(13<<2)|2,{117,86,0}},
/* 46992 */ {(13<<2)|2,{117,87,0}},
/* 46993 */ {(13<<2)|2,{117,87,0}},
/* 46994 */ {(13<<2)|2,{117,87,0}},
/* 46995 */ {(13<<2)|2,{117,87,0}},
/* 46996 */ {(13<<2)|2,{117,87,0}},
/* 46997 */ {(13<<2)|2,{117,87,0}},
/* 46998 */ {(13<<2)|2,{117,87,0}},
/* 46999 */ {(13<<2)|2,{117,87,0}},
/* 47000 */ {(13<<2)|2,{117,89,0}},
/* 47001 */ {(13<<2)|2,{117,89,0}},
/* 47002 */ {(13<<2)|2,{117,89,0}},
/* 47003 */ {(13<<2)|2,{117,89,0}},
/* 47004 */ {(13<<2)|2,{117,89,0}},
/* 47005 */ {(13<<2)|2,{117,89,0}},
/* 47006 */ {(13<<2)|2,{117,89,0}},
/* 47007 */ {(13<<2)|2,{117,89,0}},
/* 47008 */ {(13<<2)|2,{117,106,0}},
/* 47009 */ {(13<<2)|2,{117,106,0}},
/* 47010 */ {(13<<2)|2,{117,106,0}},
/* 47011 */ {(13<<2)|2,{117,106,0}},
/* 47012 */ {(13<<2)|2,{117,106,0}},
/* 47013 */ {(13<<2)|2,{117,106,0}},
/* 47014 */ {(13<<2)|2,{117,106,0}},
/* 47015 */ {(13<<2)|2,{117,106,0}},
/* 47016 */ {(13<<2)|2,{117,107,0}},
/* 47017 */ {(13<<2)|2,{117,107,0}},
/* 47018 */ {(13<<2)|2,{117,107,0}},
/* 47019 */ {(13<<2)|2,{117,107,0}},
/* 47020 */ {(13<<2)|2,{117,107,0}},
/* 47021 */ {(13<<2)|2,{117,107,0}},
/* 47022 */ {(13<<2)|2,{117,107,0}},
/* 47023 */ {(13<<2)|2,{117,107,0}},
/* 47024 */ {(13<<2)|2,{117,113,0}},
/* 47025 */ {(13<<2)|2,{117,113,0}},
/* 47026 */ {(13<<2)|2,{117,113,0}},
/* 47027 */ {(13<<2)|2,{117,113,0}},
/* 47028 */ {(13<<2)|2,{117,113,0}},
/* 47029 */ {(13<<2)|2,{117,113,0}},
/* 47030 */ {(13<<2)|2,{117,113,0}},
/* 47031 */ {(13<<2)|2,{117,113,0}},
/* 47032 */ {(13<<2)|2,{117,118,0}},
/* 47033 */ {(13<<2)|2,{117,118,0}},
/* 47034 */ {(13<<2)|2,{117,118,0}},
/* 47035 */ {(13<<2)|2,{117,118,0}},
/* 47036 */ {(13<<2)|2,{117,118,0}},
/* 47037 */ {(13<<2)|2,{117,118,0}},
/* 47038 */ {(13<<2)|2,{117,118,0}},
/* 47039 */ {(13<<2)|2,{117,118,0}},
/* 47040 */ {(13<<2)|2,{117,119,0}},
/* 47041 */ {(13<<2)|2,{117,119,0}},
/* 47042 */ {(13<<2)|2,{117,119,0}},
/* 47043 */ {(13<<2)|2,{117,119,0}},
/* 47044 */ {(13<<2)|2,{117,119,0}},
/* 47045 */ {(13<<2)|2,{117,119,0}},
/* 47046 */ {(13<<2)|2,{117,119,0}},
/* 47047 */ {(13<<2)|2,{117,119,0}},
/* 47048 */ {(13<<2)|2,{117,120,0}},
/* 47049 */ {(13<<2)|2,{117,120,0}},
/* 47050 */ {(13<<2)|2,{117,120,0}},
/* 47051 */ {(13<<2)|2,{117,120,0}},
/* 47052 */ {(13<<2)|2,{117,120,0}},
/* 47053 */ {(13<<2)|2,{117,120,0}},
/* 47054 */ {(13<<2)|2,{117,120,0}},
/* 47055 */ {(13<<2)|2,{117,120,0}},
/* 47056 */ {(13<<2)|2,{117,121,0}},
/* 47057 */ {(13<<2)|2,{117,121,0}},
/* 47058 */ {(13<<2)|2,{117,121,0}},
/* 47059 */ {(13<<2)|2,{117,121,0}},
/* 47060 */ {(13<<2)|2,{117,121,0}},
/* 47061 */ {(13<<2)|2,{117,121,0}},
/* 47062 */ {(13<<2)|2,{117,121,0}},
/* 47063 */ {(13<<2)|2,{117,121,0}},
/* 47064 */ {(13<<2)|2,{117,122,0}},
/* 47065 */ {(13<<2)|2,{117,122,0}},
/* 47066 */ {(13<<2)|2,{117,122,0}},
/* 47067 */ {(13<<2)|2,{117,122,0}},
/* 47068 */ {(13<<2)|2,{117,122,0}},
/* 47069 */ {(13<<2)|2,{117,122,0}},
/* 47070 */ {(13<<2)|2,{117,122,0}},
/* 47071 */ {(13<<2)|2,{117,122,0}},
/* 47072 */ {(14<<2)|2,{117,38,0}},
/* 47073 */ {(14<<2)|2,{117,38,0}},
/* 47074 */ {(14<<2)|2,{117,38,0}},
/* 47075 */ {(14<<2)|2,{117,38,0}},
/* 47076 */ {(14<<2)|2,{117,42,0}},
/* 47077 */ {(14<<2)|2,{117,42,0}},
/* 47078 */ {(14<<2)|2,{117,42,0}},
/* 47079 */ {(14<<2)|2,{117,42,0}},
/* 47080 */ {(14<<2)|2,{117,44,0}},
/* 47081 */ {(14<<2)|2,{117,44,0}},
/* 47082 */ {(14<<2)|2,{117,44,0}},
/* 47083 */ {(14<<2)|2,{117,44,0}},
/* 47084 */ {(14<<2)|2,{117,59,0}},
/* 47085 */ {(14<<2)|2,{117,59,0}},
/* 47086 */ {(14<<2)|2,{117,59,0}},
/* 47087 */ {(14<<2)|2,{117,59,0}},
/* 47088 */ {(14<<2)|2,{117,88,0}},
/* 47089 */ {(14<<2)|2,{117,88,0}},
/* 47090 */ {(14<<2)|2,{117,88,0}},
/* 47091 */ {(14<<2)|2,{117,88,0}},
/* 47092 */ {(14<<2)|2,{117,90,0}},
/* 47093 */ {(14<<2)|2,{117,90,0}},
/* 47094 */ {(14<<2)|2,{117,90,0}},
/* 47095 */ {(14<<2)|2,{117,90,0}},
/* 47096 */ {(16<<2)|2,{117,33,0}},
/* 47097 */ {(16<<2)|2,{117,34,0}},
/* 47098 */ {(16<<2)|2,{117,40,0}},
/* 47099 */ {(16<<2)|2,{117,41,0}},
/* 47100 */ {(16<<2)|2,{117,63,0}},
/* 47101 */ {(6<<2)|1,{117,0,0}},
/* 47102 */ {(6<<2)|1,{117,0,0}},
/* 47103 */ {(6<<2)|1,{117,0,0}},
/* 47104 */ {(12<<2)|2,{58,48,0}},
/* 47105 */ {(12<<2)|2,{58,48,0}},
/* 47106 */ {(12<<2)|2,{58,48,0}},
/* 47107 */ {(12<<2)|2,{58,48,0}},
/* 47108 */ {(12<<2)|2,{58,48,0}},
/* 47109 */ {(12<<2)|2,{58,48,0}},
/* 47110 */ {(12<<2)|2,{58,48,0}},
/* 47111 */ {(12<<2)|2,{58,48,0}},
/* 47112 */ {(12<<2)|2,{58,48,0}},
/* 47113 */ {(12<<2)|2,{58,48,0}},
/* 47114 */ {(12<<2)|2,{58,48,0}},
/* 47115 */ {(12<<2)|2,{58,48,0}},
/* 47116 */ {(12<<2)|2,{58,48,0}},
/* 47117 */ {(12<<2)|2,{58,48,0}},
/* 47118 */ {(12<<2)|2,{58,48,0}},
/* 47119 */ {(12<<2)|2,{58,48,0}},
/* 47120 */ {(12<<2)|2,{58,49,0}},
/* 47121 */ {(12<<2)|2,{58,49,0}},
/* 47122 */ {(12<<2)|2,{58,49,0}},
/* 47123 */ {(12<<2)|2,{58,49,0}},
/* 47124 */ {(12<<2)|2,{58,49,0}},
/* 47125 */ {(12<<2)|2,{58,49,0}},
/* 47126 */ {(12<<2)|2,{58,49,0}},
/* 47127 */ {(12<<2)|2,{58,49,0}},
/* 47128 */ {(12<<2)|2,{58,49,0}},
/* 47129 */ {(12<<2)|2,{58,49,0}},
/* 47130 */ {(12<<2)|2,{58,49,0}},
/* 47131 */ {(12<<2)|2,{58,49,0}},
/* 47132 */ {(12<<2)|2,{58,49,0}},
/* 47133 */ {(12<<2)|2,{58,49,0}},
/* 47134 */ {(12<<2)|2,{58,49,0}},
/* 47135 */ {(12<<2)|2,{58,49,0}},
/* 47136 */ {(12<<2)|2,{58,50,0}},
/* 47137 */ {(12<<2)|2,{58,50,0}},
/* 47138 */ {(12<<2)|2,{58,50,0}},
/* 47139 */ {(12<<2)|2,{58,50,0}},
/* 47140 */ {(12<<2)|2,{58,50,0}},
/* 47141 */ {(12<<2)|2,{58,50,0}},
/* 47142 */ {(12<<2)|2,{58,50,0}},
/* 47143 */ {(12<<2)|2,{58,50,0}},
/* 47144 */ {(12<<2)|2,{58,50,0}},
/* 47145 */ {(12<<2)|2,{58,50,0}},
/* 47146 */ {(12<<2)|2,{58,50,0}},
/* 47147 */ {(12<<2)|2,{58,50,0}},
/* 47148 */ {(12<<2)|2,{58,50,0}},
/* 47149 */ {(12<<2)|2,{58,50,0}},
/* 47150 */ {(12<<2)|2,{58,50,0}},
/* 47151 */ {(12<<2)|2,{58,50,0}},
/* 47152 */ {(12<<2)|2,{58,97,0}},
/* 47153 */ {(12<<2)|2,{58,97,0}},
/* 47154 */ {(12<<2)|2,{58,97,0}},
/* 47155 */ {(12<<2)|2,{58,97,0}},
/* 47156 */ {(12<<2)|2,{58,97,0}},
/* 47157 */ {(12<<2)|2,{58,97,0}},
/* 47158 */ {(12<<2)|2,{58,97,0}},
/* 47159 */ {(12<<2)|2,{58,97,0}},
/* 47160 */ {(12<<2)|2,{58,97,0}},
/* 47161 */ {(12<<2)|2,{58,97,0}},
/* 47162 */ {(12<<2)|2,{58,97,0}},
/* 47163 */ {(12<<2)|2,{58,97,0}},
/* 47164 */ {(12<<2)|2,{58,97,0}},
/* 47165 */ {(12<<2)|2,{58,97,0}},
/* 47166 */ {(12<<2)|2,{58,97,0}},
/* 47167 */ {(12<<2)|2,{58,97,0}},
/* 47168 */ {(12<<2)|2,{58,99,0}},
/* 47169 */ {(12<<2)|2,{58,99,0}},
/* 47170 */ {(12<<2)|2,{58,99,0}},
/* 47171 */ {(12<<2)|2,{58,99,0}},
/* 47172 */ {(12<<2)|2,{58,99,0}},
/* 47173 */ {(12<<2)|2,{58,99,0}},
/* 47174 */ {(12<<2)|2,{58,99,0}},
/* 47175 */ {(12<<2)|2,{58,99,0}},
/* 47176 */ {(12<<2)|2,{58,99,0}},
/* 47177 */ {(12<<2)|2,{58,99,0}},
/* 47178 */ {(12<<2)|2,{58,99,0}},
/* 47179 */ {(12<<2)|2,{58,99,0}},
/* 47180 */ {(12<<2)|2,{58,99,0}},
/* 47181 */ {(12<<2)|2,{58,99,0}},
/* 47182 */ {(12<<2)|2,{58,99,0}},
/* 47183 */ {(12<<2)|2,{58,99,0}},
/* 47184 */ {(12<<2)|2,{58,101,0}},
/* 47185 */ {(12<<2)|2,{58,101,0}},
/* 47186 */ {(12<<2)|2,{58,101,0}},
/* 47187 */ {(12<<2)|2,{58,101,0}},
/* 47188 */ {(12<<2)|2,{58,101,0}},
/* 47189 */ {(12<<2)|2,{58,101,0}},
/* 47190 */ {(12<<2)|2,{58,101,0}},
/* 47191 */ {(12<<2)|2,{58,101,0}},
/* 47192 */ {(12<<2)|2,{58,101,0}},
/* 47193 */ {(12<<2)|2,{58,101,0}},
/* 47194 */ {(12<<2)|2,{58,101,0}},
/* 47195 */ {(12<<2)|2,{58,101,0}},
/* 47196 */ {(12<<2)|2,{58,101,0}},
/* 47197 */ {(12<<2)|2,{58,101,0}},
/* 47198 */ {(12<<2)|2,{58,101,0}},
/* 47199 */ {(12<<2)|2,{58,101,0}},
/* 47200 */ {(12<<2)|2,{58,105,0}},
/* 47201 */ {(12<<2)|2,{58,105,0}},
/* 47202 */ {(12<<2)|2,{58,105,0}},
/* 47203 */ {(12<<2)|2,{58,105,0}},
/* 47204 */ {(12<<2)|2,{58,105,0}},
/* 47205 */ {(12<<2)|2,{58,105,0}},
/* 47206 */ {(12<<2)|2,{58,105,0}},
/* 47207 */ {(12<<2)|2,{58,105,0}},
/* 47208 */ {(12<<2)|2,{58,105,0}},
/* 47209 */ {(12<<2)|2,{58,105,0}},
/* 47210 */ {(12<<2)|2,{58,105,0}},
/* 47211 */ {(12<<2)|2,{58,105,0}},
/* 47212 */ {(12<<2)|2,{58,105,0}},
/* 47213 */ {(12<<2)|2,{58,105,0}},
/* 47214 */ {(12<<2)|2,{58,105,0}},
/* 47215 */ {(12<<2)|2,{58,105,0}},
/* 47216 */ {(12<<2)|2,{58,111,0}},
/* 47217 */ {(12<<2)|2,{58,111,0}},
/* 47218 */ {(12<<2)|2,{58,111,0}},
/* 47219 */ {(12<<2)|2,{58,111,0}},
/* 47220 */ {(12<<2)|2,{58,111,0}},
/* 47221 */ {(12<<2)|2,{58,111,0}},
/* 47222 */ {(12<<2)|2,{58,111,0}},
/* 47223 */ {(12<<2)|2,{58,111,0}},
/* 47224 */ {(12<<2)|2,{58,111,0}},
/* 47225 */ {(12<<2)|2,{58,111,0}},
/* 47226 */ {(12<<2)|2,{58,111,0}},
/* 47227 */ {(12<<2)|2,{58,111,0}},
/* 47228 */ {(12<<2)|2,{58,111,0}},
/* 47229 */ {(12<<2)|2,{58,111,0}},
/* 47230 */ {(12<<2)|2,{58,111,0}},
/* 47231 */ {(12<<2)|2,{58,111,0}},
/* 47232 */ {(12<<2)|2,{58,115,0}},
/* 47233 */ {(12<<2)|2,{58,115,0}},
/* 47234 */ {(12<<2)|2,{58,115,0}},
/* 47235 */ {(12<<2)|2,{58,115,0}},
/* 47236 */ {(12<<2)|2,{58,115,0}},
/* 47237 */ {(12<<2)|2,{58,115,0}},
/* 47238 */ {(12<<2)|2,{58,115,0}},
/* 47239 */ {(12<<2)|2,{58,115,0}},
/* 47240 */ {(12<<2)|2,{58,115,0}},
/* 47241 */ {(12<<2)|2,{58,115,0}},
/* 47242 */ {(12<<2)|2,{58,115,0}},
/* 47243 */ {(12<<2)|2,{58,115,0}},
/* 47244 */ {(12<<2)|2,{58,115,0}},
/* 47245 */ {(12<<2)|2,{58,115,0}},
/* 47246 */ {(12<<2)|2,{58,115,0}},
/* 47247 */ {(12<<2)|2,{58,115,0}},
/* 47248 */ {(12<<2)|2,{58,116,0}},
/* 47249 */ {(12<<2)|2,{58,116,0}},
/* 47250 */ {(12<<2)|2,{58,116,0}},
/* 47251 */ {(12<<2)|2,{58,116,0}},
/* 47252 */ {(12<<2)|2,{58,116,0}},
/* 47253 */ {(12<<2)|2,{58,116,0}},
/* 47254 */ {(12<<2)|2,{58,116,0}},
/* 47255 */ {(12<<2)|2,{58,116,0}},
/* 47256 */ {(12<<2)|2,{58,116,0}},
/* 47257 */ {(12<<2)|2,{58,116,0}},
/* 47258 */ {(12<<2)|2,{58,116,0}},
/* 47259 */ {(12<<2)|2,{58,116,0}},
/* 47260 */ {(12<<2)|2,{58,116,0}},
/* 47261 */ {(12<<2)|2,{58,116,0}},
/* 47262 */ {(12<<2)|2,{58,116,0}},
/* 47263 */ {(12<<2)|2,{58,116,0}},
/* 47264 */ {(13<<2)|2,{58,32,0}},
/* 47265 */ {(13<<2)|2,{58,32,0}},
/* 47266 */ {(13<<2)|2,{58,32,0}},
/* 47267 */ {(13<<2)|2,{58,32,0}},
/* 47268 */ {(13<<2)|2,{58,32,0}},
/* 47269 */ {(13<<2)|2,{58,32,0}},
/* 47270 */ {(13<<2)|2,{58,32,0}},
/* 47271 */ {(13<<2)|2,{58,32,0}},
/* 47272 */ {(13<<2)|2,{58,37,0}},
/* 47273 */ {(13<<2)|2,{58,37,0}},
/* 47274 */ {(13<<2)|2,{58,37,0}},
/* 47275 */ {(13<<2)|2,{58,37,0}},
/* 47276 */ {(13<<2)|2,{58,37,0}},
/* 47277 */ {(13<<2)|2,{58,37,0}},
/* 47278 */ {(13<<2)|2,{58,37,0}},
/* 47279 */ {(13<<2)|2,{58,37,0}},
/* 47280 */ {(13<<2)|2,{58,45,0}},
/* 47281 */ {(13<<2)|2,{58,45,0}},
/* 47282 */ {(13<<2)|2,{58,45,0}},
/* 47283 */ {(13<<2)|2,{58,45,0}},
/* 47284 */ {(13<<2)|2,{58,45,0}},
/* 47285 */ {(13<<2)|2,{58,45,0}},
/* 47286 */ {(13<<2)|2,{58,45,0}},
/* 47287 */ {(13<<2)|2,{58,45,0}},
/* 47288 */ {(13<<2)|2,{58,46,0}},
/* 47289 */ {(13<<2)|2,{58,46,0}},
/* 47290 */ {(13<<2)|2,{58,46,0}},
/* 47291 */ {(13<<2)|2,{58,46,0}},
/* 47292 */ {(13<<2)|2,{58,46,0}},
/* 47293 */ {(13<<2)|2,{58,46,0}},
/* 47294 */ {(13<<2)|2,{58,46,0}},
/* 47295 */ {(13<<2)|2,{58,46,0}},
/* 47296 */ {(13<<2)|2,{58,47,0}},
/* 47297 */ {(13<<2)|2,{58,47,0}},
/* 47298 */ {(13<<2)|2,{58,47,0}},
/* 47299 */ {(13<<2)|2,{58,47,0}},
/* 47300 */ {(13<<2)|2,{58,47,0}},
/* 47301 */ {(13<<2)|2,{58,47,0}},
/* 47302 */ {(13<<2)|2,{58,47,0}},
/* 47303 */ {(13<<2)|2,{58,47,0}},
/* 47304 */ {(13<<2)|2,{58,51,0}},
/* 47305 */ {(13<<2)|2,{58,51,0}},
/* 47306 */ {(13<<2)|2,{58,51,0}},
/* 47307 */ {(13<<2)|2,{58,51,0}},
/* 47308 */ {(13<<2)|2,{58,51,0}},
/* 47309 */ {(13<<2)|2,{58,51,0}},
/* 47310 */ {(13<<2)|2,{58,51,0}},
/* 47311 */ {(13<<2)|2,{58,51,0}},
/* 47312 */ {(13<<2)|2,{58,52,0}},
/* 47313 */ {(13<<2)|2,{58,52,0}},
/* 47314 */ {(13<<2)|2,{58,52,0}},
/* 47315 */ {(13<<2)|2,{58,52,0}},
/* 47316 */ {(13<<2)|2,{58,52,0}},
/* 47317 */ {(13<<2)|2,{58,52,0}},
/* 47318 */ {(13<<2)|2,{58,52,0}},
/* 47319 */ {(13<<2)|2,{58,52,0}},
/* 47320 */ {(13<<2)|2,{58,53,0}},
/* 47321 */ {(13<<2)|2,{58,53,0}},
/* 47322 */ {(13<<2)|2,{58,53,0}},
/* 47323 */ {(13<<2)|2,{58,53,0}},
/* 47324 */ {(13<<2)|2,{58,53,0}},
/* 47325 */ {(13<<2)|2,{58,53,0}},
/* 47326 */ {(13<<2)|2,{58,53,0}},
/* 47327 */ {(13<<2)|2,{58,53,0}},
/* 47328 */ {(13<<2)|2,{58,54,0}},
/* 47329 */ {(13<<2)|2,{58,54,0}},
/* 47330 */ {(13<<2)|2,{58,54,0}},
/* 47331 */ {(13<<2)|2,{58,54,0}},
/* 47332 */ {(13<<2)|2,{58,54,0}},
/* 47333 */ {(13<<2)|2,{58,54,0}},
/* 47334 */ {(13<<2)|2,{58,54,0}},
/* 47335 */ {(13<<2)|2,{58,54,0}},
/* 47336 */ {(13<<2)|2,{58,55,0}},
/* 47337 */ {(13<<2)|2,{58,55,0}},
/* 47338 */ {(13<<2)|2,{58,55,0}},
/* 47339 */ {(13<<2)|2,{58,55,0}},
/* 47340 */ {(13<<2)|2,{58,55,0}},
/* 47341 */ {(13<<2)|2,{58,55,0}},
/* 47342 */ {(13<<2)|2,{58,55,0}},
/* 47343 */ {(13<<2)|2,{58,55,0}},
/* 47344 */ {(13<<2)|2,{58,56,0}},
/* 47345 */ {(13<<2)|2,{58,56,0}},
/* 47346 */ {(13<<2)|2,{58,56,0}},
/* 47347 */ {(13<<2)|2,{58,56,0}},
/* 47348 */ {(13<<2)|2,{58,56,0}},
/* 47349 */ {(13<<2)|2,{58,56,0}},
/* 47350 */ {(13<<2)|2,{58,56,0}},
/* 47351 */ {(13<<2)|2,{58,56,0}},
/* 47352 */ {(13<<2)|2,{58,57,0}},
/* 47353 */ {(13<<2)|2,{58,57,0}},
/* 47354 */ {(13<<2)|2,{58,57,0}},
/* 47355 */ {(13<<2)|2,{58,57,0}},
/* 47356 */ {(13<<2)|2,{58,57,0}},
/* 47357 */ {(13<<2)|2,{58,57,0}},
/* 47358 */ {(13<<2)|2,{58,57,0}},
/* 47359 */ {(13<<2)|2,{58,57,0}},
/* 47360 */ {(13<<2)|2,{58,61,0}},
/* 47361 */ {(13<<2)|2,{58,61,0}},
/* 47362 */ {(13<<2)|2,{58,61,0}},
/* 47363 */ {(13<<2)|2,{58,61,0}},
/* 47364 */ {(13<<2)|2,{58,61,0}},
/* 47365 */ {(13<<2)|2,{58,61,0}},
/* 47366 */ {(13<<2)|2,{58,61,0}},
/* 47367 */ {(13<<2)|2,{58,61,0}},
/* 47368 */ {(13<<2)|2,{58,65,0}},
/* 47369 */ {(13<<2)|2,{58,65,0}},
/* 47370 */ {(13<<2)|2,{58,65,0}},
/* 47371 */ {(13<<2)|2,{58,65,0}},
/* 47372 */ {(13<<2)|2,{58,65,0}},
/* 47373 */ {(13<<2)|2,{58,65,0}},
/* 47374 */ {(13<<2)|2,{58,65,0}},
/* 47375 */ {(13<<2)|2,{58,65,0}},
/* 47376 */ {(13<<2)|2,{58,95,0}},
/* 47377 */ {(13<<2)|2,{58,95,0}},
/* 47378 */ {(13<<2)|2,{58,95,0}},
/* 47379 */ {(13<<2)|2,{58,95,0}},
/* 47380 */ {(13<<2)|2,{58,95,0}},
/* 47381 */ {(13<<2)|2,{58,95,0}},
/* 47382 */ {(13<<2)|2,{58,95,0}},
/* 47383 */ {(13<<2)|2,{58,95,0}},
/* 47384 */ {(13<<2)|2,{58,98,0}},
/* 47385 */ {(13<<2)|2,{58,98,0}},
/* 47386 */ {(13<<2)|2,{58,98,0}},
/* 47387 */ {(13<<2)|2,{58,98,0}},
/* 47388 */ {(13<<2)|2,{58,98,0}},
/* 47389 */ {(13<<2)|2,{58,98,0}},
/* 47390 */ {(13<<2)|2,{58,98,0}},
/* 47391 */ {(13<<2)|2,{58,98,0}},
/* 47392 */ {(13<<2)|2,{58,100,0}},
/* 47393 */ {(13<<2)|2,{58,100,0}},
/* 47394 */ {(13<<2)|2,{58,100,0}},
/* 47395 */ {(13<<2)|2,{58,100,0}},
/* 47396 */ {(13<<2)|2,{58,100,0}},
/* 47397 */ {(13<<2)|2,{58,100,0}},
/* 47398 */ {(13<<2)|2,{58,100,0}},
/* 47399 */ {(13<<2)|2,{58,100,0}},
/* 47400 */ {(13<<2)|2,{58,102,0}},
/* 47401 */ {(13<<2)|2,{58,102,0}},
/* 47402 */ {(13<<2)|2,{58,102,0}},
/* 47403 */ {(13<<2)|2,{58,102,0}},
/* 47404 */ {(13<<2)|2,{58,102,0}},
/* 47405 */ {(13<<2)|2,{58,102,0}},
/* 47406 */ {(13<<2)|2,{58,102,0}},
/* 47407 */ {(13<<2)|2,{58,102,0}},
/* 47408 */ {(13<<2)|2,{58,103,0}},
/* 47409 */ {(13<<2)|2,{58,103,0}},
/* 47410 */ {(13<<2)|2,{58,103,0}},
/* 47411 */ {(13<<2)|2,{58,103,0}},
/* 47412 */ {(13<<2)|2,{58,103,0}},
/* 47413 */ {(13<<2)|2,{58,103,0}},
/* 47414 */ {(13<<2)|2,{58,103,0}},
/* 47415 */ {(13<<2)|2,{58,103,0}},
/* 47416 */ {(13<<2)|2,{58,104,0}},
/* 47417 */ {(13<<2)|2,{58,104,0}},
/* 47418 */ {(13<<2)|2,{58,104,0}},
/* 47419 */ {(13<<2)|2,{58,104,0}},
/* 47420 */ {(13<<2)|2,{58,104,0}},
/* 47421 */ {(13<<2)|2,{58,104,0}},
/* 47422 */ {(13<<2)|2,{58,104,0}},
/* 47423 */ {(13<<2)|2,{58,104,0}},
/* 47424 */ {(13<<2)|2,{58,108,0}},
/* 47425 */ {(13<<2)|2,{58,108,0}},
/* 47426 */ {(13<<2)|2,{58,108,0}},
/* 47427 */ {(13<<2)|2,{58,108,0}},
/* 47428 */ {(13<<2)|2,{58,108,0}},
/* 47429 */ {(13<<2)|2,{58,108,0}},
/* 47430 */ {(13<<2)|2,{58,108,0}},
/* 47431 */ {(13<<2)|2,{58,108,0}},
/* 47432 */ {(13<<2)|2,{58,109,0}},
/* 47433 */ {(13<<2)|2,{58,109,0}},
/* 47434 */ {(13<<2)|2,{58,109,0}},
/* 47435 */ {(13<<2)|2,{58,109,0}},
/* 47436 */ {(13<<2)|2,{58,109,0}},
/* 47437 */ {(13<<2)|2,{58,109,0}},
/* 47438 */ {(13<<2)|2,{58,109,0}},
/* 47439 */ {(13<<2)|2,{58,109,0}},
/* 47440 */ {(13<<2)|2,{58,110,0}},
/* 47441 */ {(13<<2)|2,{58,110,0}},
/* 47442 */ {(13<<2)|2,{58,110,0}},
/* 47443 */ {(13<<2)|2,{58,110,0}},
/* 47444 */ {(13<<2)|2,{58,110,0}},
/* 47445 */ {(13<<2)|2,{58,110,0}},
/* 47446 */ {(13<<2)|2,{58,110,0}},
/* 47447 */ {(13<<2)|2,{58,110,0}},
/* 47448 */ {(13<<2)|2,{58,112,0}},
/* 47449 */ {(13<<2)|2,{58,112,0}},
/* 47450 */ {(13<<2)|2,{58,112,0}},
/* 47451 */ {(13<<2)|2,{58,112,0}},
/* 47452 */ {(13<<2)|2,{58,112,0}},
/* 47453 */ {(13<<2)|2,{58,112,0}},
/* 47454 */ {(13<<2)|2,{58,112,0}},
/* 47455 */ {(13<<2)|2,{58,112,0}},
/* 47456 */ {(13<<2)|2,{58,114,0}},
/* 47457 */ {(13<<2)|2,{58,114,0}},
/* 47458 */ {(13<<2)|2,{58,114,0}},
/* 47459 */ {(13<<2)|2,{58,114,0}},
/* 47460 */ {(13<<2)|2,{58,114,0}},
/* 47461 */ {(13<<2)|2,{58,114,0}},
/* 47462 */ {(13<<2)|2,{58,114,0}},
/* 47463 */ {(13<<2)|2,{58,114,0}},
/* 47464 */ {(13<<2)|2,{58,117,0}},
/* 47465 */ {(13<<2)|2,{58,117,0}},
/* 47466 */ {(13<<2)|2,{58,117,0}},
/* 47467 */ {(13<<2)|2,{58,117,0}},
/* 47468 */ {(13<<2)|2,{58,117,0}},
/* 47469 */ {(13<<2)|2,{58,117,0}},
/* 47470 */ {(13<<2)|2,{58,117,0}},
/* 47471 */ {(13<<2)|2,{58,117,0}},
/* 47472 */ {(14<<2)|2,{58,58,0}},
/* 47473 */ {(14<<2)|2,{58,58,0}},
/* 47474 */ {(14<<2)|2,{58,58,0}},
/* 47475 */ {(14<<2)|2,{58,58,0}},
/* 47476 */ {(14<<2)|2,{58,66,0}},
/* 47477 */ {(14<<2)|2,{58,66,0}},
/* 47478 */ {(14<<2)|2,{58,66,0}},
/* 47479 */ {(14<<2)|2,{58,66,0}},
/* 47480 */ {(14<<2)|2,{58,67,0}},
/* 47481 */ {(14<<2)|2,{58,67,0}},
/* 47482 */ {(14<<2)|2,{58,67,0}},
/* 47483 */ {(14<<2)|2,{58,67,0}},
/* 47484 */ {(14<<2)|2,{58,68,0}},
/* 47485 */ {(14<<2)|2,{58,68,0}},
/* 47486 */ {(14<<2)|2,{58,68,0}},
/* 47487 */ {(14<<2)|2,{58,68,0}},
/* 47488 */ {(14<<2)|2,{58,69,0}},
/* 47489 */ {(14<<2)|2,{58,69,0}},
/* 47490 */ {(14<<2)|2,{58,69,0}},
/* 47491 */ {(14<<2)|2,{58,69,0}},
/* 47492 */ {(14<<2)|2,{58,70,0}},
/* 47493 */ {(14<<2)|2,{58,70,0}},
/* 47494 */ {(14<<2)|2,{58,70,0}},
/* 47495 */ {(14<<2)|2,{58,70,0}},
/* 47496 */ {(14<<2)|2,{58,71,0}},
/* 47497 */ {(14<<2)|2,{58,71,0}},
/* 47498 */ {(14<<2)|2,{58,71,0}},
/* 47499 */ {(14<<2)|2,{58,71,0}},
/* 47500 */ {(14<<2)|2,{58,72,0}},
/* 47501 */ {(14<<2)|2,{58,72,0}},
/* 47502 */ {(14<<2)|2,{58,72,0}},
/* 47503 */ {(14<<2)|2,{58,72,0}},
/* 47504 */ {(14<<2)|2,{58,73,0}},
/* 47505 */ {(14<<2)|2,{58,73,0}},
/* 47506 */ {(14<<2)|2,{58,73,0}},
/* 47507 */ {(14<<2)|2,{58,73,0}},
/* 47508 */ {(14<<2)|2,{58,74,0}},
/* 47509 */ {(14<<2)|2,{58,74,0}},
/* 47510 */ {(14<<2)|2,{58,74,0}},
/* 47511 */ {(14<<2)|2,{58,74,0}},
/* 47512 */ {(14<<2)|2,{58,75,0}},
/* 47513 */ {(14<<2)|2,{58,75,0}},
/* 47514 */ {(14<<2)|2,{58,75,0}},
/* 47515 */ {(14<<2)|2,{58,75,0}},
/* 47516 */ {(14<<2)|2,{58,76,0}},
/* 47517 */ {(14<<2)|2,{58,76,0}},
/* 47518 */ {(14<<2)|2,{58,76,0}},
/* 47519 */ {(14<<2)|2,{58,76,0}},
/* 47520 */ {(14<<2)|2,{58,77,0}},
/* 47521 */ {(14<<2)|2,{58,77,0}},
/* 47522 */ {(14<<2)|2,{58,77,0}},
/* 47523 */ {(14<<2)|2,{58,77,0}},
/* 47524 */ {(14<<2)|2,{58,78,0}},
/* 47525 */ {(14<<2)|2,{58,78,0}},
/* 47526 */ {(14<<2)|2,{58,78,0}},
/* 47527 */ {(14<<2)|2,{58,78,0}},
/* 47528 */ {(14<<2)|2,{58,79,0}},
/* 47529 */ {(14<<2)|2,{58,79,0}},
/* 47530 */ {(14<<2)|2,{58,79,0}},
/* 47531 */ {(14<<2)|2,{58,79,0}},
/* 47532 */ {(14<<2)|2,{58,80,0}},
/* 47533 */ {(14<<2)|2,{58,80,0}},
/* 47534 */ {(14<<2)|2,{58,80,0}},
/* 47535 */ {(14<<2)|2,{58,80,0}},
/* 47536 */ {(14<<2)|2,{58,81,0}},
/* 47537 */ {(14<<2)|2,{58,81,0}},
/* 47538 */ {(14<<2)|2,{58,81,0}},
/* 47539 */ {(14<<2)|2,{58,81,0}},
/* 47540 */ {(14<<2)|2,{58,82,0}},
/* 47541 */ {(14<<2)|2,{58,82,0}},
/* 47542 */ {(14<<2)|2,{58,82,0}},
/* 47543 */ {(14<<2)|2,{58,82,0}},
/* 47544 */ {(14<<2)|2,{58,83,0}},
/* 47545 */ {(14<<2)|2,{58,83,0}},
/* 47546 */ {(14<<2)|2,{58,83,0}},
/* 47547 */ {(14<<2)|2,{58,83,0}},
/* 47548 */ {(14<<2)|2,{58,84,0}},
/* 47549 */ {(14<<2)|2,{58,84,0}},
/* 47550 */ {(14<<2)|2,{58,84,0}},
/* 47551 */ {(14<<2)|2,{58,84,0}},
/* 47552 */ {(14<<2)|2,{58,85,0}},
/* 47553 */ {(14<<2)|2,{58,85,0}},
/* 47554 */ {(14<<2)|2,{58,85,0}},
/* 47555 */ {(14<<2)|2,{58,85,0}},
/* 47556 */ {(14<<2)|2,{58,86,0}},
/* 47557 */ {(14<<2)|2,{58,86,0}},
/* 47558 */ {(14<<2)|2,{58,86,0}},
/* 47559 */ {(14<<2)|2,{58,86,0}},
/* 47560 */ {(14<<2)|2,{58,87,0}},
/* 47561 */ {(14<<2)|2,{58,87,0}},
/* 47562 */ {(14<<2)|2,{58,87,0}},
/* 47563 */ {(14<<2)|2,{58,87,0}},
/* 47564 */ {(14<<2)|2,{58,89,0}},
/* 47565 */ {(14<<2)|2,{58,89,0}},
/* 47566 */ {(14<<2)|2,{58,89,0}},
/* 47567 */ {(14<<2)|2,{58,89,0}},
/* 47568 */ {(14<<2)|2,{58,106,0}},
/* 47569 */ {(14<<2)|2,{58,106,0}},
/* 47570 */ {(14<<2)|2,{58,106,0}},
/* 47571 */ {(14<<2)|2,{58,106,0}},
/* 47572 */ {(14<<2)|2,{58,107,0}},
/* 47573 */ {(14<<2)|2,{58,107,0}},
/* 47574 */ {(14<<2)|2,{58,107,0}},
/* 47575 */ {(14<<2)|2,{58,107,0}},
/* 47576 */ {(14<<2)|2,{58,113,0}},
/* 47577 */ {(14<<2)|2,{58,113,0}},
/* 47578 */ {(14<<2)|2,{58,113,0}},
/* 47579 */ {(14<<2)|2,{58,113,0}},
/* 47580 */ {(14<<2)|2,{58,118,0}},
/* 47581 */ {(14<<2)|2,{58,118,0}},
/* 47582 */ {(14<<2)|2,{58,118,0}},
/* 47583 */ {(14<<2)|2,{58,118,0}},
/* 47584 */ {(14<<2)|2,{58,119,0}},
/* 47585 */ {(14<<2)|2,{58,119,0}},
/* 47586 */ {(14<<2)|2,{58,119,0}},
/* 47587 */ {(14<<2)|2,{58,119,0}},
/* 47588 */ {(14<<2)|2,{58,120,0}},
/* 47589 */ {(14<<2)|2,{58,120,0}},
/* 47590 */ {(14<<2)|2,{58,120,0}},
/* 47591 */ {(14<<2)|2,{58,120,0}},
/* 47592 */ {(14<<2)|2,{58,121,0}},
/* 47593 */ {(14<<2)|2,{58,121,0}},
/* 47594 */ {(14<<2)|2,{58,121,0}},
/* 47595 */ {(14<<2)|2,{58,121,0}},
/* 47596 */ {(14<<2)|2,{58,122,0}},
/* 47597 */ {(14<<2)|2,{58,122,0}},
/* 47598 */ {(14<<2)|2,{58,122,0}},
/* 47599 */ {(14<<2)|2,{58,122,0}},
/* 47600 */ {(15<<2)|2,{58,38,0}},
/* 47601 */ {(15<<2)|2,{58,38,0}},
/* 47602 */ {(15<<2)|2,{58,42,0}},
/* 47603 */ {(15<<2)|2,{58,42,0}},
/* 47604 */ {(15<<2)|2,{58,44,0}},
/* 47605 */ {(15<<2)|2,{58,44,0}},
/* 47606 */ {(15<<2)|2,{58,59,0}},
/* 47607 */ {(15<<2)|2,{58,59,0}},
/* 47608 */ {(15<<2)|2,{58,88,0}},
/* 47609 */ {(15<<2)|2,{58,88,0}},
/* 47610 */ {(15<<2)|2,{58,90,0}},
/* 47611 */ {(15<<2)|2,{58,90,0}},
/* 47612 */ {(7<<2)|1,{58,0,0}},
/* 47613 */ {(7<<2)|1,{58,0,0}},
/* 47614 */ {(7<<2)|1,{58,0,0}},
/* 47615 */ {(7<<2)|1,{58,0,0}},
/* 47616 */ {(12<<2)|2,{66,48,0}},
/* 47617 */ {(12<<2)|2,{66,48,0}},
/* 47618 */ {(12<<2)|2,{66,48,0}},
/* 47619 */ {(12<<2)|2,{66,48,0}},
/* 47620 */ {(12<<2)|2,{66,48,0}},
/* 47621 */ {(12<<2)|2,{66,48,0}},
/* 47622 */ {(12<<2)|2,{66,48,0}},
/* 47623 */ {(12<<2)|2,{66,48,0}},
/* 47624 */ {(12<<2)|2,{66,48,0}},
/* 47625 */ {(12<<2)|2,{66,48,0}},
/* 47626 */ {(12<<2)|2,{66,48,0}},
/* 47627 */ {(12<<2)|2,{66,48,0}},
/* 47628 */ {(12<<2)|2,{66,48,0}},
/* 47629 */ {(12<<2)|2,{66,48,0}},
/* 47630 */ {(12<<2)|2,{66,48,0}},
/* 47631 */ {(12<<2)|2,{66,48,0}},
/* 47632 */ {(12<<2)|2,{66,49,0}},
/* 47633 */ {(12<<2)|2,{66,49,0}},
/* 47634 */ {(12<<2)|2,{66,49,0}},
/* 47635 */ {(12<<2)|2,{66,49,0}},
/* 47636 */ {(12<<2)|2,{66,49,0}},
/* 47637 */ {(12<<2)|2,{66,49,0}},
/* 47638 */ {(12<<2)|2,{66,49,0}},
/* 47639 */ {(12<<2)|2,{66,49,0}},
/* 47640 */ {(12<<2)|2,{66,49,0}},
/* 47641 */ {(12<<2)|2,{66,49,0}},
/* 47642 */ {(12<<2)|2,{66,49,0}},
/* 47643 */ {(12<<2)|2,{66,49,0}},
/* 47644 */ {(12<<2)|2,{66,49,0}},
/* 47645 */ {(12<<2)|2,{66,49,0}},
/* 47646 */ {(12<<2)|2,{66,49,0}},
/* 47647 */ {(12<<2)|2,{66,49,0}},
/* 47648 */ {(12<<2)|2,{66,50,0}},
/* 47649 */ {(12<<2)|2,{66,50,0}},
/* 47650 */ {(12<<2)|2,{66,50,0}},
/* 47651 */ {(12<<2)|2,{66,50,0}},
/* 47652 */ {(12<<2)|2,{66,50,0}},
/* 47653 */ {(12<<2)|2,{66,50,0}},
/* 47654 */ {(12<<2)|2,{66,50,0}},
/* 47655 */ {(12<<2)|2,{66,50,0}},
/* 47656 */ {(12<<2)|2,{66,50,0}},
/* 47657 */ {(12<<2)|2,{66,50,0}},
/* 47658 */ {(12<<2)|2,{66,50,0}},
/* 47659 */ {(12<<2)|2,{66,50,0}},
/* 47660 */ {(12<<2)|2,{66,50,0}},
/* 47661 */ {(12<<2)|2,{66,50,0}},
/* 47662 */ {(12<<2)|2,{66,50,0}},
/* 47663 */ {(12<<2)|2,{66,50,0}},
/* 47664 */ {(12<<2)|2,{66,97,0}},
/* 47665 */ {(12<<2)|2,{66,97,0}},
/* 47666 */ {(12<<2)|2,{66,97,0}},
/* 47667 */ {(12<<2)|2,{66,97,0}},
/* 47668 */ {(12<<2)|2,{66,97,0}},
/* 47669 */ {(12<<2)|2,{66,97,0}},
/* 47670 */ {(12<<2)|2,{66,97,0}},
/* 47671 */ {(12<<2)|2,{66,97,0}},
/* 47672 */ {(12<<2)|2,{66,97,0}},
/* 47673 */ {(12<<2)|2,{66,97,0}},
/* 47674 */ {(12<<2)|2,{66,97,0}},
/* 47675 */ {(12<<2)|2,{66,97,0}},
/* 47676 */ {(12<<2)|2,{66,97,0}},
/* 47677 */ {(12<<2)|2,{66,97,0}},
/* 47678 */ {(12<<2)|2,{66,97,0}},
/* 47679 */ {(12<<2)|2,{66,97,0}},
/* 47680 */ {(12<<2)|2,{66,99,0}},
/* 47681 */ {(12<<2)|2,{66,99,0}},
/* 47682 */ {(12<<2)|2,{66,99,0}},
/* 47683 */ {(12<<2)|2,{66,99,0}},
/* 47684 */ {(12<<2)|2,{66,99,0}},
/* 47685 */ {(12<<2)|2,{66,99,0}},
/* 47686 */ {(12<<2)|2,{66,99,0}},
/* 47687 */ {(12<<2)|2,{66,99,0}},
/* 47688 */ {(12<<2)|2,{66,99,0}},
/* 47689 */ {(12<<2)|2,{66,99,0}},
/* 47690 */ {(12<<2)|2,{66,99,0}},
/* 47691 */ {(12<<2)|2,{66,99,0}},
/* 47692 */ {(12<<2)|2,{66,99,0}},
/* 47693 */ {(12<<2)|2,{66,99,0}},
/* 47694 */ {(12<<2)|2,{66,99,0}},
/* 47695 */ {(12<<2)|2,{66,99,0}},
/* 47696 */ {(12<<2)|2,{66,101,0}},
/* 47697 */ {(12<<2)|2,{66,101,0}},
/* 47698 */ {(12<<2)|2,{66,101,0}},
/* 47699 */ {(12<<2)|2,{66,101,0}},
/* 47700 */ {(12<<2)|2,{66,101,0}},
/* 47701 */ {(12<<2)|2,{66,101,0}},
/* 47702 */ {(12<<2)|2,{66,101,0}},
/* 47703 */ {(12<<2)|2,{66,101,0}},
/* 47704 */ {(12<<2)|2,{66,101,0}},
/* 47705 */ {(12<<2)|2,{66,101,0}},
/* 47706 */ {(12<<2)|2,{66,101,0}},
/* 47707 */ {(12<<2)|2,{66,101,0}},
/* 47708 */ {(12<<2)|2,{66,101,0}},
/* 47709 */ {(12<<2)|2,{66,101,0}},
/* 47710 */ {(12<<2)|2,{66,101,0}},
/* 47711 */ {(12<<2)|2,{66,101,0}},
/* 47712 */ {(12<<2)|2,{66,105,0}},
/* 47713 */ {(12<<2)|2,{66,105,0}},
/* 47714 */ {(12<<2)|2,{66,105,0}},
/* 47715 */ {(12<<2)|2,{66,105,0}},
/* 47716 */ {(12<<2)|2,{66,105,0}},
/* 47717 */ {(12<<2)|2,{66,105,0}},
/* 47718 */ {(12<<2)|2,{66,105,0}},
/* 47719 */ {(12<<2)|2,{66,105,0}},
/* 47720 */ {(12<<2)|2,{66,105,0}},
/* 47721 */ {(12<<2)|2,{66,105,0}},
/* 47722 */ {(12<<2)|2,{66,105,0}},
/* 47723 */ {(12<<2)|2,{66,105,0}},
/* 47724 */ {(12<<2)|2,{66,105,0}},
/* 47725 */ {(12<<2)|2,{66,105,0}},
/* 47726 */ {(12<<2)|2,{66,105,0}},
/* 47727 */ {(12<<2)|2,{66,105,0}},
/* 47728 */ {(12<<2)|2,{66,111,0}},
/* 47729 */ {(12<<2)|2,{66,111,0}},
/* 47730 */ {(12<<2)|2,{66,111,0}},
/* 47731 */ {(12<<2)|2,{66,111,0}},
/* 47732 */ {(12<<2)|2,{66,111,0}},
/* 47733 */ {(12<<2)|2,{66,111,0}},
/* 47734 */ {(12<<2)|2,{66,111,0}},
/* 47735 */ {(12<<2)|2,{66,111,0}},
/* 47736 */ {(12<<2)|2,{66,111,0}},
/* 47737 */ {(12<<2)|2,{66,111,0}},
/* 47738 */ {(12<<2)|2,{66,111,0}},
/* 47739 */ {(12<<2)|2,{66,111,0}},
/* 47740 */ {(12<<2)|2,{66,111,0}},
/* 47741 */ {(12<<2)|2,{66,111,0}},
/* 47742 */ {(12<<2)|2,{66,111,0}},
/* 47743 */ {(12<<2)|2,{66,111,0}},
/* 47744 */ {(12<<2)|2,{66,115,0}},
/* 47745 */ {(12<<2)|2,{66,115,0}},
/* 47746 */ {(12<<2)|2,{66,115,0}},
/* 47747 */ {(12<<2)|2,{66,115,0}},
/* 47748 */ {(12<<2)|2,{66,115,0}},
/* 47749 */ {(12<<2)|2,{66,115,0}},
/* 47750 */ {(12<<2)|2,{66,115,0}},
/* 47751 */ {(12<<2)|2,{66,115,0}},
/* 47752 */ {(12<<2)|2,{66,115,0}},
/* 47753 */ {(12<<2)|2,{66,115,0}},
/* 47754 */ {(12<<2)|2,{66,115,0}},
/* 47755 */ {(12<<2)|2,{66,115,0}},
/* 47756 */ {(12<<2)|2,{66,115,0}},
/* 47757 */ {(12<<2)|2,{66,115,0}},
/* 47758 */ {(12<<2)|2,{66,115,0}},
/* 47759 */ {(12<<2)|2,{66,115,0}},
/* 47760 */ {(12<<2)|2,{66,116,0}},
/* 47761 */ {(12<<2)|2,{66,116,0}},
/* 47762 */ {(12<<2)|2,{66,116,0}},
/* 47763 */ {(12<<2)|2,{66,116,0}},
/* 47764 */ {(12<<2)|2,{66,116,0}},
/* 47765 */ {(12<<2)|2,{66,116,0}},
/* 47766 */ {(12<<2)|2,{66,116,0}},
/* 47767 */ {(12<<2)|2,{66,116,0}},
/* 47768 */ {(12<<2)|2,{66,116,0}},
/* 47769 */ {(12<<2)|2,{66,116,0}},
/* 47770 */ {(12<<2)|2,{66,116,0}},
/* 47771 */ {(12<<2)|2,{66,116,0}},
/* 47772 */ {(12<<2)|2,{66,116,0}},
/* 47773 */ {(12<<2)|2,{66,116,0}},
/* 47774 */ {(12<<2)|2,{66,116,0}},
/* 47775 */ {(12<<2)|2,{66,116,0}},
/* 47776 */ {(13<<2)|2,{66,32,0}},
/* 47777 */ {(13<<2)|2,{66,32,0}},
/* 47778 */ {(13<<2)|2,{66,32,0}},
/* 47779 */ {(13<<2)|2,{66,32,0}},
/* 47780 */ {(13<<2)|2,{66,32,0}},
/* 47781 */ {(13<<2)|2,{66,32,0}},
/* 47782 */ {(13<<2)|2,{66,32,0}},
/* 47783 */ {(13<<2)|2,{66,32,0}},
/* 47784 */ {(13<<2)|2,{66,37,0}},
/* 47785 */ {(13<<2)|2,{66,37,0}},
/* 47786 */ {(13<<2)|2,{66,37,0}},
/* 47787 */ {(13<<2)|2,{66,37,0}},
/* 47788 */ {(13<<2)|2,{66,37,0}},
/* 47789 */ {(13<<2)|2,{66,37,0}},
/* 47790 */ {(13<<2)|2,{66,37,0}},
/* 47791 */ {(13<<2)|2,{66,37,0}},
/* 47792 */ {(13<<2)|2,{66,45,0}},
/* 47793 */ {(13<<2)|2,{66,45,0}},
/* 47794 */ {(13<<2)|2,{66,45,0}},
/* 47795 */ {(13<<2)|2,{66,45,0}},
/* 47796 */ {(13<<2)|2,{66,45,0}},
/* 47797 */ {(13<<2)|2,{66,45,0}},
/* 47798 */ {(13<<2)|2,{66,45,0}},
/* 47799 */ {(13<<2)|2,{66,45,0}},
/* 47800 */ {(13<<2)|2,{66,46,0}},
/* 47801 */ {(13<<2)|2,{66,46,0}},
/* 47802 */ {(13<<2)|2,{66,46,0}},
/* 47803 */ {(13<<2)|2,{66,46,0}},
/* 47804 */ {(13<<2)|2,{66,46,0}},
/* 47805 */ {(13<<2)|2,{66,46,0}},
/* 47806 */ {(13<<2)|2,{66,46,0}},
/* 47807 */ {(13<<2)|2,{66,46,0}},
/* 47808 */ {(13<<2)|2,{66,47,0}},
/* 47809 */ {(13<<2)|2,{66,47,0}},
/* 47810 */ {(13<<2)|2,{66,47,0}},
/* 47811 */ {(13<<2)|2,{66,47,0}},
/* 47812 */ {(13<<2)|2,{66,47,0}},
/* 47813 */ {(13<<2)|2,{66,47,0}},
/* 47814 */ {(13<<2)|2,{66,47,0}},
/* 47815 */ {(13<<2)|2,{66,47,0}},
/* 47816 */ {(13<<2)|2,{66,51,0}},
/* 47817 */ {(13<<2)|2,{66,51,0}},
/* 47818 */ {(13<<2)|2,{66,51,0}},
/* 47819 */ {(13<<2)|2,{66,51,0}},
/* 47820 */ {(13<<2)|2,{66,51,0}},
/* 47821 */ {(13<<2)|2,{66,51,0}},
/* 47822 */ {(13<<2)|2,{66,51,0}},
/* 47823 */ {(13<<2)|2,{66,51,0}},
/* 47824 */ {(13<<2)|2,{66,52,0}},
/* 47825 */ {(13<<2)|2,{66,52,0}},
/* 47826 */ {(13<<2)|2,{66,52,0}},
/* 47827 */ {(13<<2)|2,{66,52,0}},
/* 47828 */ {(13<<2)|2,{66,52,0}},
/* 47829 */ {(13<<2)|2,{66,52,0}},
/* 47830 */ {(13<<2)|2,{66,52,0}},
/* 47831 */ {(13<<2)|2,{66,52,0}},
/* 47832 */ {(13<<2)|2,{66,53,0}},
/* 47833 */ {(13<<2)|2,{66,53,0}},
/* 47834 */ {(13<<2)|2,{66,53,0}},
/* 47835 */ {(13<<2)|2,{66,53,0}},
/* 47836 */ {(13<<2)|2,{66,53,0}},
/* 47837 */ {(13<<2)|2,{66,53,0}},
/* 47838 */ {(13<<2)|2,{66,53,0}},
/* 47839 */ {(13<<2)|2,{66,53,0}},
/* 47840 */ {(13<<2)|2,{66,54,0}},
/* 47841 */ {(13<<2)|2,{66,54,0}},
/* 47842 */ {(13<<2)|2,{66,54,0}},
/* 47843 */ {(13<<2)|2,{66,54,0}},
/* 47844 */ {(13<<2)|2,{66,54,0}},
/* 47845 */ {(13<<2)|2,{66,54,0}},
/* 47846 */ {(13<<2)|2,{66,54,0}},
/* 47847 */ {(13<<2)|2,{66,54,0}},
/* 47848 */ {(13<<2)|2,{66,55,0}},
/* 47849 */ {(13<<2)|2,{66,55,0}},
/* 47850 */ {(13<<2)|2,{66,55,0}},
/* 47851 */ {(13<<2)|2,{66,55,0}},
/* 47852 */ {(13<<2)|2,{66,55,0}},
/* 47853 */ {(13<<2)|2,{66,55,0}},
/* 47854 */ {(13<<2)|2,{66,55,0}},
/* 47855 */ {(13<<2)|2,{66,55,0}},
/* 47856 */ {(13<<2)|2,{66,56,0}},
/* 47857 */ {(13<<2)|2,{66,56,0}},
/* 47858 */ {(13<<2)|2,{66,56,0}},
/* 47859 */ {(13<<2)|2,{66,56,0}},
/* 47860 */ {(13<<2)|2,{66,56,0}},
/* 47861 */ {(13<<2)|2,{66,56,0}},
/* 47862 */ {(13<<2)|2,{66,56,0}},
/* 47863 */ {(13<<2)|2,{66,56,0}},
/* 47864 */ {(13<<2)|2,{66,57,0}},
/* 47865 */ {(13<<2)|2,{66,57,0}},
/* 47866 */ {(13<<2)|2,{66,57,0}},
/* 47867 */ {(13<<2)|2,{66,57,0}},
/* 47868 */ {(13<<2)|2,{66,57,0}},
/* 47869 */ {(13<<2)|2,{66,57,0}},
/* 47870 */ {(13<<2)|2,{66,57,0}},
/* 47871 */ {(13<<2)|2,{66,57,0}},
/* 47872 */ {(13<<2)|2,{66,61,0}},
/* 47873 */ {(13<<2)|2,{66,61,0}},
/* 47874 */ {(13<<2)|2,{66,61,0}},
/* 47875 */ {(13<<2)|2,{66,61,0}},
/* 47876 */ {(13<<2)|2,{66,61,0}},
/* 47877 */ {(13<<2)|2,{66,61,0}},
/* 47878 */ {(13<<2)|2,{66,61,0}},
/* 47879 */ {(13<<2)|2,{66,61,0}},
/* 47880 */ {(13<<2)|2,{66,65,0}},
/* 47881 */ {(13<<2)|2,{66,65,0}},
/* 47882 */ {(13<<2)|2,{66,65,0}},
/* 47883 */ {(13<<2)|2,{66,65,0}},
/* 47884 */ {(13<<2)|2,{66,65,0}},
/* 47885 */ {(13<<2)|2,{66,65,0}},
/* 47886 */ {(13<<2)|2,{66,65,0}},
/* 47887 */ {(13<<2)|2,{66,65,0}},
/* 47888 */ {(13<<2)|2,{66,95,0}},
/* 47889 */ {(13<<2)|2,{66,95,0}},
/* 47890 */ {(13<<2)|2,{66,95,0}},
/* 47891 */ {(13<<2)|2,{66,95,0}},
/* 47892 */ {(13<<2)|2,{66,95,0}},
/* 47893 */ {(13<<2)|2,{66,95,0}},
/* 47894 */ {(13<<2)|2,{66,95,0}},
/* 47895 */ {(13<<2)|2,{66,95,0}},
/* 47896 */ {(13<<2)|2,{66,98,0}},
/* 47897 */ {(13<<2)|2,{66,98,0}},
/* 47898 */ {(13<<2)|2,{66,98,0}},
/* 47899 */ {(13<<2)|2,{66,98,0}},
/* 47900 */ {(13<<2)|2,{66,98,0}},
/* 47901 */ {(13<<2)|2,{66,98,0}},
/* 47902 */ {(13<<2)|2,{66,98,0}},
/* 47903 */ {(13<<2)|2,{66,98,0}},
/* 47904 */ {(13<<2)|2,{66,100,0}},
/* 47905 */ {(13<<2)|2,{66,100,0}},
/* 47906 */ {(13<<2)|2,{66,100,0}},
/* 47907 */ {(13<<2)|2,{66,100,0}},
/* 47908 */ {(13<<2)|2,{66,100,0}},
/* 47909 */ {(13<<2)|2,{66,100,0}},
/* 47910 */ {(13<<2)|2,{66,100,0}},
/* 47911 */ {(13<<2)|2,{66,100,0}},
/* 47912 */ {(13<<2)|2,{66,102,0}},
/* 47913 */ {(13<<2)|2,{66,102,0}},
/* 47914 */ {(13<<2)|2,{66,102,0}},
/* 47915 */ {(13<<2)|2,{66,102,0}},
/* 47916 */ {(13<<2)|2,{66,102,0}},
/* 47917 */ {(13<<2)|2,{66,102,0}},
/* 47918 */ {(13<<2)|2,{66,102,0}},
/* 47919 */ {(13<<2)|2,{66,102,0}},
/* 47920 */ {(13<<2)|2,{66,103,0}},
/* 47921 */ {(13<<2)|2,{66,103,0}},
/* 47922 */ {(13<<2)|2,{66,103,0}},
/* 47923 */ {(13<<2)|2,{66,103,0}},
/* 47924 */ {(13<<2)|2,{66,103,0}},
/* 47925 */ {(13<<2)|2,{66,103,0}},
/* 47926 */ {(13<<2)|2,{66,103,0}},
/* 47927 */ {(13<<2)|2,{66,103,0}},
/* 47928 */ {(13<<2)|2,{66,104,0}},
/* 47929 */ {(13<<2)|2,{66,104,0}},
/* 47930 */ {(13<<2)|2,{66,104,0}},
/* 47931 */ {(13<<2)|2,{66,104,0}},
/* 47932 */ {(13<<2)|2,{66,104,0}},
/* 47933 */ {(13<<2)|2,{66,104,0}},
/* 47934 */ {(13<<2)|2,{66,104,0}},
/* 47935 */ {(13<<2)|2,{66,104,0}},
/* 47936 */ {(13<<2)|2,{66,108,0}},
/* 47937 */ {(13<<2)|2,{66,108,0}},
/* 47938 */ {(13<<2)|2,{66,108,0}},
/* 47939 */ {(13<<2)|2,{66,108,0}},
/* 47940 */ {(13<<2)|2,{66,108,0}},
/* 47941 */ {(13<<2)|2,{66,108,0}},
/* 47942 */ {(13<<2)|2,{66,108,0}},
/* 47943 */ {(13<<2)|2,{66,108,0}},
/* 47944 */ {(13<<2)|2,{66,109,0}},
/* 47945 */ {(13<<2)|2,{66,109,0}},
/* 47946 */ {(13<<2)|2,{66,109,0}},
/* 47947 */ {(13<<2)|2,{66,109,0}},
/* 47948 */ {(13<<2)|2,{66,109,0}},
/* 47949 */ {(13<<2)|2,{66,109,0}},
/* 47950 */ {(13<<2)|2,{66,109,0}},
/* 47951 */ {(13<<2)|2,{66,109,0}},
/* 47952 */ {(13<<2)|2,{66,110,0}},
/* 47953 */ {(13<<2)|2,{66,110,0}},
/* 47954 */ {(13<<2)|2,{66,110,0}},
/* 47955 */ {(13<<2)|2,{66,110,0}},
/* 47956 */ {(13<<2)|2,{66,110,0}},
/* 47957 */ {(13<<2)|2,{66,110,0}},
/* 47958 */ {(13<<2)|2,{66,110,0}},
/* 47959 */ {(13<<2)|2,{66,110,0}},
/* 47960 */ {(13<<2)|2,{66,112,0}},
/* 47961 */ {(13<<2)|2,{66,112,0}},
/* 47962 */ {(13<<2)|2,{66,112,0}},
/* 47963 */ {(13<<2)|2,{66,112,0}},
/* 47964 */ {(13<<2)|2,{66,112,0}},
/* 47965 */ {(13<<2)|2,{66,112,0}},
/* 47966 */ {(13<<2)|2,{66,112,0}},
/* 47967 */ {(13<<2)|2,{66,112,0}},
/* 47968 */ {(13<<2)|2,{66,114,0}},
/* 47969 */ {(13<<2)|2,{66,114,0}},
/* 47970 */ {(13<<2)|2,{66,114,0}},
/* 47971 */ {(13<<2)|2,{66,114,0}},
/* 47972 */ {(13<<2)|2,{66,114,0}},
/* 47973 */ {(13<<2)|2,{66,114,0}},
/* 47974 */ {(13<<2)|2,{66,114,0}},
/* 47975 */ {(13<<2)|2,{66,114,0}},
/* 47976 */ {(13<<2)|2,{66,117,0}},
/* 47977 */ {(13<<2)|2,{66,117,0}},
/* 47978 */ {(13<<2)|2,{66,117,0}},
/* 47979 */ {(13<<2)|2,{66,117,0}},
/* 47980 */ {(13<<2)|2,{66,117,0}},
/* 47981 */ {(13<<2)|2,{66,117,0}},
/* 47982 */ {(13<<2)|2,{66,117,0}},
/* 47983 */ {(13<<2)|2,{66,117,0}},
/* 47984 */ {(14<<2)|2,{66,58,0}},
/* 47985 */ {(14<<2)|2,{66,58,0}},
/* 47986 */ {(14<<2)|2,{66,58,0}},
/* 47987 */ {(14<<2)|2,{66,58,0}},
/* 47988 */ {(14<<2)|2,{66,66,0}},
/* 47989 */ {(14<<2)|2,{66,66,0}},
/* 47990 */ {(14<<2)|2,{66,66,0}},
/* 47991 */ {(14<<2)|2,{66,66,0}},
/* 47992 */ {(14<<2)|2,{66,67,0}},
/* 47993 */ {(14<<2)|2,{66,67,0}},
/* 47994 */ {(14<<2)|2,{66,67,0}},
/* 47995 */ {(14<<2)|2,{66,67,0}},
/* 47996 */ {(14<<2)|2,{66,68,0}},
/* 47997 */ {(14<<2)|2,{66,68,0}},
/* 47998 */ {(14<<2)|2,{66,68,0}},
/* 47999 */ {(14<<2)|2,{66,68,0}},
/* 48000 */ {(14<<2)|2,{66,69,0}},
/* 48001 */ {(14<<2)|2,{66,69,0}},
/* 48002 */ {(14<<2)|2,{66,69,0}},
/* 48003 */ {(14<<2)|2,{66,69,0}},
/* 48004 */ {(14<<2)|2,{66,70,0}},
/* 48005 */ {(14<<2)|2,{66,70,0}},
/* 48006 */ {(14<<2)|2,{66,70,0}},
/* 48007 */ {(14<<2)|2,{66,70,0}},
/* 48008 */ {(14<<2)|2,{66,71,0}},
/* 48009 */ {(14<<2)|2,{66,71,0}},
/* 48010 */ {(14<<2)|2,{66,71,0}},
/* 48011 */ {(14<<2)|2,{66,71,0}},
/* 48012 */ {(14<<2)|2,{66,72,0}},
/* 48013 */ {(14<<2)|2,{66,72,0}},
/* 48014 */ {(14<<2)|2,{66,72,0}},
/* 48015 */ {(14<<2)|2,{66,72,0}},
/* 48016 */ {(14<<2)|2,{66,73,0}},
/* 48017 */ {(14<<2)|2,{66,73,0}},
/* 48018 */ {(14<<2)|2,{66,73,0}},
/* 48019 */ {(14<<2)|2,{66,73,0}},
/* 48020 */ {(14<<2)|2,{66,74,0}},
/* 48021 */ {(14<<2)|2,{66,74,0}},
/* 48022 */ {(14<<2)|2,{66,74,0}},
/* 48023 */ {(14<<2)|2,{66,74,0}},
/* 48024 */ {(14<<2)|2,{66,75,0}},
/* 48025 */ {(14<<2)|2,{66,75,0}},
/* 48026 */ {(14<<2)|2,{66,75,0}},
/* 48027 */ {(14<<2)|2,{66,75,0}},
/* 48028 */ {(14<<2)|2,{66,76,0}},
/* 48029 */ {(14<<2)|2,{66,76,0}},
/* 48030 */ {(14<<2)|2,{66,76,0}},
/* 48031 */ {(14<<2)|2,{66,76,0}},
/* 48032 */ {(14<<2)|2,{66,77,0}},
/* 48033 */ {(14<<2)|2,{66,77,0}},
/* 48034 */ {(14<<2)|2,{66,77,0}},
/* 48035 */ {(14<<2)|2,{66,77,0}},
/* 48036 */ {(14<<2)|2,{66,78,0}},
/* 48037 */ {(14<<2)|2,{66,78,0}},
/* 48038 */ {(14<<2)|2,{66,78,0}},
/* 48039 */ {(14<<2)|2,{66,78,0}},
/* 48040 */ {(14<<2)|2,{66,79,0}},
/* 48041 */ {(14<<2)|2,{66,79,0}},
/* 48042 */ {(14<<2)|2,{66,79,0}},
/* 48043 */ {(14<<2)|2,{66,79,0}},
/* 48044 */ {(14<<2)|2,{66,80,0}},
/* 48045 */ {(14<<2)|2,{66,80,0}},
/* 48046 */ {(14<<2)|2,{66,80,0}},
/* 48047 */ {(14<<2)|2,{66,80,0}},
/* 48048 */ {(14<<2)|2,{66,81,0}},
/* 48049 */ {(14<<2)|2,{66,81,0}},
/* 48050 */ {(14<<2)|2,{66,81,0}},
/* 48051 */ {(14<<2)|2,{66,81,0}},
/* 48052 */ {(14<<2)|2,{66,82,0}},
/* 48053 */ {(14<<2)|2,{66,82,0}},
/* 48054 */ {(14<<2)|2,{66,82,0}},
/* 48055 */ {(14<<2)|2,{66,82,0}},
/* 48056 */ {(14<<2)|2,{66,83,0}},
/* 48057 */ {(14<<2)|2,{66,83,0}},
/* 48058 */ {(14<<2)|2,{66,83,0}},
/* 48059 */ {(14<<2)|2,{66,83,0}},
/* 48060 */ {(14<<2)|2,{66,84,0}},
/* 48061 */ {(14<<2)|2,{66,84,0}},
/* 48062 */ {(14<<2)|2,{66,84,0}},
/* 48063 */ {(14<<2)|2,{66,84,0}},
/* 48064 */ {(14<<2)|2,{66,85,0}},
/* 48065 */ {(14<<2)|2,{66,85,0}},
/* 48066 */ {(14<<2)|2,{66,85,0}},
/* 48067 */ {(14<<2)|2,{66,85,0}},
/* 48068 */ {(14<<2)|2,{66,86,0}},
/* 48069 */ {(14<<2)|2,{66,86,0}},
/* 48070 */ {(14<<2)|2,{66,86,0}},
/* 48071 */ {(14<<2)|2,{66,86,0}},
/* 48072 */ {(14<<2)|2,{66,87,0}},
/* 48073 */ {(14<<2)|2,{66,87,0}},
/* 48074 */ {(14<<2)|2,{66,87,0}},
/* 48075 */ {(14<<2)|2,{66,87,0}},
/* 48076 */ {(14<<2)|2,{66,89,0}},
/* 48077 */ {(14<<2)|2,{66,89,0}},
/* 48078 */ {(14<<2)|2,{66,89,0}},
/* 48079 */ {(14<<2)|2,{66,89,0}},
/* 48080 */ {(14<<2)|2,{66,106,0}},
/* 48081 */ {(14<<2)|2,{66,106,0}},
/* 48082 */ {(14<<2)|2,{66,106,0}},
/* 48083 */ {(14<<2)|2,{66,106,0}},
/* 48084 */ {(14<<2)|2,{66,107,0}},
/* 48085 */ {(14<<2)|2,{66,107,0}},
/* 48086 */ {(14<<2)|2,{66,107,0}},
/* 48087 */ {(14<<2)|2,{66,107,0}},
/* 48088 */ {(14<<2)|2,{66,113,0}},
/* 48089 */ {(14<<2)|2,{66,113,0}},
/* 48090 */ {(14<<2)|2,{66,113,0}},
/* 48091 */ {(14<<2)|2,{66,113,0}},
/* 48092 */ {(14<<2)|2,{66,118,0}},
/* 48093 */ {(14<<2)|2,{66,118,0}},
/* 48094 */ {(14<<2)|2,{66,118,0}},
/* 48095 */ {(14<<2)|2,{66,118,0}},
/* 48096 */ {(14<<2)|2,{66,119,0}},
/* 48097 */ {(14<<2)|2,{66,119,0}},
/* 48098 */ {(14<<2)|2,{66,119,0}},
/* 48099 */ {(14<<2)|2,{66,119,0}},
/* 48100 */ {(14<<2)|2,{66,120,0}},
/* 48101 */ {(14<<2)|2,{66,120,0}},
/* 48102 */ {(14<<2)|2,{66,120,0}},
/* 48103 */ {(14<<2)|2,{66,120,0}},
/* 48104 */ {(14<<2)|2,{66,121,0}},
/* 48105 */ {(14<<2)|2,{66,121,0}},
/* 48106 */ {(14<<2)|2,{66,121,0}},
/* 48107 */ {(14<<2)|2,{66,121,0}},
/* 48108 */ {(14<<2)|2,{66,122,0}},
/* 48109 */ {(14<<2)|2,{66,122,0}},
/* 48110 */ {(14<<2)|2,{66,122,0}},
/* 48111 */ {(14<<2)|2,{66,122,0}},
/* 48112 */ {(15<<2)|2,{66,38,0}},
/* 48113 */ {(15<<2)|2,{66,38,0}},
/* 48114 */ {(15<<2)|2,{66,42,0}},
/* 48115 */ {(15<<2)|2,{66,42,0}},
/* 48116 */ {(15<<2)|2,{66,44,0}},
/* 48117 */ {(15<<2)|2,{66,44,0}},
/* 48118 */ {(15<<2)|2,{66,59,0}},
/* 48119 */ {(15<<2)|2,{66,59,0}},
/* 48120 */ {(15<<2)|2,{66,88,0}},
/* 48121 */ {(15<<2)|2,{66,88,0}},
/* 48122 */ {(15<<2)|2,{66,90,0}},
/* 48123 */ {(15<<2)|2,{66,90,0}},
/* 48124 */ {(7<<2)|1,{66,0,0}},
/* 48125 */ {(7<<2)|1,{66,0,0}},
/* 48126 */ {(7<<2)|1,{66,0,0}},
/* 48127 */ {(7<<2)|1,{66,0,0}},
/* 48128 */ {(12<<2)|2,{67,48,0}},
/* 48129 */ {(12<<2)|2,{67,48,0}},
/* 48130 */ {(12<<2)|2,{67,48,0}},
/* 48131 */ {(12<<2)|2,{67,48,0}},
/* 48132 */ {(12<<2)|2,{67,48,0}},
/* 48133 */ {(12<<2)|2,{67,48,0}},
/* 48134 */ {(12<<2)|2,{67,48,0}},
/* 48135 */ {(12<<2)|2,{67,48,0}},
/* 48136 */ {(12<<2)|2,{67,48,0}},
/* 48137 */ {(12<<2)|2,{67,48,0}},
/* 48138 */ {(12<<2)|2,{67,48,0}},
/* 48139 */ {(12<<2)|2,{67,48,0}},
/* 48140 */ {(12<<2)|2,{67,48,0}},
/* 48141 */ {(12<<2)|2,{67,48,0}},
/* 48142 */ {(12<<2)|2,{67,48,0}},
/* 48143 */ {(12<<2)|2,{67,48,0}},
/* 48144 */ {(12<<2)|2,{67,49,0}},
/* 48145 */ {(12<<2)|2,{67,49,0}},
/* 48146 */ {(12<<2)|2,{67,49,0}},
/* 48147 */ {(12<<2)|2,{67,49,0}},
/* 48148 */ {(12<<2)|2,{67,49,0}},
/* 48149 */ {(12<<2)|2,{67,49,0}},
/* 48150 */ {(12<<2)|2,{67,49,0}},
/* 48151 */ {(12<<2)|2,{67,49,0}},
/* 48152 */ {(12<<2)|2,{67,49,0}},
/* 48153 */ {(12<<2)|2,{67,49,0}},
/* 48154 */ {(12<<2)|2,{67,49,0}},
/* 48155 */ {(12<<2)|2,{67,49,0}},
/* 48156 */ {(12<<2)|2,{67,49,0}},
/* 48157 */ {(12<<2)|2,{67,49,0}},
/* 48158 */ {(12<<2)|2,{67,49,0}},
/* 48159 */ {(12<<2)|2,{67,49,0}},
/* 48160 */ {(12<<2)|2,{67,50,0}},
/* 48161 */ {(12<<2)|2,{67,50,0}},
/* 48162 */ {(12<<2)|2,{67,50,0}},
/* 48163 */ {(12<<2)|2,{67,50,0}},
/* 48164 */ {(12<<2)|2,{67,50,0}},
/* 48165 */ {(12<<2)|2,{67,50,0}},
/* 48166 */ {(12<<2)|2,{67,50,0}},
/* 48167 */ {(12<<2)|2,{67,50,0}},
/* 48168 */ {(12<<2)|2,{67,50,0}},
/* 48169 */ {(12<<2)|2,{67,50,0}},
/* 48170 */ {(12<<2)|2,{67,50,0}},
/* 48171 */ {(12<<2)|2,{67,50,0}},
/* 48172 */ {(12<<2)|2,{67,50,0}},
/* 48173 */ {(12<<2)|2,{67,50,0}},
/* 48174 */ {(12<<2)|2,{67,50,0}},
/* 48175 */ {(12<<2)|2,{67,50,0}},
/* 48176 */ {(12<<2)|2,{67,97,0}},
/* 48177 */ {(12<<2)|2,{67,97,0}},
/* 48178 */ {(12<<2)|2,{67,97,0}},
/* 48179 */ {(12<<2)|2,{67,97,0}},
/* 48180 */ {(12<<2)|2,{67,97,0}},
/* 48181 */ {(12<<2)|2,{67,97,0}},
/* 48182 */ {(12<<2)|2,{67,97,0}},
/* 48183 */ {(12<<2)|2,{67,97,0}},
/* 48184 */ {(12<<2)|2,{67,97,0}},
/* 48185 */ {(12<<2)|2,{67,97,0}},
/* 48186 */ {(12<<2)|2,{67,97,0}},
/* 48187 */ {(12<<2)|2,{67,97,0}},
/* 48188 */ {(12<<2)|2,{67,97,0}},
/* 48189 */ {(12<<2)|2,{67,97,0}},
/* 48190 */ {(12<<2)|2,{67,97,0}},
/* 48191 */ {(12<<2)|2,{67,97,0}},
/* 48192 */ {(12<<2)|2,{67,99,0}},
/* 48193 */ {(12<<2)|2,{67,99,0}},
/* 48194 */ {(12<<2)|2,{67,99,0}},
/* 48195 */ {(12<<2)|2,{67,99,0}},
/* 48196 */ {(12<<2)|2,{67,99,0}},
/* 48197 */ {(12<<2)|2,{67,99,0}},
/* 48198 */ {(12<<2)|2,{67,99,0}},
/* 48199 */ {(12<<2)|2,{67,99,0}},
/* 48200 */ {(12<<2)|2,{67,99,0}},
/* 48201 */ {(12<<2)|2,{67,99,0}},
/* 48202 */ {(12<<2)|2,{67,99,0}},
/* 48203 */ {(12<<2)|2,{67,99,0}},
/* 48204 */ {(12<<2)|2,{67,99,0}},
/* 48205 */ {(12<<2)|2,{67,99,0}},
/* 48206 */ {(12<<2)|2,{67,99,0}},
/* 48207 */ {(12<<2)|2,{67,99,0}},
/* 48208 */ {(12<<2)|2,{67,101,0}},
/* 48209 */ {(12<<2)|2,{67,101,0}},
/* 48210 */ {(12<<2)|2,{67,101,0}},
/* 48211 */ {(12<<2)|2,{67,101,0}},
/* 48212 */ {(12<<2)|2,{67,101,0}},
/* 48213 */ {(12<<2)|2,{67,101,0}},
/* 48214 */ {(12<<2)|2,{67,101,0}},
/* 48215 */ {(12<<2)|2,{67,101,0}},
/* 48216 */ {(12<<2)|2,{67,101,0}},
/* 48217 */ {(12<<2)|2,{67,101,0}},
/* 48218 */ {(12<<2)|2,{67,101,0}},
/* 48219 */ {(12<<2)|2,{67,101,0}},
/* 48220 */ {(12<<2)|2,{67,101,0}},
/* 48221 */ {(12<<2)|2,{67,101,0}},
/* 48222 */ {(12<<2)|2,{67,101,0}},
/* 48223 */ {(12<<2)|2,{67,101,0}},
/* 48224 */ {(12<<2)|2,{67,105,0}},
/* 48225 */ {(12<<2)|2,{67,105,0}},
/* 48226 */ {(12<<2)|2,{67,105,0}},
/* 48227 */ {(12<<2)|2,{67,105,0}},
/* 48228 */ {(12<<2)|2,{67,105,0}},
/* 48229 */ {(12<<2)|2,{67,105,0}},
/* 48230 */ {(12<<2)|2,{67,105,0}},
/* 48231 */ {(12<<2)|2,{67,105,0}},
/* 48232 */ {(12<<2)|2,{67,105,0}},
/* 48233 */ {(12<<2)|2,{67,105,0}},
/* 48234 */ {(12<<2)|2,{67,105,0}},
/* 48235 */ {(12<<2)|2,{67,105,0}},
/* 48236 */ {(12<<2)|2,{67,105,0}},
/* 48237 */ {(12<<2)|2,{67,105,0}},
/* 48238 */ {(12<<2)|2,{67,105,0}},
/* 48239 */ {(12<<2)|2,{67,105,0}},
/* 48240 */ {(12<<2)|2,{67,111,0}},
/* 48241 */ {(12<<2)|2,{67,111,0}},
/* 48242 */ {(12<<2)|2,{67,111,0}},
/* 48243 */ {(12<<2)|2,{67,111,0}},
/* 48244 */ {(12<<2)|2,{67,111,0}},
/* 48245 */ {(12<<2)|2,{67,111,0}},
/* 48246 */ {(12<<2)|2,{67,111,0}},
/* 48247 */ {(12<<2)|2,{67,111,0}},
/* 48248 */ {(12<<2)|2,{67,111,0}},
/* 48249 */ {(12<<2)|2,{67,111,0}},
/* 48250 */ {(12<<2)|2,{67,111,0}},
/* 48251 */ {(12<<2)|2,{67,111,0}},
/* 48252 */ {(12<<2)|2,{67,111,0}},
/* 48253 */ {(12<<2)|2,{67,111,0}},
/* 48254 */ {(12<<2)|2,{67,111,0}},
/* 48255 */ {(12<<2)|2,{67,111,0}},
/* 48256 */ {(12<<2)|2,{67,115,0}},
/* 48257 */ {(12<<2)|2,{67,115,0}},
/* 48258 */ {(12<<2)|2,{67,115,0}},
/* 48259 */ {(12<<2)|2,{67,115,0}},
/* 48260 */ {(12<<2)|2,{67,115,0}},
/* 48261 */ {(12<<2)|2,{67,115,0}},
/* 48262 */ {(12<<2)|2,{67,115,0}},
/* 48263 */ {(12<<2)|2,{67,115,0}},
/* 48264 */ {(12<<2)|2,{67,115,0}},
/* 48265 */ {(12<<2)|2,{67,115,0}},
/* 48266 */ {(12<<2)|2,{67,115,0}},
/* 48267 */ {(12<<2)|2,{67,115,0}},
/* 48268 */ {(12<<2)|2,{67,115,0}},
/* 48269 */ {(12<<2)|2,{67,115,0}},
/* 48270 */ {(12<<2)|2,{67,115,0}},
/* 48271 */ {(12<<2)|2,{67,115,0}},
/* 48272 */ {(12<<2)|2,{67,116,0}},
/* 48273 */ {(12<<2)|2,{67,116,0}},
/* 48274 */ {(12<<2)|2,{67,116,0}},
/* 48275 */ {(12<<2)|2,{67,116,0}},
/* 48276 */ {(12<<2)|2,{67,116,0}},
/* 48277 */ {(12<<2)|2,{67,116,0}},
/* 48278 */ {(12<<2)|2,{67,116,0}},
/* 48279 */ {(12<<2)|2,{67,116,0}},
/* 48280 */ {(12<<2)|2,{67,116,0}},
/* 48281 */ {(12<<2)|2,{67,116,0}},
/* 48282 */ {(12<<2)|2,{67,116,0}},
/* 48283 */ {(12<<2)|2,{67,116,0}},
/* 48284 */ {(12<<2)|2,{67,116,0}},
/* 48285 */ {(12<<2)|2,{67,116,0}},
/* 48286 */ {(12<<2)|2,{67,116,0}},
/* 48287 */ {(12<<2)|2,{67,116,0}},
/* 48288 */ {(13<<2)|2,{67,32,0}},
/* 48289 */ {(13<<2)|2,{67,32,0}},
/* 48290 */ {(13<<2)|2,{67,32,0}},
/* 48291 */ {(13<<2)|2,{67,32,0}},
/* 48292 */ {(13<<2)|2,{67,32,0}},
/* 48293 */ {(13<<2)|2,{67,32,0}},
/* 48294 */ {(13<<2)|2,{67,32,0}},
/* 48295 */ {(13<<2)|2,{67,32,0}},
/* 48296 */ {(13<<2)|2,{67,37,0}},
/* 48297 */ {(13<<2)|2,{67,37,0}},
/* 48298 */ {(13<<2)|2,{67,37,0}},
/* 48299 */ {(13<<2)|2,{67,37,0}},
/* 48300 */ {(13<<2)|2,{67,37,0}},
/* 48301 */ {(13<<2)|2,{67,37,0}},
/* 48302 */ {(13<<2)|2,{67,37,0}},
/* 48303 */ {(13<<2)|2,{67,37,0}},
/* 48304 */ {(13<<2)|2,{67,45,0}},
/* 48305 */ {(13<<2)|2,{67,45,0}},
/* 48306 */ {(13<<2)|2,{67,45,0}},
/* 48307 */ {(13<<2)|2,{67,45,0}},
/* 48308 */ {(13<<2)|2,{67,45,0}},
/* 48309 */ {(13<<2)|2,{67,45,0}},
/* 48310 */ {(13<<2)|2,{67,45,0}},
/* 48311 */ {(13<<2)|2,{67,45,0}},
/* 48312 */ {(13<<2)|2,{67,46,0}},
/* 48313 */ {(13<<2)|2,{67,46,0}},
/* 48314 */ {(13<<2)|2,{67,46,0}},
/* 48315 */ {(13<<2)|2,{67,46,0}},
/* 48316 */ {(13<<2)|2,{67,46,0}},
/* 48317 */ {(13<<2)|2,{67,46,0}},
/* 48318 */ {(13<<2)|2,{67,46,0}},
/* 48319 */ {(13<<2)|2,{67,46,0}},
/* 48320 */ {(13<<2)|2,{67,47,0}},
/* 48321 */ {(13<<2)|2,{67,47,0}},
/* 48322 */ {(13<<2)|2,{67,47,0}},
/* 48323 */ {(13<<2)|2,{67,47,0}},
/* 48324 */ {(13<<2)|2,{67,47,0}},
/* 48325 */ {(13<<2)|2,{67,47,0}},
/* 48326 */ {(13<<2)|2,{67,47,0}},
/* 48327 */ {(13<<2)|2,{67,47,0}},
/* 48328 */ {(13<<2)|2,{67,51,0}},
/* 48329 */ {(13<<2)|2,{67,51,0}},
/* 48330 */ {(13<<2)|2,{67,51,0}},
/* 48331 */ {(13<<2)|2,{67,51,0}},
/* 48332 */ {(13<<2)|2,{67,51,0}},
/* 48333 */ {(13<<2)|2,{67,51,0}},
/* 48334 */ {(13<<2)|2,{67,51,0}},
/* 48335 */ {(13<<2)|2,{67,51,0}},
/* 48336 */ {(13<<2)|2,{67,52,0}},
/* 48337 */ {(13<<2)|2,{67,52,0}},
/* 48338 */ {(13<<2)|2,{67,52,0}},
/* 48339 */ {(13<<2)|2,{67,52,0}},
/* 48340 */ {(13<<2)|2,{67,52,0}},
/* 48341 */ {(13<<2)|2,{67,52,0}},
/* 48342 */ {(13<<2)|2,{67,52,0}},
/* 48343 */ {(13<<2)|2,{67,52,0}},
/* 48344 */ {(13<<2)|2,{67,53,0}},
/* 48345 */ {(13<<2)|2,{67,53,0}},
/* 48346 */ {(13<<2)|2,{67,53,0}},
/* 48347 */ {(13<<2)|2,{67,53,0}},
/* 48348 */ {(13<<2)|2,{67,53,0}},
/* 48349 */ {(13<<2)|2,{67,53,0}},
/* 48350 */ {(13<<2)|2,{67,53,0}},
/* 48351 */ {(13<<2)|2,{67,53,0}},
/* 48352 */ {(13<<2)|2,{67,54,0}},
/* 48353 */ {(13<<2)|2,{67,54,0}},
/* 48354 */ {(13<<2)|2,{67,54,0}},
/* 48355 */ {(13<<2)|2,{67,54,0}},
/* 48356 */ {(13<<2)|2,{67,54,0}},
/* 48357 */ {(13<<2)|2,{67,54,0}},
/* 48358 */ {(13<<2)|2,{67,54,0}},
/* 48359 */ {(13<<2)|2,{67,54,0}},
/* 48360 */ {(13<<2)|2,{67,55,0}},
/* 48361 */ {(13<<2)|2,{67,55,0}},
/* 48362 */ {(13<<2)|2,{67,55,0}},
/* 48363 */ {(13<<2)|2,{67,55,0}},
/* 48364 */ {(13<<2)|2,{67,55,0}},
/* 48365 */ {(13<<2)|2,{67,55,0}},
/* 48366 */ {(13<<2)|2,{67,55,0}},
/* 48367 */ {(13<<2)|2,{67,55,0}},
/* 48368 */ {(13<<2)|2,{67,56,0}},
/* 48369 */ {(13<<2)|2,{67,56,0}},
/* 48370 */ {(13<<2)|2,{67,56,0}},
/* 48371 */ {(13<<2)|2,{67,56,0}},
/* 48372 */ {(13<<2)|2,{67,56,0}},
/* 48373 */ {(13<<2)|2,{67,56,0}},
/* 48374 */ {(13<<2)|2,{67,56,0}},
/* 48375 */ {(13<<2)|2,{67,56,0}},
/* 48376 */ {(13<<2)|2,{67,57,0}},
/* 48377 */ {(13<<2)|2,{67,57,0}},
/* 48378 */ {(13<<2)|2,{67,57,0}},
/* 48379 */ {(13<<2)|2,{67,57,0}},
/* 48380 */ {(13<<2)|2,{67,57,0}},
/* 48381 */ {(13<<2)|2,{67,57,0}},
/* 48382 */ {(13<<2)|2,{67,57,0}},
/* 48383 */ {(13<<2)|2,{67,57,0}},
/* 48384 */ {(13<<2)|2,{67,61,0}},
/* 48385 */ {(13<<2)|2,{67,61,0}},
/* 48386 */ {(13<<2)|2,{67,61,0}},
/* 48387 */ {(13<<2)|2,{67,61,0}},
/* 48388 */ {(13<<2)|2,{67,61,0}},
/* 48389 */ {(13<<2)|2,{67,61,0}},
/* 48390 */ {(13<<2)|2,{67,61,0}},
/* 48391 */ {(13<<2)|2,{67,61,0}},
/* 48392 */ {(13<<2)|2,{67,65,0}},
/* 48393 */ {(13<<2)|2,{67,65,0}},
/* 48394 */ {(13<<2)|2,{67,65,0}},
/* 48395 */ {(13<<2)|2,{67,65,0}},
/* 48396 */ {(13<<2)|2,{67,65,0}},
/* 48397 */ {(13<<2)|2,{67,65,0}},
/* 48398 */ {(13<<2)|2,{67,65,0}},
/* 48399 */ {(13<<2)|2,{67,65,0}},
/* 48400 */ {(13<<2)|2,{67,95,0}},
/* 48401 */ {(13<<2)|2,{67,95,0}},
/* 48402 */ {(13<<2)|2,{67,95,0}},
/* 48403 */ {(13<<2)|2,{67,95,0}},
/* 48404 */ {(13<<2)|2,{67,95,0}},
/* 48405 */ {(13<<2)|2,{67,95,0}},
/* 48406 */ {(13<<2)|2,{67,95,0}},
/* 48407 */ {(13<<2)|2,{67,95,0}},
/* 48408 */ {(13<<2)|2,{67,98,0}},
/* 48409 */ {(13<<2)|2,{67,98,0}},
/* 48410 */ {(13<<2)|2,{67,98,0}},
/* 48411 */ {(13<<2)|2,{67,98,0}},
/* 48412 */ {(13<<2)|2,{67,98,0}},
/* 48413 */ {(13<<2)|2,{67,98,0}},
/* 48414 */ {(13<<2)|2,{67,98,0}},
/* 48415 */ {(13<<2)|2,{67,98,0}},
/* 48416 */ {(13<<2)|2,{67,100,0}},
/* 48417 */ {(13<<2)|2,{67,100,0}},
/* 48418 */ {(13<<2)|2,{67,100,0}},
/* 48419 */ {(13<<2)|2,{67,100,0}},
/* 48420 */ {(13<<2)|2,{67,100,0}},
/* 48421 */ {(13<<2)|2,{67,100,0}},
/* 48422 */ {(13<<2)|2,{67,100,0}},
/* 48423 */ {(13<<2)|2,{67,100,0}},
/* 48424 */ {(13<<2)|2,{67,102,0}},
/* 48425 */ {(13<<2)|2,{67,102,0}},
/* 48426 */ {(13<<2)|2,{67,102,0}},
/* 48427 */ {(13<<2)|2,{67,102,0}},
/* 48428 */ {(13<<2)|2,{67,102,0}},
/* 48429 */ {(13<<2)|2,{67,102,0}},
/* 48430 */ {(13<<2)|2,{67,102,0}},
/* 48431 */ {(13<<2)|2,{67,102,0}},
/* 48432 */ {(13<<2)|2,{67,103,0}},
/* 48433 */ {(13<<2)|2,{67,103,0}},
/* 48434 */ {(13<<2)|2,{67,103,0}},
/* 48435 */ {(13<<2)|2,{67,103,0}},
/* 48436 */ {(13<<2)|2,{67,103,0}},
/* 48437 */ {(13<<2)|2,{67,103,0}},
/* 48438 */ {(13<<2)|2,{67,103,0}},
/* 48439 */ {(13<<2)|2,{67,103,0}},
/* 48440 */ {(13<<2)|2,{67,104,0}},
/* 48441 */ {(13<<2)|2,{67,104,0}},
/* 48442 */ {(13<<2)|2,{67,104,0}},
/* 48443 */ {(13<<2)|2,{67,104,0}},
/* 48444 */ {(13<<2)|2,{67,104,0}},
/* 48445 */ {(13<<2)|2,{67,104,0}},
/* 48446 */ {(13<<2)|2,{67,104,0}},
/* 48447 */ {(13<<2)|2,{67,104,0}},
/* 48448 */ {(13<<2)|2,{67,108,0}},
/* 48449 */ {(13<<2)|2,{67,108,0}},
/* 48450 */ {(13<<2)|2,{67,108,0}},
/* 48451 */ {(13<<2)|2,{67,108,0}},
/* 48452 */ {(13<<2)|2,{67,108,0}},
/* 48453 */ {(13<<2)|2,{67,108,0}},
/* 48454 */ {(13<<2)|2,{67,108,0}},
/* 48455 */ {(13<<2)|2,{67,108,0}},
/* 48456 */ {(13<<2)|2,{67,109,0}},
/* 48457 */ {(13<<2)|2,{67,109,0}},
/* 48458 */ {(13<<2)|2,{67,109,0}},
/* 48459 */ {(13<<2)|2,{67,109,0}},
/* 48460 */ {(13<<2)|2,{67,109,0}},
/* 48461 */ {(13<<2)|2,{67,109,0}},
/* 48462 */ {(13<<2)|2,{67,109,0}},
/* 48463 */ {(13<<2)|2,{67,109,0}},
/* 48464 */ {(13<<2)|2,{67,110,0}},
/* 48465 */ {(13<<2)|2,{67,110,0}},
/* 48466 */ {(13<<2)|2,{67,110,0}},
/* 48467 */ {(13<<2)|2,{67,110,0}},
/* 48468 */ {(13<<2)|2,{67,110,0}},
/* 48469 */ {(13<<2)|2,{67,110,0}},
/* 48470 */ {(13<<2)|2,{67,110,0}},
/* 48471 */ {(13<<2)|2,{67,110,0}},
/* 48472 */ {(13<<2)|2,{67,112,0}},
/* 48473 */ {(13<<2)|2,{67,112,0}},
/* 48474 */ {(13<<2)|2,{67,112,0}},
/* 48475 */ {(13<<2)|2,{67,112,0}},
/* 48476 */ {(13<<2)|2,{67,112,0}},
/* 48477 */ {(13<<2)|2,{67,112,0}},
/* 48478 */ {(13<<2)|2,{67,112,0}},
/* 48479 */ {(13<<2)|2,{67,112,0}},
/* 48480 */ {(13<<2)|2,{67,114,0}},
/* 48481 */ {(13<<2)|2,{67,114,0}},
/* 48482 */ {(13<<2)|2,{67,114,0}},
/* 48483 */ {(13<<2)|2,{67,114,0}},
/* 48484 */ {(13<<2)|2,{67,114,0}},
/* 48485 */ {(13<<2)|2,{67,114,0}},
/* 48486 */ {(13<<2)|2,{67,114,0}},
/* 48487 */ {(13<<2)|2,{67,114,0}},
/* 48488 */ {(13<<2)|2,{67,117,0}},
/* 48489 */ {(13<<2)|2,{67,117,0}},
/* 48490 */ {(13<<2)|2,{67,117,0}},
/* 48491 */ {(13<<2)|2,{67,117,0}},
/* 48492 */ {(13<<2)|2,{67,117,0}},
/* 48493 */ {(13<<2)|2,{67,117,0}},
/* 48494 */ {(13<<2)|2,{67,117,0}},
/* 48495 */ {(13<<2)|2,{67,117,0}},
/* 48496 */ {(14<<2)|2,{67,58,0}},
/* 48497 */ {(14<<2)|2,{67,58,0}},
/* 48498 */ {(14<<2)|2,{67,58,0}},
/* 48499 */ {(14<<2)|2,{67,58,0}},
/* 48500 */ {(14<<2)|2,{67,66,0}},
/* 48501 */ {(14<<2)|2,{67,66,0}},
/* 48502 */ {(14<<2)|2,{67,66,0}},
/* 48503 */ {(14<<2)|2,{67,66,0}},
/* 48504 */ {(14<<2)|2,{67,67,0}},
/* 48505 */ {(14<<2)|2,{67,67,0}},
/* 48506 */ {(14<<2)|2,{67,67,0}},
/* 48507 */ {(14<<2)|2,{67,67,0}},
/* 48508 */ {(14<<2)|2,{67,68,0}},
/* 48509 */ {(14<<2)|2,{67,68,0}},
/* 48510 */ {(14<<2)|2,{67,68,0}},
/* 48511 */ {(14<<2)|2,{67,68,0}},
/* 48512 */ {(14<<2)|2,{67,69,0}},
/* 48513 */ {(14<<2)|2,{67,69,0}},
/* 48514 */ {(14<<2)|2,{67,69,0}},
/* 48515 */ {(14<<2)|2,{67,69,0}},
/* 48516 */ {(14<<2)|2,{67,70,0}},
/* 48517 */ {(14<<2)|2,{67,70,0}},
/* 48518 */ {(14<<2)|2,{67,70,0}},
/* 48519 */ {(14<<2)|2,{67,70,0}},
/* 48520 */ {(14<<2)|2,{67,71,0}},
/* 48521 */ {(14<<2)|2,{67,71,0}},
/* 48522 */ {(14<<2)|2,{67,71,0}},
/* 48523 */ {(14<<2)|2,{67,71,0}},
/* 48524 */ {(14<<2)|2,{67,72,0}},
/* 48525 */ {(14<<2)|2,{67,72,0}},
/* 48526 */ {(14<<2)|2,{67,72,0}},
/* 48527 */ {(14<<2)|2,{67,72,0}},
/* 48528 */ {(14<<2)|2,{67,73,0}},
/* 48529 */ {(14<<2)|2,{67,73,0}},
/* 48530 */ {(14<<2)|2,{67,73,0}},
/* 48531 */ {(14<<2)|2,{67,73,0}},
/* 48532 */ {(14<<2)|2,{67,74,0}},
/* 48533 */ {(14<<2)|2,{67,74,0}},
/* 48534 */ {(14<<2)|2,{67,74,0}},
/* 48535 */ {(14<<2)|2,{67,74,0}},
/* 48536 */ {(14<<2)|2,{67,75,0}},
/* 48537 */ {(14<<2)|2,{67,75,0}},
/* 48538 */ {(14<<2)|2,{67,75,0}},
/* 48539 */ {(14<<2)|2,{67,75,0}},
/* 48540 */ {(14<<2)|2,{67,76,0}},
/* 48541 */ {(14<<2)|2,{67,76,0}},
/* 48542 */ {(14<<2)|2,{67,76,0}},
/* 48543 */ {(14<<2)|2,{67,76,0}},
/* 48544 */ {(14<<2)|2,{67,77,0}},
/* 48545 */ {(14<<2)|2,{67,77,0}},
/* 48546 */ {(14<<2)|2,{67,77,0}},
/* 48547 */ {(14<<2)|2,{67,77,0}},
/* 48548 */ {(14<<2)|2,{67,78,0}},
/* 48549 */ {(14<<2)|2,{67,78,0}},
/* 48550 */ {(14<<2)|2,{67,78,0}},
/* 48551 */ {(14<<2)|2,{67,78,0}},
/* 48552 */ {(14<<2)|2,{67,79,0}},
/* 48553 */ {(14<<2)|2,{67,79,0}},
/* 48554 */ {(14<<2)|2,{67,79,0}},
/* 48555 */ {(14<<2)|2,{67,79,0}},
/* 48556 */ {(14<<2)|2,{67,80,0}},
/* 48557 */ {(14<<2)|2,{67,80,0}},
/* 48558 */ {(14<<2)|2,{67,80,0}},
/* 48559 */ {(14<<2)|2,{67,80,0}},
/* 48560 */ {(14<<2)|2,{67,81,0}},
/* 48561 */ {(14<<2)|2,{67,81,0}},
/* 48562 */ {(14<<2)|2,{67,81,0}},
/* 48563 */ {(14<<2)|2,{67,81,0}},
/* 48564 */ {(14<<2)|2,{67,82,0}},
/* 48565 */ {(14<<2)|2,{67,82,0}},
/* 48566 */ {(14<<2)|2,{67,82,0}},
/* 48567 */ {(14<<2)|2,{67,82,0}},
/* 48568 */ {(14<<2)|2,{67,83,0}},
/* 48569 */ {(14<<2)|2,{67,83,0}},
/* 48570 */ {(14<<2)|2,{67,83,0}},
/* 48571 */ {(14<<2)|2,{67,83,0}},
/* 48572 */ {(14<<2)|2,{67,84,0}},
/* 48573 */ {(14<<2)|2,{67,84,0}},
/* 48574 */ {(14<<2)|2,{67,84,0}},
/* 48575 */ {(14<<2)|2,{67,84,0}},
/* 48576 */ {(14<<2)|2,{67,85,0}},
/* 48577 */ {(14<<2)|2,{67,85,0}},
/* 48578 */ {(14<<2)|2,{67,85,0}},
/* 48579 */ {(14<<2)|2,{67,85,0}},
/* 48580 */ {(14<<2)|2,{67,86,0}},
/* 48581 */ {(14<<2)|2,{67,86,0}},
/* 48582 */ {(14<<2)|2,{67,86,0}},
/* 48583 */ {(14<<2)|2,{67,86,0}},
/* 48584 */ {(14<<2)|2,{67,87,0}},
/* 48585 */ {(14<<2)|2,{67,87,0}},
/* 48586 */ {(14<<2)|2,{67,87,0}},
/* 48587 */ {(14<<2)|2,{67,87,0}},
/* 48588 */ {(14<<2)|2,{67,89,0}},
/* 48589 */ {(14<<2)|2,{67,89,0}},
/* 48590 */ {(14<<2)|2,{67,89,0}},
/* 48591 */ {(14<<2)|2,{67,89,0}},
/* 48592 */ {(14<<2)|2,{67,106,0}},
/* 48593 */ {(14<<2)|2,{67,106,0}},
/* 48594 */ {(14<<2)|2,{67,106,0}},
/* 48595 */ {(14<<2)|2,{67,106,0}},
/* 48596 */ {(14<<2)|2,{67,107,0}},
/* 48597 */ {(14<<2)|2,{67,107,0}},
/* 48598 */ {(14<<2)|2,{67,107,0}},
/* 48599 */ {(14<<2)|2,{67,107,0}},
/* 48600 */ {(14<<2)|2,{67,113,0}},
/* 48601 */ {(14<<2)|2,{67,113,0}},
/* 48602 */ {(14<<2)|2,{67,113,0}},
/* 48603 */ {(14<<2)|2,{67,113,0}},
/* 48604 */ {(14<<2)|2,{67,118,0}},
/* 48605 */ {(14<<2)|2,{67,118,0}},
/* 48606 */ {(14<<2)|2,{67,118,0}},
/* 48607 */ {(14<<2)|2,{67,118,0}},
/* 48608 */ {(14<<2)|2,{67,119,0}},
/* 48609 */ {(14<<2)|2,{67,119,0}},
/* 48610 */ {(14<<2)|2,{67,119,0}},
/* 48611 */ {(14<<2)|2,{67,119,0}},
/* 48612 */ {(14<<2)|2,{67,120,0}},
/* 48613 */ {(14<<2)|2,{67,120,0}},
/* 48614 */ {(14<<2)|2,{67,120,0}},
/* 48615 */ {(14<<2)|2,{67,120,0}},
/* 48616 */ {(14<<2)|2,{67,121,0}},
/* 48617 */ {(14<<2)|2,{67,121,0}},
/* 48618 */ {(14<<2)|2,{67,121,0}},
/* 48619 */ {(14<<2)|2,{67,121,0}},
/* 48620 */ {(14<<2)|2,{67,122,0}},
/* 48621 */ {(14<<2)|2,{67,122,0}},
/* 48622 */ {(14<<2)|2,{67,122,0}},
/* 48623 */ {(14<<2)|2,{67,122,0}},
/* 48624 */ {(15<<2)|2,{67,38,0}},
/* 48625 */ {(15<<2)|2,{67,38,0}},
/* 48626 */ {(15<<2)|2,{67,42,0}},
/* 48627 */ {(15<<2)|2,{67,42,0}},
/* 48628 */ {(15<<2)|2,{67,44,0}},
/* 48629 */ {(15<<2)|2,{67,44,0}},
/* 48630 */ {(15<<2)|2,{67,59,0}},
/* 48631 */ {(15<<2)|2,{67,59,0}},
/* 48632 */ {(15<<2)|2,{67,88,0}},
/* 48633 */ {(15<<2)|2,{67,88,0}},
/* 48634 */ {(15<<2)|2,{67,90,0}},
/* 48635 */ {(15<<2)|2,{67,90,0}},
/* 48636 */ {(7<<2)|1,{67,0,0}},
/* 48637 */ {(7<<2)|1,{67,0,0}},
/* 48638 */ {(7<<2)|1,{67,0,0}},
/* 48639 */ {(7<<2)|1,{67,0,0}},
/* 48640 */ {(12<<2)|2,{68,48,0}},
/* 48641 */ {(12<<2)|2,{68,48,0}},
/* 48642 */ {(12<<2)|2,{68,48,0}},
/* 48643 */ {(12<<2)|2,{68,48,0}},
/* 48644 */ {(12<<2)|2,{68,48,0}},
/* 48645 */ {(12<<2)|2,{68,48,0}},
/* 48646 */ {(12<<2)|2,{68,48,0}},
/* 48647 */ {(12<<2)|2,{68,48,0}},
/* 48648 */ {(12<<2)|2,{68,48,0}},
/* 48649 */ {(12<<2)|2,{68,48,0}},
/* 48650 */ {(12<<2)|2,{68,48,0}},
/* 48651 */ {(12<<2)|2,{68,48,0}},
/* 48652 */ {(12<<2)|2,{68,48,0}},
/* 48653 */ {(12<<2)|2,{68,48,0}},
/* 48654 */ {(12<<2)|2,{68,48,0}},
/* 48655 */ {(12<<2)|2,{68,48,0}},
/* 48656 */ {(12<<2)|2,{68,49,0}},
/* 48657 */ {(12<<2)|2,{68,49,0}},
/* 48658 */ {(12<<2)|2,{68,49,0}},
/* 48659 */ {(12<<2)|2,{68,49,0}},
/* 48660 */ {(12<<2)|2,{68,49,0}},
/* 48661 */ {(12<<2)|2,{68,49,0}},
/* 48662 */ {(12<<2)|2,{68,49,0}},
/* 48663 */ {(12<<2)|2,{68,49,0}},
/* 48664 */ {(12<<2)|2,{68,49,0}},
/* 48665 */ {(12<<2)|2,{68,49,0}},
/* 48666 */ {(12<<2)|2,{68,49,0}},
/* 48667 */ {(12<<2)|2,{68,49,0}},
/* 48668 */ {(12<<2)|2,{68,49,0}},
/* 48669 */ {(12<<2)|2,{68,49,0}},
/* 48670 */ {(12<<2)|2,{68,49,0}},
/* 48671 */ {(12<<2)|2,{68,49,0}},
/* 48672 */ {(12<<2)|2,{68,50,0}},
/* 48673 */ {(12<<2)|2,{68,50,0}},
/* 48674 */ {(12<<2)|2,{68,50,0}},
/* 48675 */ {(12<<2)|2,{68,50,0}},
/* 48676 */ {(12<<2)|2,{68,50,0}},
/* 48677 */ {(12<<2)|2,{68,50,0}},
/* 48678 */ {(12<<2)|2,{68,50,0}},
/* 48679 */ {(12<<2)|2,{68,50,0}},
/* 48680 */ {(12<<2)|2,{68,50,0}},
/* 48681 */ {(12<<2)|2,{68,50,0}},
/* 48682 */ {(12<<2)|2,{68,50,0}},
/* 48683 */ {(12<<2)|2,{68,50,0}},
/* 48684 */ {(12<<2)|2,{68,50,0}},
/* 48685 */ {(12<<2)|2,{68,50,0}},
/* 48686 */ {(12<<2)|2,{68,50,0}},
/* 48687 */ {(12<<2)|2,{68,50,0}},
/* 48688 */ {(12<<2)|2,{68,97,0}},
/* 48689 */ {(12<<2)|2,{68,97,0}},
/* 48690 */ {(12<<2)|2,{68,97,0}},
/* 48691 */ {(12<<2)|2,{68,97,0}},
/* 48692 */ {(12<<2)|2,{68,97,0}},
/* 48693 */ {(12<<2)|2,{68,97,0}},
/* 48694 */ {(12<<2)|2,{68,97,0}},
/* 48695 */ {(12<<2)|2,{68,97,0}},
/* 48696 */ {(12<<2)|2,{68,97,0}},
/* 48697 */ {(12<<2)|2,{68,97,0}},
/* 48698 */ {(12<<2)|2,{68,97,0}},
/* 48699 */ {(12<<2)|2,{68,97,0}},
/* 48700 */ {(12<<2)|2,{68,97,0}},
/* 48701 */ {(12<<2)|2,{68,97,0}},
/* 48702 */ {(12<<2)|2,{68,97,0}},
/* 48703 */ {(12<<2)|2,{68,97,0}},
/* 48704 */ {(12<<2)|2,{68,99,0}},
/* 48705 */ {(12<<2)|2,{68,99,0}},
/* 48706 */ {(12<<2)|2,{68,99,0}},
/* 48707 */ {(12<<2)|2,{68,99,0}},
/* 48708 */ {(12<<2)|2,{68,99,0}},
/* 48709 */ {(12<<2)|2,{68,99,0}},
/* 48710 */ {(12<<2)|2,{68,99,0}},
/* 48711 */ {(12<<2)|2,{68,99,0}},
/* 48712 */ {(12<<2)|2,{68,99,0}},
/* 48713 */ {(12<<2)|2,{68,99,0}},
/* 48714 */ {(12<<2)|2,{68,99,0}},
/* 48715 */ {(12<<2)|2,{68,99,0}},
/* 48716 */ {(12<<2)|2,{68,99,0}},
/* 48717 */ {(12<<2)|2,{68,99,0}},
/* 48718 */ {(12<<2)|2,{68,99,0}},
/* 48719 */ {(12<<2)|2,{68,99,0}},
/* 48720 */ {(12<<2)|2,{68,101,0}},
/* 48721 */ {(12<<2)|2,{68,101,0}},
/* 48722 */ {(12<<2)|2,{68,101,0}},
/* 48723 */ {(12<<2)|2,{68,101,0}},
/* 48724 */ {(12<<2)|2,{68,101,0}},
/* 48725 */ {(12<<2)|2,{68,101,0}},
/* 48726 */ {(12<<2)|2,{68,101,0}},
/* 48727 */ {(12<<2)|2,{68,101,0}},
/* 48728 */ {(12<<2)|2,{68,101,0}},
/* 48729 */ {(12<<2)|2,{68,101,0}},
/* 48730 */ {(12<<2)|2,{68,101,0}},
/* 48731 */ {(12<<2)|2,{68,101,0}},
/* 48732 */ {(12<<2)|2,{68,101,0}},
/* 48733 */ {(12<<2)|2,{68,101,0}},
/* 48734 */ {(12<<2)|2,{68,101,0}},
/* 48735 */ {(12<<2)|2,{68,101,0}},
/* 48736 */ {(12<<2)|2,{68,105,0}},
/* 48737 */ {(12<<2)|2,{68,105,0}},
/* 48738 */ {(12<<2)|2,{68,105,0}},
/* 48739 */ {(12<<2)|2,{68,105,0}},
/* 48740 */ {(12<<2)|2,{68,105,0}},
/* 48741 */ {(12<<2)|2,{68,105,0}},
/* 48742 */ {(12<<2)|2,{68,105,0}},
/* 48743 */ {(12<<2)|2,{68,105,0}},
/* 48744 */ {(12<<2)|2,{68,105,0}},
/* 48745 */ {(12<<2)|2,{68,105,0}},
/* 48746 */ {(12<<2)|2,{68,105,0}},
/* 48747 */ {(12<<2)|2,{68,105,0}},
/* 48748 */ {(12<<2)|2,{68,105,0}},
/* 48749 */ {(12<<2)|2,{68,105,0}},
/* 48750 */ {(12<<2)|2,{68,105,0}},
/* 48751 */ {(12<<2)|2,{68,105,0}},
/* 48752 */ {(12<<2)|2,{68,111,0}},
/* 48753 */ {(12<<2)|2,{68,111,0}},
/* 48754 */ {(12<<2)|2,{68,111,0}},
/* 48755 */ {(12<<2)|2,{68,111,0}},
/* 48756 */ {(12<<2)|2,{68,111,0}},
/* 48757 */ {(12<<2)|2,{68,111,0}},
/* 48758 */ {(12<<2)|2,{68,111,0}},
/* 48759 */ {(12<<2)|2,{68,111,0}},
/* 48760 */ {(12<<2)|2,{68,111,0}},
/* 48761 */ {(12<<2)|2,{68,111,0}},
/* 48762 */ {(12<<2)|2,{68,111,0}},
/* 48763 */ {(12<<2)|2,{68,111,0}},
/* 48764 */ {(12<<2)|2,{68,111,0}},
/* 48765 */ {(12<<2)|2,{68,111,0}},
/* 48766 */ {(12<<2)|2,{68,111,0}},
/* 48767 */ {(12<<2)|2,{68,111,0}},
/* 48768 */ {(12<<2)|2,{68,115,0}},
/* 48769 */ {(12<<2)|2,{68,115,0}},
/* 48770 */ {(12<<2)|2,{68,115,0}},
/* 48771 */ {(12<<2)|2,{68,115,0}},
/* 48772 */ {(12<<2)|2,{68,115,0}},
/* 48773 */ {(12<<2)|2,{68,115,0}},
/* 48774 */ {(12<<2)|2,{68,115,0}},
/* 48775 */ {(12<<2)|2,{68,115,0}},
/* 48776 */ {(12<<2)|2,{68,115,0}},
/* 48777 */ {(12<<2)|2,{68,115,0}},
/* 48778 */ {(12<<2)|2,{68,115,0}},
/* 48779 */ {(12<<2)|2,{68,115,0}},
/* 48780 */ {(12<<2)|2,{68,115,0}},
/* 48781 */ {(12<<2)|2,{68,115,0}},
/* 48782 */ {(12<<2)|2,{68,115,0}},
/* 48783 */ {(12<<2)|2,{68,115,0}},
/* 48784 */ {(12<<2)|2,{68,116,0}},
/* 48785 */ {(12<<2)|2,{68,116,0}},
/* 48786 */ {(12<<2)|2,{68,116,0}},
/* 48787 */ {(12<<2)|2,{68,116,0}},
/* 48788 */ {(12<<2)|2,{68,116,0}},
/* 48789 */ {(12<<2)|2,{68,116,0}},
/* 48790 */ {(12<<2)|2,{68,116,0}},
/* 48791 */ {(12<<2)|2,{68,116,0}},
/* 48792 */ {(12<<2)|2,{68,116,0}},
/* 48793 */ {(12<<2)|2,{68,116,0}},
/* 48794 */ {(12<<2)|2,{68,116,0}},
/* 48795 */ {(12<<2)|2,{68,116,0}},
/* 48796 */ {(12<<2)|2,{68,116,0}},
/* 48797 */ {(12<<2)|2,{68,116,0}},
/* 48798 */ {(12<<2)|2,{68,116,0}},
/* 48799 */ {(12<<2)|2,{68,116,0}},
/* 48800 */ {(13<<2)|2,{68,32,0}},
/* 48801 */ {(13<<2)|2,{68,32,0}},
/* 48802 */ {(13<<2)|2,{68,32,0}},
/* 48803 */ {(13<<2)|2,{68,32,0}},
/* 48804 */ {(13<<2)|2,{68,32,0}},
/* 48805 */ {(13<<2)|2,{68,32,0}},
/* 48806 */ {(13<<2)|2,{68,32,0}},
/* 48807 */ {(13<<2)|2,{68,32,0}},
/* 48808 */ {(13<<2)|2,{68,37,0}},
/* 48809 */ {(13<<2)|2,{68,37,0}},
/* 48810 */ {(13<<2)|2,{68,37,0}},
/* 48811 */ {(13<<2)|2,{68,37,0}},
/* 48812 */ {(13<<2)|2,{68,37,0}},
/* 48813 */ {(13<<2)|2,{68,37,0}},
/* 48814 */ {(13<<2)|2,{68,37,0}},
/* 48815 */ {(13<<2)|2,{68,37,0}},
/* 48816 */ {(13<<2)|2,{68,45,0}},
/* 48817 */ {(13<<2)|2,{68,45,0}},
/* 48818 */ {(13<<2)|2,{68,45,0}},
/* 48819 */ {(13<<2)|2,{68,45,0}},
/* 48820 */ {(13<<2)|2,{68,45,0}},
/* 48821 */ {(13<<2)|2,{68,45,0}},
/* 48822 */ {(13<<2)|2,{68,45,0}},
/* 48823 */ {(13<<2)|2,{68,45,0}},
/* 48824 */ {(13<<2)|2,{68,46,0}},
/* 48825 */ {(13<<2)|2,{68,46,0}},
/* 48826 */ {(13<<2)|2,{68,46,0}},
/* 48827 */ {(13<<2)|2,{68,46,0}},
/* 48828 */ {(13<<2)|2,{68,46,0}},
/* 48829 */ {(13<<2)|2,{68,46,0}},
/* 48830 */ {(13<<2)|2,{68,46,0}},
/* 48831 */ {(13<<2)|2,{68,46,0}},
/* 48832 */ {(13<<2)|2,{68,47,0}},
/* 48833 */ {(13<<2)|2,{68,47,0}},
/* 48834 */ {(13<<2)|2,{68,47,0}},
/* 48835 */ {(13<<2)|2,{68,47,0}},
/* 48836 */ {(13<<2)|2,{68,47,0}},
/* 48837 */ {(13<<2)|2,{68,47,0}},
/* 48838 */ {(13<<2)|2,{68,47,0}},
/* 48839 */ {(13<<2)|2,{68,47,0}},
/* 48840 */ {(13<<2)|2,{68,51,0}},
/* 48841 */ {(13<<2)|2,{68,51,0}},
/* 48842 */ {(13<<2)|2,{68,51,0}},
/* 48843 */ {(13<<2)|2,{68,51,0}},
/* 48844 */ {(13<<2)|2,{68,51,0}},
/* 48845 */ {(13<<2)|2,{68,51,0}},
/* 48846 */ {(13<<2)|2,{68,51,0}},
/* 48847 */ {(13<<2)|2,{68,51,0}},
/* 48848 */ {(13<<2)|2,{68,52,0}},
/* 48849 */ {(13<<2)|2,{68,52,0}},
/* 48850 */ {(13<<2)|2,{68,52,0}},
/* 48851 */ {(13<<2)|2,{68,52,0}},
/* 48852 */ {(13<<2)|2,{68,52,0}},
/* 48853 */ {(13<<2)|2,{68,52,0}},
/* 48854 */ {(13<<2)|2,{68,52,0}},
/* 48855 */ {(13<<2)|2,{68,52,0}},
/* 48856 */ {(13<<2)|2,{68,53,0}},
/* 48857 */ {(13<<2)|2,{68,53,0}},
/* 48858 */ {(13<<2)|2,{68,53,0}},
/* 48859 */ {(13<<2)|2,{68,53,0}},
/* 48860 */ {(13<<2)|2,{68,53,0}},
/* 48861 */ {(13<<2)|2,{68,53,0}},
/* 48862 */ {(13<<2)|2,{68,53,0}},
/* 48863 */ {(13<<2)|2,{68,53,0}},
/* 48864 */ {(13<<2)|2,{68,54,0}},
/* 48865 */ {(13<<2)|2,{68,54,0}},
/* 48866 */ {(13<<2)|2,{68,54,0}},
/* 48867 */ {(13<<2)|2,{68,54,0}},
/* 48868 */ {(13<<2)|2,{68,54,0}},
/* 48869 */ {(13<<2)|2,{68,54,0}},
/* 48870 */ {(13<<2)|2,{68,54,0}},
/* 48871 */ {(13<<2)|2,{68,54,0}},
/* 48872 */ {(13<<2)|2,{68,55,0}},
/* 48873 */ {(13<<2)|2,{68,55,0}},
/* 48874 */ {(13<<2)|2,{68,55,0}},
/* 48875 */ {(13<<2)|2,{68,55,0}},
/* 48876 */ {(13<<2)|2,{68,55,0}},
/* 48877 */ {(13<<2)|2,{68,55,0}},
/* 48878 */ {(13<<2)|2,{68,55,0}},
/* 48879 */ {(13<<2)|2,{68,55,0}},
/* 48880 */ {(13<<2)|2,{68,56,0}},
/* 48881 */ {(13<<2)|2,{68,56,0}},
/* 48882 */ {(13<<2)|2,{68,56,0}},
/* 48883 */ {(13<<2)|2,{68,56,0}},
/* 48884 */ {(13<<2)|2,{68,56,0}},
/* 48885 */ {(13<<2)|2,{68,56,0}},
/* 48886 */ {(13<<2)|2,{68,56,0}},
/* 48887 */ {(13<<2)|2,{68,56,0}},
/* 48888 */ {(13<<2)|2,{68,57,0}},
/* 48889 */ {(13<<2)|2,{68,57,0}},
/* 48890 */ {(13<<2)|2,{68,57,0}},
/* 48891 */ {(13<<2)|2,{68,57,0}},
/* 48892 */ {(13<<2)|2,{68,57,0}},
/* 48893 */ {(13<<2)|2,{68,57,0}},
/* 48894 */ {(13<<2)|2,{68,57,0}},
/* 48895 */ {(13<<2)|2,{68,57,0}},
/* 48896 */ {(13<<2)|2,{68,61,0}},
/* 48897 */ {(13<<2)|2,{68,61,0}},
/* 48898 */ {(13<<2)|2,{68,61,0}},
/* 48899 */ {(13<<2)|2,{68,61,0}},
/* 48900 */ {(13<<2)|2,{68,61,0}},
/* 48901 */ {(13<<2)|2,{68,61,0}},
/* 48902 */ {(13<<2)|2,{68,61,0}},
/* 48903 */ {(13<<2)|2,{68,61,0}},
/* 48904 */ {(13<<2)|2,{68,65,0}},
/* 48905 */ {(13<<2)|2,{68,65,0}},
/* 48906 */ {(13<<2)|2,{68,65,0}},
/* 48907 */ {(13<<2)|2,{68,65,0}},
/* 48908 */ {(13<<2)|2,{68,65,0}},
/* 48909 */ {(13<<2)|2,{68,65,0}},
/* 48910 */ {(13<<2)|2,{68,65,0}},
/* 48911 */ {(13<<2)|2,{68,65,0}},
/* 48912 */ {(13<<2)|2,{68,95,0}},
/* 48913 */ {(13<<2)|2,{68,95,0}},
/* 48914 */ {(13<<2)|2,{68,95,0}},
/* 48915 */ {(13<<2)|2,{68,95,0}},
/* 48916 */ {(13<<2)|2,{68,95,0}},
/* 48917 */ {(13<<2)|2,{68,95,0}},
/* 48918 */ {(13<<2)|2,{68,95,0}},
/* 48919 */ {(13<<2)|2,{68,95,0}},
/* 48920 */ {(13<<2)|2,{68,98,0}},
/* 48921 */ {(13<<2)|2,{68,98,0}},
/* 48922 */ {(13<<2)|2,{68,98,0}},
/* 48923 */ {(13<<2)|2,{68,98,0}},
/* 48924 */ {(13<<2)|2,{68,98,0}},
/* 48925 */ {(13<<2)|2,{68,98,0}},
/* 48926 */ {(13<<2)|2,{68,98,0}},
/* 48927 */ {(13<<2)|2,{68,98,0}},
/* 48928 */ {(13<<2)|2,{68,100,0}},
/* 48929 */ {(13<<2)|2,{68,100,0}},
/* 48930 */ {(13<<2)|2,{68,100,0}},
/* 48931 */ {(13<<2)|2,{68,100,0}},
/* 48932 */ {(13<<2)|2,{68,100,0}},
/* 48933 */ {(13<<2)|2,{68,100,0}},
/* 48934 */ {(13<<2)|2,{68,100,0}},
/* 48935 */ {(13<<2)|2,{68,100,0}},
/* 48936 */ {(13<<2)|2,{68,102,0}},
/* 48937 */ {(13<<2)|2,{68,102,0}},
/* 48938 */ {(13<<2)|2,{68,102,0}},
/* 48939 */ {(13<<2)|2,{68,102,0}},
/* 48940 */ {(13<<2)|2,{68,102,0}},
/* 48941 */ {(13<<2)|2,{68,102,0}},
/* 48942 */ {(13<<2)|2,{68,102,0}},
/* 48943 */ {(13<<2)|2,{68,102,0}},
/* 48944 */ {(13<<2)|2,{68,103,0}},
/* 48945 */ {(13<<2)|2,{68,103,0}},
/* 48946 */ {(13<<2)|2,{68,103,0}},
/* 48947 */ {(13<<2)|2,{68,103,0}},
/* 48948 */ {(13<<2)|2,{68,103,0}},
/* 48949 */ {(13<<2)|2,{68,103,0}},
/* 48950 */ {(13<<2)|2,{68,103,0}},
/* 48951 */ {(13<<2)|2,{68,103,0}},
/* 48952 */ {(13<<2)|2,{68,104,0}},
/* 48953 */ {(13<<2)|2,{68,104,0}},
/* 48954 */ {(13<<2)|2,{68,104,0}},
/* 48955 */ {(13<<2)|2,{68,104,0}},
/* 48956 */ {(13<<2)|2,{68,104,0}},
/* 48957 */ {(13<<2)|2,{68,104,0}},
/* 48958 */ {(13<<2)|2,{68,104,0}},
/* 48959 */ {(13<<2)|2,{68,104,0}},
/* 48960 */ {(13<<2)|2,{68,108,0}},
/* 48961 */ {(13<<2)|2,{68,108,0}},
/* 48962 */ {(13<<2)|2,{68,108,0}},
/* 48963 */ {(13<<2)|2,{68,108,0}},
/* 48964 */ {(13<<2)|2,{68,108,0}},
/* 48965 */ {(13<<2)|2,{68,108,0}},
/* 48966 */ {(13<<2)|2,{68,108,0}},
/* 48967 */ {(13<<2)|2,{68,108,0}},
/* 48968 */ {(13<<2)|2,{68,109,0}},
/* 48969 */ {(13<<2)|2,{68,109,0}},
/* 48970 */ {(13<<2)|2,{68,109,0}},
/* 48971 */ {(13<<2)|2,{68,109,0}},
/* 48972 */ {(13<<2)|2,{68,109,0}},
/* 48973 */ {(13<<2)|2,{68,109,0}},
/* 48974 */ {(13<<2)|2,{68,109,0}},
/* 48975 */ {(13<<2)|2,{68,109,0}},
/* 48976 */ {(13<<2)|2,{68,110,0}},
/* 48977 */ {(13<<2)|2,{68,110,0}},
/* 48978 */ {(13<<2)|2,{68,110,0}},
/* 48979 */ {(13<<2)|2,{68,110,0}},
/* 48980 */ {(13<<2)|2,{68,110,0}},
/* 48981 */ {(13<<2)|2,{68,110,0}},
/* 48982 */ {(13<<2)|2,{68,110,0}},
/* 48983 */ {(13<<2)|2,{68,110,0}},
/* 48984 */ {(13<<2)|2,{68,112,0}},
/* 48985 */ {(13<<2)|2,{68,112,0}},
/* 48986 */ {(13<<2)|2,{68,112,0}},
/* 48987 */ {(13<<2)|2,{68,112,0}},
/* 48988 */ {(13<<2)|2,{68,112,0}},
/* 48989 */ {(13<<2)|2,{68,112,0}},
/* 48990 */ {(13<<2)|2,{68,112,0}},
/* 48991 */ {(13<<2)|2,{68,112,0}},
/* 48992 */ {(13<<2)|2,{68,114,0}},
/* 48993 */ {(13<<2)|2,{68,114,0}},
/* 48994 */ {(13<<2)|2,{68,114,0}},
/* 48995 */ {(13<<2)|2,{68,114,0}},
/* 48996 */ {(13<<2)|2,{68,114,0}},
/* 48997 */ {(13<<2)|2,{68,114,0}},
/* 48998 */ {(13<<2)|2,{68,114,0}},
/* 48999 */ {(13<<2)|2,{68,114,0}},
/* 49000 */ {(13<<2)|2,{68,117,0}},
/* 49001 */ {(13<<2)|2,{68,117,0}},
/* 49002 */ {(13<<2)|2,{68,117,0}},
/* 49003 */ {(13<<2)|2,{68,117,0}},
/* 49004 */ {(13<<2)|2,{68,117,0}},
/* 49005 */ {(13<<2)|2,{68,117,0}},
/* 49006 */ {(13<<2)|2,{68,117,0}},
/* 49007 */ {(13<<2)|2,{68,117,0}},
/* 49008 */ {(14<<2)|2,{68,58,0}},
/* 49009 */ {(14<<2)|2,{68,58,0}},
/* 49010 */ {(14<<2)|2,{68,58,0}},
/* 49011 */ {(14<<2)|2,{68,58,0}},
/* 49012 */ {(14<<2)|2,{68,66,0}},
/* 49013 */ {(14<<2)|2,{68,66,0}},
/* 49014 */ {(14<<2)|2,{68,66,0}},
/* 49015 */ {(14<<2)|2,{68,66,0}},
/* 49016 */ {(14<<2)|2,{68,67,0}},
/* 49017 */ {(14<<2)|2,{68,67,0}},
/* 49018 */ {(14<<2)|2,{68,67,0}},
/* 49019 */ {(14<<2)|2,{68,67,0}},
/* 49020 */ {(14<<2)|2,{68,68,0}},
/* 49021 */ {(14<<2)|2,{68,68,0}},
/* 49022 */ {(14<<2)|2,{68,68,0}},
/* 49023 */ {(14<<2)|2,{68,68,0}},
/* 49024 */ {(14<<2)|2,{68,69,0}},
/* 49025 */ {(14<<2)|2,{68,69,0}},
/* 49026 */ {(14<<2)|2,{68,69,0}},
/* 49027 */ {(14<<2)|2,{68,69,0}},
/* 49028 */ {(14<<2)|2,{68,70,0}},
/* 49029 */ {(14<<2)|2,{68,70,0}},
/* 49030 */ {(14<<2)|2,{68,70,0}},
/* 49031 */ {(14<<2)|2,{68,70,0}},
/* 49032 */ {(14<<2)|2,{68,71,0}},
/* 49033 */ {(14<<2)|2,{68,71,0}},
/* 49034 */ {(14<<2)|2,{68,71,0}},
/* 49035 */ {(14<<2)|2,{68,71,0}},
/* 49036 */ {(14<<2)|2,{68,72,0}},
/* 49037 */ {(14<<2)|2,{68,72,0}},
/* 49038 */ {(14<<2)|2,{68,72,0}},
/* 49039 */ {(14<<2)|2,{68,72,0}},
/* 49040 */ {(14<<2)|2,{68,73,0}},
/* 49041 */ {(14<<2)|2,{68,73,0}},
/* 49042 */ {(14<<2)|2,{68,73,0}},
/* 49043 */ {(14<<2)|2,{68,73,0}},
/* 49044 */ {(14<<2)|2,{68,74,0}},
/* 49045 */ {(14<<2)|2,{68,74,0}},
/* 49046 */ {(14<<2)|2,{68,74,0}},
/* 49047 */ {(14<<2)|2,{68,74,0}},
/* 49048 */ {(14<<2)|2,{68,75,0}},
/* 49049 */ {(14<<2)|2,{68,75,0}},
/* 49050 */ {(14<<2)|2,{68,75,0}},
/* 49051 */ {(14<<2)|2,{68,75,0}},
/* 49052 */ {(14<<2)|2,{68,76,0}},
/* 49053 */ {(14<<2)|2,{68,76,0}},
/* 49054 */ {(14<<2)|2,{68,76,0}},
/* 49055 */ {(14<<2)|2,{68,76,0}},
/* 49056 */ {(14<<2)|2,{68,77,0}},
/* 49057 */ {(14<<2)|2,{68,77,0}},
/* 49058 */ {(14<<2)|2,{68,77,0}},
/* 49059 */ {(14<<2)|2,{68,77,0}},
/* 49060 */ {(14<<2)|2,{68,78,0}},
/* 49061 */ {(14<<2)|2,{68,78,0}},
/* 49062 */ {(14<<2)|2,{68,78,0}},
/* 49063 */ {(14<<2)|2,{68,78,0}},
/* 49064 */ {(14<<2)|2,{68,79,0}},
/* 49065 */ {(14<<2)|2,{68,79,0}},
/* 49066 */ {(14<<2)|2,{68,79,0}},
/* 49067 */ {(14<<2)|2,{68,79,0}},
/* 49068 */ {(14<<2)|2,{68,80,0}},
/* 49069 */ {(14<<2)|2,{68,80,0}},
/* 49070 */ {(14<<2)|2,{68,80,0}},
/* 49071 */ {(14<<2)|2,{68,80,0}},
/* 49072 */ {(14<<2)|2,{68,81,0}},
/* 49073 */ {(14<<2)|2,{68,81,0}},
/* 49074 */ {(14<<2)|2,{68,81,0}},
/* 49075 */ {(14<<2)|2,{68,81,0}},
/* 49076 */ {(14<<2)|2,{68,82,0}},
/* 49077 */ {(14<<2)|2,{68,82,0}},
/* 49078 */ {(14<<2)|2,{68,82,0}},
/* 49079 */ {(14<<2)|2,{68,82,0}},
/* 49080 */ {(14<<2)|2,{68,83,0}},
/* 49081 */ {(14<<2)|2,{68,83,0}},
/* 49082 */ {(14<<2)|2,{68,83,0}},
/* 49083 */ {(14<<2)|2,{68,83,0}},
/* 49084 */ {(14<<2)|2,{68,84,0}},
/* 49085 */ {(14<<2)|2,{68,84,0}},
/* 49086 */ {(14<<2)|2,{68,84,0}},
/* 49087 */ {(14<<2)|2,{68,84,0}},
/* 49088 */ {(14<<2)|2,{68,85,0}},
/* 49089 */ {(14<<2)|2,{68,85,0}},
/* 49090 */ {(14<<2)|2,{68,85,0}},
/* 49091 */ {(14<<2)|2,{68,85,0}},
/* 49092 */ {(14<<2)|2,{68,86,0}},
/* 49093 */ {(14<<2)|2,{68,86,0}},
/* 49094 */ {(14<<2)|2,{68,86,0}},
/* 49095 */ {(14<<2)|2,{68,86,0}},
/* 49096 */ {(14<<2)|2,{68,87,0}},
/* 49097 */ {(14<<2)|2,{68,87,0}},
/* 49098 */ {(14<<2)|2,{68,87,0}},
/* 49099 */ {(14<<2)|2,{68,87,0}},
/* 49100 */ {(14<<2)|2,{68,89,0}},
/* 49101 */ {(14<<2)|2,{68,89,0}},
/* 49102 */ {(14<<2)|2,{68,89,0}},
/* 49103 */ {(14<<2)|2,{68,89,0}},
/* 49104 */ {(14<<2)|2,{68,106,0}},
/* 49105 */ {(14<<2)|2,{68,106,0}},
/* 49106 */ {(14<<2)|2,{68,106,0}},
/* 49107 */ {(14<<2)|2,{68,106,0}},
/* 49108 */ {(14<<2)|2,{68,107,0}},
/* 49109 */ {(14<<2)|2,{68,107,0}},
/* 49110 */ {(14<<2)|2,{68,107,0}},
/* 49111 */ {(14<<2)|2,{68,107,0}},
/* 49112 */ {(14<<2)|2,{68,113,0}},
/* 49113 */ {(14<<2)|2,{68,113,0}},
/* 49114 */ {(14<<2)|2,{68,113,0}},
/* 49115 */ {(14<<2)|2,{68,113,0}},
/* 49116 */ {(14<<2)|2,{68,118,0}},
/* 49117 */ {(14<<2)|2,{68,118,0}},
/* 49118 */ {(14<<2)|2,{68,118,0}},
/* 49119 */ {(14<<2)|2,{68,118,0}},
/* 49120 */ {(14<<2)|2,{68,119,0}},
/* 49121 */ {(14<<2)|2,{68,119,0}},
/* 49122 */ {(14<<2)|2,{68,119,0}},
/* 49123 */ {(14<<2)|2,{68,119,0}},
/* 49124 */ {(14<<2)|2,{68,120,0}},
/* 49125 */ {(14<<2)|2,{68,120,0}},
/* 49126 */ {(14<<2)|2,{68,120,0}},
/* 49127 */ {(14<<2)|2,{68,120,0}},
/* 49128 */ {(14<<2)|2,{68,121,0}},
/* 49129 */ {(14<<2)|2,{68,121,0}},
/* 49130 */ {(14<<2)|2,{68,121,0}},
/* 49131 */ {(14<<2)|2,{68,121,0}},
/* 49132 */ {(14<<2)|2,{68,122,0}},
/* 49133 */ {(14<<2)|2,{68,122,0}},
/* 49134 */ {(14<<2)|2,{68,122,0}},
/* 49135 */ {(14<<2)|2,{68,122,0}},
/* 49136 */ {(15<<2)|2,{68,38,0}},
/* 49137 */ {(15<<2)|2,{68,38,0}},
/* 49138 */ {(15<<2)|2,{68,42,0}},
/* 49139 */ {(15<<2)|2,{68,42,0}},
/* 49140 */ {(15<<2)|2,{68,44,0}},
/* 49141 */ {(15<<2)|2,{68,44,0}},
/* 49142 */ {(15<<2)|2,{68,59,0}},
/* 49143 */ {(15<<2)|2,{68,59,0}},
/* 49144 */ {(15<<2)|2,{68,88,0}},
/* 49145 */ {(15<<2)|2,{68,88,0}},
/* 49146 */ {(15<<2)|2,{68,90,0}},
/* 49147 */ {(15<<2)|2,{68,90,0}},
/* 49148 */ {(7<<2)|1,{68,0,0}},
/* 49149 */ {(7<<2)|1,{68,0,0}},
/* 49150 */ {(7<<2)|1,{68,0,0}},
/* 49151 */ {(7<<2)|1,{68,0,0}},
/* 49152 */ {(12<<2)|2,{69,48,0}},
/* 49153 */ {(12<<2)|2,{69,48,0}},
/* 49154 */ {(12<<2)|2,{69,48,0}},
/* 49155 */ {(12<<2)|2,{69,48,0}},
/* 49156 */ {(12<<2)|2,{69,48,0}},
/* 49157 */ {(12<<2)|2,{69,48,0}},
/* 49158 */ {(12<<2)|2,{69,48,0}},
/* 49159 */ {(12<<2)|2,{69,48,0}},
/* 49160 */ {(12<<2)|2,{69,48,0}},
/* 49161 */ {(12<<2)|2,{69,48,0}},
/* 49162 */ {(12<<2)|2,{69,48,0}},
/* 49163 */ {(12<<2)|2,{69,48,0}},
/* 49164 */ {(12<<2)|2,{69,48,0}},
/* 49165 */ {(12<<2)|2,{69,48,0}},
/* 49166 */ {(12<<2)|2,{69,48,0}},
/* 49167 */ {(12<<2)|2,{69,48,0}},
/* 49168 */ {(12<<2)|2,{69,49,0}},
/* 49169 */ {(12<<2)|2,{69,49,0}},
/* 49170 */ {(12<<2)|2,{69,49,0}},
/* 49171 */ {(12<<2)|2,{69,49,0}},
/* 49172 */ {(12<<2)|2,{69,49,0}},
/* 49173 */ {(12<<2)|2,{69,49,0}},
/* 49174 */ {(12<<2)|2,{69,49,0}},
/* 49175 */ {(12<<2)|2,{69,49,0}},
/* 49176 */ {(12<<2)|2,{69,49,0}},
/* 49177 */ {(12<<2)|2,{69,49,0}},
/* 49178 */ {(12<<2)|2,{69,49,0}},
/* 49179 */ {(12<<2)|2,{69,49,0}},
/* 49180 */ {(12<<2)|2,{69,49,0}},
/* 49181 */ {(12<<2)|2,{69,49,0}},
/* 49182 */ {(12<<2)|2,{69,49,0}},
/* 49183 */ {(12<<2)|2,{69,49,0}},
/* 49184 */ {(12<<2)|2,{69,50,0}},
/* 49185 */ {(12<<2)|2,{69,50,0}},
/* 49186 */ {(12<<2)|2,{69,50,0}},
/* 49187 */ {(12<<2)|2,{69,50,0}},
/* 49188 */ {(12<<2)|2,{69,50,0}},
/* 49189 */ {(12<<2)|2,{69,50,0}},
/* 49190 */ {(12<<2)|2,{69,50,0}},
/* 49191 */ {(12<<2)|2,{69,50,0}},
/* 49192 */ {(12<<2)|2,{69,50,0}},
/* 49193 */ {(12<<2)|2,{69,50,0}},
/* 49194 */ {(12<<2)|2,{69,50,0}},
/* 49195 */ {(12<<2)|2,{69,50,0}},
/* 49196 */ {(12<<2)|2,{69,50,0}},
/* 49197 */ {(12<<2)|2,{69,50,0}},
/* 49198 */ {(12<<2)|2,{69,50,0}},
/* 49199 */ {(12<<2)|2,{69,50,0}},
/* 49200 */ {(12<<2)|2,{69,97,0}},
/* 49201 */ {(12<<2)|2,{69,97,0}},
/* 49202 */ {(12<<2)|2,{69,97,0}},
/* 49203 */ {(12<<2)|2,{69,97,0}},
/* 49204 */ {(12<<2)|2,{69,97,0}},
/* 49205 */ {(12<<2)|2,{69,97,0}},
/* 49206 */ {(12<<2)|2,{69,97,0}},
/* 49207 */ {(12<<2)|2,{69,97,0}},
/* 49208 */ {(12<<2)|2,{69,97,0}},
/* 49209 */ {(12<<2)|2,{69,97,0}},
/* 49210 */ {(12<<2)|2,{69,97,0}},
/* 49211 */ {(12<<2)|2,{69,97,0}},
/* 49212 */ {(12<<2)|2,{69,97,0}},
/* 49213 */ {(12<<2)|2,{69,97,0}},
/* 49214 */ {(12<<2)|2,{69,97,0}},
/* 49215 */ {(12<<2)|2,{69,97,0}},
/* 49216 */ {(12<<2)|2,{69,99,0}},
/* 49217 */ {(12<<2)|2,{69,99,0}},
/* 49218 */ {(12<<2)|2,{69,99,0}},
/* 49219 */ {(12<<2)|2,{69,99,0}},
/* 49220 */ {(12<<2)|2,{69,99,0}},
/* 49221 */ {(12<<2)|2,{69,99,0}},
/* 49222 */ {(12<<2)|2,{69,99,0}},
/* 49223 */ {(12<<2)|2,{69,99,0}},
/* 49224 */ {(12<<2)|2,{69,99,0}},
/* 49225 */ {(12<<2)|2,{69,99,0}},
/* 49226 */ {(12<<2)|2,{69,99,0}},
/* 49227 */ {(12<<2)|2,{69,99,0}},
/* 49228 */ {(12<<2)|2,{69,99,0}},
/* 49229 */ {(12<<2)|2,{69,99,0}},
/* 49230 */ {(12<<2)|2,{69,99,0}},
/* 49231 */ {(12<<2)|2,{69,99,0}},
/* 49232 */ {(12<<2)|2,{69,101,0}},
/* 49233 */ {(12<<2)|2,{69,101,0}},
/* 49234 */ {(12<<2)|2,{69,101,0}},
/* 49235 */ {(12<<2)|2,{69,101,0}},
/* 49236 */ {(12<<2)|2,{69,101,0}},
/* 49237 */ {(12<<2)|2,{69,101,0}},
/* 49238 */ {(12<<2)|2,{69,101,0}},
/* 49239 */ {(12<<2)|2,{69,101,0}},
/* 49240 */ {(12<<2)|2,{69,101,0}},
/* 49241 */ {(12<<2)|2,{69,101,0}},
/* 49242 */ {(12<<2)|2,{69,101,0}},
/* 49243 */ {(12<<2)|2,{69,101,0}},
/* 49244 */ {(12<<2)|2,{69,101,0}},
/* 49245 */ {(12<<2)|2,{69,101,0}},
/* 49246 */ {(12<<2)|2,{69,101,0}},
/* 49247 */ {(12<<2)|2,{69,101,0}},
/* 49248 */ {(12<<2)|2,{69,105,0}},
/* 49249 */ {(12<<2)|2,{69,105,0}},
/* 49250 */ {(12<<2)|2,{69,105,0}},
/* 49251 */ {(12<<2)|2,{69,105,0}},
/* 49252 */ {(12<<2)|2,{69,105,0}},
/* 49253 */ {(12<<2)|2,{69,105,0}},
/* 49254 */ {(12<<2)|2,{69,105,0}},
/* 49255 */ {(12<<2)|2,{69,105,0}},
/* 49256 */ {(12<<2)|2,{69,105,0}},
/* 49257 */ {(12<<2)|2,{69,105,0}},
/* 49258 */ {(12<<2)|2,{69,105,0}},
/* 49259 */ {(12<<2)|2,{69,105,0}},
/* 49260 */ {(12<<2)|2,{69,105,0}},
/* 49261 */ {(12<<2)|2,{69,105,0}},
/* 49262 */ {(12<<2)|2,{69,105,0}},
/* 49263 */ {(12<<2)|2,{69,105,0}},
/* 49264 */ {(12<<2)|2,{69,111,0}},
/* 49265 */ {(12<<2)|2,{69,111,0}},
/* 49266 */ {(12<<2)|2,{69,111,0}},
/* 49267 */ {(12<<2)|2,{69,111,0}},
/* 49268 */ {(12<<2)|2,{69,111,0}},
/* 49269 */ {(12<<2)|2,{69,111,0}},
/* 49270 */ {(12<<2)|2,{69,111,0}},
/* 49271 */ {(12<<2)|2,{69,111,0}},
/* 49272 */ {(12<<2)|2,{69,111,0}},
/* 49273 */ {(12<<2)|2,{69,111,0}},
/* 49274 */ {(12<<2)|2,{69,111,0}},
/* 49275 */ {(12<<2)|2,{69,111,0}},
/* 49276 */ {(12<<2)|2,{69,111,0}},
/* 49277 */ {(12<<2)|2,{69,111,0}},
/* 49278 */ {(12<<2)|2,{69,111,0}},
/* 49279 */ {(12<<2)|2,{69,111,0}},
/* 49280 */ {(12<<2)|2,{69,115,0}},
/* 49281 */ {(12<<2)|2,{69,115,0}},
/* 49282 */ {(12<<2)|2,{69,115,0}},
/* 49283 */ {(12<<2)|2,{69,115,0}},
/* 49284 */ {(12<<2)|2,{69,115,0}},
/* 49285 */ {(12<<2)|2,{69,115,0}},
/* 49286 */ {(12<<2)|2,{69,115,0}},
/* 49287 */ {(12<<2)|2,{69,115,0}},
/* 49288 */ {(12<<2)|2,{69,115,0}},
/* 49289 */ {(12<<2)|2,{69,115,0}},
/* 49290 */ {(12<<2)|2,{69,115,0}},
/* 49291 */ {(12<<2)|2,{69,115,0}},
/* 49292 */ {(12<<2)|2,{69,115,0}},
/* 49293 */ {(12<<2)|2,{69,115,0}},
/* 49294 */ {(12<<2)|2,{69,115,0}},
/* 49295 */ {(12<<2)|2,{69,115,0}},
/* 49296 */ {(12<<2)|2,{69,116,0}},
/* 49297 */ {(12<<2)|2,{69,116,0}},
/* 49298 */ {(12<<2)|2,{69,116,0}},
/* 49299 */ {(12<<2)|2,{69,116,0}},
/* 49300 */ {(12<<2)|2,{69,116,0}},
/* 49301 */ {(12<<2)|2,{69,116,0}},
/* 49302 */ {(12<<2)|2,{69,116,0}},
/* 49303 */ {(12<<2)|2,{69,116,0}},
/* 49304 */ {(12<<2)|2,{69,116,0}},
/* 49305 */ {(12<<2)|2,{69,116,0}},
/* 49306 */ {(12<<2)|2,{69,116,0}},
/* 49307 */ {(12<<2)|2,{69,116,0}},
/* 49308 */ {(12<<2)|2,{69,116,0}},
/* 49309 */ {(12<<2)|2,{69,116,0}},
/* 49310 */ {(12<<2)|2,{69,116,0}},
/* 49311 */ {(12<<2)|2,{69,116,0}},
/* 49312 */ {(13<<2)|2,{69,32,0}},
/* 49313 */ {(13<<2)|2,{69,32,0}},
/* 49314 */ {(13<<2)|2,{69,32,0}},
/* 49315 */ {(13<<2)|2,{69,32,0}},
/* 49316 */ {(13<<2)|2,{69,32,0}},
/* 49317 */ {(13<<2)|2,{69,32,0}},
/* 49318 */ {(13<<2)|2,{69,32,0}},
/* 49319 */ {(13<<2)|2,{69,32,0}},
/* 49320 */ {(13<<2)|2,{69,37,0}},
/* 49321 */ {(13<<2)|2,{69,37,0}},
/* 49322 */ {(13<<2)|2,{69,37,0}},
/* 49323 */ {(13<<2)|2,{69,37,0}},
/* 49324 */ {(13<<2)|2,{69,37,0}},
/* 49325 */ {(13<<2)|2,{69,37,0}},
/* 49326 */ {(13<<2)|2,{69,37,0}},
/* 49327 */ {(13<<2)|2,{69,37,0}},
/* 49328 */ {(13<<2)|2,{69,45,0}},
/* 49329 */ {(13<<2)|2,{69,45,0}},
/* 49330 */ {(13<<2)|2,{69,45,0}},
/* 49331 */ {(13<<2)|2,{69,45,0}},
/* 49332 */ {(13<<2)|2,{69,45,0}},
/* 49333 */ {(13<<2)|2,{69,45,0}},
/* 49334 */ {(13<<2)|2,{69,45,0}},
/* 49335 */ {(13<<2)|2,{69,45,0}},
/* 49336 */ {(13<<2)|2,{69,46,0}},
/* 49337 */ {(13<<2)|2,{69,46,0}},
/* 49338 */ {(13<<2)|2,{69,46,0}},
/* 49339 */ {(13<<2)|2,{69,46,0}},
/* 49340 */ {(13<<2)|2,{69,46,0}},
/* 49341 */ {(13<<2)|2,{69,46,0}},
/* 49342 */ {(13<<2)|2,{69,46,0}},
/* 49343 */ {(13<<2)|2,{69,46,0}},
/* 49344 */ {(13<<2)|2,{69,47,0}},
/* 49345 */ {(13<<2)|2,{69,47,0}},
/* 49346 */ {(13<<2)|2,{69,47,0}},
/* 49347 */ {(13<<2)|2,{69,47,0}},
/* 49348 */ {(13<<2)|2,{69,47,0}},
/* 49349 */ {(13<<2)|2,{69,47,0}},
/* 49350 */ {(13<<2)|2,{69,47,0}},
/* 49351 */ {(13<<2)|2,{69,47,0}},
/* 49352 */ {(13<<2)|2,{69,51,0}},
/* 49353 */ {(13<<2)|2,{69,51,0}},
/* 49354 */ {(13<<2)|2,{69,51,0}},
/* 49355 */ {(13<<2)|2,{69,51,0}},
/* 49356 */ {(13<<2)|2,{69,51,0}},
/* 49357 */ {(13<<2)|2,{69,51,0}},
/* 49358 */ {(13<<2)|2,{69,51,0}},
/* 49359 */ {(13<<2)|2,{69,51,0}},
/* 49360 */ {(13<<2)|2,{69,52,0}},
/* 49361 */ {(13<<2)|2,{69,52,0}},
/* 49362 */ {(13<<2)|2,{69,52,0}},
/* 49363 */ {(13<<2)|2,{69,52,0}},
/* 49364 */ {(13<<2)|2,{69,52,0}},
/* 49365 */ {(13<<2)|2,{69,52,0}},
/* 49366 */ {(13<<2)|2,{69,52,0}},
/* 49367 */ {(13<<2)|2,{69,52,0}},
/* 49368 */ {(13<<2)|2,{69,53,0}},
/* 49369 */ {(13<<2)|2,{69,53,0}},
/* 49370 */ {(13<<2)|2,{69,53,0}},
/* 49371 */ {(13<<2)|2,{69,53,0}},
/* 49372 */ {(13<<2)|2,{69,53,0}},
/* 49373 */ {(13<<2)|2,{69,53,0}},
/* 49374 */ {(13<<2)|2,{69,53,0}},
/* 49375 */ {(13<<2)|2,{69,53,0}},
/* 49376 */ {(13<<2)|2,{69,54,0}},
/* 49377 */ {(13<<2)|2,{69,54,0}},
/* 49378 */ {(13<<2)|2,{69,54,0}},
/* 49379 */ {(13<<2)|2,{69,54,0}},
/* 49380 */ {(13<<2)|2,{69,54,0}},
/* 49381 */ {(13<<2)|2,{69,54,0}},
/* 49382 */ {(13<<2)|2,{69,54,0}},
/* 49383 */ {(13<<2)|2,{69,54,0}},
/* 49384 */ {(13<<2)|2,{69,55,0}},
/* 49385 */ {(13<<2)|2,{69,55,0}},
/* 49386 */ {(13<<2)|2,{69,55,0}},
/* 49387 */ {(13<<2)|2,{69,55,0}},
/* 49388 */ {(13<<2)|2,{69,55,0}},
/* 49389 */ {(13<<2)|2,{69,55,0}},
/* 49390 */ {(13<<2)|2,{69,55,0}},
/* 49391 */ {(13<<2)|2,{69,55,0}},
/* 49392 */ {(13<<2)|2,{69,56,0}},
/* 49393 */ {(13<<2)|2,{69,56,0}},
/* 49394 */ {(13<<2)|2,{69,56,0}},
/* 49395 */ {(13<<2)|2,{69,56,0}},
/* 49396 */ {(13<<2)|2,{69,56,0}},
/* 49397 */ {(13<<2)|2,{69,56,0}},
/* 49398 */ {(13<<2)|2,{69,56,0}},
/* 49399 */ {(13<<2)|2,{69,56,0}},
/* 49400 */ {(13<<2)|2,{69,57,0}},
/* 49401 */ {(13<<2)|2,{69,57,0}},
/* 49402 */ {(13<<2)|2,{69,57,0}},
/* 49403 */ {(13<<2)|2,{69,57,0}},
/* 49404 */ {(13<<2)|2,{69,57,0}},
/* 49405 */ {(13<<2)|2,{69,57,0}},
/* 49406 */ {(13<<2)|2,{69,57,0}},
/* 49407 */ {(13<<2)|2,{69,57,0}},
/* 49408 */ {(13<<2)|2,{69,61,0}},
/* 49409 */ {(13<<2)|2,{69,61,0}},
/* 49410 */ {(13<<2)|2,{69,61,0}},
/* 49411 */ {(13<<2)|2,{69,61,0}},
/* 49412 */ {(13<<2)|2,{69,61,0}},
/* 49413 */ {(13<<2)|2,{69,61,0}},
/* 49414 */ {(13<<2)|2,{69,61,0}},
/* 49415 */ {(13<<2)|2,{69,61,0}},
/* 49416 */ {(13<<2)|2,{69,65,0}},
/* 49417 */ {(13<<2)|2,{69,65,0}},
/* 49418 */ {(13<<2)|2,{69,65,0}},
/* 49419 */ {(13<<2)|2,{69,65,0}},
/* 49420 */ {(13<<2)|2,{69,65,0}},
/* 49421 */ {(13<<2)|2,{69,65,0}},
/* 49422 */ {(13<<2)|2,{69,65,0}},
/* 49423 */ {(13<<2)|2,{69,65,0}},
/* 49424 */ {(13<<2)|2,{69,95,0}},
/* 49425 */ {(13<<2)|2,{69,95,0}},
/* 49426 */ {(13<<2)|2,{69,95,0}},
/* 49427 */ {(13<<2)|2,{69,95,0}},
/* 49428 */ {(13<<2)|2,{69,95,0}},
/* 49429 */ {(13<<2)|2,{69,95,0}},
/* 49430 */ {(13<<2)|2,{69,95,0}},
/* 49431 */ {(13<<2)|2,{69,95,0}},
/* 49432 */ {(13<<2)|2,{69,98,0}},
/* 49433 */ {(13<<2)|2,{69,98,0}},
/* 49434 */ {(13<<2)|2,{69,98,0}},
/* 49435 */ {(13<<2)|2,{69,98,0}},
/* 49436 */ {(13<<2)|2,{69,98,0}},
/* 49437 */ {(13<<2)|2,{69,98,0}},
/* 49438 */ {(13<<2)|2,{69,98,0}},
/* 49439 */ {(13<<2)|2,{69,98,0}},
/* 49440 */ {(13<<2)|2,{69,100,0}},
/* 49441 */ {(13<<2)|2,{69,100,0}},
/* 49442 */ {(13<<2)|2,{69,100,0}},
/* 49443 */ {(13<<2)|2,{69,100,0}},
/* 49444 */ {(13<<2)|2,{69,100,0}},
/* 49445 */ {(13<<2)|2,{69,100,0}},
/* 49446 */ {(13<<2)|2,{69,100,0}},
/* 49447 */ {(13<<2)|2,{69,100,0}},
/* 49448 */ {(13<<2)|2,{69,102,0}},
/* 49449 */ {(13<<2)|2,{69,102,0}},
/* 49450 */ {(13<<2)|2,{69,102,0}},
/* 49451 */ {(13<<2)|2,{69,102,0}},
/* 49452 */ {(13<<2)|2,{69,102,0}},
/* 49453 */ {(13<<2)|2,{69,102,0}},
/* 49454 */ {(13<<2)|2,{69,102,0}},
/* 49455 */ {(13<<2)|2,{69,102,0}},
/* 49456 */ {(13<<2)|2,{69,103,0}},
/* 49457 */ {(13<<2)|2,{69,103,0}},
/* 49458 */ {(13<<2)|2,{69,103,0}},
/* 49459 */ {(13<<2)|2,{69,103,0}},
/* 49460 */ {(13<<2)|2,{69,103,0}},
/* 49461 */ {(13<<2)|2,{69,103,0}},
/* 49462 */ {(13<<2)|2,{69,103,0}},
/* 49463 */ {(13<<2)|2,{69,103,0}},
/* 49464 */ {(13<<2)|2,{69,104,0}},
/* 49465 */ {(13<<2)|2,{69,104,0}},
/* 49466 */ {(13<<2)|2,{69,104,0}},
/* 49467 */ {(13<<2)|2,{69,104,0}},
/* 49468 */ {(13<<2)|2,{69,104,0}},
/* 49469 */ {(13<<2)|2,{69,104,0}},
/* 49470 */ {(13<<2)|2,{69,104,0}},
/* 49471 */ {(13<<2)|2,{69,104,0}},
/* 49472 */ {(13<<2)|2,{69,108,0}},
/* 49473 */ {(13<<2)|2,{69,108,0}},
/* 49474 */ {(13<<2)|2,{69,108,0}},
/* 49475 */ {(13<<2)|2,{69,108,0}},
/* 49476 */ {(13<<2)|2,{69,108,0}},
/* 49477 */ {(13<<2)|2,{69,108,0}},
/* 49478 */ {(13<<2)|2,{69,108,0}},
/* 49479 */ {(13<<2)|2,{69,108,0}},
/* 49480 */ {(13<<2)|2,{69,109,0}},
/* 49481 */ {(13<<2)|2,{69,109,0}},
/* 49482 */ {(13<<2)|2,{69,109,0}},
/* 49483 */ {(13<<2)|2,{69,109,0}},
/* 49484 */ {(13<<2)|2,{69,109,0}},
/* 49485 */ {(13<<2)|2,{69,109,0}},
/* 49486 */ {(13<<2)|2,{69,109,0}},
/* 49487 */ {(13<<2)|2,{69,109,0}},
/* 49488 */ {(13<<2)|2,{69,110,0}},
/* 49489 */ {(13<<2)|2,{69,110,0}},
/* 49490 */ {(13<<2)|2,{69,110,0}},
/* 49491 */ {(13<<2)|2,{69,110,0}},
/* 49492 */ {(13<<2)|2,{69,110,0}},
/* 49493 */ {(13<<2)|2,{69,110,0}},
/* 49494 */ {(13<<2)|2,{69,110,0}},
/* 49495 */ {(13<<2)|2,{69,110,0}},
/* 49496 */ {(13<<2)|2,{69,112,0}},
/* 49497 */ {(13<<2)|2,{69,112,0}},
/* 49498 */ {(13<<2)|2,{69,112,0}},
/* 49499 */ {(13<<2)|2,{69,112,0}},
/* 49500 */ {(13<<2)|2,{69,112,0}},
/* 49501 */ {(13<<2)|2,{69,112,0}},
/* 49502 */ {(13<<2)|2,{69,112,0}},
/* 49503 */ {(13<<2)|2,{69,112,0}},
/* 49504 */ {(13<<2)|2,{69,114,0}},
/* 49505 */ {(13<<2)|2,{69,114,0}},
/* 49506 */ {(13<<2)|2,{69,114,0}},
/* 49507 */ {(13<<2)|2,{69,114,0}},
/* 49508 */ {(13<<2)|2,{69,114,0}},
/* 49509 */ {(13<<2)|2,{69,114,0}},
/* 49510 */ {(13<<2)|2,{69,114,0}},
/* 49511 */ {(13<<2)|2,{69,114,0}},
/* 49512 */ {(13<<2)|2,{69,117,0}},
/* 49513 */ {(13<<2)|2,{69,117,0}},
/* 49514 */ {(13<<2)|2,{69,117,0}},
/* 49515 */ {(13<<2)|2,{69,117,0}},
/* 49516 */ {(13<<2)|2,{69,117,0}},
/* 49517 */ {(13<<2)|2,{69,117,0}},
/* 49518 */ {(13<<2)|2,{69,117,0}},
/* 49519 */ {(13<<2)|2,{69,117,0}},
/* 49520 */ {(14<<2)|2,{69,58,0}},
/* 49521 */ {(14<<2)|2,{69,58,0}},
/* 49522 */ {(14<<2)|2,{69,58,0}},
/* 49523 */ {(14<<2)|2,{69,58,0}},
/* 49524 */ {(14<<2)|2,{69,66,0}},
/* 49525 */ {(14<<2)|2,{69,66,0}},
/* 49526 */ {(14<<2)|2,{69,66,0}},
/* 49527 */ {(14<<2)|2,{69,66,0}},
/* 49528 */ {(14<<2)|2,{69,67,0}},
/* 49529 */ {(14<<2)|2,{69,67,0}},
/* 49530 */ {(14<<2)|2,{69,67,0}},
/* 49531 */ {(14<<2)|2,{69,67,0}},
/* 49532 */ {(14<<2)|2,{69,68,0}},
/* 49533 */ {(14<<2)|2,{69,68,0}},
/* 49534 */ {(14<<2)|2,{69,68,0}},
/* 49535 */ {(14<<2)|2,{69,68,0}},
/* 49536 */ {(14<<2)|2,{69,69,0}},
/* 49537 */ {(14<<2)|2,{69,69,0}},
/* 49538 */ {(14<<2)|2,{69,69,0}},
/* 49539 */ {(14<<2)|2,{69,69,0}},
/* 49540 */ {(14<<2)|2,{69,70,0}},
/* 49541 */ {(14<<2)|2,{69,70,0}},
/* 49542 */ {(14<<2)|2,{69,70,0}},
/* 49543 */ {(14<<2)|2,{69,70,0}},
/* 49544 */ {(14<<2)|2,{69,71,0}},
/* 49545 */ {(14<<2)|2,{69,71,0}},
/* 49546 */ {(14<<2)|2,{69,71,0}},
/* 49547 */ {(14<<2)|2,{69,71,0}},
/* 49548 */ {(14<<2)|2,{69,72,0}},
/* 49549 */ {(14<<2)|2,{69,72,0}},
/* 49550 */ {(14<<2)|2,{69,72,0}},
/* 49551 */ {(14<<2)|2,{69,72,0}},
/* 49552 */ {(14<<2)|2,{69,73,0}},
/* 49553 */ {(14<<2)|2,{69,73,0}},
/* 49554 */ {(14<<2)|2,{69,73,0}},
/* 49555 */ {(14<<2)|2,{69,73,0}},
/* 49556 */ {(14<<2)|2,{69,74,0}},
/* 49557 */ {(14<<2)|2,{69,74,0}},
/* 49558 */ {(14<<2)|2,{69,74,0}},
/* 49559 */ {(14<<2)|2,{69,74,0}},
/* 49560 */ {(14<<2)|2,{69,75,0}},
/* 49561 */ {(14<<2)|2,{69,75,0}},
/* 49562 */ {(14<<2)|2,{69,75,0}},
/* 49563 */ {(14<<2)|2,{69,75,0}},
/* 49564 */ {(14<<2)|2,{69,76,0}},
/* 49565 */ {(14<<2)|2,{69,76,0}},
/* 49566 */ {(14<<2)|2,{69,76,0}},
/* 49567 */ {(14<<2)|2,{69,76,0}},
/* 49568 */ {(14<<2)|2,{69,77,0}},
/* 49569 */ {(14<<2)|2,{69,77,0}},
/* 49570 */ {(14<<2)|2,{69,77,0}},
/* 49571 */ {(14<<2)|2,{69,77,0}},
/* 49572 */ {(14<<2)|2,{69,78,0}},
/* 49573 */ {(14<<2)|2,{69,78,0}},
/* 49574 */ {(14<<2)|2,{69,78,0}},
/* 49575 */ {(14<<2)|2,{69,78,0}},
/* 49576 */ {(14<<2)|2,{69,79,0}},
/* 49577 */ {(14<<2)|2,{69,79,0}},
/* 49578 */ {(14<<2)|2,{69,79,0}},
/* 49579 */ {(14<<2)|2,{69,79,0}},
/* 49580 */ {(14<<2)|2,{69,80,0}},
/* 49581 */ {(14<<2)|2,{69,80,0}},
/* 49582 */ {(14<<2)|2,{69,80,0}},
/* 49583 */ {(14<<2)|2,{69,80,0}},
/* 49584 */ {(14<<2)|2,{69,81,0}},
/* 49585 */ {(14<<2)|2,{69,81,0}},
/* 49586 */ {(14<<2)|2,{69,81,0}},
/* 49587 */ {(14<<2)|2,{69,81,0}},
/* 49588 */ {(14<<2)|2,{69,82,0}},
/* 49589 */ {(14<<2)|2,{69,82,0}},
/* 49590 */ {(14<<2)|2,{69,82,0}},
/* 49591 */ {(14<<2)|2,{69,82,0}},
/* 49592 */ {(14<<2)|2,{69,83,0}},
/* 49593 */ {(14<<2)|2,{69,83,0}},
/* 49594 */ {(14<<2)|2,{69,83,0}},
/* 49595 */ {(14<<2)|2,{69,83,0}},
/* 49596 */ {(14<<2)|2,{69,84,0}},
/* 49597 */ {(14<<2)|2,{69,84,0}},
/* 49598 */ {(14<<2)|2,{69,84,0}},
/* 49599 */ {(14<<2)|2,{69,84,0}},
/* 49600 */ {(14<<2)|2,{69,85,0}},
/* 49601 */ {(14<<2)|2,{69,85,0}},
/* 49602 */ {(14<<2)|2,{69,85,0}},
/* 49603 */ {(14<<2)|2,{69,85,0}},
/* 49604 */ {(14<<2)|2,{69,86,0}},
/* 49605 */ {(14<<2)|2,{69,86,0}},
/* 49606 */ {(14<<2)|2,{69,86,0}},
/* 49607 */ {(14<<2)|2,{69,86,0}},
/* 49608 */ {(14<<2)|2,{69,87,0}},
/* 49609 */ {(14<<2)|2,{69,87,0}},
/* 49610 */ {(14<<2)|2,{69,87,0}},
/* 49611 */ {(14<<2)|2,{69,87,0}},
/* 49612 */ {(14<<2)|2,{69,89,0}},
/* 49613 */ {(14<<2)|2,{69,89,0}},
/* 49614 */ {(14<<2)|2,{69,89,0}},
/* 49615 */ {(14<<2)|2,{69,89,0}},
/* 49616 */ {(14<<2)|2,{69,106,0}},
/* 49617 */ {(14<<2)|2,{69,106,0}},
/* 49618 */ {(14<<2)|2,{69,106,0}},
/* 49619 */ {(14<<2)|2,{69,106,0}},
/* 49620 */ {(14<<2)|2,{69,107,0}},
/* 49621 */ {(14<<2)|2,{69,107,0}},
/* 49622 */ {(14<<2)|2,{69,107,0}},
/* 49623 */ {(14<<2)|2,{69,107,0}},
/* 49624 */ {(14<<2)|2,{69,113,0}},
/* 49625 */ {(14<<2)|2,{69,113,0}},
/* 49626 */ {(14<<2)|2,{69,113,0}},
/* 49627 */ {(14<<2)|2,{69,113,0}},
/* 49628 */ {(14<<2)|2,{69,118,0}},
/* 49629 */ {(14<<2)|2,{69,118,0}},
/* 49630 */ {(14<<2)|2,{69,118,0}},
/* 49631 */ {(14<<2)|2,{69,118,0}},
/* 49632 */ {(14<<2)|2,{69,119,0}},
/* 49633 */ {(14<<2)|2,{69,119,0}},
/* 49634 */ {(14<<2)|2,{69,119,0}},
/* 49635 */ {(14<<2)|2,{69,119,0}},
/* 49636 */ {(14<<2)|2,{69,120,0}},
/* 49637 */ {(14<<2)|2,{69,120,0}},
/* 49638 */ {(14<<2)|2,{69,120,0}},
/* 49639 */ {(14<<2)|2,{69,120,0}},
/* 49640 */ {(14<<2)|2,{69,121,0}},
/* 49641 */ {(14<<2)|2,{69,121,0}},
/* 49642 */ {(14<<2)|2,{69,121,0}},
/* 49643 */ {(14<<2)|2,{69,121,0}},
/* 49644 */ {(14<<2)|2,{69,122,0}},
/* 49645 */ {(14<<2)|2,{69,122,0}},
/* 49646 */ {(14<<2)|2,{69,122,0}},
/* 49647 */ {(14<<2)|2,{69,122,0}},
/* 49648 */ {(15<<2)|2,{69,38,0}},
/* 49649 */ {(15<<2)|2,{69,38,0}},
/* 49650 */ {(15<<2)|2,{69,42,0}},
/* 49651 */ {(15<<2)|2,{69,42,0}},
/* 49652 */ {(15<<2)|2,{69,44,0}},
/* 49653 */ {(15<<2)|2,{69,44,0}},
/* 49654 */ {(15<<2)|2,{69,59,0}},
/* 49655 */ {(15<<2)|2,{69,59,0}},
/* 49656 */ {(15<<2)|2,{69,88,0}},
/* 49657 */ {(15<<2)|2,{69,88,0}},
/* 49658 */ {(15<<2)|2,{69,90,0}},
/* 49659 */ {(15<<2)|2,{69,90,0}},
/* 49660 */ {(7<<2)|1,{69,0,0}},
/* 49661 */ {(7<<2)|1,{69,0,0}},
/* 49662 */ {(7<<2)|1,{69,0,0}},
/* 49663 */ {(7<<2)|1,{69,0,0}},
/* 49664 */ {(12<<2)|2,{70,48,0}},
/* 49665 */ {(12<<2)|2,{70,48,0}},
/* 49666 */ {(12<<2)|2,{70,48,0}},
/* 49667 */ {(12<<2)|2,{70,48,0}},
/* 49668 */ {(12<<2)|2,{70,48,0}},
/* 49669 */ {(12<<2)|2,{70,48,0}},
/* 49670 */ {(12<<2)|2,{70,48,0}},
/* 49671 */ {(12<<2)|2,{70,48,0}},
/* 49672 */ {(12<<2)|2,{70,48,0}},
/* 49673 */ {(12<<2)|2,{70,48,0}},
/* 49674 */ {(12<<2)|2,{70,48,0}},
/* 49675 */ {(12<<2)|2,{70,48,0}},
/* 49676 */ {(12<<2)|2,{70,48,0}},
/* 49677 */ {(12<<2)|2,{70,48,0}},
/* 49678 */ {(12<<2)|2,{70,48,0}},
/* 49679 */ {(12<<2)|2,{70,48,0}},
/* 49680 */ {(12<<2)|2,{70,49,0}},
/* 49681 */ {(12<<2)|2,{70,49,0}},
/* 49682 */ {(12<<2)|2,{70,49,0}},
/* 49683 */ {(12<<2)|2,{70,49,0}},
/* 49684 */ {(12<<2)|2,{70,49,0}},
/* 49685 */ {(12<<2)|2,{70,49,0}},
/* 49686 */ {(12<<2)|2,{70,49,0}},
/* 49687 */ {(12<<2)|2,{70,49,0}},
/* 49688 */ {(12<<2)|2,{70,49,0}},
/* 49689 */ {(12<<2)|2,{70,49,0}},
/* 49690 */ {(12<<2)|2,{70,49,0}},
/* 49691 */ {(12<<2)|2,{70,49,0}},
/* 49692 */ {(12<<2)|2,{70,49,0}},
/* 49693 */ {(12<<2)|2,{70,49,0}},
/* 49694 */ {(12<<2)|2,{70,49,0}},
/* 49695 */ {(12<<2)|2,{70,49,0}},
/* 49696 */ {(12<<2)|2,{70,50,0}},
/* 49697 */ {(12<<2)|2,{70,50,0}},
/* 49698 */ {(12<<2)|2,{70,50,0}},
/* 49699 */ {(12<<2)|2,{70,50,0}},
/* 49700 */ {(12<<2)|2,{70,50,0}},
/* 49701 */ {(12<<2)|2,{70,50,0}},
/* 49702 */ {(12<<2)|2,{70,50,0}},
/* 49703 */ {(12<<2)|2,{70,50,0}},
/* 49704 */ {(12<<2)|2,{70,50,0}},
/* 49705 */ {(12<<2)|2,{70,50,0}},
/* 49706 */ {(12<<2)|2,{70,50,0}},
/* 49707 */ {(12<<2)|2,{70,50,0}},
/* 49708 */ {(12<<2)|2,{70,50,0}},
/* 49709 */ {(12<<2)|2,{70,50,0}},
/* 49710 */ {(12<<2)|2,{70,50,0}},
/* 49711 */ {(12<<2)|2,{70,50,0}},
/* 49712 */ {(12<<2)|2,{70,97,0}},
/* 49713 */ {(12<<2)|2,{70,97,0}},
/* 49714 */ {(12<<2)|2,{70,97,0}},
/* 49715 */ {(12<<2)|2,{70,97,0}},
/* 49716 */ {(12<<2)|2,{70,97,0}},
/* 49717 */ {(12<<2)|2,{70,97,0}},
/* 49718 */ {(12<<2)|2,{70,97,0}},
/* 49719 */ {(12<<2)|2,{70,97,0}},
/* 49720 */ {(12<<2)|2,{70,97,0}},
/* 49721 */ {(12<<2)|2,{70,97,0}},
/* 49722 */ {(12<<2)|2,{70,97,0}},
/* 49723 */ {(12<<2)|2,{70,97,0}},
/* 49724 */ {(12<<2)|2,{70,97,0}},
/* 49725 */ {(12<<2)|2,{70,97,0}},
/* 49726 */ {(12<<2)|2,{70,97,0}},
/* 49727 */ {(12<<2)|2,{70,97,0}},
/* 49728 */ {(12<<2)|2,{70,99,0}},
/* 49729 */ {(12<<2)|2,{70,99,0}},
/* 49730 */ {(12<<2)|2,{70,99,0}},
/* 49731 */ {(12<<2)|2,{70,99,0}},
/* 49732 */ {(12<<2)|2,{70,99,0}},
/* 49733 */ {(12<<2)|2,{70,99,0}},
/* 49734 */ {(12<<2)|2,{70,99,0}},
/* 49735 */ {(12<<2)|2,{70,99,0}},
/* 49736 */ {(12<<2)|2,{70,99,0}},
/* 49737 */ {(12<<2)|2,{70,99,0}},
/* 49738 */ {(12<<2)|2,{70,99,0}},
/* 49739 */ {(12<<2)|2,{70,99,0}},
/* 49740 */ {(12<<2)|2,{70,99,0}},
/* 49741 */ {(12<<2)|2,{70,99,0}},
/* 49742 */ {(12<<2)|2,{70,99,0}},
/* 49743 */ {(12<<2)|2,{70,99,0}},
/* 49744 */ {(12<<2)|2,{70,101,0}},
/* 49745 */ {(12<<2)|2,{70,101,0}},
/* 49746 */ {(12<<2)|2,{70,101,0}},
/* 49747 */ {(12<<2)|2,{70,101,0}},
/* 49748 */ {(12<<2)|2,{70,101,0}},
/* 49749 */ {(12<<2)|2,{70,101,0}},
/* 49750 */ {(12<<2)|2,{70,101,0}},
/* 49751 */ {(12<<2)|2,{70,101,0}},
/* 49752 */ {(12<<2)|2,{70,101,0}},
/* 49753 */ {(12<<2)|2,{70,101,0}},
/* 49754 */ {(12<<2)|2,{70,101,0}},
/* 49755 */ {(12<<2)|2,{70,101,0}},
/* 49756 */ {(12<<2)|2,{70,101,0}},
/* 49757 */ {(12<<2)|2,{70,101,0}},
/* 49758 */ {(12<<2)|2,{70,101,0}},
/* 49759 */ {(12<<2)|2,{70,101,0}},
/* 49760 */ {(12<<2)|2,{70,105,0}},
/* 49761 */ {(12<<2)|2,{70,105,0}},
/* 49762 */ {(12<<2)|2,{70,105,0}},
/* 49763 */ {(12<<2)|2,{70,105,0}},
/* 49764 */ {(12<<2)|2,{70,105,0}},
/* 49765 */ {(12<<2)|2,{70,105,0}},
/* 49766 */ {(12<<2)|2,{70,105,0}},
/* 49767 */ {(12<<2)|2,{70,105,0}},
/* 49768 */ {(12<<2)|2,{70,105,0}},
/* 49769 */ {(12<<2)|2,{70,105,0}},
/* 49770 */ {(12<<2)|2,{70,105,0}},
/* 49771 */ {(12<<2)|2,{70,105,0}},
/* 49772 */ {(12<<2)|2,{70,105,0}},
/* 49773 */ {(12<<2)|2,{70,105,0}},
/* 49774 */ {(12<<2)|2,{70,105,0}},
/* 49775 */ {(12<<2)|2,{70,105,0}},
/* 49776 */ {(12<<2)|2,{70,111,0}},
/* 49777 */ {(12<<2)|2,{70,111,0}},
/* 49778 */ {(12<<2)|2,{70,111,0}},
/* 49779 */ {(12<<2)|2,{70,111,0}},
/* 49780 */ {(12<<2)|2,{70,111,0}},
/* 49781 */ {(12<<2)|2,{70,111,0}},
/* 49782 */ {(12<<2)|2,{70,111,0}},
/* 49783 */ {(12<<2)|2,{70,111,0}},
/* 49784 */ {(12<<2)|2,{70,111,0}},
/* 49785 */ {(12<<2)|2,{70,111,0}},
/* 49786 */ {(12<<2)|2,{70,111,0}},
/* 49787 */ {(12<<2)|2,{70,111,0}},
/* 49788 */ {(12<<2)|2,{70,111,0}},
/* 49789 */ {(12<<2)|2,{70,111,0}},
/* 49790 */ {(12<<2)|2,{70,111,0}},
/* 49791 */ {(12<<2)|2,{70,111,0}},
/* 49792 */ {(12<<2)|2,{70,115,0}},
/* 49793 */ {(12<<2)|2,{70,115,0}},
/* 49794 */ {(12<<2)|2,{70,115,0}},
/* 49795 */ {(12<<2)|2,{70,115,0}},
/* 49796 */ {(12<<2)|2,{70,115,0}},
/* 49797 */ {(12<<2)|2,{70,115,0}},
/* 49798 */ {(12<<2)|2,{70,115,0}},
/* 49799 */ {(12<<2)|2,{70,115,0}},
/* 49800 */ {(12<<2)|2,{70,115,0}},
/* 49801 */ {(12<<2)|2,{70,115,0}},
/* 49802 */ {(12<<2)|2,{70,115,0}},
/* 49803 */ {(12<<2)|2,{70,115,0}},
/* 49804 */ {(12<<2)|2,{70,115,0}},
/* 49805 */ {(12<<2)|2,{70,115,0}},
/* 49806 */ {(12<<2)|2,{70,115,0}},
/* 49807 */ {(12<<2)|2,{70,115,0}},
/* 49808 */ {(12<<2)|2,{70,116,0}},
/* 49809 */ {(12<<2)|2,{70,116,0}},
/* 49810 */ {(12<<2)|2,{70,116,0}},
/* 49811 */ {(12<<2)|2,{70,116,0}},
/* 49812 */ {(12<<2)|2,{70,116,0}},
/* 49813 */ {(12<<2)|2,{70,116,0}},
/* 49814 */ {(12<<2)|2,{70,116,0}},
/* 49815 */ {(12<<2)|2,{70,116,0}},
/* 49816 */ {(12<<2)|2,{70,116,0}},
/* 49817 */ {(12<<2)|2,{70,116,0}},
/* 49818 */ {(12<<2)|2,{70,116,0}},
/* 49819 */ {(12<<2)|2,{70,116,0}},
/* 49820 */ {(12<<2)|2,{70,116,0}},
/* 49821 */ {(12<<2)|2,{70,116,0}},
/* 49822 */ {(12<<2)|2,{70,116,0}},
/* 49823 */ {(12<<2)|2,{70,116,0}},
/* 49824 */ {(13<<2)|2,{70,32,0}},
/* 49825 */ {(13<<2)|2,{70,32,0}},
/* 49826 */ {(13<<2)|2,{70,32,0}},
/* 49827 */ {(13<<2)|2,{70,32,0}},
/* 49828 */ {(13<<2)|2,{70,32,0}},
/* 49829 */ {(13<<2)|2,{70,32,0}},
/* 49830 */ {(13<<2)|2,{70,32,0}},
/* 49831 */ {(13<<2)|2,{70,32,0}},
/* 49832 */ {(13<<2)|2,{70,37,0}},
/* 49833 */ {(13<<2)|2,{70,37,0}},
/* 49834 */ {(13<<2)|2,{70,37,0}},
/* 49835 */ {(13<<2)|2,{70,37,0}},
/* 49836 */ {(13<<2)|2,{70,37,0}},
/* 49837 */ {(13<<2)|2,{70,37,0}},
/* 49838 */ {(13<<2)|2,{70,37,0}},
/* 49839 */ {(13<<2)|2,{70,37,0}},
/* 49840 */ {(13<<2)|2,{70,45,0}},
/* 49841 */ {(13<<2)|2,{70,45,0}},
/* 49842 */ {(13<<2)|2,{70,45,0}},
/* 49843 */ {(13<<2)|2,{70,45,0}},
/* 49844 */ {(13<<2)|2,{70,45,0}},
/* 49845 */ {(13<<2)|2,{70,45,0}},
/* 49846 */ {(13<<2)|2,{70,45,0}},
/* 49847 */ {(13<<2)|2,{70,45,0}},
/* 49848 */ {(13<<2)|2,{70,46,0}},
/* 49849 */ {(13<<2)|2,{70,46,0}},
/* 49850 */ {(13<<2)|2,{70,46,0}},
/* 49851 */ {(13<<2)|2,{70,46,0}},
/* 49852 */ {(13<<2)|2,{70,46,0}},
/* 49853 */ {(13<<2)|2,{70,46,0}},
/* 49854 */ {(13<<2)|2,{70,46,0}},
/* 49855 */ {(13<<2)|2,{70,46,0}},
/* 49856 */ {(13<<2)|2,{70,47,0}},
/* 49857 */ {(13<<2)|2,{70,47,0}},
/* 49858 */ {(13<<2)|2,{70,47,0}},
/* 49859 */ {(13<<2)|2,{70,47,0}},
/* 49860 */ {(13<<2)|2,{70,47,0}},
/* 49861 */ {(13<<2)|2,{70,47,0}},
/* 49862 */ {(13<<2)|2,{70,47,0}},
/* 49863 */ {(13<<2)|2,{70,47,0}},
/* 49864 */ {(13<<2)|2,{70,51,0}},
/* 49865 */ {(13<<2)|2,{70,51,0}},
/* 49866 */ {(13<<2)|2,{70,51,0}},
/* 49867 */ {(13<<2)|2,{70,51,0}},
/* 49868 */ {(13<<2)|2,{70,51,0}},
/* 49869 */ {(13<<2)|2,{70,51,0}},
/* 49870 */ {(13<<2)|2,{70,51,0}},
/* 49871 */ {(13<<2)|2,{70,51,0}},
/* 49872 */ {(13<<2)|2,{70,52,0}},
/* 49873 */ {(13<<2)|2,{70,52,0}},
/* 49874 */ {(13<<2)|2,{70,52,0}},
/* 49875 */ {(13<<2)|2,{70,52,0}},
/* 49876 */ {(13<<2)|2,{70,52,0}},
/* 49877 */ {(13<<2)|2,{70,52,0}},
/* 49878 */ {(13<<2)|2,{70,52,0}},
/* 49879 */ {(13<<2)|2,{70,52,0}},
/* 49880 */ {(13<<2)|2,{70,53,0}},
/* 49881 */ {(13<<2)|2,{70,53,0}},
/* 49882 */ {(13<<2)|2,{70,53,0}},
/* 49883 */ {(13<<2)|2,{70,53,0}},
/* 49884 */ {(13<<2)|2,{70,53,0}},
/* 49885 */ {(13<<2)|2,{70,53,0}},
/* 49886 */ {(13<<2)|2,{70,53,0}},
/* 49887 */ {(13<<2)|2,{70,53,0}},
/* 49888 */ {(13<<2)|2,{70,54,0}},
/* 49889 */ {(13<<2)|2,{70,54,0}},
/* 49890 */ {(13<<2)|2,{70,54,0}},
/* 49891 */ {(13<<2)|2,{70,54,0}},
/* 49892 */ {(13<<2)|2,{70,54,0}},
/* 49893 */ {(13<<2)|2,{70,54,0}},
/* 49894 */ {(13<<2)|2,{70,54,0}},
/* 49895 */ {(13<<2)|2,{70,54,0}},
/* 49896 */ {(13<<2)|2,{70,55,0}},
/* 49897 */ {(13<<2)|2,{70,55,0}},
/* 49898 */ {(13<<2)|2,{70,55,0}},
/* 49899 */ {(13<<2)|2,{70,55,0}},
/* 49900 */ {(13<<2)|2,{70,55,0}},
/* 49901 */ {(13<<2)|2,{70,55,0}},
/* 49902 */ {(13<<2)|2,{70,55,0}},
/* 49903 */ {(13<<2)|2,{70,55,0}},
/* 49904 */ {(13<<2)|2,{70,56,0}},
/* 49905 */ {(13<<2)|2,{70,56,0}},
/* 49906 */ {(13<<2)|2,{70,56,0}},
/* 49907 */ {(13<<2)|2,{70,56,0}},
/* 49908 */ {(13<<2)|2,{70,56,0}},
/* 49909 */ {(13<<2)|2,{70,56,0}},
/* 49910 */ {(13<<2)|2,{70,56,0}},
/* 49911 */ {(13<<2)|2,{70,56,0}},
/* 49912 */ {(13<<2)|2,{70,57,0}},
/* 49913 */ {(13<<2)|2,{70,57,0}},
/* 49914 */ {(13<<2)|2,{70,57,0}},
/* 49915 */ {(13<<2)|2,{70,57,0}},
/* 49916 */ {(13<<2)|2,{70,57,0}},
/* 49917 */ {(13<<2)|2,{70,57,0}},
/* 49918 */ {(13<<2)|2,{70,57,0}},
/* 49919 */ {(13<<2)|2,{70,57,0}},
/* 49920 */ {(13<<2)|2,{70,61,0}},
/* 49921 */ {(13<<2)|2,{70,61,0}},
/* 49922 */ {(13<<2)|2,{70,61,0}},
/* 49923 */ {(13<<2)|2,{70,61,0}},
/* 49924 */ {(13<<2)|2,{70,61,0}},
/* 49925 */ {(13<<2)|2,{70,61,0}},
/* 49926 */ {(13<<2)|2,{70,61,0}},
/* 49927 */ {(13<<2)|2,{70,61,0}},
/* 49928 */ {(13<<2)|2,{70,65,0}},
/* 49929 */ {(13<<2)|2,{70,65,0}},
/* 49930 */ {(13<<2)|2,{70,65,0}},
/* 49931 */ {(13<<2)|2,{70,65,0}},
/* 49932 */ {(13<<2)|2,{70,65,0}},
/* 49933 */ {(13<<2)|2,{70,65,0}},
/* 49934 */ {(13<<2)|2,{70,65,0}},
/* 49935 */ {(13<<2)|2,{70,65,0}},
/* 49936 */ {(13<<2)|2,{70,95,0}},
/* 49937 */ {(13<<2)|2,{70,95,0}},
/* 49938 */ {(13<<2)|2,{70,95,0}},
/* 49939 */ {(13<<2)|2,{70,95,0}},
/* 49940 */ {(13<<2)|2,{70,95,0}},
/* 49941 */ {(13<<2)|2,{70,95,0}},
/* 49942 */ {(13<<2)|2,{70,95,0}},
/* 49943 */ {(13<<2)|2,{70,95,0}},
/* 49944 */ {(13<<2)|2,{70,98,0}},
/* 49945 */ {(13<<2)|2,{70,98,0}},
/* 49946 */ {(13<<2)|2,{70,98,0}},
/* 49947 */ {(13<<2)|2,{70,98,0}},
/* 49948 */ {(13<<2)|2,{70,98,0}},
/* 49949 */ {(13<<2)|2,{70,98,0}},
/* 49950 */ {(13<<2)|2,{70,98,0}},
/* 49951 */ {(13<<2)|2,{70,98,0}},
/* 49952 */ {(13<<2)|2,{70,100,0}},
/* 49953 */ {(13<<2)|2,{70,100,0}},
/* 49954 */ {(13<<2)|2,{70,100,0}},
/* 49955 */ {(13<<2)|2,{70,100,0}},
/* 49956 */ {(13<<2)|2,{70,100,0}},
/* 49957 */ {(13<<2)|2,{70,100,0}},
/* 49958 */ {(13<<2)|2,{70,100,0}},
/* 49959 */ {(13<<2)|2,{70,100,0}},
/* 49960 */ {(13<<2)|2,{70,102,0}},
/* 49961 */ {(13<<2)|2,{70,102,0}},
/* 49962 */ {(13<<2)|2,{70,102,0}},
/* 49963 */ {(13<<2)|2,{70,102,0}},
/* 49964 */ {(13<<2)|2,{70,102,0}},
/* 49965 */ {(13<<2)|2,{70,102,0}},
/* 49966 */ {(13<<2)|2,{70,102,0}},
/* 49967 */ {(13<<2)|2,{70,102,0}},
/* 49968 */ {(13<<2)|2,{70,103,0}},
/* 49969 */ {(13<<2)|2,{70,103,0}},
/* 49970 */ {(13<<2)|2,{70,103,0}},
/* 49971 */ {(13<<2)|2,{70,103,0}},
/* 49972 */ {(13<<2)|2,{70,103,0}},
/* 49973 */ {(13<<2)|2,{70,103,0}},
/* 49974 */ {(13<<2)|2,{70,103,0}},
/* 49975 */ {(13<<2)|2,{70,103,0}},
/* 49976 */ {(13<<2)|2,{70,104,0}},
/* 49977 */ {(13<<2)|2,{70,104,0}},
/* 49978 */ {(13<<2)|2,{70,104,0}},
/* 49979 */ {(13<<2)|2,{70,104,0}},
/* 49980 */ {(13<<2)|2,{70,104,0}},
/* 49981 */ {(13<<2)|2,{70,104,0}},
/* 49982 */ {(13<<2)|2,{70,104,0}},
/* 49983 */ {(13<<2)|2,{70,104,0}},
/* 49984 */ {(13<<2)|2,{70,108,0}},
/* 49985 */ {(13<<2)|2,{70,108,0}},
/* 49986 */ {(13<<2)|2,{70,108,0}},
/* 49987 */ {(13<<2)|2,{70,108,0}},
/* 49988 */ {(13<<2)|2,{70,108,0}},
/* 49989 */ {(13<<2)|2,{70,108,0}},
/* 49990 */ {(13<<2)|2,{70,108,0}},
/* 49991 */ {(13<<2)|2,{70,108,0}},
/* 49992 */ {(13<<2)|2,{70,109,0}},
/* 49993 */ {(13<<2)|2,{70,109,0}},
/* 49994 */ {(13<<2)|2,{70,109,0}},
/* 49995 */ {(13<<2)|2,{70,109,0}},
/* 49996 */ {(13<<2)|2,{70,109,0}},
/* 49997 */ {(13<<2)|2,{70,109,0}},
/* 49998 */ {(13<<2)|2,{70,109,0}},
/* 49999 */ {(13<<2)|2,{70,109,0}},
/* 50000 */ {(13<<2)|2,{70,110,0}},
/* 50001 */ {(13<<2)|2,{70,110,0}},
/* 50002 */ {(13<<2)|2,{70,110,0}},
/* 50003 */ {(13<<2)|2,{70,110,0}},
/* 50004 */ {(13<<2)|2,{70,110,0}},
/* 50005 */ {(13<<2)|2,{70,110,0}},
/* 50006 */ {(13<<2)|2,{70,110,0}},
/* 50007 */ {(13<<2)|2,{70,110,0}},
/* 50008 */ {(13<<2)|2,{70,112,0}},
/* 50009 */ {(13<<2)|2,{70,112,0}},
/* 50010 */ {(13<<2)|2,{70,112,0}},
/* 50011 */ {(13<<2)|2,{70,112,0}},
/* 50012 */ {(13<<2)|2,{70,112,0}},
/* 50013 */ {(13<<2)|2,{70,112,0}},
/* 50014 */ {(13<<2)|2,{70,112,0}},
/* 50015 */ {(13<<2)|2,{70,112,0}},
/* 50016 */ {(13<<2)|2,{70,114,0}},
/* 50017 */ {(13<<2)|2,{70,114,0}},
/* 50018 */ {(13<<2)|2,{70,114,0}},
/* 50019 */ {(13<<2)|2,{70,114,0}},
/* 50020 */ {(13<<2)|2,{70,114,0}},
/* 50021 */ {(13<<2)|2,{70,114,0}},
/* 50022 */ {(13<<2)|2,{70,114,0}},
/* 50023 */ {(13<<2)|2,{70,114,0}},
/* 50024 */ {(13<<2)|2,{70,117,0}},
/* 50025 */ {(13<<2)|2,{70,117,0}},
/* 50026 */ {(13<<2)|2,{70,117,0}},
/* 50027 */ {(13<<2)|2,{70,117,0}},
/* 50028 */ {(13<<2)|2,{70,117,0}},
/* 50029 */ {(13<<2)|2,{70,117,0}},
/* 50030 */ {(13<<2)|2,{70,117,0}},
/* 50031 */ {(13<<2)|2,{70,117,0}},
/* 50032 */ {(14<<2)|2,{70,58,0}},
/* 50033 */ {(14<<2)|2,{70,58,0}},
/* 50034 */ {(14<<2)|2,{70,58,0}},
/* 50035 */ {(14<<2)|2,{70,58,0}},
/* 50036 */ {(14<<2)|2,{70,66,0}},
/* 50037 */ {(14<<2)|2,{70,66,0}},
/* 50038 */ {(14<<2)|2,{70,66,0}},
/* 50039 */ {(14<<2)|2,{70,66,0}},
/* 50040 */ {(14<<2)|2,{70,67,0}},
/* 50041 */ {(14<<2)|2,{70,67,0}},
/* 50042 */ {(14<<2)|2,{70,67,0}},
/* 50043 */ {(14<<2)|2,{70,67,0}},
/* 50044 */ {(14<<2)|2,{70,68,0}},
/* 50045 */ {(14<<2)|2,{70,68,0}},
/* 50046 */ {(14<<2)|2,{70,68,0}},
/* 50047 */ {(14<<2)|2,{70,68,0}},
/* 50048 */ {(14<<2)|2,{70,69,0}},
/* 50049 */ {(14<<2)|2,{70,69,0}},
/* 50050 */ {(14<<2)|2,{70,69,0}},
/* 50051 */ {(14<<2)|2,{70,69,0}},
/* 50052 */ {(14<<2)|2,{70,70,0}},
/* 50053 */ {(14<<2)|2,{70,70,0}},
/* 50054 */ {(14<<2)|2,{70,70,0}},
/* 50055 */ {(14<<2)|2,{70,70,0}},
/* 50056 */ {(14<<2)|2,{70,71,0}},
/* 50057 */ {(14<<2)|2,{70,71,0}},
/* 50058 */ {(14<<2)|2,{70,71,0}},
/* 50059 */ {(14<<2)|2,{70,71,0}},
/* 50060 */ {(14<<2)|2,{70,72,0}},
/* 50061 */ {(14<<2)|2,{70,72,0}},
/* 50062 */ {(14<<2)|2,{70,72,0}},
/* 50063 */ {(14<<2)|2,{70,72,0}},
/* 50064 */ {(14<<2)|2,{70,73,0}},
/* 50065 */ {(14<<2)|2,{70,73,0}},
/* 50066 */ {(14<<2)|2,{70,73,0}},
/* 50067 */ {(14<<2)|2,{70,73,0}},
/* 50068 */ {(14<<2)|2,{70,74,0}},
/* 50069 */ {(14<<2)|2,{70,74,0}},
/* 50070 */ {(14<<2)|2,{70,74,0}},
/* 50071 */ {(14<<2)|2,{70,74,0}},
/* 50072 */ {(14<<2)|2,{70,75,0}},
/* 50073 */ {(14<<2)|2,{70,75,0}},
/* 50074 */ {(14<<2)|2,{70,75,0}},
/* 50075 */ {(14<<2)|2,{70,75,0}},
/* 50076 */ {(14<<2)|2,{70,76,0}},
/* 50077 */ {(14<<2)|2,{70,76,0}},
/* 50078 */ {(14<<2)|2,{70,76,0}},
/* 50079 */ {(14<<2)|2,{70,76,0}},
/* 50080 */ {(14<<2)|2,{70,77,0}},
/* 50081 */ {(14<<2)|2,{70,77,0}},
/* 50082 */ {(14<<2)|2,{70,77,0}},
/* 50083 */ {(14<<2)|2,{70,77,0}},
/* 50084 */ {(14<<2)|2,{70,78,0}},
/* 50085 */ {(14<<2)|2,{70,78,0}},
/* 50086 */ {(14<<2)|2,{70,78,0}},
/* 50087 */ {(14<<2)|2,{70,78,0}},
/* 50088 */ {(14<<2)|2,{70,79,0}},
/* 50089 */ {(14<<2)|2,{70,79,0}},
/* 50090 */ {(14<<2)|2,{70,79,0}},
/* 50091 */ {(14<<2)|2,{70,79,0}},
/* 50092 */ {(14<<2)|2,{70,80,0}},
/* 50093 */ {(14<<2)|2,{70,80,0}},
/* 50094 */ {(14<<2)|2,{70,80,0}},
/* 50095 */ {(14<<2)|2,{70,80,0}},
/* 50096 */ {(14<<2)|2,{70,81,0}},
/* 50097 */ {(14<<2)|2,{70,81,0}},
/* 50098 */ {(14<<2)|2,{70,81,0}},
/* 50099 */ {(14<<2)|2,{70,81,0}},
/* 50100 */ {(14<<2)|2,{70,82,0}},
/* 50101 */ {(14<<2)|2,{70,82,0}},
/* 50102 */ {(14<<2)|2,{70,82,0}},
/* 50103 */ {(14<<2)|2,{70,82,0}},
/* 50104 */ {(14<<2)|2,{70,83,0}},
/* 50105 */ {(14<<2)|2,{70,83,0}},
/* 50106 */ {(14<<2)|2,{70,83,0}},
/* 50107 */ {(14<<2)|2,{70,83,0}},
/* 50108 */ {(14<<2)|2,{70,84,0}},
/* 50109 */ {(14<<2)|2,{70,84,0}},
/* 50110 */ {(14<<2)|2,{70,84,0}},
/* 50111 */ {(14<<2)|2,{70,84,0}},
/* 50112 */ {(14<<2)|2,{70,85,0}},
/* 50113 */ {(14<<2)|2,{70,85,0}},
/* 50114 */ {(14<<2)|2,{70,85,0}},
/* 50115 */ {(14<<2)|2,{70,85,0}},
/* 50116 */ {(14<<2)|2,{70,86,0}},
/* 50117 */ {(14<<2)|2,{70,86,0}},
/* 50118 */ {(14<<2)|2,{70,86,0}},
/* 50119 */ {(14<<2)|2,{70,86,0}},
/* 50120 */ {(14<<2)|2,{70,87,0}},
/* 50121 */ {(14<<2)|2,{70,87,0}},
/* 50122 */ {(14<<2)|2,{70,87,0}},
/* 50123 */ {(14<<2)|2,{70,87,0}},
/* 50124 */ {(14<<2)|2,{70,89,0}},
/* 50125 */ {(14<<2)|2,{70,89,0}},
/* 50126 */ {(14<<2)|2,{70,89,0}},
/* 50127 */ {(14<<2)|2,{70,89,0}},
/* 50128 */ {(14<<2)|2,{70,106,0}},
/* 50129 */ {(14<<2)|2,{70,106,0}},
/* 50130 */ {(14<<2)|2,{70,106,0}},
/* 50131 */ {(14<<2)|2,{70,106,0}},
/* 50132 */ {(14<<2)|2,{70,107,0}},
/* 50133 */ {(14<<2)|2,{70,107,0}},
/* 50134 */ {(14<<2)|2,{70,107,0}},
/* 50135 */ {(14<<2)|2,{70,107,0}},
/* 50136 */ {(14<<2)|2,{70,113,0}},
/* 50137 */ {(14<<2)|2,{70,113,0}},
/* 50138 */ {(14<<2)|2,{70,113,0}},
/* 50139 */ {(14<<2)|2,{70,113,0}},
/* 50140 */ {(14<<2)|2,{70,118,0}},
/* 50141 */ {(14<<2)|2,{70,118,0}},
/* 50142 */ {(14<<2)|2,{70,118,0}},
/* 50143 */ {(14<<2)|2,{70,118,0}},
/* 50144 */ {(14<<2)|2,{70,119,0}},
/* 50145 */ {(14<<2)|2,{70,119,0}},
/* 50146 */ {(14<<2)|2,{70,119,0}},
/* 50147 */ {(14<<2)|2,{70,119,0}},
/* 50148 */ {(14<<2)|2,{70,120,0}},
/* 50149 */ {(14<<2)|2,{70,120,0}},
/* 50150 */ {(14<<2)|2,{70,120,0}},
/* 50151 */ {(14<<2)|2,{70,120,0}},
/* 50152 */ {(14<<2)|2,{70,121,0}},
/* 50153 */ {(14<<2)|2,{70,121,0}},
/* 50154 */ {(14<<2)|2,{70,121,0}},
/* 50155 */ {(14<<2)|2,{70,121,0}},
/* 50156 */ {(14<<2)|2,{70,122,0}},
/* 50157 */ {(14<<2)|2,{70,122,0}},
/* 50158 */ {(14<<2)|2,{70,122,0}},
/* 50159 */ {(14<<2)|2,{70,122,0}},
/* 50160 */ {(15<<2)|2,{70,38,0}},
/* 50161 */ {(15<<2)|2,{70,38,0}},
/* 50162 */ {(15<<2)|2,{70,42,0}},
/* 50163 */ {(15<<2)|2,{70,42,0}},
/* 50164 */ {(15<<2)|2,{70,44,0}},
/* 50165 */ {(15<<2)|2,{70,44,0}},
/* 50166 */ {(15<<2)|2,{70,59,0}},
/* 50167 */ {(15<<2)|2,{70,59,0}},
/* 50168 */ {(15<<2)|2,{70,88,0}},
/* 50169 */ {(15<<2)|2,{70,88,0}},
/* 50170 */ {(15<<2)|2,{70,90,0}},
/* 50171 */ {(15<<2)|2,{70,90,0}},
/* 50172 */ {(7<<2)|1,{70,0,0}},
/* 50173 */ {(7<<2)|1,{70,0,0}},
/* 50174 */ {(7<<2)|1,{70,0,0}},
/* 50175 */ {(7<<2)|1,{70,0,0}},
/* 50176 */ {(12<<2)|2,{71,48,0}},
/* 50177 */ {(12<<2)|2,{71,48,0}},
/* 50178 */ {(12<<2)|2,{71,48,0}},
/* 50179 */ {(12<<2)|2,{71,48,0}},
/* 50180 */ {(12<<2)|2,{71,48,0}},
/* 50181 */ {(12<<2)|2,{71,48,0}},
/* 50182 */ {(12<<2)|2,{71,48,0}},
/* 50183 */ {(12<<2)|2,{71,48,0}},
/* 50184 */ {(12<<2)|2,{71,48,0}},
/* 50185 */ {(12<<2)|2,{71,48,0}},
/* 50186 */ {(12<<2)|2,{71,48,0}},
/* 50187 */ {(12<<2)|2,{71,48,0}},
/* 50188 */ {(12<<2)|2,{71,48,0}},
/* 50189 */ {(12<<2)|2,{71,48,0}},
/* 50190 */ {(12<<2)|2,{71,48,0}},
/* 50191 */ {(12<<2)|2,{71,48,0}},
/* 50192 */ {(12<<2)|2,{71,49,0}},
/* 50193 */ {(12<<2)|2,{71,49,0}},
/* 50194 */ {(12<<2)|2,{71,49,0}},
/* 50195 */ {(12<<2)|2,{71,49,0}},
/* 50196 */ {(12<<2)|2,{71,49,0}},
/* 50197 */ {(12<<2)|2,{71,49,0}},
/* 50198 */ {(12<<2)|2,{71,49,0}},
/* 50199 */ {(12<<2)|2,{71,49,0}},
/* 50200 */ {(12<<2)|2,{71,49,0}},
/* 50201 */ {(12<<2)|2,{71,49,0}},
/* 50202 */ {(12<<2)|2,{71,49,0}},
/* 50203 */ {(12<<2)|2,{71,49,0}},
/* 50204 */ {(12<<2)|2,{71,49,0}},
/* 50205 */ {(12<<2)|2,{71,49,0}},
/* 50206 */ {(12<<2)|2,{71,49,0}},
/* 50207 */ {(12<<2)|2,{71,49,0}},
/* 50208 */ {(12<<2)|2,{71,50,0}},
/* 50209 */ {(12<<2)|2,{71,50,0}},
/* 50210 */ {(12<<2)|2,{71,50,0}},
/* 50211 */ {(12<<2)|2,{71,50,0}},
/* 50212 */ {(12<<2)|2,{71,50,0}},
/* 50213 */ {(12<<2)|2,{71,50,0}},
/* 50214 */ {(12<<2)|2,{71,50,0}},
/* 50215 */ {(12<<2)|2,{71,50,0}},
/* 50216 */ {(12<<2)|2,{71,50,0}},
/* 50217 */ {(12<<2)|2,{71,50,0}},
/* 50218 */ {(12<<2)|2,{71,50,0}},
/* 50219 */ {(12<<2)|2,{71,50,0}},
/* 50220 */ {(12<<2)|2,{71,50,0}},
/* 50221 */ {(12<<2)|2,{71,50,0}},
/* 50222 */ {(12<<2)|2,{71,50,0}},
/* 50223 */ {(12<<2)|2,{71,50,0}},
/* 50224 */ {(12<<2)|2,{71,97,0}},
/* 50225 */ {(12<<2)|2,{71,97,0}},
/* 50226 */ {(12<<2)|2,{71,97,0}},
/* 50227 */ {(12<<2)|2,{71,97,0}},
/* 50228 */ {(12<<2)|2,{71,97,0}},
/* 50229 */ {(12<<2)|2,{71,97,0}},
/* 50230 */ {(12<<2)|2,{71,97,0}},
/* 50231 */ {(12<<2)|2,{71,97,0}},
/* 50232 */ {(12<<2)|2,{71,97,0}},
/* 50233 */ {(12<<2)|2,{71,97,0}},
/* 50234 */ {(12<<2)|2,{71,97,0}},
/* 50235 */ {(12<<2)|2,{71,97,0}},
/* 50236 */ {(12<<2)|2,{71,97,0}},
/* 50237 */ {(12<<2)|2,{71,97,0}},
/* 50238 */ {(12<<2)|2,{71,97,0}},
/* 50239 */ {(12<<2)|2,{71,97,0}},
/* 50240 */ {(12<<2)|2,{71,99,0}},
/* 50241 */ {(12<<2)|2,{71,99,0}},
/* 50242 */ {(12<<2)|2,{71,99,0}},
/* 50243 */ {(12<<2)|2,{71,99,0}},
/* 50244 */ {(12<<2)|2,{71,99,0}},
/* 50245 */ {(12<<2)|2,{71,99,0}},
/* 50246 */ {(12<<2)|2,{71,99,0}},
/* 50247 */ {(12<<2)|2,{71,99,0}},
/* 50248 */ {(12<<2)|2,{71,99,0}},
/* 50249 */ {(12<<2)|2,{71,99,0}},
/* 50250 */ {(12<<2)|2,{71,99,0}},
/* 50251 */ {(12<<2)|2,{71,99,0}},
/* 50252 */ {(12<<2)|2,{71,99,0}},
/* 50253 */ {(12<<2)|2,{71,99,0}},
/* 50254 */ {(12<<2)|2,{71,99,0}},
/* 50255 */ {(12<<2)|2,{71,99,0}},
/* 50256 */ {(12<<2)|2,{71,101,0}},
/* 50257 */ {(12<<2)|2,{71,101,0}},
/* 50258 */ {(12<<2)|2,{71,101,0}},
/* 50259 */ {(12<<2)|2,{71,101,0}},
/* 50260 */ {(12<<2)|2,{71,101,0}},
/* 50261 */ {(12<<2)|2,{71,101,0}},
/* 50262 */ {(12<<2)|2,{71,101,0}},
/* 50263 */ {(12<<2)|2,{71,101,0}},
/* 50264 */ {(12<<2)|2,{71,101,0}},
/* 50265 */ {(12<<2)|2,{71,101,0}},
/* 50266 */ {(12<<2)|2,{71,101,0}},
/* 50267 */ {(12<<2)|2,{71,101,0}},
/* 50268 */ {(12<<2)|2,{71,101,0}},
/* 50269 */ {(12<<2)|2,{71,101,0}},
/* 50270 */ {(12<<2)|2,{71,101,0}},
/* 50271 */ {(12<<2)|2,{71,101,0}},
/* 50272 */ {(12<<2)|2,{71,105,0}},
/* 50273 */ {(12<<2)|2,{71,105,0}},
/* 50274 */ {(12<<2)|2,{71,105,0}},
/* 50275 */ {(12<<2)|2,{71,105,0}},
/* 50276 */ {(12<<2)|2,{71,105,0}},
/* 50277 */ {(12<<2)|2,{71,105,0}},
/* 50278 */ {(12<<2)|2,{71,105,0}},
/* 50279 */ {(12<<2)|2,{71,105,0}},
/* 50280 */ {(12<<2)|2,{71,105,0}},
/* 50281 */ {(12<<2)|2,{71,105,0}},
/* 50282 */ {(12<<2)|2,{71,105,0}},
/* 50283 */ {(12<<2)|2,{71,105,0}},
/* 50284 */ {(12<<2)|2,{71,105,0}},
/* 50285 */ {(12<<2)|2,{71,105,0}},
/* 50286 */ {(12<<2)|2,{71,105,0}},
/* 50287 */ {(12<<2)|2,{71,105,0}},
/* 50288 */ {(12<<2)|2,{71,111,0}},
/* 50289 */ {(12<<2)|2,{71,111,0}},
/* 50290 */ {(12<<2)|2,{71,111,0}},
/* 50291 */ {(12<<2)|2,{71,111,0}},
/* 50292 */ {(12<<2)|2,{71,111,0}},
/* 50293 */ {(12<<2)|2,{71,111,0}},
/* 50294 */ {(12<<2)|2,{71,111,0}},
/* 50295 */ {(12<<2)|2,{71,111,0}},
/* 50296 */ {(12<<2)|2,{71,111,0}},
/* 50297 */ {(12<<2)|2,{71,111,0}},
/* 50298 */ {(12<<2)|2,{71,111,0}},
/* 50299 */ {(12<<2)|2,{71,111,0}},
/* 50300 */ {(12<<2)|2,{71,111,0}},
/* 50301 */ {(12<<2)|2,{71,111,0}},
/* 50302 */ {(12<<2)|2,{71,111,0}},
/* 50303 */ {(12<<2)|2,{71,111,0}},
/* 50304 */ {(12<<2)|2,{71,115,0}},
/* 50305 */ {(12<<2)|2,{71,115,0}},
/* 50306 */ {(12<<2)|2,{71,115,0}},
/* 50307 */ {(12<<2)|2,{71,115,0}},
/* 50308 */ {(12<<2)|2,{71,115,0}},
/* 50309 */ {(12<<2)|2,{71,115,0}},
/* 50310 */ {(12<<2)|2,{71,115,0}},
/* 50311 */ {(12<<2)|2,{71,115,0}},
/* 50312 */ {(12<<2)|2,{71,115,0}},
/* 50313 */ {(12<<2)|2,{71,115,0}},
/* 50314 */ {(12<<2)|2,{71,115,0}},
/* 50315 */ {(12<<2)|2,{71,115,0}},
/* 50316 */ {(12<<2)|2,{71,115,0}},
/* 50317 */ {(12<<2)|2,{71,115,0}},
/* 50318 */ {(12<<2)|2,{71,115,0}},
/* 50319 */ {(12<<2)|2,{71,115,0}},
/* 50320 */ {(12<<2)|2,{71,116,0}},
/* 50321 */ {(12<<2)|2,{71,116,0}},
/* 50322 */ {(12<<2)|2,{71,116,0}},
/* 50323 */ {(12<<2)|2,{71,116,0}},
/* 50324 */ {(12<<2)|2,{71,116,0}},
/* 50325 */ {(12<<2)|2,{71,116,0}},
/* 50326 */ {(12<<2)|2,{71,116,0}},
/* 50327 */ {(12<<2)|2,{71,116,0}},
/* 50328 */ {(12<<2)|2,{71,116,0}},
/* 50329 */ {(12<<2)|2,{71,116,0}},
/* 50330 */ {(12<<2)|2,{71,116,0}},
/* 50331 */ {(12<<2)|2,{71,116,0}},
/* 50332 */ {(12<<2)|2,{71,116,0}},
/* 50333 */ {(12<<2)|2,{71,116,0}},
/* 50334 */ {(12<<2)|2,{71,116,0}},
/* 50335 */ {(12<<2)|2,{71,116,0}},
/* 50336 */ {(13<<2)|2,{71,32,0}},
/* 50337 */ {(13<<2)|2,{71,32,0}},
/* 50338 */ {(13<<2)|2,{71,32,0}},
/* 50339 */ {(13<<2)|2,{71,32,0}},
/* 50340 */ {(13<<2)|2,{71,32,0}},
/* 50341 */ {(13<<2)|2,{71,32,0}},
/* 50342 */ {(13<<2)|2,{71,32,0}},
/* 50343 */ {(13<<2)|2,{71,32,0}},
/* 50344 */ {(13<<2)|2,{71,37,0}},
/* 50345 */ {(13<<2)|2,{71,37,0}},
/* 50346 */ {(13<<2)|2,{71,37,0}},
/* 50347 */ {(13<<2)|2,{71,37,0}},
/* 50348 */ {(13<<2)|2,{71,37,0}},
/* 50349 */ {(13<<2)|2,{71,37,0}},
/* 50350 */ {(13<<2)|2,{71,37,0}},
/* 50351 */ {(13<<2)|2,{71,37,0}},
/* 50352 */ {(13<<2)|2,{71,45,0}},
/* 50353 */ {(13<<2)|2,{71,45,0}},
/* 50354 */ {(13<<2)|2,{71,45,0}},
/* 50355 */ {(13<<2)|2,{71,45,0}},
/* 50356 */ {(13<<2)|2,{71,45,0}},
/* 50357 */ {(13<<2)|2,{71,45,0}},
/* 50358 */ {(13<<2)|2,{71,45,0}},
/* 50359 */ {(13<<2)|2,{71,45,0}},
/* 50360 */ {(13<<2)|2,{71,46,0}},
/* 50361 */ {(13<<2)|2,{71,46,0}},
/* 50362 */ {(13<<2)|2,{71,46,0}},
/* 50363 */ {(13<<2)|2,{71,46,0}},
/* 50364 */ {(13<<2)|2,{71,46,0}},
/* 50365 */ {(13<<2)|2,{71,46,0}},
/* 50366 */ {(13<<2)|2,{71,46,0}},
/* 50367 */ {(13<<2)|2,{71,46,0}},
/* 50368 */ {(13<<2)|2,{71,47,0}},
/* 50369 */ {(13<<2)|2,{71,47,0}},
/* 50370 */ {(13<<2)|2,{71,47,0}},
/* 50371 */ {(13<<2)|2,{71,47,0}},
/* 50372 */ {(13<<2)|2,{71,47,0}},
/* 50373 */ {(13<<2)|2,{71,47,0}},
/* 50374 */ {(13<<2)|2,{71,47,0}},
/* 50375 */ {(13<<2)|2,{71,47,0}},
/* 50376 */ {(13<<2)|2,{71,51,0}},
/* 50377 */ {(13<<2)|2,{71,51,0}},
/* 50378 */ {(13<<2)|2,{71,51,0}},
/* 50379 */ {(13<<2)|2,{71,51,0}},
/* 50380 */ {(13<<2)|2,{71,51,0}},
/* 50381 */ {(13<<2)|2,{71,51,0}},
/* 50382 */ {(13<<2)|2,{71,51,0}},
/* 50383 */ {(13<<2)|2,{71,51,0}},
/* 50384 */ {(13<<2)|2,{71,52,0}},
/* 50385 */ {(13<<2)|2,{71,52,0}},
/* 50386 */ {(13<<2)|2,{71,52,0}},
/* 50387 */ {(13<<2)|2,{71,52,0}},
/* 50388 */ {(13<<2)|2,{71,52,0}},
/* 50389 */ {(13<<2)|2,{71,52,0}},
/* 50390 */ {(13<<2)|2,{71,52,0}},
/* 50391 */ {(13<<2)|2,{71,52,0}},
/* 50392 */ {(13<<2)|2,{71,53,0}},
/* 50393 */ {(13<<2)|2,{71,53,0}},
/* 50394 */ {(13<<2)|2,{71,53,0}},
/* 50395 */ {(13<<2)|2,{71,53,0}},
/* 50396 */ {(13<<2)|2,{71,53,0}},
/* 50397 */ {(13<<2)|2,{71,53,0}},
/* 50398 */ {(13<<2)|2,{71,53,0}},
/* 50399 */ {(13<<2)|2,{71,53,0}},
/* 50400 */ {(13<<2)|2,{71,54,0}},
/* 50401 */ {(13<<2)|2,{71,54,0}},
/* 50402 */ {(13<<2)|2,{71,54,0}},
/* 50403 */ {(13<<2)|2,{71,54,0}},
/* 50404 */ {(13<<2)|2,{71,54,0}},
/* 50405 */ {(13<<2)|2,{71,54,0}},
/* 50406 */ {(13<<2)|2,{71,54,0}},
/* 50407 */ {(13<<2)|2,{71,54,0}},
/* 50408 */ {(13<<2)|2,{71,55,0}},
/* 50409 */ {(13<<2)|2,{71,55,0}},
/* 50410 */ {(13<<2)|2,{71,55,0}},
/* 50411 */ {(13<<2)|2,{71,55,0}},
/* 50412 */ {(13<<2)|2,{71,55,0}},
/* 50413 */ {(13<<2)|2,{71,55,0}},
/* 50414 */ {(13<<2)|2,{71,55,0}},
/* 50415 */ {(13<<2)|2,{71,55,0}},
/* 50416 */ {(13<<2)|2,{71,56,0}},
/* 50417 */ {(13<<2)|2,{71,56,0}},
/* 50418 */ {(13<<2)|2,{71,56,0}},
/* 50419 */ {(13<<2)|2,{71,56,0}},
/* 50420 */ {(13<<2)|2,{71,56,0}},
/* 50421 */ {(13<<2)|2,{71,56,0}},
/* 50422 */ {(13<<2)|2,{71,56,0}},
/* 50423 */ {(13<<2)|2,{71,56,0}},
/* 50424 */ {(13<<2)|2,{71,57,0}},
/* 50425 */ {(13<<2)|2,{71,57,0}},
/* 50426 */ {(13<<2)|2,{71,57,0}},
/* 50427 */ {(13<<2)|2,{71,57,0}},
/* 50428 */ {(13<<2)|2,{71,57,0}},
/* 50429 */ {(13<<2)|2,{71,57,0}},
/* 50430 */ {(13<<2)|2,{71,57,0}},
/* 50431 */ {(13<<2)|2,{71,57,0}},
/* 50432 */ {(13<<2)|2,{71,61,0}},
/* 50433 */ {(13<<2)|2,{71,61,0}},
/* 50434 */ {(13<<2)|2,{71,61,0}},
/* 50435 */ {(13<<2)|2,{71,61,0}},
/* 50436 */ {(13<<2)|2,{71,61,0}},
/* 50437 */ {(13<<2)|2,{71,61,0}},
/* 50438 */ {(13<<2)|2,{71,61,0}},
/* 50439 */ {(13<<2)|2,{71,61,0}},
/* 50440 */ {(13<<2)|2,{71,65,0}},
/* 50441 */ {(13<<2)|2,{71,65,0}},
/* 50442 */ {(13<<2)|2,{71,65,0}},
/* 50443 */ {(13<<2)|2,{71,65,0}},
/* 50444 */ {(13<<2)|2,{71,65,0}},
/* 50445 */ {(13<<2)|2,{71,65,0}},
/* 50446 */ {(13<<2)|2,{71,65,0}},
/* 50447 */ {(13<<2)|2,{71,65,0}},
/* 50448 */ {(13<<2)|2,{71,95,0}},
/* 50449 */ {(13<<2)|2,{71,95,0}},
/* 50450 */ {(13<<2)|2,{71,95,0}},
/* 50451 */ {(13<<2)|2,{71,95,0}},
/* 50452 */ {(13<<2)|2,{71,95,0}},
/* 50453 */ {(13<<2)|2,{71,95,0}},
/* 50454 */ {(13<<2)|2,{71,95,0}},
/* 50455 */ {(13<<2)|2,{71,95,0}},
/* 50456 */ {(13<<2)|2,{71,98,0}},
/* 50457 */ {(13<<2)|2,{71,98,0}},
/* 50458 */ {(13<<2)|2,{71,98,0}},
/* 50459 */ {(13<<2)|2,{71,98,0}},
/* 50460 */ {(13<<2)|2,{71,98,0}},
/* 50461 */ {(13<<2)|2,{71,98,0}},
/* 50462 */ {(13<<2)|2,{71,98,0}},
/* 50463 */ {(13<<2)|2,{71,98,0}},
/* 50464 */ {(13<<2)|2,{71,100,0}},
/* 50465 */ {(13<<2)|2,{71,100,0}},
/* 50466 */ {(13<<2)|2,{71,100,0}},
/* 50467 */ {(13<<2)|2,{71,100,0}},
/* 50468 */ {(13<<2)|2,{71,100,0}},
/* 50469 */ {(13<<2)|2,{71,100,0}},
/* 50470 */ {(13<<2)|2,{71,100,0}},
/* 50471 */ {(13<<2)|2,{71,100,0}},
/* 50472 */ {(13<<2)|2,{71,102,0}},
/* 50473 */ {(13<<2)|2,{71,102,0}},
/* 50474 */ {(13<<2)|2,{71,102,0}},
/* 50475 */ {(13<<2)|2,{71,102,0}},
/* 50476 */ {(13<<2)|2,{71,102,0}},
/* 50477 */ {(13<<2)|2,{71,102,0}},
/* 50478 */ {(13<<2)|2,{71,102,0}},
/* 50479 */ {(13<<2)|2,{71,102,0}},
/* 50480 */ {(13<<2)|2,{71,103,0}},
/* 50481 */ {(13<<2)|2,{71,103,0}},
/* 50482 */ {(13<<2)|2,{71,103,0}},
/* 50483 */ {(13<<2)|2,{71,103,0}},
/* 50484 */ {(13<<2)|2,{71,103,0}},
/* 50485 */ {(13<<2)|2,{71,103,0}},
/* 50486 */ {(13<<2)|2,{71,103,0}},
/* 50487 */ {(13<<2)|2,{71,103,0}},
/* 50488 */ {(13<<2)|2,{71,104,0}},
/* 50489 */ {(13<<2)|2,{71,104,0}},
/* 50490 */ {(13<<2)|2,{71,104,0}},
/* 50491 */ {(13<<2)|2,{71,104,0}},
/* 50492 */ {(13<<2)|2,{71,104,0}},
/* 50493 */ {(13<<2)|2,{71,104,0}},
/* 50494 */ {(13<<2)|2,{71,104,0}},
/* 50495 */ {(13<<2)|2,{71,104,0}},
/* 50496 */ {(13<<2)|2,{71,108,0}},
/* 50497 */ {(13<<2)|2,{71,108,0}},
/* 50498 */ {(13<<2)|2,{71,108,0}},
/* 50499 */ {(13<<2)|2,{71,108,0}},
/* 50500 */ {(13<<2)|2,{71,108,0}},
/* 50501 */ {(13<<2)|2,{71,108,0}},
/* 50502 */ {(13<<2)|2,{71,108,0}},
/* 50503 */ {(13<<2)|2,{71,108,0}},
/* 50504 */ {(13<<2)|2,{71,109,0}},
/* 50505 */ {(13<<2)|2,{71,109,0}},
/* 50506 */ {(13<<2)|2,{71,109,0}},
/* 50507 */ {(13<<2)|2,{71,109,0}},
/* 50508 */ {(13<<2)|2,{71,109,0}},
/* 50509 */ {(13<<2)|2,{71,109,0}},
/* 50510 */ {(13<<2)|2,{71,109,0}},
/* 50511 */ {(13<<2)|2,{71,109,0}},
/* 50512 */ {(13<<2)|2,{71,110,0}},
/* 50513 */ {(13<<2)|2,{71,110,0}},
/* 50514 */ {(13<<2)|2,{71,110,0}},
/* 50515 */ {(13<<2)|2,{71,110,0}},
/* 50516 */ {(13<<2)|2,{71,110,0}},
/* 50517 */ {(13<<2)|2,{71,110,0}},
/* 50518 */ {(13<<2)|2,{71,110,0}},
/* 50519 */ {(13<<2)|2,{71,110,0}},
/* 50520 */ {(13<<2)|2,{71,112,0}},
/* 50521 */ {(13<<2)|2,{71,112,0}},
/* 50522 */ {(13<<2)|2,{71,112,0}},
/* 50523 */ {(13<<2)|2,{71,112,0}},
/* 50524 */ {(13<<2)|2,{71,112,0}},
/* 50525 */ {(13<<2)|2,{71,112,0}},
/* 50526 */ {(13<<2)|2,{71,112,0}},
/* 50527 */ {(13<<2)|2,{71,112,0}},
/* 50528 */ {(13<<2)|2,{71,114,0}},
/* 50529 */ {(13<<2)|2,{71,114,0}},
/* 50530 */ {(13<<2)|2,{71,114,0}},
/* 50531 */ {(13<<2)|2,{71,114,0}},
/* 50532 */ {(13<<2)|2,{71,114,0}},
/* 50533 */ {(13<<2)|2,{71,114,0}},
/* 50534 */ {(13<<2)|2,{71,114,0}},
/* 50535 */ {(13<<2)|2,{71,114,0}},
/* 50536 */ {(13<<2)|2,{71,117,0}},
/* 50537 */ {(13<<2)|2,{71,117,0}},
/* 50538 */ {(13<<2)|2,{71,117,0}},
/* 50539 */ {(13<<2)|2,{71,117,0}},
/* 50540 */ {(13<<2)|2,{71,117,0}},
/* 50541 */ {(13<<2)|2,{71,117,0}},
/* 50542 */ {(13<<2)|2,{71,117,0}},
/* 50543 */ {(13<<2)|2,{71,117,0}},
/* 50544 */ {(14<<2)|2,{71,58,0}},
/* 50545 */ {(14<<2)|2,{71,58,0}},
/* 50546 */ {(14<<2)|2,{71,58,0}},
/* 50547 */ {(14<<2)|2,{71,58,0}},
/* 50548 */ {(14<<2)|2,{71,66,0}},
/* 50549 */ {(14<<2)|2,{71,66,0}},
/* 50550 */ {(14<<2)|2,{71,66,0}},
/* 50551 */ {(14<<2)|2,{71,66,0}},
/* 50552 */ {(14<<2)|2,{71,67,0}},
/* 50553 */ {(14<<2)|2,{71,67,0}},
/* 50554 */ {(14<<2)|2,{71,67,0}},
/* 50555 */ {(14<<2)|2,{71,67,0}},
/* 50556 */ {(14<<2)|2,{71,68,0}},
/* 50557 */ {(14<<2)|2,{71,68,0}},
/* 50558 */ {(14<<2)|2,{71,68,0}},
/* 50559 */ {(14<<2)|2,{71,68,0}},
/* 50560 */ {(14<<2)|2,{71,69,0}},
/* 50561 */ {(14<<2)|2,{71,69,0}},
/* 50562 */ {(14<<2)|2,{71,69,0}},
/* 50563 */ {(14<<2)|2,{71,69,0}},
/* 50564 */ {(14<<2)|2,{71,70,0}},
/* 50565 */ {(14<<2)|2,{71,70,0}},
/* 50566 */ {(14<<2)|2,{71,70,0}},
/* 50567 */ {(14<<2)|2,{71,70,0}},
/* 50568 */ {(14<<2)|2,{71,71,0}},
/* 50569 */ {(14<<2)|2,{71,71,0}},
/* 50570 */ {(14<<2)|2,{71,71,0}},
/* 50571 */ {(14<<2)|2,{71,71,0}},
/* 50572 */ {(14<<2)|2,{71,72,0}},
/* 50573 */ {(14<<2)|2,{71,72,0}},
/* 50574 */ {(14<<2)|2,{71,72,0}},
/* 50575 */ {(14<<2)|2,{71,72,0}},
/* 50576 */ {(14<<2)|2,{71,73,0}},
/* 50577 */ {(14<<2)|2,{71,73,0}},
/* 50578 */ {(14<<2)|2,{71,73,0}},
/* 50579 */ {(14<<2)|2,{71,73,0}},
/* 50580 */ {(14<<2)|2,{71,74,0}},
/* 50581 */ {(14<<2)|2,{71,74,0}},
/* 50582 */ {(14<<2)|2,{71,74,0}},
/* 50583 */ {(14<<2)|2,{71,74,0}},
/* 50584 */ {(14<<2)|2,{71,75,0}},
/* 50585 */ {(14<<2)|2,{71,75,0}},
/* 50586 */ {(14<<2)|2,{71,75,0}},
/* 50587 */ {(14<<2)|2,{71,75,0}},
/* 50588 */ {(14<<2)|2,{71,76,0}},
/* 50589 */ {(14<<2)|2,{71,76,0}},
/* 50590 */ {(14<<2)|2,{71,76,0}},
/* 50591 */ {(14<<2)|2,{71,76,0}},
/* 50592 */ {(14<<2)|2,{71,77,0}},
/* 50593 */ {(14<<2)|2,{71,77,0}},
/* 50594 */ {(14<<2)|2,{71,77,0}},
/* 50595 */ {(14<<2)|2,{71,77,0}},
/* 50596 */ {(14<<2)|2,{71,78,0}},
/* 50597 */ {(14<<2)|2,{71,78,0}},
/* 50598 */ {(14<<2)|2,{71,78,0}},
/* 50599 */ {(14<<2)|2,{71,78,0}},
/* 50600 */ {(14<<2)|2,{71,79,0}},
/* 50601 */ {(14<<2)|2,{71,79,0}},
/* 50602 */ {(14<<2)|2,{71,79,0}},
/* 50603 */ {(14<<2)|2,{71,79,0}},
/* 50604 */ {(14<<2)|2,{71,80,0}},
/* 50605 */ {(14<<2)|2,{71,80,0}},
/* 50606 */ {(14<<2)|2,{71,80,0}},
/* 50607 */ {(14<<2)|2,{71,80,0}},
/* 50608 */ {(14<<2)|2,{71,81,0}},
/* 50609 */ {(14<<2)|2,{71,81,0}},
/* 50610 */ {(14<<2)|2,{71,81,0}},
/* 50611 */ {(14<<2)|2,{71,81,0}},
/* 50612 */ {(14<<2)|2,{71,82,0}},
/* 50613 */ {(14<<2)|2,{71,82,0}},
/* 50614 */ {(14<<2)|2,{71,82,0}},
/* 50615 */ {(14<<2)|2,{71,82,0}},
/* 50616 */ {(14<<2)|2,{71,83,0}},
/* 50617 */ {(14<<2)|2,{71,83,0}},
/* 50618 */ {(14<<2)|2,{71,83,0}},
/* 50619 */ {(14<<2)|2,{71,83,0}},
/* 50620 */ {(14<<2)|2,{71,84,0}},
/* 50621 */ {(14<<2)|2,{71,84,0}},
/* 50622 */ {(14<<2)|2,{71,84,0}},
/* 50623 */ {(14<<2)|2,{71,84,0}},
/* 50624 */ {(14<<2)|2,{71,85,0}},
/* 50625 */ {(14<<2)|2,{71,85,0}},
/* 50626 */ {(14<<2)|2,{71,85,0}},
/* 50627 */ {(14<<2)|2,{71,85,0}},
/* 50628 */ {(14<<2)|2,{71,86,0}},
/* 50629 */ {(14<<2)|2,{71,86,0}},
/* 50630 */ {(14<<2)|2,{71,86,0}},
/* 50631 */ {(14<<2)|2,{71,86,0}},
/* 50632 */ {(14<<2)|2,{71,87,0}},
/* 50633 */ {(14<<2)|2,{71,87,0}},
/* 50634 */ {(14<<2)|2,{71,87,0}},
/* 50635 */ {(14<<2)|2,{71,87,0}},
/* 50636 */ {(14<<2)|2,{71,89,0}},
/* 50637 */ {(14<<2)|2,{71,89,0}},
/* 50638 */ {(14<<2)|2,{71,89,0}},
/* 50639 */ {(14<<2)|2,{71,89,0}},
/* 50640 */ {(14<<2)|2,{71,106,0}},
/* 50641 */ {(14<<2)|2,{71,106,0}},
/* 50642 */ {(14<<2)|2,{71,106,0}},
/* 50643 */ {(14<<2)|2,{71,106,0}},
/* 50644 */ {(14<<2)|2,{71,107,0}},
/* 50645 */ {(14<<2)|2,{71,107,0}},
/* 50646 */ {(14<<2)|2,{71,107,0}},
/* 50647 */ {(14<<2)|2,{71,107,0}},
/* 50648 */ {(14<<2)|2,{71,113,0}},
/* 50649 */ {(14<<2)|2,{71,113,0}},
/* 50650 */ {(14<<2)|2,{71,113,0}},
/* 50651 */ {(14<<2)|2,{71,113,0}},
/* 50652 */ {(14<<2)|2,{71,118,0}},
/* 50653 */ {(14<<2)|2,{71,118,0}},
/* 50654 */ {(14<<2)|2,{71,118,0}},
/* 50655 */ {(14<<2)|2,{71,118,0}},
/* 50656 */ {(14<<2)|2,{71,119,0}},
/* 50657 */ {(14<<2)|2,{71,119,0}},
/* 50658 */ {(14<<2)|2,{71,119,0}},
/* 50659 */ {(14<<2)|2,{71,119,0}},
/* 50660 */ {(14<<2)|2,{71,120,0}},
/* 50661 */ {(14<<2)|2,{71,120,0}},
/* 50662 */ {(14<<2)|2,{71,120,0}},
/* 50663 */ {(14<<2)|2,{71,120,0}},
/* 50664 */ {(14<<2)|2,{71,121,0}},
/* 50665 */ {(14<<2)|2,{71,121,0}},
/* 50666 */ {(14<<2)|2,{71,121,0}},
/* 50667 */ {(14<<2)|2,{71,121,0}},
/* 50668 */ {(14<<2)|2,{71,122,0}},
/* 50669 */ {(14<<2)|2,{71,122,0}},
/* 50670 */ {(14<<2)|2,{71,122,0}},
/* 50671 */ {(14<<2)|2,{71,122,0}},
/* 50672 */ {(15<<2)|2,{71,38,0}},
/* 50673 */ {(15<<2)|2,{71,38,0}},
/* 50674 */ {(15<<2)|2,{71,42,0}},
/* 50675 */ {(15<<2)|2,{71,42,0}},
/* 50676 */ {(15<<2)|2,{71,44,0}},
/* 50677 */ {(15<<2)|2,{71,44,0}},
/* 50678 */ {(15<<2)|2,{71,59,0}},
/* 50679 */ {(15<<2)|2,{71,59,0}},
/* 50680 */ {(15<<2)|2,{71,88,0}},
/* 50681 */ {(15<<2)|2,{71,88,0}},
/* 50682 */ {(15<<2)|2,{71,90,0}},
/* 50683 */ {(15<<2)|2,{71,90,0}},
/* 50684 */ {(7<<2)|1,{71,0,0}},
/* 50685 */ {(7<<2)|1,{71,0,0}},
/* 50686 */ {(7<<2)|1,{71,0,0}},
/* 50687 */ {(7<<2)|1,{71,0,0}},
/* 50688 */ {(12<<2)|2,{72,48,0}},
/* 50689 */ {(12<<2)|2,{72,48,0}},
/* 50690 */ {(12<<2)|2,{72,48,0}},
/* 50691 */ {(12<<2)|2,{72,48,0}},
/* 50692 */ {(12<<2)|2,{72,48,0}},
/* 50693 */ {(12<<2)|2,{72,48,0}},
/* 50694 */ {(12<<2)|2,{72,48,0}},
/* 50695 */ {(12<<2)|2,{72,48,0}},
/* 50696 */ {(12<<2)|2,{72,48,0}},
/* 50697 */ {(12<<2)|2,{72,48,0}},
/* 50698 */ {(12<<2)|2,{72,48,0}},
/* 50699 */ {(12<<2)|2,{72,48,0}},
/* 50700 */ {(12<<2)|2,{72,48,0}},
/* 50701 */ {(12<<2)|2,{72,48,0}},
/* 50702 */ {(12<<2)|2,{72,48,0}},
/* 50703 */ {(12<<2)|2,{72,48,0}},
/* 50704 */ {(12<<2)|2,{72,49,0}},
/* 50705 */ {(12<<2)|2,{72,49,0}},
/* 50706 */ {(12<<2)|2,{72,49,0}},
/* 50707 */ {(12<<2)|2,{72,49,0}},
/* 50708 */ {(12<<2)|2,{72,49,0}},
/* 50709 */ {(12<<2)|2,{72,49,0}},
/* 50710 */ {(12<<2)|2,{72,49,0}},
/* 50711 */ {(12<<2)|2,{72,49,0}},
/* 50712 */ {(12<<2)|2,{72,49,0}},
/* 50713 */ {(12<<2)|2,{72,49,0}},
/* 50714 */ {(12<<2)|2,{72,49,0}},
/* 50715 */ {(12<<2)|2,{72,49,0}},
/* 50716 */ {(12<<2)|2,{72,49,0}},
/* 50717 */ {(12<<2)|2,{72,49,0}},
/* 50718 */ {(12<<2)|2,{72,49,0}},
/* 50719 */ {(12<<2)|2,{72,49,0}},
/* 50720 */ {(12<<2)|2,{72,50,0}},
/* 50721 */ {(12<<2)|2,{72,50,0}},
/* 50722 */ {(12<<2)|2,{72,50,0}},
/* 50723 */ {(12<<2)|2,{72,50,0}},
/* 50724 */ {(12<<2)|2,{72,50,0}},
/* 50725 */ {(12<<2)|2,{72,50,0}},
/* 50726 */ {(12<<2)|2,{72,50,0}},
/* 50727 */ {(12<<2)|2,{72,50,0}},
/* 50728 */ {(12<<2)|2,{72,50,0}},
/* 50729 */ {(12<<2)|2,{72,50,0}},
/* 50730 */ {(12<<2)|2,{72,50,0}},
/* 50731 */ {(12<<2)|2,{72,50,0}},
/* 50732 */ {(12<<2)|2,{72,50,0}},
/* 50733 */ {(12<<2)|2,{72,50,0}},
/* 50734 */ {(12<<2)|2,{72,50,0}},
/* 50735 */ {(12<<2)|2,{72,50,0}},
/* 50736 */ {(12<<2)|2,{72,97,0}},
/* 50737 */ {(12<<2)|2,{72,97,0}},
/* 50738 */ {(12<<2)|2,{72,97,0}},
/* 50739 */ {(12<<2)|2,{72,97,0}},
/* 50740 */ {(12<<2)|2,{72,97,0}},
/* 50741 */ {(12<<2)|2,{72,97,0}},
/* 50742 */ {(12<<2)|2,{72,97,0}},
/* 50743 */ {(12<<2)|2,{72,97,0}},
/* 50744 */ {(12<<2)|2,{72,97,0}},
/* 50745 */ {(12<<2)|2,{72,97,0}},
/* 50746 */ {(12<<2)|2,{72,97,0}},
/* 50747 */ {(12<<2)|2,{72,97,0}},
/* 50748 */ {(12<<2)|2,{72,97,0}},
/* 50749 */ {(12<<2)|2,{72,97,0}},
/* 50750 */ {(12<<2)|2,{72,97,0}},
/* 50751 */ {(12<<2)|2,{72,97,0}},
/* 50752 */ {(12<<2)|2,{72,99,0}},
/* 50753 */ {(12<<2)|2,{72,99,0}},
/* 50754 */ {(12<<2)|2,{72,99,0}},
/* 50755 */ {(12<<2)|2,{72,99,0}},
/* 50756 */ {(12<<2)|2,{72,99,0}},
/* 50757 */ {(12<<2)|2,{72,99,0}},
/* 50758 */ {(12<<2)|2,{72,99,0}},
/* 50759 */ {(12<<2)|2,{72,99,0}},
/* 50760 */ {(12<<2)|2,{72,99,0}},
/* 50761 */ {(12<<2)|2,{72,99,0}},
/* 50762 */ {(12<<2)|2,{72,99,0}},
/* 50763 */ {(12<<2)|2,{72,99,0}},
/* 50764 */ {(12<<2)|2,{72,99,0}},
/* 50765 */ {(12<<2)|2,{72,99,0}},
/* 50766 */ {(12<<2)|2,{72,99,0}},
/* 50767 */ {(12<<2)|2,{72,99,0}},
/* 50768 */ {(12<<2)|2,{72,101,0}},
/* 50769 */ {(12<<2)|2,{72,101,0}},
/* 50770 */ {(12<<2)|2,{72,101,0}},
/* 50771 */ {(12<<2)|2,{72,101,0}},
/* 50772 */ {(12<<2)|2,{72,101,0}},
/* 50773 */ {(12<<2)|2,{72,101,0}},
/* 50774 */ {(12<<2)|2,{72,101,0}},
/* 50775 */ {(12<<2)|2,{72,101,0}},
/* 50776 */ {(12<<2)|2,{72,101,0}},
/* 50777 */ {(12<<2)|2,{72,101,0}},
/* 50778 */ {(12<<2)|2,{72,101,0}},
/* 50779 */ {(12<<2)|2,{72,101,0}},
/* 50780 */ {(12<<2)|2,{72,101,0}},
/* 50781 */ {(12<<2)|2,{72,101,0}},
/* 50782 */ {(12<<2)|2,{72,101,0}},
/* 50783 */ {(12<<2)|2,{72,101,0}},
/* 50784 */ {(12<<2)|2,{72,105,0}},
/* 50785 */ {(12<<2)|2,{72,105,0}},
/* 50786 */ {(12<<2)|2,{72,105,0}},
/* 50787 */ {(12<<2)|2,{72,105,0}},
/* 50788 */ {(12<<2)|2,{72,105,0}},
/* 50789 */ {(12<<2)|2,{72,105,0}},
/* 50790 */ {(12<<2)|2,{72,105,0}},
/* 50791 */ {(12<<2)|2,{72,105,0}},
/* 50792 */ {(12<<2)|2,{72,105,0}},
/* 50793 */ {(12<<2)|2,{72,105,0}},
/* 50794 */ {(12<<2)|2,{72,105,0}},
/* 50795 */ {(12<<2)|2,{72,105,0}},
/* 50796 */ {(12<<2)|2,{72,105,0}},
/* 50797 */ {(12<<2)|2,{72,105,0}},
/* 50798 */ {(12<<2)|2,{72,105,0}},
/* 50799 */ {(12<<2)|2,{72,105,0}},
/* 50800 */ {(12<<2)|2,{72,111,0}},
/* 50801 */ {(12<<2)|2,{72,111,0}},
/* 50802 */ {(12<<2)|2,{72,111,0}},
/* 50803 */ {(12<<2)|2,{72,111,0}},
/* 50804 */ {(12<<2)|2,{72,111,0}},
/* 50805 */ {(12<<2)|2,{72,111,0}},
/* 50806 */ {(12<<2)|2,{72,111,0}},
/* 50807 */ {(12<<2)|2,{72,111,0}},
/* 50808 */ {(12<<2)|2,{72,111,0}},
/* 50809 */ {(12<<2)|2,{72,111,0}},
/* 50810 */ {(12<<2)|2,{72,111,0}},
/* 50811 */ {(12<<2)|2,{72,111,0}},
/* 50812 */ {(12<<2)|2,{72,111,0}},
/* 50813 */ {(12<<2)|2,{72,111,0}},
/* 50814 */ {(12<<2)|2,{72,111,0}},
/* 50815 */ {(12<<2)|2,{72,111,0}},
/* 50816 */ {(12<<2)|2,{72,115,0}},
/* 50817 */ {(12<<2)|2,{72,115,0}},
/* 50818 */ {(12<<2)|2,{72,115,0}},
/* 50819 */ {(12<<2)|2,{72,115,0}},
/* 50820 */ {(12<<2)|2,{72,115,0}},
/* 50821 */ {(12<<2)|2,{72,115,0}},
/* 50822 */ {(12<<2)|2,{72,115,0}},
/* 50823 */ {(12<<2)|2,{72,115,0}},
/* 50824 */ {(12<<2)|2,{72,115,0}},
/* 50825 */ {(12<<2)|2,{72,115,0}},
/* 50826 */ {(12<<2)|2,{72,115,0}},
/* 50827 */ {(12<<2)|2,{72,115,0}},
/* 50828 */ {(12<<2)|2,{72,115,0}},
/* 50829 */ {(12<<2)|2,{72,115,0}},
/* 50830 */ {(12<<2)|2,{72,115,0}},
/* 50831 */ {(12<<2)|2,{72,115,0}},
/* 50832 */ {(12<<2)|2,{72,116,0}},
/* 50833 */ {(12<<2)|2,{72,116,0}},
/* 50834 */ {(12<<2)|2,{72,116,0}},
/* 50835 */ {(12<<2)|2,{72,116,0}},
/* 50836 */ {(12<<2)|2,{72,116,0}},
/* 50837 */ {(12<<2)|2,{72,116,0}},
/* 50838 */ {(12<<2)|2,{72,116,0}},
/* 50839 */ {(12<<2)|2,{72,116,0}},
/* 50840 */ {(12<<2)|2,{72,116,0}},
/* 50841 */ {(12<<2)|2,{72,116,0}},
/* 50842 */ {(12<<2)|2,{72,116,0}},
/* 50843 */ {(12<<2)|2,{72,116,0}},
/* 50844 */ {(12<<2)|2,{72,116,0}},
/* 50845 */ {(12<<2)|2,{72,116,0}},
/* 50846 */ {(12<<2)|2,{72,116,0}},
/* 50847 */ {(12<<2)|2,{72,116,0}},
/* 50848 */ {(13<<2)|2,{72,32,0}},
/* 50849 */ {(13<<2)|2,{72,32,0}},
/* 50850 */ {(13<<2)|2,{72,32,0}},
/* 50851 */ {(13<<2)|2,{72,32,0}},
/* 50852 */ {(13<<2)|2,{72,32,0}},
/* 50853 */ {(13<<2)|2,{72,32,0}},
/* 50854 */ {(13<<2)|2,{72,32,0}},
/* 50855 */ {(13<<2)|2,{72,32,0}},
/* 50856 */ {(13<<2)|2,{72,37,0}},
/* 50857 */ {(13<<2)|2,{72,37,0}},
/* 50858 */ {(13<<2)|2,{72,37,0}},
/* 50859 */ {(13<<2)|2,{72,37,0}},
/* 50860 */ {(13<<2)|2,{72,37,0}},
/* 50861 */ {(13<<2)|2,{72,37,0}},
/* 50862 */ {(13<<2)|2,{72,37,0}},
/* 50863 */ {(13<<2)|2,{72,37,0}},
/* 50864 */ {(13<<2)|2,{72,45,0}},
/* 50865 */ {(13<<2)|2,{72,45,0}},
/* 50866 */ {(13<<2)|2,{72,45,0}},
/* 50867 */ {(13<<2)|2,{72,45,0}},
/* 50868 */ {(13<<2)|2,{72,45,0}},
/* 50869 */ {(13<<2)|2,{72,45,0}},
/* 50870 */ {(13<<2)|2,{72,45,0}},
/* 50871 */ {(13<<2)|2,{72,45,0}},
/* 50872 */ {(13<<2)|2,{72,46,0}},
/* 50873 */ {(13<<2)|2,{72,46,0}},
/* 50874 */ {(13<<2)|2,{72,46,0}},
/* 50875 */ {(13<<2)|2,{72,46,0}},
/* 50876 */ {(13<<2)|2,{72,46,0}},
/* 50877 */ {(13<<2)|2,{72,46,0}},
/* 50878 */ {(13<<2)|2,{72,46,0}},
/* 50879 */ {(13<<2)|2,{72,46,0}},
/* 50880 */ {(13<<2)|2,{72,47,0}},
/* 50881 */ {(13<<2)|2,{72,47,0}},
/* 50882 */ {(13<<2)|2,{72,47,0}},
/* 50883 */ {(13<<2)|2,{72,47,0}},
/* 50884 */ {(13<<2)|2,{72,47,0}},
/* 50885 */ {(13<<2)|2,{72,47,0}},
/* 50886 */ {(13<<2)|2,{72,47,0}},
/* 50887 */ {(13<<2)|2,{72,47,0}},
/* 50888 */ {(13<<2)|2,{72,51,0}},
/* 50889 */ {(13<<2)|2,{72,51,0}},
/* 50890 */ {(13<<2)|2,{72,51,0}},
/* 50891 */ {(13<<2)|2,{72,51,0}},
/* 50892 */ {(13<<2)|2,{72,51,0}},
/* 50893 */ {(13<<2)|2,{72,51,0}},
/* 50894 */ {(13<<2)|2,{72,51,0}},
/* 50895 */ {(13<<2)|2,{72,51,0}},
/* 50896 */ {(13<<2)|2,{72,52,0}},
/* 50897 */ {(13<<2)|2,{72,52,0}},
/* 50898 */ {(13<<2)|2,{72,52,0}},
/* 50899 */ {(13<<2)|2,{72,52,0}},
/* 50900 */ {(13<<2)|2,{72,52,0}},
/* 50901 */ {(13<<2)|2,{72,52,0}},
/* 50902 */ {(13<<2)|2,{72,52,0}},
/* 50903 */ {(13<<2)|2,{72,52,0}},
/* 50904 */ {(13<<2)|2,{72,53,0}},
/* 50905 */ {(13<<2)|2,{72,53,0}},
/* 50906 */ {(13<<2)|2,{72,53,0}},
/* 50907 */ {(13<<2)|2,{72,53,0}},
/* 50908 */ {(13<<2)|2,{72,53,0}},
/* 50909 */ {(13<<2)|2,{72,53,0}},
/* 50910 */ {(13<<2)|2,{72,53,0}},
/* 50911 */ {(13<<2)|2,{72,53,0}},
/* 50912 */ {(13<<2)|2,{72,54,0}},
/* 50913 */ {(13<<2)|2,{72,54,0}},
/* 50914 */ {(13<<2)|2,{72,54,0}},
/* 50915 */ {(13<<2)|2,{72,54,0}},
/* 50916 */ {(13<<2)|2,{72,54,0}},
/* 50917 */ {(13<<2)|2,{72,54,0}},
/* 50918 */ {(13<<2)|2,{72,54,0}},
/* 50919 */ {(13<<2)|2,{72,54,0}},
/* 50920 */ {(13<<2)|2,{72,55,0}},
/* 50921 */ {(13<<2)|2,{72,55,0}},
/* 50922 */ {(13<<2)|2,{72,55,0}},
/* 50923 */ {(13<<2)|2,{72,55,0}},
/* 50924 */ {(13<<2)|2,{72,55,0}},
/* 50925 */ {(13<<2)|2,{72,55,0}},
/* 50926 */ {(13<<2)|2,{72,55,0}},
/* 50927 */ {(13<<2)|2,{72,55,0}},
/* 50928 */ {(13<<2)|2,{72,56,0}},
/* 50929 */ {(13<<2)|2,{72,56,0}},
/* 50930 */ {(13<<2)|2,{72,56,0}},
/* 50931 */ {(13<<2)|2,{72,56,0}},
/* 50932 */ {(13<<2)|2,{72,56,0}},
/* 50933 */ {(13<<2)|2,{72,56,0}},
/* 50934 */ {(13<<2)|2,{72,56,0}},
/* 50935 */ {(13<<2)|2,{72,56,0}},
/* 50936 */ {(13<<2)|2,{72,57,0}},
/* 50937 */ {(13<<2)|2,{72,57,0}},
/* 50938 */ {(13<<2)|2,{72,57,0}},
/* 50939 */ {(13<<2)|2,{72,57,0}},
/* 50940 */ {(13<<2)|2,{72,57,0}},
/* 50941 */ {(13<<2)|2,{72,57,0}},
/* 50942 */ {(13<<2)|2,{72,57,0}},
/* 50943 */ {(13<<2)|2,{72,57,0}},
/* 50944 */ {(13<<2)|2,{72,61,0}},
/* 50945 */ {(13<<2)|2,{72,61,0}},
/* 50946 */ {(13<<2)|2,{72,61,0}},
/* 50947 */ {(13<<2)|2,{72,61,0}},
/* 50948 */ {(13<<2)|2,{72,61,0}},
/* 50949 */ {(13<<2)|2,{72,61,0}},
/* 50950 */ {(13<<2)|2,{72,61,0}},
/* 50951 */ {(13<<2)|2,{72,61,0}},
/* 50952 */ {(13<<2)|2,{72,65,0}},
/* 50953 */ {(13<<2)|2,{72,65,0}},
/* 50954 */ {(13<<2)|2,{72,65,0}},
/* 50955 */ {(13<<2)|2,{72,65,0}},
/* 50956 */ {(13<<2)|2,{72,65,0}},
/* 50957 */ {(13<<2)|2,{72,65,0}},
/* 50958 */ {(13<<2)|2,{72,65,0}},
/* 50959 */ {(13<<2)|2,{72,65,0}},
/* 50960 */ {(13<<2)|2,{72,95,0}},
/* 50961 */ {(13<<2)|2,{72,95,0}},
/* 50962 */ {(13<<2)|2,{72,95,0}},
/* 50963 */ {(13<<2)|2,{72,95,0}},
/* 50964 */ {(13<<2)|2,{72,95,0}},
/* 50965 */ {(13<<2)|2,{72,95,0}},
/* 50966 */ {(13<<2)|2,{72,95,0}},
/* 50967 */ {(13<<2)|2,{72,95,0}},
/* 50968 */ {(13<<2)|2,{72,98,0}},
/* 50969 */ {(13<<2)|2,{72,98,0}},
/* 50970 */ {(13<<2)|2,{72,98,0}},
/* 50971 */ {(13<<2)|2,{72,98,0}},
/* 50972 */ {(13<<2)|2,{72,98,0}},
/* 50973 */ {(13<<2)|2,{72,98,0}},
/* 50974 */ {(13<<2)|2,{72,98,0}},
/* 50975 */ {(13<<2)|2,{72,98,0}},
/* 50976 */ {(13<<2)|2,{72,100,0}},
/* 50977 */ {(13<<2)|2,{72,100,0}},
/* 50978 */ {(13<<2)|2,{72,100,0}},
/* 50979 */ {(13<<2)|2,{72,100,0}},
/* 50980 */ {(13<<2)|2,{72,100,0}},
/* 50981 */ {(13<<2)|2,{72,100,0}},
/* 50982 */ {(13<<2)|2,{72,100,0}},
/* 50983 */ {(13<<2)|2,{72,100,0}},
/* 50984 */ {(13<<2)|2,{72,102,0}},
/* 50985 */ {(13<<2)|2,{72,102,0}},
/* 50986 */ {(13<<2)|2,{72,102,0}},
/* 50987 */ {(13<<2)|2,{72,102,0}},
/* 50988 */ {(13<<2)|2,{72,102,0}},
/* 50989 */ {(13<<2)|2,{72,102,0}},
/* 50990 */ {(13<<2)|2,{72,102,0}},
/* 50991 */ {(13<<2)|2,{72,102,0}},
/* 50992 */ {(13<<2)|2,{72,103,0}},
/* 50993 */ {(13<<2)|2,{72,103,0}},
/* 50994 */ {(13<<2)|2,{72,103,0}},
/* 50995 */ {(13<<2)|2,{72,103,0}},
/* 50996 */ {(13<<2)|2,{72,103,0}},
/* 50997 */ {(13<<2)|2,{72,103,0}},
/* 50998 */ {(13<<2)|2,{72,103,0}},
/* 50999 */ {(13<<2)|2,{72,103,0}},
/* 51000 */ {(13<<2)|2,{72,104,0}},
/* 51001 */ {(13<<2)|2,{72,104,0}},
/* 51002 */ {(13<<2)|2,{72,104,0}},
/* 51003 */ {(13<<2)|2,{72,104,0}},
/* 51004 */ {(13<<2)|2,{72,104,0}},
/* 51005 */ {(13<<2)|2,{72,104,0}},
/* 51006 */ {(13<<2)|2,{72,104,0}},
/* 51007 */ {(13<<2)|2,{72,104,0}},
/* 51008 */ {(13<<2)|2,{72,108,0}},
/* 51009 */ {(13<<2)|2,{72,108,0}},
/* 51010 */ {(13<<2)|2,{72,108,0}},
/* 51011 */ {(13<<2)|2,{72,108,0}},
/* 51012 */ {(13<<2)|2,{72,108,0}},
/* 51013 */ {(13<<2)|2,{72,108,0}},
/* 51014 */ {(13<<2)|2,{72,108,0}},
/* 51015 */ {(13<<2)|2,{72,108,0}},
/* 51016 */ {(13<<2)|2,{72,109,0}},
/* 51017 */ {(13<<2)|2,{72,109,0}},
/* 51018 */ {(13<<2)|2,{72,109,0}},
/* 51019 */ {(13<<2)|2,{72,109,0}},
/* 51020 */ {(13<<2)|2,{72,109,0}},
/* 51021 */ {(13<<2)|2,{72,109,0}},
/* 51022 */ {(13<<2)|2,{72,109,0}},
/* 51023 */ {(13<<2)|2,{72,109,0}},
/* 51024 */ {(13<<2)|2,{72,110,0}},
/* 51025 */ {(13<<2)|2,{72,110,0}},
/* 51026 */ {(13<<2)|2,{72,110,0}},
/* 51027 */ {(13<<2)|2,{72,110,0}},
/* 51028 */ {(13<<2)|2,{72,110,0}},
/* 51029 */ {(13<<2)|2,{72,110,0}},
/* 51030 */ {(13<<2)|2,{72,110,0}},
/* 51031 */ {(13<<2)|2,{72,110,0}},
/* 51032 */ {(13<<2)|2,{72,112,0}},
/* 51033 */ {(13<<2)|2,{72,112,0}},
/* 51034 */ {(13<<2)|2,{72,112,0}},
/* 51035 */ {(13<<2)|2,{72,112,0}},
/* 51036 */ {(13<<2)|2,{72,112,0}},
/* 51037 */ {(13<<2)|2,{72,112,0}},
/* 51038 */ {(13<<2)|2,{72,112,0}},
/* 51039 */ {(13<<2)|2,{72,112,0}},
/* 51040 */ {(13<<2)|2,{72,114,0}},
/* 51041 */ {(13<<2)|2,{72,114,0}},
/* 51042 */ {(13<<2)|2,{72,114,0}},
/* 51043 */ {(13<<2)|2,{72,114,0}},
/* 51044 */ {(13<<2)|2,{72,114,0}},
/* 51045 */ {(13<<2)|2,{72,114,0}},
/* 51046 */ {(13<<2)|2,{72,114,0}},
/* 51047 */ {(13<<2)|2,{72,114,0}},
/* 51048 */ {(13<<2)|2,{72,117,0}},
/* 51049 */ {(13<<2)|2,{72,117,0}},
/* 51050 */ {(13<<2)|2,{72,117,0}},
/* 51051 */ {(13<<2)|2,{72,117,0}},
/* 51052 */ {(13<<2)|2,{72,117,0}},
/* 51053 */ {(13<<2)|2,{72,117,0}},
/* 51054 */ {(13<<2)|2,{72,117,0}},
/* 51055 */ {(13<<2)|2,{72,117,0}},
/* 51056 */ {(14<<2)|2,{72,58,0}},
/* 51057 */ {(14<<2)|2,{72,58,0}},
/* 51058 */ {(14<<2)|2,{72,58,0}},
/* 51059 */ {(14<<2)|2,{72,58,0}},
/* 51060 */ {(14<<2)|2,{72,66,0}},
/* 51061 */ {(14<<2)|2,{72,66,0}},
/* 51062 */ {(14<<2)|2,{72,66,0}},
/* 51063 */ {(14<<2)|2,{72,66,0}},
/* 51064 */ {(14<<2)|2,{72,67,0}},
/* 51065 */ {(14<<2)|2,{72,67,0}},
/* 51066 */ {(14<<2)|2,{72,67,0}},
/* 51067 */ {(14<<2)|2,{72,67,0}},
/* 51068 */ {(14<<2)|2,{72,68,0}},
/* 51069 */ {(14<<2)|2,{72,68,0}},
/* 51070 */ {(14<<2)|2,{72,68,0}},
/* 51071 */ {(14<<2)|2,{72,68,0}},
/* 51072 */ {(14<<2)|2,{72,69,0}},
/* 51073 */ {(14<<2)|2,{72,69,0}},
/* 51074 */ {(14<<2)|2,{72,69,0}},
/* 51075 */ {(14<<2)|2,{72,69,0}},
/* 51076 */ {(14<<2)|2,{72,70,0}},
/* 51077 */ {(14<<2)|2,{72,70,0}},
/* 51078 */ {(14<<2)|2,{72,70,0}},
/* 51079 */ {(14<<2)|2,{72,70,0}},
/* 51080 */ {(14<<2)|2,{72,71,0}},
/* 51081 */ {(14<<2)|2,{72,71,0}},
/* 51082 */ {(14<<2)|2,{72,71,0}},
/* 51083 */ {(14<<2)|2,{72,71,0}},
/* 51084 */ {(14<<2)|2,{72,72,0}},
/* 51085 */ {(14<<2)|2,{72,72,0}},
/* 51086 */ {(14<<2)|2,{72,72,0}},
/* 51087 */ {(14<<2)|2,{72,72,0}},
/* 51088 */ {(14<<2)|2,{72,73,0}},
/* 51089 */ {(14<<2)|2,{72,73,0}},
/* 51090 */ {(14<<2)|2,{72,73,0}},
/* 51091 */ {(14<<2)|2,{72,73,0}},
/* 51092 */ {(14<<2)|2,{72,74,0}},
/* 51093 */ {(14<<2)|2,{72,74,0}},
/* 51094 */ {(14<<2)|2,{72,74,0}},
/* 51095 */ {(14<<2)|2,{72,74,0}},
/* 51096 */ {(14<<2)|2,{72,75,0}},
/* 51097 */ {(14<<2)|2,{72,75,0}},
/* 51098 */ {(14<<2)|2,{72,75,0}},
/* 51099 */ {(14<<2)|2,{72,75,0}},
/* 51100 */ {(14<<2)|2,{72,76,0}},
/* 51101 */ {(14<<2)|2,{72,76,0}},
/* 51102 */ {(14<<2)|2,{72,76,0}},
/* 51103 */ {(14<<2)|2,{72,76,0}},
/* 51104 */ {(14<<2)|2,{72,77,0}},
/* 51105 */ {(14<<2)|2,{72,77,0}},
/* 51106 */ {(14<<2)|2,{72,77,0}},
/* 51107 */ {(14<<2)|2,{72,77,0}},
/* 51108 */ {(14<<2)|2,{72,78,0}},
/* 51109 */ {(14<<2)|2,{72,78,0}},
/* 51110 */ {(14<<2)|2,{72,78,0}},
/* 51111 */ {(14<<2)|2,{72,78,0}},
/* 51112 */ {(14<<2)|2,{72,79,0}},
/* 51113 */ {(14<<2)|2,{72,79,0}},
/* 51114 */ {(14<<2)|2,{72,79,0}},
/* 51115 */ {(14<<2)|2,{72,79,0}},
/* 51116 */ {(14<<2)|2,{72,80,0}},
/* 51117 */ {(14<<2)|2,{72,80,0}},
/* 51118 */ {(14<<2)|2,{72,80,0}},
/* 51119 */ {(14<<2)|2,{72,80,0}},
/* 51120 */ {(14<<2)|2,{72,81,0}},
/* 51121 */ {(14<<2)|2,{72,81,0}},
/* 51122 */ {(14<<2)|2,{72,81,0}},
/* 51123 */ {(14<<2)|2,{72,81,0}},
/* 51124 */ {(14<<2)|2,{72,82,0}},
/* 51125 */ {(14<<2)|2,{72,82,0}},
/* 51126 */ {(14<<2)|2,{72,82,0}},
/* 51127 */ {(14<<2)|2,{72,82,0}},
/* 51128 */ {(14<<2)|2,{72,83,0}},
/* 51129 */ {(14<<2)|2,{72,83,0}},
/* 51130 */ {(14<<2)|2,{72,83,0}},
/* 51131 */ {(14<<2)|2,{72,83,0}},
/* 51132 */ {(14<<2)|2,{72,84,0}},
/* 51133 */ {(14<<2)|2,{72,84,0}},
/* 51134 */ {(14<<2)|2,{72,84,0}},
/* 51135 */ {(14<<2)|2,{72,84,0}},
/* 51136 */ {(14<<2)|2,{72,85,0}},
/* 51137 */ {(14<<2)|2,{72,85,0}},
/* 51138 */ {(14<<2)|2,{72,85,0}},
/* 51139 */ {(14<<2)|2,{72,85,0}},
/* 51140 */ {(14<<2)|2,{72,86,0}},
/* 51141 */ {(14<<2)|2,{72,86,0}},
/* 51142 */ {(14<<2)|2,{72,86,0}},
/* 51143 */ {(14<<2)|2,{72,86,0}},
/* 51144 */ {(14<<2)|2,{72,87,0}},
/* 51145 */ {(14<<2)|2,{72,87,0}},
/* 51146 */ {(14<<2)|2,{72,87,0}},
/* 51147 */ {(14<<2)|2,{72,87,0}},
/* 51148 */ {(14<<2)|2,{72,89,0}},
/* 51149 */ {(14<<2)|2,{72,89,0}},
/* 51150 */ {(14<<2)|2,{72,89,0}},
/* 51151 */ {(14<<2)|2,{72,89,0}},
/* 51152 */ {(14<<2)|2,{72,106,0}},
/* 51153 */ {(14<<2)|2,{72,106,0}},
/* 51154 */ {(14<<2)|2,{72,106,0}},
/* 51155 */ {(14<<2)|2,{72,106,0}},
/* 51156 */ {(14<<2)|2,{72,107,0}},
/* 51157 */ {(14<<2)|2,{72,107,0}},
/* 51158 */ {(14<<2)|2,{72,107,0}},
/* 51159 */ {(14<<2)|2,{72,107,0}},
/* 51160 */ {(14<<2)|2,{72,113,0}},
/* 51161 */ {(14<<2)|2,{72,113,0}},
/* 51162 */ {(14<<2)|2,{72,113,0}},
/* 51163 */ {(14<<2)|2,{72,113,0}},
/* 51164 */ {(14<<2)|2,{72,118,0}},
/* 51165 */ {(14<<2)|2,{72,118,0}},
/* 51166 */ {(14<<2)|2,{72,118,0}},
/* 51167 */ {(14<<2)|2,{72,118,0}},
/* 51168 */ {(14<<2)|2,{72,119,0}},
/* 51169 */ {(14<<2)|2,{72,119,0}},
/* 51170 */ {(14<<2)|2,{72,119,0}},
/* 51171 */ {(14<<2)|2,{72,119,0}},
/* 51172 */ {(14<<2)|2,{72,120,0}},
/* 51173 */ {(14<<2)|2,{72,120,0}},
/* 51174 */ {(14<<2)|2,{72,120,0}},
/* 51175 */ {(14<<2)|2,{72,120,0}},
/* 51176 */ {(14<<2)|2,{72,121,0}},
/* 51177 */ {(14<<2)|2,{72,121,0}},
/* 51178 */ {(14<<2)|2,{72,121,0}},
/* 51179 */ {(14<<2)|2,{72,121,0}},
/* 51180 */ {(14<<2)|2,{72,122,0}},
/* 51181 */ {(14<<2)|2,{72,122,0}},
/* 51182 */ {(14<<2)|2,{72,122,0}},
/* 51183 */ {(14<<2)|2,{72,122,0}},
/* 51184 */ {(15<<2)|2,{72,38,0}},
/* 51185 */ {(15<<2)|2,{72,38,0}},
/* 51186 */ {(15<<2)|2,{72,42,0}},
/* 51187 */ {(15<<2)|2,{72,42,0}},
/* 51188 */ {(15<<2)|2,{72,44,0}},
/* 51189 */ {(15<<2)|2,{72,44,0}},
/* 51190 */ {(15<<2)|2,{72,59,0}},
/* 51191 */ {(15<<2)|2,{72,59,0}},
/* 51192 */ {(15<<2)|2,{72,88,0}},
/* 51193 */ {(15<<2)|2,{72,88,0}},
/* 51194 */ {(15<<2)|2,{72,90,0}},
/* 51195 */ {(15<<2)|2,{72,90,0}},
/* 51196 */ {(7<<2)|1,{72,0,0}},
/* 51197 */ {(7<<2)|1,{72,0,0}},
/* 51198 */ {(7<<2)|1,{72,0,0}},
/* 51199 */ {(7<<2)|1,{72,0,0}},
/* 51200 */ {(12<<2)|2,{73,48,0}},
/* 51201 */ {(12<<2)|2,{73,48,0}},
/* 51202 */ {(12<<2)|2,{73,48,0}},
/* 51203 */ {(12<<2)|2,{73,48,0}},
/* 51204 */ {(12<<2)|2,{73,48,0}},
/* 51205 */ {(12<<2)|2,{73,48,0}},
/* 51206 */ {(12<<2)|2,{73,48,0}},
/* 51207 */ {(12<<2)|2,{73,48,0}},
/* 51208 */ {(12<<2)|2,{73,48,0}},
/* 51209 */ {(12<<2)|2,{73,48,0}},
/* 51210 */ {(12<<2)|2,{73,48,0}},
/* 51211 */ {(12<<2)|2,{73,48,0}},
/* 51212 */ {(12<<2)|2,{73,48,0}},
/* 51213 */ {(12<<2)|2,{73,48,0}},
/* 51214 */ {(12<<2)|2,{73,48,0}},
/* 51215 */ {(12<<2)|2,{73,48,0}},
/* 51216 */ {(12<<2)|2,{73,49,0}},
/* 51217 */ {(12<<2)|2,{73,49,0}},
/* 51218 */ {(12<<2)|2,{73,49,0}},
/* 51219 */ {(12<<2)|2,{73,49,0}},
/* 51220 */ {(12<<2)|2,{73,49,0}},
/* 51221 */ {(12<<2)|2,{73,49,0}},
/* 51222 */ {(12<<2)|2,{73,49,0}},
/* 51223 */ {(12<<2)|2,{73,49,0}},
/* 51224 */ {(12<<2)|2,{73,49,0}},
/* 51225 */ {(12<<2)|2,{73,49,0}},
/* 51226 */ {(12<<2)|2,{73,49,0}},
/* 51227 */ {(12<<2)|2,{73,49,0}},
/* 51228 */ {(12<<2)|2,{73,49,0}},
/* 51229 */ {(12<<2)|2,{73,49,0}},
/* 51230 */ {(12<<2)|2,{73,49,0}},
/* 51231 */ {(12<<2)|2,{73,49,0}},
/* 51232 */ {(12<<2)|2,{73,50,0}},
/* 51233 */ {(12<<2)|2,{73,50,0}},
/* 51234 */ {(12<<2)|2,{73,50,0}},
/* 51235 */ {(12<<2)|2,{73,50,0}},
/* 51236 */ {(12<<2)|2,{73,50,0}},
/* 51237 */ {(12<<2)|2,{73,50,0}},
/* 51238 */ {(12<<2)|2,{73,50,0}},
/* 51239 */ {(12<<2)|2,{73,50,0}},
/* 51240 */ {(12<<2)|2,{73,50,0}},
/* 51241 */ {(12<<2)|2,{73,50,0}},
/* 51242 */ {(12<<2)|2,{73,50,0}},
/* 51243 */ {(12<<2)|2,{73,50,0}},
/* 51244 */ {(12<<2)|2,{73,50,0}},
/* 51245 */ {(12<<2)|2,{73,50,0}},
/* 51246 */ {(12<<2)|2,{73,50,0}},
/* 51247 */ {(12<<2)|2,{73,50,0}},
/* 51248 */ {(12<<2)|2,{73,97,0}},
/* 51249 */ {(12<<2)|2,{73,97,0}},
/* 51250 */ {(12<<2)|2,{73,97,0}},
/* 51251 */ {(12<<2)|2,{73,97,0}},
/* 51252 */ {(12<<2)|2,{73,97,0}},
/* 51253 */ {(12<<2)|2,{73,97,0}},
/* 51254 */ {(12<<2)|2,{73,97,0}},
/* 51255 */ {(12<<2)|2,{73,97,0}},
/* 51256 */ {(12<<2)|2,{73,97,0}},
/* 51257 */ {(12<<2)|2,{73,97,0}},
/* 51258 */ {(12<<2)|2,{73,97,0}},
/* 51259 */ {(12<<2)|2,{73,97,0}},
/* 51260 */ {(12<<2)|2,{73,97,0}},
/* 51261 */ {(12<<2)|2,{73,97,0}},
/* 51262 */ {(12<<2)|2,{73,97,0}},
/* 51263 */ {(12<<2)|2,{73,97,0}},
/* 51264 */ {(12<<2)|2,{73,99,0}},
/* 51265 */ {(12<<2)|2,{73,99,0}},
/* 51266 */ {(12<<2)|2,{73,99,0}},
/* 51267 */ {(12<<2)|2,{73,99,0}},
/* 51268 */ {(12<<2)|2,{73,99,0}},
/* 51269 */ {(12<<2)|2,{73,99,0}},
/* 51270 */ {(12<<2)|2,{73,99,0}},
/* 51271 */ {(12<<2)|2,{73,99,0}},
/* 51272 */ {(12<<2)|2,{73,99,0}},
/* 51273 */ {(12<<2)|2,{73,99,0}},
/* 51274 */ {(12<<2)|2,{73,99,0}},
/* 51275 */ {(12<<2)|2,{73,99,0}},
/* 51276 */ {(12<<2)|2,{73,99,0}},
/* 51277 */ {(12<<2)|2,{73,99,0}},
/* 51278 */ {(12<<2)|2,{73,99,0}},
/* 51279 */ {(12<<2)|2,{73,99,0}},
/* 51280 */ {(12<<2)|2,{73,101,0}},
/* 51281 */ {(12<<2)|2,{73,101,0}},
/* 51282 */ {(12<<2)|2,{73,101,0}},
/* 51283 */ {(12<<2)|2,{73,101,0}},
/* 51284 */ {(12<<2)|2,{73,101,0}},
/* 51285 */ {(12<<2)|2,{73,101,0}},
/* 51286 */ {(12<<2)|2,{73,101,0}},
/* 51287 */ {(12<<2)|2,{73,101,0}},
/* 51288 */ {(12<<2)|2,{73,101,0}},
/* 51289 */ {(12<<2)|2,{73,101,0}},
/* 51290 */ {(12<<2)|2,{73,101,0}},
/* 51291 */ {(12<<2)|2,{73,101,0}},
/* 51292 */ {(12<<2)|2,{73,101,0}},
/* 51293 */ {(12<<2)|2,{73,101,0}},
/* 51294 */ {(12<<2)|2,{73,101,0}},
/* 51295 */ {(12<<2)|2,{73,101,0}},
/* 51296 */ {(12<<2)|2,{73,105,0}},
/* 51297 */ {(12<<2)|2,{73,105,0}},
/* 51298 */ {(12<<2)|2,{73,105,0}},
/* 51299 */ {(12<<2)|2,{73,105,0}},
/* 51300 */ {(12<<2)|2,{73,105,0}},
/* 51301 */ {(12<<2)|2,{73,105,0}},
/* 51302 */ {(12<<2)|2,{73,105,0}},
/* 51303 */ {(12<<2)|2,{73,105,0}},
/* 51304 */ {(12<<2)|2,{73,105,0}},
/* 51305 */ {(12<<2)|2,{73,105,0}},
/* 51306 */ {(12<<2)|2,{73,105,0}},
/* 51307 */ {(12<<2)|2,{73,105,0}},
/* 51308 */ {(12<<2)|2,{73,105,0}},
/* 51309 */ {(12<<2)|2,{73,105,0}},
/* 51310 */ {(12<<2)|2,{73,105,0}},
/* 51311 */ {(12<<2)|2,{73,105,0}},
/* 51312 */ {(12<<2)|2,{73,111,0}},
/* 51313 */ {(12<<2)|2,{73,111,0}},
/* 51314 */ {(12<<2)|2,{73,111,0}},
/* 51315 */ {(12<<2)|2,{73,111,0}},
/* 51316 */ {(12<<2)|2,{73,111,0}},
/* 51317 */ {(12<<2)|2,{73,111,0}},
/* 51318 */ {(12<<2)|2,{73,111,0}},
/* 51319 */ {(12<<2)|2,{73,111,0}},
/* 51320 */ {(12<<2)|2,{73,111,0}},
/* 51321 */ {(12<<2)|2,{73,111,0}},
/* 51322 */ {(12<<2)|2,{73,111,0}},
/* 51323 */ {(12<<2)|2,{73,111,0}},
/* 51324 */ {(12<<2)|2,{73,111,0}},
/* 51325 */ {(12<<2)|2,{73,111,0}},
/* 51326 */ {(12<<2)|2,{73,111,0}},
/* 51327 */ {(12<<2)|2,{73,111,0}},
/* 51328 */ {(12<<2)|2,{73,115,0}},
/* 51329 */ {(12<<2)|2,{73,115,0}},
/* 51330 */ {(12<<2)|2,{73,115,0}},
/* 51331 */ {(12<<2)|2,{73,115,0}},
/* 51332 */ {(12<<2)|2,{73,115,0}},
/* 51333 */ {(12<<2)|2,{73,115,0}},
/* 51334 */ {(12<<2)|2,{73,115,0}},
/* 51335 */ {(12<<2)|2,{73,115,0}},
/* 51336 */ {(12<<2)|2,{73,115,0}},
/* 51337 */ {(12<<2)|2,{73,115,0}},
/* 51338 */ {(12<<2)|2,{73,115,0}},
/* 51339 */ {(12<<2)|2,{73,115,0}},
/* 51340 */ {(12<<2)|2,{73,115,0}},
/* 51341 */ {(12<<2)|2,{73,115,0}},
/* 51342 */ {(12<<2)|2,{73,115,0}},
/* 51343 */ {(12<<2)|2,{73,115,0}},
/* 51344 */ {(12<<2)|2,{73,116,0}},
/* 51345 */ {(12<<2)|2,{73,116,0}},
/* 51346 */ {(12<<2)|2,{73,116,0}},
/* 51347 */ {(12<<2)|2,{73,116,0}},
/* 51348 */ {(12<<2)|2,{73,116,0}},
/* 51349 */ {(12<<2)|2,{73,116,0}},
/* 51350 */ {(12<<2)|2,{73,116,0}},
/* 51351 */ {(12<<2)|2,{73,116,0}},
/* 51352 */ {(12<<2)|2,{73,116,0}},
/* 51353 */ {(12<<2)|2,{73,116,0}},
/* 51354 */ {(12<<2)|2,{73,116,0}},
/* 51355 */ {(12<<2)|2,{73,116,0}},
/* 51356 */ {(12<<2)|2,{73,116,0}},
/* 51357 */ {(12<<2)|2,{73,116,0}},
/* 51358 */ {(12<<2)|2,{73,116,0}},
/* 51359 */ {(12<<2)|2,{73,116,0}},
/* 51360 */ {(13<<2)|2,{73,32,0}},
/* 51361 */ {(13<<2)|2,{73,32,0}},
/* 51362 */ {(13<<2)|2,{73,32,0}},
/* 51363 */ {(13<<2)|2,{73,32,0}},
/* 51364 */ {(13<<2)|2,{73,32,0}},
/* 51365 */ {(13<<2)|2,{73,32,0}},
/* 51366 */ {(13<<2)|2,{73,32,0}},
/* 51367 */ {(13<<2)|2,{73,32,0}},
/* 51368 */ {(13<<2)|2,{73,37,0}},
/* 51369 */ {(13<<2)|2,{73,37,0}},
/* 51370 */ {(13<<2)|2,{73,37,0}},
/* 51371 */ {(13<<2)|2,{73,37,0}},
/* 51372 */ {(13<<2)|2,{73,37,0}},
/* 51373 */ {(13<<2)|2,{73,37,0}},
/* 51374 */ {(13<<2)|2,{73,37,0}},
/* 51375 */ {(13<<2)|2,{73,37,0}},
/* 51376 */ {(13<<2)|2,{73,45,0}},
/* 51377 */ {(13<<2)|2,{73,45,0}},
/* 51378 */ {(13<<2)|2,{73,45,0}},
/* 51379 */ {(13<<2)|2,{73,45,0}},
/* 51380 */ {(13<<2)|2,{73,45,0}},
/* 51381 */ {(13<<2)|2,{73,45,0}},
/* 51382 */ {(13<<2)|2,{73,45,0}},
/* 51383 */ {(13<<2)|2,{73,45,0}},
/* 51384 */ {(13<<2)|2,{73,46,0}},
/* 51385 */ {(13<<2)|2,{73,46,0}},
/* 51386 */ {(13<<2)|2,{73,46,0}},
/* 51387 */ {(13<<2)|2,{73,46,0}},
/* 51388 */ {(13<<2)|2,{73,46,0}},
/* 51389 */ {(13<<2)|2,{73,46,0}},
/* 51390 */ {(13<<2)|2,{73,46,0}},
/* 51391 */ {(13<<2)|2,{73,46,0}},
/* 51392 */ {(13<<2)|2,{73,47,0}},
/* 51393 */ {(13<<2)|2,{73,47,0}},
/* 51394 */ {(13<<2)|2,{73,47,0}},
/* 51395 */ {(13<<2)|2,{73,47,0}},
/* 51396 */ {(13<<2)|2,{73,47,0}},
/* 51397 */ {(13<<2)|2,{73,47,0}},
/* 51398 */ {(13<<2)|2,{73,47,0}},
/* 51399 */ {(13<<2)|2,{73,47,0}},
/* 51400 */ {(13<<2)|2,{73,51,0}},
/* 51401 */ {(13<<2)|2,{73,51,0}},
/* 51402 */ {(13<<2)|2,{73,51,0}},
/* 51403 */ {(13<<2)|2,{73,51,0}},
/* 51404 */ {(13<<2)|2,{73,51,0}},
/* 51405 */ {(13<<2)|2,{73,51,0}},
/* 51406 */ {(13<<2)|2,{73,51,0}},
/* 51407 */ {(13<<2)|2,{73,51,0}},
/* 51408 */ {(13<<2)|2,{73,52,0}},
/* 51409 */ {(13<<2)|2,{73,52,0}},
/* 51410 */ {(13<<2)|2,{73,52,0}},
/* 51411 */ {(13<<2)|2,{73,52,0}},
/* 51412 */ {(13<<2)|2,{73,52,0}},
/* 51413 */ {(13<<2)|2,{73,52,0}},
/* 51414 */ {(13<<2)|2,{73,52,0}},
/* 51415 */ {(13<<2)|2,{73,52,0}},
/* 51416 */ {(13<<2)|2,{73,53,0}},
/* 51417 */ {(13<<2)|2,{73,53,0}},
/* 51418 */ {(13<<2)|2,{73,53,0}},
/* 51419 */ {(13<<2)|2,{73,53,0}},
/* 51420 */ {(13<<2)|2,{73,53,0}},
/* 51421 */ {(13<<2)|2,{73,53,0}},
/* 51422 */ {(13<<2)|2,{73,53,0}},
/* 51423 */ {(13<<2)|2,{73,53,0}},
/* 51424 */ {(13<<2)|2,{73,54,0}},
/* 51425 */ {(13<<2)|2,{73,54,0}},
/* 51426 */ {(13<<2)|2,{73,54,0}},
/* 51427 */ {(13<<2)|2,{73,54,0}},
/* 51428 */ {(13<<2)|2,{73,54,0}},
/* 51429 */ {(13<<2)|2,{73,54,0}},
/* 51430 */ {(13<<2)|2,{73,54,0}},
/* 51431 */ {(13<<2)|2,{73,54,0}},
/* 51432 */ {(13<<2)|2,{73,55,0}},
/* 51433 */ {(13<<2)|2,{73,55,0}},
/* 51434 */ {(13<<2)|2,{73,55,0}},
/* 51435 */ {(13<<2)|2,{73,55,0}},
/* 51436 */ {(13<<2)|2,{73,55,0}},
/* 51437 */ {(13<<2)|2,{73,55,0}},
/* 51438 */ {(13<<2)|2,{73,55,0}},
/* 51439 */ {(13<<2)|2,{73,55,0}},
/* 51440 */ {(13<<2)|2,{73,56,0}},
/* 51441 */ {(13<<2)|2,{73,56,0}},
/* 51442 */ {(13<<2)|2,{73,56,0}},
/* 51443 */ {(13<<2)|2,{73,56,0}},
/* 51444 */ {(13<<2)|2,{73,56,0}},
/* 51445 */ {(13<<2)|2,{73,56,0}},
/* 51446 */ {(13<<2)|2,{73,56,0}},
/* 51447 */ {(13<<2)|2,{73,56,0}},
/* 51448 */ {(13<<2)|2,{73,57,0}},
/* 51449 */ {(13<<2)|2,{73,57,0}},
/* 51450 */ {(13<<2)|2,{73,57,0}},
/* 51451 */ {(13<<2)|2,{73,57,0}},
/* 51452 */ {(13<<2)|2,{73,57,0}},
/* 51453 */ {(13<<2)|2,{73,57,0}},
/* 51454 */ {(13<<2)|2,{73,57,0}},
/* 51455 */ {(13<<2)|2,{73,57,0}},
/* 51456 */ {(13<<2)|2,{73,61,0}},
/* 51457 */ {(13<<2)|2,{73,61,0}},
/* 51458 */ {(13<<2)|2,{73,61,0}},
/* 51459 */ {(13<<2)|2,{73,61,0}},
/* 51460 */ {(13<<2)|2,{73,61,0}},
/* 51461 */ {(13<<2)|2,{73,61,0}},
/* 51462 */ {(13<<2)|2,{73,61,0}},
/* 51463 */ {(13<<2)|2,{73,61,0}},
/* 51464 */ {(13<<2)|2,{73,65,0}},
/* 51465 */ {(13<<2)|2,{73,65,0}},
/* 51466 */ {(13<<2)|2,{73,65,0}},
/* 51467 */ {(13<<2)|2,{73,65,0}},
/* 51468 */ {(13<<2)|2,{73,65,0}},
/* 51469 */ {(13<<2)|2,{73,65,0}},
/* 51470 */ {(13<<2)|2,{73,65,0}},
/* 51471 */ {(13<<2)|2,{73,65,0}},
/* 51472 */ {(13<<2)|2,{73,95,0}},
/* 51473 */ {(13<<2)|2,{73,95,0}},
/* 51474 */ {(13<<2)|2,{73,95,0}},
/* 51475 */ {(13<<2)|2,{73,95,0}},
/* 51476 */ {(13<<2)|2,{73,95,0}},
/* 51477 */ {(13<<2)|2,{73,95,0}},
/* 51478 */ {(13<<2)|2,{73,95,0}},
/* 51479 */ {(13<<2)|2,{73,95,0}},
/* 51480 */ {(13<<2)|2,{73,98,0}},
/* 51481 */ {(13<<2)|2,{73,98,0}},
/* 51482 */ {(13<<2)|2,{73,98,0}},
/* 51483 */ {(13<<2)|2,{73,98,0}},
/* 51484 */ {(13<<2)|2,{73,98,0}},
/* 51485 */ {(13<<2)|2,{73,98,0}},
/* 51486 */ {(13<<2)|2,{73,98,0}},
/* 51487 */ {(13<<2)|2,{73,98,0}},
/* 51488 */ {(13<<2)|2,{73,100,0}},
/* 51489 */ {(13<<2)|2,{73,100,0}},
/* 51490 */ {(13<<2)|2,{73,100,0}},
/* 51491 */ {(13<<2)|2,{73,100,0}},
/* 51492 */ {(13<<2)|2,{73,100,0}},
/* 51493 */ {(13<<2)|2,{73,100,0}},
/* 51494 */ {(13<<2)|2,{73,100,0}},
/* 51495 */ {(13<<2)|2,{73,100,0}},
/* 51496 */ {(13<<2)|2,{73,102,0}},
/* 51497 */ {(13<<2)|2,{73,102,0}},
/* 51498 */ {(13<<2)|2,{73,102,0}},
/* 51499 */ {(13<<2)|2,{73,102,0}},
/* 51500 */ {(13<<2)|2,{73,102,0}},
/* 51501 */ {(13<<2)|2,{73,102,0}},
/* 51502 */ {(13<<2)|2,{73,102,0}},
/* 51503 */ {(13<<2)|2,{73,102,0}},
/* 51504 */ {(13<<2)|2,{73,103,0}},
/* 51505 */ {(13<<2)|2,{73,103,0}},
/* 51506 */ {(13<<2)|2,{73,103,0}},
/* 51507 */ {(13<<2)|2,{73,103,0}},
/* 51508 */ {(13<<2)|2,{73,103,0}},
/* 51509 */ {(13<<2)|2,{73,103,0}},
/* 51510 */ {(13<<2)|2,{73,103,0}},
/* 51511 */ {(13<<2)|2,{73,103,0}},
/* 51512 */ {(13<<2)|2,{73,104,0}},
/* 51513 */ {(13<<2)|2,{73,104,0}},
/* 51514 */ {(13<<2)|2,{73,104,0}},
/* 51515 */ {(13<<2)|2,{73,104,0}},
/* 51516 */ {(13<<2)|2,{73,104,0}},
/* 51517 */ {(13<<2)|2,{73,104,0}},
/* 51518 */ {(13<<2)|2,{73,104,0}},
/* 51519 */ {(13<<2)|2,{73,104,0}},
/* 51520 */ {(13<<2)|2,{73,108,0}},
/* 51521 */ {(13<<2)|2,{73,108,0}},
/* 51522 */ {(13<<2)|2,{73,108,0}},
/* 51523 */ {(13<<2)|2,{73,108,0}},
/* 51524 */ {(13<<2)|2,{73,108,0}},
/* 51525 */ {(13<<2)|2,{73,108,0}},
/* 51526 */ {(13<<2)|2,{73,108,0}},
/* 51527 */ {(13<<2)|2,{73,108,0}},
/* 51528 */ {(13<<2)|2,{73,109,0}},
/* 51529 */ {(13<<2)|2,{73,109,0}},
/* 51530 */ {(13<<2)|2,{73,109,0}},
/* 51531 */ {(13<<2)|2,{73,109,0}},
/* 51532 */ {(13<<2)|2,{73,109,0}},
/* 51533 */ {(13<<2)|2,{73,109,0}},
/* 51534 */ {(13<<2)|2,{73,109,0}},
/* 51535 */ {(13<<2)|2,{73,109,0}},
/* 51536 */ {(13<<2)|2,{73,110,0}},
/* 51537 */ {(13<<2)|2,{73,110,0}},
/* 51538 */ {(13<<2)|2,{73,110,0}},
/* 51539 */ {(13<<2)|2,{73,110,0}},
/* 51540 */ {(13<<2)|2,{73,110,0}},
/* 51541 */ {(13<<2)|2,{73,110,0}},
/* 51542 */ {(13<<2)|2,{73,110,0}},
/* 51543 */ {(13<<2)|2,{73,110,0}},
/* 51544 */ {(13<<2)|2,{73,112,0}},
/* 51545 */ {(13<<2)|2,{73,112,0}},
/* 51546 */ {(13<<2)|2,{73,112,0}},
/* 51547 */ {(13<<2)|2,{73,112,0}},
/* 51548 */ {(13<<2)|2,{73,112,0}},
/* 51549 */ {(13<<2)|2,{73,112,0}},
/* 51550 */ {(13<<2)|2,{73,112,0}},
/* 51551 */ {(13<<2)|2,{73,112,0}},
/* 51552 */ {(13<<2)|2,{73,114,0}},
/* 51553 */ {(13<<2)|2,{73,114,0}},
/* 51554 */ {(13<<2)|2,{73,114,0}},
/* 51555 */ {(13<<2)|2,{73,114,0}},
/* 51556 */ {(13<<2)|2,{73,114,0}},
/* 51557 */ {(13<<2)|2,{73,114,0}},
/* 51558 */ {(13<<2)|2,{73,114,0}},
/* 51559 */ {(13<<2)|2,{73,114,0}},
/* 51560 */ {(13<<2)|2,{73,117,0}},
/* 51561 */ {(13<<2)|2,{73,117,0}},
/* 51562 */ {(13<<2)|2,{73,117,0}},
/* 51563 */ {(13<<2)|2,{73,117,0}},
/* 51564 */ {(13<<2)|2,{73,117,0}},
/* 51565 */ {(13<<2)|2,{73,117,0}},
/* 51566 */ {(13<<2)|2,{73,117,0}},
/* 51567 */ {(13<<2)|2,{73,117,0}},
/* 51568 */ {(14<<2)|2,{73,58,0}},
/* 51569 */ {(14<<2)|2,{73,58,0}},
/* 51570 */ {(14<<2)|2,{73,58,0}},
/* 51571 */ {(14<<2)|2,{73,58,0}},
/* 51572 */ {(14<<2)|2,{73,66,0}},
/* 51573 */ {(14<<2)|2,{73,66,0}},
/* 51574 */ {(14<<2)|2,{73,66,0}},
/* 51575 */ {(14<<2)|2,{73,66,0}},
/* 51576 */ {(14<<2)|2,{73,67,0}},
/* 51577 */ {(14<<2)|2,{73,67,0}},
/* 51578 */ {(14<<2)|2,{73,67,0}},
/* 51579 */ {(14<<2)|2,{73,67,0}},
/* 51580 */ {(14<<2)|2,{73,68,0}},
/* 51581 */ {(14<<2)|2,{73,68,0}},
/* 51582 */ {(14<<2)|2,{73,68,0}},
/* 51583 */ {(14<<2)|2,{73,68,0}},
/* 51584 */ {(14<<2)|2,{73,69,0}},
/* 51585 */ {(14<<2)|2,{73,69,0}},
/* 51586 */ {(14<<2)|2,{73,69,0}},
/* 51587 */ {(14<<2)|2,{73,69,0}},
/* 51588 */ {(14<<2)|2,{73,70,0}},
/* 51589 */ {(14<<2)|2,{73,70,0}},
/* 51590 */ {(14<<2)|2,{73,70,0}},
/* 51591 */ {(14<<2)|2,{73,70,0}},
/* 51592 */ {(14<<2)|2,{73,71,0}},
/* 51593 */ {(14<<2)|2,{73,71,0}},
/* 51594 */ {(14<<2)|2,{73,71,0}},
/* 51595 */ {(14<<2)|2,{73,71,0}},
/* 51596 */ {(14<<2)|2,{73,72,0}},
/* 51597 */ {(14<<2)|2,{73,72,0}},
/* 51598 */ {(14<<2)|2,{73,72,0}},
/* 51599 */ {(14<<2)|2,{73,72,0}},
/* 51600 */ {(14<<2)|2,{73,73,0}},
/* 51601 */ {(14<<2)|2,{73,73,0}},
/* 51602 */ {(14<<2)|2,{73,73,0}},
/* 51603 */ {(14<<2)|2,{73,73,0}},
/* 51604 */ {(14<<2)|2,{73,74,0}},
/* 51605 */ {(14<<2)|2,{73,74,0}},
/* 51606 */ {(14<<2)|2,{73,74,0}},
/* 51607 */ {(14<<2)|2,{73,74,0}},
/* 51608 */ {(14<<2)|2,{73,75,0}},
/* 51609 */ {(14<<2)|2,{73,75,0}},
/* 51610 */ {(14<<2)|2,{73,75,0}},
/* 51611 */ {(14<<2)|2,{73,75,0}},
/* 51612 */ {(14<<2)|2,{73,76,0}},
/* 51613 */ {(14<<2)|2,{73,76,0}},
/* 51614 */ {(14<<2)|2,{73,76,0}},
/* 51615 */ {(14<<2)|2,{73,76,0}},
/* 51616 */ {(14<<2)|2,{73,77,0}},
/* 51617 */ {(14<<2)|2,{73,77,0}},
/* 51618 */ {(14<<2)|2,{73,77,0}},
/* 51619 */ {(14<<2)|2,{73,77,0}},
/* 51620 */ {(14<<2)|2,{73,78,0}},
/* 51621 */ {(14<<2)|2,{73,78,0}},
/* 51622 */ {(14<<2)|2,{73,78,0}},
/* 51623 */ {(14<<2)|2,{73,78,0}},
/* 51624 */ {(14<<2)|2,{73,79,0}},
/* 51625 */ {(14<<2)|2,{73,79,0}},
/* 51626 */ {(14<<2)|2,{73,79,0}},
/* 51627 */ {(14<<2)|2,{73,79,0}},
/* 51628 */ {(14<<2)|2,{73,80,0}},
/* 51629 */ {(14<<2)|2,{73,80,0}},
/* 51630 */ {(14<<2)|2,{73,80,0}},
/* 51631 */ {(14<<2)|2,{73,80,0}},
/* 51632 */ {(14<<2)|2,{73,81,0}},
/* 51633 */ {(14<<2)|2,{73,81,0}},
/* 51634 */ {(14<<2)|2,{73,81,0}},
/* 51635 */ {(14<<2)|2,{73,81,0}},
/* 51636 */ {(14<<2)|2,{73,82,0}},
/* 51637 */ {(14<<2)|2,{73,82,0}},
/* 51638 */ {(14<<2)|2,{73,82,0}},
/* 51639 */ {(14<<2)|2,{73,82,0}},
/* 51640 */ {(14<<2)|2,{73,83,0}},
/* 51641 */ {(14<<2)|2,{73,83,0}},
/* 51642 */ {(14<<2)|2,{73,83,0}},
/* 51643 */ {(14<<2)|2,{73,83,0}},
/* 51644 */ {(14<<2)|2,{73,84,0}},
/* 51645 */ {(14<<2)|2,{73,84,0}},
/* 51646 */ {(14<<2)|2,{73,84,0}},
/* 51647 */ {(14<<2)|2,{73,84,0}},
/* 51648 */ {(14<<2)|2,{73,85,0}},
/* 51649 */ {(14<<2)|2,{73,85,0}},
/* 51650 */ {(14<<2)|2,{73,85,0}},
/* 51651 */ {(14<<2)|2,{73,85,0}},
/* 51652 */ {(14<<2)|2,{73,86,0}},
/* 51653 */ {(14<<2)|2,{73,86,0}},
/* 51654 */ {(14<<2)|2,{73,86,0}},
/* 51655 */ {(14<<2)|2,{73,86,0}},
/* 51656 */ {(14<<2)|2,{73,87,0}},
/* 51657 */ {(14<<2)|2,{73,87,0}},
/* 51658 */ {(14<<2)|2,{73,87,0}},
/* 51659 */ {(14<<2)|2,{73,87,0}},
/* 51660 */ {(14<<2)|2,{73,89,0}},
/* 51661 */ {(14<<2)|2,{73,89,0}},
/* 51662 */ {(14<<2)|2,{73,89,0}},
/* 51663 */ {(14<<2)|2,{73,89,0}},
/* 51664 */ {(14<<2)|2,{73,106,0}},
/* 51665 */ {(14<<2)|2,{73,106,0}},
/* 51666 */ {(14<<2)|2,{73,106,0}},
/* 51667 */ {(14<<2)|2,{73,106,0}},
/* 51668 */ {(14<<2)|2,{73,107,0}},
/* 51669 */ {(14<<2)|2,{73,107,0}},
/* 51670 */ {(14<<2)|2,{73,107,0}},
/* 51671 */ {(14<<2)|2,{73,107,0}},
/* 51672 */ {(14<<2)|2,{73,113,0}},
/* 51673 */ {(14<<2)|2,{73,113,0}},
/* 51674 */ {(14<<2)|2,{73,113,0}},
/* 51675 */ {(14<<2)|2,{73,113,0}},
/* 51676 */ {(14<<2)|2,{73,118,0}},
/* 51677 */ {(14<<2)|2,{73,118,0}},
/* 51678 */ {(14<<2)|2,{73,118,0}},
/* 51679 */ {(14<<2)|2,{73,118,0}},
/* 51680 */ {(14<<2)|2,{73,119,0}},
/* 51681 */ {(14<<2)|2,{73,119,0}},
/* 51682 */ {(14<<2)|2,{73,119,0}},
/* 51683 */ {(14<<2)|2,{73,119,0}},
/* 51684 */ {(14<<2)|2,{73,120,0}},
/* 51685 */ {(14<<2)|2,{73,120,0}},
/* 51686 */ {(14<<2)|2,{73,120,0}},
/* 51687 */ {(14<<2)|2,{73,120,0}},
/* 51688 */ {(14<<2)|2,{73,121,0}},
/* 51689 */ {(14<<2)|2,{73,121,0}},
/* 51690 */ {(14<<2)|2,{73,121,0}},
/* 51691 */ {(14<<2)|2,{73,121,0}},
/* 51692 */ {(14<<2)|2,{73,122,0}},
/* 51693 */ {(14<<2)|2,{73,122,0}},
/* 51694 */ {(14<<2)|2,{73,122,0}},
/* 51695 */ {(14<<2)|2,{73,122,0}},
/* 51696 */ {(15<<2)|2,{73,38,0}},
/* 51697 */ {(15<<2)|2,{73,38,0}},
/* 51698 */ {(15<<2)|2,{73,42,0}},
/* 51699 */ {(15<<2)|2,{73,42,0}},
/* 51700 */ {(15<<2)|2,{73,44,0}},
/* 51701 */ {(15<<2)|2,{73,44,0}},
/* 51702 */ {(15<<2)|2,{73,59,0}},
/* 51703 */ {(15<<2)|2,{73,59,0}},
/* 51704 */ {(15<<2)|2,{73,88,0}},
/* 51705 */ {(15<<2)|2,{73,88,0}},
/* 51706 */ {(15<<2)|2,{73,90,0}},
/* 51707 */ {(15<<2)|2,{73,90,0}},
/* 51708 */ {(7<<2)|1,{73,0,0}},
/* 51709 */ {(7<<2)|1,{73,0,0}},
/* 51710 */ {(7<<2)|1,{73,0,0}},
/* 51711 */ {(7<<2)|1,{73,0,0}},
/* 51712 */ {(12<<2)|2,{74,48,0}},
/* 51713 */ {(12<<2)|2,{74,48,0}},
/* 51714 */ {(12<<2)|2,{74,48,0}},
/* 51715 */ {(12<<2)|2,{74,48,0}},
/* 51716 */ {(12<<2)|2,{74,48,0}},
/* 51717 */ {(12<<2)|2,{74,48,0}},
/* 51718 */ {(12<<2)|2,{74,48,0}},
/* 51719 */ {(12<<2)|2,{74,48,0}},
/* 51720 */ {(12<<2)|2,{74,48,0}},
/* 51721 */ {(12<<2)|2,{74,48,0}},
/* 51722 */ {(12<<2)|2,{74,48,0}},
/* 51723 */ {(12<<2)|2,{74,48,0}},
/* 51724 */ {(12<<2)|2,{74,48,0}},
/* 51725 */ {(12<<2)|2,{74,48,0}},
/* 51726 */ {(12<<2)|2,{74,48,0}},
/* 51727 */ {(12<<2)|2,{74,48,0}},
/* 51728 */ {(12<<2)|2,{74,49,0}},
/* 51729 */ {(12<<2)|2,{74,49,0}},
/* 51730 */ {(12<<2)|2,{74,49,0}},
/* 51731 */ {(12<<2)|2,{74,49,0}},
/* 51732 */ {(12<<2)|2,{74,49,0}},
/* 51733 */ {(12<<2)|2,{74,49,0}},
/* 51734 */ {(12<<2)|2,{74,49,0}},
/* 51735 */ {(12<<2)|2,{74,49,0}},
/* 51736 */ {(12<<2)|2,{74,49,0}},
/* 51737 */ {(12<<2)|2,{74,49,0}},
/* 51738 */ {(12<<2)|2,{74,49,0}},
/* 51739 */ {(12<<2)|2,{74,49,0}},
/* 51740 */ {(12<<2)|2,{74,49,0}},
/* 51741 */ {(12<<2)|2,{74,49,0}},
/* 51742 */ {(12<<2)|2,{74,49,0}},
/* 51743 */ {(12<<2)|2,{74,49,0}},
/* 51744 */ {(12<<2)|2,{74,50,0}},
/* 51745 */ {(12<<2)|2,{74,50,0}},
/* 51746 */ {(12<<2)|2,{74,50,0}},
/* 51747 */ {(12<<2)|2,{74,50,0}},
/* 51748 */ {(12<<2)|2,{74,50,0}},
/* 51749 */ {(12<<2)|2,{74,50,0}},
/* 51750 */ {(12<<2)|2,{74,50,0}},
/* 51751 */ {(12<<2)|2,{74,50,0}},
/* 51752 */ {(12<<2)|2,{74,50,0}},
/* 51753 */ {(12<<2)|2,{74,50,0}},
/* 51754 */ {(12<<2)|2,{74,50,0}},
/* 51755 */ {(12<<2)|2,{74,50,0}},
/* 51756 */ {(12<<2)|2,{74,50,0}},
/* 51757 */ {(12<<2)|2,{74,50,0}},
/* 51758 */ {(12<<2)|2,{74,50,0}},
/* 51759 */ {(12<<2)|2,{74,50,0}},
/* 51760 */ {(12<<2)|2,{74,97,0}},
/* 51761 */ {(12<<2)|2,{74,97,0}},
/* 51762 */ {(12<<2)|2,{74,97,0}},
/* 51763 */ {(12<<2)|2,{74,97,0}},
/* 51764 */ {(12<<2)|2,{74,97,0}},
/* 51765 */ {(12<<2)|2,{74,97,0}},
/* 51766 */ {(12<<2)|2,{74,97,0}},
/* 51767 */ {(12<<2)|2,{74,97,0}},
/* 51768 */ {(12<<2)|2,{74,97,0}},
/* 51769 */ {(12<<2)|2,{74,97,0}},
/* 51770 */ {(12<<2)|2,{74,97,0}},
/* 51771 */ {(12<<2)|2,{74,97,0}},
/* 51772 */ {(12<<2)|2,{74,97,0}},
/* 51773 */ {(12<<2)|2,{74,97,0}},
/* 51774 */ {(12<<2)|2,{74,97,0}},
/* 51775 */ {(12<<2)|2,{74,97,0}},
/* 51776 */ {(12<<2)|2,{74,99,0}},
/* 51777 */ {(12<<2)|2,{74,99,0}},
/* 51778 */ {(12<<2)|2,{74,99,0}},
/* 51779 */ {(12<<2)|2,{74,99,0}},
/* 51780 */ {(12<<2)|2,{74,99,0}},
/* 51781 */ {(12<<2)|2,{74,99,0}},
/* 51782 */ {(12<<2)|2,{74,99,0}},
/* 51783 */ {(12<<2)|2,{74,99,0}},
/* 51784 */ {(12<<2)|2,{74,99,0}},
/* 51785 */ {(12<<2)|2,{74,99,0}},
/* 51786 */ {(12<<2)|2,{74,99,0}},
/* 51787 */ {(12<<2)|2,{74,99,0}},
/* 51788 */ {(12<<2)|2,{74,99,0}},
/* 51789 */ {(12<<2)|2,{74,99,0}},
/* 51790 */ {(12<<2)|2,{74,99,0}},
/* 51791 */ {(12<<2)|2,{74,99,0}},
/* 51792 */ {(12<<2)|2,{74,101,0}},
/* 51793 */ {(12<<2)|2,{74,101,0}},
/* 51794 */ {(12<<2)|2,{74,101,0}},
/* 51795 */ {(12<<2)|2,{74,101,0}},
/* 51796 */ {(12<<2)|2,{74,101,0}},
/* 51797 */ {(12<<2)|2,{74,101,0}},
/* 51798 */ {(12<<2)|2,{74,101,0}},
/* 51799 */ {(12<<2)|2,{74,101,0}},
/* 51800 */ {(12<<2)|2,{74,101,0}},
/* 51801 */ {(12<<2)|2,{74,101,0}},
/* 51802 */ {(12<<2)|2,{74,101,0}},
/* 51803 */ {(12<<2)|2,{74,101,0}},
/* 51804 */ {(12<<2)|2,{74,101,0}},
/* 51805 */ {(12<<2)|2,{74,101,0}},
/* 51806 */ {(12<<2)|2,{74,101,0}},
/* 51807 */ {(12<<2)|2,{74,101,0}},
/* 51808 */ {(12<<2)|2,{74,105,0}},
/* 51809 */ {(12<<2)|2,{74,105,0}},
/* 51810 */ {(12<<2)|2,{74,105,0}},
/* 51811 */ {(12<<2)|2,{74,105,0}},
/* 51812 */ {(12<<2)|2,{74,105,0}},
/* 51813 */ {(12<<2)|2,{74,105,0}},
/* 51814 */ {(12<<2)|2,{74,105,0}},
/* 51815 */ {(12<<2)|2,{74,105,0}},
/* 51816 */ {(12<<2)|2,{74,105,0}},
/* 51817 */ {(12<<2)|2,{74,105,0}},
/* 51818 */ {(12<<2)|2,{74,105,0}},
/* 51819 */ {(12<<2)|2,{74,105,0}},
/* 51820 */ {(12<<2)|2,{74,105,0}},
/* 51821 */ {(12<<2)|2,{74,105,0}},
/* 51822 */ {(12<<2)|2,{74,105,0}},
/* 51823 */ {(12<<2)|2,{74,105,0}},
/* 51824 */ {(12<<2)|2,{74,111,0}},
/* 51825 */ {(12<<2)|2,{74,111,0}},
/* 51826 */ {(12<<2)|2,{74,111,0}},
/* 51827 */ {(12<<2)|2,{74,111,0}},
/* 51828 */ {(12<<2)|2,{74,111,0}},
/* 51829 */ {(12<<2)|2,{74,111,0}},
/* 51830 */ {(12<<2)|2,{74,111,0}},
/* 51831 */ {(12<<2)|2,{74,111,0}},
/* 51832 */ {(12<<2)|2,{74,111,0}},
/* 51833 */ {(12<<2)|2,{74,111,0}},
/* 51834 */ {(12<<2)|2,{74,111,0}},
/* 51835 */ {(12<<2)|2,{74,111,0}},
/* 51836 */ {(12<<2)|2,{74,111,0}},
/* 51837 */ {(12<<2)|2,{74,111,0}},
/* 51838 */ {(12<<2)|2,{74,111,0}},
/* 51839 */ {(12<<2)|2,{74,111,0}},
/* 51840 */ {(12<<2)|2,{74,115,0}},
/* 51841 */ {(12<<2)|2,{74,115,0}},
/* 51842 */ {(12<<2)|2,{74,115,0}},
/* 51843 */ {(12<<2)|2,{74,115,0}},
/* 51844 */ {(12<<2)|2,{74,115,0}},
/* 51845 */ {(12<<2)|2,{74,115,0}},
/* 51846 */ {(12<<2)|2,{74,115,0}},
/* 51847 */ {(12<<2)|2,{74,115,0}},
/* 51848 */ {(12<<2)|2,{74,115,0}},
/* 51849 */ {(12<<2)|2,{74,115,0}},
/* 51850 */ {(12<<2)|2,{74,115,0}},
/* 51851 */ {(12<<2)|2,{74,115,0}},
/* 51852 */ {(12<<2)|2,{74,115,0}},
/* 51853 */ {(12<<2)|2,{74,115,0}},
/* 51854 */ {(12<<2)|2,{74,115,0}},
/* 51855 */ {(12<<2)|2,{74,115,0}},
/* 51856 */ {(12<<2)|2,{74,116,0}},
/* 51857 */ {(12<<2)|2,{74,116,0}},
/* 51858 */ {(12<<2)|2,{74,116,0}},
/* 51859 */ {(12<<2)|2,{74,116,0}},
/* 51860 */ {(12<<2)|2,{74,116,0}},
/* 51861 */ {(12<<2)|2,{74,116,0}},
/* 51862 */ {(12<<2)|2,{74,116,0}},
/* 51863 */ {(12<<2)|2,{74,116,0}},
/* 51864 */ {(12<<2)|2,{74,116,0}},
/* 51865 */ {(12<<2)|2,{74,116,0}},
/* 51866 */ {(12<<2)|2,{74,116,0}},
/* 51867 */ {(12<<2)|2,{74,116,0}},
/* 51868 */ {(12<<2)|2,{74,116,0}},
/* 51869 */ {(12<<2)|2,{74,116,0}},
/* 51870 */ {(12<<2)|2,{74,116,0}},
/* 51871 */ {(12<<2)|2,{74,116,0}},
/* 51872 */ {(13<<2)|2,{74,32,0}},
/* 51873 */ {(13<<2)|2,{74,32,0}},
/* 51874 */ {(13<<2)|2,{74,32,0}},
/* 51875 */ {(13<<2)|2,{74,32,0}},
/* 51876 */ {(13<<2)|2,{74,32,0}},
/* 51877 */ {(13<<2)|2,{74,32,0}},
/* 51878 */ {(13<<2)|2,{74,32,0}},
/* 51879 */ {(13<<2)|2,{74,32,0}},
/* 51880 */ {(13<<2)|2,{74,37,0}},
/* 51881 */ {(13<<2)|2,{74,37,0}},
/* 51882 */ {(13<<2)|2,{74,37,0}},
/* 51883 */ {(13<<2)|2,{74,37,0}},
/* 51884 */ {(13<<2)|2,{74,37,0}},
/* 51885 */ {(13<<2)|2,{74,37,0}},
/* 51886 */ {(13<<2)|2,{74,37,0}},
/* 51887 */ {(13<<2)|2,{74,37,0}},
/* 51888 */ {(13<<2)|2,{74,45,0}},
/* 51889 */ {(13<<2)|2,{74,45,0}},
/* 51890 */ {(13<<2)|2,{74,45,0}},
/* 51891 */ {(13<<2)|2,{74,45,0}},
/* 51892 */ {(13<<2)|2,{74,45,0}},
/* 51893 */ {(13<<2)|2,{74,45,0}},
/* 51894 */ {(13<<2)|2,{74,45,0}},
/* 51895 */ {(13<<2)|2,{74,45,0}},
/* 51896 */ {(13<<2)|2,{74,46,0}},
/* 51897 */ {(13<<2)|2,{74,46,0}},
/* 51898 */ {(13<<2)|2,{74,46,0}},
/* 51899 */ {(13<<2)|2,{74,46,0}},
/* 51900 */ {(13<<2)|2,{74,46,0}},
/* 51901 */ {(13<<2)|2,{74,46,0}},
/* 51902 */ {(13<<2)|2,{74,46,0}},
/* 51903 */ {(13<<2)|2,{74,46,0}},
/* 51904 */ {(13<<2)|2,{74,47,0}},
/* 51905 */ {(13<<2)|2,{74,47,0}},
/* 51906 */ {(13<<2)|2,{74,47,0}},
/* 51907 */ {(13<<2)|2,{74,47,0}},
/* 51908 */ {(13<<2)|2,{74,47,0}},
/* 51909 */ {(13<<2)|2,{74,47,0}},
/* 51910 */ {(13<<2)|2,{74,47,0}},
/* 51911 */ {(13<<2)|2,{74,47,0}},
/* 51912 */ {(13<<2)|2,{74,51,0}},
/* 51913 */ {(13<<2)|2,{74,51,0}},
/* 51914 */ {(13<<2)|2,{74,51,0}},
/* 51915 */ {(13<<2)|2,{74,51,0}},
/* 51916 */ {(13<<2)|2,{74,51,0}},
/* 51917 */ {(13<<2)|2,{74,51,0}},
/* 51918 */ {(13<<2)|2,{74,51,0}},
/* 51919 */ {(13<<2)|2,{74,51,0}},
/* 51920 */ {(13<<2)|2,{74,52,0}},
/* 51921 */ {(13<<2)|2,{74,52,0}},
/* 51922 */ {(13<<2)|2,{74,52,0}},
/* 51923 */ {(13<<2)|2,{74,52,0}},
/* 51924 */ {(13<<2)|2,{74,52,0}},
/* 51925 */ {(13<<2)|2,{74,52,0}},
/* 51926 */ {(13<<2)|2,{74,52,0}},
/* 51927 */ {(13<<2)|2,{74,52,0}},
/* 51928 */ {(13<<2)|2,{74,53,0}},
/* 51929 */ {(13<<2)|2,{74,53,0}},
/* 51930 */ {(13<<2)|2,{74,53,0}},
/* 51931 */ {(13<<2)|2,{74,53,0}},
/* 51932 */ {(13<<2)|2,{74,53,0}},
/* 51933 */ {(13<<2)|2,{74,53,0}},
/* 51934 */ {(13<<2)|2,{74,53,0}},
/* 51935 */ {(13<<2)|2,{74,53,0}},
/* 51936 */ {(13<<2)|2,{74,54,0}},
/* 51937 */ {(13<<2)|2,{74,54,0}},
/* 51938 */ {(13<<2)|2,{74,54,0}},
/* 51939 */ {(13<<2)|2,{74,54,0}},
/* 51940 */ {(13<<2)|2,{74,54,0}},
/* 51941 */ {(13<<2)|2,{74,54,0}},
/* 51942 */ {(13<<2)|2,{74,54,0}},
/* 51943 */ {(13<<2)|2,{74,54,0}},
/* 51944 */ {(13<<2)|2,{74,55,0}},
/* 51945 */ {(13<<2)|2,{74,55,0}},
/* 51946 */ {(13<<2)|2,{74,55,0}},
/* 51947 */ {(13<<2)|2,{74,55,0}},
/* 51948 */ {(13<<2)|2,{74,55,0}},
/* 51949 */ {(13<<2)|2,{74,55,0}},
/* 51950 */ {(13<<2)|2,{74,55,0}},
/* 51951 */ {(13<<2)|2,{74,55,0}},
/* 51952 */ {(13<<2)|2,{74,56,0}},
/* 51953 */ {(13<<2)|2,{74,56,0}},
/* 51954 */ {(13<<2)|2,{74,56,0}},
/* 51955 */ {(13<<2)|2,{74,56,0}},
/* 51956 */ {(13<<2)|2,{74,56,0}},
/* 51957 */ {(13<<2)|2,{74,56,0}},
/* 51958 */ {(13<<2)|2,{74,56,0}},
/* 51959 */ {(13<<2)|2,{74,56,0}},
/* 51960 */ {(13<<2)|2,{74,57,0}},
/* 51961 */ {(13<<2)|2,{74,57,0}},
/* 51962 */ {(13<<2)|2,{74,57,0}},
/* 51963 */ {(13<<2)|2,{74,57,0}},
/* 51964 */ {(13<<2)|2,{74,57,0}},
/* 51965 */ {(13<<2)|2,{74,57,0}},
/* 51966 */ {(13<<2)|2,{74,57,0}},
/* 51967 */ {(13<<2)|2,{74,57,0}},
/* 51968 */ {(13<<2)|2,{74,61,0}},
/* 51969 */ {(13<<2)|2,{74,61,0}},
/* 51970 */ {(13<<2)|2,{74,61,0}},
/* 51971 */ {(13<<2)|2,{74,61,0}},
/* 51972 */ {(13<<2)|2,{74,61,0}},
/* 51973 */ {(13<<2)|2,{74,61,0}},
/* 51974 */ {(13<<2)|2,{74,61,0}},
/* 51975 */ {(13<<2)|2,{74,61,0}},
/* 51976 */ {(13<<2)|2,{74,65,0}},
/* 51977 */ {(13<<2)|2,{74,65,0}},
/* 51978 */ {(13<<2)|2,{74,65,0}},
/* 51979 */ {(13<<2)|2,{74,65,0}},
/* 51980 */ {(13<<2)|2,{74,65,0}},
/* 51981 */ {(13<<2)|2,{74,65,0}},
/* 51982 */ {(13<<2)|2,{74,65,0}},
/* 51983 */ {(13<<2)|2,{74,65,0}},
/* 51984 */ {(13<<2)|2,{74,95,0}},
/* 51985 */ {(13<<2)|2,{74,95,0}},
/* 51986 */ {(13<<2)|2,{74,95,0}},
/* 51987 */ {(13<<2)|2,{74,95,0}},
/* 51988 */ {(13<<2)|2,{74,95,0}},
/* 51989 */ {(13<<2)|2,{74,95,0}},
/* 51990 */ {(13<<2)|2,{74,95,0}},
/* 51991 */ {(13<<2)|2,{74,95,0}},
/* 51992 */ {(13<<2)|2,{74,98,0}},
/* 51993 */ {(13<<2)|2,{74,98,0}},
/* 51994 */ {(13<<2)|2,{74,98,0}},
/* 51995 */ {(13<<2)|2,{74,98,0}},
/* 51996 */ {(13<<2)|2,{74,98,0}},
/* 51997 */ {(13<<2)|2,{74,98,0}},
/* 51998 */ {(13<<2)|2,{74,98,0}},
/* 51999 */ {(13<<2)|2,{74,98,0}},
/* 52000 */ {(13<<2)|2,{74,100,0}},
/* 52001 */ {(13<<2)|2,{74,100,0}},
/* 52002 */ {(13<<2)|2,{74,100,0}},
/* 52003 */ {(13<<2)|2,{74,100,0}},
/* 52004 */ {(13<<2)|2,{74,100,0}},
/* 52005 */ {(13<<2)|2,{74,100,0}},
/* 52006 */ {(13<<2)|2,{74,100,0}},
/* 52007 */ {(13<<2)|2,{74,100,0}},
/* 52008 */ {(13<<2)|2,{74,102,0}},
/* 52009 */ {(13<<2)|2,{74,102,0}},
/* 52010 */ {(13<<2)|2,{74,102,0}},
/* 52011 */ {(13<<2)|2,{74,102,0}},
/* 52012 */ {(13<<2)|2,{74,102,0}},
/* 52013 */ {(13<<2)|2,{74,102,0}},
/* 52014 */ {(13<<2)|2,{74,102,0}},
/* 52015 */ {(13<<2)|2,{74,102,0}},
/* 52016 */ {(13<<2)|2,{74,103,0}},
/* 52017 */ {(13<<2)|2,{74,103,0}},
/* 52018 */ {(13<<2)|2,{74,103,0}},
/* 52019 */ {(13<<2)|2,{74,103,0}},
/* 52020 */ {(13<<2)|2,{74,103,0}},
/* 52021 */ {(13<<2)|2,{74,103,0}},
/* 52022 */ {(13<<2)|2,{74,103,0}},
/* 52023 */ {(13<<2)|2,{74,103,0}},
/* 52024 */ {(13<<2)|2,{74,104,0}},
/* 52025 */ {(13<<2)|2,{74,104,0}},
/* 52026 */ {(13<<2)|2,{74,104,0}},
/* 52027 */ {(13<<2)|2,{74,104,0}},
/* 52028 */ {(13<<2)|2,{74,104,0}},
/* 52029 */ {(13<<2)|2,{74,104,0}},
/* 52030 */ {(13<<2)|2,{74,104,0}},
/* 52031 */ {(13<<2)|2,{74,104,0}},
/* 52032 */ {(13<<2)|2,{74,108,0}},
/* 52033 */ {(13<<2)|2,{74,108,0}},
/* 52034 */ {(13<<2)|2,{74,108,0}},
/* 52035 */ {(13<<2)|2,{74,108,0}},
/* 52036 */ {(13<<2)|2,{74,108,0}},
/* 52037 */ {(13<<2)|2,{74,108,0}},
/* 52038 */ {(13<<2)|2,{74,108,0}},
/* 52039 */ {(13<<2)|2,{74,108,0}},
/* 52040 */ {(13<<2)|2,{74,109,0}},
/* 52041 */ {(13<<2)|2,{74,109,0}},
/* 52042 */ {(13<<2)|2,{74,109,0}},
/* 52043 */ {(13<<2)|2,{74,109,0}},
/* 52044 */ {(13<<2)|2,{74,109,0}},
/* 52045 */ {(13<<2)|2,{74,109,0}},
/* 52046 */ {(13<<2)|2,{74,109,0}},
/* 52047 */ {(13<<2)|2,{74,109,0}},
/* 52048 */ {(13<<2)|2,{74,110,0}},
/* 52049 */ {(13<<2)|2,{74,110,0}},
/* 52050 */ {(13<<2)|2,{74,110,0}},
/* 52051 */ {(13<<2)|2,{74,110,0}},
/* 52052 */ {(13<<2)|2,{74,110,0}},
/* 52053 */ {(13<<2)|2,{74,110,0}},
/* 52054 */ {(13<<2)|2,{74,110,0}},
/* 52055 */ {(13<<2)|2,{74,110,0}},
/* 52056 */ {(13<<2)|2,{74,112,0}},
/* 52057 */ {(13<<2)|2,{74,112,0}},
/* 52058 */ {(13<<2)|2,{74,112,0}},
/* 52059 */ {(13<<2)|2,{74,112,0}},
/* 52060 */ {(13<<2)|2,{74,112,0}},
/* 52061 */ {(13<<2)|2,{74,112,0}},
/* 52062 */ {(13<<2)|2,{74,112,0}},
/* 52063 */ {(13<<2)|2,{74,112,0}},
/* 52064 */ {(13<<2)|2,{74,114,0}},
/* 52065 */ {(13<<2)|2,{74,114,0}},
/* 52066 */ {(13<<2)|2,{74,114,0}},
/* 52067 */ {(13<<2)|2,{74,114,0}},
/* 52068 */ {(13<<2)|2,{74,114,0}},
/* 52069 */ {(13<<2)|2,{74,114,0}},
/* 52070 */ {(13<<2)|2,{74,114,0}},
/* 52071 */ {(13<<2)|2,{74,114,0}},
/* 52072 */ {(13<<2)|2,{74,117,0}},
/* 52073 */ {(13<<2)|2,{74,117,0}},
/* 52074 */ {(13<<2)|2,{74,117,0}},
/* 52075 */ {(13<<2)|2,{74,117,0}},
/* 52076 */ {(13<<2)|2,{74,117,0}},
/* 52077 */ {(13<<2)|2,{74,117,0}},
/* 52078 */ {(13<<2)|2,{74,117,0}},
/* 52079 */ {(13<<2)|2,{74,117,0}},
/* 52080 */ {(14<<2)|2,{74,58,0}},
/* 52081 */ {(14<<2)|2,{74,58,0}},
/* 52082 */ {(14<<2)|2,{74,58,0}},
/* 52083 */ {(14<<2)|2,{74,58,0}},
/* 52084 */ {(14<<2)|2,{74,66,0}},
/* 52085 */ {(14<<2)|2,{74,66,0}},
/* 52086 */ {(14<<2)|2,{74,66,0}},
/* 52087 */ {(14<<2)|2,{74,66,0}},
/* 52088 */ {(14<<2)|2,{74,67,0}},
/* 52089 */ {(14<<2)|2,{74,67,0}},
/* 52090 */ {(14<<2)|2,{74,67,0}},
/* 52091 */ {(14<<2)|2,{74,67,0}},
/* 52092 */ {(14<<2)|2,{74,68,0}},
/* 52093 */ {(14<<2)|2,{74,68,0}},
/* 52094 */ {(14<<2)|2,{74,68,0}},
/* 52095 */ {(14<<2)|2,{74,68,0}},
/* 52096 */ {(14<<2)|2,{74,69,0}},
/* 52097 */ {(14<<2)|2,{74,69,0}},
/* 52098 */ {(14<<2)|2,{74,69,0}},
/* 52099 */ {(14<<2)|2,{74,69,0}},
/* 52100 */ {(14<<2)|2,{74,70,0}},
/* 52101 */ {(14<<2)|2,{74,70,0}},
/* 52102 */ {(14<<2)|2,{74,70,0}},
/* 52103 */ {(14<<2)|2,{74,70,0}},
/* 52104 */ {(14<<2)|2,{74,71,0}},
/* 52105 */ {(14<<2)|2,{74,71,0}},
/* 52106 */ {(14<<2)|2,{74,71,0}},
/* 52107 */ {(14<<2)|2,{74,71,0}},
/* 52108 */ {(14<<2)|2,{74,72,0}},
/* 52109 */ {(14<<2)|2,{74,72,0}},
/* 52110 */ {(14<<2)|2,{74,72,0}},
/* 52111 */ {(14<<2)|2,{74,72,0}},
/* 52112 */ {(14<<2)|2,{74,73,0}},
/* 52113 */ {(14<<2)|2,{74,73,0}},
/* 52114 */ {(14<<2)|2,{74,73,0}},
/* 52115 */ {(14<<2)|2,{74,73,0}},
/* 52116 */ {(14<<2)|2,{74,74,0}},
/* 52117 */ {(14<<2)|2,{74,74,0}},
/* 52118 */ {(14<<2)|2,{74,74,0}},
/* 52119 */ {(14<<2)|2,{74,74,0}},
/* 52120 */ {(14<<2)|2,{74,75,0}},
/* 52121 */ {(14<<2)|2,{74,75,0}},
/* 52122 */ {(14<<2)|2,{74,75,0}},
/* 52123 */ {(14<<2)|2,{74,75,0}},
/* 52124 */ {(14<<2)|2,{74,76,0}},
/* 52125 */ {(14<<2)|2,{74,76,0}},
/* 52126 */ {(14<<2)|2,{74,76,0}},
/* 52127 */ {(14<<2)|2,{74,76,0}},
/* 52128 */ {(14<<2)|2,{74,77,0}},
/* 52129 */ {(14<<2)|2,{74,77,0}},
/* 52130 */ {(14<<2)|2,{74,77,0}},
/* 52131 */ {(14<<2)|2,{74,77,0}},
/* 52132 */ {(14<<2)|2,{74,78,0}},
/* 52133 */ {(14<<2)|2,{74,78,0}},
/* 52134 */ {(14<<2)|2,{74,78,0}},
/* 52135 */ {(14<<2)|2,{74,78,0}},
/* 52136 */ {(14<<2)|2,{74,79,0}},
/* 52137 */ {(14<<2)|2,{74,79,0}},
/* 52138 */ {(14<<2)|2,{74,79,0}},
/* 52139 */ {(14<<2)|2,{74,79,0}},
/* 52140 */ {(14<<2)|2,{74,80,0}},
/* 52141 */ {(14<<2)|2,{74,80,0}},
/* 52142 */ {(14<<2)|2,{74,80,0}},
/* 52143 */ {(14<<2)|2,{74,80,0}},
/* 52144 */ {(14<<2)|2,{74,81,0}},
/* 52145 */ {(14<<2)|2,{74,81,0}},
/* 52146 */ {(14<<2)|2,{74,81,0}},
/* 52147 */ {(14<<2)|2,{74,81,0}},
/* 52148 */ {(14<<2)|2,{74,82,0}},
/* 52149 */ {(14<<2)|2,{74,82,0}},
/* 52150 */ {(14<<2)|2,{74,82,0}},
/* 52151 */ {(14<<2)|2,{74,82,0}},
/* 52152 */ {(14<<2)|2,{74,83,0}},
/* 52153 */ {(14<<2)|2,{74,83,0}},
/* 52154 */ {(14<<2)|2,{74,83,0}},
/* 52155 */ {(14<<2)|2,{74,83,0}},
/* 52156 */ {(14<<2)|2,{74,84,0}},
/* 52157 */ {(14<<2)|2,{74,84,0}},
/* 52158 */ {(14<<2)|2,{74,84,0}},
/* 52159 */ {(14<<2)|2,{74,84,0}},
/* 52160 */ {(14<<2)|2,{74,85,0}},
/* 52161 */ {(14<<2)|2,{74,85,0}},
/* 52162 */ {(14<<2)|2,{74,85,0}},
/* 52163 */ {(14<<2)|2,{74,85,0}},
/* 52164 */ {(14<<2)|2,{74,86,0}},
/* 52165 */ {(14<<2)|2,{74,86,0}},
/* 52166 */ {(14<<2)|2,{74,86,0}},
/* 52167 */ {(14<<2)|2,{74,86,0}},
/* 52168 */ {(14<<2)|2,{74,87,0}},
/* 52169 */ {(14<<2)|2,{74,87,0}},
/* 52170 */ {(14<<2)|2,{74,87,0}},
/* 52171 */ {(14<<2)|2,{74,87,0}},
/* 52172 */ {(14<<2)|2,{74,89,0}},
/* 52173 */ {(14<<2)|2,{74,89,0}},
/* 52174 */ {(14<<2)|2,{74,89,0}},
/* 52175 */ {(14<<2)|2,{74,89,0}},
/* 52176 */ {(14<<2)|2,{74,106,0}},
/* 52177 */ {(14<<2)|2,{74,106,0}},
/* 52178 */ {(14<<2)|2,{74,106,0}},
/* 52179 */ {(14<<2)|2,{74,106,0}},
/* 52180 */ {(14<<2)|2,{74,107,0}},
/* 52181 */ {(14<<2)|2,{74,107,0}},
/* 52182 */ {(14<<2)|2,{74,107,0}},
/* 52183 */ {(14<<2)|2,{74,107,0}},
/* 52184 */ {(14<<2)|2,{74,113,0}},
/* 52185 */ {(14<<2)|2,{74,113,0}},
/* 52186 */ {(14<<2)|2,{74,113,0}},
/* 52187 */ {(14<<2)|2,{74,113,0}},
/* 52188 */ {(14<<2)|2,{74,118,0}},
/* 52189 */ {(14<<2)|2,{74,118,0}},
/* 52190 */ {(14<<2)|2,{74,118,0}},
/* 52191 */ {(14<<2)|2,{74,118,0}},
/* 52192 */ {(14<<2)|2,{74,119,0}},
/* 52193 */ {(14<<2)|2,{74,119,0}},
/* 52194 */ {(14<<2)|2,{74,119,0}},
/* 52195 */ {(14<<2)|2,{74,119,0}},
/* 52196 */ {(14<<2)|2,{74,120,0}},
/* 52197 */ {(14<<2)|2,{74,120,0}},
/* 52198 */ {(14<<2)|2,{74,120,0}},
/* 52199 */ {(14<<2)|2,{74,120,0}},
/* 52200 */ {(14<<2)|2,{74,121,0}},
/* 52201 */ {(14<<2)|2,{74,121,0}},
/* 52202 */ {(14<<2)|2,{74,121,0}},
/* 52203 */ {(14<<2)|2,{74,121,0}},
/* 52204 */ {(14<<2)|2,{74,122,0}},
/* 52205 */ {(14<<2)|2,{74,122,0}},
/* 52206 */ {(14<<2)|2,{74,122,0}},
/* 52207 */ {(14<<2)|2,{74,122,0}},
/* 52208 */ {(15<<2)|2,{74,38,0}},
/* 52209 */ {(15<<2)|2,{74,38,0}},
/* 52210 */ {(15<<2)|2,{74,42,0}},
/* 52211 */ {(15<<2)|2,{74,42,0}},
/* 52212 */ {(15<<2)|2,{74,44,0}},
/* 52213 */ {(15<<2)|2,{74,44,0}},
/* 52214 */ {(15<<2)|2,{74,59,0}},
/* 52215 */ {(15<<2)|2,{74,59,0}},
/* 52216 */ {(15<<2)|2,{74,88,0}},
/* 52217 */ {(15<<2)|2,{74,88,0}},
/* 52218 */ {(15<<2)|2,{74,90,0}},
/* 52219 */ {(15<<2)|2,{74,90,0}},
/* 52220 */ {(7<<2)|1,{74,0,0}},
/* 52221 */ {(7<<2)|1,{74,0,0}},
/* 52222 */ {(7<<2)|1,{74,0,0}},
/* 52223 */ {(7<<2)|1,{74,0,0}},
/* 52224 */ {(12<<2)|2,{75,48,0}},
/* 52225 */ {(12<<2)|2,{75,48,0}},
/* 52226 */ {(12<<2)|2,{75,48,0}},
/* 52227 */ {(12<<2)|2,{75,48,0}},
/* 52228 */ {(12<<2)|2,{75,48,0}},
/* 52229 */ {(12<<2)|2,{75,48,0}},
/* 52230 */ {(12<<2)|2,{75,48,0}},
/* 52231 */ {(12<<2)|2,{75,48,0}},
/* 52232 */ {(12<<2)|2,{75,48,0}},
/* 52233 */ {(12<<2)|2,{75,48,0}},
/* 52234 */ {(12<<2)|2,{75,48,0}},
/* 52235 */ {(12<<2)|2,{75,48,0}},
/* 52236 */ {(12<<2)|2,{75,48,0}},
/* 52237 */ {(12<<2)|2,{75,48,0}},
/* 52238 */ {(12<<2)|2,{75,48,0}},
/* 52239 */ {(12<<2)|2,{75,48,0}},
/* 52240 */ {(12<<2)|2,{75,49,0}},
/* 52241 */ {(12<<2)|2,{75,49,0}},
/* 52242 */ {(12<<2)|2,{75,49,0}},
/* 52243 */ {(12<<2)|2,{75,49,0}},
/* 52244 */ {(12<<2)|2,{75,49,0}},
/* 52245 */ {(12<<2)|2,{75,49,0}},
/* 52246 */ {(12<<2)|2,{75,49,0}},
/* 52247 */ {(12<<2)|2,{75,49,0}},
/* 52248 */ {(12<<2)|2,{75,49,0}},
/* 52249 */ {(12<<2)|2,{75,49,0}},
/* 52250 */ {(12<<2)|2,{75,49,0}},
/* 52251 */ {(12<<2)|2,{75,49,0}},
/* 52252 */ {(12<<2)|2,{75,49,0}},
/* 52253 */ {(12<<2)|2,{75,49,0}},
/* 52254 */ {(12<<2)|2,{75,49,0}},
/* 52255 */ {(12<<2)|2,{75,49,0}},
/* 52256 */ {(12<<2)|2,{75,50,0}},
/* 52257 */ {(12<<2)|2,{75,50,0}},
/* 52258 */ {(12<<2)|2,{75,50,0}},
/* 52259 */ {(12<<2)|2,{75,50,0}},
/* 52260 */ {(12<<2)|2,{75,50,0}},
/* 52261 */ {(12<<2)|2,{75,50,0}},
/* 52262 */ {(12<<2)|2,{75,50,0}},
/* 52263 */ {(12<<2)|2,{75,50,0}},
/* 52264 */ {(12<<2)|2,{75,50,0}},
/* 52265 */ {(12<<2)|2,{75,50,0}},
/* 52266 */ {(12<<2)|2,{75,50,0}},
/* 52267 */ {(12<<2)|2,{75,50,0}},
/* 52268 */ {(12<<2)|2,{75,50,0}},
/* 52269 */ {(12<<2)|2,{75,50,0}},
/* 52270 */ {(12<<2)|2,{75,50,0}},
/* 52271 */ {(12<<2)|2,{75,50,0}},
/* 52272 */ {(12<<2)|2,{75,97,0}},
/* 52273 */ {(12<<2)|2,{75,97,0}},
/* 52274 */ {(12<<2)|2,{75,97,0}},
/* 52275 */ {(12<<2)|2,{75,97,0}},
/* 52276 */ {(12<<2)|2,{75,97,0}},
/* 52277 */ {(12<<2)|2,{75,97,0}},
/* 52278 */ {(12<<2)|2,{75,97,0}},
/* 52279 */ {(12<<2)|2,{75,97,0}},
/* 52280 */ {(12<<2)|2,{75,97,0}},
/* 52281 */ {(12<<2)|2,{75,97,0}},
/* 52282 */ {(12<<2)|2,{75,97,0}},
/* 52283 */ {(12<<2)|2,{75,97,0}},
/* 52284 */ {(12<<2)|2,{75,97,0}},
/* 52285 */ {(12<<2)|2,{75,97,0}},
/* 52286 */ {(12<<2)|2,{75,97,0}},
/* 52287 */ {(12<<2)|2,{75,97,0}},
/* 52288 */ {(12<<2)|2,{75,99,0}},
/* 52289 */ {(12<<2)|2,{75,99,0}},
/* 52290 */ {(12<<2)|2,{75,99,0}},
/* 52291 */ {(12<<2)|2,{75,99,0}},
/* 52292 */ {(12<<2)|2,{75,99,0}},
/* 52293 */ {(12<<2)|2,{75,99,0}},
/* 52294 */ {(12<<2)|2,{75,99,0}},
/* 52295 */ {(12<<2)|2,{75,99,0}},
/* 52296 */ {(12<<2)|2,{75,99,0}},
/* 52297 */ {(12<<2)|2,{75,99,0}},
/* 52298 */ {(12<<2)|2,{75,99,0}},
/* 52299 */ {(12<<2)|2,{75,99,0}},
/* 52300 */ {(12<<2)|2,{75,99,0}},
/* 52301 */ {(12<<2)|2,{75,99,0}},
/* 52302 */ {(12<<2)|2,{75,99,0}},
/* 52303 */ {(12<<2)|2,{75,99,0}},
/* 52304 */ {(12<<2)|2,{75,101,0}},
/* 52305 */ {(12<<2)|2,{75,101,0}},
/* 52306 */ {(12<<2)|2,{75,101,0}},
/* 52307 */ {(12<<2)|2,{75,101,0}},
/* 52308 */ {(12<<2)|2,{75,101,0}},
/* 52309 */ {(12<<2)|2,{75,101,0}},
/* 52310 */ {(12<<2)|2,{75,101,0}},
/* 52311 */ {(12<<2)|2,{75,101,0}},
/* 52312 */ {(12<<2)|2,{75,101,0}},
/* 52313 */ {(12<<2)|2,{75,101,0}},
/* 52314 */ {(12<<2)|2,{75,101,0}},
/* 52315 */ {(12<<2)|2,{75,101,0}},
/* 52316 */ {(12<<2)|2,{75,101,0}},
/* 52317 */ {(12<<2)|2,{75,101,0}},
/* 52318 */ {(12<<2)|2,{75,101,0}},
/* 52319 */ {(12<<2)|2,{75,101,0}},
/* 52320 */ {(12<<2)|2,{75,105,0}},
/* 52321 */ {(12<<2)|2,{75,105,0}},
/* 52322 */ {(12<<2)|2,{75,105,0}},
/* 52323 */ {(12<<2)|2,{75,105,0}},
/* 52324 */ {(12<<2)|2,{75,105,0}},
/* 52325 */ {(12<<2)|2,{75,105,0}},
/* 52326 */ {(12<<2)|2,{75,105,0}},
/* 52327 */ {(12<<2)|2,{75,105,0}},
/* 52328 */ {(12<<2)|2,{75,105,0}},
/* 52329 */ {(12<<2)|2,{75,105,0}},
/* 52330 */ {(12<<2)|2,{75,105,0}},
/* 52331 */ {(12<<2)|2,{75,105,0}},
/* 52332 */ {(12<<2)|2,{75,105,0}},
/* 52333 */ {(12<<2)|2,{75,105,0}},
/* 52334 */ {(12<<2)|2,{75,105,0}},
/* 52335 */ {(12<<2)|2,{75,105,0}},
/* 52336 */ {(12<<2)|2,{75,111,0}},
/* 52337 */ {(12<<2)|2,{75,111,0}},
/* 52338 */ {(12<<2)|2,{75,111,0}},
/* 52339 */ {(12<<2)|2,{75,111,0}},
/* 52340 */ {(12<<2)|2,{75,111,0}},
/* 52341 */ {(12<<2)|2,{75,111,0}},
/* 52342 */ {(12<<2)|2,{75,111,0}},
/* 52343 */ {(12<<2)|2,{75,111,0}},
/* 52344 */ {(12<<2)|2,{75,111,0}},
/* 52345 */ {(12<<2)|2,{75,111,0}},
/* 52346 */ {(12<<2)|2,{75,111,0}},
/* 52347 */ {(12<<2)|2,{75,111,0}},
/* 52348 */ {(12<<2)|2,{75,111,0}},
/* 52349 */ {(12<<2)|2,{75,111,0}},
/* 52350 */ {(12<<2)|2,{75,111,0}},
/* 52351 */ {(12<<2)|2,{75,111,0}},
/* 52352 */ {(12<<2)|2,{75,115,0}},
/* 52353 */ {(12<<2)|2,{75,115,0}},
/* 52354 */ {(12<<2)|2,{75,115,0}},
/* 52355 */ {(12<<2)|2,{75,115,0}},
/* 52356 */ {(12<<2)|2,{75,115,0}},
/* 52357 */ {(12<<2)|2,{75,115,0}},
/* 52358 */ {(12<<2)|2,{75,115,0}},
/* 52359 */ {(12<<2)|2,{75,115,0}},
/* 52360 */ {(12<<2)|2,{75,115,0}},
/* 52361 */ {(12<<2)|2,{75,115,0}},
/* 52362 */ {(12<<2)|2,{75,115,0}},
/* 52363 */ {(12<<2)|2,{75,115,0}},
/* 52364 */ {(12<<2)|2,{75,115,0}},
/* 52365 */ {(12<<2)|2,{75,115,0}},
/* 52366 */ {(12<<2)|2,{75,115,0}},
/* 52367 */ {(12<<2)|2,{75,115,0}},
/* 52368 */ {(12<<2)|2,{75,116,0}},
/* 52369 */ {(12<<2)|2,{75,116,0}},
/* 52370 */ {(12<<2)|2,{75,116,0}},
/* 52371 */ {(12<<2)|2,{75,116,0}},
/* 52372 */ {(12<<2)|2,{75,116,0}},
/* 52373 */ {(12<<2)|2,{75,116,0}},
/* 52374 */ {(12<<2)|2,{75,116,0}},
/* 52375 */ {(12<<2)|2,{75,116,0}},
/* 52376 */ {(12<<2)|2,{75,116,0}},
/* 52377 */ {(12<<2)|2,{75,116,0}},
/* 52378 */ {(12<<2)|2,{75,116,0}},
/* 52379 */ {(12<<2)|2,{75,116,0}},
/* 52380 */ {(12<<2)|2,{75,116,0}},
/* 52381 */ {(12<<2)|2,{75,116,0}},
/* 52382 */ {(12<<2)|2,{75,116,0}},
/* 52383 */ {(12<<2)|2,{75,116,0}},
/* 52384 */ {(13<<2)|2,{75,32,0}},
/* 52385 */ {(13<<2)|2,{75,32,0}},
/* 52386 */ {(13<<2)|2,{75,32,0}},
/* 52387 */ {(13<<2)|2,{75,32,0}},
/* 52388 */ {(13<<2)|2,{75,32,0}},
/* 52389 */ {(13<<2)|2,{75,32,0}},
/* 52390 */ {(13<<2)|2,{75,32,0}},
/* 52391 */ {(13<<2)|2,{75,32,0}},
/* 52392 */ {(13<<2)|2,{75,37,0}},
/* 52393 */ {(13<<2)|2,{75,37,0}},
/* 52394 */ {(13<<2)|2,{75,37,0}},
/* 52395 */ {(13<<2)|2,{75,37,0}},
/* 52396 */ {(13<<2)|2,{75,37,0}},
/* 52397 */ {(13<<2)|2,{75,37,0}},
/* 52398 */ {(13<<2)|2,{75,37,0}},
/* 52399 */ {(13<<2)|2,{75,37,0}},
/* 52400 */ {(13<<2)|2,{75,45,0}},
/* 52401 */ {(13<<2)|2,{75,45,0}},
/* 52402 */ {(13<<2)|2,{75,45,0}},
/* 52403 */ {(13<<2)|2,{75,45,0}},
/* 52404 */ {(13<<2)|2,{75,45,0}},
/* 52405 */ {(13<<2)|2,{75,45,0}},
/* 52406 */ {(13<<2)|2,{75,45,0}},
/* 52407 */ {(13<<2)|2,{75,45,0}},
/* 52408 */ {(13<<2)|2,{75,46,0}},
/* 52409 */ {(13<<2)|2,{75,46,0}},
/* 52410 */ {(13<<2)|2,{75,46,0}},
/* 52411 */ {(13<<2)|2,{75,46,0}},
/* 52412 */ {(13<<2)|2,{75,46,0}},
/* 52413 */ {(13<<2)|2,{75,46,0}},
/* 52414 */ {(13<<2)|2,{75,46,0}},
/* 52415 */ {(13<<2)|2,{75,46,0}},
/* 52416 */ {(13<<2)|2,{75,47,0}},
/* 52417 */ {(13<<2)|2,{75,47,0}},
/* 52418 */ {(13<<2)|2,{75,47,0}},
/* 52419 */ {(13<<2)|2,{75,47,0}},
/* 52420 */ {(13<<2)|2,{75,47,0}},
/* 52421 */ {(13<<2)|2,{75,47,0}},
/* 52422 */ {(13<<2)|2,{75,47,0}},
/* 52423 */ {(13<<2)|2,{75,47,0}},
/* 52424 */ {(13<<2)|2,{75,51,0}},
/* 52425 */ {(13<<2)|2,{75,51,0}},
/* 52426 */ {(13<<2)|2,{75,51,0}},
/* 52427 */ {(13<<2)|2,{75,51,0}},
/* 52428 */ {(13<<2)|2,{75,51,0}},
/* 52429 */ {(13<<2)|2,{75,51,0}},
/* 52430 */ {(13<<2)|2,{75,51,0}},
/* 52431 */ {(13<<2)|2,{75,51,0}},
/* 52432 */ {(13<<2)|2,{75,52,0}},
/* 52433 */ {(13<<2)|2,{75,52,0}},
/* 52434 */ {(13<<2)|2,{75,52,0}},
/* 52435 */ {(13<<2)|2,{75,52,0}},
/* 52436 */ {(13<<2)|2,{75,52,0}},
/* 52437 */ {(13<<2)|2,{75,52,0}},
/* 52438 */ {(13<<2)|2,{75,52,0}},
/* 52439 */ {(13<<2)|2,{75,52,0}},
/* 52440 */ {(13<<2)|2,{75,53,0}},
/* 52441 */ {(13<<2)|2,{75,53,0}},
/* 52442 */ {(13<<2)|2,{75,53,0}},
/* 52443 */ {(13<<2)|2,{75,53,0}},
/* 52444 */ {(13<<2)|2,{75,53,0}},
/* 52445 */ {(13<<2)|2,{75,53,0}},
/* 52446 */ {(13<<2)|2,{75,53,0}},
/* 52447 */ {(13<<2)|2,{75,53,0}},
/* 52448 */ {(13<<2)|2,{75,54,0}},
/* 52449 */ {(13<<2)|2,{75,54,0}},
/* 52450 */ {(13<<2)|2,{75,54,0}},
/* 52451 */ {(13<<2)|2,{75,54,0}},
/* 52452 */ {(13<<2)|2,{75,54,0}},
/* 52453 */ {(13<<2)|2,{75,54,0}},
/* 52454 */ {(13<<2)|2,{75,54,0}},
/* 52455 */ {(13<<2)|2,{75,54,0}},
/* 52456 */ {(13<<2)|2,{75,55,0}},
/* 52457 */ {(13<<2)|2,{75,55,0}},
/* 52458 */ {(13<<2)|2,{75,55,0}},
/* 52459 */ {(13<<2)|2,{75,55,0}},
/* 52460 */ {(13<<2)|2,{75,55,0}},
/* 52461 */ {(13<<2)|2,{75,55,0}},
/* 52462 */ {(13<<2)|2,{75,55,0}},
/* 52463 */ {(13<<2)|2,{75,55,0}},
/* 52464 */ {(13<<2)|2,{75,56,0}},
/* 52465 */ {(13<<2)|2,{75,56,0}},
/* 52466 */ {(13<<2)|2,{75,56,0}},
/* 52467 */ {(13<<2)|2,{75,56,0}},
/* 52468 */ {(13<<2)|2,{75,56,0}},
/* 52469 */ {(13<<2)|2,{75,56,0}},
/* 52470 */ {(13<<2)|2,{75,56,0}},
/* 52471 */ {(13<<2)|2,{75,56,0}},
/* 52472 */ {(13<<2)|2,{75,57,0}},
/* 52473 */ {(13<<2)|2,{75,57,0}},
/* 52474 */ {(13<<2)|2,{75,57,0}},
/* 52475 */ {(13<<2)|2,{75,57,0}},
/* 52476 */ {(13<<2)|2,{75,57,0}},
/* 52477 */ {(13<<2)|2,{75,57,0}},
/* 52478 */ {(13<<2)|2,{75,57,0}},
/* 52479 */ {(13<<2)|2,{75,57,0}},
/* 52480 */ {(13<<2)|2,{75,61,0}},
/* 52481 */ {(13<<2)|2,{75,61,0}},
/* 52482 */ {(13<<2)|2,{75,61,0}},
/* 52483 */ {(13<<2)|2,{75,61,0}},
/* 52484 */ {(13<<2)|2,{75,61,0}},
/* 52485 */ {(13<<2)|2,{75,61,0}},
/* 52486 */ {(13<<2)|2,{75,61,0}},
/* 52487 */ {(13<<2)|2,{75,61,0}},
/* 52488 */ {(13<<2)|2,{75,65,0}},
/* 52489 */ {(13<<2)|2,{75,65,0}},
/* 52490 */ {(13<<2)|2,{75,65,0}},
/* 52491 */ {(13<<2)|2,{75,65,0}},
/* 52492 */ {(13<<2)|2,{75,65,0}},
/* 52493 */ {(13<<2)|2,{75,65,0}},
/* 52494 */ {(13<<2)|2,{75,65,0}},
/* 52495 */ {(13<<2)|2,{75,65,0}},
/* 52496 */ {(13<<2)|2,{75,95,0}},
/* 52497 */ {(13<<2)|2,{75,95,0}},
/* 52498 */ {(13<<2)|2,{75,95,0}},
/* 52499 */ {(13<<2)|2,{75,95,0}},
/* 52500 */ {(13<<2)|2,{75,95,0}},
/* 52501 */ {(13<<2)|2,{75,95,0}},
/* 52502 */ {(13<<2)|2,{75,95,0}},
/* 52503 */ {(13<<2)|2,{75,95,0}},
/* 52504 */ {(13<<2)|2,{75,98,0}},
/* 52505 */ {(13<<2)|2,{75,98,0}},
/* 52506 */ {(13<<2)|2,{75,98,0}},
/* 52507 */ {(13<<2)|2,{75,98,0}},
/* 52508 */ {(13<<2)|2,{75,98,0}},
/* 52509 */ {(13<<2)|2,{75,98,0}},
/* 52510 */ {(13<<2)|2,{75,98,0}},
/* 52511 */ {(13<<2)|2,{75,98,0}},
/* 52512 */ {(13<<2)|2,{75,100,0}},
/* 52513 */ {(13<<2)|2,{75,100,0}},
/* 52514 */ {(13<<2)|2,{75,100,0}},
/* 52515 */ {(13<<2)|2,{75,100,0}},
/* 52516 */ {(13<<2)|2,{75,100,0}},
/* 52517 */ {(13<<2)|2,{75,100,0}},
/* 52518 */ {(13<<2)|2,{75,100,0}},
/* 52519 */ {(13<<2)|2,{75,100,0}},
/* 52520 */ {(13<<2)|2,{75,102,0}},
/* 52521 */ {(13<<2)|2,{75,102,0}},
/* 52522 */ {(13<<2)|2,{75,102,0}},
/* 52523 */ {(13<<2)|2,{75,102,0}},
/* 52524 */ {(13<<2)|2,{75,102,0}},
/* 52525 */ {(13<<2)|2,{75,102,0}},
/* 52526 */ {(13<<2)|2,{75,102,0}},
/* 52527 */ {(13<<2)|2,{75,102,0}},
/* 52528 */ {(13<<2)|2,{75,103,0}},
/* 52529 */ {(13<<2)|2,{75,103,0}},
/* 52530 */ {(13<<2)|2,{75,103,0}},
/* 52531 */ {(13<<2)|2,{75,103,0}},
/* 52532 */ {(13<<2)|2,{75,103,0}},
/* 52533 */ {(13<<2)|2,{75,103,0}},
/* 52534 */ {(13<<2)|2,{75,103,0}},
/* 52535 */ {(13<<2)|2,{75,103,0}},
/* 52536 */ {(13<<2)|2,{75,104,0}},
/* 52537 */ {(13<<2)|2,{75,104,0}},
/* 52538 */ {(13<<2)|2,{75,104,0}},
/* 52539 */ {(13<<2)|2,{75,104,0}},
/* 52540 */ {(13<<2)|2,{75,104,0}},
/* 52541 */ {(13<<2)|2,{75,104,0}},
/* 52542 */ {(13<<2)|2,{75,104,0}},
/* 52543 */ {(13<<2)|2,{75,104,0}},
/* 52544 */ {(13<<2)|2,{75,108,0}},
/* 52545 */ {(13<<2)|2,{75,108,0}},
/* 52546 */ {(13<<2)|2,{75,108,0}},
/* 52547 */ {(13<<2)|2,{75,108,0}},
/* 52548 */ {(13<<2)|2,{75,108,0}},
/* 52549 */ {(13<<2)|2,{75,108,0}},
/* 52550 */ {(13<<2)|2,{75,108,0}},
/* 52551 */ {(13<<2)|2,{75,108,0}},
/* 52552 */ {(13<<2)|2,{75,109,0}},
/* 52553 */ {(13<<2)|2,{75,109,0}},
/* 52554 */ {(13<<2)|2,{75,109,0}},
/* 52555 */ {(13<<2)|2,{75,109,0}},
/* 52556 */ {(13<<2)|2,{75,109,0}},
/* 52557 */ {(13<<2)|2,{75,109,0}},
/* 52558 */ {(13<<2)|2,{75,109,0}},
/* 52559 */ {(13<<2)|2,{75,109,0}},
/* 52560 */ {(13<<2)|2,{75,110,0}},
/* 52561 */ {(13<<2)|2,{75,110,0}},
/* 52562 */ {(13<<2)|2,{75,110,0}},
/* 52563 */ {(13<<2)|2,{75,110,0}},
/* 52564 */ {(13<<2)|2,{75,110,0}},
/* 52565 */ {(13<<2)|2,{75,110,0}},
/* 52566 */ {(13<<2)|2,{75,110,0}},
/* 52567 */ {(13<<2)|2,{75,110,0}},
/* 52568 */ {(13<<2)|2,{75,112,0}},
/* 52569 */ {(13<<2)|2,{75,112,0}},
/* 52570 */ {(13<<2)|2,{75,112,0}},
/* 52571 */ {(13<<2)|2,{75,112,0}},
/* 52572 */ {(13<<2)|2,{75,112,0}},
/* 52573 */ {(13<<2)|2,{75,112,0}},
/* 52574 */ {(13<<2)|2,{75,112,0}},
/* 52575 */ {(13<<2)|2,{75,112,0}},
/* 52576 */ {(13<<2)|2,{75,114,0}},
/* 52577 */ {(13<<2)|2,{75,114,0}},
/* 52578 */ {(13<<2)|2,{75,114,0}},
/* 52579 */ {(13<<2)|2,{75,114,0}},
/* 52580 */ {(13<<2)|2,{75,114,0}},
/* 52581 */ {(13<<2)|2,{75,114,0}},
/* 52582 */ {(13<<2)|2,{75,114,0}},
/* 52583 */ {(13<<2)|2,{75,114,0}},
/* 52584 */ {(13<<2)|2,{75,117,0}},
/* 52585 */ {(13<<2)|2,{75,117,0}},
/* 52586 */ {(13<<2)|2,{75,117,0}},
/* 52587 */ {(13<<2)|2,{75,117,0}},
/* 52588 */ {(13<<2)|2,{75,117,0}},
/* 52589 */ {(13<<2)|2,{75,117,0}},
/* 52590 */ {(13<<2)|2,{75,117,0}},
/* 52591 */ {(13<<2)|2,{75,117,0}},
/* 52592 */ {(14<<2)|2,{75,58,0}},
/* 52593 */ {(14<<2)|2,{75,58,0}},
/* 52594 */ {(14<<2)|2,{75,58,0}},
/* 52595 */ {(14<<2)|2,{75,58,0}},
/* 52596 */ {(14<<2)|2,{75,66,0}},
/* 52597 */ {(14<<2)|2,{75,66,0}},
/* 52598 */ {(14<<2)|2,{75,66,0}},
/* 52599 */ {(14<<2)|2,{75,66,0}},
/* 52600 */ {(14<<2)|2,{75,67,0}},
/* 52601 */ {(14<<2)|2,{75,67,0}},
/* 52602 */ {(14<<2)|2,{75,67,0}},
/* 52603 */ {(14<<2)|2,{75,67,0}},
/* 52604 */ {(14<<2)|2,{75,68,0}},
/* 52605 */ {(14<<2)|2,{75,68,0}},
/* 52606 */ {(14<<2)|2,{75,68,0}},
/* 52607 */ {(14<<2)|2,{75,68,0}},
/* 52608 */ {(14<<2)|2,{75,69,0}},
/* 52609 */ {(14<<2)|2,{75,69,0}},
/* 52610 */ {(14<<2)|2,{75,69,0}},
/* 52611 */ {(14<<2)|2,{75,69,0}},
/* 52612 */ {(14<<2)|2,{75,70,0}},
/* 52613 */ {(14<<2)|2,{75,70,0}},
/* 52614 */ {(14<<2)|2,{75,70,0}},
/* 52615 */ {(14<<2)|2,{75,70,0}},
/* 52616 */ {(14<<2)|2,{75,71,0}},
/* 52617 */ {(14<<2)|2,{75,71,0}},
/* 52618 */ {(14<<2)|2,{75,71,0}},
/* 52619 */ {(14<<2)|2,{75,71,0}},
/* 52620 */ {(14<<2)|2,{75,72,0}},
/* 52621 */ {(14<<2)|2,{75,72,0}},
/* 52622 */ {(14<<2)|2,{75,72,0}},
/* 52623 */ {(14<<2)|2,{75,72,0}},
/* 52624 */ {(14<<2)|2,{75,73,0}},
/* 52625 */ {(14<<2)|2,{75,73,0}},
/* 52626 */ {(14<<2)|2,{75,73,0}},
/* 52627 */ {(14<<2)|2,{75,73,0}},
/* 52628 */ {(14<<2)|2,{75,74,0}},
/* 52629 */ {(14<<2)|2,{75,74,0}},
/* 52630 */ {(14<<2)|2,{75,74,0}},
/* 52631 */ {(14<<2)|2,{75,74,0}},
/* 52632 */ {(14<<2)|2,{75,75,0}},
/* 52633 */ {(14<<2)|2,{75,75,0}},
/* 52634 */ {(14<<2)|2,{75,75,0}},
/* 52635 */ {(14<<2)|2,{75,75,0}},
/* 52636 */ {(14<<2)|2,{75,76,0}},
/* 52637 */ {(14<<2)|2,{75,76,0}},
/* 52638 */ {(14<<2)|2,{75,76,0}},
/* 52639 */ {(14<<2)|2,{75,76,0}},
/* 52640 */ {(14<<2)|2,{75,77,0}},
/* 52641 */ {(14<<2)|2,{75,77,0}},
/* 52642 */ {(14<<2)|2,{75,77,0}},
/* 52643 */ {(14<<2)|2,{75,77,0}},
/* 52644 */ {(14<<2)|2,{75,78,0}},
/* 52645 */ {(14<<2)|2,{75,78,0}},
/* 52646 */ {(14<<2)|2,{75,78,0}},
/* 52647 */ {(14<<2)|2,{75,78,0}},
/* 52648 */ {(14<<2)|2,{75,79,0}},
/* 52649 */ {(14<<2)|2,{75,79,0}},
/* 52650 */ {(14<<2)|2,{75,79,0}},
/* 52651 */ {(14<<2)|2,{75,79,0}},
/* 52652 */ {(14<<2)|2,{75,80,0}},
/* 52653 */ {(14<<2)|2,{75,80,0}},
/* 52654 */ {(14<<2)|2,{75,80,0}},
/* 52655 */ {(14<<2)|2,{75,80,0}},
/* 52656 */ {(14<<2)|2,{75,81,0}},
/* 52657 */ {(14<<2)|2,{75,81,0}},
/* 52658 */ {(14<<2)|2,{75,81,0}},
/* 52659 */ {(14<<2)|2,{75,81,0}},
/* 52660 */ {(14<<2)|2,{75,82,0}},
/* 52661 */ {(14<<2)|2,{75,82,0}},
/* 52662 */ {(14<<2)|2,{75,82,0}},
/* 52663 */ {(14<<2)|2,{75,82,0}},
/* 52664 */ {(14<<2)|2,{75,83,0}},
/* 52665 */ {(14<<2)|2,{75,83,0}},
/* 52666 */ {(14<<2)|2,{75,83,0}},
/* 52667 */ {(14<<2)|2,{75,83,0}},
/* 52668 */ {(14<<2)|2,{75,84,0}},
/* 52669 */ {(14<<2)|2,{75,84,0}},
/* 52670 */ {(14<<2)|2,{75,84,0}},
/* 52671 */ {(14<<2)|2,{75,84,0}},
/* 52672 */ {(14<<2)|2,{75,85,0}},
/* 52673 */ {(14<<2)|2,{75,85,0}},
/* 52674 */ {(14<<2)|2,{75,85,0}},
/* 52675 */ {(14<<2)|2,{75,85,0}},
/* 52676 */ {(14<<2)|2,{75,86,0}},
/* 52677 */ {(14<<2)|2,{75,86,0}},
/* 52678 */ {(14<<2)|2,{75,86,0}},
/* 52679 */ {(14<<2)|2,{75,86,0}},
/* 52680 */ {(14<<2)|2,{75,87,0}},
/* 52681 */ {(14<<2)|2,{75,87,0}},
/* 52682 */ {(14<<2)|2,{75,87,0}},
/* 52683 */ {(14<<2)|2,{75,87,0}},
/* 52684 */ {(14<<2)|2,{75,89,0}},
/* 52685 */ {(14<<2)|2,{75,89,0}},
/* 52686 */ {(14<<2)|2,{75,89,0}},
/* 52687 */ {(14<<2)|2,{75,89,0}},
/* 52688 */ {(14<<2)|2,{75,106,0}},
/* 52689 */ {(14<<2)|2,{75,106,0}},
/* 52690 */ {(14<<2)|2,{75,106,0}},
/* 52691 */ {(14<<2)|2,{75,106,0}},
/* 52692 */ {(14<<2)|2,{75,107,0}},
/* 52693 */ {(14<<2)|2,{75,107,0}},
/* 52694 */ {(14<<2)|2,{75,107,0}},
/* 52695 */ {(14<<2)|2,{75,107,0}},
/* 52696 */ {(14<<2)|2,{75,113,0}},
/* 52697 */ {(14<<2)|2,{75,113,0}},
/* 52698 */ {(14<<2)|2,{75,113,0}},
/* 52699 */ {(14<<2)|2,{75,113,0}},
/* 52700 */ {(14<<2)|2,{75,118,0}},
/* 52701 */ {(14<<2)|2,{75,118,0}},
/* 52702 */ {(14<<2)|2,{75,118,0}},
/* 52703 */ {(14<<2)|2,{75,118,0}},
/* 52704 */ {(14<<2)|2,{75,119,0}},
/* 52705 */ {(14<<2)|2,{75,119,0}},
/* 52706 */ {(14<<2)|2,{75,119,0}},
/* 52707 */ {(14<<2)|2,{75,119,0}},
/* 52708 */ {(14<<2)|2,{75,120,0}},
/* 52709 */ {(14<<2)|2,{75,120,0}},
/* 52710 */ {(14<<2)|2,{75,120,0}},
/* 52711 */ {(14<<2)|2,{75,120,0}},
/* 52712 */ {(14<<2)|2,{75,121,0}},
/* 52713 */ {(14<<2)|2,{75,121,0}},
/* 52714 */ {(14<<2)|2,{75,121,0}},
/* 52715 */ {(14<<2)|2,{75,121,0}},
/* 52716 */ {(14<<2)|2,{75,122,0}},
/* 52717 */ {(14<<2)|2,{75,122,0}},
/* 52718 */ {(14<<2)|2,{75,122,0}},
/* 52719 */ {(14<<2)|2,{75,122,0}},
/* 52720 */ {(15<<2)|2,{75,38,0}},
/* 52721 */ {(15<<2)|2,{75,38,0}},
/* 52722 */ {(15<<2)|2,{75,42,0}},
/* 52723 */ {(15<<2)|2,{75,42,0}},
/* 52724 */ {(15<<2)|2,{75,44,0}},
/* 52725 */ {(15<<2)|2,{75,44,0}},
/* 52726 */ {(15<<2)|2,{75,59,0}},
/* 52727 */ {(15<<2)|2,{75,59,0}},
/* 52728 */ {(15<<2)|2,{75,88,0}},
/* 52729 */ {(15<<2)|2,{75,88,0}},
/* 52730 */ {(15<<2)|2,{75,90,0}},
/* 52731 */ {(15<<2)|2,{75,90,0}},
/* 52732 */ {(7<<2)|1,{75,0,0}},
/* 52733 */ {(7<<2)|1,{75,0,0}},
/* 52734 */ {(7<<2)|1,{75,0,0}},
/* 52735 */ {(7<<2)|1,{75,0,0}},
/* 52736 */ {(12<<2)|2,{76,48,0}},
/* 52737 */ {(12<<2)|2,{76,48,0}},
/* 52738 */ {(12<<2)|2,{76,48,0}},
/* 52739 */ {(12<<2)|2,{76,48,0}},
/* 52740 */ {(12<<2)|2,{76,48,0}},
/* 52741 */ {(12<<2)|2,{76,48,0}},
/* 52742 */ {(12<<2)|2,{76,48,0}},
/* 52743 */ {(12<<2)|2,{76,48,0}},
/* 52744 */ {(12<<2)|2,{76,48,0}},
/* 52745 */ {(12<<2)|2,{76,48,0}},
/* 52746 */ {(12<<2)|2,{76,48,0}},
/* 52747 */ {(12<<2)|2,{76,48,0}},
/* 52748 */ {(12<<2)|2,{76,48,0}},
/* 52749 */ {(12<<2)|2,{76,48,0}},
/* 52750 */ {(12<<2)|2,{76,48,0}},
/* 52751 */ {(12<<2)|2,{76,48,0}},
/* 52752 */ {(12<<2)|2,{76,49,0}},
/* 52753 */ {(12<<2)|2,{76,49,0}},
/* 52754 */ {(12<<2)|2,{76,49,0}},
/* 52755 */ {(12<<2)|2,{76,49,0}},
/* 52756 */ {(12<<2)|2,{76,49,0}},
/* 52757 */ {(12<<2)|2,{76,49,0}},
/* 52758 */ {(12<<2)|2,{76,49,0}},
/* 52759 */ {(12<<2)|2,{76,49,0}},
/* 52760 */ {(12<<2)|2,{76,49,0}},
/* 52761 */ {(12<<2)|2,{76,49,0}},
/* 52762 */ {(12<<2)|2,{76,49,0}},
/* 52763 */ {(12<<2)|2,{76,49,0}},
/* 52764 */ {(12<<2)|2,{76,49,0}},
/* 52765 */ {(12<<2)|2,{76,49,0}},
/* 52766 */ {(12<<2)|2,{76,49,0}},
/* 52767 */ {(12<<2)|2,{76,49,0}},
/* 52768 */ {(12<<2)|2,{76,50,0}},
/* 52769 */ {(12<<2)|2,{76,50,0}},
/* 52770 */ {(12<<2)|2,{76,50,0}},
/* 52771 */ {(12<<2)|2,{76,50,0}},
/* 52772 */ {(12<<2)|2,{76,50,0}},
/* 52773 */ {(12<<2)|2,{76,50,0}},
/* 52774 */ {(12<<2)|2,{76,50,0}},
/* 52775 */ {(12<<2)|2,{76,50,0}},
/* 52776 */ {(12<<2)|2,{76,50,0}},
/* 52777 */ {(12<<2)|2,{76,50,0}},
/* 52778 */ {(12<<2)|2,{76,50,0}},
/* 52779 */ {(12<<2)|2,{76,50,0}},
/* 52780 */ {(12<<2)|2,{76,50,0}},
/* 52781 */ {(12<<2)|2,{76,50,0}},
/* 52782 */ {(12<<2)|2,{76,50,0}},
/* 52783 */ {(12<<2)|2,{76,50,0}},
/* 52784 */ {(12<<2)|2,{76,97,0}},
/* 52785 */ {(12<<2)|2,{76,97,0}},
/* 52786 */ {(12<<2)|2,{76,97,0}},
/* 52787 */ {(12<<2)|2,{76,97,0}},
/* 52788 */ {(12<<2)|2,{76,97,0}},
/* 52789 */ {(12<<2)|2,{76,97,0}},
/* 52790 */ {(12<<2)|2,{76,97,0}},
/* 52791 */ {(12<<2)|2,{76,97,0}},
/* 52792 */ {(12<<2)|2,{76,97,0}},
/* 52793 */ {(12<<2)|2,{76,97,0}},
/* 52794 */ {(12<<2)|2,{76,97,0}},
/* 52795 */ {(12<<2)|2,{76,97,0}},
/* 52796 */ {(12<<2)|2,{76,97,0}},
/* 52797 */ {(12<<2)|2,{76,97,0}},
/* 52798 */ {(12<<2)|2,{76,97,0}},
/* 52799 */ {(12<<2)|2,{76,97,0}},
/* 52800 */ {(12<<2)|2,{76,99,0}},
/* 52801 */ {(12<<2)|2,{76,99,0}},
/* 52802 */ {(12<<2)|2,{76,99,0}},
/* 52803 */ {(12<<2)|2,{76,99,0}},
/* 52804 */ {(12<<2)|2,{76,99,0}},
/* 52805 */ {(12<<2)|2,{76,99,0}},
/* 52806 */ {(12<<2)|2,{76,99,0}},
/* 52807 */ {(12<<2)|2,{76,99,0}},
/* 52808 */ {(12<<2)|2,{76,99,0}},
/* 52809 */ {(12<<2)|2,{76,99,0}},
/* 52810 */ {(12<<2)|2,{76,99,0}},
/* 52811 */ {(12<<2)|2,{76,99,0}},
/* 52812 */ {(12<<2)|2,{76,99,0}},
/* 52813 */ {(12<<2)|2,{76,99,0}},
/* 52814 */ {(12<<2)|2,{76,99,0}},
/* 52815 */ {(12<<2)|2,{76,99,0}},
/* 52816 */ {(12<<2)|2,{76,101,0}},
/* 52817 */ {(12<<2)|2,{76,101,0}},
/* 52818 */ {(12<<2)|2,{76,101,0}},
/* 52819 */ {(12<<2)|2,{76,101,0}},
/* 52820 */ {(12<<2)|2,{76,101,0}},
/* 52821 */ {(12<<2)|2,{76,101,0}},
/* 52822 */ {(12<<2)|2,{76,101,0}},
/* 52823 */ {(12<<2)|2,{76,101,0}},
/* 52824 */ {(12<<2)|2,{76,101,0}},
/* 52825 */ {(12<<2)|2,{76,101,0}},
/* 52826 */ {(12<<2)|2,{76,101,0}},
/* 52827 */ {(12<<2)|2,{76,101,0}},
/* 52828 */ {(12<<2)|2,{76,101,0}},
/* 52829 */ {(12<<2)|2,{76,101,0}},
/* 52830 */ {(12<<2)|2,{76,101,0}},
/* 52831 */ {(12<<2)|2,{76,101,0}},
/* 52832 */ {(12<<2)|2,{76,105,0}},
/* 52833 */ {(12<<2)|2,{76,105,0}},
/* 52834 */ {(12<<2)|2,{76,105,0}},
/* 52835 */ {(12<<2)|2,{76,105,0}},
/* 52836 */ {(12<<2)|2,{76,105,0}},
/* 52837 */ {(12<<2)|2,{76,105,0}},
/* 52838 */ {(12<<2)|2,{76,105,0}},
/* 52839 */ {(12<<2)|2,{76,105,0}},
/* 52840 */ {(12<<2)|2,{76,105,0}},
/* 52841 */ {(12<<2)|2,{76,105,0}},
/* 52842 */ {(12<<2)|2,{76,105,0}},
/* 52843 */ {(12<<2)|2,{76,105,0}},
/* 52844 */ {(12<<2)|2,{76,105,0}},
/* 52845 */ {(12<<2)|2,{76,105,0}},
/* 52846 */ {(12<<2)|2,{76,105,0}},
/* 52847 */ {(12<<2)|2,{76,105,0}},
/* 52848 */ {(12<<2)|2,{76,111,0}},
/* 52849 */ {(12<<2)|2,{76,111,0}},
/* 52850 */ {(12<<2)|2,{76,111,0}},
/* 52851 */ {(12<<2)|2,{76,111,0}},
/* 52852 */ {(12<<2)|2,{76,111,0}},
/* 52853 */ {(12<<2)|2,{76,111,0}},
/* 52854 */ {(12<<2)|2,{76,111,0}},
/* 52855 */ {(12<<2)|2,{76,111,0}},
/* 52856 */ {(12<<2)|2,{76,111,0}},
/* 52857 */ {(12<<2)|2,{76,111,0}},
/* 52858 */ {(12<<2)|2,{76,111,0}},
/* 52859 */ {(12<<2)|2,{76,111,0}},
/* 52860 */ {(12<<2)|2,{76,111,0}},
/* 52861 */ {(12<<2)|2,{76,111,0}},
/* 52862 */ {(12<<2)|2,{76,111,0}},
/* 52863 */ {(12<<2)|2,{76,111,0}},
/* 52864 */ {(12<<2)|2,{76,115,0}},
/* 52865 */ {(12<<2)|2,{76,115,0}},
/* 52866 */ {(12<<2)|2,{76,115,0}},
/* 52867 */ {(12<<2)|2,{76,115,0}},
/* 52868 */ {(12<<2)|2,{76,115,0}},
/* 52869 */ {(12<<2)|2,{76,115,0}},
/* 52870 */ {(12<<2)|2,{76,115,0}},
/* 52871 */ {(12<<2)|2,{76,115,0}},
/* 52872 */ {(12<<2)|2,{76,115,0}},
/* 52873 */ {(12<<2)|2,{76,115,0}},
/* 52874 */ {(12<<2)|2,{76,115,0}},
/* 52875 */ {(12<<2)|2,{76,115,0}},
/* 52876 */ {(12<<2)|2,{76,115,0}},
/* 52877 */ {(12<<2)|2,{76,115,0}},
/* 52878 */ {(12<<2)|2,{76,115,0}},
/* 52879 */ {(12<<2)|2,{76,115,0}},
/* 52880 */ {(12<<2)|2,{76,116,0}},
/* 52881 */ {(12<<2)|2,{76,116,0}},
/* 52882 */ {(12<<2)|2,{76,116,0}},
/* 52883 */ {(12<<2)|2,{76,116,0}},
/* 52884 */ {(12<<2)|2,{76,116,0}},
/* 52885 */ {(12<<2)|2,{76,116,0}},
/* 52886 */ {(12<<2)|2,{76,116,0}},
/* 52887 */ {(12<<2)|2,{76,116,0}},
/* 52888 */ {(12<<2)|2,{76,116,0}},
/* 52889 */ {(12<<2)|2,{76,116,0}},
/* 52890 */ {(12<<2)|2,{76,116,0}},
/* 52891 */ {(12<<2)|2,{76,116,0}},
/* 52892 */ {(12<<2)|2,{76,116,0}},
/* 52893 */ {(12<<2)|2,{76,116,0}},
/* 52894 */ {(12<<2)|2,{76,116,0}},
/* 52895 */ {(12<<2)|2,{76,116,0}},
/* 52896 */ {(13<<2)|2,{76,32,0}},
/* 52897 */ {(13<<2)|2,{76,32,0}},
/* 52898 */ {(13<<2)|2,{76,32,0}},
/* 52899 */ {(13<<2)|2,{76,32,0}},
/* 52900 */ {(13<<2)|2,{76,32,0}},
/* 52901 */ {(13<<2)|2,{76,32,0}},
/* 52902 */ {(13<<2)|2,{76,32,0}},
/* 52903 */ {(13<<2)|2,{76,32,0}},
/* 52904 */ {(13<<2)|2,{76,37,0}},
/* 52905 */ {(13<<2)|2,{76,37,0}},
/* 52906 */ {(13<<2)|2,{76,37,0}},
/* 52907 */ {(13<<2)|2,{76,37,0}},
/* 52908 */ {(13<<2)|2,{76,37,0}},
/* 52909 */ {(13<<2)|2,{76,37,0}},
/* 52910 */ {(13<<2)|2,{76,37,0}},
/* 52911 */ {(13<<2)|2,{76,37,0}},
/* 52912 */ {(13<<2)|2,{76,45,0}},
/* 52913 */ {(13<<2)|2,{76,45,0}},
/* 52914 */ {(13<<2)|2,{76,45,0}},
/* 52915 */ {(13<<2)|2,{76,45,0}},
/* 52916 */ {(13<<2)|2,{76,45,0}},
/* 52917 */ {(13<<2)|2,{76,45,0}},
/* 52918 */ {(13<<2)|2,{76,45,0}},
/* 52919 */ {(13<<2)|2,{76,45,0}},
/* 52920 */ {(13<<2)|2,{76,46,0}},
/* 52921 */ {(13<<2)|2,{76,46,0}},
/* 52922 */ {(13<<2)|2,{76,46,0}},
/* 52923 */ {(13<<2)|2,{76,46,0}},
/* 52924 */ {(13<<2)|2,{76,46,0}},
/* 52925 */ {(13<<2)|2,{76,46,0}},
/* 52926 */ {(13<<2)|2,{76,46,0}},
/* 52927 */ {(13<<2)|2,{76,46,0}},
/* 52928 */ {(13<<2)|2,{76,47,0}},
/* 52929 */ {(13<<2)|2,{76,47,0}},
/* 52930 */ {(13<<2)|2,{76,47,0}},
/* 52931 */ {(13<<2)|2,{76,47,0}},
/* 52932 */ {(13<<2)|2,{76,47,0}},
/* 52933 */ {(13<<2)|2,{76,47,0}},
/* 52934 */ {(13<<2)|2,{76,47,0}},
/* 52935 */ {(13<<2)|2,{76,47,0}},
/* 52936 */ {(13<<2)|2,{76,51,0}},
/* 52937 */ {(13<<2)|2,{76,51,0}},
/* 52938 */ {(13<<2)|2,{76,51,0}},
/* 52939 */ {(13<<2)|2,{76,51,0}},
/* 52940 */ {(13<<2)|2,{76,51,0}},
/* 52941 */ {(13<<2)|2,{76,51,0}},
/* 52942 */ {(13<<2)|2,{76,51,0}},
/* 52943 */ {(13<<2)|2,{76,51,0}},
/* 52944 */ {(13<<2)|2,{76,52,0}},
/* 52945 */ {(13<<2)|2,{76,52,0}},
/* 52946 */ {(13<<2)|2,{76,52,0}},
/* 52947 */ {(13<<2)|2,{76,52,0}},
/* 52948 */ {(13<<2)|2,{76,52,0}},
/* 52949 */ {(13<<2)|2,{76,52,0}},
/* 52950 */ {(13<<2)|2,{76,52,0}},
/* 52951 */ {(13<<2)|2,{76,52,0}},
/* 52952 */ {(13<<2)|2,{76,53,0}},
/* 52953 */ {(13<<2)|2,{76,53,0}},
/* 52954 */ {(13<<2)|2,{76,53,0}},
/* 52955 */ {(13<<2)|2,{76,53,0}},
/* 52956 */ {(13<<2)|2,{76,53,0}},
/* 52957 */ {(13<<2)|2,{76,53,0}},
/* 52958 */ {(13<<2)|2,{76,53,0}},
/* 52959 */ {(13<<2)|2,{76,53,0}},
/* 52960 */ {(13<<2)|2,{76,54,0}},
/* 52961 */ {(13<<2)|2,{76,54,0}},
/* 52962 */ {(13<<2)|2,{76,54,0}},
/* 52963 */ {(13<<2)|2,{76,54,0}},
/* 52964 */ {(13<<2)|2,{76,54,0}},
/* 52965 */ {(13<<2)|2,{76,54,0}},
/* 52966 */ {(13<<2)|2,{76,54,0}},
/* 52967 */ {(13<<2)|2,{76,54,0}},
/* 52968 */ {(13<<2)|2,{76,55,0}},
/* 52969 */ {(13<<2)|2,{76,55,0}},
/* 52970 */ {(13<<2)|2,{76,55,0}},
/* 52971 */ {(13<<2)|2,{76,55,0}},
/* 52972 */ {(13<<2)|2,{76,55,0}},
/* 52973 */ {(13<<2)|2,{76,55,0}},
/* 52974 */ {(13<<2)|2,{76,55,0}},
/* 52975 */ {(13<<2)|2,{76,55,0}},
/* 52976 */ {(13<<2)|2,{76,56,0}},
/* 52977 */ {(13<<2)|2,{76,56,0}},
/* 52978 */ {(13<<2)|2,{76,56,0}},
/* 52979 */ {(13<<2)|2,{76,56,0}},
/* 52980 */ {(13<<2)|2,{76,56,0}},
/* 52981 */ {(13<<2)|2,{76,56,0}},
/* 52982 */ {(13<<2)|2,{76,56,0}},
/* 52983 */ {(13<<2)|2,{76,56,0}},
/* 52984 */ {(13<<2)|2,{76,57,0}},
/* 52985 */ {(13<<2)|2,{76,57,0}},
/* 52986 */ {(13<<2)|2,{76,57,0}},
/* 52987 */ {(13<<2)|2,{76,57,0}},
/* 52988 */ {(13<<2)|2,{76,57,0}},
/* 52989 */ {(13<<2)|2,{76,57,0}},
/* 52990 */ {(13<<2)|2,{76,57,0}},
/* 52991 */ {(13<<2)|2,{76,57,0}},
/* 52992 */ {(13<<2)|2,{76,61,0}},
/* 52993 */ {(13<<2)|2,{76,61,0}},
/* 52994 */ {(13<<2)|2,{76,61,0}},
/* 52995 */ {(13<<2)|2,{76,61,0}},
/* 52996 */ {(13<<2)|2,{76,61,0}},
/* 52997 */ {(13<<2)|2,{76,61,0}},
/* 52998 */ {(13<<2)|2,{76,61,0}},
/* 52999 */ {(13<<2)|2,{76,61,0}},
/* 53000 */ {(13<<2)|2,{76,65,0}},
/* 53001 */ {(13<<2)|2,{76,65,0}},
/* 53002 */ {(13<<2)|2,{76,65,0}},
/* 53003 */ {(13<<2)|2,{76,65,0}},
/* 53004 */ {(13<<2)|2,{76,65,0}},
/* 53005 */ {(13<<2)|2,{76,65,0}},
/* 53006 */ {(13<<2)|2,{76,65,0}},
/* 53007 */ {(13<<2)|2,{76,65,0}},
/* 53008 */ {(13<<2)|2,{76,95,0}},
/* 53009 */ {(13<<2)|2,{76,95,0}},
/* 53010 */ {(13<<2)|2,{76,95,0}},
/* 53011 */ {(13<<2)|2,{76,95,0}},
/* 53012 */ {(13<<2)|2,{76,95,0}},
/* 53013 */ {(13<<2)|2,{76,95,0}},
/* 53014 */ {(13<<2)|2,{76,95,0}},
/* 53015 */ {(13<<2)|2,{76,95,0}},
/* 53016 */ {(13<<2)|2,{76,98,0}},
/* 53017 */ {(13<<2)|2,{76,98,0}},
/* 53018 */ {(13<<2)|2,{76,98,0}},
/* 53019 */ {(13<<2)|2,{76,98,0}},
/* 53020 */ {(13<<2)|2,{76,98,0}},
/* 53021 */ {(13<<2)|2,{76,98,0}},
/* 53022 */ {(13<<2)|2,{76,98,0}},
/* 53023 */ {(13<<2)|2,{76,98,0}},
/* 53024 */ {(13<<2)|2,{76,100,0}},
/* 53025 */ {(13<<2)|2,{76,100,0}},
/* 53026 */ {(13<<2)|2,{76,100,0}},
/* 53027 */ {(13<<2)|2,{76,100,0}},
/* 53028 */ {(13<<2)|2,{76,100,0}},
/* 53029 */ {(13<<2)|2,{76,100,0}},
/* 53030 */ {(13<<2)|2,{76,100,0}},
/* 53031 */ {(13<<2)|2,{76,100,0}},
/* 53032 */ {(13<<2)|2,{76,102,0}},
/* 53033 */ {(13<<2)|2,{76,102,0}},
/* 53034 */ {(13<<2)|2,{76,102,0}},
/* 53035 */ {(13<<2)|2,{76,102,0}},
/* 53036 */ {(13<<2)|2,{76,102,0}},
/* 53037 */ {(13<<2)|2,{76,102,0}},
/* 53038 */ {(13<<2)|2,{76,102,0}},
/* 53039 */ {(13<<2)|2,{76,102,0}},
/* 53040 */ {(13<<2)|2,{76,103,0}},
/* 53041 */ {(13<<2)|2,{76,103,0}},
/* 53042 */ {(13<<2)|2,{76,103,0}},
/* 53043 */ {(13<<2)|2,{76,103,0}},
/* 53044 */ {(13<<2)|2,{76,103,0}},
/* 53045 */ {(13<<2)|2,{76,103,0}},
/* 53046 */ {(13<<2)|2,{76,103,0}},
/* 53047 */ {(13<<2)|2,{76,103,0}},
/* 53048 */ {(13<<2)|2,{76,104,0}},
/* 53049 */ {(13<<2)|2,{76,104,0}},
/* 53050 */ {(13<<2)|2,{76,104,0}},
/* 53051 */ {(13<<2)|2,{76,104,0}},
/* 53052 */ {(13<<2)|2,{76,104,0}},
/* 53053 */ {(13<<2)|2,{76,104,0}},
/* 53054 */ {(13<<2)|2,{76,104,0}},
/* 53055 */ {(13<<2)|2,{76,104,0}},
/* 53056 */ {(13<<2)|2,{76,108,0}},
/* 53057 */ {(13<<2)|2,{76,108,0}},
/* 53058 */ {(13<<2)|2,{76,108,0}},
/* 53059 */ {(13<<2)|2,{76,108,0}},
/* 53060 */ {(13<<2)|2,{76,108,0}},
/* 53061 */ {(13<<2)|2,{76,108,0}},
/* 53062 */ {(13<<2)|2,{76,108,0}},
/* 53063 */ {(13<<2)|2,{76,108,0}},
/* 53064 */ {(13<<2)|2,{76,109,0}},
/* 53065 */ {(13<<2)|2,{76,109,0}},
/* 53066 */ {(13<<2)|2,{76,109,0}},
/* 53067 */ {(13<<2)|2,{76,109,0}},
/* 53068 */ {(13<<2)|2,{76,109,0}},
/* 53069 */ {(13<<2)|2,{76,109,0}},
/* 53070 */ {(13<<2)|2,{76,109,0}},
/* 53071 */ {(13<<2)|2,{76,109,0}},
/* 53072 */ {(13<<2)|2,{76,110,0}},
/* 53073 */ {(13<<2)|2,{76,110,0}},
/* 53074 */ {(13<<2)|2,{76,110,0}},
/* 53075 */ {(13<<2)|2,{76,110,0}},
/* 53076 */ {(13<<2)|2,{76,110,0}},
/* 53077 */ {(13<<2)|2,{76,110,0}},
/* 53078 */ {(13<<2)|2,{76,110,0}},
/* 53079 */ {(13<<2)|2,{76,110,0}},
/* 53080 */ {(13<<2)|2,{76,112,0}},
/* 53081 */ {(13<<2)|2,{76,112,0}},
/* 53082 */ {(13<<2)|2,{76,112,0}},
/* 53083 */ {(13<<2)|2,{76,112,0}},
/* 53084 */ {(13<<2)|2,{76,112,0}},
/* 53085 */ {(13<<2)|2,{76,112,0}},
/* 53086 */ {(13<<2)|2,{76,112,0}},
/* 53087 */ {(13<<2)|2,{76,112,0}},
/* 53088 */ {(13<<2)|2,{76,114,0}},
/* 53089 */ {(13<<2)|2,{76,114,0}},
/* 53090 */ {(13<<2)|2,{76,114,0}},
/* 53091 */ {(13<<2)|2,{76,114,0}},
/* 53092 */ {(13<<2)|2,{76,114,0}},
/* 53093 */ {(13<<2)|2,{76,114,0}},
/* 53094 */ {(13<<2)|2,{76,114,0}},
/* 53095 */ {(13<<2)|2,{76,114,0}},
/* 53096 */ {(13<<2)|2,{76,117,0}},
/* 53097 */ {(13<<2)|2,{76,117,0}},
/* 53098 */ {(13<<2)|2,{76,117,0}},
/* 53099 */ {(13<<2)|2,{76,117,0}},
/* 53100 */ {(13<<2)|2,{76,117,0}},
/* 53101 */ {(13<<2)|2,{76,117,0}},
/* 53102 */ {(13<<2)|2,{76,117,0}},
/* 53103 */ {(13<<2)|2,{76,117,0}},
/* 53104 */ {(14<<2)|2,{76,58,0}},
/* 53105 */ {(14<<2)|2,{76,58,0}},
/* 53106 */ {(14<<2)|2,{76,58,0}},
/* 53107 */ {(14<<2)|2,{76,58,0}},
/* 53108 */ {(14<<2)|2,{76,66,0}},
/* 53109 */ {(14<<2)|2,{76,66,0}},
/* 53110 */ {(14<<2)|2,{76,66,0}},
/* 53111 */ {(14<<2)|2,{76,66,0}},
/* 53112 */ {(14<<2)|2,{76,67,0}},
/* 53113 */ {(14<<2)|2,{76,67,0}},
/* 53114 */ {(14<<2)|2,{76,67,0}},
/* 53115 */ {(14<<2)|2,{76,67,0}},
/* 53116 */ {(14<<2)|2,{76,68,0}},
/* 53117 */ {(14<<2)|2,{76,68,0}},
/* 53118 */ {(14<<2)|2,{76,68,0}},
/* 53119 */ {(14<<2)|2,{76,68,0}},
/* 53120 */ {(14<<2)|2,{76,69,0}},
/* 53121 */ {(14<<2)|2,{76,69,0}},
/* 53122 */ {(14<<2)|2,{76,69,0}},
/* 53123 */ {(14<<2)|2,{76,69,0}},
/* 53124 */ {(14<<2)|2,{76,70,0}},
/* 53125 */ {(14<<2)|2,{76,70,0}},
/* 53126 */ {(14<<2)|2,{76,70,0}},
/* 53127 */ {(14<<2)|2,{76,70,0}},
/* 53128 */ {(14<<2)|2,{76,71,0}},
/* 53129 */ {(14<<2)|2,{76,71,0}},
/* 53130 */ {(14<<2)|2,{76,71,0}},
/* 53131 */ {(14<<2)|2,{76,71,0}},
/* 53132 */ {(14<<2)|2,{76,72,0}},
/* 53133 */ {(14<<2)|2,{76,72,0}},
/* 53134 */ {(14<<2)|2,{76,72,0}},
/* 53135 */ {(14<<2)|2,{76,72,0}},
/* 53136 */ {(14<<2)|2,{76,73,0}},
/* 53137 */ {(14<<2)|2,{76,73,0}},
/* 53138 */ {(14<<2)|2,{76,73,0}},
/* 53139 */ {(14<<2)|2,{76,73,0}},
/* 53140 */ {(14<<2)|2,{76,74,0}},
/* 53141 */ {(14<<2)|2,{76,74,0}},
/* 53142 */ {(14<<2)|2,{76,74,0}},
/* 53143 */ {(14<<2)|2,{76,74,0}},
/* 53144 */ {(14<<2)|2,{76,75,0}},
/* 53145 */ {(14<<2)|2,{76,75,0}},
/* 53146 */ {(14<<2)|2,{76,75,0}},
/* 53147 */ {(14<<2)|2,{76,75,0}},
/* 53148 */ {(14<<2)|2,{76,76,0}},
/* 53149 */ {(14<<2)|2,{76,76,0}},
/* 53150 */ {(14<<2)|2,{76,76,0}},
/* 53151 */ {(14<<2)|2,{76,76,0}},
/* 53152 */ {(14<<2)|2,{76,77,0}},
/* 53153 */ {(14<<2)|2,{76,77,0}},
/* 53154 */ {(14<<2)|2,{76,77,0}},
/* 53155 */ {(14<<2)|2,{76,77,0}},
/* 53156 */ {(14<<2)|2,{76,78,0}},
/* 53157 */ {(14<<2)|2,{76,78,0}},
/* 53158 */ {(14<<2)|2,{76,78,0}},
/* 53159 */ {(14<<2)|2,{76,78,0}},
/* 53160 */ {(14<<2)|2,{76,79,0}},
/* 53161 */ {(14<<2)|2,{76,79,0}},
/* 53162 */ {(14<<2)|2,{76,79,0}},
/* 53163 */ {(14<<2)|2,{76,79,0}},
/* 53164 */ {(14<<2)|2,{76,80,0}},
/* 53165 */ {(14<<2)|2,{76,80,0}},
/* 53166 */ {(14<<2)|2,{76,80,0}},
/* 53167 */ {(14<<2)|2,{76,80,0}},
/* 53168 */ {(14<<2)|2,{76,81,0}},
/* 53169 */ {(14<<2)|2,{76,81,0}},
/* 53170 */ {(14<<2)|2,{76,81,0}},
/* 53171 */ {(14<<2)|2,{76,81,0}},
/* 53172 */ {(14<<2)|2,{76,82,0}},
/* 53173 */ {(14<<2)|2,{76,82,0}},
/* 53174 */ {(14<<2)|2,{76,82,0}},
/* 53175 */ {(14<<2)|2,{76,82,0}},
/* 53176 */ {(14<<2)|2,{76,83,0}},
/* 53177 */ {(14<<2)|2,{76,83,0}},
/* 53178 */ {(14<<2)|2,{76,83,0}},
/* 53179 */ {(14<<2)|2,{76,83,0}},
/* 53180 */ {(14<<2)|2,{76,84,0}},
/* 53181 */ {(14<<2)|2,{76,84,0}},
/* 53182 */ {(14<<2)|2,{76,84,0}},
/* 53183 */ {(14<<2)|2,{76,84,0}},
/* 53184 */ {(14<<2)|2,{76,85,0}},
/* 53185 */ {(14<<2)|2,{76,85,0}},
/* 53186 */ {(14<<2)|2,{76,85,0}},
/* 53187 */ {(14<<2)|2,{76,85,0}},
/* 53188 */ {(14<<2)|2,{76,86,0}},
/* 53189 */ {(14<<2)|2,{76,86,0}},
/* 53190 */ {(14<<2)|2,{76,86,0}},
/* 53191 */ {(14<<2)|2,{76,86,0}},
/* 53192 */ {(14<<2)|2,{76,87,0}},
/* 53193 */ {(14<<2)|2,{76,87,0}},
/* 53194 */ {(14<<2)|2,{76,87,0}},
/* 53195 */ {(14<<2)|2,{76,87,0}},
/* 53196 */ {(14<<2)|2,{76,89,0}},
/* 53197 */ {(14<<2)|2,{76,89,0}},
/* 53198 */ {(14<<2)|2,{76,89,0}},
/* 53199 */ {(14<<2)|2,{76,89,0}},
/* 53200 */ {(14<<2)|2,{76,106,0}},
/* 53201 */ {(14<<2)|2,{76,106,0}},
/* 53202 */ {(14<<2)|2,{76,106,0}},
/* 53203 */ {(14<<2)|2,{76,106,0}},
/* 53204 */ {(14<<2)|2,{76,107,0}},
/* 53205 */ {(14<<2)|2,{76,107,0}},
/* 53206 */ {(14<<2)|2,{76,107,0}},
/* 53207 */ {(14<<2)|2,{76,107,0}},
/* 53208 */ {(14<<2)|2,{76,113,0}},
/* 53209 */ {(14<<2)|2,{76,113,0}},
/* 53210 */ {(14<<2)|2,{76,113,0}},
/* 53211 */ {(14<<2)|2,{76,113,0}},
/* 53212 */ {(14<<2)|2,{76,118,0}},
/* 53213 */ {(14<<2)|2,{76,118,0}},
/* 53214 */ {(14<<2)|2,{76,118,0}},
/* 53215 */ {(14<<2)|2,{76,118,0}},
/* 53216 */ {(14<<2)|2,{76,119,0}},
/* 53217 */ {(14<<2)|2,{76,119,0}},
/* 53218 */ {(14<<2)|2,{76,119,0}},
/* 53219 */ {(14<<2)|2,{76,119,0}},
/* 53220 */ {(14<<2)|2,{76,120,0}},
/* 53221 */ {(14<<2)|2,{76,120,0}},
/* 53222 */ {(14<<2)|2,{76,120,0}},
/* 53223 */ {(14<<2)|2,{76,120,0}},
/* 53224 */ {(14<<2)|2,{76,121,0}},
/* 53225 */ {(14<<2)|2,{76,121,0}},
/* 53226 */ {(14<<2)|2,{76,121,0}},
/* 53227 */ {(14<<2)|2,{76,121,0}},
/* 53228 */ {(14<<2)|2,{76,122,0}},
/* 53229 */ {(14<<2)|2,{76,122,0}},
/* 53230 */ {(14<<2)|2,{76,122,0}},
/* 53231 */ {(14<<2)|2,{76,122,0}},
/* 53232 */ {(15<<2)|2,{76,38,0}},
/* 53233 */ {(15<<2)|2,{76,38,0}},
/* 53234 */ {(15<<2)|2,{76,42,0}},
/* 53235 */ {(15<<2)|2,{76,42,0}},
/* 53236 */ {(15<<2)|2,{76,44,0}},
/* 53237 */ {(15<<2)|2,{76,44,0}},
/* 53238 */ {(15<<2)|2,{76,59,0}},
/* 53239 */ {(15<<2)|2,{76,59,0}},
/* 53240 */ {(15<<2)|2,{76,88,0}},
/* 53241 */ {(15<<2)|2,{76,88,0}},
/* 53242 */ {(15<<2)|2,{76,90,0}},
/* 53243 */ {(15<<2)|2,{76,90,0}},
/* 53244 */ {(7<<2)|1,{76,0,0}},
/* 53245 */ {(7<<2)|1,{76,0,0}},
/* 53246 */ {(7<<2)|1,{76,0,0}},
/* 53247 */ {(7<<2)|1,{76,0,0}},
/* 53248 */ {(12<<2)|2,{77,48,0}},
/* 53249 */ {(12<<2)|2,{77,48,0}},
/* 53250 */ {(12<<2)|2,{77,48,0}},
/* 53251 */ {(12<<2)|2,{77,48,0}},
/* 53252 */ {(12<<2)|2,{77,48,0}},
/* 53253 */ {(12<<2)|2,{77,48,0}},
/* 53254 */ {(12<<2)|2,{77,48,0}},
/* 53255 */ {(12<<2)|2,{77,48,0}},
/* 53256 */ {(12<<2)|2,{77,48,0}},
/* 53257 */ {(12<<2)|2,{77,48,0}},
/* 53258 */ {(12<<2)|2,{77,48,0}},
/* 53259 */ {(12<<2)|2,{77,48,0}},
/* 53260 */ {(12<<2)|2,{77,48,0}},
/* 53261 */ {(12<<2)|2,{77,48,0}},
/* 53262 */ {(12<<2)|2,{77,48,0}},
/* 53263 */ {(12<<2)|2,{77,48,0}},
/* 53264 */ {(12<<2)|2,{77,49,0}},
/* 53265 */ {(12<<2)|2,{77,49,0}},
/* 53266 */ {(12<<2)|2,{77,49,0}},
/* 53267 */ {(12<<2)|2,{77,49,0}},
/* 53268 */ {(12<<2)|2,{77,49,0}},
/* 53269 */ {(12<<2)|2,{77,49,0}},
/* 53270 */ {(12<<2)|2,{77,49,0}},
/* 53271 */ {(12<<2)|2,{77,49,0}},
/* 53272 */ {(12<<2)|2,{77,49,0}},
/* 53273 */ {(12<<2)|2,{77,49,0}},
/* 53274 */ {(12<<2)|2,{77,49,0}},
/* 53275 */ {(12<<2)|2,{77,49,0}},
/* 53276 */ {(12<<2)|2,{77,49,0}},
/* 53277 */ {(12<<2)|2,{77,49,0}},
/* 53278 */ {(12<<2)|2,{77,49,0}},
/* 53279 */ {(12<<2)|2,{77,49,0}},
/* 53280 */ {(12<<2)|2,{77,50,0}},
/* 53281 */ {(12<<2)|2,{77,50,0}},
/* 53282 */ {(12<<2)|2,{77,50,0}},
/* 53283 */ {(12<<2)|2,{77,50,0}},
/* 53284 */ {(12<<2)|2,{77,50,0}},
/* 53285 */ {(12<<2)|2,{77,50,0}},
/* 53286 */ {(12<<2)|2,{77,50,0}},
/* 53287 */ {(12<<2)|2,{77,50,0}},
/* 53288 */ {(12<<2)|2,{77,50,0}},
/* 53289 */ {(12<<2)|2,{77,50,0}},
/* 53290 */ {(12<<2)|2,{77,50,0}},
/* 53291 */ {(12<<2)|2,{77,50,0}},
/* 53292 */ {(12<<2)|2,{77,50,0}},
/* 53293 */ {(12<<2)|2,{77,50,0}},
/* 53294 */ {(12<<2)|2,{77,50,0}},
/* 53295 */ {(12<<2)|2,{77,50,0}},
/* 53296 */ {(12<<2)|2,{77,97,0}},
/* 53297 */ {(12<<2)|2,{77,97,0}},
/* 53298 */ {(12<<2)|2,{77,97,0}},
/* 53299 */ {(12<<2)|2,{77,97,0}},
/* 53300 */ {(12<<2)|2,{77,97,0}},
/* 53301 */ {(12<<2)|2,{77,97,0}},
/* 53302 */ {(12<<2)|2,{77,97,0}},
/* 53303 */ {(12<<2)|2,{77,97,0}},
/* 53304 */ {(12<<2)|2,{77,97,0}},
/* 53305 */ {(12<<2)|2,{77,97,0}},
/* 53306 */ {(12<<2)|2,{77,97,0}},
/* 53307 */ {(12<<2)|2,{77,97,0}},
/* 53308 */ {(12<<2)|2,{77,97,0}},
/* 53309 */ {(12<<2)|2,{77,97,0}},
/* 53310 */ {(12<<2)|2,{77,97,0}},
/* 53311 */ {(12<<2)|2,{77,97,0}},
/* 53312 */ {(12<<2)|2,{77,99,0}},
/* 53313 */ {(12<<2)|2,{77,99,0}},
/* 53314 */ {(12<<2)|2,{77,99,0}},
/* 53315 */ {(12<<2)|2,{77,99,0}},
/* 53316 */ {(12<<2)|2,{77,99,0}},
/* 53317 */ {(12<<2)|2,{77,99,0}},
/* 53318 */ {(12<<2)|2,{77,99,0}},
/* 53319 */ {(12<<2)|2,{77,99,0}},
/* 53320 */ {(12<<2)|2,{77,99,0}},
/* 53321 */ {(12<<2)|2,{77,99,0}},
/* 53322 */ {(12<<2)|2,{77,99,0}},
/* 53323 */ {(12<<2)|2,{77,99,0}},
/* 53324 */ {(12<<2)|2,{77,99,0}},
/* 53325 */ {(12<<2)|2,{77,99,0}},
/* 53326 */ {(12<<2)|2,{77,99,0}},
/* 53327 */ {(12<<2)|2,{77,99,0}},
/* 53328 */ {(12<<2)|2,{77,101,0}},
/* 53329 */ {(12<<2)|2,{77,101,0}},
/* 53330 */ {(12<<2)|2,{77,101,0}},
/* 53331 */ {(12<<2)|2,{77,101,0}},
/* 53332 */ {(12<<2)|2,{77,101,0}},
/* 53333 */ {(12<<2)|2,{77,101,0}},
/* 53334 */ {(12<<2)|2,{77,101,0}},
/* 53335 */ {(12<<2)|2,{77,101,0}},
/* 53336 */ {(12<<2)|2,{77,101,0}},
/* 53337 */ {(12<<2)|2,{77,101,0}},
/* 53338 */ {(12<<2)|2,{77,101,0}},
/* 53339 */ {(12<<2)|2,{77,101,0}},
/* 53340 */ {(12<<2)|2,{77,101,0}},
/* 53341 */ {(12<<2)|2,{77,101,0}},
/* 53342 */ {(12<<2)|2,{77,101,0}},
/* 53343 */ {(12<<2)|2,{77,101,0}},
/* 53344 */ {(12<<2)|2,{77,105,0}},
/* 53345 */ {(12<<2)|2,{77,105,0}},
/* 53346 */ {(12<<2)|2,{77,105,0}},
/* 53347 */ {(12<<2)|2,{77,105,0}},
/* 53348 */ {(12<<2)|2,{77,105,0}},
/* 53349 */ {(12<<2)|2,{77,105,0}},
/* 53350 */ {(12<<2)|2,{77,105,0}},
/* 53351 */ {(12<<2)|2,{77,105,0}},
/* 53352 */ {(12<<2)|2,{77,105,0}},
/* 53353 */ {(12<<2)|2,{77,105,0}},
/* 53354 */ {(12<<2)|2,{77,105,0}},
/* 53355 */ {(12<<2)|2,{77,105,0}},
/* 53356 */ {(12<<2)|2,{77,105,0}},
/* 53357 */ {(12<<2)|2,{77,105,0}},
/* 53358 */ {(12<<2)|2,{77,105,0}},
/* 53359 */ {(12<<2)|2,{77,105,0}},
/* 53360 */ {(12<<2)|2,{77,111,0}},
/* 53361 */ {(12<<2)|2,{77,111,0}},
/* 53362 */ {(12<<2)|2,{77,111,0}},
/* 53363 */ {(12<<2)|2,{77,111,0}},
/* 53364 */ {(12<<2)|2,{77,111,0}},
/* 53365 */ {(12<<2)|2,{77,111,0}},
/* 53366 */ {(12<<2)|2,{77,111,0}},
/* 53367 */ {(12<<2)|2,{77,111,0}},
/* 53368 */ {(12<<2)|2,{77,111,0}},
/* 53369 */ {(12<<2)|2,{77,111,0}},
/* 53370 */ {(12<<2)|2,{77,111,0}},
/* 53371 */ {(12<<2)|2,{77,111,0}},
/* 53372 */ {(12<<2)|2,{77,111,0}},
/* 53373 */ {(12<<2)|2,{77,111,0}},
/* 53374 */ {(12<<2)|2,{77,111,0}},
/* 53375 */ {(12<<2)|2,{77,111,0}},
/* 53376 */ {(12<<2)|2,{77,115,0}},
/* 53377 */ {(12<<2)|2,{77,115,0}},
/* 53378 */ {(12<<2)|2,{77,115,0}},
/* 53379 */ {(12<<2)|2,{77,115,0}},
/* 53380 */ {(12<<2)|2,{77,115,0}},
/* 53381 */ {(12<<2)|2,{77,115,0}},
/* 53382 */ {(12<<2)|2,{77,115,0}},
/* 53383 */ {(12<<2)|2,{77,115,0}},
/* 53384 */ {(12<<2)|2,{77,115,0}},
/* 53385 */ {(12<<2)|2,{77,115,0}},
/* 53386 */ {(12<<2)|2,{77,115,0}},
/* 53387 */ {(12<<2)|2,{77,115,0}},
/* 53388 */ {(12<<2)|2,{77,115,0}},
/* 53389 */ {(12<<2)|2,{77,115,0}},
/* 53390 */ {(12<<2)|2,{77,115,0}},
/* 53391 */ {(12<<2)|2,{77,115,0}},
/* 53392 */ {(12<<2)|2,{77,116,0}},
/* 53393 */ {(12<<2)|2,{77,116,0}},
/* 53394 */ {(12<<2)|2,{77,116,0}},
/* 53395 */ {(12<<2)|2,{77,116,0}},
/* 53396 */ {(12<<2)|2,{77,116,0}},
/* 53397 */ {(12<<2)|2,{77,116,0}},
/* 53398 */ {(12<<2)|2,{77,116,0}},
/* 53399 */ {(12<<2)|2,{77,116,0}},
/* 53400 */ {(12<<2)|2,{77,116,0}},
/* 53401 */ {(12<<2)|2,{77,116,0}},
/* 53402 */ {(12<<2)|2,{77,116,0}},
/* 53403 */ {(12<<2)|2,{77,116,0}},
/* 53404 */ {(12<<2)|2,{77,116,0}},
/* 53405 */ {(12<<2)|2,{77,116,0}},
/* 53406 */ {(12<<2)|2,{77,116,0}},
/* 53407 */ {(12<<2)|2,{77,116,0}},
/* 53408 */ {(13<<2)|2,{77,32,0}},
/* 53409 */ {(13<<2)|2,{77,32,0}},
/* 53410 */ {(13<<2)|2,{77,32,0}},
/* 53411 */ {(13<<2)|2,{77,32,0}},
/* 53412 */ {(13<<2)|2,{77,32,0}},
/* 53413 */ {(13<<2)|2,{77,32,0}},
/* 53414 */ {(13<<2)|2,{77,32,0}},
/* 53415 */ {(13<<2)|2,{77,32,0}},
/* 53416 */ {(13<<2)|2,{77,37,0}},
/* 53417 */ {(13<<2)|2,{77,37,0}},
/* 53418 */ {(13<<2)|2,{77,37,0}},
/* 53419 */ {(13<<2)|2,{77,37,0}},
/* 53420 */ {(13<<2)|2,{77,37,0}},
/* 53421 */ {(13<<2)|2,{77,37,0}},
/* 53422 */ {(13<<2)|2,{77,37,0}},
/* 53423 */ {(13<<2)|2,{77,37,0}},
/* 53424 */ {(13<<2)|2,{77,45,0}},
/* 53425 */ {(13<<2)|2,{77,45,0}},
/* 53426 */ {(13<<2)|2,{77,45,0}},
/* 53427 */ {(13<<2)|2,{77,45,0}},
/* 53428 */ {(13<<2)|2,{77,45,0}},
/* 53429 */ {(13<<2)|2,{77,45,0}},
/* 53430 */ {(13<<2)|2,{77,45,0}},
/* 53431 */ {(13<<2)|2,{77,45,0}},
/* 53432 */ {(13<<2)|2,{77,46,0}},
/* 53433 */ {(13<<2)|2,{77,46,0}},
/* 53434 */ {(13<<2)|2,{77,46,0}},
/* 53435 */ {(13<<2)|2,{77,46,0}},
/* 53436 */ {(13<<2)|2,{77,46,0}},
/* 53437 */ {(13<<2)|2,{77,46,0}},
/* 53438 */ {(13<<2)|2,{77,46,0}},
/* 53439 */ {(13<<2)|2,{77,46,0}},
/* 53440 */ {(13<<2)|2,{77,47,0}},
/* 53441 */ {(13<<2)|2,{77,47,0}},
/* 53442 */ {(13<<2)|2,{77,47,0}},
/* 53443 */ {(13<<2)|2,{77,47,0}},
/* 53444 */ {(13<<2)|2,{77,47,0}},
/* 53445 */ {(13<<2)|2,{77,47,0}},
/* 53446 */ {(13<<2)|2,{77,47,0}},
/* 53447 */ {(13<<2)|2,{77,47,0}},
/* 53448 */ {(13<<2)|2,{77,51,0}},
/* 53449 */ {(13<<2)|2,{77,51,0}},
/* 53450 */ {(13<<2)|2,{77,51,0}},
/* 53451 */ {(13<<2)|2,{77,51,0}},
/* 53452 */ {(13<<2)|2,{77,51,0}},
/* 53453 */ {(13<<2)|2,{77,51,0}},
/* 53454 */ {(13<<2)|2,{77,51,0}},
/* 53455 */ {(13<<2)|2,{77,51,0}},
/* 53456 */ {(13<<2)|2,{77,52,0}},
/* 53457 */ {(13<<2)|2,{77,52,0}},
/* 53458 */ {(13<<2)|2,{77,52,0}},
/* 53459 */ {(13<<2)|2,{77,52,0}},
/* 53460 */ {(13<<2)|2,{77,52,0}},
/* 53461 */ {(13<<2)|2,{77,52,0}},
/* 53462 */ {(13<<2)|2,{77,52,0}},
/* 53463 */ {(13<<2)|2,{77,52,0}},
/* 53464 */ {(13<<2)|2,{77,53,0}},
/* 53465 */ {(13<<2)|2,{77,53,0}},
/* 53466 */ {(13<<2)|2,{77,53,0}},
/* 53467 */ {(13<<2)|2,{77,53,0}},
/* 53468 */ {(13<<2)|2,{77,53,0}},
/* 53469 */ {(13<<2)|2,{77,53,0}},
/* 53470 */ {(13<<2)|2,{77,53,0}},
/* 53471 */ {(13<<2)|2,{77,53,0}},
/* 53472 */ {(13<<2)|2,{77,54,0}},
/* 53473 */ {(13<<2)|2,{77,54,0}},
/* 53474 */ {(13<<2)|2,{77,54,0}},
/* 53475 */ {(13<<2)|2,{77,54,0}},
/* 53476 */ {(13<<2)|2,{77,54,0}},
/* 53477 */ {(13<<2)|2,{77,54,0}},
/* 53478 */ {(13<<2)|2,{77,54,0}},
/* 53479 */ {(13<<2)|2,{77,54,0}},
/* 53480 */ {(13<<2)|2,{77,55,0}},
/* 53481 */ {(13<<2)|2,{77,55,0}},
/* 53482 */ {(13<<2)|2,{77,55,0}},
/* 53483 */ {(13<<2)|2,{77,55,0}},
/* 53484 */ {(13<<2)|2,{77,55,0}},
/* 53485 */ {(13<<2)|2,{77,55,0}},
/* 53486 */ {(13<<2)|2,{77,55,0}},
/* 53487 */ {(13<<2)|2,{77,55,0}},
/* 53488 */ {(13<<2)|2,{77,56,0}},
/* 53489 */ {(13<<2)|2,{77,56,0}},
/* 53490 */ {(13<<2)|2,{77,56,0}},
/* 53491 */ {(13<<2)|2,{77,56,0}},
/* 53492 */ {(13<<2)|2,{77,56,0}},
/* 53493 */ {(13<<2)|2,{77,56,0}},
/* 53494 */ {(13<<2)|2,{77,56,0}},
/* 53495 */ {(13<<2)|2,{77,56,0}},
/* 53496 */ {(13<<2)|2,{77,57,0}},
/* 53497 */ {(13<<2)|2,{77,57,0}},
/* 53498 */ {(13<<2)|2,{77,57,0}},
/* 53499 */ {(13<<2)|2,{77,57,0}},
/* 53500 */ {(13<<2)|2,{77,57,0}},
/* 53501 */ {(13<<2)|2,{77,57,0}},
/* 53502 */ {(13<<2)|2,{77,57,0}},
/* 53503 */ {(13<<2)|2,{77,57,0}},
/* 53504 */ {(13<<2)|2,{77,61,0}},
/* 53505 */ {(13<<2)|2,{77,61,0}},
/* 53506 */ {(13<<2)|2,{77,61,0}},
/* 53507 */ {(13<<2)|2,{77,61,0}},
/* 53508 */ {(13<<2)|2,{77,61,0}},
/* 53509 */ {(13<<2)|2,{77,61,0}},
/* 53510 */ {(13<<2)|2,{77,61,0}},
/* 53511 */ {(13<<2)|2,{77,61,0}},
/* 53512 */ {(13<<2)|2,{77,65,0}},
/* 53513 */ {(13<<2)|2,{77,65,0}},
/* 53514 */ {(13<<2)|2,{77,65,0}},
/* 53515 */ {(13<<2)|2,{77,65,0}},
/* 53516 */ {(13<<2)|2,{77,65,0}},
/* 53517 */ {(13<<2)|2,{77,65,0}},
/* 53518 */ {(13<<2)|2,{77,65,0}},
/* 53519 */ {(13<<2)|2,{77,65,0}},
/* 53520 */ {(13<<2)|2,{77,95,0}},
/* 53521 */ {(13<<2)|2,{77,95,0}},
/* 53522 */ {(13<<2)|2,{77,95,0}},
/* 53523 */ {(13<<2)|2,{77,95,0}},
/* 53524 */ {(13<<2)|2,{77,95,0}},
/* 53525 */ {(13<<2)|2,{77,95,0}},
/* 53526 */ {(13<<2)|2,{77,95,0}},
/* 53527 */ {(13<<2)|2,{77,95,0}},
/* 53528 */ {(13<<2)|2,{77,98,0}},
/* 53529 */ {(13<<2)|2,{77,98,0}},
/* 53530 */ {(13<<2)|2,{77,98,0}},
/* 53531 */ {(13<<2)|2,{77,98,0}},
/* 53532 */ {(13<<2)|2,{77,98,0}},
/* 53533 */ {(13<<2)|2,{77,98,0}},
/* 53534 */ {(13<<2)|2,{77,98,0}},
/* 53535 */ {(13<<2)|2,{77,98,0}},
/* 53536 */ {(13<<2)|2,{77,100,0}},
/* 53537 */ {(13<<2)|2,{77,100,0}},
/* 53538 */ {(13<<2)|2,{77,100,0}},
/* 53539 */ {(13<<2)|2,{77,100,0}},
/* 53540 */ {(13<<2)|2,{77,100,0}},
/* 53541 */ {(13<<2)|2,{77,100,0}},
/* 53542 */ {(13<<2)|2,{77,100,0}},
/* 53543 */ {(13<<2)|2,{77,100,0}},
/* 53544 */ {(13<<2)|2,{77,102,0}},
/* 53545 */ {(13<<2)|2,{77,102,0}},
/* 53546 */ {(13<<2)|2,{77,102,0}},
/* 53547 */ {(13<<2)|2,{77,102,0}},
/* 53548 */ {(13<<2)|2,{77,102,0}},
/* 53549 */ {(13<<2)|2,{77,102,0}},
/* 53550 */ {(13<<2)|2,{77,102,0}},
/* 53551 */ {(13<<2)|2,{77,102,0}},
/* 53552 */ {(13<<2)|2,{77,103,0}},
/* 53553 */ {(13<<2)|2,{77,103,0}},
/* 53554 */ {(13<<2)|2,{77,103,0}},
/* 53555 */ {(13<<2)|2,{77,103,0}},
/* 53556 */ {(13<<2)|2,{77,103,0}},
/* 53557 */ {(13<<2)|2,{77,103,0}},
/* 53558 */ {(13<<2)|2,{77,103,0}},
/* 53559 */ {(13<<2)|2,{77,103,0}},
/* 53560 */ {(13<<2)|2,{77,104,0}},
/* 53561 */ {(13<<2)|2,{77,104,0}},
/* 53562 */ {(13<<2)|2,{77,104,0}},
/* 53563 */ {(13<<2)|2,{77,104,0}},
/* 53564 */ {(13<<2)|2,{77,104,0}},
/* 53565 */ {(13<<2)|2,{77,104,0}},
/* 53566 */ {(13<<2)|2,{77,104,0}},
/* 53567 */ {(13<<2)|2,{77,104,0}},
/* 53568 */ {(13<<2)|2,{77,108,0}},
/* 53569 */ {(13<<2)|2,{77,108,0}},
/* 53570 */ {(13<<2)|2,{77,108,0}},
/* 53571 */ {(13<<2)|2,{77,108,0}},
/* 53572 */ {(13<<2)|2,{77,108,0}},
/* 53573 */ {(13<<2)|2,{77,108,0}},
/* 53574 */ {(13<<2)|2,{77,108,0}},
/* 53575 */ {(13<<2)|2,{77,108,0}},
/* 53576 */ {(13<<2)|2,{77,109,0}},
/* 53577 */ {(13<<2)|2,{77,109,0}},
/* 53578 */ {(13<<2)|2,{77,109,0}},
/* 53579 */ {(13<<2)|2,{77,109,0}},
/* 53580 */ {(13<<2)|2,{77,109,0}},
/* 53581 */ {(13<<2)|2,{77,109,0}},
/* 53582 */ {(13<<2)|2,{77,109,0}},
/* 53583 */ {(13<<2)|2,{77,109,0}},
/* 53584 */ {(13<<2)|2,{77,110,0}},
/* 53585 */ {(13<<2)|2,{77,110,0}},
/* 53586 */ {(13<<2)|2,{77,110,0}},
/* 53587 */ {(13<<2)|2,{77,110,0}},
/* 53588 */ {(13<<2)|2,{77,110,0}},
/* 53589 */ {(13<<2)|2,{77,110,0}},
/* 53590 */ {(13<<2)|2,{77,110,0}},
/* 53591 */ {(13<<2)|2,{77,110,0}},
/* 53592 */ {(13<<2)|2,{77,112,0}},
/* 53593 */ {(13<<2)|2,{77,112,0}},
/* 53594 */ {(13<<2)|2,{77,112,0}},
/* 53595 */ {(13<<2)|2,{77,112,0}},
/* 53596 */ {(13<<2)|2,{77,112,0}},
/* 53597 */ {(13<<2)|2,{77,112,0}},
/* 53598 */ {(13<<2)|2,{77,112,0}},
/* 53599 */ {(13<<2)|2,{77,112,0}},
/* 53600 */ {(13<<2)|2,{77,114,0}},
/* 53601 */ {(13<<2)|2,{77,114,0}},
/* 53602 */ {(13<<2)|2,{77,114,0}},
/* 53603 */ {(13<<2)|2,{77,114,0}},
/* 53604 */ {(13<<2)|2,{77,114,0}},
/* 53605 */ {(13<<2)|2,{77,114,0}},
/* 53606 */ {(13<<2)|2,{77,114,0}},
/* 53607 */ {(13<<2)|2,{77,114,0}},
/* 53608 */ {(13<<2)|2,{77,117,0}},
/* 53609 */ {(13<<2)|2,{77,117,0}},
/* 53610 */ {(13<<2)|2,{77,117,0}},
/* 53611 */ {(13<<2)|2,{77,117,0}},
/* 53612 */ {(13<<2)|2,{77,117,0}},
/* 53613 */ {(13<<2)|2,{77,117,0}},
/* 53614 */ {(13<<2)|2,{77,117,0}},
/* 53615 */ {(13<<2)|2,{77,117,0}},
/* 53616 */ {(14<<2)|2,{77,58,0}},
/* 53617 */ {(14<<2)|2,{77,58,0}},
/* 53618 */ {(14<<2)|2,{77,58,0}},
/* 53619 */ {(14<<2)|2,{77,58,0}},
/* 53620 */ {(14<<2)|2,{77,66,0}},
/* 53621 */ {(14<<2)|2,{77,66,0}},
/* 53622 */ {(14<<2)|2,{77,66,0}},
/* 53623 */ {(14<<2)|2,{77,66,0}},
/* 53624 */ {(14<<2)|2,{77,67,0}},
/* 53625 */ {(14<<2)|2,{77,67,0}},
/* 53626 */ {(14<<2)|2,{77,67,0}},
/* 53627 */ {(14<<2)|2,{77,67,0}},
/* 53628 */ {(14<<2)|2,{77,68,0}},
/* 53629 */ {(14<<2)|2,{77,68,0}},
/* 53630 */ {(14<<2)|2,{77,68,0}},
/* 53631 */ {(14<<2)|2,{77,68,0}},
/* 53632 */ {(14<<2)|2,{77,69,0}},
/* 53633 */ {(14<<2)|2,{77,69,0}},
/* 53634 */ {(14<<2)|2,{77,69,0}},
/* 53635 */ {(14<<2)|2,{77,69,0}},
/* 53636 */ {(14<<2)|2,{77,70,0}},
/* 53637 */ {(14<<2)|2,{77,70,0}},
/* 53638 */ {(14<<2)|2,{77,70,0}},
/* 53639 */ {(14<<2)|2,{77,70,0}},
/* 53640 */ {(14<<2)|2,{77,71,0}},
/* 53641 */ {(14<<2)|2,{77,71,0}},
/* 53642 */ {(14<<2)|2,{77,71,0}},
/* 53643 */ {(14<<2)|2,{77,71,0}},
/* 53644 */ {(14<<2)|2,{77,72,0}},
/* 53645 */ {(14<<2)|2,{77,72,0}},
/* 53646 */ {(14<<2)|2,{77,72,0}},
/* 53647 */ {(14<<2)|2,{77,72,0}},
/* 53648 */ {(14<<2)|2,{77,73,0}},
/* 53649 */ {(14<<2)|2,{77,73,0}},
/* 53650 */ {(14<<2)|2,{77,73,0}},
/* 53651 */ {(14<<2)|2,{77,73,0}},
/* 53652 */ {(14<<2)|2,{77,74,0}},
/* 53653 */ {(14<<2)|2,{77,74,0}},
/* 53654 */ {(14<<2)|2,{77,74,0}},
/* 53655 */ {(14<<2)|2,{77,74,0}},
/* 53656 */ {(14<<2)|2,{77,75,0}},
/* 53657 */ {(14<<2)|2,{77,75,0}},
/* 53658 */ {(14<<2)|2,{77,75,0}},
/* 53659 */ {(14<<2)|2,{77,75,0}},
/* 53660 */ {(14<<2)|2,{77,76,0}},
/* 53661 */ {(14<<2)|2,{77,76,0}},
/* 53662 */ {(14<<2)|2,{77,76,0}},
/* 53663 */ {(14<<2)|2,{77,76,0}},
/* 53664 */ {(14<<2)|2,{77,77,0}},
/* 53665 */ {(14<<2)|2,{77,77,0}},
/* 53666 */ {(14<<2)|2,{77,77,0}},
/* 53667 */ {(14<<2)|2,{77,77,0}},
/* 53668 */ {(14<<2)|2,{77,78,0}},
/* 53669 */ {(14<<2)|2,{77,78,0}},
/* 53670 */ {(14<<2)|2,{77,78,0}},
/* 53671 */ {(14<<2)|2,{77,78,0}},
/* 53672 */ {(14<<2)|2,{77,79,0}},
/* 53673 */ {(14<<2)|2,{77,79,0}},
/* 53674 */ {(14<<2)|2,{77,79,0}},
/* 53675 */ {(14<<2)|2,{77,79,0}},
/* 53676 */ {(14<<2)|2,{77,80,0}},
/* 53677 */ {(14<<2)|2,{77,80,0}},
/* 53678 */ {(14<<2)|2,{77,80,0}},
/* 53679 */ {(14<<2)|2,{77,80,0}},
/* 53680 */ {(14<<2)|2,{77,81,0}},
/* 53681 */ {(14<<2)|2,{77,81,0}},
/* 53682 */ {(14<<2)|2,{77,81,0}},
/* 53683 */ {(14<<2)|2,{77,81,0}},
/* 53684 */ {(14<<2)|2,{77,82,0}},
/* 53685 */ {(14<<2)|2,{77,82,0}},
/* 53686 */ {(14<<2)|2,{77,82,0}},
/* 53687 */ {(14<<2)|2,{77,82,0}},
/* 53688 */ {(14<<2)|2,{77,83,0}},
/* 53689 */ {(14<<2)|2,{77,83,0}},
/* 53690 */ {(14<<2)|2,{77,83,0}},
/* 53691 */ {(14<<2)|2,{77,83,0}},
/* 53692 */ {(14<<2)|2,{77,84,0}},
/* 53693 */ {(14<<2)|2,{77,84,0}},
/* 53694 */ {(14<<2)|2,{77,84,0}},
/* 53695 */ {(14<<2)|2,{77,84,0}},
/* 53696 */ {(14<<2)|2,{77,85,0}},
/* 53697 */ {(14<<2)|2,{77,85,0}},
/* 53698 */ {(14<<2)|2,{77,85,0}},
/* 53699 */ {(14<<2)|2,{77,85,0}},
/* 53700 */ {(14<<2)|2,{77,86,0}},
/* 53701 */ {(14<<2)|2,{77,86,0}},
/* 53702 */ {(14<<2)|2,{77,86,0}},
/* 53703 */ {(14<<2)|2,{77,86,0}},
/* 53704 */ {(14<<2)|2,{77,87,0}},
/* 53705 */ {(14<<2)|2,{77,87,0}},
/* 53706 */ {(14<<2)|2,{77,87,0}},
/* 53707 */ {(14<<2)|2,{77,87,0}},
/* 53708 */ {(14<<2)|2,{77,89,0}},
/* 53709 */ {(14<<2)|2,{77,89,0}},
/* 53710 */ {(14<<2)|2,{77,89,0}},
/* 53711 */ {(14<<2)|2,{77,89,0}},
/* 53712 */ {(14<<2)|2,{77,106,0}},
/* 53713 */ {(14<<2)|2,{77,106,0}},
/* 53714 */ {(14<<2)|2,{77,106,0}},
/* 53715 */ {(14<<2)|2,{77,106,0}},
/* 53716 */ {(14<<2)|2,{77,107,0}},
/* 53717 */ {(14<<2)|2,{77,107,0}},
/* 53718 */ {(14<<2)|2,{77,107,0}},
/* 53719 */ {(14<<2)|2,{77,107,0}},
/* 53720 */ {(14<<2)|2,{77,113,0}},
/* 53721 */ {(14<<2)|2,{77,113,0}},
/* 53722 */ {(14<<2)|2,{77,113,0}},
/* 53723 */ {(14<<2)|2,{77,113,0}},
/* 53724 */ {(14<<2)|2,{77,118,0}},
/* 53725 */ {(14<<2)|2,{77,118,0}},
/* 53726 */ {(14<<2)|2,{77,118,0}},
/* 53727 */ {(14<<2)|2,{77,118,0}},
/* 53728 */ {(14<<2)|2,{77,119,0}},
/* 53729 */ {(14<<2)|2,{77,119,0}},
/* 53730 */ {(14<<2)|2,{77,119,0}},
/* 53731 */ {(14<<2)|2,{77,119,0}},
/* 53732 */ {(14<<2)|2,{77,120,0}},
/* 53733 */ {(14<<2)|2,{77,120,0}},
/* 53734 */ {(14<<2)|2,{77,120,0}},
/* 53735 */ {(14<<2)|2,{77,120,0}},
/* 53736 */ {(14<<2)|2,{77,121,0}},
/* 53737 */ {(14<<2)|2,{77,121,0}},
/* 53738 */ {(14<<2)|2,{77,121,0}},
/* 53739 */ {(14<<2)|2,{77,121,0}},
/* 53740 */ {(14<<2)|2,{77,122,0}},
/* 53741 */ {(14<<2)|2,{77,122,0}},
/* 53742 */ {(14<<2)|2,{77,122,0}},
/* 53743 */ {(14<<2)|2,{77,122,0}},
/* 53744 */ {(15<<2)|2,{77,38,0}},
/* 53745 */ {(15<<2)|2,{77,38,0}},
/* 53746 */ {(15<<2)|2,{77,42,0}},
/* 53747 */ {(15<<2)|2,{77,42,0}},
/* 53748 */ {(15<<2)|2,{77,44,0}},
/* 53749 */ {(15<<2)|2,{77,44,0}},
/* 53750 */ {(15<<2)|2,{77,59,0}},
/* 53751 */ {(15<<2)|2,{77,59,0}},
/* 53752 */ {(15<<2)|2,{77,88,0}},
/* 53753 */ {(15<<2)|2,{77,88,0}},
/* 53754 */ {(15<<2)|2,{77,90,0}},
/* 53755 */ {(15<<2)|2,{77,90,0}},
/* 53756 */ {(7<<2)|1,{77,0,0}},
/* 53757 */ {(7<<2)|1,{77,0,0}},
/* 53758 */ {(7<<2)|1,{77,0,0}},
/* 53759 */ {(7<<2)|1,{77,0,0}},
/* 53760 */ {(12<<2)|2,{78,48,0}},
/* 53761 */ {(12<<2)|2,{78,48,0}},
/* 53762 */ {(12<<2)|2,{78,48,0}},
/* 53763 */ {(12<<2)|2,{78,48,0}},
/* 53764 */ {(12<<2)|2,{78,48,0}},
/* 53765 */ {(12<<2)|2,{78,48,0}},
/* 53766 */ {(12<<2)|2,{78,48,0}},
/* 53767 */ {(12<<2)|2,{78,48,0}},
/* 53768 */ {(12<<2)|2,{78,48,0}},
/* 53769 */ {(12<<2)|2,{78,48,0}},
/* 53770 */ {(12<<2)|2,{78,48,0}},
/* 53771 */ {(12<<2)|2,{78,48,0}},
/* 53772 */ {(12<<2)|2,{78,48,0}},
/* 53773 */ {(12<<2)|2,{78,48,0}},
/* 53774 */ {(12<<2)|2,{78,48,0}},
/* 53775 */ {(12<<2)|2,{78,48,0}},
/* 53776 */ {(12<<2)|2,{78,49,0}},
/* 53777 */ {(12<<2)|2,{78,49,0}},
/* 53778 */ {(12<<2)|2,{78,49,0}},
/* 53779 */ {(12<<2)|2,{78,49,0}},
/* 53780 */ {(12<<2)|2,{78,49,0}},
/* 53781 */ {(12<<2)|2,{78,49,0}},
/* 53782 */ {(12<<2)|2,{78,49,0}},
/* 53783 */ {(12<<2)|2,{78,49,0}},
/* 53784 */ {(12<<2)|2,{78,49,0}},
/* 53785 */ {(12<<2)|2,{78,49,0}},
/* 53786 */ {(12<<2)|2,{78,49,0}},
/* 53787 */ {(12<<2)|2,{78,49,0}},
/* 53788 */ {(12<<2)|2,{78,49,0}},
/* 53789 */ {(12<<2)|2,{78,49,0}},
/* 53790 */ {(12<<2)|2,{78,49,0}},
/* 53791 */ {(12<<2)|2,{78,49,0}},
/* 53792 */ {(12<<2)|2,{78,50,0}},
/* 53793 */ {(12<<2)|2,{78,50,0}},
/* 53794 */ {(12<<2)|2,{78,50,0}},
/* 53795 */ {(12<<2)|2,{78,50,0}},
/* 53796 */ {(12<<2)|2,{78,50,0}},
/* 53797 */ {(12<<2)|2,{78,50,0}},
/* 53798 */ {(12<<2)|2,{78,50,0}},
/* 53799 */ {(12<<2)|2,{78,50,0}},
/* 53800 */ {(12<<2)|2,{78,50,0}},
/* 53801 */ {(12<<2)|2,{78,50,0}},
/* 53802 */ {(12<<2)|2,{78,50,0}},
/* 53803 */ {(12<<2)|2,{78,50,0}},
/* 53804 */ {(12<<2)|2,{78,50,0}},
/* 53805 */ {(12<<2)|2,{78,50,0}},
/* 53806 */ {(12<<2)|2,{78,50,0}},
/* 53807 */ {(12<<2)|2,{78,50,0}},
/* 53808 */ {(12<<2)|2,{78,97,0}},
/* 53809 */ {(12<<2)|2,{78,97,0}},
/* 53810 */ {(12<<2)|2,{78,97,0}},
/* 53811 */ {(12<<2)|2,{78,97,0}},
/* 53812 */ {(12<<2)|2,{78,97,0}},
/* 53813 */ {(12<<2)|2,{78,97,0}},
/* 53814 */ {(12<<2)|2,{78,97,0}},
/* 53815 */ {(12<<2)|2,{78,97,0}},
/* 53816 */ {(12<<2)|2,{78,97,0}},
/* 53817 */ {(12<<2)|2,{78,97,0}},
/* 53818 */ {(12<<2)|2,{78,97,0}},
/* 53819 */ {(12<<2)|2,{78,97,0}},
/* 53820 */ {(12<<2)|2,{78,97,0}},
/* 53821 */ {(12<<2)|2,{78,97,0}},
/* 53822 */ {(12<<2)|2,{78,97,0}},
/* 53823 */ {(12<<2)|2,{78,97,0}},
/* 53824 */ {(12<<2)|2,{78,99,0}},
/* 53825 */ {(12<<2)|2,{78,99,0}},
/* 53826 */ {(12<<2)|2,{78,99,0}},
/* 53827 */ {(12<<2)|2,{78,99,0}},
/* 53828 */ {(12<<2)|2,{78,99,0}},
/* 53829 */ {(12<<2)|2,{78,99,0}},
/* 53830 */ {(12<<2)|2,{78,99,0}},
/* 53831 */ {(12<<2)|2,{78,99,0}},
/* 53832 */ {(12<<2)|2,{78,99,0}},
/* 53833 */ {(12<<2)|2,{78,99,0}},
/* 53834 */ {(12<<2)|2,{78,99,0}},
/* 53835 */ {(12<<2)|2,{78,99,0}},
/* 53836 */ {(12<<2)|2,{78,99,0}},
/* 53837 */ {(12<<2)|2,{78,99,0}},
/* 53838 */ {(12<<2)|2,{78,99,0}},
/* 53839 */ {(12<<2)|2,{78,99,0}},
/* 53840 */ {(12<<2)|2,{78,101,0}},
/* 53841 */ {(12<<2)|2,{78,101,0}},
/* 53842 */ {(12<<2)|2,{78,101,0}},
/* 53843 */ {(12<<2)|2,{78,101,0}},
/* 53844 */ {(12<<2)|2,{78,101,0}},
/* 53845 */ {(12<<2)|2,{78,101,0}},
/* 53846 */ {(12<<2)|2,{78,101,0}},
/* 53847 */ {(12<<2)|2,{78,101,0}},
/* 53848 */ {(12<<2)|2,{78,101,0}},
/* 53849 */ {(12<<2)|2,{78,101,0}},
/* 53850 */ {(12<<2)|2,{78,101,0}},
/* 53851 */ {(12<<2)|2,{78,101,0}},
/* 53852 */ {(12<<2)|2,{78,101,0}},
/* 53853 */ {(12<<2)|2,{78,101,0}},
/* 53854 */ {(12<<2)|2,{78,101,0}},
/* 53855 */ {(12<<2)|2,{78,101,0}},
/* 53856 */ {(12<<2)|2,{78,105,0}},
/* 53857 */ {(12<<2)|2,{78,105,0}},
/* 53858 */ {(12<<2)|2,{78,105,0}},
/* 53859 */ {(12<<2)|2,{78,105,0}},
/* 53860 */ {(12<<2)|2,{78,105,0}},
/* 53861 */ {(12<<2)|2,{78,105,0}},
/* 53862 */ {(12<<2)|2,{78,105,0}},
/* 53863 */ {(12<<2)|2,{78,105,0}},
/* 53864 */ {(12<<2)|2,{78,105,0}},
/* 53865 */ {(12<<2)|2,{78,105,0}},
/* 53866 */ {(12<<2)|2,{78,105,0}},
/* 53867 */ {(12<<2)|2,{78,105,0}},
/* 53868 */ {(12<<2)|2,{78,105,0}},
/* 53869 */ {(12<<2)|2,{78,105,0}},
/* 53870 */ {(12<<2)|2,{78,105,0}},
/* 53871 */ {(12<<2)|2,{78,105,0}},
/* 53872 */ {(12<<2)|2,{78,111,0}},
/* 53873 */ {(12<<2)|2,{78,111,0}},
/* 53874 */ {(12<<2)|2,{78,111,0}},
/* 53875 */ {(12<<2)|2,{78,111,0}},
/* 53876 */ {(12<<2)|2,{78,111,0}},
/* 53877 */ {(12<<2)|2,{78,111,0}},
/* 53878 */ {(12<<2)|2,{78,111,0}},
/* 53879 */ {(12<<2)|2,{78,111,0}},
/* 53880 */ {(12<<2)|2,{78,111,0}},
/* 53881 */ {(12<<2)|2,{78,111,0}},
/* 53882 */ {(12<<2)|2,{78,111,0}},
/* 53883 */ {(12<<2)|2,{78,111,0}},
/* 53884 */ {(12<<2)|2,{78,111,0}},
/* 53885 */ {(12<<2)|2,{78,111,0}},
/* 53886 */ {(12<<2)|2,{78,111,0}},
/* 53887 */ {(12<<2)|2,{78,111,0}},
/* 53888 */ {(12<<2)|2,{78,115,0}},
/* 53889 */ {(12<<2)|2,{78,115,0}},
/* 53890 */ {(12<<2)|2,{78,115,0}},
/* 53891 */ {(12<<2)|2,{78,115,0}},
/* 53892 */ {(12<<2)|2,{78,115,0}},
/* 53893 */ {(12<<2)|2,{78,115,0}},
/* 53894 */ {(12<<2)|2,{78,115,0}},
/* 53895 */ {(12<<2)|2,{78,115,0}},
/* 53896 */ {(12<<2)|2,{78,115,0}},
/* 53897 */ {(12<<2)|2,{78,115,0}},
/* 53898 */ {(12<<2)|2,{78,115,0}},
/* 53899 */ {(12<<2)|2,{78,115,0}},
/* 53900 */ {(12<<2)|2,{78,115,0}},
/* 53901 */ {(12<<2)|2,{78,115,0}},
/* 53902 */ {(12<<2)|2,{78,115,0}},
/* 53903 */ {(12<<2)|2,{78,115,0}},
/* 53904 */ {(12<<2)|2,{78,116,0}},
/* 53905 */ {(12<<2)|2,{78,116,0}},
/* 53906 */ {(12<<2)|2,{78,116,0}},
/* 53907 */ {(12<<2)|2,{78,116,0}},
/* 53908 */ {(12<<2)|2,{78,116,0}},
/* 53909 */ {(12<<2)|2,{78,116,0}},
/* 53910 */ {(12<<2)|2,{78,116,0}},
/* 53911 */ {(12<<2)|2,{78,116,0}},
/* 53912 */ {(12<<2)|2,{78,116,0}},
/* 53913 */ {(12<<2)|2,{78,116,0}},
/* 53914 */ {(12<<2)|2,{78,116,0}},
/* 53915 */ {(12<<2)|2,{78,116,0}},
/* 53916 */ {(12<<2)|2,{78,116,0}},
/* 53917 */ {(12<<2)|2,{78,116,0}},
/* 53918 */ {(12<<2)|2,{78,116,0}},
/* 53919 */ {(12<<2)|2,{78,116,0}},
/* 53920 */ {(13<<2)|2,{78,32,0}},
/* 53921 */ {(13<<2)|2,{78,32,0}},
/* 53922 */ {(13<<2)|2,{78,32,0}},
/* 53923 */ {(13<<2)|2,{78,32,0}},
/* 53924 */ {(13<<2)|2,{78,32,0}},
/* 53925 */ {(13<<2)|2,{78,32,0}},
/* 53926 */ {(13<<2)|2,{78,32,0}},
/* 53927 */ {(13<<2)|2,{78,32,0}},
/* 53928 */ {(13<<2)|2,{78,37,0}},
/* 53929 */ {(13<<2)|2,{78,37,0}},
/* 53930 */ {(13<<2)|2,{78,37,0}},
/* 53931 */ {(13<<2)|2,{78,37,0}},
/* 53932 */ {(13<<2)|2,{78,37,0}},
/* 53933 */ {(13<<2)|2,{78,37,0}},
/* 53934 */ {(13<<2)|2,{78,37,0}},
/* 53935 */ {(13<<2)|2,{78,37,0}},
/* 53936 */ {(13<<2)|2,{78,45,0}},
/* 53937 */ {(13<<2)|2,{78,45,0}},
/* 53938 */ {(13<<2)|2,{78,45,0}},
/* 53939 */ {(13<<2)|2,{78,45,0}},
/* 53940 */ {(13<<2)|2,{78,45,0}},
/* 53941 */ {(13<<2)|2,{78,45,0}},
/* 53942 */ {(13<<2)|2,{78,45,0}},
/* 53943 */ {(13<<2)|2,{78,45,0}},
/* 53944 */ {(13<<2)|2,{78,46,0}},
/* 53945 */ {(13<<2)|2,{78,46,0}},
/* 53946 */ {(13<<2)|2,{78,46,0}},
/* 53947 */ {(13<<2)|2,{78,46,0}},
/* 53948 */ {(13<<2)|2,{78,46,0}},
/* 53949 */ {(13<<2)|2,{78,46,0}},
/* 53950 */ {(13<<2)|2,{78,46,0}},
/* 53951 */ {(13<<2)|2,{78,46,0}},
/* 53952 */ {(13<<2)|2,{78,47,0}},
/* 53953 */ {(13<<2)|2,{78,47,0}},
/* 53954 */ {(13<<2)|2,{78,47,0}},
/* 53955 */ {(13<<2)|2,{78,47,0}},
/* 53956 */ {(13<<2)|2,{78,47,0}},
/* 53957 */ {(13<<2)|2,{78,47,0}},
/* 53958 */ {(13<<2)|2,{78,47,0}},
/* 53959 */ {(13<<2)|2,{78,47,0}},
/* 53960 */ {(13<<2)|2,{78,51,0}},
/* 53961 */ {(13<<2)|2,{78,51,0}},
/* 53962 */ {(13<<2)|2,{78,51,0}},
/* 53963 */ {(13<<2)|2,{78,51,0}},
/* 53964 */ {(13<<2)|2,{78,51,0}},
/* 53965 */ {(13<<2)|2,{78,51,0}},
/* 53966 */ {(13<<2)|2,{78,51,0}},
/* 53967 */ {(13<<2)|2,{78,51,0}},
/* 53968 */ {(13<<2)|2,{78,52,0}},
/* 53969 */ {(13<<2)|2,{78,52,0}},
/* 53970 */ {(13<<2)|2,{78,52,0}},
/* 53971 */ {(13<<2)|2,{78,52,0}},
/* 53972 */ {(13<<2)|2,{78,52,0}},
/* 53973 */ {(13<<2)|2,{78,52,0}},
/* 53974 */ {(13<<2)|2,{78,52,0}},
/* 53975 */ {(13<<2)|2,{78,52,0}},
/* 53976 */ {(13<<2)|2,{78,53,0}},
/* 53977 */ {(13<<2)|2,{78,53,0}},
/* 53978 */ {(13<<2)|2,{78,53,0}},
/* 53979 */ {(13<<2)|2,{78,53,0}},
/* 53980 */ {(13<<2)|2,{78,53,0}},
/* 53981 */ {(13<<2)|2,{78,53,0}},
/* 53982 */ {(13<<2)|2,{78,53,0}},
/* 53983 */ {(13<<2)|2,{78,53,0}},
/* 53984 */ {(13<<2)|2,{78,54,0}},
/* 53985 */ {(13<<2)|2,{78,54,0}},
/* 53986 */ {(13<<2)|2,{78,54,0}},
/* 53987 */ {(13<<2)|2,{78,54,0}},
/* 53988 */ {(13<<2)|2,{78,54,0}},
/* 53989 */ {(13<<2)|2,{78,54,0}},
/* 53990 */ {(13<<2)|2,{78,54,0}},
/* 53991 */ {(13<<2)|2,{78,54,0}},
/* 53992 */ {(13<<2)|2,{78,55,0}},
/* 53993 */ {(13<<2)|2,{78,55,0}},
/* 53994 */ {(13<<2)|2,{78,55,0}},
/* 53995 */ {(13<<2)|2,{78,55,0}},
/* 53996 */ {(13<<2)|2,{78,55,0}},
/* 53997 */ {(13<<2)|2,{78,55,0}},
/* 53998 */ {(13<<2)|2,{78,55,0}},
/* 53999 */ {(13<<2)|2,{78,55,0}},
/* 54000 */ {(13<<2)|2,{78,56,0}},
/* 54001 */ {(13<<2)|2,{78,56,0}},
/* 54002 */ {(13<<2)|2,{78,56,0}},
/* 54003 */ {(13<<2)|2,{78,56,0}},
/* 54004 */ {(13<<2)|2,{78,56,0}},
/* 54005 */ {(13<<2)|2,{78,56,0}},
/* 54006 */ {(13<<2)|2,{78,56,0}},
/* 54007 */ {(13<<2)|2,{78,56,0}},
/* 54008 */ {(13<<2)|2,{78,57,0}},
/* 54009 */ {(13<<2)|2,{78,57,0}},
/* 54010 */ {(13<<2)|2,{78,57,0}},
/* 54011 */ {(13<<2)|2,{78,57,0}},
/* 54012 */ {(13<<2)|2,{78,57,0}},
/* 54013 */ {(13<<2)|2,{78,57,0}},
/* 54014 */ {(13<<2)|2,{78,57,0}},
/* 54015 */ {(13<<2)|2,{78,57,0}},
/* 54016 */ {(13<<2)|2,{78,61,0}},
/* 54017 */ {(13<<2)|2,{78,61,0}},
/* 54018 */ {(13<<2)|2,{78,61,0}},
/* 54019 */ {(13<<2)|2,{78,61,0}},
/* 54020 */ {(13<<2)|2,{78,61,0}},
/* 54021 */ {(13<<2)|2,{78,61,0}},
/* 54022 */ {(13<<2)|2,{78,61,0}},
/* 54023 */ {(13<<2)|2,{78,61,0}},
/* 54024 */ {(13<<2)|2,{78,65,0}},
/* 54025 */ {(13<<2)|2,{78,65,0}},
/* 54026 */ {(13<<2)|2,{78,65,0}},
/* 54027 */ {(13<<2)|2,{78,65,0}},
/* 54028 */ {(13<<2)|2,{78,65,0}},
/* 54029 */ {(13<<2)|2,{78,65,0}},
/* 54030 */ {(13<<2)|2,{78,65,0}},
/* 54031 */ {(13<<2)|2,{78,65,0}},
/* 54032 */ {(13<<2)|2,{78,95,0}},
/* 54033 */ {(13<<2)|2,{78,95,0}},
/* 54034 */ {(13<<2)|2,{78,95,0}},
/* 54035 */ {(13<<2)|2,{78,95,0}},
/* 54036 */ {(13<<2)|2,{78,95,0}},
/* 54037 */ {(13<<2)|2,{78,95,0}},
/* 54038 */ {(13<<2)|2,{78,95,0}},
/* 54039 */ {(13<<2)|2,{78,95,0}},
/* 54040 */ {(13<<2)|2,{78,98,0}},
/* 54041 */ {(13<<2)|2,{78,98,0}},
/* 54042 */ {(13<<2)|2,{78,98,0}},
/* 54043 */ {(13<<2)|2,{78,98,0}},
/* 54044 */ {(13<<2)|2,{78,98,0}},
/* 54045 */ {(13<<2)|2,{78,98,0}},
/* 54046 */ {(13<<2)|2,{78,98,0}},
/* 54047 */ {(13<<2)|2,{78,98,0}},
/* 54048 */ {(13<<2)|2,{78,100,0}},
/* 54049 */ {(13<<2)|2,{78,100,0}},
/* 54050 */ {(13<<2)|2,{78,100,0}},
/* 54051 */ {(13<<2)|2,{78,100,0}},
/* 54052 */ {(13<<2)|2,{78,100,0}},
/* 54053 */ {(13<<2)|2,{78,100,0}},
/* 54054 */ {(13<<2)|2,{78,100,0}},
/* 54055 */ {(13<<2)|2,{78,100,0}},
/* 54056 */ {(13<<2)|2,{78,102,0}},
/* 54057 */ {(13<<2)|2,{78,102,0}},
/* 54058 */ {(13<<2)|2,{78,102,0}},
/* 54059 */ {(13<<2)|2,{78,102,0}},
/* 54060 */ {(13<<2)|2,{78,102,0}},
/* 54061 */ {(13<<2)|2,{78,102,0}},
/* 54062 */ {(13<<2)|2,{78,102,0}},
/* 54063 */ {(13<<2)|2,{78,102,0}},
/* 54064 */ {(13<<2)|2,{78,103,0}},
/* 54065 */ {(13<<2)|2,{78,103,0}},
/* 54066 */ {(13<<2)|2,{78,103,0}},
/* 54067 */ {(13<<2)|2,{78,103,0}},
/* 54068 */ {(13<<2)|2,{78,103,0}},
/* 54069 */ {(13<<2)|2,{78,103,0}},
/* 54070 */ {(13<<2)|2,{78,103,0}},
/* 54071 */ {(13<<2)|2,{78,103,0}},
/* 54072 */ {(13<<2)|2,{78,104,0}},
/* 54073 */ {(13<<2)|2,{78,104,0}},
/* 54074 */ {(13<<2)|2,{78,104,0}},
/* 54075 */ {(13<<2)|2,{78,104,0}},
/* 54076 */ {(13<<2)|2,{78,104,0}},
/* 54077 */ {(13<<2)|2,{78,104,0}},
/* 54078 */ {(13<<2)|2,{78,104,0}},
/* 54079 */ {(13<<2)|2,{78,104,0}},
/* 54080 */ {(13<<2)|2,{78,108,0}},
/* 54081 */ {(13<<2)|2,{78,108,0}},
/* 54082 */ {(13<<2)|2,{78,108,0}},
/* 54083 */ {(13<<2)|2,{78,108,0}},
/* 54084 */ {(13<<2)|2,{78,108,0}},
/* 54085 */ {(13<<2)|2,{78,108,0}},
/* 54086 */ {(13<<2)|2,{78,108,0}},
/* 54087 */ {(13<<2)|2,{78,108,0}},
/* 54088 */ {(13<<2)|2,{78,109,0}},
/* 54089 */ {(13<<2)|2,{78,109,0}},
/* 54090 */ {(13<<2)|2,{78,109,0}},
/* 54091 */ {(13<<2)|2,{78,109,0}},
/* 54092 */ {(13<<2)|2,{78,109,0}},
/* 54093 */ {(13<<2)|2,{78,109,0}},
/* 54094 */ {(13<<2)|2,{78,109,0}},
/* 54095 */ {(13<<2)|2,{78,109,0}},
/* 54096 */ {(13<<2)|2,{78,110,0}},
/* 54097 */ {(13<<2)|2,{78,110,0}},
/* 54098 */ {(13<<2)|2,{78,110,0}},
/* 54099 */ {(13<<2)|2,{78,110,0}},
/* 54100 */ {(13<<2)|2,{78,110,0}},
/* 54101 */ {(13<<2)|2,{78,110,0}},
/* 54102 */ {(13<<2)|2,{78,110,0}},
/* 54103 */ {(13<<2)|2,{78,110,0}},
/* 54104 */ {(13<<2)|2,{78,112,0}},
/* 54105 */ {(13<<2)|2,{78,112,0}},
/* 54106 */ {(13<<2)|2,{78,112,0}},
/* 54107 */ {(13<<2)|2,{78,112,0}},
/* 54108 */ {(13<<2)|2,{78,112,0}},
/* 54109 */ {(13<<2)|2,{78,112,0}},
/* 54110 */ {(13<<2)|2,{78,112,0}},
/* 54111 */ {(13<<2)|2,{78,112,0}},
/* 54112 */ {(13<<2)|2,{78,114,0}},
/* 54113 */ {(13<<2)|2,{78,114,0}},
/* 54114 */ {(13<<2)|2,{78,114,0}},
/* 54115 */ {(13<<2)|2,{78,114,0}},
/* 54116 */ {(13<<2)|2,{78,114,0}},
/* 54117 */ {(13<<2)|2,{78,114,0}},
/* 54118 */ {(13<<2)|2,{78,114,0}},
/* 54119 */ {(13<<2)|2,{78,114,0}},
/* 54120 */ {(13<<2)|2,{78,117,0}},
/* 54121 */ {(13<<2)|2,{78,117,0}},
/* 54122 */ {(13<<2)|2,{78,117,0}},
/* 54123 */ {(13<<2)|2,{78,117,0}},
/* 54124 */ {(13<<2)|2,{78,117,0}},
/* 54125 */ {(13<<2)|2,{78,117,0}},
/* 54126 */ {(13<<2)|2,{78,117,0}},
/* 54127 */ {(13<<2)|2,{78,117,0}},
/* 54128 */ {(14<<2)|2,{78,58,0}},
/* 54129 */ {(14<<2)|2,{78,58,0}},
/* 54130 */ {(14<<2)|2,{78,58,0}},
/* 54131 */ {(14<<2)|2,{78,58,0}},
/* 54132 */ {(14<<2)|2,{78,66,0}},
/* 54133 */ {(14<<2)|2,{78,66,0}},
/* 54134 */ {(14<<2)|2,{78,66,0}},
/* 54135 */ {(14<<2)|2,{78,66,0}},
/* 54136 */ {(14<<2)|2,{78,67,0}},
/* 54137 */ {(14<<2)|2,{78,67,0}},
/* 54138 */ {(14<<2)|2,{78,67,0}},
/* 54139 */ {(14<<2)|2,{78,67,0}},
/* 54140 */ {(14<<2)|2,{78,68,0}},
/* 54141 */ {(14<<2)|2,{78,68,0}},
/* 54142 */ {(14<<2)|2,{78,68,0}},
/* 54143 */ {(14<<2)|2,{78,68,0}},
/* 54144 */ {(14<<2)|2,{78,69,0}},
/* 54145 */ {(14<<2)|2,{78,69,0}},
/* 54146 */ {(14<<2)|2,{78,69,0}},
/* 54147 */ {(14<<2)|2,{78,69,0}},
/* 54148 */ {(14<<2)|2,{78,70,0}},
/* 54149 */ {(14<<2)|2,{78,70,0}},
/* 54150 */ {(14<<2)|2,{78,70,0}},
/* 54151 */ {(14<<2)|2,{78,70,0}},
/* 54152 */ {(14<<2)|2,{78,71,0}},
/* 54153 */ {(14<<2)|2,{78,71,0}},
/* 54154 */ {(14<<2)|2,{78,71,0}},
/* 54155 */ {(14<<2)|2,{78,71,0}},
/* 54156 */ {(14<<2)|2,{78,72,0}},
/* 54157 */ {(14<<2)|2,{78,72,0}},
/* 54158 */ {(14<<2)|2,{78,72,0}},
/* 54159 */ {(14<<2)|2,{78,72,0}},
/* 54160 */ {(14<<2)|2,{78,73,0}},
/* 54161 */ {(14<<2)|2,{78,73,0}},
/* 54162 */ {(14<<2)|2,{78,73,0}},
/* 54163 */ {(14<<2)|2,{78,73,0}},
/* 54164 */ {(14<<2)|2,{78,74,0}},
/* 54165 */ {(14<<2)|2,{78,74,0}},
/* 54166 */ {(14<<2)|2,{78,74,0}},
/* 54167 */ {(14<<2)|2,{78,74,0}},
/* 54168 */ {(14<<2)|2,{78,75,0}},
/* 54169 */ {(14<<2)|2,{78,75,0}},
/* 54170 */ {(14<<2)|2,{78,75,0}},
/* 54171 */ {(14<<2)|2,{78,75,0}},
/* 54172 */ {(14<<2)|2,{78,76,0}},
/* 54173 */ {(14<<2)|2,{78,76,0}},
/* 54174 */ {(14<<2)|2,{78,76,0}},
/* 54175 */ {(14<<2)|2,{78,76,0}},
/* 54176 */ {(14<<2)|2,{78,77,0}},
/* 54177 */ {(14<<2)|2,{78,77,0}},
/* 54178 */ {(14<<2)|2,{78,77,0}},
/* 54179 */ {(14<<2)|2,{78,77,0}},
/* 54180 */ {(14<<2)|2,{78,78,0}},
/* 54181 */ {(14<<2)|2,{78,78,0}},
/* 54182 */ {(14<<2)|2,{78,78,0}},
/* 54183 */ {(14<<2)|2,{78,78,0}},
/* 54184 */ {(14<<2)|2,{78,79,0}},
/* 54185 */ {(14<<2)|2,{78,79,0}},
/* 54186 */ {(14<<2)|2,{78,79,0}},
/* 54187 */ {(14<<2)|2,{78,79,0}},
/* 54188 */ {(14<<2)|2,{78,80,0}},
/* 54189 */ {(14<<2)|2,{78,80,0}},
/* 54190 */ {(14<<2)|2,{78,80,0}},
/* 54191 */ {(14<<2)|2,{78,80,0}},
/* 54192 */ {(14<<2)|2,{78,81,0}},
/* 54193 */ {(14<<2)|2,{78,81,0}},
/* 54194 */ {(14<<2)|2,{78,81,0}},
/* 54195 */ {(14<<2)|2,{78,81,0}},
/* 54196 */ {(14<<2)|2,{78,82,0}},
/* 54197 */ {(14<<2)|2,{78,82,0}},
/* 54198 */ {(14<<2)|2,{78,82,0}},
/* 54199 */ {(14<<2)|2,{78,82,0}},
/* 54200 */ {(14<<2)|2,{78,83,0}},
/* 54201 */ {(14<<2)|2,{78,83,0}},
/* 54202 */ {(14<<2)|2,{78,83,0}},
/* 54203 */ {(14<<2)|2,{78,83,0}},
/* 54204 */ {(14<<2)|2,{78,84,0}},
/* 54205 */ {(14<<2)|2,{78,84,0}},
/* 54206 */ {(14<<2)|2,{78,84,0}},
/* 54207 */ {(14<<2)|2,{78,84,0}},
/* 54208 */ {(14<<2)|2,{78,85,0}},
/* 54209 */ {(14<<2)|2,{78,85,0}},
/* 54210 */ {(14<<2)|2,{78,85,0}},
/* 54211 */ {(14<<2)|2,{78,85,0}},
/* 54212 */ {(14<<2)|2,{78,86,0}},
/* 54213 */ {(14<<2)|2,{78,86,0}},
/* 54214 */ {(14<<2)|2,{78,86,0}},
/* 54215 */ {(14<<2)|2,{78,86,0}},
/* 54216 */ {(14<<2)|2,{78,87,0}},
/* 54217 */ {(14<<2)|2,{78,87,0}},
/* 54218 */ {(14<<2)|2,{78,87,0}},
/* 54219 */ {(14<<2)|2,{78,87,0}},
/* 54220 */ {(14<<2)|2,{78,89,0}},
/* 54221 */ {(14<<2)|2,{78,89,0}},
/* 54222 */ {(14<<2)|2,{78,89,0}},
/* 54223 */ {(14<<2)|2,{78,89,0}},
/* 54224 */ {(14<<2)|2,{78,106,0}},
/* 54225 */ {(14<<2)|2,{78,106,0}},
/* 54226 */ {(14<<2)|2,{78,106,0}},
/* 54227 */ {(14<<2)|2,{78,106,0}},
/* 54228 */ {(14<<2)|2,{78,107,0}},
/* 54229 */ {(14<<2)|2,{78,107,0}},
/* 54230 */ {(14<<2)|2,{78,107,0}},
/* 54231 */ {(14<<2)|2,{78,107,0}},
/* 54232 */ {(14<<2)|2,{78,113,0}},
/* 54233 */ {(14<<2)|2,{78,113,0}},
/* 54234 */ {(14<<2)|2,{78,113,0}},
/* 54235 */ {(14<<2)|2,{78,113,0}},
/* 54236 */ {(14<<2)|2,{78,118,0}},
/* 54237 */ {(14<<2)|2,{78,118,0}},
/* 54238 */ {(14<<2)|2,{78,118,0}},
/* 54239 */ {(14<<2)|2,{78,118,0}},
/* 54240 */ {(14<<2)|2,{78,119,0}},
/* 54241 */ {(14<<2)|2,{78,119,0}},
/* 54242 */ {(14<<2)|2,{78,119,0}},
/* 54243 */ {(14<<2)|2,{78,119,0}},
/* 54244 */ {(14<<2)|2,{78,120,0}},
/* 54245 */ {(14<<2)|2,{78,120,0}},
/* 54246 */ {(14<<2)|2,{78,120,0}},
/* 54247 */ {(14<<2)|2,{78,120,0}},
/* 54248 */ {(14<<2)|2,{78,121,0}},
/* 54249 */ {(14<<2)|2,{78,121,0}},
/* 54250 */ {(14<<2)|2,{78,121,0}},
/* 54251 */ {(14<<2)|2,{78,121,0}},
/* 54252 */ {(14<<2)|2,{78,122,0}},
/* 54253 */ {(14<<2)|2,{78,122,0}},
/* 54254 */ {(14<<2)|2,{78,122,0}},
/* 54255 */ {(14<<2)|2,{78,122,0}},
/* 54256 */ {(15<<2)|2,{78,38,0}},
/* 54257 */ {(15<<2)|2,{78,38,0}},
/* 54258 */ {(15<<2)|2,{78,42,0}},
/* 54259 */ {(15<<2)|2,{78,42,0}},
/* 54260 */ {(15<<2)|2,{78,44,0}},
/* 54261 */ {(15<<2)|2,{78,44,0}},
/* 54262 */ {(15<<2)|2,{78,59,0}},
/* 54263 */ {(15<<2)|2,{78,59,0}},
/* 54264 */ {(15<<2)|2,{78,88,0}},
/* 54265 */ {(15<<2)|2,{78,88,0}},
/* 54266 */ {(15<<2)|2,{78,90,0}},
/* 54267 */ {(15<<2)|2,{78,90,0}},
/* 54268 */ {(7<<2)|1,{78,0,0}},
/* 54269 */ {(7<<2)|1,{78,0,0}},
/* 54270 */ {(7<<2)|1,{78,0,0}},
/* 54271 */ {(7<<2)|1,{78,0,0}},
/* 54272 */ {(12<<2)|2,{79,48,0}},
/* 54273 */ {(12<<2)|2,{79,48,0}},
/* 54274 */ {(12<<2)|2,{79,48,0}},
/* 54275 */ {(12<<2)|2,{79,48,0}},
/* 54276 */ {(12<<2)|2,{79,48,0}},
/* 54277 */ {(12<<2)|2,{79,48,0}},
/* 54278 */ {(12<<2)|2,{79,48,0}},
/* 54279 */ {(12<<2)|2,{79,48,0}},
/* 54280 */ {(12<<2)|2,{79,48,0}},
/* 54281 */ {(12<<2)|2,{79,48,0}},
/* 54282 */ {(12<<2)|2,{79,48,0}},
/* 54283 */ {(12<<2)|2,{79,48,0}},
/* 54284 */ {(12<<2)|2,{79,48,0}},
/* 54285 */ {(12<<2)|2,{79,48,0}},
/* 54286 */ {(12<<2)|2,{79,48,0}},
/* 54287 */ {(12<<2)|2,{79,48,0}},
/* 54288 */ {(12<<2)|2,{79,49,0}},
/* 54289 */ {(12<<2)|2,{79,49,0}},
/* 54290 */ {(12<<2)|2,{79,49,0}},
/* 54291 */ {(12<<2)|2,{79,49,0}},
/* 54292 */ {(12<<2)|2,{79,49,0}},
/* 54293 */ {(12<<2)|2,{79,49,0}},
/* 54294 */ {(12<<2)|2,{79,49,0}},
/* 54295 */ {(12<<2)|2,{79,49,0}},
/* 54296 */ {(12<<2)|2,{79,49,0}},
/* 54297 */ {(12<<2)|2,{79,49,0}},
/* 54298 */ {(12<<2)|2,{79,49,0}},
/* 54299 */ {(12<<2)|2,{79,49,0}},
/* 54300 */ {(12<<2)|2,{79,49,0}},
/* 54301 */ {(12<<2)|2,{79,49,0}},
/* 54302 */ {(12<<2)|2,{79,49,0}},
/* 54303 */ {(12<<2)|2,{79,49,0}},
/* 54304 */ {(12<<2)|2,{79,50,0}},
/* 54305 */ {(12<<2)|2,{79,50,0}},
/* 54306 */ {(12<<2)|2,{79,50,0}},
/* 54307 */ {(12<<2)|2,{79,50,0}},
/* 54308 */ {(12<<2)|2,{79,50,0}},
/* 54309 */ {(12<<2)|2,{79,50,0}},
/* 54310 */ {(12<<2)|2,{79,50,0}},
/* 54311 */ {(12<<2)|2,{79,50,0}},
/* 54312 */ {(12<<2)|2,{79,50,0}},
/* 54313 */ {(12<<2)|2,{79,50,0}},
/* 54314 */ {(12<<2)|2,{79,50,0}},
/* 54315 */ {(12<<2)|2,{79,50,0}},
/* 54316 */ {(12<<2)|2,{79,50,0}},
/* 54317 */ {(12<<2)|2,{79,50,0}},
/* 54318 */ {(12<<2)|2,{79,50,0}},
/* 54319 */ {(12<<2)|2,{79,50,0}},
/* 54320 */ {(12<<2)|2,{79,97,0}},
/* 54321 */ {(12<<2)|2,{79,97,0}},
/* 54322 */ {(12<<2)|2,{79,97,0}},
/* 54323 */ {(12<<2)|2,{79,97,0}},
/* 54324 */ {(12<<2)|2,{79,97,0}},
/* 54325 */ {(12<<2)|2,{79,97,0}},
/* 54326 */ {(12<<2)|2,{79,97,0}},
/* 54327 */ {(12<<2)|2,{79,97,0}},
/* 54328 */ {(12<<2)|2,{79,97,0}},
/* 54329 */ {(12<<2)|2,{79,97,0}},
/* 54330 */ {(12<<2)|2,{79,97,0}},
/* 54331 */ {(12<<2)|2,{79,97,0}},
/* 54332 */ {(12<<2)|2,{79,97,0}},
/* 54333 */ {(12<<2)|2,{79,97,0}},
/* 54334 */ {(12<<2)|2,{79,97,0}},
/* 54335 */ {(12<<2)|2,{79,97,0}},
/* 54336 */ {(12<<2)|2,{79,99,0}},
/* 54337 */ {(12<<2)|2,{79,99,0}},
/* 54338 */ {(12<<2)|2,{79,99,0}},
/* 54339 */ {(12<<2)|2,{79,99,0}},
/* 54340 */ {(12<<2)|2,{79,99,0}},
/* 54341 */ {(12<<2)|2,{79,99,0}},
/* 54342 */ {(12<<2)|2,{79,99,0}},
/* 54343 */ {(12<<2)|2,{79,99,0}},
/* 54344 */ {(12<<2)|2,{79,99,0}},
/* 54345 */ {(12<<2)|2,{79,99,0}},
/* 54346 */ {(12<<2)|2,{79,99,0}},
/* 54347 */ {(12<<2)|2,{79,99,0}},
/* 54348 */ {(12<<2)|2,{79,99,0}},
/* 54349 */ {(12<<2)|2,{79,99,0}},
/* 54350 */ {(12<<2)|2,{79,99,0}},
/* 54351 */ {(12<<2)|2,{79,99,0}},
/* 54352 */ {(12<<2)|2,{79,101,0}},
/* 54353 */ {(12<<2)|2,{79,101,0}},
/* 54354 */ {(12<<2)|2,{79,101,0}},
/* 54355 */ {(12<<2)|2,{79,101,0}},
/* 54356 */ {(12<<2)|2,{79,101,0}},
/* 54357 */ {(12<<2)|2,{79,101,0}},
/* 54358 */ {(12<<2)|2,{79,101,0}},
/* 54359 */ {(12<<2)|2,{79,101,0}},
/* 54360 */ {(12<<2)|2,{79,101,0}},
/* 54361 */ {(12<<2)|2,{79,101,0}},
/* 54362 */ {(12<<2)|2,{79,101,0}},
/* 54363 */ {(12<<2)|2,{79,101,0}},
/* 54364 */ {(12<<2)|2,{79,101,0}},
/* 54365 */ {(12<<2)|2,{79,101,0}},
/* 54366 */ {(12<<2)|2,{79,101,0}},
/* 54367 */ {(12<<2)|2,{79,101,0}},
/* 54368 */ {(12<<2)|2,{79,105,0}},
/* 54369 */ {(12<<2)|2,{79,105,0}},
/* 54370 */ {(12<<2)|2,{79,105,0}},
/* 54371 */ {(12<<2)|2,{79,105,0}},
/* 54372 */ {(12<<2)|2,{79,105,0}},
/* 54373 */ {(12<<2)|2,{79,105,0}},
/* 54374 */ {(12<<2)|2,{79,105,0}},
/* 54375 */ {(12<<2)|2,{79,105,0}},
/* 54376 */ {(12<<2)|2,{79,105,0}},
/* 54377 */ {(12<<2)|2,{79,105,0}},
/* 54378 */ {(12<<2)|2,{79,105,0}},
/* 54379 */ {(12<<2)|2,{79,105,0}},
/* 54380 */ {(12<<2)|2,{79,105,0}},
/* 54381 */ {(12<<2)|2,{79,105,0}},
/* 54382 */ {(12<<2)|2,{79,105,0}},
/* 54383 */ {(12<<2)|2,{79,105,0}},
/* 54384 */ {(12<<2)|2,{79,111,0}},
/* 54385 */ {(12<<2)|2,{79,111,0}},
/* 54386 */ {(12<<2)|2,{79,111,0}},
/* 54387 */ {(12<<2)|2,{79,111,0}},
/* 54388 */ {(12<<2)|2,{79,111,0}},
/* 54389 */ {(12<<2)|2,{79,111,0}},
/* 54390 */ {(12<<2)|2,{79,111,0}},
/* 54391 */ {(12<<2)|2,{79,111,0}},
/* 54392 */ {(12<<2)|2,{79,111,0}},
/* 54393 */ {(12<<2)|2,{79,111,0}},
/* 54394 */ {(12<<2)|2,{79,111,0}},
/* 54395 */ {(12<<2)|2,{79,111,0}},
/* 54396 */ {(12<<2)|2,{79,111,0}},
/* 54397 */ {(12<<2)|2,{79,111,0}},
/* 54398 */ {(12<<2)|2,{79,111,0}},
/* 54399 */ {(12<<2)|2,{79,111,0}},
/* 54400 */ {(12<<2)|2,{79,115,0}},
/* 54401 */ {(12<<2)|2,{79,115,0}},
/* 54402 */ {(12<<2)|2,{79,115,0}},
/* 54403 */ {(12<<2)|2,{79,115,0}},
/* 54404 */ {(12<<2)|2,{79,115,0}},
/* 54405 */ {(12<<2)|2,{79,115,0}},
/* 54406 */ {(12<<2)|2,{79,115,0}},
/* 54407 */ {(12<<2)|2,{79,115,0}},
/* 54408 */ {(12<<2)|2,{79,115,0}},
/* 54409 */ {(12<<2)|2,{79,115,0}},
/* 54410 */ {(12<<2)|2,{79,115,0}},
/* 54411 */ {(12<<2)|2,{79,115,0}},
/* 54412 */ {(12<<2)|2,{79,115,0}},
/* 54413 */ {(12<<2)|2,{79,115,0}},
/* 54414 */ {(12<<2)|2,{79,115,0}},
/* 54415 */ {(12<<2)|2,{79,115,0}},
/* 54416 */ {(12<<2)|2,{79,116,0}},
/* 54417 */ {(12<<2)|2,{79,116,0}},
/* 54418 */ {(12<<2)|2,{79,116,0}},
/* 54419 */ {(12<<2)|2,{79,116,0}},
/* 54420 */ {(12<<2)|2,{79,116,0}},
/* 54421 */ {(12<<2)|2,{79,116,0}},
/* 54422 */ {(12<<2)|2,{79,116,0}},
/* 54423 */ {(12<<2)|2,{79,116,0}},
/* 54424 */ {(12<<2)|2,{79,116,0}},
/* 54425 */ {(12<<2)|2,{79,116,0}},
/* 54426 */ {(12<<2)|2,{79,116,0}},
/* 54427 */ {(12<<2)|2,{79,116,0}},
/* 54428 */ {(12<<2)|2,{79,116,0}},
/* 54429 */ {(12<<2)|2,{79,116,0}},
/* 54430 */ {(12<<2)|2,{79,116,0}},
/* 54431 */ {(12<<2)|2,{79,116,0}},
/* 54432 */ {(13<<2)|2,{79,32,0}},
/* 54433 */ {(13<<2)|2,{79,32,0}},
/* 54434 */ {(13<<2)|2,{79,32,0}},
/* 54435 */ {(13<<2)|2,{79,32,0}},
/* 54436 */ {(13<<2)|2,{79,32,0}},
/* 54437 */ {(13<<2)|2,{79,32,0}},
/* 54438 */ {(13<<2)|2,{79,32,0}},
/* 54439 */ {(13<<2)|2,{79,32,0}},
/* 54440 */ {(13<<2)|2,{79,37,0}},
/* 54441 */ {(13<<2)|2,{79,37,0}},
/* 54442 */ {(13<<2)|2,{79,37,0}},
/* 54443 */ {(13<<2)|2,{79,37,0}},
/* 54444 */ {(13<<2)|2,{79,37,0}},
/* 54445 */ {(13<<2)|2,{79,37,0}},
/* 54446 */ {(13<<2)|2,{79,37,0}},
/* 54447 */ {(13<<2)|2,{79,37,0}},
/* 54448 */ {(13<<2)|2,{79,45,0}},
/* 54449 */ {(13<<2)|2,{79,45,0}},
/* 54450 */ {(13<<2)|2,{79,45,0}},
/* 54451 */ {(13<<2)|2,{79,45,0}},
/* 54452 */ {(13<<2)|2,{79,45,0}},
/* 54453 */ {(13<<2)|2,{79,45,0}},
/* 54454 */ {(13<<2)|2,{79,45,0}},
/* 54455 */ {(13<<2)|2,{79,45,0}},
/* 54456 */ {(13<<2)|2,{79,46,0}},
/* 54457 */ {(13<<2)|2,{79,46,0}},
/* 54458 */ {(13<<2)|2,{79,46,0}},
/* 54459 */ {(13<<2)|2,{79,46,0}},
/* 54460 */ {(13<<2)|2,{79,46,0}},
/* 54461 */ {(13<<2)|2,{79,46,0}},
/* 54462 */ {(13<<2)|2,{79,46,0}},
/* 54463 */ {(13<<2)|2,{79,46,0}},
/* 54464 */ {(13<<2)|2,{79,47,0}},
/* 54465 */ {(13<<2)|2,{79,47,0}},
/* 54466 */ {(13<<2)|2,{79,47,0}},
/* 54467 */ {(13<<2)|2,{79,47,0}},
/* 54468 */ {(13<<2)|2,{79,47,0}},
/* 54469 */ {(13<<2)|2,{79,47,0}},
/* 54470 */ {(13<<2)|2,{79,47,0}},
/* 54471 */ {(13<<2)|2,{79,47,0}},
/* 54472 */ {(13<<2)|2,{79,51,0}},
/* 54473 */ {(13<<2)|2,{79,51,0}},
/* 54474 */ {(13<<2)|2,{79,51,0}},
/* 54475 */ {(13<<2)|2,{79,51,0}},
/* 54476 */ {(13<<2)|2,{79,51,0}},
/* 54477 */ {(13<<2)|2,{79,51,0}},
/* 54478 */ {(13<<2)|2,{79,51,0}},
/* 54479 */ {(13<<2)|2,{79,51,0}},
/* 54480 */ {(13<<2)|2,{79,52,0}},
/* 54481 */ {(13<<2)|2,{79,52,0}},
/* 54482 */ {(13<<2)|2,{79,52,0}},
/* 54483 */ {(13<<2)|2,{79,52,0}},
/* 54484 */ {(13<<2)|2,{79,52,0}},
/* 54485 */ {(13<<2)|2,{79,52,0}},
/* 54486 */ {(13<<2)|2,{79,52,0}},
/* 54487 */ {(13<<2)|2,{79,52,0}},
/* 54488 */ {(13<<2)|2,{79,53,0}},
/* 54489 */ {(13<<2)|2,{79,53,0}},
/* 54490 */ {(13<<2)|2,{79,53,0}},
/* 54491 */ {(13<<2)|2,{79,53,0}},
/* 54492 */ {(13<<2)|2,{79,53,0}},
/* 54493 */ {(13<<2)|2,{79,53,0}},
/* 54494 */ {(13<<2)|2,{79,53,0}},
/* 54495 */ {(13<<2)|2,{79,53,0}},
/* 54496 */ {(13<<2)|2,{79,54,0}},
/* 54497 */ {(13<<2)|2,{79,54,0}},
/* 54498 */ {(13<<2)|2,{79,54,0}},
/* 54499 */ {(13<<2)|2,{79,54,0}},
/* 54500 */ {(13<<2)|2,{79,54,0}},
/* 54501 */ {(13<<2)|2,{79,54,0}},
/* 54502 */ {(13<<2)|2,{79,54,0}},
/* 54503 */ {(13<<2)|2,{79,54,0}},
/* 54504 */ {(13<<2)|2,{79,55,0}},
/* 54505 */ {(13<<2)|2,{79,55,0}},
/* 54506 */ {(13<<2)|2,{79,55,0}},
/* 54507 */ {(13<<2)|2,{79,55,0}},
/* 54508 */ {(13<<2)|2,{79,55,0}},
/* 54509 */ {(13<<2)|2,{79,55,0}},
/* 54510 */ {(13<<2)|2,{79,55,0}},
/* 54511 */ {(13<<2)|2,{79,55,0}},
/* 54512 */ {(13<<2)|2,{79,56,0}},
/* 54513 */ {(13<<2)|2,{79,56,0}},
/* 54514 */ {(13<<2)|2,{79,56,0}},
/* 54515 */ {(13<<2)|2,{79,56,0}},
/* 54516 */ {(13<<2)|2,{79,56,0}},
/* 54517 */ {(13<<2)|2,{79,56,0}},
/* 54518 */ {(13<<2)|2,{79,56,0}},
/* 54519 */ {(13<<2)|2,{79,56,0}},
/* 54520 */ {(13<<2)|2,{79,57,0}},
/* 54521 */ {(13<<2)|2,{79,57,0}},
/* 54522 */ {(13<<2)|2,{79,57,0}},
/* 54523 */ {(13<<2)|2,{79,57,0}},
/* 54524 */ {(13<<2)|2,{79,57,0}},
/* 54525 */ {(13<<2)|2,{79,57,0}},
/* 54526 */ {(13<<2)|2,{79,57,0}},
/* 54527 */ {(13<<2)|2,{79,57,0}},
/* 54528 */ {(13<<2)|2,{79,61,0}},
/* 54529 */ {(13<<2)|2,{79,61,0}},
/* 54530 */ {(13<<2)|2,{79,61,0}},
/* 54531 */ {(13<<2)|2,{79,61,0}},
/* 54532 */ {(13<<2)|2,{79,61,0}},
/* 54533 */ {(13<<2)|2,{79,61,0}},
/* 54534 */ {(13<<2)|2,{79,61,0}},
/* 54535 */ {(13<<2)|2,{79,61,0}},
/* 54536 */ {(13<<2)|2,{79,65,0}},
/* 54537 */ {(13<<2)|2,{79,65,0}},
/* 54538 */ {(13<<2)|2,{79,65,0}},
/* 54539 */ {(13<<2)|2,{79,65,0}},
/* 54540 */ {(13<<2)|2,{79,65,0}},
/* 54541 */ {(13<<2)|2,{79,65,0}},
/* 54542 */ {(13<<2)|2,{79,65,0}},
/* 54543 */ {(13<<2)|2,{79,65,0}},
/* 54544 */ {(13<<2)|2,{79,95,0}},
/* 54545 */ {(13<<2)|2,{79,95,0}},
/* 54546 */ {(13<<2)|2,{79,95,0}},
/* 54547 */ {(13<<2)|2,{79,95,0}},
/* 54548 */ {(13<<2)|2,{79,95,0}},
/* 54549 */ {(13<<2)|2,{79,95,0}},
/* 54550 */ {(13<<2)|2,{79,95,0}},
/* 54551 */ {(13<<2)|2,{79,95,0}},
/* 54552 */ {(13<<2)|2,{79,98,0}},
/* 54553 */ {(13<<2)|2,{79,98,0}},
/* 54554 */ {(13<<2)|2,{79,98,0}},
/* 54555 */ {(13<<2)|2,{79,98,0}},
/* 54556 */ {(13<<2)|2,{79,98,0}},
/* 54557 */ {(13<<2)|2,{79,98,0}},
/* 54558 */ {(13<<2)|2,{79,98,0}},
/* 54559 */ {(13<<2)|2,{79,98,0}},
/* 54560 */ {(13<<2)|2,{79,100,0}},
/* 54561 */ {(13<<2)|2,{79,100,0}},
/* 54562 */ {(13<<2)|2,{79,100,0}},
/* 54563 */ {(13<<2)|2,{79,100,0}},
/* 54564 */ {(13<<2)|2,{79,100,0}},
/* 54565 */ {(13<<2)|2,{79,100,0}},
/* 54566 */ {(13<<2)|2,{79,100,0}},
/* 54567 */ {(13<<2)|2,{79,100,0}},
/* 54568 */ {(13<<2)|2,{79,102,0}},
/* 54569 */ {(13<<2)|2,{79,102,0}},
/* 54570 */ {(13<<2)|2,{79,102,0}},
/* 54571 */ {(13<<2)|2,{79,102,0}},
/* 54572 */ {(13<<2)|2,{79,102,0}},
/* 54573 */ {(13<<2)|2,{79,102,0}},
/* 54574 */ {(13<<2)|2,{79,102,0}},
/* 54575 */ {(13<<2)|2,{79,102,0}},
/* 54576 */ {(13<<2)|2,{79,103,0}},
/* 54577 */ {(13<<2)|2,{79,103,0}},
/* 54578 */ {(13<<2)|2,{79,103,0}},
/* 54579 */ {(13<<2)|2,{79,103,0}},
/* 54580 */ {(13<<2)|2,{79,103,0}},
/* 54581 */ {(13<<2)|2,{79,103,0}},
/* 54582 */ {(13<<2)|2,{79,103,0}},
/* 54583 */ {(13<<2)|2,{79,103,0}},
/* 54584 */ {(13<<2)|2,{79,104,0}},
/* 54585 */ {(13<<2)|2,{79,104,0}},
/* 54586 */ {(13<<2)|2,{79,104,0}},
/* 54587 */ {(13<<2)|2,{79,104,0}},
/* 54588 */ {(13<<2)|2,{79,104,0}},
/* 54589 */ {(13<<2)|2,{79,104,0}},
/* 54590 */ {(13<<2)|2,{79,104,0}},
/* 54591 */ {(13<<2)|2,{79,104,0}},
/* 54592 */ {(13<<2)|2,{79,108,0}},
/* 54593 */ {(13<<2)|2,{79,108,0}},
/* 54594 */ {(13<<2)|2,{79,108,0}},
/* 54595 */ {(13<<2)|2,{79,108,0}},
/* 54596 */ {(13<<2)|2,{79,108,0}},
/* 54597 */ {(13<<2)|2,{79,108,0}},
/* 54598 */ {(13<<2)|2,{79,108,0}},
/* 54599 */ {(13<<2)|2,{79,108,0}},
/* 54600 */ {(13<<2)|2,{79,109,0}},
/* 54601 */ {(13<<2)|2,{79,109,0}},
/* 54602 */ {(13<<2)|2,{79,109,0}},
/* 54603 */ {(13<<2)|2,{79,109,0}},
/* 54604 */ {(13<<2)|2,{79,109,0}},
/* 54605 */ {(13<<2)|2,{79,109,0}},
/* 54606 */ {(13<<2)|2,{79,109,0}},
/* 54607 */ {(13<<2)|2,{79,109,0}},
/* 54608 */ {(13<<2)|2,{79,110,0}},
/* 54609 */ {(13<<2)|2,{79,110,0}},
/* 54610 */ {(13<<2)|2,{79,110,0}},
/* 54611 */ {(13<<2)|2,{79,110,0}},
/* 54612 */ {(13<<2)|2,{79,110,0}},
/* 54613 */ {(13<<2)|2,{79,110,0}},
/* 54614 */ {(13<<2)|2,{79,110,0}},
/* 54615 */ {(13<<2)|2,{79,110,0}},
/* 54616 */ {(13<<2)|2,{79,112,0}},
/* 54617 */ {(13<<2)|2,{79,112,0}},
/* 54618 */ {(13<<2)|2,{79,112,0}},
/* 54619 */ {(13<<2)|2,{79,112,0}},
/* 54620 */ {(13<<2)|2,{79,112,0}},
/* 54621 */ {(13<<2)|2,{79,112,0}},
/* 54622 */ {(13<<2)|2,{79,112,0}},
/* 54623 */ {(13<<2)|2,{79,112,0}},
/* 54624 */ {(13<<2)|2,{79,114,0}},
/* 54625 */ {(13<<2)|2,{79,114,0}},
/* 54626 */ {(13<<2)|2,{79,114,0}},
/* 54627 */ {(13<<2)|2,{79,114,0}},
/* 54628 */ {(13<<2)|2,{79,114,0}},
/* 54629 */ {(13<<2)|2,{79,114,0}},
/* 54630 */ {(13<<2)|2,{79,114,0}},
/* 54631 */ {(13<<2)|2,{79,114,0}},
/* 54632 */ {(13<<2)|2,{79,117,0}},
/* 54633 */ {(13<<2)|2,{79,117,0}},
/* 54634 */ {(13<<2)|2,{79,117,0}},
/* 54635 */ {(13<<2)|2,{79,117,0}},
/* 54636 */ {(13<<2)|2,{79,117,0}},
/* 54637 */ {(13<<2)|2,{79,117,0}},
/* 54638 */ {(13<<2)|2,{79,117,0}},
/* 54639 */ {(13<<2)|2,{79,117,0}},
/* 54640 */ {(14<<2)|2,{79,58,0}},
/* 54641 */ {(14<<2)|2,{79,58,0}},
/* 54642 */ {(14<<2)|2,{79,58,0}},
/* 54643 */ {(14<<2)|2,{79,58,0}},
/* 54644 */ {(14<<2)|2,{79,66,0}},
/* 54645 */ {(14<<2)|2,{79,66,0}},
/* 54646 */ {(14<<2)|2,{79,66,0}},
/* 54647 */ {(14<<2)|2,{79,66,0}},
/* 54648 */ {(14<<2)|2,{79,67,0}},
/* 54649 */ {(14<<2)|2,{79,67,0}},
/* 54650 */ {(14<<2)|2,{79,67,0}},
/* 54651 */ {(14<<2)|2,{79,67,0}},
/* 54652 */ {(14<<2)|2,{79,68,0}},
/* 54653 */ {(14<<2)|2,{79,68,0}},
/* 54654 */ {(14<<2)|2,{79,68,0}},
/* 54655 */ {(14<<2)|2,{79,68,0}},
/* 54656 */ {(14<<2)|2,{79,69,0}},
/* 54657 */ {(14<<2)|2,{79,69,0}},
/* 54658 */ {(14<<2)|2,{79,69,0}},
/* 54659 */ {(14<<2)|2,{79,69,0}},
/* 54660 */ {(14<<2)|2,{79,70,0}},
/* 54661 */ {(14<<2)|2,{79,70,0}},
/* 54662 */ {(14<<2)|2,{79,70,0}},
/* 54663 */ {(14<<2)|2,{79,70,0}},
/* 54664 */ {(14<<2)|2,{79,71,0}},
/* 54665 */ {(14<<2)|2,{79,71,0}},
/* 54666 */ {(14<<2)|2,{79,71,0}},
/* 54667 */ {(14<<2)|2,{79,71,0}},
/* 54668 */ {(14<<2)|2,{79,72,0}},
/* 54669 */ {(14<<2)|2,{79,72,0}},
/* 54670 */ {(14<<2)|2,{79,72,0}},
/* 54671 */ {(14<<2)|2,{79,72,0}},
/* 54672 */ {(14<<2)|2,{79,73,0}},
/* 54673 */ {(14<<2)|2,{79,73,0}},
/* 54674 */ {(14<<2)|2,{79,73,0}},
/* 54675 */ {(14<<2)|2,{79,73,0}},
/* 54676 */ {(14<<2)|2,{79,74,0}},
/* 54677 */ {(14<<2)|2,{79,74,0}},
/* 54678 */ {(14<<2)|2,{79,74,0}},
/* 54679 */ {(14<<2)|2,{79,74,0}},
/* 54680 */ {(14<<2)|2,{79,75,0}},
/* 54681 */ {(14<<2)|2,{79,75,0}},
/* 54682 */ {(14<<2)|2,{79,75,0}},
/* 54683 */ {(14<<2)|2,{79,75,0}},
/* 54684 */ {(14<<2)|2,{79,76,0}},
/* 54685 */ {(14<<2)|2,{79,76,0}},
/* 54686 */ {(14<<2)|2,{79,76,0}},
/* 54687 */ {(14<<2)|2,{79,76,0}},
/* 54688 */ {(14<<2)|2,{79,77,0}},
/* 54689 */ {(14<<2)|2,{79,77,0}},
/* 54690 */ {(14<<2)|2,{79,77,0}},
/* 54691 */ {(14<<2)|2,{79,77,0}},
/* 54692 */ {(14<<2)|2,{79,78,0}},
/* 54693 */ {(14<<2)|2,{79,78,0}},
/* 54694 */ {(14<<2)|2,{79,78,0}},
/* 54695 */ {(14<<2)|2,{79,78,0}},
/* 54696 */ {(14<<2)|2,{79,79,0}},
/* 54697 */ {(14<<2)|2,{79,79,0}},
/* 54698 */ {(14<<2)|2,{79,79,0}},
/* 54699 */ {(14<<2)|2,{79,79,0}},
/* 54700 */ {(14<<2)|2,{79,80,0}},
/* 54701 */ {(14<<2)|2,{79,80,0}},
/* 54702 */ {(14<<2)|2,{79,80,0}},
/* 54703 */ {(14<<2)|2,{79,80,0}},
/* 54704 */ {(14<<2)|2,{79,81,0}},
/* 54705 */ {(14<<2)|2,{79,81,0}},
/* 54706 */ {(14<<2)|2,{79,81,0}},
/* 54707 */ {(14<<2)|2,{79,81,0}},
/* 54708 */ {(14<<2)|2,{79,82,0}},
/* 54709 */ {(14<<2)|2,{79,82,0}},
/* 54710 */ {(14<<2)|2,{79,82,0}},
/* 54711 */ {(14<<2)|2,{79,82,0}},
/* 54712 */ {(14<<2)|2,{79,83,0}},
/* 54713 */ {(14<<2)|2,{79,83,0}},
/* 54714 */ {(14<<2)|2,{79,83,0}},
/* 54715 */ {(14<<2)|2,{79,83,0}},
/* 54716 */ {(14<<2)|2,{79,84,0}},
/* 54717 */ {(14<<2)|2,{79,84,0}},
/* 54718 */ {(14<<2)|2,{79,84,0}},
/* 54719 */ {(14<<2)|2,{79,84,0}},
/* 54720 */ {(14<<2)|2,{79,85,0}},
/* 54721 */ {(14<<2)|2,{79,85,0}},
/* 54722 */ {(14<<2)|2,{79,85,0}},
/* 54723 */ {(14<<2)|2,{79,85,0}},
/* 54724 */ {(14<<2)|2,{79,86,0}},
/* 54725 */ {(14<<2)|2,{79,86,0}},
/* 54726 */ {(14<<2)|2,{79,86,0}},
/* 54727 */ {(14<<2)|2,{79,86,0}},
/* 54728 */ {(14<<2)|2,{79,87,0}},
/* 54729 */ {(14<<2)|2,{79,87,0}},
/* 54730 */ {(14<<2)|2,{79,87,0}},
/* 54731 */ {(14<<2)|2,{79,87,0}},
/* 54732 */ {(14<<2)|2,{79,89,0}},
/* 54733 */ {(14<<2)|2,{79,89,0}},
/* 54734 */ {(14<<2)|2,{79,89,0}},
/* 54735 */ {(14<<2)|2,{79,89,0}},
/* 54736 */ {(14<<2)|2,{79,106,0}},
/* 54737 */ {(14<<2)|2,{79,106,0}},
/* 54738 */ {(14<<2)|2,{79,106,0}},
/* 54739 */ {(14<<2)|2,{79,106,0}},
/* 54740 */ {(14<<2)|2,{79,107,0}},
/* 54741 */ {(14<<2)|2,{79,107,0}},
/* 54742 */ {(14<<2)|2,{79,107,0}},
/* 54743 */ {(14<<2)|2,{79,107,0}},
/* 54744 */ {(14<<2)|2,{79,113,0}},
/* 54745 */ {(14<<2)|2,{79,113,0}},
/* 54746 */ {(14<<2)|2,{79,113,0}},
/* 54747 */ {(14<<2)|2,{79,113,0}},
/* 54748 */ {(14<<2)|2,{79,118,0}},
/* 54749 */ {(14<<2)|2,{79,118,0}},
/* 54750 */ {(14<<2)|2,{79,118,0}},
/* 54751 */ {(14<<2)|2,{79,118,0}},
/* 54752 */ {(14<<2)|2,{79,119,0}},
/* 54753 */ {(14<<2)|2,{79,119,0}},
/* 54754 */ {(14<<2)|2,{79,119,0}},
/* 54755 */ {(14<<2)|2,{79,119,0}},
/* 54756 */ {(14<<2)|2,{79,120,0}},
/* 54757 */ {(14<<2)|2,{79,120,0}},
/* 54758 */ {(14<<2)|2,{79,120,0}},
/* 54759 */ {(14<<2)|2,{79,120,0}},
/* 54760 */ {(14<<2)|2,{79,121,0}},
/* 54761 */ {(14<<2)|2,{79,121,0}},
/* 54762 */ {(14<<2)|2,{79,121,0}},
/* 54763 */ {(14<<2)|2,{79,121,0}},
/* 54764 */ {(14<<2)|2,{79,122,0}},
/* 54765 */ {(14<<2)|2,{79,122,0}},
/* 54766 */ {(14<<2)|2,{79,122,0}},
/* 54767 */ {(14<<2)|2,{79,122,0}},
/* 54768 */ {(15<<2)|2,{79,38,0}},
/* 54769 */ {(15<<2)|2,{79,38,0}},
/* 54770 */ {(15<<2)|2,{79,42,0}},
/* 54771 */ {(15<<2)|2,{79,42,0}},
/* 54772 */ {(15<<2)|2,{79,44,0}},
/* 54773 */ {(15<<2)|2,{79,44,0}},
/* 54774 */ {(15<<2)|2,{79,59,0}},
/* 54775 */ {(15<<2)|2,{79,59,0}},
/* 54776 */ {(15<<2)|2,{79,88,0}},
/* 54777 */ {(15<<2)|2,{79,88,0}},
/* 54778 */ {(15<<2)|2,{79,90,0}},
/* 54779 */ {(15<<2)|2,{79,90,0}},
/* 54780 */ {(7<<2)|1,{79,0,0}},
/* 54781 */ {(7<<2)|1,{79,0,0}},
/* 54782 */ {(7<<2)|1,{79,0,0}},
/* 54783 */ {(7<<2)|1,{79,0,0}},
/* 54784 */ {(12<<2)|2,{80,48,0}},
/* 54785 */ {(12<<2)|2,{80,48,0}},
/* 54786 */ {(12<<2)|2,{80,48,0}},
/* 54787 */ {(12<<2)|2,{80,48,0}},
/* 54788 */ {(12<<2)|2,{80,48,0}},
/* 54789 */ {(12<<2)|2,{80,48,0}},
/* 54790 */ {(12<<2)|2,{80,48,0}},
/* 54791 */ {(12<<2)|2,{80,48,0}},
/* 54792 */ {(12<<2)|2,{80,48,0}},
/* 54793 */ {(12<<2)|2,{80,48,0}},
/* 54794 */ {(12<<2)|2,{80,48,0}},
/* 54795 */ {(12<<2)|2,{80,48,0}},
/* 54796 */ {(12<<2)|2,{80,48,0}},
/* 54797 */ {(12<<2)|2,{80,48,0}},
/* 54798 */ {(12<<2)|2,{80,48,0}},
/* 54799 */ {(12<<2)|2,{80,48,0}},
/* 54800 */ {(12<<2)|2,{80,49,0}},
/* 54801 */ {(12<<2)|2,{80,49,0}},
/* 54802 */ {(12<<2)|2,{80,49,0}},
/* 54803 */ {(12<<2)|2,{80,49,0}},
/* 54804 */ {(12<<2)|2,{80,49,0}},
/* 54805 */ {(12<<2)|2,{80,49,0}},
/* 54806 */ {(12<<2)|2,{80,49,0}},
/* 54807 */ {(12<<2)|2,{80,49,0}},
/* 54808 */ {(12<<2)|2,{80,49,0}},
/* 54809 */ {(12<<2)|2,{80,49,0}},
/* 54810 */ {(12<<2)|2,{80,49,0}},
/* 54811 */ {(12<<2)|2,{80,49,0}},
/* 54812 */ {(12<<2)|2,{80,49,0}},
/* 54813 */ {(12<<2)|2,{80,49,0}},
/* 54814 */ {(12<<2)|2,{80,49,0}},
/* 54815 */ {(12<<2)|2,{80,49,0}},
/* 54816 */ {(12<<2)|2,{80,50,0}},
/* 54817 */ {(12<<2)|2,{80,50,0}},
/* 54818 */ {(12<<2)|2,{80,50,0}},
/* 54819 */ {(12<<2)|2,{80,50,0}},
/* 54820 */ {(12<<2)|2,{80,50,0}},
/* 54821 */ {(12<<2)|2,{80,50,0}},
/* 54822 */ {(12<<2)|2,{80,50,0}},
/* 54823 */ {(12<<2)|2,{80,50,0}},
/* 54824 */ {(12<<2)|2,{80,50,0}},
/* 54825 */ {(12<<2)|2,{80,50,0}},
/* 54826 */ {(12<<2)|2,{80,50,0}},
/* 54827 */ {(12<<2)|2,{80,50,0}},
/* 54828 */ {(12<<2)|2,{80,50,0}},
/* 54829 */ {(12<<2)|2,{80,50,0}},
/* 54830 */ {(12<<2)|2,{80,50,0}},
/* 54831 */ {(12<<2)|2,{80,50,0}},
/* 54832 */ {(12<<2)|2,{80,97,0}},
/* 54833 */ {(12<<2)|2,{80,97,0}},
/* 54834 */ {(12<<2)|2,{80,97,0}},
/* 54835 */ {(12<<2)|2,{80,97,0}},
/* 54836 */ {(12<<2)|2,{80,97,0}},
/* 54837 */ {(12<<2)|2,{80,97,0}},
/* 54838 */ {(12<<2)|2,{80,97,0}},
/* 54839 */ {(12<<2)|2,{80,97,0}},
/* 54840 */ {(12<<2)|2,{80,97,0}},
/* 54841 */ {(12<<2)|2,{80,97,0}},
/* 54842 */ {(12<<2)|2,{80,97,0}},
/* 54843 */ {(12<<2)|2,{80,97,0}},
/* 54844 */ {(12<<2)|2,{80,97,0}},
/* 54845 */ {(12<<2)|2,{80,97,0}},
/* 54846 */ {(12<<2)|2,{80,97,0}},
/* 54847 */ {(12<<2)|2,{80,97,0}},
/* 54848 */ {(12<<2)|2,{80,99,0}},
/* 54849 */ {(12<<2)|2,{80,99,0}},
/* 54850 */ {(12<<2)|2,{80,99,0}},
/* 54851 */ {(12<<2)|2,{80,99,0}},
/* 54852 */ {(12<<2)|2,{80,99,0}},
/* 54853 */ {(12<<2)|2,{80,99,0}},
/* 54854 */ {(12<<2)|2,{80,99,0}},
/* 54855 */ {(12<<2)|2,{80,99,0}},
/* 54856 */ {(12<<2)|2,{80,99,0}},
/* 54857 */ {(12<<2)|2,{80,99,0}},
/* 54858 */ {(12<<2)|2,{80,99,0}},
/* 54859 */ {(12<<2)|2,{80,99,0}},
/* 54860 */ {(12<<2)|2,{80,99,0}},
/* 54861 */ {(12<<2)|2,{80,99,0}},
/* 54862 */ {(12<<2)|2,{80,99,0}},
/* 54863 */ {(12<<2)|2,{80,99,0}},
/* 54864 */ {(12<<2)|2,{80,101,0}},
/* 54865 */ {(12<<2)|2,{80,101,0}},
/* 54866 */ {(12<<2)|2,{80,101,0}},
/* 54867 */ {(12<<2)|2,{80,101,0}},
/* 54868 */ {(12<<2)|2,{80,101,0}},
/* 54869 */ {(12<<2)|2,{80,101,0}},
/* 54870 */ {(12<<2)|2,{80,101,0}},
/* 54871 */ {(12<<2)|2,{80,101,0}},
/* 54872 */ {(12<<2)|2,{80,101,0}},
/* 54873 */ {(12<<2)|2,{80,101,0}},
/* 54874 */ {(12<<2)|2,{80,101,0}},
/* 54875 */ {(12<<2)|2,{80,101,0}},
/* 54876 */ {(12<<2)|2,{80,101,0}},
/* 54877 */ {(12<<2)|2,{80,101,0}},
/* 54878 */ {(12<<2)|2,{80,101,0}},
/* 54879 */ {(12<<2)|2,{80,101,0}},
/* 54880 */ {(12<<2)|2,{80,105,0}},
/* 54881 */ {(12<<2)|2,{80,105,0}},
/* 54882 */ {(12<<2)|2,{80,105,0}},
/* 54883 */ {(12<<2)|2,{80,105,0}},
/* 54884 */ {(12<<2)|2,{80,105,0}},
/* 54885 */ {(12<<2)|2,{80,105,0}},
/* 54886 */ {(12<<2)|2,{80,105,0}},
/* 54887 */ {(12<<2)|2,{80,105,0}},
/* 54888 */ {(12<<2)|2,{80,105,0}},
/* 54889 */ {(12<<2)|2,{80,105,0}},
/* 54890 */ {(12<<2)|2,{80,105,0}},
/* 54891 */ {(12<<2)|2,{80,105,0}},
/* 54892 */ {(12<<2)|2,{80,105,0}},
/* 54893 */ {(12<<2)|2,{80,105,0}},
/* 54894 */ {(12<<2)|2,{80,105,0}},
/* 54895 */ {(12<<2)|2,{80,105,0}},
/* 54896 */ {(12<<2)|2,{80,111,0}},
/* 54897 */ {(12<<2)|2,{80,111,0}},
/* 54898 */ {(12<<2)|2,{80,111,0}},
/* 54899 */ {(12<<2)|2,{80,111,0}},
/* 54900 */ {(12<<2)|2,{80,111,0}},
/* 54901 */ {(12<<2)|2,{80,111,0}},
/* 54902 */ {(12<<2)|2,{80,111,0}},
/* 54903 */ {(12<<2)|2,{80,111,0}},
/* 54904 */ {(12<<2)|2,{80,111,0}},
/* 54905 */ {(12<<2)|2,{80,111,0}},
/* 54906 */ {(12<<2)|2,{80,111,0}},
/* 54907 */ {(12<<2)|2,{80,111,0}},
/* 54908 */ {(12<<2)|2,{80,111,0}},
/* 54909 */ {(12<<2)|2,{80,111,0}},
/* 54910 */ {(12<<2)|2,{80,111,0}},
/* 54911 */ {(12<<2)|2,{80,111,0}},
/* 54912 */ {(12<<2)|2,{80,115,0}},
/* 54913 */ {(12<<2)|2,{80,115,0}},
/* 54914 */ {(12<<2)|2,{80,115,0}},
/* 54915 */ {(12<<2)|2,{80,115,0}},
/* 54916 */ {(12<<2)|2,{80,115,0}},
/* 54917 */ {(12<<2)|2,{80,115,0}},
/* 54918 */ {(12<<2)|2,{80,115,0}},
/* 54919 */ {(12<<2)|2,{80,115,0}},
/* 54920 */ {(12<<2)|2,{80,115,0}},
/* 54921 */ {(12<<2)|2,{80,115,0}},
/* 54922 */ {(12<<2)|2,{80,115,0}},
/* 54923 */ {(12<<2)|2,{80,115,0}},
/* 54924 */ {(12<<2)|2,{80,115,0}},
/* 54925 */ {(12<<2)|2,{80,115,0}},
/* 54926 */ {(12<<2)|2,{80,115,0}},
/* 54927 */ {(12<<2)|2,{80,115,0}},
/* 54928 */ {(12<<2)|2,{80,116,0}},
/* 54929 */ {(12<<2)|2,{80,116,0}},
/* 54930 */ {(12<<2)|2,{80,116,0}},
/* 54931 */ {(12<<2)|2,{80,116,0}},
/* 54932 */ {(12<<2)|2,{80,116,0}},
/* 54933 */ {(12<<2)|2,{80,116,0}},
/* 54934 */ {(12<<2)|2,{80,116,0}},
/* 54935 */ {(12<<2)|2,{80,116,0}},
/* 54936 */ {(12<<2)|2,{80,116,0}},
/* 54937 */ {(12<<2)|2,{80,116,0}},
/* 54938 */ {(12<<2)|2,{80,116,0}},
/* 54939 */ {(12<<2)|2,{80,116,0}},
/* 54940 */ {(12<<2)|2,{80,116,0}},
/* 54941 */ {(12<<2)|2,{80,116,0}},
/* 54942 */ {(12<<2)|2,{80,116,0}},
/* 54943 */ {(12<<2)|2,{80,116,0}},
/* 54944 */ {(13<<2)|2,{80,32,0}},
/* 54945 */ {(13<<2)|2,{80,32,0}},
/* 54946 */ {(13<<2)|2,{80,32,0}},
/* 54947 */ {(13<<2)|2,{80,32,0}},
/* 54948 */ {(13<<2)|2,{80,32,0}},
/* 54949 */ {(13<<2)|2,{80,32,0}},
/* 54950 */ {(13<<2)|2,{80,32,0}},
/* 54951 */ {(13<<2)|2,{80,32,0}},
/* 54952 */ {(13<<2)|2,{80,37,0}},
/* 54953 */ {(13<<2)|2,{80,37,0}},
/* 54954 */ {(13<<2)|2,{80,37,0}},
/* 54955 */ {(13<<2)|2,{80,37,0}},
/* 54956 */ {(13<<2)|2,{80,37,0}},
/* 54957 */ {(13<<2)|2,{80,37,0}},
/* 54958 */ {(13<<2)|2,{80,37,0}},
/* 54959 */ {(13<<2)|2,{80,37,0}},
/* 54960 */ {(13<<2)|2,{80,45,0}},
/* 54961 */ {(13<<2)|2,{80,45,0}},
/* 54962 */ {(13<<2)|2,{80,45,0}},
/* 54963 */ {(13<<2)|2,{80,45,0}},
/* 54964 */ {(13<<2)|2,{80,45,0}},
/* 54965 */ {(13<<2)|2,{80,45,0}},
/* 54966 */ {(13<<2)|2,{80,45,0}},
/* 54967 */ {(13<<2)|2,{80,45,0}},
/* 54968 */ {(13<<2)|2,{80,46,0}},
/* 54969 */ {(13<<2)|2,{80,46,0}},
/* 54970 */ {(13<<2)|2,{80,46,0}},
/* 54971 */ {(13<<2)|2,{80,46,0}},
/* 54972 */ {(13<<2)|2,{80,46,0}},
/* 54973 */ {(13<<2)|2,{80,46,0}},
/* 54974 */ {(13<<2)|2,{80,46,0}},
/* 54975 */ {(13<<2)|2,{80,46,0}},
/* 54976 */ {(13<<2)|2,{80,47,0}},
/* 54977 */ {(13<<2)|2,{80,47,0}},
/* 54978 */ {(13<<2)|2,{80,47,0}},
/* 54979 */ {(13<<2)|2,{80,47,0}},
/* 54980 */ {(13<<2)|2,{80,47,0}},
/* 54981 */ {(13<<2)|2,{80,47,0}},
/* 54982 */ {(13<<2)|2,{80,47,0}},
/* 54983 */ {(13<<2)|2,{80,47,0}},
/* 54984 */ {(13<<2)|2,{80,51,0}},
/* 54985 */ {(13<<2)|2,{80,51,0}},
/* 54986 */ {(13<<2)|2,{80,51,0}},
/* 54987 */ {(13<<2)|2,{80,51,0}},
/* 54988 */ {(13<<2)|2,{80,51,0}},
/* 54989 */ {(13<<2)|2,{80,51,0}},
/* 54990 */ {(13<<2)|2,{80,51,0}},
/* 54991 */ {(13<<2)|2,{80,51,0}},
/* 54992 */ {(13<<2)|2,{80,52,0}},
/* 54993 */ {(13<<2)|2,{80,52,0}},
/* 54994 */ {(13<<2)|2,{80,52,0}},
/* 54995 */ {(13<<2)|2,{80,52,0}},
/* 54996 */ {(13<<2)|2,{80,52,0}},
/* 54997 */ {(13<<2)|2,{80,52,0}},
/* 54998 */ {(13<<2)|2,{80,52,0}},
/* 54999 */ {(13<<2)|2,{80,52,0}},
/* 55000 */ {(13<<2)|2,{80,53,0}},
/* 55001 */ {(13<<2)|2,{80,53,0}},
/* 55002 */ {(13<<2)|2,{80,53,0}},
/* 55003 */ {(13<<2)|2,{80,53,0}},
/* 55004 */ {(13<<2)|2,{80,53,0}},
/* 55005 */ {(13<<2)|2,{80,53,0}},
/* 55006 */ {(13<<2)|2,{80,53,0}},
/* 55007 */ {(13<<2)|2,{80,53,0}},
/* 55008 */ {(13<<2)|2,{80,54,0}},
/* 55009 */ {(13<<2)|2,{80,54,0}},
/* 55010 */ {(13<<2)|2,{80,54,0}},
/* 55011 */ {(13<<2)|2,{80,54,0}},
/* 55012 */ {(13<<2)|2,{80,54,0}},
/* 55013 */ {(13<<2)|2,{80,54,0}},
/* 55014 */ {(13<<2)|2,{80,54,0}},
/* 55015 */ {(13<<2)|2,{80,54,0}},
/* 55016 */ {(13<<2)|2,{80,55,0}},
/* 55017 */ {(13<<2)|2,{80,55,0}},
/* 55018 */ {(13<<2)|2,{80,55,0}},
/* 55019 */ {(13<<2)|2,{80,55,0}},
/* 55020 */ {(13<<2)|2,{80,55,0}},
/* 55021 */ {(13<<2)|2,{80,55,0}},
/* 55022 */ {(13<<2)|2,{80,55,0}},
/* 55023 */ {(13<<2)|2,{80,55,0}},
/* 55024 */ {(13<<2)|2,{80,56,0}},
/* 55025 */ {(13<<2)|2,{80,56,0}},
/* 55026 */ {(13<<2)|2,{80,56,0}},
/* 55027 */ {(13<<2)|2,{80,56,0}},
/* 55028 */ {(13<<2)|2,{80,56,0}},
/* 55029 */ {(13<<2)|2,{80,56,0}},
/* 55030 */ {(13<<2)|2,{80,56,0}},
/* 55031 */ {(13<<2)|2,{80,56,0}},
/* 55032 */ {(13<<2)|2,{80,57,0}},
/* 55033 */ {(13<<2)|2,{80,57,0}},
/* 55034 */ {(13<<2)|2,{80,57,0}},
/* 55035 */ {(13<<2)|2,{80,57,0}},
/* 55036 */ {(13<<2)|2,{80,57,0}},
/* 55037 */ {(13<<2)|2,{80,57,0}},
/* 55038 */ {(13<<2)|2,{80,57,0}},
/* 55039 */ {(13<<2)|2,{80,57,0}},
/* 55040 */ {(13<<2)|2,{80,61,0}},
/* 55041 */ {(13<<2)|2,{80,61,0}},
/* 55042 */ {(13<<2)|2,{80,61,0}},
/* 55043 */ {(13<<2)|2,{80,61,0}},
/* 55044 */ {(13<<2)|2,{80,61,0}},
/* 55045 */ {(13<<2)|2,{80,61,0}},
/* 55046 */ {(13<<2)|2,{80,61,0}},
/* 55047 */ {(13<<2)|2,{80,61,0}},
/* 55048 */ {(13<<2)|2,{80,65,0}},
/* 55049 */ {(13<<2)|2,{80,65,0}},
/* 55050 */ {(13<<2)|2,{80,65,0}},
/* 55051 */ {(13<<2)|2,{80,65,0}},
/* 55052 */ {(13<<2)|2,{80,65,0}},
/* 55053 */ {(13<<2)|2,{80,65,0}},
/* 55054 */ {(13<<2)|2,{80,65,0}},
/* 55055 */ {(13<<2)|2,{80,65,0}},
/* 55056 */ {(13<<2)|2,{80,95,0}},
/* 55057 */ {(13<<2)|2,{80,95,0}},
/* 55058 */ {(13<<2)|2,{80,95,0}},
/* 55059 */ {(13<<2)|2,{80,95,0}},
/* 55060 */ {(13<<2)|2,{80,95,0}},
/* 55061 */ {(13<<2)|2,{80,95,0}},
/* 55062 */ {(13<<2)|2,{80,95,0}},
/* 55063 */ {(13<<2)|2,{80,95,0}},
/* 55064 */ {(13<<2)|2,{80,98,0}},
/* 55065 */ {(13<<2)|2,{80,98,0}},
/* 55066 */ {(13<<2)|2,{80,98,0}},
/* 55067 */ {(13<<2)|2,{80,98,0}},
/* 55068 */ {(13<<2)|2,{80,98,0}},
/* 55069 */ {(13<<2)|2,{80,98,0}},
/* 55070 */ {(13<<2)|2,{80,98,0}},
/* 55071 */ {(13<<2)|2,{80,98,0}},
/* 55072 */ {(13<<2)|2,{80,100,0}},
/* 55073 */ {(13<<2)|2,{80,100,0}},
/* 55074 */ {(13<<2)|2,{80,100,0}},
/* 55075 */ {(13<<2)|2,{80,100,0}},
/* 55076 */ {(13<<2)|2,{80,100,0}},
/* 55077 */ {(13<<2)|2,{80,100,0}},
/* 55078 */ {(13<<2)|2,{80,100,0}},
/* 55079 */ {(13<<2)|2,{80,100,0}},
/* 55080 */ {(13<<2)|2,{80,102,0}},
/* 55081 */ {(13<<2)|2,{80,102,0}},
/* 55082 */ {(13<<2)|2,{80,102,0}},
/* 55083 */ {(13<<2)|2,{80,102,0}},
/* 55084 */ {(13<<2)|2,{80,102,0}},
/* 55085 */ {(13<<2)|2,{80,102,0}},
/* 55086 */ {(13<<2)|2,{80,102,0}},
/* 55087 */ {(13<<2)|2,{80,102,0}},
/* 55088 */ {(13<<2)|2,{80,103,0}},
/* 55089 */ {(13<<2)|2,{80,103,0}},
/* 55090 */ {(13<<2)|2,{80,103,0}},
/* 55091 */ {(13<<2)|2,{80,103,0}},
/* 55092 */ {(13<<2)|2,{80,103,0}},
/* 55093 */ {(13<<2)|2,{80,103,0}},
/* 55094 */ {(13<<2)|2,{80,103,0}},
/* 55095 */ {(13<<2)|2,{80,103,0}},
/* 55096 */ {(13<<2)|2,{80,104,0}},
/* 55097 */ {(13<<2)|2,{80,104,0}},
/* 55098 */ {(13<<2)|2,{80,104,0}},
/* 55099 */ {(13<<2)|2,{80,104,0}},
/* 55100 */ {(13<<2)|2,{80,104,0}},
/* 55101 */ {(13<<2)|2,{80,104,0}},
/* 55102 */ {(13<<2)|2,{80,104,0}},
/* 55103 */ {(13<<2)|2,{80,104,0}},
/* 55104 */ {(13<<2)|2,{80,108,0}},
/* 55105 */ {(13<<2)|2,{80,108,0}},
/* 55106 */ {(13<<2)|2,{80,108,0}},
/* 55107 */ {(13<<2)|2,{80,108,0}},
/* 55108 */ {(13<<2)|2,{80,108,0}},
/* 55109 */ {(13<<2)|2,{80,108,0}},
/* 55110 */ {(13<<2)|2,{80,108,0}},
/* 55111 */ {(13<<2)|2,{80,108,0}},
/* 55112 */ {(13<<2)|2,{80,109,0}},
/* 55113 */ {(13<<2)|2,{80,109,0}},
/* 55114 */ {(13<<2)|2,{80,109,0}},
/* 55115 */ {(13<<2)|2,{80,109,0}},
/* 55116 */ {(13<<2)|2,{80,109,0}},
/* 55117 */ {(13<<2)|2,{80,109,0}},
/* 55118 */ {(13<<2)|2,{80,109,0}},
/* 55119 */ {(13<<2)|2,{80,109,0}},
/* 55120 */ {(13<<2)|2,{80,110,0}},
/* 55121 */ {(13<<2)|2,{80,110,0}},
/* 55122 */ {(13<<2)|2,{80,110,0}},
/* 55123 */ {(13<<2)|2,{80,110,0}},
/* 55124 */ {(13<<2)|2,{80,110,0}},
/* 55125 */ {(13<<2)|2,{80,110,0}},
/* 55126 */ {(13<<2)|2,{80,110,0}},
/* 55127 */ {(13<<2)|2,{80,110,0}},
/* 55128 */ {(13<<2)|2,{80,112,0}},
/* 55129 */ {(13<<2)|2,{80,112,0}},
/* 55130 */ {(13<<2)|2,{80,112,0}},
/* 55131 */ {(13<<2)|2,{80,112,0}},
/* 55132 */ {(13<<2)|2,{80,112,0}},
/* 55133 */ {(13<<2)|2,{80,112,0}},
/* 55134 */ {(13<<2)|2,{80,112,0}},
/* 55135 */ {(13<<2)|2,{80,112,0}},
/* 55136 */ {(13<<2)|2,{80,114,0}},
/* 55137 */ {(13<<2)|2,{80,114,0}},
/* 55138 */ {(13<<2)|2,{80,114,0}},
/* 55139 */ {(13<<2)|2,{80,114,0}},
/* 55140 */ {(13<<2)|2,{80,114,0}},
/* 55141 */ {(13<<2)|2,{80,114,0}},
/* 55142 */ {(13<<2)|2,{80,114,0}},
/* 55143 */ {(13<<2)|2,{80,114,0}},
/* 55144 */ {(13<<2)|2,{80,117,0}},
/* 55145 */ {(13<<2)|2,{80,117,0}},
/* 55146 */ {(13<<2)|2,{80,117,0}},
/* 55147 */ {(13<<2)|2,{80,117,0}},
/* 55148 */ {(13<<2)|2,{80,117,0}},
/* 55149 */ {(13<<2)|2,{80,117,0}},
/* 55150 */ {(13<<2)|2,{80,117,0}},
/* 55151 */ {(13<<2)|2,{80,117,0}},
/* 55152 */ {(14<<2)|2,{80,58,0}},
/* 55153 */ {(14<<2)|2,{80,58,0}},
/* 55154 */ {(14<<2)|2,{80,58,0}},
/* 55155 */ {(14<<2)|2,{80,58,0}},
/* 55156 */ {(14<<2)|2,{80,66,0}},
/* 55157 */ {(14<<2)|2,{80,66,0}},
/* 55158 */ {(14<<2)|2,{80,66,0}},
/* 55159 */ {(14<<2)|2,{80,66,0}},
/* 55160 */ {(14<<2)|2,{80,67,0}},
/* 55161 */ {(14<<2)|2,{80,67,0}},
/* 55162 */ {(14<<2)|2,{80,67,0}},
/* 55163 */ {(14<<2)|2,{80,67,0}},
/* 55164 */ {(14<<2)|2,{80,68,0}},
/* 55165 */ {(14<<2)|2,{80,68,0}},
/* 55166 */ {(14<<2)|2,{80,68,0}},
/* 55167 */ {(14<<2)|2,{80,68,0}},
/* 55168 */ {(14<<2)|2,{80,69,0}},
/* 55169 */ {(14<<2)|2,{80,69,0}},
/* 55170 */ {(14<<2)|2,{80,69,0}},
/* 55171 */ {(14<<2)|2,{80,69,0}},
/* 55172 */ {(14<<2)|2,{80,70,0}},
/* 55173 */ {(14<<2)|2,{80,70,0}},
/* 55174 */ {(14<<2)|2,{80,70,0}},
/* 55175 */ {(14<<2)|2,{80,70,0}},
/* 55176 */ {(14<<2)|2,{80,71,0}},
/* 55177 */ {(14<<2)|2,{80,71,0}},
/* 55178 */ {(14<<2)|2,{80,71,0}},
/* 55179 */ {(14<<2)|2,{80,71,0}},
/* 55180 */ {(14<<2)|2,{80,72,0}},
/* 55181 */ {(14<<2)|2,{80,72,0}},
/* 55182 */ {(14<<2)|2,{80,72,0}},
/* 55183 */ {(14<<2)|2,{80,72,0}},
/* 55184 */ {(14<<2)|2,{80,73,0}},
/* 55185 */ {(14<<2)|2,{80,73,0}},
/* 55186 */ {(14<<2)|2,{80,73,0}},
/* 55187 */ {(14<<2)|2,{80,73,0}},
/* 55188 */ {(14<<2)|2,{80,74,0}},
/* 55189 */ {(14<<2)|2,{80,74,0}},
/* 55190 */ {(14<<2)|2,{80,74,0}},
/* 55191 */ {(14<<2)|2,{80,74,0}},
/* 55192 */ {(14<<2)|2,{80,75,0}},
/* 55193 */ {(14<<2)|2,{80,75,0}},
/* 55194 */ {(14<<2)|2,{80,75,0}},
/* 55195 */ {(14<<2)|2,{80,75,0}},
/* 55196 */ {(14<<2)|2,{80,76,0}},
/* 55197 */ {(14<<2)|2,{80,76,0}},
/* 55198 */ {(14<<2)|2,{80,76,0}},
/* 55199 */ {(14<<2)|2,{80,76,0}},
/* 55200 */ {(14<<2)|2,{80,77,0}},
/* 55201 */ {(14<<2)|2,{80,77,0}},
/* 55202 */ {(14<<2)|2,{80,77,0}},
/* 55203 */ {(14<<2)|2,{80,77,0}},
/* 55204 */ {(14<<2)|2,{80,78,0}},
/* 55205 */ {(14<<2)|2,{80,78,0}},
/* 55206 */ {(14<<2)|2,{80,78,0}},
/* 55207 */ {(14<<2)|2,{80,78,0}},
/* 55208 */ {(14<<2)|2,{80,79,0}},
/* 55209 */ {(14<<2)|2,{80,79,0}},
/* 55210 */ {(14<<2)|2,{80,79,0}},
/* 55211 */ {(14<<2)|2,{80,79,0}},
/* 55212 */ {(14<<2)|2,{80,80,0}},
/* 55213 */ {(14<<2)|2,{80,80,0}},
/* 55214 */ {(14<<2)|2,{80,80,0}},
/* 55215 */ {(14<<2)|2,{80,80,0}},
/* 55216 */ {(14<<2)|2,{80,81,0}},
/* 55217 */ {(14<<2)|2,{80,81,0}},
/* 55218 */ {(14<<2)|2,{80,81,0}},
/* 55219 */ {(14<<2)|2,{80,81,0}},
/* 55220 */ {(14<<2)|2,{80,82,0}},
/* 55221 */ {(14<<2)|2,{80,82,0}},
/* 55222 */ {(14<<2)|2,{80,82,0}},
/* 55223 */ {(14<<2)|2,{80,82,0}},
/* 55224 */ {(14<<2)|2,{80,83,0}},
/* 55225 */ {(14<<2)|2,{80,83,0}},
/* 55226 */ {(14<<2)|2,{80,83,0}},
/* 55227 */ {(14<<2)|2,{80,83,0}},
/* 55228 */ {(14<<2)|2,{80,84,0}},
/* 55229 */ {(14<<2)|2,{80,84,0}},
/* 55230 */ {(14<<2)|2,{80,84,0}},
/* 55231 */ {(14<<2)|2,{80,84,0}},
/* 55232 */ {(14<<2)|2,{80,85,0}},
/* 55233 */ {(14<<2)|2,{80,85,0}},
/* 55234 */ {(14<<2)|2,{80,85,0}},
/* 55235 */ {(14<<2)|2,{80,85,0}},
/* 55236 */ {(14<<2)|2,{80,86,0}},
/* 55237 */ {(14<<2)|2,{80,86,0}},
/* 55238 */ {(14<<2)|2,{80,86,0}},
/* 55239 */ {(14<<2)|2,{80,86,0}},
/* 55240 */ {(14<<2)|2,{80,87,0}},
/* 55241 */ {(14<<2)|2,{80,87,0}},
/* 55242 */ {(14<<2)|2,{80,87,0}},
/* 55243 */ {(14<<2)|2,{80,87,0}},
/* 55244 */ {(14<<2)|2,{80,89,0}},
/* 55245 */ {(14<<2)|2,{80,89,0}},
/* 55246 */ {(14<<2)|2,{80,89,0}},
/* 55247 */ {(14<<2)|2,{80,89,0}},
/* 55248 */ {(14<<2)|2,{80,106,0}},
/* 55249 */ {(14<<2)|2,{80,106,0}},
/* 55250 */ {(14<<2)|2,{80,106,0}},
/* 55251 */ {(14<<2)|2,{80,106,0}},
/* 55252 */ {(14<<2)|2,{80,107,0}},
/* 55253 */ {(14<<2)|2,{80,107,0}},
/* 55254 */ {(14<<2)|2,{80,107,0}},
/* 55255 */ {(14<<2)|2,{80,107,0}},
/* 55256 */ {(14<<2)|2,{80,113,0}},
/* 55257 */ {(14<<2)|2,{80,113,0}},
/* 55258 */ {(14<<2)|2,{80,113,0}},
/* 55259 */ {(14<<2)|2,{80,113,0}},
/* 55260 */ {(14<<2)|2,{80,118,0}},
/* 55261 */ {(14<<2)|2,{80,118,0}},
/* 55262 */ {(14<<2)|2,{80,118,0}},
/* 55263 */ {(14<<2)|2,{80,118,0}},
/* 55264 */ {(14<<2)|2,{80,119,0}},
/* 55265 */ {(14<<2)|2,{80,119,0}},
/* 55266 */ {(14<<2)|2,{80,119,0}},
/* 55267 */ {(14<<2)|2,{80,119,0}},
/* 55268 */ {(14<<2)|2,{80,120,0}},
/* 55269 */ {(14<<2)|2,{80,120,0}},
/* 55270 */ {(14<<2)|2,{80,120,0}},
/* 55271 */ {(14<<2)|2,{80,120,0}},
/* 55272 */ {(14<<2)|2,{80,121,0}},
/* 55273 */ {(14<<2)|2,{80,121,0}},
/* 55274 */ {(14<<2)|2,{80,121,0}},
/* 55275 */ {(14<<2)|2,{80,121,0}},
/* 55276 */ {(14<<2)|2,{80,122,0}},
/* 55277 */ {(14<<2)|2,{80,122,0}},
/* 55278 */ {(14<<2)|2,{80,122,0}},
/* 55279 */ {(14<<2)|2,{80,122,0}},
/* 55280 */ {(15<<2)|2,{80,38,0}},
/* 55281 */ {(15<<2)|2,{80,38,0}},
/* 55282 */ {(15<<2)|2,{80,42,0}},
/* 55283 */ {(15<<2)|2,{80,42,0}},
/* 55284 */ {(15<<2)|2,{80,44,0}},
/* 55285 */ {(15<<2)|2,{80,44,0}},
/* 55286 */ {(15<<2)|2,{80,59,0}},
/* 55287 */ {(15<<2)|2,{80,59,0}},
/* 55288 */ {(15<<2)|2,{80,88,0}},
/* 55289 */ {(15<<2)|2,{80,88,0}},
/* 55290 */ {(15<<2)|2,{80,90,0}},
/* 55291 */ {(15<<2)|2,{80,90,0}},
/* 55292 */ {(7<<2)|1,{80,0,0}},
/* 55293 */ {(7<<2)|1,{80,0,0}},
/* 55294 */ {(7<<2)|1,{80,0,0}},
/* 55295 */ {(7<<2)|1,{80,0,0}},
/* 55296 */ {(12<<2)|2,{81,48,0}},
/* 55297 */ {(12<<2)|2,{81,48,0}},
/* 55298 */ {(12<<2)|2,{81,48,0}},
/* 55299 */ {(12<<2)|2,{81,48,0}},
/* 55300 */ {(12<<2)|2,{81,48,0}},
/* 55301 */ {(12<<2)|2,{81,48,0}},
/* 55302 */ {(12<<2)|2,{81,48,0}},
/* 55303 */ {(12<<2)|2,{81,48,0}},
/* 55304 */ {(12<<2)|2,{81,48,0}},
/* 55305 */ {(12<<2)|2,{81,48,0}},
/* 55306 */ {(12<<2)|2,{81,48,0}},
/* 55307 */ {(12<<2)|2,{81,48,0}},
/* 55308 */ {(12<<2)|2,{81,48,0}},
/* 55309 */ {(12<<2)|2,{81,48,0}},
/* 55310 */ {(12<<2)|2,{81,48,0}},
/* 55311 */ {(12<<2)|2,{81,48,0}},
/* 55312 */ {(12<<2)|2,{81,49,0}},
/* 55313 */ {(12<<2)|2,{81,49,0}},
/* 55314 */ {(12<<2)|2,{81,49,0}},
/* 55315 */ {(12<<2)|2,{81,49,0}},
/* 55316 */ {(12<<2)|2,{81,49,0}},
/* 55317 */ {(12<<2)|2,{81,49,0}},
/* 55318 */ {(12<<2)|2,{81,49,0}},
/* 55319 */ {(12<<2)|2,{81,49,0}},
/* 55320 */ {(12<<2)|2,{81,49,0}},
/* 55321 */ {(12<<2)|2,{81,49,0}},
/* 55322 */ {(12<<2)|2,{81,49,0}},
/* 55323 */ {(12<<2)|2,{81,49,0}},
/* 55324 */ {(12<<2)|2,{81,49,0}},
/* 55325 */ {(12<<2)|2,{81,49,0}},
/* 55326 */ {(12<<2)|2,{81,49,0}},
/* 55327 */ {(12<<2)|2,{81,49,0}},
/* 55328 */ {(12<<2)|2,{81,50,0}},
/* 55329 */ {(12<<2)|2,{81,50,0}},
/* 55330 */ {(12<<2)|2,{81,50,0}},
/* 55331 */ {(12<<2)|2,{81,50,0}},
/* 55332 */ {(12<<2)|2,{81,50,0}},
/* 55333 */ {(12<<2)|2,{81,50,0}},
/* 55334 */ {(12<<2)|2,{81,50,0}},
/* 55335 */ {(12<<2)|2,{81,50,0}},
/* 55336 */ {(12<<2)|2,{81,50,0}},
/* 55337 */ {(12<<2)|2,{81,50,0}},
/* 55338 */ {(12<<2)|2,{81,50,0}},
/* 55339 */ {(12<<2)|2,{81,50,0}},
/* 55340 */ {(12<<2)|2,{81,50,0}},
/* 55341 */ {(12<<2)|2,{81,50,0}},
/* 55342 */ {(12<<2)|2,{81,50,0}},
/* 55343 */ {(12<<2)|2,{81,50,0}},
/* 55344 */ {(12<<2)|2,{81,97,0}},
/* 55345 */ {(12<<2)|2,{81,97,0}},
/* 55346 */ {(12<<2)|2,{81,97,0}},
/* 55347 */ {(12<<2)|2,{81,97,0}},
/* 55348 */ {(12<<2)|2,{81,97,0}},
/* 55349 */ {(12<<2)|2,{81,97,0}},
/* 55350 */ {(12<<2)|2,{81,97,0}},
/* 55351 */ {(12<<2)|2,{81,97,0}},
/* 55352 */ {(12<<2)|2,{81,97,0}},
/* 55353 */ {(12<<2)|2,{81,97,0}},
/* 55354 */ {(12<<2)|2,{81,97,0}},
/* 55355 */ {(12<<2)|2,{81,97,0}},
/* 55356 */ {(12<<2)|2,{81,97,0}},
/* 55357 */ {(12<<2)|2,{81,97,0}},
/* 55358 */ {(12<<2)|2,{81,97,0}},
/* 55359 */ {(12<<2)|2,{81,97,0}},
/* 55360 */ {(12<<2)|2,{81,99,0}},
/* 55361 */ {(12<<2)|2,{81,99,0}},
/* 55362 */ {(12<<2)|2,{81,99,0}},
/* 55363 */ {(12<<2)|2,{81,99,0}},
/* 55364 */ {(12<<2)|2,{81,99,0}},
/* 55365 */ {(12<<2)|2,{81,99,0}},
/* 55366 */ {(12<<2)|2,{81,99,0}},
/* 55367 */ {(12<<2)|2,{81,99,0}},
/* 55368 */ {(12<<2)|2,{81,99,0}},
/* 55369 */ {(12<<2)|2,{81,99,0}},
/* 55370 */ {(12<<2)|2,{81,99,0}},
/* 55371 */ {(12<<2)|2,{81,99,0}},
/* 55372 */ {(12<<2)|2,{81,99,0}},
/* 55373 */ {(12<<2)|2,{81,99,0}},
/* 55374 */ {(12<<2)|2,{81,99,0}},
/* 55375 */ {(12<<2)|2,{81,99,0}},
/* 55376 */ {(12<<2)|2,{81,101,0}},
/* 55377 */ {(12<<2)|2,{81,101,0}},
/* 55378 */ {(12<<2)|2,{81,101,0}},
/* 55379 */ {(12<<2)|2,{81,101,0}},
/* 55380 */ {(12<<2)|2,{81,101,0}},
/* 55381 */ {(12<<2)|2,{81,101,0}},
/* 55382 */ {(12<<2)|2,{81,101,0}},
/* 55383 */ {(12<<2)|2,{81,101,0}},
/* 55384 */ {(12<<2)|2,{81,101,0}},
/* 55385 */ {(12<<2)|2,{81,101,0}},
/* 55386 */ {(12<<2)|2,{81,101,0}},
/* 55387 */ {(12<<2)|2,{81,101,0}},
/* 55388 */ {(12<<2)|2,{81,101,0}},
/* 55389 */ {(12<<2)|2,{81,101,0}},
/* 55390 */ {(12<<2)|2,{81,101,0}},
/* 55391 */ {(12<<2)|2,{81,101,0}},
/* 55392 */ {(12<<2)|2,{81,105,0}},
/* 55393 */ {(12<<2)|2,{81,105,0}},
/* 55394 */ {(12<<2)|2,{81,105,0}},
/* 55395 */ {(12<<2)|2,{81,105,0}},
/* 55396 */ {(12<<2)|2,{81,105,0}},
/* 55397 */ {(12<<2)|2,{81,105,0}},
/* 55398 */ {(12<<2)|2,{81,105,0}},
/* 55399 */ {(12<<2)|2,{81,105,0}},
/* 55400 */ {(12<<2)|2,{81,105,0}},
/* 55401 */ {(12<<2)|2,{81,105,0}},
/* 55402 */ {(12<<2)|2,{81,105,0}},
/* 55403 */ {(12<<2)|2,{81,105,0}},
/* 55404 */ {(12<<2)|2,{81,105,0}},
/* 55405 */ {(12<<2)|2,{81,105,0}},
/* 55406 */ {(12<<2)|2,{81,105,0}},
/* 55407 */ {(12<<2)|2,{81,105,0}},
/* 55408 */ {(12<<2)|2,{81,111,0}},
/* 55409 */ {(12<<2)|2,{81,111,0}},
/* 55410 */ {(12<<2)|2,{81,111,0}},
/* 55411 */ {(12<<2)|2,{81,111,0}},
/* 55412 */ {(12<<2)|2,{81,111,0}},
/* 55413 */ {(12<<2)|2,{81,111,0}},
/* 55414 */ {(12<<2)|2,{81,111,0}},
/* 55415 */ {(12<<2)|2,{81,111,0}},
/* 55416 */ {(12<<2)|2,{81,111,0}},
/* 55417 */ {(12<<2)|2,{81,111,0}},
/* 55418 */ {(12<<2)|2,{81,111,0}},
/* 55419 */ {(12<<2)|2,{81,111,0}},
/* 55420 */ {(12<<2)|2,{81,111,0}},
/* 55421 */ {(12<<2)|2,{81,111,0}},
/* 55422 */ {(12<<2)|2,{81,111,0}},
/* 55423 */ {(12<<2)|2,{81,111,0}},
/* 55424 */ {(12<<2)|2,{81,115,0}},
/* 55425 */ {(12<<2)|2,{81,115,0}},
/* 55426 */ {(12<<2)|2,{81,115,0}},
/* 55427 */ {(12<<2)|2,{81,115,0}},
/* 55428 */ {(12<<2)|2,{81,115,0}},
/* 55429 */ {(12<<2)|2,{81,115,0}},
/* 55430 */ {(12<<2)|2,{81,115,0}},
/* 55431 */ {(12<<2)|2,{81,115,0}},
/* 55432 */ {(12<<2)|2,{81,115,0}},
/* 55433 */ {(12<<2)|2,{81,115,0}},
/* 55434 */ {(12<<2)|2,{81,115,0}},
/* 55435 */ {(12<<2)|2,{81,115,0}},
/* 55436 */ {(12<<2)|2,{81,115,0}},
/* 55437 */ {(12<<2)|2,{81,115,0}},
/* 55438 */ {(12<<2)|2,{81,115,0}},
/* 55439 */ {(12<<2)|2,{81,115,0}},
/* 55440 */ {(12<<2)|2,{81,116,0}},
/* 55441 */ {(12<<2)|2,{81,116,0}},
/* 55442 */ {(12<<2)|2,{81,116,0}},
/* 55443 */ {(12<<2)|2,{81,116,0}},
/* 55444 */ {(12<<2)|2,{81,116,0}},
/* 55445 */ {(12<<2)|2,{81,116,0}},
/* 55446 */ {(12<<2)|2,{81,116,0}},
/* 55447 */ {(12<<2)|2,{81,116,0}},
/* 55448 */ {(12<<2)|2,{81,116,0}},
/* 55449 */ {(12<<2)|2,{81,116,0}},
/* 55450 */ {(12<<2)|2,{81,116,0}},
/* 55451 */ {(12<<2)|2,{81,116,0}},
/* 55452 */ {(12<<2)|2,{81,116,0}},
/* 55453 */ {(12<<2)|2,{81,116,0}},
/* 55454 */ {(12<<2)|2,{81,116,0}},
/* 55455 */ {(12<<2)|2,{81,116,0}},
/* 55456 */ {(13<<2)|2,{81,32,0}},
/* 55457 */ {(13<<2)|2,{81,32,0}},
/* 55458 */ {(13<<2)|2,{81,32,0}},
/* 55459 */ {(13<<2)|2,{81,32,0}},
/* 55460 */ {(13<<2)|2,{81,32,0}},
/* 55461 */ {(13<<2)|2,{81,32,0}},
/* 55462 */ {(13<<2)|2,{81,32,0}},
/* 55463 */ {(13<<2)|2,{81,32,0}},
/* 55464 */ {(13<<2)|2,{81,37,0}},
/* 55465 */ {(13<<2)|2,{81,37,0}},
/* 55466 */ {(13<<2)|2,{81,37,0}},
/* 55467 */ {(13<<2)|2,{81,37,0}},
/* 55468 */ {(13<<2)|2,{81,37,0}},
/* 55469 */ {(13<<2)|2,{81,37,0}},
/* 55470 */ {(13<<2)|2,{81,37,0}},
/* 55471 */ {(13<<2)|2,{81,37,0}},
/* 55472 */ {(13<<2)|2,{81,45,0}},
/* 55473 */ {(13<<2)|2,{81,45,0}},
/* 55474 */ {(13<<2)|2,{81,45,0}},
/* 55475 */ {(13<<2)|2,{81,45,0}},
/* 55476 */ {(13<<2)|2,{81,45,0}},
/* 55477 */ {(13<<2)|2,{81,45,0}},
/* 55478 */ {(13<<2)|2,{81,45,0}},
/* 55479 */ {(13<<2)|2,{81,45,0}},
/* 55480 */ {(13<<2)|2,{81,46,0}},
/* 55481 */ {(13<<2)|2,{81,46,0}},
/* 55482 */ {(13<<2)|2,{81,46,0}},
/* 55483 */ {(13<<2)|2,{81,46,0}},
/* 55484 */ {(13<<2)|2,{81,46,0}},
/* 55485 */ {(13<<2)|2,{81,46,0}},
/* 55486 */ {(13<<2)|2,{81,46,0}},
/* 55487 */ {(13<<2)|2,{81,46,0}},
/* 55488 */ {(13<<2)|2,{81,47,0}},
/* 55489 */ {(13<<2)|2,{81,47,0}},
/* 55490 */ {(13<<2)|2,{81,47,0}},
/* 55491 */ {(13<<2)|2,{81,47,0}},
/* 55492 */ {(13<<2)|2,{81,47,0}},
/* 55493 */ {(13<<2)|2,{81,47,0}},
/* 55494 */ {(13<<2)|2,{81,47,0}},
/* 55495 */ {(13<<2)|2,{81,47,0}},
/* 55496 */ {(13<<2)|2,{81,51,0}},
/* 55497 */ {(13<<2)|2,{81,51,0}},
/* 55498 */ {(13<<2)|2,{81,51,0}},
/* 55499 */ {(13<<2)|2,{81,51,0}},
/* 55500 */ {(13<<2)|2,{81,51,0}},
/* 55501 */ {(13<<2)|2,{81,51,0}},
/* 55502 */ {(13<<2)|2,{81,51,0}},
/* 55503 */ {(13<<2)|2,{81,51,0}},
/* 55504 */ {(13<<2)|2,{81,52,0}},
/* 55505 */ {(13<<2)|2,{81,52,0}},
/* 55506 */ {(13<<2)|2,{81,52,0}},
/* 55507 */ {(13<<2)|2,{81,52,0}},
/* 55508 */ {(13<<2)|2,{81,52,0}},
/* 55509 */ {(13<<2)|2,{81,52,0}},
/* 55510 */ {(13<<2)|2,{81,52,0}},
/* 55511 */ {(13<<2)|2,{81,52,0}},
/* 55512 */ {(13<<2)|2,{81,53,0}},
/* 55513 */ {(13<<2)|2,{81,53,0}},
/* 55514 */ {(13<<2)|2,{81,53,0}},
/* 55515 */ {(13<<2)|2,{81,53,0}},
/* 55516 */ {(13<<2)|2,{81,53,0}},
/* 55517 */ {(13<<2)|2,{81,53,0}},
/* 55518 */ {(13<<2)|2,{81,53,0}},
/* 55519 */ {(13<<2)|2,{81,53,0}},
/* 55520 */ {(13<<2)|2,{81,54,0}},
/* 55521 */ {(13<<2)|2,{81,54,0}},
/* 55522 */ {(13<<2)|2,{81,54,0}},
/* 55523 */ {(13<<2)|2,{81,54,0}},
/* 55524 */ {(13<<2)|2,{81,54,0}},
/* 55525 */ {(13<<2)|2,{81,54,0}},
/* 55526 */ {(13<<2)|2,{81,54,0}},
/* 55527 */ {(13<<2)|2,{81,54,0}},
/* 55528 */ {(13<<2)|2,{81,55,0}},
/* 55529 */ {(13<<2)|2,{81,55,0}},
/* 55530 */ {(13<<2)|2,{81,55,0}},
/* 55531 */ {(13<<2)|2,{81,55,0}},
/* 55532 */ {(13<<2)|2,{81,55,0}},
/* 55533 */ {(13<<2)|2,{81,55,0}},
/* 55534 */ {(13<<2)|2,{81,55,0}},
/* 55535 */ {(13<<2)|2,{81,55,0}},
/* 55536 */ {(13<<2)|2,{81,56,0}},
/* 55537 */ {(13<<2)|2,{81,56,0}},
/* 55538 */ {(13<<2)|2,{81,56,0}},
/* 55539 */ {(13<<2)|2,{81,56,0}},
/* 55540 */ {(13<<2)|2,{81,56,0}},
/* 55541 */ {(13<<2)|2,{81,56,0}},
/* 55542 */ {(13<<2)|2,{81,56,0}},
/* 55543 */ {(13<<2)|2,{81,56,0}},
/* 55544 */ {(13<<2)|2,{81,57,0}},
/* 55545 */ {(13<<2)|2,{81,57,0}},
/* 55546 */ {(13<<2)|2,{81,57,0}},
/* 55547 */ {(13<<2)|2,{81,57,0}},
/* 55548 */ {(13<<2)|2,{81,57,0}},
/* 55549 */ {(13<<2)|2,{81,57,0}},
/* 55550 */ {(13<<2)|2,{81,57,0}},
/* 55551 */ {(13<<2)|2,{81,57,0}},
/* 55552 */ {(13<<2)|2,{81,61,0}},
/* 55553 */ {(13<<2)|2,{81,61,0}},
/* 55554 */ {(13<<2)|2,{81,61,0}},
/* 55555 */ {(13<<2)|2,{81,61,0}},
/* 55556 */ {(13<<2)|2,{81,61,0}},
/* 55557 */ {(13<<2)|2,{81,61,0}},
/* 55558 */ {(13<<2)|2,{81,61,0}},
/* 55559 */ {(13<<2)|2,{81,61,0}},
/* 55560 */ {(13<<2)|2,{81,65,0}},
/* 55561 */ {(13<<2)|2,{81,65,0}},
/* 55562 */ {(13<<2)|2,{81,65,0}},
/* 55563 */ {(13<<2)|2,{81,65,0}},
/* 55564 */ {(13<<2)|2,{81,65,0}},
/* 55565 */ {(13<<2)|2,{81,65,0}},
/* 55566 */ {(13<<2)|2,{81,65,0}},
/* 55567 */ {(13<<2)|2,{81,65,0}},
/* 55568 */ {(13<<2)|2,{81,95,0}},
/* 55569 */ {(13<<2)|2,{81,95,0}},
/* 55570 */ {(13<<2)|2,{81,95,0}},
/* 55571 */ {(13<<2)|2,{81,95,0}},
/* 55572 */ {(13<<2)|2,{81,95,0}},
/* 55573 */ {(13<<2)|2,{81,95,0}},
/* 55574 */ {(13<<2)|2,{81,95,0}},
/* 55575 */ {(13<<2)|2,{81,95,0}},
/* 55576 */ {(13<<2)|2,{81,98,0}},
/* 55577 */ {(13<<2)|2,{81,98,0}},
/* 55578 */ {(13<<2)|2,{81,98,0}},
/* 55579 */ {(13<<2)|2,{81,98,0}},
/* 55580 */ {(13<<2)|2,{81,98,0}},
/* 55581 */ {(13<<2)|2,{81,98,0}},
/* 55582 */ {(13<<2)|2,{81,98,0}},
/* 55583 */ {(13<<2)|2,{81,98,0}},
/* 55584 */ {(13<<2)|2,{81,100,0}},
/* 55585 */ {(13<<2)|2,{81,100,0}},
/* 55586 */ {(13<<2)|2,{81,100,0}},
/* 55587 */ {(13<<2)|2,{81,100,0}},
/* 55588 */ {(13<<2)|2,{81,100,0}},
/* 55589 */ {(13<<2)|2,{81,100,0}},
/* 55590 */ {(13<<2)|2,{81,100,0}},
/* 55591 */ {(13<<2)|2,{81,100,0}},
/* 55592 */ {(13<<2)|2,{81,102,0}},
/* 55593 */ {(13<<2)|2,{81,102,0}},
/* 55594 */ {(13<<2)|2,{81,102,0}},
/* 55595 */ {(13<<2)|2,{81,102,0}},
/* 55596 */ {(13<<2)|2,{81,102,0}},
/* 55597 */ {(13<<2)|2,{81,102,0}},
/* 55598 */ {(13<<2)|2,{81,102,0}},
/* 55599 */ {(13<<2)|2,{81,102,0}},
/* 55600 */ {(13<<2)|2,{81,103,0}},
/* 55601 */ {(13<<2)|2,{81,103,0}},
/* 55602 */ {(13<<2)|2,{81,103,0}},
/* 55603 */ {(13<<2)|2,{81,103,0}},
/* 55604 */ {(13<<2)|2,{81,103,0}},
/* 55605 */ {(13<<2)|2,{81,103,0}},
/* 55606 */ {(13<<2)|2,{81,103,0}},
/* 55607 */ {(13<<2)|2,{81,103,0}},
/* 55608 */ {(13<<2)|2,{81,104,0}},
/* 55609 */ {(13<<2)|2,{81,104,0}},
/* 55610 */ {(13<<2)|2,{81,104,0}},
/* 55611 */ {(13<<2)|2,{81,104,0}},
/* 55612 */ {(13<<2)|2,{81,104,0}},
/* 55613 */ {(13<<2)|2,{81,104,0}},
/* 55614 */ {(13<<2)|2,{81,104,0}},
/* 55615 */ {(13<<2)|2,{81,104,0}},
/* 55616 */ {(13<<2)|2,{81,108,0}},
/* 55617 */ {(13<<2)|2,{81,108,0}},
/* 55618 */ {(13<<2)|2,{81,108,0}},
/* 55619 */ {(13<<2)|2,{81,108,0}},
/* 55620 */ {(13<<2)|2,{81,108,0}},
/* 55621 */ {(13<<2)|2,{81,108,0}},
/* 55622 */ {(13<<2)|2,{81,108,0}},
/* 55623 */ {(13<<2)|2,{81,108,0}},
/* 55624 */ {(13<<2)|2,{81,109,0}},
/* 55625 */ {(13<<2)|2,{81,109,0}},
/* 55626 */ {(13<<2)|2,{81,109,0}},
/* 55627 */ {(13<<2)|2,{81,109,0}},
/* 55628 */ {(13<<2)|2,{81,109,0}},
/* 55629 */ {(13<<2)|2,{81,109,0}},
/* 55630 */ {(13<<2)|2,{81,109,0}},
/* 55631 */ {(13<<2)|2,{81,109,0}},
/* 55632 */ {(13<<2)|2,{81,110,0}},
/* 55633 */ {(13<<2)|2,{81,110,0}},
/* 55634 */ {(13<<2)|2,{81,110,0}},
/* 55635 */ {(13<<2)|2,{81,110,0}},
/* 55636 */ {(13<<2)|2,{81,110,0}},
/* 55637 */ {(13<<2)|2,{81,110,0}},
/* 55638 */ {(13<<2)|2,{81,110,0}},
/* 55639 */ {(13<<2)|2,{81,110,0}},
/* 55640 */ {(13<<2)|2,{81,112,0}},
/* 55641 */ {(13<<2)|2,{81,112,0}},
/* 55642 */ {(13<<2)|2,{81,112,0}},
/* 55643 */ {(13<<2)|2,{81,112,0}},
/* 55644 */ {(13<<2)|2,{81,112,0}},
/* 55645 */ {(13<<2)|2,{81,112,0}},
/* 55646 */ {(13<<2)|2,{81,112,0}},
/* 55647 */ {(13<<2)|2,{81,112,0}},
/* 55648 */ {(13<<2)|2,{81,114,0}},
/* 55649 */ {(13<<2)|2,{81,114,0}},
/* 55650 */ {(13<<2)|2,{81,114,0}},
/* 55651 */ {(13<<2)|2,{81,114,0}},
/* 55652 */ {(13<<2)|2,{81,114,0}},
/* 55653 */ {(13<<2)|2,{81,114,0}},
/* 55654 */ {(13<<2)|2,{81,114,0}},
/* 55655 */ {(13<<2)|2,{81,114,0}},
/* 55656 */ {(13<<2)|2,{81,117,0}},
/* 55657 */ {(13<<2)|2,{81,117,0}},
/* 55658 */ {(13<<2)|2,{81,117,0}},
/* 55659 */ {(13<<2)|2,{81,117,0}},
/* 55660 */ {(13<<2)|2,{81,117,0}},
/* 55661 */ {(13<<2)|2,{81,117,0}},
/* 55662 */ {(13<<2)|2,{81,117,0}},
/* 55663 */ {(13<<2)|2,{81,117,0}},
/* 55664 */ {(14<<2)|2,{81,58,0}},
/* 55665 */ {(14<<2)|2,{81,58,0}},
/* 55666 */ {(14<<2)|2,{81,58,0}},
/* 55667 */ {(14<<2)|2,{81,58,0}},
/* 55668 */ {(14<<2)|2,{81,66,0}},
/* 55669 */ {(14<<2)|2,{81,66,0}},
/* 55670 */ {(14<<2)|2,{81,66,0}},
/* 55671 */ {(14<<2)|2,{81,66,0}},
/* 55672 */ {(14<<2)|2,{81,67,0}},
/* 55673 */ {(14<<2)|2,{81,67,0}},
/* 55674 */ {(14<<2)|2,{81,67,0}},
/* 55675 */ {(14<<2)|2,{81,67,0}},
/* 55676 */ {(14<<2)|2,{81,68,0}},
/* 55677 */ {(14<<2)|2,{81,68,0}},
/* 55678 */ {(14<<2)|2,{81,68,0}},
/* 55679 */ {(14<<2)|2,{81,68,0}},
/* 55680 */ {(14<<2)|2,{81,69,0}},
/* 55681 */ {(14<<2)|2,{81,69,0}},
/* 55682 */ {(14<<2)|2,{81,69,0}},
/* 55683 */ {(14<<2)|2,{81,69,0}},
/* 55684 */ {(14<<2)|2,{81,70,0}},
/* 55685 */ {(14<<2)|2,{81,70,0}},
/* 55686 */ {(14<<2)|2,{81,70,0}},
/* 55687 */ {(14<<2)|2,{81,70,0}},
/* 55688 */ {(14<<2)|2,{81,71,0}},
/* 55689 */ {(14<<2)|2,{81,71,0}},
/* 55690 */ {(14<<2)|2,{81,71,0}},
/* 55691 */ {(14<<2)|2,{81,71,0}},
/* 55692 */ {(14<<2)|2,{81,72,0}},
/* 55693 */ {(14<<2)|2,{81,72,0}},
/* 55694 */ {(14<<2)|2,{81,72,0}},
/* 55695 */ {(14<<2)|2,{81,72,0}},
/* 55696 */ {(14<<2)|2,{81,73,0}},
/* 55697 */ {(14<<2)|2,{81,73,0}},
/* 55698 */ {(14<<2)|2,{81,73,0}},
/* 55699 */ {(14<<2)|2,{81,73,0}},
/* 55700 */ {(14<<2)|2,{81,74,0}},
/* 55701 */ {(14<<2)|2,{81,74,0}},
/* 55702 */ {(14<<2)|2,{81,74,0}},
/* 55703 */ {(14<<2)|2,{81,74,0}},
/* 55704 */ {(14<<2)|2,{81,75,0}},
/* 55705 */ {(14<<2)|2,{81,75,0}},
/* 55706 */ {(14<<2)|2,{81,75,0}},
/* 55707 */ {(14<<2)|2,{81,75,0}},
/* 55708 */ {(14<<2)|2,{81,76,0}},
/* 55709 */ {(14<<2)|2,{81,76,0}},
/* 55710 */ {(14<<2)|2,{81,76,0}},
/* 55711 */ {(14<<2)|2,{81,76,0}},
/* 55712 */ {(14<<2)|2,{81,77,0}},
/* 55713 */ {(14<<2)|2,{81,77,0}},
/* 55714 */ {(14<<2)|2,{81,77,0}},
/* 55715 */ {(14<<2)|2,{81,77,0}},
/* 55716 */ {(14<<2)|2,{81,78,0}},
/* 55717 */ {(14<<2)|2,{81,78,0}},
/* 55718 */ {(14<<2)|2,{81,78,0}},
/* 55719 */ {(14<<2)|2,{81,78,0}},
/* 55720 */ {(14<<2)|2,{81,79,0}},
/* 55721 */ {(14<<2)|2,{81,79,0}},
/* 55722 */ {(14<<2)|2,{81,79,0}},
/* 55723 */ {(14<<2)|2,{81,79,0}},
/* 55724 */ {(14<<2)|2,{81,80,0}},
/* 55725 */ {(14<<2)|2,{81,80,0}},
/* 55726 */ {(14<<2)|2,{81,80,0}},
/* 55727 */ {(14<<2)|2,{81,80,0}},
/* 55728 */ {(14<<2)|2,{81,81,0}},
/* 55729 */ {(14<<2)|2,{81,81,0}},
/* 55730 */ {(14<<2)|2,{81,81,0}},
/* 55731 */ {(14<<2)|2,{81,81,0}},
/* 55732 */ {(14<<2)|2,{81,82,0}},
/* 55733 */ {(14<<2)|2,{81,82,0}},
/* 55734 */ {(14<<2)|2,{81,82,0}},
/* 55735 */ {(14<<2)|2,{81,82,0}},
/* 55736 */ {(14<<2)|2,{81,83,0}},
/* 55737 */ {(14<<2)|2,{81,83,0}},
/* 55738 */ {(14<<2)|2,{81,83,0}},
/* 55739 */ {(14<<2)|2,{81,83,0}},
/* 55740 */ {(14<<2)|2,{81,84,0}},
/* 55741 */ {(14<<2)|2,{81,84,0}},
/* 55742 */ {(14<<2)|2,{81,84,0}},
/* 55743 */ {(14<<2)|2,{81,84,0}},
/* 55744 */ {(14<<2)|2,{81,85,0}},
/* 55745 */ {(14<<2)|2,{81,85,0}},
/* 55746 */ {(14<<2)|2,{81,85,0}},
/* 55747 */ {(14<<2)|2,{81,85,0}},
/* 55748 */ {(14<<2)|2,{81,86,0}},
/* 55749 */ {(14<<2)|2,{81,86,0}},
/* 55750 */ {(14<<2)|2,{81,86,0}},
/* 55751 */ {(14<<2)|2,{81,86,0}},
/* 55752 */ {(14<<2)|2,{81,87,0}},
/* 55753 */ {(14<<2)|2,{81,87,0}},
/* 55754 */ {(14<<2)|2,{81,87,0}},
/* 55755 */ {(14<<2)|2,{81,87,0}},
/* 55756 */ {(14<<2)|2,{81,89,0}},
/* 55757 */ {(14<<2)|2,{81,89,0}},
/* 55758 */ {(14<<2)|2,{81,89,0}},
/* 55759 */ {(14<<2)|2,{81,89,0}},
/* 55760 */ {(14<<2)|2,{81,106,0}},
/* 55761 */ {(14<<2)|2,{81,106,0}},
/* 55762 */ {(14<<2)|2,{81,106,0}},
/* 55763 */ {(14<<2)|2,{81,106,0}},
/* 55764 */ {(14<<2)|2,{81,107,0}},
/* 55765 */ {(14<<2)|2,{81,107,0}},
/* 55766 */ {(14<<2)|2,{81,107,0}},
/* 55767 */ {(14<<2)|2,{81,107,0}},
/* 55768 */ {(14<<2)|2,{81,113,0}},
/* 55769 */ {(14<<2)|2,{81,113,0}},
/* 55770 */ {(14<<2)|2,{81,113,0}},
/* 55771 */ {(14<<2)|2,{81,113,0}},
/* 55772 */ {(14<<2)|2,{81,118,0}},
/* 55773 */ {(14<<2)|2,{81,118,0}},
/* 55774 */ {(14<<2)|2,{81,118,0}},
/* 55775 */ {(14<<2)|2,{81,118,0}},
/* 55776 */ {(14<<2)|2,{81,119,0}},
/* 55777 */ {(14<<2)|2,{81,119,0}},
/* 55778 */ {(14<<2)|2,{81,119,0}},
/* 55779 */ {(14<<2)|2,{81,119,0}},
/* 55780 */ {(14<<2)|2,{81,120,0}},
/* 55781 */ {(14<<2)|2,{81,120,0}},
/* 55782 */ {(14<<2)|2,{81,120,0}},
/* 55783 */ {(14<<2)|2,{81,120,0}},
/* 55784 */ {(14<<2)|2,{81,121,0}},
/* 55785 */ {(14<<2)|2,{81,121,0}},
/* 55786 */ {(14<<2)|2,{81,121,0}},
/* 55787 */ {(14<<2)|2,{81,121,0}},
/* 55788 */ {(14<<2)|2,{81,122,0}},
/* 55789 */ {(14<<2)|2,{81,122,0}},
/* 55790 */ {(14<<2)|2,{81,122,0}},
/* 55791 */ {(14<<2)|2,{81,122,0}},
/* 55792 */ {(15<<2)|2,{81,38,0}},
/* 55793 */ {(15<<2)|2,{81,38,0}},
/* 55794 */ {(15<<2)|2,{81,42,0}},
/* 55795 */ {(15<<2)|2,{81,42,0}},
/* 55796 */ {(15<<2)|2,{81,44,0}},
/* 55797 */ {(15<<2)|2,{81,44,0}},
/* 55798 */ {(15<<2)|2,{81,59,0}},
/* 55799 */ {(15<<2)|2,{81,59,0}},
/* 55800 */ {(15<<2)|2,{81,88,0}},
/* 55801 */ {(15<<2)|2,{81,88,0}},
/* 55802 */ {(15<<2)|2,{81,90,0}},
/* 55803 */ {(15<<2)|2,{81,90,0}},
/* 55804 */ {(7<<2)|1,{81,0,0}},
/* 55805 */ {(7<<2)|1,{81,0,0}},
/* 55806 */ {(7<<2)|1,{81,0,0}},
/* 55807 */ {(7<<2)|1,{81,0,0}},
/* 55808 */ {(12<<2)|2,{82,48,0}},
/* 55809 */ {(12<<2)|2,{82,48,0}},
/* 55810 */ {(12<<2)|2,{82,48,0}},
/* 55811 */ {(12<<2)|2,{82,48,0}},
/* 55812 */ {(12<<2)|2,{82,48,0}},
/* 55813 */ {(12<<2)|2,{82,48,0}},
/* 55814 */ {(12<<2)|2,{82,48,0}},
/* 55815 */ {(12<<2)|2,{82,48,0}},
/* 55816 */ {(12<<2)|2,{82,48,0}},
/* 55817 */ {(12<<2)|2,{82,48,0}},
/* 55818 */ {(12<<2)|2,{82,48,0}},
/* 55819 */ {(12<<2)|2,{82,48,0}},
/* 55820 */ {(12<<2)|2,{82,48,0}},
/* 55821 */ {(12<<2)|2,{82,48,0}},
/* 55822 */ {(12<<2)|2,{82,48,0}},
/* 55823 */ {(12<<2)|2,{82,48,0}},
/* 55824 */ {(12<<2)|2,{82,49,0}},
/* 55825 */ {(12<<2)|2,{82,49,0}},
/* 55826 */ {(12<<2)|2,{82,49,0}},
/* 55827 */ {(12<<2)|2,{82,49,0}},
/* 55828 */ {(12<<2)|2,{82,49,0}},
/* 55829 */ {(12<<2)|2,{82,49,0}},
/* 55830 */ {(12<<2)|2,{82,49,0}},
/* 55831 */ {(12<<2)|2,{82,49,0}},
/* 55832 */ {(12<<2)|2,{82,49,0}},
/* 55833 */ {(12<<2)|2,{82,49,0}},
/* 55834 */ {(12<<2)|2,{82,49,0}},
/* 55835 */ {(12<<2)|2,{82,49,0}},
/* 55836 */ {(12<<2)|2,{82,49,0}},
/* 55837 */ {(12<<2)|2,{82,49,0}},
/* 55838 */ {(12<<2)|2,{82,49,0}},
/* 55839 */ {(12<<2)|2,{82,49,0}},
/* 55840 */ {(12<<2)|2,{82,50,0}},
/* 55841 */ {(12<<2)|2,{82,50,0}},
/* 55842 */ {(12<<2)|2,{82,50,0}},
/* 55843 */ {(12<<2)|2,{82,50,0}},
/* 55844 */ {(12<<2)|2,{82,50,0}},
/* 55845 */ {(12<<2)|2,{82,50,0}},
/* 55846 */ {(12<<2)|2,{82,50,0}},
/* 55847 */ {(12<<2)|2,{82,50,0}},
/* 55848 */ {(12<<2)|2,{82,50,0}},
/* 55849 */ {(12<<2)|2,{82,50,0}},
/* 55850 */ {(12<<2)|2,{82,50,0}},
/* 55851 */ {(12<<2)|2,{82,50,0}},
/* 55852 */ {(12<<2)|2,{82,50,0}},
/* 55853 */ {(12<<2)|2,{82,50,0}},
/* 55854 */ {(12<<2)|2,{82,50,0}},
/* 55855 */ {(12<<2)|2,{82,50,0}},
/* 55856 */ {(12<<2)|2,{82,97,0}},
/* 55857 */ {(12<<2)|2,{82,97,0}},
/* 55858 */ {(12<<2)|2,{82,97,0}},
/* 55859 */ {(12<<2)|2,{82,97,0}},
/* 55860 */ {(12<<2)|2,{82,97,0}},
/* 55861 */ {(12<<2)|2,{82,97,0}},
/* 55862 */ {(12<<2)|2,{82,97,0}},
/* 55863 */ {(12<<2)|2,{82,97,0}},
/* 55864 */ {(12<<2)|2,{82,97,0}},
/* 55865 */ {(12<<2)|2,{82,97,0}},
/* 55866 */ {(12<<2)|2,{82,97,0}},
/* 55867 */ {(12<<2)|2,{82,97,0}},
/* 55868 */ {(12<<2)|2,{82,97,0}},
/* 55869 */ {(12<<2)|2,{82,97,0}},
/* 55870 */ {(12<<2)|2,{82,97,0}},
/* 55871 */ {(12<<2)|2,{82,97,0}},
/* 55872 */ {(12<<2)|2,{82,99,0}},
/* 55873 */ {(12<<2)|2,{82,99,0}},
/* 55874 */ {(12<<2)|2,{82,99,0}},
/* 55875 */ {(12<<2)|2,{82,99,0}},
/* 55876 */ {(12<<2)|2,{82,99,0}},
/* 55877 */ {(12<<2)|2,{82,99,0}},
/* 55878 */ {(12<<2)|2,{82,99,0}},
/* 55879 */ {(12<<2)|2,{82,99,0}},
/* 55880 */ {(12<<2)|2,{82,99,0}},
/* 55881 */ {(12<<2)|2,{82,99,0}},
/* 55882 */ {(12<<2)|2,{82,99,0}},
/* 55883 */ {(12<<2)|2,{82,99,0}},
/* 55884 */ {(12<<2)|2,{82,99,0}},
/* 55885 */ {(12<<2)|2,{82,99,0}},
/* 55886 */ {(12<<2)|2,{82,99,0}},
/* 55887 */ {(12<<2)|2,{82,99,0}},
/* 55888 */ {(12<<2)|2,{82,101,0}},
/* 55889 */ {(12<<2)|2,{82,101,0}},
/* 55890 */ {(12<<2)|2,{82,101,0}},
/* 55891 */ {(12<<2)|2,{82,101,0}},
/* 55892 */ {(12<<2)|2,{82,101,0}},
/* 55893 */ {(12<<2)|2,{82,101,0}},
/* 55894 */ {(12<<2)|2,{82,101,0}},
/* 55895 */ {(12<<2)|2,{82,101,0}},
/* 55896 */ {(12<<2)|2,{82,101,0}},
/* 55897 */ {(12<<2)|2,{82,101,0}},
/* 55898 */ {(12<<2)|2,{82,101,0}},
/* 55899 */ {(12<<2)|2,{82,101,0}},
/* 55900 */ {(12<<2)|2,{82,101,0}},
/* 55901 */ {(12<<2)|2,{82,101,0}},
/* 55902 */ {(12<<2)|2,{82,101,0}},
/* 55903 */ {(12<<2)|2,{82,101,0}},
/* 55904 */ {(12<<2)|2,{82,105,0}},
/* 55905 */ {(12<<2)|2,{82,105,0}},
/* 55906 */ {(12<<2)|2,{82,105,0}},
/* 55907 */ {(12<<2)|2,{82,105,0}},
/* 55908 */ {(12<<2)|2,{82,105,0}},
/* 55909 */ {(12<<2)|2,{82,105,0}},
/* 55910 */ {(12<<2)|2,{82,105,0}},
/* 55911 */ {(12<<2)|2,{82,105,0}},
/* 55912 */ {(12<<2)|2,{82,105,0}},
/* 55913 */ {(12<<2)|2,{82,105,0}},
/* 55914 */ {(12<<2)|2,{82,105,0}},
/* 55915 */ {(12<<2)|2,{82,105,0}},
/* 55916 */ {(12<<2)|2,{82,105,0}},
/* 55917 */ {(12<<2)|2,{82,105,0}},
/* 55918 */ {(12<<2)|2,{82,105,0}},
/* 55919 */ {(12<<2)|2,{82,105,0}},
/* 55920 */ {(12<<2)|2,{82,111,0}},
/* 55921 */ {(12<<2)|2,{82,111,0}},
/* 55922 */ {(12<<2)|2,{82,111,0}},
/* 55923 */ {(12<<2)|2,{82,111,0}},
/* 55924 */ {(12<<2)|2,{82,111,0}},
/* 55925 */ {(12<<2)|2,{82,111,0}},
/* 55926 */ {(12<<2)|2,{82,111,0}},
/* 55927 */ {(12<<2)|2,{82,111,0}},
/* 55928 */ {(12<<2)|2,{82,111,0}},
/* 55929 */ {(12<<2)|2,{82,111,0}},
/* 55930 */ {(12<<2)|2,{82,111,0}},
/* 55931 */ {(12<<2)|2,{82,111,0}},
/* 55932 */ {(12<<2)|2,{82,111,0}},
/* 55933 */ {(12<<2)|2,{82,111,0}},
/* 55934 */ {(12<<2)|2,{82,111,0}},
/* 55935 */ {(12<<2)|2,{82,111,0}},
/* 55936 */ {(12<<2)|2,{82,115,0}},
/* 55937 */ {(12<<2)|2,{82,115,0}},
/* 55938 */ {(12<<2)|2,{82,115,0}},
/* 55939 */ {(12<<2)|2,{82,115,0}},
/* 55940 */ {(12<<2)|2,{82,115,0}},
/* 55941 */ {(12<<2)|2,{82,115,0}},
/* 55942 */ {(12<<2)|2,{82,115,0}},
/* 55943 */ {(12<<2)|2,{82,115,0}},
/* 55944 */ {(12<<2)|2,{82,115,0}},
/* 55945 */ {(12<<2)|2,{82,115,0}},
/* 55946 */ {(12<<2)|2,{82,115,0}},
/* 55947 */ {(12<<2)|2,{82,115,0}},
/* 55948 */ {(12<<2)|2,{82,115,0}},
/* 55949 */ {(12<<2)|2,{82,115,0}},
/* 55950 */ {(12<<2)|2,{82,115,0}},
/* 55951 */ {(12<<2)|2,{82,115,0}},
/* 55952 */ {(12<<2)|2,{82,116,0}},
/* 55953 */ {(12<<2)|2,{82,116,0}},
/* 55954 */ {(12<<2)|2,{82,116,0}},
/* 55955 */ {(12<<2)|2,{82,116,0}},
/* 55956 */ {(12<<2)|2,{82,116,0}},
/* 55957 */ {(12<<2)|2,{82,116,0}},
/* 55958 */ {(12<<2)|2,{82,116,0}},
/* 55959 */ {(12<<2)|2,{82,116,0}},
/* 55960 */ {(12<<2)|2,{82,116,0}},
/* 55961 */ {(12<<2)|2,{82,116,0}},
/* 55962 */ {(12<<2)|2,{82,116,0}},
/* 55963 */ {(12<<2)|2,{82,116,0}},
/* 55964 */ {(12<<2)|2,{82,116,0}},
/* 55965 */ {(12<<2)|2,{82,116,0}},
/* 55966 */ {(12<<2)|2,{82,116,0}},
/* 55967 */ {(12<<2)|2,{82,116,0}},
/* 55968 */ {(13<<2)|2,{82,32,0}},
/* 55969 */ {(13<<2)|2,{82,32,0}},
/* 55970 */ {(13<<2)|2,{82,32,0}},
/* 55971 */ {(13<<2)|2,{82,32,0}},
/* 55972 */ {(13<<2)|2,{82,32,0}},
/* 55973 */ {(13<<2)|2,{82,32,0}},
/* 55974 */ {(13<<2)|2,{82,32,0}},
/* 55975 */ {(13<<2)|2,{82,32,0}},
/* 55976 */ {(13<<2)|2,{82,37,0}},
/* 55977 */ {(13<<2)|2,{82,37,0}},
/* 55978 */ {(13<<2)|2,{82,37,0}},
/* 55979 */ {(13<<2)|2,{82,37,0}},
/* 55980 */ {(13<<2)|2,{82,37,0}},
/* 55981 */ {(13<<2)|2,{82,37,0}},
/* 55982 */ {(13<<2)|2,{82,37,0}},
/* 55983 */ {(13<<2)|2,{82,37,0}},
/* 55984 */ {(13<<2)|2,{82,45,0}},
/* 55985 */ {(13<<2)|2,{82,45,0}},
/* 55986 */ {(13<<2)|2,{82,45,0}},
/* 55987 */ {(13<<2)|2,{82,45,0}},
/* 55988 */ {(13<<2)|2,{82,45,0}},
/* 55989 */ {(13<<2)|2,{82,45,0}},
/* 55990 */ {(13<<2)|2,{82,45,0}},
/* 55991 */ {(13<<2)|2,{82,45,0}},
/* 55992 */ {(13<<2)|2,{82,46,0}},
/* 55993 */ {(13<<2)|2,{82,46,0}},
/* 55994 */ {(13<<2)|2,{82,46,0}},
/* 55995 */ {(13<<2)|2,{82,46,0}},
/* 55996 */ {(13<<2)|2,{82,46,0}},
/* 55997 */ {(13<<2)|2,{82,46,0}},
/* 55998 */ {(13<<2)|2,{82,46,0}},
/* 55999 */ {(13<<2)|2,{82,46,0}},
/* 56000 */ {(13<<2)|2,{82,47,0}},
/* 56001 */ {(13<<2)|2,{82,47,0}},
/* 56002 */ {(13<<2)|2,{82,47,0}},
/* 56003 */ {(13<<2)|2,{82,47,0}},
/* 56004 */ {(13<<2)|2,{82,47,0}},
/* 56005 */ {(13<<2)|2,{82,47,0}},
/* 56006 */ {(13<<2)|2,{82,47,0}},
/* 56007 */ {(13<<2)|2,{82,47,0}},
/* 56008 */ {(13<<2)|2,{82,51,0}},
/* 56009 */ {(13<<2)|2,{82,51,0}},
/* 56010 */ {(13<<2)|2,{82,51,0}},
/* 56011 */ {(13<<2)|2,{82,51,0}},
/* 56012 */ {(13<<2)|2,{82,51,0}},
/* 56013 */ {(13<<2)|2,{82,51,0}},
/* 56014 */ {(13<<2)|2,{82,51,0}},
/* 56015 */ {(13<<2)|2,{82,51,0}},
/* 56016 */ {(13<<2)|2,{82,52,0}},
/* 56017 */ {(13<<2)|2,{82,52,0}},
/* 56018 */ {(13<<2)|2,{82,52,0}},
/* 56019 */ {(13<<2)|2,{82,52,0}},
/* 56020 */ {(13<<2)|2,{82,52,0}},
/* 56021 */ {(13<<2)|2,{82,52,0}},
/* 56022 */ {(13<<2)|2,{82,52,0}},
/* 56023 */ {(13<<2)|2,{82,52,0}},
/* 56024 */ {(13<<2)|2,{82,53,0}},
/* 56025 */ {(13<<2)|2,{82,53,0}},
/* 56026 */ {(13<<2)|2,{82,53,0}},
/* 56027 */ {(13<<2)|2,{82,53,0}},
/* 56028 */ {(13<<2)|2,{82,53,0}},
/* 56029 */ {(13<<2)|2,{82,53,0}},
/* 56030 */ {(13<<2)|2,{82,53,0}},
/* 56031 */ {(13<<2)|2,{82,53,0}},
/* 56032 */ {(13<<2)|2,{82,54,0}},
/* 56033 */ {(13<<2)|2,{82,54,0}},
/* 56034 */ {(13<<2)|2,{82,54,0}},
/* 56035 */ {(13<<2)|2,{82,54,0}},
/* 56036 */ {(13<<2)|2,{82,54,0}},
/* 56037 */ {(13<<2)|2,{82,54,0}},
/* 56038 */ {(13<<2)|2,{82,54,0}},
/* 56039 */ {(13<<2)|2,{82,54,0}},
/* 56040 */ {(13<<2)|2,{82,55,0}},
/* 56041 */ {(13<<2)|2,{82,55,0}},
/* 56042 */ {(13<<2)|2,{82,55,0}},
/* 56043 */ {(13<<2)|2,{82,55,0}},
/* 56044 */ {(13<<2)|2,{82,55,0}},
/* 56045 */ {(13<<2)|2,{82,55,0}},
/* 56046 */ {(13<<2)|2,{82,55,0}},
/* 56047 */ {(13<<2)|2,{82,55,0}},
/* 56048 */ {(13<<2)|2,{82,56,0}},
/* 56049 */ {(13<<2)|2,{82,56,0}},
/* 56050 */ {(13<<2)|2,{82,56,0}},
/* 56051 */ {(13<<2)|2,{82,56,0}},
/* 56052 */ {(13<<2)|2,{82,56,0}},
/* 56053 */ {(13<<2)|2,{82,56,0}},
/* 56054 */ {(13<<2)|2,{82,56,0}},
/* 56055 */ {(13<<2)|2,{82,56,0}},
/* 56056 */ {(13<<2)|2,{82,57,0}},
/* 56057 */ {(13<<2)|2,{82,57,0}},
/* 56058 */ {(13<<2)|2,{82,57,0}},
/* 56059 */ {(13<<2)|2,{82,57,0}},
/* 56060 */ {(13<<2)|2,{82,57,0}},
/* 56061 */ {(13<<2)|2,{82,57,0}},
/* 56062 */ {(13<<2)|2,{82,57,0}},
/* 56063 */ {(13<<2)|2,{82,57,0}},
/* 56064 */ {(13<<2)|2,{82,61,0}},
/* 56065 */ {(13<<2)|2,{82,61,0}},
/* 56066 */ {(13<<2)|2,{82,61,0}},
/* 56067 */ {(13<<2)|2,{82,61,0}},
/* 56068 */ {(13<<2)|2,{82,61,0}},
/* 56069 */ {(13<<2)|2,{82,61,0}},
/* 56070 */ {(13<<2)|2,{82,61,0}},
/* 56071 */ {(13<<2)|2,{82,61,0}},
/* 56072 */ {(13<<2)|2,{82,65,0}},
/* 56073 */ {(13<<2)|2,{82,65,0}},
/* 56074 */ {(13<<2)|2,{82,65,0}},
/* 56075 */ {(13<<2)|2,{82,65,0}},
/* 56076 */ {(13<<2)|2,{82,65,0}},
/* 56077 */ {(13<<2)|2,{82,65,0}},
/* 56078 */ {(13<<2)|2,{82,65,0}},
/* 56079 */ {(13<<2)|2,{82,65,0}},
/* 56080 */ {(13<<2)|2,{82,95,0}},
/* 56081 */ {(13<<2)|2,{82,95,0}},
/* 56082 */ {(13<<2)|2,{82,95,0}},
/* 56083 */ {(13<<2)|2,{82,95,0}},
/* 56084 */ {(13<<2)|2,{82,95,0}},
/* 56085 */ {(13<<2)|2,{82,95,0}},
/* 56086 */ {(13<<2)|2,{82,95,0}},
/* 56087 */ {(13<<2)|2,{82,95,0}},
/* 56088 */ {(13<<2)|2,{82,98,0}},
/* 56089 */ {(13<<2)|2,{82,98,0}},
/* 56090 */ {(13<<2)|2,{82,98,0}},
/* 56091 */ {(13<<2)|2,{82,98,0}},
/* 56092 */ {(13<<2)|2,{82,98,0}},
/* 56093 */ {(13<<2)|2,{82,98,0}},
/* 56094 */ {(13<<2)|2,{82,98,0}},
/* 56095 */ {(13<<2)|2,{82,98,0}},
/* 56096 */ {(13<<2)|2,{82,100,0}},
/* 56097 */ {(13<<2)|2,{82,100,0}},
/* 56098 */ {(13<<2)|2,{82,100,0}},
/* 56099 */ {(13<<2)|2,{82,100,0}},
/* 56100 */ {(13<<2)|2,{82,100,0}},
/* 56101 */ {(13<<2)|2,{82,100,0}},
/* 56102 */ {(13<<2)|2,{82,100,0}},
/* 56103 */ {(13<<2)|2,{82,100,0}},
/* 56104 */ {(13<<2)|2,{82,102,0}},
/* 56105 */ {(13<<2)|2,{82,102,0}},
/* 56106 */ {(13<<2)|2,{82,102,0}},
/* 56107 */ {(13<<2)|2,{82,102,0}},
/* 56108 */ {(13<<2)|2,{82,102,0}},
/* 56109 */ {(13<<2)|2,{82,102,0}},
/* 56110 */ {(13<<2)|2,{82,102,0}},
/* 56111 */ {(13<<2)|2,{82,102,0}},
/* 56112 */ {(13<<2)|2,{82,103,0}},
/* 56113 */ {(13<<2)|2,{82,103,0}},
/* 56114 */ {(13<<2)|2,{82,103,0}},
/* 56115 */ {(13<<2)|2,{82,103,0}},
/* 56116 */ {(13<<2)|2,{82,103,0}},
/* 56117 */ {(13<<2)|2,{82,103,0}},
/* 56118 */ {(13<<2)|2,{82,103,0}},
/* 56119 */ {(13<<2)|2,{82,103,0}},
/* 56120 */ {(13<<2)|2,{82,104,0}},
/* 56121 */ {(13<<2)|2,{82,104,0}},
/* 56122 */ {(13<<2)|2,{82,104,0}},
/* 56123 */ {(13<<2)|2,{82,104,0}},
/* 56124 */ {(13<<2)|2,{82,104,0}},
/* 56125 */ {(13<<2)|2,{82,104,0}},
/* 56126 */ {(13<<2)|2,{82,104,0}},
/* 56127 */ {(13<<2)|2,{82,104,0}},
/* 56128 */ {(13<<2)|2,{82,108,0}},
/* 56129 */ {(13<<2)|2,{82,108,0}},
/* 56130 */ {(13<<2)|2,{82,108,0}},
/* 56131 */ {(13<<2)|2,{82,108,0}},
/* 56132 */ {(13<<2)|2,{82,108,0}},
/* 56133 */ {(13<<2)|2,{82,108,0}},
/* 56134 */ {(13<<2)|2,{82,108,0}},
/* 56135 */ {(13<<2)|2,{82,108,0}},
/* 56136 */ {(13<<2)|2,{82,109,0}},
/* 56137 */ {(13<<2)|2,{82,109,0}},
/* 56138 */ {(13<<2)|2,{82,109,0}},
/* 56139 */ {(13<<2)|2,{82,109,0}},
/* 56140 */ {(13<<2)|2,{82,109,0}},
/* 56141 */ {(13<<2)|2,{82,109,0}},
/* 56142 */ {(13<<2)|2,{82,109,0}},
/* 56143 */ {(13<<2)|2,{82,109,0}},
/* 56144 */ {(13<<2)|2,{82,110,0}},
/* 56145 */ {(13<<2)|2,{82,110,0}},
/* 56146 */ {(13<<2)|2,{82,110,0}},
/* 56147 */ {(13<<2)|2,{82,110,0}},
/* 56148 */ {(13<<2)|2,{82,110,0}},
/* 56149 */ {(13<<2)|2,{82,110,0}},
/* 56150 */ {(13<<2)|2,{82,110,0}},
/* 56151 */ {(13<<2)|2,{82,110,0}},
/* 56152 */ {(13<<2)|2,{82,112,0}},
/* 56153 */ {(13<<2)|2,{82,112,0}},
/* 56154 */ {(13<<2)|2,{82,112,0}},
/* 56155 */ {(13<<2)|2,{82,112,0}},
/* 56156 */ {(13<<2)|2,{82,112,0}},
/* 56157 */ {(13<<2)|2,{82,112,0}},
/* 56158 */ {(13<<2)|2,{82,112,0}},
/* 56159 */ {(13<<2)|2,{82,112,0}},
/* 56160 */ {(13<<2)|2,{82,114,0}},
/* 56161 */ {(13<<2)|2,{82,114,0}},
/* 56162 */ {(13<<2)|2,{82,114,0}},
/* 56163 */ {(13<<2)|2,{82,114,0}},
/* 56164 */ {(13<<2)|2,{82,114,0}},
/* 56165 */ {(13<<2)|2,{82,114,0}},
/* 56166 */ {(13<<2)|2,{82,114,0}},
/* 56167 */ {(13<<2)|2,{82,114,0}},
/* 56168 */ {(13<<2)|2,{82,117,0}},
/* 56169 */ {(13<<2)|2,{82,117,0}},
/* 56170 */ {(13<<2)|2,{82,117,0}},
/* 56171 */ {(13<<2)|2,{82,117,0}},
/* 56172 */ {(13<<2)|2,{82,117,0}},
/* 56173 */ {(13<<2)|2,{82,117,0}},
/* 56174 */ {(13<<2)|2,{82,117,0}},
/* 56175 */ {(13<<2)|2,{82,117,0}},
/* 56176 */ {(14<<2)|2,{82,58,0}},
/* 56177 */ {(14<<2)|2,{82,58,0}},
/* 56178 */ {(14<<2)|2,{82,58,0}},
/* 56179 */ {(14<<2)|2,{82,58,0}},
/* 56180 */ {(14<<2)|2,{82,66,0}},
/* 56181 */ {(14<<2)|2,{82,66,0}},
/* 56182 */ {(14<<2)|2,{82,66,0}},
/* 56183 */ {(14<<2)|2,{82,66,0}},
/* 56184 */ {(14<<2)|2,{82,67,0}},
/* 56185 */ {(14<<2)|2,{82,67,0}},
/* 56186 */ {(14<<2)|2,{82,67,0}},
/* 56187 */ {(14<<2)|2,{82,67,0}},
/* 56188 */ {(14<<2)|2,{82,68,0}},
/* 56189 */ {(14<<2)|2,{82,68,0}},
/* 56190 */ {(14<<2)|2,{82,68,0}},
/* 56191 */ {(14<<2)|2,{82,68,0}},
/* 56192 */ {(14<<2)|2,{82,69,0}},
/* 56193 */ {(14<<2)|2,{82,69,0}},
/* 56194 */ {(14<<2)|2,{82,69,0}},
/* 56195 */ {(14<<2)|2,{82,69,0}},
/* 56196 */ {(14<<2)|2,{82,70,0}},
/* 56197 */ {(14<<2)|2,{82,70,0}},
/* 56198 */ {(14<<2)|2,{82,70,0}},
/* 56199 */ {(14<<2)|2,{82,70,0}},
/* 56200 */ {(14<<2)|2,{82,71,0}},
/* 56201 */ {(14<<2)|2,{82,71,0}},
/* 56202 */ {(14<<2)|2,{82,71,0}},
/* 56203 */ {(14<<2)|2,{82,71,0}},
/* 56204 */ {(14<<2)|2,{82,72,0}},
/* 56205 */ {(14<<2)|2,{82,72,0}},
/* 56206 */ {(14<<2)|2,{82,72,0}},
/* 56207 */ {(14<<2)|2,{82,72,0}},
/* 56208 */ {(14<<2)|2,{82,73,0}},
/* 56209 */ {(14<<2)|2,{82,73,0}},
/* 56210 */ {(14<<2)|2,{82,73,0}},
/* 56211 */ {(14<<2)|2,{82,73,0}},
/* 56212 */ {(14<<2)|2,{82,74,0}},
/* 56213 */ {(14<<2)|2,{82,74,0}},
/* 56214 */ {(14<<2)|2,{82,74,0}},
/* 56215 */ {(14<<2)|2,{82,74,0}},
/* 56216 */ {(14<<2)|2,{82,75,0}},
/* 56217 */ {(14<<2)|2,{82,75,0}},
/* 56218 */ {(14<<2)|2,{82,75,0}},
/* 56219 */ {(14<<2)|2,{82,75,0}},
/* 56220 */ {(14<<2)|2,{82,76,0}},
/* 56221 */ {(14<<2)|2,{82,76,0}},
/* 56222 */ {(14<<2)|2,{82,76,0}},
/* 56223 */ {(14<<2)|2,{82,76,0}},
/* 56224 */ {(14<<2)|2,{82,77,0}},
/* 56225 */ {(14<<2)|2,{82,77,0}},
/* 56226 */ {(14<<2)|2,{82,77,0}},
/* 56227 */ {(14<<2)|2,{82,77,0}},
/* 56228 */ {(14<<2)|2,{82,78,0}},
/* 56229 */ {(14<<2)|2,{82,78,0}},
/* 56230 */ {(14<<2)|2,{82,78,0}},
/* 56231 */ {(14<<2)|2,{82,78,0}},
/* 56232 */ {(14<<2)|2,{82,79,0}},
/* 56233 */ {(14<<2)|2,{82,79,0}},
/* 56234 */ {(14<<2)|2,{82,79,0}},
/* 56235 */ {(14<<2)|2,{82,79,0}},
/* 56236 */ {(14<<2)|2,{82,80,0}},
/* 56237 */ {(14<<2)|2,{82,80,0}},
/* 56238 */ {(14<<2)|2,{82,80,0}},
/* 56239 */ {(14<<2)|2,{82,80,0}},
/* 56240 */ {(14<<2)|2,{82,81,0}},
/* 56241 */ {(14<<2)|2,{82,81,0}},
/* 56242 */ {(14<<2)|2,{82,81,0}},
/* 56243 */ {(14<<2)|2,{82,81,0}},
/* 56244 */ {(14<<2)|2,{82,82,0}},
/* 56245 */ {(14<<2)|2,{82,82,0}},
/* 56246 */ {(14<<2)|2,{82,82,0}},
/* 56247 */ {(14<<2)|2,{82,82,0}},
/* 56248 */ {(14<<2)|2,{82,83,0}},
/* 56249 */ {(14<<2)|2,{82,83,0}},
/* 56250 */ {(14<<2)|2,{82,83,0}},
/* 56251 */ {(14<<2)|2,{82,83,0}},
/* 56252 */ {(14<<2)|2,{82,84,0}},
/* 56253 */ {(14<<2)|2,{82,84,0}},
/* 56254 */ {(14<<2)|2,{82,84,0}},
/* 56255 */ {(14<<2)|2,{82,84,0}},
/* 56256 */ {(14<<2)|2,{82,85,0}},
/* 56257 */ {(14<<2)|2,{82,85,0}},
/* 56258 */ {(14<<2)|2,{82,85,0}},
/* 56259 */ {(14<<2)|2,{82,85,0}},
/* 56260 */ {(14<<2)|2,{82,86,0}},
/* 56261 */ {(14<<2)|2,{82,86,0}},
/* 56262 */ {(14<<2)|2,{82,86,0}},
/* 56263 */ {(14<<2)|2,{82,86,0}},
/* 56264 */ {(14<<2)|2,{82,87,0}},
/* 56265 */ {(14<<2)|2,{82,87,0}},
/* 56266 */ {(14<<2)|2,{82,87,0}},
/* 56267 */ {(14<<2)|2,{82,87,0}},
/* 56268 */ {(14<<2)|2,{82,89,0}},
/* 56269 */ {(14<<2)|2,{82,89,0}},
/* 56270 */ {(14<<2)|2,{82,89,0}},
/* 56271 */ {(14<<2)|2,{82,89,0}},
/* 56272 */ {(14<<2)|2,{82,106,0}},
/* 56273 */ {(14<<2)|2,{82,106,0}},
/* 56274 */ {(14<<2)|2,{82,106,0}},
/* 56275 */ {(14<<2)|2,{82,106,0}},
/* 56276 */ {(14<<2)|2,{82,107,0}},
/* 56277 */ {(14<<2)|2,{82,107,0}},
/* 56278 */ {(14<<2)|2,{82,107,0}},
/* 56279 */ {(14<<2)|2,{82,107,0}},
/* 56280 */ {(14<<2)|2,{82,113,0}},
/* 56281 */ {(14<<2)|2,{82,113,0}},
/* 56282 */ {(14<<2)|2,{82,113,0}},
/* 56283 */ {(14<<2)|2,{82,113,0}},
/* 56284 */ {(14<<2)|2,{82,118,0}},
/* 56285 */ {(14<<2)|2,{82,118,0}},
/* 56286 */ {(14<<2)|2,{82,118,0}},
/* 56287 */ {(14<<2)|2,{82,118,0}},
/* 56288 */ {(14<<2)|2,{82,119,0}},
/* 56289 */ {(14<<2)|2,{82,119,0}},
/* 56290 */ {(14<<2)|2,{82,119,0}},
/* 56291 */ {(14<<2)|2,{82,119,0}},
/* 56292 */ {(14<<2)|2,{82,120,0}},
/* 56293 */ {(14<<2)|2,{82,120,0}},
/* 56294 */ {(14<<2)|2,{82,120,0}},
/* 56295 */ {(14<<2)|2,{82,120,0}},
/* 56296 */ {(14<<2)|2,{82,121,0}},
/* 56297 */ {(14<<2)|2,{82,121,0}},
/* 56298 */ {(14<<2)|2,{82,121,0}},
/* 56299 */ {(14<<2)|2,{82,121,0}},
/* 56300 */ {(14<<2)|2,{82,122,0}},
/* 56301 */ {(14<<2)|2,{82,122,0}},
/* 56302 */ {(14<<2)|2,{82,122,0}},
/* 56303 */ {(14<<2)|2,{82,122,0}},
/* 56304 */ {(15<<2)|2,{82,38,0}},
/* 56305 */ {(15<<2)|2,{82,38,0}},
/* 56306 */ {(15<<2)|2,{82,42,0}},
/* 56307 */ {(15<<2)|2,{82,42,0}},
/* 56308 */ {(15<<2)|2,{82,44,0}},
/* 56309 */ {(15<<2)|2,{82,44,0}},
/* 56310 */ {(15<<2)|2,{82,59,0}},
/* 56311 */ {(15<<2)|2,{82,59,0}},
/* 56312 */ {(15<<2)|2,{82,88,0}},
/* 56313 */ {(15<<2)|2,{82,88,0}},
/* 56314 */ {(15<<2)|2,{82,90,0}},
/* 56315 */ {(15<<2)|2,{82,90,0}},
/* 56316 */ {(7<<2)|1,{82,0,0}},
/* 56317 */ {(7<<2)|1,{82,0,0}},
/* 56318 */ {(7<<2)|1,{82,0,0}},
/* 56319 */ {(7<<2)|1,{82,0,0}},
/* 56320 */ {(12<<2)|2,{83,48,0}},
/* 56321 */ {(12<<2)|2,{83,48,0}},
/* 56322 */ {(12<<2)|2,{83,48,0}},
/* 56323 */ {(12<<2)|2,{83,48,0}},
/* 56324 */ {(12<<2)|2,{83,48,0}},
/* 56325 */ {(12<<2)|2,{83,48,0}},
/* 56326 */ {(12<<2)|2,{83,48,0}},
/* 56327 */ {(12<<2)|2,{83,48,0}},
/* 56328 */ {(12<<2)|2,{83,48,0}},
/* 56329 */ {(12<<2)|2,{83,48,0}},
/* 56330 */ {(12<<2)|2,{83,48,0}},
/* 56331 */ {(12<<2)|2,{83,48,0}},
/* 56332 */ {(12<<2)|2,{83,48,0}},
/* 56333 */ {(12<<2)|2,{83,48,0}},
/* 56334 */ {(12<<2)|2,{83,48,0}},
/* 56335 */ {(12<<2)|2,{83,48,0}},
/* 56336 */ {(12<<2)|2,{83,49,0}},
/* 56337 */ {(12<<2)|2,{83,49,0}},
/* 56338 */ {(12<<2)|2,{83,49,0}},
/* 56339 */ {(12<<2)|2,{83,49,0}},
/* 56340 */ {(12<<2)|2,{83,49,0}},
/* 56341 */ {(12<<2)|2,{83,49,0}},
/* 56342 */ {(12<<2)|2,{83,49,0}},
/* 56343 */ {(12<<2)|2,{83,49,0}},
/* 56344 */ {(12<<2)|2,{83,49,0}},
/* 56345 */ {(12<<2)|2,{83,49,0}},
/* 56346 */ {(12<<2)|2,{83,49,0}},
/* 56347 */ {(12<<2)|2,{83,49,0}},
/* 56348 */ {(12<<2)|2,{83,49,0}},
/* 56349 */ {(12<<2)|2,{83,49,0}},
/* 56350 */ {(12<<2)|2,{83,49,0}},
/* 56351 */ {(12<<2)|2,{83,49,0}},
/* 56352 */ {(12<<2)|2,{83,50,0}},
/* 56353 */ {(12<<2)|2,{83,50,0}},
/* 56354 */ {(12<<2)|2,{83,50,0}},
/* 56355 */ {(12<<2)|2,{83,50,0}},
/* 56356 */ {(12<<2)|2,{83,50,0}},
/* 56357 */ {(12<<2)|2,{83,50,0}},
/* 56358 */ {(12<<2)|2,{83,50,0}},
/* 56359 */ {(12<<2)|2,{83,50,0}},
/* 56360 */ {(12<<2)|2,{83,50,0}},
/* 56361 */ {(12<<2)|2,{83,50,0}},
/* 56362 */ {(12<<2)|2,{83,50,0}},
/* 56363 */ {(12<<2)|2,{83,50,0}},
/* 56364 */ {(12<<2)|2,{83,50,0}},
/* 56365 */ {(12<<2)|2,{83,50,0}},
/* 56366 */ {(12<<2)|2,{83,50,0}},
/* 56367 */ {(12<<2)|2,{83,50,0}},
/* 56368 */ {(12<<2)|2,{83,97,0}},
/* 56369 */ {(12<<2)|2,{83,97,0}},
/* 56370 */ {(12<<2)|2,{83,97,0}},
/* 56371 */ {(12<<2)|2,{83,97,0}},
/* 56372 */ {(12<<2)|2,{83,97,0}},
/* 56373 */ {(12<<2)|2,{83,97,0}},
/* 56374 */ {(12<<2)|2,{83,97,0}},
/* 56375 */ {(12<<2)|2,{83,97,0}},
/* 56376 */ {(12<<2)|2,{83,97,0}},
/* 56377 */ {(12<<2)|2,{83,97,0}},
/* 56378 */ {(12<<2)|2,{83,97,0}},
/* 56379 */ {(12<<2)|2,{83,97,0}},
/* 56380 */ {(12<<2)|2,{83,97,0}},
/* 56381 */ {(12<<2)|2,{83,97,0}},
/* 56382 */ {(12<<2)|2,{83,97,0}},
/* 56383 */ {(12<<2)|2,{83,97,0}},
/* 56384 */ {(12<<2)|2,{83,99,0}},
/* 56385 */ {(12<<2)|2,{83,99,0}},
/* 56386 */ {(12<<2)|2,{83,99,0}},
/* 56387 */ {(12<<2)|2,{83,99,0}},
/* 56388 */ {(12<<2)|2,{83,99,0}},
/* 56389 */ {(12<<2)|2,{83,99,0}},
/* 56390 */ {(12<<2)|2,{83,99,0}},
/* 56391 */ {(12<<2)|2,{83,99,0}},
/* 56392 */ {(12<<2)|2,{83,99,0}},
/* 56393 */ {(12<<2)|2,{83,99,0}},
/* 56394 */ {(12<<2)|2,{83,99,0}},
/* 56395 */ {(12<<2)|2,{83,99,0}},
/* 56396 */ {(12<<2)|2,{83,99,0}},
/* 56397 */ {(12<<2)|2,{83,99,0}},
/* 56398 */ {(12<<2)|2,{83,99,0}},
/* 56399 */ {(12<<2)|2,{83,99,0}},
/* 56400 */ {(12<<2)|2,{83,101,0}},
/* 56401 */ {(12<<2)|2,{83,101,0}},
/* 56402 */ {(12<<2)|2,{83,101,0}},
/* 56403 */ {(12<<2)|2,{83,101,0}},
/* 56404 */ {(12<<2)|2,{83,101,0}},
/* 56405 */ {(12<<2)|2,{83,101,0}},
/* 56406 */ {(12<<2)|2,{83,101,0}},
/* 56407 */ {(12<<2)|2,{83,101,0}},
/* 56408 */ {(12<<2)|2,{83,101,0}},
/* 56409 */ {(12<<2)|2,{83,101,0}},
/* 56410 */ {(12<<2)|2,{83,101,0}},
/* 56411 */ {(12<<2)|2,{83,101,0}},
/* 56412 */ {(12<<2)|2,{83,101,0}},
/* 56413 */ {(12<<2)|2,{83,101,0}},
/* 56414 */ {(12<<2)|2,{83,101,0}},
/* 56415 */ {(12<<2)|2,{83,101,0}},
/* 56416 */ {(12<<2)|2,{83,105,0}},
/* 56417 */ {(12<<2)|2,{83,105,0}},
/* 56418 */ {(12<<2)|2,{83,105,0}},
/* 56419 */ {(12<<2)|2,{83,105,0}},
/* 56420 */ {(12<<2)|2,{83,105,0}},
/* 56421 */ {(12<<2)|2,{83,105,0}},
/* 56422 */ {(12<<2)|2,{83,105,0}},
/* 56423 */ {(12<<2)|2,{83,105,0}},
/* 56424 */ {(12<<2)|2,{83,105,0}},
/* 56425 */ {(12<<2)|2,{83,105,0}},
/* 56426 */ {(12<<2)|2,{83,105,0}},
/* 56427 */ {(12<<2)|2,{83,105,0}},
/* 56428 */ {(12<<2)|2,{83,105,0}},
/* 56429 */ {(12<<2)|2,{83,105,0}},
/* 56430 */ {(12<<2)|2,{83,105,0}},
/* 56431 */ {(12<<2)|2,{83,105,0}},
/* 56432 */ {(12<<2)|2,{83,111,0}},
/* 56433 */ {(12<<2)|2,{83,111,0}},
/* 56434 */ {(12<<2)|2,{83,111,0}},
/* 56435 */ {(12<<2)|2,{83,111,0}},
/* 56436 */ {(12<<2)|2,{83,111,0}},
/* 56437 */ {(12<<2)|2,{83,111,0}},
/* 56438 */ {(12<<2)|2,{83,111,0}},
/* 56439 */ {(12<<2)|2,{83,111,0}},
/* 56440 */ {(12<<2)|2,{83,111,0}},
/* 56441 */ {(12<<2)|2,{83,111,0}},
/* 56442 */ {(12<<2)|2,{83,111,0}},
/* 56443 */ {(12<<2)|2,{83,111,0}},
/* 56444 */ {(12<<2)|2,{83,111,0}},
/* 56445 */ {(12<<2)|2,{83,111,0}},
/* 56446 */ {(12<<2)|2,{83,111,0}},
/* 56447 */ {(12<<2)|2,{83,111,0}},
/* 56448 */ {(12<<2)|2,{83,115,0}},
/* 56449 */ {(12<<2)|2,{83,115,0}},
/* 56450 */ {(12<<2)|2,{83,115,0}},
/* 56451 */ {(12<<2)|2,{83,115,0}},
/* 56452 */ {(12<<2)|2,{83,115,0}},
/* 56453 */ {(12<<2)|2,{83,115,0}},
/* 56454 */ {(12<<2)|2,{83,115,0}},
/* 56455 */ {(12<<2)|2,{83,115,0}},
/* 56456 */ {(12<<2)|2,{83,115,0}},
/* 56457 */ {(12<<2)|2,{83,115,0}},
/* 56458 */ {(12<<2)|2,{83,115,0}},
/* 56459 */ {(12<<2)|2,{83,115,0}},
/* 56460 */ {(12<<2)|2,{83,115,0}},
/* 56461 */ {(12<<2)|2,{83,115,0}},
/* 56462 */ {(12<<2)|2,{83,115,0}},
/* 56463 */ {(12<<2)|2,{83,115,0}},
/* 56464 */ {(12<<2)|2,{83,116,0}},
/* 56465 */ {(12<<2)|2,{83,116,0}},
/* 56466 */ {(12<<2)|2,{83,116,0}},
/* 56467 */ {(12<<2)|2,{83,116,0}},
/* 56468 */ {(12<<2)|2,{83,116,0}},
/* 56469 */ {(12<<2)|2,{83,116,0}},
/* 56470 */ {(12<<2)|2,{83,116,0}},
/* 56471 */ {(12<<2)|2,{83,116,0}},
/* 56472 */ {(12<<2)|2,{83,116,0}},
/* 56473 */ {(12<<2)|2,{83,116,0}},
/* 56474 */ {(12<<2)|2,{83,116,0}},
/* 56475 */ {(12<<2)|2,{83,116,0}},
/* 56476 */ {(12<<2)|2,{83,116,0}},
/* 56477 */ {(12<<2)|2,{83,116,0}},
/* 56478 */ {(12<<2)|2,{83,116,0}},
/* 56479 */ {(12<<2)|2,{83,116,0}},
/* 56480 */ {(13<<2)|2,{83,32,0}},
/* 56481 */ {(13<<2)|2,{83,32,0}},
/* 56482 */ {(13<<2)|2,{83,32,0}},
/* 56483 */ {(13<<2)|2,{83,32,0}},
/* 56484 */ {(13<<2)|2,{83,32,0}},
/* 56485 */ {(13<<2)|2,{83,32,0}},
/* 56486 */ {(13<<2)|2,{83,32,0}},
/* 56487 */ {(13<<2)|2,{83,32,0}},
/* 56488 */ {(13<<2)|2,{83,37,0}},
/* 56489 */ {(13<<2)|2,{83,37,0}},
/* 56490 */ {(13<<2)|2,{83,37,0}},
/* 56491 */ {(13<<2)|2,{83,37,0}},
/* 56492 */ {(13<<2)|2,{83,37,0}},
/* 56493 */ {(13<<2)|2,{83,37,0}},
/* 56494 */ {(13<<2)|2,{83,37,0}},
/* 56495 */ {(13<<2)|2,{83,37,0}},
/* 56496 */ {(13<<2)|2,{83,45,0}},
/* 56497 */ {(13<<2)|2,{83,45,0}},
/* 56498 */ {(13<<2)|2,{83,45,0}},
/* 56499 */ {(13<<2)|2,{83,45,0}},
/* 56500 */ {(13<<2)|2,{83,45,0}},
/* 56501 */ {(13<<2)|2,{83,45,0}},
/* 56502 */ {(13<<2)|2,{83,45,0}},
/* 56503 */ {(13<<2)|2,{83,45,0}},
/* 56504 */ {(13<<2)|2,{83,46,0}},
/* 56505 */ {(13<<2)|2,{83,46,0}},
/* 56506 */ {(13<<2)|2,{83,46,0}},
/* 56507 */ {(13<<2)|2,{83,46,0}},
/* 56508 */ {(13<<2)|2,{83,46,0}},
/* 56509 */ {(13<<2)|2,{83,46,0}},
/* 56510 */ {(13<<2)|2,{83,46,0}},
/* 56511 */ {(13<<2)|2,{83,46,0}},
/* 56512 */ {(13<<2)|2,{83,47,0}},
/* 56513 */ {(13<<2)|2,{83,47,0}},
/* 56514 */ {(13<<2)|2,{83,47,0}},
/* 56515 */ {(13<<2)|2,{83,47,0}},
/* 56516 */ {(13<<2)|2,{83,47,0}},
/* 56517 */ {(13<<2)|2,{83,47,0}},
/* 56518 */ {(13<<2)|2,{83,47,0}},
/* 56519 */ {(13<<2)|2,{83,47,0}},
/* 56520 */ {(13<<2)|2,{83,51,0}},
/* 56521 */ {(13<<2)|2,{83,51,0}},
/* 56522 */ {(13<<2)|2,{83,51,0}},
/* 56523 */ {(13<<2)|2,{83,51,0}},
/* 56524 */ {(13<<2)|2,{83,51,0}},
/* 56525 */ {(13<<2)|2,{83,51,0}},
/* 56526 */ {(13<<2)|2,{83,51,0}},
/* 56527 */ {(13<<2)|2,{83,51,0}},
/* 56528 */ {(13<<2)|2,{83,52,0}},
/* 56529 */ {(13<<2)|2,{83,52,0}},
/* 56530 */ {(13<<2)|2,{83,52,0}},
/* 56531 */ {(13<<2)|2,{83,52,0}},
/* 56532 */ {(13<<2)|2,{83,52,0}},
/* 56533 */ {(13<<2)|2,{83,52,0}},
/* 56534 */ {(13<<2)|2,{83,52,0}},
/* 56535 */ {(13<<2)|2,{83,52,0}},
/* 56536 */ {(13<<2)|2,{83,53,0}},
/* 56537 */ {(13<<2)|2,{83,53,0}},
/* 56538 */ {(13<<2)|2,{83,53,0}},
/* 56539 */ {(13<<2)|2,{83,53,0}},
/* 56540 */ {(13<<2)|2,{83,53,0}},
/* 56541 */ {(13<<2)|2,{83,53,0}},
/* 56542 */ {(13<<2)|2,{83,53,0}},
/* 56543 */ {(13<<2)|2,{83,53,0}},
/* 56544 */ {(13<<2)|2,{83,54,0}},
/* 56545 */ {(13<<2)|2,{83,54,0}},
/* 56546 */ {(13<<2)|2,{83,54,0}},
/* 56547 */ {(13<<2)|2,{83,54,0}},
/* 56548 */ {(13<<2)|2,{83,54,0}},
/* 56549 */ {(13<<2)|2,{83,54,0}},
/* 56550 */ {(13<<2)|2,{83,54,0}},
/* 56551 */ {(13<<2)|2,{83,54,0}},
/* 56552 */ {(13<<2)|2,{83,55,0}},
/* 56553 */ {(13<<2)|2,{83,55,0}},
/* 56554 */ {(13<<2)|2,{83,55,0}},
/* 56555 */ {(13<<2)|2,{83,55,0}},
/* 56556 */ {(13<<2)|2,{83,55,0}},
/* 56557 */ {(13<<2)|2,{83,55,0}},
/* 56558 */ {(13<<2)|2,{83,55,0}},
/* 56559 */ {(13<<2)|2,{83,55,0}},
/* 56560 */ {(13<<2)|2,{83,56,0}},
/* 56561 */ {(13<<2)|2,{83,56,0}},
/* 56562 */ {(13<<2)|2,{83,56,0}},
/* 56563 */ {(13<<2)|2,{83,56,0}},
/* 56564 */ {(13<<2)|2,{83,56,0}},
/* 56565 */ {(13<<2)|2,{83,56,0}},
/* 56566 */ {(13<<2)|2,{83,56,0}},
/* 56567 */ {(13<<2)|2,{83,56,0}},
/* 56568 */ {(13<<2)|2,{83,57,0}},
/* 56569 */ {(13<<2)|2,{83,57,0}},
/* 56570 */ {(13<<2)|2,{83,57,0}},
/* 56571 */ {(13<<2)|2,{83,57,0}},
/* 56572 */ {(13<<2)|2,{83,57,0}},
/* 56573 */ {(13<<2)|2,{83,57,0}},
/* 56574 */ {(13<<2)|2,{83,57,0}},
/* 56575 */ {(13<<2)|2,{83,57,0}},
/* 56576 */ {(13<<2)|2,{83,61,0}},
/* 56577 */ {(13<<2)|2,{83,61,0}},
/* 56578 */ {(13<<2)|2,{83,61,0}},
/* 56579 */ {(13<<2)|2,{83,61,0}},
/* 56580 */ {(13<<2)|2,{83,61,0}},
/* 56581 */ {(13<<2)|2,{83,61,0}},
/* 56582 */ {(13<<2)|2,{83,61,0}},
/* 56583 */ {(13<<2)|2,{83,61,0}},
/* 56584 */ {(13<<2)|2,{83,65,0}},
/* 56585 */ {(13<<2)|2,{83,65,0}},
/* 56586 */ {(13<<2)|2,{83,65,0}},
/* 56587 */ {(13<<2)|2,{83,65,0}},
/* 56588 */ {(13<<2)|2,{83,65,0}},
/* 56589 */ {(13<<2)|2,{83,65,0}},
/* 56590 */ {(13<<2)|2,{83,65,0}},
/* 56591 */ {(13<<2)|2,{83,65,0}},
/* 56592 */ {(13<<2)|2,{83,95,0}},
/* 56593 */ {(13<<2)|2,{83,95,0}},
/* 56594 */ {(13<<2)|2,{83,95,0}},
/* 56595 */ {(13<<2)|2,{83,95,0}},
/* 56596 */ {(13<<2)|2,{83,95,0}},
/* 56597 */ {(13<<2)|2,{83,95,0}},
/* 56598 */ {(13<<2)|2,{83,95,0}},
/* 56599 */ {(13<<2)|2,{83,95,0}},
/* 56600 */ {(13<<2)|2,{83,98,0}},
/* 56601 */ {(13<<2)|2,{83,98,0}},
/* 56602 */ {(13<<2)|2,{83,98,0}},
/* 56603 */ {(13<<2)|2,{83,98,0}},
/* 56604 */ {(13<<2)|2,{83,98,0}},
/* 56605 */ {(13<<2)|2,{83,98,0}},
/* 56606 */ {(13<<2)|2,{83,98,0}},
/* 56607 */ {(13<<2)|2,{83,98,0}},
/* 56608 */ {(13<<2)|2,{83,100,0}},
/* 56609 */ {(13<<2)|2,{83,100,0}},
/* 56610 */ {(13<<2)|2,{83,100,0}},
/* 56611 */ {(13<<2)|2,{83,100,0}},
/* 56612 */ {(13<<2)|2,{83,100,0}},
/* 56613 */ {(13<<2)|2,{83,100,0}},
/* 56614 */ {(13<<2)|2,{83,100,0}},
/* 56615 */ {(13<<2)|2,{83,100,0}},
/* 56616 */ {(13<<2)|2,{83,102,0}},
/* 56617 */ {(13<<2)|2,{83,102,0}},
/* 56618 */ {(13<<2)|2,{83,102,0}},
/* 56619 */ {(13<<2)|2,{83,102,0}},
/* 56620 */ {(13<<2)|2,{83,102,0}},
/* 56621 */ {(13<<2)|2,{83,102,0}},
/* 56622 */ {(13<<2)|2,{83,102,0}},
/* 56623 */ {(13<<2)|2,{83,102,0}},
/* 56624 */ {(13<<2)|2,{83,103,0}},
/* 56625 */ {(13<<2)|2,{83,103,0}},
/* 56626 */ {(13<<2)|2,{83,103,0}},
/* 56627 */ {(13<<2)|2,{83,103,0}},
/* 56628 */ {(13<<2)|2,{83,103,0}},
/* 56629 */ {(13<<2)|2,{83,103,0}},
/* 56630 */ {(13<<2)|2,{83,103,0}},
/* 56631 */ {(13<<2)|2,{83,103,0}},
/* 56632 */ {(13<<2)|2,{83,104,0}},
/* 56633 */ {(13<<2)|2,{83,104,0}},
/* 56634 */ {(13<<2)|2,{83,104,0}},
/* 56635 */ {(13<<2)|2,{83,104,0}},
/* 56636 */ {(13<<2)|2,{83,104,0}},
/* 56637 */ {(13<<2)|2,{83,104,0}},
/* 56638 */ {(13<<2)|2,{83,104,0}},
/* 56639 */ {(13<<2)|2,{83,104,0}},
/* 56640 */ {(13<<2)|2,{83,108,0}},
/* 56641 */ {(13<<2)|2,{83,108,0}},
/* 56642 */ {(13<<2)|2,{83,108,0}},
/* 56643 */ {(13<<2)|2,{83,108,0}},
/* 56644 */ {(13<<2)|2,{83,108,0}},
/* 56645 */ {(13<<2)|2,{83,108,0}},
/* 56646 */ {(13<<2)|2,{83,108,0}},
/* 56647 */ {(13<<2)|2,{83,108,0}},
/* 56648 */ {(13<<2)|2,{83,109,0}},
/* 56649 */ {(13<<2)|2,{83,109,0}},
/* 56650 */ {(13<<2)|2,{83,109,0}},
/* 56651 */ {(13<<2)|2,{83,109,0}},
/* 56652 */ {(13<<2)|2,{83,109,0}},
/* 56653 */ {(13<<2)|2,{83,109,0}},
/* 56654 */ {(13<<2)|2,{83,109,0}},
/* 56655 */ {(13<<2)|2,{83,109,0}},
/* 56656 */ {(13<<2)|2,{83,110,0}},
/* 56657 */ {(13<<2)|2,{83,110,0}},
/* 56658 */ {(13<<2)|2,{83,110,0}},
/* 56659 */ {(13<<2)|2,{83,110,0}},
/* 56660 */ {(13<<2)|2,{83,110,0}},
/* 56661 */ {(13<<2)|2,{83,110,0}},
/* 56662 */ {(13<<2)|2,{83,110,0}},
/* 56663 */ {(13<<2)|2,{83,110,0}},
/* 56664 */ {(13<<2)|2,{83,112,0}},
/* 56665 */ {(13<<2)|2,{83,112,0}},
/* 56666 */ {(13<<2)|2,{83,112,0}},
/* 56667 */ {(13<<2)|2,{83,112,0}},
/* 56668 */ {(13<<2)|2,{83,112,0}},
/* 56669 */ {(13<<2)|2,{83,112,0}},
/* 56670 */ {(13<<2)|2,{83,112,0}},
/* 56671 */ {(13<<2)|2,{83,112,0}},
/* 56672 */ {(13<<2)|2,{83,114,0}},
/* 56673 */ {(13<<2)|2,{83,114,0}},
/* 56674 */ {(13<<2)|2,{83,114,0}},
/* 56675 */ {(13<<2)|2,{83,114,0}},
/* 56676 */ {(13<<2)|2,{83,114,0}},
/* 56677 */ {(13<<2)|2,{83,114,0}},
/* 56678 */ {(13<<2)|2,{83,114,0}},
/* 56679 */ {(13<<2)|2,{83,114,0}},
/* 56680 */ {(13<<2)|2,{83,117,0}},
/* 56681 */ {(13<<2)|2,{83,117,0}},
/* 56682 */ {(13<<2)|2,{83,117,0}},
/* 56683 */ {(13<<2)|2,{83,117,0}},
/* 56684 */ {(13<<2)|2,{83,117,0}},
/* 56685 */ {(13<<2)|2,{83,117,0}},
/* 56686 */ {(13<<2)|2,{83,117,0}},
/* 56687 */ {(13<<2)|2,{83,117,0}},
/* 56688 */ {(14<<2)|2,{83,58,0}},
/* 56689 */ {(14<<2)|2,{83,58,0}},
/* 56690 */ {(14<<2)|2,{83,58,0}},
/* 56691 */ {(14<<2)|2,{83,58,0}},
/* 56692 */ {(14<<2)|2,{83,66,0}},
/* 56693 */ {(14<<2)|2,{83,66,0}},
/* 56694 */ {(14<<2)|2,{83,66,0}},
/* 56695 */ {(14<<2)|2,{83,66,0}},
/* 56696 */ {(14<<2)|2,{83,67,0}},
/* 56697 */ {(14<<2)|2,{83,67,0}},
/* 56698 */ {(14<<2)|2,{83,67,0}},
/* 56699 */ {(14<<2)|2,{83,67,0}},
/* 56700 */ {(14<<2)|2,{83,68,0}},
/* 56701 */ {(14<<2)|2,{83,68,0}},
/* 56702 */ {(14<<2)|2,{83,68,0}},
/* 56703 */ {(14<<2)|2,{83,68,0}},
/* 56704 */ {(14<<2)|2,{83,69,0}},
/* 56705 */ {(14<<2)|2,{83,69,0}},
/* 56706 */ {(14<<2)|2,{83,69,0}},
/* 56707 */ {(14<<2)|2,{83,69,0}},
/* 56708 */ {(14<<2)|2,{83,70,0}},
/* 56709 */ {(14<<2)|2,{83,70,0}},
/* 56710 */ {(14<<2)|2,{83,70,0}},
/* 56711 */ {(14<<2)|2,{83,70,0}},
/* 56712 */ {(14<<2)|2,{83,71,0}},
/* 56713 */ {(14<<2)|2,{83,71,0}},
/* 56714 */ {(14<<2)|2,{83,71,0}},
/* 56715 */ {(14<<2)|2,{83,71,0}},
/* 56716 */ {(14<<2)|2,{83,72,0}},
/* 56717 */ {(14<<2)|2,{83,72,0}},
/* 56718 */ {(14<<2)|2,{83,72,0}},
/* 56719 */ {(14<<2)|2,{83,72,0}},
/* 56720 */ {(14<<2)|2,{83,73,0}},
/* 56721 */ {(14<<2)|2,{83,73,0}},
/* 56722 */ {(14<<2)|2,{83,73,0}},
/* 56723 */ {(14<<2)|2,{83,73,0}},
/* 56724 */ {(14<<2)|2,{83,74,0}},
/* 56725 */ {(14<<2)|2,{83,74,0}},
/* 56726 */ {(14<<2)|2,{83,74,0}},
/* 56727 */ {(14<<2)|2,{83,74,0}},
/* 56728 */ {(14<<2)|2,{83,75,0}},
/* 56729 */ {(14<<2)|2,{83,75,0}},
/* 56730 */ {(14<<2)|2,{83,75,0}},
/* 56731 */ {(14<<2)|2,{83,75,0}},
/* 56732 */ {(14<<2)|2,{83,76,0}},
/* 56733 */ {(14<<2)|2,{83,76,0}},
/* 56734 */ {(14<<2)|2,{83,76,0}},
/* 56735 */ {(14<<2)|2,{83,76,0}},
/* 56736 */ {(14<<2)|2,{83,77,0}},
/* 56737 */ {(14<<2)|2,{83,77,0}},
/* 56738 */ {(14<<2)|2,{83,77,0}},
/* 56739 */ {(14<<2)|2,{83,77,0}},
/* 56740 */ {(14<<2)|2,{83,78,0}},
/* 56741 */ {(14<<2)|2,{83,78,0}},
/* 56742 */ {(14<<2)|2,{83,78,0}},
/* 56743 */ {(14<<2)|2,{83,78,0}},
/* 56744 */ {(14<<2)|2,{83,79,0}},
/* 56745 */ {(14<<2)|2,{83,79,0}},
/* 56746 */ {(14<<2)|2,{83,79,0}},
/* 56747 */ {(14<<2)|2,{83,79,0}},
/* 56748 */ {(14<<2)|2,{83,80,0}},
/* 56749 */ {(14<<2)|2,{83,80,0}},
/* 56750 */ {(14<<2)|2,{83,80,0}},
/* 56751 */ {(14<<2)|2,{83,80,0}},
/* 56752 */ {(14<<2)|2,{83,81,0}},
/* 56753 */ {(14<<2)|2,{83,81,0}},
/* 56754 */ {(14<<2)|2,{83,81,0}},
/* 56755 */ {(14<<2)|2,{83,81,0}},
/* 56756 */ {(14<<2)|2,{83,82,0}},
/* 56757 */ {(14<<2)|2,{83,82,0}},
/* 56758 */ {(14<<2)|2,{83,82,0}},
/* 56759 */ {(14<<2)|2,{83,82,0}},
/* 56760 */ {(14<<2)|2,{83,83,0}},
/* 56761 */ {(14<<2)|2,{83,83,0}},
/* 56762 */ {(14<<2)|2,{83,83,0}},
/* 56763 */ {(14<<2)|2,{83,83,0}},
/* 56764 */ {(14<<2)|2,{83,84,0}},
/* 56765 */ {(14<<2)|2,{83,84,0}},
/* 56766 */ {(14<<2)|2,{83,84,0}},
/* 56767 */ {(14<<2)|2,{83,84,0}},
/* 56768 */ {(14<<2)|2,{83,85,0}},
/* 56769 */ {(14<<2)|2,{83,85,0}},
/* 56770 */ {(14<<2)|2,{83,85,0}},
/* 56771 */ {(14<<2)|2,{83,85,0}},
/* 56772 */ {(14<<2)|2,{83,86,0}},
/* 56773 */ {(14<<2)|2,{83,86,0}},
/* 56774 */ {(14<<2)|2,{83,86,0}},
/* 56775 */ {(14<<2)|2,{83,86,0}},
/* 56776 */ {(14<<2)|2,{83,87,0}},
/* 56777 */ {(14<<2)|2,{83,87,0}},
/* 56778 */ {(14<<2)|2,{83,87,0}},
/* 56779 */ {(14<<2)|2,{83,87,0}},
/* 56780 */ {(14<<2)|2,{83,89,0}},
/* 56781 */ {(14<<2)|2,{83,89,0}},
/* 56782 */ {(14<<2)|2,{83,89,0}},
/* 56783 */ {(14<<2)|2,{83,89,0}},
/* 56784 */ {(14<<2)|2,{83,106,0}},
/* 56785 */ {(14<<2)|2,{83,106,0}},
/* 56786 */ {(14<<2)|2,{83,106,0}},
/* 56787 */ {(14<<2)|2,{83,106,0}},
/* 56788 */ {(14<<2)|2,{83,107,0}},
/* 56789 */ {(14<<2)|2,{83,107,0}},
/* 56790 */ {(14<<2)|2,{83,107,0}},
/* 56791 */ {(14<<2)|2,{83,107,0}},
/* 56792 */ {(14<<2)|2,{83,113,0}},
/* 56793 */ {(14<<2)|2,{83,113,0}},
/* 56794 */ {(14<<2)|2,{83,113,0}},
/* 56795 */ {(14<<2)|2,{83,113,0}},
/* 56796 */ {(14<<2)|2,{83,118,0}},
/* 56797 */ {(14<<2)|2,{83,118,0}},
/* 56798 */ {(14<<2)|2,{83,118,0}},
/* 56799 */ {(14<<2)|2,{83,118,0}},
/* 56800 */ {(14<<2)|2,{83,119,0}},
/* 56801 */ {(14<<2)|2,{83,119,0}},
/* 56802 */ {(14<<2)|2,{83,119,0}},
/* 56803 */ {(14<<2)|2,{83,119,0}},
/* 56804 */ {(14<<2)|2,{83,120,0}},
/* 56805 */ {(14<<2)|2,{83,120,0}},
/* 56806 */ {(14<<2)|2,{83,120,0}},
/* 56807 */ {(14<<2)|2,{83,120,0}},
/* 56808 */ {(14<<2)|2,{83,121,0}},
/* 56809 */ {(14<<2)|2,{83,121,0}},
/* 56810 */ {(14<<2)|2,{83,121,0}},
/* 56811 */ {(14<<2)|2,{83,121,0}},
/* 56812 */ {(14<<2)|2,{83,122,0}},
/* 56813 */ {(14<<2)|2,{83,122,0}},
/* 56814 */ {(14<<2)|2,{83,122,0}},
/* 56815 */ {(14<<2)|2,{83,122,0}},
/* 56816 */ {(15<<2)|2,{83,38,0}},
/* 56817 */ {(15<<2)|2,{83,38,0}},
/* 56818 */ {(15<<2)|2,{83,42,0}},
/* 56819 */ {(15<<2)|2,{83,42,0}},
/* 56820 */ {(15<<2)|2,{83,44,0}},
/* 56821 */ {(15<<2)|2,{83,44,0}},
/* 56822 */ {(15<<2)|2,{83,59,0}},
/* 56823 */ {(15<<2)|2,{83,59,0}},
/* 56824 */ {(15<<2)|2,{83,88,0}},
/* 56825 */ {(15<<2)|2,{83,88,0}},
/* 56826 */ {(15<<2)|2,{83,90,0}},
/* 56827 */ {(15<<2)|2,{83,90,0}},
/* 56828 */ {(7<<2)|1,{83,0,0}},
/* 56829 */ {(7<<2)|1,{83,0,0}},
/* 56830 */ {(7<<2)|1,{83,0,0}},
/* 56831 */ {(7<<2)|1,{83,0,0}},
/* 56832 */ {(12<<2)|2,{84,48,0}},
/* 56833 */ {(12<<2)|2,{84,48,0}},
/* 56834 */ {(12<<2)|2,{84,48,0}},
/* 56835 */ {(12<<2)|2,{84,48,0}},
/* 56836 */ {(12<<2)|2,{84,48,0}},
/* 56837 */ {(12<<2)|2,{84,48,0}},
/* 56838 */ {(12<<2)|2,{84,48,0}},
/* 56839 */ {(12<<2)|2,{84,48,0}},
/* 56840 */ {(12<<2)|2,{84,48,0}},
/* 56841 */ {(12<<2)|2,{84,48,0}},
/* 56842 */ {(12<<2)|2,{84,48,0}},
/* 56843 */ {(12<<2)|2,{84,48,0}},
/* 56844 */ {(12<<2)|2,{84,48,0}},
/* 56845 */ {(12<<2)|2,{84,48,0}},
/* 56846 */ {(12<<2)|2,{84,48,0}},
/* 56847 */ {(12<<2)|2,{84,48,0}},
/* 56848 */ {(12<<2)|2,{84,49,0}},
/* 56849 */ {(12<<2)|2,{84,49,0}},
/* 56850 */ {(12<<2)|2,{84,49,0}},
/* 56851 */ {(12<<2)|2,{84,49,0}},
/* 56852 */ {(12<<2)|2,{84,49,0}},
/* 56853 */ {(12<<2)|2,{84,49,0}},
/* 56854 */ {(12<<2)|2,{84,49,0}},
/* 56855 */ {(12<<2)|2,{84,49,0}},
/* 56856 */ {(12<<2)|2,{84,49,0}},
/* 56857 */ {(12<<2)|2,{84,49,0}},
/* 56858 */ {(12<<2)|2,{84,49,0}},
/* 56859 */ {(12<<2)|2,{84,49,0}},
/* 56860 */ {(12<<2)|2,{84,49,0}},
/* 56861 */ {(12<<2)|2,{84,49,0}},
/* 56862 */ {(12<<2)|2,{84,49,0}},
/* 56863 */ {(12<<2)|2,{84,49,0}},
/* 56864 */ {(12<<2)|2,{84,50,0}},
/* 56865 */ {(12<<2)|2,{84,50,0}},
/* 56866 */ {(12<<2)|2,{84,50,0}},
/* 56867 */ {(12<<2)|2,{84,50,0}},
/* 56868 */ {(12<<2)|2,{84,50,0}},
/* 56869 */ {(12<<2)|2,{84,50,0}},
/* 56870 */ {(12<<2)|2,{84,50,0}},
/* 56871 */ {(12<<2)|2,{84,50,0}},
/* 56872 */ {(12<<2)|2,{84,50,0}},
/* 56873 */ {(12<<2)|2,{84,50,0}},
/* 56874 */ {(12<<2)|2,{84,50,0}},
/* 56875 */ {(12<<2)|2,{84,50,0}},
/* 56876 */ {(12<<2)|2,{84,50,0}},
/* 56877 */ {(12<<2)|2,{84,50,0}},
/* 56878 */ {(12<<2)|2,{84,50,0}},
/* 56879 */ {(12<<2)|2,{84,50,0}},
/* 56880 */ {(12<<2)|2,{84,97,0}},
/* 56881 */ {(12<<2)|2,{84,97,0}},
/* 56882 */ {(12<<2)|2,{84,97,0}},
/* 56883 */ {(12<<2)|2,{84,97,0}},
/* 56884 */ {(12<<2)|2,{84,97,0}},
/* 56885 */ {(12<<2)|2,{84,97,0}},
/* 56886 */ {(12<<2)|2,{84,97,0}},
/* 56887 */ {(12<<2)|2,{84,97,0}},
/* 56888 */ {(12<<2)|2,{84,97,0}},
/* 56889 */ {(12<<2)|2,{84,97,0}},
/* 56890 */ {(12<<2)|2,{84,97,0}},
/* 56891 */ {(12<<2)|2,{84,97,0}},
/* 56892 */ {(12<<2)|2,{84,97,0}},
/* 56893 */ {(12<<2)|2,{84,97,0}},
/* 56894 */ {(12<<2)|2,{84,97,0}},
/* 56895 */ {(12<<2)|2,{84,97,0}},
/* 56896 */ {(12<<2)|2,{84,99,0}},
/* 56897 */ {(12<<2)|2,{84,99,0}},
/* 56898 */ {(12<<2)|2,{84,99,0}},
/* 56899 */ {(12<<2)|2,{84,99,0}},
/* 56900 */ {(12<<2)|2,{84,99,0}},
/* 56901 */ {(12<<2)|2,{84,99,0}},
/* 56902 */ {(12<<2)|2,{84,99,0}},
/* 56903 */ {(12<<2)|2,{84,99,0}},
/* 56904 */ {(12<<2)|2,{84,99,0}},
/* 56905 */ {(12<<2)|2,{84,99,0}},
/* 56906 */ {(12<<2)|2,{84,99,0}},
/* 56907 */ {(12<<2)|2,{84,99,0}},
/* 56908 */ {(12<<2)|2,{84,99,0}},
/* 56909 */ {(12<<2)|2,{84,99,0}},
/* 56910 */ {(12<<2)|2,{84,99,0}},
/* 56911 */ {(12<<2)|2,{84,99,0}},
/* 56912 */ {(12<<2)|2,{84,101,0}},
/* 56913 */ {(12<<2)|2,{84,101,0}},
/* 56914 */ {(12<<2)|2,{84,101,0}},
/* 56915 */ {(12<<2)|2,{84,101,0}},
/* 56916 */ {(12<<2)|2,{84,101,0}},
/* 56917 */ {(12<<2)|2,{84,101,0}},
/* 56918 */ {(12<<2)|2,{84,101,0}},
/* 56919 */ {(12<<2)|2,{84,101,0}},
/* 56920 */ {(12<<2)|2,{84,101,0}},
/* 56921 */ {(12<<2)|2,{84,101,0}},
/* 56922 */ {(12<<2)|2,{84,101,0}},
/* 56923 */ {(12<<2)|2,{84,101,0}},
/* 56924 */ {(12<<2)|2,{84,101,0}},
/* 56925 */ {(12<<2)|2,{84,101,0}},
/* 56926 */ {(12<<2)|2,{84,101,0}},
/* 56927 */ {(12<<2)|2,{84,101,0}},
/* 56928 */ {(12<<2)|2,{84,105,0}},
/* 56929 */ {(12<<2)|2,{84,105,0}},
/* 56930 */ {(12<<2)|2,{84,105,0}},
/* 56931 */ {(12<<2)|2,{84,105,0}},
/* 56932 */ {(12<<2)|2,{84,105,0}},
/* 56933 */ {(12<<2)|2,{84,105,0}},
/* 56934 */ {(12<<2)|2,{84,105,0}},
/* 56935 */ {(12<<2)|2,{84,105,0}},
/* 56936 */ {(12<<2)|2,{84,105,0}},
/* 56937 */ {(12<<2)|2,{84,105,0}},
/* 56938 */ {(12<<2)|2,{84,105,0}},
/* 56939 */ {(12<<2)|2,{84,105,0}},
/* 56940 */ {(12<<2)|2,{84,105,0}},
/* 56941 */ {(12<<2)|2,{84,105,0}},
/* 56942 */ {(12<<2)|2,{84,105,0}},
/* 56943 */ {(12<<2)|2,{84,105,0}},
/* 56944 */ {(12<<2)|2,{84,111,0}},
/* 56945 */ {(12<<2)|2,{84,111,0}},
/* 56946 */ {(12<<2)|2,{84,111,0}},
/* 56947 */ {(12<<2)|2,{84,111,0}},
/* 56948 */ {(12<<2)|2,{84,111,0}},
/* 56949 */ {(12<<2)|2,{84,111,0}},
/* 56950 */ {(12<<2)|2,{84,111,0}},
/* 56951 */ {(12<<2)|2,{84,111,0}},
/* 56952 */ {(12<<2)|2,{84,111,0}},
/* 56953 */ {(12<<2)|2,{84,111,0}},
/* 56954 */ {(12<<2)|2,{84,111,0}},
/* 56955 */ {(12<<2)|2,{84,111,0}},
/* 56956 */ {(12<<2)|2,{84,111,0}},
/* 56957 */ {(12<<2)|2,{84,111,0}},
/* 56958 */ {(12<<2)|2,{84,111,0}},
/* 56959 */ {(12<<2)|2,{84,111,0}},
/* 56960 */ {(12<<2)|2,{84,115,0}},
/* 56961 */ {(12<<2)|2,{84,115,0}},
/* 56962 */ {(12<<2)|2,{84,115,0}},
/* 56963 */ {(12<<2)|2,{84,115,0}},
/* 56964 */ {(12<<2)|2,{84,115,0}},
/* 56965 */ {(12<<2)|2,{84,115,0}},
/* 56966 */ {(12<<2)|2,{84,115,0}},
/* 56967 */ {(12<<2)|2,{84,115,0}},
/* 56968 */ {(12<<2)|2,{84,115,0}},
/* 56969 */ {(12<<2)|2,{84,115,0}},
/* 56970 */ {(12<<2)|2,{84,115,0}},
/* 56971 */ {(12<<2)|2,{84,115,0}},
/* 56972 */ {(12<<2)|2,{84,115,0}},
/* 56973 */ {(12<<2)|2,{84,115,0}},
/* 56974 */ {(12<<2)|2,{84,115,0}},
/* 56975 */ {(12<<2)|2,{84,115,0}},
/* 56976 */ {(12<<2)|2,{84,116,0}},
/* 56977 */ {(12<<2)|2,{84,116,0}},
/* 56978 */ {(12<<2)|2,{84,116,0}},
/* 56979 */ {(12<<2)|2,{84,116,0}},
/* 56980 */ {(12<<2)|2,{84,116,0}},
/* 56981 */ {(12<<2)|2,{84,116,0}},
/* 56982 */ {(12<<2)|2,{84,116,0}},
/* 56983 */ {(12<<2)|2,{84,116,0}},
/* 56984 */ {(12<<2)|2,{84,116,0}},
/* 56985 */ {(12<<2)|2,{84,116,0}},
/* 56986 */ {(12<<2)|2,{84,116,0}},
/* 56987 */ {(12<<2)|2,{84,116,0}},
/* 56988 */ {(12<<2)|2,{84,116,0}},
/* 56989 */ {(12<<2)|2,{84,116,0}},
/* 56990 */ {(12<<2)|2,{84,116,0}},
/* 56991 */ {(12<<2)|2,{84,116,0}},
/* 56992 */ {(13<<2)|2,{84,32,0}},
/* 56993 */ {(13<<2)|2,{84,32,0}},
/* 56994 */ {(13<<2)|2,{84,32,0}},
/* 56995 */ {(13<<2)|2,{84,32,0}},
/* 56996 */ {(13<<2)|2,{84,32,0}},
/* 56997 */ {(13<<2)|2,{84,32,0}},
/* 56998 */ {(13<<2)|2,{84,32,0}},
/* 56999 */ {(13<<2)|2,{84,32,0}},
/* 57000 */ {(13<<2)|2,{84,37,0}},
/* 57001 */ {(13<<2)|2,{84,37,0}},
/* 57002 */ {(13<<2)|2,{84,37,0}},
/* 57003 */ {(13<<2)|2,{84,37,0}},
/* 57004 */ {(13<<2)|2,{84,37,0}},
/* 57005 */ {(13<<2)|2,{84,37,0}},
/* 57006 */ {(13<<2)|2,{84,37,0}},
/* 57007 */ {(13<<2)|2,{84,37,0}},
/* 57008 */ {(13<<2)|2,{84,45,0}},
/* 57009 */ {(13<<2)|2,{84,45,0}},
/* 57010 */ {(13<<2)|2,{84,45,0}},
/* 57011 */ {(13<<2)|2,{84,45,0}},
/* 57012 */ {(13<<2)|2,{84,45,0}},
/* 57013 */ {(13<<2)|2,{84,45,0}},
/* 57014 */ {(13<<2)|2,{84,45,0}},
/* 57015 */ {(13<<2)|2,{84,45,0}},
/* 57016 */ {(13<<2)|2,{84,46,0}},
/* 57017 */ {(13<<2)|2,{84,46,0}},
/* 57018 */ {(13<<2)|2,{84,46,0}},
/* 57019 */ {(13<<2)|2,{84,46,0}},
/* 57020 */ {(13<<2)|2,{84,46,0}},
/* 57021 */ {(13<<2)|2,{84,46,0}},
/* 57022 */ {(13<<2)|2,{84,46,0}},
/* 57023 */ {(13<<2)|2,{84,46,0}},
/* 57024 */ {(13<<2)|2,{84,47,0}},
/* 57025 */ {(13<<2)|2,{84,47,0}},
/* 57026 */ {(13<<2)|2,{84,47,0}},
/* 57027 */ {(13<<2)|2,{84,47,0}},
/* 57028 */ {(13<<2)|2,{84,47,0}},
/* 57029 */ {(13<<2)|2,{84,47,0}},
/* 57030 */ {(13<<2)|2,{84,47,0}},
/* 57031 */ {(13<<2)|2,{84,47,0}},
/* 57032 */ {(13<<2)|2,{84,51,0}},
/* 57033 */ {(13<<2)|2,{84,51,0}},
/* 57034 */ {(13<<2)|2,{84,51,0}},
/* 57035 */ {(13<<2)|2,{84,51,0}},
/* 57036 */ {(13<<2)|2,{84,51,0}},
/* 57037 */ {(13<<2)|2,{84,51,0}},
/* 57038 */ {(13<<2)|2,{84,51,0}},
/* 57039 */ {(13<<2)|2,{84,51,0}},
/* 57040 */ {(13<<2)|2,{84,52,0}},
/* 57041 */ {(13<<2)|2,{84,52,0}},
/* 57042 */ {(13<<2)|2,{84,52,0}},
/* 57043 */ {(13<<2)|2,{84,52,0}},
/* 57044 */ {(13<<2)|2,{84,52,0}},
/* 57045 */ {(13<<2)|2,{84,52,0}},
/* 57046 */ {(13<<2)|2,{84,52,0}},
/* 57047 */ {(13<<2)|2,{84,52,0}},
/* 57048 */ {(13<<2)|2,{84,53,0}},
/* 57049 */ {(13<<2)|2,{84,53,0}},
/* 57050 */ {(13<<2)|2,{84,53,0}},
/* 57051 */ {(13<<2)|2,{84,53,0}},
/* 57052 */ {(13<<2)|2,{84,53,0}},
/* 57053 */ {(13<<2)|2,{84,53,0}},
/* 57054 */ {(13<<2)|2,{84,53,0}},
/* 57055 */ {(13<<2)|2,{84,53,0}},
/* 57056 */ {(13<<2)|2,{84,54,0}},
/* 57057 */ {(13<<2)|2,{84,54,0}},
/* 57058 */ {(13<<2)|2,{84,54,0}},
/* 57059 */ {(13<<2)|2,{84,54,0}},
/* 57060 */ {(13<<2)|2,{84,54,0}},
/* 57061 */ {(13<<2)|2,{84,54,0}},
/* 57062 */ {(13<<2)|2,{84,54,0}},
/* 57063 */ {(13<<2)|2,{84,54,0}},
/* 57064 */ {(13<<2)|2,{84,55,0}},
/* 57065 */ {(13<<2)|2,{84,55,0}},
/* 57066 */ {(13<<2)|2,{84,55,0}},
/* 57067 */ {(13<<2)|2,{84,55,0}},
/* 57068 */ {(13<<2)|2,{84,55,0}},
/* 57069 */ {(13<<2)|2,{84,55,0}},
/* 57070 */ {(13<<2)|2,{84,55,0}},
/* 57071 */ {(13<<2)|2,{84,55,0}},
/* 57072 */ {(13<<2)|2,{84,56,0}},
/* 57073 */ {(13<<2)|2,{84,56,0}},
/* 57074 */ {(13<<2)|2,{84,56,0}},
/* 57075 */ {(13<<2)|2,{84,56,0}},
/* 57076 */ {(13<<2)|2,{84,56,0}},
/* 57077 */ {(13<<2)|2,{84,56,0}},
/* 57078 */ {(13<<2)|2,{84,56,0}},
/* 57079 */ {(13<<2)|2,{84,56,0}},
/* 57080 */ {(13<<2)|2,{84,57,0}},
/* 57081 */ {(13<<2)|2,{84,57,0}},
/* 57082 */ {(13<<2)|2,{84,57,0}},
/* 57083 */ {(13<<2)|2,{84,57,0}},
/* 57084 */ {(13<<2)|2,{84,57,0}},
/* 57085 */ {(13<<2)|2,{84,57,0}},
/* 57086 */ {(13<<2)|2,{84,57,0}},
/* 57087 */ {(13<<2)|2,{84,57,0}},
/* 57088 */ {(13<<2)|2,{84,61,0}},
/* 57089 */ {(13<<2)|2,{84,61,0}},
/* 57090 */ {(13<<2)|2,{84,61,0}},
/* 57091 */ {(13<<2)|2,{84,61,0}},
/* 57092 */ {(13<<2)|2,{84,61,0}},
/* 57093 */ {(13<<2)|2,{84,61,0}},
/* 57094 */ {(13<<2)|2,{84,61,0}},
/* 57095 */ {(13<<2)|2,{84,61,0}},
/* 57096 */ {(13<<2)|2,{84,65,0}},
/* 57097 */ {(13<<2)|2,{84,65,0}},
/* 57098 */ {(13<<2)|2,{84,65,0}},
/* 57099 */ {(13<<2)|2,{84,65,0}},
/* 57100 */ {(13<<2)|2,{84,65,0}},
/* 57101 */ {(13<<2)|2,{84,65,0}},
/* 57102 */ {(13<<2)|2,{84,65,0}},
/* 57103 */ {(13<<2)|2,{84,65,0}},
/* 57104 */ {(13<<2)|2,{84,95,0}},
/* 57105 */ {(13<<2)|2,{84,95,0}},
/* 57106 */ {(13<<2)|2,{84,95,0}},
/* 57107 */ {(13<<2)|2,{84,95,0}},
/* 57108 */ {(13<<2)|2,{84,95,0}},
/* 57109 */ {(13<<2)|2,{84,95,0}},
/* 57110 */ {(13<<2)|2,{84,95,0}},
/* 57111 */ {(13<<2)|2,{84,95,0}},
/* 57112 */ {(13<<2)|2,{84,98,0}},
/* 57113 */ {(13<<2)|2,{84,98,0}},
/* 57114 */ {(13<<2)|2,{84,98,0}},
/* 57115 */ {(13<<2)|2,{84,98,0}},
/* 57116 */ {(13<<2)|2,{84,98,0}},
/* 57117 */ {(13<<2)|2,{84,98,0}},
/* 57118 */ {(13<<2)|2,{84,98,0}},
/* 57119 */ {(13<<2)|2,{84,98,0}},
/* 57120 */ {(13<<2)|2,{84,100,0}},
/* 57121 */ {(13<<2)|2,{84,100,0}},
/* 57122 */ {(13<<2)|2,{84,100,0}},
/* 57123 */ {(13<<2)|2,{84,100,0}},
/* 57124 */ {(13<<2)|2,{84,100,0}},
/* 57125 */ {(13<<2)|2,{84,100,0}},
/* 57126 */ {(13<<2)|2,{84,100,0}},
/* 57127 */ {(13<<2)|2,{84,100,0}},
/* 57128 */ {(13<<2)|2,{84,102,0}},
/* 57129 */ {(13<<2)|2,{84,102,0}},
/* 57130 */ {(13<<2)|2,{84,102,0}},
/* 57131 */ {(13<<2)|2,{84,102,0}},
/* 57132 */ {(13<<2)|2,{84,102,0}},
/* 57133 */ {(13<<2)|2,{84,102,0}},
/* 57134 */ {(13<<2)|2,{84,102,0}},
/* 57135 */ {(13<<2)|2,{84,102,0}},
/* 57136 */ {(13<<2)|2,{84,103,0}},
/* 57137 */ {(13<<2)|2,{84,103,0}},
/* 57138 */ {(13<<2)|2,{84,103,0}},
/* 57139 */ {(13<<2)|2,{84,103,0}},
/* 57140 */ {(13<<2)|2,{84,103,0}},
/* 57141 */ {(13<<2)|2,{84,103,0}},
/* 57142 */ {(13<<2)|2,{84,103,0}},
/* 57143 */ {(13<<2)|2,{84,103,0}},
/* 57144 */ {(13<<2)|2,{84,104,0}},
/* 57145 */ {(13<<2)|2,{84,104,0}},
/* 57146 */ {(13<<2)|2,{84,104,0}},
/* 57147 */ {(13<<2)|2,{84,104,0}},
/* 57148 */ {(13<<2)|2,{84,104,0}},
/* 57149 */ {(13<<2)|2,{84,104,0}},
/* 57150 */ {(13<<2)|2,{84,104,0}},
/* 57151 */ {(13<<2)|2,{84,104,0}},
/* 57152 */ {(13<<2)|2,{84,108,0}},
/* 57153 */ {(13<<2)|2,{84,108,0}},
/* 57154 */ {(13<<2)|2,{84,108,0}},
/* 57155 */ {(13<<2)|2,{84,108,0}},
/* 57156 */ {(13<<2)|2,{84,108,0}},
/* 57157 */ {(13<<2)|2,{84,108,0}},
/* 57158 */ {(13<<2)|2,{84,108,0}},
/* 57159 */ {(13<<2)|2,{84,108,0}},
/* 57160 */ {(13<<2)|2,{84,109,0}},
/* 57161 */ {(13<<2)|2,{84,109,0}},
/* 57162 */ {(13<<2)|2,{84,109,0}},
/* 57163 */ {(13<<2)|2,{84,109,0}},
/* 57164 */ {(13<<2)|2,{84,109,0}},
/* 57165 */ {(13<<2)|2,{84,109,0}},
/* 57166 */ {(13<<2)|2,{84,109,0}},
/* 57167 */ {(13<<2)|2,{84,109,0}},
/* 57168 */ {(13<<2)|2,{84,110,0}},
/* 57169 */ {(13<<2)|2,{84,110,0}},
/* 57170 */ {(13<<2)|2,{84,110,0}},
/* 57171 */ {(13<<2)|2,{84,110,0}},
/* 57172 */ {(13<<2)|2,{84,110,0}},
/* 57173 */ {(13<<2)|2,{84,110,0}},
/* 57174 */ {(13<<2)|2,{84,110,0}},
/* 57175 */ {(13<<2)|2,{84,110,0}},
/* 57176 */ {(13<<2)|2,{84,112,0}},
/* 57177 */ {(13<<2)|2,{84,112,0}},
/* 57178 */ {(13<<2)|2,{84,112,0}},
/* 57179 */ {(13<<2)|2,{84,112,0}},
/* 57180 */ {(13<<2)|2,{84,112,0}},
/* 57181 */ {(13<<2)|2,{84,112,0}},
/* 57182 */ {(13<<2)|2,{84,112,0}},
/* 57183 */ {(13<<2)|2,{84,112,0}},
/* 57184 */ {(13<<2)|2,{84,114,0}},
/* 57185 */ {(13<<2)|2,{84,114,0}},
/* 57186 */ {(13<<2)|2,{84,114,0}},
/* 57187 */ {(13<<2)|2,{84,114,0}},
/* 57188 */ {(13<<2)|2,{84,114,0}},
/* 57189 */ {(13<<2)|2,{84,114,0}},
/* 57190 */ {(13<<2)|2,{84,114,0}},
/* 57191 */ {(13<<2)|2,{84,114,0}},
/* 57192 */ {(13<<2)|2,{84,117,0}},
/* 57193 */ {(13<<2)|2,{84,117,0}},
/* 57194 */ {(13<<2)|2,{84,117,0}},
/* 57195 */ {(13<<2)|2,{84,117,0}},
/* 57196 */ {(13<<2)|2,{84,117,0}},
/* 57197 */ {(13<<2)|2,{84,117,0}},
/* 57198 */ {(13<<2)|2,{84,117,0}},
/* 57199 */ {(13<<2)|2,{84,117,0}},
/* 57200 */ {(14<<2)|2,{84,58,0}},
/* 57201 */ {(14<<2)|2,{84,58,0}},
/* 57202 */ {(14<<2)|2,{84,58,0}},
/* 57203 */ {(14<<2)|2,{84,58,0}},
/* 57204 */ {(14<<2)|2,{84,66,0}},
/* 57205 */ {(14<<2)|2,{84,66,0}},
/* 57206 */ {(14<<2)|2,{84,66,0}},
/* 57207 */ {(14<<2)|2,{84,66,0}},
/* 57208 */ {(14<<2)|2,{84,67,0}},
/* 57209 */ {(14<<2)|2,{84,67,0}},
/* 57210 */ {(14<<2)|2,{84,67,0}},
/* 57211 */ {(14<<2)|2,{84,67,0}},
/* 57212 */ {(14<<2)|2,{84,68,0}},
/* 57213 */ {(14<<2)|2,{84,68,0}},
/* 57214 */ {(14<<2)|2,{84,68,0}},
/* 57215 */ {(14<<2)|2,{84,68,0}},
/* 57216 */ {(14<<2)|2,{84,69,0}},
/* 57217 */ {(14<<2)|2,{84,69,0}},
/* 57218 */ {(14<<2)|2,{84,69,0}},
/* 57219 */ {(14<<2)|2,{84,69,0}},
/* 57220 */ {(14<<2)|2,{84,70,0}},
/* 57221 */ {(14<<2)|2,{84,70,0}},
/* 57222 */ {(14<<2)|2,{84,70,0}},
/* 57223 */ {(14<<2)|2,{84,70,0}},
/* 57224 */ {(14<<2)|2,{84,71,0}},
/* 57225 */ {(14<<2)|2,{84,71,0}},
/* 57226 */ {(14<<2)|2,{84,71,0}},
/* 57227 */ {(14<<2)|2,{84,71,0}},
/* 57228 */ {(14<<2)|2,{84,72,0}},
/* 57229 */ {(14<<2)|2,{84,72,0}},
/* 57230 */ {(14<<2)|2,{84,72,0}},
/* 57231 */ {(14<<2)|2,{84,72,0}},
/* 57232 */ {(14<<2)|2,{84,73,0}},
/* 57233 */ {(14<<2)|2,{84,73,0}},
/* 57234 */ {(14<<2)|2,{84,73,0}},
/* 57235 */ {(14<<2)|2,{84,73,0}},
/* 57236 */ {(14<<2)|2,{84,74,0}},
/* 57237 */ {(14<<2)|2,{84,74,0}},
/* 57238 */ {(14<<2)|2,{84,74,0}},
/* 57239 */ {(14<<2)|2,{84,74,0}},
/* 57240 */ {(14<<2)|2,{84,75,0}},
/* 57241 */ {(14<<2)|2,{84,75,0}},
/* 57242 */ {(14<<2)|2,{84,75,0}},
/* 57243 */ {(14<<2)|2,{84,75,0}},
/* 57244 */ {(14<<2)|2,{84,76,0}},
/* 57245 */ {(14<<2)|2,{84,76,0}},
/* 57246 */ {(14<<2)|2,{84,76,0}},
/* 57247 */ {(14<<2)|2,{84,76,0}},
/* 57248 */ {(14<<2)|2,{84,77,0}},
/* 57249 */ {(14<<2)|2,{84,77,0}},
/* 57250 */ {(14<<2)|2,{84,77,0}},
/* 57251 */ {(14<<2)|2,{84,77,0}},
/* 57252 */ {(14<<2)|2,{84,78,0}},
/* 57253 */ {(14<<2)|2,{84,78,0}},
/* 57254 */ {(14<<2)|2,{84,78,0}},
/* 57255 */ {(14<<2)|2,{84,78,0}},
/* 57256 */ {(14<<2)|2,{84,79,0}},
/* 57257 */ {(14<<2)|2,{84,79,0}},
/* 57258 */ {(14<<2)|2,{84,79,0}},
/* 57259 */ {(14<<2)|2,{84,79,0}},
/* 57260 */ {(14<<2)|2,{84,80,0}},
/* 57261 */ {(14<<2)|2,{84,80,0}},
/* 57262 */ {(14<<2)|2,{84,80,0}},
/* 57263 */ {(14<<2)|2,{84,80,0}},
/* 57264 */ {(14<<2)|2,{84,81,0}},
/* 57265 */ {(14<<2)|2,{84,81,0}},
/* 57266 */ {(14<<2)|2,{84,81,0}},
/* 57267 */ {(14<<2)|2,{84,81,0}},
/* 57268 */ {(14<<2)|2,{84,82,0}},
/* 57269 */ {(14<<2)|2,{84,82,0}},
/* 57270 */ {(14<<2)|2,{84,82,0}},
/* 57271 */ {(14<<2)|2,{84,82,0}},
/* 57272 */ {(14<<2)|2,{84,83,0}},
/* 57273 */ {(14<<2)|2,{84,83,0}},
/* 57274 */ {(14<<2)|2,{84,83,0}},
/* 57275 */ {(14<<2)|2,{84,83,0}},
/* 57276 */ {(14<<2)|2,{84,84,0}},
/* 57277 */ {(14<<2)|2,{84,84,0}},
/* 57278 */ {(14<<2)|2,{84,84,0}},
/* 57279 */ {(14<<2)|2,{84,84,0}},
/* 57280 */ {(14<<2)|2,{84,85,0}},
/* 57281 */ {(14<<2)|2,{84,85,0}},
/* 57282 */ {(14<<2)|2,{84,85,0}},
/* 57283 */ {(14<<2)|2,{84,85,0}},
/* 57284 */ {(14<<2)|2,{84,86,0}},
/* 57285 */ {(14<<2)|2,{84,86,0}},
/* 57286 */ {(14<<2)|2,{84,86,0}},
/* 57287 */ {(14<<2)|2,{84,86,0}},
/* 57288 */ {(14<<2)|2,{84,87,0}},
/* 57289 */ {(14<<2)|2,{84,87,0}},
/* 57290 */ {(14<<2)|2,{84,87,0}},
/* 57291 */ {(14<<2)|2,{84,87,0}},
/* 57292 */ {(14<<2)|2,{84,89,0}},
/* 57293 */ {(14<<2)|2,{84,89,0}},
/* 57294 */ {(14<<2)|2,{84,89,0}},
/* 57295 */ {(14<<2)|2,{84,89,0}},
/* 57296 */ {(14<<2)|2,{84,106,0}},
/* 57297 */ {(14<<2)|2,{84,106,0}},
/* 57298 */ {(14<<2)|2,{84,106,0}},
/* 57299 */ {(14<<2)|2,{84,106,0}},
/* 57300 */ {(14<<2)|2,{84,107,0}},
/* 57301 */ {(14<<2)|2,{84,107,0}},
/* 57302 */ {(14<<2)|2,{84,107,0}},
/* 57303 */ {(14<<2)|2,{84,107,0}},
/* 57304 */ {(14<<2)|2,{84,113,0}},
/* 57305 */ {(14<<2)|2,{84,113,0}},
/* 57306 */ {(14<<2)|2,{84,113,0}},
/* 57307 */ {(14<<2)|2,{84,113,0}},
/* 57308 */ {(14<<2)|2,{84,118,0}},
/* 57309 */ {(14<<2)|2,{84,118,0}},
/* 57310 */ {(14<<2)|2,{84,118,0}},
/* 57311 */ {(14<<2)|2,{84,118,0}},
/* 57312 */ {(14<<2)|2,{84,119,0}},
/* 57313 */ {(14<<2)|2,{84,119,0}},
/* 57314 */ {(14<<2)|2,{84,119,0}},
/* 57315 */ {(14<<2)|2,{84,119,0}},
/* 57316 */ {(14<<2)|2,{84,120,0}},
/* 57317 */ {(14<<2)|2,{84,120,0}},
/* 57318 */ {(14<<2)|2,{84,120,0}},
/* 57319 */ {(14<<2)|2,{84,120,0}},
/* 57320 */ {(14<<2)|2,{84,121,0}},
/* 57321 */ {(14<<2)|2,{84,121,0}},
/* 57322 */ {(14<<2)|2,{84,121,0}},
/* 57323 */ {(14<<2)|2,{84,121,0}},
/* 57324 */ {(14<<2)|2,{84,122,0}},
/* 57325 */ {(14<<2)|2,{84,122,0}},
/* 57326 */ {(14<<2)|2,{84,122,0}},
/* 57327 */ {(14<<2)|2,{84,122,0}},
/* 57328 */ {(15<<2)|2,{84,38,0}},
/* 57329 */ {(15<<2)|2,{84,38,0}},
/* 57330 */ {(15<<2)|2,{84,42,0}},
/* 57331 */ {(15<<2)|2,{84,42,0}},
/* 57332 */ {(15<<2)|2,{84,44,0}},
/* 57333 */ {(15<<2)|2,{84,44,0}},
/* 57334 */ {(15<<2)|2,{84,59,0}},
/* 57335 */ {(15<<2)|2,{84,59,0}},
/* 57336 */ {(15<<2)|2,{84,88,0}},
/* 57337 */ {(15<<2)|2,{84,88,0}},
/* 57338 */ {(15<<2)|2,{84,90,0}},
/* 57339 */ {(15<<2)|2,{84,90,0}},
/* 57340 */ {(7<<2)|1,{84,0,0}},
/* 57341 */ {(7<<2)|1,{84,0,0}},
/* 57342 */ {(7<<2)|1,{84,0,0}},
/* 57343 */ {(7<<2)|1,{84,0,0}},
/* 57344 */ {(12<<2)|2,{85,48,0}},
/* 57345 */ {(12<<2)|2,{85,48,0}},
/* 57346 */ {(12<<2)|2,{85,48,0}},
/* 57347 */ {(12<<2)|2,{85,48,0}},
/* 57348 */ {(12<<2)|2,{85,48,0}},
/* 57349 */ {(12<<2)|2,{85,48,0}},
/* 57350 */ {(12<<2)|2,{85,48,0}},
/* 57351 */ {(12<<2)|2,{85,48,0}},
/* 57352 */ {(12<<2)|2,{85,48,0}},
/* 57353 */ {(12<<2)|2,{85,48,0}},
/* 57354 */ {(12<<2)|2,{85,48,0}},
/* 57355 */ {(12<<2)|2,{85,48,0}},
/* 57356 */ {(12<<2)|2,{85,48,0}},
/* 57357 */ {(12<<2)|2,{85,48,0}},
/* 57358 */ {(12<<2)|2,{85,48,0}},
/* 57359 */ {(12<<2)|2,{85,48,0}},
/* 57360 */ {(12<<2)|2,{85,49,0}},
/* 57361 */ {(12<<2)|2,{85,49,0}},
/* 57362 */ {(12<<2)|2,{85,49,0}},
/* 57363 */ {(12<<2)|2,{85,49,0}},
/* 57364 */ {(12<<2)|2,{85,49,0}},
/* 57365 */ {(12<<2)|2,{85,49,0}},
/* 57366 */ {(12<<2)|2,{85,49,0}},
/* 57367 */ {(12<<2)|2,{85,49,0}},
/* 57368 */ {(12<<2)|2,{85,49,0}},
/* 57369 */ {(12<<2)|2,{85,49,0}},
/* 57370 */ {(12<<2)|2,{85,49,0}},
/* 57371 */ {(12<<2)|2,{85,49,0}},
/* 57372 */ {(12<<2)|2,{85,49,0}},
/* 57373 */ {(12<<2)|2,{85,49,0}},
/* 57374 */ {(12<<2)|2,{85,49,0}},
/* 57375 */ {(12<<2)|2,{85,49,0}},
/* 57376 */ {(12<<2)|2,{85,50,0}},
/* 57377 */ {(12<<2)|2,{85,50,0}},
/* 57378 */ {(12<<2)|2,{85,50,0}},
/* 57379 */ {(12<<2)|2,{85,50,0}},
/* 57380 */ {(12<<2)|2,{85,50,0}},
/* 57381 */ {(12<<2)|2,{85,50,0}},
/* 57382 */ {(12<<2)|2,{85,50,0}},
/* 57383 */ {(12<<2)|2,{85,50,0}},
/* 57384 */ {(12<<2)|2,{85,50,0}},
/* 57385 */ {(12<<2)|2,{85,50,0}},
/* 57386 */ {(12<<2)|2,{85,50,0}},
/* 57387 */ {(12<<2)|2,{85,50,0}},
/* 57388 */ {(12<<2)|2,{85,50,0}},
/* 57389 */ {(12<<2)|2,{85,50,0}},
/* 57390 */ {(12<<2)|2,{85,50,0}},
/* 57391 */ {(12<<2)|2,{85,50,0}},
/* 57392 */ {(12<<2)|2,{85,97,0}},
/* 57393 */ {(12<<2)|2,{85,97,0}},
/* 57394 */ {(12<<2)|2,{85,97,0}},
/* 57395 */ {(12<<2)|2,{85,97,0}},
/* 57396 */ {(12<<2)|2,{85,97,0}},
/* 57397 */ {(12<<2)|2,{85,97,0}},
/* 57398 */ {(12<<2)|2,{85,97,0}},
/* 57399 */ {(12<<2)|2,{85,97,0}},
/* 57400 */ {(12<<2)|2,{85,97,0}},
/* 57401 */ {(12<<2)|2,{85,97,0}},
/* 57402 */ {(12<<2)|2,{85,97,0}},
/* 57403 */ {(12<<2)|2,{85,97,0}},
/* 57404 */ {(12<<2)|2,{85,97,0}},
/* 57405 */ {(12<<2)|2,{85,97,0}},
/* 57406 */ {(12<<2)|2,{85,97,0}},
/* 57407 */ {(12<<2)|2,{85,97,0}},
/* 57408 */ {(12<<2)|2,{85,99,0}},
/* 57409 */ {(12<<2)|2,{85,99,0}},
/* 57410 */ {(12<<2)|2,{85,99,0}},
/* 57411 */ {(12<<2)|2,{85,99,0}},
/* 57412 */ {(12<<2)|2,{85,99,0}},
/* 57413 */ {(12<<2)|2,{85,99,0}},
/* 57414 */ {(12<<2)|2,{85,99,0}},
/* 57415 */ {(12<<2)|2,{85,99,0}},
/* 57416 */ {(12<<2)|2,{85,99,0}},
/* 57417 */ {(12<<2)|2,{85,99,0}},
/* 57418 */ {(12<<2)|2,{85,99,0}},
/* 57419 */ {(12<<2)|2,{85,99,0}},
/* 57420 */ {(12<<2)|2,{85,99,0}},
/* 57421 */ {(12<<2)|2,{85,99,0}},
/* 57422 */ {(12<<2)|2,{85,99,0}},
/* 57423 */ {(12<<2)|2,{85,99,0}},
/* 57424 */ {(12<<2)|2,{85,101,0}},
/* 57425 */ {(12<<2)|2,{85,101,0}},
/* 57426 */ {(12<<2)|2,{85,101,0}},
/* 57427 */ {(12<<2)|2,{85,101,0}},
/* 57428 */ {(12<<2)|2,{85,101,0}},
/* 57429 */ {(12<<2)|2,{85,101,0}},
/* 57430 */ {(12<<2)|2,{85,101,0}},
/* 57431 */ {(12<<2)|2,{85,101,0}},
/* 57432 */ {(12<<2)|2,{85,101,0}},
/* 57433 */ {(12<<2)|2,{85,101,0}},
/* 57434 */ {(12<<2)|2,{85,101,0}},
/* 57435 */ {(12<<2)|2,{85,101,0}},
/* 57436 */ {(12<<2)|2,{85,101,0}},
/* 57437 */ {(12<<2)|2,{85,101,0}},
/* 57438 */ {(12<<2)|2,{85,101,0}},
/* 57439 */ {(12<<2)|2,{85,101,0}},
/* 57440 */ {(12<<2)|2,{85,105,0}},
/* 57441 */ {(12<<2)|2,{85,105,0}},
/* 57442 */ {(12<<2)|2,{85,105,0}},
/* 57443 */ {(12<<2)|2,{85,105,0}},
/* 57444 */ {(12<<2)|2,{85,105,0}},
/* 57445 */ {(12<<2)|2,{85,105,0}},
/* 57446 */ {(12<<2)|2,{85,105,0}},
/* 57447 */ {(12<<2)|2,{85,105,0}},
/* 57448 */ {(12<<2)|2,{85,105,0}},
/* 57449 */ {(12<<2)|2,{85,105,0}},
/* 57450 */ {(12<<2)|2,{85,105,0}},
/* 57451 */ {(12<<2)|2,{85,105,0}},
/* 57452 */ {(12<<2)|2,{85,105,0}},
/* 57453 */ {(12<<2)|2,{85,105,0}},
/* 57454 */ {(12<<2)|2,{85,105,0}},
/* 57455 */ {(12<<2)|2,{85,105,0}},
/* 57456 */ {(12<<2)|2,{85,111,0}},
/* 57457 */ {(12<<2)|2,{85,111,0}},
/* 57458 */ {(12<<2)|2,{85,111,0}},
/* 57459 */ {(12<<2)|2,{85,111,0}},
/* 57460 */ {(12<<2)|2,{85,111,0}},
/* 57461 */ {(12<<2)|2,{85,111,0}},
/* 57462 */ {(12<<2)|2,{85,111,0}},
/* 57463 */ {(12<<2)|2,{85,111,0}},
/* 57464 */ {(12<<2)|2,{85,111,0}},
/* 57465 */ {(12<<2)|2,{85,111,0}},
/* 57466 */ {(12<<2)|2,{85,111,0}},
/* 57467 */ {(12<<2)|2,{85,111,0}},
/* 57468 */ {(12<<2)|2,{85,111,0}},
/* 57469 */ {(12<<2)|2,{85,111,0}},
/* 57470 */ {(12<<2)|2,{85,111,0}},
/* 57471 */ {(12<<2)|2,{85,111,0}},
/* 57472 */ {(12<<2)|2,{85,115,0}},
/* 57473 */ {(12<<2)|2,{85,115,0}},
/* 57474 */ {(12<<2)|2,{85,115,0}},
/* 57475 */ {(12<<2)|2,{85,115,0}},
/* 57476 */ {(12<<2)|2,{85,115,0}},
/* 57477 */ {(12<<2)|2,{85,115,0}},
/* 57478 */ {(12<<2)|2,{85,115,0}},
/* 57479 */ {(12<<2)|2,{85,115,0}},
/* 57480 */ {(12<<2)|2,{85,115,0}},
/* 57481 */ {(12<<2)|2,{85,115,0}},
/* 57482 */ {(12<<2)|2,{85,115,0}},
/* 57483 */ {(12<<2)|2,{85,115,0}},
/* 57484 */ {(12<<2)|2,{85,115,0}},
/* 57485 */ {(12<<2)|2,{85,115,0}},
/* 57486 */ {(12<<2)|2,{85,115,0}},
/* 57487 */ {(12<<2)|2,{85,115,0}},
/* 57488 */ {(12<<2)|2,{85,116,0}},
/* 57489 */ {(12<<2)|2,{85,116,0}},
/* 57490 */ {(12<<2)|2,{85,116,0}},
/* 57491 */ {(12<<2)|2,{85,116,0}},
/* 57492 */ {(12<<2)|2,{85,116,0}},
/* 57493 */ {(12<<2)|2,{85,116,0}},
/* 57494 */ {(12<<2)|2,{85,116,0}},
/* 57495 */ {(12<<2)|2,{85,116,0}},
/* 57496 */ {(12<<2)|2,{85,116,0}},
/* 57497 */ {(12<<2)|2,{85,116,0}},
/* 57498 */ {(12<<2)|2,{85,116,0}},
/* 57499 */ {(12<<2)|2,{85,116,0}},
/* 57500 */ {(12<<2)|2,{85,116,0}},
/* 57501 */ {(12<<2)|2,{85,116,0}},
/* 57502 */ {(12<<2)|2,{85,116,0}},
/* 57503 */ {(12<<2)|2,{85,116,0}},
/* 57504 */ {(13<<2)|2,{85,32,0}},
/* 57505 */ {(13<<2)|2,{85,32,0}},
/* 57506 */ {(13<<2)|2,{85,32,0}},
/* 57507 */ {(13<<2)|2,{85,32,0}},
/* 57508 */ {(13<<2)|2,{85,32,0}},
/* 57509 */ {(13<<2)|2,{85,32,0}},
/* 57510 */ {(13<<2)|2,{85,32,0}},
/* 57511 */ {(13<<2)|2,{85,32,0}},
/* 57512 */ {(13<<2)|2,{85,37,0}},
/* 57513 */ {(13<<2)|2,{85,37,0}},
/* 57514 */ {(13<<2)|2,{85,37,0}},
/* 57515 */ {(13<<2)|2,{85,37,0}},
/* 57516 */ {(13<<2)|2,{85,37,0}},
/* 57517 */ {(13<<2)|2,{85,37,0}},
/* 57518 */ {(13<<2)|2,{85,37,0}},
/* 57519 */ {(13<<2)|2,{85,37,0}},
/* 57520 */ {(13<<2)|2,{85,45,0}},
/* 57521 */ {(13<<2)|2,{85,45,0}},
/* 57522 */ {(13<<2)|2,{85,45,0}},
/* 57523 */ {(13<<2)|2,{85,45,0}},
/* 57524 */ {(13<<2)|2,{85,45,0}},
/* 57525 */ {(13<<2)|2,{85,45,0}},
/* 57526 */ {(13<<2)|2,{85,45,0}},
/* 57527 */ {(13<<2)|2,{85,45,0}},
/* 57528 */ {(13<<2)|2,{85,46,0}},
/* 57529 */ {(13<<2)|2,{85,46,0}},
/* 57530 */ {(13<<2)|2,{85,46,0}},
/* 57531 */ {(13<<2)|2,{85,46,0}},
/* 57532 */ {(13<<2)|2,{85,46,0}},
/* 57533 */ {(13<<2)|2,{85,46,0}},
/* 57534 */ {(13<<2)|2,{85,46,0}},
/* 57535 */ {(13<<2)|2,{85,46,0}},
/* 57536 */ {(13<<2)|2,{85,47,0}},
/* 57537 */ {(13<<2)|2,{85,47,0}},
/* 57538 */ {(13<<2)|2,{85,47,0}},
/* 57539 */ {(13<<2)|2,{85,47,0}},
/* 57540 */ {(13<<2)|2,{85,47,0}},
/* 57541 */ {(13<<2)|2,{85,47,0}},
/* 57542 */ {(13<<2)|2,{85,47,0}},
/* 57543 */ {(13<<2)|2,{85,47,0}},
/* 57544 */ {(13<<2)|2,{85,51,0}},
/* 57545 */ {(13<<2)|2,{85,51,0}},
/* 57546 */ {(13<<2)|2,{85,51,0}},
/* 57547 */ {(13<<2)|2,{85,51,0}},
/* 57548 */ {(13<<2)|2,{85,51,0}},
/* 57549 */ {(13<<2)|2,{85,51,0}},
/* 57550 */ {(13<<2)|2,{85,51,0}},
/* 57551 */ {(13<<2)|2,{85,51,0}},
/* 57552 */ {(13<<2)|2,{85,52,0}},
/* 57553 */ {(13<<2)|2,{85,52,0}},
/* 57554 */ {(13<<2)|2,{85,52,0}},
/* 57555 */ {(13<<2)|2,{85,52,0}},
/* 57556 */ {(13<<2)|2,{85,52,0}},
/* 57557 */ {(13<<2)|2,{85,52,0}},
/* 57558 */ {(13<<2)|2,{85,52,0}},
/* 57559 */ {(13<<2)|2,{85,52,0}},
/* 57560 */ {(13<<2)|2,{85,53,0}},
/* 57561 */ {(13<<2)|2,{85,53,0}},
/* 57562 */ {(13<<2)|2,{85,53,0}},
/* 57563 */ {(13<<2)|2,{85,53,0}},
/* 57564 */ {(13<<2)|2,{85,53,0}},
/* 57565 */ {(13<<2)|2,{85,53,0}},
/* 57566 */ {(13<<2)|2,{85,53,0}},
/* 57567 */ {(13<<2)|2,{85,53,0}},
/* 57568 */ {(13<<2)|2,{85,54,0}},
/* 57569 */ {(13<<2)|2,{85,54,0}},
/* 57570 */ {(13<<2)|2,{85,54,0}},
/* 57571 */ {(13<<2)|2,{85,54,0}},
/* 57572 */ {(13<<2)|2,{85,54,0}},
/* 57573 */ {(13<<2)|2,{85,54,0}},
/* 57574 */ {(13<<2)|2,{85,54,0}},
/* 57575 */ {(13<<2)|2,{85,54,0}},
/* 57576 */ {(13<<2)|2,{85,55,0}},
/* 57577 */ {(13<<2)|2,{85,55,0}},
/* 57578 */ {(13<<2)|2,{85,55,0}},
/* 57579 */ {(13<<2)|2,{85,55,0}},
/* 57580 */ {(13<<2)|2,{85,55,0}},
/* 57581 */ {(13<<2)|2,{85,55,0}},
/* 57582 */ {(13<<2)|2,{85,55,0}},
/* 57583 */ {(13<<2)|2,{85,55,0}},
/* 57584 */ {(13<<2)|2,{85,56,0}},
/* 57585 */ {(13<<2)|2,{85,56,0}},
/* 57586 */ {(13<<2)|2,{85,56,0}},
/* 57587 */ {(13<<2)|2,{85,56,0}},
/* 57588 */ {(13<<2)|2,{85,56,0}},
/* 57589 */ {(13<<2)|2,{85,56,0}},
/* 57590 */ {(13<<2)|2,{85,56,0}},
/* 57591 */ {(13<<2)|2,{85,56,0}},
/* 57592 */ {(13<<2)|2,{85,57,0}},
/* 57593 */ {(13<<2)|2,{85,57,0}},
/* 57594 */ {(13<<2)|2,{85,57,0}},
/* 57595 */ {(13<<2)|2,{85,57,0}},
/* 57596 */ {(13<<2)|2,{85,57,0}},
/* 57597 */ {(13<<2)|2,{85,57,0}},
/* 57598 */ {(13<<2)|2,{85,57,0}},
/* 57599 */ {(13<<2)|2,{85,57,0}},
/* 57600 */ {(13<<2)|2,{85,61,0}},
/* 57601 */ {(13<<2)|2,{85,61,0}},
/* 57602 */ {(13<<2)|2,{85,61,0}},
/* 57603 */ {(13<<2)|2,{85,61,0}},
/* 57604 */ {(13<<2)|2,{85,61,0}},
/* 57605 */ {(13<<2)|2,{85,61,0}},
/* 57606 */ {(13<<2)|2,{85,61,0}},
/* 57607 */ {(13<<2)|2,{85,61,0}},
/* 57608 */ {(13<<2)|2,{85,65,0}},
/* 57609 */ {(13<<2)|2,{85,65,0}},
/* 57610 */ {(13<<2)|2,{85,65,0}},
/* 57611 */ {(13<<2)|2,{85,65,0}},
/* 57612 */ {(13<<2)|2,{85,65,0}},
/* 57613 */ {(13<<2)|2,{85,65,0}},
/* 57614 */ {(13<<2)|2,{85,65,0}},
/* 57615 */ {(13<<2)|2,{85,65,0}},
/* 57616 */ {(13<<2)|2,{85,95,0}},
/* 57617 */ {(13<<2)|2,{85,95,0}},
/* 57618 */ {(13<<2)|2,{85,95,0}},
/* 57619 */ {(13<<2)|2,{85,95,0}},
/* 57620 */ {(13<<2)|2,{85,95,0}},
/* 57621 */ {(13<<2)|2,{85,95,0}},
/* 57622 */ {(13<<2)|2,{85,95,0}},
/* 57623 */ {(13<<2)|2,{85,95,0}},
/* 57624 */ {(13<<2)|2,{85,98,0}},
/* 57625 */ {(13<<2)|2,{85,98,0}},
/* 57626 */ {(13<<2)|2,{85,98,0}},
/* 57627 */ {(13<<2)|2,{85,98,0}},
/* 57628 */ {(13<<2)|2,{85,98,0}},
/* 57629 */ {(13<<2)|2,{85,98,0}},
/* 57630 */ {(13<<2)|2,{85,98,0}},
/* 57631 */ {(13<<2)|2,{85,98,0}},
/* 57632 */ {(13<<2)|2,{85,100,0}},
/* 57633 */ {(13<<2)|2,{85,100,0}},
/* 57634 */ {(13<<2)|2,{85,100,0}},
/* 57635 */ {(13<<2)|2,{85,100,0}},
/* 57636 */ {(13<<2)|2,{85,100,0}},
/* 57637 */ {(13<<2)|2,{85,100,0}},
/* 57638 */ {(13<<2)|2,{85,100,0}},
/* 57639 */ {(13<<2)|2,{85,100,0}},
/* 57640 */ {(13<<2)|2,{85,102,0}},
/* 57641 */ {(13<<2)|2,{85,102,0}},
/* 57642 */ {(13<<2)|2,{85,102,0}},
/* 57643 */ {(13<<2)|2,{85,102,0}},
/* 57644 */ {(13<<2)|2,{85,102,0}},
/* 57645 */ {(13<<2)|2,{85,102,0}},
/* 57646 */ {(13<<2)|2,{85,102,0}},
/* 57647 */ {(13<<2)|2,{85,102,0}},
/* 57648 */ {(13<<2)|2,{85,103,0}},
/* 57649 */ {(13<<2)|2,{85,103,0}},
/* 57650 */ {(13<<2)|2,{85,103,0}},
/* 57651 */ {(13<<2)|2,{85,103,0}},
/* 57652 */ {(13<<2)|2,{85,103,0}},
/* 57653 */ {(13<<2)|2,{85,103,0}},
/* 57654 */ {(13<<2)|2,{85,103,0}},
/* 57655 */ {(13<<2)|2,{85,103,0}},
/* 57656 */ {(13<<2)|2,{85,104,0}},
/* 57657 */ {(13<<2)|2,{85,104,0}},
/* 57658 */ {(13<<2)|2,{85,104,0}},
/* 57659 */ {(13<<2)|2,{85,104,0}},
/* 57660 */ {(13<<2)|2,{85,104,0}},
/* 57661 */ {(13<<2)|2,{85,104,0}},
/* 57662 */ {(13<<2)|2,{85,104,0}},
/* 57663 */ {(13<<2)|2,{85,104,0}},
/* 57664 */ {(13<<2)|2,{85,108,0}},
/* 57665 */ {(13<<2)|2,{85,108,0}},
/* 57666 */ {(13<<2)|2,{85,108,0}},
/* 57667 */ {(13<<2)|2,{85,108,0}},
/* 57668 */ {(13<<2)|2,{85,108,0}},
/* 57669 */ {(13<<2)|2,{85,108,0}},
/* 57670 */ {(13<<2)|2,{85,108,0}},
/* 57671 */ {(13<<2)|2,{85,108,0}},
/* 57672 */ {(13<<2)|2,{85,109,0}},
/* 57673 */ {(13<<2)|2,{85,109,0}},
/* 57674 */ {(13<<2)|2,{85,109,0}},
/* 57675 */ {(13<<2)|2,{85,109,0}},
/* 57676 */ {(13<<2)|2,{85,109,0}},
/* 57677 */ {(13<<2)|2,{85,109,0}},
/* 57678 */ {(13<<2)|2,{85,109,0}},
/* 57679 */ {(13<<2)|2,{85,109,0}},
/* 57680 */ {(13<<2)|2,{85,110,0}},
/* 57681 */ {(13<<2)|2,{85,110,0}},
/* 57682 */ {(13<<2)|2,{85,110,0}},
/* 57683 */ {(13<<2)|2,{85,110,0}},
/* 57684 */ {(13<<2)|2,{85,110,0}},
/* 57685 */ {(13<<2)|2,{85,110,0}},
/* 57686 */ {(13<<2)|2,{85,110,0}},
/* 57687 */ {(13<<2)|2,{85,110,0}},
/* 57688 */ {(13<<2)|2,{85,112,0}},
/* 57689 */ {(13<<2)|2,{85,112,0}},
/* 57690 */ {(13<<2)|2,{85,112,0}},
/* 57691 */ {(13<<2)|2,{85,112,0}},
/* 57692 */ {(13<<2)|2,{85,112,0}},
/* 57693 */ {(13<<2)|2,{85,112,0}},
/* 57694 */ {(13<<2)|2,{85,112,0}},
/* 57695 */ {(13<<2)|2,{85,112,0}},
/* 57696 */ {(13<<2)|2,{85,114,0}},
/* 57697 */ {(13<<2)|2,{85,114,0}},
/* 57698 */ {(13<<2)|2,{85,114,0}},
/* 57699 */ {(13<<2)|2,{85,114,0}},
/* 57700 */ {(13<<2)|2,{85,114,0}},
/* 57701 */ {(13<<2)|2,{85,114,0}},
/* 57702 */ {(13<<2)|2,{85,114,0}},
/* 57703 */ {(13<<2)|2,{85,114,0}},
/* 57704 */ {(13<<2)|2,{85,117,0}},
/* 57705 */ {(13<<2)|2,{85,117,0}},
/* 57706 */ {(13<<2)|2,{85,117,0}},
/* 57707 */ {(13<<2)|2,{85,117,0}},
/* 57708 */ {(13<<2)|2,{85,117,0}},
/* 57709 */ {(13<<2)|2,{85,117,0}},
/* 57710 */ {(13<<2)|2,{85,117,0}},
/* 57711 */ {(13<<2)|2,{85,117,0}},
/* 57712 */ {(14<<2)|2,{85,58,0}},
/* 57713 */ {(14<<2)|2,{85,58,0}},
/* 57714 */ {(14<<2)|2,{85,58,0}},
/* 57715 */ {(14<<2)|2,{85,58,0}},
/* 57716 */ {(14<<2)|2,{85,66,0}},
/* 57717 */ {(14<<2)|2,{85,66,0}},
/* 57718 */ {(14<<2)|2,{85,66,0}},
/* 57719 */ {(14<<2)|2,{85,66,0}},
/* 57720 */ {(14<<2)|2,{85,67,0}},
/* 57721 */ {(14<<2)|2,{85,67,0}},
/* 57722 */ {(14<<2)|2,{85,67,0}},
/* 57723 */ {(14<<2)|2,{85,67,0}},
/* 57724 */ {(14<<2)|2,{85,68,0}},
/* 57725 */ {(14<<2)|2,{85,68,0}},
/* 57726 */ {(14<<2)|2,{85,68,0}},
/* 57727 */ {(14<<2)|2,{85,68,0}},
/* 57728 */ {(14<<2)|2,{85,69,0}},
/* 57729 */ {(14<<2)|2,{85,69,0}},
/* 57730 */ {(14<<2)|2,{85,69,0}},
/* 57731 */ {(14<<2)|2,{85,69,0}},
/* 57732 */ {(14<<2)|2,{85,70,0}},
/* 57733 */ {(14<<2)|2,{85,70,0}},
/* 57734 */ {(14<<2)|2,{85,70,0}},
/* 57735 */ {(14<<2)|2,{85,70,0}},
/* 57736 */ {(14<<2)|2,{85,71,0}},
/* 57737 */ {(14<<2)|2,{85,71,0}},
/* 57738 */ {(14<<2)|2,{85,71,0}},
/* 57739 */ {(14<<2)|2,{85,71,0}},
/* 57740 */ {(14<<2)|2,{85,72,0}},
/* 57741 */ {(14<<2)|2,{85,72,0}},
/* 57742 */ {(14<<2)|2,{85,72,0}},
/* 57743 */ {(14<<2)|2,{85,72,0}},
/* 57744 */ {(14<<2)|2,{85,73,0}},
/* 57745 */ {(14<<2)|2,{85,73,0}},
/* 57746 */ {(14<<2)|2,{85,73,0}},
/* 57747 */ {(14<<2)|2,{85,73,0}},
/* 57748 */ {(14<<2)|2,{85,74,0}},
/* 57749 */ {(14<<2)|2,{85,74,0}},
/* 57750 */ {(14<<2)|2,{85,74,0}},
/* 57751 */ {(14<<2)|2,{85,74,0}},
/* 57752 */ {(14<<2)|2,{85,75,0}},
/* 57753 */ {(14<<2)|2,{85,75,0}},
/* 57754 */ {(14<<2)|2,{85,75,0}},
/* 57755 */ {(14<<2)|2,{85,75,0}},
/* 57756 */ {(14<<2)|2,{85,76,0}},
/* 57757 */ {(14<<2)|2,{85,76,0}},
/* 57758 */ {(14<<2)|2,{85,76,0}},
/* 57759 */ {(14<<2)|2,{85,76,0}},
/* 57760 */ {(14<<2)|2,{85,77,0}},
/* 57761 */ {(14<<2)|2,{85,77,0}},
/* 57762 */ {(14<<2)|2,{85,77,0}},
/* 57763 */ {(14<<2)|2,{85,77,0}},
/* 57764 */ {(14<<2)|2,{85,78,0}},
/* 57765 */ {(14<<2)|2,{85,78,0}},
/* 57766 */ {(14<<2)|2,{85,78,0}},
/* 57767 */ {(14<<2)|2,{85,78,0}},
/* 57768 */ {(14<<2)|2,{85,79,0}},
/* 57769 */ {(14<<2)|2,{85,79,0}},
/* 57770 */ {(14<<2)|2,{85,79,0}},
/* 57771 */ {(14<<2)|2,{85,79,0}},
/* 57772 */ {(14<<2)|2,{85,80,0}},
/* 57773 */ {(14<<2)|2,{85,80,0}},
/* 57774 */ {(14<<2)|2,{85,80,0}},
/* 57775 */ {(14<<2)|2,{85,80,0}},
/* 57776 */ {(14<<2)|2,{85,81,0}},
/* 57777 */ {(14<<2)|2,{85,81,0}},
/* 57778 */ {(14<<2)|2,{85,81,0}},
/* 57779 */ {(14<<2)|2,{85,81,0}},
/* 57780 */ {(14<<2)|2,{85,82,0}},
/* 57781 */ {(14<<2)|2,{85,82,0}},
/* 57782 */ {(14<<2)|2,{85,82,0}},
/* 57783 */ {(14<<2)|2,{85,82,0}},
/* 57784 */ {(14<<2)|2,{85,83,0}},
/* 57785 */ {(14<<2)|2,{85,83,0}},
/* 57786 */ {(14<<2)|2,{85,83,0}},
/* 57787 */ {(14<<2)|2,{85,83,0}},
/* 57788 */ {(14<<2)|2,{85,84,0}},
/* 57789 */ {(14<<2)|2,{85,84,0}},
/* 57790 */ {(14<<2)|2,{85,84,0}},
/* 57791 */ {(14<<2)|2,{85,84,0}},
/* 57792 */ {(14<<2)|2,{85,85,0}},
/* 57793 */ {(14<<2)|2,{85,85,0}},
/* 57794 */ {(14<<2)|2,{85,85,0}},
/* 57795 */ {(14<<2)|2,{85,85,0}},
/* 57796 */ {(14<<2)|2,{85,86,0}},
/* 57797 */ {(14<<2)|2,{85,86,0}},
/* 57798 */ {(14<<2)|2,{85,86,0}},
/* 57799 */ {(14<<2)|2,{85,86,0}},
/* 57800 */ {(14<<2)|2,{85,87,0}},
/* 57801 */ {(14<<2)|2,{85,87,0}},
/* 57802 */ {(14<<2)|2,{85,87,0}},
/* 57803 */ {(14<<2)|2,{85,87,0}},
/* 57804 */ {(14<<2)|2,{85,89,0}},
/* 57805 */ {(14<<2)|2,{85,89,0}},
/* 57806 */ {(14<<2)|2,{85,89,0}},
/* 57807 */ {(14<<2)|2,{85,89,0}},
/* 57808 */ {(14<<2)|2,{85,106,0}},
/* 57809 */ {(14<<2)|2,{85,106,0}},
/* 57810 */ {(14<<2)|2,{85,106,0}},
/* 57811 */ {(14<<2)|2,{85,106,0}},
/* 57812 */ {(14<<2)|2,{85,107,0}},
/* 57813 */ {(14<<2)|2,{85,107,0}},
/* 57814 */ {(14<<2)|2,{85,107,0}},
/* 57815 */ {(14<<2)|2,{85,107,0}},
/* 57816 */ {(14<<2)|2,{85,113,0}},
/* 57817 */ {(14<<2)|2,{85,113,0}},
/* 57818 */ {(14<<2)|2,{85,113,0}},
/* 57819 */ {(14<<2)|2,{85,113,0}},
/* 57820 */ {(14<<2)|2,{85,118,0}},
/* 57821 */ {(14<<2)|2,{85,118,0}},
/* 57822 */ {(14<<2)|2,{85,118,0}},
/* 57823 */ {(14<<2)|2,{85,118,0}},
/* 57824 */ {(14<<2)|2,{85,119,0}},
/* 57825 */ {(14<<2)|2,{85,119,0}},
/* 57826 */ {(14<<2)|2,{85,119,0}},
/* 57827 */ {(14<<2)|2,{85,119,0}},
/* 57828 */ {(14<<2)|2,{85,120,0}},
/* 57829 */ {(14<<2)|2,{85,120,0}},
/* 57830 */ {(14<<2)|2,{85,120,0}},
/* 57831 */ {(14<<2)|2,{85,120,0}},
/* 57832 */ {(14<<2)|2,{85,121,0}},
/* 57833 */ {(14<<2)|2,{85,121,0}},
/* 57834 */ {(14<<2)|2,{85,121,0}},
/* 57835 */ {(14<<2)|2,{85,121,0}},
/* 57836 */ {(14<<2)|2,{85,122,0}},
/* 57837 */ {(14<<2)|2,{85,122,0}},
/* 57838 */ {(14<<2)|2,{85,122,0}},
/* 57839 */ {(14<<2)|2,{85,122,0}},
/* 57840 */ {(15<<2)|2,{85,38,0}},
/* 57841 */ {(15<<2)|2,{85,38,0}},
/* 57842 */ {(15<<2)|2,{85,42,0}},
/* 57843 */ {(15<<2)|2,{85,42,0}},
/* 57844 */ {(15<<2)|2,{85,44,0}},
/* 57845 */ {(15<<2)|2,{85,44,0}},
/* 57846 */ {(15<<2)|2,{85,59,0}},
/* 57847 */ {(15<<2)|2,{85,59,0}},
/* 57848 */ {(15<<2)|2,{85,88,0}},
/* 57849 */ {(15<<2)|2,{85,88,0}},
/* 57850 */ {(15<<2)|2,{85,90,0}},
/* 57851 */ {(15<<2)|2,{85,90,0}},
/* 57852 */ {(7<<2)|1,{85,0,0}},
/* 57853 */ {(7<<2)|1,{85,0,0}},
/* 57854 */ {(7<<2)|1,{85,0,0}},
/* 57855 */ {(7<<2)|1,{85,0,0}},
/* 57856 */ {(12<<2)|2,{86,48,0}},
/* 57857 */ {(12<<2)|2,{86,48,0}},
/* 57858 */ {(12<<2)|2,{86,48,0}},
/* 57859 */ {(12<<2)|2,{86,48,0}},
/* 57860 */ {(12<<2)|2,{86,48,0}},
/* 57861 */ {(12<<2)|2,{86,48,0}},
/* 57862 */ {(12<<2)|2,{86,48,0}},
/* 57863 */ {(12<<2)|2,{86,48,0}},
/* 57864 */ {(12<<2)|2,{86,48,0}},
/* 57865 */ {(12<<2)|2,{86,48,0}},
/* 57866 */ {(12<<2)|2,{86,48,0}},
/* 57867 */ {(12<<2)|2,{86,48,0}},
/* 57868 */ {(12<<2)|2,{86,48,0}},
/* 57869 */ {(12<<2)|2,{86,48,0}},
/* 57870 */ {(12<<2)|2,{86,48,0}},
/* 57871 */ {(12<<2)|2,{86,48,0}},
/* 57872 */ {(12<<2)|2,{86,49,0}},
/* 57873 */ {(12<<2)|2,{86,49,0}},
/* 57874 */ {(12<<2)|2,{86,49,0}},
/* 57875 */ {(12<<2)|2,{86,49,0}},
/* 57876 */ {(12<<2)|2,{86,49,0}},
/* 57877 */ {(12<<2)|2,{86,49,0}},
/* 57878 */ {(12<<2)|2,{86,49,0}},
/* 57879 */ {(12<<2)|2,{86,49,0}},
/* 57880 */ {(12<<2)|2,{86,49,0}},
/* 57881 */ {(12<<2)|2,{86,49,0}},
/* 57882 */ {(12<<2)|2,{86,49,0}},
/* 57883 */ {(12<<2)|2,{86,49,0}},
/* 57884 */ {(12<<2)|2,{86,49,0}},
/* 57885 */ {(12<<2)|2,{86,49,0}},
/* 57886 */ {(12<<2)|2,{86,49,0}},
/* 57887 */ {(12<<2)|2,{86,49,0}},
/* 57888 */ {(12<<2)|2,{86,50,0}},
/* 57889 */ {(12<<2)|2,{86,50,0}},
/* 57890 */ {(12<<2)|2,{86,50,0}},
/* 57891 */ {(12<<2)|2,{86,50,0}},
/* 57892 */ {(12<<2)|2,{86,50,0}},
/* 57893 */ {(12<<2)|2,{86,50,0}},
/* 57894 */ {(12<<2)|2,{86,50,0}},
/* 57895 */ {(12<<2)|2,{86,50,0}},
/* 57896 */ {(12<<2)|2,{86,50,0}},
/* 57897 */ {(12<<2)|2,{86,50,0}},
/* 57898 */ {(12<<2)|2,{86,50,0}},
/* 57899 */ {(12<<2)|2,{86,50,0}},
/* 57900 */ {(12<<2)|2,{86,50,0}},
/* 57901 */ {(12<<2)|2,{86,50,0}},
/* 57902 */ {(12<<2)|2,{86,50,0}},
/* 57903 */ {(12<<2)|2,{86,50,0}},
/* 57904 */ {(12<<2)|2,{86,97,0}},
/* 57905 */ {(12<<2)|2,{86,97,0}},
/* 57906 */ {(12<<2)|2,{86,97,0}},
/* 57907 */ {(12<<2)|2,{86,97,0}},
/* 57908 */ {(12<<2)|2,{86,97,0}},
/* 57909 */ {(12<<2)|2,{86,97,0}},
/* 57910 */ {(12<<2)|2,{86,97,0}},
/* 57911 */ {(12<<2)|2,{86,97,0}},
/* 57912 */ {(12<<2)|2,{86,97,0}},
/* 57913 */ {(12<<2)|2,{86,97,0}},
/* 57914 */ {(12<<2)|2,{86,97,0}},
/* 57915 */ {(12<<2)|2,{86,97,0}},
/* 57916 */ {(12<<2)|2,{86,97,0}},
/* 57917 */ {(12<<2)|2,{86,97,0}},
/* 57918 */ {(12<<2)|2,{86,97,0}},
/* 57919 */ {(12<<2)|2,{86,97,0}},
/* 57920 */ {(12<<2)|2,{86,99,0}},
/* 57921 */ {(12<<2)|2,{86,99,0}},
/* 57922 */ {(12<<2)|2,{86,99,0}},
/* 57923 */ {(12<<2)|2,{86,99,0}},
/* 57924 */ {(12<<2)|2,{86,99,0}},
/* 57925 */ {(12<<2)|2,{86,99,0}},
/* 57926 */ {(12<<2)|2,{86,99,0}},
/* 57927 */ {(12<<2)|2,{86,99,0}},
/* 57928 */ {(12<<2)|2,{86,99,0}},
/* 57929 */ {(12<<2)|2,{86,99,0}},
/* 57930 */ {(12<<2)|2,{86,99,0}},
/* 57931 */ {(12<<2)|2,{86,99,0}},
/* 57932 */ {(12<<2)|2,{86,99,0}},
/* 57933 */ {(12<<2)|2,{86,99,0}},
/* 57934 */ {(12<<2)|2,{86,99,0}},
/* 57935 */ {(12<<2)|2,{86,99,0}},
/* 57936 */ {(12<<2)|2,{86,101,0}},
/* 57937 */ {(12<<2)|2,{86,101,0}},
/* 57938 */ {(12<<2)|2,{86,101,0}},
/* 57939 */ {(12<<2)|2,{86,101,0}},
/* 57940 */ {(12<<2)|2,{86,101,0}},
/* 57941 */ {(12<<2)|2,{86,101,0}},
/* 57942 */ {(12<<2)|2,{86,101,0}},
/* 57943 */ {(12<<2)|2,{86,101,0}},
/* 57944 */ {(12<<2)|2,{86,101,0}},
/* 57945 */ {(12<<2)|2,{86,101,0}},
/* 57946 */ {(12<<2)|2,{86,101,0}},
/* 57947 */ {(12<<2)|2,{86,101,0}},
/* 57948 */ {(12<<2)|2,{86,101,0}},
/* 57949 */ {(12<<2)|2,{86,101,0}},
/* 57950 */ {(12<<2)|2,{86,101,0}},
/* 57951 */ {(12<<2)|2,{86,101,0}},
/* 57952 */ {(12<<2)|2,{86,105,0}},
/* 57953 */ {(12<<2)|2,{86,105,0}},
/* 57954 */ {(12<<2)|2,{86,105,0}},
/* 57955 */ {(12<<2)|2,{86,105,0}},
/* 57956 */ {(12<<2)|2,{86,105,0}},
/* 57957 */ {(12<<2)|2,{86,105,0}},
/* 57958 */ {(12<<2)|2,{86,105,0}},
/* 57959 */ {(12<<2)|2,{86,105,0}},
/* 57960 */ {(12<<2)|2,{86,105,0}},
/* 57961 */ {(12<<2)|2,{86,105,0}},
/* 57962 */ {(12<<2)|2,{86,105,0}},
/* 57963 */ {(12<<2)|2,{86,105,0}},
/* 57964 */ {(12<<2)|2,{86,105,0}},
/* 57965 */ {(12<<2)|2,{86,105,0}},
/* 57966 */ {(12<<2)|2,{86,105,0}},
/* 57967 */ {(12<<2)|2,{86,105,0}},
/* 57968 */ {(12<<2)|2,{86,111,0}},
/* 57969 */ {(12<<2)|2,{86,111,0}},
/* 57970 */ {(12<<2)|2,{86,111,0}},
/* 57971 */ {(12<<2)|2,{86,111,0}},
/* 57972 */ {(12<<2)|2,{86,111,0}},
/* 57973 */ {(12<<2)|2,{86,111,0}},
/* 57974 */ {(12<<2)|2,{86,111,0}},
/* 57975 */ {(12<<2)|2,{86,111,0}},
/* 57976 */ {(12<<2)|2,{86,111,0}},
/* 57977 */ {(12<<2)|2,{86,111,0}},
/* 57978 */ {(12<<2)|2,{86,111,0}},
/* 57979 */ {(12<<2)|2,{86,111,0}},
/* 57980 */ {(12<<2)|2,{86,111,0}},
/* 57981 */ {(12<<2)|2,{86,111,0}},
/* 57982 */ {(12<<2)|2,{86,111,0}},
/* 57983 */ {(12<<2)|2,{86,111,0}},
/* 57984 */ {(12<<2)|2,{86,115,0}},
/* 57985 */ {(12<<2)|2,{86,115,0}},
/* 57986 */ {(12<<2)|2,{86,115,0}},
/* 57987 */ {(12<<2)|2,{86,115,0}},
/* 57988 */ {(12<<2)|2,{86,115,0}},
/* 57989 */ {(12<<2)|2,{86,115,0}},
/* 57990 */ {(12<<2)|2,{86,115,0}},
/* 57991 */ {(12<<2)|2,{86,115,0}},
/* 57992 */ {(12<<2)|2,{86,115,0}},
/* 57993 */ {(12<<2)|2,{86,115,0}},
/* 57994 */ {(12<<2)|2,{86,115,0}},
/* 57995 */ {(12<<2)|2,{86,115,0}},
/* 57996 */ {(12<<2)|2,{86,115,0}},
/* 57997 */ {(12<<2)|2,{86,115,0}},
/* 57998 */ {(12<<2)|2,{86,115,0}},
/* 57999 */ {(12<<2)|2,{86,115,0}},
/* 58000 */ {(12<<2)|2,{86,116,0}},
/* 58001 */ {(12<<2)|2,{86,116,0}},
/* 58002 */ {(12<<2)|2,{86,116,0}},
/* 58003 */ {(12<<2)|2,{86,116,0}},
/* 58004 */ {(12<<2)|2,{86,116,0}},
/* 58005 */ {(12<<2)|2,{86,116,0}},
/* 58006 */ {(12<<2)|2,{86,116,0}},
/* 58007 */ {(12<<2)|2,{86,116,0}},
/* 58008 */ {(12<<2)|2,{86,116,0}},
/* 58009 */ {(12<<2)|2,{86,116,0}},
/* 58010 */ {(12<<2)|2,{86,116,0}},
/* 58011 */ {(12<<2)|2,{86,116,0}},
/* 58012 */ {(12<<2)|2,{86,116,0}},
/* 58013 */ {(12<<2)|2,{86,116,0}},
/* 58014 */ {(12<<2)|2,{86,116,0}},
/* 58015 */ {(12<<2)|2,{86,116,0}},
/* 58016 */ {(13<<2)|2,{86,32,0}},
/* 58017 */ {(13<<2)|2,{86,32,0}},
/* 58018 */ {(13<<2)|2,{86,32,0}},
/* 58019 */ {(13<<2)|2,{86,32,0}},
/* 58020 */ {(13<<2)|2,{86,32,0}},
/* 58021 */ {(13<<2)|2,{86,32,0}},
/* 58022 */ {(13<<2)|2,{86,32,0}},
/* 58023 */ {(13<<2)|2,{86,32,0}},
/* 58024 */ {(13<<2)|2,{86,37,0}},
/* 58025 */ {(13<<2)|2,{86,37,0}},
/* 58026 */ {(13<<2)|2,{86,37,0}},
/* 58027 */ {(13<<2)|2,{86,37,0}},
/* 58028 */ {(13<<2)|2,{86,37,0}},
/* 58029 */ {(13<<2)|2,{86,37,0}},
/* 58030 */ {(13<<2)|2,{86,37,0}},
/* 58031 */ {(13<<2)|2,{86,37,0}},
/* 58032 */ {(13<<2)|2,{86,45,0}},
/* 58033 */ {(13<<2)|2,{86,45,0}},
/* 58034 */ {(13<<2)|2,{86,45,0}},
/* 58035 */ {(13<<2)|2,{86,45,0}},
/* 58036 */ {(13<<2)|2,{86,45,0}},
/* 58037 */ {(13<<2)|2,{86,45,0}},
/* 58038 */ {(13<<2)|2,{86,45,0}},
/* 58039 */ {(13<<2)|2,{86,45,0}},
/* 58040 */ {(13<<2)|2,{86,46,0}},
/* 58041 */ {(13<<2)|2,{86,46,0}},
/* 58042 */ {(13<<2)|2,{86,46,0}},
/* 58043 */ {(13<<2)|2,{86,46,0}},
/* 58044 */ {(13<<2)|2,{86,46,0}},
/* 58045 */ {(13<<2)|2,{86,46,0}},
/* 58046 */ {(13<<2)|2,{86,46,0}},
/* 58047 */ {(13<<2)|2,{86,46,0}},
/* 58048 */ {(13<<2)|2,{86,47,0}},
/* 58049 */ {(13<<2)|2,{86,47,0}},
/* 58050 */ {(13<<2)|2,{86,47,0}},
/* 58051 */ {(13<<2)|2,{86,47,0}},
/* 58052 */ {(13<<2)|2,{86,47,0}},
/* 58053 */ {(13<<2)|2,{86,47,0}},
/* 58054 */ {(13<<2)|2,{86,47,0}},
/* 58055 */ {(13<<2)|2,{86,47,0}},
/* 58056 */ {(13<<2)|2,{86,51,0}},
/* 58057 */ {(13<<2)|2,{86,51,0}},
/* 58058 */ {(13<<2)|2,{86,51,0}},
/* 58059 */ {(13<<2)|2,{86,51,0}},
/* 58060 */ {(13<<2)|2,{86,51,0}},
/* 58061 */ {(13<<2)|2,{86,51,0}},
/* 58062 */ {(13<<2)|2,{86,51,0}},
/* 58063 */ {(13<<2)|2,{86,51,0}},
/* 58064 */ {(13<<2)|2,{86,52,0}},
/* 58065 */ {(13<<2)|2,{86,52,0}},
/* 58066 */ {(13<<2)|2,{86,52,0}},
/* 58067 */ {(13<<2)|2,{86,52,0}},
/* 58068 */ {(13<<2)|2,{86,52,0}},
/* 58069 */ {(13<<2)|2,{86,52,0}},
/* 58070 */ {(13<<2)|2,{86,52,0}},
/* 58071 */ {(13<<2)|2,{86,52,0}},
/* 58072 */ {(13<<2)|2,{86,53,0}},
/* 58073 */ {(13<<2)|2,{86,53,0}},
/* 58074 */ {(13<<2)|2,{86,53,0}},
/* 58075 */ {(13<<2)|2,{86,53,0}},
/* 58076 */ {(13<<2)|2,{86,53,0}},
/* 58077 */ {(13<<2)|2,{86,53,0}},
/* 58078 */ {(13<<2)|2,{86,53,0}},
/* 58079 */ {(13<<2)|2,{86,53,0}},
/* 58080 */ {(13<<2)|2,{86,54,0}},
/* 58081 */ {(13<<2)|2,{86,54,0}},
/* 58082 */ {(13<<2)|2,{86,54,0}},
/* 58083 */ {(13<<2)|2,{86,54,0}},
/* 58084 */ {(13<<2)|2,{86,54,0}},
/* 58085 */ {(13<<2)|2,{86,54,0}},
/* 58086 */ {(13<<2)|2,{86,54,0}},
/* 58087 */ {(13<<2)|2,{86,54,0}},
/* 58088 */ {(13<<2)|2,{86,55,0}},
/* 58089 */ {(13<<2)|2,{86,55,0}},
/* 58090 */ {(13<<2)|2,{86,55,0}},
/* 58091 */ {(13<<2)|2,{86,55,0}},
/* 58092 */ {(13<<2)|2,{86,55,0}},
/* 58093 */ {(13<<2)|2,{86,55,0}},
/* 58094 */ {(13<<2)|2,{86,55,0}},
/* 58095 */ {(13<<2)|2,{86,55,0}},
/* 58096 */ {(13<<2)|2,{86,56,0}},
/* 58097 */ {(13<<2)|2,{86,56,0}},
/* 58098 */ {(13<<2)|2,{86,56,0}},
/* 58099 */ {(13<<2)|2,{86,56,0}},
/* 58100 */ {(13<<2)|2,{86,56,0}},
/* 58101 */ {(13<<2)|2,{86,56,0}},
/* 58102 */ {(13<<2)|2,{86,56,0}},
/* 58103 */ {(13<<2)|2,{86,56,0}},
/* 58104 */ {(13<<2)|2,{86,57,0}},
/* 58105 */ {(13<<2)|2,{86,57,0}},
/* 58106 */ {(13<<2)|2,{86,57,0}},
/* 58107 */ {(13<<2)|2,{86,57,0}},
/* 58108 */ {(13<<2)|2,{86,57,0}},
/* 58109 */ {(13<<2)|2,{86,57,0}},
/* 58110 */ {(13<<2)|2,{86,57,0}},
/* 58111 */ {(13<<2)|2,{86,57,0}},
/* 58112 */ {(13<<2)|2,{86,61,0}},
/* 58113 */ {(13<<2)|2,{86,61,0}},
/* 58114 */ {(13<<2)|2,{86,61,0}},
/* 58115 */ {(13<<2)|2,{86,61,0}},
/* 58116 */ {(13<<2)|2,{86,61,0}},
/* 58117 */ {(13<<2)|2,{86,61,0}},
/* 58118 */ {(13<<2)|2,{86,61,0}},
/* 58119 */ {(13<<2)|2,{86,61,0}},
/* 58120 */ {(13<<2)|2,{86,65,0}},
/* 58121 */ {(13<<2)|2,{86,65,0}},
/* 58122 */ {(13<<2)|2,{86,65,0}},
/* 58123 */ {(13<<2)|2,{86,65,0}},
/* 58124 */ {(13<<2)|2,{86,65,0}},
/* 58125 */ {(13<<2)|2,{86,65,0}},
/* 58126 */ {(13<<2)|2,{86,65,0}},
/* 58127 */ {(13<<2)|2,{86,65,0}},
/* 58128 */ {(13<<2)|2,{86,95,0}},
/* 58129 */ {(13<<2)|2,{86,95,0}},
/* 58130 */ {(13<<2)|2,{86,95,0}},
/* 58131 */ {(13<<2)|2,{86,95,0}},
/* 58132 */ {(13<<2)|2,{86,95,0}},
/* 58133 */ {(13<<2)|2,{86,95,0}},
/* 58134 */ {(13<<2)|2,{86,95,0}},
/* 58135 */ {(13<<2)|2,{86,95,0}},
/* 58136 */ {(13<<2)|2,{86,98,0}},
/* 58137 */ {(13<<2)|2,{86,98,0}},
/* 58138 */ {(13<<2)|2,{86,98,0}},
/* 58139 */ {(13<<2)|2,{86,98,0}},
/* 58140 */ {(13<<2)|2,{86,98,0}},
/* 58141 */ {(13<<2)|2,{86,98,0}},
/* 58142 */ {(13<<2)|2,{86,98,0}},
/* 58143 */ {(13<<2)|2,{86,98,0}},
/* 58144 */ {(13<<2)|2,{86,100,0}},
/* 58145 */ {(13<<2)|2,{86,100,0}},
/* 58146 */ {(13<<2)|2,{86,100,0}},
/* 58147 */ {(13<<2)|2,{86,100,0}},
/* 58148 */ {(13<<2)|2,{86,100,0}},
/* 58149 */ {(13<<2)|2,{86,100,0}},
/* 58150 */ {(13<<2)|2,{86,100,0}},
/* 58151 */ {(13<<2)|2,{86,100,0}},
/* 58152 */ {(13<<2)|2,{86,102,0}},
/* 58153 */ {(13<<2)|2,{86,102,0}},
/* 58154 */ {(13<<2)|2,{86,102,0}},
/* 58155 */ {(13<<2)|2,{86,102,0}},
/* 58156 */ {(13<<2)|2,{86,102,0}},
/* 58157 */ {(13<<2)|2,{86,102,0}},
/* 58158 */ {(13<<2)|2,{86,102,0}},
/* 58159 */ {(13<<2)|2,{86,102,0}},
/* 58160 */ {(13<<2)|2,{86,103,0}},
/* 58161 */ {(13<<2)|2,{86,103,0}},
/* 58162 */ {(13<<2)|2,{86,103,0}},
/* 58163 */ {(13<<2)|2,{86,103,0}},
/* 58164 */ {(13<<2)|2,{86,103,0}},
/* 58165 */ {(13<<2)|2,{86,103,0}},
/* 58166 */ {(13<<2)|2,{86,103,0}},
/* 58167 */ {(13<<2)|2,{86,103,0}},
/* 58168 */ {(13<<2)|2,{86,104,0}},
/* 58169 */ {(13<<2)|2,{86,104,0}},
/* 58170 */ {(13<<2)|2,{86,104,0}},
/* 58171 */ {(13<<2)|2,{86,104,0}},
/* 58172 */ {(13<<2)|2,{86,104,0}},
/* 58173 */ {(13<<2)|2,{86,104,0}},
/* 58174 */ {(13<<2)|2,{86,104,0}},
/* 58175 */ {(13<<2)|2,{86,104,0}},
/* 58176 */ {(13<<2)|2,{86,108,0}},
/* 58177 */ {(13<<2)|2,{86,108,0}},
/* 58178 */ {(13<<2)|2,{86,108,0}},
/* 58179 */ {(13<<2)|2,{86,108,0}},
/* 58180 */ {(13<<2)|2,{86,108,0}},
/* 58181 */ {(13<<2)|2,{86,108,0}},
/* 58182 */ {(13<<2)|2,{86,108,0}},
/* 58183 */ {(13<<2)|2,{86,108,0}},
/* 58184 */ {(13<<2)|2,{86,109,0}},
/* 58185 */ {(13<<2)|2,{86,109,0}},
/* 58186 */ {(13<<2)|2,{86,109,0}},
/* 58187 */ {(13<<2)|2,{86,109,0}},
/* 58188 */ {(13<<2)|2,{86,109,0}},
/* 58189 */ {(13<<2)|2,{86,109,0}},
/* 58190 */ {(13<<2)|2,{86,109,0}},
/* 58191 */ {(13<<2)|2,{86,109,0}},
/* 58192 */ {(13<<2)|2,{86,110,0}},
/* 58193 */ {(13<<2)|2,{86,110,0}},
/* 58194 */ {(13<<2)|2,{86,110,0}},
/* 58195 */ {(13<<2)|2,{86,110,0}},
/* 58196 */ {(13<<2)|2,{86,110,0}},
/* 58197 */ {(13<<2)|2,{86,110,0}},
/* 58198 */ {(13<<2)|2,{86,110,0}},
/* 58199 */ {(13<<2)|2,{86,110,0}},
/* 58200 */ {(13<<2)|2,{86,112,0}},
/* 58201 */ {(13<<2)|2,{86,112,0}},
/* 58202 */ {(13<<2)|2,{86,112,0}},
/* 58203 */ {(13<<2)|2,{86,112,0}},
/* 58204 */ {(13<<2)|2,{86,112,0}},
/* 58205 */ {(13<<2)|2,{86,112,0}},
/* 58206 */ {(13<<2)|2,{86,112,0}},
/* 58207 */ {(13<<2)|2,{86,112,0}},
/* 58208 */ {(13<<2)|2,{86,114,0}},
/* 58209 */ {(13<<2)|2,{86,114,0}},
/* 58210 */ {(13<<2)|2,{86,114,0}},
/* 58211 */ {(13<<2)|2,{86,114,0}},
/* 58212 */ {(13<<2)|2,{86,114,0}},
/* 58213 */ {(13<<2)|2,{86,114,0}},
/* 58214 */ {(13<<2)|2,{86,114,0}},
/* 58215 */ {(13<<2)|2,{86,114,0}},
/* 58216 */ {(13<<2)|2,{86,117,0}},
/* 58217 */ {(13<<2)|2,{86,117,0}},
/* 58218 */ {(13<<2)|2,{86,117,0}},
/* 58219 */ {(13<<2)|2,{86,117,0}},
/* 58220 */ {(13<<2)|2,{86,117,0}},
/* 58221 */ {(13<<2)|2,{86,117,0}},
/* 58222 */ {(13<<2)|2,{86,117,0}},
/* 58223 */ {(13<<2)|2,{86,117,0}},
/* 58224 */ {(14<<2)|2,{86,58,0}},
/* 58225 */ {(14<<2)|2,{86,58,0}},
/* 58226 */ {(14<<2)|2,{86,58,0}},
/* 58227 */ {(14<<2)|2,{86,58,0}},
/* 58228 */ {(14<<2)|2,{86,66,0}},
/* 58229 */ {(14<<2)|2,{86,66,0}},
/* 58230 */ {(14<<2)|2,{86,66,0}},
/* 58231 */ {(14<<2)|2,{86,66,0}},
/* 58232 */ {(14<<2)|2,{86,67,0}},
/* 58233 */ {(14<<2)|2,{86,67,0}},
/* 58234 */ {(14<<2)|2,{86,67,0}},
/* 58235 */ {(14<<2)|2,{86,67,0}},
/* 58236 */ {(14<<2)|2,{86,68,0}},
/* 58237 */ {(14<<2)|2,{86,68,0}},
/* 58238 */ {(14<<2)|2,{86,68,0}},
/* 58239 */ {(14<<2)|2,{86,68,0}},
/* 58240 */ {(14<<2)|2,{86,69,0}},
/* 58241 */ {(14<<2)|2,{86,69,0}},
/* 58242 */ {(14<<2)|2,{86,69,0}},
/* 58243 */ {(14<<2)|2,{86,69,0}},
/* 58244 */ {(14<<2)|2,{86,70,0}},
/* 58245 */ {(14<<2)|2,{86,70,0}},
/* 58246 */ {(14<<2)|2,{86,70,0}},
/* 58247 */ {(14<<2)|2,{86,70,0}},
/* 58248 */ {(14<<2)|2,{86,71,0}},
/* 58249 */ {(14<<2)|2,{86,71,0}},
/* 58250 */ {(14<<2)|2,{86,71,0}},
/* 58251 */ {(14<<2)|2,{86,71,0}},
/* 58252 */ {(14<<2)|2,{86,72,0}},
/* 58253 */ {(14<<2)|2,{86,72,0}},
/* 58254 */ {(14<<2)|2,{86,72,0}},
/* 58255 */ {(14<<2)|2,{86,72,0}},
/* 58256 */ {(14<<2)|2,{86,73,0}},
/* 58257 */ {(14<<2)|2,{86,73,0}},
/* 58258 */ {(14<<2)|2,{86,73,0}},
/* 58259 */ {(14<<2)|2,{86,73,0}},
/* 58260 */ {(14<<2)|2,{86,74,0}},
/* 58261 */ {(14<<2)|2,{86,74,0}},
/* 58262 */ {(14<<2)|2,{86,74,0}},
/* 58263 */ {(14<<2)|2,{86,74,0}},
/* 58264 */ {(14<<2)|2,{86,75,0}},
/* 58265 */ {(14<<2)|2,{86,75,0}},
/* 58266 */ {(14<<2)|2,{86,75,0}},
/* 58267 */ {(14<<2)|2,{86,75,0}},
/* 58268 */ {(14<<2)|2,{86,76,0}},
/* 58269 */ {(14<<2)|2,{86,76,0}},
/* 58270 */ {(14<<2)|2,{86,76,0}},
/* 58271 */ {(14<<2)|2,{86,76,0}},
/* 58272 */ {(14<<2)|2,{86,77,0}},
/* 58273 */ {(14<<2)|2,{86,77,0}},
/* 58274 */ {(14<<2)|2,{86,77,0}},
/* 58275 */ {(14<<2)|2,{86,77,0}},
/* 58276 */ {(14<<2)|2,{86,78,0}},
/* 58277 */ {(14<<2)|2,{86,78,0}},
/* 58278 */ {(14<<2)|2,{86,78,0}},
/* 58279 */ {(14<<2)|2,{86,78,0}},
/* 58280 */ {(14<<2)|2,{86,79,0}},
/* 58281 */ {(14<<2)|2,{86,79,0}},
/* 58282 */ {(14<<2)|2,{86,79,0}},
/* 58283 */ {(14<<2)|2,{86,79,0}},
/* 58284 */ {(14<<2)|2,{86,80,0}},
/* 58285 */ {(14<<2)|2,{86,80,0}},
/* 58286 */ {(14<<2)|2,{86,80,0}},
/* 58287 */ {(14<<2)|2,{86,80,0}},
/* 58288 */ {(14<<2)|2,{86,81,0}},
/* 58289 */ {(14<<2)|2,{86,81,0}},
/* 58290 */ {(14<<2)|2,{86,81,0}},
/* 58291 */ {(14<<2)|2,{86,81,0}},
/* 58292 */ {(14<<2)|2,{86,82,0}},
/* 58293 */ {(14<<2)|2,{86,82,0}},
/* 58294 */ {(14<<2)|2,{86,82,0}},
/* 58295 */ {(14<<2)|2,{86,82,0}},
/* 58296 */ {(14<<2)|2,{86,83,0}},
/* 58297 */ {(14<<2)|2,{86,83,0}},
/* 58298 */ {(14<<2)|2,{86,83,0}},
/* 58299 */ {(14<<2)|2,{86,83,0}},
/* 58300 */ {(14<<2)|2,{86,84,0}},
/* 58301 */ {(14<<2)|2,{86,84,0}},
/* 58302 */ {(14<<2)|2,{86,84,0}},
/* 58303 */ {(14<<2)|2,{86,84,0}},
/* 58304 */ {(14<<2)|2,{86,85,0}},
/* 58305 */ {(14<<2)|2,{86,85,0}},
/* 58306 */ {(14<<2)|2,{86,85,0}},
/* 58307 */ {(14<<2)|2,{86,85,0}},
/* 58308 */ {(14<<2)|2,{86,86,0}},
/* 58309 */ {(14<<2)|2,{86,86,0}},
/* 58310 */ {(14<<2)|2,{86,86,0}},
/* 58311 */ {(14<<2)|2,{86,86,0}},
/* 58312 */ {(14<<2)|2,{86,87,0}},
/* 58313 */ {(14<<2)|2,{86,87,0}},
/* 58314 */ {(14<<2)|2,{86,87,0}},
/* 58315 */ {(14<<2)|2,{86,87,0}},
/* 58316 */ {(14<<2)|2,{86,89,0}},
/* 58317 */ {(14<<2)|2,{86,89,0}},
/* 58318 */ {(14<<2)|2,{86,89,0}},
/* 58319 */ {(14<<2)|2,{86,89,0}},
/* 58320 */ {(14<<2)|2,{86,106,0}},
/* 58321 */ {(14<<2)|2,{86,106,0}},
/* 58322 */ {(14<<2)|2,{86,106,0}},
/* 58323 */ {(14<<2)|2,{86,106,0}},
/* 58324 */ {(14<<2)|2,{86,107,0}},
/* 58325 */ {(14<<2)|2,{86,107,0}},
/* 58326 */ {(14<<2)|2,{86,107,0}},
/* 58327 */ {(14<<2)|2,{86,107,0}},
/* 58328 */ {(14<<2)|2,{86,113,0}},
/* 58329 */ {(14<<2)|2,{86,113,0}},
/* 58330 */ {(14<<2)|2,{86,113,0}},
/* 58331 */ {(14<<2)|2,{86,113,0}},
/* 58332 */ {(14<<2)|2,{86,118,0}},
/* 58333 */ {(14<<2)|2,{86,118,0}},
/* 58334 */ {(14<<2)|2,{86,118,0}},
/* 58335 */ {(14<<2)|2,{86,118,0}},
/* 58336 */ {(14<<2)|2,{86,119,0}},
/* 58337 */ {(14<<2)|2,{86,119,0}},
/* 58338 */ {(14<<2)|2,{86,119,0}},
/* 58339 */ {(14<<2)|2,{86,119,0}},
/* 58340 */ {(14<<2)|2,{86,120,0}},
/* 58341 */ {(14<<2)|2,{86,120,0}},
/* 58342 */ {(14<<2)|2,{86,120,0}},
/* 58343 */ {(14<<2)|2,{86,120,0}},
/* 58344 */ {(14<<2)|2,{86,121,0}},
/* 58345 */ {(14<<2)|2,{86,121,0}},
/* 58346 */ {(14<<2)|2,{86,121,0}},
/* 58347 */ {(14<<2)|2,{86,121,0}},
/* 58348 */ {(14<<2)|2,{86,122,0}},
/* 58349 */ {(14<<2)|2,{86,122,0}},
/* 58350 */ {(14<<2)|2,{86,122,0}},
/* 58351 */ {(14<<2)|2,{86,122,0}},
/* 58352 */ {(15<<2)|2,{86,38,0}},
/* 58353 */ {(15<<2)|2,{86,38,0}},
/* 58354 */ {(15<<2)|2,{86,42,0}},
/* 58355 */ {(15<<2)|2,{86,42,0}},
/* 58356 */ {(15<<2)|2,{86,44,0}},
/* 58357 */ {(15<<2)|2,{86,44,0}},
/* 58358 */ {(15<<2)|2,{86,59,0}},
/* 58359 */ {(15<<2)|2,{86,59,0}},
/* 58360 */ {(15<<2)|2,{86,88,0}},
/* 58361 */ {(15<<2)|2,{86,88,0}},
/* 58362 */ {(15<<2)|2,{86,90,0}},
/* 58363 */ {(15<<2)|2,{86,90,0}},
/* 58364 */ {(7<<2)|1,{86,0,0}},
/* 58365 */ {(7<<2)|1,{86,0,0}},
/* 58366 */ {(7<<2)|1,{86,0,0}},
/* 58367 */ {(7<<2)|1,{86,0,0}},
/* 58368 */ {(12<<2)|2,{87,48,0}},
/* 58369 */ {(12<<2)|2,{87,48,0}},
/* 58370 */ {(12<<2)|2,{87,48,0}},
/* 58371 */ {(12<<2)|2,{87,48,0}},
/* 58372 */ {(12<<2)|2,{87,48,0}},
/* 58373 */ {(12<<2)|2,{87,48,0}},
/* 58374 */ {(12<<2)|2,{87,48,0}},
/* 58375 */ {(12<<2)|2,{87,48,0}},
/* 58376 */ {(12<<2)|2,{87,48,0}},
/* 58377 */ {(12<<2)|2,{87,48,0}},
/* 58378 */ {(12<<2)|2,{87,48,0}},
/* 58379 */ {(12<<2)|2,{87,48,0}},
/* 58380 */ {(12<<2)|2,{87,48,0}},
/* 58381 */ {(12<<2)|2,{87,48,0}},
/* 58382 */ {(12<<2)|2,{87,48,0}},
/* 58383 */ {(12<<2)|2,{87,48,0}},
/* 58384 */ {(12<<2)|2,{87,49,0}},
/* 58385 */ {(12<<2)|2,{87,49,0}},
/* 58386 */ {(12<<2)|2,{87,49,0}},
/* 58387 */ {(12<<2)|2,{87,49,0}},
/* 58388 */ {(12<<2)|2,{87,49,0}},
/* 58389 */ {(12<<2)|2,{87,49,0}},
/* 58390 */ {(12<<2)|2,{87,49,0}},
/* 58391 */ {(12<<2)|2,{87,49,0}},
/* 58392 */ {(12<<2)|2,{87,49,0}},
/* 58393 */ {(12<<2)|2,{87,49,0}},
/* 58394 */ {(12<<2)|2,{87,49,0}},
/* 58395 */ {(12<<2)|2,{87,49,0}},
/* 58396 */ {(12<<2)|2,{87,49,0}},
/* 58397 */ {(12<<2)|2,{87,49,0}},
/* 58398 */ {(12<<2)|2,{87,49,0}},
/* 58399 */ {(12<<2)|2,{87,49,0}},
/* 58400 */ {(12<<2)|2,{87,50,0}},
/* 58401 */ {(12<<2)|2,{87,50,0}},
/* 58402 */ {(12<<2)|2,{87,50,0}},
/* 58403 */ {(12<<2)|2,{87,50,0}},
/* 58404 */ {(12<<2)|2,{87,50,0}},
/* 58405 */ {(12<<2)|2,{87,50,0}},
/* 58406 */ {(12<<2)|2,{87,50,0}},
/* 58407 */ {(12<<2)|2,{87,50,0}},
/* 58408 */ {(12<<2)|2,{87,50,0}},
/* 58409 */ {(12<<2)|2,{87,50,0}},
/* 58410 */ {(12<<2)|2,{87,50,0}},
/* 58411 */ {(12<<2)|2,{87,50,0}},
/* 58412 */ {(12<<2)|2,{87,50,0}},
/* 58413 */ {(12<<2)|2,{87,50,0}},
/* 58414 */ {(12<<2)|2,{87,50,0}},
/* 58415 */ {(12<<2)|2,{87,50,0}},
/* 58416 */ {(12<<2)|2,{87,97,0}},
/* 58417 */ {(12<<2)|2,{87,97,0}},
/* 58418 */ {(12<<2)|2,{87,97,0}},
/* 58419 */ {(12<<2)|2,{87,97,0}},
/* 58420 */ {(12<<2)|2,{87,97,0}},
/* 58421 */ {(12<<2)|2,{87,97,0}},
/* 58422 */ {(12<<2)|2,{87,97,0}},
/* 58423 */ {(12<<2)|2,{87,97,0}},
/* 58424 */ {(12<<2)|2,{87,97,0}},
/* 58425 */ {(12<<2)|2,{87,97,0}},
/* 58426 */ {(12<<2)|2,{87,97,0}},
/* 58427 */ {(12<<2)|2,{87,97,0}},
/* 58428 */ {(12<<2)|2,{87,97,0}},
/* 58429 */ {(12<<2)|2,{87,97,0}},
/* 58430 */ {(12<<2)|2,{87,97,0}},
/* 58431 */ {(12<<2)|2,{87,97,0}},
/* 58432 */ {(12<<2)|2,{87,99,0}},
/* 58433 */ {(12<<2)|2,{87,99,0}},
/* 58434 */ {(12<<2)|2,{87,99,0}},
/* 58435 */ {(12<<2)|2,{87,99,0}},
/* 58436 */ {(12<<2)|2,{87,99,0}},
/* 58437 */ {(12<<2)|2,{87,99,0}},
/* 58438 */ {(12<<2)|2,{87,99,0}},
/* 58439 */ {(12<<2)|2,{87,99,0}},
/* 58440 */ {(12<<2)|2,{87,99,0}},
/* 58441 */ {(12<<2)|2,{87,99,0}},
/* 58442 */ {(12<<2)|2,{87,99,0}},
/* 58443 */ {(12<<2)|2,{87,99,0}},
/* 58444 */ {(12<<2)|2,{87,99,0}},
/* 58445 */ {(12<<2)|2,{87,99,0}},
/* 58446 */ {(12<<2)|2,{87,99,0}},
/* 58447 */ {(12<<2)|2,{87,99,0}},
/* 58448 */ {(12<<2)|2,{87,101,0}},
/* 58449 */ {(12<<2)|2,{87,101,0}},
/* 58450 */ {(12<<2)|2,{87,101,0}},
/* 58451 */ {(12<<2)|2,{87,101,0}},
/* 58452 */ {(12<<2)|2,{87,101,0}},
/* 58453 */ {(12<<2)|2,{87,101,0}},
/* 58454 */ {(12<<2)|2,{87,101,0}},
/* 58455 */ {(12<<2)|2,{87,101,0}},
/* 58456 */ {(12<<2)|2,{87,101,0}},
/* 58457 */ {(12<<2)|2,{87,101,0}},
/* 58458 */ {(12<<2)|2,{87,101,0}},
/* 58459 */ {(12<<2)|2,{87,101,0}},
/* 58460 */ {(12<<2)|2,{87,101,0}},
/* 58461 */ {(12<<2)|2,{87,101,0}},
/* 58462 */ {(12<<2)|2,{87,101,0}},
/* 58463 */ {(12<<2)|2,{87,101,0}},
/* 58464 */ {(12<<2)|2,{87,105,0}},
/* 58465 */ {(12<<2)|2,{87,105,0}},
/* 58466 */ {(12<<2)|2,{87,105,0}},
/* 58467 */ {(12<<2)|2,{87,105,0}},
/* 58468 */ {(12<<2)|2,{87,105,0}},
/* 58469 */ {(12<<2)|2,{87,105,0}},
/* 58470 */ {(12<<2)|2,{87,105,0}},
/* 58471 */ {(12<<2)|2,{87,105,0}},
/* 58472 */ {(12<<2)|2,{87,105,0}},
/* 58473 */ {(12<<2)|2,{87,105,0}},
/* 58474 */ {(12<<2)|2,{87,105,0}},
/* 58475 */ {(12<<2)|2,{87,105,0}},
/* 58476 */ {(12<<2)|2,{87,105,0}},
/* 58477 */ {(12<<2)|2,{87,105,0}},
/* 58478 */ {(12<<2)|2,{87,105,0}},
/* 58479 */ {(12<<2)|2,{87,105,0}},
/* 58480 */ {(12<<2)|2,{87,111,0}},
/* 58481 */ {(12<<2)|2,{87,111,0}},
/* 58482 */ {(12<<2)|2,{87,111,0}},
/* 58483 */ {(12<<2)|2,{87,111,0}},
/* 58484 */ {(12<<2)|2,{87,111,0}},
/* 58485 */ {(12<<2)|2,{87,111,0}},
/* 58486 */ {(12<<2)|2,{87,111,0}},
/* 58487 */ {(12<<2)|2,{87,111,0}},
/* 58488 */ {(12<<2)|2,{87,111,0}},
/* 58489 */ {(12<<2)|2,{87,111,0}},
/* 58490 */ {(12<<2)|2,{87,111,0}},
/* 58491 */ {(12<<2)|2,{87,111,0}},
/* 58492 */ {(12<<2)|2,{87,111,0}},
/* 58493 */ {(12<<2)|2,{87,111,0}},
/* 58494 */ {(12<<2)|2,{87,111,0}},
/* 58495 */ {(12<<2)|2,{87,111,0}},
/* 58496 */ {(12<<2)|2,{87,115,0}},
/* 58497 */ {(12<<2)|2,{87,115,0}},
/* 58498 */ {(12<<2)|2,{87,115,0}},
/* 58499 */ {(12<<2)|2,{87,115,0}},
/* 58500 */ {(12<<2)|2,{87,115,0}},
/* 58501 */ {(12<<2)|2,{87,115,0}},
/* 58502 */ {(12<<2)|2,{87,115,0}},
/* 58503 */ {(12<<2)|2,{87,115,0}},
/* 58504 */ {(12<<2)|2,{87,115,0}},
/* 58505 */ {(12<<2)|2,{87,115,0}},
/* 58506 */ {(12<<2)|2,{87,115,0}},
/* 58507 */ {(12<<2)|2,{87,115,0}},
/* 58508 */ {(12<<2)|2,{87,115,0}},
/* 58509 */ {(12<<2)|2,{87,115,0}},
/* 58510 */ {(12<<2)|2,{87,115,0}},
/* 58511 */ {(12<<2)|2,{87,115,0}},
/* 58512 */ {(12<<2)|2,{87,116,0}},
/* 58513 */ {(12<<2)|2,{87,116,0}},
/* 58514 */ {(12<<2)|2,{87,116,0}},
/* 58515 */ {(12<<2)|2,{87,116,0}},
/* 58516 */ {(12<<2)|2,{87,116,0}},
/* 58517 */ {(12<<2)|2,{87,116,0}},
/* 58518 */ {(12<<2)|2,{87,116,0}},
/* 58519 */ {(12<<2)|2,{87,116,0}},
/* 58520 */ {(12<<2)|2,{87,116,0}},
/* 58521 */ {(12<<2)|2,{87,116,0}},
/* 58522 */ {(12<<2)|2,{87,116,0}},
/* 58523 */ {(12<<2)|2,{87,116,0}},
/* 58524 */ {(12<<2)|2,{87,116,0}},
/* 58525 */ {(12<<2)|2,{87,116,0}},
/* 58526 */ {(12<<2)|2,{87,116,0}},
/* 58527 */ {(12<<2)|2,{87,116,0}},
/* 58528 */ {(13<<2)|2,{87,32,0}},
/* 58529 */ {(13<<2)|2,{87,32,0}},
/* 58530 */ {(13<<2)|2,{87,32,0}},
/* 58531 */ {(13<<2)|2,{87,32,0}},
/* 58532 */ {(13<<2)|2,{87,32,0}},
/* 58533 */ {(13<<2)|2,{87,32,0}},
/* 58534 */ {(13<<2)|2,{87,32,0}},
/* 58535 */ {(13<<2)|2,{87,32,0}},
/* 58536 */ {(13<<2)|2,{87,37,0}},
/* 58537 */ {(13<<2)|2,{87,37,0}},
/* 58538 */ {(13<<2)|2,{87,37,0}},
/* 58539 */ {(13<<2)|2,{87,37,0}},
/* 58540 */ {(13<<2)|2,{87,37,0}},
/* 58541 */ {(13<<2)|2,{87,37,0}},
/* 58542 */ {(13<<2)|2,{87,37,0}},
/* 58543 */ {(13<<2)|2,{87,37,0}},
/* 58544 */ {(13<<2)|2,{87,45,0}},
/* 58545 */ {(13<<2)|2,{87,45,0}},
/* 58546 */ {(13<<2)|2,{87,45,0}},
/* 58547 */ {(13<<2)|2,{87,45,0}},
/* 58548 */ {(13<<2)|2,{87,45,0}},
/* 58549 */ {(13<<2)|2,{87,45,0}},
/* 58550 */ {(13<<2)|2,{87,45,0}},
/* 58551 */ {(13<<2)|2,{87,45,0}},
/* 58552 */ {(13<<2)|2,{87,46,0}},
/* 58553 */ {(13<<2)|2,{87,46,0}},
/* 58554 */ {(13<<2)|2,{87,46,0}},
/* 58555 */ {(13<<2)|2,{87,46,0}},
/* 58556 */ {(13<<2)|2,{87,46,0}},
/* 58557 */ {(13<<2)|2,{87,46,0}},
/* 58558 */ {(13<<2)|2,{87,46,0}},
/* 58559 */ {(13<<2)|2,{87,46,0}},
/* 58560 */ {(13<<2)|2,{87,47,0}},
/* 58561 */ {(13<<2)|2,{87,47,0}},
/* 58562 */ {(13<<2)|2,{87,47,0}},
/* 58563 */ {(13<<2)|2,{87,47,0}},
/* 58564 */ {(13<<2)|2,{87,47,0}},
/* 58565 */ {(13<<2)|2,{87,47,0}},
/* 58566 */ {(13<<2)|2,{87,47,0}},
/* 58567 */ {(13<<2)|2,{87,47,0}},
/* 58568 */ {(13<<2)|2,{87,51,0}},
/* 58569 */ {(13<<2)|2,{87,51,0}},
/* 58570 */ {(13<<2)|2,{87,51,0}},
/* 58571 */ {(13<<2)|2,{87,51,0}},
/* 58572 */ {(13<<2)|2,{87,51,0}},
/* 58573 */ {(13<<2)|2,{87,51,0}},
/* 58574 */ {(13<<2)|2,{87,51,0}},
/* 58575 */ {(13<<2)|2,{87,51,0}},
/* 58576 */ {(13<<2)|2,{87,52,0}},
/* 58577 */ {(13<<2)|2,{87,52,0}},
/* 58578 */ {(13<<2)|2,{87,52,0}},
/* 58579 */ {(13<<2)|2,{87,52,0}},
/* 58580 */ {(13<<2)|2,{87,52,0}},
/* 58581 */ {(13<<2)|2,{87,52,0}},
/* 58582 */ {(13<<2)|2,{87,52,0}},
/* 58583 */ {(13<<2)|2,{87,52,0}},
/* 58584 */ {(13<<2)|2,{87,53,0}},
/* 58585 */ {(13<<2)|2,{87,53,0}},
/* 58586 */ {(13<<2)|2,{87,53,0}},
/* 58587 */ {(13<<2)|2,{87,53,0}},
/* 58588 */ {(13<<2)|2,{87,53,0}},
/* 58589 */ {(13<<2)|2,{87,53,0}},
/* 58590 */ {(13<<2)|2,{87,53,0}},
/* 58591 */ {(13<<2)|2,{87,53,0}},
/* 58592 */ {(13<<2)|2,{87,54,0}},
/* 58593 */ {(13<<2)|2,{87,54,0}},
/* 58594 */ {(13<<2)|2,{87,54,0}},
/* 58595 */ {(13<<2)|2,{87,54,0}},
/* 58596 */ {(13<<2)|2,{87,54,0}},
/* 58597 */ {(13<<2)|2,{87,54,0}},
/* 58598 */ {(13<<2)|2,{87,54,0}},
/* 58599 */ {(13<<2)|2,{87,54,0}},
/* 58600 */ {(13<<2)|2,{87,55,0}},
/* 58601 */ {(13<<2)|2,{87,55,0}},
/* 58602 */ {(13<<2)|2,{87,55,0}},
/* 58603 */ {(13<<2)|2,{87,55,0}},
/* 58604 */ {(13<<2)|2,{87,55,0}},
/* 58605 */ {(13<<2)|2,{87,55,0}},
/* 58606 */ {(13<<2)|2,{87,55,0}},
/* 58607 */ {(13<<2)|2,{87,55,0}},
/* 58608 */ {(13<<2)|2,{87,56,0}},
/* 58609 */ {(13<<2)|2,{87,56,0}},
/* 58610 */ {(13<<2)|2,{87,56,0}},
/* 58611 */ {(13<<2)|2,{87,56,0}},
/* 58612 */ {(13<<2)|2,{87,56,0}},
/* 58613 */ {(13<<2)|2,{87,56,0}},
/* 58614 */ {(13<<2)|2,{87,56,0}},
/* 58615 */ {(13<<2)|2,{87,56,0}},
/* 58616 */ {(13<<2)|2,{87,57,0}},
/* 58617 */ {(13<<2)|2,{87,57,0}},
/* 58618 */ {(13<<2)|2,{87,57,0}},
/* 58619 */ {(13<<2)|2,{87,57,0}},
/* 58620 */ {(13<<2)|2,{87,57,0}},
/* 58621 */ {(13<<2)|2,{87,57,0}},
/* 58622 */ {(13<<2)|2,{87,57,0}},
/* 58623 */ {(13<<2)|2,{87,57,0}},
/* 58624 */ {(13<<2)|2,{87,61,0}},
/* 58625 */ {(13<<2)|2,{87,61,0}},
/* 58626 */ {(13<<2)|2,{87,61,0}},
/* 58627 */ {(13<<2)|2,{87,61,0}},
/* 58628 */ {(13<<2)|2,{87,61,0}},
/* 58629 */ {(13<<2)|2,{87,61,0}},
/* 58630 */ {(13<<2)|2,{87,61,0}},
/* 58631 */ {(13<<2)|2,{87,61,0}},
/* 58632 */ {(13<<2)|2,{87,65,0}},
/* 58633 */ {(13<<2)|2,{87,65,0}},
/* 58634 */ {(13<<2)|2,{87,65,0}},
/* 58635 */ {(13<<2)|2,{87,65,0}},
/* 58636 */ {(13<<2)|2,{87,65,0}},
/* 58637 */ {(13<<2)|2,{87,65,0}},
/* 58638 */ {(13<<2)|2,{87,65,0}},
/* 58639 */ {(13<<2)|2,{87,65,0}},
/* 58640 */ {(13<<2)|2,{87,95,0}},
/* 58641 */ {(13<<2)|2,{87,95,0}},
/* 58642 */ {(13<<2)|2,{87,95,0}},
/* 58643 */ {(13<<2)|2,{87,95,0}},
/* 58644 */ {(13<<2)|2,{87,95,0}},
/* 58645 */ {(13<<2)|2,{87,95,0}},
/* 58646 */ {(13<<2)|2,{87,95,0}},
/* 58647 */ {(13<<2)|2,{87,95,0}},
/* 58648 */ {(13<<2)|2,{87,98,0}},
/* 58649 */ {(13<<2)|2,{87,98,0}},
/* 58650 */ {(13<<2)|2,{87,98,0}},
/* 58651 */ {(13<<2)|2,{87,98,0}},
/* 58652 */ {(13<<2)|2,{87,98,0}},
/* 58653 */ {(13<<2)|2,{87,98,0}},
/* 58654 */ {(13<<2)|2,{87,98,0}},
/* 58655 */ {(13<<2)|2,{87,98,0}},
/* 58656 */ {(13<<2)|2,{87,100,0}},
/* 58657 */ {(13<<2)|2,{87,100,0}},
/* 58658 */ {(13<<2)|2,{87,100,0}},
/* 58659 */ {(13<<2)|2,{87,100,0}},
/* 58660 */ {(13<<2)|2,{87,100,0}},
/* 58661 */ {(13<<2)|2,{87,100,0}},
/* 58662 */ {(13<<2)|2,{87,100,0}},
/* 58663 */ {(13<<2)|2,{87,100,0}},
/* 58664 */ {(13<<2)|2,{87,102,0}},
/* 58665 */ {(13<<2)|2,{87,102,0}},
/* 58666 */ {(13<<2)|2,{87,102,0}},
/* 58667 */ {(13<<2)|2,{87,102,0}},
/* 58668 */ {(13<<2)|2,{87,102,0}},
/* 58669 */ {(13<<2)|2,{87,102,0}},
/* 58670 */ {(13<<2)|2,{87,102,0}},
/* 58671 */ {(13<<2)|2,{87,102,0}},
/* 58672 */ {(13<<2)|2,{87,103,0}},
/* 58673 */ {(13<<2)|2,{87,103,0}},
/* 58674 */ {(13<<2)|2,{87,103,0}},
/* 58675 */ {(13<<2)|2,{87,103,0}},
/* 58676 */ {(13<<2)|2,{87,103,0}},
/* 58677 */ {(13<<2)|2,{87,103,0}},
/* 58678 */ {(13<<2)|2,{87,103,0}},
/* 58679 */ {(13<<2)|2,{87,103,0}},
/* 58680 */ {(13<<2)|2,{87,104,0}},
/* 58681 */ {(13<<2)|2,{87,104,0}},
/* 58682 */ {(13<<2)|2,{87,104,0}},
/* 58683 */ {(13<<2)|2,{87,104,0}},
/* 58684 */ {(13<<2)|2,{87,104,0}},
/* 58685 */ {(13<<2)|2,{87,104,0}},
/* 58686 */ {(13<<2)|2,{87,104,0}},
/* 58687 */ {(13<<2)|2,{87,104,0}},
/* 58688 */ {(13<<2)|2,{87,108,0}},
/* 58689 */ {(13<<2)|2,{87,108,0}},
/* 58690 */ {(13<<2)|2,{87,108,0}},
/* 58691 */ {(13<<2)|2,{87,108,0}},
/* 58692 */ {(13<<2)|2,{87,108,0}},
/* 58693 */ {(13<<2)|2,{87,108,0}},
/* 58694 */ {(13<<2)|2,{87,108,0}},
/* 58695 */ {(13<<2)|2,{87,108,0}},
/* 58696 */ {(13<<2)|2,{87,109,0}},
/* 58697 */ {(13<<2)|2,{87,109,0}},
/* 58698 */ {(13<<2)|2,{87,109,0}},
/* 58699 */ {(13<<2)|2,{87,109,0}},
/* 58700 */ {(13<<2)|2,{87,109,0}},
/* 58701 */ {(13<<2)|2,{87,109,0}},
/* 58702 */ {(13<<2)|2,{87,109,0}},
/* 58703 */ {(13<<2)|2,{87,109,0}},
/* 58704 */ {(13<<2)|2,{87,110,0}},
/* 58705 */ {(13<<2)|2,{87,110,0}},
/* 58706 */ {(13<<2)|2,{87,110,0}},
/* 58707 */ {(13<<2)|2,{87,110,0}},
/* 58708 */ {(13<<2)|2,{87,110,0}},
/* 58709 */ {(13<<2)|2,{87,110,0}},
/* 58710 */ {(13<<2)|2,{87,110,0}},
/* 58711 */ {(13<<2)|2,{87,110,0}},
/* 58712 */ {(13<<2)|2,{87,112,0}},
/* 58713 */ {(13<<2)|2,{87,112,0}},
/* 58714 */ {(13<<2)|2,{87,112,0}},
/* 58715 */ {(13<<2)|2,{87,112,0}},
/* 58716 */ {(13<<2)|2,{87,112,0}},
/* 58717 */ {(13<<2)|2,{87,112,0}},
/* 58718 */ {(13<<2)|2,{87,112,0}},
/* 58719 */ {(13<<2)|2,{87,112,0}},
/* 58720 */ {(13<<2)|2,{87,114,0}},
/* 58721 */ {(13<<2)|2,{87,114,0}},
/* 58722 */ {(13<<2)|2,{87,114,0}},
/* 58723 */ {(13<<2)|2,{87,114,0}},
/* 58724 */ {(13<<2)|2,{87,114,0}},
/* 58725 */ {(13<<2)|2,{87,114,0}},
/* 58726 */ {(13<<2)|2,{87,114,0}},
/* 58727 */ {(13<<2)|2,{87,114,0}},
/* 58728 */ {(13<<2)|2,{87,117,0}},
/* 58729 */ {(13<<2)|2,{87,117,0}},
/* 58730 */ {(13<<2)|2,{87,117,0}},
/* 58731 */ {(13<<2)|2,{87,117,0}},
/* 58732 */ {(13<<2)|2,{87,117,0}},
/* 58733 */ {(13<<2)|2,{87,117,0}},
/* 58734 */ {(13<<2)|2,{87,117,0}},
/* 58735 */ {(13<<2)|2,{87,117,0}},
/* 58736 */ {(14<<2)|2,{87,58,0}},
/* 58737 */ {(14<<2)|2,{87,58,0}},
/* 58738 */ {(14<<2)|2,{87,58,0}},
/* 58739 */ {(14<<2)|2,{87,58,0}},
/* 58740 */ {(14<<2)|2,{87,66,0}},
/* 58741 */ {(14<<2)|2,{87,66,0}},
/* 58742 */ {(14<<2)|2,{87,66,0}},
/* 58743 */ {(14<<2)|2,{87,66,0}},
/* 58744 */ {(14<<2)|2,{87,67,0}},
/* 58745 */ {(14<<2)|2,{87,67,0}},
/* 58746 */ {(14<<2)|2,{87,67,0}},
/* 58747 */ {(14<<2)|2,{87,67,0}},
/* 58748 */ {(14<<2)|2,{87,68,0}},
/* 58749 */ {(14<<2)|2,{87,68,0}},
/* 58750 */ {(14<<2)|2,{87,68,0}},
/* 58751 */ {(14<<2)|2,{87,68,0}},
/* 58752 */ {(14<<2)|2,{87,69,0}},
/* 58753 */ {(14<<2)|2,{87,69,0}},
/* 58754 */ {(14<<2)|2,{87,69,0}},
/* 58755 */ {(14<<2)|2,{87,69,0}},
/* 58756 */ {(14<<2)|2,{87,70,0}},
/* 58757 */ {(14<<2)|2,{87,70,0}},
/* 58758 */ {(14<<2)|2,{87,70,0}},
/* 58759 */ {(14<<2)|2,{87,70,0}},
/* 58760 */ {(14<<2)|2,{87,71,0}},
/* 58761 */ {(14<<2)|2,{87,71,0}},
/* 58762 */ {(14<<2)|2,{87,71,0}},
/* 58763 */ {(14<<2)|2,{87,71,0}},
/* 58764 */ {(14<<2)|2,{87,72,0}},
/* 58765 */ {(14<<2)|2,{87,72,0}},
/* 58766 */ {(14<<2)|2,{87,72,0}},
/* 58767 */ {(14<<2)|2,{87,72,0}},
/* 58768 */ {(14<<2)|2,{87,73,0}},
/* 58769 */ {(14<<2)|2,{87,73,0}},
/* 58770 */ {(14<<2)|2,{87,73,0}},
/* 58771 */ {(14<<2)|2,{87,73,0}},
/* 58772 */ {(14<<2)|2,{87,74,0}},
/* 58773 */ {(14<<2)|2,{87,74,0}},
/* 58774 */ {(14<<2)|2,{87,74,0}},
/* 58775 */ {(14<<2)|2,{87,74,0}},
/* 58776 */ {(14<<2)|2,{87,75,0}},
/* 58777 */ {(14<<2)|2,{87,75,0}},
/* 58778 */ {(14<<2)|2,{87,75,0}},
/* 58779 */ {(14<<2)|2,{87,75,0}},
/* 58780 */ {(14<<2)|2,{87,76,0}},
/* 58781 */ {(14<<2)|2,{87,76,0}},
/* 58782 */ {(14<<2)|2,{87,76,0}},
/* 58783 */ {(14<<2)|2,{87,76,0}},
/* 58784 */ {(14<<2)|2,{87,77,0}},
/* 58785 */ {(14<<2)|2,{87,77,0}},
/* 58786 */ {(14<<2)|2,{87,77,0}},
/* 58787 */ {(14<<2)|2,{87,77,0}},
/* 58788 */ {(14<<2)|2,{87,78,0}},
/* 58789 */ {(14<<2)|2,{87,78,0}},
/* 58790 */ {(14<<2)|2,{87,78,0}},
/* 58791 */ {(14<<2)|2,{87,78,0}},
/* 58792 */ {(14<<2)|2,{87,79,0}},
/* 58793 */ {(14<<2)|2,{87,79,0}},
/* 58794 */ {(14<<2)|2,{87,79,0}},
/* 58795 */ {(14<<2)|2,{87,79,0}},
/* 58796 */ {(14<<2)|2,{87,80,0}},
/* 58797 */ {(14<<2)|2,{87,80,0}},
/* 58798 */ {(14<<2)|2,{87,80,0}},
/* 58799 */ {(14<<2)|2,{87,80,0}},
/* 58800 */ {(14<<2)|2,{87,81,0}},
/* 58801 */ {(14<<2)|2,{87,81,0}},
/* 58802 */ {(14<<2)|2,{87,81,0}},
/* 58803 */ {(14<<2)|2,{87,81,0}},
/* 58804 */ {(14<<2)|2,{87,82,0}},
/* 58805 */ {(14<<2)|2,{87,82,0}},
/* 58806 */ {(14<<2)|2,{87,82,0}},
/* 58807 */ {(14<<2)|2,{87,82,0}},
/* 58808 */ {(14<<2)|2,{87,83,0}},
/* 58809 */ {(14<<2)|2,{87,83,0}},
/* 58810 */ {(14<<2)|2,{87,83,0}},
/* 58811 */ {(14<<2)|2,{87,83,0}},
/* 58812 */ {(14<<2)|2,{87,84,0}},
/* 58813 */ {(14<<2)|2,{87,84,0}},
/* 58814 */ {(14<<2)|2,{87,84,0}},
/* 58815 */ {(14<<2)|2,{87,84,0}},
/* 58816 */ {(14<<2)|2,{87,85,0}},
/* 58817 */ {(14<<2)|2,{87,85,0}},
/* 58818 */ {(14<<2)|2,{87,85,0}},
/* 58819 */ {(14<<2)|2,{87,85,0}},
/* 58820 */ {(14<<2)|2,{87,86,0}},
/* 58821 */ {(14<<2)|2,{87,86,0}},
/* 58822 */ {(14<<2)|2,{87,86,0}},
/* 58823 */ {(14<<2)|2,{87,86,0}},
/* 58824 */ {(14<<2)|2,{87,87,0}},
/* 58825 */ {(14<<2)|2,{87,87,0}},
/* 58826 */ {(14<<2)|2,{87,87,0}},
/* 58827 */ {(14<<2)|2,{87,87,0}},
/* 58828 */ {(14<<2)|2,{87,89,0}},
/* 58829 */ {(14<<2)|2,{87,89,0}},
/* 58830 */ {(14<<2)|2,{87,89,0}},
/* 58831 */ {(14<<2)|2,{87,89,0}},
/* 58832 */ {(14<<2)|2,{87,106,0}},
/* 58833 */ {(14<<2)|2,{87,106,0}},
/* 58834 */ {(14<<2)|2,{87,106,0}},
/* 58835 */ {(14<<2)|2,{87,106,0}},
/* 58836 */ {(14<<2)|2,{87,107,0}},
/* 58837 */ {(14<<2)|2,{87,107,0}},
/* 58838 */ {(14<<2)|2,{87,107,0}},
/* 58839 */ {(14<<2)|2,{87,107,0}},
/* 58840 */ {(14<<2)|2,{87,113,0}},
/* 58841 */ {(14<<2)|2,{87,113,0}},
/* 58842 */ {(14<<2)|2,{87,113,0}},
/* 58843 */ {(14<<2)|2,{87,113,0}},
/* 58844 */ {(14<<2)|2,{87,118,0}},
/* 58845 */ {(14<<2)|2,{87,118,0}},
/* 58846 */ {(14<<2)|2,{87,118,0}},
/* 58847 */ {(14<<2)|2,{87,118,0}},
/* 58848 */ {(14<<2)|2,{87,119,0}},
/* 58849 */ {(14<<2)|2,{87,119,0}},
/* 58850 */ {(14<<2)|2,{87,119,0}},
/* 58851 */ {(14<<2)|2,{87,119,0}},
/* 58852 */ {(14<<2)|2,{87,120,0}},
/* 58853 */ {(14<<2)|2,{87,120,0}},
/* 58854 */ {(14<<2)|2,{87,120,0}},
/* 58855 */ {(14<<2)|2,{87,120,0}},
/* 58856 */ {(14<<2)|2,{87,121,0}},
/* 58857 */ {(14<<2)|2,{87,121,0}},
/* 58858 */ {(14<<2)|2,{87,121,0}},
/* 58859 */ {(14<<2)|2,{87,121,0}},
/* 58860 */ {(14<<2)|2,{87,122,0}},
/* 58861 */ {(14<<2)|2,{87,122,0}},
/* 58862 */ {(14<<2)|2,{87,122,0}},
/* 58863 */ {(14<<2)|2,{87,122,0}},
/* 58864 */ {(15<<2)|2,{87,38,0}},
/* 58865 */ {(15<<2)|2,{87,38,0}},
/* 58866 */ {(15<<2)|2,{87,42,0}},
/* 58867 */ {(15<<2)|2,{87,42,0}},
/* 58868 */ {(15<<2)|2,{87,44,0}},
/* 58869 */ {(15<<2)|2,{87,44,0}},
/* 58870 */ {(15<<2)|2,{87,59,0}},
/* 58871 */ {(15<<2)|2,{87,59,0}},
/* 58872 */ {(15<<2)|2,{87,88,0}},
/* 58873 */ {(15<<2)|2,{87,88,0}},
/* 58874 */ {(15<<2)|2,{87,90,0}},
/* 58875 */ {(15<<2)|2,{87,90,0}},
/* 58876 */ {(7<<2)|1,{87,0,0}},
/* 58877 */ {(7<<2)|1,{87,0,0}},
/* 58878 */ {(7<<2)|1,{87,0,0}},
/* 58879 */ {(7<<2)|1,{87,0,0}},
/* 58880 */ {(12<<2)|2,{89,48,0}},
/* 58881 */ {(12<<2)|2,{89,48,0}},
/* 58882 */ {(12<<2)|2,{89,48,0}},
/* 58883 */ {(12<<2)|2,{89,48,0}},
/* 58884 */ {(12<<2)|2,{89,48,0}},
/* 58885 */ {(12<<2)|2,{89,48,0}},
/* 58886 */ {(12<<2)|2,{89,48,0}},
/* 58887 */ {(12<<2)|2,{89,48,0}},
/* 58888 */ {(12<<2)|2,{89,48,0}},
/* 58889 */ {(12<<2)|2,{89,48,0}},
/* 58890 */ {(12<<2)|2,{89,48,0}},
/* 58891 */ {(12<<2)|2,{89,48,0}},
/* 58892 */ {(12<<2)|2,{89,48,0}},
/* 58893 */ {(12<<2)|2,{89,48,0}},
/* 58894 */ {(12<<2)|2,{89,48,0}},
/* 58895 */ {(12<<2)|2,{89,48,0}},
/* 58896 */ {(12<<2)|2,{89,49,0}},
/* 58897 */ {(12<<2)|2,{89,49,0}},
/* 58898 */ {(12<<2)|2,{89,49,0}},
/* 58899 */ {(12<<2)|2,{89,49,0}},
/* 58900 */ {(12<<2)|2,{89,49,0}},
/* 58901 */ {(12<<2)|2,{89,49,0}},
/* 58902 */ {(12<<2)|2,{89,49,0}},
/* 58903 */ {(12<<2)|2,{89,49,0}},
/* 58904 */ {(12<<2)|2,{89,49,0}},
/* 58905 */ {(12<<2)|2,{89,49,0}},
/* 58906 */ {(12<<2)|2,{89,49,0}},
/* 58907 */ {(12<<2)|2,{89,49,0}},
/* 58908 */ {(12<<2)|2,{89,49,0}},
/* 58909 */ {(12<<2)|2,{89,49,0}},
/* 58910 */ {(12<<2)|2,{89,49,0}},
/* 58911 */ {(12<<2)|2,{89,49,0}},
/* 58912 */ {(12<<2)|2,{89,50,0}},
/* 58913 */ {(12<<2)|2,{89,50,0}},
/* 58914 */ {(12<<2)|2,{89,50,0}},
/* 58915 */ {(12<<2)|2,{89,50,0}},
/* 58916 */ {(12<<2)|2,{89,50,0}},
/* 58917 */ {(12<<2)|2,{89,50,0}},
/* 58918 */ {(12<<2)|2,{89,50,0}},
/* 58919 */ {(12<<2)|2,{89,50,0}},
/* 58920 */ {(12<<2)|2,{89,50,0}},
/* 58921 */ {(12<<2)|2,{89,50,0}},
/* 58922 */ {(12<<2)|2,{89,50,0}},
/* 58923 */ {(12<<2)|2,{89,50,0}},
/* 58924 */ {(12<<2)|2,{89,50,0}},
/* 58925 */ {(12<<2)|2,{89,50,0}},
/* 58926 */ {(12<<2)|2,{89,50,0}},
/* 58927 */ {(12<<2)|2,{89,50,0}},
/* 58928 */ {(12<<2)|2,{89,97,0}},
/* 58929 */ {(12<<2)|2,{89,97,0}},
/* 58930 */ {(12<<2)|2,{89,97,0}},
/* 58931 */ {(12<<2)|2,{89,97,0}},
/* 58932 */ {(12<<2)|2,{89,97,0}},
/* 58933 */ {(12<<2)|2,{89,97,0}},
/* 58934 */ {(12<<2)|2,{89,97,0}},
/* 58935 */ {(12<<2)|2,{89,97,0}},
/* 58936 */ {(12<<2)|2,{89,97,0}},
/* 58937 */ {(12<<2)|2,{89,97,0}},
/* 58938 */ {(12<<2)|2,{89,97,0}},
/* 58939 */ {(12<<2)|2,{89,97,0}},
/* 58940 */ {(12<<2)|2,{89,97,0}},
/* 58941 */ {(12<<2)|2,{89,97,0}},
/* 58942 */ {(12<<2)|2,{89,97,0}},
/* 58943 */ {(12<<2)|2,{89,97,0}},
/* 58944 */ {(12<<2)|2,{89,99,0}},
/* 58945 */ {(12<<2)|2,{89,99,0}},
/* 58946 */ {(12<<2)|2,{89,99,0}},
/* 58947 */ {(12<<2)|2,{89,99,0}},
/* 58948 */ {(12<<2)|2,{89,99,0}},
/* 58949 */ {(12<<2)|2,{89,99,0}},
/* 58950 */ {(12<<2)|2,{89,99,0}},
/* 58951 */ {(12<<2)|2,{89,99,0}},
/* 58952 */ {(12<<2)|2,{89,99,0}},
/* 58953 */ {(12<<2)|2,{89,99,0}},
/* 58954 */ {(12<<2)|2,{89,99,0}},
/* 58955 */ {(12<<2)|2,{89,99,0}},
/* 58956 */ {(12<<2)|2,{89,99,0}},
/* 58957 */ {(12<<2)|2,{89,99,0}},
/* 58958 */ {(12<<2)|2,{89,99,0}},
/* 58959 */ {(12<<2)|2,{89,99,0}},
/* 58960 */ {(12<<2)|2,{89,101,0}},
/* 58961 */ {(12<<2)|2,{89,101,0}},
/* 58962 */ {(12<<2)|2,{89,101,0}},
/* 58963 */ {(12<<2)|2,{89,101,0}},
/* 58964 */ {(12<<2)|2,{89,101,0}},
/* 58965 */ {(12<<2)|2,{89,101,0}},
/* 58966 */ {(12<<2)|2,{89,101,0}},
/* 58967 */ {(12<<2)|2,{89,101,0}},
/* 58968 */ {(12<<2)|2,{89,101,0}},
/* 58969 */ {(12<<2)|2,{89,101,0}},
/* 58970 */ {(12<<2)|2,{89,101,0}},
/* 58971 */ {(12<<2)|2,{89,101,0}},
/* 58972 */ {(12<<2)|2,{89,101,0}},
/* 58973 */ {(12<<2)|2,{89,101,0}},
/* 58974 */ {(12<<2)|2,{89,101,0}},
/* 58975 */ {(12<<2)|2,{89,101,0}},
/* 58976 */ {(12<<2)|2,{89,105,0}},
/* 58977 */ {(12<<2)|2,{89,105,0}},
/* 58978 */ {(12<<2)|2,{89,105,0}},
/* 58979 */ {(12<<2)|2,{89,105,0}},
/* 58980 */ {(12<<2)|2,{89,105,0}},
/* 58981 */ {(12<<2)|2,{89,105,0}},
/* 58982 */ {(12<<2)|2,{89,105,0}},
/* 58983 */ {(12<<2)|2,{89,105,0}},
/* 58984 */ {(12<<2)|2,{89,105,0}},
/* 58985 */ {(12<<2)|2,{89,105,0}},
/* 58986 */ {(12<<2)|2,{89,105,0}},
/* 58987 */ {(12<<2)|2,{89,105,0}},
/* 58988 */ {(12<<2)|2,{89,105,0}},
/* 58989 */ {(12<<2)|2,{89,105,0}},
/* 58990 */ {(12<<2)|2,{89,105,0}},
/* 58991 */ {(12<<2)|2,{89,105,0}},
/* 58992 */ {(12<<2)|2,{89,111,0}},
/* 58993 */ {(12<<2)|2,{89,111,0}},
/* 58994 */ {(12<<2)|2,{89,111,0}},
/* 58995 */ {(12<<2)|2,{89,111,0}},
/* 58996 */ {(12<<2)|2,{89,111,0}},
/* 58997 */ {(12<<2)|2,{89,111,0}},
/* 58998 */ {(12<<2)|2,{89,111,0}},
/* 58999 */ {(12<<2)|2,{89,111,0}},
/* 59000 */ {(12<<2)|2,{89,111,0}},
/* 59001 */ {(12<<2)|2,{89,111,0}},
/* 59002 */ {(12<<2)|2,{89,111,0}},
/* 59003 */ {(12<<2)|2,{89,111,0}},
/* 59004 */ {(12<<2)|2,{89,111,0}},
/* 59005 */ {(12<<2)|2,{89,111,0}},
/* 59006 */ {(12<<2)|2,{89,111,0}},
/* 59007 */ {(12<<2)|2,{89,111,0}},
/* 59008 */ {(12<<2)|2,{89,115,0}},
/* 59009 */ {(12<<2)|2,{89,115,0}},
/* 59010 */ {(12<<2)|2,{89,115,0}},
/* 59011 */ {(12<<2)|2,{89,115,0}},
/* 59012 */ {(12<<2)|2,{89,115,0}},
/* 59013 */ {(12<<2)|2,{89,115,0}},
/* 59014 */ {(12<<2)|2,{89,115,0}},
/* 59015 */ {(12<<2)|2,{89,115,0}},
/* 59016 */ {(12<<2)|2,{89,115,0}},
/* 59017 */ {(12<<2)|2,{89,115,0}},
/* 59018 */ {(12<<2)|2,{89,115,0}},
/* 59019 */ {(12<<2)|2,{89,115,0}},
/* 59020 */ {(12<<2)|2,{89,115,0}},
/* 59021 */ {(12<<2)|2,{89,115,0}},
/* 59022 */ {(12<<2)|2,{89,115,0}},
/* 59023 */ {(12<<2)|2,{89,115,0}},
/* 59024 */ {(12<<2)|2,{89,116,0}},
/* 59025 */ {(12<<2)|2,{89,116,0}},
/* 59026 */ {(12<<2)|2,{89,116,0}},
/* 59027 */ {(12<<2)|2,{89,116,0}},
/* 59028 */ {(12<<2)|2,{89,116,0}},
/* 59029 */ {(12<<2)|2,{89,116,0}},
/* 59030 */ {(12<<2)|2,{89,116,0}},
/* 59031 */ {(12<<2)|2,{89,116,0}},
/* 59032 */ {(12<<2)|2,{89,116,0}},
/* 59033 */ {(12<<2)|2,{89,116,0}},
/* 59034 */ {(12<<2)|2,{89,116,0}},
/* 59035 */ {(12<<2)|2,{89,116,0}},
/* 59036 */ {(12<<2)|2,{89,116,0}},
/* 59037 */ {(12<<2)|2,{89,116,0}},
/* 59038 */ {(12<<2)|2,{89,116,0}},
/* 59039 */ {(12<<2)|2,{89,116,0}},
/* 59040 */ {(13<<2)|2,{89,32,0}},
/* 59041 */ {(13<<2)|2,{89,32,0}},
/* 59042 */ {(13<<2)|2,{89,32,0}},
/* 59043 */ {(13<<2)|2,{89,32,0}},
/* 59044 */ {(13<<2)|2,{89,32,0}},
/* 59045 */ {(13<<2)|2,{89,32,0}},
/* 59046 */ {(13<<2)|2,{89,32,0}},
/* 59047 */ {(13<<2)|2,{89,32,0}},
/* 59048 */ {(13<<2)|2,{89,37,0}},
/* 59049 */ {(13<<2)|2,{89,37,0}},
/* 59050 */ {(13<<2)|2,{89,37,0}},
/* 59051 */ {(13<<2)|2,{89,37,0}},
/* 59052 */ {(13<<2)|2,{89,37,0}},
/* 59053 */ {(13<<2)|2,{89,37,0}},
/* 59054 */ {(13<<2)|2,{89,37,0}},
/* 59055 */ {(13<<2)|2,{89,37,0}},
/* 59056 */ {(13<<2)|2,{89,45,0}},
/* 59057 */ {(13<<2)|2,{89,45,0}},
/* 59058 */ {(13<<2)|2,{89,45,0}},
/* 59059 */ {(13<<2)|2,{89,45,0}},
/* 59060 */ {(13<<2)|2,{89,45,0}},
/* 59061 */ {(13<<2)|2,{89,45,0}},
/* 59062 */ {(13<<2)|2,{89,45,0}},
/* 59063 */ {(13<<2)|2,{89,45,0}},
/* 59064 */ {(13<<2)|2,{89,46,0}},
/* 59065 */ {(13<<2)|2,{89,46,0}},
/* 59066 */ {(13<<2)|2,{89,46,0}},
/* 59067 */ {(13<<2)|2,{89,46,0}},
/* 59068 */ {(13<<2)|2,{89,46,0}},
/* 59069 */ {(13<<2)|2,{89,46,0}},
/* 59070 */ {(13<<2)|2,{89,46,0}},
/* 59071 */ {(13<<2)|2,{89,46,0}},
/* 59072 */ {(13<<2)|2,{89,47,0}},
/* 59073 */ {(13<<2)|2,{89,47,0}},
/* 59074 */ {(13<<2)|2,{89,47,0}},
/* 59075 */ {(13<<2)|2,{89,47,0}},
/* 59076 */ {(13<<2)|2,{89,47,0}},
/* 59077 */ {(13<<2)|2,{89,47,0}},
/* 59078 */ {(13<<2)|2,{89,47,0}},
/* 59079 */ {(13<<2)|2,{89,47,0}},
/* 59080 */ {(13<<2)|2,{89,51,0}},
/* 59081 */ {(13<<2)|2,{89,51,0}},
/* 59082 */ {(13<<2)|2,{89,51,0}},
/* 59083 */ {(13<<2)|2,{89,51,0}},
/* 59084 */ {(13<<2)|2,{89,51,0}},
/* 59085 */ {(13<<2)|2,{89,51,0}},
/* 59086 */ {(13<<2)|2,{89,51,0}},
/* 59087 */ {(13<<2)|2,{89,51,0}},
/* 59088 */ {(13<<2)|2,{89,52,0}},
/* 59089 */ {(13<<2)|2,{89,52,0}},
/* 59090 */ {(13<<2)|2,{89,52,0}},
/* 59091 */ {(13<<2)|2,{89,52,0}},
/* 59092 */ {(13<<2)|2,{89,52,0}},
/* 59093 */ {(13<<2)|2,{89,52,0}},
/* 59094 */ {(13<<2)|2,{89,52,0}},
/* 59095 */ {(13<<2)|2,{89,52,0}},
/* 59096 */ {(13<<2)|2,{89,53,0}},
/* 59097 */ {(13<<2)|2,{89,53,0}},
/* 59098 */ {(13<<2)|2,{89,53,0}},
/* 59099 */ {(13<<2)|2,{89,53,0}},
/* 59100 */ {(13<<2)|2,{89,53,0}},
/* 59101 */ {(13<<2)|2,{89,53,0}},
/* 59102 */ {(13<<2)|2,{89,53,0}},
/* 59103 */ {(13<<2)|2,{89,53,0}},
/* 59104 */ {(13<<2)|2,{89,54,0}},
/* 59105 */ {(13<<2)|2,{89,54,0}},
/* 59106 */ {(13<<2)|2,{89,54,0}},
/* 59107 */ {(13<<2)|2,{89,54,0}},
/* 59108 */ {(13<<2)|2,{89,54,0}},
/* 59109 */ {(13<<2)|2,{89,54,0}},
/* 59110 */ {(13<<2)|2,{89,54,0}},
/* 59111 */ {(13<<2)|2,{89,54,0}},
/* 59112 */ {(13<<2)|2,{89,55,0}},
/* 59113 */ {(13<<2)|2,{89,55,0}},
/* 59114 */ {(13<<2)|2,{89,55,0}},
/* 59115 */ {(13<<2)|2,{89,55,0}},
/* 59116 */ {(13<<2)|2,{89,55,0}},
/* 59117 */ {(13<<2)|2,{89,55,0}},
/* 59118 */ {(13<<2)|2,{89,55,0}},
/* 59119 */ {(13<<2)|2,{89,55,0}},
/* 59120 */ {(13<<2)|2,{89,56,0}},
/* 59121 */ {(13<<2)|2,{89,56,0}},
/* 59122 */ {(13<<2)|2,{89,56,0}},
/* 59123 */ {(13<<2)|2,{89,56,0}},
/* 59124 */ {(13<<2)|2,{89,56,0}},
/* 59125 */ {(13<<2)|2,{89,56,0}},
/* 59126 */ {(13<<2)|2,{89,56,0}},
/* 59127 */ {(13<<2)|2,{89,56,0}},
/* 59128 */ {(13<<2)|2,{89,57,0}},
/* 59129 */ {(13<<2)|2,{89,57,0}},
/* 59130 */ {(13<<2)|2,{89,57,0}},
/* 59131 */ {(13<<2)|2,{89,57,0}},
/* 59132 */ {(13<<2)|2,{89,57,0}},
/* 59133 */ {(13<<2)|2,{89,57,0}},
/* 59134 */ {(13<<2)|2,{89,57,0}},
/* 59135 */ {(13<<2)|2,{89,57,0}},
/* 59136 */ {(13<<2)|2,{89,61,0}},
/* 59137 */ {(13<<2)|2,{89,61,0}},
/* 59138 */ {(13<<2)|2,{89,61,0}},
/* 59139 */ {(13<<2)|2,{89,61,0}},
/* 59140 */ {(13<<2)|2,{89,61,0}},
/* 59141 */ {(13<<2)|2,{89,61,0}},
/* 59142 */ {(13<<2)|2,{89,61,0}},
/* 59143 */ {(13<<2)|2,{89,61,0}},
/* 59144 */ {(13<<2)|2,{89,65,0}},
/* 59145 */ {(13<<2)|2,{89,65,0}},
/* 59146 */ {(13<<2)|2,{89,65,0}},
/* 59147 */ {(13<<2)|2,{89,65,0}},
/* 59148 */ {(13<<2)|2,{89,65,0}},
/* 59149 */ {(13<<2)|2,{89,65,0}},
/* 59150 */ {(13<<2)|2,{89,65,0}},
/* 59151 */ {(13<<2)|2,{89,65,0}},
/* 59152 */ {(13<<2)|2,{89,95,0}},
/* 59153 */ {(13<<2)|2,{89,95,0}},
/* 59154 */ {(13<<2)|2,{89,95,0}},
/* 59155 */ {(13<<2)|2,{89,95,0}},
/* 59156 */ {(13<<2)|2,{89,95,0}},
/* 59157 */ {(13<<2)|2,{89,95,0}},
/* 59158 */ {(13<<2)|2,{89,95,0}},
/* 59159 */ {(13<<2)|2,{89,95,0}},
/* 59160 */ {(13<<2)|2,{89,98,0}},
/* 59161 */ {(13<<2)|2,{89,98,0}},
/* 59162 */ {(13<<2)|2,{89,98,0}},
/* 59163 */ {(13<<2)|2,{89,98,0}},
/* 59164 */ {(13<<2)|2,{89,98,0}},
/* 59165 */ {(13<<2)|2,{89,98,0}},
/* 59166 */ {(13<<2)|2,{89,98,0}},
/* 59167 */ {(13<<2)|2,{89,98,0}},
/* 59168 */ {(13<<2)|2,{89,100,0}},
/* 59169 */ {(13<<2)|2,{89,100,0}},
/* 59170 */ {(13<<2)|2,{89,100,0}},
/* 59171 */ {(13<<2)|2,{89,100,0}},
/* 59172 */ {(13<<2)|2,{89,100,0}},
/* 59173 */ {(13<<2)|2,{89,100,0}},
/* 59174 */ {(13<<2)|2,{89,100,0}},
/* 59175 */ {(13<<2)|2,{89,100,0}},
/* 59176 */ {(13<<2)|2,{89,102,0}},
/* 59177 */ {(13<<2)|2,{89,102,0}},
/* 59178 */ {(13<<2)|2,{89,102,0}},
/* 59179 */ {(13<<2)|2,{89,102,0}},
/* 59180 */ {(13<<2)|2,{89,102,0}},
/* 59181 */ {(13<<2)|2,{89,102,0}},
/* 59182 */ {(13<<2)|2,{89,102,0}},
/* 59183 */ {(13<<2)|2,{89,102,0}},
/* 59184 */ {(13<<2)|2,{89,103,0}},
/* 59185 */ {(13<<2)|2,{89,103,0}},
/* 59186 */ {(13<<2)|2,{89,103,0}},
/* 59187 */ {(13<<2)|2,{89,103,0}},
/* 59188 */ {(13<<2)|2,{89,103,0}},
/* 59189 */ {(13<<2)|2,{89,103,0}},
/* 59190 */ {(13<<2)|2,{89,103,0}},
/* 59191 */ {(13<<2)|2,{89,103,0}},
/* 59192 */ {(13<<2)|2,{89,104,0}},
/* 59193 */ {(13<<2)|2,{89,104,0}},
/* 59194 */ {(13<<2)|2,{89,104,0}},
/* 59195 */ {(13<<2)|2,{89,104,0}},
/* 59196 */ {(13<<2)|2,{89,104,0}},
/* 59197 */ {(13<<2)|2,{89,104,0}},
/* 59198 */ {(13<<2)|2,{89,104,0}},
/* 59199 */ {(13<<2)|2,{89,104,0}},
/* 59200 */ {(13<<2)|2,{89,108,0}},
/* 59201 */ {(13<<2)|2,{89,108,0}},
/* 59202 */ {(13<<2)|2,{89,108,0}},
/* 59203 */ {(13<<2)|2,{89,108,0}},
/* 59204 */ {(13<<2)|2,{89,108,0}},
/* 59205 */ {(13<<2)|2,{89,108,0}},
/* 59206 */ {(13<<2)|2,{89,108,0}},
/* 59207 */ {(13<<2)|2,{89,108,0}},
/* 59208 */ {(13<<2)|2,{89,109,0}},
/* 59209 */ {(13<<2)|2,{89,109,0}},
/* 59210 */ {(13<<2)|2,{89,109,0}},
/* 59211 */ {(13<<2)|2,{89,109,0}},
/* 59212 */ {(13<<2)|2,{89,109,0}},
/* 59213 */ {(13<<2)|2,{89,109,0}},
/* 59214 */ {(13<<2)|2,{89,109,0}},
/* 59215 */ {(13<<2)|2,{89,109,0}},
/* 59216 */ {(13<<2)|2,{89,110,0}},
/* 59217 */ {(13<<2)|2,{89,110,0}},
/* 59218 */ {(13<<2)|2,{89,110,0}},
/* 59219 */ {(13<<2)|2,{89,110,0}},
/* 59220 */ {(13<<2)|2,{89,110,0}},
/* 59221 */ {(13<<2)|2,{89,110,0}},
/* 59222 */ {(13<<2)|2,{89,110,0}},
/* 59223 */ {(13<<2)|2,{89,110,0}},
/* 59224 */ {(13<<2)|2,{89,112,0}},
/* 59225 */ {(13<<2)|2,{89,112,0}},
/* 59226 */ {(13<<2)|2,{89,112,0}},
/* 59227 */ {(13<<2)|2,{89,112,0}},
/* 59228 */ {(13<<2)|2,{89,112,0}},
/* 59229 */ {(13<<2)|2,{89,112,0}},
/* 59230 */ {(13<<2)|2,{89,112,0}},
/* 59231 */ {(13<<2)|2,{89,112,0}},
/* 59232 */ {(13<<2)|2,{89,114,0}},
/* 59233 */ {(13<<2)|2,{89,114,0}},
/* 59234 */ {(13<<2)|2,{89,114,0}},
/* 59235 */ {(13<<2)|2,{89,114,0}},
/* 59236 */ {(13<<2)|2,{89,114,0}},
/* 59237 */ {(13<<2)|2,{89,114,0}},
/* 59238 */ {(13<<2)|2,{89,114,0}},
/* 59239 */ {(13<<2)|2,{89,114,0}},
/* 59240 */ {(13<<2)|2,{89,117,0}},
/* 59241 */ {(13<<2)|2,{89,117,0}},
/* 59242 */ {(13<<2)|2,{89,117,0}},
/* 59243 */ {(13<<2)|2,{89,117,0}},
/* 59244 */ {(13<<2)|2,{89,117,0}},
/* 59245 */ {(13<<2)|2,{89,117,0}},
/* 59246 */ {(13<<2)|2,{89,117,0}},
/* 59247 */ {(13<<2)|2,{89,117,0}},
/* 59248 */ {(14<<2)|2,{89,58,0}},
/* 59249 */ {(14<<2)|2,{89,58,0}},
/* 59250 */ {(14<<2)|2,{89,58,0}},
/* 59251 */ {(14<<2)|2,{89,58,0}},
/* 59252 */ {(14<<2)|2,{89,66,0}},
/* 59253 */ {(14<<2)|2,{89,66,0}},
/* 59254 */ {(14<<2)|2,{89,66,0}},
/* 59255 */ {(14<<2)|2,{89,66,0}},
/* 59256 */ {(14<<2)|2,{89,67,0}},
/* 59257 */ {(14<<2)|2,{89,67,0}},
/* 59258 */ {(14<<2)|2,{89,67,0}},
/* 59259 */ {(14<<2)|2,{89,67,0}},
/* 59260 */ {(14<<2)|2,{89,68,0}},
/* 59261 */ {(14<<2)|2,{89,68,0}},
/* 59262 */ {(14<<2)|2,{89,68,0}},
/* 59263 */ {(14<<2)|2,{89,68,0}},
/* 59264 */ {(14<<2)|2,{89,69,0}},
/* 59265 */ {(14<<2)|2,{89,69,0}},
/* 59266 */ {(14<<2)|2,{89,69,0}},
/* 59267 */ {(14<<2)|2,{89,69,0}},
/* 59268 */ {(14<<2)|2,{89,70,0}},
/* 59269 */ {(14<<2)|2,{89,70,0}},
/* 59270 */ {(14<<2)|2,{89,70,0}},
/* 59271 */ {(14<<2)|2,{89,70,0}},
/* 59272 */ {(14<<2)|2,{89,71,0}},
/* 59273 */ {(14<<2)|2,{89,71,0}},
/* 59274 */ {(14<<2)|2,{89,71,0}},
/* 59275 */ {(14<<2)|2,{89,71,0}},
/* 59276 */ {(14<<2)|2,{89,72,0}},
/* 59277 */ {(14<<2)|2,{89,72,0}},
/* 59278 */ {(14<<2)|2,{89,72,0}},
/* 59279 */ {(14<<2)|2,{89,72,0}},
/* 59280 */ {(14<<2)|2,{89,73,0}},
/* 59281 */ {(14<<2)|2,{89,73,0}},
/* 59282 */ {(14<<2)|2,{89,73,0}},
/* 59283 */ {(14<<2)|2,{89,73,0}},
/* 59284 */ {(14<<2)|2,{89,74,0}},
/* 59285 */ {(14<<2)|2,{89,74,0}},
/* 59286 */ {(14<<2)|2,{89,74,0}},
/* 59287 */ {(14<<2)|2,{89,74,0}},
/* 59288 */ {(14<<2)|2,{89,75,0}},
/* 59289 */ {(14<<2)|2,{89,75,0}},
/* 59290 */ {(14<<2)|2,{89,75,0}},
/* 59291 */ {(14<<2)|2,{89,75,0}},
/* 59292 */ {(14<<2)|2,{89,76,0}},
/* 59293 */ {(14<<2)|2,{89,76,0}},
/* 59294 */ {(14<<2)|2,{89,76,0}},
/* 59295 */ {(14<<2)|2,{89,76,0}},
/* 59296 */ {(14<<2)|2,{89,77,0}},
/* 59297 */ {(14<<2)|2,{89,77,0}},
/* 59298 */ {(14<<2)|2,{89,77,0}},
/* 59299 */ {(14<<2)|2,{89,77,0}},
/* 59300 */ {(14<<2)|2,{89,78,0}},
/* 59301 */ {(14<<2)|2,{89,78,0}},
/* 59302 */ {(14<<2)|2,{89,78,0}},
/* 59303 */ {(14<<2)|2,{89,78,0}},
/* 59304 */ {(14<<2)|2,{89,79,0}},
/* 59305 */ {(14<<2)|2,{89,79,0}},
/* 59306 */ {(14<<2)|2,{89,79,0}},
/* 59307 */ {(14<<2)|2,{89,79,0}},
/* 59308 */ {(14<<2)|2,{89,80,0}},
/* 59309 */ {(14<<2)|2,{89,80,0}},
/* 59310 */ {(14<<2)|2,{89,80,0}},
/* 59311 */ {(14<<2)|2,{89,80,0}},
/* 59312 */ {(14<<2)|2,{89,81,0}},
/* 59313 */ {(14<<2)|2,{89,81,0}},
/* 59314 */ {(14<<2)|2,{89,81,0}},
/* 59315 */ {(14<<2)|2,{89,81,0}},
/* 59316 */ {(14<<2)|2,{89,82,0}},
/* 59317 */ {(14<<2)|2,{89,82,0}},
/* 59318 */ {(14<<2)|2,{89,82,0}},
/* 59319 */ {(14<<2)|2,{89,82,0}},
/* 59320 */ {(14<<2)|2,{89,83,0}},
/* 59321 */ {(14<<2)|2,{89,83,0}},
/* 59322 */ {(14<<2)|2,{89,83,0}},
/* 59323 */ {(14<<2)|2,{89,83,0}},
/* 59324 */ {(14<<2)|2,{89,84,0}},
/* 59325 */ {(14<<2)|2,{89,84,0}},
/* 59326 */ {(14<<2)|2,{89,84,0}},
/* 59327 */ {(14<<2)|2,{89,84,0}},
/* 59328 */ {(14<<2)|2,{89,85,0}},
/* 59329 */ {(14<<2)|2,{89,85,0}},
/* 59330 */ {(14<<2)|2,{89,85,0}},
/* 59331 */ {(14<<2)|2,{89,85,0}},
/* 59332 */ {(14<<2)|2,{89,86,0}},
/* 59333 */ {(14<<2)|2,{89,86,0}},
/* 59334 */ {(14<<2)|2,{89,86,0}},
/* 59335 */ {(14<<2)|2,{89,86,0}},
/* 59336 */ {(14<<2)|2,{89,87,0}},
/* 59337 */ {(14<<2)|2,{89,87,0}},
/* 59338 */ {(14<<2)|2,{89,87,0}},
/* 59339 */ {(14<<2)|2,{89,87,0}},
/* 59340 */ {(14<<2)|2,{89,89,0}},
/* 59341 */ {(14<<2)|2,{89,89,0}},
/* 59342 */ {(14<<2)|2,{89,89,0}},
/* 59343 */ {(14<<2)|2,{89,89,0}},
/* 59344 */ {(14<<2)|2,{89,106,0}},
/* 59345 */ {(14<<2)|2,{89,106,0}},
/* 59346 */ {(14<<2)|2,{89,106,0}},
/* 59347 */ {(14<<2)|2,{89,106,0}},
/* 59348 */ {(14<<2)|2,{89,107,0}},
/* 59349 */ {(14<<2)|2,{89,107,0}},
/* 59350 */ {(14<<2)|2,{89,107,0}},
/* 59351 */ {(14<<2)|2,{89,107,0}},
/* 59352 */ {(14<<2)|2,{89,113,0}},
/* 59353 */ {(14<<2)|2,{89,113,0}},
/* 59354 */ {(14<<2)|2,{89,113,0}},
/* 59355 */ {(14<<2)|2,{89,113,0}},
/* 59356 */ {(14<<2)|2,{89,118,0}},
/* 59357 */ {(14<<2)|2,{89,118,0}},
/* 59358 */ {(14<<2)|2,{89,118,0}},
/* 59359 */ {(14<<2)|2,{89,118,0}},
/* 59360 */ {(14<<2)|2,{89,119,0}},
/* 59361 */ {(14<<2)|2,{89,119,0}},
/* 59362 */ {(14<<2)|2,{89,119,0}},
/* 59363 */ {(14<<2)|2,{89,119,0}},
/* 59364 */ {(14<<2)|2,{89,120,0}},
/* 59365 */ {(14<<2)|2,{89,120,0}},
/* 59366 */ {(14<<2)|2,{89,120,0}},
/* 59367 */ {(14<<2)|2,{89,120,0}},
/* 59368 */ {(14<<2)|2,{89,121,0}},
/* 59369 */ {(14<<2)|2,{89,121,0}},
/* 59370 */ {(14<<2)|2,{89,121,0}},
/* 59371 */ {(14<<2)|2,{89,121,0}},
/* 59372 */ {(14<<2)|2,{89,122,0}},
/* 59373 */ {(14<<2)|2,{89,122,0}},
/* 59374 */ {(14<<2)|2,{89,122,0}},
/* 59375 */ {(14<<2)|2,{89,122,0}},
/* 59376 */ {(15<<2)|2,{89,38,0}},
/* 59377 */ {(15<<2)|2,{89,38,0}},
/* 59378 */ {(15<<2)|2,{89,42,0}},
/* 59379 */ {(15<<2)|2,{89,42,0}},
/* 59380 */ {(15<<2)|2,{89,44,0}},
/* 59381 */ {(15<<2)|2,{89,44,0}},
/* 59382 */ {(15<<2)|2,{89,59,0}},
/* 59383 */ {(15<<2)|2,{89,59,0}},
/* 59384 */ {(15<<2)|2,{89,88,0}},
/* 59385 */ {(15<<2)|2,{89,88,0}},
/* 59386 */ {(15<<2)|2,{89,90,0}},
/* 59387 */ {(15<<2)|2,{89,90,0}},
/* 59388 */ {(7<<2)|1,{89,0,0}},
/* 59389 */ {(7<<2)|1,{89,0,0}},
/* 59390 */ {(7<<2)|1,{89,0,0}},
/* 59391 */ {(7<<2)|1,{89,0,0}},
/* 59392 */ {(12<<2)|2,{106,48,0}},
/* 59393 */ {(12<<2)|2,{106,48,0}},
/* 59394 */ {(12<<2)|2,{106,48,0}},
/* 59395 */ {(12<<2)|2,{106,48,0}},
/* 59396 */ {(12<<2)|2,{106,48,0}},
/* 59397 */ {(12<<2)|2,{106,48,0}},
/* 59398 */ {(12<<2)|2,{106,48,0}},
/* 59399 */ {(12<<2)|2,{106,48,0}},
/* 59400 */ {(12<<2)|2,{106,48,0}},
/* 59401 */ {(12<<2)|2,{106,48,0}},
/* 59402 */ {(12<<2)|2,{106,48,0}},
/* 59403 */ {(12<<2)|2,{106,48,0}},
/* 59404 */ {(12<<2)|2,{106,48,0}},
/* 59405 */ {(12<<2)|2,{106,48,0}},
/* 59406 */ {(12<<2)|2,{106,48,0}},
/* 59407 */ {(12<<2)|2,{106,48,0}},
/* 59408 */ {(12<<2)|2,{106,49,0}},
/* 59409 */ {(12<<2)|2,{106,49,0}},
/* 59410 */ {(12<<2)|2,{106,49,0}},
/* 59411 */ {(12<<2)|2,{106,49,0}},
/* 59412 */ {(12<<2)|2,{106,49,0}},
/* 59413 */ {(12<<2)|2,{106,49,0}},
/* 59414 */ {(12<<2)|2,{106,49,0}},
/* 59415 */ {(12<<2)|2,{106,49,0}},
/* 59416 */ {(12<<2)|2,{106,49,0}},
/* 59417 */ {(12<<2)|2,{106,49,0}},
/* 59418 */ {(12<<2)|2,{106,49,0}},
/* 59419 */ {(12<<2)|2,{106,49,0}},
/* 59420 */ {(12<<2)|2,{106,49,0}},
/* 59421 */ {(12<<2)|2,{106,49,0}},
/* 59422 */ {(12<<2)|2,{106,49,0}},
/* 59423 */ {(12<<2)|2,{106,49,0}},
/* 59424 */ {(12<<2)|2,{106,50,0}},
/* 59425 */ {(12<<2)|2,{106,50,0}},
/* 59426 */ {(12<<2)|2,{106,50,0}},
/* 59427 */ {(12<<2)|2,{106,50,0}},
/* 59428 */ {(12<<2)|2,{106,50,0}},
/* 59429 */ {(12<<2)|2,{106,50,0}},
/* 59430 */ {(12<<2)|2,{106,50,0}},
/* 59431 */ {(12<<2)|2,{106,50,0}},
/* 59432 */ {(12<<2)|2,{106,50,0}},
/* 59433 */ {(12<<2)|2,{106,50,0}},
/* 59434 */ {(12<<2)|2,{106,50,0}},
/* 59435 */ {(12<<2)|2,{106,50,0}},
/* 59436 */ {(12<<2)|2,{106,50,0}},
/* 59437 */ {(12<<2)|2,{106,50,0}},
/* 59438 */ {(12<<2)|2,{106,50,0}},
/* 59439 */ {(12<<2)|2,{106,50,0}},
/* 59440 */ {(12<<2)|2,{106,97,0}},
/* 59441 */ {(12<<2)|2,{106,97,0}},
/* 59442 */ {(12<<2)|2,{106,97,0}},
/* 59443 */ {(12<<2)|2,{106,97,0}},
/* 59444 */ {(12<<2)|2,{106,97,0}},
/* 59445 */ {(12<<2)|2,{106,97,0}},
/* 59446 */ {(12<<2)|2,{106,97,0}},
/* 59447 */ {(12<<2)|2,{106,97,0}},
/* 59448 */ {(12<<2)|2,{106,97,0}},
/* 59449 */ {(12<<2)|2,{106,97,0}},
/* 59450 */ {(12<<2)|2,{106,97,0}},
/* 59451 */ {(12<<2)|2,{106,97,0}},
/* 59452 */ {(12<<2)|2,{106,97,0}},
/* 59453 */ {(12<<2)|2,{106,97,0}},
/* 59454 */ {(12<<2)|2,{106,97,0}},
/* 59455 */ {(12<<2)|2,{106,97,0}},
/* 59456 */ {(12<<2)|2,{106,99,0}},
/* 59457 */ {(12<<2)|2,{106,99,0}},
/* 59458 */ {(12<<2)|2,{106,99,0}},
/* 59459 */ {(12<<2)|2,{106,99,0}},
/* 59460 */ {(12<<2)|2,{106,99,0}},
/* 59461 */ {(12<<2)|2,{106,99,0}},
/* 59462 */ {(12<<2)|2,{106,99,0}},
/* 59463 */ {(12<<2)|2,{106,99,0}},
/* 59464 */ {(12<<2)|2,{106,99,0}},
/* 59465 */ {(12<<2)|2,{106,99,0}},
/* 59466 */ {(12<<2)|2,{106,99,0}},
/* 59467 */ {(12<<2)|2,{106,99,0}},
/* 59468 */ {(12<<2)|2,{106,99,0}},
/* 59469 */ {(12<<2)|2,{106,99,0}},
/* 59470 */ {(12<<2)|2,{106,99,0}},
/* 59471 */ {(12<<2)|2,{106,99,0}},
/* 59472 */ {(12<<2)|2,{106,101,0}},
/* 59473 */ {(12<<2)|2,{106,101,0}},
/* 59474 */ {(12<<2)|2,{106,101,0}},
/* 59475 */ {(12<<2)|2,{106,101,0}},
/* 59476 */ {(12<<2)|2,{106,101,0}},
/* 59477 */ {(12<<2)|2,{106,101,0}},
/* 59478 */ {(12<<2)|2,{106,101,0}},
/* 59479 */ {(12<<2)|2,{106,101,0}},
/* 59480 */ {(12<<2)|2,{106,101,0}},
/* 59481 */ {(12<<2)|2,{106,101,0}},
/* 59482 */ {(12<<2)|2,{106,101,0}},
/* 59483 */ {(12<<2)|2,{106,101,0}},
/* 59484 */ {(12<<2)|2,{106,101,0}},
/* 59485 */ {(12<<2)|2,{106,101,0}},
/* 59486 */ {(12<<2)|2,{106,101,0}},
/* 59487 */ {(12<<2)|2,{106,101,0}},
/* 59488 */ {(12<<2)|2,{106,105,0}},
/* 59489 */ {(12<<2)|2,{106,105,0}},
/* 59490 */ {(12<<2)|2,{106,105,0}},
/* 59491 */ {(12<<2)|2,{106,105,0}},
/* 59492 */ {(12<<2)|2,{106,105,0}},
/* 59493 */ {(12<<2)|2,{106,105,0}},
/* 59494 */ {(12<<2)|2,{106,105,0}},
/* 59495 */ {(12<<2)|2,{106,105,0}},
/* 59496 */ {(12<<2)|2,{106,105,0}},
/* 59497 */ {(12<<2)|2,{106,105,0}},
/* 59498 */ {(12<<2)|2,{106,105,0}},
/* 59499 */ {(12<<2)|2,{106,105,0}},
/* 59500 */ {(12<<2)|2,{106,105,0}},
/* 59501 */ {(12<<2)|2,{106,105,0}},
/* 59502 */ {(12<<2)|2,{106,105,0}},
/* 59503 */ {(12<<2)|2,{106,105,0}},
/* 59504 */ {(12<<2)|2,{106,111,0}},
/* 59505 */ {(12<<2)|2,{106,111,0}},
/* 59506 */ {(12<<2)|2,{106,111,0}},
/* 59507 */ {(12<<2)|2,{106,111,0}},
/* 59508 */ {(12<<2)|2,{106,111,0}},
/* 59509 */ {(12<<2)|2,{106,111,0}},
/* 59510 */ {(12<<2)|2,{106,111,0}},
/* 59511 */ {(12<<2)|2,{106,111,0}},
/* 59512 */ {(12<<2)|2,{106,111,0}},
/* 59513 */ {(12<<2)|2,{106,111,0}},
/* 59514 */ {(12<<2)|2,{106,111,0}},
/* 59515 */ {(12<<2)|2,{106,111,0}},
/* 59516 */ {(12<<2)|2,{106,111,0}},
/* 59517 */ {(12<<2)|2,{106,111,0}},
/* 59518 */ {(12<<2)|2,{106,111,0}},
/* 59519 */ {(12<<2)|2,{106,111,0}},
/* 59520 */ {(12<<2)|2,{106,115,0}},
/* 59521 */ {(12<<2)|2,{106,115,0}},
/* 59522 */ {(12<<2)|2,{106,115,0}},
/* 59523 */ {(12<<2)|2,{106,115,0}},
/* 59524 */ {(12<<2)|2,{106,115,0}},
/* 59525 */ {(12<<2)|2,{106,115,0}},
/* 59526 */ {(12<<2)|2,{106,115,0}},
/* 59527 */ {(12<<2)|2,{106,115,0}},
/* 59528 */ {(12<<2)|2,{106,115,0}},
/* 59529 */ {(12<<2)|2,{106,115,0}},
/* 59530 */ {(12<<2)|2,{106,115,0}},
/* 59531 */ {(12<<2)|2,{106,115,0}},
/* 59532 */ {(12<<2)|2,{106,115,0}},
/* 59533 */ {(12<<2)|2,{106,115,0}},
/* 59534 */ {(12<<2)|2,{106,115,0}},
/* 59535 */ {(12<<2)|2,{106,115,0}},
/* 59536 */ {(12<<2)|2,{106,116,0}},
/* 59537 */ {(12<<2)|2,{106,116,0}},
/* 59538 */ {(12<<2)|2,{106,116,0}},
/* 59539 */ {(12<<2)|2,{106,116,0}},
/* 59540 */ {(12<<2)|2,{106,116,0}},
/* 59541 */ {(12<<2)|2,{106,116,0}},
/* 59542 */ {(12<<2)|2,{106,116,0}},
/* 59543 */ {(12<<2)|2,{106,116,0}},
/* 59544 */ {(12<<2)|2,{106,116,0}},
/* 59545 */ {(12<<2)|2,{106,116,0}},
/* 59546 */ {(12<<2)|2,{106,116,0}},
/* 59547 */ {(12<<2)|2,{106,116,0}},
/* 59548 */ {(12<<2)|2,{106,116,0}},
/* 59549 */ {(12<<2)|2,{106,116,0}},
/* 59550 */ {(12<<2)|2,{106,116,0}},
/* 59551 */ {(12<<2)|2,{106,116,0}},
/* 59552 */ {(13<<2)|2,{106,32,0}},
/* 59553 */ {(13<<2)|2,{106,32,0}},
/* 59554 */ {(13<<2)|2,{106,32,0}},
/* 59555 */ {(13<<2)|2,{106,32,0}},
/* 59556 */ {(13<<2)|2,{106,32,0}},
/* 59557 */ {(13<<2)|2,{106,32,0}},
/* 59558 */ {(13<<2)|2,{106,32,0}},
/* 59559 */ {(13<<2)|2,{106,32,0}},
/* 59560 */ {(13<<2)|2,{106,37,0}},
/* 59561 */ {(13<<2)|2,{106,37,0}},
/* 59562 */ {(13<<2)|2,{106,37,0}},
/* 59563 */ {(13<<2)|2,{106,37,0}},
/* 59564 */ {(13<<2)|2,{106,37,0}},
/* 59565 */ {(13<<2)|2,{106,37,0}},
/* 59566 */ {(13<<2)|2,{106,37,0}},
/* 59567 */ {(13<<2)|2,{106,37,0}},
/* 59568 */ {(13<<2)|2,{106,45,0}},
/* 59569 */ {(13<<2)|2,{106,45,0}},
/* 59570 */ {(13<<2)|2,{106,45,0}},
/* 59571 */ {(13<<2)|2,{106,45,0}},
/* 59572 */ {(13<<2)|2,{106,45,0}},
/* 59573 */ {(13<<2)|2,{106,45,0}},
/* 59574 */ {(13<<2)|2,{106,45,0}},
/* 59575 */ {(13<<2)|2,{106,45,0}},
/* 59576 */ {(13<<2)|2,{106,46,0}},
/* 59577 */ {(13<<2)|2,{106,46,0}},
/* 59578 */ {(13<<2)|2,{106,46,0}},
/* 59579 */ {(13<<2)|2,{106,46,0}},
/* 59580 */ {(13<<2)|2,{106,46,0}},
/* 59581 */ {(13<<2)|2,{106,46,0}},
/* 59582 */ {(13<<2)|2,{106,46,0}},
/* 59583 */ {(13<<2)|2,{106,46,0}},
/* 59584 */ {(13<<2)|2,{106,47,0}},
/* 59585 */ {(13<<2)|2,{106,47,0}},
/* 59586 */ {(13<<2)|2,{106,47,0}},
/* 59587 */ {(13<<2)|2,{106,47,0}},
/* 59588 */ {(13<<2)|2,{106,47,0}},
/* 59589 */ {(13<<2)|2,{106,47,0}},
/* 59590 */ {(13<<2)|2,{106,47,0}},
/* 59591 */ {(13<<2)|2,{106,47,0}},
/* 59592 */ {(13<<2)|2,{106,51,0}},
/* 59593 */ {(13<<2)|2,{106,51,0}},
/* 59594 */ {(13<<2)|2,{106,51,0}},
/* 59595 */ {(13<<2)|2,{106,51,0}},
/* 59596 */ {(13<<2)|2,{106,51,0}},
/* 59597 */ {(13<<2)|2,{106,51,0}},
/* 59598 */ {(13<<2)|2,{106,51,0}},
/* 59599 */ {(13<<2)|2,{106,51,0}},
/* 59600 */ {(13<<2)|2,{106,52,0}},
/* 59601 */ {(13<<2)|2,{106,52,0}},
/* 59602 */ {(13<<2)|2,{106,52,0}},
/* 59603 */ {(13<<2)|2,{106,52,0}},
/* 59604 */ {(13<<2)|2,{106,52,0}},
/* 59605 */ {(13<<2)|2,{106,52,0}},
/* 59606 */ {(13<<2)|2,{106,52,0}},
/* 59607 */ {(13<<2)|2,{106,52,0}},
/* 59608 */ {(13<<2)|2,{106,53,0}},
/* 59609 */ {(13<<2)|2,{106,53,0}},
/* 59610 */ {(13<<2)|2,{106,53,0}},
/* 59611 */ {(13<<2)|2,{106,53,0}},
/* 59612 */ {(13<<2)|2,{106,53,0}},
/* 59613 */ {(13<<2)|2,{106,53,0}},
/* 59614 */ {(13<<2)|2,{106,53,0}},
/* 59615 */ {(13<<2)|2,{106,53,0}},
/* 59616 */ {(13<<2)|2,{106,54,0}},
/* 59617 */ {(13<<2)|2,{106,54,0}},
/* 59618 */ {(13<<2)|2,{106,54,0}},
/* 59619 */ {(13<<2)|2,{106,54,0}},
/* 59620 */ {(13<<2)|2,{106,54,0}},
/* 59621 */ {(13<<2)|2,{106,54,0}},
/* 59622 */ {(13<<2)|2,{106,54,0}},
/* 59623 */ {(13<<2)|2,{106,54,0}},
/* 59624 */ {(13<<2)|2,{106,55,0}},
/* 59625 */ {(13<<2)|2,{106,55,0}},
/* 59626 */ {(13<<2)|2,{106,55,0}},
/* 59627 */ {(13<<2)|2,{106,55,0}},
/* 59628 */ {(13<<2)|2,{106,55,0}},
/* 59629 */ {(13<<2)|2,{106,55,0}},
/* 59630 */ {(13<<2)|2,{106,55,0}},
/* 59631 */ {(13<<2)|2,{106,55,0}},
/* 59632 */ {(13<<2)|2,{106,56,0}},
/* 59633 */ {(13<<2)|2,{106,56,0}},
/* 59634 */ {(13<<2)|2,{106,56,0}},
/* 59635 */ {(13<<2)|2,{106,56,0}},
/* 59636 */ {(13<<2)|2,{106,56,0}},
/* 59637 */ {(13<<2)|2,{106,56,0}},
/* 59638 */ {(13<<2)|2,{106,56,0}},
/* 59639 */ {(13<<2)|2,{106,56,0}},
/* 59640 */ {(13<<2)|2,{106,57,0}},
/* 59641 */ {(13<<2)|2,{106,57,0}},
/* 59642 */ {(13<<2)|2,{106,57,0}},
/* 59643 */ {(13<<2)|2,{106,57,0}},
/* 59644 */ {(13<<2)|2,{106,57,0}},
/* 59645 */ {(13<<2)|2,{106,57,0}},
/* 59646 */ {(13<<2)|2,{106,57,0}},
/* 59647 */ {(13<<2)|2,{106,57,0}},
/* 59648 */ {(13<<2)|2,{106,61,0}},
/* 59649 */ {(13<<2)|2,{106,61,0}},
/* 59650 */ {(13<<2)|2,{106,61,0}},
/* 59651 */ {(13<<2)|2,{106,61,0}},
/* 59652 */ {(13<<2)|2,{106,61,0}},
/* 59653 */ {(13<<2)|2,{106,61,0}},
/* 59654 */ {(13<<2)|2,{106,61,0}},
/* 59655 */ {(13<<2)|2,{106,61,0}},
/* 59656 */ {(13<<2)|2,{106,65,0}},
/* 59657 */ {(13<<2)|2,{106,65,0}},
/* 59658 */ {(13<<2)|2,{106,65,0}},
/* 59659 */ {(13<<2)|2,{106,65,0}},
/* 59660 */ {(13<<2)|2,{106,65,0}},
/* 59661 */ {(13<<2)|2,{106,65,0}},
/* 59662 */ {(13<<2)|2,{106,65,0}},
/* 59663 */ {(13<<2)|2,{106,65,0}},
/* 59664 */ {(13<<2)|2,{106,95,0}},
/* 59665 */ {(13<<2)|2,{106,95,0}},
/* 59666 */ {(13<<2)|2,{106,95,0}},
/* 59667 */ {(13<<2)|2,{106,95,0}},
/* 59668 */ {(13<<2)|2,{106,95,0}},
/* 59669 */ {(13<<2)|2,{106,95,0}},
/* 59670 */ {(13<<2)|2,{106,95,0}},
/* 59671 */ {(13<<2)|2,{106,95,0}},
/* 59672 */ {(13<<2)|2,{106,98,0}},
/* 59673 */ {(13<<2)|2,{106,98,0}},
/* 59674 */ {(13<<2)|2,{106,98,0}},
/* 59675 */ {(13<<2)|2,{106,98,0}},
/* 59676 */ {(13<<2)|2,{106,98,0}},
/* 59677 */ {(13<<2)|2,{106,98,0}},
/* 59678 */ {(13<<2)|2,{106,98,0}},
/* 59679 */ {(13<<2)|2,{106,98,0}},
/* 59680 */ {(13<<2)|2,{106,100,0}},
/* 59681 */ {(13<<2)|2,{106,100,0}},
/* 59682 */ {(13<<2)|2,{106,100,0}},
/* 59683 */ {(13<<2)|2,{106,100,0}},
/* 59684 */ {(13<<2)|2,{106,100,0}},
/* 59685 */ {(13<<2)|2,{106,100,0}},
/* 59686 */ {(13<<2)|2,{106,100,0}},
/* 59687 */ {(13<<2)|2,{106,100,0}},
/* 59688 */ {(13<<2)|2,{106,102,0}},
/* 59689 */ {(13<<2)|2,{106,102,0}},
/* 59690 */ {(13<<2)|2,{106,102,0}},
/* 59691 */ {(13<<2)|2,{106,102,0}},
/* 59692 */ {(13<<2)|2,{106,102,0}},
/* 59693 */ {(13<<2)|2,{106,102,0}},
/* 59694 */ {(13<<2)|2,{106,102,0}},
/* 59695 */ {(13<<2)|2,{106,102,0}},
/* 59696 */ {(13<<2)|2,{106,103,0}},
/* 59697 */ {(13<<2)|2,{106,103,0}},
/* 59698 */ {(13<<2)|2,{106,103,0}},
/* 59699 */ {(13<<2)|2,{106,103,0}},
/* 59700 */ {(13<<2)|2,{106,103,0}},
/* 59701 */ {(13<<2)|2,{106,103,0}},
/* 59702 */ {(13<<2)|2,{106,103,0}},
/* 59703 */ {(13<<2)|2,{106,103,0}},
/* 59704 */ {(13<<2)|2,{106,104,0}},
/* 59705 */ {(13<<2)|2,{106,104,0}},
/* 59706 */ {(13<<2)|2,{106,104,0}},
/* 59707 */ {(13<<2)|2,{106,104,0}},
/* 59708 */ {(13<<2)|2,{106,104,0}},
/* 59709 */ {(13<<2)|2,{106,104,0}},
/* 59710 */ {(13<<2)|2,{106,104,0}},
/* 59711 */ {(13<<2)|2,{106,104,0}},
/* 59712 */ {(13<<2)|2,{106,108,0}},
/* 59713 */ {(13<<2)|2,{106,108,0}},
/* 59714 */ {(13<<2)|2,{106,108,0}},
/* 59715 */ {(13<<2)|2,{106,108,0}},
/* 59716 */ {(13<<2)|2,{106,108,0}},
/* 59717 */ {(13<<2)|2,{106,108,0}},
/* 59718 */ {(13<<2)|2,{106,108,0}},
/* 59719 */ {(13<<2)|2,{106,108,0}},
/* 59720 */ {(13<<2)|2,{106,109,0}},
/* 59721 */ {(13<<2)|2,{106,109,0}},
/* 59722 */ {(13<<2)|2,{106,109,0}},
/* 59723 */ {(13<<2)|2,{106,109,0}},
/* 59724 */ {(13<<2)|2,{106,109,0}},
/* 59725 */ {(13<<2)|2,{106,109,0}},
/* 59726 */ {(13<<2)|2,{106,109,0}},
/* 59727 */ {(13<<2)|2,{106,109,0}},
/* 59728 */ {(13<<2)|2,{106,110,0}},
/* 59729 */ {(13<<2)|2,{106,110,0}},
/* 59730 */ {(13<<2)|2,{106,110,0}},
/* 59731 */ {(13<<2)|2,{106,110,0}},
/* 59732 */ {(13<<2)|2,{106,110,0}},
/* 59733 */ {(13<<2)|2,{106,110,0}},
/* 59734 */ {(13<<2)|2,{106,110,0}},
/* 59735 */ {(13<<2)|2,{106,110,0}},
/* 59736 */ {(13<<2)|2,{106,112,0}},
/* 59737 */ {(13<<2)|2,{106,112,0}},
/* 59738 */ {(13<<2)|2,{106,112,0}},
/* 59739 */ {(13<<2)|2,{106,112,0}},
/* 59740 */ {(13<<2)|2,{106,112,0}},
/* 59741 */ {(13<<2)|2,{106,112,0}},
/* 59742 */ {(13<<2)|2,{106,112,0}},
/* 59743 */ {(13<<2)|2,{106,112,0}},
/* 59744 */ {(13<<2)|2,{106,114,0}},
/* 59745 */ {(13<<2)|2,{106,114,0}},
/* 59746 */ {(13<<2)|2,{106,114,0}},
/* 59747 */ {(13<<2)|2,{106,114,0}},
/* 59748 */ {(13<<2)|2,{106,114,0}},
/* 59749 */ {(13<<2)|2,{106,114,0}},
/* 59750 */ {(13<<2)|2,{106,114,0}},
/* 59751 */ {(13<<2)|2,{106,114,0}},
/* 59752 */ {(13<<2)|2,{106,117,0}},
/* 59753 */ {(13<<2)|2,{106,117,0}},
/* 59754 */ {(13<<2)|2,{106,117,0}},
/* 59755 */ {(13<<2)|2,{106,117,0}},
/* 59756 */ {(13<<2)|2,{106,117,0}},
/* 59757 */ {(13<<2)|2,{106,117,0}},
/* 59758 */ {(13<<2)|2,{106,117,0}},
/* 59759 */ {(13<<2)|2,{106,117,0}},
/* 59760 */ {(14<<2)|2,{106,58,0}},
/* 59761 */ {(14<<2)|2,{106,58,0}},
/* 59762 */ {(14<<2)|2,{106,58,0}},
/* 59763 */ {(14<<2)|2,{106,58,0}},
/* 59764 */ {(14<<2)|2,{106,66,0}},
/* 59765 */ {(14<<2)|2,{106,66,0}},
/* 59766 */ {(14<<2)|2,{106,66,0}},
/* 59767 */ {(14<<2)|2,{106,66,0}},
/* 59768 */ {(14<<2)|2,{106,67,0}},
/* 59769 */ {(14<<2)|2,{106,67,0}},
/* 59770 */ {(14<<2)|2,{106,67,0}},
/* 59771 */ {(14<<2)|2,{106,67,0}},
/* 59772 */ {(14<<2)|2,{106,68,0}},
/* 59773 */ {(14<<2)|2,{106,68,0}},
/* 59774 */ {(14<<2)|2,{106,68,0}},
/* 59775 */ {(14<<2)|2,{106,68,0}},
/* 59776 */ {(14<<2)|2,{106,69,0}},
/* 59777 */ {(14<<2)|2,{106,69,0}},
/* 59778 */ {(14<<2)|2,{106,69,0}},
/* 59779 */ {(14<<2)|2,{106,69,0}},
/* 59780 */ {(14<<2)|2,{106,70,0}},
/* 59781 */ {(14<<2)|2,{106,70,0}},
/* 59782 */ {(14<<2)|2,{106,70,0}},
/* 59783 */ {(14<<2)|2,{106,70,0}},
/* 59784 */ {(14<<2)|2,{106,71,0}},
/* 59785 */ {(14<<2)|2,{106,71,0}},
/* 59786 */ {(14<<2)|2,{106,71,0}},
/* 59787 */ {(14<<2)|2,{106,71,0}},
/* 59788 */ {(14<<2)|2,{106,72,0}},
/* 59789 */ {(14<<2)|2,{106,72,0}},
/* 59790 */ {(14<<2)|2,{106,72,0}},
/* 59791 */ {(14<<2)|2,{106,72,0}},
/* 59792 */ {(14<<2)|2,{106,73,0}},
/* 59793 */ {(14<<2)|2,{106,73,0}},
/* 59794 */ {(14<<2)|2,{106,73,0}},
/* 59795 */ {(14<<2)|2,{106,73,0}},
/* 59796 */ {(14<<2)|2,{106,74,0}},
/* 59797 */ {(14<<2)|2,{106,74,0}},
/* 59798 */ {(14<<2)|2,{106,74,0}},
/* 59799 */ {(14<<2)|2,{106,74,0}},
/* 59800 */ {(14<<2)|2,{106,75,0}},
/* 59801 */ {(14<<2)|2,{106,75,0}},
/* 59802 */ {(14<<2)|2,{106,75,0}},
/* 59803 */ {(14<<2)|2,{106,75,0}},
/* 59804 */ {(14<<2)|2,{106,76,0}},
/* 59805 */ {(14<<2)|2,{106,76,0}},
/* 59806 */ {(14<<2)|2,{106,76,0}},
/* 59807 */ {(14<<2)|2,{106,76,0}},
/* 59808 */ {(14<<2)|2,{106,77,0}},
/* 59809 */ {(14<<2)|2,{106,77,0}},
/* 59810 */ {(14<<2)|2,{106,77,0}},
/* 59811 */ {(14<<2)|2,{106,77,0}},
/* 59812 */ {(14<<2)|2,{106,78,0}},
/* 59813 */ {(14<<2)|2,{106,78,0}},
/* 59814 */ {(14<<2)|2,{106,78,0}},
/* 59815 */ {(14<<2)|2,{106,78,0}},
/* 59816 */ {(14<<2)|2,{106,79,0}},
/* 59817 */ {(14<<2)|2,{106,79,0}},
/* 59818 */ {(14<<2)|2,{106,79,0}},
/* 59819 */ {(14<<2)|2,{106,79,0}},
/* 59820 */ {(14<<2)|2,{106,80,0}},
/* 59821 */ {(14<<2)|2,{106,80,0}},
/* 59822 */ {(14<<2)|2,{106,80,0}},
/* 59823 */ {(14<<2)|2,{106,80,0}},
/* 59824 */ {(14<<2)|2,{106,81,0}},
/* 59825 */ {(14<<2)|2,{106,81,0}},
/* 59826 */ {(14<<2)|2,{106,81,0}},
/* 59827 */ {(14<<2)|2,{106,81,0}},
/* 59828 */ {(14<<2)|2,{106,82,0}},
/* 59829 */ {(14<<2)|2,{106,82,0}},
/* 59830 */ {(14<<2)|2,{106,82,0}},
/* 59831 */ {(14<<2)|2,{106,82,0}},
/* 59832 */ {(14<<2)|2,{106,83,0}},
/* 59833 */ {(14<<2)|2,{106,83,0}},
/* 59834 */ {(14<<2)|2,{106,83,0}},
/* 59835 */ {(14<<2)|2,{106,83,0}},
/* 59836 */ {(14<<2)|2,{106,84,0}},
/* 59837 */ {(14<<2)|2,{106,84,0}},
/* 59838 */ {(14<<2)|2,{106,84,0}},
/* 59839 */ {(14<<2)|2,{106,84,0}},
/* 59840 */ {(14<<2)|2,{106,85,0}},
/* 59841 */ {(14<<2)|2,{106,85,0}},
/* 59842 */ {(14<<2)|2,{106,85,0}},
/* 59843 */ {(14<<2)|2,{106,85,0}},
/* 59844 */ {(14<<2)|2,{106,86,0}},
/* 59845 */ {(14<<2)|2,{106,86,0}},
/* 59846 */ {(14<<2)|2,{106,86,0}},
/* 59847 */ {(14<<2)|2,{106,86,0}},
/* 59848 */ {(14<<2)|2,{106,87,0}},
/* 59849 */ {(14<<2)|2,{106,87,0}},
/* 59850 */ {(14<<2)|2,{106,87,0}},
/* 59851 */ {(14<<2)|2,{106,87,0}},
/* 59852 */ {(14<<2)|2,{106,89,0}},
/* 59853 */ {(14<<2)|2,{106,89,0}},
/* 59854 */ {(14<<2)|2,{106,89,0}},
/* 59855 */ {(14<<2)|2,{106,89,0}},
/* 59856 */ {(14<<2)|2,{106,106,0}},
/* 59857 */ {(14<<2)|2,{106,106,0}},
/* 59858 */ {(14<<2)|2,{106,106,0}},
/* 59859 */ {(14<<2)|2,{106,106,0}},
/* 59860 */ {(14<<2)|2,{106,107,0}},
/* 59861 */ {(14<<2)|2,{106,107,0}},
/* 59862 */ {(14<<2)|2,{106,107,0}},
/* 59863 */ {(14<<2)|2,{106,107,0}},
/* 59864 */ {(14<<2)|2,{106,113,0}},
/* 59865 */ {(14<<2)|2,{106,113,0}},
/* 59866 */ {(14<<2)|2,{106,113,0}},
/* 59867 */ {(14<<2)|2,{106,113,0}},
/* 59868 */ {(14<<2)|2,{106,118,0}},
/* 59869 */ {(14<<2)|2,{106,118,0}},
/* 59870 */ {(14<<2)|2,{106,118,0}},
/* 59871 */ {(14<<2)|2,{106,118,0}},
/* 59872 */ {(14<<2)|2,{106,119,0}},
/* 59873 */ {(14<<2)|2,{106,119,0}},
/* 59874 */ {(14<<2)|2,{106,119,0}},
/* 59875 */ {(14<<2)|2,{106,119,0}},
/* 59876 */ {(14<<2)|2,{106,120,0}},
/* 59877 */ {(14<<2)|2,{106,120,0}},
/* 59878 */ {(14<<2)|2,{106,120,0}},
/* 59879 */ {(14<<2)|2,{106,120,0}},
/* 59880 */ {(14<<2)|2,{106,121,0}},
/* 59881 */ {(14<<2)|2,{106,121,0}},
/* 59882 */ {(14<<2)|2,{106,121,0}},
/* 59883 */ {(14<<2)|2,{106,121,0}},
/* 59884 */ {(14<<2)|2,{106,122,0}},
/* 59885 */ {(14<<2)|2,{106,122,0}},
/* 59886 */ {(14<<2)|2,{106,122,0}},
/* 59887 */ {(14<<2)|2,{106,122,0}},
/* 59888 */ {(15<<2)|2,{106,38,0}},
/* 59889 */ {(15<<2)|2,{106,38,0}},
/* 59890 */ {(15<<2)|2,{106,42,0}},
/* 59891 */ {(15<<2)|2,{106,42,0}},
/* 59892 */ {(15<<2)|2,{106,44,0}},
/* 59893 */ {(15<<2)|2,{106,44,0}},
/* 59894 */ {(15<<2)|2,{106,59,0}},
/* 59895 */ {(15<<2)|2,{106,59,0}},
/* 59896 */ {(15<<2)|2,{106,88,0}},
/* 59897 */ {(15<<2)|2,{106,88,0}},
/* 59898 */ {(15<<2)|2,{106,90,0}},
/* 59899 */ {(15<<2)|2,{106,90,0}},
/* 59900 */ {(7<<2)|1,{106,0,0}},
/* 59901 */ {(7<<2)|1,{106,0,0}},
/* 59902 */ {(7<<2)|1,{106,0,0}},
/* 59903 */ {(7<<2)|1,{106,0,0}},
/* 59904 */ {(12<<2)|2,{107,48,0}},
/* 59905 */ {(12<<2)|2,{107,48,0}},
/* 59906 */ {(12<<2)|2,{107,48,0}},
/* 59907 */ {(12<<2)|2,{107,48,0}},
/* 59908 */ {(12<<2)|2,{107,48,0}},
/* 59909 */ {(12<<2)|2,{107,48,0}},
/* 59910 */ {(12<<2)|2,{107,48,0}},
/* 59911 */ {(12<<2)|2,{107,48,0}},
/* 59912 */ {(12<<2)|2,{107,48,0}},
/* 59913 */ {(12<<2)|2,{107,48,0}},
/* 59914 */ {(12<<2)|2,{107,48,0}},
/* 59915 */ {(12<<2)|2,{107,48,0}},
/* 59916 */ {(12<<2)|2,{107,48,0}},
/* 59917 */ {(12<<2)|2,{107,48,0}},
/* 59918 */ {(12<<2)|2,{107,48,0}},
/* 59919 */ {(12<<2)|2,{107,48,0}},
/* 59920 */ {(12<<2)|2,{107,49,0}},
/* 59921 */ {(12<<2)|2,{107,49,0}},
/* 59922 */ {(12<<2)|2,{107,49,0}},
/* 59923 */ {(12<<2)|2,{107,49,0}},
/* 59924 */ {(12<<2)|2,{107,49,0}},
/* 59925 */ {(12<<2)|2,{107,49,0}},
/* 59926 */ {(12<<2)|2,{107,49,0}},
/* 59927 */ {(12<<2)|2,{107,49,0}},
/* 59928 */ {(12<<2)|2,{107,49,0}},
/* 59929 */ {(12<<2)|2,{107,49,0}},
/* 59930 */ {(12<<2)|2,{107,49,0}},
/* 59931 */ {(12<<2)|2,{107,49,0}},
/* 59932 */ {(12<<2)|2,{107,49,0}},
/* 59933 */ {(12<<2)|2,{107,49,0}},
/* 59934 */ {(12<<2)|2,{107,49,0}},
/* 59935 */ {(12<<2)|2,{107,49,0}},
/* 59936 */ {(12<<2)|2,{107,50,0}},
/* 59937 */ {(12<<2)|2,{107,50,0}},
/* 59938 */ {(12<<2)|2,{107,50,0}},
/* 59939 */ {(12<<2)|2,{107,50,0}},
/* 59940 */ {(12<<2)|2,{107,50,0}},
/* 59941 */ {(12<<2)|2,{107,50,0}},
/* 59942 */ {(12<<2)|2,{107,50,0}},
/* 59943 */ {(12<<2)|2,{107,50,0}},
/* 59944 */ {(12<<2)|2,{107,50,0}},
/* 59945 */ {(12<<2)|2,{107,50,0}},
/* 59946 */ {(12<<2)|2,{107,50,0}},
/* 59947 */ {(12<<2)|2,{107,50,0}},
/* 59948 */ {(12<<2)|2,{107,50,0}},
/* 59949 */ {(12<<2)|2,{107,50,0}},
/* 59950 */ {(12<<2)|2,{107,50,0}},
/* 59951 */ {(12<<2)|2,{107,50,0}},
/* 59952 */ {(12<<2)|2,{107,97,0}},
/* 59953 */ {(12<<2)|2,{107,97,0}},
/* 59954 */ {(12<<2)|2,{107,97,0}},
/* 59955 */ {(12<<2)|2,{107,97,0}},
/* 59956 */ {(12<<2)|2,{107,97,0}},
/* 59957 */ {(12<<2)|2,{107,97,0}},
/* 59958 */ {(12<<2)|2,{107,97,0}},
/* 59959 */ {(12<<2)|2,{107,97,0}},
/* 59960 */ {(12<<2)|2,{107,97,0}},
/* 59961 */ {(12<<2)|2,{107,97,0}},
/* 59962 */ {(12<<2)|2,{107,97,0}},
/* 59963 */ {(12<<2)|2,{107,97,0}},
/* 59964 */ {(12<<2)|2,{107,97,0}},
/* 59965 */ {(12<<2)|2,{107,97,0}},
/* 59966 */ {(12<<2)|2,{107,97,0}},
/* 59967 */ {(12<<2)|2,{107,97,0}},
/* 59968 */ {(12<<2)|2,{107,99,0}},
/* 59969 */ {(12<<2)|2,{107,99,0}},
/* 59970 */ {(12<<2)|2,{107,99,0}},
/* 59971 */ {(12<<2)|2,{107,99,0}},
/* 59972 */ {(12<<2)|2,{107,99,0}},
/* 59973 */ {(12<<2)|2,{107,99,0}},
/* 59974 */ {(12<<2)|2,{107,99,0}},
/* 59975 */ {(12<<2)|2,{107,99,0}},
/* 59976 */ {(12<<2)|2,{107,99,0}},
/* 59977 */ {(12<<2)|2,{107,99,0}},
/* 59978 */ {(12<<2)|2,{107,99,0}},
/* 59979 */ {(12<<2)|2,{107,99,0}},
/* 59980 */ {(12<<2)|2,{107,99,0}},
/* 59981 */ {(12<<2)|2,{107,99,0}},
/* 59982 */ {(12<<2)|2,{107,99,0}},
/* 59983 */ {(12<<2)|2,{107,99,0}},
/* 59984 */ {(12<<2)|2,{107,101,0}},
/* 59985 */ {(12<<2)|2,{107,101,0}},
/* 59986 */ {(12<<2)|2,{107,101,0}},
/* 59987 */ {(12<<2)|2,{107,101,0}},
/* 59988 */ {(12<<2)|2,{107,101,0}},
/* 59989 */ {(12<<2)|2,{107,101,0}},
/* 59990 */ {(12<<2)|2,{107,101,0}},
/* 59991 */ {(12<<2)|2,{107,101,0}},
/* 59992 */ {(12<<2)|2,{107,101,0}},
/* 59993 */ {(12<<2)|2,{107,101,0}},
/* 59994 */ {(12<<2)|2,{107,101,0}},
/* 59995 */ {(12<<2)|2,{107,101,0}},
/* 59996 */ {(12<<2)|2,{107,101,0}},
/* 59997 */ {(12<<2)|2,{107,101,0}},
/* 59998 */ {(12<<2)|2,{107,101,0}},
/* 59999 */ {(12<<2)|2,{107,101,0}},
/* 60000 */ {(12<<2)|2,{107,105,0}},
/* 60001 */ {(12<<2)|2,{107,105,0}},
/* 60002 */ {(12<<2)|2,{107,105,0}},
/* 60003 */ {(12<<2)|2,{107,105,0}},
/* 60004 */ {(12<<2)|2,{107,105,0}},
/* 60005 */ {(12<<2)|2,{107,105,0}},
/* 60006 */ {(12<<2)|2,{107,105,0}},
/* 60007 */ {(12<<2)|2,{107,105,0}},
/* 60008 */ {(12<<2)|2,{107,105,0}},
/* 60009 */ {(12<<2)|2,{107,105,0}},
/* 60010 */ {(12<<2)|2,{107,105,0}},
/* 60011 */ {(12<<2)|2,{107,105,0}},
/* 60012 */ {(12<<2)|2,{107,105,0}},
/* 60013 */ {(12<<2)|2,{107,105,0}},
/* 60014 */ {(12<<2)|2,{107,105,0}},
/* 60015 */ {(12<<2)|2,{107,105,0}},
/* 60016 */ {(12<<2)|2,{107,111,0}},
/* 60017 */ {(12<<2)|2,{107,111,0}},
/* 60018 */ {(12<<2)|2,{107,111,0}},
/* 60019 */ {(12<<2)|2,{107,111,0}},
/* 60020 */ {(12<<2)|2,{107,111,0}},
/* 60021 */ {(12<<2)|2,{107,111,0}},
/* 60022 */ {(12<<2)|2,{107,111,0}},
/* 60023 */ {(12<<2)|2,{107,111,0}},
/* 60024 */ {(12<<2)|2,{107,111,0}},
/* 60025 */ {(12<<2)|2,{107,111,0}},
/* 60026 */ {(12<<2)|2,{107,111,0}},
/* 60027 */ {(12<<2)|2,{107,111,0}},
/* 60028 */ {(12<<2)|2,{107,111,0}},
/* 60029 */ {(12<<2)|2,{107,111,0}},
/* 60030 */ {(12<<2)|2,{107,111,0}},
/* 60031 */ {(12<<2)|2,{107,111,0}},
/* 60032 */ {(12<<2)|2,{107,115,0}},
/* 60033 */ {(12<<2)|2,{107,115,0}},
/* 60034 */ {(12<<2)|2,{107,115,0}},
/* 60035 */ {(12<<2)|2,{107,115,0}},
/* 60036 */ {(12<<2)|2,{107,115,0}},
/* 60037 */ {(12<<2)|2,{107,115,0}},
/* 60038 */ {(12<<2)|2,{107,115,0}},
/* 60039 */ {(12<<2)|2,{107,115,0}},
/* 60040 */ {(12<<2)|2,{107,115,0}},
/* 60041 */ {(12<<2)|2,{107,115,0}},
/* 60042 */ {(12<<2)|2,{107,115,0}},
/* 60043 */ {(12<<2)|2,{107,115,0}},
/* 60044 */ {(12<<2)|2,{107,115,0}},
/* 60045 */ {(12<<2)|2,{107,115,0}},
/* 60046 */ {(12<<2)|2,{107,115,0}},
/* 60047 */ {(12<<2)|2,{107,115,0}},
/* 60048 */ {(12<<2)|2,{107,116,0}},
/* 60049 */ {(12<<2)|2,{107,116,0}},
/* 60050 */ {(12<<2)|2,{107,116,0}},
/* 60051 */ {(12<<2)|2,{107,116,0}},
/* 60052 */ {(12<<2)|2,{107,116,0}},
/* 60053 */ {(12<<2)|2,{107,116,0}},
/* 60054 */ {(12<<2)|2,{107,116,0}},
/* 60055 */ {(12<<2)|2,{107,116,0}},
/* 60056 */ {(12<<2)|2,{107,116,0}},
/* 60057 */ {(12<<2)|2,{107,116,0}},
/* 60058 */ {(12<<2)|2,{107,116,0}},
/* 60059 */ {(12<<2)|2,{107,116,0}},
/* 60060 */ {(12<<2)|2,{107,116,0}},
/* 60061 */ {(12<<2)|2,{107,116,0}},
/* 60062 */ {(12<<2)|2,{107,116,0}},
/* 60063 */ {(12<<2)|2,{107,116,0}},
/* 60064 */ {(13<<2)|2,{107,32,0}},
/* 60065 */ {(13<<2)|2,{107,32,0}},
/* 60066 */ {(13<<2)|2,{107,32,0}},
/* 60067 */ {(13<<2)|2,{107,32,0}},
/* 60068 */ {(13<<2)|2,{107,32,0}},
/* 60069 */ {(13<<2)|2,{107,32,0}},
/* 60070 */ {(13<<2)|2,{107,32,0}},
/* 60071 */ {(13<<2)|2,{107,32,0}},
/* 60072 */ {(13<<2)|2,{107,37,0}},
/* 60073 */ {(13<<2)|2,{107,37,0}},
/* 60074 */ {(13<<2)|2,{107,37,0}},
/* 60075 */ {(13<<2)|2,{107,37,0}},
/* 60076 */ {(13<<2)|2,{107,37,0}},
/* 60077 */ {(13<<2)|2,{107,37,0}},
/* 60078 */ {(13<<2)|2,{107,37,0}},
/* 60079 */ {(13<<2)|2,{107,37,0}},
/* 60080 */ {(13<<2)|2,{107,45,0}},
/* 60081 */ {(13<<2)|2,{107,45,0}},
/* 60082 */ {(13<<2)|2,{107,45,0}},
/* 60083 */ {(13<<2)|2,{107,45,0}},
/* 60084 */ {(13<<2)|2,{107,45,0}},
/* 60085 */ {(13<<2)|2,{107,45,0}},
/* 60086 */ {(13<<2)|2,{107,45,0}},
/* 60087 */ {(13<<2)|2,{107,45,0}},
/* 60088 */ {(13<<2)|2,{107,46,0}},
/* 60089 */ {(13<<2)|2,{107,46,0}},
/* 60090 */ {(13<<2)|2,{107,46,0}},
/* 60091 */ {(13<<2)|2,{107,46,0}},
/* 60092 */ {(13<<2)|2,{107,46,0}},
/* 60093 */ {(13<<2)|2,{107,46,0}},
/* 60094 */ {(13<<2)|2,{107,46,0}},
/* 60095 */ {(13<<2)|2,{107,46,0}},
/* 60096 */ {(13<<2)|2,{107,47,0}},
/* 60097 */ {(13<<2)|2,{107,47,0}},
/* 60098 */ {(13<<2)|2,{107,47,0}},
/* 60099 */ {(13<<2)|2,{107,47,0}},
/* 60100 */ {(13<<2)|2,{107,47,0}},
/* 60101 */ {(13<<2)|2,{107,47,0}},
/* 60102 */ {(13<<2)|2,{107,47,0}},
/* 60103 */ {(13<<2)|2,{107,47,0}},
/* 60104 */ {(13<<2)|2,{107,51,0}},
/* 60105 */ {(13<<2)|2,{107,51,0}},
/* 60106 */ {(13<<2)|2,{107,51,0}},
/* 60107 */ {(13<<2)|2,{107,51,0}},
/* 60108 */ {(13<<2)|2,{107,51,0}},
/* 60109 */ {(13<<2)|2,{107,51,0}},
/* 60110 */ {(13<<2)|2,{107,51,0}},
/* 60111 */ {(13<<2)|2,{107,51,0}},
/* 60112 */ {(13<<2)|2,{107,52,0}},
/* 60113 */ {(13<<2)|2,{107,52,0}},
/* 60114 */ {(13<<2)|2,{107,52,0}},
/* 60115 */ {(13<<2)|2,{107,52,0}},
/* 60116 */ {(13<<2)|2,{107,52,0}},
/* 60117 */ {(13<<2)|2,{107,52,0}},
/* 60118 */ {(13<<2)|2,{107,52,0}},
/* 60119 */ {(13<<2)|2,{107,52,0}},
/* 60120 */ {(13<<2)|2,{107,53,0}},
/* 60121 */ {(13<<2)|2,{107,53,0}},
/* 60122 */ {(13<<2)|2,{107,53,0}},
/* 60123 */ {(13<<2)|2,{107,53,0}},
/* 60124 */ {(13<<2)|2,{107,53,0}},
/* 60125 */ {(13<<2)|2,{107,53,0}},
/* 60126 */ {(13<<2)|2,{107,53,0}},
/* 60127 */ {(13<<2)|2,{107,53,0}},
/* 60128 */ {(13<<2)|2,{107,54,0}},
/* 60129 */ {(13<<2)|2,{107,54,0}},
/* 60130 */ {(13<<2)|2,{107,54,0}},
/* 60131 */ {(13<<2)|2,{107,54,0}},
/* 60132 */ {(13<<2)|2,{107,54,0}},
/* 60133 */ {(13<<2)|2,{107,54,0}},
/* 60134 */ {(13<<2)|2,{107,54,0}},
/* 60135 */ {(13<<2)|2,{107,54,0}},
/* 60136 */ {(13<<2)|2,{107,55,0}},
/* 60137 */ {(13<<2)|2,{107,55,0}},
/* 60138 */ {(13<<2)|2,{107,55,0}},
/* 60139 */ {(13<<2)|2,{107,55,0}},
/* 60140 */ {(13<<2)|2,{107,55,0}},
/* 60141 */ {(13<<2)|2,{107,55,0}},
/* 60142 */ {(13<<2)|2,{107,55,0}},
/* 60143 */ {(13<<2)|2,{107,55,0}},
/* 60144 */ {(13<<2)|2,{107,56,0}},
/* 60145 */ {(13<<2)|2,{107,56,0}},
/* 60146 */ {(13<<2)|2,{107,56,0}},
/* 60147 */ {(13<<2)|2,{107,56,0}},
/* 60148 */ {(13<<2)|2,{107,56,0}},
/* 60149 */ {(13<<2)|2,{107,56,0}},
/* 60150 */ {(13<<2)|2,{107,56,0}},
/* 60151 */ {(13<<2)|2,{107,56,0}},
/* 60152 */ {(13<<2)|2,{107,57,0}},
/* 60153 */ {(13<<2)|2,{107,57,0}},
/* 60154 */ {(13<<2)|2,{107,57,0}},
/* 60155 */ {(13<<2)|2,{107,57,0}},
/* 60156 */ {(13<<2)|2,{107,57,0}},
/* 60157 */ {(13<<2)|2,{107,57,0}},
/* 60158 */ {(13<<2)|2,{107,57,0}},
/* 60159 */ {(13<<2)|2,{107,57,0}},
/* 60160 */ {(13<<2)|2,{107,61,0}},
/* 60161 */ {(13<<2)|2,{107,61,0}},
/* 60162 */ {(13<<2)|2,{107,61,0}},
/* 60163 */ {(13<<2)|2,{107,61,0}},
/* 60164 */ {(13<<2)|2,{107,61,0}},
/* 60165 */ {(13<<2)|2,{107,61,0}},
/* 60166 */ {(13<<2)|2,{107,61,0}},
/* 60167 */ {(13<<2)|2,{107,61,0}},
/* 60168 */ {(13<<2)|2,{107,65,0}},
/* 60169 */ {(13<<2)|2,{107,65,0}},
/* 60170 */ {(13<<2)|2,{107,65,0}},
/* 60171 */ {(13<<2)|2,{107,65,0}},
/* 60172 */ {(13<<2)|2,{107,65,0}},
/* 60173 */ {(13<<2)|2,{107,65,0}},
/* 60174 */ {(13<<2)|2,{107,65,0}},
/* 60175 */ {(13<<2)|2,{107,65,0}},
/* 60176 */ {(13<<2)|2,{107,95,0}},
/* 60177 */ {(13<<2)|2,{107,95,0}},
/* 60178 */ {(13<<2)|2,{107,95,0}},
/* 60179 */ {(13<<2)|2,{107,95,0}},
/* 60180 */ {(13<<2)|2,{107,95,0}},
/* 60181 */ {(13<<2)|2,{107,95,0}},
/* 60182 */ {(13<<2)|2,{107,95,0}},
/* 60183 */ {(13<<2)|2,{107,95,0}},
/* 60184 */ {(13<<2)|2,{107,98,0}},
/* 60185 */ {(13<<2)|2,{107,98,0}},
/* 60186 */ {(13<<2)|2,{107,98,0}},
/* 60187 */ {(13<<2)|2,{107,98,0}},
/* 60188 */ {(13<<2)|2,{107,98,0}},
/* 60189 */ {(13<<2)|2,{107,98,0}},
/* 60190 */ {(13<<2)|2,{107,98,0}},
/* 60191 */ {(13<<2)|2,{107,98,0}},
/* 60192 */ {(13<<2)|2,{107,100,0}},
/* 60193 */ {(13<<2)|2,{107,100,0}},
/* 60194 */ {(13<<2)|2,{107,100,0}},
/* 60195 */ {(13<<2)|2,{107,100,0}},
/* 60196 */ {(13<<2)|2,{107,100,0}},
/* 60197 */ {(13<<2)|2,{107,100,0}},
/* 60198 */ {(13<<2)|2,{107,100,0}},
/* 60199 */ {(13<<2)|2,{107,100,0}},
/* 60200 */ {(13<<2)|2,{107,102,0}},
/* 60201 */ {(13<<2)|2,{107,102,0}},
/* 60202 */ {(13<<2)|2,{107,102,0}},
/* 60203 */ {(13<<2)|2,{107,102,0}},
/* 60204 */ {(13<<2)|2,{107,102,0}},
/* 60205 */ {(13<<2)|2,{107,102,0}},
/* 60206 */ {(13<<2)|2,{107,102,0}},
/* 60207 */ {(13<<2)|2,{107,102,0}},
/* 60208 */ {(13<<2)|2,{107,103,0}},
/* 60209 */ {(13<<2)|2,{107,103,0}},
/* 60210 */ {(13<<2)|2,{107,103,0}},
/* 60211 */ {(13<<2)|2,{107,103,0}},
/* 60212 */ {(13<<2)|2,{107,103,0}},
/* 60213 */ {(13<<2)|2,{107,103,0}},
/* 60214 */ {(13<<2)|2,{107,103,0}},
/* 60215 */ {(13<<2)|2,{107,103,0}},
/* 60216 */ {(13<<2)|2,{107,104,0}},
/* 60217 */ {(13<<2)|2,{107,104,0}},
/* 60218 */ {(13<<2)|2,{107,104,0}},
/* 60219 */ {(13<<2)|2,{107,104,0}},
/* 60220 */ {(13<<2)|2,{107,104,0}},
/* 60221 */ {(13<<2)|2,{107,104,0}},
/* 60222 */ {(13<<2)|2,{107,104,0}},
/* 60223 */ {(13<<2)|2,{107,104,0}},
/* 60224 */ {(13<<2)|2,{107,108,0}},
/* 60225 */ {(13<<2)|2,{107,108,0}},
/* 60226 */ {(13<<2)|2,{107,108,0}},
/* 60227 */ {(13<<2)|2,{107,108,0}},
/* 60228 */ {(13<<2)|2,{107,108,0}},
/* 60229 */ {(13<<2)|2,{107,108,0}},
/* 60230 */ {(13<<2)|2,{107,108,0}},
/* 60231 */ {(13<<2)|2,{107,108,0}},
/* 60232 */ {(13<<2)|2,{107,109,0}},
/* 60233 */ {(13<<2)|2,{107,109,0}},
/* 60234 */ {(13<<2)|2,{107,109,0}},
/* 60235 */ {(13<<2)|2,{107,109,0}},
/* 60236 */ {(13<<2)|2,{107,109,0}},
/* 60237 */ {(13<<2)|2,{107,109,0}},
/* 60238 */ {(13<<2)|2,{107,109,0}},
/* 60239 */ {(13<<2)|2,{107,109,0}},
/* 60240 */ {(13<<2)|2,{107,110,0}},
/* 60241 */ {(13<<2)|2,{107,110,0}},
/* 60242 */ {(13<<2)|2,{107,110,0}},
/* 60243 */ {(13<<2)|2,{107,110,0}},
/* 60244 */ {(13<<2)|2,{107,110,0}},
/* 60245 */ {(13<<2)|2,{107,110,0}},
/* 60246 */ {(13<<2)|2,{107,110,0}},
/* 60247 */ {(13<<2)|2,{107,110,0}},
/* 60248 */ {(13<<2)|2,{107,112,0}},
/* 60249 */ {(13<<2)|2,{107,112,0}},
/* 60250 */ {(13<<2)|2,{107,112,0}},
/* 60251 */ {(13<<2)|2,{107,112,0}},
/* 60252 */ {(13<<2)|2,{107,112,0}},
/* 60253 */ {(13<<2)|2,{107,112,0}},
/* 60254 */ {(13<<2)|2,{107,112,0}},
/* 60255 */ {(13<<2)|2,{107,112,0}},
/* 60256 */ {(13<<2)|2,{107,114,0}},
/* 60257 */ {(13<<2)|2,{107,114,0}},
/* 60258 */ {(13<<2)|2,{107,114,0}},
/* 60259 */ {(13<<2)|2,{107,114,0}},
/* 60260 */ {(13<<2)|2,{107,114,0}},
/* 60261 */ {(13<<2)|2,{107,114,0}},
/* 60262 */ {(13<<2)|2,{107,114,0}},
/* 60263 */ {(13<<2)|2,{107,114,0}},
/* 60264 */ {(13<<2)|2,{107,117,0}},
/* 60265 */ {(13<<2)|2,{107,117,0}},
/* 60266 */ {(13<<2)|2,{107,117,0}},
/* 60267 */ {(13<<2)|2,{107,117,0}},
/* 60268 */ {(13<<2)|2,{107,117,0}},
/* 60269 */ {(13<<2)|2,{107,117,0}},
/* 60270 */ {(13<<2)|2,{107,117,0}},
/* 60271 */ {(13<<2)|2,{107,117,0}},
/* 60272 */ {(14<<2)|2,{107,58,0}},
/* 60273 */ {(14<<2)|2,{107,58,0}},
/* 60274 */ {(14<<2)|2,{107,58,0}},
/* 60275 */ {(14<<2)|2,{107,58,0}},
/* 60276 */ {(14<<2)|2,{107,66,0}},
/* 60277 */ {(14<<2)|2,{107,66,0}},
/* 60278 */ {(14<<2)|2,{107,66,0}},
/* 60279 */ {(14<<2)|2,{107,66,0}},
/* 60280 */ {(14<<2)|2,{107,67,0}},
/* 60281 */ {(14<<2)|2,{107,67,0}},
/* 60282 */ {(14<<2)|2,{107,67,0}},
/* 60283 */ {(14<<2)|2,{107,67,0}},
/* 60284 */ {(14<<2)|2,{107,68,0}},
/* 60285 */ {(14<<2)|2,{107,68,0}},
/* 60286 */ {(14<<2)|2,{107,68,0}},
/* 60287 */ {(14<<2)|2,{107,68,0}},
/* 60288 */ {(14<<2)|2,{107,69,0}},
/* 60289 */ {(14<<2)|2,{107,69,0}},
/* 60290 */ {(14<<2)|2,{107,69,0}},
/* 60291 */ {(14<<2)|2,{107,69,0}},
/* 60292 */ {(14<<2)|2,{107,70,0}},
/* 60293 */ {(14<<2)|2,{107,70,0}},
/* 60294 */ {(14<<2)|2,{107,70,0}},
/* 60295 */ {(14<<2)|2,{107,70,0}},
/* 60296 */ {(14<<2)|2,{107,71,0}},
/* 60297 */ {(14<<2)|2,{107,71,0}},
/* 60298 */ {(14<<2)|2,{107,71,0}},
/* 60299 */ {(14<<2)|2,{107,71,0}},
/* 60300 */ {(14<<2)|2,{107,72,0}},
/* 60301 */ {(14<<2)|2,{107,72,0}},
/* 60302 */ {(14<<2)|2,{107,72,0}},
/* 60303 */ {(14<<2)|2,{107,72,0}},
/* 60304 */ {(14<<2)|2,{107,73,0}},
/* 60305 */ {(14<<2)|2,{107,73,0}},
/* 60306 */ {(14<<2)|2,{107,73,0}},
/* 60307 */ {(14<<2)|2,{107,73,0}},
/* 60308 */ {(14<<2)|2,{107,74,0}},
/* 60309 */ {(14<<2)|2,{107,74,0}},
/* 60310 */ {(14<<2)|2,{107,74,0}},
/* 60311 */ {(14<<2)|2,{107,74,0}},
/* 60312 */ {(14<<2)|2,{107,75,0}},
/* 60313 */ {(14<<2)|2,{107,75,0}},
/* 60314 */ {(14<<2)|2,{107,75,0}},
/* 60315 */ {(14<<2)|2,{107,75,0}},
/* 60316 */ {(14<<2)|2,{107,76,0}},
/* 60317 */ {(14<<2)|2,{107,76,0}},
/* 60318 */ {(14<<2)|2,{107,76,0}},
/* 60319 */ {(14<<2)|2,{107,76,0}},
/* 60320 */ {(14<<2)|2,{107,77,0}},
/* 60321 */ {(14<<2)|2,{107,77,0}},
/* 60322 */ {(14<<2)|2,{107,77,0}},
/* 60323 */ {(14<<2)|2,{107,77,0}},
/* 60324 */ {(14<<2)|2,{107,78,0}},
/* 60325 */ {(14<<2)|2,{107,78,0}},
/* 60326 */ {(14<<2)|2,{107,78,0}},
/* 60327 */ {(14<<2)|2,{107,78,0}},
/* 60328 */ {(14<<2)|2,{107,79,0}},
/* 60329 */ {(14<<2)|2,{107,79,0}},
/* 60330 */ {(14<<2)|2,{107,79,0}},
/* 60331 */ {(14<<2)|2,{107,79,0}},
/* 60332 */ {(14<<2)|2,{107,80,0}},
/* 60333 */ {(14<<2)|2,{107,80,0}},
/* 60334 */ {(14<<2)|2,{107,80,0}},
/* 60335 */ {(14<<2)|2,{107,80,0}},
/* 60336 */ {(14<<2)|2,{107,81,0}},
/* 60337 */ {(14<<2)|2,{107,81,0}},
/* 60338 */ {(14<<2)|2,{107,81,0}},
/* 60339 */ {(14<<2)|2,{107,81,0}},
/* 60340 */ {(14<<2)|2,{107,82,0}},
/* 60341 */ {(14<<2)|2,{107,82,0}},
/* 60342 */ {(14<<2)|2,{107,82,0}},
/* 60343 */ {(14<<2)|2,{107,82,0}},
/* 60344 */ {(14<<2)|2,{107,83,0}},
/* 60345 */ {(14<<2)|2,{107,83,0}},
/* 60346 */ {(14<<2)|2,{107,83,0}},
/* 60347 */ {(14<<2)|2,{107,83,0}},
/* 60348 */ {(14<<2)|2,{107,84,0}},
/* 60349 */ {(14<<2)|2,{107,84,0}},
/* 60350 */ {(14<<2)|2,{107,84,0}},
/* 60351 */ {(14<<2)|2,{107,84,0}},
/* 60352 */ {(14<<2)|2,{107,85,0}},
/* 60353 */ {(14<<2)|2,{107,85,0}},
/* 60354 */ {(14<<2)|2,{107,85,0}},
/* 60355 */ {(14<<2)|2,{107,85,0}},
/* 60356 */ {(14<<2)|2,{107,86,0}},
/* 60357 */ {(14<<2)|2,{107,86,0}},
/* 60358 */ {(14<<2)|2,{107,86,0}},
/* 60359 */ {(14<<2)|2,{107,86,0}},
/* 60360 */ {(14<<2)|2,{107,87,0}},
/* 60361 */ {(14<<2)|2,{107,87,0}},
/* 60362 */ {(14<<2)|2,{107,87,0}},
/* 60363 */ {(14<<2)|2,{107,87,0}},
/* 60364 */ {(14<<2)|2,{107,89,0}},
/* 60365 */ {(14<<2)|2,{107,89,0}},
/* 60366 */ {(14<<2)|2,{107,89,0}},
/* 60367 */ {(14<<2)|2,{107,89,0}},
/* 60368 */ {(14<<2)|2,{107,106,0}},
/* 60369 */ {(14<<2)|2,{107,106,0}},
/* 60370 */ {(14<<2)|2,{107,106,0}},
/* 60371 */ {(14<<2)|2,{107,106,0}},
/* 60372 */ {(14<<2)|2,{107,107,0}},
/* 60373 */ {(14<<2)|2,{107,107,0}},
/* 60374 */ {(14<<2)|2,{107,107,0}},
/* 60375 */ {(14<<2)|2,{107,107,0}},
/* 60376 */ {(14<<2)|2,{107,113,0}},
/* 60377 */ {(14<<2)|2,{107,113,0}},
/* 60378 */ {(14<<2)|2,{107,113,0}},
/* 60379 */ {(14<<2)|2,{107,113,0}},
/* 60380 */ {(14<<2)|2,{107,118,0}},
/* 60381 */ {(14<<2)|2,{107,118,0}},
/* 60382 */ {(14<<2)|2,{107,118,0}},
/* 60383 */ {(14<<2)|2,{107,118,0}},
/* 60384 */ {(14<<2)|2,{107,119,0}},
/* 60385 */ {(14<<2)|2,{107,119,0}},
/* 60386 */ {(14<<2)|2,{107,119,0}},
/* 60387 */ {(14<<2)|2,{107,119,0}},
/* 60388 */ {(14<<2)|2,{107,120,0}},
/* 60389 */ {(14<<2)|2,{107,120,0}},
/* 60390 */ {(14<<2)|2,{107,120,0}},
/* 60391 */ {(14<<2)|2,{107,120,0}},
/* 60392 */ {(14<<2)|2,{107,121,0}},
/* 60393 */ {(14<<2)|2,{107,121,0}},
/* 60394 */ {(14<<2)|2,{107,121,0}},
/* 60395 */ {(14<<2)|2,{107,121,0}},
/* 60396 */ {(14<<2)|2,{107,122,0}},
/* 60397 */ {(14<<2)|2,{107,122,0}},
/* 60398 */ {(14<<2)|2,{107,122,0}},
/* 60399 */ {(14<<2)|2,{107,122,0}},
/* 60400 */ {(15<<2)|2,{107,38,0}},
/* 60401 */ {(15<<2)|2,{107,38,0}},
/* 60402 */ {(15<<2)|2,{107,42,0}},
/* 60403 */ {(15<<2)|2,{107,42,0}},
/* 60404 */ {(15<<2)|2,{107,44,0}},
/* 60405 */ {(15<<2)|2,{107,44,0}},
/* 60406 */ {(15<<2)|2,{107,59,0}},
/* 60407 */ {(15<<2)|2,{107,59,0}},
/* 60408 */ {(15<<2)|2,{107,88,0}},
/* 60409 */ {(15<<2)|2,{107,88,0}},
/* 60410 */ {(15<<2)|2,{107,90,0}},
/* 60411 */ {(15<<2)|2,{107,90,0}},
/* 60412 */ {(7<<2)|1,{107,0,0}},
/* 60413 */ {(7<<2)|1,{107,0,0}},
/* 60414 */ {(7<<2)|1,{107,0,0}},
/* 60415 */ {(7<<2)|1,{107,0,0}},
/* 60416 */ {(12<<2)|2,{113,48,0}},
/* 60417 */ {(12<<2)|2,{113,48,0}},
/* 60418 */ {(12<<2)|2,{113,48,0}},
/* 60419 */ {(12<<2)|2,{113,48,0}},
/* 60420 */ {(12<<2)|2,{113,48,0}},
/* 60421 */ {(12<<2)|2,{113,48,0}},
/* 60422 */ {(12<<2)|2,{113,48,0}},
/* 60423 */ {(12<<2)|2,{113,48,0}},
/* 60424 */ {(12<<2)|2,{113,48,0}},
/* 60425 */ {(12<<2)|2,{113,48,0}},
/* 60426 */ {(12<<2)|2,{113,48,0}},
/* 60427 */ {(12<<2)|2,{113,48,0}},
/* 60428 */ {(12<<2)|2,{113,48,0}},
/* 60429 */ {(12<<2)|2,{113,48,0}},
/* 60430 */ {(12<<2)|2,{113,48,0}},
/* 60431 */ {(12<<2)|2,{113,48,0}},
/* 60432 */ {(12<<2)|2,{113,49,0}},
/* 60433 */ {(12<<2)|2,{113,49,0}},
/* 60434 */ {(12<<2)|2,{113,49,0}},
/* 60435 */ {(12<<2)|2,{113,49,0}},
/* 60436 */ {(12<<2)|2,{113,49,0}},
/* 60437 */ {(12<<2)|2,{113,49,0}},
/* 60438 */ {(12<<2)|2,{113,49,0}},
/* 60439 */ {(12<<2)|2,{113,49,0}},
/* 60440 */ {(12<<2)|2,{113,49,0}},
/* 60441 */ {(12<<2)|2,{113,49,0}},
/* 60442 */ {(12<<2)|2,{113,49,0}},
/* 60443 */ {(12<<2)|2,{113,49,0}},
/* 60444 */ {(12<<2)|2,{113,49,0}},
/* 60445 */ {(12<<2)|2,{113,49,0}},
/* 60446 */ {(12<<2)|2,{113,49,0}},
/* 60447 */ {(12<<2)|2,{113,49,0}},
/* 60448 */ {(12<<2)|2,{113,50,0}},
/* 60449 */ {(12<<2)|2,{113,50,0}},
/* 60450 */ {(12<<2)|2,{113,50,0}},
/* 60451 */ {(12<<2)|2,{113,50,0}},
/* 60452 */ {(12<<2)|2,{113,50,0}},
/* 60453 */ {(12<<2)|2,{113,50,0}},
/* 60454 */ {(12<<2)|2,{113,50,0}},
/* 60455 */ {(12<<2)|2,{113,50,0}},
/* 60456 */ {(12<<2)|2,{113,50,0}},
/* 60457 */ {(12<<2)|2,{113,50,0}},
/* 60458 */ {(12<<2)|2,{113,50,0}},
/* 60459 */ {(12<<2)|2,{113,50,0}},
/* 60460 */ {(12<<2)|2,{113,50,0}},
/* 60461 */ {(12<<2)|2,{113,50,0}},
/* 60462 */ {(12<<2)|2,{113,50,0}},
/* 60463 */ {(12<<2)|2,{113,50,0}},
/* 60464 */ {(12<<2)|2,{113,97,0}},
/* 60465 */ {(12<<2)|2,{113,97,0}},
/* 60466 */ {(12<<2)|2,{113,97,0}},
/* 60467 */ {(12<<2)|2,{113,97,0}},
/* 60468 */ {(12<<2)|2,{113,97,0}},
/* 60469 */ {(12<<2)|2,{113,97,0}},
/* 60470 */ {(12<<2)|2,{113,97,0}},
/* 60471 */ {(12<<2)|2,{113,97,0}},
/* 60472 */ {(12<<2)|2,{113,97,0}},
/* 60473 */ {(12<<2)|2,{113,97,0}},
/* 60474 */ {(12<<2)|2,{113,97,0}},
/* 60475 */ {(12<<2)|2,{113,97,0}},
/* 60476 */ {(12<<2)|2,{113,97,0}},
/* 60477 */ {(12<<2)|2,{113,97,0}},
/* 60478 */ {(12<<2)|2,{113,97,0}},
/* 60479 */ {(12<<2)|2,{113,97,0}},
/* 60480 */ {(12<<2)|2,{113,99,0}},
/* 60481 */ {(12<<2)|2,{113,99,0}},
/* 60482 */ {(12<<2)|2,{113,99,0}},
/* 60483 */ {(12<<2)|2,{113,99,0}},
/* 60484 */ {(12<<2)|2,{113,99,0}},
/* 60485 */ {(12<<2)|2,{113,99,0}},
/* 60486 */ {(12<<2)|2,{113,99,0}},
/* 60487 */ {(12<<2)|2,{113,99,0}},
/* 60488 */ {(12<<2)|2,{113,99,0}},
/* 60489 */ {(12<<2)|2,{113,99,0}},
/* 60490 */ {(12<<2)|2,{113,99,0}},
/* 60491 */ {(12<<2)|2,{113,99,0}},
/* 60492 */ {(12<<2)|2,{113,99,0}},
/* 60493 */ {(12<<2)|2,{113,99,0}},
/* 60494 */ {(12<<2)|2,{113,99,0}},
/* 60495 */ {(12<<2)|2,{113,99,0}},
/* 60496 */ {(12<<2)|2,{113,101,0}},
/* 60497 */ {(12<<2)|2,{113,101,0}},
/* 60498 */ {(12<<2)|2,{113,101,0}},
/* 60499 */ {(12<<2)|2,{113,101,0}},
/* 60500 */ {(12<<2)|2,{113,101,0}},
/* 60501 */ {(12<<2)|2,{113,101,0}},
/* 60502 */ {(12<<2)|2,{113,101,0}},
/* 60503 */ {(12<<2)|2,{113,101,0}},
/* 60504 */ {(12<<2)|2,{113,101,0}},
/* 60505 */ {(12<<2)|2,{113,101,0}},
/* 60506 */ {(12<<2)|2,{113,101,0}},
/* 60507 */ {(12<<2)|2,{113,101,0}},
/* 60508 */ {(12<<2)|2,{113,101,0}},
/* 60509 */ {(12<<2)|2,{113,101,0}},
/* 60510 */ {(12<<2)|2,{113,101,0}},
/* 60511 */ {(12<<2)|2,{113,101,0}},
/* 60512 */ {(12<<2)|2,{113,105,0}},
/* 60513 */ {(12<<2)|2,{113,105,0}},
/* 60514 */ {(12<<2)|2,{113,105,0}},
/* 60515 */ {(12<<2)|2,{113,105,0}},
/* 60516 */ {(12<<2)|2,{113,105,0}},
/* 60517 */ {(12<<2)|2,{113,105,0}},
/* 60518 */ {(12<<2)|2,{113,105,0}},
/* 60519 */ {(12<<2)|2,{113,105,0}},
/* 60520 */ {(12<<2)|2,{113,105,0}},
/* 60521 */ {(12<<2)|2,{113,105,0}},
/* 60522 */ {(12<<2)|2,{113,105,0}},
/* 60523 */ {(12<<2)|2,{113,105,0}},
/* 60524 */ {(12<<2)|2,{113,105,0}},
/* 60525 */ {(12<<2)|2,{113,105,0}},
/* 60526 */ {(12<<2)|2,{113,105,0}},
/* 60527 */ {(12<<2)|2,{113,105,0}},
/* 60528 */ {(12<<2)|2,{113,111,0}},
/* 60529 */ {(12<<2)|2,{113,111,0}},
/* 60530 */ {(12<<2)|2,{113,111,0}},
/* 60531 */ {(12<<2)|2,{113,111,0}},
/* 60532 */ {(12<<2)|2,{113,111,0}},
/* 60533 */ {(12<<2)|2,{113,111,0}},
/* 60534 */ {(12<<2)|2,{113,111,0}},
/* 60535 */ {(12<<2)|2,{113,111,0}},
/* 60536 */ {(12<<2)|2,{113,111,0}},
/* 60537 */ {(12<<2)|2,{113,111,0}},
/* 60538 */ {(12<<2)|2,{113,111,0}},
/* 60539 */ {(12<<2)|2,{113,111,0}},
/* 60540 */ {(12<<2)|2,{113,111,0}},
/* 60541 */ {(12<<2)|2,{113,111,0}},
/* 60542 */ {(12<<2)|2,{113,111,0}},
/* 60543 */ {(12<<2)|2,{113,111,0}},
/* 60544 */ {(12<<2)|2,{113,115,0}},
/* 60545 */ {(12<<2)|2,{113,115,0}},
/* 60546 */ {(12<<2)|2,{113,115,0}},
/* 60547 */ {(12<<2)|2,{113,115,0}},
/* 60548 */ {(12<<2)|2,{113,115,0}},
/* 60549 */ {(12<<2)|2,{113,115,0}},
/* 60550 */ {(12<<2)|2,{113,115,0}},
/* 60551 */ {(12<<2)|2,{113,115,0}},
/* 60552 */ {(12<<2)|2,{113,115,0}},
/* 60553 */ {(12<<2)|2,{113,115,0}},
/* 60554 */ {(12<<2)|2,{113,115,0}},
/* 60555 */ {(12<<2)|2,{113,115,0}},
/* 60556 */ {(12<<2)|2,{113,115,0}},
/* 60557 */ {(12<<2)|2,{113,115,0}},
/* 60558 */ {(12<<2)|2,{113,115,0}},
/* 60559 */ {(12<<2)|2,{113,115,0}},
/* 60560 */ {(12<<2)|2,{113,116,0}},
/* 60561 */ {(12<<2)|2,{113,116,0}},
/* 60562 */ {(12<<2)|2,{113,116,0}},
/* 60563 */ {(12<<2)|2,{113,116,0}},
/* 60564 */ {(12<<2)|2,{113,116,0}},
/* 60565 */ {(12<<2)|2,{113,116,0}},
/* 60566 */ {(12<<2)|2,{113,116,0}},
/* 60567 */ {(12<<2)|2,{113,116,0}},
/* 60568 */ {(12<<2)|2,{113,116,0}},
/* 60569 */ {(12<<2)|2,{113,116,0}},
/* 60570 */ {(12<<2)|2,{113,116,0}},
/* 60571 */ {(12<<2)|2,{113,116,0}},
/* 60572 */ {(12<<2)|2,{113,116,0}},
/* 60573 */ {(12<<2)|2,{113,116,0}},
/* 60574 */ {(12<<2)|2,{113,116,0}},
/* 60575 */ {(12<<2)|2,{113,116,0}},
/* 60576 */ {(13<<2)|2,{113,32,0}},
/* 60577 */ {(13<<2)|2,{113,32,0}},
/* 60578 */ {(13<<2)|2,{113,32,0}},
/* 60579 */ {(13<<2)|2,{113,32,0}},
/* 60580 */ {(13<<2)|2,{113,32,0}},
/* 60581 */ {(13<<2)|2,{113,32,0}},
/* 60582 */ {(13<<2)|2,{113,32,0}},
/* 60583 */ {(13<<2)|2,{113,32,0}},
/* 60584 */ {(13<<2)|2,{113,37,0}},
/* 60585 */ {(13<<2)|2,{113,37,0}},
/* 60586 */ {(13<<2)|2,{113,37,0}},
/* 60587 */ {(13<<2)|2,{113,37,0}},
/* 60588 */ {(13<<2)|2,{113,37,0}},
/* 60589 */ {(13<<2)|2,{113,37,0}},
/* 60590 */ {(13<<2)|2,{113,37,0}},
/* 60591 */ {(13<<2)|2,{113,37,0}},
/* 60592 */ {(13<<2)|2,{113,45,0}},
/* 60593 */ {(13<<2)|2,{113,45,0}},
/* 60594 */ {(13<<2)|2,{113,45,0}},
/* 60595 */ {(13<<2)|2,{113,45,0}},
/* 60596 */ {(13<<2)|2,{113,45,0}},
/* 60597 */ {(13<<2)|2,{113,45,0}},
/* 60598 */ {(13<<2)|2,{113,45,0}},
/* 60599 */ {(13<<2)|2,{113,45,0}},
/* 60600 */ {(13<<2)|2,{113,46,0}},
/* 60601 */ {(13<<2)|2,{113,46,0}},
/* 60602 */ {(13<<2)|2,{113,46,0}},
/* 60603 */ {(13<<2)|2,{113,46,0}},
/* 60604 */ {(13<<2)|2,{113,46,0}},
/* 60605 */ {(13<<2)|2,{113,46,0}},
/* 60606 */ {(13<<2)|2,{113,46,0}},
/* 60607 */ {(13<<2)|2,{113,46,0}},
/* 60608 */ {(13<<2)|2,{113,47,0}},
/* 60609 */ {(13<<2)|2,{113,47,0}},
/* 60610 */ {(13<<2)|2,{113,47,0}},
/* 60611 */ {(13<<2)|2,{113,47,0}},
/* 60612 */ {(13<<2)|2,{113,47,0}},
/* 60613 */ {(13<<2)|2,{113,47,0}},
/* 60614 */ {(13<<2)|2,{113,47,0}},
/* 60615 */ {(13<<2)|2,{113,47,0}},
/* 60616 */ {(13<<2)|2,{113,51,0}},
/* 60617 */ {(13<<2)|2,{113,51,0}},
/* 60618 */ {(13<<2)|2,{113,51,0}},
/* 60619 */ {(13<<2)|2,{113,51,0}},
/* 60620 */ {(13<<2)|2,{113,51,0}},
/* 60621 */ {(13<<2)|2,{113,51,0}},
/* 60622 */ {(13<<2)|2,{113,51,0}},
/* 60623 */ {(13<<2)|2,{113,51,0}},
/* 60624 */ {(13<<2)|2,{113,52,0}},
/* 60625 */ {(13<<2)|2,{113,52,0}},
/* 60626 */ {(13<<2)|2,{113,52,0}},
/* 60627 */ {(13<<2)|2,{113,52,0}},
/* 60628 */ {(13<<2)|2,{113,52,0}},
/* 60629 */ {(13<<2)|2,{113,52,0}},
/* 60630 */ {(13<<2)|2,{113,52,0}},
/* 60631 */ {(13<<2)|2,{113,52,0}},
/* 60632 */ {(13<<2)|2,{113,53,0}},
/* 60633 */ {(13<<2)|2,{113,53,0}},
/* 60634 */ {(13<<2)|2,{113,53,0}},
/* 60635 */ {(13<<2)|2,{113,53,0}},
/* 60636 */ {(13<<2)|2,{113,53,0}},
/* 60637 */ {(13<<2)|2,{113,53,0}},
/* 60638 */ {(13<<2)|2,{113,53,0}},
/* 60639 */ {(13<<2)|2,{113,53,0}},
/* 60640 */ {(13<<2)|2,{113,54,0}},
/* 60641 */ {(13<<2)|2,{113,54,0}},
/* 60642 */ {(13<<2)|2,{113,54,0}},
/* 60643 */ {(13<<2)|2,{113,54,0}},
/* 60644 */ {(13<<2)|2,{113,54,0}},
/* 60645 */ {(13<<2)|2,{113,54,0}},
/* 60646 */ {(13<<2)|2,{113,54,0}},
/* 60647 */ {(13<<2)|2,{113,54,0}},
/* 60648 */ {(13<<2)|2,{113,55,0}},
/* 60649 */ {(13<<2)|2,{113,55,0}},
/* 60650 */ {(13<<2)|2,{113,55,0}},
/* 60651 */ {(13<<2)|2,{113,55,0}},
/* 60652 */ {(13<<2)|2,{113,55,0}},
/* 60653 */ {(13<<2)|2,{113,55,0}},
/* 60654 */ {(13<<2)|2,{113,55,0}},
/* 60655 */ {(13<<2)|2,{113,55,0}},
/* 60656 */ {(13<<2)|2,{113,56,0}},
/* 60657 */ {(13<<2)|2,{113,56,0}},
/* 60658 */ {(13<<2)|2,{113,56,0}},
/* 60659 */ {(13<<2)|2,{113,56,0}},
/* 60660 */ {(13<<2)|2,{113,56,0}},
/* 60661 */ {(13<<2)|2,{113,56,0}},
/* 60662 */ {(13<<2)|2,{113,56,0}},
/* 60663 */ {(13<<2)|2,{113,56,0}},
/* 60664 */ {(13<<2)|2,{113,57,0}},
/* 60665 */ {(13<<2)|2,{113,57,0}},
/* 60666 */ {(13<<2)|2,{113,57,0}},
/* 60667 */ {(13<<2)|2,{113,57,0}},
/* 60668 */ {(13<<2)|2,{113,57,0}},
/* 60669 */ {(13<<2)|2,{113,57,0}},
/* 60670 */ {(13<<2)|2,{113,57,0}},
/* 60671 */ {(13<<2)|2,{113,57,0}},
/* 60672 */ {(13<<2)|2,{113,61,0}},
/* 60673 */ {(13<<2)|2,{113,61,0}},
/* 60674 */ {(13<<2)|2,{113,61,0}},
/* 60675 */ {(13<<2)|2,{113,61,0}},
/* 60676 */ {(13<<2)|2,{113,61,0}},
/* 60677 */ {(13<<2)|2,{113,61,0}},
/* 60678 */ {(13<<2)|2,{113,61,0}},
/* 60679 */ {(13<<2)|2,{113,61,0}},
/* 60680 */ {(13<<2)|2,{113,65,0}},
/* 60681 */ {(13<<2)|2,{113,65,0}},
/* 60682 */ {(13<<2)|2,{113,65,0}},
/* 60683 */ {(13<<2)|2,{113,65,0}},
/* 60684 */ {(13<<2)|2,{113,65,0}},
/* 60685 */ {(13<<2)|2,{113,65,0}},
/* 60686 */ {(13<<2)|2,{113,65,0}},
/* 60687 */ {(13<<2)|2,{113,65,0}},
/* 60688 */ {(13<<2)|2,{113,95,0}},
/* 60689 */ {(13<<2)|2,{113,95,0}},
/* 60690 */ {(13<<2)|2,{113,95,0}},
/* 60691 */ {(13<<2)|2,{113,95,0}},
/* 60692 */ {(13<<2)|2,{113,95,0}},
/* 60693 */ {(13<<2)|2,{113,95,0}},
/* 60694 */ {(13<<2)|2,{113,95,0}},
/* 60695 */ {(13<<2)|2,{113,95,0}},
/* 60696 */ {(13<<2)|2,{113,98,0}},
/* 60697 */ {(13<<2)|2,{113,98,0}},
/* 60698 */ {(13<<2)|2,{113,98,0}},
/* 60699 */ {(13<<2)|2,{113,98,0}},
/* 60700 */ {(13<<2)|2,{113,98,0}},
/* 60701 */ {(13<<2)|2,{113,98,0}},
/* 60702 */ {(13<<2)|2,{113,98,0}},
/* 60703 */ {(13<<2)|2,{113,98,0}},
/* 60704 */ {(13<<2)|2,{113,100,0}},
/* 60705 */ {(13<<2)|2,{113,100,0}},
/* 60706 */ {(13<<2)|2,{113,100,0}},
/* 60707 */ {(13<<2)|2,{113,100,0}},
/* 60708 */ {(13<<2)|2,{113,100,0}},
/* 60709 */ {(13<<2)|2,{113,100,0}},
/* 60710 */ {(13<<2)|2,{113,100,0}},
/* 60711 */ {(13<<2)|2,{113,100,0}},
/* 60712 */ {(13<<2)|2,{113,102,0}},
/* 60713 */ {(13<<2)|2,{113,102,0}},
/* 60714 */ {(13<<2)|2,{113,102,0}},
/* 60715 */ {(13<<2)|2,{113,102,0}},
/* 60716 */ {(13<<2)|2,{113,102,0}},
/* 60717 */ {(13<<2)|2,{113,102,0}},
/* 60718 */ {(13<<2)|2,{113,102,0}},
/* 60719 */ {(13<<2)|2,{113,102,0}},
/* 60720 */ {(13<<2)|2,{113,103,0}},
/* 60721 */ {(13<<2)|2,{113,103,0}},
/* 60722 */ {(13<<2)|2,{113,103,0}},
/* 60723 */ {(13<<2)|2,{113,103,0}},
/* 60724 */ {(13<<2)|2,{113,103,0}},
/* 60725 */ {(13<<2)|2,{113,103,0}},
/* 60726 */ {(13<<2)|2,{113,103,0}},
/* 60727 */ {(13<<2)|2,{113,103,0}},
/* 60728 */ {(13<<2)|2,{113,104,0}},
/* 60729 */ {(13<<2)|2,{113,104,0}},
/* 60730 */ {(13<<2)|2,{113,104,0}},
/* 60731 */ {(13<<2)|2,{113,104,0}},
/* 60732 */ {(13<<2)|2,{113,104,0}},
/* 60733 */ {(13<<2)|2,{113,104,0}},
/* 60734 */ {(13<<2)|2,{113,104,0}},
/* 60735 */ {(13<<2)|2,{113,104,0}},
/* 60736 */ {(13<<2)|2,{113,108,0}},
/* 60737 */ {(13<<2)|2,{113,108,0}},
/* 60738 */ {(13<<2)|2,{113,108,0}},
/* 60739 */ {(13<<2)|2,{113,108,0}},
/* 60740 */ {(13<<2)|2,{113,108,0}},
/* 60741 */ {(13<<2)|2,{113,108,0}},
/* 60742 */ {(13<<2)|2,{113,108,0}},
/* 60743 */ {(13<<2)|2,{113,108,0}},
/* 60744 */ {(13<<2)|2,{113,109,0}},
/* 60745 */ {(13<<2)|2,{113,109,0}},
/* 60746 */ {(13<<2)|2,{113,109,0}},
/* 60747 */ {(13<<2)|2,{113,109,0}},
/* 60748 */ {(13<<2)|2,{113,109,0}},
/* 60749 */ {(13<<2)|2,{113,109,0}},
/* 60750 */ {(13<<2)|2,{113,109,0}},
/* 60751 */ {(13<<2)|2,{113,109,0}},
/* 60752 */ {(13<<2)|2,{113,110,0}},
/* 60753 */ {(13<<2)|2,{113,110,0}},
/* 60754 */ {(13<<2)|2,{113,110,0}},
/* 60755 */ {(13<<2)|2,{113,110,0}},
/* 60756 */ {(13<<2)|2,{113,110,0}},
/* 60757 */ {(13<<2)|2,{113,110,0}},
/* 60758 */ {(13<<2)|2,{113,110,0}},
/* 60759 */ {(13<<2)|2,{113,110,0}},
/* 60760 */ {(13<<2)|2,{113,112,0}},
/* 60761 */ {(13<<2)|2,{113,112,0}},
/* 60762 */ {(13<<2)|2,{113,112,0}},
/* 60763 */ {(13<<2)|2,{113,112,0}},
/* 60764 */ {(13<<2)|2,{113,112,0}},
/* 60765 */ {(13<<2)|2,{113,112,0}},
/* 60766 */ {(13<<2)|2,{113,112,0}},
/* 60767 */ {(13<<2)|2,{113,112,0}},
/* 60768 */ {(13<<2)|2,{113,114,0}},
/* 60769 */ {(13<<2)|2,{113,114,0}},
/* 60770 */ {(13<<2)|2,{113,114,0}},
/* 60771 */ {(13<<2)|2,{113,114,0}},
/* 60772 */ {(13<<2)|2,{113,114,0}},
/* 60773 */ {(13<<2)|2,{113,114,0}},
/* 60774 */ {(13<<2)|2,{113,114,0}},
/* 60775 */ {(13<<2)|2,{113,114,0}},
/* 60776 */ {(13<<2)|2,{113,117,0}},
/* 60777 */ {(13<<2)|2,{113,117,0}},
/* 60778 */ {(13<<2)|2,{113,117,0}},
/* 60779 */ {(13<<2)|2,{113,117,0}},
/* 60780 */ {(13<<2)|2,{113,117,0}},
/* 60781 */ {(13<<2)|2,{113,117,0}},
/* 60782 */ {(13<<2)|2,{113,117,0}},
/* 60783 */ {(13<<2)|2,{113,117,0}},
/* 60784 */ {(14<<2)|2,{113,58,0}},
/* 60785 */ {(14<<2)|2,{113,58,0}},
/* 60786 */ {(14<<2)|2,{113,58,0}},
/* 60787 */ {(14<<2)|2,{113,58,0}},
/* 60788 */ {(14<<2)|2,{113,66,0}},
/* 60789 */ {(14<<2)|2,{113,66,0}},
/* 60790 */ {(14<<2)|2,{113,66,0}},
/* 60791 */ {(14<<2)|2,{113,66,0}},
/* 60792 */ {(14<<2)|2,{113,67,0}},
/* 60793 */ {(14<<2)|2,{113,67,0}},
/* 60794 */ {(14<<2)|2,{113,67,0}},
/* 60795 */ {(14<<2)|2,{113,67,0}},
/* 60796 */ {(14<<2)|2,{113,68,0}},
/* 60797 */ {(14<<2)|2,{113,68,0}},
/* 60798 */ {(14<<2)|2,{113,68,0}},
/* 60799 */ {(14<<2)|2,{113,68,0}},
/* 60800 */ {(14<<2)|2,{113,69,0}},
/* 60801 */ {(14<<2)|2,{113,69,0}},
/* 60802 */ {(14<<2)|2,{113,69,0}},
/* 60803 */ {(14<<2)|2,{113,69,0}},
/* 60804 */ {(14<<2)|2,{113,70,0}},
/* 60805 */ {(14<<2)|2,{113,70,0}},
/* 60806 */ {(14<<2)|2,{113,70,0}},
/* 60807 */ {(14<<2)|2,{113,70,0}},
/* 60808 */ {(14<<2)|2,{113,71,0}},
/* 60809 */ {(14<<2)|2,{113,71,0}},
/* 60810 */ {(14<<2)|2,{113,71,0}},
/* 60811 */ {(14<<2)|2,{113,71,0}},
/* 60812 */ {(14<<2)|2,{113,72,0}},
/* 60813 */ {(14<<2)|2,{113,72,0}},
/* 60814 */ {(14<<2)|2,{113,72,0}},
/* 60815 */ {(14<<2)|2,{113,72,0}},
/* 60816 */ {(14<<2)|2,{113,73,0}},
/* 60817 */ {(14<<2)|2,{113,73,0}},
/* 60818 */ {(14<<2)|2,{113,73,0}},
/* 60819 */ {(14<<2)|2,{113,73,0}},
/* 60820 */ {(14<<2)|2,{113,74,0}},
/* 60821 */ {(14<<2)|2,{113,74,0}},
/* 60822 */ {(14<<2)|2,{113,74,0}},
/* 60823 */ {(14<<2)|2,{113,74,0}},
/* 60824 */ {(14<<2)|2,{113,75,0}},
/* 60825 */ {(14<<2)|2,{113,75,0}},
/* 60826 */ {(14<<2)|2,{113,75,0}},
/* 60827 */ {(14<<2)|2,{113,75,0}},
/* 60828 */ {(14<<2)|2,{113,76,0}},
/* 60829 */ {(14<<2)|2,{113,76,0}},
/* 60830 */ {(14<<2)|2,{113,76,0}},
/* 60831 */ {(14<<2)|2,{113,76,0}},
/* 60832 */ {(14<<2)|2,{113,77,0}},
/* 60833 */ {(14<<2)|2,{113,77,0}},
/* 60834 */ {(14<<2)|2,{113,77,0}},
/* 60835 */ {(14<<2)|2,{113,77,0}},
/* 60836 */ {(14<<2)|2,{113,78,0}},
/* 60837 */ {(14<<2)|2,{113,78,0}},
/* 60838 */ {(14<<2)|2,{113,78,0}},
/* 60839 */ {(14<<2)|2,{113,78,0}},
/* 60840 */ {(14<<2)|2,{113,79,0}},
/* 60841 */ {(14<<2)|2,{113,79,0}},
/* 60842 */ {(14<<2)|2,{113,79,0}},
/* 60843 */ {(14<<2)|2,{113,79,0}},
/* 60844 */ {(14<<2)|2,{113,80,0}},
/* 60845 */ {(14<<2)|2,{113,80,0}},
/* 60846 */ {(14<<2)|2,{113,80,0}},
/* 60847 */ {(14<<2)|2,{113,80,0}},
/* 60848 */ {(14<<2)|2,{113,81,0}},
/* 60849 */ {(14<<2)|2,{113,81,0}},
/* 60850 */ {(14<<2)|2,{113,81,0}},
/* 60851 */ {(14<<2)|2,{113,81,0}},
/* 60852 */ {(14<<2)|2,{113,82,0}},
/* 60853 */ {(14<<2)|2,{113,82,0}},
/* 60854 */ {(14<<2)|2,{113,82,0}},
/* 60855 */ {(14<<2)|2,{113,82,0}},
/* 60856 */ {(14<<2)|2,{113,83,0}},
/* 60857 */ {(14<<2)|2,{113,83,0}},
/* 60858 */ {(14<<2)|2,{113,83,0}},
/* 60859 */ {(14<<2)|2,{113,83,0}},
/* 60860 */ {(14<<2)|2,{113,84,0}},
/* 60861 */ {(14<<2)|2,{113,84,0}},
/* 60862 */ {(14<<2)|2,{113,84,0}},
/* 60863 */ {(14<<2)|2,{113,84,0}},
/* 60864 */ {(14<<2)|2,{113,85,0}},
/* 60865 */ {(14<<2)|2,{113,85,0}},
/* 60866 */ {(14<<2)|2,{113,85,0}},
/* 60867 */ {(14<<2)|2,{113,85,0}},
/* 60868 */ {(14<<2)|2,{113,86,0}},
/* 60869 */ {(14<<2)|2,{113,86,0}},
/* 60870 */ {(14<<2)|2,{113,86,0}},
/* 60871 */ {(14<<2)|2,{113,86,0}},
/* 60872 */ {(14<<2)|2,{113,87,0}},
/* 60873 */ {(14<<2)|2,{113,87,0}},
/* 60874 */ {(14<<2)|2,{113,87,0}},
/* 60875 */ {(14<<2)|2,{113,87,0}},
/* 60876 */ {(14<<2)|2,{113,89,0}},
/* 60877 */ {(14<<2)|2,{113,89,0}},
/* 60878 */ {(14<<2)|2,{113,89,0}},
/* 60879 */ {(14<<2)|2,{113,89,0}},
/* 60880 */ {(14<<2)|2,{113,106,0}},
/* 60881 */ {(14<<2)|2,{113,106,0}},
/* 60882 */ {(14<<2)|2,{113,106,0}},
/* 60883 */ {(14<<2)|2,{113,106,0}},
/* 60884 */ {(14<<2)|2,{113,107,0}},
/* 60885 */ {(14<<2)|2,{113,107,0}},
/* 60886 */ {(14<<2)|2,{113,107,0}},
/* 60887 */ {(14<<2)|2,{113,107,0}},
/* 60888 */ {(14<<2)|2,{113,113,0}},
/* 60889 */ {(14<<2)|2,{113,113,0}},
/* 60890 */ {(14<<2)|2,{113,113,0}},
/* 60891 */ {(14<<2)|2,{113,113,0}},
/* 60892 */ {(14<<2)|2,{113,118,0}},
/* 60893 */ {(14<<2)|2,{113,118,0}},
/* 60894 */ {(14<<2)|2,{113,118,0}},
/* 60895 */ {(14<<2)|2,{113,118,0}},
/* 60896 */ {(14<<2)|2,{113,119,0}},
/* 60897 */ {(14<<2)|2,{113,119,0}},
/* 60898 */ {(14<<2)|2,{113,119,0}},
/* 60899 */ {(14<<2)|2,{113,119,0}},
/* 60900 */ {(14<<2)|2,{113,120,0}},
/* 60901 */ {(14<<2)|2,{113,120,0}},
/* 60902 */ {(14<<2)|2,{113,120,0}},
/* 60903 */ {(14<<2)|2,{113,120,0}},
/* 60904 */ {(14<<2)|2,{113,121,0}},
/* 60905 */ {(14<<2)|2,{113,121,0}},
/* 60906 */ {(14<<2)|2,{113,121,0}},
/* 60907 */ {(14<<2)|2,{113,121,0}},
/* 60908 */ {(14<<2)|2,{113,122,0}},
/* 60909 */ {(14<<2)|2,{113,122,0}},
/* 60910 */ {(14<<2)|2,{113,122,0}},
/* 60911 */ {(14<<2)|2,{113,122,0}},
/* 60912 */ {(15<<2)|2,{113,38,0}},
/* 60913 */ {(15<<2)|2,{113,38,0}},
/* 60914 */ {(15<<2)|2,{113,42,0}},
/* 60915 */ {(15<<2)|2,{113,42,0}},
/* 60916 */ {(15<<2)|2,{113,44,0}},
/* 60917 */ {(15<<2)|2,{113,44,0}},
/* 60918 */ {(15<<2)|2,{113,59,0}},
/* 60919 */ {(15<<2)|2,{113,59,0}},
/* 60920 */ {(15<<2)|2,{113,88,0}},
/* 60921 */ {(15<<2)|2,{113,88,0}},
/* 60922 */ {(15<<2)|2,{113,90,0}},
/* 60923 */ {(15<<2)|2,{113,90,0}},
/* 60924 */ {(7<<2)|1,{113,0,0}},
/* 60925 */ {(7<<2)|1,{113,0,0}},
/* 60926 */ {(7<<2)|1,{113,0,0}},
/* 60927 */ {(7<<2)|1,{113,0,0}},
/* 60928 */ {(12<<2)|2,{118,48,0}},
/* 60929 */ {(12<<2)|2,{118,48,0}},
/* 60930 */ {(12<<2)|2,{118,48,0}},
/* 60931 */ {(12<<2)|2,{118,48,0}},
/* 60932 */ {(12<<2)|2,{118,48,0}},
/* 60933 */ {(12<<2)|2,{118,48,0}},
/* 60934 */ {(12<<2)|2,{118,48,0}},
/* 60935 */ {(12<<2)|2,{118,48,0}},
/* 60936 */ {(12<<2)|2,{118,48,0}},
/* 60937 */ {(12<<2)|2,{118,48,0}},
/* 60938 */ {(12<<2)|2,{118,48,0}},
/* 60939 */ {(12<<2)|2,{118,48,0}},
/* 60940 */ {(12<<2)|2,{118,48,0}},
/* 60941 */ {(12<<2)|2,{118,48,0}},
/* 60942 */ {(12<<2)|2,{118,48,0}},
/* 60943 */ {(12<<2)|2,{118,48,0}},
/* 60944 */ {(12<<2)|2,{118,49,0}},
/* 60945 */ {(12<<2)|2,{118,49,0}},
/* 60946 */ {(12<<2)|2,{118,49,0}},
/* 60947 */ {(12<<2)|2,{118,49,0}},
/* 60948 */ {(12<<2)|2,{118,49,0}},
/* 60949 */ {(12<<2)|2,{118,49,0}},
/* 60950 */ {(12<<2)|2,{118,49,0}},
/* 60951 */ {(12<<2)|2,{118,49,0}},
/* 60952 */ {(12<<2)|2,{118,49,0}},
/* 60953 */ {(12<<2)|2,{118,49,0}},
/* 60954 */ {(12<<2)|2,{118,49,0}},
/* 60955 */ {(12<<2)|2,{118,49,0}},
/* 60956 */ {(12<<2)|2,{118,49,0}},
/* 60957 */ {(12<<2)|2,{118,49,0}},
/* 60958 */ {(12<<2)|2,{118,49,0}},
/* 60959 */ {(12<<2)|2,{118,49,0}},
/* 60960 */ {(12<<2)|2,{118,50,0}},
/* 60961 */ {(12<<2)|2,{118,50,0}},
/* 60962 */ {(12<<2)|2,{118,50,0}},
/* 60963 */ {(12<<2)|2,{118,50,0}},
/* 60964 */ {(12<<2)|2,{118,50,0}},
/* 60965 */ {(12<<2)|2,{118,50,0}},
/* 60966 */ {(12<<2)|2,{118,50,0}},
/* 60967 */ {(12<<2)|2,{118,50,0}},
/* 60968 */ {(12<<2)|2,{118,50,0}},
/* 60969 */ {(12<<2)|2,{118,50,0}},
/* 60970 */ {(12<<2)|2,{118,50,0}},
/* 60971 */ {(12<<2)|2,{118,50,0}},
/* 60972 */ {(12<<2)|2,{118,50,0}},
/* 60973 */ {(12<<2)|2,{118,50,0}},
/* 60974 */ {(12<<2)|2,{118,50,0}},
/* 60975 */ {(12<<2)|2,{118,50,0}},
/* 60976 */ {(12<<2)|2,{118,97,0}},
/* 60977 */ {(12<<2)|2,{118,97,0}},
/* 60978 */ {(12<<2)|2,{118,97,0}},
/* 60979 */ {(12<<2)|2,{118,97,0}},
/* 60980 */ {(12<<2)|2,{118,97,0}},
/* 60981 */ {(12<<2)|2,{118,97,0}},
/* 60982 */ {(12<<2)|2,{118,97,0}},
/* 60983 */ {(12<<2)|2,{118,97,0}},
/* 60984 */ {(12<<2)|2,{118,97,0}},
/* 60985 */ {(12<<2)|2,{118,97,0}},
/* 60986 */ {(12<<2)|2,{118,97,0}},
/* 60987 */ {(12<<2)|2,{118,97,0}},
/* 60988 */ {(12<<2)|2,{118,97,0}},
/* 60989 */ {(12<<2)|2,{118,97,0}},
/* 60990 */ {(12<<2)|2,{118,97,0}},
/* 60991 */ {(12<<2)|2,{118,97,0}},
/* 60992 */ {(12<<2)|2,{118,99,0}},
/* 60993 */ {(12<<2)|2,{118,99,0}},
/* 60994 */ {(12<<2)|2,{118,99,0}},
/* 60995 */ {(12<<2)|2,{118,99,0}},
/* 60996 */ {(12<<2)|2,{118,99,0}},
/* 60997 */ {(12<<2)|2,{118,99,0}},
/* 60998 */ {(12<<2)|2,{118,99,0}},
/* 60999 */ {(12<<2)|2,{118,99,0}},
/* 61000 */ {(12<<2)|2,{118,99,0}},
/* 61001 */ {(12<<2)|2,{118,99,0}},
/* 61002 */ {(12<<2)|2,{118,99,0}},
/* 61003 */ {(12<<2)|2,{118,99,0}},
/* 61004 */ {(12<<2)|2,{118,99,0}},
/* 61005 */ {(12<<2)|2,{118,99,0}},
/* 61006 */ {(12<<2)|2,{118,99,0}},
/* 61007 */ {(12<<2)|2,{118,99,0}},
/* 61008 */ {(12<<2)|2,{118,101,0}},
/* 61009 */ {(12<<2)|2,{118,101,0}},
/* 61010 */ {(12<<2)|2,{118,101,0}},
/* 61011 */ {(12<<2)|2,{118,101,0}},
/* 61012 */ {(12<<2)|2,{118,101,0}},
/* 61013 */ {(12<<2)|2,{118,101,0}},
/* 61014 */ {(12<<2)|2,{118,101,0}},
/* 61015 */ {(12<<2)|2,{118,101,0}},
/* 61016 */ {(12<<2)|2,{118,101,0}},
/* 61017 */ {(12<<2)|2,{118,101,0}},
/* 61018 */ {(12<<2)|2,{118,101,0}},
/* 61019 */ {(12<<2)|2,{118,101,0}},
/* 61020 */ {(12<<2)|2,{118,101,0}},
/* 61021 */ {(12<<2)|2,{118,101,0}},
/* 61022 */ {(12<<2)|2,{118,101,0}},
/* 61023 */ {(12<<2)|2,{118,101,0}},
/* 61024 */ {(12<<2)|2,{118,105,0}},
/* 61025 */ {(12<<2)|2,{118,105,0}},
/* 61026 */ {(12<<2)|2,{118,105,0}},
/* 61027 */ {(12<<2)|2,{118,105,0}},
/* 61028 */ {(12<<2)|2,{118,105,0}},
/* 61029 */ {(12<<2)|2,{118,105,0}},
/* 61030 */ {(12<<2)|2,{118,105,0}},
/* 61031 */ {(12<<2)|2,{118,105,0}},
/* 61032 */ {(12<<2)|2,{118,105,0}},
/* 61033 */ {(12<<2)|2,{118,105,0}},
/* 61034 */ {(12<<2)|2,{118,105,0}},
/* 61035 */ {(12<<2)|2,{118,105,0}},
/* 61036 */ {(12<<2)|2,{118,105,0}},
/* 61037 */ {(12<<2)|2,{118,105,0}},
/* 61038 */ {(12<<2)|2,{118,105,0}},
/* 61039 */ {(12<<2)|2,{118,105,0}},
/* 61040 */ {(12<<2)|2,{118,111,0}},
/* 61041 */ {(12<<2)|2,{118,111,0}},
/* 61042 */ {(12<<2)|2,{118,111,0}},
/* 61043 */ {(12<<2)|2,{118,111,0}},
/* 61044 */ {(12<<2)|2,{118,111,0}},
/* 61045 */ {(12<<2)|2,{118,111,0}},
/* 61046 */ {(12<<2)|2,{118,111,0}},
/* 61047 */ {(12<<2)|2,{118,111,0}},
/* 61048 */ {(12<<2)|2,{118,111,0}},
/* 61049 */ {(12<<2)|2,{118,111,0}},
/* 61050 */ {(12<<2)|2,{118,111,0}},
/* 61051 */ {(12<<2)|2,{118,111,0}},
/* 61052 */ {(12<<2)|2,{118,111,0}},
/* 61053 */ {(12<<2)|2,{118,111,0}},
/* 61054 */ {(12<<2)|2,{118,111,0}},
/* 61055 */ {(12<<2)|2,{118,111,0}},
/* 61056 */ {(12<<2)|2,{118,115,0}},
/* 61057 */ {(12<<2)|2,{118,115,0}},
/* 61058 */ {(12<<2)|2,{118,115,0}},
/* 61059 */ {(12<<2)|2,{118,115,0}},
/* 61060 */ {(12<<2)|2,{118,115,0}},
/* 61061 */ {(12<<2)|2,{118,115,0}},
/* 61062 */ {(12<<2)|2,{118,115,0}},
/* 61063 */ {(12<<2)|2,{118,115,0}},
/* 61064 */ {(12<<2)|2,{118,115,0}},
/* 61065 */ {(12<<2)|2,{118,115,0}},
/* 61066 */ {(12<<2)|2,{118,115,0}},
/* 61067 */ {(12<<2)|2,{118,115,0}},
/* 61068 */ {(12<<2)|2,{118,115,0}},
/* 61069 */ {(12<<2)|2,{118,115,0}},
/* 61070 */ {(12<<2)|2,{118,115,0}},
/* 61071 */ {(12<<2)|2,{118,115,0}},
/* 61072 */ {(12<<2)|2,{118,116,0}},
/* 61073 */ {(12<<2)|2,{118,116,0}},
/* 61074 */ {(12<<2)|2,{118,116,0}},
/* 61075 */ {(12<<2)|2,{118,116,0}},
/* 61076 */ {(12<<2)|2,{118,116,0}},
/* 61077 */ {(12<<2)|2,{118,116,0}},
/* 61078 */ {(12<<2)|2,{118,116,0}},
/* 61079 */ {(12<<2)|2,{118,116,0}},
/* 61080 */ {(12<<2)|2,{118,116,0}},
/* 61081 */ {(12<<2)|2,{118,116,0}},
/* 61082 */ {(12<<2)|2,{118,116,0}},
/* 61083 */ {(12<<2)|2,{118,116,0}},
/* 61084 */ {(12<<2)|2,{118,116,0}},
/* 61085 */ {(12<<2)|2,{118,116,0}},
/* 61086 */ {(12<<2)|2,{118,116,0}},
/* 61087 */ {(12<<2)|2,{118,116,0}},
/* 61088 */ {(13<<2)|2,{118,32,0}},
/* 61089 */ {(13<<2)|2,{118,32,0}},
/* 61090 */ {(13<<2)|2,{118,32,0}},
/* 61091 */ {(13<<2)|2,{118,32,0}},
/* 61092 */ {(13<<2)|2,{118,32,0}},
/* 61093 */ {(13<<2)|2,{118,32,0}},
/* 61094 */ {(13<<2)|2,{118,32,0}},
/* 61095 */ {(13<<2)|2,{118,32,0}},
/* 61096 */ {(13<<2)|2,{118,37,0}},
/* 61097 */ {(13<<2)|2,{118,37,0}},
/* 61098 */ {(13<<2)|2,{118,37,0}},
/* 61099 */ {(13<<2)|2,{118,37,0}},
/* 61100 */ {(13<<2)|2,{118,37,0}},
/* 61101 */ {(13<<2)|2,{118,37,0}},
/* 61102 */ {(13<<2)|2,{118,37,0}},
/* 61103 */ {(13<<2)|2,{118,37,0}},
/* 61104 */ {(13<<2)|2,{118,45,0}},
/* 61105 */ {(13<<2)|2,{118,45,0}},
/* 61106 */ {(13<<2)|2,{118,45,0}},
/* 61107 */ {(13<<2)|2,{118,45,0}},
/* 61108 */ {(13<<2)|2,{118,45,0}},
/* 61109 */ {(13<<2)|2,{118,45,0}},
/* 61110 */ {(13<<2)|2,{118,45,0}},
/* 61111 */ {(13<<2)|2,{118,45,0}},
/* 61112 */ {(13<<2)|2,{118,46,0}},
/* 61113 */ {(13<<2)|2,{118,46,0}},
/* 61114 */ {(13<<2)|2,{118,46,0}},
/* 61115 */ {(13<<2)|2,{118,46,0}},
/* 61116 */ {(13<<2)|2,{118,46,0}},
/* 61117 */ {(13<<2)|2,{118,46,0}},
/* 61118 */ {(13<<2)|2,{118,46,0}},
/* 61119 */ {(13<<2)|2,{118,46,0}},
/* 61120 */ {(13<<2)|2,{118,47,0}},
/* 61121 */ {(13<<2)|2,{118,47,0}},
/* 61122 */ {(13<<2)|2,{118,47,0}},
/* 61123 */ {(13<<2)|2,{118,47,0}},
/* 61124 */ {(13<<2)|2,{118,47,0}},
/* 61125 */ {(13<<2)|2,{118,47,0}},
/* 61126 */ {(13<<2)|2,{118,47,0}},
/* 61127 */ {(13<<2)|2,{118,47,0}},
/* 61128 */ {(13<<2)|2,{118,51,0}},
/* 61129 */ {(13<<2)|2,{118,51,0}},
/* 61130 */ {(13<<2)|2,{118,51,0}},
/* 61131 */ {(13<<2)|2,{118,51,0}},
/* 61132 */ {(13<<2)|2,{118,51,0}},
/* 61133 */ {(13<<2)|2,{118,51,0}},
/* 61134 */ {(13<<2)|2,{118,51,0}},
/* 61135 */ {(13<<2)|2,{118,51,0}},
/* 61136 */ {(13<<2)|2,{118,52,0}},
/* 61137 */ {(13<<2)|2,{118,52,0}},
/* 61138 */ {(13<<2)|2,{118,52,0}},
/* 61139 */ {(13<<2)|2,{118,52,0}},
/* 61140 */ {(13<<2)|2,{118,52,0}},
/* 61141 */ {(13<<2)|2,{118,52,0}},
/* 61142 */ {(13<<2)|2,{118,52,0}},
/* 61143 */ {(13<<2)|2,{118,52,0}},
/* 61144 */ {(13<<2)|2,{118,53,0}},
/* 61145 */ {(13<<2)|2,{118,53,0}},
/* 61146 */ {(13<<2)|2,{118,53,0}},
/* 61147 */ {(13<<2)|2,{118,53,0}},
/* 61148 */ {(13<<2)|2,{118,53,0}},
/* 61149 */ {(13<<2)|2,{118,53,0}},
/* 61150 */ {(13<<2)|2,{118,53,0}},
/* 61151 */ {(13<<2)|2,{118,53,0}},
/* 61152 */ {(13<<2)|2,{118,54,0}},
/* 61153 */ {(13<<2)|2,{118,54,0}},
/* 61154 */ {(13<<2)|2,{118,54,0}},
/* 61155 */ {(13<<2)|2,{118,54,0}},
/* 61156 */ {(13<<2)|2,{118,54,0}},
/* 61157 */ {(13<<2)|2,{118,54,0}},
/* 61158 */ {(13<<2)|2,{118,54,0}},
/* 61159 */ {(13<<2)|2,{118,54,0}},
/* 61160 */ {(13<<2)|2,{118,55,0}},
/* 61161 */ {(13<<2)|2,{118,55,0}},
/* 61162 */ {(13<<2)|2,{118,55,0}},
/* 61163 */ {(13<<2)|2,{118,55,0}},
/* 61164 */ {(13<<2)|2,{118,55,0}},
/* 61165 */ {(13<<2)|2,{118,55,0}},
/* 61166 */ {(13<<2)|2,{118,55,0}},
/* 61167 */ {(13<<2)|2,{118,55,0}},
/* 61168 */ {(13<<2)|2,{118,56,0}},
/* 61169 */ {(13<<2)|2,{118,56,0}},
/* 61170 */ {(13<<2)|2,{118,56,0}},
/* 61171 */ {(13<<2)|2,{118,56,0}},
/* 61172 */ {(13<<2)|2,{118,56,0}},
/* 61173 */ {(13<<2)|2,{118,56,0}},
/* 61174 */ {(13<<2)|2,{118,56,0}},
/* 61175 */ {(13<<2)|2,{118,56,0}},
/* 61176 */ {(13<<2)|2,{118,57,0}},
/* 61177 */ {(13<<2)|2,{118,57,0}},
/* 61178 */ {(13<<2)|2,{118,57,0}},
/* 61179 */ {(13<<2)|2,{118,57,0}},
/* 61180 */ {(13<<2)|2,{118,57,0}},
/* 61181 */ {(13<<2)|2,{118,57,0}},
/* 61182 */ {(13<<2)|2,{118,57,0}},
/* 61183 */ {(13<<2)|2,{118,57,0}},
/* 61184 */ {(13<<2)|2,{118,61,0}},
/* 61185 */ {(13<<2)|2,{118,61,0}},
/* 61186 */ {(13<<2)|2,{118,61,0}},
/* 61187 */ {(13<<2)|2,{118,61,0}},
/* 61188 */ {(13<<2)|2,{118,61,0}},
/* 61189 */ {(13<<2)|2,{118,61,0}},
/* 61190 */ {(13<<2)|2,{118,61,0}},
/* 61191 */ {(13<<2)|2,{118,61,0}},
/* 61192 */ {(13<<2)|2,{118,65,0}},
/* 61193 */ {(13<<2)|2,{118,65,0}},
/* 61194 */ {(13<<2)|2,{118,65,0}},
/* 61195 */ {(13<<2)|2,{118,65,0}},
/* 61196 */ {(13<<2)|2,{118,65,0}},
/* 61197 */ {(13<<2)|2,{118,65,0}},
/* 61198 */ {(13<<2)|2,{118,65,0}},
/* 61199 */ {(13<<2)|2,{118,65,0}},
/* 61200 */ {(13<<2)|2,{118,95,0}},
/* 61201 */ {(13<<2)|2,{118,95,0}},
/* 61202 */ {(13<<2)|2,{118,95,0}},
/* 61203 */ {(13<<2)|2,{118,95,0}},
/* 61204 */ {(13<<2)|2,{118,95,0}},
/* 61205 */ {(13<<2)|2,{118,95,0}},
/* 61206 */ {(13<<2)|2,{118,95,0}},
/* 61207 */ {(13<<2)|2,{118,95,0}},
/* 61208 */ {(13<<2)|2,{118,98,0}},
/* 61209 */ {(13<<2)|2,{118,98,0}},
/* 61210 */ {(13<<2)|2,{118,98,0}},
/* 61211 */ {(13<<2)|2,{118,98,0}},
/* 61212 */ {(13<<2)|2,{118,98,0}},
/* 61213 */ {(13<<2)|2,{118,98,0}},
/* 61214 */ {(13<<2)|2,{118,98,0}},
/* 61215 */ {(13<<2)|2,{118,98,0}},
/* 61216 */ {(13<<2)|2,{118,100,0}},
/* 61217 */ {(13<<2)|2,{118,100,0}},
/* 61218 */ {(13<<2)|2,{118,100,0}},
/* 61219 */ {(13<<2)|2,{118,100,0}},
/* 61220 */ {(13<<2)|2,{118,100,0}},
/* 61221 */ {(13<<2)|2,{118,100,0}},
/* 61222 */ {(13<<2)|2,{118,100,0}},
/* 61223 */ {(13<<2)|2,{118,100,0}},
/* 61224 */ {(13<<2)|2,{118,102,0}},
/* 61225 */ {(13<<2)|2,{118,102,0}},
/* 61226 */ {(13<<2)|2,{118,102,0}},
/* 61227 */ {(13<<2)|2,{118,102,0}},
/* 61228 */ {(13<<2)|2,{118,102,0}},
/* 61229 */ {(13<<2)|2,{118,102,0}},
/* 61230 */ {(13<<2)|2,{118,102,0}},
/* 61231 */ {(13<<2)|2,{118,102,0}},
/* 61232 */ {(13<<2)|2,{118,103,0}},
/* 61233 */ {(13<<2)|2,{118,103,0}},
/* 61234 */ {(13<<2)|2,{118,103,0}},
/* 61235 */ {(13<<2)|2,{118,103,0}},
/* 61236 */ {(13<<2)|2,{118,103,0}},
/* 61237 */ {(13<<2)|2,{118,103,0}},
/* 61238 */ {(13<<2)|2,{118,103,0}},
/* 61239 */ {(13<<2)|2,{118,103,0}},
/* 61240 */ {(13<<2)|2,{118,104,0}},
/* 61241 */ {(13<<2)|2,{118,104,0}},
/* 61242 */ {(13<<2)|2,{118,104,0}},
/* 61243 */ {(13<<2)|2,{118,104,0}},
/* 61244 */ {(13<<2)|2,{118,104,0}},
/* 61245 */ {(13<<2)|2,{118,104,0}},
/* 61246 */ {(13<<2)|2,{118,104,0}},
/* 61247 */ {(13<<2)|2,{118,104,0}},
/* 61248 */ {(13<<2)|2,{118,108,0}},
/* 61249 */ {(13<<2)|2,{118,108,0}},
/* 61250 */ {(13<<2)|2,{118,108,0}},
/* 61251 */ {(13<<2)|2,{118,108,0}},
/* 61252 */ {(13<<2)|2,{118,108,0}},
/* 61253 */ {(13<<2)|2,{118,108,0}},
/* 61254 */ {(13<<2)|2,{118,108,0}},
/* 61255 */ {(13<<2)|2,{118,108,0}},
/* 61256 */ {(13<<2)|2,{118,109,0}},
/* 61257 */ {(13<<2)|2,{118,109,0}},
/* 61258 */ {(13<<2)|2,{118,109,0}},
/* 61259 */ {(13<<2)|2,{118,109,0}},
/* 61260 */ {(13<<2)|2,{118,109,0}},
/* 61261 */ {(13<<2)|2,{118,109,0}},
/* 61262 */ {(13<<2)|2,{118,109,0}},
/* 61263 */ {(13<<2)|2,{118,109,0}},
/* 61264 */ {(13<<2)|2,{118,110,0}},
/* 61265 */ {(13<<2)|2,{118,110,0}},
/* 61266 */ {(13<<2)|2,{118,110,0}},
/* 61267 */ {(13<<2)|2,{118,110,0}},
/* 61268 */ {(13<<2)|2,{118,110,0}},
/* 61269 */ {(13<<2)|2,{118,110,0}},
/* 61270 */ {(13<<2)|2,{118,110,0}},
/* 61271 */ {(13<<2)|2,{118,110,0}},
/* 61272 */ {(13<<2)|2,{118,112,0}},
/* 61273 */ {(13<<2)|2,{118,112,0}},
/* 61274 */ {(13<<2)|2,{118,112,0}},
/* 61275 */ {(13<<2)|2,{118,112,0}},
/* 61276 */ {(13<<2)|2,{118,112,0}},
/* 61277 */ {(13<<2)|2,{118,112,0}},
/* 61278 */ {(13<<2)|2,{118,112,0}},
/* 61279 */ {(13<<2)|2,{118,112,0}},
/* 61280 */ {(13<<2)|2,{118,114,0}},
/* 61281 */ {(13<<2)|2,{118,114,0}},
/* 61282 */ {(13<<2)|2,{118,114,0}},
/* 61283 */ {(13<<2)|2,{118,114,0}},
/* 61284 */ {(13<<2)|2,{118,114,0}},
/* 61285 */ {(13<<2)|2,{118,114,0}},
/* 61286 */ {(13<<2)|2,{118,114,0}},
/* 61287 */ {(13<<2)|2,{118,114,0}},
/* 61288 */ {(13<<2)|2,{118,117,0}},
/* 61289 */ {(13<<2)|2,{118,117,0}},
/* 61290 */ {(13<<2)|2,{118,117,0}},
/* 61291 */ {(13<<2)|2,{118,117,0}},
/* 61292 */ {(13<<2)|2,{118,117,0}},
/* 61293 */ {(13<<2)|2,{118,117,0}},
/* 61294 */ {(13<<2)|2,{118,117,0}},
/* 61295 */ {(13<<2)|2,{118,117,0}},
/* 61296 */ {(14<<2)|2,{118,58,0}},
/* 61297 */ {(14<<2)|2,{118,58,0}},
/* 61298 */ {(14<<2)|2,{118,58,0}},
/* 61299 */ {(14<<2)|2,{118,58,0}},
/* 61300 */ {(14<<2)|2,{118,66,0}},
/* 61301 */ {(14<<2)|2,{118,66,0}},
/* 61302 */ {(14<<2)|2,{118,66,0}},
/* 61303 */ {(14<<2)|2,{118,66,0}},
/* 61304 */ {(14<<2)|2,{118,67,0}},
/* 61305 */ {(14<<2)|2,{118,67,0}},
/* 61306 */ {(14<<2)|2,{118,67,0}},
/* 61307 */ {(14<<2)|2,{118,67,0}},
/* 61308 */ {(14<<2)|2,{118,68,0}},
/* 61309 */ {(14<<2)|2,{118,68,0}},
/* 61310 */ {(14<<2)|2,{118,68,0}},
/* 61311 */ {(14<<2)|2,{118,68,0}},
/* 61312 */ {(14<<2)|2,{118,69,0}},
/* 61313 */ {(14<<2)|2,{118,69,0}},
/* 61314 */ {(14<<2)|2,{118,69,0}},
/* 61315 */ {(14<<2)|2,{118,69,0}},
/* 61316 */ {(14<<2)|2,{118,70,0}},
/* 61317 */ {(14<<2)|2,{118,70,0}},
/* 61318 */ {(14<<2)|2,{118,70,0}},
/* 61319 */ {(14<<2)|2,{118,70,0}},
/* 61320 */ {(14<<2)|2,{118,71,0}},
/* 61321 */ {(14<<2)|2,{118,71,0}},
/* 61322 */ {(14<<2)|2,{118,71,0}},
/* 61323 */ {(14<<2)|2,{118,71,0}},
/* 61324 */ {(14<<2)|2,{118,72,0}},
/* 61325 */ {(14<<2)|2,{118,72,0}},
/* 61326 */ {(14<<2)|2,{118,72,0}},
/* 61327 */ {(14<<2)|2,{118,72,0}},
/* 61328 */ {(14<<2)|2,{118,73,0}},
/* 61329 */ {(14<<2)|2,{118,73,0}},
/* 61330 */ {(14<<2)|2,{118,73,0}},
/* 61331 */ {(14<<2)|2,{118,73,0}},
/* 61332 */ {(14<<2)|2,{118,74,0}},
/* 61333 */ {(14<<2)|2,{118,74,0}},
/* 61334 */ {(14<<2)|2,{118,74,0}},
/* 61335 */ {(14<<2)|2,{118,74,0}},
/* 61336 */ {(14<<2)|2,{118,75,0}},
/* 61337 */ {(14<<2)|2,{118,75,0}},
/* 61338 */ {(14<<2)|2,{118,75,0}},
/* 61339 */ {(14<<2)|2,{118,75,0}},
/* 61340 */ {(14<<2)|2,{118,76,0}},
/* 61341 */ {(14<<2)|2,{118,76,0}},
/* 61342 */ {(14<<2)|2,{118,76,0}},
/* 61343 */ {(14<<2)|2,{118,76,0}},
/* 61344 */ {(14<<2)|2,{118,77,0}},
/* 61345 */ {(14<<2)|2,{118,77,0}},
/* 61346 */ {(14<<2)|2,{118,77,0}},
/* 61347 */ {(14<<2)|2,{118,77,0}},
/* 61348 */ {(14<<2)|2,{118,78,0}},
/* 61349 */ {(14<<2)|2,{118,78,0}},
/* 61350 */ {(14<<2)|2,{118,78,0}},
/* 61351 */ {(14<<2)|2,{118,78,0}},
/* 61352 */ {(14<<2)|2,{118,79,0}},
/* 61353 */ {(14<<2)|2,{118,79,0}},
/* 61354 */ {(14<<2)|2,{118,79,0}},
/* 61355 */ {(14<<2)|2,{118,79,0}},
/* 61356 */ {(14<<2)|2,{118,80,0}},
/* 61357 */ {(14<<2)|2,{118,80,0}},
/* 61358 */ {(14<<2)|2,{118,80,0}},
/* 61359 */ {(14<<2)|2,{118,80,0}},
/* 61360 */ {(14<<2)|2,{118,81,0}},
/* 61361 */ {(14<<2)|2,{118,81,0}},
/* 61362 */ {(14<<2)|2,{118,81,0}},
/* 61363 */ {(14<<2)|2,{118,81,0}},
/* 61364 */ {(14<<2)|2,{118,82,0}},
/* 61365 */ {(14<<2)|2,{118,82,0}},
/* 61366 */ {(14<<2)|2,{118,82,0}},
/* 61367 */ {(14<<2)|2,{118,82,0}},
/* 61368 */ {(14<<2)|2,{118,83,0}},
/* 61369 */ {(14<<2)|2,{118,83,0}},
/* 61370 */ {(14<<2)|2,{118,83,0}},
/* 61371 */ {(14<<2)|2,{118,83,0}},
/* 61372 */ {(14<<2)|2,{118,84,0}},
/* 61373 */ {(14<<2)|2,{118,84,0}},
/* 61374 */ {(14<<2)|2,{118,84,0}},
/* 61375 */ {(14<<2)|2,{118,84,0}},
/* 61376 */ {(14<<2)|2,{118,85,0}},
/* 61377 */ {(14<<2)|2,{118,85,0}},
/* 61378 */ {(14<<2)|2,{118,85,0}},
/* 61379 */ {(14<<2)|2,{118,85,0}},
/* 61380 */ {(14<<2)|2,{118,86,0}},
/* 61381 */ {(14<<2)|2,{118,86,0}},
/* 61382 */ {(14<<2)|2,{118,86,0}},
/* 61383 */ {(14<<2)|2,{118,86,0}},
/* 61384 */ {(14<<2)|2,{118,87,0}},
/* 61385 */ {(14<<2)|2,{118,87,0}},
/* 61386 */ {(14<<2)|2,{118,87,0}},
/* 61387 */ {(14<<2)|2,{118,87,0}},
/* 61388 */ {(14<<2)|2,{118,89,0}},
/* 61389 */ {(14<<2)|2,{118,89,0}},
/* 61390 */ {(14<<2)|2,{118,89,0}},
/* 61391 */ {(14<<2)|2,{118,89,0}},
/* 61392 */ {(14<<2)|2,{118,106,0}},
/* 61393 */ {(14<<2)|2,{118,106,0}},
/* 61394 */ {(14<<2)|2,{118,106,0}},
/* 61395 */ {(14<<2)|2,{118,106,0}},
/* 61396 */ {(14<<2)|2,{118,107,0}},
/* 61397 */ {(14<<2)|2,{118,107,0}},
/* 61398 */ {(14<<2)|2,{118,107,0}},
/* 61399 */ {(14<<2)|2,{118,107,0}},
/* 61400 */ {(14<<2)|2,{118,113,0}},
/* 61401 */ {(14<<2)|2,{118,113,0}},
/* 61402 */ {(14<<2)|2,{118,113,0}},
/* 61403 */ {(14<<2)|2,{118,113,0}},
/* 61404 */ {(14<<2)|2,{118,118,0}},
/* 61405 */ {(14<<2)|2,{118,118,0}},
/* 61406 */ {(14<<2)|2,{118,118,0}},
/* 61407 */ {(14<<2)|2,{118,118,0}},
/* 61408 */ {(14<<2)|2,{118,119,0}},
/* 61409 */ {(14<<2)|2,{118,119,0}},
/* 61410 */ {(14<<2)|2,{118,119,0}},
/* 61411 */ {(14<<2)|2,{118,119,0}},
/* 61412 */ {(14<<2)|2,{118,120,0}},
/* 61413 */ {(14<<2)|2,{118,120,0}},
/* 61414 */ {(14<<2)|2,{118,120,0}},
/* 61415 */ {(14<<2)|2,{118,120,0}},
/* 61416 */ {(14<<2)|2,{118,121,0}},
/* 61417 */ {(14<<2)|2,{118,121,0}},
/* 61418 */ {(14<<2)|2,{118,121,0}},
/* 61419 */ {(14<<2)|2,{118,121,0}},
/* 61420 */ {(14<<2)|2,{118,122,0}},
/* 61421 */ {(14<<2)|2,{118,122,0}},
/* 61422 */ {(14<<2)|2,{118,122,0}},
/* 61423 */ {(14<<2)|2,{118,122,0}},
/* 61424 */ {(15<<2)|2,{118,38,0}},
/* 61425 */ {(15<<2)|2,{118,38,0}},
/* 61426 */ {(15<<2)|2,{118,42,0}},
/* 61427 */ {(15<<2)|2,{118,42,0}},
/* 61428 */ {(15<<2)|2,{118,44,0}},
/* 61429 */ {(15<<2)|2,{118,44,0}},
/* 61430 */ {(15<<2)|2,{118,59,0}},
/* 61431 */ {(15<<2)|2,{118,59,0}},
/* 61432 */ {(15<<2)|2,{118,88,0}},
/* 61433 */ {(15<<2)|2,{118,88,0}},
/* 61434 */ {(15<<2)|2,{118,90,0}},
/* 61435 */ {(15<<2)|2,{118,90,0}},
/* 61436 */ {(7<<2)|1,{118,0,0}},
/* 61437 */ {(7<<2)|1,{118,0,0}},
/* 61438 */ {(7<<2)|1,{118,0,0}},
/* 61439 */ {(7<<2)|1,{118,0,0}},
/* 61440 */ {(12<<2)|2,{119,48,0}},
/* 61441 */ {(12<<2)|2,{119,48,0}},
/* 61442 */ {(12<<2)|2,{119,48,0}},
/* 61443 */ {(12<<2)|2,{119,48,0}},
/* 61444 */ {(12<<2)|2,{119,48,0}},
/* 61445 */ {(12<<2)|2,{119,48,0}},
/* 61446 */ {(12<<2)|2,{119,48,0}},
/* 61447 */ {(12<<2)|2,{119,48,0}},
/* 61448 */ {(12<<2)|2,{119,48,0}},
/* 61449 */ {(12<<2)|2,{119,48,0}},
/* 61450 */ {(12<<2)|2,{119,48,0}},
/* 61451 */ {(12<<2)|2,{119,48,0}},
/* 61452 */ {(12<<2)|2,{119,48,0}},
/* 61453 */ {(12<<2)|2,{119,48,0}},
/* 61454 */ {(12<<2)|2,{119,48,0}},
/* 61455 */ {(12<<2)|2,{119,48,0}},
/* 61456 */ {(12<<2)|2,{119,49,0}},
/* 61457 */ {(12<<2)|2,{119,49,0}},
/* 61458 */ {(12<<2)|2,{119,49,0}},
/* 61459 */ {(12<<2)|2,{119,49,0}},
/* 61460 */ {(12<<2)|2,{119,49,0}},
/* 61461 */ {(12<<2)|2,{119,49,0}},
/* 61462 */ {(12<<2)|2,{119,49,0}},
/* 61463 */ {(12<<2)|2,{119,49,0}},
/* 61464 */ {(12<<2)|2,{119,49,0}},
/* 61465 */ {(12<<2)|2,{119,49,0}},
/* 61466 */ {(12<<2)|2,{119,49,0}},
/* 61467 */ {(12<<2)|2,{119,49,0}},
/* 61468 */ {(12<<2)|2,{119,49,0}},
/* 61469 */ {(12<<2)|2,{119,49,0}},
/* 61470 */ {(12<<2)|2,{119,49,0}},
/* 61471 */ {(12<<2)|2,{119,49,0}},
/* 61472 */ {(12<<2)|2,{119,50,0}},
/* 61473 */ {(12<<2)|2,{119,50,0}},
/* 61474 */ {(12<<2)|2,{119,50,0}},
/* 61475 */ {(12<<2)|2,{119,50,0}},
/* 61476 */ {(12<<2)|2,{119,50,0}},
/* 61477 */ {(12<<2)|2,{119,50,0}},
/* 61478 */ {(12<<2)|2,{119,50,0}},
/* 61479 */ {(12<<2)|2,{119,50,0}},
/* 61480 */ {(12<<2)|2,{119,50,0}},
/* 61481 */ {(12<<2)|2,{119,50,0}},
/* 61482 */ {(12<<2)|2,{119,50,0}},
/* 61483 */ {(12<<2)|2,{119,50,0}},
/* 61484 */ {(12<<2)|2,{119,50,0}},
/* 61485 */ {(12<<2)|2,{119,50,0}},
/* 61486 */ {(12<<2)|2,{119,50,0}},
/* 61487 */ {(12<<2)|2,{119,50,0}},
/* 61488 */ {(12<<2)|2,{119,97,0}},
/* 61489 */ {(12<<2)|2,{119,97,0}},
/* 61490 */ {(12<<2)|2,{119,97,0}},
/* 61491 */ {(12<<2)|2,{119,97,0}},
/* 61492 */ {(12<<2)|2,{119,97,0}},
/* 61493 */ {(12<<2)|2,{119,97,0}},
/* 61494 */ {(12<<2)|2,{119,97,0}},
/* 61495 */ {(12<<2)|2,{119,97,0}},
/* 61496 */ {(12<<2)|2,{119,97,0}},
/* 61497 */ {(12<<2)|2,{119,97,0}},
/* 61498 */ {(12<<2)|2,{119,97,0}},
/* 61499 */ {(12<<2)|2,{119,97,0}},
/* 61500 */ {(12<<2)|2,{119,97,0}},
/* 61501 */ {(12<<2)|2,{119,97,0}},
/* 61502 */ {(12<<2)|2,{119,97,0}},
/* 61503 */ {(12<<2)|2,{119,97,0}},
/* 61504 */ {(12<<2)|2,{119,99,0}},
/* 61505 */ {(12<<2)|2,{119,99,0}},
/* 61506 */ {(12<<2)|2,{119,99,0}},
/* 61507 */ {(12<<2)|2,{119,99,0}},
/* 61508 */ {(12<<2)|2,{119,99,0}},
/* 61509 */ {(12<<2)|2,{119,99,0}},
/* 61510 */ {(12<<2)|2,{119,99,0}},
/* 61511 */ {(12<<2)|2,{119,99,0}},
/* 61512 */ {(12<<2)|2,{119,99,0}},
/* 61513 */ {(12<<2)|2,{119,99,0}},
/* 61514 */ {(12<<2)|2,{119,99,0}},
/* 61515 */ {(12<<2)|2,{119,99,0}},
/* 61516 */ {(12<<2)|2,{119,99,0}},
/* 61517 */ {(12<<2)|2,{119,99,0}},
/* 61518 */ {(12<<2)|2,{119,99,0}},
/* 61519 */ {(12<<2)|2,{119,99,0}},
/* 61520 */ {(12<<2)|2,{119,101,0}},
/* 61521 */ {(12<<2)|2,{119,101,0}},
/* 61522 */ {(12<<2)|2,{119,101,0}},
/* 61523 */ {(12<<2)|2,{119,101,0}},
/* 61524 */ {(12<<2)|2,{119,101,0}},
/* 61525 */ {(12<<2)|2,{119,101,0}},
/* 61526 */ {(12<<2)|2,{119,101,0}},
/* 61527 */ {(12<<2)|2,{119,101,0}},
/* 61528 */ {(12<<2)|2,{119,101,0}},
/* 61529 */ {(12<<2)|2,{119,101,0}},
/* 61530 */ {(12<<2)|2,{119,101,0}},
/* 61531 */ {(12<<2)|2,{119,101,0}},
/* 61532 */ {(12<<2)|2,{119,101,0}},
/* 61533 */ {(12<<2)|2,{119,101,0}},
/* 61534 */ {(12<<2)|2,{119,101,0}},
/* 61535 */ {(12<<2)|2,{119,101,0}},
/* 61536 */ {(12<<2)|2,{119,105,0}},
/* 61537 */ {(12<<2)|2,{119,105,0}},
/* 61538 */ {(12<<2)|2,{119,105,0}},
/* 61539 */ {(12<<2)|2,{119,105,0}},
/* 61540 */ {(12<<2)|2,{119,105,0}},
/* 61541 */ {(12<<2)|2,{119,105,0}},
/* 61542 */ {(12<<2)|2,{119,105,0}},
/* 61543 */ {(12<<2)|2,{119,105,0}},
/* 61544 */ {(12<<2)|2,{119,105,0}},
/* 61545 */ {(12<<2)|2,{119,105,0}},
/* 61546 */ {(12<<2)|2,{119,105,0}},
/* 61547 */ {(12<<2)|2,{119,105,0}},
/* 61548 */ {(12<<2)|2,{119,105,0}},
/* 61549 */ {(12<<2)|2,{119,105,0}},
/* 61550 */ {(12<<2)|2,{119,105,0}},
/* 61551 */ {(12<<2)|2,{119,105,0}},
/* 61552 */ {(12<<2)|2,{119,111,0}},
/* 61553 */ {(12<<2)|2,{119,111,0}},
/* 61554 */ {(12<<2)|2,{119,111,0}},
/* 61555 */ {(12<<2)|2,{119,111,0}},
/* 61556 */ {(12<<2)|2,{119,111,0}},
/* 61557 */ {(12<<2)|2,{119,111,0}},
/* 61558 */ {(12<<2)|2,{119,111,0}},
/* 61559 */ {(12<<2)|2,{119,111,0}},
/* 61560 */ {(12<<2)|2,{119,111,0}},
/* 61561 */ {(12<<2)|2,{119,111,0}},
/* 61562 */ {(12<<2)|2,{119,111,0}},
/* 61563 */ {(12<<2)|2,{119,111,0}},
/* 61564 */ {(12<<2)|2,{119,111,0}},
/* 61565 */ {(12<<2)|2,{119,111,0}},
/* 61566 */ {(12<<2)|2,{119,111,0}},
/* 61567 */ {(12<<2)|2,{119,111,0}},
/* 61568 */ {(12<<2)|2,{119,115,0}},
/* 61569 */ {(12<<2)|2,{119,115,0}},
/* 61570 */ {(12<<2)|2,{119,115,0}},
/* 61571 */ {(12<<2)|2,{119,115,0}},
/* 61572 */ {(12<<2)|2,{119,115,0}},
/* 61573 */ {(12<<2)|2,{119,115,0}},
/* 61574 */ {(12<<2)|2,{119,115,0}},
/* 61575 */ {(12<<2)|2,{119,115,0}},
/* 61576 */ {(12<<2)|2,{119,115,0}},
/* 61577 */ {(12<<2)|2,{119,115,0}},
/* 61578 */ {(12<<2)|2,{119,115,0}},
/* 61579 */ {(12<<2)|2,{119,115,0}},
/* 61580 */ {(12<<2)|2,{119,115,0}},
/* 61581 */ {(12<<2)|2,{119,115,0}},
/* 61582 */ {(12<<2)|2,{119,115,0}},
/* 61583 */ {(12<<2)|2,{119,115,0}},
/* 61584 */ {(12<<2)|2,{119,116,0}},
/* 61585 */ {(12<<2)|2,{119,116,0}},
/* 61586 */ {(12<<2)|2,{119,116,0}},
/* 61587 */ {(12<<2)|2,{119,116,0}},
/* 61588 */ {(12<<2)|2,{119,116,0}},
/* 61589 */ {(12<<2)|2,{119,116,0}},
/* 61590 */ {(12<<2)|2,{119,116,0}},
/* 61591 */ {(12<<2)|2,{119,116,0}},
/* 61592 */ {(12<<2)|2,{119,116,0}},
/* 61593 */ {(12<<2)|2,{119,116,0}},
/* 61594 */ {(12<<2)|2,{119,116,0}},
/* 61595 */ {(12<<2)|2,{119,116,0}},
/* 61596 */ {(12<<2)|2,{119,116,0}},
/* 61597 */ {(12<<2)|2,{119,116,0}},
/* 61598 */ {(12<<2)|2,{119,116,0}},
/* 61599 */ {(12<<2)|2,{119,116,0}},
/* 61600 */ {(13<<2)|2,{119,32,0}},
/* 61601 */ {(13<<2)|2,{119,32,0}},
/* 61602 */ {(13<<2)|2,{119,32,0}},
/* 61603 */ {(13<<2)|2,{119,32,0}},
/* 61604 */ {(13<<2)|2,{119,32,0}},
/* 61605 */ {(13<<2)|2,{119,32,0}},
/* 61606 */ {(13<<2)|2,{119,32,0}},
/* 61607 */ {(13<<2)|2,{119,32,0}},
/* 61608 */ {(13<<2)|2,{119,37,0}},
/* 61609 */ {(13<<2)|2,{119,37,0}},
/* 61610 */ {(13<<2)|2,{119,37,0}},
/* 61611 */ {(13<<2)|2,{119,37,0}},
/* 61612 */ {(13<<2)|2,{119,37,0}},
/* 61613 */ {(13<<2)|2,{119,37,0}},
/* 61614 */ {(13<<2)|2,{119,37,0}},
/* 61615 */ {(13<<2)|2,{119,37,0}},
/* 61616 */ {(13<<2)|2,{119,45,0}},
/* 61617 */ {(13<<2)|2,{119,45,0}},
/* 61618 */ {(13<<2)|2,{119,45,0}},
/* 61619 */ {(13<<2)|2,{119,45,0}},
/* 61620 */ {(13<<2)|2,{119,45,0}},
/* 61621 */ {(13<<2)|2,{119,45,0}},
/* 61622 */ {(13<<2)|2,{119,45,0}},
/* 61623 */ {(13<<2)|2,{119,45,0}},
/* 61624 */ {(13<<2)|2,{119,46,0}},
/* 61625 */ {(13<<2)|2,{119,46,0}},
/* 61626 */ {(13<<2)|2,{119,46,0}},
/* 61627 */ {(13<<2)|2,{119,46,0}},
/* 61628 */ {(13<<2)|2,{119,46,0}},
/* 61629 */ {(13<<2)|2,{119,46,0}},
/* 61630 */ {(13<<2)|2,{119,46,0}},
/* 61631 */ {(13<<2)|2,{119,46,0}},
/* 61632 */ {(13<<2)|2,{119,47,0}},
/* 61633 */ {(13<<2)|2,{119,47,0}},
/* 61634 */ {(13<<2)|2,{119,47,0}},
/* 61635 */ {(13<<2)|2,{119,47,0}},
/* 61636 */ {(13<<2)|2,{119,47,0}},
/* 61637 */ {(13<<2)|2,{119,47,0}},
/* 61638 */ {(13<<2)|2,{119,47,0}},
/* 61639 */ {(13<<2)|2,{119,47,0}},
/* 61640 */ {(13<<2)|2,{119,51,0}},
/* 61641 */ {(13<<2)|2,{119,51,0}},
/* 61642 */ {(13<<2)|2,{119,51,0}},
/* 61643 */ {(13<<2)|2,{119,51,0}},
/* 61644 */ {(13<<2)|2,{119,51,0}},
/* 61645 */ {(13<<2)|2,{119,51,0}},
/* 61646 */ {(13<<2)|2,{119,51,0}},
/* 61647 */ {(13<<2)|2,{119,51,0}},
/* 61648 */ {(13<<2)|2,{119,52,0}},
/* 61649 */ {(13<<2)|2,{119,52,0}},
/* 61650 */ {(13<<2)|2,{119,52,0}},
/* 61651 */ {(13<<2)|2,{119,52,0}},
/* 61652 */ {(13<<2)|2,{119,52,0}},
/* 61653 */ {(13<<2)|2,{119,52,0}},
/* 61654 */ {(13<<2)|2,{119,52,0}},
/* 61655 */ {(13<<2)|2,{119,52,0}},
/* 61656 */ {(13<<2)|2,{119,53,0}},
/* 61657 */ {(13<<2)|2,{119,53,0}},
/* 61658 */ {(13<<2)|2,{119,53,0}},
/* 61659 */ {(13<<2)|2,{119,53,0}},
/* 61660 */ {(13<<2)|2,{119,53,0}},
/* 61661 */ {(13<<2)|2,{119,53,0}},
/* 61662 */ {(13<<2)|2,{119,53,0}},
/* 61663 */ {(13<<2)|2,{119,53,0}},
/* 61664 */ {(13<<2)|2,{119,54,0}},
/* 61665 */ {(13<<2)|2,{119,54,0}},
/* 61666 */ {(13<<2)|2,{119,54,0}},
/* 61667 */ {(13<<2)|2,{119,54,0}},
/* 61668 */ {(13<<2)|2,{119,54,0}},
/* 61669 */ {(13<<2)|2,{119,54,0}},
/* 61670 */ {(13<<2)|2,{119,54,0}},
/* 61671 */ {(13<<2)|2,{119,54,0}},
/* 61672 */ {(13<<2)|2,{119,55,0}},
/* 61673 */ {(13<<2)|2,{119,55,0}},
/* 61674 */ {(13<<2)|2,{119,55,0}},
/* 61675 */ {(13<<2)|2,{119,55,0}},
/* 61676 */ {(13<<2)|2,{119,55,0}},
/* 61677 */ {(13<<2)|2,{119,55,0}},
/* 61678 */ {(13<<2)|2,{119,55,0}},
/* 61679 */ {(13<<2)|2,{119,55,0}},
/* 61680 */ {(13<<2)|2,{119,56,0}},
/* 61681 */ {(13<<2)|2,{119,56,0}},
/* 61682 */ {(13<<2)|2,{119,56,0}},
/* 61683 */ {(13<<2)|2,{119,56,0}},
/* 61684 */ {(13<<2)|2,{119,56,0}},
/* 61685 */ {(13<<2)|2,{119,56,0}},
/* 61686 */ {(13<<2)|2,{119,56,0}},
/* 61687 */ {(13<<2)|2,{119,56,0}},
/* 61688 */ {(13<<2)|2,{119,57,0}},
/* 61689 */ {(13<<2)|2,{119,57,0}},
/* 61690 */ {(13<<2)|2,{119,57,0}},
/* 61691 */ {(13<<2)|2,{119,57,0}},
/* 61692 */ {(13<<2)|2,{119,57,0}},
/* 61693 */ {(13<<2)|2,{119,57,0}},
/* 61694 */ {(13<<2)|2,{119,57,0}},
/* 61695 */ {(13<<2)|2,{119,57,0}},
/* 61696 */ {(13<<2)|2,{119,61,0}},
/* 61697 */ {(13<<2)|2,{119,61,0}},
/* 61698 */ {(13<<2)|2,{119,61,0}},
/* 61699 */ {(13<<2)|2,{119,61,0}},
/* 61700 */ {(13<<2)|2,{119,61,0}},
/* 61701 */ {(13<<2)|2,{119,61,0}},
/* 61702 */ {(13<<2)|2,{119,61,0}},
/* 61703 */ {(13<<2)|2,{119,61,0}},
/* 61704 */ {(13<<2)|2,{119,65,0}},
/* 61705 */ {(13<<2)|2,{119,65,0}},
/* 61706 */ {(13<<2)|2,{119,65,0}},
/* 61707 */ {(13<<2)|2,{119,65,0}},
/* 61708 */ {(13<<2)|2,{119,65,0}},
/* 61709 */ {(13<<2)|2,{119,65,0}},
/* 61710 */ {(13<<2)|2,{119,65,0}},
/* 61711 */ {(13<<2)|2,{119,65,0}},
/* 61712 */ {(13<<2)|2,{119,95,0}},
/* 61713 */ {(13<<2)|2,{119,95,0}},
/* 61714 */ {(13<<2)|2,{119,95,0}},
/* 61715 */ {(13<<2)|2,{119,95,0}},
/* 61716 */ {(13<<2)|2,{119,95,0}},
/* 61717 */ {(13<<2)|2,{119,95,0}},
/* 61718 */ {(13<<2)|2,{119,95,0}},
/* 61719 */ {(13<<2)|2,{119,95,0}},
/* 61720 */ {(13<<2)|2,{119,98,0}},
/* 61721 */ {(13<<2)|2,{119,98,0}},
/* 61722 */ {(13<<2)|2,{119,98,0}},
/* 61723 */ {(13<<2)|2,{119,98,0}},
/* 61724 */ {(13<<2)|2,{119,98,0}},
/* 61725 */ {(13<<2)|2,{119,98,0}},
/* 61726 */ {(13<<2)|2,{119,98,0}},
/* 61727 */ {(13<<2)|2,{119,98,0}},
/* 61728 */ {(13<<2)|2,{119,100,0}},
/* 61729 */ {(13<<2)|2,{119,100,0}},
/* 61730 */ {(13<<2)|2,{119,100,0}},
/* 61731 */ {(13<<2)|2,{119,100,0}},
/* 61732 */ {(13<<2)|2,{119,100,0}},
/* 61733 */ {(13<<2)|2,{119,100,0}},
/* 61734 */ {(13<<2)|2,{119,100,0}},
/* 61735 */ {(13<<2)|2,{119,100,0}},
/* 61736 */ {(13<<2)|2,{119,102,0}},
/* 61737 */ {(13<<2)|2,{119,102,0}},
/* 61738 */ {(13<<2)|2,{119,102,0}},
/* 61739 */ {(13<<2)|2,{119,102,0}},
/* 61740 */ {(13<<2)|2,{119,102,0}},
/* 61741 */ {(13<<2)|2,{119,102,0}},
/* 61742 */ {(13<<2)|2,{119,102,0}},
/* 61743 */ {(13<<2)|2,{119,102,0}},
/* 61744 */ {(13<<2)|2,{119,103,0}},
/* 61745 */ {(13<<2)|2,{119,103,0}},
/* 61746 */ {(13<<2)|2,{119,103,0}},
/* 61747 */ {(13<<2)|2,{119,103,0}},
/* 61748 */ {(13<<2)|2,{119,103,0}},
/* 61749 */ {(13<<2)|2,{119,103,0}},
/* 61750 */ {(13<<2)|2,{119,103,0}},
/* 61751 */ {(13<<2)|2,{119,103,0}},
/* 61752 */ {(13<<2)|2,{119,104,0}},
/* 61753 */ {(13<<2)|2,{119,104,0}},
/* 61754 */ {(13<<2)|2,{119,104,0}},
/* 61755 */ {(13<<2)|2,{119,104,0}},
/* 61756 */ {(13<<2)|2,{119,104,0}},
/* 61757 */ {(13<<2)|2,{119,104,0}},
/* 61758 */ {(13<<2)|2,{119,104,0}},
/* 61759 */ {(13<<2)|2,{119,104,0}},
/* 61760 */ {(13<<2)|2,{119,108,0}},
/* 61761 */ {(13<<2)|2,{119,108,0}},
/* 61762 */ {(13<<2)|2,{119,108,0}},
/* 61763 */ {(13<<2)|2,{119,108,0}},
/* 61764 */ {(13<<2)|2,{119,108,0}},
/* 61765 */ {(13<<2)|2,{119,108,0}},
/* 61766 */ {(13<<2)|2,{119,108,0}},
/* 61767 */ {(13<<2)|2,{119,108,0}},
/* 61768 */ {(13<<2)|2,{119,109,0}},
/* 61769 */ {(13<<2)|2,{119,109,0}},
/* 61770 */ {(13<<2)|2,{119,109,0}},
/* 61771 */ {(13<<2)|2,{119,109,0}},
/* 61772 */ {(13<<2)|2,{119,109,0}},
/* 61773 */ {(13<<2)|2,{119,109,0}},
/* 61774 */ {(13<<2)|2,{119,109,0}},
/* 61775 */ {(13<<2)|2,{119,109,0}},
/* 61776 */ {(13<<2)|2,{119,110,0}},
/* 61777 */ {(13<<2)|2,{119,110,0}},
/* 61778 */ {(13<<2)|2,{119,110,0}},
/* 61779 */ {(13<<2)|2,{119,110,0}},
/* 61780 */ {(13<<2)|2,{119,110,0}},
/* 61781 */ {(13<<2)|2,{119,110,0}},
/* 61782 */ {(13<<2)|2,{119,110,0}},
/* 61783 */ {(13<<2)|2,{119,110,0}},
/* 61784 */ {(13<<2)|2,{119,112,0}},
/* 61785 */ {(13<<2)|2,{119,112,0}},
/* 61786 */ {(13<<2)|2,{119,112,0}},
/* 61787 */ {(13<<2)|2,{119,112,0}},
/* 61788 */ {(13<<2)|2,{119,112,0}},
/* 61789 */ {(13<<2)|2,{119,112,0}},
/* 61790 */ {(13<<2)|2,{119,112,0}},
/* 61791 */ {(13<<2)|2,{119,112,0}},
/* 61792 */ {(13<<2)|2,{119,114,0}},
/* 61793 */ {(13<<2)|2,{119,114,0}},
/* 61794 */ {(13<<2)|2,{119,114,0}},
/* 61795 */ {(13<<2)|2,{119,114,0}},
/* 61796 */ {(13<<2)|2,{119,114,0}},
/* 61797 */ {(13<<2)|2,{119,114,0}},
/* 61798 */ {(13<<2)|2,{119,114,0}},
/* 61799 */ {(13<<2)|2,{119,114,0}},
/* 61800 */ {(13<<2)|2,{119,117,0}},
/* 61801 */ {(13<<2)|2,{119,117,0}},
/* 61802 */ {(13<<2)|2,{119,117,0}},
/* 61803 */ {(13<<2)|2,{119,117,0}},
/* 61804 */ {(13<<2)|2,{119,117,0}},
/* 61805 */ {(13<<2)|2,{119,117,0}},
/* 61806 */ {(13<<2)|2,{119,117,0}},
/* 61807 */ {(13<<2)|2,{119,117,0}},
/* 61808 */ {(14<<2)|2,{119,58,0}},
/* 61809 */ {(14<<2)|2,{119,58,0}},
/* 61810 */ {(14<<2)|2,{119,58,0}},
/* 61811 */ {(14<<2)|2,{119,58,0}},
/* 61812 */ {(14<<2)|2,{119,66,0}},
/* 61813 */ {(14<<2)|2,{119,66,0}},
/* 61814 */ {(14<<2)|2,{119,66,0}},
/* 61815 */ {(14<<2)|2,{119,66,0}},
/* 61816 */ {(14<<2)|2,{119,67,0}},
/* 61817 */ {(14<<2)|2,{119,67,0}},
/* 61818 */ {(14<<2)|2,{119,67,0}},
/* 61819 */ {(14<<2)|2,{119,67,0}},
/* 61820 */ {(14<<2)|2,{119,68,0}},
/* 61821 */ {(14<<2)|2,{119,68,0}},
/* 61822 */ {(14<<2)|2,{119,68,0}},
/* 61823 */ {(14<<2)|2,{119,68,0}},
/* 61824 */ {(14<<2)|2,{119,69,0}},
/* 61825 */ {(14<<2)|2,{119,69,0}},
/* 61826 */ {(14<<2)|2,{119,69,0}},
/* 61827 */ {(14<<2)|2,{119,69,0}},
/* 61828 */ {(14<<2)|2,{119,70,0}},
/* 61829 */ {(14<<2)|2,{119,70,0}},
/* 61830 */ {(14<<2)|2,{119,70,0}},
/* 61831 */ {(14<<2)|2,{119,70,0}},
/* 61832 */ {(14<<2)|2,{119,71,0}},
/* 61833 */ {(14<<2)|2,{119,71,0}},
/* 61834 */ {(14<<2)|2,{119,71,0}},
/* 61835 */ {(14<<2)|2,{119,71,0}},
/* 61836 */ {(14<<2)|2,{119,72,0}},
/* 61837 */ {(14<<2)|2,{119,72,0}},
/* 61838 */ {(14<<2)|2,{119,72,0}},
/* 61839 */ {(14<<2)|2,{119,72,0}},
/* 61840 */ {(14<<2)|2,{119,73,0}},
/* 61841 */ {(14<<2)|2,{119,73,0}},
/* 61842 */ {(14<<2)|2,{119,73,0}},
/* 61843 */ {(14<<2)|2,{119,73,0}},
/* 61844 */ {(14<<2)|2,{119,74,0}},
/* 61845 */ {(14<<2)|2,{119,74,0}},
/* 61846 */ {(14<<2)|2,{119,74,0}},
/* 61847 */ {(14<<2)|2,{119,74,0}},
/* 61848 */ {(14<<2)|2,{119,75,0}},
/* 61849 */ {(14<<2)|2,{119,75,0}},
/* 61850 */ {(14<<2)|2,{119,75,0}},
/* 61851 */ {(14<<2)|2,{119,75,0}},
/* 61852 */ {(14<<2)|2,{119,76,0}},
/* 61853 */ {(14<<2)|2,{119,76,0}},
/* 61854 */ {(14<<2)|2,{119,76,0}},
/* 61855 */ {(14<<2)|2,{119,76,0}},
/* 61856 */ {(14<<2)|2,{119,77,0}},
/* 61857 */ {(14<<2)|2,{119,77,0}},
/* 61858 */ {(14<<2)|2,{119,77,0}},
/* 61859 */ {(14<<2)|2,{119,77,0}},
/* 61860 */ {(14<<2)|2,{119,78,0}},
/* 61861 */ {(14<<2)|2,{119,78,0}},
/* 61862 */ {(14<<2)|2,{119,78,0}},
/* 61863 */ {(14<<2)|2,{119,78,0}},
/* 61864 */ {(14<<2)|2,{119,79,0}},
/* 61865 */ {(14<<2)|2,{119,79,0}},
/* 61866 */ {(14<<2)|2,{119,79,0}},
/* 61867 */ {(14<<2)|2,{119,79,0}},
/* 61868 */ {(14<<2)|2,{119,80,0}},
/* 61869 */ {(14<<2)|2,{119,80,0}},
/* 61870 */ {(14<<2)|2,{119,80,0}},
/* 61871 */ {(14<<2)|2,{119,80,0}},
/* 61872 */ {(14<<2)|2,{119,81,0}},
/* 61873 */ {(14<<2)|2,{119,81,0}},
/* 61874 */ {(14<<2)|2,{119,81,0}},
/* 61875 */ {(14<<2)|2,{119,81,0}},
/* 61876 */ {(14<<2)|2,{119,82,0}},
/* 61877 */ {(14<<2)|2,{119,82,0}},
/* 61878 */ {(14<<2)|2,{119,82,0}},
/* 61879 */ {(14<<2)|2,{119,82,0}},
/* 61880 */ {(14<<2)|2,{119,83,0}},
/* 61881 */ {(14<<2)|2,{119,83,0}},
/* 61882 */ {(14<<2)|2,{119,83,0}},
/* 61883 */ {(14<<2)|2,{119,83,0}},
/* 61884 */ {(14<<2)|2,{119,84,0}},
/* 61885 */ {(14<<2)|2,{119,84,0}},
/* 61886 */ {(14<<2)|2,{119,84,0}},
/* 61887 */ {(14<<2)|2,{119,84,0}},
/* 61888 */ {(14<<2)|2,{119,85,0}},
/* 61889 */ {(14<<2)|2,{119,85,0}},
/* 61890 */ {(14<<2)|2,{119,85,0}},
/* 61891 */ {(14<<2)|2,{119,85,0}},
/* 61892 */ {(14<<2)|2,{119,86,0}},
/* 61893 */ {(14<<2)|2,{119,86,0}},
/* 61894 */ {(14<<2)|2,{119,86,0}},
/* 61895 */ {(14<<2)|2,{119,86,0}},
/* 61896 */ {(14<<2)|2,{119,87,0}},
/* 61897 */ {(14<<2)|2,{119,87,0}},
/* 61898 */ {(14<<2)|2,{119,87,0}},
/* 61899 */ {(14<<2)|2,{119,87,0}},
/* 61900 */ {(14<<2)|2,{119,89,0}},
/* 61901 */ {(14<<2)|2,{119,89,0}},
/* 61902 */ {(14<<2)|2,{119,89,0}},
/* 61903 */ {(14<<2)|2,{119,89,0}},
/* 61904 */ {(14<<2)|2,{119,106,0}},
/* 61905 */ {(14<<2)|2,{119,106,0}},
/* 61906 */ {(14<<2)|2,{119,106,0}},
/* 61907 */ {(14<<2)|2,{119,106,0}},
/* 61908 */ {(14<<2)|2,{119,107,0}},
/* 61909 */ {(14<<2)|2,{119,107,0}},
/* 61910 */ {(14<<2)|2,{119,107,0}},
/* 61911 */ {(14<<2)|2,{119,107,0}},
/* 61912 */ {(14<<2)|2,{119,113,0}},
/* 61913 */ {(14<<2)|2,{119,113,0}},
/* 61914 */ {(14<<2)|2,{119,113,0}},
/* 61915 */ {(14<<2)|2,{119,113,0}},
/* 61916 */ {(14<<2)|2,{119,118,0}},
/* 61917 */ {(14<<2)|2,{119,118,0}},
/* 61918 */ {(14<<2)|2,{119,118,0}},
/* 61919 */ {(14<<2)|2,{119,118,0}},
/* 61920 */ {(14<<2)|2,{119,119,0}},
/* 61921 */ {(14<<2)|2,{119,119,0}},
/* 61922 */ {(14<<2)|2,{119,119,0}},
/* 61923 */ {(14<<2)|2,{119,119,0}},
/* 61924 */ {(14<<2)|2,{119,120,0}},
/* 61925 */ {(14<<2)|2,{119,120,0}},
/* 61926 */ {(14<<2)|2,{119,120,0}},
/* 61927 */ {(14<<2)|2,{119,120,0}},
/* 61928 */ {(14<<2)|2,{119,121,0}},
/* 61929 */ {(14<<2)|2,{119,121,0}},
/* 61930 */ {(14<<2)|2,{119,121,0}},
/* 61931 */ {(14<<2)|2,{119,121,0}},
/* 61932 */ {(14<<2)|2,{119,122,0}},
/* 61933 */ {(14<<2)|2,{119,122,0}},
/* 61934 */ {(14<<2)|2,{119,122,0}},
/* 61935 */ {(14<<2)|2,{119,122,0}},
/* 61936 */ {(15<<2)|2,{119,38,0}},
/* 61937 */ {(15<<2)|2,{119,38,0}},
/* 61938 */ {(15<<2)|2,{119,42,0}},
/* 61939 */ {(15<<2)|2,{119,42,0}},
/* 61940 */ {(15<<2)|2,{119,44,0}},
/* 61941 */ {(15<<2)|2,{119,44,0}},
/* 61942 */ {(15<<2)|2,{119,59,0}},
/* 61943 */ {(15<<2)|2,{119,59,0}},
/* 61944 */ {(15<<2)|2,{119,88,0}},
/* 61945 */ {(15<<2)|2,{119,88,0}},
/* 61946 */ {(15<<2)|2,{119,90,0}},
/* 61947 */ {(15<<2)|2,{119,90,0}},
/* 61948 */ {(7<<2)|1,{119,0,0}},
/* 61949 */ {(7<<2)|1,{119,0,0}},
/* 61950 */ {(7<<2)|1,{119,0,0}},
/* 61951 */ {(7<<2)|1,{119,0,0}},
/* 61952 */ {(12<<2)|2,{120,48,0}},
/* 61953 */ {(12<<2)|2,{120,48,0}},
/* 61954 */ {(12<<2)|2,{120,48,0}},
/* 61955 */ {(12<<2)|2,{120,48,0}},
/* 61956 */ {(12<<2)|2,{120,48,0}},
/* 61957 */ {(12<<2)|2,{120,48,0}},
/* 61958 */ {(12<<2)|2,{120,48,0}},
/* 61959 */ {(12<<2)|2,{120,48,0}},
/* 61960 */ {(12<<2)|2,{120,48,0}},
/* 61961 */ {(12<<2)|2,{120,48,0}},
/* 61962 */ {(12<<2)|2,{120,48,0}},
/* 61963 */ {(12<<2)|2,{120,48,0}},
/* 61964 */ {(12<<2)|2,{120,48,0}},
/* 61965 */ {(12<<2)|2,{120,48,0}},
/* 61966 */ {(12<<2)|2,{120,48,0}},
/* 61967 */ {(12<<2)|2,{120,48,0}},
/* 61968 */ {(12<<2)|2,{120,49,0}},
/* 61969 */ {(12<<2)|2,{120,49,0}},
/* 61970 */ {(12<<2)|2,{120,49,0}},
/* 61971 */ {(12<<2)|2,{120,49,0}},
/* 61972 */ {(12<<2)|2,{120,49,0}},
/* 61973 */ {(12<<2)|2,{120,49,0}},
/* 61974 */ {(12<<2)|2,{120,49,0}},
/* 61975 */ {(12<<2)|2,{120,49,0}},
/* 61976 */ {(12<<2)|2,{120,49,0}},
/* 61977 */ {(12<<2)|2,{120,49,0}},
/* 61978 */ {(12<<2)|2,{120,49,0}},
/* 61979 */ {(12<<2)|2,{120,49,0}},
/* 61980 */ {(12<<2)|2,{120,49,0}},
/* 61981 */ {(12<<2)|2,{120,49,0}},
/* 61982 */ {(12<<2)|2,{120,49,0}},
/* 61983 */ {(12<<2)|2,{120,49,0}},
/* 61984 */ {(12<<2)|2,{120,50,0}},
/* 61985 */ {(12<<2)|2,{120,50,0}},
/* 61986 */ {(12<<2)|2,{120,50,0}},
/* 61987 */ {(12<<2)|2,{120,50,0}},
/* 61988 */ {(12<<2)|2,{120,50,0}},
/* 61989 */ {(12<<2)|2,{120,50,0}},
/* 61990 */ {(12<<2)|2,{120,50,0}},
/* 61991 */ {(12<<2)|2,{120,50,0}},
/* 61992 */ {(12<<2)|2,{120,50,0}},
/* 61993 */ {(12<<2)|2,{120,50,0}},
/* 61994 */ {(12<<2)|2,{120,50,0}},
/* 61995 */ {(12<<2)|2,{120,50,0}},
/* 61996 */ {(12<<2)|2,{120,50,0}},
/* 61997 */ {(12<<2)|2,{120,50,0}},
/* 61998 */ {(12<<2)|2,{120,50,0}},
/* 61999 */ {(12<<2)|2,{120,50,0}},
/* 62000 */ {(12<<2)|2,{120,97,0}},
/* 62001 */ {(12<<2)|2,{120,97,0}},
/* 62002 */ {(12<<2)|2,{120,97,0}},
/* 62003 */ {(12<<2)|2,{120,97,0}},
/* 62004 */ {(12<<2)|2,{120,97,0}},
/* 62005 */ {(12<<2)|2,{120,97,0}},
/* 62006 */ {(12<<2)|2,{120,97,0}},
/* 62007 */ {(12<<2)|2,{120,97,0}},
/* 62008 */ {(12<<2)|2,{120,97,0}},
/* 62009 */ {(12<<2)|2,{120,97,0}},
/* 62010 */ {(12<<2)|2,{120,97,0}},
/* 62011 */ {(12<<2)|2,{120,97,0}},
/* 62012 */ {(12<<2)|2,{120,97,0}},
/* 62013 */ {(12<<2)|2,{120,97,0}},
/* 62014 */ {(12<<2)|2,{120,97,0}},
/* 62015 */ {(12<<2)|2,{120,97,0}},
/* 62016 */ {(12<<2)|2,{120,99,0}},
/* 62017 */ {(12<<2)|2,{120,99,0}},
/* 62018 */ {(12<<2)|2,{120,99,0}},
/* 62019 */ {(12<<2)|2,{120,99,0}},
/* 62020 */ {(12<<2)|2,{120,99,0}},
/* 62021 */ {(12<<2)|2,{120,99,0}},
/* 62022 */ {(12<<2)|2,{120,99,0}},
/* 62023 */ {(12<<2)|2,{120,99,0}},
/* 62024 */ {(12<<2)|2,{120,99,0}},
/* 62025 */ {(12<<2)|2,{120,99,0}},
/* 62026 */ {(12<<2)|2,{120,99,0}},
/* 62027 */ {(12<<2)|2,{120,99,0}},
/* 62028 */ {(12<<2)|2,{120,99,0}},
/* 62029 */ {(12<<2)|2,{120,99,0}},
/* 62030 */ {(12<<2)|2,{120,99,0}},
/* 62031 */ {(12<<2)|2,{120,99,0}},
/* 62032 */ {(12<<2)|2,{120,101,0}},
/* 62033 */ {(12<<2)|2,{120,101,0}},
/* 62034 */ {(12<<2)|2,{120,101,0}},
/* 62035 */ {(12<<2)|2,{120,101,0}},
/* 62036 */ {(12<<2)|2,{120,101,0}},
/* 62037 */ {(12<<2)|2,{120,101,0}},
/* 62038 */ {(12<<2)|2,{120,101,0}},
/* 62039 */ {(12<<2)|2,{120,101,0}},
/* 62040 */ {(12<<2)|2,{120,101,0}},
/* 62041 */ {(12<<2)|2,{120,101,0}},
/* 62042 */ {(12<<2)|2,{120,101,0}},
/* 62043 */ {(12<<2)|2,{120,101,0}},
/* 62044 */ {(12<<2)|2,{120,101,0}},
/* 62045 */ {(12<<2)|2,{120,101,0}},
/* 62046 */ {(12<<2)|2,{120,101,0}},
/* 62047 */ {(12<<2)|2,{120,101,0}},
/* 62048 */ {(12<<2)|2,{120,105,0}},
/* 62049 */ {(12<<2)|2,{120,105,0}},
/* 62050 */ {(12<<2)|2,{120,105,0}},
/* 62051 */ {(12<<2)|2,{120,105,0}},
/* 62052 */ {(12<<2)|2,{120,105,0}},
/* 62053 */ {(12<<2)|2,{120,105,0}},
/* 62054 */ {(12<<2)|2,{120,105,0}},
/* 62055 */ {(12<<2)|2,{120,105,0}},
/* 62056 */ {(12<<2)|2,{120,105,0}},
/* 62057 */ {(12<<2)|2,{120,105,0}},
/* 62058 */ {(12<<2)|2,{120,105,0}},
/* 62059 */ {(12<<2)|2,{120,105,0}},
/* 62060 */ {(12<<2)|2,{120,105,0}},
/* 62061 */ {(12<<2)|2,{120,105,0}},
/* 62062 */ {(12<<2)|2,{120,105,0}},
/* 62063 */ {(12<<2)|2,{120,105,0}},
/* 62064 */ {(12<<2)|2,{120,111,0}},
/* 62065 */ {(12<<2)|2,{120,111,0}},
/* 62066 */ {(12<<2)|2,{120,111,0}},
/* 62067 */ {(12<<2)|2,{120,111,0}},
/* 62068 */ {(12<<2)|2,{120,111,0}},
/* 62069 */ {(12<<2)|2,{120,111,0}},
/* 62070 */ {(12<<2)|2,{120,111,0}},
/* 62071 */ {(12<<2)|2,{120,111,0}},
/* 62072 */ {(12<<2)|2,{120,111,0}},
/* 62073 */ {(12<<2)|2,{120,111,0}},
/* 62074 */ {(12<<2)|2,{120,111,0}},
/* 62075 */ {(12<<2)|2,{120,111,0}},
/* 62076 */ {(12<<2)|2,{120,111,0}},
/* 62077 */ {(12<<2)|2,{120,111,0}},
/* 62078 */ {(12<<2)|2,{120,111,0}},
/* 62079 */ {(12<<2)|2,{120,111,0}},
/* 62080 */ {(12<<2)|2,{120,115,0}},
/* 62081 */ {(12<<2)|2,{120,115,0}},
/* 62082 */ {(12<<2)|2,{120,115,0}},
/* 62083 */ {(12<<2)|2,{120,115,0}},
/* 62084 */ {(12<<2)|2,{120,115,0}},
/* 62085 */ {(12<<2)|2,{120,115,0}},
/* 62086 */ {(12<<2)|2,{120,115,0}},
/* 62087 */ {(12<<2)|2,{120,115,0}},
/* 62088 */ {(12<<2)|2,{120,115,0}},
/* 62089 */ {(12<<2)|2,{120,115,0}},
/* 62090 */ {(12<<2)|2,{120,115,0}},
/* 62091 */ {(12<<2)|2,{120,115,0}},
/* 62092 */ {(12<<2)|2,{120,115,0}},
/* 62093 */ {(12<<2)|2,{120,115,0}},
/* 62094 */ {(12<<2)|2,{120,115,0}},
/* 62095 */ {(12<<2)|2,{120,115,0}},
/* 62096 */ {(12<<2)|2,{120,116,0}},
/* 62097 */ {(12<<2)|2,{120,116,0}},
/* 62098 */ {(12<<2)|2,{120,116,0}},
/* 62099 */ {(12<<2)|2,{120,116,0}},
/* 62100 */ {(12<<2)|2,{120,116,0}},
/* 62101 */ {(12<<2)|2,{120,116,0}},
/* 62102 */ {(12<<2)|2,{120,116,0}},
/* 62103 */ {(12<<2)|2,{120,116,0}},
/* 62104 */ {(12<<2)|2,{120,116,0}},
/* 62105 */ {(12<<2)|2,{120,116,0}},
/* 62106 */ {(12<<2)|2,{120,116,0}},
/* 62107 */ {(12<<2)|2,{120,116,0}},
/* 62108 */ {(12<<2)|2,{120,116,0}},
/* 62109 */ {(12<<2)|2,{120,116,0}},
/* 62110 */ {(12<<2)|2,{120,116,0}},
/* 62111 */ {(12<<2)|2,{120,116,0}},
/* 62112 */ {(13<<2)|2,{120,32,0}},
/* 62113 */ {(13<<2)|2,{120,32,0}},
/* 62114 */ {(13<<2)|2,{120,32,0}},
/* 62115 */ {(13<<2)|2,{120,32,0}},
/* 62116 */ {(13<<2)|2,{120,32,0}},
/* 62117 */ {(13<<2)|2,{120,32,0}},
/* 62118 */ {(13<<2)|2,{120,32,0}},
/* 62119 */ {(13<<2)|2,{120,32,0}},
/* 62120 */ {(13<<2)|2,{120,37,0}},
/* 62121 */ {(13<<2)|2,{120,37,0}},
/* 62122 */ {(13<<2)|2,{120,37,0}},
/* 62123 */ {(13<<2)|2,{120,37,0}},
/* 62124 */ {(13<<2)|2,{120,37,0}},
/* 62125 */ {(13<<2)|2,{120,37,0}},
/* 62126 */ {(13<<2)|2,{120,37,0}},
/* 62127 */ {(13<<2)|2,{120,37,0}},
/* 62128 */ {(13<<2)|2,{120,45,0}},
/* 62129 */ {(13<<2)|2,{120,45,0}},
/* 62130 */ {(13<<2)|2,{120,45,0}},
/* 62131 */ {(13<<2)|2,{120,45,0}},
/* 62132 */ {(13<<2)|2,{120,45,0}},
/* 62133 */ {(13<<2)|2,{120,45,0}},
/* 62134 */ {(13<<2)|2,{120,45,0}},
/* 62135 */ {(13<<2)|2,{120,45,0}},
/* 62136 */ {(13<<2)|2,{120,46,0}},
/* 62137 */ {(13<<2)|2,{120,46,0}},
/* 62138 */ {(13<<2)|2,{120,46,0}},
/* 62139 */ {(13<<2)|2,{120,46,0}},
/* 62140 */ {(13<<2)|2,{120,46,0}},
/* 62141 */ {(13<<2)|2,{120,46,0}},
/* 62142 */ {(13<<2)|2,{120,46,0}},
/* 62143 */ {(13<<2)|2,{120,46,0}},
/* 62144 */ {(13<<2)|2,{120,47,0}},
/* 62145 */ {(13<<2)|2,{120,47,0}},
/* 62146 */ {(13<<2)|2,{120,47,0}},
/* 62147 */ {(13<<2)|2,{120,47,0}},
/* 62148 */ {(13<<2)|2,{120,47,0}},
/* 62149 */ {(13<<2)|2,{120,47,0}},
/* 62150 */ {(13<<2)|2,{120,47,0}},
/* 62151 */ {(13<<2)|2,{120,47,0}},
/* 62152 */ {(13<<2)|2,{120,51,0}},
/* 62153 */ {(13<<2)|2,{120,51,0}},
/* 62154 */ {(13<<2)|2,{120,51,0}},
/* 62155 */ {(13<<2)|2,{120,51,0}},
/* 62156 */ {(13<<2)|2,{120,51,0}},
/* 62157 */ {(13<<2)|2,{120,51,0}},
/* 62158 */ {(13<<2)|2,{120,51,0}},
/* 62159 */ {(13<<2)|2,{120,51,0}},
/* 62160 */ {(13<<2)|2,{120,52,0}},
/* 62161 */ {(13<<2)|2,{120,52,0}},
/* 62162 */ {(13<<2)|2,{120,52,0}},
/* 62163 */ {(13<<2)|2,{120,52,0}},
/* 62164 */ {(13<<2)|2,{120,52,0}},
/* 62165 */ {(13<<2)|2,{120,52,0}},
/* 62166 */ {(13<<2)|2,{120,52,0}},
/* 62167 */ {(13<<2)|2,{120,52,0}},
/* 62168 */ {(13<<2)|2,{120,53,0}},
/* 62169 */ {(13<<2)|2,{120,53,0}},
/* 62170 */ {(13<<2)|2,{120,53,0}},
/* 62171 */ {(13<<2)|2,{120,53,0}},
/* 62172 */ {(13<<2)|2,{120,53,0}},
/* 62173 */ {(13<<2)|2,{120,53,0}},
/* 62174 */ {(13<<2)|2,{120,53,0}},
/* 62175 */ {(13<<2)|2,{120,53,0}},
/* 62176 */ {(13<<2)|2,{120,54,0}},
/* 62177 */ {(13<<2)|2,{120,54,0}},
/* 62178 */ {(13<<2)|2,{120,54,0}},
/* 62179 */ {(13<<2)|2,{120,54,0}},
/* 62180 */ {(13<<2)|2,{120,54,0}},
/* 62181 */ {(13<<2)|2,{120,54,0}},
/* 62182 */ {(13<<2)|2,{120,54,0}},
/* 62183 */ {(13<<2)|2,{120,54,0}},
/* 62184 */ {(13<<2)|2,{120,55,0}},
/* 62185 */ {(13<<2)|2,{120,55,0}},
/* 62186 */ {(13<<2)|2,{120,55,0}},
/* 62187 */ {(13<<2)|2,{120,55,0}},
/* 62188 */ {(13<<2)|2,{120,55,0}},
/* 62189 */ {(13<<2)|2,{120,55,0}},
/* 62190 */ {(13<<2)|2,{120,55,0}},
/* 62191 */ {(13<<2)|2,{120,55,0}},
/* 62192 */ {(13<<2)|2,{120,56,0}},
/* 62193 */ {(13<<2)|2,{120,56,0}},
/* 62194 */ {(13<<2)|2,{120,56,0}},
/* 62195 */ {(13<<2)|2,{120,56,0}},
/* 62196 */ {(13<<2)|2,{120,56,0}},
/* 62197 */ {(13<<2)|2,{120,56,0}},
/* 62198 */ {(13<<2)|2,{120,56,0}},
/* 62199 */ {(13<<2)|2,{120,56,0}},
/* 62200 */ {(13<<2)|2,{120,57,0}},
/* 62201 */ {(13<<2)|2,{120,57,0}},
/* 62202 */ {(13<<2)|2,{120,57,0}},
/* 62203 */ {(13<<2)|2,{120,57,0}},
/* 62204 */ {(13<<2)|2,{120,57,0}},
/* 62205 */ {(13<<2)|2,{120,57,0}},
/* 62206 */ {(13<<2)|2,{120,57,0}},
/* 62207 */ {(13<<2)|2,{120,57,0}},
/* 62208 */ {(13<<2)|2,{120,61,0}},
/* 62209 */ {(13<<2)|2,{120,61,0}},
/* 62210 */ {(13<<2)|2,{120,61,0}},
/* 62211 */ {(13<<2)|2,{120,61,0}},
/* 62212 */ {(13<<2)|2,{120,61,0}},
/* 62213 */ {(13<<2)|2,{120,61,0}},
/* 62214 */ {(13<<2)|2,{120,61,0}},
/* 62215 */ {(13<<2)|2,{120,61,0}},
/* 62216 */ {(13<<2)|2,{120,65,0}},
/* 62217 */ {(13<<2)|2,{120,65,0}},
/* 62218 */ {(13<<2)|2,{120,65,0}},
/* 62219 */ {(13<<2)|2,{120,65,0}},
/* 62220 */ {(13<<2)|2,{120,65,0}},
/* 62221 */ {(13<<2)|2,{120,65,0}},
/* 62222 */ {(13<<2)|2,{120,65,0}},
/* 62223 */ {(13<<2)|2,{120,65,0}},
/* 62224 */ {(13<<2)|2,{120,95,0}},
/* 62225 */ {(13<<2)|2,{120,95,0}},
/* 62226 */ {(13<<2)|2,{120,95,0}},
/* 62227 */ {(13<<2)|2,{120,95,0}},
/* 62228 */ {(13<<2)|2,{120,95,0}},
/* 62229 */ {(13<<2)|2,{120,95,0}},
/* 62230 */ {(13<<2)|2,{120,95,0}},
/* 62231 */ {(13<<2)|2,{120,95,0}},
/* 62232 */ {(13<<2)|2,{120,98,0}},
/* 62233 */ {(13<<2)|2,{120,98,0}},
/* 62234 */ {(13<<2)|2,{120,98,0}},
/* 62235 */ {(13<<2)|2,{120,98,0}},
/* 62236 */ {(13<<2)|2,{120,98,0}},
/* 62237 */ {(13<<2)|2,{120,98,0}},
/* 62238 */ {(13<<2)|2,{120,98,0}},
/* 62239 */ {(13<<2)|2,{120,98,0}},
/* 62240 */ {(13<<2)|2,{120,100,0}},
/* 62241 */ {(13<<2)|2,{120,100,0}},
/* 62242 */ {(13<<2)|2,{120,100,0}},
/* 62243 */ {(13<<2)|2,{120,100,0}},
/* 62244 */ {(13<<2)|2,{120,100,0}},
/* 62245 */ {(13<<2)|2,{120,100,0}},
/* 62246 */ {(13<<2)|2,{120,100,0}},
/* 62247 */ {(13<<2)|2,{120,100,0}},
/* 62248 */ {(13<<2)|2,{120,102,0}},
/* 62249 */ {(13<<2)|2,{120,102,0}},
/* 62250 */ {(13<<2)|2,{120,102,0}},
/* 62251 */ {(13<<2)|2,{120,102,0}},
/* 62252 */ {(13<<2)|2,{120,102,0}},
/* 62253 */ {(13<<2)|2,{120,102,0}},
/* 62254 */ {(13<<2)|2,{120,102,0}},
/* 62255 */ {(13<<2)|2,{120,102,0}},
/* 62256 */ {(13<<2)|2,{120,103,0}},
/* 62257 */ {(13<<2)|2,{120,103,0}},
/* 62258 */ {(13<<2)|2,{120,103,0}},
/* 62259 */ {(13<<2)|2,{120,103,0}},
/* 62260 */ {(13<<2)|2,{120,103,0}},
/* 62261 */ {(13<<2)|2,{120,103,0}},
/* 62262 */ {(13<<2)|2,{120,103,0}},
/* 62263 */ {(13<<2)|2,{120,103,0}},
/* 62264 */ {(13<<2)|2,{120,104,0}},
/* 62265 */ {(13<<2)|2,{120,104,0}},
/* 62266 */ {(13<<2)|2,{120,104,0}},
/* 62267 */ {(13<<2)|2,{120,104,0}},
/* 62268 */ {(13<<2)|2,{120,104,0}},
/* 62269 */ {(13<<2)|2,{120,104,0}},
/* 62270 */ {(13<<2)|2,{120,104,0}},
/* 62271 */ {(13<<2)|2,{120,104,0}},
/* 62272 */ {(13<<2)|2,{120,108,0}},
/* 62273 */ {(13<<2)|2,{120,108,0}},
/* 62274 */ {(13<<2)|2,{120,108,0}},
/* 62275 */ {(13<<2)|2,{120,108,0}},
/* 62276 */ {(13<<2)|2,{120,108,0}},
/* 62277 */ {(13<<2)|2,{120,108,0}},
/* 62278 */ {(13<<2)|2,{120,108,0}},
/* 62279 */ {(13<<2)|2,{120,108,0}},
/* 62280 */ {(13<<2)|2,{120,109,0}},
/* 62281 */ {(13<<2)|2,{120,109,0}},
/* 62282 */ {(13<<2)|2,{120,109,0}},
/* 62283 */ {(13<<2)|2,{120,109,0}},
/* 62284 */ {(13<<2)|2,{120,109,0}},
/* 62285 */ {(13<<2)|2,{120,109,0}},
/* 62286 */ {(13<<2)|2,{120,109,0}},
/* 62287 */ {(13<<2)|2,{120,109,0}},
/* 62288 */ {(13<<2)|2,{120,110,0}},
/* 62289 */ {(13<<2)|2,{120,110,0}},
/* 62290 */ {(13<<2)|2,{120,110,0}},
/* 62291 */ {(13<<2)|2,{120,110,0}},
/* 62292 */ {(13<<2)|2,{120,110,0}},
/* 62293 */ {(13<<2)|2,{120,110,0}},
/* 62294 */ {(13<<2)|2,{120,110,0}},
/* 62295 */ {(13<<2)|2,{120,110,0}},
/* 62296 */ {(13<<2)|2,{120,112,0}},
/* 62297 */ {(13<<2)|2,{120,112,0}},
/* 62298 */ {(13<<2)|2,{120,112,0}},
/* 62299 */ {(13<<2)|2,{120,112,0}},
/* 62300 */ {(13<<2)|2,{120,112,0}},
/* 62301 */ {(13<<2)|2,{120,112,0}},
/* 62302 */ {(13<<2)|2,{120,112,0}},
/* 62303 */ {(13<<2)|2,{120,112,0}},
/* 62304 */ {(13<<2)|2,{120,114,0}},
/* 62305 */ {(13<<2)|2,{120,114,0}},
/* 62306 */ {(13<<2)|2,{120,114,0}},
/* 62307 */ {(13<<2)|2,{120,114,0}},
/* 62308 */ {(13<<2)|2,{120,114,0}},
/* 62309 */ {(13<<2)|2,{120,114,0}},
/* 62310 */ {(13<<2)|2,{120,114,0}},
/* 62311 */ {(13<<2)|2,{120,114,0}},
/* 62312 */ {(13<<2)|2,{120,117,0}},
/* 62313 */ {(13<<2)|2,{120,117,0}},
/* 62314 */ {(13<<2)|2,{120,117,0}},
/* 62315 */ {(13<<2)|2,{120,117,0}},
/* 62316 */ {(13<<2)|2,{120,117,0}},
/* 62317 */ {(13<<2)|2,{120,117,0}},
/* 62318 */ {(13<<2)|2,{120,117,0}},
/* 62319 */ {(13<<2)|2,{120,117,0}},
/* 62320 */ {(14<<2)|2,{120,58,0}},
/* 62321 */ {(14<<2)|2,{120,58,0}},
/* 62322 */ {(14<<2)|2,{120,58,0}},
/* 62323 */ {(14<<2)|2,{120,58,0}},
/* 62324 */ {(14<<2)|2,{120,66,0}},
/* 62325 */ {(14<<2)|2,{120,66,0}},
/* 62326 */ {(14<<2)|2,{120,66,0}},
/* 62327 */ {(14<<2)|2,{120,66,0}},
/* 62328 */ {(14<<2)|2,{120,67,0}},
/* 62329 */ {(14<<2)|2,{120,67,0}},
/* 62330 */ {(14<<2)|2,{120,67,0}},
/* 62331 */ {(14<<2)|2,{120,67,0}},
/* 62332 */ {(14<<2)|2,{120,68,0}},
/* 62333 */ {(14<<2)|2,{120,68,0}},
/* 62334 */ {(14<<2)|2,{120,68,0}},
/* 62335 */ {(14<<2)|2,{120,68,0}},
/* 62336 */ {(14<<2)|2,{120,69,0}},
/* 62337 */ {(14<<2)|2,{120,69,0}},
/* 62338 */ {(14<<2)|2,{120,69,0}},
/* 62339 */ {(14<<2)|2,{120,69,0}},
/* 62340 */ {(14<<2)|2,{120,70,0}},
/* 62341 */ {(14<<2)|2,{120,70,0}},
/* 62342 */ {(14<<2)|2,{120,70,0}},
/* 62343 */ {(14<<2)|2,{120,70,0}},
/* 62344 */ {(14<<2)|2,{120,71,0}},
/* 62345 */ {(14<<2)|2,{120,71,0}},
/* 62346 */ {(14<<2)|2,{120,71,0}},
/* 62347 */ {(14<<2)|2,{120,71,0}},
/* 62348 */ {(14<<2)|2,{120,72,0}},
/* 62349 */ {(14<<2)|2,{120,72,0}},
/* 62350 */ {(14<<2)|2,{120,72,0}},
/* 62351 */ {(14<<2)|2,{120,72,0}},
/* 62352 */ {(14<<2)|2,{120,73,0}},
/* 62353 */ {(14<<2)|2,{120,73,0}},
/* 62354 */ {(14<<2)|2,{120,73,0}},
/* 62355 */ {(14<<2)|2,{120,73,0}},
/* 62356 */ {(14<<2)|2,{120,74,0}},
/* 62357 */ {(14<<2)|2,{120,74,0}},
/* 62358 */ {(14<<2)|2,{120,74,0}},
/* 62359 */ {(14<<2)|2,{120,74,0}},
/* 62360 */ {(14<<2)|2,{120,75,0}},
/* 62361 */ {(14<<2)|2,{120,75,0}},
/* 62362 */ {(14<<2)|2,{120,75,0}},
/* 62363 */ {(14<<2)|2,{120,75,0}},
/* 62364 */ {(14<<2)|2,{120,76,0}},
/* 62365 */ {(14<<2)|2,{120,76,0}},
/* 62366 */ {(14<<2)|2,{120,76,0}},
/* 62367 */ {(14<<2)|2,{120,76,0}},
/* 62368 */ {(14<<2)|2,{120,77,0}},
/* 62369 */ {(14<<2)|2,{120,77,0}},
/* 62370 */ {(14<<2)|2,{120,77,0}},
/* 62371 */ {(14<<2)|2,{120,77,0}},
/* 62372 */ {(14<<2)|2,{120,78,0}},
/* 62373 */ {(14<<2)|2,{120,78,0}},
/* 62374 */ {(14<<2)|2,{120,78,0}},
/* 62375 */ {(14<<2)|2,{120,78,0}},
/* 62376 */ {(14<<2)|2,{120,79,0}},
/* 62377 */ {(14<<2)|2,{120,79,0}},
/* 62378 */ {(14<<2)|2,{120,79,0}},
/* 62379 */ {(14<<2)|2,{120,79,0}},
/* 62380 */ {(14<<2)|2,{120,80,0}},
/* 62381 */ {(14<<2)|2,{120,80,0}},
/* 62382 */ {(14<<2)|2,{120,80,0}},
/* 62383 */ {(14<<2)|2,{120,80,0}},
/* 62384 */ {(14<<2)|2,{120,81,0}},
/* 62385 */ {(14<<2)|2,{120,81,0}},
/* 62386 */ {(14<<2)|2,{120,81,0}},
/* 62387 */ {(14<<2)|2,{120,81,0}},
/* 62388 */ {(14<<2)|2,{120,82,0}},
/* 62389 */ {(14<<2)|2,{120,82,0}},
/* 62390 */ {(14<<2)|2,{120,82,0}},
/* 62391 */ {(14<<2)|2,{120,82,0}},
/* 62392 */ {(14<<2)|2,{120,83,0}},
/* 62393 */ {(14<<2)|2,{120,83,0}},
/* 62394 */ {(14<<2)|2,{120,83,0}},
/* 62395 */ {(14<<2)|2,{120,83,0}},
/* 62396 */ {(14<<2)|2,{120,84,0}},
/* 62397 */ {(14<<2)|2,{120,84,0}},
/* 62398 */ {(14<<2)|2,{120,84,0}},
/* 62399 */ {(14<<2)|2,{120,84,0}},
/* 62400 */ {(14<<2)|2,{120,85,0}},
/* 62401 */ {(14<<2)|2,{120,85,0}},
/* 62402 */ {(14<<2)|2,{120,85,0}},
/* 62403 */ {(14<<2)|2,{120,85,0}},
/* 62404 */ {(14<<2)|2,{120,86,0}},
/* 62405 */ {(14<<2)|2,{120,86,0}},
/* 62406 */ {(14<<2)|2,{120,86,0}},
/* 62407 */ {(14<<2)|2,{120,86,0}},
/* 62408 */ {(14<<2)|2,{120,87,0}},
/* 62409 */ {(14<<2)|2,{120,87,0}},
/* 62410 */ {(14<<2)|2,{120,87,0}},
/* 62411 */ {(14<<2)|2,{120,87,0}},
/* 62412 */ {(14<<2)|2,{120,89,0}},
/* 62413 */ {(14<<2)|2,{120,89,0}},
/* 62414 */ {(14<<2)|2,{120,89,0}},
/* 62415 */ {(14<<2)|2,{120,89,0}},
/* 62416 */ {(14<<2)|2,{120,106,0}},
/* 62417 */ {(14<<2)|2,{120,106,0}},
/* 62418 */ {(14<<2)|2,{120,106,0}},
/* 62419 */ {(14<<2)|2,{120,106,0}},
/* 62420 */ {(14<<2)|2,{120,107,0}},
/* 62421 */ {(14<<2)|2,{120,107,0}},
/* 62422 */ {(14<<2)|2,{120,107,0}},
/* 62423 */ {(14<<2)|2,{120,107,0}},
/* 62424 */ {(14<<2)|2,{120,113,0}},
/* 62425 */ {(14<<2)|2,{120,113,0}},
/* 62426 */ {(14<<2)|2,{120,113,0}},
/* 62427 */ {(14<<2)|2,{120,113,0}},
/* 62428 */ {(14<<2)|2,{120,118,0}},
/* 62429 */ {(14<<2)|2,{120,118,0}},
/* 62430 */ {(14<<2)|2,{120,118,0}},
/* 62431 */ {(14<<2)|2,{120,118,0}},
/* 62432 */ {(14<<2)|2,{120,119,0}},
/* 62433 */ {(14<<2)|2,{120,119,0}},
/* 62434 */ {(14<<2)|2,{120,119,0}},
/* 62435 */ {(14<<2)|2,{120,119,0}},
/* 62436 */ {(14<<2)|2,{120,120,0}},
/* 62437 */ {(14<<2)|2,{120,120,0}},
/* 62438 */ {(14<<2)|2,{120,120,0}},
/* 62439 */ {(14<<2)|2,{120,120,0}},
/* 62440 */ {(14<<2)|2,{120,121,0}},
/* 62441 */ {(14<<2)|2,{120,121,0}},
/* 62442 */ {(14<<2)|2,{120,121,0}},
/* 62443 */ {(14<<2)|2,{120,121,0}},
/* 62444 */ {(14<<2)|2,{120,122,0}},
/* 62445 */ {(14<<2)|2,{120,122,0}},
/* 62446 */ {(14<<2)|2,{120,122,0}},
/* 62447 */ {(14<<2)|2,{120,122,0}},
/* 62448 */ {(15<<2)|2,{120,38,0}},
/* 62449 */ {(15<<2)|2,{120,38,0}},
/* 62450 */ {(15<<2)|2,{120,42,0}},
/* 62451 */ {(15<<2)|2,{120,42,0}},
/* 62452 */ {(15<<2)|2,{120,44,0}},
/* 62453 */ {(15<<2)|2,{120,44,0}},
/* 62454 */ {(15<<2)|2,{120,59,0}},
/* 62455 */ {(15<<2)|2,{120,59,0}},
/* 62456 */ {(15<<2)|2,{120,88,0}},
/* 62457 */ {(15<<2)|2,{120,88,0}},
/* 62458 */ {(15<<2)|2,{120,90,0}},
/* 62459 */ {(15<<2)|2,{120,90,0}},
/* 62460 */ {(7<<2)|1,{120,0,0}},
/* 62461 */ {(7<<2)|1,{120,0,0}},
/* 62462 */ {(7<<2)|1,{120,0,0}},
/* 62463 */ {(7<<2)|1,{120,0,0}},
/* 62464 */ {(12<<2)|2,{121,48,0}},
/* 62465 */ {(12<<2)|2,{121,48,0}},
/* 62466 */ {(12<<2)|2,{121,48,0}},
/* 62467 */ {(12<<2)|2,{121,48,0}},
/* 62468 */ {(12<<2)|2,{121,48,0}},
/* 62469 */ {(12<<2)|2,{121,48,0}},
/* 62470 */ {(12<<2)|2,{121,48,0}},
/* 62471 */ {(12<<2)|2,{121,48,0}},
/* 62472 */ {(12<<2)|2,{121,48,0}},
/* 62473 */ {(12<<2)|2,{121,48,0}},
/* 62474 */ {(12<<2)|2,{121,48,0}},
/* 62475 */ {(12<<2)|2,{121,48,0}},
/* 62476 */ {(12<<2)|2,{121,48,0}},
/* 62477 */ {(12<<2)|2,{121,48,0}},
/* 62478 */ {(12<<2)|2,{121,48,0}},
/* 62479 */ {(12<<2)|2,{121,48,0}},
/* 62480 */ {(12<<2)|2,{121,49,0}},
/* 62481 */ {(12<<2)|2,{121,49,0}},
/* 62482 */ {(12<<2)|2,{121,49,0}},
/* 62483 */ {(12<<2)|2,{121,49,0}},
/* 62484 */ {(12<<2)|2,{121,49,0}},
/* 62485 */ {(12<<2)|2,{121,49,0}},
/* 62486 */ {(12<<2)|2,{121,49,0}},
/* 62487 */ {(12<<2)|2,{121,49,0}},
/* 62488 */ {(12<<2)|2,{121,49,0}},
/* 62489 */ {(12<<2)|2,{121,49,0}},
/* 62490 */ {(12<<2)|2,{121,49,0}},
/* 62491 */ {(12<<2)|2,{121,49,0}},
/* 62492 */ {(12<<2)|2,{121,49,0}},
/* 62493 */ {(12<<2)|2,{121,49,0}},
/* 62494 */ {(12<<2)|2,{121,49,0}},
/* 62495 */ {(12<<2)|2,{121,49,0}},
/* 62496 */ {(12<<2)|2,{121,50,0}},
/* 62497 */ {(12<<2)|2,{121,50,0}},
/* 62498 */ {(12<<2)|2,{121,50,0}},
/* 62499 */ {(12<<2)|2,{121,50,0}},
/* 62500 */ {(12<<2)|2,{121,50,0}},
/* 62501 */ {(12<<2)|2,{121,50,0}},
/* 62502 */ {(12<<2)|2,{121,50,0}},
/* 62503 */ {(12<<2)|2,{121,50,0}},
/* 62504 */ {(12<<2)|2,{121,50,0}},
/* 62505 */ {(12<<2)|2,{121,50,0}},
/* 62506 */ {(12<<2)|2,{121,50,0}},
/* 62507 */ {(12<<2)|2,{121,50,0}},
/* 62508 */ {(12<<2)|2,{121,50,0}},
/* 62509 */ {(12<<2)|2,{121,50,0}},
/* 62510 */ {(12<<2)|2,{121,50,0}},
/* 62511 */ {(12<<2)|2,{121,50,0}},
/* 62512 */ {(12<<2)|2,{121,97,0}},
/* 62513 */ {(12<<2)|2,{121,97,0}},
/* 62514 */ {(12<<2)|2,{121,97,0}},
/* 62515 */ {(12<<2)|2,{121,97,0}},
/* 62516 */ {(12<<2)|2,{121,97,0}},
/* 62517 */ {(12<<2)|2,{121,97,0}},
/* 62518 */ {(12<<2)|2,{121,97,0}},
/* 62519 */ {(12<<2)|2,{121,97,0}},
/* 62520 */ {(12<<2)|2,{121,97,0}},
/* 62521 */ {(12<<2)|2,{121,97,0}},
/* 62522 */ {(12<<2)|2,{121,97,0}},
/* 62523 */ {(12<<2)|2,{121,97,0}},
/* 62524 */ {(12<<2)|2,{121,97,0}},
/* 62525 */ {(12<<2)|2,{121,97,0}},
/* 62526 */ {(12<<2)|2,{121,97,0}},
/* 62527 */ {(12<<2)|2,{121,97,0}},
/* 62528 */ {(12<<2)|2,{121,99,0}},
/* 62529 */ {(12<<2)|2,{121,99,0}},
/* 62530 */ {(12<<2)|2,{121,99,0}},
/* 62531 */ {(12<<2)|2,{121,99,0}},
/* 62532 */ {(12<<2)|2,{121,99,0}},
/* 62533 */ {(12<<2)|2,{121,99,0}},
/* 62534 */ {(12<<2)|2,{121,99,0}},
/* 62535 */ {(12<<2)|2,{121,99,0}},
/* 62536 */ {(12<<2)|2,{121,99,0}},
/* 62537 */ {(12<<2)|2,{121,99,0}},
/* 62538 */ {(12<<2)|2,{121,99,0}},
/* 62539 */ {(12<<2)|2,{121,99,0}},
/* 62540 */ {(12<<2)|2,{121,99,0}},
/* 62541 */ {(12<<2)|2,{121,99,0}},
/* 62542 */ {(12<<2)|2,{121,99,0}},
/* 62543 */ {(12<<2)|2,{121,99,0}},
/* 62544 */ {(12<<2)|2,{121,101,0}},
/* 62545 */ {(12<<2)|2,{121,101,0}},
/* 62546 */ {(12<<2)|2,{121,101,0}},
/* 62547 */ {(12<<2)|2,{121,101,0}},
/* 62548 */ {(12<<2)|2,{121,101,0}},
/* 62549 */ {(12<<2)|2,{121,101,0}},
/* 62550 */ {(12<<2)|2,{121,101,0}},
/* 62551 */ {(12<<2)|2,{121,101,0}},
/* 62552 */ {(12<<2)|2,{121,101,0}},
/* 62553 */ {(12<<2)|2,{121,101,0}},
/* 62554 */ {(12<<2)|2,{121,101,0}},
/* 62555 */ {(12<<2)|2,{121,101,0}},
/* 62556 */ {(12<<2)|2,{121,101,0}},
/* 62557 */ {(12<<2)|2,{121,101,0}},
/* 62558 */ {(12<<2)|2,{121,101,0}},
/* 62559 */ {(12<<2)|2,{121,101,0}},
/* 62560 */ {(12<<2)|2,{121,105,0}},
/* 62561 */ {(12<<2)|2,{121,105,0}},
/* 62562 */ {(12<<2)|2,{121,105,0}},
/* 62563 */ {(12<<2)|2,{121,105,0}},
/* 62564 */ {(12<<2)|2,{121,105,0}},
/* 62565 */ {(12<<2)|2,{121,105,0}},
/* 62566 */ {(12<<2)|2,{121,105,0}},
/* 62567 */ {(12<<2)|2,{121,105,0}},
/* 62568 */ {(12<<2)|2,{121,105,0}},
/* 62569 */ {(12<<2)|2,{121,105,0}},
/* 62570 */ {(12<<2)|2,{121,105,0}},
/* 62571 */ {(12<<2)|2,{121,105,0}},
/* 62572 */ {(12<<2)|2,{121,105,0}},
/* 62573 */ {(12<<2)|2,{121,105,0}},
/* 62574 */ {(12<<2)|2,{121,105,0}},
/* 62575 */ {(12<<2)|2,{121,105,0}},
/* 62576 */ {(12<<2)|2,{121,111,0}},
/* 62577 */ {(12<<2)|2,{121,111,0}},
/* 62578 */ {(12<<2)|2,{121,111,0}},
/* 62579 */ {(12<<2)|2,{121,111,0}},
/* 62580 */ {(12<<2)|2,{121,111,0}},
/* 62581 */ {(12<<2)|2,{121,111,0}},
/* 62582 */ {(12<<2)|2,{121,111,0}},
/* 62583 */ {(12<<2)|2,{121,111,0}},
/* 62584 */ {(12<<2)|2,{121,111,0}},
/* 62585 */ {(12<<2)|2,{121,111,0}},
/* 62586 */ {(12<<2)|2,{121,111,0}},
/* 62587 */ {(12<<2)|2,{121,111,0}},
/* 62588 */ {(12<<2)|2,{121,111,0}},
/* 62589 */ {(12<<2)|2,{121,111,0}},
/* 62590 */ {(12<<2)|2,{121,111,0}},
/* 62591 */ {(12<<2)|2,{121,111,0}},
/* 62592 */ {(12<<2)|2,{121,115,0}},
/* 62593 */ {(12<<2)|2,{121,115,0}},
/* 62594 */ {(12<<2)|2,{121,115,0}},
/* 62595 */ {(12<<2)|2,{121,115,0}},
/* 62596 */ {(12<<2)|2,{121,115,0}},
/* 62597 */ {(12<<2)|2,{121,115,0}},
/* 62598 */ {(12<<2)|2,{121,115,0}},
/* 62599 */ {(12<<2)|2,{121,115,0}},
/* 62600 */ {(12<<2)|2,{121,115,0}},
/* 62601 */ {(12<<2)|2,{121,115,0}},
/* 62602 */ {(12<<2)|2,{121,115,0}},
/* 62603 */ {(12<<2)|2,{121,115,0}},
/* 62604 */ {(12<<2)|2,{121,115,0}},
/* 62605 */ {(12<<2)|2,{121,115,0}},
/* 62606 */ {(12<<2)|2,{121,115,0}},
/* 62607 */ {(12<<2)|2,{121,115,0}},
/* 62608 */ {(12<<2)|2,{121,116,0}},
/* 62609 */ {(12<<2)|2,{121,116,0}},
/* 62610 */ {(12<<2)|2,{121,116,0}},
/* 62611 */ {(12<<2)|2,{121,116,0}},
/* 62612 */ {(12<<2)|2,{121,116,0}},
/* 62613 */ {(12<<2)|2,{121,116,0}},
/* 62614 */ {(12<<2)|2,{121,116,0}},
/* 62615 */ {(12<<2)|2,{121,116,0}},
/* 62616 */ {(12<<2)|2,{121,116,0}},
/* 62617 */ {(12<<2)|2,{121,116,0}},
/* 62618 */ {(12<<2)|2,{121,116,0}},
/* 62619 */ {(12<<2)|2,{121,116,0}},
/* 62620 */ {(12<<2)|2,{121,116,0}},
/* 62621 */ {(12<<2)|2,{121,116,0}},
/* 62622 */ {(12<<2)|2,{121,116,0}},
/* 62623 */ {(12<<2)|2,{121,116,0}},
/* 62624 */ {(13<<2)|2,{121,32,0}},
/* 62625 */ {(13<<2)|2,{121,32,0}},
/* 62626 */ {(13<<2)|2,{121,32,0}},
/* 62627 */ {(13<<2)|2,{121,32,0}},
/* 62628 */ {(13<<2)|2,{121,32,0}},
/* 62629 */ {(13<<2)|2,{121,32,0}},
/* 62630 */ {(13<<2)|2,{121,32,0}},
/* 62631 */ {(13<<2)|2,{121,32,0}},
/* 62632 */ {(13<<2)|2,{121,37,0}},
/* 62633 */ {(13<<2)|2,{121,37,0}},
/* 62634 */ {(13<<2)|2,{121,37,0}},
/* 62635 */ {(13<<2)|2,{121,37,0}},
/* 62636 */ {(13<<2)|2,{121,37,0}},
/* 62637 */ {(13<<2)|2,{121,37,0}},
/* 62638 */ {(13<<2)|2,{121,37,0}},
/* 62639 */ {(13<<2)|2,{121,37,0}},
/* 62640 */ {(13<<2)|2,{121,45,0}},
/* 62641 */ {(13<<2)|2,{121,45,0}},
/* 62642 */ {(13<<2)|2,{121,45,0}},
/* 62643 */ {(13<<2)|2,{121,45,0}},
/* 62644 */ {(13<<2)|2,{121,45,0}},
/* 62645 */ {(13<<2)|2,{121,45,0}},
/* 62646 */ {(13<<2)|2,{121,45,0}},
/* 62647 */ {(13<<2)|2,{121,45,0}},
/* 62648 */ {(13<<2)|2,{121,46,0}},
/* 62649 */ {(13<<2)|2,{121,46,0}},
/* 62650 */ {(13<<2)|2,{121,46,0}},
/* 62651 */ {(13<<2)|2,{121,46,0}},
/* 62652 */ {(13<<2)|2,{121,46,0}},
/* 62653 */ {(13<<2)|2,{121,46,0}},
/* 62654 */ {(13<<2)|2,{121,46,0}},
/* 62655 */ {(13<<2)|2,{121,46,0}},
/* 62656 */ {(13<<2)|2,{121,47,0}},
/* 62657 */ {(13<<2)|2,{121,47,0}},
/* 62658 */ {(13<<2)|2,{121,47,0}},
/* 62659 */ {(13<<2)|2,{121,47,0}},
/* 62660 */ {(13<<2)|2,{121,47,0}},
/* 62661 */ {(13<<2)|2,{121,47,0}},
/* 62662 */ {(13<<2)|2,{121,47,0}},
/* 62663 */ {(13<<2)|2,{121,47,0}},
/* 62664 */ {(13<<2)|2,{121,51,0}},
/* 62665 */ {(13<<2)|2,{121,51,0}},
/* 62666 */ {(13<<2)|2,{121,51,0}},
/* 62667 */ {(13<<2)|2,{121,51,0}},
/* 62668 */ {(13<<2)|2,{121,51,0}},
/* 62669 */ {(13<<2)|2,{121,51,0}},
/* 62670 */ {(13<<2)|2,{121,51,0}},
/* 62671 */ {(13<<2)|2,{121,51,0}},
/* 62672 */ {(13<<2)|2,{121,52,0}},
/* 62673 */ {(13<<2)|2,{121,52,0}},
/* 62674 */ {(13<<2)|2,{121,52,0}},
/* 62675 */ {(13<<2)|2,{121,52,0}},
/* 62676 */ {(13<<2)|2,{121,52,0}},
/* 62677 */ {(13<<2)|2,{121,52,0}},
/* 62678 */ {(13<<2)|2,{121,52,0}},
/* 62679 */ {(13<<2)|2,{121,52,0}},
/* 62680 */ {(13<<2)|2,{121,53,0}},
/* 62681 */ {(13<<2)|2,{121,53,0}},
/* 62682 */ {(13<<2)|2,{121,53,0}},
/* 62683 */ {(13<<2)|2,{121,53,0}},
/* 62684 */ {(13<<2)|2,{121,53,0}},
/* 62685 */ {(13<<2)|2,{121,53,0}},
/* 62686 */ {(13<<2)|2,{121,53,0}},
/* 62687 */ {(13<<2)|2,{121,53,0}},
/* 62688 */ {(13<<2)|2,{121,54,0}},
/* 62689 */ {(13<<2)|2,{121,54,0}},
/* 62690 */ {(13<<2)|2,{121,54,0}},
/* 62691 */ {(13<<2)|2,{121,54,0}},
/* 62692 */ {(13<<2)|2,{121,54,0}},
/* 62693 */ {(13<<2)|2,{121,54,0}},
/* 62694 */ {(13<<2)|2,{121,54,0}},
/* 62695 */ {(13<<2)|2,{121,54,0}},
/* 62696 */ {(13<<2)|2,{121,55,0}},
/* 62697 */ {(13<<2)|2,{121,55,0}},
/* 62698 */ {(13<<2)|2,{121,55,0}},
/* 62699 */ {(13<<2)|2,{121,55,0}},
/* 62700 */ {(13<<2)|2,{121,55,0}},
/* 62701 */ {(13<<2)|2,{121,55,0}},
/* 62702 */ {(13<<2)|2,{121,55,0}},
/* 62703 */ {(13<<2)|2,{121,55,0}},
/* 62704 */ {(13<<2)|2,{121,56,0}},
/* 62705 */ {(13<<2)|2,{121,56,0}},
/* 62706 */ {(13<<2)|2,{121,56,0}},
/* 62707 */ {(13<<2)|2,{121,56,0}},
/* 62708 */ {(13<<2)|2,{121,56,0}},
/* 62709 */ {(13<<2)|2,{121,56,0}},
/* 62710 */ {(13<<2)|2,{121,56,0}},
/* 62711 */ {(13<<2)|2,{121,56,0}},
/* 62712 */ {(13<<2)|2,{121,57,0}},
/* 62713 */ {(13<<2)|2,{121,57,0}},
/* 62714 */ {(13<<2)|2,{121,57,0}},
/* 62715 */ {(13<<2)|2,{121,57,0}},
/* 62716 */ {(13<<2)|2,{121,57,0}},
/* 62717 */ {(13<<2)|2,{121,57,0}},
/* 62718 */ {(13<<2)|2,{121,57,0}},
/* 62719 */ {(13<<2)|2,{121,57,0}},
/* 62720 */ {(13<<2)|2,{121,61,0}},
/* 62721 */ {(13<<2)|2,{121,61,0}},
/* 62722 */ {(13<<2)|2,{121,61,0}},
/* 62723 */ {(13<<2)|2,{121,61,0}},
/* 62724 */ {(13<<2)|2,{121,61,0}},
/* 62725 */ {(13<<2)|2,{121,61,0}},
/* 62726 */ {(13<<2)|2,{121,61,0}},
/* 62727 */ {(13<<2)|2,{121,61,0}},
/* 62728 */ {(13<<2)|2,{121,65,0}},
/* 62729 */ {(13<<2)|2,{121,65,0}},
/* 62730 */ {(13<<2)|2,{121,65,0}},
/* 62731 */ {(13<<2)|2,{121,65,0}},
/* 62732 */ {(13<<2)|2,{121,65,0}},
/* 62733 */ {(13<<2)|2,{121,65,0}},
/* 62734 */ {(13<<2)|2,{121,65,0}},
/* 62735 */ {(13<<2)|2,{121,65,0}},
/* 62736 */ {(13<<2)|2,{121,95,0}},
/* 62737 */ {(13<<2)|2,{121,95,0}},
/* 62738 */ {(13<<2)|2,{121,95,0}},
/* 62739 */ {(13<<2)|2,{121,95,0}},
/* 62740 */ {(13<<2)|2,{121,95,0}},
/* 62741 */ {(13<<2)|2,{121,95,0}},
/* 62742 */ {(13<<2)|2,{121,95,0}},
/* 62743 */ {(13<<2)|2,{121,95,0}},
/* 62744 */ {(13<<2)|2,{121,98,0}},
/* 62745 */ {(13<<2)|2,{121,98,0}},
/* 62746 */ {(13<<2)|2,{121,98,0}},
/* 62747 */ {(13<<2)|2,{121,98,0}},
/* 62748 */ {(13<<2)|2,{121,98,0}},
/* 62749 */ {(13<<2)|2,{121,98,0}},
/* 62750 */ {(13<<2)|2,{121,98,0}},
/* 62751 */ {(13<<2)|2,{121,98,0}},
/* 62752 */ {(13<<2)|2,{121,100,0}},
/* 62753 */ {(13<<2)|2,{121,100,0}},
/* 62754 */ {(13<<2)|2,{121,100,0}},
/* 62755 */ {(13<<2)|2,{121,100,0}},
/* 62756 */ {(13<<2)|2,{121,100,0}},
/* 62757 */ {(13<<2)|2,{121,100,0}},
/* 62758 */ {(13<<2)|2,{121,100,0}},
/* 62759 */ {(13<<2)|2,{121,100,0}},
/* 62760 */ {(13<<2)|2,{121,102,0}},
/* 62761 */ {(13<<2)|2,{121,102,0}},
/* 62762 */ {(13<<2)|2,{121,102,0}},
/* 62763 */ {(13<<2)|2,{121,102,0}},
/* 62764 */ {(13<<2)|2,{121,102,0}},
/* 62765 */ {(13<<2)|2,{121,102,0}},
/* 62766 */ {(13<<2)|2,{121,102,0}},
/* 62767 */ {(13<<2)|2,{121,102,0}},
/* 62768 */ {(13<<2)|2,{121,103,0}},
/* 62769 */ {(13<<2)|2,{121,103,0}},
/* 62770 */ {(13<<2)|2,{121,103,0}},
/* 62771 */ {(13<<2)|2,{121,103,0}},
/* 62772 */ {(13<<2)|2,{121,103,0}},
/* 62773 */ {(13<<2)|2,{121,103,0}},
/* 62774 */ {(13<<2)|2,{121,103,0}},
/* 62775 */ {(13<<2)|2,{121,103,0}},
/* 62776 */ {(13<<2)|2,{121,104,0}},
/* 62777 */ {(13<<2)|2,{121,104,0}},
/* 62778 */ {(13<<2)|2,{121,104,0}},
/* 62779 */ {(13<<2)|2,{121,104,0}},
/* 62780 */ {(13<<2)|2,{121,104,0}},
/* 62781 */ {(13<<2)|2,{121,104,0}},
/* 62782 */ {(13<<2)|2,{121,104,0}},
/* 62783 */ {(13<<2)|2,{121,104,0}},
/* 62784 */ {(13<<2)|2,{121,108,0}},
/* 62785 */ {(13<<2)|2,{121,108,0}},
/* 62786 */ {(13<<2)|2,{121,108,0}},
/* 62787 */ {(13<<2)|2,{121,108,0}},
/* 62788 */ {(13<<2)|2,{121,108,0}},
/* 62789 */ {(13<<2)|2,{121,108,0}},
/* 62790 */ {(13<<2)|2,{121,108,0}},
/* 62791 */ {(13<<2)|2,{121,108,0}},
/* 62792 */ {(13<<2)|2,{121,109,0}},
/* 62793 */ {(13<<2)|2,{121,109,0}},
/* 62794 */ {(13<<2)|2,{121,109,0}},
/* 62795 */ {(13<<2)|2,{121,109,0}},
/* 62796 */ {(13<<2)|2,{121,109,0}},
/* 62797 */ {(13<<2)|2,{121,109,0}},
/* 62798 */ {(13<<2)|2,{121,109,0}},
/* 62799 */ {(13<<2)|2,{121,109,0}},
/* 62800 */ {(13<<2)|2,{121,110,0}},
/* 62801 */ {(13<<2)|2,{121,110,0}},
/* 62802 */ {(13<<2)|2,{121,110,0}},
/* 62803 */ {(13<<2)|2,{121,110,0}},
/* 62804 */ {(13<<2)|2,{121,110,0}},
/* 62805 */ {(13<<2)|2,{121,110,0}},
/* 62806 */ {(13<<2)|2,{121,110,0}},
/* 62807 */ {(13<<2)|2,{121,110,0}},
/* 62808 */ {(13<<2)|2,{121,112,0}},
/* 62809 */ {(13<<2)|2,{121,112,0}},
/* 62810 */ {(13<<2)|2,{121,112,0}},
/* 62811 */ {(13<<2)|2,{121,112,0}},
/* 62812 */ {(13<<2)|2,{121,112,0}},
/* 62813 */ {(13<<2)|2,{121,112,0}},
/* 62814 */ {(13<<2)|2,{121,112,0}},
/* 62815 */ {(13<<2)|2,{121,112,0}},
/* 62816 */ {(13<<2)|2,{121,114,0}},
/* 62817 */ {(13<<2)|2,{121,114,0}},
/* 62818 */ {(13<<2)|2,{121,114,0}},
/* 62819 */ {(13<<2)|2,{121,114,0}},
/* 62820 */ {(13<<2)|2,{121,114,0}},
/* 62821 */ {(13<<2)|2,{121,114,0}},
/* 62822 */ {(13<<2)|2,{121,114,0}},
/* 62823 */ {(13<<2)|2,{121,114,0}},
/* 62824 */ {(13<<2)|2,{121,117,0}},
/* 62825 */ {(13<<2)|2,{121,117,0}},
/* 62826 */ {(13<<2)|2,{121,117,0}},
/* 62827 */ {(13<<2)|2,{121,117,0}},
/* 62828 */ {(13<<2)|2,{121,117,0}},
/* 62829 */ {(13<<2)|2,{121,117,0}},
/* 62830 */ {(13<<2)|2,{121,117,0}},
/* 62831 */ {(13<<2)|2,{121,117,0}},
/* 62832 */ {(14<<2)|2,{121,58,0}},
/* 62833 */ {(14<<2)|2,{121,58,0}},
/* 62834 */ {(14<<2)|2,{121,58,0}},
/* 62835 */ {(14<<2)|2,{121,58,0}},
/* 62836 */ {(14<<2)|2,{121,66,0}},
/* 62837 */ {(14<<2)|2,{121,66,0}},
/* 62838 */ {(14<<2)|2,{121,66,0}},
/* 62839 */ {(14<<2)|2,{121,66,0}},
/* 62840 */ {(14<<2)|2,{121,67,0}},
/* 62841 */ {(14<<2)|2,{121,67,0}},
/* 62842 */ {(14<<2)|2,{121,67,0}},
/* 62843 */ {(14<<2)|2,{121,67,0}},
/* 62844 */ {(14<<2)|2,{121,68,0}},
/* 62845 */ {(14<<2)|2,{121,68,0}},
/* 62846 */ {(14<<2)|2,{121,68,0}},
/* 62847 */ {(14<<2)|2,{121,68,0}},
/* 62848 */ {(14<<2)|2,{121,69,0}},
/* 62849 */ {(14<<2)|2,{121,69,0}},
/* 62850 */ {(14<<2)|2,{121,69,0}},
/* 62851 */ {(14<<2)|2,{121,69,0}},
/* 62852 */ {(14<<2)|2,{121,70,0}},
/* 62853 */ {(14<<2)|2,{121,70,0}},
/* 62854 */ {(14<<2)|2,{121,70,0}},
/* 62855 */ {(14<<2)|2,{121,70,0}},
/* 62856 */ {(14<<2)|2,{121,71,0}},
/* 62857 */ {(14<<2)|2,{121,71,0}},
/* 62858 */ {(14<<2)|2,{121,71,0}},
/* 62859 */ {(14<<2)|2,{121,71,0}},
/* 62860 */ {(14<<2)|2,{121,72,0}},
/* 62861 */ {(14<<2)|2,{121,72,0}},
/* 62862 */ {(14<<2)|2,{121,72,0}},
/* 62863 */ {(14<<2)|2,{121,72,0}},
/* 62864 */ {(14<<2)|2,{121,73,0}},
/* 62865 */ {(14<<2)|2,{121,73,0}},
/* 62866 */ {(14<<2)|2,{121,73,0}},
/* 62867 */ {(14<<2)|2,{121,73,0}},
/* 62868 */ {(14<<2)|2,{121,74,0}},
/* 62869 */ {(14<<2)|2,{121,74,0}},
/* 62870 */ {(14<<2)|2,{121,74,0}},
/* 62871 */ {(14<<2)|2,{121,74,0}},
/* 62872 */ {(14<<2)|2,{121,75,0}},
/* 62873 */ {(14<<2)|2,{121,75,0}},
/* 62874 */ {(14<<2)|2,{121,75,0}},
/* 62875 */ {(14<<2)|2,{121,75,0}},
/* 62876 */ {(14<<2)|2,{121,76,0}},
/* 62877 */ {(14<<2)|2,{121,76,0}},
/* 62878 */ {(14<<2)|2,{121,76,0}},
/* 62879 */ {(14<<2)|2,{121,76,0}},
/* 62880 */ {(14<<2)|2,{121,77,0}},
/* 62881 */ {(14<<2)|2,{121,77,0}},
/* 62882 */ {(14<<2)|2,{121,77,0}},
/* 62883 */ {(14<<2)|2,{121,77,0}},
/* 62884 */ {(14<<2)|2,{121,78,0}},
/* 62885 */ {(14<<2)|2,{121,78,0}},
/* 62886 */ {(14<<2)|2,{121,78,0}},
/* 62887 */ {(14<<2)|2,{121,78,0}},
/* 62888 */ {(14<<2)|2,{121,79,0}},
/* 62889 */ {(14<<2)|2,{121,79,0}},
/* 62890 */ {(14<<2)|2,{121,79,0}},
/* 62891 */ {(14<<2)|2,{121,79,0}},
/* 62892 */ {(14<<2)|2,{121,80,0}},
/* 62893 */ {(14<<2)|2,{121,80,0}},
/* 62894 */ {(14<<2)|2,{121,80,0}},
/* 62895 */ {(14<<2)|2,{121,80,0}},
/* 62896 */ {(14<<2)|2,{121,81,0}},
/* 62897 */ {(14<<2)|2,{121,81,0}},
/* 62898 */ {(14<<2)|2,{121,81,0}},
/* 62899 */ {(14<<2)|2,{121,81,0}},
/* 62900 */ {(14<<2)|2,{121,82,0}},
/* 62901 */ {(14<<2)|2,{121,82,0}},
/* 62902 */ {(14<<2)|2,{121,82,0}},
/* 62903 */ {(14<<2)|2,{121,82,0}},
/* 62904 */ {(14<<2)|2,{121,83,0}},
/* 62905 */ {(14<<2)|2,{121,83,0}},
/* 62906 */ {(14<<2)|2,{121,83,0}},
/* 62907 */ {(14<<2)|2,{121,83,0}},
/* 62908 */ {(14<<2)|2,{121,84,0}},
/* 62909 */ {(14<<2)|2,{121,84,0}},
/* 62910 */ {(14<<2)|2,{121,84,0}},
/* 62911 */ {(14<<2)|2,{121,84,0}},
/* 62912 */ {(14<<2)|2,{121,85,0}},
/* 62913 */ {(14<<2)|2,{121,85,0}},
/* 62914 */ {(14<<2)|2,{121,85,0}},
/* 62915 */ {(14<<2)|2,{121,85,0}},
/* 62916 */ {(14<<2)|2,{121,86,0}},
/* 62917 */ {(14<<2)|2,{121,86,0}},
/* 62918 */ {(14<<2)|2,{121,86,0}},
/* 62919 */ {(14<<2)|2,{121,86,0}},
/* 62920 */ {(14<<2)|2,{121,87,0}},
/* 62921 */ {(14<<2)|2,{121,87,0}},
/* 62922 */ {(14<<2)|2,{121,87,0}},
/* 62923 */ {(14<<2)|2,{121,87,0}},
/* 62924 */ {(14<<2)|2,{121,89,0}},
/* 62925 */ {(14<<2)|2,{121,89,0}},
/* 62926 */ {(14<<2)|2,{121,89,0}},
/* 62927 */ {(14<<2)|2,{121,89,0}},
/* 62928 */ {(14<<2)|2,{121,106,0}},
/* 62929 */ {(14<<2)|2,{121,106,0}},
/* 62930 */ {(14<<2)|2,{121,106,0}},
/* 62931 */ {(14<<2)|2,{121,106,0}},
/* 62932 */ {(14<<2)|2,{121,107,0}},
/* 62933 */ {(14<<2)|2,{121,107,0}},
/* 62934 */ {(14<<2)|2,{121,107,0}},
/* 62935 */ {(14<<2)|2,{121,107,0}},
/* 62936 */ {(14<<2)|2,{121,113,0}},
/* 62937 */ {(14<<2)|2,{121,113,0}},
/* 62938 */ {(14<<2)|2,{121,113,0}},
/* 62939 */ {(14<<2)|2,{121,113,0}},
/* 62940 */ {(14<<2)|2,{121,118,0}},
/* 62941 */ {(14<<2)|2,{121,118,0}},
/* 62942 */ {(14<<2)|2,{121,118,0}},
/* 62943 */ {(14<<2)|2,{121,118,0}},
/* 62944 */ {(14<<2)|2,{121,119,0}},
/* 62945 */ {(14<<2)|2,{121,119,0}},
/* 62946 */ {(14<<2)|2,{121,119,0}},
/* 62947 */ {(14<<2)|2,{121,119,0}},
/* 62948 */ {(14<<2)|2,{121,120,0}},
/* 62949 */ {(14<<2)|2,{121,120,0}},
/* 62950 */ {(14<<2)|2,{121,120,0}},
/* 62951 */ {(14<<2)|2,{121,120,0}},
/* 62952 */ {(14<<2)|2,{121,121,0}},
/* 62953 */ {(14<<2)|2,{121,121,0}},
/* 62954 */ {(14<<2)|2,{121,121,0}},
/* 62955 */ {(14<<2)|2,{121,121,0}},
/* 62956 */ {(14<<2)|2,{121,122,0}},
/* 62957 */ {(14<<2)|2,{121,122,0}},
/* 62958 */ {(14<<2)|2,{121,122,0}},
/* 62959 */ {(14<<2)|2,{121,122,0}},
/* 62960 */ {(15<<2)|2,{121,38,0}},
/* 62961 */ {(15<<2)|2,{121,38,0}},
/* 62962 */ {(15<<2)|2,{121,42,0}},
/* 62963 */ {(15<<2)|2,{121,42,0}},
/* 62964 */ {(15<<2)|2,{121,44,0}},
/* 62965 */ {(15<<2)|2,{121,44,0}},
/* 62966 */ {(15<<2)|2,{121,59,0}},
/* 62967 */ {(15<<2)|2,{121,59,0}},
/* 62968 */ {(15<<2)|2,{121,88,0}},
/* 62969 */ {(15<<2)|2,{121,88,0}},
/* 62970 */ {(15<<2)|2,{121,90,0}},
/* 62971 */ {(15<<2)|2,{121,90,0}},
/* 62972 */ {(7<<2)|1,{121,0,0}},
/* 62973 */ {(7<<2)|1,{121,0,0}},
/* 62974 */ {(7<<2)|1,{121,0,0}},
/* 62975 */ {(7<<2)|1,{121,0,0}},
/* 62976 */ {(12<<2)|2,{122,48,0}},
/* 62977 */ {(12<<2)|2,{122,48,0}},
/* 62978 */ {(12<<2)|2,{122,48,0}},
/* 62979 */ {(12<<2)|2,{122,48,0}},
/* 62980 */ {(12<<2)|2,{122,48,0}},
/* 62981 */ {(12<<2)|2,{122,48,0}},
/* 62982 */ {(12<<2)|2,{122,48,0}},
/* 62983 */ {(12<<2)|2,{122,48,0}},
/* 62984 */ {(12<<2)|2,{122,48,0}},
/* 62985 */ {(12<<2)|2,{122,48,0}},
/* 62986 */ {(12<<2)|2,{122,48,0}},
/* 62987 */ {(12<<2)|2,{122,48,0}},
/* 62988 */ {(12<<2)|2,{122,48,0}},
/* 62989 */ {(12<<2)|2,{122,48,0}},
/* 62990 */ {(12<<2)|2,{122,48,0}},
/* 62991 */ {(12<<2)|2,{122,48,0}},
/* 62992 */ {(12<<2)|2,{122,49,0}},
/* 62993 */ {(12<<2)|2,{122,49,0}},
/* 62994 */ {(12<<2)|2,{122,49,0}},
/* 62995 */ {(12<<2)|2,{122,49,0}},
/* 62996 */ {(12<<2)|2,{122,49,0}},
/* 62997 */ {(12<<2)|2,{122,49,0}},
/* 62998 */ {(12<<2)|2,{122,49,0}},
/* 62999 */ {(12<<2)|2,{122,49,0}},
/* 63000 */ {(12<<2)|2,{122,49,0}},
/* 63001 */ {(12<<2)|2,{122,49,0}},
/* 63002 */ {(12<<2)|2,{122,49,0}},
/* 63003 */ {(12<<2)|2,{122,49,0}},
/* 63004 */ {(12<<2)|2,{122,49,0}},
/* 63005 */ {(12<<2)|2,{122,49,0}},
/* 63006 */ {(12<<2)|2,{122,49,0}},
/* 63007 */ {(12<<2)|2,{122,49,0}},
/* 63008 */ {(12<<2)|2,{122,50,0}},
/* 63009 */ {(12<<2)|2,{122,50,0}},
/* 63010 */ {(12<<2)|2,{122,50,0}},
/* 63011 */ {(12<<2)|2,{122,50,0}},
/* 63012 */ {(12<<2)|2,{122,50,0}},
/* 63013 */ {(12<<2)|2,{122,50,0}},
/* 63014 */ {(12<<2)|2,{122,50,0}},
/* 63015 */ {(12<<2)|2,{122,50,0}},
/* 63016 */ {(12<<2)|2,{122,50,0}},
/* 63017 */ {(12<<2)|2,{122,50,0}},
/* 63018 */ {(12<<2)|2,{122,50,0}},
/* 63019 */ {(12<<2)|2,{122,50,0}},
/* 63020 */ {(12<<2)|2,{122,50,0}},
/* 63021 */ {(12<<2)|2,{122,50,0}},
/* 63022 */ {(12<<2)|2,{122,50,0}},
/* 63023 */ {(12<<2)|2,{122,50,0}},
/* 63024 */ {(12<<2)|2,{122,97,0}},
/* 63025 */ {(12<<2)|2,{122,97,0}},
/* 63026 */ {(12<<2)|2,{122,97,0}},
/* 63027 */ {(12<<2)|2,{122,97,0}},
/* 63028 */ {(12<<2)|2,{122,97,0}},
/* 63029 */ {(12<<2)|2,{122,97,0}},
/* 63030 */ {(12<<2)|2,{122,97,0}},
/* 63031 */ {(12<<2)|2,{122,97,0}},
/* 63032 */ {(12<<2)|2,{122,97,0}},
/* 63033 */ {(12<<2)|2,{122,97,0}},
/* 63034 */ {(12<<2)|2,{122,97,0}},
/* 63035 */ {(12<<2)|2,{122,97,0}},
/* 63036 */ {(12<<2)|2,{122,97,0}},
/* 63037 */ {(12<<2)|2,{122,97,0}},
/* 63038 */ {(12<<2)|2,{122,97,0}},
/* 63039 */ {(12<<2)|2,{122,97,0}},
/* 63040 */ {(12<<2)|2,{122,99,0}},
/* 63041 */ {(12<<2)|2,{122,99,0}},
/* 63042 */ {(12<<2)|2,{122,99,0}},
/* 63043 */ {(12<<2)|2,{122,99,0}},
/* 63044 */ {(12<<2)|2,{122,99,0}},
/* 63045 */ {(12<<2)|2,{122,99,0}},
/* 63046 */ {(12<<2)|2,{122,99,0}},
/* 63047 */ {(12<<2)|2,{122,99,0}},
/* 63048 */ {(12<<2)|2,{122,99,0}},
/* 63049 */ {(12<<2)|2,{122,99,0}},
/* 63050 */ {(12<<2)|2,{122,99,0}},
/* 63051 */ {(12<<2)|2,{122,99,0}},
/* 63052 */ {(12<<2)|2,{122,99,0}},
/* 63053 */ {(12<<2)|2,{122,99,0}},
/* 63054 */ {(12<<2)|2,{122,99,0}},
/* 63055 */ {(12<<2)|2,{122,99,0}},
/* 63056 */ {(12<<2)|2,{122,101,0}},
/* 63057 */ {(12<<2)|2,{122,101,0}},
/* 63058 */ {(12<<2)|2,{122,101,0}},
/* 63059 */ {(12<<2)|2,{122,101,0}},
/* 63060 */ {(12<<2)|2,{122,101,0}},
/* 63061 */ {(12<<2)|2,{122,101,0}},
/* 63062 */ {(12<<2)|2,{122,101,0}},
/* 63063 */ {(12<<2)|2,{122,101,0}},
/* 63064 */ {(12<<2)|2,{122,101,0}},
/* 63065 */ {(12<<2)|2,{122,101,0}},
/* 63066 */ {(12<<2)|2,{122,101,0}},
/* 63067 */ {(12<<2)|2,{122,101,0}},
/* 63068 */ {(12<<2)|2,{122,101,0}},
/* 63069 */ {(12<<2)|2,{122,101,0}},
/* 63070 */ {(12<<2)|2,{122,101,0}},
/* 63071 */ {(12<<2)|2,{122,101,0}},
/* 63072 */ {(12<<2)|2,{122,105,0}},
/* 63073 */ {(12<<2)|2,{122,105,0}},
/* 63074 */ {(12<<2)|2,{122,105,0}},
/* 63075 */ {(12<<2)|2,{122,105,0}},
/* 63076 */ {(12<<2)|2,{122,105,0}},
/* 63077 */ {(12<<2)|2,{122,105,0}},
/* 63078 */ {(12<<2)|2,{122,105,0}},
/* 63079 */ {(12<<2)|2,{122,105,0}},
/* 63080 */ {(12<<2)|2,{122,105,0}},
/* 63081 */ {(12<<2)|2,{122,105,0}},
/* 63082 */ {(12<<2)|2,{122,105,0}},
/* 63083 */ {(12<<2)|2,{122,105,0}},
/* 63084 */ {(12<<2)|2,{122,105,0}},
/* 63085 */ {(12<<2)|2,{122,105,0}},
/* 63086 */ {(12<<2)|2,{122,105,0}},
/* 63087 */ {(12<<2)|2,{122,105,0}},
/* 63088 */ {(12<<2)|2,{122,111,0}},
/* 63089 */ {(12<<2)|2,{122,111,0}},
/* 63090 */ {(12<<2)|2,{122,111,0}},
/* 63091 */ {(12<<2)|2,{122,111,0}},
/* 63092 */ {(12<<2)|2,{122,111,0}},
/* 63093 */ {(12<<2)|2,{122,111,0}},
/* 63094 */ {(12<<2)|2,{122,111,0}},
/* 63095 */ {(12<<2)|2,{122,111,0}},
/* 63096 */ {(12<<2)|2,{122,111,0}},
/* 63097 */ {(12<<2)|2,{122,111,0}},
/* 63098 */ {(12<<2)|2,{122,111,0}},
/* 63099 */ {(12<<2)|2,{122,111,0}},
/* 63100 */ {(12<<2)|2,{122,111,0}},
/* 63101 */ {(12<<2)|2,{122,111,0}},
/* 63102 */ {(12<<2)|2,{122,111,0}},
/* 63103 */ {(12<<2)|2,{122,111,0}},
/* 63104 */ {(12<<2)|2,{122,115,0}},
/* 63105 */ {(12<<2)|2,{122,115,0}},
/* 63106 */ {(12<<2)|2,{122,115,0}},
/* 63107 */ {(12<<2)|2,{122,115,0}},
/* 63108 */ {(12<<2)|2,{122,115,0}},
/* 63109 */ {(12<<2)|2,{122,115,0}},
/* 63110 */ {(12<<2)|2,{122,115,0}},
/* 63111 */ {(12<<2)|2,{122,115,0}},
/* 63112 */ {(12<<2)|2,{122,115,0}},
/* 63113 */ {(12<<2)|2,{122,115,0}},
/* 63114 */ {(12<<2)|2,{122,115,0}},
/* 63115 */ {(12<<2)|2,{122,115,0}},
/* 63116 */ {(12<<2)|2,{122,115,0}},
/* 63117 */ {(12<<2)|2,{122,115,0}},
/* 63118 */ {(12<<2)|2,{122,115,0}},
/* 63119 */ {(12<<2)|2,{122,115,0}},
/* 63120 */ {(12<<2)|2,{122,116,0}},
/* 63121 */ {(12<<2)|2,{122,116,0}},
/* 63122 */ {(12<<2)|2,{122,116,0}},
/* 63123 */ {(12<<2)|2,{122,116,0}},
/* 63124 */ {(12<<2)|2,{122,116,0}},
/* 63125 */ {(12<<2)|2,{122,116,0}},
/* 63126 */ {(12<<2)|2,{122,116,0}},
/* 63127 */ {(12<<2)|2,{122,116,0}},
/* 63128 */ {(12<<2)|2,{122,116,0}},
/* 63129 */ {(12<<2)|2,{122,116,0}},
/* 63130 */ {(12<<2)|2,{122,116,0}},
/* 63131 */ {(12<<2)|2,{122,116,0}},
/* 63132 */ {(12<<2)|2,{122,116,0}},
/* 63133 */ {(12<<2)|2,{122,116,0}},
/* 63134 */ {(12<<2)|2,{122,116,0}},
/* 63135 */ {(12<<2)|2,{122,116,0}},
/* 63136 */ {(13<<2)|2,{122,32,0}},
/* 63137 */ {(13<<2)|2,{122,32,0}},
/* 63138 */ {(13<<2)|2,{122,32,0}},
/* 63139 */ {(13<<2)|2,{122,32,0}},
/* 63140 */ {(13<<2)|2,{122,32,0}},
/* 63141 */ {(13<<2)|2,{122,32,0}},
/* 63142 */ {(13<<2)|2,{122,32,0}},
/* 63143 */ {(13<<2)|2,{122,32,0}},
/* 63144 */ {(13<<2)|2,{122,37,0}},
/* 63145 */ {(13<<2)|2,{122,37,0}},
/* 63146 */ {(13<<2)|2,{122,37,0}},
/* 63147 */ {(13<<2)|2,{122,37,0}},
/* 63148 */ {(13<<2)|2,{122,37,0}},
/* 63149 */ {(13<<2)|2,{122,37,0}},
/* 63150 */ {(13<<2)|2,{122,37,0}},
/* 63151 */ {(13<<2)|2,{122,37,0}},
/* 63152 */ {(13<<2)|2,{122,45,0}},
/* 63153 */ {(13<<2)|2,{122,45,0}},
/* 63154 */ {(13<<2)|2,{122,45,0}},
/* 63155 */ {(13<<2)|2,{122,45,0}},
/* 63156 */ {(13<<2)|2,{122,45,0}},
/* 63157 */ {(13<<2)|2,{122,45,0}},
/* 63158 */ {(13<<2)|2,{122,45,0}},
/* 63159 */ {(13<<2)|2,{122,45,0}},
/* 63160 */ {(13<<2)|2,{122,46,0}},
/* 63161 */ {(13<<2)|2,{122,46,0}},
/* 63162 */ {(13<<2)|2,{122,46,0}},
/* 63163 */ {(13<<2)|2,{122,46,0}},
/* 63164 */ {(13<<2)|2,{122,46,0}},
/* 63165 */ {(13<<2)|2,{122,46,0}},
/* 63166 */ {(13<<2)|2,{122,46,0}},
/* 63167 */ {(13<<2)|2,{122,46,0}},
/* 63168 */ {(13<<2)|2,{122,47,0}},
/* 63169 */ {(13<<2)|2,{122,47,0}},
/* 63170 */ {(13<<2)|2,{122,47,0}},
/* 63171 */ {(13<<2)|2,{122,47,0}},
/* 63172 */ {(13<<2)|2,{122,47,0}},
/* 63173 */ {(13<<2)|2,{122,47,0}},
/* 63174 */ {(13<<2)|2,{122,47,0}},
/* 63175 */ {(13<<2)|2,{122,47,0}},
/* 63176 */ {(13<<2)|2,{122,51,0}},
/* 63177 */ {(13<<2)|2,{122,51,0}},
/* 63178 */ {(13<<2)|2,{122,51,0}},
/* 63179 */ {(13<<2)|2,{122,51,0}},
/* 63180 */ {(13<<2)|2,{122,51,0}},
/* 63181 */ {(13<<2)|2,{122,51,0}},
/* 63182 */ {(13<<2)|2,{122,51,0}},
/* 63183 */ {(13<<2)|2,{122,51,0}},
/* 63184 */ {(13<<2)|2,{122,52,0}},
/* 63185 */ {(13<<2)|2,{122,52,0}},
/* 63186 */ {(13<<2)|2,{122,52,0}},
/* 63187 */ {(13<<2)|2,{122,52,0}},
/* 63188 */ {(13<<2)|2,{122,52,0}},
/* 63189 */ {(13<<2)|2,{122,52,0}},
/* 63190 */ {(13<<2)|2,{122,52,0}},
/* 63191 */ {(13<<2)|2,{122,52,0}},
/* 63192 */ {(13<<2)|2,{122,53,0}},
/* 63193 */ {(13<<2)|2,{122,53,0}},
/* 63194 */ {(13<<2)|2,{122,53,0}},
/* 63195 */ {(13<<2)|2,{122,53,0}},
/* 63196 */ {(13<<2)|2,{122,53,0}},
/* 63197 */ {(13<<2)|2,{122,53,0}},
/* 63198 */ {(13<<2)|2,{122,53,0}},
/* 63199 */ {(13<<2)|2,{122,53,0}},
/* 63200 */ {(13<<2)|2,{122,54,0}},
/* 63201 */ {(13<<2)|2,{122,54,0}},
/* 63202 */ {(13<<2)|2,{122,54,0}},
/* 63203 */ {(13<<2)|2,{122,54,0}},
/* 63204 */ {(13<<2)|2,{122,54,0}},
/* 63205 */ {(13<<2)|2,{122,54,0}},
/* 63206 */ {(13<<2)|2,{122,54,0}},
/* 63207 */ {(13<<2)|2,{122,54,0}},
/* 63208 */ {(13<<2)|2,{122,55,0}},
/* 63209 */ {(13<<2)|2,{122,55,0}},
/* 63210 */ {(13<<2)|2,{122,55,0}},
/* 63211 */ {(13<<2)|2,{122,55,0}},
/* 63212 */ {(13<<2)|2,{122,55,0}},
/* 63213 */ {(13<<2)|2,{122,55,0}},
/* 63214 */ {(13<<2)|2,{122,55,0}},
/* 63215 */ {(13<<2)|2,{122,55,0}},
/* 63216 */ {(13<<2)|2,{122,56,0}},
/* 63217 */ {(13<<2)|2,{122,56,0}},
/* 63218 */ {(13<<2)|2,{122,56,0}},
/* 63219 */ {(13<<2)|2,{122,56,0}},
/* 63220 */ {(13<<2)|2,{122,56,0}},
/* 63221 */ {(13<<2)|2,{122,56,0}},
/* 63222 */ {(13<<2)|2,{122,56,0}},
/* 63223 */ {(13<<2)|2,{122,56,0}},
/* 63224 */ {(13<<2)|2,{122,57,0}},
/* 63225 */ {(13<<2)|2,{122,57,0}},
/* 63226 */ {(13<<2)|2,{122,57,0}},
/* 63227 */ {(13<<2)|2,{122,57,0}},
/* 63228 */ {(13<<2)|2,{122,57,0}},
/* 63229 */ {(13<<2)|2,{122,57,0}},
/* 63230 */ {(13<<2)|2,{122,57,0}},
/* 63231 */ {(13<<2)|2,{122,57,0}},
/* 63232 */ {(13<<2)|2,{122,61,0}},
/* 63233 */ {(13<<2)|2,{122,61,0}},
/* 63234 */ {(13<<2)|2,{122,61,0}},
/* 63235 */ {(13<<2)|2,{122,61,0}},
/* 63236 */ {(13<<2)|2,{122,61,0}},
/* 63237 */ {(13<<2)|2,{122,61,0}},
/* 63238 */ {(13<<2)|2,{122,61,0}},
/* 63239 */ {(13<<2)|2,{122,61,0}},
/* 63240 */ {(13<<2)|2,{122,65,0}},
/* 63241 */ {(13<<2)|2,{122,65,0}},
/* 63242 */ {(13<<2)|2,{122,65,0}},
/* 63243 */ {(13<<2)|2,{122,65,0}},
/* 63244 */ {(13<<2)|2,{122,65,0}},
/* 63245 */ {(13<<2)|2,{122,65,0}},
/* 63246 */ {(13<<2)|2,{122,65,0}},
/* 63247 */ {(13<<2)|2,{122,65,0}},
/* 63248 */ {(13<<2)|2,{122,95,0}},
/* 63249 */ {(13<<2)|2,{122,95,0}},
/* 63250 */ {(13<<2)|2,{122,95,0}},
/* 63251 */ {(13<<2)|2,{122,95,0}},
/* 63252 */ {(13<<2)|2,{122,95,0}},
/* 63253 */ {(13<<2)|2,{122,95,0}},
/* 63254 */ {(13<<2)|2,{122,95,0}},
/* 63255 */ {(13<<2)|2,{122,95,0}},
/* 63256 */ {(13<<2)|2,{122,98,0}},
/* 63257 */ {(13<<2)|2,{122,98,0}},
/* 63258 */ {(13<<2)|2,{122,98,0}},
/* 63259 */ {(13<<2)|2,{122,98,0}},
/* 63260 */ {(13<<2)|2,{122,98,0}},
/* 63261 */ {(13<<2)|2,{122,98,0}},
/* 63262 */ {(13<<2)|2,{122,98,0}},
/* 63263 */ {(13<<2)|2,{122,98,0}},
/* 63264 */ {(13<<2)|2,{122,100,0}},
/* 63265 */ {(13<<2)|2,{122,100,0}},
/* 63266 */ {(13<<2)|2,{122,100,0}},
/* 63267 */ {(13<<2)|2,{122,100,0}},
/* 63268 */ {(13<<2)|2,{122,100,0}},
/* 63269 */ {(13<<2)|2,{122,100,0}},
/* 63270 */ {(13<<2)|2,{122,100,0}},
/* 63271 */ {(13<<2)|2,{122,100,0}},
/* 63272 */ {(13<<2)|2,{122,102,0}},
/* 63273 */ {(13<<2)|2,{122,102,0}},
/* 63274 */ {(13<<2)|2,{122,102,0}},
/* 63275 */ {(13<<2)|2,{122,102,0}},
/* 63276 */ {(13<<2)|2,{122,102,0}},
/* 63277 */ {(13<<2)|2,{122,102,0}},
/* 63278 */ {(13<<2)|2,{122,102,0}},
/* 63279 */ {(13<<2)|2,{122,102,0}},
/* 63280 */ {(13<<2)|2,{122,103,0}},
/* 63281 */ {(13<<2)|2,{122,103,0}},
/* 63282 */ {(13<<2)|2,{122,103,0}},
/* 63283 */ {(13<<2)|2,{122,103,0}},
/* 63284 */ {(13<<2)|2,{122,103,0}},
/* 63285 */ {(13<<2)|2,{122,103,0}},
/* 63286 */ {(13<<2)|2,{122,103,0}},
/* 63287 */ {(13<<2)|2,{122,103,0}},
/* 63288 */ {(13<<2)|2,{122,104,0}},
/* 63289 */ {(13<<2)|2,{122,104,0}},
/* 63290 */ {(13<<2)|2,{122,104,0}},
/* 63291 */ {(13<<2)|2,{122,104,0}},
/* 63292 */ {(13<<2)|2,{122,104,0}},
/* 63293 */ {(13<<2)|2,{122,104,0}},
/* 63294 */ {(13<<2)|2,{122,104,0}},
/* 63295 */ {(13<<2)|2,{122,104,0}},
/* 63296 */ {(13<<2)|2,{122,108,0}},
/* 63297 */ {(13<<2)|2,{122,108,0}},
/* 63298 */ {(13<<2)|2,{122,108,0}},
/* 63299 */ {(13<<2)|2,{122,108,0}},
/* 63300 */ {(13<<2)|2,{122,108,0}},
/* 63301 */ {(13<<2)|2,{122,108,0}},
/* 63302 */ {(13<<2)|2,{122,108,0}},
/* 63303 */ {(13<<2)|2,{122,108,0}},
/* 63304 */ {(13<<2)|2,{122,109,0}},
/* 63305 */ {(13<<2)|2,{122,109,0}},
/* 63306 */ {(13<<2)|2,{122,109,0}},
/* 63307 */ {(13<<2)|2,{122,109,0}},
/* 63308 */ {(13<<2)|2,{122,109,0}},
/* 63309 */ {(13<<2)|2,{122,109,0}},
/* 63310 */ {(13<<2)|2,{122,109,0}},
/* 63311 */ {(13<<2)|2,{122,109,0}},
/* 63312 */ {(13<<2)|2,{122,110,0}},
/* 63313 */ {(13<<2)|2,{122,110,0}},
/* 63314 */ {(13<<2)|2,{122,110,0}},
/* 63315 */ {(13<<2)|2,{122,110,0}},
/* 63316 */ {(13<<2)|2,{122,110,0}},
/* 63317 */ {(13<<2)|2,{122,110,0}},
/* 63318 */ {(13<<2)|2,{122,110,0}},
/* 63319 */ {(13<<2)|2,{122,110,0}},
/* 63320 */ {(13<<2)|2,{122,112,0}},
/* 63321 */ {(13<<2)|2,{122,112,0}},
/* 63322 */ {(13<<2)|2,{122,112,0}},
/* 63323 */ {(13<<2)|2,{122,112,0}},
/* 63324 */ {(13<<2)|2,{122,112,0}},
/* 63325 */ {(13<<2)|2,{122,112,0}},
/* 63326 */ {(13<<2)|2,{122,112,0}},
/* 63327 */ {(13<<2)|2,{122,112,0}},
/* 63328 */ {(13<<2)|2,{122,114,0}},
/* 63329 */ {(13<<2)|2,{122,114,0}},
/* 63330 */ {(13<<2)|2,{122,114,0}},
/* 63331 */ {(13<<2)|2,{122,114,0}},
/* 63332 */ {(13<<2)|2,{122,114,0}},
/* 63333 */ {(13<<2)|2,{122,114,0}},
/* 63334 */ {(13<<2)|2,{122,114,0}},
/* 63335 */ {(13<<2)|2,{122,114,0}},
/* 63336 */ {(13<<2)|2,{122,117,0}},
/* 63337 */ {(13<<2)|2,{122,117,0}},
/* 63338 */ {(13<<2)|2,{122,117,0}},
/* 63339 */ {(13<<2)|2,{122,117,0}},
/* 63340 */ {(13<<2)|2,{122,117,0}},
/* 63341 */ {(13<<2)|2,{122,117,0}},
/* 63342 */ {(13<<2)|2,{122,117,0}},
/* 63343 */ {(13<<2)|2,{122,117,0}},
/* 63344 */ {(14<<2)|2,{122,58,0}},
/* 63345 */ {(14<<2)|2,{122,58,0}},
/* 63346 */ {(14<<2)|2,{122,58,0}},
/* 63347 */ {(14<<2)|2,{122,58,0}},
/* 63348 */ {(14<<2)|2,{122,66,0}},
/* 63349 */ {(14<<2)|2,{122,66,0}},
/* 63350 */ {(14<<2)|2,{122,66,0}},
/* 63351 */ {(14<<2)|2,{122,66,0}},
/* 63352 */ {(14<<2)|2,{122,67,0}},
/* 63353 */ {(14<<2)|2,{122,67,0}},
/* 63354 */ {(14<<2)|2,{122,67,0}},
/* 63355 */ {(14<<2)|2,{122,67,0}},
/* 63356 */ {(14<<2)|2,{122,68,0}},
/* 63357 */ {(14<<2)|2,{122,68,0}},
/* 63358 */ {(14<<2)|2,{122,68,0}},
/* 63359 */ {(14<<2)|2,{122,68,0}},
/* 63360 */ {(14<<2)|2,{122,69,0}},
/* 63361 */ {(14<<2)|2,{122,69,0}},
/* 63362 */ {(14<<2)|2,{122,69,0}},
/* 63363 */ {(14<<2)|2,{122,69,0}},
/* 63364 */ {(14<<2)|2,{122,70,0}},
/* 63365 */ {(14<<2)|2,{122,70,0}},
/* 63366 */ {(14<<2)|2,{122,70,0}},
/* 63367 */ {(14<<2)|2,{122,70,0}},
/* 63368 */ {(14<<2)|2,{122,71,0}},
/* 63369 */ {(14<<2)|2,{122,71,0}},
/* 63370 */ {(14<<2)|2,{122,71,0}},
/* 63371 */ {(14<<2)|2,{122,71,0}},
/* 63372 */ {(14<<2)|2,{122,72,0}},
/* 63373 */ {(14<<2)|2,{122,72,0}},
/* 63374 */ {(14<<2)|2,{122,72,0}},
/* 63375 */ {(14<<2)|2,{122,72,0}},
/* 63376 */ {(14<<2)|2,{122,73,0}},
/* 63377 */ {(14<<2)|2,{122,73,0}},
/* 63378 */ {(14<<2)|2,{122,73,0}},
/* 63379 */ {(14<<2)|2,{122,73,0}},
/* 63380 */ {(14<<2)|2,{122,74,0}},
/* 63381 */ {(14<<2)|2,{122,74,0}},
/* 63382 */ {(14<<2)|2,{122,74,0}},
/* 63383 */ {(14<<2)|2,{122,74,0}},
/* 63384 */ {(14<<2)|2,{122,75,0}},
/* 63385 */ {(14<<2)|2,{122,75,0}},
/* 63386 */ {(14<<2)|2,{122,75,0}},
/* 63387 */ {(14<<2)|2,{122,75,0}},
/* 63388 */ {(14<<2)|2,{122,76,0}},
/* 63389 */ {(14<<2)|2,{122,76,0}},
/* 63390 */ {(14<<2)|2,{122,76,0}},
/* 63391 */ {(14<<2)|2,{122,76,0}},
/* 63392 */ {(14<<2)|2,{122,77,0}},
/* 63393 */ {(14<<2)|2,{122,77,0}},
/* 63394 */ {(14<<2)|2,{122,77,0}},
/* 63395 */ {(14<<2)|2,{122,77,0}},
/* 63396 */ {(14<<2)|2,{122,78,0}},
/* 63397 */ {(14<<2)|2,{122,78,0}},
/* 63398 */ {(14<<2)|2,{122,78,0}},
/* 63399 */ {(14<<2)|2,{122,78,0}},
/* 63400 */ {(14<<2)|2,{122,79,0}},
/* 63401 */ {(14<<2)|2,{122,79,0}},
/* 63402 */ {(14<<2)|2,{122,79,0}},
/* 63403 */ {(14<<2)|2,{122,79,0}},
/* 63404 */ {(14<<2)|2,{122,80,0}},
/* 63405 */ {(14<<2)|2,{122,80,0}},
/* 63406 */ {(14<<2)|2,{122,80,0}},
/* 63407 */ {(14<<2)|2,{122,80,0}},
/* 63408 */ {(14<<2)|2,{122,81,0}},
/* 63409 */ {(14<<2)|2,{122,81,0}},
/* 63410 */ {(14<<2)|2,{122,81,0}},
/* 63411 */ {(14<<2)|2,{122,81,0}},
/* 63412 */ {(14<<2)|2,{122,82,0}},
/* 63413 */ {(14<<2)|2,{122,82,0}},
/* 63414 */ {(14<<2)|2,{122,82,0}},
/* 63415 */ {(14<<2)|2,{122,82,0}},
/* 63416 */ {(14<<2)|2,{122,83,0}},
/* 63417 */ {(14<<2)|2,{122,83,0}},
/* 63418 */ {(14<<2)|2,{122,83,0}},
/* 63419 */ {(14<<2)|2,{122,83,0}},
/* 63420 */ {(14<<2)|2,{122,84,0}},
/* 63421 */ {(14<<2)|2,{122,84,0}},
/* 63422 */ {(14<<2)|2,{122,84,0}},
/* 63423 */ {(14<<2)|2,{122,84,0}},
/* 63424 */ {(14<<2)|2,{122,85,0}},
/* 63425 */ {(14<<2)|2,{122,85,0}},
/* 63426 */ {(14<<2)|2,{122,85,0}},
/* 63427 */ {(14<<2)|2,{122,85,0}},
/* 63428 */ {(14<<2)|2,{122,86,0}},
/* 63429 */ {(14<<2)|2,{122,86,0}},
/* 63430 */ {(14<<2)|2,{122,86,0}},
/* 63431 */ {(14<<2)|2,{122,86,0}},
/* 63432 */ {(14<<2)|2,{122,87,0}},
/* 63433 */ {(14<<2)|2,{122,87,0}},
/* 63434 */ {(14<<2)|2,{122,87,0}},
/* 63435 */ {(14<<2)|2,{122,87,0}},
/* 63436 */ {(14<<2)|2,{122,89,0}},
/* 63437 */ {(14<<2)|2,{122,89,0}},
/* 63438 */ {(14<<2)|2,{122,89,0}},
/* 63439 */ {(14<<2)|2,{122,89,0}},
/* 63440 */ {(14<<2)|2,{122,106,0}},
/* 63441 */ {(14<<2)|2,{122,106,0}},
/* 63442 */ {(14<<2)|2,{122,106,0}},
/* 63443 */ {(14<<2)|2,{122,106,0}},
/* 63444 */ {(14<<2)|2,{122,107,0}},
/* 63445 */ {(14<<2)|2,{122,107,0}},
/* 63446 */ {(14<<2)|2,{122,107,0}},
/* 63447 */ {(14<<2)|2,{122,107,0}},
/* 63448 */ {(14<<2)|2,{122,113,0}},
/* 63449 */ {(14<<2)|2,{122,113,0}},
/* 63450 */ {(14<<2)|2,{122,113,0}},
/* 63451 */ {(14<<2)|2,{122,113,0}},
/* 63452 */ {(14<<2)|2,{122,118,0}},
/* 63453 */ {(14<<2)|2,{122,118,0}},
/* 63454 */ {(14<<2)|2,{122,118,0}},
/* 63455 */ {(14<<2)|2,{122,118,0}},
/* 63456 */ {(14<<2)|2,{122,119,0}},
/* 63457 */ {(14<<2)|2,{122,119,0}},
/* 63458 */ {(14<<2)|2,{122,119,0}},
/* 63459 */ {(14<<2)|2,{122,119,0}},
/* 63460 */ {(14<<2)|2,{122,120,0}},
/* 63461 */ {(14<<2)|2,{122,120,0}},
/* 63462 */ {(14<<2)|2,{122,120,0}},
/* 63463 */ {(14<<2)|2,{122,120,0}},
/* 63464 */ {(14<<2)|2,{122,121,0}},
/* 63465 */ {(14<<2)|2,{122,121,0}},
/* 63466 */ {(14<<2)|2,{122,121,0}},
/* 63467 */ {(14<<2)|2,{122,121,0}},
/* 63468 */ {(14<<2)|2,{122,122,0}},
/* 63469 */ {(14<<2)|2,{122,122,0}},
/* 63470 */ {(14<<2)|2,{122,122,0}},
/* 63471 */ {(14<<2)|2,{122,122,0}},
/* 63472 */ {(15<<2)|2,{122,38,0}},
/* 63473 */ {(15<<2)|2,{122,38,0}},
/* 63474 */ {(15<<2)|2,{122,42,0}},
/* 63475 */ {(15<<2)|2,{122,42,0}},
/* 63476 */ {(15<<2)|2,{122,44,0}},
/* 63477 */ {(15<<2)|2,{122,44,0}},
/* 63478 */ {(15<<2)|2,{122,59,0}},
/* 63479 */ {(15<<2)|2,{122,59,0}},
/* 63480 */ {(15<<2)|2,{122,88,0}},
/* 63481 */ {(15<<2)|2,{122,88,0}},
/* 63482 */ {(15<<2)|2,{122,90,0}},
/* 63483 */ {(15<<2)|2,{122,90,0}},
/* 63484 */ {(7<<2)|1,{122,0,0}},
/* 63485 */ {(7<<2)|1,{122,0,0}},
/* 63486 */ {(7<<2)|1,{122,0,0}},
/* 63487 */ {(7<<2)|1,{122,0,0}},
/* 63488 */ {(13<<2)|2,{38,48,0}},
/* 63489 */ {(13<<2)|2,{38,48,0}},
/* 63490 */ {(13<<2)|2,{38,48,0}},
/* 63491 */ {(13<<2)|2,{38,48,0}},
/* 63492 */ {(13<<2)|2,{38,48,0}},
/* 63493 */ {(13<<2)|2,{38,48,0}},
/* 63494 */ {(13<<2)|2,{38,48,0}},
/* 63495 */ {(13<<2)|2,{38,48,0}},
/* 63496 */ {(13<<2)|2,{38,49,0}},
/* 63497 */ {(13<<2)|2,{38,49,0}},
/* 63498 */ {(13<<2)|2,{38,49,0}},
/* 63499 */ {(13<<2)|2,{38,49,0}},
/* 63500 */ {(13<<2)|2,{38,49,0}},
/* 63501 */ {(13<<2)|2,{38,49,0}},
/* 63502 */ {(13<<2)|2,{38,49,0}},
/* 63503 */ {(13<<2)|2,{38,49,0}},
/* 63504 */ {(13<<2)|2,{38,50,0}},
/* 63505 */ {(13<<2)|2,{38,50,0}},
/* 63506 */ {(13<<2)|2,{38,50,0}},
/* 63507 */ {(13<<2)|2,{38,50,0}},
/* 63508 */ {(13<<2)|2,{38,50,0}},
/* 63509 */ {(13<<2)|2,{38,50,0}},
/* 63510 */ {(13<<2)|2,{38,50,0}},
/* 63511 */ {(13<<2)|2,{38,50,0}},
/* 63512 */ {(13<<2)|2,{38,97,0}},
/* 63513 */ {(13<<2)|2,{38,97,0}},
/* 63514 */ {(13<<2)|2,{38,97,0}},
/* 63515 */ {(13<<2)|2,{38,97,0}},
/* 63516 */ {(13<<2)|2,{38,97,0}},
/* 63517 */ {(13<<2)|2,{38,97,0}},
/* 63518 */ {(13<<2)|2,{38,97,0}},
/* 63519 */ {(13<<2)|2,{38,97,0}},
/* 63520 */ {(13<<2)|2,{38,99,0}},
/* 63521 */ {(13<<2)|2,{38,99,0}},
/* 63522 */ {(13<<2)|2,{38,99,0}},
/* 63523 */ {(13<<2)|2,{38,99,0}},
/* 63524 */ {(13<<2)|2,{38,99,0}},
/* 63525 */ {(13<<2)|2,{38,99,0}},
/* 63526 */ {(13<<2)|2,{38,99,0}},
/* 63527 */ {(13<<2)|2,{38,99,0}},
/* 63528 */ {(13<<2)|2,{38,101,0}},
/* 63529 */ {(13<<2)|2,{38,101,0}},
/* 63530 */ {(13<<2)|2,{38,101,0}},
/* 63531 */ {(13<<2)|2,{38,101,0}},
/* 63532 */ {(13<<2)|2,{38,101,0}},
/* 63533 */ {(13<<2)|2,{38,101,0}},
/* 63534 */ {(13<<2)|2,{38,101,0}},
/* 63535 */ {(13<<2)|2,{38,101,0}},
/* 63536 */ {(13<<2)|2,{38,105,0}},
/* 63537 */ {(13<<2)|2,{38,105,0}},
/* 63538 */ {(13<<2)|2,{38,105,0}},
/* 63539 */ {(13<<2)|2,{38,105,0}},
/* 63540 */ {(13<<2)|2,{38,105,0}},
/* 63541 */ {(13<<2)|2,{38,105,0}},
/* 63542 */ {(13<<2)|2,{38,105,0}},
/* 63543 */ {(13<<2)|2,{38,105,0}},
/* 63544 */ {(13<<2)|2,{38,111,0}},
/* 63545 */ {(13<<2)|2,{38,111,0}},
/* 63546 */ {(13<<2)|2,{38,111,0}},
/* 63547 */ {(13<<2)|2,{38,111,0}},
/* 63548 */ {(13<<2)|2,{38,111,0}},
/* 63549 */ {(13<<2)|2,{38,111,0}},
/* 63550 */ {(13<<2)|2,{38,111,0}},
/* 63551 */ {(13<<2)|2,{38,111,0}},
/* 63552 */ {(13<<2)|2,{38,115,0}},
/* 63553 */ {(13<<2)|2,{38,115,0}},
/* 63554 */ {(13<<2)|2,{38,115,0}},
/* 63555 */ {(13<<2)|2,{38,115,0}},
/* 63556 */ {(13<<2)|2,{38,115,0}},
/* 63557 */ {(13<<2)|2,{38,115,0}},
/* 63558 */ {(13<<2)|2,{38,115,0}},
/* 63559 */ {(13<<2)|2,{38,115,0}},
/* 63560 */ {(13<<2)|2,{38,116,0}},
/* 63561 */ {(13<<2)|2,{38,116,0}},
/* 63562 */ {(13<<2)|2,{38,116,0}},
/* 63563 */ {(13<<2)|2,{38,116,0}},
/* 63564 */ {(13<<2)|2,{38,116,0}},
/* 63565 */ {(13<<2)|2,{38,116,0}},
/* 63566 */ {(13<<2)|2,{38,116,0}},
/* 63567 */ {(13<<2)|2,{38,116,0}},
/* 63568 */ {(14<<2)|2,{38,32,0}},
/* 63569 */ {(14<<2)|2,{38,32,0}},
/* 63570 */ {(14<<2)|2,{38,32,0}},
/* 63571 */ {(14<<2)|2,{38,32,0}},
/* 63572 */ {(14<<2)|2,{38,37,0}},
/* 63573 */ {(14<<2)|2,{38,37,0}},
/* 63574 */ {(14<<2)|2,{38,37,0}},
/* 63575 */ {(14<<2)|2,{38,37,0}},
/* 63576 */ {(14<<2)|2,{38,45,0}},
/* 63577 */ {(14<<2)|2,{38,45,0}},
/* 63578 */ {(14<<2)|2,{38,45,0}},
/* 63579 */ {(14<<2)|2,{38,45,0}},
/* 63580 */ {(14<<2)|2,{38,46,0}},
/* 63581 */ {(14<<2)|2,{38,46,0}},
/* 63582 */ {(14<<2)|2,{38,46,0}},
/* 63583 */ {(14<<2)|2,{38,46,0}},
/* 63584 */ {(14<<2)|2,{38,47,0}},
/* 63585 */ {(14<<2)|2,{38,47,0}},
/* 63586 */ {(14<<2)|2,{38,47,0}},
/* 63587 */ {(14<<2)|2,{38,47,0}},
/* 63588 */ {(14<<2)|2,{38,51,0}},
/* 63589 */ {(14<<2)|2,{38,51,0}},
/* 63590 */ {(14<<2)|2,{38,51,0}},
/* 63591 */ {(14<<2)|2,{38,51,0}},
/* 63592 */ {(14<<2)|2,{38,52,0}},
/* 63593 */ {(14<<2)|2,{38,52,0}},
/* 63594 */ {(14<<2)|2,{38,52,0}},
/* 63595 */ {(14<<2)|2,{38,52,0}},
/* 63596 */ {(14<<2)|2,{38,53,0}},
/* 63597 */ {(14<<2)|2,{38,53,0}},
/* 63598 */ {(14<<2)|2,{38,53,0}},
/* 63599 */ {(14<<2)|2,{38,53,0}},
/* 63600 */ {(14<<2)|2,{38,54,0}},
/* 63601 */ {(14<<2)|2,{38,54,0}},
/* 63602 */ {(14<<2)|2,{38,54,0}},
/* 63603 */ {(14<<2)|2,{38,54,0}},
/* 63604 */ {(14<<2)|2,{38,55,0}},
/* 63605 */ {(14<<2)|2,{38,55,0}},
/* 63606 */ {(14<<2)|2,{38,55,0}},
/* 63607 */ {(14<<2)|2,{38,55,0}},
/* 63608 */ {(14<<2)|2,{38,56,0}},
/* 63609 */ {(14<<2)|2,{38,56,0}},
/* 63610 */ {(14<<2)|2,{38,56,0}},
/* 63611 */ {(14<<2)|2,{38,56,0}},
/* 63612 */ {(14<<2)|2,{38,57,0}},
/* 63613 */ {(14<<2)|2,{38,57,0}},
/* 63614 */ {(14<<2)|2,{38,57,0}},
/* 63615 */ {(14<<2)|2,{38,57,0}},
/* 63616 */ {(14<<2)|2,{38,61,0}},
/* 63617 */ {(14<<2)|2,{38,61,0}},
/* 63618 */ {(14<<2)|2,{38,61,0}},
/* 63619 */ {(14<<2)|2,{38,61,0}},
/* 63620 */ {(14<<2)|2,{38,65,0}},
/* 63621 */ {(14<<2)|2,{38,65,0}},
/* 63622 */ {(14<<2)|2,{38,65,0}},
/* 63623 */ {(14<<2)|2,{38,65,0}},
/* 63624 */ {(14<<2)|2,{38,95,0}},
/* 63625 */ {(14<<2)|2,{38,95,0}},
/* 63626 */ {(14<<2)|2,{38,95,0}},
/* 63627 */ {(14<<2)|2,{38,95,0}},
/* 63628 */ {(14<<2)|2,{38,98,0}},
/* 63629 */ {(14<<2)|2,{38,98,0}},
/* 63630 */ {(14<<2)|2,{38,98,0}},
/* 63631 */ {(14<<2)|2,{38,98,0}},
/* 63632 */ {(14<<2)|2,{38,100,0}},
/* 63633 */ {(14<<2)|2,{38,100,0}},
/* 63634 */ {(14<<2)|2,{38,100,0}},
/* 63635 */ {(14<<2)|2,{38,100,0}},
/* 63636 */ {(14<<2)|2,{38,102,0}},
/* 63637 */ {(14<<2)|2,{38,102,0}},
/* 63638 */ {(14<<2)|2,{38,102,0}},
/* 63639 */ {(14<<2)|2,{38,102,0}},
/* 63640 */ {(14<<2)|2,{38,103,0}},
/* 63641 */ {(14<<2)|2,{38,103,0}},
/* 63642 */ {(14<<2)|2,{38,103,0}},
/* 63643 */ {(14<<2)|2,{38,103,0}},
/* 63644 */ {(14<<2)|2,{38,104,0}},
/* 63645 */ {(14<<2)|2,{38,104,0}},
/* 63646 */ {(14<<2)|2,{38,104,0}},
/* 63647 */ {(14<<2)|2,{38,104,0}},
/* 63648 */ {(14<<2)|2,{38,108,0}},
/* 63649 */ {(14<<2)|2,{38,108,0}},
/* 63650 */ {(14<<2)|2,{38,108,0}},
/* 63651 */ {(14<<2)|2,{38,108,0}},
/* 63652 */ {(14<<2)|2,{38,109,0}},
/* 63653 */ {(14<<2)|2,{38,109,0}},
/* 63654 */ {(14<<2)|2,{38,109,0}},
/* 63655 */ {(14<<2)|2,{38,109,0}},
/* 63656 */ {(14<<2)|2,{38,110,0}},
/* 63657 */ {(14<<2)|2,{38,110,0}},
/* 63658 */ {(14<<2)|2,{38,110,0}},
/* 63659 */ {(14<<2)|2,{38,110,0}},
/* 63660 */ {(14<<2)|2,{38,112,0}},
/* 63661 */ {(14<<2)|2,{38,112,0}},
/* 63662 */ {(14<<2)|2,{38,112,0}},
/* 63663 */ {(14<<2)|2,{38,112,0}},
/* 63664 */ {(14<<2)|2,{38,114,0}},
/* 63665 */ {(14<<2)|2,{38,114,0}},
/* 63666 */ {(14<<2)|2,{38,114,0}},
/* 63667 */ {(14<<2)|2,{38,114,0}},
/* 63668 */ {(14<<2)|2,{38,117,0}},
/* 63669 */ {(14<<2)|2,{38,117,0}},
/* 63670 */ {(14<<2)|2,{38,117,0}},
/* 63671 */ {(14<<2)|2,{38,117,0}},
/* 63672 */ {(15<<2)|2,{38,58,0}},
/* 63673 */ {(15<<2)|2,{38,58,0}},
/* 63674 */ {(15<<2)|2,{38,66,0}},
/* 63675 */ {(15<<2)|2,{38,66,0}},
/* 63676 */ {(15<<2)|2,{38,67,0}},
/* 63677 */ {(15<<2)|2,{38,67,0}},
/* 63678 */ {(15<<2)|2,{38,68,0}},
/* 63679 */ {(15<<2)|2,{38,68,0}},
/* 63680 */ {(15<<2)|2,{38,69,0}},
/* 63681 */ {(15<<2)|2,{38,69,0}},
/* 63682 */ {(15<<2)|2,{38,70,0}},
/* 63683 */ {(15<<2)|2,{38,70,0}},
/* 63684 */ {(15<<2)|2,{38,71,0}},
/* 63685 */ {(15<<2)|2,{38,71,0}},
/* 63686 */ {(15<<2)|2,{38,72,0}},
/* 63687 */ {(15<<2)|2,{38,72,0}},
/* 63688 */ {(15<<2)|2,{38,73,0}},
/* 63689 */ {(15<<2)|2,{38,73,0}},
/* 63690 */ {(15<<2)|2,{38,74,0}},
/* 63691 */ {(15<<2)|2,{38,74,0}},
/* 63692 */ {(15<<2)|2,{38,75,0}},
/* 63693 */ {(15<<2)|2,{38,75,0}},
/* 63694 */ {(15<<2)|2,{38,76,0}},
/* 63695 */ {(15<<2)|2,{38,76,0}},
/* 63696 */ {(15<<2)|2,{38,77,0}},
/* 63697 */ {(15<<2)|2,{38,77,0}},
/* 63698 */ {(15<<2)|2,{38,78,0}},
/* 63699 */ {(15<<2)|2,{38,78,0}},
/* 63700 */ {(15<<2)|2,{38,79,0}},
/* 63701 */ {(15<<2)|2,{38,79,0}},
/* 63702 */ {(15<<2)|2,{38,80,0}},
/* 63703 */ {(15<<2)|2,{38,80,0}},
/* 63704 */ {(15<<2)|2,{38,81,0}},
/* 63705 */ {(15<<2)|2,{38,81,0}},
/* 63706 */ {(15<<2)|2,{38,82,0}},
/* 63707 */ {(15<<2)|2,{38,82,0}},
/* 63708 */ {(15<<2)|2,{38,83,0}},
/* 63709 */ {(15<<2)|2,{38,83,0}},
/* 63710 */ {(15<<2)|2,{38,84,0}},
/* 63711 */ {(15<<2)|2,{38,84,0}},
/* 63712 */ {(15<<2)|2,{38,85,0}},
/* 63713 */ {(15<<2)|2,{38,85,0}},
/* 63714 */ {(15<<2)|2,{38,86,0}},
/* 63715 */ {(15<<2)|2,{38,86,0}},
/* 63716 */ {(15<<2)|2,{38,87,0}},
/* 63717 */ {(15<<2)|2,{38,87,0}},
/* 63718 */ {(15<<2)|2,{38,89,0}},
/* 63719 */ {(15<<2)|2,{38,89,0}},
/* 63720 */ {(15<<2)|2,{38,106,0}},
/* 63721 */ {(15<<2)|2,{38,106,0}},
/* 63722 */ {(15<<2)|2,{38,107,0}},
/* 63723 */ {(15<<2)|2,{38,107,0}},
/* 63724 */ {(15<<2)|2,{38,113,0}},
/* 63725 */ {(15<<2)|2,{38,113,0}},
/* 63726 */ {(15<<2)|2,{38,118,0}},
/* 63727 */ {(15<<2)|2,{38,118,0}},
/* 63728 */ {(15<<2)|2,{38,119,0}},
/* 63729 */ {(15<<2)|2,{38,119,0}},
/* 63730 */ {(15<<2)|2,{38,120,0}},
/* 63731 */ {(15<<2)|2,{38,120,0}},
/* 63732 */ {(15<<2)|2,{38,121,0}},
/* 63733 */ {(15<<2)|2,{38,121,0}},
/* 63734 */ {(15<<2)|2,{38,122,0}},
/* 63735 */ {(15<<2)|2,{38,122,0}},
/* 63736 */ {(16<<2)|2,{38,38,0}},
/* 63737 */ {(16<<2)|2,{38,42,0}},
/* 63738 */ {(16<<2)|2,{38,44,0}},
/* 63739 */ {(16<<2)|2,{38,59,0}},
/* 63740 */ {(16<<2)|2,{38,88,0}},
/* 63741 */ {(16<<2)|2,{38,90,0}},
/* 63742 */ {(8<<2)|1,{38,0,0}},
/* 63743 */ {(8<<2)|1,{38,0,0}},
/* 63744 */ {(13<<2)|2,{42,48,0}},
/* 63745 */ {(13<<2)|2,{42,48,0}},
/* 63746 */ {(13<<2)|2,{42,48,0}},
/* 63747 */ {(13<<2)|2,{42,48,0}},
/* 63748 */ {(13<<2)|2,{42,48,0}},
/* 63749 */ {(13<<2)|2,{42,48,0}},
/* 63750 */ {(13<<2)|2,{42,48,0}},
/* 63751 */ {(13<<2)|2,{42,48,0}},
/* 63752 */ {(13<<2)|2,{42,49,0}},
/* 63753 */ {(13<<2)|2,{42,49,0}},
/* 63754 */ {(13<<2)|2,{42,49,0}},
/* 63755 */ {(13<<2)|2,{42,49,0}},
/* 63756 */ {(13<<2)|2,{42,49,0}},
/* 63757 */ {(13<<2)|2,{42,49,0}},
/* 63758 */ {(13<<2)|2,{42,49,0}},
/* 63759 */ {(13<<2)|2,{42,49,0}},
/* 63760 */ {(13<<2)|2,{42,50,0}},
/* 63761 */ {(13<<2)|2,{42,50,0}},
/* 63762 */ {(13<<2)|2,{42,50,0}},
/* 63763 */ {(13<<2)|2,{42,50,0}},
/* 63764 */ {(13<<2)|2,{42,50,0}},
/* 63765 */ {(13<<2)|2,{42,50,0}},
/* 63766 */ {(13<<2)|2,{42,50,0}},
/* 63767 */ {(13<<2)|2,{42,50,0}},
/* 63768 */ {(13<<2)|2,{42,97,0}},
/* 63769 */ {(13<<2)|2,{42,97,0}},
/* 63770 */ {(13<<2)|2,{42,97,0}},
/* 63771 */ {(13<<2)|2,{42,97,0}},
/* 63772 */ {(13<<2)|2,{42,97,0}},
/* 63773 */ {(13<<2)|2,{42,97,0}},
/* 63774 */ {(13<<2)|2,{42,97,0}},
/* 63775 */ {(13<<2)|2,{42,97,0}},
/* 63776 */ {(13<<2)|2,{42,99,0}},
/* 63777 */ {(13<<2)|2,{42,99,0}},
/* 63778 */ {(13<<2)|2,{42,99,0}},
/* 63779 */ {(13<<2)|2,{42,99,0}},
/* 63780 */ {(13<<2)|2,{42,99,0}},
/* 63781 */ {(13<<2)|2,{42,99,0}},
/* 63782 */ {(13<<2)|2,{42,99,0}},
/* 63783 */ {(13<<2)|2,{42,99,0}},
/* 63784 */ {(13<<2)|2,{42,101,0}},
/* 63785 */ {(13<<2)|2,{42,101,0}},
/* 63786 */ {(13<<2)|2,{42,101,0}},
/* 63787 */ {(13<<2)|2,{42,101,0}},
/* 63788 */ {(13<<2)|2,{42,101,0}},
/* 63789 */ {(13<<2)|2,{42,101,0}},
/* 63790 */ {(13<<2)|2,{42,101,0}},
/* 63791 */ {(13<<2)|2,{42,101,0}},
/* 63792 */ {(13<<2)|2,{42,105,0}},
/* 63793 */ {(13<<2)|2,{42,105,0}},
/* 63794 */ {(13<<2)|2,{42,105,0}},
/* 63795 */ {(13<<2)|2,{42,105,0}},
/* 63796 */ {(13<<2)|2,{42,105,0}},
/* 63797 */ {(13<<2)|2,{42,105,0}},
/* 63798 */ {(13<<2)|2,{42,105,0}},
/* 63799 */ {(13<<2)|2,{42,105,0}},
/* 63800 */ {(13<<2)|2,{42,111,0}},
/* 63801 */ {(13<<2)|2,{42,111,0}},
/* 63802 */ {(13<<2)|2,{42,111,0}},
/* 63803 */ {(13<<2)|2,{42,111,0}},
/* 63804 */ {(13<<2)|2,{42,111,0}},
/* 63805 */ {(13<<2)|2,{42,111,0}},
/* 63806 */ {(13<<2)|2,{42,111,0}},
/* 63807 */ {(13<<2)|2,{42,111,0}},
/* 63808 */ {(13<<2)|2,{42,115,0}},
/* 63809 */ {(13<<2)|2,{42,115,0}},
/* 63810 */ {(13<<2)|2,{42,115,0}},
/* 63811 */ {(13<<2)|2,{42,115,0}},
/* 63812 */ {(13<<2)|2,{42,115,0}},
/* 63813 */ {(13<<2)|2,{42,115,0}},
/* 63814 */ {(13<<2)|2,{42,115,0}},
/* 63815 */ {(13<<2)|2,{42,115,0}},
/* 63816 */ {(13<<2)|2,{42,116,0}},
/* 63817 */ {(13<<2)|2,{42,116,0}},
/* 63818 */ {(13<<2)|2,{42,116,0}},
/* 63819 */ {(13<<2)|2,{42,116,0}},
/* 63820 */ {(13<<2)|2,{42,116,0}},
/* 63821 */ {(13<<2)|2,{42,116,0}},
/* 63822 */ {(13<<2)|2,{42,116,0}},
/* 63823 */ {(13<<2)|2,{42,116,0}},
/* 63824 */ {(14<<2)|2,{42,32,0}},
/* 63825 */ {(14<<2)|2,{42,32,0}},
/* 63826 */ {(14<<2)|2,{42,32,0}},
/* 63827 */ {(14<<2)|2,{42,32,0}},
/* 63828 */ {(14<<2)|2,{42,37,0}},
/* 63829 */ {(14<<2)|2,{42,37,0}},
/* 63830 */ {(14<<2)|2,{42,37,0}},
/* 63831 */ {(14<<2)|2,{42,37,0}},
/* 63832 */ {(14<<2)|2,{42,45,0}},
/* 63833 */ {(14<<2)|2,{42,45,0}},
/* 63834 */ {(14<<2)|2,{42,45,0}},
/* 63835 */ {(14<<2)|2,{42,45,0}},
/* 63836 */ {(14<<2)|2,{42,46,0}},
/* 63837 */ {(14<<2)|2,{42,46,0}},
/* 63838 */ {(14<<2)|2,{42,46,0}},
/* 63839 */ {(14<<2)|2,{42,46,0}},
/* 63840 */ {(14<<2)|2,{42,47,0}},
/* 63841 */ {(14<<2)|2,{42,47,0}},
/* 63842 */ {(14<<2)|2,{42,47,0}},
/* 63843 */ {(14<<2)|2,{42,47,0}},
/* 63844 */ {(14<<2)|2,{42,51,0}},
/* 63845 */ {(14<<2)|2,{42,51,0}},
/* 63846 */ {(14<<2)|2,{42,51,0}},
/* 63847 */ {(14<<2)|2,{42,51,0}},
/* 63848 */ {(14<<2)|2,{42,52,0}},
/* 63849 */ {(14<<2)|2,{42,52,0}},
/* 63850 */ {(14<<2)|2,{42,52,0}},
/* 63851 */ {(14<<2)|2,{42,52,0}},
/* 63852 */ {(14<<2)|2,{42,53,0}},
/* 63853 */ {(14<<2)|2,{42,53,0}},
/* 63854 */ {(14<<2)|2,{42,53,0}},
/* 63855 */ {(14<<2)|2,{42,53,0}},
/* 63856 */ {(14<<2)|2,{42,54,0}},
/* 63857 */ {(14<<2)|2,{42,54,0}},
/* 63858 */ {(14<<2)|2,{42,54,0}},
/* 63859 */ {(14<<2)|2,{42,54,0}},
/* 63860 */ {(14<<2)|2,{42,55,0}},
/* 63861 */ {(14<<2)|2,{42,55,0}},
/* 63862 */ {(14<<2)|2,{42,55,0}},
/* 63863 */ {(14<<2)|2,{42,55,0}},
/* 63864 */ {(14<<2)|2,{42,56,0}},
/* 63865 */ {(14<<2)|2,{42,56,0}},
/* 63866 */ {(14<<2)|2,{42,56,0}},
/* 63867 */ {(14<<2)|2,{42,56,0}},
/* 63868 */ {(14<<2)|2,{42,57,0}},
/* 63869 */ {(14<<2)|2,{42,57,0}},
/* 63870 */ {(14<<2)|2,{42,57,0}},
/* 63871 */ {(14<<2)|2,{42,57,0}},
/* 63872 */ {(14<<2)|2,{42,61,0}},
/* 63873 */ {(14<<2)|2,{42,61,0}},
/* 63874 */ {(14<<2)|2,{42,61,0}},
/* 63875 */ {(14<<2)|2,{42,61,0}},
/* 63876 */ {(14<<2)|2,{42,65,0}},
/* 63877 */ {(14<<2)|2,{42,65,0}},
/* 63878 */ {(14<<2)|2,{42,65,0}},
/* 63879 */ {(14<<2)|2,{42,65,0}},
/* 63880 */ {(14<<2)|2,{42,95,0}},
/* 63881 */ {(14<<2)|2,{42,95,0}},
/* 63882 */ {(14<<2)|2,{42,95,0}},
/* 63883 */ {(14<<2)|2,{42,95,0}},
/* 63884 */ {(14<<2)|2,{42,98,0}},
/* 63885 */ {(14<<2)|2,{42,98,0}},
/* 63886 */ {(14<<2)|2,{42,98,0}},
/* 63887 */ {(14<<2)|2,{42,98,0}},
/* 63888 */ {(14<<2)|2,{42,100,0}},
/* 63889 */ {(14<<2)|2,{42,100,0}},
/* 63890 */ {(14<<2)|2,{42,100,0}},
/* 63891 */ {(14<<2)|2,{42,100,0}},
/* 63892 */ {(14<<2)|2,{42,102,0}},
/* 63893 */ {(14<<2)|2,{42,102,0}},
/* 63894 */ {(14<<2)|2,{42,102,0}},
/* 63895 */ {(14<<2)|2,{42,102,0}},
/* 63896 */ {(14<<2)|2,{42,103,0}},
/* 63897 */ {(14<<2)|2,{42,103,0}},
/* 63898 */ {(14<<2)|2,{42,103,0}},
/* 63899 */ {(14<<2)|2,{42,103,0}},
/* 63900 */ {(14<<2)|2,{42,104,0}},
/* 63901 */ {(14<<2)|2,{42,104,0}},
/* 63902 */ {(14<<2)|2,{42,104,0}},
/* 63903 */ {(14<<2)|2,{42,104,0}},
/* 63904 */ {(14<<2)|2,{42,108,0}},
/* 63905 */ {(14<<2)|2,{42,108,0}},
/* 63906 */ {(14<<2)|2,{42,108,0}},
/* 63907 */ {(14<<2)|2,{42,108,0}},
/* 63908 */ {(14<<2)|2,{42,109,0}},
/* 63909 */ {(14<<2)|2,{42,109,0}},
/* 63910 */ {(14<<2)|2,{42,109,0}},
/* 63911 */ {(14<<2)|2,{42,109,0}},
/* 63912 */ {(14<<2)|2,{42,110,0}},
/* 63913 */ {(14<<2)|2,{42,110,0}},
/* 63914 */ {(14<<2)|2,{42,110,0}},
/* 63915 */ {(14<<2)|2,{42,110,0}},
/* 63916 */ {(14<<2)|2,{42,112,0}},
/* 63917 */ {(14<<2)|2,{42,112,0}},
/* 63918 */ {(14<<2)|2,{42,112,0}},
/* 63919 */ {(14<<2)|2,{42,112,0}},
/* 63920 */ {(14<<2)|2,{42,114,0}},
/* 63921 */ {(14<<2)|2,{42,114,0}},
/* 63922 */ {(14<<2)|2,{42,114,0}},
/* 63923 */ {(14<<2)|2,{42,114,0}},
/* 63924 */ {(14<<2)|2,{42,117,0}},
/* 63925 */ {(14<<2)|2,{42,117,0}},
/* 63926 */ {(14<<2)|2,{42,117,0}},
/* 63927 */ {(14<<2)|2,{42,117,0}},
/* 63928 */ {(15<<2)|2,{42,58,0}},
/* 63929 */ {(15<<2)|2,{42,58,0}},
/* 63930 */ {(15<<2)|2,{42,66,0}},
/* 63931 */ {(15<<2)|2,{42,66,0}},
/* 63932 */ {(15<<2)|2,{42,67,0}},
/* 63933 */ {(15<<2)|2,{42,67,0}},
/* 63934 */ {(15<<2)|2,{42,68,0}},
/* 63935 */ {(15<<2)|2,{42,68,0}},
/* 63936 */ {(15<<2)|2,{42,69,0}},
/* 63937 */ {(15<<2)|2,{42,69,0}},
/* 63938 */ {(15<<2)|2,{42,70,0}},
/* 63939 */ {(15<<2)|2,{42,70,0}},
/* 63940 */ {(15<<2)|2,{42,71,0}},
/* 63941 */ {(15<<2)|2,{42,71,0}},
/* 63942 */ {(15<<2)|2,{42,72,0}},
/* 63943 */ {(15<<2)|2,{42,72,0}},
/* 63944 */ {(15<<2)|2,{42,73,0}},
/* 63945 */ {(15<<2)|2,{42,73,0}},
/* 63946 */ {(15<<2)|2,{42,74,0}},
/* 63947 */ {(15<<2)|2,{42,74,0}},
/* 63948 */ {(15<<2)|2,{42,75,0}},
/* 63949 */ {(15<<2)|2,{42,75,0}},
/* 63950 */ {(15<<2)|2,{42,76,0}},
/* 63951 */ {(15<<2)|2,{42,76,0}},
/* 63952 */ {(15<<2)|2,{42,77,0}},
/* 63953 */ {(15<<2)|2,{42,77,0}},
/* 63954 */ {(15<<2)|2,{42,78,0}},
/* 63955 */ {(15<<2)|2,{42,78,0}},
/* 63956 */ {(15<<2)|2,{42,79,0}},
/* 63957 */ {(15<<2)|2,{42,79,0}},
/* 63958 */ {(15<<2)|2,{42,80,0}},
/* 63959 */ {(15<<2)|2,{42,80,0}},
/* 63960 */ {(15<<2)|2,{42,81,0}},
/* 63961 */ {(15<<2)|2,{42,81,0}},
/* 63962 */ {(15<<2)|2,{42,82,0}},
/* 63963 */ {(15<<2)|2,{42,82,0}},
/* 63964 */ {(15<<2)|2,{42,83,0}},
/* 63965 */ {(15<<2)|2,{42,83,0}},
/* 63966 */ {(15<<2)|2,{42,84,0}},
/* 63967 */ {(15<<2)|2,{42,84,0}},
/* 63968 */ {(15<<2)|2,{42,85,0}},
/* 63969 */ {(15<<2)|2,{42,85,0}},
/* 63970 */ {(15<<2)|2,{42,86,0}},
/* 63971 */ {(15<<2)|2,{42,86,0}},
/* 63972 */ {(15<<2)|2,{42,87,0}},
/* 63973 */ {(15<<2)|2,{42,87,0}},
/* 63974 */ {(15<<2)|2,{42,89,0}},
/* 63975 */ {(15<<2)|2,{42,89,0}},
/* 63976 */ {(15<<2)|2,{42,106,0}},
/* 63977 */ {(15<<2)|2,{42,106,0}},
/* 63978 */ {(15<<2)|2,{42,107,0}},
/* 63979 */ {(15<<2)|2,{42,107,0}},
/* 63980 */ {(15<<2)|2,{42,113,0}},
/* 63981 */ {(15<<2)|2,{42,113,0}},
/* 63982 */ {(15<<2)|2,{42,118,0}},
/* 63983 */ {(15<<2)|2,{42,118,0}},
/* 63984 */ {(15<<2)|2,{42,119,0}},
/* 63985 */ {(15<<2)|2,{42,119,0}},
/* 63986 */ {(15<<2)|2,{42,120,0}},
/* 63987 */ {(15<<2)|2,{42,120,0}},
/* 63988 */ {(15<<2)|2,{42,121,0}},
/* 63989 */ {(15<<2)|2,{42,121,0}},
/* 63990 */ {(15<<2)|2,{42,122,0}},
/* 63991 */ {(15<<2)|2,{42,122,0}},
/* 63992 */ {(16<<2)|2,{42,38,0}},
/* 63993 */ {(16<<2)|2,{42,42,0}},
/* 63994 */ {(16<<2)|2,{42,44,0}},
/* 63995 */ {(16<<2)|2,{42,59,0}},
/* 63996 */ {(16<<2)|2,{42,88,0}},
/* 63997 */ {(16<<2)|2,{42,90,0}},
/* 63998 */ {(8<<2)|1,{42,0,0}},
/* 63999 */ {(8<<2)|1,{42,0,0}},
/* 64000 */ {(13<<2)|2,{44,48,0}},
/* 64001 */ {(13<<2)|2,{44,48,0}},
/* 64002 */ {(13<<2)|2,{44,48,0}},
/* 64003 */ {(13<<2)|2,{44,48,0}},
/* 64004 */ {(13<<2)|2,{44,48,0}},
/* 64005 */ {(13<<2)|2,{44,48,0}},
/* 64006 */ {(13<<2)|2,{44,48,0}},
/* 64007 */ {(13<<2)|2,{44,48,0}},
/* 64008 */ {(13<<2)|2,{44,49,0}},
/* 64009 */ {(13<<2)|2,{44,49,0}},
/* 64010 */ {(13<<2)|2,{44,49,0}},
/* 64011 */ {(13<<2)|2,{44,49,0}},
/* 64012 */ {(13<<2)|2,{44,49,0}},
/* 64013 */ {(13<<2)|2,{44,49,0}},
/* 64014 */ {(13<<2)|2,{44,49,0}},
/* 64015 */ {(13<<2)|2,{44,49,0}},
/* 64016 */ {(13<<2)|2,{44,50,0}},
/* 64017 */ {(13<<2)|2,{44,50,0}},
/* 64018 */ {(13<<2)|2,{44,50,0}},
/* 64019 */ {(13<<2)|2,{44,50,0}},
/* 64020 */ {(13<<2)|2,{44,50,0}},
/* 64021 */ {(13<<2)|2,{44,50,0}},
/* 64022 */ {(13<<2)|2,{44,50,0}},
/* 64023 */ {(13<<2)|2,{44,50,0}},
/* 64024 */ {(13<<2)|2,{44,97,0}},
/* 64025 */ {(13<<2)|2,{44,97,0}},
/* 64026 */ {(13<<2)|2,{44,97,0}},
/* 64027 */ {(13<<2)|2,{44,97,0}},
/* 64028 */ {(13<<2)|2,{44,97,0}},
/* 64029 */ {(13<<2)|2,{44,97,0}},
/* 64030 */ {(13<<2)|2,{44,97,0}},
/* 64031 */ {(13<<2)|2,{44,97,0}},
/* 64032 */ {(13<<2)|2,{44,99,0}},
/* 64033 */ {(13<<2)|2,{44,99,0}},
/* 64034 */ {(13<<2)|2,{44,99,0}},
/* 64035 */ {(13<<2)|2,{44,99,0}},
/* 64036 */ {(13<<2)|2,{44,99,0}},
/* 64037 */ {(13<<2)|2,{44,99,0}},
/* 64038 */ {(13<<2)|2,{44,99,0}},
/* 64039 */ {(13<<2)|2,{44,99,0}},
/* 64040 */ {(13<<2)|2,{44,101,0}},
/* 64041 */ {(13<<2)|2,{44,101,0}},
/* 64042 */ {(13<<2)|2,{44,101,0}},
/* 64043 */ {(13<<2)|2,{44,101,0}},
/* 64044 */ {(13<<2)|2,{44,101,0}},
/* 64045 */ {(13<<2)|2,{44,101,0}},
/* 64046 */ {(13<<2)|2,{44,101,0}},
/* 64047 */ {(13<<2)|2,{44,101,0}},
/* 64048 */ {(13<<2)|2,{44,105,0}},
/* 64049 */ {(13<<2)|2,{44,105,0}},
/* 64050 */ {(13<<2)|2,{44,105,0}},
/* 64051 */ {(13<<2)|2,{44,105,0}},
/* 64052 */ {(13<<2)|2,{44,105,0}},
/* 64053 */ {(13<<2)|2,{44,105,0}},
/* 64054 */ {(13<<2)|2,{44,105,0}},
/* 64055 */ {(13<<2)|2,{44,105,0}},
/* 64056 */ {(13<<2)|2,{44,111,0}},
/* 64057 */ {(13<<2)|2,{44,111,0}},
/* 64058 */ {(13<<2)|2,{44,111,0}},
/* 64059 */ {(13<<2)|2,{44,111,0}},
/* 64060 */ {(13<<2)|2,{44,111,0}},
/* 64061 */ {(13<<2)|2,{44,111,0}},
/* 64062 */ {(13<<2)|2,{44,111,0}},
/* 64063 */ {(13<<2)|2,{44,111,0}},
/* 64064 */ {(13<<2)|2,{44,115,0}},
/* 64065 */ {(13<<2)|2,{44,115,0}},
/* 64066 */ {(13<<2)|2,{44,115,0}},
/* 64067 */ {(13<<2)|2,{44,115,0}},
/* 64068 */ {(13<<2)|2,{44,115,0}},
/* 64069 */ {(13<<2)|2,{44,115,0}},
/* 64070 */ {(13<<2)|2,{44,115,0}},
/* 64071 */ {(13<<2)|2,{44,115,0}},
/* 64072 */ {(13<<2)|2,{44,116,0}},
/* 64073 */ {(13<<2)|2,{44,116,0}},
/* 64074 */ {(13<<2)|2,{44,116,0}},
/* 64075 */ {(13<<2)|2,{44,116,0}},
/* 64076 */ {(13<<2)|2,{44,116,0}},
/* 64077 */ {(13<<2)|2,{44,116,0}},
/* 64078 */ {(13<<2)|2,{44,116,0}},
/* 64079 */ {(13<<2)|2,{44,116,0}},
/* 64080 */ {(14<<2)|2,{44,32,0}},
/* 64081 */ {(14<<2)|2,{44,32,0}},
/* 64082 */ {(14<<2)|2,{44,32,0}},
/* 64083 */ {(14<<2)|2,{44,32,0}},
/* 64084 */ {(14<<2)|2,{44,37,0}},
/* 64085 */ {(14<<2)|2,{44,37,0}},
/* 64086 */ {(14<<2)|2,{44,37,0}},
/* 64087 */ {(14<<2)|2,{44,37,0}},
/* 64088 */ {(14<<2)|2,{44,45,0}},
/* 64089 */ {(14<<2)|2,{44,45,0}},
/* 64090 */ {(14<<2)|2,{44,45,0}},
/* 64091 */ {(14<<2)|2,{44,45,0}},
/* 64092 */ {(14<<2)|2,{44,46,0}},
/* 64093 */ {(14<<2)|2,{44,46,0}},
/* 64094 */ {(14<<2)|2,{44,46,0}},
/* 64095 */ {(14<<2)|2,{44,46,0}},
/* 64096 */ {(14<<2)|2,{44,47,0}},
/* 64097 */ {(14<<2)|2,{44,47,0}},
/* 64098 */ {(14<<2)|2,{44,47,0}},
/* 64099 */ {(14<<2)|2,{44,47,0}},
/* 64100 */ {(14<<2)|2,{44,51,0}},
/* 64101 */ {(14<<2)|2,{44,51,0}},
/* 64102 */ {(14<<2)|2,{44,51,0}},
/* 64103 */ {(14<<2)|2,{44,51,0}},
/* 64104 */ {(14<<2)|2,{44,52,0}},
/* 64105 */ {(14<<2)|2,{44,52,0}},
/* 64106 */ {(14<<2)|2,{44,52,0}},
/* 64107 */ {(14<<2)|2,{44,52,0}},
/* 64108 */ {(14<<2)|2,{44,53,0}},
/* 64109 */ {(14<<2)|2,{44,53,0}},
/* 64110 */ {(14<<2)|2,{44,53,0}},
/* 64111 */ {(14<<2)|2,{44,53,0}},
/* 64112 */ {(14<<2)|2,{44,54,0}},
/* 64113 */ {(14<<2)|2,{44,54,0}},
/* 64114 */ {(14<<2)|2,{44,54,0}},
/* 64115 */ {(14<<2)|2,{44,54,0}},
/* 64116 */ {(14<<2)|2,{44,55,0}},
/* 64117 */ {(14<<2)|2,{44,55,0}},
/* 64118 */ {(14<<2)|2,{44,55,0}},
/* 64119 */ {(14<<2)|2,{44,55,0}},
/* 64120 */ {(14<<2)|2,{44,56,0}},
/* 64121 */ {(14<<2)|2,{44,56,0}},
/* 64122 */ {(14<<2)|2,{44,56,0}},
/* 64123 */ {(14<<2)|2,{44,56,0}},
/* 64124 */ {(14<<2)|2,{44,57,0}},
/* 64125 */ {(14<<2)|2,{44,57,0}},
/* 64126 */ {(14<<2)|2,{44,57,0}},
/* 64127 */ {(14<<2)|2,{44,57,0}},
/* 64128 */ {(14<<2)|2,{44,61,0}},
/* 64129 */ {(14<<2)|2,{44,61,0}},
/* 64130 */ {(14<<2)|2,{44,61,0}},
/* 64131 */ {(14<<2)|2,{44,61,0}},
/* 64132 */ {(14<<2)|2,{44,65,0}},
/* 64133 */ {(14<<2)|2,{44,65,0}},
/* 64134 */ {(14<<2)|2,{44,65,0}},
/* 64135 */ {(14<<2)|2,{44,65,0}},
/* 64136 */ {(14<<2)|2,{44,95,0}},
/* 64137 */ {(14<<2)|2,{44,95,0}},
/* 64138 */ {(14<<2)|2,{44,95,0}},
/* 64139 */ {(14<<2)|2,{44,95,0}},
/* 64140 */ {(14<<2)|2,{44,98,0}},
/* 64141 */ {(14<<2)|2,{44,98,0}},
/* 64142 */ {(14<<2)|2,{44,98,0}},
/* 64143 */ {(14<<2)|2,{44,98,0}},
/* 64144 */ {(14<<2)|2,{44,100,0}},
/* 64145 */ {(14<<2)|2,{44,100,0}},
/* 64146 */ {(14<<2)|2,{44,100,0}},
/* 64147 */ {(14<<2)|2,{44,100,0}},
/* 64148 */ {(14<<2)|2,{44,102,0}},
/* 64149 */ {(14<<2)|2,{44,102,0}},
/* 64150 */ {(14<<2)|2,{44,102,0}},
/* 64151 */ {(14<<2)|2,{44,102,0}},
/* 64152 */ {(14<<2)|2,{44,103,0}},
/* 64153 */ {(14<<2)|2,{44,103,0}},
/* 64154 */ {(14<<2)|2,{44,103,0}},
/* 64155 */ {(14<<2)|2,{44,103,0}},
/* 64156 */ {(14<<2)|2,{44,104,0}},
/* 64157 */ {(14<<2)|2,{44,104,0}},
/* 64158 */ {(14<<2)|2,{44,104,0}},
/* 64159 */ {(14<<2)|2,{44,104,0}},
/* 64160 */ {(14<<2)|2,{44,108,0}},
/* 64161 */ {(14<<2)|2,{44,108,0}},
/* 64162 */ {(14<<2)|2,{44,108,0}},
/* 64163 */ {(14<<2)|2,{44,108,0}},
/* 64164 */ {(14<<2)|2,{44,109,0}},
/* 64165 */ {(14<<2)|2,{44,109,0}},
/* 64166 */ {(14<<2)|2,{44,109,0}},
/* 64167 */ {(14<<2)|2,{44,109,0}},
/* 64168 */ {(14<<2)|2,{44,110,0}},
/* 64169 */ {(14<<2)|2,{44,110,0}},
/* 64170 */ {(14<<2)|2,{44,110,0}},
/* 64171 */ {(14<<2)|2,{44,110,0}},
/* 64172 */ {(14<<2)|2,{44,112,0}},
/* 64173 */ {(14<<2)|2,{44,112,0}},
/* 64174 */ {(14<<2)|2,{44,112,0}},
/* 64175 */ {(14<<2)|2,{44,112,0}},
/* 64176 */ {(14<<2)|2,{44,114,0}},
/* 64177 */ {(14<<2)|2,{44,114,0}},
/* 64178 */ {(14<<2)|2,{44,114,0}},
/* 64179 */ {(14<<2)|2,{44,114,0}},
/* 64180 */ {(14<<2)|2,{44,117,0}},
/* 64181 */ {(14<<2)|2,{44,117,0}},
/* 64182 */ {(14<<2)|2,{44,117,0}},
/* 64183 */ {(14<<2)|2,{44,117,0}},
/* 64184 */ {(15<<2)|2,{44,58,0}},
/* 64185 */ {(15<<2)|2,{44,58,0}},
/* 64186 */ {(15<<2)|2,{44,66,0}},
/* 64187 */ {(15<<2)|2,{44,66,0}},
/* 64188 */ {(15<<2)|2,{44,67,0}},
/* 64189 */ {(15<<2)|2,{44,67,0}},
/* 64190 */ {(15<<2)|2,{44,68,0}},
/* 64191 */ {(15<<2)|2,{44,68,0}},
/* 64192 */ {(15<<2)|2,{44,69,0}},
/* 64193 */ {(15<<2)|2,{44,69,0}},
/* 64194 */ {(15<<2)|2,{44,70,0}},
/* 64195 */ {(15<<2)|2,{44,70,0}},
/* 64196 */ {(15<<2)|2,{44,71,0}},
/* 64197 */ {(15<<2)|2,{44,71,0}},
/* 64198 */ {(15<<2)|2,{44,72,0}},
/* 64199 */ {(15<<2)|2,{44,72,0}},
/* 64200 */ {(15<<2)|2,{44,73,0}},
/* 64201 */ {(15<<2)|2,{44,73,0}},
/* 64202 */ {(15<<2)|2,{44,74,0}},
/* 64203 */ {(15<<2)|2,{44,74,0}},
/* 64204 */ {(15<<2)|2,{44,75,0}},
/* 64205 */ {(15<<2)|2,{44,75,0}},
/* 64206 */ {(15<<2)|2,{44,76,0}},
/* 64207 */ {(15<<2)|2,{44,76,0}},
/* 64208 */ {(15<<2)|2,{44,77,0}},
/* 64209 */ {(15<<2)|2,{44,77,0}},
/* 64210 */ {(15<<2)|2,{44,78,0}},
/* 64211 */ {(15<<2)|2,{44,78,0}},
/* 64212 */ {(15<<2)|2,{44,79,0}},
/* 64213 */ {(15<<2)|2,{44,79,0}},
/* 64214 */ {(15<<2)|2,{44,80,0}},
/* 64215 */ {(15<<2)|2,{44,80,0}},
/* 64216 */ {(15<<2)|2,{44,81,0}},
/* 64217 */ {(15<<2)|2,{44,81,0}},
/* 64218 */ {(15<<2)|2,{44,82,0}},
/* 64219 */ {(15<<2)|2,{44,82,0}},
/* 64220 */ {(15<<2)|2,{44,83,0}},
/* 64221 */ {(15<<2)|2,{44,83,0}},
/* 64222 */ {(15<<2)|2,{44,84,0}},
/* 64223 */ {(15<<2)|2,{44,84,0}},
/* 64224 */ {(15<<2)|2,{44,85,0}},
/* 64225 */ {(15<<2)|2,{44,85,0}},
/* 64226 */ {(15<<2)|2,{44,86,0}},
/* 64227 */ {(15<<2)|2,{44,86,0}},
/* 64228 */ {(15<<2)|2,{44,87,0}},
/* 64229 */ {(15<<2)|2,{44,87,0}},
/* 64230 */ {(15<<2)|2,{44,89,0}},
/* 64231 */ {(15<<2)|2,{44,89,0}},
/* 64232 */ {(15<<2)|2,{44,106,0}},
/* 64233 */ {(15<<2)|2,{44,106,0}},
/* 64234 */ {(15<<2)|2,{44,107,0}},
/* 64235 */ {(15<<2)|2,{44,107,0}},
/* 64236 */ {(15<<2)|2,{44,113,0}},
/* 64237 */ {(15<<2)|2,{44,113,0}},
/* 64238 */ {(15<<2)|2,{44,118,0}},
/* 64239 */ {(15<<2)|2,{44,118,0}},
/* 64240 */ {(15<<2)|2,{44,119,0}},
/* 64241 */ {(15<<2)|2,{44,119,0}},
/* 64242 */ {(15<<2)|2,{44,120,0}},
/* 64243 */ {(15<<2)|2,{44,120,0}},
/* 64244 */ {(15<<2)|2,{44,121,0}},
/* 64245 */ {(15<<2)|2,{44,121,0}},
/* 64246 */ {(15<<2)|2,{44,122,0}},
/* 64247 */ {(15<<2)|2,{44,122,0}},
/* 64248 */ {(16<<2)|2,{44,38,0}},
/* 64249 */ {(16<<2)|2,{44,42,0}},
/* 64250 */ {(16<<2)|2,{44,44,0}},
/* 64251 */ {(16<<2)|2,{44,59,0}},
/* 64252 */ {(16<<2)|2,{44,88,0}},
/* 64253 */ {(16<<2)|2,{44,90,0}},
/* 64254 */ {(8<<2)|1,{44,0,0}},
/* 64255 */ {(8<<2)|1,{44,0,0}},
/* 64256 */ {(13<<2)|2,{59,48,0}},
/* 64257 */ {(13<<2)|2,{59,48,0}},
/* 64258 */ {(13<<2)|2,{59,48,0}},
/* 64259 */ {(13<<2)|2,{59,48,0}},
/* 64260 */ {(13<<2)|2,{59,48,0}},
/* 64261 */ {(13<<2)|2,{59,48,0}},
/* 64262 */ {(13<<2)|2,{59,48,0}},
/* 64263 */ {(13<<2)|2,{59,48,0}},
/* 64264 */ {(13<<2)|2,{59,49,0}},
/* 64265 */ {(13<<2)|2,{59,49,0}},
/* 64266 */ {(13<<2)|2,{59,49,0}},
/* 64267 */ {(13<<2)|2,{59,49,0}},
/* 64268 */ {(13<<2)|2,{59,49,0}},
/* 64269 */ {(13<<2)|2,{59,49,0}},
/* 64270 */ {(13<<2)|2,{59,49,0}},
/* 64271 */ {(13<<2)|2,{59,49,0}},
/* 64272 */ {(13<<2)|2,{59,50,0}},
/* 64273 */ {(13<<2)|2,{59,50,0}},
/* 64274 */ {(13<<2)|2,{59,50,0}},
/* 64275 */ {(13<<2)|2,{59,50,0}},
/* 64276 */ {(13<<2)|2,{59,50,0}},
/* 64277 */ {(13<<2)|2,{59,50,0}},
/* 64278 */ {(13<<2)|2,{59,50,0}},
/* 64279 */ {(13<<2)|2,{59,50,0}},
/* 64280 */ {(13<<2)|2,{59,97,0}},
/* 64281 */ {(13<<2)|2,{59,97,0}},
/* 64282 */ {(13<<2)|2,{59,97,0}},
/* 64283 */ {(13<<2)|2,{59,97,0}},
/* 64284 */ {(13<<2)|2,{59,97,0}},
/* 64285 */ {(13<<2)|2,{59,97,0}},
/* 64286 */ {(13<<2)|2,{59,97,0}},
/* 64287 */ {(13<<2)|2,{59,97,0}},
/* 64288 */ {(13<<2)|2,{59,99,0}},
/* 64289 */ {(13<<2)|2,{59,99,0}},
/* 64290 */ {(13<<2)|2,{59,99,0}},
/* 64291 */ {(13<<2)|2,{59,99,0}},
/* 64292 */ {(13<<2)|2,{59,99,0}},
/* 64293 */ {(13<<2)|2,{59,99,0}},
/* 64294 */ {(13<<2)|2,{59,99,0}},
/* 64295 */ {(13<<2)|2,{59,99,0}},
/* 64296 */ {(13<<2)|2,{59,101,0}},
/* 64297 */ {(13<<2)|2,{59,101,0}},
/* 64298 */ {(13<<2)|2,{59,101,0}},
/* 64299 */ {(13<<2)|2,{59,101,0}},
/* 64300 */ {(13<<2)|2,{59,101,0}},
/* 64301 */ {(13<<2)|2,{59,101,0}},
/* 64302 */ {(13<<2)|2,{59,101,0}},
/* 64303 */ {(13<<2)|2,{59,101,0}},
/* 64304 */ {(13<<2)|2,{59,105,0}},
/* 64305 */ {(13<<2)|2,{59,105,0}},
/* 64306 */ {(13<<2)|2,{59,105,0}},
/* 64307 */ {(13<<2)|2,{59,105,0}},
/* 64308 */ {(13<<2)|2,{59,105,0}},
/* 64309 */ {(13<<2)|2,{59,105,0}},
/* 64310 */ {(13<<2)|2,{59,105,0}},
/* 64311 */ {(13<<2)|2,{59,105,0}},
/* 64312 */ {(13<<2)|2,{59,111,0}},
/* 64313 */ {(13<<2)|2,{59,111,0}},
/* 64314 */ {(13<<2)|2,{59,111,0}},
/* 64315 */ {(13<<2)|2,{59,111,0}},
/* 64316 */ {(13<<2)|2,{59,111,0}},
/* 64317 */ {(13<<2)|2,{59,111,0}},
/* 64318 */ {(13<<2)|2,{59,111,0}},
/* 64319 */ {(13<<2)|2,{59,111,0}},
/* 64320 */ {(13<<2)|2,{59,115,0}},
/* 64321 */ {(13<<2)|2,{59,115,0}},
/* 64322 */ {(13<<2)|2,{59,115,0}},
/* 64323 */ {(13<<2)|2,{59,115,0}},
/* 64324 */ {(13<<2)|2,{59,115,0}},
/* 64325 */ {(13<<2)|2,{59,115,0}},
/* 64326 */ {(13<<2)|2,{59,115,0}},
/* 64327 */ {(13<<2)|2,{59,115,0}},
/* 64328 */ {(13<<2)|2,{59,116,0}},
/* 64329 */ {(13<<2)|2,{59,116,0}},
/* 64330 */ {(13<<2)|2,{59,116,0}},
/* 64331 */ {(13<<2)|2,{59,116,0}},
/* 64332 */ {(13<<2)|2,{59,116,0}},
/* 64333 */ {(13<<2)|2,{59,116,0}},
/* 64334 */ {(13<<2)|2,{59,116,0}},
/* 64335 */ {(13<<2)|2,{59,116,0}},
/* 64336 */ {(14<<2)|2,{59,32,0}},
/* 64337 */ {(14<<2)|2,{59,32,0}},
/* 64338 */ {(14<<2)|2,{59,32,0}},
/* 64339 */ {(14<<2)|2,{59,32,0}},
/* 64340 */ {(14<<2)|2,{59,37,0}},
/* 64341 */ {(14<<2)|2,{59,37,0}},
/* 64342 */ {(14<<2)|2,{59,37,0}},
/* 64343 */ {(14<<2)|2,{59,37,0}},
/* 64344 */ {(14<<2)|2,{59,45,0}},
/* 64345 */ {(14<<2)|2,{59,45,0}},
/* 64346 */ {(14<<2)|2,{59,45,0}},
/* 64347 */ {(14<<2)|2,{59,45,0}},
/* 64348 */ {(14<<2)|2,{59,46,0}},
/* 64349 */ {(14<<2)|2,{59,46,0}},
/* 64350 */ {(14<<2)|2,{59,46,0}},
/* 64351 */ {(14<<2)|2,{59,46,0}},
/* 64352 */ {(14<<2)|2,{59,47,0}},
/* 64353 */ {(14<<2)|2,{59,47,0}},
/* 64354 */ {(14<<2)|2,{59,47,0}},
/* 64355 */ {(14<<2)|2,{59,47,0}},
/* 64356 */ {(14<<2)|2,{59,51,0}},
/* 64357 */ {(14<<2)|2,{59,51,0}},
/* 64358 */ {(14<<2)|2,{59,51,0}},
/* 64359 */ {(14<<2)|2,{59,51,0}},
/* 64360 */ {(14<<2)|2,{59,52,0}},
/* 64361 */ {(14<<2)|2,{59,52,0}},
/* 64362 */ {(14<<2)|2,{59,52,0}},
/* 64363 */ {(14<<2)|2,{59,52,0}},
/* 64364 */ {(14<<2)|2,{59,53,0}},
/* 64365 */ {(14<<2)|2,{59,53,0}},
/* 64366 */ {(14<<2)|2,{59,53,0}},
/* 64367 */ {(14<<2)|2,{59,53,0}},
/* 64368 */ {(14<<2)|2,{59,54,0}},
/* 64369 */ {(14<<2)|2,{59,54,0}},
/* 64370 */ {(14<<2)|2,{59,54,0}},
/* 64371 */ {(14<<2)|2,{59,54,0}},
/* 64372 */ {(14<<2)|2,{59,55,0}},
/* 64373 */ {(14<<2)|2,{59,55,0}},
/* 64374 */ {(14<<2)|2,{59,55,0}},
/* 64375 */ {(14<<2)|2,{59,55,0}},
/* 64376 */ {(14<<2)|2,{59,56,0}},
/* 64377 */ {(14<<2)|2,{59,56,0}},
/* 64378 */ {(14<<2)|2,{59,56,0}},
/* 64379 */ {(14<<2)|2,{59,56,0}},
/* 64380 */ {(14<<2)|2,{59,57,0}},
/* 64381 */ {(14<<2)|2,{59,57,0}},
/* 64382 */ {(14<<2)|2,{59,57,0}},
/* 64383 */ {(14<<2)|2,{59,57,0}},
/* 64384 */ {(14<<2)|2,{59,61,0}},
/* 64385 */ {(14<<2)|2,{59,61,0}},
/* 64386 */ {(14<<2)|2,{59,61,0}},
/* 64387 */ {(14<<2)|2,{59,61,0}},
/* 64388 */ {(14<<2)|2,{59,65,0}},
/* 64389 */ {(14<<2)|2,{59,65,0}},
/* 64390 */ {(14<<2)|2,{59,65,0}},
/* 64391 */ {(14<<2)|2,{59,65,0}},
/* 64392 */ {(14<<2)|2,{59,95,0}},
/* 64393 */ {(14<<2)|2,{59,95,0}},
/* 64394 */ {(14<<2)|2,{59,95,0}},
/* 64395 */ {(14<<2)|2,{59,95,0}},
/* 64396 */ {(14<<2)|2,{59,98,0}},
/* 64397 */ {(14<<2)|2,{59,98,0}},
/* 64398 */ {(14<<2)|2,{59,98,0}},
/* 64399 */ {(14<<2)|2,{59,98,0}},
/* 64400 */ {(14<<2)|2,{59,100,0}},
/* 64401 */ {(14<<2)|2,{59,100,0}},
/* 64402 */ {(14<<2)|2,{59,100,0}},
/* 64403 */ {(14<<2)|2,{59,100,0}},
/* 64404 */ {(14<<2)|2,{59,102,0}},
/* 64405 */ {(14<<2)|2,{59,102,0}},
/* 64406 */ {(14<<2)|2,{59,102,0}},
/* 64407 */ {(14<<2)|2,{59,102,0}},
/* 64408 */ {(14<<2)|2,{59,103,0}},
/* 64409 */ {(14<<2)|2,{59,103,0}},
/* 64410 */ {(14<<2)|2,{59,103,0}},
/* 64411 */ {(14<<2)|2,{59,103,0}},
/* 64412 */ {(14<<2)|2,{59,104,0}},
/* 64413 */ {(14<<2)|2,{59,104,0}},
/* 64414 */ {(14<<2)|2,{59,104,0}},
/* 64415 */ {(14<<2)|2,{59,104,0}},
/* 64416 */ {(14<<2)|2,{59,108,0}},
/* 64417 */ {(14<<2)|2,{59,108,0}},
/* 64418 */ {(14<<2)|2,{59,108,0}},
/* 64419 */ {(14<<2)|2,{59,108,0}},
/* 64420 */ {(14<<2)|2,{59,109,0}},
/* 64421 */ {(14<<2)|2,{59,109,0}},
/* 64422 */ {(14<<2)|2,{59,109,0}},
/* 64423 */ {(14<<2)|2,{59,109,0}},
/* 64424 */ {(14<<2)|2,{59,110,0}},
/* 64425 */ {(14<<2)|2,{59,110,0}},
/* 64426 */ {(14<<2)|2,{59,110,0}},
/* 64427 */ {(14<<2)|2,{59,110,0}},
/* 64428 */ {(14<<2)|2,{59,112,0}},
/* 64429 */ {(14<<2)|2,{59,112,0}},
/* 64430 */ {(14<<2)|2,{59,112,0}},
/* 64431 */ {(14<<2)|2,{59,112,0}},
/* 64432 */ {(14<<2)|2,{59,114,0}},
/* 64433 */ {(14<<2)|2,{59,114,0}},
/* 64434 */ {(14<<2)|2,{59,114,0}},
/* 64435 */ {(14<<2)|2,{59,114,0}},
/* 64436 */ {(14<<2)|2,{59,117,0}},
/* 64437 */ {(14<<2)|2,{59,117,0}},
/* 64438 */ {(14<<2)|2,{59,117,0}},
/* 64439 */ {(14<<2)|2,{59,117,0}},
/* 64440 */ {(15<<2)|2,{59,58,0}},
/* 64441 */ {(15<<2)|2,{59,58,0}},
/* 64442 */ {(15<<2)|2,{59,66,0}},
/* 64443 */ {(15<<2)|2,{59,66,0}},
/* 64444 */ {(15<<2)|2,{59,67,0}},
/* 64445 */ {(15<<2)|2,{59,67,0}},
/* 64446 */ {(15<<2)|2,{59,68,0}},
/* 64447 */ {(15<<2)|2,{59,68,0}},
/* 64448 */ {(15<<2)|2,{59,69,0}},
/* 64449 */ {(15<<2)|2,{59,69,0}},
/* 64450 */ {(15<<2)|2,{59,70,0}},
/* 64451 */ {(15<<2)|2,{59,70,0}},
/* 64452 */ {(15<<2)|2,{59,71,0}},
/* 64453 */ {(15<<2)|2,{59,71,0}},
/* 64454 */ {(15<<2)|2,{59,72,0}},
/* 64455 */ {(15<<2)|2,{59,72,0}},
/* 64456 */ {(15<<2)|2,{59,73,0}},
/* 64457 */ {(15<<2)|2,{59,73,0}},
/* 64458 */ {(15<<2)|2,{59,74,0}},
/* 64459 */ {(15<<2)|2,{59,74,0}},
/* 64460 */ {(15<<2)|2,{59,75,0}},
/* 64461 */ {(15<<2)|2,{59,75,0}},
/* 64462 */ {(15<<2)|2,{59,76,0}},
/* 64463 */ {(15<<2)|2,{59,76,0}},
/* 64464 */ {(15<<2)|2,{59,77,0}},
/* 64465 */ {(15<<2)|2,{59,77,0}},
/* 64466 */ {(15<<2)|2,{59,78,0}},
/* 64467 */ {(15<<2)|2,{59,78,0}},
/* 64468 */ {(15<<2)|2,{59,79,0}},
/* 64469 */ {(15<<2)|2,{59,79,0}},
/* 64470 */ {(15<<2)|2,{59,80,0}},
/* 64471 */ {(15<<2)|2,{59,80,0}},
/* 64472 */ {(15<<2)|2,{59,81,0}},
/* 64473 */ {(15<<2)|2,{59,81,0}},
/* 64474 */ {(15<<2)|2,{59,82,0}},
/* 64475 */ {(15<<2)|2,{59,82,0}},
/* 64476 */ {(15<<2)|2,{59,83,0}},
/* 64477 */ {(15<<2)|2,{59,83,0}},
/* 64478 */ {(15<<2)|2,{59,84,0}},
/* 64479 */ {(15<<2)|2,{59,84,0}},
/* 64480 */ {(15<<2)|2,{59,85,0}},
/* 64481 */ {(15<<2)|2,{59,85,0}},
/* 64482 */ {(15<<2)|2,{59,86,0}},
/* 64483 */ {(15<<2)|2,{59,86,0}},
/* 64484 */ {(15<<2)|2,{59,87,0}},
/* 64485 */ {(15<<2)|2,{59,87,0}},
/* 64486 */ {(15<<2)|2,{59,89,0}},
/* 64487 */ {(15<<2)|2,{59,89,0}},
/* 64488 */ {(15<<2)|2,{59,106,0}},
/* 64489 */ {(15<<2)|2,{59,106,0}},
/* 64490 */ {(15<<2)|2,{59,107,0}},
/* 64491 */ {(15<<2)|2,{59,107,0}},
/* 64492 */ {(15<<2)|2,{59,113,0}},
/* 64493 */ {(15<<2)|2,{59,113,0}},
/* 64494 */ {(15<<2)|2,{59,118,0}},
/* 64495 */ {(15<<2)|2,{59,118,0}},
/* 64496 */ {(15<<2)|2,{59,119,0}},
/* 64497 */ {(15<<2)|2,{59,119,0}},
/* 64498 */ {(15<<2)|2,{59,120,0}},
/* 64499 */ {(15<<2)|2,{59,120,0}},
/* 64500 */ {(15<<2)|2,{59,121,0}},
/* 64501 */ {(15<<2)|2,{59,121,0}},
/* 64502 */ {(15<<2)|2,{59,122,0}},
/* 64503 */ {(15<<2)|2,{59,122,0}},
/* 64504 */ {(16<<2)|2,{59,38,0}},
/* 64505 */ {(16<<2)|2,{59,42,0}},
/* 64506 */ {(16<<2)|2,{59,44,0}},
/* 64507 */ {(16<<2)|2,{59,59,0}},
/* 64508 */ {(16<<2)|2,{59,88,0}},
/* 64509 */ {(16<<2)|2,{59,90,0}},
/* 64510 */ {(8<<2)|1,{59,0,0}},
/* 64511 */ {(8<<2)|1,{59,0,0}},
/* 64512 */ {(13<<2)|2,{88,48,0}},
/* 64513 */ {(13<<2)|2,{88,48,0}},
/* 64514 */ {(13<<2)|2,{88,48,0}},
/* 64515 */ {(13<<2)|2,{88,48,0}},
/* 64516 */ {(13<<2)|2,{88,48,0}},
/* 64517 */ {(13<<2)|2,{88,48,0}},
/* 64518 */ {(13<<2)|2,{88,48,0}},
/* 64519 */ {(13<<2)|2,{88,48,0}},
/* 64520 */ {(13<<2)|2,{88,49,0}},
/* 64521 */ {(13<<2)|2,{88,49,0}},
/* 64522 */ {(13<<2)|2,{88,49,0}},
/* 64523 */ {(13<<2)|2,{88,49,0}},
/* 64524 */ {(13<<2)|2,{88,49,0}},
/* 64525 */ {(13<<2)|2,{88,49,0}},
/* 64526 */ {(13<<2)|2,{88,49,0}},
/* 64527 */ {(13<<2)|2,{88,49,0}},
/* 64528 */ {(13<<2)|2,{88,50,0}},
/* 64529 */ {(13<<2)|2,{88,50,0}},
/* 64530 */ {(13<<2)|2,{88,50,0}},
/* 64531 */ {(13<<2)|2,{88,50,0}},
/* 64532 */ {(13<<2)|2,{88,50,0}},
/* 64533 */ {(13<<2)|2,{88,50,0}},
/* 64534 */ {(13<<2)|2,{88,50,0}},
/* 64535 */ {(13<<2)|2,{88,50,0}},
/* 64536 */ {(13<<2)|2,{88,97,0}},
/* 64537 */ {(13<<2)|2,{88,97,0}},
/* 64538 */ {(13<<2)|2,{88,97,0}},
/* 64539 */ {(13<<2)|2,{88,97,0}},
/* 64540 */ {(13<<2)|2,{88,97,0}},
/* 64541 */ {(13<<2)|2,{88,97,0}},
/* 64542 */ {(13<<2)|2,{88,97,0}},
/* 64543 */ {(13<<2)|2,{88,97,0}},
/* 64544 */ {(13<<2)|2,{88,99,0}},
/* 64545 */ {(13<<2)|2,{88,99,0}},
/* 64546 */ {(13<<2)|2,{88,99,0}},
/* 64547 */ {(13<<2)|2,{88,99,0}},
/* 64548 */ {(13<<2)|2,{88,99,0}},
/* 64549 */ {(13<<2)|2,{88,99,0}},
/* 64550 */ {(13<<2)|2,{88,99,0}},
/* 64551 */ {(13<<2)|2,{88,99,0}},
/* 64552 */ {(13<<2)|2,{88,101,0}},
/* 64553 */ {(13<<2)|2,{88,101,0}},
/* 64554 */ {(13<<2)|2,{88,101,0}},
/* 64555 */ {(13<<2)|2,{88,101,0}},
/* 64556 */ {(13<<2)|2,{88,101,0}},
/* 64557 */ {(13<<2)|2,{88,101,0}},
/* 64558 */ {(13<<2)|2,{88,101,0}},
/* 64559 */ {(13<<2)|2,{88,101,0}},
/* 64560 */ {(13<<2)|2,{88,105,0}},
/* 64561 */ {(13<<2)|2,{88,105,0}},
/* 64562 */ {(13<<2)|2,{88,105,0}},
/* 64563 */ {(13<<2)|2,{88,105,0}},
/* 64564 */ {(13<<2)|2,{88,105,0}},
/* 64565 */ {(13<<2)|2,{88,105,0}},
/* 64566 */ {(13<<2)|2,{88,105,0}},
/* 64567 */ {(13<<2)|2,{88,105,0}},
/* 64568 */ {(13<<2)|2,{88,111,0}},
/* 64569 */ {(13<<2)|2,{88,111,0}},
/* 64570 */ {(13<<2)|2,{88,111,0}},
/* 64571 */ {(13<<2)|2,{88,111,0}},
/* 64572 */ {(13<<2)|2,{88,111,0}},
/* 64573 */ {(13<<2)|2,{88,111,0}},
/* 64574 */ {(13<<2)|2,{88,111,0}},
/* 64575 */ {(13<<2)|2,{88,111,0}},
/* 64576 */ {(13<<2)|2,{88,115,0}},
/* 64577 */ {(13<<2)|2,{88,115,0}},
/* 64578 */ {(13<<2)|2,{88,115,0}},
/* 64579 */ {(13<<2)|2,{88,115,0}},
/* 64580 */ {(13<<2)|2,{88,115,0}},
/* 64581 */ {(13<<2)|2,{88,115,0}},
/* 64582 */ {(13<<2)|2,{88,115,0}},
/* 64583 */ {(13<<2)|2,{88,115,0}},
/* 64584 */ {(13<<2)|2,{88,116,0}},
/* 64585 */ {(13<<2)|2,{88,116,0}},
/* 64586 */ {(13<<2)|2,{88,116,0}},
/* 64587 */ {(13<<2)|2,{88,116,0}},
/* 64588 */ {(13<<2)|2,{88,116,0}},
/* 64589 */ {(13<<2)|2,{88,116,0}},
/* 64590 */ {(13<<2)|2,{88,116,0}},
/* 64591 */ {(13<<2)|2,{88,116,0}},
/* 64592 */ {(14<<2)|2,{88,32,0}},
/* 64593 */ {(14<<2)|2,{88,32,0}},
/* 64594 */ {(14<<2)|2,{88,32,0}},
/* 64595 */ {(14<<2)|2,{88,32,0}},
/* 64596 */ {(14<<2)|2,{88,37,0}},
/* 64597 */ {(14<<2)|2,{88,37,0}},
/* 64598 */ {(14<<2)|2,{88,37,0}},
/* 64599 */ {(14<<2)|2,{88,37,0}},
/* 64600 */ {(14<<2)|2,{88,45,0}},
/* 64601 */ {(14<<2)|2,{88,45,0}},
/* 64602 */ {(14<<2)|2,{88,45,0}},
/* 64603 */ {(14<<2)|2,{88,45,0}},
/* 64604 */ {(14<<2)|2,{88,46,0}},
/* 64605 */ {(14<<2)|2,{88,46,0}},
/* 64606 */ {(14<<2)|2,{88,46,0}},
/* 64607 */ {(14<<2)|2,{88,46,0}},
/* 64608 */ {(14<<2)|2,{88,47,0}},
/* 64609 */ {(14<<2)|2,{88,47,0}},
/* 64610 */ {(14<<2)|2,{88,47,0}},
/* 64611 */ {(14<<2)|2,{88,47,0}},
/* 64612 */ {(14<<2)|2,{88,51,0}},
/* 64613 */ {(14<<2)|2,{88,51,0}},
/* 64614 */ {(14<<2)|2,{88,51,0}},
/* 64615 */ {(14<<2)|2,{88,51,0}},
/* 64616 */ {(14<<2)|2,{88,52,0}},
/* 64617 */ {(14<<2)|2,{88,52,0}},
/* 64618 */ {(14<<2)|2,{88,52,0}},
/* 64619 */ {(14<<2)|2,{88,52,0}},
/* 64620 */ {(14<<2)|2,{88,53,0}},
/* 64621 */ {(14<<2)|2,{88,53,0}},
/* 64622 */ {(14<<2)|2,{88,53,0}},
/* 64623 */ {(14<<2)|2,{88,53,0}},
/* 64624 */ {(14<<2)|2,{88,54,0}},
/* 64625 */ {(14<<2)|2,{88,54,0}},
/* 64626 */ {(14<<2)|2,{88,54,0}},
/* 64627 */ {(14<<2)|2,{88,54,0}},
/* 64628 */ {(14<<2)|2,{88,55,0}},
/* 64629 */ {(14<<2)|2,{88,55,0}},
/* 64630 */ {(14<<2)|2,{88,55,0}},
/* 64631 */ {(14<<2)|2,{88,55,0}},
/* 64632 */ {(14<<2)|2,{88,56,0}},
/* 64633 */ {(14<<2)|2,{88,56,0}},
/* 64634 */ {(14<<2)|2,{88,56,0}},
/* 64635 */ {(14<<2)|2,{88,56,0}},
/* 64636 */ {(14<<2)|2,{88,57,0}},
/* 64637 */ {(14<<2)|2,{88,57,0}},
/* 64638 */ {(14<<2)|2,{88,57,0}},
/* 64639 */ {(14<<2)|2,{88,57,0}},
/* 64640 */ {(14<<2)|2,{88,61,0}},
/* 64641 */ {(14<<2)|2,{88,61,0}},
/* 64642 */ {(14<<2)|2,{88,61,0}},
/* 64643 */ {(14<<2)|2,{88,61,0}},
/* 64644 */ {(14<<2)|2,{88,65,0}},
/* 64645 */ {(14<<2)|2,{88,65,0}},
/* 64646 */ {(14<<2)|2,{88,65,0}},
/* 64647 */ {(14<<2)|2,{88,65,0}},
/* 64648 */ {(14<<2)|2,{88,95,0}},
/* 64649 */ {(14<<2)|2,{88,95,0}},
/* 64650 */ {(14<<2)|2,{88,95,0}},
/* 64651 */ {(14<<2)|2,{88,95,0}},
/* 64652 */ {(14<<2)|2,{88,98,0}},
/* 64653 */ {(14<<2)|2,{88,98,0}},
/* 64654 */ {(14<<2)|2,{88,98,0}},
/* 64655 */ {(14<<2)|2,{88,98,0}},
/* 64656 */ {(14<<2)|2,{88,100,0}},
/* 64657 */ {(14<<2)|2,{88,100,0}},
/* 64658 */ {(14<<2)|2,{88,100,0}},
/* 64659 */ {(14<<2)|2,{88,100,0}},
/* 64660 */ {(14<<2)|2,{88,102,0}},
/* 64661 */ {(14<<2)|2,{88,102,0}},
/* 64662 */ {(14<<2)|2,{88,102,0}},
/* 64663 */ {(14<<2)|2,{88,102,0}},
/* 64664 */ {(14<<2)|2,{88,103,0}},
/* 64665 */ {(14<<2)|2,{88,103,0}},
/* 64666 */ {(14<<2)|2,{88,103,0}},
/* 64667 */ {(14<<2)|2,{88,103,0}},
/* 64668 */ {(14<<2)|2,{88,104,0}},
/* 64669 */ {(14<<2)|2,{88,104,0}},
/* 64670 */ {(14<<2)|2,{88,104,0}},
/* 64671 */ {(14<<2)|2,{88,104,0}},
/* 64672 */ {(14<<2)|2,{88,108,0}},
/* 64673 */ {(14<<2)|2,{88,108,0}},
/* 64674 */ {(14<<2)|2,{88,108,0}},
/* 64675 */ {(14<<2)|2,{88,108,0}},
/* 64676 */ {(14<<2)|2,{88,109,0}},
/* 64677 */ {(14<<2)|2,{88,109,0}},
/* 64678 */ {(14<<2)|2,{88,109,0}},
/* 64679 */ {(14<<2)|2,{88,109,0}},
/* 64680 */ {(14<<2)|2,{88,110,0}},
/* 64681 */ {(14<<2)|2,{88,110,0}},
/* 64682 */ {(14<<2)|2,{88,110,0}},
/* 64683 */ {(14<<2)|2,{88,110,0}},
/* 64684 */ {(14<<2)|2,{88,112,0}},
/* 64685 */ {(14<<2)|2,{88,112,0}},
/* 64686 */ {(14<<2)|2,{88,112,0}},
/* 64687 */ {(14<<2)|2,{88,112,0}},
/* 64688 */ {(14<<2)|2,{88,114,0}},
/* 64689 */ {(14<<2)|2,{88,114,0}},
/* 64690 */ {(14<<2)|2,{88,114,0}},
/* 64691 */ {(14<<2)|2,{88,114,0}},
/* 64692 */ {(14<<2)|2,{88,117,0}},
/* 64693 */ {(14<<2)|2,{88,117,0}},
/* 64694 */ {(14<<2)|2,{88,117,0}},
/* 64695 */ {(14<<2)|2,{88,117,0}},
/* 64696 */ {(15<<2)|2,{88,58,0}},
/* 64697 */ {(15<<2)|2,{88,58,0}},
/* 64698 */ {(15<<2)|2,{88,66,0}},
/* 64699 */ {(15<<2)|2,{88,66,0}},
/* 64700 */ {(15<<2)|2,{88,67,0}},
/* 64701 */ {(15<<2)|2,{88,67,0}},
/* 64702 */ {(15<<2)|2,{88,68,0}},
/* 64703 */ {(15<<2)|2,{88,68,0}},
/* 64704 */ {(15<<2)|2,{88,69,0}},
/* 64705 */ {(15<<2)|2,{88,69,0}},
/* 64706 */ {(15<<2)|2,{88,70,0}},
/* 64707 */ {(15<<2)|2,{88,70,0}},
/* 64708 */ {(15<<2)|2,{88,71,0}},
/* 64709 */ {(15<<2)|2,{88,71,0}},
/* 64710 */ {(15<<2)|2,{88,72,0}},
/* 64711 */ {(15<<2)|2,{88,72,0}},
/* 64712 */ {(15<<2)|2,{88,73,0}},
/* 64713 */ {(15<<2)|2,{88,73,0}},
/* 64714 */ {(15<<2)|2,{88,74,0}},
/* 64715 */ {(15<<2)|2,{88,74,0}},
/* 64716 */ {(15<<2)|2,{88,75,0}},
/* 64717 */ {(15<<2)|2,{88,75,0}},
/* 64718 */ {(15<<2)|2,{88,76,0}},
/* 64719 */ {(15<<2)|2,{88,76,0}},
/* 64720 */ {(15<<2)|2,{88,77,0}},
/* 64721 */ {(15<<2)|2,{88,77,0}},
/* 64722 */ {(15<<2)|2,{88,78,0}},
/* 64723 */ {(15<<2)|2,{88,78,0}},
/* 64724 */ {(15<<2)|2,{88,79,0}},
/* 64725 */ {(15<<2)|2,{88,79,0}},
/* 64726 */ {(15<<2)|2,{88,80,0}},
/* 64727 */ {(15<<2)|2,{88,80,0}},
/* 64728 */ {(15<<2)|2,{88,81,0}},
/* 64729 */ {(15<<2)|2,{88,81,0}},
/* 64730 */ {(15<<2)|2,{88,82,0}},
/* 64731 */ {(15<<2)|2,{88,82,0}},
/* 64732 */ {(15<<2)|2,{88,83,0}},
/* 64733 */ {(15<<2)|2,{88,83,0}},
/* 64734 */ {(15<<2)|2,{88,84,0}},
/* 64735 */ {(15<<2)|2,{88,84,0}},
/* 64736 */ {(15<<2)|2,{88,85,0}},
/* 64737 */ {(15<<2)|2,{88,85,0}},
/* 64738 */ {(15<<2)|2,{88,86,0}},
/* 64739 */ {(15<<2)|2,{88,86,0}},
/* 64740 */ {(15<<2)|2,{88,87,0}},
/* 64741 */ {(15<<2)|2,{88,87,0}},
/* 64742 */ {(15<<2)|2,{88,89,0}},
/* 64743 */ {(15<<2)|2,{88,89,0}},
/* 64744 */ {(15<<2)|2,{88,106,0}},
/* 64745 */ {(15<<2)|2,{88,106,0}},
/* 64746 */ {(15<<2)|2,{88,107,0}},
/* 64747 */ {(15<<2)|2,{88,107,0}},
/* 64748 */ {(15<<2)|2,{88,113,0}},
/* 64749 */ {(15<<2)|2,{88,113,0}},
/* 64750 */ {(15<<2)|2,{88,118,0}},
/* 64751 */ {(15<<2)|2,{88,118,0}},
/* 64752 */ {(15<<2)|2,{88,119,0}},
/* 64753 */ {(15<<2)|2,{88,119,0}},
/* 64754 */ {(15<<2)|2,{88,120,0}},
/* 64755 */ {(15<<2)|2,{88,120,0}},
/* 64756 */ {(15<<2)|2,{88,121,0}},
/* 64757 */ {(15<<2)|2,{88,121,0}},
/* 64758 */ {(15<<2)|2,{88,122,0}},
/* 64759 */ {(15<<2)|2,{88,122,0}},
/* 64760 */ {(16<<2)|2,{88,38,0}},
/* 64761 */ {(16<<2)|2,{88,42,0}},
/* 64762 */ {(16<<2)|2,{88,44,0}},
/* 64763 */ {(16<<2)|2,{88,59,0}},
/* 64764 */ {(16<<2)|2,{88,88,0}},
/* 64765 */ {(16<<2)|2,{88,90,0}},
/* 64766 */ {(8<<2)|1,{88,0,0}},
/* 64767 */ {(8<<2)|1,{88,0,0}},
/* 64768 */ {(13<<2)|2,{90,48,0}},
/* 64769 */ {(13<<2)|2,{90,48,0}},
/* 64770 */ {(13<<2)|2,{90,48,0}},
/* 64771 */ {(13<<2)|2,{90,48,0}},
/* 64772 */ {(13<<2)|2,{90,48,0}},
/* 64773 */ {(13<<2)|2,{90,48,0}},
/* 64774 */ {(13<<2)|2,{90,48,0}},
/* 64775 */ {(13<<2)|2,{90,48,0}},
/* 64776 */ {(13<<2)|2,{90,49,0}},
/* 64777 */ {(13<<2)|2,{90,49,0}},
/* 64778 */ {(13<<2)|2,{90,49,0}},
/* 64779 */ {(13<<2)|2,{90,49,0}},
/* 64780 */ {(13<<2)|2,{90,49,0}},
/* 64781 */ {(13<<2)|2,{90,49,0}},
/* 64782 */ {(13<<2)|2,{90,49,0}},
/* 64783 */ {(13<<2)|2,{90,49,0}},
/* 64784 */ {(13<<2)|2,{90,50,0}},
/* 64785 */ {(13<<2)|2,{90,50,0}},
/* 64786 */ {(13<<2)|2,{90,50,0}},
/* 64787 */ {(13<<2)|2,{90,50,0}},
/* 64788 */ {(13<<2)|2,{90,50,0}},
/* 64789 */ {(13<<2)|2,{90,50,0}},
/* 64790 */ {(13<<2)|2,{90,50,0}},
/* 64791 */ {(13<<2)|2,{90,50,0}},
/* 64792 */ {(13<<2)|2,{90,97,0}},
/* 64793 */ {(13<<2)|2,{90,97,0}},
/* 64794 */ {(13<<2)|2,{90,97,0}},
/* 64795 */ {(13<<2)|2,{90,97,0}},
/* 64796 */ {(13<<2)|2,{90,97,0}},
/* 64797 */ {(13<<2)|2,{90,97,0}},
/* 64798 */ {(13<<2)|2,{90,97,0}},
/* 64799 */ {(13<<2)|2,{90,97,0}},
/* 64800 */ {(13<<2)|2,{90,99,0}},
/* 64801 */ {(13<<2)|2,{90,99,0}},
/* 64802 */ {(13<<2)|2,{90,99,0}},
/* 64803 */ {(13<<2)|2,{90,99,0}},
/* 64804 */ {(13<<2)|2,{90,99,0}},
/* 64805 */ {(13<<2)|2,{90,99,0}},
/* 64806 */ {(13<<2)|2,{90,99,0}},
/* 64807 */ {(13<<2)|2,{90,99,0}},
/* 64808 */ {(13<<2)|2,{90,101,0}},
/* 64809 */ {(13<<2)|2,{90,101,0}},
/* 64810 */ {(13<<2)|2,{90,101,0}},
/* 64811 */ {(13<<2)|2,{90,101,0}},
/* 64812 */ {(13<<2)|2,{90,101,0}},
/* 64813 */ {(13<<2)|2,{90,101,0}},
/* 64814 */ {(13<<2)|2,{90,101,0}},
/* 64815 */ {(13<<2)|2,{90,101,0}},
/* 64816 */ {(13<<2)|2,{90,105,0}},
/* 64817 */ {(13<<2)|2,{90,105,0}},
/* 64818 */ {(13<<2)|2,{90,105,0}},
/* 64819 */ {(13<<2)|2,{90,105,0}},
/* 64820 */ {(13<<2)|2,{90,105,0}},
/* 64821 */ {(13<<2)|2,{90,105,0}},
/* 64822 */ {(13<<2)|2,{90,105,0}},
/* 64823 */ {(13<<2)|2,{90,105,0}},
/* 64824 */ {(13<<2)|2,{90,111,0}},
/* 64825 */ {(13<<2)|2,{90,111,0}},
/* 64826 */ {(13<<2)|2,{90,111,0}},
/* 64827 */ {(13<<2)|2,{90,111,0}},
/* 64828 */ {(13<<2)|2,{90,111,0}},
/* 64829 */ {(13<<2)|2,{90,111,0}},
/* 64830 */ {(13<<2)|2,{90,111,0}},
/* 64831 */ {(13<<2)|2,{90,111,0}},
/* 64832 */ {(13<<2)|2,{90,115,0}},
/* 64833 */ {(13<<2)|2,{90,115,0}},
/* 64834 */ {(13<<2)|2,{90,115,0}},
/* 64835 */ {(13<<2)|2,{90,115,0}},
/* 64836 */ {(13<<2)|2,{90,115,0}},
/* 64837 */ {(13<<2)|2,{90,115,0}},
/* 64838 */ {(13<<2)|2,{90,115,0}},
/* 64839 */ {(13<<2)|2,{90,115,0}},
/* 64840 */ {(13<<2)|2,{90,116,0}},
/* 64841 */ {(13<<2)|2,{90,116,0}},
/* 64842 */ {(13<<2)|2,{90,116,0}},
/* 64843 */ {(13<<2)|2,{90,116,0}},
/* 64844 */ {(13<<2)|2,{90,116,0}},
/* 64845 */ {(13<<2)|2,{90,116,0}},
/* 64846 */ {(13<<2)|2,{90,116,0}},
/* 64847 */ {(13<<2)|2,{90,116,0}},
/* 64848 */ {(14<<2)|2,{90,32,0}},
/* 64849 */ {(14<<2)|2,{90,32,0}},
/* 64850 */ {(14<<2)|2,{90,32,0}},
/* 64851 */ {(14<<2)|2,{90,32,0}},
/* 64852 */ {(14<<2)|2,{90,37,0}},
/* 64853 */ {(14<<2)|2,{90,37,0}},
/* 64854 */ {(14<<2)|2,{90,37,0}},
/* 64855 */ {(14<<2)|2,{90,37,0}},
/* 64856 */ {(14<<2)|2,{90,45,0}},
/* 64857 */ {(14<<2)|2,{90,45,0}},
/* 64858 */ {(14<<2)|2,{90,45,0}},
/* 64859 */ {(14<<2)|2,{90,45,0}},
/* 64860 */ {(14<<2)|2,{90,46,0}},
/* 64861 */ {(14<<2)|2,{90,46,0}},
/* 64862 */ {(14<<2)|2,{90,46,0}},
/* 64863 */ {(14<<2)|2,{90,46,0}},
/* 64864 */ {(14<<2)|2,{90,47,0}},
/* 64865 */ {(14<<2)|2,{90,47,0}},
/* 64866 */ {(14<<2)|2,{90,47,0}},
/* 64867 */ {(14<<2)|2,{90,47,0}},
/* 64868 */ {(14<<2)|2,{90,51,0}},
/* 64869 */ {(14<<2)|2,{90,51,0}},
/* 64870 */ {(14<<2)|2,{90,51,0}},
/* 64871 */ {(14<<2)|2,{90,51,0}},
/* 64872 */ {(14<<2)|2,{90,52,0}},
/* 64873 */ {(14<<2)|2,{90,52,0}},
/* 64874 */ {(14<<2)|2,{90,52,0}},
/* 64875 */ {(14<<2)|2,{90,52,0}},
/* 64876 */ {(14<<2)|2,{90,53,0}},
/* 64877 */ {(14<<2)|2,{90,53,0}},
/* 64878 */ {(14<<2)|2,{90,53,0}},
/* 64879 */ {(14<<2)|2,{90,53,0}},
/* 64880 */ {(14<<2)|2,{90,54,0}},
/* 64881 */ {(14<<2)|2,{90,54,0}},
/* 64882 */ {(14<<2)|2,{90,54,0}},
/* 64883 */ {(14<<2)|2,{90,54,0}},
/* 64884 */ {(14<<2)|2,{90,55,0}},
/* 64885 */ {(14<<2)|2,{90,55,0}},
/* 64886 */ {(14<<2)|2,{90,55,0}},
/* 64887 */ {(14<<2)|2,{90,55,0}},
/* 64888 */ {(14<<2)|2,{90,56,0}},
/* 64889 */ {(14<<2)|2,{90,56,0}},
/* 64890 */ {(14<<2)|2,{90,56,0}},
/* 64891 */ {(14<<2)|2,{90,56,0}},
/* 64892 */ {(14<<2)|2,{90,57,0}},
/* 64893 */ {(14<<2)|2,{90,57,0}},
/* 64894 */ {(14<<2)|2,{90,57,0}},
/* 64895 */ {(14<<2)|2,{90,57,0}},
/* 64896 */ {(14<<2)|2,{90,61,0}},
/* 64897 */ {(14<<2)|2,{90,61,0}},
/* 64898 */ {(14<<2)|2,{90,61,0}},
/* 64899 */ {(14<<2)|2,{90,61,0}},
/* 64900 */ {(14<<2)|2,{90,65,0}},
/* 64901 */ {(14<<2)|2,{90,65,0}},
/* 64902 */ {(14<<2)|2,{90,65,0}},
/* 64903 */ {(14<<2)|2,{90,65,0}},
/* 64904 */ {(14<<2)|2,{90,95,0}},
/* 64905 */ {(14<<2)|2,{90,95,0}},
/* 64906 */ {(14<<2)|2,{90,95,0}},
/* 64907 */ {(14<<2)|2,{90,95,0}},
/* 64908 */ {(14<<2)|2,{90,98,0}},
/* 64909 */ {(14<<2)|2,{90,98,0}},
/* 64910 */ {(14<<2)|2,{90,98,0}},
/* 64911 */ {(14<<2)|2,{90,98,0}},
/* 64912 */ {(14<<2)|2,{90,100,0}},
/* 64913 */ {(14<<2)|2,{90,100,0}},
/* 64914 */ {(14<<2)|2,{90,100,0}},
/* 64915 */ {(14<<2)|2,{90,100,0}},
/* 64916 */ {(14<<2)|2,{90,102,0}},
/* 64917 */ {(14<<2)|2,{90,102,0}},
/* 64918 */ {(14<<2)|2,{90,102,0}},
/* 64919 */ {(14<<2)|2,{90,102,0}},
/* 64920 */ {(14<<2)|2,{90,103,0}},
/* 64921 */ {(14<<2)|2,{90,103,0}},
/* 64922 */ {(14<<2)|2,{90,103,0}},
/* 64923 */ {(14<<2)|2,{90,103,0}},
/* 64924 */ {(14<<2)|2,{90,104,0}},
/* 64925 */ {(14<<2)|2,{90,104,0}},
/* 64926 */ {(14<<2)|2,{90,104,0}},
/* 64927 */ {(14<<2)|2,{90,104,0}},
/* 64928 */ {(14<<2)|2,{90,108,0}},
/* 64929 */ {(14<<2)|2,{90,108,0}},
/* 64930 */ {(14<<2)|2,{90,108,0}},
/* 64931 */ {(14<<2)|2,{90,108,0}},
/* 64932 */ {(14<<2)|2,{90,109,0}},
/* 64933 */ {(14<<2)|2,{90,109,0}},
/* 64934 */ {(14<<2)|2,{90,109,0}},
/* 64935 */ {(14<<2)|2,{90,109,0}},
/* 64936 */ {(14<<2)|2,{90,110,0}},
/* 64937 */ {(14<<2)|2,{90,110,0}},
/* 64938 */ {(14<<2)|2,{90,110,0}},
/* 64939 */ {(14<<2)|2,{90,110,0}},
/* 64940 */ {(14<<2)|2,{90,112,0}},
/* 64941 */ {(14<<2)|2,{90,112,0}},
/* 64942 */ {(14<<2)|2,{90,112,0}},
/* 64943 */ {(14<<2)|2,{90,112,0}},
/* 64944 */ {(14<<2)|2,{90,114,0}},
/* 64945 */ {(14<<2)|2,{90,114,0}},
/* 64946 */ {(14<<2)|2,{90,114,0}},
/* 64947 */ {(14<<2)|2,{90,114,0}},
/* 64948 */ {(14<<2)|2,{90,117,0}},
/* 64949 */ {(14<<2)|2,{90,117,0}},
/* 64950 */ {(14<<2)|2,{90,117,0}},
/* 64951 */ {(14<<2)|2,{90,117,0}},
/* 64952 */ {(15<<2)|2,{90,58,0}},
/* 64953 */ {(15<<2)|2,{90,58,0}},
/* 64954 */ {(15<<2)|2,{90,66,0}},
/* 64955 */ {(15<<2)|2,{90,66,0}},
/* 64956 */ {(15<<2)|2,{90,67,0}},
/* 64957 */ {(15<<2)|2,{90,67,0}},
/* 64958 */ {(15<<2)|2,{90,68,0}},
/* 64959 */ {(15<<2)|2,{90,68,0}},
/* 64960 */ {(15<<2)|2,{90,69,0}},
/* 64961 */ {(15<<2)|2,{90,69,0}},
/* 64962 */ {(15<<2)|2,{90,70,0}},
/* 64963 */ {(15<<2)|2,{90,70,0}},
/* 64964 */ {(15<<2)|2,{90,71,0}},
/* 64965 */ {(15<<2)|2,{90,71,0}},
/* 64966 */ {(15<<2)|2,{90,72,0}},
/* 64967 */ {(15<<2)|2,{90,72,0}},
/* 64968 */ {(15<<2)|2,{90,73,0}},
/* 64969 */ {(15<<2)|2,{90,73,0}},
/* 64970 */ {(15<<2)|2,{90,74,0}},
/* 64971 */ {(15<<2)|2,{90,74,0}},
/* 64972 */ {(15<<2)|2,{90,75,0}},
/* 64973 */ {(15<<2)|2,{90,75,0}},
/* 64974 */ {(15<<2)|2,{90,76,0}},
/* 64975 */ {(15<<2)|2,{90,76,0}},
/* 64976 */ {(15<<2)|2,{90,77,0}},
/* 64977 */ {(15<<2)|2,{90,77,0}},
/* 64978 */ {(15<<2)|2,{90,78,0}},
/* 64979 */ {(15<<2)|2,{90,78,0}},
/* 64980 */ {(15<<2)|2,{90,79,0}},
/* 64981 */ {(15<<2)|2,{90,79,0}},
/* 64982 */ {(15<<2)|2,{90,80,0}},
/* 64983 */ {(15<<2)|2,{90,80,0}},
/* 64984 */ {(15<<2)|2,{90,81,0}},
/* 64985 */ {(15<<2)|2,{90,81,0}},
/* 64986 */ {(15<<2)|2,{90,82,0}},
/* 64987 */ {(15<<2)|2,{90,82,0}},
/* 64988 */ {(15<<2)|2,{90,83,0}},
/* 64989 */ {(15<<2)|2,{90,83,0}},
/* 64990 */ {(15<<2)|2,{90,84,0}},
/* 64991 */ {(15<<2)|2,{90,84,0}},
/* 64992 */ {(15<<2)|2,{90,85,0}},
/* 64993 */ {(15<<2)|2,{90,85,0}},
/* 64994 */ {(15<<2)|2,{90,86,0}},
/* 64995 */ {(15<<2)|2,{90,86,0}},
/* 64996 */ {(15<<2)|2,{90,87,0}},
/* 64997 */ {(15<<2)|2,{90,87,0}},
/* 64998 */ {(15<<2)|2,{90,89,0}},
/* 64999 */ {(15<<2)|2,{90,89,0}},
/* 65000 */ {(15<<2)|2,{90,106,0}},
/* 65001 */ {(15<<2)|2,{90,106,0}},
/* 65002 */ {(15<<2)|2,{90,107,0}},
/* 65003 */ {(15<<2)|2,{90,107,0}},
/* 65004 */ {(15<<2)|2,{90,113,0}},
/* 65005 */ {(15<<2)|2,{90,113,0}},
/* 65006 */ {(15<<2)|2,{90,118,0}},
/* 65007 */ {(15<<2)|2,{90,118,0}},
/* 65008 */ {(15<<2)|2,{90,119,0}},
/* 65009 */ {(15<<2)|2,{90,119,0}},
/* 65010 */ {(15<<2)|2,{90,120,0}},
/* 65011 */ {(15<<2)|2,{90,120,0}},
/* 65012 */ {(15<<2)|2,{90,121,0}},
/* 65013 */ {(15<<2)|2,{90,121,0}},
/* 65014 */ {(15<<2)|2,{90,122,0}},
/* 65015 */ {(15<<2)|2,{90,122,0}},
/* 65016 */ {(16<<2)|2,{90,38,0}},
/* 65017 */ {(16<<2)|2,{90,42,0}},
/* 65018 */ {(16<<2)|2,{90,44,0}},
/* 65019 */ {(16<<2)|2,{90,59,0}},
/* 65020 */ {(16<<2)|2,{90,88,0}},
/* 65021 */ {(16<<2)|2,{90,90,0}},
/* 65022 */ {(8<<2)|1,{90,0,0}},
/* 65023 */ {(8<<2)|1,{90,0,0}},
/* 65024 */ {(15<<2)|2,{33,48,0}},
/* 65025 */ {(15<<2)|2,{33,48,0}},
/* 65026 */ {(15<<2)|2,{33,49,0}},
/* 65027 */ {(15<<2)|2,{33,49,0}},
/* 65028 */ {(15<<2)|2,{33,50,0}},
/* 65029 */ {(15<<2)|2,{33,50,0}},
/* 65030 */ {(15<<2)|2,{33,97,0}},
/* 65031 */ {(15<<2)|2,{33,97,0}},
/* 65032 */ {(15<<2)|2,{33,99,0}},
/* 65033 */ {(15<<2)|2,{33,99,0}},
/* 65034 */ {(15<<2)|2,{33,101,0}},
/* 65035 */ {(15<<2)|2,{33,101,0}},
/* 65036 */ {(15<<2)|2,{33,105,0}},
/* 65037 */ {(15<<2)|2,{33,105,0}},
/* 65038 */ {(15<<2)|2,{33,111,0}},
/* 65039 */ {(15<<2)|2,{33,111,0}},
/* 65040 */ {(15<<2)|2,{33,115,0}},
/* 65041 */ {(15<<2)|2,{33,115,0}},
/* 65042 */ {(15<<2)|2,{33,116,0}},
/* 65043 */ {(15<<2)|2,{33,116,0}},
/* 65044 */ {(16<<2)|2,{33,32,0}},
/* 65045 */ {(16<<2)|2,{33,37,0}},
/* 65046 */ {(16<<2)|2,{33,45,0}},
/* 65047 */ {(16<<2)|2,{33,46,0}},
/* 65048 */ {(16<<2)|2,{33,47,0}},
/* 65049 */ {(16<<2)|2,{33,51,0}},
/* 65050 */ {(16<<2)|2,{33,52,0}},
/* 65051 */ {(16<<2)|2,{33,53,0}},
/* 65052 */ {(16<<2)|2,{33,54,0}},
/* 65053 */ {(16<<2)|2,{33,55,0}},
/* 65054 */ {(16<<2)|2,{33,56,0}},
/* 65055 */ {(16<<2)|2,{33,57,0}},
/* 65056 */ {(16<<2)|2,{33,61,0}},
/* 65057 */ {(16<<2)|2,{33,65,0}},
/* 65058 */ {(16<<2)|2,{33,95,0}},
/* 65059 */ {(16<<2)|2,{33,98,0}},
/* 65060 */ {(16<<2)|2,{33,100,0}},
/* 65061 */ {(16<<2)|2,{33,102,0}},
/* 65062 */ {(16<<2)|2,{33,103,0}},
/* 65063 */ {(16<<2)|2,{33,104,0}},
/* 65064 */ {(16<<2)|2,{33,108,0}},
/* 65065 */ {(16<<2)|2,{33,109,0}},
/* 65066 */ {(16<<2)|2,{33,110,0}},
/* 65067 */ {(16<<2)|2,{33,112,0}},
/* 65068 */ {(16<<2)|2,{33,114,0}},
/* 65069 */ {(16<<2)|2,{33,117,0}},
/* 65070 */ {(10<<2)|1,{33,0,0}},
/* 65071 */ {(10<<2)|1,{33,0,0}},
/* 65072 */ {(10<<2)|1,{33,0,0}},
/* 65073 */ {(10<<2)|1,{33,0,0}},
/* 65074 */ {(10<<2)|1,{33,0,0}},
/* 65075 */ {(10<<2)|1,{33,0,0}},
/* 65076 */ {(10<<2)|1,{33,0,0}},
/* 65077 */ {(10<<2)|1,{33,0,0}},
/* 65078 */ {(10<<2)|1,{33,0,0}},
/* 65079 */ {(10<<2)|1,{33,0,0}},
/* 65080 */ {(10<<2)|1,{33,0,0}},
/* 65081 */ {(10<<2)|1,{33,0,0}},
/* 65082 */ {(10<<2)|1,{33,0,0}},
/* 65083 */ {(10<<2)|1,{33,0,0}},
/* 65084 */ {(10<<2)|1,{33,0,0}},
/* 65085 */ {(10<<2)|1,{33,0,0}},
/* 65086 */ {(10<<2)|1,{33,0,0}},
/* 65087 */ {(10<<2)|1,{33,0,0}},
/* 65088 */ {(15<<2)|2,{34,48,0}},
/* 65089 */ {(15<<2)|2,{34,48,0}},
/* 65090 */ {(15<<2)|2,{34,49,0}},
/* 65091 */ {(15<<2)|2,{34,49,0}},
/* 65092 */ {(15<<2)|2,{34,50,0}},
/* 65093 */ {(15<<2)|2,{34,50,0}},
/* 65094 */ {(15<<2)|2,{34,97,0}},
/* 65095 */ {(15<<2)|2,{34,97,0}},
/* 65096 */ {(15<<2)|2,{34,99,0}},
/* 65097 */ {(15<<2)|2,{34,99,0}},
/* 65098 */ {(15<<2)|2,{34,101,0}},
/* 65099 */ {(15<<2)|2,{34,101,0}},
/* 65100 */ {(15<<2)|2,{34,105,0}},
/* 65101 */ {(15<<2)|2,{34,105,0}},
/* 65102 */ {(15<<2)|2,{34,111,0}},
/* 65103 */ {(15<<2)|2,{34,111,0}},
/* 65104 */ {(15<<2)|2,{34,115,0}},
/* 65105 */ {(15<<2)|2,{34,115,0}},
/* 65106 */ {(15<<2)|2,{34,116,0}},
/* 65107 */ {(15<<2)|2,{34,116,0}},
/* 65108 */ {(16<<2)|2,{34,32,0}},
/* 65109 */ {(16<<2)|2,{34,37,0}},
/* 65110 */ {(16<<2)|2,{34,45,0}},
/* 65111 */ {(16<<2)|2,{34,46,0}},
/* 65112 */ {(16<<2)|2,{34,47,0}},
/* 65113 */ {(16<<2)|2,{34,51,0}},
/* 65114 */ {(16<<2)|2,{34,52,0}},
/* 65115 */ {(16<<2)|2,{34,53,0}},
/* 65116 */ {(16<<2)|2,{34,54,0}},
/* 65117 */ {(16<<2)|2,{34,55,0}},
/* 65118 */ {(16<<2)|2,{34,56,0}},
/* 65119 */ {(16<<2)|2,{34,57,0}},
/* 65120 */ {(16<<2)|2,{34,61,0}},
/* 65121 */ {(16<<2)|2,{34,65,0}},
/* 65122 */ {(16<<2)|2,{34,95,0}},
/* 65123 */ {(16<<2)|2,{34,98,0}},
/* 65124 */ {(16<<2)|2,{34,100,0}},
/* 65125 */ {(16<<2)|2,{34,102,0}},
/* 65126 */ {(16<<2)|2,{34,103,0}},
/* 65127 */ {(16<<2)|2,{34,104,0}},
/* 65128 */ {(16<<2)|2,{34,108,0}},
/* 65129 */ {(16<<2)|2,{34,109,0}},
/* 65130 */ {(16<<2)|2,{34,110,0}},
/* 65131 */ {(16<<2)|2,{34,112,0}},
/* 65132 */ {(16<<2)|2,{34,114,0}},
/* 65133 */ {(16<<2)|2,{34,117,0}},
/* 65134 */ {(10<<2)|1,{34,0,0}},
/* 65135 */ {(10<<2)|1,{34,0,0}},
/* 65136 */ {(10<<2)|1,{34,0,0}},
/* 65137 */ {(10<<2)|1,{34,0,0}},
/* 65138 */ {(10<<2)|1,{34,0,0}},
/* 65139 */ {(10<<2)|1,{34,0,0}},
/* 65140 */ {(10<<2)|1,{34,0,0}},
/* 65141 */ {(10<<2)|1,{34,0,0}},
/* 65142 */ {(10<<2)|1,{34,0,0}},
/* 65143 */ {(10<<2)|1,{34,0,0}},
/* 65144 */ {(10<<2)|1,{34,0,0}},
/* 65145 */ {(10<<2)|1,{34,0,0}},
/* 65146 */ {(10<<2)|1,{34,0,0}},
/* 65147 */ {(10<<2)|1,{34,0,0}},
/* 65148 */ {(10<<2)|1,{34,0,0}},
/* 65149 */ {(10<<2)|1,{34,0,0}},
/* 65150 */ {(10<<2)|1,{34,0,0}},
/* 65151 */ {(10<<2)|1,{34,0,0}},
/* 65152 */ {(15<<2)|2,{40,48,0}},
/* 65153 */ {(15<<2)|2,{40,48,0}},
/* 65154 */ {(15<<2)|2,{40,49,0}},
/* 65155 */ {(15<<2)|2,{40,49,0}},
/* 65156 */ {(15<<2)|2,{40,50,0}},
/* 65157 */ {(15<<2)|2,{40,50,0}},
/* 65158 */ {(15<<2)|2,{40,97,0}},
/* 65159 */ {(15<<2)|2,{40,97,0}},
/* 65160 */ {(15<<2)|2,{40,99,0}},
/* 65161 */ {(15<<2)|2,{40,99,0}},
/* 65162 */ {(15<<2)|2,{40,101,0}},
/* 65163 */ {(15<<2)|2,{40,101,0}},
/* 65164 */ {(15<<2)|2,{40,105,0}},
/* 65165 */ {(15<<2)|2,{40,105,0}},
/* 65166 */ {(15<<2)|2,{40,111,0}},
/* 65167 */ {(15<<2)|2,{40,111,0}},
/* 65168 */ {(15<<2)|2,{40,115,0}},
/* 65169 */ {(15<<2)|2,{40,115,0}},
/* 65170 */ {(15<<2)|2,{40,116,0}},
/* 65171 */ {(15<<2)|2,{40,116,0}},
/* 65172 */ {(16<<2)|2,{40,32,0}},
/* 65173 */ {(16<<2)|2,{40,37,0}},
/* 65174 */ {(16<<2)|2,{40,45,0}},
/* 65175 */ {(16<<2)|2,{40,46,0}},
/* 65176 */ {(16<<2)|2,{40,47,0}},
/* 65177 */ {(16<<2)|2,{40,51,0}},
/* 65178 */ {(16<<2)|2,{40,52,0}},
/* 65179 */ {(16<<2)|2,{40,53,0}},
/* 65180 */ {(16<<2)|2,{40,54,0}},
/* 65181 */ {(16<<2)|2,{40,55,0}},
/* 65182 */ {(16<<2)|2,{40,56,0}},
/* 65183 */ {(16<<2)|2,{40,57,0}},
/* 65184 */ {(16<<2)|2,{40,61,0}},
/* 65185 */ {(16<<2)|2,{40,65,0}},
/* 65186 */ {(16<<2)|2,{40,95,0}},
/* 65187 */ {(16<<2)|2,{40,98,0}},
/* 65188 */ {(16<<2)|2,{40,100,0}},
/* 65189 */ {(16<<2)|2,{40,102,0}},
/* 65190 */ {(16<<2)|2,{40,103,0}},
/* 65191 */ {(16<<2)|2,{40,104,0}},
/* 65192 */ {(16<<2)|2,{40,108,0}},
/* 65193 */ {(16<<2)|2,{40,109,0}},
/* 65194 */ {(16<<2)|2,{40,110,0}},
/* 65195 */ {(16<<2)|2,{40,112,0}},
/* 65196 */ {(16<<2)|2,{40,114,0}},
/* 65197 */ {(16<<2)|2,{40,117,0}},
/* 65198 */ {(10<<2)|1,{40,0,0}},
/* 65199 */ {(10<<2)|1,{40,0,0}},
/* 65200 */ {(10<<2)|1,{40,0,0}},
/* 65201 */ {(10<<2)|1,{40,0,0}},
/* 65202 */ {(10<<2)|1,{40,0,0}},
/* 65203 */ {(10<<2)|1,{40,0,0}},
/* 65204 */ {(10<<2)|1,{40,0,0}},
/* 65205 */ {(10<<2)|1,{40,0,0}},
/* 65206 */ {(10<<2)|1,{40,0,0}},
/* 65207 */ {(10<<2)|1,{40,0,0}},
/* 65208 */ {(10<<2)|1,{40,0,0}},
/* 65209 */ {(10<<2)|1,{40,0,0}},
/* 65210 */ {(10<<2)|1,{40,0,0}},
/* 65211 */ {(10<<2)|1,{40,0,0}},
/* 65212 */ {(10<<2)|1,{40,0,0}},
/* 65213 */ {(10<<2)|1,{40,0,0}},
/* 65214 */ {(10<<2)|1,{40,0,0}},
/* 65215 */ {(10<<2)|1,{40,0,0}},
/* 65216 */ {(15<<2)|2,{41,48,0}},
/* 65217 */ {(15<<2)|2,{41,48,0}},
/* 65218 */ {(15<<2)|2,{41,49,0}},
/* 65219 */ {(15<<2)|2,{41,49,0}},
/* 65220 */ {(15<<2)|2,{41,50,0}},
/* 65221 */ {(15<<2)|2,{41,50,0}},
/* 65222 */ {(15<<2)|2,{41,97,0}},
/* 65223 */ {(15<<2)|2,{41,97,0}},
/* 65224 */ {(15<<2)|2,{41,99,0}},
/* 65225 */ {(15<<2)|2,{41,99,0}},
/* 65226 */ {(15<<2)|2,{41,101,0}},
/* 65227 */ {(15<<2)|2,{41,101,0}},
/* 65228 */ {(15<<2)|2,{41,105,0}},
/* 65229 */ {(15<<2)|2,{41,105,0}},
/* 65230 */ {(15<<2)|2,{41,111,0}},
/* 65231 */ {(15<<2)|2,{41,111,0}},
/* 65232 */ {(15<<2)|2,{41,115,0}},
/* 65233 */ {(15<<2)|2,{41,115,0}},
/* 65234 */ {(15<<2)|2,{41,116,0}},
/* 65235 */ {(15<<2)|2,{41,116,0}},
/* 65236 */ {(16<<2)|2,{41,32,0}},
/* 65237 */ {(16<<2)|2,{41,37,0}},
/* 65238 */ {(16<<2)|2,{41,45,0}},
/* 65239 */ {(16<<2)|2,{41,46,0}},
/* 65240 */ {(16<<2)|2,{41,47,0}},
/* 65241 */ {(16<<2)|2,{41,51,0}},
/* 65242 */ {(16<<2)|2,{41,52,0}},
/* 65243 */ {(16<<2)|2,{41,53,0}},
/* 65244 */ {(16<<2)|2,{41,54,0}},
/* 65245 */ {(16<<2)|2,{41,55,0}},
/* 65246 */ {(16<<2)|2,{41,56,0}},
/* 65247 */ {(16<<2)|2,{41,57,0}},
/* 65248 */ {(16<<2)|2,{41,61,0}},
/* 65249 */ {(16<<2)|2,{41,65,0}},
/* 65250 */ {(16<<2)|2,{41,95,0}},
/* 65251 */ {(16<<2)|2,{41,98,0}},
/* 65252 */ {(16<<2)|2,{41,100,0}},
/* 65253 */ {(16<<2)|2,{41,102,0}},
/* 65254 */ {(16<<2)|2,{41,103,0}},
/* 65255 */ {(16<<2)|2,{41,104,0}},
/* 65256 */ {(16<<2)|2,{41,108,0}},
/* 65257 */ {(16<<2)|2,{41,109,0}},
/* 65258 */ {(16<<2)|2,{41,110,0}},
/* 65259 */ {(16<<2)|2,{41,112,0}},
/* 65260 */ {(16<<2)|2,{41,114,0}},
/* 65261 */ {(16<<2)|2,{41,117,0}},
/* 65262 */ {(10<<2)|1,{41,0,0}},
/* 65263 */ {(10<<2)|1,{41,0,0}},
/* 65264 */ {(10<<2)|1,{41,0,0}},
/* 65265 */ {(10<<2)|1,{41,0,0}},
/* 65266 */ {(10<<2)|1,{41,0,0}},
/* 65267 */ {(10<<2)|1,{41,0,0}},
/* 65268 */ {(10<<2)|1,{41,0,0}},
/* 65269 */ {(10<<2)|1,{41,0,0}},
/* 65270 */ {(10<<2)|1,{41,0,0}},
/* 65271 */ {(10<<2)|1,{41,0,0}},
/* 65272 */ {(10<<2)|1,{41,0,0}},
/* 65273 */ {(10<<2)|1,{41,0,0}},
/* 65274 */ {(10<<2)|1,{41,0,0}},
/* 65275 */ {(10<<2)|1,{41,0,0}},
/* 65276 */ {(10<<2)|1,{41,0,0}},
/* 65277 */ {(10<<2)|1,{41,0,0}},
/* 65278 */ {(10<<2)|1,{41,0,0}},
/* 65279 */ {(10<<2)|1,{41,0,0}},
/* 65280 */ {(15<<2)|2,{63,48,0}},
/* 65281 */ {(15<<2)|2,{63,48,0}},
/* 65282 */ {(15<<2)|2,{63,49,0}},
/* 65283 */ {(15<<2)|2,{63,49,0}},
/* 65284 */ {(15<<2)|2,{63,50,0}},
/* 65285 */ {(15<<2)|2,{63,50,0}},
/* 65286 */ {(15<<2)|2,{63,97,0}},
/* 65287 */ {(15<<2)|2,{63,97,0}},
/* 65288 */ {(15<<2)|2,{63,99,0}},
/* 65289 */ {(15<<2)|2,{63,99,0}},
/* 65290 */ {(15<<2)|2,{63,101,0}},
/* 65291 */ {(15<<2)|2,{63,101,0}},
/* 65292 */ {(15<<2)|2,{63,105,0}},
/* 65293 */ {(15<<2)|2,{63,105,0}},
/* 65294 */ {(15<<2)|2,{63,111,0}},
/* 65295 */ {(15<<2)|2,{63,111,0}},
/* 65296 */ {(15<<2)|2,{63,115,0}},
/* 65297 */ {(15<<2)|2,{63,115,0}},
/* 65298 */ {(15<<2)|2,{63,116,0}},
/* 65299 */ {(15<<2)|2,{63,116,0}},
/* 65300 */ {(16<<2)|2,{63,32,0}},
/* 65301 */ {(16<<2)|2,{63,37,0}},
/* 65302 */ {(16<<2)|2,{63,45,0}},
/* 65303 */ {(16<<2)|2,{63,46,0}},
/* 65304 */ {(16<<2)|2,{63,47,0}},
/* 65305 */ {(16<<2)|2,{63,51,0}},
/* 65306 */ {(16<<2)|2,{63,52,0}},
/* 65307 */ {(16<<2)|2,{63,53,0}},
/* 65308 */ {(16<<2)|2,{63,54,0}},
/* 65309 */ {(16<<2)|2,{63,55,0}},
/* 65310 */ {(16<<2)|2,{63,56,0}},
/* 65311 */ {(16<<2)|2,{63,57,0}},
/* 65312 */ {(16<<2)|2,{63,61,0}},
/* 65313 */ {(16<<2)|2,{63,65,0}},
/* 65314 */ {(16<<2)|2,{63,95,0}},
/* 65315 */ {(16<<2)|2,{63,98,0}},
/* 65316 */ {(16<<2)|2,{63,100,0}},
/* 65317 */ {(16<<2)|2,{63,102,0}},
/* 65318 */ {(16<<2)|2,{63,103,0}},
/* 65319 */ {(16<<2)|2,{63,104,0}},
/* 65320 */ {(16<<2)|2,{63,108,0}},
/* 65321 */ {(16<<2)|2,{63,109,0}},
/* 65322 */ {(16<<2)|2,{63,110,0}},
/* 65323 */ {(16<<2)|2,{63,112,0}},
/* 65324 */ {(16<<2)|2,{63,114,0}},
/* 65325 */ {(16<<2)|2,{63,117,0}},
/* 65326 */ {(10<<2)|1,{63,0,0}},
/* 65327 */ {(10<<2)|1,{63,0,0}},
/* 65328 */ {(10<<2)|1,{63,0,0}},
/* 65329 */ {(10<<2)|1,{63,0,0}},
/* 65330 */ {(10<<2)|1,{63,0,0}},
/* 65331 */ {(10<<2)|1,{63,0,0}},
/* 65332 */ {(10<<2)|1,{63,0,0}},
/* 65333 */ {(10<<2)|1,{63,0,0}},
/* 65334 */ {(10<<2)|1,{63,0,0}},
/* 65335 */ {(10<<2)|1,{63,0,0}},
/* 65336 */ {(10<<2)|1,{63,0,0}},
/* 65337 */ {(10<<2)|1,{63,0,0}},
/* 65338 */ {(10<<2)|1,{63,0,0}},
/* 65339 */ {(10<<2)|1,{63,0,0}},
/* 65340 */ {(10<<2)|1,{63,0,0}},
/* 65341 */ {(10<<2)|1,{63,0,0}},
/* 65342 */ {(10<<2)|1,{63,0,0}},
/* 65343 */ {(10<<2)|1,{63,0,0}},
/* 65344 */ {(16<<2)|2,{39,48,0}},
/* 65345 */ {(16<<2)|2,{39,49,0}},
/* 65346 */ {(16<<2)|2,{39,50,0}},
/* 65347 */ {(16<<2)|2,{39,97,0}},
/* 65348 */ {(16<<2)|2,{39,99,0}},
/* 65349 */ {(16<<2)|2,{39,101,0}},
/* 65350 */ {(16<<2)|2,{39,105,0}},
/* 65351 */ {(16<<2)|2,{39,111,0}},
/* 65352 */ {(16<<2)|2,{39,115,0}},
/* 65353 */ {(16<<2)|2,{39,116,0}},
/* 65354 */ {(11<<2)|1,{39,0,0}},
/* 65355 */ {(11<<2)|1,{39,0,0}},
/* 65356 */ {(11<<2)|1,{39,0,0}},
/* 65357 */ {(11<<2)|1,{39,0,0}},
/* 65358 */ {(11<<2)|1,{39,0,0}},
/* 65359 */ {(11<<2)|1,{39,0,0}},
/* 65360 */ {(11<<2)|1,{39,0,0}},
/* 65361 */ {(11<<2)|1,{39,0,0}},
/* 65362 */ {(11<<2)|1,{39,0,0}},
/* 65363 */ {(11<<2)|1,{39,0,0}},
/* 65364 */ {(11<<2)|1,{39,0,0}},
/* 65365 */ {(11<<2)|1,{39,0,0}},
/* 65366 */ {(11<<2)|1,{39,0,0}},
/* 65367 */ {(11<<2)|1,{39,0,0}},
/* 65368 */ {(11<<2)|1,{39,0,0}},
/* 65369 */ {(11<<2)|1,{39,0,0}},
/* 65370 */ {(11<<2)|1,{39,0,0}},
/* 65371 */ {(11<<2)|1,{39,0,0}},
/* 65372 */ {(11<<2)|1,{39,0,0}},
/* 65373 */ {(11<<2)|1,{39,0,0}},
/* 65374 */ {(11<<2)|1,{39,0,0}},
/* 65375 */ {(11<<2)|1,{39,0,0}},
/* 65376 */ {(16<<2)|2,{43,48,0}},
/* 65377 */ {(16<<2)|2,{43,49,0}},
/* 65378 */ {(16<<2)|2,{43,50,0}},
/* 65379 */ {(16<<2)|2,{43,97,0}},
/* 65380 */ {(16<<2)|2,{43,99,0}},
/* 65381 */ {(16<<2)|2,{43,101,0}},
/* 65382 */ {(16<<2)|2,{43,105,0}},
/* 65383 */ {(16<<2)|2,{43,111,0}},
/* 65384 */ {(16<<2)|2,{43,115,0}},
/* 65385 */ {(16<<2)|2,{43,116,0}},
/* 65386 */ {(11<<2)|1,{43,0,0}},
/* 65387 */ {(11<<2)|1,{43,0,0}},
/* 65388 */ {(11<<2)|1,{43,0,0}},
/* 65389 */ {(11<<2)|1,{43,0,0}},
/* 65390 */ {(11<<2)|1,{43,0,0}},
/* 65391 */ {(11<<2)|1,{43,0,0}},
/* 65392 */ {(11<<2)|1,{43,0,0}},
/* 65393 */ {(11<<2)|1,{43,0,0}},
/* 65394 */ {(11<<2)|1,{43,0,0}},
/* 65395 */ {(11<<2)|1,{43,0,0}},
/* 65396 */ {(11<<2)|1,{43,0,0}},
/* 65397 */ {(11<<2)|1,{43,0,0}},
/* 65398 */ {(11<<2)|1,{43,0,0}},
/* 65399 */ {(11<<2)|1,{43,0,0}},
/* 65400 */ {(11<<2)|1,{43,0,0}},
/* 65401 */ {(11<<2)|1,{43,0,0}},
/* 65402 */ {(11<<2)|1,{43,0,0}},
/* 65403 */ {(11<<2)|1,{43,0,0}},
/* 65404 */ {(11<<2)|1,{43,0,0}},
/* 65405 */ {(11<<2)|1,{43,0,0}},
/* 65406 */ {(11<<2)|1,{43,0,0}},
/* 65407 */ {(11<<2)|1,{43,0,0}},
/* 65408 */ {(16<<2)|2,{124,48,0}},
/* 65409 */ {(16<<2)|2,{124,49,0}},
/* 65410 */ {(16<<2)|2,{124,50,0}},
/* 65411 */ {(16<<2)|2,{124,97,0}},
/* 65412 */ {(16<<2)|2,{124,99,0}},
/* 65413 */ {(16<<2)|2,{124,101,0}},
/* 65414 */ {(16<<2)|2,{124,105,0}},
/* 65415 */ {(16<<2)|2,{124,111,0}},
/* 65416 */ {(16<<2)|2,{124,115,0}},
/* 65417 */ {(16<<2)|2,{124,116,0}},
/* 65418 */ {(11<<2)|1,{124,0,0}},
/* 65419 */ {(11<<2)|1,{124,0,0}},
/* 65420 */ {(11<<2)|1,{124,0,0}},
/* 65421 */ {(11<<2)|1,{124,0,0}},
/* 65422 */ {(11<<2)|1,{124,0,0}},
/* 65423 */ {(11<<2)|1,{124,0,0}},
/* 65424 */ {(11<<2)|1,{124,0,0}},
/* 65425 */ {(11<<2)|1,{124,0,0}},
/* 65426 */ {(11<<2)|1,{124,0,0}},
/* 65427 */ {(11<<2)|1,{124,0,0}},
/* 65428 */ {(11<<2)|1,{124,0,0}},
/* 65429 */ {(11<<2)|1,{124,0,0}},
/* 65430 */ {(11<<2)|1,{124,0,0}},
/* 65431 */ {(11<<2)|1,{124,0,0}},
/* 65432 */ {(11<<2)|1,{124,0,0}},
/* 65433 */ {(11<<2)|1,{124,0,0}},
/* 65434 */ {(11<<2)|1,{124,0,0}},
/* 65435 */ {(11<<2)|1,{124,0,0}},
/* 65436 */ {(11<<2)|1,{124,0,0}},
/* 65437 */ {(11<<2)|1,{124,0,0}},
/* 65438 */ {(11<<2)|1,{124,0,0}},
/* 65439 */ {(11<<2)|1,{124,0,0}},
/* 65440 */ {(12<<2)|1,{35,0,0}},
/* 65441 */ {(12<<2)|1,{35,0,0}},
/* 65442 */ {(12<<2)|1,{35,0,0}},
/* 65443 */ {(12<<2)|1,{35,0,0}},
/* 65444 */ {(12<<2)|1,{35,0,0}},
/* 65445 */ {(12<<2)|1,{35,0,0}},
/* 65446 */ {(12<<2)|1,{35,0,0}},
/* 65447 */ {(12<<2)|1,{35,0,0}},
/* 65448 */ {(12<<2)|1,{35,0,0}},
/* 65449 */ {(12<<2)|1,{35,0,0}},
/* 65450 */ {(12<<2)|1,{35,0,0}},
/* 65451 */ {(12<<2)|1,{35,0,0}},
/* 65452 */ {(12<<2)|1,{35,0,0}},
/* 65453 */ {(12<<2)|1,{35,0,0}},
/* 65454 */ {(12<<2)|1,{35,0,0}},
/* 65455 */ {(12<<2)|1,{35,0,0}},
/* 65456 */ {(12<<2)|1,{62,0,0}},
/* 65457 */ {(12<<2)|1,{62,0,0}},
/* 65458 */ {(12<<2)|1,{62,0,0}},
/* 65459 */ {(12<<2)|1,{62,0,0}},
/* 65460 */ {(12<<2)|1,{62,0,0}},
/* 65461 */ {(12<<2)|1,{62,0,0}},
/* 65462 */ {(12<<2)|1,{62,0,0}},
/* 65463 */ {(12<<2)|1,{62,0,0}},
/* 65464 */ {(12<<2)|1,{62,0,0}},
/* 65465 */ {(12<<2)|1,{62,0,0}},
/* 65466 */ {(12<<2)|1,{62,0,0}},
/* 65467 */ {(12<<2)|1,{62,0,0}},
/* 65468 */ {(12<<2)|1,{62,0,0}},
/* 65469 */ {(12<<2)|1,{62,0,0}},
/* 65470 */ {(12<<2)|1,{62,0,0}},
/* 65471 */ {(12<<2)|1,{62,0,0}},
/* 65472 */ {(13<<2)|1,{0,0,0}},
/* 65473 */ {(13<<2)|1,{0,0,0}},
/* 65474 */ {(13<<2)|1,{0,0,0}},
/* 65475 */ {(13<<2)|1,{0,0,0}},
/* 65476 */ {(13<<2)|1,{0,0,0}},
/* 65477 */ {(13<<2)|1,{0,0,0}},
/* 65478 */ {(13<<2)|1,{0,0,0}},
/* 65479 */ {(13<<2)|1,{0,0,0}},
/* 65480 */ {(13<<2)|1,{36,0,0}},
/* 65481 */ {(13<<2)|1,{36,0,0}},
/* 65482 */ {(13<<2)|1,{36,0,0}},
/* 65483 */ {(13<<2)|1,{36,0,0}},
/* 65484 */ {(13<<2)|1,{36,0,0}},
/* 65485 */ {(13<<2)|1,{36,0,0}},
/* 65486 */ {(13<<2)|1,{36,0,0}},
/* 65487 */ {(13<<2)|1,{36,0,0}},
/* 65488 */ {(13<<2)|1,{64,0,0}},
/* 65489 */ {(13<<2)|1,{64,0,0}},
/* 65490 */ {(13<<2)|1,{64,0,0}},
/* 65491 */ {(13<<2)|1,{64,0,0}},
/* 65492 */ {(13<<2)|1,{64,0,0}},
/* 65493 */ {(13<<2)|1,{64,0,0}},
/* 65494 */ {(13<<2)|1,{64,0,0}},
/* 65495 */ {(13<<2)|1,{64,0,0}},
/* 65496 */ {(13<<2)|1,{91,0,0}},
/* 65497 */ {(13<<2)|1,{91,0,0}},
/* 65498 */ {(13<<2)|1,{91,0,0}},
/* 65499 */ {(13<<2)|1,{91,0,0}},
/* 65500 */ {(13<<2)|1,{91,0,0}},
/* 65501 */ {(13<<2)|1,{91,0,0}},
/* 65502 */ {(13<<2)|1,{91,0,0}},
/* 65503 */ {(13<<2)|1,{91,0,0}},
/* 65504 */ {(13<<2)|1,{93,0,0}},
/* 65505 */ {(13<<2)|1,{93,0,0}},
/* 65506 */ {(13<<2)|1,{93,0,0}},
/* 65507 */ {(13<<2)|1,{93,0,0}},
/* 65508 */ {(13<<2)|1,{93,0,0}},
/* 65509 */ {(13<<2)|1,{93,0,0}},
/* 65510 */ {(13<<2)|1,{93,0,0}},
/* 65511 */ {(13<<2)|1,{93,0,0}},
/* 65512 */ {(13<<2)|1,{126,0,0}},
/* 65513 */ {(13<<2)|1,{126,0,0}},
/* 65514 */ {(13<<2)|1,{126,0,0}},
/* 65515 */ {(13<<2)|1,{126,0,0}},
/* 65516 */ {(13<<2)|1,{126,0,0}},
/* 65517 */ {(13<<2)|1,{126,0,0}},
/* 65518 */ {(13<<2)|1,{126,0,0}},
/* 65519 */ {(13<<2)|1,{126,0,0}},
/* 65520 */ {(14<<2)|1,{94,0,0}},
/* 65521 */ {(14<<2)|1,{94,0,0}},
/* 65522 */ {(14<<2)|1,{94,0,0}},
/* 65523 */ {(14<<2)|1,{94,0,0}},
/* 65524 */ {(14<<2)|1,{125,0,0}},
/* 65525 */ {(14<<2)|1,{125,0,0}},
/* 65526 */ {(14<<2)|1,{125,0,0}},
/* 65527 */ {(14<<2)|1,{125,0,0}},
/* 65528 */ {(15<<2)|1,{60,0,0}},
/* 65529 */ {(15<<2)|1,{60,0,0}},
/* 65530 */ {(15<<2)|1,{96,0,0}},
/* 65531 */ {(15<<2)|1,{96,0,0}},
/* 65532 */ {(15<<2)|1,{123,0,0}},
/* 65533 */ {(15<<2)|1,{123,0,0}},
/* 65534 */ {0,{0,0,0,}},
/* 65535 */ {0,{0,0,0,}},
};


#define SHORTEST_CODE 5


/* The decoder is optimized for the common case.  Most of the time, we decode
 * data whose encoding is 16 bits or shorter.  This lets us use a 64 KB table
 * indexed by two bytes of input and outputs 1, 2, or 3 bytes at a time.
 *
 * In the case a longer code is encoutered, we fall back to the original
 * Huffman decoder that supports all code lengths.
 */
int
lshpack_dec_huff_decode (const unsigned char *src, int src_len,
                                            unsigned char *dst, int dst_len)
{
    unsigned char *const orig_dst = dst;
    const unsigned char *const src_end = src + src_len;
    unsigned char *const dst_end = dst + dst_len;
    uintptr_t buf;                  /* OK not to initialize */
    unsigned avail_bits, len;
    struct hdec hdec;
    uint16_t idx;
    int r;

    avail_bits = 0;
    while (1)
    {
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wuninitialized"        
        if (src + sizeof(buf) <= src_end)
        {
            len = (sizeof(buf) * 8 - avail_bits) >> 3;
            avail_bits += len << 3;
            switch (len)
            {
#if UINTPTR_MAX == 18446744073709551615ull
            case 8:
                buf <<= 8;
                buf |= (uintptr_t) *src++;
            case 7:
                buf <<= 8;
                buf |= (uintptr_t) *src++;
            default:
                buf <<= 48;
                buf |= (uintptr_t) *src++ << 40;
                buf |= (uintptr_t) *src++ << 32;
                buf |= (uintptr_t) *src++ << 24;
                buf |= (uintptr_t) *src++ << 16;
#else
            case 4:
                buf <<= 8;
                buf |= (uintptr_t) *src++;
            case 3:
                buf <<= 8;
                buf |= (uintptr_t) *src++;
            default:
                buf <<= 16;
#endif
                buf |= (uintptr_t) *src++ <<  8;
                buf |= (uintptr_t) *src++ <<  0;
            }
        }
        else if (src < src_end)
            do
            {
                buf <<= 8;
                buf |= (uintptr_t) *src++;
                avail_bits += 8;
            }
            while (src < src_end && avail_bits <= sizeof(buf) * 8 - 8);
        else
            break;  /* Normal case terminating condition: out of input */
#pragma GCC diagnostic pop

        if (dst_end - dst >= (ptrdiff_t) (8 * sizeof(buf) / SHORTEST_CODE)
                                                            && avail_bits >= 16)
        {
            /* Fast path: don't check destination bounds */
            do
            {
                idx = buf >> (avail_bits - 16);
                hdec = hdecs[idx];
                dst[0] = hdec.out[0];
                dst[1] = hdec.out[1];
                dst[2] = hdec.out[2];
                dst += hdec.lens & 3;
                avail_bits -= hdec.lens >> 2;
            }
            while (avail_bits >= 16 && hdec.lens);
            if (avail_bits < 16)
                continue;
            goto slow_path;
        }
        else
            while (avail_bits >= 16)
            {
                idx = buf >> (avail_bits - 16);
                hdec = hdecs[idx];
                len = hdec.lens & 3;
                if (len && dst + len <= dst_end)
                {
                    switch (len)
                    {
                    case 3:
                        *dst++ = hdec.out[0];
                        *dst++ = hdec.out[1];
                        *dst++ = hdec.out[2];
                        break;
                    case 2:
                        *dst++ = hdec.out[0];
                        *dst++ = hdec.out[1];
                        break;
                    default:
                        *dst++ = hdec.out[0];
                        break;
                    }
                    avail_bits -= hdec.lens >> 2;
                }
                else if (dst + len > dst_end)
                    return -2;
                else
                    goto slow_path;
            }
    }

    if (avail_bits >= SHORTEST_CODE)
    {
        idx = buf << (16 - avail_bits);
        idx |= (1 << (16 - avail_bits)) - 1;    /* EOF */
        if (idx == 0xFFFF && avail_bits < 8)
            goto end;
        /* If a byte or more of input is left, this mean there is a valid
         * encoding, not just EOF.
         */
        hdec = hdecs[idx];
        len = hdec.lens & 3;
        if ((hdec.lens >> 2) > avail_bits)
            return -1;
        if (len && dst + len <= dst_end)
        {
            switch (len)
            {
            case 3:
                *dst++ = hdec.out[0];
                *dst++ = hdec.out[1];
                *dst++ = hdec.out[2];
                break;
            case 2:
                *dst++ = hdec.out[0];
                *dst++ = hdec.out[1];
                break;
            default:
                *dst++ = hdec.out[0];
                break;
            }
            avail_bits -= hdec.lens >> 2;
        }
        else if (dst + len > dst_end)
            return -2;
        else
            /* This must be an invalid code, otherwise it would have fit */
            return -1;
    }

    if (avail_bits > 0)
    {
        if (((1u << avail_bits) - 1) != (buf & ((1u << avail_bits) - 1)))
            return -1;  /* Not EOF as expected */
    }

  end:
    return dst - orig_dst;

  slow_path:
    /* Find previous byte boundary and finish decoding thence. */
    while ((avail_bits & 7) && dst > orig_dst)
        avail_bits += encode_table[ *--dst ].bits;
    assert((avail_bits & 7) == 0);
    src -= avail_bits >> 3;
    r = lshpack_dec_huff_decode_full(src, src_end - src, dst, dst_end - dst);
    if (r >= 0)
        return dst - orig_dst + r;
    else
        return r;
}
